/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_ADAPTIVE_ESTIMATOR
#define MPEGI_ADAPTIVE_ESTIMATOR

#include <MPEGI/flex_range_coder.h>
#include <algorithm>
#include <cmath>
#include <cstdio>
#include <vector>

namespace MPEGI {

//! Adaptive KT probability estimator (increment 2, offset 1) with frequency rescaling
class AdaptiveEstimator
{
private:
    //! complete binary tree, root node at 1, node n has child nodes 2 * n and 2 * n + 1
    std::vector<unsigned> freqs;
    unsigned symbol_count;
    unsigned total_freq;
    unsigned max_total_freq;
    unsigned leaf_offset; //!< base offset for leaf symbol nodes in the tree

    unsigned traverseUpdateTree(unsigned node);
    void rescaleFreqs();

public:
    void create(unsigned alphabet_size, unsigned max_freq);
    inline void encodeUpdate(unsigned sym, FlexRangeEncoder* fre);
    inline unsigned decodeUpdate(FlexRangeDecoder* frd);
    inline double getLength(unsigned sym);
    void resetFreqs();
};

inline void AdaptiveEstimator::encodeUpdate(unsigned sym, FlexRangeEncoder* fre)
{
    unsigned code = sym + leaf_offset;
    unsigned cum_freq = 0;

    do {
        unsigned lsb = code & 1;
        code >>= 1;
        if (lsb != 0) {
            cum_freq += freqs[code];
        } else {
            freqs[code] += 2;
        }
    } while (code > 1);

    fre_encode(fre, cum_freq, freqs[sym + leaf_offset], total_freq);
    freqs[sym + leaf_offset] += 2;
    total_freq += 2;

    if (total_freq >= max_total_freq) {
        // rescale all freqs by 0.5
        rescaleFreqs();
    }
}

inline double AdaptiveEstimator::getLength(unsigned sym)
{
    return log2((double)total_freq / freqs[sym + leaf_offset]);
}

inline unsigned AdaptiveEstimator::decodeUpdate(FlexRangeDecoder* frd)
{
    unsigned cum_freq = frd_get_cumulative(frd, total_freq);
    unsigned code = 1;
    unsigned sum = 0;

    do {
        if (cum_freq >= freqs[code] + sum) {
            sum += freqs[code];
            code = (code << 1) + 1;
        } else {
            freqs[code] += 2;
            code <<= 1;
        }
    } while (code < leaf_offset);

    frd_update(frd, sum, freqs[code]);
    freqs[code] += 2;
    total_freq += 2;

    if (total_freq >= max_total_freq) {
        // rescale all freqs by 0.5
        rescaleFreqs();
    }

    return code - leaf_offset;
}

inline void AdaptiveEstimator::create(unsigned alphabet_size, unsigned max_freq)
{
    symbol_count = alphabet_size;
    leaf_offset = 1u << ceil_log2(alphabet_size);
    // when alphabet_size is 1, leaf_offset is also 1, which overlaps with the root node at 1
    leaf_offset = std::max(leaf_offset, 2u);
    max_total_freq = max_freq;

    // unused space at 0 and from symbol_count + leaf_offset to 2 * leaf_offset - 1 inclusive
    freqs.assign(2 * leaf_offset, 0);

    resetFreqs();
}

inline void AdaptiveEstimator::resetFreqs()
{
    for (unsigned sym = 0; sym < symbol_count; sym++) {
        freqs[sym + leaf_offset] = 1;
    }
    total_freq = symbol_count;

    traverseUpdateTree(1); // 1 is the root node of the tree
}

inline void AdaptiveEstimator::rescaleFreqs()
{
    total_freq = 0;
    for (unsigned sym = 0; sym < symbol_count; sym++) {
        freqs[sym + leaf_offset] = 1 + ((freqs[sym + leaf_offset] - 1) >> 1);
        total_freq += freqs[sym + leaf_offset];
    }

    traverseUpdateTree(1); // 1 is the root node of the tree
}

inline unsigned AdaptiveEstimator::traverseUpdateTree(unsigned node)
{
    if (node >= leaf_offset)
        return freqs[node];

    unsigned sum_left = traverseUpdateTree(2 * node);
    unsigned sum_right = traverseUpdateTree(2 * node + 1);
    freqs[node] = sum_left;

    return sum_left + sum_right;
}

} // namespace MPEGI

#endif /* MPEGI_ADAPTIVE_ESTIMATOR */
