/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_GEOMETRY_CODING
#define MPEGI_GEOMETRY_CODING

#include "flex_range_coder.h"
#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <memory>
#include <string>
#include <vector>

namespace MPEGI {

class VertexListQuant;
class TriangleList;
class BoundingBoxQuant;

class VertexList
{
public:
    int count = 0;

    std::vector<float> x;
    std::vector<float> y;
    std::vector<float> z;

    void print(FILE* f);

    // internal only members
    int duplicates_removed = 0;
    int exact_duplicates_removed = 0;

    void quantize(VertexListQuant& vlq, float step);
};

class VertexListQuant
{
public:
    int count = 0;
    float step = 0.0f; // invalid value for default initialization

    std::vector<int> x;
    std::vector<int> y;
    std::vector<int> z;

    int duplicates_removed = 0;
    int exact_duplicates_removed = 0;

    std::vector<int> xu;
    std::vector<int> yu;
    std::vector<int> zu;

    void print(FILE* f);
    void compute_unique_coordinates(FILE* f);
    int encode_raw_baseline(FlexRangeEncoder* fre, BoundingBoxQuant& bbq, bool est_only);
    int encode_optimized_v1(FlexRangeEncoder* fre, BoundingBoxQuant& bbq, bool est_only);
    int encode_optimized_v2(FlexRangeEncoder* fre, BoundingBoxQuant& bbq, bool est_only);

    void decode_raw_baseline(FlexRangeDecoder* frd, BoundingBoxQuant& bbq, int vl_count);
    void decode_optimized_v1(FlexRangeDecoder* frd, BoundingBoxQuant& bbq, int vl_quant);

    void dequantize(VertexList& vl);
};

class TriangleList
{
public:
    int count = 0;
    int shared_material_count = 0;

    std::vector<int> v1;
    std::vector<int> v2;
    std::vector<int> v3;
    std::vector<int> mi;

    void print(FILE* f);

    // internal only members
    int duplicates_removed = 0;
    int exact_duplicates_removed = 0;
    int degenerates_removed = 0;
    int exact_degenerates_removed = 0;

    int encode_raw_baseline(FlexRangeEncoder* fre, int vl_count, int ml_count, bool est_only);
    int encode_optimized_v1(FlexRangeEncoder* fre, int vl_count, int ml_count, bool est_only);

    void decode_raw_baseline(FlexRangeDecoder* frd, int vl_count, int tl_count, int ml_count);
    void decode_optimized_v1(FlexRangeDecoder* frd, int vl_count, int tl_count, int ml_count);

    void sort_vertexes();
    void print_double_sorted(FILE* f);
    void sort_triangles(int vl_count);
};

class AcousticMaterial
{
public:
    int freq_count = 0; // invalid value for default initialization

    int freq_grid_index = 0;

    std::vector<float> c1;
    std::vector<float> c2;
    std::vector<float> c3;
    std::vector<float> c4;

    void print(FILE* f);

    // internal only members
    std::string name;

    float encode_raw_baseline(FlexRangeEncoder* fre, int cx_steps, bool est_only);
    float encode_optimized_v1(FlexRangeEncoder* fre, int cx_steps, bool est_only);

    void decode_raw_baseline(FlexRangeDecoder* frd, int cx_steps);
    void decode_optimized_v1(FlexRangeDecoder* frd, int cx_steps);
};

class AcousticMaterialList
{
public:
    int count = 0;

    std::vector<std::unique_ptr<AcousticMaterial>> list;

    void print(FILE* f);

    // internal only members
    float step = 0.0f; // invalid value for default initialization

    int encode_raw_baseline(FlexRangeEncoder* fre, bool est_only);

    void decode_raw_baseline(FlexRangeDecoder* frd, int ml_count);
};

class BoundingBoxQuant
{
public:
    int x_min = 0, x_max = 0;
    int y_min = 0, y_max = 0;
    int z_min = 0, z_max = 0;

    void compute(VertexListQuant& vlq);
    void print(FILE* f);
    int encode_raw_baseline(FlexRangeEncoder* fre);
    int encode_optimized_v1(FlexRangeEncoder* fre);
    int encode_optimized_v2(FlexRangeEncoder* fre);

    int decode_raw_baseline(FlexRangeDecoder* frd);
};

class PreprocessorGCD
{
public:
    int gcd_x = 1;
    int gcd_y = 1;
    int gcd_z = 1;

    void compute(VertexListQuant& vlq, BoundingBoxQuant& bbq);
    void preprocess(VertexListQuant& vlq, BoundingBoxQuant& bbq);
    void print(FILE* f);
    void encode_raw_baseline(FlexRangeEncoder* fre, int bits_needed);
    void encode_optimized_v1(FlexRangeEncoder* fre, BoundingBoxQuant& bbq, int bits_needed);

    void decode_raw_baseline(FlexRangeDecoder* frd, int bits_needed);
    void adjust_bounding_box(BoundingBoxQuant& bbq);
    void reverse_preprocess(VertexListQuant& vlq, BoundingBoxQuant& bbq);
};

//! Used for encoding, decoding, and accesing an entire geometry data (vertexes, triangles, and acoustic materials)
/*!
 * For encoding, see GeometryEncodingTest.py for an example using the C wrapper in GeometryEncodingWrapper.h
 * For decoding, after calling decodeFromBuffer function, direct access to vl, tl, and ml members is available
 */
class GeometryData
{
public:
    //! Add the vertex list, to be called at most once (can be skipped if ver_count is 0)
    /*!
     * Vertex 0 has coordinates x[0], y[0], z[0], measured in meters
     *
     * \param ver_count     number of vertexes (valid range 0 .. 2 ^ 18 - 1 inclusive)
     * \param x             x coordinate values array [m]
     * \param y             y coordinate values array [m]
     * \param z             z coordinate values array [m]
     */
    void addVertexList(int ver_count, float* x, float* y, float* z);

    //! Add the triangle list, to be called at most once (can be skipped if tri_count is 0)
    /*!
     * Triangle 0 has vertexes v1_index[0], v2_index[0], v3_index[0], in order, and material index mat_index[0]
     * Material index -1 is always used to indicate no material, meaning completely transparent acoustically
     *
     * \param tri_count     number of triangles (valid range 0 .. 2 ^ 18 - 1 inclusive)
     * \param v1_index      vertex 1 index coordinate values array (valid range 0 .. ver_count - 1 inclusive)
     * \param v2_index      vertex 2 index coordinate values array (valid range 0 .. ver_count - 1 inclusive)
     * \param v3_index      vertex 3 index coordinate values array (valid range 0 .. ver_count - 1 inclusive)
     * \param mat_index     material index values array (valid range -1 .. shr_mat_count - 1 inclusive)
     * \param shr_mat_count shared total number of acoustic materials (valid range 0 .. 2 ^ 10 - 1 inclusive)
     */
    void addTriangleList(int tri_count, int* v1_index, int* v2_index, int* v3_index,
        int shr_mat_count, int* mat_index);

    //! Add a new acoustic material, to be called exactly once for each of the material_count acoustic materials
    /*!
     * The call can be skipped if mat_count is 0, when only material index -1 is used (indicating no material)
     * Frequency 0 with value freqs[0] has coefficients c1[0], c2[0], c3[0], c4[0]
     *
     * \param mat_count     total number of acoustic materials (valid range 0 .. 2 ^ 10 - 1 inclusive)
     * \param mat_index     index of the acoustic material (valid range 0 .. material_count - 1 inclusive)
     * \param freq_count    number of frequencies in the material specification (valid range 1 to 64 inclusive)
     * \param freqs         frequency values array [Hz] (valid range 1.0 to 24000.0 inclusive)
     * \param c1            first coefficient values array (valid range 0.0 to 1.0 inclusive)
     * \param c2            second coefficient values array (valid range 0.0 to 1.0 inclusive)
     * \param c3            third coefficient values array (valid range 0.0 to 1.0 inclusive)
     * \param c4            fourth coefficient values array (valid range 0.0 to 1.0 inclusive)
     */
    void addAcousticMaterial(int mat_count, int mat_index, int freq_count, int freq_grid_index,
        float* c1, float* c2, float* c3, float* c4);

    //! Set encoder options, like quantization step sizes and tolerances for removal of duplicates and degenerates
    /*!
     * A value of 0.0 for vertex_tol will remove only exact duplicate vertexes (with the same float coordinates)
     * A value of 0.0 for triangle_tol will remove only exact degenerate triangles (with an edge of length 0)
     *
     * \param quant_step    quantization step size [m] for vertex coordinates (valid range 0.0001 to 0.1 inclusive)
     * \param vertex_tol    tolerance for vertex coordinates [m] for duplicates (valid range 0.0 to 0.1 inclusive)
     * \param triangle_tol  tolerance for triangle edges [m] for degenerates (valid range 0.0 to 0.1 inclusive)
     * \param mat_qnt_step  quantization step size for acoustic material coeffs (valid range 0.0001 to 0.1 inclusive)
     * \param tri_rm_nomat  whether to remove triangles with no acoustic material (completely transparent)
     */
    void setEncoderOptions(float quant_step, float vertex_tol, float triangle_tol,
        float mat_qnt_step, bool tri_rm_nomat);

    //! Get the required buffer size that is always large enough to accommodate the entire encoded data
    /*!
     * \return              size in bytes of the required buffer size to be passed as buffer_data
     */
    int getRequiredBufferSize();

    //! Encode all the previously added geometry data to the provided memory buffer
    /*!
     * \param buffer_size   size in bytes of the provided memory buffer
     * \param buffer_data   the provided memory buffer
     * \return              encoded data size in bytes, or its negated value if greater than buffer_size
     */
    int encodeToBuffer(int buffer_size, unsigned char* buffer_data);

    //! Decode the entire geometry data from the provided memory buffer
    /*!
     * \param buffer_size   exact size in bytes of the encoded data contained in the provided memory buffer
     * \param buffer_data   the provided memory buffer
     * \return              whether the decoding was successful or an error occured during decoding
     */
    bool decodeFromBuffer(int buffer_size, unsigned char* buffer_data);

    //! Enable automated text dump of the entire geometry data after calling encodeToBuffer
    /*!
     * The generated file name format is <file_prefix>-<ver_count>v-<tri_count>t-<mat_count>m-<encoded_size>.geo
     * The parameter file_prefix should be ideally of the form <scene_name>-<object_name> for easy identification
     *
     * \param file_prefix   prefix of the generated file name for the automated text dump
     */
    void enableAutomatedTextDump(const char* file_prefix);

    void print(FILE* f);

    VertexList vl;
    TriangleList tl;
    AcousticMaterialList ml;

private:
    void check_and_fix_mesh(float tol_v, float tol_t);

    // internal encoder options
    bool enable_dump = false; //!< whether to enable automated text dump after encoding
    std::string dump_file_prefix = ""; //!< file prefix used for the automated text dump

    float quantization_step = 0.001f; //!< quantization step size for vertex coordinates
    float vertex_tolerance = 0.0f; //!< tolerance for vertex coordinates for removal of duplicates
    float triangle_tolerance = 0.0f; //!< tolerance for triangle edges for removal of degenerates
    //! quantization step size for acoustic material coefficients
    float material_quantization_step = 0.01f;
    //! whether to remove triangles with no acoustic material
    bool triangle_remove_nomaterial = true;

    // internal encoder state
    //! whether the duplicates and degenerates were processed
    bool duplicates_and_degenerates_processed = false;
    //! log file for internal diagnostics (actual output categories enabled by macros)
    FILE* log = stdout;
};

} // namespace MPEGI

#endif /* MPEGI_GEOMETRY_CODING */
