/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef __FLEX_RANGE_CODER_H__
#define __FLEX_RANGE_CODER_H__

#include <stdio.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C"
{
#endif

    int clz_u32(unsigned int n);
    int floor_log2(unsigned int n);
    int ceil_log2(unsigned int n);

    /* callback function to write byte_val using the output_state handle;
       it is the responsibility of output_state to track errors such as
       write, out of disk space, or memory buffer too small errors;
       see fre_write_byte_default for a default implementation with FILE */
    typedef void (*fre_write_byte_type)(void* output_state, unsigned char byte_val);
    /* callback function to read a byte using the input_state handle;
       returns the byte read, -1 for EOF, and -2 for a read error;
       see frd_read_byte_default for a default implementation with FILE */
    typedef int (*frd_read_byte_type)(void* input_state);

    /* State of the range encoder */
    typedef struct
    {
        unsigned int low;
        unsigned int range;
        int pending;
        int has_carry;
        int carry_count;

        int byte_count;
        int last_byte_bit_count;

        fre_write_byte_type write_byte;
        void* output_state;
    } FlexRangeEncoder;

    /* State of the range decoder */
    typedef struct
    {
        unsigned int low;
        unsigned int range;

        int bit_count;
        int max_allowable_bit_count;

        int eof_bytes_read_count;
        int bit_error_detected;

        frd_read_byte_type read_byte;
        void* input_state;
    } FlexRangeDecoder;

    /*
     * After finalizing the range encoder, all the produced bits had been written
     * using the fre->write_byte function. The bits are packed into bytes, MSB first,
     * as byte_array[0].bit[7], ..., byte_array[0].bit[0], byte_array[1].bit[7], ...
     * The last byte may contain less than 8 bits, therefore it can have from 0 to 7
     * inclusive zero padding bits in the LSB part. The total bit count is returned
     * by the fre_finish function but can also be computed afterwards as
     * total_bit_count = (fre->byte_count - 1) * 8 + fre->last_byte_bit_count
     */

    /* Initalize the range encoder */
    void fre_init(FlexRangeEncoder* fre, /* i/o: Encoder state */
        fre_write_byte_type write_byte, void* output_state);

    /* Encode given cumulative count and symbol count */
    void fre_encode(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int cum_count, /* i: Cumulative count up to symbol */
        unsigned int sym_count, /* i: Symbol count */
        unsigned int tot_count /* i: Total count */
    );

    /* Encode given cumulative count and symbol count when total count is a power of 2 */
    void fre_encode_fast(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int cum_count, /* i: Cumulative count up to symbol */
        unsigned int sym_count, /* i: Symbol count */
        unsigned int tot_shift /* i: Total count as a power of 2 */
    );

    /* Encode an alphabet symbol when total count is a power of 2 */
    void fre_encode_symbol_fast(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int symbol, /* i: Symbol to encode */
        const unsigned short table[], /* i: Cumulative counts table */
        unsigned int tot_shift /* i: Total count as a power of 2 */
    );

    /* Encode one bit with probability count0 / 2 ^ tot_shift for symbol 0 */
    void fre_encode_bit_prob_fast(FlexRangeEncoder* fre, /* i/o: Encoder State */
        unsigned int value, /* i: Bit value to encode */
        int count0, /* i: Count for symbol 0 */
        unsigned int tot_shift /* i: Total count as a power of 2 */
    );

    /* Encode up to 16 bits with uniform probability */
    void fre_encode_bits(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int value, /* i: Value to encode */
        unsigned int bits /* i: Number of bits */
    );

    /* Encode up to 32 bits with uniform probability */
    void fre_encode_bits_long(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int value, /* i: Value to encode */
        unsigned int bits /* i: Number of bits */
    );

    /* Encode value from an alphabet of size total with uniform distribution */
    void fre_encode_uniform(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int value, /* i: Value to encode */
        unsigned int total /* i: Alphabet size */
    );

    /* Encode value from an alphabet of size total with uniform distribution */
    void fre_encode_uniform_long(FlexRangeEncoder* fre, /* i/o: Encoder state */
        unsigned int value, /* i: Value to encode */
        unsigned int total /* i: Alphabet size */
    );

    /* Save the state of the range encoder (without output_state position) */
    void fre_save_state(FlexRangeEncoder* fre, /* i: Encoder state */
        FlexRangeEncoder* dst /* o: Saved encoder state */
    );

    /* Restore the state of the range encoder (without output_state position) */
    void fre_restore_state(FlexRangeEncoder* fre, /* o: Encoder state */
        FlexRangeEncoder* src /* i: Saved encoder state */
    );

    /* Get the total number of bits that would be produced by finalization */
    int fre_virtual_finish(/* o: Total number of bits produced */
        FlexRangeEncoder* fre /* i: Encoder state */
    );

    /* Finalize the range encoder */
    int fre_finish(/* o: Total number of bits produced */
        FlexRangeEncoder* fre /* i/o: Encoder state */
    );

    /* Default callback to write byte_val byte to the FILE* output_state */
    void fre_write_byte_default(void* output_state, unsigned char byte_val);

    /*
     * Up to 30 bits are read internally from bit_buffer as look-ahead more than the
     * declared amount of max_available_bits, the total maximum bits available.
     * Therefore, the caller must ensure that bit_buffer has an allocated size of at
     * least max_available_bits + 30 entries, and also that the extra 30 padding
     * entries contain only binary (0 or 1) values, e.g., by setting them to zero.
     * The parameter max_available_bits does not represent the total number of bits
     * that will be consumed by the range coder, but the usually larger total number
     * of bits that have been read from the bitstream and have meaningful values.
     */

    /* Initialize the range decoder */
    void frd_init(FlexRangeDecoder* frd, /* i/o: Decoder State */
        int max_available_bits, /* i: Total maximum bits available */
        frd_read_byte_type read_byte, void* input_state);

    /* Get cumulative count */
    unsigned int frd_get_cumulative(/* o: Obtained cumulative count */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int tot_count /* i: Total count */
    );

    /* Get cumulative count when total count is a power of 2 */
    unsigned int frd_get_cumulative_fast(/* o: Obtained cumulative count */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int tot_shift /* i: Total count as a power of 2 */
    );

    /* Read an alphabet symbol when total count is a power of 2 */
    unsigned int frd_read_symbol_fast(/* o: Read symbol */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        const unsigned short table[], /* i: Cumulative counts table */
        unsigned int alphabet_size, /* i: Number of symbols in the alphabet */
        unsigned int tot_shift /* i: Total count as a power of 2 */
    );

    /* Update the state for the symbol found after using get_cumulative */
    void frd_update(FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int cum_count, /* i: Cumulative count */
        unsigned int sym_count /* i: Symbol count */
    );

    /* Read one bit with uniform probability */
    unsigned int frd_read_bit(/* o: Read bit */
        FlexRangeDecoder* frd /* i/o: Decoder State */
    );

    /* Read one bit with probability count0 / 2 ^ tot_shift for symbol 0 */
    unsigned int frd_read_bit_prob_fast(/* o: Read bit */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        int count0, /* i: Count for symbol 0 */
        unsigned int tot_shift /* i: Total count as a power of 2 */
    );

    /* Read up to 16 bits with uniform probability */
    unsigned int frd_read_bits(/* o: Read bits */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int bits /* i: Number of bits */
    );

    /* Read up to 32 bits with uniform probability */
    unsigned int frd_read_bits_long(/* i: Read bits */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int bits /* i: Number of bits */
    );

    /* Read value from an alphabet of size total with uniform distribution */
    unsigned int frd_read_uniform(/* i: Read value */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int total /* i: Alphabet size */
    );

    /* Read value from an alphabet of size total with uniform distribution */
    unsigned int frd_read_uniform_long(/* i: Read value */
        FlexRangeDecoder* frd, /* i/o: Decoder State */
        unsigned int total /* i: Alphabet size */
    );

    /* Get the total number of bits that would be consumed by finalization */
    int frd_virtual_finish(/* o: Total number of bits consumed */
        FlexRangeDecoder* frd /* i: Decoder state */
    );

    /* Finalize the range decoder */
    int frd_finish(/* o: Total number of bits consumed */
        FlexRangeDecoder* frd /* i/o: Decoder State */
    );

    /* Default callback to read a byte from the FILE* input_state */
    int frd_read_byte_default(void* input_state);

#ifdef __cplusplus
}
#endif

#endif
