#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <functional>
#include <map>
#include <memory>
#include <vector>

#include <MPEGI/AdaptiveEstimator.h>
#include <MPEGI/CodingUtils.h>
#include <MPEGI/flex_range_coder.h>

struct Alphabet
{
    Alphabet(std::string _n, std::size_t _s, std::size_t _f)
        : name(_n)
        , maxNumOfSymbols(_s)
        , maxFreq(_f)
    {}

    std::string name;
    std::size_t maxNumOfSymbols;
    std::size_t maxFreq;

    bool operator==(const Alphabet& other) const { return name == other.name; }
};

struct AlphabetHasher
{
    std::size_t operator()(const Alphabet& alphabet) const
    {
        return std::hash<std::string>{}(alphabet.name);
    }
};

template <> struct std::hash<Alphabet>
{
    std::size_t operator()(const Alphabet& alph) const noexcept
    {
        return std::hash<std::string>{}(alph.name);
    }
};

class EstimatorContainer
{
public:
    EstimatorContainer() = default;

    MPEGI::AdaptiveEstimator& getEstimator(const Alphabet& alph)
    {
        if (estimators.find(alph) == estimators.end()) {
            auto ae = MPEGI::AdaptiveEstimator();
            ae.create(alph.maxNumOfSymbols, alph.maxFreq);
            estimators.insert({ alph, ae });
        }

        return estimators.at(alph);
    }

private:
    std::unordered_map<Alphabet, MPEGI::AdaptiveEstimator> estimators;
};

struct ResizingBuffer
{
    ResizingBuffer()
        : buffer(512, '\x00')
        , pos(0)
    {}

    void writeByte(unsigned char byte)
    {
        if (pos >= buffer.size()) {
            std::size_t currentSize = buffer.size();
            buffer.resize(currentSize * 1.5);
        }

        buffer[pos++] = byte;
    }

    std::vector<unsigned char> buffer;
    std::size_t pos;
};

void write_byte_to_buffer(void* output_state, unsigned char byte_val)
{
    ((ResizingBuffer*)output_state)->writeByte(byte_val);
}

class FlexEncoder
{
public:
    FlexEncoder() { fre_init(&fre, write_byte_to_buffer, &buffer); }

    FlexEncoder(const FlexEncoder& other)
        : buffer(other.buffer)
        , estimatorContainer(other.estimatorContainer)
    {
        FlexRangeEncoder* fre_ptr = const_cast<FlexRangeEncoder*>(&other.fre);
        fre_save_state(fre_ptr, &fre);
    }

    void encodeSymbol(unsigned int symbol, const Alphabet& alph)
    {
        auto& estimator = estimatorContainer.getEstimator(alph);
        estimator.encodeUpdate(symbol, &fre);
    }

    void encodeSymbolUniform(unsigned int symbol, const Alphabet& alph)
    {
        fre_encode_uniform(&fre, symbol, alph.maxNumOfSymbols);
    }

    void encodeSymbolUniformLong(unsigned int symbol, const Alphabet& alph)
    {
        fre_encode_uniform_long(&fre, symbol, alph.maxNumOfSymbols);
    }

    void encodeBits(unsigned int value, unsigned int width) { fre_encode_bits(&fre, value, width); }

    void encodeBitsLong(unsigned int value, unsigned int width)
    {
        fre_encode_bits_long(&fre, value, width);
    }

    py::bytes finish()
    {
        fre_finish(&fre);
        const char* bufferPtr = (const char*)buffer.buffer.data();
        return py::bytes(bufferPtr, buffer.pos);
    }

private:
    FlexRangeEncoder fre;
    ResizingBuffer buffer;
    EstimatorContainer estimatorContainer;
};

class FlexDecoder
{
public:
    FlexDecoder(py::bytes inputBuffer)
    {
        py::buffer_info info(py::buffer(inputBuffer).request());
        mb.setup(info.size, (unsigned char*)info.ptr);
        frd_init(&frd, -1, MPEGI::frd_read_byte_from_memory_buffer, &mb);
    }

    unsigned int decodeSymbol(const Alphabet& alph)
    {
        auto& estimator = estimatorContainer.getEstimator(alph);
        return estimator.decodeUpdate(&frd);
    }

    unsigned int decodeSymbolUniform(const Alphabet& alph)
    {
        return frd_read_uniform(&frd, alph.maxNumOfSymbols);
    }

    unsigned int decodeSymbolUniformLong(const Alphabet& alph)
    {
        return frd_read_uniform_long(&frd, alph.maxNumOfSymbols);
    }

    unsigned int decodeBits(std::size_t width) { return frd_read_bits(&frd, width); }

    unsigned int decodeBitsLong(std::size_t width) { return frd_read_bits_long(&frd, width); }

    std::size_t finish() { return frd_finish(&frd); }

private:
    FlexRangeDecoder frd;
    MPEGI::MemoryBuffer mb;
    EstimatorContainer estimatorContainer;
};
