#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include "directivity.h"
#include "flex.h"
#include "material.h"
#include "mesh.h"

PYBIND11_MODULE(mcode, module)
{
    // Define modules and submodules
    module.doc() = "Python bindings for the Metadata Coding Library.";
    py::module_ direcMod = module.def_submodule("directivity", "Efficient Directivity Coding.");
    py::module_ matMod = module.def_submodule("material", "Efficient Acoustic Material Coding.");
    py::module_ meshMod = module.def_submodule("mesh", "Efficient Mesh Coding.");
    py::module_ flexMod = module.def_submodule("flex", "Flex Range Coding.");

    /* -------------------------------------------------------------------------- */
    /*                            DIRECTIVITY SUBMODULE                           */
    /* -------------------------------------------------------------------------- */

    py::class_<Cover>(direcMod, "Cover")
        .def(py::init<float, std::vector<float>>(), "Construct a Directivity Cover.", py::kw_only(),
            py::arg("freq"), py::arg("gains"))
        .def_readwrite("frequency", &Cover::frequency)
        .def_readwrite("gains", &Cover::gains);

    direcMod.def("encode", &encodeCovers, "Encode a list of covers into bytes.", py::arg("covers"),
        py::kw_only(), py::arg("deg_step"), py::arg("db_step"));

    /* -------------------------------------------------------------------------- */
    /*                             MATERIAL SUBMODULE                             */
    /* -------------------------------------------------------------------------- */

    py::class_<Coeffs>(matMod, "Coeffs")
        .def(py::init<float, float, float, float>(),
            "Construct a Frequency Coefficent Pair for Acoustic Materials.", py::kw_only(),
            py::arg("r"), py::arg("s"), py::arg("t"), py::arg("c"))
        .def_readwrite("r", &Coeffs::c1)
        .def_readwrite("s", &Coeffs::c2)
        .def_readwrite("t", &Coeffs::c3)
        .def_readwrite("c", &Coeffs::c4);

    py::class_<Material>(matMod, "Material")
        .def(py::init<int, std::vector<Coeffs>>(), "Construct an Acoustic Material.",
            py::arg("freq_grid_index"), py::arg("Coeffs"))
        .def_readwrite("freq_grid_index", &Material::freq_grid_index)
        .def_readwrite("coeffs", &Material::coeffs);

    matMod.def("encode", &encodeMaterials, "Encode a list of materials into bytes.",
        py::arg("materials"), py::kw_only(), py::arg("mat_qnt_step"));

    /* -------------------------------------------------------------------------- */
    /*                               MESH SUBMODULE                               */
    /* -------------------------------------------------------------------------- */

    py::class_<Mesh>(meshMod, "Mesh")
        .def(py::init<std::vector<Vertex>, std::vector<Face>>(), "Construct a Mesh.", py::kw_only(),
            py::arg("vertices"), py::arg("faces"))
        .def_readwrite("vertices", &Mesh::vertices)
        .def_readwrite("faces", &Mesh::faces);

    py::class_<Vertex>(meshMod, "Vertex")
        .def(py::init<float, float, float>(), "Construct a Vertex.", py::kw_only(), py::arg("x"),
            py::arg("y"), py::arg("z"))
        .def_readwrite("x", &Vertex::x)
        .def_readwrite("y", &Vertex::y)
        .def_readwrite("z", &Vertex::z);

    py::class_<Face>(meshMod, "Face")
        .def(py::init<std::optional<int>, int, int, int>(), "Construct a Face.", py::kw_only(),
            py::arg("mat_idx").none(true), py::arg("v1"), py::arg("v2"), py::arg("v3"))
        .def_readwrite("mat_idx", &Face::matIdx)
        .def_readwrite("v1", &Face::v1)
        .def_readwrite("v2", &Face::v2)
        .def_readwrite("v3", &Face::v3);

    meshMod.def("encode", &encodeMesh, "Encode a Mesh into bytes.", py::arg("mesh"), py::kw_only(),
        py::arg("vertex_qnt_step"), py::arg("vertex_tolerance"), py::arg("face_tolerance"),
        py::arg("remove_faces_without_mats"));

    /* -------------------------------------------------------------------------- */
    /*                               FLEX SUBMODULE                               */
    /* -------------------------------------------------------------------------- */

    py::class_<Alphabet>(flexMod, "Alphabet")
        .def(py::init<std::string, std::size_t, std::size_t>(), py::arg("name"), py::kw_only(),
            py::arg("size"), py::arg("max_freq") = 32768)
        .def_readwrite("name", &Alphabet::name)
        .def_readwrite("size", &Alphabet::maxNumOfSymbols)
        .def_readwrite("max_freq", &Alphabet::maxFreq);

    py::class_<FlexEncoder>(flexMod, "FlexEncoder")
        .def(py::init<>(), "Construct a Flex Range Encoder")
        .def("encode_symbol", &FlexEncoder::encodeSymbol, py::arg("symbol"), py::kw_only(),
            py::arg("alpha"))
        .def("encode_symbol_uniform", &FlexEncoder::encodeSymbolUniform, py::arg("symbol"),
            py::kw_only(), py::arg("alpha"))
        .def("encode_symbol_uniform_long", &FlexEncoder::encodeSymbolUniformLong, py::arg("symbol"),
            py::kw_only(), py::arg("alpha"))
        .def("encode_bits", &FlexEncoder::encodeBits, py::arg("value"), py::kw_only(),
            py::arg("width"))
        .def("encode_bits_long", &FlexEncoder::encodeBitsLong, py::arg("value"), py::kw_only(),
            py::arg("width"))
        .def("finish", &FlexEncoder::finish)
        .def("__copy__", [](const FlexEncoder& self) { return FlexEncoder(self); })
        .def(
            "__deepcopy__",
            [](const FlexEncoder& self, py::dict memo) { return FlexEncoder(self); },
            py::arg("memo"));

    py::class_<FlexDecoder>(flexMod, "FlexDecoder")
        .def(py::init<py::bytes>(), "Construct a Flex Range Decoder")
        .def("decode_symbol", &FlexDecoder::decodeSymbol, py::arg("alpha"))
        .def("decode_symbol_uniform", &FlexDecoder::decodeSymbolUniform, py::arg("alpha"))
        .def("decode_symbol_uniform_long", &FlexDecoder::decodeSymbolUniformLong, py::arg("alpha"))
        .def("decode_bits", &FlexDecoder::decodeBits, py::kw_only(), py::arg("width"))
        .def("decode_bits_long", &FlexDecoder::decodeBitsLong, py::kw_only(), py::arg("width"))
        .def("finish", &FlexDecoder::finish);
}
