/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/flex_range_coder.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#define FRD_DUMP_EOF 0

/*-------------------------------------------------------------------*
 * Local function prototypes
 *-------------------------------------------------------------------*/

static unsigned int frd_read(FlexRangeDecoder* frd);

/*-------------------------------------------------------------------*
 * frd_init()
 *
 * Initialize the range decoder
 *-------------------------------------------------------------------*/

void frd_init(FlexRangeDecoder* frd, /* i/o: Decoder State */
    int max_available_bits, /* i: Total maximum bits available */
    frd_read_byte_type read_byte, void* input_state)
{
    int i;

    frd->low = 0;
    frd->range = 0xFFFFFFFF;

    frd->bit_count = 0;
    if (max_available_bits != -1) {
        /* the renormalization procedure reads at most 32 - 2 extra bits than available */
        frd->max_allowable_bit_count = max_available_bits + 30;
    } else {
        /* no information about the total maximum bits available, set it to "unlimited" */
        frd->max_allowable_bit_count = 0x7FFFFFFF;
    }

    frd->eof_bytes_read_count = 0;
    frd->bit_error_detected = 0;

    frd->read_byte = read_byte;
    frd->input_state = input_state;

    for (i = 0; i < 4; i++) {
        frd->low = (frd->low << 8) + frd_read(frd);
    }
}

/*-------------------------------------------------------------------*
 * frd_get_cumulative()
 *
 * Get cumulative count
 *-------------------------------------------------------------------*/

unsigned int frd_get_cumulative(/* o: Obtained cumulative count */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int tot_count /* i: Total count */
)
{
    unsigned int val;
    assert(tot_count <= 65536);
    assert(frd->range >= 0x01000000); /* range is normalized */

    frd->range /= tot_count;

    val = frd->low / frd->range;

    /* in case of bitstream errors it is possible that val >= tot_count */
    if (val >= tot_count) {
        frd->bit_error_detected = 1;

        /* for valid bistreams, always low < range, therefore setting low = range */
        /* will always reach the bitstream error condition branch on the next call */
        frd->range = 0xFFFFFFFF;
        frd->low = frd->range;

        /* the current value and all the following values are very likely incorrect */
        return 0; /* return the minimum valid value for the output */
    }

    return val;
}

/*-------------------------------------------------------------------*
 * frd_get_cumulative_fast()
 *
 * Get cumulative count when total count is a power of 2
 *-------------------------------------------------------------------*/

unsigned int frd_get_cumulative_fast(/* o: Obtained cumulative count */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int tot_shift /* i: Total count as a power of 2 */
)
{
    unsigned int val;
    assert(tot_shift <= 16);
    assert(frd->range >= 0x01000000); /* range is normalized */

    frd->range >>= tot_shift;

    val = frd->low / frd->range;

    /* in case of bitstream errors it is possible that val >= (1 << tot_shift) */
    if ((val >> tot_shift) != 0) {
        frd->bit_error_detected = 1;

        /* for valid bistreams, always low < range, therefore setting low = range */
        /* will always reach the bitstream error condition branch on the next call */
        frd->range = 0xFFFFFFFF;
        frd->low = frd->range;

        /* the current value and all the following values are very likely incorrect */
        return 0; /* return the minimum valid value for the output */
    }

    return val;
}

/*-------------------------------------------------------------------*
 * frd_read_symbol_fast()
 *
 * Read an alphabet symbol when total count is a power of 2
 *-------------------------------------------------------------------*/

unsigned int frd_read_symbol_fast(/* o: Read symbol */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    const unsigned short table[], /* i: Cumulative counts table */
    unsigned int alphabet_size, /* i: Number of symbols in the alphabet */
    unsigned int tot_shift /* i: Total count as a power of 2 */
)
{
    unsigned int sym_begin;
    unsigned int sym_end;
    unsigned int sym_middle;
    unsigned int ceil_log2_alphabet_size; /* ceil(log2(alphabet_size)) */
    unsigned int step;
    assert(tot_shift <= 16);
    assert(frd->range >= 0x01000000); /* range is normalized */

    frd->range >>= tot_shift;

    /* the cumulative value is val = low / range */
    /* the condition val >= level is equivalent with low >= range * level */

    /* in case of bitstream errors it is possible that val >= (1 << tot_shift) */
    if (frd->low >= (frd->range << tot_shift)) {
        frd->bit_error_detected = 1;

        /* for valid bistreams, always low < range, therefore setting low = range */
        /* will always reach the bitstream error condition branch on the next call */
        frd->range = 0xFFFFFFFF;
        frd->low = frd->range;

        /* the current value and all the following values are very likely incorrect */
        return 0; /* return the minimum valid value for the output */
    }

    /* do a binary search to find the symbol read */
    sym_begin = 0;
    sym_end = alphabet_size; /* the possible values are {sym_begin, .., sym_end - 1} */

    /* see ceil_log2(n) function for the derivation, used for any 2 <= alphabet_size */
    ceil_log2_alphabet_size = 32 - clz_u32(alphabet_size - 1); /* ceil(log2(alphabet_size)) */

    /* at most ceil_log2_alphabet_size steps are needed so that sym_end - sym_begin == 1 */
    for (step = 0; step < ceil_log2_alphabet_size; step++) {
        sym_middle = (sym_begin + sym_end) >> 1;
        if (frd->range * table[sym_middle] <= frd->low) {
            sym_begin = sym_middle;
        } else {
            sym_end = sym_middle;
        }
    }

    /* sym_begin contains the symbol read */

    /* an extra entry with value 2 ^ tot_shift must be present at the end of table */
    frd_update(frd, table[sym_begin], table[sym_begin + 1] - table[sym_begin]);

    return sym_begin;
}

/*-------------------------------------------------------------------*
 * frd_update()
 *
 * Update the state for the symbol found after using get_cumulative
 *-------------------------------------------------------------------*/

void frd_update(FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int cum_count, /* i: Cumulative count */
    unsigned int sym_count /* i: Symbol count */
)
{
    frd->low -= cum_count * frd->range;
    frd->range *= sym_count;

    /* range was shifted right by up to 16, so at most two renormalizations are needed */
    if (frd->range < 0x01000000) {
        frd->low = (frd->low << 8) + frd_read(frd);
        frd->range <<= 8;
        if (frd->range < 0x01000000) {
            frd->low = (frd->low << 8) + frd_read(frd);
            frd->range <<= 8;
        }
    }
}

/*-------------------------------------------------------------------*
 * frd_read_bit()
 *
 * Read one bit with uniform probability
 *-------------------------------------------------------------------*/

unsigned int frd_read_bit(/* o: Read bit */
    FlexRangeDecoder* frd /* i/o: Decoder State */
)
{
    unsigned int val;
    assert(frd->range >= 0x01000000); /* range is normalized */

    frd->range >>= 1;

    /* in case of bitstream errors it is possible that low >= (range << 1) */
    val = 0;
    if (frd->low >= frd->range) {
        val++;
        frd->low -= frd->range;

        /* range was already subtracted once from low */
        if (frd->low >= frd->range) {
            frd->bit_error_detected = 1;

            /* for valid bistreams, always low < range, therefore setting low = range */
            /* will always reach the bitstream error condition branch on the next call */
            frd->range = 0xFFFFFFFF;
            frd->low = frd->range;

            /* the current value and all the following values are very likely incorrect */
            return 0; /* return the minimum valid value for the output */
        }
    }

    /* range was shifted right by 1, so at most one renormalization is needed */
    if (frd->range < 0x01000000) {
        frd->low = (frd->low << 8) + frd_read(frd);
        frd->range <<= 8;
    }

    return val;
}

/*-------------------------------------------------------------------*
 * frd_read_bit_prob_fast()
 *
 * Read one bit with probability count0 / 2 ^ tot_shift for symbol 0
 *-------------------------------------------------------------------*/

unsigned int frd_read_bit_prob_fast(/* o: Read bit */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    int count0, /* i: Count for symbol 0 */
    unsigned int tot_shift /* i: Total count as a power of 2 */
)
{
    unsigned int val;
    assert(tot_shift <= 16);
    assert(frd->range >= 0x01000000); /* range is normalized */

    frd->range >>= tot_shift;

    /* in case of bitstream errors it is possible that low >= (1 << tot_shift) * range */
    if (frd->low >= (frd->range << tot_shift)) /* equivalent condition */
    {
        frd->bit_error_detected = 1;

        /* for valid bistreams, always low < range, therefore setting low = range */
        /* will always reach the bitstream error condition branch on the next call */
        frd->range = 0xFFFFFFFF;
        frd->low = frd->range;

        /* the current value and all the following values are very likely incorrect */
        return 0; /* return the minimum valid value for the output */
    }

    val = 0;
    if (frd->low >= frd->range * count0) {
        val++;

        frd->low -= frd->range * count0;
        frd->range *= (1 << tot_shift) - count0; /* count1 = (1 << tot_shift) - count0 */
    } else {
        frd->range *= count0;
    }

    /* range was shifted right by up to 16, so at most two renormalizations are needed */
    if (frd->range < 0x01000000) {
        frd->low = (frd->low << 8) + frd_read(frd);
        frd->range <<= 8;
        if (frd->range < 0x01000000) {
            frd->low = (frd->low << 8) + frd_read(frd);
            frd->range <<= 8;
        }
    }

    return val;
}

/*-------------------------------------------------------------------*
 * frd_read_bits()
 *
 * Read up to 16 bits with uniform probability
 *-------------------------------------------------------------------*/

unsigned int frd_read_bits(/* o: Read bits */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int bits /* i: Number of bits */
)
{
    unsigned int val;
    assert(bits <= 16);
    assert(frd->range >= 0x01000000); /* range is normalized */

    frd->range >>= bits;

    val = frd->low / frd->range;

    /* in case of bitstream errors it is possible that val >= (1 << bits) */
    if ((val >> bits) != 0) /* equivalent condition */
    {
        frd->bit_error_detected = 1;

        /* for valid bistreams, always low < range, therefore setting low = range */
        /* will always reach the bitstream error condition branch on the next call */
        frd->range = 0xFFFFFFFF;
        frd->low = frd->range;

        /* the current value and all the following values are very likely incorrect */
        return 0; /* return the minimum valid value for the output */
    }

    frd->low -= val * frd->range;

    /* range was shifted right by up to 16, so at most two renormalizations are needed */
    if (frd->range < 0x01000000) {
        frd->low = (frd->low << 8) + frd_read(frd);
        frd->range <<= 8;
        if (frd->range < 0x01000000) {
            frd->low = (frd->low << 8) + frd_read(frd);
            frd->range <<= 8;
        }
    }

    return val;
}

/*-------------------------------------------------------------------*
 * frd_read_bits_long()
 *
 * Read up to 32 bits with uniform probability
 *-------------------------------------------------------------------*/

unsigned int frd_read_bits_long(/* i: Read bits */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int bits /* i: Number of bits */
)
{
    unsigned int value;

    if (bits > 16) {
        value = frd_read_bits(frd, bits - 16);
        value = (value << 16) | frd_read_bits(frd, 16);
    } else {
        value = frd_read_bits(frd, bits);
    }

    return value;
}

/*-------------------------------------------------------------------*
 * frd_read_uniform()
 *
 * Read value from an alphabet of size total with uniform distribution
 *-------------------------------------------------------------------*/

unsigned int frd_read_uniform(/* i: Read value */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int total /* i: Alphabet size */
)
{
    unsigned int value;
    assert((total >= 1) && (total <= 65536));

    value = frd_get_cumulative(frd, total);
    frd_update(frd, value, 1);

    return value;
}

/*-------------------------------------------------------------------*
 * frd_read_uniform_long()
 *
 * Read value from an alphabet of size total with uniform distribution
 *-------------------------------------------------------------------*/

unsigned int frd_read_uniform_long(/* i: Read value */
    FlexRangeDecoder* frd, /* i/o: Decoder State */
    unsigned int total /* i: Alphabet size */
)
{
    unsigned int value;
    int n;
    unsigned int mask;
    unsigned int extra_bin;
    assert((total >= 1) && (total <= (1 << 24)));
    assert(sizeof(total) == 4); /* ensure the "pseudo" sign bit computation is correct */

    if (total <= 256) {
        value = frd_get_cumulative(frd, total);
        frd_update(frd, value, 1);
    } else { /* 256 < total <= 2 ^ 24 */
        /* see ceil_log2(n) function for the derivation, valid for any 2 <= total */
        n = 32 - clz_u32(total - 1); /* n = ceil(log2(total)) */
        assert((n >= 9) && (n <= 24));

        n -= 8;
        mask = (1 << n) - 1;

        /* add an extra bin at the end if total is not a multiple of 2 ^ n */
        extra_bin = 1 - (((total & mask) - 1) >> 31);
        /* completely equivalent with extra_bin = ((total & mask) != 0) */

        /* recompose total from (total >> n) + extra_bin equal bins of size 2 ^ n */
        value = frd_get_cumulative(frd, (total >> n) + extra_bin);
        frd_update(frd, value, 1);
        value = (value << n) | frd_read_bits(frd, n);
    }

    return value;
}

/*-------------------------------------------------------------------*
 * frd_finish()
 *
 * Finalize the range decoder
 *-------------------------------------------------------------------*/

int frd_finish(/* o: Total number of bits consumed */
    FlexRangeDecoder* frd /* i/o: Decoder State */
)
{
    int total_bit_count;
    int bits;
    assert(frd->range >= 0x01000000); /* range is normalized */

    bits = clz_u32(frd->range) + 1; /* bits = 32 - floor(log2(range)) */
    /* 32 - floor(log2(y)) = 32 - [31 - clz_u32(y)] = clz_u32(y) + 1 */

    bits++; /* conservative number of bits, because the decoder only has range available */

    assert((bits >= 2) && (bits <= 9)); /* depends on range, which is normalized */

    total_bit_count = (frd->bit_count - 32) + bits;

    return total_bit_count;
}

/*-------------------------------------------------------------------*
 * frd_virtual_finish()
 *
 * Get the total number of bits that would be consumed by finalization
 *-------------------------------------------------------------------*/

int frd_virtual_finish(/* o: Total number of bits consumed */
    FlexRangeDecoder* frd /* i: Decoder state */
)
{
    /* the function is completely equivalent with frd_finish */
    return frd->bit_count + clz_u32(frd->range) - 30;
}

/*-------------------------------------------------------------------*
 * frd_read()
 *
 * Read one byte from the bit stream (internal function)
 *-------------------------------------------------------------------*/

static unsigned int frd_read(/* o: Byte read */
    FlexRangeDecoder* frd /* i/o: Decoder State */
)
{
    int byte_read;

    byte_read = frd->read_byte(frd->input_state);
    frd->bit_count += 8;

    /*
     * In case of bitstream errors the number the bits read may be larger than the total
     * number of bits that have been read from the bitstream and have meaningful values
     * plus the extra 30 padding bits that must be appended to the buffer by the caller.
     */
    if ((byte_read < 0) || (frd->bit_count > frd->max_allowable_bit_count)) {
        if (byte_read == -1) /* EOF byte read */
        {
            frd->eof_bytes_read_count += 1;
            /* 3 additional bytes are read by the finalization of the range decoder */
            if (frd->eof_bytes_read_count <= 3) {
                return 0; /* return a dummy 0 byte used for finalization */
            }
        }

        frd->bit_error_detected = 1;

        /* for valid bistreams, always low < range, therefore setting low = range */
        /* will always reach the bitstream error condition branch on the next call */
        frd->range = 0xFFFFFFFF;
        frd->low = frd->range;

        return 0; /* reading the 8 bits would trigger an out-of-bounds array access */
    }

    return (unsigned int)byte_read;
}

/*-------------------------------------------------------------------*
 * frd_read_byte_default()
 *
 * Default callback to read a byte from the FILE* input_state
 *-------------------------------------------------------------------*/

int frd_read_byte_default(void* input_state)
{
    int byte = fgetc((FILE*)input_state);

    if (byte != EOF) {
        return byte;
    } else {
        if (ferror((FILE*)input_state)) {
#if FRD_DUMP_EOF
            printf("read error\n");
#endif

            return -2;
        } else if (feof((FILE*)input_state)) {
#if FRD_DUMP_EOF
            printf("read eof\n");
#endif

            return -1;
        } else {
#if FRD_DUMP_EOF
            printf("read inconsistent\n");
#endif

            assert(0); /* inconsistent, should never happen */
            return -2;
        }
    }
}
