# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import cProfile
import inspect
import logging
import os
import pathlib
import sys
from typing import List, Optional

from mpegi import cli, coding, commands, encoder, framework, geometry, metadata, sofa, voxel


def main(args: Optional[List[str]] = None) -> None:
    defaultConfigFile = pathlib.Path(__file__).parents[2].joinpath("default.conf")

    argv = args if args is not None else sys.argv
    config = cli.Config(argv, str(defaultConfigFile))
    config.parser.prog = "python -m mpegi"
    config.parser.description = "MPEG-I Tools - A collection of tools to aid MPEG-I development"

    # Setup command line arguments and config files
    verbosityArgs = config.parser.add_mutually_exclusive_group()
    verbosityArgs.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose logging"
    )
    verbosityArgs.add_argument("-vv", action="store_true", help="Enable super verbose logging")
    config.parser.add_argument(
        "-l",
        "--log",
        metavar="PATH",
        help="Outputs a log to the given file path",
    )
    config.parser.add_argument(
        "--profile",
        action="store_true",
        help="Dumps profiling stats in pstats format",
    )

    # Programatically register all commands and register arguments
    subcommandCallbacks = {}
    commandsModule = commands.__name__
    foundClasses = inspect.getmembers(sys.modules[commandsModule], inspect.isclass)
    foundCommands = [c[1] for c in foundClasses if issubclass(c[1], cli.Command)]
    for command in foundCommands:
        subcommandName = command.getCommandName()
        subparser = config.addSubcommand(subcommandName)
        command.registerArguments(subparser)
        subcommandCallbacks[subcommandName] = command.execute

    # Setup logging
    if not config["verbose"] and not config["vv"] and not config["log"]:
        logging.getLogger().disabled = True
    else:
        logLevel = logging.WARNING
        logFormat = "%(asctime)s %(message)s"
        if config["verbose"]:
            logLevel = logging.INFO
        elif config["vv"]:
            logLevel = logging.DEBUG

        logging.getLogger().setLevel(logLevel)

        # Setup file logger
        if config["log"]:
            logFile = config["log"]
            if not os.path.isabs(logFile):
                logFile = os.getcwd() + os.sep + config["log"]

            with open(logFile, "w"):
                pass  # Clear the log

            fileHandler = logging.FileHandler(logFile)
            fileHandler.setFormatter(logging.Formatter(logFormat))
            fileHandler.setLevel(logLevel)
            logging.getLogger().addHandler(fileHandler)

        # Setup console logger
        if config["verbose"] or config["vv"]:
            consoleHandler = logging.StreamHandler()
            consoleHandler.setLevel(logLevel)
            consoleHandler.setFormatter(logging.Formatter(logFormat))
            logging.getLogger().addHandler(consoleHandler)

    # Execute the subcommand with the previously registered entry point
    assert config["subcommand"] != None, "A subcommand must be called"
    try:
        if config["profile"]:  # Dump profiling stats
            command = 'subcommandCallbacks[config["subcommand"]](config)'
            cProfile.runctx(command, globals(), locals(), sort="time")
        else:
            subcommandCallbacks[config["subcommand"]](config)
    except Exception as exception:
        logging.exception(exception)
        raise exception
