# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse

from .config import Config


class Command:
    """Command declares an interface to implement a subcommand on the terminal.
    A command requires three things:

    - A Name - this will be the string that is called on the terminal
    - Argument Registration - the command can register what arguments it needs
    - Entry Point - the code to be executed when the command is called
    """

    @staticmethod
    def execute(config: Config) -> None:
        """When the command is called on the command, execute() will be the
        entry point for the command

        :param config: The config instance in which the arguments can be accessed
        :raises NotImplementedError: Commands must implement this function
        """
        raise NotImplementedError("Command class must implement its entry point")

    @staticmethod
    def getCommandName() -> str:
        """Returns the name of the command. This will be the string that is
        entered on the terminal.

        :raises NotImplementedError: Commands must implement this function
        :return: The name of the command
        :rtype: str
        """
        raise NotImplementedError("Command class must declare its name")

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        """An argument parser is passed as an argument and through it arguments
        can be registered. Look at argparse documentation for more details.
        https://docs.python.org/3/library/argparse.html

        :param argParser: The argument parser instance
        :type argParser: argparse.ArgumentParser
        :raises NotImplementedError: All commands must implement this function
        """
        raise NotImplementedError("Command Class must register to arguments")
