# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import copy
from typing import Any, Dict, List


class NewlineFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith("\n"):
            return text[1:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)


class Config:
    """Config provides functionality for command line arguments and config
    files. Typical usage would be to add a "subcommand" and then add all
    of the subcommand parameters, similar to *git*, e.g "git commit" or
    "git status" or "git log".

    To add a global argument, access the "parser" member variable and call the
    'add_argument()' function. See argparse documentation for more details.
    https://docs.python.org/3/library/argparse.html

    This class can be accessed via the "[]" operator, much like a dictionary.
    It will return the value of the argument which was registered prior.
    """

    def __init__(self, argv, defaultConfigFile):
        self.argv = argv

        self.defaultConfigFile = defaultConfigFile

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument(
            "-c",
            "--config",
            metavar="PATH",
            help=(
                "Specify a path to a config file that will completely "
                "overwrite the default config."
            ),
        )

        self.subparserGroup = None  # Lazy loaded
        self.subcommands = []

    def addSubcommand(self, subcommand: str) -> argparse.ArgumentParser:
        """Adds a subcommand to the command line parsing

        :param subcommand: The string in which the subcommand is invoked
        :type subcommand: str
        :return: An argparse ArgumentParser - refer argparse documentation
            for more details https://docs.python.org/3/library/argparse.html
        :rtype: argparse.ArgumentParser
        """
        if self.subparserGroup == None:
            self.subparserGroup = self.parser.add_subparsers(dest="subcommand")

        self.subcommands.append(subcommand)
        return self.subparserGroup.add_parser(subcommand, formatter_class=NewlineFormatter)

    def getArgs(self) -> Dict[str, Any]:
        """Returns all of the arguments that were passed on the command line

        :return: A dictionary where the key is the argument and value is the
            was the option passed into said argument
        :rtype: Dict[str, Any]
        :raises Exception: Raises exception if the config file is invalid
        """
        if "-h" in self.argv or "--help" in self.argv:
            self.parser.parse_args()

        configFile = self._getConfigFile()
        argvWithConfig = self._formatArgvWithConfig(self.argv, configFile)

        return vars(self.parser.parse_args(argvWithConfig))

    def addArg(self, key: str):
        self.argv.append(key)

    def _formatArgvWithConfig(self, argv, configFile) -> List[str]:
        """Place all config args found from config file into the argv list"""
        if "mpegi" in argv[0]:
            argv.pop(0)
        configFileArgs = self._getConfigFileArgs(configFile)
        args = copy.deepcopy(argv)
        if len(args) > 0:
            if ".py" in argv[0]:
                args.pop(0)

        subcommands = [s for s in self.subcommands if s in args]
        args[0:0] = configFileArgs.get("global", [])
        for subcommand in subcommands:
            insertIndex = args.index(subcommand) + 1
            args[insertIndex:insertIndex] = configFileArgs.get(subcommand, [])

        return args

    def _getConfigFile(self) -> str:
        """Try to find a config file in argv, otherwise return the default file"""
        configFileParser = argparse.ArgumentParser()
        configFileParser.add_argument("-c", "--config")
        knownArgs, _ = configFileParser.parse_known_args(self.argv)
        configFile = knownArgs.config
        if knownArgs.config == None:
            configFile = self.defaultConfigFile
        return configFile

    def _getConfigFileArgs(self, configFile) -> Dict[str, List[str]]:
        """Grab all configs in file as a dict where the key is the subcommand
        and the value being a list of all config options
        """
        with open(configFile, "r") as file:
            fileContents = file.readlines()

        configFileArgs = {}
        section = None
        for line in fileContents:
            line = line.strip()

            if line.startswith("#") or line.startswith(";"):
                continue

            if line.startswith("[") and line.endswith("]"):
                section = line[1:-1]
                configFileArgs[section] = []
                continue

            if line != "":
                assert section != None, f"All Config File args must be in a section"
                configFileArgs[section].append(line)

        return configFileArgs

    def __getitem__(self, key: str):
        return self.getArgs()[key]
