# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import logging
from math import ceil, cos, floor, pi
from typing import List

import mcode
import numpy as np

from mpegi.sofa import SOFAReader


class DirectivityEncoder:
    """Code source directivities as a binary blob"""

    def __init__(self):
        # Used as a cache for faster access
        self.magnitudes = {}
        self.path: str = None
        self.lowestPossibleMagnitudeValue = 10 ** (-127 / 20)  # -127dB as Magnitude

    def encode(
        self,
        path: str,
        deg_step_req: float,
        db_step_req: float,
    ) -> bytes:
        """Code a source directivity as a binary blob.

        :param path: The path to the .sofa file of the source directivity
        :param deg_step_req: Quantization step of degrees in the source directivity
        :param db_step_req: Quantization step of decibels in the source directivity
        :param dump: If true, produces data dumps as text files, defaults to False
        :return: The source directivity as bytes
        """
        self.magnitudes = {}  # clear cache
        self.path = path
        directivity_object = SOFAReader(path)

        covers: List[mcode.directivity.Cover] = []

        for freqIndex, freq in enumerate(directivity_object.getVar("N")):
            cover_list = self._extractSphereCovering(directivity_object, deg_step_req, freqIndex)
            covers.append(mcode.directivity.Cover(freq=freq, gains=cover_list))

        encoded_data = mcode.directivity.encode(covers, deg_step=deg_step_req, db_step=db_step_req)

        return encoded_data

    def _extractSphereCovering(self, directivity_object, deg_step_req, freqIndex):
        assert (deg_step_req >= 2.0) and (deg_step_req <= 90.0)

        cover = []

        # add 0.01 to avoid reaching the next finer grid if deg_step_req
        # is supplied already rounded down using 2 fractional digits
        intervals_per_90 = ceil(90.0 / (deg_step_req + 0.01))

        # the actual degree step that results (used for elevations and also at the equator)
        deg_step = 90.0 / intervals_per_90

        el_cnt = 2 * intervals_per_90 + 1  # count of elevation levels
        # each elevation level has a different count of azimuth points, proportional to el_circle_ratio
        az_cnt = [0] * el_cnt
        # maximum count of azimuth points (at the equator), also named cover_width in C++ code
        az_cnt_max = 4 * intervals_per_90

        for el_idx in range(el_cnt):
            el_ang = (el_idx - intervals_per_90) * deg_step
            # relative circle lentgth at the current elevation level (1 at the equator)
            el_circle_ratio = cos(el_ang * pi / 180.0)
            # floor(x + 0.5), for x >= 0.0, is (int) lround(x) in C++ (nearest integer, rounding halfway cases away from zero)
            # count of azimuth points
            az_cnt[el_idx] = max(floor(el_circle_ratio * az_cnt_max + 0.5), 1)

            az_step = 360.0 / az_cnt[el_idx]  # azimuth degree step at the current elevation level
            for az_idx in range(az_cnt[el_idx]):
                az_ang = az_idx * az_step

                directivity_in_dB = self._getDirectivity(
                    directivity_object, el_ang, az_ang, freqIndex
                )
                cover.append(directivity_in_dB)

        return cover

    def _getDirectivity(self, directivity_object, el, az, freqIndex):
        # the actual value of the directivity in dB for elevation el and azimuth al should be retrieved from directivity_object
        try:
            magnitude = self.magnitudes[(el, az)][freqIndex].real
        except:
            self.magnitudes[(el, az)] = directivity_object.getTF([az, el, 1])[0]
            magnitude = self.magnitudes[(el, az)][freqIndex].real

        if magnitude < self.lowestPossibleMagnitudeValue:
            newMagnitude = self.lowestPossibleMagnitudeValue
            logging.error(
                f"Magnitude of '{magnitude}' in directivity {self.path} "
                "is too low of a value to be stored as dB. Magnitude was set "
                f"to {self.lowestPossibleMagnitudeValue} before storing."
            )
            magnitude = newMagnitude

        return 20 * np.log10(magnitude)
