# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Code Geometries into a binary blob."""


from typing import List

import mcode

from mpegi.metadata import eifplus


class GeometryCoder:
    """Code geometries into a binary blob.

    This class has text dump functionality which is disabled by default.
    """

    # Non-runtime tweakable parameters
    vertexQuantizationStep: float = 0.001
    """Quantization step for vertex coordinates in meters"""
    vertexTolerance: float = 0.0
    """The tolerance in which to remove duplicate vertices in meters"""
    faceTolerance: float = 0.0
    """The tolerance in which to remove degenerate faces in meters"""

    def __init__(self) -> None:
        """Contructs a GeometryCoder."""
        pass

    def code_mesh(
        self,
        mesh: eifplus.Mesh,
        materialIds: List[int],
        removeNoMaterialFaces: bool = True,
    ) -> bytes:
        """Code a mesh as a binary blob.

        :param mesh: The mesh to code
        :param materials: All materials used in the mesh - no more, no less
        :param removeNoMaterialFaces: If true, all faces which do not have an
            acoustic material assigned to them will be removed. Defaults to True.
        :return: The mesh as bytes
        """
        vertices = [mcode.mesh.Vertex(x=m[0], y=m[1], z=m[2]) for m in mesh.vertices]
        faces = [
            mcode.mesh.Face(mat_idx=None if m < 0 else m, v1=f[0], v2=f[1], v3=f[2])
            for f, m in zip(mesh.faces, materialIds)
        ]
        marshalled_mesh = mcode.mesh.Mesh(vertices=vertices, faces=faces)

        encoded_data = mcode.mesh.encode(
            marshalled_mesh,
            vertex_qnt_step=self.vertexQuantizationStep,
            vertex_tolerance=self.vertexTolerance,
            face_tolerance=self.faceTolerance,
            remove_faces_without_mats=removeNoMaterialFaces,
        )

        return encoded_data
