# This software module was originally developed by
#
# Ericsson AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Ericsson AB retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2023.

import logging
import os
from math import ceil, cos, floor, pi

import numpy as np

from mpegi.granular import GrainDBReader
from mpegi.framework import Bitstream, Component


class GrainDBEncoder:
    """Code a Granular database as a list of values"""

    # def __init__(self):
    def encode(self, path: str):
        """Code a Granular database as a list of values
        :param path: The path to the .h5 file of granular database
        :return: The Granular database read from the .h5 file as a list of arrays, the composition of bitstream is done in the payload.
        """
        self.path = path
        grain_object = GrainDBReader(path)
        numGrain = grain_object.getDim("M")
        numChannel = grain_object.getDim("Ch")
        descriptor_coordinates = grain_object.getVar("DescriptorCoordinates")
        dim = grain_object.getDim("C")
        # normalize descriptor coordinates in every dimension:
        for ind in range(dim):
            descriptor_coordinates[:, ind] = (
                descriptor_coordinates[:, ind] - min(descriptor_coordinates[:, ind])
            ) / (max(descriptor_coordinates[:, ind]) - min(descriptor_coordinates[:, ind]))

        All_grains = grain_object.getVar("grains")
        if numChannel > 1:
            grainLenAll = [len(grain[0]) for grain in All_grains]
        else:
            grainLenAll = [len(grain) for grain in All_grains]

        grain_metadata = grain_object.getMetaData()

        encoded_data = {}  # empty dictionary of granular data which has various attributes
        encoded_data["DescriptorCoordinates"] = descriptor_coordinates
        encoded_data["Dimensions"] = dim
        encoded_data["Channels"] = numChannel
        encoded_data["numGrains"] = numGrain
        encoded_data["grainLens"] = grainLenAll
        encoded_data["GranularAudioEncodingMode"] = 0
        encoded_data["Grains"] = All_grains

        # get all the defined global parameters and metadata from the granular database:
        encoded_data["coordinateScaling"] = grain_object.getVar("coordinateScaling")
        encoded_data["smoothingTargetCoordinates"] = grain_object.getVar(
            "smoothingTargetCoordinates"
        )
        encoded_data["searchRadius"] = grain_object.getVar("searchRadius")
        encoded_data["NumMixRule"] = grain_object.getVar(
            "NumMixRule"
        )  # number of locations in the descriptor coordinate space where type of mixing rule is defined
        encoded_data["grainMixingRule"] = grain_object.getVar("grainMixingRule")
        encoded_data["grainMixingRuleCoordinates"] = grain_object.getVar(
            "grainMixingRuleCoordinates"
        )
        encoded_data["grainHistoryLength"] = grain_object.getVar("grainHistoryLength")

        # exponents for probability shaping-compulsory:
        encoded_data["probabilityExponentPosition"] = grain_object.getVar(
            "probabilityExponentPosition"
        )
        encoded_data["probabilityExponentHistory"] = grain_object.getVar(
            "probabilityExponentHistory"
        )

        # weights for probability shaping-compulsory:
        encoded_data["probabilityWeightPosition"] = grain_object.getVar("probabilityWeightPosition")
        encoded_data["probabilityWeightHistory"] = grain_object.getVar("probabilityWeightHistory")

        # probability groups:
        if "probabilityGroups" in grain_metadata and grain_metadata["probabilityGroups"]:
            encoded_data["probabilityGroups"] = True
            encoded_data["P"] = grain_object.getVar("P")  # number of probability groups
            encoded_data["probabilityGroupVal"] = grain_object.getVar("probabilityGroup")
            encoded_data["grainProbabilityGroup"] = grain_object.getVar("grainProbabilityGroup")
            encoded_data["perGrainProbabilityVal"] = grain_object.getVar("perGrainProbabilityVal")
        else:
            encoded_data["probabilityGroups"] = False

        # random selection:
        if "randomSelection" in grain_metadata and grain_metadata["randomSelection"]:
            encoded_data["randomSelection"] = True
        else:
            encoded_data["randomSelection"] = False

        # fast grain switch:
        if (
            "fastGrainSwitchThresholdAvailable" in grain_metadata
            and grain_metadata["fastGrainSwitchThresholdAvailable"]
        ):
            encoded_data["fastGrainSwitchThresholdAvailable"] = True
            encoded_data["fastGrainSwitchThreshold"] = grain_object.getVar(
                "fastGrainSwitchThreshold"
            )
        else:
            encoded_data["fastGrainSwitchThresholdAvailable"] = False
            encoded_data["fastGrainSwitchThreshold"] = 1  # default value

        if "trend" in grain_metadata and grain_metadata["trend"]:
            encoded_data["DescTrend"] = True
            encoded_data["DescriptorCoordinatesTrend"] = grain_object.getVar(
                "DescriptorCoordinatesTrend"
            )
            encoded_data["probabilityExponentTrend"] = grain_object.getVar(
                "probabilityExponentTrend"
            )
            encoded_data["probabilityWeightTrend"] = grain_object.getVar("probabilityWeightTrend")
            encoded_data["descrTrendMethod"] = grain_object.getVar("descrTrendMethod")
        else:
            encoded_data["DescTrend"] = False
            encoded_data["DescriptorCoordinatesTrend"] = []
            encoded_data["probabilityExponentTrend"] = []
            encoded_data["probabilityWeightTrend"] = []
            encoded_data["descrTrendMethod"] = []

        if "position_in_file" in grain_metadata and grain_metadata["position_in_file"]:
            encoded_data["OrigPosition"] = True
            encoded_data["PositionWithinAudioFile"] = grain_object.getVar("PositionWithinAudioFile")
            encoded_data["probabilityExponentOriginalPos"] = grain_object.getVar(
                "probabilityExponentOriginalPos"
            )
            encoded_data["probabilityWeightOriginalPos"] = grain_object.getVar(
                "probabilityWeightOriginalPos"
            )
        else:
            encoded_data["OrigPosition"] = False
            encoded_data["PositionWithinAudioFile"] = []
            encoded_data["probabilityExponentOriginalPos"] = []
            encoded_data["probabilityWeightOriginalPos"] = []

        if "grain_overlap_amount" in grain_metadata and grain_metadata["grain_overlap_amount"]:
            encoded_data["grain_overlap_flag"] = True
            encoded_data["GrainOverlapAmount"] = grain_object.getVar("GrainOverlapAmount")
        else:
            encoded_data["grain_overlap_flag"] = False
            encoded_data["GrainOverlapAmount"] = []

        # interpolation:
        if "useInterpolation" in grain_metadata and grain_metadata["useInterpolation"]:
            encoded_data["useInterpolation"] = True
            encoded_data["grainInterpDistThreshold"] = grain_object.getVar(
                "grainInterpDistThreshold"
            )
            encoded_data["interpMethod"] = grain_object.getVar("interpMethod")
            pitchDimension = grain_object.getVar("pitchDimension")
            pitchDimension = pitchDimension.astype(bool)
            encoded_data["pitchDimension"] = pitchDimension.flatten()
            if (
                "useInterpolationResampling" in grain_metadata
                and grain_metadata["useInterpolationResampling"]
            ):
                encoded_data["useInterpolationResampling"] = True
            else:
                encoded_data["useInterpolationResampling"] = False

        else:
            encoded_data["useInterpolation"] = False
            encoded_data["grainInterpDistThreshold"] = []
            encoded_data["interpMethod"] = []
            encoded_data["pitchDimension"] = []

        return encoded_data
