# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

"""Count bits of an encoded token list."""

from mpegi.framework import Bitstream
from mpegi.framework.generic_codebook import (
    EncoderConfig,
    encode_symbols,
    serialize_codebook,
)


class TokenBitLengthCounter:
    """Count bits of an encoded token list."""

    def __init__(self) -> None:
        """Constructs a TokenBitLengthCounter."""
        self._bs = Bitstream()
        self._tokenList = list()

    def addBool(self, value: bool) -> None:
        """Add a boolean value to the token list and the bitstream."""
        self._bs.addBool(value)
        self._tokenList.append(value)

    def addInt(self, value: int, width: int) -> None:
        """Add an unsigned integer value to the token list and the bitstream."""
        self._bs.addInt(value, width)
        self._tokenList.append(value)

    def addSignedInt(self, value: int, width: int) -> None:
        """Add a signed integer value to the token list and the bitstream."""
        self._bs.addSignedInt(value, width)
        self._tokenList.append(value)

    def addFloat32(self, value: float) -> None:
        """Add a 32bit floating point value to the token list and the bitstream."""
        self._bs.addFloat32(value)
        self._tokenList.append(value)
        pass

    def addCodebook(self, config: EncoderConfig) -> None:
        """Add a codebook to the token list and the bitstream."""
        self._bs.appendBitstream(serialize_codebook(config))
        self._tokenList.append(config)

    def addSymbol(self, config: EncoderConfig, value: int) -> None:
        """Add a codebook symbol to the token list and the bitstream."""
        self._bs.appendBitstream(encode_symbols(config, [value]))
        self._tokenList.append(value)

    def append(self, other) -> None:
        """Append a TokenBitLengthCounter to the token list and the bitstream."""
        self._bs.appendBitstream(other.bs)
        self._tokenList.extend(other.tokenList)

    def getLength(self) -> int:
        """Returns the length of bitstream in bits."""
        return self._bs.getLength()

    def __lt__(self, other):
        """Bit length comparison: less than"""
        return self._bs.getLength() < other.bs.getLength()

    def __gt__(self, other):
        """Bit length comparison: greater than"""
        return self._bs.getLength() > other.bs.getLength()

    def __le__(self, other):
        """Bit length comparison: less than or equal"""
        return self._bs.getLength() <= other.bs.getLength()

    def __ge__(self, other):
        """Bit length comparison: greater than or equal"""
        return self._bs.getLength() >= other.bs.getLength()

    @property
    def bs(self) -> Bitstream:
        """Returns the encoded bitstream."""
        return self._bs

    @property
    def tokenList(self) -> list:
        """Returns the encoded bitstream."""
        return self._tokenList
