# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

"""Code Geometries into a binary blob."""

import logging
from collections import Counter
from enum import Enum
from typing import Tuple

import numpy as np

from mpegi.framework.generic_codebook import (
    SymbolHistogram,
    get_encoder_config,
)
from mpegi.coding._tokenBitLengthCounter import TokenBitLengthCounter

_bitsForNumVoxels = 32  # FIXME: this should be defined in mpegi.encoder._constants


class EncodingMode(Enum):
    none = 0
    x_axis = 1
    y_axis = 2
    z_axis = 3


class EncodedPathList:
    """Class for storing an encoded set of reflection sequences."""

    def __init__(self) -> None:
        """Constructs a VoxelDataEarlyEncoder."""

        self.bitCounter = TokenBitLengthCounter()
        self.counterNumPaths = Counter()
        self.counterEdgeId = Counter()
        self.counterPathId = Counter()
        self.counterIndicesRemoved = Counter()


class ValidPathDiffrEncoder:
    """Code geometries into a binary blob."""

    def __init__(self) -> None:
        """Constructs a VoxelDataEarlyEncoder."""
        pass

    def encode(self, voxel_dict: dict) -> TokenBitLengthCounter:
        # generate initial codebooks
        voxel_dict_sorted = _reorder_voxel_dict(voxel_dict)
        self._initial_codebooks(voxel_dict_sorted)

        for n in range(2):
            encoded_path_list = self._encode_iteration(voxel_dict_sorted)

        return encoded_path_list.bitCounter

    def _encode_iteration(self, voxel_dict_sorted: dict) -> EncodedPathList:
        # return value: EncodedPathList with token list, bitstream and symbol distribution
        ret = EncodedPathList()

        # cache for path lists
        voxel_cache = dict()

        # voxel index differences of neighbors
        delta_x = -np.array([1, 0, 0])
        delta_y = -np.array([0, 1, 0])
        delta_z = -np.array([0, 0, 1])
        delta_list = [0, delta_x, delta_y, delta_z]

        # add codebooks
        ret.bitCounter.addCodebook(self.encoder_vcX)
        ret.bitCounter.addCodebook(self.encoder_vcY)
        ret.bitCounter.addCodebook(self.encoder_vcZ)
        ret.bitCounter.addCodebook(self.encoder_NumPaths)
        ret.bitCounter.addCodebook(self.encoder_EdgeId)
        ret.bitCounter.addCodebook(self.encoder_PathId)
        ret.bitCounter.addCodebook(self.encoder_IndicesRemoved)

        # voxel coordinate declaration
        vcX = -1
        vcY = -1
        vcZ = -1

        # encode voxel data
        ret.bitCounter.addInt(len(voxel_dict_sorted), _bitsForNumVoxels)
        for vc, paths in voxel_dict_sorted.items():
            # encode voxel coordinate
            vcZ += 1
            has_voxel_coord_z = (vc[0] != vcX) or (vc[1] != vcY) or (vc[2] != vcZ)
            ret.bitCounter.addBool(has_voxel_coord_z)
            if has_voxel_coord_z:
                vcZ = vc[2]
                ret.bitCounter.addSymbol(self.encoder_vcZ, vcZ)

                vcY += 1
                has_voxel_coord_y = (vc[0] != vcX) or (vc[1] != vcY)
                ret.bitCounter.addBool(has_voxel_coord_y)
                if has_voxel_coord_y:
                    vcY = vc[1]
                    ret.bitCounter.addSymbol(self.encoder_vcY, vcY)

                    vcX += 1
                    has_voxel_coord_x = vc[0] != vcX
                    ret.bitCounter.addBool(has_voxel_coord_x)
                    if has_voxel_coord_x:
                        vcX = vc[0]
                        ret.bitCounter.addSymbol(self.encoder_vcX, vcX)

            encodedPaths = EncodedPathList()
            for mode in [
                EncodingMode.x_axis,
                EncodingMode.y_axis,
                EncodingMode.z_axis,
                EncodingMode.none,
            ]:
                vc_ref = tuple(np.array(vc) + delta_list[mode.value])
                if vc_ref in voxel_cache:
                    paths_ref = voxel_cache[vc_ref]
                else:
                    if mode is not EncodingMode.none:
                        continue  # skip this neighbor, if it does not exist in cache
                    paths_ref = []

                # try differential encoding variant
                encodedPaths_new = self._encode_path_list(mode, paths_ref, paths)
                if (
                    encodedPaths.bitCounter.getLength() == 0
                    or encodedPaths_new.bitCounter <= encodedPaths.bitCounter
                ):
                    encodedPaths = encodedPaths_new

            # add path list to cache
            voxel_cache[vc] = paths

            # add path list
            ret.bitCounter.append(encodedPaths.bitCounter)
            ret.counterNumPaths += encodedPaths.counterNumPaths
            ret.counterEdgeId += encodedPaths.counterEdgeId
            ret.counterPathId += encodedPaths.counterPathId
            ret.counterIndicesRemoved += encodedPaths.counterIndicesRemoved

        # tweak symbol distribution
        freqMapNumPaths = ret.counterNumPaths + Counter(range(self.max_num_paths + 1))
        freqMapEdgeId = ret.counterEdgeId
        freqMapPathId = ret.counterPathId
        freqMapIndicesRemoved = ret.counterIndicesRemoved + Counter(range(self.max_num_paths + 1))

        # update codebooks
        histogram_NumPaths = SymbolHistogram.from_histogram(freqMapNumPaths)
        histogram_EdgeId = SymbolHistogram.from_histogram(freqMapEdgeId)
        histogram_PathId = SymbolHistogram.from_histogram(freqMapPathId)
        histogram_IndicesRemoved = SymbolHistogram.from_histogram(freqMapIndicesRemoved)
        self.encoder_NumPaths = get_encoder_config(histogram_NumPaths)
        self.encoder_EdgeId = get_encoder_config(histogram_EdgeId)
        self.encoder_PathId = get_encoder_config(histogram_PathId)
        self.encoder_IndicesRemoved = get_encoder_config(histogram_IndicesRemoved)

        logging.info(
            "ValidPathDiffr:   _encode_iteration size = "
            + str(ret.bitCounter.getLength() / 8.0)
            + " bytes"
        )

        return ret

    def _encode_path_list(self, mode, paths_ref, paths) -> EncodedPathList:

        ret = EncodedPathList()

        ret.bitCounter.addInt(mode.value, 2)

        if mode is EncodingMode.none:
            listIndicesRemoved = []
            paths_added = paths
        else:
            # determine differences between the path lists
            indices_removed, indices_added = _list_diff(paths_ref, paths)
            paths_added = [paths[n] for n in indices_added]

            # encode list of removed path indices
            tmp = [-1] + indices_removed
            listIndicesRemoved = list(np.diff(tmp))
            listIndicesRemoved.append(0)  # indicates "end of list"
            for index in listIndicesRemoved:
                ret.bitCounter.addSymbol(self.encoder_IndicesRemoved, int(index))

        # encode list of added paths
        list_edge_id = []
        list_path_id = []
        num_paths = len(paths_added)
        ret.bitCounter.addSymbol(self.encoder_NumPaths, int(num_paths))
        for path_pair in paths_added:
            assert len(path_pair) == 2
            edge_id = path_pair[0]
            path_id = path_pair[1]
            ret.bitCounter.addSymbol(self.encoder_EdgeId, int(edge_id))
            ret.bitCounter.addSymbol(self.encoder_PathId, int(path_id))
            list_edge_id.append(edge_id)
            list_path_id.append(path_id)

        ret.counterNumPaths = Counter([num_paths])
        ret.counterEdgeId = Counter(list_edge_id)
        ret.counterPathId = Counter(list_path_id)
        ret.counterIndicesRemoved = Counter(listIndicesRemoved)

        return ret

    def _initial_codebooks(self, voxel_dict_sorted: dict):
        list_vc_x = list()
        list_vc_y = list()
        list_vc_z = list()
        list_num_paths = [0]
        list_edge_id = list()
        list_path_id = list()

        # initialize voxel coordinate lists
        vcX = -1
        vcY = -1
        vcZ = -1
        for vc in voxel_dict_sorted.keys():
            vcZ += 1
            has_voxel_coord_z = (vc[0] != vcX) or (vc[1] != vcY) or (vc[2] != vcZ)
            if has_voxel_coord_z:
                vcZ = vc[2]
                list_vc_z.append(vcZ)

                vcY += 1
                has_voxel_coord_y = (vc[0] != vcX) or (vc[1] != vcY)
                if has_voxel_coord_y:
                    vcY = vc[1]
                    list_vc_y.append(vcY)

                    vcX += 1
                    has_voxel_coord_x = vc[0] != vcX
                    if has_voxel_coord_x:
                        vcX = vc[0]
                        list_vc_x.append(vcX)

        # initialize num_paths, edge_id, and path_id lists
        for vc, paths in voxel_dict_sorted.items():
            num_paths = len(paths)
            list_num_paths.append(num_paths)
            for path_pair in paths:
                assert len(path_pair) == 2
                edge_id = path_pair[0]
                path_id = path_pair[1]
                list_edge_id.append(edge_id)
                list_path_id.append(path_id)

        self.max_num_paths = max(list_num_paths)

        # enforce 1 bit for symbol '0'
        list_indices_removed = [0] * self.max_num_paths + list(range(self.max_num_paths + 1))

        # generate histograms
        histogram_vc_x = SymbolHistogram.from_plain(list_vc_x)
        histogram_vc_y = SymbolHistogram.from_plain(list_vc_y)
        histogram_vc_z = SymbolHistogram.from_plain(list_vc_z)
        histogram_num_paths = SymbolHistogram.from_plain(list(range(self.max_num_paths + 1)))
        histogram_edge_id = SymbolHistogram.from_plain(list_edge_id)
        histogram_path_id = SymbolHistogram.from_plain(list_path_id)
        histogram_indices_removed = SymbolHistogram.from_plain(list_indices_removed)

        # initialize encoder configs
        self.encoder_vcX = get_encoder_config(histogram_vc_x)
        self.encoder_vcY = get_encoder_config(histogram_vc_y)
        self.encoder_vcZ = get_encoder_config(histogram_vc_z)
        self.encoder_NumPaths = get_encoder_config(histogram_num_paths)
        self.encoder_EdgeId = get_encoder_config(histogram_edge_id)
        self.encoder_PathId = get_encoder_config(histogram_path_id)
        self.encoder_IndicesRemoved = get_encoder_config(histogram_indices_removed)


def _list_diff(list_1, list_2) -> Tuple[list, list]:
    indices_removed = list()
    indices_added = list()
    n2 = 0

    for n1, seq_1 in enumerate(list_1):
        while n2 < len(list_2) and list_2[n2] < seq_1:
            indices_added.append(n2)
            n2 += 1
        if n2 < len(list_2) and list_2[n2] == seq_1:
            n2 += 1
        else:
            indices_removed.append(n1)

    # add remaining elements from list_2
    while n2 < len(list_2):
        indices_added.append(n2)
        n2 += 1

    return indices_removed, indices_added


def _reorder_voxel_dict(voxel_dict: dict) -> dict:

    voxel_dict_sorted = {}
    for string in voxel_dict.keys():
        vc = eval(string)
        voxel_dict_sorted[vc] = sorted(voxel_dict[string])

    return voxel_dict_sorted
