# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

"""Code Geometries into a binary blob."""

import logging
from collections import Counter, OrderedDict
from enum import Enum

import numpy as np

from mpegi.framework import Bitstream
from mpegi.framework.generic_codebook import (
    SymbolHistogram,
    encode_symbols,
    get_encoder_config,
    serialize_codebook,
)
from mpegi.metadata.early.voxelDataEarlyDict import VoxelDataEarlyDict
from mpegi.voxel import VoxelGrid, voxelCoordinateToVoxelIndex

_bitsForNumVoxelPairs = 32  # FIXME: this should be defined in mpegi.encoder._constants


class EncodingMode(Enum):
    none = 0
    x_axis = 1
    y_axis = 2
    z_axis = 3


class EncodedPathList:
    """Class for storing an encoded set of reflection sequences."""

    def __init__(self) -> None:
        """Constructs a VoxelDataEarlyEncoder."""

        self.bs = Bitstream()
        self.counterNumPaths = Counter()
        self.counterOrder = Counter()
        self.counterSurf = Counter()
        self.counterIndicesRemoved = Counter()


class VoxelDataEarlyEncoder:
    """Code geometries into a binary blob."""

    def __init__(self) -> None:
        """Constructs a VoxelDataEarlyEncoder."""
        pass

    def encode(self, vde: VoxelDataEarlyDict, voxel_grid: VoxelGrid) -> Bitstream:
        # generate initial codebooks
        vde_sorted = _reorder_voxel_data(vde)
        self._initial_codebooks(vde_sorted)

        for n in range(2):
            encoded_path_list = self._encode_iteration(vde_sorted, voxel_grid)

        return encoded_path_list.bs

    def _encode_iteration(
        self, vde_sorted: VoxelDataEarlyDict, voxel_grid: VoxelGrid
    ) -> EncodedPathList:
        # return value: EncodedPathList with bitstream and symbol distribution
        ret = EncodedPathList()

        # cache for path lists
        path_cache = dict()

        # voxel index differences of neighbors
        delta_x = -voxelCoordinateToVoxelIndex(voxel_grid, [1, 0, 0])
        delta_y = -voxelCoordinateToVoxelIndex(voxel_grid, [0, 1, 0])
        delta_z = -voxelCoordinateToVoxelIndex(voxel_grid, [0, 0, 1])
        delta_list = [0, delta_x, delta_y, delta_z]

        # add codebooks
        ret.bs.appendBitstream(serialize_codebook(self.encoderL))
        ret.bs.appendBitstream(serialize_codebook(self.encoderS))
        ret.bs.appendBitstream(serialize_codebook(self.encoderNumPaths))
        ret.bs.appendBitstream(serialize_codebook(self.encoderOrder))
        ret.bs.appendBitstream(serialize_codebook(self.encoderSurf))
        ret.bs.appendBitstream(serialize_codebook(self.encoderIndicesRemoved))

        # encode voxel data
        ret.bs.addInt(len(vde_sorted.lut), _bitsForNumVoxelPairs)
        for (l, s), path_set in vde_sorted.lut.items():
            ret.bs.appendBitstream(encode_symbols(self.encoderL, [l]))
            ret.bs.appendBitstream(encode_symbols(self.encoderS, [s]))
            path_list = sorted(list(path_set), key=lambda seq: (len(seq), seq))

            encodedPaths = EncodedPathList()
            for mode in [
                EncodingMode.x_axis,
                EncodingMode.y_axis,
                EncodingMode.z_axis,
                EncodingMode.none,
            ]:
                l_ref = l + delta_list[mode.value]
                if l_ref in path_cache:
                    path_list_ref = path_cache[l_ref]
                else:
                    if mode is not EncodingMode.none:
                        continue  # skip this neighbor, if it does not exist in cache
                    path_list_ref = []

                # try differential encoding variant
                encodedPaths_new = self._encode_path_list(mode, path_list_ref, path_list)
                if len(encodedPaths.bs) == 0 or len(encodedPaths_new.bs) < len(encodedPaths.bs):
                    encodedPaths = encodedPaths_new

            # add path list to cache
            path_cache[l] = path_list

            # add path list
            ret.bs.appendBitstream(encodedPaths.bs)
            ret.counterNumPaths += encodedPaths.counterNumPaths
            ret.counterOrder += encodedPaths.counterOrder
            ret.counterSurf += encodedPaths.counterSurf
            ret.counterIndicesRemoved += encodedPaths.counterIndicesRemoved

        # tweak symbol distribution
        freqMapNumPaths = ret.counterNumPaths + Counter(range(self.max_num_paths + 1))
        freqMapOrder = ret.counterOrder + Counter(range(self.max_order + 1))
        freqMapSurf = ret.counterSurf
        freqMapIndicesRemoved = ret.counterIndicesRemoved + Counter(range(self.max_num_paths + 1))

        # update codebooks
        histogram_NumPaths = SymbolHistogram.from_histogram(freqMapNumPaths)
        histogram_Order = SymbolHistogram.from_histogram(freqMapOrder)
        histogram_Surf = SymbolHistogram.from_histogram(freqMapSurf)
        histogram_IndicesRemoved = SymbolHistogram.from_histogram(freqMapIndicesRemoved)
        self.encoderNumPaths = get_encoder_config(histogram_NumPaths)
        self.encoderOrder = get_encoder_config(histogram_Order)
        self.encoderSurf = get_encoder_config(histogram_Surf)
        self.encoderIndicesRemoved = get_encoder_config(histogram_IndicesRemoved)

        logging.info("ER:   _encode_iteration size = " + str(len(ret.bs) / 8.0) + " bytes")

        return ret

    def _encode_path_list(self, mode, path_list_ref, path_list) -> EncodedPathList:

        ret = EncodedPathList()
        ret.bs.addInt(mode.value, 2)

        if mode is EncodingMode.none:
            listIndicesRemoved = []
            paths_added = path_list
        else:
            # determine differences between the path lists
            indices_removed, indices_added = _list_diff(path_list_ref, path_list)
            paths_added = [path_list[n] for n in indices_added]

            # encode list of removed path indices
            tmp = [-1] + indices_removed
            listIndicesRemoved = list(np.diff(tmp))
            listIndicesRemoved.append(0)  # indicates "end of list"
            for index in listIndicesRemoved:
                ret.bs.appendBitstream(encode_symbols(self.encoderIndicesRemoved, [index]))

        # encode list of added paths
        listOrder = list()
        listSurf = list()
        num_paths = len(paths_added)
        ret.bs.appendBitstream(encode_symbols(self.encoderNumPaths, [num_paths]))
        for seq_surf in paths_added:
            order = len(seq_surf)
            ret.bs.appendBitstream(encode_symbols(self.encoderOrder, [order]))
            listOrder.append(order)
            for surf in seq_surf:
                ret.bs.appendBitstream(encode_symbols(self.encoderSurf, [surf]))
                listSurf.append(surf)

        ret.counterNumPaths = Counter([num_paths])
        ret.counterOrder = Counter(listOrder)
        ret.counterSurf = Counter(listSurf)
        ret.counterIndicesRemoved = Counter(listIndicesRemoved)

        return ret

    def _initial_codebooks(self, vde: VoxelDataEarlyDict):
        list_l = list()
        list_s = list()
        list_num_paths = [0]
        list_order = [0]
        list_surf = list()
        for (l, s), path_set in vde.lut.items():
            list_l.append(l)
            list_s.append(s)
            num_paths = len(path_set)
            list_num_paths.append(num_paths)
            for seq_surf in path_set:
                order = len(seq_surf)
                list_order.append(order)
                for surf in list(seq_surf):
                    list_surf.append(surf)

        self.max_num_paths = max(list_num_paths)
        self.max_order = max(list_order)

        # enforce 1 bit for symbol '0'
        list_indices_removed = [0] * self.max_num_paths + list(range(self.max_num_paths + 1))

        histogram_L = SymbolHistogram.from_plain(list_l)
        histogram_S = SymbolHistogram.from_plain(list_s)
        histogram_NumPaths = SymbolHistogram.from_plain(list(range(self.max_num_paths + 1)))
        histogram_Order = SymbolHistogram.from_plain(list(range(self.max_order + 1)))
        histogram_Surf = SymbolHistogram.from_plain(list_surf)
        histogram_IndicesRemoved = SymbolHistogram.from_plain(list_indices_removed)

        self.encoderL = get_encoder_config(histogram_L)
        self.encoderS = get_encoder_config(histogram_S)
        self.encoderNumPaths = get_encoder_config(histogram_NumPaths)
        self.encoderOrder = get_encoder_config(histogram_Order)
        self.encoderSurf = get_encoder_config(histogram_Surf)
        self.encoderIndicesRemoved = get_encoder_config(histogram_IndicesRemoved)


def _list_diff(list_1, list_2) -> (list, list):
    indices_removed = list()
    indices_added = list()
    n2 = 0

    for n1, seq_1 in enumerate(list_1):
        while n2 < len(list_2) and list_2[n2] < seq_1:
            indices_added.append(n2)
            n2 += 1
        if n2 < len(list_2) and list_2[n2] == seq_1:
            n2 += 1
        else:
            indices_removed.append(n1)

    # add remaining elements from list_2
    while n2 < len(list_2):
        indices_added.append(n2)
        n2 += 1

    return indices_removed, indices_added


def _reorder_voxel_data(vde: VoxelDataEarlyDict) -> VoxelDataEarlyDict:

    order_sl = OrderedDict()
    for (l, s) in vde.lut.keys():
        if s not in order_sl:
            order_sl[s] = list()
        order_sl[s].append(l)

    vde_sorted = VoxelDataEarlyDict()
    for s in sorted(order_sl.keys()):
        for l in sorted(order_sl[s]):
            if not vde.lut[(l, s)]:
                vde_sorted.append(l, s, [])
                continue
            for seq_surf in vde.lut[(l, s)]:
                vde_sorted.append(l, s, seq_surf)

    return vde_sorted
