# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import logging
import pathlib
import time
from typing import List, Optional, Type

from mpegi.cli import Command, Config
from mpegi.encoder import SceneConfigPacket, ScenePayloadPacket, SceneUpdatePacket, components
from mpegi.framework import Bitstream, Component, ComponentBuilder, Inventory, build_packets


import xml.etree.ElementTree as ET


class EncodeCommand(Command):
    @staticmethod  # Inherited from cli.Command
    def getCommandName() -> str:
        return "encode"

    @staticmethod  # Inherited from cli.Command
    def execute(config: Config) -> None:
        start = time.time()

        version = 30

        _printAndLog(
            f"Encoding: EIF File (.xml) {config['scene']} -> MPEG-I v{version} bitstream (.bit)"
        )

        components_to_build: List[Type[Component]] = [
            components.SceneComponent,
            components.ScenePlusComponent,
            components.DirectivityComponent,
            components.ReverbComponent,
            components.BitstreamIdComponent,
            components.PreprocessorComponent,
            components.ERComponent,
            components.DiffractionComponent,
            components.AudioplusComponent,
            components.VoxelSceneComponent,
            components.FreqGridComponent,
            components.AirflowComponent,
            components.RasterMapComponent,
            components.GranularDatabaseComponent,
            components.LCERComponent,
        ]

        _preprocessEIF(config)

        if config["disableER"] or config["useLowComplexityER"]:
            components_to_build.pop(components_to_build.index(components.ERComponent))

        if config["disableDIFFR"]:
            components_to_build.pop(components_to_build.index(components.DiffractionComponent))

        if not config["useLowComplexityER"]:
            components_to_build.pop(components_to_build.index(components.LCERComponent))

        component_builder = ComponentBuilder(components_to_build)
        scene_name = config["scene"].stem
        cache_directory = pathlib.Path(__file__).parents[3].joinpath(".cache", scene_name)
        inventory = Inventory(config, str(cache_directory))
        component_builder.process(inventory, config["force"])

        config_packet = build_packets(packet_to_build=SceneConfigPacket, inv=inventory)
        payload_packet = build_packets(packet_to_build=ScenePayloadPacket, inv=inventory)
        update_packets = build_packets(packet_to_build=SceneUpdatePacket, inv=inventory)
        packets = [*config_packet, *payload_packet, *update_packets]

        mpegi_bstream = Bitstream()
        mpegi_bstream.addString("MPEGI")
        mpegi_bstream.addInt(version, 16)
        for packet in packets:
            mpegi_bstream.appendBitstream(packet)

        output_file: Optional[pathlib.Path] = config["output"]
        out_file: pathlib.Path
        if output_file is None:
            out_file = config["scene"].with_suffix(".bit")
        elif output_file.is_dir():
            out_file = output_file.joinpath(f'{config["scene"].stem}.bit')
        else:
            out_file = output_file

        mpegi_bstream.writeToFile(str(out_file))
        _printAndLog(f"Finished! Time taken (s): {time.time() - start:.3f}")
        _printAndLog(f"Wrote bitstream file to {out_file.resolve()}")

    @staticmethod  # Inherited from cli.Command
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        def componentForceList(string: str):
            if string == "":
                return None
            elif string == "all":
                return []
            else:
                return [cName for cName in string.split(",")]

        argParser.add_argument(
            "--disableDIFFR", action="store_true", help="Disable Diffraction Component"
        )
        argParser.add_argument(
            "--disableER", action="store_true", help="Disable Early Reflection Component"
        )
        argParser.add_argument(
            "--useLowComplexityER",
            action="store_true",
            help="Encode Bitstream using Low Complexity Early Reflections",
        )

        argParser.add_argument(
            "-f",
            "--force",
            default="",
            type=componentForceList,
            metavar="COMPONENT_LIST",
            help=(
                "Force construction of all or some encoder components, "
                "regardless if the cache exists on the disk. A list of comma "
                "seperated class names are expected. If 'all' is given, then "
                "all components are forcefully built."
            ),
        )
        argParser.add_argument(
            "-o",
            "--output",
            metavar="PATH",
            type=pathlib.Path,
            default=None,
            help=(
                "Path to output file. If an output file is not specified, the "
                "output file will be written in the same place as the scene file"
            ),
        )

        argParser.add_argument(
            "--serializeURR",
            action="store_true",
            help="If true, user reachable region will be serialize into bitstream.",
        )

        SceneConfigPacket.register_arguments(argParser)
        components.SceneComponent.registerArguments(argParser)
        components.ScenePlusComponent.registerArguments(argParser)
        components.DirectivityComponent.registerArguments(argParser)
        components.PreprocessorComponent.registerArguments(argParser)
        components.ReverbComponent.registerArguments(argParser)
        components.DiffractionComponent.registerArguments(argParser)
        components.ERComponent.registerArguments(argParser)
        components.VoxelSceneComponent.registerArguments(argParser)
        components.RasterMapComponent.registerArguments(argParser)
        components.GranularDatabaseComponent.registerArguments(argParser)
        components.LCERComponent.registerArguments(argParser)


def _printAndLog(msg: str) -> None:
    print(msg)
    logging.info(msg)


def _preprocessEIF(config: Config) -> None:
    root = ET.parse(config["scene"]).getroot()
    voxel_DIFFR_enabled = len(root.findall("VoxelScene")) > 0
    faces_with_materials = len(root.findall("Mesh/Face[@material]")) > 0
    voxel_ER_enabled = voxel_DIFFR_enabled and not faces_with_materials

    if voxel_DIFFR_enabled:
        logging.info(f"Voxel based diffraction enabled")
        if not config["disableDIFFR"]:
            config.addArg("--disableDIFFR")
    if voxel_ER_enabled:
        logging.info(f"Voxel based early reflection enabled")
        if not config["disableER"]:
            config.addArg("--disableER")
