# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import pathlib
from io import BufferedReader, BytesIO
from typing import List, Tuple

from mpegi.cli import Command, Config
from mpegi.framework import MHASPacketType, ScenePayloadPacketType


class StatCommand(Command):
    @staticmethod  # Inherited from cli.Command
    def execute(config: Config) -> None:
        bs_path: pathlib.Path = config["file"]
        assert bs_path.suffix == ".bit", "File must end in .bit"

        with open(bs_path, "rb") as bs_file:
            if bs_file.read1(5) != "MPEGI".encode():
                raise Exception(f"{bs_path} is not an MPEGI Bitstream")
            bs_file.read1(1)  # Skip the null terminated byte

            version = int.from_bytes(bs_file.read1(2), "big")

            payload_packets: List[bytes] = []
            update_packets: List[bytes] = []
            config_packets: List[bytes] = []
            while not StatCommand.isEOF(bs_file):
                packet_name, packet = StatCommand.read_packet(bs_file)
                if packet_name == "PACTYP_PLD":
                    payload_packets.append(packet)
                elif packet_name == "PACTYP_UPD":
                    update_packets.append(packet)
                elif packet_name == "PACTYP_CFG":
                    config_packets.append(packet)
                else:
                    raise ValueError(f"Unrecognized Packet Type: {packet_name}")

        # -------- FORMATTING ---------- #
        table_width = 80
        first_col_width = 16
        second_col_width = 40  # table_width - first_col_width
        third_col_width = table_width - first_col_width - second_col_width
        widths = (first_col_width, second_col_width, third_col_width)
        print("BITSTREAM STATUS".center(table_width))
        print("")
        print(f"FILE PATH: {bs_path}")
        print(f"VERSION: {version}")
        print(f"NUM OF CONFIG PACKETS: {len(config_packets)}")
        print(f"NUM OF UPDATE PACKETS: {len(update_packets)}")
        print(f"NUM OF PAYLOAD PACKETS: {len(payload_packets)}")

        print("/" + ("-" * (table_width - 2)) + "\\")
        StatCommand.print_three_columns(widths, ("PACKET TYPE", "PACKET CONTENTS", "SIZE (BYTES)"))
        for payload in payload_packets:
            print("|" + ("-" * (table_width - 2)) + "|")
            payload_elements, packet_lengths = StatCommand.read_payloads(BytesIO(payload))
            StatCommand.print_three_columns(widths, ("PACTYP_PLD", "PAYLOADS:", ""))
            # for p_ele in payload_elements:
            for k in range(0, len(payload_elements)):
                StatCommand.print_three_columns(
                    widths, ("", payload_elements[k], packet_lengths[k])
                )
            # StatCommand.print_two_columns(widths, ("", element))
            print("\\" + ("-" * (table_width - 2)) + "/")
            print("")

    @staticmethod  # Inherited from cli.Command
    def getCommandName() -> str:
        return "stat"

    @staticmethod  # Inherited from cli.Command
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        argParser.add_argument("file", metavar="PATH", type=pathlib.Path, help="Path to .bit file")

    @staticmethod
    def isEOF(file: BufferedReader) -> bool:
        """Returns True if at end of file."""
        return True if file.peek(1) == "".encode() else False

    @staticmethod
    def read_packet(file: BufferedReader) -> Tuple[str, bytes]:
        packet_header_length = 9
        packet_header = file.read1(packet_header_length)
        packet_type = int.from_bytes(packet_header, "big") >> 61
        packet_type_name = MHASPacketType(packet_type).name

        mask_length = 13
        mask = int("1" * mask_length, 2) << ((packet_header_length * 8) - mask_length)
        packet_length = int.from_bytes(packet_header, "big") & ~(mask)
        return (packet_type_name, file.read(packet_length))

    @staticmethod
    def read_payloads(file: BytesIO) -> List[str]:
        packet_elements: List[str] = []
        packet_lengths: List[str] = []
        file.read(2)  # Skip ID
        num_of_payloads = int.from_bytes(file.read(2), "big")
        for _ in range(num_of_payloads):
            payload_header_length = 9
            payload_header = file.read(payload_header_length)
            payload_type = int.from_bytes(payload_header, "big") >> 61
            packet_type_name = ScenePayloadPacketType(payload_type).name
            packet_elements.append(packet_type_name)

            mask_length = 13
            mask = int("1" * mask_length, 2) << ((payload_header_length * 8) - mask_length)
            payload_length = int.from_bytes(payload_header, "big") & ~(mask)
            packet_lengths.append(str(payload_length))
            file.read(payload_length)  # Skip rest of the payload

        return packet_elements, packet_lengths

    @staticmethod
    def print_three_columns(
        col_widths: Tuple[int, int, int], col_content: Tuple[str, str, str]
    ) -> None:
        print(
            "|"
            + col_content[0].center(col_widths[0] - 2)
            + "|"
            + col_content[1].center(col_widths[1] - 1)
            + "|"
            + col_content[2].center(col_widths[2] - 1)
            + "|"
        )
