# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from pathlib import Path
from typing import Any, Tuple

import numpy as np
import numpy.typing as npt
from scipy.io import wavfile


def convert_32_to_float(data_int32: npt.NDArray[np.int32]) -> npt.NDArray[np.float64]:
    """Convert int32 numppy array to float64 with range -1.0 - +1.0."""
    data_float64: npt.NDArray[np.float64] = data_int32 / 2 ** 31
    data_float64 = data_float64.astype(np.float64)
    return data_float64


def convert_16_to_float(data_int16: npt.NDArray[np.int16]) -> npt.NDArray[np.float64]:
    """Convert int16 numppy array to float64 with range -1.0 - +1.0."""
    data_float64: npt.NDArray[np.float64] = data_int16 / 2 ** 15
    data_float64 = data_float64.astype(np.float64)
    return data_float64


def convert_u8_to_float(data_uint8: npt.NDArray[np.uint8]) -> npt.NDArray[np.float64]:
    """Convert uint8 numppy array to float64 with range -1.0 - +1.0."""
    data_int8: npt.NDArray[np.int8] = data_uint8 - 2 ** 7
    data_float64: npt.NDArray[np.float64] = data_int8 / 2 ** 7
    data_float64 = data_float64.astype(np.float64)
    return data_float64


def read_wav(file: str) -> Tuple[int, npt.NDArray[np.float64]]:
    """Read a wav file and return its data.

    :return: A tuple where the first element is sample rate and second
        is the sample data
    """
    file = Path(file).resolve(True)
    read_file = wavfile.read(file)
    sample_rate: int = read_file[0]
    data: npt.NDArray[Any] = read_file[1]

    if str(data.dtype) == "float32":
        return (sample_rate, data)
    elif str(data.dtype) == "int32":
        return (sample_rate, convert_32_to_float(data))
    elif str(data.dtype) == "int16":
        return (sample_rate, convert_16_to_float(data))
    elif str(data.dtype) == "uint8":
        return (sample_rate, convert_u8_to_float(data))
    else:
        raise TypeError(f"Wavfile with datatype '{data.dtype}' not supported")
