# This software module was originally developed by
#
# Fraunhofer IIS and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS and Koninklijke Philips N.V. retain full right to modify
# and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import argparse
from enum import Enum
from typing import Any, Dict, List

from mpegi.framework import Bitstream, BitstreamComposer, Inventory, MHASPacketType, Packet, tokens
from . import _constants

humidity_default = 40
temperature_default = 20
speed_of_sound_default = 340
culling_threshold = {-100: 0, -90: 1, -80: 2, -70: 3, -60: 4, -50: 5, -40: 6, -30: 7}
distance_max_delay = {
    10000: 0,
    20000: 1,
    30000: 2,
    50000: 3,
    100000: 4,
    200000: 5,
    500000: 6,
    1000000: 7,
}
vdl_method_default = "spline"

FreqGridData: List[tokens.Token] = [
    tokens.writeCountOrIndex("fgdNrGrids"),
    tokens.Repeat(
        "fgdNrGrids",
        syntax=[
            tokens.UnsignedInteger("fgdMethod", width=2),
            tokens.Switch(
                "fgdMethod",
                cases=[
                    tokens.SwitchCase(
                        0,
                        syntax=[
                            tokens.writeCountOrIndex("fgdNrBands"),
                            tokens.Repeat(
                                "fgdNrBands", syntax=[tokens.writeFrequency("fgdCenterFreq")]
                            ),
                        ],
                    ),  # individual freq
                    tokens.SwitchCase(
                        1,
                        syntax=[
                            tokens.writeCountOrIndex("fgdNrBands"),
                            tokens.writeFrequency("fgdCenterFreq"),
                            tokens.writeValueFromLUT(
                                "frequencyHop", tableName="frequency_hop_codes"
                            ),
                        ],
                    ),  # start-hop
                    tokens.SwitchCase(
                        2,
                        syntax=[
                            tokens.UnsignedInteger("fgdDefaultGrid", width=4),
                            tokens.Flag("fgdIsSubGrid"),
                            tokens.Conditional(
                                "fgdIsSubGrid",
                                false=[],
                                true=[
                                    tokens.UnsignedInteger("fgdDefaultGridOffset", width=3),
                                    tokens.UnsignedInteger("fgdDefaultGridNrBands", width=6),
                                ],
                            ),
                        ],
                    ),  # Default Banding
                    tokens.SwitchCase(3, syntax=[]),  # reserved
                ],
            ),
        ],
    ),
]

early_tuning_parameters: List[tokens.Token] = [
    tokens.Flag("hasTuningGainER"),
    tokens.Conditional(
        "hasTuningGainER",
        true=[
            tokens.UnsignedInteger("tuningGainER", width=6),
        ],
        false=[],
    ),
    tokens.Flag("useLowComplexityER"),
]

HeadphoneData: List[tokens.Token] = [
    tokens.UnsignedInteger("specificationType", width=3),
    tokens.Switch(
        "specificationType",
        cases=[
            tokens.SwitchCase(
                0,
                syntax=[
                    tokens.writeCountOrIndex("freqGridIdx"),
                    tokens.Repeat(
                        "headphoneNrBands",
                        syntax=[
                            tokens.writeGain("ccHpGain", isHiPrecGain=True),
                        ],
                    ),
                ],
            ),
            tokens.SwitchCase(
                1,
                syntax=[
                    tokens.writeCountOrIndex("freqGridIdx"),
                    tokens.Repeat(
                        "headphoneNrBands",
                        syntax=[
                            tokens.UnsignedInteger("ccHpParametircFilterType", width=2),
                            tokens.writeGain("ccHpGain", isHiPrecGain=True),
                            tokens.QuantizedFloat("ccHpQ", quant_range=_constants.ccHpQ_qrange),
                        ],
                    ),
                ],
            ),
            tokens.SwitchCase(2, syntax=[tokens.UnsignedInteger("LUTIndex", width=7)]),
        ],
    ),
]

loudness_culling_parameters: List[tokens.Token] = [
    tokens.Flag("hasLoudnessCullingThreshold"),
    tokens.Conditional(
        "hasLoudnessCullingThreshold",
        true=[
            tokens.UnsignedInteger("loudnessCullingThreshold", width=6),
        ],
        false=[],
    ),
]
write_config_packet = tokens.Function(
    "writeConfigPacket",
    syntax=[
        tokens.writeCountOrIndex("entityCount"),
        tokens.Repeat(
            "entityCount",
            syntax=[
                tokens.writeID("integerId"),
                tokens.CString("stringId"),
            ],
        ),
        tokens.UnsignedInteger(
            "delayBufferSize",
            width=3,
            code_page=distance_max_delay,
        ),
        tokens.UnsignedInteger(
            "gainCullingThreshold",
            width=3,
            code_page=culling_threshold,
        ),
        tokens.Flag("enableCullingReflectionRI"),
        tokens.Flag("overrideSpeedOfSound"),
        tokens.Conditional(
            "overrideSpeedOfSound",
            false=[],
            true=[tokens.UnsignedInteger("speedOfSound", width=13)],
        ),
        tokens.Flag("overrideTemperature"),
        tokens.Conditional(
            "overrideTemperature",
            false=[],
            true=[tokens.UnsignedInteger("temperature", width=5)],
        ),
        tokens.Flag("overrideHumidity"),
        tokens.Conditional(
            "overrideHumidity",
            false=[],
            true=[tokens.UnsignedInteger("humidity", width=4)],
        ),
        tokens.Flag("isSmallScene"),
        *FreqGridData,
        *early_tuning_parameters,
        tokens.Flag("hasHeadphoneEQ"),
        tokens.Conditional("hasHeadphoneEQ", false=[], true=[*HeadphoneData]),
        *loudness_culling_parameters,
        tokens.Flag("hasConsolidationThreshold"),
        tokens.Conditional(
            "hasConsolidationThreshold",
            false=[],
            true=[tokens.UnsignedInteger("consolidationThreshold", width=5)],
        ),
        tokens.Flag("enableLocallyCapturedAudio"),
        tokens.Conditional(
            "enableLocallyCapturedAudio",
            true=[
                tokens.Flag("overrideListenerVoiceDirectivity"),
                tokens.Conditional(
                    "overrideListenerVoiceDirectivity",
                    false=[],
                    true=[tokens.writeID("listenerVoiceDirectivityId")],
                ),
            ],
            false=[],
        ),
    ],
)


class FGDMethod(Enum):
    INDIVIDUAL_FREQUENCIES = 0
    START_HOP_AMOUNT = 1
    DEFAULT_BANDING = 2
    RESERVED = 3


class HeadphoneSpecType(Enum):
    freqgain = 0
    parametricfilters = 1
    preset = 2


class HeadphoneFilterType(Enum):
    lowshelf = 0
    peak = 1
    highshelf = 2


class SceneConfigPacket(Packet):
    """Packet to configure the scene via the bitstream."""

    def __init__(self, args: Dict[str, Any]) -> None:
        """Construct a SceneConfigPacket."""
        self.delay_buffer_size = args["delayBufferSize"]
        self.culling_threshold = args["gainCullingThreshold"]
        self.loudness_culling_threshold = args["loudnessCullingThreshold"]
        self.humidity = args["humidity"]
        self.temperature = args["temperature"]
        self.speed_of_sound = args["speedOfSound"]
        self.use_low_complexity_er = args["useLowComplexityER"]
        if args["disableCullingReflectionRI"] or args["disableER"]:
            self.cullingReflectionRI = False
        else:
            self.cullingReflectionRI = True
        self.consolidation_threshold = args["consolidationThreshold"]

    @staticmethod
    def get_packet_type() -> MHASPacketType:
        """Return PACTYPE_CFG MHAS structure."""
        return MHASPacketType.PACTYP_CFG

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        """Return a list of all required arguments."""
        return [
            "delayBufferSize",
            "gainCullingThreshold",
            "disableCullingReflectionRI",
            "loudnessCullingThreshold",
            "humidity",
            "temperature",
            "speedOfSound",
            "disableER",
            "useLowComplexityER",
            "consolidationThreshold",
        ]

    @staticmethod
    def register_arguments(arg_parser: argparse.ArgumentParser) -> None:
        config_parser = arg_parser.add_argument_group("Renderer Config")
        config_parser.add_argument(
            "--delayBufferSize",
            default=200000,
            type=int,
            choices=distance_max_delay.keys(),
            metavar="SAMPLES",
            help="Maximum size of buffers in samples for delay propogation.",
        )

        config_parser.add_argument(
            "--gainCullingThreshold",
            default=-30,
            type=int,
            choices=culling_threshold.keys(),
            metavar="DECIBEL",
            help="Cull all sources below given threshold in decibels.",
        )

        config_parser.add_argument(
            "--disableCullingReflectionRI",
            action="store_true",
            help="Disable Culling ReflectionRI (This flag is automatically turned on using the option --disableER)",
        )

        config_parser.add_argument(
            "--loudnessCullingThreshold",
            default=0,
            type=int,
            choices=range(-73, -9),
            metavar="DECIBEL",
            help="Threshold for loudness-based culling in decibels.",
        )

        config_parser.add_argument(
            "--humidity",
            default=humidity_default,
            type=int,
            metavar="%(0-100)",
            help="The value of humidity used for medium attenuation",
        )

        config_parser.add_argument(
            "--temperature",
            default=temperature_default,
            type=int,
            metavar="CELSIUS (-50 - +105)",
            help="The value of the temperature for medium attenuation.",
        )

        config_parser.add_argument(
            "--speedOfSound",
            default=speed_of_sound_default,
            type=int,
            metavar="m/s (Meters Per Second)",
            help="The value of the speed of sound for propogation calculation delay.",
        )

        config_parser.add_argument(
            "--consolidationThreshold",
            default=5,
            type=int,
            metavar="PERCEPTUALDISTANCE (0 - 31)",
            help="Threshold for consolidation of Render Items (default=5)",
        )

    def serialize(self, inventory: Inventory) -> List[Bitstream]:
        """Serialize the Scene Config packet payload.

        .. note::
            Currently, the scene config payload is left blank. It will be
            fleshed out in the future once subscene decomposition is
            investigated further.
        """
        composer = BitstreamComposer([write_config_packet])

        bs_ids = inventory.getArtifacts(["bs_ids"])["bs_ids"]
        composer.push(bs_ids.numOfItems)
        for string_id, int_id in bs_ids.anyToIntIdMap.items():
            composer.push(int_id)
            composer.push(string_id)

        composer.push(self.delay_buffer_size)
        composer.push(self.culling_threshold)
        composer.push(self.cullingReflectionRI)
        override_speed_of_sound = self.speed_of_sound != speed_of_sound_default
        composer.push(override_speed_of_sound)
        if override_speed_of_sound:
            composer.push(self.speed_of_sound)

        override_temperature = self.temperature != temperature_default
        composer.push(override_temperature)
        if override_temperature:
            composer.push(round((self.temperature + 50) / 5))

        override_humidity = self.humidity != humidity_default
        composer.push(override_humidity)
        if override_humidity:
            composer.push(round(self.humidity / 6.667))

        scene_info = inventory.artifacts["scene_info"]
        composer.push(scene_info["isSmallScene"])

        self.writeFreqGridData(inventory.artifacts["frequency_grids"], composer)

        # Acoustic tuning parameters
        acoustic_tuning_parms = inventory.getArtifacts(["eif_scene"])[
            "eif_scene"
        ].acousticTuningParams

        early_reflection_tuning_gain_db: float = acoustic_tuning_parms.erGainDb
        has_early_reflection_tuning: bool = early_reflection_tuning_gain_db != -8.0
        composer.push(has_early_reflection_tuning)

        if has_early_reflection_tuning:
            er_level: float = round(2 * early_reflection_tuning_gain_db + 43)
            composer.push(int(er_level))

        composer.push(self.use_low_complexity_er)

        hasHeadphoneEQ = True if inventory.artifacts["headphoneEQ"][0].specType != None else False
        composer.push(hasHeadphoneEQ)
        if hasHeadphoneEQ:
            self.writeHeadphoneEQ(
                inventory.artifacts["headphoneEQ"][0],
                inventory.artifacts["frequency_grids"],
                inventory.artifacts["frequency_grid_mapping"],
                composer,
            )

        has_loudness_culling_threshold: bool = self.loudness_culling_threshold != 0
        composer.push(has_loudness_culling_threshold)

        if has_loudness_culling_threshold:
            composer.push(self.loudness_culling_threshold + 73)

        has_consolidation_threshold: bool = self.consolidation_threshold != 5
        composer.push(has_consolidation_threshold)

        if has_consolidation_threshold:
            consolidation_threshold: float = self.consolidation_threshold
            if consolidation_threshold > 31:
                consolidation_threshold = 31
            composer.push(int(consolidation_threshold))

        composer.push(scene_info["enableLocallyCapturedAudio"])
        if scene_info["enableLocallyCapturedAudio"]:
            composer.push(scene_info["overrideListenerVoiceDirectivity"])
            if scene_info["overrideListenerVoiceDirectivity"]:
                composer.push(bs_ids[scene_info["listenerVoiceDirectivityId"]])
        bitstream = composer.get_bitstream()
        bitstream.pad()
        return [bitstream]

    def writeFreqGridData(self, freqGridData: FreqGridData, composer: BitstreamComposer):
        composer.push(freqGridData.fgdNrGrids)
        for g in range(freqGridData.fgdNrGrids):
            composer.push(freqGridData.fgdMethod[g]["method"].value)
            if freqGridData.fgdMethod[g]["method"].name == FGDMethod.INDIVIDUAL_FREQUENCIES.name:
                freqGrid = freqGridData.uniqueFrequencyLayouts[g]
                fgdNrBands = len(freqGrid)
                composer.push(fgdNrBands)
                for b in range(fgdNrBands):
                    composer.push(freqGrid[b])
            elif freqGridData.fgdMethod[g]["method"].name == FGDMethod.DEFAULT_BANDING.name:
                composer.push(freqGridData.fgdMethod[g]["value"])
                isSubGrid = freqGridData.fgdMethod[g].keys().__contains__("offset")
                composer.push(isSubGrid)
                if isSubGrid:
                    composer.push(freqGridData.fgdMethod[g]["offset"])
                    composer.push(freqGridData.fgdMethod[g]["length"] - 1)
                # if composer.scope.__contains__("fgdNrBands"):
                #     composer.scope["fgdNrBands"].append(len(freqGridData.uniqueFrequencyLayouts[g]))
                # else:
                #     composer.scope["fgdNrBands"] = [len(freqGridData.uniqueFrequencyLayouts[g])]
            elif freqGridData.fgdMethod[g]["method"].name == FGDMethod.START_HOP_AMOUNT.name:
                composer.push(len(freqGridData.uniqueFrequencyLayouts[g]))
                composer.push(freqGridData.fgdMethod[g]["start"])
                composer.push(freqGridData.fgdMethod[g]["key"])
            else:
                raise SyntaxError("Unknown frequency grid method used")

    def writeHeadphoneEQ(
        self, headphoneEQ, freqGridData, freqGridMapping, composer: BitstreamComposer
    ):
        if headphoneEQ.specType == HeadphoneSpecType.freqgain.name:
            composer.push(HeadphoneSpecType.freqgain.value)
            numBands = len(freqGridData.uniqueFrequencyLayouts[freqGridMapping["headphoneEQ"]])
            composer.scope["headphoneNrBands"] = numBands
            composer.push(freqGridMapping["headphoneEQ"])
            for g in headphoneEQ.gain:
                composer.push(g)
        elif headphoneEQ.specType == HeadphoneSpecType.parametricfilters.name:
            composer.push(HeadphoneSpecType.parametricfilters.value)
            numBands = len(freqGridData.uniqueFrequencyLayouts[freqGridMapping["headphoneEQ"]])
            composer.scope["headphoneNrBands"] = numBands
            composer.push(freqGridMapping["headphoneEQ"])
            for n in range(0, numBands):
                composer.push(HeadphoneFilterType[headphoneEQ.filtType[n]].value)
                composer.push(headphoneEQ.gain[n])
                composer.push(headphoneEQ.q[n])
        elif headphoneEQ.specType == HeadphoneSpecType.preset.name:
            composer.push(HeadphoneSpecType.preset.value)
            composer.push(headphoneEQ.preset)
        else:
            raise ValueError("Unknown HeadphoneEQ Spec Type")
