# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import Any, Dict, List
from distutils.util import strtobool

from mpegi.framework import Bitstream, BitstreamComposer, Inventory, MHASPacketType, Packet, tokens
from mpegi.metadata.eifplus import Update

from . import _constants
from .components import BitstreamIdentifiers

position_modification_syntax: List[tokens.Token] = [
    tokens.Repeat(
        3,
        syntax=[
            tokens.Flag("isParameterVariable"),
            tokens.Conditional(
                "isParameterVariable",
                true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
                false=[
                    tokens.writeDistance("positionValue"),
                    tokens.Flag("isNegative"),
                ],
            ),
        ],
    )
]

orientation_modification_syntax: List[tokens.Token] = [
    tokens.Repeat(
        3,
        syntax=[
            tokens.Flag("isParameterVariable"),
            tokens.Conditional(
                "isParameterVariable",
                true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
                false=[tokens.Float32("positionValue")],
            ),
        ],
    )
]

cspace_modification_syntax: List[tokens.Token] = [tokens.Flag("cspace")]

active_modification_syntax: List[tokens.Token] = [tokens.Flag("active")]

gain_modification_syntax: List[tokens.Token] = [
    tokens.Flag("isParameterVariable"),
    tokens.Conditional(
        "isParameterVariable",
        true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
        false=[tokens.writeGain("gainValue", isHiPrecGain=True)],
    ),
]

signal_modification_syntax: List[tokens.Token] = [tokens.writeID("signalId")]

extent_modification_syntax: List[tokens.Token] = [tokens.writeID("extentId")]

material_modification_syntax: List[tokens.Token] = [tokens.writeID("materialID")]

directivity_modification_syntax: List[tokens.Token] = [tokens.writeID("directivityId")]

directiveness_modification_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat("directivenessValue", quant_range=_constants.directiveness_qrange)
]

loudness_modification_syntax: List[tokens.Token] = [
    tokens.Flag("isParameterVariable"),
    tokens.Conditional(
        "isParameterVariable",
        true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
        false=[tokens.Float32("loudnessDb")],
    ),
]

play_modification_syntax: List[tokens.Token] = [tokens.Flag("playValue")]

group_modification_syntax: List[tokens.Token] = [tokens.writeID("groupId")]

region_modification_syntax: List[tokens.Token] = [tokens.writeID("regionId")]

size_modification_syntax: List[tokens.Token] = [
    tokens.writeDistance("sizeX"),
    tokens.writeDistance("sizeY"),
    tokens.writeDistance("sizeZ"),
]

speed_modification_syntax: List[tokens.Token] = [
    tokens.Flag("isParameterVariable"),
    tokens.Conditional(
        "isParameterVariable",
        true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
        false=[tokens.QuantizedFloat("speed", quant_range=_constants.airflow_speed_range)],
    ),
]

granular_pos_modification_syntax: List[tokens.Token] = [
    tokens.Repeat(
        3,
        syntax=[
            tokens.Flag("isParameterVariable"),
            tokens.Conditional(
                "isParameterVariable",
                true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
                false=[
                    tokens.QuantizedFloat(
                        "granularPosValue", quant_range=_constants.granularDB_descriptor_qrange
                    )
                ],
            ),
        ],
    )
]

granular_group_prob_modification_syntax: List[tokens.Token] = [
    tokens.Flag("isParameterVariable"),
    tokens.Conditional(
        "isParameterVariable",
        true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
        false=[
            tokens.QuantizedFloat(
                "granular_prob_group_no", quant_range=_constants.grainProbabilityGroupNo_qrange
            )
        ],
    ),
    tokens.Flag("isParameterVariable"),
    tokens.Conditional(
        "isParameterVariable",
        true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
        false=[
            tokens.QuantizedFloat(
                "granular_group_prob", quant_range=_constants.grainProbabilityValue_qrange
            )
        ],
    ),
]

reverb_gain_modification_syntax: List[tokens.Token] = [
    tokens.Flag("isParameterVariable"),
    tokens.Conditional(
        "isParameterVariable",
        true=[tokens.UnsignedInteger("parameterVariableIndex", width=4)],
        false=[tokens.writeGain("reverbGainDb", isHiPrecGain=True)],
    ),
]

no_reverb_modification_syntax: List[tokens.Token] = [tokens.Flag("noReverb")]

modification_syntax: List[tokens.Token] = [
    tokens.writeID("targetId"),
    tokens.Flag("hasDuration"),
    tokens.Conditional(
        "hasDuration",
        false=[],
        true=[
            tokens.writeDuration("duration", isLongRangeMode=True),
        ],
    ),
    tokens.UnsignedInteger("changesCount", width=5),
    tokens.Repeat(
        "changesCount",
        syntax=[
            tokens.UnsignedInteger("targetAttribute", width=6),
            tokens.Switch(
                "targetAttribute",
                cases=[
                    tokens.SwitchCase(0, syntax=position_modification_syntax),
                    tokens.SwitchCase(1, syntax=orientation_modification_syntax),
                    tokens.SwitchCase(2, syntax=cspace_modification_syntax),
                    tokens.SwitchCase(3, syntax=active_modification_syntax),
                    tokens.SwitchCase(4, syntax=gain_modification_syntax),
                    tokens.SwitchCase(5, syntax=signal_modification_syntax),
                    tokens.SwitchCase(6, syntax=extent_modification_syntax),
                    tokens.SwitchCase(7, syntax=directivity_modification_syntax),
                    tokens.SwitchCase(8, syntax=directiveness_modification_syntax),
                    tokens.SwitchCase(9, syntax=play_modification_syntax),
                    tokens.SwitchCase(10, syntax=group_modification_syntax),
                    tokens.SwitchCase(11, syntax=region_modification_syntax),
                    tokens.SwitchCase(12, syntax=size_modification_syntax),
                    tokens.SwitchCase(13, syntax=speed_modification_syntax),
                    tokens.SwitchCase(14, syntax=material_modification_syntax),
                    tokens.SwitchCase(15, syntax=reverb_gain_modification_syntax),
                    tokens.SwitchCase(16, syntax=no_reverb_modification_syntax),
                    tokens.SwitchCase(17, syntax=loudness_modification_syntax),
                    tokens.SwitchCase(18, syntax=granular_pos_modification_syntax),
                    tokens.SwitchCase(19, syntax=granular_group_prob_modification_syntax),
                ],
            ),
        ],
    ),
]

timed_update_syntax: List[tokens.Token] = [
    tokens.Flag("hasId"),
    tokens.Conditional(
        "hasId",
        false=[],
        true=[tokens.writeID("updateId")],
    ),
    tokens.Flag("hasIndex"),
    tokens.Conditional(
        "hasIndex",
        false=[],
        true=[tokens.writeCountOrIndex("updateIndex")],
    ),
    tokens.writeDuration("time", isLongRangeMode=True),  # TODO: get fs from scene info
]

conditional_update_syntax: List[tokens.Token] = [
    tokens.Flag("hasId"),
    tokens.Conditional(
        "hasId",
        false=[],
        true=[tokens.writeID("updateId")],
    ),
    tokens.Flag("hasIndex"),
    tokens.Conditional(
        "hasIndex",
        false=[],
        true=[tokens.writeCountOrIndex("updateIndex")],
    ),
    tokens.Flag("fireOn"),
    tokens.Flag("hasDelay"),
    tokens.Conditional(
        "hasDelay",
        false=[],
        true=[tokens.writeDuration("delay", isLongRangeMode=True)],
    ),
    tokens.writeID("conditionId"),
]

triggered_update_syntax: List[tokens.Token] = [
    tokens.writeID("updateId"),
    tokens.writeCountOrIndex("updateIndex"),
]

dynamic_update_syntax: List[tokens.Token] = [
    tokens.writeID("updateId"),
    tokens.writeCountOrIndex("updateIndex"),
]

write_updates = tokens.Function(
    "writeUpdates",
    syntax=[
        tokens.writeCountOrIndex("updatesCount"),
        tokens.Repeat(
            "updatesCount",
            syntax=[
                tokens.writeCountOrIndex("modificationsCount"),
                tokens.Repeat("modificationsCount", syntax=modification_syntax),
                tokens.UnsignedInteger("updateType", width=2),
                tokens.Switch(
                    "updateType",
                    cases=[
                        tokens.SwitchCase(0, timed_update_syntax),
                        tokens.SwitchCase(1, conditional_update_syntax),
                        tokens.SwitchCase(2, triggered_update_syntax),
                        tokens.SwitchCase(3, dynamic_update_syntax),
                    ],
                ),
            ],
        ),
    ],
)


def _serialize_timed_update(
    composer: BitstreamComposer, update: Update, bs_ids: BitstreamIdentifiers
) -> None:
    has_id = update.id is not None
    composer.push(has_id)
    if has_id:
        composer.push(bs_ids[update])

    has_index = update.index is not None
    composer.push(has_index)
    if has_index:
        composer.push(update.index)

    composer.push(update.time)


def _serialize_conditional_update(
    composer: BitstreamComposer, update: Update, bs_ids: BitstreamIdentifiers
) -> None:
    has_id = update.id is not None
    composer.push(has_id)
    if has_id:
        composer.push(bs_ids[update])

    has_index = update.index is not None
    composer.push(has_index)
    if has_index:
        composer.push(update.index)

    composer.push(update.fireOn)

    has_delay = update.delay is not None
    composer.push(has_delay)
    if has_delay:
        composer.push(update.delay)

    composer.push(bs_ids[update.condition])


def _serialize_triggered_update(
    composer: BitstreamComposer, update: Update, bs_ids: BitstreamIdentifiers
) -> None:
    composer.push(bs_ids[update])
    composer.push(update.index)


def _serialize_dynamic_update(
    composer: BitstreamComposer, update: Update, bs_ids: BitstreamIdentifiers
) -> None:
    composer.push(bs_ids[update])
    composer.push(update.index)


def _serialize_position_modification(
    composer: BitstreamComposer, value: str, _, isSmallScene: bool
) -> None:
    for val in value.split(" "):
        is_placeholder = val[0] == "$"
        composer.push(is_placeholder)
        if is_placeholder:
            composer.push(int(val[1:]))
        else:
            composer.push({"distance": abs(float(val)), "isSmallScene": isSmallScene})
            composer.push(float(val) < 0)


def _serialize_orientation_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    for val in value.split(" "):
        is_placeholder = val[0] == "$"
        composer.push(is_placeholder)
        if is_placeholder:
            composer.push(int(val[1:]))
        else:
            angle = float(val)
            wrapped_angle = ((angle + 180.0) % 360) - 180.0
            composer.push(wrapped_angle)


def _serialize_size_modification(
    composer: BitstreamComposer, value: str, _, isSmallScene: bool
) -> None:
    val = value.split(" ")
    composer.push({"distance": float(val[0]), "isSmallScene": isSmallScene})
    composer.push({"distance": float(val[1]), "isSmallScene": isSmallScene})
    composer.push({"distance": float(val[2]), "isSmallScene": isSmallScene})
    # composer.push(float(val[1]))
    # composer.push(float(val[2]))


def _serialize_flag_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    composer.push(bool(strtobool(value)))


def _serialize_cspace_modification(composer: BitstreamComposer, cspace: str, _, __) -> None:
    if cspace == "relative":
        composer.push(False)
    elif cspace == "user":
        composer.push(True)


def _serialize_gain_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    is_placeholder = value[0] == "$"
    composer.push(is_placeholder)
    if is_placeholder:
        composer.push(int(value[1:]))
    else:
        composer.push(float(value))


def _serialize_directiveness_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    composer.push(float(value))


def _serialize_id_modification(
    composer: BitstreamComposer, value: str, bs_ids: BitstreamIdentifiers, _
) -> None:
    composer.push(bs_ids[value])


def _serialize_speed_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    is_placeholder = value[0] == "$"
    composer.push(is_placeholder)
    if is_placeholder:
        composer.push(int(value[1:]))
    else:
        composer.push(float(value))


def _serialize_granular_pos_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    for val in value.split(" "):
        is_placeholder = val[0] == "$"
        composer.push(is_placeholder)
        if is_placeholder:
            composer.push(int(val[1:]))
        else:
            composer.push(float(val))


def _serialize_granular_group_prob_modification(
    composer: BitstreamComposer, value: str, _, __
) -> None:
    val = value.split(" ")
    is_placeholder = val[0][0] == "$"
    composer.push(is_placeholder)
    if is_placeholder:
        composer.push(int(val[0][1:]))
    else:
        composer.push(float(val[0]))
    is_placeholder = val[1][0] == "$"
    composer.push(is_placeholder)
    if is_placeholder:
        composer.push(int(val[1][1:]))
    else:
        composer.push(float(val[1]))


def _serialize_material_modification(
    composer: BitstreamComposer, value: str, bs_ids: BitstreamIdentifiers, _
) -> None:
    composer.push(bs_ids[value])


def _serialize_reverb_gain_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    is_placeholder = value[0] == "$"
    composer.push(is_placeholder)
    if is_placeholder:
        composer.push(int(value[1:]))
    else:
        composer.push(float(value))


def _serialize_loudness_modification(composer: BitstreamComposer, value: str, _, __) -> None:
    is_placeholder = value[0] == "$"
    composer.push(is_placeholder)
    if is_placeholder:
        composer.push(int(value[1:]))
    else:
        composer.push(float(value))


_targetAttSwitch = {
    "position": _serialize_position_modification,
    "orientation": _serialize_orientation_modification,
    "cspace": _serialize_cspace_modification,
    "active": _serialize_flag_modification,
    "gaindb": _serialize_gain_modification,
    "signal": _serialize_id_modification,
    "extent": _serialize_id_modification,
    "directivity": _serialize_id_modification,
    "directiveness": _serialize_directiveness_modification,
    "play": _serialize_flag_modification,
    "group": _serialize_id_modification,
    "region": _serialize_id_modification,
    "size": _serialize_size_modification,
    "speed": _serialize_speed_modification,
    "material": _serialize_material_modification,
    "reverbgaindb": _serialize_reverb_gain_modification,
    "mutereverb": _serialize_flag_modification,
    "loudnessdb": _serialize_loudness_modification,
    "granularpos": _serialize_granular_pos_modification,
    "granulargroupprob": _serialize_granular_group_prob_modification,
}


class SceneUpdatePacket(Packet):
    """Packet detailing how to update a scene thruogh its runtime."""

    def __init__(self, args: Dict[str, Any]) -> None:
        pass

    @staticmethod
    def get_packet_type() -> MHASPacketType:
        """Return PACTYPE_UPD MHAS structure."""
        return MHASPacketType.PACTYP_UPD

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        """Return a list of all required arguments."""
        return []

    def serialize(self, inventory: Inventory) -> List[Bitstream]:
        """Serialize SceneUpdate packet."""
        bs_ids: BitstreamIdentifiers = inventory.getArtifacts(["bs_ids"])["bs_ids"]
        updates: List[Update] = inventory.getArtifacts(["updates"])["updates"]
        updates.sort(key=lambda upd: -1 if upd.time == None else upd.time)
        isSmallScene = inventory.artifacts["scene_info"]["isSmallScene"]

        composer = BitstreamComposer([write_updates])
        composer.push(len(updates))
        for upd in updates:
            composer.push(len(upd.modifications))
            for mod in upd.modifications:
                composer.push(bs_ids[mod.id])

                has_duration = mod.duration is not None
                composer.push(has_duration)
                if has_duration:
                    composer.push(mod.duration)

                composer.push(len(mod.changes))
                for key, value in mod.changes.items():
                    target_attribute = list(_targetAttSwitch.keys()).index(key)
                    composer.push(target_attribute)
                    _targetAttSwitch[key](composer, value, bs_ids, isSmallScene)

            if upd.type == "timed":
                composer.push(0)
                _serialize_timed_update(composer, upd, bs_ids)
            elif upd.type == "conditional":
                composer.push(1)
                _serialize_conditional_update(composer, upd, bs_ids)
            elif upd.type == "triggered":
                composer.push(2)
                _serialize_triggered_update(composer, upd, bs_ids)
            elif upd.type == "dynamic":
                composer.push(3)
                _serialize_dynamic_update(composer, upd, bs_ids)
            else:
                raise Exception()

        return [composer.get_bitstream()]
