# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import argparse
import logging
from typing import Any, Dict, List
from dataclasses import dataclass

from mpegi.framework import Component


@dataclass
class AirflowFrequencyGrid:
    gain: List[float]
    q: List[float]
    fgdNrBands: int
    frequencies: List[float]
    speed: float
    azimuth: float
    elevation: float


class AirflowComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        super().__init__(args, artifacts)

        airflowSources = artifacts["airflow_sources"]

        self.frequencyGrids = {}

        for src in airflowSources:
            if src.AirflowProfile is not None:
                for profile in src.AirflowProfile:
                    freqGrid = AirflowFrequencyGrid(
                        gain=profile.gainDb,
                        q=profile.q,
                        frequencies=profile.freqs,
                        fgdNrBands=len(profile.freqs),
                        speed=profile.speed,
                        azimuth=profile.azimuth,
                        elevation=profile.elevation,
                    )
                    self.frequencyGrids[profile.id] = freqGrid

        speedProfiles = [x for x in self.frequencyGrids if self.frequencyGrids[x].speed is not None]
        maxGain = 0
        for grid in speedProfiles:
            maxGain = max([maxGain, max(self.frequencyGrids[grid].gain)])
        for grid in speedProfiles:
            self.frequencyGrids[grid].normGain = [
                x - maxGain for x in self.frequencyGrids[grid].gain
            ]

        aziProfiles = [x for x in self.frequencyGrids if self.frequencyGrids[x].azimuth is not None]
        maxGain = 0
        for grid in aziProfiles:
            maxGain = max([maxGain, max(max(self.frequencyGrids[grid].gain))])
        for grid in aziProfiles:
            self.frequencyGrids[grid].normGain = [
                [y - maxGain for y in x] for x in self.frequencyGrids[grid].gain
            ]

        elevProfiles = [
            x for x in self.frequencyGrids if self.frequencyGrids[x].elevation is not None
        ]
        maxGain = 0
        for grid in elevProfiles:
            maxGain = max([maxGain, max(self.frequencyGrids[grid].gain)])
        for grid in elevProfiles:
            self.frequencyGrids[grid].normGain = [
                x - maxGain for x in self.frequencyGrids[grid].gain
            ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return ["scene"]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return ["airflow_sources", "airflow_generators"]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return ["airflow_frequency_grids"]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "airflow_frequency_grids": self.frequencyGrids,
        }

    @staticmethod
    def getVersion() -> Any:
        return 1
