# This software module was originally developed by
#
# Ericsson AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Ericsson AB retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import pathlib
from typing import Any, Dict, List

import numpy as np
from scipy import signal

from mpegi.dsp import read_wav
from mpegi.framework import Component


class AudioplusComponent(Component):
    """A component is a small compartmental unit of the encoder. It is responsible
    for creating artifacts (typically) from preexisting artifacts created from
    other components.
    """

    # Note that the analysis are done using the original waw files, not the MPEG-H encoded signals
    # This can of course be modified, but then ecah channel is in a separate wav file
    # and the read function for the audio file will need to be updated
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        object_sources = []
        if "object_sources" in artifacts.keys():
            object_sources = [
                item for item in artifacts["object_sources"] if not item.extent is None
            ]
        if "hoa_sources" in artifacts.keys():
            hoa_sources = [
                item
                for item in artifacts["hoa_sources"]
                if item.extentTransform & (not item.extent is None)
            ]
        audio_streams = artifacts["audio_streams"]
        self.data = []
        for object_source in object_sources:
            audio_stream_id = object_source.signal
            found_stream = [stream for stream in audio_streams if stream.id == audio_stream_id]
            channels = 0
            if len(found_stream) > 0:
                audio_file_path = str(
                    pathlib.Path(args["scene"]).parent.joinpath(found_stream[0].file)
                )
                channels = len(found_stream[0].aepInputChannels)

            if channels > 1:
                # Multichannel wave file
                # Run analysis program on audio_file_path and deliver analysis parameters
                # This needs to be updated if the analysis should be done on the encoded audio signal
                samplerate, data = read_wav(audio_file_path)
                # Perform correlation
                correlation = self.correlationAnalysis(data, samplerate)
                # Map to diffuseness
                hasDiffuseness = True
                diffuseness = 1.0 - abs(correlation[0])
                # Perform spatial sparsity analysis
                distanceGainModel = self.directionalityAnalysis(data, samplerate)
                audioplus_params = [
                    object_source.id,
                    distanceGainModel,
                    hasDiffuseness,
                    diffuseness,
                ]
                self.data.append(audioplus_params)
            else:
                # Single channel signal or not a wave file
                distanceGainModel = False
                hasDiffuseness = False
                diffuseness = 0
                audioplus_params = [
                    object_source.id,
                    distanceGainModel,
                    hasDiffuseness,
                    diffuseness,
                ]
                self.data.append(audioplus_params)

        xx = 1
        for hoa_source in hoa_sources:
            audio_stream_id = hoa_source.signal
            found_stream = [stream for stream in audio_streams if stream.id == audio_stream_id]
            channels = 0
            if len(found_stream) > 0:
                audio_file_path = str(
                    pathlib.Path(args["scene"]).parent.joinpath(found_stream[0].file)
                )
                channels = len(found_stream[0].aepInputChannels)

            if channels > 1:
                audio_file_path = str(
                    pathlib.Path(args["scene"]).parent.joinpath(found_stream[0].file)
                )
                # Run analysis program on audio_file_path and deliver analysis parameters
                # This needs to be updated if the analysis should be done on the encoded audio signal
                samplerate, data = read_wav(audio_file_path)
                # Perform correlation
                correlation = self.correlationAnalysis(data, samplerate)
                # Map to diffuseness
                hasDiffuseness = True
                diffuseness = 1.0 - abs(correlation[0])
                # Perform spatial sparsity analysis
                distanceGainModel = self.directionalityAnalysis(data, samplerate)
                audioplus_params = [hoa_source.id, distanceGainModel, hasDiffuseness, diffuseness]
                self.data.append(audioplus_params)
            else:
                # Not a wave file
                distanceGainModel = False
                hasDiffuseness = False
                diffuseness = 0
                audioplus_params = [
                    hoa_source.id,
                    distanceGainModel,
                    hasDiffuseness,
                    diffuseness,
                ]
                self.data.append(audioplus_params)
        yy = 1

    ##### util functions #####
    ## calculate energy per frame
    def energyAnalysis(self, matrix):
        e = np.mean(matrix ** 2, axis=0)
        return e

    ## emulate mid/side calculation and compare energy values
    def midSideRelationAnalysis(self, matrix):
        channels = matrix.shape[1]
        mid = np.mean(matrix, axis=1)
        if channels == 2:
            ## do proper side calculation for stereo signals (l-r)
            side = matrix[:, 0] - matrix[:, 1]
        elif channels > 2:
            ## pseudo side calculation
            side = (matrix.T - mid).T
        ## energy calculation
        e_mid = np.mean(mid ** 2)
        e_side = np.mean(np.mean(side ** 2, axis=0))
        return e_side / e_mid

    ## perform SVD and calculate relation
    def spatialAnalysis(self, matrix):
        D = np.linalg.svd(matrix, full_matrices=False, compute_uv=False)
        Dgt0 = [item for item in D if item > 0]
        # relDir = max(D) / min(D)
        relDir = max(D) / min(Dgt0)
        return relDir

    ##### analysis functions #####
    def correlationAnalysis(self, data, samplerate):
        ## split into frames
        framelength = int(samplerate / 100)  # 10ms
        frames = int(np.ceil(data.shape[0] / framelength))
        channels = data.shape[1]

        corrCoefs = []
        for i in range(0, frames):
            frame = data[i * framelength : (i + 1) * framelength]
            coefs = []
            for c in range(0, channels - 1):
                channel1 = frame[:, c]
                channel2 = frame[:, c + 1]
                coef = np.corrcoef(channel1, channel2)[0, 1]  # correlation coefficient
                coefs.append(coef)

            corrCoefs.append(coefs)

        corrCoefs = np.array(corrCoefs)

        return np.mean(corrCoefs, axis=0)

    def directionalityAnalysis(self, data, samplerate):
        ## apply 1kHz highpass
        hp_coef = signal.firwin(41, 1000, window=("kaiser", 5.0), pass_zero=False, fs=samplerate)
        data = signal.lfilter(hp_coef, 1, data, axis=0)

        ## split into frames
        framelength = int(samplerate / 10)  # 100ms
        frames = int(np.ceil(data.shape[0] / framelength))

        ## per frame analysis
        dirs = []  # relative directionality per frame
        energies = []  # energies per frame and channel
        energy_relations = []  # energy relation between mid / side
        for i in range(0, frames):
            frame = data[i * framelength : (i + 1) * framelength]
            ## perform analysis
            relDir = self.spatialAnalysis(frame)
            e = self.energyAnalysis(frame)
            rel = self.midSideRelationAnalysis(frame)

            ## store statistics
            dirs.append(relDir)
            energies.append(e)
            energy_relations.append(rel)

        ## convert statistics to numpy arrays
        dirs = np.array(dirs)
        energies = np.array(energies)
        energy_relations = np.array(energy_relations)

        ## combine svd directionality and energy relation to final score
        scores = dirs * energy_relations  # + np.var(dirs * energy_relations)
        ## weighted average with energies per frame
        score = np.sum((scores * np.mean(energies, axis=1)) / np.sum(np.mean(energies, axis=1)))

        ## Thresholding
        threshold = 20
        spatially_sparse = score >= threshold

        return spatially_sparse

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return ["scene"]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return ["object_sources", "hoa_sources", "audio_streams"]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return ["audioplus"]

    def getArtifacts(self) -> Dict[str, Any]:
        return {"audioplus": self.data}

    @staticmethod
    def getVersion() -> Any:
        return 1

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        pass
