# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
from typing import Any, Dict, List

from mpegi.framework import Component
from mpegi.metadata.eifplus import Entity, Update


class BitstreamIdentifiers:
    """Provides a way to access the same integer value with multiple keys"""

    def __init__(self) -> None:
        self.anyToIntIdMap: Dict[Any, int] = {}
        self.numOfItems = 0

    def add(self, key: Any) -> int:
        """Add an object to assign it a bitstream id

        :return: Returns the bitstream identifier for the given key
        :rtype: int
        """
        key = self._getKeyFromObj(key)
        assert key not in self.anyToIntIdMap.keys()  # should never be reached
        self.anyToIntIdMap[key] = self.numOfItems
        self.numOfItems += 1
        return self.numOfItems - 1

    def getId(self, key: Any) -> int:
        """Return the bitstream identifier of the previously registered keys

        :param key: A key that was previously registered
        :type key: Any
        :return: The bitstream identifier
        :rtype: int
        """
        key = self._getKeyFromObj(key)
        return self.anyToIntIdMap[key]

    def __getitem__(self, key) -> int:
        """Provides dict-like access with "[]" operator"""
        return self.getId(key)

    @staticmethod
    def _getKeyFromObj(key: Any) -> str:
        # TODO: In the future, EIF entities should have __hash__ and __eq__ implemented
        #       such that they themselves can be used as dict keys. Then this stops
        #       this ugly function from existing :D.
        if key == str:
            return key
        elif isinstance(key, Entity) and not isinstance(key, Update):
            return key.id
        elif isinstance(key, Update):
            stringKey = ""
            if key.id != None:
                stringKey = key.id
            elif key.index != None:
                stringKey = f"update-index:{key.index}"
            elif key.time != None:
                stringKey = f"update-time:{key.time}"
            elif key.condition != None:
                fireOnString = f"{key.fireOn}" if key.fireOn != None else f"{True}"
                stringKey = f"update-condition:{key.condition}|fireOn:{fireOnString}"
            else:
                raise Exception("Invalid Update format")  # should never be reached
            return stringKey
        else:
            return key


class BitstreamIdComponent(Component):
    """This component is responsible for keeping tracking of all of the integer
    ids within the bitstream. Bitstream identifiers can be registered with
    different keys. This is useful for getting a bitstream identifier of an
    entity with either the python object representation or its stringId.
    """

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        super().__init__(args, artifacts)

        # Register Objects
        self.bitstreamIds = BitstreamIdentifiers()
        allEntities = [
            *artifacts["acoustic_environments"],
            *artifacts["acoustic_materials"],
            *artifacts["anchors"],
            *artifacts["audio_streams"],
            *artifacts["boxes"],
            *artifacts["channel_sources"],
            *artifacts["cylinders"],
            *artifacts["directivities"],
            *artifacts["informed_sources"],
            *artifacts["hoa_groups"],
            *artifacts["hoa_sources"],
            *artifacts["meshes"],
            *artifacts["object_sources"],
            *artifacts["spheres"],
            *artifacts["transforms"],
            *artifacts["updates"],
            *artifacts["airflow_sources"],
            *artifacts["airflow_generators"],
            *artifacts["granular_sound"],
        ]
        for ent in allEntities:
            self.bitstreamIds.add(ent)
        if artifacts["portal_info"] != []:
            for portal in artifacts["portal_info"].portalGeometries:
                self.bitstreamIds.add(portal)
        if artifacts["eif_scene"].portals != []:
            for portal in artifacts["eif_scene"].portals:
                self.bitstreamIds.add(portal.id)
        for urr in artifacts["user_regions"]:
            self.bitstreamIds.add(urr)

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return ["serializeURR", "disableDefaultAE"]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "acoustic_environments",
            "acoustic_materials",
            "anchors",
            "audio_streams",
            "boxes",
            "channel_sources",
            "cylinders",
            "directivities",
            "informed_sources",
            "eif_scene",
            "hoa_groups",
            "hoa_sources",
            "meshes",
            "object_sources",
            "spheres",
            "transforms",
            "updates",
            "portal_info",
            "user_regions",
            "airflow_sources",
            "airflow_generators",
            "granular_sound",
        ]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return ["bs_ids"]

    def getArtifacts(self) -> Dict[str, Any]:
        return {"bs_ids": self.bitstreamIds}

    @staticmethod
    def getVersion() -> Any:
        return 5

    @staticmethod
    def registerArguments(_: argparse.ArgumentParser) -> None:
        pass
