# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import logging
from typing import Any, Dict, List
import numpy as np

from mpegi import voxel
from mpegi.framework import Component
from mpegi.metadata import diffraction


class DiffractionComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Constructor to set up Diffraction module"""
        super().__init__(args, artifacts)

        sceneDict = diffraction.SceneDictionary(artifacts, False)

        scene_trimesh_DIFFR = sceneDict.sceneDictionary["sceneTrimesh_DIFFR"]

        voxelizer_DIFFR = voxel.VoxelizerDIFFR(sceneDict.sceneDictionary, args)
        self.earLevelVoxelGrid = voxelizer_DIFFR.voxelGrid_earLevel_DIFFR
        earLevelVoxelPositions = voxelizer_DIFFR.voxelPositions_earLevel_DIFFR

        if (
            sceneDict.sceneDictionary["_flag_emtpySceneTrimesh_DIFFR"]
            or sceneDict.sceneDictionary["_flag_noObjectSources"]
        ):
            logging.info(f" - sceneDictionary['_flag_emtpySceneTrimesh_DIFFR'] = True")
            self.static_el = []
            self.static_pd = {}
            self.sourceVoxelDict = {}
            self.listenerVoxelDict = {}
            self.cell = {}
            self.edgeDict = {}
            self.pathDict = {}
            self.validPathDict = {}
            return

        sceneTrimeshName_DIFFR = "sceneTrimesh_DIFFR"

        # Note: do not change the processing sequence
        elg_perScene = diffraction.EdgeListGenPerScene(
            sceneTrimeshName_DIFFR,
            self.earLevelVoxelGrid,
            sceneDict.sceneDictionary,
            args,
        )

        vdg_source = diffraction.SourceVoxelDictGen(
            elg_perScene.edgeListGenDict[sceneTrimeshName_DIFFR],
            sceneDict.sceneDictionary,
            self.earLevelVoxelGrid,
        )

        pdg_perScene = diffraction.PathDictGenPerScene(
            sceneTrimeshName_DIFFR, elg_perScene, vdg_source, sceneDict.sceneDictionary, args
        )

        vdg_listener = diffraction.ListenerVoxelDictGen(
            elg_perScene.edgeListGenDict[sceneTrimeshName_DIFFR],
            sceneDict.sceneDictionary,
            self.earLevelVoxelGrid,
            earLevelVoxelPositions,
        )

        lpdg = diffraction.ValidPathGen(
            vdg_source.sourceVoxelDict,
            vdg_listener.listenerVoxelDict,
            pdg_perScene.dictOfPathDicts,
            pdg_perScene.dictOfPathDictsByEdgeIds,
            self.earLevelVoxelGrid,
            sceneDict.sceneDictionary,
        )

        cg_listener = diffraction.CellForListenerVoxels(vdg_listener.listenerVoxelDict)

        self.static_el = elg_perScene.edgeListGenDict[sceneTrimeshName_DIFFR].edgeList
        self.static_pd = pdg_perScene.dictOfPathDictGens[sceneTrimeshName_DIFFR].pathDict
        self.sourceVoxelDict = vdg_source.sourceVoxelDict
        self.listenerVoxelDict = vdg_listener.listenerVoxelDict
        self.cell = cg_listener.cell

        self.edgeDict = elg_perScene.edgeListDict
        # self.edgeDict[sceneTrimeshName_DIFFR] = self.static_el  # update
        self.pathDict = pdg_perScene.dictOfPathDicts
        # self.pathDict[sceneTrimeshName_DIFFR] = self.static_pd  # update
        self.validPathDict = lpdg.validPathDict

        # re-map vertex and triangle IDs of static edge lists
        mesh_unproc = artifacts["scene_trimesh_DIFFR_unproc_as_eif_mesh"]
        self._remapCleanMesh(scene_trimesh_DIFFR, mesh_unproc)

    def _remapCleanMesh(self, scene_trimesh_DIFFR, mesh_unproc) -> None:
        """re-map vertex and triangle IDs of static edge lists"""
        from ...geometry.trimesh import convertGeometriesToTrimesh

        trimesh_unproc = convertGeometriesToTrimesh([mesh_unproc], faceAngle=30, process=False)[0]

        # generate pos -> vtxID map
        vertex_id_unproc = {}
        for id, vertex_pos in enumerate(trimesh_unproc.vertices):
            pos = vertex_pos.astype("single")
            vertex_id_unproc[tuple(pos)] = id

        # generate vtxID_proc -> vtxID_unproc map
        vertex_map = [None] * len(scene_trimesh_DIFFR.vertices)
        for id_proc, vertex_pos in enumerate(scene_trimesh_DIFFR.vertices):
            pos = vertex_pos.astype("single")
            id_unproc = vertex_id_unproc[tuple(pos)]
            vertex_map[id_proc] = id_unproc

        # generate face_finder map
        face_finder = dict()  # dict: ((vtxID1, vtxID2), fnormal) -> faceID
        for id, face_ref in enumerate(trimesh_unproc.faces):
            # use vtxID -> pos -> vtxID to avoid problems with duplicate vertices
            vertex0_pos = trimesh_unproc.vertices[face_ref[0]].astype("single")
            vertex1_pos = trimesh_unproc.vertices[face_ref[1]].astype("single")
            vertex2_pos = trimesh_unproc.vertices[face_ref[2]].astype("single")
            face = [0, 0, 0]
            face[0] = vertex_id_unproc[tuple(vertex0_pos)]
            face[1] = vertex_id_unproc[tuple(vertex1_pos)]
            face[2] = vertex_id_unproc[tuple(vertex2_pos)]

            # store all three triangle edges in face_finder
            vtxIDs0 = sorted([face[0], face[1]])
            vtxIDs1 = sorted([face[1], face[2]])
            vtxIDs2 = sorted([face[2], face[0]])
            fnormal = trimesh_unproc.face_normals[id].astype("single")
            face_finder[(tuple(vtxIDs0), tuple(fnormal))] = id
            face_finder[(tuple(vtxIDs1), tuple(fnormal))] = id
            face_finder[(tuple(vtxIDs2), tuple(fnormal))] = id

        # re-map vertex and triangle IDs of static edge lists
        miss_counter_vertex = 0
        miss_counter_face = 0
        for edge_ref in self.static_el:
            # copy face ID
            vtx0 = edge_ref.vtxIDs[0]
            vtx1 = edge_ref.vtxIDs[1]
            fid0 = edge_ref.fIDs[0]
            fid1 = edge_ref.fIDs[1]
            fnormal0 = scene_trimesh_DIFFR.face_normals[fid0].astype("single")
            fnormal1 = scene_trimesh_DIFFR.face_normals[fid1].astype("single")

            # identify first vertex
            try:
                # fast method: use vertex_map
                vtx0_new = vertex_map[vtx0]
            except IndexError:
                # slow method: fuzzy search
                miss_counter_vertex += 1
                pos = scene_trimesh_DIFFR.vertices[vtx0]
                vtx0_new = self._getSimilarVertex(trimesh_unproc, pos)

            # identify second vertex
            try:
                # fast method: use vertex_map
                vtx1_new = vertex_map[vtx1]
            except IndexError:
                # slow method: fuzzy search
                miss_counter_vertex += 1
                pos = scene_trimesh_DIFFR.vertices[vtx1]
                vtx1_new = self._getSimilarVertex(trimesh_unproc, pos)

            vtxIDs_new = sorted([vtx0_new, vtx1_new])

            # identify first triangle
            try:
                # fast method: use face_finder dictionary
                fid0_new = face_finder[(tuple(vtxIDs_new), tuple(fnormal0))]
            except KeyError:
                # slow method: fuzzy search
                miss_counter_face += 1
                fid0_new = self._getSimilarFace(trimesh_unproc, vtx0_new, vtx1_new, fnormal0)

            # identify second triangle
            try:
                # fast method: use face_finder dictionary
                fid1_new = face_finder[(tuple(vtxIDs_new), tuple(fnormal1))]
            except KeyError:
                # slow method: fuzzy search
                miss_counter_face += 1
                fid1_new = self._getSimilarFace(trimesh_unproc, vtx0_new, vtx1_new, fnormal1)

            edge_ref.vtxIDs[0] = vtx0_new
            edge_ref.vtxIDs[1] = vtx1_new
            edge_ref.fIDs[0] = fid0_new
            edge_ref.fIDs[1] = fid1_new

            # only needed for validation
            edge_ref.fvtxIDs[0] = list(trimesh_unproc.faces[edge_ref.fIDs[0]])
            edge_ref.fvtxIDs[1] = list(trimesh_unproc.faces[edge_ref.fIDs[1]])
            A0 = trimesh_unproc.vertices[edge_ref.fvtxIDs[0][0]]
            B0 = trimesh_unproc.vertices[edge_ref.fvtxIDs[0][1]]
            C0 = trimesh_unproc.vertices[edge_ref.fvtxIDs[0][2]]
            A1 = trimesh_unproc.vertices[edge_ref.fvtxIDs[1][0]]
            B1 = trimesh_unproc.vertices[edge_ref.fvtxIDs[1][1]]
            C1 = trimesh_unproc.vertices[edge_ref.fvtxIDs[1][2]]
            fcentroid0 = (A0 + B0 + C0) / 3.0
            fcentroid1 = (A1 + B1 + C1) / 3.0
            edge_ref.fcentroids = [fcentroid0, fcentroid1]

    def _getSimilarVertex(self, trimesh, pos) -> int:
        delta = 1e-4
        for vtx, vertex_ref in enumerate(trimesh.vertices):
            if np.linalg.norm(vertex_ref - pos) < delta:
                return vtx

        return None

    def _getSimilarFace(self, trimesh, vtx0, vtx1, fnormal) -> int:
        delta = 1e-4
        fnormal = np.array(fnormal)
        pos0 = trimesh.vertices[vtx0]
        pos1 = trimesh.vertices[vtx1]
        for fid, fnormal_ref in enumerate(trimesh.face_normals):
            if np.linalg.norm(fnormal_ref - fnormal) < delta:
                face_ref = trimesh.faces[fid]
                A = trimesh.vertices[face_ref[0]]
                B = trimesh.vertices[face_ref[1]]
                C = trimesh.vertices[face_ref[2]]
                if (
                    np.linalg.norm(A - pos0) < delta
                    or np.linalg.norm(B - pos0) < delta
                    or np.linalg.norm(C - pos0) < delta
                ):
                    if (
                        np.linalg.norm(A - pos1) < delta
                        or np.linalg.norm(B - pos1) < delta
                        or np.linalg.norm(C - pos1) < delta
                    ):
                        return fid

        return None

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return [
            "max_order_DIFFR",
            "max_iter_DIFFR",
            "max_order_moving_source_DIFFR",
            "min_edge_length_DIFFR",
        ]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "scene_trimesh_DIFFR_as_eif_mesh",
            "scene_trimesh_DIFFR_unproc_as_eif_mesh",
            "scene_metadata",
            "scene_info",
            "acoustic_environments",
            "boxes",
            "cylinders",
            "meshes",
            "object_sources",
            "spheres",
            "updates",
            "user_regions",
        ]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return [
            "earLevelVoxelGrid",
            "diffraction_edges",
            "diffraction_paths",
            "diffraction_sourceVoxels",
            "diffraction_listenerVoxels",
            "diffraction_cell",
            "diffraction_edgeDict",
            "diffraction_pathDict",
            "diffraction_validPathDict",
        ]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "earLevelVoxelGrid": self.earLevelVoxelGrid,
            "diffraction_edges": self.static_el,
            "diffraction_paths": self.static_pd,
            "diffraction_sourceVoxels": self.sourceVoxelDict,
            "diffraction_listenerVoxels": self.listenerVoxelDict,
            "diffraction_cell": self.cell,
            "diffraction_edgeDict": self.edgeDict,
            "diffraction_pathDict": self.pathDict,
            "diffraction_validPathDict": self.validPathDict,
        }

    @staticmethod
    def getVersion() -> Any:
        return 25

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        diffractionParser = argParser.add_argument_group("Diffraction")
        diffractionParser.add_argument(
            "--max_order_DIFFR",
            default=5,
            type=int,
            metavar="ORDER",
            help="Maximum diffraction order for static sources (default=5)",
        )

        diffractionParser.add_argument(
            "--max_iter_DIFFR",
            default=16,
            type=int,
            metavar="LIMIT",
            help="Iteration limit for diffraction mesh voxelization, increase if encoding errors with max_iter exceeded (default=16)",
        )

        diffractionParser.add_argument(
            "--max_order_moving_source_DIFFR",
            default=3,
            type=int,
            metavar="ORDER",
            help="Maximum diffraction order allowed for moving sources (default=5)",
        )

        diffractionParser.add_argument(
            "--min_edge_length_DIFFR",
            default=0.0,
            type=float,
            metavar="LENGTH",
            help="Minimum length for potential edges is optional (default = 0)",
        )
