# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import pathlib
from typing import Any, Dict, List

from mpegi.coding import DirectivityEncoder
from mpegi.framework import Bitstream, Component


class DirectivityComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        self.directivitiesCoded = {}
        encoder = DirectivityEncoder()
        for directivity in artifacts["directivities"]:
            path = str(pathlib.Path(args["scene"]).parent.joinpath(directivity.file))
            directivityAsBytes = encoder.encode(path, args["deg"], args["db"])
            bitstream = Bitstream.fromBytes(directivityAsBytes)
            self.directivitiesCoded[directivity.id] = bitstream

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return ["scene", "db", "deg"]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return ["directivities"]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return ["directivities_coded"]

    def getArtifacts(self) -> Dict[str, Any]:
        return {"directivities_coded": self.directivitiesCoded}

    @staticmethod
    def getVersion() -> Any:
        return 5

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        directivityParser = argParser.add_argument_group("Directivity")
        directivityParser.add_argument(
            "--deg",
            default=15.0,
            type=float,
            metavar="DEGREE",
            help=(
                "Degree resolution for directivity compression. "
                "Cannot be either: less than 2; greater than 90. (default=15.0)"
            ),
        )
        directivityParser.add_argument(
            "--db",
            default=1.0,
            type=float,
            metavar="DECIBEL",
            help=(
                "Decibel resolution for directivity compression. "
                "Cannot be either: less than 0.25; greater than 3. (default=1.0)"
            ),
        )
