# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse

import numpy as np
import trimesh

from mpegi.framework import Component
from mpegi.geometry.trimesh import convertGeometriesToTrimesh
from mpegi.metadata.early.ERComponentData import ERComponentData
from mpegi.metadata.early.voxelDataGeneratorEarly import VoxelDataGeneratorEarly
from mpegi.voxel import VoxelizerER


class ERComponent(Component):
    def __init__(self, args, artifacts):
        # Initialize default values that this Component "Manifests"
        self.erData = ERComponentData()
        self.erData.surface = []
        self.erData.surfaceConnected = []

        scene_trimesh_ER_as_eif_mesh = artifacts["scene_trimesh_ER_as_eif_mesh"]
        scene_trimesh_ER = convertGeometriesToTrimesh(
            [scene_trimesh_ER_as_eif_mesh], faceAngle=30, process=False
        )[0]
        artifacts["scene_trimesh_ER"] = scene_trimesh_ER

        # Compute the area that the listener can reach in voxel form
        if scene_trimesh_ER.bounds is None:
            # creaty dummy voxel grid if scene contains no mesh
            voxel_encoding = trimesh.voxel.encoding.DenseEncoding(np.zeros((1, 1, 1), dtype=bool))
            self.reachableVoxels = trimesh.voxel.VoxelGrid(encoding=voxel_encoding)
            self.reachableVoxels.apply_scale(args["voxel_size_ER"])
            return
        else:
            erVoxelizer = VoxelizerER(args, artifacts)
            voxel_size_ER = args["voxel_size_ER"]
            user_regions = artifacts["user_regions"]
            erVoxelizer.analyze_listener_voxels(scene_trimesh_ER, voxel_size_ER, user_regions)
            self.reachableVoxels = erVoxelizer.reachableVoxels

        # get static source positions
        scene_metadata = artifacts["scene_metadata"]
        object_sources = artifacts["object_sources"]
        src_pos_list = [
            list(src.position) for src in object_sources if scene_metadata.isEntityStatic(src.id)
        ]

        # Early Reflection Voxel Data Generator
        grid_factor = 4
        self.generator = VoxelDataGeneratorEarly(
            args, artifacts, self.reachableVoxels, src_pos_list, grid_factor
        )

        # Use parallel processing
        self._run()

        # re-map triangle IDs of surface data
        mesh_unproc = artifacts["scene_trimesh_ER_unproc_as_eif_mesh"]
        self._remapCleanMesh(scene_trimesh_ER, mesh_unproc)

        # result data structure
        self.erData = self._getErComponentData()

    @staticmethod
    def getArgumentSubscription():
        return [
            "voxel_size_ER",
            "max_order_ER_indoor",
            "max_order_ER_outdoor",
            "max_dist_ER_triangle_1",
            "max_dist_ER_triangle_2",
            "max_dist_ER_source_1",
            "max_dist_ER_source_2",
        ]

    @staticmethod
    def getArtifactSubscription():
        return [
            "scene_trimesh_ER_as_eif_mesh",
            "scene_trimesh_ER_unproc_as_eif_mesh",
            "scene_metadata",
            "object_sources",
            "acoustic_environments",
            "acoustic_materials",
            "meshes",
            "user_regions",
        ]

    @staticmethod
    def getArtifactManifest():
        return ["ER_data", "reachable_voxels"]

    def getArtifacts(self):
        return {"ER_data": self.erData, "reachable_voxels": self.reachableVoxels}

    @staticmethod
    def getVersion():
        return 8

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        erParser = argParser.add_argument_group("Early Reflections")
        erParser.add_argument(
            "--max_order_ER_indoor",
            default=2,
            type=int,
            metavar="ORDER",
            help="Maximum reflection order for Early Reflections, indoor (default=2)",
        )
        erParser.add_argument(
            "--max_order_ER_outdoor",
            default=1,
            type=int,
            metavar="ORDER",
            help="Maximum reflection order for Early Reflections, outdoor (default=1)",
        )
        erParser.add_argument(
            "--max_dist_ER_triangle_1",
            default=45,
            type=float,
            metavar="DIST",
            help="Culling distance in m for triangles, 1st-order (default=45m)",
        )
        erParser.add_argument(
            "--max_dist_ER_triangle_2",
            default=15,
            type=float,
            metavar="DIST",
            help="Culling distance in m for triangles, 2nd-order (default=15m)",
        )
        erParser.add_argument(
            "--max_dist_ER_source_1",
            default=90,
            type=float,
            metavar="DIST",
            help="Culling distance in m for sources, 1st-order (default=90m)",
        )
        erParser.add_argument(
            "--max_dist_ER_source_2",
            default=30,
            type=float,
            metavar="DIST",
            help="Culling distance in m for sources, 2nd-order (default=30m)",
        )
        erParser.add_argument(
            "--voxel_size_ER",
            default=1.0,
            type=float,
            metavar="SIZE",
            help=("Determines the length, width, and depth of a Voxel (default=1m)"),
        )

    def _run(self):
        # generate voxel data
        self.generator.process_parallel()
        # self.generator.process_profile()

        # export as json file
        # with open("voxel_data_early.json", "w") as output_file:
        #    output_file.write(self.generator.vde.to_json())

    def _getErComponentData(self):
        erComponentData = ERComponentData()

        # add voxel data
        erComponentData.addVoxelDataEarly(self.generator.vde)

        # add geometric data: surfaces (facets)
        erComponentData.addSurface(self.generator.surface)

        return erComponentData

    def _remapCleanMesh(self, scene_trimesh_ER, mesh_unproc) -> None:
        """re-map triangle IDs of surface data"""
        from ...geometry.trimesh import convertGeometriesToTrimesh

        trimesh_unproc = convertGeometriesToTrimesh([mesh_unproc], faceAngle=30, process=False)[0]

        # generate pos -> vtxID map
        vertex_id_unproc = {}
        for id, vertex_pos in enumerate(trimesh_unproc.vertices):
            pos = vertex_pos.astype("single")
            vertex_id_unproc[tuple(pos)] = id

        # generate vtxID_proc -> vtxID_unproc map
        vertex_map = [None] * len(scene_trimesh_ER.vertices)
        for id_proc, vertex_pos in enumerate(scene_trimesh_ER.vertices):
            pos = vertex_pos.astype("single")
            id_unproc = vertex_id_unproc[tuple(pos)]
            vertex_map[id_proc] = id_unproc

        surface_list = self.generator.surface
        num_surfaces = len(surface_list)
        valid_vertex_list = [set()] * num_surfaces
        faceIdxNew_list = [list()] * num_surfaces
        for s in range(num_surfaces):
            valid_vertex_list[s] = set()
            faceIdxNew_list[s] = list()

        # analyze which vertices belong to which surface
        for s, surface_ref in enumerate(surface_list):
            for fid in surface_ref.faceIdx:
                face = scene_trimesh_ER.faces[fid]
                for vtx in face:
                    try:
                        # fast method: use vertex_map
                        vtx_new = vertex_map[vtx]
                    except IndexError:
                        # slow method: fuzzy search
                        pos = scene_trimesh_ER.vertices[vtx]
                        vtx_new = self._getSimilarVertex(trimesh_unproc, pos)
                    valid_vertex_list[s].add(vtx_new)

        for fid_new, face_new_ref in enumerate(trimesh_unproc.faces):
            # needed for face orientation check
            fnormal_new = trimesh_unproc.face_normals[fid_new]

            # use vtxID -> pos -> vtxID to avoid problems with duplicate vertices
            vertex0_pos = trimesh_unproc.vertices[face_new_ref[0]].astype("single")
            vertex1_pos = trimesh_unproc.vertices[face_new_ref[1]].astype("single")
            vertex2_pos = trimesh_unproc.vertices[face_new_ref[2]].astype("single")
            face_new = [0, 0, 0]
            face_new[0] = vertex_id_unproc[tuple(vertex0_pos)]
            face_new[1] = vertex_id_unproc[tuple(vertex1_pos)]
            face_new[2] = vertex_id_unproc[tuple(vertex2_pos)]

            for s, valid_vertex_ref in enumerate(valid_vertex_list):
                # needed for face orientation check
                fid0 = surface_list[s].faceIdx[0]
                fnormal0 = scene_trimesh_ER.face_normals[fid0]

                if (
                    face_new[0] in valid_vertex_ref
                    and face_new[1] in valid_vertex_ref
                    and face_new[2] in valid_vertex_ref
                    and np.dot(fnormal_new, fnormal0) > 0
                ):
                    faceIdxNew_list[s].append(fid_new)

        # replace faceIdx list
        for s, surface_ref in enumerate(surface_list):
            surface_ref.faceIdx = faceIdxNew_list[s]

    def _getSimilarVertex(self, mesh, pos) -> int:
        delta = 1e-4
        for vtx, vertex_ref in enumerate(trimesh.vertices):
            if np.linalg.norm(np.array(vertex_ref) - pos) < delta:
                return vtx

        return None
