# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import copy
import logging
import math
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Tuple
import itertools

import numpy as np

from mpegi.framework import Component

from mpegi.metadata.eifplus import SceneMetadata
import mpegi.framework._lookup_tables as LUT


class FGDMethod(Enum):
    INDIVIDUAL_FREQUENCIES = 0
    START_HOP_AMOUNT = 1
    DEFAULT_BANDING = 2
    RESERVED = 3


@dataclass
class freqGridData:
    fgdNrGrids: int
    fgdMethod: List[dict]
    uniqueFrequencyLayouts: List[List]


class FreqGridComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]) -> None:
        # TODO: remove includeMaterials bool and checks once fully integrated.
        includeMaterials = True
        self.uniqueFrequencyLayouts = []
        self.frequencyGrids = {}

        reverb_freq_grids = artifacts["reverb_frequency_grids"]
        airflow_freq_grids = artifacts["airflow_frequency_grids"]
        materials = artifacts["acoustic_materials"]
        headphoneEQ = artifacts["headphoneEQ"]
        metadata: SceneMetadata = artifacts["scene_metadata"]
        materials = [mat for mat in materials if metadata.isMaterialUsed(mat.id)]

        # add all grids used in the reverb AEs to the list of unique grids.
        for ae in reverb_freq_grids:
            self.uniqueFrequencyLayouts.append(reverb_freq_grids[ae].frequencies)

        for ae in airflow_freq_grids:
            self.uniqueFrequencyLayouts.append(airflow_freq_grids[ae].frequencies)

        if includeMaterials:
            for mat in materials:
                freqs = []
                for f in mat.frequencies:
                    freqs.append(f.f)
                mat.freqs = freqs
                self.uniqueFrequencyLayouts.append(freqs)

        if headphoneEQ[0].specType in ["freqgain", "parametricfilters"]:
            self.uniqueFrequencyLayouts.append(headphoneEQ[0].f)

        # remove duplicate freq layouts
        self.uniqueFrequencyLayouts.sort()
        self.uniqueFrequencyLayouts = list(
            self.uniqueFrequencyLayouts
            for self.uniqueFrequencyLayouts, _ in itertools.groupby(self.uniqueFrequencyLayouts)
        )

        # find which uniqueFrequenyGrid each AE uses, add mapping to frequency grids dict.
        for ae in reverb_freq_grids:
            for j in range(len(self.uniqueFrequencyLayouts)):
                if len(reverb_freq_grids[ae].frequencies) == len(
                    self.uniqueFrequencyLayouts[j]
                ) and np.allclose(
                    reverb_freq_grids[ae].frequencies,
                    self.uniqueFrequencyLayouts[j],
                    atol=0.01,
                ):
                    self.frequencyGrids.update({ae: j})
                    break

        for ae in airflow_freq_grids:
            for j in range(len(self.uniqueFrequencyLayouts)):
                if len(airflow_freq_grids[ae].frequencies) == len(
                    self.uniqueFrequencyLayouts[j]
                ) and np.allclose(
                    airflow_freq_grids[ae].frequencies,
                    self.uniqueFrequencyLayouts[j],
                    atol=0.01,
                ):
                    self.frequencyGrids.update({ae: j})
                    break

        # find which unique grid each material uses, and add mapping to dict.
        if includeMaterials:
            for mat in materials:
                for j in range(len(self.uniqueFrequencyLayouts)):
                    if len(mat.freqs) == len(self.uniqueFrequencyLayouts[j]) and np.allclose(
                        mat.freqs,
                        self.uniqueFrequencyLayouts[j],
                        atol=0.01,
                    ):
                        self.frequencyGrids.update({mat.id: j})
                        break

        if headphoneEQ[0].specType in ["freqgain", "parametricfilters"]:
            for j in range(len(self.uniqueFrequencyLayouts)):
                if len(headphoneEQ[0].f) == len(self.uniqueFrequencyLayouts[j]) and np.allclose(
                    headphoneEQ[0].f,
                    self.uniqueFrequencyLayouts[j],
                    atol=0.01,
                ):
                    self.frequencyGrids.update({"headphoneEQ": j})
                    break

        self.fgdNrGrids = len(self.uniqueFrequencyLayouts)
        self.fgdMethod = []

        # For each unique grid, find the type of grid.

        for j in range(len(self.uniqueFrequencyLayouts)):
            keyfound = False
            # first check if the freqGrid exists as a default grid
            for k in range(len(LUT.frequency_default_grids)):
                if len(self.uniqueFrequencyLayouts[j]) == len(
                    LUT.frequency_default_grids[k]
                ) and np.allclose(
                    self.uniqueFrequencyLayouts[j], LUT.frequency_default_grids[k], atol=0.01
                ):
                    self.fgdMethod.append(
                        {
                            "method": FGDMethod.DEFAULT_BANDING,
                            "value": k,
                        }
                    )
                    keyfound = True
                    break
            # if no full default grid, see if there is a partial grid.
            if keyfound == False:
                for k in range(len(LUT.frequency_default_grids)):
                    if all(
                        np.any(np.isclose(item, LUT.frequency_default_grids[k], atol=0.01))
                        for item in self.uniqueFrequencyLayouts[j]
                    ):
                        offset = LUT.frequency_default_grids[k].index(
                            self.uniqueFrequencyLayouts[j][0]
                        )
                        self.fgdMethod.append(
                            {
                                "method": FGDMethod.DEFAULT_BANDING,
                                "value": k,
                                "offset": offset,
                                "length": len(self.uniqueFrequencyLayouts[j]),
                            }
                        )
                        keyfound = True
                        break
            # if no grid at all, see if a freq hop code can be used
            if keyfound == False:
                for key in LUT.frequency_hop_codes.keys():
                    freqHopGrid = self.genFreqHopGrid(
                        key, self.uniqueFrequencyLayouts[j][0], len(self.uniqueFrequencyLayouts[j])
                    )
                    if np.allclose(self.uniqueFrequencyLayouts[j], freqHopGrid, atol=0.01):
                        self.fgdMethod.append(
                            {
                                "method": FGDMethod.START_HOP_AMOUNT,
                                "start": self.uniqueFrequencyLayouts[j][0],
                                "key": key,
                            }
                        )
                        keyfound = True
                        break
            # finally, default to individual frequencies.
            if keyfound is False:
                self.fgdMethod.append({"method": FGDMethod.INDIVIDUAL_FREQUENCIES})
        self.freqGridData = freqGridData(
            self.fgdNrGrids, self.fgdMethod, self.uniqueFrequencyLayouts
        )
        pass

    def genFreqHopGrid(self, hop, start, length):
        freqHopGrid = [start]
        for n in range(length - 1):
            freqHopGrid.append(freqHopGrid[-1] * hop)
        return freqHopGrid

    @staticmethod
    def getVersion() -> Any:
        return 1

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return [
            "scene",
        ]  ## subscribe the new arg

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "scene_metadata",
            "acoustic_environments",
            "acoustic_materials",
            "reverb_params",
            "reverb_frequency_grids",
            "airflow_frequency_grids",
            "headphoneEQ",
        ]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return [
            "frequency_grids",
            "frequency_grid_mapping",
        ]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "frequency_grids": self.freqGridData,
            "frequency_grid_mapping": self.frequencyGrids,
        }
