# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
from typing import List

from mpegi.framework import Component
from mpegi.metadata.eifplus import Box
from mpegi.metadata.eifplus.scene import SceneType
from mpegi.metadata.lcearly import (
    LCERComponentData,
    LCEREnvironment,
    LCEROutdoorData,
    new_er_outdoor,
    simple_er_analysis,
)


class LCERComponent(Component):
    def __init__(self, args, artifacts):
        self.LCER_predelay1 = 60
        self.LCER_predelay2 = 120
        self.LCER_ampFac1 = 0
        self.LCER_ampFac2 = 0

        self.addOutdoorData(artifacts)

    def addOutdoorData(self, artifacts):
        self.lcerComponentData = LCERComponentData()

        if not self.scene_needs_er_analysis(artifacts):
            return

        geometry = [
            *artifacts["boxes"],
            *artifacts["spheres"],
            *artifacts["cylinders"],
            *artifacts["meshes"],
        ]
        audio_sources = [
            *artifacts["object_sources"],
            *artifacts["channel_sources"],
            *artifacts["hoa_sources"],
        ]
        analysis = simple_er_analysis(
            scene_metadata=artifacts["scene_metadata"],
            geometry=geometry,
            env_regions=[ae.region for ae in artifacts["acoustic_environments"]],
            transforms=artifacts["transforms"],
            audio_sources=audio_sources,
            user_regions=artifacts["user_regions"],
        )
        if analysis[0] == None:
            self.lcerComponentData = LCERComponentData()
            return

        userRegionsDict = analysis[0]
        valuesArray = analysis[1]

        userRegionsForGeometries = artifacts["user_regions"]
        userRegionsForGeometriesDict = {urr.id: urr for urr in userRegionsForGeometries}

        for userRegion, id in userRegionsDict.items():
            box = Box(
                id="BoundingBox",
                position=userRegionsForGeometriesDict[userRegion].position,
                size=userRegionsForGeometriesDict[userRegion].size,
                orientation=userRegionsForGeometriesDict[userRegion].orientation,
                cspace=None,
                material=None,
                transform=None,
                anchor=None,
            )

            val = valuesArray[id]
            result = new_er_outdoor(val["distances"], val["prominences"])
            environment = LCEREnvironment(priority=0)

            # Acoustic tuning parameters
            acoustic_tuning_parms = artifacts["acoustic_tuning_params"]

            if (
                acoustic_tuning_parms[0].lcErOutdoorTuning is not None
            ):  # hasattr(acoustic_tuning_parms[0], 'lcErOutdoorTuning'):
                self.LCER_ampFac1 = acoustic_tuning_parms[0].lcErOutdoorTuning.ampFac1Db
                self.LCER_ampFac2 = acoustic_tuning_parms[0].lcErOutdoorTuning.ampFac2Db
                self.LCER_predelay1 = acoustic_tuning_parms[0].lcErOutdoorTuning.predelay1
                self.LCER_predelay2 = acoustic_tuning_parms[0].lcErOutdoorTuning.predelay2
            else:
                self.LCER_ampFac1 = result["ampFac1Db"]
                self.LCER_ampFac2 = result["ampFac2Db"]
                self.LCER_predelay1 = result["preDelay1"]
                self.LCER_predelay2 = result["preDelay2"]

            # round gains to one decimal
            self.LCER_ampFac1 = 10 * round(self.LCER_ampFac1 / 10)
            self.LCER_ampFac2 = 10 * round(self.LCER_ampFac2 / 10)

            # limit predelay to bitstream field range
            self.LCER_predelay1 = max(0, min(65535, self.LCER_predelay1))
            self.LCER_predelay2 = max(0, min(65535, self.LCER_predelay2))

            outdoorData = LCEROutdoorData(
                ampFac1=float(self.LCER_ampFac1),
                ampFac2=float(self.LCER_ampFac2),
                predelay1=int(self.LCER_predelay1),
                predelay2=int(self.LCER_predelay2),
                boundingBox=box,
            )
            environment.addOutdoorData(outdoorData)
            self.lcerComponentData.addLCEREnvironment(environment)

    def scene_needs_er_analysis(self, artifacts) -> bool:
        if artifacts["scene_info"]["type"] == SceneType.AR:
            return False
        return True

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return []

    @staticmethod
    def getArtifactSubscription():
        return [
            "acoustic_environments",
            "scene_info",
            "boxes",
            "spheres",
            "cylinders",
            "meshes",
            "scene_metadata",
            "transforms",
            "channel_sources",
            "hoa_sources",
            "object_sources",
            "user_regions",
            "acoustic_tuning_params",
        ]

    @staticmethod
    def getArtifactManifest():
        return ["lcerComponentData"]

    def getArtifacts(self):
        return {"lcerComponentData": self.lcerComponentData}

    @staticmethod
    def getVersion():
        return 2

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        LCERGroup = argParser.add_argument_group("Low Complexity Early Reflections")
