# This software module was originally developed jointly by
#
# Fraunhofer IIS and Qualcomm Technologies, Inc.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS and Qualcomm Technologies, Inc. retain
# full right to modify and use the code for their own purpose,
# assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import copy
import logging
import math
from typing import Any, Dict, List, Tuple

import numpy as np
import trimesh

from mpegi.framework import Component
from mpegi.geometry import createPortals
from mpegi.geometry.transforms import getOffsetAndAnglesFromMatrixTRS, transformMatrixTRS_4x4
from mpegi.geometry.trimesh import convertGeometriesToTrimesh, trimeshToEifMesh
from mpegi.metadata import eifplus
from mpegi.encoder import _constants as Constants

import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element


class PreprocessorComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]) -> None:
        self.sceneMetadata = eifplus.SceneMetadata()
        entitiesToAnalyze = []
        entitiesToAnalyze.extend(artifacts["boxes"])
        entitiesToAnalyze.extend(artifacts["cylinders"])
        entitiesToAnalyze.extend(artifacts["spheres"])
        entitiesToAnalyze.extend(artifacts["meshes"])
        entitiesToAnalyze.extend(artifacts["acoustic_materials"])
        entitiesToAnalyze.extend(artifacts["audio_streams"])
        entitiesToAnalyze.extend(artifacts["granular_sound"])
        entitiesToAnalyze.extend(artifacts["object_sources"])
        entitiesToAnalyze.extend(artifacts["channel_sources"])
        entitiesToAnalyze.extend(artifacts["hoa_sources"])
        entitiesToAnalyze.extend(artifacts["updates"])

        """Build list of all voxel blocks"""
        voxelBlocksToAnalyze = []
        root = ET.parse(args["scene"]).getroot()
        voxel_subscenes = root.findall("VoxelScene")

        for subscene in voxel_subscenes:
            voxelBlocksToAnalyze = voxelBlocksToAnalyze + subscene.findall("VoxelBlock")

        entitiesToAnalyze.extend(artifacts["portals"])
        self.sceneMetadata.analyze(
            entitiesToAnalyze, voxelBlocksToAnalyze, rootFolder=str(args["scene"].parent)
        )

        meshes = artifacts["meshes"]

        geometry = meshes + artifacts["boxes"] + artifacts["cylinders"] + artifacts["spheres"]
        environments = artifacts["acoustic_environments"]
        envRegions = [env.region for env in environments]

        transforms = artifacts["transforms"]
        geometry = PreprocessorComponent.applyParentTransforms(geometry, transforms)

        self.objectSources = artifacts["object_sources"]
        self.channelSources = artifacts["channel_sources"]
        self.hoaSources = artifacts["hoa_sources"]
        PreprocessorComponent.applyAudioSourceRescale(
            self.objectSources, self.channelSources, self.hoaSources, artifacts["anchors"]
        )

        # automatic detection of portals only if there are no portals explicit defined
        if len(artifacts["portals"]) == 0:
            portalGeo = artifacts["boxes"] + meshes
            envNames = [env.id for env in environments]
            self.portalInfo = createPortals(
                environments,
                portalGeo,
                artifacts["anchors"],
                False,
                artifacts["updates"],
                args["reverbAsSess"],
            )
            entitiesToAnalyze.extend(self.portalInfo.portalGeometries)
            self.sceneMetadata.analyze(
                entitiesToAnalyze, voxelBlocksToAnalyze, rootFolder=str(args["scene"].parent)
            )
        else:
            self.portalInfo = []

        self.sceneMetadata.earlyStaticMeshIds = self.getEarlyStaticMeshIds(
            geometry, envRegions, self.sceneMetadata
        )

        unproc_mesh, proc_mesh = self.makeSceneTrimeshDiffr(geometry, self.sceneMetadata)
        self.sceneTrimeshDiffrUnprocAsEifMesh = unproc_mesh
        self.sceneTrimeshDiffrAsEifMesh = proc_mesh

        self.sceneTrimeshERUnprocAsEifMesh, self.sceneTrimeshERAsEifMesh = self.makeSceneTrimeshER(
            geometry, self.sceneMetadata
        )

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return [
            "scene",
            "reverbAsSess",
        ]  ## subscribe the new arg

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "acoustic_environments",
            "acoustic_materials",
            "anchors",
            "audio_streams",
            "granular_sound",
            "boxes",
            "channel_sources",
            "cylinders",
            "hoa_sources",
            "meshes",
            "object_sources",
            "spheres",
            "transforms",
            "updates",
            "portals",
        ]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return [
            "scene_metadata",
            "scene_trimesh_ER_as_eif_mesh",
            "scene_trimesh_ER_unproc_as_eif_mesh",
            "scene_trimesh_DIFFR_as_eif_mesh",
            "scene_trimesh_DIFFR_unproc_as_eif_mesh",
            "portal_info",
        ]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "scene_metadata": self.sceneMetadata,
            "scene_trimesh_ER_as_eif_mesh": self.sceneTrimeshERAsEifMesh,
            "scene_trimesh_ER_unproc_as_eif_mesh": self.sceneTrimeshERUnprocAsEifMesh,
            "scene_trimesh_DIFFR_as_eif_mesh": self.sceneTrimeshDiffrAsEifMesh,
            "scene_trimesh_DIFFR_unproc_as_eif_mesh": self.sceneTrimeshDiffrUnprocAsEifMesh,
            "portal_info": self.portalInfo,
            "object_sources": self.objectSources,
            "channel_sources": self.channelSources,
            "hoa_sources": self.hoaSources,
        }

    @staticmethod
    def getVersion() -> Any:
        return 12

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        preprocParser = argParser.add_argument_group("Preprocess")
        preprocParser.add_argument(
            "--reverbAsSess",
            action="store_true",
            help="If true, export reverb as portal extent.",
        )

    @staticmethod
    def makeSceneTrimeshDiffr(
        geometry: List[eifplus.Geometry],
        metadata: eifplus.SceneMetadata,
    ) -> (eifplus.Mesh, eifplus.Mesh):
        # append cylinders at the end of the Diffr geometry
        cylinderGeometries = PreprocessorComponent.onlyCylinders(geometry, metadata)
        geometry = [mesh for mesh in geometry if mesh.id in metadata.earlyStaticMeshIds]
        for cg in cylinderGeometries:
            geometry.append(cg)

        # convert to Trimesh without cleanup (first mesh variant)
        sceneTrimeshDiffrUnproc = convertGeometriesToTrimesh(
            geometry, faceAngle=30, subdivisions=0, process=False
        )[0]

        # convert combined Trimesh to EIF mesh
        parentAnchors = [g.anchor for g in geometry if g.id in geometry]
        if len(set(parentAnchors)) > 1:
            raise AssertionError("SceneTrimeshDiffr cannot have multiple parent anchors")
        parentAnchor = None if len(parentAnchors) == 0 else parentAnchors[0]
        sceneTrimeshDiffrUnprocAsEifMesh = trimeshToEifMesh(
            sceneTrimeshDiffrUnproc, "mesh:_combinedDiffractionMesh", parentAnchor
        )

        # cleanup geometries
        geometry = PreprocessorComponent.removeDiminuitive(geometry)

        # convert to Trimesh and cleanup (second mesh variant)
        sceneTrimeshDiffr = convertGeometriesToTrimesh(
            geometry, faceAngle=30, subdivisions=0, process=True
        )[0]
        PreprocessorComponent._cleanupMesh(sceneTrimeshDiffr)

        # convert combined clean Trimesh to EIF mesh
        parentAnchors = [g.anchor for g in geometry if g.id in geometry]
        if len(set(parentAnchors)) > 1:
            raise AssertionError("SceneTrimeshDiffr cannot have multiple parent anchors")
        parentAnchor = None if len(parentAnchors) == 0 else parentAnchors[0]
        sceneTrimeshDiffrAsEifMesh = trimeshToEifMesh(
            sceneTrimeshDiffr, "mesh:_combinedDiffractionMeshCleaned", parentAnchor
        )

        return sceneTrimeshDiffrUnprocAsEifMesh, sceneTrimeshDiffrAsEifMesh

    @staticmethod
    def makeSceneTrimeshER(
        geometry: List[eifplus.Geometry],
        metadata: eifplus.SceneMetadata,
    ) -> Tuple[trimesh.Trimesh, List[str]]:
        # list of geometric objects which belong to the static ER mesh according to getEarlyStaticMeshIds()
        geometryER = [mesh for mesh in geometry if mesh.id in metadata.earlyStaticMeshIds]

        # convert to combined Trimesh
        sceneTrimeshERUnproc = convertGeometriesToTrimesh(
            geometryER, faceAngle=30, subdivisions=0, process=False
        )[0]

        # convert combined Trimesh to EIF mesh
        parentAnchors = [g.anchor for g in geometry if g.id in geometryER]
        if len(set(parentAnchors)) > 1:
            raise AssertionError("SceneTrimeshER cannot have multiple parent anchors")
        parentAnchor = None if len(parentAnchors) == 0 else parentAnchors[0]
        sceneTrimeshERUnprocAsEifMesh = trimeshToEifMesh(
            sceneTrimeshERUnproc, "mesh:_combinedEarlyReflectionMesh", parentAnchor
        )

        # convert to combined clean Trimesh
        sceneTrimeshER = convertGeometriesToTrimesh(
            geometryER, faceAngle=30, subdivisions=0, process=True
        )[0]
        PreprocessorComponent._cleanupMesh(sceneTrimeshER)

        # convert combined clean Trimesh to EIF mesh
        parentAnchors = [g.anchor for g in geometry if g.id in geometryER]
        if len(set(parentAnchors)) > 1:
            raise AssertionError("SceneTrimeshER cannot have multiple parent anchors")
        parentAnchor = None if len(parentAnchors) == 0 else parentAnchors[0]
        sceneTrimeshERAsEifMesh = trimeshToEifMesh(
            sceneTrimeshER, "mesh:_combinedEarlyReflectionMesh", parentAnchor
        )

        return sceneTrimeshERUnprocAsEifMesh, sceneTrimeshERAsEifMesh

    @staticmethod
    def getEarlyStaticMeshIds(
        geometry: List[eifplus.Geometry],
        envIds: List[str],
        metadata: eifplus.SceneMetadata,
    ) -> List[str]:
        geometry = copy.deepcopy(geometry)
        geometry = PreprocessorComponent.removeDynamic(geometry, metadata)
        geometry = PreprocessorComponent.removeAcousticEnvironments(geometry, envIds)
        geometry = PreprocessorComponent.removeTransparent(geometry)
        geometry = PreprocessorComponent.removeCylinders(geometry)

        earlyStaticMeshIds = [mesh.id for mesh in geometry]

        return earlyStaticMeshIds

    @staticmethod
    def applyParentTransforms(
        geometry: List[eifplus.Geometry], transforms: List[eifplus.Transform]
    ):
        for geo in geometry:
            if geo.transform != None:
                # get parent transform
                transform = [t for t in transforms if t.id == geo.transform][0]

                Tgeo = transformMatrixTRS_4x4(geo.position, np.deg2rad(geo.orientation), [1, 1, 1])
                Ttransform = transformMatrixTRS_4x4(
                    transform.position, np.deg2rad(transform.orientation), [1, 1, 1]
                )
                Tglobal = np.matmul(Ttransform, Tgeo)
                newPos, newOri = getOffsetAndAnglesFromMatrixTRS(Tglobal)

                geo.position = newPos.tolist()
                geo.orientation = np.rad2deg(newOri).tolist()

        return geometry

    @staticmethod
    def applyAudioSourceRescale(
        object_sources: List[eifplus.ObjectSource],
        channel_sources: List[eifplus.ChannelSource],
        hoa_sources: List[eifplus.HOASource],
        anchors: List[eifplus.Anchor],
    ):
        for anchor in anchors:
            if anchor.autoRescale != True:
                for obj in object_sources:
                    if obj.anchor == anchor.id:
                        obj.position = obj.position * anchor.audioSourceRescale
                for ch in channel_sources:
                    if ch.anchor == anchor.id:
                        ch.position = ch.position * anchor.audioSourceRescale
                for hoa in hoa_sources:
                    if hoa.anchor == anchor.id:  # verifying that the source belongs to the anchor
                        hoa.position = hoa.position * anchor.audioSourceRescale

    @staticmethod
    def removeAcousticEnvironments(
        geometry: List[eifplus.Geometry], envRegions: List[str]
    ) -> List[eifplus.Geometry]:
        for envRegion in envRegions:
            if envRegion in [geo.id for geo in geometry]:
                geometryIds = [geo.id for geo in geometry]
                index = geometryIds.index(envRegion)
                geometry.pop(index)
        return geometry

    @staticmethod
    def removeTransparent(geometry: List[eifplus.Geometry]) -> List[eifplus.Geometry]:
        return [g for g in geometry if not g.isTransparent]

    @staticmethod
    def removeDiminuitive(
        geometry: List[eifplus.Geometry], minimumLength: float = 0.5
    ) -> List[eifplus.Geometry]:
        """Removes geometry if the size of bounding box is smaller than the
        minimum length in two or three dimesions
        """
        finalGeometry = []
        for g in geometry:
            gTrimesh = convertGeometriesToTrimesh([g])[0]
            bounds = gTrimesh.bounds
            boundingBoxSize = np.abs(np.subtract(bounds[0], bounds[1]))
            if np.count_nonzero(boundingBoxSize > minimumLength) > 1:
                finalGeometry.append(g)
            else:
                logging.debug(
                    ">> Removing geometry that is shorter than "
                    f"{minimumLength} in two or three dimenions: {g.id, type(g)}"
                )
        return finalGeometry

    @staticmethod
    def removeCylinders(
        geometry: List[eifplus.Geometry],
    ) -> List[eifplus.Geometry]:
        return [g for g in geometry if not isinstance(g, eifplus.Cylinder)]

    @staticmethod
    def onlyCylinders(
        geometry: List[eifplus.Geometry], metadata: eifplus.SceneMetadata
    ) -> List[eifplus.Geometry]:
        return [
            g
            for g in geometry
            if isinstance(g, eifplus.Cylinder)
            and metadata.isEntityStatic(g.id)
            and not g.isTransparent
        ]

    @staticmethod
    def removeDynamic(
        geometry: List[eifplus.Geometry], metadata: eifplus.SceneMetadata
    ) -> List[eifplus.Geometry]:
        return [g for g in geometry if metadata.isEntityStatic(g.id)]

    @staticmethod
    def enlargeCylinders(
        geometry: List[eifplus.Geometry], faceAngle: float = 30
    ) -> List[eifplus.Geometry]:
        for geo in geometry:
            if isinstance(geo, eifplus.Cylinder):
                geo.size[0] *= 1 / math.cos((faceAngle / 2) / 180 * math.pi)
                geo.size[2] *= 1 / math.cos((faceAngle / 2) / 180 * math.pi)
        return geometry

    @staticmethod
    def cleanupEifplusMesh(mesh: eifplus.Mesh) -> None:
        # save position, orientation, and size of the mesh and reset it to no transformation
        saved_location = copy.deepcopy(mesh.location)
        mesh.location.position = [0, 0, 0]
        mesh.location.orientation = [0, 0, 0]
        if hasattr(mesh, "size"):
            old_size = mesh.size
            mesh.size = [1, 1, 1]

        # conversion to Trimesh and cleanup
        tmesh = convertGeometriesToTrimesh([mesh])[0]
        PreprocessorComponent._cleanupMesh(tmesh)

        # replace faces and vertices by cleaned up variants from Trimesh
        matIndices = tmesh.face_attributes["matIndices"]
        materials = tmesh.metadata["materials"]
        mesh.materials = [materials[idx] for idx in matIndices]
        mesh.faces = tmesh.faces.copy()
        mesh.vertices = tmesh.vertices.copy()

        # restore position, orientation, and size
        mesh.location = saved_location
        if hasattr(mesh, "size"):
            mesh.size = old_size

    @staticmethod
    def _cleanupMesh(tmesh: trimesh.Trimesh) -> None:
        tmesh.remove_degenerate_faces()
        tmesh.remove_duplicate_faces()
        tmesh.remove_infinite_values()
        tmesh.remove_unreferenced_vertices()
