# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, List, Optional
import argparse
import logging
import numpy as np
from dataclasses import dataclass

import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element

from mpegi.framework import Component

from .._payloads._vox_diffraction import vox_diffraction


@dataclass
class SubSceneStruct:
    """Class for voxel sub scene data."""

    pm: np.ndarray
    vox_sources: dict
    voxel_size: float


class DiffractionTracer:
    """Class for tracing diffracted sound using a voxel grid."""

    def __init__(self, is_air: np.ndarray, layer: int):
        """Constructor"""

        self.width = is_air.shape[0]
        self.height = is_air.shape[1]
        self.layer = layer
        self.isAir = is_air
        self.dist = -np.ones((self.width, self.height, layer))
        self.waypoint = -np.ones((self.width, self.height, layer, 3), dtype=np.int_)

        # FIXME: remove this!
        framecounter = 0

    def trace(self, pos: tuple) -> None:
        # reset distance matrix
        self.dist.fill(float(self.width * self.height * self.layer))
        self.dist[pos] = 0

        # set waypoint
        self.waypoint[:, :, :, 0] = -1  # x-coordinate
        self.waypoint[:, :, :, 1] = -1  # y-coordinate
        self.waypoint[:, :, :, 2] = 0  # l-coordinate
        self.waypoint[pos] = pos

        # add source position to update list
        updateList = [pos]

        # perform initial tracing algorithm
        self._traceImpl(0, updateList)

        # perform re-tracing for each additional layer
        for new_layer in range(1, self.layer):
            self._retrace(new_layer)

        # remove irrelevant waypoints
        self._reduce(pos)

        # FIXME: remove this debug code!
        # undef = self.width * self.height * self.layer
        # max_dist = np.max(self.dist[self.dist != undef]) + 1.0
        # for d in np.arange(0.0, max_dist, 0.5):
        #    self._plotWaveFront(d)

    def _retrace(self, new_layer: int) -> None:
        updateList = []
        old_layer = new_layer - 1

        # check neighbors: left-right
        for y in range(self.height):
            for x in range(self.width - 1):
                posLeft = (x, y, old_layer)
                posRight = (x + 1, y, old_layer)
                wpLeft = tuple(self.waypoint[posLeft])
                wpRight = tuple(self.waypoint[posRight])
                if not self.isAir[x, y] or not self.isAir[x + 1, y] or wpLeft == wpRight:
                    continue
                if not self._isSameWave(posLeft, posRight):
                    # wave left->right
                    posNew = (posRight[0], posRight[1], new_layer)
                    origin = wpLeft
                    if not self._isVisible(origin, posNew):
                        origin = posLeft
                    delta = np.array(posNew) - np.array(origin)
                    deltaDist = np.sqrt([delta[0] * delta[0] + delta[1] * delta[1]])
                    distNew = self.dist[origin] + deltaDist
                    if distNew < self.dist[posNew]:
                        self.dist[posNew] = distNew
                        self.waypoint[posNew] = origin
                        updateList.append(posNew)

                    # wave right->left
                    posNew = (posLeft[0], posLeft[1], new_layer)
                    origin = wpRight
                    if not self._isVisible(origin, posNew):
                        origin = posRight
                    delta = np.array(posNew) - np.array(origin)
                    deltaDist = np.sqrt([delta[0] * delta[0] + delta[1] * delta[1]])
                    distNew = self.dist[origin] + deltaDist
                    if distNew < self.dist[posNew]:
                        self.dist[posNew] = distNew
                        self.waypoint[posNew] = origin
                        updateList.append(posNew)

        # check neighbors: up-down
        for y in range(self.height - 1):
            for x in range(0, self.width):
                posUp = (x, y, old_layer)
                posDown = (x, y + 1, old_layer)
                wpUp = tuple(self.waypoint[posUp])
                wpDown = tuple(self.waypoint[posDown])
                if not self.isAir[x, y] or not self.isAir[x, y + 1] or wpUp == wpDown:
                    continue

                if not self._isSameWave(posUp, posDown):
                    # wave up->down
                    posNew = (posDown[0], posDown[1], new_layer)
                    origin = wpUp
                    if not self._isVisible(origin, posNew):
                        origin = posUp
                    delta = np.array(posNew) - np.array(origin)
                    deltaDist = np.sqrt([delta[0] * delta[0] + delta[1] * delta[1]])
                    distNew = self.dist[origin] + deltaDist
                    if distNew < self.dist[posNew]:
                        self.dist[posNew] = distNew
                        self.waypoint[posNew] = origin
                        updateList.append(posNew)

                    # wave down->up
                    posNew = (posUp[0], posUp[1], new_layer)
                    origin = wpDown
                    if not self._isVisible(origin, posNew):
                        origin = posDown
                    delta = np.array(posNew) - np.array(origin)
                    deltaDist = np.sqrt([delta[0] * delta[0] + delta[1] * delta[1]])
                    distNew = self.dist[origin] + deltaDist
                    if distNew < self.dist[posNew]:
                        self.dist[posNew] = distNew
                        self.waypoint[posNew] = origin
                        updateList.append(posNew)

        # FIXME: remove this debug code
        # self._plotUpdateList(updateList)

        self._traceImpl(new_layer, updateList)

    def _traceImpl(self, new_layer: int, _updateList: List[tuple]) -> None:
        updateList = list(_updateList)
        neighborList = [
            np.array([1, 0, 0]),
            np.array([0, 1, 0]),
            np.array([-1, 0, 0]),
            np.array([0, -1, 0]),
        ]

        isSource = np.zeros((self.width, self.height), dtype=np.bool_)
        for pos in updateList:
            isSource[pos[0], pos[1]] = True

        while len(updateList) > 0:

            updateListNew = []

            for pos in updateList:
                posNPA = np.array(pos)
                for neighbor in neighborList:
                    posNew = tuple(posNPA + neighbor)

                    # skip out-of-bounds and non-air neighbors and new positions with old waypoints
                    if (
                        posNew[0] < 0
                        or posNew[1] < 0
                        or posNew[0] >= self.width
                        or posNew[1] >= self.height
                        or not self.isAir[posNew[0], posNew[1]]
                        or isSource[posNew[0], posNew[1]]
                        or (self.waypoint[posNew] == self.waypoint[pos]).all()
                    ):
                        continue

                    # check if the waypoint for the current position is visible at the new position
                    origin = tuple(self.waypoint[pos])
                    if not self._isVisible(origin, posNew):
                        origin = pos

                    # check if we found a shorter distance
                    delta = np.array(posNew) - np.array(origin)
                    deltaDist = np.sqrt([delta[0] * delta[0] + delta[1] * delta[1]])
                    distNew = self.dist[origin] + deltaDist
                    if distNew < self.dist[posNew]:
                        self.dist[posNew] = distNew
                        self.waypoint[posNew] = origin
                        updateListNew.append(posNew)

            # self._plotResult(new_layer)

            # replace update list
            updateList = updateListNew

    def _isVisible(self, pos1: tuple, pos2: tuple) -> bool:
        quick_test = True
        if pos1[0] < pos2[0]:
            quick_test = (self.waypoint[(pos2[0] - 1, pos2[1], pos2[2])] == pos1).all()
        elif pos1[0] > pos2[0]:
            quick_test = (self.waypoint[(pos2[0] + 1, pos2[1], pos2[2])] == pos1).all()
        if pos1[1] < pos2[1]:
            quick_test &= (self.waypoint[(pos2[0], pos2[1] - 1, pos2[2])] == pos1).all()
        elif pos1[1] > pos2[1]:
            quick_test &= (self.waypoint[(pos2[0], pos2[1] + 1, pos2[2])] == pos1).all()
        if quick_test:
            return True

        x = pos1[0]
        y = pos1[1]
        dx = abs(pos2[0] - x)
        sx = 1 if x < pos2[0] else -1
        dy = -abs(pos2[1] - y)
        sy = 1 if y < pos2[1] else -1
        err = dx + dy

        while x != pos2[0] or y != pos2[1]:
            e2 = 2 * err
            if e2 > dy:
                err += dy
                x += sx
            if e2 < dx:
                err += dx
                y += sy
            if not self.isAir[x, y]:
                return False

        return True

    def _isSameWave(self, pos1: tuple, pos2: tuple) -> bool:
        wp1 = tuple(self.waypoint[pos1])
        wp2 = tuple(self.waypoint[pos2])
        if wp1 == wp2:
            return True

        # generate roundtrip path
        path1 = self._getPath(pos1)
        path2 = self._getPath(pos2)
        mid_pos = 0.5 * (np.array(pos1) + np.array(pos2))
        roundtrip = list(reversed(path1)) + [mid_pos] + path2
        x = [p[0] for p in roundtrip]
        y = [p[1] for p in roundtrip]

        # compute enclosed area and thickness measure
        area = 0.0
        length = 0.0
        num_points = len(x)
        for n in range(1, num_points):
            a_x = x[n - 1]
            a_y = y[n - 1]
            b_x = x[n]
            b_y = y[n]
            delta_x = b_x - a_x
            delta_y = b_y - a_y
            area += a_x * b_y - a_y * b_x  # z-component of the 3D cross-product
            length += np.sqrt(delta_x * delta_x + delta_y * delta_y)
        area *= 0.5
        thickness = np.fabs(area) / (0.5 * length)

        return thickness < 1.0

    def _isNeighbor(self, pos1: tuple, pos2: tuple) -> bool:
        return abs(pos1[0] - pos2[0]) <= 1 and abs(pos1[1] - pos2[1]) <= 1

    def _getPath(self, pos: tuple) -> List:
        path = []
        wp = tuple(self.waypoint[pos])
        while pos[0] != wp[0] or pos[1] != wp[1]:
            pos = wp
            wp = tuple(self.waypoint[wp])
            path.append(pos)

        return path

    def _reduce(self, src_pos: tuple) -> None:

        wp_matrix = -np.ones((self.width, self.height, self.layer, 3), dtype=np.int_)
        wp_matrix[:, :, :, 0] = -1  # x-coordinate
        wp_matrix[:, :, :, 1] = -1  # y-coordinate
        wp_matrix[:, :, :, 2] = 0  # l-coordinate
        wp_matrix[src_pos] = src_pos

        # generate visibility matrix
        visible_matrix = np.zeros((self.width, self.height), dtype=np.int_)
        for y in range(self.height):
            for x in range(self.width):
                if not self.isAir[x, y]:
                    continue
                pos = (x, y, 0)
                wp = tuple(self.waypoint[pos])
                if wp == src_pos:
                    visible_matrix[x, y] = 1

        # reduce visible area by 1 voxel
        for y in range(self.height):
            vec_in = visible_matrix[:, y]
            vec_out = np.convolve(vec_in, np.array([1, 1, 1]))
            vec_out[1] += vec_out[0]
            vec_out[-2] += vec_out[-1]
            vec_out = np.floor_divide(vec_out[1:-1], 3)
            visible_matrix[:, y] = vec_out
        for x in range(self.width):
            vec_in = visible_matrix[x, :]
            vec_out = np.convolve(vec_in, np.array([1, 1, 1]))
            vec_out[1] += vec_out[0]
            vec_out[-2] += vec_out[-1]
            vec_out = np.floor_divide(vec_out[1:-1], 3)
            visible_matrix[x, :] = vec_out

        # first layer
        for y in range(self.height):
            for x in range(self.width):
                pos = (x, y, 0)
                wp = tuple(self.waypoint[pos])

                # check if voxel has a valid waypoint
                if wp[0] < 0 or wp[1] < 0:
                    continue

                # check if voxel is visible
                if visible_matrix[x, y]:
                    continue

                # add all waypoints of the diffraction path
                wp_matrix[pos] = wp
                for wp in self._getPath(pos):
                    wp_matrix[wp] = self.waypoint[wp]

        # second and higher layers
        for l in range(1, self.layer):
            for y in range(self.height):
                for x in range(self.width):
                    pos = (x, y, l)
                    wp = tuple(self.waypoint[pos])

                    # check if voxel has a valid waypoint
                    if wp[0] < 0 or wp[1] < 0:
                        continue

                    # check if voxel is visible
                    if visible_matrix[x, y]:
                        continue

                    # add all waypoints of the diffraction path
                    wp_matrix[pos] = wp
                    for wp in self._getPath(pos):
                        wp_matrix[wp] = self.waypoint[wp]

        self.waypoint = wp_matrix

    def _plotResult(self, l) -> None:
        """Debug routine: plot result"""
        import matplotlib.pyplot as plt

        undef = self.width * self.height * self.layer

        tmp = np.copy(self.dist[:, :, l])
        tmp[tmp == undef] = -1.0
        plt.clf()
        plt.matshow(tmp, fignum=1)
        plt.title("Distance")
        plt.draw()
        plt.pause(0.010)

    def _plotWaveFront(self, d) -> None:
        """Debug routine: plot wave front"""
        import matplotlib.pyplot as plt

        tmp = 0.0 - self.isAir
        for l in range(self.layer):
            tmp += np.maximum(1.0 - np.fabs(d - self.dist[:, :, l]), 0.0)
        plt.clf()
        plt.matshow(tmp, fignum=1)
        plt.title("Wave Front")
        plt.draw()
        plt.pause(0.010)

        try:
            filename = "video_frames/frame_" + str(self.framecounter).zfill(6) + ".png"
            plt.savefig(filename)
            self.framecounter += 1
        except:
            pass

    def _plotUpdateList(self, updateList) -> None:
        """Debug routine: plot update list"""
        import matplotlib.pyplot as plt

        tmp = 1 - self.isAir
        for pos in updateList:
            tmp[pos[0], pos[1]] = 2
        plt.clf()
        plt.matshow(tmp, fignum=1)
        plt.title("UpdateList")
        plt.draw()
        plt.pause(5.0)


class RasterMapComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]) -> None:
        super().__init__(args, artifacts)

        # Get encoding mode for static path data
        voxel_config = artifacts["voxel_config"]
        if artifacts["voxel_mode"]:
            voxel_config["mode"] = artifacts["voxel_mode"]
        elif "mode" not in voxel_config:
            voxel_config["mode"] = "low_bitrate"

        # check encoding mode: ["low_bitrate", "jps_selective", "jps_full", "multi_path", "multi_path_or_jps"]
        skip_rastermap = not self._isStatic(artifacts)
        skip_rastermap |= voxel_config["mode"] == "low_bitrate"
        skip_rastermap |= voxel_config["mode"] == "jps_selective"
        skip_rastermap |= voxel_config["mode"] == "jps_full"
        if skip_rastermap:
            self.rasterMapList = None
            return

        self.bs_ids = artifacts["bs_ids"]
        self.object_sources = artifacts["object_sources"]

        self.transparency_map = {"": True}
        for material in artifacts["acoustic_materials"]:
            is_transparent = any(band.t >= 0.5 for band in material.frequencies)
            self.transparency_map[material.id] = is_transparent

        # FIXME: replace this
        self.sss_list = self.__load_subscene_data(args, artifacts)

        num_layers = args["raster_map_order"]
        self.rasterMapList = self._generateRasterMap(num_layers)

    def _isStatic(self, artifacts: Dict[str, Any]) -> bool:
        updates = artifacts["updates"]
        isStatic = len(updates) == 0
        if not isStatic:
            logging.getLogger("RasterMapComponent").warning("Voxel scene is not static")

        return isStatic

    def _generateRasterMap(self, num_layers: int):
        subsceneRasterMapList = []

        # FIXME: remove this!
        framecounter = 0

        # iterate over all voxel sub-scenes
        for sss in self.sss_list:

            rasterMapDict = {}
            is_air = sss.pm == 0

            # iterate over all sources which belong to this voxel sub-scene
            for src_id, voxel_coord in sss.vox_sources.items():
                pos = (voxel_coord[0], voxel_coord[1], 0)
                if pos in rasterMapDict:
                    rasterMapDict[pos]["src_id"].append(src_id)
                    continue
                else:
                    rasterMapDict[pos] = {"src_id": [src_id]}

                tracer = DiffractionTracer(is_air, num_layers)
                tracer.framecounter = framecounter
                tracer.trace(pos)
                framecounter = tracer.framecounter

                rasterMapDict[pos]["wp_matrix"] = tracer.waypoint

            rasterMapList = []
            for rmDict in rasterMapDict.values():
                rasterMapList.append(rmDict)

            subsceneRasterMapList.append(rasterMapList)

        return subsceneRasterMapList

    def __load_subscene_data(self, args, artifacts) -> List[SubSceneStruct]:
        # FIXME: this part should be provided by the VoxelSceneComponent as artifact

        # return value
        sss_list = []

        logger: logging.Logger = logging.getLogger("RasterMapComponent")

        voxel_config = artifacts["voxel_config"]

        if not "subscenes" in voxel_config:
            voxel_config["subscenes"] = []

        try:
            root = ET.parse(args["scene"]).getroot()

            voxel_subscenes = root.findall("VoxelScene")

            for s, subscene in enumerate(voxel_subscenes):
                # collect objects voxel coords for this subscene
                vox_sources = {}
                for obj in self.object_sources:
                    src_id = self.bs_ids[obj.id]
                    if obj.extent is None:
                        obj_pos = obj.position
                    else:
                        # use mesh position for sources with extent
                        meshes = artifacts["meshes"]
                        mesh_match = (mesh for mesh in meshes if mesh.id == obj.extent)
                        mesh_extent = next(mesh_match)
                        obj_pos = mesh_extent.position
                    voxel_coord = self.__position_to_voxel_coord(subscene, obj_pos)
                    if len(voxel_coord):
                        vox_sources[src_id] = voxel_coord

                blocks = subscene.findall("VoxelBlock")
                if not blocks:
                    raise Exception("No voxel blocks found in {}".format(args["scene"]))

                voxel_blocks = self.__load_voxel_blocks(blocks, self.transparency_map)

                # Subscene config
                subscene_config: Optional[Dict[str, Any]] = next(
                    filter(
                        lambda ss: ss["id"] == subscene.attrib["id"],
                        voxel_config["subscenes"],
                    ),
                    {},
                )

                voxel_dimensions: List[int] = [
                    int(val) for val in subscene.attrib["vox_resolution"].split()
                ]

                # Setup projection map
                pm = vox_diffraction.projection_map(voxel_dimensions[0], voxel_dimensions[1])

                # Diffraction data (diffraction map and pre-computed path data)
                diffraction_map: Dict[str, Any] = (
                    subscene_config["diffraction_map"]
                    if "diffraction_map" in subscene_config
                    else []
                )

                if diffraction_map == []:
                    logger.warning(
                        "No diffraction map provided. Generating diffraction map from voxel blocks"
                    )

                    # build projection map from voxel blocks
                    z = int(voxel_dimensions[2] / 2)

                    for block in voxel_blocks:
                        start: List[int] = block["grid_coordinate_s"]
                        end: List[int] = block["grid_coordinate_e"]
                        value = block["diffraction_enabled"]

                        for x in range(start[0] - 1, end[0]):
                            for y in range(start[1] - 1, end[1]):
                                if start[2] - 1 <= z < end[2]:
                                    pm.set(x, y, chr(value))
                else:
                    for el in diffraction_map:
                        value = bool(int(el["enabled"]))
                        x0 = int(el["p0"][0])
                        y0 = int(el["p0"][1])
                        x1 = int(el["p1"][0])
                        y1 = int(el["p1"][1])

                        # Indices from the bitstream are in the range [1..dim]
                        for x in range(x0, x1 + 1, 1):
                            for y in range(y0, y1 + 1, 1):
                                pm.set(x - 1, y - 1, chr(value))

                # encode RasterMaps
                pm_nparray = np.zeros((voxel_dimensions[0], voxel_dimensions[1]), dtype=np.int_)
                for x in range(voxel_dimensions[0]):
                    for y in range(voxel_dimensions[1]):
                        val_bytes = bytes(pm.get(x, y), "utf-8")
                        pm_nparray[x, y] = int.from_bytes(val_bytes, "little")

                # extract voxel_size
                pos_s: List[int] = [float(val) for val in subscene.attrib["position_s"].split()]
                pos_e: List[int] = [float(val) for val in subscene.attrib["position_e"].split()]
                voxel_size = (pos_e[0] - pos_s[0]) / voxel_dimensions[0]

                sss_list.append(SubSceneStruct(pm_nparray, vox_sources, voxel_size))

        except Exception as exc:
            logger.exception("Error parsing voxel scene")

        return sss_list

    @staticmethod
    def __load_voxel_blocks(blocks: List[Element], transparency_map: Dict) -> List:
        # FIXME: this is only needed by __load_subscene_data() which should be replaced
        """
        Voxel blocks composing the scene geometry. A block defines a region in the voxel-scene
        that has the same acoustic properties (e.g. material, sound-medium)
        """

        voxel_blocks = []

        for block in blocks:

            mat: str = block.attrib["material"]

            voxel_coord_s: List[int] = [
                int(val) for val in block.attrib["grid_coordinate_s"].split()
            ]

            voxel_coord_e: List[int] = [
                int(val) for val in block.attrib["grid_coordinate_e"].split()
            ]

            # Store voxel-blocks in case they are needed to build the diffraction map later on
            voxel_blocks.append(
                {
                    "material": mat,
                    "grid_coordinate_s": voxel_coord_s,
                    "grid_coordinate_e": voxel_coord_e,
                    "diffraction_enabled": not transparency_map[mat],
                }
            )

        return voxel_blocks

    @staticmethod
    def __position_to_voxel_coord(subscene: Element, position: List[float]) -> List[int]:
        pos_s: List[float] = [float(val) for val in subscene.attrib["position_s"].split()]
        pos_e: List[float] = [float(val) for val in subscene.attrib["position_e"].split()]
        voxel_res: List[int] = [int(val) for val in subscene.attrib["vox_resolution"].split()]
        voxel_size: List[float] = [
            (pos_e[0] - pos_s[0]) / voxel_res[0],
            (pos_e[1] - pos_s[1]) / voxel_res[1],
            (pos_e[2] - pos_s[2]) / voxel_res[2],
        ]

        vox_position: List[float] = [
            position[2],
            position[0],
            position[1],
        ]

        voxel_coord: List[int] = []
        if (
            vox_position[0] >= pos_s[0]
            and vox_position[1] >= pos_s[1]
            and vox_position[2] >= pos_s[2]
            and vox_position[0] <= pos_e[0]
            and vox_position[1] <= pos_e[1]
            and vox_position[2] <= pos_e[2]
        ):
            for index, xyz in enumerate(vox_position):
                voxel_coord.append(int((vox_position[index] - pos_s[index]) / voxel_size[index]))
        return voxel_coord

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return [
            "scene",
            "raster_map_order",
        ]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "acoustic_materials",
            "bs_ids",
            "meshes",
            "object_sources",
            "updates",
            "voxel_config",
            "voxel_mode",
        ]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return ["rastermap"]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "rastermap": self.rasterMapList,
        }

    @staticmethod
    def getVersion() -> Any:
        return 1

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        rmParser = argParser.add_argument_group("RasterMap")
        rmParser.add_argument(
            "--raster_map_order",
            default=2,
            type=int,
            metavar="ORDER",
            help="number of supperted voxel diffraction paths (default=2)",
        )
