# This software module was originally developed by
#
# Nokia
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Nokia retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import argparse
import logging
from typing import Any, Dict, List
from dataclasses import dataclass

from mpegi.framework import Component
from mpegi.metadata.reverb import Reverberator, MultiroomCalculator
from mpegi.metadata.eif.acoustTuningParameters import DefaultAE, GlobalAEControl


@dataclass
class ReverbFrequencyGrid:
    acousticEnvironmentId: str
    rt60: List[float]
    dsr: List[float]
    gridIdx: int
    fgdNrBands: int
    frequencies: List[float]
    preDelay: float
    complexityLevel: int


class ReverbComponent(Component):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        super().__init__(args, artifacts)

        self.reverb_params = None
        if args["disableReverberator"]:
            logging.warning("REVERBERATOR IS DISABLED! Reverb will be omitted from bitstream!")
            return

        hasAcousticEnvironments = (
            artifacts["acoustic_environments"] is not None
            and len(artifacts["acoustic_environments"]) > 0
        )
        eifFilepath = args["scene"]
        sceneType = artifacts["scene_info"]["type"]

        if not hasAcousticEnvironments and args["disableDefaultAE"]:
            logging.info(
                f"Scene {eifFilepath} contains no AE's and default AE is disabled. Can't calculate reverb parameters"
            )
            return

        acousticEnvironments = artifacts["acoustic_environments"]
        configDisableDefaultAE = args["disableDefaultAE"]
        configDisablePortalLos = args["disablePortalLos"]
        configDisablePortal2PortalFactors = args["disablePortal2PortalFactors"]

        eifScene = artifacts["eif_scene"]

        if artifacts["acoustic_tuning_params"][0].defaultAE is None:
            defaultAE = DefaultAE()
        else:
            defaultAE = artifacts["acoustic_tuning_params"][0].defaultAE

        if artifacts["acoustic_tuning_params"][0].revRendererControl is None:
            revRendererControl = GlobalAEControl()
        else:
            revRendererControl = artifacts["acoustic_tuning_params"][0].revRendererControl

        if configDisableDefaultAE:
            defaultAE.enabled = False

        reverberator = Reverberator(defaultAE)

        explicitPortals = artifacts["eif_scene"].portals
        meshes = artifacts["meshes"]
        geometryById = eifScene.geometryByID
        isVoxelScene = len(artifacts["eif_scene"].voxelSceneIDs) > 0

        self.reverb_params = reverberator.createReverbParameters(
            acousticEnvironments,
            eifScene,
            revRendererControl,
            explicitPortals,
            meshes,
            geometryById,
            sceneType,
            configDisablePortalLos,
            isVoxelScene,
            eifFilepath,
        )
        if len(explicitPortals) != 0 and configDisablePortal2PortalFactors != True:
            mrc = MultiroomCalculator()
            self.portal2portalFactors = mrc.getPortal2PortalFactorMatrix(
                explicitPortals, geometryById
            )
        else:
            self.portal2portalFactors = None

        self.frequencyGrids = {}

        for ae in self.reverb_params.acousticEnvironments:
            freqs = []
            rt60 = []
            dsr = []

            for freq in ae.parameters[0].frequencies:
                freqs.append(freq.frequency)
                rt60.append(freq.rt60)
                dsr.append(freq.dsr)

            freqGrid = ReverbFrequencyGrid(
                acousticEnvironmentId=ae.id,
                rt60=rt60,
                dsr=dsr,
                frequencies=freqs,
                fgdNrBands=len(freqs),
                gridIdx=None,
                preDelay=ae.parameters[0].predelay,
                complexityLevel=ae.parameters[0].complexityLevel,
            )

            self.frequencyGrids[ae.id] = freqGrid

        if self.reverb_params.defaultAE.environment is not None:
            freqs = []
            rt60 = []
            dsr = []

            for freq in defaultAE.environment.parameters[0].frequencies:
                freqs.append(freq.frequency)
                rt60.append(freq.rt60)
                dsr.append(freq.dsr)

            freqGrid = ReverbFrequencyGrid(
                acousticEnvironmentId=defaultAE.environment.id,
                rt60=rt60,
                dsr=dsr,
                frequencies=freqs,
                fgdNrBands=len(freqs),
                gridIdx=None,
                preDelay=defaultAE.environment.parameters[0].predelay,
                complexityLevel=defaultAE.environment.parameters[0].complexityLevel,
            )

            self.frequencyGrids[defaultAE.environment.id] = freqGrid

        pass

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        # Reverberator Parameters
        reverberatorGroup = argParser.add_argument_group("Reverberator")
        reverberatorGroup.add_argument(
            "--disableReverberator",
            action="store_true",
            help="Disables the reverberator.",
        )

        reverberatorGroup.add_argument(
            "--disableDefaultAE",
            action="store_true",
            help="If enabled, provides a fallback reverb where none is authored in the scene",
        )

        reverberatorGroup.add_argument(
            "--disablePortalLos",
            action="store_true",
            help="Disables portal LOS",
        )

        reverberatorGroup.add_argument(
            "--disablePortal2PortalFactors",
            action="store_true",
            help="Disables calculation of Portal-to-Portal factors",
        )

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return [
            "scene",
            "disableReverberator",
            "disableDefaultAE",
            "disablePortalLos",
            "disablePortal2PortalFactors",
        ]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "acoustic_environments",
            "scene_info",
            "boxes",
            "eif_scene",
            "acoustic_tuning_params",
            "meshes",
            "boxes",
        ]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return [
            "reverb_params",
            "default_acoustic_environment",
            "reverb_frequency_grids",
            "portal_to_portal_factors",
        ]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "reverb_params": self.reverb_params,
            "default_acoustic_environment": (
                self.reverb_params.defaultAE if self.reverb_params else None
            ),
            "reverb_frequency_grids": self.frequencyGrids,
            "portal_to_portal_factors": self.portal2portalFactors,
        }

    @staticmethod
    def getVersion() -> Any:
        return 9
