# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import Any, Dict, List

import copy

from mpegi.framework import Component
from mpegi.metadata.eifplus import Scene, Mesh
from mpegi.encoder.components.preprocess import PreprocessorComponent


class ScenePlusComponent(Component):
    def __init__(self, _: Dict[str, Any], artifacts: Dict[str, Any]) -> None:
        self.scene = Scene.fromEif(artifacts["eif_scene"])
        isSmallScene = Scene.detectIsSmallScene(self.scene)
        self.sceneInfo = {
            "id": self.scene.id,
            "duration": self.scene.duration,
            "type": self.scene.type,
            "isSmallScene": isSmallScene,
            "enableLocallyCapturedAudio": self.scene.enableLocallyCapturedAudio,
            "overrideListenerVoiceDirectivity": self.scene.overrideListenerVoiceDirectivity,
            "listenerVoiceDirectivityId": self.scene.listenerVoiceDirectivityId,
        }
        self._generateCleanMeshes()

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return []

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return ["eif_scene"]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return [
            "scene_info",
            "acoustic_environments",
            "acoustic_materials",
            "anchors",
            "audio_streams",
            "boxes",
            "channel_sources",
            "cylinders",
            "directivities",
            "informed_sources",
            "hoa_groups",
            "hoa_sources",
            "raw_meshes",
            "meshes",
            "object_sources",
            "spheres",
            "transforms",
            "updates",
            "user_regions",
            "airflow_sources",
            "airflow_generators",
            "granular_sound",
        ]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "scene_info": self.sceneInfo,
            "acoustic_environments": self.scene.getEnvironments(),
            "acoustic_materials": self.scene.getMaterials(),
            "anchors": self.scene.getAnchors(),
            "audio_streams": self.scene.getAudioStreams(),
            "boxes": self.scene.getBoxes(),
            "channel_sources": self.scene.getChannelSources(),
            "cylinders": self.scene.getCylinders(),
            "directivities": self.scene.getDirectivities(),
            "informed_sources": self.scene.getInformedSources(),
            "hoa_groups": self.scene.getHoaGroups(),
            "hoa_sources": self.scene.getHoaSources(),
            "raw_meshes": self.scene.getMeshes(),
            "meshes": self.clean_meshes,
            "object_sources": self.scene.getObjectSources(),
            "portals": self.scene.getPortals(),
            "spheres": self.scene.getSpheres(),
            "transforms": self.scene.getTransforms(),
            "updates": self.scene.getUpdates(),
            "user_regions": self.scene.getUserRegions(),
            "acoustic_tuning_params": self.scene.getAcoustTuningParams(),
            "airflow_sources": self.scene.getAirflowSources(),
            "airflow_generators": self.scene.getAirflowGenerators(),
            "headphoneEQ": self.scene.getHeadphoneEQ(),
            "granular_sound": self.scene.getGranularSound(),
        }

    def _generateCleanMeshes(self) -> None:
        self.clean_meshes = [copy.deepcopy(mesh) for mesh in self.scene.getMeshes()]
        for mesh in self.clean_meshes:
            if not mesh.isTransparent:
                PreprocessorComponent.cleanupEifplusMesh(mesh)

    @staticmethod
    def getVersion() -> Any:
        return 7

    @staticmethod
    def registerArguments(_) -> None:
        pass
