# This software module was originally developed by
#
# Dolby International AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes and
# its performance may not have been optimized.  This software module is an
# implementation of one or more tools as specified by the ISO/IEC 23090-4
# standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
# copyright license to copy, distribute, and make derivative works of this
# software module or modifications thereof for use in implementations or
# products claiming conformance to ISO/IEC standards incorporating ISO/IEC
# 23090-4 reference software and which satisfy any specified conformance
# criteria of that standard.
#
# Those intending to use this software module in products are advised that its
# use may infringe existing patents and that the license in the previous
# paragraph grants no licenses under such patents.  ISO/IEC and the developers
# of this software module have no liability for use of this software module or
# modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Dolby International AB retains full right to
# modify and use the code for its own purpose, assign or donate the code to
# a third party without consent of and without accounting to third parties,
# and to inhibit third parties from using the code for products that do not
# conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative works.
#
# Copyright (c) ISO/IEC 2022.
import argparse
import json
import pathlib
from mpegi.framework import Component
from typing import List, Dict, Any
from jsonschema import validate

# ---------------------------------------------------------------------------------------------------
class VoxelSceneComponent(Component):
    switch_voxel_config = "voxconf"
    switch_voxel_mode = "voxmode"

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]) -> None:
        super().__init__(args, artifacts)

        # Voxel encoder configuration
        self.voxel_config = {}
        if self.switch_voxel_config in args and not args[self.switch_voxel_config] is None:
            with open(args[self.switch_voxel_config], "r", encoding="utf-8") as f_config:
                with open(
                    pathlib.Path(__file__).parent / "voxconf.schema.json", "r", encoding="utf-8"
                ) as f_schema:
                    self.voxel_config = json.load(f_config)
                    validate(self.voxel_config, schema=json.load(f_schema))

        self._mode = args[self.switch_voxel_mode]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return [
            VoxelSceneComponent.switch_voxel_config,
            VoxelSceneComponent.switch_voxel_mode,
        ]

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return ["acoustic_materials"]

    @staticmethod
    def getArtifactManifest() -> List[str]:
        return ["voxel_config", "voxel_mode"]

    def getArtifacts(self) -> Dict[str, Any]:
        return {
            "voxel_config": self.voxel_config,
            "voxel_mode": self._mode,
        }

    @staticmethod
    def getVersion() -> Any:
        return "0.0.1"

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        vox_parser = argParser.add_argument_group("Voxel-based Occlusion and Diffraction")

        vox_parser.add_argument(
            f"--{VoxelSceneComponent.switch_voxel_config}",
            metavar="config.json",
            type=pathlib.Path,
            default=None,
            help="Path to voxel encoder configuration (see voxconf.schema.json)",
        )

        help_text = (
            "\n"
            '"low_bitrate":       no payload data (default)\n'
            '"jps_full":          JPS data for the complete scene\n'
            '"jps_selective":     JPS data for selected scene area\n'
            '"multi_path":        RasterMap data for static scene\n'
            "                     or else no payload data\n"
            '"multi_path_or_jps": RasterMap data for static scene\n'
            '                     or else "jps_full" data'
        )
        vox_parser.add_argument(
            f"--{VoxelSceneComponent.switch_voxel_mode}",
            required=False,
            type=str,
            choices=["low_bitrate", "jps_selective", "jps_full", "multi_path", "multi_path_or_jps"],
            help=help_text,
        )
