# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

"""Use this module to serialise primitive data type into bool array.

Furthermore, floating point numbers can be quantized to integers suitable for
binary storage. Create a QuantizationRange data structure and then use that in
conjuction with the appropriate functions to encode/decode floating points.
"""

from __future__ import annotations

import enum
import struct
from dataclasses import dataclass
from typing import Tuple, Optional
import copy

import bitarray
import bitarray.util
import numpy as np

import hashlib


class Bitstream:
    """Create, modify, use a boolean array with this class.

    Use this to easily serialise data into a binary format, then write it to a
    file.

    .. warning:: The responsibility of type checking is left to the user. It
        must be ensured that in any function where a value is used as an
        argument that it must be the exact expected type. For example, addInt()
        only accepts an `int` and not `np.int64` or `bool`. It is critical that
        this is done otherwise the byte representation of the data maybe false -
        pythons dynamic typing cannot be relied upon.
    """

    # --= Construction =-- #

    def __init__(self) -> None:
        """Construct a Bitstream."""
        self.bitarray: bitarray.bitarray = bitarray.bitarray()

    @staticmethod
    def from01(binaryStr: str) -> Bitstream:
        """Construct a Bitstream from a binary string"""
        bitstream = Bitstream()
        bitstream.bitarray = bitarray.bitarray(binaryStr)
        return bitstream

    @staticmethod
    def fromBytes(bytesObj: bytes) -> Bitstream:
        """Construct a Bitstream from a bytes object"""
        bitstream = Bitstream()
        bitstream.bitarray.frombytes(bytesObj)
        return bitstream

    # --= Operations =-- #

    def addBool(self, value: bool) -> None:
        """Add a boolean value to the bitstream.

        :param value: The boolean value to add
        :param name: The name to be used for debugging, defaults to ""
        """
        self.addInt(int(value), 1)

    def addInt(self, value: int, width: int = 8) -> None:
        """Add an unsigned integer in the bitstream.

        :param value: The integer to add
        :param width: The width of the binary word in bits.
        """
        integer = bitarray.util.int2ba(value, width)
        self.bitarray.extend(integer)

    def addSignedInt(self, value: int, width: int = 8) -> None:
        """Add a signed integer in the bitstream as a two's complement.

        :param value: The integer to add
        :param length: The length of the binary word in bits, defaults to 8
        """
        integer = bitarray.util.int2ba(value, width, signed=True)
        self.bitarray.extend(integer)

    def addEscapedInt(self, value: int, widths: Tuple[int, int, int]) -> None:
        """Add an escaped integer in the bitstream.

        This allows for variable size binary words depending on how large the
        value is.

        :param value: The integer to add
        :param lengths: Three possible lengths given as an integer list
        :raises OverflowError: Value is too big to be stored with the given lengths
        """
        if value > self.getLargestValuePossibleForEscapedLengths(widths):
            raise OverflowError(
                f"Value '{value}' cannot be stored in lengths {widths[0]},"
                f"{widths[1]},{widths[2]}. The maximum possible value that can "
                f"be stored in the given lengths is {self.getLargestValuePossibleForEscapedLengths(widths)}"
            )

        bits_written = 0
        all_ones_bit = bitarray.bitarray([True for _ in range(widths[0])])
        all_ones_int = bitarray.util.ba2int(all_ones_bit)

        if value < all_ones_int:
            bits_written += widths[0]
            value_as_bits = bitarray.util.int2ba(value, widths[0])
            self.bitarray.extend(value_as_bits)
        else:
            bits_written += widths[0]
            self.bitarray.extend(all_ones_bit)
            value -= all_ones_int

            all_ones_bit = bitarray.bitarray([True for _ in range(widths[1])])
            all_ones_int = bitarray.util.ba2int(all_ones_bit)

            if value < all_ones_int:
                bits_written += widths[1]
                value_as_bits = bitarray.util.int2ba(value, widths[1])
                self.bitarray.extend(value_as_bits)
            else:
                bits_written += widths[1]
                self.bitarray.extend(all_ones_bit)
                value -= all_ones_int

                bits_written += widths[2]
                value_as_bits = bitarray.util.int2ba(value, widths[2])
                self.bitarray.extend(value_as_bits)

    def addFloat32(self, value: float) -> None:
        """Add a 32 bit floating point value (IEEE754 format) to the bitstream.

        :param value: The floating point value to add to the bitstream
        """
        value_as_long_int = struct.unpack(">L", struct.pack(">f", value))[0]
        self.addInt(value_as_long_int, 32)

    def addString(self, value: str) -> None:
        """Add a C-Style string into the bitstream.

        Each character is represented as 8bit ASCII char which is null
        terminated.

        :param value: The string to add to the bitstream
        """
        for letter in value:
            char = ord(letter)
            self.addInt(char, 8)
        self.addInt(0, 8)

    def appendBitstream(self, bitstream: Bitstream) -> None:
        """Appends another bitstream onto this bitstream.

        :param bitstream: The bitstream to append
        """
        self.bitarray.extend(bitstream.bitarray)

    def prependBitstream(self, bitstream: Bitstream) -> None:
        """Prepends another bitstream onto this bitstream.

        :param bitstream: The bitstream to prepend
        """
        bitstream.bitarray.extend(self.bitarray)
        self.bitarray = bitstream.bitarray

    def pad(self) -> None:
        """Byte-align the bitstream.

        The method adds zeros at the end of the bitstream so that the length of
        the bitstream becomes an integer multiple of eight.
        """
        self.bitarray.fill()

    def writeToFile(self, path: str) -> None:
        """Writes the bitstream to file to the specified path.

        :param path: The absolute or relative path for the file to be written
            to.
        """
        with open(path, "w+b") as file:
            self.bitarray.tofile(file)

    def pop(self, start: int = 0, numberOfBits: int = 1) -> Bitstream:
        """Remove a range of bits and return it as a new bitstream.

        .. note::
            This pop function by default pops the front of the bitstream and not the back like one
            would expected

        :param start: The index of the pop, defaults to 0
        :param numberOfBits: Give a range of bits to pop, defaults to 1
        :return: The popped bits
        """
        newBitarray = self.bitarray[start:numberOfBits]
        self.bitarray = self.bitarray[start + numberOfBits :]

        newBitstream = Bitstream()
        newBitstream.bitarray = newBitarray
        return newBitstream

    def popBool(self) -> bool:
        """Pop a boolean value return it.

        :return: The popped boolean value.
        """
        return bool(self.popInt(1))

    def popInt(self, width: int = 8) -> int:
        """Pop an unsigned integer and return it.

        :param width: The width of the binary word in bits.
        :return: The popped int read integer
        """
        temp = bitarray.bitarray()
        for _ in range(width):
            temp.append(self.bitarray.pop(0))
        value = bitarray.util.ba2int(temp, signed=False)
        return value

    def popSignedInt(self, length: int = 8) -> int:
        """Pop a signed integer and return it.

        :param length: The length of the binary word in bits, defaults to 8
        :return: The popped signed integer
        """
        temp = bitarray.bitarray()
        for _ in range(length):
            temp.append(self.bitarray.pop(0))
        value = bitarray.util.ba2int(temp, signed=True)
        return value

    def copy(self, start: int = 0, length: Optional[int] = None) -> Bitstream:
        """Deepcopy the bitstream into a new object.

        Optioanally provide start and length to "extract" chunks out of
        the bitstream.

        :param start: At which index to begin the copy, defaults to 0
        :param length: Length in bits of the desired copy, is None is
            given, then the original length will be used, defaults to
            None.
        :return: The deepcopied bitstream
        """
        new_bitstream = Bitstream()
        new_bitstream.bitarray = copy.deepcopy(self.bitarray[start:])

        if length is not None:
            new_bitstream.bitarray = new_bitstream.bitarray[:length]

        return new_bitstream

    # --= Information Getters =-- #

    def getLength(self) -> int:
        """Returns the length of bitstream in bits.

        .. note::
            NOT BYTES

        :return: Length of Bitstream
        """
        return len(self.bitarray)

    # --= Utility =-- #

    @staticmethod
    def getLargestValuePossibleForEscapedLengths(lengths: Tuple[int, int, int]) -> int:
        """Returns the largest possible value that can be stored in an escaped
        integer via the three given lengths

        :param lengths: The three binary word lengths as a list in ascending order
        :return: The maximum possible value that can be stored with the given lengths
        """
        return (2 ** lengths[0] - 1) + (2 ** lengths[1] - 1) + (2 ** lengths[2] - 1)

    # --= Representations =-- #

    def to01(self) -> str:
        """Returns a string of '0's and '1's that represent this bitstream.

        :return: Bitstream as a string
        """
        return self.bitarray.to01()

    def toBytes(self) -> bytes:
        """Returns a bytearray that represents this bitstream.

        .. note:: This function will pad the bitstream with 0s to the nearest
            byte.

        :return: The bitstream as a bytearray
        """
        self.pad()
        return self.bitarray.tobytes()

    def toBool(self) -> bool:
        """Return bitstream as a bool.

        Length must be exactly one bit long.

        :raises ValueError: Bitstream must be exactly one bit long
        :return: The bitstream as a boolean value
        """
        if self.getLength() != 1:
            raise ValueError("Must have length of 1 bit to cast to bool")
        return bool(self.bitarray[0])

    def toInt(self) -> int:
        """Return the entire bitstream represented as an unsigned integer."""
        return bitarray.util.ba2int(self.bitarray)

    def toSignedInt(self) -> int:
        """Return the entire bitstream represented as a signed integer."""
        return bitarray.util.ba2int(self.bitarray, True)

    def toFloat(self) -> float:
        """Return the entire bitstream represented as a IEEE 745 float."""
        assert self.getLength() == 32, "bitstream must be 32 bits long"
        value = self.toInt()
        return struct.unpack(">f", struct.pack(">L", value))[0]

    def toAscii(self) -> str:
        """Return the entire bitstream as an ascii string."""
        return self.toBytes().decode("ascii")

    def to_md5(self) -> str:
        """Return hex digest of a md5 hash for this bitstream."""
        return hashlib.md5(self.toBytes()).hexdigest()

    # --= "Private" =-- #

    def __len__(self) -> int:
        """Returns the length of bitstream in bits.

        .. note::
            NOT BYTES

        :return: Length of Bitstream
        """
        return self.getLength()

    def __str__(self) -> str:
        """Returns a string of '0's and '1's that represent this bitstream."""
        return self.to01()

    def __bool__(self) -> bool:
        """Return bitstream as a bool."""
        return self.toBool()

    def __int__(self) -> int:
        """Return bitstream as an integer."""
        return self.toInt()


class QuantizationScaling(enum.IntEnum):
    """Different types of QuanitizationScaling."""

    LINEAR = 0


@dataclass
class QuantizationRange:
    """Data structure to hold all required info for quantization for a value."""

    min: float
    """Minimum value for the range."""
    max: float
    """Maximum value for the range."""
    num_bits: int
    """The number of bits to quantise a value to."""
    scaling_type: QuantizationScaling = QuantizationScaling.LINEAR
    """Quantization Scaling type - see QuantizationScaling for more info."""

    @staticmethod
    def from_tolerance(min_val: float, max_val: float, tolerance: float) -> QuantizationRange:
        """Construct a range with a given tolerance.

        The range that's constructed will have a quantisation error less than
        the tolerance. This is useful for determining ranges where you know the
        desired step size/quantization, but don't know the number of bits
        required for such a range.

        :param min_val: The minimum value for the range.
        :param max_val: The maximum value for the range.
        :param tolerance: The range returned will have a quantization error less than this
        :return: The designed QuantizationRange with the given tolerance
        """
        num_bits = get_num_bits_required_for_quantization(min_val, max_val, tolerance)
        return QuantizationRange(min=min_val, max=max_val, num_bits=num_bits)

    def is_within_range(self, value: float) -> bool:
        """Return true if the value is in between the range."""
        return np.greater_equal(value, self.min) and np.less_equal(value, self.max)


def encode_float(value: float, quantization_range: QuantizationRange) -> int:
    """Encode a floating point value to a given QuantizationRange.

    :param value: The value to encode
    :param quantisation_range: The data structure which holds data necessary for encoding
    :raises ValueError: Cannot encode a value that is not within range
    :return: The encoded float as an int
    """
    min_val = quantization_range.min
    max_val = quantization_range.max
    bits = quantization_range.num_bits

    if not quantization_range.is_within_range(value):
        raise ValueError(f"Value '{value}' must be between '{min_val}' and '{max_val}'")

    return int(np.round((value - min_val) / (max_val - min_val) * (2 ** bits - 1)))


def decode_float(value: int, quantization_range: QuantizationRange) -> float:
    """Decode an integer value from a given QuantizationRange.

    :param value: The value to decode
    :param quantisation_range: The data structure which holds data necessary for decoding
    :raises ValueError: Cannot decode a value that is not within range
    :return: The decoded value as a float
    """
    min_val = quantization_range.min
    max_val = quantization_range.max
    bits = quantization_range.num_bits

    result = (max_val - min_val) * value / float(2 ** bits - 1) + min_val

    if not quantization_range.is_within_range(result):
        raise ValueError(f"Value '{value}' must be between '{min_val}' and '{max_val}'")

    return result


def round_value_to_range(value: float, quantization_range: QuantizationRange) -> float:
    """Quantizes and the unquantizes a value.

    This is useful so that the float will always fit the quantization range with
    no quantization error.

    :param value: The value to found
    :param quantization_range: The range in which to round to
    :return: The rounded value
    """
    value_as_int = encode_float(value, quantization_range)
    return decode_float(value_as_int, quantization_range)


def get_quantization_error(value: float, quanisation_range: QuantizationRange) -> float:
    """Return the quantization error for a value for a given quantization range."""
    temp = encode_float(value, quanisation_range)
    return value - decode_float(temp, quanisation_range)


def get_max_quantization_error(quantization_range: QuantizationRange) -> float:
    """Return the maximum possible quantization erroe for a value for a given range."""
    # we are using a mid-tread quantization here, so the error is maximal
    # when the value is exactly at the boundary of a step
    min_val = quantization_range.min
    max_val = quantization_range.max
    bits = quantization_range.num_bits

    scale_to_max = float(2 ** bits - 1)
    step_size = (max_val - min_val) / scale_to_max
    max_error = np.abs(get_quantization_error(min_val + (step_size / 2), quantization_range))
    return float(max_error)


def get_num_bits_required_for_quantization(min_val: float, max_val: float, tolerance: float) -> int:
    """Return the number of bits required to store a value with a given tolerance.

    The quantization error will always be smaller than the tolerance.
    """
    # if the error should always be smaller than tolerance
    # the step size needs to be smaller than 2*tolerance
    return int(np.ceil(np.log2((max_val - min_val) / (2 * tolerance))))
