# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""This module uses the :class:`~mpegi.framework.BitstreamComposer` together the
:mod:`mpegi.framework.tokens`. Please also read the documentation in the Tokens module. This module
provides a way to create a :class:`~mpegi.framework.Bitstream` using semantic elements. Read
:class:`~mpegi.framework.BitstreamComposer` for further details
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from . import tokens
from .bitstream import Bitstream


class BitstreamComposer:
    """The BitstreamComposer is used to create a :class:`~mpegi.framework.Bitstream` through the use
    of semantic units. These units are called :class:`Tokens <mpegi.framework.tokens>`. Upon
    construction, the Composer is given a list of Tokens. Then values are then pushed to the
    composer. Using the information supplied by the Tokens, a Bitstream is created. This way, the
    creation of an abstract syntax vs a concerete bitstream is seperated. Note that if too much or
    not enough data is pushed to Composer the creation of the Bitstream will fail. This is to ensure
    that a bitstream has been created correctly according the grammar of the syntax.

    When creating a Syntax, take care to give all of the Tokens unique and descriptive labels."""

    def __init__(self, syntax: List[tokens.Token]):
        """A syntax is represented as a list of tokens
        :param syntax: A list of Tokens
        """
        self.scope: Dict[str, Any] = {}
        self.token_generator = tokens.TokenGenerator(syntax, self.scope)
        self.bitstream = Bitstream()

    def push(self, value: Any, append: bool = False) -> None:
        """Push a value to the Composer.

        .. warning:: Make sure that the order of what data is pushed to the Composer directly
            coressponds to the data that the Token expected. For example, if you push a float value
            when the syntax expected an integer, then this will break.

        :param value: The value to push
        :raises OverflowError: Too much data was supplied
        """
        token = self.token_generator.get_next_token()
        if token is None:
            raise OverflowError("Too much data supplied")
        # if append == True:
        if hasattr(token, "storeValue"):
            if token.storeValue == True:
                if token.label in self.scope and isinstance(self.scope[token.label], list):
                    self.scope[token.label].append(value)
                else:
                    self.scope[token.label] = [value]
            else:
                self.scope[token.label] = value
        else:
            self.scope[token.label] = value
        bitstream = token.to_bitstream(value)
        self.bitstream.appendBitstream(bitstream)

    def get_bitstream(self) -> Bitstream:
        """Once all of the data has been pushed to the Composer, call this function to retrieve a
        :class:`~mpegi.framework.Bitstream`.

        .. warning:: Make sure that _all_ tokens have had values pushed to them before calling this
            method.

        :raises NotEnoughDataPushedError: More data is required to satisfy the syntax.
        :return: The concrete Bitstream that was created from an abstract syntax
        """
        if self.token_generator.get_next_token() is not None:
            raise NotEnoughDataPushedError("More data is required to satisfy the syntax")

        return self.bitstream


class BitstreamDecomposer:
    """Decompose a Bitstream into a small usable chunks.

    By giving a bitstream and a syntax tree for said bitstream, the
    bitstream can be decomposed and broken into human readable parts.
    Certain parts of the bitstream can be queried by parsing the
    bitstream and interpretting the bits underneath.
    """

    def __init__(self, bitstream: Bitstream, syntax: List[tokens.Token]) -> None:
        """Constuct a BitstreamDecomposer.

        :param bitstream: The bitstream to decompose
        :param syntax: The abstract syntax tree to interpret the
            bitstream with
        """
        self.scope: Dict[str, Any] = {}
        self.token_generator = tokens.TokenGenerator(syntax, self.scope)
        self.token_generator.ignore_markers = False
        self.bitstream = bitstream
        self.start: int = 0

    def pop(self) -> Bitstream:
        next_token = self.token_generator.get_next_token()
        length: int = next_token.width
        new_bitstream = self.bitstream.copy(self.start, length=length)
        self.start += length
        return new_bitstream

    def extract_between_markers(self) -> Bitstream:
        """Extract a bitstream indicated by start and stop markers.

        :return: The extracted bitstream
        """
        length: Optional[int] = None
        token_start = 0

        while True:
            token = self.token_generator.get_next_token()
            if isinstance(token, tokens.StartMarker):
                length = 0
                continue

            if isinstance(token, tokens.EndMarker):
                break

            assert token is not None
            bits, value = token.from_bitstream(self.bitstream, token_start)
            if length is None:
                self.start += bits
                token_start = self.start
            else:
                length += bits
                token_start += bits
            if hasattr(token, "storeValue"):
                if token.storeValue == True:
                    if token.label in self.scope:
                        if type(self.scope[token.label]) != list:
                            self.scope[token.label] = [self.scope[token.label]]
                        self.scope[token.label].append(value)
                    else:
                        self.scope[token.label] = value
                else:
                    self.scope[token.label] = value

            else:
                self.scope[token.label] = value

        return self.bitstream.copy(self.start, length)


class NotEnoughDataPushedError(Exception):
    """This exception is raised when the :class:`~mpegi.framework.BitstreamComposer` didn't recieve
    enough data to fully satisfy all of the requirements as dictated by the Syntax. For example,
    when a syntax declares three flags, and only one bool was pushed to the Composer, this exception
    is raised"""

    pass
