# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, List

import argparse


class Component:
    """A component is a small compartmental unit of the encoder. It is responsible
    for creating artifacts (typically) from preexisting artifacts created from
    other components.
    """

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """All subscribed arguments and artifacts are given to a component on
        construction.

        :param args: The subscribed arguments in a dictionary
        :type args: Dict[str, Any]
        :param artifacts: The subscribed artifacts in a dictionary
        :type artifacts: Dict[str, Any]
        """
        pass

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        """A component can subscribe to arguments given on the command line.
        Return an empty list if no arguments are required.

        :raises NotImplementedError: Child class must implement this function
        :return: A list of a strings previously registered with Config
        :rtype: List[str]
        """
        raise NotImplementedError("Component must implement getArgumentSubscription()")

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        """A component can subscribe to artifacts that were previously made
        from other components within the encoder.

        :raises NotImplementedError: Child class of component must implement
            this function
        :return: A list of strings previously registered via component manifests
        :rtype: List[str]
        """
        raise NotImplementedError("Component must implement getArtifactSubscription()")

    @staticmethod
    def getArtifactManifest() -> List[str]:
        """A component must declare which artifacts it produces so that other
        components can subscribe to the them

        :raises NotImplementedError: Child class of component must implement
            this function
        :return: A list of strings that other components can subscribe to
        :rtype: List[str]
        """
        raise NotImplementedError("Component must implement getArtifactManifest()")

    def getArtifacts(self) -> Dict[str, Any]:
        """Returns all of the artifacts in a dictionary where the key is the string
        declared in the component manifest and the value is the corresponding
        artifact object.

        Please ensure that all artifacts can be pickled. See here for more details:
        https://docs.python.org/3/library/pickle.html#pickle-picklable

        :raises NotImplementedError: Child class of component must implement
            this function
        :return: Artifacts in a dictionary format
        :rtype: Dict[str, Any]
        """
        raise NotImplementedError("Component must implement getArtifacts()")

    @staticmethod
    def getVersion() -> Any:
        """Returns a version number for the component. To invalidate the cache
        for a component, change the value that this function returns. The cache
        must be invalidated whenever the component semantics change. Component
        semantics means how artifacts are created and how they behave.

        :raises NotImplementedError: Child class of component must implement
        :return: This can be any value, so long as it is different from all of
            values that have come before it, i.e. not repeat values.
        :rtype: Any
        """
        raise NotImplementedError("Component must implement getVersion()")

    @staticmethod
    def registerArguments(argParser: argparse.ArgumentParser) -> None:
        """An argument parser is passed as an argument and through it arguments
        can be registered. Look at argparse documentation for more details.
        https://docs.python.org/3/library/argparse.html

        :param argParser: The argument parser instance
        :type argParser: argparse.ArgumentParser
        :raises NotImplementedError: All Components must implement this function
        """
        raise NotImplementedError("Component must register to arguments")
