# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Type, Optional

from .component import Component
from .dependencygraph import DependencyGraph
from .inventory import Inventory


class ComponentBuilder:
    """The component builder is responsible for constructing each component.
    Construction can be skipped if it is available via the inventory. Alternatively
    construction can be optionally forced every time.
    """

    def __init__(self, componentList: List[Type[Component]]):
        self.dependencyGraph = DependencyGraph.fromComponents(componentList)

    def process(self, inventory: Inventory, forceBuild: Optional[List[str]] = None) -> None:
        """Begins processing of all components, deciding whether or not to
        construct them based on cache availablility or user command

        :param inventory: The artifacts produced from the components will be
            stored in the inventory.
        :param forceBuild: Forcefully builds all components regardless of diskcache
            availability. If an empty list is given, all components are forcefully
            built, otherwise the strings in said list must coresspond to the
            class names of the components being constructed, defaults to None
        :type forceBuild: Optional[List[str]], optional
        """
        if forceBuild != None:
            allComponentNames = [c.__name__ for c in self.dependencyGraph.getAllNodes()]
            for component in forceBuild:
                if component not in allComponentNames:
                    raise KeyError(
                        f"Cannot force build Component '{component}' " "because it does not exist"
                    )

        for componentLayer in self.dependencyGraph:
            for ComponentType in componentLayer:
                force = self._shouldForceBuildComponent(ComponentType, forceBuild)
                if force == False:
                    try:
                        inventory.load(ComponentType)
                    except FileNotFoundError:
                        self._createAndStoreArtifacts(ComponentType, inventory, force)
                else:
                    self._createAndStoreArtifacts(ComponentType, inventory, force)

    @staticmethod
    def _createAndStoreArtifacts(
        ComponentType: Type[Component], inventory: Inventory, force: bool = False
    ) -> None:
        """A helper function to retrieve arguments and artifacts from the inventory,
        use them to create a component, then finally store the resulting artifacts
        from the component back into the inventory
        """
        try:
            args, artifacts = inventory.getArgsAndArtifacts(ComponentType)
        except KeyError as ex:
            errMsg = f"Invalid arg/artifact request from {ComponentType.__name__}: {ex}"
            raise KeyError(errMsg)

        concreteComponent = ComponentType(args, artifacts)
        inventory.store(ComponentType, concreteComponent.getArtifacts(), force)

    def _shouldForceBuildComponent(
        self, ComponentType: Type[Component], forceBuild: Optional[List[str]]
    ) -> bool:
        """Determines whether or not a component should be built forcefully or
        not. To forcefully build a component is to construct it regardless
        if its cache can be retreived or not
        """
        if self.dependencyGraph.isRootNode(ComponentType):
            return True

        if forceBuild == None:
            return False

        if ComponentType.__name__ in forceBuild or len(forceBuild) == 0:
            return True

        return False
