# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, List, Tuple, Type

from .component import Component


class AdjacencyList:
    """Detail/Helper class for dependency graph. Analyses adjacent nodes in a graph"""

    def __init__(self) -> None:
        self.elements: Dict[Any, Dict[str, List[Any]]] = {}

    def addElement(self, element: Any, input: List[Any], output: List[Any]) -> None:
        assert element not in self.elements.keys(), "Duplicate Element"
        for o in output:
            for e in self.elements.keys():
                assert o not in self.elements[e]["output"], f"Duplicate Output Artifact: {o}"

        self.elements[element] = {"input": input, "output": output}

    def getList(self) -> List[Tuple[Any, List[int]]]:
        allElements = list(self.elements.keys())
        allElements = sorted(allElements, key=lambda x: len(self.elements[x]["input"]))

        adjList: List[Tuple[Any, List[Any]]] = []
        for element in allElements:
            elementDependants = self._getElementDependants(element)
            elementDependantIndicies = [allElements.index(e) for e in elementDependants]
            adjList.append((element, elementDependantIndicies))
        return adjList

    def _getElementDependants(self, element) -> List[Any]:
        elementDependendants = []
        for output in self.elements[element]["output"]:
            elementDependendants[-1:-1] = self._getAllElementsWithInput(output)
        return list(set(elementDependendants))

    def _getAllElementsWithInput(self, input: Any) -> List[Any]:
        elementsWithInput = []
        for key, value in self.elements.items():
            if input in value["input"]:
                elementsWithInput.append(key)
        return elementsWithInput


class DependencyGraph:
    """A DepedencyGraph is a directed acylic graph between all of the components
    within the encoder. It identifies all dependencies between all components
    via a toplogical sort. By iterating over the graph, all components which can
    be constructed in parallel will be returned.
    """

    def __init__(self, adjacencyList: AdjacencyList):
        self.graph = self._topologicalSort(adjacencyList)

    def getAllNodes(self) -> List[Any]:
        """Get all nodes as a 1D list"""
        return sum(self.graph, [])

    def isRootNode(self, node: Any) -> bool:
        """Returns true if the node is in the first layer, i.e. at the top"""
        return node in self.graph[0]

    @staticmethod
    def fromComponents(components: List[Type[Component]]):
        """Construct DepencyGraph from Components"""
        graph = AdjacencyList()
        for component in components:
            componentInput = component.getArtifactSubscription()
            componentOutput = component.getArtifactManifest()
            graph.addElement(component, componentInput, componentOutput)
        return DependencyGraph(graph)

    def __iter__(self):
        yield from self.graph

    @staticmethod
    def _topologicalSort(adjacencyList: AdjacencyList) -> List[List[Any]]:
        """Slight modification on Kahn's topological sort, returns a list of lists.
        Each sub list gives all nodes that can be performed at once, i.e all
        dependencies have been excuted before hand. This means that each element
        within the sub list can be performed in parallel.
        """
        adjList = adjacencyList.getList()
        components = [i[0] for i in adjList]
        adjList = [i[1] for i in adjList]

        numberOfNodes = len(adjList)
        indegree = [0 for _ in range(numberOfNodes)]
        for i in range(numberOfNodes):
            for to in adjList[i]:
                indegree[to] += 1

        queue = []
        for i in range(numberOfNodes):
            if indegree[i] == 0:
                queue.append(i)

        index = 0
        order = []
        while len(queue) > 0:
            order.append([])

            while len(queue) > 0:
                at = queue.pop()
                order[-1].append(at)
                index += 1

            for at in order[-1]:
                for to in adjList[at]:
                    indegree[to] -= 1
                    if indegree[to] == 0:
                        queue.append(to)

        if index != numberOfNodes:
            raise Exception("Cyclic Dependency Between Components Found")

        graph = []
        for o in order:
            graph.append([])
            for i in o:
                graph[-1].append(components[i])
        return graph
