# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict
import hashlib
import pathlib
import pickle


class DiskCache:
    """Provides a storage solution for artifacts of components between encodings.
    Artifacts can be stored and retrieved from disk.
    """

    def __init__(self, directory: str):
        """Instanciate the diskcache with the given directory

        :param directory: The directory in which the cache should be using
        :type directory: str
        """
        self.cacheDir = pathlib.Path(directory)
        pathlib.Path.mkdir(self.cacheDir, parents=True, exist_ok=True)

    def retrieve(
        self,
        componentName: str,
        componentVersion: Any,
        subscribedObjects: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Returns all artifacts for a given component. The cache file is named
        after the hash of the subscribed arguments and artifacts and the component
        name itself.

        :param componentName: The class name of the component
        :type componentName: str
        :param componentVersion: The component version
        :type componentVersion: Any
        :param subscribedObjects: The instanciate objects of arguments and artifacts
            for the component
        :type subscribedObjects: Dict[str, Any]
        :raises FileNotFoundError: The cache file could not be found
        :return: The output artifacts coressponding to the incoming subscribed objects
        :rtype: Dict[str, Any]
        """
        # Check if folder with same hash exists
        inputHash = self._getHash(subscribedObjects, componentVersion)
        cacheDir = self.cacheDir.joinpath(f"{componentName}-{inputHash}")
        if not cacheDir.is_dir():
            raise FileNotFoundError(f"Cache not found")

        # Return artefacts from cache
        artifacts = {}
        cachesInDir = list(cacheDir.glob("*.art"))
        for cache in cachesInDir:
            with open(cache, "r+b") as cacheFile:
                artifact: Dict[str, Any] = pickle.load(cacheFile)
                artifacts.update(artifact)
        return artifacts

    def save(
        self,
        componentName: str,
        componentVersion: Any,
        subscribedObjects: Dict[str, Any],
        artifacts: Dict[str, Any],
        force: bool = False,
    ) -> None:
        """Save artifacts to disk. The file name is the combination of the hash
        of the subscribed objects and the component name and the file contents
        is a binary representation of the artifacts themselves.

        :param componentName: The name of the component class
        :type componentName: str
        :param componentVersion: The version of the component class
        :type componentVersion: str
        :param subscribedObjects: The instanciated objects of the subscribed
            arguments and artifacts for the component.
        :type subscribedObjects: Dict[str, Any]
        :param artifacts: The artifacts to save to disk
        :type artifacts: Dict[str, Any]
        :param force: If true, save to disk even if the cache already exists,
            defaults to False
        :type force: bool, optional
        """
        # Create a folder with the name of the input data hash
        inputHash = self._getHash(subscribedObjects, componentVersion)
        cacheDirName = f"{componentName}-{inputHash}"
        newCacheDir = self.cacheDir.joinpath(cacheDirName)
        if not newCacheDir.is_dir():
            pathlib.Path.mkdir(newCacheDir)

        for key, value in artifacts.items():
            artifactToCache = {key: value}
            artifactHash = DiskCache._getHash(artifactToCache, componentVersion)
            cacheName = f"{key.upper()}-{artifactHash}.art"
            cache = newCacheDir.joinpath(cacheName)
            if not cache.is_file() or force:
                with open(cache, "wb") as cacheFile:
                    artifactAsBytes = pickle.dumps(artifactToCache)
                    cacheFile.write(artifactAsBytes)

    @staticmethod
    def _getHash(artifacts: Dict[str, Any], salt: Any) -> str:
        """Helper function to get the md5 hash of a dictionary of artifacts

        :param artifacts: The artifacts to get the hash of
        :type artifacts: Dict[str, Any]
        :param salt: A salt that is applied before hashing
        :type salt: Any
        :return: The hash in a string
        :rtype: str
        """
        dictToHash = {"SALT_FOR_HASHING": salt}
        for key, value in artifacts.items():
            dictToHash[key] = DiskCache._getHashFromArtifact(value)
        dictToHashAsBytes = pickle.dumps(dictToHash)
        return hashlib.md5(dictToHashAsBytes).hexdigest()

    @staticmethod
    def _getHashFromArtifact(artifact: Any) -> str:
        """Helper function to handle hashing of different types"""
        if isinstance(artifact, pathlib.Path):
            hash = DiskCache._getHashFromFile(artifact)
        elif getattr(artifact, "md5", None) is not None:
            hash = artifact.md5()
        else:
            artifactAsBytes = pickle.dumps(artifact)
            hash = hashlib.md5(artifactAsBytes).hexdigest()
        return hash

    @staticmethod
    def _getHashFromFile(filePath: pathlib.Path) -> str:
        with open(filePath, "rb") as file:
            fileContents = file.read()

        carriageReturn = bytes.fromhex("0d")  # "CR" or "\r"
        fileContents = fileContents.replace(carriageReturn, b"")  # strip \r
        md5 = hashlib.md5(fileContents)
        return md5.hexdigest()
