# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Create Codebooks, Encode Symbols, Select Efficient Codebook Encoding.

This module provides a generic codebook tool for encoding data series.
A SymbolHistogram is first created from either a list of plain symbols
or from a given histogram of symbols. The SymbolHistogram is then used
to generate an EncoderConfig which results in the smallest possible
encoding of a plain text that has as many symbols specified in the given
histogram. This module also includes functions for serializing the code
book and for encoding lists of symbols.

This module also provides a class to decode the codebooks as well.
"""

from __future__ import annotations

from collections import Counter, OrderedDict
from dataclasses import dataclass
from typing import Dict, List

from bitarray import bitarray
from bitarray.util import huffman_code, int2ba

from mpegi.framework import Bitstream


@dataclass(frozen=True)
class SymbolHistogram:
    """Data structure for symbol histograms."""

    freq_map: Counter[int]

    @staticmethod
    def from_histogram(freq_map: Counter[int]) -> SymbolHistogram:
        """Construct a SymbolHistogram from a Counter."""

        return SymbolHistogram(freq_map)

    @staticmethod
    def from_plain(val_list: List[int]) -> SymbolHistogram:
        """Construct a SymbolHistogram from a list of plain symbols.

        The list of symbols should include repeated occurrences of
        symbols so that a probability distribution can be determined.
        """
        val_list_int = [int(val) for val in val_list]

        # work-around for empty sets
        if len(val_list_int) == 0:
            val_list_int = [int(0)]

        return SymbolHistogram.from_histogram(Counter(val_list_int))

    def min(self) -> int:
        """Return the lowest symbol in the codebook."""
        return int(min(self.freq_map.keys()))

    def max(self) -> int:
        """Return the highest symbol in the codebook."""
        return int(max(self.freq_map.keys()))


@dataclass(frozen=True)
class EncoderConfig:
    """Data structure for encoder configurations."""

    # flag indicating the usage of fixed or variable length encoding (0: Huffman, 1: fixed length)
    fixed_length: bool

    # optional offset value
    offset: int

    # codebook: map symbol list -> code word list
    codebook: OrderedDict[int, bitarray]

    def get_codebook_size(self) -> int:
        """Return the number of bits required to serialize the codebook"""
        bs = serialize_codebook(self)
        return bs.getLength()


def get_encoder_config(histogram: SymbolHistogram) -> EncoderConfig:
    """Select the most efficient encoder and return the resulting encoder configuration."""

    # workaround for empty histograms
    if len(histogram.freq_map) == 0:
        histogram = SymbolHistogram.from_plain([0])

    (config_fixed_length, num_bits_fixed_length) = get_fixed_length_config(histogram)
    (config_huffman, num_bits_huffman) = get_huffman_config(histogram)
    if num_bits_fixed_length <= num_bits_huffman:
        return config_fixed_length
    else:
        return config_huffman


def get_fixed_length_config(histogram: SymbolHistogram) -> tuple(EncoderConfig, int):
    """Generate a fixed length encoder configuration."""
    fixed_length = True
    symbol_list = [int(symbol) for symbol in histogram.freq_map.keys()]

    # config without offset
    offset = 0
    word_size = max(1, int(histogram.max() - offset).bit_length())
    code_list = [int2ba(symbol - offset, length=word_size) for symbol in symbol_list]
    codebook = OrderedDict(zip(symbol_list, code_list))
    config_without_offset = EncoderConfig(fixed_length, offset, codebook)

    # config with offset
    offset = histogram.min()
    word_size = max(1, int(histogram.max() - offset).bit_length())
    code_list = [int2ba(symbol - offset, length=word_size) for symbol in symbol_list]
    codebook = OrderedDict(zip(symbol_list, code_list))
    config_with_offset = EncoderConfig(fixed_length, offset, codebook)

    # select config
    num_bits_with_offset = estimate_bitstream_size(config_with_offset, histogram)
    num_bits_without_offset = estimate_bitstream_size(config_without_offset, histogram)
    if num_bits_with_offset < num_bits_without_offset:
        return config_with_offset, num_bits_with_offset
    else:
        return config_without_offset, num_bits_without_offset


def get_huffman_config(histogram: SymbolHistogram) -> tuple(EncoderConfig, int):
    """Generate a Huffman encoder configuration."""
    fixed_length = False

    # workaround for identical symbols
    freq_map = histogram.freq_map
    single_symbol = len(histogram.freq_map) == 1
    if single_symbol:
        symbols = list(freq_map.keys())
        freq_map[symbols[0] + 1] = 1
        histogram = SymbolHistogram.from_histogram(freq_map)

    # get symbol and code word list in tree traversal order
    codebook_unordered = huffman_code(histogram.freq_map)
    codebook_inverse = _get_codebook_inverse(codebook_unordered)
    bitstream = Bitstream()
    symbol_list = []
    code_list = []
    _serialize_tree(codebook_inverse, bitstream, symbol_list, code_list)
    codebook = OrderedDict(zip(symbol_list, code_list))

    # config without offset
    offset = 0
    config_without_offset = EncoderConfig(fixed_length, offset, codebook)

    # config without offset
    offset = histogram.min()
    config_with_offset = EncoderConfig(fixed_length, offset, codebook)

    # select config
    too_much = int(single_symbol)
    num_bits_with_offset = estimate_bitstream_size(config_with_offset, histogram) - too_much
    num_bits_without_offset = estimate_bitstream_size(config_without_offset, histogram) - too_much
    if num_bits_with_offset < num_bits_without_offset:
        return config_with_offset, num_bits_with_offset
    else:
        return config_without_offset, num_bits_without_offset


def serialize_codebook(config: EncoderConfig) -> Bitstream:
    """Serialize codebook to bitstream."""

    bitstream = Bitstream()

    # flagFixedLength
    flagFixedLength = config.fixed_length
    bitstream.addBool(flagFixedLength)

    # flagOffset
    flagOffset = config.offset > 0
    bitstream.addBool(flagOffset)

    # add "offset" value to codebook header if needed
    if flagOffset:
        # wordSizeOffset
        wordSizeOffset = max(1, config.offset.bit_length())
        bitstream.addInt(wordSizeOffset, HuffmanDecoder.bitsForWordSize)

        # offset
        offset = config.offset
        bitstream.addInt(offset, wordSizeOffset)

    # wordSize
    wordSize = max(
        1, max([int(symbol - config.offset).bit_length() for symbol in config.codebook.keys()])
    )
    bitstream.addInt(wordSize, HuffmanDecoder.bitsForWordSize)

    if not config.fixed_length:
        # serialize decoding tree
        codebook_inverse = _get_codebook_inverse(config.codebook)
        symbol_list = []
        code_list = []
        _serialize_tree(codebook_inverse, bitstream, symbol_list, code_list)

        # serialize symbol list
        for symbol in symbol_list:
            bitstream.addInt(symbol - config.offset, wordSize)

    return bitstream


def estimate_bitstream_size(config: EncoderConfig, histogram: SymbolHistogram) -> int:
    """Return the number of bits of a bitstream encoded by a given encoder configuration for a given number of symbols."""
    num_bits = config.get_codebook_size()
    for symbol in histogram.freq_map.keys():
        freq = histogram.freq_map[symbol]
        code = config.codebook[symbol]
        bits_per_code = len(code)
        num_bits += freq * bits_per_code

    return num_bits


def encode_symbols(config: EncoderConfig, symbols: List[int]) -> Bitstream:
    """Encode symbols into a bitstream according to a codebook."""
    bitstream = Bitstream()
    for symbol in symbols:
        bs_symbol = Bitstream()
        bs_symbol.bitarray = config.codebook[symbol]
        bitstream.appendBitstream(bs_symbol)

    return bitstream


def _serialize_tree(
    codebook_inverse: Dict[str, int],
    bitstream: Bitstream,
    symbol_list: List[int],
    code_list: List[bitarray],
    code: str = "",
) -> None:
    """Encode tree into a bitstream.

    :param codebook_inverse: Dictionary which contains the inverse
        Huffman codebook, in argument
    :param bitstream: The resulting bitstream for the tree, in/out
        argument
    :param symbol_list: The symbols will be in the order that the
        tree was traversed, in/out argument
    :param code_list: The code words will be in the order that the
        tree was traversed, in/out argument
    """
    if code in codebook_inverse.keys():
        bitstream.addBool(True)
        codeword = bitarray(code)
        symbol = codebook_inverse[code]
        symbol_list.append(symbol)
        code_list.append(codeword)
    else:
        bitstream.addBool(False)
        _serialize_tree(codebook_inverse, bitstream, symbol_list, code_list, code + "0")
        _serialize_tree(codebook_inverse, bitstream, symbol_list, code_list, code + "1")


def _get_codebook_inverse(codebook: Dict[int, bitarray]) -> Dict[str, int]:
    """Return an inverse codebook where the code words are given as strings.

    :param codebook: Dictionary which contains the codebook, in argument
    """
    code_list_str = [code.to01() for code in codebook.values()]
    symbol_list_int = [int(symbol) for symbol in codebook.keys()]
    codebook_inverse = dict(zip(code_list_str, symbol_list_int))

    return codebook_inverse


class HuffmanDecoder:
    """Huffman decoder based on bitarray
    :var __codebook:            dict, Huffman code book (symbol -> code)
    :var __codebookInverse:     dict, inverse Huffman code book (code -> symbol)
    :var __fixed_length:        bool, flag: use fixed length encoding instead of Huffman encoding
    :var __word_size:           int, word size for fixed length encoding / symbol table entries
    """

    bitsForWordSize = 6  # number of bits for the word_size value

    def __init__(self, bs_codebook=None):
        """Constructor
        :var val_list:          list(Object), symbol list
        """
        self.__codebook = None
        self.__fixed_length = False
        self.__word_size = 0

        if bs_codebook is not None:
            self.deserializeCodebook(bs_codebook)

    def deserializeCodebook(self, bs):
        """Deserialize codebook
        :param bs:              Bitstream, code book bitstream
        """

        # read "fixed_length" flag from codebook header
        self.__fixed_length = bs.popBool()

        if self.__fixed_length == True:
            self.__deserializeCodebookFixedLength(bs)
        else:
            self.__deserializeCodebookHuffman(bs)

    def decodeSingle(self, bs):
        """decode single symbol from bitstream
        :param bs:              Bitstream, bitstream data
        :return:                int, plain text symbol
        """
        if self.__fixed_length == True:
            return bs.popInt(self.__word_size) + self.__offset
        else:
            code = str()
            while code not in self.__codebookInverse:
                code += str(bs.popInt(1))
            return self.__codebookInverse[code]

    def decode(self, bs):
        """decode bitstream data
        :param bs:              Bitstream, bitstream data
        :return:                list(int), plain text
        """
        plain = list()
        while len(bs) > 0:
            plain.append(self.decodeSingle(bs))
        return plain

    def __deserializeCodebookFixedLength(self, bs):
        """Serialize fixed length codebook
        :param bs:              Bitstream, code book bitstream
        """

        # read "offset" value
        use_offset = bs.popBool()
        if use_offset:
            word_size_offset = bs.popInt(HuffmanDecoder.bitsForWordSize)
            self.__offset = bs.popInt(word_size_offset)
        else:
            self.__offset = 0

        # read "word_size" value from codebook header
        self.__word_size = bs.popInt(HuffmanDecoder.bitsForWordSize)

    def __deserializeCodebookHuffman(self, bs):
        """Serialize Huffman codebook
        :param bs:              Bitstream, code book bitstream
        """

        # read "offset" value
        use_offset = bs.popBool()
        if use_offset:
            word_size_offset = bs.popInt(HuffmanDecoder.bitsForWordSize)
            self.__offset = bs.popInt(word_size_offset)
        else:
            self.__offset = 0

        # read "word_size" value from codebook header
        word_size = bs.popInt(HuffmanDecoder.bitsForWordSize)

        # deserialize decoding tree
        code_list = list()
        self.__traverseTree(bs, code_list)
        symbol_list = list()
        for n in range(len(code_list)):
            symbol = bs.popInt(word_size)
            symbol_list.append(symbol + self.__offset)

        # generate codebook
        self.__codebook = {symbol: bitarray(code) for symbol, code in zip(symbol_list, code_list)}
        self.__codebookInverse = {code: symbol for symbol, code in zip(symbol_list, code_list)}

    def __traverseTree(self, bs, code_list, code=None):
        """traverse the Huffman code tree and interprete "0" as branches and "1" as leafs
        :param bs:              Bitstream, encoded bitstream using fixed length encoding
        :param code_list:       list(int), generated list with code words (output)
        :param code:            str, internal parameter which contains the iteration code
        """
        if code is None:
            code = str()
        is_leaf = bs.popBool()
        if is_leaf:
            code_list.append(code)
        else:
            self.__traverseTree(bs, code_list, code + "0")
            self.__traverseTree(bs, code_list, code + "1")

    @property
    def codebook(self):
        """Get Huffman code book
        :return:                dict, Huffman code book
        """
        return self.__codebook

    @property
    def fixed_length(self):
        """Get "fixed_length" flag"
        :return:                bool, flag: use fixed length encoding instead of Huffman encoding
        """
        return self.__fixed_length

    @property
    def offset(self):
        """Get offset
        :return:                int, offset for encoded values
        """
        return self.__offset
