# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import copy
from typing import Any, Dict, List, Tuple, Type

import trimesh

from mpegi.cli import Config
from mpegi.geometry.trimesh import copy_trimesh

from .component import Component
from .diskcache import DiskCache


class Inventory:
    """The inventory is a place where components store and retrieve their artifacts.
    In addition, this is also where packets retrieve the artifacts.
    """

    def __init__(self, config: Config, cacheDirectory: str):
        """Instanciate the inventory with the preexisting cli configuration.
        The inventory also instanciates the disk cache with the name of the scene.

        :param config: The cli Config
        """
        self.config = config
        self.artifacts: Dict[str, Any] = {}
        self.diskCache = DiskCache(cacheDirectory)

    def load(self, Component: Type[Component]) -> None:
        """Tries to load a components artifacts from disk and then stores them
        in the inventory.

        :param Component: The component to load
        :raises FileNotFoundError: Component artifacts don't exist on disk
        """
        try:
            subscribedObjects = {}
            args, inputArtifacts = self.getArgsAndArtifacts(Component)
            subscribedObjects.update(args)
            subscribedObjects.update(inputArtifacts)
            outputArtifacts = self.diskCache.retrieve(
                Component.__name__, Component.getVersion(), subscribedObjects
            )
        except:
            raise FileNotFoundError(f"Artifacts for {Component.__name__} do not exist")

        self.store(Component, outputArtifacts)

    def store(
        self,
        Component: Type[Component],
        outputArtifacts: Dict[str, Any],
        force: bool = False,
    ) -> None:
        """Stores the artifacts of a component in the inventory and saves them
        to the disk cache.

        :param Component: The component to store
        :param outputArtifacts: The output artifacts of the given component
        :type outputArtifacts: Dict[str, Any]
        :param force: If true, overwrites cache files, defaults to False
        :type force: bool, optional
        """
        # Store to inventory
        self.artifacts.update(outputArtifacts)

        # Store to disk
        subscribedObjects = {}
        args, inputArtifacts = self.getArgsAndArtifacts(Component)
        subscribedObjects.update(args)
        subscribedObjects.update(inputArtifacts)
        self.diskCache.save(
            Component.__name__,
            Component.getVersion(),
            subscribedObjects,
            outputArtifacts,
            force,
        )

    def getArtifacts(self, subscription: List[str]) -> Dict[str, Any]:
        """Helper function to retrieve artifacts from a subscription

        :param subscription: The string of the requested artifacts
        :type subscription: List[str]
        :raises KeyError: Artifact doesn't exist in inventory
        :return: A dict where the keys are the given subscription and the values
            are the coresponding artifacts
        :rtype: Dict[str, Any]
        """
        newArtifacts = {}
        for requested_artifact in subscription:
            try:
                # Artifacts are intended to be immutable - python doesn't like that
                # Deep copies achieve the same function at memory expense
                artifact = self.artifacts[requested_artifact]
                if isinstance(artifact, trimesh.Trimesh):  # TODO: This is not good... framework
                    artifactCopy = copy_trimesh(artifact)  # should be artifact agnostic
                else:
                    artifactCopy = copy.deepcopy(artifact)
                newArtifacts[requested_artifact] = artifactCopy
            except KeyError:
                raise KeyError(f"The artifact '{requested_artifact}' does not exist.")
        return newArtifacts

    def getArguments(self, subscription: List[str]) -> Dict[str, Any]:
        """Helper function to retrieve arguments from a subscription

        :param subscription: The string of the requested artifacts
        :type subscription: List[str]
        :return: A dict where the keys are the given subscription and the values
            are the coresponding arguments
        :rtype: Dict[str, Any]
        """
        newArguments = {}
        for arg in subscription:
            newArguments[arg] = copy.deepcopy(self.config[arg])
        return newArguments

    def getArgsAndArtifacts(
        self, Component: Type[Component]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Helper function to return the arguments and artifacts from a components
        subscription

        :param Component: The component from which to return the args/artifacts from
        :return: First element of the tuple is arguments and second artifacts
        :rtype: Tuple[Dict[str, Any], Dict[str, Any]]
        """
        subbedArgs = self.getArguments(Component.getArgumentSubscription())
        subbedArtifacts = self.getArtifacts(Component.getArtifactSubscription())
        return (subbedArgs, subbedArtifacts)
