# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Module that contains classes, methods, and data structures to make MHAS packets."""

import argparse
import enum
from abc import abstractmethod
from typing import Any, Dict, List, Type

from .bitstream import Bitstream
from .inventory import Inventory


class MHASPacketType(enum.IntEnum):
    """MPEG-I packet types for the MHAS packet structure.

    15-127 types are free for ISO use as is determined by the MPEG-H
    specification.
    """

    PACTYP_CFG = 23
    PACTYP_PLD = 24
    PACTYP_UPD = 25


class Packet:
    """Interface to declare a packet."""

    @abstractmethod
    def __init__(self, args: Dict[str, Any]) -> None:
        """Construct a packet."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_packet_type() -> MHASPacketType:
        """Return Packet type."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def getArgumentSubscription() -> List[str]:
        """Return a list of all required arguments."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def register_arguments(arg_parser: argparse.ArgumentParser) -> None:
        """Register arguments with argparse for CLI configuration.

        See https://docs.python.org/3/library/argparse.html for more
        details
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize(self, inventory: Inventory) -> List[Bitstream]:
        """Serialize the payload for this packet."""
        raise NotImplementedError()


def build_packets(packet_to_build: Type[Packet], inv: Inventory) -> List[Bitstream]:
    """Build and serialize packets with a given inventory.

    It could be the case that the number of packet bitstreams is more than one
    depending on how the packets are composed.

    :param packet_to_build: The given packet to build
    :param inv: The inventory that is already filled
    :return: The packets as bitstreams.
    """
    subbed_args = inv.getArguments(packet_to_build.getArgumentSubscription())
    concrete_packet = packet_to_build(subbed_args)
    packet_contents = concrete_packet.serialize(inv)

    packets: List[Bitstream] = []
    for packet_content in packet_contents:
        packet_content.pad()
        packet_header = _serialize_packet_header(
            packet_to_build.get_packet_type(),
            int(packet_content.getLength() / 8),
        )

        packet_as_bstream = Bitstream()
        packet_as_bstream.appendBitstream(packet_header)
        packet_as_bstream.appendBitstream(packet_content)
        packets.append(packet_as_bstream)

    return packets


def _serialize_packet_header(packet_type: MHASPacketType, content_length: int) -> Bitstream:
    """Return a bitstream for MHAS 'header' for a given packet type.

    Header information includes data about the packet type, the label, the
    length of the packet itself.

    :param packet_type: The packet type in which to create the header for
    :param content_length: The length of the content in bytes
    :return: The bitstream header for the packet
    """
    header = Bitstream()
    header.addInt(int(packet_type), 11)
    header.addInt(0, 2)  # Label not currently used for anything
    header.addInt(content_length, 59)
    return header
