# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

from . import tokens


class SyntaxFormatter:
    def __init__(self, syntax: List[tokens.Token], is_scope_open: List[bool] = []) -> None:
        # TODO: isScopeOpen is not a good name - think of a better one
        self.is_scope_open: List[bool] = is_scope_open
        self.syntax = syntax
        self.token_formatters = {
            tokens.CString: SyntaxFormatter.cstring_to_md,
            tokens.Blob: SyntaxFormatter.blob_to_md,
            # tokens.EscapedInteger: SyntaxFormatter.escapedIntToMarkdown,
            tokens.Label: SyntaxFormatter.label_to_md,
            tokens.Comment: SyntaxFormatter.comment_to_md,
            tokens.Flag: SyntaxFormatter.flag_to_md,
            tokens.Float32: SyntaxFormatter.float_to_md,
            tokens.Conditional: SyntaxFormatter.conditional_to_md,
            tokens.SignedInteger: SyntaxFormatter.signed_int_to_md,
            tokens.Repeat: SyntaxFormatter.repeat_to_md,
            tokens.UnsignedInteger: SyntaxFormatter.int_to_md,
            tokens.Switch: SyntaxFormatter.switch_to_md,
            tokens.QuantizedFloat: SyntaxFormatter.quantized_float_to_md,
        }

    def __str__(self) -> str:
        md_string = ""

        if len(self.syntax) == 0:
            return md_string

        if not isinstance(self.syntax[0], tokens.Label) and self.is_in_root_scope():
            raise SyntaxError("First element in bitstream syntax must be a Label")

        if self.is_in_root_scope():
            md_string = SyntaxFormatter.get_header("MPEGI Bitstream Syntax")
            md_string += "\n\n"

        for token in self.syntax:
            if not self.is_in_root_scope() and isinstance(token, tokens.Label):
                md_string += self.close_scope() + "\n"

            token_to_md_converter = self.token_formatters[type(token)]
            md_string += token_to_md_converter(self, token)

        md_string += self.close_scope()

        if self.is_in_root_scope():
            md_string += "\n"

            all_tokens = tokens.flatten_tokens(self.syntax)
            quantized_float_tokens = [t for t in all_tokens if isinstance(t, tokens.QuantizedFloat)]
            for q_float_token in quantized_float_tokens:
                md_string += SyntaxFormatter.get_quantized_float_appendix(q_float_token)
                md_string += "\n"

            md_string = SyntaxFormatter.pad_tables(md_string)

        return md_string

    # --- Elements -> Markdown --- #

    def label_to_md(self, label: tokens.Label) -> str:
        scope_depth = len(self.is_scope_open)
        self.is_scope_open.append(True)

        label_as_string = SyntaxFormatter.get_header(label.label)
        label_as_string += "\n\n"
        label_as_string += SyntaxFormatter.get_table_header(label.label.upper())
        label_as_string += SyntaxFormatter.get_table_row(f"{label.label}() {{", "", scope_depth, "")
        return label_as_string

    def close_scope(self) -> str:
        self.is_scope_open.pop()
        scope_depth = len(self.is_scope_open)
        return self.get_table_row("}", "", scope_depth, "")

    def flag_to_md(self, flag: tokens.Flag) -> str:
        scope_depth = len(self.is_scope_open)
        return SyntaxFormatter.get_table_row(flag.label + ";", _Mnemonics.bslbf, scope_depth, "1")

    def int_to_md(self, integer: tokens.UnsignedInteger) -> str:
        scope_depth = len(self.is_scope_open)
        return SyntaxFormatter.get_table_row(
            integer.label + ";", _Mnemonics.uimsbf, scope_depth, str(integer.width)
        )

    def signed_int_to_md(self, signed_int: tokens.SignedInteger) -> str:
        scope_depth = len(self.is_scope_open)
        return SyntaxFormatter.get_table_row(
            signed_int.label + ";", _Mnemonics.tcimsbf, scope_depth, str(signed_int.width)
        )

    # def escapedIntToMarkdown(self, escapedInt: tokens.EscapedInteger) -> str:
    #     scopeDepth = len(self.isScopeOpen)
    #     lengths = f"{escapedInt.width[0]}, {escapedInt.width[1]}, {escapedInt.width[2]}"
    #     return SyntaxFormatter.getTableRow(escapedInt.label, _Mnemonics.escval, scopeDepth, lengths)

    def float_to_md(self, float_point: tokens.Float32) -> str:
        scope_depth = len(self.is_scope_open)
        return SyntaxFormatter.get_table_row(
            float_point.label + ";", _Mnemonics.float, scope_depth, "32"
        )

    def cstring_to_md(self, cstring: tokens.CString) -> str:
        scope_depth = len(self.is_scope_open)
        return SyntaxFormatter.get_table_row(
            cstring.label + ";", _Mnemonics.vlclbf, scope_depth, "8..*"
        )

    def blob_to_md(self, blob: tokens.Blob) -> str:
        scope_depth = len(self.is_scope_open)
        label_length = blob.label + "Length"
        blob_length = f"{tokens.Blob.bitsForBlobLength}"
        md = SyntaxFormatter.get_table_row(
            label_length + ";", _Mnemonics.uimsbf, scope_depth, blob_length
        )
        md += SyntaxFormatter.get_table_row(
            blob.label + ";", _Mnemonics.bslbf, scope_depth, label_length
        )
        return md

    def comment_to_md(self, comment: tokens.Comment) -> str:
        scope_depth = len(self.is_scope_open)
        return SyntaxFormatter.get_table_row(comment.comment, "", scope_depth)

    def scope_to_md(self, scope_header: str, scope: List[tokens.Token]) -> str:
        scope_depth = len(self.is_scope_open)
        self.is_scope_open.append(True)

        scope_as_md = SyntaxFormatter.get_table_row(scope_header, "", scope_depth)

        is_scope_open = [True for _ in self.is_scope_open]
        scope_as_markdown = str(SyntaxFormatter(scope, is_scope_open))
        scope_as_md += scope_as_markdown

        self.is_scope_open.pop()

        return scope_as_md

    def repeat_to_md(self, repeat: tokens.Repeat) -> str:
        scope_depth = len(self.is_scope_open) - 1
        incrementer = chr(ord("i") + scope_depth)
        scope_header = (
            f"for (int {incrementer} = 0; "
            f"{incrementer} < {repeat.iterations}; "
            f"{incrementer}++) {{"
        )
        scope_as_markdown = self.scope_to_md(scope_header, repeat.syntax)
        return scope_as_markdown

    def conditional_to_md(self, conditonal: tokens.Conditional) -> str:
        scope_header = f"if ({conditonal.flag_label}) {{"
        scope_as_md = self.scope_to_md(scope_header, conditonal.true)

        if len(conditonal.false) != 0:
            scope_header = f"else {{"
            scope_as_md += self.scope_to_md(scope_header, conditonal.false)

        return scope_as_md

    def switch_to_md(self, switch: tokens.Switch) -> str:
        case_content: str = ""
        for case_int, case_scope in switch.cases.items():
            case_header = f"case {case_int}: {{"
            case_syntax: List[tokens.Token] = [*case_scope.syntax, tokens.Comment("break;")]
            case_scope = self.scope_to_md(case_header, case_syntax)
            case_content += case_scope

        switch_start = self.get_table_row(f"switch({switch.label}) {{", "", len(self.is_scope_open))
        switch_end = self.get_table_row("}", "", len(self.is_scope_open))
        foo = switch_start + case_content + switch_end
        return foo

    def quantized_float_to_md(self, q_float: tokens.QuantizedFloat) -> str:
        word_width = q_float.quant_range.num_bits
        int_token = tokens.UnsignedInteger(q_float.label, width=word_width)
        return self.int_to_md(int_token)

    # --- Markdown Helpers --- #

    @staticmethod
    def get_header(header: str, level: int = 1) -> str:
        return f"{'#' * level} {header}"

    @staticmethod
    def get_table_header(headerTitle: str) -> str:
        header = f"| {headerTitle} | NO. OF BITS | MNEMONIC |\n"
        header += "| --- | --- | --- |\n"
        return header

    @staticmethod
    def get_table_row(name: str, mnemonic: str, scopeDepth: int, width: str = "") -> str:
        tabbed_string = SyntaxFormatter.get_tabbed_string(scopeDepth)
        width_formatted = width + " " if width != "" else ""
        mnemonic_formatted = mnemonic + " " if mnemonic != "" else ""
        return f"| `{tabbed_string}{name} ` | {width_formatted}| {mnemonic_formatted}|\n"

    @staticmethod
    def get_tabbed_string(tabs: int) -> str:
        tabDepth = 4
        return " " * tabs * tabDepth

    @staticmethod
    def get_quantized_float_appendix(q_float: tokens.QuantizedFloat) -> str:
        return (
            f"## {q_float.label}\n"
            "integerized floating point representing a range between:\n"
            f"min={q_float.quant_range.min}f\n"
            f"max={q_float.quant_range.max}f\n"
        )

    # --- Helpers --- #

    def is_in_root_scope(self) -> bool:
        return len(self.is_scope_open) == 0

    @staticmethod
    def pad_tables(md_string: str) -> str:
        split_lines = md_string.split("\n")
        split_lines = [l for l in split_lines if len(l) != 0]
        split_lines = [l for l in split_lines if l[0] == "|"]

        tables: List[List[str]] = []
        for line in split_lines:
            if "MNEMONIC" in line:
                tables.append([])

            tables[-1].append(line)

        if len(tables) == 0:
            return md_string

        tables = [SyntaxFormatter._pad_table(t) for t in tables]

        new_md_string = ""
        table = tables.pop(0)
        for line in md_string.split("\n"):
            if len(line) == 0:
                new_md_string += "\n"
                continue

            if line[0] != "|":
                new_md_string += line + "\n"
                continue

            new_md_string += table.pop(0) + "\n"
            if len(table) == 0 and len(tables) != 0:
                table = tables.pop(0)

        new_md_string = new_md_string[:-1]

        return new_md_string

    @staticmethod
    def _pad_table(table: List[str]) -> List[str]:
        for column in [0, 1, 2]:
            longest_length = SyntaxFormatter._find_longest_length_in_column(table, column)
            table = SyntaxFormatter._pad_cell(table, column, 0, " ", longest_length)
            table = SyntaxFormatter._pad_cell(table, column, 1, "-", longest_length)
            for row in range(2, len(table)):
                table = SyntaxFormatter._pad_cell(table, column, row, " ", longest_length)

        return table

    @staticmethod
    def _find_longest_length_in_column(table: List[str], column: int) -> int:
        length = 0
        for row in table:
            content = SyntaxFormatter._get_content_from_column(row, column)
            if len(content) > length:
                length = len(content)
        return length

    @staticmethod
    def _pad_cell(table: List[str], column: int, row: int, pad_char: str, limit: int) -> List[str]:
        cell_content = SyntaxFormatter._get_content_from_column(table[row], column)
        padded_cell_content = cell_content

        is_backtick = False
        if len(padded_cell_content) != 0:
            is_backtick = padded_cell_content[-1] == "`"

        if is_backtick:
            padded_cell_content = padded_cell_content[:-1]
            limit -= 1

        while len(padded_cell_content) < limit:
            padded_cell_content += pad_char

        if is_backtick:
            padded_cell_content += "`"

        table[row] = SyntaxFormatter._delete_cell_content(table[row], column)
        table[row] = SyntaxFormatter._insert_cell_content(table[row], column, padded_cell_content)

        return table

    @staticmethod
    def _get_content_from_column(row: str, column: int) -> str:
        for _ in range(column):
            row = row[1:]
            while row[0] != "|":
                row = row[1:]

        start = row.find("|") + 2
        end = row.find("|", 1) - 1
        return row[start:end]

    @staticmethod
    def _delete_cell_content(row: str, column: int) -> str:
        index = 0
        while column > 0:
            index = row.find("|", index + 1)
            column -= 1

        while row[index + 2] != "|":
            row = row[: index + 1] + row[index + 2 :]

        return row

    @staticmethod
    def _insert_cell_content(row: str, column: int, content: str) -> str:
        index = 0
        while column > 0:
            index = row.find("|", index + 1)
            column -= 1

        return row[: index + 1] + f" {content} " + row[index + 2 :]


class _Mnemonics:
    """Holds constants of MPEG mnemonics"""

    uimsbf = "uimsbf"  # unsigned integer most significant bit first
    tcimsbf = "tcimsbf"  # twos complement integer most significant bit first
    vlclbf = "vlclbf"  # variable code length left bit first
    bslbf = "bslbf"  # bit string left bit first
    escval = "escval"  # escaped value - this is not an MPEG Mnemonic
    float = "float"  # floating point - this is not an MPEG Mnemonic
