# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, Optional, Protocol, Tuple, Union

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token


class Operator(Protocol):
    def perform_operation(self, scope: Dict[str, Any]) -> Any:
        ...


class Assign(Token):
    def __init__(self, label: str, operator: Operator) -> None:
        super().__init__(label)
        self.label = label
        self.operator = operator

    def to_bitstream(self, value: Any) -> Bitstream:
        return super().to_bitstream(value)

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        return super().from_bitstream(bitstream, start)

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        new_value = self.operator.perform_operation(token_generator.scope)
        token_generator.scope[self.label] = new_value

        return token_generator.get_next_token()


class Add:
    def __init__(self, left: Union[str, int], right: Union[str, int]) -> None:
        self.left = left
        self.right = right

    def perform_operation(self, scope: Dict[str, Any]) -> Any:
        right = scope[self.right] if isinstance(self.right, str) else self.right
        left = scope[self.left] if isinstance(self.left, str) else self.left
        return left + right


class Lookup:
    def __init__(self, val: Any) -> None:
        self.val = val

    def perform_operation(self, scope: Dict[str, Any]) -> Any:
        return scope[self.val]


class Value:
    def __init__(self, val: Any) -> None:
        self.val = val

    def perform_operation(self, scope: Dict[str, Any]) -> Any:
        return self.val
