# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import copy
from typing import Any, List, Optional, Tuple

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token


class Conditional(Token):
    def __init__(self, flag_label: str, *, true: List[Token], false: List[Token]) -> None:
        """Condtionally introduce tokens into a syntax.

        This token takes a flag label upon construction. The composer
        will then take said label and resolve it to find the value,
        either True or False. That value determines whether to use
        the True or False token list.

        :param flag_labe: The label of the flag to be used for the
            conditional resolution.
        :param true: The list of tokens to use if the condition is true.
        :param false: The list of tokens if the condition is false.

        """
        super().__init__(flag_label)
        self.flag_label = flag_label
        self.true = true
        self.false = false

    def to_bitstream(self, value: Any) -> Bitstream:
        return super().to_bitstream(value)

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        return super().from_bitstream(bitstream, start)

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        condition = token_generator.scope[self.flag_label]

        if not isinstance(condition, (bool, int)):
            raise TypeError(
                f"ConditionalToken only supports boolean and int references. Invalid type: {condition}"
            )

        if condition:
            tokens = copy.deepcopy(self.true)
        else:
            tokens = copy.deepcopy(self.false)
        tokens.reverse()
        token_generator.token_stack.extend(tokens)
        return token_generator.get_next_token()
