# This software module was originally developed by
#
# Dolby International AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes and
# its performance may not have been optimized.  This software module is an
# implementation of one or more tools as specified by the ISO/IEC 23090-4
# standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
# copyright license to copy, distribute, and make derivative works of this
# software module or modifications thereof for use in implementations or
# products claiming conformance to ISO/IEC standards incorporating ISO/IEC
# 23090-4 reference software and which satisfy any specified conformance
# criteria of that standard.
#
# Those intending to use this software module in products are advised that its
# use may infringe existing patents and that the license in the previous
# paragraph grants no licenses under such patents.  ISO/IEC and the developers
# of this software module have no liability for use of this software module or
# modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Dolby International AB retains full right to
# modify and use the code for its own purpose, assign or donate the code to
# a third party without consent of and without accounting to third parties,
# and to inhibit third parties from using the code for products that do not
# conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, Optional, Tuple

from ..bitstream import Bitstream
from ._checktype import checkType
from ._generator import TokenGenerator
from ._token import Token

from typing import Any, Dict, Optional, Tuple


class EscapedUnsignedInteger(Token):
    """Escaped coded unsigned integer"""

    def __init__(
        self,
        label: str,
        *,
        widths: Tuple[int, int, int] = [8, 16, 32],
        code_page: Optional[Dict[int, int]] = None,
    ) -> None:
        """Construct an EscapedUnsignedInteger Token.

        If a code_page is provided, the pushed value will resolve via
        the code_page. This is useful for encoding symbols. The key of
        the dictionary should be the unencoded value and the value of
        the dictionary should be the encoded value. The values should
        not exceed the width provided.

        :param label: The first parameter
        :param widths: The width of the binary word to represent this
            int, defaults to 8, 16, 32
        :param code_page: The dictionary to map input values to output
            values
        """
        if code_page is not None:
            if max(code_page.values()) > 2 ** 8 - 1:
                raise Exception("Invalid code page: Values cannot be encoded with the given width")

        super().__init__(label)
        self.widths = widths
        self.code_page = code_page

    def to_bitstream(self, value: int) -> Bitstream:
        """Takes the incoming value and translates it into a bitstream

        :param value: The value to put into a bitstream
        :type value: int
        :return: The bitstream representation of the value
        """
        checkType(self, int, value)

        if self.code_page is not None:
            value = self.code_page[value]

        if value < 0:
            raise TypeError(
                f"EscapedUnsignedInteger Token cannot serialize negative integers: {value}"
            )

        bitstream = Bitstream()
        bitstream.addEscapedInt(value, self.widths)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        width = self.widths[0]
        token_as_bitstream = bitstream.copy(start, self.widths[0])
        value = token_as_bitstream.toInt()
        if value == ((1 << self.widths[0]) - 1):
            token_as_bitstream = bitstream.copy(start + width, self.widths[1])
            valueAdd = token_as_bitstream.toInt()
            value += valueAdd
            width += self.widths[1]
            if valueAdd == ((1 << self.widths[1]) - 1):
                token_as_bitstream = bitstream.copy(start + width, self.widths[2])
                valueAdd = token_as_bitstream.toInt()
                value += valueAdd
                width += self.widths[2]
        return (width, value)

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
