# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, List, Optional

from ._token import CommentToken, Token


class TokenGenerator:
    """Generators Tokens from a syntax through recursive resolution."""

    def __init__(self, syntax: List[Token], scope: Dict[str, Any]) -> None:
        """Construct a Token Generator.

        :param syntax: A list of Tokens
        :param scope: An object reference to scope which is updated by
            the BitstreamComposer. The key is the label of the token and
            the value is the value that is pushed to the
            :class:mpegi.framework.BitstreamCompser
        """
        token_stack = [t for t in syntax if not isinstance(t, CommentToken)]
        token_stack.reverse()
        self.token_stack: List[Token] = token_stack
        self.scope = scope
        self.ignore_markers: bool = True

    def get_next_token(self) -> Optional[Token]:
        """Return the next token in the stack.

        This function defers the decision making to the token itself.
        If the token needs to resolve something, it will take the
        state of this generator, modify the state if needed, and/or
        call recursively call this exact function to continue.

        :return: Returns the next token, returns None if at end of stack
        """
        try:
            token = self.token_stack.pop()
        except IndexError:
            return None

        return token.get_token(self)
