# This software module was originally developed by
#
# Dolby International AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes and
# its performance may not have been optimized.  This software module is an
# implementation of one or more tools as specified by the ISO/IEC 23090-4
# standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
# copyright license to copy, distribute, and make derivative works of this
# software module or modifications thereof for use in implementations or
# products claiming conformance to ISO/IEC standards incorporating ISO/IEC
# 23090-4 reference software and which satisfy any specified conformance
# criteria of that standard.
#
# Those intending to use this software module in products are advised that its
# use may infringe existing patents and that the license in the previous
# paragraph grants no licenses under such patents.  ISO/IEC and the developers
# of this software module have no liability for use of this software module or
# modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Dolby International AB retains full right to
# modify and use the code for its own purpose, assign or donate the code to
# a third party without consent of and without accounting to third parties,
# and to inhibit third parties from using the code for products that do not
# conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, List, Optional, Tuple, Union

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token


class PackedVoxelCoord(Token):
    """Packed voxel coordinate (2D / 3D)"""

    def __init__(self, label: str = "", dimensions: List[int] = None):
        super().__init__(label)
        self.label = label
        self.dimensions = dimensions

    def to_bitstream(self, value: Any) -> Bitstream:

        if isinstance(value, list):
            voxel_cord = value
            voxel_dim = self.dimensions
        elif isinstance(value, dict):
            voxel_cord = value.get("value")
            voxel_dim = value.get("dimensions")
        else:
            raise TypeError(f"PackedVoxelCoord Token cannot serialize type: {type(value)}")

        assert len(voxel_cord) == len(voxel_dim)
        """
        Determinate packed voxel coordinate and bit length
        """
        packedVoxelCoordMax: int = 1
        packedVoxelCoord: int = 0
        multiplier: int = 1
        i = 0

        for dim in voxel_dim:
            # check if voxel coordinates are valid
            assert 0 <= voxel_cord[i] - 1 < dim
            packedVoxelCoordMax *= dim
            packedVoxelCoord = packedVoxelCoord + ((voxel_cord[i] - 1) * multiplier)
            multiplier = multiplier * dim
            i += 1

        packedVoxelCoordMax -= 1
        n_bits = packedVoxelCoordMax.bit_length()

        bitstream = Bitstream()
        bitstream.addInt(packedVoxelCoord, n_bits)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        assert self.dimensions != None
        """
        Determinate packed voxel bit length
        """
        voxel_dim = self.dimensions
        packedVoxelCoordMax: int = 1

        for dim in voxel_dim:
            packedVoxelCoordMax *= dim
        packedVoxelCoordMax -= 1
        n_bits = packedVoxelCoordMax.bit_length()

        token_as_bitstream = bitstream.copy(start, n_bits)
        packedVoxelCoord = token_as_bitstream.toInt()

        """
        Determinate unpacked voxel cord
        """
        unpackedVoxelCoord = []

        for xyz in voxel_dim:
            unpackedVoxelCoord.append((packedVoxelCoord % xyz) + 1)
            packedVoxelCoord = packedVoxelCoord // xyz
        return (n_bits, unpackedVoxelCoord)

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
