# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Tuple, Optional

from ..bitstream import Bitstream, QuantizationRange, encode_float, decode_float
from ._token import Token
from ._generator import TokenGenerator


class QuantizedFloat(Token):
    """Token to represent a intergerized floating point value.

    This takes a floating point value and turns it into an integer based
    on the maximum and minmum range of values and the number of bits
    provided for it to be stored on.
    """

    def __init__(self, label: str, *, quant_range: QuantizationRange) -> None:
        """Construct a Quantized floating point bitstream token.

        :param label: The label of this Token
        :param quant_range: The quantization ranges and number of bits
            to use for this token.
        """
        self.label = label
        self.quant_range = quant_range

    def to_bitstream(self, value: Any) -> Bitstream:
        """Return a bitstream value from the incoming value.

        :return: The concrete bitstream that represents this token
        """
        try:
            encoded_value = encode_float(value, self.quant_range)
        except ValueError as ex:
            raise ValueError(f"Invalid quantized float '{self.label}': {ex}") from ex

        try:
            bitstream = Bitstream()
            bitstream.addInt(encoded_value, width=self.quant_range.num_bits)
        except Exception as exc:
            raise Exception(
                "Couldn't serialise bitstream element - "
                f"BitstreamElement: {__class__.__name__}, "
                f"Label: {self.label} "
            ) from exc
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        token_as_bitstream = bitstream.copy(start, self.quant_range.num_bits)
        value = token_as_bitstream.toInt()
        return self.quant_range.num_bits, decode_float(value, self.quant_range)

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
