# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Tuple

from ..bitstream import Bitstream

if TYPE_CHECKING:
    from ._generator import TokenGenerator


class Token(ABC):
    """Interface for syntactical tokens."""

    def __init__(self, label: str) -> None:
        """Every token must be labelled such that it can be used in Contexts for other Tokens

        :param label: The label of this Token
        """
        self.label = label

    @abstractmethod
    def to_bitstream(self, value: Any) -> Bitstream:
        """Takes the incoming arguments and constructs a bitstream from them

        :raises NotImplementedError: Must implement this
        :return: The concrete bitstream that represents this token
        """
        raise NotImplementedError()

    @abstractmethod
    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        """Return value and length of token from a given bitstream.

        :param bitstream: The Bitstream to analyze from
        :param start: Offset in bits in which to begin the resolution
        :raises NotImplementedError: Abstract method needs implementing
        :return: Returns Tuple of length of bits and the value itself
        """
        raise NotImplementedError()

    @abstractmethod
    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        raise NotImplementedError()


class CommentToken(Token):
    """This token should be completely ignored when creating a concrete bitstream"""

    def __init__(self, label: str) -> None:
        super().__init__(label)

    def to_bitstream(self, value: Any) -> Bitstream:
        """Raise not implemented error - Comments don't produce bitstreams"""
        raise NotImplementedError()

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        raise NotImplementedError()

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        raise NotImplementedError()

    def __eq__(self, other: CommentToken) -> bool:
        return self.label == other.label
