# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Utility Functions for handling syntax tokens."""

from typing import List

from ._token import Token
from ._conditional import Conditional
from ._switch import Switch
from ._repeat import Repeat
from ._function import Function
from ..bitstream import Bitstream


def flatten_tokens(token_list: List[Token]) -> List[Token]:
    """Flatten a list of tokens and return it.

    Some tokens have a list of tokens within them. This will 'unpack'
    all of these nested tokens and put them into one list. This is
    useful for unrolling a syntax out to have access to tokens that
    would otherwise be in for-loops or if statements
    """
    flattened_list: List[Token] = []

    for token in token_list:
        flattened_list.append(token)
        if isinstance(token, Conditional):
            flattened_list.extend(flatten_tokens(token.true))
            flattened_list.extend(flatten_tokens(token.false))
        elif isinstance(token, Switch):
            for case in token.cases.values():
                flattened_list.extend(flatten_tokens(case.syntax))
        elif isinstance(token, Repeat):
            flattened_list.extend(flatten_tokens(token.syntax))
        elif isinstance(token, Function):
            flattened_list.extend(flatten_tokens(token.syntax))

    return flattened_list


def readCodeword(bitstream: Bitstream, start: int, table):
    inv_table = {code: value for value, code in table.items()}
    maxLen = len(max(table.values(), key=len))
    word = ""
    num_bits = 0
    wordFound = False
    while not wordFound and num_bits <= maxLen:
        word += bitstream.copy(start + num_bits, 1).to01()
        num_bits += 1
        if word in inv_table:
            value = inv_table[word]
            wordFound = True
    if wordFound is False:
        num_bits = 0
        value = None
    return num_bits, value
