# This software module was originally developed jointly by
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


from typing import Any, Optional, Tuple

from mpegi.framework import _lookup_tables as LUT

from ..bitstream import Bitstream
from ._checktype import checkType
from ._generator import TokenGenerator
from ._token import Token
from ._util import readCodeword


class writeCountOrIndex(Token):
    """encode integer using CountOrIndex LUTs"""

    def __init__(self, label: str = "", storeValue: bool = False):
        super().__init__(label)
        self.label = label
        self.storeValue = storeValue

    def to_bitstream(self, value: int) -> Bitstream:
        checkType(self, int, value)
        if value > 1023:
            raise ValueError(
                "{}: Maximum value exceeded: Tried to encode {} (max = 1023)".format(
                    self.__class__.__name__, value
                )
            )
        threshold = 64
        isLargerNumber = value > max(LUT.count_or_index_lo_code.keys())
        bs = ""

        hiCode = int(value / threshold)
        loCode = value % threshold

        bs += LUT.count_or_index_lo_code[loCode]
        bs += str(int(isLargerNumber))

        if isLargerNumber:
            bs += LUT.count_or_index_hi_code[hiCode]

        bitstream = Bitstream()
        bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        num_bits, value = readCodeword(bitstream, start, LUT.count_or_index_lo_code)
        largerNumber = int(bitstream.copy(start + num_bits, 1).to01(), 2)
        num_bits += 1
        if largerNumber:
            hi_num_bits, hiValue = readCodeword(
                bitstream, start + num_bits, LUT.count_or_index_hi_code
            )
            value += hiValue * 64
            num_bits += hi_num_bits
        return num_bits, value

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
