# This software module was originally developed jointly by
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

# Copyright (c) ISO/IEC 2022.


from typing import Any, Optional, Tuple

import numpy as np

from mpegi.framework import _lookup_tables as LUT

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token
from ._util import readCodeword


class writeDistance(Token):
    """Write a distance as bitstring using LUTs"""

    def __init__(self, label: str = "", *, isSmallScene: bool = False):
        super().__init__(label)
        self.label = label
        self.isSmallScene = isSmallScene

    def to_bitstream(self, value) -> Bitstream:
        bs = ""
        if isinstance(value, (float, int)):
            distance = value
        elif isinstance(value, dict):
            distance = value.get("distance")
            self.isSmallScene = value.get("isSmallScene")
        if self.isSmallScene and distance > 100:
            raise ValueError(f"Small scene should not have coordinates beyond 100m")
        if distance < 0:
            raise ValueError(f"writeDistance Token cannot support negative numbers")

        eps = 1e-12
        kilometers = np.floor(distance / 1000 + eps)
        distance -= kilometers * 1000
        hectometers = np.floor(distance / 100 + eps)
        distance -= hectometers * 100
        meters = np.floor(distance + eps)
        distance -= meters
        centimeters = np.floor(distance * 100 + eps)
        distance -= centimeters / 100
        millimeters = round(distance * 1000 + eps)
        distance -= millimeters / 1000

        if millimeters == 10:
            millimeters = 0
            if centimeters == 100:
                centimeters = 0
                meters += 1
                if meters == 100:
                    meters = 0
                    hectometers += 1
                    if hectometers == 10:
                        hectometers = 0
                        kilometers += 1

        if distance.round(5) < -0.5 * 0.001 or distance.round(5) > 0.5 * 0.001:
            raise ValueError("Something went wrong breaking down the distance.")

        bs += LUT.meters_code[meters]

        if not self.isSmallScene:
            if hectometers > 0 or kilometers > 0:
                addHectoMeters = True
                bs += str(int(addHectoMeters))
                bs += LUT.hecto_meters_code[hectometers]

                kmMax = max(LUT.kilo_meters_code.keys())
                if kilometers > 0:
                    while kilometers > 0:
                        addKilometers = True
                        bs += str(int(addKilometers))
                        val = min(10, kilometers)
                        bs += LUT.kilo_meters_code[val]
                        kilometers -= val
                    bs += str(int(0))
                else:
                    addKilometers = False
                    bs += str(int(addKilometers))

            else:
                addHectometers = False
                bs += str(int(addHectometers))

        if centimeters > 0 or millimeters > 0:
            addCentimeters = True
            bs += str(int(addCentimeters))
            bs += LUT.centi_meters_code[centimeters]

            if millimeters > 0:
                addMillimeters = True
                bs += str(int(addMillimeters))
                bs += LUT.milli_meters_code[millimeters]

            else:
                addMillimeters = False
                bs += str(int(addMillimeters))

        else:
            addCentimeters = False
            bs += str(int(addCentimeters))

        bitstream = Bitstream()
        bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(
        self, bitstream: Bitstream, start: int = 0, isSmallScene: bool = None
    ) -> Tuple[int, Any]:
        if isSmallScene is not None:
            self.isSmallScene = isSmallScene
        num_bits, meters = readCodeword(bitstream, start, LUT.meters_code)
        distance = meters
        if self.isSmallScene is False:
            addHectometers = int(bitstream.copy(start + num_bits, 1).to01(), 2)
            num_bits += 1
            if addHectometers:
                num_bits_word, hectometers = readCodeword(
                    bitstream, start + num_bits, LUT.hecto_meters_code
                )
                num_bits += num_bits_word
                distance += 100 * hectometers
                addKilometers = int(bitstream.copy(start + num_bits, 1).to01(), 2)
                num_bits += 1
                while addKilometers:
                    num_bits_word, kilometers = readCodeword(
                        bitstream, start + num_bits, LUT.kilo_meters_code
                    )
                    num_bits += num_bits_word
                    distance += kilometers * 1000
                    addKilometers = int(bitstream.copy(start + num_bits, 1).to01(), 2)
                    num_bits += 1
        addCentimeters = int(bitstream.copy(start + num_bits, 1).to01(), 2)
        num_bits += 1
        if addCentimeters:
            num_bits_word, centimeters = readCodeword(
                bitstream, start + num_bits, LUT.centi_meters_code
            )
            num_bits += num_bits_word
            distance += centimeters / 100
            addMillimeters = int(bitstream.copy(start + num_bits, 1).to01(), 2)
            num_bits += 1
            if addMillimeters:
                num_bits_word, millimeters = readCodeword(
                    bitstream, start + num_bits, LUT.milli_meters_code
                )
                num_bits += num_bits_word
                distance += millimeters / 1000

        return num_bits, round(distance, 3)

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
