# This software module was originally developed jointly by
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


from typing import Any, Optional, Tuple

from mpegi.framework import _lookup_tables as LUT

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token
from ._util import readCodeword


class writeDuration(Token):
    """Write a duration in seconds as bitstring using LUTs"""

    def __init__(self, label: str, *, isLongRangeMode: bool = True) -> None:
        self.label = label
        self.islongRangeMode = isLongRangeMode

    def to_bitstream(self, duration: float) -> Bitstream:
        bs = ""
        if isinstance(duration, float):
            numDecimals = len(str(duration).split(".")[1])
        elif isinstance(duration, int):
            numDecimals = 0
        else:
            raise ValueError(f"writeDuration token only accepts float or int")
        originalDuration = duration
        addSamples = False
        rawSeconds = 0

        # This assumes that value is always as seconds
        if duration >= 1:
            rawSeconds = int(str(duration).split(".")[0])
            originalDuration = round(originalDuration - rawSeconds, numDecimals)

        addHours = duration >= 3600
        rawHours = int(duration / 3600)
        addSeconds = rawSeconds > 0
        addMilliseconds = numDecimals >= 2

        deciSecondsCode = float(str(originalDuration)[0:3])
        bs += LUT.deci_seconds_code[deciSecondsCode]

        bs += str(int(addMilliseconds))
        if addMilliseconds:
            milliSecondsCode = float(str(round((originalDuration - deciSecondsCode), 4))[0:5])
            bs += LUT.milli_seconds_code[milliSecondsCode]

            microseconds = round((originalDuration - deciSecondsCode - milliSecondsCode), 5)
            addMicroseconds = microseconds > 0

            bs += str(int(addMicroseconds))
            if addMicroseconds:
                bs += LUT.microseconds_code[microseconds]

        bs += str(int(addSeconds))
        if addSeconds:
            if self.islongRangeMode:
                rawSeconds = int(duration % 3600)
                secondsPart = "{0:012b}".format(rawSeconds)
                if len(secondsPart) != 12:
                    raise ValueError(f"Invalid seconds bit length: {len(secondsPart)}")
                bs += secondsPart
                bs += str(int(addHours))
                if addHours:
                    hourPart = "{0:05b}".format(rawHours - 1)
                    if len(hourPart) != 5:
                        raise ValueError(f"Invalid hours bit length: {len(hourPart)}")
                    bs += hourPart
            else:
                bs += LUT.seconds_code[rawSeconds]

        bitstream = Bitstream()
        bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        num_bits, deciSeconds = readCodeword(bitstream, start, LUT.deci_seconds_code)
        duration = deciSeconds
        addMilliseconds = int(bitstream.copy(start + num_bits, 1).to01(), 2)
        num_bits += 1
        if addMilliseconds:
            num_bits_word, milliSeconds = readCodeword(
                bitstream, start + num_bits, LUT.milli_seconds_code
            )
            num_bits += num_bits_word
            duration += milliSeconds
            addMicroseconds = int(bitstream.copy(start + num_bits, 1).to01(), 2)
            num_bits += 1
            if addMicroseconds:
                num_bits_word, microSeconds = readCodeword(
                    bitstream, start + num_bits, LUT.microseconds_code
                )
                num_bits += num_bits_word
                duration += microSeconds
        addSeconds = int(bitstream.copy(start + num_bits, 1).to01(), 2)
        num_bits += 1
        if addSeconds:
            if self.islongRangeMode:
                rawSeconds = int(bitstream.copy(start + num_bits, 12).to01(), 2)
                num_bits += 12
                duration += rawSeconds
                addHours = int(bitstream.copy(start + num_bits, 1).to01(), 2)
                num_bits += 1
                if addHours:
                    rawHours = int(bitstream.copy(start + num_bits, 5).to01(), 2)
                    num_bits += 5
                    duration += (rawHours + 1) * 3600
            else:
                num_bits_word, seconds = readCodeword(bitstream, start + num_bits, LUT.seconds_code)
                num_bits += num_bits_word
                duration += seconds

        return num_bits, duration

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
