# This software module was originally developed jointly by
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


from typing import Any, Optional, Tuple

import numpy as np

from mpegi.framework import _lookup_tables as LUT

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token
from ._util import readCodeword


class writeGain(Token):
    """write gain value using LUTs"""

    def __init__(self, label: str = "", *, isHiPrecGain: bool = False):
        super().__init__(label)
        self.label = label
        self.isHiPrecGain = isHiPrecGain

    def to_bitstream(self, gain) -> Bitstream:
        bs = ""
        coarseKeyArr = np.asarray(list(LUT.coarse_gain_code.keys()))
        fineKeyArr = np.asarray(list(LUT.fine_gain_code.keys()))
        addFineGain = not coarseKeyArr.__contains__(gain)

        if gain < min(coarseKeyArr) + min(fineKeyArr):
            raise ValueError(
                "writeGain does not support values < {} dB".format(
                    min(coarseKeyArr) + min(fineKeyArr)
                )
            )
        elif gain > max(coarseKeyArr):
            raise ValueError("writeGain does not support values > {} dB".format(max(coarseKeyArr)))

        if isinstance(gain, float):
            addDecimalGain = not float.is_integer(gain)
        elif isinstance(gain, int):
            addDecimalGain = False

        if self.isHiPrecGain:
            if addDecimalGain:
                # gain value is float with decimal != .0
                decimalValue = round(gain - int(gain), 2)
                if decimalValue < 0:
                    gain = int(gain // 1)
                    decimalValue = round(decimalValue + 1, 2)
                addFineGain = not coarseKeyArr.__contains__(int(gain))
                decimalCode = LUT.deci_seconds_code[decimalValue]

        gain = int(gain)

        if addFineGain:
            idxFound = False
            for i in range(len(coarseKeyArr)):
                if idxFound:
                    break
                coarseVal = coarseKeyArr[i]
                for j in range(len(fineKeyArr)):
                    fineVal = fineKeyArr[j]
                    if fineVal + coarseVal == gain:
                        coarseIdx = list(coarseKeyArr).index(coarseVal)
                        fineIdx = list(fineKeyArr).index(fineVal)
                        coarseValue = coarseKeyArr[coarseIdx]
                        fineValue = fineKeyArr[fineIdx]
                        idxFound = True
                        break

            coarseCode = LUT.coarse_gain_code[coarseValue]
            fineCode = LUT.fine_gain_code[fineValue]

        else:
            coarseCode = LUT.coarse_gain_code[gain]

        bs += coarseCode

        bs += str(int(addFineGain))
        if addFineGain:
            bs += fineCode

        if self.isHiPrecGain:
            bs += str(int(addDecimalGain))
            if addDecimalGain:
                bs += decimalCode

        bitstream = Bitstream()
        bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        num_bits, gain = readCodeword(bitstream, start, LUT.coarse_gain_code)
        addFineGain = int(bitstream.copy(start + num_bits, 1).to01(), 2)
        num_bits += 1
        if addFineGain:
            num_bits_word, fineGain = readCodeword(bitstream, start + num_bits, LUT.fine_gain_code)
            num_bits += num_bits_word
            gain += fineGain
        if self.isHiPrecGain:
            addDecimalGain = int(bitstream.copy(start + num_bits, 1).to01(), 2)
            num_bits += 1
            if addDecimalGain:
                num_bits_word, decimalGain = readCodeword(
                    bitstream, start + num_bits, LUT.deci_seconds_code
                )
                num_bits += num_bits_word
                gain += decimalGain
        return num_bits, gain

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
