# This software module was originally developed jointly by
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


from typing import Any, Optional, Tuple

from ..bitstream import Bitstream
from ._checktype import checkType
from ._generator import TokenGenerator
from ._token import Token
import numpy as np


class writeID(Token):
    """write ID as a varlaible length bitstring"""

    def __init__(self, label: str = ""):
        super().__init__(label)
        self.label = label

    def to_bitstream(self, value: int) -> Bitstream:
        checkType(self, int, value)
        id = value
        if id < 0:
            raise ValueError(f"id value cannot be negative: {id}")

        bs = ""
        if id == 0:
            # Special case
            scale = 1
        else:
            scale = np.power(2, (7 * np.floor(np.log2(id) / 7)))

        while scale >= 1:
            idVal = np.floor(id / scale)
            bs += "{0:07b}".format(int(idVal))
            id = id - idVal * scale
            scale /= 128
            if scale >= 1:
                largerValue = 1
                bs += "{0:01b}".format(largerValue)
            else:
                largerValue = 0
                bs += "{0:01b}".format(largerValue)

        bitstream = Bitstream()
        bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        id = int(bitstream.copy(start, 7).to01(), 2)

        largerValue = int(bitstream.copy(start + 7, 1).to01(), 2)
        num_bits = 8

        while largerValue:
            idVal = int(bitstream.copy(start + num_bits, 7).to01(), 2)
            id = id * 128 + idVal
            largerValue = int(bitstream.copy(start + num_bits + 7, 1).to01(), 2)
            num_bits += 8

        return num_bits, id

        # print(Bitstream)
        # raise NotImplementedError(
        #     f'Class {self.__class__.__name__} has no "from_bitstream" function implemented'
        # )

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
