# This software module was originally developed jointly by
#
# Fraunhofer IIS and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


from typing import Any, Optional, Tuple

import numpy as np

from mpegi.framework import _lookup_tables as LUT

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token
from ._util import readCodeword


class writeOrientation(Token):
    """write ID as a varlaible length bitstring"""

    def __init__(self, label: str = "", *, highResolution: bool = False):
        super().__init__(label)
        self.label = label
        self.highResolution = highResolution

    def to_bitstream(self, value: list) -> Bitstream:
        bs = ""
        # first test if the orientation is in the default orientation code list - convert to string first.

        oriStr = "{:s}".format(",".join(["{:.3f}".format(x) for x in value]))
        if LUT.orientation_code.keys().__contains__(oriStr):
            bs += LUT.orientation_code[oriStr]
            bitstream = Bitstream()
            bitstream = bitstream.from01(bs)
        # if not, see if all 3 values are in their respective default lists
        elif (
            LUT.default_yaw_code.keys().__contains__(value[0])
            and LUT.default_pitch_code.keys().__contains__(value[1])
            and LUT.default_roll_code.keys().__contains__(value[2])
        ):
            bs += "110"
            bs += LUT.default_yaw_code[value[0]]
            bs += LUT.default_pitch_code[value[1]]
            bs += LUT.default_roll_code[value[2]]
            bitstream = Bitstream()
            bitstream = bitstream.from01(bs)
        # else, construct if from the explicit angles...
        else:
            if self.highResolution is False:
                bs += "111"
                for angle in value:
                    if angle == 180:
                        angle = -180
                    if LUT.coarse_angle_code.keys().__contains__(angle):
                        bs += LUT.coarse_angle_code[angle]
                        bs += "0"
                    else:
                        coarseValue, fineValue = self.findIndex(angle)
                        bs += LUT.coarse_angle_code[coarseValue]
                        bs += "1"
                        bs += LUT.fine_angle_code[fineValue]
                bitstream = Bitstream()
                bitstream = bitstream.from01(bs)
            else:
                bs += "100"
                bitstream = Bitstream()
                bitstream = bitstream.from01(bs)
                for angle in value:
                    bitstream.addFloat32(angle)

        # bitstream = Bitstream()
        # bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        orientation = []
        num_bits, ori_code = readCodeword(bitstream, start, LUT.orientation_code)
        if ori_code is None:
            ori_code = bitstream.copy(start + num_bits, 3).to01()
            num_bits += 3
            if ori_code == "110":
                num_bits_word, value = readCodeword(
                    bitstream, start + num_bits, LUT.default_yaw_code
                )
                num_bits += num_bits_word
                orientation.append(value)
                num_bits_word, value = readCodeword(
                    bitstream, start + num_bits, LUT.default_pitch_code
                )
                num_bits += num_bits_word
                orientation.append(value)
                num_bits_word, value = readCodeword(
                    bitstream, start + num_bits, LUT.default_roll_code
                )
                num_bits += num_bits_word
                orientation.append(value)
            elif ori_code == "111":
                for o in range(0, 3):
                    num_bits_word, coarseAngle = readCodeword(
                        bitstream, start + num_bits, LUT.coarse_angle_code
                    )
                    angle = coarseAngle
                    num_bits += num_bits_word
                    addFineAngle = int(bitstream.copy(start + num_bits, 1).to01(), 2)
                    num_bits += 1
                    if addFineAngle:
                        num_bits_word, fineAngle = readCodeword(
                            bitstream, start + num_bits, LUT.coarse_angle_code
                        )
                        num_bits += num_bits_word
                        angle += fineAngle
                    orientation.append(angle)
            elif ori_code == "100":
                for o in range(0, 3):
                    angle = bitstream.copy(start + num_bits, 32).toFloat()
                    num_bits += 32
                    orientation.append(angle)
        else:
            orientation = [float(i) for i in ori_code.split(",")]
        return num_bits, orientation

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self

    def findIndex(self, angle):
        coarseKeyArr = np.asarray(list(LUT.coarse_angle_code.keys()))
        fineKeyArr = np.asarray(list(LUT.fine_angle_code.keys()))
        coarseValue = 0
        fineValue = 0
        diff = 1
        for i in range(len(coarseKeyArr)):
            coarseVal = coarseKeyArr[i]
            for j in range(len(fineKeyArr)):
                fineVal = fineKeyArr[j]
                if abs((fineVal + coarseVal) - angle) < diff:
                    coarseValue = coarseKeyArr[list(coarseKeyArr).index(coarseVal)]
                    fineValue = fineKeyArr[list(fineKeyArr).index(fineVal)]
                    diff = abs((fineVal + coarseVal) - angle)
        if not diff == 0:
            print("Couldn't find an exact angle match, mismatch by {:.6f} radians".format(diff))
        return coarseValue, fineValue
