# This software module was originally developed jointly by
#
# Fraunhofer IIS and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


import logging
from typing import Any, Optional, Tuple

import numpy as np

from mpegi.framework import _lookup_tables as LUT

from ..bitstream import Bitstream
from ._generator import TokenGenerator
from ._token import Token
from ._util import readCodeword


class writeValueFromLUT(Token):
    """write DSR as a varlaible length bitstring, assumes DSR in the Linear domain"""

    def __init__(self, label: str = "", tableName: str = ""):
        super().__init__(label)
        self.label = label
        self.lookupTable = tableName

    def to_bitstream(self, value: float) -> Bitstream:
        # raise NotImplementedError(
        #     f'Class {self.__class__.__name__} not implemented')
        bs = ""

        try:
            table = getattr(LUT, self.lookupTable)
        except:
            raise ValueError(f"Lookup Table: {self.lookupTable} does not exist.")

        if not table.keys().__contains__(value):
            table_vals = np.asarray([int(key) for key in table])
            closestIdx = (np.abs(table_vals - value)).argmin()
            diff = value - table_vals[closestIdx]
            new_value = table_vals[closestIdx]
            logging.info(
                f"Value: {value} was not found in table {self.lookupTable}, using nearest value of: {new_value}, a diff of {diff}"
            )
            value = new_value

        bs += table[value]

        bitstream = Bitstream()
        bitstream = bitstream.from01(bs)
        return bitstream

    def from_bitstream(self, bitstream: Bitstream, start: int = 0) -> Tuple[int, Any]:
        try:
            table = getattr(LUT, self.lookupTable)
        except:
            raise ValueError(f"Lookup Table: {self.lookupTable} does not exist.")

        num_bits, value = readCodeword(bitstream, start, table)

        return num_bits, value

    def get_token(self, token_generator: TokenGenerator) -> Optional[Token]:
        return self
