# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
#  Various intersection examples
#

import numpy as np


def intersectionRayPlane(x0, d, p0, p1, p2):
    """Ray-plane intersection test
    Computes the intersection of the ray x(t) = x0 + t*d
    with the plane in parametric form y(u,v) = p0 + u*(p1-p0) + v*(p2-p0)

    :param x0: Ray origin (vector3)
    :param d: Ray direction (vector3)
    :param p0: Plane point (vector3)
    :param p1: Plane point (vector3)
    :param p2: Plane point (vector3)

    :returns [t, u, v] (vector3) when the ray intersects the plane,
             None otherwise

    Source: https://en.wikipedia.org/wiki/Line%E2%80%93plane_intersection
    """

    x0 = np.array(x0)
    d = np.array(d)
    p0 = np.array(p0)
    p1 = np.array(p1)
    p2 = np.array(p2)

    p01 = p1 - p0
    p02 = p2 - p0

    # Intermediate terms
    p01_x_p02 = np.cross(p01, p02)

    # Intersection anyway? Check via determinant...
    # Determinant = 0  -->  Ray and plane are parallel  -->  no intersection
    det = np.dot(-d, p01_x_p02)
    if np.abs(det) < 1e-9:
        return None

    # Compute the solution
    s = np.array(
        [
            np.dot(p01_x_p02, x0 - p0) / det,
            np.dot(np.cross(p02, -d), x0 - p0) / det,
            np.dot(np.cross(-d, p01), x0 - p0) / det,
        ]
    )

    # Debug: Intersection point
    # q = p0 + s[1]*p01 + s[2]*p02
    # print("Intersection point:", q)

    return s


def intersectionRayVoxel(x0, d, z, s):
    """Ray-voxel intersection test
    Computes the intersection of the ray x(t) = x0 + t*d
    with the voxel (cube) with corner point p0 and edge length s.

    :param x0: Ray origin (vector3)
    :param d: Ray direction (vector3)
    :param z: Voxel corner point (vector3)
    :param s: Voxel edge lengths (vector3)

    :returns Does the ray intersects the voxel (boolean)
    """

    # Define the eight corner points of the voxel
    v0 = np.array(z[0], z[1], z[2])
    v1 = np.array(z[0] + s[0], z[1], z[2])
    v2 = np.array(z[0], z[1] + s[1], z[2])
    v3 = np.array(z[0] + s[0], z[1] + s[1], z[2])
    v4 = np.array(z[0], z[1], z[2] + s[2])
    v5 = np.array(z[0] + s[0], z[1], z[2] + s[2])
    v6 = np.array(z[0], z[1] + s[1], z[2] + s[2])
    v7 = np.array(z[0] + s[0], z[1] + s[1], z[2] + s[2])

    # Successively check all six sides for intersections
    s = intersectionRayPlane(x0, d, v0, v1, v2)
    if s and s[0] >= 0 and s[0] <= 1 and s[1] >= 0 and s[1] <= 1:
        return True

    s = intersectionRayPlane(x0, d, v0, v1, v4)
    if s and s[0] >= 0 and s[0] <= 1 and s[1] >= 0 and s[1] <= 1:
        return True

    s = intersectionRayPlane(x0, d, v0, v4, v2)
    if s and s[0] >= 0 and s[0] <= 1 and s[1] >= 0 and s[1] <= 1:
        return True

    s = intersectionRayPlane(x0, d, v7, v6, v5)
    if s and s[0] >= 0 and s[0] <= 1 and s[1] >= 0 and s[1] <= 1:
        return True

    s = intersectionRayPlane(x0, d, v7, v3, v6)
    if s and s[0] >= 0 and s[0] <= 1 and s[1] >= 0 and s[1] <= 1:
        return True

    s = intersectionRayPlane(x0, d, v7, v3, v5)
    if s and s[0] >= 0 and s[0] <= 1 and s[1] >= 0 and s[1] <= 1:
        return True

    return False
