# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Use this module to detect acoustic portal in a scene."""

import copy
from typing import Dict, List

import numpy as np
from trimesh import Trimesh
from trimesh.points import PointCloud
from trimesh.primitives import Box
from trimesh.proximity import ProximityQuery, signed_distance
from trimesh.ray.ray_pyembree import RayMeshIntersector
from trimesh.util import concatenate
from trimesh.voxel.ops import matrix_to_points

from mpegi.metadata.eifplus import AcousticEnvironment, Geometry, Mesh, Update, Anchor

from .transforms import transformMatrixTRS_4x4
from .trimesh import convertGeometriesToTrimesh, trimeshToEifMesh
from .utils import recenterMesh, is_building


class Portals:
    """A portal can be used to transmit audio between two spaces (currently only used for AE)"""

    def __init__(self, id: int, extent: str, sourceId: int, sst: int) -> None:
        """Construct Portal

        :param id: unique id count of the portal
        :param extent: geometry id of the portal's extent
        :param sourceId: id of the AE, from which this portal's downmix is created
        :param sst: enum for the signal generation rule (SST) of this portal
        :param eq: List[float], eq of this portal (currently empty)
        """
        self.id = id
        self.extent = extent
        self.sourceId = sourceId
        self.sst = sst
        self.eq = []


class InternalInfo:
    """A data strucutre that helps document and pass around the intermediate results"""

    def __init__(self, connectionInfo, openingInfo) -> None:
        """construct InternalInfo

        :param connectionInfo: dict, documenting which AEs (value) have an opening to each AE (key)
        :param openingInfo: dict, documenting the plane area and center position of the corresponding opening
            openingInfo's data structure corresponds to connectionInfo's. ex.
            connectionInfo = ``{'acEnv:1': ['acEnv:2']}``
            openingInfo = ``{'acEnv:1': [surface area of the opening between acEnv 1&2, center Pos of the opening]}``
        :param portalFactor: a list of lists, documents the 1. surface area and 2. absorption coefs for each AE
        """
        self.conInfo = connectionInfo
        self.openInfo = openingInfo
        self.portalFactor = []


class PortalInfo:
    """Struct that holds portal related information in the encoder"""

    def __init__(
        self,
        portalGoemetries: List[Geometry],
        allPortals: List[Portals],
        portalAEMap: Dict,
        openFactor: Dict,
        openLoc: Dict,
        openInfo: Dict,
        conInfo: Dict,
    ) -> None:
        """Construct PortalInfo

        :param portalGeometries: geometries representing the extent of portals
        :param allPortals: holds all portal struct in the scene
        :param portalAEMap: documents which portals belongs to which AE
        :param openFactor: documents the portal factor of each opening. 0 if there's no opening between AEs.
        :param openLoc: documents the location of the opening. None if there's no opening between AEs.
        """
        self.portalGeometries = portalGoemetries
        self.allPortals = allPortals
        self.portalAEMap = portalAEMap
        self.openFactor = openFactor
        self.openLoc = openLoc
        self.openInfo = openInfo
        self.conInfo = conInfo


def createPortals(
    acousticEnvs: List[AcousticEnvironment],
    geometries: List[Geometry],
    anchors: List[Anchor],
    isSimplified: bool,
    updates: List[Update],
    reverbAsSess: bool,
):
    """Detect portals

    Creates the portal geometries based on the acoustic environment in the scene

    :param acousticEnvs: a list of all acoustic environments in the scene
    :param geometries: a list of relevant geoemtrues in the scene
    :param isSimplied: whether the scene went through the geometry simplification algorithm
    :param update: a list of all updates in the scene
    :param reverbAsSess: whether the scene enables the flag reverbAsSess
    :return portalInfo: a class containing all the relevant portal information in the scene
    """

    acousEnvGeo = [env.region for env in acousticEnvs]
    acousticEnvNames = [env.id for env in acousticEnvs]
    interInfo = getConnectInfo(acousEnvGeo, acousticEnvNames, geometries, isSimplified, updates)
    anchorIds = [a.id for a in anchors]

    # get the geometry of the AE and convert to trimesh
    geoOfAE = [g for g in geometries if g.id in acousEnvGeo]
    tri_AEs = convertGeometriesToTrimesh(geoOfAE, includeTransparent=True, concatenate=False)

    portalAsTrimesh = []
    portalNames = []
    for i in range(len(tri_AEs)):
        if isinstance(geoOfAE[i], Mesh):
            portalAsTrimesh.append(
                Trimesh(
                    vertices=tri_AEs[i].vertices + 0.5 * tri_AEs[i].vertex_normals,
                    faces=[np.flip(f) for f in tri_AEs[i].faces],
                )
            )
        else:
            for idx in range(len(tri_AEs[i].vertices)):
                vn = [0.5 * x for x in tri_AEs[i].vertex_normals[idx]]
                tri_AEs[i].vertices[idx] = [x - y for x, y in zip(tri_AEs[i].vertices[idx], vn)]

            portalAsTrimesh.append(tri_AEs[i])

        portalNames.append(acousticEnvNames[acousEnvGeo.index(geoOfAE[i].id)])

        # use portal size & absorption coefficients to estimate portal factor (first half done here,
        # the other half requires the surface area of the opening between AEs, and will be done in createPortalStruct)
        dimensions = portalAsTrimesh[-1].extents
        surfArea = 0
        for a in range(len(dimensions)):
            for b in range(a + 1, len(dimensions)):
                surfArea += 2 * dimensions[a] * dimensions[b]

        rt_ave = np.mean([f.rt60 for f in acousticEnvs[i].parameters[0].frequencies])
        abCoef = np.max([0, np.min([1, 0.161 * np.prod(dimensions) / (surfArea * rt_ave)])])
        interInfo.portalFactor.append([surfArea, abCoef])

    # organize all portal related data
    allPortals, portalAEMap, mapOpeningFactor, mapOpeningLoc = createPortalStruct(
        portalNames, acousticEnvNames, interInfo, reverbAsSess
    )

    portalGeometries = [
        recenterMesh(
            trimeshToEifMesh(pTri, "portal-" + portalNames[i]),
            pTri.center_mass.tolist(),
        )
        for i, pTri in enumerate(portalAsTrimesh)
    ]

    # Copy the anchor on the original acoustic environment onto the portal
    for i in range(len(portalGeometries)):
        if geoOfAE[i].anchor in anchorIds:
            portalGeometries[i].anchor = anchors[anchorIds.index(geoOfAE[i].anchor)]

    portalInfo = PortalInfo(
        portalGeometries,
        allPortals,
        portalAEMap,
        mapOpeningFactor,
        mapOpeningLoc,
        interInfo.openInfo,
        interInfo.conInfo,
    )

    return portalInfo


def createPortalStruct(
    portalNames: List[str], AENames: List[str], interInfo: InternalInfo, reverbAsSess: bool
):
    """create the portal struct

    construct the portal dictionary which documents which portals existed in each AE

    :param portalNames: list of portal names
    :param AENames: list of names of the acoustic environments in the scene
    :param interInfo: Intermediate information for portal creation
    :param reverbAsSess: whether the scene enables the flag reverbAsSess
    :return allPortals: a list of portals created
    :return portalAEMap: a dictionary documenting which portals exist in which acoustic environment
    :return mapOpeningFactor: a dictionary documenting the coupling factor of the openings
    :return mapOpeningLoc: a dictionary documenting the center location of the openings
    """

    # The portal Factor is currently calculated as
    # (surface area of the opening / surface area of the adjacent AE )[also called openingRatio] +
    # (1 - openingRatio) * (1 - absorption coefficients) * 'nonOpenRatio'
    # Basically this nonOpenRatio decides what's the maximum ratio of sound energy that will be preserved for
    # the sound that does not propogate through the opening directly (hit the walls etc.))
    nonOpenRatio = 0.5

    allPortals = []
    portalAEMap = {}
    mapOpeningFactor = {}
    mapOpeningLoc = {}

    pCount = 0
    portalAEMap["outside"] = []
    for p in portalNames:
        curPortal = Portals(pCount, "portal-" + p, p, 0)
        allPortals.append(curPortal)
        portalAEMap["outside"].append(curPortal.id)
        pCount += 1

    for ae in AENames:
        if reverbAsSess:
            portals = allPortals
        else:
            portals = [n for n in allPortals if n.sourceId != ae]
        if portals:
            portalAEMap[ae] = []
            mapOpeningFactor[ae] = []
            mapOpeningLoc[ae] = []
        for p in portals:
            portalAEMap[ae].append(p.id)
            # write the extra info of 1. center position of an opening between AEs (if exists)
            # 2.portal factor for each portal [This is calculated as surface area of opening / absoption area]
            if ae in interInfo.conInfo and p.sourceId in interInfo.conInfo[ae]:
                curPortalData = interInfo.portalFactor[portalNames.index(p.sourceId)]
                openingRatio = (
                    interInfo.openInfo[ae][interInfo.conInfo[ae].index(p.sourceId)][0]
                    / curPortalData[0]
                )

                mapOpeningFactor[ae].append(
                    np.round(
                        openingRatio + (1 - curPortalData[1]) * nonOpenRatio * (1 - openingRatio),
                        6,
                    )
                )
                mapOpeningLoc[ae].append(
                    interInfo.openInfo[ae][interInfo.conInfo[ae].index(p.sourceId)][1]
                )
            else:
                mapOpeningFactor[ae].append(0)
                mapOpeningLoc[ae].append([None])

    return allPortals, portalAEMap, mapOpeningFactor, mapOpeningLoc


class PortalMethodStore:
    def __init__(self) -> None:
        self.proximityQueries = {}
        self.potentialPort = {}
        self.closePtoAE = {}
        self.surfacePoints = {}


def getConnectInfo(
    acousticEnvs: List[Geometry],
    acousticEnvNames: List[str],
    geometries: List[Geometry],
    isSimplified: bool,
    updates: List[Update],
):
    """Construct and return connection info.

    Identify the topological connection between acoustic environments in the scene

    :param acousticEnvs: a list of the geometry ids of all acoustic environments
    :param acousticEnvNames: a list of all acoustic environment names
    :param geometries: a list of relevant geoemtries in the scene
    :param isSimplified: whether the scene went through the geometry simplification algorithm
    :param update: a list of all updates in the scene
    :return portalInfo: a class containing all the intermediate information in this module
    """
    if len(acousticEnvs) == 1:
        return InternalInfo({}, {})

    if len(acousticEnvs) == 2:
        vPitch = 0.8
    else:
        vPitch = 0.2

    rayDirections = [
        [50, 0, 0],
        [-50, 0, 0],
        [0, 0, 50],
        [0, 0, -50],
        [50, 0, 50],
        [50, 0, -50],
        [-50, 0, 50],
        [-50, 0, -50],
        [50, 0, 25],
        [50, 0, -25],
        [-50, 0, 25],
        [-50, 0, -25],
        [25, 0, 50],
        [25, 0, -50],
        [-25, 0, 50],
        [-25, 0, -50],
    ]

    geoOfAE = []
    relevantGeo = []
    updateList = []

    # This part takes care of checking updates and excluding irrelavent geometries
    for u in updates:
        updateList.extend(u.modifications)
    uIdList = [u.id for u in updateList]

    for g in geometries:
        if g.id in uIdList or g.transform in uIdList:
            continue

        if g.id in acousticEnvs:
            geoOfAE.append(g)
            continue
        if isSimplified:
            # only building-like mesh or box is relevant if polygon reduction is enabled
            if isinstance(g, Mesh) and is_building(g):
                relevantGeo.append(g)
        else:
            relevantGeo.append(g)

    # convert eif geometries into trimeshes
    triGeo = convertGeometriesToTrimesh(relevantGeo, concatenate=False)
    tri_AEs = convertGeometriesToTrimesh(geoOfAE, includeTransparent=True, concatenate=False)

    portalStore = PortalMethodStore()

    #### This part use ray-casting to find the contacting points between a pair of AE (if there is)
    # Loop through all acoustic environment as first instance
    for i in range(len(tri_AEs)):

        # initiaize a list populated with the center of current AE as ray origin for later
        AE_center = [tri_AEs[i].center_mass for x in rayDirections]

        # Loop through other AEs as second instance
        for j in range(len(tri_AEs)):
            if i == 0:
                voxel_ae = (tri_AEs[j].voxelized(vPitch)).hollow()
                portalStore.proximityQueries[j] = ProximityQuery(tri_AEs[j])
                portalStore.surfacePoints[str(j)] = matrix_to_points(
                    voxel_ae.matrix, pitch=vPitch, origin=voxel_ae.origin
                )

            if i == j:
                continue
            else:
                curId = str(i) + "/" + str(j) if i < j else str(j) + "/" + str(i)

            # find and store the connecting-surface voxel between pairs of AE
            voxelCloseAE = portalStore.proximityQueries[i].signed_distance(
                portalStore.surfacePoints[str(j)]
            )
            smallestDist = np.min(np.abs(voxelCloseAE))
            closeAEVP = [
                portalStore.surfacePoints[str(j)][a]
                for a, x in enumerate(voxelCloseAE)
                if abs(x) == smallestDist
            ]

            if curId not in portalStore.surfacePoints:
                portalStore.surfacePoints[curId] = []
            portalStore.surfacePoints[curId].extend(closeAEVP)

            # create a ProximityQuery and Intersector for the second instance of AE
            AE_intersector = RayMeshIntersector(tri_AEs[j])

            # shoot all directions from the center of first AE and see if it hits another AE, record the hit location
            pos, ray, tri = AE_intersector.intersects_location(
                AE_center, rayDirections, multiple_hits=False
            )
            if pos.size != 0:
                # make sure that the hit point is close enough to the original AE
                distToOrigAE = portalStore.proximityQueries[i].signed_distance(pos)
                pos = [
                    p for i, p in enumerate(pos) if distToOrigAE[i] > -0.5 and distToOrigAE[i] < 0.5
                ]
                # for now just grab the closest three for proof of concept and speed
                diffDist = [sum(map(lambda x, c: (x - c) ** 2, p, AE_center[0])) for p in pos]
                smallestIdx = (
                    np.argsort(diffDist)[:3] if len(diffDist) > 3 else range(len(diffDist))
                )
                for idx in smallestIdx:
                    if curId not in portalStore.closePtoAE:
                        portalStore.closePtoAE[curId] = []
                    portalStore.closePtoAE[curId].append(pos[idx])

    ## This part detect "occluded" and "opened" portal for each pair of AEs (if contact points exist between them)
    for key in sorted(portalStore.closePtoAE, key=len):
        value = portalStore.closePtoAE[key]

        # if there is only one contacting point found for a pair of AE, skip it
        if len(value) < 2:
            continue

        potentialMesh = None
        for k in range(len(relevantGeo)):
            geoDistToHitP = signed_distance(triGeo[k], value)
            # If this geometry is close enough to any of the contact points between a pair of AEs
            # or simply the surface of a single AE
            if len(key) > 1 and any(-1 < x and 0.5 > x for x in geoDistToHitP):

                if potentialMesh == None:
                    potentialMesh = copy.deepcopy(triGeo[k])
                else:
                    potentialMesh = concatenate([potentialMesh, triGeo[k]])

        # find the opening based on the surface points + all relevant geometry (not using each occluded geo)
        if potentialMesh != None and len(key) > 1:
            portalStore.potentialPort[key] = {}
            curQuery = ProximityQuery(potentialMesh)

            # opened portal detection
            distQ = curQuery.signed_distance(portalStore.surfacePoints[key])
            newOpening = [
                portalStore.surfacePoints[key][a] for a, b in enumerate(distQ) if b < -vPitch * 0.8
            ]
            hasOcclude = len(newOpening) < 0.8 * len(distQ)

            if len(newOpening) != 0:
                portalStore.potentialPort[key]["opened"] = []
                actualOpening = PointCloud(newOpening)
                opening_bound = actualOpening.bounds
                opening_size = [
                    y - x + vPitch for (x, y) in zip(opening_bound[0], opening_bound[1])
                ]
                opening_center = [(x + y) / 2 for (x, y) in zip(opening_bound[0], opening_bound[1])]

                # the extra criteria is to exclude horizontal portal (doesn't exist for current scenes)
                if len(newOpening) != 0 and opening_bound[1][1] - opening_bound[0][1] >= vPitch:

                    portalStore.potentialPort[key]["opened"].append(
                        Box(
                            extents=opening_size,
                            transform=transformMatrixTRS_4x4(
                                opening_center, np.deg2rad([0, 0, 0]), [1, 1, 1]
                            ),
                        )
                    )

            # occluded portal detection
            if hasOcclude:
                occludedPC = PointCloud(portalStore.surfacePoints[key])
                occluded_bound = occludedPC.bounds
                occluded_size = [
                    y - x + vPitch for (x, y) in zip(occluded_bound[0], occluded_bound[1])
                ]

                occluded_center = [
                    (x + y) / 2 for (x, y) in zip(occluded_bound[0], occluded_bound[1])
                ]
                portalStore.potentialPort[key]["occluded"] = []
                portalStore.potentialPort[key]["occluded"].append(
                    Box(
                        extents=occluded_size,
                        transform=transformMatrixTRS_4x4(
                            occluded_center, np.deg2rad([0, 0, 0]), [1, 1, 1]
                        ),
                    )
                )

    opening = []
    occluded = []
    connectionInfo = {}
    openingInfo = {}

    # finalize opening and occluded portal between a pair of AEs (and document the connection between
    # all AEs and outside)
    for key, value in portalStore.potentialPort.items():
        if "opened" in value:
            if len(value["opened"]) > 1:
                combinedP = concatenate(value["opened"])
                opening.append(combinedP)
            else:
                opening.extend(value["opened"])

            nameOne = acousticEnvNames[acousticEnvs.index(geoOfAE[int(key[0])].id)]
            nameTwo = acousticEnvNames[acousticEnvs.index(geoOfAE[int(key[2])].id)]
            openExtent = opening[-1].extents
            openSurface = np.round(np.prod(openExtent[openExtent != np.min(openExtent)]), 1)

            if nameOne not in connectionInfo:
                connectionInfo[nameOne] = []
                openingInfo[nameOne] = []
            connectionInfo[nameOne].append(nameTwo)
            openingInfo[nameOne].append([])
            openingInfo[nameOne][-1].append(openSurface)
            openingInfo[nameOne][-1].append(opening[-1].center_mass)
            openingInfo[nameOne][-1].append(opening[-1])
            if nameTwo not in connectionInfo:
                connectionInfo[nameTwo] = []
                openingInfo[nameTwo] = []
            connectionInfo[nameTwo].append(nameOne)
            openingInfo[nameTwo].append([])
            openingInfo[nameTwo][-1].append(openSurface)
            openingInfo[nameTwo][-1].append(opening[-1].center_mass)
            openingInfo[nameTwo][-1].append(opening[-1])

        if "occluded" in value:
            combinedP = concatenate(value["occluded"])
            occluded.append(combinedP)

    return InternalInfo(connectionInfo, openingInfo)
