# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np


class surfaceStruct:
    """data structure for surfaces
    :var N0:                np.array(3,), normal vector
    :var d:                 float, plane distance
    :var surfaceIdx:        int, surface index
    :var faceIdx:           [int, ..., int], list of face indices
    """

    def __init__(self, N0, d, surfaceIdx, faceIdx):
        self.N0 = N0
        self.d = d
        self.surfaceIdx = surfaceIdx
        self.faceIdx = faceIdx


def same_plane(normal1, dist1, normal2, dist2, thresh=0.001):
    return abs(dist2 - dist1) < thresh and np.linalg.norm(normal2 - normal1) < thresh


def get_plane(trimesh, fIdx):
    A = np.array(trimesh.vertices[trimesh.faces[fIdx][0]])
    B = np.array(trimesh.vertices[trimesh.faces[fIdx][1]])
    C = np.array(trimesh.vertices[trimesh.faces[fIdx][2]])
    N0 = np.cross((B - A), (C - A))
    N0 = N0 / np.linalg.norm(N0)
    d = np.dot(N0, A)
    return (N0, d)


def preprocSurface(trimesh):
    """Geometry pre-processor: extracts co-planar fids
    :param trimesh:         Trimesh, geometry data
    :return:                list(surfaceStruct), surfaces with co-planar fids
    """
    surface = list()

    num_triangles = len(trimesh.faces)
    if num_triangles < 1:
        return surface

    # initialize first surface
    surfaceIndex = 0
    fIdx = 0
    N0, d = get_plane(trimesh, fIdx)
    surface.append(surfaceStruct(N0, d, len(surface), list([fIdx])))

    # extract surfaces
    for fIdx in range(1, num_triangles):
        N0, d = get_plane(trimesh, fIdx)
        found = False
        s = 0
        for s in range(len(surface)):
            found = same_plane(N0, d, surface[s].N0, surface[s].d)
            if found:
                break
        if found:
            surface[s].faceIdx.append(fIdx)
        else:
            surface.append(surfaceStruct(N0, d, len(surface), list([fIdx])))

    # FIXME: add can_occlude flags

    return surface


def mapFace2Surface(trimesh, surface):
    """generate map: triangle (face) index -> connected surface index
    :param trimesh:             Trimesh, geometry data
    :param surface:             list(surfaceStruct), surfaces with co-planar fids
    :return:                    [int, ..., int], map: triangle (face) index -> surface index
    """
    num_triangles = len(trimesh.faces)
    face2surf = [-1] * (num_triangles + 1)  # make sure that face2surf[-1] maps to -1
    for s in range(len(surface)):
        for fIdx in surface[s].faceIdx:
            face2surf[fIdx] = s
    return face2surf
