# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
#  Affine transforms and homogeneous matrices
#

import numpy as np

# --= Homogeneous transform matrices =--


def translationMatrix_4x4(offset):
    """Homogeneous rotation matrix: yaw-pitch-roll convention
    :param offset = [dx, dy, dz]
    """
    T = np.identity(4)
    T[0:3, 3] = offset
    return T


def scaleMatrix_4x4(scale):
    """Homogeneous scaling matrix
    :param scale = [sx, sy, sz]
    """
    S = np.zeros((4, 4))
    S[0, 0] = scale[0]
    S[1, 1] = scale[1]
    S[2, 2] = scale[2]
    S[3, 3] = 1
    return S


def rotationMatrixX_4x4(phi):
    """Homogeneous rotation matrix: ccw-rotation around x axis"""
    return np.array(
        [
            [1, 0, 0, 0],
            [0, np.cos(phi), -np.sin(phi), 0],
            [0, np.sin(phi), np.cos(phi), 0],
            [0, 0, 0, 1],
        ]
    )


def rotationMatrixY_4x4(phi):
    """Homogeneous rotation matrix: ccw-rotation around y axis"""
    return np.array(
        [
            [np.cos(phi), 0, np.sin(phi), 0],
            [0, 1, 0, 0],
            [-np.sin(phi), 0, np.cos(phi), 0],
            [0, 0, 0, 1],
        ]
    )


def rotationMatrixZ_4x4(phi):
    """Homogeneous rotation matrix: ccw-rotation around z axis"""
    return np.array(
        [
            [np.cos(phi), -np.sin(phi), 0, 0],
            [np.sin(phi), np.cos(phi), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ]
    )


def rotationMatrixYPR_4x4(angles):
    """Homogeneous rotation matrix: yaw-pitch-roll convention
    :param angles = [yaw, pitch, roll]
    """
    return np.matmul(
        rotationMatrixY_4x4(angles[0]),
        np.matmul(rotationMatrixX_4x4(angles[1]), rotationMatrixZ_4x4(angles[2])),
    )


def transformMatrixTRS_4x4(offset, angles, scales):
    """Homogeneous transform matrix: 1. scale, 2. rotate, 3. translate
    :param angles = [yaw, pitch, roll] in rad
    :param scales = [sx, sy, sz]
    """
    return np.matmul(
        translationMatrix_4x4(offset),
        np.matmul(
            rotationMatrixY_4x4(angles[0]),
            np.matmul(
                rotationMatrixX_4x4(angles[1]),
                np.matmul(rotationMatrixZ_4x4(angles[2]), scaleMatrix_4x4(scales)),
            ),
        ),
    )


def applyTransformMatrix4x4(T, v):
    """Apply a homogeneous 4x4 transform matrix to a 3d vector"""
    v4 = np.ones((4,))
    v4[0:3] = v
    u4 = np.matmul(T, v4)
    return u4[0:3]


def getOffsetAndAnglesFromMatrixTRS(matrix):
    """Get offset and angles from an existing TRS matrix.
    IMPORTANT: scaling needs to be [1,1,1]!

    :param matrix: a TRS matrix with scaling [1,1,1]
    :type matrix: (4,4)
    :return: a tuple of offset (in cartesian coordinates) and angles (in Euler rad)
    """

    offset = matrix[:3, 3]

    angles = np.zeros((3))
    angles[0] = np.arctan2(matrix[0, 2], matrix[2, 2])
    angles[1] = -1 * np.arcsin(matrix[1, 2])
    angles[2] = np.arctan2(matrix[1, 0], matrix[1, 1])

    return offset, angles
