# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
#  Akenine-Möller triangle-box overlap test
#  Source: http://fileadmin.cs.lth.se/cs/Personal/Tomas_Akenine-Moller/code/tribox3.txt
#  Python port by Frank Wefers (Frank.Wefers@iis.fraunhofer.de)
#

import numpy as np


def triBoxOverlap(boxcenter, boxhalfsize, triverts):
    """Triangle-box intersection test
    :param boxcenter: ndarray-3
    :param boxhalfsize: ndarray-3
    :param triverts: triple of ndarray-3
    """

    # use separating axis theorem to test overlap between triangle and box
    # need to test for overlap in these directions:
    #
    # 1) the {x,y,z}-directions (actually, since we use the AABB of the triangle
    #    we do not even need to test these)
    # 2) normal of the triangle
    # 3) crossproduct(edge from tri, {x,y,z}-directin)
    #    this gives 3x3=9 more examples

    # move everything so that the boxcenter is in (0, 0, 0)
    v0 = triverts[0] - boxcenter
    v1 = triverts[1] - boxcenter
    v2 = triverts[2] - boxcenter

    # compute triangle edges
    e0 = v1 - v0  # tri edge 0
    e1 = v2 - v1  # tri edge 1
    e2 = v0 - v2  # tri edge 2

    # Bullet 3:
    #  test the 9 examples first (this was faster)

    fe = np.abs(e0)

    # AXISTEST_X01(e0[Z], e0[Y], fez, fey)
    p0 = e0[2] * v0[1] - e0[1] * v0[2]
    p2 = e0[2] * v2[1] - e0[1] * v2[2]
    rad = fe[2] * boxhalfsize[1] + fe[1] * boxhalfsize[2]
    if min(p0, p2) > rad or max(p0, p2) < -rad:
        return False

    # AXISTEST_Y02(e0[Z], e0[X], fez, fex)
    p0 = -e0[2] * v0[0] + e0[0] * v0[2]
    p2 = -e0[2] * v2[0] + e0[0] * v2[2]
    rad = fe[2] * boxhalfsize[0] + fe[0] * boxhalfsize[2]
    if min(p0, p2) > rad or max(p0, p2) < -rad:
        return False

    # AXISTEST_Z12(e0[Y], e0[X], fey, fex)
    p1 = e0[1] * v1[0] - e0[0] * v1[1]
    p2 = e0[1] * v2[0] - e0[0] * v2[1]
    rad = fe[1] * boxhalfsize[0] + fe[0] * boxhalfsize[1]
    if min(p1, p2) > rad or max(p1, p2) < -rad:
        return False

    fe = np.abs(e1)

    # AXISTEST_X01(e1[Z], e1[Y], fez, fey)
    p0 = e1[2] * v0[1] - e1[1] * v0[2]
    p2 = e1[2] * v2[1] - e1[1] * v2[2]
    rad = fe[2] * boxhalfsize[1] + fe[1] * boxhalfsize[2]
    if min(p0, p2) > rad or max(p0, p2) < -rad:
        return False

    # AXISTEST_Y02(e1[Z], e1[X], fez, fex)
    p0 = -e1[2] * v0[0] + e1[0] * v0[2]
    p2 = -e1[2] * v2[0] + e1[0] * v2[2]
    rad = fe[2] * boxhalfsize[0] + fe[0] * boxhalfsize[2]
    if min(p0, p2) > rad or max(p0, p2) < -rad:
        return False

    # AXISTEST_Z0(e1[Y], e1[X], fey, fex)
    p0 = e1[1] * v0[0] - e1[0] * v0[1]
    p1 = e1[1] * v1[0] - e1[0] * v1[1]
    rad = fe[1] * boxhalfsize[0] + fe[0] * boxhalfsize[1]
    if min(p0, p1) > rad or max(p0, p1) < -rad:
        return False

    fe = np.abs(e2)

    # AXISTEST_X2(e2[Z], e2[Y], fez, fey)
    p0 = e2[2] * v0[1] - e2[1] * v0[2]
    p1 = e2[2] * v1[1] - e2[1] * v1[2]
    rad = fe[2] * boxhalfsize[1] + fe[1] * boxhalfsize[2]
    if min(p0, p1) > rad or max(p0, p1) < -rad:
        return False

    # AXISTEST_Y1(e2[Z], e2[X], fez, fex)
    p0 = -e2[2] * v0[0] + e2[0] * v0[2]
    p1 = -e2[2] * v1[0] + e2[0] * v1[2]
    rad = fe[2] * boxhalfsize[0] + fe[0] * boxhalfsize[2]
    if min(p0, p1) > rad or max(p0, p1) < -rad:
        return False

    # AXISTEST_Z12(e2[Y], e2[X], fey, fex)
    p1 = e2[1] * v1[0] - e2[0] * v1[1]
    p2 = e2[1] * v2[0] - e2[0] * v2[1]
    rad = fe[1] * boxhalfsize[0] + fe[0] * boxhalfsize[1]
    if min(p1, p2) > rad or max(p1, p2) < -rad:
        return False

    # Bullet 1:
    #  first test overlap in the {x,y,z}-directions
    #  find min, max of the triangle each direction, and test for overlap in
    #  that direction -- this is equivalent to testing a minimal AABB around
    #  the triangle against the AABB

    # test in X-direction
    if min(v0[0], v1[0], v2[0]) > boxhalfsize[0] or max(v0[0], v1[0], v2[0]) < -boxhalfsize[0]:
        return False

    # test in Y-direction
    if min(v0[1], v1[1], v2[1]) > boxhalfsize[1] or max(v0[1], v1[1], v2[1]) < -boxhalfsize[1]:
        return False

    # test in Z-direction
    if min(v0[2], v1[2], v2[2]) > boxhalfsize[2] or max(v0[2], v1[2], v2[2]) < -boxhalfsize[2]:
        return False

    # Bullet 2:
    #  test if the box intersects the plane of the triangle
    #  compute plane equation of triangle: normal*x+d=0
    normal = np.cross(e0, e1)
    if not planeBoxOverlap(normal, v0, boxhalfsize):
        return False

    return True  # box and triangle overlaps


def planeBoxOverlap(normal, vert, maxbox):

    vmin = np.empty((3,))
    vmax = np.empty((3,))

    for q in [0, 1, 2]:
        if normal[q] > 0:
            vmin[q] = -maxbox[q] - vert[q]
            vmax[q] = maxbox[q] - vert[q]
        else:
            vmin[q] = maxbox[q] - vert[q]
            vmax[q] = -maxbox[q] - vert[q]

    if np.dot(normal, vmin) > 0:
        return False
    if np.dot(normal, vmax) > 0:
        return True

    return False

    # Old functions, before inlining
    # # ======================== X-examples ========================
    #
    # def AXISTEST_X01(a, b, fa, fb):
    #     p0 = a * v0[1] - b * v0[2]
    #     p2 = a * v2[1] - b * v2[2]
    #
    #     rad = fa * boxhalfsize[1] + fb * boxhalfsize[2]
    #     if min(p0, p2) > rad or max(p0, p2) < -rad: return False
    #     return True
    #
    # def AXISTEST_X2(a, b, fa, fb):
    #     p0 = a * v0[1] - b * v0[2]
    #     p1 = a * v1[1] - b * v1[2]
    #
    #     rad = fa * boxhalfsize[1] + fb * boxhalfsize[2]
    #     if min(p0, p1) > rad or max(p0, p1) < -rad: return False
    #     return True
    #
    # # ======================== Y-examples =======================
    #
    # def AXISTEST_Y02(a, b, fa, fb):
    #     p0 = -a * v0[0] + b * v0[2];
    #     p2 = -a * v2[0] + b * v2[2];
    #
    #     rad = fa * boxhalfsize[0] + fb * boxhalfsize[2]
    #     if min(p0, p2) > rad or max(p0, p2) < -rad: return False
    #     return True
    #
    # def AXISTEST_Y1(a, b, fa, fb):
    #     p0 = -a * v0[0] + b * v0[2];
    #     p1 = -a * v1[0] + b * v1[2];
    #
    #     rad = fa * boxhalfsize[0] + fb * boxhalfsize[2]
    #     if min(p0, p1) > rad or max(p0, p1) < -rad: return False
    #     return True
    #
    # # ======================== Z-examples ========================
    #
    # def AXISTEST_Z12(a, b, fa, fb):
    #     p1 = a*v1[0] - b*v1[1]
    #     p2 = a*v2[0] - b*v2[1]
    #
    #     rad = fa * boxhalfsize[0] + fb * boxhalfsize[1]
    #     if min(p1, p2) > rad or max(p1, p2) < -rad: return False
    #     return True
    #
    # def AXISTEST_Z0(a, b, fa, fb):
    #     p0 = a*v0[0] - b*v0[1]
    #     p1 = a*v1[0] - b*v1[1]
    #
    #     rad = fa * boxhalfsize[0] + fb * boxhalfsize[1]
    #     if min(p0, p1) >rad or max(p0, p1) < -rad: return False
    #     return True
