# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Optional, Tuple

import numpy as np
import numpy.linalg as la
import math
import trimesh

from mpegi.geometry import transforms
from mpegi.metadata import eif, eifplus
from mpegi.metadata.eifplus import (
    AcousticMaterial,
    Box,
    Cylinder,
    Geometry,
    Mesh,
    Sphere,
    UserRegion,
)
from mpegi.metadata.eifplus.basetypes import GeometryType

# TODO: In each _convert(Box|Sphere|Cylinder|Mesh)ToTrimesh function, there's a repeated pattern
# The pattern is calling apply_transform() and _apply_metadata
# Extract this functionality into a seperate function -> DRY principle


def mergeTrimeshes(
    a: trimesh.Trimesh, b: trimesh.Trimesh, transform=None, process=True
) -> trimesh.Trimesh:
    """Merges two trimeshes together, whilst retaining material metadata

    :param a: The left hand trimesh
    :type a: trimesh.Trimesh
    :param b: The right hand trimesh
    :type b: trimesh.Trimesh
    :param transform: Apply a transform to the B trimesh before merging, defaults to None
    :type transform: (4, 4) float, optional
    :return: Merged trimesh
    :rtype: trimesh.Trimesh
    """
    if transform is not None:
        b.apply_transform(transform)

    meshes = [a, b]
    toBeMergedVertices = [m.vertices.copy() for m in meshes]
    toBeMergedFaces = [m.faces.copy() for m in meshes]

    # Append faces and vertices of meshes
    vertices, faces = trimesh.util.append_faces(toBeMergedVertices, toBeMergedFaces)

    # Take materials that already exist
    mergedMaterials = a.metadata[_Consts._MATS]

    # And merge the right side materials whilst noting the new indices
    bFaceMatIdxs = []
    for faceIndex in b.face_attributes[_Consts._MATIDXS]:
        # If material of 'B' face does not exist in 'A' mesh
        if b.metadata[_Consts._MATS][faceIndex] not in mergedMaterials:
            mergedMaterials.append(b.metadata[_Consts._MATS][faceIndex])

        matIdx = mergedMaterials.index(b.metadata[_Consts._MATS][faceIndex])
        bFaceMatIdxs.append(matIdx)

    # Must be of type np.ndarray because trimesh does fancy masking
    # Cannot simply concat two lists like newList = [1, 2] + [3, 4]
    mergedFaceMatIdxs = np.concatenate(
        (a.face_attributes[_Consts._MATIDXS], bFaceMatIdxs), axis=None
    ).astype(int)

    newFaceAttributes = {_Consts._MATIDXS: mergedFaceMatIdxs}
    newMetadata = {_Consts._MATS: mergedMaterials}

    # Create the mesh object
    return trimesh.Trimesh(
        vertices=vertices,
        faces=faces,
        face_attributes=newFaceAttributes,
        metadata=newMetadata,
        process=process,
    )


# TODO: convertGeometriesToTrimesh and convertToTrimesh have awful names. convertGeometriesToTrimesh
#       should be named something else. Something to the effect of "convertToTrimeshScene" or
#       "convertToTrimeshAndTransformPositions". Not sure what a good name is here ...


def convertGeometriesToTrimesh(
    geometry: List[Geometry],
    includeTransparent: bool = False,
    faceAngle: float = 10,
    subdivisions: int = 2,
    concatenate: bool = True,
    process: bool = True,
) -> List[trimesh.Trimesh]:
    """Converts an encoder eifplus into a Trimesh eifplus. The materials of the
    of the mesh can be found in the "metadata" dict found of a trimesh object.
    E.g. myTrimeshScene.metadata["materials"]. Then, in
    face_attributes["matIndices"], is an array where each element has an index which
    can be looked up in the metadata dict. The order of this
    face_attribute["matIndices"] is the same as the faces array. For example
    myMesh.face_attribute["matIndices"][4] refers to the same face as myMesh.faces[4]

    :param geometry: The list of geometry to convert
    :param includeTransparent: If true, include all faces without an acoustic
        material assigned to it, defaults to False
    :type includeTransparent: bool, optional
    :param faceAngle: The face angle for sampling primitives in degrees,
        defaults to 10
    :type faceAngle: float, optional
    :param subdivisions: The number of iterations of sub-dividing the icosaeder mesh, defaults to 2
    :type subdivisions: int, optional
    :param concatenate: Merge all entities into a single mesh, defaults to False
    :type concatenate: bool, optional
    :param process: Activate Trimesh mesh processing (re-ordering and cleanup)
    :type concatenate: bool, optional
    :return: A list of converted trimeshes. If concatenate is true, there'll be
        only one trimesh in the list.
    """
    # Convert all geometry to trimeshes
    convertedMeshes = convertToTrimesh(
        geometry, includeTransparent, faceAngle, subdivisions, process
    )

    if concatenate == False:
        for index, convertedMesh in enumerate(convertedMeshes):
            tform = transforms.transformMatrixTRS_4x4(
                geometry[index].position, np.deg2rad([0, 0, 0]), [1, 1, 1]
            )

            convertedMesh.apply_transform(tform)

        return convertedMeshes
    else:  # Concatenate all meshes into one meshes
        trimeshScene = trimesh.Trimesh()
        trimeshScene.metadata[_Consts._MATS] = [None]
        trimeshScene.face_attributes[_Consts._MATIDXS] = []

        for index, convertedMesh in enumerate(convertedMeshes):
            tform = transforms.transformMatrixTRS_4x4(
                geometry[index].position, np.deg2rad([0, 0, 0]), [1, 1, 1]
            )

            trimeshScene = mergeTrimeshes(trimeshScene, convertedMesh, tform, process)

        return [trimeshScene]


def convertToTrimesh(
    geometry: List[Geometry],
    includeTransparent: bool = False,
    faceAngle: float = 10,
    subdivisions: int = 2,
    process: bool = True,
) -> List[trimesh.Trimesh]:
    """Converts an eifplus Geometry into a trimesh. Geometry can either be a eifplus
    Box, Sphere, Cylinder, Mesh, or UserRegion. The trimesh will respect the size and
    orientation of the geometry, but not the position.

    :param geometry: Either one or a list of Box, Sphere, Cylinder, or Mesh to convert
    :type geometry: Union[eifplus.Box, eifplus.Sphere, eifplus.Cylinder, eifplus.Mesh]
    :param includeTransparent: If true, include all faces without an acoustic
        material assigned to it, defaults to False, defaults to False
    :type includeTransparent: bool, optional
    :param faceAngle: The face angle for sampling primitives in degrees, defaults to 10
    :type faceAngle: float, optional
    :param subdivisions: The number of iterations of sub-dividing the icosaeder mesh, defaults to 2
    :type subdivisions: int, optional
    :param process: Activate Trimesh mesh processing (re-ordering and cleanup)
    :type process: bool, optional
    :raises TypeError: If raised, an invalid type was given
    :return: The geometry represented as a trimesh
    """
    convertedGeometry: List[trimesh.Trimesh] = []
    for geo in geometry:
        if isinstance(geo, Box):
            convertedGeometry.append(_convertBoxToTrimesh(geo))
        elif isinstance(geo, Sphere):
            convertedGeometry.append(_convertSphereToTrimesh(geo, subdivisions))
        elif isinstance(geo, Cylinder):
            convertedGeometry.append(_convertCylinderToTrimesh(geo, faceAngle))
        elif isinstance(geo, Mesh):
            convertedGeometry.append(_convertMeshToTrimesh(geo, includeTransparent, process))
        elif isinstance(geo, UserRegion):
            convertedGeometry.append(_convertUserRegionToTrimesh(geo, faceAngle, subdivisions))
        else:
            # Something went terribly wrong if this exception is raised...
            raise TypeError(f"Type '{type(geometry)}' is invalid for conversion")
    return convertedGeometry


def _convertBoxToTrimesh(box: Box) -> trimesh.Trimesh:
    # trimeshBox = trimesh.creation.box([1, 1, 1])
    vertices = [
        [-0.5, 0.5, 0.5],
        [0.5, 0.5, 0.5],
        [-0.5, -0.5, 0.5],
        [0.5, -0.5, 0.5],
        [-0.5, 0.5, -0.5],
        [0.5, 0.5, -0.5],
        [-0.5, -0.5, -0.5],
        [0.5, -0.5, -0.5],
    ]
    faces = [
        [1, 0, 2],
        [1, 2, 3],
        [4, 5, 7],
        [4, 7, 6],
        [5, 4, 0],
        [5, 0, 1],
        [3, 2, 6],
        [3, 6, 7],
        [0, 4, 6],
        [0, 6, 2],
        [5, 1, 3],
        [5, 3, 7],
    ]
    trimeshBox = trimesh.Trimesh(
        vertices=vertices,
        faces=faces,
        process=False,
    )

    trimeshBox.apply_transform(_getTransform(box))

    material = box.material if hasattr(box, "material") else None
    _applyMetadata(
        trimeshBox,
        box.id,
        [material for face in trimeshBox.faces],
    )

    return trimeshBox


def _convertSphereToTrimesh(sphere: Sphere, subdivisions: int = 2) -> trimesh.Trimesh:
    # trimeshSphere = trimesh.creation.icosphere(radius=0.5, subdivisions=0)
    vertices, faces = _create_icosphere(subdivisions)

    # scale sphere to get a better polygonal approximation
    p1 = np.array(vertices[faces[0][0]])
    p2 = np.array(vertices[faces[0][1]])
    p3 = np.array(vertices[faces[0][2]])
    triangle_center = (p1 + p2 + p3) / 3
    scale = 2.0 / (1.0 + la.norm(triangle_center))
    vertices = list(scale * 0.5 * np.array(vertices))

    trimeshSphere = trimesh.Trimesh(
        vertices=vertices,
        faces=faces,
        process=False,
    )

    # radius must be identical in all dimensions
    if hasattr(sphere, "size"):
        assert sphere.size[0] == sphere.size[1]
        assert sphere.size[1] == sphere.size[2]

    trimeshSphere.apply_transform(_getTransform(sphere))

    material = sphere.material if hasattr(sphere, "material") else None
    _applyMetadata(
        trimeshSphere,
        sphere.id,
        [material for face in trimeshSphere.faces],
    )

    return trimeshSphere


def _create_icosphere(subdivisions=0) -> Tuple[list, list]:
    def _get_edge_vertex(
        v1: int,
        v2: int,
        edge_dict: dict,
        vertices: list,
    ) -> int:
        edge = (min(v1, v2), max(v1, v2))
        if edge in edge_dict:
            v = edge_dict[edge]
        else:
            v = len(vertices)
            edge_dict[edge] = v
            p1 = np.array(vertices[v1])
            p2 = np.array(vertices[v2])
            p12 = p1 + p2
            p12 = p12 * (la.norm(p1) + la.norm(p2)) / (2 * la.norm(p12))
            vertices.append(list(p12))
        return v

    phi = 1.6180339887  # golden ratio
    a = 1.0 / math.sqrt(1.0 + phi * phi)
    b = phi * a
    vertices = [
        [-a, b, 0],
        [a, b, 0],
        [-a, -b, 0],
        [a, -b, 0],
        [0, -a, b],
        [0, a, b],
        [0, -a, -b],
        [0, a, -b],
        [b, 0, -a],
        [b, 0, a],
        [-b, 0, -a],
        [-b, 0, a],
    ]

    faces = [
        [0, 11, 5],
        [0, 5, 1],
        [0, 1, 7],
        [0, 7, 10],
        [0, 10, 11],
        [1, 5, 9],
        [5, 11, 4],
        [11, 10, 2],
        [10, 7, 6],
        [7, 1, 8],
        [3, 9, 4],
        [3, 4, 2],
        [3, 2, 6],
        [3, 6, 8],
        [3, 8, 9],
        [4, 9, 5],
        [2, 4, 11],
        [6, 2, 10],
        [8, 6, 7],
        [9, 8, 1],
    ]

    for n in range(subdivisions):
        edge_dict = {}
        fine_faces = []

        for v in faces:
            v_01 = _get_edge_vertex(v[0], v[1], edge_dict, vertices)
            v_12 = _get_edge_vertex(v[1], v[2], edge_dict, vertices)
            v_20 = _get_edge_vertex(v[2], v[0], edge_dict, vertices)
            fine_faces.append([v_01, v_12, v_20])
            fine_faces.append([v[0], v_01, v_20])
            fine_faces.append([v[1], v_12, v_01])
            fine_faces.append([v[2], v_20, v_12])

        faces = fine_faces

    return vertices, faces


def _convertCylinderToTrimesh(cylinder: Cylinder, faceAngle: float = 10) -> trimesh.Trimesh:
    assert faceAngle > 0 and faceAngle <= 90, f"Face Angle of '{faceAngle}' is not between 0 and 90"

    cyl_mesh_subdivs = int(round(360.0 / faceAngle))
    vertices = list()
    faces = list()

    # scale circle to get a better polygonal approximation
    scale = 2.0 / (1.0 + math.cos(math.pi / cyl_mesh_subdivs))

    # create vertices
    for i in range(2 * cyl_mesh_subdivs):
        x = 0.5 * scale * math.cos((2 * math.pi * i) / cyl_mesh_subdivs)
        y = 0.5 if i < cyl_mesh_subdivs else -0.5
        z = 0.5 * scale * math.sin((2 * math.pi * i) / cyl_mesh_subdivs)
        vertices.append([cylinder.size[0] * x, cylinder.size[1] * y, cylinder.size[2] * z])
    vertices.append([0, 0.5 * cylinder.size[1], 0])
    vertices.append([0, -0.5 * cylinder.size[1], 0])

    # side faces with outward facing normals
    for i in range(cyl_mesh_subdivs):
        j = i - 1 if i > 0 else cyl_mesh_subdivs - 1
        faces.append([j, i, i + cyl_mesh_subdivs])
        faces.append([j, i + cyl_mesh_subdivs, j + cyl_mesh_subdivs])

    # top faces
    for i in range(cyl_mesh_subdivs):
        j = (i + 1) % cyl_mesh_subdivs
        faces.append([2 * cyl_mesh_subdivs, j, i])

    # bottom faces
    for i in range(cyl_mesh_subdivs):
        j = (i + 1) % cyl_mesh_subdivs
        faces.append([2 * cyl_mesh_subdivs + 1, i + cyl_mesh_subdivs, j + cyl_mesh_subdivs])

    materials = [cylinder.material if hasattr(cylinder, "material") else None] * len(faces)

    mesh = Mesh(
        id=cylinder.id,
        position=cylinder.position,
        orientation=cylinder.orientation,
        cspace=cylinder.cspace,
        vertices=vertices,
        faces=faces,
        materials=materials,
        transform=cylinder.transform,
        anchor=cylinder.anchor,
    )

    trimeshCylinder = _convertMeshToTrimesh(mesh, process=False)

    return trimeshCylinder


def _convertMeshToTrimesh(
    mesh: Mesh, includeTransparent: bool = False, process: bool = True
) -> trimesh.Trimesh:
    faces = []
    materials = []

    for index, face in enumerate(mesh.faces):
        if includeTransparent == False and mesh.materials[index] == None:
            continue  # Skip faces that don't have materials

        faces.append(face)
        materials.append(mesh.materials[index])

    # process=False because Trimesh silently rearranges vertices/faces through cleanup by default
    convertedMesh = trimesh.Trimesh(mesh.vertices, faces, process=process)

    convertedMesh.apply_transform(_getTransform(mesh))
    _applyMetadata(convertedMesh, mesh.id, materials)

    return convertedMesh


def _convertUserRegionToTrimesh(
    userRegion: UserRegion, faceAngle: float = 10.0, subdivisions: int = 2
) -> trimesh.Trimesh:
    if userRegion.geoType == GeometryType.Box:
        tmesh = _convertBoxToTrimesh(userRegion)
    elif userRegion.geoType == GeometryType.Sphere:
        tmesh = _convertSphereToTrimesh(userRegion, subdivisions)
    elif userRegion.geoType == GeometryType.Cylinder:
        tmesh = _convertCylinderToTrimesh(userRegion, faceAngle)
    else:
        raise Exception("UserRegion is not a Box, Sphere, or Cylinder")  # should never be reached
    return tmesh


def _getTransform(entity):
    # Rotate and scale the entity
    try:
        size = entity.size
    except AttributeError:  # mesh doesn't have size attribute
        size = [1, 1, 1]

    return transforms.transformMatrixTRS_4x4([0, 0, 0], np.deg2rad(entity.orientation), size)


def _applyMetadata(mesh: trimesh.Trimesh, id: str, materials: List[AcousticMaterial]):
    """Applies metadata to the trimesh. There are two dicts in a trimesh: metadata
    and face_attributes. In metadata, values are placed with the keys "id" and
    "materials". "id" will give the id of the entity and "materials" will give
    all of the acoustic materials that exist within the entity. face_attributes
    will have a key of "matIndices" which will be a list that corresponds to the
    faces list. Each element is an index to "materials" in metadata.

    :param mesh: The trimesh to apply the metadata to
    :type mesh: trimesh.Trimesh
    :param id: The string identifier for this entity
    :type id: str
    :param materials: A list of acoustic materials. Must correspond to the faces
        list
    :type materials: List[scene.AcousticMaterial]
    :raises ValueError: Faces and Materials list must be exactly the same
    """
    if len(mesh.faces) != len(materials):
        raise ValueError("List of faces and materials must be equal in length")

    allMatsInMesh = [None]
    faceMatIndices = []

    if len(set(materials)) == 1:  # Only one unique material
        if materials[0] != None:
            allMatsInMesh.append(materials[0])
            faceMatIndices = [allMatsInMesh.index(materials[0]) for face in mesh.faces]
        else:
            faceMatIndices = [allMatsInMesh.index(materials[0]) for face in mesh.faces]
    else:
        for index in range(len(mesh.faces)):
            faceMaterial = materials[index]

            if faceMaterial not in allMatsInMesh:
                allMatsInMesh.append(faceMaterial)

            faceMatIndices.append(allMatsInMesh.index(faceMaterial))

    mesh.metadata[_Consts._ID] = id
    mesh.metadata[_Consts._MATS] = allMatsInMesh
    mesh.face_attributes[_Consts._MATIDXS] = np.array(faceMatIndices, dtype=int)


def trimeshToEifMesh(
    mesh: trimesh.Trimesh, meshId: str, anchorId: Optional[str] = None
) -> eifplus.Mesh:
    vertices = mesh.vertices.tolist()
    faces = mesh.faces.tolist()
    materials = []
    if len(mesh.face_attributes.keys()) != 0:
        materials = [
            mesh.metadata[_Consts._MATS][matId] for matId in mesh.face_attributes[_Consts._MATIDXS]
        ]
    newMesh = eif.Mesh(meshId, vertices, faces, materials, anchorID=anchorId)
    newMesh = eifplus.EntityFactory.manufactureEntity(newMesh)
    return newMesh


def copy_trimesh(tmesh: trimesh.Trimesh) -> trimesh.Trimesh:
    """Deepcopy a trimesh and return the copy.

    This function retains all of the material and face data that we embed into
    trimesh.

    :param tmesh: The trimesh to copy.
    :return: The deepcopied trimesh.
    """
    return trimesh.Trimesh(
        vertices=tmesh.vertices,
        faces=tmesh.faces,
        metadata=tmesh.metadata,
        face_attributes=tmesh.face_attributes,
    )


class _Consts:
    """Constants to avoid hard coding the same string in multiple places.
    Only for private use. These variables were put in a class, as opposed to
    putting them in a global scope, because modules shouldn't have side affects.
    """

    _ID = "id"
    _MATS = "materials"
    _MATIDXS = "matIndices"
