# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np

from mpegi.metadata import eifplus


def isPointInBounds(x, a, b):
    return a[0] <= x[0] <= b[0] and a[1] <= x[1] <= b[1] and a[2] <= x[2] <= b[2]


def updateBoundsByPoint(bounds, x):
    """Update axis-aligned bounds by a new point"""
    if not bounds:
        return [np.array(x), np.array(x)]
    else:
        return [np.minimum(bounds[0], x), np.maximum(bounds[1], x)]


def recenterMesh(mesh, meshCenter):
    for i in range(len(mesh.vertices)):
        mesh.vertices[i] = [x - y for x, y in zip(mesh.vertices[i], meshCenter)]

    mesh.location.position = meshCenter

    return mesh


def is_building(mesh: eifplus.Mesh) -> bool:
    """test whether the mesh is a building-like strucutre, which has faces that are mostly parallel or vertical to each other
    param mesh: the mesh to be tested
    """
    # setable parameters
    # bFactor - identify as a building if the percentage of building-like structure is higher than this
    bFactor = 0.7

    likeBuilding = 0
    step = int(len(mesh.faces) / 100) + 1 if len(mesh.faces) >= 100 else 1
    # Iterate over face normals (jump in steps if there's too many faces)
    for i in range(0, len(mesh.faces), step):
        normal_i = mesh.getFaceNormal(i)
        for j in range(i + step, len(mesh.faces), step):
            # Compute the angles between each pair of face normal
            normal_j = mesh.getFaceNormal(j)
            angle = abs(np.arccos(round(np.inner(normal_i, normal_j), 2)) / np.pi * 180)
            # count the occasion if it's a parellel or vertical angle
            if angle <= 2 or angle >= 178 or (angle <= 92 and angle >= 88):
                likeBuilding += 1

    allOccasion = len(mesh.faces) / step / 2 * (len(mesh.faces) / step - 1)
    isBuilding = likeBuilding / allOccasion >= bFactor

    return isBuilding


def is_complex(mesh: eifplus.Mesh) -> bool:
    """test whether mesh qualifies as complex if its vertices per unit distance exceeds cFactor (tunable parameter)
    param mesh: the mesh to be tested
    """
    # setable parameters
    # cFactor - identify as a complex if the number of vertices per unit distance is higher than this
    cFactor = 10

    dist = 0
    step = int(len(mesh.vertices) / 100) + 1 if len(mesh.vertices) >= 100 else 1
    # Iterate over vertices (jump in steps if there's too many faces)
    for i in range(0, len(mesh.vertices), step):
        for j in range(i + step, len(mesh.vertices), step):
            # Compute the distance between pair of vertices and add up
            temp = [a - b for (a, b) in zip(mesh.vertices[i], mesh.vertices[j])]
            dist += np.sqrt(temp[0] * temp[0] + temp[1] * temp[1] + temp[2] * temp[2])

    allOccasion = len(mesh.vertices) / step / 2 * (len(mesh.vertices) / step - 1)
    # get a "average distance between vertices" for this mesh and devide the total number of vertices with it
    numOfVPerUnitD = len(mesh.vertices) / (dist / allOccasion)
    isComplex = numOfVPerUnitD >= cFactor

    return isComplex
