# This software module was originally developed by
#
# Ericsson AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Ericsson AB retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""
Read and write hdf5 files of granular database 

"""
import h5py
import numpy as np
from scipy.spatial import KDTree
from typing import List, Dict
import traceback


class GrainDBReader(object):
    def __init__(self, file):
        self.h5 = h5py.File(file, "r")
        # Cache is used for frequently accessed variables with "getVar()"
        self._cache = {}

    def __enter__(self):
        return self

    def getDim(self, dimensionName: str) -> int:
        """


        Parameters
        ----------
        dimensionName : str
            Name of the dimension. Available dimensions:

            * C: descriptor space coordinate dimension
            * I: singleton scalar value (=1),
            * M: number of grains in the database,
            * Ch: number of channels

        Returns
        -------
        int
            Size of the dimension

        """
        return self.h5[dimensionName][0]  # all dimension based datasets have a single value only

    def getAvailableVars(self) -> List[str]:
        """Returns a list of available variables

        :return: Available variables
        :rtype: List[str]
        """
        return self.h5.keys()

    def getVar(self, variableName: str, attribute: bool = False):
        """Return variables in Grain DB h5 file

        :param variableName: Available variables: DescriptorCoordinates, Grains, weights/probability
        :type variableName: str
        :param attribute: defaults to False
        :type attribute: bool, optional
        """
        try:
            var = self._cache[variableName]
        except:
            self._cache[variableName] = self.h5[variableName]
            var = self._cache[variableName]

        return var[:] if not attribute else var.attrs[attribute]

    def getMetaData(self) -> Dict:
        """Returns a dictionary of  metadata stored in the Granular Database file (we may not need it currently but can be used to store metadata regarding cross-fade for example)

        :return: Metadata of GranularDB file
        :rtype: Dict
        """
        meta = {}
        for attr, value in self.h5.attrs.items():
            meta[attr] = value
        return meta

    def getClosestGrainCoordinates(self, coordinate: List[float]) -> List[float]:
        """Returns the Grain Descriptor Space coordinates closest to the ones given

        :param coordinates: C-dimensional Descriptor space coordinates
        :type coordinates: List[float]
        :return: Closest descriptor space coordinates
        :rtype: List[float]

        """

        try:
            _, index = self._cache["KDTree"].query(coordinate)
        except:

            DescCoord = self.getVar("DescriptorCoordinates")
            self._cache["KDTree"] = KDTree(DescCoord)  # temporary solution for 1-D database
            _, index = self._cache["KDTree"].query(coordinate)

        return self.getVar("DescriptorCoordinates")[index]

    def getGrain(self, coordinates: List[float]) -> np.ndarray:
        """Returns the audio grain at a given descriptor space coordinate

        :param coordinates: DescriptorCoordinate
        :type coordinates: List[float]
        :raises ValueError: No grain found for given coordinates
        :return: grain at descriptor space coordinate
        :rtype: np.ndarray
        """
        meas = self.__findFirst(self.getVar("DescriptorCoordinates"), np.array(coordinates))
        if meas != None:

            return self.getVar("Grains")[meas]

        else:
            raise ValueError("No grain found for given coordinates ")

    def __findFirst(self, array, pattern):
        """Find first occurance of a pattern (b) in a numpy array (a)"""
        array = array.flatten()
        lb, la = len(pattern), len(array)
        lb = lb - 1
        mask = array[lb:la] == pattern[lb]
        for i in range(lb):
            mask *= array[i : la - lb + i] == pattern[i]
        return None if not np.any(mask) else np.argmax(mask) / len(pattern)

    def __exit__(self, errType, errVal, errTrace):
        if errType:
            print(errType, errVal)
            traceback.print_tb(errTrace)
        return True

    def __del__(self):
        self.h5.close()


## Write functionality:


class GrainDBWrite(object):
    apiName = "Ericsson Granular Database Writer for Python"
    apiVersion = "1.0.0"

    def __init__(self, file):
        self.h5 = h5py.File(file, "w")

    def __enter__(self):
        return self

    def getVar(self, variableName: str, attribute: bool = False):
        """Return variables in GranularDB file

        :param variableName: Available variables: DescriptorCoordinates, Grains, weights/probability
        :type variableName: str
        :param attribute: defaults to False
        :type attribute: bool, optional
        """
        var = self.h5[variableName]
        return var[:] if not attribute else var.attrs[attribute]

    def setDimWithValues(self, dimensionName, vals):
        """


        Parameters
        ----------
        dimensionName : type str
            name of the dimension
        vals : TYPE: int
            integer value for the dimension

        Returns
        -------
        sets the dimensionName with an int value which is later used to define tuples for granular database

        Example:
            C: coordinate dimensions: number of descriptors or dimension of descriptor space
            M: number of audio grains in the granular database
            I: singleton scalar value (=1)
            Ch: number of channels

        """
        v = self.h5.create_dataset(
            dimensionName, 1, int
        )  # create_variable is a netCDF function, not used for grains, only dimensions
        v[:] = vals
        return self

    def setMetaData(self, meta, parent=False):
        """
        Parameters
        ----------
        meta : DICT
            DESCRIPTION.

        parent : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        item = self.h5 if not parent else self.h5[parent]

        for attr, value in meta.items():
            item.attrs[attr] = value
        return self

    def setVar(self, variableName, values, dimensionNames, attributes=False, dtype=float):

        dimensions = (self.h5[dimensionNames[0]][0], self.h5[dimensionNames[1]][0])
        if variableName not in self.h5.keys() and variableName == "grains":
            if "DescriptorCoordinates" not in self.h5:
                raise ValueError("Descriptor coordinate dataset is not initialized.")
            # dtype = values.dtype if not dtype else dtype
            # generic dtype is float but in the case of grains, special dtype: is a double with varying length of files
            vardoub = h5py.special_dtype(vlen=np.dtype("double"))  # variable length of the grains
            if dimensions[1] > 1:
                var = self.h5.create_dataset(
                    variableName,
                    (
                        dimensions[0],
                        dimensions[1],
                    ),
                    dtype=vardoub,
                )
                # var[:] this does not work! have to set each channel separately...
                for ch_ind in range(dimensions[1]):
                    var[:, ch_ind] = values[:, ch_ind]
            else:
                var = self.h5.create_dataset(variableName, (dimensions[0],), dtype=vardoub)
                var[:] = values

        elif variableName not in self.h5.keys():
            var = self.h5.create_dataset(variableName, dimensions, dtype=dtype)
            var[:] = values

        return self

    def setGrain(self, coordinates, grain):  # specifically to be used for grains
        if "DescriptorCoordinates" not in self.h5:
            raise ValueError("Descriptor coordinate dataset is not initialized.")

        meas = self.__findFirst(self.getVar("DescriptorCoordinates"), np.array(coordinates))
        if meas:
            self.h5["grains"][meas] = grain

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_value:
            print(exc_value)  # for debug
        return True

    def __del__(self):
        self.h5.attrs["APIName"] = np.string_(GrainDBWrite.apiName)
        self.h5.attrs["APIVersion"] = np.string_(GrainDBWrite.apiVersion)

        self.h5.close()

    #
    # private functions
    #

    # find first occurence of a pattern (b) in numpy array (a)
    def __findFirst(self, array, pattern):
        array = array.flatten()
        lb, la = len(pattern), len(array)
        lb = lb - 1
        mask = array[lb:la] == pattern[lb]
        for i in range(lb):
            mask *= array[i : la - lb + i] == pattern[i]
        return None if not np.any(mask) else np.argmax(mask) / len(pattern)


# -----------------------------------testing read and write functions-----------------------------------#

## Read functions-values in the old file
# GD = GrainDBReader('granular_database_car_engine.h5')
# Descriptor1 = GD.getVar('fundamental_frequency')
# Grains = GD.getVar('grains')

# # They only work with later files
# #closestGrain = GD.getClosestGrainCoordinates([31,])
# #grain = GD.getGrain([31.7898195])

# ## Write functions:
# granularDB = GrainDBWrite('new_grain_database_car_engine.h5') #creation of granular database file
# # define the dimensionality:
# granularDB.setDimWithValues("C",1) #1 dimnesional descriptor coordinate
# granularDB.setDimWithValues("M",796) # number of grains/grain coordinate values
# granularDB.setDimWithValues("Ch",1) # number of channels in the audio
# granularDB.setDimWithValues("I",1)    #singleton if ever needed

# # data to be written:
# Freq =  Descriptor1.reshape((len(Descriptor1),1))
# Grains = Grains.reshape((len(Grains),1))

# #write descriptor coordinates
# granularDB.setVar("DescriptorCoordinates",Freq,("M","C"))
# #write grains-one grain per coordinate-second dimension refers to number of channels
# granularDB.setVar("grains",Grains,("M","Ch"))

# # Write a grain to a particular coordinate:
# granularDB.setGrain([31.7898195], Grains[0])

# ## Read this new file:
# GQ = GrainDBReader('new_grain_database_car_engine.h5')
# closestGrain = GQ.getClosestGrainCoordinates([31,])
# C = GQ.getDim("C")
