# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import trimesh
from typing import List, Union

import logging
from math import sin, cos, tan, atan, pi

from .enumclasses import *
from .arbitraryaxisrotation import *


class AngleCalculator:
    def __init__(self, trimeshScene: trimesh.Trimesh = None, visualize: bool = False):
        """
        AngleCalculator provides the interface to calculate the inner angle between two adjacent faces,
        a vertex and a face, two connected vectors

        @param trimeshScene: trimesh object instantiated by EncScene (eifplus.scene.Scene)
        """

        self.__trimesh = trimeshScene
        self.__coord = None  # todo: check if the coordinate system change is requisite
        self.__config_type = None  # AngleCalculator (in Encoder) supports TriangleToTriangle Case

        self.__v_origin = None
        self.__v_bias = None
        self.__v_origin2bias = None

        # reference face
        self.reference_face_index = None
        self.Normal = None
        self.Ui_ref = None

        # target face or target location
        self.Uj_tar = None
        self.target = None

        self.__internal_angle = None
        self.__axisRotationInstance = None

        self.visualize = visualize

    def set(
        self,
        coord,
        config_type,
        v_origin,
        v_bias,
        target,
        fid=None,
        triangleCentroid=None,
        triangleNormal=None,
    ):
        """
        set the relevant class attributes

        @param triangleNormal: surface normal of a triangle which will be used to calculate the target point angle
        @type triangleNormal: None or Vector3
        @param triangleCentroid: centroid coordinates of a triangle which will be used to calculate the target point angle
        @type triangleCentroid: None or Vector3
        @param v_origin: one of the edge vertices coordinates which will be used as the reference point to rotate the target
        @param v_bias: the other edge vertex coordinate
        @param coord: type of coordinates system, that is currently XYZ or OpenGL)
        @param config_type: type of geometric configuration
        @param fid: index of the face which will be used as reference plane to calculate the angle
        @param target: the coordinate of the target point to which the angle will be calculated
        :return: none
        """

        self.__coord = coord
        # note: TriangleToTriangle type/case is only the case we can encounter in the encoding stage
        self.__config_type = config_type

        if isinstance(v_origin, list) and isinstance(v_bias, list):
            v_origin = np.array(v_origin)
            v_bias = np.array(v_bias)

        if isinstance(v_origin, np.ndarray) and isinstance(v_bias, np.ndarray):
            if len(v_origin) == 3 and len(v_bias) == 3:
                self.__v_origin = v_origin
                self.__v_bias = v_bias
                self.__v_origin2bias = self.__v_bias - self.__v_origin
            else:
                raise TypeError("Required size of v_origin/v_bias::3")
        else:
            raise TypeError("Required type of v_origin/v_bias::np.ndarray, list")

        if (fid is not None) and isinstance(fid, (np.ndarray, np.generic)):
            fid = int(fid)

        if isinstance(self.__trimesh, trimesh.Trimesh) and isinstance(fid, int):
            if fid < len(self.__trimesh.faces):
                self.reference_face_index = fid
                self.Ui_ref = self.__trimesh.triangles_center[self.reference_face_index]
                self.Normal = self.__trimesh.face_normals[self.reference_face_index]
            else:
                raise IndexError("Requirement:: [fid < number of trimeshScene.fids]")
        elif isinstance(self.__trimesh, trimesh.Trimesh) and not isinstance(fid, int):
            raise TypeError("Required type of fid::int", fid, type(fid))
        elif not isinstance(self.__trimesh, trimesh.Trimesh):
            self.reference_face_index = -1
            self.Ui_ref = triangleCentroid
            self.Normal = triangleNormal

        if isinstance(target, np.ndarray):
            if np.size(target) == 3:
                self.target = target  # no usage so far
                self.Uj_tar = target
            else:
                raise AttributeError("Requirement:: [target_coord::an coordinate in 3D space]")
        elif isinstance(target, List):
            if np.size(target) == 3:
                self.target = np.array(target)  # no usage so far
                self.Uj_tar = np.array(target)
        else:
            raise TypeError(f"Required type of target_coord::numpy.ndarray: {target, type(target)}")

    @staticmethod
    def cal_internalAngle_onXY(vector3: List[float]) -> float:
        """
        return internal angle between two vectors on the projected 2-D plane

        @param vector3: List of 3 floats representing a location in 3-D space
        """
        if np.linalg.norm(vector3) < 1e-100:
            # logging.warning(f"np.linalg.norm(vector3) == 0")
            return 0
        else:
            n_vec3 = vector3 / np.linalg.norm(vector3)

        if abs(n_vec3[0]) < 1e-100:
            if n_vec3[1] > 0:
                return 90
            elif n_vec3[1] < 0:
                return 270
            else:
                # raise ValueError("projected point is on the Z-axis, the rotating axis")
                # logging.info(f" _____ info: {n_vec3}")
                return 0

        angle = atan(abs(n_vec3[1]) / abs(n_vec3[0])) * (180 / pi)
        if (n_vec3[0] <= 0) and (n_vec3[1] >= 0):
            return 180 - angle
        elif (n_vec3[0] <= 0) and (n_vec3[1] <= 0):
            return 180 + angle
        elif (n_vec3[0] >= 0) and (n_vec3[1] <= 0):
            return 360 - angle
        elif (n_vec3[0] >= 0) and (n_vec3[1] >= 0):
            return angle

        return 0
        # raise RuntimeError("unexpected case when calculating internal angles", n_vec3, vector3, np.linalg.norm(vector3))

    def cal_internalAngle(self) -> float:
        """returns the internal angle between reference plane and target (Uj_tar)"""

        # original scripts are created to project target/reference vectors onto the X-Y plane
        if self.__coord == CoordinateSystems.OpenGL:
            self.__v_origin = np.roll(self.__v_origin, 1)
            self.__v_bias = np.roll(self.__v_bias, 1)
            self.__v_origin2bias = np.roll(self.__v_origin2bias, 1)
            self.Ui_ref = np.roll(self.Ui_ref, 1)
            self.Uj_tar = np.roll(self.Uj_tar, 1)
            self.target = np.roll(self.target, 1)
            self.Normal = np.roll(self.Normal, 1)

            self.__coord = CoordinateSystems.XYZ
        else:
            pass

        M = self.__v_origin.tolist()
        N = self.__v_bias.tolist()
        rot_vec = np.array([M, N])

        self.__axisRotationInstance = ArbitraryAxisRotation(
            rot_vector=rot_vec, coord=CoordinateSystems.XYZ, angle=0.0
        )

        # rot_target_coord = self.__axisRotationInstance.projectTarget_onXY(self.target)
        rot_v_origin = self.__axisRotationInstance.projectTarget_onXY(np.array(self.__v_origin))
        rot_v_bias = self.__axisRotationInstance.projectTarget_onXY(np.array(self.__v_bias))
        rot_Ui_ref = self.__axisRotationInstance.projectTarget_onXY(np.array(self.Ui_ref))
        rot_Uj_tar = self.__axisRotationInstance.projectTarget_onXY(np.array(self.Uj_tar))
        rot_Normal = self.__axisRotationInstance.projectTarget_onXY(
            np.array(self.Normal + self.__v_origin)
        )

        norm = np.linalg.norm(rot_v_bias)
        if norm < 1e-100:
            return 0

        Proj_V_rot_Ui_ref = np.dot(rot_v_bias, rot_Ui_ref) / norm / norm * rot_v_bias
        Ui_ref_onXY = rot_Ui_ref - Proj_V_rot_Ui_ref
        Proj_V_rot_Uj_tar = np.dot(rot_v_bias, rot_Uj_tar) / norm / norm * rot_v_bias
        Uj_tar_onXY = rot_Uj_tar - Proj_V_rot_Uj_tar

        angleMeasurementDirection = np.cross(
            Ui_ref_onXY, rot_Normal
        )  # by the right hand rule. CW or CCW

        ref_angle = self.cal_internalAngle_onXY(Ui_ref_onXY.tolist()[0:3])
        tar_angle = self.cal_internalAngle_onXY(Uj_tar_onXY.tolist()[0:3])

        if angleMeasurementDirection[2] > 0:
            internal_angle = (tar_angle - ref_angle) % 360
        else:
            internal_angle = (ref_angle - tar_angle) % 360

        if (abs(internal_angle) < Constants.PrecisionErrorThreshold) or (
            abs(internal_angle) > 360 - Constants.PrecisionErrorThreshold
        ):
            return 0

        return internal_angle

    # note: without configurationType, it will be removed
    def getCoord(self):
        return self.__coord

    def __str__(self):
        return 'AngleCalculation "%s" {\n\t attributes: %s\n}' % (self.__name__, self.__dict__)
