# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np
from math import cos, sin, pi
from enum import Enum


class CoordinateSystems(Enum):
    # OpenGL/Cartesian coordinate system (right-handed, +y points up)
    # Usages: EIF or Graphics
    OpenGL = 0
    # Cartesian Coordinate System: (right-handed, +z points up)
    XYZ = 1


def target_verifier(func):
    """
    target_verifier guarantees that given positional target is well formulated for the function
    for instance, if a target is 3-element list -> 4-element np.ndarray
    @param func:  functions which require a positional data in 4 element list/np.ndarray
    @return: return func(self=args[0], target=target) where the target will be 4 element np.ndarray
    """

    def helper(*args):
        """This is the helper function inside the target_verifier function"""
        target = list(args)[1]
        if not isinstance(target, np.ndarray):
            if isinstance(target, list):
                target = np.array(target)
            else:
                raise TypeError("Required type of args::list or np.ndarray")

        if target.size == 3:
            target = np.append(target, 1)
        elif target.size == 4:
            pass
        else:
            raise AttributeError("3 or 4 as target.size required")

        return func(self=args[0], target=target)

    helper.__name__ = func.__name__
    return helper


class ArbitraryAxisRotation(object):
    def __init__(
        self,
        coord: CoordinateSystems = CoordinateSystems.OpenGL,
        rot_vector: np.ndarray = np.ndarray(shape=(2, 3)),
        angle: float = 0.0,
    ):
        """
        Create the matrix for transformation about an arbitrary axis in 3D space
        including for six translational and rotational motions inside
        ref: https://www.engr.uvic.ca/~mech410/lectures/4_2_RotateArbi.pdf

        @param coord: coordinate system (will be removed)
        @param rot_vector: rotating axis written as two vertex coordinates
        @param angle: rotating angle by rot_vector
        """

        self.__rot_vector = rot_vector
        self.__coord = coord

        self.angle = angle
        self.rot_vector_without_offset = np.subtract(rot_vector[1], rot_vector[0])
        self.__TTFM = np.zeros((4, 4))
        self.Individual_TFM = np.zeros((7, 4, 4))

        self.__set_matrix_identity_all()
        self.__update_ITFMs()

    @property
    def rot_vector(self):
        return self.__rot_vector

    @rot_vector.setter
    def rot_vector(self, rot_vector):
        assert isinstance(rot_vector, np.ndarray)
        raw, col = np.shape(rot_vector)
        if (raw, col) != (2, 3):
            raise TypeError("required rot_vector dimension: 2 by 3")
        self.__rot_vector = rot_vector

    @property
    def coord(self):
        return self.__coord

    @coord.setter
    def coord(self, given_coord):
        if given_coord != CoordinateSystems.XYZ:
            raise TypeError("XYZ Coordinate System Required")
        self.__coord = given_coord

    def __set_matrix_identity(self, index):
        """
        set one of the matrices as Identity according to the given "index"
        :param index: 0, 1, 2, 3, 4, 5, 6, 7
            0: Translation Matrix
            1: Rotation Matrix by X axis
            2: Rotation Matrix by Y axis
            3: Rotation Matrix by Z axis
            4: Inverse Rotation Matrix by Y axis
            5: Inverse Rotation Matrix by X axis
            6: Inverse Translation Matrix
            7: 4 by 4 Transformation Matrix
        """
        if (index > 7) or (index < 0):
            ValueError("set_matrix_identity's index: out of range")

        if index == 7:
            self.__TTFM = np.zeros((4, 4))
            for i in range(4):
                for j in range(4):
                    if i is j:
                        self.__TTFM[i, j] = 1.0
        else:
            self.Individual_TFM[index] = np.zeros((4, 4))
            for i in range(4):
                for j in range(4):
                    if i == j:
                        self.Individual_TFM[index, i, j] = 1.0

    def __set_matrix_identity_all(self):
        """set all the related matrices as Identity matrix"""
        for index in range(8):  # 0, 1, 2, [3], 4, 5, 6 + 7
            self.__set_matrix_identity(index)

    def __update_ITFMs(self):
        """set the translation and rotation matrices"""
        if self.__coord != CoordinateSystems.XYZ:
            raise TypeError("set_rotational_matrix requires CoordinateSystems.XYZ")

        self.__set_matrix_identity(0)
        self.Individual_TFM[0][0][3] = -self.__rot_vector[0][0]
        self.Individual_TFM[0][1][3] = -self.__rot_vector[0][1]
        self.Individual_TFM[0][2][3] = -self.__rot_vector[0][2]

        self.__set_matrix_identity(6 - 0)
        self.Individual_TFM[6 - 0][0][3] = self.__rot_vector[0][0]
        self.Individual_TFM[6 - 0][1][3] = self.__rot_vector[0][1]
        self.Individual_TFM[6 - 0][2][3] = self.__rot_vector[0][2]

        x_comp = self.rot_vector_without_offset[0]
        y_comp = self.rot_vector_without_offset[1]
        z_comp = self.rot_vector_without_offset[2]
        L = np.linalg.norm(self.rot_vector_without_offset)  # the norm of target vector
        V = np.sqrt(np.power(y_comp, 2) + np.power(z_comp, 2))

        rot_X_axis = True
        rot_Y_axis = True

        if V == 0:
            # raise ZeroDivisionError
            # In this case: it should skip the rotation operation by the X-axis only
            # print("self.rot_vector_without_offset, L, V", self.rot_vector_without_offset, L, V)
            rot_X_axis = False

        if L == 0:
            # raise ZeroDivisionError
            # In this case: it should skip the rotation operation by the X-axis only
            # print("self.rot_vector_without_offset, L, V", self.rot_vector_without_offset, L, V)
            rot_Y_axis = False

        if rot_X_axis:
            self.__set_matrix_identity(1)
            self.Individual_TFM[1][1][1] = z_comp / V
            self.Individual_TFM[1][1][2] = -y_comp / V
            self.Individual_TFM[1][2][1] = y_comp / V
            self.Individual_TFM[1][2][2] = z_comp / V

            self.__set_matrix_identity(6 - 1)
            self.Individual_TFM[6 - 1][1][1] = z_comp / V
            self.Individual_TFM[6 - 1][1][2] = y_comp / V
            self.Individual_TFM[6 - 1][2][1] = -y_comp / V
            self.Individual_TFM[6 - 1][2][2] = z_comp / V

        if rot_Y_axis:
            self.__set_matrix_identity(2)
            self.Individual_TFM[2][0][0] = V / L
            self.Individual_TFM[2][0][2] = -x_comp / L
            self.Individual_TFM[2][2][0] = x_comp / L
            self.Individual_TFM[2][2][2] = V / L

            self.__set_matrix_identity(6 - 2)
            self.Individual_TFM[6 - 2][0][0] = V / L
            self.Individual_TFM[6 - 2][0][2] = x_comp / L
            self.Individual_TFM[6 - 2][2][0] = -x_comp / L
            self.Individual_TFM[6 - 2][2][2] = V / L

        self.__set_matrix_identity(3)
        self.Individual_TFM[3][0][0] = cos(self.angle * pi / 180)
        self.Individual_TFM[3][0][1] = -sin(self.angle * pi / 180)
        self.Individual_TFM[3][1][0] = sin(self.angle * pi / 180)
        self.Individual_TFM[3][1][1] = cos(self.angle * pi / 180)

    def __update_TTFM(self):
        prod = np.identity(4, dtype=None)
        for index in [0, 1, 2, 3, 4, 5, 6]:
            prod = np.matmul(self.Individual_TFM[index], prod)
        self.__TTFM = prod

    def __update_TTFM_for_xyp_projection(self):
        prod = np.identity(4, dtype=None)
        for index in [0, 1, 2]:
            prod = np.matmul(self.Individual_TFM[index], prod)
        self.__TTFM = prod

    @target_verifier
    def rotate(self, target):
        self.__update_TTFM()  # T = invT*invRy*invRx*Rz*Ry*Rx*T
        return np.matmul(self.__TTFM, target)

    @target_verifier
    def projectTarget_onXY(self, target):
        self.__update_TTFM_for_xyp_projection()  # T = Rz*Ry*Rx*T
        projected_target = np.matmul(self.__TTFM, target)

        return projected_target[0:3]

    def getTTFM(self):
        self.__update_TTFM()  # TM = inv(Rt)*inv(Rx)*inv(Ry)*Rz*Ry*Rx*Rt
        return self.__TTFM

    def getTTFM_onXY(self):
        self.__update_TTFM_for_xyp_projection()  # TM = Rz*Ry*Rx*Rt
        return self.__TTFM

    def __str__(self):
        return "Arbitrary Axis Rotation Object {\nrot_vector: %s, \nTTFM: %s\n}" % (
            self.__rot_vector,
            self.__TTFM,
        )
