# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import copy
import numpy as np
from tqdm import tqdm
import multiprocessing as mp
from typing import List, Union, Dict, Tuple

from .utils import writePreprocessedData

from .enumclasses import DiffractionTypes, ConfigurationType, CoordinateSystems
from .anglecalculator import AngleCalculator
from .edgelistgenpermesh import EdgeListGenPerMesh
from .pathdictgenpermesh import PathDictGenPerMesh, Path
from .listenervoxeldictgen import ListenerVoxelDictGen
from .cellforlistenervoxels import CellForListenerVoxels
from .validpathgen import ValidPathGen
from .sourcevoxeldictgen import *


class DiffractionDataOptimizer:
    def __init__(
        self,
        staticEdgeListGenerator: EdgeListGenPerMesh,
        sourceVoxelDictGenerator: SourceVoxelDictGen,
        listenerVoxelDictGenerator: ListenerVoxelDictGen,
        staticPathDictGenerator: PathDictGenPerMesh,
        cellGen: CellForListenerVoxels,
        validPathGenerator: ValidPathGen,
        earLevelVoxelGrid,
        sceneDict: Dict,
        optimizationEnergyThreshold: float = 0.99,  # at most, 1 percent missing sound field pressure to be accepted
    ):
        """
        DiffractionDataOptimizer provides the attributes and methods for optimized diffraction data generation

        :param staticEdgeListGenerator: instantiated edgeListGen Object
        :param sourceVoxelDictGenerator: instantiated sourceVoxelDictGenerator Object
        :param listenerVoxelDictGenerator: instantiated listenerVoxelDictGenerator Object
        :param staticPathDictGenerator: instantiated staticPathDictGenerator Object
        :param cellGen: instantiated cellGen Object
        :param optimizationEnergyThreshold: energy threshold level for field analysis-based optimization
        """

        # Set instance attributes from arguments
        self.__static_pdg = staticPathDictGenerator
        self.__optimizationEnergyThreshold = optimizationEnergyThreshold  # threshold

        self.__edgeList = staticEdgeListGenerator.edgeList
        self.__pathDict = self.__static_pdg.pathDict
        self.__listenerVoxelDict = listenerVoxelDictGenerator.listenerVoxelDict
        self.__sourceVoxelDict = sourceVoxelDictGenerator.sourceVoxelDict
        self.__cell = cellGen.cell
        self.__validPathDict = validPathGenerator.validPathDict

        if sceneDict["sceneInfo"]["type"] == SceneType.AR:
            return

        # Voxel grid information to generate voxel indices from voxel center position
        self.__voxel_resolution = earLevelVoxelGrid.pitch
        self.__voxel_grid_origin_position = staticEdgeListGenerator.sceneTrimesh.bounds[0]

        self._flag_allSourcesStatic = sceneDict["_flag_allSourcesStatic"]
        self._flag_optimizePacket = (
            len(staticEdgeListGenerator.edgeList) != 0
            and len(listenerVoxelDictGenerator.listenerVoxelDict.keys()) != 0
            and sourceVoxelDictGenerator.allSourcesStatic
            and self.__getCentralizedCell()
        )

        # Todo: check if the optimization procedure is redundant when using validPathList data in static scenes
        #  if it is, able to save much encoding time
        self._flag_optimizePacket = False

        self.__angleCalNoTrimesh = AngleCalculator(
            trimeshScene=None
        )  # for parallel processing w/o visibility test

        noe_before, nop_before = self.stateReport(self.__edgeList, self.__pathDict)

        if self._flag_optimizePacket:
            self.__optimize()

        noe_after, nop_after = self.stateReport(self.__edgeList, self.__pathDict)

        logging.info(
            f"Optimization Process Report: "
            f"\nNumber of relevant/potential edges: {noe_before}"
            f"\tNumber of relevant/potential paths: {nop_before}"
            f"\nNumber of relevant/potential edges: {noe_after}"
            f"\tNumber of relevant/potential paths: {nop_after}"
        )

        # writePreprocessedData(self, "_" + sceneDict["sceneInfo"]['id'])

    @staticmethod
    def stateReport(eList, pDict):
        nofRelevantEdges = 0
        for e in eList:
            if e.relevant:
                nofRelevantEdges += 1

        nofPaths = 0
        for _, pList in pDict.items():
            nofPaths += len(pList)

        return nofRelevantEdges, nofPaths

    def __getCentralizedCell(self):
        self.__centralizedCell: Dict = {}
        for lrEdgeIDs, lvIndexListInString in self.__cell.items():
            lvIndexCenter = self.getCentralizedVoxelIndex(lvIndexListInString=lvIndexListInString)
            self.__centralizedCell[lrEdgeIDs] = (lvIndexCenter, len(lvIndexListInString))
        return True

    @staticmethod
    def getCentralizedVoxelIndex(lvIndexListInString: List):
        lvIndexCenter = np.zeros(shape=(1, 3), dtype=float)
        for lvIndexStr in lvIndexListInString:
            lvIndexList = [lvIndexPerAxis.strip() for lvIndexPerAxis in lvIndexStr.split(",")]
            lvIndexArray = np.array(lvIndexList, dtype=float)
            lvIndexCenter = np.add(lvIndexCenter, lvIndexArray)
        return np.true_divide(lvIndexCenter, len(lvIndexListInString))

    def __optimize(self):
        # Initialize propagationGainDict for sound field analysis based optimization
        # ################################################################################################################
        # # TODO: WILL BE CRYSTALLIZED FOR SOUND FIELD ANALYSIS. - IT IS ONLY GOOD TO HAVE FOR DEBUGGING PURPOSE!!
        # self.__audibleVoxelPositions = listenerVoxelDictGenerator.audibleVoxelPositions
        # self.__propagationGainDict = {}  # propagation gain distribution per source over audibleVoxelPositions
        # for src in self.__edgeListGen.sources:
        #     for voxelPosition in self.__audibleVoxelPositions:
        #         key_components = [src.id, str(voxelPosition).strip('[]')]
        #         key = '+'.join(key_components)
        #         self.__propagationGainDict[key] = 1.0
        # ################################################################################################################

        # Initialize edgeContributionGainList
        self.__edgeContributionGainList = [0 for idx in range(len(self.__edgeList))]

        # Prepare for parallel processing
        self.pool = mp.Pool(mp.cpu_count())

        # Update self.__edgeContributionGainList[eID] += attenuationGain for each source-cell pair
        self.__updateEdgeContributionGainList()

        # Set corresponding edgeContributionGain as zero and set zero-contribution edge's relevant as false
        self.__updateEdgeRelevantOfNegligibleEdges()

        # Remove less relevant data by gain analysis results
        self.__removeLessRelevantDataByGainAnalysis()

    def __updateEdgeRelevantOfNegligibleEdges(self):
        """
        Set corresponding edgeContributionGain as zero and set zero-contribution edge's relevant as false
        """

        gainList = [
            i
            for i in sorted(
                enumerate(self.__edgeContributionGainList), reverse=True, key=lambda x: x[1]
            )
        ]
        # for gainPair in gainList:
        #     print(f"gainIndex, gainValue: {gainPair[0], gainPair[1]}")

        cumGain = 0.0
        minIndexToZero = 0
        totalGain = sum(self.__edgeContributionGainList)
        for index, gain in enumerate(gainList):
            cumGain += gain[1]
            if cumGain > self.__optimizationEnergyThreshold * totalGain:
                minIndexToZero = min(index + 1, len(self.__edgeContributionGainList) - 1)
                break

        for indexToZero in range(minIndexToZero, len(gainList), 1):
            self.__edgeContributionGainList[gainList[indexToZero][0]] = 0

        # Update edge's relevant flag for negligible edges
        for edge in self.__edgeList:
            if self.__edgeContributionGainList[edge.eID] == 0:
                edge.relevant = False

    def __removeLessRelevantDataByGainAnalysis(self):
        """
        less relevant data indicate negligible edges and paths according to the given threshold
        """

        for edge in self.__edgeList:
            if not edge.relevant:
                negEdgeID = edge.eID
                # From path dictionary
                _removedPathPairDict = self.__pathDict.pop(negEdgeID, f"no Key Found: {negEdgeID}")
                self.__static_pdg.nofPaths -= len(_removedPathPairDict)

                # From listener's voxel dictionary
                for lvIndex, lvREIDs in self.__listenerVoxelDict.items():
                    try:
                        lvREIDs.remove(negEdgeID)
                    except ValueError:
                        pass

                # From sourece's voxel dictionary
                for sID, iSVD in self.__sourceVoxelDict.items():
                    for svIndex, sREIDs in iSVD.items():
                        try:
                            sREIDs.remove(negEdgeID)
                        except ValueError:
                            pass

    def strVIndexToFloatArray(self, vIndexKeyInString):
        vIndex = []
        if not isinstance(vIndexKeyInString, str):
            raise TypeError(
                f"ERROR: strVIndexToFloatArray, "
                f"\nnot isinstance(vIndexKeyInString, str) = True,"
                f"\nvIndexKeyInString: {vIndexKeyInString}"
            )

        if vIndexKeyInString == "":
            return None

        for vIndexPerAxis in vIndexKeyInString.split(","):
            vIndex.append(vIndexPerAxis.strip())
        return np.array(vIndex, dtype=float)

    def __updateEdgeContributionGainList(self):
        """calculates the propagation gain from every source to every voxel (listenerVoxels)"""

        nofSets = 0
        argsListPerSet = {nofSets: []}
        argsListForLoop = []

        for sID, iSVD in tqdm(
            self.__sourceVoxelDict.items(),
            ncols=100,
            desc=f">> PacketOptimizer::CalculateEdgeContributionList with {len(self.__sourceVoxelDict.keys())} sources",
        ):
            for svIndex, sREIDs in iSVD.items():
                svIndexArray = self.strVIndexToFloatArray(svIndex)
                sPos = np.multiply(svIndexArray, self.__voxel_resolution) + np.array(
                    self.__voxel_grid_origin_position
                )

                if sREIDs == "":
                    continue

                # centralizedCell[lrEdgeIDs] = (lvIndexCenter, len(lvIndexListInString)), Here: lrEdgeIDs = lREIDs
                for lREIDs, lvIndexTuple in self.__centralizedCell.items():
                    lvIndexArrayCenter = lvIndexTuple[0]
                    nofIncludedVoxels = lvIndexTuple[1]
                    lPos = np.multiply(lvIndexArrayCenter, self.__voxel_resolution) + np.array(
                        self.__voxel_grid_origin_position
                    )

                    # No visible paths
                    if lREIDs == "":
                        continue

                    # List[List[Path]] contains all the possible path lists starting from reachable edges from source (sPos)
                    if isinstance(lREIDs, str):
                        lREIDs = list(self.strVIndexToFloatArray(lREIDs))
                    if isinstance(sREIDs, str):
                        sREIDs = list(self.strVIndexToFloatArray(sREIDs))
                    # print(f"lREIDs, sREIDs: {lREIDs, sREIDs}")

                    sVPLs = []
                    for sREID in sREIDs:
                        # srcReachablePathList
                        sRPL: List = self.__pathDict[sREID]
                        for p in sRPL:
                            if p.edgeIDSeq[-1] in lREIDs:
                                sVPLs.append(p)

                    args = (
                        sREIDs,
                        lREIDs,
                        lPos,
                        sPos,
                        sVPLs,
                        self.__angleCalNoTrimesh,
                        nofIncludedVoxels,
                    )

                    argsListPerSet[nofSets].append(args)
                    if len(argsListPerSet[nofSets]) == 1e2:
                        edgePropagationGainListWithZeros = [0 for i in range(len(self.__edgeList))]
                        # (argsList, zero gainList)
                        argsListForLoop.append(
                            (argsListPerSet[nofSets], edgePropagationGainListWithZeros)
                        )
                        nofSets += 1
                        argsListPerSet[nofSets] = []  # new argsList for the next set

        if len(list(argsListPerSet.values())[-1]) < 1e2:
            edgePropagationGainListWithZeros = [0 for i in range(len(self.__edgeList))]
            argsListForLoop.append((argsListPerSet[nofSets], edgePropagationGainListWithZeros))

        # ################################################################################################################
        self.__returns = self.pool.map(
            self.collectSeparatePropagationGainDistribution_per_set, argsListForLoop
        )

        # Update propagation gain distribution
        for updatedEdgePropagationGainList in self.__returns:
            for edge in self.__edgeList:
                eID = edge.eID
                updateThis = updatedEdgePropagationGainList[eID]
                self.__edgeContributionGainList[eID] += updateThis
        # ###############################################################################################################

    @staticmethod
    def collectSeparatePropagationGainDistribution_per_set(argsList: Tuple):

        edgeContributionGainListProt = argsList[1]  # default
        for args in argsList[0]:
            # visible path list from a source to the listener (voxel)
            s2lRPL = DiffractionDataOptimizer.calculatePropagationGainDistribution_source_voxel(
                args
            )
            nofIncludedVoxels = args[-1]

            for cp in s2lRPL:
                # self.__propagationGain[key] = cp.getAttenuationGain()

                # edge contribution updates
                for edge in cp.edgeSeq:
                    edgeContributionGainListProt[edge.eID] += (
                        cp.getAttenuationGain() * nofIncludedVoxels
                    )
                cp.resetValidated()
        return edgeContributionGainListProt

    @staticmethod
    def calculatePropagationGainDistribution_source_voxel(args: Tuple):

        # vel_source = args[0]
        vel_listener = args[1]
        voxelPosition = args[2]
        sourcePosition = args[3]
        sRPL = args[4]  # source's reachable path list - not lists
        __angleCalNoTrimesh = args[5]
        # __angleCalNoTrimesh = AngleCalculator(None)

        vpl_s2l = []  # validated path list from a source to a listener
        for sRP in sRPL:
            edgeIDSeq: List = sRP.edgeIDSeq
            edgeSeq: List = sRP.edgeSeq
            faceIndicatorSeq: List = sRP.faceIndicatorSeq
            angleSeq: List = sRP.angleSeq

            if edgeIDSeq[-1] not in vel_listener:
                continue

            v2e = edgeSeq[-1].cp - voxelPosition
            v2s = sourcePosition - voxelPosition
            if np.inner(np.array(v2e), np.array(v2s)) < 0:
                continue

            if sRP.diffractionType == DiffractionTypes(1):
                # Simple filtering
                eFNormals = edgeSeq[0].fnormals
                fCentroids = edgeSeq[0].fcentroids

                inner1_source = np.inner(sourcePosition - fCentroids[0], eFNormals[0])
                inner2_source = np.inner(sourcePosition - fCentroids[1], eFNormals[1])
                if (inner1_source > 0) and (inner2_source > 0):
                    continue

                inner1_listener = np.inner(voxelPosition - fCentroids[0], eFNormals[0])
                inner2_listener = np.inner(voxelPosition - fCentroids[1], eFNormals[1])
                if (inner1_listener > 0) and (inner2_listener > 0):
                    continue

                ####################################################################################################
                __angleCalNoTrimesh.set(
                    coord=CoordinateSystems.OpenGL,
                    config_type=ConfigurationType.VertexToTriangle,
                    v_origin=edgeSeq[0].vtxPositions[0],
                    v_bias=edgeSeq[0].vtxPositions[1],
                    target=sourcePosition,
                    fid=None,
                    triangleCentroid=edgeSeq[0].fcentroids[faceIndicatorSeq[0]],
                    triangleNormal=edgeSeq[0].fnormals[faceIndicatorSeq[0]],
                )
                angle_f2s = __angleCalNoTrimesh.cal_internalAngle()

                __angleCalNoTrimesh.set(
                    coord=CoordinateSystems.OpenGL,
                    config_type=ConfigurationType.VertexToTriangle,
                    v_origin=edgeSeq[-1].vtxPositions[0],
                    v_bias=edgeSeq[-1].vtxPositions[1],
                    target=voxelPosition,
                    fid=None,
                    triangleCentroid=edgeSeq[-1].fcentroids[faceIndicatorSeq[-1]],
                    triangleNormal=edgeSeq[-1].fnormals[faceIndicatorSeq[-1]],
                )
                angle_f2l = __angleCalNoTrimesh.cal_internalAngle()
                ####################################################################################################

                angle_s2l = angle_f2l - (angle_f2s + 180)
                if (angle_f2l > 180) and (angle_f2s < 180) and (angle_s2l > 0):
                    # Note: do not copy vp directly which will change vp instance in the database as well
                    # -----------------------------------------------------------------------------------
                    completedPath = copy.deepcopy(sRP)
                    # -----------------------------------------------------------------------------------
                    # completedPath: will be used to calculate the propagation attenuation gain for optimization
                    completedPath.complete([angle_s2l], [sourcePosition, voxelPosition])
                    vpl_s2l.append(completedPath)

            elif sRP.diffractionType == DiffractionTypes(2):

                # Simple filtering
                eFNormals = edgeSeq[0].fnormals
                fCentroids = edgeSeq[0].fcentroids

                inner1_source = np.inner(sourcePosition - fCentroids[0], eFNormals[0])
                inner2_source = np.inner(sourcePosition - fCentroids[1], eFNormals[1])
                if (inner1_source > 0) and (inner2_source > 0):
                    continue

                eFNormals = edgeSeq[-1].fnormals
                fCentroids = edgeSeq[-1].fcentroids

                inner1_listener = np.inner(voxelPosition - fCentroids[0], eFNormals[0])
                inner2_listener = np.inner(voxelPosition - fCentroids[1], eFNormals[1])
                if (inner1_listener > 0) and (inner2_listener > 0):
                    continue

                ####################################################################################################
                __angleCalNoTrimesh.set(
                    coord=CoordinateSystems.OpenGL,
                    config_type=ConfigurationType.VertexToTriangle,
                    v_origin=edgeSeq[0].vtxPositions[0],
                    v_bias=edgeSeq[0].vtxPositions[1],
                    target=sourcePosition,
                    fid=None,
                    triangleCentroid=edgeSeq[0].fcentroids[faceIndicatorSeq[0]],
                    triangleNormal=edgeSeq[0].fnormals[faceIndicatorSeq[0]],
                )
                angle_f2s = __angleCalNoTrimesh.cal_internalAngle()

                __angleCalNoTrimesh.set(
                    coord=CoordinateSystems.OpenGL,
                    config_type=ConfigurationType.VertexToTriangle,
                    v_origin=edgeSeq[-1].vtxPositions[0],
                    v_bias=edgeSeq[-1].vtxPositions[1],
                    target=voxelPosition,
                    fid=None,
                    triangleCentroid=edgeSeq[-1].fcentroids[faceIndicatorSeq[-1]],
                    triangleNormal=edgeSeq[-1].fnormals[faceIndicatorSeq[-1]],
                )
                angle_f2l = __angleCalNoTrimesh.cal_internalAngle()
                ####################################################################################################

                angle_atSource = angleSeq[0] - angle_f2s
                if angle_atSource < 0:
                    continue

                angle_atListener = angle_f2l - angleSeq[-1]
                if angle_atListener < 0:
                    continue

                # Note: do not copy vp directly which will change vp instance in the database as well
                # -----------------------------------------------------------------------------------
                completedPath = copy.deepcopy(sRP)
                # -----------------------------------------------------------------------------------
                # completedPath: will be used to calculate the propagation attenuation gain for optimization
                completedPath.complete(
                    [angle_atSource, angle_atListener], [sourcePosition, voxelPosition]
                )
                vpl_s2l.append(completedPath)

        return vpl_s2l

    @property
    def edgeList(self):
        return self.__edgeList

    @property
    def pathDict(self):
        return self.__pathDict

    @property
    def listenerVoxelDict(self):
        return self.__listenerVoxelDict

    @property
    def sourceVoxelDict(self):
        return self.__sourceVoxelDict

    @property
    def cell(self):
        return self.__cell

    @property
    def validPathDict(self):
        return self.__validPathDict
