# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List


class Edge:
    def __init__(
        self,
        eID: int,
        vtxIDs: List[int],
        vtxPositions: List[List[float]],
        edgeLength: float,
        edgeVector: List[float],
        normalizedEdgeVector: List[float],
        fIDs: List[int],
        fvtxIDs: List[List[int]],
        fcentroids: List[List[float]],
        fnormals: List[List[float]],
        edgeAngle: float,
        cp: List[float],
        cp_vt: List[float],
        eIDAdj: int = 0,
    ):
        """
        Edge: it contains unique-edge relevant information

        @param eID: edge ID
        @param vtxIDs: vertex IDs of the end points
        @param vtxPositions: positions of two end points
        @param edgeLength: edge length
        @param edgeVector: edge vector
        @param normalizedEdgeVector: normalized edge vector
        @param fIDs: adjacent triangle's IDs (faceIDs)
        @param fvtxIDs: vertex IDs of adjacent triangles
        @param fcentroids: centroids of adjacent triangles
        @param fnormals: surface normals of adjacent triangles
        @param edgeAngle: the internal angle between adjacent triangles considering triangle's surface normals
        @param cp: center point of the edge
        @param cp_vt: center point of the edge with bias for visibility test
        @param eIDAdj: adjacent edge's ID(s)
        """

        self.eID = eID
        self.eIDAdj = eIDAdj
        self.vtxIDs = vtxIDs
        self.vtxPositions = vtxPositions
        self.edgeLength = edgeLength
        self.edgeVector = edgeVector
        self.normalizedEdgeVector = normalizedEdgeVector

        self.fIDs = fIDs
        self.fvtxIDs = fvtxIDs
        self.fcentroids = fcentroids
        self.fnormals = fnormals

        self.edgeAngle = edgeAngle

        self.cp = cp
        self.cp_vt = cp_vt

        self.visibleEdgeIDsFromThis = []
        self.visibleEdgeIDsToLimit = []

        self.rounded = False  # any good idea to utilize it for efficient RT rendering?
        self.relevant = True  # will be updated during optimization
        self.isVisibleFromStaticSources = False

    def getAttributesAll(self):
        return [
            self.eID,
            self.fIDs,
            self.fvtxIDs,
            self.vtxIDs,
            self.vtxPositions,
            self.edgeAngle,
            self.edgeVector,
            self.edgeLength,
            self.fnormals[0],
            self.fnormals[1],
            self.cp_vt,
            self.visibleEdgeIDsFromThis,
            self.rounded,
            self.relevant,
        ]

    def __str__(self):
        return "Edge::in Scene::edgeID: %s::vtxIDs: %s::FaceIDs: %s::visibleEdgeIDsFromThis: %s" % (
            self.eID,
            self.vtxIDs,
            self.fIDs,
            self.visibleEdgeIDsFromThis,
        )
