# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Union, List
from math import sqrt
import numpy as np
import trimesh

from .anglecalculator import AngleCalculator
from .enumclasses import Constants, CoordinateSystems, ConfigurationType
from .edge import Edge


class EdgeGen(object):
    def __init__(
        self,
        eID: int,
        vIDs: Union[List[int], np.ndarray],
        fIDs: Union[List[int], np.ndarray],
        trimeshScene: trimesh.Trimesh,
    ):
        """
        Edge generator includes return a new edge object based on eid, vids, fids, and trimeshScene.

        @param eid: edge id in interger starting from 0
        @param vids: edge vertex ids,2 integers
        @param fids: face ids of faces which are containing edge vertices, 2 integers
        @param trimeshScene: trimesh object instantiated by EncScene (eifplus.scene.Scene)
        """

        self.__eID = eID

        if len(vIDs) != 2:
            raise TypeError("Required length of vids::2")

        if not isinstance(trimeshScene, trimesh.Trimesh):
            raise TypeError(
                f"Required type of trimeshScene: trimesh.Trimesh, type: {type(trimesh)}"
            )
        self.__trimesh = trimeshScene

        self.__vtxIDs = []
        self.__vtxPositions = []
        for vID in vIDs:
            if vID >= len(self.__trimesh.vertices):
                raise ValueError(
                    f"Vertex ID is out of bound: vid={vID}, NofVertices={len(self.__trimesh.vertices)}"
                )
            self.__vtxIDs.append(vID)
            self.__vtxPositions.append(self.__trimesh.vertices[vID])

        self.__edgeVector = self.__vtxPositions[1] - self.__vtxPositions[0]
        self.__edgeLength = np.linalg.norm(np.array(self.__edgeVector))
        self.__normalizedEdgeVector = (
            np.array(self.__edgeVector) / np.linalg.norm(self.__edgeVector)
        ).tolist()

        if len(fIDs) != 2:
            raise TypeError("Required length of fids::2")

        self.__fIDs = []
        self.__fvtxIDs = []
        self.__fcentroids = []
        self.__fnormals = []

        for fID in fIDs:
            if fID >= len(self.__trimesh.faces):
                raise ValueError("Face ID is out of bound")
            self.__fIDs.append(fID)
            self.__fvtxIDs.append(self.__trimesh.faces[fID])
            self.__fcentroids.append(self.__trimesh.triangles_center[fID])
            self.__fnormals.append(self.__trimesh.face_normals[fID])  # np.float64

        self.__angleCal = AngleCalculator(self.__trimesh)
        self.__edgeAngle = self.updateEdgeAngle()

        self.__cp = (self.__vtxPositions[0] + self.__vtxPositions[1]) / 2
        self.__cp_vt = self.getCenterPointForVisibilityTest()

        self.__visibleEdgeIDsFromThis = []
        self.__connectedEdgeIDsToThis = []

        self.__rounded = False  # what if this is quite __rounded? idea for diffrent approaches?
        self.__isLessRelevant = False  # it indicates how relevant this edge would be to the scene.

    def getEdge(self):
        return Edge(
            self.__eID,
            self.__vtxIDs,
            self.__vtxPositions,
            self.__edgeLength,
            self.__edgeVector,
            self.__normalizedEdgeVector,
            self.__fIDs,
            self.__fvtxIDs,
            self.__fcentroids,
            self.__fnormals,
            self.__edgeAngle,
            self.__cp,
            self.__cp_vt,
        )

    def updateEdgeAngle(self) -> float:
        target = self.__trimesh.triangles_center[self.__fIDs[1]]
        self.__angleCal.set(
            coord=CoordinateSystems.OpenGL,
            config_type=ConfigurationType.VertexToTriangle,
            v_origin=self.__vtxPositions[0],
            v_bias=self.__vtxPositions[1],
            fid=self.__fIDs[0],
            target=target,
        )
        return self.__angleCal.cal_internalAngle()

    def getCenterPointForVisibilityTest(self):
        forward_projection = Constants.PrecisionErrorThreshold * (
            (self.__fnormals[0] + self.__fnormals[1]) / sqrt(2)
        )
        return self.__cp + forward_projection

    def __str__(self):
        return "EdgeGen::in Scene::edgeID: %s::vtxIDs: %s::fIDs: %s::visibleEdgeIDsFromThis: %s" % (
            self.__eID,
            self.__vtxIDs,
            self.__fIDs,
            self.__visibleEdgeIDsFromThis,
        )
