# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import json
import trimesh
import logging
import numpy as np
from tqdm import tqdm
from itertools import permutations, combinations
from typing import List, Dict, AnyStr, Tuple, Union, Optional

from mpegi.voxel.voxelizer_diffr import *
from mpegi.metadata.eifplus import Scene as eifScene
from mpegi.metadata.eifplus import SceneMetadata
from mpegi.metadata.eifplus.scene import SceneType

from .edge import *
from .edgegen import *
from .enumclasses import Constants
from .anglecalculator import *
from .utils import NumpyVariableToPythonVariable, shortestDistance


class EdgeListGenPerMesh:
    def __init__(
        self,
        staticSceneTrimeshName: AnyStr,
        sceneTrimeshName: AnyStr,
        sceneTrimesh: trimesh.Trimesh,
        sceneDict: Dict,
        voxelEdgeLength_DIFFR: Union[List[float], float, np.ndarray],
        minEdgeLength: float = 0.0,
    ):
        """EdgeListGenPerMesh provides the attributes and methods to generate the list of edges around the given mesh
        and the edges in the list would be either static or dynamic in nature

        @param staticSceneTrimeshName: the name of sceneTrimesh_DIFFR
        @param sceneTrimeshName: the name of the given trimesh object
        @param sceneTrimesh: the given trimesh object
        @param sceneDict: the scene dictionary enriched and instantiated from artifacts
        @param voxelEdgeLength_DIFFR: voxel edge length of voxelGrid_DIFFR
        """

        # Set default attributes from arguments
        self.__sceneTrimeshName = sceneTrimeshName
        self.__sceneTrimesh = sceneTrimesh
        self.__voxel_resolution_DIFFR = voxelEdgeLength_DIFFR
        self.__acousticEnvBoundsDict_DIFFR = sceneDict["acousticEnvBoundsDict"]
        self.minEdgeLength = minEdgeLength
        self.__sceneDict = sceneDict

        # Define relevant attributes
        self.__angleCal = AngleCalculator(self.__sceneTrimesh)
        self._isARScene = sceneDict["_flag_isARScene"]
        self._sceneId = sceneDict["sceneInfo"]["id"]
        self.__sources = sceneDict["objectSources"]

        # Attributes for edge candidate storage
        self.__edgeList = []

        # Set flags
        self.__flag_isStaticMesh = self.__sceneTrimeshName == staticSceneTrimeshName
        self.__flag_isARScene = sceneDict["_flag_isARScene"]
        self.__flag_allSourcesStatic = sceneDict["_flag_allSourcesStatic"]
        self.__flag_doubleAudibleRange = sceneDict["_flag_doubleAudibleRange"]

        self.__flag_edgeListUpdated = False
        self.__flag_eCpsUpdated = False

        self.__updateEdgeList()
        self.__updateReachableEdgeIDsPerEdge()

    def getVisibleEdgeIDsFrom(
        self, p: Union[List[float], np.ndarray], backface_culling: bool = False
    ) -> List[int]:
        """returns visible edge ids from a given location
        @param p: position where we want to check which edges are visible
        @return: list of visible edge ids from the given position

        For sources, backface culling required
        https://trimsh.org/trimesh.ray.ray_pyembree.html
        """

        if not self._flag_edgeListUpdated:
            raise RuntimeError(">> not self._flag_edgeListUpdated")

        if not self.__flag_isStaticMesh:
            raise RuntimeError(">> ERROR: Non-static mesh-generator not allowed to be called")

        if not self.__flag_eCpsUpdated:
            self.__eCps: List[List[float]] = [edge.cp_vt for edge in self.__edgeList]
            self.__flag_eCpsUpdated = True

        if len(self.__edgeList) == 0:
            return []

        vec_tp = np.ones((len(self.__eCps), 1)) * p
        vec_t2e = self.__eCps - vec_tp
        vec_t2e_distance = np.linalg.norm(vec_t2e, axis=1)
        (
            x_locations,
            idx_indicatingEdgeIDs,
            idx_xfaceIDs,
        ) = self.__sceneTrimesh.ray.intersects_location(
            ray_origins=vec_tp, ray_directions=vec_t2e, multiple_hits=False
        )

        peIDsFromThis = [e.eID for e in self.edgeList if e.eID not in idx_indicatingEdgeIDs]
        veIDsFromThis = []
        vec_t2x = x_locations - np.ones((len(x_locations), 1)) * p
        vec_t2x_distance = np.linalg.norm(vec_t2x, axis=1)
        for eID, t2x_distance in zip(idx_indicatingEdgeIDs, vec_t2x_distance):
            # condition 1: out of audible range
            if vec_t2e_distance[eID] > Constants.MaxDistance_FromTheFirstEdgeToTheLastEdge_InPath:
                continue

            # condition 2: no obstacles
            if t2x_distance > vec_t2e_distance[eID] + Constants.PrecisionErrorThreshold:
                veIDsFromThis.append(eID)

        for eID in veIDsFromThis:
            e = self.__edgeList[eID]

            if not e.relevant:
                continue

            # Simple filtering using dark zone analysis
            inner1_source = np.inner(p - e.cp, e.fnormals[0]) / np.linalg.norm(p - e.cp)
            inner2_source = np.inner(p - e.cp, e.fnormals[1]) / np.linalg.norm(p - e.cp)
            if inner1_source * inner2_source > 0:
                # for edges visible from the given position, but not in the dark zone of the edge
                continue

            # when the target is in one of the dark zones around the edge for edges visible from the given position
            peIDsFromThis.append(int(eID))

        if backface_culling:
            if len(peIDsFromThis) != 0:
                return peIDsFromThis
        else:
            return peIDsFromThis

        logging.info(f"********** back-face culling test: required")
        # To handle back-face culling for a source inside a enclosed space
        (
            x_locations,
            idx_indicatingEdgeIDs,
            idx_xfaceIDs,
        ) = self.__sceneTrimesh.ray.intersects_location(
            ray_origins=vec_tp, ray_directions=vec_t2e, multiple_hits=True
        )
        peIDsFromThis = [e.eID for e in self.edgeList if e.eID not in idx_indicatingEdgeIDs]
        xDict = {}
        for xIndex, eID in enumerate(idx_indicatingEdgeIDs):
            # Two audibility checks
            # When a source is outside audible range (no test required before)
            if vec_t2e_distance[eID] > Constants.MaxDistance_FromSource:
                continue

            # When an intersection point is outside audible range
            p2xDistance = np.linalg.norm(np.subtract(p, x_locations[xIndex]))
            if p2xDistance > Constants.MaxDistance_FromSource:
                continue

            # Simple dark zone analysis
            e: Edge = self.__edgeList[eID]
            inner1_source = np.inner(p - e.cp, e.fnormals[0]) / np.linalg.norm(p - e.cp)
            inner2_source = np.inner(p - e.cp, e.fnormals[1]) / np.linalg.norm(p - e.cp)
            if inner1_source * inner2_source > 0:
                continue

            # When a shoot ray intersects with other occluders before reaching the target edge
            isxFaceCulled = False
            if p2xDistance < vec_t2e_distance[eID] - Constants.PrecisionErrorThreshold:
                # Exceptional case: back-face culling
                fn = self.__sceneTrimesh.face_normals[idx_xfaceIDs[xIndex]]
                if np.inner(np.subtract(x_locations[xIndex], p), fn) > 0:
                    isxFaceCulled = True

            if eID not in peIDsFromThis:
                peIDsFromThis.append(int(eID))

            try:
                xDict[eID]
            except KeyError:
                xDict[eID] = []
            finally:
                xDict[eID].append([p2xDistance, vec_t2e_distance[eID], isxFaceCulled])

        for eID, xList in xDict.items():
            bList = [bCulled for _a, _b, bCulled in xList]
            try:
                bList.index(False)
            except ValueError:
                continue

            if (len(bList) - 1) != bList.index(False):
                peIDsFromThis.remove(eID)
                continue

        return peIDsFromThis

    def __updateEdgeList(self):
        """validate collected edges for diffraction and creates a new edge object to update self.__edgeList"""

        # Collection of edge candidates
        edges_unique = self.__sceneTrimesh.edges_unique.tolist()
        self.__eVTX_IDs_faceAdj = self.__sceneTrimesh.face_adjacency_edges.tolist()
        self.__fIDs_faceAdj = self.__sceneTrimesh.face_adjacency.tolist()

        if len(self.__eVTX_IDs_faceAdj) != len(self.__fIDs_faceAdj):
            raise RuntimeError(">> ERROR: NUMBERS OF FACES AND EDGES ARE DIFFERENT")

        if len(edges_unique) != len(self.__eVTX_IDs_faceAdj):
            # Note: some of the edges were deleted because of duplicated triangles/faces around it

            for edge_unique in edges_unique:
                if (edge_unique in self.__eVTX_IDs_faceAdj) or (
                    edge_unique[::-1] in self.__eVTX_IDs_faceAdj
                ):  # reverse
                    continue

                # ---------------------------------------------------------
                fIDs_shared = []
                for fID in self.__sceneTrimesh.vertex_faces[edge_unique[0]]:
                    if fID == -1:
                        break

                    if fID in self.__sceneTrimesh.vertex_faces[edge_unique[1]]:
                        fIDs_shared.append(fID)
                # ---------------------------------------------------------
                fIDIndices = [0, 1, 2, 3]
                if len(fIDs_shared) == 4:
                    normals = np.array(
                        [self.__sceneTrimesh.face_normals[fIDs_shared[i]] for i in range(4)]
                    )
                    for normal in normals:
                        normal /= np.linalg.norm(normal)

                    # Find the duplicated face
                    for i, j in combinations(fIDIndices, 2):
                        if (
                            abs(np.inner(normals[i], normals[j]))
                            > 1 - Constants.PrecisionErrorThreshold
                        ):
                            fIDIndices.pop(j)
                            fIDIndices.pop(i)  # in the revered order

                            fIDs_shared.pop(j)
                            fIDs_shared.pop(i)
                            break
                else:
                    continue

                if len(fIDs_shared) == 2:
                    self.__eVTX_IDs_faceAdj.append(edge_unique)
                    self.__fIDs_faceAdj.append(fIDs_shared)
                # ---------------------------------------------------------

        args = []
        for i, eVTX_IDs in tqdm(
            enumerate(self.__eVTX_IDs_faceAdj),
            desc=f">> Edge-Sample-Collection [{self.__sceneTrimesh}]",
        ):

            fIDs = self.__fIDs_faceAdj[i]
            eVTXs = self.__sceneTrimesh.vertices[eVTX_IDs]
            eCp = np.divide(np.add(eVTXs[0], eVTXs[1]), 2)

            if self.__sceneDict["_flag_isARScene"]:
                args.append((eVTX_IDs, fIDs, i))
                continue

            # Condition 1: minimum length
            if np.linalg.norm(np.subtract(eVTXs[0], eVTXs[1]), axis=0) < self.minEdgeLength:
                continue

            # condition 2: coplanar
            if self.__isCoplanar(fIDs):
                continue

            # condition 3: distance-culling
            if self.__flag_allSourcesStatic:
                if not self._isAudible(self.__sceneTrimesh.vertices[eVTX_IDs]):
                    continue

            # condition 4: boundary check for static mesh
            if self.__flag_isStaticMesh:
                self.__sceneDict["sceneTrimesh_DIFFR"]: trimesh.Trimesh
                nofeVTXsOnBoundary = 0
                for bound in self.__sceneDict["sceneTrimesh_DIFFR"].bounds:
                    for eVTX in eVTXs:
                        if any(np.abs(np.subtract(bound, eVTX)) < 0.5):
                            nofeVTXsOnBoundary = nofeVTXsOnBoundary + 1

                    cVTX = np.subtract(eVTXs[0], eVTXs[1])
                    if any(np.abs(np.subtract(bound, cVTX)) < 0.5):
                        nofeVTXsOnBoundary = nofeVTXsOnBoundary + 1

                if nofeVTXsOnBoundary > 2:
                    continue

            # condition 5: inside/reachable test in/from AEs or URRs
            margin = [Constants.MaxDistance_FromSource / 16 for i in range(3)]

            samples = []
            for alpha in range(10):
                sample = np.add(eVTXs[0] * (alpha / 10), eVTXs[1] * (1 - alpha / 10))
                samples.append(list(sample))

            reachableFromAEs = False
            if self.__sceneDict["_flag_acEnvExists"]:
                for trimesh_AE in self.__sceneDict["trimeshes_AcousticEnvironments"]:
                    trimesh_bounds_wMargin = []
                    for i, AEs_bound in enumerate(trimesh_AE.bounds):
                        if i == 0:
                            trimesh_bounds_wMargin.append(np.subtract(AEs_bound, margin))
                        else:
                            trimesh_bounds_wMargin.append(np.add(AEs_bound, margin))

                    for sample in samples:
                        reachableFromAEs |= VoxelizerDIFFR.withinEnv(
                            sample, trimesh_bounds_wMargin, ndim=3
                        )

            reachableFromURRs = False
            for trimesh_URR in self.__sceneDict["trimeshes_URRs"]:
                trimesh_bounds_wMargin = []
                for i, URR_bound in enumerate(trimesh_URR.bounds):
                    if i == 0:
                        trimesh_bounds_wMargin.append(np.subtract(URR_bound, margin))
                    else:
                        trimesh_bounds_wMargin.append(np.add(URR_bound, margin))

                for sample in samples:
                    reachableFromURRs |= VoxelizerDIFFR.withinEnv(
                        sample, trimesh_bounds_wMargin, ndim=3
                    )

            if not (reachableFromAEs or reachableFromURRs):
                continue

            args.append((eVTX_IDs, fIDs, i))

        for arg in tqdm(
            args,
            desc=f">> Edge-List-Generation [SampleN: {len(args)}]",
        ):
            self.__createEdge(arg)

        self._resetEdgeIDs()

    def __updateReachableEdgeIDsPerEdge(self, projMargin: float = 1.0) -> None:
        """update "visibleEdgeIDsFromThis" which will be important when generating higher-order (multiple-point) paths.
        @param projMargin: projection margins when allowing marginal spaces for edge-to-edge propagation
        """

        # the list of edge's center points for visibility test
        eCps = [e.cp_vt for e in self.__edgeList]
        nofReEdgeIDsOnAvg = 0
        for e in tqdm(
            self.__edgeList,
            desc=f">> VisibleEdgeList-Update-Per-Edge [N: {len(self.__edgeList)}]",
        ):
            if e.edgeLength < 0.314:
                e.visibleEdgeIDsFromThis = []
                continue

            # Condition 1: secure the edge-to-edge visibility
            reCps = np.ones((len(self.__edgeList), 1)) * e.cp_vt
            re2eVs = eCps - reCps
            re2eDs = np.linalg.norm(re2eVs, axis=1)

            xLocations, eIndexes, xtriIndexes = self.__sceneTrimesh.ray.intersects_location(
                ray_origins=reCps, ray_directions=re2eVs, multiple_hits=False
            )

            vaEdgeIDsFromThis = [idx for idx in range(len(self.__edgeList)) if idx not in eIndexes]
            re2itxVs = xLocations - np.ones((len(xLocations), 1)) * e.cp_vt
            re2itxDs = np.linalg.norm(re2itxVs, axis=1)
            for eid, dis in zip(eIndexes, re2itxDs):
                if re2eDs[eid] > Constants.MaxDistance_FromTheFirstEdgeToTheLastEdge_InPath:
                    continue

                if dis > re2eDs[eid] - Constants.PrecisionErrorThreshold:
                    vaEdgeIDsFromThis.append(eid)
            e.visibleEdgeIDsToLimit = vaEdgeIDsFromThis

            # Simple filtering by dark zone analysis:
            # Different geometrical analysis will be used to limit the odd/unrealistic/ignorable propagations
            # There might be more in the near future for tuning the paths - perceptually
            # But, for now let's make them as simple as possible.
            reachableEIDsFromThis = []
            for veid in vaEdgeIDsFromThis:
                # Condition 2: reachability by diffraction
                ve = self.__edgeList[veid]
                eFNormals = ve.fnormals

                tarPos = e.cp_vt  # A to B
                inner1_source = np.inner(tarPos - ve.cp, eFNormals[0]) / np.linalg.norm(
                    tarPos - ve.cp
                )
                inner2_source = np.inner(tarPos - ve.cp, eFNormals[1]) / np.linalg.norm(
                    tarPos - ve.cp
                )
                if inner1_source * inner2_source > 0:
                    continue

                eFNormals = e.fnormals
                tarPos = ve.cp_vt  # B to A
                inner1_source1 = np.inner(tarPos - e.cp, eFNormals[0]) / np.linalg.norm(
                    tarPos - e.cp
                )
                inner2_source1 = np.inner(tarPos - e.cp, eFNormals[1]) / np.linalg.norm(
                    tarPos - e.cp
                )
                if inner1_source1 * inner2_source1 >= 0:
                    continue

                # Condition 3: Limitation of oblique propagation up to 45 degree
                # oblique diffraction can cause distorted source location - precomputed angle sequence needs to be modified
                # Think about the fluid streamlines when a object submerged in water moves;
                #   - Which streamline is reachable from a wave source to a receiver?
                #   - Here, need to approximate the streamlines as a ray - a straight line for practical reason
                n_ev1 = np.array(e.edgeVector) / np.linalg.norm(np.array(e.edgeVector))
                n_ev2 = np.array(ve.edgeVector) / np.linalg.norm(np.array(ve.edgeVector))
                if abs(np.inner(n_ev1, n_ev2)) < 0.707:  # 45 deg
                    continue

                # Condition 4: Limitation of propagation when
                # Propagation between edges is allowed only when two edges are visible and
                # overlapped along the shared rotation axis - need of pictorial explanation?
                # marginal oblique propagation along a diffraction sequence by distance along the edge axis
                projs = [np.inner(n_ev1, ve.vtxPositions[i] - e.cp) for i in range(2)]
                if not np.any([np.linalg.norm(projs) < (e.edgeLength / 2 + projMargin)]):
                    if not (
                        projs[0] * projs[1] < 0
                        and np.all([np.linalg.norm(projs) > (e.edgeLength / 2 + projMargin)])
                    ):
                        continue

                # # only check if the center position of the next edge is out of the range (for simplicity)
                # if abs(np.inner(n_ev1, ve.cp - e.vtxPositions[0])) > (e.edgeLength / 2 + 1):
                #     continue

                reachableEIDsFromThis.append(veid)

            e.visibleEdgeIDsFromThis = reachableEIDsFromThis
            nofReEdgeIDsOnAvg += len(e.visibleEdgeIDsFromThis)

        if len(self.edgeList) != 0:
            self.nofReEdgeIDsOnAvg = nofReEdgeIDsOnAvg / len(self.__edgeList)
        else:
            self.nofReEdgeIDsOnAvg = 0

        self._flag_edgeListUpdated = True

    def _resetEdgeIDs(self) -> None:
        """
        reset the edge id as the order in the edgeList
        """

        for newID, e in enumerate(self.__edgeList):
            try:
                e.eID = newID
            except AttributeError:
                logging.error(f" __ error: the given edgeList includes non-Edge type element")

    def _isAudible(self, ls: List[List[float]]) -> bool:
        """
        check if the given point (cp_vt) is in the audible range from sources
        @return: True when the given point is in the audible range
        """

        ls1, ls2 = ls
        for source in self.__sources:
            if source.cspace == "user":
                logging.error(f" __ error: source.cspace == 'user': check this out")

            if self.__flag_doubleAudibleRange:
                if (
                    shortestDistance(source.position, ls1, ls2)
                    < Constants.MaxDistance_FromSource_noAE
                ):
                    return True
            else:
                if shortestDistance(source.position, ls1, ls2) < Constants.MaxDistance_FromSource:
                    return True

        return False

    def __createEdge(self, arg: Tuple) -> None:
        """
        create edge with given parameters and pass it to the edgeList after validation passes
        @rtype: None
        @type arg: tuple of List[int], List[int], int
        """
        eVtxIDs, fIDs, trimeshEdgeIndex = arg
        fvtxIDs = []
        for fID in fIDs:
            fvtxIDs.append(self.__sceneTrimesh.faces[fID])

        eGen = EdgeGen(
            eID=trimeshEdgeIndex, vIDs=eVtxIDs, fIDs=fIDs, trimeshScene=self.__sceneTrimesh
        )
        edgeObj = eGen.getEdge()

        if edgeObj.edgeAngle > Constants.MinEdgeAngle:
            self.__edgeList.append(edgeObj)

    def __isCoplanar(self, fids: List[int]) -> bool:
        normals = np.array(self.__sceneTrimesh.face_normals[fids])
        for n in normals:
            n /= np.linalg.norm(n)
        return np.inner(normals[0], normals[1]) < -0.9848  # cos 10 <------.------>

    @staticmethod
    def writeEdgeList(data: Dict, edge: Edge):
        """
        @param data: Dictionary with the key "edgeList"
        @param edge: Edge object whose information will be written in Json
        @return:
        """

        if not isinstance(data, Dict):
            raise TypeError("Required type of data::dict")

        if "edgeList" not in data.keys():
            raise AttributeError("Required Attribute in data dictionary::edgeList")

        if isinstance(edge, Edge):
            entityNames = [
                "eid",
                "fids",
                "fvids",
                "vids",
                "vtxPositions",
                "edgeAngle",
                "edgeVector",
                "edgeLength",
                "normal_adjFaceL",
                "normal_adjFaceR",
                "cp_vt",
                "nextPotentialEdgeIndices",
                "flag_rounded",
                "flag_isValid",
            ]

            edgeEntities = edge.getAttributesAll()

            entityDict = {}
            for entityIndex in range(len(edgeEntities)):
                edgeEntities[entityIndex] = NumpyVariableToPythonVariable(edgeEntities[entityIndex])
                entityDict[entityNames[entityIndex]] = edgeEntities[entityIndex]
            data["edgeList"].append(entityDict)

    def writeJSONFile(self, jsonFileNamePrefix: AnyStr = "json_EdgeListForDiffraction_"):
        """write a list of edges found in the given scene"""

        encoderOutput = {"edgeList": []}
        for edge in self.__edgeList:
            EdgeListGenPerMesh.writeEdgeList(data=encoderOutput, edge=edge)

        with open(
            jsonFileNamePrefix + self._sceneId.replace(".xml", "") + ".json", "w", encoding="utf-8"
        ) as f:
            json.dump(encoderOutput, f, ensure_ascii=False, indent=4)

    @property
    def sceneTrimesh(self):
        return self.__sceneTrimesh

    @property
    def edgeList(self):
        return self.__edgeList

    @property
    def angleCal(self):
        return self.__angleCal

    @property
    def sceneTrimeshName(self):
        return self.__sceneTrimeshName

    @property
    def sources(self):
        return self.__sources
