# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, AnyStr, Dict

from mpegi.geometry.trimesh import convertGeometriesToTrimesh

from .scenedictionary import *


class EdgeListGenPerScene:
    """Provides attributes and methods to generate edge lists from meshes."""

    def __init__(
        self,
        staticTrimeshName_DIFFR: AnyStr,
        earLevelVoxelGrid,
        sceneDict: Dict,
        args: Dict[str, Any],
    ):
        """
        Construct an EdgeListGenPerScene instance.

        EdgeListCreatorPerScene provides the attributes and methods to generate
        single/multiple edge list(s) from the meshes: static sceneTrimeshDIFFR
        and non-transparent dynamic meshes.

        :param staticTrimeshName_DIFFR: name of sceneTrimesh for diffraction component
        :param sceneDict: customized scene dictionary including artifact-relevant information
        """

        # Set the attributes from arguments
        self.__sceneTrimesh_DIFFR = sceneDict["sceneTrimesh_DIFFR"]  # be called from other classes
        self.minEdgeLength = args["min_edge_length_DIFFR"]

        if earLevelVoxelGrid == None:
            voxelEdgeLength_DIFFR = np.array([1.0, 1.0, 1.0])
        else:
            voxelEdgeLength_DIFFR = earLevelVoxelGrid.pitch

        # Define trimesh dictionary to test independently
        self.__trimDict = {staticTrimeshName_DIFFR: self.__sceneTrimesh_DIFFR}

        # Create geometry list to check if an included geometry is non-transparent and inherently dynamic
        geometryList = []
        [geometryList.append(mesh) for mesh in sceneDict["meshes"]]
        [geometryList.append(box) for box in sceneDict["boxes"]]
        [geometryList.append(cylinder) for cylinder in sceneDict["cylinders"]]

        for geo in geometryList:
            # dynamic and non-transparent geometry
            if (not sceneDict["sceneMeta"].isEntityStatic(geo.id)) and (not geo.isTransparent):
                # supposed to handle non-transparent dynamic geometry
                dynamicTrimesh = convertGeometriesToTrimesh(
                    geometry=[geo], includeTransparent=False, concatenate=False
                )
                self.__trimDict[geo.id] = dynamicTrimesh[0]

        # Define list of edge list generators from self.__meshesToAnalyze
        self.__edgeListGenDict = {}
        for tmID, tm in self.__trimDict.items():
            self.__edgeListGenDict[tmID] = EdgeListGenPerMesh(
                staticSceneTrimeshName=staticTrimeshName_DIFFR,
                sceneTrimeshName=tmID,
                sceneTrimesh=tm,
                sceneDict=sceneDict,
                voxelEdgeLength_DIFFR=voxelEdgeLength_DIFFR,
                minEdgeLength=self.minEdgeLength,
            )
            logging.info(
                f">> Mesh with ID[{tmID}] has {len(self.__edgeListGenDict[tmID].edgeList)} edges"
            )

        # Define the Dictionary of edge lists with mesh's name with it
        self.__edgeListDict = {}
        for tmID, eLG in self.__edgeListGenDict.items():
            self.__edgeListDict[tmID] = eLG.edgeList

    @property
    def edgeListGenDict(self):
        return self.__edgeListGenDict

    @property
    def edgeListDict(self):
        return self.__edgeListDict
