# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from tqdm import tqdm
from typing import Dict, List
import multiprocessing as mp

from mpegi.geometry.trimesh import trimeshToEifMesh, convertGeometriesToTrimesh

from .edgelistgenpermesh import *
from .enumclasses import Constants


class ListenerVoxelDictGen:
    def __init__(
        self,
        staticEdgeListGenerator: EdgeListGenPerMesh,
        sceneDict: Dict,
        earLevelVoxelGrid,
        earLevelVoxelPositions,
    ):
        """
        ListenerVoxelDictGen provides the attributes and methods to create the dictionary where its key is a string-type
        (listener) voxel index and reachable edgeIDs

        @param staticEdgeListGenerator: instantiated edgeListGenerator
        """

        if sceneDict["sceneInfo"]["type"] == SceneType.AR:
            self.__listenerVoxelDict = {}
            self.__audibleVoxelPositions = []
            return

        if staticEdgeListGenerator.sceneTrimeshName != "sceneTrimesh_DIFFR":
            raise ValueError(f">> ERROR: ListenerVoxelDictGen requires static sceneTrimesh_DIFFR")

        self.__static_elg = staticEdgeListGenerator
        self.__earLevelVoxelPositions = earLevelVoxelPositions
        self.earLevelVoxelGrid: trimesh.voxel.VoxelGrid = earLevelVoxelGrid
        self.__areAllSourcesStatic = all(
            [sceneDict["sceneMeta"].isEntityStatic(s.id) for s in self.__static_elg.sources]
        )

        # Voxel grid information to generate voxel indices from voxel center position
        self.__voxel_grid_origin_position = self.__static_elg.sceneTrimesh.bounds[0]
        self.__voxel_resolution = earLevelVoxelGrid.pitch

        # Prepare for parallel processing
        self.pool = mp.Pool(min(mp.cpu_count(), 4))
        self.sceneTrimesh_DIFFR: trimesh.Trimesh = sceneDict["sceneTrimesh_DIFFR"]

        self.__initializeVoxelDict()
        self.__updateVoxelDict()

    def __initializeVoxelDict(self):
        """Initialize voxelDict and related quantities"""

        self.__audibleVoxelPositions = []
        self.__listenerVoxelDict = {}
        for voxelPosition in self.__earLevelVoxelPositions:  # key: voxel index, not voxel position
            lvIndex = np.divide(
                np.array(voxelPosition) - np.array(self.__voxel_grid_origin_position),
                self.__voxel_resolution,
            )
            lvIndex = np.rint(lvIndex)
            lvIndexKey = str(np.array(lvIndex, dtype=int).tolist()).strip("[]")

            self.__audibleVoxelPositions.append(voxelPosition)
            self.__listenerVoxelDict[lvIndexKey] = []

    def __updateVoxelDict(self):

        self.__eCps: List[List[float]] = [edge.cp_vt for edge in self.__static_elg.edgeList]
        sceneEifMesh_DIFFR = trimeshToEifMesh(self.sceneTrimesh_DIFFR, "sceneEifMesh_DIFFR")

        nofSets = 0
        argsListPerSet = {nofSets: []}
        argsListForParallelProcessing = []

        voxel_count = 1
        nofVoxels = len(self.__audibleVoxelPositions)
        for voxelPosition in tqdm(
            self.__audibleVoxelPositions, desc=f">> Listener-Voxel-Generation"
        ):
            lvIndex = np.divide(
                np.array(voxelPosition) - np.array(self.__voxel_grid_origin_position),
                self.__voxel_resolution,
            )
            lvIndex = np.rint(lvIndex)
            lvIndexKey = str(np.array(lvIndex, dtype=int).tolist()).strip("[]")

            args = (
                lvIndexKey,
                voxelPosition,
                self.__voxel_resolution,
                self.finerDataNeeded(lvIndex),
            )

            argsListPerSet[nofSets].append(args)

            if (
                len(argsListPerSet[nofSets]) == int(nofVoxels / mp.cpu_count())
                or voxel_count == nofVoxels
            ):
                argsListForParallelProcessing.append(
                    (
                        self.__static_elg.edgeList,
                        self.__eCps,
                        sceneEifMesh_DIFFR,
                        argsListPerSet[nofSets],
                    )
                )
                nofSets += 1
                argsListPerSet[nofSets] = []  # new argsList for the next set
            voxel_count += 1

        results = []
        for args in argsListForParallelProcessing:
            r = self.pool.apply_async(ListenerVoxelDictGen.collectVisibleEdgeIDs_per_set, args=args)
            results.append(r)
        self.pool.close()
        self.pool.join()
        self.__returns = [r.get() for r in results]

        # self.__returns: List[Dict] = \
        #     self.pool.map(collectVisibleEdgeIDs_per_set,
        #                   argsListForParallelProcessing
        #                   )

        # Update listenerVoxelDict
        for listenerVoxelDictPerSet in self.__returns:
            listenerVoxelDictPerSet: Dict[int : List[int]]

            for lvIndexKey, lREIDs in listenerVoxelDictPerSet.items():
                self.__listenerVoxelDict[lvIndexKey] = lREIDs

        logging.info(f">> listener, nofListenerPositions: {len(self.__listenerVoxelDict.keys())}")

    def finerDataNeeded(self, lvIndex):
        # Only on the x-z plane
        for i in [-1, 0, 1]:
            for k in [-1, 0, 1]:
                if i == 0 and k == 0:
                    continue

                if (self.earLevelVoxelGrid.shape[0] - 1 < int(lvIndex[0]) + i) or (
                    self.earLevelVoxelGrid.shape[2] - 1 < int(lvIndex[2]) + k
                ):
                    continue

                if not self.earLevelVoxelGrid.matrix[int(lvIndex[0]) + i][int(lvIndex[1]) + 0][
                    int(lvIndex[2]) + k
                ]:
                    return True
        return False

    @staticmethod
    def collectVisibleEdgeIDs_per_set(
        edgeList: List[Edge], eCps: List[List[float]], sceneEifMesh_DIFFR, argsList
    ):

        trimeshList: List[trimesh.Trimesh] = convertGeometriesToTrimesh(
            [sceneEifMesh_DIFFR], True, 30, True
        )
        listenerVoxelDictPerSet = {}
        for lvIndexKey, voxelPosition, voxelResolution, finerDataNeeded in argsList:
            listenerVoxelDictPerSet[lvIndexKey] = []

            listenerVoxelDictPerSet[lvIndexKey] = ListenerVoxelDictGen.getVisibleEdgeIDsFrom(
                edgeList, eCps, trimeshList[0], voxelPosition
            )

            # Forced to not to make denser data
            finerDataNeeded = False
            if not finerDataNeeded:
                continue

            # To get more finer data to handle an area close to complex geometries,
            # where a visible edge list changes sensitively
            for i in [-2 / 3, 2 / 3]:
                for k in [-2 / 3, 2 / 3]:
                    voxelPosition[0] += i * voxelResolution[0]
                    voxelPosition[2] += k * voxelResolution[0]
                    visibleEdgeIDsFrom = ListenerVoxelDictGen.getVisibleEdgeIDsFrom(
                        edgeList, eCps, trimeshList[0], voxelPosition
                    )
                    for veId in visibleEdgeIDsFrom:
                        if veId not in listenerVoxelDictPerSet[lvIndexKey]:
                            listenerVoxelDictPerSet[lvIndexKey].append(veId)

        return listenerVoxelDictPerSet

    @staticmethod
    def getVisibleEdgeIDsFrom(
        edgeList: List[Edge],
        eCps: List[List[float]],
        sceneTrimesh_DIFFR: trimesh.Trimesh,
        p: List[float],
    ):

        vec_tp = np.ones((len(eCps), 1)) * p
        vec_t2e = eCps - vec_tp
        vec_t2e_distance = np.linalg.norm(vec_t2e, axis=1)
        (
            x_locations,
            idx_indicatingEdgeIDs,
            idx_xfaceIDs,
        ) = sceneTrimesh_DIFFR.ray.intersects_location(
            ray_origins=vec_tp, ray_directions=vec_t2e, multiple_hits=False
        )

        peIDsFromThis = [e.eID for e in edgeList if e.eID not in idx_indicatingEdgeIDs]
        veIDsFromThis = []
        vec_t2x = x_locations - np.ones((len(x_locations), 1)) * p
        vec_t2x_distance = np.linalg.norm(vec_t2x, axis=1)
        for eID, t2x_distance in zip(idx_indicatingEdgeIDs, vec_t2x_distance):
            # condition 1: out of audible range
            if vec_t2e_distance[eID] > Constants.MaxDistance_FromTheFirstEdgeToTheLastEdge_InPath:
                continue

            # condition 2: no obstacles
            if t2x_distance > vec_t2e_distance[eID] + Constants.PrecisionErrorThreshold:
                veIDsFromThis.append(eID)

        for eID in veIDsFromThis:
            e = edgeList[eID]

            if not e.relevant:
                continue

            # Simple filtering using dark zone analysis
            inner1_source = np.inner(p - e.cp, e.fnormals[0]) / np.linalg.norm(p - e.cp)
            inner2_source = np.inner(p - e.cp, e.fnormals[1]) / np.linalg.norm(p - e.cp)
            if inner1_source * inner2_source > 0:
                # for edges visible from the given position, but not in the dark zone of the edge
                continue

            # when the target is in one of the dark zones around the edge for edges visible from the given position
            peIDsFromThis.append(int(eID))
        return peIDsFromThis

    @property
    def listenerVoxelDict(self):
        return self.__listenerVoxelDict

    @property
    def audibleVoxelPositions(self):
        return self.__audibleVoxelPositions
