# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import copy
import math
import logging
import numpy as np
from typing import Union, List

from .edge import Edge
from .enumclasses import Constants, DiffractionTypes


class Path:
    def __init__(self, pid: int, edgeSeq: list, faceIndicatorSeq: list, angleSeq: list):

        """Create a specific diffractor path within a given audioScene"""
        if isinstance(pid, int):
            self.pid = pid
        else:
            raise TypeError("Required type of pid::int")

        if len(edgeSeq) == 1:
            self.__diffractionType: DiffractionTypes = DiffractionTypes.FirstOrder
        else:
            self.__diffractionType: DiffractionTypes = DiffractionTypes.HigherOrder

        if isinstance(edgeSeq, list):
            self.__edgeSeq = copy.deepcopy(edgeSeq)
            self.__edgeIDSeq = [edge.eID for edge in self.__edgeSeq if isinstance(edge, Edge)]
        else:
            raise TypeError("Required type of edgeSeq::list")

        if isinstance(faceIndicatorSeq, list):
            self.__faceIndicatorSeq = copy.deepcopy(faceIndicatorSeq)
        else:
            raise TypeError("Required type of faceSeq::list")

        if isinstance(angleSeq, list):
            self.__angleSeq = copy.deepcopy(angleSeq)
        else:
            raise TypeError("Required type of angleSeq::list")

        # the following attributes are only for optimization process
        self.__pathEndPoints = []
        self.__propagationDistance = 0.0
        self.__attenuationByDistance = 1.0
        self.__attenuationByDiffraction = 1.0
        self.__attenuation = 1.0

        self.__validated = False
        self.propagationDistance()

    def propagationDistance(self):
        for edgeIndex in range(len(self.__edgeSeq) - 1):
            fEdge = self.__edgeSeq[edgeIndex]
            nEdge = self.__edgeSeq[edgeIndex + 1]
            self.__propagationDistance += np.linalg.norm(fEdge.cp - nEdge.cp)

    def attenuationByDistance(self, sPos: np.ndarray, lPos: np.ndarray) -> float:
        return float(
            1
            / (
                self.__propagationDistance
                + np.linalg.norm(sPos - self.__edgeSeq[0].cp)
                + np.linalg.norm(lPos - self.__edgeSeq[-1].cp)
            )
        )

    def complete(self, angleList: Union[List], endPoints: List[List[float]]):
        """
        completes the path information for edge contribution/effect analysis during the geometry preprocessing stage
        @param angleList: single-entity list for a first order diffraction and two-entity list for a higher order diffraction
        @type angleList: list of float(s)
        @param endPoints: locations of the engaged source and the listener
        @type endPoints: list of list of 3-floats
        """
        if self.__validated:
            raise AttributeError(
                f"self.__validated: True -> already updated and consolidated with angles at source and listener sides"
            )

        if len(endPoints) == 2:
            self.__pathEndPoints.append(endPoints[0])
            self.__pathEndPoints.append(endPoints[1])
        else:
            raise TypeError(f"endPoints: List of two List[3 floats]")

        # update distance and attenuation gains by distance and diffraction
        self.__attenuationByDistance = 1 / (
            self.__propagationDistance
            + np.linalg.norm(self.__pathEndPoints[0] - self.__edgeSeq[0].cp)
            + np.linalg.norm(self.__pathEndPoints[1] - self.__edgeSeq[-1].cp)
        )

        _angleSeq = self.__angleSeq
        if len(angleList) == 1:
            _angleSeq[0] = angleList[0]
        elif len(angleList) == 2:
            _angleSeq[0] = angleList[0]
            _angleSeq[-1] = angleList[1]

        for angleToDiffract in _angleSeq:
            if math.floor(angleToDiffract) <= 179:
                self.__attenuationByDiffraction *= Constants.fCoef_1kHz[math.floor(angleToDiffract)]
            else:
                self.__attenuationByDiffraction *= Constants.fCoef_1kHz[179]
                # TODO: "Check where the angleToDiffract > 180"
                # print(f"\nangeToDiffract: {angleToDiffract, math.floor(angleToDiffract)}")

        self.__attenuation = self.__attenuationByDiffraction * self.__attenuationByDistance
        self.__validated = True

    def resetValidated(self):
        """reset the validated flag as False and the initial and the last angleSeq as 0 for the analysis"""

        self.__validated = False

    def getAttenuationGain(self):
        """returns the attenuation gain"""

        if self.__validated:
            return self.__attenuation
        else:
            logging.info(f"self.__validated: False")
            return 0

    @property
    def edgeSeq(self):
        return self.__edgeSeq

    @property
    def edgeIDSeq(self):
        return self.__edgeIDSeq

    @property
    def faceIndicatorSeq(self):
        return self.__faceIndicatorSeq

    @property
    def angleSeq(self):
        return self.__angleSeq

    @property
    def validated(self):
        return self.__validated

    @property
    def diffractionType(self):
        return self.__diffractionType

    def __str__(self):
        return "<< pathID: %s, edgeIdSeq: %s, faceIndSeq: %s, angleSeq: %s" % (
            self.pid,
            self.__edgeIDSeq,
            self.__faceIndicatorSeq,
            self.__angleSeq,
        )
