# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from tqdm import tqdm

from .edge import Edge
from .enumclasses import ConfigurationType, Constants, CoordinateSystems
from .path import Path
from .sourcevoxeldictgen import *


class PathDictGenPerMesh:
    def __init__(
        self,
        edgeListGenerator: EdgeListGenPerMesh,
        sourceVoxelDictGen: Union[SourceVoxelDictGen, None],
        maximumDiffractionOrder: int = 5,
        maximumDiffractionOrderForMovingSources: int = 3,
        dynamicMesh: bool = False,
        sceneDict: Dict = None,
        edgeLengthThresholdForFirstOrder: float = 0.314,
    ):
        """
        PathDictGenPerMesh provides the attributes and methods to generate a dictionary which consists of trimesh name as its key
        and a dictionary of pairs of an edgeID and potential paths starting from the edgeID

        @param edgeListGenerator: instantiated edgeListGenerator
        @param sourceVoxelDictGen: instantiated sourceVoxelDictGen with static edgeListGen only
        @param maximumDiffractionOrder: maximum number of edges in a single path
        @param dynamicMesh: false by default for the case of static meshes and true for dynamic non-transparent meshes
        """

        # Set instance attributes from arguments
        self.__maximumDiffractionOrder = maximumDiffractionOrder
        self.__maximumDiffractionOrderForMovingSources = maximumDiffractionOrderForMovingSources
        self.edgeListGenerator = edgeListGenerator
        self.edgeLengthThresholdForFirstOrder = edgeLengthThresholdForFirstOrder

        # Get edgeList and related quantities
        self.__edgeList = edgeListGenerator.edgeList
        self.__sourceVoxelDict = sourceVoxelDictGen.sourceVoxelDict
        self.__angleCal = edgeListGenerator.angleCal
        self.__sceneTrimesh = edgeListGenerator.sceneTrimesh  # for visualization
        self.__sceneDict = sceneDict
        self._sceneId = sceneDict["sceneInfo"]["id"]

        self.__flag_isARScene = sceneDict["sceneInfo"]["type"] == SceneType.AR
        self.__flag_allSourcesStatic = sourceVoxelDictGen.allSourcesStatic
        self.__flag_isDynamicMesh = dynamicMesh
        self.__flag_morePaths = sceneDict["_flag_morePaths"]
        self.__flag_allSourcesDynamic = sceneDict["_flag_allSourcesDynamic"]

        self.__reachableEdgeIDsFromSources = []

        # For static mesh in VR scene
        self.__edgeIDsForPathDictKeys = []

        if self.__flag_isDynamicMesh or self.__flag_isARScene or (not self.__flag_allSourcesStatic):
            self.__edgeIDsForPathDictKeys = [i for i in range(len(self.__edgeList))]
        else:
            for (
                sID,
                iSVD,
            ) in self.__sourceVoxelDict.items():  # sourceID, individualSourceVoxelDictionary (iSVD)
                for (
                    svIndexKey,
                    sREIDs,
                ) in iSVD.items():  # sourcePositionAsKey, sourceReachableEdgeIDs (sREIDs)
                    for rEID in sREIDs:
                        if rEID not in self.__edgeIDsForPathDictKeys:
                            self.__edgeIDsForPathDictKeys.append(rEID)

        self.__initialize()
        self.__update()
        self.__updateEdgeRelevantForUnreachableEdges()

    def __initialize(self):
        # Initialize pathDict
        self.nofPaths = 0
        self.__pathDict = {}
        for edgeID in self.__edgeIDsForPathDictKeys:
            self.__pathDict[edgeID] = []

        # Initialize pathDictByEdgeIds
        self.nofPathsByEdgeIds = 0
        self.__pathDictByEdgeIds = {}

    def __update(self):
        """
        update self.__pathDict and a new path has the following information:
            1) key: self.__pathDict[edgeID] where edgeID is the initial edge id (int) of the included path
            2) value: self.__pathDict[edgeID].append(newPath) where newPath is a new Path object
        """

        for eid in tqdm(
            self.__pathDict.keys(),
            desc=f">> Path-Dictionary-Generation [{self.edgeListGenerator.sceneTrimeshName}]",
        ):
            nofAllowedMaxOrder = -1
            # Per-edge max path order assignment
            if self.__flag_isARScene:
                nofAllowedMaxOrder = self.__maximumDiffractionOrder  # 5 by default
            else:
                if self.__flag_isDynamicMesh:
                    # For dynamic edges
                    nofAllowedMaxOrder = self.__maximumDiffractionOrder  # 5 by default
                else:
                    # For static edges
                    if self.__flag_allSourcesStatic:
                        if self.__edgeList[eid].isVisibleFromStaticSources:
                            # for reachable edges from static sources
                            if self.__flag_morePaths:
                                nofAllowedMaxOrder = 7
                            else:
                                nofAllowedMaxOrder = self.__maximumDiffractionOrder  # 5 by default
                        else:
                            continue
                    elif self.__flag_allSourcesDynamic:
                        nofAllowedMaxOrder = 5
                    else:
                        if self.__edgeList[eid].isVisibleFromStaticSources:
                            nofAllowedMaxOrder = self.__maximumDiffractionOrder  # 5 by default
                        else:
                            eVTXs = self.__edgeList[eid].vtxPositions
                            samples = []
                            for alpha in range(10):
                                samples.append(
                                    np.add(eVTXs[0] * (alpha / 10), eVTXs[1] * (1 - alpha / 10))
                                )

                            reachableFromAEs = False
                            for trimesh_AE in self.__sceneDict["trimeshes_AcousticEnvironments"]:
                                for sample in samples:
                                    reachableFromAEs |= VoxelizerDIFFR.withinEnv(
                                        sample, trimesh_AE.bounds, ndim=2
                                    )

                            reachableFromURRs = False
                            for trimesh_URR in self.__sceneDict["trimeshes_URRs"]:
                                for sample in samples:
                                    reachableFromURRs |= VoxelizerDIFFR.withinEnv(
                                        sample, trimesh_URR.bounds, ndim=2
                                    )

                            if reachableFromAEs or reachableFromURRs:
                                # for edges inside URRs or AEs (full support)
                                nofAllowedMaxOrder = self.__maximumDiffractionOrderForMovingSources
                            else:
                                # for edges outside URRs or AEs (1st order support)
                                nofAllowedMaxOrder = 1  # 1 by default

            if nofAllowedMaxOrder == -1:
                raise ValueError(f"Path-Dict-Gen-Per-Mesh: [nofAllowedMaxOrder == -1] = True")

            frontEdge: Edge = self.__edgeList[eid]
            if frontEdge.edgeLength < self.edgeLengthThresholdForFirstOrder:
                nofAllowedMaxOrder = 1

            tempEdges = []
            tempFaceIndicators = []
            tempAngles = []
            faceIndicator = [0, 1]

            # Forward direction
            frontEdge
            tempAngleOffset = 0.0
            tempEdges.append(frontEdge)
            tempFaceIndicators.append(faceIndicator[0])
            tempAngles

            self.__recursivePathFinder(
                frontEdge,
                tempAngleOffset,
                tempEdges,
                tempFaceIndicators,
                tempAngles,
                nofAllowedMaxOrder,
            )

            # Backward direction
            frontEdge = self.__edgeList[eid]
            tempAngleOffset = 0.0
            tempEdges.clear()
            tempEdges.append(frontEdge)
            tempFaceIndicators.clear()
            tempFaceIndicators.append(faceIndicator[1])
            tempAngles.clear()

            self.__recursivePathFinder(
                frontEdge,
                tempAngleOffset,
                tempEdges,
                tempFaceIndicators,
                tempAngles,
                nofAllowedMaxOrder,
            )

    def __checkDarkZonePresenceByAngle(self, target, refEdge, refFaceIndicator):
        """
        returns the internal angle between a specified face of reference edge and target
        so that the returned float value can be used for dark zone presence

        @param target: target point location
        @type target: list of three floats
        @param refEdge: reference edge which will be used as a rotating axis
        @type refEdge: Edge
        @param refFaceIndicator: 0 or 1 which indicates which face within the adjacent faces should be used
        @type refFaceIndicator: int
        @return: internal angle
        @rtype: float
        """
        # todo: return type must be a boolean value? (might be straightforward)
        self.__angleCal.set(
            coord=CoordinateSystems.OpenGL,
            config_type=ConfigurationType.VertexToTriangle,
            v_origin=refEdge.vtxPositions[0],
            v_bias=refEdge.vtxPositions[1],
            target=target,
            fid=refEdge.fIDs[refFaceIndicator],
        )

        return self.__angleCal.cal_internalAngle()

    def __isNextEdgeinTheDarkZone(self, frontEdge, nextEdge, indicator, tempAngleOffset):
        """
        check the next edge is inside of the dark zone from the front edge location
        it can tell if the next edge is within the range of diffraction or not.

        @param frontEdge: reference edge where sound propagates around
        @type frontEdge: Edge
        @param nextEdge: the next edge to be checked for the possibility of diffraction
        @type nextEdge: Edge
        @param indicator: 0 or 1 which indicates which face within the adjacent faces should be used
        @type indicator: int
        @param tempAngleOffset: offset angle
        @type tempAngleOffset: float
        @return: True if the next edge is in the dark zone of the front edge
        @rtype: bool
        """
        # check if the next edge is present within the dark zone of the reference (front) edge
        # ---------------------------- #
        target = nextEdge.cp
        # ---------------------------- #
        angle_f2e = self.__checkDarkZonePresenceByAngle(target, frontEdge, indicator)
        Tr1 = angle_f2e > 180.0 + tempAngleOffset + Constants.MinOffsetAngle_PerEdge
        Tr2 = angle_f2e < frontEdge.edgeAngle + 1e-3

        if not (Tr1 and Tr2):
            return False, angle_f2e
        return True, angle_f2e

    def __isDiffractionable(self, frontEdge: Union[Edge, List, np.ndarray], nextEdge):
        """
        check if sound can diffract around the next edge from the front edge
        @param frontEdge: reference edge which sound propagates from
        @type frontEdge: Edge
        @param nextEdge: the next edge to be checked for the possibility of diffraction
        @type nextEdge: Edge
        @return: True if the next edge is in the dark zone of the front edge for possible diffraction
        @rtype: bool
        """
        # check if sound can diffract around the next edge
        # ---------------------------- #
        target = frontEdge
        if isinstance(frontEdge, Edge):
            target = frontEdge.cp_vt
        elif isinstance(frontEdge, List):
            target = np.array(frontEdge)
        # ---------------------------- #

        _angle_Face2EdgeCP_Face0 = self.__checkDarkZonePresenceByAngle(target, nextEdge, 0)
        _angle_Face2EdgeCP_Face1 = self.__checkDarkZonePresenceByAngle(target, nextEdge, 1)

        Tr1 = _angle_Face2EdgeCP_Face0 > 180.0 + Constants.MinOffsetAngle_PerEdge
        Tr2 = _angle_Face2EdgeCP_Face0 < nextEdge.edgeAngle + 1e-3
        Tr3 = _angle_Face2EdgeCP_Face1 < 180.0 - Constants.MinOffsetAngle_PerEdge

        if Tr1 and Tr2 and Tr3:
            return True, _angle_Face2EdgeCP_Face1, 1

        Tr1 = _angle_Face2EdgeCP_Face1 > 180.0 + Constants.MinOffsetAngle_PerEdge
        Tr2 = _angle_Face2EdgeCP_Face1 < nextEdge.edgeAngle + 1e-3
        Tr3 = _angle_Face2EdgeCP_Face0 < 180.0 - Constants.MinOffsetAngle_PerEdge

        if Tr1 and Tr2 and Tr3:
            return True, _angle_Face2EdgeCP_Face0, 0

        return False, -1, -1

    def __recursivePathFinder(
        self,
        frontEdge,
        tempAngleOffset,
        tempEdges,
        tempFaceIndicators,
        tempAngles,
        nofAllowedMaxOrder,
    ):
        """
        path finder is designed to keep looking for dark zones from the front Edge to unvisited edges.
        and if one of the ending conditions is satisfied, then its accumulated path information defined as three
        sequences will create a new path object and it will be added to the pathDict attribute.
        """
        for nextEdgeID in frontEdge.visibleEdgeIDsFromThis:
            nextEdge = self.__edgeList[nextEdgeID]
            if nextEdge in tempEdges:
                continue

            if frontEdge.eID == nextEdge.eID:
                continue

            if len(tempEdges) == nofAllowedMaxOrder:
                break

            isVisible = False
            for e in tempEdges[1:-1]:
                if nextEdgeID in e.visibleEdgeIDsToLimit:
                    isVisible = True
                    break

            if isVisible:
                continue

            isNextEdgeInTheDarkZone, angle_f2e = self.__isNextEdgeinTheDarkZone(
                frontEdge, nextEdge, tempFaceIndicators[-1], tempAngleOffset
            )
            if not isNextEdgeInTheDarkZone:
                continue

            if self.isThePathLongerThanLimit(tempEdges):
                continue

            isDiffractionable, angleOffset, faceIndicator = self.__isDiffractionable(
                frontEdge, nextEdge
            )
            if not isDiffractionable:
                continue

            # reset the arguments for next diffraction test
            frontEdgeForNextIteration = nextEdge
            tempEdges.append(frontEdgeForNextIteration)
            tempFaceIndicators.append(faceIndicator)
            tempAngles.append(angle_f2e - (180 + tempAngleOffset))
            newTempAngleOffset = angleOffset

            self.__recursivePathFinder(
                frontEdgeForNextIteration,
                newTempAngleOffset,
                tempEdges,
                tempFaceIndicators,
                tempAngles,
                nofAllowedMaxOrder,
            )

            tempEdges.pop()
            tempFaceIndicators.pop()
            tempAngles.pop()

        # create a new diffractionPath
        tempAngles.append(tempAngleOffset + 180.0)
        newPath = Path(
            pid=self.nofPaths,
            edgeSeq=tempEdges,
            faceIndicatorSeq=tempFaceIndicators,
            angleSeq=tempAngles,
        )
        tempAngles.pop()

        # Update pathDict: Dict{int: List[Path]}
        iEdgeID = tempEdges[0].eID
        self.__pathDict[iEdgeID].append(newPath)
        self.nofPaths += 1

        # Update pathDictByEdgeIds: Dict{str(int, int): List[Path]}
        fEdgeId = tempEdges[-1].eID
        try:
            self.__pathDictByEdgeIds[str([iEdgeID, fEdgeId])]
        except KeyError:
            self.__pathDictByEdgeIds[str([iEdgeID, fEdgeId])]: List = []
        finally:
            self.__pathDictByEdgeIds[str([iEdgeID, fEdgeId])].append(newPath)
            self.nofPathsByEdgeIds += 1

    @staticmethod
    def isThePathLongerThanLimit(tempEdges):
        """
        check if the given path by tempEdges list is longer than pre-assigned maximum distance for a valid path

        @param tempEdges:  list of edges to define a sample path
        @type tempEdges: list of edges
        @return: true if the given path is longer than limit. Otherwise false.
        @rtype: bool
        """
        tempDiffractionOrder = len(tempEdges)
        propagationDistance = 0.0
        for i in range(0, tempDiffractionOrder - 1, 1):
            propagationDistance += np.linalg.norm(tempEdges[i + 1].cp_vt - tempEdges[i + 0].cp_vt)
        if propagationDistance > Constants.MaxDistance_FromTheFirstEdgeToTheLastEdge_InPath:
            return True
        return False

    # will be deleted
    @staticmethod
    def writeDiffractionPathDict(EncoderOutput: Dict, pathDict: Dict):
        """
        # General procedure in mind as shown below
        encoderOutput["pathDict"] = {}
        and each dictionary contains all the possible paths starting from one of the edges::

            data["pathDict"] = {"eid_1":[entityDict_1, entityDict_2, ..., entityDict_m],
                                "eid_2":[],
                                ...,
                                "eid_n":[]}

            entityDict = ['_diffractionEdgeIDSequence': [],
                          '_diffractionFaceIndicatorSequence': [],
                          '_diffractionAngleSequence': []]
            data["pathDict"]["eid"].append(entityDict)"""

        if not isinstance(pathDict, dict):
            raise TypeError("Required type of pathDict::dict")

        for key, pathList in pathDict.items():
            if not isinstance(pathList, list):
                print("key: %s, pathList: %s" % (key, pathList))
                raise TypeError("Required type of pathDict[key]::list")

            for path in pathList:
                if not isinstance(path, Path):
                    print("path: %s, type(path): %s" % (path, type(path)))
                    raise TypeError("Required type of path::DiffractionPath")

        if isinstance(EncoderOutput, dict):
            if "pathDict" not in EncoderOutput.keys():
                raise AttributeError("Required key in the given data::'pathDict'")

            for key in pathDict.keys():
                EncoderOutput["pathDict"][key] = []

        for key, pathList in pathDict.items():
            for path in pathList:
                entityDict = {
                    "_diffractionEdgeIDSequence": path.edgeIDSeq,
                    "_diffractionFaceIndicatorSequence": path.faceIndicatorSeq,
                    "_diffractionAngleSequence": path.angleSeq,
                }

                EncoderOutput["pathDict"][key].append(entityDict)

    # will be deleted
    def writeJSONFile(self, jsonFileNamePrefix: AnyStr = "json_PathDictionaryForDiffraction_"):
        """
        write json file to save the path database (dictionary)
        """
        # write a dictionary of precomputed paths
        encoderOutput = {"pathDict": {}}
        self.writeDiffractionPathDict(EncoderOutput=encoderOutput, pathDict=self.__pathDict)

        # write the json with diffracted path information

        with open(
            jsonFileNamePrefix + self._sceneId.replace(".xml", "") + ".json", "w", encoding="utf-8"
        ) as f:
            json.dump(encoderOutput, f, ensure_ascii=False, indent=4)

    def __updateReachableEdgeIDsFromSources(self):
        for sourceID, iSVD in self.__sourceVoxelDict.items():
            for svIndexKey, sVEL in iSVD.items():
                # precomputed path list with the sVE as the initial edge
                for eID in sVEL:  # sourceVisibleEdgeList (sVEL)
                    sVPL: List[Path] = self.__pathDict[eID]
                    for sVP in sVPL:
                        edgeIDSeq = sVP.edgeIDSeq
                        if edgeIDSeq[-1] not in self.__reachableEdgeIDsFromSources:
                            self.__reachableEdgeIDsFromSources.append(edgeIDSeq[-1])

    def __updateEdgeRelevantForUnreachableEdges(self):
        if self.__flag_isDynamicMesh or self.__flag_isARScene or (not self.__flag_allSourcesStatic):
            return

        # Assume that we're handling a static mesh in the scene
        # where all the sources are static
        self.__updateReachableEdgeIDsFromSources()
        for edge in self.__edgeList:
            if edge.eID not in self.__reachableEdgeIDsFromSources:
                edge.relevant = False

    @property
    def pathDict(self):
        return self.__pathDict

    @property
    def pathDictByEdgeIds(self):
        return self.__pathDictByEdgeIds

    @property
    def reachableEdgeIDsFromSources(self):
        if self.__flag_isDynamicMesh:
            return None
        return self.__reachableEdgeIDsFromSources
