# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from mpegi.metadata.eifplus.mesh import Mesh
from mpegi.metadata.eifplus.update import Update
from mpegi.metadata.eifplus.metadata import SceneMetadata
from mpegi.metadata.eifplus.objectsource import ObjectSource
from mpegi.metadata.eifplus.primitives import Box, Sphere, Cylinder
from mpegi.metadata.eifplus.environments import AcousticEnvironment, AcousticParameters
from mpegi.geometry.trimesh import convertGeometriesToTrimesh

from .edgelistgenpermesh import *


class SceneDictionary:
    def __init__(self, artifacts, loggingPrints: bool = True):
        """
        SceneDictionary returns customized dictionary containing diffraction-relevant information directly from artifacts
        or preprocessed information from them.
        """

        _sceneTrimesh_DIFFR_as_eif_mesh = artifacts["scene_trimesh_DIFFR_as_eif_mesh"]
        sceneTrimesh_DIFFR: trimesh.Trimesh = convertGeometriesToTrimesh(
            [_sceneTrimesh_DIFFR_as_eif_mesh], faceAngle=30, process=False
        )[0]
        sceneMeta: SceneMetadata = artifacts["scene_metadata"]
        sceneInfo: Dict = artifacts["scene_info"]
        objectSources: List[ObjectSource] = artifacts["object_sources"]
        boxes: List[Box] = artifacts["boxes"]
        meshes: List[Mesh] = artifacts["meshes"]
        spheres: List[Sphere] = artifacts["spheres"]
        cylinders: List[Cylinder] = artifacts["cylinders"]
        acousticEnv: List[AcousticEnvironment] = artifacts["acoustic_environments"]
        updates: List[Update] = artifacts["updates"]
        user_regions = artifacts["user_regions"]
        anchors = None
        audio_streams = None
        channel_sources = None
        acoustic_materials = None
        directivities = None
        transforms = None
        hoa_groups = None
        hoa_sources = None

        logging.info(f"********** SceneDictionary:")

        geometries = [*boxes, *spheres, *cylinders, *meshes]
        geoIds = [geo.id for geo in geometries]
        acousticEnvRegionIds = [env.region for env in acousticEnv]
        acousticEnvGeometries = list()
        for regionId in acousticEnvRegionIds:
            if regionId is None:
                continue
            index = geoIds.index(regionId)
            acousticEnvGeometries.append(geometries[index])

        self._sceneDictionary = {
            "sceneTrimesh_DIFFR": sceneTrimesh_DIFFR,
            "sceneMeta": sceneMeta,
            "sceneInfo": sceneInfo,
            "objectSources": objectSources,
            "boxes": boxes,
            "spheres": spheres,
            "cylinders": cylinders,
            "meshes": meshes,
            "updates": updates,
            "geometries": geometries,
            "acousticEnv": acousticEnv,
            "acousticEnvRegionIds": acousticEnvRegionIds,
            "acousticEnvGeometries": acousticEnvGeometries,
            "user_regions": user_regions,
        }

        # AE-relevant derived parameters
        trimesh_AcousticEnvironments = convertGeometriesToTrimesh(
            self._sceneDictionary["acousticEnvGeometries"],
            concatenate=True,
            includeTransparent=True,
        )[0]
        self._sceneDictionary["trimesh_AcousticEnvironments"] = trimesh_AcousticEnvironments

        self._sceneDictionary["trimeshes_AcousticEnvironments"] = []
        for AE in self._sceneDictionary["acousticEnvGeometries"]:
            trimesh_AE = convertGeometriesToTrimesh(
                [AE],
                concatenate=False,
                includeTransparent=True,
            )[0]
            self._sceneDictionary["trimeshes_AcousticEnvironments"].append(trimesh_AE)

        self._sceneDictionary["AEs_points"] = []
        if not trimesh_AcousticEnvironments.is_empty:
            for trimesh_AE in self._sceneDictionary["trimeshes_AcousticEnvironments"]:
                AE_bounds = trimesh_AE.bounds
                AE_points = []
                AE_points.append(AE_bounds[0])
                AE_points.append(AE_bounds[1])
                AE_points.append([AE_bounds[1][0], AE_bounds[0][1], AE_bounds[0][2]])
                AE_points.append([AE_bounds[1][0], AE_bounds[0][1], AE_bounds[1][2]])
                AE_points.append([AE_bounds[1][0], AE_bounds[1][1], AE_bounds[0][2]])
                AE_points.append([AE_bounds[0][0], AE_bounds[0][1], AE_bounds[1][2]])
                AE_points.append([AE_bounds[0][0], AE_bounds[1][1], AE_bounds[1][2]])
                AE_points.append([AE_bounds[0][0], AE_bounds[1][1], AE_bounds[0][2]])

                self._sceneDictionary["AEs_points"].append(AE_points)

        # URR-relevant derived parameters
        trimesh_UserReachableRegions = convertGeometriesToTrimesh(
            user_regions, concatenate=True, includeTransparent=True
        )[0]
        self._sceneDictionary["trimesh_UserReachableRegions"] = trimesh_UserReachableRegions

        self._sceneDictionary["trimeshes_URRs"] = []
        for URR in user_regions:
            trimesh_URR = convertGeometriesToTrimesh(
                [URR],
                concatenate=False,
                includeTransparent=True,
            )[0]
            self._sceneDictionary["trimeshes_URRs"].append(trimesh_URR)

        self._sceneDictionary["URRs_points"] = []
        if self._sceneDictionary["sceneInfo"]["type"] == SceneType.VR:
            for trimesh_URR in self._sceneDictionary["trimeshes_URRs"]:
                URR_bounds = trimesh_URR.bounds

                URR_points = []
                URR_points.append(URR_bounds[0])
                URR_points.append(URR_bounds[1])
                URR_points.append([URR_bounds[1][0], URR_bounds[0][1], URR_bounds[0][2]])
                URR_points.append([URR_bounds[1][0], URR_bounds[0][1], URR_bounds[1][2]])
                URR_points.append([URR_bounds[1][0], URR_bounds[1][1], URR_bounds[0][2]])
                URR_points.append([URR_bounds[0][0], URR_bounds[0][1], URR_bounds[1][2]])
                URR_points.append([URR_bounds[0][0], URR_bounds[1][1], URR_bounds[1][2]])
                URR_points.append([URR_bounds[0][0], URR_bounds[1][1], URR_bounds[0][2]])

                self._sceneDictionary["URRs_points"].append(URR_points)

        _acousticEnvBoundsDict = dict()
        for i, acousticEnvGeo in enumerate(self._sceneDictionary["acousticEnvGeometries"]):
            _acousticEnvTrimesh: List[trimesh.Trimesh] = convertGeometriesToTrimesh(
                geometry=[acousticEnvGeo], includeTransparent=True
            )

            _acousticEnvRegionId = self._sceneDictionary["acousticEnvRegionIds"][i]
            _acousticEnvVoxelGrid: trimesh.voxel.VoxelGrid = _acousticEnvTrimesh[0].voxelized(
                pitch=1.0
            )
            _acousticEnvBoundsDict[_acousticEnvRegionId] = _acousticEnvVoxelGrid.bounds
        self._sceneDictionary["acousticEnvBoundsDict"]: Dict[AnyStr:List] = _acousticEnvBoundsDict

        # Set flags
        # For AR scenes
        _flag_isARScene = self._sceneDictionary["sceneInfo"]["type"] == SceneType.AR

        # For VR scenes
        _flag_emtpySceneTrimesh_DIFFR = self._sceneDictionary["sceneTrimesh_DIFFR"].is_empty
        _flag_acEnvExists = len(self._sceneDictionary["acousticEnvBoundsDict"].keys()) != 0
        _flag_noObjectSources = len(self._sceneDictionary["objectSources"]) == 0
        _flag_allSourcesStatic = all(
            [
                self._sceneDictionary["sceneMeta"].isEntityStatic(s.id)
                for s in self._sceneDictionary["objectSources"]
            ]
        )
        _flag_allSourcesDynamic = all(
            [
                not self._sceneDictionary["sceneMeta"].isEntityStatic(s.id)
                for s in self._sceneDictionary["objectSources"]
            ]
        )

        _flag_isDynamicGeoIncluded = False
        geometryList = []
        [geometryList.append(mesh) for mesh in meshes]
        [geometryList.append(box) for box in boxes]
        [geometryList.append(cylinder) for cylinder in cylinders]
        for geo in geometryList:
            # dynamic and non-transparent geometry
            if (not sceneMeta.isEntityStatic(geo.id)) and (not geo.isTransparent):
                _flag_isDynamicGeoIncluded = True
                break

        _flag_airMediumExists = not _flag_allSourcesStatic and _flag_acEnvExists
        # set as False, but can be true when airMedium generation needs less computation
        # and it is fair enough to keep the code for later usages.
        _flag_airMediumExists = False
        _flag_doubleAudibleRange = _flag_allSourcesStatic and not _flag_acEnvExists

        _flag_morePaths = False
        if _flag_acEnvExists:
            AE: AcousticEnvironment
            for AE in acousticEnv:
                for para in AE.parameters:
                    if para.frequencies[0].dsr == None:
                        continue

                    _flag_morePaths |= para.frequencies[0].dsr < 1e-9

        self._sceneDictionary["_flag_isARScene"] = _flag_isARScene
        self._sceneDictionary["_flag_acEnvExists"] = _flag_acEnvExists
        self._sceneDictionary["_flag_noObjectSources"] = _flag_noObjectSources
        self._sceneDictionary["_flag_allSourcesStatic"] = _flag_allSourcesStatic
        self._sceneDictionary["_flag_allSourcesDynamic"] = _flag_allSourcesDynamic
        self._sceneDictionary["_flag_isDynamicGeoIncluded"] = _flag_isDynamicGeoIncluded
        self._sceneDictionary["_flag_airMediumExists"] = _flag_airMediumExists
        self._sceneDictionary["_flag_emtpySceneTrimesh_DIFFR"] = _flag_emtpySceneTrimesh_DIFFR
        self._sceneDictionary["_flag_doubleAudibleRange"] = _flag_doubleAudibleRange
        self._sceneDictionary["_flag_morePaths"] = _flag_morePaths

        if not loggingPrints:
            return

        logging.info(f"The given scene's general status can be described by the following flags")
        logging.info(f">> ---------------------------------------------------------")
        logging.info(f">> _flag_isARScene               : {_flag_isARScene}")
        logging.info(f">> _flag_acEnvExists             : {_flag_acEnvExists}")
        logging.info(f">> _flag_allSourcesStatic        : {_flag_allSourcesStatic}")
        logging.info(f">> _flag_allSourcesDynamic       : {_flag_allSourcesDynamic}")
        logging.info(f">> _flag_noObjectSources         : {_flag_noObjectSources}")
        logging.info(f">> _flag_airMediumExists         : {_flag_airMediumExists}")
        logging.info(f">> _flag_emtpySceneTrimesh_DIFFR : {_flag_emtpySceneTrimesh_DIFFR}")
        logging.info(f">> _flag_doubleAudibleRange      : {_flag_doubleAudibleRange}")
        logging.info(f">> _flag_morePaths               : {_flag_morePaths}")
        logging.info(f">> ---------------------------------------------------------")

    @property
    def sceneDictionary(self):
        return self._sceneDictionary
