# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from tqdm import tqdm

from .edgelistgenpermesh import *
from ..eifplus.objectsource import ObjectSource


class SourceVoxelDictGen:
    def __init__(
        self, staticEdgeListGenerator: EdgeListGenPerMesh, sceneDict: Dict, earLevelVoxelGrid
    ):
        """
        SourceVoxelDictGen provides the attributes and methods to create the dictionary where its key is source id (str)
        and its corresponding value is a dictionary of pairs of stringized source voxel index and reachable edgeIDs

        @param staticEdgeListGenerator: instantiated edgeListGenerator
        @param sceneDict: dictionary containing artifacts and derived secondary information
        """

        logging.info(f"********** SourceVoxelDictGen:")

        if sceneDict["sceneInfo"]["type"] == SceneType.AR:
            self.__sourceVoxelDict = {}
            self.__allSourcesStatic = False
            return

        if staticEdgeListGenerator.sceneTrimeshName != "sceneTrimesh_DIFFR":
            raise ValueError(
                f">> ERROR: SourceVoxelDictGen requires edgeListGenerator with sceneTrimesh_DIFFR"
            )

        self.__staticEdgeListGen = staticEdgeListGenerator

        # Assign necessary member attributes for conditional process
        self._sceneDict = sceneDict
        self.__sources = sceneDict["objectSources"]
        self.__updates = sceneDict["updates"]
        self.__allSourcesStatic = sceneDict["_flag_allSourcesStatic"]

        # Voxel grid information to generate voxel indices from voxel center position
        self.__voxel_grid_origin_position = staticEdgeListGenerator.sceneTrimesh.bounds[0]
        self.__voxel_resolution = earLevelVoxelGrid.pitch

        self.__initialize()
        self.__update()

    def __initialize(self):
        """Initialize voxelDict and related quantities"""

        self.__timedSourceUpdates = {}
        for update in self.__updates:
            if update.type != "Timed":
                continue

            modification = update.modifications[0]
            if (modification.id.find("objSrc") != -1) and (
                modification.changes.find("position") != -1
            ):
                self.__timedSourceUpdates[modification.id] = update.modifications

        # for k, v in self.__timedSourceUpdates.items():
        #     print(f"source's timed modification on positions: {k, v}")

        self.__sourceVoxelDict = {}
        for source in tqdm(
            self.__sources, desc=f">> Source-Voxel-Generation [N: {len(self.__sources)}]"
        ):

            ############################################################################
            # TODO: GENERATE SOURCE GRID FROM TIMED UPDATES INFORMATION ON POSITION.
            #  Q: WHICH SCENES HAVE TIMED-UPDATED SOURCES
            # sourceGrid = SourceVoxelGrid(self.__timedSourceUpdates[key])
            ############################################################################
            # if key in self.__timedSourceUpdates.keys():
            #     print(f"self.__timedSourceUpdates[key]: {self.__timedSourceUpdates[key]}")
            #     raise NotImplementedError(
            #         "ERROR: From self.__timedSourceUpdates[key]: need to generate trajectory of source positions and update sVEL for each location")
            # else:
            ############################################################################

            if not self._sceneDict["sceneMeta"].isEntityStatic(source.id):
                continue

            if source.extent != None:
                isSourceExtentStatic = self._sceneDict["sceneMeta"].isEntityStatic(source.extent)
                if not isSourceExtentStatic:
                    continue

            vIndex = np.divide(
                np.array(source.position) - np.array(self.__voxel_grid_origin_position),
                self.__voxel_resolution,
            )
            if any(vIndex < 0):
                logging.info(
                    f">> DIFFR: skip out-of-range source voxel of sourceID[{source.id}], sourceLocation[{source.position}]"
                )
                continue

            vIndex = np.rint(vIndex)
            vIndexKey = str(np.array(vIndex, dtype=int).tolist()).strip(
                "[]"
            )  # str(List[int, int, int])
            self.__sourceVoxelDict[source.id] = {vIndexKey: []}

    def __update(self):
        """
        update source voxel dictionary = {"source.id": {svIndex: List[int]}}
            - iSVD (individualSourceVoxelDictionary) = {svIndex: List[int]}
            - sv (source's voxel) = str(List[int, int, int])
        """

        sourcePosition = []
        for srcID, iSVD in self.__sourceVoxelDict.items():

            s: ObjectSource
            for s in self.__sources:
                if s.id == srcID:
                    sourcePosition = s.position
                    break
            isvIndex = np.divide(
                np.array(sourcePosition) - np.array(self.__voxel_grid_origin_position),
                self.__voxel_resolution,
            )
            isvIndex = np.rint(isvIndex)

            for svIndexKey in iSVD.keys():
                svIndex = []
                for svIndexPerAxis in svIndexKey.split(","):
                    svIndex.append(svIndexPerAxis.strip())
                svIndex = np.array(svIndex, dtype=float)

                if np.array_equal(isvIndex, svIndex):
                    svPosition = sourcePosition
                else:
                    svPosition = np.multiply(svIndex, self.__voxel_resolution) + np.array(
                        self.__voxel_grid_origin_position
                    )

                iSVD[svIndexKey] = self.__staticEdgeListGen.getVisibleEdgeIDsFrom(
                    svPosition, backface_culling=True
                )

                for eid in iSVD[svIndexKey]:
                    self.__staticEdgeListGen.edgeList[eid].isVisibleFromStaticSources = True

                logging.info(
                    f">> sourceID: {srcID}, "
                    f"sourcePosition: {svPosition}, "
                    f"sourceVoxelIndex: {svIndex}, "
                    f"visibleEdgeList: {iSVD[svIndexKey]}, "
                    f"nofEdges: {len(iSVD[svIndexKey])}"
                )

    @property
    def sourceVoxelDict(self):
        return self.__sourceVoxelDict

    @property
    def allSourcesStatic(self):
        return self.__allSourcesStatic
