# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import json
import numpy as np
from typing import Union, List, Dict, AnyStr
from .path import Path


def getTypeAsString(obj) -> str:
    if isinstance(obj, int):
        return "int"
    elif isinstance(obj, float):
        return "float"
    elif isinstance(obj, np.int32):
        return "np.int32"
    elif isinstance(obj, np.int64):
        return "np.int64"
    elif isinstance(obj, str):
        return "str"
    elif isinstance(obj, list):
        return "list"
    elif isinstance(obj, tuple):
        return "tuple"
    elif isinstance(obj, dict):
        return "dict"
    elif isinstance(obj, np.ndarray):
        return "np.ndarray"
    else:
        print("Obj's Type: ", type(obj))
        return ""


def typeChecker(obj, typeStr="") -> str:
    nextIteration = True
    try:
        iter(obj)
    except TypeError:
        nextIteration = False

    typeStr += getTypeAsString(obj)
    if not nextIteration:
        return typeStr
    else:
        # for a nested dictionary, it won't call isIterable function recursively
        if isinstance(obj, dict):
            return typeStr

        typeStrforNestedElements = "("
        for nestedElement in obj:
            typeStrforNestedElements += typeChecker(nestedElement, typeStr="") + ", "
        typeStrforNestedElements = typeStrforNestedElements[:-2]
        typeStrforNestedElements += ")"
        return typeStr + typeStrforNestedElements


def NumpyVariableToPythonVariable(data):
    """
    numpy variable has to be changed into python-readable variable
    """
    if isinstance(data, int) or isinstance(data, float):
        return data

    if isinstance(data, np.ndarray):
        data = data.tolist()

    if isinstance(data, list):
        for index in range(len(data)):
            if isinstance(data[index], np.int32) or isinstance(data[index], np.int64):
                data[index] = data[index].item()
            else:
                data[index] = NumpyVariableToPythonVariable(data[index])
    return data


def shortestDistance(p, ls1, ls2):
    """
    calculate the shortest distance from a p point to a line segment defined by ls1 and ls2 points

    @param p: target position in 3D
    @type p: np.array (3,1)
    @param ls1: coordinate of one of the endpoints of the line segment
    @type ls1: np.array (3,1)
    @param ls2: coordinate of the other endpoint of the line segment
    @type ls2:np.array (3,1)
    @return: the shortest distance
    @rtype: float
    """
    lsv = ls2 - ls1
    nlsv = lsv / np.linalg.norm(lsv)

    lp = p - ls1
    dp = np.dot(lp, nlsv)
    proj = dp * nlsv
    orth = lp - proj
    if dp < 0:
        distance = np.linalg.norm(lp)
    elif dp > np.linalg.norm(lsv):
        distance = np.linalg.norm(p - ls2)
    else:
        distance = np.linalg.norm(orth)

    return distance


def writeJSONFile(JSONFileName: AnyStr, dictToWrite: Union[Dict, List, AnyStr]):

    if isinstance(dictToWrite, List):
        listToDict = dictToWrite
        dictToWrite = {}
        for e in listToDict:
            dictToWrite[e.eID] = str(e.getAttributesAll())
    else:
        isPath = False
        for k, v in dictToWrite.items():
            for p in v:
                if isinstance(p, Path):
                    isPath = True
                    break

            if isPath:
                break

        if isPath:
            dictToDict = dictToWrite
            dictToWrite = {}
            for k, pList in dictToDict.items():
                dictToWrite[k] = []
                for p in pList:
                    dictToWrite[k].append(str(p))

    with open(JSONFileName + ".json", "w", encoding="utf-8") as f:
        f.write(json.dumps(dictToWrite, ensure_ascii=False, indent=4))


def writePreprocessedData(diffractionDataOptimizer, eifplusSceneID: AnyStr):

    # writeJSONFile(JSONFileName="edgeList_" + eifplusSceneID,
    #               dictToWrite=diffractionDataOptimizer.edgeList)
    #
    # writeJSONFile(JSONFileName="pathDict_" + eifplusSceneID,
    #               dictToWrite=diffractionDataOptimizer.pathDict)
    #
    # writeJSONFile(JSONFileName="pathDictByEdges_" + eifplusSceneID,
    #               dictToWrite=diffractionDataOptimizer.pathDictByEdgeIds)
    #
    # writeJSONFile(JSONFileName="sourceDict_" + eifplusSceneID,
    #               dictToWrite=diffractionDataOptimizer.sourceVoxelDict)
    #
    # writeJSONFile(JSONFileName="listenerVoxelDict_" + eifplusSceneID,
    #               dictToWrite=diffractionDataOptimizer.listenerVoxelDict)
    #
    # writeJSONFile(JSONFileName="Cell_" + eifplusSceneID,
    #               dictToWrite=diffractionDataOptimizer.cell)
    #
    writeJSONFile(
        JSONFileName="listenerPathDict_" + eifplusSceneID,
        dictToWrite=diffractionDataOptimizer.validPathDict,
    )


if __name__ == "__main__":
    ls1 = np.array([0, 0, 0])
    ls2 = np.array([3, 0, 0])
    for i in range(-4, 8):
        distance = shortestDistance(np.array([i, 3, 0]), ls1, ls2)
        print(f"distance: {distance}")
