# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from .listenervoxeldictgen import *
from .edgelistgenpermesh import *


class ValidPathGen:
    def __init__(
        self,
        sourceVoxelDict: Dict,
        listenerVoxelDict: Dict,
        pathDictDict: Dict,
        pathDictDictByEdgeIds: Dict,
        earLevelVoxelGrid,
        sceneDict: Dict,
    ):

        """
        ValidPathGen provides the attributes and methods to create the dictionary where you can retrieve valid
        paths from fixed source location to a given listener location assuming that no occluders block the propagation
        in-between edges in a given path

        @param sourceVoxelDict:dictionary with source location as voxel index and visible eIds from it
        @param listenerVoxelDict: dictionary with reachable listener location as voxel index and visible eIds from it
        @param pathDictDict: precomputed path dictionary on sceneTrimesh_DIFFR
        @param sceneDict: scene-relevant information container
        """

        logging.info(f"********** ValidPathGen:")

        # Abolish the process for AR or VR with empty static meshes
        self.validPathDict = {}
        if sceneDict["sceneInfo"]["type"] == SceneType.AR:
            return

        if sceneDict["_flag_emtpySceneTrimesh_DIFFR"]:
            raise ValueError(f"<< ERROR: sceneDict['_flag_emtpySceneTrimesh_DIFFR']")

        # Set parallel processing variables
        self.earLevelVoxelGrid: trimesh.voxel.VoxelGrid = earLevelVoxelGrid
        self.sceneTrimeshDIFFR: trimesh.Trimesh = sceneDict["sceneTrimesh_DIFFR"]
        self.staticPathDict = pathDictDict["sceneTrimesh_DIFFR"]
        self.staticPathDictByEdgeIds = pathDictDictByEdgeIds["sceneTrimesh_DIFFR"]

        for srcId, svDictPerSource in sourceVoxelDict.items():
            self.validPathDict[srcId] = {}
            for svIndex, sVEL in svDictPerSource.items():
                for lvIndex, lVEL in tqdm(
                    listenerVoxelDict.items(), desc=f">> Valid-Path-Generation [srcId: {srcId}]"
                ):
                    # distance culling from a static source to the listener voxel
                    svPosition = self.earLevelVoxelGrid.indices_to_points(
                        np.array([int(val) for val in svIndex.split(",")])
                    )

                    lvPosition = self.earLevelVoxelGrid.indices_to_points(
                        np.array([int(val) for val in lvIndex.split(",")])
                    )

                    if (
                        np.linalg.norm(np.subtract(svPosition, lvPosition))
                        > Constants.MaxDistance_FromSource
                    ):
                        continue

                    args = (sVEL, lVEL, svIndex, lvIndex)
                    validPathList = self.getAllValidatedPathList(args)

                    if len(validPathList) == 0:
                        continue

                    self.validPathDict[srcId][lvIndex]: List = []
                    # validPathList.append([[seId, leId], p.pid, attenuation])
                    validPathList.sort(reverse=True, key=self.returnAttenuation)

                    maxPropagationGain = validPathList[0][2]
                    nofPaths = 0
                    for seid_leid, pid, attenuation in validPathList:
                        if attenuation < maxPropagationGain * 0.01:
                            break

                        if nofPaths > 10:
                            break

                        seid = seid_leid[0]  # source visible edge
                        leid = seid_leid[1]  # listene visible edge

                        pList = self.staticPathDict[seid]  # potential path list with seid key
                        for lpIndex, p in enumerate(pList):
                            if pid == p.pid:
                                # pathIndex = [seid, lpIndex]
                                self.validPathDict[srcId][lvIndex].append([seid, lpIndex])
                                nofPaths = nofPaths + 1
                                break

        logging.info(f">> len(self.validPathDictPerStaticSource): {len(self.validPathDict)}")

    def getAllValidatedPathList(self, args) -> List:
        # args = (sVEL, lVEL, svIndex, lvIndex)
        svel, lvel, svIndex, lvIndex = args

        svPosition = self.earLevelVoxelGrid.indices_to_points(
            np.array([int(val) for val in svIndex.split(",")])
        )

        lvPosition = self.earLevelVoxelGrid.indices_to_points(
            np.array([int(val) for val in lvIndex.split(",")])
        )

        # Find the valid path list per possible source-listener pair
        validPathList = []
        for seId in svel:
            for leId in lvel:

                try:
                    pList = self.staticPathDictByEdgeIds[str([seId, leId])]
                except KeyError:
                    continue

                for p in pList:
                    attenuation = p.attenuationByDistance(svPosition, lvPosition)
                    validPathList.append([[seId, leId], p.pid, attenuation])

        return validPathList

    @staticmethod
    def returnAttenuation(item: List):
        return item[2]
