# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#!/usr/bin/env python

from mpegi.voxel import (
    voxelIndexToVoxelCoordinate,
    voxelCoordinateToVoxelIndex,
    voxelIndexToPoint,
    pointToVoxelIndex,
)
from .voxelSubGridGenerator import voxelSubgridGenerator as subgrid
from .multiprocessor import Multiprocessor
import numpy as np
import logging
from tqdm import tqdm
import time
import multiprocessing


class HigherOrderAccelerator:
    """Helper class for accelerating voxelDataGeneratorEarly
    :var trimesh_noAE:          Trimesh, geometry data
    :var reachableVoxels:       VoxelGrid, Voxel grid containing those voxels with are reachable by the listener
    :var surface:               list(surfaceStruct), surfaces
    :var time:                  float, time code
    """

    def __init__(self, trimesh_noAE, reachableVoxels, surface):
        # Import done here because pyembree is required for this class
        # Building pyembree on windows is a pain - not everyone needs this class
        from trimesh.ray.ray_pyembree import RayMeshIntersector

        """ Constructor
        :param trimesh_noAE:        Trimesh, geometry data
        :param reachableVoxels:     VoxelGrid, Voxel grid containing those voxels with are reachable by the listener
        :param surface:             list(surfaceStruct), surfaces
        """

        # reduced Trimesh representation without acoustic environments etc.
        self.trimesh_noAE = trimesh_noAE

        # voxel grid containing those voxels with are reachable by the listener
        self.reachableVoxels = reachableVoxels

        # surfaces
        self.surface = surface

        # get current time
        self.time = time.time()

    def analyze_surface_surface(self, max_dist=-1):
        """Analyze surface-to-surface visibility
        :param max_dist:        float, maximum surface-to-surface distance
        :return:                N x N numpy array, binary visibility mask
        """
        logging.info("ER: analyzing surface-to-surface visibility")

        # reset 2D array which contains the result
        num_surfaces = len(self.surface)
        self.testSurfaceSurface = np.zeros((num_surfaces, num_surfaces), dtype=bool)

        # parallel processing
        self.num_workers = multiprocessing.cpu_count() - 1
        mp = Multiprocessor(
            func_supervisor=self.process_supervisor_ss,
            func_worker=self.process_worker_ss,
            func_collector=self.process_collector_ss,
        )
        mp.run(self.num_workers)
        logging.info("ER: parallel processing finnished")

        # valid entries must be True for both involved surfaces
        self.testSurfaceSurface &= self.testSurfaceSurface.transpose()

        # limit surface-surface rays to maximum distance
        if max_dist > 0:
            self.testSurfaceSurface = self.limit_surface_surface(self.testSurfaceSurface, max_dist)

        num_true = np.sum(self.testSurfaceSurface.astype(int))
        if num_true > 0:
            ratio = (num_surfaces * num_surfaces) / num_true
            ratio_round = np.round(ratio * 100) / 100
            logging.info("ER: expected speedup factor: " + str(ratio_round))

        return self.testSurfaceSurface

    def process_supervisor_ss(self, task_queue):
        """generate bundles of computation tasks; executed by a background process
        :var task_queue:        multiprocessing.queue, task queue
        """
        num_surfaces = len(self.surface)

        # set bundle size
        bundle_size = np.round(num_surfaces / (20 * self.num_workers)).astype(
            int
        )  # try to generate 20 task bundles per worker
        bundle_size = max(bundle_size, 1)  # limit bundle size to the range [1,500]
        bundle_size = min(bundle_size, 500)

        # generate task list
        task_list = list()
        bundle_list = list()
        for surfaceID in range(num_surfaces):
            if len(bundle_list) == bundle_size:
                task_list.append(bundle_list)
                bundle_list = list()
            bundle_list.append(surfaceID)
        if len(bundle_list) > 0:
            task_list.append(bundle_list)

        # add tasks to task queue
        for bundle_list in tqdm(task_list, desc="parallel processing"):
            task_queue.put(bundle_list)

        # send end-of-task signal
        task_queue.put(None)

    def process_collector_ss(self, result_list):
        """collect the result of a computation task and add it to the internal dictionary
        :var result_list:           list([int, numpy.array(N,)]), unique surface index, visibility mask
        """
        for result in result_list:
            surfaceID = result[0]
            mask = result[1]
            self.testSurfaceSurface[surfaceID] = mask

    def process_worker_ss(self, bundle_list):
        """process a computation task; executed by a background process
        :var bundle_list:       list(int), structure which defines the computation task
        :return:                list([int, numpy.array(N,)]), unique surface index, visibility mask
        """
        result_list = []

        for surfaceID in bundle_list:
            mask = self.process_job_surface_surface(surfaceID)
            result_list.append([surfaceID, mask])

        return result_list

    def limit_surface_surface(self, mask, max_dist):
        """limit surface-to-surface distances
        :param mask:            N x N numpy array, binary visibility mask
        :param max_dist:        float, maximum surface-to-surface distance
        :return:                N x N numpy array, binary visibility mask
        """
        mask_limit = mask

        logging.info("ER: limiting surface-to-surface distance")

        num_surfaces = len(self.surface)
        for surf1 in range(num_surfaces):
            for surf2 in [i for i, element in enumerate(mask_limit[surf1]) if element]:
                face_list_1 = self.surface[surf1].faceIdx
                face_list_2 = self.surface[surf2].faceIdx
                subMesh1 = self.trimesh_noAE.submesh([face_list_1], append=True)
                subMesh2 = self.trimesh_noAE.submesh([face_list_2], append=True)

                # first iteration: start at all triangle vertices of the first surface
                vtxSet = set()
                for triangleID in face_list_1:
                    vtx = self.trimesh_noAE.faces[triangleID]
                    vtxSet.add(vtx[0])
                    vtxSet.add(vtx[1])
                    vtxSet.add(vtx[2])
                points1 = self.trimesh_noAE.vertices[list(vtxSet)]
                (points2, distances, triangle_id) = subMesh2.nearest.on_surface(points1)
                minEntry = np.argmin(distances)
                points1 = [points1[minEntry]]
                points2 = [points2[minEntry]]

                # iteratively find shortest distance
                for loop in range(20):
                    old_point = points1[0]
                    (points1, distances, triangle_id) = subMesh1.nearest.on_surface(points2)
                    if np.linalg.norm(old_point - points1[0]) < 0.001:
                        break
                    old_point = points2[0]
                    (points2, distances, triangle_id) = subMesh2.nearest.on_surface(points1)
                    if np.linalg.norm(old_point - points2[0]) < 0.001:
                        break

                if distances[0] > max_dist:
                    mask_limit[surf1][surf2] = False
                    mask_limit[surf2][surf1] = False

        return mask_limit

    def process_job_surface_surface(self, i):
        """Process a given work task
        :var i:                 int, work task ID
        :return:                [Bool, ..., Bool], binary visibility mask
        """
        num_surfaces = len(self.surface)
        mask = np.zeros(num_surfaces, dtype=bool)

        if i >= num_surfaces:
            return mask

        # get all vertices of the i-th surface
        vertex_set = set()
        my_surf = self.surface[i]
        for fIdx in my_surf.faceIdx:
            vertex_set.add(self.trimesh_noAE.faces[fIdx][0])
            vertex_set.add(self.trimesh_noAE.faces[fIdx][1])
            vertex_set.add(self.trimesh_noAE.faces[fIdx][2])
        my_vertex_list = [self.trimesh_noAE.vertices[vIdx] for vIdx in vertex_set]

        # check all surfaces
        for s in range(num_surfaces):
            surf = self.surface[s]
            N0 = surf.N0
            d = surf.d

            # check if any of the vertices is on the positive side of this surface (threshold: 1mm)
            for P in my_vertex_list:
                dist = np.dot(N0, P) - d
                if dist > 0.001:
                    mask[s] = True
                    break

        # self-reflection is not possible
        mask[i] = False

        return mask

    def analyze_voxel_surface(self, src_pos_list=None):
        """Analyze voxel-to-surface visibility
        :param src_pos_list:    list(Float3), list with source positions in Cartesian coordinates
        :return:                dict(int, [Bool, ..., Bool]), binary visibility mask
        """
        logging.info("ER: analyzing voxel-to-surface visibility")

        # store list of source positions
        self.src_pos_list = src_pos_list

        # reset dictionary which contains the result
        self.testVoxelSurface = dict()

        # parallel processing
        self.num_workers = multiprocessing.cpu_count() - 1
        mp = Multiprocessor(
            func_supervisor=self.process_supervisor_vs,
            func_worker=self.process_worker_vs,
            func_collector=self.process_collector_vs,
        )
        mp.run(self.num_workers)
        logging.info("ER: parallel processing finnished")

        # determine speedup factor
        num_entries = 0
        num_true = 0
        for mask in self.testVoxelSurface.values():
            num_entries += len(mask)
            num_true += np.sum(mask.astype(int))
        if num_true > 0:
            ratio = num_entries / num_true
            ratio_round = np.round(ratio * 100) / 100
            logging.info("ER: expected speedup factor: " + str(ratio_round))

        return self.testVoxelSurface

    def process_supervisor_vs(self, task_queue):
        """generate bundles of computation tasks; executed by a background process
        :var task_queue:        multiprocessing.queue, task queue
        """

        # initialize extended list with voxel indices including source voxels
        voxelIDs = [
            voxelCoordinateToVoxelIndex(self.reachableVoxels, vc)
            for vc in self.reachableVoxels.sparse_indices
        ]
        if not self.src_pos_list is None:
            for pos in self.src_pos_list:
                try:
                    src_voxel = pointToVoxelIndex(self.reachableVoxels, pos)
                    vc = voxelIndexToVoxelCoordinate(self.reachableVoxels, src_voxel)
                    voxelIDs.append(src_voxel)
                except:
                    print("ER: skipping out-of-range source voxel at position " + str(pos))

        # set bundle size
        bundle_size = np.round(len(voxelIDs) / (20 * self.num_workers)).astype(
            int
        )  # try to generate 20 task bundles per worker
        bundle_size = max(bundle_size, 1)  # limit bundle size to the range [1,500]
        bundle_size = min(bundle_size, 500)

        # generate task list
        task_list = list()
        bundle_list = list()
        for voxelID in voxelIDs:
            if len(bundle_list) == bundle_size:
                task_list.append(bundle_list)
                bundle_list = list()
            bundle_list.append(voxelID)
        if len(bundle_list) > 0:
            task_list.append(bundle_list)

        # add tasks to task queue
        for bundle_list in tqdm(task_list, desc="parallel processing"):
            task_queue.put(bundle_list)

        # send end-of-task signal
        task_queue.put(None)

    def process_collector_vs(self, result_list):
        """collect the result of a computation task and add it to the internal dictionary
        :var result_list:           list([int, numpy.array(N,)]), unique voxel index, visibility mask
        """
        for result in result_list:
            voxelID = result[0]
            mask = result[1]
            self.testVoxelSurface[voxelID] = mask

    def process_worker_vs(self, bundle_list):
        """process a computation task; executed by a background process
        :var bundle_list:       list(int), structure which defines the computation task
        :return:                list([int, numpy.array(N,)]), unique voxel index, visibility mask
        """
        result_list = []

        for voxelID in bundle_list:
            mask = self.process_job_voxel_surface(voxelID)
            result_list.append([voxelID, mask])

        return result_list

    # def limit_voxel_surface(self, matrix_VS, max_dist):
    #     """ limit voxel-to-surface distances
    #     :param matrix_VS:       numpy.array((num_voxel,num_surface), dtype=bool), binary visibility mask
    #     :param max_dist:        float, maximum voxel-to-surface distance
    #     :return:                numpy.array((num_voxel,num_surface), dtype=bool), binary visibility mask
    #     """
    #     logging.info('ER: limiting voxel-to-surface distance')
    #
    #     num_surface = len(self.surface)
    #     num_voxel = self.reachableVoxels.filled_count
    #     voxel_centers = self.reachableVoxels.points
    #     centers_T = np.transpose(voxel_centers)
    #     xvec = centers_T[0]
    #     yvec = centers_T[1]
    #     zvec = centers_T[2]
    #
    #     # matrices used for computation
    #     matrix_SV = np.transpose(matrix_VS)                         # input
    #     ret_SV = np.zeros((num_surface,num_voxel), dtype=bool)      # output
    #
    #     for s, surf in enumerate(self.surface):
    #         face_list = surf.faceIdx
    #         sub_mesh = self.trimesh_noAE.submesh([face_list], append=True)
    #         bbox = sub_mesh.bounds
    #
    #         # pre-selection of voxels via bounding box
    #         bool_vec = matrix_SV[s]
    #         bool_vec &= (xvec > bbox[0][0] - max_dist)    # xmin
    #         bool_vec &= (xvec < bbox[1][0] + max_dist)    # xmax
    #         bool_vec &= (yvec > bbox[0][1] - max_dist)    # ymin
    #         bool_vec &= (yvec < bbox[1][1] + max_dist)    # ymax
    #         bool_vec &= (zvec > bbox[0][2] - max_dist)    # zmin
    #         bool_vec &= (zvec < bbox[1][2] + max_dist)    # zmax
    #
    #         # get array of boolVec elements which are true
    #         voxel_indices = np.nonzero(bool_vec)[0]
    #         if len(voxel_indices) == 0:
    #             continue
    #
    #         # compute voxel-to-surface distances for pre-selected voxels
    #         voxel_points = voxel_centers[voxel_indices]
    #         (closest_points, distances, triangle_id) = sub_mesh.nearest.on_surface(voxel_points)
    #
    #         # set elements of the returned matrix to true, if distance is less than given threshold
    #         voxel_indices_lim = voxel_indices[distances < max_dist]
    #         ret_SV[s][voxel_indices_lim] = True
    #
    #         #self.plot_VS(sub_mesh, voxel_indices_lim)
    #
    #     return np.transpose(ret_SV)

    def process_job_voxel_surface(self, voxel_index):
        """Process a given work task
        :var voxel_index:       int, unique voxel index
        :return:                [Bool, ..., Bool], binary visibility mask
        """
        num_surfaces = len(self.surface)
        mask = np.zeros(num_surfaces, dtype=bool)

        voxel_center = voxelIndexToPoint(self.reachableVoxels, voxel_index)

        # get all vertices of the i-th reachable voxel
        my_vertex_list = list()
        for pos in subgrid(pos=voxel_center, pitch=self.reachableVoxels.pitch, factor=1, delta=0):
            my_vertex_list.append(pos)

        # check all surfaces
        for s in range(num_surfaces):
            surf = self.surface[s]
            N0 = surf.N0
            d = surf.d

            # check if any of the vertices is on the positive side of this surface (threshold: 1mm)
            for P in my_vertex_list:
                dist = np.dot(N0, P) - d
                if dist > 0.001:
                    mask[s] = True
                    break

        return mask

    def log_message(self, msg):
        """Print a log message with a timestamp
        :var msg:               str, text message
        """
        timestamp = time.time() - self.time
        print("" + format(timestamp, ".3f") + "  " + msg)
