# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import time
from tqdm import tqdm
import numpy as np
import multiprocessing
import psutil
import enum


class WorkerStatus(enum.Enum):
    """Enum class for worker status
    :var uninitialized:         the worker is not initialized and not ready to run
    :var initialized:           the worker is initialized and ready to run
    :var ok:                    the worker processed a given task
    :var error:                 the worker aborted with an error
    :var finished:              the worker finished normally
    """

    uninitialized = -1
    initialized = 0
    ok = 1
    error = 2
    finished = 3


class WorkerResult:
    """Data structure which contains the result of a worker
    :var worker_id:             int, unique worker ID
    :var status:                enum, WorkerStatus which signals the status of the worker
    :var result:                object, result of the worker
    """

    def __init__(self, worker_id, status, result):
        """Constructor
        :param worker_id:           int, unique worker ID
        :param status:              enum, WorkerStatus which signals the status of the worker
        :param result:              object, result of the worker
        """
        self.worker_id = worker_id
        self.status = status
        self.result = result


class Multiprocessor:
    """Helper class for parallelizetion (simple joblib replacement)
    :var func_supervisor:       function, supervisor function with a single multiprocessing.queue parameter which fills the given task queue (None for end-of-queue); runs in the background
    :var func_worker:           function, worker function with a single task parameter which returns a single task result; runs in the background
    :var func_collector:        function, collector function with a single task result parameter; runs in the current process
    """

    def __init__(self, func_supervisor, func_worker, func_collector):
        """Constructor
        :param func_supervisor:     function, supervisor function with a single multiprocessing.queue parameter which fills the given task queue (None for end-of-queue); runs in the background
        :param func_worker:         function, worker function with a single task parameter which returns a single task result; runs in the background
        :param func_collector:      function, collector function with a single task result parameter; runs in the current process
        """
        self.func_supervisor = func_supervisor
        self.func_worker = func_worker
        self.func_collector = func_collector

    def run(self, num_worker=None):
        """Execute the parallel processing jobs
        :param num_worker:          int, number of worker processes
        """

        if num_worker is None:
            num_worker = multiprocessing.cpu_count() - 1

        self.worker_status = np.full(num_worker, WorkerStatus.uninitialized.value)

        # setup work queues
        task_queue = multiprocessing.Queue(maxsize=num_worker)
        result_queue = multiprocessing.Queue(maxsize=num_worker)

        # start the worker processes
        procs = []
        task_lock = multiprocessing.Lock()
        for worker_id in tqdm(range(num_worker), desc="spawning workers"):
            proc = multiprocessing.Process(
                target=self._worker, args=(task_queue, result_queue, worker_id, task_lock)
            )
            proc.start()
            procs.append(proc)
            res = result_queue.get()
            if res.status != WorkerStatus.initialized:
                raise RuntimeError("Worker was not properly initialized")
            else:
                self.worker_status[res.worker_id] = res.status.value

        # start the supervisor process
        proc = multiprocessing.Process(target=self._supervisor, args=(task_queue, num_worker))
        proc.start()
        procs.append(proc)

        # collect results until all workers finished their jobs
        while not np.all(self.worker_status == WorkerStatus.finished.value):
            self.__collector(result_queue)

        # drain task queue
        task = task_queue.get()
        if not task is None:
            raise RuntimeError("Multiprocessor: leftover task detected!!!")

        if not task_queue.empty():
            raise RuntimeError("Multiprocessor: Task queue is NOT empty.")

        if not result_queue.empty():
            raise RuntimeError("Multiprocessor: Result queue is NOT empty.")

        # join background processes
        for proc in tqdm(multiprocessing.active_children(), desc="terminating processes"):
            proc.terminate()

    def _worker(self, task_queue, result_queue, worker_id, task_lock):
        """Worker function running in the backround; calls external worker function
        :var task_queue:            multiprocessing.queue, task queue
        :var result_queue:          multiprocessing.queue, result queue
        :var worker_id:             int, worker ID
        :var task_lock:             multiprocessing.Lock, lock for accessing the task queue
        """
        # pin this process to a certain CPU core
        self.__pin_cpu_core(worker_id)

        # signal that we are ready
        result_queue.put(WorkerResult(worker_id, WorkerStatus.initialized, None))

        fwd_except = None

        try:
            while True:
                task_lock.acquire()  # critical section which shall not be interrupted by other workers
                task = task_queue.get()
                if task is None:
                    task_queue.put(None)  # re-insert end-of-task signal and finnish up
                    task_lock.release()
                    break
                task_lock.release()

                # call the external worker function and add the result to the result queue
                result = self.func_worker(task)
                result_queue.put(WorkerResult(worker_id, WorkerStatus.ok, result))
        except BaseException as err:
            result_queue.put(WorkerResult(worker_id, WorkerStatus.error, err))
            return

        # signal that we are done
        result_queue.put(WorkerResult(worker_id, WorkerStatus.finished, None))

    def _supervisor(self, task_queue, num_worker):
        """Supervisor function running in the backround; calls external supervisor function
        :var task_queue:            multiprocessing.queue, task queue
        :var num_worker:            int, number of workers
        """
        # pin this process to a certain CPU core
        self.__pin_cpu_core(num_worker)

        self.func_supervisor(task_queue)

    def __collector(self, result_queue):
        """Collector function; calls external collector function
        :var result_queue:          multiprocessing.queue, result queue
        """
        res = result_queue.get()
        if res.status == WorkerStatus.ok:
            self.func_collector(res.result)
        elif res.status == WorkerStatus.error:
            for proc in tqdm(multiprocessing.active_children(), desc="killing processes"):
                proc.terminate()  # kill all processes to avoid misleading additional exceptions from workers
            raise RuntimeError(res.result)
        elif res.status == WorkerStatus.finished:
            self.worker_status[res.worker_id] = res.status.value
        else:
            raise RuntimeError("Received unexpected worker status message")

    def __pin_cpu_core(self, n):
        """pin process to a certain (virtual) CPU core
        :var n:                     int, CPU core
        """
        # pin this process to a certain CPU core
        cpu_count = psutil.cpu_count()
        my_cpu_core = int(np.mod(n, cpu_count))
        my_process = psutil.Process()
        try:
            my_process.cpu_affinity([my_cpu_core])
        except AttributeError:
            # Setting CPU affinity is not available on macOS, see
            # https://psutil.readthedocs.io/en/latest/index.html#psutil.Process.cpu_affinity
            pass
