# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np


class RayData:
    """Data structure for RayTracing.
    :var surf_seq_list:    numpy.array(N,O), surface reflection sequences (padded with -1)
    :var L_list:            numpy.array(N,3), listener positions
    :var max_order:         int, maximum reflection order
    :var num_ray:           int, number of rays
    :var order:             numpy.array(N,), reflection orders
    :var S_img_list:        numpy.array(N,3), image source positions
    :var seq_list:          numpy.array(N,O), reflection sequences (padded with -1)
    :var src_pos:           numpy.array(3,), source position
    :var visible:           numpy.array(N,), outcome of the visibility test based on ray tracing
    """

    def __init__(
        self,
        src_pos,
        src_voxel,
        listener_pos_list,
        listener_voxel,
        seqIt,
        testVoxelSurface,
        testSurfaceSurface,
    ):
        """Constructur
        :param src_pos:             numpy.array(3,), source position
        :param src_voxel:           int, unique voxel ID of the source position
        :param listener_pos_list:   numpy.array(N,3), Cartesian coordinates of the listener positions
        :param listener_voxel:      int, unique voxel ID of the listener positions
        :param seqIt:               SequenceIterator, iterator over all possible reflection sequences
        :param testVoxelSurface:    dict(int, [Bool, ..., Bool]), binary voxel-to-surface visibility mask
        :param testSurfaceSurface:  numpy.array(S,S), binary surface-to-surface visibility mask
        """
        self.src_pos = src_pos
        self.L_list = np.zeros((0, 3))
        self.S_img_list = np.zeros((0, 3))
        self.max_order = seqIt.max_order
        self.order = np.zeros((0,), dtype=int)
        self.seq_list = np.zeros((0, seqIt.max_order), dtype=int)

        num_listener_pos = listener_pos_list.shape[0]

        for seq in seqIt:
            # check if listener is on the correct side of the first reflecting surface and if source is on the correct side of the last reflecting surface
            if (
                not testVoxelSurface[listener_voxel][seq[0]]
                or not testVoxelSurface[src_voxel][seq[-1]]
            ):  # check if source is on the correct side of the last reflecting surface
                continue

            # check if the 2 involved surfaces are on the correct side
            if len(seq) == 2 and not testSurfaceSurface[seq[0]][seq[1]]:
                continue

            # append ray starting points
            self.L_list = np.append(self.L_list, listener_pos_list, axis=0)

            # append ray directions: D = S_img - M
            S_img = seqIt.getImage(src_pos)
            S_img_mat = np.tile(S_img.reshape((1, 3)), (num_listener_pos, 1))
            self.S_img_list = np.append(self.S_img_list, S_img_mat, axis=0)

            # append ray order
            order = np.repeat(len(seq), num_listener_pos)
            self.order = np.append(self.order, order)

            # append reflection sequence
            seq_padded = -np.ones((1, seqIt.max_order), dtype=int)
            seq_padded[0][0 : len(seq)] = np.reshape(seq, (1, len(seq)))
            seq_padded = np.repeat(seq_padded, num_listener_pos, axis=0)
            self.seq_list = np.append(self.seq_list, seq_padded, axis=0)

            pass

        self.num_ray = self.order.shape[0]
        self.surf_seq_list = -np.ones((self.num_ray, seqIt.max_order), dtype=int)
        self.visible = np.zeros((self.num_ray,), dtype=bool)

    def reset_visible(self):
        """Reset raytracing data structure and limit it to rays which are marked as visible"""
        keep = np.nonzero(self.visible)[0]
        self.num_ray = len(keep)
        self.L_list = self.L_list[keep]
        self.S_img_list = self.S_img_list[keep]
        self.surf_seq_list = self.surf_seq_list[keep]
        self.order = self.order[keep]
        self.seq_list = self.seq_list[keep]
        self.visible = np.zeros((self.num_ray,), dtype=bool)
