# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np


class SequenceIterator:
    """Class for iterating over all possible reflection sequences of a geometric model (up to a given order)
    :var alphabet:              numpy.array(N), array with allowed surface indices
    :var counter:               list(int), N-digit counter
    :var max_order:             int, maximum reflection order
    :var start_order:           int, initial reflection order
    :var surface:               list(surfaceStruct), surfaces with co-planar faces
    :var tform:                 numpy.array(4,4), transformation matrix
    :var tform_cache:           dict(int, numpy.array(4,4)), cache for transformation matrices for first order reflections
    """

    def __init__(self, surface, max_order, alphabet=None, start_order=1):
        """Constructor
        :param surface:         list(surfaceStruct), surfaces with co-planar faces
        :param max_order:       int, maximum reflection order
        :param alphabet:        numpy.array(N), array with allowed surface indices (default: all surfaces)
        :param start_order:     int, initial reflection order
        """
        if alphabet is None:
            num_surface = len(surface)
            self.alphabet = np.arange(num_surface)
        else:
            self.alphabet = np.array(alphabet)
        num_alphabet = len(self.alphabet)

        self.max_order = max_order
        self.start_order = start_order
        self.surface = surface
        self.reset()

    def __iter__(self):
        return self

    def __next__(self):
        """next iteration: next reflection sequence"""
        num_alphabet = len(self.alphabet)
        order = len(self.counter)

        if order == 0:
            self.counter = [0]
            self.update_tform()
            return self.alphabet[self.counter]

        # iterate the reflection sequence
        self.counter[order - 1] += 1
        for n in range(order - 1, -1, -1):
            if self.counter[n] >= num_alphabet:
                self.counter[n] = 0
                if n > 0:
                    self.counter[n - 1] += 1
                else:
                    self.counter.insert(0, 0)
                    if len(self.counter) > self.max_order:
                        raise StopIteration  # signals "the end"
                    order += 1
                    n += 1

        # skip invalid sequences
        for n in range(order - 1):
            if self.counter[n] == self.counter[n + 1]:
                return self.__next__()

        self.update_tform()

        return self.alphabet[self.counter]

    def reset(self):
        """reset the object to initial state"""
        num_alphabet = len(self.alphabet)
        counter_len = self.start_order - 1
        self.counter = [num_alphabet - 1] * counter_len
        self.tform = np.eye(4, dtype=float)
        self.tform_cache = dict()

    def update_tform(self):
        """update the transformation matrix"""
        it = iter(self.alphabet[self.counter])
        tform = self.get_mirror_tform(next(it))
        for idx in it:
            tform = np.matmul(self.get_mirror_tform(idx), tform)
        self.tform = tform

    def get_mirror_tform(self, idx):
        """get a transformation matrix for first order reflections (cached)
        :param idx:             int, surface index
        :returns:               numpy.array(4,4), transformation matrix
        """
        if idx in self.tform_cache:
            return self.tform_cache[idx]

        tform = np.zeros((4, 4), dtype=float)
        N0 = self.surface[idx].N0
        d = self.surface[idx].d
        O = (2 * d) * N0
        tform[0, :-1] = (2 * (d - N0[0])) * N0 - O  # A - O
        tform[1, :-1] = (2 * (d - N0[1])) * N0 - O  # B - O
        tform[2, :-1] = (2 * (d - N0[2])) * N0 - O  # C - O
        tform[3, :-1] = O
        tform += np.eye(4)
        self.tform_cache[idx] = tform
        return tform

    def getImage(self, x):
        """Returns the mirrored versions of the given points
        :param x:               numpy.array(3,:), points in Cartesian coordinates
        :returns:               numpy.array(3,:), mirrored points in Cartesian coordinates
        """
        x = np.array(x)
        nd = x.ndim
        if nd == 1:  # convert column vector to matrix
            N = 1
            x = x.reshape((1, 3))
        else:
            N = len(x)

        x_homo = np.c_[x, np.ones(N)]
        y_homo = np.matmul(x_homo, self.tform)
        if nd == 1:  # convert to matrix if input is a vector
            y = y_homo[:, 0:3].reshape(3)
        else:
            y = y_homo[:, 0:3]

        return y
