# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import copy
import json


class VoxelDataEarlyDict:
    """Lookup Table for Early Reflections. This variant of the VoxelDataEarly class uses a nested dictionary.
    :var lut:               list(list(dict(set(tuple())))), lookup table: lut[listener_voxel_ID][source_voxel_ID][tuple([surf, ..., surf])]
    """

    def __init__(self):
        """Constructur"""

        # initialize lookup table
        self.clear()

    def clear(self):
        self.lut = {}

    def union(self, other):
        """Union of two lookup table instances
        :param other:           VoxelDataEarlyDict, other lookup table object that will be merged into this lookup table object
        """
        for voxel_combi in sorted(other.lut.keys()):
            if self.lut.get(voxel_combi, {}) == {}:
                self.lut[voxel_combi] = copy.deepcopy(other.lut[voxel_combi])
            else:
                l = voxel_combi[0]
                s = voxel_combi[1]
                for seq_surf in other.lut[voxel_combi]:
                    self.append(l, s, seq_surf)

    def copy(self):
        """Create a deep copy
        :return:            VoxelDataEarlyDict, deep copy
        """
        ret = VoxelDataEarlyDict()
        for voxel_combi in self.lut.keys():
            ret.lut[voxel_combi] = copy.deepcopy(self.lut[voxel_combi])
        return ret

    def to_dict(self):
        """Convert internal data structure to a dictionary that can easily be exported to JSON"""
        ret = dict()

        for voxel_combi in sorted(self.lut.keys()):
            l = voxel_combi[0]
            s = voxel_combi[1]
            str_l = "L_" + str(l)
            str_s = "S_" + str(s)
            if ret.get(str_l, {}) == {}:
                ret[str_l] = {}
            if ret[str_l].get(str_s, []) == []:
                ret[str_l][str_s] = []
            reflections = []
            for seq_surf in sorted(sorted(self.lut[voxel_combi]), key=len):
                reflections.append(str(list(seq_surf)))
            ret[str_l][str_s] = reflections
        return ret

    def to_json(self):
        """Convert internal data structure to a JSON formatted string
        :return:            str, json export
        """
        str_json = json.dumps(self.to_dict(), indent=4)
        str_json = str_json.replace('"[', "[")
        str_json = str_json.replace(']"', "]")
        return str_json

    def from_json(self, str_json):
        """Convert a JSON formatted string to internal data structure
        :var str_json:      str, json import
        """
        self.clear()
        database_json = json.loads(str_json)
        for key_l in database_json.keys():
            assert key_l[0:2] == "L_"
            l = int(key_l[2:])
            for key_s in database_json[key_l].keys():
                assert key_s[0:2] == "S_"
                s = int(key_s[2:])
                for seq_surf in database_json[key_l][key_s]:
                    self.append(l, s, seq_surf)

    def append(self, voxel_listener, voxel_src, seq_surf):
        """Append an entry to the lookup table
        :param voxel_listener:      int, voxel ID of the listener position
        :param voxel_src:           int, voxel ID of the source position
        :param seq_surf:            tuple(int), reflection sequence: surface IDs
        """
        seq_surf = tuple(seq_surf)

        voxel_combi = (voxel_listener, voxel_src)
        if self.lut.get(voxel_combi, set()) == set():
            self.lut[voxel_combi] = set()
        if len(seq_surf) != 0:
            self.lut[voxel_combi].add(seq_surf)

    def save(self, filename):
        """Save lookup table data to a text file
        :param filename:            str, filename
        """
        with open(filename, "w") as outfile:
            outfile.write(str(self.lut))

    def load(self, filename):
        """Load lookup table data from a text file
        :param filename:            str, filename
        """
        with open(filename) as infile:
            self.lut = eval(infile.read())

    def is_complete(self, voxel_listener, voxel_src, seq_surf):
        """Returns if the given lookup table entry is complete and contains all possible combinations of connected surfaces
        :param voxel_listener:      int, voxel ID of the listener position
        :param voxel_src:           int, voxel ID of the source position
        :param seq_surf:            tuple(int), reflection sequence: surface IDs, will be derived from csurf2surf map if empty
        :return:                    bool, is the given lookup table entry complete and contains all possible combinations of connected surfaces?
        """
        seq_surf = tuple(seq_surf)

        num_max = 1
        for s in list(seq_surf):
            num_max *= self.surf_count[s]
        try:
            num_found = len(self.lut[(voxel_listener, voxel_src)][seq_surf])
            return num_found == num_max
        except KeyError:
            return False

    def __str__(self):
        return str(self.to_dict())
