# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#!/usr/bin/env python
from tqdm import tqdm
import cProfile
import io
import logging
import multiprocessing
import numpy as np
import pstats
import time

from mpegi.geometry import transforms
from mpegi.geometry.trimesh import convertToTrimesh
from mpegi.geometry.preproc_surface import preprocSurface, mapFace2Surface
from mpegi.voxel import (
    voxelIndexToVoxelCoordinate,
    voxelCoordinateToVoxelIndex,
    voxelCoordinateToPoint,
    voxelIndexToPoint,
    pointToVoxelIndex,
)

from .voxelDataEarlyDict import VoxelDataEarlyDict
from .voxelSubGridGenerator import voxelSubgridGenerator as subgrid
from .sequenceIterator import SequenceIterator
from .higherOrderAccelerator import HigherOrderAccelerator
from .rayData import RayData
from .multiprocessor import Multiprocessor


def profile(fnc):
    """A decorator that uses cProfile to profile a function"""

    def inner(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()
        retval = fnc(*args, **kwargs)
        pr.disable()
        s = io.StringIO()
        sortby = "cumtime"
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
        pr.dump_stats("profile.pstats")
        return retval

    return inner


class VoxelDataGeneratorEarly:
    """Raytracer for accelerating visibility tests using voxels
    :var face2surf:             [int, ..., int], map: triangle (face) index -> surface index
    :var grid_factor:           int, over-sampling factor of the voxel grid
    :var intersector:           pyembree ray-mesh intersector
    :var is_small_scene:        bool, is the extend of the mesh greater than max_dist_triangle?
    :var jobs:                  list(ComputationJob), list of computation jobs
    :var max_dist_source:       dict(int: float), source culling distance in m for a specific reflection order
    :var max_dist_triangle:     dict(int: float), triangle culling distance in m for a specific reflection order
    :var max_order_indoor:      int, indoor maximum reflection order
    :var max_order_outdoor:     int, outdoor maximum reflection order
    :var N0_list:               numpy.array(N,3), list of all surface normals
    :var reachableVoxels:       VoxelGrid, voxels reachable by the listener
    :var src_pos_list:          list(Float3), list with source positions in Cartesian coordinates
    :var src_voxel_list:        list(int), list of source voxel IDs
    :var subgrid_offset:        numpy.array(N,3), voxel sub-grid positions
    :var subgrid_offset_coarse: numpy.array(N,3), coarse voxel sub-grid positions for source = listener (e.g. finger snaps)
    :var surface:               list(surfaceStruct), surfaces
    :var testSurfaceSurface:    N x N Bool array, possible surface-to-surface visibility
    :var testVoxelSurface:      dict(N x 1 Bool array), possible voxel-to-surface visibility
    :var time:                  float, timestamp used by log_message()
    :var trimesh_noAE:          Trimesh, geometry data
    :var vde:                   VoxelDataEarlyDict, voxel data for Early Reflections
    :var vi2point:              dict(int, Float3), map voxel IDs to voxel center positions
    """

    class ComputationJob:
        """Data structure for computation jobs
        :var src_pos:           Float3, source position in Cartesian coordinates
        :var listener_voxel:    int, listener voxel ID
        :var order:             int, reflection order
        :var src_is_listener:   bool, 0:normal computation job, 1:finger snap or listener's voice
        """

        def __init__(self, src_pos, listener_voxel, order, src_is_listener=False):
            self.src_pos = src_pos
            self.listener_voxel = listener_voxel
            self.order = order
            self.src_is_listener = src_is_listener

    def __init__(self, args, artifacts, reachable_voxels, src_pos_list, grid_factor=4):
        # Import done here because pyembree is required for this class
        # Building pyembree on windows is a pain - not everyone needs this class
        from trimesh.ray.ray_pyembree import RayMeshIntersector

        """ Constructor
        :param args:                dict, input parameters
        :param artifacts:           dict, output parameters (of this component or others)
        :param reachable_voxels:    trimesh.VoxelGrid, voxels that the listener can reach
        :param src_pos_list:        list(Float3), list with source positions in Cartesian coordinates
        :param grid_factor:         int, over-sampling factor of the voxel grid
        """

        # artifacts used by this component
        trimesh_noAE = artifacts["scene_trimesh_ER"]
        acoustic_environments = artifacts["acoustic_environments"]
        acoustic_materials = artifacts["acoustic_materials"]
        meshes = artifacts["meshes"]

        # command line parameters used by this component
        max_order_indoor = args["max_order_ER_indoor"]
        max_order_outdoor = args["max_order_ER_outdoor"]
        max_dist_triangle_1 = args["max_dist_ER_triangle_1"]
        max_dist_triangle_2 = args["max_dist_ER_triangle_2"]
        max_dist_source_1 = args["max_dist_ER_source_1"]
        max_dist_source_2 = args["max_dist_ER_source_2"]

        # reduced Trimesh representation without acoustic environments etc.
        self.trimesh_noAE = trimesh_noAE

        # generate transparency filter
        self.__generate_transparency_filter(acoustic_materials)

        # run geometry pre-processor to detect surfaces
        self.surface = preprocSurface(self.trimesh_noAE)

        # remove tiny surfaces
        self.__cleanup_surfaces()

        # generate map: triangle (face) ID -> surface ID
        self.face2surf = np.array(mapFace2Surface(self.trimesh_noAE, self.surface))
        self.face2surf_unfiltered = np.array(mapFace2Surface(self.trimesh_noAE, self.surface))

        # distance culling limits
        self.max_order_indoor = max_order_indoor
        self.max_order_outdoor = max_order_outdoor
        self.max_dist_triangle = {1: max_dist_triangle_1, 2: max_dist_triangle_2}
        self.max_dist_source = {1: max_dist_source_1, 2: max_dist_source_2}

        # voxel sub-positions
        self.grid_factor = grid_factor

        # add voxel data
        self.reachableVoxels = reachable_voxels
        # self.plot_voxels()

        src_pos_list = np.array(src_pos_list)
        self.src_pos_list = src_pos_list
        self.src_voxel_list = list()
        for pos in src_pos_list:
            try:
                src_voxel = pointToVoxelIndex(self.reachableVoxels, pos)
                vc = voxelIndexToVoxelCoordinate(self.reachableVoxels, src_voxel)
                self.src_voxel_list.append(src_voxel)
            except:
                logging.info("ER: skipping out-of-range source voxel at position " + str(pos))

        # look-up-table for voxel center positions
        self.vi2point = {}
        reachableVoxelIndices = sorted(
            [
                voxelCoordinateToVoxelIndex(self.reachableVoxels, vc)
                for vc in self.reachableVoxels.sparse_indices
            ]
        )
        for vi in reachableVoxelIndices:
            self.vi2point[vi] = voxelIndexToPoint(self.reachableVoxels, vi)

        # initialize voxel database
        self.vde = VoxelDataEarlyDict()

        # add empty voxel data entries for all source and listener voxels
        for src_pos in self.src_pos_list:
            src_voxel = pointToVoxelIndex(self.reachableVoxels, src_pos)
            for listener_voxel in reachableVoxelIndices:
                self.vde.append(listener_voxel, src_voxel, ())

        # initialize Embree ray-mesh-intersector; ignore transparent triangles
        self.intersector, self.face2surf = self.__generate_global_intersector()

        # initialize higher-order accelerator
        booster = HigherOrderAccelerator(self.trimesh_noAE, self.reachableVoxels, self.surface)
        self.testSurfaceSurface = booster.analyze_surface_surface()
        self.testVoxelSurface = booster.analyze_voxel_surface(src_pos_list)

        # generate voxel sub-grid positions
        self.__generate_subpos()

        # generate bounding boxes for all triangles
        geo_extend = trimesh_noAE.bounds[1] - trimesh_noAE.bounds[0]
        self.is_small_scene = np.all(geo_extend < max_dist_triangle_1)
        if not self.is_small_scene:
            self.__generate_triangle_bbox()

        # store list of normal vectors
        self.N0_list = np.array([surf.N0 for surf in self.surface])

        # extract boundaries of the acoustic environments
        geometry_dict = dict()
        for ae_mesh in meshes:
            geometry_dict[ae_mesh.id] = ae_mesh
        geometry_list = [
            geometry_dict[env.region]
            for env in acoustic_environments
            if env.region in geometry_dict
        ]
        self.ae_trimesh_list = convertToTrimesh(geometry=geometry_list, includeTransparent=True)
        for geometry, trimesh in zip(geometry_list, self.ae_trimesh_list):
            # Convert geometry to trimesh and apply translation
            tform = transforms.transformMatrixTRS_4x4(
                geometry.position, np.deg2rad([0, 0, 0]), [1, 1, 1]
            )
            trimesh.apply_transform(tform)

        self.time = time.time()

        logging.info("ER: voxelDataGeneratorEarly initialized")

    def process_parallel(self, num_workers=0):
        """High-level function which generates the voxel data for early reflections using parallel-processing
        :var num_workers:       int, number of parallel processes (default: all available CPU threads)
        """
        if num_workers < 1:
            num_workers = multiprocessing.cpu_count() - 1
        logging.info("ER: generating voxel data for early reflections using process_parallel()")

        self.ray_counter = 0
        self.ray_segment_counter = 0

        self.num_workers = multiprocessing.cpu_count() - 1
        mp = Multiprocessor(
            func_supervisor=self.process_supervisor,
            func_worker=self.process_worker,
            func_collector=self.process_collector,
        )
        mp.run(num_workers)
        logging.info("ER: parallel processing finnished")
        logging.info("ER: process_parallel() finnished")

    @profile
    def process_profile(self):
        """High-level function which profiles the generation of the voxel data for early reflections"""

        logging.info("ER: generating voxel data for early reflections using process_profile()")

        self.ray_counter = 0
        self.ray_segment_counter = 0

        # generate task list for static sound sources
        task_list = list()
        lpos_T = np.transpose(self.reachableVoxels.points)
        for src_pos in self.src_pos_list:
            dist = np.sqrt(
                np.square(lpos_T[0] - src_pos[0])
                + np.square(lpos_T[1] - src_pos[1])
                + np.square(lpos_T[2] - src_pos[2])
            )
            for order in self.max_dist_source.keys():
                for listener_vc in self.reachableVoxels.sparse_indices[
                    dist < self.max_dist_source[order]
                ]:  # loop over reachable voxels which are close to the listener
                    listener_pos = voxelCoordinateToPoint(self.reachableVoxels, listener_vc)
                    if order <= self.__get_location_based_order(listener_pos):
                        listener_voxel = voxelCoordinateToVoxelIndex(
                            self.reachableVoxels, listener_vc
                        )
                        task_list.append(
                            VoxelDataGeneratorEarly.ComputationJob(src_pos, listener_voxel, order)
                        )

        # generate task list for finger snaps
        src_is_listener = True
        for order in self.max_dist_source.keys():
            for listener_vc in self.reachableVoxels.sparse_indices:  # loop over reachable voxels
                src_pos = voxelCoordinateToPoint(self.reachableVoxels, listener_vc)
                if order <= self.__get_location_based_order(src_pos):
                    listener_voxel = voxelCoordinateToVoxelIndex(self.reachableVoxels, listener_vc)
                    task_list.append(
                        VoxelDataGeneratorEarly.ComputationJob(
                            src_pos, listener_voxel, order, src_is_listener
                        )
                    )

        for task in tqdm(task_list, desc="sequential processing"):
            result = self.process_worker([task])
            self.process_collector(result)

        logging.info("ER: process_profile() finnished")

    def process_supervisor(self, task_queue):
        """generate bundles of computation tasks; executed by a background process
        :var task_queue:        multiprocessing.queue, task queue
        """
        # generate (unstructured) task list for static sound sources
        unstructured_list = list()
        lpos_T = np.transpose(self.reachableVoxels.points)
        for src_pos in self.src_pos_list:
            dist = np.sqrt(
                np.square(lpos_T[0] - src_pos[0])
                + np.square(lpos_T[1] - src_pos[1])
                + np.square(lpos_T[2] - src_pos[2])
            )
            for order in self.max_dist_source.keys():
                for listener_vc in self.reachableVoxels.sparse_indices[
                    dist < self.max_dist_source[order]
                ]:  # loop over reachable voxels which are close to the listener
                    listener_pos = voxelCoordinateToPoint(self.reachableVoxels, listener_vc)
                    if order <= self.__get_location_based_order(listener_pos):
                        listener_voxel = voxelCoordinateToVoxelIndex(
                            self.reachableVoxels, listener_vc
                        )
                        unstructured_list.append(
                            VoxelDataGeneratorEarly.ComputationJob(src_pos, listener_voxel, order)
                        )

        # create bundles of computation jobs as task list
        bundle_size = max(
            np.round(len(unstructured_list) / (50 * self.num_workers)).astype(int), 1
        )  # try to generate 50 task bundles per worker
        task_list = list()
        bundle_list = list()
        for computation_job in unstructured_list:
            if len(bundle_list) == bundle_size:
                task_list.append(bundle_list)
                bundle_list = list()
            bundle_list.append(computation_job)
        del unstructured_list

        # add final task bundle to task list
        if len(bundle_list) > 0:
            task_list.append(bundle_list)
            bundle_list = list()

        # generate task list for finger snaps
        src_is_listener = True
        # adapt number of tasks per bundle (coarser subgrid positions)
        bundle_size *= np.round(self.subgrid_offset.shape[0] / self.subgrid_offset_coarse.shape[0])
        for order in self.max_dist_source.keys():
            for listener_vc in self.reachableVoxels.sparse_indices:  # loop over reachable voxels
                src_pos = voxelCoordinateToPoint(self.reachableVoxels, listener_vc)
                if order <= self.__get_location_based_order(src_pos):
                    if len(bundle_list) == bundle_size:
                        task_list.append(bundle_list)
                        bundle_list = list()
                    listener_voxel = voxelCoordinateToVoxelIndex(self.reachableVoxels, listener_vc)
                    bundle_list.append(
                        VoxelDataGeneratorEarly.ComputationJob(
                            src_pos, listener_voxel, order, src_is_listener
                        )
                    )

        # add final task bundle to task list
        if len(bundle_list) > 0:
            task_list.append(bundle_list)

        # add tasks to task queue
        for bundle_list in tqdm(task_list, desc="parallel processing"):
            task_queue.put(bundle_list)

        # send end-of-task signal
        task_queue.put(None)

    def process_worker(self, bundle_list):
        """process a computation task; executed by a background process
        :var bundle_list:       list(ComputationJob), structure which defines the computation task
        :return:                [int, int, int, int, list()], number of rays, number of ray segments, listener voxel index, source voxel index, list of visible "seq" sequences
        """
        result_list = list()
        for task in bundle_list:
            src_pos = task.src_pos
            listener_voxel = task.listener_voxel
            order = task.order
            src_voxel = pointToVoxelIndex(self.reachableVoxels, src_pos)
            listener_pos = voxelIndexToPoint(self.reachableVoxels, listener_voxel)
            if task.src_is_listener:
                num_listener_pos = self.subgrid_offset_coarse.shape[0]
                listener_pos_list = (
                    np.tile(listener_pos.reshape((1, 3)), (num_listener_pos, 1))
                    + self.subgrid_offset_coarse
                )
            else:
                num_listener_pos = self.subgrid_offset.shape[0]
                listener_pos_list = (
                    np.tile(listener_pos.reshape((1, 3)), (num_listener_pos, 1))
                    + self.subgrid_offset
                )

            # generate PyEmbree Intersector for local sub-mesh which contains triangles close to the listener position
            if self.is_small_scene:
                # do not use sub-mesh for small scenes
                local_intersector, local_face2surf = self.intersector, self.face2surf
                local_surf_list = np.arange(len(self.surface))
            else:
                local_intersector, local_face2surf = self.__generate_local_intersector(
                    listener_pos, src_pos, order
                )
                local_surf_list = np.unique(local_face2surf[local_face2surf != -1].astype(int))

            # generate configuration data for ray tracing (all combinations of voxel sub-grid positions and reduced list of reflection sequences)
            seqIt = SequenceIterator(self.surface, order, local_surf_list, order)
            ray_data = RayData(
                src_pos,
                src_voxel,
                listener_pos_list,
                listener_voxel,
                seqIt,
                self.testVoxelSurface,
                self.testSurfaceSurface,
            )

            # perform ray tracing with local sub-mesh
            ray_counter, ray_segment_counter = self.__raytrace(
                ray_data, local_intersector, local_face2surf
            )

            # perform ray tracing again with complete mesh and rays that were visible for the local mesh
            if not self.is_small_scene:
                ray_data.reset_visible()
                _ray_counter, _ray_segment_counter = self.__raytrace(
                    ray_data, self.intersector, self.face2surf
                )
                ray_counter += _ray_counter
                ray_segment_counter += _ray_segment_counter

            # append visible reflection sequences
            seq_set = set()
            ray_data.reset_visible()
            for order, seq_surf_padded in zip(ray_data.order, ray_data.surf_seq_list):
                seq_surf = seq_surf_padded[0:order]
                seq_set.add(tuple(seq_surf))

            result_list.append(
                [ray_counter, ray_segment_counter, listener_voxel, src_voxel, list(seq_set)]
            )

        return result_list

    def process_collector(self, result_list):
        """collect the result of a computation task and add it to the internal database
        :var result_list:           list([int, int, int, int, list()]), number of rays, number of ray segments, listener voxel index, source voxel index, list of visible "seq" sequences
        """
        for result in result_list:
            self.ray_counter += result[0]
            self.ray_segment_counter += result[1]
            listener_voxel = result[2]
            src_voxel = result[3]
            seq_list = result[4]
            for seq_surf in seq_list:
                self.vde.append(listener_voxel, src_voxel, np.array(seq_surf))

    def __get_location_based_order(self, listener_pos):
        """Check if the given position is inside an acoustic environment and return the specified maximum reflection order
        :var listener_pos:      Float3, listener position in Cartesian coordinates for which a local sub-mesh is generated
        :return:                int, maximum reflection order
        """
        points = listener_pos.reshape((1, 3))
        for trimesh in self.ae_trimesh_list:
            if trimesh.contains(points)[0]:
                return self.max_order_indoor

        return self.max_order_outdoor

    def __raytrace(self, ray_data, intersector, face2surf):
        """Perform ray tracing
        :var ray_data:          RayData, ray tracing configuration
        :var intersector:       RayMeshIntersector, PyEmbree Intersector for reduced local mesh
        :var face2surf:         numpy.array(N,), map: triangle (face) index -> surface index
        :return:                (int, int), number of rays and number of ray segments
        """

        # variable naming convention used in this function:
        #
        # *_RDI     ray_data indices
        # *_II      intersector indices

        active_RDI = np.arange(ray_data.num_ray)
        ray_counter = ray_data.num_ray
        ray_segment_counter = 0

        M = ray_data.L_list
        D = ray_data.S_img_list - ray_data.L_list
        # self.__plot_rays(M, D)
        for order in range(ray_data.max_order):

            # perform raytracing for the current ray segment
            num_rays = len(M)
            ray_segment_counter += num_rays
            hit_pos, hit_II, hit_faces = intersector.intersects_location(M, D, multiple_hits=False)

            # reduce rays to those who hit a surface
            M = M[hit_II, :]
            D = D[hit_II, :]
            active_RDI = active_RDI[hit_II]

            # store surface indices which were hit
            ray_data.surf_seq_list[active_RDI, order] = face2surf[hit_faces]

            # check if we hit the correct surface
            wanted_surfaces = ray_data.seq_list[active_RDI, order]
            hit_surfaces = face2surf[hit_faces]
            valid_II = np.nonzero(hit_surfaces == wanted_surfaces)[0]
            if len(valid_II) == 0:
                return ray_counter, ray_segment_counter

            # reduce rays to those who hit the correct surface and crop ray length
            M_old = M[valid_II, :]
            M = hit_pos[valid_II, :]
            D = D[valid_II, :] + M_old - M
            active_RDI = active_RDI[valid_II]

            # mirror rays
            D = self.__mirrorRays(D, self.N0_list[wanted_surfaces[valid_II], :])
            # self.__plot_rays(M, D)

            # 1mm offset to avoid that the ray hits the ray origin face
            D_norm = np.reshape(1 / np.linalg.norm(D, axis=1), (len(D), 1))
            D_0 = np.tile(D_norm, (1, 3)) * D
            M += 0.001 * D_0

            # check last segments
            final_II = np.nonzero(ray_data.order[active_RDI] <= order + 1)[0]
            num_rays = len(final_II)
            if num_rays > 0:
                # trace last ray segment from source position to last refelection point
                final_RDI = active_RDI[final_II]
                M_final = np.tile(ray_data.src_pos.reshape((1, 3)), (num_rays, 1))
                D_final = -2 * D[final_II, :]  # make sure that we hit the surface
                # self.__plot_rays(M_final, D_final)
                ray_segment_counter += num_rays
                hit_pos, hit_II, hit_faces = intersector.intersects_location(
                    M_final, D_final, multiple_hits=False
                )
                hit_RDI = final_RDI[hit_II]

                # check if we hit the correct surface
                wanted_surfaces = ray_data.seq_list[hit_RDI, order]
                hit_surfaces = face2surf[hit_faces]
                valid_II = np.nonzero(hit_surfaces == wanted_surfaces)[0]
                valid_RDI = hit_RDI[valid_II]
                ray_data.visible[valid_RDI] = True

            # filter out final ray segments
            if order < ray_data.max_order - 1:
                valid_II = np.nonzero(ray_data.order[active_RDI] > order + 1)[0]
                M = M[valid_II, :]
                D = D[valid_II, :]
                active_RDI = active_RDI[valid_II]
                # self.__plot_rays(M, D)

        return ray_counter, ray_segment_counter

    def __mirrorRays(self, D_in, N0):
        """Mirror rays
        :var D_in:              np.array(N,3), ray directions
        :var N0:                np.array(N,3), unit normal vectors of the mirror planes
        :return:                np.array(N,3), mirrored ray directions
        """

        # D_out[n] = D_in[n] - 2 * np.dot(D_in[n], N0[n]) * N0[n]
        dot_list = (D_in * N0).sum(axis=1, keepdims=True)
        scale = np.tile(-2 * dot_list, (1, 3))
        D_out = D_in + scale * N0

        return D_out

    def __generate_subpos(self):
        """Generate list with voxel sub-grid positions; modifies self.subgrid_offset and self.subgrid_offset_coarse"""
        pos_list = list()
        for pos in subgrid(pitch=self.reachableVoxels.pitch, factor=self.grid_factor):
            pos_list.append(pos)
        self.subgrid_offset = np.array(pos_list)

        # Generate list with voxel sub-grid positions for finger snaps"""
        pos_list_fingersnap = list()
        for pos in subgrid(pitch=self.reachableVoxels.pitch, factor=1):
            pos_list_fingersnap.append(pos)
        self.subgrid_offset_coarse = np.array(pos_list_fingersnap)

    def __generate_triangle_bbox(self):
        """Generate bounding boxes for all triangles"""
        num_triangle = len(self.trimesh_noAE.faces)
        self.triangle_bbox_xmin = dict()
        self.triangle_bbox_xmax = dict()
        self.triangle_bbox_ymin = dict()
        self.triangle_bbox_ymax = dict()
        self.triangle_bbox_zmin = dict()
        self.triangle_bbox_zmax = dict()

        for order in self.max_dist_triangle.keys():

            triangle_bbox_xmin = np.zeros(num_triangle)
            triangle_bbox_xmax = np.zeros(num_triangle)
            triangle_bbox_ymin = np.zeros(num_triangle)
            triangle_bbox_ymax = np.zeros(num_triangle)
            triangle_bbox_zmin = np.zeros(num_triangle)
            triangle_bbox_zmax = np.zeros(num_triangle)

            for f, face in enumerate(self.trimesh_noAE.faces):
                V = self.trimesh_noAE.vertices[face]
                max_dist = self.max_dist_triangle[
                    order
                ]  # change this to utilize material-specific distance culling
                bbox_min = V.min(axis=0) - max_dist
                bbox_max = V.max(axis=0) + max_dist
                triangle_bbox_xmin[f] = bbox_min[0]
                triangle_bbox_xmax[f] = bbox_max[0]
                triangle_bbox_ymin[f] = bbox_min[1]
                triangle_bbox_ymax[f] = bbox_max[1]
                triangle_bbox_zmin[f] = bbox_min[2]
                triangle_bbox_zmax[f] = bbox_max[2]

            self.triangle_bbox_xmin[order] = triangle_bbox_xmin
            self.triangle_bbox_xmax[order] = triangle_bbox_xmax
            self.triangle_bbox_ymin[order] = triangle_bbox_ymin
            self.triangle_bbox_ymax[order] = triangle_bbox_ymax
            self.triangle_bbox_zmin[order] = triangle_bbox_zmin
            self.triangle_bbox_zmax[order] = triangle_bbox_zmax

    def __generate_transparency_filter(self, acoustic_materials):
        """Generate transparency filter by checking the transmission coefficients of all triangles"""

        # check materials
        materialList = self.trimesh_noAE.metadata["materials"]
        reflectiveMaterial = np.zeros(len(materialList), dtype=bool)
        for m, materialName in enumerate(materialList):
            if materialName is None:
                continue
            for mat in acoustic_materials:
                if mat.id == materialName:
                    pwr_trans = [val.t * val.t for val in mat.frequencies]
                    pwr_refl = [val.r * val.r for val in mat.frequencies]
                    reflectiveMaterial[m] = np.sum(pwr_refl) > np.sum(pwr_trans)
                    break

        # set transparency filter
        matIndices = self.trimesh_noAE.face_attributes["matIndices"]
        self.transparency_filter = reflectiveMaterial[matIndices]

    def __generate_local_intersector(self, listener_pos, source_pos, order):
        """Generate PyEmbree Intersector for local sub-mesh which contains triangles close to the given position
        :var listener_pos:      Float3, listener position in Cartesian coordinates for which a local sub-mesh is generated
        :var source_pos:        Float3, source position in Cartesian coordinates for which a local sub-mesh is generated
        :var order:             int, reflection order
        :return:                (RayMeshIntersector, numpy.array(N,)), PyEmbree Intersector for reduced local mesh, map: triangle (face) index -> surface index
        """
        from trimesh.ray.ray_pyembree import RayMeshIntersector

        listener_boolvec = listener_pos[0] > self.triangle_bbox_xmin[order]
        listener_boolvec &= listener_pos[0] < self.triangle_bbox_xmax[order]
        listener_boolvec &= listener_pos[1] > self.triangle_bbox_ymin[order]
        listener_boolvec &= listener_pos[1] < self.triangle_bbox_ymax[order]
        listener_boolvec &= listener_pos[2] > self.triangle_bbox_zmin[order]
        listener_boolvec &= listener_pos[2] < self.triangle_bbox_zmax[order]
        source_boolvec = source_pos[0] > self.triangle_bbox_xmin[order]
        source_boolvec &= source_pos[0] < self.triangle_bbox_xmax[order]
        source_boolvec &= source_pos[1] > self.triangle_bbox_ymin[order]
        source_boolvec &= source_pos[1] < self.triangle_bbox_ymax[order]
        source_boolvec &= source_pos[2] > self.triangle_bbox_zmin[order]
        source_boolvec &= source_pos[2] < self.triangle_bbox_zmax[order]
        local_faces = np.nonzero((listener_boolvec | source_boolvec) & self.transparency_filter)
        submesh = self.trimesh_noAE.submesh(local_faces, append=True)
        intersector = RayMeshIntersector(submesh)
        face2surf = self.face2surf_unfiltered[local_faces[0]]

        return intersector, face2surf

    def __generate_global_intersector(self):
        """Generate PyEmbree Intersector for local sub-mesh which contains reflective triangles
        :return:                (RayMeshIntersector, numpy.array(N,)), PyEmbree Intersector for reduced local mesh, map: triangle (face) index -> connected surface index
        """
        from trimesh.ray.ray_pyembree import RayMeshIntersector

        local_faces = np.nonzero(self.transparency_filter)
        submesh = self.trimesh_noAE.submesh(local_faces, append=True)
        intersector = RayMeshIntersector(submesh)
        face2surf = self.face2surf_unfiltered[local_faces[0]]

        return intersector, face2surf

    def __cleanup_surfaces(
        self,
        round_angle_threshold=31.0,
        round_radius_threshold=0.60,
        area_thresh=1.0,
        width_thresh=0.5,
    ):
        """Remove surfaces which are acoustically irrelevant; modifies self.surfaces
        :var round_angle_threshold:     float, minimum angle in degree for rounded edges (should be identical with EAR_ROUND_EDGE_ALPHA_MAX)
        :var round_radius_threshold:    float, minimum radius in m for rounded edges
        :var area_thresh:               float, minimum surface area in m^2
        :var width_thresh:              float, minimum width
        """

        trimesh_noAE = self.trimesh_noAE
        num_triangle = len(trimesh_noAE.faces)
        triangle2facet = -np.ones(num_triangle, dtype=int)
        faces_in_facets = set()
        for f, facet in enumerate(trimesh_noAE.facets):
            for t in facet:
                triangle2facet[t] = f
                faces_in_facets.add(t)

        # keep facets with round edges
        keep_facets = set()
        for f, triangle_pair in enumerate(trimesh_noAE.face_adjacency):
            facet_1 = triangle2facet[triangle_pair[0]]
            facet_2 = triangle2facet[triangle_pair[1]]
            if facet_1 != facet_2:
                angle = trimesh_noAE.face_adjacency_angles[f]
                valid_angle = abs(angle) * 180.0 / np.pi < round_angle_threshold
                valid_radius = trimesh_noAE.face_adjacency_radius[f] >= round_radius_threshold
                if valid_angle and valid_radius:
                    keep_facets.add(facet_1)
                    keep_facets.add(facet_2)

        # keep facets with minimum surface area and minimum width
        for facet in np.nonzero(trimesh_noAE.facets_area >= area_thresh)[0]:
            face_list = trimesh_noAE.facets[facet]
            sub_mesh = self.trimesh_noAE.submesh([face_list])
            bbox = sub_mesh[0].bounds
            extend = bbox[1] - bbox[0]
            # bounding box must have at least 2 dimensions which are greater than the threshold
            if sum(abs(extend) >= width_thresh) >= 2:
                keep_facets.add(facet)

        # map facets to triangle IDs
        keep_triangles = set()
        for facet in keep_facets:
            for triangle in trimesh_noAE.facets[facet]:
                keep_triangles.add(triangle)

        # same procedure for all triangles which are not contained in the facets list
        missed_triangles = [t for t in range(num_triangle) if t not in faces_in_facets]
        for triangle in missed_triangles:
            if trimesh_noAE.area_faces[triangle] < area_thresh:
                continue
            tri = trimesh_noAE.faces[triangle]
            V = trimesh_noAE.vertices[tri]
            extend = V.max(axis=0) - V.min(axis=0)
            # bounding box must have at least 2 dimensions which are greater than the threshold
            if sum(extend >= width_thresh) >= 2:
                keep_triangles.add(triangle)

        delete_surfaces = list()
        for s, surf in enumerate(self.surface):
            faceIdxNew = [triangle for triangle in surf.faceIdx if triangle in keep_triangles]
            self.surface[s].faceIdx = faceIdxNew
            if len(faceIdxNew) == 0:
                delete_surfaces.append(s)
        for s in reversed(delete_surfaces):
            self.surface.pop(s)
        for s in range(len(self.surface)):
            self.surface[s].surfaceIdx = s

    def __log_message(self, msg):
        """Print a log message with a timestamp
        :var msg:               str, text message
        """
        timestamp = time.time() - self.time
        print("" + format(timestamp, ".3f") + "  " + msg)
