# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#!/usr/bin/env python

import numpy as np


def voxelSubgridGenerator(pos=None, pitch=None, factor=4, delta=1 / 100):
    """Generator for voxel sub-positions
    :param pos:                 Float3: voxel position (default: np.zeros(3))
    :param pitch:               Float3: pitch of the voxel grid (default: np.ones(3))
    :param factor:              int: sub-sampling factor (default: 4)
    :param delta:               float: border spacing (default: 1/100)
    """

    if pos is None:
        pos = np.zeros(3)
    if pitch is None:
        pitch = np.ones(3)

    step = 1 / factor
    vec = np.arange(-0.5, 0.5 + step / 2, step)
    vec[0] += delta * step
    vec[-1] -= delta * step
    xvec = pitch[0] * vec + pos[0]
    yvec = pitch[1] * vec + pos[1]
    zvec = pitch[2] * vec + pos[2]

    # bottom layer
    for y in yvec:
        for x in xvec:
            yield np.array([x, y, zvec[0]])

    # intermediate layers
    for z in zvec[1:-1]:
        for x in xvec[:-1]:
            yield np.array([x, yvec[0], z])
        for y in yvec[:-1]:
            yield np.array([xvec[-1], y, z])
        for x in reversed(xvec[1:]):
            yield np.array([x, yvec[-1], z])
        for y in reversed(yvec[1:]):
            yield np.array([xvec[0], y, z])

    # top layer
    for y in yvec:
        for x in xvec:
            yield np.array([x, y, zvec[-1]])
