# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from xml.dom.minidom import Element
from typing import List, Optional, Dict


class VoxOcclusionControl:
    """Tuning parameter for voxel occclusion recovery effect per material"""

    attr_value: str = "value"
    attr_material: str = "material"
    elem_setting: str = "Setting"

    def __init__(self, settings: Dict[str, float]):
        self._settings = settings

    @staticmethod
    def fromxml(node: Element):
        settings: Dict[str, float] = {}

        settings_nodes: List[Element] = (
            n for n in node.childNodes if n.nodeName == VoxOcclusionControl.elem_setting
        )

        for n in settings_nodes:
            if not n.hasAttribute(VoxOcclusionControl.attr_material):
                raise Exception(
                    f"VoxOcclusionControl: Missing attribute `{VoxOcclusionControl.attr_material}`"
                )

            if not n.hasAttribute(VoxOcclusionControl.attr_value):
                raise Exception(
                    f"VoxOcclusionControl: Missing attribute `{VoxOcclusionControl.attr_value}`"
                )

            value: float = float(n.getAttribute(VoxOcclusionControl.attr_value).strip())
            if value < 0.0 or value > 100.0:
                raise ValueError(
                    f"Invalid voxDirectOcclusionControl value '{value}': Value must be between 0.0 and 100.0"
                )

            settings[n.getAttribute(VoxOcclusionControl.attr_material).strip()] = value

        return VoxOcclusionControl(settings)


class VoxDiffractionControl:
    """Tuning parameter for voxel diffraction 'strength'"""

    attr_value: str = "value"

    def __init__(self, value: float):
        self._value = value

    @staticmethod
    def fromxml(node: Element):
        if not node.hasAttribute(VoxDiffractionControl.attr_value):
            raise Exception(
                f"VoxDiffractionControl: Missing attribute `{VoxDiffractionControl.attr_value}`"
            )

        value: float = float(node.getAttribute(VoxDiffractionControl.attr_value).strip())
        if value < 0.0 or value > 1.0:
            raise ValueError(
                f"Invalid voxDiffractionControl value '{value}': Value must be between 0.0 and 1.0"
            )

        return VoxDiffractionControl(value)


class VoxEarlyReflectionControl:
    """Tuning parameters for voxel Early Reflection modelling"""

    attr_value_tag: str = "value"
    num_origins_tag: str = "NumOrigins"
    num_clusters_tag: str = "NumClusters"
    combine_reflections_per_cluster_tag: str = "CombineReflectionsPerCluster"
    max_gain_indoor_tag: str = "EarlyTuningIndoorGainDb"
    max_gain_outdoor_tag: str = "EarlyTuningOutdoorGainDb"

    def __init__(
        self,
        num_clusters: Optional[int],
        num_origins: Optional[int],
        combine_reflections_per_cluster: bool,
        indoor_gain_db: Optional[float],
        outdoor_gain_db: Optional[float],
    ):
        self._num_clusters = num_clusters
        self._num_origins = num_origins
        self._combine_reflections_per_cluster = combine_reflections_per_cluster
        self._indoor_gain_db = indoor_gain_db
        self._outdoor_gain_db = outdoor_gain_db

    @property
    def num_clusters(self) -> Optional[int]:
        return self._num_clusters

    @property
    def num_origins(self) -> Optional[int]:
        return self._num_origins

    @property
    def indoor_gain_db(self) -> Optional[float]:
        return self._indoor_gain_db

    @property
    def outdoor_gain_db(self) -> Optional[float]:
        return self._outdoor_gain_db

    @staticmethod
    def fromxml(node: Element):
        # Number of origins
        num_origins_node: List[Element] = [
            n for n in node.childNodes if n.nodeName == VoxEarlyReflectionControl.num_origins_tag
        ]

        n_origins: Optional[int] = None
        if len(num_origins_node) == 1:
            n_origins = int(
                num_origins_node[0].getAttribute(VoxEarlyReflectionControl.attr_value_tag).strip()
            )

            if n_origins < 2 or n_origins > 32:
                raise ValueError(
                    f"Invalid NumOrigins value '{n_origins}': Value must be between 2 and 32"
                )

        # Number of clusters
        num_clusters_node: List[Element] = [
            n for n in node.childNodes if n.nodeName == VoxEarlyReflectionControl.num_clusters_tag
        ]

        n_clusters: Optional[int] = None
        if len(num_clusters_node) == 1:
            n_clusters = int(
                num_clusters_node[0].getAttribute(VoxEarlyReflectionControl.attr_value_tag).strip()
            )

            if n_clusters < 2 or n_clusters > 32:
                raise ValueError(
                    f"Invalid NumClusters value '{n_clusters}': Value must be between 2 and 32"
                )

        # Allow multiple reflections per sector
        combine_reflections_per_cluster_node: List[Element] = [
            n
            for n in node.childNodes
            if n.nodeName == VoxEarlyReflectionControl.combine_reflections_per_cluster_tag
        ]

        combine_reflections_per_cluster: bool = len(combine_reflections_per_cluster_node) > 0

        # Indoor/outdoor gains
        indoor_gain_node: List[Element] = [
            n
            for n in node.childNodes
            if n.nodeName == VoxEarlyReflectionControl.max_gain_indoor_tag
        ]

        outdoor_gain_node: List[Element] = [
            n
            for n in node.childNodes
            if n.nodeName == VoxEarlyReflectionControl.max_gain_outdoor_tag
        ]

        if 1 == len(indoor_gain_node) and 1 == len(outdoor_gain_node):
            indoor_gain_db = float(
                indoor_gain_node[0].getAttribute(VoxEarlyReflectionControl.attr_value_tag).strip()
            )

            outdor_gain_db = float(
                outdoor_gain_node[0].getAttribute(VoxEarlyReflectionControl.attr_value_tag).strip()
            )
        elif not indoor_gain_node and not outdoor_gain_node:
            indoor_gain_db = None
            outdor_gain_db = None
        else:
            raise Exception(
                f'VoxEarlyReflectionControl: Either exactly one instance of "{VoxEarlyReflectionControl.max_gain_indoor_tag}" and "{VoxEarlyReflectionControl.max_gain_outdoor_tag}" must be present or neither of the two'
            )

        return VoxEarlyReflectionControl(
            n_clusters, n_origins, combine_reflections_per_cluster, indoor_gain_db, outdor_gain_db
        )


class EarlyTuningGainDb:
    """Early reflections tuning gain"""

    attr_value: str = "value"

    def __init__(self, value: float):
        self._value = value

    @staticmethod
    def fromxml(node: Element):
        if not node.hasAttribute("value"):
            raise Exception(
                f"EarlyTuningGainDb: Missing attribute `{EarlyTuningGainDb.attr_value}`"
            )

        value: float = float(node.getAttribute(EarlyTuningGainDb.attr_value).strip())
        if value < -21.5 or value > 10.0:
            raise ValueError(
                f"Invalid EarlyTuningGainDb value '{value}': Value must be between -21.5 and 10.0"
            )

        return EarlyTuningGainDb(value)


class LCEarlyOutdoorTuning:
    """LC Early reflections tuning parameters"""

    def __init__(
        self,
        predelay1=None,
        predelay2=None,
        ampFac1Db=None,
        ampFac2Db=None,
    ):
        self.predelay1 = predelay1
        self.predelay2 = predelay2
        self.ampFac1Db = ampFac1Db
        self.ampFac2Db = ampFac2Db

    @staticmethod
    def fromxml(node):
        predelay1 = None
        predelay2 = None
        ampFac1Db = None
        ampFac2Db = None

        if node.hasAttribute("predelay1"):
            predelay1 = float(node.getAttribute("predelay1").strip())
        else:
            raise SyntaxError("Missing required parameter in LCEarlyOutdoorTuning: predelay1")
        if node.hasAttribute("predelay2"):
            predelay2 = float(node.getAttribute("predelay2").strip())
        else:
            raise SyntaxError("Missing required parameter in LCEarlyOutdoorTuning: predelay2")
        if node.hasAttribute("ampFac1Db"):
            ampFac1Db = float(node.getAttribute("ampFac1Db").strip())
        else:
            raise SyntaxError("Missing required parameter in LCEarlyOutdoorTuning: ampFac1Db")
        if node.hasAttribute("ampFac2Db"):
            ampFac2Db = float(node.getAttribute("ampFac2Db").strip())
        else:
            raise SyntaxError("Missing required parameter in LCEarlyOutdoorTuning: ampFac2Db")

        return LCEarlyOutdoorTuning(predelay1, predelay2, ampFac1Db, ampFac2Db)


# --------------------------------------------------------------------------------------------------
class AcousticTuningParameters:
    def __init__(
        self,
        defaultAE=None,
        revRendererControl=None,
        voxOccCtrl: VoxOcclusionControl = None,
        voxDiffrCtrl: VoxDiffractionControl = None,
        voxEarlyCtrl: VoxEarlyReflectionControl = None,
        erGainDb: EarlyTuningGainDb = None,
        lcErOutdoorTuning: LCEarlyOutdoorTuning = None,
    ):
        """Creates an Acoustic Tuning Parameter set"""

        self.defaultAE = defaultAE
        self.revRendererControl = revRendererControl
        self.voxOccCtrl = voxOccCtrl._settings if voxOccCtrl else {}
        self.voxDiffrCtrl = voxDiffrCtrl._value if voxDiffrCtrl else 1.0
        self.voxEarlyCtrl = voxEarlyCtrl if voxEarlyCtrl else {}
        self.erGainDb = erGainDb._value if erGainDb else -8.0
        self.lcErOutdoorTuning = lcErOutdoorTuning

    @staticmethod
    def fromxml(node: Element):
        """Deserialises from an XML node"""
        defaultAE: Optional[DefaultAE] = None
        DefaultAENodes: List[Element] = node.getElementsByTagName("DefaultAE")

        if len(DefaultAENodes) > 1:
            raise Exception(f"Only exactly one instance of `DefaultAE` is allowed")
        elif len(DefaultAENodes) == 1:
            defaultAE = DefaultAE.fromxml(DefaultAENodes[0])

        revRendererControl: Optional[GlobalAEControl] = None
        revRendererControlNodes: List[Element] = node.getElementsByTagName("GlobalAEControl")

        if len(revRendererControlNodes) > 1:
            raise Exception(f"Only exactly one instance of `GlobalAEControl` is allowed")
        elif len(revRendererControlNodes) == 1:
            revRendererControl = GlobalAEControl.fromxml(revRendererControlNodes[0])

        # Voxel direct occlusion control
        voxDirectOcclusionControl: Optional[VoxOcclusionControl] = None
        voxDirectOcclusionControlNodes: List[Element] = node.getElementsByTagName(
            "VoxDirectOcclusionControl"
        )

        if len(voxDirectOcclusionControlNodes) > 1:
            raise Exception(f"Only exactly one instance of `VoxDirectOcclusionControl` is allowed")
        elif len(voxDirectOcclusionControlNodes) == 1:
            voxDirectOcclusionControl = VoxOcclusionControl.fromxml(
                voxDirectOcclusionControlNodes[0]
            )

        # Voxel diffraction control
        voxDiffractionControl: Optional[VoxDiffractionControl] = None
        voxDiffractionControlNodes: List[Element] = node.getElementsByTagName(
            "VoxDiffractionControl"
        )

        if len(voxDiffractionControlNodes) > 1:
            raise Exception(f"Only exactly one instance of `VoxDiffractionControl` is allowed")
        elif len(voxDiffractionControlNodes) == 1:
            voxDiffractionControl = VoxDiffractionControl.fromxml(voxDiffractionControlNodes[0])

        # Voxel early reflection control
        voxEarlyReflectionControl: Optional[VoxEarlyReflectionControl] = None
        voxEarlyReflectionControlNodes: List[Element] = node.getElementsByTagName(
            "VoxEarlyReflectionControl"
        )

        if len(voxEarlyReflectionControlNodes) > 1:
            raise Exception(f"Only exactly one instance of `VoxEarlyReflectionControl` is allowed")
        elif len(voxEarlyReflectionControlNodes) == 1:
            voxEarlyReflectionControl = VoxEarlyReflectionControl.fromxml(
                voxEarlyReflectionControlNodes[0]
            )

        # Early reflections tuning gain
        earlyReflectionTuning: Optional[EarlyTuningGainDb] = None
        earlyReflectionTuningNodes: List[Element] = node.getElementsByTagName("EarlyTuningGainDb")

        if len(earlyReflectionTuningNodes) > 1:
            raise Exception(f"Only exactly one instance of `EarlyTuningGainDb` is allowed")
        elif len(earlyReflectionTuningNodes) == 1:
            earlyReflectionTuning = EarlyTuningGainDb.fromxml(earlyReflectionTuningNodes[0])

        # Low complexity early reflections tuning parameters
        lcErOutdoorTuning: Optional[LCEarlyOutdoorTuning] = None
        lcErOutdoorTuningNodes: List[Element] = node.getElementsByTagName("LCEarlyOutdoorTuning")

        if len(lcErOutdoorTuningNodes) > 1:
            raise Exception(f"Only exactly one instance of `LCEarlyOutdoorTuning` is allowed")
        elif len(lcErOutdoorTuningNodes) == 1:
            lcErOutdoorTuning = LCEarlyOutdoorTuning.fromxml(lcErOutdoorTuningNodes[0])

        return AcousticTuningParameters(
            defaultAE,
            revRendererControl,
            voxDirectOcclusionControl,
            voxDiffractionControl,
            voxEarlyReflectionControl,
            earlyReflectionTuning,
            lcErOutdoorTuning,
        )

    def toxml(self, dom):
        node = dom.createElement("AcousticTuningParameters")
        if self.defaultAE != None:
            node.appendChild(DefaultAE.toxml(self.defaultAE, dom))

        if self.revRendererControl != None:
            node.appendChild(GlobalAEControl.toxml(self.revRendererControl, dom))

        if self.voxOccCtrl != None:
            voxOcclCtrlNode: Element = dom.createElement("VoxDirectOcclusionControl")
            for mat, value in self.voxOccCtrl.items():
                settingNode: Element = dom.createElement("Setting")
                settingNode.setAttribute("material", mat)
                settingNode.setAttribute("value", str(value))
                voxOcclCtrlNode.appendChild(settingNode)
            node.appendChild(voxOcclCtrlNode)

        if self.voxDiffrCtrl != None:
            voxDiffrCtrlNode: Element = dom.createElement("VoxDiffractionControl")
            voxDiffrCtrlNode.setAttribute("value", str(self.voxDiffrCtrl))
            node.appendChild(voxDiffrCtrlNode)

        if self.erGainDb != None:
            erGainDbNode: Element = dom.createElement("EarlyTuningGainDb")
            erGainDbNode.setAttribute("value", str(self.erGainDb))
            node.appendChild(erGainDbNode)

        if self.lcErOutdoorTuning != None:
            lcErOutdoorTuningNode: Element = dom.createElement("LCEarlyOutdoorTuning")
            lcErOutdoorTuningNode.setAttribute("predelay1", str(self.lcErOutdoorTuning.predelay1))
            lcErOutdoorTuningNode.setAttribute("predelay2", str(self.lcErOutdoorTuning.predelay2))
            lcErOutdoorTuningNode.setAttribute("ampFac1Db", str(self.lcErOutdoorTuning.ampFac1Db))
            lcErOutdoorTuningNode.setAttribute("ampFac2Db", str(self.lcErOutdoorTuning.ampFac2Db))
            node.appendChild(lcErOutdoorTuningNode)

        return node

    def __str__(self):
        s = "AcousticTuningParameters {"
        s += 'defaultAE = "%s", ' % self.defaultAE
        s += 'voxDirectOcclusionControl = "%s", ' % self.voxOccCtrl
        s += 'vocDiffractionControl = "%s", ' % self.voxDiffrCtrl
        s += 'earlyTuningGainDb = "%s" }' % self.erGainDb
        s += 'lcErOutdoorTuning = "%s" }' % self.lcErOutdoorTuning

        return s


class DefaultAE:
    def __init__(
        self,
        enabled=False,
        preset=None,
        environment=None,
        gainDb=None,
        density=None,
        delayMinMs=None,
        distanceDropDb=None,
        minDist=None,
    ):
        if preset != None and environment != None:
            raise SyntaxError(
                "DefaultAE cannot specify both a preset and an environment in EIF file."
            )
        self.enabled = enabled
        self.preset = preset
        self.environment = environment
        self.gainDb = gainDb
        self.density = density
        self.delayMinMs = delayMinMs
        self.distanceDropDb = distanceDropDb
        self.minDist = minDist

    @staticmethod
    def fromxml(node):
        enabled = False
        preset = None
        environment = None
        gainDb = None
        density = None
        delayMinMs = None
        distanceDropDb = None
        minDist = None

        if node.hasAttribute("enabled"):
            enabled = utils.parseBool(node.getAttribute("enabled").strip())
        if node.hasAttribute("preset"):
            preset = node.getAttribute("preset").strip()
        if node.hasAttribute("environment"):
            environment = node.getAttribute("environment").strip()
        if node.hasAttribute("gainDb"):
            gainDb = float(node.getAttribute("gainDb").strip())
        if node.hasAttribute("density"):
            density = float(node.getAttribute("density").strip())
        if node.hasAttribute("delayMinMs"):
            delayMinMs = float(node.getAttribute("delayMinMs").strip())
        if node.hasAttribute("distanceDropDb"):
            distanceDropDb = float(node.getAttribute("distanceDropDb").strip())
        if node.hasAttribute("minDist"):
            minDist = float(node.getAttribute("minDist").strip())
        return DefaultAE(
            enabled, preset, environment, gainDb, density, delayMinMs, distanceDropDb, minDist
        )

    def toxml(self, dom):
        defaultAENode: Element = dom.createElement("DefaultAE")

        if self.enabled != None:
            defaultAENode.setAttribute("enabled", str(self.enabled))
        if self.preset != None:
            defaultAENode.setAttribute("preset", str(self.preset))
        if self.environment != None:
            defaultAENode.setAttribute("environment", str(self.environment))
        if self.gainDb != None:
            defaultAENode.setAttribute("gainDb", str(self.gainDb))
        if self.density != None:
            defaultAENode.setAttribute("density", str(self.density))
        if self.delayMinMs != None:
            defaultAENode.setAttribute("delayMinMs", str(self.delayMinMs))
        if self.distanceDropDb != None:
            defaultAENode.setAttribute("distanceDropDb", str(self.distanceDropDb))
        if self.minDist != None:
            defaultAENode.setAttribute("minDist", str(self.minDist))
        return defaultAENode


class GlobalAEControl:
    def __init__(
        self,
        distanceDropDb=None,
        minDist=None,
        fadeInDistance=None,
    ):
        self.distanceDropDb = distanceDropDb
        self.minDist = minDist
        self.fadeInDistance = fadeInDistance

    @staticmethod
    def fromxml(node):
        distanceDropDb = None
        minDist = None
        if node.hasAttribute("distanceDropDb"):
            distanceDropDb = float(node.getAttribute("distanceDropDb").strip())
        if node.hasAttribute("minDist"):
            minDist = float(node.getAttribute("minDist").strip())
        if node.hasAttribute("fadeInDistance"):
            fadeInDistance = float(node.getAttribute("fadeInDistance").strip())
        return GlobalAEControl(distanceDropDb, minDist, fadeInDistance)

    def toxml(self, dom):
        GlobalAEControlNode: Element = dom.createElement("GlobalAEControl")
        if self.distanceDropDb != None:
            GlobalAEControlNode.setAttribute("distanceDropDb", str(self.distanceDropDb))
        if self.minDist != None:
            GlobalAEControlNode.setAttribute("minDist", str(self.minDist))
        if self.minDist != None:
            GlobalAEControlNode.setAttribute("fadeInDistance", str(self.fadeInDistance))

        return GlobalAEControlNode
