# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from . import validation


class AirflowGenerator:
    def __init__(
        self,
        id,
        generatorType,
        orientation,
        speed,
        position=None,
        gainDb=0,
        active=True,
        azimuthLimit=None,
        elevationLimit=None,
        height=None,
        width=None,
        baseRadius=None,
        topRadius=0.0,
        length=None,
        distanceAttenuationFactor=1.0,
        edgeRolloff=50,
        transformID=None,
        anchorID=None,
    ):
        self.id = validation.validateID(id)
        self.generatorType = generatorType

        self.position = position
        self.orientation = orientation
        self.speed = speed
        self.gainDb = gainDb
        self.active = validation.validateBool(active, f"active flag in '{self.id}'")
        self.azimuthLimit = (azimuthLimit,)
        self.elevationLimit = (elevationLimit,)
        self.height = height
        self.width = width
        self.baseRadius = baseRadius
        self.topRadius = topRadius
        self.length = length
        self.distanceAttenuationFactor = distanceAttenuationFactor
        self.edgeRolloff = edgeRolloff
        self.transform = transformID
        self.anchor = anchorID

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""

        if not node.hasAttribute("id"):
            raise SyntaxError("AirflowGenerator is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("type"):
            raise SyntaxError("AirflowGenerator is missing type in EIF file")
        generatorType = node.getAttribute("type").strip()

        if node.hasAttribute("position"):
            position = validation.validateVector3(
                node.getAttribute("position").strip(), f"position in AirflowGenerator '{id}'"
            )
        else:
            position = None

        if not node.hasAttribute("orientation"):
            raise SyntaxError("AirflowGenerator is missing orientation in EIF file")
        else:
            orientation = validation.validateVector3(
                node.getAttribute("orientation").strip(), f"orientation in AirflowGenerator '{id}'"
            )

        if not node.hasAttribute("speed"):
            raise SyntaxError("AirflowGenerator is missing speed in EIF file")
        else:
            speed = float(node.getAttribute("speed").strip())

        if node.hasAttribute("gainDb"):
            gainDb = float(node.getAttribute("gainDb").strip())
        else:
            gainDb = 0

        if node.hasAttribute("active"):
            active = utils.parseBool(node.getAttribute("active").strip())
        else:
            active = True

        if node.hasAttribute("azimuthLimit"):
            azimuthLimit = float(node.getAttribute("azimuthLimit").strip())
        else:
            azimuthLimit = None

        if node.hasAttribute("elevationLimit"):
            elevationLimit = float(node.getAttribute("elevationLimit").strip())
        else:
            elevationLimit = None

        if node.hasAttribute("height"):
            height = float(node.getAttribute("height").strip())
        else:
            height = None

        if node.hasAttribute("width"):
            width = float(node.getAttribute("width").strip())
        else:
            width = None

        if node.hasAttribute("baseRadius"):
            baseRadius = float(node.getAttribute("baseRadius").strip())
        else:
            baseRadius = None

        if node.hasAttribute("topRadius"):
            topRadius = float(node.getAttribute("topRadius").strip())
        else:
            topRadius = 0.0

        if node.hasAttribute("length"):
            length = float(node.getAttribute("length").strip())
        else:
            length = None

        if node.hasAttribute("distanceAttenuationFactor"):
            distanceAttenuationFactor = float(
                node.getAttribute("distanceAttenuationFactor").strip()
            )
        else:
            distanceAttenuationFactor = 1.0

        if node.hasAttribute("edgeRolloff"):
            edgeRolloff = int(node.getAttribute("edgeRolloff").strip())
        else:
            edgeRolloff = 50

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        return AirflowGenerator(
            id,
            generatorType=generatorType,
            orientation=orientation,
            speed=speed,
            position=position,
            gainDb=gainDb,
            active=active,
            azimuthLimit=azimuthLimit,
            elevationLimit=elevationLimit,
            height=height,
            width=width,
            baseRadius=baseRadius,
            topRadius=topRadius,
            length=length,
            distanceAttenuationFactor=distanceAttenuationFactor,
            edgeRolloff=edgeRolloff,
            transformID=transform,
            anchorID=anchor,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        raise NotImplementedError("AirflowGenerator not yet implemented toXML")

    def __str__(self):
        raise NotImplementedError("AirflowGenerator not yet implemented __str__")
