# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import validation


def TrueXor(*args):
    return sum(bool(a) for a in args) == 1


class AirflowSource:
    def __init__(
        self,
        id,
        airflowGenerator,
        defaultAirflowProfile=1,
        activeOutsideAEs=False,
        ActiveEnvironment=None,
        AirflowProfile=None,
    ):
        self.id = validation.validateID(id)
        self.airflowGenerator = airflowGenerator
        self.defaultAirflowProfile = defaultAirflowProfile
        self.activeOutsideAEs = validation.validateBool(
            activeOutsideAEs, f"activeOutsideAEs in Airflow Source '{id}'"
        )
        self.ActiveEnvironment = ActiveEnvironment
        self.AirflowProfiles = AirflowProfile
        pass

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""

        if not node.hasAttribute("id"):
            raise SyntaxError("AirflowSource is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("airflowGenerator"):
            raise SyntaxError("AirflowSource is missing airflowGenerator reference in EIF file")
        airflowGenerator = node.getAttribute("airflowGenerator").strip()

        if node.hasAttribute("defaultAirflowProfile"):
            defaultAirflowProfile = int(node.getAttribute("defaultAirflowProfile").strip())
        else:
            defaultAirflowProfile = 0

        if node.hasAttribute("ActiveOutsideAEs"):
            activeOutsideAEs = node.getAttribute("ActiveOutsideAEs").strip()
        else:
            activeOutsideAEs = False

        ActiveEnvironmentElements = node.getElementsByTagName("ActiveEnvironment")
        if len(ActiveEnvironmentElements) > 0:
            ActiveEnvironment = []
            for n in ActiveEnvironmentElements:
                ActiveEnvironment.append(AirflowActiveEnv.fromxml(n))
        else:
            ActiveEnvironment = None

        AirflowProfileElements = node.getElementsByTagName("AirflowProfile")
        if len(AirflowProfileElements) > 0:
            AirflowProfiles = []
            for n in AirflowProfileElements:
                AirflowProfiles.append(AirflowProfile.fromxml(n))
            if len(set([a.profileType for a in AirflowProfiles])) > 1:
                raise SyntaxError("All AirflowProfiles must be of the same type")
        else:
            AirflowProfiles = None

        return AirflowSource(
            id,
            airflowGenerator=airflowGenerator,
            defaultAirflowProfile=defaultAirflowProfile,
            activeOutsideAEs=activeOutsideAEs,
            ActiveEnvironment=ActiveEnvironment,
            AirflowProfile=AirflowProfiles,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        raise NotImplementedError("AirflowSource not yet implemented toXML")
        node = dom.createElement("ObjectSource")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)

        return node

    def __str__(self):
        raise NotImplementedError("AirflowSource not yet implemented __str__ function")


class AirflowProfile:
    def __init__(
        self,
        id=None,
        profileType=0,
        speed=None,
        azimuth=None,
        elevation=None,
        freqs=None,
        gainDb=None,
        q=None,
    ):
        self.id = id
        self.profileType = profileType
        self.speed = speed
        self.azimuth = azimuth
        self.elevation = elevation
        self.freqs = freqs
        self.gainDb = gainDb
        self.q = q

    @staticmethod
    def fromxml(node):
        if not node.hasAttribute("id"):
            raise SyntaxError("AirflowProfile is missing ID in EIF file")
        else:
            id = node.getAttribute("id").strip()

        if not node.hasAttribute("profileType"):
            raise SyntaxError("AirflowProfile is missing ID in EIF file")
        else:
            profileType = int(node.getAttribute("profileType").strip())

        if node.hasAttribute("speed"):
            speed = float(node.getAttribute("speed").strip())
        else:
            speed = None

        if node.hasAttribute("azimuth"):
            azimuth = float(node.getAttribute("azimuth").strip())
        else:
            azimuth = None

        if node.hasAttribute("elevation"):
            elevation = float(node.getAttribute("elevation").strip())
        else:
            elevation = None

        if TrueXor(speed is not None, azimuth is not None, elevation is not None) is False:
            raise SyntaxError(
                "Only one of Speed, Azimuth, or Elevation may be specified in an Airflow Profile"
            )

        freqs = []
        gainDb = []
        q = []
        for n in node.getElementsByTagName("AirflowFrequency"):
            freqs.append(float(n.getAttribute("f").strip()))
            if azimuth is None:
                gainDb.append(float(n.getAttribute("gainDb").strip()))
            else:
                gains = n.getAttribute("gainDb").strip().split(" ")
                if len(gains) != 2:
                    raise SyntaxError(
                        "Azimuth profiles require exactly 2 gains per frequency in an Airflow Profile"
                    )
                gainDb.append([float(x) for x in gains])
            if n.hasAttribute("q"):
                if azimuth is None:
                    q.append(float(n.getAttribute("q").strip()))
                else:
                    qs = n.getAttribute("q").strip().split(" ")
                    if len(qs) != 2:
                        raise SyntaxError(
                            "Azimuth profiles require exactly 2 q values per frequency in an Airflow Profile"
                        )
                    q.append([float(x) for x in qs])

        return AirflowProfile(id, profileType, speed, azimuth, elevation, freqs, gainDb, q)


class AirflowActiveEnv:
    def __init__(
        self,
        AcEnvId=None,
    ):
        self.AcEnvId = AcEnvId

    def fromxml(node):
        if not node.hasAttribute("AcEnvId"):
            raise SyntaxError("Airflow ActiveEnvironment is missing AcEnvId in EIF file")
        else:
            AcEnvId = node.getAttribute("AcEnvId").strip()

        return AirflowActiveEnv(AcEnvId)
