# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import validation
from . import utils


class Anchor:
    def __init__(
        self,
        id,
        lsdfRef,
        autoRescale=False,
        tolerance=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        audioSourceRescale=[1, 1, 1],
        boxIDs=[],
        sphereIDs=[],
        cylinderIDs=[],
        meshIDs=[],
        objectSourceIDs=[],
        hoaSourceIDs=[],
        channelSourceIDs=[],
    ):
        """Creates a spatial transform"""
        self.id = validation.validateID(id)
        self.lsdfRef = lsdfRef
        self.autoRescale = validation.validateBool(autoRescale, f"autoRescale flag in Anchor {id}")
        self.audioSourceRescale = validation.validateVector3(
            audioSourceRescale, f"audioSourceRescale in Anchor '{self.id}'"
        )
        self.tolerance = validation.validateNumberList(
            tolerance, f"tolerance in Anchor '{self.id}'"
        )
        self.boxIDs = boxIDs
        self.sphereIDs = sphereIDs
        self.cylinderIDs = cylinderIDs
        self.meshIDs = meshIDs

        self.objectSourceIDs = objectSourceIDs
        self.hoaSourceIDs = hoaSourceIDs
        self.channelSourceIDs = channelSourceIDs

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Anchor has missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("lsdf_ref"):
            raise SyntaxError(f"Anchor '{id}' has missing lsdf_ref in EIF file")
        lsdfRef = node.getAttribute("lsdf_ref").strip()

        if node.hasAttribute("audioSourceRescale"):
            audioSourceRescale = node.getAttribute("audioSourceRescale").strip()
        else:
            audioSourceRescale = [1, 1, 1]

        if node.hasAttribute("autoRescale"):
            autoRescale = utils.parseBool(node.getAttribute("autoRescale").strip())
        else:
            autoRescale = False

        if node.hasAttribute("tolerance"):
            tolerance = node.getAttribute("tolerance").strip()
        else:
            tolerance = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        boxIDs = []
        sphereIDs = []
        cylinderIDs = []
        meshIDs = []
        objectSourceIDs = []
        hoaSourceIDs = []
        channelSourceIDs = []

        def _getChildrenIDs(node, childName, IDs):
            if len(node.getElementsByTagName(childName)) > 0:
                for child in node.getElementsByTagName(childName):
                    IDs.append(child.getAttribute("id").strip())

        _getChildrenIDs(node, "Box", boxIDs)
        _getChildrenIDs(node, "Sphere", sphereIDs)
        _getChildrenIDs(node, "Cylinder", cylinderIDs)
        _getChildrenIDs(node, "Mesh", meshIDs)
        _getChildrenIDs(node, "ObjectSource", objectSourceIDs)
        _getChildrenIDs(node, "HOASource", hoaSourceIDs)
        _getChildrenIDs(node, "ChannelSource", channelSourceIDs)

        return Anchor(
            id,
            lsdfRef,
            autoRescale,
            tolerance,
            audioSourceRescale,
            boxIDs=boxIDs,
            sphereIDs=sphereIDs,
            cylinderIDs=cylinderIDs,
            meshIDs=meshIDs,
            objectSourceIDs=objectSourceIDs,
            hoaSourceIDs=hoaSourceIDs,
            channelSourceIDs=channelSourceIDs,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Anchor")

        node.setAttribute("id", self.id)
        node.setAttribute("lsdf_ref", self.lsdfRef)
        if self.autoRescale != False:
            node.setAttribute("autoRescale", "true")
        node.setAttribute("tolerance".self.tolerance)
        node.setAttribute("audioSourceRescale".self.audioSourceRescale)

        # TODO: Add children here

        return node

    def __str__(self):
        children = "\n"
        for boxId in self.boxIDs:
            children += f"\tBox: '{boxId}'\n"
        for sphereId in self.sphereIDs:
            children += f"\tSphere: '{sphereId}'\n"
        for cylinderId in self.cylinderIDs:
            children += f"\tCylinder: '{cylinderId}'\n"
        for objSrcId in self.objectSourceIDs:
            children += f"\tObjectSource: '{objSrcId}'\n"
        for hoaSrcId in self.hoaSourceIDs:
            children += f"\tHOASource: '{hoaSrcId}'\n"
        for chanSrcId in self.channelSourceIDs:
            children += f"\tChannelSource: '{chanSrcId}'\n"

        children = "" if children == "\n" else children

        return f"Anchor '{self.id}': LSDF_REF-'{self.lsdfRef}': autoRescale-'{self.autoRescale}': audioSoureRescale-'{self.audioSourceRescale}': tolerance-'{self.tolerance}': {{{children}}}"
