# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import validation
from . import utils

import numpy as np


class Loudspeaker:
    def __init__(
        self,
        position,
        channel,
        orientation=None,
        directivity=None,
        directiveness=1.0,
    ):
        self.position = validation.validateVector3(position, "position in Loudspeaker")
        self.orientation = orientation
        if type(channel) != int:
            raise ValueError(f"Channel must be an integer and not {type(channel)}")
        self.channel = channel
        self.directivity = directivity
        if directiveness < 0:
            raise ValueError(f"Directiveness value of '{directiveness}' must not be negative")
        self.directiveness = directiveness

    @staticmethod
    def fromxml(node):
        if not node.hasAttribute("position"):
            raise SyntaxError("Loudspeaker missing position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = None

        if not node.hasAttribute("channel"):
            raise ValueError("Loudspeaker missing channel in EIF file")
        channel = int(node.getAttribute("channel").strip())
        if channel < 0:
            raise ValueError(f"Invalid Loudspeaker Channel value: {channel}")

        if node.hasAttribute("directivity"):
            directivity = node.getAttribute("directivity").strip()
        else:
            directivity = None

        if node.hasAttribute("directiveness"):
            directiveness = float(node.getAttribute("directiveness").strip())
            if directiveness < 0:
                raise ValueError(f"Invalid directiveness value: '{directiveness}'")
        else:
            directiveness = 1

        return Loudspeaker(
            position=position,
            orientation=orientation,
            channel=channel,
            directivity=directivity,
            directiveness=directiveness,
        )

    def toxml(self, dom):
        node = dom.createElement("Loudspeaker")

        node.setAttribute(
            "position",
            "%s %s %s"
            % (
                utils.numberToStr(self.position[0]),
                utils.numberToStr(self.position[1]),
                utils.numberToStr(self.position[2]),
            ),
        )
        if self.orientation[0] != 0 or self.orientation[1] != 0 or self.orientation[2] != 0:
            node.setAttribute(
                "orientation",
                "%s %s %s"
                % (
                    utils.numberToStr(self.orientation[0]),
                    utils.numberToStr(self.orientation[1]),
                    utils.numberToStr(self.orientation[2]),
                ),
            )
        node.setAttribute("channel", utils.numberToStr(self.channel))
        if self.directivity is not None:
            node.setAttribute("directivity", self.directivity)
        if self.directiveness != 1:
            node.setAttribute("directiveness", utils.numberToStr(self.directiveness))

        return node

    def __str__(self):
        pass


class ChannelSource:
    def __init__(
        self,
        loudspeakers,
        id,
        inputLayout,
        signal,
        position=[0, 0, 0],
        orientation=[0, 0, 0],
        cspace="relative",
        active=True,
        gainDb=0.0,
        refDistance=1.0,
        aparams=None,
        mode="continuous",
        play=False,
        transformID=None,
        anchorID=None,
        vdlMethod=0,
        hasReverbGain=False,
        reverbGainDb=0.0,
        noReverb=False,
        granularPos=None,
    ):
        self.id = validation.validateID(id)

        if any([not isinstance(l, Loudspeaker) for l in loudspeakers]):
            typesString = ""
            for l in loudspeakers:
                typesString += str(type(l)) + " "

            raise ValueError(
                f"Loudspeakers in ChannelSource '{id}' must be a "
                f"list of Loudspeakers and not of types: {typesString}"
            )

        self.position = validation.validateVector3(position, f"position in ChannelSource '{id}'")
        self.orientation = validation.validateVector3(
            orientation, f"orientation in ChannelSource '{id}'"
        )
        self.cspace = validation.validateCSpace(cspace)

        for ls in loudspeakers:
            if ls.orientation is None:
                ls.orientation = _look_orientation(ls.position, [0, 0, 0])
            ls.orientation = validation.validateVector3(
                ls.orientation, "orientation in Loudspeaker"
            )

        self.loudspeakers = loudspeakers

        if type(inputLayout) != int:
            raise ValueError(
                f"inputLayout in ChannelSource '{id}' must be an integer and "
                f"not of type {type(inputLayout)}"
            )
        if not 0 <= inputLayout <= 127:
            raise ValueError(
                f"inputLayout in ChannelSource '{id}' must be a CICP" "index between 0 and 127"
            )
        self.inputLayout = inputLayout
        self.active = validation.validateBool(active, f"active in ChannelSource '{id}'")
        self.gainDb = gainDb
        if refDistance <= 0:
            raise ValueError(
                f"Reference distance in ChannelSource '{id}' must be positive: "
                f"RefDistance '{refDistance}'"
            )
        self.refDistance = refDistance
        if signal == "" or not isinstance(signal, str):
            raise ValueError(
                f"Signal in ChannelSource '{id}' must be a string and not empty: "
                f"Type '{type(signal)}', Value '{signal}'"
            )
        self.signal = signal
        self.aparams = aparams
        self.mode = validation.validatePlaybackMode(mode)
        self.play = validation.validateBool(play, f"play flag in ChannelSource {id}")
        self.transform = transformID
        self.anchor = anchorID
        self.vdlMethod = vdlMethod
        self.hasReverbGain = hasReverbGain
        self.reverbGainDb = reverbGainDb
        self.noReverb = noReverb
        if granularPos != None:
            self.granularPos = validation.validateVector3(
                granularPos, f"Granular position in HOASource '{self.id}'"
            )
        else:
            self.granularPos = granularPos

    @staticmethod
    def fromxml(node):
        if not node.hasAttribute("id"):
            raise SyntaxError("ChannelSource has missing ID in EIF file")
        id = node.getAttribute("id").strip()

        elements = node.getElementsByTagName("Loudspeaker")
        if len(elements) < 1:
            raise SyntaxError(f"ChannelSource '{id}' must have at least one loudspeaker")
        loudspeakers = [Loudspeaker.fromxml(e) for e in elements]

        if not node.hasAttribute("position"):
            raise SyntaxError(f"ChannelSource '{id}' has Missing position EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0, 0, 0]

        if node.hasAttribute("cspace"):
            cspace = node.getAttribute("cspace").strip()
        else:
            cspace = "relative"

        if not node.hasAttribute("inputLayout"):
            raise SyntaxError(f"ChannelSource '{id}' must have an inputlayout")
        inputLayout = int(node.getAttribute("inputLayout").strip())

        if node.hasAttribute("active"):
            active = utils.parseBool(node.getAttribute("active").strip())
        else:
            active = True

        if node.hasAttribute("gainDb"):
            gainDb = float(node.getAttribute("gainDb").strip())
        else:
            gainDb = 0

        if node.hasAttribute("refDistance"):
            refDistance = float(node.getAttribute("refDistance").strip())
        else:
            refDistance = 1

        if not node.hasAttribute("signal"):
            raise SyntaxError(f"Channel Source '{id}' has missing signal in EIF file")
        signal = node.getAttribute("signal")

        if node.hasAttribute("aparams"):
            aparams = utils.parseAuthoringParameters(node.getAttribute("aparams").strip())
        else:
            aparams = None

        if node.hasAttribute("mode"):
            mode = utils.parsePlaybackMode(node.getAttribute("mode").strip())
        else:
            mode = "continuous"

        if node.hasAttribute("play"):
            play = utils.parseBool(node.getAttribute("play").strip())
        else:
            play = False

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        if node.hasAttribute("vdlMethod"):
            from ...framework import _lookup_tables as LUT

            vdlMethod = LUT.vdl_method_codes[node.getAttribute("vdlMethod").strip()]
        else:
            vdlMethod = 0  # Default 0 means "spline" method

        if node.hasAttribute("granularPos"):
            granularPos = node.getAttribute("granularPos").strip()
        else:
            granularPos = None

        if node.hasAttribute("reverbGainDb"):
            reverbGainDb = float(node.getAttribute("reverbGainDb").strip())
            assert (
                reverbGainDb >= -119
            ), "reverbGainDb cannot be smaller than -119 dB ! Source: " + str(id)
            assert reverbGainDb <= 12, "reverbGainDb cannot be grater than +12 dB ! Source: " + str(
                id
            )
            hasReverbGain = True
        else:
            hasReverbGain = False
            reverbGainDb = 0.0

        if node.hasAttribute("muteReverb"):
            noReverb = utils.parseBool(node.getAttribute("muteReverb").strip())
        else:
            noReverb = False

        return ChannelSource(
            loudspeakers=loudspeakers,
            id=id,
            inputLayout=inputLayout,
            signal=signal,
            position=position,
            orientation=orientation,
            cspace=cspace,
            active=active,
            gainDb=gainDb,
            refDistance=refDistance,
            aparams=aparams,
            mode=mode,
            play=play,
            transformID=transform,
            anchorID=anchor,
            vdlMethod=vdlMethod,
            hasReverbGain=hasReverbGain,
            reverbGainDb=reverbGainDb,
            noReverb=noReverb,
            granularPos=granularPos,
        )

    def toxml(self, dom):
        node = dom.createElement("ChannelSource")

        for loudspeaker in self.loudspeakers:
            node.appendChild(loudspeaker.toxml(dom))

        node.setAttribute("id", self.id)
        node.setAttribute(
            "position",
            "%s %s %s"
            % (
                utils.numberToStr(self.position[0]),
                utils.numberToStr(self.position[1]),
                utils.numberToStr(self.position[2]),
            ),
        )
        if self.orientation[0] != 0 or self.orientation[1] != 0 or self.orientation[2] != 0:
            node.setAttribute(
                "orientation",
                "%s %s %s"
                % (
                    utils.numberToStr(self.orientation[0]),
                    utils.numberToStr(self.orientation[1]),
                    utils.numberToStr(self.orientation[2]),
                ),
            )
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)
        node.setAttribute("inputLayout", utils.numberToStr(self.inputLayout))
        if self.active != True:
            node.setAttribute("active", "False")
        if self.gainDb != 0:
            node.setAttribute("gainDb", utils.numberToStr(self.gainDb))
        if self.refDistance != 1:
            node.setAttribute("refDistance", utils.numberToStr(self.refDistance))
        node.setAttribute("signal", self.signal)
        if self.mode != "continuous":
            node.setAttribute("mode", self.mode)
        if self.play != False:
            node.setAttribute("play", "true")
        if self.aparams is not None:
            node.setAttribute("aparams", self.aparams)
        if self.vdlMethod is not None:
            node.setAttribute("vdlMethod", utils.numberToStr(self.vdlMethod))
        if self.reverbGainDb is not None:
            assert (
                self.reverbGainDb >= -119
            ), "reverbGainDb cannot be smaller than -119 dB ! Source: " + str(id)
            assert (
                self.reverbGainDb <= 12
            ), "reverbGainDb cannot be grater than +12 dB ! Source: " + str(id)
            node.setAttribute("reverbGainDb", utils.numberToStr(self.reverbGainDb))
            self.hasReverbGain = True
        if self.noReverb != False:
            node.setAttribute("muteReverb", "true")
        if self.granularPos is not None:
            node.setAttribute(
                "granularPos",
                "%s %s %s"
                % (
                    utils.numberToStr(self.granularPos[0]),
                    utils.numberToStr(self.granularPos[1]),
                    utils.numberToStr(self.granularPos[2]),
                ),
            )

        return node

    def __str__(self):
        pass


def _look_orientation(from_pos, to_pos):
    # NB: importing this at the top of the file led to cyclic dependencies:
    from mpegi.geometry.transforms import getOffsetAndAnglesFromMatrixTRS

    # vector in look direction, default look direction is -z
    minus_z = np.asarray(to_pos) - np.asarray(from_pos)
    z_length = np.linalg.norm(minus_z)
    if z_length < 0.000001:
        raise ValueError("_look_orientation(): points are too close")
    minus_z /= z_length
    up = 0, 1, 0
    up_ness = np.dot(minus_z, up)
    if up_ness > 0.999999:
        # Looking up, remaining DoF is chosen arbitrarily. This is a singularity!
        return [0, 90, 0]
    elif up_ness < -0.999999:
        # Looking down, remaining DoF is chosen arbitrarily. This is a singularity!
        return [0, -90, 0]
    x = np.cross(minus_z, up)
    x /= np.linalg.norm(x)
    y = np.cross(x, minus_z)
    mat = np.identity(4)
    mat[:3, 0] = x
    mat[:3, 1] = y
    mat[:3, 2] = -minus_z
    _, angles = getOffsetAndAnglesFromMatrixTRS(mat)
    return list(map(np.degrees, angles))
