# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from . import validation
import warnings


class AcousticEnvironment:
    def __init__(self, id, parameters=None, region=None):
        """Creates an Acoustic Environment"""
        if not parameters:
            raise ValueError(
                "Acoustic Enviroment must have at least \
                              one acoustic parameter"
            )

        self.id = validation.validateID(id)
        self.parameters = parameters
        self.region = region

    @staticmethod
    def fromxml(node):
        """Deserialises from an XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("AcousticEnvironment is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if node.hasAttribute("region"):
            region = node.getAttribute("region").strip()
        else:
            region = None  # global environment

        parameters = []
        for n in node.getElementsByTagName("AcousticParameters"):
            p = AcousticParameters.fromxml(n)
            parameters.append(p)
        if len(parameters) > 1:
            if any(x for x in parameters if x.position == None):
                raise SyntaxError(
                    "Acoustic Environments with multiple parameters must specify a positino for each parameter set"
                )

        return AcousticEnvironment(id, parameters, region)

    def toxml(self, dom):
        node = dom.createElement("AcousticEnvironment")
        node.setAttribute("id", self.id)
        if self.region != None:
            node.setAttribute("region", self.region)

        for p in self.parameters:
            node.appendChild(p.toxml(dom))

        return node

    def __str__(self):
        if self.region:
            regionName = self.region.id
        else:
            regionName = "None"

        s = 'AcousticEnvironment "%s" {\n\tregion = %s\n' % (self.id, regionName)
        for x in self.parameters:
            s += utils.indent(str(x)) + "\n"
        s += "}"
        return s


class AcousticParameters:
    def __init__(
        self, position=None, predelay=0, complexityLevel=None, freqs=None, rt60=None, dsr=None
    ):
        """Create Acoustic Parameters"""
        if type(freqs) != list or type(rt60) != list or type(dsr) != list:
            raise ValueError("freqs, rt60, and dsr must be lists")
        if len(freqs) != len(rt60) or len(freqs) != len(dsr):
            raise ValueError("freqs, rt60, add dsr all must have the same number of elements.py")

        self.position = position
        self.predelay = predelay
        self.complexityLevel = complexityLevel
        self.freqs = freqs
        self.rt60 = rt60
        self.dsr = dsr

    @staticmethod
    def fromxml(node):
        """Deserialise from an XML node"""
        if node.hasAttribute("position"):
            position = utils.parsePositionFromString(node.getAttribute("position").strip())
        else:
            position = None

        if node.hasAttribute("predelay"):
            predelay = float(node.getAttribute("predelay").strip())
        else:
            predelay = None

        if node.hasAttribute("complexityLevel"):
            complexityLevel = float(node.getAttribute("complexityLevel").strip())
        else:
            complexityLevel = None

        freqs = []
        rt60 = []
        dsr = []

        warnings.simplefilter("always", DeprecationWarning)

        for n in node.getElementsByTagName("Frequency"):
            # TODO: Validate values
            freqs.append(float(n.getAttribute("f").strip()))
            rt60.append(float(n.getAttribute("rt60").strip()))
            if n.hasAttribute("dsr"):
                assert predelay != None, (
                    "AcousticParameters cannot specify a "
                    "dsr value without first specifying a predelay value"
                )
                dsr.append(float(n.getAttribute("dsr").strip()))
            elif n.hasAttribute("ddr"):
                assert predelay != None, (
                    "AcousticParameters cannot specify a "
                    "dsr value without first specifying a predelay value"
                )
                warnings.warn(
                    "DDR valus have been deprectaed, update EIFs to use DSRs instead.",
                    DeprecationWarning,
                )
                warnings.simplefilter("ignore", DeprecationWarning)
                dsr.append(float(n.getAttribute("ddr").strip()))
            else:
                # Default DSR value
                dsr.append(None)
        warnings.simplefilter("ignore", DeprecationWarning)
        return AcousticParameters(position, predelay, complexityLevel, freqs, rt60, dsr)

    def toxml(self, dom):
        node = dom.createElement("AcousticParameters")

        node.setAttribute(
            "position",
            "%s %s %s"
            % (
                utils.numberToStr(self.position[0]),
                utils.numberToStr(self.position[1]),
                utils.numberToStr(self.position[2]),
            ),
        )
        if self.predelay != 0:
            node.setAttribute("predelay", utils.numberToStr(self.predelay))
        if self.complexityLevel != None:
            node.setAttribute("complexityLevel", utils.numberToStr(self.complexityLevel))

        for i in range(len(self.freqs)):
            fnode = dom.createElement("Frequency")
            fnode.setAttribute("f", utils.numberToStr(self.freqs[i]))
            fnode.setAttribute("r", utils.numberToStr(self.rt60[i]))
            if self.dsr[i] != 0:
                fnode.setAttribute("dsr", utils.numberToStr(self.dsr[i]))
            node.appendChild(fnode)

        return node

    def __str__(self):
        return (
            "AcousticParameters {\n\tposition = %s,\n\tpredelay = %f,\n\tcomplexityLevel = %f,\n\tfreqs = %s,\n\trt60 = %s,\n\tdsr = %s\n}"
            % (self.position, self.predelay, self.complexityLevel, self.freqs, self.rt60, self.dsr)
        )
