# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2023.
import numpy as np

from . import validation
from . import utils


class HeadphoneEQ:
    def __init__(
        self,
        specType=None,
        f=None,
        gain=None,
        q=None,
        filtType=None,
        preset=None,
    ):
        """Creates a content creator headphone EQ
        :param specType         Defines how the EQ is specified, either 'freq-gain', 'parametricFilters', 'preset'.
        :param f                List of frequencies [Hz]
        :param gain             List of gains [dB]
        :param q                List of filter q [float, 0..10]
        :param filtType         List of filter types if specType is parametricFilters [string, 'lowshelf', 'peak', and 'highshelf']
        :param preset           Integer index into the table of presets [int]

        ::note Length of gain and (optionally) q must match the list of frequencies.
        """
        self.specType = None
        self.f = None
        self.gain = None
        self.q = None
        self.filtType = None
        self.preset = None

        if specType != None:
            if specType.lower() in ["freqgain", "parametricfilters", "preset"]:
                self.specType = specType.lower()
            else:
                raise ValueError(f"Unknown headphone EQ specification type")

            if self.specType == "freqgain" or self.specType == "parametricfilters":
                self.f = validation.validateNumberList(f, "Headphone EQ frequencies")
                if np.any(self.f < 0):
                    raise ValueError(
                        f"HeadphoneEQ: Frequencies must be greater than or equal to zero"
                    )
                self.f = list(self.f)
                self.gain = validation.validateNumberList(gain, "Headphone EQ Gain")
                if len(self.gain) != len(self.f):
                    raise ValueError(f"Headphone EQ: Wrong number of gains")
                self.gain = list(self.gain)

                if self.specType == "parametricfilters":
                    self.q = validation.validateNumberList(q, "Headphone EQ filter Q values")
                    if len(self.q) != len(self.f):
                        raise ValueError(f"Headphone EQ: Wrong number of filter Q values.")
                    self.filtType = list(filtType)
                    if len(self.filtType) != len(self.f):
                        raise ValueError(
                            f"Headphone EQ: Wrong number of parametric filter types specified."
                        )
                    self.q = list(self.q)
            elif self.specType == "preset":
                self.preset = int(preset)

    @staticmethod
    def fromxml(node):
        """Construct an AcousticMaterial from the given XML node"""

        specType = node.getAttribute("specificationType").strip()
        f, gain, q, filtType = [], [], [], []

        for n in node.getElementsByTagName("FrequencyGain"):
            f.append(float(n.getAttribute("f").strip()))
            if n.hasAttribute("gainDb"):
                gain.append(float(n.getAttribute("gainDb").strip()))
            if n.hasAttribute("Q"):
                q.append(float(n.getAttribute("Q").strip()))
            if n.hasAttribute("type"):
                filtType.append(n.getAttribute("type").strip())

        if node.hasAttribute("preset"):
            preset = int(node.getAttribute("preset").strip())
        else:
            preset = None

        return HeadphoneEQ(
            specType,
            f=f,
            gain=gain,
            q=q,
            filtType=filtType,
            preset=preset,
        )

    def toxml(self, dom):
        node = dom.createElement("ContentCreatorHeadphones")
        node.setAttribute("specificationType", self.specType)

        for i in range(len(self.f)):
            fnode = dom.createElement("FrequencyGain")
            fnode.setAttribute("f", utils.numberToStr(self.f[i]))
            if self.gain[i] != 0:
                fnode.setAttribute("gainDb", utils.numberToStr(self.gain[i]))
            if self.q[i] != 0:
                fnode.setAttribute("Q", utils.numberToStr(self.q[i]))
            if self.filtType[i] != None:
                fnode.setAttribute("filtType", self.filtType[i])
        if self.preset != None:
            node.setAttribute("preset", self.preset)

        return node

    def __str__(self):
        return "HeadphoneEQ {\n\tspecType = %s,\n\tf = %s,\n\tgain = %s,\n\tq = %s,\n\tfilterType = %s,\n\tpreset = %s}" % (
            self.specType,
            self.f,
            self.gain,
            self.q,
            self.filtType,
            self.preset,
        )
