# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from . import validation

from .informedsource import InformedSource


class HoaGroup:
    def __init__(
        self,
        id,
        region=None,
        coSources=[],
        informedSources=[],
        exteriorProjectionRadius=4.0,
        adaptiveExteriorProjectionRadius=True,
    ):
        """Creates a HOAGroup"""
        self.id = validation.validateID(id)
        self.region = region
        self.coSources = coSources
        self.informedSources = informedSources
        self.exteriorProjectionRadius = exteriorProjectionRadius
        self.adaptiveExteriorProjectionRadius = adaptiveExteriorProjectionRadius

    @staticmethod
    def fromxml(node):
        """Deserialises a HOAGroup from an XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("HOAGroup is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if node.hasAttribute("region"):
            region = node.getAttribute("region").strip()
        else:
            region = None
        coSources = []
        if node.hasAttribute("coSources"):
            coSourcesString = node.getAttribute("coSources").strip()
            for coSource in coSourcesString.split(", "):
                coSources.append(coSource)

        informedSources = []
        for n in node.getElementsByTagName("InformedSource"):
            src = InformedSource.fromxml(n)
            informedSources.append(src)

        exteriorProjectionRadius = 4.0
        if node.hasAttribute("exteriorProjectionRadius"):
            exteriorProjectionRadius = float(node.getAttribute("exteriorProjectionRadius").strip())

        adaptiveExteriorProjectionRadius = True
        if node.hasAttribute("adaptiveExteriorProjectionRadius"):
            adaptiveExteriorProjectionRadius = utils.parseBool(
                node.getAttribute("adaptiveExteriorProjectionRadius").strip()
            )

        return HoaGroup(
            id,
            region,
            coSources,
            informedSources,
            exteriorProjectionRadius,
            adaptiveExteriorProjectionRadius,
        )

    def toxml(self, dom):
        """Serialises a HOAGroup to an XML node"""
        node = dom.createElement("HOAGroup")

        node.setAttribute("id", self.id)

        if self.region != None:
            node.setAttribute("region", self.region)

        if self.coSources != []:
            coSources = ", ".join(self.coSources)
            node.setAttribute("coSources", coSources)

        for src in self.informedSources:
            node.appendChild(src.toxml(dom))

        node.setAttribute("exteriorProjectionRadius", self.exteriorProjectionRadius)
        node.setAttribute("adaptiveExteriorProjectionRadius", self.adaptiveExteriorProjectionRadius)

        return node

    def validate(self, objSources):
        """Validate the cosources by checking if they exist in the object sources"""
        for coSource in self.coSources:
            if objSources.get(coSource) == None:
                raise ValueError(f"The coSource '{coSource}' does not exist")

    def __str__(self):
        pass


class HoaSource:
    def __init__(
        self,
        id,
        signal,
        representation,
        position=[0.0, 0.0, 0.0],
        orientation=[0.0, 0.0, 0.0],
        cspace="relative",
        active=True,
        gainDb=0.0,
        refDistance=None,
        extent=None,
        extentTransform=True,
        transitionDistance=0.0,
        group=None,
        aparams=None,
        mode="continuous",
        play=False,
        transformID=None,
        anchorID=None,
        vdlMethod=0,
        hasReverbGain=False,
        reverbGainDb=0.0,
        noReverb=False,
        granularPos=None,
    ):
        """Creates a HOASource"""

        self.id = validation.validateID(id)
        self.position = validation.validateVector3(position, f"position in HOASource '{self.id}'")
        self.orientation = validation.validateVector3(
            orientation, f"orientation in HOASource '{self.id}'"
        )
        self.cspace = validation.validateCSpace(cspace)
        self.active = active
        self.gainDb = gainDb

        self.signal = signal
        self.extent = extent
        self.extentTransform = extentTransform
        if (representation != 0) and (representation != 1):
            raise ValueError(
                f"Representation in HOASource '{self.id}'must have either a value of 0 or 1"
            )
        self.representation = representation
        if refDistance == None:
            refDistance = self.representation  # either 0 or 1 implied
        elif refDistance < 0:
            raise ValueError(f"Invalid RefDistance in HOASource '{self.id}': {refDistance}")
        self.refDistance = refDistance
        if transitionDistance < 0:
            raise ValueError(f"Transition distance in HOASource '{self.id}' must be positive")
        self.transitionDistance = transitionDistance
        self.group = group
        self.aparams = aparams
        self.mode = validation.validatePlaybackMode(mode)
        self.play = validation.validateBool(play, f"play flag {self.id}")
        self.transform = transformID
        self.anchor = anchorID
        self.vdlMethod = vdlMethod
        self.hasReverbGain = hasReverbGain
        self.reverbGainDb = reverbGainDb
        self.noReverb = noReverb
        if granularPos != None:
            self.granularPos = validation.validateVector3(
                granularPos, f"Granular position in HOASource '{self.id}'"
            )
        else:
            self.granularPos = granularPos

    @staticmethod
    def fromxml(node):
        """Deserialises a HOASource from an XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("HOASource has missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("position"):
            raise SyntaxError(f"HOASource '{id}' has missing Position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0.0, 0.0, 0.0]

        if node.hasAttribute("cspace"):
            cspace = node.getAttribute("cspace").strip()
        else:
            cspace = "relative"

        if node.hasAttribute("active"):
            active = node.getAttribute("active").strip()
        else:
            active = True

        if node.hasAttribute("gainDb"):
            gainDb = float(node.getAttribute("gainDb").strip())
        else:
            gainDb = 0.0

        if node.hasAttribute("refDistance"):
            refDistance = float(node.getAttribute("refDistance").strip())
        else:
            refDistance = 1.0

        if not node.hasAttribute("signal"):
            raise SyntaxError(f"HOASource '{id}' has missing signal in EIF file")
        signal = node.getAttribute("signal").strip()

        if node.hasAttribute("extent"):
            extent = node.getAttribute("extent").strip()
        else:
            extent = None

        if node.hasAttribute("extentTransform"):
            extentTransform = utils.parseBool(node.getAttribute("extentTransform").strip())
        else:
            extentTransform = True

        if not node.hasAttribute("representation"):
            raise SyntaxError(f"HOASource '{id}' has missing representation attribute in EIF file")
        else:
            representation = int(node.getAttribute("representation").strip())

        if node.hasAttribute("transitionDistance"):
            transitionDistance = float(node.getAttribute("transitionDistance").strip())
        else:
            transitionDistance = 1.0

        if node.hasAttribute("group"):
            group = node.getAttribute("group").strip()
        else:
            group = None

        if node.hasAttribute("aparams"):
            aparams = utils.parseAuthoringParameters(node.getAttribute("aparams").strip())
        else:
            aparams = None

        if node.hasAttribute("mode"):
            mode = utils.parsePlaybackMode(node.getAttribute("mode").strip())
        else:
            mode = "continuous"

        if node.hasAttribute("play"):
            play = utils.parseBool(node.getAttribute("play").strip())
        else:
            play = False

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        if node.hasAttribute("vdlMethod"):
            from ...framework import _lookup_tables as LUT

            vdlMethod = LUT.vdl_method_codes[node.getAttribute("vdlMethod").strip()]
        else:
            vdlMethod = 0  # Default 0 means "spline" method

        if node.hasAttribute("granularPos"):
            granularPos = node.getAttribute("granularPos").strip()
        else:
            granularPos = None

        if node.hasAttribute("reverbGainDb"):
            reverbGainDb = float(node.getAttribute("reverbGainDb").strip())
            assert (
                reverbGainDb >= -119
            ), "reverbGainDb cannot be smaller than -119 dB ! Source: " + str(id)
            assert reverbGainDb <= 12, "reverbGainDb cannot be grater than +12 dB ! Source: " + str(
                id
            )
            hasReverbGain = True
        else:
            hasReverbGain = False
            reverbGainDb = 0.0

        if node.hasAttribute("muteReverb"):
            noReverb = utils.parseBool(node.getAttribute("muteReverb").strip())
        else:
            noReverb = False

        return HoaSource(
            id,
            signal=signal,
            representation=representation,
            position=position,
            orientation=orientation,
            cspace=cspace,
            active=active,
            gainDb=gainDb,
            refDistance=refDistance,
            extent=extent,
            extentTransform=extentTransform,
            transitionDistance=transitionDistance,
            group=group,
            aparams=aparams,
            mode=mode,
            play=play,
            transformID=transform,
            anchorID=anchor,
            vdlMethod=vdlMethod,
            hasReverbGain=hasReverbGain,
            reverbGainDb=reverbGainDb,
            noReverb=noReverb,
            granularPos=granularPos,
        )

    def toxml(self, dom):
        """Serialises a HOASource to an XML node"""
        node = dom.createElement("HOASource")

        node.setAttribute("id", self.id)
        node.setAttribute(
            "position",
            "%s %s %s"
            % (
                utils.numberToStr(self.position[0]),
                utils.numberToStr(self.position[1]),
                utils.numberToStr(self.position[2]),
            ),
        )
        if self.orientation[0] != 0 or self.orientation[1] != 0 or self.orientation[2] != 0:
            node.setAttribute(
                "orientation",
                "%s %s %s"
                % (
                    utils.numberToStr(self.orientation[0]),
                    utils.numberToStr(self.orientation[1]),
                    utils.numberToStr(self.orientation[2]),
                ),
            )
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)
        if self.active != True:
            node.setAttribute("active", "False")
        if self.gainDb != 0:
            node.setAttribute("gainDb", utils.numberToStr(self.gainDb))
        if self.refDistance != 1:
            node.setAttribute("refDistance", utils.numberToStr(self.refDistance))
        node.setAttribute("signal", self.signal)
        if self.extent is not None:
            node.setAttribute("extent", self.extent.id)
        if self.extentTransform is not True:
            node.setAttribute("extentTransform", self.extentTransform)
        node.setAttribute("representation", self.representation)
        if self.transitionDistance != 1.0:
            node.setAttribute("transitionDistance", self.transitionDistance)
        if self.group is not None:
            node.setAttribute("group", self.group)
        if self.mode != "continuous":
            node.setAttribute("mode", self.mode)
        if self.play != False:
            node.setAttribute("play", "true")
        if self.aparams is not None:
            node.setAttribute("aparams", self.aparams)
        if self.vdlMethod is not None:
            node.setAttribute("vdlMethod", utils.numberToStr(self.vdlMethod))
        if self.reverbGainDb is not None:
            assert (
                self.reverbGainDb >= -119
            ), "reverbGainDb cannot be smaller than -119 dB ! Source: " + str(id)
            assert (
                self.reverbGainDb <= 12
            ), "reverbGainDb cannot be grater than +12 dB ! Source: " + str(id)
            node.setAttribute("reverbGainDb", utils.numberToStr(self.reverbGainDb))
            self.hasReverbGain = True
        if self.noReverb != False:
            node.setAttribute("muteReverb", "true")
        if self.granularPos is not None:
            node.setAttribute(
                "granularPos",
                "%s %s %s"
                % (
                    utils.numberToStr(self.granularPos[0]),
                    utils.numberToStr(self.granularPos[1]),
                    utils.numberToStr(self.granularPos[2]),
                ),
            )
        pass

    def validate(self, hoaGroups):
        """Validates if the parent hoaGroup exists"""
        if self.group != None:
            if hoaGroups.get(self.group) == None:
                raise ValueError(f"The HOAGroup '{self.group}' does not exist.")

    def __str__(self):
        pass
