from . import utils
from . import validation


class InformedSource:
    def __init__(
        self,
        id,
        position=[0.0, 0.0, 0.0],
        listenerThreshold=3.0,
        hoaSourceThreshold=3.0,
        priorityValue=0,
    ):
        """Creates an InformedSource"""

        self.id = validation.validateID(id)
        self.position = validation.validateVector3(
            position, f"position in InformedSource '{self.id}'"
        )

        self.listenerThreshold = listenerThreshold
        self.hoaSourceThreshold = hoaSourceThreshold
        self.priorityValue = priorityValue

    @staticmethod
    def fromxml(node):
        """Deserialises an InformedSource from an XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("InformedSource has missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("position"):
            raise SyntaxError(f"InformedSource '{id}' has missing Position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("listenerThreshold"):
            listenerThreshold = float(node.getAttribute("listenerThreshold").strip())
            if listenerThreshold < 0.0:
                raise SyntaxError(f"InformedSource '{id}' has listenerThreshold that is negative")
        else:
            listenerThreshold = 3.0

        if node.hasAttribute("hoaSourceThreshold"):
            hoaSourceThreshold = float(node.getAttribute("hoaSourceThreshold").strip())
            if hoaSourceThreshold < 0.0:
                raise SyntaxError(f"InformedSource '{id}' has hoaSourceThreshold that is negative")
        else:
            hoaSourceThreshold = 3.0

        if node.hasAttribute("priorityValue"):
            priorityValue = int(node.getAttribute("priorityValue").strip())
        else:
            priorityValue = 0

        return InformedSource(
            id,
            position=position,
            listenerThreshold=listenerThreshold,
            hoaSourceThreshold=hoaSourceThreshold,
            priorityValue=priorityValue,
        )

    def toxml(self, dom):
        """Serialises a InformedSource to an XML node"""
        node = dom.createElement("InformedSource")

        node.setAttribute("id", self.id)
        node.setAttribute(
            "position",
            "%s %s %s"
            % (
                utils.numberToStr(self.position[0]),
                utils.numberToStr(self.position[1]),
                utils.numberToStr(self.position[2]),
            ),
        )
        if self.listenerThreshold != 3.0:
            node.setAttribute("listenerThreshold", self.listenerThreshold)
        if self.hoaSourceThreshold != 3.0:
            node.setAttribute("hoaSourceThreshold", self.hoaSourceThreshold)
        if self.priorityValue != 0:
            node.setAttribute("priorityValue", self.priorityValue)
        pass

    def __str__(self):
        return 'InformedSource "%s" {\n\tposition = %s,\n\tlistenerThreshold = %s,\n\thoaSourceThreshold = %s,\n\tpriorityValue = %s\n}' % (
            self.id,
            utils.positionToString(self.position),
            self.listenerThreshold,
            self.hoaSourceThreshold,
            self.priorityValue,
        )
