# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np

from . import validation
from . import utils


class Material:
    def __init__(self, id, f, r=None, s=None, t=None, c=None):
        """Creates an acoustic material
        :param id   Material ID
        :param f    List of frequencies [Hz]
        :param r    List of reflection factors [0..1]
        :param s    List of scattering coefficients [0..1]
        :param t    List of transmission coefficients [0..1]
        :param c    List of coupling coefficients [0..1]

        ::note All lists must match the length of the frequencies
        """
        self.id = validation.validateID(id)
        self.f = validation.validateNumberList(f, "frequencies")
        if np.any(self.f < 0):
            raise ValueError(f"Material '{id}': Frequencies must be greater than or equal to zero")

        if not r and not s and not t and not c:
            raise ValueError(
                f"Material '{id}': At least one parameter (r, s, t, c) must be specified"
            )

        if r:
            self.r = validation.validateNumberList(r, f"reflection factors in Material '{id}'")
            if len(self.r) != len(self.f):
                raise ValueError(f"Material '{id}': Wrong number of reflection factors")
            if np.any(self.r < 0) or np.any(self.r > 1):
                raise ValueError(f"Material '{id}': Reflection factors must be in range [0..1]")
        else:
            self.r = np.zeros_like(self.f)

        if s:
            self.s = validation.validateNumberList(s, f"scattering coefficients in Material '{id}'")
            if len(self.s) != len(self.f):
                raise ValueError(f"Material '{id}': Wrong number of scattering coefficients")
            if np.any(self.s < 0) or np.any(self.s > 1):
                raise ValueError(
                    f"Material '{id}': Scattering coefficients must be in range [0..1]"
                )
        else:
            self.s = np.zeros_like(self.f)

        if t:
            self.t = validation.validateNumberList(
                t, f"transmission coefficients in Material '{id}'"
            )
            if len(self.t) != len(self.f):
                raise ValueError(f"Material '{id}': Wrong number of transmission coefficients")
            if np.any(self.t < 0) or np.any(self.t > 1):
                raise ValueError(
                    f"Material '{id}': Transmission coefficients must be in range [0..1]"
                )
        else:
            self.t = np.zeros_like(self.f)

        if c:
            self.c = validation.validateNumberList(c, f"coupling coefficients in Material '{id}'")
            if len(self.c) != len(self.f):
                raise ValueError(f"Material '{id}': Wrong number of coupling coefficients")
            if np.any(self.c < 0) or np.any(self.c > 1):
                raise ValueError(f"Material '{id}': Coupling coefficients must be in range [0..1]")
        else:
            self.c = np.zeros_like(self.f)

    @staticmethod
    def fromxml(node):
        """Construct an AcousticMaterial from the given XML node"""

        if not node.hasAttribute("id"):
            raise SyntaxError("AcousticMaterial is missing ID in EIF file")
        id = node.getAttribute("id")
        f, r, s, t, c = [], [], [], [], []

        for n in node.getElementsByTagName("Frequency"):
            f.append(float(n.getAttribute("f").strip()))
            if n.hasAttribute("r"):
                r.append(float(n.getAttribute("r").strip()))
            else:
                r.append(0)
            if n.hasAttribute("s"):
                s.append(float(n.getAttribute("s").strip()))
            else:
                s.append(0)
            if n.hasAttribute("t"):
                t.append(float(n.getAttribute("t").strip()))
            else:
                t.append(0)
            if n.hasAttribute("c"):
                c.append(float(n.getAttribute("c").strip()))
            else:
                c.append(0)

        return Material(id, f, r=r, s=s, t=t, c=c)

    def toxml(self, dom):
        node = dom.createElement("AcousticMaterial")
        node.setAttribute("id", self.id)

        for i in range(len(self.f)):
            fnode = dom.createElement("Frequency")
            fnode.setAttribute("f", utils.numberToStr(self.f[i]))
            if self.r[i] != 0:
                fnode.setAttribute("r", utils.numberToStr(self.r[i]))
            if self.s[i] != 0:
                fnode.setAttribute("s", utils.numberToStr(self.s[i]))
            if self.t[i] != 0:
                fnode.setAttribute("t", utils.numberToStr(self.t[i]))
            if self.c[i] != 0:
                fnode.setAttribute("c", utils.numberToStr(self.c[i]))
            node.appendChild(fnode)

        return node

    def __str__(self):
        return 'Material "%s" {\n\tf = %s,\n\tr = %s,\n\ts = %s,\n\tt = %s,\n\tc = %s\n}' % (
            self.id,
            self.f,
            self.r,
            self.s,
            self.t,
            self.c,
        )
