# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from . import validation

import numpy as np


class Mesh:
    def __init__(
        self,
        id,
        vertices=None,
        faces=None,
        materials=None,
        position=[0, 0, 0],
        orientation=[0, 0, 0],
        cspace="relative",
        transformID=None,
        anchorID=None,
    ):
        """Creates a mesh
        :param transform Parent spatial transform
        """
        self.id = validation.validateID(id)
        self.position = validation.validateVector3(position, f"position in Mesh '{self.id}'")
        self.orientation = validation.validateVector3(
            orientation, f"orientation in Mesh '{self.id}'"
        )
        self.cspace = validation.validateCSpace(cspace)

        self.__vertices = []
        self.__verticesGC = []  # Vertices expressed in global coordinates
        self.faces = []
        self.materials = []

        self.transform = transformID
        self.anchor = anchorID

        if vertices:
            if not isinstance(vertices, list):
                raise ValueError(f"Mesh '{self.id}': Vertices must be a list")
            for v in vertices:
                self.addVertex(v)

        if faces:
            if not vertices:
                raise ValueError(f"Mesh '{self.id}': Faces cannot be defined without vertices")
            if not isinstance(faces, list):
                raise ValueError(f"Mesh '{self.id}': Faces must be a list")

            if materials:
                if not isinstance(materials, list):
                    raise ValueError(f"Mesh '{self.id}': Materials must be a list")
                if not len(materials) == len(faces):
                    raise ValueError(f"Mesh '{self.id}': Materials must be of same length as Faces")

            for f, m in zip(faces, materials or [None] * len(faces)):
                self.addFace(f, material=m) if m else self.addFace(f)

    @property
    def vertices(self):
        """Return list of vertices (mesh coordinates)"""
        return self.__vertices

    @property
    def verticesGC(self):
        """Return list of vertices (global coordinates)"""
        return self.__verticesGC

    def addVertex(self, position):
        """Adds a vertex and returns its index"""
        idx = len(self.vertices)
        v = validation.validateVector3(position, f"position in Mesh '{self.id}'")
        self.vertices.append(v)

        return idx

    def addFace(self, vertices, material=None):
        """Adds a face and returns its index"""
        idx = len(self.faces)
        vids = validation.validateIntList(vertices, "vertices")
        if len(vids) != 3:
            raise ValueError(f"Mesh '{self.id}': A face must have three vertex indices")
        for vid in vids:
            if vid < 0 or vid >= len(self.vertices):
                raise ValueError(f"Mesh '{self.id}': Vertex index '{vid}' out of range")
        self.faces.append(vids)
        self.materials.append(material)
        return idx

    def removeFace(self, faceIdx):
        """Removes a face"""
        self.faces.pop(faceIdx)
        self.materials.pop(faceIdx)

    def getFaceNormal(self, faceIdx):
        """Returns the normal vector of a face according to OpenGL/EIF convention
        n = (v2 - v1) x (v3 - v1)
        :param faceIdx  the face index as returned by addFace
        :returns        numpy vector of length 3 with normal in global coordinates
        """
        v1 = np.array(self.vertices[self.faces[faceIdx][0]])
        v2 = np.array(self.vertices[self.faces[faceIdx][1]])
        v3 = np.array(self.vertices[self.faces[faceIdx][2]])
        n = np.cross((v2 - v1), (v3 - v1))
        return n / np.linalg.norm(n)

    def getFaceCentroid(self, faceIdx):
        """Returns the centroid of a face
        :param faceIdx  the face index as returned by addFace
        :returns        numpy vector of length 3 with centroid in global coordinates
        """
        v1 = np.array(self.vertices[self.faces[faceIdx][0]])
        v2 = np.array(self.vertices[self.faces[faceIdx][1]])
        v3 = np.array(self.vertices[self.faces[faceIdx][2]])
        return (v1 + v2 + v3) / 3

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Mesh has Missing ID from EIF file")
        id = node.getAttribute("id").strip()

        if node.hasAttribute("position"):
            position = node.getAttribute("position").strip()
        else:
            position = [0, 0, 0]

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0, 0, 0]

        if node.hasAttribute("cspace"):
            cspace = node.getAttribute("cspace").strip()
        else:
            cspace = "relative"

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        # Construct the mesh
        mesh = Mesh(
            id=id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            transformID=transform,
            anchorID=anchor,
        )

        # Add the vertices, while keeping track of their labels (ids)
        vertexNodes = node.getElementsByTagName("Vertex")
        if not vertexNodes:
            raise SyntaxError(f"Mesh '{id}' contains no vertices")
        vertexIndexMap = {}
        for v in vertexNodes:
            if not v.hasAttribute("index"):
                raise SyntaxError(f"Mesh '{id}', Vertex '{v}': missing index in EIF file")
            src_idx = int(v.getAttribute("index").strip())

            if not v.hasAttribute("position"):
                raise SyntaxError(f"Mesh '{id}', Vertex '{v}': missing position in EIF file")
            position = v.getAttribute("position").strip()

            dst_idx = mesh.addVertex(position)
            vertexIndexMap[src_idx] = dst_idx

        # Add the faces, thereby translate the vertex indices
        faceNodes = node.getElementsByTagName("Face")
        if not faceNodes:
            raise SyntaxError(f"Mesh '{id}' contains no faces in EIF file")
        for f in faceNodes:
            if not f.hasAttribute("vertices"):
                raise SyntaxError(f"Mesh '{id}', Face '{f}': missing vertices in EIF file")
            v = validation.validateNumberList(f.getAttribute("vertices").strip(), "vertices")
            new_idx = []
            for vidx in v:
                if vidx not in vertexIndexMap.keys():
                    raise ValueError(f"Mesh '{id}', Face '{f}': references undefined vertex")
                new_idx.append(vertexIndexMap[vidx])

            if f.hasAttribute("material"):
                material = f.getAttribute("material").strip()
            else:
                material = None

            mesh.addFace(new_idx, material)

        return mesh

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Mesh")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        node.setAttribute("position", utils.vectorToAttributeStr(self.position))
        if np.any(self.orientation != 0):
            node.setAttribute("orientation", utils.vectorToAttributeStr(self.orientation))
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)

        for i in range(len(self.vertices)):
            vnode = dom.createElement("Vertex")
            vnode.setAttribute("index", utils.numberToStr(i))
            vnode.setAttribute("position", utils.vectorToAttributeStr(self.vertices[i]))
            node.appendChild(vnode)

        for i in range(len(self.faces)):
            fnode = dom.createElement("Face")
            fnode.setAttribute("index", utils.numberToStr(i))
            fnode.setAttribute(
                "vertices", "%d %d %d" % (self.faces[i][0], self.faces[i][1], self.faces[i][2])
            )
            if self.materials[i] != None:
                fnode.setAttribute("material", self.materials[i])
            node.appendChild(fnode)

        return node

    def __str__(self):
        return 'Mesh "%s" {\n\tposition = %s,\n\torientation = %s,\n\tcspace = %s,\n\t%d vertices,\n\t%d faces\n}' % (
            self.id,
            utils.positionToString(self.position),
            utils.orientationToString(self.orientation),
            self.cspace,
            len(self.vertices),
            len(self.faces),
        )


class MeshFace:
    """
    Data class addressing a single face within an EIF mesh

    :note This class is handy when addressing individual faces in a mesh (e.g. for Trimesh)
    """

    def __init__(self, mesh, face_index):
        if not isinstance(mesh, Mesh):
            raise ValueError("Expecting an EIF mesh")
        if not isinstance(face_index, int):
            raise ValueError("Face index must be an integer")
        if face_index < 0 or face_index >= len(mesh.faces):
            raise ValueError("Face index out of range")

        self.__mesh = mesh
        self.__face_index = face_index

    @property
    def mesh(self):
        return self.__mesh

    @property
    def face_index(self):
        return self.__face_index

    def __str__(self):
        return 'MeshFace { mesh "%s", face %d }' % (self.mesh.id, self.face_index)
