# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from . import validation


class InputLayout:
    def __init__(self, alignment, positioning):
        """Creates an InputLayout"""
        self.alignment = validation.validateAlignment(alignment)
        self.positioning = validation.validatePositioning(positioning)

    @staticmethod
    def fromxml(node):
        """Deserialises an InputLayout from an XML node"""
        if not node.hasAttribute("alignment"):
            raise ValueError("InputLayout must have an alignment attribute")
        alignment = node.getAttribute("alignment").strip()

        if not node.hasAttribute("positioning"):
            raise ValueError("InputLayout must have a positioning attribute")
        positioning = node.getAttribute("positioning").strip()

        return InputLayout(alignment, positioning)

    def toxml(self, dom):
        """Serialises to an XML node"""
        node = dom.createElement("InputLayout")

        node.setAttribute("alignment", self.alignment)
        node.setAttribute("positioning", self.positioning)

        return node

    def __str__(self):
        return f"(alignment = {self.alignment}, positioning = {self.positioning})"


class ObjectSource:
    def __init__(
        self,
        id,
        position,
        signal,
        orientation=[0, 0, 0],
        cspace="relative",
        active=True,
        gainDb=0,
        refDistance=1,
        extent=None,
        directivity=None,
        directiveness=1,
        mode="continuous",
        play=False,
        inputLayout=None,
        aparams=None,
        transformID=None,
        anchorID=None,
        vdlMethod=0,
        hasReverbGain=False,
        reverbGainDb=0.0,
        noReverb=False,
        granularPos=None,
    ):
        self.id = validation.validateID(id)
        self.position = validation.validateVector3(
            position, f"position in ObjectSource '{self.id}'"
        )
        self.orientation = validation.validateVector3(
            orientation, f"orientation in ObjectSource '{self.id}'"
        )
        self.cspace = validation.validateCSpace(cspace)
        self.active = validation.validateBool(active, f"active flag in '{self.id}'")
        self.gainDb = gainDb
        if refDistance <= 0:
            raise ValueError(
                f"ObjectSource '{self.id}': Reference distance must be positive "
                f"and not '{refDistance}'"
            )
        self.refDistance = refDistance
        self.signal = signal
        self.extent = extent
        self.directivity = directivity
        if directiveness < 0:
            raise ValueError(
                f"ObjectSource '{self.id}': Directiveness must not be negative "
                "and not '{directiveness}'"
            )
        self.directiveness = directiveness
        self.mode = validation.validatePlaybackMode(mode)
        self.play = validation.validateBool(play, f"play flag in ObjectSource '{self.id}'")
        self.inputLayout = inputLayout
        self.aparams = aparams
        self.transform = transformID
        self.anchor = anchorID
        self.vdlMethod = vdlMethod
        self.hasReverbGain = hasReverbGain
        self.reverbGainDb = reverbGainDb
        self.noReverb = noReverb
        if granularPos != None:
            self.granularPos = validation.validateVector3(
                granularPos, f"Granular position in ObjectSource '{self.id}'"
            )
        else:
            self.granularPos = granularPos

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""

        if not node.hasAttribute("id"):
            raise SyntaxError("ObjectSource is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("position"):
            raise SyntaxError(f"ObjectSource '{id}' is missing position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0, 0, 0]

        if node.hasAttribute("cspace"):
            cspace = node.getAttribute("cspace").strip()
        else:
            cspace = "relative"

        if node.hasAttribute("active"):
            active = utils.parseBool(node.getAttribute("active").strip())
        else:
            active = True

        if node.hasAttribute("gainDb"):
            gainDb = float(node.getAttribute("gainDb").strip())
        else:
            gainDb = 0

        if node.hasAttribute("refDistance"):
            refDistance = float(node.getAttribute("refDistance").strip())
        else:
            refDistance = 1

        if not node.hasAttribute("signal"):
            raise SyntaxError(f"ObjectSource '{id}' has missing signal in EIF file")
        signal = node.getAttribute("signal").strip()

        if node.hasAttribute("extent"):
            extent = node.getAttribute("extent").strip()
        else:
            extent = None

        if node.hasAttribute("directivity"):
            directivity = node.getAttribute("directivity").strip()
        else:
            directivity = None

        if node.hasAttribute("directiveness"):
            directiveness = float(node.getAttribute("directiveness").strip())
            if directiveness < 0:
                raise ValueError(f"Invalid directiveness value - Less than 0: {directiveness}")
        else:
            directiveness = 1

        if node.hasAttribute("mode"):
            mode = utils.parsePlaybackMode(node.getAttribute("mode").strip())
        else:
            mode = "continuous"

        if node.hasAttribute("play"):
            play = utils.parseBool(node.getAttribute("play").strip())
        else:
            play = False

        inputLayoutNode = node.getElementsByTagName("InputLayout")
        if len(inputLayoutNode) > 1:
            raise SyntaxError(f"ObjectSource '{id}' has more than one input layout")
        elif len(inputLayoutNode) == 1:
            inputLayout = InputLayout.fromxml(inputLayoutNode[0])
        else:
            inputLayout = None

        if node.hasAttribute("aparams"):
            aparams = utils.parseAuthoringParameters(node.getAttribute("aparams").strip())
        else:
            aparams = None

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        if node.hasAttribute("vdlMethod"):
            from ...framework import _lookup_tables as LUT

            vdlMethod = LUT.vdl_method_codes[node.getAttribute("vdlMethod").strip()]
        else:
            vdlMethod = 0  # Default 0 means "spline" method

        if node.hasAttribute("granularPos"):
            granularPos = node.getAttribute("granularPos").strip()
        else:
            granularPos = None

        if node.hasAttribute("reverbGainDb"):
            reverbGainDb = float(node.getAttribute("reverbGainDb").strip())
            assert (
                reverbGainDb >= -119
            ), "reverbGainDb cannot be smaller than -119 dB ! Source: " + str(id)
            assert reverbGainDb <= 12, "reverbGainDb cannot be grater than +12 dB ! Source: " + str(
                id
            )
            hasReverbGain = True
        else:
            hasReverbGain = False
            reverbGainDb = 0.0

        if node.hasAttribute("muteReverb"):
            noReverb = utils.parseBool(node.getAttribute("muteReverb").strip())
        else:
            noReverb = False

        return ObjectSource(
            id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            active=active,
            gainDb=gainDb,
            refDistance=refDistance,
            signal=signal,
            extent=extent,
            directivity=directivity,
            directiveness=directiveness,
            mode=mode,
            play=play,
            inputLayout=inputLayout,
            aparams=aparams,
            transformID=transform,
            anchorID=anchor,
            vdlMethod=vdlMethod,
            hasReverbGain=hasReverbGain,
            reverbGainDb=reverbGainDb,
            noReverb=noReverb,
            granularPos=granularPos,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("ObjectSource")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        node.setAttribute(
            "position",
            "%s %s %s"
            % (
                utils.numberToStr(self.position[0]),
                utils.numberToStr(self.position[1]),
                utils.numberToStr(self.position[2]),
            ),
        )
        if self.orientation[0] != 0 or self.orientation[1] != 0 or self.orientation[2] != 0:
            node.setAttribute(
                "orientation",
                "%s %s %s"
                % (
                    utils.numberToStr(self.orientation[0]),
                    utils.numberToStr(self.orientation[1]),
                    utils.numberToStr(self.orientation[2]),
                ),
            )
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)
        if self.active != True:
            node.setAttribute("active", "False")
        if self.gainDb != 0:
            node.setAttribute("gainDb", utils.numberToStr(self.gainDb))
        if self.refDistance != 1:
            node.setAttribute("refDistance", utils.numberToStr(self.refDistance))
        node.setAttribute("signal", self.signal)
        if self.extent is not None:
            node.setAttribute("extent", self.extent)
        if self.directivity is not None:
            node.setAttribute("directivity", self.directivity)
        if self.directiveness != 1:
            node.setAttribute("directiveness", utils.numberToStr(self.directiveness))
        if self.mode != "continuous":
            node.setAttribute("mode", self.mode)
        if self.play != False:
            node.setAttribute("play", "true")
        if self.inputLayout is not None:
            node.appendChild(self.inputLayout.toxml(dom))
        if self.aparams is not None:
            node.setAttribute("aparams", self.aparams)
        if self.vdlMethod is not None:
            node.setAttribute("vdlMethod", utils.numberToStr(self.vdlMethod))
        if self.reverbGainDb is not None:
            assert (
                self.reverbGainDb >= -119
            ), "reverbGainDb cannot be smaller than -119 dB ! Source: " + str(id)
            assert (
                self.reverbGainDb <= 12
            ), "reverbGainDb cannot be grater than +12 dB ! Source: " + str(id)
            node.setAttribute("reverbGainDb", utils.numberToStr(self.reverbGainDb))
            self.hasReverbGain = True
        if self.noReverb != False:
            node.setAttribute("muteReverb", "true")
        if self.granularPos is not None:
            node.setAttribute(
                "granularPos",
                "%s %s %s"
                % (
                    utils.numberToStr(self.granularPos[0]),
                    utils.numberToStr(self.granularPos[1]),
                    utils.numberToStr(self.granularPos[2]),
                ),
            )
        return node

    def __str__(self):
        if self.extent:
            extentID = self.extent.id
        else:
            extentID = None
        if self.directivity:
            directivityID = self.directivity
        else:
            directivityID = None
        return 'ObjectSource "%s" {\n\tposition = %s,\n\torientation = %s,\n\tcspace = %s,\n\tactive = %s,\n\tgainDb = %f,\n\tsignal = %s,\n\trefDistance = %f,\n\textent = %s,\n\tdirectivitiy = %s,\n\tdirectiveness = %f,\n\tinputLayout = %s,\n\taparams = %s,\n\tmode = %s,\n\tplay = %s, \n\tvdlMethod = %s, \n\treverbGainDb = %s, \n\tnoReverb = %s\n}' % (
            self.id,
            utils.positionToString(self.position),
            utils.orientationToString(self.orientation),
            self.cspace,
            self.active,
            self.gainDb,
            self.signal,
            self.refDistance,
            extentID,
            directivityID,
            self.directiveness,
            self.inputLayout,
            self.aparams,
            self.mode,
            self.play,
            self.vdlMethod,
            self.hasReverbGain,
            self.reverbGainDb,
            self.noReverb,
            self.granularPos,
        )
