# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import validation
from . import utils
import numpy as np


class Portal:
    def __init__(
        self,
        id,
        portalExtent,
        portalNormal,
        acousticEnvironment1,
        acousticEnvironment2,
        material,
        portalPosition,
        portalSize,
    ):
        """Creates a Portal"""
        self.id = validation.validateID(id)
        self.portalExtent = portalExtent
        self.portalNormal = validation.validateVector3(
            portalNormal, f"portalNormal in Portal '{self.id}'"
        )
        self.acousticEnvironment1 = acousticEnvironment1
        self.acousticEnvironment2 = acousticEnvironment2
        self.material = material
        self.portalPosition = (
            validation.validateVector3(portalPosition, f"portalPosition in Portal '{self.id}'")
            if portalPosition is not None
            else None
        )
        self.portalSize = (
            validation.validateVector3(portalSize, f"portalSize in Portal '{self.id}'")
            if portalSize is not None
            else None
        )

        temp = [portalSize != None, portalPosition != None]
        if any(temp) and not all(temp):
            raise SyntaxError("Portal cannot specify only one of portalSize and portalPosition")
        elif all(temp):
            self.hasExplicitPortal = True
        else:
            self.hasExplicitPortal = False

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Portal is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if node.hasAttribute("portalExtent"):
            portalExtent = node.getAttribute("portalExtent").strip()
        else:
            portalExtent = None

        if node.hasAttribute("portalNormal"):
            portalNormal = node.getAttribute("portalNormal").strip()
        else:
            raise SyntaxError(
                f"Portal '{id}' has no portalNormal defined in EIF file as is required."
            )

        if node.hasAttribute("acousticEnvironment1"):
            acousticEnvironment1 = node.getAttribute("acousticEnvironment1").strip()
        else:
            acousticEnvironment1 = None

        if node.hasAttribute("acousticEnvironment2"):
            acousticEnvironment2 = node.getAttribute("acousticEnvironment2").strip()
        else:
            acousticEnvironment2 = None

        if node.hasAttribute("material"):
            material = node.getAttribute("material").strip()
        else:
            material = None

        if node.hasAttribute("portalPosition"):
            portalPosition = node.getAttribute("portalPosition").strip()
        else:
            portalPosition = None

        if node.hasAttribute("portalSize"):
            portalSize = node.getAttribute("portalSize").strip()
        else:
            portalSize = None

        return Portal(
            id,
            portalExtent,
            portalNormal,
            acousticEnvironment1,
            acousticEnvironment2,
            material,
            portalPosition,
            portalSize,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Portal")
        node.setAttribute("id", self.id)
        node.setAttribute("portalExtent", self.portalExtent)
        if np.any(self.portalNormal != 0):
            node.setAttribute("portalNormal", self.aepInputChannels)
        if self.acousticEnvironment1 is not None:
            node.setAttribue("acousticEnvironment1", self.acousticEnvironment1)
        if self.acousticEnvironment2 is not None:
            node.setAttribue("acousticEnvironment2", self.acousticEnvironment2)
        if self.material is not None:
            node.setAttribue("material", self.material)
        return node

    def __str__(self):
        return 'Portal "%s" {\n\tportalExtent = "%s" \n\tportalNormal = "%s" \n\tacousticEnvironment1 = "%s" \n\tacousticEnvironment2 = "%s" \n\tmaterial = "%s" }' % (
            self.id,
            self.portalExtent,
            utils.orientationToString(self.portalNormal),
            self.acousticEnvironment1,
            self.acousticEnvironment2,
            self.material,
        )
