# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np

from . import validation
from . import utils


class Box:
    def __init__(
        self,
        id,
        position,
        size,
        orientation=[0, 0, 0],
        cspace="relative",
        material=None,
        transformID=None,
        anchorID=None,
    ):
        """Creates a box"""
        self.id = validation.validateID(id)
        self.position = validation.validateVector3(position, f"position in Box '{self.id}'")
        self.orientation = validation.validateVector3(
            orientation, f"orientation in Box '{self.id}'"
        )
        self.size = validation.validateSize(size)
        self.cspace = validation.validateCSpace(cspace)
        self.material = material
        self.transform = transformID
        self.anchor = anchorID

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Box is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("position"):
            raise SyntaxError(f"Box '{id}' is missing position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0.0, 0.0, 0.0]

        if node.hasAttribute("cspace"):
            cspace = utils.parseCSpace(node.getAttribute("cspace").strip())
        else:
            cspace = "relative"

        if not node.hasAttribute("size"):
            raise SyntaxError(f"Box '{id}' is missing size in EIF file")
        size = node.getAttribute("size").strip()

        if node.hasAttribute("material"):
            material = node.getAttribute("material").strip()
        else:
            material = None

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        return Box(
            id,
            position,
            size,
            orientation=orientation,
            cspace=cspace,
            material=material,
            transformID=transform,
            anchorID=anchor,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Box")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        node.setAttribute("position", utils.vectorToAttributeStr(self.position))
        if np.any(self.orientation != 0):
            node.setAttribute("orientation", utils.vectorToAttributeStr(self.orientation))
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)
        node.setAttribute("size", utils.vectorToAttributeStr(self.size))
        if self.material is not None:
            node.setAttribute("material", self.material)

        return node

    def __str__(self):
        if self.material:
            materialID = self.material
        else:
            materialID = None
        return 'Box "%s" {\n\tposition = %s,\n\torientation = %s,\n\tcspace = %s,\n\tsize = %s,\n\tmaterial = %s\n}' % (
            self.id,
            utils.positionToString(self.position),
            utils.orientationToString(self.orientation),
            self.cspace,
            utils.positionToString(self.size),
            materialID,
        )


class Sphere:
    def __init__(
        self,
        id,
        position,
        size,
        orientation=[0, 0, 0],
        cspace="relative",
        material=None,
        transformID=None,
        anchorID=None,
    ):
        """Creates a sphere"""
        self.id = validation.validateID(id)
        self.position = validation.validateVector3(position, f"position in Sphere '{id}'")
        self.orientation = validation.validateVector3(orientation, f"orientation in Sphere '{id}'")
        self.size = validation.validateSize(size)
        self.cspace = validation.validateCSpace(cspace)
        self.material = material
        self.transform = transformID
        self.anchor = anchorID

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Sphere is missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("position"):
            raise SyntaxError(f"Sphere '{id}' is missing position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0, 0, 0]

        if node.hasAttribute("cspace"):
            cspace = utils.parseCSpace(node.getAttribute("cspace").strip())
        else:
            cspace = "relative"

        if not node.hasAttribute("size"):
            raise SyntaxError(f"Sphere '{id}' is missing size in EIF file")
        size = node.getAttribute("size").strip()

        if node.hasAttribute("material"):
            material = node.getAttribute("material").strip()
        else:
            material = None

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        return Sphere(
            id,
            position,
            size,
            orientation=orientation,
            cspace=cspace,
            material=material,
            transformID=transform,
            anchorID=anchor,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Sphere")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        node.setAttribute("position", utils.vectorToAttributeStr(self.position))
        if np.any(self.orientation != 0):
            node.setAttribute("orientation", utils.vectorToAttributeStr(self.orientation))
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)
        node.setAttribute("size", utils.vectorToAttributeStr(self.size))
        if self.material is not None:
            node.setAttribute("material", self.material)

        return node

    def __str__(self):
        if self.material:
            materialID = self.material
        else:
            materialID = None
        return 'Sphere "%s" {\n\tposition = %s,\n\torientation = %s,\n\tcspace = %s,\n\tsize = %s,\n\tmaterial = %s\n}' % (
            self.id,
            utils.positionToString(self.position),
            utils.orientationToString(self.orientation),
            self.cspace,
            utils.positionToString(self.size),
            materialID,
        )


class Cylinder:
    def __init__(
        self,
        id,
        position,
        size,
        orientation=[0, 0, 0],
        cspace="relative",
        material=None,
        transformID=None,
        anchorID=None,
    ):
        """Creates a cylinder"""
        self.id = validation.validateID(id)
        self.position = validation.validateVector3(position, f"position in Cylinder '{self.id}'")
        self.orientation = validation.validateVector3(
            orientation, f"orientation in Cylinder '{self.id}'"
        )
        self.size = validation.validateSize(size)
        self.cspace = validation.validateCSpace(cspace)
        self.material = material
        self.transform = transformID
        self.anchor = anchorID

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Cylinder missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if not node.hasAttribute("position"):
            raise SyntaxError(f"Cylinder '{id}' missing position in EIF file")
        position = node.getAttribute("position").strip()

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0, 0, 0]

        if node.hasAttribute("cspace"):
            cspace = utils.parseCSpace(node.getAttribute("cspace").strip())
        else:
            cspace = "relative"

        if not node.hasAttribute("size"):
            raise SyntaxError(f"Cylinder '{id}' is missing size in EIF file")
        size = node.getAttribute("size").strip()

        if node.hasAttribute("material"):
            material = node.getAttribute("material").strip()
        else:
            material = None

        if node.parentNode.nodeName == "Transform":
            transform = node.parentNode.getAttribute("id").strip()
        else:
            transform = None

        if node.parentNode.nodeName == "Anchor":
            anchor = node.parentNode.getAttribute("id").strip()
        else:
            anchor = None

        return Cylinder(
            id,
            position,
            size,
            orientation=orientation,
            cspace=cspace,
            material=material,
            transformID=transform,
            anchorID=anchor,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Cylinder")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        node.setAttribute("position", utils.vectorToAttributeStr(self.position))
        if np.any(self.orientation != 0):
            node.setAttribute("orientation", utils.vectorToAttributeStr(self.orientation))
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)
        node.setAttribute("size", utils.vectorToAttributeStr(self.size))
        if self.material is not None:
            node.setAttribute("material", self.material)

        return node

    def __str__(self):
        if self.material:
            materialID = self.material
        else:
            materialID = None
        return 'Cylinder "%s" {\n\tposition = %s,\n\torientation = %s,\n\tcspace = %s,\n\tsize = %s,\n\tmaterial = %s\n}' % (
            self.id,
            utils.positionToString(self.position),
            utils.orientationToString(self.orientation),
            self.cspace,
            utils.positionToString(self.size),
            materialID,
        )
