# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from xml.dom.minidom import Document, parse

from . import utils
from .acoustTuningParameters import AcousticTuningParameters
from .anchor import Anchor
from .channelsource import ChannelSource
from .conditions import ListenerProximityCondition
from .directivity import Directivity
from .granularsound import GranularSound
from .environments import AcousticEnvironment
from .material import Material
from .mesh import Mesh
from .objectsource import ObjectSource
from .primitives import Box, Cylinder, Sphere
from .hoasource import HoaSource, HoaGroup
from .informedsource import InformedSource
from .stream import AudioStream
from .transform import Transform
from .update import Update
from .airflowgenerator import AirflowGenerator
from .airflowsource import AirflowSource
from .portal import Portal
from .headphoneEQ import HeadphoneEQ


class AudioScene:
    def __init__(
        self,
        id,
        duration=0,
        type="VR",
        enableLocallyCapturedAudio=True,
        overrideListenerVoiceDirectivity=False,
        listenerVoiceDirectivityId="",
    ):
        """Creates an empty audio eifplus"""
        self.id = id.strip()
        if not self.id:
            raise SyntaxError("ID must not be empty")

        self.duration = duration
        if self.duration < 0:
            raise SyntaxError("Duration must not be negative")

        self.type = type
        if self.type != "VR" and self.type != "AR":
            raise SyntaxError(f"AudioScene type must be either 'VR' or 'AR', and not '{self.type}'")

        self.enableLocallyCapturedAudio = enableLocallyCapturedAudio
        self.overrideListenerVoiceDirectivity = overrideListenerVoiceDirectivity
        self.listenerVoiceDirectivityId = listenerVoiceDirectivityId

        self.objects = []
        self.objectsByID = {}

        self.materials = []
        self.materialsByID = {}

        self.proximityConditions = []
        self.proximityConditionsByID = {}

        self.streams = []
        self.streamsByID = {}

        self.directivities = []
        self.directivitiesByID = {}

        self.granularsound = []
        self.granularsoundByID = {}

        self.geometry = []
        self.geometryByID = {}
        self.meshes = []
        self.meshesByID = {}

        self.sources = []
        self.sourcesByID = {}

        self.objectSources = []
        self.objectSourcesByID = {}

        self.hoaGroups = []
        self.hoaGroupsByID = {}

        self.hoaSources = []
        self.hoaSourcesByID = {}

        self.informedSources = []
        self.informedSourcesByID = {}

        self.channelSources = []
        self.channelSourcesByID = {}

        self.transforms = []
        self.transformsByID = {}

        self.anchors = []
        self.anchorsByID = {}

        self.environments = []
        self.environmentsByID = {}

        self.updates = []
        self.updatesByID = {}

        self.acousticTuningParams = AcousticTuningParameters()

        self.airflowSources = []
        self.airflowSourcesByID = {}

        self.airflowGenerators = []
        self.airflowGeneratorsByID = {}

        self.portals = []
        self.portalsByID = {}

        self.headphoneEQ = HeadphoneEQ()

        self.voxelSceneIDs = []

        self.map(id, self)

    def add(self, *argv):
        """Add object to the eifplus"""
        for obj in argv:
            if obj in self.objects:
                continue
            if hasattr(obj, "id") and obj.id in self.objectsByID.keys() and obj.id is not None:
                raise ValueError('Object with ID "%s" already contained' % obj.id)

            if isinstance(obj, Directivity):
                self.directivities.append(obj)
                self.directivitiesByID[obj.id] = obj

            elif isinstance(obj, GranularSound):
                self.granularsound.append(obj)
                self.granularsoundByID[obj.id] = obj
            elif isinstance(obj, AcousticEnvironment):
                self.environments.append(obj)
                self.environmentsByID[obj.id] = obj
            elif isinstance(obj, Material):
                self.materials.append(obj)
                self.materialsByID[obj.id] = obj
            elif (
                isinstance(obj, Box)
                or isinstance(obj, Cylinder)
                or isinstance(obj, Sphere)
                or isinstance(obj, Mesh)
            ):
                self.geometry.append(obj)
                self.geometryByID[obj.id] = obj
                if isinstance(obj, Mesh):
                    self.meshes.append(obj)
                    self.meshesByID[obj.id] = obj
            elif isinstance(obj, ObjectSource):
                self.sources.append(obj)
                self.sourcesByID[obj.id] = obj

                self.objectSources.append(obj)
                self.objectSourcesByID[obj.id] = obj
            elif isinstance(obj, HoaGroup):
                self.hoaGroups.append(obj)
                self.hoaGroupsByID[obj.id] = obj
            elif isinstance(obj, HoaSource):
                self.sources.append(obj)
                self.sourcesByID[obj.id] = obj

                self.hoaSources.append(obj)
                self.hoaSourcesByID[obj.id] = obj
            elif isinstance(obj, InformedSource):
                self.informedSources.append(obj)
                self.informedSourcesByID[obj.id] = obj
            elif isinstance(obj, ChannelSource):
                self.sources.append(obj)
                self.sourcesByID[obj.id] = obj

                self.channelSources.append(obj)
                self.channelSourcesByID[obj.id] = obj
            elif isinstance(obj, Transform):
                self.transforms.append(obj)
                self.transformsByID[obj.id] = obj
            elif isinstance(obj, Anchor):
                self.anchors.append(obj)
                self.anchorsByID[obj.id] = obj
            elif isinstance(obj, ListenerProximityCondition):
                self.proximityConditions.append(obj)
                self.proximityConditionsByID[obj.id] = obj
            elif isinstance(obj, AudioStream):
                self.streams.append(obj)
                self.streamsByID[obj.id] = obj
            elif isinstance(obj, Update):
                self.updates.append(obj)
                self.updatesByID[obj.id] = obj
            elif isinstance(obj, AcousticTuningParameters):
                self.acousticTuningParams = obj
            elif isinstance(obj, AirflowSource):
                self.airflowSources.append(obj)
                self.airflowSourcesByID[obj.id] = obj
            elif isinstance(obj, AirflowGenerator):
                self.airflowGenerators.append(obj)
                self.airflowGeneratorsByID[obj.id] = obj
            elif isinstance(obj, Portal):
                self.portals.append(obj)
                self.portalsByID[obj.id] = obj
            elif isinstance(obj, HeadphoneEQ):
                if self.headphoneEQ.specType != None:
                    raise SyntaxError("Only one headphone EQ allowed")
                self.headphoneEQ = obj
            else:
                raise ValueError("Unsupported type (%s)" % obj.__class__.__name__)

            self.objects.append(obj)
            if hasattr(obj, "id"):
                self.objectsByID[obj.id] = obj

    def remove(self, *argv):
        """Remove object from eifplus"""
        for obj in argv:
            if obj not in self.objects:
                continue

            if isinstance(obj, Directivity):
                self.directivities.remove(obj)
                self.directivitiesByID.pop(obj.id)
            elif isinstance(obj, AcousticEnvironment):
                self.environments.remove(obj)
                self.environmentsByID.pop(obj.id)
            elif isinstance(obj, Material):
                self.materials.remove(obj)
                self.materialsByID.pop(obj.id)
            elif (
                isinstance(obj, Box)
                or isinstance(obj, Cylinder)
                or isinstance(obj, Sphere)
                or isinstance(obj, Mesh)
            ):
                self.geometry.remove(obj)
                self.geometryByID.pop(obj.id)
            elif isinstance(obj, ObjectSource):
                self.sources.remove(obj)
                self.sourcesByID.pop(obj.id)

                self.objectSources.remove(obj)
                self.objectSourcesByID.pop(obj.id)
            elif isinstance(obj, InformedSource):
                self.informedSources.remove(obj)
                self.sourcesByID.pop(obj.id)
            elif isinstance(obj, HoaSource):
                self.hoaGroups.remove(obj)
                self.hoaGroupsByID.pop(obj.id)
            elif isinstance(obj, HoaSource):
                self.sources.remove(obj)
                self.sourcesByID.pop(obj.id)

                self.hoaSources.remove(obj)
                self.hoaSourcesByID.pop(obj.id)
            elif isinstance(obj, ChannelSource):
                self.sources.remove(obj)
                self.sourcesByID.pop(obj.id)

                self.channelSources.remove(obj)
                self.channelSourcesByID.pop(obj.id)
            elif isinstance(obj, Transform):
                self.transforms.remove(obj)
                self.transformsByID.pop(obj.id)
            elif isinstance(obj, Anchor):
                self.anchors.remove(obj)
                self.anchorsByID.pop(obj.id)
            elif isinstance(obj, AudioStream):
                self.streams.remove(obj)
                self.streamsByID.pop(obj.id)
            elif isinstance(obj, ListenerProximityCondition):
                self.proximityConditions.remove(obj)
                self.proximityConditionsByID.pop(obj.id)
            elif isinstance(obj, Update):
                self.updates.remove(obj)
                self.updatesByID.pop(obj.id)
            elif isinstance(obj, AcousticTuningParameters):
                self.acousticTuningParams = AcousticTuningParameters()
            elif isinstance(obj, AirflowSource):
                self.airflowSources.remove(obj)
                self.airflowSourcesByID.pop(obj.id)
            elif isinstance(obj, AirflowGenerator):
                self.airflowGenerators.remove(obj)
                self.airflowGeneratorsByID.pop(obj.id)
            elif isinstance(obj, Portal):
                self.portals.remove(obj)
                self.portalsByID.pop(obj.id)
            elif isinstance(obj, HeadphoneEQ):
                self.headphoneEQ = HeadphoneEQ()

            self.objects.remove(obj)
            self.objectsByID.pop(obj.id)

    def strip(self):
        """Remove all geometries without a specified material"""
        geoKillList = []
        for n in range(len(self.geometry)):
            obj = self.geometry[n]
            if isinstance(obj, Box) or isinstance(obj, Cylinder) or isinstance(obj, Sphere):
                if obj.material == None:
                    geoKillList.append(n)
            elif isinstance(obj, Mesh):
                num_materials = len(obj.materials)
                if num_materials == 0 or obj.materials == [None] * num_materials:
                    geoKillList.append(n)

        for n in reversed(sorted(geoKillList)):
            self.geometry.pop(n)

    @staticmethod
    def load(path):
        """Loads an audio eifplus from an EIF file

        Parameters:
        path (string): Path the to XML file
        """
        dom = parse(path)

        sceneNodes = dom.getElementsByTagName("AudioScene")
        if not sceneNodes:
            raise SyntaxError("File contains no <AudioScene> node")
        if len(sceneNodes) > 1:
            raise SyntaxError("File contains multiple <AudioScene> node")
        sceneNode = sceneNodes[0]

        if not sceneNode.hasAttribute("id"):
            raise SyntaxError("Missing ID")
        id = sceneNode.getAttribute("id").strip()

        if sceneNode.hasAttribute("duration"):
            duration = float(sceneNode.getAttribute("duration"))
        else:
            duration = 0

        if sceneNode.hasAttribute("type"):
            sceneType = sceneNode.getAttribute("type").strip()
        else:
            sceneType = "VR"

        if sceneNode.hasAttribute("enableLocallyCapturedAudio"):
            enableLocallyCapturedAudio = utils.parseBool(
                sceneNode.getAttribute("enableLocallyCapturedAudio").strip()
            )
        else:
            enableLocallyCapturedAudio = True

        if sceneNode.hasAttribute("overrideListenerVoiceDirectivity"):
            listenerVoiceDirectivityId = sceneNode.getAttribute(
                "overrideListenerVoiceDirectivity"
            ).strip()
            if listenerVoiceDirectivityId == "":
                raise SyntaxError(
                    f"Id for overriding listener voice directivity is missing in EIF file"
                )
            overrideListenerVoiceDirectivity = True
        else:
            overrideListenerVoiceDirectivity = False
            listenerVoiceDirectivityId = ""

        scene = AudioScene(
            id,
            duration,
            sceneType,
            enableLocallyCapturedAudio,
            overrideListenerVoiceDirectivity,
            listenerVoiceDirectivityId,
        )

        deserializers = {
            "AcousticMaterial": Material.fromxml,
            "AudioStream": AudioStream.fromxml,
            "SourceDirectivity": Directivity.fromxml,
            "Box": Box.fromxml,
            "Sphere": Sphere.fromxml,
            "Cylinder": Cylinder.fromxml,
            "Mesh": Mesh.fromxml,
            "ObjectSource": ObjectSource.fromxml,
            "Update": Update.fromxml,
            "AcousticEnvironment": AcousticEnvironment.fromxml,
            "InformedSource": InformedSource.fromxml,
            "HOAGroup": HoaGroup.fromxml,
            "HOASource": HoaSource.fromxml,
            "ChannelSource": ChannelSource.fromxml,
            "Transform": Transform.fromxml,
            "Anchor": Anchor.fromxml,
            "ListenerProximityCondition": ListenerProximityCondition.fromxml,
            "AcousticTuningParameters": AcousticTuningParameters.fromxml,
            "AirflowSource": AirflowSource.fromxml,
            "AirflowGenerator": AirflowGenerator.fromxml,
            "Portal": Portal.fromxml,
            "ContentCreatorHeadphone": HeadphoneEQ.fromxml,
            "GranularSound": GranularSound.fromxml,
        }

        # Parse document and create eifplus entities
        for name, creator in deserializers.items():
            for n in sceneNode.getElementsByTagName(name):
                obj = creator(n)
                scene.add(obj)

                if type(obj) == HoaGroup:
                    obj.validate(scene.objectSourcesByID)

                if type(obj) == HoaSource:
                    obj.validate(scene.hoaGroupsByID)

        # Store voxel-scene IDs, if any
        scene.voxelSceneIDs = [
            scene.getAttribute("id") for scene in dom.getElementsByTagName("VoxelScene")
        ]

        return scene

    def store(self, filename):
        """Stores an audio eifplus to an EIF file"""
        doc = Document()
        root = doc.createElement("AudioScene")
        root.setAttribute("id", self.id)
        if self.duration != 0:
            root.setAttribute("duration", utils.numberToStr(self.duration))
        if self.type != "VR":
            root.setAttribute("type", self.type)

        # --= Materials =--

        for x in self.materials:
            root.appendChild(x.toxml(doc))

        # --= Conditions =--

        for x in self.proximityConditions:
            root.appendChild(x.toxml(doc))

        # --= Streams =--

        for x in self.streams:
            root.appendChild(x.toxml(doc))

        # --= Directivities =--

        for x in self.directivities:
            root.appendChild(x.toxml(doc))

        # --= Geometry =--

        for x in self.geometry:
            root.appendChild(x.toxml(doc))

        # --= Sources =--

        for x in self.sources:
            root.appendChild(x.toxml(doc))

        # --= Environments =--

        for x in self.environments:
            root.appendChild(x.toxml(doc))

        # --= Updates =--

        for x in self.updates:
            root.appendChild(x.toxml(doc))

        # --= Transforms =--

        for x in self.transforms:
            root.appendChild(x.toxml(doc))

        # --= Anchors =--

        for x in self.anchors:
            root.appendChild(x.toxml(doc))

        # --= AcousticTuningParameters =--

        for x in self.acousticTuningParams:
            root.appendChild(x.toxml(doc))

        # --= Portals =--

        for x in self.portals:
            root.appendChild(x.toxml(doc))

        # --= Content Creator Headphone EQ =--
        for x in self.headphoneEQ:
            root.appendChild(x.toxml(doc))

        # TODO: Updates

        # TODO: AirflowSources

        # TODO: AirflowGenerators

        doc.appendChild(root)
        doc.writexml(open(filename, "w"), addindent="\t", newl="\n")

    def __str__(self):
        s = 'AudioScene "%s" {\n\tduration = %f\n' % (self.id, self.duration)
        for x in self.materials:
            s += utils.indent(str(x)) + "\n"
        for x in self.directivities:
            s += utils.indent(str(x)) + "\n"
        for x in self.streams:
            s += utils.indent(str(x)) + "\n"
        for x in self.geometry:
            s += utils.indent(str(x)) + "\n"
        for x in self.sources:
            s += utils.indent(str(x)) + "\n"
        for x in self.environments:
            s += utils.indent(str(x)) + "\n"
        for x in self.updates:
            s += utils.indent(str(x)) + "\n"
        for x in self.portals:
            s += utils.indent(str(x)) + "\n"
        s += "}"
        return s

    def map(self, id, obj):
        """Validate an ID (string) and map it to an object"""
        if len(str(id)) == 0:
            raise SyntaxError("Invalid ID")
        if id in self.objectsByID.keys():
            raise SyntaxError("Duplicate IDs")
        self.objectsByID[id] = obj
