# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np

from . import utils
from . import validation

# from mpegi.geometry import transformMatrixTRS_4x4 TODO: Move to EIF+


class Transform:
    def __init__(
        self,
        id,
        position=[0, 0, 0],
        orientation=[0, 0, 0],
        cspace="relative",
        boxIDs=[],
        sphereIDs=[],
        cylinderIDs=[],
        meshIDs=[],
        objectSourceIDs=[],
        hoaSourceIDs=[],
        channelSourceIDs=[],
    ):
        """Creates a spatial transform"""
        self.id = validation.validateID(id)
        self.position = validation.validateVector3(position, f"position in Transform '{self.id}'")
        self.orientation = validation.validateVector3(
            orientation, f"orientation in Transforms '{self.id}'"
        )
        self.cspace = validation.validateCSpace(cspace)

        self.boxIDs = boxIDs
        self.sphereIDs = sphereIDs
        self.cylinderIDs = cylinderIDs
        self.meshIDs = meshIDs

        self.objectSourceIDs = objectSourceIDs
        self.hoaSourceIDs = hoaSourceIDs
        self.channelSourceIDs = channelSourceIDs

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        if not node.hasAttribute("id"):
            raise SyntaxError("Transform has missing ID in EIF file")
        id = node.getAttribute("id").strip()

        if node.hasAttribute("position"):
            position = node.getAttribute("position").strip()
        else:
            position = [0, 0, 0]

        if node.hasAttribute("orientation"):
            orientation = node.getAttribute("orientation").strip()
        else:
            orientation = [0, 0, 0]

        if node.hasAttribute("cspace"):
            cspace = utils.parseCSpace(node.getAttribute("cspace").strip())
        else:
            cspace = "relative"

        boxIDs = []
        sphereIDs = []
        cylinderIDs = []
        meshIDs = []
        objectSourceIDs = []
        hoaSourceIDs = []
        channelSourceIDs = []

        def _getChildrenIDs(node, childName, IDs):
            if len(node.getElementsByTagName(childName)) > 0:
                for child in node.getElementsByTagName(childName):
                    IDs.append(child.getAttribute("id"))

        _getChildrenIDs(node, "Box", boxIDs)
        _getChildrenIDs(node, "Sphere", sphereIDs)
        _getChildrenIDs(node, "Cylinder", cylinderIDs)
        _getChildrenIDs(node, "Mesh", meshIDs)
        _getChildrenIDs(node, "ObjectSource", objectSourceIDs)
        _getChildrenIDs(node, "HOASource", hoaSourceIDs)
        _getChildrenIDs(node, "ChannelSource", channelSourceIDs)

        return Transform(
            id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            boxIDs=boxIDs,
            sphereIDs=sphereIDs,
            cylinderIDs=cylinderIDs,
            meshIDs=meshIDs,
            objectSourceIDs=objectSourceIDs,
            hoaSourceIDs=hoaSourceIDs,
            channelSourceIDs=channelSourceIDs,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Transform")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        if np.any(self.position != 0):
            node.setAttribute("position", utils.vectorToAttributeStr(self.position))
        if np.any(self.orientation != 0):
            node.setAttribute("orientation", utils.vectorToAttributeStr(self.orientation))
        if self.cspace != "relative":
            node.setAttribute("cspace", self.cspace)

        return node

    def __str__(self):
        return 'Transform "%s" {\n\tposition = %s,\n\torientation = %s,\n\tcspace = %s\n}' % (
            self.id,
            utils.positionToString(self.position),
            utils.orientationToString(self.orientation),
            self.cspace,
        )
