# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from . import utils
from . import validation


class Modification:
    def __init__(self, id, duration=0.0, changes={}):
        """Creates a eifplus modification
        :param id           ID of the entity to be modified
        :param duration     Transition duration [s]
        :param changes      Dictionary attribute-value to set for the entity
        """
        self.id = validation.validateID(id)
        if not isinstance(changes, dict):
            raise ValueError("Changes must be a dictionary with string keys")

        if not all(isinstance(x, str) for x in changes.keys()):
            raise ValueError("Changes must be a dictionary with string keys")
        self.changes = changes

        if duration < 0:
            raise ValueError(
                f"Invalid Duration: Modification '{self.id}' mustn't have "
                f"negative value for duration: '{duration}'"
            )
        self.duration = duration

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        id = None
        duration = 0.0
        changes = {}

        # Note: When attributes id, transition, duration are specified,
        # the first occurence defined the sets the modification's attribute,
        # any further occurence this of the target object.
        for attr in node.attributes.values():
            name = attr.name.lower()
            if name == "id":
                if id is None:
                    id = attr.value
                else:
                    raise ValueError("Modifications may not set ID of an entity")
            elif name == "duration":
                duration = float(attr.value)
            else:
                changes[name] = attr.value.strip()

        return Modification(id, duration, changes)

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Modify")

        # Note: Attribute with default values are skipped for compactness
        node.setAttribute("id", self.id)
        if self.duration != 0:
            node.setAttribute("duration", utils.numberToStr(self.duration))
        for k, v in self.changes.items():
            node.setAttribute(k, v)

        return node

    def __str__(self):
        s = 'Modification { id = "%s" ' % (self.id)
        if self.duration > 0:
            s += ", duration = %f" % self.duration
        for k, v in self.changes.items():
            if isinstance(v, str):
                s += ', %s = "%s"' % (k, v)
            else:
                s += ", %s = %s" % (k, v)
        s += " }"
        return s


class Update:
    """Base class for eifplus updates"""

    def __init__(
        self,
        id,
        modifications=[],
        index=None,
        time=None,
        condition=None,
        fireOn=True,
        delay=0,
    ):
        """Creates an update"""
        self.id = id
        self.index = index
        if time:
            self.time = float(time)
            if self.time < 0:
                raise ValueError("Invalid time")
        else:
            self.time = None

        if condition:
            self.condition = validation.validateID(condition)
        else:
            self.condition = None
        self.fireOn = validation.validateBool(fireOn, "fireOn flag")
        self.delay = float(delay)
        if self.delay < 0:
            raise ValueError("Invalid delay")

        # Modal checks
        if self.time and self.condition:
            ValueError("An update can have either a time or a condition, but not both")

        self.modifications = modifications

        def _checkIfModsHaveDynamicParams(mods):
            for mod in mods:
                for change in mod.changes.values():
                    if "$" in change:
                        return True

            return False

        self.type = None
        if self.time != None:
            self.type = "timed"
        elif self.condition != None:
            self.type = "conditional"
        elif _checkIfModsHaveDynamicParams(self.modifications):
            for mod in self.modifications:
                if mod.duration != 0.0:
                    raise ValueError(
                        "Cannot have both dynamic paramater "
                        "and duration variables in modifications "
                        "of an update."
                    )

            self.type = "dynamic"
        else:
            self.type = "triggered"

    @staticmethod
    def fromxml(node):
        """Deserialize from XML node"""
        # if not node.hasAttribute("id"): raise SyntaxError("Missing ID")
        # TODO: Raise exceptions to check different Update Types
        id = node.getAttribute("id")
        if id == "":
            id = None

        if node.hasAttribute("index"):
            index = int(node.getAttribute("index").strip())
        else:
            index = None

        if node.hasAttribute("time"):
            time = node.getAttribute("time").strip()
        else:
            time = None

        if node.hasAttribute("condition"):
            condition = node.getAttribute("condition").strip()
        else:
            condition = None

        if node.hasAttribute("fireOn"):
            fireOn = validation.validateBool(node.getAttribute("fireOn").strip(), "fireOn flag")
        else:
            fireOn = True

        if node.hasAttribute("delay"):
            delay = node.getAttribute("delay").strip()
        else:
            delay = 0

        modifications = []
        for n in node.getElementsByTagName("Modify"):
            modification = Modification.fromxml(n)
            modifications.append(modification)

        return Update(
            id,
            modifications=modifications,
            index=index,
            time=time,
            condition=condition,
            fireOn=fireOn,
            delay=delay,
        )

    def toxml(self, dom):
        """Serialize to XML node"""
        node = dom.createElement("Update")

        # Note: Attribute with default values are skipped for compactness
        if self.index:
            node.setAttribute("index", utils.numberToStr(self.index))
        if self.time:
            node.setAttribute("time", utils.numberToStr(self.time))
        if self.condition:
            node.setAttribute("condition", self.condition)
        if self.fireOn != True:
            node.setAttribute("fireOn", "false")
        if self.delay:
            node.setAttribute("delay", utils.numberToStr(self.delay))

        for m in self.modifications:
            node.appendChild(m.toxml(dom))

        return node

    def __str__(self):
        s = 'Update "%s" { ' % self.id
        if self.index:
            s += "index = %d, " % self.index
        if self.time:
            s += "time = %s, " % utils.numberToStr(self.time)
        if self.condition:
            s += "condition = %s, " % self.condition
        s += "fireOn = %s, delay = %s\n" % (self.fireOn, utils.numberToStr(self.delay))
        for m in self.modifications:
            s += utils.indent(str(m)) + "\n"
        s += "}\n"
        return s
