# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
#  Validation and type checking functions
#

import numpy as np


def validateID(x):
    """Validate an entity ID"""
    # Remove whitespaces
    y = x.strip()
    if not y:
        raise ValueError('Invalid ID ("%s")' % x)
    return y


def validateFilename(x):
    """Validate a filename
    note:: Does not check if the file exists
    """
    y = x.strip()
    if not y:
        raise ValueError('Invalid filename ("%s")' % x)
    return y


def validateAepInputChannels(x):
    """Validate an aepInputChannels parameter
    note:: Does not check the range
    """
    y = x.strip()
    if not isinstance(x, str) or not y:
        raise ValueError('Invalid aepInputChannels parameter ("%s")' % x)
    return y


def validateLoudness(x):
    """Validate a loudness
    note:: Does not check the range
    """
    y = x.strip()
    if not y:
        raise ValueError('Invalid loudness ("%s")' % x)
    return float(y)


def validateNumChannels(x):
    """Validate numChannels parameter in granular sound
    note:: Does not check the range
    """
    y = x.strip()
    if not isinstance(x, str) or not y:
        raise ValueError('Invalid NumChannels parameter ("%s")' % x)
    return y


def validateVector3(x, valueName):
    """Validate a position vector (3D)"""
    if isinstance(x, str):
        y = np.array([float(s) for s in x.strip().split(" ")])
    elif isinstance(x, list):
        y = np.array(x)
    elif isinstance(x, np.ndarray):
        y = x
    else:
        raise ValueError("Invalid %s (%s)" % (valueName, str(x)))

    if y.ndim != 1 or len(y) != 3 or not np.all(np.isreal(y)):
        raise ValueError("Invalid %s (%s)" % (valueName, str(x)))

    return y


def validateIntVector3(x, valueName):
    """Validate a three-element integer vector (3D)"""
    y = validateVector3(x, valueName)
    yi = np.round(y).astype(np.int_)

    # Check for integers
    e = np.abs(y - yi)
    if np.any(e > 1e-6):
        raise ValueError("Invalid %s (%s)" % (valueName, str(x)))

    return yi


def validateSize(x):
    """Validate a size vector (3D)"""
    y = validateVector3(x, "size")

    # Speciality: Values must be >=0 and only two values may be 0
    if np.any(y < 0) or np.all(y == 0):
        raise ValueError("Invalid size (%s)" % str(x))

    return y


def validateCSpace(x):
    """Validate a coordinate space attribute"""
    y = x.strip().lower()
    if not y in ["relative", "user"]:
        raise ValueError('Invalid coordinate space ("%s")' % str(x))
    return y


def validateNumberList(x, valueName):
    """Validate a list of numbers"""
    if isinstance(x, str):
        y = np.array([float(s) for s in x.strip().split(" ")])
    elif isinstance(x, list):
        y = np.array(x)
    elif isinstance(x, np.ndarray):
        y = x
    else:
        raise ValueError("Invalid %s (%s)" % (valueName, str(x)))

    if y.ndim != 1 or len(y) == 0 or not np.all(np.isreal(y)):
        raise ValueError("Invalid %s (%s)" % (valueName, str(x)))

    return y


def validateIntList(x, valueName):
    """Validate a list of integers"""
    if isinstance(x, str):
        y = np.array([int(s) for s in x.strip().split(" ")])
    elif isinstance(x, list):
        y = [int(z) for z in x]
    else:
        raise ValueError("Invalid %s (%s)" % (valueName, str(x)))

    return y


def validateBool(x, valueName):
    if isinstance(x, bool):
        return x
    elif isinstance(x, str):
        if x.lower() == "true":
            return True
        elif x.lower() == "false":
            return False

    raise ValueError('Invalid %s ("%s")' % (valueName, str(x)))


def validatePlaybackMode(x):
    s = x.strip().lower()
    if s in ["continuous", "event"]:
        return s
    raise SyntaxError('Invalid playback mode ("%s")' % x)


def validateAlignment(x):
    s = x.strip().lower()
    if s in ["user", "object"]:
        return s
    raise SyntaxError('Invalid alignment ("%s")' % x)


def validatePositioning(x):
    string = x.strip().upper()
    for s in string.split():
        s.strip()
        if (
            s != "TL"
            and s != "T"
            and s != "TR"
            and s != "L"
            and s != "C"
            and s != "R"
            and s != "BL"
            and s != "B"
            and s != "BR"
        ):
            raise SyntaxError('Invalid Position ("%s")' % x)

    return string
