# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Optional

from .aparams import AParams
from .location import Location
from .mode import Mode


class AudioElement:
    """Representation of an audio entity, i.e. ObjectSources, HOASources,
    and ChannelSources.
    """

    def __init__(
        self,
        location: Location,
        active: bool,
        gainDb: float,
        refDistance: float,
        signal: str,
        aparams: AParams,
        mode: Mode,
        play: bool,
        vdlMethod: int,
        hasReverbGain: bool,
        reverbGainDb: float,
        noReverb: bool,
    ):
        """Constructor

        :param active: If true, render this source
        :type active: bool
        :param gainDb: Gain in Decibels
        :type gainDb: float
        :param refDistance: The reference distance for distance attentuation
        :type refDistance: float
        :param signal: Unique ID for AudioStream
        :type signal: str
        :param aparams: A list of authoring parameters
        :param mode: Either "continuous" or "event"
        :param play: Is playback enabled
        :type play: bool
        """
        self.location = location
        self.active = active
        self.gainDb = gainDb
        self.refDistance = refDistance
        self.signal = signal
        self.aparams = aparams
        self.mode = mode
        self.play = play
        self.vdlMethod = vdlMethod
        self.hasReverbGain = hasReverbGain
        self.reverbGainDb = reverbGainDb
        self.noReverb = noReverb

    @property
    def position(self):
        return self.location.position

    @position.setter
    def position(self, newPosition):
        self.location.position = newPosition

    @property
    def orientation(self):
        return self.location.orientation

    @orientation.setter
    def orientation(self, newOrientation):
        self.location.orientation = newOrientation

    @property
    def cspace(self):
        return self.location.cspace

    @cspace.setter
    def cspace(self, newCSpace):
        self.location.cspace = newCSpace

    @property
    def transform(self):
        return self.location.transform

    @transform.setter
    def transform(self, newTransform):
        self.location.transform = newTransform

    @property
    def anchor(self):
        return self.location.anchor

    @anchor.setter
    def anchor(self, newAnchor):
        self.location.anchor = newAnchor
