# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Optional

from .basetypes import AudioElement, CSpace, Entity, Location


class Loudspeaker:
    """A Loudspeaker based on an EIF Loudspeaker within a ChannelSource.
    Enhanced with extra metadata for the purposes of encoding to an MPEGI bitstream.
    """

    def __init__(
        self,
        position: List[float],
        orientation: List[float],
        channel: int,
        directivity: str,
        directiveness: float,
    ):
        """Creates a Loudspeaker within a ChannelSource with an EntityFactory.
        Do not construct without entityFactory

        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: list
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: list
        :param channel: Channel Index in the associated AudioStream
        :type channel: int
        :param directivity: Unique identifier to a SourceDirectivity
        :type directivity: str
        :param directiveness: Intensify factor of directiveness
        :type directiveness: float
        """
        self.position = position
        self.orientation = orientation
        self.channel = channel
        self.directivity = directivity
        self.directiveness = directiveness


class ChannelSource(Entity, AudioElement):
    """A ChannelSource based on an EIF ChannelSource that is enhanced with
    extra metadata for the purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        loudspeakers: List[Loudspeaker],
        id: str,
        position: List[float],
        orientation: List[float],
        cspace: CSpace,
        inputLayout: int,
        active: bool,
        gainDb: float,
        refDistance: float,
        signal: str,
        aparams: Optional[str],
        mode: str,
        play: bool,
        transform: str,
        anchor: str,
        vdlMethod: int,
        hasReverbGain: bool,
        reverbGainDb: float,
        noReverb: bool,
        granularPos: List[float],
    ):
        """Creates a ChannelSource with EntityFactor. Do not construct without
        entityFactory.

        :param loudspeakers: A list of Loudspeaker() (not EIF)
        :type loudspeakers: List[Loudspeaker]
        :param id: Unique Indentifier
        :type id: str
        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: List[float]
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: List[float]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :type cspace: eifplus.CSpace
        :param inputLayout: CICP Layout (0 - 127) as specified in
            ISO/IEC 23091-3:2018 CICP-Audio
        :type inputLayout: int
        :param active: A flag to render this source
        :type active: bool
        :param gainDb: The gain of this source expressed in dB
        :type gainDb: float
        :param refDistance: The reference distance for distance attenuation law
        :type refDistance: float
        :param signal: Unique identifier for the relevant AudioStream
        :type signal: str
        :param aparams: Authoring Parameters expressed as a list of strings
        :type aparams: Optional[str]
        :param mode: A flag describing the playback mode for this object
        :type mode: str
        :param play: A flag for whether or not playback is enabled
        :type play: bool
        :param transform: Unique Identifier for parent transform
        :type transform: str
        :param anchor: Unique Indentifier for parent anchor
        :type anchor: str
        :param granularPos: Initial granular position
        :type granularPos: List[float]
        """
        Entity.__init__(self, id)
        AudioElement.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            active,
            gainDb,
            refDistance,
            signal,
            aparams,
            mode,
            play,
            vdlMethod,
            hasReverbGain,
            reverbGainDb,
            noReverb,
        )

        self.loudspeakers = loudspeakers
        self.inputLayout = inputLayout

        self.id  # Inherited by Entity

        # Inherited by AudioElement
        self.position
        self.orientation
        self.cspace
        self.active
        self.gainDb
        self.refDistance
        self.signal
        self.aparams
        self.mode
        self.play
        self.transform
        self.anchor
        self.vdlMethod
        self.hasReverbGain
        self.reverbGainDb
        self.noReverb
        self.granularPos = granularPos
