# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Union

from mpegi.metadata import eif

from .airflowsource import AirflowSource
from .airflowgenerator import AirflowGenerator
from .anchor import Anchor
from .audiostream import AudioStream
from .basetypes import AParams, CSpace, Mode, GeometryType
from .channelsource import ChannelSource, Loudspeaker
from .directivity import Directivity
from .environments import AcousticEnvironment, AcousticParameters, Frequency
from .hoasource import HOASource, HOAGroup
from .headphoneEQ import HeadphoneEQ
from .informedsource import InformedSource
from .material import AcousticMaterial, MaterialFrequency
from .mesh import Mesh
from .objectsource import ObjectSource, InputLayout, Alignment, Positioning
from .portal import Portal
from .primitives import Box, Sphere, Cylinder
from .transform import Transform
from .update import Update, Modification
from .userregion import UserRegion
from .acoustTuningParameters import AcousticTuningParameters
from .granularsound import GranularSound


class EntityFactory:
    """A factory that creates entities and holds them for a base eifplus
    from EIF objects. It keeps track and holds the indices/identifiers of the
    entities and can return entities/indices/identifiers from a given
    entity/index/identifier.
    """

    @staticmethod
    def manufactureEntity(eifEntity):
        """Manufactures an entity based on the incoming eifEntity and returns it

        :param eifEntity: A fully instanciated EIF entity object
        :type eifEntity: eif.entity
        :return: A Base Scene entity object
        :rtype: eifplus.entity
        """
        if (
            isinstance(eifEntity, (eif.Box, eif.Cylinder, eif.Sphere))
            and "userregion" in eifEntity.id.lower()
        ):
            return EntityFactory._createUserRegion(eifEntity)

        entitySwitch = {
            eif.AcousticEnvironment: EntityFactory._createEnvironment,
            eif.Anchor: EntityFactory._createAnchor,
            eif.AudioStream: EntityFactory._createAudioStream,
            eif.Box: EntityFactory._createPrimitive,
            eif.ChannelSource: EntityFactory._createChannelSource,
            eif.Cylinder: EntityFactory._createPrimitive,
            eif.Directivity: EntityFactory._createDirectivity,
            eif.HoaGroup: EntityFactory._createHoaGroup,
            eif.HoaSource: EntityFactory._createHoaSource,
            eif.Material: EntityFactory._createMaterial,
            eif.Mesh: EntityFactory._createMesh,
            eif.ObjectSource: EntityFactory._createObjectSource,
            eif.Sphere: EntityFactory._createPrimitive,
            eif.Transform: EntityFactory._createTransform,
            eif.Update: EntityFactory._createUpdate,
            eif.Portal: EntityFactory._createPortal,
            eif.AcousticTuningParameters: EntityFactory._createAcousticTuningParams,
            eif.AirflowSource: EntityFactory._createAirflowSource,
            eif.AirflowGenerator: EntityFactory._createAirflowGenerator,
            eif.HeadphoneEQ: EntityFactory._createHeadphoneEQ,
            eif.GranularSound: EntityFactory._createGranularSound,
        }

        return entitySwitch[type(eifEntity)](eifEntity)

    @staticmethod
    def _createEnvironment(eifEnv: eif.AcousticEnvironment) -> AcousticEnvironment:
        id = eifEnv.id
        region = eifEnv.region

        params = []
        for eifParam in eifEnv.parameters:
            if eifParam.position is not None:
                position = [float(p) for p in eifParam.position]
            else:
                position = []
            predelay = eifParam.predelay
            if predelay is not None:
                predelay = float(eifParam.predelay)

            complexityLevel = eifParam.complexityLevel
            if complexityLevel is not None:
                complexityLevel = int(eifParam.complexityLevel)

            frequencies = []
            for i in range(len(eifParam.freqs)):
                newFreq = Frequency(eifParam.freqs[i], eifParam.rt60[i], eifParam.dsr[i])
                frequencies.append(newFreq)

            param = AcousticParameters(
                position=position,
                predelay=predelay,
                complexityLevel=complexityLevel,
                frequencies=frequencies,
            )
            params.append(param)

        return AcousticEnvironment(id=id, region=region, acousticParameters=params)

    @staticmethod
    def _createAnchor(eifAnchor: eif.Anchor) -> Anchor:
        """Creates an anchor and returns it as well as adding it to the
        entity map and setting the index within it. Private function - do not
        access from outside this class

        :param eifAnchor: The EIF Anchor to instanciate from
        :return: New BaseScene Anchor
        """
        id = eifAnchor.id
        lsdfRef = eifAnchor.lsdfRef
        autoRescale = eifAnchor.autoRescale
        audioSourceRescale = eifAnchor.audioSourceRescale
        tolerance = eifAnchor.tolerance

        return Anchor(
            id=id,
            lsdfRef=lsdfRef,
            autoRescale=autoRescale,
            audioSourceRescale=audioSourceRescale,
            tolerance=tolerance,
        )

    def _createAcousticTuningParams(
        eifAcoustTuningParams: eif.AcousticTuningParameters,
    ) -> AcousticTuningParameters:
        """Creates an AcousticTuningParamers and returns it as well as adding it to the
        entity map and setting the index within it. Private function - do not
        access from outside this class

        :param eifAcoutsicTuningParam: The EIF Acoustic Tuning Parameters to instanciate from
        :return: New BaseScene AcousticTuningParameters
        """
        defaultAE = eifAcoustTuningParams.defaultAE
        erGainDb = eifAcoustTuningParams.erGainDb
        voxDiffrCtrl = eifAcoustTuningParams.voxDiffrCtrl
        voxOccCtrl = eifAcoustTuningParams.voxOccCtrl
        voxEarlyCtrl = eifAcoustTuningParams.voxEarlyCtrl
        revRendererControl = eifAcoustTuningParams.revRendererControl
        lcErOutdoorTuning = eifAcoustTuningParams.lcErOutdoorTuning
        return AcousticTuningParameters(
            defaultAE=defaultAE,
            revRendererControl=revRendererControl,
            voxOccCtrl=voxOccCtrl,
            voxDiffrCtrl=voxDiffrCtrl,
            voxEarlyCtrl=voxEarlyCtrl,
            erGainDb=erGainDb,
            lcErOutdoorTuning=lcErOutdoorTuning,
        )

    @staticmethod
    def _createAirflowSource(eifAirflowSource: eif.AirflowSource) -> AirflowSource:
        return AirflowSource(
            id=eifAirflowSource.id,
            airflowGenerator=eifAirflowSource.airflowGenerator,
            defaultAirflowProfile=eifAirflowSource.defaultAirflowProfile,
            activeOutsideAEs=eifAirflowSource.activeOutsideAEs,
            ActiveEnvironment=eifAirflowSource.ActiveEnvironment,
            AirflowProfile=eifAirflowSource.AirflowProfiles,
        )

    @staticmethod
    def _createAirflowGenerator(eifAirflowGenerator: eif.AirflowGenerator) -> AirflowGenerator:
        return AirflowGenerator(
            id=eifAirflowGenerator.id,
            generatorType=eifAirflowGenerator.generatorType,
            orientation=eifAirflowGenerator.orientation,
            speed=eifAirflowGenerator.speed,
            position=eifAirflowGenerator.position,
            gainDb=eifAirflowGenerator.gainDb,
            active=eifAirflowGenerator.active,
            azimuthLimit=eifAirflowGenerator.azimuthLimit,
            elevationLimit=eifAirflowGenerator.elevationLimit,
            height=eifAirflowGenerator.height,
            width=eifAirflowGenerator.width,
            baseRadius=eifAirflowGenerator.baseRadius,
            topRadius=eifAirflowGenerator.topRadius,
            length=eifAirflowGenerator.length,
            distanceAttenuationFactor=eifAirflowGenerator.distanceAttenuationFactor,
            edgeRolloff=eifAirflowGenerator.edgeRolloff,
            transformID=eifAirflowGenerator.transform,
            anchorID=eifAirflowGenerator.anchor,
        )

    @staticmethod
    def _createAudioStream(eifStream: eif.AudioStream) -> AudioStream:
        """Creates an audioStream and returns it as well as adding it to the
        entity map and setting the index within it. Private function - do not
        access from outside this class

        :param eifStream: The EIF AudioStream to instanciate from
        :return: New Basescene AudioScene
        """
        id = eifStream.id
        isLocallyCaptured = eifStream.isLocallyCaptured

        if not isLocallyCaptured:
            file = eifStream.file

            channels = []
            if eifStream.aepInputChannels.find(",") != -1:
                channels = eifStream.aepInputChannels.split(",")
                channels = [int(i) for i in channels]
            elif eifStream.aepInputChannels.find(":") != -1:
                channelRange = eifStream.aepInputChannels.split(":")
                channelRange = [int(i) for i in channelRange]

                rangeLength = channelRange[1] - channelRange[0]

                for i in range(rangeLength + 1):
                    channels.append(channelRange[0] + i)
            else:
                channels.append(int(eifStream.aepInputChannels))
        else:
            file = ""
            channels = []

        return AudioStream(
            id=id, isLocallyCaptured=isLocallyCaptured, file=file, aepInputChannels=channels
        )

    @staticmethod
    def _createPrimitive(
        eifPrimitive: Union[eif.Box, eif.Sphere, eif.Cylinder]
    ) -> Union[Box, Sphere, Cylinder]:
        """Creats a primitive and returns it as well as adding it to the entity
        map and setting the indices within it. Private function - do not access
        from outside this class

        :param eifPrimitive: The EIF Primitive to instanciate from
        :raises AssertionError: Indices could not be found with the given
            identifiers. Make sure that the entity that the identifier references
            is added before this entity.
        :return: New BaseScene Primitive
        """
        id = eifPrimitive.id
        position = eifPrimitive.position
        orientation = eifPrimitive.orientation
        size = eifPrimitive.size
        cspace = CSpace.fromString(eifPrimitive.cspace)
        material = eifPrimitive.material
        transform = eifPrimitive.transform
        anchor = eifPrimitive.anchor

        if type(eifPrimitive) == eif.Box:
            primitive = Box(
                id=id,
                position=position,
                orientation=orientation,
                size=size,
                cspace=cspace,
                material=material,
                transform=transform,
                anchor=anchor,
            )
        elif type(eifPrimitive) == eif.Sphere:
            primitive = Sphere(
                id=id,
                position=position,
                orientation=orientation,
                size=size,
                cspace=cspace,
                material=material,
                transform=transform,
                anchor=anchor,
            )
        elif type(eifPrimitive) == eif.Cylinder:
            primitive = Cylinder(
                id=id,
                position=position,
                orientation=orientation,
                size=size,
                cspace=cspace,
                material=material,
                transform=transform,
                anchor=anchor,
            )

        return primitive

    @staticmethod
    def _createChannelSource(eifChanSrc: eif.ChannelSource) -> ChannelSource:
        """Creates a ChannelSource and returns it as well as adding it to the
        entity map and setting the indices within it. Private function - do
        not access from outside this class

        :param eifChanSrc: The EIF ChannelSource to instanciate from
        :raises AssertionError: Indices could not be found with the given
            identifiers. Make sure that the entity that the identifier references
            is added before this entity.
        :return: New ChannelSource
        """
        loudspeakers = []
        for loudspeaker in eifChanSrc.loudspeakers:
            position = [float(p) for p in loudspeaker.position]
            orientation = [float(p) for p in loudspeaker.orientation]
            channel = loudspeaker.channel
            directivity = loudspeaker.directivity
            directiveness = float(loudspeaker.directiveness)

            ldspkr = Loudspeaker(
                position=position,
                orientation=orientation,
                channel=channel,
                directivity=directivity,
                directiveness=directiveness,
            )

            loudspeakers.append(ldspkr)

        id = eifChanSrc.id
        position = [float(p) for p in eifChanSrc.position]
        orientation = [float(p) for p in eifChanSrc.orientation]
        cspace = CSpace.fromString(eifChanSrc.cspace)
        inputLayout = eifChanSrc.inputLayout
        active = eifChanSrc.active
        gainDb = float(eifChanSrc.gainDb)
        refDistance = float(eifChanSrc.refDistance)
        signal = eifChanSrc.signal
        aparams = AParams.fromString(eifChanSrc.aparams)
        mode = Mode.fromString(eifChanSrc.mode)
        play = eifChanSrc.play
        transform = eifChanSrc.transform
        anchor = eifChanSrc.anchor
        vdlMethod = eifChanSrc.vdlMethod
        hasReverbGain = eifChanSrc.hasReverbGain
        reverbGainDb = float(eifChanSrc.reverbGainDb)
        noReverb = eifChanSrc.noReverb
        granularPos = eifChanSrc.granularPos

        channelSource = ChannelSource(
            loudspeakers=loudspeakers,
            id=id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            inputLayout=inputLayout,
            active=active,
            gainDb=gainDb,
            refDistance=refDistance,
            signal=signal,
            aparams=aparams,
            mode=mode,
            play=play,
            transform=transform,
            anchor=anchor,
            vdlMethod=vdlMethod,
            hasReverbGain=hasReverbGain,
            reverbGainDb=reverbGainDb,
            noReverb=noReverb,
            granularPos=granularPos,
        )

        return channelSource

    @staticmethod
    def _createDirectivity(eifDirec: eif.Directivity) -> Directivity:
        """Creates a Directivity and returns it as well as adding it to the
        entity map and setting the indices within it. Private function - do
        not access from outside this class.

        :param eifDirec: The EIF Direcitivity to instanciate from
        :return: New BaseScene Directivity
        """
        id = eifDirec.id
        file = eifDirec.file

        return Directivity(id=id, file=file)

    @staticmethod
    def _createGranularSound(eifGrain: eif.GranularSound) -> GranularSound:
        id = eifGrain.id
        file = eifGrain.file
        numChannels = eifGrain.numChannels

        return GranularSound(id=id, file=file, numChannels=int(numChannels))

    @staticmethod
    def _createInformedSource(eifInformedSource: eif.InformedSource) -> InformedSource:
        """Creates an InformedSource and returns it as well as adding it to the entity
        map and setting the indices within it. Private function - do not access
        from outside the class.

        :param eifInformedSource: The EIF InformedSource to instanciate from
        :return: New BaseScene InformedSource
        """
        id = eifInformedSource.id
        position = eifInformedSource.position
        listenerThreshold = eifInformedSource.listenerThreshold
        hoaSourceThreshold = eifInformedSource.hoaSourceThreshold
        priorityValue = eifInformedSource.priorityValue

        informedSource = InformedSource(
            id=id,
            position=position,
            listenerThreshold=listenerThreshold,
            hoaSourceThreshold=hoaSourceThreshold,
            priorityValue=priorityValue,
        )

        return informedSource

    @staticmethod
    def _createHeadphoneEQ(eifHpEq: eif.HeadphoneEQ) -> HeadphoneEQ:
        """Creates a HeadphoneEQ and returns it as well as adding it to the entity
        map, and setting internal values. Private function - do not access
        from outside the class.

        :param eifHeadphoneEQ: The EIF HeadphoneEQ to instanciate from
        :return: New BaseScene HeadphoneEQ
        """

        return HeadphoneEQ(
            specType=eifHpEq.specType,
            f=eifHpEq.f,
            gain=eifHpEq.gain,
            q=eifHpEq.q,
            filtType=eifHpEq.filtType,
            preset=eifHpEq.preset,
        )

    @staticmethod
    def _createHoaGroup(eifGroup: eif.HoaGroup) -> HOAGroup:
        """Creates a HOAGroup and returns it as well as adding it to the entity
        map and setting the indices within it. Private function - do not access
        from outside the class.

        :param eifGroup: The EIF HOAGroup to instanciate from
        :raises AssertionError: Indices could not be found with the given
            identifiers. Make sure that the entity that the identifier references
            is added before this entity.
        :return: New BaseScene HOAGroup
        """
        id = eifGroup.id
        region = eifGroup.region
        coSources = eifGroup.coSources
        informedSources = []
        exteriorProjectionRadius = eifGroup.exteriorProjectionRadius
        adaptiveExteriorProjectionRadius = eifGroup.adaptiveExteriorProjectionRadius
        for src in eifGroup.informedSources:
            informedSources.append(EntityFactory._createInformedSource(src))

        return HOAGroup(
            id=id,
            region=region,
            coSources=coSources,
            informedSources=informedSources,
            exteriorProjectionRadius=exteriorProjectionRadius,
            adaptiveExteriorProjectionRadius=adaptiveExteriorProjectionRadius,
        )

    @staticmethod
    def _createHoaSource(eifHoa: eif.HoaSource) -> HOASource:
        """Creates a HOASource and returns it as well as adding it to the
        entity map and setting the indices within it. Private function - do
        not access from outside the class.

        :param eifHoa: The EIF HOASource to instanciate from
        :return: New BaseScene HOASource
        """
        id = eifHoa.id
        position = eifHoa.position
        orientation = eifHoa.orientation
        cspace = CSpace.fromString(eifHoa.cspace)
        active = eifHoa.active
        gainDb = eifHoa.gainDb
        refDistance = eifHoa.refDistance
        signal = eifHoa.signal
        extent = eifHoa.extent
        extentTransform = eifHoa.extentTransform
        representation = eifHoa.representation
        transitionDistance = eifHoa.transitionDistance
        group = eifHoa.group
        aparams = AParams.fromString(eifHoa.aparams)
        mode = Mode.fromString(eifHoa.mode)
        play = eifHoa.play
        transform = eifHoa.transform
        anchor = eifHoa.anchor
        vdlMethod = eifHoa.vdlMethod
        hasReverbGain = eifHoa.hasReverbGain
        reverbGainDb = eifHoa.reverbGainDb
        noReverb = eifHoa.noReverb
        granularPos = (eifHoa.granularPos,)

        hoaSource = HOASource(
            id=id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            active=active,
            gainDb=gainDb,
            refDistance=refDistance,
            signal=signal,
            extent=extent,
            extentTransform=extentTransform,
            representation=representation,
            transitionDistance=transitionDistance,
            group=group,
            aparams=aparams,
            mode=mode,
            play=play,
            transform=transform,
            anchor=anchor,
            vdlMethod=vdlMethod,
            hasReverbGain=hasReverbGain,
            reverbGainDb=reverbGainDb,
            noReverb=noReverb,
            granularPos=granularPos,
        )

        return hoaSource

    @staticmethod
    def _createMaterial(eifMaterial: eif.Material) -> AcousticMaterial:
        """Creates an AcousticMaterial and returns it as well as adding it to
        the entity map and setting the indices within it.
        Private function - do not access from outside the class.

        :param eifMaterial: The EIF AcousticMaterial to instanciate from
        :return: New BaseScene AcousticMaterial
        """

        id = eifMaterial.id

        frequencies = []
        for i in range(len(eifMaterial.f)):
            freq = MaterialFrequency(
                float(eifMaterial.f[i]),
                float(eifMaterial.r[i]),
                float(eifMaterial.s[i]),
                float(eifMaterial.t[i]),
                float(eifMaterial.c[i]),
            )
            frequencies.append(freq)

        return AcousticMaterial(id=id, frequencies=frequencies)

    @staticmethod
    def _createMesh(eifMesh) -> Mesh:
        id = eifMesh.id
        position = [float(p) for p in eifMesh.position]
        orientation = [float(o) for o in eifMesh.orientation]
        cspace = CSpace.fromString(eifMesh.cspace)

        vertices = []
        for vertex in eifMesh.vertices:
            vertices.append([float(v) for v in vertex])

        faces = []
        for face in eifMesh.faces:
            faces.append([int(f) for f in face])

        materials = eifMesh.materials
        transform = eifMesh.transform
        anchor = eifMesh.anchor

        mesh = Mesh(
            id=id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            vertices=vertices,
            faces=faces,
            materials=materials,
            transform=transform,
            anchor=anchor,
        )
        return mesh

    @staticmethod
    def _createPortal(eifSource: eif.Portal) -> Portal:
        id = eifSource.id
        extent = eifSource.portalExtent
        normal = eifSource.portalNormal
        ae1 = eifSource.acousticEnvironment1
        ae2 = eifSource.acousticEnvironment2
        mat = eifSource.material
        portalPosition = eifSource.portalPosition
        portalSize = eifSource.portalSize
        hasExplicitPortal = eifSource.hasExplicitPortal
        return Portal(
            id=id,
            portalExtent=extent,
            portalNormal=normal,
            acousticEnvironment1=ae1,
            acousticEnvironment2=ae2,
            material=mat,
            portalPosition=portalPosition,
            portalSize=portalSize,
            hasExplicitPortal=hasExplicitPortal,
        )

    @staticmethod
    def _createObjectSource(eifSource: eif.ObjectSource) -> ObjectSource:
        """Creates an ObjectSource and returns it as well as adding it to the
        entity map and setting the indices within it. Private function - do not
        access from outside the class

        :param eifSource: The EIF ObjectSource to instanciate from
        :return: New Basescene ObjectSource
        """
        id = eifSource.id
        position = [float(p) for p in eifSource.position]
        orientation = [float(o) for o in eifSource.orientation]
        cspace = CSpace.fromString(eifSource.cspace)
        active = eifSource.active
        gainDb = float(eifSource.gainDb)
        refDistance = float(eifSource.refDistance)
        signal = eifSource.signal
        extent = eifSource.extent
        directivity = eifSource.directivity
        directiveness = float(eifSource.directiveness)
        mode = Mode.fromString(eifSource.mode)
        play = eifSource.play

        inputLayout = None
        if eifSource.inputLayout != None:
            inputLayout = InputLayout(
                Alignment.fromString(eifSource.inputLayout.alignment),
                Positioning.fromString(eifSource.inputLayout.positioning),
            )

        aparams = AParams.fromString(eifSource.aparams)
        transformID = eifSource.transform
        anchorID = eifSource.anchor
        vdlMethod = eifSource.vdlMethod
        hasReverbGain = eifSource.hasReverbGain
        reverbGainDb = float(eifSource.reverbGainDb)
        noReverb = eifSource.noReverb
        granularPos = eifSource.granularPos

        objSource = ObjectSource(
            id=id,
            position=position,
            orientation=orientation,
            cspace=cspace,
            active=active,
            gainDb=gainDb,
            refDistance=refDistance,
            signal=signal,
            extent=extent,
            directivity=directivity,
            directiveness=directiveness,
            mode=mode,
            play=play,
            inputLayout=inputLayout,
            aparams=aparams,
            transform=transformID,
            anchor=anchorID,
            vdlMethod=vdlMethod,
            hasReverbGain=hasReverbGain,
            reverbGainDb=reverbGainDb,
            noReverb=noReverb,
            granularPos=granularPos,
        )
        return objSource

    @staticmethod
    def _createTransform(eifTransform: eif.Transform) -> Transform:
        """Creates a Transform and returns it as well as adding it to the
        entity map and setting the indices within it. Private function - do not
        access from outside the class

        :param eifTransform: The EIF Transform to instanciate from
        :return: New BaseScene Transform
        """
        id = eifTransform.id
        position = eifTransform.position
        orientation = eifTransform.orientation
        cspace = CSpace.fromString(eifTransform.cspace)

        transform = Transform(id=id, position=position, orientation=orientation, cspace=cspace)

        return transform

    @staticmethod
    def _createUpdate(eifUpdate: eif.Update) -> Update:
        """Creates an Update and returns it as well as adding it to the entity
        map and setting the indices within it. Private function - do not access
        from outside the class

        :param eifUpdate: The EIF Update to instanciate from
        :return: New BaseScene Update
        """
        id = eifUpdate.id
        index = eifUpdate.index
        time = eifUpdate.time
        condition = eifUpdate.condition
        fireOn = eifUpdate.fireOn
        delay = eifUpdate.delay
        type = eifUpdate.type

        modifications = []
        for eifMod in eifUpdate.modifications:
            mod = Modification(eifMod.id, eifMod.duration, eifMod.changes)
            modifications.append(mod)

        update = Update(
            id=id,
            index=index,
            time=time,
            condition=condition,
            fireOn=fireOn,
            delay=delay,
            type=type,
            modifications=modifications,
        )

        return update

    @staticmethod
    def _createUserRegion(eifPrimitive) -> UserRegion:
        geoType = {
            eif.Box: GeometryType.Box,
            eif.Sphere: GeometryType.Sphere,
            eif.Cylinder: GeometryType.Cylinder,
        }

        return UserRegion(
            id=eifPrimitive.id,
            position=eifPrimitive.position,
            orientation=eifPrimitive.orientation,
            size=eifPrimitive.size,
            geoType=geoType[type(eifPrimitive)],
        )
