# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Optional

from .basetypes import Entity


class Frequency:
    """An EIF Frequency found in AcousticParameters enhanced with extra
    metadata for the purposes of encoding to an MPEGI bitstream.
    """

    def __init__(self, frequency: float, rt60: float, dsr: Optional[float]):
        """Creates a Frequency..

        :param frequency: Frequency in Hertz
        :type frequency: float
        :param rt60: The time it takes for reverb to decay to 60dB in seconds
        :type rt60: float
        :param dsr: The diffuse to source ratio
        :type dsr: Optional[float]
        """
        self.frequency = frequency
        self.rt60 = rt60
        self.dsr = dsr


class AcousticParameters:
    """EIF AcousticParameters that are enhanced with extra metadata for the
    purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        position: List[float],
        predelay: Optional[float],
        complexityLevel: Optional[int],
        frequencies: List[Frequency],
    ):
        """Creates AcousticParameters. Typically fromEIFParam() is used to
        instanciate this class instead.

        :param position: Spatial position described as a list of floats ([X, Y, Z])
        :type position: List[float]
        :param predelay: The predelay time in seconds
        :type predelay: Optional[float]
        :param complexityLevel: Computational complexity of the reverb (1, 2, ..., 5)
        :type complexityLevel: Optional[int]
        :param frequencies: A list of Frequencies
        :type frequencies: List[Frequency]

        """
        self.frequencies = frequencies

        self.position = position
        self.predelay = predelay
        self.complexityLevel = complexityLevel


class AcousticEnvironment(Entity):
    """An EIF AcousticEnvironment that is enhanced with extra metadata for the
    purpposes of encoding to an MPEGI bitstream
    """

    def __init__(self, id: str, region: str, acousticParameters: List[AcousticParameters]):
        """Creates an AcousticEnvironment. Typically fromEIFEnvironment is used
        to instanciate this class instead

        :param id: Unique identifier
        :param region: Unique identifier for the geometry to describe this spatial region
        :param acousticParameters: A list of AcousticParameters()
        """
        Entity.__init__(self, id)

        self.parameters = acousticParameters

        self.id  # Inherited by Entity

        self.region = region
