# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Optional

from .basetypes import AudioElement, CSpace, Entity, Location

from .informedsource import InformedSource


class HOASource(Entity, AudioElement):
    """An EIF HOASource that is enhanced with extra metadata for the purposes of
    encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: str,
        position: List[float],
        orientation: List[float],
        cspace: CSpace,
        active: bool,
        gainDb: float,
        refDistance: float,
        signal: str,
        extent: Optional[str],
        extentTransform: bool,
        representation: bool,
        transitionDistance: float,
        group: Optional[str],
        aparams: Optional[str],
        mode: str,
        play: bool,
        transform: str,
        anchor: str,
        vdlMethod: int,
        hasReverbGain: bool,
        reverbGainDb: float,
        noReverb: bool,
        granularPos: List[float],
    ):
        """Creates a HOASource. Typically, fromEIFHoaSource is used to instanciate
        this class instead.

        :param id: Unique identifier
        :type id: str
        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: List[float]
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: List[float]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :type cspace: eifplus.CSpace
        :param active: A flag to render this object
        :type active: bool
        :param gainDb: The gain of this source expressed in dB
        :type gainDb: float
        :param refDistance: The reference distance for distance attenuation law
        :type refDistance: float
        :param signal: Unique identifier for the relevant AudioStream
        :type signal: str
        :param extent: Unique identifier for the relevant Geometry for the
            spatial extent of this source
        :type extent: Union[str, None]
        :param extentTransform: A flag to turn on/off external/internal
            rendering of an interior/exterior source
        :type extentTransform: bool
        :param representation: Interior or Exterior HOA expansion
        :type representation: int
        :param transitionDistance: Transition area between internal and
            external representation
        :type transitionDistance: float
        :param group: Unique identifier for parent HOAGroup
        :type group: str
        :param aparams: Authoring parameters expressed as a list of strings
        :type aparams: Optional[str]
        :param mode: A flag describing the playback mode for this object
        :type mode: str
        :param play: A flag for whether or not playback is enabled
        :type play: bool
        :param transform: Unique identifier for parent transform
        :type transform: str
        :param anchor: Unique identifier for parent anchor
        :type anchor: str
        :param granularPos: Initial granular position
        :type granularPos: List[float]
        """
        Entity.__init__(self, id)
        AudioElement.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            active,
            gainDb,
            refDistance,
            signal,
            aparams,
            mode,
            play,
            vdlMethod,
            hasReverbGain,
            reverbGainDb,
            noReverb,
        )

        self.id  # Inherited by Entity

        # Inherited by AudioElement
        self.position
        self.orientation
        self.cspace
        self.active
        self.gainDb
        self.refDistance
        self.signal
        self.aparams
        self.mode
        self.play
        self.transform
        self.anchor
        self.vdlMethod
        self.hasReverbGain
        self.reverbGainDb
        self.noReverb

        self.extent = extent
        self.extentTransform = extentTransform
        self.representation = representation
        self.transitionDistance = transitionDistance
        self.group = group
        self.granularPos = granularPos


class HOAGroup(Entity):
    def __init__(
        self,
        id: str,
        region: str,
        coSources: List[str],
        informedSources: List[InformedSource],
        exteriorProjectionRadius: float,
        adaptiveExteriorProjectionRadius: bool,
    ):
        """A HOAGroup based on an EIF HOAGroup that is enhanced with extra
        metadata for the purposes of encoding to an MPEGI bitstream.

        :param id: Unique Identifier
        :type id: str
        :param region: Unique Identifier for the relevant geometry
        :type region: str
        :param coSources: A list of ObjectSources within this HOAGroup
        :type coSources: list
        :param informedSources: A list of known audio sources within this HOAGroup
        :type informedSources: list
        :param exteriorProjectionRadius: Exterior rendering projection radius
        :type exteriorProjectionRadius: float
        :param adaptiveExteriorProjectionRadius: A flag for adaptive exterior rendering
            projection radius (used when informed sources are present)
        :type adaptiveExteriorProjectionRadius: bool
        """
        Entity.__init__(self, id)

        self.id  # Inherited by Entity

        self.region = region
        self.coSources = coSources
        self.informedSources = informedSources
        self.exteriorProjectionRadius = exteriorProjectionRadius
        self.adaptiveExteriorProjectionRadius = adaptiveExteriorProjectionRadius
