# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

import numpy as np

from .basetypes import CSpace, Entity, Geometry, GeometryType, Location


class Mesh(Entity, Geometry):
    """A Mesh based on an EIF Mesh that is enhanced with extra metadata for the
    purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: str,
        position: List[float],
        orientation: List[float],
        cspace: CSpace,
        vertices: List[List[float]],
        faces: List[List[int]],
        materials: List[str],
        transform: str,
        anchor: str,
    ):
        """Creates a Mesh with EntityFactory. Do not construct without EntityFactory.

        :param id: Unique Identifier
        :param position: Spatial position in cartesian coordinates expressed as a list of three
            floats [X, Y, Z]
        :param orientation: Spatial orientation in cartesian coordinates expressed as a list of
            three floats [Yaw, Pitch, Roll]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :param vertices: Each element in the list is also a list which contains cartesian
            coordinates for the vertex
        :param faces: Each element in the list is also a list which contains three indieces to
            vertices which comprises the face
        :param materials: A list of string identifiers for AcousticMaterials
        :param transform: the ID of the parent transform of this mesh
        :param anchor: the ID of the parent anchor of this mesh
        """
        Entity.__init__(self, id)
        Geometry.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            GeometryType.Mesh,
        )

        self.id  # Inherited by Entity

        # Inherited by Geometry
        self.position
        self.orientation
        self.cspace
        self.anchor
        self.transform

        self.vertices = vertices
        self.faces = faces
        self.materials = materials

    @property
    def isTransparent(self) -> bool:
        """Returns a bool based on whether or not the mesh is Acoustically
        "Transparent", i.e. acoustically irrelevant; it has no materials

        :return: Returns true if the mesh is acoustically transparent
        :rtype: bool
        """
        return all([mat == None for mat in self.materials])

    def getFaceNormal(self, faceIdx):
        """Returns the normal vector of a face according to OpenGL/EIF convention
        n = (v2 - v1) x (v3 - v1)
        :param faceIdx  the face index as returned by addFace
        :returns        numpy vector of length 3 with normal in global coordinates
        """
        v1 = np.array(self.vertices[self.faces[faceIdx][0]])
        v2 = np.array(self.vertices[self.faces[faceIdx][1]])
        v3 = np.array(self.vertices[self.faces[faceIdx][2]])
        n = np.cross((v2 - v1), (v3 - v1))
        mag_n = np.linalg.norm(n)
        return n / mag_n if mag_n != 0 else [0, 0, 0]

    def getFaceCentroid(self, faceIdx):
        """Returns the centroid of a face
        :param faceIdx  the face index as returned by addFace
        :returns        numpy vector of length 3 with centroid in global coordinates
        """
        v1 = np.array(self.vertices[self.faces[faceIdx][0]])
        v2 = np.array(self.vertices[self.faces[faceIdx][1]])
        v3 = np.array(self.vertices[self.faces[faceIdx][2]])
        return (v1 + v2 + v3) / 3
