# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

from ..basetypes import AudioElement, Geometry, Mode
from ..mesh import Mesh
from ..primitives import Primitive
from ..update import Update
from ..util import getAllChanges
from ...eifplus.audiostream import AudioStream

from xml.etree.ElementTree import Element
import soundfile as sf

"""These functions analyse a scene and give information on whether or not
some information is redundant, for example if an audio source would ever
be rendered in a scene.
"""


def isMaterialUsed(materialId: str, geometry: List[Geometry], voxelBlocks: List[Element]) -> bool:
    """Returns true if another geometry refers to an acoustic material

    :param materialId: The string identifier of the material
    :type materialId: str
    :param geometry: A list of geometry to check
    :param voxelBlocks: A list of voxelblocks to check
    :return: returns false if the material is never used by geometry or voxel Block
    :rtype: bool
    """
    allMaterialsInUse = []
    for geo in geometry:
        if isinstance(geo, Mesh):
            allMaterialsInUse[-1:-1] = geo.materials
        elif isinstance(geo, Primitive):
            allMaterialsInUse.append(geo.material)
        else:
            raise Exception("Invalid Geometry")  # should never be reached

    for block in voxelBlocks:
        if block.attrib["material"] not in allMaterialsInUse:
            allMaterialsInUse.append(block.attrib["material"])

    return materialId in allMaterialsInUse


def isStreamUsed(streamId: str, sources: List[AudioElement], updates: List[Update]) -> bool:
    """Returns true if an audio source refers to this stream or if an update
    changes an audio source to use this audio stream.

    :param streamId: The string identifier of the audio stream
    :type streamId: str
    :param sources: A list of all audio elements in the scene
    :param updates: A list of all updates to check
    :return: Returns false if the stream is never used or updated to be used
    :rtype: bool
    """
    changes = getAllChanges(updates)
    allStreamsInUse = [source.signal for source in sources]
    allStreamsInUse += set([c["value"] for c in changes if c["attr"] == "signal"])
    return streamId in allStreamsInUse


def isStreamUsedForEvent(
    stream: AudioStream, sources: List[AudioElement], updates: List[Update], rootPath
) -> bool:
    """Returns true if an audio source in event mode refers to this stream or if an update
    changes an audio source in event mode to use this audio stream.

    :param streamId: The string identifier of the audio stream
    :type streamId: str
    :param sources: A list of all audio elements in the scene
    :param updates: A list of all updates to check
    :return: Returns false if the stream is never used or updated to be used by an event mode audio source
    :rtype: bool
    """
    streamId = stream.id
    changes = getAllChanges(updates)
    allStreamsInUse = [source.signal for source in sources if source.mode == Mode.Event]
    eventSources = [source.id for source in sources if source.mode == Mode.Event]
    allStreamsInUse += set(
        [c["value"] for c in changes if c["attr"] == "signal" and c["id"] in eventSources]
    )
    isUsedForEvent = streamId in allStreamsInUse
    duration = 0.0
    if isUsedForEvent:
        f = sf.SoundFile(rootPath + "/" + stream.file)
        duration = f.frames / f.samplerate

    return isUsedForEvent, duration


def isAudioSourceMadeActive(audioSourceId: str, updates: List[Update]) -> bool:
    """Returns true if the audiosource's 'active' attribute was made true by an update

    :param audioSourceId: The string identifier of the audio source to check
    :type audioSourceId: str
    :param updates: A list of updates to check if the audio source is rendered
    :return: Returns false if the audio source is never made active by updates
    :rtype: bool
    """
    changes = getAllChanges(updates)
    allSourcesMadeActive = [
        c["id"] for c in changes if c["attr"] == "active" and c["value"] == "true"
    ]
    return audioSourceId in allSourcesMadeActive
