# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Dict, List, Any, Union

from ..audiostream import AudioStream
from ..basetypes import AudioElement, Geometry
from ..channelsource import ChannelSource
from ..granularsound import GranularSound
from ..hoasource import HOASource
from ..material import AcousticMaterial
from ..mesh import Mesh
from ..objectsource import ObjectSource
from ..primitives import Box, Cylinder, Sphere
from ..update import Update
from ..portal import Portal
from .redundancy import isAudioSourceMadeActive, isMaterialUsed, isStreamUsed, isStreamUsedForEvent
from .staticity import isEntityStatic

from xml.etree.ElementTree import Element


class SceneMetadata:
    """Provides second order data about the Entities in a Scene , i.e. data
    not stated in the EIF, but rather data derived from the EIF.
    """

    def __init__(self) -> None:
        self.entityStaticity = {}
        self.usedMaterials = {}
        self.usedStreams = {}
        self.usedSources = {}
        self.eventStreams = {}
        self.eventDurations = {}
        self.usedGranularSounds = {}

    def analyze(self, entities, voxelBlocks, rootFolder) -> None:
        """Analyze all previously added entities. add() and analyze() can be
        called in any order"""
        boxes = [b for b in entities if type(b) is Box]
        cylinders = [c for c in entities if type(c) is Cylinder]
        spheres = [s for s in entities if type(s) is Sphere]
        meshes = [m for m in entities if type(m) is Mesh]
        materials = [m for m in entities if type(m) is AcousticMaterial]
        channelSources = [c for c in entities if type(c) is ChannelSource]
        hoaSources = [h for h in entities if type(h) is HOASource]
        objectSources = [o for o in entities if type(o) is ObjectSource]
        streams = [s for s in entities if type(s) is AudioStream]
        granularsounds = [s for s in entities if type(s) is GranularSound]
        updates = [u for u in entities if type(u) is Update]
        portals = [p for p in entities if type(p) is Portal]

        geometry = [*boxes, *spheres, *cylinders, *meshes]
        sources = [*objectSources, *channelSources, *hoaSources]

        self.entityStaticity = self._getStaticityData([*geometry, *sources], updates)
        self.usedMaterials = self._getUsedMaterials(
            materials, geometry, voxelBlocks, portals, updates
        )
        self.usedSources = self._getRenderedSources(sources, updates)
        self.usedStreams = self._getUsedStreams(streams, sources, updates)
        self.eventStreams, self.eventDurations = self._getEventStreams(
            streams, sources, updates, rootFolder
        )
        self.usedGranularSounds = self._getUsedGranularSounds(granularsounds, sources, updates)

    def isMaterialUsed(self, materialId: str) -> bool:
        """Returns true if the AcousticMaterial is used by other Entities in a
        Scene. If false, the AcousticMaterial can be safely ignored.

        :param materialId: The string identifier for the AcousticMaterial
        :type material: str
        :return: True if used within the scene
        :rtype: bool
        """
        return self.usedMaterials[materialId]

    def isAudioStreamUsed(self, audioStreamId: str) -> bool:
        """Returns true if an AudioStream is used by other Entities in a Scene.
        If false, the AudioStream can be safely ignored.

        :param audioStreamId: The string identifier for the AudioStream
        :type audioStreamId: str
        :return: True if used within the scene
        :rtype: bool
        """
        return self.usedStreams[audioStreamId]

    def isGranularSoundUsed(self, granularId: str) -> bool:
        """Returns true if a granular sound source (database) is used by other Entities in a Scene.
        If false, the GranularSound can be safely ignored.

        :param granularId: The string identifier for the granular sound
        :type granularId: str
        :return: True if used within the scene
        :rtype: bool
        """
        return self.usedGranularSounds[granularId]

    def isAudioSourceRendered(self, audioSourceId: str) -> bool:
        """Returns true if the AudioElement is used by other Entities in Scene.
        If false, the AudioElement can be safely ignored.

        :param audioSourceId: The string identifier for the AudioElement
        :type audioSourceId: str
        :return: True if used within the scene
        :rtype: bool
        """
        return self.usedSources[audioSourceId]

    def isEntityStatic(self, entityId: str) -> bool:
        """Returns true if the Geometry or AudioElement never moves in space
        throughout the entire duration of the scene.

        :param entityId: The string identifier for the Geometry or AudioElement
        :type entityId: str
        :return: True if Entity never moves in space, False if Entity is dynamic
        :rtype: bool
        """
        return self.entityStaticity[entityId]

    @staticmethod
    def _getStaticityData(
        geometryAndAudioSources: List[Union[Geometry, AudioElement]],
        updates: List[Update],
    ) -> Dict[str, bool]:
        staticityData = {}
        for ent in geometryAndAudioSources:
            staticityData[ent.id] = isEntityStatic(ent.id, ent.transform, updates)
        return staticityData

    @staticmethod
    def _getUsedMaterials(
        materials: List[AcousticMaterial],
        geometry: List[Geometry],
        voxelBlocks: List[Element],
        portals: List[Portal],
        updates: List[Update],
    ) -> Dict[str, bool]:
        materialsUsed = {}
        for mat in materials:
            materialsUsed[mat.id] = isMaterialUsed(mat.id, geometry, voxelBlocks)
        for upd in updates:
            for mods in upd.modifications:
                if "material" in mods.changes.keys():
                    materialsUsed[mods.changes["material"]] = True
        for portal in portals:
            materialsUsed[portal.material] = True
        return materialsUsed

    @staticmethod
    def _getRenderedSources(sources: List[AudioElement], updates: List[Update]) -> Dict[str, bool]:
        sourcesRendered = {}
        for source in sources:
            isAlreadyActive = source.active
            isMadeActive = isAudioSourceMadeActive(source.id, updates)
            sourcesRendered[source.id] = isAlreadyActive or isMadeActive
        return sourcesRendered

    @staticmethod
    def _getUsedStreams(
        streams: List[AudioStream], sources: List[AudioElement], updates: List[Update]
    ) -> Dict[str, bool]:
        streamsUsed = {}
        for stream in streams:
            streamsUsed[stream.id] = isStreamUsed(stream.id, sources, updates)
        return streamsUsed

    @staticmethod
    def _getEventStreams(
        streams: List[AudioStream], sources: List[AudioElement], updates: List[Update], rootPath
    ) -> Dict[str, bool]:
        streamsUsed = {}
        durations = {}
        for stream in streams:
            streamsUsed[stream.id], durations[stream.id] = isStreamUsedForEvent(
                stream, sources, updates, rootPath
            )
        return streamsUsed, durations

    @staticmethod
    def _getUsedGranularSounds(
        streams: List[GranularSound], sources: List[AudioElement], updates: List[Update]
    ) -> Dict[str, bool]:
        streamsUsed = {}
        for stream in streams:
            streamsUsed[stream.id] = isStreamUsed(stream.id, sources, updates)
        return streamsUsed
