# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from __future__ import annotations
from enum import IntEnum
from typing import List

from .basetypes import AudioElement, CSpace, Entity, Location


class Alignment(IntEnum):
    """Indicates anchoring mode. "user" indicates that the labels are anchored
    by the user viewing direction. "object" indicates that the labels are anchored
    by the position/orientation of the ObjectSource element
    """

    User = 0
    Object = 1

    @staticmethod
    def fromString(alignmentString: str) -> Alignment:
        """Returns an Alignment value based upon the string

        :param alignmentString: The alignment string to base the return value from
        :type alignmentString: str
        :raises SyntaxError: Invalid alignment string
        :return: The alignment value
        """
        aString = alignmentString.strip().lower()
        if aString == "user":
            return Alignment.User
        elif aString == "object":
            return Alignment.Object
        else:
            raise SyntaxError(f"'{aString}' is invalid Alignment string")


class Positioning:
    """An input layout position indicates spatial assignment of waveforms"""

    def __init__(self):
        self.TL: bool = False  # Top-Left
        self.T: bool = False  # Top
        self.TR: bool = False  # Top-Right

        self.L: bool = False  # Left
        self.C: bool = False  # Center
        self.R: bool = False  # Right

        self.BL: bool = False  # Bottom-Left
        self.B: bool = False  # Bottom
        self.BR: bool = False  # Bottom-Right

    @staticmethod
    def fromString(positioningString: str) -> Positioning:
        position = Positioning()

        if positioningString == None or positioningString == "":
            return position

        possiblePositionLayouts = ["TL", "T", "TR", "L", "C", "R", "BL", "B", "BR"]
        for pos in positioningString.split(" "):
            if pos.upper() not in possiblePositionLayouts:
                raise SyntaxError(f"'{positioningString}' is an invalid InputLayoutPosition string")

            setattr(position, pos.upper(), True)

        return position


class InputLayout:
    """An InputLayout based on EIF InputLayout that is enhanced with extra
    metadata for the purposes of encoding to an MPEGI bitstream
    """

    def __init__(self, alignment: Alignment, positioning: Positioning):
        """Creates an InputLayout for an ObjectSource with EntityFactor. Do not
        construct without EntityFactory

        :param alignment: Indicates anchoring mode "user"/"object"
        :type alignment: str
        :param positioning: Intended spatial assignment of waveforms
        :type positioning: InputLayoutPosition
        """
        self.alignment = alignment
        self.positioning = positioning


class ObjectSource(Entity, AudioElement):
    """An ObjectSource based on an EIF ObjectSource that is enhanced with extra
    metadata for the purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: str,
        position: List[float],
        orientation: List[float],
        cspace: CSpace,
        active: bool,
        gainDb: float,
        refDistance: float,
        signal: str,
        extent: str,
        directivity: str,
        directiveness: float,
        mode: str,
        play: bool,
        inputLayout: InputLayout,
        aparams: List[str],
        transform: str,
        anchor: str,
        vdlMethod: int,
        hasReverbGain: bool,
        reverbGainDb: float,
        noReverb: bool,
        granularPos: List[float],
    ):
        """Creates an ObjectSource with EntityFactory. Do not construct without
        entityFactory

        :param id: Unique Indentifier
        :type id: str
        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: List[float]
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: List[float]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :type cspace: eifplus.CSpace
        :param active: A flag to render this source
        :type active: bool
        :param gainDb: The gain of this source expressed in dB
        :type gainDb: float
        :param refDistance: The reference distance for distance attenuation law
        :type refDistance: float
        :param signal: Unique identifier for the relevant AudioStream
        :type signal: str
        :param extent: Unique identifier for the relevant Geometry for the
            spatial extent of this source
        :type extent: str
        :param directivity: Unique identifier for the relavant directivity
        :type directivity: str
        :param directiveness: A scaling factor for the intensity of the
            directivity
        :type directiveness: float
        :param mode: A flag describing the playback mode for this object
        :type mode: str
        :param play: A flag for whether or not playback is enabled
        :type play: bool
        :param inputLayout: Describes the layout of associated waveforms
        :type inputLayout: eifplus.InputLayout
        :param aparams: Authoring parameters expressed as a list of strings
        :type aparams: List[str]
        :param transform: Unique identifier for parent transform
        :type transform: str
        :param anchor: Unique identifier for parent anchor
        :type anchor: str
        :param granularPos: Initial granular position
        :type granularPos: List[float]
        """
        Entity.__init__(self, id)
        AudioElement.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            active,
            gainDb,
            refDistance,
            signal,
            aparams,
            mode,
            play,
            vdlMethod,
            hasReverbGain,
            reverbGainDb,
            noReverb,
        )

        self.id  # Inherited by Entity

        # Inherited by AudioElement
        self.position
        self.orientation
        self.cspace
        self.active
        self.gainDb
        self.refDistance
        self.signal
        self.aparams
        self.mode
        self.play
        self.transform
        self.anchor
        self.hasReverbGain
        self.reverbGainDb
        self.noReverb

        self.extent = extent
        self.directivity = directivity
        self.directiveness = directiveness
        self.inputLayout = inputLayout
        self.granularPos = granularPos
