# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

from .basetypes import CSpace, Entity, Geometry, GeometryType, Location


class Primitive:
    def __init__(
        self,
        size: List[float],
        material: str,
    ):
        """A base class to represent a primitive, i.e Boxes, Cylinders, Spheres

        :param size: Size of dimensions in x, y, z
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type size: List[float]
        :param material: Unique identifier for the relavent AcousticMaterial
        :type material: str
        """
        self.size = size
        self.material = material

    @property
    def isTransparent(self):
        """Returns a bool based on whether or not the mesh is Acoustically
        "Transparent", i.e. acoustically irrelevant; it has no materials

        :return: Returns true if the mesh is acoustically transparent
        :rtype: bool
        """
        return self.material == None


class Box(Entity, Geometry, Primitive):
    """An EIF Box that is enhanced with extra metadata for the
    purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: str,
        position: List[float],
        orientation: List[float],
        size: List[float],
        cspace: CSpace,
        material: str,
        transform: str,
        anchor: str,
    ):
        """Creates a Box Primitive.

        :param id: Unique Indentifier
        :type id: str
        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: List[float]
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: List[float]
        :param size: Size of dimensions in x, y, z
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type size: List[float]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :type cspace: eifplus.CSpace
        :param material: Unique identifier for the relavent AcousticMaterial
        :type material: str
        :param transform: the ID of the parent transform of this mesh
        :type transform: str
        :param anchor: the ID of the parent anchor of this mesh
        :type anchor: str
        """
        Entity.__init__(self, id)
        Primitive.__init__(self, size, material)
        Geometry.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            GeometryType.Box,
        )


class Sphere(Entity, Geometry, Primitive):
    """An EIF Sphere that is enhanced with extra metadata for the
    purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: str,
        position: List[float],
        orientation: List[float],
        size: List[float],
        cspace: CSpace,
        material: str,
        transform: str,
        anchor: str,
    ):
        """Creates a Sphere Primitive.

        :param id: Unique Indentifier
        :type id: str
        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: List[float]
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: List[float]
        :param size: Size of dimensions in x, y, z
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type size: List[float]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :type cspace: eifplus.CSpace
        :param material: Unique identifier for the relavent AcousticMaterial
        :type material: str
        :param transform: the ID of the parent transform of this mesh
        :type transform: str
        :param anchor: the ID of the parent anchor of this mesh
        :type anchor: str
        """
        Entity.__init__(self, id)
        Primitive.__init__(self, size, material)
        Geometry.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            GeometryType.Sphere,
        )


class Cylinder(Entity, Geometry, Primitive):
    """An EIF Cylinder that is enhanced with extra metadata for the
    purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: str,
        position: List[float],
        orientation: List[float],
        size: List[float],
        cspace: CSpace,
        material: str,
        transform: str,
        anchor: str,
    ):
        """Creates a Cylinder Primitive.

        :param id: Unique Indentifier
        :type id: str
        :param position: Spatial position in cartesian coordinates expressed as
            a list of three floats [X, Y, Z]
        :type position: List[float]
        :param orientation: Spatial orientation in cartesian coordinates
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type orientation: List[float]
        :param size: Size of dimensions in x, y, z
            expressed as a list of three floats [Yaw, Pitch, Roll]
        :type size: List[float]
        :param cspace: The frame of reference for coordinate spaces (user/relative)
        :type cspace: eifplus.CSpace
        :param material: Unique identifier for the relavent AcousticMaterial
        :type material: str
        :param transform: the ID of the parent transform of this mesh
        :type transform: str
        :param anchor: the ID of the parent anchor of this mesh
        :type anchor: str
        """
        Entity.__init__(self, id)
        Primitive.__init__(self, size, material)
        Geometry.__init__(
            self,
            Location(position, orientation, cspace, transform, anchor),
            GeometryType.Cylinder,
        )
