# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from enum import IntEnum
from pathlib import Path
from typing import Iterable, List, Union

from mpegi.metadata import eif

from .acoustTuningParameters import AcousticTuningParameters
from .anchor import Anchor
from .audiostream import AudioStream
from .basetypes import AudioElement, Entity
from .channelsource import ChannelSource
from .directivity import Directivity
from .granularsound import GranularSound
from .entityfactory import EntityFactory
from .environments import AcousticEnvironment
from .headphoneEQ import HeadphoneEQ
from .hoasource import HOAGroup, HOASource
from .informedsource import InformedSource
from .material import AcousticMaterial
from .mesh import Mesh
from .objectsource import ObjectSource
from .portal import Portal
from .primitives import Box, Cylinder, Sphere
from .transform import Transform
from .update import Update
from .userregion import UserRegion
from .validation import (
    raiseExceptionIfAcousticEnvironmentIsMovedByUpdates,
    raiseExceptionIfDuplicateUpdateIndiciesExist,
)
from .airflowsource import AirflowSource
from .airflowgenerator import AirflowGenerator


class SceneType(IntEnum):
    """A type to indicate whether a scene is a VR scene or an AR scene"""

    VR = 0
    AR = 1


class Scene:
    """An EIF Scene that is enhanced with extra metadata to turn it into an
    encoder eifplus for encoding to an MPEGI bitstream. Once entities are finished
    being added to the eifplus, call the preprocess function to generate
    information that can only be made once a eifplus is finalised.
    """

    def __init__(
        self,
        id: str,
        entities: Iterable[Entity],
        duration: float = 0.0,
        sceneType: SceneType = SceneType.VR,
        enableLocallyCapturedAudio: bool = True,
        overrideListenerVoiceDirectivity: bool = False,
        listenerVoiceDirectivityId: str = "",
    ):
        """Creates an AudioScene

        :param id: A Scene Identifier
        :type id: str
        :param entities: A list of entities to add to the scene
        :param duration: Duration of scene in scenes. If 0, it will loop. Defaults to 0.0.
        :type duration: float, optional
        :param sceneType: Options are either "VR" or "AR" for "Virtual Reality"
            and "Augmented Reality" respectively, defaults to "VR"
        """
        self.id = id
        self.duration = duration
        self.type = sceneType
        self.enableLocallyCapturedAudio = enableLocallyCapturedAudio
        self.overrideListenerVoiceDirectivity = overrideListenerVoiceDirectivity
        self.listenerVoiceDirectivityId = listenerVoiceDirectivityId

        self.entities = {
            AcousticEnvironment: [],
            Anchor: [],
            AudioStream: [],
            Box: [],
            ChannelSource: [],
            Cylinder: [],
            Directivity: [],
            HOAGroup: [],
            HOASource: [],
            InformedSource: [],
            AcousticMaterial: [],
            Mesh: [],
            ObjectSource: [],
            Sphere: [],
            Transform: [],
            Update: [],
            UserRegion: [],
            Portal: [],
            AcousticTuningParameters: [],
            GranularSound: [],
            AirflowSource: [],
            AirflowGenerator: [],
            HeadphoneEQ: [],
        }

        for entity in entities:
            self.entities[type(entity)].append(entity)

        assert duration >= 0, "Duration cannot be negative seconds"

        raiseExceptionIfDuplicateUpdateIndiciesExist(self.entities[Update])
        raiseExceptionIfAcousticEnvironmentIsMovedByUpdates(
            self.entities[Update], self.entities[AcousticEnvironment]
        )

    @staticmethod
    def fromEif(eifScene: Union[eif.AudioScene, str]):
        """Instanciate an Encoder Scene from an EIF Scene

        :param eifScene: Either an EIF object or a path to an EIF file.
        :return: An encoder eifplus with the entities found in the given EIF eifplus
        :rtype: Scene
        """
        if isinstance(eifScene, str):
            eifScene = eif.AudioScene.load(eifScene)
        elif not isinstance(eifScene, eif.AudioScene):
            raise ValueError(
                "Cannot instanciate Encoder Scene with object of "
                f"type '{type(eifScene)}'. It is not an EIF AudioScene or string"
            )

        sceneType = SceneType.VR if eifScene.type.lower() == "vr" else SceneType.AR

        for condition in eifScene.proximityConditions:
            filtered_updates = [upd for upd in eifScene.updates if upd.condition == condition.id]
            for upd in filtered_updates:
                upd.condition = condition.region

        # Voxel-scene updates are encoded in VOXDATA payload
        if len(eifScene.voxelSceneIDs) > 0:
            for upd in eifScene.updates:
                upd.modifications = list(
                    filter(lambda mod: mod.id not in eifScene.voxelSceneIDs, upd.modifications)
                )

            # Remove updates w/o modifications
            eifScene.updates = list(
                filter(lambda upd: len(upd.modifications) > 0, eifScene.updates)
            )

        # The order of the entities here matters - change with great care
        eifEntities = (
            eifScene.transforms
            + eifScene.anchors
            + eifScene.streams
            + eifScene.materials
            + eifScene.directivities
            + eifScene.geometry
            + eifScene.environments
            + eifScene.objectSources
            + eifScene.hoaGroups
            + eifScene.hoaSources
            + eifScene.channelSources
            + eifScene.updates
            + eifScene.portals
            + [eifScene.acousticTuningParams]
            + eifScene.granularsound
            + eifScene.airflowSources
            + eifScene.airflowGenerators
            + [eifScene.headphoneEQ]
        )

        eifplusEntities = []
        for eifEntity in eifEntities:
            eifplusEntity = EntityFactory.manufactureEntity(eifEntity)
            eifplusEntities.append(eifplusEntity)

            if isinstance(eifEntity, eif.HoaGroup):
                for informedSourceEntity in eifplusEntity.informedSources:
                    eifplusEntities.append(informedSourceEntity)

        return Scene(
            eifScene.id,
            eifplusEntities,
            eifScene.duration,
            sceneType,
            eifScene.enableLocallyCapturedAudio,
            eifScene.overrideListenerVoiceDirectivity,
            eifScene.listenerVoiceDirectivityId,
        )

    def detectIsSmallScene(scene):
        isSmallScene = True

        for cylinder in scene.getCylinders():
            if any(abs(cylinder.position > 100)) or any(cylinder.size > 100):
                isSmallScene = False
                return isSmallScene
        for box in scene.getBoxes():
            if any(abs(box.position > 100)) or any(box.size > 100):
                isSmallScene = False
                return isSmallScene
        for sphere in scene.getSpheres():
            if any(abs(sphere.position > 100)) or any(sphere.size > 100):
                isSmallScene = False
                return isSmallScene

        for mesh in scene.getMeshes():
            for verts in mesh.vertices:
                if any([v for v in verts if abs(v) > 100]):
                    isSmallScene = False
                    return isSmallScene
            if any([coord for coord in mesh.position if abs(coord) > 100]):
                isSmallScene = False
                return isSmallScene

        for objects in scene.getObjectSources():
            if any([coord for coord in objects.position if abs(coord) > 100]):
                isSmallScene = False
                return isSmallScene

        for channels in scene.getChannelSources():
            if any([coord for coord in channels.position if abs(coord) > 100]):
                isSmallScene = False
                return isSmallScene
            for speakers in channels.loudspeakers:
                if any([coord for coord in speakers.position if abs(coord) > 100]):
                    isSmallScene = False
                    return isSmallScene

        for hoa in scene.getHoaSources():
            if any(abs(hoa.position) > 100):
                isSmallScene = False
                return isSmallScene

        for transforms in scene.getTransforms():
            if any([coord for coord in transforms.position if abs(coord) > 100]):
                isSmallScene = False
                return isSmallScene

        for update in scene.getUpdates():
            for mod in update.modifications:
                if "position" in mod.changes:
                    if mod.changes["position"][0] != "$":
                        position = [float(c) for c in mod.changes["position"].split()]
                        if any(abs(coord) > 100 for coord in position):
                            isSmallScene = False
                            return isSmallScene
                    else:
                        isSmallScene = False
                        return isSmallScene

        return isSmallScene

    def get(self, entityId: str) -> Entity:
        """Return the entity that matches the given id

        :param entityId: The id of the entity to be returned
        :type entityId: str
        :raises KeyError: Entity does not exist in scene
        :return: The matched entity
        """
        if not self.isInScene(entityId):
            raise KeyError(f"Entity {entityId} does not exist in scene")

        entityIds = [e.id for e in self.getAllEntities()]
        index = entityIds.index(entityId)
        return self.getAllEntities()[index]

    def getAnchors(self) -> List[Anchor]:
        """:return: A list of base eifplus Anchors"""
        return self.entities[Anchor]

    def getAudioStreams(self) -> List[AudioStream]:
        """:return: A list of base eifplus Audio Streams"""
        return self.entities[AudioStream]

    def getAudioSources(self) -> List[AudioElement]:
        """Returns a all audio sources within the eifplus: ObjectSources,
        ChannelSources, HOASources, as a list of lists

        :return: A list of all audio source lists
        """
        objSources = self.entities[ObjectSource]
        chanSources = self.entities[ChannelSource]
        hoaSources = self.entities[HOASource]

        return sum([objSources, chanSources, hoaSources], [])

    def getBoxes(self) -> List[Box]:
        """:return: A list of base eifplus Boxes"""
        return self.entities[Box]

    def getChannelSources(self) -> List[ChannelSource]:
        """:return: A list of base eifplus ChannelSources"""
        return self.entities[ChannelSource]

    def getCylinders(self) -> List[Cylinder]:
        """:return: A list of base eifplus Cylinders"""
        return self.entities[Cylinder]

    def getDirectivities(self) -> List[Directivity]:
        """:return: A list of base eifplus Directivities"""
        return self.entities[Directivity]

    def getGranularSound(self) -> List[GranularSound]:
        """:return: A list of base eifplus GranularDB"""
        return self.entities[GranularSound]

    def getEnvironments(self) -> List[AcousticEnvironment]:
        """:return: A list of base eifplus Environments"""
        return self.entities[AcousticEnvironment]

    def getGeometry(self):
        """:return: A list of geometry lists
        :rtype: List[List[eifplus.Mesh], List[eifplus.Box], List[eifplus.Cylinder],
            List[eifplus.Sphere]]
        """
        boxList = self.entities[Box]
        cylinderList = self.entities[Cylinder]
        sphereList = self.entities[Sphere]
        meshList = self.entities[Mesh]

        return [meshList, boxList, cylinderList, sphereList]

    def getHeadphoneEQ(self) -> List[HeadphoneEQ]:
        """:return: A list of base eifplus HeadphoneEQs"""
        return self.entities[HeadphoneEQ]

    def getHoaGroups(self) -> List[HOAGroup]:
        """:return: A list of base eifplus HOAGroups"""
        return self.entities[HOAGroup]

    def getHoaSources(self) -> List[HOASource]:
        """:return: A list of base eifplus HOASources"""
        return self.entities[HOASource]

    def getInformedSources(self) -> List[InformedSource]:
        """:return: A list of base eifplus InformedSources"""
        return self.entities[InformedSource]

    def getMaterials(self) -> List[AcousticMaterial]:
        """:return: A list of base eifplus Materials"""
        return self.entities[AcousticMaterial]

    def getMeshes(self) -> List[Mesh]:
        """:return: A list of base eifplus Meshes"""
        return self.entities[Mesh]

    def getObjectSources(self) -> List[ObjectSource]:
        """:return: A list of base eifplus ObjectSources"""
        return self.entities[ObjectSource]

    def getPortals(self) -> List[Portal]:
        """:return: A list of base eifplus Portals"""
        return self.entities[Portal]

    def getSpheres(self) -> List[Sphere]:
        """:return: A list of base eifplus Spheres"""
        return self.entities[Sphere]

    def getTransforms(self) -> List[Transform]:
        """:return: A list of base eifplus Transforms"""
        return self.entities[Transform]

    def getUserRegions(self) -> List[UserRegion]:
        """:return: A list of base eifplus User Regions"""
        return self.entities[UserRegion]

    def getUpdates(self) -> List[Update]:
        """:return: A list of base eifplus updates"""
        return self.entities[Update]

    def getAcoustTuningParams(self) -> List[AcousticTuningParameters]:
        return self.entities[AcousticTuningParameters]

    def getAirflowSources(self) -> List[AirflowSource]:
        return self.entities[AirflowSource]

    def getAirflowGenerators(self) -> List[AirflowGenerator]:
        return self.entities[AirflowGenerator]

    def getAllEntities(self) -> List[Entity]:
        """:return: Returns all entities in the scene"""
        return sum(self.entities.values(), [])

    def isInScene(self, entityId: str) -> bool:
        """Returns true if the if matches to an existing entity in the scene

        :param entityId: Unique Entity ID String
        :type entityId: str
        :return: False if cannot be found
        :rtype: bool
        """
        return entityId in [e.id for e in self.getAllEntities()]
