# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Optional

from .basetypes import Entity


class Modification:
    """A Modification of an Update based on an EIF Modification that is enhanced
    with extra metdata for the purposes of encoding to an MPEGI bitstream
    """

    def __init__(self, id: str, duration: float, changes):
        """Creates a Modification with EntityFactory. Do not construct without
        entityFactory

        :param id: Identifier of the target entity to modify
        :type id: str
        :param duration: The length of time to adapt the old values to the new
            values
        :type duration: float
        :param changes: A dict of the changes to an entity. The key corresponds
            to the name of the value to change and the value is the new value
            which could be of types int, float, list, str, or bool.
        :type changes: dict
        """
        self.id = id
        self.duration = duration
        self.changes = changes


class Update(Entity):
    """An Update based on an EIF Update that is enhanced with extra metadata
    for the purposes of encoding to an MPEGI bitstream
    """

    def __init__(
        self,
        id: Optional[str],
        index: Optional[int],
        time: Optional[float],
        condition: Optional[str],
        fireOn: Optional[bool],
        delay: Optional[float],
        type: str,
        modifications: List[Modification],
    ):
        """Creates an Update with EntityFactory. Do not construct without
        entityFactory

        :param id: Unique Identifier of this update
        :type id: Optional[str]
        :param index: Unique Global Index of this update
        :type index: Optional[int]
        :param time: The time in which the update should be executed in seconds
        :type time: Optional[float]
        :param condition: Unique Idenfifier of the ListenerProximityCondition
        :type condition: Optional[str]
        :param fireOn: Update fires when this state is reached
        :type fireOn: Optional[bool]
        :param delay: Postpone the update in seconds
        :type delay: Optional[float]
        :param type: Either Timed, Conditonal, Triggered, Dynamic
        :type type: str
        :param modifications: A list of Modification()
        :type modifications: list
        """
        Entity.__init__(self, id)

        self.modifications = modifications

        self.id  # Inherited by Entity

        self.index = index
        self.time = time
        self.condition = condition
        self.fireOn = fireOn
        self.delay = delay
        self.type = type
