# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Scene geometric analysis for determining parameters for simple early reflections."""
import math
from typing import Dict, List, Optional, Tuple

import numpy as np
import scipy
import trimesh

from mpegi.geometry import trimesh as mpegtri
from mpegi.geometry.transforms import getOffsetAndAnglesFromMatrixTRS, transformMatrixTRS_4x4
from mpegi.metadata import eifplus


def get_xz_rays(n: int) -> np.ndarray:
    """Returns n ray directions (unit length) spaced equally on horizontal (X-Z) plane

    :param n: number of ray directions to return
    :type n: int
    :return: numpy.ndarray of shape (n, 3), where each row is one ray direction
    :rtype: numpy.ndarray
    """

    def pol2cart(r, phi):
        x = r * np.cos(phi)
        y = r * np.sin(phi)
        return x, y

    angles = np.linspace(start=0, stop=2 * np.pi, num=n, endpoint=False)
    x, z = pol2cart(r=1, phi=angles)

    ray_dirs = np.zeros((n, 3))
    ray_dirs[:, 0] = x
    ray_dirs[:, 2] = z
    return ray_dirs


def simple_elevation(vecs: np.ndarray) -> np.ndarray:
    """Return the elevation from X-Z plane

    :param vecs: [X, Y, X] directions, shape (n, 3)
    :type vecs: numpy.ndarray
    :return: numpy.ndarray of shape (n,), where elevation angles from horizontal plane in degrees
    :rtype: numpy.ndarray
    """
    if len(vecs) == 0:
        return vecs
    hxy = np.hypot(vecs[:, 0], vecs[:, 2])  # lengths of these 2-vectors
    ele = np.arctan2(vecs[:, 1], hxy)
    return ele / (2.0 * np.pi) * 360.0


def get_sampling_points(
    urr_trimesh: trimesh.Trimesh, division_level: Optional[int] = 1
) -> np.ndarray:
    """Given a user-reachable region box as a trimesh, provide a number of sampling
    points from inside it.
    The number of sampling points is determined by division_level:
    - 1 -> only the center of the box is given
    - 2 -> 1 + 8
    - 3 -> 1 + 2*8
    The points are spaced equidistantly in the 3 dimensions depending on the
    subdivision level.

    :param urr_trimesh: input user-reachable region trimesh
    :type urr_trimesh: trimesh.Trimesh
    :param division_level: level of sampling pointn sub-divisions
    :type division_level: int
    :return: coordinates of the sampling points, shape (n_points, 3)
    :rtype: numpy.ndarray
    """
    sample_points = [urr_trimesh.centroid[None, :]]  # centroid
    if division_level > 1:
        # oriented bounding box
        to_origin, extents = trimesh.bounds.oriented_bounds(
            obj=urr_trimesh, angle_digits=1
        )  # to_origin: (4, 4) TRS, extents: (3, )

        # invert transform matrix -> inverse translation and rotation
        inv_trans = np.linalg.pinv(to_origin)

        for sub_level in range(1, division_level):
            sz_scaler = float(sub_level) / float(division_level)
            new_mesh = trimesh.creation.box(extents=extents * sz_scaler, transform=inv_trans)
            new_mesh.merge_vertices()
            sample_points.append(new_mesh.vertices)

    all_points = np.concatenate(sample_points, axis=0)
    return all_points


def aa_box_overlap(mesh_list: List[trimesh.Trimesh]) -> Tuple[int, List[int]]:
    """Axis-aligned bounding box overlap -based grouping.

    For each mesh, take the AABB and assign meshes with overlapping
    AABBs into the same group.

    A more accurate version would use oriented bounding boxes,
    but implementing that is (for now) so difficult that this
    approximation should be enough.

    :param mesh_list: meshes whose overlap is approximated using axis-aligned boxes
    :type mesh_list: list of trimesh.Trimesh
    :return:
        - n_groups: int: number of unique groups of overlapping meshes
        - group_labels: list of int: length is equal to the length of mesh_list. each element denotes the group in which the mesh belongs to
    """

    def is_overlapping_1d(range1: List[float], range2: List[float]) -> bool:
        """1D overlap detection

        :param range1: min and max coordinate in the target axis
        :type range1: list or tuple of float
        :param range2: min and max coordinate in the target axis
        :type range2: list or tuple of float
        :return: are the two ranges overlapping
        :rtype: bool
        """
        return (max(range1) >= min(range2)) and (max(range2) >= min(range1))

    n_meshes = len(mesh_list)
    # collect all axis-aligned boundign boxes into a list.
    # a better approach would be to use oriented BBs, but the overlap
    # check will be much more difficult.
    all_bbs = []
    for one_mesh in mesh_list:
        all_bbs.append(one_mesh.bounds)  # (2, 3) float

    # N*N bool matrix if the two meshes (potentially) overlap
    has_overlap = np.zeros((n_meshes, n_meshes), dtype=bool)
    for m1_idx in range(0, n_meshes):
        bb1 = all_bbs[m1_idx]
        for m2_idx in range(m1_idx, n_meshes):
            bb2 = all_bbs[m2_idx]
            overlap_in_x = is_overlapping_1d(bb1[:, 0], bb2[:, 0])
            overlap_in_y = is_overlapping_1d(bb1[:, 1], bb2[:, 1])
            overlap_in_z = is_overlapping_1d(bb1[:, 2], bb2[:, 2])
            has_overlap[m1_idx, m2_idx] = overlap_in_x & overlap_in_y & overlap_in_z
            has_overlap[m2_idx, m1_idx] = has_overlap[m1_idx, m2_idx]

    # treating the overlappedness information as adjacency matrix, solve connected components,
    # i.e., the urrs that are (partly) overlapping, including transitivity
    n_groups, group_labels = scipy.sparse.csgraph.connected_components(has_overlap, directed=False)

    return n_groups, group_labels


def mirror_peak_picking(
    x: np.ndarray, peak_dist_thr: Optional[int] = 1, n_max_peaks: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Peak-picking, but uses scaled mirrored copies to handle edge cases.

    :param x: input signal
    :type x: numpy.ndarray
    :param peak_dist_thr: minimum distance between peaks
    :type peak_dist_thr: int, optional
    :param n_max_peaks: maximum number of peaks to return
    :type n_max_peaks: int, optional

    :return:
        - peak_indices : numpy.ndarray
        - peak_prominences : numpy.ndarray: prominence value for the peaks after scaling the input signal to peak==1
    """
    eps = 1e-15
    x_len = len(x)
    x_col = x.flatten()
    # the concatenation handles edges
    g = 0.9  # <1 to prevent selecting mirrored peaks over true ones at edges
    y = np.concatenate((g * x_col[::-1], x_col, g * x_col[::-1]))

    this_peaks, _ = scipy.signal.find_peaks(y, distance=peak_dist_thr)

    # normalize the values to peak=1 so that the prominences become more comaprable
    norm_val = np.max(y)
    if norm_val < eps:
        norm_val = 1.0
    peak_prominences = scipy.signal.peak_prominences(y / norm_val, this_peaks)[0]

    this_peaks = this_peaks - x_len
    if -1 in this_peaks:
        this_peaks[this_peaks == -1] = 0

    # remove the peaks from the mirrored parts
    peak_prominences = peak_prominences[this_peaks >= 0]
    this_peaks = this_peaks[this_peaks >= 0]
    peak_prominences = peak_prominences[this_peaks < x_len]
    this_peaks = this_peaks[this_peaks < x_len]

    # return only n_max_peaks most prominent peaks
    if (n_max_peaks is not None) and (len(this_peaks) > n_max_peaks):
        sort_indices = np.argsort(peak_prominences)
        this_peaks = this_peaks[sort_indices[-1 : -(n_max_peaks + 1) : -1]]
        peak_prominences = peak_prominences[sort_indices[-1 : -(n_max_peaks + 1) : -1]]

    return this_peaks, peak_prominences


def simple_er_analysis(
    *,
    scene_metadata: eifplus.SceneMetadata,
    geometry: List[eifplus.Geometry],
    env_regions: List[str],
    transforms: List[eifplus.Transform],
    audio_sources: List[eifplus.AudioElement],
    user_regions: List[eifplus.UserRegion],
) -> Tuple[Dict[str, int], List[Dict[str, List[float]]]]:
    """This is the main routine of this module. Analyzes the entire EIF scene and
    produces aprameters for simple early reflections.

    Returns None
        - if no sound sources outside acoustic environments.
        - if scene trimesh is empty (Beach)
        - if no user-reachable areas are defined

    :param scene_metadata: Metadata about the scene
    :param geometry: All geometry within the scene
    :param env_regions: All ids for the regions of acoustic environments
    :param audio_sources: All audio sources that are in the scene
    :param user_regions: All user reachable regions in the scene
    :return:
        - urr_group_map : dict(str, int): user-reachable region id to group_idx mapping
        - group_er_params : list of dict(str, float) for each URR group, ER parameter dictionary: {'distance': [float, float], 'prominence': [float̋, float])
    """
    # increase the AE size by this amount for more robust "contains" test
    ae_size_eps = 0.5

    # definitions for faces relevant for the simple early reflections
    max_vert_angle_dev = 5  # max accepted deviation from vertical, in degrees
    n_rays = 60  # number of sampling rays in each sampling point

    # determine sampling points within user-reachable regions.
    # 1: just the centroid, 2: centroid + 8 intermediate, 3: centroid + 2*8 intermediate
    sampling_level = 2

    # for the ding reflective area statistics
    max_dist = 50  # m
    ring_width = 3  #
    do_ring_overlap = False
    # peak selection from the ring statistics
    n_max_peaks = 2
    peak_dist_m = 5

    # keeping ring size constant, adjust max distance for a round number
    if do_ring_overlap:
        n_rings = math.ceil(max_dist / (ring_width / 2))
        max_dist_eff = n_rings * ring_width / 2
    else:
        n_rings = math.ceil(max_dist / ring_width)
        max_dist_eff = n_rings * ring_width

    # define the rings
    n_borders = math.ceil((1 + do_ring_overlap) * n_rings + 1)
    ring_borders, ring_bin_width = np.linspace(
        0, max_dist_eff, num=n_borders, endpoint=True, retstep=True
    )
    n_rings = n_borders - (1 + do_ring_overlap)
    ring_centers = np.zeros((n_rings,), dtype=float)
    peak_dist_thr = math.ceil(
        peak_dist_m / ring_bin_width
    )  # peaks closer that 10m are not considered

    for dist_idx in range(n_rings):
        # hits within this distance "ring"
        end_idx = dist_idx + (1 + do_ring_overlap)
        ring_centers[dist_idx] = (ring_borders[dist_idx] + ring_borders[end_idx]) / 2

    geometry = _apply_parent_transforms(geometry, transforms)

    # combined mesh of acoustic environments
    ae_geos = []
    for one_geo in geometry:
        if one_geo.id in env_regions:
            if hasattr(one_geo, "size"):
                # increase the AE size slightly
                one_geo.size += ae_size_eps
            ae_geos.append(one_geo)
    ae_trimeshes = mpegtri.convertGeometriesToTrimesh(
        ae_geos, includeTransparent=True, concatenate=False
    )

    # check for audio sources outside acoustic environments
    src_list = audio_sources
    # need to apply the parent transforms on source positions, too
    src_list = _apply_parent_transforms(src_list, transforms)
    src_outside_aes = len(src_list) > 0
    if (len(src_list) > 0) and (len(ae_geos) > 0):
        # audio sources and acoustic environments exist => can check for overlap
        src_positions = np.zeros((len(src_list), 3))
        for src_idx, one_src in enumerate(src_list):
            src_positions[src_idx, :] = one_src.position

        src_in_ae = np.zeros((len(src_list),), dtype=bool)
        # need to check each AE separately, checking the merged mesh doesn't work
        for one_ae in ae_trimeshes:
            src_in_ae = np.logical_or(src_in_ae, one_ae.contains(src_positions))

        src_outside_aes = sum(src_in_ae) != len(src_in_ae)

    if not src_outside_aes:
        # print('INFO: AudioSources do not exit outside AcousticEnvironments. Assuming an in-doors scene. Not continuing.')
        return None, None

    # these are the step from makeSceneTrimeshER(), except that here we keep the cylinders
    geometry = [g for g in geometry if scene_metadata.isEntityStatic(g.id)]
    for env_region in env_regions:  # Remove acoustic environments
        if env_region in [geo.id for geo in geometry]:
            geometryIds = [geo.id for geo in geometry]
            index = geometryIds.index(env_region)
            geometry.pop(index)
    geometry = [g for g in geometry if not g.isTransparent]

    # geometry = preprocess.PreprocessorComponent.removeSpheresAndCylinders(geometry)  # keep the pillars in Park
    scene_trimesh_er = mpegtri.convertGeometriesToTrimesh(geometry)[0]

    # if the function does in-place cleanup, ret_val is None. otherwise ret_val is the cleaned mesh
    scene_trimesh_er.remove_degenerate_faces()
    scene_trimesh_er.remove_duplicate_faces()
    scene_trimesh_er.remove_infinite_values()
    scene_trimesh_er.remove_unreferenced_vertices()

    if scene_trimesh_er.is_empty:  # e.g., Beach
        # print('WARNING: Scene trimesh is empty. Assuming no ERs.')
        return None, None

    # determine user-reachable regions
    user_regions = _apply_parent_transforms(user_regions, transforms)

    if len(user_regions) == 0:
        # no user-reachable regions defined, e.g., Offices
        # print('WARNING: No user-reachable regions defined. Unable to proceed.')
        return None, None

    # convert user-reachable regions into trimeshes
    urr_meshes = [mpegtri.convertGeometriesToTrimesh([urr])[0] for urr in user_regions]
    # group overlapping URRs
    n_urr_groups, urr_group_labels = aa_box_overlap(urr_meshes)

    # dictionary from urr id str to int group index
    urr_group_map = {}
    for urr_idx, urr in enumerate(user_regions):
        urr_group_map[urr.id] = urr_group_labels[urr_idx]

    # mapping from face to a facet, if one exists
    n_faces = scene_trimesh_er.faces.shape[0]
    face_to_facet = [None] * n_faces
    facet_heights = np.zeros((len(scene_trimesh_er.facets), 1))  # Y-dim range of the facet
    for facet_idx, this_faces in enumerate(scene_trimesh_er.facets):
        for face_idx, one_face in enumerate(this_faces):
            face_to_facet[one_face] = facet_idx
            one_min_y = np.min(scene_trimesh_er.triangles[one_face, :, 1])
            one_max_y = np.max(scene_trimesh_er.triangles[one_face, :, 1])
            if face_idx == 0:
                min_y = one_min_y
                max_y = one_max_y
            else:
                min_y = min(min_y, one_min_y)
                max_y = max(max_y, one_max_y)

        facet_heights[facet_idx] = max_y - min_y

    # height of each face: size of the range on y-axis
    face_heights = np.max(scene_trimesh_er.triangles[:, :, 1], axis=1) - np.min(
        scene_trimesh_er.triangles[:, :, 1], axis=1
    )

    # setup the ray caster
    ray_caster = trimesh.ray.ray_triangle.RayMeshIntersector(scene_trimesh_er)
    # create a circle of equidistant ray directions
    ray_dirs = get_xz_rays(n=n_rays)

    # loop over URR groups. within each group, accumulate the results
    group_er_params = [None] * n_urr_groups
    for urr_grp in range(n_urr_groups):
        this_meshes = np.nonzero(urr_group_labels == urr_grp)[0].tolist()  # => list of mesh indices

        # statistics over this group
        all_ring_areas = []
        sample_outside_ae = False

        # loop over all URRs in this group
        for urr_idx in this_meshes:
            one_urr = urr_meshes[urr_idx]  # => trimesh.Trimesh
            # get a number of sampling points, equally distributed inside the URR
            sampling_points = get_sampling_points(one_urr, division_level=sampling_level)
            n_sampling_points = sampling_points.shape[0]

            # loop over the sampling points in this URR
            for point_idx in range(n_sampling_points):
                sampling_point = sampling_points[point_idx, :]

                if (len(ae_trimeshes) > 0) and (
                    any([ae.contains(sampling_point[None, :]) for ae in ae_trimeshes])
                ):
                    # there's an acoustic environment defined, and the user-reachable area in (partly) covered by it
                    continue
                else:
                    # this sampling point is outside of AE => should be checked
                    sample_outside_ae = True

                # shoot rays around the samplign point
                ray_origins = np.tile(np.array(sampling_point), (n_rays, 1))
                index_tri, index_ray, hit_locs = ray_caster.intersects_id(
                    ray_origins=ray_origins,
                    ray_directions=ray_dirs,
                    return_locations=True,  # we need hit locations for distance
                    multiple_hits=False,
                )  # take only the first hit

                # check that the hits are on the front of the triangles
                back_hits = (
                    np.sum(
                        np.multiply(
                            ray_dirs[index_ray, :], scene_trimesh_er.face_normals[index_tri, :]
                        ),
                        axis=1,
                    )
                    > 0
                )
                front_hits = np.logical_not(back_hits)

                if np.all(back_hits):
                    # no point checking in more detail
                    continue

                # discard all hits that are not on the front of the mesh
                index_tri = index_tri[front_hits]
                index_ray = index_ray[front_hits]
                hit_locs = hit_locs[front_hits, :]

                n_hits = len(index_tri)

                # distance of the hit point from the origin
                ray_lens = np.sqrt(np.sum(np.square(np.abs(hit_locs - ray_origins[0, :])), axis=1))

                # how vertical is the face
                hit_normals = scene_trimesh_er.face_normals[index_tri]
                normal_els = simple_elevation(hit_normals)

                # for each hit, how big is the face or facet it hit, what's the height of the hit surface
                hit_heights = np.zeros((n_hits,))
                for hit_idx, face_idx in enumerate(index_tri):
                    this_facet = face_to_facet[face_idx]
                    if this_facet is not None:
                        # this face is a part of a facet
                        hit_heights[hit_idx] = facet_heights[this_facet]
                    else:
                        # lone face
                        hit_heights[hit_idx] = face_heights[face_idx]

                # bool filter of hit surfaces approximately vertical
                vert_hits = np.logical_and(
                    (normal_els > -max_vert_angle_dev), (normal_els < max_vert_angle_dev)
                )

                # check all vertical face hits and compute area descriptors for rings
                # of distances from the center. these describe the amount of potential
                # reflective energy from that distance => delay
                ring_areas = np.zeros((n_rings,))
                for dist_idx in range(n_rings):
                    # hits within this distance "ring"
                    end_idx = dist_idx + (1 + do_ring_overlap)

                    ring_hits = np.logical_and(
                        ray_lens >= ring_borders[dist_idx], ray_lens < ring_borders[end_idx]
                    )
                    vert_ring_hits = np.logical_and(ring_hits, vert_hits)
                    if np.sum(vert_ring_hits) > 0:
                        # what is the equivalent area of the sector hit.
                        # with constant absolute height, the equivalent area behaves as 1/r.
                        # this version clips the effective height to 45 degree angle, i.e., distance
                        max_h = max_dist_eff
                        # what is the maximum area per ray: apply height limit at the far end of the ring
                        max_area_per_ray = min(max_h, ring_borders[end_idx]) / ring_borders[end_idx]
                        max_area = n_rays * max_area_per_ray
                        # for each vertical hit ray in this ring, the area contibution, clipped by the maximum value
                        clipped_ray_areas = np.minimum(
                            max_area_per_ray, hit_heights[vert_ring_hits] / ray_lens[vert_ring_hits]
                        )
                        # sum over all rays and normalize by theoretical maximum
                        ring_areas[dist_idx] = np.sum(clipped_ray_areas) / max_area

                all_ring_areas.append(ring_areas)
            # end of sampling point loop

        if not sample_outside_ae:  # np.all(sample_in_ae):
            # print('INFO: All sampling points in URR group {} are inside acoustic environments.'.format(urr_grp+1))
            # will not analyse the scene further
            continue

        if len(all_ring_areas) == 0:
            # print('WARNING: No acceptable hits found in URR group {}.'.format(urr_grp+1))
            pass

        else:
            # merge results from all sampling points
            all_ring_areas_arr = np.stack(all_ring_areas, axis=1)

            # average ring areas over all sampling points and urrs
            mean_ring_areas = np.mean(all_ring_areas_arr, axis=1)

            this_peaks = mirror_peak_picking(
                mean_ring_areas, peak_dist_thr=peak_dist_thr, n_max_peaks=n_max_peaks
            )[0]

            peak_ord = np.argsort(this_peaks)
            this_peaks = this_peaks[peak_ord]

            n_found_peaks = len(this_peaks)
            out_dists = [None] * n_max_peaks
            out_prominences = [None] * n_max_peaks
            out_dists[:n_found_peaks] = ring_centers[this_peaks]
            out_prominences[:n_found_peaks] = mean_ring_areas[this_peaks]

            urr_result = {"distances": out_dists, "prominences": out_prominences}

            group_er_params[urr_grp] = urr_result

    return urr_group_map, group_er_params


def new_er_outdoor(peak_dists: List[float], peak_heights: List[float]) -> Dict:
    """Mapping from potential reflection distances and heights into ER parameters.
    based on v1.0 MATLAB version

    :param peak_dists: [m] distance of first 2 peaks from central listening area
    :type peak_dists: list of float
    :param peak_heights: 2 normalized height values of the two peaks
    :type peak_heights: list of float
    :return: SimpleER parameters:
        - 'algo_type'
        - 'preDelay1'
        - 'preDelay2'
        - 'ampFac1db'
        - 'ampFac2db'
    :rtype: dict

    % Scene/URR     p1dist, p2dist  p1height, p2height
    % ParkingLot
    % [3,57,1,1.2,0.81] = NewEROutdoor([1.5, 19.5],[0.00333, 0.004104])
    % Rec_PianoBar
    % [3,66,0.3,1.4,0.96] = NewEROutdoor([1.5, 22.5],[0.10345 0.001096])
    % Rec_RockConc
    % [3,144,0.68,2.2,1] = NewEROutdoor([19.5, 49.5],[0.00008, 0.000014])
    % Rec_Waterfall
    % [3,66,0.6,1.4,0.93] = NewEROutdoor([1.5, 22.5],[0.01443, 0.001446])
    % Canyon
    % [3,57,0.6,1.3,0,87] = NewEROutdoor([4.5, 19.5],[0.01811, 0.002303])
    % DowntownDrum
    % [3,109,0.43,1,0,65] = NewEROutdoor([4.5, 37.5],[0.13001, 0.042821])
    % Park
    % [3,57,0.9,1,0,58] = NewEROutdoor([10.5, 19.5],[0.02639, 0.024268])
    % Battle
    % [3,144,0.1,1.4,0,9] = NewEROutdoor([1.5, 49.5],[0.54726, 0.004216])

    """

    n_peaks = len(peak_dists)
    assert n_peaks == 2

    if sum(peak_dists) / n_peaks < 0.001:
        algo_type = 0  # no reflections
        preDelay1 = None
        preDelay2 = None
        ampFac1Db = None
        ampFac2Db = None
    else:
        algo_type = 3  # four quad reflections
        # pre_delay = round(1000.0 * peak_dists[1] / c)  # 2nd peak defines predelay of algoType 3

        if peak_heights[0] < 0.001:
            peak_heights[0] = 0.001
        if peak_heights[1] < 0.001:
            peak_heights[1] = 0.001

        # tuned mapping from analysis to pattern parameters
        preDelay1 = round(25 * peak_dists[0])
        preDelay2 = round(3 * peak_dists[1])

        ampFac1_lin = 100 * peak_heights[0]
        ampFac2_lin = 100 * peak_heights[1]

        if ampFac1_lin > 3:
            ampFac1_lin = 3
        if ampFac2_lin > 3:
            ampFac2_lin = 3

        ampFac1Db = 20 * math.log10(ampFac1_lin)
        ampFac2Db = 20 * math.log10(ampFac2_lin)

    return {
        "algo_type": algo_type,
        "preDelay1": preDelay1,
        "preDelay2": preDelay2,
        "ampFac1Db": ampFac1Db,
        "ampFac2Db": ampFac2Db,
    }


def _apply_parent_transforms(
    geometry: List[eifplus.Geometry], transforms: List[eifplus.Transform]
) -> List[eifplus.Geometry]:
    for geo in geometry:
        if geo.transform != None:
            # get parent transform
            transform = [t for t in transforms if t.id == geo.transform][0]

            Tgeo = transformMatrixTRS_4x4(geo.position, np.deg2rad(geo.orientation), [1, 1, 1])
            Ttransform = transformMatrixTRS_4x4(
                transform.position, np.deg2rad(transform.orientation), [1, 1, 1]
            )
            Tglobal = np.matmul(Ttransform, Tgeo)
            newPos, newOri = getOffsetAndAnglesFromMatrixTRS(Tglobal)

            geo.position = newPos.tolist()
            geo.orientation = np.rad2deg(newOri).tolist()

    return geometry
