# This software module was originally developed by
#
# Nokia
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Nokia retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2024.

from typing import List

from mpegi.metadata.reverb.reverbutils import is_point_beyond
from trimesh import Trimesh
from trimesh.ray.ray_pyembree import RayMeshIntersector


def get_los_factor_p2p_multi(
    ray_origins: List[List[float]],
    meshes: List[Trimesh],
    target_vertices: List[float],
    dist_atol: float,
):
    n_rays = len(ray_origins)
    n_target_v = len(target_vertices)

    if n_rays == 3 and not isinstance(ray_origins[0], list):
        ray_origins = [ray_origins]

    if n_rays != n_target_v:
        raise Exception(
            f"Ray origin and target vertices lengths must match: {n_rays}, {n_target_v}"
        )

    is_single = len(ray_origins) == 1 and len(ray_origins[0]) == 3
    end_points = []
    divisions = []
    ray_origins_new = []

    for i in range(len(target_vertices)):
        target_v = target_vertices[i]
        target_v_len = len(target_v)
        divisions.append(target_v_len)
        pos = ray_origins[i]
        for v in target_v:
            end_points.append([v[0], v[1], v[2]])
            ray_origins_new.append(pos)

    ray_origins = ray_origins_new

    del ray_origins_new

    n_rays = len(end_points)
    ray_directions = []

    for i in range(n_rays):
        sp = ray_origins[i]
        ep = end_points[i]
        ray_directions.append([ep[0] - sp[0], ep[1] - sp[1], ep[2] - sp[2]])

    ray_los_truth_table = [True] * n_rays

    for mesh_id, trimesh in meshes:
        intersector = RayMeshIntersector(trimesh)
        hit_data, ray_idx_data, tri_idx_data = intersector.intersects_location(
            ray_origins=ray_origins, ray_directions=ray_directions, multiple_hits=False
        )
        hit = len(hit_data) > 0
        if hit:
            for i in range(len(hit_data)):
                hit_pos = hit_data[i]
                ray_idx = ray_idx_data[i]
                end_point = end_points[ray_idx]
                start_point = ray_origins[ray_idx]

                if ray_los_truth_table[ray_idx] == False:
                    continue

                # Compare the distance between p1 -> end_point, p1 -> hit_point
                has_los = is_point_beyond(start_point, end_point, hit_pos, atol=dist_atol)

                if not has_los:
                    ray_los_truth_table[ray_idx] = has_los

    del ray_origins
    del end_points
    del ray_directions

    los_factors = []
    ray_idx = 0
    for div in divisions:
        ray_los_data = []
        for i in range(div):
            ray_truth = ray_los_truth_table[ray_idx]
            ray_los_data.append(ray_truth)
            ray_idx += 1

        passed_rays = len([ray for ray in ray_los_data if ray == True])
        los_factor = passed_rays / div

        los_factors.append(los_factor)

    return los_factors if not is_single else los_factors[0]


# Single point los check to target points, slower
def check_los(
    ray_origin: List[float],
    ray_destinations: List[float],
    meshes_to_check: List[Trimesh],
    dist_atol: float,
):
    n_rays = len(ray_destinations)
    ray_directions = [None] * n_rays
    ray_origin_new = [ray_origin] * n_rays

    for i in range(n_rays):
        ray_directions[i] = [
            ray_destinations[i][0] - ray_origin_new[i][0],
            ray_destinations[i][1] - ray_origin_new[i][1],
            ray_destinations[i][2] - ray_origin_new[i][2],
        ]

    ray_los_truth_table = [True] * n_rays
    for mesh in meshes_to_check:
        ray_mesh_intersector = RayMeshIntersector(mesh)
        hit, ray_idx_data, tri_idx = ray_mesh_intersector.intersects_location(
            ray_origins=ray_origin_new, ray_directions=ray_directions
        )
        for i in range(len(hit)):
            hit_pos = hit[i]
            ray_idx = ray_idx_data[i]
            origin = ray_origin_new[ray_idx]
            destination = ray_destinations[ray_idx]

            has_los = is_point_beyond(origin, destination, hit_pos, atol=dist_atol)

            if not has_los:
                ray_los_truth_table[ray_idx] = False

    passed_rays = len([ray for ray in ray_los_truth_table if ray == True])
    return passed_rays / n_rays


def get_los_factor_p2p(
    start_point: List[float],
    end_points: List[List[float]],
    meshes: List[Trimesh],
    dist_atol: float,
):
    n_rays = len(end_points)
    ray_origins = [start_point] * n_rays
    ray_data = []
    x1, y1, z1 = start_point
    ray_directions = [[ep[0] - x1, ep[1] - y1, ep[2] - z1] for ep in end_points]
    ray_los_truth_table = [True] * n_rays
    no_los_ray_ids = []

    for trimesh in meshes:
        intersector = RayMeshIntersector(trimesh)
        hit_data, ray_idx_data, tri_idx_data = intersector.intersects_location(
            ray_origins=ray_origins, ray_directions=ray_directions, multiple_hits=False
        )
        hit = len(hit_data) > 0
        if hit:
            for i in range(len(hit_data)):
                hit_pos = hit_data[i]
                ray_idx = ray_idx_data[i]
                end_point = end_points[ray_idx]

                # Compare the distance between p1 -> end_point, p1 -> hit_point
                has_los = is_point_beyond(start_point, end_point, hit_pos, atol=dist_atol)

                if not has_los:
                    ray_data.append([hit_pos])
                    no_los_ray_ids.append(ray_idx)
                    ray_los_truth_table[ray_idx] = has_los

    n_rays_passed = len([ray for ray in ray_los_truth_table if ray == True])
    los_factor = n_rays_passed / n_rays

    return los_factor
