# This software module was originally developed by
#
# Nokia
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Nokia retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2024.


# encode binary mask
# using runlength + vertical runlength
# entropy coding:
#   make two conext:
#      positive numbers
#      signed numbers

import io
import struct
import numpy as np
import bmcrans
import os
from pathlib import Path


def get_cdfs():
    # prepare cdfs
    cdfs_width = 512
    cdfs_height = 2  # unsigned, signed
    cdfs = np.zeros([cdfs_height, cdfs_width], dtype=np.int32)
    offsets = np.zeros(cdfs_height, dtype=np.int32)
    cdf_sizes = np.zeros(cdfs_height, dtype=np.int32)
    abs_f_path = Path(os.path.abspath(__file__))
    path_parsed = list(abs_f_path.parts[0 : len(abs_f_path.parts) - 5])
    path = ""

    for str in path_parsed:
        path += str + os.sep

    cdfs_data_dir = f"{path}external{os.sep}mask_codec{os.sep}EntropyCodec{os.sep}cdfsdata{os.sep}"
    cdfs = np.loadtxt(cdfs_data_dir + "cdfs.csv", delimiter=",").astype(np.int32)
    offsets = np.loadtxt(cdfs_data_dir + "offsets.csv", delimiter=",").astype(np.int32)

    # set cdf_sizes
    for cdf_idx, cdf in enumerate(cdfs):
        cdf_sizes[cdf_idx] = np.where(cdf > 0)[0][-1] + 1

    __cdfs_data = (cdfs, offsets, cdf_sizes)

    return __cdfs_data


def encode_mask(mask):
    """
    Encode a mask image. Returns a bytes array. The first 4 bytes are width and height
    """
    # output symbol list
    # each entry is (length, stype)
    # stype: symbol type
    #   0: unsigned length
    #   1: signed length
    #
    output = []
    # assume the 0 row is a zero row
    width = mask.shape[1]
    height = mask.shape[0]
    prev_row = np.array([0] * width, dtype=np.uint8)
    row_ll = 0

    prev_start = 0

    for row in mask:
        # encode a row
        if np.array_equal(prev_row, row):
            # same as previous row
            row_ll += 1
        else:
            # different row
            output.append((row_ll, 0))

            # find the staring and ending compared to the previous row
            pix_diff = np.where(prev_row != row)[0]
            start_idx = pix_diff[0]
            end_idx = pix_diff[-1] + 1

            output.append((start_idx - prev_start, 1))
            output.append((end_idx - start_idx, 0))
            prev_start = start_idx

            # encode this row
            prev_v = row[start_idx]
            ll = 1
            for pix in row[start_idx + 1 : end_idx]:
                if pix == prev_v:
                    ll += 1
                else:
                    output.append((ll, 0))
                    prev_v = pix
                    ll = 1
            # ending of the row
            output.append((ll, 0))

            # reset previous row
            prev_row = row
            row_ll = 0

    output.append((row_ll, 0))

    # next entropy encoding output
    # entropy coding also has overhead, it is working only if the number of symbols is large > 300?

    # split symbols and indexes
    output_0 = np.array([x for x, stype in output if stype == 0])
    output_1 = np.array([x for x, stype in output if stype == 1])
    output, indexes = zip(*output)
    output = np.array(output)
    indexes = np.array(indexes)

    assert min(output_0) >= 0, print("all symbols shall be positive")

    # encode output

    __cdfs_data = get_cdfs()
    cdfs, offsets, cdf_sizes = __cdfs_data

    entropy_coder = bmcrans.RansEntropyCoder()
    symbols = output

    entropy_coder.encode_with_indexes(symbols, indexes, cdfs, cdf_sizes, offsets, oor_bypass=True)
    bitstream = entropy_coder.flush()

    buf = io.BytesIO()
    buf.write(struct.pack(">H", width))
    buf.write(struct.pack(">H", height))
    buf.write(bitstream)
    buf.seek(0)
    buf.getbuffer()
    return buf.getbuffer().tobytes()


def decode_mask(bitstream):
    """
    Decode a image mask from a bitstream
    The first 4 bytes in the bitstream are width and height
    """

    __cdfs_data = get_cdfs()
    cdfs, offsets, cdf_sizes = __cdfs_data

    width = struct.unpack(">H", bitstream[:2])[0]
    height = struct.unpack(">H", bitstream[2:4])[0]

    # decode bitstream
    entropy_decoder = bmcrans.RansEntropyCoder()
    entropy_decoder.set_stream(bitstream[4:])

    # get symbol
    def get_symbol(index):
        out = entropy_decoder.decode_stream([index], cdfs, cdf_sizes, offsets, oor_bypass=True)
        return out[0]

    # reconstruct the mask from the output
    rec = np.zeros((height, width), dtype=np.uint8)
    prev_row = np.zeros(width, dtype=np.uint8)
    row_idx = 0

    prev_start = 0
    prev_end = 0

    v_idx = 0
    while True:
        row_ll = get_symbol(0)

        for row in range(row_ll):
            rec[row_idx] = prev_row
            row_idx += 1

        if row_idx >= height:
            break

        # reconstruct a row
        row = np.zeros(width, dtype=np.uint8)
        start_idx = prev_start + get_symbol(1)
        prev_start = start_idx
        seg_len = get_symbol(0)
        end_idx = start_idx + seg_len

        # repeat the begining of the previous row
        pix_idx = start_idx
        row[:start_idx] = prev_row[:start_idx]
        prev_v = prev_row[start_idx]

        while pix_idx < end_idx:
            ll = get_symbol(0)
            row[pix_idx : pix_idx + ll] = 1 - prev_v
            prev_v = 1 - prev_v
            pix_idx += ll
        # fill row ending
        row[end_idx:] = prev_row[end_idx:]

        rec[row_idx] = row
        row_idx += 1
        if row_idx >= height:
            break

        prev_row = row

    return rec
