# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2024.

from typing import List
from ast import Return
from mpegi.geometry.transforms import applyTransformMatrix4x4
from mpegi.metadata.eif import Mesh, Box
from mpegi.metadata.eifplus import Portal
import numpy as np
import trimesh
import trimesh.creation
from mpegi.geometry.trimesh import _convertMeshToTrimesh, _convertBoxToTrimesh


class MultiroomCalculator:
    def __init__(self):
        Return

    def rotate_vector_to(self, v1: List[float], v2: List[float]):

        v1 /= np.sqrt(sum(v1 ** 2))
        v2 /= np.sqrt(sum(v2 ** 2))
        axis = np.cross(v1, v2)
        cos_a = np.dot(v1, v2)
        if cos_a < -0.9999:
            k = 0
        else:
            k = 1 / (1 + cos_a)

        a_x = axis[0]
        a_y = axis[1]
        a_z = axis[2]

        x = [(a_x * a_y * k) + cos_a, (a_x * a_y * k) + a_z, (a_x * a_z * k) - a_y]
        y = [(a_y * a_x * k) - a_z, (a_y * a_y * k) + cos_a, (a_y * a_z * k) + a_x]
        z = [(a_z * a_x * k) + a_y, (a_z * a_y * k) - a_x, (a_z * a_z * k) + cos_a]
        o = [0, 0, 0]

        return self.homogen_mat(x, y, z, o)

    def homogen_mat(self, x, y, z, o):
        return np.asarray(
            [x[0], y[0], z[0], o[0], x[1], y[1], z[1], o[1], x[2], y[2], z[2], o[2], 0, 0, 0, 1]
        ).reshape((4, 4))

    def max_with_sign(self, x: float, y: float):
        y_abs = abs(y)
        return min(x, -y_abs) if x < 0 else max(x, y_abs)

    def calculateSourceContribution(
        self,
        portalPos: List[float],
        portalWidth: float,
        portalHeight: float,
        portalNormal: List[float],
        sourcePos: List[float],
    ):
        eps = 1e-9
        source_position = np.array(sourcePos)
        portal_position = np.array(portalPos)
        portal_normal = np.array(portalNormal)
        source_portal_vector = source_position - portal_position
        target_vector = np.array([0.0, 0.0, -1.0])
        rot_matrix = self.rotate_vector_to(portal_normal, target_vector)

        portal_normal = applyTransformMatrix4x4(rot_matrix, portal_normal)
        if not (abs(portal_normal - target_vector)).all() < 1e-6:
            raise Exception("Error in portal normal - target vec")

        src_portal_vec = applyTransformMatrix4x4(rot_matrix, source_portal_vector)

        src_to_portal_dist_azi = max(
            np.sqrt(src_portal_vec[0] * src_portal_vec[0] + src_portal_vec[2] * src_portal_vec[2]),
            eps,
        )
        src_to_portal_dist_ele = max(
            np.sqrt(src_portal_vec[1] * src_portal_vec[1] + src_portal_vec[2] * src_portal_vec[2]),
            eps,
        )

        # Calculating cosine of the angle between the portal's normal and the source-portal vector in the horizontal plane.
        num_a = src_portal_vec[0] * portal_normal[0] + src_portal_vec[2] * portal_normal[2]
        den_a = src_to_portal_dist_azi * np.sqrt(
            portal_normal[0] * portal_normal[0] + portal_normal[2] * portal_normal[2]
        )
        cos_phi_a = np.copysign(max(abs(num_a), eps), num_a) / max(den_a, eps)

        # Calculating cosine of the purely vertical angle between the portal's normal and the source-portal vector.
        num_e = src_portal_vec[1] * portal_normal[1] + src_portal_vec[2] * portal_normal[2]
        den_e = src_to_portal_dist_ele * np.sqrt(
            portal_normal[1] * portal_normal[1] + portal_normal[2] * portal_normal[2]
        )
        cos_phi_e = np.copysign(max(abs(num_e), eps), num_e) / max(den_e, eps)

        # Ensure that acos(cosPhiA) and acos(cosPhiE) are in a range of 0..1/2*pi (0 .. 90 degrees).
        # TODO: Check as 0 <= cosPhi <= 1, because acos(eps) / M_PI_2 seems to be larger than 0.5.
        # Otherwise they would be outside the environment.
        cos_phi_a = min(max(cos_phi_a, 0), 1)
        cos_phi_e = min(max(cos_phi_e, 0), 1)

        # Calculate azimuthal and elevational angle ranges that the portal covers on a circle centered on
        # the source position and going through the portal position.

        w = portalWidth
        h = portalHeight

        theta_a = np.arctan(
            src_to_portal_dist_azi
            * w
            / self.max_with_sign(
                src_to_portal_dist_azi * src_to_portal_dist_azi - 0.25 * w * w, eps
            )
            * max(cos_phi_a, eps)
        )
        if theta_a < 0:
            theta_a += np.pi

        theta_e = np.arctan(
            src_to_portal_dist_ele
            * h
            / self.max_with_sign(
                src_to_portal_dist_ele * src_to_portal_dist_ele - 0.25 * h * h, eps
            )
            * max(cos_phi_e, eps)
        )
        if theta_e < 0:
            theta_e += np.pi

        # Calculate source energy contribution
        portal_contribution = theta_a * np.sin(0.5 * theta_e) / (2 * np.pi)

        return portal_contribution

    def getPortalProperties(self, portal: Portal, geometryById: dict):
        if portal.portalPosition is None or portal.portalSize is None:
            extent = geometryById[portal.portalExtent]

            if isinstance(extent, Mesh):
                portal_extent_mesh = _convertMeshToTrimesh(extent, includeTransparent=True)
                portal.portalPosition = portal_extent_mesh.bounding_box.centroid + extent.position
            elif isinstance(extent, Box):
                portal_extent_mesh = _convertBoxToTrimesh(extent)
                portal_extent_mesh.apply_transform(
                    trimesh.transformations.translation_matrix(direction=extent.position)
                )

                portal.portalPosition = portal_extent_mesh.bounding_box.centroid
            R = self.rotate_vector_to(portal.portalNormal, np.array([0.0, 0.0, -1.0]))
            portal_extent_mesh.apply_transform(R)
            portal.portalSize = (
                portal_extent_mesh.bounding_box.bounds[1]
                - portal_extent_mesh.bounding_box.bounds[0]
            )
        portal.portalWidth = portal.portalSize[0]
        portal.portalHeight = portal.portalSize[1]

        portal.portalCorners = self.getPortalCorners(portal)

        return portal

    def getPortalCorners(self, portal: Portal):
        R = np.array(
            [
                [0.0, 0.0, 1.0],
                [0.0, 1.0, 0.0],
                [-1.0, 0.0, 0.0],
            ]
        )

        portalWidthVec = np.matmul(portal.portalNormal, R)
        portalCorners = [None] * 4
        portalCorners[0] = (
            portal.portalPosition
            + (portalWidthVec * (portal.portalWidth / 2))
            + (np.array([0.0, 1.0, 0.0]) * portal.portalHeight / 2)
        )
        portalCorners[1] = (
            portal.portalPosition
            - (portalWidthVec * (portal.portalWidth / 2))
            + (np.array([0.0, 1.0, 0.0]) * portal.portalHeight / 2)
        )
        portalCorners[2] = (
            portal.portalPosition
            + (portalWidthVec * (portal.portalWidth / 2))
            - (np.array([0.0, 1.0, 0.0]) * portal.portalHeight / 2)
        )
        portalCorners[3] = (
            portal.portalPosition
            - (portalWidthVec * (portal.portalWidth / 2))
            - (np.array([0.0, 1.0, 0.0]) * portal.portalHeight / 2)
        )
        return portalCorners

    def calculatePortalToPortalFactor(self, fromPortal: Portal, toPortal: Portal):
        normalDirection = 1
        sc = np.zeros(5)

        sc[0] = self.calculateSourceContribution(
            toPortal.portalPosition,
            toPortal.portalWidth,
            toPortal.portalHeight,
            toPortal.portalNormal,
            fromPortal.portalPosition,
        )

        altSc = self.calculateSourceContribution(
            toPortal.portalPosition,
            toPortal.portalWidth,
            toPortal.portalHeight,
            -1 * toPortal.portalNormal,
            fromPortal.portalPosition,
        )

        if altSc > sc[0]:
            sc[0] = altSc
            normalDirection = -1

        for cornerIdx in range(0, 4):
            sourcePos = fromPortal.portalCorners[cornerIdx]
            sc[cornerIdx + 1] = self.calculateSourceContribution(
                toPortal.portalPosition,
                toPortal.portalWidth,
                toPortal.portalHeight,
                normalDirection * toPortal.portalNormal,
                sourcePos,
            )

        portal2portalFactor = sum(sc * [4, 1, 1, 1, 1]) / 4

        return portal2portalFactor

    def getPortal2PortalFactorMatrix(self, portals: List[Portal], geometryById: dict):

        # check that all portals have been pre-processed and we have the corner data
        for portal in portals:
            if not hasattr(portal, "portalCorners"):
                portal = self.getPortalProperties(portal, geometryById)
        nrPortals = len(portals)
        portal2portalFactors = np.full((nrPortals, nrPortals), None)

        for i in range(0, nrPortals):
            for j in range(0, nrPortals):
                if i != j:
                    if (
                        (
                            portals[i].acousticEnvironment1 == portals[j].acousticEnvironment1
                            and portals[i].acousticEnvironment2 != portals[j].acousticEnvironment2
                        )
                        or (
                            portals[i].acousticEnvironment2 == portals[j].acousticEnvironment2
                            and portals[i].acousticEnvironment1 != portals[j].acousticEnvironment1
                        )
                        or (
                            portals[i].acousticEnvironment1 == portals[j].acousticEnvironment2
                            and portals[i].acousticEnvironment2 != portals[j].acousticEnvironment1
                        )
                        or (
                            portals[i].acousticEnvironment2 == portals[j].acousticEnvironment1
                            and portals[i].acousticEnvironment1 != portals[j].acousticEnvironment2
                        )
                    ):
                        portal2portalFactors[i][j] = self.calculatePortalToPortalFactor(
                            portals[i], portals[j]
                        )
        return portal2portalFactors
