# This software module was originally developed by
#
# Nokia
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Nokia retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2024.

import logging
from typing import List

import mpegi.metadata.reverb.constants as constants

if constants.debug_enable_visualization:
    import matplotlib.pyplot as plt

from multiprocessing import Pool
from os import cpu_count
import numpy as np
from trimesh import Trimesh, creation, transformations, PointCloud
from mpegi.metadata.reverb.los_helpers import get_los_factor_p2p_multi
from mpegi.geometry.trimesh import _convertMeshToTrimesh
from mpegi.metadata.eif import Mesh, Portal
from mpegi.metadata.eifplus import AcousticEnvironment

from ..reverb.reverbutils import (
    is_point_inside_box,
    calculate_centroid,
    distance_3d,
    get_3d_plot,
    get_portal_geom,
    get_portal_open_faces,
    ae_geom_as_trimesh,
    get_outside_aes,
    get_ae_portals,
    get_outside_portals,
    get_diagonal_corners,
    set_box_height,
    expand_box_horizontal,
    populate_box_positions,
    get_meshes_in_box,
    get_trimesh_faces_normals_areas,
    determine_floor_face,
    arr_to_nd_bin_mask,
    nd_bin_mask_to_arr,
)
from mpegi.metadata.reverb.mask_codec import encode_mask, decode_mask


class PortalLosAnalyzer:
    def __init__(
        self,
        explicitPortals: {},
        meshes: List[Mesh],
        geometry_by_id,
        acoustic_environments: List[AcousticEnvironment],
        eif_scene: str,
        isVoxelScene: bool,
        eifFilePath: str,
    ):

        logging.getLogger("trimesh").setLevel(level=logging.CRITICAL)
        self.explicit_portals = explicitPortals
        self.geometry_by_id = geometry_by_id
        self.acoustic_environments = acoustic_environments
        self.ae_verts = []
        self.eif_scene = eif_scene
        self.ae_bounding_geoms = {}
        self.ae_geoms = {}
        self.portal_open_face_map = {}
        self.portal_open_face_centroids = {}
        self.meshes_as_trimeshes = {}
        self.portal_geoms = {}
        self.is_voxel_scene = isVoxelScene

        for portal in self.explicit_portals:
            portal_geom = get_portal_geom(portal, self.geometry_by_id[portal.portalExtent])
            p_open_faces = get_portal_open_faces(portal, self.geometry_by_id[portal.portalExtent])
            self.portal_geoms[portal.id] = portal_geom
            self.portal_open_face_map[portal.id] = p_open_faces

            p_of_1 = p_open_faces[0]
            p_of_2 = p_open_faces[1]

            p_c1 = calculate_centroid(p_of_1)
            p_c2 = calculate_centroid(p_of_2)
            self.portal_open_face_centroids[portal.id] = [p_c1, p_c2]

        for ae in self.acoustic_environments:
            ae_geometry = self.geometry_by_id[ae.region]
            ae_geom, ae_b_geom = ae_geom_as_trimesh(ae_geometry)

            self.ae_geoms[ae.id] = ae_geom
            self.ae_bounding_geoms[ae.id] = ae_b_geom

        if self.is_voxel_scene:
            logging.info("Converting voxel scene into mesh for LoS analysis")
            self.voxel_data = self.voxel_scene_as_mesh(eifFilePath)
            for subscene_id, subscene in self.voxel_data.items():
                vox_idx = 0
                for vox_block in subscene:
                    self.meshes_as_trimeshes[subscene_id + "_" + str(vox_idx)] = vox_block
                    vox_idx += 1

        else:
            for mesh in meshes:
                mesh_as_trimesh = _convertMeshToTrimesh(mesh, includeTransparent=True)
                mesh_as_trimesh.apply_transform(
                    transformations.translation_matrix(direction=mesh.position)
                )
                self.meshes_as_trimeshes[mesh.id] = mesh_as_trimesh

        logging.getLogger("trimesh").setLevel(level=logging.INFO)

    def vox_subscene_as_mesh(
        self, subscene, vox_idx_start: List[float], vox_idx_end: List[float]
    ) -> List[float]:
        corner_p1: List[float] = [float(val) for val in subscene.attrib["position_s"].split()]
        corner_p2: List[float] = [float(val) for val in subscene.attrib["position_e"].split()]
        corner_p1 = [corner_p1[1], corner_p1[2], corner_p1[0]]
        corner_p2 = [corner_p2[1], corner_p2[2], corner_p2[0]]

        vox_res: List[int] = [int(val) for val in subscene.attrib["vox_resolution"].split()]
        vox_res = [vox_res[1], vox_res[2], vox_res[0]]

        voxel_size: List[float] = [
            (corner_p2[0] - corner_p1[0]) / vox_res[0],
            (corner_p2[1] - corner_p1[1]) / vox_res[1],
            (corner_p2[2] - corner_p1[2]) / vox_res[2],
        ]

        x, y, z = vox_idx_start
        x1, y1, z1 = vox_idx_end
        xd, yd, zd = [x1 - x, y1 - y, z1 - z]

        pos_s = [
            corner_p1[0] + (voxel_size[0] * x) - (voxel_size[0] / 2),
            corner_p1[1] + (voxel_size[1] * y) - (voxel_size[1] / 2),
            corner_p1[2] + (voxel_size[2] * z) - (voxel_size[2] / 2),
        ]

        z_val = pos_s[2]
        x_val = pos_s[0]
        y_val = pos_s[1]

        trimesh_vox = []
        for i in range(yd + 1):
            for j in range(zd + 1):
                for k in range(xd + 1):
                    pos = [x_val, y_val, z_val]
                    trimesh_vox.append(
                        creation.box(voxel_size, transform=transformations.translation_matrix(pos))
                    )
                    x_val += voxel_size[0]
                z_val += voxel_size[2]
                x_val = pos_s[0]

            y_val += voxel_size[1]
            z_val = pos_s[2]
            x_val = pos_s[0]

        verts = []
        for block in trimesh_vox:
            for v in block.vertices:
                verts.append([v[0], v[1], v[2]])

        # Fuse individual blocks into one mesh for performance
        return PointCloud(verts).bounding_box

    def voxel_scene_as_mesh(self, eifFilePath: str):
        import xml.etree.ElementTree as ET

        root = ET.parse(eifFilePath).getroot()
        voxel_subscenes = root.findall("VoxelScene")

        if not voxel_subscenes:
            raise Exception("no subscenes")

        subscenes = {}
        for subscene in voxel_subscenes:
            subscene_id = subscene.attrib["id"]
            voxel_blocks = subscene.findall("VoxelBlock")
            trimesh_vox_scene = []
            for v_b in voxel_blocks:
                block_coord_s: List[int] = [
                    int(val) for val in v_b.attrib["grid_coordinate_s"].split()
                ]
                block_coord_e: List[int] = [
                    int(val) for val in v_b.attrib["grid_coordinate_e"].split()
                ]
                block_coord_s = [block_coord_s[1], block_coord_s[2], block_coord_s[0]]
                block_coord_e = [block_coord_e[1], block_coord_e[2], block_coord_e[0]]

                vox_block_mesh = self.vox_subscene_as_mesh(subscene, block_coord_s, block_coord_e)
                trimesh_vox_scene.append(vox_block_mesh)
            subscenes[subscene_id] = trimesh_vox_scene

        return subscenes

    def decode_3d_bin_mask(self, bitmask_enc_3d):
        bm_dec = []
        for mask_2d in bitmask_enc_3d:
            bm_dec.append(decode_mask(mask_2d))
        return bm_dec

    def get_ae_los_voxel_data(
        self,
        los_voxel_size_m,
        los_area_height_m: float,
        dist_atol: float,
        ae_mesh_ins_prc: int,
        los_ratio_thr: float,
        multiprocess: bool,
    ):

        logging.info(
            f"pla: Generating AE los-voxel-grid for {len(self.acoustic_environments)} AEs with parameters:\nvox_side_len: {los_voxel_size_m}m\n{los_area_height_m=}\n{ae_mesh_ins_prc=}\n{los_ratio_thr=}\n{multiprocess=}"
        )
        los_voxel_grid = {}
        if multiprocess:
            pool = Pool(cpu_count() - 2)
            los_vox_args = [
                (
                    ae.id,
                    los_voxel_size_m,
                    los_area_height_m,
                    dist_atol,
                    ae_mesh_ins_prc,
                    los_ratio_thr,
                )
                for ae in self.acoustic_environments
            ]
            data = pool.starmap(self.get_los_voxel_data_ae, los_vox_args)
            for ae in self.acoustic_environments:
                for data_set in data:
                    data_ae = data_set["ae_id"]
                    if data_ae == ae.id:
                        los_voxel_grid[ae.id] = data_set

        else:
            for ae in self.acoustic_environments:
                data = self.get_los_voxel_data_ae(
                    ae.id,
                    los_voxel_size_m,
                    los_area_height_m,
                    dist_atol,
                    ae_mesh_ins_prc,
                    los_ratio_thr,
                )
                los_voxel_grid[ae.id] = data

        # Initialize voxel grid as member variable
        self.los_voxel_grid = los_voxel_grid

        return los_voxel_grid

    def los_grid_visualize(self, voxels_to_visualize: str):
        for ae_id, los_data in self.los_voxel_grid.items():
            los_voxels = los_data["portal_los_voxels"]
            n_voxels_x, n_voxels_y, n_voxels_z = los_data["ae_n_voxels"]
            los_area_corners = los_data["ae_corner_points"]
            los_voxel_size_m = los_data["los_voxel_side_len_m"]
            n_voxels_total = n_voxels_z * n_voxels_y * n_voxels_x
            ae_meshes = get_meshes_in_box(
                self.ae_bounding_geoms[ae_id],
                self.meshes_as_trimeshes,
                constants.pla_ae_mesh_inside_percentage,
            )
            ae_portals = los_data["ae_portals"]

            for p_idx in range(len(ae_portals)):
                ae_p = ae_portals[p_idx]
                ae_portal = ae_portals[p_idx]
                ax, fig = get_3d_plot(True)
                portal_geom = self.portal_geoms[ae_portal.id]
                ax.plot_trisurf(
                    portal_geom.vertices[:, 0],
                    portal_geom.vertices[:, 2],
                    portal_geom.vertices[:, 1],
                    alpha=0.3,
                    color="blue",
                    triangles=portal_geom.faces,
                )

                for mesh_id, mesh in ae_meshes:
                    ax.plot_trisurf(
                        mesh.vertices[:, 0],
                        mesh.vertices[:, 2],
                        mesh.vertices[:, 1],
                        alpha=0.3,
                        color="black",
                        triangles=mesh.faces,
                    )

                nolos_vox = []
                los_vox = []
                for vox_idx in range(n_voxels_total):
                    portal_los_truth = los_voxels[ae_p.id][vox_idx]
                    vox_pos = self._los_voxel_coord_by_idx(
                        vox_idx,
                        los_area_corners,
                        n_voxels_x,
                        n_voxels_y,
                        n_voxels_z,
                        los_voxel_size_m,
                    )
                    vox = creation.box(
                        extents=[los_voxel_size_m, los_voxel_size_m, los_voxel_size_m],
                        transform=transformations.translation_matrix(
                            [vox_pos[0], vox_pos[1], vox_pos[2]]
                        ),
                    )

                    if portal_los_truth:
                        los_vox.append(vox)
                    else:
                        nolos_vox.append(vox)

                if voxels_to_visualize == "nolos":
                    for vox in nolos_vox:
                        ax.plot_trisurf(
                            vox.vertices[:, 0],
                            vox.vertices[:, 2],
                            vox.vertices[:, 1],
                            alpha=0.3,
                            color="red",
                            triangles=vox.faces,
                        )

                elif voxels_to_visualize == "los":
                    for vox in los_vox:
                        ax.plot_trisurf(
                            vox.vertices[:, 0],
                            vox.vertices[:, 2],
                            vox.vertices[:, 1],
                            alpha=0.3,
                            color="green",
                            triangles=vox.faces,
                        )

                elif voxels_to_visualize == "all":
                    for vox in [*los_vox, *nolos_vox]:
                        if vox in nolos_vox:
                            color = "red"
                        elif vox in los_vox:
                            color = "green"
                        ax.plot_trisurf(
                            vox.vertices[:, 0],
                            vox.vertices[:, 2],
                            vox.vertices[:, 1],
                            alpha=0.3,
                            color=color,
                            triangles=vox.faces,
                        )
                else:
                    raise Exception(f"Available los plot parameters: 'nolos', 'los', 'all'")

                plt.show()

    def rle_encode_nd_bin_arr(self, bin_arr: List[int], n_height, n_col, n_row):
        masks = arr_to_nd_bin_mask(bin_arr, n_height, n_col, n_row)
        masks_enc = [None] * len(masks)
        for i in range(len(masks)):
            mask = masks[i]
            masks_enc[i] = encode_mask(mask)

        return np.asarray(masks_enc)

    def get_outside_meshes(self, return_meshes: bool = False):
        outside_meshes = []
        for mesh_id, mesh in self.meshes_as_trimeshes.items():
            is_inside_truths = []
            for ae_id, ae_b_geom in self.ae_bounding_geoms.items():
                for v in mesh.vertices:
                    is_inside_truths.append(is_point_inside_box(ae_b_geom.vertices, v))

            if not any(is_inside_truths):
                outside_meshes.append([mesh_id, mesh])

        outside_meshes_only = []
        if return_meshes:
            for mesh_id, mesh in outside_meshes:
                outside_meshes_only.append(mesh)
            outside_meshes = outside_meshes_only

        return outside_meshes

    def get_portal_los_truths(
        self,
        portal_or_idx: Portal,
        interp_resol: float,
        mesh_to_check: Trimesh,
        meshes_to_raycast: List[Trimesh],
        dist_atol: float,
        ae_id: str,
        as_los_ratios: bool,
        los_ratio_thr: float,
    ):
        if isinstance(portal_or_idx, int):
            portal_or_idx = self.explicit_portals[portal_or_idx]

        mesh_populated = populate_box_positions(interp_resol, mesh_to_check)
        los_ratios = self.get_portal_los_factors(
            mesh_populated, meshes_to_raycast, portal_or_idx, dist_atol, ae_id
        )
        portal_los_truths = []
        for l_r in los_ratios:
            if as_los_ratios:
                portal_los_truths.append(l_r)
            else:
                has_los = l_r >= los_ratio_thr
                portal_los_truths.append(has_los)

        return portal_los_truths

    def get_outside_los_data(
        self,
        interp_resol: float,
        los_area_height_m: float,
        dist_atol: float,
        los_ratio_thr: float,
        outside_area_extension_m: float = 10,
    ):
        logging.info(
            f"pla: Generating Outside los-voxel-grid with parameters:\nvox_side_len: {interp_resol}m\n"
        )

        outside_aes = get_outside_aes(self.explicit_portals)

        if len(outside_aes) == 0:
            logging.info(f"{self.eif_scene} has no outside AE's")
            return None

        outside_ae_los_data = {}
        for outside_ae in outside_aes:

            # Expand and limit height
            outside_bounds = self.expand_ae_bounds(outside_ae, outside_area_extension_m)
            outside_bounds = set_box_height(outside_bounds, los_area_height_m)

            outside_portals = get_outside_portals(self.explicit_portals)

            meshes_to_raycast = []

            # Raycast only outside meshes and other ae b geoms
            outside_meshes = self.get_outside_meshes()

            for mesh_id, o_mesh in outside_meshes:
                meshes_to_raycast.append([mesh_id, o_mesh])
            for ae_id, ae_geom in self.ae_geoms.items():
                meshes_to_raycast.append([ae_id, ae_geom])

            # interp_resol = vox side len
            n_voxels_x, n_voxels_y, n_voxels_z = self.get_n_voxels_by_extents(
                outside_bounds, interp_resol
            )

            portal_bitmasks_rle_encoded = {}
            portal_los_truths = {}
            for o_portal in outside_portals:
                portal_los_data = self.get_portal_los_truths(
                    o_portal,
                    interp_resol,
                    outside_bounds,
                    meshes_to_raycast,
                    dist_atol=dist_atol,
                    ae_id=None,
                    as_los_ratios=False,
                    los_ratio_thr=los_ratio_thr,
                )
                portal_los_binary_mask = arr_to_nd_bin_mask(
                    portal_los_data, n_voxels_y, n_voxels_z, n_voxels_x
                )

                # Sanity checks
                portal_los_binary_mask_flattened = nd_bin_mask_to_arr(portal_los_binary_mask)
                if not np.array_equal(portal_los_data, portal_los_binary_mask_flattened):
                    raise Exception("Flattened los bin arr not same as original")

                for mask in portal_los_binary_mask:
                    mask_enc = encode_mask(mask)
                    mask_dec = decode_mask(mask_enc)
                    arrays_equal = np.array_equal(mask_dec, mask)
                    if not arrays_equal:
                        raise Exception("Bitmasks not equal")

                portal_bitmasks_rle_encoded[o_portal.id] = self.rle_encode_nd_bin_arr(
                    portal_los_data, n_voxels_y, n_voxels_z, n_voxels_x
                )
                portal_los_truths[o_portal.id] = portal_los_data

            # Get diagonal corner points from bounds
            diag_points = get_diagonal_corners(outside_bounds)

            data = {
                "portal_los_voxels_rle_encoded": portal_bitmasks_rle_encoded,
                "portal_los_voxels": portal_los_truths,
                "n_voxels": [n_voxels_x, n_voxels_y, n_voxels_z],
                "area_portals": outside_portals,
                "diag_corner_points": diag_points,
                "los_voxel_side_len_cm": int(interp_resol * 100),
            }

            outside_ae_los_data[outside_ae] = data

        self.outside_los_voxel_grid = outside_ae_los_data

        return outside_ae_los_data

    def _los_voxel_coord_by_idx(
        self,
        voxel_idx: int,
        corner_points: List[float],
        n_voxels_x,
        n_voxels_y,
        n_voxels_z,
        vox_size_m,
    ):
        x = corner_points[0][0] + (voxel_idx % n_voxels_x) * vox_size_m
        y = corner_points[0][1] - int(voxel_idx / (n_voxels_x * n_voxels_z)) * vox_size_m
        z = corner_points[0][2] + int(voxel_idx / n_voxels_x) % n_voxels_z * vox_size_m
        pos = [x, y, z]

        return pos

    def expand_ae_bounds(self, ae_id: str, expansion_m: float):

        bounding_geom = self.ae_bounding_geoms[ae_id]
        bg_c = bounding_geom.centroid

        l, h, d = bounding_geom.extents
        ext_expanded = [l + expansion_m * 2, h, d + expansion_m * 2]

        bounds_expanded = creation.box(
            extents=ext_expanded, transform=transformations.translation_matrix(direction=bg_c)
        )

        return bounds_expanded

    def get_n_voxels_by_extents(self, extents_or_trimesh, vox_size_m: float):
        if isinstance(extents_or_trimesh, Trimesh):
            extents_or_trimesh = extents_or_trimesh.extents
        return [
            int(extents_or_trimesh[0] / vox_size_m) + 1,
            int(extents_or_trimesh[1] / vox_size_m) + 1,
            int(extents_or_trimesh[2] / vox_size_m) + 1,
        ]

    def get_los_voxel_data_ae(
        self,
        ae_id: str,
        interp_resol,
        los_area_height_m: float,
        dist_atol: float,
        ae_mesh_ins_prc: int,
        los_ratio_thr: float,
    ):
        logging.info(f"Analyzing LoS-grid for {ae_id}")
        ae_b_geom = self.ae_bounding_geoms[ae_id]
        ae_inside_meshes = get_meshes_in_box(ae_b_geom, self.meshes_as_trimeshes, ae_mesh_ins_prc)
        ae_portals = get_ae_portals(ae_id, self.explicit_portals)

        if los_area_height_m > 0:
            ae_b_geom = set_box_height(ae_b_geom, los_area_height_m)

        ae_b_geom_widened = expand_box_horizontal(ae_b_geom, 1, 1)
        ae_b_geom = ae_b_geom_widened

        ae_b_geom_populated = populate_box_positions(interp_resol, ae_b_geom)
        n_voxels = len(ae_b_geom_populated)
        n_voxels_x, n_voxels_y, n_voxels_z = self.get_n_voxels_by_extents(ae_b_geom, interp_resol)

        ae_portal_ids = []
        portal_bitmasks_rle_encoded = {}
        portal_los_truths = {}
        for portal in ae_portals:
            p_id = portal.id
            ae_portal_ids.append(p_id)
            portal_los_data = self.get_portal_los_truths(
                portal,
                interp_resol,
                ae_b_geom,
                ae_inside_meshes,
                dist_atol,
                ae_id,
                False,
                los_ratio_thr,
            )

            portal_los_binary_mask = arr_to_nd_bin_mask(
                portal_los_data, n_voxels_y, n_voxels_z, n_voxels_x
            )

            # Sanity checks
            portal_los_binary_mask_flattened = nd_bin_mask_to_arr(portal_los_binary_mask)
            if not np.array_equal(portal_los_data, portal_los_binary_mask_flattened):
                raise Exception("Flattened los bin arr not same as original")

            for mask in portal_los_binary_mask:
                mask_enc = encode_mask(mask)
                mask_dec = decode_mask(mask_enc)
                arrays_equal = np.array_equal(mask_dec, mask)
                if not arrays_equal:
                    raise Exception("Bitmasks not equal")

            portal_bitmasks_rle_encoded[portal.id] = self.rle_encode_nd_bin_arr(
                portal_los_data, n_voxels_y, n_voxels_z, n_voxels_x
            )
            portal_los_truths[portal.id] = portal_los_data

        corner_p1, corner_p2 = get_diagonal_corners(ae_b_geom)
        n_voxels_x, n_voxels_y, n_vozels_z = self.get_n_voxels_by_extents(ae_b_geom, interp_resol)

        los_data = {
            "ae_id": ae_id,
            "portal_los_voxels": portal_los_truths,
            "portal_los_voxels_rle_encoded": portal_bitmasks_rle_encoded,
            "ae_n_voxels": [n_voxels_x, n_voxels_y, n_vozels_z],
            "ae_corner_points": [corner_p1, corner_p2],
            "ae_portals": ae_portals,
            "los_voxel_side_len_cm": int(interp_resol * 100),
            "los_voxel_side_len_m": interp_resol,
            "n_voxels_total": n_voxels,
        }

        return los_data

    def _get_portal_open_face_by_pos(self, pos: List[float], portal: Portal):
        p_o_f1, p_o_f2 = self.portal_open_face_map[portal.id]
        p_o_f1_c, p_o_f2_c = (
            self.portal_open_face_centroids[portal.id][0],
            self.portal_open_face_centroids[portal.id][1],
        )

        d1, d2 = distance_3d(
            pos[0], pos[1], pos[2], p_o_f1_c[0], p_o_f1_c[1], p_o_f1_c[2]
        ), distance_3d(pos[0], pos[1], pos[2], p_o_f2_c[0], p_o_f2_c[1], p_o_f2_c[2])

        return p_o_f1 if d1 < d2 else p_o_f2

    def get_portal_los_factors(
        self,
        positions: List[float],
        meshes: List[Trimesh],
        portal: Portal,
        dist_atol: float,
        ae_id: str,
    ):
        portal_faces = []
        tol = 0.05
        try:
            ae_mesh = self.ae_geoms[ae_id]
            ae_geom_f, _, __ = get_trimesh_faces_normals_areas(ae_mesh)
            ae_floor = determine_floor_face(ae_geom_f)
            ae_floor_y_vals = [v[1] for v in ae_floor]
            ae_floor_min_y = min(ae_floor_y_vals)

            portal_geom = self.portal_geoms[portal.id]

            p_geom_f, _, __ = get_trimesh_faces_normals_areas(portal_geom)
            p_floor = determine_floor_face(p_geom_f)
            p_floor_y_vals = [v[1] for v in p_floor]
            p_floor_min_y = min(p_floor_y_vals)

            is_portal_under_ae = p_floor_min_y < ae_floor_min_y

        except KeyError:
            is_portal_under_ae = False

        for pos in positions:
            closest_p_face = self._get_portal_open_face_by_pos(pos, portal)
            # If portal is under ae floor, elevate all vertices to ae floor level + tolerance
            if is_portal_under_ae:
                closest_p_face_new = []
                for v in closest_p_face:
                    if v[1] < ae_floor_min_y:
                        closest_p_face_new.append([v[0], ae_floor_min_y + tol, v[2]])
                    else:
                        closest_p_face_new.append(v)
                closest_p_face = closest_p_face_new

            # Due to voxel scene conversion inaccuracies, sometimes portal corner vertices are partly occluded
            # Translate corner vertices towards center for tolerance
            if self.is_voxel_scene:
                c = calculate_centroid(closest_p_face)
                f_new = []
                for v in closest_p_face:
                    len = distance_3d(v[0], v[1], v[2], c[0], c[1], c[2]) / 4
                    dir = [c[0] - v[0], c[1] - v[1], c[2] - v[2]]
                    dir_norm = dir / np.sqrt(np.sum([dir[0] ** 2, dir[1] ** 2, dir[2] ** 2]))
                    v1 = [
                        v[0] + len * dir_norm[0],
                        v[1] + len * dir_norm[1],
                        v[2] + len * dir_norm[2],
                    ]
                    f_new.append(v1)
                closest_p_face = f_new
            portal_faces.append(closest_p_face)  # Target vertices

        return get_los_factor_p2p_multi(positions, meshes, portal_faces, dist_atol)
