# This software module was originally developed by
#
# Nokia
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Nokia retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


import math
from ast import Return
from dataclasses import dataclass
from typing import List

import mpegi.metadata.reverb.constants as constants
import mpegi.metadata.reverb.dsrHelpers as dsrHelpers
from mpegi.metadata.eif import Box
from mpegi.metadata.eif.scene import AudioScene
from mpegi.metadata.eifplus import AcousticEnvironment, Mesh
from mpegi.metadata.eifplus.scene import SceneType
from mpegi.metadata.reverb.portal_los_analyzer import PortalLosAnalyzer


class Reverberator:
    def __init__(self, defaultAE):
        self.defaultAE = defaultAE
        Return

    def createReverbParameters(
        self,
        acousticEnvironments: List[AcousticEnvironment],
        eifScene: AudioScene,
        revRendererControl: list,
        explicitPortals: {},
        meshes: List[Mesh],
        geometries,
        sceneType,
        disableLos: bool,
        isVoxelScene: bool,
        eifFilePath: str,
    ):
        plaAeLosVoxelData = None
        plaOutsideLosVoxelData = None

        if not disableLos:
            hasPortals = len(explicitPortals) > 0
            if hasPortals:
                pla = PortalLosAnalyzer(
                    explicitPortals,
                    meshes,
                    geometries,
                    acousticEnvironments,
                    eifScene.id,
                    isVoxelScene,
                    eifFilePath,
                )

                plaAeLosVoxelData = pla.get_ae_los_voxel_data(
                    constants.pla_los_area_voxel_side_cm / 100,
                    constants.pla_los_area_grid_height_m,
                    constants.los_check_dist_atol,
                    constants.pla_ae_mesh_inside_percentage,
                    constants.pla_los_ratio_threshold,
                    True,
                )

                plaOutsideLosVoxelData = pla.get_outside_los_data(
                    constants.pla_outside_los_area_voxel_side_cm / 100,
                    constants.pla_los_area_grid_height_m,
                    constants.los_check_dist_atol,
                    constants.pla_los_ratio_threshold,
                    constants.pla_outside_los_area_extension,
                )

        if self.defaultAE.environment is not None:
            for ae in acousticEnvironments:
                if ae.id == self.defaultAE.environment:
                    self.defaultAE.environment = ae
                    acousticEnvironments.remove(ae)
                    break

        for ae in acousticEnvironments:
            if ae.parameters[0].predelay is None or any(
                f for f in ae.parameters[0].frequencies if f.dsr is None
            ):
                geom = eifScene.geometryByID[ae.region]
                if type(geom) == Box:
                    bounds = geom.size
                elif type(geom) == Mesh:
                    bounds = dsrHelpers.calculateBoundingVolumeOfMesh(geom)
                else:
                    raise TypeError("Unknown geometry type for acoustic environment")

                roomVolume = bounds[0] * bounds[1] * bounds[2]

                ae.parameters[0].predelay = dsrHelpers.calculateEIFpredelay(bounds)
                rendererPredelay = dsrHelpers.calculateRendererPredelay(
                    bounds, ae.parameters[0].predelay
                )
                for f in ae.parameters[0].frequencies:
                    if f.dsr is None:
                        f.dsr = self.calculateTheoreticalDsr(
                            f.rt60, roomVolume, ae.parameters[0].predelay
                        )
                        scaleFactor = dsrHelpers.calculateDSRscaling(
                            f.rt60, rendererPredelay, ae.parameters[0].predelay
                        )
                        f.dsr /= scaleFactor

        return ReverbMetaData(
            acousticEnvironments=acousticEnvironments,
            defaultAE=self.defaultAE,
            sceneType=sceneType,
            revRendererControl=revRendererControl,
            aeLosVoxelData=plaAeLosVoxelData,
            outsideLosVoxelData=plaOutsideLosVoxelData,
        )

    def calculateTheoreticalDsr(self, rt60, roomVolume, eifPredelay):
        if rt60 > 0:
            rdrLog = 10 * math.log10(rt60 / roomVolume) - 60 * (eifPredelay / rt60) + 25
        else:
            rdrLog = -60

        dsrLog = (
            rdrLog + constants.lsdf_rdr_to_ddr_db + constants.lsdf_theoretical_rdr_to_actual_rdr
        )
        dsrLin = math.pow(10, dsrLog / 10)
        return dsrLin


@dataclass
class ReverbMetaData:
    acousticEnvironments: List[AcousticEnvironment]
    defaultAE: bool
    sceneType: SceneType
    revRendererControl: list
    aeLosVoxelData: list
    outsideLosVoxelData: list
