# This software module was originally developed by
#
# Nokia
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Nokia retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2024.

import itertools
from typing import List

import mpegi.metadata.reverb.constants as constants
from mpegi.geometry.trimesh import _convertMeshToTrimesh, _convertBoxToTrimesh
from mpegi.metadata.eif import Portal, Mesh, Box, AcousticEnvironment
from trimesh import creation, Trimesh, transformations
from trimesh.ray.ray_pyembree import RayMeshIntersector

if constants.debug_enable_visualization:
    import matplotlib.pyplot as plt
import numpy as np
import trimesh as trimesh

# Plotting parameters
dspr_plot3d_axis_ticks = [x for x in np.arange(-10, 10, 1)]
dspr_plot3d_figure_size = (15, 15)


def populate_box_positions(interp_resol: float, mesh: Trimesh, debug: bool = False):
    if len(mesh.vertices) != 8:
        raise Exception("mesh has to be a box")

    length, height, depth = mesh.extents

    x_elems = mesh.vertices[:, 0]
    y_elems = mesh.vertices[:, 1]
    z_elems = mesh.vertices[:, 2]

    min_x, max_x = min(x_elems), max(x_elems)
    min_y, max_y = min(y_elems), max(y_elems)
    min_z, max_z = min(z_elems), max(z_elems)

    # Starting vertex (top-left)
    v_s_x, v_s_y, v_s_z = [min_x, max_y, min_z]

    n_l_interp = int(length / interp_resol) + 1
    n_d_interp = int(depth / interp_resol) + 1
    n_h_interp = int(height / interp_resol) + 1

    points = []
    for i in range(n_l_interp * n_d_interp * n_h_interp):
        if i % n_l_interp == 0 and i > 0:
            v_s_z += interp_resol
            v_s_x = min_x

        if i % (n_l_interp * n_d_interp) == 0 and i > 0:
            v_s_z = min_z
            v_s_y -= interp_resol

        point = [v_s_x, v_s_y, v_s_z]
        points.append(point)
        v_s_x += interp_resol

    if debug:
        ax, fig = get_3d_plot(True)
        for p in points:
            print(p)
            x, y, z = p
            ax.scatter(x, z, y, color="black")
        points = np.asarray(points)
        max_x, min_x = max(points[:, 0]), min(points[:, 0])
        max_y, min_y = max(points[:, 1]), min(points[:, 1])
        max_z, min_z = max(points[:, 2]), min(points[:, 2])

        ax.set_title(
            f"{mesh.extents}\n" f"{max_x=}, {min_x=}\n{max_y=}, {min_y=}\n{max_z=}, {min_z=}"
        )
        plt.show()

    return points


def determine_floor_face(faces: []):
    face_centroids = []

    for i in range(len(faces)):
        face = faces[i]
        face_centroid = calculate_centroid(face)
        face_centroids.append(face_centroid[1])  # Y

    floor_face = faces[face_centroids.index(min(face_centroids))]

    return floor_face


def is_point_beyond(p1: List[float], p2: List[float], test_point: List[float], atol: float = 0):
    x1, y1, z1 = p1
    x2, y2, z2 = p2
    t_x, t_y, t_z = test_point

    vec_d = distance_3d(x1, y1, z1, x2, y2, z2)
    test_d = distance_3d(x1, y1, z1, t_x, t_y, t_z)

    if test_d >= vec_d:
        return True

    elif test_d < vec_d:
        if atol > 0:
            if np.isclose(test_d, vec_d, atol=atol):
                return True

        return False


def determine_ceiling_face(faces: []):
    face_centroids = []

    for i in range(len(faces)):
        face = faces[i]
        face_centroid = calculate_centroid(face)
        face_centroids.append(face_centroid[1])  # Y

    floor_ceiling = faces[face_centroids.index(max(face_centroids))]
    return floor_ceiling


def distance_3d(x1, y1, z1, x2, y2, z2):
    a = np.power(x1 - x2, 2)
    b = np.power(y1 - y2, 2)
    c = np.power(z1 - z2, 2)

    return np.power((a + b + c), 0.5)


def triangle_area_3d(x1, y1, z1, x2, y2, z2, x3, y3, z3):
    x1_d = distance_3d(x1, y1, z1, x2, y2, z2)
    x2_d = distance_3d(x2, y2, z2, x3, y3, z3)
    x3_d = distance_3d(x3, y3, z3, x1, y1, z1)

    s = (x1_d + x2_d + x3_d) / 2
    return np.power((s * (s - x1_d) * (s - x2_d) * (s - x3_d)), 0.5)


def is_point_inside_box(vertices: [], point: List[float]):
    # Checks if point lies inside 8 vertices, only for axis-oriented box
    if len(vertices) != 8:
        raise ValueError("Invalid amount of vertices")

    vertices_np = np.asarray(vertices)

    x_coords = vertices_np[:, 0]
    y_coords = vertices_np[:, 1]
    z_coords = vertices_np[:, 2]
    max_x, max_y, max_z = max(x_coords), max(y_coords), max(z_coords)
    min_x, min_y, min_z = min(x_coords), min(y_coords), min(z_coords)

    return min_x <= point[0] <= max_x and min_y <= point[1] <= max_y and min_z <= point[2] <= max_z


def get_meshes_in_box(
    box: Trimesh,
    meshes_to_check,
    inside_percentage_thr: int = 100,
    expand: bool = True,
    return_meshes: bool = False,
):
    box_meshes = []

    # Expand box to add tolerance
    if expand:
        exp = 0.5
        extents_e = [box.extents[0] + exp, box.extents[1] + exp, box.extents[2] + exp]
        box_e = creation.box(
            extents=extents_e, transform=transformations.translation_matrix(direction=box.centroid)
        )
    else:
        box_e = box

    for mesh_id, mesh in meshes_to_check.items():
        i_c = 0
        for v in mesh.vertices:
            if is_point_inside_box(box_e.vertices, v):
                i_c += 1

        if i_c == 0:
            continue

        if inside_percentage_thr == 0 and i_c > 0:
            box_meshes.append([mesh_id, mesh])
        else:
            i_prt = i_c / len(mesh.vertices) * 100
            if i_prt >= inside_percentage_thr:
                box_meshes.append([mesh_id, mesh])

    if return_meshes:
        box_meshes_new = []
        for mesh_id, mesh in box_meshes:
            box_meshes_new.append(mesh)
        box_meshes = box_meshes_new

    return box_meshes


def plot_trimeshes_3d(
    meshes: [trimesh.Trimesh],
    points: List[List[float]] = [],
    m_color: str = "grey",
    p_color: str = "red",
    auto_scale: bool = True,
    crd_sys: str = "OGL",
    alpha: float = 1,
    label: str = "",
):
    ax, fig = get_3d_plot(auto_scale=auto_scale, crd_sys=crd_sys)
    ax.set_title(label)
    set_axis_labels(crd_sys, ax)

    if not isinstance(meshes, List):
        meshes = [meshes]

    for mesh in meshes:
        if crd_sys == "OGL":
            ax.plot_trisurf(
                mesh.vertices[:, 0],
                mesh.vertices[:, 2],
                mesh.vertices[:, 1],
                alpha=alpha,
                triangles=mesh.faces,
                color=m_color,
            )

        elif crd_sys == "XYZ":
            ax.plot_trisurf(
                mesh.vertices[:, 0],
                mesh.vertices[:, 1],
                mesh.vertices[:, 2],
                alpha=alpha,
                triangles=mesh.faces,
                color=m_color,
            )

    for p in points:
        x, y, z = p
        if crd_sys == "OGL":
            ax.scatter(x, z, y, alpha=alpha, color=p_color)
        elif crd_sys == "XYZ":
            ax.scatter(x, y, z, alpha=alpha, color=p_color)

    plt.show()


def set_axis_labels(crd_sys: str, ax):
    if crd_sys == "OGL":
        ax.set_xlabel("X")
        ax.set_ylabel("Z")
        ax.set_zlabel("Y")

    elif crd_sys == "XYZ":
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
    else:
        raise ValueError(f"Invalid coord system: {crd_sys}")


def get_3d_plot(auto_scale: bool = False, crd_sys: str = "OGL"):
    if crd_sys != "OGL" and crd_sys != "XYZ":
        raise ValueError(f"Supported coordinate systems: OGL, XYZ")

    fig = plt.figure(figsize=dspr_plot3d_figure_size)
    ax = fig.add_subplot(111, projection="3d")
    set_axis_labels(crd_sys, ax)

    ax.autoscale(auto_scale, axis="x")
    ax.autoscale(auto_scale, axis="y")
    ax.autoscale(auto_scale, axis="z")

    if not auto_scale:
        plt.xticks(dspr_plot3d_axis_ticks)
        plt.yticks(dspr_plot3d_axis_ticks)
        ax.set_zticks(dspr_plot3d_axis_ticks)

    return ax, fig


def get_trimesh_faces_normals_areas(mesh: trimesh.Trimesh):
    if not isinstance(mesh, trimesh.Trimesh):
        raise TypeError("Only trimesh supported")

    mesh_faces = []
    mesh_normals = []
    mesh_face_areas = []
    for i in range(len(mesh.facets)):
        mesh_f1 = mesh.faces[mesh.facets[i][0]]
        mesh_f2 = mesh.faces[mesh.facets[i][1]]
        mesh_f1_area = mesh.facets_area[i]
        mesh_f1_v = [
            list(mesh.vertices[mesh_f1[0]]),
            list(mesh.vertices[mesh_f1[1]]),
            list(mesh.vertices[mesh_f1[2]]),
        ]
        mesh_f2_v = [
            list(mesh.vertices[mesh_f2[0]]),
            list(mesh.vertices[mesh_f2[1]]),
            list(mesh.vertices[mesh_f2[2]]),
        ]

        mesh_normal = mesh.facets_normal[i]

        # This gives 6 vertices [/], remove duplicates to get 4 []
        mesh_quad_face = mesh_f1_v + mesh_f2_v

        for j in range(len(mesh_quad_face)):
            for k in range(3):
                mesh_quad_face[j][k] = float(mesh_quad_face[j][k])

        # Remove duplicates
        mesh_quad_face.sort()
        mesh_quad_face_nodupes = list(
            mesh_quad_face for mesh_quad_face, _ in itertools.groupby(mesh_quad_face)
        )
        mesh_normals.append(mesh_normal)
        mesh_face_areas.append(mesh_f1_area)
        mesh_faces.append(mesh_quad_face_nodupes)

    return mesh_faces, mesh_normals, mesh_face_areas


def calculate_centroid(vertices: []):
    face_x, face_y, face_z = (
        [v[0] for v in vertices],
        [v[1] for v in vertices],
        [v[2] for v in vertices],
    )
    return [sum(face_x) / len(vertices), sum(face_y) / len(vertices), sum(face_z) / len(vertices)]


def expand_box_horizontal(mesh: Trimesh, amount_meters_x: float, amount_meters_z: float):
    if len(mesh.vertices) != 8:
        raise Exception("Mesh needs to be a box")

    centroid = calculate_centroid(mesh.vertices)
    w, h, l = mesh.extents

    return creation.box(
        extents=[w + amount_meters_x, h, l + amount_meters_z],
        transform=transformations.translation_matrix(direction=centroid),
    )


def set_box_height(mesh: Trimesh, amount_meters: float):

    if len(mesh.vertices) != 8:
        raise Exception("Mesh needs to be a box")

    faces, _, __ = get_trimesh_faces_normals_areas(mesh)
    floor_f = determine_floor_face(faces)
    ceiling_f = determine_ceiling_face(faces)
    elev = floor_f[0][1]

    v_new = []
    for v in ceiling_f:
        x, y, z = v
        v_new.append([x, elev + amount_meters, z])

    for v in floor_f:
        v_new.append(v)

    v_new = np.asarray(v_new)
    mesh_new = Trimesh(
        v_new,
        mesh.faces,
        mesh.face_normals,
        mesh.vertex_normals,
        None,
        None,
        mesh.face_attributes,
        mesh.vertex_attributes,
    )

    return mesh_new


def homogen_mat(x, y, z, o):
    return np.asarray(
        [x[0], y[0], z[0], o[0], x[1], y[1], z[1], o[1], x[2], y[2], z[2], o[2], 0, 0, 0, 1]
    ).reshape((4, 4))


def get_portal_geom(portal: Portal, portal_ext_mesh: Mesh):
    if portal.portalExtent == None:
        portal_ext_trimesh = creation.box(
            extents=[portal.portalSize[0], portal.portalSize[1], portal.portalSize[2]]
        )
    else:
        if isinstance(portal_ext_mesh, Mesh):
            portal_ext_trimesh = _convertMeshToTrimesh(portal_ext_mesh, includeTransparent=True)
        elif isinstance(portal_ext_mesh, Box):
            portal_ext_trimesh = _convertBoxToTrimesh(portal_ext_mesh)

    portal_ext_trimesh.apply_transform(
        trimesh.transformations.translation_matrix(direction=portal_ext_mesh.position)
    )

    return portal_ext_trimesh


def nd_bin_mask_to_arr(nd_bin_arr):
    arr = np.asarray(nd_bin_arr)
    return arr.flatten()


def arr_to_nd_bin_mask(bin_arr: List[int], n_height, n_col, n_row):
    mask = np.zeros(shape=(n_height, n_col, n_row), dtype=np.int8)
    x_row = 0
    y_row = 0
    z_row = 0

    for i in range(n_row * n_col * n_height):
        truth = bin_arr[i]

        if i > 0 and i % n_row == 0:
            x_row = 0
            z_row += 1

        if z_row > 0 and z_row == n_col:
            z_row = 0
            x_row = 0
            y_row += 1

            if y_row == n_height:
                break

        mask[y_row][z_row][x_row] = truth
        x_row += 1

    return mask


# Returns top-left and bottom-right corner points
def get_diagonal_corners(mesh_or_vertices):

    if isinstance(mesh_or_vertices, trimesh.Trimesh) or isinstance(mesh_or_vertices, Mesh):
        mesh_or_vertices = mesh_or_vertices.vertices

    max_x, min_x = max(mesh_or_vertices[:, 0]), min(mesh_or_vertices[:, 0])
    max_y, min_y = max(mesh_or_vertices[:, 1]), min(mesh_or_vertices[:, 1])
    max_z, min_z = max(mesh_or_vertices[:, 2]), min(mesh_or_vertices[:, 2])

    diagonal_p1 = [min_x, max_y, min_z]
    diagonal_p2 = [max_x, min_y, max_z]

    return [diagonal_p1, diagonal_p2]


def get_outside_portals(portals: List[Portal]):
    return [
        portal
        for portal in portals
        if portal.acousticEnvironment1 == None or portal.acousticEnvironment2 == None
    ]


def get_outside_aes(portals: List[Portal]):
    return [
        portal.acousticEnvironment1
        for portal in portals
        if portal.acousticEnvironment1 == None or portal.acousticEnvironment2
        for portal in portals
        if portal.acousticEnvironment2 == None
    ]


def get_ae_portals(ae_id: str, portals: List[Portal]):
    return [
        exp_p
        for exp_p in portals
        if exp_p.acousticEnvironment1 == ae_id or exp_p.acousticEnvironment2 == ae_id
    ]


def get_ae_portal_ids(ae_id: str, portals: List[Portal]):
    return [
        exp_p.id
        for exp_p in portals
        if exp_p.acousticEnvironment1 == ae_id or exp_p.acousticEnvironment2 == ae_id
    ]


def get_portal_by_id(p_id: str, portals: List[Portal]):
    return [exp_p for exp_p in portals if exp_p.id == p_id][0]


def get_ae_by_portal(portal: Portal, aes: List[AcousticEnvironment]):
    portal_ae = None
    for ae in aes:
        ae_portals = get_ae_portals(ae.id)
        for ae_portal in ae_portals:
            if ae_portal == portal:
                portal_ae = ae
    return portal_ae


def get_ae_outside_portals(ae_id: str, portals: List[Portal]):
    o_portals = get_outside_portals(portals)
    ae_o_portals = []
    for p in o_portals:
        if p.acousticEnvironment1 == ae_id or p.acousticEnvironment2 == ae_id:
            ae_o_portals.append(p)

    return ae_o_portals


def ae_geom_as_trimesh(ae_geom: Mesh):
    ae_geom_trimesh = None
    ae_bounding_geom_trimesh = None

    if isinstance(ae_geom, Box):
        ae_geom_trimesh = _convertBoxToTrimesh(ae_geom)

    elif isinstance(ae_geom, Mesh) and len(ae_geom.vertices) == 8:
        ae_geom_trimesh = _convertMeshToTrimesh(ae_geom, includeTransparent=True)

    # If geom is a mesh and not a box, convert it to trimesh and use it's bounding box
    elif isinstance(ae_geom, Mesh) and len(ae_geom.vertices) != 8:
        ae_geom_trimesh = _convertMeshToTrimesh(ae_geom, includeTransparent=True)

    ae_geom_trimesh.apply_transform(
        trimesh.transformations.translation_matrix(direction=ae_geom.position)
    )

    if not np.array_equal(ae_geom.orientation, np.asarray([0, 0, 0])):
        # Means that we use oriented bb
        ae_bounding_geom_trimesh = ae_geom_trimesh.bounding_box_oriented
    else:
        ae_bounding_geom_trimesh = ae_geom_trimesh.bounding_box

    if ae_geom_trimesh is None or ae_bounding_geom_trimesh is None:
        raise Exception("Can't determine ae geometry")

    return [ae_geom_trimesh, ae_bounding_geom_trimesh]


def get_portal_open_faces(portal: Portal, portal_ext_mesh: Mesh):
    if list(portal.portalNormal) == [0, 0, 0]:
        raise Exception(f"portal {portal.id} has no normal defined")

    n_x, n_y, n_z = portal.portalNormal
    portal_geom = get_portal_geom(portal, portal_ext_mesh)
    ray_origins = [list(portal_geom.centroid) for i in range(2)]
    ray_directions = [[n_x, n_y, n_z], [-n_x, -n_y, -n_z]]

    pgeom_intersector = RayMeshIntersector(portal_geom)
    for i in range(2):
        hit_pos, ray, tri = pgeom_intersector.intersects_location(
            ray_origins=ray_origins, ray_directions=ray_directions, multiple_hits=False
        )

        if len(hit_pos) < 2:
            # In some rare occasions ray hit doesnt register the other face, nudge ray origins a bit and try again
            eps = 1e-06
            ray_origins[0] = [
                ray_origins[0][0] + eps,
                ray_origins[0][1] + eps,
                ray_origins[0][2] + eps,
            ]
            ray_origins[1] = [
                ray_origins[1][0] + eps,
                ray_origins[1][1] + eps,
                ray_origins[1][2] + eps,
            ]
            continue

        t1, t2 = tri
        f1, f2 = None, None

        for facet in portal_geom.facets:
            if facet[0] == t1 or facet[1] == t1:
                f1 = facet
            elif facet[0] == t2 or facet[1] == t2:
                f2 = facet

        if f1 is None or f2 is None:
            raise Exception(f"No portal open faces found for portal: {portal.id}")

        face1_tris = [portal_geom.faces[f1[0]], portal_geom.faces[f1[1]]]
        face2_tris = [portal_geom.faces[f2[0]], portal_geom.faces[f2[1]]]
        face1_verts = []
        face2_verts = []
        for i in range(2):
            tri1 = list(face1_tris[i])
            tri2 = list(face2_tris[i])
            for j in range(3):
                tri1_idx = tri1[j]
                tri2_idx = tri2[j]
                x, y, z = portal_geom.vertices[tri1_idx]
                x1, y1, z1 = portal_geom.vertices[tri2_idx]
                face1_verts.append([x, y, z])
                face2_verts.append([x1, y1, z1])

        face1_verts.sort()
        face1_verts = list(face1_verts for face1_verts, _ in itertools.groupby(face1_verts))
        face2_verts.sort()
        face2_verts = list(face2_verts for face2_verts, _ in itertools.groupby(face2_verts))

        if len(face1_verts) != 4 or len(face2_verts) != 4:
            raise Exception(f"Invalid number of portal open face vertices")

        return face1_verts, face2_verts

    return None
