# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
# SOFAReader/Writer reads/writes files bsaed on the
# Spatially Oriented Format for Acoustics
# as defined in AES69-2015.
#
# @author Simon Schwaer (sws@iis.fraunhofer.de)
# @required
#    h5py
#    h5netcdf
#    numpy
#

import h5py
import h5netcdf
import numpy as np
import traceback
from scipy.spatial import KDTree
from typing import Dict, List


class SOFAReader(object):
    def __init__(self, file):
        self.sofa = h5py.File(file, "r")

        # Cache is used for frequently accessed variables with "getVar()"
        self._cache = {}

    def __enter__(self):
        return self

    def getDim(self, dimensionName: str) -> int:
        """Get dimensions

        :param dimensionName: Available dimensions
            C: coordinate dimensions,
            E: number of Emitters,
            I: singleton scalar value (=1),
            M: number of measurement,
            N: number of data samples per measurement,
            R: number of receivers
        :type dimensionName: str
        :return: Size of dimension
        :rtype: int
        """
        return self.sofa[dimensionName].len()

    def getMetaData(self) -> Dict:
        """Returns a dictionary of all global metadata stored in the SOFA file

        :return: Metadata of SOFA file
        :rtype: Dict
        """
        meta = {}
        for attr, value in self.sofa.attrs.items():
            meta[attr] = value
        return meta

    def getAvailableVars(self) -> List[str]:
        """Returns a list of available variables

        :return: Available variables
        :rtype: List[str]
        """
        return self.sofa.keys()

    def getVar(self, variableName: str, attribute: bool = False):
        """Return variable in SOFA file

        :param variableName: Available variables: Data.Delay, Data.IR,
            Data.SamplingRate, EmitterPosition, ListenerPosition, ListenerUp,
            ListenerView, ReceiverPosition, SourcePosition
        :type variableName: str
        :param attribute: defaults to False
        :type attribute: bool, optional
        """
        try:
            var = self._cache[variableName]
        except:
            self._cache[variableName] = self.sofa[variableName]
            var = self._cache[variableName]

        return var[:] if not attribute else var.attrs[attribute]

    def getClosestDirection(self, direction: List[float]) -> List[float]:
        """Returns the SourcePosition coordinates closest to the ones given

        :param direction: Polar direction [Azimuth, Elevation, Distance]
        :type direction: List[float]
        :return: Closest SourcePosition in Polar Direction
        :rtype: List[float]
        """

        try:
            sourcePos = self.polarToCart(direction[0], direction[1], direction[2])
            _, index = self._cache["KDTree"].query(sourcePos)
        except:
            cartesianSourcePos = [
                self.polarToCart(x[0], x[1], x[2]) for x in self.getVar("SourcePosition")
            ]
            self._cache["KDTree"] = KDTree(cartesianSourcePos)

            sourcePos = self.polarToCart(direction[0], direction[1], direction[2])
            _, index = self._cache["KDTree"].query(sourcePos)

        return self.getVar("SourcePosition")[index]

    def getFIR(self, coordinates: List[float]) -> np.ndarray:
        """Returns the FIR (shape: R x N) at given SourcePosition coordinates

        :param coordinates: SourcePosition
        :type coordinates: List[float]
        :raises ValueError: No HRTF for given coordinates found
        :return: FIR at given SourcePosition
        :rtype: np.ndarray
        """
        meas = self.__findFirst(self.getVar("SourcePosition"), np.array(coordinates))
        if meas != None:
            if self.sofa.attrs["DataType"] == "FIR":
                return self.getVar("Data.IR")[meas]

            elif self.sofa.attrs["DataType"] == "TF":
                real = self.getVar("Data.Real")[meas]
                imag = self.getVar("Data.Imag")[meas]
                return np.fft.irfft(real + 1j * imag)
        else:
            raise ValueError("No HRTF for given coordinates found.")

    def getTF(
        self, coordinates: List[float], minLength: bool = False, maxLength: bool = False
    ) -> np.ndarray:
        """Returns the TF (abs value) (shape: R x N) at given SourcePosition
        coordinates

        :param coordinates: SourcePosition coordinates
        :type coordinates: List[float]
        :param minLength: Minimum Length, defaults to False
        :type minLength: bool, optional
        :param maxLength: Maximum Length, defaults to False
        :type maxLength: bool, optional
        :raises ValueError: No HRTF for given coordinates found
        :return: Transfer Function
        :rtype: complex np.ndarray
        """
        meas = self.__findFirst(
            self.getVar("SourcePosition"),
            self.getClosestDirection(np.array(coordinates)),
        )
        if meas != None:
            # Decode needed because DataType attribute is of type numpy.bytes_ for some reason...
            if self.sofa.attrs["DataType"] == bytes("FIR", "utf-8"):
                fir = self.getVar("Data.IR")[meas]

                if minLength and len(fir[0]) < minLength:
                    zeroLength = minLength - len(fir[0])
                    fir = np.lib.pad(
                        fir, ((0, 0), (0, zeroLength)), "constant"
                    )  # append zeros to reach length of minLength

                if maxLength and len(fir[0]) > maxLength:
                    fir = fir[:, :maxLength]

                tf = np.empty(
                    (2, (len(fir[0]) / 2 + 1)), dtype=np.complex64
                )  # numpy real fft returns only the positive half of the hermitian spectrum
                tf[0] = np.fft.rfft(fir[0])
                tf[1] = np.fft.rfft(fir[1])
                return tf

            # Decode needed because DataType attribute is of type numpy.bytes_ for some reason...
            elif self.sofa.attrs["DataType"] == bytes("TF", "utf-8"):
                real = self.getVar("Data.Real")[int(np.rint(meas))]
                imag = self.getVar("Data.Imag")[int(np.rint(meas))]
                return real + 1j * imag
        else:
            raise ValueError("No HRTF for given coordinates found.")

    def __exit__(self, errType, errVal, errTrace):
        if errType:
            print(errType, errVal)
            traceback.print_tb(errTrace)
        return True

    def __del__(self):
        self.sofa.close()

    #
    # private functions
    #

    def __findFirst(self, array, pattern):
        """Find first occurance of a pattern (b) in a numpy array (a)"""
        array = array.flatten()
        lb, la = len(pattern), len(array)
        lb = lb - 1
        mask = array[lb:la] == pattern[lb]
        for i in range(lb):
            mask *= array[i : la - lb + i] == pattern[i]
        return None if not np.any(mask) else np.argmax(mask) / len(pattern)

    @staticmethod
    def polarToCart(azi, ele, dist):
        piOver180 = np.pi / 180.0
        x = -1.0 * dist * np.sin(azi * piOver180) * np.cos(ele * piOver180)
        y = dist * np.sin(ele * piOver180)
        z = -1.0 * dist * np.cos(azi * piOver180) * np.cos(ele * piOver180)
        return [x, y, z]


class SOFAWriter(object):
    apiName = "Fraunhofer IIS SOFAWriter for Python"
    apiVersion = "3.0.0"

    def __init__(self, file):
        self.sofa = h5netcdf.File(file, "w")

    def __enter__(self):
        return self

    def getVar(self, variableName: str, attribute: bool = False):
        """Return variable in SOFA file

        :param variableName: Available variables: Data.Delay, Data.IR,
            Data.SamplingRate, EmitterPosition, ListenerPosition, ListenerUp,
            ListenerView, ReceiverPosition, SourcePosition
        :type variableName: str
        :param attribute: defaults to False
        :type attribute: bool, optional
        """
        var = self.sofa[variableName]
        return var[:] if not attribute else var.attrs[attribute]

    def setDim(self, dimensionName, size):
        self.sofa.dimensions.update({dimensionName: size})
        return self

    def setDimWithValues(self, dimensionName, vals):
        self.sofa.dimensions.update({dimensionName: len(vals)})
        v = self.sofa.create_variable(dimensionName, (dimensionName), float)
        v[:] = vals
        return self

    def setMetaData(self, meta, parent=False):
        item = self.sofa if not parent else self.sofa[parent]

        for attr, value in meta.items():
            item.attrs[attr] = np.string_(value)
        return self

    def setVar(self, variableName, values, dimensions, attributes=False, dtype=float):
        if variableName not in self.sofa.keys():
            # dtype = values.dtype if not dtype else dtype

            var = self.sofa.create_variable(variableName, dimensions, dtype=dtype)

        var[:] = values

        if attributes and isinstance(attributes, dict):
            self.setMetaData(attributes, parent=variableName)

        return self

    def setFIR(self, coordinates, fir):
        if "SourcePosition" not in self.sofa:
            raise ValueError("SourcePosition dataset is not initialized.")

        meas = self.__findFirst(self.getVar("SourcePosition"), np.array(coordinates))
        if meas:
            if self.sofa.attrs["DataType"] == "FIR":
                if "Data.IR" not in self.sofa:
                    raise ValueError("Dataset for payload is not initialized.")

                self.sofa["Data.IR"][meas] = fir

            elif self.sofa.attrs["DataType"] == "TF":
                if "Data.Real" not in self.sofa:
                    raise ValueError("Dataset for payload is not initialized.")

                tf = np.empty(
                    (2, (len(fir[0]) / 2 + 1)), dtype=np.complex64
                )  # numpy real fft returns only the positive half of the hermitian spectrum
                tf[0] = np.fft.rfft(fir[0])
                tf[1] = np.fft.rfft(fir[1])
                self.sofa["Data.Real"][meas] = np.real(tf)
                self.sofa["Data.Imag"][meas] = np.imag(tf)

        return self

    def setTF(self, coordinates, tf):
        if "SourcePosition" not in self.sofa:
            raise ValueError("SourcePosition dataset is not initialized.")

        meas = self.__findFirst(self.getVar("SourcePosition"), np.array(coordinates))
        if meas:
            if self.sofa.attrs["DataType"] == "FIR":
                if "Data.IR" not in self.sofa:
                    raise ValueError("Dataset for payload is not initialized.")

                self.sofa["Data.IR"][meas] = np.fft.irfft(tf)

            elif self.sofa.attrs["DataType"] == "TF":
                if "Data.Real" not in self.sofa:
                    raise ValueError("Dataset for payload is not initialized.")

                self.sofa["Data.Real"][meas] = np.real(tf)
                self.sofa["Data.Imag"][meas] = np.imag(tf)

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_value:
            print(exc_value)  # for debug
        return True

    def __del__(self):
        self.sofa.attrs["APIName"] = np.string_(SOFAWriter.apiName)
        self.sofa.attrs["APIVersion"] = np.string_(SOFAWriter.apiVersion)

        self.sofa.close()

    #
    # private functions
    #

    # find first occurence of a pattern (b) in numpy array (a)
    def __findFirst(self, array, pattern):
        array = array.flatten()
        lb, la = len(pattern), len(array)
        lb = lb - 1
        mask = array[lb:la] == pattern[lb]
        for i in range(lb):
            mask *= array[i : la - lb + i] == pattern[i]
        return None if not np.any(mask) else np.argmax(mask) / len(pattern)
