# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from .voxelgrid import VoxelGrid


class VoxelData:
    """Sparse data storage for information attributed to voxels in a grid

    This class realized an encapsulated storage for data attributed
    to voxels in a voxel grid. The data is represented by a freely
    usable Python dictionary.

    As a voxel grid is usually sparsely populated, dictionaries (hashes) are
    the best way to store the information, without wasting memory.
    The reason to realize this functionality in an own class is implementation freedom.
    The user of the class does not need to consider a certain storage scheme in his code.
    """

    # Note: After several performance examples I found, that the voxel data is best
    #       stored in a dictionary (hash table). Normal arrays become to big for
    #       large scenes and high voxel resolutions. Even for moderate resolutions,
    #       the dictionary has no real disadvantage.

    def __init__(self, grid):
        """Create an empty voxel data for the given voxel grid"""
        if not isinstance(grid, VoxelGrid):
            raise ValueError("Expecting a voxel grid")
        self.__grid = grid
        self.__storage = dict()

    def tryGet(self, index):
        """Try getting the data for a voxel

        :note This function does not create data, if it does not yet exist and returns None in this case.
              It is designed for queries, where it is useful not to leave a memory footprint.

        :param index Voxel coordinate or voxel index (auto-converted)

        :returns Dict or None (if no data present)
        """
        # Auto-convert a voxel coordinate into a voxel index
        vi = self.__grid.asVoxelIndex(index)
        return self.__storage.get(vi)

    def get(self, index):
        """Gets the data for a voxel or initialize it, when not present

        :note This function creates data, if it does not yet exist.
              This will increase the memory footprint.

        :param index Voxel coordinate or voxel index (auto-converted)

        """
        # Auto-convert a voxel coordinate into a voxel index
        vi = self.__grid.asVoxelIndex(index)
        vd = self.__storage.get(vi)
        if vd:
            return vd

        # Lazy initialization on first access
        vd = dict()
        self.__storage[vi] = vd
        return vd

    @property
    def populated(self):
        """Returns the voxels with contain any data
        :returns List of voxel indices
        """
        return self.__storage.keys()
