# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import numpy as np
from mpegi.metadata.eif.utils import positionToString
from mpegi.metadata.eif.validation import validateVector3, validateIntVector3
from mpegi.geometry.tribox import triBoxOverlap


class VoxelGrid:
    """
    This class formally describes a voxel grid (origin, edge length, etc.)
    and provides a rich set of functions to work with such grids. This
    includes coordinate transformations, neighbour queries, etc.
    It does not perform any voxelization operations or store voxel-based data.

    To clarify terms:

    Spatial coordinate
        Point in 3D space (double-3)

    Voxel coordinate
        Voxel position in the grid (integer-3)

    Voxel index
        Linear (1D) representation of a voxel coordinate introduced by a certain traversal scheme.
    """

    def __init__(self, origin, edge_length, size):
        """
        Creates a uniform voxel grid in three dimensions

        :param origin:      Origin (vector-3)
        :param edge_length: Voxel edge length (>0) in each dimension (vector-3)
        :param size:        Number of voxels in each dimension (int-3)

        :note The origin is a corner point of the voxel (0,0,0), not its center point
        :note Points on the outer margin (origin + size*spacing) are included in the grid as a special case
        """

        self.__x0 = validateVector3(origin, "origin")
        self.__dx = validateVector3(edge_length, "edge length")
        if np.any(self.__dx <= 0):
            raise ValueError("Invalid edge length")
        self.__size = validateIntVector3(size, "size")
        if np.any(self.__size <= 0):
            raise ValueError("Invalid size")

        self.__x1 = self.__x0 + self.__size * self.__dx

        # Total number of voxels
        self.__nvoxels = np.prod(self.__size)

    @property
    def origin(self):
        """Grid origin (vector-3)"""
        return self.__x0

    @property
    def edge_length(self):
        """Voxel edge length (vector-3)"""
        return self.__dx

    @property
    def vc_max(self):
        """Voxel grid size (vector-3)"""
        return self.__size

    @property
    def nvoxels(self):
        """Total number of voxels (integer)"""
        return self.__nvoxels

    # --= Coordinate calculations =--

    def point2VoxelCoordinates(self, x):
        """Compute the (three-integer) voxel coordinate of a point (three-float) in space

        .. note::
            For performance reasons, this function does not check if the points lies within the grid
            The returned coordinate values can thus be out of range.

        :param x:  Point (vector-3)
        :returns: nparray(int,3) Voxel coordinate if x lies in the grid boundaries, None otherwise
        """

        # # TODO: Arguably not the fastest code. Improvde speed when needed...
        # i = np.divide(x - self.__origin, self.__spacing)
        #
        # if np.any(i < 0):
        #     return None
        # if np.any(self.__size - i < 0):
        #     return None
        #
        # # Note: Checking < 0 above (instead of <= 0) is very important!
        # #       Ending up here, np.any(self.__size - i == 0) can happen here,
        # #       when the point lies exactly on the grid's outer margin.
        # #       In this exception case, the point is associated with the nearby margin voxel.
        # #       This is taken care of by clipping the value (see below).
        # #
        # # Example:
        # #    vg = VoxelGrid([0,0,0], [1,1,1], [1, 2, 3])
        # #    vg.getVoxelCoordinates([1, 2, 3]) => [0, 1, 2]
        # #
        # #    Without the rule above, [1, 2, 3] would lie outside of the grid!
        # #
        # return np.clip(np.floor(i).astype(np.int_), [0, 0, 0], self.__size - 1)

        # Performance: This code is 2.5x faster than the above code
        vc = np.zeros((3,), dtype=int)

        # Not needed: if x[0] < self.__x0[0] or x[0] > self.__x1[0]: return None
        if x[0] == self.__x1[0]:
            vc[0] = self.__size[0] - 1
        else:
            vc[0] = (x[0] - self.__x0[0]) / self.__dx[0]

        # Not needed: if x[1] < self.__x0[1] or x[1] > self.__x1[1]: return None
        if x[1] == self.__x1[1]:
            vc[1] = self.__size[1] - 1
        else:
            vc[1] = (x[1] - self.__x0[1]) / self.__dx[1]

        # Not needed: if x[2] < self.__x0[2] or x[2] > self.__x1[2]: return None
        if x[2] == self.__x1[2]:
            vc[2] = self.__size[2] - 1
        else:
            vc[2] = (x[2] - self.__x0[2]) / self.__dx[2]

        return vc

    # def point2VoxelIndex(self, x):
    #     """Compute the (linear) voxel index of a point (three-float) in space
    #     :param x:  Point (vector-3)
    #     :returns integer: Voxel index if x lies within the grid boundaries,
    #              None otherwise
    #     :note For performance reasons, this function does not check if the points lie inside the grid
    #           The returned coordinate values can thus be out of range.
    #     """
    #
    #     vc = np.zeros((3,), dtype=int)
    #
    #     # Not needed: if x[0] < self.__x0[0] or x[0] > self.__x1[0]: return None
    #     if x[0] == self.__x1[0]:
    #         vc[0] = self.__size[0] - 1
    #     else:
    #         vc[0] = (x[0] - self.__x0[0]) / self.__dx[0]
    #
    #     # Not needed: if x[1] < self.__x0[1] or x[1] > self.__x1[1]: return None
    #     if x[1] == self.__x1[1]:
    #         vc[1] = self.__size[1] - 1
    #     else:
    #         vc[1] = (x[1] - self.__x0[1]) / self.__dx[1]
    #
    #     # Not needed: if x[2] < self.__x0[2] or x[2] > self.__x1[2]: return None
    #     if x[2] == self.__x1[2]:
    #         vc[2] = self.__size[2] - 1
    #     else:
    #         vc[2] = (x[2] - self.__x0[2]) / self.__dx[2]
    #
    #     return vc[0] + self.__size[0] * (vc[1] + self.__size[1] * vc[2])

    def voxelCoordinates2VoxelIndex(self, vc):
        """Computes the (linear) voxel index of a (three-integer) voxel coordinate

        .. note::
            This function checks if the point lies in the grid and throws an exception if this is
            not the case

        :param vc: Voxel coordinates (integer3)
        :returns: Voxel index (integer)
        """
        # Performance: Enabling validate makes the code 1.5x slower
        # y = validateIntVector3(i, "i")
        # if np.any(y < 0) or np.any(self.__size - y <= 0):
        #     raise ValueError("Invalid voxel coordinates (out-of-range)")
        # return y[0] + self.__size[0] * (y[1] + self.__size[1] * y[2])

        # Performance: This variant is significantly faster
        if not self.isVoxelCoordinateInside(vc):
            raise ValueError("Voxel coordinates out-of-range")
        return vc[0] + self.__size[0] * (vc[1] + self.__size[1] * vc[2])

    def voxelIndexToVoxelCoordinate(self, vi):
        """Computes the (three-integer) voxel coordinate of a (linear) voxel index
        :param vi: Voxel index (integer)
        :returns Voxel coordinates (integer3)
        """
        if vi < 0 or vi >= self.nvoxels:
            raise ValueError("Voxel index out-of-range")

        i = np.zeros((3,), dtype=np.int_)
        r = vi
        i[0] = r % self.__size[0]
        r /= self.__size[0]
        i[1] = r % self.__size[1]
        i[2] = r / self.__size[1]
        return i

    def asVoxelIndex(self, index):
        """Auto-convert a voxel coordinate into a voxel index"""
        if isinstance(index, (int, np.int32, np.int64)):
            # Already a linear voxel index...
            return index
        else:
            # Got a voxel coordinate, try to convert it...
            return self.voxelCoordinates2VoxelIndex(index)

    # --= Range checks =--

    def isPointInside(self, x):
        """Returns if a point in space lies within the grid"""
        return (
            self.__x0[0] <= x[0] <= self.__x1[0]
            and self.__x0[1] <= x[1] <= self.__x1[1]
            and self.__x0[2] <= x[2] <= self.__x1[2]
        )

    def isVoxelCoordinateInside(self, vc):
        """Returns if a (three-integer) voxel coordinate lies inside the grid"""
        return (
            0 <= vc[0] < self.__size[0]
            and 0 <= vc[1] < self.__size[1]
            and 0 <= vc[2] < self.__size[2]
        )

    # --= Voxel geometry =--

    def getVoxelCorner(self, vc):
        """Returns the spatial corner point of a voxel coordinate"""
        return np.array(
            [
                self.__x0[0] + vc[0] * self.__dx[0],
                self.__x0[1] + vc[1] * self.__dx[1],
                self.__x0[2] + vc[2] * self.__dx[2],
            ]
        )

    def getVoxelCenter(self, vc):
        """Returns the spatial center point of a voxel coordinate"""
        return self.getVoxelCorner(vc) + self.__dx / 2

    # TODO: Check if really needed...
    def getVoxelAxialNeighbors(self, i):
        """Returns axial neighbors of a voxel coordinate
        :param i: Voxel coordinate (integer-3)
        :returns list(integer-3) Neighbor voxel coordinates
        """
        n = []

        def neighbor(d0, d1, d2):
            j = i + np.array([d0, d1, d2], dtype=np.int_)
            if self.isVoxelCoordinateInside(j):
                n.append(j)

        neighbor(-1, 0, 0)
        neighbor(+1, 0, 0)
        neighbor(0, -1, 0)
        neighbor(0, +1, 0)
        neighbor(0, 0, -1)
        neighbor(0, 0, +1)

        return n

    def getNeighbors(self, i):
        """Returns all neighbors of a voxel (axial, diagonal, side-diagonal)
        :param i: Voxel coordinate (integer-3)
        :returns list(integer-3) Neighbor voxel coordinates
        """
        n = []
        for d0 in [-1, 0, +1]:
            for d1 in [-1, 0, +1]:
                for d2 in [-1, 0, +1]:
                    if d0 == 0 and d1 == 0 and d2 == 0:
                        continue
                    j = i + np.array([d0, d1, d2], dtype=np.int_)
                    if self.isVoxelCoordinateInside(j):
                        n.append(j)

        return n

    # --= Voxelization functions =--

    # TODO: Move into voxelscene. Wrong place here...
    def voxelizeTriangle(self, v0, v1, v2, verbose=False):
        """Returns all voxels that are intersected by a triangle"""

        # Map the triangle points to voxel indices
        vi0 = self.voxelCoordinates2VoxelIndex(self.point2VoxelCoordinates(v0))
        vi1 = self.voxelCoordinates2VoxelIndex(self.point2VoxelCoordinates(v1))
        vi2 = self.voxelCoordinates2VoxelIndex(self.point2VoxelCoordinates(v2))

        # Scan neighboring voxels successively
        # Rule: If a voxel is intersected, also examples all of its neighbors
        voxels_intersect = set([vi0, vi1, vi2])
        voxels_scanned = set()
        voxels_todo = set([vi0])

        ntests = 0

        while voxels_todo:
            voxels_next = set()
            for vi in voxels_todo:
                voxels_scanned.add(vi)
                vc = self.voxelIndexToVoxelCoordinate(vi)

                if triBoxOverlap(self.getVoxelCenter(vc), self.edge_length / 2, [v0, v1, v2]):
                    ntests += 1
                    # This voxel is intersected. Now we have to check all unvisited neighbors.
                    voxels_intersect.add(vi)

                    # Important: Checking only the axial neighbors first,
                    # reduces the number of intersection examples by almost a factor of two.
                    neighbors = [
                        self.voxelCoordinates2VoxelIndex(x) for x in self.getVoxelAxialNeighbors(vc)
                    ]
                    # neighbors = [self.voxelCoordinates2VoxelIndex(x) for x in self.getNeighbors(vc)]
                    for vj in neighbors:
                        # Neighbor already visited?
                        if not vj in voxels_scanned:
                            voxels_next.add(vj)

            voxels_todo = voxels_next

        assert len(voxels_intersect) > 0
        assert vi0 in voxels_intersect
        assert vi1 in voxels_intersect
        assert vi2 in voxels_intersect

        if verbose:
            print("Voxel-triangle intersection examples: %d" % ntests)
            print("Voxels intersected: %d" % len(voxels_intersect))

        return voxels_intersect

    # TODO: Move into voxelscene. Wrong place here...
    def voxelizeLine(self, v0, v1, probe_neighbors=False):
        """Returns all voxels that are intersected by a line
        :param v0:  vertex of the line start point in Cartesian coordinates
        :param v1:  vertex of the line end point in Cartesian coordinates
        :returns:   set(integer), set of intersected voxel IDs
        """
        v0 = np.array(v0)
        v1 = np.array(v1)

        # map the line's start and end point to voxel coordinates
        vc0 = self.point2VoxelCoordinates(v0)
        vc1 = self.point2VoxelCoordinates(v1)

        # add voxel indices of line's start and end point to the intersection set
        vi0 = self.voxelCoordinates2VoxelIndex(vc0)
        vi1 = self.voxelCoordinates2VoxelIndex(vc1)
        voxels_intersect = set([vi0, vi1])

        # loop over all 3 dimensions
        for dim in range(3):
            # skip this dimension, if both points are in the same voxel level
            if vc0[dim] == vc1[dim]:
                continue

            # sort start and end point according to the current dimension
            if vc0[dim] < vc1[dim]:
                v_start = v0
                v_end = v1
                l_start = vc0[dim]
                l_end = vc1[dim]
            else:
                v_start = v1
                v_end = v0
                l_start = vc1[dim]
                l_end = vc0[dim]

            v_diff = v_end - v_start
            num_levels = l_end - l_start

            probe = list()
            if probe_neighbors:
                delta = [0.001, 0.001, 0.001]
                delta[dim] = (
                    self.__dx[dim] / 2
                )  # normal to the intersection plane with half voxel length
                for s0 in [-1, 1]:
                    for s1 in [-1, 1]:
                        for s2 in [-1, 1]:
                            v = np.array([s0 * delta[0], s1 * delta[1], s2 * delta[2]])
                            probe.append(v)
            else:
                delta = np.array([0.0, 0.0, 0.0])
                delta[dim] = (
                    self.__dx[dim] / 2
                )  # normal to the intersection plane with half voxel length
                probe.append(delta)
                probe.append(-delta)

            plane_dist = (
                self.__x0[dim] + l_start * self.__dx[dim]
            )  # distance of the first intersection plane to the origin (in world coordinates)
            alpha = (plane_dist - v_start[dim]) / v_diff[dim]
            x = (
                v_start + alpha * v_diff
            )  # intersection point between the line and the first intersection plane
            inc = (
                v_diff * self.__dx[dim] / v_diff[dim]
            )  # vector increment to the next intersection plane

            # loop over all intersection planes with intersection point x
            for n in range(num_levels):
                x += inc
                for p in probe:
                    try:
                        vc = self.point2VoxelCoordinates(x + np.array(p))
                        vi = self.voxelCoordinates2VoxelIndex(vc)
                        voxels_intersect.add(vi)
                    except ValueError:
                        pass

        return voxels_intersect

    def __str__(self):
        return "VoxelGrid {\n\torigin = %s,\n\tspacing = %s,\n\tsize = %dx%dx%d (%d voxels)\n}" % (
            positionToString(self.__x0),
            positionToString(self.__dx),
            self.__size[0],
            self.__size[1],
            self.__size[2],
            self.__nvoxels,
        )
