# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import multiprocessing

import numpy as np
from joblib import Parallel, delayed

from mpegi.metadata.eif.utils import isVector3

from .voxeldata import VoxelData
from .voxelgrid import VoxelGrid


def voxelize(trimesh, edge_length=1, verbose=False, parallel=False):
    """
    Voxelizes a Trimesh and creates a convenient representation of it.

    Fragen in welchem Voxel eine Quelle ist
    Fragen welche Voxel ein Mesh abdeckt, u.A.
    Frage was ist in einem Voxel drinne...

    :param eifplus:     AudioScene (vector-3)
    :param edge_length: Voxel edge length (>0) in each dimension (scalar or vector-3)
    :param verbose:     Debug output?
    :param parallel:    Multithreading?
    """

    if isVector3(edge_length):
        edge_length = np.array(edge_length)
    else:
        edge_length = edge_length * np.ones((3,))

    # Calculate the bounds of the entire eifplus and set up the voxel grid
    span = np.abs(trimesh.bounds[1] - trimesh.bounds[0])
    size = np.ceil(span / edge_length)
    x0 = 0.5 * (trimesh.bounds[0] + trimesh.bounds[1] - size * edge_length)

    grid = VoxelGrid(x0, edge_length, size)
    if verbose:
        print(grid)

    # Create an empty voxel dataset
    data = VoxelData(grid)

    # Voxelize all meshes
    def voxelizeFace(fi):
        f = trimesh.faces[fi]
        v0 = trimesh.vertices[f[0]]
        v1 = trimesh.vertices[f[1]]
        v2 = trimesh.vertices[f[2]]
        return grid.voxelizeTriangle(v0, v1, v2)

    if not parallel:
        faceVoxels = [None] * len(trimesh.faces)
        for fi in range(len(trimesh.faces)):
            if verbose:
                print("Voxelizing face #%d of #%d" % (fi + 1, len(trimesh.faces)))
            faceVoxels[fi] = voxelizeFace(fi)
    else:
        num_cores = multiprocessing.cpu_count()
        faceVoxels = Parallel(n_jobs=num_cores)(
            delayed(voxelizeFace)(fi) for fi in range(len(trimesh.faces))
        )

    # This is executes single-threaded only
    for fi in range(len(trimesh.faces)):
        for vi in faceVoxels[fi]:
            vd = data.get(vi)
            if not "fids" in vd:
                vd["fids"] = []
            vd["fids"].append(fi)

    return grid, data
