# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
#  Voxelizer component of the encoder
#

import logging
import numpy as np
from mpegi.metadata.eif.utils import positionToString
from mpegi.geometry.transforms import scaleMatrix_4x4, translationMatrix_4x4
from mpegi.geometry.trimesh import convertGeometriesToTrimesh
import trimesh
from tqdm import tqdm
from typing import Dict


#
#  TODO:
#  - How to configure voxel resolution form cli?
#


class Voxelizer:
    def __init__(self, args, artifacts):
        logging.getLogger().setLevel(logging.INFO)
        self.artifacts = artifacts

        # Default settings
        self.__voxel_edge_length = np.array([1, 1, 1])
        self.__listener_height_range = np.array([1.5, 2.0])

    @property
    def voxel_edge_length(self):
        return self.__voxel_edge_length

    @voxel_edge_length.setter
    def voxel_edge_length(self, voxel_edge_length):
        if np.isscalar(voxel_edge_length):
            self.__voxel_edge_length = np.ones((3,)) * voxel_edge_length
        else:
            self.__voxel_edge_length = np.array(voxel_edge_length)

    @property
    def listener_height_range(self):
        """Height range (in meters) above the ground that the listener can reach
        [min_height, max_height]
        """
        return self.__listener_height_range

    @listener_height_range.setter
    def listener_height_range(self, height_range):
        self.__listener_height_range = height_range

    def voxelize(self):
        """Voxelizes the scene provided by the encoder environment"""

        # Get the acoustically-relevant geometry as one big Trimesh
        scene_trimesh_DIFFR_as_eif_mesh = self.artifacts["scene_trimesh_DIFFR_as_eif_mesh"]
        self.mesh = convertGeometriesToTrimesh(
            [scene_trimesh_DIFFR_as_eif_mesh], faceAngle=30, process=False
        )[0]
        origin = np.copy(self.mesh.bounds[0])
        extent = np.copy(self.mesh.extents)

        # LATER:
        # Trimesh sometimes did not include vertices on the scene boundary into the outer boundary voxels.
        # Hence we voxelize with a little safety margin here.
        # epsilon = 0.001
        # origin -= epsilon * np.ones((3,))
        # extent += 2 * epsilon * np.ones((3,))

        logging.info(f"Scene origin: {positionToString(origin)}")
        logging.info(f"Scene extent: {positionToString(extent)}")

        edge_length = self.voxel_edge_length

        # Handle automatic values
        # 0 indicates only one voxel layer in this axis direction
        for i in range(3):
            if edge_length[i] == 0:
                edge_length[i] = extent[i]

        logging.info(f"Voxel edge lengths: {positionToString(edge_length)}")

        # Number of voxels per dimension
        size = np.ceil(extent / edge_length).astype(np.int_)
        nvoxels = np.prod(size)

        logging.info(f"Voxel grid size: {size[0]}x{size[1]}x{size[2]} ({nvoxels} voxels)")

        #
        # Standard function used by trimesh for mesh voxelization: voxelize_subdivide
        # Unfortunately, this function does not allow to specify a custom origin and
        # independent pitch (individual voxel edge lengths along each dimension).
        #
        # We achieve this by first translating/scaling the model accordingly and
        # then assigning in inverse transform to Trimesh's voxel grid later
        #
        # 1. Translate the mesh so that the origin is [0, 0, 0]
        # 2. Then scale it to match pseudo-uniform voxels of size [1, 1, 1]
        #

        transform = np.matmul(scaleMatrix_4x4(1 / edge_length), translationMatrix_4x4(-origin))
        inverse_transform = np.linalg.inv(transform)
        self.mesh.apply_transform(transform)

        # Now voxelize the mesh (uniform pitch)
        self.voxelGrid = self.mesh.voxelized(pitch=1)
        self.voxelGrid.transform = inverse_transform

        # Debug:
        # print("Origin: ", self.grid.origin)
        # print("Scale: ", self.grid.scale)
        # print("Pitch: ", self.grid.scale)
        # print("Voxel volume: ", self.grid.element_volume)

        #
        #  Examples:
        #  grid.points - Center points of all populated voxels
        #                Get the n'th voxel center: grid.points[n,:]
        #

        self.sceneVoxels = self.voxelGrid

        # First attempt for reachability: All unpopulated voxels
        reachable = np.logical_not(self.voxelGrid.matrix)
        reachable_encoded = trimesh.voxel.encoding.SparseEncoding.from_dense(reachable)
        self.reachableVoxels = trimesh.voxel.VoxelGrid(
            reachable_encoded, transform=self.voxelGrid.transform
        )

    def analyzeReachablity(self):
        """Analyzes the (user) reachable voxels in the scene"""
        logging.info("Analyzing reachable voxels")

        def first_true(a):
            for i in range(len(a)):
                if a[i] == True:
                    return i
            return None

        def next_false(a, i):
            if i >= len(a):
                return None
            for i in range(len(a)):
                if a[i] == True:
                    return i
            return None

        reachable = np.zeros_like(self.voxelGrid.matrix, dtype=np.bool)

        listener_height = [0.0, 2.0]
        num_height_voxels = int(
            np.ceil((listener_height[1] - listener_height[0]) / self.voxelGrid.pitch[1])
        )

        # Scan each y-axis column along increasing heights
        for x in range(self.voxelGrid.shape[0]):
            for z in range(self.voxelGrid.shape[2]):
                # Get the column
                c = self.voxelGrid.matrix[x, :, z]
                # print(f"Column {x},{z}: Fields = {c}")

                # Find the first populated voxel in this column...
                g = None
                for i in range(0, len(c) - 1):
                    if c[i] == True:
                        g = i
                        break

                if g is None:
                    # If all voxels are unpopulated in this column, there is no ground
                    # (a listener could stand on) and the entire column is unreachable
                    continue

                # Now find the first next voxel which is empty
                a = None
                for i in range(g + 1, len(c) - 1):
                    if c[i] == False:
                        a = i
                        break

                if a is None:
                    # Not further empty voxel => we're done with this column
                    continue

                p = [i, None]

                # Find the next voxel which is solid again
                g = None
                for i in range(a + 1, len(c) - 1):
                    if c[i] == True:
                        g = i
                        break

                if g:
                    p[1] = g - 1
                else:
                    p[1] = len(c) - 1

                # print(f"Column {x},{z}: Range = {p}")

                # Mark those voxels in the height range above the ground as reachable
                reachable[x, p[0] : min(p[0] + num_height_voxels, p[1]), z] = True

        unpopulated = np.logical_not(self.voxelGrid.matrix)
        avx = trimesh.voxel.encoding.SparseEncoding.from_dense(reachable)
        avg = trimesh.voxel.VoxelGrid(avx, transform=self.voxelGrid.transform)
        # hhh = avg.hollow()
        # avg.show()
        # self.voxelGrid.show()
        avg.show()

        # airVoxelGrid = self.voxelGrid.copy()
        # airVoxelGrid.matrix = unpopulated

        pass
