# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import time
import logging
import trimesh
import numpy as np
from tqdm import tqdm
from typing import List, Dict

from mpegi.metadata.diffraction.enumclasses import Constants


class VoxelizerDIFFR:
    def __init__(
        self,
        sceneDict: Dict,
        args: Dict,
    ):

        # Set default attributes from arguments
        self._sceneDict = sceneDict
        self._maxIter = args["max_iter_DIFFR"]

        # Dummy value for AR scenes
        self._voxelGrid_earLevel_DIFFR = None
        self._voxelPositions_earLevel_DIFFR = None
        self._voxelResolution_earLevel_DIFFR = np.array([0.5, 0.5, 0.5])

        # For AR scene or empty sceneTrimesh_DIFFR, skip voxelize process
        if self._sceneDict["_flag_isARScene"] or self._sceneDict["_flag_emtpySceneTrimesh_DIFFR"]:
            return

        # Ear-level voxel grid for the listener
        if not self._sceneDict["_flag_acEnvExists"] and self._sceneDict["_flag_allSourcesDynamic"]:
            self._voxelResolution_earLevel_DIFFR = np.array([1.0, 1.0, 1.0])
        try:
            self._voxelGrid_DIFFR = self._sceneDict["sceneTrimesh_DIFFR"].voxelized(
                pitch=self._voxelResolution_earLevel_DIFFR[0], max_iter=self._maxIter
            )
        except ValueError as error:
            if str(error) == "max_iter exceeded!":
                raise ValueError(
                    str(error)
                    + f" Value was {self._maxIter}. Try increasing it with the --max_iter_DIFFR option of the encode command."
                )
            else:
                raise error
        # Set the possible listener ear level
        self.__earRange = np.array([1.0, 1.8])
        self.__earRangeVoxelIndexRangeFromGL = (
            self.__earRange / self._voxelResolution_earLevel_DIFFR[1]
        )

        self.analyzeEarLevelVoxelGrid_DIFFR()
        self.updateEarLevelVoxelPositions_DIFFR()

    def analyzeEarLevelVoxelGrid_DIFFR(self):
        """voxelize to return voxelGrid where a listener can walk/jump (not fly) around"""

        shape = self._voxelGrid_DIFFR.matrix.shape
        self.voxelGridMatrix_DIFFR = np.zeros(shape, dtype=bool)

        for idx_x in tqdm(
            range(shape[0]),
            desc=f">> Ear-Level-Voxel-Generation [VoRes: {self._voxelResolution_earLevel_DIFFR}]",
        ):
            for idx_z in range(shape[2]):
                column_xz = self._voxelGrid_DIFFR.matrix[idx_x, :, idx_z]
                unoccupiedVRY = self.unoccupiedVoxelRangeInYAxis(column_xz)

                if len(unoccupiedVRY) == 0:
                    continue

                for uvr in unoccupiedVRY:
                    if (uvr[1] - uvr[0]) < self.__earRangeVoxelIndexRangeFromGL[0]:
                        continue

                    uvIndexList = range(uvr[1] - uvr[0] + 1)
                    for voxel_offset in uvIndexList:
                        if (self.__earRangeVoxelIndexRangeFromGL[0] < voxel_offset + 0.5) and (
                            self.__earRangeVoxelIndexRangeFromGL[1] > voxel_offset - 0.5
                        ):
                            idx_y = uvr[0] + voxel_offset

                            if idx_y > (shape[1] - 1):
                                raise ValueError("ERROR: idx_y can't be larger than shape[1] - 1")

                            self.voxelGridMatrix_DIFFR[idx_x, idx_y, idx_z] = True
                            self.__occupiedVoxels = True

        earLevelVoxel_encodedHR = trimesh.voxel.encoding.SparseEncoding.from_dense(
            self.voxelGridMatrix_DIFFR
        )
        self._voxelGrid_earLevel_DIFFR = trimesh.voxel.VoxelGrid(earLevelVoxel_encodedHR)
        self._voxelGrid_earLevel_DIFFR.apply_scale(self._voxelGrid_DIFFR.pitch)
        self._voxelGrid_earLevel_DIFFR.apply_translation(
            self._sceneDict["sceneTrimesh_DIFFR"].bounds[0]
        )
        # self._voxelGrid_earLevel_DIFFR.show()

        # Union of AEs and URRs && Voxels on Listener Ear Height/Level
        unconstrainedEarLevelVoxelIndices = self._voxelGrid_earLevel_DIFFR.sparse_indices
        unconstrainedEarLevelVoxelPoints = self._voxelGrid_earLevel_DIFFR.indices_to_points(
            self._voxelGrid_earLevel_DIFFR.sparse_indices
        )

        isIncludedInAEs = np.zeros([len(unconstrainedEarLevelVoxelPoints), 1], dtype=bool)
        if self._sceneDict["_flag_acEnvExists"]:
            for trimesh_AE in self._sceneDict["trimeshes_AcousticEnvironments"]:
                isIncludedInAEs_temp = trimesh_AE.contains(unconstrainedEarLevelVoxelPoints)

                for i, b in enumerate(isIncludedInAEs_temp):
                    isIncludedInAEs[i] = isIncludedInAEs[i] or b

        isIncludedInURRs = np.zeros([len(unconstrainedEarLevelVoxelPoints), 1], dtype=bool)
        if len(self._sceneDict["user_regions"]) != 0:
            for trimesh_URR in self._sceneDict["trimeshes_URRs"]:
                isIncludedInURRs_temp = trimesh_URR.contains(unconstrainedEarLevelVoxelPoints)

                for i, b in enumerate(isIncludedInURRs_temp):
                    isIncludedInURRs[i] = isIncludedInURRs[i] or b

        if not self._sceneDict["_flag_acEnvExists"] and len(self._sceneDict["user_regions"]) == 0:
            logging.info(f" __ no acoustic environments and no URRs")

        isIncluded = [(a or b) for a, b in zip(isIncludedInAEs, isIncludedInURRs)]

        for i, b in enumerate(isIncluded):
            if np.any(np.abs(unconstrainedEarLevelVoxelIndices[i]) >= shape):
                continue

            self.voxelGridMatrix_DIFFR[unconstrainedEarLevelVoxelIndices[i][0]][
                unconstrainedEarLevelVoxelIndices[i][1]
            ][unconstrainedEarLevelVoxelIndices[i][2]] = b

        earLevelVoxel_encodedHR = trimesh.voxel.encoding.SparseEncoding.from_dense(
            self.voxelGridMatrix_DIFFR
        )
        self._voxelGrid_earLevel_DIFFR = trimesh.voxel.VoxelGrid(earLevelVoxel_encodedHR)
        self._voxelGrid_earLevel_DIFFR.apply_scale(self._voxelGrid_DIFFR.pitch)
        self._voxelGrid_earLevel_DIFFR.apply_translation(
            self._sceneDict["sceneTrimesh_DIFFR"].bounds[0]
        )
        # self._voxelGrid_earLevel_DIFFR.show()

    def updateEarLevelVoxelPositions_DIFFR(self):
        self._voxelPositions_earLevel_DIFFR = []

        _flag_visibility_check = self._sceneDict["_flag_allSourcesStatic"] and (
            not self._sceneDict["_flag_isDynamicGeoIncluded"]
        )
        srcLocations = [s.position for s in self._sceneDict["objectSources"]]

        for vi in self._voxelGrid_earLevel_DIFFR.sparse_indices:
            vLoc = self._voxelGrid_earLevel_DIFFR.indices_to_points(np.array(vi))

            _update_voxel_position = False
            if _flag_visibility_check:
                vec_tp = np.ones((len(srcLocations), 1)) * vLoc
                vec_t2s = srcLocations - vec_tp
                vec_t2s_distance = np.linalg.norm(vec_t2s, axis=1)
                (coord_xs, idx_srcs, idx_xfaceIDs,) = self._sceneDict[
                    "sceneTrimesh_DIFFR"
                ].ray.intersects_location(
                    ray_origins=vec_tp, ray_directions=vec_t2s, multiple_hits=False
                )

                veIDsFromThis = []
                vec_t2x = coord_xs - np.ones((len(coord_xs), 1)) * vLoc
                vec_t2x_distance = np.linalg.norm(vec_t2x, axis=1)
                for idx_src, t2x_distance in zip(idx_srcs, vec_t2x_distance):
                    if t2x_distance < vec_t2s_distance[idx_src] - Constants.PrecisionErrorThreshold:
                        _update_voxel_position = True
                        break
            else:
                _update_voxel_position = True

            if _update_voxel_position:
                self._voxelPositions_earLevel_DIFFR.append(vLoc)

    @staticmethod
    def withinEnv(point: List[float], bounds: List[List[float]], ndim: int = 3) -> bool:
        insideX = (bounds[0][0] - point[0]) * (bounds[1][0] - point[0]) <= 0
        insideY = (bounds[0][1] - point[1]) * (bounds[1][1] - point[1]) <= 0
        insideZ = (bounds[0][2] - point[2]) * (bounds[1][2] - point[2]) <= 0

        if ndim == 3:
            return insideX and insideY and insideZ
        elif ndim == 2:
            return insideX and insideZ
        else:
            raise ValueError(">> ERROR: ndim = [2, 3]")

    @staticmethod
    def unoccupiedVoxelRangeInYAxis(column_xz):
        yIdx_init = 0
        __floorExists = False
        unoccupiedVoxelRangeinYaxis = []
        for idx_y, filled in enumerate(column_xz):
            if (not filled) and (not __floorExists):
                yIdx_init = idx_y
                __floorExists = True

            if filled and __floorExists:
                end_idx_y = idx_y - 1
                if yIdx_init != 0:
                    unoccupiedVoxelRangeinYaxis.append((yIdx_init, end_idx_y))
                __floorExists = False

        if __floorExists and (yIdx_init != 0):
            unoccupiedVoxelRangeinYaxis.append((yIdx_init, len(column_xz) - 1))

        return unoccupiedVoxelRangeinYaxis

    @property
    def voxelGrid_earLevel_DIFFR(self):
        return self._voxelGrid_earLevel_DIFFR

    @property
    def voxelPositions_earLevel_DIFFR(self):
        return self._voxelPositions_earLevel_DIFFR

    @property
    def voxelResolution_earLevel_DIFFR(self):
        return self._voxelResolution_earLevel_DIFFR
