# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
#
#  Voxelizer component of the encoder
#

import logging
import numpy as np
from mpegi.voxel.voxelutils import voxelCoordinateToPoint
from mpegi.geometry.trimesh import convertToTrimesh, convertGeometriesToTrimesh
from mpegi.geometry import transforms
import trimesh
from tqdm import tqdm


class VoxelizerER:
    """Voxelizer for early reflections
    :var reachableVoxels:           VoxelGrid, voxels reachable by the listener
    :var __listener_height_range:   [int, int], range of the listener's ear level
    """

    def __init__(self, args, artifacts):
        """Constructor
        :param args:                dict, input parameters
        :param artifacts:           dict, output parameters
        """
        logging.getLogger().setLevel(logging.INFO)

        # Default settings
        self.__listener_height_range = np.array([1.40, 1.90])

        # extract boundaries of the acoustic environments
        geometry_dict = dict()
        for ae_mesh in artifacts["meshes"]:
            geometry_dict[ae_mesh.id] = ae_mesh

        geometry_list = [
            geometry_dict[env.region]
            for env in artifacts["acoustic_environments"]
            if env.region in geometry_dict
        ]
        self.ae_trimesh_list = convertToTrimesh(geometry=geometry_list, includeTransparent=True)
        for geometry, trimesh in zip(geometry_list, self.ae_trimesh_list):
            # Convert geometry to trimesh and apply translation
            tform = transforms.transformMatrixTRS_4x4(
                geometry.position, np.deg2rad([0, 0, 0]), [1, 1, 1]
            )
            trimesh.apply_transform(tform)

    @property
    def listener_height_range(self):
        """Height range (in meters) above the ground that the listener can reach
        [min_height, max_height]
        """
        return self.__listener_height_range

    @listener_height_range.setter
    def listener_height_range(self, height_range):
        self.__listener_height_range = height_range

    def analyze_listener_voxels(self, mesh, voxel_size, user_regions):
        """Voxelizes the scene provided by the encoder environment and stores it in self.reachableVoxels
        :var mesh:                      Trimesh, geometry data
        :var voxel_size:                Float3, voxel size in m
        :var user_regions:              list(UserRegion): list of UserRegion objects that describe the user reachable areas
        """
        logging.info("Generating voxel grid")

        # enforce vector format
        if np.isscalar(voxel_size):
            voxel_size = np.ones((3,)) * voxel_size
        else:
            voxel_size = np.array(voxel_size)

        # align voxel grid
        origin, shape = self.__get_aligned_grid(mesh, voxel_size)

        # determine reachable voxels
        reachable_flags = self.__get_reachable_flags(mesh, voxel_size, origin, shape)

        # limit reachable voxels to user regions
        reachable_flags &= self.__get_user_region_flags(
            reachable_flags, user_regions, voxel_size, origin, shape
        )

        # generate VoxelGrid
        voxel_encoding = trimesh.voxel.encoding.DenseEncoding(reachable_flags)
        self.reachableVoxels = trimesh.voxel.VoxelGrid(encoding=voxel_encoding)
        self.reachableVoxels.apply_scale(voxel_size)
        self.reachableVoxels.apply_translation(origin)

        # Debug: plot reachable voxels
        # self.__plot_voxels(mesh, self.reachableVoxels)

        # Number of voxels per dimension
        size = np.array(self.reachableVoxels.shape)
        nvoxels = np.prod(size)
        logging.info(f"Voxel grid size: {size[0]}x{size[1]}x{size[2]} ({nvoxels} voxels)")
        logging.info(f"Number of reachable voxels: {self.reachableVoxels.filled_count}")

    def __get_aligned_grid(self, mesh, voxel_size):
        """Return axis aligned voxel grid parameters
        :var mesh:                  Trimesh, geometry data
        :var voxel_size:            Float3, voxel size in m
        :return:                    (Float3, Int3), center position of voxel [0,0,0] and number of voxels in each dimension
        """

        # headroom: literally a "head" room
        delta = 0.001  # just a small number on top of the headroom
        headroom = np.zeros((3,))
        headroom[1] += self.__listener_height_range[1] + delta

        # initial axis alignment
        offset = np.zeros((3,))
        for dim in range(3):
            offset[dim] = mesh.bounds[0, dim] - self.__get_aligned_axis(mesh, dim, voxel_size)

        # extract floor height and find best vertical alignment
        floor_height = self.__get_floor_height(mesh)
        if floor_height.size > 0:
            # test various y-offsets
            granularity = 100
            offset_cost = np.zeros((granularity,), dtype=int)
            for n in range(granularity):
                offset_y = voxel_size[1] * n / granularity
                origin_y = mesh.bounds[0, 1] + voxel_size[1] / 2 - offset_y
                ny1 = np.round(
                    (floor_height + self.__listener_height_range[0] - origin_y) / voxel_size[1]
                ).astype(int)
                ny2 = (
                    np.round(
                        (floor_height + self.__listener_height_range[1] - origin_y) / voxel_size[1]
                    ).astype(int)
                    + 1
                )
                offset_cost[n] = np.sum(ny2 - ny1)

            # select alignment offset with minimum cost
            n_best = self.__get_minimum_pos(offset_cost)
            offset[1] = voxel_size[1] * n_best / granularity

        # shape: number of voxels in each dimension
        extent = mesh.extents + offset + headroom
        shape = np.ceil(extent / voxel_size).astype(int)

        origin = mesh.bounds[0] + voxel_size / 2 - offset

        return origin, shape

    def __get_aligned_axis(self, mesh, dim, voxel_size):
        """Return axis alignment for 1 dimension
        :var mesh:                  Trimesh, geometry data
        :var dim:                   int, dimension
        :var voxel_size:            Float3, voxel size in m
        :return:                    float, proposed origin value for this dimension
        """
        thresh = np.cos(2 * np.pi / 180)  # maximum tilt of 2°
        parallel_faces = np.nonzero(mesh.face_normals[:, dim] > thresh)
        parallel_mesh = mesh.submesh(parallel_faces, append=True)
        if type(parallel_mesh) != trimesh.Trimesh:  # no floor? -> use whole mesh
            parallel_mesh = mesh
        val = parallel_mesh.bounds[0, dim]
        while val > mesh.bounds[0, dim]:
            val -= voxel_size[dim]
        return val

    def __get_floor_height(self, mesh):
        # Import done here because pyembree is required for this class
        # Building pyembree on windows is a pain - not everyone needs this class
        from trimesh.ray.ray_pyembree import RayMeshIntersector

        """Return a list of floor points
        :var mesh:                  Trimesh, geometry data
        :return:                    numpy.array(N,), array with height values of floor points within the acoustic environments
        """
        logging.info("Extracting floor height")

        # declaration of the return value
        height = np.zeros((0,), dtype=float)

        for trimesh in self.ae_trimesh_list:
            # generate parameters for ray-tracing
            xv = np.arange(trimesh.bounds[0, 0] + 0.5, trimesh.bounds[1, 0])
            zv = np.arange(trimesh.bounds[0, 2] + 0.5, trimesh.bounds[1, 2])
            xv, zv = np.meshgrid(xv, zv)  # linear 2D grid with 1m spacing
            xv = xv.flatten()
            zv = zv.flatten()
            yv = (
                np.zeros(xv.shape) + (trimesh.bounds[0, 1] + trimesh.bounds[1, 1]) / 2
            )  # height: mean between minimum and maximum height
            ray_start = np.array([xv, yv, zv]).transpose()
            ray_start = ray_start[
                trimesh.contains(ray_start), :
            ]  # make sure that the ray starting points are really inside the region of the acoustic environment
            ray_dir = np.zeros(ray_start.shape)
            ray_dir[:, 1] = -1

            # find floor points via ray-tracing
            intersector = RayMeshIntersector(mesh)
            posArray, rayArray, triArray = intersector.intersects_location(
                ray_start, ray_dir, multiple_hits=False
            )
            height = np.concatenate((height, posArray[:, 1]))

        return height

    def __get_minimum_pos(self, cost):
        """Return array index with minimum value (mean index for consecutive minimum values).
        :var cost:                  numpy.array(N), array with cost values
        :return:                    int, array index with minimum value
        """
        min_cost = np.min(cost)
        mask = np.concatenate((cost, cost)) == min_cost  # repeat cost array to handle wrap around

        # find longest sequence of minimum values
        flip_pos = [n + 1 for n in range(len(mask) - 1) if mask[n + 1] != mask[n]]
        optim_len = 0
        optim_pos = 0
        for k in range(len(flip_pos) - 1):
            n = flip_pos[k]
            if not mask[n]:
                continue
            curr_len = flip_pos[k + 1] - flip_pos[k]
            if curr_len > optim_len:
                optim_len = curr_len
                optim_pos = n

        optim_pos = np.floor(optim_pos + optim_len / 2).astype(
            int
        )  # middle of the minimum sequence
        optim_pos = np.mod(optim_pos, len(cost))  # apply modulo to handle wrap around
        return optim_pos

    def __get_reachable_flags(self, mesh, voxel_size, origin, shape):
        # Import done here because pyembree is required for this class
        # Building pyembree on windows is a pain - not everyone needs this class
        from trimesh.ray.ray_pyembree import RayMeshIntersector

        """Return axis aligned voxel grid parameters
        :var mesh:                  Trimesh, geometry data
        :var voxel_size:            Float3, voxel size in m
        :var origin:                Float3, center position of voxel [0,0,0]
        :var shape:                 Int3, number of voxels in each dimension
        :return:                    numpy.array((X,Y,Z), dtype=bool), binary mask indicating the reachability of the corresponding voxel
        """
        logging.info("Analyzing reachability")

        # memory allocation
        reachable_flags = np.zeros(shape, dtype=bool)

        # initialize Embree ray-mesh-intersectors for floor and ceiling triangles
        floor_thresh = np.cos(30 * np.pi / 180)  # floors may not be steeper than 30°
        ceil_thresh = 0.0
        floor_faces = np.nonzero(mesh.face_normals[:, 1] > floor_thresh)
        ceil_faces = np.nonzero(mesh.face_normals[:, 1] < ceil_thresh)
        floor_mesh = mesh.submesh(floor_faces, append=True)
        floor_mesh = self.__cleanup_trimesh(floor_mesh)  # remove small surfaces
        ceil_mesh = mesh.submesh(ceil_faces, append=True)
        if type(floor_mesh) == trimesh.Trimesh:
            floor_intersector = RayMeshIntersector(floor_mesh)
        if type(ceil_mesh) == trimesh.Trimesh:
            ceil_intersector = RayMeshIntersector(ceil_mesh)

        y_max = mesh.bounds[1, 1] + 1

        num_subpos = 5
        subpos = np.zeros((num_subpos, 3))
        subpos[:, 0] = np.array([0, -1, -1, 1, 1]) * voxel_size[0] * 0.49
        subpos[:, 1] = 0
        subpos[:, 2] = np.array([0, -1, 1, -1, 1]) * voxel_size[2] * 0.49
        subpos_list = list(subpos)

        # memory allocation for voxel coordinate tables and ray parameters
        ray_start = np.zeros((1, 3), dtype=float)
        ray_dir = np.zeros((1, 3), dtype=float)
        ray_dir[:, 1] = -1

        # find valid floor height levels
        for nx in tqdm(range(shape[0])):
            for nz in range(shape[2]):
                for sp in subpos_list:
                    # floor and ceiling ray-mesh intersection tests
                    ny = shape[1]  # center of the first voxel layer above the voxel grid
                    ray_start[0, :] = np.array([nx, ny, nz]) * voxel_size + origin
                    if type(floor_mesh) == trimesh.Trimesh:
                        (
                            floor_posArray,
                            floor_rayArray,
                            floor_triArray,
                        ) = floor_intersector.intersects_location(
                            ray_start, ray_dir, multiple_hits=True
                        )
                    else:
                        floor_posArray, floor_rayArray, floor_triArray = (
                            np.zeros((0, 3), dtype=float),
                            np.zeros((0,), dtype=int),
                            np.zeros((0,), dtype=int),
                        )
                    if floor_posArray.size == 0:
                        continue
                    if type(ceil_mesh) == trimesh.Trimesh:
                        (
                            ceil_posArray,
                            ceil_rayArray,
                            ceil_triArray,
                        ) = ceil_intersector.intersects_location(
                            ray_start, ray_dir, multiple_hits=True
                        )
                    else:
                        ceil_posArray, ceil_rayArray, ceil_triArray = (
                            np.zeros((0, 3), dtype=float),
                            np.zeros((0,), dtype=int),
                            np.zeros((0,), dtype=int),
                        )
                    floor_height = np.sort(floor_posArray[:, 1])
                    ceil_height = ceil_posArray[:, 1]
                    for y in floor_height:
                        dist = ceil_height - y
                        invalid = (dist > 0) & (dist < self.__listener_height_range[1])
                        if dist.size == 0 or not np.any(invalid):
                            ny1 = np.round(
                                (y + self.__listener_height_range[0] - origin[1]) / voxel_size[1]
                            ).astype(int)
                            ny2 = (
                                np.round(
                                    (y + self.__listener_height_range[1] - origin[1])
                                    / voxel_size[1]
                                ).astype(int)
                                + 1
                            )
                            reachable_flags[nx, ny1:ny2, nz] = True
                            break

        return reachable_flags

    def __get_user_region_flags(self, reachable_flags, user_regions, voxel_size, origin, shape):
        """Generate binary mask indicating the reachability of the corresponding voxel
        :var reachable_flags:       numpy.array((X,Y,Z), dtype=bool), binary mask indicating the reachability of the corresponding voxel
        :var user_regions:          list(UserRegion), geometry data
        :var voxel_size:            Float3, voxel size in m
        :var origin:                Float3, center position of voxel [0,0,0]
        :var shape:                 Int3, number of voxels in each dimension
        :return:                    numpy.array((X,Y,Z), dtype=bool), binary mask indicating the reachability of the corresponding voxel
        """
        logging.info("Limit reachable voxels to user regions")

        # No user region? No limitation!
        if len(user_regions) == 0:
            return True

        # intermediate voxel specification
        voxel_encoding = trimesh.voxel.encoding.DenseEncoding(reachable_flags)
        reachableVoxels = trimesh.voxel.VoxelGrid(encoding=voxel_encoding)
        reachableVoxels.apply_scale(voxel_size)
        reachableVoxels.apply_translation(origin)

        # extract list of reachable voxel coordinates inside the user regions
        is_inside = False
        for region in convertGeometriesToTrimesh(user_regions, concatenate=False):
            is_inside |= region.contains(reachableVoxels.points)
        vc_inside = reachableVoxels.sparse_indices[is_inside, :]

        # generate binary 3D mask
        reachable_flags = np.zeros(shape, dtype=bool)
        reachable_flags[vc_inside[:, 0], vc_inside[:, 1], vc_inside[:, 2]] = True

        return reachable_flags

    def __cleanup_trimesh(self, trimesh, area_thresh=1.0, width_thresh=1.0):
        """Remove small surfaces
        :var trimesh:               Trimesh, geometry data
        :var area_thresh:           float, minimum surface area in m^2
        :var width_thresh:          float, minimum width
        :return:                    Trimesh, reduced geometry data
        """
        num_triangle = len(trimesh.faces)
        triangle2facet = -np.ones(num_triangle, dtype=int)
        faces_in_facets = set()
        for f, facet in enumerate(trimesh.facets):
            for t in facet:
                triangle2facet[t] = f
                faces_in_facets.add(t)

        # keep facets with minimum surface area and minimum width
        keep_facets = set()
        for facet in np.nonzero(trimesh.facets_area >= area_thresh)[0]:
            face_list = trimesh.facets[facet]
            sub_mesh = trimesh.submesh([face_list])
            bbox = sub_mesh[0].bounds
            extend = bbox[1] - bbox[0]
            if (
                sum(abs(extend) >= width_thresh) >= 2
            ):  # bounding box must have at least 2 dimensions which are greater than the threshold
                keep_facets.add(facet)

        # map facets to triangle IDs
        keep_triangles = set()
        for facet in keep_facets:
            for triangle in trimesh.facets[facet]:
                keep_triangles.add(triangle)

        # same procedure for all triangles which are not contained in the facets list
        missed_triangles = [t for t in range(num_triangle) if t not in faces_in_facets]
        for triangle in missed_triangles:
            if trimesh.area_faces[triangle] < area_thresh:
                continue
            tri = trimesh.faces[triangle]
            V = trimesh.vertices[tri]
            extend = V.max(axis=0) - V.min(axis=0)
            if (
                sum(extend >= width_thresh) >= 2
            ):  # bounding box must have at least 2 dimensions which are greater than the threshold
                keep_triangles.add(triangle)

        reduced_mesh = trimesh.submesh([list(keep_triangles)], append=True)
        return reduced_mesh
