# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import pathlib
import unittest

from mpegi.cli import Config


class TestConfig(unittest.TestCase):
    def setUp(self):
        testDir = pathlib.Path(__file__).parent
        self.defaultConfigFile = testDir.joinpath("default.conf")
        self.posConfigFile = testDir.joinpath("positionalarg.conf")
        self.optConfigFile = testDir.joinpath("optionalarg.conf")
        self.flagConfigFile = testDir.joinpath("flagarg.conf")
        self.posSubcommandConfigFile = testDir.joinpath("positionalsubcommandarg.conf")
        self.optSubcommandConfigFile = testDir.joinpath("optionalsubcommandarg.conf")
        self.flagSubcommandConfigFile = testDir.joinpath("flagsubcommandarg.conf")
        self.multiSectionConfigFile = testDir.joinpath("multiplesections.conf")

    def test_positionalArgument(self):
        argv = ["bar"]
        config = Config(argv, self.defaultConfigFile)
        config.parser.add_argument("foo")

        self.assertEqual("bar", config["foo"])

    def test_optionalArgument(self):
        argv = ["--bar=foo"]
        config = Config(argv, self.defaultConfigFile)
        config.parser.add_argument("--bar")

        self.assertEqual("foo", config["bar"])

    def test_flagArgument(self):
        argv = ["--foo"]
        config = Config(argv, self.defaultConfigFile)
        config.parser.add_argument("--foo", action="store_true")

        self.assertEqual(True, config["foo"])

    def test_subcommandPositionalArgument(self):
        subcommandString = "mysubcommand"
        argv = [subcommandString, "bar"]
        config = Config(argv, self.defaultConfigFile)
        subparser = config.addSubcommand(subcommandString)
        subparser.add_argument("foo")

        self.assertEqual("bar", config["foo"])

    def test_subcommandOptionalArgument(self):
        subcommandString = "mysubcommand"
        argv = [subcommandString, "--bar=foo"]
        config = Config(argv, self.defaultConfigFile)
        subparser = config.addSubcommand(subcommandString)
        subparser.add_argument("--bar")

        self.assertEqual("foo", config["bar"])

    def test_positionalArgumentWithConfigFile(self):
        config = Config(["-c", str(self.posConfigFile)], self.defaultConfigFile)
        config.parser.add_argument("bar")

        self.assertEqual("foo", config["bar"])

    def test_optionalArgumentWithConfigFile(self):
        config = Config(["-c", str(self.optConfigFile)], self.defaultConfigFile)
        config.parser.add_argument("--foo")

        self.assertEqual("bar", config["foo"])

    def test_flagArgumentWithConfigFile(self):
        config = Config(["-c", str(self.flagConfigFile)], self.defaultConfigFile)
        config.parser.add_argument("--foo", action="store_true")

        self.assertEqual(True, config["foo"])

    def test_positionalArgumentWithConfigFileAndSubcommand(self):
        conf = str(self.posSubcommandConfigFile)
        config = Config(["-c", conf, "mysubcommand"], self.defaultConfigFile)
        subparser = config.addSubcommand("mysubcommand")
        subparser.add_argument("bar")

        self.assertEqual("foo", config["bar"])

    def test_optionalArgumentWithConfigFileAndSubcommand(self):
        conf = str(self.optSubcommandConfigFile)
        config = Config(["-c", conf, "mysubcommand"], self.defaultConfigFile)
        subparser = config.addSubcommand("mysubcommand")
        subparser.add_argument("--foo")

        self.assertEqual("bar", config["foo"])

    def test_flagArgumentWithConfigFileAndSubcommand(self):
        conf = str(self.flagSubcommandConfigFile)
        config = Config(["-c", conf, "mysubcommand"], self.defaultConfigFile)
        subparser = config.addSubcommand("mysubcommand")
        subparser.add_argument("--foo", action="store_true")

        self.assertEqual(True, config["foo"])

    def test_multipleSectionsInConfigFile(self):
        conf = str(self.multiSectionConfigFile)
        config = Config(["-c", conf, "mysubcommand"], self.defaultConfigFile)
        config.addSubcommand("mysubcommand")
        config.addSubcommand("anothersubcommand")

        config.getArgs()


if __name__ == "__main__":
    unittest.main()
