# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Type
import pathlib
import shutil
import sys
import unittest
import unittest.mock as mock

from mpegi import cli, framework

sys.path.append(str(pathlib.Path(__file__).parent))
from mockcomponents import *
import mockcomponents


class TestEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.cacheDir = pathlib.Path(__file__).parent.joinpath(".cache")
        cls.cacheDir.mkdir()
        configFile = pathlib.Path(__file__).parent.joinpath("default.conf").resolve()
        cls.config = cli.Config([], str(configFile))
        FooComponent.registerArguments(cls.config.parser)

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree(cls.cacheDir)

    def tearDown(self) -> None:
        allSubDirs = self.cacheDir.glob("*")
        for subDir in allSubDirs:
            shutil.rmtree(subDir)

    def test_basicComponentBuilder(self):
        components = [FooComponent, BarComponent, BazComponent, FooBarBazComponent]
        builder = framework.ComponentBuilder(components)
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, None)

        artifactsSub = ["primary", "secondary1", "secondary2", "tertiary"]
        artifacts = inventory.getArtifacts(artifactsSub)
        self.assertEqual(artifacts["primary"], "primaryartifact")
        self.assertEqual(artifacts["secondary1"], "secondaryartifact1")
        self.assertEqual(artifacts["secondary2"], "secondaryartifact2")
        self.assertEqual(artifacts["tertiary"], "secondaryartifact1secondaryartifact2")

    def test_unorderedComponentBuilder(self):
        components = [FooBarBazComponent, BazComponent, BarComponent, FooComponent]
        builder = framework.ComponentBuilder(components)
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, None)

        artifactsSub = ["primary", "secondary1", "secondary2", "tertiary"]
        artifacts = inventory.getArtifacts(artifactsSub)
        self.assertEqual(artifacts["primary"], "primaryartifact")
        self.assertEqual(artifacts["secondary1"], "secondaryartifact1")
        self.assertEqual(artifacts["secondary2"], "secondaryartifact2")
        self.assertEqual(artifacts["tertiary"], "secondaryartifact1secondaryartifact2")

    @mock.patch("mockcomponents.BazComponent.__init__", return_value=None)
    @mock.patch("mockcomponents.BarComponent.__init__", return_value=None)
    @mock.patch("mockcomponents.FooComponent.__init__", return_value=None)
    def test_forceConstruction(self, fooInitMock, barInitMock, bazInitMock):
        components = [FooComponent, BarComponent, BazComponent]
        builder = framework.ComponentBuilder(components)

        # First run - Construct all components
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, None)

        self.assertEqual(fooInitMock.call_count, 1)
        self.assertEqual(barInitMock.call_count, 1)
        self.assertEqual(bazInitMock.call_count, 1)

        # Second run - Construct no components because of caching
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, None)

        self.assertEqual(fooInitMock.call_count, 2)
        self.assertEqual(barInitMock.call_count, 1)
        self.assertEqual(bazInitMock.call_count, 1)

        # Third run - Construct all components through force
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(
            inventory,
            [FooComponent.__name__, BarComponent.__name__, BazComponent.__name__],
        )

        self.assertEqual(fooInitMock.call_count, 3)
        self.assertEqual(barInitMock.call_count, 2)
        self.assertEqual(bazInitMock.call_count, 2)

        # Fourth run - Construct all components through force via empty list
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, [])

        self.assertEqual(fooInitMock.call_count, 4)
        self.assertEqual(barInitMock.call_count, 3)
        self.assertEqual(bazInitMock.call_count, 3)

        # Fifth run - Construct only one component through force
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, [FooComponent.__name__])

        self.assertEqual(fooInitMock.call_count, 5)
        self.assertEqual(barInitMock.call_count, 3)
        self.assertEqual(bazInitMock.call_count, 3)

        with self.assertRaises(Exception):
            inventory = framework.Inventory(self.config, str(self.cacheDir))
            builder.process(inventory, ["componentnamethatdoesntexist"])

    def test_componentVersionControl(self):
        # First run - Build v1 component
        components: List[Type[framework.Component]] = [FooComponent, BarComponent]
        builder = framework.ComponentBuilder(components)
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, None)
        artifacts = inventory.getArtifacts(["primary", "secondary1"])
        self.assertEqual(artifacts["primary"], "primaryartifact")
        self.assertEqual(artifacts["secondary1"], "secondaryartifact1")

        # Second run - Build v1 component with new artifact but expect old artifact
        #              because the version number hasn't increased
        artifactGetterImport = "mockcomponents.BarComponent.getArtifacts"
        with mock.patch(artifactGetterImport) as artifactGetterMock:
            artifactGetterMock.return_value = {"secondary1": "secondarychanged"}
            builder = framework.ComponentBuilder(components)
            inventory = framework.Inventory(self.config, str(self.cacheDir))
            builder.process(inventory, None)
            artifacts = inventory.getArtifacts(["secondary1"])
            self.assertEqual(artifacts["secondary1"], "secondaryartifact1")

        # Third run - Build v2 component and expect new artifact
        versionGetterImport = "mockcomponents.BarComponent.getVersion"
        with mock.patch(artifactGetterImport) as artifactGetterMock:
            with mock.patch(versionGetterImport) as versionGetterMock:
                versionGetterMock.return_value = 2  # instead of one
                artifactGetterMock.return_value = {"secondary1": "secondarychanged"}
                builder = framework.ComponentBuilder(components)
                inventory = framework.Inventory(self.config, str(self.cacheDir))
                builder.process(inventory, None)
                artifacts = inventory.getArtifacts(["secondary1"])
                self.assertEqual(artifacts["secondary1"], "secondarychanged")

    def test_cyclicDependency(self):
        components = [BazCyclicComponent, BarCyclicComponent]
        with self.assertRaises(Exception):
            framework.ComponentBuilder(components)

    def test_duplicateArtifacts(self):
        components = [BarDuplicateComponent, BazDuplicateComponent]
        with self.assertRaises(Exception):
            framework.ComponentBuilder(components)

    def test_rootComponentChangeShouldPropogateRegardlessOfVersionNumber(self):
        # First run - Root component produces "A" artifact
        components: List[Type[framework.Component]] = [FooComponent, BarComponent]
        builder = framework.ComponentBuilder(components)
        inventory = framework.Inventory(self.config, str(self.cacheDir))
        builder.process(inventory, None)
        artifacts = inventory.getArtifacts(["primary"])
        self.assertEqual(artifacts["primary"], "primaryartifact")

        # Second run - Root component produces "B" artifact without changing version
        artifactGetterImport = "mockcomponents.FooComponent.getArtifacts"
        with mock.patch(artifactGetterImport) as artifactGetterMock:
            artifactGetterMock.return_value = {"primary": "newPrimaryArtifact"}
            components: List[Type[framework.Component]] = [FooComponent, BarComponent]
            builder = framework.ComponentBuilder(components)
            inventory = framework.Inventory(self.config, str(self.cacheDir))
            builder.process(inventory, None)
            artifacts = inventory.getArtifacts(["primary"])
            self.assertEqual(artifacts["primary"], "newPrimaryArtifact")


if __name__ == "__main__":
    unittest.main()
