# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import unittest

from mpegi.metadata import eifplus


class TestStaticity(unittest.TestCase):
    def test_staticBox(self):
        box = eifplus.Box(
            "box",
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0],
            eifplus.CSpace.fromString("relative"),
            "dummymaterial",
            "",
            "",
        )
        metadata = eifplus.SceneMetadata()

        metadata.analyze([box], [], "")

        self.assertEqual(metadata.isEntityStatic(box.id), True)

    def test_staticChannelSource(self):
        loudspeaker = eifplus.Loudspeaker([1.0, 0.0, 0.0], [0.0, 0.0, 0.0], 0, "", 1.0)
        channelSource = eifplus.ChannelSource(
            [loudspeaker],
            "chanSrc",
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            eifplus.CSpace.Relative,
            0,
            True,
            0.0,
            1.0,
            "dummySignal",
            None,
            "continuous",
            True,
            "",
            "",
            vdlMethod=0,
            hasReverbGain=False,
            reverbGainDb=0.0,
            noReverb=False,
            granularPos=[0.0, 0.0, 0.0],
        )

        metadata = eifplus.SceneMetadata()

        metadata.analyze([channelSource], [], "")

        self.assertEqual(metadata.isEntityStatic(channelSource.id), True)

    def test_dynamicSphere(self):
        dynamicSphere = eifplus.Sphere(
            "dynamicSphere",
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0],
            eifplus.CSpace.fromString("relative"),
            "dummymaterial",
            "",
            "",
        )
        mod = eifplus.Modification(dynamicSphere.id, 1.0, {"position": "1.0 1.0 1.0"})
        updateSphere = eifplus.Update("moveSphere", None, 1.0, None, None, None, "Timed", [mod])
        entities = [dynamicSphere, updateSphere]
        metadata = eifplus.SceneMetadata()

        metadata.analyze(entities, [], "")

        self.assertEqual(metadata.isEntityStatic(dynamicSphere.id), False)

    def test_dynamicObjectSource(self):
        alignment = eifplus.Alignment.fromString("object")
        positioning = eifplus.Positioning.fromString("L C R")
        inputLayout = eifplus.InputLayout(alignment, positioning)
        dynamicObjSrc = eifplus.ObjectSource(
            id="dynamicObjSrc",
            position=[0.0, 0.0, 0.0],
            orientation=[0.0, 0.0, 0.0],
            cspace=eifplus.CSpace.fromString("relative"),
            active=True,
            gainDb=0.0,
            refDistance=1.0,
            signal="dummystream",
            extent="",
            directivity="",
            directiveness=1.0,
            mode="continuous",
            play=True,
            inputLayout=inputLayout,
            aparams=None,
            transform="",
            anchor="",
            vdlMethod=0,
            hasReverbGain=False,
            reverbGainDb=0.0,
            noReverb=False,
            granularPos=[0.0, 0.0, 0.0],
        )
        mod = eifplus.Modification(dynamicObjSrc.id, 1.0, {"orientation": "1.0 1.0 1.0"})
        updateObjSrc = eifplus.Update("moveobj", None, 2.0, None, None, None, "Timed", [mod])
        entities = [dynamicObjSrc, updateObjSrc]
        metadata = eifplus.SceneMetadata()

        metadata.analyze(entities, [], "")

        self.assertEqual(metadata.isEntityStatic(dynamicObjSrc.id), False)

    def test_meshMovedByTransform(self):
        transform = eifplus.Transform(
            "transform", [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], eifplus.CSpace.Relative
        )
        meshInTransform = eifplus.Mesh(
            "meshInTransform",
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            eifplus.CSpace.fromString("relative"),
            [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
            [[0, 1, 2]],
            ["dummymaterial"],
            "transform",
            "",
        )
        mod = eifplus.Modification(transform.id, 1.0, {"position": "1.0 1.0 1.0"})
        updateTransformPosition = eifplus.Update(
            "movetform", None, 3.0, None, None, None, "Timed", [mod]
        )
        entities = [meshInTransform, updateTransformPosition]
        metadata = eifplus.SceneMetadata()

        metadata.analyze(entities, [], "")

        self.assertEqual(metadata.isEntityStatic(meshInTransform.id), False)

    def test_hoaSourceMovedByTransform(self):
        transform = eifplus.Transform(
            "transform", [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], eifplus.CSpace.Relative
        )
        hoaSourceInTransform = eifplus.HOASource(
            "hoaSource",
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            eifplus.CSpace.Relative,
            True,
            0.0,
            1.0,
            "dummySignal",
            None,
            True,
            True,
            1.0,
            None,
            None,
            "continuous",
            True,
            "transform",
            "",
            vdlMethod=0,
            hasReverbGain=False,
            reverbGainDb=0.0,
            noReverb=False,
            granularPos=[0.0, 0.0, 0.0],
        )
        mod = eifplus.Modification(transform.id, 0.0, {"cspace": "user"})
        updateTransformCSpace = eifplus.Update(
            "cspacetformupdate", None, 4.0, None, None, None, "Timed", [mod]
        )
        entities = [hoaSourceInTransform, updateTransformCSpace]
        metadata = eifplus.SceneMetadata()

        metadata.analyze(entities, [], "")

        self.assertEqual(metadata.isEntityStatic(hoaSourceInTransform.id), False)


if __name__ == "__main__":
    unittest.main()
