from pathlib import Path

import pytest

from mpegi.metadata import eif
from mpegi.metadata.eif.acoustTuningParameters import DefaultAE, GlobalAEControl
from xml.dom.minidom import Document, Element


def test_correct_types_when_parsing_tuning_params(tmp_path: Path):
    test_eif = """<?xml version="1.0" encoding="utf-8" standalone="yes"?>
<AudioScene id="my_scene" duration="30">
    <AcousticTuningParameters>
        <VoxDirectOcclusionControl>
            <Setting material="mat:test" value="1" />
        </VoxDirectOcclusionControl>
        <VoxDiffractionControl value="1" />
        <EarlyTuningGainDb value="-4" />
        <DefaultAE enabled="true" environment="Outside" gainDb="-2" density="2.5" delayMinMs="38.19" distanceDropDb="5" minDist="2" />
    </AcousticTuningParameters>
</AudioScene>"""

    tmp_eif = tmp_path.joinpath("my_scene.xml")
    tmp_eif.write_text(test_eif)

    scene = eif.AudioScene.load(str(tmp_eif))

    assert scene.acousticTuningParams.defaultAE.enabled == True
    assert isinstance(scene.acousticTuningParams.defaultAE.enabled, bool)
    assert scene.acousticTuningParams.voxOccCtrl == {"mat:test": 1.0}
    assert isinstance(scene.acousticTuningParams.voxOccCtrl, dict)
    assert scene.acousticTuningParams.voxDiffrCtrl == 1.0
    assert isinstance(scene.acousticTuningParams.voxDiffrCtrl, float)
    assert scene.acousticTuningParams.erGainDb == -4.0
    assert isinstance(scene.acousticTuningParams.erGainDb, float)


def test_default_values_when_parsing_tuning_params(tmp_path: Path):
    test_eif = (
        '<?xml version="1.0" encoding="utf-8" standalone="yes"?>\n'
        '<AudioScene id="my_scene" duration="30">\n'
        "</AudioScene>\n"
    )
    tmp_eif = tmp_path.joinpath("my_scene.xml")
    tmp_eif.write_text(test_eif)

    scene = eif.AudioScene.load(str(tmp_eif))

    assert scene.acousticTuningParams.defaultAE == None
    assert scene.acousticTuningParams.voxOccCtrl == {}
    assert isinstance(scene.acousticTuningParams.voxOccCtrl, dict)
    assert scene.acousticTuningParams.voxDiffrCtrl == 1.0
    assert isinstance(scene.acousticTuningParams.voxDiffrCtrl, float)
    assert scene.acousticTuningParams.erGainDb == -8.0
    assert isinstance(scene.acousticTuningParams.erGainDb, float)


@pytest.mark.parametrize("occlCtrlVal", [100.1, -0.1])
def test_throw_exception_when_exceeding_valid_occlusion_control(tmp_path: Path, occlCtrlVal: float):
    test_eif = """<?xml version="1.0" encoding="utf-8" standalone="yes"?>
<AudioScene id="my_scene" duration="30">
    <AcousticTuningParameters defaultAE="true">
        <VoxDirectOcclusionControl>
            <Setting material="mat:test" value="{occlCtrlVal}" />
        </VoxDirectOcclusionControl>
    </AcousticTuningParameters>
</AudioScene>"""
    tmp_eif = tmp_path.joinpath("my_scene.xml")
    tmp_eif.write_text(test_eif)

    with pytest.raises(ValueError):
        eif.AudioScene.load(str(tmp_eif))


@pytest.mark.parametrize("voxDiffrCtrl", [1.01, -0.1])
def test_throw_exception_when_exceeding_valid_diffraction_control(
    tmp_path: Path, voxDiffrCtrl: float
):
    test_eif = """<?xml version="1.0" encoding="utf-8" standalone="yes"?>
<AudioScene id="my_scene" duration="30">
    <AcousticTuningParameters defaultAE="true">
        <VoxDiffractionControl value="{voxDiffrCtrl}" />
    </AcousticTuningParameters>
</AudioScene>"""

    tmp_eif = tmp_path.joinpath("my_scene.xml")
    tmp_eif.write_text(test_eif)

    with pytest.raises(ValueError):
        eif.AudioScene.load(str(tmp_eif))


@pytest.mark.parametrize("early_db", [-21.6, 10.1])
def test_throw_exception_when_exceeding_valid_diffraction_control(tmp_path: Path, early_db: float):
    test_eif = """<?xml version="1.0" encoding="utf-8" standalone="yes"?>
<AudioScene id="my_scene" duration="30">
    <AcousticTuningParameters defaultAE="true">
        <EarlyTuningGainDb value="{early_db}" />
    </AcousticTuningParameters>
</AudioScene>"""

    tmp_eif = tmp_path.joinpath("my_scene.xml")
    tmp_eif.write_text(test_eif)

    with pytest.raises(ValueError):
        eif.AudioScene.load(str(tmp_eif))


def test_tuning_params_toxml(tmp_path: Path):
    tuning_params: eif.AcousticTuningParameters = eif.AcousticTuningParameters()
    tuning_params.defaultAE: DefaultAE = DefaultAE()
    tuning_params.defaultAE.enabled = True
    tuning_params.defaultAE.preset = "LargeRoom"
    tuning_params.defaultAE.environment = "AE:Outdoor"
    tuning_params.defaultAE.density = 4
    tuning_params.defaultAE.delayMinMs = 80.0
    tuning_params.defaultAE.gainDb = -3
    tuning_params.defaultAE.distanceDropDb = 5
    tuning_params.defaultAE.minDist = 2
    tuning_params.revRendererControl: GlobalAEControl = GlobalAEControl()
    tuning_params.revRendererControl.distanceDropDb = 5
    tuning_params.revRendererControl.minDist = 2
    tuning_params.voxOccCtrl = {"mat:test": 42.0, "mat:test2": 3.1415}
    tuning_params.voxDiffrCtrl = 0.1234
    tuning_params.erGainDb = -10.1234

    doc: Document = Document()
    n: Element = tuning_params.toxml(doc)

    assert n.nodeName == "AcousticTuningParameters"

    assert n.getElementsByTagName("DefaultAE") != None
    assert (
        bool(n.getElementsByTagName("DefaultAE")[0].getAttribute("enabled"))
        == tuning_params.defaultAE.enabled
    )
    assert (
        n.getElementsByTagName("DefaultAE")[0].getAttribute("preset")
        == tuning_params.defaultAE.preset
    )
    assert (
        n.getElementsByTagName("DefaultAE")[0].getAttribute("environment")
        == tuning_params.defaultAE.environment
    )
    assert (
        float(n.getElementsByTagName("DefaultAE")[0].getAttribute("density"))
        == tuning_params.defaultAE.density
    )
    assert (
        float(n.getElementsByTagName("DefaultAE")[0].getAttribute("delayMinMs"))
        == tuning_params.defaultAE.delayMinMs
    )
    assert (
        float(n.getElementsByTagName("DefaultAE")[0].getAttribute("gainDb"))
        == tuning_params.defaultAE.gainDb
    )
    assert (
        float(n.getElementsByTagName("DefaultAE")[0].getAttribute("distanceDropDb"))
        == tuning_params.defaultAE.distanceDropDb
    )
    assert (
        float(n.getElementsByTagName("DefaultAE")[0].getAttribute("minDist"))
        == tuning_params.defaultAE.minDist
    )

    assert n.getElementsByTagName("GlobalAEControl") != None
    assert (
        float(n.getElementsByTagName("GlobalAEControl")[0].getAttribute("minDist"))
        == tuning_params.revRendererControl.minDist
    )
    assert (
        float(n.getElementsByTagName("GlobalAEControl")[0].getAttribute("distanceDropDb"))
        == tuning_params.revRendererControl.distanceDropDb
    )


if __name__ == "__main__":
    test_tuning_params_toxml()


def test_tuning_params_toxml(tmp_path: Path):
    tuning_params: eif.AcousticTuningParameters = eif.AcousticTuningParameters()
    tuning_params.voxOccCtrl = {"mat:test": 42.0, "mat:test2": 3.1415}
    tuning_params.voxDiffrCtrl = 0.1234
    tuning_params.erGainDb = -10.1234

    doc: Document = Document()
    n: Element = tuning_params.toxml(doc)

    assert n.nodeName == "AcousticTuningParameters"
    assert n.getElementsByTagName("VoxDirectOcclusionControl") != None
    assert n.getElementsByTagName("VoxDiffractionControl") != None
    assert n.getElementsByTagName("EarlyTuningGainDb") != None

    assert (
        float(n.getElementsByTagName("EarlyTuningGainDb")[0].getAttribute("value"))
        == tuning_params.erGainDb
    )

    assert (
        float(n.getElementsByTagName("VoxDiffractionControl")[0].getAttribute("value"))
        == tuning_params.voxDiffrCtrl
    )

    assert len(n.getElementsByTagName("VoxDirectOcclusionControl")[0].childNodes) == 2
    setting_0: Element = n.getElementsByTagName("VoxDirectOcclusionControl")[0].childNodes[0]
    assert setting_0.getAttribute("material") == "mat:test"
    assert float(setting_0.getAttribute("value")) == 42

    setting_1: Element = n.getElementsByTagName("VoxDirectOcclusionControl")[0].childNodes[1]
    assert setting_1.getAttribute("material") == "mat:test2"
    assert float(setting_1.getAttribute("value")) == 3.1415
