# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import pathlib

import pytest

from mpegi.framework import Bitstream


def test_addBool():
    trueBitstream = Bitstream()
    trueBitstream.addBool(True)

    assert trueBitstream.to01() == "1"

    falseBitstream = Bitstream()
    falseBitstream.addBool(False)

    assert falseBitstream.to01() == "0"


def test_addInt():
    bitstream = Bitstream()
    bitstream.addInt(15, 4)

    assert bitstream.to01() == "1111"

    bitstream.addInt(0, 2)
    assert bitstream.to01() == "111100"

    bitstream.addInt(10, 4)
    assert bitstream.to01() == "1111001010"

    # Add a number that is higher than the bit word length capacity
    with pytest.raises(OverflowError):
        bitstream.addInt(42, 1)


def test_addSignedInt():
    length = 4
    bitstreams = {}
    for i in range(-8, 8):
        bitstreams[i] = Bitstream()
        bitstreams[i].addSignedInt(i, length)

    assert bitstreams[0].to01() == "0000"
    assert bitstreams[1].to01() == "0001"
    assert bitstreams[2].to01() == "0010"
    assert bitstreams[3].to01() == "0011"
    assert bitstreams[4].to01() == "0100"
    assert bitstreams[5].to01() == "0101"
    assert bitstreams[6].to01() == "0110"
    assert bitstreams[7].to01() == "0111"
    assert bitstreams[-1].to01() == "1111"
    assert bitstreams[-2].to01() == "1110"
    assert bitstreams[-3].to01() == "1101"
    assert bitstreams[-4].to01() == "1100"
    assert bitstreams[-5].to01() == "1011"
    assert bitstreams[-6].to01() == "1010"
    assert bitstreams[-7].to01() == "1001"
    assert bitstreams[-8].to01() == "1000"

    with pytest.raises(OverflowError):
        Bitstream().addSignedInt(8, length)

    with pytest.raises(OverflowError):
        Bitstream().addSignedInt(-9, length)


def test_escapedLength():
    length = (2, 3, 4)
    bitstreams = []
    for i in range(26):
        bitstreams.append(Bitstream())
        bitstreams[i].addEscapedInt(i, length)

    assert bitstreams[0].to01() == "00"
    assert bitstreams[1].to01() == "01"
    assert bitstreams[2].to01() == "10"
    assert bitstreams[3].to01() == "11000"
    assert bitstreams[4].to01() == "11001"
    assert bitstreams[5].to01() == "11010"
    assert bitstreams[6].to01() == "11011"
    assert bitstreams[7].to01() == "11100"
    assert bitstreams[8].to01() == "11101"
    assert bitstreams[9].to01() == "11110"
    assert bitstreams[10].to01() == "111110000"
    assert bitstreams[11].to01() == "111110001"
    assert bitstreams[12].to01() == "111110010"
    assert bitstreams[13].to01() == "111110011"
    assert bitstreams[14].to01() == "111110100"
    assert bitstreams[15].to01() == "111110101"
    assert bitstreams[16].to01() == "111110110"
    assert bitstreams[17].to01() == "111110111"
    assert bitstreams[18].to01() == "111111000"
    assert bitstreams[19].to01() == "111111001"
    assert bitstreams[20].to01() == "111111010"
    assert bitstreams[21].to01() == "111111011"
    assert bitstreams[22].to01() == "111111100"
    assert bitstreams[23].to01() == "111111101"
    assert bitstreams[24].to01() == "111111110"
    assert bitstreams[25].to01() == "111111111"

    with pytest.raises(OverflowError):
        Bitstream().addEscapedInt(26, length)


def test_addFloat():
    # Floating point is meant to adhere to IEEE-754 standard
    # Here's a useful converter https://www.h-schmidt.net/FloatConverter/IEEE754.html
    a = Bitstream()
    a.addFloat32(1.25)
    assert a.to01() == "00111111101000000000000000000000"

    b = Bitstream()
    b.addFloat32(-31.806400299072265625)
    assert b.to01() == "11000001111111100111001110000010"


def test_addString():
    bitstream = Bitstream()
    bitstream.addString("foo")

    fAsciiCode = "01100110"
    oAsciiCode = "01101111"
    nullAsciiCode = "00000000"
    fooAsCStyleString = fAsciiCode + oAsciiCode + oAsciiCode + nullAsciiCode

    assert bitstream.to01() == fooAsCStyleString


def test_appendBitstream():
    left = Bitstream()
    left.addInt(7, 6)

    right = Bitstream()
    right.addInt(56, 6)
    left.appendBitstream(right)

    assert left.to01() == "000111111000"


def test_prependBitstream():
    left = Bitstream()
    left.addInt(7, 6)

    right = Bitstream()
    right.addInt(56, 6)
    right.prependBitstream(left)

    assert right.to01() == "000111111000"


def test_getLength():
    bitstream = Bitstream()
    bitstream.addInt(1)
    assert bitstream.getLength() == 8


def test_pad():
    bitstream = Bitstream()
    bitstream.addInt(7, 3)
    bitstream.pad()
    assert bitstream.to01() == "11100000"


def test_toBytes():
    bitstream = Bitstream()
    bitstream.addInt(7, 3)
    bytesObject = bytes([224])  # 1110 0000 is 224
    assert bitstream.toBytes() == bytesObject


def test_fromBytes():
    bytesObject = bytes([15])  # 15 is 0000 1111
    bitstream = Bitstream.fromBytes(bytesObject)
    assert bitstream.to01() == "00001111"

    bytesObject = bytes([128, 0])
    bitstream = Bitstream.fromBytes(bytesObject)
    assert bitstream.to01() == "1000000000000000"


def test_writeToFile(tmp_path: pathlib.Path):
    numberToAddToBitstream = 7
    bitstream = Bitstream()
    bitstream.addInt(numberToAddToBitstream)
    outputFilePath = tmp_path.joinpath("test.bit")
    bitstream.writeToFile(str(outputFilePath))

    with open(outputFilePath, "r+b") as file:
        intFromFile = int.from_bytes(file.read(), "little")

    assert intFromFile == numberToAddToBitstream


@pytest.mark.parametrize("value", [True, False])
def test_toBool(value):
    bitstream = Bitstream()
    bitstream.addBool(value)

    assert bool(bitstream) == value
    assert bitstream.toBool() == value


def test_castingABitstreamThatIsMoreThanOneBitLongToABoolShouldFail():
    bitstream = Bitstream()
    bitstream.addBool(True)
    bitstream.addBool(True)

    with pytest.raises(ValueError):
        bool(bitstream)

    with pytest.raises(ValueError):
        bitstream.toBool()


def test_castingABitstreamThatIsEmptyToABoolShouldFail():
    bitstream = Bitstream()

    with pytest.raises(ValueError):
        bool(bitstream)

    with pytest.raises(ValueError):
        bitstream.toBool()


def test_toInt():
    bitstream = Bitstream()
    bitstream.addBool(True)

    assert int(bitstream) == 1
    assert bitstream.toInt() == 1

    bitstream = Bitstream()
    bitstream.addInt(15)

    assert int(bitstream) == 15
    assert bitstream.toInt() == 15

    bitstream = Bitstream()
    bitstream.addFloat32(1.25)

    assert int(bitstream) == 1067450368
    assert bitstream.toInt() == 1067450368


def test_toSignedInt():
    bitstream = Bitstream()
    bitstream.addBool(False)
    bitstream.addBool(True)

    assert bitstream.toSignedInt() == 1

    bitstream = Bitstream()
    bitstream.addSignedInt(-41)

    assert bitstream.toSignedInt() == -41


@pytest.mark.parametrize("value", [1.25, -3.1530001163482666, 34.0])
def test_toFloat(value):
    bitstream = Bitstream()
    bitstream.addFloat32(value)

    assert bitstream.toFloat() == value


def test_toAscii():
    bitstream = Bitstream()
    bitstream.addString("foo")

    assert bitstream.toAscii() == "foo\x00"


def test_pop():
    bitstream = Bitstream()
    bitstream.addBool(True)
    bitstream.addBool(False)

    poppedBit = bitstream.pop()

    assert poppedBit.toBool() == True
    assert bitstream.toBool() == False


def test_popWithInts():
    bitstream = Bitstream()
    bitstream.addInt(15)

    poppedBits = bitstream.pop(numberOfBits=4)

    assert poppedBits.toInt() == 0
    assert bitstream.toInt() == 15


def test_copy():
    bitstream = Bitstream()

    new_bitstream = bitstream.copy()

    assert bitstream.to01() == new_bitstream.to01()
    assert id(bitstream) != id(new_bitstream)


def test_copy_with_range_specified():
    bitstream = Bitstream()
    bitstream.addInt(15)

    new_bitstream = bitstream.copy(start=2, length=4)

    assert bitstream.toInt() == 15
    assert new_bitstream.toInt() == 3
