# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, List

import pytest

from mpegi.framework import (
    Bitstream,
    BitstreamComposer,
    NotEnoughDataPushedError,
    QuantizationRange,
    tokens,
)
from mpegi.framework.generic_codebook import (
    HuffmanDecoder,
    SymbolHistogram,
    get_encoder_config,
)

# ---------------------------------------------------------------------------- #
#                                   Fixtures                                   #
# ---------------------------------------------------------------------------- #


@pytest.fixture
def flagValue() -> Dict[str, Any]:
    return {"bool": True, "binStr": "1"}


@pytest.fixture()
def uIntValue() -> Dict[str, Any]:
    return {"uInt": 1, "width": 8, "binStr": "00000001"}


@pytest.fixture()
def sIntValue() -> Dict[str, Any]:
    return {"sInt": -4, "width": 8, "binStr": "11111100"}


@pytest.fixture()
def floatValue() -> Dict[str, Any]:
    # Useful tool for IEEE754 float conversions - https://www.h-schmidt.net/FloatConverter/IEEE754.html
    return {"float": 1.25, "binStr": "00111111101000000000000000000000"}


@pytest.fixture()
def quantizedFloatValue() -> Dict[str, Any]:
    return {"float": 0.89, "bin_str": "001", "q_range": QuantizationRange(0, 7, 3)}


@pytest.fixture()
def cStringValue() -> Dict[str, Any]:
    fAsAsciiBinary = f"{ord('f'):b}".zfill(8)
    oAsAsciiBinary = f"{ord('o'):b}".zfill(8)
    nullByte = "0" * 8
    binStr = fAsAsciiBinary + oAsAsciiBinary + oAsAsciiBinary + nullByte
    return {"str": "foo", "bytes": b"foo\x00", "binStr": binStr}


@pytest.fixture
def blobValue() -> Dict[str, Any]:
    blob = Bitstream()
    blob.addInt(255)

    bitsForBlobLength = 32
    blobLength = "1".zfill(bitsForBlobLength)
    blobValue = "1" * 8
    return {"blob": blob, "binStr": blobLength + blobValue}


# ---------------------------------------------------------------------------- #
#                          Tests for Primitive Tokens                          #
# ---------------------------------------------------------------------------- #


def test_addFlag(flagValue):
    syntax: List[tokens.Token] = [tokens.Flag("firstFlag")]
    composer = BitstreamComposer(syntax)
    composer.push(value=flagValue["bool"])

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == flagValue["binStr"]


def test_addUnsignedInt(uIntValue):
    syntax: List[tokens.Token] = [tokens.UnsignedInteger("firstUInt")]
    composer = BitstreamComposer(syntax)
    composer.push(value=uIntValue["uInt"])

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == uIntValue["binStr"]


def test_addUnsignedIntWithSpecifiedWidth():
    syntax: List[tokens.Token] = [tokens.UnsignedInteger("firstUInt", width=13)]
    composer = BitstreamComposer(syntax)
    composer.push(value=15)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "0000000001111"


@pytest.mark.parametrize(["value", "expected"], [(42, "00"), (61, "01"), (93, "10"), (1, "11")])
def test_addUnsignedIntWithCodePage(value, expected):
    code_page = {42: 0, 61: 1, 93: 2, 1: 3}
    syntax: List[tokens.Token] = [tokens.UnsignedInteger("uint", width=2, code_page=code_page)]
    composer = BitstreamComposer(syntax)
    composer.push(value)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == expected


def test_addSignedInt(sIntValue):
    syntax: List[tokens.Token] = [tokens.SignedInteger("SignedInt")]
    composer = BitstreamComposer(syntax)
    composer.push(value=sIntValue["sInt"])

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == sIntValue["binStr"]


def test_addSignedIntWithSpecifiedWidth():
    syntax: List[tokens.Token] = [tokens.SignedInteger("SignedInt", width=23)]
    composer = BitstreamComposer(syntax)
    composer.push(value=-2414953)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "10110110010011010010111"


def test_addFloat(floatValue):
    syntax: List[tokens.Token] = [tokens.Float32("FloatingPoint")]
    composer = BitstreamComposer(syntax)
    composer.push(value=floatValue["float"])

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == floatValue["binStr"]


def test_addQuantizedFloat(quantizedFloatValue: Dict[str, Any]):
    q_range = quantizedFloatValue["q_range"]
    syntax: List[tokens.Token] = [tokens.QuantizedFloat("QFloat", quant_range=q_range)]
    composer = BitstreamComposer(syntax)
    composer.push(value=quantizedFloatValue["float"])

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == quantizedFloatValue["bin_str"]


def test_addCString(cStringValue):
    syntax: List[tokens.Token] = [tokens.CString("ACstyleString")]
    composer = BitstreamComposer(syntax)
    composer.push(value=cStringValue["str"])

    bitstream = composer.get_bitstream()

    assert bitstream.toBytes() == cStringValue["bytes"]


def test_addBlob(blobValue):
    syntax: List[tokens.Token] = [tokens.Blob("ArbitraryBinBlob")]
    composer = BitstreamComposer(syntax)
    composer.push(value=blobValue["blob"])

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == blobValue["binStr"]


def test_addLabel():
    syntax: List[tokens.Token] = [tokens.Label("Label")]
    composer = BitstreamComposer(syntax)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == ""


def test_addAllPrimitiveTokens(
    flagValue,
    uIntValue,
    sIntValue,
    floatValue,
    cStringValue,
    blobValue,
):
    syntax: List[tokens.Token] = [
        tokens.Label("Here is every single primitive token added one after another"),
        tokens.Flag("Flag"),
        tokens.UnsignedInteger("UInt"),
        tokens.SignedInteger("SignedInt"),
        tokens.Label("Here's another label for good measure"),
        tokens.Float32("Float"),
        tokens.CString("CString"),
        tokens.Blob("blob"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(value=flagValue["bool"])
    composer.push(value=uIntValue["uInt"])
    composer.push(value=sIntValue["sInt"])
    composer.push(value=floatValue["float"])
    composer.push(value=cStringValue["str"])
    composer.push(value=blobValue["blob"])
    bitstream = composer.get_bitstream()

    allTokensAsBinStr = (
        flagValue["binStr"]
        + uIntValue["binStr"]
        + sIntValue["binStr"]
        + floatValue["binStr"]
        + cStringValue["binStr"]
        + blobValue["binStr"]
    )
    assert bitstream.to01() == allTokensAsBinStr


# ---------------------------------------------------------------------------- #
#                         Tests for Control-Flow tokens                        #
# ---------------------------------------------------------------------------- #


@pytest.mark.parametrize("repeats", [1, 3, 42, 0])
def test_repeatTokenWithInt(repeats):
    syntax: List[tokens.Token] = [tokens.Repeat(repeats, [tokens.Flag("Flag")])]
    composer = BitstreamComposer(syntax)
    for _ in range(repeats):
        composer.push(value=True)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "1" * repeats


def test_repeatTokenWithTokenReference():
    fooValue = 3
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("Foo"),
        tokens.Repeat("Foo", [tokens.Flag("Flag")]),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(value=fooValue)
    for _ in range(fooValue):
        composer.push(value=True)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "00000011111"


def test_repeatTokenWithTokenReferenceThatResolvesToZero():
    fooValue = 0
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("Foo"),
        tokens.Repeat("Foo", [tokens.Flag("Flag")]),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(value=fooValue)
    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "00000000"


def test_repeatTokenThatResolvesCodePage():
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("foo", width=2, code_page={3: 0, 66: 1, 91: 2, 100: 3}),
        tokens.Repeat("foo", syntax=[tokens.Flag("bar")]),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    composer.push(True)
    composer.push(True)
    composer.push(True)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "00111"


@pytest.mark.parametrize(["values", "expected"], [([True, True], "11"), ([False, 7], "0111")])
def test_conditionalToken(values, expected):
    syntax: List[tokens.Token] = [
        tokens.Flag("foo"),
        tokens.Conditional(
            "foo",
            true=[tokens.Flag("IfTrue")],
            false=[tokens.UnsignedInteger("IfFalse", width=3)],
        ),
    ]
    composer = BitstreamComposer(syntax)
    for value in values:
        composer.push(value)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == expected


@pytest.mark.parametrize(
    ["values", "expected"],
    [
        ([0, 2], "00" + "10"),
        ([1, True, False], "01" + "1" + "0"),
        ([2, "k"], "10" + f"{ord('k'):b}".zfill(8) + ("0" * 8)),
    ],
)
def test_switchToken(values, expected):
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("condition", width=2),
        tokens.Switch(
            "condition",
            cases=[
                tokens.SwitchCase(case=0, syntax=[tokens.UnsignedInteger("first", width=2)]),
                tokens.SwitchCase(case=1, syntax=[tokens.Flag("second"), tokens.Flag("third")]),
                tokens.SwitchCase(case=2, syntax=[tokens.CString("fourth")]),
            ],
        ),
    ]
    composer = BitstreamComposer(syntax)
    for value in values:
        composer.push(value)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == expected


def test_expression_token():
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("foo", width=4),
        tokens.UnsignedInteger("bar", width=4),
        tokens.Assign("result", tokens.Add("foo", "bar")),
        tokens.Assign("result", tokens.Add("result", 1)),
        tokens.Assign("result", tokens.Add(-1, "result")),
        tokens.Repeat("result", syntax=[tokens.Flag("baz")]),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(1)
    composer.push(2)
    for _ in range(3):  # result == 3
        composer.push(True)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "00010010111"


def test_while_token():
    syntax: List[tokens.Token] = [
        tokens.Assign("do_loop", tokens.Value(1)),
        tokens.While(
            "do_loop",
            syntax=[
                tokens.UnsignedInteger("some_int", width=2),
                tokens.Assign("do_loop", tokens.Lookup("some_int")),
            ],
        ),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    composer.push(2)
    composer.push(1)
    composer.push(0)

    bitstream = composer.get_bitstream()

    assert bitstream.to01() == "11100100"


# ---------------------------------------------------------------------------- #
#                            Tests for invalid input                           #
# ---------------------------------------------------------------------------- #


def test_unsignedIntWithoutWidthKeywordShouldFail():
    with pytest.raises(TypeError):
        tokens.UnsignedInteger("BreakWithoutWidthKeyword", 3)


def test_signedIntWithoutWidthKeywordShouldFail():
    with pytest.raises(TypeError):
        tokens.SignedInteger("BreakWithoutWidthKeyword", 412)


def test_ingestingAValueWhenThereAreNoMoreTokensLeftShouldFail():
    syntax: List[tokens.Token] = [
        tokens.Flag("firstFlag"),
        tokens.Flag("secondFlag"),
        tokens.Flag("thirdFlag"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(value=True)
    composer.push(value=True)
    composer.push(value=True)

    with pytest.raises(OverflowError):
        composer.push(value=True)


def test_gettingABitstreamWithNotEnoughIngestedDataShouldFail():
    syntax: List[tokens.Token] = [
        tokens.Flag("firstFlag"),
        tokens.Flag("secondFlag"),
        tokens.Flag("thirdFlag"),
    ]
    composer = BitstreamComposer(syntax)

    with pytest.raises(NotEnoughDataPushedError):
        composer.get_bitstream()


def test_addingNonBoolToFlagTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.Flag("Flag")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=42)


def test_addingNonIntToUIntTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.UnsignedInteger("Uint")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=42.3)


def test_addingNonIntToSIntTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.SignedInteger("Sint")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=42.3)


def test_addingNonFloatToFloatTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.Float32("Float32")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=True)


def test_addingNonStringToStringTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.CString("String")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=True)


def test_addingNonBlobToBlobTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.Blob("Blob")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=True)


def test_addingNegativeIntegerToUIntTokenShouldFail():
    syntax: List[tokens.Token] = [tokens.UnsignedInteger("Int")]
    composer = BitstreamComposer(syntax)

    with pytest.raises(TypeError):
        composer.push(value=-13)


def test_codePageExceedingGivenWidthShouldFail():
    with pytest.raises(Exception):
        tokens.UnsignedInteger("foo", width=8, code_page={15: 256})


def test_add_codebook_config():
    syntax: List[tokens.Token] = [tokens.CodebookConfig("codebook")]
    composer = BitstreamComposer(syntax)
    histogram = SymbolHistogram({0: 2, 1: 4, 2: 7})
    encoder_config = get_encoder_config(histogram)
    composer.push(encoder_config)

    bitstream = composer.get_bitstream()
    decoded_codebook = HuffmanDecoder(bitstream)

    assert decoded_codebook.fixed_length == encoder_config.fixed_length
    assert decoded_codebook.offset == encoder_config.offset


def test_add_codebook_symbol():
    syntax: List[tokens.Token] = [
        tokens.CodebookConfig("codebook"),
        tokens.CodebookSymbol("symbol", config="codebook"),
    ]
    composer = BitstreamComposer(syntax)
    histogram = SymbolHistogram({0: 2, 1: 4, 2: 7})
    encoder_config = get_encoder_config(histogram)
    composer.push(encoder_config)
    composer.push(2)

    bitstream = composer.get_bitstream()
    decoded_codebook = HuffmanDecoder(bitstream)

    assert decoded_codebook.fixed_length == encoder_config.fixed_length
    assert decoded_codebook.offset == encoder_config.offset
