# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

import pytest

from mpegi.framework import (
    Bitstream,
    BitstreamComposer,
    BitstreamDecomposer,
    QuantizationRange,
    tokens,
)


def test_pop_bitstream_chunks():
    bitstream = Bitstream()
    bitstream.addInt(15)
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("first", width=4),
        tokens.UnsignedInteger("second", width=4),
    ]
    decomposer = BitstreamDecomposer(bitstream, syntax)

    first = decomposer.pop()
    second = decomposer.pop()

    assert first.to01() == "0000"
    assert second.to01() == "1111"


def test_bitstream_extraction():
    bitstream = Bitstream()
    bitstream.addInt(15)
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("first", width=2),
        tokens.StartMarker(),
        tokens.UnsignedInteger("second", width=4),
        tokens.EndMarker(),
        tokens.UnsignedInteger("third", width=2),
    ]
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.to01() == "0011"


def test_extract_unsigned_int():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.UnsignedInteger("foo", width=8),
        tokens.EndMarker("end"),
    ]
    bitstream = Bitstream()
    bitstream.addInt(255)
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.to01() == "11111111"


def test_extract_esc_unsigned_int():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.EscapedUnsignedInteger("foo", widths=[4, 8, 32]),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(4711)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_extract_signed_int():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.SignedInteger("foo", width=8),
        tokens.EndMarker("end"),
    ]
    bitstream = Bitstream()
    bitstream.addSignedInt(-34)
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.to01() == "11011110"


def test_extract_cstring():
    syntax: List[tokens.Token] = [
        tokens.StartMarker(),
        tokens.CString("foo"),
        tokens.EndMarker(),
    ]
    bitstream = Bitstream()
    bitstream.addString("foo")
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.getLength() == (4 * 8)


def test_extract_blob():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.Blob("blob"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    blob = Bitstream()
    blob.addInt(255, width=8)
    composer.push(blob)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.getLength() == (tokens.Blob.bitsForBlobLength + 8)


def test_extract_quant_float():
    q_range = QuantizationRange(min=0.0, max=8.0, num_bits=8)
    syntax: List[tokens.Token] = [
        tokens.StartMarker(),
        tokens.QuantizedFloat("qfloat", quant_range=q_range),
        tokens.EndMarker(),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(8.0)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.getLength() == 8


def test_extract_flag():
    syntax: List[tokens.Token] = [
        tokens.StartMarker(),
        tokens.Flag("flag"),
        tokens.EndMarker(),
    ]
    bitstream = Bitstream()
    bitstream.addBool(True)
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.to01() == "1"


def test_extract_expression():
    syntax: List[tokens.Token] = [
        tokens.StartMarker(),
        tokens.UnsignedInteger("foo", width=4),
        tokens.UnsignedInteger("bar", width=4),
        tokens.Assign("result", tokens.Add("foo", "bar")),
        tokens.Repeat("result", syntax=[tokens.Flag("baz")]),
        tokens.EndMarker(),
    ]
    bitstream = Bitstream()
    bitstream.addInt(1, width=4)
    bitstream.addInt(2, width=4)
    for _ in range(3):
        bitstream.addBool(True)
    decomposer = BitstreamDecomposer(bitstream, syntax)

    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.to01() == bitstream.to01()


def test_bitstream_extraction_with_many_different_elements():
    syntax: List[tokens.Token] = [
        tokens.UnsignedInteger("dummy_int"),
        tokens.StartMarker(),
        tokens.UnsignedInteger("iterator", width=3),
        tokens.Repeat(
            "iterator",
            [
                tokens.CString("foo"),
                tokens.Float32("float"),
                tokens.Flag("flag"),
                tokens.Conditional(
                    "flag",
                    true=[tokens.UnsignedInteger("short", width=8)],
                    false=[tokens.UnsignedInteger("long", width=16)],
                ),
            ],
        ),
        tokens.EndMarker(),
        tokens.Float32("dummy_float"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(0)  # dummy
    composer.push(2)  # for loop

    composer.push("foo")  # cstring
    composer.push(0.0)  # float
    composer.push(False)  # condition
    composer.push(65535)  # long

    composer.push("foo")  # cstring
    composer.push(0.0)  # float
    composer.push(True)  # condition
    composer.push(255)  # short

    composer.push(0.0)  # dummy

    bitstream = composer.get_bitstream()

    expected_length: int = 3 + (32 + (4 * 8) + 1 + 16) + (32 + (4 * 8) + 1 + 8)
    decomposer = BitstreamDecomposer(bitstream, syntax)
    extracted_bitstream = decomposer.extract_between_markers()

    assert extracted_bitstream.getLength() == expected_length
