# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import pathlib

import pytest

from typing import List
from mpegi.framework import Bitstream, BitstreamComposer, tokens

write_function = tokens.Function(
    "writeFunction",
    syntax=[
        tokens.writeCountOrIndex("testFunctionCount"),
        tokens.Repeat(
            "testFunctionCount",
            syntax=[tokens.writeID("testFunctionID")],
        ),
        tokens.writeFrequency("testFrequency"),
    ],
)


def test_writeID():
    syntax: List[tokens.Token] = [tokens.writeID("testID")]
    composer = BitstreamComposer(syntax)
    composer.push(42)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "01010100"

    syntax: List[tokens.Token] = [tokens.writeID("testID")]
    composer = BitstreamComposer(syntax)
    composer.push(16385)
    payload_as_bitstream = composer.get_bitstream()
    # "0000001 1 0000000 1 0000001 0"
    assert payload_as_bitstream.to01() == "000000110000000100000010"


def test_writeCountOrIndex():
    syntax: List[tokens.Token] = [tokens.writeCountOrIndex("testCountOrIndex")]
    composer = BitstreamComposer(syntax)
    composer.push(42)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "11001010"

    syntax: List[tokens.Token] = [tokens.writeCountOrIndex("testCountOrIndex2")]
    composer = BitstreamComposer(syntax)
    composer.push(1)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1000"

    syntax: List[tokens.Token] = [tokens.writeCountOrIndex("testCountOrIndex3")]
    composer = BitstreamComposer(syntax)
    composer.push(65)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1001001"


def test_writeDuration():

    isLongRangeMode = False
    syntax: List[tokens.Token] = [
        tokens.writeDuration("testDuration", isLongRangeMode=isLongRangeMode)
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.000)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1100111100"

    syntax: List[tokens.Token] = [
        tokens.writeDuration("testDuration2", isLongRangeMode=isLongRangeMode)
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.300)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "01100111100"

    syntax: List[tokens.Token] = [
        tokens.writeDuration("testDuration2a", isLongRangeMode=isLongRangeMode)
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.30012)
    payload_as_bitstream = composer.get_bitstream()
    # deciSeconds addMilliSeconds miliseconds addSamples samplesCode addseconds seconds
    # 0110        1               1111010     1          10111      1          11100
    assert payload_as_bitstream.to01() == "011011111010110111111100"

    syntax: List[tokens.Token] = [
        tokens.writeDuration("testDuration3", isLongRangeMode=isLongRangeMode)
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.327)
    payload_as_bitstream = composer.get_bitstream()
    # deciSeconds addMilliSeconds miliseconds addSamples addseconds seconds
    # 0110        1               0100011     0          1          11100
    assert payload_as_bitstream.to01() == "0110101000110111100"

    syntax: List[tokens.Token] = [tokens.writeDuration("testDuration4", isLongRangeMode=True)]
    composer = BitstreamComposer(syntax)
    composer.push(3618.327)
    payload_as_bitstream = composer.get_bitstream()
    # deciSeconds addMilliSeconds miliseconds addSamples addseconds rawseconds   addHours hoursCode
    # 0110        1               0100011     0          1          000000010010 1        00000
    assert payload_as_bitstream.to01() == "01101010001101000000010010100000"


def test_writeFrequency():
    syntax: List[tokens.Token] = [tokens.writeFrequency("testFrequency")]
    composer = BitstreamComposer(syntax)
    composer.push(1250)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "0100100"

    syntax: List[tokens.Token] = [tokens.writeFrequency("testFrequency2")]
    composer = BitstreamComposer(syntax)
    composer.push(1100)
    payload_as_bitstream = composer.get_bitstream()
    # 1100 1 0110 = 1000Hz True 6
    assert payload_as_bitstream.to01() == "110010110"


def test_writeDistance():
    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance", isSmallScene=True)]
    composer = BitstreamComposer(syntax)
    composer.push(50)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "10001000"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance2", isSmallScene=True)]
    composer = BitstreamComposer(syntax)
    composer.push(50.42)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1000100100010000"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance3", isSmallScene=True)]
    composer = BitstreamComposer(syntax)
    composer.push(50.428)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1000100100010001100"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance3b")]
    composer = BitstreamComposer(syntax)
    composer.push({"distance": 50.428, "isSmallScene": True})
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1000100100010001100"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance4", isSmallScene=False)]
    composer = BitstreamComposer(syntax)
    composer.push(350.428)
    payload_as_bitstream = composer.get_bitstream()
    # metersCode addHm hmCode addKm addCm cmCode  addmm mmCode
    # 1000100    1     111    0     1     0001000 1     100'
    assert payload_as_bitstream.to01() == "100010011110100010001100"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance5", isSmallScene=False)]
    composer = BitstreamComposer(syntax)
    composer.push(9350.428)
    payload_as_bitstream = composer.get_bitstream()
    # metersCode addHm hmCode addKm kmCode addkm addCm cmCode  addmm mmCode
    # 1000100    1     111    1     11001  0     1     0001000 1     100
    assert payload_as_bitstream.to01() == "100010011111110010100010001100"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance6", isSmallScene=False)]
    composer = BitstreamComposer(syntax)
    composer.push(19350.428)
    payload_as_bitstream = composer.get_bitstream()
    # metersCode addHm hmCode addKm kmCode addKM kmCode addkm addCm cmCode  addmm mmCode
    # 1000100    1     111    1     11000  1     11001  0     1     0001000 1     100
    assert payload_as_bitstream.to01() == "100010011111110001110010100010001100"

    syntax: List[tokens.Token] = [tokens.writeDistance("testDistance6", isSmallScene=False)]
    composer = BitstreamComposer(syntax)
    composer.push(9050.428)
    payload_as_bitstream = composer.get_bitstream()
    # metersCode addHm hmCode addKm kmCode addKm addCm cmCode  addmm mmCode
    # 1000100    1     000    1     11001  0     1     0001000 1     100
    assert payload_as_bitstream.to01() == "100010010001110010100010001100"


def test_writeGain():
    syntax: List[tokens.Token] = [tokens.writeGain("testGain", isHiPrecGain=False)]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "0100"

    syntax: List[tokens.Token] = [tokens.writeGain("testGain2", isHiPrecGain=False)]
    composer = BitstreamComposer(syntax)
    composer.push(5)
    payload_as_bitstream = composer.get_bitstream()
    # coarseGainCode addFineGain fineGainCode
    # 1101           1           11
    assert payload_as_bitstream.to01() == "1101111"

    syntax: List[tokens.Token] = [tokens.writeGain("testGain3", isHiPrecGain=True)]
    composer = BitstreamComposer(syntax)
    composer.push(4.2)
    payload_as_bitstream = composer.get_bitstream()
    # coarseGainCode addFineGain fineGainCode addDeciGain deciGainCode
    # 1101           1           10           1           101
    assert payload_as_bitstream.to01() == "11011101101"

    syntax: List[tokens.Token] = [tokens.writeGain("testGain3", isHiPrecGain=False)]
    composer = BitstreamComposer(syntax)
    composer.push(-42)
    payload_as_bitstream = composer.get_bitstream()
    # coarseGainCode    addFineGain fineGainCode addDeciGain deciGainCode
    # 0110101          1           001111
    assert payload_as_bitstream.to01() == "01101011001111"

    syntax: List[tokens.Token] = [tokens.writeGain("testGain4", isHiPrecGain=True)]
    composer = BitstreamComposer(syntax)
    composer.push(-4.2)
    payload_as_bitstream = composer.get_bitstream()
    # coarseGainCode addFineGain fineGainCode addDeciGain deciGainCode
    # 101            1           10           1           0010
    # -3             +           -2           +           0.8 = -4.2
    assert payload_as_bitstream.to01() == "10111010010"


def test_writePosition():
    syntax: List[tokens.Token] = [tokens.writePosition("testPosition", isSmallScene=True)]
    composer = BitstreamComposer(syntax)
    composer.push([12, 4, 7])
    payload_as_bitstream = composer.get_bitstream()
    # dist isNeg dist isNeg dist isNeg
    # 0010010 0 1100010 0 1101000 0
    assert payload_as_bitstream.to01() == "001001001100010011010000"

    syntax: List[tokens.Token] = [tokens.writePosition("testPosition", isSmallScene=True)]
    composer = BitstreamComposer(syntax)
    composer.push([-12, 4, -7])
    payload_as_bitstream = composer.get_bitstream()
    # dist    isNeg dist    isNeg dist    isNeg
    # 0010010 1     1100010 0     1101000 1
    assert payload_as_bitstream.to01() == "001001011100010011010001"

    syntax: List[tokens.Token] = [tokens.writePosition("testPosition", isSmallScene=True)]
    composer = BitstreamComposer(syntax)
    composer.push([-50.428, 27.3, -7.001])
    payload_as_bitstream = composer.get_bitstream()
    # dist                    isNeg   dist                isNeg  dist                    isNeg
    # 1000100 1 0001000 1 100 // 1 // 011000 1 0111100 0 // 0 // 110100 1 110010 1 001 // 1
    assert payload_as_bitstream.to01() == "100010010001000110010110001011110000110100111001010011"


def test_writeOrientation():
    syntax: List[tokens.Token] = [tokens.writeOrientation("testOrientation")]
    composer = BitstreamComposer(syntax)
    composer.push([0, 0, 0])
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "01"

    syntax: List[tokens.Token] = [tokens.writeOrientation("testOrientation2")]
    composer = BitstreamComposer(syntax)
    composer.push([0, -90, 0])
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "1010"

    syntax: List[tokens.Token] = [tokens.writeOrientation("testOrientation3")]
    composer = BitstreamComposer(syntax)
    composer.push([180, 60, 180])
    payload_as_bitstream = composer.get_bitstream()
    # 111 010010 0 110010 0 010010 0'
    assert payload_as_bitstream.to01() == "111010010011001000100100"

    syntax: List[tokens.Token] = [tokens.writeOrientation("testOrientation4", highResolution=True)]
    composer = BitstreamComposer(syntax)
    composer.push([180, 60, 180])
    payload_as_bitstream = composer.get_bitstream()
    # 111 010010 0 110010 0 010010 0
    # 100 01000011001101000000000000000000 01000010011100000000000000000000 01000011001101000000000000000000
    # 100010000110011010000000000000000000100001001110000000000000000000001000011001101000000000000000000
    assert (
        payload_as_bitstream.to01()
        == "100010000110011010000000000000000000100001001110000000000000000000001000011001101000000000000000000"
    )

    syntax: List[tokens.Token] = [tokens.writeOrientation("testOrientation5")]
    composer = BitstreamComposer(syntax)
    composer.push([60, 90, 90])
    payload_as_bitstream = composer.get_bitstream()
    # The LUT contains an entry for 0.33333, but as 60deg is 1/3 pi, it fails to find it due to rounding errors.
    # explicitAngles // coarseYaw // addfine // coarsePitch // addfine // coarseRoll // addFine
    # '111           // 110010    // 0       // 110100      // 0       // 110100     // 0'
    # '111              111101       1          11110          110100 0 110100 0'
    # '111              110010       0          110100         0          110100        0'
    assert payload_as_bitstream.to01() == "111110010011010001101000"


def test_repeat():
    syntax: List[tokens.Token] = [
        tokens.writeCountOrIndex("testRepeatCount"),
        tokens.Repeat("testRepeatCount", syntax=[tokens.writeID("testRepeatID")]),
    ]

    composer = BitstreamComposer(syntax)
    composer.push(5)
    for n in range(0, 5):
        composer.push(n)
    payload_as_bitstream = composer.get_bitstream()
    # 010110 00000000 00000010 00000100 00000110 00001000
    assert payload_as_bitstream.to01() == "0101100000000000000010000001000000011000001000"


def test_function():
    composer = BitstreamComposer([write_function])
    composer.push(5)
    for n in range(0, 5):
        composer.push(n)
    composer.push(1250)
    payload_as_bitstream = composer.get_bitstream()
    # 010110 00000000 00000010 00000100 00000110 00001000
    assert payload_as_bitstream.to01() == "01011000000000000000100000010000000110000010000100100"


def test_FullSyntax():
    syntax: List[tokens.Token] = [
        tokens.writeID("testID"),
        tokens.writeCountOrIndex("testCountOrIndex"),
        tokens.writeDuration("testDuration", isLongRangeMode=True),
        tokens.writeFrequency("testFrequency"),
        tokens.writeDistance("testDistance", isSmallScene=False),
        tokens.writeGain("testGain", isHiPrecGain=False),
        tokens.writePosition("testPosition", isSmallScene=True),
        tokens.writeOrientation("testOrientation"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(42)
    composer.push(65)
    composer.push(3618.327)
    composer.push(1250)
    composer.push(350.428)
    composer.push(3)
    composer.push([-50.428, 27.3, -7.001])
    composer.push([60, 90, 90])

    # "01010100"
    # "1001001"
    # "01101010001101000000010010100000"
    # "0100100"
    # "100010011110100010001100"
    # "0100"
    # "100010010001000110010110001011110000110100111001010011"
    #  111110010011010001101000

    payload_as_bitstream = composer.get_bitstream()

    assert (
        payload_as_bitstream.to01()
        == "0101010010010010110101000110100000001001010000001001001000100111101000100011000100100010010001000110010110001011110000110100111001010011111110010011010001101000"
    )


def test_writeValueFromLUT():
    syntax: List[tokens.Token] = [
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="dsr_code")
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-140)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "10000110"

    syntax: List[tokens.Token] = [
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="dsr_code")
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-160)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "10001100"

    syntax: List[tokens.Token] = [
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="dsr_code")
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-140.4)
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "10000110"

    syntax: List[tokens.Token] = [
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="frequency_hop_codes")
    ]
    composer = BitstreamComposer(syntax)
    composer.push(pow(2, (1 / 3)))
    payload_as_bitstream = composer.get_bitstream()
    assert payload_as_bitstream.to01() == "01"


def test_repeatWithIndex():
    syntax: List[tokens.Token] = [
        tokens.writeCountOrIndex("forEachCount", storeValue=True),
        tokens.writeCountOrIndex("forEachCount", storeValue=True),
        tokens.writeCountOrIndex("idx"),
        tokens.repeatWithIndex("forEachCount", "idx", syntax=[tokens.writeID("testIDs")]),
        tokens.writeCountOrIndex("idx"),
        tokens.repeatWithIndex("forEachCount", "idx", syntax=[tokens.writeID("testIDs")]),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    composer.push(5)
    composer.push(0)
    for n in range(3):
        composer.push(42)
    composer.push(1)
    for n in range(5):
        composer.push(24)
    payload_as_bitstream = composer.get_bitstream()
    # 011010 010110 01110 01010100 01010100 01010100 1000 00110000 00110000 00110000 00110000 00110000
    # 0110100101100111001010100010101000101010010000011000000110000001100000011000000110000
    assert (
        payload_as_bitstream.to01()
        == "0110100101100111001010100010101000101010010000011000000110000001100000011000000110000"
    )


def main():
    test_writeID()
    test_writeCountOrIndex()
    test_writeDuration()
    test_writeFrequency()
    test_writePosition()
    test_writeDistance()
    test_writeOrientation()
    test_writeGain()
    test_FullSyntax()
    test_repeat()
    test_function()
    test_writeValueFromLUT()
    test_repeatWithIndex()


if __name__ == "__main__":
    main()
