# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import pathlib

import pytest

from typing import List
from mpegi.framework import Bitstream, BitstreamComposer, tokens, BitstreamDecomposer


def test_writeID():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeID("testID"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(42)
    bitstream = composer.get_bitstream()

    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeID("testID"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(16385)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeCountOrIndex():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeCountOrIndex("testCountOrIndex"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(42)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert decomposer.scope["testCountOrIndex"] == 42

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeCountOrIndex("testCountOrIndex3"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(65)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeDuration():

    isLongRangeMode = False
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDuration("testDuration", isLongRangeMode=isLongRangeMode),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.000)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDuration("testDuration2", isLongRangeMode=isLongRangeMode),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.300)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDuration("testDuration2a", isLongRangeMode=isLongRangeMode),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.30012)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDuration("testDuration3", isLongRangeMode=isLongRangeMode),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(18.327)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDuration("testDuration4", isLongRangeMode=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3618.327)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeFrequency():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeFrequency("testFrequency"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(1250)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeFrequency("testFrequency2"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(1100)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeDistance():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(50)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(50.42)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=True),
        tokens.EndMarker("end"),
    ]
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push({"distance": 50.428, "isSmallScene": True})
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope["testDistance"]["distance"] == decomposer.scope["testDistance"]

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(350.428)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(9350.428)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(19350.428)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeDistance("testDistance", isSmallScene=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(9050.428)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeGain():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeGain("testGain", isHiPrecGain=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeGain("testGain", isHiPrecGain=False),
        tokens.writeGain("testGain2", isHiPrecGain=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    composer.push(5)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeGain("testGain", isHiPrecGain=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(4.2)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeGain("testGain", isHiPrecGain=False),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-42)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeGain("testGain", isHiPrecGain=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-4.2)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writePosition():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writePosition("testPosition", isSmallScene=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([12, 4, 7])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writePosition("testPosition", isSmallScene=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([-12, 4, -7])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writePosition("testPosition", isSmallScene=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([-50.428, 27.3, -7.001])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeOrientation():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeOrientation("testOrientation"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([0, 0, 0])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeOrientation("testOrientation"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([0, -90, 0])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeOrientation("testOrientation"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([-180, 60, -180])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeOrientation("testOrientation", highResolution=True),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([180, 60, 180])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeOrientation("testOrientation"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push([60, 90, 90])
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_FullSyntax():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeID("testID"),
        tokens.writeCountOrIndex("testCountOrIndex"),
        tokens.writeDuration("testDuration", isLongRangeMode=True),
        tokens.writeFrequency("testFrequency"),
        tokens.writeDistance("testDistance", isSmallScene=False),
        tokens.writeGain("testGain", isHiPrecGain=False),
        tokens.writePosition("testPosition", isSmallScene=True),
        tokens.writeOrientation("testOrientation"),
        tokens.writeID("testID2"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(42)
    composer.push(65)
    composer.push(3618.327)
    composer.push(1250)
    composer.push(350.428)
    composer.push(3)
    composer.push([-50.428, 27.3, -7.001])
    composer.push([60, 90, 90])
    composer.push(12)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_writeValueFromLUT():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="dsr_code"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-140)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="dsr_code"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-160)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert decomposer.scope["testwriteValueFromLUT"] == -150

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="dsr_code"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(-140.4)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert decomposer.scope["testwriteValueFromLUT"] == -140

    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeValueFromLUT("testwriteValueFromLUT", tableName="frequency_hop_codes"),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(pow(2, (1 / 3)))
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_repeat():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeCountOrIndex("testRepeatCount"),
        tokens.Repeat("testRepeatCount", syntax=[tokens.writeID("testRepeatID")]),
        tokens.EndMarker("end"),
    ]

    composer = BitstreamComposer(syntax)
    composer.push(5)
    for n in range(0, 5):
        composer.push(n)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def test_repeatWithIndex():
    syntax: List[tokens.Token] = [
        tokens.StartMarker("start"),
        tokens.writeCountOrIndex("forEachCount", storeValue=True),
        tokens.writeCountOrIndex("forEachCount", storeValue=True),
        tokens.writeCountOrIndex("idx"),
        tokens.repeatWithIndex("forEachCount", "idx", syntax=[tokens.writeID("testIDs")]),
        tokens.writeCountOrIndex("idx"),
        tokens.repeatWithIndex("forEachCount", "idx", syntax=[tokens.writeID("testIDs")]),
        tokens.EndMarker("end"),
    ]
    composer = BitstreamComposer(syntax)
    composer.push(3)
    composer.push(5)
    composer.push(0)
    for n in range(3):
        composer.push(42)
    composer.push(1)
    for n in range(5):
        composer.push(24)
    bitstream = composer.get_bitstream()
    decomposer = BitstreamDecomposer(bitstream, syntax)
    decomposer.extract_between_markers()
    assert composer.scope == decomposer.scope


def main():
    test_writeID()
    test_writeCountOrIndex()
    test_writeDuration()
    test_writeFrequency()
    test_writeDistance()
    test_writePosition()
    test_writeOrientation()
    test_writeGain()
    test_FullSyntax()
    test_writeValueFromLUT()
    test_repeat()
    test_repeatWithIndex()


if __name__ == "__main__":
    main()
