# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from bitarray import bitarray

from mpegi.framework import Bitstream
from mpegi.framework.generic_codebook import (
    HuffmanDecoder,
    SymbolHistogram,
    encode_symbols,
    get_encoder_config,
    get_fixed_length_config,
    get_huffman_config,
    serialize_codebook,
)


def test_codebook():
    plain_samples = [ord(ch) for ch in list("aasasdasdfasdfgasdfghasdfghjasdfghjk")]
    histogram = SymbolHistogram.from_plain(plain_samples)
    encoder = get_encoder_config(histogram)

    # check frequency map
    target_freq_map = {
        ord("a"): 8,
        ord("s"): 7,
        ord("d"): 6,
        ord("f"): 5,
        ord("g"): 4,
        ord("h"): 3,
        ord("j"): 2,
        ord("k"): 1,
    }
    freq_map = histogram.freq_map
    assert freq_map == target_freq_map

    # check codebook
    target_codebook = {
        ord("s"): bitarray("00"),
        ord("a"): bitarray("01"),
        ord("g"): bitarray("100"),
        ord("f"): bitarray("101"),
        ord("h"): bitarray("1100"),
        ord("k"): bitarray("11010"),
        ord("j"): bitarray("11011"),
        ord("d"): bitarray("111"),
    }
    codebook = encoder.codebook
    assert codebook == target_codebook

    # check decoding with code book
    target_plaintext = [ord(ch) for ch in list("asdfghjk")]
    bs = bitarray()
    bs.encode(codebook, target_plaintext)
    plaintext = bs.decode(codebook)
    assert plaintext == target_plaintext


def test_encodeFixedLength():
    plain_samples = [ord(ch) for ch in list("aasasdasdfasdfgasdfghasdfghjasdfghjk")]
    histogram = SymbolHistogram.from_plain(plain_samples)
    (encoder, num_bits) = get_fixed_length_config(histogram)

    # encoding
    target_plaintext = plain_samples
    bs_codebook = serialize_codebook(encoder)
    bs_data = encode_symbols(encoder, target_plaintext)

    # decoding
    decoder = HuffmanDecoder(bs_codebook)
    decoded_plaintext = decoder.decode(bs_data)
    assert decoded_plaintext == target_plaintext


def test_encodeHuffman():
    plain_samples = [ord(ch) for ch in list("aasasdasdfasdfgasdfghasdfghjasdfghjk")]
    histogram = SymbolHistogram.from_plain(plain_samples)
    (encoder, num_bits) = get_huffman_config(histogram)

    # encoding
    target_plaintext = plain_samples
    bs_codebook = serialize_codebook(encoder)
    bs_data = encode_symbols(encoder, target_plaintext)

    # decoding
    decoder = HuffmanDecoder(bs_codebook)
    decoded_plaintext = decoder.decode(bs_data)
    assert decoded_plaintext == target_plaintext


def test_encodeAuto():
    plain_samples = [ord(ch) for ch in list("aasasdasdfasdfgasdfghasdfghjasdfghjk")]
    histogram = SymbolHistogram.from_plain(plain_samples)
    encoder = get_encoder_config(histogram)

    # encoding
    target_plaintext = plain_samples
    bs_codebook = serialize_codebook(encoder)
    bs_data = encode_symbols(encoder, target_plaintext)

    # decoding
    decoder = HuffmanDecoder(bs_codebook)
    decoded_plaintext = decoder.decode(bs_data)
    assert decoded_plaintext == target_plaintext


def test_encodeSingleFixedLength():
    plain_samples = [ord(ch) for ch in list("aasasdasdfasdfgasdfghasdfghjasdfghjk")]
    histogram = SymbolHistogram.from_plain(plain_samples)
    (encoder, num_bits) = get_fixed_length_config(histogram)

    # encoding
    target_plaintext = plain_samples
    bs_codebook = serialize_codebook(encoder)
    bs_data = Bitstream()
    for symbol in plain_samples:
        bs_symbol = encode_symbols(encoder, [symbol])
        bs_data.appendBitstream(bs_symbol)

    # decoding
    decoder = HuffmanDecoder(bs_codebook)
    decoded_plaintext = []
    while len(bs_data) > 0:
        symbol = decoder.decodeSingle(bs_data)
        decoded_plaintext.append(symbol)
    assert decoded_plaintext == target_plaintext


def test_encodeSingleHuffman():
    plain_samples = [ord(ch) for ch in list("aasasdasdfasdfgasdfghasdfghjasdfghjk")]
    histogram = SymbolHistogram.from_plain(plain_samples)
    (encoder, num_bits) = get_huffman_config(histogram)

    # encoding
    target_plaintext = plain_samples
    bs_codebook = serialize_codebook(encoder)
    bs_data = Bitstream()
    for symbol in plain_samples:
        bs_symbol = encode_symbols(encoder, [symbol])
        bs_data.appendBitstream(bs_symbol)

    # decoding
    decoder = HuffmanDecoder(bs_codebook)
    decoded_plaintext = []
    while len(bs_data) > 0:
        symbol = decoder.decodeSingle(bs_data)
        decoded_plaintext.append(symbol)
    assert decoded_plaintext == target_plaintext


def test_voxelComponents():
    import numpy as np

    # generate some pseudo voxel indices given as voxel components
    num_voxels = 100
    vcX = list(np.random.randint(30, 60, size=num_voxels))
    vcY = list(np.random.randint(2, 5, size=num_voxels))
    vcZ = list(np.random.randint(10, 20, size=num_voxels))
    voxel_list = [(x, y, z) for x, y, z in zip(vcX, vcY, vcZ)]

    # register all values so that the Huffman encoder knows the distribution of the symbols
    vcX, vcY, vcZ = ([], [], [])
    for vc in voxel_list:
        vcX.append(vc[0])
        vcY.append(vc[1])
        vcZ.append(vc[2])

    # analyze symbol distributions
    histogram_VcX = SymbolHistogram.from_plain(vcX)
    histogram_VcY = SymbolHistogram.from_plain(vcY)
    histogram_VcZ = SymbolHistogram.from_plain(vcZ)

    # generate codebooks for voxel components
    encoderVcX = get_encoder_config(histogram_VcX)
    encoderVcY = get_encoder_config(histogram_VcY)
    encoderVcZ = get_encoder_config(histogram_VcZ)

    # write codebook headers to bitstream
    bs = Bitstream()
    bs.appendBitstream(serialize_codebook(encoderVcX))
    bs.appendBitstream(serialize_codebook(encoderVcY))
    bs.appendBitstream(serialize_codebook(encoderVcZ))

    # encode actual voxel component data sequentially
    for vc in voxel_list:
        vc_x_symbol = encode_symbols(encoderVcX, [vc[0]])
        vc_y_symbol = encode_symbols(encoderVcY, [vc[1]])
        vc_z_symbol = encode_symbols(encoderVcZ, [vc[2]])

        bs.appendBitstream(vc_x_symbol)
        bs.appendBitstream(vc_y_symbol)
        bs.appendBitstream(vc_z_symbol)
        # [...] add some more bitstream data that belongs to this voxel

    ####################################

    # read codebook headers from bitstream
    decoderVcX = HuffmanDecoder()
    decoderVcY = HuffmanDecoder()
    decoderVcZ = HuffmanDecoder()
    decoderVcX.deserializeCodebook(bs)
    decoderVcY.deserializeCodebook(bs)
    decoderVcZ.deserializeCodebook(bs)

    # read actual voxel component data sequentially
    decodedVcX = list()
    decodedVcY = list()
    decodedVcZ = list()
    while len(bs) > 0:
        x = decoderVcX.decodeSingle(bs)
        y = decoderVcY.decodeSingle(bs)
        z = decoderVcZ.decodeSingle(bs)
        decodedVcX.append(x)
        decodedVcY.append(y)
        decodedVcZ.append(z)
        # [...] read some more bitstream data that belongs to this voxel

    # check decoded values
    assert decodedVcX == vcX
    assert decodedVcY == vcY
    assert decodedVcZ == vcZ
