# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import pytest
from typing import List
import numpy as np
from numpy.testing import assert_almost_equal
from mpegi.metadata.eifplus import Portal
from mpegi.metadata.eif import Box, Mesh

from mpegi.metadata.reverb import MultiroomCalculator


def getHospitalPortals():
    geometryByID = {}
    geometryByID["geo:portal1"] = Box(id="geo:portal1", position=[-5.2, 1, 2.6], size=[1.2, 2, 0.2])
    geometryByID["geo:portal2"] = Box(
        id="geo:portal2", position=[-3.6, 1, 1.4], size=[1.2, 2, 0.2], orientation=[90.0, 0.0, 0.0]
    )
    geometryByID["geo:portal3"] = Box(id="geo:portal3", position=[0, 1, 2.6], size=[2.4, 2, 0.2])

    portals = []
    portals.append(
        Portal(
            id="prtl1:nurseroom2corridor",
            portalExtent="geo:portal1",
            portalNormal=np.array([0.0, 0.0, -1.0]),
            acousticEnvironment1="NurseRoom",
            acousticEnvironment2="Corridor",
            material="AcMat:PaintedWood",
            portalPosition=None,
            portalSize=None,
            hasExplicitPortal=True,
        )
    )
    portals.append(
        Portal(
            id="prtl2:nurseroom2patientroom",
            portalExtent="geo:portal2",
            portalNormal=np.array([-1.0, 0.0, 0.0]),
            acousticEnvironment1="NurseRoom",
            acousticEnvironment2="PatientRoom",
            material="AcMat:PaintedWood",
            portalPosition=None,
            portalSize=None,
            hasExplicitPortal=True,
        )
    )
    portals.append(
        Portal(
            id="prtl3:patientroom2corridor",
            portalExtent="geo:portal3",
            portalNormal=np.array([0.0, 0.0, -1.0]),
            acousticEnvironment1="PatientRoom",
            acousticEnvironment2="Corridor",
            material="AcMat:Air",
            portalPosition=None,
            portalSize=None,
            hasExplicitPortal=True,
        )
    )

    return portals, geometryByID


def getCathedralPortals():
    portals = []
    portals.append(
        Portal(
            id="prt:OutsideToPreRoom",
            portalExtent="mesh:cathedral_ARM_portalOuter",
            portalNormal=np.array([1.0, 0.0, 0.0]),
            acousticEnvironment1="acEnv:preRoom",
            acousticEnvironment2=None,
            material=None,
            portalPosition=None,
            portalSize=None,
            hasExplicitPortal=True,
        )
    )

    portals.append(
        Portal(
            id="prt:preRoomToMainRoom",
            portalExtent="mesh:cathedral_ARM_portalInner",
            portalNormal=np.array([-1.0, 0.0, 0.0]),
            acousticEnvironment1="acEnv:preRoom",
            acousticEnvironment2="acEnv:cathdral",
            material=None,
            portalPosition=None,
            portalSize=None,
            hasExplicitPortal=True,
        )
    )

    geometryById = {}
    geometryById["mesh:cathedral_ARM_portalInner"] = Mesh(
        id="mesh:cathedral_ARM_portalInner",
        position=np.array([-0.02, -0.04, -0.13]),
        orientation=np.array([0.00, 89.98, 0.00]),
    )
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-16.9335, -22.3535, -5.862])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-18.2657, -22.3535, -5.862])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-18.2657, -22.3535, 0.192])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-16.9335, -22.3535, 0.192])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-18.2657, -17.4082, -5.862])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-16.9335, -17.4082, -5.862])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-16.9335, -17.4082, 0.192])
    geometryById["mesh:cathedral_ARM_portalInner"].addVertex([-18.2657, -17.4083, 0.192])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([2, 1, 0])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([3, 2, 0])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([6, 5, 4])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([7, 6, 4])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([0, 1, 4])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([5, 0, 4])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([0, 5, 6])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([3, 0, 6])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([7, 4, 1])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([2, 7, 1])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([3, 6, 7])
    geometryById["mesh:cathedral_ARM_portalInner"].addFace([2, 3, 7])

    geometryById["mesh:cathedral_ARM_portalOuter"] = Mesh(
        id="mesh:cathedral_ARM_portalOuter",
        position=np.array([-24.64, 0.09, -0.20]),
        orientation=np.array([0.00, 89.98, 0.00]),
    )
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-3.9913, -22.2946, -6.8276])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-4.1779, -22.2946, -6.8276])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-4.1779, -22.2946, 0.1884])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-3.9913, -22.2946, 0.1884])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-4.1779, -17.3298, -6.8276])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-3.9913, -17.3298, -6.8276])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-3.9913, -17.3298, 0.1884])
    geometryById["mesh:cathedral_ARM_portalOuter"].addVertex([-4.1779, -17.3298, 0.1884])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([2, 1, 0])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([3, 2, 0])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([0, 1, 4])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([5, 0, 4])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([6, 5, 4])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([7, 6, 4])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([3, 0, 5])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([6, 3, 5])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([7, 2, 3])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([6, 7, 3])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([7, 4, 1])
    geometryById["mesh:cathedral_ARM_portalOuter"].addFace([2, 7, 1])

    return portals, geometryById


def test_calculateSourceContribution():
    sourcePos = [-4.193, 0.883, -1.976]
    portalPos = [-3.6, 1, 1.4]
    portalWidth = 1.2
    portalHeight = 2
    portalNormal = [-1.0, 0.0, 0.0]

    expectedContribution = 0.008519569628728

    calculator = MultiroomCalculator()

    sourceContribution = calculator.calculateSourceContribution(
        portalPos=portalPos,
        portalWidth=portalWidth,
        portalHeight=portalHeight,
        portalNormal=portalNormal,
        sourcePos=sourcePos,
    )

    assert_almost_equal(sourceContribution, expectedContribution, decimal=10)


def test_getPortalProperties():

    portals, geometryByID = getHospitalPortals()

    expectedPortalCorners = []
    expectedPortalCorners.append(
        [
            np.array([-4.6, 2, 2.6]),
            np.array([-5.8, 2, 2.6]),
            np.array([-4.6, 0, 2.6]),
            np.array([-5.8, 0, 2.6]),
        ]
    )

    expectedPortalCorners.append(
        [
            np.array([-3.6, 2, 0.8]),
            np.array([-3.6, 2, 2.0]),
            np.array([-3.6, 0, 0.8]),
            np.array([-3.6, 0, 2.0]),
        ]
    )

    expectedPortalCorners.append(
        [
            np.array([1.2, 2, 2.6]),
            np.array([-1.2, 2, 2.6]),
            np.array([1.2, 0, 2.6]),
            np.array([-1.2, 0, 2.6]),
        ]
    )

    calculator = MultiroomCalculator()
    for i in range(0, 3):
        portal = calculator.getPortalProperties(portals[i], geometryById=geometryByID)
        assert all((portal.portalCorners[0] - expectedPortalCorners[i][0]) < 0.00001)
        assert all((portal.portalCorners[1] - expectedPortalCorners[i][1]) < 0.00001)
        assert all((portal.portalCorners[2] - expectedPortalCorners[i][2]) < 0.00001)
        assert all((portal.portalCorners[3] - expectedPortalCorners[i][3]) < 0.00001)


def test_calculatePortalToPortalFactors_Hospital():

    # expectedResults = [fromPortalIndex, toPortalIndex, portal2portalFactor]
    expectedResults = [
        [1, 0, 0.063643],
        [0, 1, 0.074872],
        [2, 0, 0.000000],
        [0, 2, 0.000000],
        [2, 1, 0.027170],
        [1, 2, 0.037228],
    ]
    portals, geometryByID = getHospitalPortals()
    calculator = MultiroomCalculator()

    for portal in portals:
        portal = calculator.getPortalProperties(portal=portal, geometryById=geometryByID)

    for result in expectedResults:
        portal2portalFactor = calculator.calculatePortalToPortalFactor(
            fromPortal=portals[result[0]], toPortal=portals[result[1]]
        )
        assert_almost_equal(portal2portalFactor, result[2], decimal=6)


def test_calculatePortalToPortalFactors_Cathedral():

    expectedPortalCorners = []
    expectedPortalCorners.append(
        [
            np.array([-28.7246, 6.9116, -17.5297]),
            np.array([-28.7246, 6.9116, -22.4970]),
            np.array([-28.7246, -0.1062, -17.5297]),
            np.array([-28.7246, -0.1062, -22.4970]),
        ]
    )

    expectedPortalCorners.append(
        [
            np.array([-17.6196, 5.8159, -22.4856]),
            np.array([-17.6196, 5.8159, -17.5381]),
            np.array([-17.6196, -0.2398, -22.4856]),
            np.array([-17.6196, -0.2398, -17.5381]),
        ]
    )

    expectedResults = [[1, 0, 0.040010], [0, 1, 0.034408]]

    portals, geometryById = getCathedralPortals()
    calculator = MultiroomCalculator()

    for portal in portals:
        portal = calculator.getPortalProperties(portal=portal, geometryById=geometryById)

    for i in range(0, len(portals)):
        assert all((portals[i].portalCorners[0] - expectedPortalCorners[i][0]) < 0.0001)
        assert all((portals[i].portalCorners[1] - expectedPortalCorners[i][1]) < 0.0001)
        assert all((portals[i].portalCorners[2] - expectedPortalCorners[i][2]) < 0.0001)
        assert all((portals[i].portalCorners[3] - expectedPortalCorners[i][3]) < 0.0001)

    for result in expectedResults:
        portal2portalFactor = calculator.calculatePortalToPortalFactor(
            fromPortal=portals[result[0]], toPortal=portals[result[1]]
        )
        assert_almost_equal(portal2portalFactor, result[2], decimal=6)


def test_calculatePortalToPortalFactorMatrix_Hospital():
    expectedFactors = [
        [None, 0.074872, 0.000000],
        [0.063643, None, 0.037228],
        [0.000000, 0.027170, None],
    ]
    portals, geometryById = getHospitalPortals()
    calculator = MultiroomCalculator()
    p2pFactors = calculator.getPortal2PortalFactorMatrix(portals, geometryById)
    for i in range(0, len(p2pFactors)):
        for j in range(0, len(p2pFactors[0])):
            if p2pFactors[i][j] == None:
                assert p2pFactors[i][j] == expectedFactors[i][j]
            else:
                assert_almost_equal(p2pFactors[i][j], expectedFactors[i][j], decimal=6)


def test_calculatePortalToPortalFactorMatrix_Cathedral():
    expectedFactors = [
        [None, 0.034408],
        [0.040010, None],
    ]
    portals, geometryById = getCathedralPortals()
    calculator = MultiroomCalculator()
    p2pFactors = calculator.getPortal2PortalFactorMatrix(portals, geometryById)
    for i in range(0, len(p2pFactors)):
        for j in range(0, len(p2pFactors[0])):
            if p2pFactors[i][j] == None:
                assert p2pFactors[i][j] == expectedFactors[i][j]
            else:
                assert_almost_equal(p2pFactors[i][j], expectedFactors[i][j], decimal=6)


if __name__ == "__main__":
    test_calculateSourceContribution()
    test_getPortalProperties()
    test_calculatePortalToPortalFactors_Hospital()
    test_calculatePortalToPortalFactors_Cathedral()
    test_calculatePortalToPortalFactorMatrix_Hospital()
    test_calculatePortalToPortalFactorMatrix_Cathedral()
