# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
import pytest

from mpegi.framework import (
    QuantizationRange,
    decode_float,
    encode_float,
    get_max_quantization_error,
    round_value_to_range,
)


@pytest.mark.parametrize(
    ["value_to_encode", "result"],
    [(0.1, 0), (0.4, 0), (0.45, 0), (0.51, 1), (0.9, 1), (1.1, 1), (3.14, 3), (14.9, 15)],
)
def test_encode_float(value_to_encode: float, result: int):
    q_range = QuantizationRange(min=0, max=15, num_bits=4)

    encoded_float = encode_float(value_to_encode, q_range)

    assert encoded_float == result


def test_encoding_float_below_min_value_should_fail():
    q_range = QuantizationRange(min=0, max=15, num_bits=4)

    with pytest.raises(ValueError):
        encode_float(-1.23, q_range)


def test_encoding_float_above_max_value_should_fail():
    q_range = QuantizationRange(min=0, max=15, num_bits=4)

    with pytest.raises(ValueError):
        encode_float(15.1, q_range)


@pytest.mark.parametrize(
    ["value_to_decode", "result"],
    [(0.0, 0), (1.0, 1), (2.0, 2), (3.0, 3), (4.0, 4), (5.0, 5), (6.0, 6), (7.0, 7)],
)
def test_decode_float(value_to_decode: int, result: float):
    q_range = QuantizationRange(min=0, max=7, num_bits=3)

    decoded_float = decode_float(value_to_decode, q_range)

    assert decoded_float == result


def test_decode_float_additional():
    q_range = QuantizationRange(min=0.0, max=8.0, num_bits=8)

    decoded_float = decode_float(255, q_range)

    assert decoded_float == 8.0


@pytest.mark.parametrize(
    ["value", "result"],
    [
        (0.0, 0.0),
        (0.1, 0.0),
        (0.4, 0.0),
        (0.49, 0.0),
        (0.51, 1.0),
        (0.9, 1.0),
        (0.99, 1.0),
        (1.0, 1.0),
    ],
)
def test_round_to_value(value: float, result: float):
    q_range = QuantizationRange(min=0.0, max=7.0, num_bits=3)

    rounded_value = round_value_to_range(value, q_range)

    assert rounded_value == result


def test_decoding_below_min_range_should_fail():
    q_range = QuantizationRange(min=0, max=7, num_bits=3)

    with pytest.raises(ValueError):
        decode_float(-1, q_range)


def test_decoding_above_max_range_should_fail():
    q_range = QuantizationRange(min=0, max=7, num_bits=3)

    with pytest.raises(ValueError):
        decode_float(8, q_range)


def test_max_quantisation_error():
    q_range = QuantizationRange(min=0, max=15, num_bits=4)

    quant_error = get_max_quantization_error(q_range)

    assert quant_error == 0.5
