# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

import pytest

from mpegi.framework import QuantizationRange, SyntaxFormatter, tokens


def test_label():
    syntax: List[tokens.Token] = [tokens.Label("foo")]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO        | NO. OF BITS | MNEMONIC |\n"
        "| ---------- | ----------- | -------- |\n"
        "| `foo() { ` |             |          |\n"
        "| `}       ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_multipleLabels():
    syntax: List[tokens.Token] = [tokens.Label("foo"), tokens.Label("bar")]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO        | NO. OF BITS | MNEMONIC |\n"
        "| ---------- | ----------- | -------- |\n"
        "| `foo() { ` |             |          |\n"
        "| `}       ` |             |          |\n"
        "\n"
        "# bar\n"
        "\n"
        "| BAR        | NO. OF BITS | MNEMONIC |\n"
        "| ---------- | ----------- | -------- |\n"
        "| `bar() { ` |             |          |\n"
        "| `}       ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_flag():
    syntax = [tokens.Label("foo"), tokens.Flag("bar")]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO         | NO. OF BITS | MNEMONIC |\n"
        "| ----------- | ----------- | -------- |\n"
        "| `foo() {  ` |             |          |\n"
        "| `    bar; ` | 1           | bslbf    |\n"
        "| `}        ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_int():
    syntax = [tokens.Label("foo"), tokens.UnsignedInteger("bar", width=4)]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO         | NO. OF BITS | MNEMONIC |\n"
        "| ----------- | ----------- | -------- |\n"
        "| `foo() {  ` |             |          |\n"
        "| `    bar; ` | 4           | uimsbf   |\n"
        "| `}        ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_signedInt():
    syntax = [tokens.Label("foo"), tokens.SignedInteger("bar", width=12)]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO         | NO. OF BITS | MNEMONIC |\n"
        "| ----------- | ----------- | -------- |\n"
        "| `foo() {  ` |             |          |\n"
        "| `    bar; ` | 12          | tcimsbf  |\n"
        "| `}        ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


# def test_escapedInt():
#     syntax = [tokens.Label("foo"), tokens.EscapedInteger("bar", width=(2, 3, 4))]
#     formatter = SyntaxFormatter(syntax)

#     syntaxAsMarkdown = str(formatter)

#     expectedResult = (
#         "# MPEGI Bitstream Syntax\n"
#         "\n"
#         "# foo\n"
#         "\n"
#         "| FOO | NO. OF BITS | MNEMONIC |\n"
#         "| --- | --- | --- |\n"
#         "| `foo() { ` | | |\n"
#         "| `    bar ` | 2, 3, 4 | escval |\n"
#         "| `} ` | | |\n"
#         "\n"
#     )
#     assert syntaxAsMarkdown == expectedResult


def test_float():
    syntax = [tokens.Label("bar"), tokens.Float32("baz")]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# bar\n"
        "\n"
        "| BAR         | NO. OF BITS | MNEMONIC |\n"
        "| ----------- | ----------- | -------- |\n"
        "| `bar() {  ` |             |          |\n"
        "| `    baz; ` | 32          | float    |\n"
        "| `}        ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_cstring():
    syntax = [tokens.Label("bar"), tokens.CString("baz")]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# bar\n"
        "\n"
        "| BAR         | NO. OF BITS | MNEMONIC |\n"
        "| ----------- | ----------- | -------- |\n"
        "| `bar() {  ` |             |          |\n"
        "| `    baz; ` | 8..*        | vlclbf   |\n"
        "| `}        ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_blob():
    syntax = [tokens.Label("bar"), tokens.Blob("baz")]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# bar\n"
        "\n"
        "| BAR               | NO. OF BITS | MNEMONIC |\n"
        "| ----------------- | ----------- | -------- |\n"
        "| `bar() {        ` |             |          |\n"
        "| `    bazLength; ` | 32          | uimsbf   |\n"
        "| `    baz;       ` | bazLength   | bslbf    |\n"
        "| `}              ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_repeatWithInt():
    syntax = [tokens.Label("bar"), tokens.Repeat(3, [tokens.Flag("baz")])]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# bar\n"
        "\n"
        "| BAR                                  | NO. OF BITS | MNEMONIC |\n"
        "| ------------------------------------ | ----------- | -------- |\n"
        "| `bar() {                           ` |             |          |\n"
        "| `    for (int i = 0; i < 3; i++) { ` |             |          |\n"
        "| `        baz;                      ` | 1           | bslbf    |\n"
        "| `    }                             ` |             |          |\n"
        "| `}                                 ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_repeatWithVariableReference():
    syntax = [
        tokens.Label("foo"),
        tokens.UnsignedInteger("bar", width=4),
        tokens.Repeat("bar", [tokens.Flag("baz")]),
    ]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO                                    | NO. OF BITS | MNEMONIC |\n"
        "| -------------------------------------- | ----------- | -------- |\n"
        "| `foo() {                             ` |             |          |\n"
        "| `    bar;                            ` | 4           | uimsbf   |\n"
        "| `    for (int i = 0; i < bar; i++) { ` |             |          |\n"
        "| `        baz;                        ` | 1           | bslbf    |\n"
        "| `    }                               ` |             |          |\n"
        "| `}                                   ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_nestedRepeats():
    syntax = [
        tokens.Label("foo"),
        tokens.Repeat(
            3,
            [
                tokens.Repeat(
                    4,
                    [
                        tokens.Repeat(
                            5,
                            [
                                tokens.UnsignedInteger("bar", width=4),
                            ],
                        )
                    ],
                )
            ],
        ),
    ]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO                                          | NO. OF BITS | MNEMONIC |\n"
        "| -------------------------------------------- | ----------- | -------- |\n"
        "| `foo() {                                   ` |             |          |\n"
        "| `    for (int i = 0; i < 3; i++) {         ` |             |          |\n"
        "| `        for (int j = 0; j < 4; j++) {     ` |             |          |\n"
        "| `            for (int k = 0; k < 5; k++) { ` |             |          |\n"
        "| `                bar;                      ` | 4           | uimsbf   |\n"
        "| `            }                             ` |             |          |\n"
        "| `        }                                 ` |             |          |\n"
        "| `    }                                     ` |             |          |\n"
        "| `}                                         ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_conditional():
    syntax = [
        tokens.Label("foo"),
        tokens.Flag("bar"),
        tokens.Conditional(
            "bar",
            true=[tokens.Flag("baz")],
            false=[tokens.UnsignedInteger("bam", width=4)],
        ),
    ]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO               | NO. OF BITS | MNEMONIC |\n"
        "| ----------------- | ----------- | -------- |\n"
        "| `foo() {        ` |             |          |\n"
        "| `    bar;       ` | 1           | bslbf    |\n"
        "| `    if (bar) { ` |             |          |\n"
        "| `        baz;   ` | 1           | bslbf    |\n"
        "| `    }          ` |             |          |\n"
        "| `    else {     ` |             |          |\n"
        "| `        bam;   ` | 4           | uimsbf   |\n"
        "| `    }          ` |             |          |\n"
        "| `}              ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_conditional_with_one_empty_condition():
    syntax = [
        tokens.Label("foo"),
        tokens.Flag("bar"),
        tokens.Conditional("bar", true=[tokens.Flag("baz")], false=[]),
    ]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO               | NO. OF BITS | MNEMONIC |\n"
        "| ----------------- | ----------- | -------- |\n"
        "| `foo() {        ` |             |          |\n"
        "| `    bar;       ` | 1           | bslbf    |\n"
        "| `    if (bar) { ` |             |          |\n"
        "| `        baz;   ` | 1           | bslbf    |\n"
        "| `    }          ` |             |          |\n"
        "| `}              ` |             |          |\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


def test_switch():
    syntax: List[tokens.Token] = [
        tokens.Label("switch"),
        tokens.UnsignedInteger("foo"),
        tokens.Switch(
            "foo",
            cases=[
                tokens.SwitchCase(0, [tokens.Flag("bar")]),
                tokens.SwitchCase(1, [tokens.Flag("baz")]),
            ],
        ),
    ]
    formatter = SyntaxFormatter(syntax)

    syntax_as_markdown = str(formatter)

    expected_result = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# switch\n"
        "\n"
        "| SWITCH               | NO. OF BITS | MNEMONIC |\n"
        "| -------------------- | ----------- | -------- |\n"
        "| `switch() {        ` |             |          |\n"
        "| `    foo;          ` | 8           | uimsbf   |\n"
        "| `    switch(foo) { ` |             |          |\n"
        "| `    case 0: {     ` |             |          |\n"
        "| `        bar;      ` | 1           | bslbf    |\n"
        "| `        break;    ` |             |          |\n"
        "| `    }             ` |             |          |\n"
        "| `    case 1: {     ` |             |          |\n"
        "| `        baz;      ` | 1           | bslbf    |\n"
        "| `        break;    ` |             |          |\n"
        "| `    }             ` |             |          |\n"
        "| `    }             ` |             |          |\n"
        "| `}                 ` |             |          |\n"
        "\n"
    )
    assert syntax_as_markdown == expected_result


def test_quantized_float():
    syntax: List[tokens.Token] = [
        tokens.Label("foo"),
        tokens.QuantizedFloat("bar", quant_range=QuantizationRange(min=-8.0, max=8.0, num_bits=8)),
    ]
    formatter = SyntaxFormatter(syntax)

    syntaxAsMarkdown = str(formatter)

    expectedResult = (
        "# MPEGI Bitstream Syntax\n"
        "\n"
        "# foo\n"
        "\n"
        "| FOO         | NO. OF BITS | MNEMONIC |\n"
        "| ----------- | ----------- | -------- |\n"
        "| `foo() {  ` |             |          |\n"
        "| `    bar; ` | 8           | uimsbf   |\n"
        "| `}        ` |             |          |\n"
        "\n"
        "## bar\n"
        "integerized floating point representing a range between:\n"
        "min=-8.0f\n"
        "max=8.0f\n"
        "\n"
    )
    assert syntaxAsMarkdown == expectedResult


# ---- Test Incorrect Cases ---- #
#                                #
#                                #
# ------------------------------ #


def test_breakIfFirstElementIsNotAlabel():
    syntax: List[tokens.Token] = [tokens.UnsignedInteger("bar", width=4)]
    formatter = SyntaxFormatter(syntax)

    with pytest.raises(SyntaxError):
        str(formatter)
