/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ivas_limiter.h"

#include <math.h>
#include <stdio.h>

/*----------------------------------------------------------------------------------*
 * Local constants
 *----------------------------------------------------------------------------------*/

#define EPSILON 0.000000000000001f

/*-------------------------------------------------------------------*
 * limiter_mc()
 *
 * In-place saturation control for multichannel buffers with adaptive release time
 *-------------------------------------------------------------------*/

void limiter_mc(
    float** output,                                    /* i/o: input/output buffer                                 */
    float limiter_mem[2],                              /* i/o: limiter_mem[2] should be initialized to [1.f, 0.f]  */
    const int16_t output_frame,                        /* i  : number of samples per channel in the buffer         */
    const int16_t nchan_out,                           /* i  : number of channels in the buffer                    */
    const int32_t output_Fs,                           /* i  : sampling rate                                       */
    const float attack_seconds,                        /* i  : attack in seconds                                        */
    const float threshold                              /* i  : threshold                                           */
)
{
    int16_t i, c;
    float tmp, max_val;
    float *sample;
    float gain, frame_gain, attack_constant, release_constant;
    float releaseHeuristic;

    /*-----------------------------------------------------------------*
     * Calculate gain factor to avoid saturation
     *-----------------------------------------------------------------*/
    max_val = 0.f;

    for ( i = 0; i < output_frame; i++ )
    {
        for ( c = 0; c < nchan_out; c++ )
        {
            tmp = fabsf( output[c][i] );
            if ( tmp > max_val )
            {
                max_val = tmp;
            }
        }
    }

    /* Release heuristic
     * 
     * Value ranging from 0.f to 1.f. Increases on each frame that contains
     * a sample with a value above threshold and decreases on each frame
     * with all sample values below threshold.
     * 
     * Values of 0 and 1 map to the shortest and longest release time, respectively.
     * 
     * The goal of this heuristic is to avoid the "pumping" effect when only
     * sharp transients exceed the threshold (use short release time), but also
     * keep the gain curve smoother if the threshold is exceeded in many frames
     * in a short span of time.
     */
    releaseHeuristic = limiter_mem[1];

    if ( max_val > threshold )
    {
        frame_gain = threshold / max_val;

        releaseHeuristic = fminf( 1.f, releaseHeuristic + ( 4.f * output_frame / output_Fs ) );
        /* Unoptimized code for reference */
        /* releaseHeuristic = min( 1.f, releaseHeuristic + ( (float) 2.f * output_frame / output_Fs / adaptiveReleaseWindowLengthInSeconds ) );
         *                                                            ^
         *                                                          React faster when release time should be increased
         */
    }
    else
    {
        releaseHeuristic = fmaxf( 0.f, releaseHeuristic - ( (float) output_frame / output_Fs ) );
        /* Unoptimized code for reference */
        /* releaseHeuristic = max( 0.f, releaseHeuristic - ( (float) 0.5f * output_frame / output_Fs / adaptiveReleaseWindowLengthInSeconds ) );
         *                                                            ^
         *                                                          React slower when release time should be decreased
         */

        /* No samples above threshold and gain from previous frame is already 1.f,
         * therefore gain == 1.f for the entire frame. Skip processing. */
        if ( *limiter_mem >= 1.f - EPSILON )
        {
            limiter_mem[0] = 1.f;
            limiter_mem[1] = releaseHeuristic;
            return;
        }

        /* No samples above threshold but gain from previous frame is not 1.f,
         * transition to gain == 1.f */
        frame_gain = 1.f;
    }

    /* Get last gain value from previous frame */
    gain = limiter_mem[0];

    /* 99% time constants of the gain curve
     * 
     * The denominator of the second argument determines after how many
     * samples the gain curve will reach 99% of its target value
     */
    attack_constant = powf( 0.01f, 1.0f / ( attack_seconds * output_Fs ) );
    release_constant = powf( 0.01f, 1.0f / ( 0.005f * powf( 200.f, releaseHeuristic ) * output_Fs ) );

    /* Unoptimized code for reference */
    /* releaseTimeInSeconds = 0.005f * powf(200.f, releaseHeuristic); <-- Map heuristic value (0; 1) exponentially to range (0.005; 1)
     * release_constant = powf( 0.01f, 1.0f / ( releaseTimeInSeconds *   ) );
     */

    /*-----------------------------------------------------------------*
     * Apply limiting
     *-----------------------------------------------------------------*/
    for ( i = 0; i < output_frame; i++ )
    {
        /* update gain */
        if ( frame_gain < gain )
        {
            gain = attack_constant * ( gain - frame_gain ) + frame_gain;
        }
        else
        {
            gain = release_constant * ( gain - frame_gain ) + frame_gain;
        }

        for ( c = 0; c < nchan_out; c++ )
        {
            sample = &output[c][i];

            /* Apply gain */
            *sample = gain * ( *sample );

            if ( *sample > 1.0f )
            {
                *sample = 1.0f;
            }
            else if ( *sample < -1.0f )
            {
                *sample = -1.0f;
            }
        }
    }

    /* Save last gain and release heuristic values for next frame */
    limiter_mem[0] = gain;
    limiter_mem[1] = releaseHeuristic;

    return;
}
