/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_AIRFLOW_BASE
#define EAR_AIRFLOW_BASE

#include <EAR/BaseTypes.h>
#include <EAR/PchipInterpolator.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

#include <array>
#include <memory>
#include <ostream>
#include <vector>

namespace EAR {

// Convenience function to translate a vector in world coordinates into the rotated frame specified by the Euler angles.
// NOTE: {0,0,-1} = "forward", {0,1,0} = "up", {1, 0, 0} = "right".
inline Vector3 eulerToVec3(const Euler& euler, const Vector3& worldVector = { 0, 0, -1 })
{
    const Quaternion quaternion(euler);
    return quaternion.rotate(worldVector);
}

// Computes the relative coordinates that target is relative to reference. I.e. if the origin of the coordinate system was defined to be reference, what would the coordinates of target be?
inline Vector3 getRelativePosition(const Position& reference, const Vector3& target)
{
    const HomogenMatrix refHomogenMatrix(reference);

    return refHomogenMatrix.inverseGeometric() * target;
}

// Gets the orientation airflow (given in global coordinates) is coming from, relative to listenerOrientation.
// NOTE: This returns the direction the airflow comes from, so this is the opposite direction of airflow (which points in the direction it is blowing towards).
inline Polar getAirflowOrigin(const Euler listenerOrientation, const Vector3& airflow)
{
    return cartesianToPolar(
        getRelativePosition({ { 0, 0, 0 }, listenerOrientation }, -1 * airflow));
}

class AirflowGeneratorBase : public SceneObject
{
public:
    //! Constructor
    AirflowGeneratorBase(ObjectID oid, std::string name = "", Position pos = Position())
        : SceneObject(oid, name, pos)
    {
        staticFlag.store(true); // geometry (in contrast to other SceneObjects) is static by default
    };

    //! Destructor
    virtual ~AirflowGeneratorBase(){};

    //! Common epsilon for tests etc
    static const float EPSILON;

    //! Seperate different types of geometry
    enum class Type
    {
        Global,
        Point,
        Cone,
        Plane
    };

    // Computes the airspeed experienced at listenerPosition. Result is in world coordinates.
    virtual Vector3 getAirspeed(const Vector3& listenerPosition) const = 0;

    //! Returns the type of the geometry
    virtual Type getType() const = 0;

    //! Get gain value of SceneObject (thread-safe)
    /*!
     * \return          current gain in dB (may return -inf, if factor is 0)
     */
    float getGainDb() const { return gaindB.load(); }

    //! Get gain value of SceneObject (thread-safe)
    /*!
     * \return          current gain as a factor (not dB!)
     */
    float getGainFactor() const { return gain.load(); }

    //! Set gain value of SceneObject (thread-safe)
    /*!
     * Default is 0 dB. It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * But you can set it to -inf if you want.
     * \param newGain  new gain in dB
     */
    void setGainDb(float newGain)
    {
        if (newGain != gaindB.load()) {
            float newFactor = std::pow(10, newGain / 20);
            gain.store(newFactor);
            gaindB.store(newGain);
            moved.store(true);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

    //! Set gain value of SceneObject (thread-safe)
    /*!
     * Default is 1. It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * \param newFactor     new gain as a factor (not dB!)
     */
    void setGainFactor(float newFactor)
    {
        if (newFactor != gain.load()) {
            float newGain = -1 * std::numeric_limits<float>::infinity();

            if (newFactor > 0) {
                newGain = 20 * std::log10(newFactor);
            }

            gain.store(newFactor);
            gaindB.store(newGain);
            moved.store(true);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

    void setSpeed(float newSpeed)
    {
        if (newSpeed != speed.load()) {
            speed.store(newSpeed);
            notify(SceneObjectObserver::Property::Speed);
        }
    }

    float getSpeed() const { return speed.load(); }

    void setDistanceAttenuationFactor(float newVal)
    {
        if (newVal != distanceAttenuationFactor.load()) {
            hasDistanceAttenuationFactor = true;
            distanceAttenuationFactor.store(newVal);
        }
    }

    float getDistanceAttenuationFactor() const { return distanceAttenuationFactor.load(); }

    bool hasDistanceAttenuationFactor{ false };

    void setStringId(std::string id) { this->stringID = id; }

    void setBsId(int id) { this->bsID = id; }

    int getBsId() { return bsID; }

    std::string getStringId() { return stringID; }

private:
    //! gain factor (not dB!) that shall be applied in a rendering scenario
    std::atomic<float> gain{ 1.f };

    //! gain in dB (factor and dB are stored to make either get function cheaper)
    std::atomic<float> gaindB{ 0.f };

    //! speed in m/s.
    std::atomic<float> speed{ 0.0f };

    std::atomic<float> distanceAttenuationFactor{ 1.0f };

    std::string stringID{ "" };
    std::atomic<int> bsID{ 0 };
};

} /* namespace EAR*/

#endif /* EAR_GEOMETRY_BASE */
