/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_ANCHOR_OBJECT
#define EAR_ANCHOR_OBJECT

#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

namespace EAR {
//! Represents an anchor object with no acoustic functionality
/*!
 * AnchorObjects can be used as parents for other SceneObjects to allow a spatial hierarchy
 * and joint transforms of multiple SceneObjects.
 */
class AnchorObject : public SceneObject
{
public:
    //! Constructor (only accessible through SceneObjectFactory)
    /*!
     * \param oid           fixed object ID (always assigned by factory method!)
     * \param name          arbitrary name of object (optional, defaults to "")
     * \param pos           Position struct with initial location and orientation (optional, defaults to 0,0,0,0,0,0)
     * \param autoRescale   Enable automatic rescaling of audio elements (optional, defaults to false)
     */
    AnchorObject(ObjectID oid, std::string name = "", Position pos = Position())
        : SceneObject(oid, name, pos){};
    //! Destructor
    virtual ~AnchorObject(){};
    // factory method can access private constructor
    friend SceneObjectFactory<AnchorObject>;

    //! Check if the autoRescale is enabled
    /*!
     * If set to true, the renderer rescales the source positions
     * \return          
     */
    bool isAutoRescale() const;

    //! Change the state of autoRescale
    /*!
     * If set to true, the renderer rescales the source positions
     * \return          
     */
    virtual void setAutoRescale(bool newState);

    //! Get the values of tolerance
    /*!
     * Tolerance is the percentage of scene region that remains outside of LSDF region
     * \return          
     */
    std::vector<float> getTolerance();

    //! Change the values of tolerance
    /*!
     * Sets the percentage of scene region that remains outside of LSDF region
     * \return          
     */
    virtual void setTolerance(std::vector<float> newTolerance);

protected:
    //! if set to true it enables the renderer to modify, based on listener's playback dimension
    // and the source dimension, a new location of the source.
    std::atomic<bool> autoRescale{ false };

    //!
    //
    std::vector<float> tolerance{ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
};

} // namespace EAR

#endif /* EAR_ANCHOR_OBJECT */
