/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_AUDIOFILE
#define EAR_AUDIOFILE

#include <EAR/RealFrame.h>
#include <memory>
#include <string>

namespace EAR {

//! read/write for wav files (useful for debugging, etc.)
/*!
 * AudioFile can read and write 8/16/24 bit uncompressed wavs into and from an internal Frame.
 * When writing to file, values should be clipped to the range [-1.0,1.0] to avoid integer overflows. AudioFile does NOT take care of this.
 */
class AudioFile
{
public:
    //! Constructor
    AudioFile();

    //! Destructor
    ~AudioFile();

    //! load a wav file from disk
    /*!
     * \param path    Absolute file path
     */
    void load(const std::string& path);

    //! save internal RealFrame contents to wav file on disk
    /*!
     * \param path          Absolute file path (overwrites existing files)
     * \param _bitDepth     Bit depth of written file (float values in Frame are converted accordingly)
     * \param _sampleRate   Sample rate of Frame data [Hz]
     */
    void save(const std::string& path, int _bitDepth, float _sampleRate);

    //! return sample rate of loaded file [Hz]
    float getSampleRate() const;

    //! return bit depth of loaded file
    int getBitDepth() const;

    //! return pointer to internal Frame
    RealFrame* getFrame();

protected:
    //! internal Frame representation
    RealFrame frame;
    //! sample rate of loaded file (not used for saving)
    float sampleRate = 0.f;
    //! bit depth of loaded file (not used for saving)
    int bitDepth = 0;

private:
    //! find string in byte representation of loaded file
    int findString(const std::string& needle, const std::vector<uint8_t>& haystack);

    //! 16 bit read
    int16_t readTwoBytes(std::vector<uint8_t>& data, int pos);

    //! 24 bit read
    int32_t readThreeBytes(std::vector<uint8_t>& data, int pos);

    //! 32 bit read
    int32_t readFourBytes(std::vector<uint8_t>& data, int pos);

    //! read sample in 32 bit IEEE float format
    float readFloatSample(std::vector<uint8_t>& data, int pos);

    //! read sample according to bit depth of file
    float readSample(std::vector<uint8_t>& data, int pos);

    //! create byte representation of string
    void stringToData(std::vector<uint8_t>& data, std::string val);

    //! create byte representation of 16 bit int
    void int16ToData(std::vector<uint8_t>& data, int16_t val);

    //! create byte representation of 24 bit int
    void int24ToData(std::vector<uint8_t>& data, int32_t val);

    //! create byte representation of 32 bit int
    void int32ToData(std::vector<uint8_t>& data, int32_t val);
};

} // namespace EAR

#endif /* EAR_AUDIOFILE */
