/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_BASETYPES
#define EAR_BASETYPES

#include <EAR/Exception.h>

#include <array>
#include <cmath>
#include <iterator>
#include <numeric>
#include <ostream>
#include <vector>

// TODO: make this configurable
#define SIMD_ALIGN alignas(16)

namespace EAR {

class Matrix3_3; //forward declare
//! Coordinate system declarator
/*!
 * Indicates the frame of reference of a position (location + orientation).
 */
enum class CSpace
{
    WCS, //!< World Coordinate System
    RCS, //!< Parent-relative Coordinate System
    LCS, //!< Listener-relative Coordinate System
};

//! stringify CSpace when piped into ostream
/*!
 * \param os    output stream
 * \param cs    coordinate system to be printed
 */
std::ostream& operator<<(std::ostream& os, const CSpace& cs);

//! A 3d float vector representation that supports common algebraic expressions.
/*!
 * In EAR, they are mainly used for the cartesian representation of points in 3D space. EAR uses a right-handed coordinate system in unison with for example OpenGL.
 * This results in the following axis directions:
 * 1. the x-axis with positive values increasing to the right
 * 2. the y-axis with positive values increasing upwards
 * 3. the z-axis with positive values increasing to the back (out of the screen)
 *
 */
class SIMD_ALIGN Vector3
{

public:
    //! Empty Constructor (zero-initializing)
    Vector3() { set(0, 0, 0); };

    //! Constructor to initialize from single component values
    /*!
     * \param _x   first component
     * \param _y   second component
     * \param _z   third component
     */
    Vector3(float _x, float _y, float _z) { set(_x, _y, _z); };

    //! Constructor to initialize from float array of length 3
    /*!
     * \param _val   float array of length 3
     */
    Vector3(float _val[3]) { set(_val); };

    //! Constructor to initialize from an std::array<T, 3>
    /*!
     * \param _val   std::array<T, 3> where T is convertible to float
     */
    template <typename T>
    Vector3(std::array<T, 3> const& _val)
        : Vector3(
            static_cast<float>(_val[0]), static_cast<float>(_val[1]), static_cast<float>(_val[2]))
    {
        static_assert(std::is_convertible_v<T, float>, "Must be convertible to float");
    }

    //! Copy constructor
    Vector3(const Vector3& rhs) = default;

    //! Destructor
    ~Vector3(){};

    //! Set all components from outside
    /*!
     * \param _x   first component
     * \param _y   second component
     * \param _z   third component
     */
    void set(float _x, float _y, float _z);

    //! Set all components from outside
    /*!
     * \param _val   float array of length 3
     */
    void set(float _val[3]);

    /*Vector3& operator=(const Vector3& other); // copy
    Vector3& operator=(Vector3&& other); // move*/

    //! perform vector addition
    /*!
     * \param rhs   other Vector3 to be added
     */
    Vector3& operator+=(const Vector3& rhs);

    //! perform vector subtraction
    /*!
     * \param rhs   other Vector3 to be subtracted
     */
    Vector3& operator-=(const Vector3& rhs);

    //! perform vector multiplication (cross product!)
    /*!
     * \param rhs   other Vector3 to be multiplied
     */
    Vector3& operator*=(const Vector3& rhs);

    //! perform vector scaling
    /*!
     * \param rhs   scalar value
     */
    Vector3& operator*=(float rhs);

    //! direct access to components
    /*!
     * \param idx   index (0, 1, 2, 3)
     *                    (index 3 always returns 1)
     */
    inline float& operator[](int idx) { return val[idx]; };

    //! direct const access to components
    /*!
     * \param idx   index (0, 1, 2, 3)
     *                    (index 3 always returns 1)
     */
    inline float operator[](int idx) const { return val[idx]; };

    friend Vector3 operator+(Vector3 lhs, const Vector3& rhs);
    friend Vector3 operator-(Vector3 lhs, const Vector3& rhs);
    friend Vector3 operator*(Vector3 lhs, const Vector3& rhs);
    friend Vector3 operator*(Vector3 lhs, float rhs);
    friend Vector3 operator-(const Vector3& rhs);

    //! calculate dot product and return new Vector3 with result
    /*!
     * \param rhs   other Vector3 to be multiplied
     */
    float dot(const Vector3& rhs) const;

    //! calculate cross product and return new Vector3 with result
    /*!
     * \param rhs   other Vector3 to be multiplied
     */
    Vector3 cross(const Vector3& rhs) const;

    //! Returns true when two Vector3 are within a tolerance value and false otherwise
    /*!
     * \param rhs   other Vector3 to be compared against
     * \param tolerance float value that must be exceeded for return value to be true.
     */
    bool isClose(const Vector3& rhs, float tolerance) const;

    //! Normalize vector (abs value 1) and return new Vector3 with result
    Vector3 normalize() const;

    //! get absolute value of vector (in Euclidian sense)
    float abs() const;

    //! get raw pointer to underlying data
    inline float* data() { return &val[0]; }

    //! get const raw pointer to underlying data
    inline const float* data() const { return &val[0]; }

    //! mirror a point on a plane given in Hesse normal form
    /*!
     * \param N0        normal vector of the plane
     * \param d         distance to the origin
     */
    Vector3 mirror(const Vector3& N0, float d) const;

    //! Conversion to std::array
    operator std::array<float, 3>() const { return { val[0], val[1], val[2] }; }

private:
    //! internal data store (fourth value is always 1 to allow homogeneous transformation)
    float val[4];
};

//! perform vector addition and return new Vector3
/*!
 * \param lhs   first Vector3 to be added
 * \param rhs   second Vector3 to be added
 */
Vector3 operator+(Vector3 lhs, const Vector3& rhs);

//! perform vector subtraction and return new Vector3
/*!
 * \param lhs   first Vector3 to be subtracted from
 * \param rhs   second Vector3 to subtract
 */
Vector3 operator-(Vector3 lhs, const Vector3& rhs);

//! perform vector multiplication (cross product!) and return new Vector3
/*!
 * \param lhs   first Vector3 to be multiplied
 * \param rhs   second Vector3 to be multiplied
 */
Vector3 operator*(Vector3 lhs, const Vector3& rhs);

//! Scale Vector3 by scalar (duh)
/*!
 * \param lhs   Vector3 to be scaled
 * \param rhs   scalar value
 */
Vector3 operator*(Vector3 lhs, float rhs);

//! Scale Vector3 by scalar (duh)
/*!
 * \param lhs   scalar value
 * \param rhs   Vector3 to be scaled
 */
Vector3 operator*(float lhs, const Vector3& rhs);

//! stringify Vector3 when piped into ostream
/*!
 * \param os    output stream
 * \param vec   Vector3 to be printed
 */
std::ostream& operator<<(std::ostream& os, const Vector3& vec);

//! boolean compare
bool operator==(const Vector3& lhs, const Vector3& rhs);

//! boolean compare
bool operator!=(const Vector3& lhs, const Vector3& rhs);

//! Stores Euler angles (yaw, pitch, roll) in Tait-Bryant convention in degrees from 0-360.
/*!
 * According to the right-handed coordinate system:
 * 1. yaw is left-positive (counter-clockwise rotation around the y-axis)
 * 2. pitch is up-positive (clockwise rotation around the x-axis)
 * 3. roll is right-positive (clockwise rotation around the z-axis)
 *
 * Rotations shall be performed in the order above.
 */
class Euler
{
public:
    //! Empty Constructor (zero-initializing)
    Euler() { set(0, 0, 0); };

    //! Constructor to initialize from single angles
    /*!
     * \param y   yaw (0 to 360 degrees)
     * \param p   pitch (0 to 360 degrees)
     * \param r   roll (0 to 360 degrees)
     */
    Euler(float y, float p, float r) { set(y, p, r); };

    //! Constructor to initialize from float array of length 3
    /*!
     * \param _angles   float array of length 3
     */
    Euler(float _angles[3]) { set(_angles); };

    //! Copy constructor
    Euler(const Euler& rhs) = default;

    //! Destructor
    ~Euler(){};

    //! Set all angles from outside
    /*!
     * \param y   yaw (0 to 360 degrees)
     * \param p   pitch (0 to 360 degrees)
     * \param r   roll (0 to 360 degrees)
     */
    void set(float y, float p, float r);

    //! Set all angles from outside
    /*!
     * \param _angles   float array of length 3
     */
    void set(float _angles[3]);

    // Returns the minimum angle needed to rotate between this and rhs in degrees.
    float angleBetween(const Euler& rhs) const;

    //! Returns true when the minimum rotation angle to go from this to rhs are within a threshold value and false otherwise
    /*!
     * \param rhs       other Euler to be compared against.
     * \param threshold Returns true if the angle between this and rhs is smaller than threshold (in degrees) and false otherwise.
     */
    bool isClose(const Euler& rhs, float threshold) const;

    //! direct access to components
    /*!
     * \param idx   index (0 => yaw, 1 => pitch, 2 => roll)
     */
    float& operator[](int idx) { return angles[idx]; };

    //! direct const access to components
    /*!
     * \param idx   index (0 => yaw, 1 => pitch, 2 => roll)
     */
    float operator[](int idx) const { return angles[idx]; };

    //! "Adding" Euler angles. Actually composes the rotation that would be induced by the Euler angles, and give the Euler angles of the resulting rotation. Note that the composed rotation consists of first rotating by rhs, and then by *this. Also note operator+ is not commutative.
    Euler operator+(const Euler& rhs);
    //! get raw pointer to underlying data
    float* data() { return &angles[0]; }

    //! get const raw pointer to underlying data
    const float* data() const { return &angles[0]; }

private:
    //! internal data store
    float angles[3];
};

//! stringify Euler when piped into ostream
/*!
 * \param os    output stream
 * \param eul   Euler to be printed
 */
std::ostream& operator<<(std::ostream& os, const Euler& eul);

//! boolean compare
//! NOTE: Does simple elementwise comparison of the euler angles, and does not take into account equivalent (but different) euler angle representations of the same rotation. Use angleBetween or closeTo functions for a proper comparison.
bool operator==(const Euler& lhs, const Euler& rhs);

//! boolean compare
//! NOTE: Does simple elementwise comparison of the euler angles, and does not take into account equivalent (but different) euler angle representations of the same rotation. Use angleBetween or closeTo functions for a proper comparison.
bool operator!=(const Euler& lhs, const Euler& rhs);

//! Stores a rotation in Quaternion format
class Quaternion
{
public:
    //! Empty Constructor (zero-initializing)
    Quaternion() { set(0, 0, 0); };

    //! Constructor to initialize from existing quaternion values
    /*!
     * \param w   real component
     * \param x   i component
     * \param y   j component
     * \param z   k component
     */
    Quaternion(float w, float x, float y, float z) { set(w, x, y, z); };

    //! Constructor to initialize from single angles
    /*!
     * \param y   yaw (0 to 360 degrees)
     * \param p   pitch (0 to 360 degrees)
     * \param r   roll (0 to 360 degrees)
     */
    Quaternion(float y, float p, float r) { set(y, p, r); };

    //! Constructor to initialize from single angles in array
    /*!
     * \param vals  float[3] array of yaw, pitch, roll angles
     */
    Quaternion(float vals[3]) { set(vals[0], vals[1], vals[2]); };

    //! Constructor to initialize from Euler type
    /*!
     * \param angles   EAR::Euler instance
     */
    Quaternion(const Euler& angles) { set(angles); };

    //! Copy constructor
    Quaternion(const Quaternion& rhs) = default;

    //! Destructor
    ~Quaternion(){};

    //! Set quaternion to represent same rotation as given yaw, pitch, roll angles
    /*!
     * \param y   yaw (0 to 360 degrees)
     * \param p   pitch (0 to 360 degrees)
     * \param r   roll (0 to 360 degrees)
     */
    void set(float y, float p, float r);

    //! Set quaternion from existing quaternion values
    /*!
     * \param w   real component
     * \param x   i component
     * \param y   j component
     * \param z   k component
     */
    void set(float w, float x, float y, float z);

    //! Set quaternion to represent same rotation as given Euler angles
    /*!
     * \param angles   EAR::Euler instance
     */
    void set(const Euler& angles);

    //! direct access to components
    /*!
     * \param idx   index (0 => w (real), 1 => x (i), 2 => y (j), 3 => z (k))
     */
    float& operator[](int idx) { return values[idx]; };

    //! direct const access to components
    /*!
     * \param idx   index (0 => w (real), 1 => x (i), 2 => y (j), 3 => z (k))
     */
    float operator[](int idx) const { return values[idx]; };

    //! Rotate a vector
    Vector3 rotate(const Vector3& v) const;

    //! Get EAR::Euler representation of rotation
    Euler getEuler() const;

    //! Get conjugate Quaternion (w + ix + jy + kz => w - ix - jy - kz)
    Quaternion conj() const;

    //! Multiply with another Quaternion
    /*!
     * \param rhs   other Quaternion to be multiplied
     */
    Quaternion& operator*=(const Quaternion& rhs);

    //! Multiply with a scalar
    /*!
     * \param rhs   float scalar to be multiplied
     */
    Quaternion& operator*=(float rhs);

    //! Calculate dot product of two Quaternions
    float dot(const Quaternion& other) const;

    //! Normalize the quaternion in place
    void normalize();

    friend Quaternion operator*(const Quaternion& lhs, const Quaternion& rhs);
    friend std::ostream& operator<<(std::ostream& os, const Quaternion& qua);
    friend bool operator==(const Quaternion& lhs, const Quaternion& rhs);
    friend bool operator!=(const Quaternion& lhs, const Quaternion& rhs);
    friend Quaternion operator*(Quaternion lhs, float rhs);
    friend Quaternion operator*(float lhs, const Quaternion& rhs);
    friend Quaternion operator+(Quaternion lhs, const Quaternion& rhs);
    friend Quaternion operator-(Quaternion lhs, const Quaternion& rhs);

private:
    //! internal data store
    union
    {
        struct
        {
            float w;
            float x;
            float y;
            float z;
        } q;
        float values[4];
    };
};

//! Perform spherical linear interpolation with two normalized Quaternions
/*!
 * \param lhs   first Quaternion to be interpolated between
 * \param rhs   second Quaternion to be interpolated between
 * \param t     float [0;1]
 */
Quaternion slerp(const Quaternion& lhs, const Quaternion& rhs, float t);

//! Multiply two Quaternions (not commutative!)
/*!
 * \param lhs   first Quaternion to be multiplied
 * \param rhs   second Quaternion to be multiplied
 */
Quaternion operator*(const Quaternion& lhs, const Quaternion& rhs);

//! Scale Quaternion by scalar (duh)
/*!
 * \param lhs   Quaternion to be scaled
 * \param rhs   scalar value
 */
Quaternion operator*(Quaternion lhs, float rhs);

//! Scale Quaternion by scalar (duh)
/*!
 * \param lhs   scalar value
 * \param rhs   Quaternion to be scaled
 */
Quaternion operator*(float lhs, const Quaternion& rhs);

//! perform Quaternion addition and return new Quaternion
/*!
 * \param lhs   first Quaternion to be added
 * \param rhs   second Quaternion to be added
 */
Quaternion operator+(Quaternion lhs, const Quaternion& rhs);

//! perform Quaternion subtraction and return new Quaternion
/*!
 * \param lhs   first Quaternion to be subtracted from
 * \param rhs   second Quaternion to subtract
 */
Quaternion operator-(Quaternion lhs, const Quaternion& rhs);

//! stringify Quaternions when piped into ostream
/*!
 * \param os    output stream
 * \param qua   Quaternions to be printed
 */
std::ostream& operator<<(std::ostream& os, const Quaternion& qua);

//! boolean compare
bool operator==(const Quaternion& lhs, const Quaternion& rhs);

//! boolean compare
bool operator!=(const Quaternion& lhs, const Quaternion& rhs);

//! Stores Polar coordinates (azimuth, elevation, distance).
/*!
 * In EAR, this representation is mainly used for SceneObject-centric positions of other SceneObjects.
 *
 * Both angles can take values in degrees from -180 to 180, so there are multiple ways to express the same location in space.
 */

class Polar
{
public:
    //! Empty Constructor (zero-initializing)
    Polar() { set(0, 0, 0); };

    //! Constructor to initialize from single components
    /*!
     * \param a   azimuth angle (-180 to 180 degrees)
     * \param e   elevation angle (-180 to 180 degrees)
     * \param d   radius (distance in m)
     */
    Polar(float a, float e, float d) { set(a, e, d); };

    //! Constructor to initialize from float array of length 3
    /*!
     * \param _values   float array of length 3
     */
    Polar(float _values[3]) { set(_values); };

    //! Copy constructor
    Polar(const Polar& rhs) = default;

    //! Destructor
    ~Polar(){};

    //! Set all components from outside
    /*!
     * \param a   azimuth angle (-180 to 180 degrees)
     * \param e   elevation angle (-180 to 180 degrees)
     * \param d   roll
     */
    void set(float a, float e, float d);

    //! Set all components from outside
    /*!
     * \param _values   float array of length 3
     */
    void set(float _values[3]);

    //! direct access to components
    /*!
     * \param idx   index (0 => azimuth, 1 => elevation, 2 => distance)
     */
    float& operator[](int idx) { return values[idx]; };

    //! direct const access to components
    /*!
     * \param idx   index (0 => azimuth, 1 => elevation, 2 => distance)
     */
    float operator[](int idx) const { return values[idx]; };

    //! get raw pointer to underlying data
    float* data() { return &values[0]; }

    //! get const raw pointer to underlying data
    const float* data() const { return &values[0]; }

private:
    //! internal data store
    float values[3];
};

//! stringify Polar when piped into ostream
/*!
 * \param os    output stream
 * \param pol   Polar to be printed
 */
std::ostream& operator<<(std::ostream& os, const Polar& pol);

//! boolean compare
bool operator==(const Polar& lhs, const Polar& rhs);

//! boolean compare
bool operator!=(const Polar& lhs, const Polar& rhs);

//! The combination of the LOCATION in space (a cartesian Vector3) and the ORIENTATION (currently Euler angles).
typedef struct Position
{
    //! cartesian
    Vector3 location;
    //! Euler for now, will probably change to quaternions
    Euler orientation;
    //! Reference coordinate system
    CSpace cSpace;

    Position()
        : location()
        , orientation()
        , cSpace(CSpace::WCS)
    {}
    Position(Vector3 loc, Euler ori)
        : location(loc)
        , orientation(ori)
        , cSpace(CSpace::WCS)
    {}
} Position;

//! stringify Position when piped into ostream
/*!
 * \param os    output stream
 * \param pos   Position to be printed
 */
std::ostream& operator<<(std::ostream& os, const Position& pos);

/*! \file BaseTypes.h
    \brief Base types helper functions
*/

//! Helper function to convert from Polar to cartesian Vector3
Vector3 polarToCartesian(const Polar& pol);

//! Helper function to convert from cartesian Vector3 to Polar
Polar cartesianToPolar(const Vector3& cart);

// Helper function to check if the given value is in an angular range
/*!
 * The function can deal with the transition from positive to negative angles (e.g. the range from -5 to 5 contains 0)
 * The range is always the shorter way around the circle (e.g. range from -170 to 170 contains 180)
 */
inline bool isInAngularRange(float test, float low, float high)
{
    if (low > high) {
        std::swap(low, high);
    }

    float center = (low + high) / 2;
    float diff = std::fmod(test - center + 540, 360) - 180;
    if ((high - low) > 180) { // take the shorter way around
        return (diff <= (low - center) || diff >= (high - center));
    } else {
        return (diff >= (low - center) && diff <= (high - center));
    }
}

//! Homogeneous tranformation matrix wrapper
/*!
 * All member functions assume that no scaling factor is present in the transformation.
 * This is relevant for setRotation(), getRotation() and inverse(), because the calculations would become much less straight-forward.
 * If you introduce a scaling factor by writing the matrix components directly, make sure not to rely on these member functions!
 */
class HomogenMatrix
{

public:
    //! Empty Constructor
    HomogenMatrix();

    //! Constructor from position struct
    HomogenMatrix(const Position& pos);

    //! Constructor from location (= translation) and orientation (= intrinsic rotation)
    HomogenMatrix(const Vector3& loc, const Euler& ori);

    //! Constructor from 4 vectors
    HomogenMatrix(const Vector3& x, const Vector3& y, const Vector3& z, const Vector3& o);

    //! Set the intrinsic rotation (= orientation) component of the matrix
    /*!
     * \param rot       intrinsic rotation described in Euler angles
     */
    void setRotation(const Euler& rot);

    //! Set the translation (= location) component of the matrix
    /*!
     * \param rot       translation as a Vector3
     */
    void setTranslation(const Vector3& tra);

    //! Return the rotatory component as a triplet of Euler angles
    Euler getRotation() const;

    //! Return the translatory component as a Vector3
    Vector3 getTranslation() const;

    //! Direct access to matrix components
    /*!
     * \param idx   index (row * 4 + col)
     */
    inline float& operator[](int idx) { return matrix[idx]; };

    //! Direct const access to matrix components
    /*!
     * \param idx   index (row * 4 + col)
     */
    inline float operator[](int idx) const { return matrix[idx]; };

    //! Perform transformation
    /*!
     * \param rhs       input Vector3
     * \return          transformed Vector3
     */
    Vector3 transform(const Vector3& rhs) const;

    //! Multiplication with Vector3 (= perform transformation)
    Vector3 operator*(const Vector3& rhs) const;

    //! Multiplication with float
    HomogenMatrix operator*(const float& rhs) const;

    //! concatenate two transformations into one matrix (using post-multiplication like OpenGL)
    /*!
     * \param rhs       other HomogenMatrix
     * \return          a new HomogenMatrix that combines both transformations
     */
    HomogenMatrix concatenate(const HomogenMatrix& rhs) const;

    //! Multiplication with another homogeneous matrix (= concatenate two transformations)
    HomogenMatrix operator*(const HomogenMatrix& rhs) const;

    //! Check if two matrices are identical
    /*!
     * \param rhs           other HomogenMatrix to compare
     * \param threshold     every element in the matrix is considered to be identical, if the absolute difference is smaller than this threshold (default 0.001)
     * \return              true, if both matrices are identical
     */
    bool isIdentical(const HomogenMatrix& rhs, float threshold = 0.001) const;

    //! Check if two matrices are identical (see isIdentical())
    bool operator==(const HomogenMatrix& rhs) const;

    //! Return the geometrical inverse of this transformation matrix
    /*!
     * Calculated using geometric principles of inverse rotation and translation. 
     * The output of this function is undefined, if the original matrix encodes more 
     * than just a rotation and translation (e.g. scaling)!
     *
     * \return          the inverse transformation matrix
     */
    HomogenMatrix inverseGeometric() const;

    //! Return the inverse of a 3x3 matrix stored as homogeneous matrix
    /*!
     * \return          the inverse matrix
     */
    HomogenMatrix inverse3x3() const;

    //! Return the basis vectors in world-coordinates
    /*!
     * The homogeneous matrix captures information regarding translation and rotation.
     * This information is also included in the basis, though it is not represented in
     * useful form for linear algebra.
     */
    std::vector<EAR::Vector3> getBasis() const;

    //! Return all 3x3 submatrices of this matrix
    /*!
     * \return          a vector including all 3x3 submatrices
     */
    std::vector<Matrix3_3> getSubmatrices() const;

    //! Return the adjugate of this matrix
    HomogenMatrix getAdjugate() const;

    //! Return the transposed matrix
    HomogenMatrix T() const;

    //! Return the determinant
    float det() const;

    //! Return inverse of this matrix
    /*!
     * Calculated via the adjugate – this algorithm fails, if the matrix is not full rank
     * (which should not happen with the 3D transformations this class is built for)
     * 
     * Source of algorithm: https://de.wikipedia.org/wiki/Adjunkte#Berechnung_der_Inversen_einer_Matrix
     */
    HomogenMatrix inverse() const;

private:
    float matrix[4 * 4];
};

//! Matrix class for playing with 3x3 matrices and their operations
/*!
 * It includes basic matrix operations. Separate matrix values can be accessed
 * via matrix[index_row][index_column]. Separate matrix values can be set via
 * matrix[index_row][index_column] = 5.
 */

class Matrix3_3
{
public:
    //! MutableReference class
    /*!
     *  Prupose: Establishs direct matrix access in accordance to
     *  matrix[index_row][index_column]
     */
    class MutableReference
    {
    public:
        //! Constructor
        MutableReference(Matrix3_3& parent_, int row_)
            : parent(parent_)
            , row(row_)
        {}

        //! direct element access regarding one specific row in matrix
        /*!
         *  param index equals the element in row which is accessed
         */
        float& operator[](const int column) { return parent.m_matrix[row][column]; }

    private:
        Matrix3_3& parent;
        int row;
    };

    //! ImmutableReference class
    /*!
     *  Purpose: Establishes const direct matrix access in accordance to
     *  matrix[index_row][index_column]
     */
    class ImmutableReference
    {
    public:
        //! Constructor
        ImmutableReference(const Matrix3_3& parent_, int row_)
            : parent(parent_)
            , row(row_)
        {}
        //! direct element access regarding one specific row in matrix
        /*!
        *  param index equals the element in row which is accessed
        */
        const float& operator[](const int column) const { return parent.m_matrix[row][column]; }

    private:
        const Matrix3_3& parent;
        int row;
    };

    //! Empty Constructor
    Matrix3_3()
        : m_matrix{ { { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 } } } {};

    //! Constructor using initializer lists
    Matrix3_3(const std::initializer_list<std::initializer_list<float>> _val)
    {
        EAR_CHECK_REL(
            _val.size() == m_matrix.size(), "Invalid size, matrix size does not match input size");

        std::array<std::array<float, 3>, 3>::iterator i = m_matrix.begin();

        for (const std::initializer_list<float>& lst : _val) {
            std::array<float, 3>::iterator j = i->begin();

            for (const float& v : lst) {
                *j = v;
                ++j;
            }

            ++i;
        }
    };

    //! Constructor using std::arrays
    Matrix3_3(const std::array<std::array<float, 3>, 3> _val)
    {
        EAR_CHECK_REL(
            _val.size() == m_matrix.size(), "Invalid size, matrix size does not match input size");

        std::array<std::array<float, 3>, 3>::iterator i = m_matrix.begin();

        for (const std::array<float, 3>& lst : _val) {
            std::copy(lst.begin(), lst.end(), i->begin());
            ++i;
        }
    };

    //! Constructor using C-style array
    Matrix3_3(const float _val[3][3])
    {
        std::copy(&_val[0][0], &_val[0][0] + 9, &m_matrix[0][0]); //possible is memory continous???
    };

    //! Deconstructor
    ~Matrix3_3(){};

    //! direct access to matrix element
    /*!
     * \param index equals the row which is accessed
     *
     * This function is used to establish direct matrix element access
     * in accordance to matrix[index_row][index_column]
     */
    MutableReference operator[](const int indexRow) { return MutableReference(*this, indexRow); }

    //! const direct access to matrix element
    /*!
     * \param index equals the row which is accessed
     *
     * This function is used to establish direct matrix element access
     * in accordance to matrix[index_row][index_column]
     */
    ImmutableReference operator[](const int indexRow) const
    {
        return ImmutableReference(*this, indexRow);
    }

    //! perform matrix addition
    /*!
     * \param rhs added to matrix
    */
    Matrix3_3& operator+=(const Matrix3_3& rhs);

    //! perform matrix subtraction
    /*!
     * \param rhs subtracted from matrix
    */
    Matrix3_3& operator-=(const Matrix3_3& rhs);

    //! perform matrix multiplication
    /*!
     * \param rhs multiplicated with matrix
    */
    Matrix3_3& operator*=(const Matrix3_3& rhs);

    //! perform matrix assignment
    /*!
     * \param rhs assigned to lhs
    */
    EAR::Matrix3_3& operator=(const EAR::Matrix3_3& rhs);

    //! returns transposed of matrix
    /*!
     *
    */
    Matrix3_3 getTranspose() const;

    //! returns determinant of matrix
    /*!
     *
    */
    float getDeterminant() const;

    //! returns inverse of matrix
    /*!
     *
    */
    Matrix3_3 getInverse() const;

    //! returns trace of matrix
    /*!
     *
    */
    float getTrace() const;

    //! perform matrix-scalar addition
    /*!
     * \param scalar value rhs added to separate matrix values
    */
    Matrix3_3& operator+=(const float rhs);

    //! perform matrix-scalar subtraction
    /*!
     * \param scalar value rhs subtracted from separate matrix values
    */
    Matrix3_3& operator-=(const float rhs);

    //! perform matrix-scalar multiplication
    /*!
     * \param scalar value rhs multiplied with separate matrix values
    */
    Matrix3_3& operator*=(const float rhs);

    //! perform matrix-scalar devision
    /*!
     * \param separate matrix values divided by scalar value rhs
    */
    Matrix3_3& operator/=(const float rhs);

    //! perform matrix-scalar multiplication
    /*!
     * \param separate matrix values multiplied by scalar value rhs
    */
    Matrix3_3& operator*(const float rhs);

    //! perform matrix-matrix subtraction
    /*!
     * \param separate matrix values subtracted by rhs
    */
    Matrix3_3& operator-(const Matrix3_3& rhs);

    //! perform matrix-matrix addition
    /*!
     * \param separate matrix values added by rhs
    */
    Matrix3_3& operator+(const Matrix3_3& rhs);

    //! returns dedicated row
    /*!
     * \param idx equals the index of the row to be returned
    */
    std::array<float, 3> getRow(const int idx) const;

    //! returns dedicated column
    /*!
     * \param idx equals the index of the column to be returned
    */
    std::array<float, 3> getColumn(const int idx) const;

private:
    std::array<std::array<float, 3>, 3> m_matrix;
    const unsigned int m_rowSize = 3;
    const unsigned int m_colSize = 3;
};

Vector3 operator*(const Matrix3_3 lhs, const Vector3 rhs);

//! stringify HomogenMatrix when piped into ostream
/*!
 * \param os    output stream
 * \param mat   HomogenMatrix to be printed
 */
std::ostream& operator<<(std::ostream& os, const HomogenMatrix& mat);

//! stringify Matrix3_3 when piped into ostream
/*!
 * \param os    output stream
 * \param mat   Matrix3_3 to be printed
 */
std::ostream& operator<<(std::ostream& os, const Matrix3_3& mat);

//! Helper function to print arbitrary std::vectors to ostream
template <typename T> std::ostream& operator<<(std::ostream& os, const std::vector<T>& vec)
{
    os << "[";

    for (unsigned n = 0; n < vec.size(); n++) {
        os << vec[n];

        if (n + 1 < vec.size()) {
            os << ", ";
        }
    }

    return os << "]";
}

//! Helper function to calculate the rotation matrix between two vectors.
/*! 
* \param v1     Vector to rotate from
* \param v2     Vector to rotate onto
*
* Note that this problem does not have a unique solution, so this function will give a 
* HomogenMatrix that satisfies the constraints, but it is not guaranteed to be 
* "optimal" in any way and may induce roll around the vectors.
*/
HomogenMatrix rotateVectorTo(const Vector3 v1, const Vector3 v2);

} /* namespace EAR*/

#endif /* EAR_BASETYPES */
