/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_BUFFER_BASE
#define EAR_BUFFER_BASE

#include <cstddef>
#include <memory>
#include <vector>

#include <EAR/AlignedAllocatorFwd.h>

namespace EAR {

//TODO: operator[] is very slow, because it's not inlined and thereby prevents compiler optimizations...

//! Base class for specializations with float (RealBuffer) and std::complex<float> (ComplexBuffer) values.
/*!
 * It provides convenient access to memory with common operations
 * and makes sure it is aligned to the needs of SIMD operations.
 */
template <typename T> class Buffer
{
public:
    //! Create an empty buffer
    Buffer();

    //! Create a buffer initialized with zeros
    /*!
     * \param length    buffer length [samples]
     */
    Buffer(std::size_t length);

    //! Copy constructor
    Buffer(const Buffer<T>& other);

    //! Copy constructor from a raw array pointer
    /*!
     * \param data      pointer to sample data to copy
     * \param length    buffer size [samples]
     */
    Buffer(const T* data, std::size_t length);

    //! Destructor
    virtual ~Buffer();

    //! Assignment
    Buffer<T>& operator=(const Buffer<T>& other);

    // Move
    // Buffer<T>& operator=(Buffer<T>&& other);

    //! In-place multiplication with a constant
    /*!
     * \param c    constant factor
     */
    void mul(T c);

    //! In-place multiplication of a specified range with constant
    /*!
     * \param c         constant factor
     * \param offset    first sample to multiply
     * \param count     number of samples to multiply
     */
    void mul(T c, std::size_t offset, std::size_t count);

    //! In-place sample-wise multiplication with other buffer
    /*!
     * \param src    other buffer of the same length
     */
    void mul(const Buffer<T>& src);

    //! In-place sample-wise multiplication with other buffer
    /*!
     * \param src         other buffer
     * \param srcOffset   first sample to multiply from the other buffer
     * \param dstOffset   first sample to multiply from the internal buffer
     * \param count       number of samples to multiply
     */
    void mul(const Buffer<T>& src, std::size_t srcOffset, std::size_t dstOffset, std::size_t count);

    //! In-place sample-wise multiplication with a linear gain ramp
    /*!
     * \param startC    factor at start
     * \param endC      factor at end
     */
    void rampMul(T startC, T endC);

    //! In-place sample-wise multiplication with a linear gain ramp
    /*!
     * \param startC    factor at start
     * \param endC      factor at end
     * \param offset    first sample to multiply
     * \param count     number of samples to multiply
     */
    void rampMul(T startC, T endC, std::size_t offset, std::size_t count);

    //! In-place addition with a constant
    /*!
     * \param c    constant value
     */
    void add(T c);

    //! In-place addition of a specified range with constant
    /*!
     * \param c         constant value
     * \param offset    first sample to add c to
     * \param count     number of samples to add c to
     */
    void add(T c, std::size_t offset, std::size_t count);

    //! In-place sample-wise addition with other buffer
    /*!
     * \param src    other buffer of the same length
     */
    void add(const Buffer<T>& src);

    //! In-place sample-wise addition with other buffer
    /*!
     * \param src         other buffer
     * \param srcOffset   first sample to add from the other buffer
     * \param dstOffset   first sample to add from the internal buffer
     * \param count       number of samples to add
     */
    void add(const Buffer<T>& src, std::size_t srcOffset, std::size_t dstOffset, std::size_t count);

    // TODO: operator overloads for convenience?

    //! Get raw const pointer to underlying data
    const T* data() const;

    //! Get raw pointer to underlying data
    T* data();

    //! Direct const access to sample values
    /*!
     * \param index   index
     */
    const T& operator[](std::size_t index) const;

    //! Direct access to sample values
    /*!
     * \param index   index
     */
    T& operator[](std::size_t index);

    //! Direct const access to sample values
    /*!
     * \param index   index
     */
    const T& at(std::size_t index) const;

    //! Direct access to sample values
    /*!
     * \param index   index
     */
    T& at(std::size_t index);

    //! Read a range of sample values from the buffer
    /*!
     * \param dst     pointer to write to (make sure to allocate sufficient memory beforehand)
     * \param offset  first sample to read
     * \param count   number of samples to read
     */
    void read(T* dst, std::size_t offset, std::size_t count) const;

    //! Insert given samples at specified position by overwriting current data
    /*!
     * \param src     pointer to new sample data
     * \param offset  first sample to write to
     * \param count   number of samples to write
     */
    void write(const T* src, std::size_t offset, std::size_t count);

    //! Copy all values from other buffer (receiving buffer is resized accordingly - resize is a slow operation!)
    /*!
     * \param src     other buffer
     */
    void copyFrom(const Buffer<T>& src);

    //! Copy a range of values from other buffer
    /*!
     * \param src         other buffer
     * \param srcOffset   first sample to copy from the other buffer
     * \param dstOffset   first sample to copy into the internal buffer
     * \param count       number of samples to copy
     */
    void copyFrom(
        const Buffer<T>& src, std::size_t srcOffset, std::size_t dstOffset, std::size_t count);

    /* TODO: cyclic read/write
    void readCyclic(float* dst, std::size_t offset, std::size_t count) const;
    void writeCyclic(const float* src, std::size_t offset, std::size_t count);
    */

    //! Cyclic shift of samples to the left (first samples added to the end again).
    /*!
     * \param count    number of samples to shift
     */
    void shiftLeft(std::size_t count);

    //! Cyclic shift of samples to the right (last samples added to the front again).
    /*!
     * \param count    number of samples to shift
     */
    void shiftRight(std::size_t count);

    //! Cyclic shift of samples to the left and write new samples (faster than first calling shift and then write).
    /*!
     * \param src      pointer to new sample data
     * \param count    number of samples to shift
     */
    void shiftLeftInsert(const T* src, std::size_t count);

    //! Cyclic shift of samples to the right and write new samples (faster than first calling shift and then write).
    /*!
     * \param src      pointer to new sample data
     * \param count    number of samples to shift
     */
    void shiftRightInsert(const T* src, std::size_t count);

    //! Resize allocated memory to new length. This is a SLOW operation!
    /*!
     * \param newLength     new length of buffer in samples
     */
    void resize(std::size_t newLength);

    //! Fill complete buffer with a constant value
    /*!
     * \param value     constant value to fill buffer with
     */
    void fill(T value);

    //! Fill a specified range of the buffer with a constant value
    /*!
     * \param value     constant value to fill buffer with
     * \param offset    first sample to overwrite
     * \param count     number of samples to overwrite
     */
    void fill(T value, std::size_t offset, std::size_t count);

    //! Fill complete buffer with zeros
    void zero();

    //! Fill a specified range of the buffer with zeros
    /*!
     * \param offset    first sample to overwrite
     * \param count     number of samples to overwrite
     */
    void zero(std::size_t offset, std::size_t count);

    //! Return current length of buffer
    std::size_t length() const;

protected:
    //! internal data storage
    /*!
     * The vector has a custom allocator taking care of memory alignment (AlignedAllocator is private in the lib)
     */
    std::unique_ptr<std::vector<T, AlignedAllocator<T>>> buffer;
};

} // namespace EAR

#endif /* EAR_BUFFER_BASE */
