/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_CARTESIANKDTREE
#define EAR_CARTESIANKDTREE

#include <EAR/BaseTypes.h>
#include <EAR/GeometryBase.h>

#include <memory>
#include <vector>

// forward declaration for private includes
namespace nanoflann {
template <class T, class DataSource, typename _DistanceType> struct L2_Simple_Adaptor;
template <typename Distance, class DatasetAdaptor, int DIM, typename IndexType>
class KDTreeSingleIndexAdaptor;
struct EarPointCloud;
} // namespace nanoflann

namespace EAR {

//! Cartestian based KD Tree - Space-Partitioned data structure for point cloud
/*!
 *  Upon construction of CartesianKDTree, a list of cartesian points (x, y, z)
 *  is given as an argument. The CartesianKDTree can then be queried to see which
 *  point in the tree is closest to the given point via a nearest neighbour search.
 */
class CartesianKDTree
{
public:
    //! Constructor
    /*!
     * \param points    A list of cartesian points (x, y, z)
     */
    CartesianKDTree(const std::vector<Vector3>& points);

    ~CartesianKDTree();

    //! Return all points that were given on upon construction
    const std::vector<Vector3>& getPoints() const;

    //! Perform nearest neighbour search for the given point
    /*!
     * \param point     The cartesian point (x, y, z) to be queried
     *
     * \return          The point closest to the given point
     */
    const Vector3& getNearestPoint(const Vector3& point) const;

    //! Perform nearest neighbour search for the given point
    /*!
     * \param point     The cartesian point (x, y, z) to be queried
     *
     * \return          The index to the closest point in the list of points given on construction
     */
    std::size_t getNearestIndex(const Vector3& point) const;

    //! Perform nearest neightbour search and return the distance to the closest point
    /*!
     * \param point     The cartesian point (x, y, z) to be quried
     *
     * \return          The distance (m) to the closest point
     */
    float getDistanceToNearestPoint(const Vector3& point) const;

    //! Perform nearest neighbour search for the given point within a given radius
    /*!
     * \param point     The cartesian point (x, y, z) to be queried
     * \param radius    The radius from the point to check nearby neighbours
     *
     * \return          All points that are within the radius for the given point
     */
    std::vector<Vector3> getAllPointsWithinRadius(const Vector3& point, float radius) const;

    //! Perform nearest neighbour search for the given point within a given radius
    /*!
     * \param point     The cartesian point (x, y, z) to be queried
     * \param radius    The radius from the point to check nearby neighbours
     *
     * \return          All indices to the points that are within the radius for the given points on construction
     */
    std::vector<std::size_t> getAllIndicesWithinRadius(const Vector3& point, float radius) const;

    //! Perform nearest neighbour search for the given point within a given radius
    /*!
     * \param point     The cartesian point (x, y, z) to be queried
     * \param radius    The radius from the point to check nearby neighbours
     *
     * \return          All indices and distances to the points that are within the radius for the given points on construction
     */
    std::vector<std::pair<std::size_t, float>> getAllIndicesAndDistancesWithinRadius(
        const Vector3& point, float radius) const;

    //! Return a bounding box which encloses all the points in the dataset - the min and max coordinate dimensions
    Geometry::BoundingBox getBoundingBox();

private:
    typedef nanoflann::KDTreeSingleIndexAdaptor<
        // euclidean distance setup
        nanoflann::L2_Simple_Adaptor<float, nanoflann::EarPointCloud, float>,
        nanoflann::EarPointCloud, // data structure
        3, //dimensions
        std::size_t>
        CartKDTree;

    std::unique_ptr<CartKDTree> lookupTree;
    std::unique_ptr<nanoflann::EarPointCloud> points;
};

} // namespace EAR

#endif // EAR_CARTESIANKDTREE
