/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_COMPLEXBUFFER
#define EAR_COMPLEXBUFFER

#include <EAR/Buffer.h>
#include <complex>
#include <cstddef>

namespace EAR {

//! Specialization of Buffer with std::complex<float> values.
/*!
 * Data management etc. is inherited from its base class Buffer.
 * Complex-specific operations are currently only writing to real and imag part separately and my be extended in the future.
 */
class ComplexBuffer : public Buffer<std::complex<float>>
{
public:
    //! Create an empty buffer
    ComplexBuffer();

    //! Create a buffer initialized with zeros
    /*!
     * \param length    buffer  [samples]
     */
    ComplexBuffer(std::size_t length);

    //! Copy constructor
    ComplexBuffer(const Buffer<std::complex<float>>& other);

    //! Copy constructor from a raw array pointer
    /*!
     * \param data      pointer to sample data to copy
     * \param length    buffer size [samples]
     */
    ComplexBuffer(const std::complex<float>* data, std::size_t length);

    //! Destructor
    virtual ~ComplexBuffer();

    //! Write real part of samples
    /*!
     * \param src       pointer to new real sample data
     * \param offset    first sample to write to
     * \param count     number of samples to write
     */
    void writeReal(const float* src, std::size_t offset, std::size_t count);

    //! Write imaginary part of samples
    /*!
     * \param src       pointer to new real sample data
     * \param offset    first sample to write to
     * \param count     number of samples to write
     */
    void writeImag(const float* src, std::size_t offset, std::size_t count);

    //! Read real part of samples
    /*!
     * \param dst       pointer to write to
     * \param offset    first sample to read
     * \param count     number of samples to read
     */
    void readReal(float* dst, std::size_t offset, std::size_t count);

    //! Read imaginary part of samples
    /*!
     * \param dst       pointer to write to
     * \param offset    first sample to read
     * \param count     number of samples to read
     */
    void readImag(float* dst, std::size_t offset, std::size_t count);

    //! Mirror spectrum along the DC band, so that negative and positive spectrum components are symmetric
    void symmetrizeSpectrum();
};

} // namespace EAR

#endif /* EAR_COMPLEXBUFFER */
