/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_GEOMETRY_CYLINDER
#define EAR_GEOMETRY_CYLINDER

#include <EAR/GeometryBase.h>

namespace EAR {

// fwd declare
class Mesh;

//! Cylinder primitive geometry, defined by a elliptical ground plane in x-z and a height in y direction
class Cylinder : public Geometry
{
private:
    //! Constructor for a cylinder primitive with circular base (only accessible through SceneObjectFactory)
    /*!
     * The cylinder height is parallel to the y axis in local coordinates.
     *
     * \param oid        fixed object ID (always assigned by factory method!)
     * \param radius     the distance from the centre of the base to the edge
     * \param height     the perpendicular distance between the bases
     * \param name       arbitrary name of object
     * \param pos        Position struct with initial location and orientation
     */
    Cylinder(ObjectID oid, float radius, float height, const std::string& name = "",
        Position pos = Position());

    //! Constructor for a cylinder primitive with elliptical base (only accessible through SceneObjectFactory)
    /*!
     * The cylinder height is parallel to the y axis in local coordinates.
     *
     * \param oid        fixed object ID (always assigned by factory method!)
     * \param a          radius of the base ellipsis in local x direction
     * \param b          radius of the base ellipsis in local z direction
     * \param height     the perpendicular distance between the bases
     * \param name       arbitrary name of object
     * \param pos        Position struct with initial location and orientation
     */
    Cylinder(ObjectID oid, float a, float b, float height, const std::string& name = "",
        Position pos = Position());

    //! Constructor for a cylinder primitive with default dimensions (only accessible through SceneObjectFactory)
    /*!
     * Default dimensions are radius=1, height=1.
     * The cylinder height is parallel to the y axis in local coordinates.
     *
     * \param oid        fixed object ID (always assigned by factory method!)
     * \param name       arbitrary name of object
     */
    Cylinder(ObjectID oid, const std::string& name = "");

    friend SceneObjectFactory<Cylinder>;

public:
    //! Sets the material for this object
    /*!
    * \param material  the new material of this primitive
    */
    void setMaterial(std::shared_ptr<Material> material);

    //! Returns the material for this primitive
    std::shared_ptr<Material> getMaterial() const;

    //! Inherited from Geometry
    bool isInside(const EAR::Vector3& point) const override;

    //! Inherited from Geometry
    Vector3 closestPointTo(const EAR::Vector3& point) const override;

    //! Inherited from Geometry
    Type getType() const override { return Type::Cylinder; };

    //! Inherited from Geometry
    float getVolume() const override;

    //! Inherited from Geometry
    float getArea() const override;

    //! Inherited from SceneObject
    void setOrientation(Euler newOrient, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setLocation(Vector3 newLoc, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setPosition(
        Position newPos, SceneObject* parent = nullptr, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setStatic(bool newState) override;

    //! Return the two center points of the base circles on top and bottom of the cylinder
    std::pair<Vector3, Vector3> getCenterOfBases() const;

    //! Return the minor radius of the Cylinder base ellipse (same as getMajorRadius() for circular base)
    float getMinorRadius() const;

    //! Return the major radius of the Cylinder base ellipse (same as getMinorRadius() for circular base)
    float getMajorRadius() const;

    //! Return the major radius of the Cylinder base ellipse (same as getMajorRadius(), but available as legacy interface)
    float getRadius() const;

    //! Return true if the base of the Cylinder is circular
    bool hasCircularBase() const;

    //! Return the height of the Cylinder
    float getHeight() const;

    //! Inherited from Geometry
    const Mesh* getMesh() const override;

    //! Inherited from Geometry
    float greatestDistance() const override;

    //! Inherited from Geometry
    void getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
        EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    void getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    BoundingBox getBoundingBox() const override;

    //! Inherited from Geometry
    BoundingBox getGlobalBoundingBox() const override;

    EAR::Vector3 getCenterPosition() const override;

private:
    //! Calculates the nearest point on the ellipse to an abritrary point outside the ellipse (based on Newton method)
    /*
     * \param point        Point of interest
     */
    // Vector3 closestPoint2Ellipse_Newton(const Vector3& point) const;

    //! Calculates the nearest point on the ellipse to an abritrary point outside the ellipse (based on curvature)
    /*
     * \param point        Point of interest
     * \param nInter       Number of iteration cycles (4 is a good value)
     */
    Vector3 closestPoint2Ellipse_Curvature(
        const EAR::Vector3& point, const unsigned int& nIter) const;

    inline void createMesh();

    float height;
    float majorR;
    float minorR;
    bool isCircular;
    //! The material that this cylinder is made of
    std::shared_ptr<Material> material;

    //! Mesh representation of the cylinder
    std::unique_ptr<Mesh> mesh;

    friend std::ostream& operator<<(std::ostream& os, Cylinder& c);
};

std::ostream& operator<<(std::ostream& os, Cylinder& c);

} /* namespace EAR*/

#endif /* EAR_GEOMETRY_CYLINDER */
