/***************************************************************************
This software module was originally developed jointly by

Ericsson AB and Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB and Fraunhofer IIS retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_DISTANCE_MODEL
#define EAR_DISTANCE_MODEL

namespace EAR {

//! Volumetric distance gain model
/*!
 * The volumetric distance gain model calculates the distance gain for a source with extent.
 *
 * The DistanceGainVolumetric class contains two functions:
 *      - VolumetricGainModel: the "core" volumetric distance gain model
 *      - getGainForDistance: shell function that calculates the overall distance gain value, taking into account reference distance, close-distance gain limitation and large-distance gain ramp-down
 *
 * The core function has three input parameters:
 *      - the distance: any suitable representation of the distance from the evaluation point to the source.
 *      - two size parameters representing the size in two orthogonal dimensions of a suitable representation of the extent.
 * If both size parameters are set to zero, the model produces the 1/r point-source distance gain.
 */
class DistanceGainVolumetric
{
public:
    //! Constructor.
    /*!
     * \param _dampingStart  the radius of the area where no damping is applied
     */
    DistanceGainVolumetric(float _dampingStart);

    //! returns the overall distance gain value within the pre-defined bounds
    /*!
     * \param dist        distance between evaluation point and source [m]
     * \param dim1        size of source extent in first dimension [m]
     * \param dim2        size of source extent in second dimension [m]
     * \param refDist     reference distance, where gain shall be 1 [m]
     * \param gainComp    with near-field gain compensation or not
     * \param portalComp  distance compensation for portal items
     * \param distExp   exponent to shape the distance attenuation 
     */
    float getGainForDistance(float dist, float dim1, float dim2, float refDist,
        float portalComp = 0.f, bool gainComp = false, float distExp = 1.f);

    //! returns the "core" volumetric distance gain
    /*!
     * \param dist      distance between evaluation point and source [m]
     * \param dim1      size of source extent in first dimension [m]
     * \param dim2      size of source extent in second dimension [m]
     */
    float VolumetricGainModel(float dist, float dim1, float dim2);

private:
    float dampingStart, gainCompDist, maxCompGain;
};

} // namespace EAR

#endif /* EAR_DISTANCE_MODEL */
