/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_EMBREE_TRACER
#define EAR_EMBREE_TRACER

#include <EAR/BaseTypes.h>
#include <EAR/SceneObjectObserver.h>

#include <memory>
#include <mutex>
#include <unordered_map>
#include <vector>

typedef struct RTCDeviceTy* RTCDevice;
typedef struct RTCSceneTy* RTCScene;
struct RTCRayHit;

namespace EAR {

class Mesh;

class EmbreeTracer : public SceneObjectObserver
{
public:
    //! Constructor from vector
    /*!
     * \param geometry          vector of pointers to the meshes
     *                          (tracer determines from Mesh::isStatic() whether it should be handled as moving object or not)
     * \param staticMatters     if set to true, embree will not observe possible movement of geometries
     *                          that are declared as static (default is true)
     * \param ignoreBackface    if set to true, intersections with rays only count if the front face
     *                          (i.e. the direction in which the normal points) is hit (default is false)
     * \param bundleWidth       Maximum distance of each ray in a bundle to the given origin in meters
     *                          (default: 0.15m, resulting in a 30x30 cm frustrum base)
     */
    EmbreeTracer(const std::vector<const Mesh*> geometry, bool staticMatters = true,
        bool ignoreBackface = false, float bundleWidth = 0.3);

    //! Constructor from single mesh* instance
    /*!
     * \param geometry          pointer to the mesh
     *                          (tracer determines from Mesh::isStatic() whether it should be handled as moving object or not)
     * \param staticMatters     if set to true, embree will not observe possible movement of geometries
     *                          that are declared as static (default is true)
     * \param ignoreBackface    if set to true, intersections with rays only count if the front face
     *                          (i.e. the direction in which the normal points) is hit (default is false)
     * \param bundleWidth       Maximum distance of each ray in a bundle to the given origin in meters
     *                          (default: 0.15m, resulting in a 30x30 cm frustrum base)
     */
    EmbreeTracer(const Mesh* geometry, bool staticMatters = true, bool ignoreBackface = false,
        float bundleWidth = 0.15);

    //! Destructor
    ~EmbreeTracer();

    //! Check if a direct line of sight between two points exists
    /*!
     * \param origin            starting point
     * \param target            end point
     * \param threshold         allowed deviation from the actual target in meters
     */
    bool isOccluded(const Vector3& origin, const Vector3& target, float threshold = 0);

    //! Direct access to the embree device (for your own experiments with the Embree API)
    /*!
     * The device is singleton across the whole application.
     * Caution: you are now responsible to make your calls and changes on the scene sequential when using multiple threads.
     */
    static RTCDevice getDevice();

    //! Direct access to the static part of the embree scene (for your own experiments with the Embree API)
    /*!
     * Caution: you are now responsible to make your calls and changes on the scene sequential when using multiple threads.
     */
    RTCScene getScene();

    //! Return a pointer to the first geometry that is hit when shooting a ray in a specified direction
    /*!
     * \param origin        starting point
     * \param direction     Polar angle giving the direction of the ray
     * \param maxDist       boolean whether or not to consider the distance component of the Polar coordinates as the maximum distance of the ray
     *                      (default false)
     * \return              pointer to a Mesh instance (or nullptr if nothing is hit)
     */
    const EAR::Mesh* firstHit(const Vector3& origin, const Polar& direction, bool maxDist = false);

    //! Like firstHit(), but shoots a bundle of rays and calculates the fraction of rays that hits the same geometry.
    /*!
     * The bundle is created as a frustrum with all rays meeting in the maximum distance point of the direction.
     * 8 rays are constructed in a retangular pattern with a maximum distance to the origin given by the constructor parameter bundleWidth.
     *
     * \param origin        starting point
     * \param direction     Polar angle giving the direction and maximum distance of the rays
     * \param hitMesh       pointer to the Mesh that was hit first (corresponds to return value of firstHit())
     * \return              float 0-1, representing the amount of covered rays
     */
    float firstHitBundle(const Vector3& origin, const Polar& direction, const EAR::Mesh*& hitMesh);

    //! Wheter or not a point is occluded by a specific mesh from a given point of view
    /*!
	 * \param origin        starting point (point of view)
	 * \param target		location of the target point
	 * \param givenMesh		mesh of interest
	 * \return              true only if the ray is occluded by the givenMesh (and not if the ray is occluded by other meshes only)
	 */
    bool occludedByGivenMesh(const Vector3& origin, const Vector3& target, const Mesh* givenMesh);

    //! Determine whether a point is inside or outside of the geometry (which is assumed to be a closed mesh for this)
    /*!
     * \param point         Point to test
     */
    bool isInside(const Vector3& point);

    //! Inherited from SceneObjectObserver
    void objectChanged(
        EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

    //! Return the distance from the hit point to the starting point of the ray that was shot last (e.g. by firstHit())
    float getDistanceToLastHit();

    //! Return the primitive id from the hit produced by the ray that was shot last (e.g. by firstHit())
    /*!
     * Returns -1 if no last hit is available.
     */
    int getPrimitiveOfLastHit();

private:
    static RTCDevice device;

    RTCScene scene;
    std::unordered_map<unsigned int, const Mesh*> geo2mesh;
    std::unordered_map<const Mesh*, unsigned int> mesh2geo;

    std::mutex sceneMutex;

    std::vector<Vector3> bundleRaster; // used for bundle coverage test

    std::unique_ptr<RTCRayHit> rayhit;
    float lastDist;
    int lastPrimID;

    //! Initialize the rayhit member variable for intersection tests
    inline void initRayHitForIntersection(const Vector3& origin, const Vector3& direction,
        float maxDist = std::numeric_limits<float>::infinity());

    //! Attach geometries to scenes
    inline void attachGeometry(const Mesh*& geo, bool staticMatters, bool ignoreBackface);

    bool isInsideImpl(const Vector3& point, const Vector3& direction);
};

} // namespace EAR

#endif /* EAR_EMBREE_TRACER */
