/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_FFT
#define EAR_FFT

#include <complex>
#include <memory>
#include <vector>

namespace EAR {

//! Interface and factory for real-data forward FFT [real-to-complex]
class RFFT
{
public:
    //! Factory method, returns a real-to-complex FFT processor based on the CMake config "EAR_FFT_LIB"
    static std::unique_ptr<RFFT> create(int length);

    //! return the bin center frequencies of a real fft of specified length and samplerate
    /*!
     * \param length    number of frequency bins
     * \param fs        sampling rate [Hz]
     */
    static std::vector<float> binFrequencies(int length, float fs);

    //! Destructor
    virtual ~RFFT(){};

    //! DFT transform size
    const int N;

    //! Number of DFT spectral coefficients
    // As the transform is complex-conjugate symmetric: K = ceil(N/2 + 1)
    const int K;

    //! Perform transformation (input and output may point to the same memory location)
    /*!
     * \param input     pointer to real input data (not modified)
     * \param output    pointer to complex output data
     */
    virtual void transform(const float* input, std::complex<float>* output) = 0;

protected:
    //! Protected constructor to prohibit direct instanciation
    RFFT(int length);
};

//! Interface and factory for real-data inverse (backward) FFT [complex-to-real]
/*!
 * See RFFT for available backends and their constraints.
 *
 * CAUTION: The result is NOT normalized with 1/N as would be required for getting an exact inverse.
 */
class IRFFT
{
public:
    //! Factory method, returns a real-to-complex IFFT processor based on the CMake config
    static std::unique_ptr<IRFFT> create(int length);

    //! Destructor
    virtual ~IRFFT(){};

    //! DFT transform size
    const int N;

    //! Number of DFT spectral coefficients
    /*!
     *As the transform is complex-conjugate symmetric: K = ceil(N/2 + 1)
    */
    const int K;

    //! Perform transformation (input and output may point to the same memory location)
    /*!
     * \param input     pointer to complex input data (not modified)
     * \param output    pointer to real output data
     */
    virtual void transform(const std::complex<float>* input, float* output) = 0;

protected:
    //! Protected constructor to prohibit direct instanciation
    IRFFT(int length);
};

//! Interface and factory for complex-valued forward FFT [complex-to-complex]
class FFT
{
public:
    //! Factory method, returns a complex-to-complex FFT processor based on the CMake config "EAR_FFT_LIB"
    static std::unique_ptr<FFT> create(int length);

    //! return the bin center frequencies of a real fft of specified length and samplerate
    /*!
     * \param length    number of frequency bins
     * \param fs        sampling rate [Hz]
     */
    static std::vector<float> binFrequencies(int length, float fs);

    //! Destructor
    virtual ~FFT(){};

    //! DFT transform size
    const int N;

    //! Perform transformation (input and output may point to the same memory location)
    /*!
     * \param input     pointer to complex input data (not modified)
     * \param output    pointer to complex output data (may be the same as input for in-place operation)
     */
    virtual void transform(const std::complex<float>* input, std::complex<float>* output) = 0;

protected:
    //! Protected constructor to prohibit direct instanciation
    FFT(int length);
};

//! Interface and factory for complex-valued inverse (backward) FFT [complex-to-complex]
/*!
 * See FFT for available backends and their constraints.
 *
 * CAUTION: The result is NOT normalized with 1/N as would be required for getting an exact inverse.
 */
class IFFT
{
public:
    //! Factory method, returns a complex-to-complex IFFT processor based on the CMake config
    /*!
     * Current option is only IPP (see FFT::create)
     */
    static std::unique_ptr<IFFT> create(int length);

    //! Destructor
    virtual ~IFFT(){};

    //! DFT transform size
    const int N;

    //! Perform transformation (input and output may point to the same memory location)
    /*!
     * \param input     pointer to complex input data (not modified)
     * \param output    pointer to real output data
     */
    virtual void transform(const std::complex<float>* input, std::complex<float>* output) = 0;

protected:
    //! Protected constructor to prohibit direct instanciation
    IFFT(int length);
};

} // namespace EAR

#endif /* EAR_FFT */
