/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/
#ifndef EAR_FIR2_MINPHASE_FILTER_DESIGNER
#define EAR_FIR2_MINPHASE_FILTER_DESIGNER

#include <EAR/ComplexBuffer.h>
#include <EAR/Exception.h>
#include <EAR/FFT.h>
#include <EAR/FilterDesigner.h>
#include <EAR/RealBuffer.h>

#include <algorithm>
#include <cstddef>
#include <memory>
#include <vector>

namespace EAR {

class Hilbert
{
public:
    Hilbert(std::size_t N);

    void transform(const RealBuffer& b, ComplexBuffer& r);
    void transform(const ComplexBuffer& b, ComplexBuffer& r);

private:
    int N; // signal length
    ComplexBuffer w; // hilbert window
    ComplexBuffer in; // fft buffers
    ComplexBuffer out; // fft buffers

    std::unique_ptr<EAR::FFT> fft;
    std::unique_ptr<EAR::IFFT> ifft;
};

class MinimumPhase
{
public:
    MinimumPhase(std::size_t N);

    void transform(const RealBuffer& b, RealBuffer& r);

private:
    const float eps = std::numeric_limits<float>::epsilon();
    std::size_t N;
    std::unique_ptr<FFT> fft;
    std::unique_ptr<IFFT> ifft;
    std::unique_ptr<Hilbert> hilbert;
    ComplexBuffer in; // fft buffers
    ComplexBuffer out; // fft buffers
    ComplexBuffer Hin;
    ComplexBuffer Hout;
};

//! FIR Filter designer using the frequency sampling method
class FIR2MinphaseFilterDesigner
{
public:
    FIR2MinphaseFilterDesigner(std::size_t Ntaps, float fs);

    //! Destructor
    virtual ~FIR2MinphaseFilterDesigner();

    /*! Designs a minimum phase filter given a frequency vector and a gain using the frequency sampling method.
        frequency vector f runs from [DC ... nyquist] in (Hz)
        gain vector contains gains per frequency  (dimensionless)
        returns b, vector of length Ntaps with the filtercoefficients.
    */
    void design(const RealBuffer& f, const RealBuffer& gains, RealBuffer& b);

    void design(const RealBuffer& measuredFrequencies, const RealBuffer& measuredGains,
        const RealBuffer& targetFrequencies, const RealBuffer& targetGains, RealBuffer& b);

    void design(const RealBuffer& measuredFrequencies, const RealBuffer& measuredGains,
        const RealBuffer& targetFrequencies, const RealBuffer& targetGains,
        const RealBuffer& compensationFrequencies, const RealBuffer& compensationGains,
        RealBuffer& b);

    void phaseAlign(RealBuffer& ch1, RealBuffer& ch2, const size_t maxShift);

    /*! applies Hamming window on RealBuffer b
        note: made public for testing
    */
    void applyHamming(RealBuffer& b);

private:
    float fs; // samplerate
    std::size_t Ntaps; // number of taps filter
    std::size_t Nf; // Number of frequencies of onesided spectrum
    std::size_t Nds; // Number of frequencies of doublesided spectrum (or fft length)
    RealBuffer wn; // normalized frequencies
    RealBuffer k; // interpolated gains
    ComplexBuffer H; // interpolated spectrum
    RealBuffer h; // inverse fft result
    std::unique_ptr<FFT> fft;
    std::unique_ptr<IRFFT> ifft; // inverse fft
    std::unique_ptr<MinimumPhase> minimumPhase; // minimuphase estimator
};

/*! Piecewise linear interpolation, returns the interpolated values v, for every value in u given the vectors x and y.
*/
void piecelin(const RealBuffer& x, const RealBuffer& y, const RealBuffer& u, RealBuffer& v);

} // namespace EAR

#endif /* EAR_PCHIP_MINPHASE_FILTER_DESIGNER */
