/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_FIR_FILTER
#define EAR_FIR_FILTER

#include <atomic>
#include <memory>
#include <mutex>

namespace EAR {

class ComplexBuffer;
class ComplexFrame;
class RealBuffer;
class RealFrame;
class RFFT;
class IRFFT;

//! Fast general purpose FIR filtering (convolution) with arbitrary length filters
/*!
 * Implementation: GUPOLS (Generalized Uniformly Partitioned Overlap Save) Circular Convolution
 * Ref: F. Wefers 2014: Partitioned convolution algorithms for real-time auralization, Chapter 5
 */
class FIRFilter
{
public:
    //! Convolver initialization
    /*!
     * \param h      pointer to length N array with filter sample data
     * \param _B     number of samples per in/out block
     * \param _N     number of samples in filter
     * \param _K     number of samples of internal representation
     * \param _L     number of samples of filter partitioning
     */
    FIRFilter(const float* h, int _B, int _N, int _K, int _L);

    //! Convolver initialization without filter initialization
    /*!
     * \param _B     number of samples per in/out block
     * \param _N     number of samples in filter
     * \param _K     number of samples of internal represenation
     * \param _L     number of samples of filter partitioning
     */
    FIRFilter(int _B, int _N, int _K, int _L);

    //! Convolver initialization (K and L (see above) are chosen for optimal performance)
    /*!
     * \param h      pointer to length N array with filter sample data
     * \param _B     number of samples per in/out block
     * \param _N     number of samples in filter
     */
    FIRFilter(const float* h, int _B, int _N);

    //! Convolver initialization without filter initialization (K and L (see above) are chosen for optimal performance)
    /*!
     * \param _B     number of samples per in/out block
     * \param _N     number of samples in filter
     */
    FIRFilter(int _B, int _N);

    //! Destructor
    ~FIRFilter();

    //! Exchange filter used for convolution (initiates filter exchange in next processing block)
    /*!
     * Can savely be called from another thread as the process function
     *
     * \param _h     pointer to length N array with filter sample data
     */
    void setFilter(const float* _h);

    //! Perform convolution on input stream (input and output may point to the same memory location -> in-place operation)
    /*!
     * \param in     pointer to length B array with input sample data
     * \param out    pointer to length B array where output sample data should be written
     */
    void process(const float* in, float* out);

    //! Clear all input buffers without changing the filter
    void flush();

private:
    //! Internal state initialization
    void init();

    //! Helper function to partition and prepare filter for convolution
    void fillFilterBuffers(const float* h, RealFrame* timeDomainBuf, ComplexFrame* freqDomainBuf);

    //! Helper function to iterate over filter partitions
    void filterFDL(ComplexFrame* filterBuf, RealBuffer* result);

    //! flags for concurrency control
    bool running = false;
    std::atomic<bool> filterChanged{ false };
    std::mutex filterChangeMutex;

    //! length constants calculated upon instanciation
    const int B, N, K, L, P, Krealres;

    //! correctly padded time domain filter partitions
    std::unique_ptr<RealFrame> filt;
    //! frequency domain filter partitions
    std::unique_ptr<ComplexFrame> filtH;
    //! time domain filter partitions (filter exchange)
    std::unique_ptr<RealFrame> sidechain_filt;
    //! frequency domain filter partitions (filter exchange)
    std::unique_ptr<ComplexFrame> sidechain_filtH;

    //! time domain input buffer with history
    std::unique_ptr<RealBuffer> inBuffer;
    //! frequency delay line
    std::unique_ptr<ComplexFrame> fdl;
    //! frequency delay line read/write cursor
    int fdlPos = 0;

    //! sum of cplx multiplications with all partitions
    std::unique_ptr<ComplexBuffer> spectralSum;
    //! time domain result
    std::unique_ptr<RealBuffer> signalSum;
    //! time domain result (filter exchange)
    std::unique_ptr<RealBuffer> sidechain_signalSum;

    //! fade envelopes (predefined on init)
    std::unique_ptr<RealBuffer> fadeIn, fadeOut;

    //! real to complex FFT processor to be used in setFilter() (on the update thread)
    std::unique_ptr<RFFT> fftU;
    //! real to complex FFT processor to be used in process() (on the audio thread)
    std::unique_ptr<RFFT> fftP;
    //! complex to real IFFT processor
    std::unique_ptr<IRFFT> ifft;
};

} // namespace EAR

#endif /* EAR_FIR_FILTER */
