/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_FIXED_HRIR_PROVIDER
#define EAR_FIXED_HRIR_PROVIDER

#include <EAR/BaseTypes.h>
#include <EAR/IRProvider.h>

#include <cstddef>

namespace EAR {

// fwd declare
class CartesianKDTree;

//! Provides directional IRs from an HRIR dataset compiled into EAR
/*!
 * This dataset is taken from the FABIAN head-related transfer function database.
 * History: Resampled at 48kHz; Diffuse-field equalized using the filter provided in 
 *      FABIAN_CTF_measured_inverted_smoothed.sofa; Processed by Ericsson;
 * License: Creative Commons (CC-BY) Visit http://creativecommons.org/licenses/by/4.0/ 
 *      for licence details.
 * Organization: Audio Communication Group, TU Berlin, Germany (www.ak.tu-berlin.de)
 * Origin : HRIRs were acoustically measured in an anechoic chamber using swept sines, 
 *      and spectral deconvolution.
 * References : F.Brinkmann et al.(2017) : 
 *         "A High Resolution and Full-Spherical Head-Related Transfer Function Database for "
 *         "Different Head-Above-Torso Orientations.\"
 *         J.Audio Eng.Soc.65(10) : 841 - 848. DOI : 10.17743/ jaes .2017.0033
 */
class FixedHRIRProvider : public IRProvider
{

public:
    //! Constructor
    FixedHRIRProvider();

    //! Destructor
    ~FixedHRIRProvider();

    //! Inherited from IRProvider
    void getNearestIRForLocation(RealFrame& ir, Vector3 loc) const override;

    //! Inherited from IRProvider
    std::unique_ptr<const float*[]> getIRPointerForLocation(Vector3 loc) const override;

    //! Inherited from IRProvider
    virtual std::unique_ptr<const float*[]> getIRPointerForIndex(int index) const override;

    //! Inherited from IRProvider
    virtual std::vector<int> getAllIndicesInRange(
        const EAR::Polar& bottomLeft, const EAR::Polar& topRight) const override;

    //! Inherited from IRProvider
    int irLength() const override;

    //! Inherited from IRProvider
    int irCount() const override;

    //! Inherited from IRProvider
    float getSampleRate() const override;

    //! Inherited from IRProvider
    Vector3 getNearestLocation(Vector3& loc) const override;

    //! Inherited from IRProvider
    Vector3 getLocationForIndex(int idx) const override;

    //! Returns the distance of the nearest IR to the given location
    /*!
     * \param loc       A Vector3 point with the source location. This point
     *                  should be in a coordinate system relative to the listener.
     * \return          Distance to closest match [meters]
     */
    float getDistanceToNearestIndex(Vector3& loc) const;

private:
    //! Returns the index of the nearest IR to the given location
    unsigned long getNearestIndex(Vector3& loc) const;

    const float* getPointerToIndexForLeft(std::size_t idx) const;
    const float* getPointerToIndexForRight(std::size_t idx) const;

    const unsigned long N, M;
    const float fs;

    std::vector<EAR::Vector3> sourcePoints;
    std::unique_ptr<CartesianKDTree> lookupTree;

    const float* cache;
    const float* endOfCache;
    const float* directions;
};

} // namespace EAR

#endif /* EAR_FIXED_HRIR_PROVIDER */
