/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_FRAME_BASE
#define EAR_FRAME_BASE

#include <cstddef>
#include <memory>
#include <vector>

namespace EAR {

//! A group of same-length Buffer instances
/*!
 * Base class for specializations with float/RealBuffer (RealFrame) and std::complex<float>/ComplexBuffer (ComplexFrame) buffers.
 */
template <typename T, class B> class Frame
{
public:
    //! Initialize empty buffer
    Frame();

    //! Initialize with zeros
    /*!
     * \param channels  number of buffers
     * \param length    buffer length [samples]
     */
    Frame(std::size_t channels, std::size_t length);

    //! Copy constructor
    Frame(const Frame<T, B>& other);

    //! Destructor
    virtual ~Frame();

    //! Assignment
    Frame<T, B>& operator=(const Frame<T, B>& other);

    // Move
    // Frame<T, B>& operator=(Frame<T, B>&& other);

    //! In-place multiplication with a constant
    /*!
     * \param c    constant factor
     */
    void mul(T c);

    //! In-place addition with a constant
    /*!
     * \param c    constant value
     */
    void add(T c);

    //! In-place linear interpolation with another frame
    /*!
     * \param c    constant value
     */
    void lerp(const Frame<T, B>& tgt, float weight);

    //! Get const 2d pointer to underlying data
    std::unique_ptr<const T*[]> data() const;

    //! Get 2d pointer to underlying data
    std::unique_ptr<T*[]> data();

    //! Get 2d pointer to underlying data
    T** dataRaw();

    //! Direct const access to buffers
    /*!
     * \param index   index
     */
    const B& operator[](std::size_t index) const;

    //! Direct access to buffers
    /*!
     * \param index   index
     */
    B& operator[](std::size_t index);

    //! Direct const access to buffers
    /*!
     * \param index   index
     */
    const B& ch(std::size_t index) const;

    //! Direct access to buffers
    /*!
     * \param index   index
     */
    B& ch(std::size_t index);

    //! Copy all values from all buffers in other Frame (receiving buffers are resized accordingly - resize is a slow operation!)
    /*!
     * \param src     other Frame
     */
    void copyFrom(const Frame<T, B>& src);

    //! Write values into this Frame from interleaved source (Frame will be resized to fit).
    /*!
     * \param src           memory location of interleaved samples to be read
     * \param channels      number of channels
     * \param length        number of samples per channel
     */
    void writeInterleaved(const T* src, std::size_t channels, std::size_t length);

    //! Read a range of sample values from the Frame into an interleaved buffer.
    /*!
     * \param dst           memory location where samples are written interleaved
     * \param chOffset      first channel to copy from
     * \param chCount       number of channels to copy
     * \param smplOffset    first sample to copy in each channel
     * \param smplCount     number of samples to copy in each channel
     */
    void readInterleaved(T* dst, std::size_t chOffset, std::size_t chCount, std::size_t smplOffset,
        std::size_t smplCount);

    //! Cyclic shift of samples in all buffers to the left (first samples added to the end again).
    /*!
     * \param count    number of samples to shift
     */
    void shiftLeft(std::size_t count);

    //! Cyclic shift of samples in all buffers to the right (last samples added to the front again).
    /*!
     * \param count    number of samples to shift
     */
    void shiftRight(std::size_t count);

    // --= Channel operations =--

    //! Cyclic shift up of Buffers in the Frame (last channels are put to the front again).
    /*!
     * \param count    number of channels to shift
     */
    void shiftUp(std::size_t count);

    //! Cyclic shift down of Buffers in the Frame (first channels are put to the end again).
    /*!
     * \param count    number of channels to shift
     */
    void shiftDown(std::size_t count);

    //! Resize allocated frame and buffers to new length. This is a slow operation and may invalidate pointers acquired with frame->data()!
    /*!
     * \param newChannels   new number of buffers
     * \param newLength     new length of buffer [samples]
     */
    void resize(std::size_t newChannels, std::size_t newLength);

    //! Fill all buffers in frame with a constant value
    /*!
     * \param value     constant value to fill buffers with
     */
    void fill(T value);

    //! Fill a specified channel with a constant value
    /*!
     * \param value     constant value to fill buffer with
     * \param channel   channel to fill
     */
    void fill(T value, std::size_t channel);

    //! Fill a specified range of a specified channel with a constant value
    /*!
     * \param value     constant value to fill buffer range with
     * \param channel   channel to operate on
     * \param offset    first sample in buffer to overwrite
     * \param count     number of samples in buffer to overwrite
     */
    void fill(T value, std::size_t channel, std::size_t offset, std::size_t count);

    //! Fill all buffers in frame with zeros
    void zero();

    //! Fill a specified channel with zeros
    /*!
     * \param channel   channel to fill
     */
    void zero(std::size_t channel);

    //! Fill a specified range of a specified channel with zeros
    /*!
     * \param channel   channel to operate on
     * \param offset    first sample in buffer to overwrite
     * \param count     number of samples in buffer to overwrite
     */
    void zero(std::size_t channel, std::size_t offset, std::size_t count);

    //! Return current length of all buffers [samples]
    std::size_t length() const;
    //! Return number of Buffers in the Frame
    std::size_t channels() const;

protected:
    //! underlying data structure
    std::vector<B> frame;
    std::vector<T*> bufPtrs;
};

} // namespace EAR

#endif /* EAR_FRAME_BASE */
