/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_GEOMETRY_BASE
#define EAR_GEOMETRY_BASE

#include <EAR/BaseTypes.h>
#include <EAR/PchipInterpolator.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

#include <array>
#include <memory>
#include <ostream>
#include <vector>

// fwd declares for KD-tree implementation used to retrieve adjacent vertices in a Mesh
namespace nanoflann {
template <typename Distance, class DatasetAdaptor, int DIM, typename IndexType>
class KDTreeSingleIndexAdaptor;
template <class T, class DataSource, typename _DistanceType> struct L2_Simple_Adaptor;
struct EarPointCloud;
} // namespace nanoflann

typedef nanoflann::KDTreeSingleIndexAdaptor<
    nanoflann::L2_Simple_Adaptor<float, nanoflann::EarPointCloud,
        float>, // euclidean distance setup
    nanoflann::EarPointCloud, // data structure
    3, //dimensions
    std::size_t>
    CartKDTree;

namespace EAR {
class Mesh;

class Geometry : public SceneObject
{
public:
    //! Constructor
    Geometry(ObjectID oid, std::string name = "", Position pos = Position())
        : SceneObject(oid, name, pos)
    {
        staticFlag.store(true); // geometry (in contrast to other SceneObjects) is static by default
    };

    //! Destructor
    virtual ~Geometry(){};

    //! Common epsilon for tests etc
    static const float EPSILON;

    //! Seperate different types of geometry
    enum class Type
    {
        Mesh,
        Box,
        Sphere,
        Cylinder
    };

    struct BoundingBoxDimension
    {
        float low{ 0 };
        float high{ 0 };
    };

    //! Bounding box of the geometry
    /*!
     * This is the size of the bounding box in three dimensions. Position or orientation of the box are not important.
     */
    using BoundingBox = std::array<BoundingBoxDimension, 3>;

    //! Detects whether or not a point is within the space of this object
    /*!
    * \param point     a point in space to be tested within this object
    * \return          true if the point is within the object, false if not
    */
    virtual bool isInside(const EAR::Vector3& point) const = 0;

    //! Return the closestPoint on a Geometry to the given point
    virtual Vector3 closestPointTo(const EAR::Vector3& point) const = 0;

    //! Returns the type of the geometry
    virtual Type getType() const = 0;

    //! Returns the surface area of this mesh
    virtual float getArea() const = 0;

    //! Returns the volume of this mesh
    virtual float getVolume() const = 0;

    //! Returns an underlying mesh if it exists. If not it returns a nullptr
    virtual const Mesh* getMesh() const = 0;

    //! Returns greatest distance between two points which lay on the surface and are furthest apart
    virtual float greatestDistance() const = 0;

    //! Return the extremal points of the contour of the geometry as seen from the given point of view
    /*!
     * If the point of view is inside the object, points that covers the frontal hemisphere
     * of the listener are returned.
     *
     * \param pov                point of view as EAR::Position (probably the listener position)
     * \param resultBottomLeft   cartesian coordinates of the bottom-left corner of the contour relative to the POV
     * \param resultBottomRight  cartesian coordinates of the bottom-right corner of the contour relative to the POV
     * \param resultTopLeft      cartesian coordinates of the top-left corner of the contour relative to the POV
     * \param resultTopRight     cartesian coordinates of the top-right corner of the contour relative to the POV
     * \param isInside           point of view is inside the geometry
     * \param closestPoint       the closest point on the geometry surface to the POV point (in world coordinates, NOT relative to the POV!)
     */
    virtual void getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
        EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
        bool& povIsInside, EAR::Vector3& closestPoint) const = 0;

    virtual void getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
        bool& povIsInside, EAR::Vector3& closestPoint) const = 0;

    //! Returns the bounding box of the geometry
    virtual BoundingBox getBoundingBox() const = 0;

    //! Returns the bounding box of the geometry in global coordinates. This may differ in size to the bounding box in local CSpace depending on global rotations.
    virtual BoundingBox getGlobalBoundingBox() const = 0;

    //! Returns the center position of the geometry in global coordinats.
    virtual EAR::Vector3 getCenterPosition() const = 0;
};

//! Describe the acoustic properties of an isotropic, locally-reacting material
class Material
{
public:
    Material(int bsId = -1, const std::string& name = "")
        : bsId(bsId)
        , name(name)
    {}
    //! Add a tuple of frequency and coeffs (coeffs sum to <= 1, rest is dissipated energy)
    /*!
    * \param f 	frequency [Hz]
    * \param r 	reflection coefficient at f (0-1)
    * \param s 	scattering coefficient at f (0-1)
    * \param t     transmission coefficient at f (0-1)
    * \param c     coupling coefficient at f (0-1)
    */
    void append(float f, float r, float s, float t, float c);

    //! Returns vector of all frequency values where this material has coefficients defined
    inline const std::vector<float>& getFrequencies() { return frequencies; }

    //! Returns an interpolated reflection coefficient for a given frequency
    /*!
     * \param f         frequency at which the reflection coefficient shall be calculated
     */
    inline float getReflectionCoeffForFreq(float f) const
    {
        return reflectionInterpolator->interpolate(f);
    }

    //! Returns an interpolated transmission coefficient for a given frequency
    /*!
     * \param f         frequency at which the reflection coefficient shall be calculated
     */
    inline float getTransmissionCoeffForFreq(float f) const
    {
        return transmissionInterpolator->interpolate(f);
    }

    //! Returns an interpolated coupling coefficient for a given frequency
    /*!
     * \param f         frequency at which the coupling coefficient shall be calculated
     */
    inline float getCouplingCoeffForFreq(float f) const
    {
        return couplingInterpolator->interpolate(f);
    }

    //! Returns vector of all defined reflection coefficients relating to the frequencies given by getFrequencies()
    inline const std::vector<float>& getReflectionCoeffs() const { return reflection; }

    //! Returns vector of all defined scattering coefficients relating to the frequencies given by getFrequencies()
    inline const std::vector<float>& getScatteringCoeffs() const { return scattering; }

    //! Returns vector of all defined transmission coefficients relating to the frequencies given by getFrequencies()
    inline const std::vector<float>& getTransmissionCoeffs() const { return transmission; }

    //! Returns vector of all defined coupling coefficients relating to the frequencies given by getFrequencies()
    inline const std::vector<float>& getCouplingCoeffs() const { return coupling; }

    //! Returns the optional human-readable name of the material (may be empty string)
    inline const std::string& getName() { return name; }

    //! Returns the unique material bitstream id
    inline const int getBsId() { return bsId; }

    //! Sets the optional human-readable name of the material
    inline void setName(std::string matName) { name = matName; }

private:
    int bsId; //!< Unique bitstream id
    std::string name = ""; //!< Displayed name (optional)
    std::vector<float>
        frequencies; //!< Frequency values at which absorption and/or scattering coefficients are given [Hz]
    std::vector<float>
        reflection; //!< Reflection coefficients as factor from 0.0 (no specular reflection) to 1.0 (all energy is specularly reflected)
    std::vector<float>
        scattering; //!< Scattering coefficients as factor from 0.0 (no energy scattered) to 1.0 (all energy is randomly scattered)
    std::vector<float>
        transmission; //!< Transmission coefficients as factor from 0.0 (no energy is transmitted) to 1.0 (material is acoustically invisible)
    std::vector<float>
        coupling; //!< Coupling coefficients as factor from 0.0 (no energy is coupled) to 1.0 (all energy is used to resonate in the material)

    std::unique_ptr<PchipInterpolator> reflectionInterpolator =
        std::make_unique<PchipInterpolator>(PchipInterpolator(true));
    std::unique_ptr<PchipInterpolator> transmissionInterpolator =
        std::make_unique<PchipInterpolator>(PchipInterpolator(true));
    std::unique_ptr<PchipInterpolator> couplingInterpolator =
        std::make_unique<PchipInterpolator>(PchipInterpolator(true));
};

std::ostream& operator<<(std::ostream& os, Material& m);

} /* namespace EAR*/

#endif /* EAR_GEOMETRY_BASE */
