/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include <EAR/GeometryBase.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

#include <cmath>
#include <limits>

namespace EAR {
//! Represents an audio object with high order ambisonics.
/*!
 * Instanciation must be handled by the SceneObjectFactory.
 * Position handling is inherited from the SceneObject base class.
 */
class HOASource : public SceneObject
{
public:
    // Representation of HOA expansion
    enum class Representation
    {
        Interior,
        Exterior
    };

private:
    //! Constructor (only accessible through SceneObjectFactory)
    /*!
     * \param oid              fixed object ID (always assigned by factory method!)
     * \param representation   interior or exterior HOA expansion
     * \param group            group id to which this source belongs (for multipoint, -1 -> no group)
     * \param extentTransform  whether the source should do extent transform (default false)
     * \param transitionDist   optional transition distance for the extent transform
     * \param name             arbitrary name of object (optional, defaults to "")
     * \param pos              Position struct with initial location and orientation (optional, defaults to 0,0,0,0,0,0)
     */
    HOASource(ObjectID oid, Representation representation, int group = -1,
        bool extentTransform = false, float transitionDist = 0.f, std::string name = "",
        Position pos = Position())
        : SceneObject(oid, name, pos)
        , representation(representation)
        , group(group)
        , extentTransform(extentTransform)
        , transitionDistance(transitionDist)
        , granularGroupProb(-1.0, 0.0, 0.0){};

    // factory method can access private constructor
    friend SceneObjectFactory<HOASource>;

public:
    //! Set if the HOASource should be attenuated based on its distance to the listener
    /*!
     * \param isDistanceEnabled     true to enable the attenuation with a distance model
     */
    void setDistanceModel(bool isDistanceEnabled)
    {
        if (isDistanceEnabled != distanceModel.load()) {
            distanceModel.store(isDistanceEnabled);
            moved.store(true);
        }
    }

    //! Indicates if the HOASource should be attenuated based on its distance to the listener
    /*!
     * \return          true if source should be attenuated
     */
    bool isDistanceModelEnabled() { return distanceModel.load(); }

    //! Get gain value of SceneObject (thread-safe)
    /*!
     * \return          current gain in dB (may return -inf, if factor is 0)
     */
    float getGainDb() const { return gaindB.load(); }

    //! Get loudness value of SceneObject(thread - safe)
    /*!
     * \return          current loudness in dB (may return -inf)
     */
    float getLoudnessDb() const { return loudnessdB.load(); }

    //! Get gain value of SceneObject (thread-safe)
    /*!
     * \return          current gain as a factor (not dB!)
     */
    float getGainFactor() const { return gain.load(); }

    //! Get reverb gain value of SceneObject (thread-safe)
    /*!
     * \return          current reverb gain as a factor (not dB!)
     */
    float getReverbGain() const { return itemReverbGain.load(); }

    //! Get noReverb bool of SceneObject (thread-safe)
    /*!
     * \return          current noReverb as a bool
     */
    bool getNoReverb() const { return noReverb.load(); }
    //! Get loudness value of SceneObject (thread-safe)
    /*!
     * \return          current loudness as a factor (not dB!)
     */
    std::optional<float> getLoudnessFactor() const { return loudness; }

    //! Set gain value of SceneObject (thread-safe)
    /*!
     * Default is 0 dB. It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * But you can set it to -inf if you want.
     * \param newGain  new gain in dB
     */
    void setGainDb(float newGain)
    {
        if (newGain != gaindB.load()) {
            float newFactor = std::pow(10, newGain / 20);
            gain.store(newFactor);
            gaindB.store(newGain);
            //  moved.store(true);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

    //!Set loudness value of SceneObject(thread - safe)
    /*!
     * \param newLoudness  new loudness in dB
     */
    void setLoudnessDb(float newLoudness)
    {
        if (newLoudness != loudnessdB.load()) {
            auto const newFactor = std::pow(10.f, newLoudness / 20.f);
            if (!loudness.has_value()) {
                loudness.emplace(newFactor);
            } else {
                loudness.value().store(newFactor);
            }

            loudnessdB.store(newLoudness);
            notify(SceneObjectObserver::Property::Loudness);
        }
    }

    //! Set gain value of SceneObject (thread-safe)
    /*!
     * Default is 1. It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * \param newFactor     new gain as a factor (not dB!)
     */
    void setGainFactor(float newFactor)
    {
        if (newFactor != gain.load()) {
            float newGain = -1 * std::numeric_limits<float>::infinity();

            if (newFactor > 0) {
                newGain = 20 * std::log10(newFactor);
            }

            gain.store(newFactor);
            gaindB.store(newGain);
            moved.store(true);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

    //! Set reverb gain value of SceneObject (thread-safe)
    /*!
     * \param itemReverbGain     new gain as a factor (not dB!)
     */
    void setReverbGain(float newItemReverbGain)
    {
        // clamp reverbGain value between -119 and +12 dB
        if (newItemReverbGain > 3.9810717055349725077) {
            newItemReverbGain = 3.9810717055349725077;
        }
        if (newItemReverbGain < 0.00000112201845430196) {
            newItemReverbGain = 0.00000112201845430196;
        }
        if (newItemReverbGain != itemReverbGain.exchange(newItemReverbGain)) {
            notify(SceneObjectObserver::Property::ReverbGain);
        }
    }

    //! Set noReverb of SceneObject (thread-safe)
    /*!
     * \param noReverb     new bool 
     */
    void setNoReverb(bool newNoReverb)
    {
        if (newNoReverb != noReverb.exchange(newNoReverb)) {
            notify(SceneObjectObserver::Property::NoReverb);
        }
    }

    //! Set the representation of this HOASource
    void setRepresentation(Representation newRep) { representation = newRep; }

    //! Get the representation of this HOASource
    Representation getRepresentation() const noexcept { return representation; }

    //! Annotate the Source with a new Extent geometry (use nullptr to remove an existing extent)
    /*!
     * \param newExtent    pointer to an EAR::Geometry type or nullptr
     */
    void setExtent(Geometry* newExtent)
    {
        if (newExtent != extent.load()) {
            extent.store(newExtent);
            moved.store(true);
            notify(SceneObjectObserver::Property::Extent);
        }
    }

    //! Returns a Geometry object that defines the extent of this PointSource (or nullptr in non-existent)
    /*!
     * \return    pointer to an EAR::Geometry type or nullptr
     */
    Geometry* getExtent() const { return extent.load(); };

    //! Set the reference distance
    /*!
     * The reference distance specifies the distance at which the attenuation factor of the distance model shall be 1.
     */
    void setReferenceDistance(float newRefDist)
    {
        if (newRefDist != refDistance.load()) {
            refDistance.store(newRefDist);
            moved.store(true);
            notify(SceneObjectObserver::Property::ReferenceDistance);
        }
    }

    //! Returns the current reference distance
    float getReferenceDistance() const { return refDistance.load(); }

    //! Indicates whether the HOA source should be handled with extent transform
    bool doesExtentTransform() const { return extentTransform; }

    //! Returns the transition distance for an optional extent transform
    float getTransitionDistance() const { return transitionDistance; }

    //! Returns the group id of the multipoint HOA configuration this source may be affiliated with
    /*!
     * Group -1 means the source is not in a group (i.e. it is not a multipoint source)
     */
    int getGroup() const noexcept { return group; }

    //! Returns `true` if the source is a 6DoF HOA source
    /*!
     * If `true`, the source is rendered in 6DoF within its region of validity (or everywhere if no
     * extent is defined).
     *
     * Also enables the following attributes when `true`: group, refDistance
     */
    bool getIs6DoF() const noexcept { return is6DoF; }

    //! Set Variable Delay Line interpolation method to be used
    void setVDLMethod(VDLMethod newVDLMethod)
    {
        if (newVDLMethod != vdlMethod.exchange(newVDLMethod)) {
            moved.store(true);
            notify(SceneObjectObserver::Property::VDLMethod);
        }
    };

    //! Set the granular descriptor space position
    void setGranularPos(Vector3 newGranularPos)
    {
        if (newGranularPos != granularPos) {
            granularPos = newGranularPos;
            moved.store(true);
            notify(SceneObjectObserver::Property::GranularPos);
        }
    }

    //! Set the probability of the granular probability group specified by GranularProbGroupNo
    void setGranularGroupProb(Vector3 newGranularGroupProb)
    {
        if (newGranularGroupProb != granularGroupProb) {
            granularGroupProb = newGranularGroupProb;
            moved.store(true);
            notify(SceneObjectObserver::Property::GranularGroupProb);
        }
    }

    //! Returns the vdl method
    VDLMethod getVDLMethod() { return vdlMethod.load(); };

    //! Returns the granular position
    Vector3 getGranularPos() { return granularPos; };

    //! Returns the granular group probability
    Vector3 getGranularGroupProb() { return granularGroupProb; };

private:
    //! gain factor (not dB!) that shall be applied in a rendering scenario
    std::atomic<float> gain{ 1.f };

    //! gain in dB (factor and dB are stored to make either get function cheaper)
    std::atomic<float> gaindB{ 0.f };

    //! gain factor (not dB!) that shall be applied in a reverb rendering scenario
    std::atomic<float> itemReverbGain{ 1.f };

    //! noReverb flag that shall be applied in a reverb rendering scenario
    std::atomic<bool> noReverb{ false };

    //! loudness factor (not dB!) that shall be applied in a rendering scenario
    std::optional<std::atomic<float>> loudness{ std::nullopt };

    //! loudness in dB (factor and dB are stored to make either get function cheaper)
    std::atomic<float> loudnessdB{ 0.f };

    //! flag to indicate if the HOASource should be attenuated based on its distance to the listener
    std::atomic<bool> distanceModel{ true };

    //! Reference distance for distance models
    std::atomic<float> refDistance{ 1.f };

    //! Interior or exterior HOA expansion
    Representation representation;

    //! Is this a 6DoF HOA source
    bool is6DoF{ false };

    //! Group id (-1 for none, cannot be changed at runtime)
    const int group;

    //! Transition distance for an optional extent transform
    const bool extentTransform;

    //! Transition distance for an optional extent transform
    const float transitionDistance;

    //! Pointer to a Geometry object that defines the source's spatial extent
    std::atomic<Geometry*> extent{ nullptr };

    //! VDL method to be used per object
    std::atomic<VDLMethod> vdlMethod;

    //! Granular descriptor space target position
    Vector3 granularPos;

    //! Updated probability of the group specified by granularProbGroup
    Vector3 granularGroupProb;
};

} // namespace EAR
