/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/
#ifndef EAR_HEADPHONE_EQ
#define EAR_HEADPHONE_EQ

#include <EAR/Exception.h>
#include <EAR/FIRFilter.h>
#include <map>
#include <vector>

namespace EAR {

enum class HeadphoneSpecType
{
    unset = -1,
    freqgain = 0,
    parametricfilters = 1,
    preset = 2,
};

enum class HeadphoneFilterType
{
    lowshelf = 0,
    peak = 1,
    highshelf = 2,
};

struct HeadphoneEQData
{
    HeadphoneEQData() {}

    HeadphoneSpecType specType = HeadphoneSpecType::unset;
    std::vector<float> centerFreq;
    std::vector<float> gain;
    std::vector<float> normalizedGainLin;
    std::vector<float> q;
    std::vector<HeadphoneFilterType> filterType;
    int preset;
};

class HeadphoneEQPresets
{
    //! EQpresets, provides frequency vector and gain depending on chosen preset
private:
    std::vector<float> frequencies = { 70, 180, 320, 600, 1000, 3000, 6000, 12000, 14000, 16000 };
    std::map<int, std::vector<float>> gains = { { 0, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } },
        { 1, { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -6.5, -6.5, -6.5, -8.5 } },
        { 2, { 0.0, 0.0, 2.5, 4.5, 4.5, 4.5, 2.5, 0.0, 0.0, 0.0 } },
        { 3, { 7.5, 5.5, 1.5, -0.5, -0.5, -5.5, -6.5, -6.5, -0.5, -0.5 } },
        { 4, { 3.5, 8.5, 4.0, -3.5, -2.5, 1.0, 3.5, 7.5, 10.0, 11.5 } },
        { 5, { 8.0, 8.0, 4.5, 4.5, 0.0, -4.5, -4.5, -4.5, 0.0, 0.0 } },
        { 6, { 5.5, 5.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 5.5, 5.5 } },
        { 7, { -2.0, 3.5, 5.5, 6.0, 4.0, -1.5, -2.5, -2.5, -2.0, -2.0 } },
        { 8, { 0.0, 0.0, -1.0, -5.5, 0.0, 5.0, 5.0, 0.0, 0.0, 0.0 } },
        { 9, { 6.0, 3.5, -5.0, -7.0, -3.5, 3.0, 7.0, 8.5, 8.5, 8.5 } },
        { 10, { 3.5, 1.0, -1.5, -2.5, -1.5, 3.0, 6.5, 7.5, 8.5, 9.5 } },
        { 11, { -2.5, -4.5, -4.0, -1.0, 3.0, 4.5, 7.0, 7.5, 8.5, 7.5 } },
        { 12, { 7.5, 7.5, 7.5, 4.5, 1.0, -4.0, -7.5, -9.0, -9.5, -9.5 } },
        { 13, { 3.0, 3.0, 1.5, -1.0, -4.0, -5.0, -3.5, -1.0, 2.0, 7.0 } },
        { 14, { -8.5, -8.5, -8.5, -4.0, 2.0, 8.5, 12.5, 12.5, 12.5, 13.5 } },
        { 15, { 5.5, 4.5, 0.0, -6.5, -4.5, 1.0, 6.5, 8.5, 9.5, 9.5 } },
        { 16, { -4.5, 0.0, 3.0, 4.0, 4.5, 4.5, 3.0, 2.0, 2.0, 1.5 } },
        { 17, { 6.0, 4.5, 0.0, -5.0, -4.5, 0.0, 6.0, 7.5, 7.5, 7.0 } } };

public:
    HeadphoneEQPresets(){};

    enum PresetIndex
    {
        flat = 0,
        classical = 1,
        club = 2,
        dance = 3,
        midcut = 4,
        large_hall = 5,
        party = 6,
        pop = 7,
        reggae = 8,
        rock = 9,
        soft = 10,
        ska = 11,
        full_bass = 12,
        soft_rock = 13,
        full_treble = 14,
        full_bass_treble = 15,
        live = 16,
        techno = 17,
    };

    const std::vector<float> getFreqs() { return frequencies; }
    const std::vector<float> getGains(int index)
    {
        if (auto it{ gains.find(index) }; it != std::end(gains)) {
            return gains.at(index);
        } else {
            EAR_FAIL("Headphone EQ preset index not found in list");
        }
    }
    const std::tuple<std::vector<float>, std::vector<float>> getPreset(int index)
    {
        if (auto it{ gains.find(index) }; it != std::end(gains)) {
            return { frequencies, gains.at(index) };
        } else {
            EAR_FAIL("Headphone EQ preset index not found in list");
        }
    }
};

class HeadphoneEQ
{

public:
    static std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> calcFilters(
        const HeadphoneEQData& targetEQ, const HeadphoneEQData& measuredEQ, float fs, float B,
        const size_t Ntaps = 2048);

    static std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> calcFilters(
        const HeadphoneEQData& targetEQ, const HeadphoneEQData& measuredEQ,
        const HeadphoneEQData& spectralCompensation, float fs, float B, const size_t Ntaps = 2048);

    static std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> calcFilters(
        const HeadphoneEQData& targetEQ, const HeadphoneEQData& measuredEQ,
        const HeadphoneEQData& spectralCompensationLeft,
        const HeadphoneEQData& spectralCompensationRight, float fs, float B,
        const size_t Ntaps = 2048);

    static std::vector<float> convertGainsToLin(
        std::vector<float> dBGains, const bool substractAverage = true);

    static std::tuple<std::vector<float>, std::vector<float>> getGainsFromBiquad(
        const HeadphoneEQData& filters, float fs, float nTargetFreqs);
};

} // namespace EAR

#endif
