/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_IIRFILTER
#define EAR_IIRFILTER

#include <memory>

namespace EAR {

//! Infinite impulse response (IIR) filter
/*!
 * Base class for different implementations:
 * - Basic implementation of the direct form II in time domain.
 * - Fast implementation utilizing IPP functions
 */
class IIRFilter
{
public:
    //! Factory method, returns an IIRFilter instance
    /*!
     * Current options are:
     * - Basic implementation of the direct form II in time domain.
     * - Fast implementation utilizing IPP functions (used if IPP is found on the compiling machine)
     *
     * \param _b    FIR coefficients
     * \param nb    Number of FIR coefficients
     * \param _a    IIR coefficients
     * \param na    Number of IIR coefficients
     */
    static std::unique_ptr<IIRFilter> create(const float* _b, int nb, const float* _a, int na);

    //! Destructor
    virtual ~IIRFilter(){};

    //! Empty all internal buffers
    virtual void clear() = 0;

    //! Process a block of samples
    /*!
     * \param input     pointer to input samples
     * \param output    pointer to output memory location
     * \param nsamples  number of samples to process (arbitrary length)
     */
    virtual void process(const float* input, float* output, int nsamples) = 0;

    //! Process a block of samples with the possibility for inherent upsampling
    /*!
     * \param input     pointer to input samples
     * \param output    pointer to output memory location
     *                  (required length differs if factor is != 1, and can be calculated by factor*nsamples)
     * \param nsamples  number of samples to process from the input buffer (arbitrary length)
     * \param factor    upsampling factor (only integer ratios, downsampling currently not supported)
     */
    // virtual void process(const float* input, float* output, int nsamples, int factor=1) = 0;

protected:
    //! Protected constructor to avoid direct instanciation
    IIRFilter(){};
};

} // namespace EAR

#endif /* EAR_IIRFILTER */
