/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_IR_PROVIDER
#define EAR_IR_PROVIDER

#include <memory>
#include <vector>

namespace EAR {

// fwd declarations
class Polar;
class RealFrame;
class Vector3;

//! Interface for impulse response (IR) providers based on spatial lookup
class IRProvider
{

public:
    //! Destructor
    virtual ~IRProvider() = default;

    //! Look up the directional filter that corresponds to the given cartesian source position
    /*!
     * Polar coordinates would be more human-readable here, but we stick to the convention of having all-cartesian interfaces
     *
     * \param ir        An AudioFrame where the HRIRs should be written to (will be resized to fit).
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     */
    virtual void getNearestIRForLocation(RealFrame& ir, Vector3 loc) const = 0;

    //! Look up the directional filter that corresponds to the given cartesian source position and returns actual cache pointer.
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     * \return          unique pointer to the actual IR cache position
     */
    virtual std::unique_ptr<const float*[]> getIRPointerForLocation(Vector3 loc) const = 0;

    //! Returns actual cache pointer for a given IR index.
    /*!
     * \param index     an index as returned by getAllIndicesInRange() (range is not checked!)
     * \return          unique pointer to the actual IR cache position
     */
    virtual std::unique_ptr<const float*[]> getIRPointerForIndex(int index) const = 0;

    //! Returns a list of indices of all IRs that are inside of a given azimuth and elevation range
    /*!
     * \param bottomLeft    polar coordinates of the bottom left corner of the search range
     * \param topRight      polar coordinates of the top right corner of the search range
     */
    virtual std::vector<int> getAllIndicesInRange(
        const EAR::Polar& bottomLeft, const EAR::Polar& topRight) const = 0;

    //! Returns the length in taps of the filters available.
    virtual int irLength() const = 0;

    //! Returns the number of available filters for different directions available.
    virtual int irCount() const = 0;

    //! Returns the sample rate of the filters in the given SOFA file.
    virtual float getSampleRate() const = 0;

    //! Returns the location of the nearest IR to the given location
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     * \return          Vector3 location of the closest measured point
     */
    virtual Vector3 getNearestLocation(Vector3& loc) const = 0;

    //! Returns the location of a hrir point using its index
    /*!
     * \param idx       the index of the hrir point
     * \return          its location
     */
    virtual Vector3 getLocationForIndex(int idx) const = 0;
};

} // namespace EAR

#endif /* EAR_IR_PROVIDER */
