/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_INTERPOLATOR
#define EAR_INTERPOLATOR

#include <cstddef>
#include <vector>

namespace EAR {

//! Interface for Interpolators
/*!
 * Generates a mapping from arbitrary input of type Tx onto output of type Ty based on a number of supporting points.
 */
template <typename Tx, typename Ty> class Interpolator
{
public:
    //! Destructor
    virtual ~Interpolator(){};

    //! Provide supporting points to generate the mapping
    /*!
     * \param x       x values of the supporting points (length n)
     * \param y       y values of the supporting points (length n)
     * \param n       number of supporting points
     */
    virtual void initialize(const Tx* x, const Ty* y, std::size_t n) = 0;

    //! Insert a new supporting point and reinitialize if required
    /*!
     * \param x      single new x value of supporting point
     * \param y      single new y value of supporting point
     */
    virtual void addSupportingPoint(Tx x, Ty y) = 0;

    //! Insert a new supporting point and reinitialize if required (with specified interpolation)
    /*!
     * Default function, so derived classes do not have to implement these.
     * \param x             single new x value of supporting point
     * \param y             single new y value of supporting point
     * \param _interpolate  whether or not the values between this and the previous x should be interpolated
     */
    virtual void addSupportingPoint(Tx x, Ty y, bool _interpolate)
    {
        (void)_interpolate; // unused in this implementation, but kept to have consistent interface
        addSupportingPoint(x, y);
    };

    //! Returns true if this x is in a region with active interpolation (see addSupportingPoint)
    /*!
     * Default function that returns always true, so derived classes do not have to implement these if they always interpolate.
     *
     * \param x             x value of interest
     * \param px            previous x that may be given in order to find out if a 
     *                      non-interpolating step happened between x and px
     */
    virtual bool isChanging(Tx x, Tx px = 0)
    {
        (void)x; // unused in this implementation, but kept to have consistent interface
        (void)px; // unused in this implementation, but kept to have consistent interface
        return true;
    };

    //! Return the slope of y at x
    /*!
     * Default function, that always returns the 0-element of Ty, so derived classes do not have to implement this.
     */
    virtual Ty getSlope(Tx x)
    {
        (void)x; // unused in this implementation, but kept to have consistent interface
        return Ty();
    };

    //! Provide interpolated y value for a single x
    /*!
     * \param x       x value at which the polynomials shall be evaluated (length n)
     */
    virtual Ty interpolate(Tx x) = 0;

    //! Provide interpolated y value for a single x
    /*!
     * \param x       x value at which the polynomials shall be evaluated (length n)
     */
    inline Ty operator()(Tx x) { return interpolate(x); };
};

} // namespace EAR

#endif /* EAR_INTERPOLATOR */
