/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_LISTENER
#define EAR_LISTENER

#include <EAR/RealFrame.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

#include <cmath>
#include <limits>
#include <vector>

namespace EAR {
//! Represents a binaural receiver in the scene.
/*!
 * Instanciation shall be handled by the SceneObjectFactory.
 * Position handling is inherited from the SceneObject base class.
 */
class Listener : public SceneObject
{

    //! Constructor (only accessible through SceneObjectFactory)
    /*!
     * \param oid              fixed object ID (always assigned by factory method!)
     * \param name             arbitrary name of object (optional, defaults to "")
     * \param pos              Position struct with initial location and orientation (optional, defaults to 0,0,0,0,0,0)
     */
    Listener(ObjectID oid, std::string name = "", Position pos = Position())
        : SceneObject(oid, name, pos){};

    // factory method can access private constructor
    friend SceneObjectFactory<Listener>;

public:
    //! Get gain value of SceneObject (thread-safe)
    /*!
     * \return          current gain in dB (may return -inf, if factor is 0)
     */
    float getGainDb() const { return gaindB.load(); }

    //! Get reverb gain value of SceneObject (thread-safe)
    /*!
     * \return          current reverb gain 
     */
    float getReverbGain() const { return itemReverbGain.load(); }

    //! Get noReverb of SceneObject (thread-safe)
    /*!
     * \return   noReverb       current noReverb
     */
    bool getNoReverb() const { return noReverb.load(); }

    //! Get gain value of SceneObject (thread-safe)
    /*!
     * \return          current gain as a factor (not dB!)
     */
    float getGainFactor() const { return gain.load(); }

    //! Set gain value of SceneObject (thread-safe)
    /*!
     * Default is 0 dB. It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * But you can set it to -inf if you want.
     * \param newGain  new gain in dB
     */
    void setGainDb(float newGain)
    {
        if (newGain != gaindB.load()) {
            float newFactor = std::pow(10, newGain / 20);
            gain.store(newFactor);
            gaindB.store(newGain);
            // moved.store(true);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

    //! Set reverb gain value of SceneObject (thread-safe)
    /*!
     * Default is 1 . It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * But you can set it to -inf if you want.
     * \param newGain  new gain 
     */
    void setReverbGain(float newItemReverbGain)
    {
        // clamp reverbGain value between -119 and +12 dB
        if (newItemReverbGain > 3.9810717055349725077) {
            newItemReverbGain = 3.9810717055349725077;
        }
        if (newItemReverbGain < 0.00000112201845430196) {
            newItemReverbGain = 0.00000112201845430196;
        }
        if (newItemReverbGain != itemReverbGain.exchange(newItemReverbGain)) {
            notify(SceneObjectObserver::Property::ReverbGain);
        }
    }

    //! Set noReverb of SceneObject (thread-safe)
    /*!
     * Default is false . 
     * \param noReverb  new gain 
     */
    void setNoReverb(bool newNoReverb)
    {
        if (newNoReverb != noReverb.exchange(newNoReverb)) {
            notify(SceneObjectObserver::Property::NoReverb);
        }
    }

    //! Set gain value of SceneObject (thread-safe)
    /*!
     * Default is 1. It is advised to turn the SceneObject off with setActive(false) instead of providing very low gain values (below -96 dB or so).
     * \param newFactor     new gain as a factor (not dB!)
     */
    void setGainFactor(float newFactor)
    {
        if (newFactor != gain.load()) {
            float newGain = -1 * std::numeric_limits<float>::infinity();

            if (newFactor > 0) {
                newGain = 20 * std::log10(newFactor);
            }

            gain.store(newFactor);
            gaindB.store(newGain);
            // moved.store(true);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

    //! The position of the real listener in the listening room
    Position realPosition;

private:
    //! gain factor (not dB!) that shall be applied in a rendering scenario
    std::atomic<float> gain{ 1.f };

    //! gain in dB (factor and dB are stored to make either get function cheaper)
    std::atomic<float> gaindB{ 0.f };

    //! reverb gain factor (not dB!) that shall be applied in a rendering scenario
    std::atomic<float> itemReverbGain{ 1.f };

    //! reverb gain factor (not dB!) that shall be applied in a rendering scenario
    std::atomic<bool> noReverb{ false };
};

} // namespace EAR

#endif /* EAR_LISTENER */
