/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_LISTENERVOICESOURCE
#define EAR_LISTENERVOICESOURCE

#include <EAR/PointSource.h>
#include <EAR/SceneObjectFactory.h>
namespace EAR {

//! Represents an audio object to reproduce the user/listener voice in the scene.
/*!
 * Instanciation must be handled by the SceneObjectFactory.
 * Position handling is inherited from the SceneObject base class.
 */
class ListenerVoiceSource : public PointSource
{
    ListenerVoiceSource(ObjectID oid, std::string name = "", Position pos = Position(),
        std::shared_ptr<DirectivityModel> directivity = nullptr)
        : PointSource(oid, name, pos, directivity)
    {}

    // factory method can access private constructor
    friend SceneObjectFactory<ListenerVoiceSource>;

public:
    //! Get Direct sound gain value of Listener Voice object (thread-safe)
    /*!
     * \return          current gain in dB (may return -inf, if factor is 0)
     */
    float getDirectGainDb() const { return directGaindB.load(); }

    //! Set Direct sound gain value of Listener Voice object (thread-safe)
    /*!
     * \param newGain  new gain in dB
     */
    void setDirectGainDb(float newGain)
    {
        if (newGain != directGaindB.load()) {
            directGaindB.store(newGain);
            notify(SceneObjectObserver::Property::Gain);
        }
    }
    //! Get early reflections gain value of Listener Voice object (thread-safe)
    /*!
     * \return          current gain in dB (may return -inf, if factor is 0)
     */
    float getERGainDb() const { return erGaindB.load(); }

    //! Set early reflections gain value of Listener Voice object (thread-safe)
    /*!
     * \param newGain  new gain in dB
     */
    void setERGainDb(float newGain)
    {
        if (newGain != erGaindB.load()) {
            erGaindB.store(newGain);
            notify(SceneObjectObserver::Property::Gain);
        }
    }

private:
    //! Direct sound gain in dB
    std::atomic<float> directGaindB{ 0.f };

    //! Early Reflection gain in dB
    std::atomic<float> erGaindB{ 0.f };
};

} // namespace EAR

#endif /* EAR_LISTENERVOICESOURCE */
