/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include <EAR/PointSource.h>
#include <EAR/SceneObject.h>

#include <vector>

namespace EAR {

//! A set of loudspeakers that can be placed in space
/*!
 *  Instanciation shall be handled by the SceneObjectFactory
 */
class LoudspeakerLayout : public SceneObject
{
private:
    //! Constructor
    /*!
     * \param oid           fixed object ID (always assigned by factory method!)
     * \param layout        the desired layout
     * \param radius        the distance between the centre and the loudspeakers in meters
     * \param name          human readable name for this layout
     * \param pos           the center position (sweet spot of the LS layout) in space
     */
    LoudspeakerLayout(ObjectID oid, int layout, float radius = 1.f, std::string name = "",
        Position pos = Position());

    friend SceneObjectFactory<LoudspeakerLayout>;

public:
    //! Getter for all loudspeakers within
    const std::vector<std::shared_ptr<PointSource>>& getLoudspeakerPointSources() const;

    //! Inherited from SceneObject (overriden to also update loudspeakers)
    void setActive(bool newState) override;

    //! Set the distance between the centre of this layout and all of the loudspeakers in meters
    /*!
     * This deletes existing loudspeaker SceneObjects, which will invalidate observers of their changes.
     */
    void setRadius(float distance);

    //! Return the distance between the centre of this layout and all of the loudspeakers in meters
    float getRadius() const;

    //! Return the layout index; please check ISO/IEC 23091-3:2018 CICP-Audio for reference
    int getLayoutIndex() const;

    //! Inherited from SceneObject (overriden to also notify loudspeakers about changes on layout)
    void notify(SceneObjectObserver::Property mod) override;

    //! Set gain of all loudspeakers.
    void setGainDb(float newGain);

    float getReverbGain() { return itemReverbGain.load(); };

    //! Set reverb gain of all loudspeakers.
    void setReverbGain(float newItemReverbGain);

    bool getNoReverb() { return noReverb.load(); };

    //! Set noReverb of all loudspeakers.
    void setNoReverb(bool newNoReverb);

    //! Set the vdl method
    void setVDLMethod(VDLMethod newVDLMethod);

    //! Returns the vdl method
    VDLMethod getVDLMethod() { return vdlMethod.load(); };

    //! A list of all loudspeakers
    std::vector<std::shared_ptr<PointSource>> loudspeakers;

private:
    //! Calculates positions to place loudspeakers
    inline void constructLoudspeakers();

    //! The current layout of loudspeakers, please check ISO/IEC 23091-3:2018 CICP-Audio for layout index
    const int layout;

    //! Distance between the loudspeaker and the centre (sweet spot) of this layout
    float radius;

    //! VDL method to be used per channel source
    std::atomic<VDLMethod> vdlMethod;

    //! Reverb gain as factor
    std::atomic<float> itemReverbGain;

    //! noReverb as bool
    std::atomic<bool> noReverb;
};

} // namespace EAR
