/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_GEOMETRY_MESH
#define EAR_GEOMETRY_MESH

#include <EAR/GeometryBase.h>

#include <EAR/EmbreeTracer.h>
#include <EAR/config.h>

namespace EAR {

// fwd declare
class CartesianKDTree;

//! Arbitrary mesh geometry
class Mesh : public Geometry
{

    //! Constructor for empty Mesh (only accessible through SceneObjectFactory)
    /*!
    * \param oid        fixed object ID (always assigned by factory method!)
    * \param name       arbitrary name of object
    * \param pos        Position struct with initial location and orientation
    */
    Mesh(ObjectID oid, const std::string& name = "", Position pos = Position());

    // factory method can access private constructor
    friend SceneObjectFactory<Mesh>;

public:
    ~Mesh();
    //! Determines desired frame of reference
    enum class CSpace
    {
        Global, //!< Coordinates relative to a global origin
        Local //!< Coordinates relative to a local reference, e.g. the center point of a mesh
    };

    //! A triangle is described by three vertex indices.
    struct Triangle
    {
        //! Constructor

        Triangle(unsigned int v1, unsigned int v2, unsigned int v3, Mesh* parent,
            std::shared_ptr<Material> material = std::shared_ptr<Material>(nullptr),
            unsigned int surfaceId = -1)
            : v1(v1)
            , v2(v2)
            , v3(v3)
            , parent(parent)
            , material(material)
            , surfaceId(surfaceId){};

        unsigned int v1, v2, v3; //!< Vertex indices from parent Mesh

        Mesh* parent; //!< Pointer to parent Mesh holding the actual vertices used

        std::shared_ptr<Material>
            material; //!< An optional shared_ptr to the Material of the frontal side

        //! The ID of the surface this triangle belongs to (is set in rebuildSurfaceStructure())
        unsigned int surfaceId;

        //! Return the normal vector of the Triangle
        Vector3 getNormal(CSpace coordinateSpace = CSpace::Local) const;

        //! Return the centroid of the Triangle
        Vector3 getCentroid(CSpace coordinateSpace = CSpace::Local) const;

    private:
        friend Mesh;
    };

    struct Plane
    {
        Vector3 N0; //!< normal vector
        float d; //!< distance to the origin
    };

    struct Surface
    {

        Surface(unsigned int id, const std::vector<unsigned int>& triIds, const Plane& p, bool occ)
            : surfaceId(id)
            , triangleId(triIds)
            , plane(p)
            , canOcclude(occ)
            , canOccludeSelf(occ)
            , canOccludeOthers(false){};

        unsigned int surfaceId;
        std::vector<unsigned int>
            triangleId; //!< triangle mesh defined by vertex indices    //FIXME: rename -> triangleIds
        Plane plane; //!< plane given as Hesse normal form
        bool canOcclude; //!< is this surface generally a potential occluder?
        bool canOccludeSelf; //!< is this surface a potential occluder for itself?
        bool canOccludeOthers; //!< is this surface a potential occluder for other objects?
        //! Lists of triangle ids of sub surfaces (connected subsets of the whole surface)
        std::vector<std::vector<unsigned int>> subSurfaceTriangleIds;
    };

    struct RoundEdge
    {

        RoundEdge(unsigned int _vtxFrom, unsigned int _vtxTo, unsigned int _triLeft,
            unsigned int _triRight, float _alpha)
            : vtxFrom(_vtxFrom)
            , vtxTo(_vtxTo)
            , triLeft(_triLeft)
            , triRight(_triRight)
            , alpha(_alpha){};

        unsigned int vtxFrom; //!< first vertex index
        unsigned int vtxTo; //!< second vertex index
        unsigned int triLeft; //!< ID of the left-hand triangle
        unsigned int triRight; //!< ID of the right-hand triangle
        float alpha; //!< angle between the triangle normals
    };

    struct RoundCorner
    {

        RoundCorner(unsigned int _vtx, const std::vector<unsigned int>& _tidList)
            : vtx(_vtx)
            , tidList(_tidList){};

        unsigned int vtx; //!< vertex index
        std::vector<unsigned int> tidList; //!< set of triangle IDs
    };

    //! Inherited from SceneObject
    void setOrientation(Euler newOrient, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setLocation(Vector3 newLoc, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setCSpace(
        EAR::CSpace newCSpace, SceneObject* newParent = nullptr, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setPosition(
        Position newPos, SceneObject* parent = nullptr, bool notifyObs = true) override;

    //! Add a new vertex
    /*!
    * \param v 		a Vector3 determining the new vertex position in local coordinate space
    *
    * \return 			a vertex index for future reference
    */
    int addVertex(const Vector3& v);

    //! Add a new vertex with index (existing vertices may be overwritten)
    /*!
    * \param v         a Vector3 determining the new vertex position in local coordinate space
    * \param index     index of the given vertex (>= 0)
    */
    void addVertex(const Vector3& v, unsigned int idx);

    //! Return coordinates of a single vertex
    /*!
     * \param vertexid          ID of desired vertex (as returned by addVertex)
     * \param coordinateSpace   choose whether global or local coordinates of the vertex should be returned
     *                          (see CSpace enum)
     */
    const Vector3& getVertex(int vertexid, CSpace coordinateSpace = CSpace::Local) const;

    //! Add a new triangle
    /*!
    * The order of the vertex indices determines the front face via the normal direction.
    *
    * \param v1 		first vertex
    * \param v2 		second vertex
    * \param v3 		third vertex
    * \param material   optional material (default is nullptr)
    *
    * \return 			a triangle index for future reference
    */
    int addTriangle(unsigned int v1, unsigned int v2, unsigned int v3,
        std::shared_ptr<Material> material = std::shared_ptr<Material>(nullptr));

    //! Return a single Triangle
    /*!
     * \param triangleid        ID of desired triangle (as returned by addTriangle)
     */
    Triangle& getTriangle(int triangleid);

    //! Return a single Triangle (const access)
    /*!
     * \param triangleid        ID of desired triangle (as returned by addTriangle)
     */
    const Triangle& getTriangle(int triangleid) const;

    //! Retuns a pointer to the list of vertices required to describe the geometry object
    /*!
     * \param coordinateSpace   choose whether global or local coordinates of the vertices should be returned
     *                          (see CSpace enum)
     */
    const std::vector<Vector3>& getVertices(CSpace coordinateSpace = CSpace::Local) const;

    //! Returns a list of triangles
    const std::vector<Triangle>& getTriangles() const;

    //! Returns a list of surfaces
    const std::vector<Surface>& getSurfaces() const;

    //! Returns a list of round edges
    const std::vector<RoundEdge>& getRoundEdges() const;

    //! Returns a list of round corners
    const std::vector<RoundCorner>& getRoundCorners() const;

    //! Returns a plane definition given in Hesse normal form
    /*!
     * \param tri               A triangle in that plane
     * \param coordinateSpace   choose whether global or local coordinates of the vertices should be returned
     *                          (see CSpace enum)
     */
    Plane getPlane(const Triangle& tri, CSpace coordinateSpace = CSpace::Local) const;

    //! Returns a plane definition given in Hesse normal form
    /*!
     * \param A                 first point of the triangle
     * \param B                 second point of the triangle
     * \param C                 third point of the triangle
     * \return                  plane definition in Hesse normal form
     */
    Plane getPlane(const Vector3& A, const Vector3& B, const Vector3& C) const;

    //! Returns whether two planes are identical within a certain threshold
    bool samePlane(const Plane& p1, const Plane& p2);

    //! Returns the distance from a point to a plane
    float pointPlaneDist(const Plane& plane, Vector3 P) const;

    //! Returns whether a plane is a potential occluder
    bool canPlaneOcclude(Plane& plane, CSpace coordinateSpace = CSpace::Local);

    //! Returns whether a plane is a potential occluder for another mesh
    bool canPlaneOccludeOther(const Plane& plane, const Mesh* other);

    //! Rebuilds the Surface structure and sets the canOcclude flags
    /*!
    * \param coordinateSpace    use local or global coordinate system (default: global)
    * \return                   number of surfaces
    */
    unsigned int rebuildSurfaceStructure(CSpace coordinateSpace = CSpace::Global);

    //! Copies a given list of surfaces and sets the canOcclude flags
    /*!
    * \param _surfaces          given list of surfaces
    * \param coordinateSpace    use local or global coordinate system (default: global)
    * \return                   number of surfaces
    */
    unsigned int rebuildSurfaceStructureExtern(const std::vector<Mesh::Surface>& _surfaces,
        bool replaceDist = false, CSpace coordinateSpace = CSpace::Global);

    //! rebuild the data structures for round edges and corners
    void rebuildRoundStructure();

    //! update internal plane definitions
    void updatePlanes();

    //! Updates the canOccludeOthers flags
    /*!
    * \param other              check if we potentiall occlude this other mesh
    */
    void updateOccludeOthers(const Mesh* other);

    //! reset all canOccludeOthers flags
    void resetOccludeOthers();

    //! Determines whether a point is inside or outside. This function is only permitted to be
    //  executed after the lock function has been invoked.
    /*!
    */
    bool isInside(const EAR::Vector3& point) const override;

    //! get the shortest distance from this mesh to a given point
    Vector3 closestPointTo(const EAR::Vector3& point) const override;

    //! Inherited from geometry
    Type getType() const override { return Type::Mesh; };

    //! Inherited from Geometry
    float getArea() const override;

    //! Inherited from Geometry
    float getVolume() const override;

    //! Returns whether this mesh requires an occlusion test
    bool getOcclusionFlag(void) const { return needsOcclusionTest; }

    //! After invocation, geometry is not allowed to be changed.
    /*!
     * This function sets the lock variable and rebuilds some internal structures based on the state of the vertices/faces.
     */
    void lock();

    //! After invocation, geometry is allowed to be changed again.
    /*!
     * This function unsets the lock variable.
     */
    void unlock();

    //! Inherited from Geometry
    const Mesh* getMesh() const override;

    //! Inherited from Geometry
    float greatestDistance() const override;

    //! Inherited from Geometry
    void getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
        EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeftm,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    void getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    BoundingBox getBoundingBox() const override;

    BoundingBox getGlobalBoundingBox() const override;

    EAR::Vector3 getCenterPosition() const override;

private:
    //! Inherited from SceneObject
    void parentMoved(bool notifyObs) override;

    //! Generate subsurface information within given Surface (this is only called from rebuildSurfaceStructure)
    void getSubSurfaces(Surface& surf);

    //! Returns whether two SubSurfaces are connected
    bool connectedSubSurfs(
        std::vector<unsigned int> sub_surface_i, std::vector<unsigned int> sub_surface_j);

    //! Returns whether two triangles are connected
    bool connectedTris(Triangle& Tri_i, Triangle& Tri_j);

    void rebuildGlobalVertices();
    Vector3 local2global(const Vector3& v);
    bool isInsideImpl(const EAR::Vector3& point, const EAR::Vector3& rayVector) const;

    std::vector<Vector3> localVertices;
    std::vector<Vector3>
        globalVertices; //!< are transformed according to the position and orientation of the whole Mesh
    std::vector<Triangle> triangles;
    std::vector<Surface> surfaces;
    std::unique_ptr<CartesianKDTree> lookupTree;
    std::vector<RoundEdge> roundEdges;
    std::vector<RoundCorner> roundCorners;

    bool needsOcclusionTest = false;
    bool isLocked = false;

    std::unique_ptr<EmbreeTracer> rt = NULL;

    friend std::ostream& operator<<(std::ostream& os, Mesh& m);
};

std::ostream& operator<<(std::ostream& os, Mesh& m);

std::ostream& operator<<(std::ostream& os, Mesh::RoundEdge& e);

std::ostream& operator<<(std::ostream& os, Mesh::RoundCorner& c);

} /* namespace EAR*/

#endif /* EAR_GEOMETRY_MESH */
