/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_OSC
#define EAR_OSC

#include <EAR/Variant.h>

#include <atomic>
#include <chrono>
#include <functional>
#include <memory>
#include <mutex>
#include <thread>
#include <vector>

// fwd declares
class IpEndpointName;
class UdpTransmitSocket;
class UdpListeningReceiveSocket;

namespace osc {
class OutboundPacketStream;
class OscPacketListener;
class ReceivedBundle;
class ReceivedMessage;
class ReceivedPacket;
} //namespace osc

namespace EAR {

//! A class to receive/create/handle OSC messages and bundles.
/*!
 *  Outgoing messages/bundles are sent as unicast UDP packets to the IP address and port given on construction.
 *  At the same time, the class allows listening to incoming OSC messages/bundles on the given port.
 *  Incoming messages can be recorded and exported in the OSC::Message and OSC::Bundle format.
 *  This class will stay connected as long as its own lifetime.
 */
class OSC
{

private:
    //! A class that receives all incoming osc packets and calls the appropriate functions to handle them
    class PacketListener;

public:
    //! A class representing an OSC message to be sent or received
    class Message
    {

    public:
        //!Constructor
        /*
         * \param oscAddress    create a message that will be sent on this osc address; convention is to use forward slashes to seperate the address e.g "/this/is/an/osc/address"
         */
        Message(const std::string& oscAddress);

        //!Copy constructor
        Message(const Message& message);

        //!Destructor
        ~Message();

        //! Add an argument to this message
        /*!
         *  If an argument is already created then it can be added by calling this function
         *  The OSC server will receieve the arguments in the order that you call these functions
         * \param argument      add an already created argument
         * \return              a reference to this message, this allows chaining e.g. message.addArgument(true).addArgument(3);
         */
        Message& addArgument(Variant argument);

        //! Get the osc address that this message will be sent on
        /*!
         * \return              a string reference to the osc address that was given on construction
         */
        const std::string& getOscAddress() const;

        //! Get all of the currently added arguments in the message in the order that they will be sent
        /*!
         * \return              a vector of all arguments within the message
         */
        const std::vector<Variant> getArguments() const;

        //! Allows for iteration over arguments in this message using range-based for loops e.g. for (auto argument : message) { argument.getBoolValue(); }
        std::vector<Variant>::const_iterator begin() const;

        //! Allows for iteration over arguments in this message using range-based for loops e.g. for (auto argument : message) { argument.getBoolValue(); }
        std::vector<Variant>::const_iterator end() const;

        //! Allows for iteration over arguments in this message using range-based for loops e.g. for (auto argument : message) { argument.getBoolValue(); }
        std::vector<Variant>::const_iterator cbegin() const;

        //! Allows for iteration over arguments in this message using range-based for loops e.g. for (auto argument : message) { argument.getBoolValue(); }
        std::vector<Variant>::const_iterator cend() const;

        //! Return a string representation of message with all arguments
        std::string serialize() const;

        //! Set the message according to a serialized representation
        /*!
         * \param val       string in format as received from Message::serialize()
         * \return          a new OSC::Message according to the serialized representation
         */
        static OSC::Message deserialize(const std::string& val);

    private:
        friend OSC;

        //! Container to hold all of the arguments that will be sent with this message
        std::vector<Variant> arguments;

        //! Byte count of all arguments needed to create a packet with the correct size
        int byteSizeOfArguments;

        //! OscAddress that this message will be sent on
        std::string address;
    };

    //! A class representing an OSC bundle to be sent or received
    /*!
     *  An OSC bundle can have two elements within it, a message or another nested bundle
     */
    class Bundle
    {
    public:
        //! A class representing an OSC element; either an OSCMessage or another OSCBundle
        /*!
        *   A public constructor is not available for the element, please use a bundle and call the addElement() function
        */
        class Element
        {
        public:
            //! Constructor
            /*
             *  Create an element within a bundle that contains a message
             */
            Element(const Message& message);

            //! Constructor
            /*
             *  Create an element within a bundle that contains a message
             */
            Element(const Bundle& bundle);

            //! Checks if this element is a bundle
            /*
             * \return      true if the element is a bundle, false if the element is a message
             */
            bool isBundle() const;

            //! Checks if this element is a message
            /*
             * \return      true if the element is a message, false if the element is a bundle
             */
            bool isMessage() const;

            //! Get the message within this element
            /*
             * Warning! This function does not perform checks in a release build!
             * Verify that this element is a message, using either isBundle/Message(), before calling this function
             * \return      message within this element
             */
            const Message& getMessage() const;

            //! Get the bundle within this element
            /*
             * Warning! This function does not perform checks in a release build!
             * Verify that this element is a message, using either isBundle/Message(), before calling this function
             * \return      bundle within this element
             */
            const Bundle& getBundle() const;

        private:
            friend OSC;

            std::shared_ptr<Message> message;
            std::shared_ptr<Bundle> bundle;
        };

        //TODO: Add functionality for time tags in bundles
        //Bundle(specifiedTime) { timeTag = constructTimeTag(specifiedTime)};

        //! Add an element to the bundle
        /*
         *  The OSC server will receieve the elements in the order that you call these functions
         * \param element       a message that was created beforehand
         */
        Bundle& addElement(const Message& element);

        //! Add an element to the bundle
        /*
         *  The OSC server will receieve the elements in the order that you call these functions
         *  Note: it is possible to have nested bundles
         * \param element       a bundle that was created beforehand
         */
        Bundle& addElement(const Bundle& element);

        //! Get all of the elements currently in this bundle
        /*
         * \return          vector of elements within the bundle
         */
        const std::vector<Bundle::Element>& getElements() const;

        //! Allows for iteration over elements in this bundle using range-based for loops e.g. for (auto element : bundle) { element.getMessage(); }
        std::vector<Bundle::Element>::const_iterator begin() const;

        //! Allows for iteration over elements in this bundle using range-based for loops e.g. for (auto element : bundle) { element.getMessage(); }
        std::vector<Bundle::Element>::const_iterator end() const;

        //! Allows for iteration over elements in this bundle using range-based for loops e.g. for (auto element : bundle) { element.getMessage(); }
        std::vector<Bundle::Element>::const_iterator cbegin() const;

        //! Allows for iteration over elements in this bundle using range-based for loops e.g. for (auto element : bundle) { element.getMessage(); }
        std::vector<Bundle::Element>::const_iterator cend() const;

        //! Return a string representation of Bundle with all arguments
        std::string serialize() const;

        //! Create a Bundle according to a serialized representation
        /*!
         * \param val       string in format as received from Bundle::serialize()
         * \return          a new OSC::Bundle according to the serialized representation
         */
        static OSC::Bundle deserialize(const std::string& val);

    private:
        friend OSC;

        //! Container to hold all elements that will be sent with this message
        std::vector<Bundle::Element> elements;

        //! An OSC timetag
        uint64_t timetag = 0;
    };

    //! Constructor
    /*
     *  This class will listen and send on the port and address provided in this constructor
     * \param ipAddress             an ipAddress to listen and send OSC packets on, use the form "192.168.1.1"
     * \param port                  a UDP port to listen and send OSC packets on, please supply a valid UDP port and avoid any port below 1024
     * \param defaultPacketSize     the size of the packet in bytes, the packet should resize by itself but if you are using large packets you can set the size here
     * \param disableListening      set to true to avoid occupying the port to listen on it, i.e. only sending is possible (defaults to false)
     * \param autoConnect           set to false to not automatically block the port in the constructor (manual call to connect() is needed then, default true)
     */
    OSC(std::string ipAddress = "127.0.0.1", int port = 9001, int defaultPacketSize = 1024,
        bool disableListening = false, bool autoConnect = true);

    //! Destructor
    ~OSC();

    //! Send an OSC bundle on the given address and port in this class
    /*
     * \param      a preconstructed bundle
     */
    void send(const Bundle& oscBundle);

    //! Send an OSC message on the given address and port in this class
    /*
     * \param oscMessage            a preconstructed message
     */
    void send(const Message& oscMessage);

    //TODO: should a function be added to listen to bundles?
    //! Add a listener with a callback
    /*
     * The given callback will be executed if the message address matches the address given
     * \param oscAddress            execute the callback with message addresses matching this one
     * \param callback              supply a lambda that will be called when the addresses match
     */
    void addListener(std::string oscAddress, std::function<void(const Message& message)> callback);

    //! Begin recording all OSC messages and bundles receieved
    /*
     *  Sent messages and bundles are not recorded.
     *  Calling this function will clear any previous recordings
     */
    void startRecordingReceivedPackets();

    //! Stop recording all OSC messages and bundles receieved
    /*
     *  Only call this function if startRecordingReceiedPackets() was called beforehand
     */
    void stopRecordingReceivedPackets();

    //! Return a serialized string of the recorded OSC elements
    /*!
     * This cannot be called while a recording is running and will throw an exception.
     */
    std::string serializeRecording() const;

    //! Return a vector of deserialized recorded OSC elements
    /*!
     * The first element in the std::pair is the receive time, the second element is the actual Element (Message or Bundle)
     *
     * \param val   the complete string as returned by OSC::serializeRecording()
     */
    static std::vector<std::pair<double, OSC::Bundle::Element>> deserializeRecording(
        const std::string& val);

    //! Get all recorded OSC messages and bundles
    /*
     * This cannot be called while a recording is running and will throw an exception.
     * The returned vector reference is invalidated if a new recording is started. 
     *
     * \return      a float value in seconds at which the given element was recorded at
     */
    const std::vector<std::pair<float, Bundle::Element>>& getRecordedOscElements() const;

    //! Change the port to send and receive messages on
    /*
     * \param port          a valid UPD port, please avoid using UDP ports below 1024
     * \param autoConnect   set to false to not automatically block the port in the constructor (manual call to connect() is needed then, default true)
     */
    void changePort(int port, bool autoConnect = true);

    //! Change the ipAddress to send and receive message on
    /*
     * \param ipAddress     a valid ipAddress, use the form "127.0.0.1"
     * \param autoConnect   set to false to not automatically block the port in the constructor (manual call to connect() is needed then, default true)
     */
    void changeIpAddress(std::string ipAddress, bool autoConnect = true);

    //! Connect the sockets to the configured ipAddress and port
    void connect();

private:
    typedef std::function<void(const Message& message)> Callback;
    typedef std::pair<std::string, Callback> CallbackWithAddress;

    //! Receive a message and call all callbacks with addresses that the address within this function
    /*
    * \param message               a message passed from the packet listener
    * \param isRecursiveCall       if this function is called recursively, then it does not record that element whilst recording
    */
    void receive(const osc::ReceivedMessage& message, bool isRecursiveCall = false);

    //! Receive a bundle and process all elements within it
    /*
    * \param bundle                 a bundle passed from the packet listener
    * \param isRecursiveCall        if this function is called recursively, then it does not record that element whilst recording
    */
    void receive(const osc::ReceivedBundle& bundle, bool isRecursiveCall = false);

    //! Finalise a bundle and stream all of it's elements into the packet before it is sent
    void streamBundleIntoPacket(const Bundle& bundle);

    //! Finalise a message and stream all of it's arguments into the packet before it is sent
    void streamMessageIntoPacket(const Message& message);

    //! Construct an EAR OSC message from an OSCPack OSC message
    Message getMessageFromReceivedMessage(const osc::ReceivedMessage& receivedMessage);

    //! Construct an EAR OSC bundle from an OSCPack OSC bundle
    Bundle getBundleFromReceivedBundle(const osc::ReceivedBundle& receivedBundle);

    //! Sometimes messages can be bigger than the packet; call this function in this case
    void resizePacketToAccomodateMessage(const Message& message);

    //! All listener callbacks to be executed
    std::vector<CallbackWithAddress> listenerCallbacks;

    //! All outgoing network data is transmitted through this socket
    std::unique_ptr<UdpTransmitSocket> transmitSocket;

    //! All incoming network data is received through this socket
    std::unique_ptr<UdpListeningReceiveSocket> receiveSocket;

    //! A default size for the packet; it is arbitrary and is resized if, in the rare case, a message/bundle is bigger than the value stated
    int packetSize;

    //! A packet buffer thats given to the packet
    std::unique_ptr<char[]> packetBuffer;

    //! All data from the messages and bundles is streamed into this packet before the packet is then sent
    std::unique_ptr<osc::OutboundPacketStream> packet;

    //! The thread that runs the listener function required to receive osc packets
    std::unique_ptr<std::thread> listenerThread;

    //! An instance of the packetListener
    std::unique_ptr<PacketListener> listener;

    //! The current IP Address that is being used
    std::string ipAddress;

    //! The UPD port that is being used; avoid using ports below 1024
    int port;

    //! True if currently recording
    bool recording;

    //! Only send
    bool disableListening;

    //! Timestamp is set when recording starts; used to create time stamps
    std::chrono::time_point<std::chrono::system_clock> recordingStartTime;

    //! All recorded osc elements while recording is true, will be cleared if startRecordingReceivedPackets() is called.
    std::vector<std::pair<float, Bundle::Element>> recordedOscElements;

    //! Mutex for accessing receive-related variables
    std::mutex receiveMutex;
};

} //namespace EAR

#endif /* EAR_OSC */
