/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_PCHIP_INTERPOLATOR
#define EAR_PCHIP_INTERPOLATOR

#include <EAR/Interpolator.h>

#include <vector>

namespace EAR {

//! Piecewise Cubic Hermite Interpolating Polynomial
/*!
 * Creates a cubic interpolant for every section between two supporting points allowing evaluation at every value. Extrapolates linearly.
 * PCHIP produces good result for interpolating spectra, as there is no overshooting, etc. as with other polynomial interpolations.
 *
 * Ref: F.N. Fritsch, R.E. Carlson: Monotone Piecewise Cubic Interpolation, SIAM J. Numer. Anal. Vol 17, No. 2, April 1980
 */
// TODO: Make generic x and y types
class PchipInterpolator : public Interpolator<float, float>
{

public:
    //! Constructor without initialization (calls to interpolate naturally only make sense after initialization)
    /*!
     * \param bindBoundaryValues_       determines whether linearly (false) or fixed
                                        value (true) extrapolation is used
     */
    PchipInterpolator(bool bindBoundaryValues_);

    //! Constructor with initialization
    /*!
     * \param x                       x values of the supporting points (length n)
     * \param y                       y values of the supporting points (length n)
     * \param n                       number of supporting points
     * \param bindBoundaryValues_     determines whether linearly (false) or
                                      fixed value (true) extrapolation is used
     */
    PchipInterpolator(const float* x, const float* y, std::size_t n, bool bindBoundaryValues_);

    //! Provide supporting points to generate the mapping
    /*!
     * \param x       x values of the supporting points (length n)
     * \param y       y values of the supporting points (length n)
     * \param n       number of supporting points
     */
    void initialize(const float* x, const float* y, std::size_t n) override;

    //! Insert a new supporting point and reinitialize if required
    /*!
     * \param x      single new x value of supporting point
     * \param y      single new y value of supporting point
     */
    void addSupportingPoint(float x, float y) override;

    //! Provide interpolated y value for a single x
    /*!
     * \param x       x value at which the polynomials shall be evaluated
     */
    float interpolate(float x) override;

private:
    //! find section in which x lies
    std::size_t findSection(float x) const;

    //! return the bigger of the abs values of a and b
    float absmax(float a, float b) const;

    //! return the smaller of the abs values of a and b
    float absmin(float a, float b) const;

    //! compare signs of a and b: -1 -> sign is different; 0 -> a or b (or both) are 0; 1 -> sign is the same
    int signTest(float a, float b) const;

    //! vector of supporting x values
    std::vector<float> xs;

    //! vector of supporting y values
    std::vector<float> ys;

    //! vector of slopes in supporting points (1st derivative)
    std::vector<float> dy;

    //determines whether linearly or fixed value extrapolation is used
    bool bindBoundaryValues;
};

} // namespace EAR

#endif /* EAR_PCHIP_INTERPOLATOR */
