/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_PCHIP_MINPHASE_FILTER_DESIGNER
#define EAR_PCHIP_MINPHASE_FILTER_DESIGNER

#include <EAR/FilterDesigner.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/FFT.h>
#include <EAR/PchipInterpolator.h>
#include <EAR/RealBuffer.h>

#include <cstddef>
#include <memory>
#include <vector>

namespace EAR {

//! Filter designer to interpolate a magnitude spectrum with PCHIP and generate a minimum phase FIR
/*!
 * The minimum phase approach allows for zero-latency filtering and is achieved with a windowing in Cepstrum domain.
 * Ref [1]: A. V. Oppenheim and R. W. Schafer, 'Discrete-time signal processing', pp. 781-787
 */
class PchipMinphaseFilterDesigner : public FilterDesigner
{
public:
    //! Constructor
    /*!
     * \param f     float pointer to array of length H containing the frequency values at which the H magnitudes describe the desired frequency response
     * \param H     number of supporting magnitude values
     * \param N     length of the resulting FIR filter in taps
     * \param fs    sample rate in Hz
     */
    PchipMinphaseFilterDesigner(const float* f, std::size_t H, std::size_t N, float fs);

    //! Constructor with default third-octave magnitude description
    /*!
     * \param N     length of the resulting FIR filter in taps
     * \param fs    sample rate in Hz
     */
    PchipMinphaseFilterDesigner(std::size_t N, float fs);

    //! Destructor
    ~PchipMinphaseFilterDesigner();

    //! design a FIR filter with a given magnitude response
    /*!
     * \param m       float pointer to array of length H containing the desired magnitude values at the previously defined frequencies
     * \param result  reference to RealBuffer where the resulting filter is written
     *                (Caution: assumed to already have the correct size (N) to avoid the expensive resizing in time-critical situations)
     */
    void generate(const float* m, RealBuffer& result) override;

private:
    //! length of corresponding real spectrum for filter length N
    std::size_t K;
    //! bin frequencies of real spectrum for filter length N
    std::vector<float> binf;
    //! Interpolator to generate full mag spectrum from band magnitudes
    std::unique_ptr<PchipInterpolator> pchip;

    //! temp buffer for RFFT operations
    ComplexBuffer temp;
    //! temp buffer for log operations
    RealBuffer logSignal;
    //! buffer for interpolated spectrum domain values
    RealBuffer spectrum;
    //! buffer for cepstrum domain values
    RealBuffer cepstrum;
    //! window to be applied to the cepstrum
    RealBuffer cepstrumWin;
    //! real to complex FFT processor
    std::unique_ptr<RFFT> fft;
    //! complex to real IFFT processor
    std::unique_ptr<IRFFT> ifft;

    static const std::vector<float> defaultCenterFrequencies;
};

} // namespace EAR

#endif /* EAR_PCHIP_MINPHASE_FILTER_DESIGNER */
