/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#ifndef EAR_PORTAL
#define EAR_PORTAL

#include <EAR/BaseTypes.h>
#include <EAR/PortalReverbExtent.h>
#include <EAR/SceneObject.h>

#include <unordered_map>

namespace EAR {

// Portal class which is part of the SceneState so that it may receive updates.
class Portal : public SceneObject
{
public:
    //! Constructor
    Portal(ObjectID oid, std::string name = "", Position pos = Position())
        : SceneObject(oid, name, pos)
    {
        staticFlag.store(true); // Portals don't move
    };

    Portal(ObjectID oid, std::string name, Position pos, const int portalId,
        std::shared_ptr<EAR::Material> material,
        const std::array<std::shared_ptr<PortalReverbExtent>, 2>& externalReverbExtents)
        : SceneObject(oid, name, pos)
        , portalId(portalId)
        , material(material)
        , externalReverbExtents(externalReverbExtents)
    {
        staticFlag.store(true); // Portals don't move
    };

    void setPortalId(int portalId) { this->portalId = portalId; }

    int getPortalId() { return portalId; }

    void setMaterial(std::shared_ptr<EAR::Material> material)
    {
        this->material = material;
        notify(EAR::SceneObjectObserver::Property::Material);
    }

    inline void setExtentOffset(const std::array<float, 2>& offset) { this->extentOffset = offset; }
    inline std::array<float, 2> getExtentOffset() { return this->extentOffset; }

    inline void setPortalDims(EAR::Vector3 dimensionsWHD) { this->dimensions = dimensionsWHD; }
    inline EAR::Vector3 getPortalDims(void) { return this->dimensions; }

    std::shared_ptr<EAR::Material> getMaterial() { return material; }

    void setExternalReverbExtents(
        const std::array<std::shared_ptr<PortalReverbExtent>, 2>& externalReverbExtents)
    {
        this->externalReverbExtents = externalReverbExtents;
    }

    std::array<std::shared_ptr<PortalReverbExtent>, 2> getExternalReverbExtents()
    {
        return externalReverbExtents;
    }

private:
    //! Unique id of this portal
    int portalId;
    //! material associated with the ID for calculation of energy passing through the portal.
    std::shared_ptr<EAR::Material> material;
    // ! Extents used for reverb of external sources. Contains two extents, one for each Acoustic Environment the portal is connected to. The first extent corresponds to the first accoustic environment (i.e. MPEGI::Portal::portalEnvironments.first), and the second extent to the second accoustic environment.
    std::array<std::shared_ptr<PortalReverbExtent>, 2> externalReverbExtents;

    // The distance from the portal where the reverb extents are created
    std::array<float, 2> extentOffset = { 1.0f, 1.0f };

    // Dimensions of the portal: Width, Height, Depth
    EAR::Vector3 dimensions;
};

} // namespace EAR
#endif
