/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#ifndef EAR_GEOMETRY_PORTAL_EXTENT
#define EAR_GEOMETRY_PORTAL_EXTENT

#include <EAR/GeometryBase.h>

namespace EAR {

// fwd declare
class Mesh;

//! Portal Extent geometry as a mesh, derived from a portal box.
class PortalReverbExtent : public Geometry
{
private:
    //! Constructor for a portal extent geometry (only accessible through SceneObjectFactory)
    /*!
    * \param oid                    fixed object ID (always assigned by factory method!)
    * \param size                   the size of the portal box expressed as three elements (x, y, z / width, height, depth)
    * \param name                   arbitrary name of object
    * \param pos                    Position struct with initial location and orientation
    * \param segments               Number of segments to subdivide the curve (1,2 or 3)
    * \param offset                 Distance to offset the portal extent mesh from the portal box.
    * \param backFaceOffset         Thickness of the portal extent. In order to make the portal extent closed, it needs a backside. This backsize is placed at backFaceOffset distance from the front side.
    * \param constrainingGeometry   Extent geometry outside constrainingGeometry will be clipped to the closest point on the constrainingGeometry. Used to constrain portal extent to be inside the room. No constraints used when nullptr.
    */

    PortalReverbExtent(ObjectID oid, EAR::Vector3 size, float& revExtOffset,
        const std::string& name = "", Position pos = Position(), int segments = 1,
        float backFaceOffset = 0.0f, const Geometry* constrainingGeometry = nullptr);

    friend SceneObjectFactory<PortalReverbExtent>;

public:
    //! Inherited from Geometry
    const Mesh* getMesh() const override;

    //! Inherited from Geometry
    bool isInside(const EAR::Vector3& point) const override;

    //! Inherited from Geometry
    Vector3 closestPointTo(const EAR::Vector3& point) const override;

    //! Inherited from geometry
    Type getType() const override { return Type::Box; };

    //! Inherited from Geometry
    float getVolume() const override;

    //! Inherited from Geometry
    float getArea() const override;

    //! get the dimensions of the portal mesh
    Vector3 getSize() const;

    //! Inherited from SceneObject
    void setOrientation(Euler newOrient, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setLocation(Vector3 newLoc, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setPosition(
        Position newPos, SceneObject* parent = nullptr, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setStatic(bool newState) override;

    //! Inherited from Geometry
    float greatestDistance() const override;

    //! Inherited from Geometry
    void getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
        EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    void getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    BoundingBox getBoundingBox() const override;

    BoundingBox getGlobalBoundingBox() const override;

    EAR::Vector3 getCenterPosition() const override;

private:
    //! The volume expressed as three elements (x, y, z)
    EAR::Vector3 volume;

    //! Mesh representation of the portal extent
    std::unique_ptr<Mesh> mesh;

    friend std::ostream& operator<<(std::ostream& os, PortalReverbExtent& b);
};

std::ostream& operator<<(std::ostream& os, PortalReverbExtent& b);

} /* namespace EAR*/

#endif /* EAR_GEOMETRY_PORTAL_EXTENT */
