/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_SAMPLE_AND_HOLD_INTERPOLATOR
#define EAR_SAMPLE_AND_HOLD_INTERPOLATOR

#include <EAR/Interpolator.h>

#include <algorithm>
#include <deque>
#include <vector>

namespace EAR {

//! Simple sample-and-hold interpolation that allows to dynamically add new supporting points
/*!
 * Supporting points are expected to be ordered by their x value.
 * Type for x only make sense if it is numerable with a zero-element (int, long, float, double, ...)
 * Type for y doesn't matter, because the value of the nearest supporting point with an x value smaller than the input is returned unchanged.
 */
template <typename Tx, typename Ty> class SampleAndHoldInterpolator : public Interpolator<Tx, Ty>
{
public:
    SampleAndHoldInterpolator() = default;
    SampleAndHoldInterpolator(const Tx* x, const Ty* y, std::size_t N) { initialize(x, y, N); }
    ~SampleAndHoldInterpolator() = default;

    void initialize(const Tx* x, const Ty* y, std::size_t N) override
    {
        xs.resize(N);
        std::copy(x, x + N, xs.data());
        ys.resize(N);
        std::copy(y, y + N, ys.data());
    }

    void addSupportingPoint(Tx x, Ty y) override
    {
        std::vector<Tx> newxs = xs;
        std::vector<Ty> newys = ys;

        // if x already exists in the sorted vector, replace the corresponding y
        if (std::binary_search(newxs.begin(), newxs.end(), x)) {
            int idx = std::find(newxs.begin(), newxs.end(), x) - newxs.begin();
            newys[idx] = y;
        } else { // else find new spot for x in newxs and insert into newxs and newys accordingly
            int idx = std::lower_bound(newxs.begin(), newxs.end(),
                          x)
                - newxs.begin(); // taking the detour over index to use it for newys too
            newxs.insert(newxs.begin() + idx, x);
            newys.insert(newys.begin() + idx, y);
        }

        initialize(newxs.data(), newys.data(), newxs.size());
    };

    Ty interpolate(Tx x) override
    {
        int idx = std::upper_bound(xs.begin(), xs.end(), x) - xs.begin();

        if (--idx < 0) {
            idx = 0;
        }

        return ys[idx];
    }

private:
    std::vector<Tx> xs;
    std::vector<Ty> ys;
};

// explicit specialization for Tx = double, Ty = bool
// vector<bool> is a special case that doesn't work for our purpose, so we replace it with std::deque
template <> // partial specialization only for Ty not allowed in C++
class SampleAndHoldInterpolator<double, bool> : public Interpolator<double, bool>
{
public:
    SampleAndHoldInterpolator() = default;
    SampleAndHoldInterpolator(const double* x, const bool* y, std::size_t N)
    {
        initialize(x, y, N);
    }
    ~SampleAndHoldInterpolator() = default;

    void initialize(const double* x, const bool* y, std::size_t N) override
    {
        xs.resize(N);
        std::copy(x, x + N, xs.data());
        ys.resize(N);

        for (auto i = 0; i < N; i++) {
            ys[i] = y[i];
        }
    }

    void addSupportingPoint(double x, bool y) override
    {
        std::vector<double> newxs = xs;
        std::deque<bool> newys = ys;

        // if x already exists in the sorted vector, replace the corresponding y
        if (std::binary_search(newxs.begin(), newxs.end(), x)) {
            int idx = std::find(newxs.begin(), newxs.end(), x) - newxs.begin();
            newys[idx] = y;
        } else { // else find new spot for x in newxs and insert into newxs and newys accordingly
            int idx = std::lower_bound(newxs.begin(), newxs.end(),
                          x)
                - newxs.begin(); // taking the detour over index to use it for newys too
            newxs.insert(newxs.begin() + idx, x);
            newys.insert(newys.begin() + idx, y);
        }

        xs.resize(newxs.size());
        std::copy(newxs.data(), newxs.data() + newxs.size(), xs.data());
        ys.resize(newxs.size());

        for (auto i = 0; i < newxs.size(); i++) {
            ys[i] = newys[i];
        }
    };

    bool interpolate(double x) override
    {
        int idx = std::upper_bound(xs.begin(), xs.end(), x) - xs.begin();

        if (--idx < 0) {
            idx = 0;
        }

        return ys[idx];
    }

private:
    std::vector<double> xs;
    std::deque<bool> ys;
};

} // namespace EAR

#endif /* EAR_SAMPLE_AND_HOLD_INTERPOLATOR */
