/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_SCENE_OBJECT
#define EAR_SCENE_OBJECT

#include <EAR/BaseTypes.h>
#include <EAR/Uncopyable.h>

#include <EAR/SceneObjectObserver.h>

#include <atomic>
#include <optional>
#include <string>
#include <vector>

enum class VDLMethod
{
    spline = 0,
    linear = 1,
    spline_erLinear = 2,
};

namespace EAR {

//! The type used for unique object IDs
typedef int ObjectID;

//! Base class for every object in an auditory scene
/*!
 * Provides object identification attributes and a full 6 DoF position with relative transformation.
 * See the BaseTypes Vector3 and Euler for information on coordinate axis conventions.
 * Instantiation shall be handled by the SceneObjectFactory.
 */
class SceneObject : Uncopyable
{

    // TODO: remove the rotation stuff in favour of the CoordinateSystem class

public:
    //! Constructor
    /*!
     * \param oid        fixed object ID (always assigned by factory method!)
     * \param name       arbitrary name of object
     * \param pos        Position struct with initial location and orientation
     */
    SceneObject(ObjectID oid, std::string name = "", Position pos = Position())
        : oid(oid)
        , name(name)
        , position(pos)
        , matrix(pos.location, pos.orientation)
        , bsId(std::nullopt){};

    //! Destructor
    virtual ~SceneObject(){};

    //! Register a new observer
    /*!
     * Make sure to call detach() (or to destruct the Scene) before deleting the object that is observing the Scene.
     *
     * \param observer        Pointer to a SceneObjectObserver instance that will be notified if the Scene changes
     */
    void attach(SceneObjectObserver* observer);

    //! Remove a registered observer
    /*!
     * \param observer        Pointer to a SceneObjectObserver instance to be removed from the list of observers. If this is not a valid observer, nothing happens.
     */
    void detach(SceneObjectObserver* observer);

    //! Return the unique ID of the object
    ObjectID getID() const;

    //! Return the bitstream ID of the object
    std::optional<int> getBsID() const;

    //! Return the name of the object
    std::string getName() const;

    //! Return the Position struct of the object in the current CSpace
    Position getPosition() const;

    //! Return the cartesian Vector3 location of the object in global coordinates
    /*!
     * Note that the CSpace still indicates LCS if this is set.
     */
    Position getGlobalPosition() const;

    //! Return the cartesian Vector3 location of the object in the current CSpace
    Vector3 getLocation() const;

    //! Return the cartesian Vector3 location of the object in global coordinates
    Vector3 getGlobalLocation() const;

    //! Return the intrinsic Euler orientation of the object in the current CSpace
    Euler getOrientation() const;

    //! Return the intrinsic Euler orientation of the object in global coordinates
    Euler getGlobalOrientation() const;

    //! Return the reference coordinate system of the SceneObject's position
    CSpace getCSpace() const;

    //! Return a pointer to the cspace parent (may be nullptr, e.g. if CSpace::WCS)
    const SceneObject* getParent() const;

    //! Return the 3D velocity of the object in the current CSpace
    Vector3 getVelocity() const;

    //! Return the 3D velocity of the object relative to the global coordinate system
    /*!
     * This includes the velocities of potential parent objects
     */
    Vector3 getGlobalVelocity() const;

    //! Return the cartesian Vector3 location of the given other location relative to this object's position (translated and rotated)
    /*!
     * \param other      other location
     */
    Vector3 getRelativeLocation(const Vector3& other) const;

    //! Return the timestamp of the last position update
    /*!
     * The value of this timestamp is entirely arbitrary and has to be updated manually
     * by the entity that modifies the position.
     */
    double getPositionTimestamp() const;

    //! Check if the SceneObject is active (thread-safe)
    /*!
     * If set to false, all processing shall behave as if the object would not exist.
     * \return          true if object is active
     */
    bool isActive() const;

    //! Change activity status of SceneObject (thread-safe)
    /*!
     * If set to false, all processing shall behave as if the object would not exist.
     * \param newState  true if object should be active (if newState == oldState, nothing happens)
     */
    virtual void setActive(bool newState);

    //! Check if the SceneObject is static or dynamic (i.e. allowed to move) (thread-safe)
    /*!
     * If set to true, various optimizations based on the static location in global coordinates are possible.
     * However, this property is NOT ENFORCED, so the position can still be changed with static = true.
     * \return          true if object is static (non-movable)
     */
    bool isStatic() const;

    //! Change whether SceneObject is static or dynamic (thread-safe)
    /*!
     * If set to true, various optimizations based on the static location in global coordinates are possible.
     * However, this property is NOT ENFORCED, so the position can still be changed with static = true.
     * Changes of this property cannot be monitored with a SceneObjectObserver.
     * \param newState  true if object should be static (if newState == oldState, nothing happens)
     */
    virtual void setStatic(bool newState);

    //! Check if the SceneObject is in listener-relative (local) coordinate space (thread-safe)
    /*!
     * false -> global coordinate space, true -> listener-relative coordinate space
     * This might be of no relevance for some SceneObjects (e.g. Listener).
     * \return          true if object is listener-relative
     */
    bool isListenerRelative() const;

    //! Change coordinate-space status of SceneObject (thread-safe)
    /*!
     * false -> global coordinate space, true -> listener-relative coordinate space
     * This might be of no relevance for some SceneObjects (e.g. Listener).
     * \param newState  true if object should be listener-relative (if newState == oldState, nothing happens)
     */
    void setListenerRelative(bool newState);

    //! Check if object has moved (thread-safe)
    /*!
     * This can be used by renderers to determine if something changed.
     */
    bool hasMoved() const;

    //! Check if object has moved larger distance (thread-safe)
    /*!
     * This can be used by renderers to determine if something changed beyond a larger threshold.
     */
    bool hasCoarseMoved() const;

    //! Check if object has rotated larger amount (thread-safe)
    /*!
     * This can be used by renderers to determine if something changed beyond a larger threshold.
     */
    bool hasCoarseRotated() const;

    //! Set object-moved to true (thread-safe)
    /*!
     * This can be used by renderers to indicate that an object was moved (e.g. for filter updates).
     */
    void setMoved();

    //! Set object-moved to false (thread-safe)
    /*!
     * This can be used by renderers to indicate that an object movement has been handled.
     */
    void resetMoved();

    //! Change the reference coordinate system
    /*!
     * A new parent has to be given, as a potential previous parent is not saved when the CSpace is changed.
     *
     * \param newCSpace  new coordinate space
     * \param parent     parent SceneObject for CSpace::RCS or LCS
     * \param notifyObs  boolean whether or not to notify observers of this SceneObject (default true)
     */
    virtual void setCSpace(CSpace newCSpace, SceneObject* parent = nullptr, bool notifyObs = true);

    //! Change the orientation of the object
    /*!
     * \param newOrient  new orientation
     * \param notifyObs  boolean whether or not to notify observers of this SceneObject (default true)
     */
    virtual void setOrientation(Euler newOrient, bool notifyObs = true);

    //! Change the location of the object
    /*!
     * \param newLoc     new location
     * \param notifyObs  boolean whether or not to notify observers of this SceneObject (default true)
     */
    virtual void setLocation(Vector3 newLoc, bool notifyObs = true);

    //! Change the position of the object
    /*!
     * \param newPos     new position
     * \param parent     parent SceneObject for CSpace::RCS or LCS
     * \param notifyObs  boolean whether or not to notify observers of this SceneObject (default true)
     */
    virtual void setPosition(Position newPos, SceneObject* parent = nullptr, bool notifyObs = true);

    //! Change the 3D velocity vector of the object
    /*!
     * \param newPos     new velocity
     * \param notifyObs  boolean whether or not to notify observers of this SceneObject (default true)
     */
    virtual void setVelocity(Vector3 newPos, bool notifyObs = true);

    //! Set the timestamp of the last position update
    /*!
     * The value of this timestamp is entirely arbitrary and has to be updated manually
     * by the entity that modifies the position.
     *
     * \param ts          new timestamp with arbitrary unit
     */
    void setPositionTimestamp(double ts);

    //! Get a HomogenMatrix instance that represents the rotation and translation of the current local coordinate system of the object (in relation to global coordinates)
    HomogenMatrix getLocalCSTransform() const;

    //! Get a HomogenMatrix instance that represents the rotation and translation from the current CS into global coordinate system
    HomogenMatrix getGlobalCSTransform() const;

    //! Returns 3 Vector3 instances that correspond to local coordinate system basis vectors expressed in global coordinates
    std::vector<Vector3> getBasis() const;

    //! notify observers
    virtual void notify(SceneObjectObserver::Property mod);

protected:
    //! Called by parent object in CSpace::RCS and LCS when the parent position changed
    virtual void parentMoved(bool notifyObs);

    //! Add a new child with CSpace::RCS or LCS
    void addChild(SceneObject* child);

    //! Remove a child with CSpace::RCS or LCS
    void removeChild(SceneObject* child);

    //! List of active observers
    std::vector<SceneObjectObserver*> observers;

    //! if set to false, all processing shall behave as if the object would not exist
    std::atomic<bool> active{ true };

    //! if set to false, the SceneObject is rendered according to global coordinate space, otherwise according to listener-relative coordinate space
    std::atomic<bool> listenerRelative{ false }; // TODO: deprecated

    //! indicator if the object position changed
    std::atomic<bool> moved{ false };

    //! indicator if the object position changed significantly
    std::atomic<bool> coarseMoved{ false };

    //! indicator if the object rotation changed significantly
    std::atomic<bool> coarseRotated{ false };

    //! indicator if the object is allowed to move in the scene
    std::atomic<bool> staticFlag{ false };

    //! unique ID
    const ObjectID oid;

    //! human readable name
    std::string name = "";

    //! 6DoF position described by location (cartesian) and orientation (polar)
    Position position;

    //! 6DoF position described by location (cartesian) and orientation (polar) since the last significant change
    Position oldPosition;

    //! Parent SceneObject for CSpace::LCS or CSpace::RCS
    SceneObject* parent = nullptr;

    //! Children of this SceneObject for CSpace::LCS or CSpace::RCS
    std::vector<SceneObject*> children; // TODO: thread safety?

    //! 3D Velocity vector (values in m/s)
    Vector3 velocity;

    //! Rotation and translation matrix
    HomogenMatrix matrix;

    //! Timestamp of last position update
    std::atomic<double> positionTimestamp;

    //! VDL method
    std::atomic<VDLMethod> vdlMethod;
    //! unique bitstream ID, if any
    std::optional<int> bsId;

    //! Reverb gain as factor
    std::atomic<float> itemReverbGain;

    //! noReverb as bool
    std::atomic<bool> noReverb;
};

} // namespace EAR

#endif /* EAR_SCENE_OBJECT */
