/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_SOFA_HRIR_PROVIDER
#define EAR_SOFA_HRIR_PROVIDER

#include <EAR/BaseTypes.h>
#include <EAR/IRProvider.h>
#include <EAR/RealBuffer.h>
#include <EAR/SofaReader.h>

#include <memory>
#include <string>

namespace EAR {

//! Read SOFA files following the "SimpleFreeFieldHRIR" convention (implements IRProvider interface)
/*!
 * Some checks are included upon opening the file to enforce this convention, but it actually just takes the IR data and the position information for now.
 * More complex conventions like MultiSpeakerBRIR are not currently supported and will lead to uncontrolled behaviour!
 */
class SofaHRIRProvider : public IRProvider
{

public:
    //! Constructor
    /*!
     * Throws EAR::Exception if file could not be opened or something goes wrong in constructing the search tree.
     *
     * \param path      absolute or relative path string to the SOFA file location
     */
    SofaHRIRProvider(const std::string& path);

    //! Destructor
    ~SofaHRIRProvider();

    //! Inherited from IRProvider
    void getNearestIRForLocation(RealFrame& ir, Vector3 loc) const override;

    //! Inherited from IRProvider
    std::unique_ptr<const float*[]> getIRPointerForLocation(Vector3 loc) const override;

    //! Inherited from IRProvider
    virtual std::unique_ptr<const float*[]> getIRPointerForIndex(int index) const override;

    //! Inherited from IRProvider
    virtual std::vector<int> getAllIndicesInRange(
        const EAR::Polar& bottomLeft, const EAR::Polar& topRight) const override;

    //! Inherited from IRProvider
    int irLength() const override;

    //! Inherited from IRProvider
    int irCount() const override;

    //! Inherited from IRProvider
    float getSampleRate() const override;

    //! Inherited from IRProvider
    Vector3 getNearestLocation(Vector3& loc) const override;

    //! Inherited from IRProvider
    Vector3 getLocationForIndex(int idx) const override;

    //! Returns the distance of the nearest IR to the given location
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     * \return          distance to closest match [meters]
     */
    float getDistanceToNearestIndex(Vector3& loc) const;

private:
    SofaReader<float, RealBuffer> reader;
    float fs;
};

} // namespace EAR

#endif /* EAR_SOFA_HRIR_PROVIDER */
