/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_SOFA_READER
#define EAR_SOFA_READER

#include <EAR/BaseTypes.h>
#include <EAR/Frame.h>

#include <cstddef>
#include <memory>
#include <mutex>
#include <string>

#ifndef DOXYGEN_SHOULD_SKIP_THIS

// forward declaration for private includes
namespace HighFive {
class File;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

namespace EAR {

// fwd declares
class RealBuffer;
class CartesianKDTree;
class ComplexBuffer;

//! Read SOFA files.
/*!
 * The two template arguments specify, which data type can be expected from the SOFA file. Allowed combinations are:
 * 1. SofaReader<float, RealBuffer>
 * 2. SofaReader<std::complex<float>, ComplexBuffer>
 */
template <typename DataT = float, class BufferT = RealBuffer> class SofaReader
{

public:
    //! List of all supported Sofa conventions.
    /*enum class Convention {
        SimpleFreeFieldHRIR = 0,
        GeneralTF
    };*/

    //! Constructor
    /*!
     * Throws EAR::Exception if file could not be opened or something goes wrong in constructing the search tree.
     *
     * \param path        absolute or relative path string to the SOFA file location
     * \param convention  the expected convention (Exception is thrown, when SOFA file does not match this)
     */
    SofaReader(const std::string& path); //, Convention convention);

    //! Destructor
    ~SofaReader();

    //! Return the convention that is used in the opened file (see enum Convention for supported types)
    // Convention getConvention();

    //! Look up the directional values that corresponds to the given cartesian source position
    /*!
     * \param data      A frame where the values should be written to (needs to have the right size and type (real or complex) for the requested values dimensionality.
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     */
    void getNearestValueForLocation(Frame<DataT, BufferT>& data, const Vector3& loc) const;

    //! Look up the directional filter that corresponds to the given cartesian source position and return actual cache pointer without any copying.
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     * \return          unique pointer to the actual HRIR cache position
     */
    std::unique_ptr<const DataT*[]> getValuePointerForLocation(const Vector3& loc) const;

    //! Look up the directional values that corresponds to the given cartesian source position
    /*!
     * \param data      A frame where the values should be written to (needs to have the right size and type (real or complex) for the requested values dimensionality.
     * \param idx       The cache index as given by getNearestIndex (not sanity checked, so don't increase or decrease this yourself!)
     */
    void getValueForIndex(Frame<DataT, BufferT>& data, unsigned long idx) const;

    //! Return actual cache pointer without any copying.
    /*!
     * \param idx       The cache index as given by getNearestIndex (not sanity checked, so don't increase or decrease this yourself!)
     * \return          unique pointer to the actual HRIR cache position (type depends on the convention (real or complex))
     */
    std::unique_ptr<const DataT*[]> getValuePointerForIndex(unsigned long idx) const;

    //! Returns the size of a given dimension
    /*!
     * \param dim       The dimension name
     */
    int getDimension(std::string dim) const;

    //! Returns a vector with the center frequencies of the data
    const std::vector<float>& getCenterFreqs() const;

    //! Returns the value corresponding value to a given meta data identifier. Templated to the type of metadata to be read.
    /*!
     * \param name      The meta data identifier
     * \param data      Pointer to memory of templated type, where the meta data should be read to. Make sure to provide enough space there!
     */
    template <typename T> void getMetaData(std::string name, T* data) const;

    //! Returns the location of the nearest available value to the given location
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     * \return          Vector3 location of the closest measured point
     */
    Vector3 getNearestLocation(const Vector3& loc) const;

    //! Returns the location of a hrir point using its index
    /*!
     * \param idx       the index of the hrir point
     * \return          its location
     */
    Vector3 getLocationForIndex(int idx) const;

    //! Returns the distance of the nearest available value to the given location
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     * \return          distance to closest match [meters]
     */
    float getDistanceToNearestIndex(const Vector3& loc) const;

    //! Returns the pointer to the raw cache buffer
    BufferT* getCacheBuffer();

    //! Returns the pointer to the raw directions buffer
    const float* getDirectionsBuffer() const;

    //! Returns the index of the nearest IR to the given location
    /*!
     * \param loc       A Vector3 point with the source location. This point should be in a coordinate system relative to the listener.
     */
    unsigned long getNearestIndex(const Vector3& loc) const;

private:
    //! load the actual filter contents into the cache (implementation depends on template parameters)
    void fillCache();

    unsigned long N, M, R; //!< dimension metadata

    std::unique_ptr<HighFive::File> file;

    std::vector<Vector3> sourcePoints;
    std::unique_ptr<CartesianKDTree> lookupTree;

    std::unique_ptr<float[]> directions;

    std::unique_ptr<BufferT> cache;

    std::vector<float> centerFreqs;

    // unfortunately, the HDF5 API is not thread-safe so we ensure sequential use manually
    static std::mutex hdf5Mutex;

    // Convention convention; //!< convention according to SofaReader::Convention enum
};

} // namespace EAR

#endif /* EAR_SOFA_READER */
