/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_GEOMETRY_SPHERE
#define EAR_GEOMETRY_SPHERE

#include <EAR/GeometryBase.h>

namespace EAR {

// fwd declare
class Mesh;

//! Sphere primitive geometry, defined only by a radius (TODO: implement Ellipsoid)
class Sphere : public Geometry
{
private:
    //! Constructor for a sphere primitive (only accessible through SceneObjectFactory)
    /*!
    * \param oid            fixed object ID (always assigned by factory method!)
    * \param radius         the distance from the centre of the sphere to the face
    * \param name           arbitrary name of object
    * \param pos            Position struct with initial location and orientation
    * \param subdivisions   Number of iterations for sub-dividing the icosahedron mesh
    */
    Sphere(ObjectID oid, float radius, const std::string& name = "", Position pos = Position(),
        unsigned int subdivisions = 0);

    friend SceneObjectFactory<Sphere>;

public:
    //! Sets the material for this object
    /*!
    * \param material  the new material of this primitive
    */
    void setMaterial(std::shared_ptr<Material> material);

    //! Returns the material for this primitive
    std::shared_ptr<Material> getMaterial() const;

    //! Inherited from Geometry
    bool isInside(const EAR::Vector3& point) const override;

    //! Inherited from Geometry
    Vector3 closestPointTo(const EAR::Vector3& point) const override;

    //! Inherited from geometry
    Type getType() const override { return Type::Sphere; };

    //! Inherited from Geometry
    float getVolume() const override;

    //! Inherited from Geometry
    float getArea() const override;

    //! Return the radius of the Sphere
    inline float getRadius() const { return radius; };

    //! Inherited from Geometry
    const Mesh* getMesh() const override;

    //! Inherited from SceneObject
    void setOrientation(Euler newOrient, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setLocation(Vector3 newLoc, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setPosition(
        Position newPos, SceneObject* parent = nullptr, bool notifyObs = true) override;

    //! Inherited from SceneObject
    void setStatic(bool newState) override;

    //! Inherited from Geometry
    float greatestDistance() const override;

    //! Inherited from Geometry
    void getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
        EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    void getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
        bool& povIsInside, EAR::Vector3& closestPoint) const override;

    //! Inherited from Geometry
    BoundingBox getBoundingBox() const override;

    //! Inherited from Geometry
    BoundingBox getGlobalBoundingBox() const override;

    EAR::Vector3 getCenterPosition() const override;

private:
    //! The radius of the sphere
    float radius;

    //! The material that this sphere is made of
    std::shared_ptr<Material> material;

    //! Mesh representation of the sphere
    std::unique_ptr<Mesh> mesh;

    friend std::ostream& operator<<(std::ostream& os, Sphere& s);
};

std::ostream& operator<<(std::ostream& os, Sphere& s);

} /* namespace EAR*/

#endif /* EAR_GEOMETRY_SPHERE */
