/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_VDL
#define EAR_VDL

#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <cstddef>
#include <map>
#include <memory>
#include <vector>

namespace EAR {

class IIRFilter;
class FIRFilter;

//! Variable delay-line
/**
 * Delay a signal by arbitrary fractional number of samples. For this, interpolation is needed to calculate the intersample values.
 * Currently, VDL offers high and low quality versions of the Lagrange interpolation (verbose) and the Lagrange interpolation using the
 * farrow structure. Additionally, the VDL provides spline interpolation using basis functions.
 * Also, the 2x upsampling can be turned off. This results in a maximum relative velocity of the sound source of ~86 m/s (= ~300 km/h), before aliasing effects may appear.
 * Under all circumstances, the anti-aliasing low-pass filters have a cutoff frequency of 18 kHz.
 */

class VDL
{
public:
    //! Setup flag: Farrow interpolation with continuous intersample interpolation based on Lagrange polynomials (default behaviour)
    static const int InterpolationFarrow_Lagrange = 0b00000;

    //! Setup flag: Farrow interpolation with continuous intersample interpolation based on basis fuctions
    static const int InterpolationFarrow_Spline = 0b00001;

    //! Setup flag: Lagrange interpolation with precomposed FIR filters for quantized intersample values
    static const int InterpolationLagrange = 0b00010;

    //! Setup flag: Linear interpolation with precomposed FIR filters for quantized intersample values
    static const int InterpolationLinear = 0b00100;

    //! Setup flag: Newton interpolation with continuous intersample interpolation based on Lagrange polynomials
    static const int InterpolationNewton_Lagrange = 0b01000;

    //! Setup flag: Linear interpolation with continuous intersample interpolation based on basis fuctions
    static const int InterpolationNewton_Spline = 0b10000;

    //! Setup flag: instantaneous delay change with crossfading afterwards over a full block (this always disables upsampling)
    static const int DelayFade = 0b100000;

    //! Setup flag: Lower quality, but faster linear interpolation between samples (does not apply for Lagrange interpolation)
    static const int QualityLow = 0b000000;

    //! Setup flag: Higher quality interpolation with higher order (default behaviour)
    static const int QualityHigh = 0b1000000;

    //! Setup flag: Use prefiltering in the context of spline interpolation. This prefiltering turns the spline approximation into an interpolation
    static const int CorrectionFilterEnabled = 0b10000000;

    //! Token for read cursors
    /*!
     * Data container passed to the VDL instance so that multiple read positions 
     * are possible for a single instance using a shared signal source.
     */
    struct ReadCursor;

    //! Constructor
    /*!
     * \param samplerate    Sampling rate used for filter selection [Hz]
     * \param blocklength   Streaming block length [samples]
     * \param maxdelay      Maximum possible delay [samples]
     * \param nChannels     Channel count in the 2D data buffers
     * \param upsampling    Disable or enables upsampling (Default: upsampling enabled)
     */
    VDL(float samplerate, int blocklength, float maxDelay, std::size_t nChannels,
        bool upsampling = true);

    //! Destructor
    ~VDL();

    //! Return maximum possible delay [samples]
    float getMaximumDelay();

    //! Empty all internal audio buffers
    /*!
     * This does not reset the current delay time.
     * The function is NOT thread-safe w.r.t. process().
     */
    void clear();

    //! Creates a read cursor
    /*!
     * \param initialDelay      Initial delay
     * \return                  A token that needs to be save and passed to read() and setDelay()
     *                          to allow multiple read cursors on one buffer
    */
    VDL::ReadCursor registerReadCursor(float initialDelay, int flags);

    //! Set the delay time
    /*!
    * \param accessor      ReadCursor the VDL is acting on
    * \param delay         new delay value [samples]
    * \param jump          crossfade to new delay instead of adapting
    *                      (e.g. for "teleport", default false, no effect in DelayFade mode)
    */
    void setDelay(VDL::ReadCursor& accessor, float delay, bool jump = false);

    /*! Copying samples of the audio signal into the VDL buffer
    *
    * \param input          2D pointer for input samples written into the VDL buffer
    */
    void write(const float* const* input);

    /*! Interpolates the data taking accessor and VDL specifications into account
    *   
    * \param accessors     ReadCursor the VDL is acting on
    * \param output        2D output buffer into which the VDL writes the interpolated data
    */
    void read(VDL::ReadCursor& accessor, float** output);

    /*! Interpolating the underlying data (audio signal) taking the specific accessor and the VDL
    * interpolation algorithm into account
    *
    * Note: The process function has to act on all accessors which correspond to one VDL.
    *       Resistance will lead to write-cursor (wc) mismatches.
    *
    * \param accessors     list of data containers the VDL is acting on
    * \param input         input samples written into the VDL buffer
    * \param output        list of output buffers into which the VDL writes the interpolated data
    */
    // void process(VDL::Accessor* accessors, const float* input, float** output);

protected:
    /*! Update the backward differences used for the interpolation algorithms in newton structure
    *
    * \param backwardDiffs     Reference to the backward differences
    * \param k0-k1             Latest samples
    */
    inline void updateBackwardDifferences(
        float* backwardDiffs, const float& k0, const float& k1, const float& k2, const float& k3);

    //! IIR prefilter for anti-aliasing upon upsampling
    std::vector<std::unique_ptr<IIRFilter>> aaPreFilter;

    //! Group delay of aaprefilter
    float gd_PreFilter;

    //! upsampling enabled flag
    bool useUpsampling;

    //! data buffer
    RealFrame data;

    //! temporary buffer for output
    RealBuffer out;

    //! length of data buffer
    int datalen;

    //! Blocksize
    int B;

    //! Sample rate [Hz]
    float fs;

    //! Delay values
    float Dmax;

    //! Number of supported channels
    std::size_t nChannels;

    //! write pointer to data
    std::size_t wc;

    std::unique_ptr<EAR::RealBuffer> fadeIn;
    std::unique_ptr<EAR::RealBuffer> fadeOut;
};

struct VDL::ReadCursor
{ // TODO: allow different interpolation methods for different read cursors
public:
    ~ReadCursor() = default;

    //! Get the minimum delay for this read cursor [samples]
    /*!
         * Depends on interpolation method used.
         */
    float getMinimumDelay();

private:
    friend VDL; // all members only accesible via parent VDL

    //! Constructor
    /*!
        * \param initDelay     initial delay [samples], no adaption will be performed to that value
        * \param vdl           VDL instance the read cursor is passed to
        * \param flags         Configure the VDL algorithm.
        *                      Currently, interpolation quality and delay change mode can be set and concatenated with bitwise operator |
        *                      (Default is InterpolationFarrow_Lagrange | QualityLow | DelayAdapt)
        */
    ReadCursor(float initialDelay, VDL* vdl,
        int flags = VDL::InterpolationFarrow_Lagrange | VDL::QualityLow);

    //! Delay values
    float Dcur;
    float Dnew;
    float Dmin;

    //! IIR postfilter for anti-aliasing upon upsampling
    std::vector<std::shared_ptr<IIRFilter>> aaPostFilter;

    //! used for Newton structure
    float backwardDiffs[4];

    //! interpolation filters
    const float* C;

    //! number of coefficients in interpolation filters
    int nC;

    //! number of discrete intersample delays
    int mC;

    //! group delay imposed by lagrange or farrow interpolation
    int gd;

    //! Container to store adjacent samples corresponding to delay
    std::shared_ptr<int> cursorPositions;

    //!FIR filter for turning spline approximation into spline interpolation
    std::vector<std::shared_ptr<FIRFilter>> cFilter;

    bool farrowInterpolation_Lagrange, farrowInterpolation_Spline, crossfade,
        newtonInterpolation_Lagrange, linearInterpolation, newtonInterpolation_Spline, singleJump,
        useCorrectionFilter;
};

} // namespace EAR

#endif /* EAR_VDLIMPL */
