/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_VARIANT
#define EAR_VARIANT

#include <EAR/BaseTypes.h>

#include <string>
#include <vector>

namespace EAR {
//! Var provides an abstraction to hold different varities of primitive values.
class Variant
{
public:
    //! List of all available types within this Variant
    enum class Type
    {
        None = 0,
        Bool = 1,
        Int,
        Float,
        Euler,
        Quaternion,
        Vector3,
        String,
        FloatPtr,
        FloatVector,
    };

    //! Constructor
    /*!
     *  Default constructor
     */
    Variant()
        : type(EAR::Variant::Type::None){};

    //! Constructor
    /*!
     * \param value     instanciate class with integer type
     */
    Variant(int value);

    //! Constructor
    /*!
     * \param value     instanciate class with CSpace (integer) type
     */
    Variant(CSpace value);

    //! Constructor
    /*!
     * \param value     instanciate class with bool type
     */
    Variant(bool value);

    //! Constructor
    /*!
     * \param value     instanciate class with float type
     */
    Variant(float value);

    //! Constructor
    /*!
     * \param value     instance of EAR::Vector3
     */
    Variant(const EAR::Vector3& tripletValue);

    //! Constructor
    /*!
     * \param value     instance of EAR::Euler
     */
    Variant(const EAR::Euler& tripletValue);

    //! Constructor
    /*!
     * \param value     instance of EAR::Quaternion
     */
    Variant(const EAR::Quaternion& v);

    //! Constructor
    /*!
     * \param value     instanciate class with a std::string type
     */
    Variant(const std::string& value);

    //! Constructor
    /*!
     * \param value     instanciate class with a void* type
     */
    Variant(const float* value);

    //! Constructor
    /*!
     * \param value     instanciate class with a vector of floats
     */
    Variant(const std::vector<float>& value);

    //! Deconstructor
    ~Variant(){};

    //! Cast to bool.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator bool() const { return boolValue; };

    //! Cast to int
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator int() const { return intValue; };

    //! Cast to float.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator float() const { return floatValue; };

    //! Cast to Vector3.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator EAR::Vector3() const
    {
        return EAR::Vector3(triplet[0], triplet[1], triplet[2]);
    };

    //! Cast to Euler.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator EAR::Euler() const { return EAR::Euler(triplet[0], triplet[1], triplet[2]); };

    //! Cast to Quaternion.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator EAR::Quaternion() const
    {
        return EAR::Quaternion(quadruplet[0], quadruplet[1], quadruplet[2], quadruplet[3]);
    };

    //! Cast to CSpace.
    /*!
     *  Helper to directly return a CSpace enum type instead of int
     *
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator EAR::CSpace() const { return static_cast<CSpace>(intValue); };

    //! Cast to string reference.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator const std::string&() const
    {
        return strValue; //return *reinterpret_cast<std::string*>(const_cast<char*>(stringValue));
    };

    //! Cast to float*.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator const float*() const { return ptrValue; };

    //! Cast to std::vector<float> reference.
    /*!
     *  Warning! No sanity checking is done.
     *  Check that this type is correct first, if not, data could be destroyed
     */
    inline operator const std::vector<float>&() const { return floatVectorValue; };

    //! Get memory address of the Variant store (don't use if you don't know what you are doing...)
    inline void* getRawData() { return reinterpret_cast<void*>(&floatValue); };

    //! Get the type held in this variant
    const Type& getVariantType() const;

    //! Return a string version of this variant that can later reconstruct the same data
    std::string serialize() const;

    //! Set this variant to the state given by the serialized data
    /*!
     * Caution: This overrides everything that was previously stored in this Variant!
     *
     * \param str       string as formatted by Variant::serialize() (copied by value to operate on it)
     */
    void deserialize(std::string str);

private:
    union
    {
        bool boolValue;
        int intValue;
        float floatValue;
        float triplet[3];
        float quadruplet[4];
        const float* ptrValue;
        // char stringValue[sizeof(std::string)];
    };

    std::string strValue;
    std::vector<float> floatVectorValue;

    Type type;
};

} // namespace EAR

#endif // EAR_VARIANT
