/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_ALIGNED_ALLOCATOR
#define EAR_ALIGNED_ALLOCATOR

#include <EAR/Exception.h>
#include <EAR/config.h>

#if defined(MALLOC_H_AVAILABLE)
#include <malloc.h>
#endif
#if defined(STDLIB_H_AVAILABLE)
#include <cstdlib>
#endif

namespace EAR {

// this is inspired by fftw
#if defined(SIMD_KCVI) || defined(SIMD_AVX512)
#define ALIGNMENT_BYTES 64
#elif defined(SIMD_AVX) || defined(SIMD_AVX2)
#define ALIGNMENT_BYTES 32
#else
#define ALIGNMENT_BYTES 16
#endif

namespace internal {
inline void* allocAligned(std::size_t size);
inline void freeAligned(void* ptr) noexcept;
} // namespace internal

template <typename T> class AlignedAllocator;

/*!
 * void type specialization of aligned allocator
 */
template <> class AlignedAllocator<void>
{
public:
    typedef void* pointer;
    typedef const void* const_pointer;
    typedef void value_type;

    template <class U> struct rebind
    {
        typedef AlignedAllocator<U> other;
    };
};

/*!
 * Aligned memory allocation inspired by the top answer by "Florian" in https://stackoverflow.com/questions/12942548/making-stdvector-allocate-aligned-memory
 * Also see http://en.cppreference.com/w/cpp/memory/allocator for more info.
 */
template <typename T> class AlignedAllocator
{
public:
    typedef T value_type;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef T& reference;
    typedef const T& const_reference;
    typedef std::size_t size_type;
    typedef ptrdiff_t difference_type;

    typedef std::true_type propagate_on_container_move_assignment;

    template <class U> struct rebind
    {
        typedef AlignedAllocator<U> other;
    };

    AlignedAllocator() noexcept {}

    template <class U> AlignedAllocator(const AlignedAllocator<U>&) noexcept {}

    size_type max_size() const noexcept
    {
        return (size_type(~0) - size_type(ALIGNMENT_BYTES)) / sizeof(T);
    }

    pointer address(reference x) const noexcept { return std::addressof(x); }

    const_pointer address(const_reference x) const noexcept { return std::addressof(x); }

    pointer allocate(size_type n, typename AlignedAllocator<void>::const_pointer = 0)
    {
        void* ptr = internal::allocAligned(n * sizeof(T));

        return reinterpret_cast<pointer>(ptr);
    }

    void deallocate(pointer p, size_type) noexcept { return internal::freeAligned(p); }

    template <class U, class... Args> void construct(U* p, Args&&... args)
    {
        ::new (reinterpret_cast<void*>(p)) U(std::forward<Args>(args)...);
    }

    void destroy(pointer p) { p->~T(); }
};

/*!
 * const type specialization of aligned allocator
 */
template <typename T> class AlignedAllocator<const T>
{
public:
    typedef T value_type;
    typedef const T* pointer;
    typedef const T* const_pointer;
    typedef const T& reference;
    typedef const T& const_reference;
    typedef std::size_t size_type;
    typedef ptrdiff_t difference_type;

    typedef std::true_type propagate_on_container_move_assignment;

    template <class U> struct rebind
    {
        typedef AlignedAllocator<U> other;
    };

    AlignedAllocator() noexcept {}

    template <class U> AlignedAllocator(const AlignedAllocator<U>&) noexcept {}

    size_type max_size() const noexcept
    {
        return (size_type(~0) - size_type(ALIGNMENT_BYTES)) / sizeof(T);
    }

    const_pointer address(const_reference x) const noexcept { return std::addressof(x); }

    pointer allocate(size_type n, typename AlignedAllocator<void>::const_pointer = 0)
    {
        void* ptr = internal::allocAligned(n * sizeof(T));

        return reinterpret_cast<pointer>(ptr);
    }

    void deallocate(pointer p, size_type) noexcept { return internal::freeAligned(p); }

    template <class U, class... Args> void construct(U* p, Args&&... args)
    {
        ::new (reinterpret_cast<void*>(p)) U(std::forward<Args>(args)...);
    }

    void destroy(pointer p) { p->~T(); }
};

// every AlignedAllocator can always destroy memory from other AlignedAllocator, because the alignment is fixed at compile time
template <typename T, typename U>
inline bool operator==(const AlignedAllocator<T>&, const AlignedAllocator<U>&) noexcept
{
    return true;
}
template <typename T, typename U>
inline bool operator!=(const AlignedAllocator<T>&, const AlignedAllocator<U>&) noexcept
{
    return false;
}

inline void* internal::allocAligned(std::size_t size)
{
    EAR_CHECK(size > 0, "Length of allocated memory must be positive.");

    void* ptr = nullptr;

    // inspired by fftw
    // FreeBSD and OSX already do 16-byte-aligned allocations
    // for greater alignments, they usually have posix_memalign
#if defined(__FreeBSD__) && (ALIGNMENT_BYTES <= 16)
    ptr = malloc(size);
#elif (defined(__MACOSX__) || defined(__APPLE__)) && (ALIGNMENT_BYTES <= 16)
    ptr = malloc(size);
#elif defined(MEMALIGN_AVAILABLE) // is found out by cmake
    ptr = memalign(ALIGNMENT_BYTES, size);
#elif defined(POSIX_MEMALIGN_AVAILABLE) // is found out by cmake
    posix_memalign(&ptr, ALIGNMENT_BYTES, size);
#elif defined(__ICC) || defined(__INTEL_COMPILER) || defined(HAVE__MM_MALLOC)
    ptr = (void*)_mm_malloc(size, ALIGNMENT_BYTES);
#elif defined(_MSC_VER)
    ptr = (void*)_aligned_malloc(size, ALIGNMENT_BYTES);
#else
    EAR_FAIL("Trying to compile for a system with unknown aligned memory allocation.");
#endif

    return ptr;
}

inline void internal::freeAligned(void* ptr) noexcept
{

#if defined(__ICC) || defined(__INTEL_COMPILER) || defined(HAVE__MM_MALLOC)
    _mm_free(ptr);
#elif defined(_MSC_VER)
    _aligned_free(ptr);
#else
    free(ptr);
#endif
}

} /* namespace EAR */

#endif /* EAR_ALIGNED_ALLOCATOR */
