/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/AudioFile.h>
#include <EAR/Exception.h>
#include <cstdint>
#include <fstream>
#include <iterator>

namespace EAR {

AudioFile::AudioFile() {}

AudioFile::~AudioFile() {}

void AudioFile::load(const std::string& path)
{
    std::ifstream file(path, std::ios::binary);

    if (file.good()) {
        file.seekg(0, std::ios::end);
        auto totalSize = file.tellg();
        std::vector<uint8_t> data(totalSize);
        file.seekg(0);
        file.read(reinterpret_cast<char*>(data.data()), totalSize);

        // copying the file contents (above) is much faster than using ifstream iterators

        // file.unsetf(std::ios::skipws);
        // std::istream_iterator<uint8_t> begin(file), end;
        // std::vector<uint8_t> data(begin, end);

        std::string riff(data.begin(), data.begin() + 4);
        std::string wave(data.begin() + 8, data.begin() + 12);

        int dataStart = findString("data", data);
        int fmtStart = findString("fmt", data);

        if (riff != "RIFF" || wave != "WAVE" || dataStart == -1 || fmtStart == -1) {
            EAR_FAIL("The given file is not a valid WAVE file.");
        }

        int nChannels = (int)readTwoBytes(data, fmtStart + 10);
        sampleRate = static_cast<float>((int)readFourBytes(data, fmtStart + 12));
        bitDepth = (int)readTwoBytes(data, fmtStart + 22);
        int byteDepth = bitDepth / 8;

        int16_t audioFormat = readTwoBytes(data, fmtStart + 8);
        int16_t blocksize = readTwoBytes(data, fmtStart + 20);

        if (blocksize != (nChannels * bitDepth / 8)) {
            EAR_FAIL("The given file is not a valid WAVE file.");
        }

        int32_t chunksize = readFourBytes(data, dataStart + 4);
        int nSamples = chunksize / (nChannels * byteDepth);

        frame.resize(nChannels, nSamples);
        auto framePtr = frame.data();

        int16_t subFormat = 1;
        if (audioFormat == -2) { // -2 == 0xFFFE == extensible wav
            // the 33rd and 34th byte after "fmt " contain the actual format in extensible format
            // see: http://www-mmsp.ece.mcgill.ca/Documents/AudioFormats/WAVE/WAVE.html
            subFormat = readTwoBytes(data, fmtStart + 32);
        }

        int si = dataStart + 8;
        if (audioFormat == 1 || (audioFormat == -2 && subFormat == 1)) {
            for (auto s = 0; s < nSamples; s++) {
                for (auto ch = 0; ch < nChannels; ch++) {
                    framePtr[ch][s] = readSample(data, si);
                    si += byteDepth;
                }
            }
        } else if (audioFormat == 3 || (audioFormat == -2 && subFormat == 3)) {
            for (auto s = 0; s < nSamples; s++) {
                for (auto ch = 0; ch < nChannels; ch++) {
                    framePtr[ch][s] = readFloatSample(data, si);
                    si += byteDepth;
                }
            }
        } else {
            EAR_FAIL("Unsupported WAVE file format (i.e. not PCM or IEEE float samples).");
        }
    } else {
        EAR_FAIL("Not able to read the given WAVE file.");
    }
}

void AudioFile::save(const std::string& path, int _bitDepth, float _sampleRate)
{
    bitDepth = _bitDepth;
    sampleRate = _sampleRate;

    std::vector<uint8_t> data;
    std::ofstream file(path, std::ios::binary);

    int32_t chunksize = static_cast<int>(frame.length() * frame.channels()) * bitDepth / 8;

    stringToData(data, "RIFF");

    // filesize = header size + fmt chunk + data chunk meta + data chunk
    int32_t filesize = 4 + 24 + 8 + chunksize;
    int32ToData(data, filesize);

    stringToData(data, "WAVE");

    stringToData(data, "fmt ");
    int32ToData(data, 16); // format chunk size
    int16ToData(data, 1); // audio format
    int16ToData(data, (int16_t)frame.channels());
    int32ToData(data, (int32_t)sampleRate);

    // bytes per second
    int32ToData(data, (int32_t)((frame.channels() * sampleRate * bitDepth) / 8));

    int16_t blockbytes = (static_cast<int>(frame.channels()) * (bitDepth / 8));
    int16ToData(data, blockbytes);
    int16ToData(data, (int16_t)bitDepth);

    stringToData(data, "data");
    int32ToData(data, chunksize);

    for (unsigned int s = 0; s < frame.length(); s++) {
        for (unsigned int ch = 0; ch < frame.channels(); ch++) {
            float floatval = frame[ch][s];

            switch (bitDepth) {
            case 8: {
                int intval = static_cast<int>(floatval * INT8_MAX) + INT8_MAX;
                data.push_back((uint8_t)intval);
                break;
            }

            case 16:
                int16ToData(data, static_cast<int16_t>(floatval * INT16_MAX));
                break;

            case 24:
                int24ToData(data, static_cast<int32_t>(floatval * 8388607)); // INT24_MAX = 8388607
                break;

            case 32:
                int32ToData(data, static_cast<int32_t>(floatval * INT32_MAX));
                break;

            default:
                EAR_FAIL("The given bit depth is not supported.");
                break;
            }
        }
    }

    if ((frame.length() * blockbytes) % 2 != 0) { // padding byte for odd number of samples
        data.push_back(0x00);
    }

    if (filesize != static_cast<int>(data.size() - 8)) {
        EAR_FAIL("The given file is not a valid WAVE file.");
    }

    if (file.is_open()) {
        for (unsigned int i = 0; i < data.size(); i++) {
            char val = (char)data[i];
            file.write(&val, sizeof(char));
        }

        file.close();
    } else {
        EAR_FAIL("Not able to write to disk.");
    }
}

float AudioFile::getSampleRate() const
{
    return sampleRate;
}

int AudioFile::getBitDepth() const
{
    return bitDepth;
}

RealFrame* AudioFile::getFrame()
{
    return &frame;
}

int AudioFile::findString(const std::string& needle, const std::vector<uint8_t>& haystack)
{
    int res = -1;
    int l = (int)needle.length();

    for (unsigned int i = 0; i < haystack.size() - l; i++) {
        std::string window(haystack.begin() + i, haystack.begin() + i + l);

        if (needle == window) {
            res = i;
            break;
        }
    }

    return res;
}

inline int16_t AudioFile::readTwoBytes(std::vector<uint8_t>& data, int pos)
{
    return ((data[pos + 1] << 8) | data[pos]);
}

inline int32_t AudioFile::readThreeBytes(std::vector<uint8_t>& data, int pos)
{
    return (0x0000 << 24) | (data[pos + 2] << 16) | (data[pos + 1] << 8) | data[pos];
}

inline int32_t AudioFile::readFourBytes(std::vector<uint8_t>& data, int pos)
{
    return (data[pos + 3] << 24) | (data[pos + 2] << 16) | (data[pos + 1] << 8) | data[pos];
}

inline float AudioFile::readFloatSample(std::vector<uint8_t>& data, int pos)
{
    EAR_CHECK_REL(bitDepth == 32, "Only 32 bit IEEE float is supported for WAVE files.");
    int32_t bytes = readFourBytes(data, pos);
    return *reinterpret_cast<float*>(&bytes);
}

inline float AudioFile::readSample(std::vector<uint8_t>& data, int pos)
{
    switch (bitDepth) {
    case 8: {
        uint8_t intval = data[pos];

        return (((float)intval - INT8_MAX) / INT8_MAX);
    }

    case 16: {
        int16_t intval = readTwoBytes(data, pos);

        return ((float)intval / INT16_MAX);
    }

    case 24: {
        int32_t intval = readThreeBytes(data, pos);

        if (intval & 0x800000) { // convert 24 bit sign to 32 bit sign
            intval = intval | ~0xFFFFFF;
        }

        return ((float)intval / 8388607); // INT24_MAX = 8388607
    }

    case 32: {
        int32_t intval = readFourBytes(data, pos);

        return ((float)intval / INT32_MAX);
    }

    default:
        EAR_FAIL("The bit depth of the loaded WAVE file is not supported.");
        return 0.f;
        break;
    }
}

inline void AudioFile::stringToData(std::vector<uint8_t>& data, std::string val)
{
    for (unsigned int c = 0; c < val.length(); c++) {
        data.push_back((uint8_t)val[c]);
    }
}

inline void AudioFile::int16ToData(std::vector<uint8_t>& data, int16_t val)
{
    data.push_back(val & 0xFF);
    data.push_back((val >> 8) & 0xFF);
}

inline void AudioFile::int24ToData(std::vector<uint8_t>& data, int32_t val)
{
    data.push_back(val & 0xFF);
    data.push_back((val >> 8) & 0xFF);
    data.push_back((val >> 16) & 0xFF);
}

inline void AudioFile::int32ToData(std::vector<uint8_t>& data, int32_t val)
{
    data.push_back(val & 0xFF);
    data.push_back((val >> 8) & 0xFF);
    data.push_back((val >> 16) & 0xFF);
    data.push_back((val >> 24) & 0xFF);
}

} // namespace EAR
