/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Box.h>

#include <EAR/Mesh.h>

#include <algorithm>
#include <cmath>

namespace EAR {

Box::Box(ObjectID oid, EAR::Vector3 size, const std::string& name, Position pos)
    : Geometry(oid, name, pos)
    , volume(size)
    , mesh(std::unique_ptr<Mesh>(SceneObjectFactory<Mesh>::getNewObject(name, pos)))
{
    // Vertex Notation: L = Left, R = Right, U = Up, D = Down, F = Front, B = Back
    int LUF = mesh->addVertex(Vector3(-1 * (volume[0] / 2), (volume[1] / 2), (volume[2] / 2)));
    int RUF = mesh->addVertex(Vector3((volume[0] / 2), (volume[1] / 2), (volume[2] / 2)));
    int LDF = mesh->addVertex(Vector3(-1 * (volume[0] / 2), -1 * (volume[1] / 2), (volume[2] / 2)));
    int RDF = mesh->addVertex(Vector3((volume[0] / 2), -1 * (volume[1] / 2), (volume[2] / 2)));
    int LUB = mesh->addVertex(Vector3(-1 * (volume[0] / 2), (volume[1] / 2), -1 * (volume[2] / 2)));
    int RUB = mesh->addVertex(Vector3((volume[0] / 2), (volume[1] / 2), -1 * (volume[2] / 2)));
    int LDB =
        mesh->addVertex(Vector3(-1 * (volume[0] / 2), -1 * (volume[1] / 2), -1 * (volume[2] / 2)));
    int RDB = mesh->addVertex(Vector3((volume[0] / 2), -1 * (volume[1] / 2), -1 * (volume[2] / 2)));

    //Each side is split into two right angle triangles to construct a square
    // Front side
    mesh->addTriangle(RUF, LUF, LDF, material);
    mesh->addTriangle(RUF, LDF, RDF, material);
    // Back Side
    mesh->addTriangle(LUB, RUB, RDB, material);
    mesh->addTriangle(LUB, RDB, LDB, material);
    // Up Side
    mesh->addTriangle(RUB, LUB, LUF, material);
    mesh->addTriangle(RUB, LUF, RUF, material);
    // Down Side
    mesh->addTriangle(RDF, LDF, LDB, material);
    mesh->addTriangle(RDF, LDB, RDB, material);
    // Left Side
    mesh->addTriangle(LUF, LUB, LDB, material);
    mesh->addTriangle(LUF, LDB, LDF, material);
    // Right Side
    mesh->addTriangle(RUB, RUF, RDF, material);
    mesh->addTriangle(RUB, RDF, RDB, material);
    mesh->lock();
}

Box::Box(ObjectID oid, float xs, float ys, float zs, const std::string& name, Position pos)
    : Box(oid, Vector3(xs, ys, zs), name, pos)
{}

bool Box::isInside(const EAR::Vector3& point) const
{
    auto relPoint = getRelativeLocation(point);

    // Check if the absolute relative point exceeds half the volume of all three dimensions
    return ((std::fabs(relPoint[0]) < (volume[0] / 2.f))
        && (std::fabs(relPoint[1]) < (volume[1] / 2.f))
        && (std::fabs(relPoint[2]) < (volume[2] / 2.f)));
}

Vector3 Box::closestPointTo(const EAR::Vector3& point) const
{
    return this->getMesh()->closestPointTo(point);
}

void Box::setMaterial(std::shared_ptr<Material> _material)
{
    material = _material;
    for (auto& tri : mesh->getTriangles()) {
        const_cast<Mesh::Triangle&>(tri).material = _material;
    }
}

std::shared_ptr<Material> Box::getMaterial() const
{
    return material;
}

const Mesh* Box::getMesh() const
{
    return mesh.get();
}

float Box::getVolume() const
{
    return (volume[0] * volume[1] * volume[2]);
}

float Box::getArea() const
{
    return (
        2.f * volume[0] * volume[1] + 2.f * volume[0] * volume[2] + 2.f * volume[1] * volume[2]);
}

Vector3 Box::getSize() const
{
    return volume;
}

void Box::setOrientation(Euler newOrient, bool notifyObs)
{
    if (newOrient != position.orientation) {
        this->mesh->setOrientation(newOrient, notifyObs);
        SceneObject::setOrientation(newOrient,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void Box::setLocation(Vector3 newLoc, bool notifyObs)
{
    if (newLoc != position.location) {
        this->mesh->setLocation(newLoc, notifyObs);
        SceneObject::setLocation(newLoc,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}
void Box::setPosition(Position newPos, SceneObject* parent, bool notifyObs)
{
    if ((newPos.location != position.location) || (newPos.orientation != position.orientation)) {
        this->mesh->setPosition(newPos, parent, notifyObs);
        SceneObject::setPosition(newPos, parent,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void Box::setStatic(bool newState)
{
    this->mesh->setStatic(newState);
    SceneObject::setStatic(newState);
}

float Box::greatestDistance() const
{
    return std::sqrt(std::pow(volume[0], 2) + std::pow(volume[1], 2) + std::pow(volume[2], 2));
}

void Box::getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
    EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    this->mesh->getPOVContourEdges(pov, resultBottomLeft, resultTopRight, resultBottomRight,
        resultTopLeft, povIsInside, closestPoint);
}

void Box::getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    this->mesh->getApparentSize(pov, resWidth, resHeight, povIsInside, closestPoint);
}

Geometry::BoundingBox Box::getBoundingBox() const
{
    return mesh->getBoundingBox();
}

Geometry::BoundingBox Box::getGlobalBoundingBox() const
{
    return mesh->getGlobalBoundingBox();
}

EAR::Vector3 Box::getCenterPosition() const
{
    return mesh->getCenterPosition();
}

std::ostream& operator<<(std::ostream& os, Box& b)
{
    os << "Box \"" << b.name << "\":" << std::endl;
    os << "* Size: " << b.volume[0] << "m, " << b.volume[1] << "m, " << b.volume[2] << "m"
       << std::endl;
    os << "* Location :" << b.position.location[0] << "m, " << b.position.location[1] << "m, "
       << b.position.location[2] << "m" << std::endl;
    os << "* Orientation :" << b.position.orientation[0] << "deg, " << b.position.orientation[1]
       << "deg, " << b.position.orientation[2] << "deg" << std::endl;
    os << "* Material: \"" << b.material->getName() << "\" " << std::endl << std::endl;

    return os;
}

} /* namespace EAR */
