/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Buffer.h>
#include <EAR/Exception.h>

#include <algorithm>
#include <complex>
#include <functional> // needed by Visual Studio to have std::placeholders available

#include "AlignedAllocator.h"

namespace EAR {

template <typename T>
Buffer<T>::Buffer()
    : Buffer(0)
{}

template <typename T> Buffer<T>::Buffer(std::size_t length)
{
    buffer = std::make_unique<std::vector<T, AlignedAllocator<T>>>();
    resize(length);
}

template <typename T> Buffer<T>::Buffer(const Buffer<T>& other)
{
    buffer = std::make_unique<std::vector<T, AlignedAllocator<T>>>();
    *this = other;
}

template <typename T>
Buffer<T>::Buffer(const T* data, std::size_t length)
    : Buffer(length)
{
    write(data, 0, length);
}

template <typename T> Buffer<T>::~Buffer() {}

template <typename T> const T* Buffer<T>::data() const
{
    return buffer->data();
}

template <typename T> T* Buffer<T>::data()
{
    return buffer->data();
}

template <typename T> Buffer<T>& Buffer<T>::operator=(const Buffer<T>& other)
{ // copy
    EAR_CHECK(this != &other, "Self-assignment not allowed.");

    buffer->resize(other.buffer->size());
    std::copy(other.buffer->begin(), other.buffer->end(), buffer->begin());

    return *this;
}

/*template<typename T>
Buffer<T>& Buffer<T>::operator=(Buffer<T>&& other) { // move
    EAR_CHECK(this != &other, "Self-assignment not allowed.");

    // buffer->resize(other.buffer->size());
    // std::copy(other.buffer->begin(), other.buffer->end(), buffer->begin());
    std::move(other.buffer->begin(), other.buffer->end(), buffer->begin());

    return *this;
}*/

template <typename T> void Buffer<T>::mul(T c)
{
    if (c != 1.f) { // no need to do stuff with the unitary element
        std::transform(buffer->begin(), buffer->end(), buffer->begin(),
            std::bind(std::multiplies<T>(), std::placeholders::_1, c));
    }
}

template <typename T> void Buffer<T>::mul(T c, std::size_t offset, std::size_t count)
{
    if (c != 1.f) {
        std::transform(buffer->begin() + offset, buffer->begin() + offset + count,
            buffer->begin() + offset, std::bind(std::multiplies<T>(), std::placeholders::_1, c));
    }
}

template <typename T> void Buffer<T>::mul(const Buffer<T>& src)
{
    EAR_CHECK(src.length() == buffer->size(), "Buffer lengths don't match.");

    std::transform(
        buffer->begin(), buffer->end(), src.data(), buffer->begin(), std::multiplies<T>());
}

template <typename T>
void Buffer<T>::mul(
    const Buffer<T>& src, std::size_t srcOffset, std::size_t dstOffset, std::size_t count)
{
    EAR_CHECK((src.length() >= srcOffset + count) || (buffer->size() >= dstOffset + count),
        "Buffer lengths don't match.");

    std::transform(buffer->begin() + dstOffset, buffer->begin() + dstOffset + count,
        src.data() + srcOffset, buffer->begin() + dstOffset, std::multiplies<T>());
}

template <typename T> void Buffer<T>::rampMul(T startC, T endC)
{
    rampMul(startC, endC, 0, length());
}

template <typename T>
void Buffer<T>::rampMul(T startC, T endC, std::size_t offset, std::size_t count)
{
    if (startC == endC) {
        mul(startC, offset, count);
        return;
    }

    T step = (endC - startC) / static_cast<T>(count);

    for (std::size_t i = 0; i < count; i++) {
        T weight = startC + static_cast<T>(i) * step;
        buffer->operator[](offset + i) *= weight;
    }
}

template <typename T> void Buffer<T>::add(T c)
{
    std::transform(buffer->begin(), buffer->end(), buffer->begin(),
        std::bind(std::plus<T>(), std::placeholders::_1, c));
}

template <typename T> void Buffer<T>::add(T c, std::size_t offset, std::size_t count)
{
    std::transform(buffer->begin() + offset, buffer->begin() + offset + count,
        buffer->begin() + offset, std::bind(std::plus<T>(), std::placeholders::_1, c));
}

template <typename T> void Buffer<T>::add(const Buffer<T>& src)
{
    EAR_CHECK(src.length() == buffer->size(), "Buffer lengths don't match.");

    std::transform(buffer->begin(), buffer->end(), src.data(), buffer->begin(), std::plus<T>());
}

template <typename T>
void Buffer<T>::add(
    const Buffer<T>& src, std::size_t srcOffset, std::size_t dstOffset, std::size_t count)
{
    EAR_CHECK((src.length() >= srcOffset + count) || (buffer->size() >= dstOffset + count),
        "Buffer lengths don't match.");

    std::transform(buffer->begin() + dstOffset, buffer->begin() + dstOffset + count,
        src.data() + srcOffset, buffer->begin() + dstOffset, std::plus<T>());
}

template <typename T> const T& Buffer<T>::operator[](std::size_t index) const
{
    return (*buffer)[index];
}

template <typename T> T& Buffer<T>::operator[](std::size_t index)
{
    return (*buffer)[index];
}

template <typename T> const T& Buffer<T>::at(std::size_t index) const
{
    return (*buffer)[index];
}

template <typename T> T& Buffer<T>::at(std::size_t index)
{
    return (*buffer)[index];
}

template <typename T> void Buffer<T>::read(T* dst, std::size_t offset, std::size_t count) const
{
    std::copy(buffer->data() + offset, buffer->data() + offset + count, dst);
}

template <typename T> void Buffer<T>::write(const T* src, std::size_t offset, std::size_t count)
{
    std::copy(src, src + count, buffer->data() + offset);
}

template <typename T> void Buffer<T>::copyFrom(const Buffer<T>& src)
{
    resize(src.length());
    write(src.data(), 0, buffer->size());
}

template <typename T>
void Buffer<T>::copyFrom(
    const Buffer<T>& src, std::size_t srcOffset, std::size_t dstOffset, std::size_t count)
{
    EAR_CHECK((src.length() >= srcOffset + count) || (buffer->size() >= dstOffset + count),
        "Buffer lengths do not match.");
    write(src.data() + srcOffset, dstOffset, count);
}

template <typename T> void Buffer<T>::shiftLeft(std::size_t count)
{
    std::rotate(buffer->begin(), buffer->begin() + count, buffer->end());
}

template <typename T> void Buffer<T>::shiftRight(std::size_t count)
{
    std::rotate(buffer->rbegin(), buffer->rbegin() + count, buffer->rend());
}

template <typename T> void Buffer<T>::shiftLeftInsert(const T* src, std::size_t count)
{
    std::copy(buffer->begin() + count, buffer->end(), buffer->begin());
    std::copy(src, src + count, buffer->begin() + (buffer->size() - count));
}

template <typename T> void Buffer<T>::shiftRightInsert(const T* src, std::size_t count)
{
    std::copy(buffer->begin(), buffer->end() - count, buffer->begin() + count);
    std::copy(src, src + count, buffer->begin());
}

template <typename T> void Buffer<T>::resize(std::size_t newLength)
{
    buffer->resize(newLength);
}

template <typename T> void Buffer<T>::fill(T value)
{
    std::fill(buffer->begin(), buffer->end(), value);
}

template <typename T> void Buffer<T>::fill(T value, std::size_t offset, std::size_t count)
{
    std::fill(buffer->begin() + offset, buffer->begin() + offset + count, value);
}

template <typename T> void Buffer<T>::zero()
{
    std::fill(buffer->begin(), buffer->end(), T(0));
}

template <typename T> void Buffer<T>::zero(std::size_t offset, std::size_t count)
{
    std::fill(buffer->begin() + offset, buffer->begin() + offset + count, T(0));
}

template <typename T> std::size_t Buffer<T>::length() const
{
    return buffer->size();
}

// allowed specializations:
template class Buffer<float>;
template class Buffer<std::complex<float>>;

} // namespace EAR
