/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/ConeAirflowGenerator.h>

#include <cmath>

namespace EAR {

ConeAirflowGenerator::ConeAirflowGenerator(ObjectID oid, const std::string& name, Position pos)
    : AirflowGeneratorBase(oid, name, pos)
{}

Vector3 ConeAirflowGenerator::getAirspeed(const Vector3& listenerPosition) const
{
    const Vector3 coneDirection = eulerToVec3(getOrientation());
    const Vector3 baseOfCone = getLocation() + coneDirection * length;

    // Distance of point along cone axis
    const float coneDist = (listenerPosition - getLocation()).dot(coneDirection);

    if (coneDist > length || coneDist < 0.f) {
        // Listener is outside the cone
        return { 0, 0, 0 };
    }

    float coneRadius = 0.f; // Radius of cone at listener position
    float topRadiusScale = 0.f; // How much smaller the top radius of the cone is
    if (!hasTopRadius || topRadius == 0.f) {
        // Cone has no top
        coneRadius = baseRadius * coneDist / length;
        topRadiusScale = 0.f;
    } else {
        coneRadius = (baseRadius - topRadius) * (coneDist / length) + topRadius;
        topRadiusScale = topRadius / baseRadius;
    }

    const Vector3 closestPointOnAxis = getLocation() + coneDirection * coneDist;
    const Vector3 orthogonal =
        listenerPosition - closestPointOnAxis; // Vector orthogonal to the cone axis to the listener

    // Position on the top of the cone which "corresponds" with the listener position. I.e. intersection point of the (line between the cone "origin" and listenerposition) and the cone top.
    const Vector3 perceivedTopPosition = getLocation() + orthogonal * topRadiusScale;

    const float orthogonalDistance = orthogonal.abs();
    const float fadeRadius = coneRadius * (edgeRollOff / 100.f);

    if (orthogonalDistance > coneRadius + fadeRadius) {
        // Outside the cone and outside the fade radius
        return { 0, 0, 0 };
    }

    float fadeFactor = 1.f;
    if (orthogonalDistance > coneRadius) {
        // Outside cone, but inside fade region
        fadeFactor = 1 - (orthogonalDistance - coneRadius) / fadeRadius;
    }

    // Limit minimal distance so that airflow is also limited
    const float distance = (perceivedTopPosition - listenerPosition).abs();
    const float maxSpeed = fadeFactor * getSpeed();
    float airspeed = getSpeed() * (1 / std::pow((distance + 1), 2 * getDistanceAttenuationFactor()))
        * fadeFactor;
    if (airspeed > maxSpeed) {
        airspeed = maxSpeed;
    }
    const Vector3 airDirection = (listenerPosition - perceivedTopPosition).normalize();

    return airspeed * airDirection;
}

void ConeAirflowGenerator::setTopRadius(float topRadius)
{
    this->topRadius = topRadius;
    this->hasTopRadius = true;
}

void ConeAirflowGenerator::removeTopRadius()
{
    this->hasTopRadius = false;
    this->topRadius = 0.f;
}

} /* namespace EAR */
