/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Cylinder.h>

#include <EAR/Mesh.h>

#include <cmath>

namespace EAR {

#define CYL_MESH_SUBDIVS 12 // must be divisible by 2

namespace detail {

bool fcmp(float a, float b, float epsilon = Geometry::EPSILON)
{
    return (std::fabs(a - b) <= epsilon);
}

} /* namespace detail */

Cylinder::Cylinder(
    ObjectID oid, float _radius, float _height, const std::string& name, Position pos)
    : Geometry(oid, name, pos)
    , height(_height)
    , majorR(_radius)
    , minorR(_radius)
    , isCircular(true)
    , mesh(std::unique_ptr<Mesh>(SceneObjectFactory<Mesh>::getNewObject(name, pos)))
{
    createMesh();
}

Cylinder::Cylinder(
    ObjectID oid, float a, float b, float _height, const std::string& name, Position pos)
    : Geometry(oid, name, pos)
    , height(_height)
    , majorR(a)
    , minorR(b)
    , mesh(std::unique_ptr<Mesh>(SceneObjectFactory<Mesh>::getNewObject(name, pos)))
{

    if (detail::fcmp(a, b)) {
        isCircular = true;
    } else {
        isCircular = false;
    }

    createMesh();
}

Cylinder::Cylinder(ObjectID oid, const std::string& name)
    : Cylinder(oid, 1.f, 1.f, name)
{}

void Cylinder::createMesh()
{
    EAR::Vector3 p;

    // scale circle to get a better polygonal approximation
    float scale = 2.0f / (1.0 + std::cos(static_cast<float>(M_PI) / CYL_MESH_SUBDIVS));

    // create vertices
    for (int i = 0; i < CYL_MESH_SUBDIVS * 2; i++) {
        p[0] = scale * majorR * std::cos(static_cast<float>(2 * M_PI * i) / CYL_MESH_SUBDIVS);
        p[1] = height / 2;
        p[2] = scale * minorR * std::sin(static_cast<float>(2 * M_PI * i) / CYL_MESH_SUBDIVS);
        if (i >= CYL_MESH_SUBDIVS) {
            p[1] = -p[1];
        }
        mesh->addVertex(p);
    }
    int topCenterIdx = mesh->addVertex({ 0.f, height / 2, 0.f });
    int bottomCenterIdx = mesh->addVertex({ 0.f, -height / 2, 0.f });

    // side faces with outward facing normals
    for (int i = 0; i < CYL_MESH_SUBDIVS; i++) {
        int j = (i > 0) ? i - 1 : CYL_MESH_SUBDIVS - 1;
        mesh->addTriangle(j, i, i + CYL_MESH_SUBDIVS, material);
        mesh->addTriangle(j, i + CYL_MESH_SUBDIVS, j + CYL_MESH_SUBDIVS, material);
    }

    // top faces
    for (int i = 0; i < CYL_MESH_SUBDIVS; i++) {
        int j = (i + 1) % CYL_MESH_SUBDIVS;
        mesh->addTriangle(topCenterIdx, j, i, material);
    }

    // bottom faces
    for (int i = 0; i < CYL_MESH_SUBDIVS; i++) {
        int j = (i + 1) % CYL_MESH_SUBDIVS;
        mesh->addTriangle(bottomCenterIdx, i + CYL_MESH_SUBDIVS, j + CYL_MESH_SUBDIVS, material);
    }

    mesh->lock();
}

float Cylinder::greatestDistance() const
{
    return std::sqrt(std::pow(2 * majorR, 2) + std::pow(height, 2));
}

void Cylinder::getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
    EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    this->mesh->getPOVContourEdges(pov, resultBottomLeft, resultTopRight, resultBottomRight,
        resultTopLeft, povIsInside, closestPoint);
}

void Cylinder::getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    this->mesh->getApparentSize(pov, resWidth, resHeight, povIsInside, closestPoint);
}

bool Cylinder::isInside(const EAR::Vector3& point) const
{
    auto relPoint = getRelativeLocation(point);

    if (isCircular) {
        return (Vector3(relPoint[0], 0, relPoint[2]).abs() + 0.00001 < majorR
            && std::fabs(relPoint[1]) + EPSILON < (height / 2.f));
    } else { // elliptical base implied
        float norm = relPoint[0] * relPoint[0] / (majorR * majorR)
            + relPoint[2] * relPoint[2] / (minorR * minorR) + EPSILON;
        return ((norm < 1.f) && (std::fabs(relPoint[1]) + EPSILON < (height / 2.f)));
    }
}

/* Vector3 Cylinder::closestPoint2Ellipse_Newton(const Vector3& point) const {
    Vector3 e_phi, pointOnEllipse, point2d, point2point;
    float error = 0.0001;
    float alpha = M_PI / 4;
    float errMetric = 0;
    float dAlpha = M_PI / 8;

    //projection to 2-dimensional plane and first quadrant
    point2d[0] = (point[0] > 0) ? point[0] : -point[0];
    point2d[2] = (point[2] > 0) ? point[2] : -point[2];

    do {
        e_phi[0] = -majorR * std::sin(alpha);
        e_phi[2] = minorR * std::cos(alpha);
        e_phi *= 1 / e_phi.abs();

        pointOnEllipse[0] = majorR * std::cos(alpha);
        pointOnEllipse[2] = minorR * std::sin(alpha);

        point2point = point2d - pointOnEllipse;
        point2point *= 1 / point2point.abs();

        errMetric = point2point.dot(e_phi);

        if (errMetric > 0) {
            //go in positive angle direction
            alpha += dAlpha;
            dAlpha /= 2;
        } else {
            // go in negative angle direction
            alpha -= dAlpha;
            dAlpha /= 2;
        }

        alpha = (alpha > M_PI / 2) ? M_PI / 2 : alpha;
        alpha = (alpha < 0) ? 0 : alpha;

    } while (std::abs(errMetric) > error);

    pointOnEllipse[0] = (point[0] < 0) ? -pointOnEllipse[0] : pointOnEllipse[0];
    pointOnEllipse[2] = (point[2] < 0) ? -pointOnEllipse[2] : pointOnEllipse[2];


    return pointOnEllipse;
} */

Vector3 Cylinder::closestPoint2Ellipse_Curvature(
    const EAR::Vector3& point, const unsigned int& nIter) const
{
    EAR::Vector3 finalClosestPoint;
    Vector3 r; //point on ellipse
    Vector3 q; //vector from evolute to point of interest
    Vector3 ev; //evolute of ellipse
    Vector3 p; //point of interest in 2d plane

    float dc, dt;

    float t = M_PI / 4; //The first guess given in degree
    p[0] = (point[0] > 0) ? point[0] : -point[0];
    p[2] = (point[2] > 0) ? point[2] : -point[2];

    for (unsigned int i = 0; i < nIter; i++) {
        finalClosestPoint[0] = majorR * std::cos(t);
        finalClosestPoint[2] = minorR * std::sin(t);

        //evolute of ellipse
        ev[0] = (majorR * majorR - minorR * minorR) * std::pow(std::cos(t), 3) / majorR;
        ev[2] = (minorR * minorR - majorR * majorR) * std::pow(std::sin(t), 3) / minorR;

        //point on ellipse
        r[0] = finalClosestPoint[0] - ev[0];
        r[2] = finalClosestPoint[2] - ev[2];

        //vector from evolute to point of interest
        q[0] = p[0] - ev[0];
        q[2] = p[2] - ev[2];

        //get delta t
        dc = r.abs() * std::asin((r[0] * q[2] - r[2] * q[0]) / (r.abs() * q.abs()));
        dt = dc
            / std::sqrt(majorR * majorR + minorR * minorR - std::pow(finalClosestPoint.abs(), 2));

        t += dt;
        t = (t < M_PI / 2) ? t : M_PI / 2;
        t = (t > 0) ? t : 0;
    }

    //get the correct quadrant
    finalClosestPoint[0] = (point[0] < 0) ? -finalClosestPoint[0] : finalClosestPoint[0];
    finalClosestPoint[2] = (point[2] < 0) ? -finalClosestPoint[2] : finalClosestPoint[2];

    return finalClosestPoint;
}

Vector3 Cylinder::closestPointTo(const EAR::Vector3& point) const
{
    auto relPoint = getRelativeLocation(point);
    Vector3 finalClosestPoint;

    //find closest point
    if (std::abs(relPoint[1]) <= (height / 2.f)) {
        //distinguish between cricle and ellipse
        if (isCircular) {
            Vector3 newOrig(0, relPoint[1], 0);
            finalClosestPoint =
                newOrig + (relPoint - newOrig) * (majorR / (relPoint - newOrig).abs());
        } else {
            finalClosestPoint = closestPoint2Ellipse_Curvature(relPoint, 4);
            finalClosestPoint[1] = relPoint[1];
        }
    } else if (relPoint[1] > (height / 2.f)) {
        Vector3 projPoint(relPoint[0], height / 2.f, relPoint[2]);

        if ((relPoint[0] * relPoint[0] / majorR / majorR
                + relPoint[2] * relPoint[2] / minorR / minorR)
            <= 1.f) {
            finalClosestPoint = projPoint;
        } else {
            if (isCircular) {
                Vector3 newOrig(0, height / 2.f, 0);
                finalClosestPoint =
                    newOrig + (projPoint - newOrig) * (majorR / (projPoint - newOrig).abs());
            } else {
                finalClosestPoint = closestPoint2Ellipse_Curvature(relPoint, 4);
                finalClosestPoint[1] = relPoint[1];
            }
        }
    } else {
        Vector3 projPoint(relPoint[0], -height / 2.f, relPoint[2]);

        if ((relPoint[0] * relPoint[0] / majorR / majorR
                + relPoint[2] * relPoint[2] / minorR / minorR)
            <= 1.f) {
            finalClosestPoint = projPoint;
        } else {
            if (isCircular) {
                Vector3 newOrig(0, -height / 2.f, 0);
                finalClosestPoint =
                    newOrig + (projPoint - newOrig) * (majorR / (projPoint - newOrig).abs());
            } else {
                finalClosestPoint = closestPoint2Ellipse_Curvature(relPoint, 4);
                finalClosestPoint[1] = relPoint[1];
            }
        }
    }

    finalClosestPoint = getGlobalCSTransform() * finalClosestPoint;
    return finalClosestPoint;
}

void Cylinder::setMaterial(std::shared_ptr<Material> _material)
{
    material = _material;
    for (auto& tri : mesh->getTriangles()) {
        const_cast<Mesh::Triangle&>(tri).material = _material;
    }
}

std::shared_ptr<Material> Cylinder::getMaterial() const
{
    return material;
}

float Cylinder::getVolume() const
{
    return M_PI * minorR * majorR * height;
}

float Cylinder::getArea() const
{
    if (isCircular) {
        return 2.f * M_PI * std::pow(majorR, 2) + height * 2.f * M_PI * majorR;
    } else {
        //the length of edge of ellipse is calculated according to the approximation
        //formula of Ramanujan (the more elliptical, the greater the error)
        float lambda = (majorR - minorR) / (majorR + minorR);

        float circumference = M_PI * (minorR + majorR)
            * (1.f + 3.f * lambda * lambda / (10.f + std::sqrt(4.f - 3.f * lambda * lambda)));
        return 2.f * M_PI * majorR * minorR + height * circumference;
    }
}

const Mesh* Cylinder::getMesh() const
{
    return mesh.get();
}

void Cylinder::setOrientation(Euler newOrient, bool notifyObs)
{
    if (newOrient != position.orientation) {
        this->mesh->setOrientation(newOrient, notifyObs);
        SceneObject::setOrientation(newOrient,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void Cylinder::setLocation(Vector3 newLoc, bool notifyObs)
{
    if (newLoc != position.location) {
        this->mesh->setLocation(newLoc, notifyObs);
        SceneObject::setLocation(newLoc,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}
void Cylinder::setPosition(Position newPos, SceneObject* parent, bool notifyObs)
{
    if ((newPos.location != position.location) || (newPos.orientation != position.orientation)) {
        this->mesh->setPosition(newPos, parent, notifyObs);
        SceneObject::setPosition(newPos, parent,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void Cylinder::setStatic(bool newState)
{
    this->mesh->setStatic(newState);
    SceneObject::setStatic(newState);
}

std::ostream& operator<<(std::ostream& os, Cylinder& c)
{
    os << "Cylinder \"" << c.name << "\":" << std::endl;
    os << "* Major Axis: " << c.majorR << "m, " << std::endl;
    os << "* Minor Axis: " << c.minorR << "m, " << std::endl;
    os << "* Height: " << c.height << "m, " << std::endl;
    os << "* Location :" << c.position.location[0] << "m, " << c.position.location[1] << "m, "
       << c.position.location[2] << "m" << std::endl;
    os << "* Orientation :" << c.position.orientation[0] << "deg, " << c.position.orientation[1]
       << "deg, " << c.position.orientation[2] << "deg" << std::endl;
    os << "* Material: \"" << c.material->getName() << "\" " << std::endl << std::endl;

    return os;
};

std::pair<Vector3, Vector3> Cylinder::getCenterOfBases() const
{
    Vector3 upper(0, height / 2, 0);
    Vector3 lower(0, -height / 2, 0);

    HomogenMatrix transform = this->getGlobalCSTransform();

    return std::pair<Vector3, Vector3>(transform * lower, transform * upper);
}

float Cylinder::getMinorRadius() const
{
    return minorR;
}

float Cylinder::getMajorRadius() const
{
    return majorR;
}

float Cylinder::getRadius() const
{
    return getMajorRadius();
}

bool Cylinder::hasCircularBase() const
{
    return isCircular;
}

float Cylinder::getHeight() const
{
    return height;
}

Geometry::BoundingBox Cylinder::getBoundingBox() const
{
    // This may be simply `mesh->getBoundingBox();` but hasn't been tested.

    EAR_FAIL("Cylinder::getBoundingBox is not implemented");
}

Geometry::BoundingBox Cylinder::getGlobalBoundingBox() const
{
    // This may be simply `mesh->getGlobalBoundingBox();` but hasn't been tested.

    EAR_FAIL("Cylinder::getBoundingBox is not implemented");
}

EAR::Vector3 Cylinder::getCenterPosition() const
{
    return mesh->getCenterPosition();
}

} /* namespace EAR */
