/***************************************************************************
This software module was originally developed jointly by

Ericsson AB and Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB and Fraunhofer IIS retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/DistanceModels.h>

#include <cmath>

namespace EAR {

DistanceGainVolumetric::DistanceGainVolumetric(float _dampingStart)
    : dampingStart(_dampingStart)
{
    // These are empirically derived numbers
    gainCompDist = _dampingStart + 0.5f;
    maxCompGain = 2.f;
}

float DistanceGainVolumetric::getGainForDistance(float dist, float dim1, float dim2, float refDist,
    float portalComp, bool gainComp, float distExp)
{
    float gain = 0.f; // outside of rendering area
    float refDistGain = 1.0;
    float dampingstartGain = 1.0;
    float extraGain = 0.f;

    //shaping the distance attenuation:
    // distExp > 1: distance will have larger effect on gains
    // distExp < 1: distance will have smaller effect on gains
    dist = std::pow(dist, distExp);

    // make sure that refererence distance is never smaller than the minimum distance
    if (refDist < dampingStart) {
        refDist = dampingStart;
    }

    // Special tuning for primary Homogeneous extent and portal
    if (gainComp && dist < gainCompDist) {
        extraGain = maxCompGain * (gainCompDist - dist);
    } else if (portalComp != 0.f) {
        dist = (dist <= portalComp ? 0 : dist - portalComp);
    }

    // calculate distance gain at reference distance
    refDistGain = VolumetricGainModel(refDist, dim1, dim2);

    if (dist <= dampingStart) { // closer than the minimum distance
        // calculate distance gain at minimum distance, taking reference distance into account
        gain = (VolumetricGainModel(dampingStart, dim1, dim2) + 0.0001f) / (refDistGain + 0.0001f);
        if (gainComp) {
            gain += maxCompGain * 0.5f;
        }
    } else { // default case
        // avoiding division by zero with a constant fraction expansion
        gain =
            (VolumetricGainModel(dist, dim1, dim2) + 0.0001f) / (refDistGain + 0.0001f) + extraGain;
    }

    return gain;
}

float DistanceGainVolumetric::VolumetricGainModel(float dist, float dim1, float dim2)
{
    float L1, L2, volumetricGain;

    if (dim1 > dim2) {
        L1 = dim1;
        L2 = dim2;
    } else {
        L1 = dim2;
        L2 = dim1;
    }

    if (dist > L1) {
        volumetricGain = std::pow(dist, -1);
    } else if (dist > fmax(L2, (L1 / 6))) {
        volumetricGain = std::pow(L1, -.25) * std::pow(dist, -.75);
    } else if (dist > (L2 / 6)) {
        volumetricGain = std::pow(L1 * fmax(L2, (L1 / 6)), -.25) * std::pow(dist, -.5);
    } else {
        volumetricGain = std::pow(L1 * fmax(L2, (L1 / 6)), -.25) * std::pow((L2 / 6), -.375)
            * std::pow(dist, -.125);
    }

    return volumetricGain;
}

} // namespace EAR
