/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/EmbreeTracer.h>

#include <EAR/Exception.h>
#include <EAR/Geometry.h>

// This file should probably be moved to avoid the strange include:
#include "../../core/include/MPEGI/Helpers/EmbreeHelper.h"

#include <embree3/rtcore.h>

#include <cassert>
#include <cstdint>
#include <cstdlib>
#include <limits>
#include <string>

#include <iostream>

namespace EAR {

namespace detail {
// function that will be used to filter backface hits if neccessary
// (adapted from the example on https://github.com/embree/embree/blob/master/tutorials/intersection_filter/)
void backfaceFilter(const RTCFilterFunctionNArguments* args)
{
    // avoid crashing when debug visualizations are used
    if (args->context == nullptr)
        return;
    RTCRay* ray = (RTCRay*)args->ray; // casting from RayN to Ray
    RTCHit* hit = (RTCHit*)args->hit; // casting from HitN to Hit

    EAR_CHECK_REL(args->N == 1, "An Embree misconfiguration breaks the backface filter.");

    // ignore inactive rays
    if (args->valid[0] != -1)
        return;

    // calculate dot product of normal direction and ray direction
    float dot = ray->dir_x * hit->Ng_x + ray->dir_y * hit->Ng_y + ray->dir_z * hit->Ng_z;
    // if they are pointing in opposite directions, reject the hit
    if (dot > 0) {
        args->valid[0] = 0;
    }
}
} // namespace detail

RTCDevice EmbreeTracer::device = nullptr;

EmbreeTracer::EmbreeTracer(const std::vector<const Mesh*> geometry, bool staticMatters,
    bool ignoreBackface, float bundleWidth)
    : lastDist(0.f)
    , lastPrimID(-1)
{
    // construct bundle raster for spaced ray bundles
    bundleRaster.push_back(EAR::Vector3(0, 0, 0));
    for (int i = 0; i < 8; i++) {
        float r = bundleWidth / 2;
        if (i % 2 == 1) {
            r = bundleWidth / 4;
        }
        bundleRaster.push_back(EAR::polarToCartesian(EAR::Polar(90, i * 360.f / 8, r)));
    }

    scene = rtcNewScene(getDevice());
    rtcSetSceneFlags(scene,
        RTC_SCENE_FLAG_DYNAMIC | RTC_SCENE_FLAG_ROBUST | RTC_SCENE_FLAG_CONTEXT_FILTER_FUNCTION);
    rtcSetSceneBuildQuality(scene, RTC_BUILD_QUALITY_LOW);

    for (const Mesh* geo : geometry) {
        attachGeometry(geo, staticMatters, ignoreBackface);
    }

    rtcJoinCommitScene(scene);
    rayhit = std::make_unique<RTCRayHit>();
}

EmbreeTracer::EmbreeTracer(
    const Mesh* geometry, bool staticMatters, bool ignoreBackface, float bundleWidth)
    : lastDist(0.f)
    , lastPrimID(-1)
{
    // construct bundle raster for spaced ray bundles
    bundleRaster.push_back(EAR::Vector3(0, 0, 0));
    for (int i = 0; i < 8; i++) {
        float r = bundleWidth / 2;
        if (i % 2 == 1) {
            r = bundleWidth / 4;
        }
        bundleRaster.push_back(EAR::polarToCartesian(EAR::Polar(90, i * 360.f / 8, r)));
    }

    scene = rtcNewScene(getDevice());
    rtcSetSceneFlags(scene,
        RTC_SCENE_FLAG_DYNAMIC | RTC_SCENE_FLAG_ROBUST | RTC_SCENE_FLAG_CONTEXT_FILTER_FUNCTION);
    rtcSetSceneBuildQuality(scene, RTC_BUILD_QUALITY_LOW);

    attachGeometry(geometry, staticMatters, ignoreBackface);

    rtcJoinCommitScene(scene);
    rayhit = std::make_unique<RTCRayHit>();
}

EmbreeTracer::~EmbreeTracer()
{
    rtcReleaseScene(scene);
    scene = nullptr;

    // remove observers from EAR::SceneObjects
    for (auto& pair : mesh2geo) {
        const_cast<Mesh*>(pair.first)->detach(this);
    }
}

void EmbreeTracer::attachGeometry(const Mesh*& geo, bool staticMatters, bool ignoreBackface)
{
    RTCGeometry mesh = rtcNewGeometry(getDevice(), RTC_GEOMETRY_TYPE_TRIANGLE);

    rtcSetSharedGeometryBuffer(mesh, RTC_BUFFER_TYPE_VERTEX, 0, RTC_FORMAT_FLOAT3,
        geo->getVertices(Mesh::CSpace::Global).data(), 0, sizeof(Vector3),
        geo->getVertices().size());
    rtcSetSharedGeometryBuffer(mesh, RTC_BUFFER_TYPE_INDEX, 0, RTC_FORMAT_UINT3,
        geo->getTriangles().data(), 0, sizeof(Mesh::Triangle), geo->getTriangles().size());

    if (geo->isStatic() && staticMatters) {
        rtcSetGeometryBuildQuality(mesh, RTC_BUILD_QUALITY_MEDIUM);

        rtcCommitGeometry(mesh);
        unsigned int id = rtcAttachGeometry(scene, mesh);
        geo2mesh[id] = geo;
    } else {
        rtcSetGeometryBuildQuality(mesh, RTC_BUILD_QUALITY_REFIT); // RTC_BUILD_QUALITY_LOW

        rtcCommitGeometry(mesh);
        unsigned int id = rtcAttachGeometry(scene, mesh);
        geo2mesh[id] = geo;
        mesh2geo[geo] = id;

        const_cast<Mesh*>(geo)->attach(this); // add change observer to EAR::SceneObject
    }

    if (ignoreBackface) {
        rtcSetGeometryIntersectFilterFunction(mesh, detail::backfaceFilter);
        rtcSetGeometryOccludedFilterFunction(mesh, detail::backfaceFilter);
    }

    rtcReleaseGeometry(mesh);
}

void EmbreeTracer::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    const Mesh* geo = dynamic_cast<const Mesh*>(obj);
    if (modification == EAR::SceneObjectObserver::Property::Position
        && mesh2geo.find(geo) != mesh2geo.end()) {
        std::lock_guard<std::mutex> guard(sceneMutex);

        RTCGeometry mesh = rtcGetGeometry(
            scene, mesh2geo[geo]); // does not increment reference count, so no release needed
        rtcUpdateGeometryBuffer(mesh, RTC_BUFFER_TYPE_VERTEX, 0);
        rtcCommitGeometry(mesh);
        rtcCommitScene(scene); // this is the expensive bit TODO: maybe do this only every ~0.5 s
    }
}

RTCDevice EmbreeTracer::getDevice()
{
    if (!device) {
        device = rtcNewDevice("threads=1,start_threads=1");
    }
    return device;
}

RTCScene EmbreeTracer::getScene()
{
    return scene;
}

void EmbreeTracer::initRayHitForIntersection(
    const Vector3& origin, const Vector3& direction, float maxDist)
{

    Vector3 relativeTarget = direction.abs() ? direction.normalize() : Vector3{};

    rayhit->ray.org_x = origin[0];
    rayhit->ray.org_y = origin[1];
    rayhit->ray.org_z = origin[2];
    rayhit->ray.tnear = 0.f;
    rayhit->ray.dir_x = relativeTarget[0];
    rayhit->ray.dir_y = relativeTarget[1];
    rayhit->ray.dir_z = relativeTarget[2];
    rayhit->ray.tfar = maxDist;
    rayhit->ray.flags = 0;

    rayhit->hit.geomID = RTC_INVALID_GEOMETRY_ID;
    rayhit->hit.primID = RTC_INVALID_GEOMETRY_ID;
    rayhit->hit.instID[0] = -1;
}

bool EmbreeTracer::isOccluded(const Vector3& origin, const Vector3& target, float threshold)
{
    float dist = target.abs();

    RTCIntersectContext context;
    rtcInitIntersectContext(&context);

    RTCRay* ray = &(rayhit->ray);

    ray->org_x = origin[0];
    ray->org_y = origin[1];
    ray->org_z = origin[2];
    ray->tnear = 0.f;
    ray->dir_x = target[0] / dist; // normalized direction vector
    ray->dir_y = target[1] / dist;
    ray->dir_z = target[2] / dist;
    ray->tfar = dist - threshold; // after hitting, it will change to -inf
    ray->flags = 0; // has to be set as 0

    std::lock_guard<std::mutex> guard(sceneMutex);

    rtcOccluded1(scene, &context, ray);

    lastDist = ray->tfar;
    lastPrimID = rayhit->hit.primID;

    return (lastDist <= 0);
}

const EAR::Mesh* EmbreeTracer::firstHit(const Vector3& origin, const Polar& direction, bool maxDist)
{
    RTCIntersectContext context;
    rtcInitIntersectContext(&context);

    if (maxDist) {
        initRayHitForIntersection(origin, polarToCartesian(direction), direction[2]);
    } else {
        initRayHitForIntersection(origin, polarToCartesian(direction));
    }

    std::lock_guard<std::mutex> guard(sceneMutex);

    rtcIntersect1(scene, &context, rayhit.get());
    lastDist = rayhit->ray.tfar;
    lastPrimID = rayhit->hit.primID;

    if (geo2mesh.find(rayhit->hit.geomID) != geo2mesh.end()) {
        return geo2mesh[rayhit->hit.geomID];
    } else {
        return nullptr;
    }
}

float EmbreeTracer::firstHitBundle(
    const Vector3& origin, const Polar& direction, const EAR::Mesh*& hitMesh)
{
    RTCIntersectContext context;
    rtcInitIntersectContext(&context);

    // TODO: use ray bundle for this

    Vector3 relativeTarget = polarToCartesian(direction);
    HomogenMatrix rotation(Vector3(), Euler(direction[0], direction[1], 0));

    std::lock_guard<std::mutex> guard(sceneMutex);

    float minDist = direction[2];
    int closestGeo = RTC_INVALID_GEOMETRY_ID;
    std::vector<int> geosHit;

    for (unsigned int i = 0; i < bundleRaster.size(); i++) {
        Vector3 displacement = rotation * bundleRaster[i];
        Vector3 target = (relativeTarget - displacement);
        float dist = target.abs();

        rayhit->ray.org_x = origin[0] + displacement[0];
        rayhit->ray.org_y = origin[1] + displacement[1];
        rayhit->ray.org_z = origin[2] + displacement[2];
        rayhit->ray.tnear = 0.f;
        rayhit->ray.dir_x = target[0] / dist;
        rayhit->ray.dir_y = target[1] / dist;
        rayhit->ray.dir_z = target[2] / dist;
        rayhit->ray.tfar = dist;
        rayhit->ray.flags = 0;

        rayhit->hit.geomID = RTC_INVALID_GEOMETRY_ID;
        rayhit->hit.primID = RTC_INVALID_GEOMETRY_ID;
        rayhit->hit.instID[0] = -1;

        rtcIntersect1(scene, &context, rayhit.get());

        if (rayhit->ray.tfar < minDist && rayhit->hit.geomID != RTC_INVALID_GEOMETRY_ID) {
            minDist = rayhit->ray.tfar;
            closestGeo = rayhit->hit.geomID;
            lastPrimID = rayhit->hit.primID;
        }
        geosHit.push_back(rayhit->hit.geomID);
    }
    lastDist = minDist;

    int count = 0;
    if (geo2mesh.find(closestGeo) != geo2mesh.end()) {
        hitMesh = geo2mesh[closestGeo];

        // see how many rays hit the same
        for (int gid : geosHit) {
            if (gid == closestGeo) {
                count++;
            }
        }
    } else {
        hitMesh = nullptr; // set to nullptr if nothing is hit
    }

    return static_cast<float>(count) / bundleRaster.size();
}

bool EmbreeTracer::occludedByGivenMesh(
    const Vector3& origin, const Vector3& target, const EAR::Mesh* givenMesh)
{
    float targetDistance = (target - origin).abs();
    if (targetDistance < 0.001f) {
        return false;
    }

    std::lock_guard<std::mutex> guard(sceneMutex);

    RTCIntersectContext context;
    rtcInitIntersectContext(&context);
    initRayHitForIntersection(origin, target - origin);
    rayhit->ray.tfar = targetDistance;

    while (true) {
        rtcIntersect1(scene, &context, rayhit.get());

        if (geo2mesh.find(rayhit->hit.geomID) != geo2mesh.end()) {
            // hit by something
            const EAR::Mesh* hitMesh = geo2mesh.find(rayhit->hit.geomID)->second;
            if (hitMesh == givenMesh) {
                lastDist = rayhit->ray.tfar;
                lastPrimID = rayhit->hit.primID;
                return true;
            }

            // reset ray status
            rayhit->ray.tnear = rayhit->ray.tfar + 0.001f;
            rayhit->ray.tfar = targetDistance;
        } else {
            // no hit by anything
            return false;
        }
    }
}

float EmbreeTracer::getDistanceToLastHit()
{
    return lastDist;
}

int EmbreeTracer::getPrimitiveOfLastHit()
{
    return lastPrimID;
}

bool EmbreeTracer::isInside(const Vector3& point)
{
    float azimuth = (std::rand() % 360) - 180;
    float elevation = (std::rand() % 360) - 180;

    std::lock_guard<std::mutex> guard(sceneMutex);

    return isInsideImpl(point, polarToCartesian(Polar(azimuth, elevation, 1)));
}

// Anonymous namespace for local use
namespace {

void countHits(const struct RTCFilterFunctionNArguments* args);

struct CountHitsContext
{
    CountHitsContext()
    {
        rtcInitIntersectContext(&this->ctx);
        // incoherent is the default
        this->ctx.flags = RTC_INTERSECT_CONTEXT_FLAG_INCOHERENT;
        this->ctx.filter = countHits;
    }
    RTCIntersectContext ctx; // This must be the first field of the struct!
    uint8_t hits = 0;
};

void countHits(const struct RTCFilterFunctionNArguments* args)
{
    assert(args->N == 1);

    auto* ray = (RTCRay*)args->ray;
    auto* hit = (RTCHit*)args->hit;
    auto* ctx = (CountHitsContext*)args->context;

    constexpr float ray_eps = 0.0001f;

    // treat all points on the surface as external
    // if embree does not guarantee a hit when the ray starts near a surface,
    // this seems to be the only consistent choice
    if (ray->tfar < ray_eps) {
        // Stop working on this ray, mark as "outside" (set counter to even number)
        ctx->hits = 0;
        args->valid[0] = -1;
        return;
    }

    // test if an edge or a vertex has been hit (if yes, abort
    // and another ray will be shot in random direction)
    if (hit->u == 0.0f || hit->u == 1.f || hit->v == 0.f || hit->u + hit->v == 1.f) {
        ctx->hits = UINT8_MAX;
        args->valid[0] = -1;
        return;
    }

    ++ctx->hits;

    if (ctx->hits == UINT8_MAX) {
        // Too many hits, stop handling this ray
        args->valid[0] = -1;
        return;
    }

    // Continue working on the current ray:
    args->valid[0] = 0;
}

} // anonymous namespace

bool EmbreeTracer::isInsideImpl(const Vector3& origin, const Vector3& direction)
{
    using MPEGI::EmbreeHelper::setDir;
    using MPEGI::EmbreeHelper::setOrg;

    CountHitsContext ctx{};
    RTCRay ray{};
    setOrg(ray, origin);
    setDir(ray, direction);

    ray.tnear = 0;
    ray.tfar = std::numeric_limits<float>::infinity();
    ray.flags = 0;
    rtcOccluded1(scene, &ctx.ctx, &ray);

    if (ctx.hits == UINT8_MAX) {
        float azimuth = (rand() % 360) - 180;
        float elevation = (rand() % 360) - 180;
        // NB: There is no base case. With a lot of bad luck, this can lead to a stack overflow.
        return isInsideImpl(origin, polarToCartesian(Polar(azimuth, elevation, 1)));
    }

    // Odd number of hits means "is inside":
    return ctx.hits % 2;
}

} // namespace EAR
