/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/FIR2MinphaseFilterDesigner.h>

#include <EAR/config.h>

#include <algorithm>
#include <cmath>
#include <complex>
#include <numeric>

namespace EAR {

void piecelin(const float x[], const float y[], int K, const float u[], float v[], int N)
{
    const float eps = std::numeric_limits<float>::epsilon();

    // get derivative dy/dx, making sure repeated x values are not causing division by zero
    std::vector<float> dydx = std::vector<float>(K - 1);
    for (size_t k = 0; k < K - 1; k++) {
        const float dx = x[k + 1] - x[k];
        dydx[k] = (y[k + 1] - y[k]) / (dx == 0 ? eps : dx);
    }

    // get interpolated values
    int k = 0;
    for (size_t n = 0; n < N; n++) {
        // get subinterval index
        while ((k < K - 2) && (u[n] >= x[k + 1])) {
            k++;
        }

        // linear interpolation
        v[n] = y[k] + (u[n] - x[k]) * dydx[k];
    }
}

void piecelin(const RealBuffer& x, const RealBuffer& y, const RealBuffer& u, RealBuffer& v)
{
    EAR_CHECK(x.length() == y.length(), "Sample point buffers must have equal length");
    EAR_CHECK(u.length() == v.length(), "Query point buffers must have equal length");

    // perform linear interpolation of sample points (x,y), to query points (u,v)
    piecelin(x.data(), y.data(), x.length(), u.data(), v.data(), u.length());
}

Hilbert::Hilbert(std::size_t N)
    : N(N)
    , w(N)
    , in(N)
    , out(N)
{
    fft = FFT::create(N);
    ifft = IFFT::create(N);

    // index of nyquist frequency
    int Nq = N / 2;

    // init hilbert window
    w.zero();
    w[0] = 1; // DC
    if (N % 2 == 0) {
        // even length
        w.fill(2, 1, Nq - 1);
        w[Nq] = 1; // Nyquist
    } else {
        // odd length
        w.fill(2, 1, Nq);
    }

    // ifft normalization
    w.mul(std::complex(1.0f / N));
}

void Hilbert::transform(const RealBuffer& b, ComplexBuffer& r)
{
    EAR_CHECK(b.length() == r.length(), "Buffers must have equal length");

    // assign real part
    in.zero();
    in.writeReal(b.data(), 0, N);

    // hilbert transform
    transform(in, r);
}

void Hilbert::transform(const ComplexBuffer& b, ComplexBuffer& r)
{
    EAR_CHECK(b.length() == r.length(), "Buffers must have equal length");

    // hilbert transform
    fft->transform(b.data(), out.data());
    out.mul(w);
    ifft->transform(out.data(), r.data());
}

MinimumPhase::MinimumPhase(std::size_t N)
    : N(N)
    , Hin(N * 2)
    , Hout(N * 2)
    , in(N * 2)
    , out(N * 2)
{
    fft = FFT::create(N * 2);
    ifft = IFFT::create(N * 2);
    hilbert = std::make_unique<Hilbert>(N * 2);
}

void MinimumPhase::transform(const RealBuffer& b, RealBuffer& r)
{
    // zeropad
    in.zero();
    in.writeReal(b.data(), 0, N);

    // goto frequencydomain
    fft->transform(in.data(), Hin.data());

    // log amplitude
    for (int n = 0; n < 2 * N; n++) {
        Hin[n] = std::log(std::abs(Hin[n]) + eps);
    }

    // hilbert
    hilbert->transform(Hin, Hout);

    // exp(conj())
    for (size_t n = 0; n < 2 * N; n++) {
        Hout[n] = std::exp(std::conj(Hout[n]));
    }

    // go back to timedomain
    ifft->transform(Hout.data(), out.data());

    // remove zeropad
    out.readReal(r.data(), 0, N);

    // fft scaling
    r.mul(0.5f / N);
}

FIR2MinphaseFilterDesigner::FIR2MinphaseFilterDesigner(std::size_t Ntaps, float fs)
    : fs(fs)
    , Ntaps(Ntaps)
{
    minimumPhase = std::make_unique<MinimumPhase>(Ntaps);

    // Nextpow 2 length of Ntaps(scipy method)
    Nf = 1 + std::pow(2.0, std::ceil(std::log2((Ntaps - 1))));
    Nds = (Nf - 1) * 2;

    wn = RealBuffer(Nf);
    k = RealBuffer(Nf);
    H = ComplexBuffer(Nf);
    h = RealBuffer(Nds);
    ifft = IRFFT::create(Nds);
    fft = FFT::create(Nds);

    // create uniform frequency mesh of length Nf
    for (size_t n = 0; n < Nf; n++) {
        wn[n] = static_cast<float>(n) / (Nf - 1) * (fs / 2.0f);
    }
}

//! Destructor
FIR2MinphaseFilterDesigner::~FIR2MinphaseFilterDesigner() {}

/*! Designs a minimum phase filter given a frequency vector and a gain using the frequency sampling method.
        frequency vector f runs from [DC ... nyquist] in (Hz)
        gain vector contains gains per frequency  (dimensionless)
        returns b, vector of length Ntaps with the filtercoefficients.
    */
void FIR2MinphaseFilterDesigner::design(const RealBuffer& f, const RealBuffer& gains, RealBuffer& b)
{
    EAR_CHECK(f.length() == gains.length(), "Frequency and gains buffers must have equal lengths.");

    // make sure filtercoefficients fit
    b.resize(Ntaps);

    // Linearly interpolate the desired response to a onesided spectrum
    // with a uniform mesh wn of length Ntaps
    piecelin(f, gains, wn, k);

    // Adjust the phases of the coefficients so that the first `ntaps` of the
    // inverse FFT are the desired filter coefficients.
    int phi = Ntaps % 2 == 0 ? Ntaps : Ntaps - 1;
    for (size_t n = 0; n < Nf; n++) {
        H[n] = k[n] * std::exp(wn[n] * std::complex<float>(0.0f, -(phi * M_PI)) / fs);
    }

    // inverse fft of onesided spectrum
    ifft->transform(H.data(), h.data());
    h.mul(1.0f / Nds);

    // crop and apply window on impulse response
    b.write(h.data(), 0, Ntaps);
    applyHamming(b);

    // get minimumphase of impulse response
    minimumPhase->transform(b, b);
}

void FIR2MinphaseFilterDesigner::design(const RealBuffer& measuredFrequencies,
    const RealBuffer& measuredGains, const RealBuffer& targetFrequencies,
    const RealBuffer& targetGains, RealBuffer& b)
{
    EAR_CHECK(measuredFrequencies.length() == measuredGains.length(),
        "Measured frequency and gains buffers must have equal lengths.");
    EAR_CHECK(targetFrequencies.length() == targetGains.length(),
        "Target frequency and gains buffers must have equal lengths.");

    // make sure filtercoefficients fit
    b.resize(Ntaps);

    // Linearly interpolate the desired response to a onesided spectrum
    // with a uniform mesh wn of length Ntaps
    RealBuffer kMeasured = RealBuffer(Nf);
    RealBuffer kTarget = RealBuffer(Nf);
    piecelin(measuredFrequencies, measuredGains, wn, kMeasured);
    piecelin(targetFrequencies, targetGains, wn, kTarget);

    // get transfer function (limit between -20 and 20 dB)
    for (size_t n = 0; n < Nf; n++) {
        k[n] = kTarget[n] / kMeasured[n];
        k[n] = std::max(k[n], 0.1f);
        k[n] = std::min(k[n], 10.0f);
    }

    // Adjust the phases of the coefficients so that the first `ntaps` of the
    // inverse FFT are the desired filter coefficients.
    int phi = Ntaps % 2 == 0 ? Ntaps : Ntaps - 1;
    for (size_t n = 0; n < Nf; n++) {
        H[n] = k[n] * std::exp(wn[n] * std::complex<float>(0.0f, -(phi * M_PI)) / fs);
    }

    // inverse fft of onesided spectrum
    ifft->transform(H.data(), h.data());
    h.mul(1.0f / Nds);

    // crop and apply window on impulse response
    b.write(h.data(), 0, Ntaps);
    applyHamming(b);

    // get minimumphase of impulse response
    minimumPhase->transform(b, b);
}

void FIR2MinphaseFilterDesigner::design(const RealBuffer& measuredFrequencies,
    const RealBuffer& measuredGains, const RealBuffer& targetFrequencies,
    const RealBuffer& targetGains, const RealBuffer& compensationFrequencies,
    const RealBuffer& compensationGains, RealBuffer& b)
{
    EAR_CHECK(measuredFrequencies.length() == measuredGains.length(),
        "Measured frequency and gains buffers must have equal lengths.");
    EAR_CHECK(targetFrequencies.length() == targetGains.length(),
        "Target frequency and gains buffers must have equal lengths.");
    EAR_CHECK(compensationFrequencies.length() == compensationGains.length(),
        "Compensation frequency and gains buffers must have equal lengths.");

    // make sure filtercoefficients fit
    b.resize(Ntaps);

    // Linearly interpolate the desired response to a onesided spectrum
    // with a uniform mesh wn of length Ntaps
    RealBuffer kMeasured = RealBuffer(Nf);
    RealBuffer kTarget = RealBuffer(Nf);
    RealBuffer kCompensation = RealBuffer(Nf);
    piecelin(measuredFrequencies, measuredGains, wn, kMeasured);
    piecelin(targetFrequencies, targetGains, wn, kTarget);
    piecelin(compensationFrequencies, compensationGains, wn, kCompensation);

    // get transfer function (limit between -20 and 20 dB)
    for (size_t n = 0; n < Nf; n++) {
        k[n] = kCompensation[n] * kTarget[n] / kMeasured[n];
        k[n] = std::max(k[n], 0.1f);
        k[n] = std::min(k[n], 10.0f);
    }

    // Adjust the phases of the coefficients so that the first `ntaps` of the
    // inverse FFT are the desired filter coefficients.
    int phi = Ntaps % 2 == 0 ? Ntaps : Ntaps - 1;
    for (size_t n = 0; n < Nf; n++) {
        H[n] = k[n] * std::exp(wn[n] * std::complex<float>(0.0f, -(phi * M_PI)) / fs);
    }

    // inverse fft of onesided spectrum
    ifft->transform(H.data(), h.data());
    h.mul(1.0f / Nds);

    // crop and apply window on impulse response
    b.write(h.data(), 0, Ntaps);
    applyHamming(b);

    // get minimumphase of impulse response
    minimumPhase->transform(b, b);
}

void FIR2MinphaseFilterDesigner::phaseAlign(RealBuffer& ch1, RealBuffer& ch2, const size_t maxShift)
{
    ComplexBuffer in1;
    ComplexBuffer in2;
    ComplexBuffer IN1;
    ComplexBuffer IN2;
    RealBuffer temp1;
    RealBuffer temp2;
    float phi;

    EAR_CHECK(
        ch1.length() == ch2.length(), "Filter lengths for phase alignment are not identical.");
    EAR_CHECK(maxShift < Nf, "Maximum allowed group delay is too large.");

    in1.resize(Nds);
    in2.resize(Nds);
    IN1.resize(Nds);
    IN2.resize(Nds);
    temp1.resize(Nds);
    temp2.resize(Nds);

    in1.writeReal(ch1.data(), 0, Ntaps);
    in2.writeReal(ch2.data(), 0, Ntaps);

    fft->transform(in1.data(), IN1.data());
    fft->transform(in2.data(), IN2.data());

    bool isAlreadyAligned = true;
    for (size_t n = 1; n < Nf; n++) {
        phi = std::arg(IN1[n] * std::conj(IN2[n])); //TODO: unwrap?
        if (phi != 0.f) {
            isAlreadyAligned = false;
            if (phi > 0.0f) {
                IN1[n] *= std::exp(std::complex<float>(0.0f, -phi));
            } else {
                IN2[n] *= std::exp(std::complex<float>(0.0f, phi));
            }
        }
    }
    if (!isAlreadyAligned) {
        IN1.resize(Nf);
        IN2.resize(Nf);
        ifft->transform(IN1.data(), temp1.data());
        ifft->transform(IN2.data(), temp2.data());

        //determine circshift
        const float thres = 0.0001;

        float t = temp1.peak() * thres;
        size_t kl = 1;
        while (std::fabs(temp1[Nds - kl]) > t && kl < maxShift) {
            kl++;
        }
        t = temp2.peak() * thres;
        size_t kr = 1;
        while (std::fabs(temp2[Nds - kr]) > t && kr < maxShift) {
            kr++;
        }
        if (kr > kl) {
            kl = kr;
        }
        //circshift by kl-1 samples
        if (kl > 1) {
            phi = -2.f * M_PI / Nds * (kl - 1);
            for (size_t n = 1; n < Nf; n++) {
                IN1[n] *= std::exp((float)n * std::complex<float>(0.0f, phi));
                IN2[n] *= std::exp((float)n * std::complex<float>(0.0f, phi));
            }
            ifft->transform(IN1.data(), temp1.data());
            ifft->transform(IN2.data(), temp2.data());
        }
        ch1.write(temp1.data(), 0, Ntaps);
        ch1.mul(1.f / Nds);

        ch2.write(temp2.data(), 0, Ntaps);
        ch2.mul(1.f / Nds);
    }
}

void FIR2MinphaseFilterDesigner::applyHamming(RealBuffer& b)
{
    for (size_t n = 0; n < Ntaps; n++) {
        b[n] *= 0.54f - 0.46f * cos((2.0f * M_PI * n) / (Ntaps - 1));
    }
}

} // namespace EAR
