/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/FixedHRIRProvider.h>

#include <EAR/CartesianKDTree.h>
#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include "fabian_fixdef.h"

#include <functional>

namespace EAR {

namespace detail {

bool isPointerIsWithinBounds(const float* pointer, const float* begin, const float* end)
{
    return std::greater_equal<const float*>()(pointer, begin)
        && std::less_equal<const float*>()(pointer, end);
}

} // namespace detail

FixedHRIRProvider::FixedHRIRProvider()
    : N(FABIAN_DIFFEQ_N)
    , M(FABIAN_DIFFEQ_M)
    , fs(FABIAN_DIFFEQ_FS)
{
    cache = &FABIAN_DIFFEQ_HRIRS[0][0][0];
    endOfCache = &FABIAN_DIFFEQ_HRIRS[FABIAN_DIFFEQ_M - 1][1][FABIAN_DIFFEQ_N];
    directions = &FABIAN_DIFFEQ_POLAR_COORDS[0][0];

    sourcePoints.resize(M);
    for (unsigned int i = 0; i < M; i++) {
        Polar direction = Polar(directions[i * 3], directions[i * 3 + 1], directions[i * 3 + 2]);
        sourcePoints[i] = polarToCartesian(direction);
    }
    lookupTree = std::make_unique<EAR::CartesianKDTree>(sourcePoints);
}

FixedHRIRProvider::~FixedHRIRProvider() {}

void FixedHRIRProvider::getNearestIRForLocation(RealFrame& ir, Vector3 loc) const
{
    ir.resize(2, N);

    const unsigned long idx = getNearestIndex(loc);

    ir[0].write(getPointerToIndexForLeft(idx), 0, N);
    ir[1].write(getPointerToIndexForRight(idx), 0, N);
}

std::unique_ptr<const float* []> FixedHRIRProvider::getIRPointerForLocation(Vector3 loc) const {
    auto ir = std::make_unique<const float*[]>(2);
    const unsigned long idx = getNearestIndex(loc);

    ir[0] = getPointerToIndexForLeft(idx);
    ir[1] = getPointerToIndexForRight(idx);

    return ir;
}

std::unique_ptr<const float* []> FixedHRIRProvider::getIRPointerForIndex(int index) const {
    auto ir = std::make_unique<const float*[]>(2);
    ir[0] = getPointerToIndexForLeft(index);
    ir[1] = getPointerToIndexForRight(index);

    return ir;
}

std::vector<int> FixedHRIRProvider::getAllIndicesInRange(
    const EAR::Polar& bottomLeft, const EAR::Polar& topRight) const
{
    // TODO: this a crude way of implementing this, a kd tree range search would be much more efficient
    std::vector<int> result;

    for (unsigned int i = 0; i < M; i++) {
        if (isInAngularRange(directions[i * 3], bottomLeft[0], topRight[0])
            && isInAngularRange(directions[i * 3 + 1], bottomLeft[1], topRight[1])) {
            result.push_back(i);
        }
    }

    return result;
}

unsigned long FixedHRIRProvider::getNearestIndex(Vector3& loc) const
{
    return static_cast<unsigned long>(lookupTree->getNearestIndex(loc));
}

const float* FixedHRIRProvider::getPointerToIndexForLeft(std::size_t idx) const
{
    const float* pointerAtLeft = cache + 2 * N * idx;
    EAR_CHECK_REL(detail::isPointerIsWithinBounds(pointerAtLeft, cache, endOfCache),
        "Out of bounds error in HRIR access")
    return pointerAtLeft;
}

const float* FixedHRIRProvider::getPointerToIndexForRight(std::size_t idx) const
{
    const float* pointerAtRight = cache + 2 * N * idx + N;
    EAR_CHECK_REL(detail::isPointerIsWithinBounds(pointerAtRight, cache, endOfCache),
        "Out of bounds error in HRIR access")
    return pointerAtRight;
}

float FixedHRIRProvider::getDistanceToNearestIndex(Vector3& loc) const
{
    return lookupTree->getDistanceToNearestPoint(loc);
}

Vector3 FixedHRIRProvider::getNearestLocation(Vector3& loc) const
{
    return lookupTree->getNearestPoint(loc);
}

Vector3 FixedHRIRProvider::getLocationForIndex(int idx) const
{
    return sourcePoints[idx];
}

int FixedHRIRProvider::irLength() const
{
    return static_cast<int>(N);
}

int FixedHRIRProvider::irCount() const
{
    return static_cast<int>(M);
}

float FixedHRIRProvider::getSampleRate() const
{
    return static_cast<float>(fs);
}

} // namespace EAR
