/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/RealFrame.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>

#include <algorithm>
#include <complex>

namespace EAR {

template <typename T, class B>
Frame<T, B>::Frame()
    : Frame(0, 0)
{}

template <typename T, class B> Frame<T, B>::Frame(std::size_t channels, std::size_t length)
{
    resize(channels, length);
}

template <typename T, class B> Frame<T, B>::Frame(const Frame<T, B>& other)
{
    *this = other;
}

template <typename T, class B> Frame<T, B>::~Frame() {}

template <typename T, class B> Frame<T, B>& Frame<T, B>::operator=(const Frame<T, B>& other)
{
    EAR_CHECK(this != &other, "Self-assignment not allowed.");

    resize(other.channels(), other.length());
    std::copy(other.frame.begin(), other.frame.end(), frame.begin());

    bufPtrs.clear();
    for (auto& buffer : frame) {
        bufPtrs.push_back(buffer.data());
    }

    return *this;
}

/*template<typename T, class B>
Frame<T, B>& Frame<T, B>::operator=(Frame<T, B>&& other) {
    EAR_CHECK(this != &other, "Self-assignment not allowed.");

    //resize(other.channels(), other.length());
    std::move(other.frame->begin(), other.frame->end(), frame->begin());
    //std::copy(other.frame->begin(), other.frame->end(), frame->begin());

    return *this;
}*/

template <typename T, class B> void Frame<T, B>::mul(T c)
{
    for (auto& buffer : frame) {
        buffer.mul(c);
    }
}

template <typename T, class B> void Frame<T, B>::add(T c)
{
    for (auto& buffer : frame) {
        buffer.add(c);
    }
}

template <typename T, class B> void Frame<T, B>::lerp(const Frame<T, B>& tgt, float weight)
{
    EAR_CHECK(this->channels() == tgt.channels() && this->length() == tgt.length(),
        "Canot lerp incompatible frames");
    const size_t len = length();
    const size_t chan = channels();
    for (unsigned int ch = 0; ch < chan; ch++) {
        T* thisPtr = frame[ch].data();
        const T* tgtPtr = tgt[ch].data();
        for (int s = 0; s < len; s++)
            thisPtr[s] = thisPtr[s] * weight + tgtPtr[s] * (1.f - weight);
    }
}

template <typename T, class B>
std::unique_ptr<const T* []> Frame<T, B>::data() const {
    auto ptrs = std::make_unique<const T*[]>(frame.size());

    for (unsigned int ch = 0; ch < frame.size(); ch++) {
        ptrs[ch] = frame[ch].data();
    }

    return ptrs;
}

template <typename T, class B>
std::unique_ptr<T* []> Frame<T, B>::data() {
    auto ptrs = std::make_unique<T*[]>(frame.size());

    for (unsigned int ch = 0; ch < frame.size(); ch++) {
        ptrs[ch] = frame[ch].data();
    }

    return ptrs;
}

template <typename T, class B>
T** Frame<T, B>::dataRaw()
{
    return bufPtrs.data();
}

template <typename T, class B> const B& Frame<T, B>::operator[](std::size_t index) const
{
    return frame[index];
}

template <typename T, class B> B& Frame<T, B>::operator[](std::size_t index)
{
    return frame[index];
}

template <typename T, class B> const B& Frame<T, B>::ch(std::size_t index) const
{
    return frame[index];
}

template <typename T, class B> B& Frame<T, B>::ch(std::size_t index)
{
    return frame[index];
}

template <typename T, class B> void Frame<T, B>::copyFrom(const Frame<T, B>& src)
{
    resize(src.channels(), src.length());

    int i = 0;

    for (auto& buffer : frame) {
        buffer.write(src[i].data(), 0, length());
        i++;
    }
}

template <typename T, class B>
void Frame<T, B>::writeInterleaved(const T* src, std::size_t channels, std::size_t length)
{
    resize(channels, length);
    auto data = frame.data();

    for (std::size_t s = 0; s < length; s++) {
        for (std::size_t c = 0; c < channels; c++) {
            data[c][s] = *(src + (s * channels) + c);
        }
    }
}

template <typename T, class B>
void Frame<T, B>::readInterleaved(T* dst, std::size_t chOffset, std::size_t chCount,
    std::size_t smplOffset, std::size_t smplCount)
{
    auto data = frame.data();

    for (uint32_t s = static_cast<uint32_t>(smplOffset); s < (smplOffset + smplCount); s++) {
        for (uint32_t c = static_cast<uint32_t>(chOffset); c < (chOffset + chCount); c++) {
            *(dst) = data[c][s];
            dst++;
        }
    }
}

template <typename T, class B> void Frame<T, B>::shiftLeft(std::size_t count)
{
    for (auto& buffer : frame) {
        buffer.shiftLeft(count);
    }
}

template <typename T, class B> void Frame<T, B>::shiftRight(std::size_t count)
{
    for (auto& buffer : frame) {
        buffer.shiftRight(count);
    }
}

template <typename T, class B> void Frame<T, B>::shiftUp(std::size_t count)
{
    std::rotate(frame.begin(), frame.end() - count, frame.end());
}

template <typename T, class B> void Frame<T, B>::shiftDown(std::size_t count)
{
    std::rotate(frame.begin(), frame.begin() + count, frame.end());
}

template <typename T, class B>
void Frame<T, B>::resize(std::size_t newChannels, std::size_t newLength)
{
    frame.resize(newChannels);
    bufPtrs.reserve(newChannels);
    bufPtrs.clear();

    for (auto& buffer : frame) {
        buffer.resize(newLength);
        bufPtrs.push_back(buffer.data());
    }
}

template <typename T, class B> void Frame<T, B>::fill(T value)
{
    for (auto& buffer : frame) {
        buffer.fill(value);
    }
}

template <typename T, class B> void Frame<T, B>::fill(T value, std::size_t channel)
{
    frame[channel].fill(value);
}

template <typename T, class B>
void Frame<T, B>::fill(T value, std::size_t channel, std::size_t offset, std::size_t count)
{
    frame[channel].fill(value, offset, count);
}

template <typename T, class B> void Frame<T, B>::zero()
{
    for (auto& buffer : frame) {
        buffer.zero();
    }
}

template <typename T, class B> void Frame<T, B>::zero(std::size_t channel)
{
    frame[channel].zero();
}

template <typename T, class B>
void Frame<T, B>::zero(std::size_t channel, std::size_t offset, std::size_t count)
{
    frame[channel].zero(offset, count);
}

template <typename T, class B> std::size_t Frame<T, B>::length() const
{
    if (frame.size() > 0) {
        return frame[0].length();
    }

    return 0;
}

template <typename T, class B> std::size_t Frame<T, B>::channels() const
{
    return frame.size();
}

// allowed specializations:
template class Frame<float, RealBuffer>;
template class Frame<std::complex<float>, ComplexBuffer>;

} // namespace EAR
