/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Exception.h>
#include <EAR/GeometryPreprocessor.h>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES

#include <algorithm>
#include <cmath>
#include <iostream>
#include <vector>

namespace EAR {

// private method declarations

namespace GeometryPreprocessor {

//! extend "forward" edge graph
/*!
     * \param edge              input/output: forward graph
     * \param tid               input: triangle ID
     * \param vid1              input: first vertex ID
     * \param vid2              input: second vertex ID
     */
inline void addEdgeForward(std::vector<EdgeInfo>& edge, int tid, int vid1, int vid2);

//! extend "reverse" edge graph
/*!
     * \param edge              input/output: forward graph
     * \param tid               input: triangle ID
     * \param vid1              input: first vertex ID
     * \param vid2              input: second vertex ID
     */
inline void addEdgeReverse(std::vector<EdgeInfo>& edge, int tid, int vid1, int vid2);

//! compute angle between two triangles
/*!
     * \param vertex            input: vertex set
     * \param triangle          input: triangle data set
     * \param normal            input: set of normal vectors
     * \param from              input: first vertex ID
     * \param to                input: second vertex ID
     * \param left              input: left triangle ID
     * \param right             input: right triangle ID
     * \return                  angle in degree between two triangles
     */
inline float getEdgeAngle(const std::vector<Vector3>& vertex,
    const std::vector<Mesh::Triangle>& triangle, const std::vector<Vector3>& normal,
    unsigned int from, unsigned int to, unsigned int left, unsigned int right);

//! estimate radius between two adjacent triangles on a curved surface
/*!
     * \param vertex            input: vertex set
     * \param triangle          input: triangle data set
     * \param from              input: first vertex ID
     * \param to                input: second vertex ID
     * \param left              input: left triangle ID
     * \param right             input: right triangle ID
     * \param alpha             input: angle in degree between the triangles
     * \return                  angle in degree between two triangles
     */
float getEdgeRadius(const std::vector<Vector3>& vertex, const std::vector<Mesh::Triangle>& triangle,
    unsigned int from, unsigned int to, unsigned int left, unsigned int right, float alpha);

//! convert edge graphs into roundEdge and roundCorner data structures
/*!
     * \param roundEdges        output: list of round edges
     * \param roundCorners      output: list of round corners
     * \param edge_forward      input/output: forward graph
     * \param edge_reverse      input/output: reverse graph
     * \param vertex            input: vertex set
     * \param triangle          input: triangle data set
     * \param normal            input: set of noprmal vectors
     */
inline void convertEdge(std::vector<Mesh::RoundEdge>& roundEdges,
    std::vector<Mesh::RoundCorner>& roundCorners, std::vector<EdgeInfo>& edge_forward,
    std::vector<EdgeInfo>& edge_reverse, const std::vector<Vector3>& vertex,
    const std::vector<Mesh::Triangle>& triangle, const std::vector<Vector3>& normal);

} // namespace GeometryPreprocessor

/*********************************/
/*        public  methods        */
/*********************************/

void GeometryPreprocessor::getRoundStructure(std::vector<Mesh::RoundEdge>& roundEdges,
    std::vector<Mesh::RoundCorner>& roundCorners, const Mesh* mesh)
{

    auto triangle = mesh->getTriangles();
    auto vertex = mesh->getVertices(Mesh::CSpace::Local);

    // get normal vectors
    std::vector<Vector3> normal;
    for (auto& tri : triangle) {
        Vector3 N = tri.getNormal(Mesh::CSpace::Local);
        normal.push_back(N.normalize());
    }

    // memory allocation for forward and reverse graph
    auto num_vertex = vertex.size();
    std::vector<EdgeInfo> edge_forward(num_vertex);
    std::vector<EdgeInfo> edge_reverse(num_vertex);

    // initialize forward and reverse graph
    auto num_triangle = triangle.size();
    for (unsigned int tid = 0; tid < num_triangle; tid++) {
        std::vector<unsigned int> tri{ triangle[tid].v1, triangle[tid].v2, triangle[tid].v3,
            triangle[tid].v1 };
        for (unsigned int n = 0; n < 3; n++) {
            unsigned int vid1 = tri[n];
            unsigned int vid2 = tri[n + 1];
            addEdgeForward(edge_forward, tid, vid1, vid2);
            addEdgeReverse(edge_reverse, tid, vid1, vid2);
        }
    }

    // convert forward and reverse graph to list of rounded edges and corners
    convertEdge(roundEdges, roundCorners, edge_forward, edge_reverse, vertex, triangle, normal);
}

/*********************************/
/*        private methods        */
/*********************************/

void GeometryPreprocessor::addEdgeForward(std::vector<EdgeInfo>& edge, int tid, int vid1, int vid2)
{

    int vid_min = std::min(vid1, vid2);
    int vid_max = std::max(vid1, vid2);

    unsigned int num_entry = edge[vid_min].entry.size();
    unsigned int found = 0;
    for (found = 0; found < num_entry; found++) {
        if (edge[vid_min].entry[found] == vid_max)
            break;
    }
    if (found == num_entry) {
        edge[vid_min].left.push_back(-1);
        edge[vid_min].right.push_back(-1);
        edge[vid_min].entry.push_back(vid_max);
    }

    if (vid2 > vid1) {
        edge[vid_min].left[found] = tid;
    } else {
        edge[vid_min].right[found] = tid;
    }
}

void GeometryPreprocessor::addEdgeReverse(std::vector<EdgeInfo>& edge, int tid, int vid1, int vid2)
{

    int vid_min = std::min(vid1, vid2);
    int vid_max = std::max(vid1, vid2);

    unsigned int num_entry = edge[vid_max].entry.size();
    unsigned int found = 0;
    for (found = 0; found < num_entry; found++) {
        if (edge[vid_max].entry[found] == vid_min)
            break;
    }
    if (found == num_entry) {
        edge[vid_max].left.push_back(-1);
        edge[vid_max].right.push_back(-1);
        edge[vid_max].entry.push_back(vid_min);
    }

    if (vid2 > vid1) {
        edge[vid_max].right[found] = tid;
    } else {
        edge[vid_max].left[found] = tid;
    }
}

float GeometryPreprocessor::getEdgeAngle(const std::vector<Vector3>& vertex,
    const std::vector<Mesh::Triangle>& triangle, const std::vector<Vector3>& normal,
    unsigned int from, unsigned int to, unsigned int left, unsigned int right)
{

    Vector3 N_left = normal[left];
    Vector3 N_right = normal[right];
    Vector3 A = vertex[triangle[left].v1];
    float d_left = N_left.dot(A);

    float d_test = 0.f;
    std::vector<unsigned int> tri = { triangle[right].v1, triangle[right].v2, triangle[right].v3 };
    for (auto vid : tri) {
        if ((vid != from) && (vid != to)) {
            d_test = vertex[vid].dot(N_left) - d_left;
            break;
        }
    }

    float alpha = acos(N_left.dot(N_right)) * EAR_SGN(-d_test) * 180 / M_PI;

    return alpha;
}

float GeometryPreprocessor::getEdgeRadius(const std::vector<Vector3>& vertex,
    const std::vector<Mesh::Triangle>& triangle, unsigned int from, unsigned int to,
    unsigned int left, unsigned int right, float alpha)
{
    Vector3 AB0 = (vertex[to] - vertex[from]).normalize();

    float l_left = 0.f;
    std::vector<unsigned int> tri = { triangle[left].v1, triangle[left].v2, triangle[left].v3 };
    for (auto vid : tri) {
        if ((vid != from) && (vid != to)) {
            Vector3 AC = vertex[vid] - vertex[from];
            Vector3 AP = AB0 * AB0.dot(AC);
            l_left = (AC - AP).abs();
            break;
        }
    }

    float l_right = 0.f;
    tri = { triangle[right].v1, triangle[right].v2, triangle[right].v3 };
    for (auto vid : tri) {
        if ((vid != from) && (vid != to)) {
            Vector3 AC = vertex[vid] - vertex[from];
            Vector3 AP = AB0 * AB0.dot(AC);
            l_right = (AC - AP).abs();
            break;
        }
    }

    float radius = std::min(l_left, l_right) / (std::abs(alpha) * M_PI / 180.f);

    return radius;
}

void GeometryPreprocessor::convertEdge(std::vector<Mesh::RoundEdge>& roundEdges,
    std::vector<Mesh::RoundCorner>& roundCorners, std::vector<EdgeInfo>& edge_forward,
    std::vector<EdgeInfo>& edge_reverse, const std::vector<Vector3>& vertex,
    const std::vector<Mesh::Triangle>& triangle, const std::vector<Vector3>& normal)
{

    // first part: extract all rounded edges and count them for the forward graph
    for (unsigned int from = 0; from < edge_forward.size(); from++) {
        unsigned int num_entries_forward = edge_forward[from].entry.size();
        for (unsigned int n = 0; n < num_entries_forward; n++) {
            int to = edge_forward[from].entry[n];
            int left = edge_forward[from].left[n];
            int right = edge_forward[from].right[n];
            if ((left != -1) && (right != -1)) {
                float alpha = getEdgeAngle(vertex, triangle, normal, from, to, left, right);
                if ((alpha >= EAR_ROUND_EDGE_ALPHA_MIN) && (alpha <= EAR_ROUND_EDGE_ALPHA_MAX)) {
                    float radius = getEdgeRadius(vertex, triangle, from, to, left, right, alpha);
                    if (radius >= EAR_ROUND_EDGE_RADIUS_MIN) {
                        roundEdges.push_back(Mesh::RoundEdge(from, to, left, right, alpha));
                        edge_forward[from].num_round_edges++;
                    }
                }
            }
        }
    }

    // second part: count rounded edges of the reverse graph
    for (unsigned int from = 0; from < edge_reverse.size(); from++) {
        unsigned int num_entries_forward = edge_reverse[from].entry.size();
        for (unsigned int n = 0; n < num_entries_forward; n++) {
            int to = edge_reverse[from].entry[n];
            int left = edge_reverse[from].left[n];
            int right = edge_reverse[from].right[n];
            if ((left != -1) && (right != -1)) {
                float alpha = getEdgeAngle(vertex, triangle, normal, from, to, left, right);
                if ((alpha >= EAR_ROUND_EDGE_ALPHA_MIN) && (alpha <= EAR_ROUND_EDGE_ALPHA_MAX)) {
                    float radius = getEdgeRadius(vertex, triangle, from, to, left, right, alpha);
                    if (radius >= EAR_ROUND_EDGE_RADIUS_MIN) {
                        edge_reverse[from].num_round_edges++;
                    }
                }
            }
        }
    }

    // extract rounded corners
    for (unsigned int from = 0; from < edge_forward.size(); from++) {
        unsigned int num_entries_forward = edge_forward[from].entry.size();
        unsigned int num_entries_reverse = edge_reverse[from].entry.size();
        if ((num_entries_forward + num_entries_reverse > 0)
            && (edge_forward[from].num_round_edges == num_entries_forward)
            && (edge_reverse[from].num_round_edges
                == num_entries_reverse)) { // all edges connected to this vertex are rounded

            // left_list = {edge_forward[from].left, edge_reverse[from].left};
            auto left_list = edge_forward[from].left;
            left_list.insert(
                left_list.end(), edge_reverse[from].left.begin(), edge_reverse[from].left.end());

            // right_list = {edge_forward[from].right, edge_reverse[from].right};
            auto right_list = edge_forward[from].right;
            right_list.insert(
                right_list.end(), edge_reverse[from].right.begin(), edge_reverse[from].right.end());

            // generate tid_list
            unsigned int num_tid = left_list.size();
            std::vector<unsigned int> tid_list;
            unsigned int pos = 0;
            int tid = left_list[pos];
            for (unsigned int k = 0; k < num_tid; k++) {
                tid_list.push_back(tid);
                tid = right_list[pos];
                for (pos = 0; pos < left_list.size(); pos++) {
                    if (left_list[pos] == tid)
                        break;
                }
            }

            // add round corner to roundCorners list
            roundCorners.push_back(Mesh::RoundCorner(from, tid_list));
        }
    }
}

/*********************************/
/*         print methods         */
/*********************************/

std::ostream& operator<<(std::ostream& os, GeometryPreprocessor::EdgeInfo& e)
{

    os << "left:  " << e.left << std::endl;
    os << "right: " << e.right << std::endl;
    os << "entry: " << e.entry << std::endl;
    os << "num_round_edges: " << e.num_round_edges << std::endl;

    return os;
}

} // namespace EAR
