/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#include <EAR/Biquad.h>
#include <EAR/FIR2MinphaseFilterDesigner.h>
#include <EAR/FIRFilter.h>
#include <EAR/HeadphoneEQ.h>

#include <numeric>
#include <vector>

namespace EAR {
std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> HeadphoneEQ::calcFilters(
    const HeadphoneEQData& targetEQ, const HeadphoneEQData& measuredEQ, float fs, float B,
    const size_t Ntaps)
{
    EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

    EAR::RealBuffer targetFreqs;
    EAR::RealBuffer targetGains;
    int nTargetFreqs;
    std::vector<float> freqs;
    std::vector<float> gains;

    HeadphoneEQPresets presets;

    switch (targetEQ.specType) {
    case HeadphoneSpecType::freqgain: {
        nTargetFreqs = static_cast<int>(targetEQ.centerFreq.size());
        gains = convertGainsToLin(targetEQ.gain);
        freqs = targetEQ.centerFreq;
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nTargetFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(targetEQ, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    case HeadphoneSpecType::preset: {
        const auto [calcfreqs, calcgains] = presets.getPreset(targetEQ.preset);
        freqs = calcfreqs;
        nTargetFreqs = freqs.size();
        gains = convertGainsToLin(calcgains);
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown Headphone spec type given");
        break;
    }
    }

    targetFreqs.resize(nTargetFreqs);
    targetGains.resize(nTargetFreqs);
    targetFreqs.write(freqs.data(), 0, nTargetFreqs);
    targetGains.write(gains.data(), 0, nTargetFreqs);

    EAR::RealBuffer measuredFreqs;
    EAR::RealBuffer measuredGains;
    int nMeasuredFreqs;
    freqs.clear();
    gains.clear();

    switch (measuredEQ.specType) {
    case HeadphoneSpecType::freqgain: {
        nMeasuredFreqs = static_cast<int>(measuredEQ.centerFreq.size());
        freqs = measuredEQ.centerFreq;
        gains = convertGainsToLin(measuredEQ.gain);
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nMeasuredFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(measuredEQ, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    case HeadphoneSpecType::preset: {
        const auto [calcfreqs, calcgains] = presets.getPreset(measuredEQ.preset);
        freqs = calcfreqs;
        nMeasuredFreqs = freqs.size();
        gains = convertGainsToLin(calcgains);
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown Headphone spec type given");
        break;
    }
    }

    measuredFreqs.resize(nMeasuredFreqs);
    measuredGains.resize(nMeasuredFreqs);
    measuredFreqs.write(freqs.data(), 0, nMeasuredFreqs);
    measuredGains.write(gains.data(), 0, nMeasuredFreqs);

    EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
    fdsgn.design(measuredFreqs, measuredGains, targetFreqs, targetGains, b);

    // create
    std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> filters;
    filters.first = std::make_unique<EAR::FIRFilter>(b.data(), B, Ntaps);
    filters.second = std::make_unique<EAR::FIRFilter>(b.data(), B, Ntaps);
    return filters;
}

std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> HeadphoneEQ::calcFilters(
    const HeadphoneEQData& targetEQ, const HeadphoneEQData& measuredEQ,
    const HeadphoneEQData& spectralCompensation, float fs, float B, const size_t Ntaps)
{
    EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

    EAR::RealBuffer targetFreqs;
    EAR::RealBuffer targetGains;
    EAR::RealBuffer spectralCompensationFreqs;
    EAR::RealBuffer spectralCompensationGains;
    int nFreqs;
    std::vector<float> freqs;
    std::vector<float> gains;

    HeadphoneEQPresets presets;

    switch (targetEQ.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(targetEQ.centerFreq.size());
        gains = convertGainsToLin(targetEQ.gain);
        freqs = targetEQ.centerFreq;
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(targetEQ, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    case HeadphoneSpecType::preset: {
        const auto [calcfreqs, calcgains] = presets.getPreset(targetEQ.preset);
        freqs = calcfreqs;
        nFreqs = freqs.size();
        gains = convertGainsToLin(calcgains);
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown Headphone spec type given");
        break;
    }
    }

    targetFreqs.resize(nFreqs);
    targetGains.resize(nFreqs);
    targetFreqs.write(freqs.data(), 0, nFreqs);
    targetGains.write(gains.data(), 0, nFreqs);

    freqs.clear();
    gains.clear();

    switch (spectralCompensation.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(spectralCompensation.centerFreq.size());
        gains = convertGainsToLin(spectralCompensation.gain, false);
        freqs = spectralCompensation.centerFreq;
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(spectralCompensation, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown compensation spec type given");
        break;
    }
    }
    spectralCompensationFreqs.resize(nFreqs);
    spectralCompensationGains.resize(nFreqs);
    spectralCompensationFreqs.write(freqs.data(), 0, nFreqs);
    spectralCompensationGains.write(gains.data(), 0, nFreqs);

    EAR::RealBuffer measuredFreqs;
    EAR::RealBuffer measuredGains;

    freqs.clear();
    gains.clear();

    switch (measuredEQ.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(measuredEQ.centerFreq.size());
        freqs = measuredEQ.centerFreq;
        gains = convertGainsToLin(measuredEQ.gain);
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(measuredEQ, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    case HeadphoneSpecType::preset: {
        const auto [calcfreqs, calcgains] = presets.getPreset(measuredEQ.preset);
        freqs = calcfreqs;
        nFreqs = freqs.size();
        gains = convertGainsToLin(calcgains);
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown Headphone spec type given");
        break;
    }
    }

    measuredFreqs.resize(nFreqs);
    measuredGains.resize(nFreqs);
    measuredFreqs.write(freqs.data(), 0, nFreqs);
    measuredGains.write(gains.data(), 0, nFreqs);

    EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
    //fdsgn.design(measuredFreqs, measuredGains, targetFreqs, targetGains, b);
    fdsgn.design(measuredFreqs, measuredGains, targetFreqs, targetGains, spectralCompensationFreqs,
        spectralCompensationGains, b);

    // create
    std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> filters;
    filters.first = std::make_unique<EAR::FIRFilter>(b.data(), B, Ntaps);
    filters.second = std::make_unique<EAR::FIRFilter>(b.data(), B, Ntaps);
    return filters;
}

std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> HeadphoneEQ::calcFilters(
    const HeadphoneEQData& targetEQ, const HeadphoneEQData& measuredEQ,
    const HeadphoneEQData& spectralCompensationLeft,
    const HeadphoneEQData& spectralCompensationRight, float fs, float B, const size_t Ntaps)
{
    EAR::RealBuffer bLeft = EAR::RealBuffer(Ntaps);
    EAR::RealBuffer bRight = EAR::RealBuffer(Ntaps);

    EAR::RealBuffer targetFreqs;
    EAR::RealBuffer targetGains;
    EAR::RealBuffer spectralCompensationFreqs;
    EAR::RealBuffer spectralCompensationGains;
    int nFreqs;
    std::vector<float> freqs;
    std::vector<float> gains;

    HeadphoneEQPresets presets;

    switch (targetEQ.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(targetEQ.centerFreq.size());
        gains = convertGainsToLin(targetEQ.gain);
        freqs = targetEQ.centerFreq;
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(targetEQ, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    case HeadphoneSpecType::preset: {
        const auto [calcfreqs, calcgains] = presets.getPreset(targetEQ.preset);
        freqs = calcfreqs;
        nFreqs = freqs.size();
        gains = convertGainsToLin(calcgains);
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown Headphone spec type given");
        break;
    }
    }

    targetFreqs.resize(nFreqs);
    targetGains.resize(nFreqs);
    targetFreqs.write(freqs.data(), 0, nFreqs);
    targetGains.write(gains.data(), 0, nFreqs);

    freqs.clear();
    gains.clear();

    switch (spectralCompensationLeft.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(spectralCompensationLeft.centerFreq.size());
        gains = convertGainsToLin(spectralCompensationLeft.gain, false);
        freqs = spectralCompensationLeft.centerFreq;
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(spectralCompensationLeft, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown compensation spec type given");
        break;
    }
    }
    spectralCompensationFreqs.resize(nFreqs);
    spectralCompensationGains.resize(nFreqs);
    spectralCompensationFreqs.write(freqs.data(), 0, nFreqs);
    spectralCompensationGains.write(gains.data(), 0, nFreqs);

    EAR::RealBuffer measuredFreqs;
    EAR::RealBuffer measuredGains;
    freqs.clear();
    gains.clear();

    switch (measuredEQ.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(measuredEQ.centerFreq.size());
        freqs = measuredEQ.centerFreq;
        gains = convertGainsToLin(measuredEQ.gain);
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] = getGainsFromBiquad(measuredEQ, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    case HeadphoneSpecType::preset: {
        const auto [calcfreqs, calcgains] = presets.getPreset(measuredEQ.preset);
        freqs = calcfreqs;
        nFreqs = freqs.size();
        gains = convertGainsToLin(calcgains);
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown Headphone spec type given");
        break;
    }
    }

    measuredFreqs.resize(nFreqs);
    measuredGains.resize(nFreqs);
    measuredFreqs.write(freqs.data(), 0, nFreqs);
    measuredGains.write(gains.data(), 0, nFreqs);

    EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
    fdsgn.design(measuredFreqs, measuredGains, targetFreqs, targetGains, spectralCompensationFreqs,
        spectralCompensationGains, bLeft);

    //Right Ear
    freqs.clear();
    gains.clear();
    switch (spectralCompensationRight.specType) {
    case HeadphoneSpecType::freqgain: {
        nFreqs = static_cast<int>(spectralCompensationRight.centerFreq.size());
        gains = convertGainsToLin(spectralCompensationRight.gain, false);
        freqs = spectralCompensationLeft.centerFreq;
        break;
    }
    case HeadphoneSpecType::parametricfilters: {
        nFreqs = (int)Ntaps / 2;
        const auto [calcfreqs, calcgains] =
            getGainsFromBiquad(spectralCompensationRight, fs, Ntaps);
        freqs = calcfreqs;
        gains = calcgains;
        break;
    }
    default: {
        // should never reach here.
        EAR_FAIL("Unknown compensation spec type given");
        break;
    }
    }
    spectralCompensationFreqs.resize(nFreqs);
    spectralCompensationGains.resize(nFreqs);
    spectralCompensationFreqs.write(freqs.data(), 0, nFreqs);
    spectralCompensationGains.write(gains.data(), 0, nFreqs);

    fdsgn.design(measuredFreqs, measuredGains, targetFreqs, targetGains, spectralCompensationFreqs,
        spectralCompensationGains, bRight);

    //Phase Alignment
    fdsgn.phaseAlign(bLeft, bRight, (size_t)Ntaps / 2);

    // create
    std::pair<std::unique_ptr<FIRFilter>, std::unique_ptr<FIRFilter>> filters;
    filters.first = std::make_unique<EAR::FIRFilter>(bLeft.data(), B, Ntaps);
    filters.second = std::make_unique<EAR::FIRFilter>(bRight.data(), B, Ntaps);
    return filters;
}

std::vector<float> HeadphoneEQ::convertGainsToLin(
    std::vector<float> dBGains, const bool substractAverage)
{
    std::vector<float> linGains;
    linGains.reserve(dBGains.size());
    if (substractAverage) {
        const float average = std::accumulate(dBGains.begin(), dBGains.end(), 0.0) / dBGains.size();
        for (const float& dbVal : dBGains) {
            linGains.push_back(std::pow(10.0F, 0.05F * (dbVal - average)));
        }
    } else {
        for (const float& dbVal : dBGains) {
            linGains.push_back(std::pow(10.0F, 0.05F * dbVal));
        }
    }

    return linGains;
}

std::tuple<std::vector<float>, std::vector<float>> HeadphoneEQ::getGainsFromBiquad(
    const HeadphoneEQData& filterData, float fs, float nTargetFreqs)
{
    const float step = fs / 2 / (nTargetFreqs - 1);
    std::vector<float> freqs;
    std::vector<float> gains;

    std::vector<Biquad> filters;

    for (int i = 0; i < nTargetFreqs; ++i) {
        freqs.push_back(i * step);
    }

    for (int i = 0; i < filterData.centerFreq.size(); ++i) {
        switch (filterData.filterType[i]) {
        case HeadphoneFilterType::lowshelf: {
            filters.push_back(Biquad(bq_type_lowshelf, filterData.centerFreq[i] / fs,
                filterData.q[i], filterData.gain[i]));
            break;
        }
        case HeadphoneFilterType::peak: {
            filters.push_back(Biquad(
                bq_type_peak, filterData.centerFreq[i] / fs, filterData.q[i], filterData.gain[i]));
            break;
        }
        case HeadphoneFilterType::highshelf: {
            filters.push_back(Biquad(bq_type_highshelf, filterData.centerFreq[i] / fs,
                filterData.q[i], filterData.gain[i]));
            break;
        }
        }
    }

    for (const auto& f : freqs) {
        gains.push_back(filters[0].calcResponse(f / fs));
    }

    for (int f = 1; f < filters.size(); ++f) {
        for (int freqIdx = 0; freqIdx < nTargetFreqs; ++freqIdx) {
            gains[freqIdx] *= filters[f].calcResponse(freqs[freqIdx] / fs);
        }
    }

    return { freqs, gains };
}

} // namespace EAR
