/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/IIRFilter.h>

#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>

#include <algorithm>

namespace EAR {

class IIRFilter_fallback : public IIRFilter
{
public:
    IIRFilter_fallback(const float* _b, int nb, const float* _a, int na)
    {
        // if numerator or denominator is longer, the shorter is zero-padded to make process easier
        b.resize(std::max(nb, na));
        a.resize(std::max(nb, na));
        z.resize(std::max(nb, na) + 1);

        // fill b/a coefficients and normalize to a[0] = 1
        for (auto i = 0; i < nb; i++) {
            b[i] = _b[i] / _a[0];
        }

        for (auto i = 0; i < na; i++) {
            a[i] = _a[i] / _a[0];
        }

        bLen = b.length();
    }

    void clear() { z.zero(); }

    // transposed direct form II
    void process(const float* input, float* output, int nsamples)
    {
        // operate on raw pointers to be much faster by allowing compiler optimizations
        float* zptr = z.data();
        float* bptr = b.data();
        float* aptr = a.data();

        // iterate over the samples
        for (auto n = 0; n < nsamples; n++) {

            // new input + old state
            // saved in temporary var to allow inputptr = outputptr
            zptr[0] = bptr[0] * input[n] + zptr[1];

            // add to states
            // new state = b * x + a * y + previous state of next step
            // z[0] is the next output value
            // last z is always 0
            for (unsigned int i = 1; i < bLen; i++) {
                zptr[i] = bptr[i] * input[n] - aptr[i] * zptr[0] + zptr[i + 1];
            }

            output[n] = zptr[0];
        }
    }

private:
    //! Filter Length
    std::size_t bLen;

    //! Filter coefficients
    RealBuffer b, a;

    //! Accumulators
    RealBuffer z;
};

} // namespace EAR
