/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/LinearInterpolator.h>

namespace EAR {

template <typename Tx, typename Ty>
LinearInterpolator<Tx, Ty>::LinearInterpolator(const Tx* x, const Ty* y, std::size_t N)
{
    initialize(x, y, N);
}

template <typename Tx, typename Ty>
void LinearInterpolator<Tx, Ty>::initialize(const Tx* x, const Ty* y, std::size_t N)
{
    xs.resize(N);
    std::copy(x, x + N, xs.data());
    ys.resize(N);
    std::copy(y, y + N, ys.data());
    shouldInterpolate.resize(N, true);
}

template <typename Tx, typename Ty> void LinearInterpolator<Tx, Ty>::addSupportingPoint(Tx x, Ty y)
{
    std::vector<Tx> newxs = xs;
    std::vector<Ty> newys = ys;

    // if x already exists in the sorted vector, replace the corresponding y
    if (std::binary_search(newxs.begin(), newxs.end(), x)) {
        int idx = std::find(newxs.begin(), newxs.end(), x) - newxs.begin();
        newys[idx] = y;
        shouldInterpolate[idx] = true;
    } else { // else find new spot for x in newxs and insert into newxs and newys accordingly
        int idx = std::lower_bound(newxs.begin(), newxs.end(),
                      x)
            - newxs.begin(); // taking the detour over index to use it for newys too
        newxs.insert(newxs.begin() + idx, x);
        newys.insert(newys.begin() + idx, y);
        shouldInterpolate.insert(shouldInterpolate.begin() + idx, true);
    }

    initialize(newxs.data(), newys.data(), newxs.size());
};

template <typename Tx, typename Ty>
void LinearInterpolator<Tx, Ty>::addSupportingPoint(Tx x, Ty y, bool _interpolate)
{
    std::vector<Tx> newxs = xs;
    std::vector<Ty> newys = ys;

    // if x already exists in the sorted vector, replace the corresponding y
    if (std::binary_search(newxs.begin(), newxs.end(), x)) {
        int idx = std::find(newxs.begin(), newxs.end(), x) - newxs.begin();
        newys[idx] = y;
        shouldInterpolate[idx] = _interpolate;
    } else { // else find new spot for x in newxs and insert into newxs and newys accordingly
        int idx = std::lower_bound(newxs.begin(), newxs.end(),
                      x)
            - newxs.begin(); // taking the detour over index to use it for newys too
        newxs.insert(newxs.begin() + idx, x);
        newys.insert(newys.begin() + idx, y);
        shouldInterpolate.insert(shouldInterpolate.begin() + idx, _interpolate);
    }

    initialize(newxs.data(), newys.data(), newxs.size());
};

template <typename Tx, typename Ty> Ty LinearInterpolator<Tx, Ty>::interpolate(Tx x)
{
    if (ys.size() == 0) {
        return Ty();
    } else if (ys.size() == 1) {
        return ys[0];
    }

    auto a = std::lower_bound(xs.begin(), xs.end(), x);
    int idx = a - xs.begin();

    if (idx == 0) { // x is smaller or equal to the first x supporting point
        return ys[0];
    }

    if (idx >= static_cast<int>(ys.size())) { // x is bigger or equal to the last x supporting point
        return ys.back();
    }

    if (*a == x) { // x is exactly on a supporting point
        return ys[idx];
    }

    if (!shouldInterpolate[idx]) { // should not interpolate here
        return ys[idx - 1];
    }

    // all edge cases considered, we now do the actual linear interpolation
    double t = static_cast<double>(x - xs[idx - 1]) / static_cast<double>(xs[idx] - xs[idx - 1]);
    return (1 - t) * ys[idx - 1] + t * ys[idx];
};

template <typename Tx, typename Ty> bool LinearInterpolator<Tx, Ty>::isChanging(Tx x, Tx px)
{

    auto a = std::lower_bound(xs.begin(), xs.end(), x);
    int idx = a - xs.begin();

    if (idx == 0) { // x is smaller or equal to the first x supporting point
        return shouldInterpolate[0];
    }

    bool stepHappened = false;
    if (px != 0 && px < x) {
        Ty y1 = interpolate(x);
        Ty y2 = interpolate(px);

        stepHappened = (y1 != y2);
    }

    if (idx >= static_cast<int>(ys.size())) { // x is bigger or equal to the last x supporting point
        return shouldInterpolate.back() || stepHappened;
    }

    return shouldInterpolate[idx] || stepHappened;
};

template <typename Tx, typename Ty> Ty LinearInterpolator<Tx, Ty>::getSlope(Tx x)
{
    if (!isChanging(x)) {
        return Ty(); // zero element
    }

    auto a = std::lower_bound(xs.begin(), xs.end(), x);

    if (a == xs.begin()) {
        return Ty();
    }

    a -= 1;
    auto b = std::upper_bound(xs.begin(), xs.end(), x);

    if (b == xs.end() || a == b) {
        return Ty();
    }

    int ai = a - xs.begin();
    int bi = b - xs.begin();

    return static_cast<Ty>((ys[bi] - ys[ai]) * (1 / (xs[bi] - xs[ai])));
};

// Explicit specialization for slerp interpolation of Quaternions
template <> Quaternion LinearInterpolator<double, Quaternion>::interpolate(double x)
{
    if (ys.size() == 0) {
        return Quaternion();
    } else if (ys.size() == 1) {
        return ys[0];
    }

    auto a = std::lower_bound(xs.begin(), xs.end(), x);
    int idx = a - xs.begin();

    if (idx == 0) { // x is smaller or equal to the first x supporting point
        return ys[0];
    }

    if (idx >= static_cast<int>(ys.size())) { // x is bigger or equal to the last x supporting point
        return ys.back();
    }

    if (*a == x) { // x is exactly on a supporting point
        return ys[idx];
    }

    if (!shouldInterpolate[idx]) { // should not interpolate here
        return ys[idx - 1];
    }

    // all edge cases considered, we now do the actual slerp interpolation
    // https://web.mit.edu/2.998/www/QuaternionReport1.pdf
    double t = static_cast<double>(x - xs[idx - 1]) / static_cast<double>(xs[idx] - xs[idx - 1]);
    return slerp(ys[idx - 1], ys[idx], t);
};

// allowed specialisations
template class LinearInterpolator<int, long>;
template class LinearInterpolator<float, float>;
template class LinearInterpolator<double, float>;
template class LinearInterpolator<double, Vector3>;
template class LinearInterpolator<double, Quaternion>;

} // namespace EAR
