/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Log.h>

#include <fstream>
#include <sstream>

#include <spdlog/async.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/sinks/stdout_color_sinks.h>
#include <spdlog/spdlog.h>

#include <EAR/Exception.h>
#include <EAR/config.h>

namespace EAR {

Log* Log::instance = nullptr;

Log::Log()
{
#ifdef EAR_LOG_LEVEL_DEBUG
    spdlog::set_level(spdlog::level::debug);
#endif
#ifdef EAR_LOG_LEVEL_INFO
    spdlog::set_level(spdlog::level::info);
#endif
#ifdef EAR_LOG_LEVEL_WARN
    spdlog::set_level(spdlog::level::warn);
#endif
#ifdef EAR_LOG_LEVEL_ERROR
    spdlog::set_level(spdlog::level::err);
#endif

    consoleLogger = spdlog::stdout_color_mt<spdlog::async_factory>("console");
}

Log* Log::getInstance()
{
    if (instance == nullptr) {
        instance = new Log;
    }

    return instance;
}

void Log::log(const std::string& message, Log::Level logLevel)
{
    switch (logLevel) {
    case Log::Level::DBG:
        consoleLogger->debug(message);
        if (fileLogger) {
            fileLogger->debug(message);
        }
        break;
    case Log::Level::INFO:
        consoleLogger->info(message);
        if (fileLogger) {
            fileLogger->info(message);
        }
        break;
    case Log::Level::WARN:
        consoleLogger->warn(message);
        if (fileLogger) {
            fileLogger->warn(message);
        }
        break;
    case Log::Level::ERR:
        consoleLogger->error(message);
        if (fileLogger) {
            fileLogger->error(message);
        }
        break;
    };
}

void Log::enableFileLogger(const std::string& path)
{
    try {
        fileLogger = spdlog::basic_logger_mt<spdlog::async_factory>("file", path);

        // If file already exists, clear it
        if (std::ifstream(path).good()) {
            std::ofstream ofs;
            ofs.open(path, std::ofstream::out | std::ofstream::trunc);
            ofs.close();
        }

    } catch (const spdlog::spdlog_ex& exc) {
        std::stringstream sstr;
        sstr << "Log init failed: " << exc.what();
        EAR_FAIL(sstr.str());
    }
}

} // namespace EAR
