/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/LoudspeakerLayout.h>

#include <EAR/CICP_Layouts.h>
#include <EAR/Exception.h>

#include <map>

namespace EAR {

LoudspeakerLayout::LoudspeakerLayout(
    ObjectID oid, int layout, float radius, std::string name, Position pos)
    : SceneObject(oid, name, pos)
    , layout(layout)
    , radius(radius)
{
    EAR_CHECK_REL(detail::CICP_LAYOUTS.find(layout) != detail::CICP_LAYOUTS.end(),
        "Unknown CICP layout with id " + std::to_string(layout) + ".");

    constructLoudspeakers();
}

const std::vector<std::shared_ptr<PointSource>>&
LoudspeakerLayout::getLoudspeakerPointSources() const
{
    return loudspeakers;
}

void LoudspeakerLayout::setActive(bool newState)
{
    for (auto& ls : loudspeakers) {
        ls->setActive(newState);
    }
    SceneObject::setActive(newState);
}

void LoudspeakerLayout::setRadius(float radius)
{
    this->radius = radius;

    constructLoudspeakers();
}

float LoudspeakerLayout::getRadius() const
{
    return radius;
}

int LoudspeakerLayout::getLayoutIndex() const
{
    return layout;
}

void LoudspeakerLayout::setVDLMethod(VDLMethod vdlMethod)
{
    this->vdlMethod = vdlMethod;
}

inline void LoudspeakerLayout::constructLoudspeakers()
{
    loudspeakers.clear();

    int i = 1;
    for (const auto& speakerPolar : detail::CICP_LAYOUTS.at(layout)) {

        loudspeakers.push_back(
            SceneObjectFactory<PointSource>::getNewObject(getName() + "_LS" + std::to_string(i++)));
        loudspeakers.back()->setCSpace(CSpace::RCS, this);

        // Set location according to the polar coordinates from the CICP Layouts with specified radius
        loudspeakers.back()->setLocation(
            polarToCartesian(Polar(speakerPolar[0], speakerPolar[1], radius)));

        // Reorient the speaker to point to the centre
        Polar relativeOrient =
            cartesianToPolar(loudspeakers.back()->getRelativeLocation(Vector3(0, 0, 0)));
        loudspeakers.back()->setOrientation(Euler(relativeOrient[0], relativeOrient[1], 0));
    }
}

void LoudspeakerLayout::notify(SceneObjectObserver::Property mod)
{
    if (mod != SceneObjectObserver::Property::Position) {
        // this is already handled by the transform tree for position attributes
        for (auto& ls : loudspeakers) {
            ls->notify(mod);
        }
    }
    SceneObject::notify(mod);
}

void LoudspeakerLayout::setGainDb(float newGain)
{
    for (auto& ls : loudspeakers) {
        ls->setGainDb(newGain);
    }
}

void LoudspeakerLayout::setReverbGain(float newItemReverbGain)
{
    // clamp reverbGain value between -119 and +12 dB
    if (newItemReverbGain > 3.9810717055349725077) {
        newItemReverbGain = 3.9810717055349725077;
    }
    if (newItemReverbGain < 0.00000112201845430196) {
        newItemReverbGain = 0.00000112201845430196;
    }
    for (auto& ls : loudspeakers) {
        ls->setReverbGain(newItemReverbGain);
    }
}

void LoudspeakerLayout::setNoReverb(bool newNoReverb)
{
    for (auto& ls : loudspeakers) {
        ls->setNoReverb(newNoReverb);
    }
}

} // namespace EAR
