/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Mesh.h>

#include <EAR/CartesianKDTree.h>
#include <EAR/GeometryPreprocessor.h>

#include <algorithm>
#include <cmath>
#include <limits>
#include <numeric>
#include <set>

namespace EAR {

Vector3 Mesh::Triangle::getNormal(CSpace coordinateSpace) const
{
    const Vector3& v1l = parent->getVertex(v1, coordinateSpace);
    return (parent->getVertex(v2, coordinateSpace) - v1l)
        * (parent->getVertex(v3, coordinateSpace) - v1l);
}

Vector3 Mesh::Triangle::getCentroid(CSpace coordinateSpace) const
{
    Vector3 temp;

    temp = parent->getVertex(v1, coordinateSpace) + parent->getVertex(v2, coordinateSpace)
        + parent->getVertex(v3, coordinateSpace);

    temp[0] /= 3;
    temp[1] /= 3;
    temp[2] /= 3;

    return temp;
}

Mesh::Mesh(ObjectID oid, const std::string& name, Position pos)
    : Geometry(oid, name, pos)
{}

Mesh::~Mesh() {}

void Mesh::setOrientation(Euler newOrient, bool notifyObs)
{
    if (newOrient != position.orientation) {
        SceneObject::setOrientation(newOrient,
            false); // without notifiying observers to rebuild vertices first
        rebuildGlobalVertices();

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }
    }
}

void Mesh::setLocation(Vector3 newLoc, bool notifyObs)
{
    if (newLoc != position.location) {
        SceneObject::setLocation(
            newLoc, false); // without notifiying observers to rebuild vertices first
        rebuildGlobalVertices();

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }
    }
}

void Mesh::setPosition(Position newPos, SceneObject* parent, bool notifyObs)
{
    if (newPos.location != position.location || newPos.orientation != position.orientation
        || newPos.cSpace != position.cSpace) {
        SceneObject::setPosition(newPos, parent,
            false); // without notifiying observer to rebuild vertices first
        rebuildGlobalVertices();

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        // TODO: this might lead to problems with RCS meshes and notified observers?
    }
}

void Mesh::setCSpace(EAR::CSpace newCSpace, SceneObject* newParent, bool notifyObs)
{
    if (newCSpace != position.cSpace || newParent != parent) {
        // without notifiying observer to rebuild vertices first
        SceneObject::setCSpace(newCSpace, newParent, false);
        rebuildGlobalVertices();

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }
    }
}

int Mesh::addVertex(const Vector3& v)
{
    localVertices.push_back(v);
    globalVertices.push_back(local2global(v));
    return static_cast<int>(localVertices.size() - 1);
}

void Mesh::addVertex(const Vector3& v, unsigned int idx)
{
    if (localVertices.size() <= idx) {
        localVertices.resize(idx + 1);
        globalVertices.resize(idx + 1);
    }

    localVertices[idx] = v;
    globalVertices[idx] = local2global(v);
}

const Vector3& Mesh::getVertex(int vertexid, CSpace coordinateSpace) const
{
    if (coordinateSpace == CSpace::Local) {
        return localVertices[vertexid];
    } else { // CSpace::Global
        return globalVertices[vertexid];
    }
}

Mesh::Triangle& Mesh::getTriangle(int triangleid)
{
    return triangles[triangleid];
}

const Mesh::Triangle& Mesh::getTriangle(int triangleid) const
{
    return triangles[triangleid];
}

int Mesh::addTriangle(
    unsigned int v1, unsigned int v2, unsigned int v3, std::shared_ptr<Material> material)
{
    if (isLocked) {
        EAR_FAIL("addTriangle not executable while isLocked = true")
    }

    EAR_CHECK_REL(
        v1 < localVertices.size() && v2 < localVertices.size() && v3 < localVertices.size(),
        "Index out of range.");
    triangles.emplace_back(v1, v2, v3, this, material);

    return static_cast<int>(triangles.size() - 1);
}

const std::vector<Vector3>& Mesh::getVertices(CSpace coordinateSpace) const
{
    if (coordinateSpace == CSpace::Local) {
        return localVertices;
    } else { // CSpace::Global
        return globalVertices;
    }
}

const std::vector<Mesh::Triangle>& Mesh::getTriangles() const
{
    return triangles;
}

const std::vector<Mesh::Surface>& Mesh::getSurfaces() const
{
    return surfaces;
}

const std::vector<Mesh::RoundEdge>& Mesh::getRoundEdges() const
{
    return roundEdges;
}

const std::vector<Mesh::RoundCorner>& Mesh::getRoundCorners() const
{
    return roundCorners;
}

Mesh::Plane Mesh::getPlane(const Triangle& tri, CSpace coordinateSpace) const
{
    Plane ret;

    // compute unit normal vector
    ret.N0 = tri.getNormal(coordinateSpace);
    float len = ret.N0.abs();
    ret.N0[0] /= len;
    ret.N0[1] /= len;
    ret.N0[2] /= len;

    const Vector3& A = getVertex(tri.v1, coordinateSpace);
    ret.d = ret.N0[0] * A[0] + ret.N0[1] * A[1] + ret.N0[2] * A[2];

    return ret;
}

Mesh::Plane Mesh::getPlane(const Vector3& A, const Vector3& B, const Vector3& C) const
{
    Plane ret;

    // compute unit normal vector
    ret.N0 = ((B - A) * (C - A)).normalize();
    float len = ret.N0.abs();
    ret.N0[0] /= len;
    ret.N0[1] /= len;
    ret.N0[2] /= len;

    ret.d = ret.N0[0] * A[0] + ret.N0[1] * A[1] + ret.N0[2] * A[2];

    return ret;
}

bool Mesh::samePlane(const Plane& p1, const Plane& p2)
{
    bool ret = false;
    const float thresh = 0.01f;

    ret = (std::fabs(p2.d - p1.d) < thresh) && ((p2.N0 - p1.N0).abs() < thresh);
    return ret;
}

bool Mesh::connectedTris(Triangle& Tri_i, Triangle& Tri_j)
{
    bool ret = false;

    EAR::Vector3 Tri_i_vertices(Tri_i.v1, Tri_i.v2, Tri_i.v3);
    EAR::Vector3 Tri_j_vertices(Tri_j.v1, Tri_j.v2, Tri_j.v3);

    int count = 0;

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            if (Tri_i_vertices[i] == Tri_j_vertices[j]) {
                count++;
            }
        }
    }

    if (count == 2) {
        ret = true;
    }

    return ret;
}

bool Mesh::connectedSubSurfs(
    std::vector<unsigned int> sub_surface_i, std::vector<unsigned int> sub_surface_j)
{
    bool ret = false;

    for (auto& triID_i : sub_surface_i) {
        for (auto& triID_j : sub_surface_j) {
            Triangle Tri_i = triangles[triID_i];
            Triangle Tri_j = triangles[triID_j];

            if (connectedTris(Tri_i, Tri_j)) {
                return true;
            }
        }
    }

    return ret;
}

float Mesh::pointPlaneDist(const Plane& plane, Vector3 P) const
{
    return plane.N0[0] * P[0] + plane.N0[1] * P[1] + plane.N0[2] * P[2] - plane.d;
}

bool Mesh::canPlaneOcclude(Plane& plane, CSpace coordinateSpace)
{
    const std::vector<Vector3>& vertices = getVertices(coordinateSpace);
    const float thresh = 0.001f;

    for (auto& vtx : vertices) {
        float plane_dist = pointPlaneDist(plane, vtx);

        if (plane_dist < -thresh) {
            return true;
        }
    }

    return false;
}

bool Mesh::canPlaneOccludeOther(const Plane& plane, const Mesh* other)
{
    const std::vector<Vector3>& vertices = other->getVertices(CSpace::Global);
    const float thresh = 0.001f;

    for (auto& vtx : vertices) {
        float plane_dist = pointPlaneDist(plane, vtx);

        if (plane_dist < -thresh) {
            return true;
        }
    }

    return false;
}

void Mesh::getSubSurfaces(Surface& surf)
{
    std::vector<EAR::Vector3> centroidList;

    for (auto& triID : surf.triangleId) {
        Triangle tri = triangles[triID];
        centroidList.push_back(tri.getCentroid());
    }

    std::vector<int> index(centroidList.size(), 0);

    for (std::size_t i = 0; i != index.size(); i++) {
        index[i] = i;
    }

    std::vector<float> distList;
    EAR::Vector3 ref_0 = centroidList[0];
    EAR::Vector3 relDist;

    for (auto& centroid : centroidList) {
        relDist.set(ref_0[0] - centroid[0], ref_0[1] - centroid[1], ref_0[2] - centroid[2]);
        distList.push_back(relDist.abs());
    }

    sort(index.begin(), index.end(),
        [&](const int& a, const int& b) { return (distList[a] < distList[b]); });

    for (std::size_t i = 0; i < index.size(); i++) {
        unsigned int new_triID = surf.triangleId[index[i]];
        Triangle new_tri = triangles[new_triID];

        if (surf.subSurfaceTriangleIds.size() == 0) {
            std::vector<unsigned int> new_sub_surface;
            new_sub_surface.push_back(new_triID);
            surf.subSurfaceTriangleIds.push_back(new_sub_surface);
        } else {
            bool NEED_OF_CHECK = true;

            for (auto& sub_surface :
                surf.subSurfaceTriangleIds) { // goes into individual vector of triIDs
                if (NEED_OF_CHECK) {
                    for (auto& sub_surface_triID : sub_surface) { // into individual triID

                        unsigned int existing_triID = sub_surface_triID;
                        Triangle existing_tri = triangles[existing_triID];

                        if (connectedTris(existing_tri, new_tri)) {
                            sub_surface.push_back(surf.triangleId[index[i]]);
                            NEED_OF_CHECK = false;
                            break;
                        }
                    }
                }
            }

            if (NEED_OF_CHECK) {
                std::vector<unsigned int> new_sub_surface;
                new_sub_surface.push_back(new_triID);
                surf.subSurfaceTriangleIds.push_back(new_sub_surface);
            }
        }
    }

    std::vector<std::pair<unsigned int, unsigned int>> integration;
    int NOF_SUB_SURFACES = surf.subSurfaceTriangleIds.size();

    for (int i = 0; i < NOF_SUB_SURFACES; i++) {
        for (int j = i + 1; j < NOF_SUB_SURFACES; j++) {

            if (connectedSubSurfs(surf.subSurfaceTriangleIds[i], surf.subSurfaceTriangleIds[j])) {
                integration.push_back(std::make_pair(i, j));
            }
        }
    }

    std::vector<unsigned int> delete_Sub_SurfaceIndex;

    for (std::size_t i = 0; i < integration.size(); i++) {
        for (std::size_t j = i + 1; j < integration.size(); j++) {

            if (integration[i].second == integration[j].second) {
                integration[i].second = integration[j].first;

                for (auto& a : surf.subSurfaceTriangleIds[integration[j].second]) {
                    surf.subSurfaceTriangleIds[integration[j].first].push_back(a);
                }

                delete_Sub_SurfaceIndex.push_back(integration[j].second);
            }
        }
    }

    for (auto& del : delete_Sub_SurfaceIndex) {
        surf.subSurfaceTriangleIds[del].clear();
    }

    for (int i = integration.size() - 1; i > -1; i--) {
        for (auto& a : surf.subSurfaceTriangleIds[integration[i].second]) {
            surf.subSurfaceTriangleIds[integration[i].first].push_back(a);
        }

        surf.subSurfaceTriangleIds[integration[i].second].clear();
    }

    int NOF_SUBSURFACES = surf.subSurfaceTriangleIds.size();

    for (int i = NOF_SUBSURFACES - 1; i > -1; i--) {
        if (surf.subSurfaceTriangleIds[i].size() == 0) {
            surf.subSurfaceTriangleIds.erase(surf.subSurfaceTriangleIds.begin() + i);
        }
    }
}

unsigned int Mesh::rebuildSurfaceStructure(CSpace coordinateSpace)
{
    // clear surface data structure
    surfaces.clear();

    unsigned int numTriangles = triangles.size();

    if (numTriangles == 0) {
        return surfaces.size();
    }

    // extract surfaces
    unsigned int count = 0;

    for (unsigned int tid = 0; tid < numTriangles; tid++) {
        Triangle& tri = triangles[tid];
        Plane currentPlane = getPlane(tri, coordinateSpace);
        bool found = false;

        unsigned int surfaceId = 0;

        for (auto& surf : surfaces) {
            found = samePlane(currentPlane, surf.plane);

            if (found) {
                // add triangle to existing surface
                surf.triangleId.push_back(tid);
                break;
            }

            surfaceId++;
        }

        if (!found) {
            // generate new surface
            std::vector<unsigned int> triIds;
            triIds.push_back(tid);
            surfaces.emplace_back(count++, triIds, currentPlane, false);
        }

        tri.surfaceId = surfaceId;
    }

    // determine potential occluders
    needsOcclusionTest = false;

    for (auto& surf : surfaces) {
        surf.canOccludeSelf = canPlaneOcclude(surf.plane, coordinateSpace);
        surf.canOccludeOthers = false;
        surf.canOcclude = surf.canOccludeSelf;
        needsOcclusionTest |= surf.canOccludeSelf;
    }

    // divide a surface into multiple sub-surfaces only if they are separated
    for (auto& surf : surfaces) {
        getSubSurfaces(surf);
    }

    return surfaces.size();
}

unsigned int Mesh::rebuildSurfaceStructureExtern(
    const std::vector<Mesh::Surface>& _surfaces, bool replaceDist, CSpace coordinateSpace)
{
    // copy surface data structure
    surfaces = _surfaces;

    // determine potential occluders
    needsOcclusionTest = false;

    for (auto& surf : surfaces) {
        surf.canOccludeSelf = canPlaneOcclude(surf.plane, coordinateSpace);
        surf.canOccludeOthers = false;
        surf.canOcclude = surf.canOccludeSelf;
        needsOcclusionTest |= surf.canOccludeSelf;

        // Special feature: extract d parameter from vertices (plane distance in Hesse normal form).
        // This is useful, when the surface data comes from a bitstream without d values.
        if (replaceDist) {
            std::set<unsigned int> vertex_set;
            for (auto faceId : surf.triangleId) {
                auto tri = getTriangle(faceId);
                vertex_set.insert(tri.v1);
                vertex_set.insert(tri.v2);
                vertex_set.insert(tri.v3);
            }
            float d = 0;
            for (auto vertexId : vertex_set) {
                EAR::Vector3 N0 = surf.plane.N0;
                EAR::Vector3 V = getVertex(vertexId);
                d += N0.dot(V);
            }
            d /= vertex_set.size(); // mean distance
            surf.plane.d = d;
        }
    }

    // divide a surface into multiple sub-surfaces only if they are separated
    for (auto& surf : surfaces) {
        getSubSurfaces(surf);
    }

    // reset surfaceId for all triangles
    for (auto& tri : triangles) {
        tri.surfaceId = -1;
    }
    for (auto& surf : surfaces) {
        for (auto tid : surf.triangleId) {
            triangles[tid].surfaceId = surf.surfaceId;
        }
    }

    return surfaces.size();
}

void Mesh::rebuildRoundStructure()
{
    roundEdges.clear();
    roundCorners.clear();
    GeometryPreprocessor::getRoundStructure(roundEdges, roundCorners, this);
}

void Mesh::updatePlanes()
{
    for (auto& surf : surfaces) {
        auto tri = getTriangle(surf.triangleId[0]);
        surf.plane = getPlane(tri, EAR::Mesh::CSpace::Global);
    }
}

void Mesh::updateOccludeOthers(const Mesh* other)
{
    if (other == this) {
        return;
    }

    // check for each surface of this mesh, if it potentially occludes the other geometry
    for (auto& surf : surfaces) {
        if (surf.canOcclude == false) {
            surf.canOccludeOthers = canPlaneOccludeOther(surf.plane, other);
            surf.canOcclude =
                surf.canOccludeOthers; // canOccludeSelf can be skipped as it must be false, if canOcclude is false
            needsOcclusionTest |= surf.canOcclude;
        }
    }
}

void Mesh::resetOccludeOthers()
{
    needsOcclusionTest = false;

    for (auto& surf : surfaces) {
        surf.canOccludeOthers = false;
        surf.canOcclude = surf.canOccludeSelf;
        needsOcclusionTest |= surf.canOcclude;
    }
}

float Mesh::getArea() const
{

    float area = 0;

    for (auto& tri : triangles) {
        const Vector3& first = localVertices[tri.v1];
        const Vector3& second = localVertices[tri.v2];
        const Vector3& third = localVertices[tri.v3];

        Vector3 crossproduct = (second - first) * (third - first);

        area += 0.5f * crossproduct.abs();
    }

    return area;
}

const Mesh* Mesh::getMesh() const
{
    return this;
}

float Mesh::getVolume() const
{
    float volume = 0.f;

    for (const auto& tri : triangles) {
        const Vector3& p1 = localVertices[tri.v1];
        const Vector3& p2 = localVertices[tri.v2];
        const Vector3& p3 = localVertices[tri.v3];
        volume += p1.dot(p2 * p3);
    }

    return std::abs(volume / 6.f);
}

Vector3 Mesh::local2global(const Vector3& v)
{
    return getGlobalCSTransform() * v;
}

void Mesh::rebuildGlobalVertices()
{
    globalVertices.resize(localVertices.size());
    std::transform(localVertices.begin(), localVertices.end(), globalVertices.begin(),
        [this](Vector3& v) { return local2global(v); });

    // update plane definitions of all surfaces
    updatePlanes();
}

void Mesh::lock()
{
    lookupTree = std::make_unique<CartesianKDTree>(localVertices);
    this->isLocked = true;

    this->rt = std::make_unique<EmbreeTracer>(this, false);

    return;
}

void Mesh::unlock()
{
    this->isLocked = false;
    this->rt.reset();
}

void Mesh::parentMoved(bool notifyObs)
{
    if (position.cSpace != EAR::CSpace::WCS) {
        rebuildGlobalVertices();
    }
    SceneObject::parentMoved(notifyObs);
}

std::ostream& operator<<(std::ostream& os, Mesh& m)
{
    os << "Mesh \"" << m.name << "\" (" << m.triangles.size() << " Triangles, "
       << m.localVertices.size() << " Vertices)";

    return os;
}

std::ostream& operator<<(std::ostream& os, Mesh::RoundEdge& e)
{
    os << "RoundEdge(vtxFrom = " << e.vtxFrom << ", ";
    os << "vtxTo = " << e.vtxTo << ", ";
    os << "triLeft = " << e.triLeft << ", ";
    os << "triRight = " << e.triRight << ", ";
    os << "alpha = " << e.alpha << ")";

    return os;
}

std::ostream& operator<<(std::ostream& os, Mesh::RoundCorner& c)
{
    os << "RoundCorner(vtx = " << c.vtx << ", ";
    os << "tidList = " << c.tidList << ")";

    return os;
}

bool Mesh::isInside(const Vector3& point) const
{
    //embree algorithm
    if (!this->isLocked) {
        EAR_FAIL("isInsideMesh not executable while isLocked = false");
    }

    return rt->isInside(point);
}

// David Eberly, Geometric Tools, Redmond WA 98052
// https://www.geometrictools.com/index.html
// Copyright (c) 1998-2022
// Distributed under the Boost Software License, Version 1.0.
// Version: 6.0.2023.08.08
Vector3 Mesh::closestPointTo(const EAR::Vector3& point) const
{
    Vector3 relPoint = getRelativeLocation(point);

    float distToClosestPoint = lookupTree->getDistanceToNearestPoint(relPoint);
    auto bb = lookupTree->getBoundingBox();
    auto bbMax =
        std::max(bb[0].high - bb[0].low, std::max(bb[1].high - bb[1].low, bb[2].high - bb[2].low));
    // The nearest point may not be on a triangle that uses the nearest vertex, especially with concave meshes.
    // This finds all verts within half the max bounding box dimension of the nearest vert, and adds them to the
    // list of verts to test with.
    auto allClosePoints =
        lookupTree->getAllIndicesWithinRadius(relPoint, distToClosestPoint + (bbMax / 2));

    float shortestDist = std::numeric_limits<float>::infinity();
    Vector3 finalClosestPoint;

    // Loop through to find the triangles that has the closest vertice(s) as their vertex
    for (unsigned int i = 0; i < triangles.size(); i++) {

        const Triangle& tri = triangles[i];
        bool needTest = false;
        for (const auto& idxToTest : allClosePoints) {
            if (idxToTest == tri.v1 || idxToTest == tri.v2 || idxToTest == tri.v3) {
                needTest = true;
                break;
            }
        }

        if (!needTest) {
            continue;
        } else {

            // // calculate the distance from this triangle to point
            const Vector3& vertex0 = getVertex(tri.v1, CSpace::Local);
            const Vector3& vertex1 = getVertex(tri.v2, CSpace::Local);
            const Vector3& vertex2 = getVertex(tri.v3, CSpace::Local);
            Vector3 edge0 = vertex1 - vertex0;
            Vector3 edge1 = vertex2 - vertex0;

            float a = edge0.dot(edge0);
            float b = edge0.dot(edge1);
            float c = edge1.dot(edge1);
            float d = edge0.dot(vertex0 - relPoint);
            float e = edge1.dot(vertex0 - relPoint);
            float det = a * c - b * b;
            float s = b * e - c * d;
            float t = b * d - a * e;
            float invDet, tmp0, tmp1, denom, numer;

            if (s + t <= det) {
                if (s < 0) {
                    if (t < 0) // region 4
                    {
                        if (d < 0) {
                            t = 0;
                            if (-d >= a) {
                                s = 1;
                            } else {
                                s = -d / a;
                            }
                        } else {
                            s = 0;
                            if (e >= 0) {
                                t = 0;
                            } else if (-e >= c) {
                                t = 1;
                            } else {
                                t = -e / c;
                            }
                        }
                    } else // region 3
                    {
                        s = 0;
                        if (e >= 0) {
                            t = 0;
                        } else if (-e >= c) {
                            t = 1;
                        } else {
                            t = -e / c;
                        }
                    }
                } else if (t < 0) // region 5
                {
                    t = 0;
                    if (d >= 0) {
                        s = 0;
                    } else if (-d >= a) {
                        s = 1;
                    } else {
                        s = -d / a;
                    }
                } else // region 0
                {
                    // minimum at interior point
                    s /= det;
                    t /= det;
                }
            } else {
                if (s < 0) // region 2
                {
                    tmp0 = b + d;
                    tmp1 = c + e;
                    if (tmp1 > tmp0) {
                        numer = tmp1 - tmp0;
                        denom = a - 2 * b + c;
                        if (numer >= denom) {
                            s = 1;
                            t = 0;
                        } else {
                            s = numer / denom;
                            t = 1 - s;
                        }
                    } else {
                        s = 0;
                        if (tmp1 <= 0) {
                            t = 1;
                        } else if (e >= 0) {
                            t = 0;
                        } else {
                            t = -e / c;
                        }
                    }
                } else if (t < 0) // region 6
                {
                    tmp0 = b + e;
                    tmp1 = a + d;
                    if (tmp1 > tmp0) {
                        numer = tmp1 - tmp0;
                        denom = a - 2 * b + c;
                        if (numer >= denom) {
                            t = 1;
                            s = 0;
                        } else {
                            t = numer / denom;
                            s = 1 - t;
                        }
                    } else {
                        t = 0;
                        if (tmp1 <= 0) {
                            s = 1;
                        } else if (d >= 0) {
                            s = 0;
                        } else {
                            s = -d / a;
                        }
                    }
                } else // region 1
                {
                    numer = c + e - b - d;
                    if (numer <= 0) {
                        s = 0;
                        t = 1;
                    } else {
                        denom = a - 2 * b + c;
                        if (numer >= denom) {
                            s = 1;
                            t = 0;
                        } else {
                            s = numer / denom;
                            t = 1 - s;
                        }
                    }
                }
            }

            Vector3 closestPoint(vertex0 + s * edge0 + t * edge1);
            float distance = (relPoint - closestPoint).abs();

            // Compare current close point on this triangle with the recorded closest one
            if (distance < shortestDist) {
                shortestDist = distance;
                finalClosestPoint = closestPoint;
            }
        }
    }

    // Convert the calculated point back to global coordinate
    finalClosestPoint = getGlobalCSTransform() * finalClosestPoint;
    return finalClosestPoint;
}

bool Mesh::isInsideImpl(const Vector3& point, const EAR::Vector3& rayVector) const
{
    Vector3 relPoint = getRelativeLocation(point);
    int ind = 0;

    for (const Triangle& tri : triangles) {
        const Vector3& vertex0 = getVertex(tri.v1, CSpace::Local);
        const Vector3& vertex1 = getVertex(tri.v2, CSpace::Local);
        const Vector3& vertex2 = getVertex(tri.v3, CSpace::Local);

        Vector3 edge1, edge2, h, s, q, temp;
        float det, u, v, t;

        edge1 = vertex1 - vertex0;
        edge2 = vertex2 - vertex0;

        h = rayVector * edge2;
        det = edge1.dot(h);

        if (det > -EPSILON && det < EPSILON) {
            // This ray is parallel to this triangle.
            continue;
        }

        s = relPoint - vertex0;
        u = s.dot(h) / det;

        if (u < 0.0f || u > 1.0) {
            //outside of triangle
            continue;
        }

        q = s * edge1;
        v = q.dot(rayVector) / det;

        if (v < 0.0f || u + v > 1.0) {
            //outside of triangle
            continue;
        }

        t = edge2.dot(q) / det;

        if (t < 0.0f) {
            //only positive ray direction
            continue;
        } else if (t == 0.0f) {
            //point on surface
            ind = 0;
            break;
        }

        if (u == 0.0 || u == 1.0) {
            //new shot
            float azimuth = (rand() % 360) - 180;
            float elevation = (rand() % 360) - 180;
            return isInsideImpl(point, polarToCartesian(Polar(azimuth, elevation, 1)));
        }

        if (v == 0.0 || u + v == 1.0) {
            //new shot
            float azimuth = (rand() % 360) - 180;
            float elevation = (rand() % 360) - 180;
            return isInsideImpl(point, polarToCartesian(Polar(azimuth, elevation, 1)));
        }

        //hit triangle
        ind ^= 1;
    }

    return (ind != 0);
}

float Mesh::greatestDistance() const
{
    float greatestDistance = 0;
    const std::vector<Vector3>& vertices = this->getVertices();

    for (const auto& vertex : vertices) {
        std::for_each(
            vertices.begin(), vertices.end(), [&vertex, &greatestDistance](Vector3 point) {
                greatestDistance = greatestDistance > (vertex - point).abs()
                    ? greatestDistance
                    : (vertex - point).abs();
            });
    }

    return greatestDistance;
}

void Mesh::getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    povIsInside = false;
    // create coordinate transform matrix with POV as origin
    // and rotation in the direction of the closest point of the mesh
    closestPoint = closestPointTo(pov.location);
    EAR::Polar closestPInPov = cartesianToPolar(closestPoint - pov.location);
    if (isInside(pov.location)) {
        povIsInside = true;
        auto bb = getBoundingBox();
        resWidth = bb[0].high - bb[0].low;
        resHeight = bb[1].high - bb[1].low;
        return;
    }
    auto projCamera = HomogenMatrix(pov.location, Euler(closestPInPov[0], closestPInPov[1], 0))
                          .inverseGeometric();

    Vector3 pv;
    float minW = 1000, maxW = -1000, minH = 1000, maxH = -1000;

    // find outer edge vertices on the projection plane
    for (auto& v : globalVertices) {
        // bring vertex in projection camera coordinate system
        pv = projCamera * v;

        // find the maxima of x and y coordinates in the projection plane
        if (pv[0] < minW) {
            minW = pv[0];
        } else if (pv[0] > maxW) {
            maxW = pv[0];
        }

        if (pv[1] < minH) {
            minH = pv[1];
        } else if (pv[1] > maxH) {
            maxH = pv[1];
        }
    }

    resWidth = maxW - minW;
    resHeight = maxH - minH;
}

void Mesh::getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
    EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    povIsInside = false;
    // create coordinate transform matrix with POV as origin
    // and rotation in the direction of the closest point of the mesh
    closestPoint = closestPointTo(pov.location);
    EAR::Polar closestPInPov = cartesianToPolar(closestPoint - pov.location);
    if (isInside(pov.location)) {
        povIsInside = true;
        resultBottomLeft.set(-1, -1, 0);
        resultBottomRight.set(1, -1, 0);
        resultTopLeft.set(-1, 1, 0);
        resultTopRight.set(1, 1, 0);
        return;
    }
    auto projCamera = HomogenMatrix(pov.location, Euler(closestPInPov[0], closestPInPov[1], 0))
                          .inverseGeometric();
    HomogenMatrix povCamera(Vector3(), pov.orientation);
    povCamera = povCamera.inverseGeometric() * projCamera.inverseGeometric();
    povCamera.setTranslation(Vector3());
    float d = closestPInPov[2];

    Vector3 pv, resBL, resTR, resBR, resTL;

    // find outer edge vertices on the projection plane
    for (auto& v : globalVertices) {
        // bring vertex in projection camera coordinate system
        pv = projCamera * v;

        // calculate projected x and y
        float z = -1 * pv[2];
        pv[0] *= d / z;
        pv[1] *= d / z;
        pv[2] = -1 * d;

        if (pv[0] < resBL[0]) {
            resBL[0] = pv[0];
        } else if (pv[0] > resTR[0]) {
            resTR[0] = pv[0];
        }

        if (pv[1] < resBL[1]) {
            resBL[1] = pv[1];
        } else if (pv[1] > resTR[1]) {
            resTR[1] = pv[1];
        }
    }
    resBR[0] = resTR[0];
    resBR[1] = resBL[1];
    resTL[0] = resBL[0];
    resTL[1] = resTR[1];

    resBL[2] = -1 * d;
    resTR[2] = -1 * d;
    resBR[2] = -1 * d;
    resTL[2] = -1 * d;

    // pov-relative cartesian coordinates, incorporating the pov rotation
    resultBottomLeft = povCamera * resBL;
    resultTopRight = povCamera * resTR;
    resultBottomRight = povCamera * resBR;
    resultTopLeft = povCamera * resTL;
}

Mesh::BoundingBox Mesh::getBoundingBox() const
{
    return lookupTree->getBoundingBox();
}

Mesh::BoundingBox Mesh::getGlobalBoundingBox() const
{

    EAR::Geometry::BoundingBox boundingBox;

    const auto xExtremes = std::minmax_element(globalVertices.begin(), globalVertices.end(),
        [](const EAR::Vector3& lhs, const EAR::Vector3 rhs) { return lhs[0] < rhs[0]; });

    const auto yExtremes = std::minmax_element(globalVertices.begin(), globalVertices.end(),
        [](const EAR::Vector3& lhs, const EAR::Vector3& rhs) { return lhs[1] < rhs[1]; });

    const auto zExtremes = std::minmax_element(globalVertices.begin(), globalVertices.end(),
        [](const EAR::Vector3& lhs, const EAR::Vector3& rhs) { return lhs[2] < rhs[2]; });

    boundingBox[0].low = xExtremes.first->data()[0];
    boundingBox[0].high = xExtremes.second->data()[0];
    boundingBox[1].low = yExtremes.first->data()[1];
    boundingBox[1].high = yExtremes.second->data()[1];
    boundingBox[2].low = zExtremes.first->data()[2];
    boundingBox[2].high = zExtremes.second->data()[2];

    return boundingBox;
}

EAR::Vector3 Mesh::getCenterPosition() const
{
    auto bb = this->getGlobalBoundingBox();
    return EAR::Vector3(
        (bb[0].low + bb[0].high) / 2, (bb[1].low + bb[1].high) / 2, (bb[2].low + bb[2].high) / 2);
}

} /* namespace EAR */
