/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/MpegiDecorrelator.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/ComplexIIRFilter.h>
#include <EAR/FFT.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <algorithm>
#include <cmath>
#include <numeric>
#include <utility>

namespace EAR {

namespace detail {

static const int blockSize = 256;

// The delay is currently 4 for all frequency bins
static const int numOfDelay = 4;

// variables for t/f envelop shaping
static const float alpha = 0.4f;
static const float beta = 1.5f;
static const float normFactor = 1.1f;

// variables for transient detection
static const float delta = 0.6f;
static const int startBand = 3;
static const int holdCount = 8;
static const float transThreshold = 2.8f;

// define the coefficients of the all-pass filters (originally designed to be 4 serial filters,
// but based on the current IIR filter method, one longer filter is cheaper)
static const int FILTER_COEFFS_N = 17;
static const float FREQ_DOMAIN_FILTER_COEFFS_A[FILTER_COEFFS_N] = { 1.f, 0.7f, 0.7f, 1.19f, 0.49f,
    1.19f, 0.833f, 0.49f, 0.833f, 0.343f, 0.343f, 0.2401f };
static const float FREQ_DOMAIN_FILTER_COEFFS_B[FILTER_COEFFS_N] = { 0.2401f, 0.343f, 0.343f, 0.833f,
    0.49f, 0.833f, 1.19f, 0.49f, 1.19f, 0.7f, 0.7f, 1.f };
} // namespace detail

MpegiDecorrelator::MpegiDecorrelator(std::size_t B)
    : hB(detail::blockSize / 2)
    , B(B)
{
    decoFFT = EAR::RFFT::create(detail::blockSize);
    decoIFFT = EAR::IRFFT::create(detail::blockSize);
    K = decoFFT->K;

    // container in time-domain
    ringBuffer = std::make_unique<EAR::RealBuffer>(B + detail::blockSize);
    curInData = std::make_unique<EAR::RealBuffer>(detail::blockSize);
    windowedData = std::make_unique<EAR::RealBuffer>(detail::blockSize);
    preDecorData = std::make_unique<EAR::RealBuffer>(detail::blockSize / 2);
    tmpOutput = std::make_unique<EAR::RealBuffer>(hB);
    tmpFrame = std::make_unique<EAR::RealFrame>(2, hB);
    outFrame = std::make_unique<EAR::RealFrame>(2, B + detail::blockSize / 2);

    // container in frequency-domain
    freqData = std::make_unique<EAR::ComplexBuffer>(K);
    direcFreqData = std::make_unique<EAR::RealBuffer>(K);
    preDecorEnergy = std::make_unique<EAR::RealBuffer>(K);
    preDirecEnergy = std::make_unique<EAR::RealBuffer>(K);

    // prepare the windowing
    window = std::make_unique<EAR::RealBuffer>(detail::blockSize);
    float step = M_PI / detail::blockSize;
    for (int i = 0; i < detail::blockSize; i++) {
        (*window)[i] = std::sin((i + 0.5f) * step);
    }

    // set the read & write cursor value for the input and output buffer respectively
    inRC = 0;
    inWC = detail::blockSize / 2;
    outRC = 0;
    outWC = detail::blockSize / 2;
    delayRC = 0;
    delayWC = detail::numOfDelay;

    // prepare the storing container for delayed frequency conponents and the all-pass filters
    aPFilters.reserve(K);
    for (int i = 0; i < K; i++) {
        delayStorage.push_back(EAR::ComplexBuffer(detail::numOfDelay + 1));
        aPFilters.emplace_back(detail::FREQ_DOMAIN_FILTER_COEFFS_B, detail::FILTER_COEFFS_N,
            detail::FREQ_DOMAIN_FILTER_COEFFS_A, detail::FILTER_COEFFS_N);
    }

    numToProcess = 0;
    noneProcessed = true;
    preIsTrans = false;
    writeDirect = false;
    inhibitCount = 56;

#ifndef MPEGI_CFP_SUBMISSION
    preTranSum = 0.f;
    transCount = 0;
    noTransCount = 0;
#endif
}

void MpegiDecorrelator::process(const float* in, float** out)
{

    // read current data into the overlap buffer
    auto bPtr = ringBuffer->data();
    const std::size_t inBuffLen = ringBuffer->length();
    if (inWC + B <= inBuffLen) {
        std::copy(in, in + B, bPtr + inWC);
        inWC = (inWC + B == inBuffLen ? 0 : inWC + B);
    } else {
        std::copy(in, in + inBuffLen - inWC, bPtr + inWC);
        std::copy(in + inBuffLen - inWC, in + B, bPtr);
        inWC += (B - inBuffLen);
    }

    numToProcess += B;

    auto winPtr = window->data();
    auto curIPtr = curInData->data();
    auto winedPtr = windowedData->data();
    auto freqPtr = freqData->data();
    auto direcFPtr = direcFreqData->data();
    auto pDecorEgPtr = preDecorEnergy->data();
    auto pDirecEgPtr = preDirecEnergy->data();
    auto prePtr = preDecorData->data();
    auto tmpPtr = tmpOutput->data();
    auto tmpFPtr = tmpFrame->data();
    auto outPtr = outFrame->data();

    const std::size_t outBuffLen = outFrame->length();
    double sqrtTwo = std::sqrt(2);

    // start processing
    while (numToProcess >= hB) {
        noneProcessed = false;
        writeDirect = true;
        writeProcessed = true;

        // read from overlap buffer into current process buffer
        if (inRC + detail::blockSize <= inBuffLen) {
            std::copy(bPtr + inRC, bPtr + inRC + detail::blockSize, curIPtr);
            inRC += hB;
        } else {
            std::copy(bPtr + inRC, bPtr + inBuffLen, curIPtr);
            std::copy(
                bPtr, bPtr + detail::blockSize - inBuffLen + inRC, curIPtr + inBuffLen - inRC);
            inRC = (inRC + hB >= inBuffLen ? inRC + hB - inBuffLen : inRC + hB);
        }
        // do windowing (but keep the unwindowed current buffer for later operation)
        std::transform(curIPtr, curIPtr + detail::blockSize, winPtr, winedPtr,
            [](float a, float b) { return a * b; });

        // spectral analysis
        decoFFT->transform(winedPtr, freqPtr);
        std::transform(freqPtr, freqPtr + K, direcFPtr,
            [](std::complex<float> a) { return std::pow(std::abs(a), 2); });

        // delay and all-pass chains
        for (int i = 0; i < K; i++) {
            delayStorage[i][delayWC] = freqPtr[i];
            freqPtr[i] = delayStorage[i][delayRC];
            freqPtr[i] = aPFilters[i].process(freqPtr[i]);
        }
        delayRC = (delayRC == detail::numOfDelay ? 0 : delayRC + 1);
        delayWC = (delayWC == detail::numOfDelay ? 0 : delayWC + 1);

        // transient detection
        float curTranSum = std::accumulate(direcFPtr + detail::startBand, direcFPtr + K, 0.f);
        curTranSum = detail::delta * preTranSum + (1.f - detail::delta) * curTranSum;

        transCount = (transCount == detail::holdCount ? 0 : transCount);
        noTransCount = (noTransCount == inhibitCount ? 0 : noTransCount);
        transCount = (transCount > 0 ? transCount + 1 : 0);
        noTransCount = (noTransCount > 0 ? noTransCount + 1 : 0);

        if ((detail::transThreshold * preTranSum) < curTranSum && noTransCount > 0) {
            noTransCount = 1;
            inhibitCount = 64;
        }

        if ((detail::transThreshold * preTranSum) < curTranSum && noTransCount == 0) {
            transCount = 1;
            noTransCount = 1;
            inhibitCount = 56;
        }

        preTranSum = curTranSum;

        // t/f envelop shaping (possible tuning on frequency resolution here)
        float tmpDecorEg, tmpDirecEg;
        for (int i = 0; i < K; i++) {
            tmpDecorEg = detail::alpha * std::pow(std::abs(freqPtr[i]), 2)
                + (1 - detail::alpha) * pDecorEgPtr[i];
            tmpDirecEg = detail::alpha * direcFPtr[i] + (1 - detail::alpha) * pDirecEgPtr[i];
            // the actually t/f shaping, skip to save computation if it's not needed (transient detected)
            if (transCount == 0) {
                // amplify if decorrelated signal is too weak
                if (tmpDirecEg > detail::beta * tmpDecorEg && tmpDecorEg > 0.f) {
                    freqPtr[i] *=
                        std::sqrt(tmpDirecEg / (detail::beta * tmpDecorEg)) * detail::normFactor;
                } // attenuate if decorrelated signal is too strong
                else if (detail::beta * tmpDirecEg < tmpDecorEg && tmpDecorEg > 0.f) {
                    freqPtr[i] *=
                        std::sqrt(detail::beta * tmpDirecEg / tmpDecorEg) * detail::normFactor;
                } // boost the signal slightly if it's neither of the above sitution
                else {
                    freqPtr[i] *= detail::normFactor;
                }
            }
            pDecorEgPtr[i] = tmpDecorEg;
            pDirecEgPtr[i] = tmpDirecEg;
        }

        tmpFrame->zero();
        // obtain output based on whether transient count is running in the current frame
        if (transCount > 0) {
            //take coherent addition into account
            windowedData->mul(sqrtTwo - 1.f);

            std::transform(winedPtr, winedPtr + hB, winPtr, tmpFPtr[0],
                [](float a, float b) { return a * b; });

            if (preIsTrans) {
                // when transient is detected in the previous and this frame
                for (int i = 0; i < hB; i++) {
                    tmpFPtr[0][i] += prePtr[i] * winPtr[hB + i];
                }
                writeProcessed = false;
            } else {
                // when transient is detected in this frame but not the previous one
                std::transform(prePtr, prePtr + hB, winPtr + hB, tmpPtr,
                    [](float a, float b) { return a * b; });
            }

        } else {
            // spectral synthesis
            decoIFFT->transform(freqPtr, winedPtr);
            windowedData->mul(1.0f / detail::blockSize);

            std::transform(
                winedPtr, winedPtr + hB, winPtr, tmpPtr, [](float a, float b) { return a * b; });
            if (preIsTrans) {
                // when transient is detected in previous frame, but not this one
                std::transform(prePtr, prePtr + hB, winPtr + hB, tmpFPtr[0],
                    [](float a, float b) { return a * b; });
            } else {
                // when transient is not detected in this or previous frame
                for (int i = 0; i < hB; i++) {
                    tmpPtr[i] += prePtr[i] * winPtr[hB + i];
                }
                writeDirect = false;
            }
        }

        // Produce the final output 1. add current input
        std::transform(tmpFPtr[0], tmpFPtr[0] + hB, curIPtr, tmpFPtr[0],
            [](float a, float b) { return a + b; });
        tmpFrame->ch(1).write(tmpFPtr[0], 0, hB);

        // 2. for transient output, both channel are original input plus the windowed original intput
        // which is already done in the preceeding section

        // 3. For non-transient output, left and right channel is original input add and minus the processed
        // output respectively
        if (writeProcessed) {
            std::transform(tmpFPtr[0], tmpFPtr[0] + hB, tmpPtr, tmpFPtr[0],
                [](float a, float b) { return a + b; });
            std::transform(tmpFPtr[1], tmpFPtr[1] + hB, tmpPtr, tmpFPtr[1],
                [](float a, float b) { return a - b; });
        }

        // normalize and write final output into the ringBuffer
        if (outWC + hB <= outBuffLen) {
            std::transform(tmpFPtr[0], tmpFPtr[0] + hB, outPtr[0] + outWC,
                [num = sqrtTwo](float a) { return a / num; });
            std::transform(tmpFPtr[1], tmpFPtr[1] + hB, outPtr[1] + outWC,
                [num = sqrtTwo](float a) { return a / num; });
            outWC = (outWC + hB == outBuffLen ? 0 : outWC + hB);
        } else {
            std::transform(tmpFPtr[0], tmpFPtr[0] + outBuffLen - outWC, outPtr[0] + outWC,
                [num = sqrtTwo](float a) { return a / num; });
            std::transform(tmpFPtr[0] + outBuffLen - outWC, tmpFPtr[0] + hB, outPtr[0],
                [num = sqrtTwo](float a) { return a / num; });
            std::transform(tmpFPtr[1], tmpFPtr[1] + outBuffLen - outWC, outPtr[1] + outWC,
                [num = sqrtTwo](float a) { return a / num; });
            std::transform(tmpFPtr[1] + outBuffLen - outWC, tmpFPtr[1] + hB, outPtr[1],
                [num = sqrtTwo](float a) { return a / num; });
            outWC += (hB - outBuffLen);
        }

        // save current output for next round
        windowedData->read(prePtr, hB, hB);

        // keep track of how many samples are left to be processed and trasient detection
        numToProcess -= hB;
        preIsTrans = (transCount > 0);
    }

    // write the decorrelated output into the final output
    if (outRC + B <= outBuffLen) {
        std::copy(outPtr[0] + outRC, outPtr[0] + outRC + B, out[0]);
        std::copy(outPtr[1] + outRC, outPtr[1] + outRC + B, out[1]);
        outRC = (outRC + B == outBuffLen ? 0 : outRC + B);
    } else {
        std::copy(outPtr[0] + outRC, outPtr[0] + outBuffLen, out[0]);
        std::copy(outPtr[0], outPtr[0] + B - outBuffLen + outRC, out[0] + outBuffLen - outRC);
        std::copy(outPtr[1] + outRC, outPtr[1] + outBuffLen, out[1]);
        std::copy(outPtr[1], outPtr[1] + B - outBuffLen + outRC, out[1] + outBuffLen - outRC);
        outRC += (B - outBuffLen);
    }
}

} // namespace EAR
