/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/OSC.h>

#include <EAR/Exception.h>

#include <OscOutboundPacketStream.h>
#include <OscPacketListener.h>
#include <UdpSocket.h>

#include <algorithm>
#include <cctype>
#include <cstdlib>
#include <iomanip>
#include <sstream>

/*#include <pthread.h>
#include <sched.h>*/

namespace EAR {

namespace detail {

inline void eraseSubStr(std::string& mainStr, const std::string& toErase)
{
    std::size_t pos = mainStr.find(toErase);
    // delete all occurences
    while (pos != std::string::npos) {
        mainStr.erase(pos, toErase.length());

        pos = mainStr.find(toErase);
    }
};

inline bool string2bool(const std::string& v)
{
    return !v.empty() && (v == "true" || std::atoi(v.c_str()) != 0);
}

inline void ltrim(std::string& s)
{
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](int ch) { return !std::isspace(ch); }));
}

inline void rtrim(std::string& s)
{
    s.erase(std::find_if(s.rbegin(), s.rend(), [](int ch) { return !std::isspace(ch); }).base(),
        s.end());
}

inline void trim(std::string& s)
{
    ltrim(s);
    rtrim(s);
}

} // namespace detail

class OSC::PacketListener : public osc::OscPacketListener
{
public:
    PacketListener(OSC& _parent)
        : parent(_parent){};

    //Pure virtual function not used. Messages are processed outside of the listener.
    void ProcessMessage(const osc::ReceivedMessage&, const IpEndpointName&) override {}

    void ProcessPacket(const char* data, int size, const IpEndpointName& remoteEndpoint) override
    {
        (void)
            remoteEndpoint; // unused in this implementation, but kept to have consistent interface

        osc::ReceivedPacket packet(data, size);

        if (packet.IsMessage()) {
            parent.receive(osc::ReceivedMessage(packet));
        } else { //packet.IsBundle() implied
            parent.receive(osc::ReceivedBundle(packet));
        }
    }

private:
    OSC& parent;
};

//==============================================================================

OSC::Message::Message(const std::string& oscAddress)
    : byteSizeOfArguments(0)
    , address(oscAddress)
{
    byteSizeOfArguments += sizeof(osc::BeginMessage(oscAddress.c_str())) + sizeof(osc::EndMessage);
}

OSC::Message::Message(const Message& message)
    : Message(message.address)
{
    arguments = message.arguments;
    byteSizeOfArguments = message.byteSizeOfArguments;
}

OSC::Message::~Message()
{
    byteSizeOfArguments = 0;
}

OSC::Message& OSC::Message::addArgument(Variant argument)
{
    arguments.push_back(argument);

    //Increment byteSizeOfArguments by the size of the type of argument that was passed
    switch (argument.getVariantType()) {
    case Variant::Type::Bool:
        byteSizeOfArguments += sizeof(bool);
        break;

    case Variant::Type::Int:
        byteSizeOfArguments += sizeof(int);
        break;

    case Variant::Type::Float:
        byteSizeOfArguments += sizeof(float);
        break;

    case Variant::Type::String:
        byteSizeOfArguments += static_cast<std::string>(argument).size();
        break;

    default:
        EAR_FAIL("Unsupported argument type added to OSC message.");
        break;
    }

    //OSCPack includes a char flag with each type
    byteSizeOfArguments += sizeof(char);

    return *this;
}

const std::string& OSC::Message::getOscAddress() const
{
    return address;
}

const std::vector<Variant> OSC::Message::getArguments() const
{
    return arguments;
}

std::vector<Variant>::const_iterator OSC::Message::begin() const
{
    return arguments.begin();
}

std::vector<Variant>::const_iterator OSC::Message::end() const
{
    return arguments.end();
}

std::vector<Variant>::const_iterator OSC::Message::cbegin() const
{
    return arguments.cbegin();
}

std::vector<Variant>::const_iterator OSC::Message::cend() const
{
    return arguments.cend();
}

//==============================================================================

bool OSC::Bundle::Element::isBundle() const
{
    return bundle != nullptr;
}

bool OSC::Bundle::Element::isMessage() const
{
    return message != nullptr;
}

const OSC::Message& OSC::Bundle::Element::getMessage() const
{
    EAR_CHECK(isMessage(), "Element is not a message");
    return *message;
}

const OSC::Bundle& OSC::Bundle::Element::getBundle() const
{
    EAR_CHECK(isBundle(), "Element is not a bundle");
    return *bundle;
}

OSC::Bundle::Element::Element(const OSC::Message& _message)
    : message(std::make_shared<OSC::Message>(_message))
{}

OSC::Bundle::Element::Element(const OSC::Bundle& _bundle)
    : bundle(std::make_shared<OSC::Bundle>(_bundle))
{}

//==============================================================================

OSC::Bundle& OSC::Bundle::addElement(const OSC::Bundle& element)
{
    elements.push_back(OSC::Bundle::Element(element));

    return *this;
}

OSC::Bundle& OSC::Bundle::addElement(const OSC::Message& element)
{
    elements.push_back(OSC::Bundle::Element(element));

    return *this;
}

const std::vector<OSC::Bundle::Element>& OSC::Bundle::getElements() const
{
    return elements;
}

std::vector<OSC::Bundle::Element>::const_iterator OSC::Bundle::begin() const
{
    return elements.begin();
}

std::vector<OSC::Bundle::Element>::const_iterator OSC::Bundle::end() const
{
    return elements.end();
}

std::vector<OSC::Bundle::Element>::const_iterator OSC::Bundle::cbegin() const
{
    return elements.cbegin();
}

std::vector<OSC::Bundle::Element>::const_iterator OSC::Bundle::cend() const
{
    return elements.cend();
}

std::string OSC::Bundle::serialize() const
{
    std::stringstream out;
    out << "Bundle(" << timetag << "," << std::endl;

    for (auto& el : elements) {
        if (el.isMessage()) {
            out << el.getMessage().serialize() << "," << std::endl;
        } else { // is bundle
            out << el.getBundle().serialize() << "," << std::endl;
        }
    }

    out << ")";

    return out.str();
}

OSC::Bundle OSC::Bundle::deserialize(const std::string& val)
{
    OSC::Bundle result;

    std::stringstream input(val);

    std::string line;
    std::getline(input, line); // first line (expected to be "Bundle(t,")
    EAR_CHECK_REL(line.find("Bundle(") != std::string::npos,
        "Unexpected string format in OSC::Bundle::deserialize().")

    // Parse the first bundle line
    detail::eraseSubStr(line, "Bundle(");
    detail::eraseSubStr(line, ",");
    result.timetag = std::stoi(line); // TODO: problems with stoi and uint64?

    // handle subelements
    std::vector<std::pair<bool, std::string>> subElements;

    int openBrackets = 0;
    while (std::getline(input, line)) {
        if (line.find("Bundle(") != std::string::npos) {
            if (openBrackets == 0) {
                subElements.emplace_back(true, "");
            }

            openBrackets++;
        } else if (line.find("Message(") != std::string::npos) {
            if (openBrackets == 0) {
                subElements.emplace_back(false, "");
            }

            openBrackets++;
        } else if (line.find(")") != std::string::npos) {
            openBrackets--;
        }

        subElements.back().second += line + "\n";
    }

    for (auto& pair : subElements) {
        if (pair.first) { // is bundle
            result.addElement(OSC::Bundle::deserialize(pair.second));
        } else { // is message
            result.addElement(OSC::Message::deserialize(pair.second));
        }
    }

    return result;
}

std::string OSC::Message::serialize() const
{
    std::stringstream out;
    out << "Message(\"" << address << "\"," << std::endl;

    for (auto& arg : arguments) {
        switch (arg.getVariantType()) {
        case Variant::Type::Bool:
            out << "   bool: " << static_cast<bool>(arg) << "," << std::endl;
            break;
        case Variant::Type::Float:
            out << std::fixed << std::setprecision(6) << "   float: " << static_cast<float>(arg)
                << "," << std::endl;
            break;
        case Variant::Type::Int:
            out << "   int: " << static_cast<int>(arg) << "," << std::endl;
            break;
        case Variant::Type::String:
            out << "   str: \"" << static_cast<std::string>(arg) << "\"," << std::endl;
            break;
        default:
            EAR_FAIL("Unexpected argument in OSC::Message::serialize().");
            break;
        }
    }

    out << ")";

    return out.str();
}

OSC::Message OSC::Message::deserialize(const std::string& val)
{
    std::stringstream input(val);

    OSC::Message result("");

    std::string line;
    while (std::getline(input, line)) {
        detail::eraseSubStr(line, ",");

        if (line.find("Message(") != std::string::npos) {
            detail::eraseSubStr(line, "Message(");
            detail::eraseSubStr(line, "\"");
            line.erase(std::remove_if(line.begin(), line.end(), isspace), line.end());
            result.address = line;
            result.byteSizeOfArguments = 0;
            result.byteSizeOfArguments +=
                sizeof(osc::BeginMessage(address.c_str())) + sizeof(osc::EndMessage);
        } else if (line.find("bool:") != std::string::npos) {
            detail::eraseSubStr(line, "bool:");
            line.erase(std::remove_if(line.begin(), line.end(), isspace), line.end());
            result.addArgument(EAR::Variant(detail::string2bool(line)));
        } else if (line.find("float:") != std::string::npos) {
            detail::eraseSubStr(line, "float:");
            line.erase(std::remove_if(line.begin(), line.end(), isspace), line.end());
            result.addArgument(EAR::Variant(std::stof(line)));
        } else if (line.find("int:") != std::string::npos) {
            detail::eraseSubStr(line, "int:");
            line.erase(std::remove_if(line.begin(), line.end(), isspace), line.end());
            result.addArgument(EAR::Variant(std::stoi(line)));
        } else if (line.find("str:") != std::string::npos) {
            detail::eraseSubStr(line, "str:");
            detail::eraseSubStr(line, "\"");
            detail::trim(line);
            result.addArgument(EAR::Variant(line));
        }
    }

    return result;
}

//==============================================================================

OSC::OSC(std::string _ipAddress, int _port, int defaultPacketSize, bool disableListening,
    bool autoConnect)
    : packetSize(defaultPacketSize)
    , packetBuffer(std::make_unique<char[]>(packetSize))
    , packet(std::make_unique<osc::OutboundPacketStream>(packetBuffer.get(), packetSize))
    , listener(std::make_unique<PacketListener>(*this))
    , ipAddress(_ipAddress)
    , port(_port)
    , recording(false)
    , disableListening(disableListening)
{

    if (autoConnect) {
        connect();
    }
}

OSC::~OSC()
{
    //Break the thread that was running and wait for it to join
    if (listenerThread) {
        receiveSocket->AsynchronousBreak();
        listenerThread->join();
    }

    transmitSocket.reset(nullptr);
    receiveSocket.reset(nullptr);
    listenerThread.reset(nullptr);
}

void OSC::send(const Bundle& oscBundle)
{
    packet->Clear();

    streamBundleIntoPacket(oscBundle);

    transmitSocket->Send(packet->Data(), packet->Size());
}

void OSC::send(const Message& message)
{
    EAR_CHECK(transmitSocket.get() != nullptr, "Not connected to network.");
    packet->Clear();

    streamMessageIntoPacket(message);
    transmitSocket->Send(packet->Data(), packet->Size());
}

void OSC::addListener(std::string oscAddress, std::function<void(const Message& message)> callback)
{
    listenerCallbacks.push_back(CallbackWithAddress(oscAddress, callback));
}

void OSC::startRecordingReceivedPackets()
{
    std::lock_guard<std::mutex> guard(receiveMutex);

    recordedOscElements.clear();
    recordingStartTime = std::chrono::system_clock::now();
    recording = true;
}

void OSC::stopRecordingReceivedPackets()
{
    std::lock_guard<std::mutex> guard(receiveMutex);
    recording = false;
}

std::string OSC::serializeRecording() const
{
    EAR_CHECK_REL(!recording, "OSC Recording can not be serialized while it is running.");

    std::stringstream out;

    for (const auto& element : recordedOscElements) {
        out << "Record(t=" << element.first << "," << std::endl;

        if (element.second.isBundle()) {
            out << element.second.bundle->serialize() << std::endl;
        } else { // isMessage() == true implied
            out << element.second.message->serialize() << std::endl;
        }

        out << ")" << std::endl;
    }

    return out.str();
}

std::vector<std::pair<double, OSC::Bundle::Element>> OSC::deserializeRecording(
    const std::string& val)
{

    std::vector<std::pair<double, OSC::Bundle::Element>> elements;

    std::stringstream input(val);
    std::string line;

    std::getline(input, line);
    EAR_CHECK_REL(line.find("Record(") != std::string::npos,
        "Unexpected string format in OSC::deserializeRecording().");
    input.seekg(0, std::ios::beg);

    std::vector<std::pair<double, std::string>> records;

    int openBrackets = 0;
    while (std::getline(input, line)) {
        if (line.find("Record(") != std::string::npos) {
            if (openBrackets == 0) {
                detail::eraseSubStr(line, "Record(t=");
                detail::eraseSubStr(line, ",");
                records.emplace_back(std::stof(line), "");
            } else {
                EAR_FAIL("Unexpected string format in OSC::deserializeRecording().")
            }

            openBrackets++;
            continue;
        } else if (line.find("(") != std::string::npos) {
            openBrackets++;
        } else if (line.find(")") != std::string::npos) {
            openBrackets--;
        }

        if (openBrackets != 0) {
            records.back().second += line + "\n";
        }
    }

    for (auto& pair : records) {
        std::size_t posBundle = pair.second.find("Bundle(");
        std::size_t posMessage = pair.second.find("Message(");
        if (posBundle == std::string::npos) {
            posBundle = posMessage + 1;
        }

        if (posBundle > posMessage) { // is message
            elements.emplace_back(pair.first, OSC::Message::deserialize(pair.second));
        } else { // is bundle
            elements.emplace_back(pair.first, OSC::Bundle::deserialize(pair.second));
        }
    }

    return elements;
}

const std::vector<std::pair<float, OSC::Bundle::Element>>& OSC::getRecordedOscElements() const
{
    EAR_CHECK_REL(!recording, "OSC Recording can not accessed while it is running.");

    return recordedOscElements;
}

void OSC::changePort(int newPort, bool autoConnect)
{
    port = newPort;
    if (autoConnect) {
        connect();
    }
}

void OSC::changeIpAddress(std::string newIpAddress, bool autoConnect)
{
    ipAddress = newIpAddress;
    if (autoConnect) {
        connect();
    }
}

void OSC::connect()
{
    if (listenerThread) {
        //Break the listener then wait for the other thread to join
        receiveSocket->AsynchronousBreak();
        listenerThread->join();
    }

    //Connect sockets to current IP address and UDP port
    transmitSocket.reset(new UdpTransmitSocket(IpEndpointName(ipAddress.c_str(), port)));
    if (!disableListening) {
        receiveSocket.reset(
            new UdpListeningReceiveSocket(IpEndpointName(ipAddress.c_str(), port), listener.get()));

        //Create a thread to listen to incoming packets
        listenerThread.reset(new std::thread(&UdpListeningReceiveSocket::Run, receiveSocket.get()));

        // try to increase listener thread priority
        /*struct sched_param params;
        params.sched_priority = sched_get_priority_max(SCHED_RR);
        int ret = pthread_setschedparam(listenerThread->native_handle(), SCHED_RR, &params);*/
    }
}

void OSC::receive(const osc::ReceivedMessage& message, bool isRecursiveCall)
{
    if (!isRecursiveCall) {
        receiveMutex.lock();
    }

    //Check the message address with all callbacks and execute the callback if they match
    for (auto callback : listenerCallbacks) {
        if (callback.first == message.AddressPattern()) {
            //Execute the callback and pass it an EAR OSC message created from the received OSCPack message
            callback.second(getMessageFromReceivedMessage(message));
        }
    }

    if (recording && !isRecursiveCall) {
        auto now = std::chrono::system_clock::now();
        float diffTime =
            std::chrono::duration_cast<std::chrono::microseconds>(now - recordingStartTime).count();

        recordedOscElements.push_back(std::make_pair(
            diffTime / 1000000.f, Bundle::Element(getMessageFromReceivedMessage(message))));
    }

    if (!isRecursiveCall) {
        receiveMutex.unlock();
    }
}

void OSC::receive(const osc::ReceivedBundle& bundle, bool isRecursiveCall)
{
    if (!isRecursiveCall) {
        receiveMutex.lock();
    }

    for (auto element = bundle.ElementsBegin(); element != bundle.ElementsEnd(); ++element) {
        if (element->IsBundle()) {
            receive(osc::ReceivedBundle(*element), true);
        } else { //element->IsMessage() implied
            receive(osc::ReceivedMessage(*element), true);
        }
    }

    if (recording && !isRecursiveCall) {
        auto now = std::chrono::system_clock::now();
        float diffTime =
            std::chrono::duration_cast<std::chrono::microseconds>(now - recordingStartTime).count();

        recordedOscElements.push_back(std::make_pair(
            diffTime / 1000000.f, Bundle::Element(getBundleFromReceivedBundle(bundle))));
    }

    if (!isRecursiveCall) {
        receiveMutex.unlock();
    }
}

void OSC::streamBundleIntoPacket(const Bundle& bundle)
{
    *packet << osc::BeginBundleImmediate;

    for (auto element : bundle) {
        if (element.isBundle()) {
            streamBundleIntoPacket(*element.bundle);
        } else { //element.isMessage() implied
            streamMessageIntoPacket(*element.message);
        }
    }

    *packet << osc::EndBundle;
}

void OSC::streamMessageIntoPacket(const Message& message)
{
    try {
        *packet << osc::BeginMessage(message.address.c_str());

        for (auto argument : message) {
            switch (argument.getVariantType()) {
            case Variant::Type::Bool:
                *packet << static_cast<bool>(argument);
                break;

            case Variant::Type::Int:
                *packet << static_cast<int>(argument);
                break;

            case Variant::Type::Float:
                *packet << static_cast<float>(argument);
                break;

            case Variant::Type::String:
                *packet << static_cast<std::string>(argument).c_str();
                break;

            default:
                break;
            }
        }

        *packet << osc::EndMessage;
    } catch (osc::Exception& e) {
        //If the packet is too small for the message, resize the packet and restream the message
        resizePacketToAccomodateMessage(message);
        streamMessageIntoPacket(message);
    }
}

OSC::Message OSC::getMessageFromReceivedMessage(const osc::ReceivedMessage& rMessage)
{
    Message message(rMessage.AddressPattern());

    for (auto arg = rMessage.ArgumentsBegin(); arg != rMessage.ArgumentsEnd(); arg++) {
        if (arg->IsBool()) {
            message.addArgument(arg->AsBool());
        } else if (arg->IsInt32()) {
            message.addArgument(arg->AsInt32());
        } else if (arg->IsFloat()) {
            message.addArgument(arg->AsFloat());
        } else if (arg->IsString()) {
            message.addArgument(static_cast<std::string>(arg->AsString()));
        } else {
            EAR_FAIL("Invalid argument.");
        }
    }

    return message;
}

OSC::Bundle OSC::getBundleFromReceivedBundle(const osc::ReceivedBundle& rBundle)
{
    Bundle bundle;

    //Iterate through all received elements within a bundle and add them to the newly created bundle
    for (auto element = rBundle.ElementsBegin(); element != rBundle.ElementsEnd(); element++) {
        const osc::ReceivedPacket packet(element->Contents(), element->Size());

        if (element->IsBundle()) {
            osc::ReceivedBundle rNestedBundle(packet);
            Bundle nestedBundle(getBundleFromReceivedBundle(rNestedBundle));
            nestedBundle.timetag = rBundle.TimeTag();
            bundle.addElement(nestedBundle);
        } else { // element->IsMessage() implied
            osc::ReceivedMessage rMessage(packet);
            Message message(getMessageFromReceivedMessage(rMessage));
            bundle.addElement(message);
        }
    }

    return bundle;
}

void OSC::resizePacketToAccomodateMessage(const Message& message)
{
    //+ 2 is arbitrarily defined in the Size() of OutboundPacketStream
    packetSize = message.byteSizeOfArguments + 2;
    packetSize += message.address.size();
    packetBuffer.reset(new char[packetSize]);
    packet.reset(new osc::OutboundPacketStream(packetBuffer.get(), packetSize));
}

} // namespace EAR
