/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/PchipInterpolator.h>

#include <EAR/Exception.h>

#include <algorithm> // for std::lower_bound, std::binary_search, std::find, std::max
#include <cmath> // for std::abs

#include <iostream>

namespace EAR {

PchipInterpolator::PchipInterpolator(bool bindBoundaryValues_)
{
    bindBoundaryValues = bindBoundaryValues_;
}

PchipInterpolator::PchipInterpolator(
    const float* x, const float* y, std::size_t n, bool bindBoundaryValues_)
{
    bindBoundaryValues = bindBoundaryValues_;
    initialize(x, y, n);
}

void PchipInterpolator::initialize(const float* x, const float* y, std::size_t n)
{

    xs.resize(n);
    std::copy(x, x + n, xs.data());

    ys.resize(n);
    std::copy(y, y + n, ys.data());

    if (n >= 2) {

        // make sure to initialize dy with all zeros
        dy.clear();
        dy.resize(n);

        float step1 = xs[1] - xs[0];
        float d1 = (ys[1] - ys[0]) / step1;

        // only two supporting point reduce to linear interpolation
        if (n == 2) {
            dy[0] = dy[1] = d1;
            return;
        }

        // from here we're sure to have at least 3 supporting points
        // and start to calculate the 1st derivative at the first supporting point
        float step2 = xs[2] - xs[1];
        float d2 = (ys[2] - ys[1]) / step2;

        // non-centered three-point formula, adjusted to be shape-preserving.
        float step = step1 + step2;
        float w1 = (step1 + step) / step;
        float w2 = -1 * step1 / step;
        dy[0] = w1 * d1 + w2 * d2;

        float dmin, dmax;

        if (signTest(dy[0], d1) <= 0) {
            dy[0] = 0;
        } else if (signTest(d1, d2) < 0) {
            // if the function is not monotone (sign of slopes between points switches),
            // we need to limit the 1st derivative in the previous point to allow a smooth transition at the following point
            dmax = 3 * d1;

            if (std::abs(dy[0]) > std::abs(dmax)) {
                dy[0] = dmax;
            }
        }

        // now for the supporting points with two neighbours:
        for (std::size_t i = 1; i < dy.size() - 1; i++) {
            // calculate step lengths and slopes for next step
            if (i > 1) {
                step1 = step2;
                step2 = xs[i + 1] - xs[i];
                step = step1 + step2;
                d1 = d2;
                d2 = (ys[i + 1] - ys[i]) / step2;
            }

            if ((signTest(d1, d2) == -1 && d2 != 0) || signTest(d1, d2) == 0) {
                // sign switches between segments or both segments have 0 slope: derivative shall stay zero (as dy is initialized)
            } else if (signTest(d1, d2) == 1) {
                // same direction of slope for both segments
                // use brodlie modification of butland formula.
                w1 = (1 + step1 / step) / 3;
                w2 = (1 + step2 / step) / 3;
                dmax = absmax(d1, d2);
                dmin = absmin(d1, d2);
                dy[i] = dmin / (w1 * (d1 / dmax) + w2 * (d2 / dmax));
            }
        }

        // ...and the last one
        // three-point formula as above for first point
        w1 = -1 * step2 / step;
        w2 = (step2 + step) / step;
        dy[n - 1] = w1 * d1 + w2 * d2;

        if (signTest(dy[n - 1], d2) <= 0) {
            dy[n - 1] = 0;
        } else if (signTest(d1, d2) < 0) {
            dmax = 3 * d2;

            if (std::abs(dy[n - 1]) > std::abs(dmax)) {
                dy[n - 1] = dmax;
            }
        }
    }
}

void PchipInterpolator::addSupportingPoint(float x, float y)
{
    std::vector<float> newxs = xs;
    std::vector<float> newys = ys;

    // if x already exists in the sorted vector, replace the corresponding y
    if (std::binary_search(newxs.begin(), newxs.end(), x)) {
        int idx = std::find(newxs.begin(), newxs.end(), x) - newxs.begin();
        newys[idx] = y;
    } else { // else find new spot for x in newxs and insert into newxs and newys accordingly
        int idx = std::lower_bound(newxs.begin(), newxs.end(),
                      x)
            - newxs.begin(); // taking the detour over index to use it for newys too
        newxs.insert(newxs.begin() + idx, x);
        newys.insert(newys.begin() + idx, y);
    }

    initialize(newxs.data(), newys.data(), newxs.size());
}

float PchipInterpolator::interpolate(float x)
{
    if (xs.size() == 0) {
        return 0.f;
    }

    if (xs.size() == 1) {
        return ys[0];
    }

    // extrapolate if x value is out of bounds
    // if bindBoundaryValues equals false: linearly
    // if bindBoundaryValues equals true: fix value
    if (x < xs.front()) {
        if (bindBoundaryValues) {
            return ys.front();
        } else {
            return ys.front() + (x - xs.front()) * dy.front();
        }
    }

    if (x > xs.back()) {
        if (bindBoundaryValues) {
            return ys.back();
        } else {
            return ys.back() + (x - xs.back()) * dy.back();
        }
    }

    auto i = findSection(x);

    // calculate the first 4 cubic Hermite basis functions at the local x (relative to selected section)
    float H[4];
    float xlocal = x - xs[i];
    float xnorm = xlocal / (xs[i + 1] - xs[i]);
    H[1] = xnorm * xnorm * (3 - (2 * xnorm));
    H[0] = 1 - H[1];
    H[2] = xlocal * (xnorm * (xnorm - 2) + 1);
    H[3] = xlocal * xnorm * (xnorm - 1);

    // ...and evaluate with our precalculated support values
    return H[0] * ys[i] + H[1] * ys[i + 1] + H[2] * dy[i] + H[3] * dy[i + 1];
}

inline std::size_t PchipInterpolator::findSection(float x) const
{
    auto i = std::lower_bound(xs.begin(), xs.end(), x) - xs.begin();

    if (i > 0) {
        --i;
    }

    return static_cast<std::size_t>(i);
}

inline float PchipInterpolator::absmax(float a, float b) const
{
    float res = std::abs(a);

    if (res < std::abs(b)) {
        res = std::abs(b);
    }

    return res;
}

inline float PchipInterpolator::absmin(float a, float b) const
{
    float res = std::abs(a);

    if (res > std::abs(b)) {
        res = std::abs(b);
    }

    return res;
}

inline int PchipInterpolator::signTest(float a, float b) const
{
    int sa = 0;
    int sb = 0;

    if (a > 0) {
        sa = 1;
    } else if (a < 0) {
        sa = -1;
    }

    if (b > 0) {
        sb = 1;
    } else if (b < 0) {
        sb = -1;
    }

    return sa * sb;
}

} // namespace EAR
