/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/PchipMinphaseFilterDesigner.h>

#include <EAR/config.h>

#include <algorithm>
#include <cmath>
#include <complex>
#include <numeric>

#if defined(EAR_ENABLE_IPP)
#include <ipp.h>
#include <ipps.h>
#endif

namespace EAR {

const std::vector<float> PchipMinphaseFilterDesigner::defaultCenterFrequencies({ 19.69f, 24.80f,
    31.25f, 39.37f, 49.61f, 62.50f, 78.75f, 99.21f, 125.00f, 157.49f, 198.43f, 250.00f, 314.98f,
    396.85f, 500.00f, 629.96f, 793.70f, 1000.00f, 1259.92f, 1587.40f, 2000.00f, 2519.84f, 3174.80f,
    4000.00f, 5039.68f, 6349.60f, 8000.00f, 10079.37f, 12699.21f, 16000.00f, 20158.74f });

PchipMinphaseFilterDesigner::PchipMinphaseFilterDesigner(
    const float* f, std::size_t H, std::size_t N, float fs)
    : FilterDesigner(f, H, N)
    , K((N + 2) / 2)
    , temp(K)
    , logSignal(K)
    , spectrum(K)
    , cepstrum(N)
    , cepstrumWin(N)
{

    fft = RFFT::create(N);
    ifft = IRFFT::create(N);

    pchip = std::make_unique<PchipInterpolator>(false);
    binf = EAR::RFFT::binFrequencies(K, fs);

    // create the window to get the left half of the cepstrum according to Ref [1]
    int odd = N % 2;
    cepstrumWin[0] = 1;
    cepstrumWin.fill(2, 1, std::floor((N + odd) / 2 - 1));

    if (!odd) {
        // we're sure that N is even
        cepstrumWin[N / 2] = 1;
    }
}

PchipMinphaseFilterDesigner::PchipMinphaseFilterDesigner(std::size_t N, float fs)
    : PchipMinphaseFilterDesigner(
        defaultCenterFrequencies.data(), defaultCenterFrequencies.size(), N, fs)
{}

PchipMinphaseFilterDesigner::~PchipMinphaseFilterDesigner() {}

void PchipMinphaseFilterDesigner::generate(const float* m, RealBuffer& result)
{
    float magnitudeSum = std::accumulate(m, m + H, 0.f);

    if (magnitudeSum != 0) { // pchip does not work if all supporting points are 0

        // 1. interpolate spectrum
        pchip->initialize(centerFreqs.data(), m, H);

        for (unsigned int i = 0; i < K; i++) {
            spectrum[i] = std::max(0.00000001f, pchip->interpolate(binf[i]));
        }

        // 2. cepstrum transform
        temp.zero();

#if defined(EAR_ENABLE_IPP)

        // ipp real natural logarithm function with precision to 3 decimal digits (could be replaced by A21 for 6 decimal digits or A24)
        ippsLn_32f_A11(spectrum.data(), logSignal.data(), K);
        temp.writeReal(logSignal.data(), 0, K);

#else // fallback logarithm

        temp.writeReal(spectrum.data(), 0, K);
        std::transform(temp.data(), temp.data() + K, temp.data(), [](auto a) {
            // TODO: use a fast approximation of real natural log
            a.real(std::log(a.real()));
            return a;
        });

#endif

        ifft->transform(temp.data(), cepstrum.data());
        // denormalization is required directly after the ifft to continue with correct values in the nonlinear exp function
        cepstrum.mul(1.f / (2 * K));

        // 3. apply window
        cepstrum.mul(cepstrumWin);

        // 4. inverse cepstrum transform
        fft->transform(cepstrum.data(), temp.data());

#if defined(EAR_ENABLE_IPP)

        // ipp complex exponential function with precision to 3 decimal digits (could be replaced by A21 for 6 decimal digits or A24)
        ippsExp_32fc_A11(
            reinterpret_cast<Ipp32fc*>(temp.data()), reinterpret_cast<Ipp32fc*>(temp.data()), K);

#else // fallback exponential

        std::transform(temp.data(), temp.data() + K, temp.data(), [](auto a) {
            // TODO: use a fast approximation of complex exp function
            return std::exp(a);
        });

#endif

        // inverse fourier transform to time domain
        ifft->transform(temp.data(), result.data());
        result.mul(1.f / (2 * K));

    } else { // all values in m are zero
        result.zero();
    }
}

} // namespace EAR
