/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/PlaneAirflowGenerator.h>

#include <cmath>

namespace EAR {

PlaneAirflowGenerator::PlaneAirflowGenerator(ObjectID oid, const std::string& name, Position pos)
    : AirflowGeneratorBase(oid, name, pos)
{}

Vector3 PlaneAirflowGenerator::getAirspeed(const Vector3& listenerPosition) const
{
    // Define a local coordinate system along the plane
    const Vector3 planeX = eulerToVec3(
        getOrientation(), { 1, 0, 0 }); // Goes towards "right" (relative to the generator)
    const Vector3 planeY =
        eulerToVec3(getOrientation(), { 0, 1, 0 }); // Goes towards "up" (relative to the generator)
    const Vector3 planeNormal = eulerToVec3(
        getOrientation(), { 0, 0, -1 }); // Goes towards "forwards" (relative to the generator)

    const Vector3 relative = listenerPosition - getLocation();

    // Compute relative dimensions in the local plane coordinate system
    const float x = relative.dot(planeX);
    const float y = relative.dot(planeY);
    const float z = relative.dot(planeNormal);

    const float halfWidth = 0.5f * width;
    const float halfHeight = 0.5f * height;

    const float rollOffWidth = halfWidth * edgeRollOff / 100.f;
    const float rollOffHeight = halfHeight * edgeRollOff / 100.f;

    if (std::abs(x) > halfWidth + rollOffWidth || std::abs(y) > halfHeight + rollOffHeight
        || (z < 0)) {
        // Listener is either outside the plane and outside the edge rolloff area, and/or behind the plane.
        return { 0, 0, 0 };
    }

    float fadeFactor = 1.f;

    if (std::abs(x) > halfWidth) {
        // Within x rolloff area
        fadeFactor = 1.f - ((std::abs(x) - halfWidth) / rollOffWidth);
    }

    if (std::abs(y) > halfHeight) {
        // Within y rolloff area. Use minimum of x and y rolloff.
        fadeFactor = std::min(fadeFactor, 1.f - ((std::abs(y) - halfHeight) / rollOffHeight));
    }

    const float maxSpeed = getSpeed() * fadeFactor;
    float airspeed =
        getSpeed() * (1 / std::pow((z + 1), 2 * getDistanceAttenuationFactor())) * fadeFactor;
    if (airspeed > maxSpeed) {
        airspeed = maxSpeed;
    }
    return airspeed * planeNormal;
}

} /* namespace EAR */
