/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include <EAR/PortalReverbExtent.h>

#include <EAR/Log.h>
#include <EAR/Mesh.h>

#include <algorithm>
#include <cmath>
#include <iostream>
#include <math.h>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES
namespace EAR {

int addVertexConstrained(Mesh& mesh, const Geometry* constrainingGeometry,
    const Vector3& localVertex, const bool insideFace, const float backFaceOffset = 0)
{
    // minimum distance that the mesh should be inside of the constraining geometry.
    // TODO: decide what this should be, currenlty chosen to ensure hospital works with current AE / portal definitions.
    float minDistInside = 0.15f;
    EAR::Vector3 vertexToAdd;
    if (!insideFace && backFaceOffset > 0.0f) {
        auto t = (localVertex.abs() + backFaceOffset) / localVertex.abs();
        vertexToAdd.set(
            (t * localVertex.data()[0]), (t * localVertex.data()[1]), (t * localVertex.data()[2]));
    } else {
        vertexToAdd = localVertex;
    }

    if (constrainingGeometry) {
        // Check if vertex is inside or outside constraining geometry.
        const Vector3 globalVertex = mesh.getGlobalCSTransform() * vertexToAdd;
        bool doConstraint = true;
        if (constrainingGeometry->isInside(globalVertex)) {
            // make sure it is at least 5cm inside the geometry.
            if ((constrainingGeometry->closestPointTo(globalVertex) - globalVertex).abs()
                > minDistInside + 0.001f) {
                doConstraint = false;
            }
        }

        if (doConstraint) {
            // Point is outside constraining geometry. Find closest point that is inside and add that instead.
            const float factor = insideFace ? minDistInside + backFaceOffset : minDistInside;
            const Vector3 globalClosest = constrainingGeometry->closestPointTo(globalVertex);
            const Vector3 localClosest = (mesh.getLocalCSTransform() * globalClosest); //
            const Vector3 vecToClosest = localClosest - vertexToAdd;
            Vector3 offsetAmount;
            offsetAmount[0] = std::copysign(factor, vertexToAdd.data()[0]);
            offsetAmount[1] = std::copysign(factor, vertexToAdd.data()[1]);
            // only shift in Z dimension if needed. Should only happen if the portal is not in a flat wall.
            offsetAmount[2] = std::abs(vecToClosest[2]) > 0.005f
                ? std::copysign(factor, -vertexToAdd.data()[2])
                : 0;
            vertexToAdd = localClosest - offsetAmount;
        }
    }
    return mesh.addVertex(vertexToAdd);
}

PortalReverbExtent::PortalReverbExtent(ObjectID oid, EAR::Vector3 size, float& offset,
    const std::string& name, Position pos, int segments, float backFaceOffset,
    const Geometry* constrainingGeometry)
    : Geometry(oid, name, pos)
    , volume(size)
    , mesh(std::unique_ptr<Mesh>(SceneObjectFactory<Mesh>::getNewObject(name + "_mesh", pos)))
{
    float hw = size[0] / 2;
    float hh = size[1] / 2;
    float hd = (size[2] / 2)
        + 0.1; // moves the mesh 10cm inside of the room, to ensure not touching the wall.
    float irt2 = 1 / std::sqrt(2.f);

    if (constrainingGeometry) {
        // If needed, reduce the reverb extent offset so that we're never further into the room than half the available room.

        // Define four points at the corners of the portal, 2x the reverb extent into the room
        Vector3 checkCorners[4] = { Vector3(-hw, +hh, -hd - 2.0f * offset),
            Vector3(-hw, -hh, -hd - 2.0f * offset), Vector3(+hw, +hh, -hd - 2.0f * offset),
            Vector3(+hw, -hh, -hd - 2.0f * offset) };

        for (int cornerIdx = 0; cornerIdx < 4; cornerIdx++) {
            const Vector3 globalVertex = mesh->getGlobalCSTransform() * checkCorners[cornerIdx];
            if (!constrainingGeometry->isInside(globalVertex)) {
                // Point is outside constraining geometry.
                const Vector3 globalClosest = constrainingGeometry->closestPointTo(globalVertex);
                const Vector3 localClosest = (mesh->getLocalCSTransform() * globalClosest);

                offset = std::min(offset, -1.0f * ((localClosest[2] + hd) / 2.0f));
            }
        }
    }

    float ho = offset / 2;
    float cos30o = offset * std::cos(30 * M_PI / 180.0f);

    if (!(segments == 0 || segments == 1 || segments == 2 || segments == 3)) {
        EAR_FAIL(
            "PortalReverbExtent segments other than 0, 1, 2 or 3 are currently not implemented")
    }
    if (segments == 0) {
        if (backFaceOffset > 0.0f) {
            //  Flat box +1m in width / height dimension.
            //    4----------5
            //   /|         /|
            //  0 -------- 1 |
            //  | |        | |
            //  | |        | |
            //  | |        | |
            //  | 6--------|-7
            //  |/         |/
            //  2 -------- 3

            int v0 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(-hw - 1.f, hh + 1.f, -hd - offset), true);
            int v1 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(hw + 1.f, hh + 1.f, -hd - offset), true);
            int v2 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(-hw - 1.f, -hh - 1.f, -hd - offset), true);
            int v3 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(hw + 1.f, -hh - 1.f, -hd - offset), true);

            int v4 = addVertexConstrained(*mesh, constrainingGeometry,
                Vector3(-hw - 1.f, hh + 1.f, -hd - offset - backFaceOffset), false);
            int v5 = addVertexConstrained(*mesh, constrainingGeometry,
                Vector3(hw + 1.f, hh + 1.f, -hd - offset - backFaceOffset), false);
            int v6 = addVertexConstrained(*mesh, constrainingGeometry,
                Vector3(-hw - 1.f, -hh - 1.f, -hd - offset - backFaceOffset), false);
            int v7 = addVertexConstrained(*mesh, constrainingGeometry,
                Vector3(hw + 1.f, -hh - 1.f, -hd - offset - backFaceOffset), false);

            // Front face
            mesh->addTriangle(v0, v2, v1);
            mesh->addTriangle(v1, v2, v3);

            // Top face
            mesh->addTriangle(v0, v5, v4);
            mesh->addTriangle(v0, v1, v5);

            // Right face
            mesh->addTriangle(v1, v3, v5);
            mesh->addTriangle(v3, v7, v5);

            // Left face
            mesh->addTriangle(v0, v4, v2);
            mesh->addTriangle(v4, v6, v2);

            // Back face
            mesh->addTriangle(v5, v7, v4);
            mesh->addTriangle(v4, v7, v6);

            // Bottom face
            mesh->addTriangle(v6, v7, v2);
            mesh->addTriangle(v2, v7, v3);

        } else {

            //  Flat plane +1m in all directions.
            //  0 -------- 1
            //  |          |
            //  |          |
            //  |          |
            //  |          |
            //  |          |
            //  2 -------- 3

            // (-hw-1, hd+offset, hh+1), # 0
            // ( hw+1, hd+offset, hh+1),  # 1
            // (-hw-1, hd+offset, -hh-1), # 2
            // ( hw+1, hd+offset, -hh-1), # 3

            int v0 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(-hw - 1.f, hh + 1.f, -hd - offset), false);
            int v1 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(hw + 1.f, hh + 1.f, -hd - offset), false);
            int v2 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(-hw - 1.f, -hh - 1.f, -hd - offset), false);
            int v3 = addVertexConstrained(
                *mesh, constrainingGeometry, Vector3(hw + 1.f, -hh - 1.f, -hd - offset), false);

            mesh->addTriangle(v0, v2, v1);
            mesh->addTriangle(v1, v2, v3);
        }

    } else if (segments == 1) {

        std::vector<EAR::Vector3> VertCoords{ Vector3(-hw, hh, -hd - offset),
            Vector3(hw, hh, -hd - offset), Vector3(-hw, -hh, -hd - offset),
            Vector3(hw, -hh, -hd - offset), Vector3(-hw, hh + offset, -hd),
            Vector3(hw, hh + offset, -hd), Vector3(-hw - offset, hh, -hd),
            Vector3(hw + offset, hh, -hd), Vector3(-hw - offset, -hh, -hd),
            Vector3(hw + offset, -hh, -hd), Vector3(-hw, -hh - offset, -hd),
            Vector3(hw, -hh - offset, -hd) };

        std::vector<std::vector<int>> faceIdxs{ { 0, 2, 1 }, { 1, 2, 3 }, { 0, 5, 4 }, { 1, 5, 0 },
            { 0, 6, 8 }, { 0, 8, 2 }, { 3, 2, 10 }, { 3, 10, 11 }, { 7, 1, 3 }, { 7, 3, 9 },
            { 4, 6, 0 }, { 5, 1, 7 }, { 2, 8, 10 }, { 9, 3, 11 } };

        for (EAR::Vector3 v : VertCoords) {
            addVertexConstrained(*mesh, constrainingGeometry, v, true, backFaceOffset);
        }

        for (std::vector<int> v : faceIdxs) {
            mesh->addTriangle(v[0], v[1], v[2]);
        }

        if (backFaceOffset > 0.0f) {
            //
            //            16---------17
            //           /|         /|
            //          4 -------- 5 |
            //     18---|-12 ------|-13-----19
            //    /|    |/|        |/|     /|
            //   6 ---- 0 -------- 1 ---- 7 |
            //   | |    | |        | |    | |
            //   | |    | |        | |    | |
            //   | |    | |        | |    | |
            //   | 20---|-14-------|-15---|-21
            //   |/     |/|        |/|    |/
            //   8 ---- 2 -------- 3 ---- 9
            //          | 22-------|-23
            //          |/         |/
            //          10 ------ 11
            // (-hw, hd+offset, hh), # 0
            // ( hw, hd+offset, hh),  # 1
            // (-hw, hd+offset, -hh), # 2
            // ( hw, hd+offset, -hh), # 3
            // (-hw, hd, hh+offset), # 4
            // ( hw, hd, hh+offset), # 5
            // (-hw-offset, hd, hh), # 6
            // ( hw+offset, hd, hh), # 7
            // (-hw-offset, hd, -hh), # 8
            // ( hw+offset, hd, -hh), # 9
            // (-hw, hd, -hh-offset), # 10
            // ( hw, hd, -hh-offset), # 11

            for (EAR::Vector3 v : VertCoords) {
                addVertexConstrained(*mesh, constrainingGeometry, v, false, backFaceOffset);
            }
            int idxOffset = VertCoords.size();
            for (std::vector<int> v : faceIdxs) {
                mesh->addTriangle(v[0] + idxOffset, v[2] + idxOffset, v[1] + idxOffset);
            }

            mesh->addTriangle(17, 16, 4);
            mesh->addTriangle(17, 4, 5);
            mesh->addTriangle(10, 22, 23);
            mesh->addTriangle(23, 11, 10);
            mesh->addTriangle(19, 7, 9);
            mesh->addTriangle(19, 9, 21);
            mesh->addTriangle(6, 18, 20);
            mesh->addTriangle(6, 20, 8);
            mesh->addTriangle(17, 5, 7);
            mesh->addTriangle(17, 7, 19);
            mesh->addTriangle(9, 11, 23);
            mesh->addTriangle(9, 23, 21);
            mesh->addTriangle(8, 20, 22);
            mesh->addTriangle(8, 22, 10);
            mesh->addTriangle(4, 16, 18);
            mesh->addTriangle(4, 18, 6);
        }
    } else if (segments == 2) {
        //   20       4 -------- 5        22
        //            |          |
        //       21   12 ------ 13   23
        //            |          |
        //   6 --16-- 0 -------- 1 --18-- 7
        //   |        |          |        |
        //   |        |          |        |
        //   |        |          |        |
        //   |        |          |        |
        //   |        |          |        |
        //   8 --17-- 2 -------- 3 --19-- 9
        //            |          |
        //       25   14 ------ 15   27
        //            |          |
        //   24       10 ------ 11        26

        std::vector<EAR::Vector3> VertCoords{
            Vector3(-hw, hh, -hd - offset),
            Vector3(hw, hh, -hd - offset),
            Vector3(-hw, -hh, -hd - offset),
            Vector3(hw, -hh, -hd - offset),
            Vector3(-hw, hh + offset, -hd),
            Vector3(hw, hh + offset, -hd),
            Vector3(-hw - offset, hh, -hd),
            Vector3(hw + offset, hh, -hd),
            Vector3(-hw - offset, -hh, -hd),
            Vector3(hw + offset, -hh, -hd),
            Vector3(-hw, -hh - offset, -hd),
            Vector3(hw, -hh - offset, -hd),
            Vector3(-hw, hh + (offset * irt2), -hd - (offset * irt2)),
            Vector3(hw, hh + (offset * irt2), -hd - (offset * irt2)),
            Vector3(-hw, -hh - (offset * irt2), -hd - (offset * irt2)),
            Vector3(hw, -hh - (offset * irt2), -hd - (offset * irt2)),
            Vector3(-hw - (offset * irt2), hh, -hd - (offset * irt2)),
            Vector3(-hw - (offset * irt2), -hh, -hd - (offset * irt2)),
            Vector3(hw + (offset * irt2), hh, -hd - (offset * irt2)),
            Vector3(hw + (offset * irt2), -hh, -hd - (offset * irt2)),
            Vector3(-hw - (offset * irt2), hh + (offset * irt2), -hd),
            Vector3(-hw - ho, hh + (irt2 * offset), -hd - ho),
            Vector3(hw + (offset * irt2), hh + (offset * irt2), -hd),
            Vector3(hw + ho, hh + (irt2 * offset), -hd - ho),
            Vector3(-hw - (offset * irt2), -hh - (offset * irt2), -hd),
            Vector3(-hw - ho, -hh - (irt2 * offset), -hd - ho),
            Vector3(hw + (offset * irt2), -hh - (offset * irt2), -hd),
            Vector3(hw + ho, -hh - (irt2 * offset), -hd - ho),
        };

        for (EAR::Vector3 v : VertCoords) {
            addVertexConstrained(*mesh, constrainingGeometry, v, true, backFaceOffset);
        }

        std::vector<std::vector<int>> faceIdxs{ { 0, 2, 1 }, { 1, 2, 3 }, { 5, 4, 12 },
            { 5, 12, 13 }, { 13, 12, 0 }, { 13, 0, 1 }, { 3, 2, 14 }, { 3, 14, 15 }, { 15, 14, 10 },
            { 15, 10, 11 }, { 16, 6, 8 }, { 16, 8, 17 }, { 0, 16, 17 }, { 0, 17, 2 }, { 18, 1, 3 },
            { 18, 3, 19 }, { 7, 18, 19 }, { 7, 19, 9 }, { 21, 16, 0 }, { 12, 21, 0 }, { 22, 18, 7 },
            { 18, 22, 23 }, { 4, 21, 12 }, { 4, 20, 21 }, { 23, 13, 1 }, { 23, 1, 18 },
            { 20, 6, 16 }, { 22, 5, 23 }, { 5, 13, 23 }, { 16, 21, 20 }, { 2, 17, 25 },
            { 2, 25, 14 }, { 24, 25, 17 }, { 24, 17, 8 }, { 14, 25, 10 }, { 25, 24, 10 },
            { 3, 15, 27 }, { 3, 27, 19 }, { 27, 15, 11 }, { 27, 11, 26 }, { 19, 27, 26 },
            { 9, 19, 26 } };

        for (std::vector<int> v : faceIdxs) {
            mesh->addTriangle(v[0], v[1], v[2]);
        }

        if (backFaceOffset > 0.0f) {

            for (EAR::Vector3 v : VertCoords) {
                addVertexConstrained(*mesh, constrainingGeometry, v, false, backFaceOffset);
            }
            int idxOffset = VertCoords.size();

            for (std::vector<int> v : faceIdxs) {
                mesh->addTriangle(v[0] + idxOffset, v[2] + idxOffset, v[1] + idxOffset);
            }

            mesh->addTriangle(4, 5, 33);
            mesh->addTriangle(33, 32, 4);
            mesh->addTriangle(4, 32, 20);
            mesh->addTriangle(32, 48, 20);
            mesh->addTriangle(20, 48, 6);
            mesh->addTriangle(48, 34, 6);
            mesh->addTriangle(6, 34, 8);
            mesh->addTriangle(34, 36, 8);
            mesh->addTriangle(8, 36, 24);
            mesh->addTriangle(36, 52, 24);
            mesh->addTriangle(24, 52, 10);
            mesh->addTriangle(52, 38, 10);
            mesh->addTriangle(10, 38, 11);
            mesh->addTriangle(38, 39, 11);
            mesh->addTriangle(11, 39, 26);
            mesh->addTriangle(39, 54, 26);
            mesh->addTriangle(26, 54, 9);
            mesh->addTriangle(54, 37, 9);
            mesh->addTriangle(9, 37, 7);
            mesh->addTriangle(37, 35, 7);
            mesh->addTriangle(7, 35, 22);
            mesh->addTriangle(35, 50, 22);
            mesh->addTriangle(22, 50, 5);
            mesh->addTriangle(50, 33, 5);
        }

    } else if (segments == 3) {

        std::vector<EAR::Vector3> VertCoords{
            Vector3(-hw, hh, -hd - offset),
            Vector3(hw, hh, -hd - offset),
            Vector3(-hw, -hh, -hd - offset),
            Vector3(hw, -hh, -hd - offset),
            Vector3(-hw, hh + offset, -hd),
            Vector3(hw, hh + offset, -hd),
            Vector3(-hw - offset, hh, -hd),
            Vector3(hw + offset, hh, -hd),
            Vector3(-hw - offset, -hh, -hd),
            Vector3(hw + offset, -hh, -hd),
            Vector3(-hw, -hh - offset, -hd),
            Vector3(hw, -hh - offset, -hd),
            Vector3(-hw, hh + cos30o, -hd - ho),
            Vector3(hw, hh + cos30o, -hd - ho),
            Vector3(-hw, hh + ho, -hd - cos30o),
            Vector3(hw, hh + ho, -hd - cos30o),
            Vector3(-hw, -hh - ho, -hd - cos30o),
            Vector3(hw, -hh - ho, -hd - cos30o),
            Vector3(-hw, -hh - cos30o, -hd - ho),
            Vector3(hw, -hh - cos30o, -hd - ho),
            Vector3(-hw - cos30o, hh, -hd - ho),
            Vector3(-hw - ho, hh, -hd - cos30o),
            Vector3(-hw - cos30o, -hh, -hd - ho),
            Vector3(-hw - ho, -hh, -hd - cos30o),
            Vector3(hw + ho, hh, -hd - cos30o),
            Vector3(hw + cos30o, hh, -hd - ho),
            Vector3(hw + ho, -hh, -hd - cos30o),
            Vector3(hw + cos30o, -hh, -hd - ho),
            Vector3(-hw - cos30o, hh + ho, -hd),
            Vector3(-hw - ho, hh + cos30o, -hd),
            Vector3(-hw - (offset * 0.75), hh + ho, -hd - (offset * 0.4330)),
            Vector3(-hw - (offset * 0.4330), hh + cos30o, -hd - (offset * 0.25)),
            Vector3(-hw - (offset * 0.25), hh + cos30o, -hd - (offset * 0.4330)),
            Vector3(-hw - (offset * 0.4330), hh + ho, -hd - (offset * 0.75)),
            Vector3(hw + cos30o, hh + ho, -hd),
            Vector3(hw + ho, hh + cos30o, -hd),
            Vector3(hw + (offset * 0.75), hh + ho, -hd - (offset * 0.4330)),
            Vector3(hw + (offset * 0.4330), hh + cos30o, -hd - (offset * 0.25)),
            Vector3(hw + (offset * 0.25), hh + cos30o, -hd - (offset * 0.4330)),
            Vector3(hw + (offset * 0.4330), hh + ho, -hd - (offset * 0.75)),
            Vector3(-hw - cos30o, -hh - ho, -hd),
            Vector3(-hw - ho, -hh - cos30o, -hd),
            Vector3(-hw - (offset * 0.75), -hh - ho, -hd - (offset * 0.4330)),
            Vector3(-hw - (offset * 0.4330), -hh - cos30o, -hd - (offset * 0.25)),
            Vector3(-hw - (offset * 0.25), -hh - cos30o, -hd - (offset * 0.4330)),
            Vector3(-hw - (offset * 0.4330), -hh - ho, -hd - (offset * 0.75)),
            Vector3(hw + cos30o, -hh - ho, -hd),
            Vector3(hw + ho, -hh - cos30o, -hd),
            Vector3(hw + (offset * 0.75), -hh - ho, -hd - (offset * 0.4330)),
            Vector3(hw + (offset * 0.4330), -hh - cos30o, -hd - (offset * 0.25)),
            Vector3(hw + (offset * 0.25), -hh - cos30o, -hd - (offset * 0.4330)),
            Vector3(hw + (offset * 0.4330), -hh - ho, -hd - (offset * 0.75)),
        };

        for (EAR::Vector3 v : VertCoords) {
            addVertexConstrained(*mesh, constrainingGeometry, v, true, backFaceOffset);
        }

        std::vector<std::vector<int>> faceIdxs{ { 0, 2, 1 }, { 1, 2, 3 }, { 12, 5, 4 },
            { 12, 13, 5 }, { 14, 13, 12 }, { 14, 15, 13 }, { 0, 15, 14 }, { 0, 1, 15 },
            { 3, 2, 16 }, { 3, 16, 17 }, { 17, 16, 18 }, { 17, 18, 19 }, { 19, 18, 10 },
            { 19, 10, 11 }, { 20, 6, 8 }, { 20, 8, 22 }, { 21, 20, 22 }, { 21, 22, 23 },
            { 0, 21, 23 }, { 0, 23, 2 }, { 24, 1, 3 }, { 24, 3, 26 }, { 25, 24, 26 },
            { 25, 26, 27 }, { 7, 25, 27 }, { 7, 27, 9 }, { 4, 32, 12 }, { 4, 31, 32 },
            { 4, 29, 31 }, { 33, 14, 32 }, { 14, 12, 32 }, { 31, 29, 30 }, { 30, 29, 28 },
            { 21, 0, 33 }, { 14, 33, 0 }, { 32, 31, 33 }, { 33, 31, 30 }, { 5, 13, 38 },
            { 5, 38, 37 }, { 5, 37, 35 }, { 38, 13, 15 }, { 38, 15, 39 }, { 37, 38, 39 },
            { 37, 39, 36 }, { 35, 37, 36 }, { 35, 36, 34 }, { 39, 15, 1 }, { 39, 1, 24 },
            { 36, 39, 24 }, { 36, 24, 25 }, { 34, 36, 25 }, { 34, 25, 7 }, { 10, 43, 41 },
            { 10, 44, 43 }, { 10, 18, 44 }, { 41, 42, 40 }, { 43, 42, 41 }, { 43, 45, 42 },
            { 44, 45, 43 }, { 44, 16, 45 }, { 18, 16, 44 }, { 40, 22, 8 }, { 42, 22, 40 },
            { 42, 23, 22 }, { 45, 23, 42 }, { 45, 2, 23 }, { 16, 2, 45 }, { 11, 50, 19 },
            { 11, 49, 50 }, { 11, 47, 49 }, { 17, 51, 3 }, { 3, 51, 26 }, { 49, 47, 48 },
            { 48, 47, 46 }, { 48, 46, 27 }, { 46, 9, 27 }, { 51, 48, 26 }, { 26, 48, 27 },
            { 51, 50, 49 }, { 49, 48, 51 }, { 19, 50, 17 }, { 17, 50, 51 }, { 30, 28, 20 },
            { 20, 28, 6 }, { 21, 30, 20 }, { 33, 30, 21 } };

        for (std::vector<int> v : faceIdxs) {
            mesh->addTriangle(v[0], v[1], v[2]);
        }

        if (backFaceOffset > 0.0f) {
            for (EAR::Vector3 v : VertCoords) {
                addVertexConstrained(*mesh, constrainingGeometry, v, false, backFaceOffset);
            }
            int idxOffset = VertCoords.size();

            for (std::vector<int> v : faceIdxs) {
                mesh->addTriangle(v[0] + idxOffset, v[2] + idxOffset, v[1] + idxOffset);
            }
            mesh->addTriangle(7, 9, 61);
            mesh->addTriangle(61, 59, 7);
            mesh->addTriangle(7, 59, 34);
            mesh->addTriangle(59, 86, 34);
            mesh->addTriangle(34, 86, 35);
            mesh->addTriangle(86, 87, 35);
            mesh->addTriangle(35, 87, 5);
            mesh->addTriangle(87, 57, 5);
            mesh->addTriangle(5, 57, 4);
            mesh->addTriangle(57, 56, 4);
            mesh->addTriangle(4, 56, 29);
            mesh->addTriangle(56, 81, 29);
            mesh->addTriangle(29, 81, 28);
            mesh->addTriangle(81, 80, 28);
            mesh->addTriangle(80, 58, 28);
            mesh->addTriangle(28, 58, 6);
            mesh->addTriangle(58, 60, 6);
            mesh->addTriangle(6, 60, 8);
            mesh->addTriangle(60, 92, 8);
            mesh->addTriangle(8, 92, 40);
            mesh->addTriangle(92, 93, 40);
            mesh->addTriangle(40, 93, 41);
            mesh->addTriangle(93, 62, 41);
            mesh->addTriangle(41, 62, 10);
            mesh->addTriangle(62, 63, 10);
            mesh->addTriangle(10, 63, 11);
            mesh->addTriangle(63, 99, 11);
            mesh->addTriangle(11, 99, 47);
            mesh->addTriangle(99, 98, 47);
            mesh->addTriangle(47, 98, 46);
            mesh->addTriangle(98, 61, 46);
            mesh->addTriangle(9, 46, 61);
        }
    }

    mesh->lock();
}

bool PortalReverbExtent::isInside(const EAR::Vector3& point) const
{
    return mesh->isInside(point);
}

Vector3 PortalReverbExtent::closestPointTo(const EAR::Vector3& point) const
{
    return this->getMesh()->closestPointTo(point);
}

const Mesh* PortalReverbExtent::getMesh() const
{
    return mesh.get();
}

float PortalReverbExtent::getVolume() const
{
    EAR_LOG("getVolume() for portal extent is likely incorrect", EAR::Log::Level::WARN);
    return (volume[0] * volume[1] * volume[2]);
}

float PortalReverbExtent::getArea() const
{
    return mesh->getArea();
}

Vector3 PortalReverbExtent::getSize() const
{
    EAR_LOG("getSize() for portal extent is likely incorrect", EAR::Log::Level::WARN);
    return volume;
}

void PortalReverbExtent::setOrientation(Euler newOrient, bool notifyObs)
{
    if (newOrient != position.orientation) {
        this->mesh->setOrientation(newOrient, notifyObs);
        SceneObject::setOrientation(newOrient,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void PortalReverbExtent::setLocation(Vector3 newLoc, bool notifyObs)
{
    if (newLoc != position.location) {
        this->mesh->setLocation(newLoc, notifyObs);
        SceneObject::setLocation(newLoc,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}
void PortalReverbExtent::setPosition(Position newPos, SceneObject* parent, bool notifyObs)
{
    if ((newPos.location != position.location) || (newPos.orientation != position.orientation)) {
        this->mesh->setPosition(newPos, parent, notifyObs);
        SceneObject::setPosition(newPos, parent,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void PortalReverbExtent::setStatic(bool newState)
{
    this->mesh->setStatic(newState);
    SceneObject::setStatic(newState);
}

float PortalReverbExtent::greatestDistance() const
{
    EAR_LOG("greatestDistance() for portal extents is likely incorrect", EAR::Log::Level::WARN);
    return std::sqrt(std::pow(volume[0], 2) + std::pow(volume[1], 2) + std::pow(volume[2], 2));
}

void PortalReverbExtent::getPOVContourEdges(const EAR::Position& pov,
    EAR::Vector3& resultBottomLeft, EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight,
    EAR::Vector3& resultTopLeft, bool& povIsInside, EAR::Vector3& closestPoint) const
{
    this->mesh->getPOVContourEdges(pov, resultBottomLeft, resultTopRight, resultBottomRight,
        resultTopLeft, povIsInside, closestPoint);
}

void PortalReverbExtent::getApparentSize(const EAR::Position& pov, float& resWidth,
    float& resHeight, bool& povIsInside, EAR::Vector3& closestPoint) const
{
    this->mesh->getApparentSize(pov, resWidth, resHeight, povIsInside, closestPoint);
}

Geometry::BoundingBox PortalReverbExtent::getBoundingBox() const
{
    return mesh->getBoundingBox();
}

Geometry::BoundingBox PortalReverbExtent::getGlobalBoundingBox() const
{
    return mesh->getGlobalBoundingBox();
}

EAR::Vector3 PortalReverbExtent::getCenterPosition() const
{
    return mesh->getCenterPosition();
}

std::ostream& operator<<(std::ostream& os, PortalReverbExtent& b)
{
    os << "Box \"" << b.name << "\":" << std::endl;
    os << "* Size: " << b.volume[0] << "m, " << b.volume[1] << "m, " << b.volume[2] << "m"
       << std::endl;
    os << "* Location :" << b.position.location[0] << "m, " << b.position.location[1] << "m, "
       << b.position.location[2] << "m" << std::endl;
    os << "* Orientation :" << b.position.orientation[0] << "deg, " << b.position.orientation[1]
       << "deg, " << b.position.orientation[2] << "deg" << std::endl
       << std::endl;
    return os;
}

} /* namespace EAR */
