/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/SceneObject.h>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES

#define LOCATION_COARSE_EPSILON 0.01f // in meters
#define ROTATION_COARSE_EPSILON 1.0f // in degrees

#include <algorithm>
#include <cmath>
#include <optional>

namespace EAR {

ObjectID SceneObject::getID() const
{
    return oid;
}

std::optional<int> SceneObject::getBsID() const
{
    return bsId;
};

std::string SceneObject::getName() const
{
    return name;
}

Position SceneObject::getPosition() const
{
    return position;
}

Position SceneObject::getGlobalPosition() const
{
    Position gpos;

    if (parent != nullptr && position.cSpace == CSpace::RCS) {
        gpos.cSpace = parent->getGlobalPosition().cSpace;
    } else {
        gpos.cSpace = position.cSpace;
    }

    gpos.location = getGlobalCSTransform().getTranslation();
    gpos.orientation = getGlobalCSTransform().getRotation();

    return gpos;
}

Vector3 SceneObject::getLocation() const
{
    return position.location;
}

Vector3 SceneObject::getGlobalLocation() const
{
    return getGlobalCSTransform().getTranslation();
}

Euler SceneObject::getOrientation() const
{
    return position.orientation;
}

Euler SceneObject::getGlobalOrientation() const
{
    return getGlobalCSTransform().getRotation();
}

CSpace SceneObject::getCSpace() const
{
    return position.cSpace;
}

const SceneObject* SceneObject::getParent() const
{
    return parent;
}

Vector3 SceneObject::getVelocity() const
{
    return velocity;
}

Vector3 SceneObject::getGlobalVelocity() const
{
    if (position.cSpace != CSpace::WCS && parent) {
        // recursively sum up global velocities
        auto transform = getGlobalCSTransform();
        transform.setTranslation(Vector3(0, 0, 0));
        return parent->getGlobalVelocity() + transform * velocity;
    } else {
        return velocity;
    }
}

double SceneObject::getPositionTimestamp() const
{
    return positionTimestamp.load();
}

void SceneObject::setCSpace(CSpace newCSpace, SceneObject* newParent, bool notifyObs)
{
    if (newCSpace != position.cSpace || parent != newParent) {
        position.cSpace = newCSpace;

        if (parent != newParent && parent) {
            parent->removeChild(this);
        }

        if (newCSpace == CSpace::WCS) {
            parent = nullptr;
        } else {
            EAR_CHECK_REL(newParent != nullptr,
                "The parent of the SceneObject \"" + getName()
                    + "\" with CSpace::RCS or LCS may not be nullptr.");
            EAR_CHECK_REL(newParent->position.cSpace != CSpace::RCS,
                "The parent of the SceneObject \"" + getName()
                    + "\" with CSpace::RCS may not be CSpace::RCS itself.");
            parent = newParent;
            parent->addChild(this);
        }

        if (newCSpace == CSpace::LCS) {
            listenerRelative.store(true);
        } else {
            listenerRelative.store(false);
        }

        moved.store(true);

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        for (auto child : children) {
            child->parentMoved(notifyObs);
        }
    }
}

void SceneObject::setOrientation(Euler newOrient, bool notifyObs)
{
    if (!oldPosition.orientation.isClose(newOrient, ROTATION_COARSE_EPSILON)) {
        oldPosition.orientation = newOrient;
        coarseRotated.store(true);
    } else {
        coarseRotated.store(false);
    }
    if (newOrient != position.orientation) {
        position.orientation = newOrient;
        matrix.setRotation(newOrient);
        moved.store(true);

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        for (auto child : children) {
            child->parentMoved(notifyObs);
        }
    }
}

void SceneObject::setLocation(Vector3 newLoc, bool notifyObs)
{
    if (!oldPosition.location.isClose(newLoc, LOCATION_COARSE_EPSILON)) {
        oldPosition.location = newLoc;
        coarseMoved.store(true);
    } else {
        coarseMoved.store(false);
    }
    if (newLoc != position.location) {
        position.location = newLoc;
        matrix.setTranslation(newLoc);
        moved.store(true);

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        for (auto child : children) {
            child->parentMoved(notifyObs);
        }
    }
}

void SceneObject::setPosition(Position newPos, SceneObject* parent, bool notifyObs)
{
    if (newPos.location != position.location || newPos.orientation != position.orientation
        || newPos.cSpace != position.cSpace) {

        position = newPos;
        matrix.setRotation(newPos.orientation);
        matrix.setTranslation(newPos.location);
        moved.store(true);

        if (newPos.cSpace != position.cSpace) {
            // call setCSpace without observer notification to update parent/child tree structure
            setCSpace(newPos.cSpace, parent, false);
        }

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        for (auto child : children) {
            child->parentMoved(notifyObs);
        }
    }
}

void SceneObject::setVelocity(Vector3 newVel, bool notifyObs)
{
    if (newVel != velocity) {
        velocity = newVel;
        moved.store(true);

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        for (auto child : children) {
            child->parentMoved(notifyObs);
        }
    }
}

void SceneObject::setPositionTimestamp(double ts)
{
    positionTimestamp.store(ts);

    for (auto child : children) {
        child->setPositionTimestamp(ts);
    }
}

Vector3 SceneObject::getRelativeLocation(const Vector3& other) const
{
    return getLocalCSTransform() * other; // TODO: performance optimization?
}

bool SceneObject::isActive() const
{
    return active.load();
}

void SceneObject::setActive(bool newState)
{
    if (newState != active.load()) {
        active.store(newState);
        moved.store(true);
        notify(SceneObjectObserver::Property::Activity);
    }
}

bool SceneObject::isStatic() const
{
    return staticFlag.load();
}

void SceneObject::setStatic(bool newState)
{
    if (newState != staticFlag.load()) {
        staticFlag.store(newState);
        notify(SceneObjectObserver::Property::Staticity);
    }

    if (newState == false) {
        // if the parent becomes non-static, children are too
        // as they may move in absolute coords
        for (auto child : children) {
            child->setStatic(false);
        }
    }
}

bool SceneObject::isListenerRelative() const
{
    return listenerRelative.load();
}

void SceneObject::setListenerRelative(bool newState)
{
    listenerRelative.store(newState);
}

bool SceneObject::hasMoved() const
{
    return moved.load();
}

bool SceneObject::hasCoarseMoved() const
{
    return coarseMoved.load();
}

bool SceneObject::hasCoarseRotated() const
{
    return coarseRotated.load();
}

void SceneObject::setMoved()
{
    moved.store(true);
}

void SceneObject::resetMoved()
{
    moved.store(false);
}

HomogenMatrix SceneObject::getLocalCSTransform() const
{
    return getGlobalCSTransform().inverseGeometric();
}

HomogenMatrix SceneObject::getGlobalCSTransform() const
{
    HomogenMatrix result = matrix;

    if (position.cSpace != CSpace::WCS && parent) {
        // recurse through multi-hierarchical relative coordinates
        result = parent->getGlobalCSTransform() * result;
    }

    return result;
}

std::vector<Vector3> SceneObject::getBasis() const
{
    HomogenMatrix transform = getGlobalCSTransform();

    std::vector<Vector3> basis{ transform * Vector3(1, 0, 0), transform * Vector3(0, 1, 0),
        transform * Vector3(0, 0, 1) };

    return basis;
}

void SceneObject::attach(SceneObjectObserver* observer)
{
    observers.push_back(observer);
}

void SceneObject::detach(SceneObjectObserver* observer)
{
    auto it = std::find(observers.begin(), observers.end(), observer);

    if (it != observers.end()) {
        observers.erase(it);
    }
}

void SceneObject::notify(SceneObjectObserver::Property mod)
{
    for (auto obs : observers) {
        obs->objectChanged(this, mod);
    }
}

void SceneObject::parentMoved(bool notifyObs)
{
    if (position.cSpace != CSpace::WCS) {
        moved.store(true);

        if (notifyObs) {
            notify(SceneObjectObserver::Property::Position);
        }

        for (auto child : children) {
            child->parentMoved(notifyObs);
        }
    }
}

void SceneObject::addChild(SceneObject* child)
{
    children.push_back(child);
}

void SceneObject::removeChild(SceneObject* child)
{
    auto it = std::find(children.begin(), children.end(), child);

    if (it != children.end()) {
        children.erase(it);
    }
}

} // namespace EAR
