/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/SofaReader.h>

#include <EAR/CartesianKDTree.h>
#include <EAR/ComplexBuffer.h>
#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>

#include <highfive/H5DataSpace.hpp>
#include <highfive/H5File.hpp>

#include <complex>
#include <type_traits>

namespace EAR {

template <typename DataT, class BufferT> std::mutex SofaReader<DataT, BufferT>::hdf5Mutex;

template <typename DataT, class BufferT>
SofaReader<DataT, BufferT>::SofaReader(const std::string& path)
{ //, Convention convention) : convention(convention) {

    try {
        std::lock_guard<std::mutex> lock(hdf5Mutex);
        file = std::make_unique<HighFive::File>(
            path); // HighFive::File::ReadOnly is default and omitted
    } catch (...) {
        EAR_FAIL("SOFA file \"" + path + "\" could not be opened.");
    }

    // auto temp = file->getGroup("/").getAttribute("SOFAConventions").getDataType();

    // determine convention
    /*try {
        std::string conventionName;
        file->getGroup("/").getAttribute("SOFAConventions").getDataType();
        file->getGroup("/").getAttribute("SOFAConventions").read(conventionName);

        if (convention == Convention::SimpleFreeFieldHRIR) {
            EAR_CHECK_REL(conventionName.compare("SimpleFreeFieldHRIR") == 0,
                "The given SOFA convention \"" + conventionName + "\" is not compatible to the expected convention \"SimpleFreeFieldHRIR\".");
        } else if (convention == Convention::GeneralTF) {
            EAR_CHECK_REL(conventionName.compare("GeneralTF") == 0,
                "The given SOFA convention \"" + conventionName + "\" is not compatible to the expected convention \"GeneralTF\".");
        }
    } catch (...) {
        EAR_FAIL("The convention used in this SOFA file could not be determined.")
    }*/

    // load interesting meta data
    try {
        N = static_cast<unsigned long>(getDimension("N"));
        M = static_cast<unsigned long>(getDimension("M"));
        R = static_cast<unsigned long>(getDimension("R"));

        centerFreqs.resize(N);
        getMetaData("N", centerFreqs.data());

    } catch (...) {
        EAR_FAIL("Failed while reading SOFA file meta data.");
    }

    EAR_CHECK(N > 0, "Invalid value dimensionality in SOFA dataset.");
    EAR_CHECK(M > 0, "Invalid number of directions in SOFA dataset.");

    // create lookup table for available directions
    try {
        directions = std::unique_ptr<float[]>(new float[M * 3]);

        {
            std::lock_guard<std::mutex> lock(hdf5Mutex);
            file->getDataSet("SourcePosition").read(directions.get());
        }

        sourcePoints.resize(M);
        for (unsigned long i = 0; i < M; i++) {
            Polar direction(directions[i * 3], directions[i * 3 + 1], directions[i * 3 + 2]);
            sourcePoints[i] = polarToCartesian(direction);
        }

        // TODO: loading from disk is very expensive and may be handled more elegant?
        cache = std::make_unique<BufferT>(M * R * N);

        // discriminate between the two template option combinations
        // <float, RealBuffer> and <std::complex<float, ComplexBuffer>
        // by specialization of this private member function
        fillCache();

        lookupTree = std::make_unique<CartesianKDTree>(sourcePoints);
    } catch (...) {
        EAR_FAIL("Could not create lookup for available directions in SOFA file.");
    }
}

template <typename DataT, class BufferT> SofaReader<DataT, BufferT>::~SofaReader()
{
    std::lock_guard<std::mutex> lock(hdf5Mutex);
    file.reset();
}

template <> void SofaReader<float, RealBuffer>::fillCache()
{
    std::lock_guard<std::mutex> lock(hdf5Mutex);
    file->getDataSet("Data.IR").select({ 0, 0, 0 }, { M, R, N }).read(cache->data());
}

template <> void SofaReader<std::complex<float>, ComplexBuffer>::fillCache()
{
    std::lock_guard<std::mutex> lock(hdf5Mutex);
    std::vector<float> tempCache(M * R * N);

    file->getDataSet("Data.Real").select({ 0, 0, 0 }, { M, R, N }).read(tempCache.data());
    cache->writeReal(tempCache.data(), 0, M * R * N);
    file->getDataSet("Data.Imag").select({ 0, 0, 0 }, { M, R, N }).read(tempCache.data());
    cache->writeImag(tempCache.data(), 0, M * R * N);
}

template <typename DataT, class BufferT>
void SofaReader<DataT, BufferT>::getNearestValueForLocation(
    Frame<DataT, BufferT>& data, const Vector3& loc) const
{
    const unsigned long idx = getNearestIndex(loc);

    getValueForIndex(data, idx);
}

template <typename DataT, class BufferT>
std::unique_ptr<const DataT* []> SofaReader<DataT, BufferT>::getValuePointerForLocation(
    const Vector3& loc) const {
    const unsigned long idx = getNearestIndex(loc);

    return getValuePointerForIndex(idx);
}

template <typename DataT, class BufferT>
void SofaReader<DataT, BufferT>::getValueForIndex(
    Frame<DataT, BufferT>& data, unsigned long idx) const
{
    data.resize(R, N);

    // TODO: check if in bounds
    for (auto i = 0; i < static_cast<int>(R); i++) {
        data[i].write(cache->data() + R * N * idx + (i * N), 0, N);
    }
}

template <typename DataT, class BufferT>
std::unique_ptr<const DataT* []> SofaReader<DataT, BufferT>::getValuePointerForIndex(
    unsigned long idx) const {
    auto data = std::make_unique<const DataT*[]>(R);

    // TODO: check if in bounds

    for (auto i = 0; i < static_cast<int>(R); i++) {
        data[i] = cache->data() + R * N * idx + (i * N);
    }

    return data;
}

template <typename DataT, class BufferT>
unsigned long SofaReader<DataT, BufferT>::getNearestIndex(const Vector3& loc) const
{
    return static_cast<unsigned long>(lookupTree->getNearestIndex(loc));
}

template <typename DataT, class BufferT>
float SofaReader<DataT, BufferT>::getDistanceToNearestIndex(const Vector3& loc) const
{
    return lookupTree->getDistanceToNearestPoint(loc);
}

template <typename DataT, class BufferT>
Vector3 SofaReader<DataT, BufferT>::getNearestLocation(const Vector3& loc) const
{
    return lookupTree->getNearestPoint(loc);
}

template <typename DataT, class BufferT>
Vector3 SofaReader<DataT, BufferT>::getLocationForIndex(int idx) const
{
    return sourcePoints[idx];
}

template <typename DataT, class BufferT>
int SofaReader<DataT, BufferT>::getDimension(std::string dim) const
{
    std::lock_guard<std::mutex> lock(hdf5Mutex);
    return static_cast<int>(file->getDataSet(dim).getSpace().getDimensions()[0]);
}

template <typename DataT, class BufferT>
const std::vector<float>& SofaReader<DataT, BufferT>::getCenterFreqs() const
{
    return centerFreqs;
}

template <typename DataT, class BufferT>
template <typename T>
void SofaReader<DataT, BufferT>::getMetaData(std::string name, T* data) const
{
    std::lock_guard<std::mutex> lock(hdf5Mutex);
    file->getDataSet(name).read(data);
}
// allowed specializations:
template void SofaReader<float, RealBuffer>::getMetaData(std::string name, float* data) const;
template void SofaReader<float, RealBuffer>::getMetaData(std::string name, int* data) const;
template void SofaReader<float, RealBuffer>::getMetaData(std::string name, char* data) const;
template void SofaReader<float, RealBuffer>::getMetaData(
    std::string name, unsigned long* data) const;
template void SofaReader<std::complex<float>, ComplexBuffer>::getMetaData(
    std::string name, float* data) const;
template void SofaReader<std::complex<float>, ComplexBuffer>::getMetaData(
    std::string name, int* data) const;
template void SofaReader<std::complex<float>, ComplexBuffer>::getMetaData(
    std::string name, char* data) const;
template void SofaReader<std::complex<float>, ComplexBuffer>::getMetaData(
    std::string name, unsigned long* data) const;

template <typename DataT, class BufferT> BufferT* SofaReader<DataT, BufferT>::getCacheBuffer()
{
    return cache.get();
}

template <typename DataT, class BufferT>
const float* SofaReader<DataT, BufferT>::getDirectionsBuffer() const
{
    return directions.get();
}

/*template<typename DataT, class BufferT>
typename SofaReader<DataT, BufferT>::Convention SofaReader<DataT, BufferT>::getConvention() {
    return convention;
}*/

// allowed specializations:
template class SofaReader<float, RealBuffer>;
template class SofaReader<std::complex<float>, ComplexBuffer>;

} // namespace EAR
