/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Sphere.h>

#include <EAR/Mesh.h>

#include <algorithm>
#include <array>
#include <cmath>
#include <map>
#include <utility>

namespace {

using VertexIdx = unsigned int;
using Face = std::array<VertexIdx, 3>;
using Edge = std::array<VertexIdx, 2>;
using Vertex = EAR::Vector3;

VertexIdx create_unit_midpoint_vertex(VertexIdx v1, VertexIdx v2,
    std::map<Edge, VertexIdx>& midpoint_vertices, std::vector<Vertex>& vertices)
{
    // edge defining vertices must be sorted
    Edge edge = { v1, v2 };
    std::sort(edge.begin(), edge.end());

    auto [it, new_vertex] = midpoint_vertices.try_emplace(edge, vertices.size());
    VertexIdx v = it->second;
    if (!new_vertex) {
        return v;
    }

    const Vertex& p1 = vertices[v1];
    const Vertex& p2 = vertices[v2];
    Vertex p12 = (p1 + p2).normalize();
    vertices.push_back(p12);

    return v;
}

void subdivide_unit_sphere(
    std::vector<Face>& faces, std::vector<Vertex>& vertices, unsigned int subdivisions)
{
    std::vector<Face> old_faces;

    // number of faces: size * 4 ^ subdivisions
    size_t num_faces = faces.size() * 1L << (2 * subdivisions);
    size_t num_faces_quarter = num_faces / 4;
    if (subdivisions % 2 == 0) {
        // even number of sub-divisions: faces is the final vector
        old_faces.reserve(num_faces_quarter);
        faces.reserve(num_faces);
    } else {
        // odd number of sub-divisions: old_faces is the final vector
        old_faces.reserve(num_faces);
        faces.reserve(num_faces_quarter);
    }

    for (unsigned int n = 0; n < subdivisions; n++) {
        std::map<Edge, VertexIdx> midpoint_vertices;
        std::swap(faces, old_faces);
        faces.clear();

        for (const auto& v : old_faces) {
            VertexIdx v_01 = create_unit_midpoint_vertex(v[0], v[1], midpoint_vertices, vertices);
            VertexIdx v_12 = create_unit_midpoint_vertex(v[1], v[2], midpoint_vertices, vertices);
            VertexIdx v_20 = create_unit_midpoint_vertex(v[2], v[0], midpoint_vertices, vertices);
            faces.push_back({ v_01, v_12, v_20 });
            faces.push_back({ v[0], v_01, v_20 });
            faces.push_back({ v[1], v_12, v_01 });
            faces.push_back({ v[2], v_20, v_12 });
        }
    }
}

} // namespace

namespace EAR {

Sphere::Sphere(
    ObjectID oid, float _radius, const std::string& name, Position pos, unsigned int subdivisions)
    : Geometry(oid, name, pos)
    , radius(_radius)
    , mesh(std::unique_ptr<Mesh>(SceneObjectFactory<Mesh>::getNewObject(name, pos)))
{
    // number of faces: 20 * 4 ^ subdivisions
    size_t num_faces = 20 * 1L << (2 * subdivisions);
    size_t num_vertices = 2 + num_faces / 2;

    // create an icosahedron as mesh representation (e.g. for ray tracing purposes)
    const float phi = 1.6180339887; // golden ratio
    const float a = 1.f / std::sqrt(1.f + phi * phi);
    const float b = phi * a;
    std::vector<Vertex> vertices;
    vertices.reserve(num_vertices);
    vertices.push_back({ -a, b, 0 });
    vertices.push_back({ a, b, 0 });
    vertices.push_back({ -a, -b, 0 });
    vertices.push_back({ a, -b, 0 });
    vertices.push_back({ 0, -a, b });
    vertices.push_back({ 0, a, b });
    vertices.push_back({ 0, -a, -b });
    vertices.push_back({ 0, a, -b });
    vertices.push_back({ b, 0, -a });
    vertices.push_back({ b, 0, a });
    vertices.push_back({ -b, 0, -a });
    vertices.push_back({ -b, 0, a });

    std::vector<Face> faces;
    faces.reserve(20);
    faces.push_back({ 0, 11, 5 });
    faces.push_back({ 0, 5, 1 });
    faces.push_back({ 0, 1, 7 });
    faces.push_back({ 0, 7, 10 });
    faces.push_back({ 0, 10, 11 });
    faces.push_back({ 1, 5, 9 });
    faces.push_back({ 5, 11, 4 });
    faces.push_back({ 11, 10, 2 });
    faces.push_back({ 10, 7, 6 });
    faces.push_back({ 7, 1, 8 });
    faces.push_back({ 3, 9, 4 });
    faces.push_back({ 3, 4, 2 });
    faces.push_back({ 3, 2, 6 });
    faces.push_back({ 3, 6, 8 });
    faces.push_back({ 3, 8, 9 });
    faces.push_back({ 4, 9, 5 });
    faces.push_back({ 2, 4, 11 });
    faces.push_back({ 6, 2, 10 });
    faces.push_back({ 8, 6, 7 });
    faces.push_back({ 9, 8, 1 });

    // sub-divide icosahedron
    if (subdivisions > 0) {
        subdivide_unit_sphere(faces, vertices, subdivisions);
    }

    // scale sphere to get a better polygonal approximation
    Vertex& p1 = vertices[faces[0][0]];
    Vertex& p2 = vertices[faces[0][1]];
    Vertex& p3 = vertices[faces[0][2]];
    Vertex triangle_center = (p1 + p2 + p3) * (1.f / 3.f);
    float scale = 2.f / (1.f + triangle_center.abs());
    float scaled_radius = scale * _radius;

    for (const auto& vertex : vertices) {
        mesh->addVertex(vertex * scaled_radius);
    }
    for (const auto& v : faces) {
        mesh->addTriangle(v[0], v[1], v[2], material);
    }

    mesh->lock();
}

bool Sphere::isInside(const EAR::Vector3& point) const
{
    auto distance = point - position.location;

    return radius > distance.abs();
}

Vector3 Sphere::closestPointTo(const EAR::Vector3& point) const
{
    auto diff = point - position.location;

    return position.location + diff * (radius / diff.abs());
}

void Sphere::setMaterial(std::shared_ptr<Material> _material)
{
    material = _material;
    for (auto& tri : mesh->getTriangles()) {
        const_cast<Mesh::Triangle&>(tri).material = _material;
    }
}

float Sphere::getVolume() const
{
    return 4.f / 3.f * M_PI * std::pow(radius, 3);
}

float Sphere::getArea() const
{
    return 4.f * M_PI * std::pow(radius, 2);
}

float Sphere::greatestDistance() const
{
    return 2 * radius;
}

void Sphere::getPOVContourEdges(const EAR::Position& pov, EAR::Vector3& resultBottomLeft,
    EAR::Vector3& resultTopRight, EAR::Vector3& resultBottomRight, EAR::Vector3& resultTopLeft,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    povIsInside = false;
    // create coordinate transform matrix with POV as origin
    // and rotation in the direction of the closest point of the mesh
    closestPoint = closestPointTo(pov.location);
    EAR::Polar closestPInPov = cartesianToPolar(closestPoint - pov.location);
    if (isInside(pov.location)) {
        povIsInside = true;
        resultBottomLeft.set(-1, -1, 0);
        resultBottomRight.set(1, -1, 0);
        resultTopLeft.set(-1, 1, 0);
        resultTopRight.set(1, 1, 0);
        return;
    }

    HomogenMatrix povCamera(Vector3(), pov.orientation);
    povCamera = povCamera.inverseGeometric()
        * HomogenMatrix(Vector3(), Euler(closestPInPov[0], closestPInPov[1], 0));
    float d = closestPInPov[2];

    // second term of the sum is a heuristic correction of the inaccuracy of the first term
    // when POV is close to the sphere
    float projRadius = radius * d / (d + radius) + 0.1 * (radius / d);
    Vector3 resBL(-projRadius, -projRadius, -d);
    Vector3 resBR(projRadius, -projRadius, -d);
    Vector3 resTL(-projRadius, projRadius, -d);
    Vector3 resTR(projRadius, projRadius, -d);

    // pov-relative cartesian coordinates, incorporating the pov rotation
    resultBottomLeft = povCamera * resBL;
    resultBottomRight = povCamera * resBR;
    resultTopLeft = povCamera * resTL;
    resultTopRight = povCamera * resTR;
}

void Sphere::getApparentSize(const EAR::Position& pov, float& resWidth, float& resHeight,
    bool& povIsInside, EAR::Vector3& closestPoint) const
{
    closestPoint = closestPointTo(pov.location);
    povIsInside = isInside(pov.location);

    resWidth = 2 * radius;
    resHeight = 2 * radius;
}

void Sphere::setOrientation(Euler newOrient, bool notifyObs)
{
    if (newOrient != position.orientation) {
        this->mesh->setOrientation(newOrient, notifyObs);
        SceneObject::setOrientation(newOrient,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void Sphere::setLocation(Vector3 newLoc, bool notifyObs)
{
    if (newLoc != position.location) {
        this->mesh->setLocation(newLoc, notifyObs);
        SceneObject::setLocation(newLoc,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}
void Sphere::setPosition(Position newPos, SceneObject* parent, bool notifyObs)
{
    if ((newPos.location != position.location) || (newPos.orientation != position.orientation)) {
        this->mesh->setPosition(newPos, parent, notifyObs);
        SceneObject::setPosition(newPos, parent,
            notifyObs); // mesh first to rebuild vertices before notifiying observers
    }
}

void Sphere::setStatic(bool newState)
{
    this->mesh->setStatic(newState);
    SceneObject::setStatic(newState);
}

std::shared_ptr<Material> Sphere::getMaterial() const
{
    return material;
}

const Mesh* Sphere::getMesh() const
{
    return mesh.get();
}

Geometry::BoundingBox Sphere::getBoundingBox() const
{
    // This may be simply `mesh->getBoundingBox();` but hasn't been tested.

    EAR_FAIL("Sphere::getBoundingBox is not implemented");
}

Geometry::BoundingBox Sphere::getGlobalBoundingBox() const
{
    // This may be simply `mesh->getGlobalBoundingBox();` but hasn't been tested.

    EAR_FAIL("Sphere::getGlobalBoundingBox is not implemented");
}

EAR::Vector3 Sphere::getCenterPosition() const
{
    return mesh->getCenterPosition();
}

std::ostream& operator<<(std::ostream& os, Sphere& s)
{
    os << "Sphere \"" << s.name << "\":" << std::endl;
    os << "* Radius: " << s.radius << "m, " << std::endl;
    os << "* Location :" << s.position.location[0] << "m, " << s.position.location[1] << "m, "
       << s.position.location[2] << "m" << std::endl;
    os << "* Material: \"" << s.material->getName() << "\" " << std::endl << std::endl;

    return os;
}

} /* namespace EAR */
