/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/StereoFIRFilter.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/ComplexFrame.h>
#include <EAR/Exception.h>
#include <EAR/FFT.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>
#include <EAR/config.h>

#include <cmath>
#include <cstddef>

#if defined(EAR_ENABLE_IPP)
#include <ipp.h>
#include <ipps.h>
#endif

namespace EAR {

StereoFIRFilter::StereoFIRFilter(const float* const* h, int _B, int _N, int _K, int _L)
    : StereoFIRFilter(_B, _N, _K, _L)
{

    setFilter(h);
}

StereoFIRFilter::StereoFIRFilter(const float* const* h, int _B, int _N)
    : StereoFIRFilter(_B, _N)
{

    setFilter(h);
}

StereoFIRFilter::StereoFIRFilter(int _B, int _N, int _K, int _L)
    : B(_B)
    , N(_N)
    , K(_K)
    , L(_L)
    , P((N + L - 1) / L)
    , smoothingWeight(0.f)
{ // P = ceil(N/L), Krealres = ceil(K/2 + 1)

    EAR_CHECK(B > 0, "Blocksize must be > 0.");
    EAR_CHECK(N > 0, "Filter length must be > 0.");
    EAR_CHECK(K >= B, "Transform size must be bigger or equal to blocksize.");
    EAR_CHECK(K >= L, "Transform size must be bigger or equal to filter partition size.");
    init();
}

StereoFIRFilter::StereoFIRFilter(int _B, int _N)
    : B(_B)
    , N(_N)
    , K(2 * B)
    , L(B)
    , P((N + L - 1) / L)
    , smoothingWeight(0.f)
{
    // TODO: we could implement smarter strategies to find the best parameters...

    EAR_CHECK(B > 0, "Blocksize must be > 0.");
    EAR_CHECK(N > 0, "Filter length must be > 0.");

    init();
}

StereoFIRFilter::~StereoFIRFilter() {}

void StereoFIRFilter::setFilter(const float* const* _h)
{
    if (running) { // convolution is already running, so we crossfade to new filter in the next "convolve" call
        // note: we write to the same time domain buffer, because we don't need it anymore after the FFT
        std::lock_guard<std::mutex> guard(filterChangeMutex);
        fillFilterBuffers(_h[0], _h[1], filt.get(), sidechain_filtH.get());
        filterChanged.store(true);
    } else {
        fillFilterBuffers(_h[0], _h[1], filt.get(), filtH.get());
        running = true;
    }
}

void StereoFIRFilter::setFilter(const float* h_left, const float* h_right)
{
    if (running) { // convolution is already running, so we crossfade to new filter in the next "convolve" call
        // note: we write to the same time domain buffer, because we don't need it anymore after the FFT
        std::lock_guard<std::mutex> guard(filterChangeMutex);
        fillFilterBuffers(h_left, h_right, filt.get(), sidechain_filtH.get());
        filterChanged.store(true);
    } else {
        fillFilterBuffers(h_left, h_right, filt.get(), filtH.get());
        running = true;
    }
}

void StereoFIRFilter::process(const float* in, float** out)
{
    // append new samples
    inBuffer->shiftLeftInsert(in, B);

    // perform delay step
    // fdl->shiftUp(1); is not used because it is slow...
    fdlPos = (fdlPos + 1) % fdl->channels();

    // write fft of inBuffer to first channel of fdl
    rfft->transform(inBuffer->data(), fdl->ch(fdlPos).data());
    // mirror spectrum along DC axis
    fdl->ch(fdlPos).symmetrizeSpectrum();

    filterFDL(filtH.get(), signalSum.get());

    if (filterChanged.load()
        && filterChangeMutex.try_lock()) { // switch in next block if exchange mutex still locked
        // TODO: this could be simplified by crossfading over only a few samples
        filterFDL(sidechain_filtH.get(), sidechain_signalSum.get());

        // time domain crossfade

        auto sumPtr = signalSum->data();
        auto finPtr = fadeIn->data();
        auto foutPtr = fadeOut->data();
        auto sc_sumPtr = sidechain_signalSum->data();

        if (smoothingWeight != 0) {
            const float w0 = smoothingWeight;
            const float w1 = 1.f - w0;

            // multiplying complex with real here
            for (int s = 0; s < B; s++) {
                sumPtr[K - B + s] = sumPtr[K - B + s] * (w0 + w1 * foutPtr[s])
                    + sc_sumPtr[K - B + s] * (w1 * finPtr[s]);
            }

            filtH->lerp(*sidechain_filtH, w0);
        } else {
            // multiplying complex with real here
            for (int s = 0; s < B; s++) {
                sumPtr[K - B + s] =
                    sumPtr[K - B + s] * foutPtr[s] + sc_sumPtr[K - B + s] * finPtr[s];
            }

            std::swap(filtH, sidechain_filtH);
            filterChanged.store(false);
        }

        filterChangeMutex.unlock();
    }

    // copy last B samples of signalSum to *out
    signalSum->readReal(out[0], K - B, B);
    signalSum->readImag(out[1], K - B, B);
}

void StereoFIRFilter::processFD(const float* in, std::complex<float>* out,
    std::complex<float>* outSCOut, std::complex<float>* outSCIn)
{
    // append new samples
    inBufferFD->shiftLeftInsert(in, B);

    // perform delay step
    // fdl->shiftUp(1); is not used because it is slow...
    fdlPosFD = (fdlPosFD + 1) % fdlFD->channels();

    // write fft of inBuffer to first channel of fdl
    rfft->transform(inBufferFD->data(), fdlFD->ch(fdlPosFD).data());
    // mirror spectrum along DC axis
    fdlFD->ch(fdlPosFD).symmetrizeSpectrum();

    std::complex<float>* const signalSumPtr = signalSum->data();
    filterFDL_FD(filtH->dataRaw(), signalSumPtr);

    if (filterChanged.load()
        && filterChangeMutex.try_lock()) { // switch in next block if exchange mutex still locked
        // TODO: this could be simplified by crossfading over only a few samples
        std::complex<float>* const sidechainSignalSumPtr = sidechain_signalSum->data();
        filterFDL_FD(sidechain_filtH->dataRaw(), sidechainSignalSumPtr);

        std::swap(filtH, sidechain_filtH);
        filterChanged.store(false);
        filterChangeMutex.unlock();
        for (int s = 0; s < K; s++) {
            outSCOut[s] += signalSumPtr[s];
            outSCIn[s] += sidechainSignalSumPtr[s];
        }
    } else {
        for (int s = 0; s < K; s++) {
            out[s] += signalSumPtr[s];
        }
    }
}

void StereoFIRFilter::flush()
{
    inBuffer->zero();
    inBufferFD->zero();
}

void StereoFIRFilter::setSmoothing(float smoothing)
{
    smoothingWeight = std::fmax(std::fmin(smoothing, 1.f), 0.f);
}

float StereoFIRFilter::getSmoothing() const
{
    return smoothingWeight;
}

inline void StereoFIRFilter::init()
{
    fft = FFT::create(K);
    rfft = RFFT::create(K);
    ifft = IFFT::create(K);

    filt = std::make_unique<ComplexFrame>(P, K);
    filtH = std::make_unique<ComplexFrame>(P, K);
    sidechain_filtH = std::make_unique<ComplexFrame>(P, K);

    inBuffer = std::make_unique<RealBuffer>(K);
    inBufferFD = std::make_unique<RealBuffer>(K);

    fdl = std::make_unique<ComplexFrame>(((P - 1) * L / B) + 1, K);
    fdlFD = std::make_unique<ComplexFrame>(((P - 1) * L / B) + 1, K);

    spectralSum = std::make_unique<ComplexBuffer>(K);

    signalSum = std::make_unique<ComplexBuffer>(K);
    sidechain_signalSum = std::make_unique<ComplexBuffer>(K);

    // predefine gain ramps for crossfading
    fadeIn = std::make_unique<RealBuffer>(B);
    fadeOut = std::make_unique<RealBuffer>(B);
    const float step = static_cast<float>(M_PI) / (2.0f * (float)(B - 1));

    for (int i = 0; i < B; i++) {
        (*fadeIn)[i] = sin((float)i * step) * sin((float)i * step);
        (*fadeOut)[i] = cos((float)i * step) * cos((float)i * step);
    }
}

inline void StereoFIRFilter::fillFilterBuffers(
    const float* h1, const float* h2, ComplexFrame* timeDomainBuf, ComplexFrame* freqDomainBuf)
{

    // no need to zero our buffers here, because we always write at the same location
    // in every channel as long as the convolver exists

    for (int p = 0; p < P; p++) {
        int writelen = L;

        // don't exceed h's length and leave zeros at the end if there's space left in Ps part of K
        if (p * L + L > N) {
            writelen = N - p * L;
        }

        timeDomainBuf->ch(p).writeReal(h1 + (p * L), (p * L % B), writelen);
        timeDomainBuf->ch(p).writeImag(h2 + (p * L), (p * L % B), writelen);
    }

    // normalize filters to counteract fft denormalization
    // this is done here to spare this operation on every input block
    timeDomainBuf->mul(1.0f / (K));

    for (std::size_t ch = 0; ch < timeDomainBuf->channels(); ch++) {
        fft->transform(timeDomainBuf->ch(ch).data(), freqDomainBuf->ch(ch).data());
    }
}

inline void StereoFIRFilter::filterFDL(ComplexFrame* filterBuf, ComplexBuffer* result)
{
    // we're operating on pointers here to minimize function call overhead
    auto sumPtr = spectralSum->data();
    const auto fdlPtr = fdl->data();
    const auto filtPtr = filterBuf->data();

    // first complex multiplication is done outside of a loop to avoid zeroing the sum buffer
    int p = 0;
    int ch = (fdlPos - (p * L / B) + static_cast<int>(fdl->channels())) % fdl->channels();

#if defined(EAR_ENABLE_IPP)

    ippsMul_32fc(reinterpret_cast<const Ipp32fc*>(fdlPtr[ch]),
        reinterpret_cast<const Ipp32fc*>(filtPtr[p]), reinterpret_cast<Ipp32fc*>(sumPtr), K);

#else // fallback complex multiply

    for (int b = 0; b < K; b++) {
        // this manual complex multiplication is significantly faster than stdlib
        // spectralSum[b] += fdlPtr[b] * filtPtr[b];

        sumPtr[b].real(fdlPtr[ch][b].real() * filtPtr[p][b].real()
            - fdlPtr[ch][b].imag() * filtPtr[p][b].imag());
        sumPtr[b].imag(fdlPtr[ch][b].real() * filtPtr[p][b].imag()
            + fdlPtr[ch][b].imag() * filtPtr[p][b].real());
    }

#endif

    // select fdl slot for every subfilter and multiply spectra
    for (p = 1; p < P; p++) {
        // get the correct fdl channel to match with the current filter partition
        // this proceeds in reverse order starting from the current read/write pointer and wraps around the fdl boundaries
        ch = (fdlPos - (p * L / B) + static_cast<int>(fdl->channels())) % fdl->channels();

#if defined(EAR_ENABLE_IPP)

        ippsAddProduct_32fc(reinterpret_cast<const Ipp32fc*>(fdlPtr[ch]),
            reinterpret_cast<const Ipp32fc*>(filtPtr[p]), reinterpret_cast<Ipp32fc*>(sumPtr), K);

#else // fallback complex multiply

        for (int b = 0; b < K; b++) {
            // this manual complex multiplication is significantly faster than stdlib
            // spectralSum[b] += fdlPtr[b] * filtPtr[b];

            sumPtr[b].real(sumPtr[b].real() + fdlPtr[ch][b].real() * filtPtr[p][b].real()
                - fdlPtr[ch][b].imag() * filtPtr[p][b].imag());
            sumPtr[b].imag(sumPtr[b].imag() + fdlPtr[ch][b].real() * filtPtr[p][b].imag()
                + fdlPtr[ch][b].imag() * filtPtr[p][b].real());
        }

#endif
    }

    ifft->transform(sumPtr, result->data());
}

inline void StereoFIRFilter::filterFDL_FD(
    std::complex<float>** filterBuf, std::complex<float>* resultFD)
{
    // we're operating on pointers here to minimize function call overhead
    auto sumPtr = resultFD;
    const auto fdlPtr = fdlFD->dataRaw();

    // first complex multiplication is done outside of a loop to avoid zeroing the sum buffer
    int p = 0;
    int ch = (fdlPosFD - (p * L / B) + static_cast<int>(fdlFD->channels())) % fdlFD->channels();

#if defined(EAR_ENABLE_IPP)

    ippsMul_32fc(reinterpret_cast<const Ipp32fc*>(fdlPtr[ch]),
        reinterpret_cast<const Ipp32fc*>(filterBuf[p]), reinterpret_cast<Ipp32fc*>(sumPtr), K);

#else // fallback complex multiply

    for (int b = 0; b < K; b++) {
        // this manual complex multiplication is significantly faster than stdlib
        // spectralSum[b] += fdlPtr[b] * filterBuf[b];

        sumPtr[b].real(fdlPtr[ch][b].real() * filterBuf[p][b].real()
            - fdlPtr[ch][b].imag() * filterBuf[p][b].imag());
        sumPtr[b].imag(fdlPtr[ch][b].real() * filterBuf[p][b].imag()
            + fdlPtr[ch][b].imag() * filterBuf[p][b].real());
    }

#endif

    // select fdl slot for every subfilter and multiply spectra
    for (p = 1; p < P; p++) {
        // get the correct fdl channel to match with the current filter partition
        // this proceeds in reverse order starting from the current read/write pointer and wraps around the fdl boundaries
        ch = (fdlPosFD - (p * L / B) + static_cast<int>(fdlFD->channels())) % fdlFD->channels();

#if defined(EAR_ENABLE_IPP)

        ippsAddProduct_32fc(reinterpret_cast<const Ipp32fc*>(fdlPtr[ch]),
            reinterpret_cast<const Ipp32fc*>(filterBuf[p]), reinterpret_cast<Ipp32fc*>(sumPtr), K);

#else // fallback complex multiply

        for (int b = 0; b < K; b++) {
            // this manual complex multiplication is significantly faster than stdlib
            // spectralSum[b] += fdlPtr[b] * filterBuf[b];

            sumPtr[b].real(sumPtr[b].real() + fdlPtr[ch][b].real() * filterBuf[p][b].real()
                - fdlPtr[ch][b].imag() * filterBuf[p][b].imag());
            sumPtr[b].imag(sumPtr[b].imag() + fdlPtr[ch][b].real() * filterBuf[p][b].imag()
                + fdlPtr[ch][b].imag() * filterBuf[p][b].real());
        }

#endif
    }
}

} // namespace EAR
