/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/VDL.h>

#include <EAR/Exception.h>
#include <EAR/FIRFilter.h>
#include <EAR/IIRFilter.h>

#include "VDL_definitions.h"

#include <algorithm>
#include <cmath>

#define SUBSAMPLE_RESOLUTION 4096
#define SUBSAMPLE_RESOLUTION_FLOAT 4096.f

namespace EAR {

VDL::VDL(float samplerate, int blocklength, float maxDelay, std::size_t nChannels, bool upsampling)
    : B(blocklength)
    , fs(samplerate)
    , Dmax(maxDelay)
    , nChannels(nChannels)
{

    EAR_CHECK(B > 0, "Invalid block length");
    //extract delay line specifications

    useUpsampling = upsampling;

    // Supported sampling rates: 44.1 and 48
    for (std::size_t ch = 0; ch < nChannels; ch++) {
        if (samplerate == 44100.f) {
            if (useUpsampling) {
                aaPreFilter.push_back(
                    IIRFilter::create(LPAA_IIR_44100_18000_2x[0], LPAA_IIR_44100_18000_2x_LENGTH,
                        LPAA_IIR_44100_18000_2x[1], LPAA_IIR_44100_18000_2x_LENGTH));
                gd_PreFilter = LPAA_IIR_44100_18000_2x_GROUPDELAY;
            } else {
                aaPreFilter.push_back(
                    IIRFilter::create(LPAA_IIR_44100_18000_1x[0], LPAA_IIR_44100_18000_1x_LENGTH,
                        LPAA_IIR_44100_18000_1x[1], LPAA_IIR_44100_18000_1x_LENGTH));
                gd_PreFilter = LPAA_IIR_44100_18000_1x_GROUPDELAY;
            }
        } else if (samplerate == 48000.f) {
            if (useUpsampling) {
                aaPreFilter.push_back(
                    IIRFilter::create(LPAA_IIR_48000_18000_2x[0], LPAA_IIR_48000_18000_2x_LENGTH,
                        LPAA_IIR_48000_18000_2x[1], LPAA_IIR_48000_18000_2x_LENGTH));
                gd_PreFilter = LPAA_IIR_48000_18000_2x_GROUPDELAY;
            } else {
                aaPreFilter.push_back(
                    IIRFilter::create(LPAA_IIR_48000_18000_1x[0], LPAA_IIR_48000_18000_1x_LENGTH,
                        LPAA_IIR_48000_18000_1x[1], LPAA_IIR_48000_18000_1x_LENGTH));
                gd_PreFilter = LPAA_IIR_48000_18000_1x_GROUPDELAY;
            }
        } else {
            EAR_FAIL("Samplerate not supported by VDL");
        }
    }

    int Bl = B;
    if (useUpsampling) {
        Bl = 2 * B;
    }

    // predefine gain ramps for crossfading
    fadeIn = std::make_unique<EAR::RealBuffer>(Bl);
    fadeOut = std::make_unique<EAR::RealBuffer>(Bl);
    const float step = M_PI / (2.0f * Bl);
    for (int i = 0; i < Bl; i++) {
        (*fadeIn)[i] = std::sin(i * step) * std::sin(i * step);
        (*fadeOut)[i] = std::cos(i * step) * std::cos(i * step);
    }

    out = EAR::RealBuffer(Bl);

    // Align the maximum delay to be an integer multiple of the block length,
    // thus no wrap-around writes occur (*2 due to 2x upsampling)
    datalen = (std::ceil(maxDelay / blocklength) + 1) * blocklength;

    if (useUpsampling) {
        datalen *= 2;
    }

    data = EAR::RealFrame(nChannels, datalen);
    wc = 0;
}

VDL::~VDL() {}

float VDL::getMaximumDelay()
{
    return Dmax;
}

void VDL::setDelay(VDL::ReadCursor& accessor, float delay, bool jump)
{
    EAR_CHECK_REL(delay >= accessor.Dmin && delay <= Dmax, "Delay out of bounds.");

    accessor.Dnew = delay;

    if (useUpsampling) {
        accessor.Dnew *= 2;
    }

    accessor.singleJump = jump;
}

void VDL::clear()
{
    for (unsigned int ch = 0; ch < nChannels; ch++) {
        aaPreFilter[ch]->clear();
    }

    data.zero();
    wc = 0;
}

inline void VDL::updateBackwardDifferences(
    float* backwardDiffs, const float& k0, const float& k1, const float& k2, const float& k3)
{
    backwardDiffs[0] = k0;
    backwardDiffs[1] = k0 - k1;
    backwardDiffs[2] = k0 - 2 * k1 + k2;
    backwardDiffs[3] = k0 - 3 * k1 + 3 * k2 - k3;
    return;
}

void VDL::write(const float* const* input)
{

    int Bl = B;
    if (useUpsampling) {
        Bl = 2 * B;
    }

    auto dataptr2d = data.data();

    for (std::size_t ch = 0; ch < nChannels; ch++) {

        float* dataptr = dataptr2d[ch];

        // Upsampling of the new input samples
        if (useUpsampling) {
            for (auto i = 0; i < B; i++) {
                dataptr[wc + 2 * i] = input[ch][i];
                dataptr[wc + 2 * i + 1] = 0;
            }

            aaPreFilter[ch]->process(dataptr + wc, dataptr + wc, Bl);

        } else {
            aaPreFilter[ch]->process(input[ch], dataptr + wc,
                Bl); // filter still needed to remove aliasing by doppler shifts to freqs higher than Nyquist
        }
    }

    // update write cursor
    wc += Bl;

    if (wc >= static_cast<std::size_t>(datalen)) {
        wc -= datalen;
    }
}

void VDL::read(VDL::ReadCursor& accessor, float** output)
{

    int Bl = B;
    if (useUpsampling) {
        Bl = 2 * B;
    }

    auto dataptr2d = data.data();

    if (accessor.crossfade || accessor.singleJump) { // DelayFade mode

        int dprev = static_cast<int>(std::round(accessor.Dcur));
        int dnext = static_cast<int>(std::round(accessor.Dnew));

        for (std::size_t ch = 0; ch < nChannels; ch++) {
            float* outptr = output[ch];
            if (useUpsampling) {
                outptr = out.data();
            }
            float* dataptr = dataptr2d[ch];

            if (dprev != dnext) {
                auto finPtr = fadeIn->data();
                auto foutPtr = fadeOut->data();
                for (auto i = 0; i < Bl; i++) {

                    int cprev = static_cast<int>(wc) + i - dprev - Bl;

                    if (cprev >= datalen || cprev < 0) {
                        cprev = (cprev + static_cast<int>(datalen)) % datalen;
                    }

                    int cnext = static_cast<int>(wc) + i - dnext - Bl;

                    if (cnext >= datalen || cnext < 0) {
                        cnext = (cnext + static_cast<int>(datalen)) % datalen;
                    }
                    outptr[i] = foutPtr[i] * dataptr[cprev] + finPtr[i] * dataptr[cnext];
                }
            } else {
                for (auto i = 0; i < Bl; i++) {
                    int c = static_cast<int>(wc) + i - dprev - Bl;

                    if (c >= datalen || c < 0) {
                        c = (c + static_cast<int>(datalen)) % datalen;
                    }

                    outptr[i] = dataptr[c];
                }
            }

            // AA filtering and decimation of the output
            if (useUpsampling) {
                accessor.aaPostFilter[ch]->process(outptr, outptr, Bl);
                // downsampling
                for (int i = 0; i < B; i++) {
                    // TODO: find a way to tweak the AA filters to get rid of this *2 multiplication
                    output[ch][i] = outptr[2 * i] * 2;
                }
            }

        } // endfor channels

        accessor.singleJump = false;
        accessor.Dcur = accessor.Dnew;

    } else { // DelayAdapt mode
        // delay change per sample in this block
        float dd = ((accessor.Dnew - accessor.Dcur) * SUBSAMPLE_RESOLUTION_FLOAT) / Bl;
        // total delay at the beginning of this block
        int32_t diff = (accessor.Dcur - accessor.Dmin) * SUBSAMPLE_RESOLUTION;

        // fixed-point types used for delay calculations to avoid expensive float rounding in per-sample loop
        int32_t d; // total delay per sample
        int32_t fd; // fractional (sub-sample) part of total delay per sample
        int32_t id; // integer part of total delay per sample

        for (std::size_t ch = 0; ch < nChannels; ch++) {
            float* outptr = output[ch];
            if (useUpsampling) {
                outptr = out.data();
            }
            float* dataptr = dataptr2d[ch];
            for (auto i = 0; i < Bl; i++) {
                d = diff + i * dd;
                id = d / SUBSAMPLE_RESOLUTION; // implicit floor
                fd = d & (SUBSAMPLE_RESOLUTION - 1); // interval [0,SUBSAMPLE_RESOLUTION[

                outptr[i] = 0;

                int cursor = static_cast<int>(wc) + i - id - Bl;

                if ((fd <= 1.f)
                    && !(accessor.farrowInterpolation_Spline
                        || accessor.newtonInterpolation_Spline)) {
                    // fallback to integer ring buffer read

                    cursor = cursor - accessor.gd;

                    if (cursor >= datalen) {
                        cursor = cursor - datalen;
                    } else if (cursor < 0) {
                        cursor = cursor + datalen;
                    }

                    outptr[i] = dataptr[cursor];
                    continue;
                } else {
                    // calculate positions in ring buffer (this is done the verbose way to ensure correct wrap around)
                    for (auto col = 0; col < accessor.nC; col++) {
                        (accessor.cursorPositions.get())[col] = cursor - col;
                        // the conditional construct is much faster, because the wrap around occurs only rarely...
                        // https://www.youtube.com/watch?v=nXaxk27zwlk&feature=youtu.be&t=56m34s
                        if ((accessor.cursorPositions.get())[col] >= datalen) {
                            (accessor.cursorPositions.get())[col] =
                                (accessor.cursorPositions.get())[col] - static_cast<int>(datalen);
                        } else if ((accessor.cursorPositions.get())[col] < 0) {
                            (accessor.cursorPositions.get())[col] =
                                (accessor.cursorPositions.get())[col] + static_cast<int>(datalen);
                        }

                        if ((fd <= 1.f) && accessor.farrowInterpolation_Spline) {
                            //in the case of spline interpolation in farrow structure
                            outptr[i] +=
                                dataptr[(accessor.cursorPositions.get())[col]] * accessor.C[col];

                            if (col == accessor.nC - 1) {
                                continue;
                            }
                        }
                    }
                }

                // interpolate the intersample value
                if (accessor.farrowInterpolation_Lagrange || accessor.farrowInterpolation_Spline) {
                    for (auto row = accessor.nC - 1; row > 0; row--) {
                        for (auto col = 0; col < accessor.nC; col++) {
                            outptr[i] += accessor.C[accessor.nC * row + col]
                                * dataptr[(accessor.cursorPositions.get())[col]];
                        }

                        outptr[i] *= fd / SUBSAMPLE_RESOLUTION_FLOAT;
                    }

                    for (auto col = 0; col < accessor.nC; col++) {
                        outptr[i] +=
                            accessor.C[col] * dataptr[(accessor.cursorPositions.get())[col]];
                    }

                } else if (accessor.newtonInterpolation_Lagrange
                    || accessor.newtonInterpolation_Spline) {
                    float temp = 0;

                    // Note: When delay points two consecutive times to the same interval, then, the back differences remain the same
                    if (dataptr[(accessor.cursorPositions.get())[1]] == accessor.backwardDiffs[0]) {
                        temp = accessor.backwardDiffs[3] =
                            dataptr[(accessor.cursorPositions.get())[0]];

                        for (int col = 0; col < accessor.nC; col++) {
                            accessor.backwardDiffs[3] -= accessor.backwardDiffs[col];
                            accessor.backwardDiffs[col] = temp;
                            temp = accessor.backwardDiffs[3];
                        }
                    } else {
                        // the previous backdifferences have to be updated, because they do not correspond
                        // to the current sample the integer delay is pointing to
                        updateBackwardDifferences(accessor.backwardDiffs,
                            dataptr[(accessor.cursorPositions.get())[0]],
                            dataptr[(accessor.cursorPositions.get())[1]],
                            dataptr[(accessor.cursorPositions.get())[2]],
                            dataptr[(accessor.cursorPositions.get())[3]]);
                    }

                    if (accessor.newtonInterpolation_Lagrange) {
                        // newton formula in horner scheme (this is done the verbose way to save operations)
                        outptr[i] = dataptr[(accessor.cursorPositions.get())[0]]
                            + (-fd / SUBSAMPLE_RESOLUTION_FLOAT - 1.f)
                                * (accessor.backwardDiffs[1]
                                    + (-fd / SUBSAMPLE_RESOLUTION_FLOAT)
                                        * (accessor.C[0] * accessor.backwardDiffs[2]
                                            + (-fd / SUBSAMPLE_RESOLUTION_FLOAT + 1.f)
                                                * accessor.C[1] * accessor.backwardDiffs[3]));

                    } else { // newtonInterpolation_Spline implied
                        // newton formula in horner scheme (this is done the verbose way to save operations)
                        outptr[i] = dataptr[(accessor.cursorPositions.get())[0]]
                            + accessor.C[1]
                                * (accessor.backwardDiffs[2] + accessor.backwardDiffs[3])
                            + (-fd / SUBSAMPLE_RESOLUTION_FLOAT - 1.f)
                                * (accessor.backwardDiffs[1]
                                    + accessor.C[1] * accessor.backwardDiffs[3]
                                    + (-fd / SUBSAMPLE_RESOLUTION_FLOAT)
                                        * (accessor.C[0] * accessor.backwardDiffs[2]
                                            + (-fd / SUBSAMPLE_RESOLUTION_FLOAT + 1.f)
                                                * accessor.C[1] * accessor.backwardDiffs[3]));
                    }
                } else if (accessor.linearInterpolation) {
                    float dy = (dataptr[(accessor.cursorPositions.get())[1]]
                                   - dataptr[(accessor.cursorPositions.get())[0]])
                        * fd / SUBSAMPLE_RESOLUTION_FLOAT;
                    outptr[i] = dataptr[(accessor.cursorPositions.get())[0]] + dy;
                } else { // Lagrange interpolation
                    int fdidx = fd * (accessor.mC - 1) / SUBSAMPLE_RESOLUTION; // interval [0,mC-1]

                    for (int col = 0; col < accessor.nC / 2; col++) {
                        outptr[i] += accessor.C[fdidx + col * accessor.mC]
                            * dataptr[(accessor.cursorPositions.get())[col]]; //from begin
                        outptr[i] += accessor.C[accessor.mC * (col + 1) - fdidx - 1]
                            * dataptr[(accessor.cursorPositions
                                           .get())[accessor.nC - (col + 1)]]; //from end
                    }
                }
            }

            if (accessor.useCorrectionFilter) {
                //prefiltering by B(z) = 6z/(1+4z+z**2)
                accessor.cFilter[ch]->process(outptr, outptr);
            }

            // AA filtering and decimation of the output
            if (useUpsampling) {
                accessor.aaPostFilter[ch]->process(outptr, outptr, Bl);
                // downsampling
                for (int i = 0; i < B; i++) {
                    // TODO: find a way to tweak the AA filters to get rid of this *2 multiplication
                    output[ch][i] = outptr[2 * i] * 2;
                }
            }
        } // endfor channels

        accessor.Dcur = accessor.Dnew;
    } // endif crossfade vs. adapt
}

VDL::ReadCursor VDL::registerReadCursor(float initialDelay, int flags)
{
    return VDL::ReadCursor(std::fmin(initialDelay, Dmax), this, flags);
}

VDL::ReadCursor::ReadCursor(float initialDelay, VDL* vdl, int flags)
    : backwardDiffs{ 0.f, 0.f, 0.f, 0.f }
    , singleJump(true)
{

    Dmin = vdl->gd_PreFilter;
    for (std::size_t ch = 0; ch < vdl->nChannels; ch++) {
        if (vdl->useUpsampling) {
            //initialize aaPostFilter
            //Dmin affected by postfilter and prefilter
            if (vdl->fs == 44100.f) {
                aaPostFilter.push_back(std::shared_ptr<IIRFilter>(
                    IIRFilter::create(LPAA_IIR_44100_18000_2x[0], LPAA_IIR_44100_18000_2x_LENGTH,
                        LPAA_IIR_44100_18000_2x[1], LPAA_IIR_44100_18000_2x_LENGTH)));
                Dmin += LPAA_IIR_44100_18000_2x_GROUPDELAY;
            } else if (vdl->fs == 48000.f) {
                aaPostFilter.push_back(std::shared_ptr<IIRFilter>(
                    IIRFilter::create(LPAA_IIR_48000_18000_2x[0], LPAA_IIR_48000_18000_2x_LENGTH,
                        LPAA_IIR_48000_18000_2x[1], LPAA_IIR_48000_18000_2x_LENGTH)));
                Dmin += LPAA_IIR_48000_18000_2x_GROUPDELAY;
            } else {
                EAR_FAIL("Samplerate not supported by VDL");
            }
        }
    }

    farrowInterpolation_Lagrange = false;
    farrowInterpolation_Spline = false;
    linearInterpolation = false;
    newtonInterpolation_Lagrange = false;
    newtonInterpolation_Spline = false;
    crossfade = false;
    useCorrectionFilter = true;

    if (flags & VDL::InterpolationLagrange) {
        // Lagrange interpolation (only supports even number of control points)
        if (flags & VDL::QualityHigh) {
            gd = FD_LAGRANGE_3_GD;
            C = &FD_LAGRANGE_16_3_C[0][0];
            nC = FD_LAGRANGE_3_N;
            mC = FD_LAGRANGE_3_M;
            ;
        } else {
            gd = FD_LAGRANGE_1_GD;
            C = &FD_LAGRANGE_16_1_C[0];
            nC = FD_LAGRANGE_1_N;
            mC = FD_LAGRANGE_1_M;
        }
    } else if (flags & VDL::InterpolationLinear) {
        linearInterpolation = true;
        nC = 2;
        gd = 0;
    } else if (flags & VDL::InterpolationFarrow_Spline) {
        //Note that there are 2 types of spline interpolation approaches:
        //1. The spline approximiation which does not include prefiltering by B(z) = 6z/(1+4z+z**2)
        //2. The spline interpolation which does include the prefilter
        farrowInterpolation_Spline = true;
        if (flags & VDL::QualityHigh) {
            gd = FD_FARROW_3_GD;
            C = &FD_FARROW_SPLINE_3_C[0][0];
            nC = FD_FARROW_3_N;
        } else {
            EAR_FAIL("Spline interpolation does only support high quality mode!");
        }
    } else if (flags & VDL::InterpolationNewton_Lagrange) {
        //Newton interpolation based on lagrangian polynomials
        newtonInterpolation_Lagrange = true;

        if (flags & VDL::QualityHigh) {
            gd = FD_NEWTON_3_GD;
            C = &FD_NEWTON_LAGRANGE_3_C[0];
            nC = FD_NEWTON_3_N;
        } else {
            EAR_FAIL("Lagragian Newton interpolation does only support high quality mode!");
        }
    } else if (flags & VDL::InterpolationNewton_Spline) {
        //Note that there are 2 types of spline interpolation approaches:
        //1. The spline approximiation which does not include prefiltering by B(z) = 6z/(1+4z+z**2)
        //2. The spline interpolation which does include the prefilter
        newtonInterpolation_Spline = true;
        if (flags & VDL::QualityHigh) {
            gd = FD_NEWTON_3_GD;
            C = &FD_NEWTON_SPLINE_3_C[0];
            nC = FD_NEWTON_3_N;
        } else {
            EAR_FAIL("Spline Newton interpolation does only support high quality mode!");
        }
    } else if (flags & VDL::DelayFade) {
        crossfade = true;
        gd = 0;
        nC = 0;
        C = nullptr;

    } else {
        farrowInterpolation_Lagrange = true;

        if (flags & VDL::QualityHigh) {
            gd = FD_FARROW_3_GD;
            C = &FD_FARROW_LAGRANGE_3_C[0][0];
            nC = FD_FARROW_3_N;
        } else {
            gd = FD_FARROW_1_GD;
            C = &FD_FARROW_LAGRANGE_1_C[0][0];
            nC = FD_FARROW_1_N;
        }
    }

    useCorrectionFilter = (flags & VDL::CorrectionFilterEnabled);

    if (useCorrectionFilter && !(newtonInterpolation_Spline || farrowInterpolation_Spline)) {
        EAR_FAIL("Only spline interpolation does support correction filter application!")
    }

    if (useCorrectionFilter) {
        //group delay and filter creation

        Dmin +=
            CF_FIR_GROUPDELAY; // not necessary as integer delay is not supported by spline interpolation
        for (unsigned int ch = 0; ch < vdl->nChannels; ch++) {
            if (vdl->useUpsampling) {
                cFilter.push_back(
                    std::make_shared<EAR::FIRFilter>(CF_FIR, vdl->B * 2, CF_FIR_LENGTH));
            } else {
                cFilter.push_back(std::make_shared<EAR::FIRFilter>(CF_FIR, vdl->B, CF_FIR_LENGTH));
            }
        }
    }

    Dmin += gd; // additional group delay imposed by the IIR filters for interpolation
    cursorPositions = std::shared_ptr<int>(new int[nC], std::default_delete<int[]>());

    Dcur = Dnew = Dmin;
    if (initialDelay > Dcur) {
        Dcur = initialDelay;
    }

    if (Dcur >= vdl->Dmax) {
        // TODO: should this be handled as an exception?
        Dcur = vdl->Dmax;
    }
}

float VDL::ReadCursor::getMinimumDelay()
{
    return Dmin;
}

} // namespace EAR
