/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Variant.h>

#include <algorithm>
#include <cctype>
#include <sstream>

namespace detail {

inline void eraseSubStr(std::string& mainStr, const std::string& toErase)
{
    std::size_t pos = mainStr.find(toErase);
    // delete all occurences
    while (pos != std::string::npos) {
        mainStr.erase(pos, toErase.length());

        pos = mainStr.find(toErase);
    }
};

inline bool string2bool(const std::string& v)
{
    return !v.empty() && (v == "true" || std::atoi(v.c_str()) != 0);
}

inline void ltrim(std::string& s)
{
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](int ch) { return !std::isspace(ch); }));
}

inline void rtrim(std::string& s)
{
    s.erase(std::find_if(s.rbegin(), s.rend(), [](int ch) { return !std::isspace(ch); }).base(),
        s.end());
}

inline void trim(std::string& s)
{
    ltrim(s);
    rtrim(s);
}

} // namespace detail

EAR::Variant::Variant(bool v)
    : boolValue(v)
    , type(EAR::Variant::Type::Bool)
{}

EAR::Variant::Variant(int v)
    : intValue(v)
    , type(EAR::Variant::Type::Int)
{}

EAR::Variant::Variant(CSpace v)
    : intValue(static_cast<int>(v))
    , type(EAR::Variant::Type::Int)
{}

EAR::Variant::Variant(float v)
    : floatValue(v)
    , type(EAR::Variant::Type::Float)
{}

EAR::Variant::Variant(const EAR::Vector3& v)
    : triplet{ v[0], v[1], v[2] }
    , type(EAR::Variant::Type::Vector3)
{}

EAR::Variant::Variant(const EAR::Euler& v)
    : triplet{ v[0], v[1], v[2] }
    , type(EAR::Variant::Type::Euler)
{}

EAR::Variant::Variant(const EAR::Quaternion& v)
    : quadruplet{ v[0], v[1], v[2], v[3] }
    , type(EAR::Variant::Type::Quaternion)
{}

EAR::Variant::Variant(const std::string& v)
    : type(EAR::Variant::Type::String)
{
    strValue = v;
    // new (stringValue) std::string(v);
}

EAR::Variant::Variant(const float* v)
    : type(EAR::Variant::Type::FloatPtr)
{
    ptrValue = v;
    // new (stringValue) std::string(v);
}

EAR::Variant::Variant(const std::vector<float>& v)
    : type(EAR::Variant::Type::FloatVector)
{
    floatVectorValue = v;
}

const EAR::Variant::Type& EAR::Variant::getVariantType() const
{
    return type;
}

std::string EAR::Variant::serialize() const
{
    std::stringstream out;
    out << "Variant(";
    switch (type) {
    case Type::None:
        out << "null, null";
        break;
    case Type::Bool:
        out << "bool, " << (boolValue ? "true" : "false");
        break;
    case Type::Int:
        out << "int, " << intValue;
        break;
    case Type::Float:
        out << "float, " << floatValue;
        break;
    case Type::Euler:
        out << "EAR::Euler, (" << triplet[0] << ", " << triplet[1] << ", " << triplet[2] << ")";
        break;
    case Type::Quaternion:
        out << "EAR::Quaternion, (" << quadruplet[0] << ", " << quadruplet[1] << ", "
            << quadruplet[2] << ", " << quadruplet[3] << ")";
        break;
    case Type::Vector3:
        out << "EAR::Vector3, (" << triplet[0] << ", " << triplet[1] << ", " << triplet[2] << ")";
        break;
    case Type::String:
        out << "string, " << strValue;
        break;
    case Type::FloatPtr:
        EAR_FAIL("Variant: A pointer can and should not be serialized.");
        break;
    case Type::FloatVector:
        out << "vector<float>, (";
        auto const c = floatVectorValue.size();

        if (c > 0) {
            out << floatVectorValue[0];

            for (std::size_t i{ 1 }; i < c; ++i) {
                out << "," << floatVectorValue[i];
            }
        }
        out << ")";
        break;
    }
    out << ")";
    return out.str();
}

void EAR::Variant::deserialize(std::string str)
{
    EAR_CHECK_REL(str.find("Variant(") != std::string::npos,
        "Unexpected string format in Variant::deserialize().");
    detail::eraseSubStr(str, "Variant(");
    std::size_t pos = str.find(","); // first occurence
    std::string typeStr = str.substr(0, pos);
    str.erase(0, pos + 1); // remove type label and ","
    detail::trim(str);
    str.erase(str.length() - 1); // remove last ")"

    if (typeStr == "None") {
        type = Type::None;
    } else if (typeStr == "bool") {
        type = Type::Bool;
        intValue = detail::string2bool(str);
    } else if (typeStr == "int") {
        type = Type::Int;
        intValue = std::stoi(str);
    } else if (typeStr == "float") {
        type = Type::Float;
        floatValue = std::stof(str);
    } else if (typeStr == "EAR::Euler") {
        type = Type::Euler;
        str.erase(0, 1); // remove "("
        str.erase(str.length() - 1); // remove ")"
        for (int i = 0; i < 3; i++) {
            pos = str.find(", ");
            if (pos != std::string::npos) {
                triplet[i] = std::stof(str.substr(0, pos));
                str.erase(0, pos + 2);
            } else {
                triplet[i] = std::stof(str);
            }
        }
    } else if (typeStr == "EAR::Quaternion") {
        type = Type::Quaternion;
        str.erase(0, 1); // remove "("
        str.erase(str.length() - 1); // remove ")"
        for (int i = 0; i < 4; i++) {
            pos = str.find(", ");
            if (pos != std::string::npos) {
                quadruplet[i] = std::stof(str.substr(0, pos));
                str.erase(0, pos + 2);
            } else {
                quadruplet[i] = std::stof(str);
            }
        }
    } else if (typeStr == "EAR::Vector3") {
        type = Type::Vector3;
        str.erase(0, 1); // remove "("
        str.erase(str.length() - 1); // remove ")"
        for (int i = 0; i < 3; i++) {
            pos = str.find(", ");
            if (pos != std::string::npos) {
                triplet[i] = std::stof(str.substr(0, pos));
                str.erase(0, pos + 2);
            } else {
                triplet[i] = std::stof(str);
            }
        }
    } else if (typeStr == "string") {
        type = Type::String;
        strValue = str;
    } else if (typeStr == "vector<float>") {
        type = Type::FloatVector;
        str.erase(0, 1); // remove "("
        str.erase(str.length() - 1); // remove ")"
        if (str.find_first_not_of(" ") == std::string::npos) {
            return;
        }
        pos = str.find(",");
        while (pos != std::string::npos) {
            floatVectorValue.push_back(std::stof(str.substr(0, pos)));
            str.erase(0, pos + 1);
            pos = str.find(",");
        }
        floatVectorValue.push_back(std::stof(str));
    }
}
