/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef EAR_FFT_IPP
#define EAR_FFT_IPP

#include <EAR/FFT.h>
#include <EAR/Uncopyable.h>

#include <cmath>

#include <ipp.h>
#include <ipps.h>

namespace EAR {

class RFFT_IPP : public RFFT, Uncopyable
{
public:
    RFFT_IPP(int length)
        : RFFT(length)
    {
        int specSize, initBufSize, workBufSize;
        // arguments: length, normalization method (we do this ourselves), something deprecated, size spec, size init, size work
        ippsDFTGetSize_R_32f(
            N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, &specSize, &initBufSize, &workBufSize);

        // allocate IPP buffers
        specBuf = reinterpret_cast<IppsDFTSpec_R_32f*>(ippsMalloc_8u(specSize));
        initBuf = ippsMalloc_8u(initBufSize);
        workBuf = ippsMalloc_8u(workBufSize);

        // arguments: length, normalization method (we do this ourselves), something deprecated, spec buffer, init buffer
        ippsDFTInit_R_32f(N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, specBuf, initBuf);

        if (initBuf) {
            ippFree(initBuf);
        }
    }

    ~RFFT_IPP()
    {
        if (workBuf) {
            ippFree(workBuf);
        }

        if (specBuf) {
            ippFree(specBuf);
        }
    }

    void transform(const float* input, std::complex<float>* output) override
    {
        if (ippsDFTFwd_RToCCS_32f(input, reinterpret_cast<float*>(output), specBuf, workBuf)
            != ippStsNoErr) {
            EAR_FAIL("Intel IPP FFT failed.");
        }
    }

private:
    IppsDFTSpec_R_32f* specBuf = nullptr;
    Ipp8u* initBuf = nullptr;
    Ipp8u* workBuf = nullptr;
};

class IRFFT_IPP : public IRFFT
{
public:
    IRFFT_IPP(int length)
        : IRFFT(length)
    {
        int specSize, initBufSize, workBufSize;
        // arguments: length, normalization method (we do this ourselves), something deprecated, size spec, size init, size work
        ippsDFTGetSize_R_32f(
            N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, &specSize, &initBufSize, &workBufSize);

        // allocate IPP buffers
        specBuf = reinterpret_cast<IppsDFTSpec_R_32f*>(ippsMalloc_8u(specSize));
        initBuf = ippsMalloc_8u(initBufSize);
        workBuf = ippsMalloc_8u(workBufSize);

        // arguments: length, normalization method (we do this ourselves), something deprecated, spec buffer, init buffer
        ippsDFTInit_R_32f(N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, specBuf, initBuf);

        if (initBuf) {
            ippFree(initBuf);
        }
    }

    ~IRFFT_IPP()
    {
        if (workBuf) {
            ippFree(workBuf);
        }

        if (specBuf) {
            ippFree(specBuf);
        }
    }

    void transform(const std::complex<float>* input, float* output) override
    {
        if (ippsDFTInv_CCSToR_32f(reinterpret_cast<const float*>(input), output, specBuf, workBuf)
            != ippStsNoErr) {
            EAR_FAIL("Intel IPP IFFT failed.");
        }
    }

private:
    IppsDFTSpec_R_32f* specBuf;
    Ipp8u* initBuf = nullptr;
    Ipp8u* workBuf = nullptr;
};

class FFT_IPP : public FFT, Uncopyable
{
public:
    FFT_IPP(int length)
        : FFT(length)
    {
        int specSize, initBufSize, workBufSize;
        // arguments: length, normalization method (we do this ourselves), something deprecated, size spec, size init, size work
        ippsDFTGetSize_C_32fc(
            N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, &specSize, &initBufSize, &workBufSize);

        // allocate IPP buffers
        specBuf = reinterpret_cast<IppsDFTSpec_C_32fc*>(ippsMalloc_8u(specSize));
        initBuf = ippsMalloc_8u(initBufSize);
        workBuf = ippsMalloc_8u(workBufSize);

        // arguments: length, normalization method (we do this ourselves), something deprecated, spec buffer, init buffer
        ippsDFTInit_C_32fc(N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, specBuf, initBuf);

        if (initBuf) {
            ippFree(initBuf);
        }
    }

    ~FFT_IPP()
    {
        if (workBuf) {
            ippFree(workBuf);
        }

        if (specBuf) {
            ippFree(specBuf);
        }
    }

    void transform(const std::complex<float>* input, std::complex<float>* output) override
    {
        if (ippsDFTFwd_CToC_32fc(reinterpret_cast<const Ipp32fc*>(input),
                reinterpret_cast<Ipp32fc*>(output), specBuf, workBuf)
            != ippStsNoErr) {
            EAR_FAIL("Intel IPP FFT failed.");
        }
    }

private:
    IppsDFTSpec_C_32fc* specBuf = nullptr;
    Ipp8u* initBuf = nullptr;
    Ipp8u* workBuf = nullptr;
};

class IFFT_IPP : public IFFT
{
public:
    IFFT_IPP(int length)
        : IFFT(length)
    {
        int specSize, initBufSize, workBufSize;
        // arguments: length, normalization method (we do this ourselves), something deprecated, size spec, size init, size work
        ippsDFTGetSize_C_32fc(
            N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, &specSize, &initBufSize, &workBufSize);

        // allocate IPP buffers
        specBuf = reinterpret_cast<IppsDFTSpec_C_32fc*>(ippsMalloc_8u(specSize));
        initBuf = ippsMalloc_8u(initBufSize);
        workBuf = ippsMalloc_8u(workBufSize);

        // arguments: length, normalization method (we do this ourselves), something deprecated, spec buffer, init buffer
        ippsDFTInit_C_32fc(N, IPP_FFT_NODIV_BY_ANY, ippAlgHintNone, specBuf, initBuf);

        if (initBuf) {
            ippFree(initBuf);
        }
    }

    ~IFFT_IPP()
    {
        if (workBuf) {
            ippFree(workBuf);
        }

        if (specBuf) {
            ippFree(specBuf);
        }
    }

    void transform(const std::complex<float>* input, std::complex<float>* output) override
    {
        if (ippsDFTInv_CToC_32fc(reinterpret_cast<const Ipp32fc*>(input),
                reinterpret_cast<Ipp32fc*>(output), specBuf, workBuf)
            != ippStsNoErr) {
            EAR_FAIL("Intel IPP IFFT failed.");
        }
    }

private:
    IppsDFTSpec_C_32fc* specBuf;
    Ipp8u* initBuf = nullptr;
    Ipp8u* workBuf = nullptr;
};

/// Implementation of (R)FFT/I(R)FFT factory methods
std::unique_ptr<RFFT> RFFT::create(int length)
{
    return std::make_unique<RFFT_IPP>(length);
}

std::unique_ptr<IRFFT> IRFFT::create(int length)
{
    return std::make_unique<IRFFT_IPP>(length);
}

std::unique_ptr<FFT> FFT::create(int length)
{
    return std::make_unique<FFT_IPP>(length);
}

std::unique_ptr<IFFT> IFFT::create(int length)
{
    return std::make_unique<IFFT_IPP>(length);
}

} // namespace EAR

#endif /* EAR_FFT_IPP */
