/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Nokia Technologies Ltd. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#ifndef EAR_NE10_FFT
#define EAR_NE10_FFT

#include <EAR/Exception.h>
#include <EAR/FFT.h>
#include <EAR/Uncopyable.h>

#include <cmath>

#include "NE10.h"

#include <iostream>

namespace EAR {

// To avoid copying std::complex<float> data is passed directly as NE10 structs.
// This is slightly unsafe, so make at least sure that the data type sizes match.
static_assert(sizeof(ne10_fft_cpx_float32_t) == sizeof(std::complex<float>));

class RFFT_NE10 : public RFFT, Uncopyable
{
public:
    RFFT_NE10(int length)
        : RFFT(length)
    {
        EAR_CHECK_REL(ne10_init() == NE10_OK, "Failed to initialise Ne10.");
        m_cfg = ne10_fft_alloc_r2c_float32(length);
    }

    ~RFFT_NE10() { ne10_fft_destroy_r2c_float32(m_cfg); }

    void transform(const float* input, std::complex<float>* output) override
    {
        ne10_fft_r2c_1d_float32(reinterpret_cast<ne10_fft_cpx_float32_t*>(output),
            reinterpret_cast<ne10_float32_t*>(const_cast<float*>(input)), m_cfg);
    }

private:
    ne10_fft_r2c_cfg_float32_t m_cfg;
};

class IRFFT_NE10 : public IRFFT
{
public:
    IRFFT_NE10(int length)
        : IRFFT(length)
    {
        EAR_CHECK_REL(ne10_init() == NE10_OK, "Failed to initialise Ne10.");
        m_cfg = ne10_fft_alloc_r2c_float32(length);
    }

    ~IRFFT_NE10() { ne10_fft_destroy_r2c_float32(m_cfg); }

    void transform(const std::complex<float>* input, float* output) override
    {
        ne10_fft_c2r_1d_float32(reinterpret_cast<ne10_float32_t*>(output),
            reinterpret_cast<ne10_fft_cpx_float32_t*>(const_cast<std::complex<float>*>(input)),
            m_cfg);
        // Inverse transform requires scaling by length of the buffer
        for (unsigned int i = 0; i < m_cfg->nfft; i++) {
            output[i] *= static_cast<float>(m_cfg->nfft);
        }
    }

private:
    ne10_fft_r2c_cfg_float32_t m_cfg;
};

class FFT_NE10 : public FFT, Uncopyable
{
public:
    FFT_NE10(int length)
        : FFT(length)
    {
        EAR_CHECK_REL(ne10_init() == NE10_OK, "Failed to initialise Ne10.");
        m_cfg = ne10_fft_alloc_c2c_float32(length);
    }

    ~FFT_NE10() { ne10_fft_destroy_c2c_float32(m_cfg); }

    void transform(const std::complex<float>* input, std::complex<float>* output) override
    {
        ne10_fft_c2c_1d_float32(reinterpret_cast<ne10_fft_cpx_float32_t*>(output),
            reinterpret_cast<ne10_fft_cpx_float32_t*>(const_cast<std::complex<float>*>(input)),
            m_cfg, 0);
    }

private:
    ne10_fft_cfg_float32_t m_cfg;
};

class IFFT_NE10 : public IFFT
{
public:
    IFFT_NE10(int length)
        : IFFT(length)
    {
        EAR_CHECK_REL(ne10_init() == NE10_OK, "Failed to initialise Ne10.");
        m_cfg = ne10_fft_alloc_c2c_float32(length);
    }

    ~IFFT_NE10() { ne10_fft_destroy_c2c_float32(m_cfg); }

    void transform(const std::complex<float>* input, std::complex<float>* output) override
    {
        ne10_fft_c2c_1d_float32(reinterpret_cast<ne10_fft_cpx_float32_t*>(output),
            reinterpret_cast<ne10_fft_cpx_float32_t*>(const_cast<std::complex<float>*>(input)),
            m_cfg, 1);
        // Inverse transform requires scaling by length of the buffer
        for (unsigned int i = 0; i < m_cfg->nfft; i++) {
            output[i] *= static_cast<float>(m_cfg->nfft);
        }
    }

private:
    ne10_fft_cfg_float32_t m_cfg;
};

/// Implementation of (R)FFT/I(R)FFT factory methods
std::unique_ptr<RFFT> RFFT::create(int length)
{
    return std::make_unique<RFFT_NE10>(length);
}

std::unique_ptr<IRFFT> IRFFT::create(int length)
{
    return std::make_unique<IRFFT_NE10>(length);
}

std::unique_ptr<FFT> FFT::create(int length)
{
    return std::make_unique<FFT_NE10>(length);
}

std::unique_ptr<IFFT> IFFT::create(int length)
{
    return std::make_unique<IFFT_NE10>(length);
}

} // namespace EAR

#endif /* EAR_NE10_FFT */
