/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Exception.h>
#include <EAR/Geometry.h>
#include <EAR/SceneObjectFactory.h>
#include <EAR/TimeMeasure.h>

#include <algorithm>
#include <ctime>
#include <iostream>

using namespace EAR;
using namespace std;

static void genTestMesh(Mesh& mesh);
static void comparisonEmbreeBoxIsInside();
static void performanceTestClosestPointOnEllipse();

int main(int argc, char* argv[])
{

    //  try {
    //      auto mesh = SceneObjectFactory<Mesh>::getNewObject("TestMesh");
    //      genTestMesh(*mesh);
    //      mesh->rebuildSurfaceStructure();
    //  } catch (EAR::Exception e) {
    //      cout << e.toString() << endl;
    //  }

    // isInsidePerformanceTest();
    // comparisonEmbreeBoxIsInside();
    performanceTestClosestPointOnEllipse();
}

static void performanceTestClosestPointOnEllipse()
{
    //elliptical cylinder
    float a = 5;
    float b = 1;
    float h = 4;
    unsigned int iterN = 3;
    unsigned int Ncycles = 10000;

    auto tubeElliptical = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(a, b, h);
    std::vector<float> durations;

    EAR::Vector3 arbitraryPoint;
    //newton approach (This has been disabled because the below appraoch is more efficient)
    for (unsigned int i = 0; i < 10000; i++) {
        float x = std::rand() % 10 + 5.1;
        float y = std::rand() % 10 + 2.1;
        float z = std::rand() % 10 + 1.1;
        arbitraryPoint = EAR::Vector3(x, y, z);
        TimeMeasure::startTimer();
        tubeElliptical->closestPointTo(arbitraryPoint);
        TimeMeasure::stopTimer();

        durations.push_back(TimeMeasure::getElapsedMicroseconds());
    }

    std::cout << "Performance test closestPoint on ellipse allgorithms: "
              << tubeElliptical->getName() << std::endl;

    float mean = 0;
    float median = 0;
    for (int i = 0; i < Ncycles; i++) {
        mean += durations[i];
    }

    mean = mean / durations.size();

    std::sort(durations.begin(), durations.end());

    median = durations[int(durations.size() / 2)];

    std::cout << "Mean: " << mean << std::endl;
    std::cout << "Median: " << median << std::endl;
}

static void genTestMesh(Mesh& mesh)
{
    mesh.addVertex({ -314.9606, 0, 314.9606 });
    mesh.addVertex({ -196.8504, 0, 314.9606 });
    mesh.addVertex({ -196.8504, 0, 118.1102 });
    mesh.addVertex({ -314.9606, 0, 0 });
    mesh.addVertex({ 0, 0, 0 });
    mesh.addVertex({ 0, 0, 118.1102 });
    mesh.addVertex({ 0, 118.1102, 118.1102 });
    mesh.addVertex({ 0, 0, 118.1102 });
    mesh.addVertex({ -196.8504, 0, 118.1102 });
    mesh.addVertex({ -196.8504, 118.1102, 118.1102 });
    mesh.addVertex({ -196.8504, 118.1102, 314.9607 });
    mesh.addVertex({ -196.8504, 118.1102, 118.1102 });
    mesh.addVertex({ -196.8504, 0, 118.1102 });
    mesh.addVertex({ -196.8504, 0, 314.9606 });
    mesh.addVertex({ -196.8504, 118.1102, 314.9607 });
    mesh.addVertex({ -196.8504, 0, 314.9606 });
    mesh.addVertex({ -314.9606, 0, 314.9606 });
    mesh.addVertex({ -314.9606, 118.1102, 314.9607 });
    mesh.addVertex({ -314.9606, 0, 314.9606 });
    mesh.addVertex({ -314.9606, 0, 0 });
    mesh.addVertex({ -314.9606, 118.1102, 0 });
    mesh.addVertex({ -314.9606, 118.1102, 314.9607 });
    mesh.addVertex({ -314.9606, 118.1102, 0 });
    mesh.addVertex({ -314.9606, 0, 0 });
    mesh.addVertex({ 0, 0, 0 });
    mesh.addVertex({ 0, 118.1102, 0 });
    mesh.addVertex({ 0, 118.1102, 118.1102 });
    mesh.addVertex({ 0, 118.1102, 0 });
    mesh.addVertex({ 0, 0, 0 });
    mesh.addVertex({ 0, 0, 118.1102 });
    mesh.addVertex({ -314.9606, 118.1102, 0 });
    mesh.addVertex({ 0, 118.1102, 0 });
    mesh.addVertex({ 0, 118.1102, 118.1102 });
    mesh.addVertex({ -196.8504, 118.1102, 118.1102 });
    mesh.addVertex({ -196.8504, 118.1102, 314.9607 });
    mesh.addVertex({ -314.9606, 118.1102, 314.9607 });
    mesh.addTriangle(0, 1, 2, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(3, 0, 2, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(4, 3, 2, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(2, 5, 4, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(6, 7, 8, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(8, 9, 6, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(10, 11, 12, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(12, 13, 10, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(14, 15, 16, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(16, 17, 14, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(18, 19, 20, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(20, 21, 18, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(22, 23, 24, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(24, 25, 22, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(26, 27, 28, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(28, 29, 26, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(30, 31, 32, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(32, 33, 30, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(33, 34, 30, std::unique_ptr<Material>(new Material()));
    mesh.addTriangle(34, 35, 30, std::unique_ptr<Material>(new Material()));
}

static void comparisonEmbreeBoxIsInside()
{
    auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

    mesh->addVertex(EAR::Vector3(1, 0, 0));
    mesh->addVertex(EAR::Vector3(0, 0, 0));
    mesh->addVertex(EAR::Vector3(1, 1, 0));
    mesh->addVertex(EAR::Vector3(0, 1, 0));

    mesh->addVertex(EAR::Vector3(1, 1, 1));
    mesh->addVertex(EAR::Vector3(0, 1, 1));
    mesh->addVertex(EAR::Vector3(1, 0, 1));
    mesh->addVertex(EAR::Vector3(0, 0, 1));

    mesh->addTriangle(0, 2, 3);
    mesh->addTriangle(0, 3, 1);
    mesh->addTriangle(2, 4, 5);
    mesh->addTriangle(2, 5, 3);
    mesh->addTriangle(4, 6, 7);
    mesh->addTriangle(4, 7, 5);
    mesh->addTriangle(6, 0, 1);
    mesh->addTriangle(6, 1, 7);
    mesh->addTriangle(1, 3, 5);
    mesh->addTriangle(1, 5, 7);
    mesh->addTriangle(6, 4, 2);
    mesh->addTriangle(6, 2, 0);

    mesh->lock();

    auto cube = EAR::SceneObjectFactory<EAR::Box>::getNewObject(1.f, 1.f, 1.f);

    cube->setLocation(EAR::Vector3(0.5, 0.5, 0.5));

    bool embree;
    bool box;
    int counter = 0;
    srand((int)time(0));
    auto X = 4;
    for (int i = 0; i < 10000; i++) {
        //float x = (std::rand() % 5) - 2;
        //float y = (std::rand() % 5) - 2;
        //float z = (std::rand() % 5) - 2;

        float x = (static_cast<float>(rand()) / (static_cast<float>(RAND_MAX / X))) - 2;
        float y = (static_cast<float>(rand()) / (static_cast<float>(RAND_MAX / X))) - 2;
        float z = (static_cast<float>(rand()) / (static_cast<float>(RAND_MAX / X))) - 2;

        Vector3 point(x, y, z);
        embree = mesh->isInside(point);
        box = cube->isInside(point);

        if (embree != box) {

            std::cout << "Point: " << point << "return: " << embree << std::endl;
            counter++;
        }
    }

    std::cout << "counter: " << counter << std::endl;

    return;
}
