/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/Geometry.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

namespace MPEGI {

using Vertex = EAR::Vector3;
using Face = std::array<unsigned int, 3>;

TEST_CASE("Icosahedron Mesh")
{
    std::vector<Vertex> target_vertices = {
        { -2.92942810, 4.73991442, 0.00000000 },
        { 2.92942810, 4.73991442, 0.00000000 },
        { -2.92942810, -4.73991442, 0.00000000 },
        { 2.92942810, -4.73991442, 0.00000000 },
        { 0.00000000, -2.92942810, 4.73991442 },
        { 0.00000000, 2.92942810, 4.73991442 },
        { 0.00000000, -2.92942810, -4.73991442 },
        { 0.00000000, 2.92942810, -4.73991442 },
        { 4.73991442, 0.00000000, -2.92942810 },
        { 4.73991442, 0.00000000, 2.92942810 },
        { -4.73991442, 0.00000000, -2.92942810 },
        { -4.73991442, 0.00000000, 2.92942810 },
    };
    std::vector<Face> target_faces = {
        { 0, 11, 5 },
        { 0, 5, 1 },
        { 0, 1, 7 },
        { 0, 7, 10 },
        { 0, 10, 11 },
        { 1, 5, 9 },
        { 5, 11, 4 },
        { 11, 10, 2 },
        { 10, 7, 6 },
        { 7, 1, 8 },
        { 3, 9, 4 },
        { 3, 4, 2 },
        { 3, 2, 6 },
        { 3, 6, 8 },
        { 3, 8, 9 },
        { 4, 9, 5 },
        { 2, 4, 11 },
        { 6, 2, 10 },
        { 8, 6, 7 },
        { 9, 8, 1 },
    };

    float radius = 5.f;
    const std::string& name = "sphere:TestSphere";
    EAR::Position pos({ 1.f, 2.f, 3.f }, { 0.1f, 0.2f, 0.3f });
    unsigned int subdivisions = 0;
    auto sph = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(radius, name, pos, subdivisions);

    const auto& mesh = sph->getMesh();
    const auto& vertices = mesh->getVertices();
    const auto& triangles = mesh->getTriangles();

    CHECK(vertices.size() == target_vertices.size());
    for (int n = 0; n < vertices.size(); n++) {
        CHECK(vertices[n] == target_vertices[n]);
    }

    CHECK(triangles.size() == target_faces.size());
    for (int n = 0; n < triangles.size(); n++) {
        CHECK(triangles[n].v1 == target_faces[n][0]);
        CHECK(triangles[n].v2 == target_faces[n][1]);
        CHECK(triangles[n].v3 == target_faces[n][2]);
    }
}

TEST_CASE("Icosahedron Mesh Sub-Division")
{
    std::vector<Vertex> target_vertices = {
        { -2.65219784, 4.29134607, 0.00000000 },
        { 2.65219784, 4.29134607, 0.00000000 },
        { -2.65219784, -4.29134607, 0.00000000 },
        { 2.65219784, -4.29134607, 0.00000000 },
        { 0.00000000, -2.65219784, 4.29134607 },
        { 0.00000000, 2.65219784, 4.29134607 },
        { 0.00000000, -2.65219784, -4.29134607 },
        { 0.00000000, 2.65219784, -4.29134607 },
        { 4.29134607, 0.00000000, -2.65219784 },
        { 4.29134607, 0.00000000, 2.65219784 },
        { -4.29134607, 0.00000000, -2.65219784 },
        { -4.29134607, 0.00000000, 2.65219784 },
        { -4.08131218, 2.52238989, 1.55892253 },
        { -2.52238989, 1.55892253, 4.08131218 },
        { -1.55892253, 4.08131218, 2.52238989 },
        { 1.55892253, 4.08131218, 2.52238989 },
        { 0.00000000, 5.04477978, 0.00000000 },
        { 1.55892253, 4.08131218, -2.52238989 },
        { -1.55892253, 4.08131218, -2.52238989 },
        { -2.52238989, 1.55892253, -4.08131218 },
        { -4.08131218, 2.52238989, -1.55892253 },
        { -5.04477978, 0.00000000, 0.00000000 },
        { 2.52238989, 1.55892253, 4.08131218 },
        { 4.08131218, 2.52238989, 1.55892253 },
        { -2.52238989, -1.55892253, 4.08131218 },
        { 0.00000000, 0.00000000, 5.04477978 },
        { -4.08131218, -2.52238989, -1.55892253 },
        { -4.08131218, -2.52238989, 1.55892253 },
        { 0.00000000, 0.00000000, -5.04477978 },
        { -2.52238989, -1.55892253, -4.08131218 },
        { 4.08131218, 2.52238989, -1.55892253 },
        { 2.52238989, 1.55892253, -4.08131218 },
        { 4.08131218, -2.52238989, 1.55892253 },
        { 2.52238989, -1.55892253, 4.08131218 },
        { 1.55892253, -4.08131218, 2.52238989 },
        { -1.55892253, -4.08131218, 2.52238989 },
        { 0.00000000, -5.04477978, 0.00000000 },
        { -1.55892253, -4.08131218, -2.52238989 },
        { 1.55892253, -4.08131218, -2.52238989 },
        { 2.52238989, -1.55892253, -4.08131218 },
        { 4.08131218, -2.52238989, -1.55892253 },
        { 5.04477978, 0.00000000, 0.00000000 },
        { -3.47177148, 2.14567280, 2.96524668 },
        { -2.14567280, 2.96524668, 3.47177148 },
        { -2.96524668, 3.47177148, 2.14567280 },
        { -3.49996972, 3.54166985, 0.810302675 },
        { -2.18887210, 4.35197258, 1.31109738 },
        { -3.54166985, 0.810302675, 3.49996972 },
        { -4.35197258, 1.31109738, 2.18887210 },
        { -0.810302675, 3.49996972, 3.54166985 },
        { -1.31109738, 2.18887210, 4.35197258 },
        { 0.00000000, 4.29134607, 2.65219760 },
        { 0.819574177, 4.79787064, 1.32609892 },
        { -0.819574177, 4.79787064, 1.32609892 },
        { -1.37856936, 4.85276699, 0.00000000 },
        { 0.810302675, 3.49996972, 3.54166985 },
        { 1.37856936, 4.85276699, 0.00000000 },
        { 2.18887210, 4.35197258, 1.31109738 },
        { 0.819574177, 4.79787064, -1.32609892 },
        { 0.00000000, 4.29134607, -2.65219760 },
        { -0.819574177, 4.79787064, -1.32609892 },
        { -2.18887210, 4.35197258, -1.31109738 },
        { 2.18887210, 4.35197258, -1.31109738 },
        { -0.810302675, 3.49996972, -3.54166985 },
        { 0.810302675, 3.49996972, -3.54166985 },
        { -2.14567280, 2.96524668, -3.47177148 },
        { -3.47177148, 2.14567280, -2.96524668 },
        { -2.96524668, 3.47177148, -2.14567280 },
        { -3.49996972, 3.54166985, -0.810302675 },
        { -1.31109738, 2.18887210, -4.35197258 },
        { -4.35197258, 1.31109738, -2.18887210 },
        { -3.54166985, 0.810302675, -3.49996972 },
        { -4.79787064, 1.32609892, -0.819574177 },
        { -4.79787064, 1.32609892, 0.819574177 },
        { -4.29134607, 2.65219760, 0.00000000 },
        { -4.85276699, 0.00000000, -1.37856936 },
        { -4.85276699, 0.00000000, 1.37856936 },
        { 2.14567280, 2.96524668, 3.47177148 },
        { 3.47177148, 2.14567280, 2.96524668 },
        { 2.96524668, 3.47177148, 2.14567280 },
        { 3.49996972, 3.54166985, 0.810302675 },
        { 1.31109738, 2.18887210, 4.35197258 },
        { 4.35197258, 1.31109738, 2.18887210 },
        { 3.54166985, 0.810302675, 3.49996972 },
        { -2.65219760, 0.00000000, 4.29134607 },
        { -1.32609892, -0.819574177, 4.79787064 },
        { -1.32609892, 0.819574177, 4.79787064 },
        { 0.00000000, 1.37856936, 4.85276699 },
        { -3.54166985, -0.810302675, 3.49996972 },
        { 0.00000000, -1.37856936, 4.85276699 },
        { -1.31109738, -2.18887210, 4.35197258 },
        { -4.79787064, -1.32609892, -0.819574177 },
        { -4.29134607, -2.65219760, 0.00000000 },
        { -4.79787064, -1.32609892, 0.819574177 },
        { -4.35197258, -1.31109738, 2.18887210 },
        { -4.35197258, -1.31109738, -2.18887210 },
        { -3.49996972, -3.54166985, 0.810302675 },
        { -3.49996972, -3.54166985, -0.810302675 },
        { -1.32609892, 0.819574177, -4.79787064 },
        { -1.32609892, -0.819574177, -4.79787064 },
        { -2.65219760, 0.00000000, -4.29134607 },
        { -3.54166985, -0.810302675, -3.49996972 },
        { 0.00000000, 1.37856936, -4.85276699 },
        { -1.31109738, -2.18887210, -4.35197258 },
        { 0.00000000, -1.37856936, -4.85276699 },
        { 2.96524668, 3.47177148, -2.14567280 },
        { 3.47177148, 2.14567280, -2.96524668 },
        { 2.14567280, 2.96524668, -3.47177148 },
        { 1.31109738, 2.18887210, -4.35197258 },
        { 3.49996972, 3.54166985, -0.810302675 },
        { 3.54166985, 0.810302675, -3.49996972 },
        { 4.35197258, 1.31109738, -2.18887210 },
        { 3.47177148, -2.14567280, 2.96524668 },
        { 2.14567280, -2.96524668, 3.47177148 },
        { 2.96524668, -3.47177148, 2.14567280 },
        { 3.49996972, -3.54166985, 0.810302675 },
        { 2.18887210, -4.35197258, 1.31109738 },
        { 3.54166985, -0.810302675, 3.49996972 },
        { 4.35197258, -1.31109738, 2.18887210 },
        { 0.810302675, -3.49996972, 3.54166985 },
        { 1.31109738, -2.18887210, 4.35197258 },
        { 0.00000000, -4.29134607, 2.65219760 },
        { -0.819574177, -4.79787064, 1.32609892 },
        { 0.819574177, -4.79787064, 1.32609892 },
        { 1.37856936, -4.85276699, 0.00000000 },
        { -0.810302675, -3.49996972, 3.54166985 },
        { -1.37856936, -4.85276699, 0.00000000 },
        { -2.18887210, -4.35197258, 1.31109738 },
        { -0.819574177, -4.79787064, -1.32609892 },
        { 0.00000000, -4.29134607, -2.65219760 },
        { 0.819574177, -4.79787064, -1.32609892 },
        { 2.18887210, -4.35197258, -1.31109738 },
        { -2.18887210, -4.35197258, -1.31109738 },
        { 0.810302675, -3.49996972, -3.54166985 },
        { -0.810302675, -3.49996972, -3.54166985 },
        { 2.14567280, -2.96524668, -3.47177148 },
        { 3.47177148, -2.14567280, -2.96524668 },
        { 2.96524668, -3.47177148, -2.14567280 },
        { 3.49996972, -3.54166985, -0.810302675 },
        { 1.31109738, -2.18887210, -4.35197258 },
        { 4.35197258, -1.31109738, -2.18887210 },
        { 3.54166985, -0.810302675, -3.49996972 },
        { 4.79787064, -1.32609892, -0.819574177 },
        { 4.79787064, -1.32609892, 0.819574177 },
        { 4.29134607, -2.65219760, 0.00000000 },
        { 4.85276699, 0.00000000, -1.37856936 },
        { 4.85276699, 0.00000000, 1.37856936 },
        { 2.65219760, 0.00000000, 4.29134607 },
        { 1.32609892, 0.819574177, 4.79787064 },
        { 1.32609892, -0.819574177, 4.79787064 },
        { -2.14567280, -2.96524668, 3.47177148 },
        { -3.47177148, -2.14567280, 2.96524668 },
        { -2.96524668, -3.47177148, 2.14567280 },
        { -2.96524668, -3.47177148, -2.14567280 },
        { -3.47177148, -2.14567280, -2.96524668 },
        { -2.14567280, -2.96524668, -3.47177148 },
        { 1.32609892, -0.819574177, -4.79787064 },
        { 1.32609892, 0.819574177, -4.79787064 },
        { 2.65219760, 0.00000000, -4.29134607 },
        { 4.79787064, 1.32609892, -0.819574177 },
        { 4.29134607, 2.65219760, 0.00000000 },
        { 4.79787064, 1.32609892, 0.819574177 },
    };
    std::vector<Face> target_faces = {
        { 42, 43, 44 },
        { 12, 42, 44 },
        { 13, 43, 42 },
        { 14, 44, 43 },
        { 45, 44, 46 },
        { 0, 45, 46 },
        { 12, 44, 45 },
        { 14, 46, 44 },
        { 47, 42, 48 },
        { 11, 47, 48 },
        { 13, 42, 47 },
        { 12, 48, 42 },
        { 49, 43, 50 },
        { 5, 49, 50 },
        { 14, 43, 49 },
        { 13, 50, 43 },
        { 51, 52, 53 },
        { 14, 51, 53 },
        { 15, 52, 51 },
        { 16, 53, 52 },
        { 46, 53, 54 },
        { 0, 46, 54 },
        { 14, 53, 46 },
        { 16, 54, 53 },
        { 55, 51, 49 },
        { 5, 55, 49 },
        { 15, 51, 55 },
        { 14, 49, 51 },
        { 56, 52, 57 },
        { 1, 56, 57 },
        { 16, 52, 56 },
        { 15, 57, 52 },
        { 58, 59, 60 },
        { 16, 58, 60 },
        { 17, 59, 58 },
        { 18, 60, 59 },
        { 54, 60, 61 },
        { 0, 54, 61 },
        { 16, 60, 54 },
        { 18, 61, 60 },
        { 62, 58, 56 },
        { 1, 62, 56 },
        { 17, 58, 62 },
        { 16, 56, 58 },
        { 63, 59, 64 },
        { 7, 63, 64 },
        { 18, 59, 63 },
        { 17, 64, 59 },
        { 65, 66, 67 },
        { 18, 65, 67 },
        { 19, 66, 65 },
        { 20, 67, 66 },
        { 61, 67, 68 },
        { 0, 61, 68 },
        { 18, 67, 61 },
        { 20, 68, 67 },
        { 69, 65, 63 },
        { 7, 69, 63 },
        { 19, 65, 69 },
        { 18, 63, 65 },
        { 70, 66, 71 },
        { 10, 70, 71 },
        { 20, 66, 70 },
        { 19, 71, 66 },
        { 72, 73, 74 },
        { 20, 72, 74 },
        { 21, 73, 72 },
        { 12, 74, 73 },
        { 68, 74, 45 },
        { 0, 68, 45 },
        { 20, 74, 68 },
        { 12, 45, 74 },
        { 75, 72, 70 },
        { 10, 75, 70 },
        { 21, 72, 75 },
        { 20, 70, 72 },
        { 48, 73, 76 },
        { 11, 48, 76 },
        { 12, 73, 48 },
        { 21, 76, 73 },
        { 77, 78, 79 },
        { 15, 77, 79 },
        { 22, 78, 77 },
        { 23, 79, 78 },
        { 57, 79, 80 },
        { 1, 57, 80 },
        { 15, 79, 57 },
        { 23, 80, 79 },
        { 81, 77, 55 },
        { 5, 81, 55 },
        { 22, 77, 81 },
        { 15, 55, 77 },
        { 82, 78, 83 },
        { 9, 82, 83 },
        { 23, 78, 82 },
        { 22, 83, 78 },
        { 84, 85, 86 },
        { 13, 84, 86 },
        { 24, 85, 84 },
        { 25, 86, 85 },
        { 50, 86, 87 },
        { 5, 50, 87 },
        { 13, 86, 50 },
        { 25, 87, 86 },
        { 88, 84, 47 },
        { 11, 88, 47 },
        { 24, 84, 88 },
        { 13, 47, 84 },
        { 89, 85, 90 },
        { 4, 89, 90 },
        { 25, 85, 89 },
        { 24, 90, 85 },
        { 91, 92, 93 },
        { 21, 91, 93 },
        { 26, 92, 91 },
        { 27, 93, 92 },
        { 76, 93, 94 },
        { 11, 76, 94 },
        { 21, 93, 76 },
        { 27, 94, 93 },
        { 95, 91, 75 },
        { 10, 95, 75 },
        { 26, 91, 95 },
        { 21, 75, 91 },
        { 96, 92, 97 },
        { 2, 96, 97 },
        { 27, 92, 96 },
        { 26, 97, 92 },
        { 98, 99, 100 },
        { 19, 98, 100 },
        { 28, 99, 98 },
        { 29, 100, 99 },
        { 71, 100, 101 },
        { 10, 71, 101 },
        { 19, 100, 71 },
        { 29, 101, 100 },
        { 102, 98, 69 },
        { 7, 102, 69 },
        { 28, 98, 102 },
        { 19, 69, 98 },
        { 103, 99, 104 },
        { 6, 103, 104 },
        { 29, 99, 103 },
        { 28, 104, 99 },
        { 105, 106, 107 },
        { 17, 105, 107 },
        { 30, 106, 105 },
        { 31, 107, 106 },
        { 64, 107, 108 },
        { 7, 64, 108 },
        { 17, 107, 64 },
        { 31, 108, 107 },
        { 109, 105, 62 },
        { 1, 109, 62 },
        { 30, 105, 109 },
        { 17, 62, 105 },
        { 110, 106, 111 },
        { 8, 110, 111 },
        { 31, 106, 110 },
        { 30, 111, 106 },
        { 112, 113, 114 },
        { 32, 112, 114 },
        { 33, 113, 112 },
        { 34, 114, 113 },
        { 115, 114, 116 },
        { 3, 115, 116 },
        { 32, 114, 115 },
        { 34, 116, 114 },
        { 117, 112, 118 },
        { 9, 117, 118 },
        { 33, 112, 117 },
        { 32, 118, 112 },
        { 119, 113, 120 },
        { 4, 119, 120 },
        { 34, 113, 119 },
        { 33, 120, 113 },
        { 121, 122, 123 },
        { 34, 121, 123 },
        { 35, 122, 121 },
        { 36, 123, 122 },
        { 116, 123, 124 },
        { 3, 116, 124 },
        { 34, 123, 116 },
        { 36, 124, 123 },
        { 125, 121, 119 },
        { 4, 125, 119 },
        { 35, 121, 125 },
        { 34, 119, 121 },
        { 126, 122, 127 },
        { 2, 126, 127 },
        { 36, 122, 126 },
        { 35, 127, 122 },
        { 128, 129, 130 },
        { 36, 128, 130 },
        { 37, 129, 128 },
        { 38, 130, 129 },
        { 124, 130, 131 },
        { 3, 124, 131 },
        { 36, 130, 124 },
        { 38, 131, 130 },
        { 132, 128, 126 },
        { 2, 132, 126 },
        { 37, 128, 132 },
        { 36, 126, 128 },
        { 133, 129, 134 },
        { 6, 133, 134 },
        { 38, 129, 133 },
        { 37, 134, 129 },
        { 135, 136, 137 },
        { 38, 135, 137 },
        { 39, 136, 135 },
        { 40, 137, 136 },
        { 131, 137, 138 },
        { 3, 131, 138 },
        { 38, 137, 131 },
        { 40, 138, 137 },
        { 139, 135, 133 },
        { 6, 139, 133 },
        { 39, 135, 139 },
        { 38, 133, 135 },
        { 140, 136, 141 },
        { 8, 140, 141 },
        { 40, 136, 140 },
        { 39, 141, 136 },
        { 142, 143, 144 },
        { 40, 142, 144 },
        { 41, 143, 142 },
        { 32, 144, 143 },
        { 138, 144, 115 },
        { 3, 138, 115 },
        { 40, 144, 138 },
        { 32, 115, 144 },
        { 145, 142, 140 },
        { 8, 145, 140 },
        { 41, 142, 145 },
        { 40, 140, 142 },
        { 118, 143, 146 },
        { 9, 118, 146 },
        { 32, 143, 118 },
        { 41, 146, 143 },
        { 147, 148, 149 },
        { 33, 147, 149 },
        { 22, 148, 147 },
        { 25, 149, 148 },
        { 120, 149, 89 },
        { 4, 120, 89 },
        { 33, 149, 120 },
        { 25, 89, 149 },
        { 83, 147, 117 },
        { 9, 83, 117 },
        { 22, 147, 83 },
        { 33, 117, 147 },
        { 87, 148, 81 },
        { 5, 87, 81 },
        { 25, 148, 87 },
        { 22, 81, 148 },
        { 150, 151, 152 },
        { 35, 150, 152 },
        { 24, 151, 150 },
        { 27, 152, 151 },
        { 127, 152, 96 },
        { 2, 127, 96 },
        { 35, 152, 127 },
        { 27, 96, 152 },
        { 90, 150, 125 },
        { 4, 90, 125 },
        { 24, 150, 90 },
        { 35, 125, 150 },
        { 94, 151, 88 },
        { 11, 94, 88 },
        { 27, 151, 94 },
        { 24, 88, 151 },
        { 153, 154, 155 },
        { 37, 153, 155 },
        { 26, 154, 153 },
        { 29, 155, 154 },
        { 134, 155, 103 },
        { 6, 134, 103 },
        { 37, 155, 134 },
        { 29, 103, 155 },
        { 97, 153, 132 },
        { 2, 97, 132 },
        { 26, 153, 97 },
        { 37, 132, 153 },
        { 101, 154, 95 },
        { 10, 101, 95 },
        { 29, 154, 101 },
        { 26, 95, 154 },
        { 156, 157, 158 },
        { 39, 156, 158 },
        { 28, 157, 156 },
        { 31, 158, 157 },
        { 141, 158, 110 },
        { 8, 141, 110 },
        { 39, 158, 141 },
        { 31, 110, 158 },
        { 104, 156, 139 },
        { 6, 104, 139 },
        { 28, 156, 104 },
        { 39, 139, 156 },
        { 108, 157, 102 },
        { 7, 108, 102 },
        { 31, 157, 108 },
        { 28, 102, 157 },
        { 159, 160, 161 },
        { 41, 159, 161 },
        { 30, 160, 159 },
        { 23, 161, 160 },
        { 146, 161, 82 },
        { 9, 146, 82 },
        { 41, 161, 146 },
        { 23, 82, 161 },
        { 111, 159, 145 },
        { 8, 111, 145 },
        { 30, 159, 111 },
        { 41, 145, 159 },
        { 80, 160, 109 },
        { 1, 80, 109 },
        { 23, 160, 80 },
        { 30, 109, 160 },
    };

    float radius = 5.f;
    const std::string& name = "sphere:TestSphere";
    EAR::Position pos({ 1.f, 2.f, 3.f }, { 0.1f, 0.2f, 0.3f });
    unsigned int subdivisions = 2;
    auto sph = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(radius, name, pos, subdivisions);

    const auto& mesh = sph->getMesh();
    const auto& vertices = mesh->getVertices();
    const auto& triangles = mesh->getTriangles();

    CHECK(vertices.size() == target_vertices.size());
    for (int n = 0; n < vertices.size(); n++) {
        CHECK(vertices[n] == target_vertices[n]);
    }

    CHECK(triangles.size() == target_faces.size());
    for (int n = 0; n < triangles.size(); n++) {
        CHECK(triangles[n].v1 == target_faces[n][0]);
        CHECK(triangles[n].v2 == target_faces[n][1]);
        CHECK(triangles[n].v3 == target_faces[n][2]);
    }
}

} // namespace MPEGI
