import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import sys
from scikits.audiolab import Sndfile
from scipy.signal import stft

"""Helper functions"""

octave_center_freqs = np.array([16, 31.5, 63, 125, 250, 500, 1000, 2000, 4000, 8000, 16000]);
octave_center_freqs_names = np.array(['16', '31.5', '63', '125', '250', '500', '1k', '2k', '4k', '8k', '16k']);

def loadWavefile(filename):
    """Loads a WAV file from the harddisk"""
    file = Sndfile(filename, mode='r')
    x = file.read_frames(file.nframes, dtype=np.float32)
    if len(x.shape) == 1:
        x = np.array([x])
    else:
        x = x.T
    return x, file.samplerate

def plot(h, fs):
    """Plot time-domain representation of a signal"""
    fig = plt.figure()

    t = np.arange(0, h.shape[1])/float(fs)*1000  # Time axis

    ax = plt.subplot2grid((1, 1), (0, 0), rowspan=1, colspan=1)
    for ch in h:
        plt.plot(t, ch)

    plt.xlim(0, np.max(t))
    m = np.max(np.abs(h))
    plt.ylim(-m*1.2, m*1.2)
    plt.xlabel('Time [ms]')
    plt.ylabel('Amplitude')
    ax.xaxis.grid(True);


def plotFT(h, fs):
    """Plot frequency-domain representation of a signal"""
    fig = plt.figure()

    N = h.shape[1] # Length
    H = np.fft.rfft(h) # Spectral coefficients (all channels)
    
    K = int(np.ceil(N/2))
    f = np.arange(K)*float(fs)/N # Bin frequencies [Hz] 

    ax = plt.subplot2grid((1, 1), (0, 0), rowspan=1, colspan=1)
    a = 10
    b = a - 180

    for ch in H:
        H_abs = np.abs(ch[0:K])
        H_db = 20*np.log10(H_abs/np.max(H_abs))
        plt.semilogx( f, H_db )


    plt.xticks(octave_center_freqs, octave_center_freqs_names)
    plt.yticks(np.arange(b, a, 10))
    plt.grid(True)
    plt.xlim([np.max([f[0], 8.]), fs/2])
    plt.ylim(b, a)
    plt.xlabel('Frequency [Hz]')
    plt.ylabel('Modulus [dB]')


def plotSTFT(h, fs):
    fig = plt.figure()
    if len(h.shape) > 1: # use only first channel
        h = h[0]
    f, t, H = stft(h, fs=fs, nperseg=8192*2, noverlap=8192)
    H_db = np.clip(20*np.log10(np.abs(H)), -120, 0) # clip at -120 dB

    plt.yscale('log')
    plt.ylim([np.max([f[0], 8.]), fs/2])

    plt.pcolormesh(t, f, H_db, cmap=sns.cubehelix_palette(reverse=True, dark=0, light=1, as_cmap=True))

    plt.yticks(octave_center_freqs, octave_center_freqs_names)
    plt.xlabel('Time [s]')
    plt.ylabel('Frequency [Hz]')

    cbar = plt.colorbar()
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel('Magnitude [dBFS]', rotation=270)



data, fs = loadWavefile(sys.argv[1])
# data2, fs2 = loadWavefile(sys.argv[2])

# plotFT(data, fs)
# plotFT(data2, fs)

plotSTFT(data, fs)
# plotSTFT(data2, fs2)

# plotFT(data, fs)
    
plt.show()
