/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/BaseTypes.h>
// #include <EAR/TimeMeasure.h>

#include <cstdlib>
#include <iostream>
#include <vector>

TEST_CASE("Base Types")
{
    SUBCASE("Vector3 arithmetic")
    {
        EAR::Vector3 a(1, 1, 1);
        EAR::Vector3 b(-2, -1, 0);
        EAR::Vector3 c;

        // zero init?
        CHECK(c[0] == 0.f);
        CHECK(c[1] == 0.f);
        CHECK(c[2] == 0.f);

        CHECK(a.abs() == doctest::Approx(1.73205)); // sqrt(1+1+1)

        c = a + b;
        CHECK(c[0] == -1); // a0+b0
        CHECK(c[1] == 0); // a1+b1
        CHECK(c[2] == 1); // a2+b2

        c = a - b;
        CHECK(c[0] == 3); // a0-b0
        CHECK(c[1] == 2); // a1-b1
        CHECK(c[2] == 1); // a2-b2

        c = a * b;
        CHECK(c[0] == 1); // a1*b2 - a2*b1
        CHECK(c[1] == -2); // a2*b0 - a0*b2
        CHECK(c[2] == 1); // a0*b1 - a1*b0

        c = a.cross(b);
        CHECK(c[0] == 1); // a1*b2 - a2*b1
        CHECK(c[1] == -2); // a2*b0 - a0*b2
        CHECK(c[2] == 1); // a0*b1 - a1*b0

        CHECK(a.dot(b) == -3); // a0*b0 +  a1*b1 + a2*b2

        c = 2 * a;
        CHECK(c[0] == 2);
        CHECK(c[1] == 2);
        CHECK(c[2] == 2);

        c = a * -3;
        CHECK(c[0] == -3);
        CHECK(c[1] == -3);
        CHECK(c[2] == -3);
    }

    SUBCASE("cartesian -> polar -> cartesian")
    {
        EAR::Vector3 a(1, 1, 0);

        EAR::Polar p = cartesianToPolar(a);

        CHECK(p[0] == doctest::Approx(-90.f)); // ccw azimuth
        CHECK(p[1] == doctest::Approx(45.f)); // up positive
        CHECK(p[2] == doctest::Approx(1.41421)); //sqrt(2)

        // back to cartesian should bring us where we started
        EAR::Vector3 a_ret = polarToCartesian(p);
        CHECK(a_ret[0] == doctest::Approx(a[0]));
        CHECK(a_ret[1] == doctest::Approx(a[1]));
        CHECK(a_ret[2] == doctest::Approx(a[2]));
    }

    SUBCASE("homogeneous rotations")
    {
        EAR::HomogenMatrix m;
        m.setRotation(EAR::Euler(90, 0, 0));
        m.setTranslation(EAR::Vector3(0, 0, 0));

        EAR::Vector3 a(1, 0, 0);
        EAR::Vector3 b = m * a;

        CHECK(b[0] == doctest::Approx(0));
        CHECK(b[1] == doctest::Approx(0));
        CHECK(b[2] == doctest::Approx(-1));

        m.setTranslation(EAR::Vector3(2, -1, 1));
        b = m.transform(a);

        CHECK(b[0] == doctest::Approx(2));
        CHECK(b[1] == doctest::Approx(-1));
        CHECK(b[2] == doctest::Approx(0));

        m.setRotation(EAR::Euler(90, 180, 45));
        m.setTranslation(EAR::Vector3(0, 0, 0));
        b = m * a;

        CHECK(b[0] == doctest::Approx(0));
        CHECK(b[1] == doctest::Approx(-0.70711));
        CHECK(b[2] == doctest::Approx(-0.70711));

        m.setRotation(EAR::Euler(0, 90, 0));
        m.setTranslation(EAR::Vector3(-1, 1, -1));
        EAR::HomogenMatrix m2;
        m2.setRotation(EAR::Euler(0, 0, 90));
        m2.setTranslation(EAR::Vector3(0, 0, 0));
        EAR::HomogenMatrix mc = m * m2;

        b = mc * a;

        CHECK(b[0] == doctest::Approx(-1));
        CHECK(b[1] == doctest::Approx(1));
        CHECK(b[2] == doctest::Approx(0));

        m.setRotation(EAR::Euler(90, 180, 45));
        m.setTranslation(EAR::Vector3(-2, 4, -17));
        a.set(1.5, -0.25, 3);
        EAR::HomogenMatrix mi = m.inverse();
        b = m * a;
        b = mi * b;

        CHECK(b[0] == doctest::Approx(a[0]));
        CHECK(b[1] == doctest::Approx(a[1]));
        CHECK(b[2] == doctest::Approx(a[2]));

        m.setRotation(EAR::Euler(45, -45, 90));
        m.setTranslation(EAR::Vector3(-2, 4, -17));

        EAR::Euler rot = m.getRotation();
        EAR::Vector3 tra = m.getTranslation();

        CHECK(rot[0] == doctest::Approx(45));
        CHECK(rot[1] == doctest::Approx(-45));
        CHECK(rot[2] == doctest::Approx(90));

        CHECK(tra[0] == doctest::Approx(-2));
        CHECK(tra[1] == doctest::Approx(4));
        CHECK(tra[2] == doctest::Approx(-17));

        m.setRotation(EAR::Euler(45, -45, 90));
        m.setTranslation(EAR::Vector3(-2, 4, -17));
        m2.setRotation(EAR::Euler(45, -45, 90));
        m2.setTranslation(EAR::Vector3(-2, 4, -17));

        CHECK(m2 == m);
    }
    SUBCASE("Matrix3_3")
    {
        // check default constructor and (const) direct element access
        EAR::Matrix3_3 m;
        m[0][0] = 5;

        CHECK(m[0][0] == 5);
        CHECK(m[0][1] == 0);

        //check basic operators +(=),-(=),*(=),/(=) witf scalar and constructor form array
        float c[3][3] = { { 1, 2, 3 }, { 4, 5, 6 }, { 7, 8, 9 } };

        EAR::Matrix3_3 n(c);

        n += 1;

        CHECK(n[0][0] == 2);
        CHECK(n[0][1] == 3);

        n -= 1;

        CHECK(n[0][0] == 1);
        CHECK(n[0][1] == 2);

        n *= 2;

        CHECK(n[0][0] == 2);
        CHECK(n[0][1] == 4);

        n /= 2;

        CHECK(n[0][0] == 1);
        CHECK(n[0][1] == 2);

        //check basic operations (...) with matrices and constructor form initializer lists
        EAR::Matrix3_3 p({ { 1, 2, 3 }, { 4, 5, 6 }, { 7, 8, 9 } });

        p += p;

        CHECK(p[0][0] == 2);
        CHECK(p[0][1] == 4);

        p -= n;

        CHECK(p[0][0] == 1);
        CHECK(p[0][1] == 2);

        p *= p;

        CHECK(p[0][0] == 30);
        CHECK(p[0][1] == 36);

        //check matrix operations (determinant, inverse, trace, transpose)

        CHECK(n.getTrace() == 15);
        CHECK(n.getDeterminant() == 0);
        CHECK(n.getTranspose()[0][1] == 4);

        //n.getInverse(); --> inverse does not exist, therefore EAR_CHECK_REL throws exception

        EAR::Matrix3_3 transformation({ { 1, 2, 3 }, { 4, 5, 6 }, { 7, 8, 9 } });

        auto transformedVector = transformation * EAR::Vector3(1, 1, 1);
        CHECK(transformedVector == EAR::Vector3(6, 15, 24));
    }

    SUBCASE("Quaternions")
    {
        std::vector<EAR::Euler> tests;

        const EAR::Vector3 orig(0, 1, 0);

        tests.emplace_back(90, 0, 0);
        tests.emplace_back(0, 90, 0); // north pole singularity
        tests.emplace_back(0, -90, 0); // south pole singularity
        tests.emplace_back(0, 0, 90);
        tests.emplace_back(179, -45, 7);

        for (const EAR::Euler& a : tests) {
            EAR::Quaternion b(a);
            EAR::Euler c = b.getEuler();

            CHECK(a[0] == doctest::Approx(c[0]));
            CHECK(a[1] == doctest::Approx(c[1]));
            CHECK(a[2] == doctest::Approx(c[2]));

            EAR::HomogenMatrix m;
            m.setRotation(a);
            m.setTranslation(EAR::Vector3(0, 0, 0));

            EAR::Vector3 rotaRef = m * orig;
            EAR::Vector3 rota = b.rotate(orig);

            CHECK(rotaRef[0] == doctest::Approx(rota[0]));
            CHECK(rotaRef[1] == doctest::Approx(rota[1]));
            CHECK(rotaRef[2] == doctest::Approx(rota[2]));
        }

        EAR::Quaternion q1(EAR::Euler(90, 0, 0));
        EAR::Quaternion q2(EAR::Euler(0, -45, 0));

        EAR::Quaternion q12 = q1 * q2;
        EAR::Quaternion q21 = q2 * q1;

        EAR::Vector3 rota = q12.rotate(orig);
        EAR::Vector3 rota2 = q1.rotate(orig);
        rota2 = q2.rotate(rota2);

        CHECK(rota2[0] == doctest::Approx(rota[0]));
        CHECK(rota2[1] == doctest::Approx(rota[1]));
        CHECK(rota2[2] == doctest::Approx(rota[2]));

        rota = q21.rotate(orig);
        rota2 = q2.rotate(orig);
        rota2 = q1.rotate(rota2);

        CHECK(rota2[0] == doctest::Approx(rota[0]));
        CHECK(rota2[1] == doctest::Approx(rota[1]));
        CHECK(rota2[2] == doctest::Approx(rota[2]));
    }
    SUBCASE("homogeneous inverse, transpose, determinate and adjugate")
    {

        int testNumber = 100;

        //check on determinant
        for (int i = 0; i < testNumber; i++) {
            float x = std::rand() % 100;
            float y = std::rand() % 100;
            float z = std::rand() % 100;
            float ex = std::rand() % 100;
            float ey = std::rand() % 100;
            float ez = std::rand() % 100;
            EAR::HomogenMatrix h(EAR::Position(EAR::Vector3(x, y, z), EAR::Euler(ex, ey, ez)));
            float det = h.det();
            //must be 1 because only translation, and rotation
            float error_metric = (1 - det) * (1 - det);
            CHECK(error_metric <= 0.05 * 0.05);
        }

        //check on transpose
        for (int i = 0; i < testNumber; i++) {
            float x = std::rand() % 100;
            float y = std::rand() % 100;
            float z = std::rand() % 100;
            float ex = std::rand() % 100;
            float ey = std::rand() % 100;
            float ez = std::rand() % 100;
            EAR::HomogenMatrix h(EAR::Position(EAR::Vector3(x, y, z), EAR::Euler(ex, ey, ez)));
            EAR::HomogenMatrix t = h.T();
            t = t.T();

            CHECK(t == h);
        }

        //check on adjugate
        for (int i = 0; i < testNumber; i++) {
            float x = std::rand() % 100;
            float y = std::rand() % 100;
            float z = std::rand() % 100;
            float ex = std::rand() % 100;
            float ey = std::rand() % 100;
            float ez = std::rand() % 100;
            EAR::HomogenMatrix h(EAR::Position(EAR::Vector3(x, y, z), EAR::Euler(ex, ey, ez)));
            x = std::rand() % 100;
            y = std::rand() % 100;
            z = std::rand() % 100;
            ex = std::rand() % 100;
            ey = std::rand() % 100;
            ez = std::rand() % 100;
            EAR::HomogenMatrix a(EAR::Position(EAR::Vector3(x, y, z), EAR::Euler(ex, ey, ez)));

            //test one regarding the adjugate
            EAR::HomogenMatrix adj = h.getAdjugate();
            EAR::HomogenMatrix id(EAR::Position(EAR::Vector3(0, 0, 0), EAR::Euler(0, 0, 0)));

            bool result = (adj * h == id * h.det());
            CHECK(result);

            //test two regarding the adjugate
            EAR::HomogenMatrix adj_1 = (h * a).getAdjugate();
            EAR::HomogenMatrix adj_2 = a.getAdjugate() * h.getAdjugate();
            result = (adj_1 == adj_2);
            CHECK(result);
        }

        //check algebraic inverse
        // EAR::TimeMeasure::startTimer();
        for (int i = 0; i < testNumber; i++) {
            float x = std::rand() % 100;
            float y = std::rand() % 100;
            float z = std::rand() % 100;
            float ex = std::rand() % 100;
            float ey = std::rand() % 100;
            float ez = std::rand() % 100;
            EAR::HomogenMatrix h(EAR::Position(EAR::Vector3(x, y, z), EAR::Euler(ex, ey, ez)));
            EAR::HomogenMatrix inv = h.inverse();
            EAR::HomogenMatrix id(EAR::Position(EAR::Vector3(0, 0, 0), EAR::Euler(0, 0, 0)));
            bool result = (id == h * inv);
            CHECK(result);
        }
        // EAR::TimeMeasure::stopTimer();
        // std::cout << "Algebraic inverse took on average " << EAR::TimeMeasure::getElapsedMicroseconds() / testNumber << " µs." << std::endl;

        //check geometric inverse
        // EAR::TimeMeasure::startTimer();
        for (int i = 0; i < testNumber; i++) {
            float x = std::rand() % 100;
            float y = std::rand() % 100;
            float z = std::rand() % 100;
            float ex = std::rand() % 100;
            float ey = std::rand() % 100;
            float ez = std::rand() % 100;
            EAR::HomogenMatrix h(EAR::Position(EAR::Vector3(x, y, z), EAR::Euler(ex, ey, ez)));
            EAR::HomogenMatrix inv = h.inverseGeometric();
            EAR::HomogenMatrix id(EAR::Position(EAR::Vector3(0, 0, 0), EAR::Euler(0, 0, 0)));
            bool result = (id == h * inv);
            CHECK(result);
        }
        // EAR::TimeMeasure::stopTimer();
        // std::cout << "Geometric inverse took on average " << EAR::TimeMeasure::getElapsedMicroseconds() / testNumber << " µs." << std::endl;
    }

    SUBCASE("Euler normal case")
    {
        const EAR::Euler first{ 10, 0, 0 };
        const EAR::Euler second{ 11, 0, 0 };

        CHECK(first.isClose(second, 2) == true);
        CHECK(first.angleBetween(second) == doctest::Approx(1.f).epsilon(0.01));
    }

    SUBCASE("Euler singularity")
    {
        // These two Euler representations represent the same rotation.
        const EAR::Euler first{ 45, 90, 45 };
        const EAR::Euler second{ 90, 90, 90 };

        CHECK(first.isClose(second, 0.01) == true);
        CHECK(first.angleBetween(second) == doctest::Approx(0.f).epsilon(0.01));
    }

    SUBCASE("Euler across 360 degree boundary")
    {
        const EAR::Euler first{ 1, 0, 0 };
        const EAR::Euler second{ 359, 0, 0 };

        CHECK(first.isClose(second, 3) == true);
        CHECK(first.angleBetween(second) == doctest::Approx(2.f).epsilon(0.01));
    }
}