/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/RealFrame.h>

TEST_CASE("class RealFrame")
{
    EAR::RealFrame dut(2, 8);
    float sig[] = { 1, -1, 0.5, 0.3333, -0.23546, 0.56, 0.1, -0.01 };
    float sig2[] = { -14.5, 13.3, 7.6, 0.275 };

    SUBCASE("get counts")
    {
        CHECK(dut.channels() == 2);
        CHECK(dut.length() == 8);
    }

    SUBCASE("write to channel")
    {
        CHECK(dut[0][7] == 0.f); // is zero initialized?
        dut[0].write(sig, 0, 4); // partly
        CHECK(dut[0][7] != sig[7]);
        dut[0].write(sig, 0, 8); // full
        CHECK(dut[0][7] == sig[7]);
        dut[0].write(sig, 2, 6); // offset write
        CHECK(dut[0][7] == sig[5]);
        CHECK(dut[0][1] == sig[1]);
    }

    SUBCASE("shifting")
    {
        dut[1].write(sig, 0, 8);
        dut[1].shiftLeft(2); // first two samples should be found at end
        CHECK(dut[1][0] == sig[2]);
        CHECK(dut[1][1] == sig[3]);
        CHECK(dut[1][2] == sig[4]);
        CHECK(dut[1][6] == sig[0]);
        CHECK(dut[1][7] == sig[1]);

        dut[1].write(sig, 0, 8);
        dut[1].shiftRight(2); // last two samples should be found at beginning
        CHECK(dut[1][0] == sig[6]);
        CHECK(dut[1][1] == sig[7]);
        CHECK(dut[1][2] == sig[0]);
        CHECK(dut[1][6] == sig[4]);
        CHECK(dut[1][7] == sig[5]);

        dut[0].write(sig2, 0, 4);
        dut[1].write(sig, 0, 4);
        dut.shiftUp(1); // channels should be switched
        CHECK(dut[0][0] == sig[0]);
        CHECK(dut[1][0] == sig2[0]);

        dut.shiftDown(1); // channels should be switched again
        CHECK(dut[1][0] == sig[0]);
        CHECK(dut[0][0] == sig2[0]);
    }

    SUBCASE("shifting and inserting")
    {
        dut[1].write(sig, 0, 8);
        dut[1].shiftLeftInsert(sig + 2, 2); // samples 3+4 should be found at end
        CHECK(dut[1][0] == sig[2]);
        CHECK(dut[1][1] == sig[3]);
        CHECK(dut[1][2] == sig[4]);
        CHECK(dut[1][6] == sig[2]);
        CHECK(dut[1][7] == sig[3]);

        dut[1].write(sig, 0, 8);
        dut[1].shiftRightInsert(sig + 2, 2); // last two samples should be found at beginning
        CHECK(dut[1][0] == sig[2]);
        CHECK(dut[1][1] == sig[3]);
        CHECK(dut[1][2] == sig[0]);
        CHECK(dut[1][6] == sig[4]);
        CHECK(dut[1][7] == sig[5]);
    }

    SUBCASE("clearing")
    {
        dut[0].write(sig, 0, 8);
        dut.zero(0, 2, 4); // middle 4 samples of first buffer should be cleared
        CHECK(dut[0][0] == sig[0]);
        CHECK(dut[0][2] == 0.f);
        CHECK(dut[0][5] == 0.f);
        CHECK(dut[0][6] == sig[6]);
    }

    SUBCASE("resize")
    {
        dut.resize(5, 123);
        CHECK(dut.channels() == 5);
        CHECK(dut.length() == 123);
    }

    SUBCASE("multiplication/addition")
    {
        dut[0].write(sig, 0, 8);
        dut[1].write(sig, 0, 8);
        dut[0].mul(2); // multiply first channel by 2
        dut[1].add(0.5); // add 0.5 to every sample in second channel
        CHECK(dut[0][0] == doctest::Approx(2 * sig[0]));
        CHECK(dut[1][0] == doctest::Approx(sig[0] + 0.5));
        CHECK(dut[0][5] == doctest::Approx(2 * sig[5]));
        CHECK(dut[1][6] == doctest::Approx(sig[6] + 0.5));
    }

    SUBCASE("min/max/peak/rms")
    {
        dut[0].write(sig, 0, 8);
        dut[1].write(sig2, 0, 4);

        CHECK(dut.mini() == doctest::Approx(-14.5));
        CHECK(dut.maxi() == doctest::Approx(13.3));
        CHECK(dut.peak() == doctest::Approx(14.5));

        // values from numpy.sqrt(numpy.mean(numpy.square(sig)))
        CHECK(dut[0].rms() == doctest::Approx(0.585259));
        CHECK(dut[1].rms() == doctest::Approx(7.458012));
    }

    SUBCASE("interleaved read/write")
    {
        // read from interleaved buffer with 4 channels and two samples per channel (total length 8)
        dut.writeInterleaved(sig, 4, 2);

        CHECK(dut.channels() == 4);
        CHECK(dut.length() == 2);
        CHECK(dut[0][0] == sig[0]);
        CHECK(dut[1][0] == sig[1]);
        CHECK(dut[2][1] == sig[6]);
        CHECK(dut[3][1] == sig[7]);

        // interleave into float array of length 8
        float res[8];
        dut.readInterleaved(res, 0, 4, 0, 2);
        CHECK(res[0] == sig[0]);
        CHECK(res[1] == sig[1]);
        CHECK(res[6] == sig[6]);
        CHECK(res[7] == sig[7]);

        // interleave only second sample of channels 2 & 3 into float array of length 2
        float res2[2];
        dut.readInterleaved(res2, 1, 2, 1, 1);
        CHECK(res2[0] == sig[5]);
        CHECK(res2[1] == sig[6]);
    }

    SUBCASE("write only to real or imag part")
    {
        EAR::ComplexBuffer dut2(4);
        dut2.writeReal(sig2, 0, 4);
        dut2.writeImag(sig2, 2, 2);

        CHECK(dut2[0].real() == sig2[0]);
        CHECK(dut2[3].real() == sig2[3]);
        CHECK(dut2[0].imag() == 0.f);
        CHECK(dut2[3].imag() == sig2[1]);
    }
}
