/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/CartesianKDTree.h>

namespace detail {
template <typename T> bool isValueInListOfValues(const T& value, const std::vector<T>& listOfValues)
{
    for (const auto& valuesInList : listOfValues) {
        if (value == valuesInList) {
            return true;
        }
    }

    return false;
}
} // namespace detail

TEST_CASE("Find points in a point cloud through a KDTree")
{
    SUBCASE("Points that are used on construction are the same that are returned")
    {
        std::vector<EAR::Vector3> points{ EAR::Vector3(0.f, 0.f, 0.f),
            EAR::Vector3(1.5f, 1.5f, 1.5f), EAR::Vector3(-2.3f, -2.3f, -2.3f) };
        EAR::CartesianKDTree kdtree(points);

        auto returnedPoints = kdtree.getPoints();

        CHECK(points == returnedPoints);
    }

    SUBCASE("Get Nearest Point")
    {
        EAR::Vector3 pointOne(0.f, 0.f, 0.f);
        EAR::Vector3 pointTwo(1.5f, 1.5f, 1.5f);
        EAR::Vector3 pointThree(-2.3f, -2.3f, -2.3f);
        std::vector<EAR::Vector3> points{ pointOne, pointTwo, pointThree };
        EAR::CartesianKDTree kdtree(points);

        auto returnedPointOne = kdtree.getNearestPoint(EAR::Vector3(0.1f, -0.1f, 0.05f));
        auto returnedPointTwo = kdtree.getNearestPoint(EAR::Vector3(1.4f, 1.6f, 1.45f));
        auto returnedPointThree = kdtree.getNearestPoint(EAR::Vector3(-2.4f, -2.3f, -2.25f));

        CHECK(pointOne == returnedPointOne);
        CHECK(pointTwo == returnedPointTwo);
        CHECK(pointThree == returnedPointThree);
    }

    SUBCASE("Get Nearest Point with only one point")
    {
        EAR::Vector3 point(0.f, 0.f, 0.f);
        std::vector<EAR::Vector3> points{ point };
        EAR::CartesianKDTree kdtree(points);

        auto returnedPoint =
            kdtree.getNearestPoint(EAR::Vector3(20395823049.1f, -02394558.1f, 45789409.05f));

        CHECK(point == returnedPoint);
    }

    SUBCASE("Get Nearest Index")
    {
        EAR::Vector3 pointOne(0.f, 0.f, 0.f);
        EAR::Vector3 pointTwo(1.5f, 1.5f, 1.5f);
        EAR::Vector3 pointThree(-2.3f, -2.3f, -2.3f);
        std::vector<EAR::Vector3> points{ pointOne, pointTwo, pointThree };
        EAR::CartesianKDTree kdtree(points);

        auto returnedIndexOne = kdtree.getNearestIndex(EAR::Vector3(0.1f, -0.1f, 0.05f));
        auto returnedIndexTwo = kdtree.getNearestIndex(EAR::Vector3(1.4f, 1.6f, 1.45f));
        auto returnedIndexThree = kdtree.getNearestIndex(EAR::Vector3(-2.4f, -2.3f, -2.25f));

        CHECK(points[0] == points[returnedIndexOne]);
        CHECK(points[1] == points[returnedIndexTwo]);
        CHECK(points[2] == points[returnedIndexThree]);
    }

    SUBCASE("Get Nearest Points within a given radius")
    {
        EAR::Vector3 pointOne(0.f, 0.f, 0.f);
        EAR::Vector3 pointTwo(1.5f, 1.5f, 1.5f);
        EAR::Vector3 pointThree(-10.3f, -10.3f, -10.3f);
        std::vector<EAR::Vector3> points{ pointOne, pointTwo, pointThree };
        EAR::CartesianKDTree kdtree(points);

        auto returnedPoints = kdtree.getAllPointsWithinRadius(EAR::Vector3(0.75f, 0.75f, 0.75f), 2);

        CHECK(detail::isValueInListOfValues<EAR::Vector3>(pointOne, returnedPoints) == true);
        CHECK(detail::isValueInListOfValues<EAR::Vector3>(pointTwo, returnedPoints) == true);
        CHECK(detail::isValueInListOfValues<EAR::Vector3>(pointThree, returnedPoints) == false);
    }

    SUBCASE("Get Nearest indices within a given radius")
    {
        EAR::Vector3 pointOne(1.5f, 1.5f, 1.5f);
        EAR::Vector3 pointTwo(-10.3f, -10.3f, -10.3f);
        EAR::Vector3 pointThree(0.f, 0.f, 0.f);
        std::vector<EAR::Vector3> points{ pointOne, pointTwo, pointThree };
        EAR::CartesianKDTree kdtree(points);

        auto returnedIndices =
            kdtree.getAllIndicesWithinRadius(EAR::Vector3(0.75f, 0.75f, 0.75f), 2);

        CHECK(detail::isValueInListOfValues<std::size_t>(0, returnedIndices) == true);
        CHECK(detail::isValueInListOfValues<std::size_t>(1, returnedIndices) == false);
        CHECK(detail::isValueInListOfValues<std::size_t>(2, returnedIndices) == true);
    }

    SUBCASE("Break on no points")
    {
        std::vector<EAR::Vector3> points;
        bool exceptionRaised = false;

        try {
            EAR::CartesianKDTree kdtree(points);
        } catch (const EAR::Exception& e) {
            exceptionRaised = true;
        }

        CHECK(exceptionRaised == true);
    }
}
