/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/EmbreeTracer.h>
#include <EAR/Geometry.h>

#include <iostream>

using namespace EAR;

TEST_CASE("EmbreeTracer test")
{

    SUBCASE("occlusion by mesh cube with outward facing normals")
    {
        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        int a = mesh->addVertex(EAR::Vector3(1, 1, 1));
        int b = mesh->addVertex(EAR::Vector3(1, 1, -1));
        int c = mesh->addVertex(EAR::Vector3(1, -1, 1));
        int d = mesh->addVertex(EAR::Vector3(-1, 1, 1));
        int e = mesh->addVertex(EAR::Vector3(1, -1, -1));
        int f = mesh->addVertex(EAR::Vector3(-1, -1, 1));
        int g = mesh->addVertex(EAR::Vector3(-1, 1, -1));
        int h = mesh->addVertex(EAR::Vector3(-1, -1, -1));

        mesh->addTriangle(c, b, a);
        mesh->addTriangle(e, b, c);
        mesh->addTriangle(c, a, d);
        mesh->addTriangle(d, f, c);
        mesh->addTriangle(a, b, d);
        mesh->addTriangle(b, g, d);
        mesh->addTriangle(f, h, c);
        mesh->addTriangle(h, e, c);
        mesh->addTriangle(f, d, h);
        mesh->addTriangle(d, g, h);
        mesh->addTriangle(g, b, e);
        mesh->addTriangle(h, g, e);

        mesh->lock();

        EmbreeTracer tracer(mesh.get());

        // points are on opposite sides of the cube -> should be occluded
        CHECK(tracer.isOccluded(EAR::Vector3(0, 0, -2), EAR::Vector3(0, 0, 2)));
        // points are both on the same side of the cube -> should not be occluded
        CHECK(!tracer.isOccluded(EAR::Vector3(0, 0, -4), EAR::Vector3(0, 0, -1.5)));

        // shoot in direction of cube -> should hit the mesh
        CHECK(tracer.firstHit(EAR::Vector3(0, 0, -2), EAR::Polar(180, 0, 1)) == mesh.get());
        // shoot sideways -> should not hit any mesh
        CHECK(tracer.firstHit(EAR::Vector3(0, 0, -2), EAR::Polar(90, 0, 1)) == nullptr);

        // both points are on the same side of the cube -> should not be occluded by the mesh
        CHECK(!tracer.occludedByGivenMesh(
            EAR::Vector3(0, 0, -2), EAR::Vector3(2, 0, -2), mesh.get()));
        // points are on opposite sides of the cube -> should be occluded by the mesh
        CHECK(
            tracer.occludedByGivenMesh(EAR::Vector3(0, 0, -2), EAR::Vector3(0, 0, 2), mesh.get()));
        // target point is inside the cube -> should be occluded by the mesh
        CHECK(tracer.occludedByGivenMesh(EAR::Vector3(0, 0, 4), EAR::Vector3(0, 0, 0), mesh.get()));
        // both points are on opposite sides of the cube (source far away) -> should be occluded by the mesh
        CHECK(
            tracer.occludedByGivenMesh(EAR::Vector3(0, 0, -20), EAR::Vector3(0, 0, 2), mesh.get()));

        CHECK(tracer.getDistanceToLastHit() == doctest::Approx(19));
    }

    SUBCASE("backface hit turned on or off")
    {
        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        int a = mesh->addVertex(EAR::Vector3(1, 1, 0));
        int b = mesh->addVertex(EAR::Vector3(1, -1, 0));
        int c = mesh->addVertex(EAR::Vector3(-1, -1, 0));
        int d = mesh->addVertex(EAR::Vector3(-1, 1, 0));

        mesh->addTriangle(d, a, c);
        mesh->addTriangle(a, b, c);

        mesh->lock();

        EmbreeTracer tracer(
            mesh.get(), true, true); // second argument sets tracer to ignore backfaces

        // normals pointing towards the source (away from the target) -> should be occluded
        CHECK(tracer.isOccluded(EAR::Vector3(0, 0, -2), EAR::Vector3(0, 0, 2)));
        // normals pointing away from the source (towards the target) -> should not be occluded
        CHECK(!tracer.isOccluded(EAR::Vector3(0, 0, 2), EAR::Vector3(0, 0, -2)));

        // shoot against the direction of normals -> should hit the mesh
        CHECK(tracer.firstHit(EAR::Vector3(0, 0, -2), EAR::Polar(180, 0, 1)) == mesh.get());
        // shoot in direction of normals -> should not hit the mesh
        CHECK(tracer.firstHit(EAR::Vector3(0, 0, 2), EAR::Polar(0, 0, 1)) == nullptr);

        EmbreeTracer tracer2(mesh.get(), true,
            false); // second argument sets tracer to consider backfaces as normal hits

        // normals pointing towards the source (away from the target) -> should be occluded
        CHECK(tracer2.isOccluded(EAR::Vector3(0, 0, -2), EAR::Vector3(0, 0, 2)));
        // normals pointing away from the source (towards the target) -> should be occluded
        CHECK(tracer2.isOccluded(EAR::Vector3(0, 0, 2), EAR::Vector3(0, 0, -2)));

        // shoot against the direction of normals -> should hit the mesh
        CHECK(tracer2.firstHit(EAR::Vector3(0, 0, -2), EAR::Polar(180, 0, 1)) == mesh.get());
        // shoot in direction of normals -> should hit the mesh
        CHECK(tracer2.firstHit(EAR::Vector3(0, 0, 2), EAR::Polar(0, 0, 1)) == mesh.get());
    }
}
