/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/FFT.h>
#include <EAR/config.h>

#include <complex>
#include <iostream>

TEST_CASE("Real-data FFT")
{

    SUBCASE("Length 4 forward")
    {
        const float in[] = { 1, 2, 3, 4 };
        const std::complex<float> ref[] = { { 10, 0 }, { -2, 2 }, { -2, 0 } };

        auto fft = EAR::RFFT::create(4);
        auto out = new std::complex<float>[3];

        fft->transform(in, out);

        CHECK(out[0] == ref[0]);
        CHECK(out[1] == ref[1]);
        CHECK(out[2] == ref[2]);
    }

    SUBCASE("Length 4 backward")
    {
        const std::complex<float> in[] = { { 10, 0 }, { -2, 2 }, { -2, 0 } };
        const float ref[] = { 1, 2, 3, 4 };

        auto ifft = EAR::IRFFT::create(4);
        auto out = new float[4];

        ifft->transform(in, out);

        // not normalized -> needs to be divided by fft length 4
        CHECK(out[0] / 4 == doctest::Approx(ref[0]));
        CHECK(out[1] / 4 == doctest::Approx(ref[1]));
        CHECK(out[2] / 4 == doctest::Approx(ref[2]));
        CHECK(out[3] / 4 == doctest::Approx(ref[3]));
    }

    SUBCASE("Center Frequencies (length 9)")
    {
        float ref[9] = { 0.00f, 3000.00f, 6000.00f, 9000.00f, 12000.00f, 15000.00f, 18000.00f,
            21000.00f, 24000.00f };
        auto test = EAR::RFFT::binFrequencies(9, 48000.f);

        for (int i = 0; i < 9; i++) {
            CHECK(test[i] == doctest::Approx(ref[i]));
        }
    }

    SUBCASE("Length 4 forward C2C")
    {
        const std::complex<float> in[] = { { 1, 1 }, { 2, 2 }, { 3, 3 }, { 4, 4 } };
        const std::complex<float> ref[] = { { 10, 10 }, { -4, 0 }, { -2, -2 }, { 0, -4 } };

        auto fft = EAR::FFT::create(4);
        auto out = new std::complex<float>[4];

        fft->transform(in, out);

        CHECK(out[0] == ref[0]);
        CHECK(out[1] == ref[1]);
        CHECK(out[2] == ref[2]);
        CHECK(out[3] == ref[3]);
    }

    SUBCASE("Length 4 forward-backward C2C")
    {
        const std::complex<float> in[] = { { 1, 1 }, { 2, 2 }, { 3, 3 }, { 4, 4 } };
        const std::complex<float> ref[] = { { 1, 1 }, { 2, 2 }, { 3, 3 }, { 4, 4 } };

        auto fft = EAR::FFT::create(4);
        auto ifft = EAR::IFFT::create(4);
        auto Out = new std::complex<float>[4];
        auto out = new std::complex<float>[4];

        fft->transform(in, Out);
        ifft->transform(Out, out);

        out[0] /= 4.f;
        out[1] /= 4.f;
        out[2] /= 4.f;
        out[3] /= 4.f;

        // not normalized -> needs to be divided by fft length 4
        CHECK(out[0] == ref[0]);
        CHECK(out[1] == ref[1]);
        CHECK(out[2] == ref[2]);
        CHECK(out[3] == ref[3]);
    }

    SUBCASE("Length 4 backward C2C")
    {
        const std::complex<float> in[] = { { 10, 10 }, { -4, 0 }, { -2, -2 }, { 0, -4 } };
        const std::complex<float> ref[] = { { 1, 1 }, { 2, 2 }, { 3, 3 }, { 4, 4 } };

        auto ifft = EAR::IFFT::create(4);
        auto out = new std::complex<float>[4];

        ifft->transform(in, out);

        // not normalized -> needs to be divided by fft length 4
        CHECK(out[0] / 4.f == ref[0]);
        CHECK(out[1] / 4.f == ref[1]);
        CHECK(out[2] / 4.f == ref[2]);
        CHECK(out[3] / 4.f == ref[3]);
    }

    SUBCASE("C2C Center Frequencies (length 9)")
    {
        float ref[9] = { 0.f, 5333.33f, 10666.67f, 16000.f, 21333.33f, -21333.33f, -16000.f,
            -10666.67f, -5333.33f };
        auto test = EAR::FFT::binFrequencies(9, 48000.f);

        for (int i = 0; i < 9; i++) {
            CHECK(test[i] == doctest::Approx(ref[i]));
        }
    }

    SUBCASE("C2C Center Frequencies (length 8)")
    {
        float ref[8] = { 0.f, 6000.f, 12000.f, 18000.f, -24000.f, -18000.f, -12000.f, -6000.f };
        auto test = EAR::FFT::binFrequencies(8, 48000.f);

        for (int i = 0; i < 8; i++) {
            CHECK(test[i] == doctest::Approx(ref[i]));
        }
    }
}
