/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#include <doctest/doctest.h>

#include "DataFile.h"
#include "EAR/FIR2MinphaseFilterDesigner.h"
#include <EAR/FFT.h>
#include <EAR/FIRFilter.h>
#include <EAR/HeadphoneEQ.h>
#include <EAR/RealFrame.h>

#define WRITE_TO_FILE false

TEST_CASE("FIR2MinphaseFilterDesigner test")
{

    SUBCASE("Piecelin with extrapolation")
    {
        const float xf[] = { 1, 2, 3, 4 };
        const float yf[] = { 2, 3, 4, 5 };
        const float uf[] = { 0, 1, 2, 3, 4, 5 };
        const float ef[] = { 1, 2, 3, 4, 5, 6 };
        EAR::RealBuffer x = EAR::RealBuffer(4);
        EAR::RealBuffer y = EAR::RealBuffer(4);
        EAR::RealBuffer u = EAR::RealBuffer(6);
        EAR::RealBuffer v = EAR::RealBuffer(6);
        EAR::RealBuffer e = EAR::RealBuffer(6);

        x.write(xf, 0, 4);
        y.write(yf, 0, 4);
        u.write(uf, 0, 6);
        EAR::piecelin(x, y, u, v);

        e.write(ef, 0, 6);
        for (int n = 0; n < 6; n++) {
            CHECK(v[n] == doctest::Approx(e[n]));
        }
    }

    SUBCASE("Piecelin with interpolation")
    {
        const float xf[] = { 1, 4 };
        const float yf[] = { 2, 5 };
        const float uf[] = { 0, 1, 2, 3, 4, 5 };
        const float ef[] = { 1, 2, 3, 4, 5, 6 };
        EAR::RealBuffer x = EAR::RealBuffer(2);
        EAR::RealBuffer y = EAR::RealBuffer(2);
        EAR::RealBuffer u = EAR::RealBuffer(6);
        EAR::RealBuffer v = EAR::RealBuffer(6);
        EAR::RealBuffer e = EAR::RealBuffer(6);

        x.write(xf, 0, 2);
        y.write(yf, 0, 2);
        u.write(uf, 0, 6);
        EAR::piecelin(x, y, u, v);

        e.write(ef, 0, 6);
        for (int n = 0; n < 6; n++) {
            CHECK(v[n] == doctest::Approx(e[n]));
        }
    }

    SUBCASE("Minimum Phase even length")
    {
        const float y[] = { 0, 0, 1, 0, 0, 0, 0, 0 };
        EAR::MinimumPhase minPhase = EAR::MinimumPhase(8);
        EAR::RealBuffer in = EAR::RealBuffer(8);
        EAR::RealBuffer out = EAR::RealBuffer(8);

        in.write(y, 0, 8);
        minPhase.transform(in, out);

        const float ref[] = { 1, 0, 0, 0, 0, 0, 0, 0 };

        for (int n = 0; n < 8; n++) {
            CHECK(out[n] == doctest::Approx(ref[n]));
        }
    }

    SUBCASE("Minimum Phase odd length")
    {
        const float y[] = { 0, 0, 1, 0, 0, 0, 0 };
        EAR::MinimumPhase minPhase = EAR::MinimumPhase(7);
        EAR::RealBuffer in = EAR::RealBuffer(7);
        EAR::RealBuffer out = EAR::RealBuffer(7);

        in.write(y, 0, 7);
        minPhase.transform(in, out);

        const float ref[] = { 1, 0, 0, 0, 0, 0, 0 };

        for (int n = 0; n < 7; n++) {
            CHECK(out[n] == doctest::Approx(ref[n]));
        }
    }

    SUBCASE("Hilbert even length")
    {
        const float y[] = { 1, 2, 3, 4, 5, 6, 7, 8 };
        EAR::Hilbert hilbert = EAR::Hilbert(8);
        EAR::RealBuffer b = EAR::RealBuffer(8);
        EAR::ComplexBuffer h = EAR::ComplexBuffer(8);

        b.write(y, 0, 8);
        hilbert.transform(b, h);

        const std::complex<float> ref[] = {
            { 1, 3.828427124746190 },
            { 2, -1 },
            { 3, -1 },
            { 4, -1.828427124746190 },
            { 5, -1.828427124746190 },
            { 6, -1 },
            { 7, -1 },
            { 8, 3.828427124746190 },
        };

        for (int n = 0; n < 8; n++) {
            CHECK(h[n].real() == doctest::Approx(ref[n].real()));
            CHECK(h[n].imag() == doctest::Approx(ref[n].imag()));
        }
    }

    SUBCASE("Hilbert odd length")
    {
        const float y[] = { 1, 2, 3, 4, 5, 6, 7 };
        EAR::Hilbert hilbert = EAR::Hilbert(7);
        EAR::RealBuffer b = EAR::RealBuffer(7);
        EAR::ComplexBuffer h = EAR::ComplexBuffer(7);

        b.write(y, 0, 8);
        hilbert.transform(b, h);

        const std::complex<float> ref[] = {
            { 1, 3.102238259844891 },
            { 2, -1.279048007689934 },
            { 3, -0.797473388882403 },
            { 4, -2.051433726545107 },
            { 5, -0.797473388882403 },
            { 6, -1.279048007689934 },
            { 7, 3.102238259844891 },
        };

        for (int n = 0; n < 7; n++) {
            CHECK(h[n].real() == doctest::Approx(ref[n].real()));
            CHECK(h[n].imag() == doctest::Approx(ref[n].imag()));
        }
    }

    SUBCASE("IFFT normalization")
    {
        // flat spectrum
        int Nfreqs = 8;
        EAR::ComplexBuffer H = EAR::ComplexBuffer(Nfreqs); // frequency domain
        EAR::RealBuffer h = EAR::RealBuffer(Nfreqs); // time domain
        const float real[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0 };
        const float imag[] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
        H.writeReal(real, 0, 8);
        H.writeImag(imag, 0, 8);

        // inverse FFT
        std::unique_ptr<EAR::IRFFT> ifft = EAR::IRFFT::create(Nfreqs);
        ifft->transform(H.data(), h.data());

        CHECK(h[0] == doctest::Approx(Nfreqs));
        for (int n = 1; n < Nfreqs; n++) {
            CHECK(h[n] == doctest::Approx(0.0));
        }
    }

    SUBCASE("Hamming even")
    {
        // ones
        int Nfreqs = 8;
        EAR::RealBuffer h = EAR::RealBuffer(Nfreqs);
        h.fill(1.0);

        // apply hamming
        EAR::FIR2MinphaseFilterDesigner(Nfreqs, 1.0).applyHamming(h);

        // check results
        const float ref[] = { 0.08, 0.25319, 0.64236, 0.95445, 0.95445, 0.64236, 0.25319, 0.08 };
        for (int n = 0; n < Nfreqs; n++) {
            CHECK(h[n] == doctest::Approx(ref[n]).epsilon(0.0001));
        }
    }

    SUBCASE("Hamming odd")
    {
        // flat spectrum
        int Nfreqs = 7;
        EAR::RealBuffer h = EAR::RealBuffer(Nfreqs);
        h.fill(1.0);

        // apply hamming
        EAR::FIR2MinphaseFilterDesigner(Nfreqs, 1.0).applyHamming(h);

        // check results
        const float ref[] = { 0.08, 0.31, 0.77, 1.0, 0.77, 0.31, 0.08 };
        for (int n = 0; n < Nfreqs; n++) {
            CHECK(h[n] == doctest::Approx(ref[n]).epsilon(0.0001));
        }
    }

    SUBCASE("Low pass even")
    {
        const float fs = 48000.0;
        const size_t Ntaps = 8;
        const int Nfreqs = 4;
        const float f[] = { 0, 8000, 8000, fs / 2 }; // frequencies
        const float k[] = { 1, 1, 0, 0 }; // gains

        EAR::RealBuffer frq = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer gains = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

        frq.write(f, 0, Nfreqs);
        gains.write(k, 0, Nfreqs);

        EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
        fdsgn.design(frq, gains, b);

        // check results
        const float ref[] = { 0.081864, 0.24241, 0.30181, 0.18615, 0.050553, 0.00013333, -0.0014930,
            -0.00083635 };

        for (int n = 0; n < Ntaps; n++) {
            CHECK(b[n] == doctest::Approx(ref[n]).epsilon(0.0001));
        }
    }

    SUBCASE("Low pass odd")
    {
        const float fs = 48000.0;
        const size_t Ntaps = 7;
        const int Nfreqs = 4;
        const float f[] = { 0, 8000, 8000, fs / 2 }; // frequencies
        const float k[] = { 1, 1, 0, 0 }; // gains

        EAR::RealBuffer frq = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer gains = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

        frq.write(f, 0, Nfreqs);
        gains.write(k, 0, Nfreqs);

        EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
        fdsgn.design(frq, gains, b);

        // check results
        const float ref[] = { 0.092200, 0.24713, 0.29025, 0.15701, 0.032430, -0.00020076,
            0.0029442 };
        for (int n = 0; n < Ntaps; n++) {
            CHECK(b[n] == doctest::Approx(ref[n]).epsilon(0.0001));
        }
    }

    SUBCASE("High pass even")
    {
        const float fs = 48000.0;
        const size_t Ntaps = 8;
        const int Nfreqs = 4;
        const float f[] = { 0, 8000, 8000, fs / 2 }; // frequencies
        const float k[] = { 0, 0, 1, 1 }; // gains

        EAR::RealBuffer frq = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer gains = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

        frq.write(f, 0, Nfreqs);
        gains.write(k, 0, Nfreqs);

        EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
        fdsgn.design(frq, gains, b);

        // check results
        const float ref[] = { 0.54754, -0.48131, -0.036006, 0.049012, 0.013638, -0.0020800,
            -0.0017016, -0.0020654 };
        for (int n = 0; n < Ntaps; n++) {
            CHECK(b[n] == doctest::Approx(ref[n]).epsilon(0.0001));
        }
    }

    SUBCASE("High pass")
    {
        const float fs = 48000.0;
        const size_t Ntaps = 64;
        const int Nfreqs = 4;
        const float f[] = { 0, 8000, 8000, fs / 2 }; // frequencies
        const float k[] = { 0, 0, 1, 1 }; // gains

        EAR::RealBuffer frq = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer gains = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

        frq.write(f, 0, Nfreqs);
        gains.write(k, 0, Nfreqs);

        EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
        fdsgn.design(frq, gains, b);

        const float ref[] = { 0.11804, -0.43001, 0.53675, -0.099682, -0.26126, 0.012213, 0.17641,
            0.070024, -0.093286, -0.10836, 0.0038839, 0.089022, 0.059910, -0.028859, -0.071285,
            -0.031210, 0.034896, 0.054163, 0.014753, -0.033392, -0.040133, -0.0055538, 0.028957,
            0.029168, 0.00068781, -0.023612, -0.020778, 0.0015688, 0.018310, 0.014483, -0.0022650,
            -0.013526, -0.0098146, 0.0017282, 0.0098205, 0.0070090, -0.0018021, -0.0067817,
            -0.0043983, 0.0012227, 0.0042652, 0.0028258, -0.00059579, -0.0025465, -0.0018333,
            0.00016029, 0.0014197, 0.0011598, 7.4362e-05, 0.00071027, -0.00068507, -0.00015195,
            0.00030018, 0.00035945, 0.00012999, 0.00010427, -0.00016089, -5.1358e-05, 0.00010443,
            0.00019978, 0.00019734, 0.00011830, 1.0969e-05, -7.3001e-05 };
        // TODO: epsilon would ideally be 0.0001, but the IPP IFFT implementation used by the filter designer
        // gives slightly different values depending on build platform (AMD vs Intel / Windows vs Linux etc)
        for (int n = 0; n < Ntaps; n++) {
            CHECK(b[n] == doctest::Approx(ref[n]).epsilon(0.01));
        }

#if (WRITE_TO_FILE)
        DataFile::write("C:\\Trash\\fir2design_hp.dat", b, fs);
#endif
    }

    SUBCASE("Headphone 1 to headphone 2")
    {
        const float fs = 48000.0;
        const size_t Ntaps = 512;
        const int Nfreqs = 4;
        const float f1[] = { 0, 4000, 8000, fs / 2 }; // frequencies
        const float k1[] = { 0.5, 0.5, 0.5, 0.5 }; // gains
        const float f2[] = { 0, 4000, 8000, fs / 2 }; // frequencies
        const float k2[] = { 1, 1, 1, 1 }; // gains

        EAR::RealBuffer frq1 = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer gains1 = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer frq2 = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer gains2 = EAR::RealBuffer(Nfreqs);
        EAR::RealBuffer b = EAR::RealBuffer(Ntaps);

        frq1.write(f1, 0, Nfreqs);
        gains1.write(k1, 0, Nfreqs);
        frq2.write(f2, 0, Nfreqs);
        gains2.write(k2, 0, Nfreqs);

        EAR::FIR2MinphaseFilterDesigner fdsgn = EAR::FIR2MinphaseFilterDesigner(Ntaps, fs);
        fdsgn.design(frq1, gains1, frq2, gains2, b);

        b[0] == doctest::Approx(2.0).epsilon(0.0001);
        for (int n = 1; n < Ntaps; n++) {
            CHECK(b[n] == doctest::Approx(0.0).epsilon(0.0001));
        }

#if (WRITE_TO_FILE)
        DataFile::write("C:\\Trash\\fir2design_hp1_to_hp2.dat", b, fs);
#endif
    }
}
