/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/AnchorObject.h>
#include <EAR/Geometry.h>
#include <EAR/SceneObjectFactory.h>
#include <math.h>

#include <iostream>

TEST_CASE("Points within geometry")
{

    SUBCASE("Point within cube")
    {
        auto cube = EAR::SceneObjectFactory<EAR::Box>::getNewObject(5.f, 5.f, 5.f);

        cube->setLocation(EAR::Vector3(10.f, 10.f, 10.f));

        EAR::Vector3 centre(10.f, 10.f, 10.f);
        EAR::Vector3 botLeft(7.5f, 7.5f, 7.5f);
        EAR::Vector3 topRight(12.5f, 12.5f, 12.5f);

        CHECK(cube->isInside(centre));
        CHECK(!cube->isInside(botLeft));
        CHECK(!cube->isInside(topRight));

        EAR::Vector3 above(10.f, 13.f, 10.f);
        EAR::Vector3 below(10.f, 7.f, 10.f);
        EAR::Vector3 sceneOrigin(0.f, 0.f, 0.f);

        CHECK(!cube->isInside(above));
        CHECK(!cube->isInside(below));
        CHECK(!cube->isInside(sceneOrigin));

        cube->setOrientation(EAR::Euler(45.f, 0.f, 0.f));

        EAR::Vector3 outsideRotation(10.f - (5.f * 1.42f), 10.f, 10.f);
        EAR::Vector3 insideRotation(10.f - (5.f * 1.4f), 10.f, 10.f);

        CHECK(!cube->isInside(outsideRotation));
        CHECK(!cube->isInside(insideRotation));

        CHECK(cube->getCenterPosition() == centre);
    }

    SUBCASE("Point within sphere")
    {
        auto sphere = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(4.f);

        sphere->setLocation(EAR::Vector3(8.f, 8.f, -8.f));

        EAR::Vector3 centre(8.f, 8.f, -8.f);
        EAR::Vector3 top(8.f, 12.f, -8.f);
        EAR::Vector3 bot(8.f, 4.f, -8.f);
        EAR::Vector3 left(4.f, 8.f, -8.f);
        EAR::Vector3 right(12.f, 8.f, -8.f);
        EAR::Vector3 front(8.f, 8.f, -4.f);
        EAR::Vector3 back(8.f, 8.f, -12.f);

        EAR::Vector3 justOutsideTop(8.f, 13.f, -8.f);
        EAR::Vector3 justOutsideBot(8.f, 3.f, -8.f);
        EAR::Vector3 justOutsideLeft(3.f, 8.f, -8.f);
        EAR::Vector3 justOutsideRight(13.f, 8.f, -8.f);
        EAR::Vector3 justOutsideFront(8.f, 8.f, -3.f);
        EAR::Vector3 justOutsideBack(8.f, 8.f, -13.f);

        CHECK(sphere->isInside(centre));
        CHECK(!sphere->isInside(top));
        CHECK(!sphere->isInside(bot));
        CHECK(!sphere->isInside(left));
        CHECK(!sphere->isInside(right));
        CHECK(!sphere->isInside(front));
        CHECK(!sphere->isInside(back));

        CHECK(!sphere->isInside(justOutsideTop));
        CHECK(!sphere->isInside(justOutsideBot));
        CHECK(!sphere->isInside(justOutsideLeft));
        CHECK(!sphere->isInside(justOutsideRight));
        CHECK(!sphere->isInside(justOutsideFront));
        CHECK(!sphere->isInside(justOutsideBack));

        CHECK(sphere->getCenterPosition() == centre);
    }
    SUBCASE("Point within cylinder ellipsoid")
    {

        //tube with ellipsoid as plate (major axis = 3, minor axis = 2)
        auto tube = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(3.f, 2.f, 4.f);

        tube->setOrientation(EAR::Euler(0, -90.f, 0));
        tube->setLocation(EAR::Vector3(0, 2.f, -2.f));

        //inside
        EAR::Vector3 centre(0.f, 2.f, -2.f);

        //points on the surface --> outside according to definition
        EAR::Vector3 left(0.f, 4.f, -2.f);
        EAR::Vector3 right(0.f, 0.f, -2.f);
        EAR::Vector3 bot(-3.f, 2.f, -2.f);
        EAR::Vector3 top(3.f, 2.f, -2.f);
        EAR::Vector3 front(0.f, 2.f, -4.f);
        EAR::Vector3 back(0.f, 2.f, 0.f);

        //points outside
        EAR::Vector3 justOutsideTop(0.f, 5.f, -2.f);
        EAR::Vector3 justOutsideBot(0.f, -1.f, -2.f);
        EAR::Vector3 justOutsideLeft(-4.f, 2.f, -2.f);
        EAR::Vector3 justOutsideRight(4.f, 2.f, -2.f);
        EAR::Vector3 justOutsideFront(0.f, 2.f, -5.f);
        EAR::Vector3 justOutsideBack(0.f, 2.f, 1.f);

        CHECK(tube->isInside(centre));

        CHECK(!tube->isInside(top));
        CHECK(!tube->isInside(front));
        CHECK(!tube->isInside(back));
        CHECK(!tube->isInside(bot));
        CHECK(!tube->isInside(left));
        CHECK(!tube->isInside(right));

        CHECK(!tube->isInside(justOutsideTop));
        CHECK(!tube->isInside(justOutsideBot));
        CHECK(!tube->isInside(justOutsideLeft));
        CHECK(!tube->isInside(justOutsideRight));
        CHECK(!tube->isInside(justOutsideFront));
        CHECK(!tube->isInside(justOutsideBack));

        CHECK(tube->getCenterPosition() == centre);
    }

    SUBCASE("Point within cylinder")
    {
        auto cylinder = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(2.f, 2.f, 4.f);

        cylinder->setOrientation(EAR::Euler(0.f, -90.f, 0.f)); // y,x,z
        cylinder->setLocation(EAR::Vector3(0.f, 2.f, -2.f));

        //is inside
        EAR::Vector3 centre(0.f, 2.f, -2.f);

        //from now on, outside
        EAR::Vector3 left(0.f, 4.f, -2.f);
        EAR::Vector3 right(0.f, 0.f, -2.f);
        EAR::Vector3 bot(-2.f, 2.f, -2.f);
        EAR::Vector3 top(2.f, 2.f, -2.f);
        EAR::Vector3 front(0.f, 2.f, -4.f);
        EAR::Vector3 back(0.f, 2.f, 0.f);

        EAR::Vector3 justOutsideTop(0.f, 5.f, -2.f);
        EAR::Vector3 justOutsideBot(0.f, -1.f, -2.f);
        EAR::Vector3 justOutsideLeft(-3.f, 2.f, -2.f);
        EAR::Vector3 justOutsideRight(3.f, 2.f, -2.f);
        EAR::Vector3 justOutsideFront(0.f, 2.f, -5.f);
        EAR::Vector3 justOutsideBack(0.f, 2.f, 1.f);

        CHECK(cylinder->isInside(centre));

        CHECK(!cylinder->isInside(top));
        CHECK(!cylinder->isInside(bot));
        CHECK(!cylinder->isInside(left));
        CHECK(!cylinder->isInside(right));
        CHECK(!cylinder->isInside(front));
        CHECK(!cylinder->isInside(back));

        CHECK(!cylinder->isInside(justOutsideTop));
        CHECK(!cylinder->isInside(justOutsideBot));
        CHECK(!cylinder->isInside(justOutsideLeft));
        CHECK(!cylinder->isInside(justOutsideRight));
        CHECK(!cylinder->isInside(justOutsideFront));
        CHECK(!cylinder->isInside(justOutsideBack));

        CHECK(cylinder->getCenterPosition() == centre);
    }

    SUBCASE("Point within (moving) Mesh")
    {
        //define mesh
        auto transform = EAR::SceneObjectFactory<EAR::AnchorObject>::getNewObject();
        transform->setLocation(EAR::Vector3(0, 0, 0));

        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();
        mesh->setCSpace(EAR::CSpace::RCS, transform.get());
        mesh->setStatic(false);

        mesh->addVertex(EAR::Vector3(0.5, 0, -0.5));
        mesh->addVertex(EAR::Vector3(-0.5, 0, -0.5));
        mesh->addVertex(EAR::Vector3(0.5, 1, -0.5));
        mesh->addVertex(EAR::Vector3(-0.5, 1, -0.5));

        mesh->addVertex(EAR::Vector3(0.5, 1, 0.5));
        mesh->addVertex(EAR::Vector3(-0.5, 1, 0.5));
        mesh->addVertex(EAR::Vector3(0.5, 0, 0.5));
        mesh->addVertex(EAR::Vector3(-0.5, 0, 0.5));

        mesh->addTriangle(0, 2, 3);
        mesh->addTriangle(0, 3, 1);
        mesh->addTriangle(2, 4, 5);
        mesh->addTriangle(2, 5, 3);
        mesh->addTriangle(4, 6, 7);
        mesh->addTriangle(4, 7, 5);
        mesh->addTriangle(6, 0, 1);
        mesh->addTriangle(6, 1, 7);
        mesh->addTriangle(1, 3, 5);
        mesh->addTriangle(1, 5, 7);
        mesh->addTriangle(6, 4, 2);
        mesh->addTriangle(6, 2, 0);

        //test whether point is inside or outside

        mesh->lock();
        mesh->setLocation(EAR::Vector3(10, 0, 0));

        //inside
        EAR::Vector3 p0(10, 0.4, 0.1);
        CHECK(mesh->isInside(p0));
        //CHECK(cube->isInside(p0));

        //outside
        EAR::Vector3 p1(15, 5, 5);
        CHECK(!mesh->isInside(p1));

        //outside, inital ray || to surface
        EAR::Vector3 p3(10.5, -2, -0.5);
        CHECK(!mesh->isInside(p3));

        EAR::Vector3 p5(10.5, -1.5, -0.5);
        CHECK(!mesh->isInside(p5));

        EAR::Vector3 p6(10.5, -2.1, -0.5);
        CHECK(!mesh->isInside(p6));

        EAR::Vector3 p7(10.5, -0.001, -0.5);
        CHECK(!mesh->isInside(p7));

        //barely inside
        EAR::Vector3 p4(10.49, 0.1, -0.49);
        CHECK(mesh->isInside(p4));

        //
        EAR::Vector3 p8(10, 0, 2);
        CHECK(!mesh->isInside(p8));

        // test after moving
        mesh->setLocation(EAR::Vector3(10.5, -0.5, -0.5));
        CHECK(mesh->isInside(p7));

        // test after moving the parent SceneObject
        transform->setLocation(EAR::Vector3(3, 0, 0));
        CHECK(!mesh->isInside(p7));
    }

    SUBCASE("Point within Box-turned-Mesh")
    {
        auto cube = EAR::SceneObjectFactory<EAR::Box>::getNewObject(2.f, 2.f, 2.f);
        auto cMesh = cube->getMesh();

        cube->setLocation(EAR::Vector3(0.f, 0.f, 4.f));

        EAR::Vector3 p1(0, 0, 0);
        CHECK(!cube->isInside(p1));
        CHECK(!cMesh->isInside(p1));

        cube->setLocation(EAR::Vector3(4.f, 0.f, 0.f));
        CHECK(!cube->isInside(p1));
        CHECK(!cMesh->isInside(p1));
    }
}

TEST_CASE("Cylinder::getCenterOfBases")
{
    float r = 4.f;
    float h = 4.f;

    auto tube = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(r, h);
    std::pair<EAR::Vector3, EAR::Vector3> baseCenters = tube->getCenterOfBases();

    CHECK(baseCenters.first == EAR::Vector3(0, -2, 0)); //lower
    CHECK(baseCenters.second == EAR::Vector3(0, 2, 0)); //upper
}

TEST_CASE("greatestDistance()")
{
    //cylinder
    auto tubeCircle = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(4.f, 4.f);
    CHECK(std::round(tubeCircle->greatestDistance() * 10000) == std::round(std::sqrt(80) * 10000));

    auto tubeEllipse = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(10.f, 1.f, 4.f);
    CHECK(std::round(tubeEllipse->greatestDistance() * 10000)
        == std::round(std::sqrt(20.f * 20.f + 16.f) * 10000));

    //mesh
    auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

    mesh->addVertex(EAR::Vector3(5, 0, -5));
    mesh->addVertex(EAR::Vector3(-5, 0, -5));
    mesh->addVertex(EAR::Vector3(5, 1, -5));
    mesh->addVertex(EAR::Vector3(-5, 1, -5));

    mesh->addVertex(EAR::Vector3(5, 1, 5));
    mesh->addVertex(EAR::Vector3(-5, 1, 5));
    mesh->addVertex(EAR::Vector3(5, 0, 5));
    mesh->addVertex(EAR::Vector3(-5, 0, 5));

    mesh->addTriangle(0, 2, 3);
    mesh->addTriangle(0, 3, 1);
    mesh->addTriangle(2, 4, 5);
    mesh->addTriangle(2, 5, 3);
    mesh->addTriangle(4, 6, 7);
    mesh->addTriangle(4, 7, 5);
    mesh->addTriangle(6, 0, 1);
    mesh->addTriangle(6, 1, 7);
    mesh->addTriangle(1, 3, 5);
    mesh->addTriangle(1, 5, 7);
    mesh->addTriangle(6, 4, 2);
    mesh->addTriangle(6, 2, 0);

    CHECK(std::round(mesh->greatestDistance() * 10000) == std::round(std::sqrt(200 + 1) * 10000));

    //sphere
    auto sp = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(4.f);
    CHECK(sp->greatestDistance() == 8);

    //box
    EAR::Vector3 v(4, 4, 4);
    auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(v);
    CHECK(std::round(box->greatestDistance() * 10000) == std::round(std::sqrt(32 + 16) * 10000));
}

TEST_CASE("Area and Volume")
{

    SUBCASE("Area and Volume of Mesh")
    {
        float xs = 3.f;
        float ys = 5.f;
        float zs = 7.f;
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(xs, ys, zs);
        box->setLocation(EAR::Vector3(1.0, 0.7, 0.5));
        const auto& mesh = box->getMesh();

        CHECK(mesh->getArea() == 2 * (xs * ys + xs * zs + ys * zs));
        CHECK(mesh->getVolume() == xs * ys * zs);
    }

    SUBCASE("Area and Volume of Box")
    {
        float r = 4.f;
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(r, r, r);

        CHECK(box->getArea() == 6 * std::pow(r, 2));
        CHECK(box->getVolume() == std::pow(r, 3));
    }

    SUBCASE("Area and Volume of Sphere")
    {
        float r = 4.f;
        auto sphere = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(r);

        CHECK(sphere->getArea() == static_cast<float>(4.f * M_PI * std::pow(r, 2)));
        CHECK(sphere->getVolume() == static_cast<float>(4.f / 3.f * M_PI * std::pow(r, 3)));
    }

    SUBCASE("Area and Volume of Cylinder")
    {
        float r = 4.f;
        float h = 4.f;

        auto tubeCircle = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(r, h);

        CHECK(tubeCircle->getArea()
            == static_cast<float>(2 * M_PI * r * h + 2 * M_PI * std::pow(r, 2)));
        CHECK(tubeCircle->getVolume() == static_cast<float>(M_PI * std::pow(r, 2) * h));

        float a = 10;
        float b = 1;

        auto tubeEllipse = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(a, b, h);
        CHECK(tubeEllipse->getVolume() - static_cast<float>(h * a * b * M_PI)
            < EAR::Geometry::EPSILON);
        CHECK(tubeEllipse->getArea() - (40.639 * h) - M_PI * a * b * 2.f < 0.1);
    }
}

TEST_CASE("Orientation")
{
    SUBCASE("Provide Mesh with Orientation")
    {
        //mesh which is rotate by 90° around z-axis
        EAR::Position meshPosition;
        float rotationAngle = 90.f / 180 * M_PI;
        meshPosition.location = EAR::Vector3(0, 0, 0);
        meshPosition.orientation = EAR::Euler(0, 0, 90);

        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        mesh->addVertex(EAR::Vector3(5, 0, -5)); // vertex_inx = 0
        mesh->addVertex(EAR::Vector3(-5, 0, -5)); //.
        mesh->addVertex(EAR::Vector3(5, 1, -5)); //.
        mesh->addVertex(EAR::Vector3(-5, 1, -5)); //.

        mesh->addVertex(EAR::Vector3(5, 1, 5)); //.
        mesh->addVertex(EAR::Vector3(-5, 1, 5)); //.
        mesh->addVertex(EAR::Vector3(5, 0, 5)); //.
        mesh->addVertex(EAR::Vector3(-5, 0, 5)); // vertex_inx = 7

        mesh->addTriangle(0, 2, 3);
        mesh->addTriangle(0, 3, 1);
        mesh->addTriangle(2, 4, 5);
        mesh->addTriangle(2, 5, 3);
        mesh->addTriangle(4, 6, 7);
        mesh->addTriangle(4, 7, 5);
        mesh->addTriangle(6, 0, 1);
        mesh->addTriangle(6, 1, 7);
        mesh->addTriangle(1, 3, 5);
        mesh->addTriangle(1, 5, 7);
        mesh->addTriangle(6, 4, 2);
        mesh->addTriangle(6, 2, 0);

        mesh->setOrientation(meshPosition.orientation);
        EAR::Matrix3_3 rotationMatrix({ { std::cos(rotationAngle), -std::sin(rotationAngle), 0 },
            { std::sin(rotationAngle), std::cos(rotationAngle), 0 }, { 0, 0, 1 } });

        // Evaluate vertices positions by metric (Used metric: distance)
        for (int i = 0; i < mesh->getVertices().size(); i++) {
            EAR::Vector3 vertexLoc = mesh->getVertex(i, EAR::Mesh::CSpace::Local);
            EAR::Vector3 vertexGlo = mesh->getVertex(i, EAR::Mesh::CSpace::Global);

            float distance = (vertexGlo - rotationMatrix * vertexLoc).abs();

            CHECK(distance < 0.0001);
        }
    }
}

TEST_CASE("Surfaces and Their Geometries")
{
    SUBCASE("Meshes and Surfaces")
    {
        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        mesh->addVertex(EAR::Vector3(5, 0, -5)); // vertex_inx = 0
        mesh->addVertex(EAR::Vector3(-5, 0, -5)); //.
        mesh->addVertex(EAR::Vector3(5, 1, -5)); //.
        mesh->addVertex(EAR::Vector3(-5, 1, -5)); //.

        mesh->addVertex(EAR::Vector3(5, 1, 5)); //.
        mesh->addVertex(EAR::Vector3(-5, 1, 5)); //.
        mesh->addVertex(EAR::Vector3(5, 0, 5)); //.
        mesh->addVertex(EAR::Vector3(-5, 0, 5)); // vertex_inx = 7

        mesh->addTriangle(0, 2, 3);
        mesh->addTriangle(0, 3, 1);
        mesh->addTriangle(2, 4, 5);
        mesh->addTriangle(2, 5, 3);
        mesh->addTriangle(4, 6, 7);
        mesh->addTriangle(4, 7, 5);
        mesh->addTriangle(6, 0, 1);
        mesh->addTriangle(6, 1, 7);
        mesh->addTriangle(1, 3, 5);
        mesh->addTriangle(1, 5, 7);
        mesh->addTriangle(6, 4, 2);
        mesh->addTriangle(6, 2, 0);
        mesh->rebuildSurfaceStructure();

        int numberOfSurfaces = mesh->getSurfaces().size();
        CHECK(numberOfSurfaces == 6);
    }
    SUBCASE("Box and Surfaces")
    {
        EAR::Vector3 volume(30, 30, 30);
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(volume);
        auto meshOfBox = box->getMesh();

        //build new mesh by using the box's mesh
        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        auto verticesOfBox = meshOfBox->getVertices(EAR::Mesh::CSpace::Global);

        for (const auto& vertex : verticesOfBox) {
            mesh->addVertex(vertex);
        }

        auto trianglesOfBox = meshOfBox->getTriangles();
        for (const auto& triangle : trianglesOfBox) {
            mesh->addTriangle(triangle.v1, triangle.v2, triangle.v3);
        }

        mesh->rebuildSurfaceStructure();
        //test on number of surface
        CHECK(mesh->getSurfaces().size() == 6);
    }
}

TEST_CASE("distance to Mesh")
{
    SUBCASE("Point to cube")
    {
        auto cube = EAR::SceneObjectFactory<EAR::Box>::getNewObject(5.f, 5.f, 5.f);

        cube->setLocation(EAR::Vector3(10.f, 10.f, 10.f));

        EAR::Vector3 above(10.f, 13.f, 10.f);
        EAR::Vector3 below(10.f, 6.f, 10.f);
        EAR::Vector3 sceneOrigin(0.f, 0.f, 0.f);

        CHECK((((cube->closestPointTo(above)) - above).abs() - 0.5f) < 0.001);
        CHECK((((cube->closestPointTo(below)) - below).abs() - 1.5f) < 0.001);
        CHECK((((cube->closestPointTo(sceneOrigin)) - sceneOrigin).abs() - 12.99f) < 0.001);
    }

    SUBCASE("Point to cube 2")
    {
        // box is hallway from Hospital scene
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(40.f, 2.5f, 2.0f);
        box->setLocation(EAR::Vector3(0.f, 1.25f, 3.7f));

        const EAR::Vector3 point(-3.73397446f, 0.f, 3.19999981f);

        // Point is just barely not on the box.
        const EAR::Vector3 expectedClosestPoint(-3.734f, 0.f, 3.2f);

        const EAR::Vector3 closestPoint = box->closestPointTo(point);

        // Verify computed point is within 0.1 of actual closest point
        CHECK(closestPoint[0] == doctest::Approx(expectedClosestPoint[0]).epsilon(0.1));
        CHECK(closestPoint[1] == doctest::Approx(expectedClosestPoint[1]).epsilon(0.1));
        CHECK(closestPoint[2] == doctest::Approx(expectedClosestPoint[2]).epsilon(0.1));
    }

    SUBCASE("Point to cube 3")
    {
        // box is patient room from Hospital scene
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(7.f, 2.5f, 5.0f);
        box->setLocation(EAR::Vector3(0.f, 1.25f, 0.f));

        const EAR::Vector3 point(2.20000005f, 0.f, 2.39999986f);

        // Point is approximately on the box mesh
        const EAR::Vector3 expectedClosestPoint = point;

        const EAR::Vector3 closestPoint = box->closestPointTo(point);
        ;
        // Verify computed point is within 0.1 of actual closest point
        CHECK(closestPoint[0] == doctest::Approx(expectedClosestPoint[0]).epsilon(0.1));
        CHECK(closestPoint[1] == doctest::Approx(expectedClosestPoint[1]).epsilon(0.1));
        CHECK(closestPoint[2] == doctest::Approx(expectedClosestPoint[2]).epsilon(0.1));
    }

    SUBCASE("Point to sphere")
    {
        auto sphere = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(4.f);

        sphere->setLocation(EAR::Vector3(8.f, 8.f, -8.f));

        EAR::Vector3 justOutsideTop(8.f, 13.f, -8.f);
        EAR::Vector3 justOutsideBot(8.f, 2.5f, -8.f);
        EAR::Vector3 justOutsideBack(8.f, 8.f, -13.f);

        CHECK(((sphere->closestPointTo(justOutsideTop)) - justOutsideTop).abs() - 1.f < 0.001);
        CHECK(((sphere->closestPointTo(justOutsideBot)) - justOutsideBot).abs() - 1.5f < 0.001);
        CHECK(((sphere->closestPointTo(justOutsideBack)) - justOutsideBack).abs() - 1.f < 0.001);
    }

    SUBCASE("Point to cylinder")
    {
        //circular cylinder
        auto cylinder = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(2.f, 4.f);

        cylinder->setOrientation(EAR::Euler(0.f, 90.f, 0.f));
        cylinder->setLocation(EAR::Vector3(0.f, 2.f, -2.f));

        EAR::Vector3 justOutsideTop(0.f, 5.f, -2.f);
        EAR::Vector3 justOutsideBot(0.f, -1.f, 0.f);
        EAR::Vector3 justOutsideLeft(1.f, 1.f, 1.f);
        EAR::Vector3 justOutside(2.f, 0.f, -2.f);

        CHECK(((cylinder->closestPointTo(justOutsideTop)) - justOutsideTop).abs() - 1.f < 0.001);
        CHECK(((cylinder->closestPointTo(justOutsideBot)) - justOutsideBot).abs() - 1.f < 0.001);
        CHECK(((cylinder->closestPointTo(justOutsideLeft)) - justOutsideLeft).abs() - 1.f < 0.001);
        CHECK(((cylinder->closestPointTo(justOutside)) - justOutside).abs()
            < std::sqrt(8) - 2 + 0.001);

        //elliptical cylinder
        float a = 10;
        float b = 1;
        float h = 4;
        unsigned int iterN = 3;

        auto tubeElliptical = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(a, b, h);

        tubeElliptical->setOrientation(EAR::Euler(0, 90, 0));
        tubeElliptical->setLocation(EAR::Vector3(0, 2.f, -2.f));

        EAR::Vector3 justOutsideRight = EAR::Vector3(0.f, 4.f, -2.f);
        justOutsideLeft = EAR::Vector3(0.f, 0.f, -2.f);
        EAR::Vector3 justOutsideFront = EAR::Vector3(1.f, 1.f, 1.f);
        EAR::Vector3 justOutsideBack = EAR::Vector3(0, 3.f, -5.f);
        justOutsideTop = EAR::Vector3(11.f, 2.f, -2.f);
        justOutsideBot = EAR::Vector3(-11.f, 2.f, -2.f);

        CHECK((tubeElliptical->closestPointTo(justOutsideRight) - justOutsideRight).abs() - 1.f
            < 0.1);
        CHECK(
            (tubeElliptical->closestPointTo(justOutsideLeft) - justOutsideLeft).abs() - 1.f < 0.1);
        CHECK((tubeElliptical->closestPointTo(justOutsideFront) - justOutsideFront).abs() - 1.f
            < 0.1);
        CHECK(
            (tubeElliptical->closestPointTo(justOutsideBack) - justOutsideBack).abs() - 1.f < 0.1);
        CHECK((tubeElliptical->closestPointTo(justOutsideTop) - justOutsideTop).abs() - 1.f < 0.1);
        CHECK((tubeElliptical->closestPointTo(justOutsideBot) - justOutsideBot).abs() - 1.f < 0.1);
    }

    /* SUBCASE("Point to ellipse") {

        //elliptical cylinder
        float a = 5;
        float b = 1;
        float h = 4;
        unsigned int iterN = 4;

        auto tubeElliptical = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(a, b, h);

        EAR::Vector3 arbitraryPoint;
        for(unsigned int i = 0; i < 10000; i++){
            float x = -std::rand () % 10 - 5.1;
            float y = -std::rand () % 10 - 2.1;
            float z = -std::rand () % 10 - 1.1;
            arbitraryPoint = EAR::Vector3(x,y,z);
            CHECK((tubeElliptical->closestPoint2Ellipse_Newton(arbitraryPoint) - tubeElliptical->closestPoint2Ellipse_Curvature(arbitraryPoint, iterN)).abs() < 0.1);
        }
    } */
}

TEST_CASE("2D Projection")
{
    SUBCASE("Mesh")
    {
        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        mesh->addVertex(EAR::Vector3(2, -2, -2));
        mesh->addVertex(EAR::Vector3(-2, -2, -2));
        mesh->addVertex(EAR::Vector3(2, 2, -2));
        mesh->addVertex(EAR::Vector3(-2, 2, -2));
        mesh->addVertex(EAR::Vector3(2, 2, 2));
        mesh->addVertex(EAR::Vector3(-2, 2, 2));
        mesh->addVertex(EAR::Vector3(2, -2, 2));
        mesh->addVertex(EAR::Vector3(-2, -2, 2));

        mesh->addTriangle(0, 2, 3);
        mesh->addTriangle(0, 3, 1);
        mesh->addTriangle(2, 4, 5);
        mesh->addTriangle(2, 5, 3);
        mesh->addTriangle(4, 6, 7);
        mesh->addTriangle(4, 7, 5);
        mesh->addTriangle(6, 0, 1);
        mesh->addTriangle(6, 1, 7);
        mesh->addTriangle(1, 3, 5);
        mesh->addTriangle(1, 5, 7);
        mesh->addTriangle(6, 4, 2);
        mesh->addTriangle(6, 2, 0);
        mesh->lock();

        EAR::Position listener;
        listener.location = EAR::Vector3(0, 0, 4);
        listener.orientation = EAR::Euler(0, 0, 0);
        EAR::Vector3 resBL, resTR, resBR, resTL, closestP;
        bool povIsInside;

        mesh->getPOVContourEdges(listener, resBL, resTR, resBR, resTL, povIsInside, closestP);

        CHECK(resBL[0] == doctest::Approx(-2.f));
        CHECK(resBL[1] == doctest::Approx(-2.f));
        CHECK(resBL[2] == doctest::Approx(-2.f));

        CHECK(resBR[0] == doctest::Approx(2.f));
        CHECK(resBR[1] == doctest::Approx(-2.f));
        CHECK(resBR[2] == doctest::Approx(-2.f));

        CHECK(resTL[0] == doctest::Approx(-2.f));
        CHECK(resTL[1] == doctest::Approx(2.f));
        CHECK(resTL[2] == doctest::Approx(-2.f));

        CHECK(resTR[0] == doctest::Approx(2.f));
        CHECK(resTR[1] == doctest::Approx(2.f));
        CHECK(resTR[2] == doctest::Approx(-2.f));

        CHECK(povIsInside == false);

        CHECK(closestP[0] == doctest::Approx(0.f));
        CHECK(closestP[1] == doctest::Approx(0.f));
        CHECK(closestP[2] == doctest::Approx(2.f));
    }

    SUBCASE("Box")
    {
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(EAR::Vector3(4, 4, 4));

        EAR::Position listener;
        listener.location = EAR::Vector3(0, 0, 4);
        listener.orientation = EAR::Euler(45, 0, 0);
        EAR::Vector3 resBL, resTR, resBR, resTL, closestP;
        bool povIsInside;

        box->getPOVContourEdges(listener, resBL, resTR, resBR, resTL, povIsInside, closestP);

        CHECK(resBL[0] == doctest::Approx(0.f));
        CHECK(resBL[1] == doctest::Approx(-2.f));
        CHECK(resBL[2] == doctest::Approx(-2.82843f));

        CHECK(resBR[0] == doctest::Approx(2.82843f));
        CHECK(resBR[1] == doctest::Approx(-2.f));
        CHECK(resBR[2] == doctest::Approx(0.f));

        CHECK(resTL[0] == doctest::Approx(0.f));
        CHECK(resTL[1] == doctest::Approx(2.f));
        CHECK(resTL[2] == doctest::Approx(-2.82843f));

        CHECK(resTR[0] == doctest::Approx(2.82843f));
        CHECK(resTR[1] == doctest::Approx(2.f));
        CHECK(resTR[2] == doctest::Approx(0.f));

        CHECK(povIsInside == false);

        CHECK(closestP[0] == doctest::Approx(0.f));
        CHECK(closestP[1] == doctest::Approx(0.f));
        CHECK(closestP[2] == doctest::Approx(2.f));
    }

    SUBCASE("Sphere")
    {
        auto sphere = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(1);

        EAR::Position listener;
        listener.location = EAR::Vector3(0, 0, 2);
        listener.orientation = EAR::Euler(0, 0, 0);
        EAR::Vector3 resBL, resTR, resBR, resTL, closestP;
        bool povIsInside;

        sphere->getPOVContourEdges(listener, resBL, resTR, resBR, resTL, povIsInside, closestP);

        CHECK(resBL[0] == doctest::Approx(-0.6f));
        CHECK(resBL[1] == doctest::Approx(-0.6f));
        CHECK(resBL[2] == doctest::Approx(-1.f));

        CHECK(resBR[0] == doctest::Approx(0.6f));
        CHECK(resBR[1] == doctest::Approx(-0.6f));
        CHECK(resBR[2] == doctest::Approx(-1.f));

        CHECK(resTL[0] == doctest::Approx(-0.6f));
        CHECK(resTL[1] == doctest::Approx(0.6f));
        CHECK(resTL[2] == doctest::Approx(-1.f));

        CHECK(resTR[0] == doctest::Approx(0.6f));
        CHECK(resTR[1] == doctest::Approx(0.6f));
        CHECK(resTR[2] == doctest::Approx(-1.f));

        CHECK(povIsInside == false);

        CHECK(closestP[0] == doctest::Approx(0.f));
        CHECK(closestP[1] == doctest::Approx(0.f));
        CHECK(closestP[2] == doctest::Approx(1.f));
    }

    SUBCASE("Cylinder")
    {
        auto cylinder = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(1, 2);
        cylinder->setOrientation(EAR::Euler(90, 90, 0));
        cylinder->setLocation(EAR::Vector3(0, 0, 0));

        EAR::Position listener;
        listener.location = EAR::Vector3(4, 0, 0);
        listener.orientation = EAR::Euler(0, 0, 0);
        EAR::Vector3 resBL, resTR, resBR, resTL, closestP;
        bool povIsInside;

        cylinder->getPOVContourEdges(listener, resBL, resTR, resBR, resTL, povIsInside, closestP);

        float scale = 1.01733243;

        CHECK(resBL[0] == doctest::Approx(-3.f));
        CHECK(resBL[1] == doctest::Approx(-1.f * scale));
        CHECK(resBL[2] == doctest::Approx(1.f * scale));

        CHECK(resBR[0] == doctest::Approx(-3.f));
        CHECK(resBR[1] == doctest::Approx(-1.f * scale));
        CHECK(resBR[2] == doctest::Approx(-1.f * scale));

        CHECK(resTL[0] == doctest::Approx(-3.f));
        CHECK(resTL[1] == doctest::Approx(1.f * scale));
        CHECK(resTL[2] == doctest::Approx(1.f * scale));

        CHECK(resTR[0] == doctest::Approx(-3.f));
        CHECK(resTR[1] == doctest::Approx(1.f * scale));
        CHECK(resTR[2] == doctest::Approx(-1.f * scale));

        CHECK(povIsInside == false);

        CHECK(closestP[0] == doctest::Approx(1.f));
        CHECK(closestP[1] == doctest::Approx(0.f));
        CHECK(closestP[2] == doctest::Approx(0.f));
    }

    SUBCASE("elliptical cylinder")
    {
        auto cylinder = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(10, 4, 2);
        cylinder->setOrientation(EAR::Euler(90, 90, 0));
        cylinder->setLocation(EAR::Vector3(-4, 0, 0));

        EAR::Position listener;
        listener.location = EAR::Vector3(0, 0, 0);
        listener.orientation = EAR::Euler(0, 0, 0);
        EAR::Vector3 resBL, resTR, resBR, resTL, closestP;
        bool povIsInside;

        cylinder->getPOVContourEdges(listener, resBL, resTR, resBR, resTL, povIsInside, closestP);

        float scale = 1.01733243;

        CHECK(resBL[0] == doctest::Approx(-3.f));
        CHECK(resBL[1] == doctest::Approx(-4.f * scale));
        CHECK(resBL[2] == doctest::Approx(10.f * scale));

        CHECK(resBR[0] == doctest::Approx(-3.f));
        CHECK(resBR[1] == doctest::Approx(-4.f * scale));
        CHECK(resBR[2] == doctest::Approx(-10.f * scale));

        CHECK(resTL[0] == doctest::Approx(-3.f));
        CHECK(resTL[1] == doctest::Approx(4.f * scale));
        CHECK(resTL[2] == doctest::Approx(10.f * scale));

        CHECK(resTR[0] == doctest::Approx(-3.f));
        CHECK(resTR[1] == doctest::Approx(4.f * scale));
        CHECK(resTR[2] == doctest::Approx(-10.f * scale));

        CHECK(povIsInside == false);

        CHECK(closestP[0] == doctest::Approx(-3.f));
        CHECK(closestP[1] == doctest::Approx(0.f));
        CHECK(closestP[2] == doctest::Approx(0.f));
    }
}

TEST_CASE("Geometry::getBoundingBox")
{
    SUBCASE("Bounding box for Mesh")
    {
        // Mesh properties from "mesh:acousticEnvironment" in EIF_SingerInTheLab.xml

        EAR::Position position{};
        position.location = { 0.F, 1.5F, 0.F };

        auto mesh = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();
        mesh->setPosition(position);

        mesh->addVertex({ 4.2407F, -2.6157F, -5.1061F });
        mesh->addVertex({ 4.2407F, 2.6157F, -5.1061F });
        mesh->addVertex({ 4.2407F, 2.6157F, 5.1061F });
        mesh->addVertex({ 4.2407F, -2.6157F, 5.1061F });
        mesh->addVertex({ -4.2407F, 2.6157F, 5.1061F });
        mesh->addVertex({ -4.2407F, -2.6157F, 5.1061F });
        mesh->addVertex({ -4.2407F, 2.6157F, -5.1061F });
        mesh->addVertex({ -4.2407F, -2.6157F, -5.1061F });

        mesh->addTriangle(2, 1, 0);
        mesh->addTriangle(3, 2, 0);
        mesh->addTriangle(4, 2, 3);
        mesh->addTriangle(5, 4, 3);
        mesh->addTriangle(6, 4, 5);
        mesh->addTriangle(7, 6, 5);
        mesh->addTriangle(1, 6, 7);
        mesh->addTriangle(0, 1, 7);
        mesh->addTriangle(7, 5, 3);
        mesh->addTriangle(0, 7, 3);
        mesh->addTriangle(1, 2, 4);
        mesh->addTriangle(6, 1, 4);

        // Lock the Mesh to get it to rebuild its internals based on the added vertices and triangles.
        mesh->lock();
        mesh->unlock();

        auto bb = mesh->getBoundingBox();
        CHECK(bb[0].high - bb[0].low == 4.2407F - -4.2407F);
        CHECK(bb[1].high - bb[1].low == 2.6157F - -2.6157F);
        CHECK(bb[2].high - bb[2].low == 5.1061F - -5.1061F);
    }

    SUBCASE("Bounding box for Box")
    {
        // Box properties from "box::inMaze" in EIF_SimpleMaze.xml

        EAR::Position position{};
        position.location = { 0.F, 1.5F, 0.F };
        auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(14.F, 3.F, 14.F);
        box->setPosition(position);

        auto bb = box->getBoundingBox();
        CHECK(bb[0].high - bb[0].low == 14.F);
        CHECK(bb[1].high - bb[1].low == 3.F);
        CHECK(bb[2].high - bb[2].low == 14.F);
    }

    SUBCASE("Bounding box for Cylinder is not implemented")
    {
        // Circle::getBoundingBox hasn't been needed, so it's not implemented.

        auto cylinder = EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(0.F, 0.F, 0.F);

        CHECK_THROWS(cylinder->getBoundingBox());
    }

    SUBCASE("Bounding box for Sphere is not implemented")
    {
        // Sphere::getBoundingBox hasn't been needed, so it's not implemented.

        auto sphere = EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(4.f);

        CHECK_THROWS(sphere->getBoundingBox());
    }
}
