/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/ComplexIIRFilter.h>
#include <EAR/Exception.h>
#include <EAR/IIRFilter.h>

#include <complex>
#include <iostream>

using namespace EAR;

TEST_CASE("IIR filter test")
{

    SUBCASE("unity filter")
    {
        const float b[] = { 1 };
        const float a[] = { 1 };
        const float x[] = { 1, 2, 3, 4 };
        const float ref[] = { 1, 2, 3, 4 };
        float y[4];

        try { // the IPP implementation does not support order 0
            auto filter = IIRFilter::create(b, 1, a, 1);
            filter->process(x, y, 4);

            for (auto i = 0; i < 4; i++) {
                CHECK(y[i] == doctest::Approx(ref[i]).epsilon(1e-9f));
            }
        } catch (const Exception& e) {
            std::cout << "Caught exception: " << e.what() << std::endl;
        }
    }

    SUBCASE("2-tap MA filter")
    {
        const float b[] = { 1, 2 };
        const float a[] = { 1 };
        const float x[] = { 1, 2, 3, 4 };
        const float ref[] = { 1, 4, 7, 10 };
        float y[4];

        auto filter = IIRFilter::create(b, 2, a, 1);
        filter->process(x, y, 4);

        for (auto i = 0; i < 4; i++) {
            CHECK(y[i] == doctest::Approx(ref[i]).epsilon(1e-9f));
        }
    }

    SUBCASE("2-tap feedback filter")
    {
        const float b[] = { 1 };
        const float a[] = { 1, -1 };
        const float x[] = { 1, 2, 3, 4 };
        const float ref[] = { 1, 3, 6, 10 };
        float y[4];

        auto filter = IIRFilter::create(b, 1, a, 2);
        filter->process(x, y, 4);

        for (auto i = 0; i < 4; i++) {
            CHECK(y[i] == doctest::Approx(ref[i]).epsilon(1e-9f));
        }
    }

    SUBCASE("4-tap pole/zero filter")
    {
        const float b[] = { 0.5, 0.5, 1, 0.25 };
        const float a[] = { 1, -1, -0.5, -0.25 };
        const float x[] = { 1, 2, 3, 4 };
        const float ref[] = { 0.5, 2, 5.75, 12.625 };
        float y[4];

        auto filter = IIRFilter::create(b, 4, a, 4);
        filter->process(x, y, 4);

        for (auto i = 0; i < 4; i++) {
            CHECK(y[i] == doctest::Approx(ref[i]).epsilon(1e-9f));
        }
    }

    SUBCASE("denormalized 4-tap pole/zero filter")
    {
        const float b[] = { 2 * 0.5, 2 * 0.5, 2 * 1, 2 * 0.25 };
        const float a[] = { 2 * 1, 2 * -1, 2 * -0.5, 2 * -0.25 };
        const float x[] = { 1, 2, 3, 4 };
        const float ref[] = { 0.5, 2, 5.75, 12.625 };
        float y[4];

        auto filter = IIRFilter::create(b, 4, a, 4);
        filter->process(x, y, 4);

        for (auto i = 0; i < 4; i++) {
            CHECK(y[i] == doctest::Approx(ref[i]).epsilon(1e-9f));
        }
    }

    SUBCASE("compare real and complex IIRfilter")
    {
        const float b[] = { 0.7, 0, 0, 1.f };
        const float a[] = { 1.f, 0, 0, 0.7 };
        const float x_real[] = { 1, 0, 0, 0 };
        const std::complex<float> x_complex[] = { 1, 0, 0, 0 };
        float y_real[4];
        std::complex<float> y_complex[4];

        auto realFilter = IIRFilter::create(b, 4, a, 4);
        auto complexFilter = ComplexIIRFilter(b, 4, a, 4);

        realFilter->process(x_real, y_real, 4);

        for (auto i = 0; i < 4; i++) {
            y_complex[i] = complexFilter.process(x_complex[i]);
            CHECK(y_real[i] == doctest::Approx(y_complex[i].real()).epsilon(1e-9f));
        }
    }
}
