/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>
#include <string>

#include <EAR/Exception.h>
#include <EAR/OSC.h>
#include <EAR/TimeMeasure.h>
#include <EAR/Variant.h>

#include <thread>

#include <iostream>

TEST_CASE("OSC Messages and Bundles")
{

    SUBCASE("Serialize/Deserialize")
    {
        EAR::OSC::Message message("/hello/world");

        bool testBoolArg = true;
        int testIntArg = 6;
        float testFloatArg = 3.142f;
        std::string testStringArg = "I'm an argument!";

        message.addArgument(EAR::Variant(testBoolArg));
        message.addArgument(EAR::Variant(testIntArg)).addArgument(EAR::Variant(testFloatArg));
        message.addArgument(EAR::Variant(testStringArg));

        auto newMessage = EAR::OSC::Message::deserialize(message.serialize());
        CHECK(newMessage.serialize() == message.serialize());

        std::string testStringMsg1Arg1 = "I live in message 1!";
        std::string testStringMsg1Arg2 = "I also live in message 1!";
        std::string testStringMsg1Arg3 = "I too live in message 1!";

        std::string testStringMsg2Arg1 = "I live in message 2!";
        std::string testStringMsg2Arg2 = "I also live in message 2!";

        std::string testStringMsg3Arg1 = "I live in message 3!";

        EAR::OSC::Bundle masterBundle;
        EAR::OSC::Bundle nestedBundle;
        EAR::OSC::Message message1("/message/1");
        EAR::OSC::Message message2("/message/2");
        EAR::OSC::Message message3("/message/3");

        message1.addArgument(EAR::Variant(testStringMsg1Arg1))
            .addArgument(EAR::Variant(testStringMsg1Arg2))
            .addArgument(EAR::Variant(testStringMsg1Arg3));

        message2.addArgument(EAR::Variant(testStringMsg2Arg1))
            .addArgument(EAR::Variant(testStringMsg2Arg2));

        message3.addArgument(EAR::Variant(testStringMsg3Arg1));

        nestedBundle.addElement(message1).addElement(message2);

        masterBundle.addElement(nestedBundle).addElement(message3);

        auto newBundle = EAR::OSC::Bundle::deserialize(masterBundle.serialize());
        CHECK(newBundle.serialize() == masterBundle.serialize());
    }

    SUBCASE("Send and wait to receive message")
    {
        EAR::OSC osc("127.0.0.1", 9001);

        bool testBoolArg = true;
        int testIntArg = 6;
        float testFloatArg = 3.142f;
        std::string testStringArg = "I'm an argument!";

        //Create message
        EAR::OSC::Message message("/hello/world");
        message.addArgument(EAR::Variant(testBoolArg));
        message.addArgument(EAR::Variant(testIntArg)).addArgument(EAR::Variant(testFloatArg));
        message.addArgument(EAR::Variant(testStringArg));

        bool boolArg;
        int intArg;
        float floatArg;
        std::string stringArg;

        //Create listener
        osc.addListener("/hello/world",
            [&boolArg, &intArg, &floatArg, &stringArg](const EAR::OSC::Message& message) {
                auto args = message.getArguments();

                CHECK(EAR::Variant::Type::Bool == args[0].getVariantType());
                boolArg = static_cast<bool>(args[0]);

                CHECK(EAR::Variant::Type::Int == args[1].getVariantType());
                intArg = static_cast<int>(args[1]);

                CHECK(EAR::Variant::Type::Float == args[2].getVariantType());
                floatArg = static_cast<float>(args[2]);

                CHECK(EAR::Variant::Type::String == args[3].getVariantType());
                stringArg = static_cast<std::string>(args[3]);
            });

        osc.send(message);

        // wait to be sure that the message was received
        using namespace std::chrono_literals;
        std::this_thread::sleep_for(0.5s);

        CHECK(testBoolArg == boolArg);
        CHECK(testIntArg == intArg);
        CHECK(testFloatArg == floatArg);
        CHECK(testStringArg == stringArg);
    }

    SUBCASE("Send and receive message")
    {
        EAR::OSC osc;

        bool testBoolArg = true;
        int testIntArg = 6;
        float testFloatArg = 3.142f;
        std::string testStringArg = "I'm an argument!";

        //Create message
        EAR::OSC::Message message("/hello/world");
        message.addArgument(EAR::Variant(testBoolArg));
        message.addArgument(EAR::Variant(testIntArg)).addArgument(EAR::Variant(testFloatArg));
        message.addArgument(EAR::Variant(testStringArg));

        bool boolArg;
        int intArg;
        float floatArg;
        std::string stringArg;

        //Create listener
        osc.addListener("/hello/world",
            [&boolArg, &intArg, &floatArg, &stringArg](const EAR::OSC::Message& message) {
                auto args = message.getArguments();

                CHECK(EAR::Variant::Type::Bool == args[0].getVariantType());
                boolArg = static_cast<bool>(args[0]);

                CHECK(EAR::Variant::Type::Int == args[1].getVariantType());
                intArg = static_cast<int>(args[1]);

                CHECK(EAR::Variant::Type::Float == args[2].getVariantType());
                floatArg = static_cast<float>(args[2]);

                CHECK(EAR::Variant::Type::String == args[3].getVariantType());
                stringArg = static_cast<std::string>(args[3]);
            });

        osc.send(message);

        // wait to be sure that the message was received
        using namespace std::chrono_literals;
        std::this_thread::sleep_for(0.5s);

        CHECK(testBoolArg == boolArg);
        CHECK(testIntArg == intArg);
        CHECK(testFloatArg == floatArg);
        CHECK(testStringArg == stringArg);
    }

    SUBCASE("Send and receive bundle")
    {
        EAR::OSC osc;

        //Create test variables
        std::string testStringMsg1Arg1 = "I live in message 1!";
        std::string testStringMsg1Arg2 = "I also live in message 1!";
        std::string testStringMsg1Arg3 = "I too live in message 1!";

        std::string testStringMsg2Arg1 = "I live in message 2!";
        std::string testStringMsg2Arg2 = "I also live in message 2!";

        std::string testStringMsg3Arg1 = "I live in message 3!";

        //Init messages and bundles
        EAR::OSC::Bundle masterBundle;
        EAR::OSC::Bundle nestedBundle;
        EAR::OSC::Message message1("/message/1");
        EAR::OSC::Message message2("/message/2");
        EAR::OSC::Message message3("/message/3");

        message1.addArgument(EAR::Variant(testStringMsg1Arg1))
            .addArgument(EAR::Variant(testStringMsg1Arg2))
            .addArgument(EAR::Variant(testStringMsg1Arg3));

        message2.addArgument(EAR::Variant(testStringMsg2Arg1))
            .addArgument(EAR::Variant(testStringMsg2Arg2));

        message3.addArgument(EAR::Variant(testStringMsg3Arg1));

        //Nest message 1 and 2 into one bundle and nest message 3 into the master bundle
        nestedBundle.addElement(message1).addElement(message2);

        masterBundle.addElement(nestedBundle).addElement(message3);

        std::string checkStringMsg1Arg1, checkStringMsg1Arg2, checkStringMsg1Arg3;
        std::string checkStringMsg2Arg1, checkStringMsg2Arg2;
        std::string checkStringMsg3Arg1;

        osc.addListener("/message/1", [&](const EAR::OSC::Message& message) {
            auto args = message.getArguments();

            checkStringMsg1Arg1 = static_cast<std::string>(args[0]);
            checkStringMsg1Arg2 = static_cast<std::string>(args[1]);
            checkStringMsg1Arg3 = static_cast<std::string>(args[2]);
        });

        osc.addListener("/message/2", [&](const EAR::OSC::Message& message) {
            auto args = message.getArguments();

            checkStringMsg2Arg1 = static_cast<std::string>(args[0]);
            checkStringMsg2Arg2 = static_cast<std::string>(args[1]);
        });

        osc.addListener("/message/3", [&](const EAR::OSC::Message& message) {
            auto args = message.getArguments();

            checkStringMsg3Arg1 = static_cast<std::string>(args[0]);
        });

        osc.send(masterBundle);

        using namespace std::chrono_literals;
        std::this_thread::sleep_for(0.5s);

        CHECK(testStringMsg1Arg1 == checkStringMsg1Arg1);
        CHECK(testStringMsg1Arg2 == checkStringMsg1Arg2);
        CHECK(testStringMsg1Arg3 == checkStringMsg1Arg3);

        CHECK(testStringMsg2Arg1 == checkStringMsg2Arg1);
        CHECK(testStringMsg2Arg2 == checkStringMsg2Arg2);

        CHECK(testStringMsg3Arg1 == checkStringMsg3Arg1);
    }
}
