/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include "PortalExtentTest.h"
#include <EAR/Box.h>
#include <EAR/Mesh.h>
#include <EAR/Portal.h>
#include <EAR/PortalReverbExtent.h>
#include <MPEGI/Bitstream/EIFPayload.h>
#include <doctest/doctest.h>
#include <vector>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif

#define LOG_TO_XML false

namespace logging {
bool writeMeshToEIF(FILE* EIF, const EAR::Mesh* mesh, std::string meshName)
{
    auto verts = mesh->getVertices(EAR::Mesh::CSpace::Local);
    auto tris = mesh->getTriangles();
    auto location = mesh->getLocation();
    auto orientation = mesh->getOrientation();
    fprintf(EIF, "<Mesh id=\"%s\" position=\"%f %f %f\" orientation=\"%f %f %f\">\n",
        meshName.c_str(), location[0], location[1], location[2], orientation[0], orientation[1],
        orientation[2]);

    for (int i = 0; i < verts.size(); ++i) {
        fprintf(EIF, "<Vertex index=\"%i\" position=\"%f %f %f\"/>\n", i, verts[i][0], verts[i][1],
            verts[i][2]);
    }

    for (int i = 0; i < tris.size(); ++i) {
        fprintf(EIF, "<Face index=\"%i\" vertices=\"%i %i %i\"/>\n", i, tris[i].v1, tris[i].v2,
            tris[i].v3);
    }
    fprintf(EIF, "</Mesh>\n");

    return true;
}

bool writePlaneToEIF(FILE* EIF, MPEGI::IntersectPlane plane, std::string planeName)
{
    fprintf(EIF, "<Mesh id=\"%s\" position=\"0 0 0\" orientation=\"0 0 0\" center=\"%f %f %f\">\n",
        planeName.c_str(), plane.planeCenter[0], plane.planeCenter[1], plane.planeCenter[2]);

    fprintf(EIF, "<Vertex index=\"%i\" position=\"%f %f %f\"/>\n", 0, plane.upperLeftCorner[0],
        plane.upperLeftCorner[1], plane.upperLeftCorner[2]);
    fprintf(EIF, "<Vertex index=\"%i\" position=\"%f %f %f\"/>\n", 1, plane.lowerLeftCorner[0],
        plane.lowerLeftCorner[1], plane.lowerLeftCorner[2]);
    fprintf(EIF, "<Vertex index=\"%i\" position=\"%f %f %f\"/>\n", 2, plane.lowerRightCorner[0],
        plane.lowerRightCorner[1], plane.lowerRightCorner[2]);
    fprintf(EIF, "<Vertex index=\"%i\" position=\"%f %f %f\"/>\n", 2, plane.upperRightCorner[0],
        plane.upperRightCorner[1], plane.upperRightCorner[2]);

    fprintf(EIF, "<Face index=\"%i\" vertices=\"%i %i %i\"/>\n", 0, 0, 1, 2);
    fprintf(EIF, "<Face index=\"%i\" vertices=\"%i %i %i\"/>\n", 1, 0, 2, 3);
    fprintf(EIF, "</Mesh>\n");

    return true;
}
} // namespace logging

void checkExtentDistance(const EAR::Mesh* portalMesh,
    std::shared_ptr<EAR::PortalReverbExtent> reverbExtent, float meshOffset)
{
    float portalDistance;
    auto portalExtentVerts = reverbExtent->getMesh()->getVertices(EAR::Mesh::CSpace::Global);
    for (auto vert : portalExtentVerts) {
        portalDistance = (portalMesh->closestPointTo(vert) - vert).abs();
        CHECK((std::abs(portalDistance - meshOffset) - 0.1f) < 0.075f);
    }
}

void checkExtentDistance(const EAR::Mesh* portalMesh,
    const std::array<std::shared_ptr<EAR::PortalReverbExtent>, 2>& reverbExtents,
    const std::array<float, 2>& meshOffsets)
{
    for (int AEid = 0; AEid < 2; ++AEid) {
        checkExtentDistance(portalMesh, reverbExtents[AEid], meshOffsets[AEid]);
    }
}

namespace MPEGI {

// Create public method from private method of base class
void DummySceneState::createBox2(const EIFPayload::BoxDesc& boxDesc)
{
    SceneState::createBox(boxDesc);
}

// Create public method from private method of base class
void DummySceneState::createPortal2(MPEGI::Portal* portal)
{
    SceneState::createPortal(portal);
}

// Create public method from private method of base class
void DummySceneState::createMesh2(const EIFPayload::MeshDesc& meshDesc)
{
    SceneState::createMesh(meshDesc);
}

DummySceneState::DummySceneState(bool hasDefaultAe)
{
    if (hasDefaultAe) {
        ReverbEnv env = ReverbEnv(DEFAULT_AE_ID_NAME, nullptr);
        env.isLsdf = false;
        env.isSceneBasedAE = true;
        env.predelay = 0;
        env.acousticEnvironmentBsId = DEFAULT_AE_ID;
        dummyReverbEnvs.push_back(std::make_unique<ReverbEnv>(std::move(env)));
    }
}

void DummySceneState::addAcousticEnvironment(
    std::optional<std::string> aeId, EAR::Geometry* enclosure)
{
    ReverbEnv env = ReverbEnv(aeId.value_or("0"), enclosure);
    env.isLsdf = false;
    env.isSceneBasedAE = true;
    env.predelay = 0;
    env.acousticEnvironmentBsId = std::stoi(aeId.value_or("0"));
    dummyReverbEnvs.push_back(std::make_unique<ReverbEnv>(std::move(env)));
}

TEST_CASE("Portal Reverb Extent tests")
{
    SUBCASE("Check portal extent normal vectors are consistent")
    {
        const EAR::Vector3 portalSize = { 1.2, 2, 0.2 };
        const EAR::Position portalPosition = { { -5.2, 1.0, 2.6 }, { 0, 0, 0 } };
        const EAR::Vector3 portalNormal = { 0, 0, -1 };
        for (int segments = 1; segments < 4; ++segments) {

            float offset = 1.0f;
            std::unique_ptr<EAR::PortalReverbExtent> portalExtent =
                EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                    portalSize, offset, "PortalReverbExtent", portalPosition, segments);

            const auto& triangles = portalExtent->getMesh()->getTriangles();
            REQUIRE(triangles.size() > 0);
            // All triangles should point in the same direction. Test to see that the inner product with the portalNormal always has the same sign.
            const bool signOfFirstTriangle = triangles[0].getNormal().dot(portalNormal) < 0;
            for (int triangleId = 1; triangleId < triangles.size(); ++triangleId) {
                CAPTURE(segments);
                CAPTURE(triangleId);
                const auto triangle = triangles[triangleId];
                CHECK_EQ(triangle.getNormal().dot(portalNormal) < 0, signOfFirstTriangle);
            }
        }
    }

    SUBCASE("Portal as Box")
    {
        const EAR::Vector3 portalPosition(0.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalVector(0.0f, 0.0f, 1.0f);
        const EAR::Vector3 portalBoxSize(2.0f, 2.0f, 0.1f);
        float offset = 1.f;
        auto portalPtr = EAR::SceneObjectFactory<EAR::Portal>::getNewObject("portal");
        auto extent = EAR::SceneObjectFactory<EAR::Box>::getNewObject(portalBoxSize, "extentBox");
        extent->setLocation(portalPosition);

        std::array<std::shared_ptr<PortalReverbExtent>, 2> externalReverbExtents;

        const EAR::Geometry::BoundingBox portalBox = extent->getBoundingBox();
        const EAR::Vector3 portalSize(portalBox[0].high - portalBox[0].low,
            portalBox[1].high - portalBox[1].low, portalBox[2].high - portalBox[2].low);

        EAR::Position extentPosition = extent->getPosition();

        {
            // portal->normalVector always points to portalEnvironments[0]. But we need to check that portalPosition.orientation points in the same direction as that is actually used for the mesh.
            const EAR::Vector3 extentNormal =
                EAR::Quaternion(portalPtr->getPosition().orientation)
                    .rotate(EAR::Vector3(0, 0, -1)); // (0, 0, -1) is "forward".
            if (extentNormal.dot(portalVector) < 0) {
                // Extent orientation and portal normalvector are in opposite directions. Switch extent orientation 180 degrees to fix.
                extentPosition.orientation[0] += 180.f;
            }

            auto reverbExtent1 = std::shared_ptr<EAR::PortalReverbExtent>(
                EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                    portalSize, offset, "PortalExtent1", extentPosition, 3));

            externalReverbExtents[0] = reverbExtent1;
        }

        // Create reverb extent in the second accoustic environment
        {
            extentPosition.orientation[0] += 180.f; // Opposite direction as the first extent

            auto reverbExtent2 = std::shared_ptr<EAR::PortalReverbExtent>(
                EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                    portalSize, offset, "PortalExtent2", extentPosition, 3));

            externalReverbExtents[1] = reverbExtent2;
        }

        portalPtr->setExternalReverbExtents(externalReverbExtents);

        CHECK(portalPtr->getLocation() == portalPosition);

        checkExtentDistance(extent->getMesh(), externalReverbExtents[0], offset);
        checkExtentDistance(extent->getMesh(), externalReverbExtents[1], offset);
        CHECK(
            (externalReverbExtents[0]->getCenterPosition() - EAR::Vector3(0.0f, 0.0f, 0.65f)).abs()
            < 0.001f);
        CHECK(
            (externalReverbExtents[1]->getCenterPosition() - EAR::Vector3(0.0f, 0.0f, -0.65f)).abs()
            < 0.001f);
    }

    SUBCASE("Portal as Box in Ceiling using SceneState")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsBoxInCeilingUsingScenestate.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(0.0f, 2.5f, 0.0f);
        const EAR::Vector3 portalVector(0.0f, -1.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(2.0f, 0.1, 2.0f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        const EAR::Position room1Pos(EAR::Vector3(0.0f, -roomSize / 2.f, 0.f), EAR::Euler());
        const EAR::Position room2Pos(EAR::Vector3(0.0f, roomSize / 2.f, 0.f), EAR::Euler());

        const EAR::Vector3 roomCube(roomSize, roomSize, roomSize);
        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 2", room2Pos);
        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        // Create portal extent
        EIFPayload::BoxDesc portalBox;
        portalBox.position = EAR::Position(portalPosition, EAR::Euler());
        portalBox.size = portalBoxSize;
        portalBox.id = "0";
        portalBox.bsId = 0;

        sceneState.createBox2(portalBox);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(
            portalLogFile, sceneState.getAllGeometry()[0]->getMesh(), "OriginalExtentMesh");
#endif

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[0], reverbExtentMeshOffset[0]);
        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[1], reverbExtentMeshOffset[1]);
        EAR::Vector3 calcFirstMeshCenter = portalPosition
            + (portalVector * (reverbExtentMeshOffset[0] / 2)) + EAR::Vector3(0.0f, -0.17f, 0.0f);
        EAR::Vector3 calcSecondMeshCenter = portalPosition
            - (portalVector * (reverbExtentMeshOffset[1] / 2)) - EAR::Vector3(0.0f, -0.17f, 0.0f);
        auto temp = reverbExtentMeshes[0]->getCenterPosition();
        CHECK((reverbExtentMeshes[0]->getCenterPosition() - calcFirstMeshCenter).abs() < 0.01f);
        CHECK((reverbExtentMeshes[1]->getCenterPosition() - calcSecondMeshCenter).abs() < 0.01f);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Portal From Hospital using SceneState")
    {
        // portal between nurse room and patient room in hospital
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalHospitalNurseroomUsingScenestate.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(-3.6f, 1.0f, 1.4f);
        const EAR::Vector3 portalVector(-1.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(1.2f, 2.0, 0.2f);
        const EAR::Euler portalRotation(90.f, 0.f, 0.f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        // nurse room
        const EAR::Position room1Pos(EAR::Vector3(-4.950f, 1.250f, 0.000f), EAR::Euler());
        const EAR::Vector3 room1Size(2.700f, 2.500f, 5.200f);

        // patient room
        const EAR::Position room2Pos(EAR::Vector3(0.000f, 1.250f, -0.000f), EAR::Euler());
        const EAR::Vector3 room2Size(7.200f, 2.500f, 5.200f);

        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room1Size, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room2Size, "Room 2", room2Pos);
        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        // Create portal extent
        EIFPayload::BoxDesc portalBox;
        portalBox.position = EAR::Position(portalPosition, portalRotation);
        portalBox.size = portalBoxSize;
        portalBox.id = "0";
        portalBox.bsId = 0;

        sceneState.createBox2(portalBox);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(
            portalLogFile, sceneState.getAllGeometry()[0]->getMesh(), "OriginalExtentMesh");
#endif

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
#if (LOG_TO_XML)
        logging::writePlaneToEIF(
            portalLogFile, portal->portalPlanes.at(portal->portalEnvironments[0]), "plane1");
        logging::writePlaneToEIF(portalLogFile,
            portal->portalPlanes.at(portal->portalEnvironments[1].value()), "plane2");
#endif

        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(portalBoxPtr->getMesh(), reverbExtentMeshes, reverbExtentMeshOffset);

        EAR::Vector3 lowerLeftCorner{ -3.7f, 0.0f, 2.0f };
        EAR::Vector3 lowerRightCorner{ -3.7f, 0.0f, 0.8f };
        EAR::Vector3 upperLeftCorner{ -3.7f, 2.0f, 2.0f };
        EAR::Vector3 upperRightCorner{ -3.7f, 2.0f, 0.8f };
        EAR::Vector3 planeCenter{ -3.7f, 1.0f, 1.4f };
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room1AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

        lowerLeftCorner.set(-3.5f, 0.0f, 2.0f);
        lowerRightCorner.set(-3.5f, 0.0f, 0.8f);
        upperLeftCorner.set(-3.5f, 2.0f, 2.0f);
        upperRightCorner.set(-3.5f, 2.0f, 0.8f);
        planeCenter.set(-3.5f, 1.0f, 1.4f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room2AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Patient room to corridor portal From Hospital using SceneState")
    {
        // portal between patient room and corridor in hospital
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalHospitalPatientRoomToCorridor.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(0.0f, 1.0f, 2.6f);
        const EAR::Vector3 portalVector(0.0f, 0.0f, -1.0f);
        const EAR::Vector3 portalBoxSize(2.4f, 2.0, 0.2f);
        const EAR::Euler portalRotation(0.f, 0.f, 0.f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        // Patient room
        const EAR::Position room1Pos(EAR::Vector3(0.0f, 1.250f, 0.000f), EAR::Euler());
        const EAR::Vector3 room1Size(7.200f, 2.500f, 5.200f);

        // Corridor
        const EAR::Position room2Pos(EAR::Vector3(0.000f, 1.250f, 3.700f), EAR::Euler());
        const EAR::Vector3 room2Size(40.000f, 2.500f, 2.200f);

        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room1Size, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room2Size, "Room 2", room2Pos);
        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        // Create portal extent
        EIFPayload::BoxDesc portalBox;
        portalBox.position = EAR::Position(portalPosition, portalRotation);
        portalBox.size = portalBoxSize;
        portalBox.id = "0";
        portalBox.bsId = 0;

        sceneState.createBox2(portalBox);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(
            portalLogFile, sceneState.getAllGeometry()[0]->getMesh(), "OriginalExtentMesh");
#endif

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments[0] = room1AcousticEnvironmentId;
        portal->portalEnvironments[1] = room2AcousticEnvironmentId;
        portal->extentId = portalBox.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
#if (LOG_TO_XML)
        logging::writePlaneToEIF(
            portalLogFile, portal->portalPlanes.at(portal->portalEnvironments[0]), "plane1");
        logging::writePlaneToEIF(portalLogFile,
            portal->portalPlanes.at(portal->portalEnvironments[1].value()), "plane2");
#endif

        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[0], reverbExtentMeshOffset[0]);
        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[1], reverbExtentMeshOffset[1]);

        EAR::Vector3 lowerLeftCorner{ -1.2f, 0.0f, 2.5f };
        EAR::Vector3 lowerRightCorner{ 1.2f, 0.0f, 2.5f };
        EAR::Vector3 upperLeftCorner{ -1.2f, 2.0f, 2.5f };
        EAR::Vector3 upperRightCorner{ 1.2f, 2.0f, 2.5f };
        EAR::Vector3 planeCenter{ 0.0f, 1.0f, 2.5f };
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room1AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

        lowerLeftCorner.set(-1.2f, 0.0f, 2.7f);
        lowerRightCorner.set(1.2f, 0.0f, 2.7f);
        upperLeftCorner.set(-1.2f, 2.0f, 2.7f);
        upperRightCorner.set(1.2f, 2.0f, 2.7f);
        planeCenter.set(0.0f, 1.0f, 2.7f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room2AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Patient room to corridor portal From Hospital using SceneState as explicit box")
    {
        // portal between patient room and corridor in hospital
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalHospitalPatientRoomToCorridorExplicitBox.xml",
            "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(0.0f, 1.0f, 2.6f);
        const EAR::Vector3 portalVector(0.0f, 0.0f, -1.0f);
        const EAR::Vector3 portalBoxSize(2.4f, 2.0, 0.2f);
        const EAR::Euler portalRotation(0.f, 0.f, 0.f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        // Patient room
        const EAR::Position room1Pos(EAR::Vector3(0.0f, 1.250f, 0.000f), EAR::Euler());
        const EAR::Vector3 room1Size(7.200f, 2.500f, 5.200f);

        // Corridor
        const EAR::Position room2Pos(EAR::Vector3(0.000f, 1.250f, 3.700f), EAR::Euler());
        const EAR::Vector3 room2Size(40.000f, 2.500f, 2.200f);

        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room1Size, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room2Size, "Room 2", room2Pos);
        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments[0] = room1AcousticEnvironmentId;
        portal->portalEnvironments[1] = room2AcousticEnvironmentId;
        portal->normalVector = portalVector;
        portal->position = portalPosition;
        portal->equivalentPortalSize = portalBoxSize;

        sceneState.createPortal2(portal.get());
#if (LOG_TO_XML)
        logging::writePlaneToEIF(
            portalLogFile, portal->portalPlanes.at(portal->portalEnvironments[0]), "plane1");
        logging::writePlaneToEIF(portalLogFile,
            portal->portalPlanes.at(portal->portalEnvironments[1].value()), "plane2");
#endif

        portalPayload->addPortal(std::move(portal));

        // const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        const EAR::Geometry* portalBoxPtr =
            portalPayload->getPortals()[0]->equivalentBoxPointer.get();
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[0], reverbExtentMeshOffset[0]);
        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[1], reverbExtentMeshOffset[1]);

        EAR::Vector3 lowerLeftCorner{ -1.2f, 0.0f, 2.5f };
        EAR::Vector3 lowerRightCorner{ 1.2f, 0.0f, 2.5f };
        EAR::Vector3 upperLeftCorner{ -1.2f, 2.0f, 2.5f };
        EAR::Vector3 upperRightCorner{ 1.2f, 2.0f, 2.5f };
        EAR::Vector3 planeCenter{ 0.0f, 1.0f, 2.5f };
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room1AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

        lowerLeftCorner.set(-1.2f, 0.0f, 2.7f);
        lowerRightCorner.set(1.2f, 0.0f, 2.7f);
        upperLeftCorner.set(-1.2f, 2.0f, 2.7f);
        upperRightCorner.set(1.2f, 2.0f, 2.7f);
        planeCenter.set(0.0f, 1.0f, 2.7f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room2AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Portal as Box using SceneState")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsBoxUsingScenestate.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(0.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalVector(-1.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(0.1f, 2.0f, 2.0f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        const EAR::Position room1Pos(EAR::Vector3(-roomSize / 2.f, 0.f, 0.f), EAR::Euler());
        const EAR::Position room2Pos(EAR::Vector3(roomSize / 2.f, 0.f, 0.f), EAR::Euler());

        const EAR::Vector3 roomCube(roomSize, roomSize, roomSize);
        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 2", room2Pos);
        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        // Create portal extent
        EIFPayload::BoxDesc portalBox;
        portalBox.position = EAR::Position(portalPosition, EAR::Euler());
        portalBox.size = portalBoxSize;
        portalBox.id = "0";
        portalBox.bsId = 0;

        sceneState.createBox2(portalBox);

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(portalBoxPtr->getMesh(), reverbExtentMeshes, reverbExtentMeshOffset);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Portal as explicit defined Box using SceneState")
    {
        // portal from nurse room to patient room
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsExplicitBoxUsingScenestate.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(-3.6f, 1.0f, 1.4f);
        const EAR::Vector3 portalVector(-1.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(1.2f, 2.0, 0.2f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        // nurse room
        const EAR::Position room1Pos(EAR::Vector3(-4.950f, 1.250f, 0.000f), EAR::Euler());
        const EAR::Vector3 room1Size(2.700f, 2.500f, 5.200f);

        // patient room
        const EAR::Position room2Pos(EAR::Vector3(0.000f, 1.250f, -0.000f), EAR::Euler());
        const EAR::Vector3 room2Size(7.200f, 2.500f, 5.200f);

        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room1Size, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room2Size, "Room 2", room2Pos);

        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments[0] = room1AcousticEnvironmentId;
        portal->portalEnvironments[1] = room2AcousticEnvironmentId;

        portal->normalVector = portalVector;
        portal->position = portalPosition;
        portal->equivalentPortalSize = portalBoxSize;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr =
            portalPayload->getPortals()[0]->equivalentBoxPointer.get();
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[0], reverbExtentMeshOffset[0]);
        checkExtentDistance(
            portalBoxPtr->getMesh(), reverbExtentMeshes[1], reverbExtentMeshOffset[1]);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Cathedral pre-room to main room as Box")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsBox.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const EAR::Vector3 portalBoxPosition(-17.62f, 2.7881f, -20.012f);
        const EAR::Euler portalBoxOrientation(90.f, 0.f, 0.f);
        EAR::Vector3 portalNormal(-1.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(4.95f, 6.05f, 1.33f);
        float offset = 1.0f;

        auto portalPtr = EAR::SceneObjectFactory<EAR::Portal>::getNewObject("portal");
        auto extent = EAR::SceneObjectFactory<EAR::Box>::getNewObject(portalBoxSize, "extentBox");
        extent->setLocation(portalBoxPosition);
        extent->setOrientation(portalBoxOrientation);

        EAR::Geometry::BoundingBox portalBox = extent->getGlobalBoundingBox();
        EAR::Vector3 portalLocation; //centre point of the portal, before axis alignment.
        portalLocation[0] = portalBox[0].low + (portalBox[0].high - portalBox[0].low) / 2;
        portalLocation[1] = portalBox[1].low + (portalBox[1].high - portalBox[1].low) / 2;
        portalLocation[2] = portalBox[2].low + (portalBox[2].high - portalBox[2].low) / 2;
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, extent->getMesh(), "extentMeshBeforeRotation");
#endif
        EAR::Vector3 targetVector(0.f, 0.f, -1.f);
        EAR::HomogenMatrix rotMatrix = EAR::rotateVectorTo(portalNormal, targetVector);
        EAR::Euler currRotation = extent->getOrientation();
        extent->setOrientation(currRotation + rotMatrix.getRotation());
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, extent->getMesh(), "extentMeshAfterRotation");
#endif
        portalBox = extent->getGlobalBoundingBox();
        EAR::Vector3 portalSize;
        portalSize[0] = portalBox[0].high - portalBox[0].low;
        portalSize[1] = portalBox[1].high - portalBox[1].low;
        portalSize[2] = portalBox[2].high - portalBox[2].low;

        EAR::Position portalPosition(portalLocation, rotMatrix.inverse().getRotation());

        auto reverbExtent1 = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalSize, offset, "PortalExtent1", portalPosition, 3));
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, reverbExtent1->getMesh(), "reverbExtent1");
#endif
        portalPosition.orientation[0] += 180.f;
        auto reverbExtent2 = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalSize, offset, "PortalExtent2", portalPosition, 3));
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, reverbExtent2->getMesh(), "reverbExtent2");
#endif
        //return extent to original location
        extent->setOrientation(currRotation);
        checkExtentDistance(extent->getMesh(), reverbExtent1, offset);
        checkExtentDistance(extent->getMesh(), reverbExtent2, offset);

#if (LOG_TO_XML)
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Cathedral pre-room to main room as mesh")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsMesh.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        // data from EIF_Cathedral with mesh-based portals
        auto portalPtr = EAR::SceneObjectFactory<EAR::Portal>::getNewObject("portal");
        const EAR::Vector3 portalNormal(-1.0f, 0.0f, 0.0f);

        float offset = 1.f;
        const float backfaceOffset = 0.1f;

        const EAR::Vector3 portalMeshLocation(-0.02f, -0.04f, -0.13f);
        const EAR::Euler portalMeshOrientation(0.00f, 89.98f, 0.00f);
        auto extent = EAR::SceneObjectFactory<EAR::Mesh>::getNewObject();

        int v0 = extent->addVertex(EAR::Vector3(-16.9335f, -22.3535f, -5.862f));
        int v1 = extent->addVertex(EAR::Vector3(-18.2657f, -22.3535f, -5.862f));
        int v2 = extent->addVertex(EAR::Vector3(-18.2657f, -22.3535f, 0.192f));
        int v3 = extent->addVertex(EAR::Vector3(-16.9335f, -22.3535f, 0.192f));
        int v4 = extent->addVertex(EAR::Vector3(-18.2657f, -17.4082f, -5.862f));
        int v5 = extent->addVertex(EAR::Vector3(-16.9335f, -17.4082f, -5.862f));
        int v6 = extent->addVertex(EAR::Vector3(-16.9335f, -17.4082f, 0.192f));
        int v7 = extent->addVertex(EAR::Vector3(-18.2657f, -17.4083f, 0.192f));

        extent->addTriangle(v2, v1, v0);
        extent->addTriangle(v3, v2, v0);
        extent->addTriangle(v6, v5, v4);
        extent->addTriangle(v7, v6, v4);
        extent->addTriangle(v0, v1, v4);
        extent->addTriangle(v5, v0, v4);
        extent->addTriangle(v0, v5, v6);
        extent->addTriangle(v3, v0, v6);
        extent->addTriangle(v7, v4, v1);
        extent->addTriangle(v2, v7, v1);
        extent->addTriangle(v3, v6, v7);
        extent->addTriangle(v2, v3, v7);

        extent->setPosition(EAR::Position(portalMeshLocation, portalMeshOrientation));

        extent->lock();
        extent->unlock();

        EAR::Geometry::BoundingBox portalBox = extent->getGlobalBoundingBox();
        EAR::Vector3 portalLocation; //centre point of the portal, before axis alignment.
        portalLocation[0] = portalBox[0].low + (portalBox[0].high - portalBox[0].low) / 2;
        portalLocation[1] = portalBox[1].low + (portalBox[1].high - portalBox[1].low) / 2;
        portalLocation[2] = portalBox[2].low + (portalBox[2].high - portalBox[2].low) / 2;

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, extent->getMesh(), "extentMeshBeforeRotation");
#endif

        // rotate the portal so that the portal normal is pointing forward.
        EAR::Vector3 targetVector(0.f, 0.f, -1.f);

        EAR::HomogenMatrix rotMatrix = EAR::rotateVectorTo(portalNormal, targetVector);
        EAR::Euler currRotation = extent->getOrientation();
        extent->setOrientation(currRotation + rotMatrix.getRotation());
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, extent->getMesh(), "extentMeshAfterRotMatrix");
#endif
        portalBox = extent->getGlobalBoundingBox();
        EAR::Vector3 portalSize;
        portalSize[0] = portalBox[0].high - portalBox[0].low;
        portalSize[1] = portalBox[1].high - portalBox[1].low;
        portalSize[2] = portalBox[2].high - portalBox[2].low;

        EAR::Position portalPosition(portalLocation, rotMatrix.inverse().getRotation());

        auto reverbExtent1 = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalSize, offset, "PortalExtent1", portalPosition, 3));
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, reverbExtent1->getMesh(), "reverbExtent1");
#endif
        portalPosition.orientation[0] += 180.f;
        auto reverbExtent2 = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalSize, offset, "PortalExtent2", portalPosition, 3));
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, reverbExtent2->getMesh(), "reverbExtent2");
#endif
        //return extent to original location
        extent->setOrientation(currRotation);

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, extent->getMesh(), "extentMeshBeforeTesting");
#endif

        checkExtentDistance(extent->getMesh(), reverbExtent1, offset);
        checkExtentDistance(extent->getMesh(), reverbExtent2, offset);

#if (LOG_TO_XML)
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    // TODO: Re-enable below test once CreatePortals in SceneState uses segmented meshes. The verts of the 0 segment mesh are not uniformly distant from the portal, so the below fails.
    SUBCASE("Cathedral pre-room to main room as mesh using SceneState")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsMeshUsingScenestate.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalMeshLocation(-0.02f, -0.04f, -0.13f);
        const EAR::Euler portalMeshOrientation(0.00f, 89.98f, 0.00f);
        const EAR::Position portalMeshPosition(portalMeshLocation, portalMeshOrientation);
        const EAR::Vector3 portalVector(-1.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(0.1f, 2.0f, 2.0f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        const EAR::Position room1Pos(EAR::Vector3(-roomSize / 2.f, 0.f, 0.f), EAR::Euler());
        const EAR::Position room2Pos(EAR::Vector3(roomSize / 2.f, 0.f, 0.f), EAR::Euler());

        const EAR::Vector3 roomCube(roomSize, roomSize, roomSize);
        auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 1", room1Pos);
        auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 2", room2Pos);
        sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
        sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

        EIFPayload::MeshDesc portalMesh({});
        EIFPayload::MeshDesc::VertexDesc portalVerts({});
        portalMesh.position = portalMeshPosition;
        portalMesh.id = "0";
        portalMesh.bsId = 0;
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 0, { -16.9335f, -22.3535f, -5.862f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 1, { -18.2657f, -22.3535f, -5.862f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 2, { -18.2657f, -22.3535f, 0.192f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 3, { -16.9335f, -22.3535f, 0.192f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 4, { -18.2657f, -17.4082f, -5.862f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 5, { -16.9335f, -17.4082f, -5.862f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 6, { -16.9335f, -17.4082f, 0.192f } });
        portalMesh.vertices.emplace_back(
            EIFPayload::MeshDesc::VertexDesc{ 7, { -18.2657f, -17.4083f, 0.192f } });

        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 0, 2, 1, 0, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 1, 3, 2, 0, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 2, 6, 5, 4, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 3, 7, 6, 4, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 4, 0, 1, 4, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 5, 5, 0, 4, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 6, 0, 5, 6, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 7, 3, 0, 6, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 8, 7, 4, 1, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 9, 2, 7, 1, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 10, 3, 6, 7, nullptr });
        portalMesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{ 11, 2, 3, 7, nullptr });

        sceneState.createMesh2(portalMesh);

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalMesh.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(portalBoxPtr->getMesh(), reverbExtentMeshes, reverbExtentMeshOffset);

        EAR::Vector3 lowerLeftCorner{ -18.286f, -0.23894f, -17.539f };
        EAR::Vector3 lowerRightCorner{ -18.286f, -0.23894f, -22.484f };
        EAR::Vector3 upperLeftCorner{ -18.286f, 5.8151f, -17.539f };
        EAR::Vector3 upperRightCorner{ -18.286f, 5.8151f, -22.484f };
        EAR::Vector3 planeCenter{ -18.286f, 2.7881f, -20.012f };
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room1AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room1AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

        lowerLeftCorner.set(-16.953f, -0.23894f, -17.539f);
        lowerRightCorner.set(-16.953f, -0.23894f, -22.484f);
        upperLeftCorner.set(-16.953f, 5.8151f, -17.539f);
        upperRightCorner.set(-16.953f, 5.8151f, -22.484f);
        planeCenter.set(-16.953f, 2.7881f, -20.012f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerLeftCorner
                  - lowerLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .lowerRightCorner
                  - lowerRightCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperLeftCorner
                  - upperLeftCorner)
                  .abs()
            < 0.001f);
        CHECK((portalPayload->getPortals()[0]
                      ->portalPlanes.at(room2AcousticEnvironmentId)
                      .upperRightCorner
                  - upperRightCorner)
                  .abs()
            < 0.001f);
        CHECK(
            (portalPayload->getPortals()[0]->portalPlanes.at(room2AcousticEnvironmentId).planeCenter
                - planeCenter)
                .abs()
            < 0.001f);

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        logging::writePlaneToEIF(portalLogFile,
            portalPayload->getPortals()[0]->portalPlanes.at(room1AcousticEnvironmentId),
            room1AcousticEnvironmentId);
        logging::writePlaneToEIF(portalLogFile,
            portalPayload->getPortals()[0]->portalPlanes.at(room2AcousticEnvironmentId),
            room2AcousticEnvironmentId);
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Cathedral pre-room to main room box closest point test")
    {
        const std::vector<std::pair<EAR::Vector3, EAR::Vector3>> testPositions = {
            { { -12.000000, 3.300000, -22.500000 }, { -15.955001, 3.300000, -22.487001 } },
            { { -12.092000, 3.300000, -22.480000 }, { -15.955001, 3.300000, -22.480000 } },
            { { -12.184000, 3.300000, -22.460000 }, { -15.955001, 3.300000, -22.459999 } },
            { { -12.276000, 3.300000, -22.440000 }, { -15.955001, 3.300000, -22.440001 } },
            { { -12.368000, 3.300000, -22.420000 }, { -15.955001, 3.300000, -22.420000 } },
            { { -12.460000, 3.300000, -22.400000 }, { -15.955001, 3.300000, -22.400000 } },
            { { -12.552000, 3.300000, -22.380000 }, { -15.955001, 3.300000, -22.379999 } },
            { { -12.644000, 3.300000, -22.360000 }, { -15.955001, 3.300000, -22.360001 } },
            { { -12.736000, 3.300000, -22.340000 }, { -15.955001, 3.300000, -22.340000 } },
            { { -12.828000, 3.300000, -22.320000 }, { -15.955001, 3.300000, -22.320000 } },
            { { -12.920000, 3.300000, -22.300000 }, { -15.955001, 3.300000, -22.299999 } },
            { { -13.012000, 3.300000, -22.280000 }, { -15.955001, 3.300000, -22.280001 } },
            { { -13.104000, 3.300000, -22.260000 }, { -15.955001, 3.300000, -22.260000 } },
            { { -13.196000, 3.300000, -22.240000 }, { -15.955001, 3.300000, -22.240000 } },
            { { -13.288000, 3.300000, -22.220000 }, { -15.955001, 3.300000, -22.219999 } },
            { { -13.380000, 3.300000, -22.200000 }, { -15.955001, 3.300000, -22.200001 } },
            { { -13.472000, 3.300000, -22.180000 }, { -15.955001, 3.300000, -22.180000 } },
            { { -13.564000, 3.300000, -22.160000 }, { -15.955001, 3.300000, -22.160000 } },
            { { -13.656000, 3.300000, -22.140000 }, { -15.955001, 3.300000, -22.139999 } },
            { { -13.748000, 3.300000, -22.120000 }, { -15.955001, 3.300000, -22.120001 } },
            { { -13.840000, 3.300000, -22.100000 }, { -15.955001, 3.300000, -22.100000 } },
            { { -13.932000, 3.300000, -22.080000 }, { -15.955001, 3.300000, -22.080000 } },
            { { -14.024000, 3.300000, -22.060000 }, { -15.955001, 3.300000, -22.059999 } },
            { { -14.116000, 3.300000, -22.040000 }, { -15.955001, 3.300000, -22.040001 } },
            { { -14.208000, 3.300000, -22.020000 }, { -15.955001, 3.300000, -22.020000 } },
            { { -14.300000, 3.300000, -22.000000 }, { -15.955001, 3.300000, -22.000000 } },
            { { -14.392000, 3.300000, -21.980000 }, { -15.955001, 3.300000, -21.980000 } },
            { { -14.484000, 3.300000, -21.960000 }, { -15.955001, 3.300000, -21.959999 } },
            { { -14.576000, 3.300000, -21.940000 }, { -15.955001, 3.300000, -21.940001 } },
            { { -14.668000, 3.300000, -21.920000 }, { -15.955001, 3.300000, -21.920000 } },
            { { -14.760000, 3.300000, -21.900000 }, { -15.955001, 3.300000, -21.900000 } },
            { { -14.852000, 3.300000, -21.880000 }, { -15.955001, 3.300000, -21.879999 } },
            { { -14.944000, 3.300000, -21.860000 }, { -15.955001, 3.300000, -21.860001 } },
            { { -15.036000, 3.300000, -21.840000 }, { -15.955001, 3.300000, -21.840000 } },
            { { -15.128000, 3.300000, -21.820000 }, { -15.955001, 3.300000, -21.820000 } },
            { { -15.220000, 3.300000, -21.800000 }, { -15.955001, 3.300000, -21.799999 } },
            { { -15.312000, 3.300000, -21.780000 }, { -15.955001, 3.300000, -21.780001 } },
            { { -15.404000, 3.300000, -21.760000 }, { -15.955001, 3.300000, -21.760000 } },
            { { -15.496000, 3.300000, -21.740000 }, { -15.955001, 3.300000, -21.740000 } },
            { { -15.588000, 3.300000, -21.720000 }, { -15.955001, 3.300000, -21.719999 } },
            { { -15.680000, 3.300000, -21.700000 }, { -15.955001, 3.300000, -21.700001 } },
            { { -15.772000, 3.300000, -21.680000 }, { -15.955001, 3.300000, -21.680000 } },
            { { -15.864000, 3.300000, -21.660000 }, { -15.955001, 3.300000, -21.660000 } },
            { { -15.956000, 3.300000, -21.640000 }, { -15.955001, 3.300000, -21.639999 } },
            { { -16.048000, 3.300000, -21.620000 }, { -15.955001, 3.300000, -21.620001 } },
            { { -16.140000, 3.300000, -21.600000 }, { -15.955001, 3.300000, -21.600000 } },
            { { -16.232000, 3.300000, -21.580000 }, { -15.955001, 3.300000, -21.580000 } },
            { { -16.324000, 3.300000, -21.560000 }, { -15.955001, 3.300000, -21.559999 } },
            { { -16.416000, 3.300000, -21.540000 }, { -15.955001, 3.300000, -21.540001 } },
            { { -16.508000, 3.300000, -21.520000 }, { -15.955001, 3.300000, -21.520000 } },
            { { -16.600000, 3.300000, -21.500000 }, { -15.955001, 3.300000, -21.500000 } },
            { { -16.692000, 3.300000, -21.480000 }, { -15.955001, 3.300000, -21.480000 } },
            { { -16.784000, 3.300000, -21.460000 }, { -15.955001, 3.300000, -21.459999 } },
            { { -16.876000, 3.300000, -21.440000 }, { -15.955001, 3.300000, -21.440001 } },
            { { -16.968000, 3.300000, -21.420000 }, { -15.955001, 3.300000, -21.420000 } },
            { { -17.060000, 3.300000, -21.400000 }, { -15.955001, 3.300000, -21.400000 } },
            { { -17.152000, 3.300000, -21.380000 }, { -15.955001, 3.300000, -21.379999 } },
            { { -17.244000, 3.300000, -21.360000 }, { -15.955001, 3.300000, -21.360001 } },
            { { -17.336000, 3.300000, -21.340000 }, { -15.955001, 3.300000, -21.340000 } },
            { { -17.428000, 3.300000, -21.320000 }, { -15.955001, 3.300000, -21.320000 } },
            { { -17.520000, 3.300000, -21.300000 }, { -15.955001, 3.300000, -21.299999 } },
            { { -17.612000, 3.300000, -21.280000 }, { -15.955001, 3.300000, -21.280001 } },
            { { -17.704000, 3.300000, -21.260000 }, { -15.955001, 3.300000, -21.260000 } },
            { { -17.796000, 3.300000, -21.240000 }, { -15.955001, 3.300000, -21.240000 } },
            { { -17.888000, 3.300000, -21.220000 }, { -15.955001, 3.300000, -21.219999 } },
            { { -17.980000, 3.300000, -21.200000 }, { -15.955001, 3.300000, -21.200001 } },
            { { -18.072000, 3.300000, -21.180000 }, { -15.955001, 3.300000, -21.180000 } },
            { { -18.164000, 3.300000, -21.160000 }, { -15.955001, 3.300000, -21.160000 } },
            { { -18.256000, 3.300000, -21.140000 }, { -15.955001, 3.300000, -21.139999 } },
            { { -18.348000, 3.300000, -21.120000 }, { -15.955001, 3.300000, -21.120001 } },
            { { -18.440000, 3.300000, -21.100000 }, { -15.955001, 3.300000, -21.100000 } },
            { { -18.532000, 3.300000, -21.080000 }, { -15.955001, 3.300000, -21.080000 } },
            { { -18.624000, 3.300000, -21.060000 }, { -15.955001, 3.300000, -21.059999 } },
            { { -18.716000, 3.300000, -21.040000 }, { -15.955001, 3.300000, -21.040001 } },
            { { -18.808000, 3.300000, -21.020000 }, { -15.955001, 3.300000, -21.020000 } },
            { { -18.900000, 3.300000, -21.000000 }, { -15.955001, 3.300000, -21.000000 } },
            { { -18.992000, 3.300000, -20.980000 }, { -15.955001, 3.300000, -20.980000 } },
            { { -19.084000, 3.300000, -20.960000 }, { -15.955001, 3.300000, -20.959999 } },
            { { -19.176000, 3.300000, -20.940000 }, { -15.955001, 3.300000, -20.940001 } },
            { { -19.268000, 3.300000, -20.920000 }, { -15.955001, 3.300000, -20.920000 } },
            { { -19.360000, 3.300000, -20.900000 }, { -15.955001, 3.300000, -20.900000 } },
            { { -19.452000, 3.300000, -20.880000 }, { -15.955001, 3.300000, -20.879999 } },
            { { -19.544000, 3.300000, -20.860000 }, { -15.955001, 3.300000, -20.860001 } },
            { { -19.636000, 3.300000, -20.840000 }, { -15.955001, 3.300000, -20.840000 } },
            { { -19.728000, 3.300000, -20.820000 }, { -15.955001, 3.300000, -20.820000 } },
            { { -19.820000, 3.300000, -20.800000 }, { -15.955001, 3.300000, -20.799999 } },
            { { -19.912000, 3.300000, -20.780000 }, { -15.955001, 3.300000, -20.780001 } },
            { { -20.004000, 3.300000, -20.760000 }, { -15.955001, 3.300000, -20.760000 } },
            { { -20.096000, 3.300000, -20.740000 }, { -15.955001, 3.300000, -20.740000 } },
            { { -20.188000, 3.300000, -20.720000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.280000, 3.300000, -20.700000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.372000, 3.300000, -20.680000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.464000, 3.300000, -20.660000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.556000, 3.300000, -20.640000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.648000, 3.300000, -20.620000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.740000, 3.300000, -20.600000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.832000, 3.300000, -20.580000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -20.924000, 3.300000, -20.560000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.016000, 3.300000, -20.540000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.108000, 3.300000, -20.520000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.200000, 3.300000, -20.500000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.292000, 3.300000, -20.480000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.384000, 3.300000, -20.460000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.476000, 3.300000, -20.440000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.568000, 3.300000, -20.420000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.660000, 3.300000, -20.400000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.752000, 3.300000, -20.380000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.844000, 3.300000, -20.360000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -21.936000, 3.300000, -20.340000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.028000, 3.300000, -20.320000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.120000, 3.300000, -20.300000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.212000, 3.300000, -20.280000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.304000, 3.300000, -20.260000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.396000, 3.300000, -20.240000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.488000, 3.300000, -20.220000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.580000, 3.300000, -20.200000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.672000, 3.300000, -20.180000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.764000, 3.300000, -20.160000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.856000, 3.300000, -20.140000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -22.948000, 3.300000, -20.120000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -23.040000, 3.300000, -20.100000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -23.132000, 3.300000, -20.080000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -23.224000, 3.300000, -20.060000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -23.316000, 3.300000, -20.040000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -23.408000, 3.300000, -20.020000 }, { -16.955002, 3.300000, -23.487001 } },
            { { -23.500000, 3.300000, -20.000000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -23.592000, 3.300000, -19.980000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -23.684000, 3.300000, -19.960000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -23.776000, 3.300000, -19.940000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -23.868000, 3.300000, -19.920000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -23.960000, 3.300000, -19.900000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.052000, 3.300000, -19.880000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.144000, 3.300000, -19.860000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.236000, 3.300000, -19.840000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.328000, 3.300000, -19.820000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.420000, 3.300000, -19.800000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.512000, 3.300000, -19.780000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.604000, 3.300000, -19.760000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.696000, 3.300000, -19.740000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.788000, 3.300000, -19.720000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.880000, 3.300000, -19.700000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -24.972000, 3.300000, -19.680000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.064000, 3.300000, -19.660000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.156000, 3.300000, -19.640000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.248000, 3.300000, -19.620000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.340000, 3.300000, -19.600000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.432000, 3.300000, -19.580000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.524000, 3.300000, -19.560000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.616000, 3.300000, -19.540000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.708000, 3.300000, -19.520000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.800000, 3.300000, -19.500000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.892000, 3.300000, -19.480000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -25.984000, 3.300000, -19.460000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.076000, 3.300000, -19.440000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.168000, 3.300000, -19.420000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.260000, 3.300000, -19.400000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.352000, 3.300000, -19.380000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.444000, 3.300000, -19.360000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.536000, 3.300000, -19.340000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.628000, 3.300000, -19.320000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.720000, 3.300000, -19.300000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.812000, 3.300000, -19.280000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.904000, 3.300000, -19.260000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -26.996000, 3.300000, -19.240000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.088000, 3.300000, -19.220000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.180000, 3.300000, -19.200000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.272000, 3.300000, -19.180000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.364000, 3.300000, -19.160000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.456000, 3.300000, -19.140000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.548000, 3.300000, -19.120000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.640000, 3.300000, -19.100000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.732000, 3.300000, -19.080000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.824000, 3.300000, -19.060000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -27.916000, 3.300000, -19.040000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.008000, 3.300000, -19.020000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.100000, 3.300000, -19.000000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.192000, 3.300000, -18.980000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.284000, 3.300000, -18.960000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.376000, 3.300000, -18.940000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.468000, 3.300000, -18.920000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.560000, 3.300000, -18.900000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.652000, 3.300000, -18.880000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.744000, 3.300000, -18.860000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.836000, 3.300000, -18.840000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -28.928000, 3.300000, -18.820000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.020000, 3.300000, -18.800000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.112000, 3.300000, -18.780000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.204000, 3.300000, -18.760000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.296000, 3.300000, -18.740000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.388000, 3.300000, -18.720000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.480000, 3.300000, -18.700000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.572000, 3.300000, -18.680000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.664000, 3.300000, -18.660000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.756000, 3.300000, -18.640000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.848000, 3.300000, -18.620000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -29.940000, 3.300000, -18.600000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.032000, 3.300000, -18.580000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.124000, 3.300000, -18.560000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.216000, 3.300000, -18.540000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.308000, 3.300000, -18.520000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.400000, 3.300000, -18.500000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.492000, 3.300000, -18.480000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.584000, 3.300000, -18.460000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.676000, 3.300000, -18.440000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.768000, 3.300000, -18.420000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.860000, 3.300000, -18.400000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -30.952000, 3.300000, -18.380000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.044000, 3.300000, -18.360000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.136000, 3.300000, -18.340000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.228000, 3.300000, -18.320000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.320000, 3.300000, -18.300000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.412000, 3.300000, -18.280000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.504000, 3.300000, -18.260000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.596000, 3.300000, -18.240000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.688000, 3.300000, -18.220000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.780000, 3.300000, -18.200000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.872000, 3.300000, -18.180000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -31.964000, 3.300000, -18.160000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.056000, 3.300000, -18.140000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.148000, 3.300000, -18.120000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.240000, 3.300000, -18.100000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.332000, 3.300000, -18.080000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.424000, 3.300000, -18.060000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.516000, 3.300000, -18.040000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.608000, 3.300000, -18.020000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.700000, 3.300000, -18.000000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.792000, 3.300000, -17.980000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.884000, 3.300000, -17.960000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -32.976000, 3.300000, -17.940000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.068000, 3.300000, -17.920000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.160000, 3.300000, -17.900000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.252000, 3.300000, -17.880000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.344000, 3.300000, -17.860000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.436000, 3.300000, -17.840000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.528000, 3.300000, -17.820000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.620000, 3.300000, -17.800000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.712000, 3.300000, -17.780000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.804000, 3.300000, -17.760000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.896000, 3.300000, -17.740000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -33.988000, 3.300000, -17.720000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.080000, 3.300000, -17.700000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.172000, 3.300000, -17.680000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.264000, 3.300000, -17.660000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.356000, 3.300000, -17.640000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.448000, 3.300000, -17.620000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.540000, 3.300000, -17.600000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.632000, 3.300000, -17.580000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.724000, 3.300000, -17.560000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.816000, 3.300000, -17.540000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -34.908000, 3.300000, -17.520000 }, { -16.955000, 3.300000, -16.537001 } },
            { { -35.000000, 3.300000, -17.500000 }, { -16.955000, 3.300000, -16.537001 } },
        };
        const EAR::Vector3 portalBoxSize(4.95, 6.05f, 1.33f);
        const EAR::Vector3 portalBoxPosition(-17.62f, 2.7881f, -20.012f);
        const EAR::Euler portalBoxRotation(90.f, 0.f, 0.f);

        float offset = 1.f;

        EAR::Position extentPosition(portalBoxPosition, portalBoxRotation);

        // Create reverb extent
        extentPosition.orientation[0] += 180.f; // Opposite direction to the portal normal

        auto reverbExtent = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalBoxSize, offset, "PortalExtent2", extentPosition, 3));

        // externalReverbExtents.second = reverbExtent2;

        EAR::Vector3 calcClosestPos;
        for (auto [listPos, closestPos] : testPositions) {
            calcClosestPos = reverbExtent->getMesh()->closestPointTo(listPos);
            auto temp = (calcClosestPos - closestPos).abs();
            CHECK(temp < 0.101f);
        }
    }

    SUBCASE("Portal as Box with constraing geometry")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsBoxWithConstraints.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float epsilon = 1e-9f;
        const EAR::Vector3 portalBoxPosition(2.0f, -3.0f, 5.0f);
        const EAR::Euler portalBoxOrientation(0.f, 0.f, 0.f);
        EAR::Vector3 portalNormal(-1.0f, 0.0f, 0.0f);
        const EAR::Vector3 portalBoxSize(.1f, 2.0f, 1.0f);
        float offset = 1.0f;

        auto portalPtr = EAR::SceneObjectFactory<EAR::Portal>::getNewObject("portal");
        auto extent = EAR::SceneObjectFactory<EAR::Box>::getNewObject(portalBoxSize, "extentBox");
        extent->setLocation(portalBoxPosition);
        extent->setOrientation(portalBoxOrientation);

        const EAR::Vector3 room1Size(4.0f, 2.5f, 4.0f);
        EAR::Vector3 room1Position(0.0f, -3.f, 5.0f);
        const auto room1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room1Size, "room1Box");
        room1->setLocation(room1Position);
        auto room1Mesh = room1->getMesh();

        const EAR::Vector3 room2Size(4.0f, 4.0f, 1.5f);
        EAR::Vector3 room2Position(4.0f, -3.f, 5.0f);
        // room2Position += portalBoxPosition;
        auto room2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(room2Size, "room1Box");
        room2->setLocation(room2Position);
        auto room2Mesh = room2->getMesh();

#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, room1->getMesh(), "room1");
        logging::writeMeshToEIF(portalLogFile, room2->getMesh(), "room2");
#endif

        EAR::Geometry::BoundingBox portalBox = extent->getGlobalBoundingBox();
        EAR::Vector3 portalLocation; //centre point of the portal, before axis alignment.
        portalLocation[0] = portalBox[0].low + (portalBox[0].high - portalBox[0].low) / 2;
        portalLocation[1] = portalBox[1].low + (portalBox[1].high - portalBox[1].low) / 2;
        portalLocation[2] = portalBox[2].low + (portalBox[2].high - portalBox[2].low) / 2;
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, extent->getMesh(), "extentMeshBeforeRotation");
#endif
        EAR::Vector3 targetVector(0.f, 0.f, -1.f);

        EAR::HomogenMatrix rotMatrix = EAR::rotateVectorTo(portalNormal, targetVector);
        EAR::Euler currRotation = extent->getOrientation();
        extent->setOrientation(currRotation + rotMatrix.getRotation());

        portalBox = extent->getGlobalBoundingBox();
        EAR::Vector3 portalSize;
        portalSize[0] = portalBox[0].high - portalBox[0].low;
        portalSize[1] = portalBox[1].high - portalBox[1].low;
        portalSize[2] = portalBox[2].high - portalBox[2].low;

        EAR::Position portalPosition(portalLocation, rotMatrix.inverse().getRotation());

        auto reverbExtent1 = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalSize, offset, "PortalExtent1", portalPosition, 3, 0.0f, room1Mesh));
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, reverbExtent1->getMesh(), "reverbExtent1");
#endif
        portalPosition.orientation[0] += 180.f;
        auto reverbExtent2 = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(
                portalSize, offset, "PortalExtent2", portalPosition, 3, 0.0f, room2Mesh));
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, reverbExtent2->getMesh(), "reverbExtent2");
#endif
        //return extent to original location
        extent->setOrientation(currRotation);

        float portalDistance;
        auto portalExtentVerts = reverbExtent1->getMesh()->getVertices(EAR::Mesh::CSpace::Global);
        for (auto vert : portalExtentVerts) {
            portalDistance = (extent->getMesh()->closestPointTo(vert) - vert).abs();
            CHECK(portalDistance - 0.005f <= offset + 0.1f);
            CHECK(room1Mesh->isInside(vert) == true);
        }

        portalExtentVerts = reverbExtent2->getMesh()->getVertices(EAR::Mesh::CSpace::Global);
        for (auto vert : portalExtentVerts) {
            portalDistance = (extent->getMesh()->closestPointTo(vert) - vert).abs();
            CHECK(portalDistance - 0.005f <= offset + 0.1f);
            CHECK(room2Mesh->isInside(vert) == true);
        }

#if (LOG_TO_XML)
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Portal as Box using SceneState - none-axis aligned initial vector")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(&portalLogFile, "C:/temp/reverbPortalAsBoxUsingScenestate-notAxisAligned.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(0.0f, 0.0f, 0.0f);
        const EAR::Euler portalOrientation(142.2f, 0.f, 0.f);
        const EAR::Vector3 portalVector(-0.61566f, 0.f, 0.78801f);
        const EAR::Vector3 portalBoxSize(2.0f, 2.0f, 0.1f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        // Create portal extent
        EIFPayload::BoxDesc portalBox;
        portalBox.position = EAR::Position(portalPosition, portalOrientation);
        portalBox.size = portalBoxSize;
        portalBox.id = "0";
        portalBox.bsId = 0;

        sceneState.createBox2(portalBox);

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(portalBoxPtr->getMesh(), reverbExtentMeshes, reverbExtentMeshOffset);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }

    SUBCASE("Portal as Box using SceneState - none-axis aligned initial vector 2")
    {
#if (LOG_TO_XML)
        FILE* portalLogFile = 0;
        fopen_s(
            &portalLogFile, "C:/temp/reverbPortalAsBoxUsingScenestate-notAxisAligned-2.xml", "w");
        fprintf(portalLogFile, "<AudioScene id=\"\">\n");
#endif
        const float roomSize = 5.f;
        const bool hasDefaultAE = false;
        DummySceneState sceneState(hasDefaultAE);
        const EAR::Vector3 portalPosition(10.f, -3.0f, 4.0f);
        const EAR::Euler portalOrientation(-23.f, 0.f, 0.f);
        const EAR::Vector3 portalVector(0.39073f, 0.f, -0.9205f);
        const EAR::Vector3 portalBoxSize(2.0f, 2.0f, 0.1f);

        const std::string room1AcousticEnvironmentId(std::to_string(1));
        const std::string room2AcousticEnvironmentId(std::to_string(2));

        // Create portal extent
        EIFPayload::BoxDesc portalBox;
        portalBox.position = EAR::Position(portalPosition, portalOrientation);
        portalBox.size = portalBoxSize;
        portalBox.id = "0";
        portalBox.bsId = 0;

        sceneState.createBox2(portalBox);

        auto portalPayload = std::make_shared<DummyPortalPayload>();
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalVector;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));

        const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];
        auto portalBoxMesh = portalBoxPtr->getMesh();
        auto reverbExtentMeshes =
            portalPayload->getPortals()[0]->scenePortal->getExternalReverbExtents();
        auto reverbExtentMeshOffset =
            portalPayload->getPortals()[0]->scenePortal->getExtentOffset();

        checkExtentDistance(portalBoxPtr->getMesh(), reverbExtentMeshes, reverbExtentMeshOffset);
#if (LOG_TO_XML)
        logging::writeMeshToEIF(portalLogFile, portalBoxPtr->getMesh(), "extentMesh");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[0]->getMesh(), "reverbMesh1");
        logging::writeMeshToEIF(portalLogFile, reverbExtentMeshes[1]->getMesh(), "reverbMesh2");
        fprintf(portalLogFile, "</AudioScene>");
        fclose(portalLogFile);
#endif
    }
}

} // namespace MPEGI
