/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include "MPEGI/Helpers/RayPlaneIntersection.h"
#include "MPEGI/Renderer/PerceptualPointDistribution.h"
#include <doctest/doctest.h>
#include <vector>

#define writeToXml false

namespace MPEGI {

TEST_CASE("Single Ray Plane Intersection")
{
    SUBCASE("Unbounded plane directly in front")
    {
        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, 1.0f };

        EAR::Vector3 rayOrigin{ .0f, .0f, 4.0f };
        EAR::Vector3 rayVector{ 0.0f, 0.0f, -1.0f };

        EAR::Vector3 hitLocation{ 0.0f, 0.0f, 0.0f };
        IntersectPlane plane{ planeCenter, planeNormal };
        plane.planeCenter = planeCenter;
        plane.planeNormal = planeNormal;

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == true);
        CHECK((result.second - hitLocation).abs() < 0.001f);

        rayVector.set(0.0f, 1.0f, 0.0f);
        result = MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);
        CHECK(result.first == false);
    }

    SUBCASE("Unbounded plane, ray at angle to plane")
    {
        EAR::Vector3 planeCenter{ 1.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, 1.0f };

        EAR::Vector3 rayOrigin{ -3.0f, -3.0f, 4.0f };
        EAR::Vector3 rayVector{ 1.f, 1.0f, -1.0f };

        EAR::Vector3 hitLocation{ 1.0f, 1.0f, 0.0f };

        IntersectPlane plane{ planeCenter, planeNormal };

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == true);
        CHECK((result.second - hitLocation).abs() < 0.001f);
    }

    SUBCASE("Unbounded plane directly in front, ray pointing away")
    {
        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, 1.0f };

        EAR::Vector3 rayOrigin{ .0f, .0f, 4.0f };
        EAR::Vector3 rayVector{ 0.0f, 0.0f, 1.0f };

        IntersectPlane plane{ planeCenter, planeNormal };

        EAR::Vector3 hitLocation{ 0.0f, 0.0f, 0.0f };

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == false);
        CHECK((result.second - hitLocation).abs() < 0.001f);
    }

    SUBCASE("Unbounded plane directly in front, plane normal away from ray origin")
    {
        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, -1.0f };

        EAR::Vector3 rayOrigin{ .0f, .0f, 4.0f };
        EAR::Vector3 rayVector{ 0.0f, 0.0f, -1.0f };

        EAR::Vector3 hitLocation{ 0.0f, 0.0f, 0.0f };
        IntersectPlane plane{ planeCenter, planeNormal };
        plane.planeCenter = planeCenter;
        plane.planeNormal = planeNormal;

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == true);
        CHECK((result.second - hitLocation).abs() < 0.001f);
    }

    SUBCASE("Bounded plane directly in front")
    {
        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, 1.0f };
        EAR::Vector3 plane_lowerLeft{ -2.f, -2.f, 0.f };
        EAR::Vector3 plane_lowerRight{ 2.f, -2.f, 0.f };
        EAR::Vector3 plane_upperLeft{ -2.f, 2.f, 0.f };
        EAR::Vector3 plane_upperRight{ 2.f, 2.f, 0.f };
        IntersectPlane plane{ planeCenter, planeNormal };
        plane.setBounds(plane_lowerLeft, plane_lowerRight, plane_upperLeft, plane_upperRight);

        EAR::Vector3 rayOrigin{ -3.0f, .0f, 4.0f };
        EAR::Vector3 rayVector{ 0.0f, 0.0f, -1.0f };

        EAR::Vector3 hitLocation{ 0.0f, 0.0f, 0.0f };

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == false);
    }

    SUBCASE("Bounded plane directly in front, ray missing")
    {
        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, 1.0f };
        EAR::Vector3 plane_lowerLeft{ -2.f, -2.f, 0.f };
        EAR::Vector3 plane_lowerRight{ 2.f, -2.f, 0.f };
        EAR::Vector3 plane_upperLeft{ -2.f, 2.f, 0.f };
        EAR::Vector3 plane_upperRight{ 2.f, 2.f, 0.f };
        IntersectPlane plane{ planeCenter, planeNormal };
        plane.setBounds(plane_lowerLeft, plane_lowerRight, plane_upperLeft, plane_upperRight);

        EAR::Vector3 rayOrigin{ .0f, .0f, 4.0f };
        EAR::Vector3 rayVector{ 0.0f, 0.0f, -1.0f };

        EAR::Vector3 hitLocation{ 0.0f, 0.0f, 0.0f };

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == true);
        CHECK((result.second - hitLocation).abs() < 0.001f);
    }

    SUBCASE("Bounded plane directly in front facing away")
    {
        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, -1.0f };
        EAR::Vector3 plane_lowerLeft{ 2.f, -2.f, 0.f };
        EAR::Vector3 plane_lowerRight{ -2.f, -2.f, 0.f };
        EAR::Vector3 plane_upperLeft{ 2.f, 2.f, 0.f };
        EAR::Vector3 plane_upperRight{ -2.f, 2.f, 0.f };
        IntersectPlane plane{ planeCenter, planeNormal };
        plane.setBounds(plane_lowerLeft, plane_lowerRight, plane_upperLeft, plane_upperRight);

        EAR::Vector3 rayOrigin{ .0f, .0f, 4.0f };
        EAR::Vector3 rayVector{ 0.0f, 0.0f, -1.0f };

        EAR::Vector3 hitLocation{ 0.0f, 0.0f, 0.0f };

        std::pair<bool, EAR::Vector3> result =
            MPEGI::checkRayPlaneIntersection(rayOrigin, rayVector, plane);

        CHECK(result.first == true);
        CHECK((result.second - hitLocation).abs() < 0.001f);
    }
}

TEST_CASE("Vectors of RayHits")
{
    SUBCASE("")
    {

#if (writeToXml)
        FILE* hitLog = 0;
        fopen_s(&hitLog, "C:/Temp/planeHitLog.xml", "w");
        fprintf(hitLog, "<hitLog>\n");
#endif
        std::vector<MPEGI::RayHit> allRays;
        for (unsigned int i = 0; i < 4096; i++) {
            allRays.push_back(RayHit{ i, 10.f, 1.f });
        }

        EAR::Vector3 planeCenter{ 0.0f, 0.0f, 0.0f };
        EAR::Vector3 planeNormal{ 0.0f, 0.0f, 1.0f };
        EAR::Vector3 plane_lowerLeft{ -1.f, -1.f, 0.f };
        EAR::Vector3 plane_lowerRight{ 1.f, -1.f, 0.f };
        EAR::Vector3 plane_upperLeft{ -1.f, 1.f, 0.f };
        EAR::Vector3 plane_upperRight{ 1.f, 1.f, 0.f };
        IntersectPlane plane{ planeCenter, planeNormal };
        plane.setBounds(plane_lowerLeft, plane_lowerRight, plane_upperLeft, plane_upperRight);

        EAR::Vector3 listenerLocation{ 0.f, 0.f, 3.f };
        EAR::Quaternion listenerRotation{ EAR::Euler{} };

        std::vector<RayHit> resultRays =
            checkRayHitsPlaneIntersection(allRays, plane, listenerLocation, listenerRotation);

        for (MPEGI::RayHit rh : resultRays) {
            EAR::Vector3 rotatedRay = listenerRotation.rotate(SESS_RAY_DISTRIBUTION[rh.rayID]);
            std::pair<bool, EAR::Vector3> result =
                MPEGI::checkRayPlaneIntersection(listenerLocation, rotatedRay, plane);
            CHECK(result.first == true);
#if (writeToXml)
            fprintf(hitLog, "<rayHit index=\"%i\" location=\"%f %f %f\"/>\n", rh.rayID,
                result.second[0], result.second[1], result.second[2]);
#endif
        }
#if (writeToXml)
        fprintf(hitLog, "</hitLog>\n");
        fclose(hitLog);
#endif
    }
}

} // namespace MPEGI