# MPEG-I Renderer

## Build Instructions

### Preparations

On Windows, Microsoft Visual Studio has to be installed.
Version 2017 is supported, other versions may or may not work.
The commands mentioned below should be entered into the
*x64 Native Tools Command Prompt for Visual Studio*
(if multiple versions of Visual Studio are installed,
each one has its separate command prompt).
Alternatively, the `vcvarsall.bat` script can be used to select the desired version.
The Ninja build tool is already part of the Visual Studio installation.

On macOS, the Clang compiler can be installed via the *XCode Command Line Tools*.
The Ninja build tool can be installed via the package manager (e.g. `brew`).

On all platforms, [CMake](https://cmake.org/) is needed for generating the build system.

The IPP and MKL libraries have to be installed on the system
(<https://www.intel.com/content/www/us/en/developer/tools/oneapi/base-toolkit-download.html>).

On the Windows command prompt, their path can be added like this:

    set MKL_ROOT=C:\Program Files (x86)\Intel\oneAPI\mkl\latest
    set IPP_ROOT=C:\Program Files (x86)\Intel\oneAPI\ipp\latest

It is recommended to add those environment variables via the Control Panel,
which makes them available also for future sessions (but re-booting might be necessary).

On macOS and Linux, their path can be added to the shell configuration file like this:

    export IPP_ROOT=/opt/intel/oneapi/ipp/latest
    export MKL_ROOT=/opt/intel/oneapi/mkl/latest

HDF5 v1.10.5 has to be installed
(Windows: [download](https://docs.hdfgroup.org/archive/support/ftp/HDF5/releases/hdf5-1.10/hdf5-1.10.5/bin/index.html);
macOS: `brew install hdf5`;
Linux: `apt-get install libhdf5-dev`).

If HDF5 is not found during the build process,
its location can be specified explicitly, for example by adding
`-D HDF5_PATH="C:\Program Files\HDF_Group\HDF5\1.10.5\cmake\hdf5"`
to the CMake invocation.

NetCDF-C v4.7.4 has to be installed
(macOS: `brew install netcdf`;
Linux: `apt-get install libnetcdf-dev`),
except on Windows, where this library is contained in the `renderer` repository.

If NetCDF is not found during the build process,
its location can be specified explicitly, for example by adding
`-D NetCDF_DIR=/usr/local/Cellar/netcdf/4.7.4_2/lib/cmake/NetCDF`
to the CMake invocation.

Note that NetCDF depends on HDF5 and the versions of the standalone HDF5 installation
and the one NetCDF refers (or contains) should match
so if you change the HDF5 version, make sure the NetCDF version is changed accordingly.
On Linux and Mac this should be taken care of by installing the libraries
using system package management (e.g. Apt and Brew),
on Windows the correct HDF5 version has to be installed
to match the NetCDF version from the repository.


### Build Directory

In the command prompt, navigate to the main source code directory.
Create a new build directory and navigate into it:

    mkdir build
    cd build

The build directory can have an arbitrary name and location.
In the following commands, we assume the directory name `build`
(relative to the main source code directory).


### Configure/Generate

Here we show the command line usage, but alternatively, the
*CMake-GUI* application can be used for the "configure" and "generate" steps.
The (recommended) Ninja generator can be selected with:

    cmake .. -G Ninja

Here, the `..` is the path to the main source code directory.

The available generators depend on the operating system and can be listed with:

    cmake --help

On Windows, a Visual Studio solution file can be created with
(specifying the desired Visual Studio version and machine architecture):

    cmake .. -G "Visual Studio 15 2017" -A x64

Regardless of the selected generator, the software can be compiled directly with CMake,
as shown in the following section.

Any CMake configuration variables can be passed with the `-D` flag, e.g.:

    cmake .. -G Ninja -D MPEGI_BUNDLE_NAME=ME_P13

See further below for available options.

Note that additional 3rd party dependencies are downloaded and set up during the first configuration. Therefore, internet connectivity is required. It is possible to work offline (after initial configuration) by setting:

```
cmake .. -D FETCHCONTENT_FULLY_DISCONNECTED=Yes
```


### Build

In the build directory, simply call:

    cmake --build .

The `.` is the path to the build directory.

Use the `--verbose` flag to show the full compilation commands,
to make sure the expected optimization flags are used:

    cmake --build . --verbose

When using the `"Visual Studio ..."` generator
(which is a so-called "multi-config generator"),
you should select the `Release` configuration to enable optimizations:

    cmake --build . --config Release


### Change Configuration

Any CMake configuration variables can be changed from the build directory, e.g.:

    cmake . -D MPEGI_BUNDLE_NAME=experiment42

This will override the mentioned variable(s), but all other variables stay unchanged.


### CMake Configuration Variables

Those variables can be set with the `-D` command line option or via the CMake-GUI tool.

#### `HDF5_PATH`

If the HDF5 library (which has to be installed on the system) is not found automatically,
its path can be provided manually.

#### `MKL_ROOT`

If the MKL library (which has to be installed on the system) is not found automatically,
its path can be provided manually.
Alternatively, `MKL_ROOT` can also be defined as environment variable.

#### `MPEGI_BUNDLE_NAME`

The name of the Max external can be specified
to allow multiple different builds of the renderer running in parallel
(MaxMSP doesn't allow this if they have the same name).
For example, with this flag set to "internal-test", the external is called "internal-test~".
Default is "mpegi", resulting in "mpegi~".
Note that the `OSCPort` config value for each instance must be different,
because multiple listeners cannot bind to the same UDP port.

#### `PYTHON_EXECUTABLE`

To specify the exact python version used to compile python bindings. For interoperability
with the MPEG-I Audio encoder, this should be set to the python executable contained in the
corresponding `mpegi` conda-environment.

#### `MPEGI_MEASURE_STAGE_TIMES`

To enable the complexity measurement tool from within the CLI renderer.
`-D MPEGI_MEASURE_STAGE_TIMES=ON`

### Troubleshooting

#### My Git submodules are messed up!

Whenever there are changes to the Git submodules, you should run this:

    git submodule update

If that's not enough, you can try the more thorough option:

    git submodule sync --recursive
    git submodule update --init --recursive

#### My CMake settings are messed up!

It's probably easiest to remove the whole build directory and start again with a fresh one.

#### My whole Git workdir is messed up!

It's always possible to fix things with Git, but sometimes it's just easier
to remove the whole Git workdir and start with a fresh clone.


### Faster Builds

Compiling the dependencies (especially the Embree library) can take quite a while.
There are two options for reducing build times:

#### Conan

If the [Conan package manager](https://conan.io/) is installed,
it should be automatically detected by CMake and the Embree library should be built with Conan.

#### sccache

If [sccache](https://github.com/mozilla/sccache) is installed and the environment variables
`CMAKE_C_COMPILER_LAUNCHER` and `CMAKE_CXX_COMPILER_LAUNCHER` are both set to `sccache`
(or the full path to the `sccache` executable), all compiled object files will be cached.
This will make subsequent builds much faster.


### Building the Source Code Documentation

The low-level API documentation can be generated with Doxygen like this:

    cd doc
    doxygen

### Apple macOS computers with Apple Silicon CPU
#### Rosetta emulation environment build

For mac computers with ARM chips, some additional configuration steps are needed. First, install
Rosetta 2 and homebrew for it 
(https://www.wisdomgeek.com/development/installing-intel-based-packages-using-homebrew-on-the-m1-mac/). 

Through the Rosetta homebrew, install HDF5 and NetCDF packages, e.g. using `arch -x86_64 /usr/local/homebrew/bin/brew 
install <package name>`, if you configured homebrew for a single terminal.

For CMake process, add flags that point to the Intel based packages, e.g. 
`-D HDF5_ROOT=/usr/local/homebrew/Cellar/hdf5/1.12.2_2` and 
`-D NetCDF_DIR=NetCDF_DIR=/usr/local/homebrew/Cellar/netcdf/4.9.0/lib/cmake/netCDF`. Additionally, configure CMake to
build for x64 with a flag ` -D CMAKE_OSX_ARCHITECTURES=x86_64`.

### Native Apple Silicon (ARM) build

Using the native Homebrew for Apple Silicon, install ARM versions of HDF5 and NetCDF with `brew install hdf5` and ` brew install netcdf`.
Configure CMake to build a native ARM binary for macOS with flags ` -D CMAKE_OSX_ARCHITECTURES=arm64 -D EAR_FFT_LIB=ne10`.

## HRIR Data

Head-related impulse responses from the FABIAN head-related transfer function database
are included in the file `fabian_fixdef.h`.
The data has been provided by Audio Communication Group, TU Berlin, Germany (www.ak.tu-berlin.de)
under the Creative Commons (CC-BY) license, see http://creativecommons.org/licenses/by/4.0/.


## Python bindings for MPEG-I Audio Encoder

To compile python bindings for the MPEG-I Audio Encoder, the python version used in the `mpegi` conda environment must
be specified during configuration, e.g.,

```
cmake -DPYTHON_EXECUTABLE="C:\Miniconda3\envs\mpegi\python.exe" ..
```

*Note*: On Linux, the `python3-dev` package must be installed

To make the module available to the encoder, the path to the shared-objects (`pyd` on Windows) must be added to `PYTHONPATH`,
e.g., `PowerShell`:

```
$env:PYTHONPATH +=";${CMAKE_BINARY_DIR}\stages\VoxelDiffractionStage" # Single-config generators, e.g., Ninja
$env:PYTHONPATH +=";${CMAKE_BINARY_DIR}\stages\VoxelDiffractionStage\[Debug|Release|RelWithDebInfo] # MSVC generator
```

## JSON trajectories for renderer-cli

### Setup

1. Clone nlohmann-json repo:
```
git clone https://github.com/nlohmann/json.git
```

2. Build
```
mkdir build
cmake -DJSON_BuildTests=OFF ..
cmake --build .
```

3. Configure MPEG-I renderer
```
cmake -DCMAKE_PREFIX_PATH="/path/to/json/build" -DUSE_JSON_TRAJECTORY=Yes ..
```

4. Check cmake output
```
-- Using JSON trajectories
```

## Format

Example:
```
{
    "source": "Source file name, if converted from AEP log",
    "records": [
        {
            "timestamp": 0.0,
            "address": "/interactionEvent",
            "args": [
                131,
                -74.97
            ]
        },
        {
            "address": "/userPose",
            "timestamp": 0.0,
            "args": [
                0.0,
                2.0,
                10.0,
                180.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0
            ]
        }
    ]
}
```
