/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <algorithm>
#include <cctype>
#define _USE_MATH_DEFINES
#include <cmath>
#include <cstddef>
#include <iostream>
#include <string>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayloadHandler.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/ComplexFrame.h>
#include <EAR/FFT.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

char* getCmdOption(char** begin, char** end, const std::string& option)
{
    char** it = std::find(begin, end, option);

    if (it != end && ++it != end) {
        return *it;
    }

    return 0;
}

std::vector<float> parsePositionCmdOption(char** begin, char** end, const std::string& option)
{
    char** it = std::find(begin, end, option);

    std::vector<float> numbers;

    if (it != end) {
        while (++it != end && (it[0][0] != '-' || std::isdigit(it[0][1]))) {
            try {
                numbers.push_back(std::stof(*it));
            } catch (...) {
                std::cout << "Invalid number in positional option " << option << ": " << *it
                          << std::endl;
                exit(1);
            }
        }
    }

    return numbers;
}

bool cmdOptionExists(char** begin, char** end, const std::string& option)
{
    return std::find(begin, end, option) != end;
}

void printStats(std::vector<double>& runtimes, const std::string& name, float outlierThreshold)
{
    int blocks = runtimes.size();

    if (blocks < 1) {
        std::cout << "Not enough blocks for meaningful statistics." << std::endl;
        return;
    }

    double unfiltMin = *std::min_element(runtimes.begin(), runtimes.end());
    double unfiltMax = *std::max_element(runtimes.begin(), runtimes.end());

    // filter outliers in place
    double unfiltSum = std::accumulate(runtimes.begin(), runtimes.end(), 0.0);
    double unfiltMean = unfiltSum / blocks;
    double unfiltSqSum =
        std::inner_product(runtimes.begin(), runtimes.end(), runtimes.begin(), 0.0);
    double unfiltStdev = std::sqrt(unfiltSqSum / blocks - unfiltMean * unfiltMean);

    double range = outlierThreshold * unfiltStdev;
    int outlierCount = 0;
    for (auto it = runtimes.begin(); it != runtimes.end();) {
        if (*it > unfiltMean + range || *it < unfiltMean - range) {
            it = runtimes.erase(it);
            outlierCount++;
        } else {
            it++;
        }
    }

    blocks = runtimes.size(); // update number of remaining blocks

    if (blocks < 1) {
        std::cout << "Not enough blocks for meaningful statistics." << std::endl;
        return;
    }

    // calculate and print statistics
    double medianRuntime;
    std::sort(runtimes.begin(), runtimes.end());
    if (blocks % 2 == 0) {
        medianRuntime = (runtimes[blocks / 2 - 1] + runtimes[blocks / 2]) / 2;
    } else {
        medianRuntime = runtimes[blocks / 2];
    }
    double minimumRuntime = runtimes[0];
    double maximumRuntime = runtimes[blocks - 1];

    double sum = std::accumulate(runtimes.begin(), runtimes.end(), 0.0);
    double meanRuntime = sum / blocks;

    double sq_sum = std::inner_product(runtimes.begin(), runtimes.end(), runtimes.begin(), 0.0);
    double stdev = std::sqrt(sq_sum / blocks - meanRuntime * meanRuntime);

    std::cout << "Processing statistics (" << name << "):" << std::endl;
    std::cout << "Mean: " << meanRuntime << " us" << std::endl;
    std::cout << "Stddev: " << stdev << " us" << std::endl;
    std::cout << "Median: " << medianRuntime << " us" << std::endl;
    std::cout << "Minimum: " << minimumRuntime << " us" << std::endl;
    std::cout << "Maximum: " << maximumRuntime << " us" << std::endl;
    std::cout << "Filtered " << outlierCount << " outliers (more than " << outlierThreshold
              << " standard deviations from the mean, "
              << "unfiltered max: " << unfiltMax << " us, min: " << unfiltMin << " us)"
              << std::endl;
    std::cout << std::endl;
}

std::size_t nextPowerOf2(float num)
{
    return std::pow(2, std::ceil(std::log2(num)));
}

//! Overwriting the EIF payload handler to only create a single source for IR generation
class IRCustomEIFPayloadHandler : public MPEGI::EIFPayloadHandler
{
public:
    IRCustomEIFPayloadHandler(const std::string& basePath, const EAR::Position& sourcePos)
        : MPEGI::EIFPayloadHandler(basePath)
        , sourcePos(sourcePos){};

    //! Inherited from Bitstream::PayloadHandler
    virtual std::shared_ptr<const MPEGI::Bitstream::Payload> unpack(
        const MPEGI::Bitstream::RawPayload& payload,
        const MPEGI::BitstreamConfig& bsConfig) override
    {
        auto eifPayload = std::const_pointer_cast<MPEGI::EIFPayload>(
            std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
                MPEGI::EIFPayloadHandler::unpack(payload, bsConfig)));
        eifPayload->streams.clear();
        eifPayload->objectSources.clear();
        eifPayload->hoaGroups.clear();
        eifPayload->hoaSources.clear();
        eifPayload->channelSources.clear();

        // create new audio stream for sweep
        eifPayload->streams.push_back(MPEGI::EIFPayload::AudioStreamDesc());

        eifPayload->streams.back().id = "sweep";
        eifPayload->streams.back().aepInputChannels.push_back(0);

        // create object source at given sweep source position
        eifPayload->objectSources.push_back(MPEGI::EIFPayload::ObjectSourceDesc());

        eifPayload->objectSources.back().id = "SweepSource";
        eifPayload->objectSources.back().position = sourcePos;
        eifPayload->objectSources.back().active = true;
        eifPayload->objectSources.back().gainDb = 0;
        eifPayload->objectSources.back().refDistance = 1;
        eifPayload->objectSources.back().signal = &(eifPayload->streams.back());
        eifPayload->objectSources.back().extent = nullptr;
        eifPayload->objectSources.back().directivity = nullptr;
        eifPayload->objectSources.back().directiveness = 1.f;
        // eifPayload->objectSources.back().aparams.push_back(AudioElement::AuthoringParameter::noDistance)

        return std::dynamic_pointer_cast<MPEGI::Bitstream::Payload>(eifPayload);
    };

private:
    const EAR::Position sourcePos;
};
