/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <algorithm>
#include <atomic>
#include <fstream>
#include <iostream>
#include <numeric>
#include <sstream>
#include <stdexcept>
#include <streambuf>
#include <string>
#include <utility>
#include <vector>

#include <chrono>
#include <thread>

#include <MPEGI/Helpers/Clock.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/DenormalPrevention.h>
#include <MPEGI/Helpers/Helpers.h>
#include <MPEGI/Helpers/ReverbConstants.h>
#include <MPEGI/Renderer.h>
#include <MPEGI/Renderer/Spatializer.h>
#include <MPEGI/Scene.h>
#include <MPEGI/Scene/OSCListener.h>
#include <MPEGI/Stream/FileStreamManager.h>
#include <MPEGI/Stream/VSTStreamManager.h>

#include <EAR/AudioFile.h>
#include <EAR/BaseTypes.h>
#include <EAR/Exception.h>
#include <EAR/Log.h>
#include <EAR/RealFrame.h>
#include <EAR/TimeMeasure.h>

#include <LSDF/LSDFParser.h>

#include <MPEGI/Renderer/RenderStageRegistry.h>

#include "helpers.h"

#define BLOCKSIZE 256
#define SAMPLERATE 48000
#define OUTLIER_THRESHOLD 5 // allowed runtime distance from the mean in stddevs

// constants only used in IR mode
#define IR_LENGTH 240000 // impulse response length in samples
#define IR_OFFSET 4096 // number of samples before the impulse starts (to avoid boundary effects)
#define IR_START_FREQ 16 // lowest frequency in sweep in Hz
#define IR_STOP_FREQ 24000 // highest frequency in sweep in Hz

int main(int argc, char* argv[])
{
    std::cout << std::endl << "MPEG-I CLI Renderer " << MPEGI_VERSION_STRING << std::endl;

    MPEGI::registerStages();
    MPEGI::registerPayloadHandlers();

    // argument parsing
    const char *sceneFile = nullptr, *lsdfFile = nullptr, *sofaFile = nullptr, *outFile = nullptr,
               *trajectoryFile = nullptr, *configFile = nullptr;
    float runtime = -1;
    int16_t outBitDepth = 24;
    bool printAudioUtilization = false;

    if (argc <= 2) {
        std::cout << "\n"
                  << "Not enough input arguments.\n"
                  << "Usage:\n"
                  << "    ./renderer-cli -scene scene-file -out output-file "
                     "[-sofa hrir-file] "
                     "[-traj trajectory-recording-file] "
                     "[-conf config-file] "
                     "[-runtime length-in-seconds] "
                     "[-lsdf lsdf-file] "
                     "[-lst-pos listener-position] "
                     "[-ir-src-pos source-position] "
                     "[-out-bit-depth wave-file-bit-depth] "
#ifdef MPEGI_MEASURE_STAGE_TIMES
                     "[-stageTimeLog logfile-time-per-stage]"
                     "[-stageTimeLogPerFrame logfile-time-per-frame]"
#endif
                     " [-print-audio-utilization]"
                     "\n"
                  << std::endl;
        exit(0);
    }

    std::cout << std::endl;

    sceneFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-scene"));

    if (sceneFile) {
        std::cout << "Scene file: " << sceneFile << std::endl;
    } else {
        std::cout << "Error parsing argument \"-scene\"." << std::endl;
        exit(1);
    }

    if (cmdOptionExists(argv, argv + argc, "-lsdf")) {
        lsdfFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-lsdf"));

        if (lsdfFile) {
            std::cout << "LSDF file: " << lsdfFile << std::endl;
        } else {
            std::cout << "Error parsing argument \"-lsdf\"." << std::endl;
            exit(1);
        }
    }

    outFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-out"));

    if (outFile) {
        std::cout << "Output file: " << outFile << std::endl;
    } else {
        std::cout << "Error parsing argument \"-out\"." << std::endl;
        exit(1);
    }

    if (cmdOptionExists(argv, argv + argc, "-out-bit-depth")) {
        outBitDepth = atoi(getCmdOption(argv, argv + argc, "-out-bit-depth"));
        if (outBitDepth == 8 || outBitDepth == 16 || outBitDepth == 24 || outBitDepth == 32) {
            std::cout << "Output WAV file Bitdepth: " << outBitDepth << std::endl;
        } else {
            std::cout << "Error parsing argument \"-out-bit-depth\", only values 8, 16, 24, or 32 "
                         "are valid."
                      << std::endl;
            exit(1);
        }
    }

    if (cmdOptionExists(argv, argv + argc, "-sofa")) {
        sofaFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-sofa"));

        if (sofaFile) {
            std::cout << "SOFA HRIR file: " << sofaFile << std::endl;
        } else {
            std::cout << "Error parsing argument \"-sofa\"." << std::endl;
            exit(1);
        }
    }

    if (cmdOptionExists(argv, argv + argc, "-traj")) {
        trajectoryFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-traj"));

        if (trajectoryFile) {
            // check if file exists
            std::ifstream f(trajectoryFile);
            if (!f.good()) {
                std::cout << "Trajectory file \"" << trajectoryFile << "\" does not exist."
                          << std::endl;
                exit(1);
            }
            std::cout << "Trajectory Recording file: " << trajectoryFile << std::endl;
        } else {
            std::cout << "Error parsing argument \"-traj\"." << std::endl;
            exit(1);
        }
    }

    if (cmdOptionExists(argv, argv + argc, "-conf")) {
        configFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-conf"));

        if (configFile) {
            // check if file exists
            std::ifstream f(configFile);
            if (!f.good()) {
                std::cout << "Config file \"" << configFile << "\" does not exist." << std::endl;
                exit(1);
            }
            std::cout << "Config file: " << configFile << std::endl;
        } else {
            std::cout << "Error parsing argument \"-conf\"." << std::endl;
            exit(1);
        }
    }

    if (cmdOptionExists(argv, argv + argc, "-runtime")) {
        std::string arg = const_cast<char*>(getCmdOption(argv, argv + argc, "-runtime"));

        std::size_t pos;
        runtime = std::stof(arg, &pos);

        if (runtime >= 0) {
            std::cout << "Scene runtime: " << runtime << "s" << std::endl;
        } else {
            std::cout << "Error parsing argument \"-runtime\"." << std::endl;
            exit(1);
        }
    }

    std::unique_ptr<EAR::Position> listenerPos;
    std::unique_ptr<EAR::Position> irSourcePos;

    if (cmdOptionExists(argv, argv + argc, "-lst-pos")) {
        std::vector<float> nums = parsePositionCmdOption(argv, argv + argc, "-lst-pos");

        if (nums.size() != 6) {
            std::cout << "Too few arguments for the listener position." << std::endl;
            exit(1);
        }

        listenerPos = std::make_unique<EAR::Position>(
            EAR::Vector3(nums[0], nums[1], nums[2]), EAR::Euler(nums[3], nums[4], nums[5]));
    } else {
        listenerPos = std::make_unique<EAR::Position>();
    }

    if (cmdOptionExists(argv, argv + argc, "-ir-src-pos")) {
        std::vector<float> nums = parsePositionCmdOption(argv, argv + argc, "-ir-src-pos");

        if (nums.size() != 6) {
            std::cout << "Too few arguments for the IR generator source position." << std::endl;
            exit(1);
        }

        irSourcePos = std::make_unique<EAR::Position>(
            EAR::Vector3(nums[0], nums[1], nums[2]), EAR::Euler(nums[3], nums[4], nums[5]));
    }

#ifdef MPEGI_MEASURE_STAGE_TIMES
    const char* stageTimeLogFile = nullptr;
    const char* stageTimeLogFilePerFrame = nullptr;

    if (cmdOptionExists(argv, argv + argc, "-stageTimeLog")) {
        stageTimeLogFile = const_cast<char*>(getCmdOption(argv, argv + argc, "-stageTimeLog"));

        if (stageTimeLogFile) {
            std::cout << "stage time log file: " << stageTimeLogFile << std::endl;
        } else {
            std::cout << "Error parsing argument \"-stageTimeLog\"." << std::endl;
            exit(1);
        }
    }

    if (cmdOptionExists(argv, argv + argc, "-stageTimeLogPerFrame")) {
        stageTimeLogFilePerFrame =
            const_cast<char*>(getCmdOption(argv, argv + argc, "-stageTimeLogPerFrame"));

        if (stageTimeLogFilePerFrame) {
            std::cout << "stage time log file PerFrame: " << stageTimeLogFilePerFrame << std::endl;
        } else {
            std::cout << "Error parsing argument \"-stageTimeLogPerFrame\"." << std::endl;
            exit(1);
        }
    }
#endif

    bool irMode = false;
    if (listenerPos && irSourcePos) {
        irMode = true;
        std::cout << "Measuring binaural room impulse response with listener at " << *listenerPos
                  << " and source at " << *irSourcePos << std::endl;
    } else {
        std::cout << "Setting initial listener position for rendering to " << *listenerPos << "."
                  << std::endl;
    }

    std::cout << std::endl;

    std::unique_ptr<EAR::RealBuffer> irImpulse;
    if (irMode) {
        irImpulse = std::make_unique<EAR::RealBuffer>(IR_LENGTH + IR_OFFSET);
        irImpulse->at(IR_OFFSET) = 0.5f;
    }

    if (cmdOptionExists(argv, argv + argc, "-print-audio-utilization")) {
        printAudioUtilization = true;
    }

    // the whole processing is in a try/catch block to display all possible EAR::Exceptions
    try {
        std::cout << "Available stages: ";
        for (auto& name : MPEGI::RenderStageRegistry::getAvailableStages())
            std::cout << name << ", ";
        std::cout << std::endl;
        std::cout << "Available spatializers: ";
        for (auto& name : MPEGI::RenderStageRegistry::getAvailableSpatializers())
            std::cout << name << ", ";
        std::cout << std::endl;

        // Set config if given
        if (configFile) {
            std::ifstream file(configFile);
            std::ostringstream stringStream;
            stringStream << file.rdbuf();
            MPEGI::Config::deserialize(stringStream.str());

            const std::string& logPath = MPEGI::Config::getStringDef("LogPath", "");
            if (logPath != "") {
                EAR::Log::getInstance()->enableFileLogger(logPath);
            }
        }

        // load scene description
        std::string sf = sceneFile;

        // find base path (needed to locate relative wav and directivity files)
        std::string basePath = MPEGI::Helpers::getBasePath(sf);

        EAR::TimeMeasure::startTimer();
        MPEGI::Bitstream bitstream(sf);

        if (irMode) {
            bitstream.registerHandler(
                std::make_shared<IRCustomEIFPayloadHandler>(basePath, *irSourcePos),
                MPEGI::Bitstream::PLD_SCENE);
        } else {
            bitstream.registerHandler(
                std::make_shared<MPEGI::EIFPayloadHandler>(basePath), MPEGI::Bitstream::PLD_SCENE);
        }

        // EIF payload is needed below to set the scene duration correctly
        auto eifPayload = std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
            bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE));

        std::unique_ptr<LSDF::LSDFParser> lsdf;
        if (lsdfFile) {
            std::string lsdfPath = lsdfFile;
            auto predelayOffset = MPEGI::Config::getFloatDef(
                "Reverb:Predelay_Offset", MPEGI::DEFAULT_REVERB_PREDELAY_OFFSET);
            lsdf = std::make_unique<LSDF::LSDFParser>(lsdfPath, predelayOffset, false, true);
        }

        if (sofaFile) {
            MPEGI::Config::set("HrirSofaFile", std::string(sofaFile));
        }

        // build scene & renderer pipeline
        auto clock = std::make_shared<MPEGI::ExternalClock>();
        // params: EIF payload, clock, offlineMode, OSC listening, LSDF
        MPEGI::Scene scene(bitstream, clock, true, true, lsdf.get());

        if (trajectoryFile) {
            std::ifstream file(trajectoryFile);
            std::ostringstream stringStream;
            stringStream << file.rdbuf();
            double lastUpdateTime = scene.getOSCListener()->playbackRecording(stringStream.str());
            if (runtime <= 0) {
                runtime = lastUpdateTime;
            }
        }

        if (listenerPos) {
            MPEGI::Update upd;

            upd.addModification("listener", "location", listenerPos->location, false);
            upd.addModification("listener", "orientation", listenerPos->orientation, false);

            scene.triggerUpdate(upd);
        }

        std::shared_ptr<MPEGI::StreamManager> streamManager;
        if (irMode) {
            streamManager =
                std::make_shared<MPEGI::VSTStreamManager>(bitstream, SAMPLERATE, BLOCKSIZE);
        } else {
            streamManager = std::make_shared<MPEGI::FileStreamManager>(
                bitstream, SAMPLERATE, BLOCKSIZE, lsdf.get());
        }

        MPEGI::Renderer renderer(bitstream, scene.getState(), streamManager);

        auto spatializer = MPEGI::RenderStageRegistry::createSpatializer(
            MPEGI::Config::getStringDef("Spatializer", MPEGI_DEFAULT_SPATIALIZER));

        // check if spatializer supports binauralization and load optional HRIRs
        auto binauralizer = std::dynamic_pointer_cast<MPEGI::Binauralizer>(spatializer);
        if (binauralizer && sofaFile) {
            binauralizer->loadHRIRFile(sofaFile);
        }

        renderer.addSpatializer(spatializer, bitstream);

        double sceneTime = 0.0;
        clock->start();

        // create output frame
        EAR::AudioFile output;
        EAR::RealFrame* outFrame = output.getFrame();

        // prepare blockwise processing
        int lenInSamples;
        if (irMode) {
            lenInSamples = IR_LENGTH + IR_OFFSET;
        } else {
            lenInSamples = static_cast<int>(
                dynamic_cast<MPEGI::FileStreamManager*>(streamManager.get())->maxLength());
            if (eifPayload->getSceneInfo().duration > 0) {
                // a runtime is specified in the scene, so we overwrite how many blocks to process
                lenInSamples = eifPayload->getSceneInfo().duration * SAMPLERATE;
            }
            if (runtime >= 0) {
                // a runtime is specified by hand, so we overwrite how many blocks to process
                lenInSamples = runtime * SAMPLERATE;
            }
        }

        int blocks = lenInSamples / BLOCKSIZE; // floor
        int blockRemainder = lenInSamples % BLOCKSIZE;

        outFrame->resize(spatializer->getNbrOfChannels(), lenInSamples);

        EAR::RealFrame blockOutFrame(outFrame->channels(), BLOCKSIZE);

        // save measured runtimes for statistics
        std::vector<double> processBlockRuntimes(blocks);
        std::vector<double> sceneControlRuntimes(blocks);
        std::vector<double> renderUpdatesRuntimes(blocks);

#ifdef MPEGI_MEASURE_STAGE_TIMES
        renderer.initStageTimePerFrame(blocks);
#endif

        // EAR::Euler listenerOrient;

        // blockwise processing loop
        EAR::TimeMeasure::stopTimer();
        double initTime = EAR::TimeMeasure::getElapsedMicroseconds() / 1000000.f;
        std::cout << std::endl;
        std::cout << "Scene: " << sceneFile << ", Intialization time (s): " << initTime
                  << std::endl;
        std::cout << std::endl;

        std::cout << "Processing..." << std::endl;
        double tpb = static_cast<double>(BLOCKSIZE) / SAMPLERATE;
        double sceneDuration = eifPayload->getSceneInfo().duration;

        if (printAudioUtilization) {
            std::cout << std::fixed;
            std::cout << std::setprecision(2);
        }
        for (int b = 0; b < blocks; b++) {

            renderer.triggerControlStep();

            EAR::TimeMeasure::startTimer();
            if (irMode) {
                float* input[1];
                input[0] = irImpulse->data() + (b * BLOCKSIZE);
                dynamic_cast<MPEGI::VSTStreamManager*>(streamManager.get())->ingest(input);
            }
            {
                auto guard = MPEGI::TemporarilyDisableDenormals{};
                renderer.processBlock(blockOutFrame.data().get());
            }
            EAR::TimeMeasure::stopTimer();

            sceneTime += tpb;
            if (sceneDuration > 0 && sceneTime > sceneDuration) {
                sceneTime -= sceneDuration;
            }
            scene.setTime(sceneTime);

            for (int i = 0; i < blockOutFrame.channels(); i++) {
                outFrame->ch(i).write(blockOutFrame[i].data(), b * BLOCKSIZE, BLOCKSIZE);
            }

            processBlockRuntimes[b] = EAR::TimeMeasure::getElapsedMicroseconds();
            sceneControlRuntimes[b] = scene.getLastControlRuntime() * 1e6;
            renderUpdatesRuntimes[b] = renderer.getLastControlRuntime() * 1e6;

            if (printAudioUtilization) {
                if (b % 20 == 0) {
                    std::cout << "Timestamp: " << sceneTime;
                    MPEGI::AudioUtilization audioUtilization = renderer.getAudioUtilization();
                    if (audioUtilization.activeMPHOASources.size() > 0) {
                        std::cout << " - ACTIVE MP-HOA SOURCES: ";
                    }
                    for (int i = 0; i < audioUtilization.activeMPHOASources.size(); i++) {
                        std::cout << audioUtilization.activeMPHOASources[i].id << "("
                                  << audioUtilization.activeMPHOASources[i].order << "), ";
                    }
                    if (audioUtilization.notActiveSources.size() > 0) {
                        std::cout << " - INACTIVE SOURCES: ";
                    }
                    for (int i = 0; i < audioUtilization.notActiveSources.size(); i++) {
                        std::cout << audioUtilization.notActiveSources[i] << " ";
                    }
                    std::cout << "\n";
                }
            }
            if (b % 10 == 0) {
                std::cout << "[" << std::setw(3) << int(100 * (float)b / (float)blocks) << " %]\r";
            }
        }

        // process remainder samples
        if (blockRemainder > 0) {
            renderer.triggerControlStep();
            blockOutFrame.zero();
            if (irMode) {
                float* input[1];
                input[0] = irImpulse->data() + (blocks * BLOCKSIZE);
                dynamic_cast<MPEGI::VSTStreamManager*>(streamManager.get())->ingest(input);
            }
            renderer.processBlock(blockOutFrame.data().get());

            for (int i = 0; i < blockOutFrame.channels(); i++) {
                outFrame->ch(i).write(blockOutFrame[i].data(), blocks * BLOCKSIZE, blockRemainder);
            }
        }

        std::cout << "\33[2K"
                  << "Done!" << std::endl
                  << std::endl;

#ifdef MPEGI_MEASURE_STAGE_TIMES
        std::cout << "Stage time statistics" << std::endl;
        renderer.printStageTimesAvg(std::cout);
        std::cout << std::endl;
        if (stageTimeLogFile) {
            std::ofstream logFile(stageTimeLogFile);
            if (logFile) {
                renderer.printStageTimes(logFile);
                logFile.close();
            }
        }
        if (stageTimeLogFilePerFrame) {
            std::ofstream logFile(stageTimeLogFilePerFrame);
            if (logFile) {
                renderer.printStageTimesPerFrame(logFile);
                logFile.close();
            }
        }
#endif

        // scene.stopController();

        if (irMode) {
            outFrame->mul(2.f); // makeup gain for impulse having -6 dBFS
        }

        // save output file
        output.save(outFile, outBitDepth, SAMPLERATE);

        // print processing statistics
        printStats(sceneControlRuntimes, "Scene Control", OUTLIER_THRESHOLD);
        printStats(renderUpdatesRuntimes, "Renderer Updates", OUTLIER_THRESHOLD);
        printStats(processBlockRuntimes, "Audio Thread", OUTLIER_THRESHOLD);

        std::cout << "Blocks processed: " << blocks << std::endl;
        std::cout << "Single block audio duration: "
                  << static_cast<float>(BLOCKSIZE) / SAMPLERATE * 1000 * 1000 << " us" << std::endl
                  << std::endl;

        // std::cout << "Peak value in inputs: " << inFrame.peak() << std::endl;
        std::cout << "Peak value in output: " << outFrame->peak() << std::endl << std::endl;

        // std::cout << "RMS value in inputs: " << inFrame.rms() << std::endl;
        std::cout << "RMS value in output: " << outFrame->rms() << std::endl;
    } catch (const EAR::Exception& e) {
        std::cout << "Processing failed with error: " << std::endl << e.toString() << std::endl;
        exit(1);
    } catch (const LSDF::Exception& e) {
        std::cout << "Processing failed with error: " << std::endl << e.toString() << std::endl;
        exit(1);
    }

    std::cout << std::endl << "Processing successful." << std::endl << std::endl;

    return 0;
}
