#pragma once
//--------------------------------------------------------------------------------------------------
#include "accu.h"
#include "timer.h"
#include "notify.h"

#include <numeric>
//--------------------------------------------------------------------------------------------------
template<
    typename T,
    template<typename, typename> class Timer,
    template<typename> class Notify
    >
struct detector
{
    using value_t = T;
    using notify_t = Notify<T>;
    using timer_t = Timer<T, detector>;

protected:
    accumulator<T> _acc;
    notify_t _notify;

public:
    template<typename... Args>
    detector(
        std::size_t window_size,
        T threshold,
        Args&&... args)
        : _acc(window_size)
        , _notify(threshold, std::forward<Args>(args)...)
    {
    }

    ~detector() = default;
    detector(detector const&) = delete;
    detector& operator=(detector const&) = delete;

    std::size_t get_window_size() const { return _acc.get_window_size(); }

    void operator()(T duration)
    {
        _acc(duration);

        auto const avg =
            std::accumulate(_acc.begin(), _acc.end(), value_t{ 0 }) / _acc.get_window_size();

        if(avg > _notify.get_threshold())
        {
            _notify(avg);
        }
    }

    timer_t start()
    {
        return timer_t(*this);
    }
};

//--------------------------------------------------------------------------------------------------
/**
 * @brief Detector configuration
 */
using detector_t =
    detector<std::chrono::microseconds, chrono_timer, notify_maxmsp>;

//--------------------------------------------------------------------------------------------------
template<typename Detector, typename... Args>
Detector* make_detector(Args&&... args)
{
    return new Detector(std::forward<Args>(args)...);
}

template<typename Detector>
void free_detector(Detector **detector)
{
    if(!detector || !*detector) return;

    delete *detector;
    *detector = nullptr;
}