/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "fhg-mpegi-renderer_tilde.h"

#include <MPEGI/Renderer.h>

#include <LSDF/LSDFParser.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayload.h>
#include <MPEGI/Bitstream/EIFPayloadHandler.h>

#include <MPEGI/Scene.h>
#include <MPEGI/Scene/Update.h>

#ifndef MPEGI_DISABLE_OSC_LISTENER
#include <MPEGI/Scene/OSCListener.h>
#endif

#include <MPEGI/Helpers/Clock.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/DenormalPrevention.h>
#include <MPEGI/Helpers/Helpers.h>

#include <MPEGI/Stream/MaxStreamManager.h>

#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <MPEGI/Renderer/RenderStageRegistry.h>
#include <MPEGI/Renderer/Spatializer.h>

#include <exception>
#include <fstream>
#include <ostream>
#include <sstream>
#include <string>
#include <thread>

#include "helpers.h"

#define MPEGI_BITSTREAM_SUFFIX ".bit"

#define MPEGI_RELATIVE_LSDF_PATH                                                                   \
    ".." + LSDF_DIRECTORY_SEPARATOR + ".." + LSDF_DIRECTORY_SEPARATOR + "LSDF"                     \
        + LSDF_DIRECTORY_SEPARATOR + "LSDF.xml"

/* Helper functions */

// Send status message without changing the status identifier
void printStatusMessage(t_renderer* self, const std::string& message)
{
    t_atom statusMessage;
    atom_setsym(&statusMessage, gensym(message.c_str()));
    outlet_anything(self->messageOutlet, gensym("statusMsg"), 1, &statusMessage);
}

// This function can be used from inside the renderer to indicate errors to the outside Max world
void setStatus(t_renderer* self, int status, std::string message = "")
{
    t_atom statusValue;
    atom_setlong(&statusValue, status);
    outlet_anything(self->messageOutlet, gensym("status"), 1, &statusValue);

    if (message.length() > 0) {
        printStatusMessage(self, message);
    }
}

// Returns true if the file exists (doesn't say anything about the contents)
bool checkIfFileExists(const std::string& path)
{
    std::ifstream fileStream(path.c_str());
    return fileStream.good();
}

int Renderer::initializeRenderer(t_renderer* self)
{
    std::lock_guard<std::mutex> lock(processBlockMutex);

    renderer.reset();
    streamManager.reset();
    scene.reset();
    temp.reset();
    MPEGI::Config::reset();

    if (fs == 0 || B == 0 || sceneFile.empty() || hrirFile.empty()) {
        return 0;
    }

    try {
        // all required data is available

        if (configFile != "") {
            std::ifstream file(configFile);
            EAR_CHECK_REL(file.good(), "Config file \"" + configFile + "\" could not be opened.");
            std::ostringstream stringStream;
            stringStream << file.rdbuf();
            MPEGI::Config::deserialize(stringStream.str());
        }

        MPEGI::Config::set("HrirSofaFile", EAR::Variant(hrirFile));

        std::string basePath = MPEGI::Helpers::getBasePath(sceneFile);

        // bitstream memory is released after initialization is complete
        auto bitstream = std::make_unique<MPEGI::Bitstream>(sceneFile);
        bitstream->registerHandler(std::make_shared<MPEGI::EIFPayloadHandler>(basePath, false),
            MPEGI::Bitstream::PLD_SCENE);

        // EIF payload is needed below to determine if its AR (with LSDF) or VR
        auto eifPayload = std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
            bitstream->getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE));

        LSDF::LSDFParser* lsdfPtr = lsdfObject.get();

        loopStart = 0.0;
        loopEnd = eifPayload->getSceneInfo().duration;
        clock = std::make_shared<MPEGI::ExternalClock>();
        const bool enableOSCListener =
#ifdef MPEGI_DISABLE_OSC_LISTENER
            false;
#else // MPEGI_DISABLE_OSC_LISTENER
            true;
#endif // MPEGI_DISABLE_OSC_LISTENER
        scene =
            std::make_unique<MPEGI::Scene>(*bitstream, clock, false, enableOSCListener, lsdfPtr);
        streamManager = std::make_shared<MPEGI::MaxStreamManager>(
            *bitstream, fs, B, self->numberOfInputs, lsdfPtr);
        printStatusMessage(
            self, "Available Streams for Override: " + streamManager->getOverridableStreamIDs());

        renderer = std::make_unique<MPEGI::Renderer>(*bitstream, scene->getState(), streamManager);

        auto spatializer = MPEGI::RenderStageRegistry::createSpatializer(
            MPEGI::Config::getStringDef("Spatializer", MPEGI_DEFAULT_SPATIALIZER));

        // check if spatializer supports binauralization and load HRIRs if yes
        auto binauralizer = std::dynamic_pointer_cast<MPEGI::Binauralizer>(spatializer);
        if (binauralizer) {
            binauralizer->loadHRIRFile(hrirFile);
        }

        renderer->addSpatializer(spatializer, *bitstream);

        auto channels = spatializer->getNbrOfChannels();
        if (channels > MPEGI_MAX_OUTPUT_CHANNELS) {
            EAR_FAIL("Too many spatializer channels");
        }
        this->temp = std::make_unique<EAR::RealFrame>(channels, B);

        scene->setControlRate(MPEGI_CONTROL_RATE);
        scene->startController();

        // run controller by hand once so we definitely have a first valid RenderList when the audio process starts
        renderer->triggerControlStep();

        renderer->setControlRate(MPEGI_CONTROL_RATE);
        renderer->startController();

        tpb = B / fs;
        sceneTime = 0;
        clock->start();

    } catch (EAR::Exception& e) {
        printStatusMessage(self, e.toString());

        return 2;
    } catch (LSDF::Exception& e) {
        printStatusMessage(self, e.toString());

        return 3;
    } catch (std::exception& e) {
        printStatusMessage(self, std::string(e.what()));

        return 4;
    }

    return 0;
}

/* global class pointer variable */
static t_class* renderer_class = NULL;

/* Entry point */
void ext_main(void* r)
{
    if (!renderer_class) {
        MPEGI::registerStages();
        MPEGI::registerPayloadHandlers();
        renderer_class = class_new(MPEGI_BUNDLE_NAME, (method)renderer_new, (method)renderer_free,
            (long)sizeof(t_renderer), 0L, A_GIMME, 0);

        class_dspinit(renderer_class);

        class_addmethod(renderer_class, (method)renderer_dsp64, "dsp64", A_CANT, 0);
        class_addmethod(renderer_class, (method)renderer_assist, "assist", A_CANT, 0);

        // Register AEP specific message callbacks
        class_addmethod(renderer_class, (method)renderer_loadScene, "loadScene", A_GIMME, 0);
        class_addmethod(renderer_class, (method)renderer_loadLSDF, "loadLSDF", A_GIMME, 0);

        class_addmethod(renderer_class, (method)renderer_loadHRIRSet, "loadHRIRSet", A_GIMME, 0);

        class_addmethod(renderer_class, (method)renderer_setMaxDelay, "setMaxDelay", A_LONG, 0);

        // class_addmethod(renderer_class, (method) renderer_setLoop, "setLoop", A_GIMME, 0); // start time, end time
        // class_addmethod(renderer_class, (method) renderer_clearLoop, "clearLoop", 0);

        class_addmethod(renderer_class, (method)renderer_userPose, "userPose", A_GIMME, 0);

        class_addmethod(
            renderer_class, (method)renderer_accessibility, "accessibility", A_GIMME, 0);

        // params: id, x, y, z
        class_addmethod(renderer_class, (method)renderer_position, "position", A_GIMME, 0);
        // params: id, y, p, r
        class_addmethod(renderer_class, (method)renderer_orientation, "orientation", A_GIMME, 0);
        // params: id, g
        class_addmethod(renderer_class, (method)renderer_gain, "gain", A_GIMME, 0);
        // params: id (int), [placeholder (float), ...]
        class_addmethod(
            renderer_class, (method)renderer_triggerUpdate, "triggerUpdate", A_GIMME, 0);
        // params: id (string), [placeholder (float), ...]
        class_addmethod(renderer_class, (method)renderer_triggerUpdateString, "triggerUpdateString",
            A_GIMME, 0);

#ifndef MPEGI_DISABLE_OSC_LISTENER
        class_addmethod(renderer_class, (method)renderer_startRecordingOsc, "startRecordingOsc", 0);
        class_addmethod(renderer_class, (method)renderer_stopRecordingOsc, "stopRecordingOsc", 0);
        class_addmethod(
            renderer_class, (method)renderer_saveRecordedOsc, "saveRecordedOsc", A_SYM, 0);
#endif

        class_register(CLASS_BOX, renderer_class);
    }
}

void* renderer_new(t_symbol* sym, long argc, t_atom* argv)
{
    t_renderer* self = (t_renderer*)object_alloc(renderer_class);
#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "renderer_new thread id: " << std::this_thread::get_id();

    t_atom* ap;
    int i;
    for (i = 0, ap = argv; i < argc; i++, ap++) {
        switch (atom_gettype(ap)) {
        case A_LONG:
            ss << ", arg long: " << atom_getlong(ap);
            break;
        case A_FLOAT:
            ss << ", arg float: " << atom_getfloat(ap);
            break;
        case A_SYM:
            ss << ", arg sym: " << atom_getsym(ap)->s_name;
            break;
        default:
            ss << ", arg of unknown type: " << atom_gettype(ap);
            break;
        }
    }

    printStatusMessage(self, ss.str());
#endif

    if (!self) {
        return NULL;
    }

    // first argument is number of input channels
    self->numberOfInputs = 2; // default value
    if (argc > 0) {
        self->numberOfInputs = atom_getlong(argv);
    }
    self->connectedInlets = 0;

    dsp_setup((t_pxobject*)self, self->numberOfInputs);

    self->messageOutlet = outlet_new((t_object*)self, NULL);

    // second argument is number of output channels. default is 2 for headphones
    self->numberOfOutputs = (argc > 1) ? atom_getlong(&argv[1]) : 2;
    if (self->numberOfOutputs > MPEGI_MAX_OUTPUT_CHANNELS) {
        printStatusMessage(self, "Too many output channels");
        return nullptr;
    }

    // third argument is detector treshold
    using namespace std::chrono_literals;
    auto const detector_threshold =
        (argc > 2) ? std::chrono::microseconds(atom_getlong(&argv[2])) : 5000us;

    // fourth argument is detector window size
    auto const window_size = (argc > 3) ? static_cast<int>(atom_getlong(&argv[3])) : 11;

    self->detector =
        make_detector<detector_t>(window_size, detector_threshold, self->messageOutlet);
    if (!self->detector) {
        renderer_free(self);
        return nullptr;
    }

    for (int i = 0; i < self->numberOfOutputs; ++i) {
        outlet_new(self, "signal");
    }

    // self->proxyInlet = proxy_new((t_object *)self, 0, &self->proxyNumber);

    self->rendererPayload = static_cast<void*>(new Renderer);
    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    // fifth argument is config file path (not part of the official MPEG-I requirements)
    std::string configFile = "";
    if (argc > 4) {
        configFile = std::string(atom_getsym(&argv[4])->s_name);
    }
    renderer->configFile = configFile;

    std::stringstream versionString;
    versionString << "MPEG-I renderer Max External v" << MPEGI_VERSION_STRING << ", bundle name: \""
                  << MPEGI_BUNDLE_NAME << "\"";
    object_post((t_object*)self, versionString.str().c_str());

    return self;
}

void renderer_free(t_renderer* self)
{
    dsp_free((t_pxobject*)self);

    delete static_cast<Renderer*>(self->rendererPayload);

    free_detector(&self->detector);
}

void renderer_assist(t_renderer* self, void* unused, long io, long index, char* string_dest)
{
    switch (io) {
    case ASSIST_INLET:
        if (index == self->numberOfInputs) {
            sprintf(string_dest, "Message inlet");
        } else {
            sprintf(string_dest, "Signal inlet (ch. %ld)", index + 1);
        }
        break;
    case ASSIST_OUTLET:
        if (index == self->numberOfOutputs) {
            sprintf(string_dest, "Message outlet (status)");
        } else {
            sprintf(string_dest, "Signal outlet (%ld)", index + 1);
        }
        break;
    default:
        break;
    }
}

void renderer_loadScene(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    t_symbol* sym = atom_getsym(argv);
    auto sceneName = std::string{ sym->s_name };

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "loadScene from path: " << sceneName;
    printStatusMessage(self, ss.str());
#endif

    setStatus(self, -1);

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    auto bundleName = std::string_view{ MPEGI_BUNDLE_NAME };
    // If bundleName starts with ME_Pxy, append _Pxy to sceneName, unless it's already there
    if (bundleName.rfind("ME_P", 0) == 0) {
        auto proponentSuffix = bundleName.substr(2, 4); // including leading underscore
        if (proponentSuffix.size() == 4
            && (sceneName.size() < 4
                || sceneName.substr(sceneName.size() - 4, 4) != proponentSuffix)) {
            sceneName += proponentSuffix;
        }
    }

    renderer->sceneFile =
        getDirectory() + LSDF_DIRECTORY_SEPARATOR + sceneName + MPEGI_BITSTREAM_SUFFIX;
    int err = renderer->initializeRenderer(self);

    // checking the scene for errors
    if (err != 0 || renderer->fs == 0 || renderer->B == 0) {
        try {
            auto test = std::make_unique<MPEGI::Bitstream>(renderer->sceneFile);
        } catch (EAR::Exception& e) {
            printStatusMessage(self, e.toString());
            err = 6;
        } catch (std::exception& e) {
            printStatusMessage(self, std::string(e.what()));
            err = 7;
        }
    }

    if (renderer->streamManager
        && self->connectedInlets < renderer->streamManager->getNbrOfChannels()) {
        std::stringstream ss;
        ss << "Not enough inlets connected (EIF config requires at least "
           << renderer->streamManager->getNbrOfChannels() << " channels).";
        printStatusMessage(self, ss.str());
        err = 5;
    }

    if (renderer->hrirFile.empty()) {
        printStatusMessage(self, "No HRIR file loaded.");
        err = 8;
    }

    setStatus(self, err);

    if (err == 0) {
        std::stringstream ss;
        ss << "Successfully loaded scene: " << renderer->sceneFile;
        printStatusMessage(self, ss.str());
    }
}

void renderer_loadLSDF(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    t_symbol* sym = atom_getsym(argv);

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "loadLSDF from path: " << sym->s_name;
    printStatusMessage(self, ss.str());
#endif

    setStatus(self, -1);

    std::string filepath(sym->s_name);
    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    // empty string resets the LSDF file
    if (filepath == "") {
        renderer->lsdfObject = std::unique_ptr<LSDF::LSDFParser>(nullptr);
        printStatusMessage(self, "LSDF file reset (none will be used for next scene load).");
        return;
    }

    int err = 0;
    try {
        auto predelayOffset = MPEGI::Config::getFloatDef("Reverb:Predelay_Offset", 0.0f);
        renderer->lsdfObject =
            std::make_unique<LSDF::LSDFParser>(filepath, predelayOffset, false, true);
    } catch (LSDF::Exception& e) {
        printStatusMessage(self, e.toString());
        err = 9;
    } catch (std::exception& e) {
        printStatusMessage(self, std::string(e.what()));
        err = 10;
    }

    setStatus(self, err);

    if (err == 0) {
        std::stringstream ss;
        ss << "Successfully loaded LSDF file: " << filepath;
        printStatusMessage(self, ss.str());

        // try to reinitialize scene if it is already given
        if (renderer->sceneFile != "") {
            err = renderer->initializeRenderer(self);
            setStatus(self, err);
        }
    }
}

void renderer_loadHRIRSet(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    t_symbol* sym = atom_getsym(argv);

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss2;
    ss2 << "loadHRIRset from path: " << sym->s_name;
    printStatusMessage(self, ss2.str());
#endif

    auto renderer = static_cast<Renderer*>(self->rendererPayload);
    renderer->hrirFile = std::string(sym->s_name);

    if (!checkIfFileExists(renderer->hrirFile)) {
        std::stringstream ss;
        ss << "HRIR file \"" << renderer->hrirFile << "\" does not exist.";
        printStatusMessage(self, ss.str());
        renderer->hrirFile = "";
        return;
    }

    int err = renderer->initializeRenderer(self);
    setStatus(self, err);

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss3;
    ss3 << " -- initialized renderer with error: " << err;
    printStatusMessage(self, ss3.str());
#endif
    if (err == 0) {
        std::stringstream ss;
        ss << "Loaded new HRIR set: " << renderer->hrirFile;
        printStatusMessage(self, ss.str());
    }
}

void renderer_setMaxDelay(t_renderer* self, long dly)
{
    // noop
#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "Set maximum delay to " << dly << " samples (noop)";
    printStatusMessage(self, ss.str());
#endif
}

// void renderer_setLoop(
//     t_renderer* self,
//     t_symbol* s, long argc, t_atom* argv
// ) {
//     if (argc != 2) {
//         printStatusMessage(self, "Wrong number of arguments sent to \"setLoop\".");

//         return;
//     }

//     // TODO: implement

// #if MPEGI_VERBOSE_LOGGING
//     std::stringstream ss;
//     ss << "Setting loop from " << atom_getfloat(argv) << " to " << atom_getfloat(argv + 1) << " seconds";
//     printStatusMessage(self, ss.str());
// #endif
// }

// void renderer_clearLoop(
//     t_renderer* self
// ) {

//     // TODO: implement

// #if MPEGI_VERBOSE_LOGGING
//     printStatusMessage(self, "Clearing loop");
// #endif
// }

void renderer_userPose(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    if (argc != 13) {
        printStatusMessage(self, "Wrong number of arguments sent to \"userPose\".");

        return;
    }

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    if (renderer->scene) {
        MPEGI::Update upd;

        upd.addModification("realListener", "location",
            EAR::Vector3(atom_getfloat(argv + 6), atom_getfloat(argv + 7), atom_getfloat(argv + 8)),
            true);

        upd.addModification("realListener", "orientation",
            EAR::Euler(atom_getfloat(argv + 9), atom_getfloat(argv + 10), atom_getfloat(argv + 11)),
            true);

        upd.addModification("listener", "location",
            EAR::Vector3(atom_getfloat(argv), atom_getfloat(argv + 1), atom_getfloat(argv + 2)),
            (atom_getlong(argv + 12) == 1));

        upd.addModification("listener", "orientation",
            EAR::Euler(atom_getfloat(argv + 3), atom_getfloat(argv + 4), atom_getfloat(argv + 5)),
            false);

        renderer->scene->triggerUpdate(upd);
    }
}

void renderer_accessibility(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    if (argc != 14) {
        printStatusMessage(self, "Wrong number of arguments sent to \"accessibility\".");
        return;
    }

    MPEGI::Config::set("EarlyReflectionsEnabled", EAR::Variant((bool)atom_getlong(argv)));
    MPEGI::Config::set("ErAttenuationDb", EAR::Variant((float)atom_getfloat(argv + 1)));
    MPEGI::Config::set("Reverb:Enabled", EAR::Variant((bool)atom_getlong(argv + 2)));
    MPEGI::Config::set("ReverbAttenuationDb", EAR::Variant((float)atom_getfloat(argv + 3)));
    //MPEGI::Config::set("Distance_hasDistExp", EAR::Variant((bool)atom_getlong(argv + 4)));
    MPEGI::Config::set("Distance_distExp", EAR::Variant((float)atom_getfloat(argv + 5)));
    MPEGI::Config::set("DirectionalFocus_enabled", EAR::Variant((bool)atom_getlong(argv + 6)));
    MPEGI::Config::set("DirectionalFocus_Aperture", EAR::Variant((float)atom_getfloat(argv + 7)));
    MPEGI::Config::set(
        "DirectionalFocus_TransitionWidth", EAR::Variant((float)atom_getfloat(argv + 8)));
    MPEGI::Config::set(
        "DirectionalFocus_StopbandAttenuationDb", EAR::Variant((float)atom_getfloat(argv + 9)));
    MPEGI::Config::set(
        "DirectionalFocus_hasDirection", EAR::Variant((bool)atom_getlong(argv + 10)));
    MPEGI::Config::set(
        "DirectionalFocus_directionYaw", EAR::Variant((float)atom_getfloat(argv + 11)));
    MPEGI::Config::set(
        "DirectionalFocus_directionPitch", EAR::Variant((float)atom_getfloat(argv + 12)));
}

void renderer_position(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    if (argc != 5) {
        printStatusMessage(self, "Wrong number of arguments sent to \"position\".");

        return;
    }

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "called position update";

    t_atom* ap;
    int i;
    for (i = 0, ap = argv; i < argc; i++, ap++) {
        switch (atom_gettype(ap)) {
        case A_LONG:
            ss << ", arg long: " << atom_getlong(ap);
            break;
        case A_FLOAT:
            ss << ", arg float: " << atom_getfloat(ap);
            break;
        case A_SYM:
            ss << ", arg sym: " << atom_getsym(ap)->s_name;
            break;
        default:
            ss << ", arg of unknown type: " << atom_gettype(ap);
            break;
        }
    }

    printStatusMessage(self, ss.str());
#endif

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    if (renderer->scene) {
        MPEGI::Update upd;

        upd.addModification(atom_getsym(argv)->s_name, "location",
            EAR::Vector3(atom_getfloat(argv + 1), atom_getfloat(argv + 2), atom_getfloat(argv + 3)),
            (atom_getlong(argv + 4) == 1));
        // TODO: Velocity
        // upd.addModification("listener", "velocity", EAR::Vector3(args[4], args[5], args[6]), false);

        renderer->scene->triggerUpdate(upd);
    }
}

void renderer_orientation(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    if (argc != 4) {
        printStatusMessage(self, "Wrong number of arguments sent to \"orientation\".");

        return;
    }

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "called orientation update";

    t_atom* ap;
    int i;
    for (i = 0, ap = argv; i < argc; i++, ap++) {
        switch (atom_gettype(ap)) {
        case A_LONG:
            ss << ", arg long: " << atom_getlong(ap);
            break;
        case A_FLOAT:
            ss << ", arg float: " << atom_getfloat(ap);
            break;
        case A_SYM:
            ss << ", arg sym: " << atom_getsym(ap)->s_name;
            break;
        default:
            ss << ", arg of unknown type: " << atom_gettype(ap);
            break;
        }
    }

    printStatusMessage(self, ss.str());
#endif

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    if (renderer->scene) {
        MPEGI::Update upd;

        upd.addModification(atom_getsym(argv)->s_name, "orientation",
            EAR::Euler(atom_getfloat(argv + 1), atom_getfloat(argv + 2), atom_getfloat(argv + 3)),
            false);

        renderer->scene->triggerUpdate(upd);
    }
}

void renderer_gain(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    if (argc != 2) {
        printStatusMessage(self, "Wrong number of arguments sent to \"gain\".");

        return;
    }

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "called gain update";

    t_atom* ap;
    int i;
    for (i = 0, ap = argv; i < argc; i++, ap++) {
        switch (atom_gettype(ap)) {
        case A_LONG:
            ss << ", arg long: " << atom_getlong(ap);
            break;
        case A_FLOAT:
            ss << ", arg float: " << atom_getfloat(ap);
            break;
        case A_SYM:
            ss << ", arg sym: " << atom_getsym(ap)->s_name;
            break;
        default:
            ss << ", arg of unknown type: " << atom_gettype(ap);
            break;
        }
    }

    printStatusMessage(self, ss.str());
#endif

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    if (renderer->scene) {
        MPEGI::Update upd;

        upd.addModification(
            atom_getsym(argv)->s_name, "gain", static_cast<float>(atom_getfloat(argv + 1)), false);

        renderer->scene->triggerUpdate(upd);
    }
}

void renderer_triggerUpdate(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    int index = atom_getlong(argv);

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "Trigger update by int index " << index;
    printStatusMessage(self, ss.str());
#endif

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    std::vector<std::pair<int, float>> placeholderValues;

    for (int i = 1; i < argc; i++) {
        placeholderValues.push_back(std::make_pair(i - 1, atom_getfloat(argv + i)));
    }

    if (renderer->scene) {
        try {
            renderer->scene->triggerUpdateByIndex(index, placeholderValues);
        } catch (EAR::Exception& e) {
            printStatusMessage(self, e.toString());
        }
    }
}

void renderer_triggerUpdateString(t_renderer* self, t_symbol* s, long argc, t_atom* argv)
{

    t_symbol* sym = atom_getsym(argv);

#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "update triggered: " << sym->s_name << ", ";
    ss << "update thread id: " << std::this_thread::get_id();
    printStatusMessage(self, ss.str());
#endif

    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    std::vector<std::pair<int, float>> placeholderValues;

    for (int i = 1; i < argc; i++) {
        placeholderValues.push_back(std::make_pair(i - 1, atom_getfloat(argv + i)));
    }

    if (renderer->scene) {
        try {
            renderer->scene->triggerUpdateById(std::string(sym->s_name), placeholderValues);
        } catch (EAR::Exception& e) {
            printStatusMessage(self, e.toString());
        }
    }
}

#ifndef MPEGI_DISABLE_OSC_LISTENER
void renderer_startRecordingOsc(t_renderer* self)
{
    auto renderer = static_cast<Renderer*>(self->rendererPayload);
    if (renderer->scene) {
        renderer->scene->getOSCListener()->startRecordingReceivedPackets();
    }
}

void renderer_stopRecordingOsc(t_renderer* self)
{
    auto renderer = static_cast<Renderer*>(self->rendererPayload);
    if (renderer->scene) {
        renderer->scene->getOSCListener()->stopRecordingReceivedPackets();
    }
}

void renderer_saveRecordedOsc(t_renderer* self, t_symbol* filename)
{
    auto renderer = static_cast<Renderer*>(self->rendererPayload);
    if (renderer->scene) {
        std::ofstream out(filename->s_name);
        out << renderer->scene->getOSCListener()->serializeRecording();
        out.close();
    }
}
#endif

void renderer_dsp64(t_renderer* self, t_object* dsp64, short* count, double samplerate,
    long maxvectorsize, long flags)
{
#if MPEGI_VERBOSE_LOGGING
    std::stringstream ss;
    ss << "dsp64 with fs=" << samplerate << ", B=" << maxvectorsize;
    printStatusMessage(self, ss.str());
#endif
    setStatus(self, -1);

    self->connectedInlets = 0;
    for (int i = 0; i < self->numberOfInputs; i++) {
        if (count[i]) {
            self->connectedInlets++;
        }
    }
    auto renderer = static_cast<Renderer*>(self->rendererPayload);
    renderer->fs = samplerate;
    renderer->B = maxvectorsize;
    int err = renderer->initializeRenderer(self);

    if (renderer->streamManager
        && self->connectedInlets < renderer->streamManager->getNbrOfChannels()) {
        std::stringstream ss;
        ss << "Not enough inlets connected (EIF config requires at least "
           << renderer->streamManager->getNbrOfChannels() << " channels).";
        printStatusMessage(self, ss.str());
        err = 5;
    }

    if (renderer->hrirFile.empty()) {
        printStatusMessage(self, "No HRIR file loaded.");
        err = 8;
    }

    auto number_spatializer_channels = renderer->temp->channels();
    if (number_spatializer_channels != self->numberOfOutputs) {
        std::stringstream ss;
        ss << "Mismatch between number of spatializer channels and external outlets is not allowed "
              "! "
           << number_spatializer_channels << " spatializer channels (LSDF) and "
           << self->numberOfOutputs << " outlets !";
        printStatusMessage(self, ss.str());
    }

    setStatus(self, err);

    object_method_direct(void, (t_object*, t_object*, t_perfroutine64, long, void*), dsp64,
        gensym("dsp_add64"), (t_object*)self, (t_perfroutine64)renderer_perform64, 0, NULL);
}

void renderer_perform64(t_renderer* self, t_object* dsp64, double** ins, long numins, double** outs,
    long numouts, long sampleframes, long flags, void* userparam)
{
    auto const timer = self->detector->start();
    auto renderer = static_cast<Renderer*>(self->rendererPayload);

    if (!renderer->processBlockMutex.try_lock()) {
        return;
    }

    if (renderer->scene) {

        if (self->connectedInlets < renderer->streamManager->getNbrOfChannels()) {
            renderer->processBlockMutex.unlock();
            return;
        }

        // ingest new samples from MaxMSP
        renderer->streamManager->ingest(ins);
        {
            auto guard = MPEGI::TemporarilyDisableDenormals{};
            renderer->renderer->processBlock(renderer->temp->data().get());
        }

        auto renderer_channels = renderer->temp->channels();
        assert(renderer_channels <= numouts);
        for (int ch = 0; ch < renderer_channels; ch++) {
            renderer->temp->ch(ch).readDbl(outs[ch], 0, renderer->B);
        }
        for (int ch = renderer_channels; ch < numouts; ch++) {
            std::fill(outs[ch], outs[ch] + sampleframes, 0.0);
        }

        renderer->sceneTime += renderer->tpb;
        if (renderer->loopEnd > renderer->loopStart && renderer->sceneTime >= renderer->loopEnd) {
            // accounts for blocksize not fitting into loop
            renderer->sceneTime = renderer->loopStart + (renderer->sceneTime - renderer->loopEnd);
        }
        renderer->scene->setTime(renderer->sceneTime);
    }

    renderer->processBlockMutex.unlock();
}
