/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef RENDERER_TILDE_H
#define RENDERER_TILDE_H

#include <memory>
#include <mutex>
#include <string>

#include "detector/detector.h"

#include "c74_msp.h"
using namespace c74::max;

#define MPEGI_CONTROL_RATE 60 // Hz
#define MPEGI_VERBOSE_LOGGING 0 // bool

namespace EAR {
class RealFrame;
}

namespace MPEGI {
class Clock;
class MaxStreamManager;
class Renderer;
class Scene;
} // namespace MPEGI

namespace LSDF {
class LSDFParser;
}

// struct to represent the object's state
typedef struct _renderer
{
    t_pxobject ob;

    long numberOfInputs;
    long numberOfOutputs;

    void* messageOutlet;

    detector_t* detector; // detector instance

    long proxyNumber; // just used by SDKs proxy function
    void* proxyInlet;

    long connectedInlets;

    void* rendererPayload; // this is all data needed by the C++ renderer implementation
} t_renderer;

struct Renderer
{
    float fs = 0;
    std::size_t B = 0;
    float tpb = 0;
    double sceneTime = 0, loopStart = 0, loopEnd = 0;

    std::string sceneFile;
    std::string hrirFile;
    std::string configFile;

    std::unique_ptr<EAR::RealFrame> temp;

    std::unique_ptr<LSDF::LSDFParser> lsdfObject = std::unique_ptr<LSDF::LSDFParser>(nullptr);
    std::shared_ptr<MPEGI::Clock> clock;
    std::unique_ptr<MPEGI::Scene> scene;
    std::unique_ptr<MPEGI::Renderer> renderer;
    std::shared_ptr<MPEGI::MaxStreamManager> streamManager;

    std::mutex processBlockMutex;

    int initializeRenderer(t_renderer* self);
};

// method prototypes
void* renderer_new(t_symbol* sym, long argc, t_atom* argv);

void renderer_free(t_renderer* self);

void renderer_assist(t_renderer* self, void* unused, long io, long index, char* string_dest);

void renderer_loadScene(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_loadLSDF(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_loadHRIRSet(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

// Registered callback for the "setMaxDelay" message
void renderer_setMaxDelay(t_renderer* self, // pointer to the renderer data structure defined above
    long dly // maximum delay in samples
);

// // Registered callback for the "setLoop" message
// void renderer_setLoop(
//     t_renderer* self, // pointer to the renderer data structure defined above
//     t_symbol* s, // pointer to the symbol that holds the message name
//     long argc, // number of arguments in the A_GIMME message
//     t_atom* argv // pointer to arguments (1. float start time [s], 2. float end time [s])
// );

// // Registered callback for the "clearLoop" message
// void renderer_clearLoop(
//     t_renderer* self // pointer to the renderer data structure defined above
// );

void renderer_userPose(t_renderer* self, // pointer to the renderer instance
    t_symbol* s, // pointer to the symbol that holds the message name
    long argc, // number of arguments in the A_GIMME message (should be 13)
    t_atom*
        argv // pointer to arguments (wcs_x, wcs_y, wcs_z, wcs_yaw, wcs_pitch, wcs_roll, rcs_x, rcs_y, rcs_z, rcs_yaw, rcs_pitch, rcs_roll, teleport)
);

void renderer_accessibility(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_position(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_orientation(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_gain(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_triggerUpdate(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

void renderer_triggerUpdateString(t_renderer* self, t_symbol* s, long argc, t_atom* argv);

#ifndef MPEGI_DISABLE_OSC_LISTENER
void renderer_startRecordingOsc(t_renderer* self);
void renderer_stopRecordingOsc(t_renderer* self);
void renderer_saveRecordedOsc(t_renderer* self, t_symbol* filename);
#endif

void renderer_dsp64(t_renderer* self, t_object* dsp64, short* count, double samplerate,
    long maxvectorsize, long flags);

void renderer_perform64(t_renderer* self, t_object* dsp64, double** ins, long numins, double** outs,
    long numouts, long sampleframes, long flags, void* userparam);

#endif /* RENDERER_TILDE_H */
