/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_BINCURSOR
#define MPEGI_BINCURSOR

#include <MPEGI/Bitstream/Globals.h>

#include <cstdint>
#include <memory>
#include <string>
#include <vector>

namespace MPEGI {

//! Binary Cursor - load binary data, use bitwise traversal, and read C++ representations of the data
class BinCursor
{
public:
    //! Constructor
    BinCursor(const std::uint8_t* data, std::size_t length);

    //! Copy Constructor
    BinCursor(const BinCursor& cursor);

    //! Traverse the stream and return the value of the traversed bits as a boolean.
    /*!
     * Traversal length is 1.
     */
    bool readBool();

    //! Traverse the stream and return the value of the traversed bits as an unsigned int.
    /*!
     * \param length        the number of bits for traversal and integer representation
     */
    unsigned long long int readUInt(std::size_t length = 8);

    //! Traverse the stream and return the value of the traversed bits as an int.
    /*!
     * \param length        the number of bits for traversal and integer representation
     */
    long long int readSInt(std::size_t length = 8);

    //! Traverse the stream and return the value of the traversed bits as an int.
    /*!
     * Number of bits used scales with the value
     *
     * \param l1    minimum number of bits
     * \param l2    if value is >= (2^l1 - 1), the new size is l1+l2 bits
     * \param l3    if value is >= (2^l1 - 1 + 2^l2 - 1), the new size is l1+l2+l3 bits
     *
     * Example:
     * Using "readEscapedUInt(2, 3, 4)", these values result in the following bitstream representation:
     *
     * 1  -> 0b01
     * 2  -> 0b10
     * 3  -> 0b11000
     * 4  -> 0b11001
     * 9  -> 0b11110
     * 10 -> 0b111110000
     * 25 -> 0b111111111
     *
     * The maximum value that you can represent in this format is the following:
     * maxValue = (2^l1 - 1) + (2^l2 - 1) + (2^l3 - 1)
     */
    unsigned int readEscapedUInt(std::size_t l1, std::size_t l2, std::size_t l3);

    //! Traverse the stream and return the value of the traversed bits as a float.
    /*!
     * Traversal length is 32.
     */
    float readFloat();

    //! Traverse the stream and return the value of the traversed bits as a decoded integer.
    /*!
     * \param quantizationRange     the data struct that lists, min, max, bit width values for quantisation
     */
    float readQuantFloat(QuantizationRange qRange);

    //! Return a sequence of raw bytes read from the bitstream
    /*!
     * The data is returned as a copy, as the start of the data is not necessarily byte-aligned.
     *
     * \param length   length of blob in bytes will be written into the value given
     */
    std::unique_ptr<std::uint8_t[]> readBlob(std::size_t& lengthOfBlob);

    //! Return a sequence of raw bytes read from the bitstream
    /*!
     * The data is returned as a copy, as the start of the data is not necessarily byte-aligned.
     */
    std::shared_ptr<std::vector<std::uint8_t>> readBlob();

    //! Traverse the stream and return the value of the traversed bits as a std::string.
    /*!
     * Traversal length is variable until 0-terminating byte is reached.
     */
    std::string readString();

    //! Move the cursor position ahead effectively skipping data
    /*!
     * \param length    The number of bits to skip ahead
     */
    void movePosRight(std::size_t length);

    //! Move the cursor position behind effectively skipping data
    /*!
     * \param length    The number of bits to skip behind
     */
    void movePosLeft(std::size_t length);

    //! Traverse the stream and return the first variable length data as a unsigned int that
    //! represents an index or size/count.
    /*!
     * Traversal length is variable.
     */
    unsigned int getCountOrIndex();

    //! Traverse the stream and return the first variable length data as a float that
    //! represents a duration.
    /*!
     * Traversal length is variable.
     */
    float getDuration(bool isLongRangeMode);

    //! Traverse the stream and return the first variable length data as a unsigned int that
    //! represents an ID.
    /*!
     * Traversal length is variable.
     */
    unsigned int getID();

    //! Traverse the stream and return the first variable length data as a unsigned int that
    //! represents an ID.
    /*!
     * Traversal length is variable.
     */
    float getDistance(bool isSmallScene);

    //! Traverse the stream and return the first variable length data as an array with 3 values that
    //! represent a position (x, y, z).
    /*!
     * Traversal length is 96.
     */
    std::array<float, 3> getPosition(bool isSmallScene);

    //! Traverse the stream and return the first variable length data as an array with 3 values that
    //! represent an orientation (x, y, z).
    /*!
     * Traversal length is variable.
     */
    std::array<float, 3> getOrientation();

    //! Traverse the stream and return the first variable length data as a float that
    //! represents a gain.
    /*!
     * Traversal length is variable.
     */
    float getGain(bool isHiPrecGain);

    //! Traverse the stream and return the first variable length data as a float that
    //! represents a frequency.
    /*!
     * Traversal length is variable.
     */
    float getFrequency();

    //! Reads a code word on the lookup table with index 'tableName' and returns its value if found.
    //! If the value is not found, an exception of type EAR_FAIL is triggered.
    /*!
     * Traversal length is variable.
     * \param tableName    The index of the lookup table.
     */
    CodeWordResult readCodeWord(const std::string& tableName);

    //! Resize the end of data marker for the cursor - Exceptions are thrown when reading beyond EOF
    /*!
     * \param length    The number of bytes to be the new length of the cursor
     */
    inline void resize(std::size_t length) { dataEnd = length * 8; };

    //! Return the current cursor position in bits
    inline std::size_t getPosition() const { return position; }

    //! Move cursor to the start of the next byte
    inline void byteAlign()
    {
        while (!this->isByteAligned()) {
            this->movePosRight(1);
        }
    }

    //! Returns true if the cursor position is a integer multiple of 8, i.e. aligned to the byte
    inline bool isByteAligned() const { return position % 8 == 0; };

    //! Returns true if the cursor has begun traversing the last byte of the data
    /*
     *  Note: there is no "isConsumed" function because there's no way to know if
     *        the last byte is padded. Since the encoder is using bitwise data
     *        it may be the case that the encoder pads the last byte.
     */
    inline bool atLastByte() const { return (position > (dataEnd - 8)) && (position <= dataEnd); }

private:
    const std::uint8_t* data;
    std::size_t dataStart; //!< start of data in bits
    std::size_t dataEnd; //!< end of data in bits
    std::size_t position; //!< position in bits

    template <typename T> T readBits(std::size_t nBits);
};

} // namespace MPEGI

#endif // MPEGI_BINCURSOR
