/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_BITSTREAM
#define MPEGI_BITSTREAM

#include <EAR/HeadphoneEQ.h>
#include <MPEGI/Bitstream/Globals.h>
#include <MPEGI/Bitstream/Update.h>
#include <MPEGI/Helpers/Types.h>

#include <cstddef>
#include <map>
#include <memory>
#include <optional>
#include <string>
#include <utility>
#include <vector>

namespace MPEGI {

// This is implemented in the auto-generated file RegisterPayloadHandlers.cpp:
void registerPayloadHandlers();

struct FreqGrid
{
    FreqGrid(const unsigned int _nrBands, const std::vector<float>& _centerFreq)
        : nrBands(_nrBands)
        , centerFreq(_centerFreq)
    {}

    unsigned int nrBands;
    std::vector<float> centerFreq;
};

class Bitstream;

//! A container for all configutation parameters deserialised from config packets
class BitstreamConfig
{
public:
    //! Default ER tuning gain [dB]
    inline static float const early_tuning_gain_default_db = -7.95880017344075f;

    //! Number of samples for distance propogatio buffers.
    unsigned int delayBufferSize;

    //! Threshold for culling quiet sources in magnitude
    float gainCullingThreshold;

    bool enableCullingReflectionRI;

    //! Threshold for loudness based culling (magnitude)
    float loudnessCullingThreshold{ 1.f };

    //! TODO: Add docs
    int humidity;

    //! TODO: Add docs
    int temperature;

    //! TODO: Add docs
    int speedOfSound;

    bool isSmallScene;

    bool hasConsolidationThreshold;

    int consolidationThreshold;

    //! Flag that allows adding locally captured elements to the scene
    bool enableLocallyCapturedAudio;

    //! Directivity ID to be used for Listener Voice. If not given by bitstream renderer default is used
    std::string listenerVoiceDirectivityId;
    //! Return map that resolves integer ids to string ids
    inline const std::map<int, std::string>& getIdMap() const { return intIdToStringId; };

    const std::vector<std::shared_ptr<FreqGrid>>& GetFreqGrid() const { return freqGridData; }

    /**
     * Get the Early Reflection tuning gain
     * @return Tuning gain in dB
     */
    float getEarlyTuningGain() const
    {
        return earlyTuningGainDb.has_value() ? earlyTuningGainDb.value()
                                             : early_tuning_gain_default_db;
    }

    bool useLowComplexityER;

    bool hasHeadphoneEQ;

    const std::shared_ptr<EAR::HeadphoneEQData>& GetContentCreatorHeadphoneEQ() const
    {
        return contentCreatorHP;
    }

private:
    friend Bitstream;

    BitstreamConfig() = default;

    std::map<int, std::string> intIdToStringId;

    std::vector<std::shared_ptr<FreqGrid>> freqGridData;

    std::optional<float> earlyTuningGainDb; //!< Early reflection tuning gain [dB], if present

    std::shared_ptr<EAR::HeadphoneEQData> contentCreatorHP;
};

//! A class to load and read from the bitstream
class Bitstream
{
public:
    //! A Bitstream::RawPayload represents an unparsed MHAS payload (should only be used by PayloadHandlers)
    class RawPayload;

    //! A Bitstream::RawPayloadForTesting represents an unparsed MHAS payload (should only be used for testing)
    class RawPayloadForTesting;

    //! A Bitstream::Payload is the base class for readily parsed payloads
    class Payload;

    //! A Bistream::PayloadHandler is an interface for a class that reads received payloads of a specific type
    class PayloadHandler;

    //! MHAS Packets
    enum MHASPacketType
    {
        PACTYP_FILLDATA = 0,
        PACTYP_MPEGH3DACFG = 1,
        PACTYP_MPEGH3DAFRAME = 2,
        PACTYP_SYNC = 6,
        PACTYP_SYNCGAP = 7,
        PACTYP_MARKER = 8,
        PACTYP_CRC16 = 9,
        PACTYP_CRC32 = 10,
        PACTYP_DESCRIPTOR = 11,
        PACTYP_USERINTERACTION = 12,
        PACTYP_LOUDNESS_DRC = 13,
        PACTYP_BUFFERINFO = 14,

        // MPEG-I Defined Packets
        PACTYP_SCENE_CFG = 23,
        PACTYP_SCENE_PLD = 24,
        PACTYP_SCENE_UPD = 25,
    };

    //! MHAS ScenePayloadPacket Payload element types
    enum PayloadType
    {
        PLD_ACOUSTIC_ENVIRONMENT = 0,
        PLD_ACOUSTIC_MATERIAL = 1,
        PLD_ANCHOR = 2,
        PLD_AUDIO_STREAM = 3,
        PLD_CHANNEL_SOURCE = 4,
        PLD_CONDITION = 5,
        PLD_DIRECTIVITY = 6,
        PLD_HOA_SOURCE = 7,
        PLD_MESH = 8,
        PLD_OBJECT_SOURCE = 9,
        PLD_DIFFR = 10,
        PLD_ER = 11,
        PLD_PORTAL = 12,
        PLD_MPHOA = 13,
        PLD_REVERB = 14,
        PLD_AUDIO_PLUS = 15,
        PLD_LCER = 16,
        PLD_DISPF = 17,
        PLD_VOXDATA = 18,
        PLD_AIRFLOW = 19,
        PLD_RASTERMAP = 20,
        PLD_GRANULAR = 22,

        // These types are used during development and should be removed in the future
        PLD_SCENE = 31,

        PLD_RMS = 30,
    };

    static constexpr const char* payloadTypeToString(PayloadType payloadType)
    {
        switch (payloadType) {
        case PayloadType::PLD_ACOUSTIC_ENVIRONMENT:
            return "AcousticEnvironment";
        case PayloadType::PLD_ACOUSTIC_MATERIAL:
            return "AcousticMaterial";
        case PayloadType::PLD_AIRFLOW:
            return "Airflow";
        case PayloadType::PLD_ANCHOR:
            return "Anchor";
        case PayloadType::PLD_AUDIO_STREAM:
            return "AudioStream";
        case PayloadType::PLD_CHANNEL_SOURCE:
            return "ChannelSource";
        case PayloadType::PLD_CONDITION:
            return "Condition";
        case PayloadType::PLD_DIRECTIVITY:
            return "Directivity";
        case PayloadType::PLD_HOA_SOURCE:
            return "HOASource";
        case PayloadType::PLD_MESH:
            return "Mesh";
        case PayloadType::PLD_OBJECT_SOURCE:
            return "ObjectSource";
        case PayloadType::PLD_DIFFR:
            return "Diffraction";
        case PayloadType::PLD_ER:
            return "Early Reflection";
        case PayloadType::PLD_VOXDATA:
            return "Voxel Data";
        case PayloadType::PLD_PORTAL:
            return "Portal";
        case PayloadType::PLD_MPHOA:
            return "MultipointHOA";
        case PayloadType::PLD_REVERB:
            return "Reverb";
        case PayloadType::PLD_AUDIO_PLUS:
            return "Audio plus";
        case PayloadType::PLD_LCER:
            return "Low Complexity Early Reflection";
        case PayloadType::PLD_GRANULAR:
            return "Granular Synthesis";
        case PayloadType::PLD_SCENE:
            return "Scene";
        case PayloadType::PLD_RMS:
            return "RMS";
        case PayloadType::PLD_RASTERMAP:
            return "RasterMap";
        default:
            return "Unknown Payload Type";
        }
    }

    //! Constructor with bitstring
    /*!
     * \param filepath          full path to a bitstream file
     */
    Bitstream(const std::string& filepath);

    struct PayloadHandlerAndType
    {
        std::shared_ptr<PayloadHandler> handler;
        PayloadType type;
    };

    using CreateHandlerInstanceFunc = PayloadHandlerAndType (*)();

    //! Introduce a class derived from PayloadHandler so it can be registered at construction of a bitstream
    /*!
     * usage: Bitstream::register<MyPayloadHandler>(PayloadType);
     * \param type              the payload type of interest for this handler (see enum Bitstream::PayloadTypes)
     */
    static void introduceHandler(CreateHandlerInstanceFunc func)
    {
        getHandlerList().push_back(func);
    }

    static void clearHandlerList();

    //! Register as a handler for a specific payload type
    /*!
     * \param handler           shared pointer to a handler instance to receive all payloads of a given type
     * \param type              the payload type of interest (see enum Bitstream::PayloadTypes)
     * \param acceptStreamed    if true, the handler is also called when a payload of interest
     *                          is streamed at runtime (TODO: not implemented)
     */
    void registerHandler(std::shared_ptr<PayloadHandler> handler, PayloadType type);

    //! Get all payloads of a given type
    /*!
     * \param type              the payload type of interest (see enum Bitstream::PayloadTypes)
     * \return                  vector of shared_ptrs to Payload instances, that can be casted to the derived type
     */
    virtual std::vector<std::shared_ptr<const Payload>> getAllPayloadsOfType(
        PayloadType type) const;

    //! Get the only payload of a given type
    /*!
     * This function terminates with the first payload of the given type found.
     * It throws an exception if there is not exactly one of the type.
     *
     * \param type              the payload type of interest (see enum Bitstream::PayloadTypes)
     * \return                  shared_ptr to Payload instance, that can be casted to the derived type
     */
    virtual std::shared_ptr<const Payload> getOnlyPayloadOfType(PayloadType type) const;

    inline const std::vector<BitstreamUpdate>& getUpdates() const { return updates; };

    BitstreamConfig config;

    inline const bool& getEnableLocallyCapturedAudio() const
    {
        return config.enableLocallyCapturedAudio;
    }

    inline const std::string& getListenerVoiceDirectivityId() const
    {
        return config.listenerVoiceDirectivityId;
    }

    // TODO: add option to receive new data at runtime

protected:
    //! Constructs an empty bitstream
    Bitstream() = default;

private:
    using HandlerList = std::vector<CreateHandlerInstanceFunc>;

    static HandlerList& getHandlerList();

    std::size_t totalSize; // in bytes
    std::unique_ptr<const std::uint8_t[]> buffer; //!< Immutable buffer for the file data
    std::vector<RawPayload> rawPayloads;
    std::map<PayloadType, std::vector<std::shared_ptr<const Payload>>> payloads;
    std::vector<BitstreamUpdate> updates;
};

class Bitstream::RawPayload
{
public:
    ~RawPayload(){};

    //! Traverse the stream and return the value of the traversed bits as a boolean.
    /*!
     * Traversal length is 1.
     */
    bool readBool() const;

    //! Traverse the stream and return the value of the traversed bits as an int.
    /*!
     * \param length        the number of bits used for representing the integer
     */
    int readInt(std::size_t length) const;

    //! Traverse the stream and return the value of the traversed bits as T.
    /*!
     * T can be either a signed or an unsigned integer type.
     * Size of T needs to be at least N bits.
     * Traversal length is N.
     */
    template <typename T, int N> T readInteger() const
    {
        static_assert(sizeof(T) * 8 >= N, "Target type too short");

        // Use `readUInt` to read the bits as `readBits` is in `detail` namespace.
        static_assert(sizeof(unsigned int) * 8 >= N, "Requested bit size too high");
        struct
        {
            T v : N;
        } extend{ static_cast<T>(readUInt(N)) };

        return extend.v;
    }

    //! Traverse the stream and return the value of the traversed bits as an unsigned int.
    /*!
     * \param length        the number of bits used for representing the integer
     */
    unsigned int readUInt(std::size_t length) const;

    //! Traverse the stream and return the value of the traversed bits as a float.
    /*!
     * Traversal length is 32.
     */
    float readFloat() const;

    //! Traverse the stream and return the value of the traversed bits as a decoded integer.
    /*!
     * \param quantizationRange     the data struct that lists, min, max, bit width values for quantisation
     */
    float readQuantFloat(QuantizationRange quantizationRange) const;

    //! Traverse the stream and return the value of the traversed bits as a std::string.
    /*!
     * Traversal length is variable until 0-terminating byte is reached.
     */
    std::string readString() const;

    //! Traverse the stream and return the value of the traversed bits as an int.
    /*!
     * Number of bits used scales with the value
     *
     * \param l1    minimum number of bits
     * \param l2    if value is >= (2^l1 - 1), the new size is l1+l2 bits
     * \param l3    if value is >= (2^l1 - 1 + 2^l2 - 1), the new size is l1+l2+l3 bits
     *
     * Example:
     * Using "readEscapedUInt(2, 3, 4)", these values result in the following bitstream representation:
     *
     * 1  -> 0b01
     * 2  -> 0b10
     * 3  -> 0b11000
     * 4  -> 0b11001
     * 9  -> 0b11110
     * 10 -> 0b111110000
     * 25 -> 0b111111111
     *
     * The maximum value that you can represent in this format is the following:
     * maxValue = (2^l1 - 1) + (2^l2 - 1) + (2^l3 - 1)
     */
    int readEscapedUInt(std::size_t l1, std::size_t l2, std::size_t l3) const;

    //! Return a sequence of raw bytes read from the bitstream
    /*!
     * The data is returned as a copy, as the start of the data is not necessarily byte-aligned.
     *
     * \param length   length of blob in bytes will be written into the value given
     */
    std::unique_ptr<std::uint8_t[]> readBlob(std::size_t& length) const;

    //! Return a sequence of raw bytes read from the bitstream
    /*!
     * The data is returned as a copy, as the start of the data is not necessarily byte-aligned.
     */
    std::shared_ptr<std::vector<std::uint8_t>> readBlob() const;

    //! Return the size of the Payload in bytes
    inline std::size_t getSize() const { return bufferSize; }

    //! Return the payload label
    inline int getLabel() const { return label; }

    //! Return the payload type (see enum Bitstream::PayloadType)
    inline Bitstream::PayloadType getType() const { return type; }

    //! Returns true, if the position in the stream is within the last byte
    inline bool atLastByte() const
    {
        return (bitsRead > (bufferSize - 1) * 8) && (bitsRead <= bufferSize * 8);
    }

    //! Returns true if the whole rawPayload has been consumed (up to the buffer's byte size)
    bool isConsumed() const
    {
        auto const bytesRead = (bitsRead + 7) / 8;
        return bytesRead == bufferSize;
    }

    //! Byte-align the read-pointer
    void byteAlign() const;

    //! Traverse the stream and return the first variable length data as a unsigned int that
    //! represents an index or size/count.
    /*!
     * Traversal length is variable.
     */
    unsigned int getCountOrIndex() const;

    //! Traverse the stream and return the first variable length data as a float that
    //! represents a duration.
    /*!
     * Traversal length is variable.
     */
    float getDuration(bool isLongRangeMode) const;

    //! Traverse the stream and return the first variable length data as a unsigned int that
    //! represents an ID.
    /*!
     * Traversal length is variable.
     */
    unsigned int getID() const;

    //! Traverse the stream and return the first variable length data as a unsigned int that
    //! represents an ID.
    /*!
     * Traversal length is variable.
     */
    float getDistance(bool isSmallScene) const;

    //! Traverse the stream and return the first variable length data as an array with 3 values that
    //! represent a position (x, y, z).
    /*!
     * Traversal length is 96.
     */
    std::array<float, 3> getPosition(bool isSmallScene) const;

    //! Traverse the stream and return the first variable length data as an array with 3 values that
    //! represent an orientation (x, y, z).
    /*!
     * Traversal length is variable.
     */
    std::array<float, 3> getOrientation() const;

    //! Traverse the stream and return the first variable length data as a float that
    //! represents a gain.
    /*!
     * Traversal length is variable.
     */
    float getGain(bool isHiPrecGain) const;

    //! Traverse the stream and return the first variable length data as an array with 6 values that
    //! represent the percentage of scene region that remains outside of LSDF region,
    //! (max x, min x, max y, min y, max z, min z).
    /*!
     * Traversal length is variable.
     */
    std::array<float, 6> getTolerance() const;

    //! Traverse the stream and return the first variable length data as a float that
    //! represents a frequency.
    /*!
     * Traversal length is variable.
     */
    float getFrequency() const;

    //! Reads a code word on the lookup table with index 'tableName' and returns its value if found.
    //! If the value is not found, an exception of type EAR_FAIL is triggered.
    /*!
     * Traversal length is variable.
     * \param tableName    The index of the lookup table.
     */
    CodeWordResult readCodeWord(const std::string& tableName) const;

private:
    friend Bitstream; // this is the only class that can create a new payload

    RawPayload(Bitstream::PayloadType type, std::size_t length, const std::uint8_t* buffer);

    std::size_t bufferSize; // in bytes
    int label;
    Bitstream::PayloadType type;

    const std::uint8_t* buffer;
    std::size_t bitsRead;
};

class Bitstream::RawPayloadForTesting : public Bitstream::RawPayload
{
public:
    RawPayloadForTesting(std::string dataString);

    ~RawPayloadForTesting(){};

    const std::uint8_t* getBuffer() { return buffer; };

private:
    std::vector<std::uint8_t> vector_buffer;
};

class Bitstream::PayloadHandler
{
public:
    virtual ~PayloadHandler(){};

    //! Interface function that is called when a new RawPayload needs to be parsed
    /*!
     * \param payload        reference to the RawPacked to be parsed
     *                      (guaranteed to be of the type this Handler registered for)
     * \return              should return a parsed payload that inherits from Bitstream::Payload
     */
    virtual std::shared_ptr<const Bitstream::Payload> unpack(
        const Bitstream::RawPayload& payload, const BitstreamConfig& config) = 0;
};

class Bitstream::Payload
{
public:
    virtual ~Payload(){};
};

} // namespace MPEGI

#endif // MPEGI_BITSTREAM
