/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DIFFRACTION_PAYLOAD
#define MPEGI_DIFFRACTION_PAYLOAD

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/DiffractionEdge.h>
#include <MPEGI/Renderer/DiffractionPath.h>

#include <EAR/BaseTypes.h>
#include <EAR/Exception.h>

#include <unordered_map>
#include <utility>

namespace MPEGI {

// Provide a accessible C++ representation of an MPEG-I Diffraction type MHAS payload (temporary)
struct DiffractionPayload : public Bitstream::Payload
{
public:
    using VoxelPathDict = std::map<uint64_t, std::vector<std::pair<int, int>>>;
    using VoxelDict = std::map<std::vector<int>, std::vector<int>>;

protected:
    friend class DiffractionPayloadHandler; // Handler is allowed to create new payloads and fill in data
    friend class DiffractionMetadata; // Metadata interface is allowed to access member variables

    //! Constructor (only visible to DiffractionPayloadHandler)
    DiffractionPayload() = default;

private:
    // List of diffraction edges
    std::vector<DiffractionEdge> edgeList;

    // Dictionary of paths
    std::map<int, std::vector<DiffractionPath>> pathDict;

    // Dictionary of listener voxels
    VoxelDict listenerVoxelDict;

    // Dictionary of source voxels
    std::map<int, VoxelDict> sourceVoxelDicts; //sourceIdx, sourceVoxelDict

    // Dictionary of validated paths for possible source-listener pair - in case of stationary source
    std::map<int, VoxelPathDict> validPathDict;

    // For dynamic geometry handling
    // Map of dynamic diffraction edge lists
    std::map<int, std::vector<DiffractionEdge>> dynamicEdgeLists;

    // Map of dynamic diffraction path dictionaries
    std::map<int, std::map<int, std::vector<DiffractionPath>>> dynamicPathDicts;

    // List of static source indices
    std::vector<int> staticSourceIdxList;

    // Voxel grid information
    EAR::Vector3 voxelGridOrigin;
    EAR::Vector3 voxelGridPitch;
    VoxelCoord voxelGridShape;

    // Resolve geometry name -> bitstream ID
    std::unordered_map<std::string, int> geoName2BitsId;
};

constexpr uint64_t PackVoxelCoords(int i, int j, int k)
{
    // The triple (-1, -1, -1), which indicates an out-of-bounds position,
    // gets mapped to (1ull<<63), which does not correspond to any valid voxel.
    const unsigned int shift_width = (sizeof(int) * 8 - 1);
    const unsigned int sign_bit = 1 << shift_width;
    const unsigned int oob_correction =
        ((unsigned int)(i) & (unsigned int)(j) & (unsigned int)(k)&sign_bit) >> shift_width;
    const uint64_t stride = 1ull << 21;
    const uint64_t idx_mask = stride - 1;
    return ((k & idx_mask) * stride + (j & idx_mask)) * stride + (i & idx_mask) + oob_correction;
}

} // namespace MPEGI

#endif // MPEGI_DIFFRACTION_PAYLOAD
