/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_EIF_PAYLOAD
#define MPEGI_EIF_PAYLOAD

#include <EAR/SceneObject.h>
#include <MPEGI/Bitstream/Bitstream.h>

#include <EAR/BaseTypes.h>
#include <EAR/Variant.h>

#include <algorithm>
#include <map>
#include <memory>
#include <optional>
#include <string>
#include <vector>

// fwd declare
class IRCustomEIFPayloadHandler; // this hack is needed to make the IR generator in the CLI work

namespace MPEGI {

//! Provide a accessible C++ representation of an MPEG-I EIF type MHAS payload (temporary)
struct EIFPayload : public Bitstream::Payload
{
public:
    // Foward declares
    struct AcousticMaterialDesc;
    struct AnchorDesc;
    struct AudioStreamDesc;
    struct HoaSourceDesc;
    struct ObjectSourceDesc;
    struct SourceDirectivityDesc;
    struct TransformDesc;
    struct GranularSoundDesc;

    //! Superclass of every description element
    struct Description
    {
        //! unique indentifier as used in EIF
        std::string id;

        //! Bitstream-internal ID
        int bsId;
    };

    //! Representation of the scene metadata
    struct SceneInfo
    {
        //! Representation of either an AR or a VR scene
        enum class SceneType
        {
            VR = 0,
            AR
        };

        //! Duration of the scene in seconds. If zero, it will not loop.
        float duration = 0.f;

        //! See enum SceneType
        SceneType type;
    };

    //! Representation of an audio element (ObjectSources, HOASources, and ChannelSources)
    struct AudioElementDesc
    {
        //! Playback mode
        enum Mode
        {
            Continuous, //!< Will play in a loop from the start of the scene
            Event //!< Will play once when play is set to true. After playback is finished, play is set to false automatically
        };

        //! Configure behavior of the renderer for the individual AudioElement
        enum AuthoringParameter
        {
            noDoppler,
            noDistance,
            noDirectionalFocus,
            noCulling
        };

        //! If false, do not render this AudioElement
        bool active = true;

        //! Global gain in dB applied to the playback through this audio element
        float gainDb = 0.f;

        //! Gain (factor) applied to the reverb for this particular audio element
        float itemReverbGain = 1.f;

        //! The distance in m at which the distance attenuation/gain is 0dB
        float refDistance = 1.f;

        //! The distance in m at which the distance attenuation/gain is 0dB
        bool noReverb = false;

        //! A pointer to the AudioStream description used for this audio element
        AudioStreamDesc* signal = nullptr;

        //! A pointer to the GranularSound description used for this audio element
        GranularSoundDesc* granularSignal = nullptr;

        //! Authoring parameters for this AudioElement (see enum AuthoringParameter)
        std::vector<AuthoringParameter> aparams;

        //! Playback mode (see enum Mode)
        Mode mode = Mode::Continuous;

        //! Play flag (has not effect in Mode::Continuous)
        bool play = false;

        //! id of the vdl method per object
        VDLMethod vdlMethod;

        //! Parent transform if the entity is nested in a Transform node (nullptr if not nested)
        TransformDesc* parentTransform = nullptr;

        //! Parent anchor if the entity is nested in an Anchor node (nullptr if not nested)
        AnchorDesc* parentAnchor = nullptr;

        //! Whether or not the object will move in the scene
        bool isStatic = false;
    };

    //! Representation of a geometry element (Primitives and Meshes)
    struct GeometryDesc : Description
    {
        //! Type of the geometry
        enum class GeometryType
        {
            Box = 0,
            Sphere,
            Cylinder,
            Mesh
        };

        //! Position (location, orientation and coordinate space) of the geometry (i.e. its center point)
        EAR::Position position;

        //! Parent transform if the entity is nested in a Transform node (nullptr if not nested)
        TransformDesc* parentTransform = nullptr;

        //! Parent anchor if the entity is nested in an Anchor node (nullptr if not nested)
        AnchorDesc* parentAnchor = nullptr;

        //! Type (see enum GeometryType)
        GeometryType geometryType;

        GeometryDesc(){};

        GeometryDesc(GeometryType type)
            : geometryType(type)
        {
            position.cSpace = EAR::CSpace::RCS;
        };

        //! Whether or not the object will move in the scene
        bool isStatic = false;

        //! Whether or not the object belongs to the early reflection geometry
        bool isEarlyReflectionGeometry = false;
    };

    //! Representation of Primitive
    struct PrimitiveDesc : GeometryDesc
    {

        EAR::Vector3 size; //!< Spatial extent (in x, y, z direction)

        //! Pointer to the AcousticMaterial assigned to this geometry (may be nullptr)
        AcousticMaterialDesc* material = nullptr;

        PrimitiveDesc(){};
    };

    //! Representation of an Acoustic Environment
    struct AcousticEnvironmentDesc : Description
    {

        //! Representation of acoustic parameters in a specific point within an acoustic environment
        struct AcousticParameterDesc
        {

            //! Representation of all acoustic parameters at a specific frequency
            EAR::Position
                position; //!< The position in space at which the acoustic environment is described by the parameters

            //! Default constructor
            AcousticParameterDesc()
            {
                position.cSpace = EAR::CSpace::RCS; // default coordinate space is relative
            };
        };

        //! Pointer to the Geometry that describes the region of validity (nullptr if valid everywhere)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(region->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(region); }
         */
        GeometryDesc* region = nullptr;

        //! List of all AcousticParameterDesc instances in the region of validity
        /*!
         * See struct AcousticParametersDesc for details.
         */
        std::vector<AcousticParameterDesc> parameters;
    };

    //! Representation of an Acoustic Material
    struct AcousticMaterialDesc : Description
    {

        //! Representation of the material coefficients at a specific frequency
        /*!
         *  Note that r + s + t + c must be <= 1
         */
        struct FrequencyDesc
        {
            //! Frequency [Hz]
            float f;

            //! Specular reflection coefficient, range 0-1
            float r = 0.f;

            //! Diffuse scattering coefficient, range 0-1
            float s = 0.f;

            //! Transmission coefficient, range 0-1
            float t = 0.f;

            //! Coupling coefficient, range 0-1
            float c = 0.f;
        };

        //! List containing all coefficients across frequencies
        std::vector<FrequencyDesc> frequencies;
    };

    //! Representation of an Anchor node
    struct AnchorDesc : Description
    {
        //! Identifier of the corresponding LSDF anchor
        std::string lsdfRef;

        //! Whether to automatically rescale audio source positions based on LSDF anchor positions
        bool autoRescale = false;

        //! Control attribute for autoRescale which defines the percentage of scene region that remains outside of LSDF region
        std::vector<float> tolerance;
    };

    //! Representation of an audiostream
    struct AudioStreamDesc : Description
    {
        //!Flag to mark Audio Stream as locally captured
        bool isLocallyCaptured = false;
        std::string file; //!< File path to the audio file (relative to EIF file location)

        //! Input channel indices where the audio for this stream is provided in the MPEG-I 6DoF AEP
        std::vector<int> aepInputChannels;

        //! Bool to indicate if the stream is used by an event-triggered source
        bool isSoundEffect;

        //! Duration of the audio stream in seconds
        float duration{ -1.0f };

        //! Duration in the audio stream in blocks.
        int durationInBlocks{ -1 };
    };
    //! Representation of an audiostream
    struct GranularSoundDesc : Description
    {

        std::string file; //!< File path to the audio file (relative to EIF file location)

        //! number of channels specified
        int numChannels;
    };

    struct AirflowGeneratorDesc : Description
    {
        float speed;
        EAR::Position position;
        float gainDb = 0.f;
        bool active = false;
        bool hasDistanceAttenuationFactor = false;
        float distanceAttenuationFactor = 1.0f;

        AirflowGeneratorDesc() {}
    };

    struct GlobalAirflowGeneratorDesc : public AirflowGeneratorDesc
    {
        GlobalAirflowGeneratorDesc(AirflowGeneratorDesc airflowGenerator)
            : AirflowGeneratorDesc(airflowGenerator)
        {}
    };

    struct PointAirflowGeneratorDesc : public AirflowGeneratorDesc
    {
        bool hasLimits = false;
        float aziRange = 360.0f;
        float elevRange = 360.0f;

        PointAirflowGeneratorDesc(AirflowGeneratorDesc airflowGenerator)
            : AirflowGeneratorDesc(airflowGenerator)
        {}
    };

    struct ConeAirflowGeneratorDesc : public AirflowGeneratorDesc
    {
        float baseRadius = 1.0f;
        float length = 1.0f;
        bool hasTopRadius = false;
        float topRadius = 0.0f;
        float edgeRollOff = 50.0f;

        ConeAirflowGeneratorDesc(AirflowGeneratorDesc airflowGenerator)
            : AirflowGeneratorDesc(airflowGenerator)
        {}
    };

    struct PlaneAirflowGeneratorDesc : public AirflowGeneratorDesc
    {
        float width = 1.0f;
        float height = 1.0f;
        float edgeRollOff = 50.0f;

        PlaneAirflowGeneratorDesc(AirflowGeneratorDesc airflowGenerator)
            : AirflowGeneratorDesc(airflowGenerator)
        {}
    };

    //! Representation of a Box primitive
    struct BoxDesc : PrimitiveDesc
    {
        //! Default Constructor
        BoxDesc(PrimitiveDesc primitive = {})
            : PrimitiveDesc(primitive)
        {
            geometryType = GeometryType::Box;
        }
    };

    //! Representation of a Channel Source
    struct ChannelSourceDesc : Description, AudioElementDesc
    {

        //! Representation of a loudspeaker within a Channel Source
        struct LoudspeakerDesc
        {

            int channelIndex; //!< Index of channel from AudioStream that is played through this Loudspeaker

            EAR::Position
                position; //!< The position, always in RCS relative to the parent ChannelSource

            SourceDirectivityDesc* directivity =
                nullptr; //!< Pointer to the SourceDirectivity (nullptr if no directivity)

            float directiveness =
                1.f; //!< Control the intensity of the directivity over all frequency bands

            LoudspeakerDesc() { position.cSpace = EAR::CSpace::RCS; }
        };

        EAR::Position position; //!< The channel source position (LS positions are relative to this)

        int inputLayout; //!< CICP channel layout number as specified in ISO/IEC 23091-3:2018 CICP-Audio

        std::vector<LoudspeakerDesc>
            loudspeakers; //!< List of virtual loudspeakers (see struct LoudspeakerDesc)

        EAR::Vector3 granularPos; //!< Granular position in three dimensions
    };

    //! Representation of a box primitive
    struct CylinderDesc : PrimitiveDesc
    {
        //! Default Constructor
        CylinderDesc(PrimitiveDesc primitive)
            : PrimitiveDesc(primitive)
        {
            this->geometryType = GeometryType::Box;
        };
    };

    //! Representation of a Source Directivity
    struct SourceDirectivityDesc : Description
    {};

    struct InformedSourceDesc : Description
    {
        EAR::Position position; //!< The position of this source in space

        //! Turns on/off the threshold distance from the listener to the informed source
        bool listenerThresholdPresent = true;

        //! Turns on/off the threshold distance from HOA sources to the informed source
        bool hoaSourceThresholdPresent = true;

        //! Distance from the listener to the informed source where the informed source is active
        float listenerThreshold = 3.0f;

        //! Distance from the listener to HOA sources where the informed source is active
        float hoaSourceThreshold = 3.0f;

        //! Criteria used to disable active informed sources in the rendering
        int disableInformedSourceType = 3;

        //! Distance threshold after the informed source is inactive
        float disableInformedSourceThreshold = 3.0f;

        //! Determines the processing priority between informed sources that are close to each other
        int priorityValue = 0;
    };

    struct HoaGroupDesc : Description
    {
        //! Pointer to the Geometry that describes the region of validity (nullptr if valid everywhere)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(region->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(region); }
         */
        GeometryDesc* region = nullptr;

        //! List of HOA Source Representations that belong to this group
        std::vector<HoaSourceDesc*> hoaSources;

        //! List of ObjectSources sampled simultaneously with the HOA Sources
        std::vector<ObjectSourceDesc*> coSources;

        //! Parameters used to reduce the complexity of MPHOA rendering
        std::vector<int> complexityReductionParameters;

        //! Exterior rendering parameters
        float exteriorProjectionRadius;
        bool adaptiveExteriorProjectionRadius;

        //! Maximum number of simultaneous active informed sources
        int maxSimulInformedSources = 10;

        //! List of informed sources (i.e. known audio source positions)
        std::vector<InformedSourceDesc> informedSources;

        //! Low complexity MPHOA renderer parameters
        float distanceFactor = 1.0f;
        int panningGridOrder = 0;
    };

    struct HoaSourceDesc : Description, AudioElementDesc
    {

        EAR::Position position; //!< The position of this source in space

        //! Pointer to corresponding HOAGroup (may be nullptr)
        HoaGroupDesc* group;

        //! Pointer to the Geometry that describes the spatial extent of this source (nullptr if no extent)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(extent->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(extent); }
         */
        GeometryDesc* extent = nullptr;

        bool extentTransform =
            true; //!< Turns on/off the external rendering of an interior source or vice versa

        int representation; //!< Representation type; 1 => Exterior, otherwise interior

        float transitionDistance =
            0.f; //!< Determines the transition radius between external and internal representation [m]

        //! Granular position in three dimensions
        EAR::Vector3 granularPos;
    };

    //! Representation of a Mesh
    struct MeshDesc : GeometryDesc
    {
        struct VertexDesc
        {
            //! Unique index >= 0
            int index = -1;

            //! Local coordinates of the vertex
            EAR::Vector3 position;
        };

        struct FaceDesc
        {
            //! Unique index >= 0
            int index;

            //! The 3 indexes of the previously declared Vertices
            int v1, v2, v3;

            //! Pointer to the AcousticMaterial assigned to this geometry (may be nullptr)
            AcousticMaterialDesc* material;
        };

        //! List of all Vertices of this Mesh
        std::vector<VertexDesc> vertices;

        //! List of all Faces of this Mesh
        std::vector<FaceDesc> faces;

        MeshDesc()
            : GeometryDesc(GeometryType::Mesh){};
    };

    //! Representation of an ObjectSource
    struct ObjectSourceDesc : Description, AudioElementDesc
    {
        //! The InputLayout gives hints for the positioning of the audio stream channels w.r.t. spatial extent of the source
        struct InputLayoutDesc
        {
            //! The alignment specifies the reference frame of the positioning
            enum class AlignmentDesc
            {
                User = 0, //!< Positionings are relative to the user's viewing direction
                Object //!< Positionings are fixed to the position/orientation of the ObjectSource
            };

            //! Positioning Mnemonics (see EIF specification for further details)
            enum class PositioningDesc
            {
                TL = 0, //!< top left
                T, //!< top
                TR, //!< top right
                L, //!< left
                C, //!< center
                R, //!< right
                BL, //!< bottom left
                B, //!< bottom
                BR //!< bottom right
            };

            AlignmentDesc alignment; //!< Anchoring mode (see InputLayout::Alignment)

            std::vector<PositioningDesc>
                positioning; //!< Positioning of the channels from the audio stream
        };

        EAR::Position position; //!< The position of this source in space

        //! ObjectSource input layout (see ObjectSourceDesc::InputLayout)
        /*!
         * May be nullptr if no input layout is specified.
         */
        std::unique_ptr<InputLayoutDesc> inputLayout;

        //! Pointer to the Geometry that describes the spatial extent of this source (nullptr if no extent)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(extent->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(extent); }
         */
        GeometryDesc* extent = nullptr;

        SourceDirectivityDesc* directivity =
            nullptr; //!< Pointer to the SourceDirectivity (nullptr if no directivity)

        float directiveness =
            1.f; //!< Control the intensity of the directivity over all frequency bands

        EAR::Vector3 granularPos; //!< Granular position in three dimensions
    };

    //! Representation of a box primitive
    struct SphereDesc : PrimitiveDesc
    {
        //! Default Constructor
        SphereDesc(PrimitiveDesc primitive)
            : PrimitiveDesc(primitive)
        {
            this->geometryType = GeometryType::Box;
        };
    };

    //! Representation of a transform
    struct TransformDesc : Description
    {
        //! Position (location, orientation and coordinate space) of the transform
        EAR::Position position;
    };

    //! Destructor
    ~EIFPayload(){};

    //! Return SceneInfo struct
    inline const SceneInfo& getSceneInfo() const { return scene; }

    //! Returns all AcousticEnvironments in the scene
    inline const std::vector<AcousticEnvironmentDesc>& getAcousticEnvironments() const
    {
        return environments;
    };

    //! Returns all AcousticMaterials in the scene
    inline const std::vector<AcousticMaterialDesc>& getAcousticMaterials() const
    {
        return materials;
    };

    //! Returns all AudioStreams in the scene
    inline const std::vector<AudioStreamDesc>& getAudioStreams() const { return streams; };

    //! Returns all GranularSounds in the scene
    inline const std::vector<GranularSoundDesc>& getGranularSound() const
    {
        return granularSounds;
    };

    //! Returns all Anchors in the scene
    inline const std::vector<AnchorDesc>& getAnchors() const { return anchors; };

    //! Returns all Boxes in the scene
    inline const std::vector<BoxDesc>& getBoxes() const { return boxes; };

    //! Returns all ChannelSources in the scene
    inline const std::vector<ChannelSourceDesc>& getChannelSources() const
    {
        return channelSources;
    };

    //! Returns all Cylinders in the scene
    inline const std::vector<CylinderDesc>& getCylinders() const { return cylinders; };

    //! Returns all Directivities in the scene
    inline const std::vector<SourceDirectivityDesc>& getDirectivities() const
    {
        return directivities;
    };

    //! Returns all HoaGroups in the scene
    inline const std::vector<HoaGroupDesc>& getHoaGroups() const { return hoaGroups; };

    //! Returns all HoaSources in the scene
    inline const std::vector<HoaSourceDesc>& getHoaSources() const { return hoaSources; };

    //! Returns all Meshes in the scene
    inline const std::vector<MeshDesc>& getMeshes() const { return meshes; };

    //! Returns all Spheres in the scene
    inline const std::vector<SphereDesc>& getSpheres() const { return spheres; };

    //! Returns all ObjectSources in the scene
    inline const std::vector<ObjectSourceDesc>& getObjectSources() const { return objectSources; };

    //! Returns all Transforms in the scene
    inline const std::vector<TransformDesc>& getTransforms() const { return transforms; };

    inline const std::vector<GlobalAirflowGeneratorDesc>& getGlobalAirflowGenerators() const
    {
        return globalAirflowGenerators;
    };
    inline const std::vector<PointAirflowGeneratorDesc>& getPointAirflowGenerators() const
    {
        return pointAirflowGenerators;
    };
    inline const std::vector<ConeAirflowGeneratorDesc>& getConeAirflowGenerators() const
    {
        return coneAirflowGenerators;
    };
    inline const std::vector<PlaneAirflowGeneratorDesc>& getPlaneAirflowGenerators() const
    {
        return planeAirflowGenerators;
    };

protected:
    friend class EIFPayloadHandler; // Handler is allowed to create new payloads and fill in data
    friend IRCustomEIFPayloadHandler;

    //! Constructor (only visible to EIFPayloadHandler)
    /*!
     * \param payload        raw MHAS payload from bitstream
     * \param basePath      path to the loaded bitstream (required for checking relative files,
     *                      e.g. wav or directivity – will not be needed eventually)
     * \param checkFiles    whether or not to throw an exception when a linked file is not found (default: true)
     */
    EIFPayload(const std::string& _basePath, bool _checkFiles = true)
        : basePath(_basePath)
        , checkFiles(_checkFiles){};

    SceneInfo scene;
    std::string basePath;
    bool checkFiles;

    std::vector<AcousticEnvironmentDesc> environments;
    std::vector<AcousticMaterialDesc> materials;
    std::vector<AudioStreamDesc> streams;
    std::vector<AnchorDesc> anchors;
    std::vector<BoxDesc> boxes;
    std::vector<ChannelSourceDesc> channelSources;
    std::vector<CylinderDesc> cylinders;
    std::vector<SourceDirectivityDesc> directivities;
    std::vector<HoaGroupDesc> hoaGroups;
    std::vector<HoaSourceDesc> hoaSources;
    std::vector<MeshDesc> meshes;
    std::vector<ObjectSourceDesc> objectSources;
    std::vector<SphereDesc> spheres;
    std::vector<TransformDesc> transforms;
    std::vector<GlobalAirflowGeneratorDesc> globalAirflowGenerators;
    std::vector<PointAirflowGeneratorDesc> pointAirflowGenerators;
    std::vector<ConeAirflowGeneratorDesc> coneAirflowGenerators;
    std::vector<PlaneAirflowGeneratorDesc> planeAirflowGenerators;
    std::vector<GranularSoundDesc> granularSounds;
};

} // namespace MPEGI

#endif // MPEGI_EIF_PAYLOAD
