/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_EARLY_PAYLOAD
#define MPEGI_EARLY_PAYLOAD

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/BaseTypes.h>
#include <EAR/Geometry.h>

#include <unordered_map>
#include <vector>

typedef std::vector<std::vector<int>> EarlyDataSet;
typedef std::unordered_map<int, std::unordered_map<int, EarlyDataSet>> VoxelDataEarly;

namespace MPEGI {

//! Provide a accessible C++ representation of an MPEG-I Early type MHAS payload (temporary)
struct EarlyPayload : public Bitstream::Payload
{
public:
    //! Destructor
    ~EarlyPayload(){};

    //! Return culling distances for ER voxel data
    inline int getMaxDistTriangle1() const { return max_dist_triangle_1; }
    inline int getMaxDistTriangle2() const { return max_dist_triangle_2; }
    inline int getMaxDistSource1() const { return max_dist_source_1; }
    inline int getMaxDistSource2() const { return max_dist_source_2; }

    //! Return voxel grid origin
    inline const EAR::Vector3& getVoxelGridOrigin(void) const { return voxelGridOrigin; }

    //! Return voxel grid shape
    inline const EAR::Vector3& getVoxelGridPitch(void) const { return voxelGridPitch; }

    //! Return voxel grid shape
    inline const VoxelCoord& getVoxelGridShape(void) const { return voxelGridShape; }

    //! compute the three-integer voxel coordinate for a Cartesian coordinate
    VoxelCoord pointToVoxelCoordinate(const EAR::Vector3& pos) const
    {
        VoxelCoord vc;
        EAR::Vector3 shifted = pos - voxelGridOrigin;
        vc[0] = int(shifted[0] / voxelGridPitch[0] + 0.5);
        vc[1] = int(shifted[1] / voxelGridPitch[1] + 0.5);
        vc[2] = int(shifted[2] / voxelGridPitch[2] + 0.5);

        if ((vc[0] < 0) || (vc[0] >= voxelGridShape[0]) || (vc[1] < 0)
            || (vc[1] >= voxelGridShape[1]) || (vc[2] < 0) || (vc[2] >= voxelGridShape[2])) {
            vc = VoxelCoord(-1, -1, -1);
        }

        return vc;
    };

    //! compute the unique integer index for a three-integer voxel coordinate
    int voxelCoordinateToVoxelIndex(const VoxelCoord& vc) const
    {
        return vc[0] + voxelGridShape[0] * (vc[1] + voxelGridShape[1] * vc[2]);
    };

    //! compute the unique integer index for a Cartesian coordinate
    int pointToVoxelIndex(const EAR::Vector3& pos) const
    {
        return voxelCoordinateToVoxelIndex(pointToVoxelCoordinate(pos));
    };

    //! Return surface list
    inline const std::vector<EAR::Mesh::Surface>& getSurfaceList() const { return surfaceList; };

    //! Return voxel data for the given listener and source position
    const EarlyDataSet& getLocalDataSet(const EAR::Vector3& posL, const EAR::Vector3& posS) const
    {
        int l = pointToVoxelIndex(posL);
        int s = pointToVoxelIndex(posS);
        return voxelDataEarly.at(l).at(s);
    };

protected:
    friend class EarlyPayloadHandler; // Handler is allowed to create new payloads and fill in data

    //! Constructor (only visible to EarlyPayloadHandler)
    EarlyPayload()
        : max_dist_triangle_1(-1)
        , max_dist_triangle_2(-1)
        , max_dist_source_1(-1)
        , max_dist_source_2(-1){};

    //! encoder parameters: culling distances for ER voxel data
    int max_dist_triangle_1;
    int max_dist_triangle_2;
    int max_dist_source_1;
    int max_dist_source_2;

    //! Voxel grid information
    EAR::Vector3 voxelGridOrigin;
    EAR::Vector3 voxelGridPitch;
    VoxelCoord voxelGridShape;

    //! list of surfaces
    std::vector<EAR::Mesh::Surface> surfaceList;

    //! voxel data of the ER stage
    VoxelDataEarly voxelDataEarly;
};

} // namespace MPEGI

#endif // MPEGI_EARLY_PAYLOAD
