/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_BITSTREAM_GLOBALS
#define MPEGI_BITSTREAM_GLOBALS

#include <EAR/Exception.h>

#include <algorithm>
#include <array>
#include <string>
#include <unordered_map>
#include <vector>

// Magic numbers defined here for the bitstream. Useful to globally change/tweak
// numbers without having to manually change 10s of lines

namespace MPEGI {

struct cwLUT
{
    std::vector<std::vector<float>> values;
    std::vector<unsigned int> code;
    std::vector<unsigned int> codeLen;
    std::vector<std::string> descriptions;
};

struct CodeWordResult
{
    std::vector<float> value;
    std::string description;
};

struct LUTs
{
    static const std::unordered_map<std::string, cwLUT> lut;

    static bool readCodeWord(const std::string& tableIdx, const unsigned int code,
        unsigned int codeLen, CodeWordResult& result)
    {
        const auto& table = lut.at(tableIdx);
        unsigned int maxCodeLen = 0;
        for (int i = 0; i < table.code.size(); ++i) {
            if (code == table.code[i] && codeLen == table.codeLen[i]) {

                result.value = table.values[i];
                result.description = table.descriptions[i];
                return true;
            }
            maxCodeLen = std::max(maxCodeLen, table.codeLen[i]);
        }

        if (codeLen >= maxCodeLen)
            EAR_FAIL("Trying to parse unknown codeword.");

        return false;
    }
};

enum QuantizationScaling
{
    LINEAR = 0
};

struct QuantizationRange
{
    float min;
    float max;
    int bits;
    QuantizationScaling scalingType;
};

static const int ID_SIZE = 16;
static const int NUM_OF_BITS_FOR_MAX_STREAMS = 16;
static const int NUM_OF_BITS_FOR_VOXEL_GRID_SHAPE_X = 32;
static const int NUM_OF_BITS_FOR_VOXEL_GRID_SHAPE_Y = 32;
static const int NUM_OF_BITS_FOR_VOXEL_GRID_SHAPE_Z = 32;
static const int NUM_OF_BITS_FOR_NORMAL_AZI = 12;
static const int NUM_OF_BITS_FOR_NORMAL_ELE = 11;
static const int NUM_OF_BITS_FOR_PLANE_DIST = 22;
static const int NUM_OF_BITS_FOR_NUM_VOXEL_PAIRS = 32;
static const int NUM_OF_BITS_FOR_MAX_FACES = 16;
static const int NUM_OF_BITS_FOR_BLOB_LENGTH = 32;
static const int NUM_OF_BITS_FOR_AUDIO_DIFFUSENESS = 7;
static const int NUM_OF_POLYNOMIAL_COEFFS = 10;
static const std::array<float, NUM_OF_POLYNOMIAL_COEFFS> POLYNOMIAL_COEFF_SCALERS{ 70.0f, 5.5f,
    5.0f, 1.5f, 3.f, 3.f, 0.05f, 0.07f, 0.7f, 1.f };

static const int MAX_COUNT_OR_INDEX_VALUE = 1023;

typedef QuantizationRange QRange;
typedef QuantizationScaling QScale;

static const QRange X_COORD_QUANTIZATION_RANGE = QRange{ -2000.f, 2000.f, 24, QScale::LINEAR };
static const QRange Y_COORD_QUANTIZATION_RANGE = QRange{ -2000.f, 2000.f, 24, QScale::LINEAR };
static const QRange Z_COORD_QUANTIZATION_RANGE = QRange{ -2000.f, 2000.f, 24, QScale::LINEAR };
static const QRange X_LENGTH_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange Y_LENGTH_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange Z_LENGTH_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange X_PRIMITIVE_SIZE_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange Y_PRIMITIVE_SIZE_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange Z_PRIMITIVE_SIZE_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange REF_DIST_QUANTIZATION_RANGE = QRange{ 0.f, 100.f, 10, QScale::LINEAR };
static const QRange YAW_QUANTIZATION_RANGE = QRange{ -180.f, 180.f, 10, QScale::LINEAR };
static const QRange PITCH_QUANTIZATION_RANGE = QRange{ -180.f, 180.f, 10, QScale::LINEAR };
static const QRange ROLL_QUANTIZATION_RANGE = QRange{ -180.f, 180.f, 10, QScale::LINEAR };
static const QRange X_VOXEL_GRID_PITCH_QUANTIZATION_RANGE =
    QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange Y_VOXEL_GRID_PITCH_QUANTIZATION_RANGE =
    QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange Z_VOXEL_GRID_PITCH_QUANTIZATION_RANGE =
    QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange EDGE_LENGTH_QUANTIZATION_RANGE = QRange{ 0.f, 500.f, 19, QScale::LINEAR };
static const QRange EDGE_ANGLE_QUANTIZATION_RANGE = QRange{ -180.f, 180.f, 10, QScale::LINEAR };
static const QRange EDGE_VECTOR_LENGTH_QUANTIZATION_RANGE =
    QRange{ -500.f, 500.f, 29, QScale::LINEAR };
static const QRange EDGE_NORMALIZED_VECTOR_LENGTH_QUANTIZATION_RANGE =
    QRange{ -1.f, 1.f, 19, QScale::LINEAR };
static const QRange DIRECTIVENESS_QUANTIZATION_RANGE = QRange{ 0.f, 20.f, 8, QScale::LINEAR };
static const QRange DECIBEL_QUANTIZATION_RANGE = QRange{ -127.f, 127.f, 12, QScale::LINEAR };
static const QRange SCENE_DURATION_QUANTIZATION_RANGE = QRange{ 0.f, 180.f, 8, QScale::LINEAR };
static const QRange FREQUENCY_QUANTIZATION_RANGE = QRange{ 0.f, 24000.f, 15, QScale::LINEAR };
static const QRange RT60_QUANTIZATION_RANGE = QRange{ 0.f, 30.f, 18, QScale::LINEAR };
static const QRange MODIFICATION_DURATION_QUANTIZATION_RANGE =
    QRange{ 0.f, 180.f, 17, QScale::LINEAR };
static const QRange UPDATE_TIME_QUANTIZATION_RANGE = QRange{ 0.f, 180.f, 16, QScale::LINEAR };
static const QRange UPDATE_DELAY_QUANTIZATION_RANGE = QRange{ 0.f, 10.f, 9, QScale::LINEAR };
static const QRange PREDELAY_QUANTIZATION_RANGE = QRange{ 0.f, 1.f, 10, QScale::LINEAR };
static const QRange PORTAL_FACTOR_QUANTIZATION_RANGE = QRange{ 0.f, 1.f, 19, QScale::LINEAR };
static const QRange AUDIO_DIFFUSENESS_QUANTIZATION_RANGE = QRange{ 0.f, 1.f, 7, QScale::LINEAR };
static const QRange GEQ_SOS_QUANTIZATION_RANGE = QRange{ -4.f, 4.f, 31, QScale::LINEAR };
static const QRange GEQ_F_QUANTIZATION_RANGE = QRange{ -4.f, 4.f, 31, QScale::LINEAR };
static const QRange GEQ_LEVELDB_QUANTIZATION_RANGE = QRange{ -127.f, 127.f, 31, QScale::LINEAR };
static const QRange REVERB_DISTANCE_GAIN_DROP_QUANTIZATION_RANGE =
    QRange{ 0.f, 128.f, 16, QScale::LINEAR };
static const QRange REVERB_MINIMUM_DISTANCE_QUANTIZATION_RANGE =
    QRange{ 0.f, 64.f, 12, QScale::LINEAR };
static const QRange PREDELAY_FACTOR_QUANTIZATION_RANGE =
    QRange{ 0.f, 4294967295.f, 31, QScale::LINEAR };
static const QRange AE_GAIN_QUANTIZATION_RANGE = QRange{ 0.f, 1.f, 16, QScale::LINEAR };
static const QRange FADE_IN_DISTANCE_QUANTIZATION_RANGE = QRange{ 0.f, 128.f, 16, QScale::LINEAR };
static const QRange LSDF_RDR_ADJUSTMENT_QUANTIZATION_RANGE =
    QRange{ -50.f, 50.f, 16, QScale::LINEAR };
static const QRange SPEED_OF_SOUND_QUANTIZATION_RANGE = QRange{ 150.f, 700.f, 31, QScale::LINEAR };
static const QRange INFORMED_SOURCE_DISTANCE_THRESHOLD_QUANTIZATION_RANGE =
    QRange{ 0.f, 100.f, 16, QScale::LINEAR };
static const QRange AIRFLOW_SPEED_RANGE = QRange{ 0.f, 100.f, 10, QScale::LINEAR };
static const QRange AIRFLOW_FILTER_Q_RANGE = QRange{ 0.f, 10.f, 7, QScale::LINEAR };
static const QRange AIRFLOW_ANGLE_RANGE = QRange{ 0.f, 360.f, 9, QScale::LINEAR };
static const QRange AIRFLOW_AZI_ELEV_LIMIT_RANGE = QRange{ 0.f, 180.f, 8, QScale::LINEAR };
static const QRange AIRFLOW_DISTANCE_ATTENUATION_FACTOR_RANGE =
    QRange{ 0.0f, 5.f, 8, QScale::LINEAR };
static const QRange GRANULAR_POS_RANGE = QRange{ 0.f, 1.f, 16, QScale::LINEAR };
static const QRange GRANULAR_PROB_GRP_NO_RANGE = QRange{ 0.f, 255.f, 8, QScale::LINEAR };
static const QRange GRANULAR_GROUP_PROB_RANGE = QRange{ 0.f, 1000.f, 16, QScale::LINEAR };

static const QRange LOWCOMPMPHOA_DISTANCE_FACTOR = QRange{ 0.25f, 4.f, 8, QScale::LINEAR };
static const QRange HEADPHONE_EQ_FILTER_Q_RANGE = QRange{ 0.0f, 10.0f, 7, QScale::LINEAR };
} // namespace MPEGI

#endif // MPEGI_BITSTREAM_GLOBALS
