/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#ifndef MPEGI_GRANULAR_PAYLOAD
#define MPEGI_GRANULAR_PAYLOAD

#include <EAR/FIRFilter.h>
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Helpers/Types.h>
#include <fstream>

namespace MPEGI {

class Grain
{
public:
    unsigned int Dim;
    EAR::Vector3 descriptorCoordinates;
    unsigned short grainLen;
    EAR::Vector3 coordTrend;
    float orgPosition;
    std::vector<std::vector<float>> samples;
    unsigned int grainProbabilityGroup;
    float perGrainProbabilityValue;
    float mixingRule;

    //! Constructor
    Grain(unsigned int dimensions, EAR::Vector3 coord, unsigned short len)
    {
        Dim = dimensions;
        descriptorCoordinates = coord;
        grainLen = len;
        coordTrend = { 0.0, 0.0, 0.0 };
        orgPosition = 0.0f;
        grainProbabilityGroup = 0.0f;
        perGrainProbabilityValue = 1.0f;
        mixingRule = 1.0;
    };

    void SetCoordTrend(EAR::Vector3 cTrend) { coordTrend = cTrend; }

    void SetOrgPos(float oPos) { orgPosition = oPos; }

    void SetProbGroup(int prob) { grainProbabilityGroup = prob; }

    void SetGrainProbVal(float prob) { perGrainProbabilityValue = prob; }

    //! Destructor
    ~Grain(){};
};

class GranularDatabase
{
public:
    std::string granularId;
    unsigned int dimensions;
    unsigned int numberOfGrains;
    unsigned int channels;
    std::vector<Grain> grains;

    // Additional parameters:
    bool descrTrendAvailable;
    EAR::Vector3 coordScaling;
    EAR::Vector3 smoothingTargetCoordinates;
    float searchRadius;
    unsigned int grainHistoryLength;
    // grainMixRule
    unsigned int numMixRule;
    std::vector<float> grainMixRule;
    std::vector<EAR::Vector3> grainMixingRuleCoordinates;
    //fast grain switch
    bool fastGrainSwitchActive;
    float fastGrainSwitchThreshold;
    // probability weights
    float probabilityWeightPosition;
    float probabilityWeightHistory;
    float probabilityWeightTrend;
    float probabilityWeightOrigPos;
    // probability weights
    float probabilityExponentPosition;
    float probabilityExponentHistory;
    float probabilityExponentTrend;
    float probabilityExponentOrigPos;
    int descTrendMethod;
    // probability groups
    bool probabilityGroupsActive;
    int numProbGroups;
    std::vector<float> probabilityGroupVal;
    // random selection
    bool randomSelection;
    // interpolation
    bool useInterpolation;
    float grainInterpolationDistThreshold;
    bool useInterpolationResampling;
    std::vector<bool> pitchDimension;
    int InterpolationMethod;
    // grain overlap amount
    float grainOverlapAmount;
    bool originalPositionAvailable;

    //! Empty Constructor (zero-initializing)
    GranularDatabase(unsigned int dim, unsigned int nrOfGrains, int nrOfChannels)
    {
        dimensions = dim;
        numberOfGrains = nrOfGrains;
        channels = nrOfChannels;
        descrTrendAvailable = false;
        coordScaling = { 1.0, 1.0, 1.0 };
        smoothingTargetCoordinates = { 1.0, 1.0, 1.0 };
        searchRadius = 1.0;
        grainHistoryLength = 5;
        numMixRule = 1;
        grainMixRule = { 0.0 };
        fastGrainSwitchActive = false;
        fastGrainSwitchThreshold = 1.0;
        probabilityWeightPosition = 1.0;
        probabilityWeightHistory = 1.0;
        probabilityWeightTrend = 1.0;
        probabilityWeightOrigPos = 1.0;
        probabilityExponentPosition = 0.0;
        probabilityExponentHistory = 0.0;
        probabilityExponentTrend = 0.0;
        probabilityExponentOrigPos = 0.0;
        descTrendMethod = 0;
        numProbGroups = 0;
        probabilityGroupsActive = false;
        probabilityGroupVal = { 0.0 };
        randomSelection = true;
        useInterpolation = true;
        grainInterpolationDistThreshold = 0.1;
        useInterpolationResampling = false;
        pitchDimension = { false, false, false };
        InterpolationMethod = 0;
        grainOverlapAmount = 0.5;
        originalPositionAvailable = false;
    };

    void SetCoordScaling(EAR::Vector3 scale) { coordScaling = scale; }

    void SetSmoothTargetCoord(EAR::Vector3 smoothTarget)
    {
        smoothingTargetCoordinates = smoothTarget;
    }

    void SetSearchRadius(float radius) { searchRadius = radius; }

    void SetGrainHistoryLength(unsigned int len) { grainHistoryLength = len; }

    void SetGrainMixRule(
        unsigned int R, std::vector<float> mixRule, std::vector<EAR::Vector3> mixRuleCoord)
    {
        numMixRule = R;
        grainMixRule = mixRule;
        grainMixingRuleCoordinates = mixRuleCoord;
    }

    void SetFastGrainSwitchThreshold(float thresh)
    {
        fastGrainSwitchThreshold = thresh;
        fastGrainSwitchActive = true;
    }

    void SetDescTrendMethod(int desctrend) { descTrendMethod = desctrend; }

    void SetProbabilityGroup(int P, std::vector<float> probgroupval)
    {
        numProbGroups = P;
        probabilityGroupVal = probgroupval;
    }

    void SetGrainInterpParams(
        bool interp, float dist, bool useResamp, std::vector<bool> pitchDim, int interpMethod)
    {
        useInterpolation = interp;
        grainInterpolationDistThreshold = dist;
        useInterpolationResampling = useResamp;
        pitchDimension = pitchDim;
        InterpolationMethod = interpMethod;
    }

    void SetGrainOverlapAmount(float overlap) { grainOverlapAmount = overlap; }

    void SetRandomSelection(bool randsel) { randomSelection = randsel; }

    //! Destructor
    ~GranularDatabase(){};
};

struct GranularPayload : public Bitstream::Payload
{
public:
    //! Destructor
    ~GranularPayload() = default;

    //! Returns all granular items in the payload
    inline const std::map<std::string, std::shared_ptr<GranularDatabase>>& getGranular() const
    {
        return granularMap;
    };

private:
    //! Handler is allowed to create new payloads and fill in data
    friend class GranularPayloadHandler;

    //! Constructor (only visible to GranularPayloadHandler)
    GranularPayload() = default;

    std::map<std::string, std::shared_ptr<GranularDatabase>> granularMap;
};

} // namespace MPEGI

#endif // MPEGI_GRANULAR_PAYLOAD
