/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_HUFFMANDECODER_H
#define MPEGI_HUFFMANDECODER_H

#include <MPEGI/Bitstream/BinCursor.h>

#include <cstddef>
#include <string>
#include <unordered_map>
#include <vector>

#include <EAR/Exception.h>

#include <cassert>
#include <cmath>
#include <iostream>

namespace MPEGI {

constexpr std::size_t BITS_FOR_WORD_SIZE{ 6 };

class HuffmanDecoder
{
public:
    template <typename P> void deserializeCodebook(P& payload);

    template <typename P> std::vector<unsigned int> decode(P& payload, size_t len);

    template <typename P> unsigned int decodeSingle(P& payload);

private:
    template <typename P> void deserializeCodebookFixedLength(P& payload);

    template <typename P> void deserializeCodebookHuffman(P& payload);

    template <typename P>
    void traverseTree(P& payload, std::vector<std::string>& codeList, std::string code);

private:
    std::unordered_map<unsigned int, std::string> codebook;
    std::unordered_map<std::string, unsigned int> codebookInverse;
    bool fixedLength = false;
    uint32_t wordSize = 0;
    std::size_t bitsRead = 0;
    unsigned int offset = 0;
};

template <typename P> void HuffmanDecoder::deserializeCodebook(P& payload)
{
    fixedLength = payload.readBool();

    if (fixedLength)
        deserializeCodebookFixedLength(payload);
    else
        deserializeCodebookHuffman(payload);
}

template <typename P> std::vector<unsigned int> HuffmanDecoder::decode(P& payload, size_t len)
{
    std::vector<unsigned int> resultList;
    while (payload.getPosition() < len) {
        unsigned int result = decodeSingle(payload);
        resultList.push_back(result);
    }
    return resultList;
}

template <typename P> unsigned int HuffmanDecoder::decodeSingle(P& payload)
{
    //single int
    if (fixedLength) {
        unsigned int symbol = payload.readUInt(wordSize);
        return symbol + offset;
    } else {
        std::string code = "";
        while (codebookInverse.find(code) == codebookInverse.end()) {
            // We expect that payload.readUInt(1) throws an exception when reading
            // beyond the end of the bitstream. Hence, this loop will also terminate
            // in case of a transmission error.
            unsigned int nextCodePart = payload.readUInt(1);
            code += std::to_string(nextCodePart);
        }
        return codebookInverse[code];
    }
}

template <typename P> void HuffmanDecoder::deserializeCodebookFixedLength(P& payload)
{
    bool use_offset = payload.readBool();
    if (use_offset) {
        unsigned int wordSizeOffset = payload.readUInt(BITS_FOR_WORD_SIZE);
        offset = payload.readUInt(wordSizeOffset);
    } else {
        offset = 0;
    }

    // read "word_size" value from codebook header
    wordSize = payload.readUInt(BITS_FOR_WORD_SIZE);
}

template <typename P> void HuffmanDecoder::deserializeCodebookHuffman(P& payload)
{
    bool use_offset = payload.readBool();
    if (use_offset) {
        unsigned int wordSizeOffset = payload.readUInt(BITS_FOR_WORD_SIZE);
        offset = payload.readUInt(wordSizeOffset);
    } else {
        offset = 0;
    }

    // read "word_size" value from codebook header
    wordSize = payload.readUInt(BITS_FOR_WORD_SIZE);
    std::vector<std::string> codeList;
    std::string code = "";
    traverseTree(payload, codeList, code);
    std::vector<unsigned int> symbolList;
    for (auto& entry : codeList) {
        unsigned int symbol = payload.readUInt(wordSize);
        symbolList.push_back(symbol + offset);
    }
    assert(codeList.size() == symbolList.size());
    for (int it = 0; it < codeList.size(); it++) {
        codebook.insert(std::pair<unsigned int, std::string>(symbolList.at(it), codeList.at(it)));
        codebookInverse.insert(
            std::pair<std::string, unsigned int>(codeList.at(it), symbolList.at(it)));
    }
}

template <typename P>
void HuffmanDecoder::traverseTree(P& payload, std::vector<std::string>& codeList, std::string code)
{
    bool isLeaf = payload.readBool();

    if (isLeaf) {
        codeList.push_back(code);
    } else {
        std::string code0 = code + "0";
        std::string code1 = code + "1";
        traverseTree(payload, codeList, code0);
        traverseTree(payload, codeList, code1);
    }
}

} // namespace MPEGI

#endif //MPEGI_HUFFMANDECODER_H
