/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_LCER_PAYLOAD
#define MPEGI_LCER_PAYLOAD

#include <MPEGI/Bitstream/Bitstream.h>

#include <EAR/BaseTypes.h>

#include <iostream>
#include <map>
#include <optional>
#include <vector>

namespace MPEGI {

class LCERPayload : public Bitstream::Payload
{
public:
    enum LCEREnvironmentType
    {
        NotSet = 0,
        Indoor = 1,
        Outdoor = 2,
        PredefinedPositions = 3
    };

    struct Position
    {
        float azimuth;
        float elevation;
        float radius;
        float ampFac;
    };

    struct BoundingBox
    {
        EAR::Position position;
        EAR::Vector3 size;
    };

    //! Outdoor data structure
    struct LCEROutdoorData
    {
        //! preDelay1
        float preDelay1;

        //! preDelay2
        float preDelay2;

        //! ampFac1
        float ampFac1;

        //! ampFac2
        float ampFac2;

        //! bounding box
        BoundingBox boundingBox;

        friend std::ostream& operator<<(std::ostream& os, const LCEROutdoorData& outdoorData)
        {
            os << "Outdoor Data:" << std::endl;
            os << "preDelay1: " << outdoorData.preDelay1 << std::endl;
            os << "preDelay2: " << outdoorData.preDelay1 << std::endl;
            os << "ampFac1: " << outdoorData.ampFac1 << std::endl;
            os << "ampFac2: " << outdoorData.ampFac2 << std::endl;
            os << "box: ";
            os << outdoorData.boundingBox.position << std::endl;
            os << outdoorData.boundingBox.size << std::endl;
            return os;
        }
    };

    //! Indoor data structure
    struct LCERIndoorData
    {
        //! preDelay
        float preDelay;

        //! ampFac
        float ampFac;

        //! absorption
        float absorption;

        //! patternAzimuth
        float patternAzimuth;

        //! bounding box
        BoundingBox boundingBox;

        friend std::ostream& operator<<(std::ostream& os, const LCERIndoorData& indoorData)
        {
            os << "Indoor Data:" << std::endl;
            os << "preDelay: " << indoorData.preDelay << std::endl;
            os << "ampFac: " << indoorData.ampFac << std::endl;
            os << "absorption: " << indoorData.absorption << std::endl;
            os << "patternAzimuth: " << indoorData.patternAzimuth << std::endl;
            os << "box: ";
            os << indoorData.boundingBox.position << std::endl;
            os << indoorData.boundingBox.size << std::endl;
            return os;
        }
    };

    //! Indoor data structure
    struct LCERPositionData
    {
        //! LC ER positions
        std::vector<Position> positions;

        //! bounding box
        BoundingBox boundingBox;

        friend std::ostream& operator<<(std::ostream& os, const LCERPositionData& positionData)
        {
            os << "Position Data:" << std::endl;
            os << std::endl;
            os << "positions: ";
            for (auto& position : positionData.positions) {
                os << "{" << position.azimuth << "," << position.elevation << "," << position.radius
                   << "," << position.ampFac << "},";
            }
            os << std::endl;
            os << "box: ";
            os << positionData.boundingBox.position << std::endl;
            os << positionData.boundingBox.size << std::endl;
            return os;
        }
    };

    //! Area data structure
    struct LCEREnvironment
    {
        int priority;

        LCEREnvironmentType type;

        LCERPositionData positionData;

        LCEROutdoorData outdoorData;

        LCERIndoorData indoorData;

        bool operator>(const LCEREnvironment& env) const { return (priority > env.priority); }

        friend std::ostream& operator<<(std::ostream& os, const LCEREnvironment& env)
        {
            os << "LC ER environment hasPriority: " << env.priority << " and is of type "
               << static_cast<int>(env.type) << std::endl;
            switch (env.type) {
            case LCEREnvironmentType::Outdoor:
                os << env.outdoorData;
                break;
            case LCEREnvironmentType::Indoor:
                os << env.indoorData;
                break;
            case LCEREnvironmentType::PredefinedPositions:
                os << env.positionData;
                break;
            case LCEREnvironmentType::NotSet:
                os << "LC ER environment undefined";
                break;
            }
            return os;
        }
    };

    //! Destructor
    ~LCERPayload(){};

    friend std::ostream& operator<<(std::ostream& os, const LCERPayload& payload)
    {
        std::string delimiter = "--------------------------------------------------------";
        os << "LCERPayload contains " << payload.envs.size() << " LC ER environments" << std::endl;
        for (auto& envList : payload.envs) {
            os << delimiter << std::endl;
            for (auto& env : envList.second) {
                os << env << std::endl;
            }
        }
        return os;
    }

public:
    //! assign areas to reverbids if available
    std::map<std::string, std::vector<LCEREnvironment>> envs;

private:
    //! Handler is allowed to create new payloads and fill in data
    friend class LCERPayloadHandler;

    //! Constructor (only visible to LCERPayloadHandler)
    LCERPayload(){};
};

} // namespace MPEGI

#endif // MPEGI_LCER_PAYLOAD
