/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#ifndef MPEGI_RASTERMAP_PAYLOAD
#define MPEGI_RASTERMAP_PAYLOAD

#include <algorithm>
#include <iostream>
#include <stdexcept>
#include <vector>

#include <MPEGI/Bitstream/Bitstream.h>

namespace MPEGI {

using RasterMapCoord = std::array<int, 3>;

struct WaypointMatrix
{
public:
    WaypointMatrix(int _width, int _height, int _layer)
        : width(_width)
        , height(_height)
        , layer(_layer)
        , buf(_width * _height * _layer, RasterMapCoord{ -1, -1, 0 })
    {}

    RasterMapCoord& at(int x, int y, int l)
    {
        if (x < 0 || x >= width || y < 0 || y >= height || l < 0 || l >= layer)
            throw std::out_of_range("Invalid WaypointMatrix Coordinate");

        return buf.at(x + width * (y + height * l));
    }

    RasterMapCoord& at(const RasterMapCoord& pos) { return at(pos[0], pos[1], pos[2]); }

    const RasterMapCoord& at(const RasterMapCoord& pos) const
    {
        WaypointMatrix* that = const_cast<WaypointMatrix*>(this);
        return that->at(pos);
    }

    RasterMapCoord& operator[](const RasterMapCoord& pos)
    {
        return buf[pos[0] + width * (pos[1] + height * pos[2])];
    }

    std::vector<RasterMapCoord> getPath(RasterMapCoord pos) const
    {
        std::vector<RasterMapCoord> path;

        RasterMapCoord wp = at(pos);
        while (pos != wp) {
            pos = wp;
            wp = at(wp);
            path.push_back(pos);
        }

        return path;
    }

    float getPathLength(RasterMapCoord pos, const std::vector<RasterMapCoord>& path) const
    {
        float length = 0.f;

        for (const RasterMapCoord& wp : path) {
            float delta_x = float(wp[0] - pos[0]);
            float delta_y = float(wp[1] - pos[1]);
            length += sqrtf(delta_x * delta_x + delta_y * delta_y);
            pos = wp;
        }

        return length;
    }

#if 0
    float getPathAngleDeg(RasterMapCoord pos, const std::vector<RasterMapCoord>& path) const
    {
        float angle = 0.f;

        if (path.size() < 2) {
            return angle;
        }

        std::vector<float> dir_x;
        std::vector<float> dir_y;
        dir_x.reserve(path.size());
        dir_y.reserve(path.size());

        for (const RasterMapCoord& wp : path) {
            float x = float(wp[0] - pos[0]);
            float y = float(wp[1] - pos[1]);
            float len = sqrtf(x * x + y * y);
            dir_x.push_back(x / len);
            dir_y.push_back(y / len);
            pos = wp;
        }

        for (int n = 0; n < dir_x.size() - 1; n++) {
            angle += acosf(dir_x[n] * dir_x[n + 1] + dir_y[n] * dir_y[n + 1]);
        }

        return angle * float(180.0 / M_PI);
    }
#endif

private:
    int width;
    int height;
    int layer;

    //! internal buffer
    std::vector<RasterMapCoord> buf;
};

struct SubsceneData
{
public:
    std::map<int, std::vector<RasterMapCoord>> availableCoords;
    std::vector<WaypointMatrix> waypointMatrix;
    std::map<int, int> source2entry;

    const WaypointMatrix& getWaypointMatrix(int src_id) const
    {
        int entry = source2entry.at(src_id);
        return waypointMatrix[entry];
    }

private:
};

// Provide a accessible C++ representation of an MPEG-I RasterMap type MHAS payload (temporary)
class RasterMapPayload : public Bitstream::Payload
{
public:
    const std::vector<SubsceneData>& getSubsceneData() const { return subsceneData; }

protected:
    friend class RasterMapPayloadHandler; // Handler is allowed to create new payloads and fill in data

    //! Constructor (only visible to RasterMapPayloadHandler)
    RasterMapPayload() = default;

    std::vector<SubsceneData> subsceneData;

private:
};

} // namespace MPEGI

#endif // MPEGI_RASTERMAP_PAYLOAD
