#ifndef MPEGI_BITSTREAM_UPDATE
#define MPEGI_BITSTREAM_UPDATE

#include <EAR/Variant.h>

#include <map>
#include <string>
#include <vector>

namespace MPEGI {

//! Representation of a single Modification (a pair of ID and *one* modified parameter)
struct Modification
{
    //! Which type of entity this modification targets
    enum class TargetEntityType
    {
        ObjectSource = 0,
        HoaSource,
        HoaGroup,
        ChannelSource,
        Transform,
        Geometry,
        Portal,
    };

    enum class TargetAttribute
    {
        Position = 0,
        Orientation,
        CSpace,
        Active,
        GainDb,
        Signal,
        Extent,
        Directivity,
        Directiveness,
        Play,
        Group,
        Region,
        Size,
        Speed,
        Material,
        ReverbGain,
        NoReverb,
        LoudnessDb,
        GranularPos,
        GranularGroupProb
    };

    //! Property to describe the interpolation mode between Updates
    enum class Transition
    {
        Continuous = 0, //!< Continuous change of the value (e.g. motion trajectory)
        Immediate //!< Immediate change the value (e.g. teleport to a new position)
    };

    //! Descriptor of the placeholder position in a dynamic update
    enum class Placeholder
    {
        PositionX = 0,
        PositionY,
        PositionZ,
        OrientationY,
        OrientationP,
        OrientationR,
        GainDb,
        Speed,
        ReverbGain,
        NoReverb,
        LoudnessDb,
        GranularPosX,
        GranularPosY,
        GranularPosZ,
        GranularProbGroupNo,
        GranularGroupProb
    };

    std::string id; //!< ID of a modifiable entity

    Transition transition = Transition::Continuous; //!< see enum Transition

    float duration = 0.f; //!< Duration of interpolation in seconds (if Transition::Continuous)

    //!< The name of the attribute being modified, e.g. "position" or "orientation"
    TargetAttribute attribute;

    //!< The new value of the modification (may be empty for a dynamic update with placeholders)
    EAR::Variant targetValue;

    //!< Placeholders ("parameter indicators") for dynamic updates (map key is the placeholder number)
    std::map<int, Placeholder> placeholders;
};

//! Representation of an Update
struct BitstreamUpdate
{
    //! Type of the update, inferred from the given attributes
    enum class Type
    {
        Timed = 0, //!< Update executed at specific time (required attributes: time)
        Conditional, //!< Update executed when condition is met (required attributes: condition, fireOn, delay)
        Triggered, //!< Update executed by external trigger, possibly with placeholder values for dynamic update (required attributes: id, index)
        Dynamic //!< Update with placeholder values (required attributes: id, index)
    };

    Type type; //!< Update type as described in enum Type

    std::string id; //! Unique identifier of this update

    //!< Globally unique identifier of this Update (required for triggered and dynamic updates, -1 indicates that id is not present)
    int index = -1;

    float time = -1; //!< Update execution time in seconds (negative value if not a timed update)

    std::string condition; //!< Condition (nullptr if not a conditional update)

    bool fireOn = true; //!< Condition flank negation (only for conditional update)

    float delay = 0.f; //!< Postpone update after trigger in seconds

    std::vector<Modification> modifications; //!< All atomic modifications within this update
};

} // namespace MPEGI

#endif // MPEGI_BITSTREAM_UPDATE
