/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include <array>
#include <map>
#include <memory>
#include <optional>

#include "MPEGI/Bitstream/Bitstream.h"
#include "MPEGI/Helpers/VoxelTypes.h"

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
/// Data structures and classes for voxel-based occlusion and diffraction processing
namespace Voxel {
/// Data structures representing the payload for voxel-based occlusion and diffraction (payloadVoxData())
namespace Payload {

/**
 * Unpack 2/3D indices
 *
 * @param packed Packed indices read from the bitstream
 * @param dim Dimensions of the 2/3D space
 * @return Unpacked 2/3D indices
 */
template <std::size_t N>
std::array<voxel_index_val_t, N> unpack(
    voxel_index_val_t packed, std::array<voxel_index_val_t, N> const& dim)
{
    static_assert(N == 2 || N == 3, "Only 2 and 3D supported for now");
    std::array<voxel_index_val_t, N> idx;

    auto i = 0U;
    for (auto& xyz : idx) {
        xyz = (packed % dim[i]) + 1;
        packed /= dim[i];
        i++;
    }
    return idx;
}

/**
 * Pack 2/3D indices
 *
 * @param unpacked indices
 * @param dim Dimensions of the 2/3D space
 * @return packed 2/3D indices
 */
template <std::size_t N>
voxel_index_val_t pack(
    std::array<voxel_index_val_t, N> voxel_cord, std::array<voxel_index_val_t, N> const& dim)
{
    static_assert(N == 2 || N == 3, "Only 2 and 3D supported for now");

    unsigned int packedVoxelCoord = 0;
    unsigned int multiplier = 1;

    for (auto i = 0; i < N; i++) {
        packedVoxelCoord = packedVoxelCoord + ((voxel_cord[i] - 1) * multiplier);
        multiplier = multiplier * dim[i];
    }

    return packedVoxelCoord;
}

//--------------------------------------------------------------------------------------------------
/**
 * Calculate number of bits used for a packed index (2/3D)
 *
 * @tparam N Dimension of the map/grid, i.e. 2/3D
 * @param r Resolution of the grid
 * @return Number of bits used to encode a packed index
 */
template <std::size_t N> std::size_t calc_bit_length(std::array<voxel_index_val_t, N> const& r)
{
    static_assert(N == 2 || N == 3, "Only 2 and 3D supported for now");

    auto value_max =
        std::accumulate(std::cbegin(r), std::cend(r), 1, std::multiplies<voxel_index_val_t>()) - 1;

    auto num_bits = 1U;
    while (value_max >>= 1) {
        ++num_bits;
    }

    return num_bits;
}

//--------------------------------------------------------------------------------------------------
/**
 * Block of voxels with same material (e.g. air, wood, concrete, etc.)
 */
template <int N> struct block
{
    static_assert(N == 2 || N == 3, "Only 2 and 3D blocks supported for now");

    /// Number of bits to read from the stream
    enum
    {
        n_bits_esc1 = 4,
        n_bits_esc2 = 8,
        n_bits_esc3 = 32
    };

    int value{ 0 }; //!< Value, e.g. material ID
    std::array<voxel_index_val_t, N> p0{ 0 }; //!< Block corner 1 (voxel IDs)
    std::array<voxel_index_val_t, N> p1{ 0 }; //!< Block corner 2 (voxel IDs)

    block() = default;
    block(Bitstream::RawPayload const& payload, std::array<voxel_index_val_t, N> const& dim,
        std::size_t n_bits_packed_coord)
    {
        value = read_value(payload, dim);

        p0 = unpack(payload.readUInt(n_bits_packed_coord), dim);
        p1 = unpack(payload.readUInt(n_bits_packed_coord), dim);
    }

    block(int _value, std::array<voxel_index_val_t, N> _p0, std::array<voxel_index_val_t, N> _p1)
    {
        value = _value;
        for (auto i = 0; i < N; i++) {
            // Bitstream block indices are 1-based
            p0[i] = _p0[i] + 1;
            p1[i] = _p1[i] + 1;
        }
    }

    /**
     * Overload for 2D data
     *
     * @param payload Bitstream payload
     * @param dim Map dimension (2D)
     * @return Value
     */
    int read_value(
        Bitstream::RawPayload const& payload, std::array<voxel_index_val_t, 2> const& dim) const
    {
        return payload.readBool();
    }

    /**
     * Overload for 3D data (ESC coding)
     *
     * @param payload Bitstream payload
     * @param dim Grid dimension (3D)
     * @return Value
     */
    int read_value(
        Bitstream::RawPayload const& payload, std::array<voxel_index_val_t, 3> const& dim) const
    {
        return payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
    }

    int get_value() const { return value; }
    int& get_value() { return value; }
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing material
 */
struct scene_material
{
    /// Number of bits to read from the stream
    enum
    {
        n_bits_esc1 = 8,
        n_bits_esc2 = 16,
        n_bits_esc3 = 32
    };

    /**
     * Constructor
     *
     * @param payload The bitstream payload
     * @param id Material ID is implied by order not read from bitstream
     */
    scene_material(Bitstream::RawPayload const& payload, const BitstreamConfig& config, int id)
    {
        bs_id = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);

        EAR_CHECK(
            config.getIdMap().count(bs_id) > 0, "VoxDataPayload: Invalid material bitstream id");
        this->id = id;
    }

    int id{ -1 }; //!< Material ID, starting at 1
    int bs_id{ 0 }; //!< Bitstream ID (EIF) for material mappings
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a pre computed path data on the diffraction map
 */
struct pre_computed_path
{
    enum
    {
        n_bits_delta = 2
    };

    /**
     * Constructor
     *
     * @param payload The bitstream payload
     * @param dim Dimension of the diffraction map (x,y)
     * @param n_bits_packed_coord Number of bits used to encode packed coordinates
     */
    pre_computed_path(Bitstream::RawPayload const& payload,
        std::array<voxel_index_val_t, 2> const& start, std::array<voxel_index_val_t, 2> const& end,
        std::pair<std::array<voxel_index_val_t, 2>, float>& prev,
        std::array<voxel_index_val_t, 2> const& dim, std::size_t n_bits_packed_coord)
        : start_xy(start)
        , end_xy(end)
    {
        auto const exists = payload.readBool();
        if (exists) {
            auto const use_prev = payload.readBool();
            if (!use_prev) {
                src_xy = unpack(payload.readUInt(n_bits_packed_coord), dim);
                length = payload.readFloat();
                prev.first = src_xy.value();
            } else {
                static std::array<float, 4> const delta{ 1.f, M_SQRT2 - 1, 1 - M_SQRT2, -1.f };

                src_xy = prev.first;
                length = prev.second + delta[payload.readUInt(n_bits_delta)];
            }

            prev.second = length.value();
        }
    }

    pre_computed_path(Bitstream::RawPayload const& payload,
        std::array<voxel_index_val_t, 2> const& start,
        std::array<voxel_index_val_t, 2>& prev_corner, std::array<voxel_index_val_t, 2> const& dim,
        std::size_t n_bits_packed_coord)
        : start_xy(start)
    {
        end_xy = unpack(payload.readUInt(n_bits_packed_coord), dim);
        if (payload.readBool()) {
            if (payload.readBool()) {
                src_xy = prev_corner;
            } else {
                src_xy = unpack(payload.readUInt(n_bits_packed_coord), dim);
                prev_corner = src_xy.value();
            }

            length = payload.readFloat();
        }
    }

    dmap_index_t start_xy{ 0 };
    dmap_index_t end_xy{ 0 };
    std::optional<dmap_index_t> src_xy;
    std::optional<float> length;
};

//--------------------------------------------------------------------------------------------------
struct instruction_occlusion
{
    inline static float recovery_default = 0.f;

    instruction_occlusion(int material_id_, float recovery_ = recovery_default)
        : material_id(material_id_)
        , recovery(recovery_)
    {}

    int material_id; //!< Material ID to that this instruction applies
    float recovery; //!< Occlusion recovery setting
};

//--------------------------------------------------------------------------------------------------
struct instruction_diffraction
{
    instruction_diffraction(int material_id_, bool diffraction_enable_ = true)
        : material_id(material_id_)
        , diffraction_enable(diffraction_enable_)
    {}

    int material_id; //!< Material ID to that this instruction applies
    bool diffraction_enable; //!< Enable/disable diffraction
};

//--------------------------------------------------------------------------------------------------
struct source_material_dependency_occlusion
{
    /// Number of bits to read from the stream for escaped values
    enum
    {
        n_bits_esc1 = 8,
        n_bits_esc2 = 16,
        n_bits_esc3 = 32
    };

    source_material_dependency_occlusion() = default;
    source_material_dependency_occlusion(Bitstream::RawPayload const& payload)
    {
        auto const source_id = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
        auto const material_id = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);

        source_material_pair = std::make_pair(source_id, material_id);
    };

    /// Non-relevant materials per source
    std::pair<int, int> source_material_pair;
};

//--------------------------------------------------------------------------------------------------
/**
 * Struct representing data for voxel-based occlusion processing
 */
struct occlusion_data
{
    using shared = std::shared_ptr<occlusion_data>;

    /// Number of bits to read from the stream for escaped values
    enum
    {
        n_bits_esc1 = 8,
        n_bits_esc2 = 16,
        n_bits_esc3 = 32
    };

    /**
     * Fill data structure with occlusion-related data from the bitstream
     *
     * @param payload Bitstream handle
     * @param materials List of materials
     */
    occlusion_data(
        Bitstream::RawPayload const& payload, std::vector<scene_material> const& materials);

protected:
    void read_instructions(
        Bitstream::RawPayload const& payload, std::vector<scene_material> const& materials);

    void read_source_material_dependency(Bitstream::RawPayload const& payload);

public:
    /// Renderer instructions, i.e., voxSceneOcclusionRecoveryValue per material
    std::vector<instruction_occlusion> instructions_occlusion;
    /// Source-material dependency (voxSceneOcclusionSourceMaterialDependency())
    std::vector<source_material_dependency_occlusion> source_material_dependency;
};

//--------------------------------------------------------------------------------------------------
/**
 * Struct representing diffraction data
 */
struct diffraction_data
{
    using shared = std::shared_ptr<diffraction_data>;

    /// Number of bits to read from the stream for escaped values
    enum
    {
        n_bits_esc1 = 8,
        n_bits_esc2 = 16,
        n_bits_esc3 = 32,
        n_bits_diffraction_map_heigth = 16
    };

    /**
     * Constructor reads data from the bitstream
     *
     * @param payload Bitstream handle
     * @param materials List of materials
     * @param res Resolution of the diffraction map (x,y)
     */
    diffraction_data(Bitstream::RawPayload const& payload,
        std::vector<scene_material> const& materials, dmap_index_t const& res);

protected:
    void read_instructions(
        Bitstream::RawPayload const& payload, std::vector<scene_material> const& materials);
    void read_effect_control(Bitstream::RawPayload const& payload);
    void read_diffraction_map(Bitstream::RawPayload const& payload, dmap_index_t const& res);
    void read_diffraction_map_height(Bitstream::RawPayload const& payload);
    void read_pre_computed_path_data(Bitstream::RawPayload const& payload, dmap_index_t const& res);

public:
    using dmap_element_t = std::vector<block<2>>;

    dmap_index_t resolution; //!< Resolution of the diffraction map
    dmap_element_t blocks_dmap; //!< Block elements (2D) composing diffraction map
    std::vector<pre_computed_path> pre_computed_path_data; //!< Pre computed path_data
    std::vector<instruction_diffraction>
        instructions_diffraction; //!< Renderer instructions (Diffraction)

    float diffraction_effect_control{ 1.f }; //!< Control diffraction "strength"

    std::optional<unsigned int>
        diffraction_map_height; //!< Scene height (in voxels) to be used for automatic generation of diffraction map
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a sequence point within a scene update
 */
struct sequence_point
{
    sequence_point(int duration_, std::vector<block<3>>&& blocks_,
        occlusion_data::shared const& data_occlusion_,
        diffraction_data::shared const& data_diffraction_)
        : duration(duration_)
        , blocks(std::move(blocks_))
        , data_occlusion(data_occlusion_)
        , data_diffraction(data_diffraction_)
    {}

    int duration{ 0 }; //!< Duration of the sequence point
    std::vector<block<3>> blocks; //!< Blocks to be updated
    diffraction_data::shared data_diffraction; //!< Diffraction data update, e.g., diffraction map
    occlusion_data::shared data_occlusion; //!< Occlusion data update, e.g., occlusion recovery
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing an update to a voxel (sub-)scene
 */
struct scene_update
{
    enum
    {
        n_bits_update_id = 8,
        n_bits_has_reference_point = 1,
        n_bits_esc1 = 8,
        n_bits_esc2 = 16,
        n_bits_esc3 = 32,
        n_bits_update_duration = 16,
        n_bits_num_block_elements = 16
    };

    scene_update(Bitstream::RawPayload const& payload, const BitstreamConfig& config,
        std::vector<scene_material> const& materials, Voxel::voxel_index_t const& dim);

    int id; //!< Update ID
    std::vector<sequence_point> sequence; //!< Scene update sequence

    using reference_point_opt_t = std::optional<std::array<float, 3>>;
    reference_point_opt_t p0; //!< Optional new reference point for the scene
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a voxel (sub-)scene
 */
struct subscene
{
    /**
     * Read voxSceneElements() defining the voxel scene geometry
     *
     * @param payload Bitstream handle
     * @param res Voxel scene resolution (x,y,z)
     * @return Non-empty vector on success
     */
    static std::vector<block<3>> readVoxSceneElements(
        Bitstream::RawPayload const& payload, Voxel::voxel_index_t const& res);

    enum
    {
        n_bits_grid_resolution = 16,
        n_bits_esc1 = 8,
        n_bits_esc2 = 16,
        n_bits_esc3 = 32
    };

    /// Construct a subscene
    subscene(Bitstream::RawPayload const& payload, const BitstreamConfig& config);

    struct
    {
        std::array<float, 3> p0{ 0 }; //!< scene bounding box p0 (xyz)
        std::array<float, 3> p1{ 0 }; //!< scene bounding box p1 (xyz)
        Voxel::voxel_index_t dim{ 0 }; //!< voxel grid resolution (xyz)

    } scene;

    std::vector<scene_update> scene_updates; //!< Scene updates
    std::vector<block<3>> blocks; //!< Block elements composing scene geometry
    std::vector<scene_material> materials; //!< Material descriptions
    diffraction_data::shared data_diffraction; //!< Data for voxel-based diffraction
    occlusion_data::shared data_occlusion; //!< Data for voxel-based occlusion

protected:
    /// Read voxSceneHeader() providing basic voxel scene information
    void readVoxSceneHeader(Bitstream::RawPayload const& payload, const BitstreamConfig& config);
    /// Read voxSceneElements() defining the voxel scene geometry
    void readVoxSceneElements(Bitstream::RawPayload const& payload);
    /// Read voxSceneMaterials() defining the materials used in the voxel scene
    void readVoxSceneMaterials(Bitstream::RawPayload const& payload, const BitstreamConfig& config);
    /// Read voxSceneInstructionsForOcclusion() providing supplemental information for occlusion processing
    void readVoxSceneInstructionsForOcclusion(Bitstream::RawPayload const& payload);
    /// Read voxSceneInstructionsForDiffraction() providing supplemental information for diffraction processing
    void readVoxSceneInstructionsForDiffraction(Bitstream::RawPayload const& payload);
    /// Read voxSceneUpdates() defining updates to the voxel scene
    void readVoxSceneUpdates(Bitstream::RawPayload const& payload, const BitstreamConfig& config);
};

//--------------------------------------------------------------------------------------------------
} // namespace Payload
} // namespace Voxel
//--------------------------------------------------------------------------------------------------

class VoxDataPayload : public Bitstream::Payload
{
    friend class VoxDataPayloadHandler;

    VoxDataPayload(Bitstream::RawPayload const& payload, const BitstreamConfig& config);

    VoxDataPayload(VoxDataPayload const&) = delete;
    VoxDataPayload& operator=(VoxDataPayload const&) = delete;

public:
    enum
    {
        n_bits_num_subscenes_esc1 = 8,
        n_bits_num_subscenes_esc2 = 16,
        n_bits_num_subscenes_esc3 = 32,
        n_bits_vox_er_num_origins = 5,
        n_bits_vox_er_num_clusters = 5
    };

    std::vector<Voxel::Payload::subscene> subscenes;

public:
    using shared = std::shared_ptr<VoxDataPayload>;
    using const_shared = std::shared_ptr<VoxDataPayload const>;

protected:
    void set_global_parameters(Bitstream::RawPayload const& payload) const;
};

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------