/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/AirflowPayload.h>
#include <MPEGI/Bitstream/AirflowPayloadHandler.h>
#include <MPEGI/Bitstream/Globals.h>

#include <EAR/Exception.h>
#include <MPEGI/Helpers/Config.h>
#include <iostream>
#include <math.h>

namespace MPEGI {

Bitstream::PayloadHandlerAndType makeAirflowPayloadHandler()
{
    return { std::make_shared<AirflowPayloadHandler>(), Bitstream::PLD_AIRFLOW };
}

std::shared_ptr<const Bitstream::Payload> AirflowPayloadHandler::unpack(
    const Bitstream::RawPayload& rawPayload, const BitstreamConfig& config)
{
    auto payload = std::shared_ptr<AirflowPayload>(new AirflowPayload());
    payload->freqGridData = config.GetFreqGrid();

    // Control params
    const bool isHiPrecGain = true;

    const int numOfAirflowSources = rawPayload.getCountOrIndex();

    readAirflowParameters(numOfAirflowSources, rawPayload, payload.get(), isHiPrecGain, config);

    return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
}

void AirflowPayloadHandler::readAirflowParameters(const int& numOfAirflowSources,
    const MPEGI::Bitstream::RawPayload& rawPayload, MPEGI::AirflowPayload* payload,
    bool isHiPrecGain, const MPEGI::BitstreamConfig& config)
{
    for (int e = 0; e < numOfAirflowSources; ++e) {
        const unsigned int srcID = rawPayload.getID();
        const bool hasCustomProfile = rawPayload.readBool();
        unsigned int defaultProfile = 0;
        std::vector<AirflowProfileParameters> customAirflowProfiles;
        if (hasCustomProfile) {
            customAirflowProfiles = ReadCustomAirflowProfiles(rawPayload, payload, isHiPrecGain);
        } else {
            defaultProfile = rawPayload.readUInt(2);
        }
        const unsigned int generatorID = rawPayload.getID();
        const std::string generatorStringID = config.getIdMap().at(generatorID);
        const bool activeInDefaultRegion = rawPayload.readBool();
        const bool hasValidRegions = rawPayload.readBool();

        std::vector<int> activeRegionIDs;

        if (hasValidRegions) {
            const size_t numRegions = rawPayload.getCountOrIndex();
            activeRegionIDs.resize(numRegions);
            for (int b = 0; b < numRegions; ++b) {
                activeRegionIDs[b] = rawPayload.getID();
            }
        }

        payload->airflowSources[srcID] =
            std::make_shared<AirflowSource>(srcID, generatorID, generatorStringID, defaultProfile,
                hasValidRegions, customAirflowProfiles, activeInDefaultRegion, activeRegionIDs);
    }
}

std::vector<MPEGI::AirflowProfileParameters> AirflowPayloadHandler::ReadCustomAirflowProfiles(
    const MPEGI::Bitstream::RawPayload& rawPayload, const MPEGI::AirflowPayload* payload,
    bool isHiPrecGain)
{
    std::vector<MPEGI::AirflowProfileParameters> customAirflowProfiles;

    const unsigned int profileType = rawPayload.readUInt(2);
    const auto profileTypeEnum = static_cast<AirflowProfileParameters::ProfileType>(profileType);

    switch (profileTypeEnum) {
    case AirflowProfileParameters::ProfileType::Freq_Gain_Pairs: {
        const unsigned int numSpeedProfiles = rawPayload.getCountOrIndex();
        const unsigned int numAzimuthProfiles = rawPayload.getCountOrIndex();
        const unsigned int numElevProfiles = rawPayload.getCountOrIndex();
        std::vector<std::pair<float, float>> gain;
        std::vector<std::pair<float, float>> q;

        for (int b = 0; b < numSpeedProfiles; ++b) {
            const auto profileModifies = AirflowProfileParameters::ProfileModification::speed;
            const unsigned int freqGridIdx = rawPayload.getCountOrIndex();
            const float refSpeed = rawPayload.readQuantFloat(AIRFLOW_SPEED_RANGE);
            const size_t fgdNrBands = payload->freqGridData[freqGridIdx]->nrBands;
            gain.resize(fgdNrBands);
            for (int g = 0; g < fgdNrBands; ++g) {
                gain[g] = std::make_pair(rawPayload.getGain(isHiPrecGain), 0.0f);
            }
            customAirflowProfiles.emplace_back(
                profileTypeEnum, freqGridIdx, profileModifies, refSpeed, gain, q);
        }
        for (int b = 0; b < numAzimuthProfiles; ++b) {
            const auto profileModifies = AirflowProfileParameters::ProfileModification::azimuth;
            const unsigned int freqGridIdx = rawPayload.getCountOrIndex();
            const float refAzimuth = rawPayload.readQuantFloat(AIRFLOW_ANGLE_RANGE);
            const size_t fgdNrBands = payload->freqGridData[freqGridIdx]->nrBands;
            gain.resize(fgdNrBands);
            for (int g = 0; g < fgdNrBands; ++g) {
                auto gainL = rawPayload.getGain(isHiPrecGain);
                auto gainR = rawPayload.getGain(isHiPrecGain);
                gain[g] = std::make_pair(gainL, gainR);
            }
            customAirflowProfiles.emplace_back(
                profileTypeEnum, freqGridIdx, profileModifies, refAzimuth, gain, q);
        }
        for (int b = 0; b < numElevProfiles; ++b) {
            const auto profileModifies = AirflowProfileParameters::ProfileModification::elevation;
            const unsigned int freqGridIdx = rawPayload.getCountOrIndex();
            const float refElev = rawPayload.readQuantFloat(AIRFLOW_ANGLE_RANGE);
            const size_t fgdNrBands = payload->freqGridData[freqGridIdx]->nrBands;
            gain.resize(fgdNrBands);
            for (int g = 0; g < fgdNrBands; ++g) {
                gain[g] = std::make_pair(rawPayload.getGain(isHiPrecGain), 0.f);
            }
            customAirflowProfiles.emplace_back(
                profileTypeEnum, freqGridIdx, profileModifies, refElev, gain, q);
        }
        break;
    }
    case AirflowProfileParameters::ProfileType::Filter_Params: {
        const unsigned int numSpeedProfiles = rawPayload.getCountOrIndex();
        const unsigned int numAzimuthProfiles = rawPayload.getCountOrIndex();
        const unsigned int numElevProfiles = rawPayload.getCountOrIndex();
        std::vector<std::pair<float, float>> gain;
        std::vector<std::pair<float, float>> q;

        for (int b = 0; b < numSpeedProfiles; ++b) {
            const auto profileModifies = AirflowProfileParameters::ProfileModification::speed;
            const unsigned int freqGridIdx = rawPayload.getCountOrIndex();
            const float refSpeed = rawPayload.readQuantFloat(AIRFLOW_SPEED_RANGE);
            const size_t fgdNrBands = payload->freqGridData[freqGridIdx]->nrBands;
            gain.resize(fgdNrBands);
            q.resize(fgdNrBands);
            for (int g = 0; g < fgdNrBands; ++g) {
                gain[g] = std::make_pair(rawPayload.getGain(isHiPrecGain), 0.f);
                q[g] = std::make_pair(rawPayload.readQuantFloat(AIRFLOW_FILTER_Q_RANGE), 0.f);
            }
            customAirflowProfiles.emplace_back(
                profileTypeEnum, freqGridIdx, profileModifies, refSpeed, gain, q);
        }
        for (int b = 0; b < numAzimuthProfiles; ++b) {
            const auto profileModifies = AirflowProfileParameters::ProfileModification::azimuth;
            const unsigned int freqGridIdx = rawPayload.getCountOrIndex();
            const float refAzimuth = rawPayload.readQuantFloat(AIRFLOW_ANGLE_RANGE);
            const size_t fgdNrBands = payload->freqGridData[freqGridIdx]->nrBands;
            gain.resize(fgdNrBands);
            q.resize(fgdNrBands);
            for (int g = 0; g < fgdNrBands; ++g) {
                auto gainL = rawPayload.getGain(isHiPrecGain);
                auto gainR = rawPayload.getGain(isHiPrecGain);
                gain[g] = std::make_pair(gainL, gainR);
                auto qL = rawPayload.readQuantFloat(AIRFLOW_FILTER_Q_RANGE);
                auto qR = rawPayload.readQuantFloat(AIRFLOW_FILTER_Q_RANGE);
                q[g] = std::make_pair(qL, qR);
            }
            customAirflowProfiles.emplace_back(
                profileTypeEnum, freqGridIdx, profileModifies, refAzimuth, gain, q);
        }
        for (int b = 0; b < numElevProfiles; ++b) {
            const auto profileModifies = AirflowProfileParameters::ProfileModification::elevation;
            const unsigned int freqGridIdx = rawPayload.getCountOrIndex();
            const float refElev = rawPayload.readQuantFloat(AIRFLOW_ANGLE_RANGE);
            const size_t fgdNrBands = payload->freqGridData[freqGridIdx]->nrBands;
            gain.resize(fgdNrBands);
            q.resize(fgdNrBands);
            for (int g = 0; g < fgdNrBands; ++g) {
                gain[g] = std::make_pair(rawPayload.getGain(isHiPrecGain), 0.f);
                q[g] = std::make_pair(rawPayload.readQuantFloat(AIRFLOW_FILTER_Q_RANGE), 0.f);
            }
            customAirflowProfiles.emplace_back(
                profileTypeEnum, freqGridIdx, profileModifies, refElev, gain, q);
        }
        break;
    }
    default:
        EAR_FAIL("Unsupported airflow profile type");
        break;
    }
    return customAirflowProfiles;
}

} // namespace MPEGI
