/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/BinCursor.h>

#include <EAR/Exception.h>

#include <array>
#include <cmath>

namespace MPEGI {

BinCursor::BinCursor(const std::uint8_t* _data, std::size_t length)
    : data(_data)
    , dataStart(0)
    , dataEnd(length * 8)
    , position(0)
{}

BinCursor::BinCursor(const BinCursor& cursor)
    : data(cursor.data)
    , dataStart(cursor.position)
    , dataEnd(cursor.dataEnd)
    , position(0)
{}

bool BinCursor::readBool()
{
    return readBits<bool>(1);
}

unsigned long long int BinCursor::readUInt(std::size_t lengthToRead)
{
    return readBits<unsigned long long int>(lengthToRead);
}

long long int BinCursor::readSInt(std::size_t lengthToRead)
{
    unsigned long long int val = readUInt(lengthToRead);
    // Converting two's complement unsigned representation into signed integer representation
    if (val >= 1 << (lengthToRead - 1)) {
        val -= 1 << lengthToRead;
    }

    return (long long int)val;
}

unsigned int BinCursor::readEscapedUInt(std::size_t l1, std::size_t l2, std::size_t l3)
{
    unsigned int val = readUInt(l1);

    if (val == (1 << l1) - 1) {
        int valueAdd = readUInt(l2);
        val += valueAdd;

        if (valueAdd == (1 << l2) - 1) {
            valueAdd = readUInt(l3);
            val += valueAdd;
        }
    }

    return val;
}

float BinCursor::readFloat()
{
    unsigned int val = readUInt(32);
    return *(float*)(&val);
}

float BinCursor::readQuantFloat(QuantizationRange qRange)
{
    unsigned int encodedValue = readUInt(qRange.bits);
    return (qRange.max - qRange.min) * static_cast<float>(encodedValue)
        / float(std::pow(2, qRange.bits) - 1)
        + qRange.min;
}

std::unique_ptr<std::uint8_t[]> BinCursor::readBlob(std::size_t& lengthOfBlob)
{
    lengthOfBlob = readUInt(32);

    auto blob = std::make_unique<std::uint8_t[]>(lengthOfBlob);

    for (int i = 0; i < lengthOfBlob; i++) {
        std::uint8_t byte = static_cast<std::uint8_t>(readUInt(8));
        blob[i] = byte;
    }

    return blob;
}

std::shared_ptr<std::vector<std::uint8_t>> BinCursor::readBlob()
{
    std::size_t lengthOfBlob = readUInt(32);

    auto blob = std::make_shared<std::vector<std::uint8_t>>(lengthOfBlob);

    auto* data = blob->data();
    for (int i = 0; i < lengthOfBlob; i++) {
        std::uint8_t byte = static_cast<std::uint8_t>(readUInt(8));
        data[i] = byte;
    }

    return blob;
}

std::string BinCursor::readString()
{
    std::string string;

    char character = 0x0;
    while (true) {
        character = static_cast<char>(readUInt(8));

        if (character == 0x0) { // Null terminator in ASCII string
            return string;
        }

        string += std::string(1, character);
    }
}

void BinCursor::movePosRight(std::size_t lengthToTraverse)
{
    EAR_CHECK_REL(
        position + lengthToTraverse <= dataEnd, "Bitstream: Cannot move cursor beyond data end");

    position += lengthToTraverse;
}

void BinCursor::movePosLeft(std::size_t lengthToReceed)
{
    EAR_CHECK_REL((int)position - (int)lengthToReceed >= 0,
        "Bitstream: Cannot move cursor behind data start");

    position -= lengthToReceed;
}

template <typename T> T BinCursor::readBits(std::size_t nBits)
{
    EAR_CHECK_REL(
        dataStart + position + nBits <= dataEnd, "Bitstream: Cannot read beyond length of data");

    EAR_CHECK_REL(nBits <= (sizeof(T) * 8),
        "Bitstream: Cannot read more than " + std::to_string((sizeof(T) * 8))
            + " bits into given type.");

    T value = 0;

    for (std::size_t i = 0; i < nBits; i++) {
        int byteNum = std::floor((dataStart + position + i) / 8.0);
        std::uint8_t byte = data[byteNum];
        std::uint8_t readBit = byte >> (7 - ((dataStart + position + i) % 8)) & 1;
        value = (value << 1) + readBit;
    }

    position += nBits;

    return value;
}

CodeWordResult BinCursor::readCodeWord(const std::string& tableName)
{
    std::uint16_t word = 0;

    uint64_t length = 0;
    CodeWordResult result;

    do {
        uint8_t bit = readBits<uint8_t>(1);
        word = (word << 1) | bit;
        length++;

    } while (!LUTs::readCodeWord(tableName, word, length, result));

    return result;
}

unsigned int BinCursor::getID()
{
    unsigned int id = readUInt(7);

    bool largerValue = readBool();
    while (largerValue) {
        unsigned int idVal = readUInt(7);

        id = id * 128 + idVal;

        largerValue = readBool();
    }

    return id;
}

unsigned int BinCursor::getCountOrIndex()
{
    unsigned int number = readCodeWord("countOrIndexLoCode").value[0];

    bool isLargeNumber = readBool();
    if (isLargeNumber) {
        unsigned int numberHi = readCodeWord("countOrIndexHiCode").value[0];
        number = number + numberHi * 64;
    }

    return number;
}

float BinCursor::getDuration(bool isLongRangeMode)
{
    float deciSeconds = readCodeWord("deciSecondsCode").value[0];
    float duration = deciSeconds;

    bool addMs = readBool();
    if (addMs) {
        float miliSeconds = readCodeWord("milliSecondsCode").value[0];
        duration = duration + miliSeconds;
        bool addMicroseconds = readBool();
        if (addMicroseconds) {
            float microsecs = readCodeWord("microsecondsCode").value[0];
            duration = duration + microsecs;
        }
    }

    bool addSecs = readBool();
    if (addSecs) {
        if (isLongRangeMode) {
            unsigned int rawSeconds = readUInt(12);
            duration = duration + rawSeconds;
            bool addHours = readBool();
            if (addHours) {
                unsigned int rawHours = readUInt(5);
                duration = duration + (rawHours + 1) * 3600;
            }
        } else {
            float seconds = readCodeWord("secondsCode").value[0];
            duration = duration + seconds;
        }
    }

    return duration;
}

float BinCursor::getFrequency()
{
    float frequency = readCodeWord("frequencyCode").value[0];

    bool moreAccuracy = readBool();
    if (moreAccuracy) {
        float frequencyRefine = readUInt(4);
        frequency = frequency * std::pow(2.0f, ((frequencyRefine + 1) / 51.0f));
    }
    return frequency;
}

std::array<float, 3> BinCursor::getPosition(bool isSmallScene)
{
    std::array<float, 3> position{ 0, 0, 0 };

    for (int i = 0; i < 3; ++i) {
        float coord = getDistance(isSmallScene);
        bool isNegative = readBool();
        if (isNegative) {
            position[i] = -1 * coord;
        } else {
            position[i] = coord;
        }
    }

    return position;
}

float BinCursor::getDistance(bool isSmallScene)
{
    float meters = readCodeWord("metersCode").value[0];

    float distance = meters;
    if (!isSmallScene) {
        bool addHectometers = readBool();
        if (addHectometers) {
            float hectometers = readCodeWord("hectometersCode").value[0];
            distance = distance + 100 * hectometers;

            bool addKilometers = readBool();
            while (addKilometers) {
                float kilometers = readCodeWord("kilometersCode").value[0];
                distance = distance + kilometers * 1000;

                addKilometers = readBool();
            }
        }
    }

    bool addCentimeters = readBool();
    if (addCentimeters) {
        // centimetersCode
        float centimeters = readCodeWord("centimetersCode").value[0];
        distance = distance + centimeters / 100.0f;

        // addMillimeters
        bool addMillimeters = readBool();
        if (addMillimeters) {
            // millimetersCode
            float millimeters = readCodeWord("millimetersCode").value[0];
            distance = distance + millimeters / 1000.0f;
        }
    }

    return distance;
}

std::array<float, 3> BinCursor::getOrientation()
{
    CodeWordResult r = readCodeWord("orientationCode");
    std::array<float, 3> orientation{ r.value[0], r.value[1], r.value[2] };
    if (r.description.find("Default angles") != std::string::npos) {
        orientation[0] = readCodeWord("defaultYawCode").value[0];
        orientation[1] = readCodeWord("defaultPitchCode").value[0];
        orientation[2] = readCodeWord("defaultRollCode").value[0];
    } else if (r.description.find("Explicit angles") != std::string::npos) {
        for (unsigned int o = 0; o < 3; ++o) {
            float coarseAngle = readCodeWord("coarseAngleCode").value[0];
            float angle = coarseAngle;
            bool addFineAngle = readBool();
            if (addFineAngle) {
                float fineAngle = readCodeWord("fineAngleCode").value[0];
                angle = angle + fineAngle;
            }
            orientation[o] = angle;
        }
    } else if (r.description.find("High resolution angles") != std::string::npos) {
        for (unsigned int o = 0; o < 3; ++o) {
            float angleVal = readFloat();
            orientation[o] = angleVal;
        }
    }

    return orientation;
}

float BinCursor::getGain(bool isHiPrecGain)
{
    float gain = readCodeWord("coarseGainCode").value[0];

    bool addFineGain = readBool();
    if (addFineGain) {
        float fineGain = readCodeWord("fineGainCode").value[0];

        gain = gain + fineGain;
    }
    if (isHiPrecGain) {
        bool addDecimalGain = readBool();
        if (addDecimalGain) {
            float decimalGain = readCodeWord("deciSecondsCode").value[0];
            gain = gain + decimalGain;
        }
    }

    return gain;
}

} // namespace MPEGI
