/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/Bitstream.h>

#include <MPEGI/Bitstream/BinCursor.h>
#include <MPEGI/Helpers/Config.h>

#include <EAR/Exception.h>
#include <EAR/HeadphoneEQ.h>

#include <cmath>
#include <fstream>
#include <streambuf>

namespace MPEGI {

namespace detail {

inline std::string bitstreamVersionToString(int version)
{
    return "v" + std::to_string(version);
}

template <typename T>
inline void readBits(T& target, const std::uint8_t* data, std::size_t& start, std::size_t nBits)
{
    EAR_CHECK_REL(nBits <= (sizeof(T) * 8),
        "Bitstream: Cannot read more than " + std::to_string((sizeof(T) * 8))
            + " bits into given type.");

    for (unsigned int i = 0; i < nBits; i++) {
        int byteNum = (start + i) >> 3;
        std::uint8_t byte = data[byteNum];
        std::uint8_t readBit = byte >> (7 - ((start + i) % 8)) & 1;
        target = (target << 1) + readBit;
    }

    start += nBits;
}

inline int readDelayBufferSize(BinCursor& cursor)
{
    int bufferSize = cursor.readUInt(3);
    switch (bufferSize) {
    case 0:
        return 10000;
        break;
    case 1:
        return 20000;
        break;
    case 2:
        return 30000;
        break;
    case 3:
        return 50000;
        break;
    case 4:
        return 100000;
        break;
    case 5:
        return 200000;
        break;
    case 6:
        return 500000;
        break;
    case 7:
        return 1000000;
        break;
    default:
        EAR_FAIL("Error reading max distance delay");
    }
}

inline float readCullingThreshold(BinCursor& cursor)
{
    int threshold = cursor.readUInt(3);
    return std::pow(10.f, threshold - 10.f);
}

inline float readLoudnessCullingThreshold(BinCursor& cursor)
{
    auto const threshold = cursor.readUInt(6);
    return std::pow(10.f, (-73.f + threshold) / 10.f);
}

inline int readSpeedOfSound(BinCursor& cursor)
{
    bool overrideSpeedOfSound = cursor.readBool();
    if (overrideSpeedOfSound) {
        return cursor.readUInt(13);
    } else {
        return 340;
    }
}

inline int readTemperature(BinCursor& cursor)
{
    bool overrideTemperature = cursor.readBool();
    if (overrideTemperature) {
        return 5 * cursor.readUInt(5) - 50;
    } else {
        return 20;
    }
}

inline float readHumidity(BinCursor& cursor)
{
    bool overrideHumidity = cursor.readBool();
    if (overrideHumidity) {
        return 6.667 * cursor.readUInt(4);
    } else {
        return 40.f;
    }
}

enum class fgdMethodT : unsigned int
{
    Individual_Frequencies = 0u,
    Start_Hop_Amount,
    Default_Banding,
    Reserved
};

inline std::vector<std::shared_ptr<FreqGrid>> readFreqGrid(BinCursor& cursor)
{
    std::vector<std::shared_ptr<FreqGrid>> freqGrids;
    std::vector<unsigned int> fgdNrBands;
    const unsigned int fgdNrGrids = cursor.getCountOrIndex();
    for (unsigned int i = 0; i < fgdNrGrids; ++i) {
        const unsigned int fgdMethod = cursor.readUInt(2);
        if (fgdMethod == (unsigned int)fgdMethodT::Individual_Frequencies) {
            fgdNrBands.push_back(cursor.getCountOrIndex());
            std::vector<float> fgdCenterFreq;
            fgdCenterFreq.reserve(fgdNrBands[i]);
            for (unsigned int j = 0; j < fgdNrBands[i]; ++j) {
                fgdCenterFreq.push_back(cursor.getFrequency());
            }
            std::shared_ptr<FreqGrid> data =
                std::make_shared<FreqGrid>(fgdNrBands[i], fgdCenterFreq);
            freqGrids.push_back(data);
        } else if (fgdMethod == (unsigned int)fgdMethodT::Start_Hop_Amount) {
            fgdNrBands.push_back(cursor.getCountOrIndex());
            std::vector<float> fgdCenterFreq;
            fgdCenterFreq.reserve(fgdNrBands[i]);
            fgdCenterFreq.push_back(cursor.getFrequency());
            const float frequencyHop = cursor.readCodeWord("frequencyHopCode").value[0];
            for (unsigned int j = 1; j < fgdNrBands[i]; ++j) {
                fgdCenterFreq.push_back(fgdCenterFreq[j - 1] * frequencyHop);
            }

            std::shared_ptr<FreqGrid> data =
                std::make_shared<FreqGrid>(fgdNrBands[i], fgdCenterFreq);
            freqGrids.push_back(data);
        } else if (fgdMethod == (unsigned int)fgdMethodT::Default_Banding) {
            CodeWordResult grid = cursor.readCodeWord("fgdDefaultGrid");
            const bool fgdIsSubGrid = cursor.readBool();
            if (fgdIsSubGrid) {
                unsigned int fgdDefaultGridOffset = cursor.readUInt(3);
                unsigned int fgdDefaultGridNrBands = cursor.readUInt(6);
                std::vector<float> fgdCenterFreq;
                fgdNrBands.push_back(fgdDefaultGridNrBands + 1);
                for (unsigned int j = 0; j < fgdNrBands[i]; ++j) {
                    fgdCenterFreq.push_back(grid.value[fgdDefaultGridOffset + j]);
                }
                std::shared_ptr<FreqGrid> data =
                    std::make_shared<FreqGrid>(fgdNrBands[i], fgdCenterFreq);
                freqGrids.push_back(data);
            } else {
                fgdNrBands.push_back(grid.value.size());
                std::shared_ptr<FreqGrid> data =
                    std::make_shared<FreqGrid>(fgdNrBands[i], grid.value);
                freqGrids.push_back(data);
            }

        } else {
            EAR_FAIL("Bitstream signals mode that is not supported yet.");
        }
    }
    return freqGrids;
}

inline std::shared_ptr<EAR::HeadphoneEQData> readHeadphoneEQ(
    BinCursor& cursor, const BitstreamConfig& bsConfig)
{
    const unsigned int specType = cursor.readUInt(3);
    std::shared_ptr<EAR::HeadphoneEQData> data = std::make_shared<EAR::HeadphoneEQData>();
    if (specType == (unsigned int)EAR::HeadphoneSpecType::freqgain) {
        auto freqGridData = bsConfig.GetFreqGrid();
        int freqGridIdx = cursor.getCountOrIndex();
        int nrBands = freqGridData[freqGridIdx]->nrBands;
        std::vector<float> gains;
        for (unsigned int i = 0; i < nrBands; ++i) {
            gains.push_back(cursor.getGain(true));
        }

        data->specType = EAR::HeadphoneSpecType::freqgain;
        data->centerFreq = freqGridData[freqGridIdx]->centerFreq;
        data->gain = gains;
    } else if (specType == (unsigned int)EAR::HeadphoneSpecType::parametricfilters) {
        auto freqGridData = bsConfig.GetFreqGrid();
        int freqGridIdx = cursor.getCountOrIndex();
        int nrBands = freqGridData[freqGridIdx]->nrBands;
        std::vector<EAR::HeadphoneFilterType> filterType;
        std::vector<float> gains;
        std::vector<float> q;
        for (unsigned int i = 0; i < nrBands; ++i) {
            filterType.push_back(static_cast<EAR::HeadphoneFilterType>(cursor.readUInt(2)));
            gains.push_back(cursor.getGain(true));
            q.push_back(cursor.readQuantFloat(HEADPHONE_EQ_FILTER_Q_RANGE));
        }
        data->specType = EAR::HeadphoneSpecType::parametricfilters;
        data->filterType = filterType;
        data->centerFreq = freqGridData[freqGridIdx]->centerFreq;
        data->gain = gains;
        data->q = q;
    } else if (specType == (unsigned int)EAR::HeadphoneSpecType::preset) {
        data->preset = cursor.readUInt(7);
        data->specType = EAR::HeadphoneSpecType::preset;
    }
    return data;
}

inline BitstreamUpdate readUpdate(BinCursor& cursor, const BitstreamConfig& bsConfig)
{
    const bool isLongRangeMode = true; // specified in WD text
    const bool isHiPrecGain = true; // specified in WD text
    MPEGI::BitstreamUpdate update = MPEGI::BitstreamUpdate();

    int numberOfMods = cursor.getCountOrIndex();
    while (numberOfMods--) {
        std::string modId = bsConfig.getIdMap().at(cursor.getID());

        float modDuration = 0.f;
        bool hasDuration = cursor.readBool();
        if (hasDuration) {
            modDuration = cursor.getDuration(isLongRangeMode);
        }

        int numberOfChanges = cursor.readUInt(5);
        while (numberOfChanges--) {
            update.modifications.push_back(Modification());
            Modification& mod = update.modifications.back();

            mod.id = modId;
            mod.duration = modDuration;

            using TargetAttribute = Modification::TargetAttribute;
            using Placeholder = Modification::Placeholder;
            mod.attribute = static_cast<TargetAttribute>(cursor.readUInt(6));

            switch (mod.attribute) {
            case TargetAttribute::Position: {
                EAR::Vector3 value(0.f, 0.f, 0.f);

                for (int i = 0; i < 3; i++) {
                    bool isPlaceholder = cursor.readBool();
                    if (isPlaceholder) {
                        mod.placeholders.insert(
                            std::make_pair(cursor.readUInt(4), static_cast<Placeholder>(i)));
                    } else {
                        value[i] = cursor.getDistance(bsConfig.isSmallScene);
                        bool isNegative = cursor.readBool();
                        if (isNegative)
                            value[i] = -value[i];
                    }
                }

                mod.targetValue = EAR::Variant(value);

                break;
            }
            case TargetAttribute::Orientation: {
                EAR::Euler value(0.f, 0.f, 0.f);

                for (int i = 0; i < 3; i++) {
                    bool isPlaceholder = cursor.readBool();
                    if (isPlaceholder) {
                        mod.placeholders.insert(
                            std::make_pair(cursor.readUInt(4), static_cast<Placeholder>(i + 3)));
                    } else {
                        value[i] = cursor.readFloat();
                    }
                }

                mod.targetValue = EAR::Variant(value);
                break;
            }
            case TargetAttribute::CSpace: {
                bool cspace = cursor.readBool();
                if (cspace == true) {
                    mod.targetValue = EAR::Variant(EAR::CSpace::LCS);
                } else {
                    mod.targetValue = EAR::Variant(EAR::CSpace::RCS);
                }
                break;
            }
            case TargetAttribute::Active:
            case TargetAttribute::Play: {
                mod.targetValue = EAR::Variant(cursor.readBool());
                break;
            }
            case TargetAttribute::GainDb: {
                bool isPlaceholder = cursor.readBool();
                if (isPlaceholder) {
                    mod.placeholders.insert(
                        std::make_pair(cursor.readUInt(4), Placeholder::GainDb));
                } else {
                    float newGain = cursor.getGain(isHiPrecGain);
                    mod.targetValue = EAR::Variant(newGain);
                }
                break;
            }
            case TargetAttribute::Signal:
            case TargetAttribute::Extent:
            case TargetAttribute::Directivity:
            case TargetAttribute::Group:
            case TargetAttribute::Region: {
                mod.targetValue = EAR::Variant(bsConfig.getIdMap().at(cursor.getID()));
                break;
            }
            case TargetAttribute::Directiveness: {
                mod.targetValue =
                    EAR::Variant(cursor.readQuantFloat(DIRECTIVENESS_QUANTIZATION_RANGE));
                break;
            }
            case TargetAttribute::Size: {
                float sizeX = cursor.getDistance(bsConfig.isSmallScene);
                float sizeY = cursor.getDistance(bsConfig.isSmallScene);
                float sizeZ = cursor.getDistance(bsConfig.isSmallScene);
                mod.targetValue = EAR::Variant(EAR::Vector3(sizeX, sizeY, sizeZ));
                break;
            }
            case TargetAttribute::Speed: {
                bool isPlaceholder = cursor.readBool();
                if (isPlaceholder) {
                    mod.placeholders.insert(std::make_pair(cursor.readUInt(4), Placeholder::Speed));
                } else {
                    float newSpeed = cursor.readQuantFloat(AIRFLOW_SPEED_RANGE);
                    mod.targetValue = EAR::Variant(newSpeed);
                }
                break;
            }
            case TargetAttribute::Material: {
                mod.targetValue = EAR::Variant(bsConfig.getIdMap().at(cursor.getID()));
                break;
            }
            case TargetAttribute::ReverbGain: {
                bool isPlaceholder = cursor.readBool();
                if (isPlaceholder) {
                    mod.placeholders.insert(
                        std::make_pair(cursor.readUInt(4), Placeholder::ReverbGain));
                } else {
                    float newGain = pow(10, cursor.getGain(isHiPrecGain) / 20);
                    mod.targetValue = EAR::Variant(newGain);
                }
                break;
            }
            case TargetAttribute::NoReverb: {
                mod.targetValue = EAR::Variant(cursor.readBool());
                break;
            }
            case TargetAttribute::LoudnessDb: {
                bool isPlaceholder = cursor.readBool();
                if (isPlaceholder) {
                    mod.placeholders.insert(
                        std::make_pair(cursor.readUInt(4), Placeholder::LoudnessDb));
                } else {
                    mod.targetValue = EAR::Variant(cursor.readFloat());
                }
                break;
            }
            case TargetAttribute::GranularPos: {
                EAR::Vector3 value(0.f, 0.f, 0.f);

                for (int i = 0; i < 3; i++) {
                    bool isPlaceholder = cursor.readBool();
                    if (isPlaceholder) {
                        mod.placeholders.insert(std::make_pair(cursor.readUInt(4),
                            static_cast<Placeholder>((int)Placeholder::GranularPosX + i)));
                    } else {
                        value[i] = cursor.readQuantFloat(GRANULAR_POS_RANGE);
                    }
                }

                mod.targetValue = EAR::Variant(value);
                break;
            }
            case TargetAttribute::GranularGroupProb: {
                EAR::Vector3 value(-1.f, 0.f, 0.f);

                bool isPlaceholder = cursor.readBool();
                if (isPlaceholder) {
                    mod.placeholders.insert(
                        std::make_pair(cursor.readUInt(4), static_cast<Placeholder>(0)));
                } else {
                    value[0] = cursor.readQuantFloat(GRANULAR_PROB_GRP_NO_RANGE);
                }
                isPlaceholder = cursor.readBool();
                if (isPlaceholder) {
                    mod.placeholders.insert(
                        std::make_pair(cursor.readUInt(4), static_cast<Placeholder>(1)));
                } else {
                    value[1] = cursor.readQuantFloat(GRANULAR_GROUP_PROB_RANGE);
                }

                mod.targetValue = EAR::Variant(value);
                break;
            }
            }
        }
    }

    update.type = static_cast<BitstreamUpdate::Type>(cursor.readUInt(2));

    switch (update.type) {
    case BitstreamUpdate::Type::Timed: {
        bool hasId = cursor.readBool();
        if (hasId) {
            update.id = bsConfig.getIdMap().at(cursor.getID());
        }

        bool hasIndex = cursor.readBool();
        if (hasIndex) {
            update.index = cursor.getCountOrIndex();
        }

        update.time = cursor.getDuration(isLongRangeMode);
        break;
    }
    case BitstreamUpdate::Type::Conditional: {
        bool hasId = cursor.readBool();
        if (hasId) {
            update.id = bsConfig.getIdMap().at(cursor.getID());
        }

        bool hasIndex = cursor.readBool();
        if (hasIndex) {
            update.index = cursor.getCountOrIndex();
        }

        update.fireOn = cursor.readBool();

        bool hasDelay = cursor.readBool();
        if (hasDelay) {
            update.delay = cursor.getDuration(isLongRangeMode);
        }

        update.condition = bsConfig.getIdMap().at(cursor.getID());
        break;
    }
    case BitstreamUpdate::Type::Triggered:
    case BitstreamUpdate::Type::Dynamic: {
        int id = cursor.getID();
        update.id = bsConfig.getIdMap().at(id);
        update.index = cursor.getCountOrIndex();
        break;
    }
    }

    return update;
}

/**
 * Decode early reflection tuning gain if present
 *
 * @param cursor BinCursor instance
 * @return Tuning gain in dB, if present
 */
std::optional<float> decodeEarlyTuningGain(BinCursor& cursor)
{
    auto const has_tuning_gain = cursor.readBool();
    return has_tuning_gain ? std::optional<float>((static_cast<int>(cursor.readUInt(6)) - 43) * .5f)
                           : std::nullopt;
};

} // namespace detail

Bitstream::Bitstream(const std::string& filepath)
{
    // load file and resize buffer according to its size
    std::ifstream file(filepath, std::ios::binary);
    EAR_CHECK_REL(file.is_open(),
        "Invalid scene file - Could not be opened: Please check the input file and its path")

    file.seekg(0, std::ios::end);
    totalSize = file.tellg();
    buffer = std::make_unique<const std::uint8_t[]>(totalSize);

    // write file contents into buffer
    // this intentionally circumvents constness to make sure nobody else does this accidentally
    file.seekg(0);
    file.read(reinterpret_cast<char*>(const_cast<unsigned char*>(buffer.get())), totalSize);

    BinCursor cursor = BinCursor(buffer.get(), totalSize);

    // Parse header - Should contain "MPEGI" as a string and 2 bytes for the version number
    std::uint64_t bitstreamIdentifier = cursor.readUInt(48);
    EAR_CHECK_REL(bitstreamIdentifier == 0x4d5045474900, // "MPEGI" in hexadecimal representation
        "Invalid bitstream format: Incorrect header information; must be an MPEGI bitstream");

    std::uint16_t bitstreamVersion = cursor.readUInt(16);
    std::uint16_t expectedBitstreamVersion = MPEGI_BS_VERSION;
    EAR_CHECK_REL(bitstreamVersion == expectedBitstreamVersion,
        "Invalid bitstream format: Renderer expects "
            + detail::bitstreamVersionToString(expectedBitstreamVersion)
            + ", but received bitstream version "
            + detail::bitstreamVersionToString(bitstreamVersion));

    // create payload instances
    bool isConfigPacketRead = false;
    while (!cursor.atLastByte()) {
        MHASPacketType packetType = static_cast<MHASPacketType>(cursor.readUInt(11));
        int label = cursor.readUInt(2);
        std::size_t packetLength = cursor.readUInt(59);

        auto const cursor_beg = cursor.getPosition();
        switch (packetType) {
        case MHASPacketType::PACTYP_SCENE_CFG: {
            int idCount = cursor.getCountOrIndex();
            while (idCount--) {
                int intId = cursor.getID();
                config.intIdToStringId.emplace(std::make_pair(intId, cursor.readString()));
            }

            config.delayBufferSize = detail::readDelayBufferSize(cursor);
            config.gainCullingThreshold = detail::readCullingThreshold(cursor);
            config.enableCullingReflectionRI = cursor.readBool();
            config.speedOfSound = detail::readSpeedOfSound(cursor);
            config.temperature = detail::readTemperature(cursor);
            config.humidity = detail::readHumidity(cursor);
            config.isSmallScene = cursor.readBool();

            config.freqGridData = detail::readFreqGrid(cursor);
            config.earlyTuningGainDb = detail::decodeEarlyTuningGain(cursor);

            config.useLowComplexityER = cursor.readBool();

            config.hasHeadphoneEQ = cursor.readBool();
            if (config.hasHeadphoneEQ) {
                config.contentCreatorHP = detail::readHeadphoneEQ(cursor, config);
            } else {
                config.contentCreatorHP = std::make_shared<EAR::HeadphoneEQData>();
            }

            config.loudnessCullingThreshold =
                cursor.readBool() ? detail::readLoudnessCullingThreshold(cursor) : 0.f;

            config.hasConsolidationThreshold = cursor.readBool();
            config.consolidationThreshold = 5;
            if (config.hasConsolidationThreshold)
                config.consolidationThreshold = cursor.readUInt(5);

            config.enableLocallyCapturedAudio = cursor.readBool();
            if (config.enableLocallyCapturedAudio) {
                const bool overrideListenerVoiceDirectivity = cursor.readBool();
                if (overrideListenerVoiceDirectivity) {
                    const int bsId = cursor.getID();
                    config.listenerVoiceDirectivityId = config.getIdMap().at(bsId);
                } else {
                    config.listenerVoiceDirectivityId = "";
                }
            }

            cursor.byteAlign();

            isConfigPacketRead = true;
        } break;
        case MHASPacketType::PACTYP_SCENE_UPD: {
            EAR_CHECK_REL(isConfigPacketRead,
                "Incorrect Packet Ordering: Config Packet must be serialized before update "
                "packet")

            std::size_t updCount = cursor.getCountOrIndex();
            updates.reserve(updCount);
            while (updCount--) {
                updates.push_back(detail::readUpdate(cursor, config));
            }

        } break;
        case MHASPacketType::PACTYP_SCENE_PLD: {
            unsigned int packetId = cursor.readUInt(16); // Not currently used
            std::size_t numOfPayloads = cursor.readUInt(16);

            while (numOfPayloads--) {
                PayloadType payloadType = static_cast<PayloadType>(cursor.readUInt(11));
                unsigned int payloadId = cursor.readUInt(2); // Not currently used
                std::size_t payloadLength = cursor.readUInt(59);
                EAR_CHECK_REL(payloadLength > 0,
                    "Cannot have a payload of length 0 - Should have been removed in the "
                    "encoder")

                rawPayloads.push_back(RawPayload(payloadType, payloadLength,
                    buffer.get() + static_cast<std::size_t>(cursor.getPosition() / 8)));

                cursor.movePosRight(payloadLength * 8);
            }
        } break;
        default:
            // skip unknown packet
            cursor.movePosRight(packetLength * 8);
            break;
        };

        // Skip remaining bits in packet, implicit byte-alignment
        cursor.movePosRight((packetLength << 3) - cursor.getPosition() + cursor_beg);
    }

    EAR_CHECK_REL(cursor.atLastByte(), "Invalid bitstream format: Bytes left at EOF");

    // register previously introduced handlers by invoking local registration function
    for (auto* func : this->getHandlerList()) {
        auto [handler, type] = func();
        this->registerHandler(std::move(handler), type);
    }

    MPEGI::Config::setDef("Distance_MaxDelay", (int)config.delayBufferSize);
    MPEGI::Config::setDef("SimpleCulling_DeactivationThreshold", config.gainCullingThreshold);
    MPEGI::Config::setDef("EnableCullingReflectionRI", config.enableCullingReflectionRI);
    MPEGI::Config::setDef("Humidity", config.humidity);
    MPEGI::Config::setDef("Temperature", config.temperature);
    MPEGI::Config::setDef("SpeedOfSound", config.speedOfSound);

    auto const early_tuning_gain_db = config.getEarlyTuningGain();
    auto const early_tuning_gain_lin = std::pow(10.f, early_tuning_gain_db * .05f);

    MPEGI::Config::setDef("EarlyTuningGain", early_tuning_gain_lin);
    MPEGI::Config::setDef("VoxelEarlyReflectionLevel", early_tuning_gain_lin);

    MPEGI::Config::setDef("useLowComplexityER", config.useLowComplexityER);
    MPEGI::Config::setDef(
        "SimpleCulling_DeactivationThresholdRel", config.loudnessCullingThreshold);
    MPEGI::Config::setDef("Consolidation_Threshold", config.consolidationThreshold);
}

void Bitstream::registerHandler(std::shared_ptr<PayloadHandler> handler, PayloadType type)
{
    // call new handler for payloads that already have been parsed
    for (auto& payload : rawPayloads) {
        if (payload.getType() == type) {
            payloads[type].push_back(handler->unpack(payload, config));
        }
    }
}

Bitstream::HandlerList& Bitstream::getHandlerList()
{
    static HandlerList mylist;
    return mylist;
}

void Bitstream::clearHandlerList()
{
    auto& list = getHandlerList();
    list.clear();
}

std::vector<std::shared_ptr<const Bitstream::Payload>> Bitstream::getAllPayloadsOfType(
    PayloadType type) const
{
    if (payloads.find(type) != payloads.end()) {
        return payloads.at(type);
    } else { // return empty vector when nothing is found
        return std::vector<std::shared_ptr<const Bitstream::Payload>>();
    }
}

std::shared_ptr<const Bitstream::Payload> Bitstream::getOnlyPayloadOfType(PayloadType type) const
{
    if (payloads.find(type) == payloads.end()) {
        auto pldAsString = std::string{ payloadTypeToString(type) };
        EAR_FAIL("Zero payloads found for requested type: " + pldAsString);
    } else if (payloads.find(type) != payloads.end() && payloads.at(type).size() > 1) {
        auto pldAsString = std::string{ payloadTypeToString(type) };
        EAR_FAIL("Many payloads found where exactly one was requested: " + pldAsString);
    }
    return payloads.at(type)[0];
}

// -----------------------------------------------------------------------------

Bitstream::RawPayload::RawPayload(
    Bitstream::PayloadType type, std::size_t length, const std::uint8_t* buffer)
    : type(type)
    , bufferSize(length)
    , buffer(buffer)
    , bitsRead(0)
{}

bool Bitstream::RawPayload::readBool() const
{
    std::uint8_t val = 0;
    // conciously bypassing constness of the function to update the read cursor
    // (but disencouraging any other modification of members)
    detail::readBits<std::uint8_t>(val, buffer, const_cast<RawPayload*>(this)->bitsRead, 1);
    return static_cast<bool>(val);
}

int Bitstream::RawPayload::readInt(std::size_t length) const
{
    int val = readUInt(length);
    // Converting two's complement unsigned representation into signed integer representation
    if (val >= 1 << (length - 1)) {
        val -= 1 << length;
    }

    return (int)val;
}

unsigned int Bitstream::RawPayload::readUInt(std::size_t length) const
{
    unsigned int val = 0;
    detail::readBits<unsigned int>(val, buffer, const_cast<RawPayload*>(this)->bitsRead, length);
    return val;
}

float Bitstream::RawPayload::readFloat() const
{
    std::uint32_t val = 0;
    detail::readBits<std::uint32_t>(val, buffer, const_cast<RawPayload*>(this)->bitsRead, 32);
    return *(float*)(&val);
}

float Bitstream::RawPayload::readQuantFloat(QuantizationRange qRange) const
{
    unsigned int encodedValue = readUInt(qRange.bits);
    return (qRange.max - qRange.min) * static_cast<float>(encodedValue)
        / float(std::pow(2, qRange.bits) - 1)
        + qRange.min;
}

std::string Bitstream::RawPayload::readString() const
{
    std::string string;

    std::uint8_t character = 0x0;
    while (true) {
        detail::readBits<std::uint8_t>(
            character, buffer, const_cast<RawPayload*>(this)->bitsRead, 8);

        if (character == 0x0) { // Null terminator in ASCII string
            return string;
        }

        string += std::string(1, character);
    }
}

int Bitstream::RawPayload::readEscapedUInt(std::size_t l1, std::size_t l2, std::size_t l3) const
{
    int val = 0;
    detail::readBits<int>(val, buffer, const_cast<RawPayload*>(this)->bitsRead, l1);

    if (val == (1 << l1) - 1) {
        int valueAdd = 0;
        detail::readBits<int>(valueAdd, buffer, const_cast<RawPayload*>(this)->bitsRead, l2);
        val += valueAdd;

        if (valueAdd == (1 << l2) - 1) {
            valueAdd = 0;
            detail::readBits<int>(valueAdd, buffer, const_cast<RawPayload*>(this)->bitsRead, l3);
            val += valueAdd;
        }
    }

    return val;
}

std::unique_ptr<std::uint8_t[]> Bitstream::RawPayload::readBlob(std::size_t& length) const
{
    detail::readBits<std::size_t>(length, buffer, const_cast<RawPayload*>(this)->bitsRead, 32);

    auto blob = std::make_unique<std::uint8_t[]>(length);

    for (int i = 0; i < length; i++) {
        std::uint8_t byte = 0;
        detail::readBits<std::uint8_t>(byte, buffer, const_cast<RawPayload*>(this)->bitsRead, 8);
        blob[i] = byte;
    }

    return blob;
}

std::shared_ptr<std::vector<std::uint8_t>> Bitstream::RawPayload::readBlob() const
{
    std::size_t length{ 0 };
    detail::readBits<std::size_t>(length, buffer, const_cast<RawPayload*>(this)->bitsRead, 32);

    auto blob = std::make_shared<std::vector<std::uint8_t>>(length);

    auto* data = blob->data();
    for (int i = 0; i < length; i++) {
        std::uint8_t byte = 0;
        detail::readBits<std::uint8_t>(byte, buffer, const_cast<RawPayload*>(this)->bitsRead, 8);
        data[i] = byte;
    }

    return blob;
}

void Bitstream::RawPayload::byteAlign() const
{
    auto const mod = bitsRead % 8;
    if (mod) {
        auto& bitsRead = const_cast<RawPayload*>(this)->bitsRead;
        bitsRead += (8 - mod);
    }
}

CodeWordResult Bitstream::RawPayload::readCodeWord(const std::string& tableName) const
{
    std::uint16_t word = 0;
    // conciously bypassing constness of the function to update the read cursor
    // (but disencouraging any other modification of members)

    uint64_t length = 0;
    CodeWordResult result;
    while (true) {

        detail::readBits<std::uint16_t>(word, buffer, const_cast<RawPayload*>(this)->bitsRead, 1);
        length++;

        // LUT = {  codeword, corresponding value;
        //          codeword, corresponding value;
        //          codeword, corresponding value}

        if (LUTs::readCodeWord(tableName, word, length, result))
            break;
    }

    return std::move(result);
}

unsigned int Bitstream::RawPayload::getID() const
{
    unsigned int id = readUInt(7);

    bool largerValue = readBool();
    while (largerValue) {
        unsigned int idVal = readUInt(7);

        id = id * 128 + idVal;

        largerValue = readBool();
    }

    return id;
}

unsigned int Bitstream::RawPayload::getCountOrIndex() const
{
    unsigned int number = readCodeWord("countOrIndexLoCode").value[0];

    bool isLargeNumber = readBool();
    if (isLargeNumber) {
        unsigned int numberHi = readCodeWord("countOrIndexHiCode").value[0];
        number = number + numberHi * 64;
    }

    return number;
}

float Bitstream::RawPayload::getDuration(bool isLongRangeMode) const
{
    float deciSeconds = readCodeWord("deciSecondsCode").value[0];
    float duration = deciSeconds;

    bool addMs = readBool();
    if (addMs) {
        float miliSeconds = readCodeWord("milliSecondsCode").value[0];
        duration = duration + miliSeconds;
        bool addMicroseconds = readBool();
        if (addMicroseconds) {
            float microSecs = readCodeWord("microsecondsCode").value[0];
            duration = duration + microSecs;
        }
    }

    bool addSecs = readBool();
    if (addSecs) {
        if (isLongRangeMode) {
            unsigned int rawSeconds = readUInt(12);
            duration = duration + rawSeconds;
            bool addHours = readBool();
            if (addHours) {
                unsigned int rawHours = readUInt(5);
                duration = duration + (rawHours + 1) * 3600;
            }
        } else {
            float seconds = readCodeWord("secondsCode").value[0];
            duration = duration + seconds;
        }
    }

    return duration;
}

float Bitstream::RawPayload::getFrequency() const
{
    float frequency = readCodeWord("frequencyCode").value[0];

    bool moreAccuracy = readBool();
    if (moreAccuracy) {
        float frequencyRefine = readUInt(4);
        frequency = frequency * std::pow(2.0f, ((frequencyRefine + 1) / 51.0f));
    }
    return frequency;
}

std::array<float, 3> Bitstream::RawPayload::getPosition(bool isSmallScene) const
{
    std::array<float, 3> position{ 0, 0, 0 };

    for (int i = 0; i < 3; ++i) {
        float coord = getDistance(isSmallScene);
        bool isNegative = readBool();
        if (isNegative) {
            position[i] = -1 * coord;
        } else {
            position[i] = coord;
        }
    }

    return position;
}

float Bitstream::RawPayload::getDistance(bool isSmallScene) const
{
    float meters = readCodeWord("metersCode").value[0];

    float distance = meters;
    if (!isSmallScene) {
        bool addHectometers = readBool();
        if (addHectometers) {
            float hectometers = readCodeWord("hectometersCode").value[0];
            distance = distance + 100 * hectometers;

            bool addKilometers = readBool();
            while (addKilometers) {
                float kilometers = readCodeWord("kilometersCode").value[0];
                distance = distance + kilometers * 1000;

                addKilometers = readBool();
            }
        }
    }

    bool addCentimeters = readBool();
    if (addCentimeters) {
        // centimetersCode
        float centimeters = readCodeWord("centimetersCode").value[0];
        distance = distance + centimeters / 100.0f;

        // addMillimeters
        bool addMillimeters = readBool();
        if (addMillimeters) {
            // millimetersCode
            float millimeters = readCodeWord("millimetersCode").value[0];
            distance = distance + millimeters / 1000.0f;
        }
    }

    return distance;
}

std::array<float, 3> Bitstream::RawPayload::getOrientation() const
{
    CodeWordResult r = readCodeWord("orientationCode");
    std::array<float, 3> orientation{ r.value[0], r.value[1], r.value[2] };
    if (r.description.find("Default angles") != std::string::npos) {
        orientation[0] = readCodeWord("defaultYawCode").value[0];
        orientation[1] = readCodeWord("defaultPitchCode").value[0];
        orientation[2] = readCodeWord("defaultRollCode").value[0];
    } else if (r.description.find("Explicit angles") != std::string::npos) {
        for (unsigned int o = 0; o < 3; ++o) {
            float coarseAngle = readCodeWord("coarseAngleCode").value[0];
            float angle = coarseAngle;
            bool addFineAngle = readBool();
            if (addFineAngle) {
                float fineAngle = readCodeWord("fineAngleCode").value[0];
                angle = angle + fineAngle;
            }
            orientation[o] = angle;
        }
    } else if (r.description.find("High resolution angles") != std::string::npos) {
        for (unsigned int o = 0; o < 3; ++o) {
            float angleVal = readFloat();
            orientation[o] = angleVal;
        }
    }

    return orientation;
}

float Bitstream::RawPayload::getGain(bool isHiPrecGain) const
{
    float gain = readCodeWord("coarseGainCode").value[0];

    bool addFineGain = readBool();
    if (addFineGain) {
        float fineGain = readCodeWord("fineGainCode").value[0];

        gain = gain + fineGain;
    }
    if (isHiPrecGain) {
        bool addDecimalGain = readBool();
        if (addDecimalGain) {
            float decimalGain = readCodeWord("deciSecondsCode").value[0];
            gain = gain + decimalGain;
        }
    }

    return gain;
}

std::array<float, 6> Bitstream::RawPayload::getTolerance() const
{
    std::array<float, 6> tolerance{ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
    for (int i = 0; i < 6; ++i) {
        tolerance[i] = readCodeWord("decimalCode").value[0];
    }
    return tolerance;
}

Bitstream::RawPayloadForTesting::RawPayloadForTesting(std::string dataString)
    : RawPayload(Bitstream::PLD_ACOUSTIC_ENVIRONMENT, 0, nullptr)
{
    constexpr size_t bitsPerByte{ 8 };

    // Add padding at the end of the data
    while (dataString.size() % bitsPerByte != 0) {
        dataString += "0";
    }

    // initialize member variables
    bufferSize = dataString.size() / bitsPerByte;
    vector_buffer.resize(bufferSize);
    buffer = vector_buffer.data();
    for (int i = 0; i < bufferSize; i++) {
        std::string substr = dataString.substr(i * bitsPerByte, bitsPerByte);
        vector_buffer.at(i) = std::stoi(substr, nullptr, 2);
    }
}

} // namespace MPEGI
