/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/CompressedDirectivity.h>

#include <MPEGI/DirectivityCoding.h>

#include <EAR/LinearInterpolator.h>

#include <algorithm>
#include <cmath>
#include <functional>

namespace MPEGI {

CompressedDirectivity::CompressedDirectivity(std::uint8_t* data, std::size_t length)
{
    coverList.decodeFromBuffer(length, data);

    for (int i = 0; i < coverList.getFrequencyCount(); i++) {
        frequencies.push_back(coverList.getFrequency(i));
    }
};

float CompressedDirectivity::get(float freq, const EAR::Polar& coordinates,
    bool shouldInterpolateFreq, bool shouldInterpolateSpace)
{
    EAR_CHECK(coordinates[0] >= 0.f && coordinates[0] < 360.f,
        "Azimuth angle for compressed directivity lookup out of range.");

    // find nearest frequency index in data
    auto lit = std::lower_bound(frequencies.begin(), frequencies.end(), freq);
    int idx = lit - frequencies.begin();

    // use either space-interpolating function or nearest neighbor
    typedef float (DirectivityCoverList::*GetterFn)(int idx, float ele, float azi) const;
    GetterFn getMagnitude = (shouldInterpolateSpace)
        ? &DirectivityCoverList::getDirectivityInterpolated
        : &DirectivityCoverList::getDirectivityNearestPoint;

    float magnitude;
    if (idx == 0) {
        magnitude = std::invoke(getMagnitude, coverList, 0, coordinates[1], coordinates[0]);

    } else if (idx >= frequencies.size()) {
        magnitude = std::invoke(
            getMagnitude, coverList, frequencies.size() - 1, coordinates[1], coordinates[0]);
    } else if (*lit == freq) {
        magnitude = std::invoke(getMagnitude, coverList, idx, coordinates[1], coordinates[0]);
    } else {
        float t = (freq - frequencies[idx - 1]) / (frequencies[idx] - frequencies[idx - 1]);
        float m1 = std::invoke(getMagnitude, coverList, idx - 1, coordinates[1], coordinates[0]);
        float m2 = std::invoke(getMagnitude, coverList, idx, coordinates[1], coordinates[0]);
        if (shouldInterpolateFreq) {
            // interpolate linearly
            // TODO: interpolation happens in dB here instead of magnitudes. Is that a problem?
            magnitude = (1 - t) * m1 + t * m2;
        } else {
            // nearest neighbor
            magnitude = (t <= 0.5f) ? m1 : m2;
        }
    }
    return magnitude;
}

} // namespace MPEGI
