/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/DiffractionPayload.h>
#include <MPEGI/Bitstream/HuffmanDecoder.h>

#include <EAR/Exception.h>

#include <algorithm>
#include <iostream>

namespace {

static const int NUM_OF_BITS_FOR_MAX_OBJECT_SOURCES = 16;
static const int NUM_OF_BITS_FOR_MAX_VERTICES = 15;
static const int NUM_OF_BITS_FOR_MAX_FACES = 16;

static const int NUM_OF_BITS_FOR_MAX_EDGES = 16;
static const int NUM_OF_BITS_FOR_MAX_PATHS_PER_KEY = 16;
static const int NUM_OF_BITS_FOR_MAX_EDGES_PER_PATH = 8;
static const int NUM_OF_BITS_FOR_MAX_GRIDS_PER_AXIS = 16;
static const int NUM_OF_BITS_FOR_MAX_EDGES_PER_VOXEL = 16;
static const int NUM_OF_BITS_FOR_MAX_VOXELS_PER_SOURCE = 16;
static const int NUM_OF_BITS_FOR_MAX_LISTENER_VOXELS = 32;
static const int NUM_OF_BITS_FOR_MAX_PATHS = 32;
static const int NUM_OF_BITS_FOR_ANGLE_BITS = 6;

MPEGI::DiffractionEdge readEdge(const MPEGI::Bitstream::RawPayload& bs,
    MPEGI::HuffmanDecoder& decoderEdgeID, MPEGI::HuffmanDecoder& decoderVtxID,
    MPEGI::HuffmanDecoder& decoderTriID)
{
    MPEGI::DiffractionEdge edge;
    edge.edgeID = decoderEdgeID.decodeSingle(bs);
    edge.parentMesh = nullptr;
    edge.meshID = -1; // "sceneTrimeshDIFFR";

    edge.vtxIDs.first = decoderVtxID.decodeSingle(bs);
    edge.vtxIDs.second = decoderVtxID.decodeSingle(bs);

    edge.vtxCoords.first[0] = -1.f;
    edge.vtxCoords.first[1] = -1.f;
    edge.vtxCoords.first[2] = -1.f;

    edge.vtxCoords.second[0] = -1.f;
    edge.vtxCoords.second[1] = -1.f;
    edge.vtxCoords.second[2] = -1.f;

    edge.edgeVector[0] = -1.f;
    edge.edgeVector[1] = -1.f;
    edge.edgeVector[2] = -1.f;

    edge.normalizedEdgeDirection[0] = -1.f;
    edge.normalizedEdgeDirection[1] = -1.f;
    edge.normalizedEdgeDirection[2] = -1.f;

    edge.edgeLength = -1.f;

    // Adjacent triangles/faces
    edge.adjacentTriIDs.first = decoderTriID.decodeSingle(bs);
    edge.adjacentTriIDs.second = decoderTriID.decodeSingle(bs);

    edge.adjacentTriCentroids.first[0] = -1.f;
    edge.adjacentTriCentroids.first[1] = -1.f;
    edge.adjacentTriCentroids.first[2] = -1.f;

    edge.adjacentTriCentroids.second[0] = -1.f;
    edge.adjacentTriCentroids.second[1] = -1.f;
    edge.adjacentTriCentroids.second[2] = -1.f;

    edge.adjacentTriNormals.first[0] = -1.f;
    edge.adjacentTriNormals.first[1] = -1.f;
    edge.adjacentTriNormals.first[2] = -1.f;

    edge.adjacentTriNormals.second[0] = -1.f;
    edge.adjacentTriNormals.second[1] = -1.f;
    edge.adjacentTriNormals.second[2] = -1.f;

    // Edge
    edge.edgeAngle = -1.f;

    // Center points
    edge.cp[0] = -1.f;
    edge.cp[1] = -1.f;
    edge.cp[2] = -1.f;

    edge.cp_vt[0] = -1.f;
    edge.cp_vt[1] = -1.f;
    edge.cp_vt[2] = -1.f;

    // Flags
    edge.roundedEdge = bs.readBool(); // roundedEdge
    edge.relevant = bs.readBool(); // relevant

    return edge;
}

} // namespace

namespace MPEGI {

class DiffractionPayloadHandler : public Bitstream::PayloadHandler
{
public:
    //! Constructor
    DiffractionPayloadHandler() = default;

    //! Destructor
    ~DiffractionPayloadHandler() = default;

    void unpackVoxelGrid(const Bitstream::RawPayload& bs, const BitstreamConfig& bsConfig,
        DiffractionPayload& payload)
    {
        auto v = bs.getPosition(bsConfig.isSmallScene);
        payload.voxelGridOrigin[0] = v[0];
        payload.voxelGridOrigin[1] = v[1];
        payload.voxelGridOrigin[2] = v[2];

        payload.voxelGridPitch[0] = bs.getDistance(bsConfig.isSmallScene);
        payload.voxelGridPitch[1] = bs.getDistance(bsConfig.isSmallScene);
        payload.voxelGridPitch[2] = bs.getDistance(bsConfig.isSmallScene);

        payload.voxelGridShape[0] = bs.getID();
        payload.voxelGridShape[1] = bs.getID();
        payload.voxelGridShape[2] = bs.getID();
    }

    void unpackEdgeList(const Bitstream::RawPayload& bs, std::vector<DiffractionEdge>& edgeList)
    {
        // codebooks for diffrStaticEdgeList
        HuffmanDecoder decoderEdgeID;
        HuffmanDecoder decoderVtxID;
        HuffmanDecoder decoderTriID;
        decoderEdgeID.deserializeCodebook(bs);
        decoderVtxID.deserializeCodebook(bs);
        decoderTriID.deserializeCodebook(bs);

        int nEdges = bs.getID(); // Number of Static Edges
        for (int i = 0; i < nEdges; i++) {
            edgeList.push_back(readEdge(bs, decoderEdgeID, decoderVtxID, decoderTriID));
        }
    }

    void unpackStaticEdgeList(const Bitstream::RawPayload& bs, DiffractionPayload& payload)
    {
        payload.edgeList.clear();

        // flag: has_static_edge_data
        bool has_static_edge_data = bs.readBool();
        if (has_static_edge_data == false) {
            return;
        }

        unpackEdgeList(bs, payload.edgeList);
    }

    void unpackPathDict(const Bitstream::RawPayload& bs,
        std::map<int, std::vector<DiffractionPath>>& pathDict,
        std::vector<DiffractionEdge>& edgeList)
    {
        // deserialize codebooks
        HuffmanDecoder decoderEdgeIDSeqLen;
        HuffmanDecoder decoderEdgeIDSeq;
        HuffmanDecoder decoderAngleSeq;
        decoderEdgeIDSeqLen.deserializeCodebook(bs);
        decoderEdgeIDSeq.deserializeCodebook(bs);
        decoderAngleSeq.deserializeCodebook(bs);

        // angle de-quantization constant
        int numBitsAngle = bs.readUInt(NUM_OF_BITS_FOR_ANGLE_BITS);
        int num_steps_angle = 1 << numBitsAngle;
        float quant2angle = float(360.0 / double(num_steps_angle));

        // Number of relevant edges in path dict
        int nRelevantEdges = bs.getID();

        for (int i = 0; i < nRelevantEdges; i++) {
            // Number of Paths with {eID}-th Edge List
            int nPaths = bs.getID();

            for (int j = 0; j < nPaths; j++) {
                // number of edges in the path
                int nPoints = decoderEdgeIDSeqLen.decodeSingle(bs);

                std::vector<DiffractionPoint> pointSeq;
                for (int k = 0; k < nPoints; k++) {
                    // edge Id in the given edgeIDSeq
                    int eId = decoderEdgeIDSeq.decodeSingle(bs);

                    // face indicator in the given faceIndicatorSeq
                    int fIndicator = bs.readBool();

                    // angle in the given angleSeq
                    int bs_angle = decoderAngleSeq.decodeSingle(bs);
                    float angle = bs_angle * quant2angle;

                    DiffractionPoint p = { eId, &edgeList[eId], fIndicator, angle };
                    pointSeq.push_back(p);
                }
                DiffractionPath newPath(pointSeq);

                int pathDictKey = newPath.getKeyId(); // = the initial edge Id
                if (pathDict.find(pathDictKey) != pathDict.end()) {
                    pathDict.at(pathDictKey).push_back(newPath);
                } else {
                    std::vector<DiffractionPath> paths = { newPath };
                    pathDict.insert({ pathDictKey, paths });
                }
            }
        }
    }

    void unpackStaticPathDict(const Bitstream::RawPayload& bs, DiffractionPayload& payload)
    {
        payload.pathDict.clear();

        // flag: has_static_edge_data
        bool has_static_paths_data = bs.readBool();
        if (has_static_paths_data == false) {
            return;
        }

        // parse static path dict
        std::map<int, std::vector<DiffractionPath>>& pathDict = payload.pathDict;
        std::vector<DiffractionEdge>& edgeList = payload.edgeList;
        unpackPathDict(bs, pathDict, edgeList);
    }

    void unpackListenerVoxelDict(const Bitstream::RawPayload& bs, DiffractionPayload& payload)
    {
        payload.listenerVoxelDict.clear();

        // flag: has_listener_voxel_data
        bool has_listener_voxel_data = bs.readBool();
        if (has_listener_voxel_data == false) {
            return;
        }

        // cache for edge lists
        std::unordered_map<unsigned int, std::vector<int>> voxelCache;

        // voxel index differences of neighbors
        std::vector<VoxelCoord> delta_list = { VoxelCoord(0, 0, 0), VoxelCoord(-1, 0, 0),
            VoxelCoord(0, -1, 0), VoxelCoord(0, 0, -1) };

        // deserialize codebooks
        HuffmanDecoder decoderVcX;
        HuffmanDecoder decoderVcY;
        HuffmanDecoder decoderVcZ;
        HuffmanDecoder decoderNumEdges;
        HuffmanDecoder decoderEdgeId;
        HuffmanDecoder decoderIndicesRemoved;
        decoderVcX.deserializeCodebook(bs);
        decoderVcY.deserializeCodebook(bs);
        decoderVcZ.deserializeCodebook(bs);
        decoderNumEdges.deserializeCodebook(bs);
        decoderEdgeId.deserializeCodebook(bs);
        decoderIndicesRemoved.deserializeCodebook(bs);

        // voxel coordinate declaration
        int vcX = -1;
        int vcY = -1;
        int vcZ = -1;

        // Number of listener voxels
        unsigned int nListenerVoxels = bs.getID();

        for (unsigned int n = 0; n < nListenerVoxels; n++) {

            // decode voxel coordinate
            vcZ += 1;
            bool hasVoxelCoordZ = bs.readBool();
            if (hasVoxelCoordZ) {
                vcZ = decoderVcZ.decodeSingle(bs);

                vcY += 1;
                bool hasVoxelCoordY = bs.readBool();
                if (hasVoxelCoordY) {
                    vcY = decoderVcY.decodeSingle(bs);

                    vcX += 1;
                    bool hasVoxelCoordX = bs.readBool();
                    if (hasVoxelCoordX) {
                        vcX = decoderVcX.decodeSingle(bs);
                    }
                }
            }

            // Dictionary key: voxel coordinate
            std::vector<int> dictKey = { vcX, vcY, vcZ };

            // voxel coordinate
            VoxelCoord vc(vcX, vcY, vcZ);

            // edge list for the current voxel
            std::vector<int> edge_list;

            // read mode
            unsigned int mode = bs.readUInt(2);

            if (mode != 0) {
                // get reference path list
                int* vc_ref = (vc + delta_list[mode]).data();
                std::vector<int> key_ref = { vc_ref[0], vc_ref[1], vc_ref[2] };
                edge_list = payload.listenerVoxelDict[key_ref];

                // generate list with removed items in reverse order
                std::vector<int> listIndicesRemoved;
                int val = -1;
                int inc;
                while ((inc = decoderIndicesRemoved.decodeSingle(bs)) != 0) {
                    val += inc;
                    listIndicesRemoved.insert(listIndicesRemoved.begin(), val);
                }

                // remove items
                for (int pos : listIndicesRemoved) {
                    edge_list.erase(edge_list.begin() + pos);
                }
            }

            // add items
            int num_edges = decoderNumEdges.decodeSingle(bs);
            for (int n = 0; n < num_edges; n++) {
                int edgeId = decoderEdgeId.decodeSingle(bs);
                edge_list.push_back(edgeId);
            }

            // sort edge list
            std::sort(edge_list.begin(), edge_list.end());

            // add edge list to listener voxel dict
            payload.listenerVoxelDict.insert({ dictKey, std::move(edge_list) });
        }
    }

    int unpackSourceVoxelDict(const Bitstream::RawPayload& bs, DiffractionPayload& payload)
    {
        using SourceVoxelDict = std::map<std::vector<int>, std::vector<int>>;

        payload.sourceVoxelDicts.clear();

        // flag: has_sourcevoxel_data
        bool has_sourcevoxel_data = bs.readBool();
        if (has_sourcevoxel_data == false) {
            return 0;
        }

        int nStaticSources = bs.getID(); // Number of static sources
        for (int i = 0; i < nStaticSources; i++) {
            int srcIdIdx = bs.getID(); //Source Index: {sourceID, sourceIdx}
            // number of voxels per source
            int nSourceVoxels = bs.getID();

            SourceVoxelDict sourceVoxelDictPerSource;
            float count = 0;
            while (count < nSourceVoxels) {
                int xGridIndex = bs.getID();
                int yGridIndex = bs.getID();
                int zGridIndex = bs.getID();
                std::vector<int> dictKey = { xGridIndex, yGridIndex,
                    zGridIndex }; // Voxel grid index

                int nEdgesPerVoxel = bs.getID(); // Number of edges in voxel grid

                std::vector<int> visibleEdgesAtSourceVoxel;
                for (int i = 0; i < nEdgesPerVoxel; i++) {
                    int veId = bs.getID();
                    visibleEdgesAtSourceVoxel.push_back(veId);
                }
                sourceVoxelDictPerSource.insert({ dictKey, visibleEdgesAtSourceVoxel });
                count += 1;
            }

            payload.sourceVoxelDicts.insert({ srcIdIdx, sourceVoxelDictPerSource });
        }

        return nStaticSources;
    }

    void unpackValidPathDict(
        const Bitstream::RawPayload& bs, DiffractionPayload& payload, int nStaticSources)
    {
        using PathList = std::vector<std::pair<int, int>>;
        using VoxelPathDict = std::map<uint64_t, PathList>;

        // valid path dictionary data in bitstream
        payload.validPathDict.clear();
        payload.staticSourceIdxList.clear();

        // flag: has_valid_paths
        bool has_valid_paths = bs.readBool();
        if (has_valid_paths == false) {
            return;
        }

        int nStaticSources1 = bs.getID();
        if (nStaticSources != 0) {
            EAR_CHECK_REL(nStaticSources == nStaticSources1,
                ">> Error: the numbers of static sources assigned in valid path dictionay and "
                "source "
                "voxel "
                "dictionary are different - check the encoder statement");
        }

        for (int s = 0; s < nStaticSources1; s++) {
            int srcIdIdx = bs.getID(); //Source index: {sourceID, sourceIdx}
            payload.staticSourceIdxList.push_back(srcIdIdx);

            // cache for edge lists
            std::unordered_map<unsigned int, std::vector<int>> voxelCache;

            // voxel index differences of neighbors
            std::vector<VoxelCoord> delta_list = { VoxelCoord(0, 0, 0), VoxelCoord(-1, 0, 0),
                VoxelCoord(0, -1, 0), VoxelCoord(0, 0, -1) };

            // deserialize codebooks
            HuffmanDecoder decoderVcX;
            HuffmanDecoder decoderVcY;
            HuffmanDecoder decoderVcZ;
            HuffmanDecoder decoderNumPaths;
            HuffmanDecoder decoderEdgeId;
            HuffmanDecoder decoderPathId;
            HuffmanDecoder decoderIndicesRemoved;
            decoderVcX.deserializeCodebook(bs);
            decoderVcY.deserializeCodebook(bs);
            decoderVcZ.deserializeCodebook(bs);
            decoderNumPaths.deserializeCodebook(bs);
            decoderEdgeId.deserializeCodebook(bs);
            decoderPathId.deserializeCodebook(bs);
            decoderIndicesRemoved.deserializeCodebook(bs);

            // voxel coordinate declaration
            int vcX = -1;
            int vcY = -1;
            int vcZ = -1;

            // Number of listener voxels
            unsigned int nListenerVoxels = bs.getID();
            VoxelPathDict validPathDictPerSource;

            for (unsigned int v = 0; v < nListenerVoxels; v++) {

                // decode voxel coordinate
                vcZ += 1;
                bool hasVoxelCoordZ = bs.readBool();
                if (hasVoxelCoordZ) {
                    vcZ = decoderVcZ.decodeSingle(bs);

                    vcY += 1;
                    bool hasVoxelCoordY = bs.readBool();
                    if (hasVoxelCoordY) {
                        vcY = decoderVcY.decodeSingle(bs);

                        vcX += 1;
                        bool hasVoxelCoordX = bs.readBool();
                        if (hasVoxelCoordX) {
                            vcX = decoderVcX.decodeSingle(bs);
                        }
                    }
                }

                // Dictionary key: voxel coordinate
                std::vector<int> dictKey = { vcX, vcY, vcZ };

                // voxel coordinate
                VoxelCoord vc(vcX, vcY, vcZ);
                const uint64_t vhash = PackVoxelCoords(vcX, vcY, vcZ);

                // path list for the current voxel
                std::vector<std::pair<int, int>> validPathList;

                // read mode
                unsigned int mode = bs.readUInt(2);

                if (mode != 0) {
                    // get reference path list
                    int* vc_ref = (vc + delta_list[mode]).data();
                    const uint64_t vhash_ref = PackVoxelCoords(vc_ref[0], vc_ref[1], vc_ref[2]);
                    validPathList = validPathDictPerSource[vhash_ref];

                    // generate list with removed items in reverse order
                    std::vector<int> listIndicesRemoved;
                    int val = -1;
                    int inc;
                    while ((inc = decoderIndicesRemoved.decodeSingle(bs)) != 0) {
                        val += inc;
                        listIndicesRemoved.insert(listIndicesRemoved.begin(), val);
                    }

                    // remove items
                    for (int pos : listIndicesRemoved) {
                        validPathList.erase(validPathList.begin() + pos);
                    }
                }

                // Number of paths per source-listener pair
                int nPathsPerSourceListenerPair = decoderNumPaths.decodeSingle(bs);
                for (int p = 0; p < nPathsPerSourceListenerPair; p++) {
                    int veId = decoderEdgeId.decodeSingle(bs);
                    int pLocalIndex = decoderPathId.decodeSingle(bs);
                    validPathList.push_back(std::make_pair(veId, pLocalIndex));
                }

                // sort path list
                std::sort(validPathList.begin(), validPathList.end());

                // add valid path list for the current voxel
                validPathDictPerSource.insert({ vhash, validPathList });
            }
            payload.validPathDict.insert({ srcIdIdx, validPathDictPerSource });
        }
    }

    void unpackDynamicEdges(const Bitstream::RawPayload& bs, DiffractionPayload& payload)
    {
        // dynamic edge lists
        payload.dynamicEdgeLists.clear();

        // flag: has_dynamic_edge_data
        bool has_dynamic_edge_data = bs.readBool();
        if (has_dynamic_edge_data == false) {
            return;
        }

        // dynamic edge lists
        int nDynEdgeLists = bs.getID(); // Number of Dynamic Edge Lists
        for (int j = 0; j < nDynEdgeLists; j++) {
            int geoIdIdx = bs.getID(); // Geometry Index: {geoId, geoIdx}

            std::vector<DiffractionEdge> edgeListToUpdate;
            unpackEdgeList(bs, edgeListToUpdate);

            // Update dynamic edge lists map with entity-Id and its corresponding edge list
            payload.dynamicEdgeLists.insert({ geoIdIdx, edgeListToUpdate });
        }
    }

    void unpackDynamicPaths(const Bitstream::RawPayload& bs, DiffractionPayload& payload)
    {
        // dynamic path dictionaries
        payload.dynamicPathDicts.clear();

        // flag: has_dynamic_paths
        bool has_dynamic_paths = bs.readBool();
        if (has_dynamic_paths == false) {
            return;
        }

        int nDynPathDicts = bs.getID(); // Number of Dynamic Path Dicts
        for (int j = 0; j < nDynPathDicts; j++) {
            int geoIdIdx = bs.getID(); // Geometry Index: {geoId, geoIdx}

            // parse dynamic path dict
            std::map<int, std::vector<DiffractionPath>>& pathDict =
                payload.dynamicPathDicts[geoIdIdx];
            std::vector<DiffractionEdge>& edgeList = payload.dynamicEdgeLists[geoIdIdx];
            unpackPathDict(bs, pathDict, edgeList);
        }
    }

    //! Inherited from Bitstream::PayloadHandler
    virtual std::shared_ptr<const Bitstream::Payload> unpack(
        const Bitstream::RawPayload& bs, const BitstreamConfig& bsConfig) override
    {
        auto payload = std::shared_ptr<DiffractionPayload>(new DiffractionPayload());

        // unpack diffrVoxelGrid data
        unpackVoxelGrid(bs, bsConfig, *payload);

        // unpack diffrStaticEdgeList data
        unpackStaticEdgeList(bs, *payload);

        // unpack diffrStaticPathDict data
        unpackStaticPathDict(bs, *payload);

        // unpack diffrListenerVoxelDict data
        unpackListenerVoxelDict(bs, *payload);

        // unpack diffrSourceVoxelDict data
        int nStaticSources = unpackSourceVoxelDict(bs, *payload);

        // unpack diffrValidPathDict data
        unpackValidPathDict(bs, *payload, nStaticSources);

        // unpack diffrDynamicEdges data
        unpackDynamicEdges(bs, *payload);

        // unpack diffrDynamicPaths data
        unpackDynamicPaths(bs, *payload);

        EAR_CHECK_REL(bs.atLastByte(),
            "Unspecified error while parsing when parsing DiffractionPayload (too much data)");

        // name resolution for bitstream elements
        for (const auto& it : bsConfig.getIdMap()) {
            payload->geoName2BitsId.insert({ it.second, it.first });
        }

        return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
    }
};

Bitstream::PayloadHandlerAndType makeDiffractionPayloadHandler()
{
    return { std::make_shared<DiffractionPayloadHandler>(), Bitstream::PLD_DIFFR };
}

} // namespace MPEGI
