/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/EIFPayloadHandler.h>

#include <MPEGI/Bitstream/Globals.h>
#include <MPEGI/GeometryCoding.h>
#include <MPEGI/Helpers/Config.h>

#include <EAR/Exception.h>

#include <algorithm>
#include <fstream>
#include <iomanip> //cli cout
#include <iostream>

#if defined(WIN32)
#define MPEGI_DIRECTORY_SEPARATOR '\\'
#else
#define MPEGI_DIRECTORY_SEPARATOR '/'
#endif

namespace detail {

EAR::Vector3 readLocation(const MPEGI::Bitstream::RawPayload& cursor)
{
    float posX = cursor.readFloat();
    float posY = cursor.readFloat();
    float posZ = cursor.readFloat();
    return EAR::Vector3(posX, posY, posZ);
}

EAR::Euler readOrientation(const MPEGI::Bitstream::RawPayload& cursor)
{
    float oriX = cursor.readFloat();
    float oriY = cursor.readFloat();
    float oriZ = cursor.readFloat();
    return EAR::Euler(oriX, oriY, oriZ);
}

}; // namespace detail

namespace MPEGI {

EIFPayloadHandler::EIFPayloadHandler(const std::string& basePath, bool checkFiles)
    : basePath(basePath)
    , checkFiles(checkFiles)
    , payload(std::shared_ptr<EIFPayload>(new EIFPayload(basePath, checkFiles)))
{}

std::shared_ptr<const Bitstream::Payload> EIFPayloadHandler::unpack(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    createSceneData(cursor);
    createTransforms(cursor, bsConfig);
    createAnchors(cursor, bsConfig);
    createAudioStreams(cursor, bsConfig);
    createGranularSound(cursor, bsConfig);
    createAcousticMaterials(cursor, bsConfig);
    createDirectivities(cursor, bsConfig);
    createMeshes(cursor, bsConfig);
    createPrimitives(cursor, bsConfig);
    createAcousticEnvironments(cursor, bsConfig);
    createObjectSources(cursor, bsConfig);
    createHoaGroups(cursor, bsConfig);
    createHoaSources(cursor, bsConfig);
    createChannelSources(cursor, bsConfig);
    createAirflowGenerators(cursor, bsConfig);

    EAR_CHECK_REL(cursor.atLastByte(),
        "Unspecified error while parsing when parsing EIFPayload (too much data)");

    return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
}

void EIFPayloadHandler::createSceneData(const Bitstream::RawPayload& cursor)
{
    payload->scene.duration = cursor.readQuantFloat(SCENE_DURATION_QUANTIZATION_RANGE);
    payload->scene.type = static_cast<EIFPayload::SceneInfo::SceneType>(cursor.readUInt(1));
}

void EIFPayloadHandler::createAcousticEnvironments(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfEnvironments = cursor.getCountOrIndex();
    std::vector<int> boundedEnvIds;
    int idDefaultAE;

    for (int i = 0; i < numberOfEnvironments; i++) {
        EIFPayload::AcousticEnvironmentDesc env;

        int numberOfParameters = cursor.getCountOrIndex();
        for (int j = 0; j < numberOfParameters; j++) {
            EIFPayload::AcousticEnvironmentDesc::AcousticParameterDesc param;

            auto v = cursor.getPosition(bsConfig.isSmallScene);
            param.position.location = { v[0], v[1], v[2] };

            env.parameters.push_back(param);
        }

        env.bsId = cursor.getID();
        boundedEnvIds.push_back(env.bsId);
        env.id = bsConfig.getIdMap().at(env.bsId);

        env.region = findGeometry(cursor.getID());

        payload->environments.push_back(env);
    }

    // A default acoustic environment may be defined in the ReverbPayload.
    // Since ReverbPayload is not available at this point we append a new acoustic environment to the list of envoronments just in case.
    // Note that in the default acoustic environment region == nullptr.
    EIFPayload::AcousticEnvironmentDesc env;
    env.parameters.emplace_back();
    // making sure that idDefaultAE has unique value within the scope of AEs
    // TODO: Revisit ID management in reverb stage and for default AE in particular
    if (boundedEnvIds.size() < 1) {
        idDefaultAE = 0;
    } else {
        idDefaultAE = *max_element(boundedEnvIds.begin(), boundedEnvIds.end()) + 1;
    }
    // we assign a dummy ID that doesn't exist in the bitstream
    env.bsId = idDefaultAE;
    env.id = "defaultAE";
    payload->environments.push_back(env);
}

void EIFPayloadHandler::createAcousticMaterials(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfMaterials = cursor.getCountOrIndex();
    const auto& freqGridData = bsConfig.GetFreqGrid();
    for (int i = 0; i < numberOfMaterials; i++) {
        EIFPayload::AcousticMaterialDesc material;

        material.bsId = cursor.getID();
        material.id = bsConfig.getIdMap().at(material.bsId);

        std::size_t length = 0;
        auto materialBlob = cursor.readBlob(length);
        auto decoder = GeometryData();
        if (!decoder.decodeFromBuffer(length, materialBlob.get())) {
            EAR_FAIL("Decoding Error: Acoustic Material '" + material.id + "' couldn't be decoded.")
        }

        for (int j = 0; j < decoder.ml.list[0]->freq_count; j++) {

            material.frequencies.emplace_back(EIFPayload::AcousticMaterialDesc::FrequencyDesc{
                freqGridData[decoder.ml.list[0]->freq_grid_index]->centerFreq[j],
                decoder.ml.list[0]->c1[j],
                decoder.ml.list[0]->c2[j],
                decoder.ml.list[0]->c3[j],
                decoder.ml.list[0]->c4[j],
            });
        }

        payload->materials.push_back(material);
    }
}

void EIFPayloadHandler::createAirflowGenerators(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    const bool isHiPrecGain = false;
    const bool isSmallScene = bsConfig.isSmallScene;
    int airflowGeneratorCount = cursor.getCountOrIndex();

    for (int i = 0; i < airflowGeneratorCount; i++) {
        EIFPayload::AirflowGeneratorDesc airflowGenerator;
        airflowGenerator.bsId = cursor.getID();
        airflowGenerator.id = bsConfig.getIdMap().at(airflowGenerator.bsId);
        int type = cursor.readUInt(3);
        airflowGenerator.speed = cursor.readQuantFloat(AIRFLOW_SPEED_RANGE);
        auto o = cursor.getOrientation();
        airflowGenerator.position.orientation = { o[0], o[1], o[2] };
        airflowGenerator.gainDb = cursor.getGain(isHiPrecGain);
        airflowGenerator.active = cursor.readBool();
        switch (type) {
        case 0: {
            EIFPayload::GlobalAirflowGeneratorDesc global(airflowGenerator);
            global.hasDistanceAttenuationFactor = cursor.readBool();
            if (global.hasDistanceAttenuationFactor) {
                global.distanceAttenuationFactor = cursor.readFloat();
            }
            payload->globalAirflowGenerators.push_back(global);
            break;
        }
        case 1: {
            EIFPayload::PointAirflowGeneratorDesc point(airflowGenerator);
            auto v = cursor.getPosition(isSmallScene);
            point.position.location = { v[0], v[1], v[2] };
            point.hasLimits = cursor.readBool();
            if (point.hasLimits) {
                point.aziRange = cursor.readQuantFloat(AIRFLOW_AZI_ELEV_LIMIT_RANGE);
                point.elevRange = cursor.readQuantFloat(AIRFLOW_AZI_ELEV_LIMIT_RANGE);
            }
            point.hasDistanceAttenuationFactor = cursor.readBool();
            if (point.hasDistanceAttenuationFactor) {
                point.distanceAttenuationFactor =
                    cursor.readQuantFloat(AIRFLOW_DISTANCE_ATTENUATION_FACTOR_RANGE);
            }
            payload->pointAirflowGenerators.push_back(point);
            break;
        }
        case 2: {
            EIFPayload::ConeAirflowGeneratorDesc cone(airflowGenerator);
            auto v = cursor.getPosition(isSmallScene);
            cone.position.location = { v[0], v[1], v[2] };
            cone.baseRadius = cursor.getDistance(isSmallScene);
            cone.length = cursor.getDistance(isSmallScene);
            cone.hasTopRadius = cursor.readBool();
            if (cone.hasTopRadius) {
                cone.topRadius = cursor.getDistance(isSmallScene);
            }
            cone.edgeRollOff = cursor.readUInt(10);
            cone.hasDistanceAttenuationFactor = cursor.readBool();
            if (cone.hasDistanceAttenuationFactor) {
                cone.distanceAttenuationFactor =
                    cursor.readQuantFloat(AIRFLOW_DISTANCE_ATTENUATION_FACTOR_RANGE);
            }
            payload->coneAirflowGenerators.push_back(cone);
            break;
        }
        case 3: {
            EIFPayload::PlaneAirflowGeneratorDesc plane(airflowGenerator);
            auto v = cursor.getPosition(isSmallScene);
            plane.position.location = { v[0], v[1], v[2] };
            plane.width = cursor.getDistance(isSmallScene);
            plane.height = cursor.getDistance(isSmallScene);
            plane.edgeRollOff = cursor.readUInt(10);
            plane.hasDistanceAttenuationFactor = cursor.readBool();
            if (plane.hasDistanceAttenuationFactor) {
                plane.distanceAttenuationFactor =
                    cursor.readQuantFloat(AIRFLOW_DISTANCE_ATTENUATION_FACTOR_RANGE);
            }
            payload->planeAirflowGenerators.push_back(plane);
            break;
        }
        }
    }
}
void EIFPayloadHandler::createAudioStreams(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfAudioStreams = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfAudioStreams; i++) {
        EIFPayload::AudioStreamDesc stream;

        stream.bsId = cursor.getID();
        stream.id = bsConfig.getIdMap().at(stream.bsId);
        if (bsConfig.enableLocallyCapturedAudio) {
            stream.isLocallyCaptured = cursor.readBool();
        } else {
            stream.isLocallyCaptured = false;
        }
        if (!stream.isLocallyCaptured) {
            stream.file = checkIfFileExistsRelatively(cursor.readString(), stream.id);

            int numberOfInputChannels = cursor.getCountOrIndex();
            for (int i = 0; i < numberOfInputChannels; i++) {
                stream.aepInputChannels.push_back(cursor.getCountOrIndex());
            }
            stream.isSoundEffect = cursor.readBool();
            if (stream.isSoundEffect) {
                stream.duration = cursor.getDuration(true);
            }
        }
        payload->streams.push_back(stream);
    }
}

void EIFPayloadHandler::createGranularSound(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfGranularSound = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfGranularSound; i++) {
        EIFPayload::GranularSoundDesc granularSound;

        granularSound.bsId = cursor.getID();
        granularSound.id = bsConfig.getIdMap().at(granularSound.bsId);

        granularSound.file = cursor.readString();

        granularSound.numChannels = cursor.getCountOrIndex();

        payload->granularSounds.push_back(granularSound);
    }
}

void EIFPayloadHandler::createAnchors(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfAnchors = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfAnchors; i++) {
        EIFPayload::AnchorDesc anchor;

        anchor.bsId = cursor.getID();
        anchor.id = bsConfig.getIdMap().at(anchor.bsId);

        anchor.lsdfRef = cursor.readString();

        anchor.autoRescale = cursor.readBool();

        auto v = cursor.getTolerance();
        anchor.tolerance = { v[0], v[1], v[2], v[3], v[4], v[5] };

        payload->anchors.push_back(anchor);
    }
}

void EIFPayloadHandler::createChannelSources(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{

    const bool isHiPrecGain = true;
    int numberOfChannelSources = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfChannelSources; i++) {
        EIFPayload::ChannelSourceDesc chanSource;

        int numberOfLoudspeakers = cursor.getCountOrIndex();
        for (int j = 0; j < numberOfLoudspeakers; j++) {
            chanSource.loudspeakers.push_back(EIFPayload::ChannelSourceDesc::LoudspeakerDesc());
            // ID Not used - reserved for future use
            int loudspeakerID = cursor.getID();

            auto v = cursor.getPosition(bsConfig.isSmallScene);
            chanSource.loudspeakers.back().position.location = { v[0], v[1], v[2] };

            auto o = cursor.getOrientation();
            chanSource.loudspeakers.back().position.orientation = { o[0], o[1], o[2] };

            chanSource.loudspeakers.back().channelIndex = cursor.getCountOrIndex();

            bool hasDirectivity = cursor.readBool();
            if (hasDirectivity) {
                chanSource.loudspeakers.back().directivity = findDirectivity(cursor.getID());
            }

            chanSource.loudspeakers.back().directiveness =
                cursor.readQuantFloat(DIRECTIVENESS_QUANTIZATION_RANGE);
        }

        chanSource.bsId = cursor.getID();
        chanSource.id = bsConfig.getIdMap().at(chanSource.bsId);

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        chanSource.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        chanSource.position.orientation = { o[0], o[1], o[2] };

        bool cspace = cursor.readBool();
        if (cspace == true) {
            chanSource.position.cSpace = EAR::CSpace::LCS;
        } else { // cspace == false implied
            chanSource.position.cSpace = EAR::CSpace::RCS;
        }

        chanSource.inputLayout = cursor.readUInt(7);
        chanSource.active = cursor.readBool();
        chanSource.gainDb = cursor.getGain(isHiPrecGain);
        if (cursor.readBool()) { // hasReverbGain
            chanSource.itemReverbGain = pow(10, cursor.getGain(isHiPrecGain) / 20);
        }
        chanSource.refDistance = cursor.getDistance(bsConfig.isSmallScene);
        int ID = cursor.getID();
        chanSource.signal = findStream(ID);
        if (chanSource.signal == nullptr) {
            // it is a Granular Sound not an audio stream
            chanSource.granularSignal = findGranularSound(ID);
        }

        chanSource.noReverb = cursor.readBool();

        if (cursor.readBool()) {
            chanSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDoppler);
        }
        if (cursor.readBool()) {
            chanSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDistance);
        }

        if (cursor.readBool()) {
            chanSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noCulling);
        }

        if (cursor.readBool()) {
            chanSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDirectionalFocus);
        }

        if (cursor.readBool()) {
            chanSource.mode = EIFPayload::AudioElementDesc::Mode::Event;
        } else {
            chanSource.mode = EIFPayload::AudioElementDesc::Mode::Continuous;
        }

        chanSource.play = cursor.readBool();

        chanSource.vdlMethod = VDLMethod(cursor.readUInt(3));

        bool hasSpatialTransform = cursor.readBool();
        if (hasSpatialTransform) {
            if (cursor.readBool()) {
                chanSource.parentAnchor = findAnchor(cursor.getID());
            } else {
                chanSource.parentTransform = findTransform(cursor.getID());
            }
        }

        chanSource.isStatic = cursor.readBool();

        bool hasGranularPos = cursor.readBool();
        if (hasGranularPos) {
            for (int i = 0; i < 3; i++) {
                chanSource.granularPos[i] = cursor.readQuantFloat(GRANULAR_POS_RANGE);
            }
        }

        payload->channelSources.push_back(chanSource);
    }
}

void EIFPayloadHandler::createDirectivities(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfDirectivities = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfDirectivities; i++) {
        EIFPayload::SourceDirectivityDesc directivity;

        directivity.bsId = cursor.getID();
        directivity.id = bsConfig.getIdMap().at(directivity.bsId);

        payload->directivities.push_back(directivity);
    }
}

void EIFPayloadHandler::createPrimitives(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{

    int numberOfPrimitives = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfPrimitives; i++) {
        int primType = cursor.readUInt(2);

        EIFPayload::PrimitiveDesc primitive;

        primitive.bsId = cursor.getID();
        primitive.id = bsConfig.getIdMap().at(primitive.bsId);

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        primitive.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        primitive.position.orientation = { o[0], o[1], o[2] };

        bool cspace = cursor.readBool();
        if (cspace == true) {
            primitive.position.cSpace = EAR::CSpace::LCS;
        } else { // cspace == false implied
            primitive.position.cSpace = EAR::CSpace::RCS;
        }

        primitive.size[0] = cursor.getDistance(bsConfig.isSmallScene);
        primitive.size[1] = cursor.getDistance(bsConfig.isSmallScene);
        primitive.size[2] = cursor.getDistance(bsConfig.isSmallScene);

        bool hasMaterial = cursor.readBool();
        if (hasMaterial) {
            primitive.material = findMaterial(cursor.getID());
        }

        bool hasSpatialTransform = cursor.readBool();
        if (hasSpatialTransform) {
            if (cursor.readBool()) {
                primitive.parentAnchor = findAnchor(cursor.getID());
            } else {
                primitive.parentTransform = findTransform(cursor.getID());
            }
        }

        primitive.isStatic = cursor.readBool();

        primitive.isEarlyReflectionGeometry = cursor.readBool();

        switch (primType) {
        case 0: {
            EIFPayload::BoxDesc box(primitive);
            payload->boxes.push_back(box);
            break;
        }
        case 1: {
            EIFPayload::SphereDesc sphere(primitive);
            payload->spheres.push_back(sphere);
            break;
        }
        case 2: {
            EIFPayload::CylinderDesc cylinder(primitive);
            payload->cylinders.push_back(cylinder);
            break;
        }
        }
    }
}

void EIFPayloadHandler::createMeshes(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{

    int numberOfMeshes = cursor.getCountOrIndex();
    for (int i = 0; i < numberOfMeshes; i++) {
        EIFPayload::MeshDesc mesh;

        mesh.bsId = cursor.getID();
        mesh.id = bsConfig.getIdMap().at(mesh.bsId);

        std::size_t length = 0;
        std::unique_ptr<uint8_t[]> meshBlob = cursor.readBlob(length);
        auto decoder = GeometryData();
        if (!decoder.decodeFromBuffer(length, meshBlob.get())) {
            EAR_FAIL("Decoding Error: Mesh '" + mesh.id + "' couldn't be decoded.")
        }

        for (int j = 0; j < decoder.vl.count; j++) {
            mesh.vertices.emplace_back(EIFPayload::MeshDesc::VertexDesc{
                j, // VertexDesc.index
                {
                    decoder.vl.x[j], // VertexDesc.position.x
                    decoder.vl.y[j], // VertexDesc.position.y
                    decoder.vl.z[j], // VertexDesc.position.z
                },
            });
        }

        for (int j = 0; j < decoder.tl.count; j++) {
            mesh.faces.emplace_back(EIFPayload::MeshDesc::FaceDesc{
                j, // FaceDesc.index
                decoder.tl.v1[j], // FaceDesc.v1
                decoder.tl.v2[j], // FaceDesc.v2
                decoder.tl.v3[j], // FaceDesc.v3
                decoder.tl.mi[j] == -1 ? nullptr //FaceDesc.material
                                       : findMaterial(decoder.tl.mi[j]),
            });
        }

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        mesh.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        mesh.position.orientation = { o[0], o[1], o[2] };

        bool cspace = cursor.readBool();
        if (cspace == true) {
            mesh.position.cSpace = EAR::CSpace::LCS;
        } else { // cspace == false implied
            mesh.position.cSpace = EAR::CSpace::RCS;
        }

        bool hasSpatialTransform = cursor.readBool();
        if (hasSpatialTransform) {
            if (cursor.readBool()) {
                mesh.parentAnchor = findAnchor(cursor.getID());
            } else {
                mesh.parentTransform = findTransform(cursor.getID());
            }
        }

        mesh.isStatic = cursor.readBool();

        mesh.isEarlyReflectionGeometry = cursor.readBool();

        payload->meshes.push_back(mesh);
    }
}

void EIFPayloadHandler::createInformedSource(const Bitstream::RawPayload& cursor,
    const BitstreamConfig& bsConfig, EIFPayload::InformedSourceDesc& src)
{
    src.bsId = cursor.getID();
    src.id = bsConfig.getIdMap().at(src.bsId);

    auto v = cursor.getPosition(bsConfig.isSmallScene);
    src.position.location = { v[0], v[1], v[2] };

    src.listenerThresholdPresent = cursor.readBool();
    src.hoaSourceThresholdPresent = cursor.readBool();

    if (src.listenerThresholdPresent) {
        src.listenerThreshold =
            cursor.readQuantFloat(INFORMED_SOURCE_DISTANCE_THRESHOLD_QUANTIZATION_RANGE);
    }
    if (src.hoaSourceThresholdPresent) {
        src.hoaSourceThreshold =
            cursor.readQuantFloat(INFORMED_SOURCE_DISTANCE_THRESHOLD_QUANTIZATION_RANGE);
    }

    src.priorityValue = cursor.readUInt(6);
}

void EIFPayloadHandler::createHoaGroups(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    int numberOfHoaGroups = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfHoaGroups; i++) {
        EIFPayload::HoaGroupDesc group;

        group.bsId = cursor.getID();
        group.id = bsConfig.getIdMap().at(group.bsId);

        bool hasRegion = cursor.readBool();
        if (hasRegion) {
            group.region = findGeometry(cursor.getID());
        }

        int numberOfCoSources = cursor.getCountOrIndex();

        for (int j = 0; j < numberOfCoSources; j++) {
            group.coSources.push_back(findObjectSource(cursor.getID()));
        }

        bool hasComplexityReductionParameters = cursor.readBool();
        if (hasComplexityReductionParameters) {
            int highestSingleBinBandsIndex = cursor.readUInt(8);
            int lowestHighBandIndex = cursor.readUInt(8);
            int intermediateBandsERBWidth = cursor.readUInt(8);
            group.complexityReductionParameters.push_back(highestSingleBinBandsIndex);
            group.complexityReductionParameters.push_back(lowestHighBandIndex);
            group.complexityReductionParameters.push_back(intermediateBandsERBWidth);
        }

        group.exteriorProjectionRadius = cursor.getDistance(bsConfig.isSmallScene);

        bool hasInformedSources = cursor.readBool();
        if (hasInformedSources) {
            int numberOfInformedSources = cursor.getCountOrIndex();
            group.maxSimulInformedSources = cursor.readUInt(8);
            for (int n = 0; n < numberOfInformedSources; n++) {
                EIFPayload::InformedSourceDesc src;
                createInformedSource(cursor, bsConfig, src);
                group.informedSources.push_back(src);
            }
            group.adaptiveExteriorProjectionRadius = cursor.readBool();
        }

        bool hasLowProfileParameters = cursor.readBool();
        if (hasLowProfileParameters) {
            group.distanceFactor = cursor.readQuantFloat(LOWCOMPMPHOA_DISTANCE_FACTOR);
            group.panningGridOrder = cursor.readUInt(8);
        }

        payload->hoaGroups.push_back(group);
    }
}

void EIFPayloadHandler::createHoaSources(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{

    const bool isHiPrecGain = true;
    int numberOfHoaSources = cursor.getCountOrIndex();
    for (int i = 0; i < numberOfHoaSources; i++) {
        payload->hoaSources.push_back(EIFPayload::HoaSourceDesc());
        EIFPayload::HoaSourceDesc& hoaSource = payload->hoaSources.back();

        hoaSource.bsId = cursor.getID();
        hoaSource.id = bsConfig.getIdMap().at(hoaSource.bsId);

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        hoaSource.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        hoaSource.position.orientation = { o[0], o[1], o[2] };

        bool cspace = cursor.readBool();
        if (cspace == true) {
            hoaSource.position.cSpace = EAR::CSpace::LCS;
        } else { // cspace == false implied
            hoaSource.position.cSpace = EAR::CSpace::RCS;
        }

        hoaSource.active = cursor.readBool();
        hoaSource.gainDb = cursor.getGain(isHiPrecGain);
        if (cursor.readBool()) { // hasReverbGain
            hoaSource.itemReverbGain = pow(10, cursor.getGain(isHiPrecGain) / 20);
        }
        hoaSource.refDistance = cursor.getDistance(bsConfig.isSmallScene);
        int ID = cursor.getID();
        hoaSource.signal = findStream(ID);
        if (hoaSource.signal == nullptr) {
            // it is a Granular Sound not an audio stream
            hoaSource.granularSignal = findGranularSound(ID);
        }

        bool hasExtent = cursor.readBool();
        if (hasExtent) {
            hoaSource.extent = findGeometry(cursor.getID());
        }

        hoaSource.extentTransform = cursor.readBool();
        hoaSource.representation = cursor.readBool();

        bool hasHoaGroup = cursor.readBool();
        if (hasHoaGroup) {
            hoaSource.group = findHoaGroupAndRegister(cursor.getID(), &hoaSource);
        }

        hoaSource.noReverb = cursor.readBool();

        if (cursor.readBool()) {
            hoaSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDoppler);
        }
        if (cursor.readBool()) {
            hoaSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDistance);
        }

        if (cursor.readBool()) {
            hoaSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noCulling);
        }

        if (cursor.readBool()) {
            hoaSource.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDirectionalFocus);
        }

        if (cursor.readBool()) {
            hoaSource.mode = EIFPayload::AudioElementDesc::Mode::Event;
        } else {
            hoaSource.mode = EIFPayload::AudioElementDesc::Mode::Continuous;
        }

        hoaSource.play = cursor.readBool();

        hoaSource.vdlMethod = VDLMethod(cursor.readUInt(3));

        bool hasSpatialTransform = cursor.readBool();
        if (hasSpatialTransform) {
            if (cursor.readBool()) {
                hoaSource.parentAnchor = findAnchor(cursor.getID());
            } else {
                hoaSource.parentTransform = findTransform(cursor.getID());
            }
        }

        hoaSource.isStatic = cursor.readBool();

        bool hasGranularPos = cursor.readBool();
        if (hasGranularPos) {
            for (int i = 0; i < 3; i++) {
                hoaSource.granularPos[i] = cursor.readQuantFloat(GRANULAR_POS_RANGE);
            }
        }
    }
}

void EIFPayloadHandler::createObjectSources(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{

    const bool isHiPrecGain = true;
    int numberOfOjectSources = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfOjectSources; i++) {
        payload->objectSources.push_back(EIFPayload::ObjectSourceDesc());
        EIFPayload::ObjectSourceDesc& objSrc = payload->objectSources.back();

        bool hasInputLayout = cursor.readBool();
        if (hasInputLayout) {
            typedef EIFPayload::ObjectSourceDesc::InputLayoutDesc InLayoutDesc;

            objSrc.inputLayout = std::make_unique<InLayoutDesc>();

            bool alignment = cursor.readBool();
            if (alignment) {
                objSrc.inputLayout->alignment = InLayoutDesc::AlignmentDesc::Object;
            } else {
                objSrc.inputLayout->alignment = InLayoutDesc::AlignmentDesc::User;
            }

            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::TL);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::T);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::TR);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::L);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::C);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::R);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::BL);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::B);
            }
            if (cursor.readBool()) {
                objSrc.inputLayout->positioning.push_back(InLayoutDesc::PositioningDesc::BR);
            }
        }

        objSrc.bsId = cursor.getID();
        objSrc.id = bsConfig.getIdMap().at(objSrc.bsId);

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        objSrc.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        objSrc.position.orientation = { o[0], o[1], o[2] };

        bool cspace = cursor.readBool();
        if (cspace == true) {
            objSrc.position.cSpace = EAR::CSpace::LCS;
        } else { // cspace == false implied
            objSrc.position.cSpace = EAR::CSpace::RCS;
        }

        objSrc.active = cursor.readBool();
        objSrc.gainDb = cursor.getGain(isHiPrecGain);
        if (cursor.readBool()) { // hasReverbGain
            objSrc.itemReverbGain = pow(10, cursor.getGain(isHiPrecGain) / 20);
        }

        objSrc.refDistance = cursor.getDistance(bsConfig.isSmallScene);
        int ID = cursor.getID();
        objSrc.signal = findStream(ID);
        if (objSrc.signal == nullptr) {
            // it is a Granular Sound not an audio stream
            objSrc.granularSignal = findGranularSound(ID);
        }

        bool hasExtent = cursor.readBool();
        if (hasExtent) {
            objSrc.extent = findGeometry(cursor.getID());
        }

        bool hasDirectivity = cursor.readBool();
        if (hasDirectivity) {
            objSrc.directivity = findDirectivity(cursor.getID());
        }

        objSrc.directiveness = cursor.readQuantFloat(DIRECTIVENESS_QUANTIZATION_RANGE);

        objSrc.noReverb = cursor.readBool();

        if (cursor.readBool()) {
            objSrc.aparams.push_back(EIFPayload::AudioElementDesc::AuthoringParameter::noDoppler);
        }
        if (cursor.readBool()) {
            objSrc.aparams.push_back(EIFPayload::AudioElementDesc::AuthoringParameter::noDistance);
        }
        if (cursor.readBool()) {
            objSrc.aparams.push_back(EIFPayload::AudioElementDesc::AuthoringParameter::noCulling);
        }
        if (cursor.readBool()) {
            objSrc.aparams.push_back(
                EIFPayload::AudioElementDesc::AuthoringParameter::noDirectionalFocus);
        }

        if (cursor.readBool()) {
            objSrc.mode = EIFPayload::AudioElementDesc::Mode::Event;
        } else {
            objSrc.mode = EIFPayload::AudioElementDesc::Mode::Continuous;
        }

        objSrc.play = cursor.readBool();

        objSrc.vdlMethod = VDLMethod(cursor.readUInt(3));

        bool hasSpatialTransform = cursor.readBool();
        if (hasSpatialTransform) {
            if (cursor.readBool()) {
                objSrc.parentAnchor = findAnchor(cursor.getID());
            } else {
                objSrc.parentTransform = findTransform(cursor.getID());
            }
        }

        objSrc.isStatic = cursor.readBool();

        bool hasGranularPos = cursor.readBool();
        if (hasGranularPos) {
            for (int i = 0; i < 3; i++) {
                objSrc.granularPos[i] = cursor.readQuantFloat(GRANULAR_POS_RANGE);
            }
        }
    }
}

void EIFPayloadHandler::createTransforms(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{

    int numberOfTransforms = cursor.getCountOrIndex();

    for (int i = 0; i < numberOfTransforms; i++) {
        EIFPayload::TransformDesc transform;

        transform.bsId = cursor.getID();
        transform.id = bsConfig.getIdMap().at(transform.bsId);

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        transform.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        transform.position.orientation = { o[0], o[1], o[2] };

        bool cspace = cursor.readBool();
        if (cspace == true) {
            transform.position.cSpace = EAR::CSpace::LCS;
        } else { // cspace == false implied
            transform.position.cSpace = EAR::CSpace::WCS;
        }

        payload->transforms.push_back(transform);
    }
}

EIFPayload::AnchorDesc* EIFPayloadHandler::findAnchor(int anchorId)
{
    auto anchorIt = std::find_if(payload->anchors.begin(), payload->anchors.end(),
        [&anchorId](const EIFPayload::AnchorDesc& anchor) { return anchor.bsId == anchorId; });

    return &*anchorIt;
}

EIFPayload::SourceDirectivityDesc* EIFPayloadHandler::findDirectivity(int directivityId)
{
    auto directivityIt = std::find_if(payload->directivities.begin(), payload->directivities.end(),
        [&directivityId](const EIFPayload::SourceDirectivityDesc& directivity) {
            return directivity.bsId == directivityId;
        });

    return &*directivityIt;
}

EIFPayload::AcousticMaterialDesc* EIFPayloadHandler::findMaterial(int materialId)
{
    auto materialIt = std::find_if(payload->materials.begin(), payload->materials.end(),
        [&materialId](const EIFPayload::AcousticMaterialDesc& material) {
            return material.bsId == materialId;
        });

    return &*materialIt;
}

EIFPayload::GeometryDesc* EIFPayloadHandler::findGeometry(int geometryId)
{
    auto boxIt = std::find_if(payload->boxes.begin(), payload->boxes.end(),
        [geometryId](const EIFPayload::BoxDesc& subject) { return subject.bsId == geometryId; });

    auto sphereIt = std::find_if(payload->spheres.begin(), payload->spheres.end(),
        [geometryId](const EIFPayload::SphereDesc& subject) { return subject.bsId == geometryId; });

    auto cylinderIt = std::find_if(payload->cylinders.begin(), payload->cylinders.end(),
        [geometryId](
            const EIFPayload::CylinderDesc& subject) { return subject.bsId == geometryId; });

    auto meshIt = std::find_if(payload->meshes.begin(), payload->meshes.end(),
        [geometryId](const EIFPayload::MeshDesc& subject) { return subject.bsId == geometryId; });

    if (boxIt != payload->boxes.end()) {
        return &*boxIt;
    } else if (sphereIt != payload->spheres.end()) {
        return &*sphereIt;
    } else if (cylinderIt != payload->cylinders.end()) {
        return &*cylinderIt;
    } else if (meshIt != payload->meshes.end()) {
        return &*meshIt;
    } else {
        return nullptr;
    }
}

EIFPayload::HoaGroupDesc* EIFPayloadHandler::findHoaGroupAndRegister(
    int hoaGroupId, EIFPayload::HoaSourceDesc* hoaSourcePtr)
{
    std::vector<EIFPayload::HoaGroupDesc>::iterator hoaGroupIt =
        std::find_if(payload->hoaGroups.begin(), payload->hoaGroups.end(),
            [&hoaGroupId](
                const EIFPayload::HoaGroupDesc& hoaGroup) { return hoaGroup.bsId == hoaGroupId; });

    hoaGroupIt->hoaSources.push_back(hoaSourcePtr);

    return &*hoaGroupIt;
}

EIFPayload::ObjectSourceDesc* EIFPayloadHandler::findObjectSource(int objectSourceId)
{
    auto objectSourceIt = std::find_if(payload->objectSources.begin(), payload->objectSources.end(),
        [&objectSourceId](const EIFPayload::ObjectSourceDesc& objectSource) {
            return objectSource.bsId == objectSourceId;
        });

    return &*objectSourceIt;
}

EIFPayload::AudioStreamDesc* EIFPayloadHandler::findStream(int streamId)
{
    auto streamIt = std::find_if(payload->streams.begin(), payload->streams.end(),
        [&streamId](const EIFPayload::AudioStreamDesc& stream) { return stream.bsId == streamId; });

    if (streamIt != payload->streams.end()) {
        return &*streamIt;
    } else {
        return nullptr;
    }
}

EIFPayload::GranularSoundDesc* EIFPayloadHandler::findGranularSound(int granularSoundId)
{
    auto granularSoundIt =
        std::find_if(payload->granularSounds.begin(), payload->granularSounds.end(),
            [&granularSoundId](const EIFPayload::GranularSoundDesc& granularSound) {
                return granularSound.bsId == granularSoundId;
            });

    if (granularSoundIt != payload->granularSounds.end()) {
        return &*granularSoundIt;
    } else {
        return nullptr;
    }
}

EIFPayload::TransformDesc* EIFPayloadHandler::findTransform(int transformId)
{
    auto transformIt = std::find_if(payload->transforms.begin(), payload->transforms.end(),
        [&transformId](
            const EIFPayload::TransformDesc& transform) { return transform.bsId == transformId; });

    return &*transformIt;
}

std::string EIFPayloadHandler::checkIfFileExistsRelatively(
    const std::string& relativeFile, const std::string& id) const
{
    auto fileToFind = basePath + MPEGI_DIRECTORY_SEPARATOR + relativeFile;
#if defined(WIN32)
    std::replace(fileToFind.begin(), fileToFind.end(), '/', '\\');
#endif
    std::ifstream fileStream(fileToFind.c_str());
    EAR_CHECK_REL(!checkFiles || fileStream.good(),
        "File '" + fileToFind + "' not found in entity '" + id + "'.");

    return fileToFind;
}

} // namespace MPEGI
