/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/BinCursor.h>
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EarlyPayload.h>
#include <MPEGI/Bitstream/Globals.h>

#include <MPEGI/Bitstream/HuffmanDecoder.h>

#include <EAR/Exception.h>

#include <algorithm>
#include <cmath>
#include <iostream>
#include <unordered_map>

namespace MPEGI {

class EarlyPayloadHandler : public Bitstream::PayloadHandler
{
public:
    //! Constructor
    EarlyPayloadHandler() = default;

    //! Destructor
    ~EarlyPayloadHandler() = default;

    //! Inherited from Bitstream::PayloadHandler
    virtual std::shared_ptr<const Bitstream::Payload> unpack(
        const Bitstream::RawPayload& payload, const BitstreamConfig& bsConfig) override
    {
        auto earlyPayload = std::shared_ptr<EarlyPayload>(new EarlyPayload());

        // parse culling distances for ER voxel data
        earlyPayload->max_dist_triangle_1 = payload.readInt(8);
        earlyPayload->max_dist_triangle_2 = payload.readInt(8);
        earlyPayload->max_dist_source_1 = payload.readInt(8);
        earlyPayload->max_dist_source_2 = payload.readInt(8);

        // parse voxel grid info
        auto v = payload.getPosition(bsConfig.isSmallScene);
        earlyPayload->voxelGridOrigin[0] = v[0];
        earlyPayload->voxelGridOrigin[1] = v[1];
        earlyPayload->voxelGridOrigin[2] = v[2];

        earlyPayload->voxelGridPitch[0] = payload.getDistance(bsConfig.isSmallScene);
        earlyPayload->voxelGridPitch[1] = payload.getDistance(bsConfig.isSmallScene);
        earlyPayload->voxelGridPitch[2] = payload.getDistance(bsConfig.isSmallScene);

        earlyPayload->voxelGridShape[0] = payload.readUInt(NUM_OF_BITS_FOR_VOXEL_GRID_SHAPE_X);
        earlyPayload->voxelGridShape[1] = payload.readUInt(NUM_OF_BITS_FOR_VOXEL_GRID_SHAPE_Y);
        earlyPayload->voxelGridShape[2] = payload.readUInt(NUM_OF_BITS_FOR_VOXEL_GRID_SHAPE_Z);

        // parse surface data
        unpackSurface(payload, earlyPayload);

        // parse ER voxel data
        unpackVoxelDataEarly(payload, earlyPayload);

        EAR_CHECK_REL(payload.atLastByte(),
            "Unspecified error while parsing EarlyPayload (length missmatch)");

        return std::dynamic_pointer_cast<Bitstream::Payload>(earlyPayload);
    }

private:
    //! parse surface data
    void unpackSurface(
        const Bitstream::RawPayload& payload, std::shared_ptr<EarlyPayload> earlyPayload)
    {
        // flag: has_surface_data
        bool has_surface_data = payload.readBool();
        if (has_surface_data == false) {
            return;
        }

        // get blob data
        auto blob = payload.readBlob();
        auto bitBuffer = BinCursor(blob.get()->data(), blob.get()->size());

        // codebooks for faceList
        HuffmanDecoder decoderLengthFaceIdx;
        HuffmanDecoder decoderFaceIdx;
        decoderLengthFaceIdx.deserializeCodebook(bitBuffer);
        decoderFaceIdx.deserializeCodebook(bitBuffer);

        // codebooks for N0
        HuffmanDecoder decoderAzi;
        HuffmanDecoder decoderEle;
        decoderAzi.deserializeCodebook(bitBuffer);
        decoderEle.deserializeCodebook(bitBuffer);

        // codebook for d
        HuffmanDecoder decoderDist;
        decoderDist.deserializeCodebook(bitBuffer);

        // de-quantization constants
        int max_steps_azi = 1 << NUM_OF_BITS_FOR_NORMAL_AZI;
        int max_steps_ele = 1 << NUM_OF_BITS_FOR_NORMAL_ELE;
        int num_steps_azi = 144 * (max_steps_azi / 144);
        int num_steps_ele = 72 * (max_steps_ele / 72);
        int shift_ele = num_steps_ele / 2;
        float quant2azi = double(2.0 * M_PI) / double(num_steps_azi);
        float quant2ele = double(M_PI) / double(num_steps_ele);
        float quant2dist = 0.001f;

        // distance offset
        int dist_offset = bitBuffer.readSInt(NUM_OF_BITS_FOR_PLANE_DIST);

        // number of surfaces
        int num_surfaces = bitBuffer.readUInt(NUM_OF_BITS_FOR_MAX_FACES);

        for (int surfaceId = 0; surfaceId < num_surfaces; surfaceId++) {
            // de-serialize faceList
            std::vector<unsigned int> faceList;
            unsigned int num_faces = decoderLengthFaceIdx.decodeSingle(bitBuffer);
            for (int f = 0; f < num_faces; f++) {
                unsigned int face = decoderFaceIdx.decodeSingle(bitBuffer);
                faceList.push_back(face);
            }

            // de-serialize N0
            int nAzi = decoderAzi.decodeSingle(bitBuffer);
            int nEle = decoderEle.decodeSingle(bitBuffer);
            float azi = nAzi * quant2azi;
            float ele = (nEle - shift_ele) * quant2ele;
            EAR::Vector3 N0 = sph2cart(azi, ele, 1.0f);

            // de-serialize d
            int nDist = decoderDist.decodeSingle(bitBuffer);
            float d = (nDist + dist_offset) * quant2dist;

            // add surface to surface list
            EAR::Mesh::Surface surf(surfaceId, faceList, { N0, d }, false);
            earlyPayload->surfaceList.push_back(surf);
        }
    }

    //! parse ER voxel data
    void unpackVoxelDataEarly(
        const Bitstream::RawPayload& payload, std::shared_ptr<EarlyPayload> earlyPayload)
    {
        // flag: has_voxel_data
        bool has_voxel_data = payload.readBool();
        if (has_voxel_data == false) {
            return;
        }

        // get blob data
        auto blob = payload.readBlob();
        auto bitBuffer = BinCursor(blob.get()->data(), blob.get()->size());

        // cache for path lists
        std::unordered_map<unsigned int, EarlyDataSet> pathCache;

        // voxel index differences of neighbors
        int delta_x = -earlyPayload->voxelCoordinateToVoxelIndex(VoxelCoord(1, 0, 0));
        int delta_y = -earlyPayload->voxelCoordinateToVoxelIndex(VoxelCoord(0, 1, 0));
        int delta_z = -earlyPayload->voxelCoordinateToVoxelIndex(VoxelCoord(0, 0, 1));
        std::vector<int> delta_list = { 0, delta_x, delta_y, delta_z };

        // read codebooks
        HuffmanDecoder decoderL;
        HuffmanDecoder decoderS;
        HuffmanDecoder decoderNumPaths;
        HuffmanDecoder decoderOrder;
        HuffmanDecoder decoderSurf;
        HuffmanDecoder decoderIndicesRemoved;
        decoderL.deserializeCodebook(bitBuffer);
        decoderS.deserializeCodebook(bitBuffer);
        decoderNumPaths.deserializeCodebook(bitBuffer);
        decoderOrder.deserializeCodebook(bitBuffer);
        decoderSurf.deserializeCodebook(bitBuffer);
        decoderIndicesRemoved.deserializeCodebook(bitBuffer);

        // read number of voxel pairs
        unsigned int num_voxel_pairs = bitBuffer.readUInt(NUM_OF_BITS_FOR_NUM_VOXEL_PAIRS);

        for (unsigned int n = 0; n < num_voxel_pairs; n++) {
            // path list for the current voxel pair
            EarlyDataSet path_list;

            // read listener voxel index
            unsigned int l = decoderL.decodeSingle(bitBuffer);

            // read source voxel index
            unsigned int s = decoderS.decodeSingle(bitBuffer);

            // read mode
            unsigned int mode = bitBuffer.readUInt(2);

            if (mode != 0) {
                // get reference path list
                unsigned int l_ref = l + delta_list[mode];
                path_list = pathCache[l_ref];

                // generate list with removed items in reverse order
                std::vector<int> listIndicesRemoved;
                int val = -1;
                int inc;
                while ((inc = decoderIndicesRemoved.decodeSingle(bitBuffer)) != 0) {
                    val += inc;
                    listIndicesRemoved.insert(listIndicesRemoved.begin(), val);
                }

                // remove items
                for (int pos : listIndicesRemoved) {
                    path_list.erase(path_list.begin() + pos);
                }
            }

            // add items
            int num_paths = decoderNumPaths.decodeSingle(bitBuffer);
            for (int p = 0; p < num_paths; p++) {
                int order = decoderOrder.decodeSingle(bitBuffer);
                std::vector<int> path;
                path.reserve(order);
                for (int o = 0; o < order; o++) {
                    int surf = decoderSurf.decodeSingle(bitBuffer);
                    path.push_back(surf);
                }
                path_list.push_back(std::move(path));
            }

            // sort path list
            std::sort(path_list.begin(), path_list.end(), [](const auto& one, const auto& two) {
                return std::make_tuple(one.size(), std::cref(one))
                    < std::make_tuple(two.size(), std::cref(two));
            });

            // add path list to voxelDataEarly
            earlyPayload->voxelDataEarly[l][s] = path_list;

            // store path list in path cache
            pathCache[l] = std::move(path_list);
        }
    }

    //! convert spherical coordinates into Cartesian coordinates
    EAR::Vector3 sph2cart(float azi, float ele, float dist)
    {
        EAR::Vector3 cart;

        cart[0] = -1.0 * dist * sin(azi) * cos(ele);
        cart[1] = dist * sin(ele);
        cart[2] = -1.0 * dist * cos(azi) * cos(ele);

        return cart;
    }
};

Bitstream::PayloadHandlerAndType makeEarlyPayloadHandler()
{
    return { std::make_shared<EarlyPayloadHandler>(), Bitstream::PLD_ER };
}

} // namespace MPEGI
