/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#include <EAR/Exception.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/GranularPayload.h>

namespace MPEGI {

class GranularPayloadHandler : public Bitstream::PayloadHandler
{
public:
    //! Constructor
    GranularPayloadHandler() = default;

    //! Destructor
    ~GranularPayloadHandler() = default;

    //! Inherited from Bitstream::PayloadHandler
    virtual std::shared_ptr<const Bitstream::Payload> unpack(
        const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig) override

    {
        auto payload = std::shared_ptr<GranularPayload>(new GranularPayload());

        int numberOfDatabases = cursor.getCountOrIndex();
        for (int i = 0; i < numberOfDatabases; i++) {
            const std::string granularDatabaseId = bsConfig.getIdMap().at(cursor.getID());
            int granularEncodingMode = cursor.getCountOrIndex();
            int dimensions = cursor.getCountOrIndex();
            int channels = cursor.getCountOrIndex();
            auto numOfGrains = cursor.readUInt(20);

            // write into granularMap payload header :

            payload->granularMap[granularDatabaseId] =
                std::make_shared<GranularDatabase>(dimensions, numOfGrains, channels);

            //coordinate-scaling:
            EAR::Vector3 coordScaling(0.0, 0.0, 0.0);
            auto qRangeCoordScaling =
                QuantizationRange{ 0.0f, 100.0f, 10, QuantizationScaling::LINEAR };
            for (int dim = 0; dim < dimensions; dim++) {
                coordScaling[dim] = cursor.readQuantFloat(qRangeCoordScaling);
            }

            //smoothing target coordinates:
            EAR::Vector3 smoothingTargetCoordinates(0.0, 0.0, 0.0);
            auto qRangesmoothingTargetCoordinates =
                QuantizationRange{ 0.0f, 1.0f, 10, QuantizationScaling::LINEAR };
            for (int dim = 0; dim < dimensions; dim++) {
                smoothingTargetCoordinates[dim] =
                    cursor.readQuantFloat(qRangesmoothingTargetCoordinates);
            }
            // search radius
            auto qRangeSearchRadius =
                QuantizationRange{ 0.0f, 1.0f, 10, QuantizationScaling::LINEAR };
            float searchRadius = cursor.readQuantFloat(qRangeSearchRadius);
            //grain history length
            int grainHistoryLength = cursor.getCountOrIndex();

            // write into granularMap:
            payload->granularMap[granularDatabaseId]->SetCoordScaling(coordScaling);
            payload->granularMap[granularDatabaseId]->SetSmoothTargetCoord(
                smoothingTargetCoordinates);
            payload->granularMap[granularDatabaseId]->SetSearchRadius(searchRadius);
            payload->granularMap[granularDatabaseId]->SetGrainHistoryLength(grainHistoryLength);

            // number of mixing rules (number of points where overlap windows are defined
            int numMixRule = cursor.getCountOrIndex();
            //grainMixingRule
            std::vector<float> grainMixRule(numMixRule, 0.0);
            auto qRangeGrainMix = QuantizationRange{ 0.0f, 1.0f, 8, QuantizationScaling::LINEAR };
            std::vector<EAR::Vector3> grainMixingRuleCoordinates;
            auto qRange = QuantizationRange{ 0.0f, 1.0f, 16, QuantizationScaling::LINEAR };
            for (int ind = 0; ind < numMixRule; ind++) {
                grainMixRule[ind] = cursor.readQuantFloat(qRangeGrainMix);
                EAR::Vector3 tmp(0.0, 0.0, 0.0);
                for (int dim = 0; dim < dimensions; dim++) {
                    tmp[dim] = cursor.readQuantFloat(qRange);
                }
                grainMixingRuleCoordinates.push_back(tmp);
            }

            //write into granular Map:
            payload->granularMap[granularDatabaseId]->SetGrainMixRule(
                numMixRule, grainMixRule, grainMixingRuleCoordinates);

            //fast grain switch:
            bool fastGrainSwitchThresholdAvailable = cursor.readBool();
            if (fastGrainSwitchThresholdAvailable) {
                //fast grain switch threshold:
                auto qRangeFastGrainSwitchThreshold =
                    QuantizationRange{ 0.0f, 2.0f, 10, QuantizationScaling::LINEAR };
                float fastGrainSwitchThreshold =
                    cursor.readQuantFloat(qRangeFastGrainSwitchThreshold);
                //write into granularMap
                payload->granularMap[granularDatabaseId]->SetFastGrainSwitchThreshold(
                    fastGrainSwitchThreshold);
            }

            // probability exponents and weights :
            auto qRangeProbabilityExponent =
                QuantizationRange{ 0.0f, 10.0f, 8, QuantizationScaling::LINEAR };
            auto qRangeProbabilityWeight =
                QuantizationRange{ 0.0f, 100.0f, 16, QuantizationScaling::LINEAR };

            // for position  and history:
            payload->granularMap[granularDatabaseId]->probabilityExponentPosition =
                cursor.readQuantFloat(qRangeProbabilityExponent);
            payload->granularMap[granularDatabaseId]->probabilityExponentHistory =
                cursor.readQuantFloat(qRangeProbabilityExponent);
            payload->granularMap[granularDatabaseId]->probabilityWeightPosition =
                cursor.readQuantFloat(qRangeProbabilityWeight);
            payload->granularMap[granularDatabaseId]->probabilityWeightHistory =
                cursor.readQuantFloat(qRangeProbabilityWeight);

            // descriptor trend
            bool descrTrendAvailable = cursor.readBool();
            if (descrTrendAvailable) {
                payload->granularMap[granularDatabaseId]->probabilityExponentTrend =
                    cursor.readQuantFloat(qRangeProbabilityExponent);
                payload->granularMap[granularDatabaseId]->probabilityWeightTrend =
                    cursor.readQuantFloat(qRangeProbabilityWeight);
                int descTrendMethod = cursor.getCountOrIndex();
                payload->granularMap[granularDatabaseId]->SetDescTrendMethod(descTrendMethod);
                payload->granularMap[granularDatabaseId]->descrTrendAvailable = true;
            }

            // original recording position within file
            bool originalPositionAvailable = cursor.readBool();
            if (originalPositionAvailable) {
                payload->granularMap[granularDatabaseId]->probabilityExponentOrigPos =
                    cursor.readQuantFloat(qRangeProbabilityExponent);
                payload->granularMap[granularDatabaseId]->probabilityWeightOrigPos =
                    cursor.readQuantFloat(qRangeProbabilityWeight);
                payload->granularMap[granularDatabaseId]->originalPositionAvailable = true;
            }

            // probability Groups:
            bool probabilityGroupsAvailable = cursor.readBool();
            if (probabilityGroupsAvailable) {
                //number of probability groups:
                int numProbGroups = cursor.getCountOrIndex();
                auto qRangeProbabilityVal =
                    QuantizationRange{ 0.0f, 1000.0f, 16, QuantizationScaling::LINEAR };
                std::vector<float> probabilityGroupVal(numProbGroups, 0);
                for (int ind = 0; ind < numProbGroups; ind++) {
                    probabilityGroupVal[ind] = cursor.readQuantFloat(qRangeProbabilityVal);
                }
                //write into granularMap
                payload->granularMap[granularDatabaseId]->SetProbabilityGroup(
                    numProbGroups, probabilityGroupVal);
                payload->granularMap[granularDatabaseId]->probabilityGroupsActive = true;
            }

            // random selection
            bool randomSelection = cursor.readBool();
            payload->granularMap[granularDatabaseId]->SetRandomSelection(randomSelection);

            // interpolation
            bool useInterpolation = cursor.readBool();

            if (useInterpolation) {
                auto qRangeGrainInterpolationDistThreshold =
                    QuantizationRange{ 0.0f, 1.0f, 10, QuantizationScaling::LINEAR };
                float grainInterpolationDistThreshold =
                    cursor.readQuantFloat(qRangeGrainInterpolationDistThreshold);
                bool useInterpolationResampling = cursor.readBool();
                int InterpolationMethod = cursor.getCountOrIndex();
                std::vector<bool> pitchDimension(dimensions, 1);
                for (int i = 0; i < dimensions; i++) {
                    pitchDimension[i] = cursor.readBool();
                }

                payload->granularMap[granularDatabaseId]->SetGrainInterpParams(useInterpolation,
                    grainInterpolationDistThreshold, useInterpolationResampling, pitchDimension,
                    InterpolationMethod);
            } else {
                payload->granularMap[granularDatabaseId]->SetGrainInterpParams(
                    useInterpolation, 0.0, false, { false }, 0);
            }

            //grain overlap amount
            bool grainOverlapAmountAvailable = cursor.readBool();
            if (grainOverlapAmountAvailable) {
                auto qRangeOverlap =
                    QuantizationRange{ 0.0f, 1.0f, 10, QuantizationScaling::LINEAR };
                float grainOverlapAmount = cursor.readQuantFloat(qRangeOverlap);
                payload->granularMap[granularDatabaseId]->SetGrainOverlapAmount(grainOverlapAmount);
            }

            // Unpack each grain
            for (int grainInd = 0; grainInd < numOfGrains; grainInd++) {

                EAR::Vector3 descCoord(0.0, 0.0, 0.0);
                EAR::Vector3 trend(0.0, 0.0, 0.0);
                float orgPos = 0.0f;

                // Read descriptor coordinates
                auto qRange = QuantizationRange{ 0.0f, 1.0f, 16, QuantizationScaling::LINEAR };
                for (int dim = 0; dim < dimensions; dim++) {
                    descCoord[dim] = cursor.readQuantFloat(qRange);
                }

                // Read grain length
                int grainLen = cursor.readUInt(18);

                // new grain created
                Grain newGrain(dimensions, descCoord, grainLen);
                payload->granularMap[granularDatabaseId]->grains.push_back(newGrain);

                // Read descriptor coordinate trend
                if (descrTrendAvailable) {
                    auto trendQRange =
                        QuantizationRange{ -5.0f, 5.0f, 16, QuantizationScaling::LINEAR };
                    for (int dim = 0; dim < dimensions; dim++) {
                        trend[dim] = cursor.readQuantFloat(trendQRange);
                    }
                    payload->granularMap[granularDatabaseId]->grains.back().SetCoordTrend(trend);
                }

                // Read position in the original recording
                if (originalPositionAvailable) {
                    auto posQRange =
                        QuantizationRange{ 0.0f, 20.0f, 16, QuantizationScaling::LINEAR };
                    orgPos = cursor.readQuantFloat(posQRange);
                    payload->granularMap[granularDatabaseId]->grains.back().SetOrgPos(orgPos);
                }

                // Read probability group index

                if (probabilityGroupsAvailable) {
                    int grainProbabilityGroup = cursor.getCountOrIndex();
                    auto qRangePerGrainProbabilityValue =
                        QuantizationRange{ 0.0f, 1000.0f, 16, QuantizationScaling::LINEAR };
                    float perGrainProbabilityValue =
                        cursor.readQuantFloat(qRangePerGrainProbabilityValue);
                    payload->granularMap[granularDatabaseId]->grains.back().SetProbGroup(
                        grainProbabilityGroup);
                    payload->granularMap[granularDatabaseId]->grains.back().SetGrainProbVal(
                        perGrainProbabilityValue);
                }

                // Read samples for this grain
                auto sampleQRange =
                    QuantizationRange{ -1.0f, 1.0f, 16, QuantizationScaling::LINEAR };

                switch (granularEncodingMode) {
                default: {
                    for (int chan = 0; chan < channels; chan++) {
                        std::vector<float> sample_vec1;
                        payload->granularMap[granularDatabaseId]->grains.back().samples.push_back(
                            sample_vec1);
                        for (int sampleInd = 0; sampleInd < grainLen; sampleInd++) {
                            float sample = 0;
                            payload->granularMap[granularDatabaseId]
                                ->grains.back()
                                .samples[chan]
                                .push_back(sample);
                        }
                    }
                } break;
                case 0: //no compression

                {
                    for (int chan = 0; chan < channels; chan++) {
                        std::vector<float> sample_vec;
                        payload->granularMap[granularDatabaseId]->grains.back().samples.push_back(
                            sample_vec);
                        for (int sampleInd = 0; sampleInd < grainLen; sampleInd++) {
                            float sample = cursor.readQuantFloat(sampleQRange);
                            payload->granularMap[granularDatabaseId]
                                ->grains.back()
                                .samples[chan]
                                .push_back(sample);
                        }
                    }
                } break;
                }
            }
        }

        EAR_CHECK_REL(cursor.atLastByte(),
            "Unspecified error while parsing when parsing GranularPayload (too much data)");

        return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
    }
};

Bitstream::PayloadHandlerAndType makeGranularPayloadHandler()
{
    return { std::make_shared<GranularPayloadHandler>(), Bitstream::PLD_GRANULAR };
}

} // namespace MPEGI
