/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/Bitstream.h>

#include <MPEGI/Bitstream/LCERPayload.h>

#include <EAR/Exception.h>

#include <algorithm>

namespace MPEGI {

class LCERPayloadHandler : public Bitstream::PayloadHandler
{
private:
    LCERPayload::LCEREnvironment createLCERIndoorData(
        const Bitstream::RawPayload& cursor, LCERPayload::LCEREnvironmentType type)
    {
        return LCERPayload::LCEREnvironment();
    }

    LCERPayload::LCEREnvironment createLCERPositionData(
        const Bitstream::RawPayload& cursor, LCERPayload::LCEREnvironmentType type)
    {
        return LCERPayload::LCEREnvironment();
    }

    LCERPayload::LCEREnvironment createLCEROutdoorData(const Bitstream::RawPayload& cursor,
        LCERPayload::LCEREnvironmentType type, const BitstreamConfig& bsConfig)
    {
        LCERPayload::LCEREnvironment environment;
        environment.type = type;
        environment.priority = cursor.readInt(8);

        float lcEarlyAmpFac1db = cursor.getGain(true);
        float lcEarlyAmpFac2db = cursor.getGain(true);
        int lcEarlyPreDelay1 = cursor.readInt(16);
        int lcEarlyPreDelay2 = cursor.readInt(16);

        environment.outdoorData.ampFac1 = std::pow(10.f, 0.05f * lcEarlyAmpFac1db);
        environment.outdoorData.ampFac2 = std::pow(10.f, 0.05f * lcEarlyAmpFac2db);
        environment.outdoorData.preDelay1 = (float)lcEarlyPreDelay1;
        environment.outdoorData.preDelay2 = (float)lcEarlyPreDelay2;

        auto v = cursor.getPosition(bsConfig.isSmallScene);
        environment.outdoorData.boundingBox.position.location = { v[0], v[1], v[2] };

        auto o = cursor.getOrientation();
        environment.outdoorData.boundingBox.position.orientation = { o[0], o[1], o[2] };

        auto s = cursor.getPosition(bsConfig.isSmallScene);
        environment.outdoorData.boundingBox.size = { s[0], s[1], s[2] };

        return environment;
    }

public:
    //! Constructor
    LCERPayloadHandler(){};

    //! Destructor
    ~LCERPayloadHandler(){};

    //! Inherited from Bitstream::Payloadhandler
    virtual std::shared_ptr<const Bitstream::Payload> unpack(
        const Bitstream::RawPayload& cursor, const BitstreamConfig& config) override
    {
        auto lcerPayload = std::shared_ptr<LCERPayload>(new LCERPayload());

        int numberOfEnvironments = cursor.readInt(16);
        for (int i = 0; i < numberOfEnvironments; i++) {
            typedef LCERPayload::LCEREnvironmentType EnvType;
            EnvType type = static_cast<EnvType>(cursor.readInt(8));
            if (type != EnvType::Outdoor) {
                std::cout << "unexpected LC ER env type" << std::endl;
                continue;
            }
            auto environment = createLCEROutdoorData(cursor, type, config);
            lcerPayload->envs[""].push_back(environment);
        }
        //order by priority (highest to lowest per acoustic environment) for environment hit test
        for (auto& list : lcerPayload->envs) {
            auto& envVec = list.second;
            std::sort(envVec.begin(), envVec.end(), std::greater<LCERPayload::LCEREnvironment>());
        }

        EAR_CHECK_REL(cursor.atLastByte(),
            "Unspecified error while parsing when parsing LCERPayloadPayload (too much data)");

        return std::dynamic_pointer_cast<Bitstream::Payload>(lcerPayload);
    }
};

Bitstream::PayloadHandlerAndType makeLCERPayloadHandler()
{
    return { std::make_shared<LCERPayloadHandler>(), Bitstream::PLD_LCER };
}

} // namespace MPEGI
