/***************************************************************************
This software module was originally developed jointly by

Nokia Technologies Ltd. and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "rans_interface.hpp"
#include <EAR/Exception.h>
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Bitstream/PortalPayloadHandler.h>
#include <MPEGI/Helpers/Config.h>
#include <cstdint>
#include <iostream>
#include <vector>

using std::int32_t;

namespace MPEGI {

bool arePortalsConnected(const MPEGI::Portal& p1, const MPEGI::Portal& p2)
{
    // check there is exactly one common AE between the two portals.
    const int match0 = std::count(
        p2.portalEnvironments.begin(), p2.portalEnvironments.end(), p1.portalEnvironments[0]);
    const int match1 = std::count(
        p2.portalEnvironments.begin(), p2.portalEnvironments.end(), p1.portalEnvironments[1]);
    return match0 + match1 == 1;
}

//! Constructor
PortalPayloadHandler::PortalPayloadHandler()
    : m_cdfSizes{ 258, 245 }
    , m_offsets{ 0, -128 }
{
    // These are the pre-calculated values of the cumulative distribution function (cdf)
    // that are used by the compression algorithm for the line of sight voxel data.
    // Putting them in an initializer list would format them on one line, so they are here
    // in order to fit horizontally.
    std::vector<std::vector<int32_t>> cdfs{
        { 0, 3782, 7564, 11346, 15128, 17436, 19744, 22052, 24360, 26054, 27748, 29442, 31136,
            32241, 33346, 34451, 35556, 36710, 37864, 39018, 40172, 40736, 41300, 41864, 42428,
            42722, 43016, 43310, 43604, 43898, 44192, 44486, 44780, 45246, 45712, 46178, 46644,
            47012, 47380, 47748, 48116, 48312, 48508, 48704, 48900, 48998, 49096, 49194, 49292,
            49365, 49438, 49511, 49584, 49780, 49976, 50172, 50368, 50417, 50466, 50515, 50564,
            50662, 50760, 50858, 50956, 51103, 51250, 51397, 51544, 51642, 51740, 51838, 51936,
            51960, 51984, 52008, 52032, 52130, 52228, 52326, 52424, 52448, 52472, 52496, 52520,
            52667, 52814, 52961, 53108, 53181, 53254, 53327, 53400, 53473, 53546, 53619, 53692,
            53790, 53888, 53986, 54084, 54206, 54328, 54450, 54572, 54670, 54768, 54866, 54964,
            55013, 55062, 55111, 55160, 55258, 55356, 55454, 55552, 55601, 55650, 55699, 55748,
            55772, 55796, 55820, 55844, 55942, 56040, 56138, 56236, 56260, 56284, 56308, 56332,
            56356, 56380, 56404, 56428, 56452, 56476, 56500, 56524, 56597, 56670, 56743, 56816,
            56840, 56864, 56888, 56912, 56961, 57010, 57059, 57108, 57132, 57156, 57180, 57204,
            57253, 57302, 57351, 57400, 57473, 57546, 57619, 57692, 57716, 57740, 57764, 57788,
            57861, 57934, 58007, 58080, 58104, 58128, 58152, 58176, 58225, 58274, 58323, 58372,
            58421, 58470, 58519, 58568, 58617, 58666, 58715, 58764, 58862, 58960, 59058, 59156,
            59278, 59400, 59522, 59644, 59668, 59692, 59716, 59740, 59764, 59788, 59812, 59836,
            59860, 59884, 59908, 59932, 59956, 59980, 60004, 60028, 60052, 60076, 60100, 60124,
            60148, 60172, 60196, 60220, 60269, 60318, 60367, 60416, 60440, 60464, 60488, 60512,
            60536, 60560, 60584, 60608, 60632, 60656, 60680, 60704, 60728, 60752, 60776, 60800,
            60824, 60848, 60872, 60896, 60920, 60944, 60968, 60992, 61041, 61090, 61139, 61188,
            61319, 61450, 61581, 63496, 65536, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0 },
        { 0, 15, 30, 45, 60, 75, 90, 105, 120, 135, 150, 165, 180, 195, 210, 225, 240, 255, 270,
            340, 410, 480, 550, 643, 736, 829, 899, 969, 1039, 1109, 1202, 1295, 1388, 1458, 1528,
            1598, 1668, 1761, 1854, 1947, 2017, 2087, 2157, 2227, 2297, 2367, 2437, 2507, 2600,
            2693, 2786, 2856, 2926, 2996, 3066, 3159, 3252, 3345, 3415, 3485, 3555, 3625, 3718,
            3811, 3904, 4044, 4184, 4324, 4464, 4557, 4650, 4743, 4813, 4883, 4953, 5023, 5233,
            5443, 5653, 5863, 5956, 6049, 6142, 6212, 6282, 6352, 6422, 6515, 6608, 6701, 6771,
            6841, 6911, 6981, 7168, 7355, 7542, 7682, 7822, 7962, 8102, 8242, 8382, 8522, 8662,
            8755, 8848, 8941, 9292, 9643, 9994, 10345, 10720, 11095, 11470, 12102, 12734, 13366,
            13998, 14841, 15684, 16527, 17370, 18213, 19056, 19899, 20367, 20835, 21303, 21654,
            22005, 22356, 22707, 22917, 23127, 23337, 23547, 23640, 23733, 23826, 23896, 23966,
            24036, 24106, 24199, 24292, 24385, 24525, 24665, 24805, 24945, 25038, 25131, 25224,
            25294, 25364, 25434, 25504, 25644, 25784, 25924, 26064, 26157, 26250, 26343, 26413,
            26483, 26553, 26623, 26716, 26809, 26902, 26972, 27042, 27112, 27182, 27275, 27368,
            27461, 27531, 27601, 27671, 27741, 27834, 27927, 28020, 28090, 28160, 28230, 28300,
            28370, 28440, 28510, 28580, 28673, 28766, 28859, 28929, 28999, 29069, 29139, 29232,
            29325, 29418, 29488, 29558, 29628, 29698, 29791, 29884, 29977, 30047, 30117, 30187,
            30257, 30327, 30397, 30467, 30537, 30630, 30723, 30816, 30886, 30956, 31026, 31096,
            31189, 31282, 31375, 31445, 31515, 31585, 31655, 31748, 31841, 31934, 32004, 32074,
            32144, 32214, 32307, 32400, 32493, 48947, 65536, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
    };
    m_cdfs = cdfs;
    m_ransDecoder = std::make_unique<RansDecoder>();
};

//! Destructor
PortalPayloadHandler::~PortalPayloadHandler(){};

//! Inherited from Bitstream::PayloadHandler
std::shared_ptr<const Bitstream::Payload> PortalPayloadHandler::unpack(
    const Bitstream::RawPayload& cursor, const BitstreamConfig& bsConfig)
{
    auto payload = std::shared_ptr<PortalPayload>(new PortalPayload());
    bool isExplicitPortalMode = false;
    int numberOfPortals = cursor.getCountOrIndex();
    std::map<std::pair<int, int>, float> portal2portalFactors;

    if (numberOfPortals > 0) {
        isExplicitPortalMode = cursor.readBool();
    }

    for (int i = 0; i < numberOfPortals; i++) {
        int portalId = cursor.getID();
        const std::string stringID = bsConfig.getIdMap().at(portalId);
        bool hasExtent = cursor.readBool();
        std::optional<std::string> portalExtentStringId;
        if (hasExtent) {
            const int portalExtentId = cursor.getID();
            portalExtentStringId = bsConfig.getIdMap().at(portalExtentId);
        }
        if (isExplicitPortalMode) {
            bool isConnectedToAE = cursor.readBool();
            bool isConnectedToSecondAE = false;
            std::string portalEnvironment1;
            std::string portalEnvironment2;
            if (isConnectedToAE) {
                portalEnvironment1 = bsConfig.getIdMap().at(cursor.getID());
                isConnectedToSecondAE = cursor.readBool();
                if (isConnectedToSecondAE)
                    portalEnvironment2 = bsConfig.getIdMap().at(cursor.getID());
            }

            bool hasMaterial = cursor.readBool();
            std::string materialId;
            if (hasMaterial) {
                auto matBsId = cursor.getID();
                materialId = bsConfig.getIdMap().at(matBsId);
            }
            auto normVec = cursor.getPosition(true);
            EAR::Vector3 normalVector = EAR::Vector3(normVec[0], normVec[1], normVec[2]);

            bool explicitPortalProperties = cursor.readBool();
            EAR::Vector3 portalCentrePos;
            std::optional<EAR::Vector3> equivalentPortalSize;
            if (explicitPortalProperties) {
                auto v = cursor.getPosition(bsConfig.isSmallScene);
                portalCentrePos = EAR::Vector3(v[0], v[1], v[2]);
                EAR::Vector3 size;
                size[0] = cursor.getDistance(bsConfig.isSmallScene);
                size[1] = cursor.getDistance(bsConfig.isSmallScene);
                size[2] = cursor.getDistance(bsConfig.isSmallScene);
                equivalentPortalSize = size;
            }

            auto portal = std::make_unique<Portal>();
            portal->portalId = portalId;
            portal->portalStringId = stringID;
            portal->extentId = portalExtentStringId;
            portal->sourceId = ""; // n.a.
            portal->portalEnvironments[0] = portalEnvironment1;
            portal->portalEnvironments[1] = portalEnvironment2;
            portal->type = PortalType::ReverbOnly;
            portal->position = portalCentrePos;
            portal->normalVector = normalVector;
            portal->materialId = materialId;
            portal->equivalentPortalSize = equivalentPortalSize;
            payload->portals.push_back(std::move(portal));
            payload->portalMap[portalEnvironment1].insert(portalId);
            payload->portalMap[portalEnvironment2].insert(portalId);
            bool hasPortal2PortalFactors = cursor.readBool();
            if (hasPortal2PortalFactors) {
                for (int j = 0; j < i; j++) {
                    if (arePortalsConnected(*payload->portals.at(i), *payload->portals.at(j))) {
                        bool hasAcousticLink = cursor.readBool();
                        if (hasAcousticLink) {
                            float portalFactor1 =
                                cursor.readCodeWord("portal2PortalFactorCode").value[0];
                            float portalFactor2 =
                                cursor.readCodeWord("portal2PortalFactorCode").value[0];
                            portalFactor1 = std::pow(10.0f, (portalFactor1 / 10.0f));
                            portalFactor2 = std::pow(10.0f, (portalFactor2 / 10.0f));
                            portal2portalFactors.emplace(std::pair(payload->portals.at(i)->portalId,
                                                             payload->portals.at(j)->portalId),
                                portalFactor1);
                            portal2portalFactors.emplace(std::pair(payload->portals.at(j)->portalId,
                                                             payload->portals.at(i)->portalId),
                                portalFactor2);
                        } else {
                            portal2portalFactors.emplace(std::pair(payload->portals.at(i)->portalId,
                                                             payload->portals.at(j)->portalId),
                                0.0f);
                            portal2portalFactors.emplace(std::pair(payload->portals.at(j)->portalId,
                                                             payload->portals.at(i)->portalId),
                                0.0f);
                        }
                    }
                }
            }
        } else {
            int portalSourceId = cursor.getID();
            const std::string portalSourceStringId = bsConfig.getIdMap().at(portalSourceId);
            const std::string portalSourceStringDefaultAEId = "";
            auto portal = std::make_unique<Portal>();
            portal->portalId = portalId;
            portal->portalStringId = stringID;
            portal->extentId = portalExtentStringId;
            portal->sourceId = portalSourceStringId;

            portal->portalEnvironments = { "", "" }; // n.a.
            portal->type = PortalType::ReverbOnly; // n.a.
            portal->position = { 0, 0, 0 }; // n.a.
            portal->normalVector = { 1, 1, 1 }; // n.a.
            portal->materialId = ""; // n.a.

            payload->portals.push_back(std::move(portal));
            payload->portalMap[portalSourceStringId].insert(portalId);
            // insert portal id also for default AE
            payload->portalMap[portalSourceStringDefaultAEId].insert(portalId);
        }
    }
    payload->portal2portalFactors = portal2portalFactors;
    payload->isExplicitPortalMode = isExplicitPortalMode;
    if (isExplicitPortalMode) {
        bool hasAELosVoxelData = cursor.readBool();
        if (hasAELosVoxelData) {
            auto numAEs = cursor.getCountOrIndex();
            for (unsigned int iAE = 0; iAE < numAEs; iAE++) {
                unsigned int aeBsId = cursor.getCountOrIndex();
                auto losArea = readLineOfSightArea(cursor, bsConfig.isSmallScene);
                payload->aeLosAreas[aeBsId] = std::move(losArea);
            }
        }
        bool hasOutsideLosVoxelData = cursor.readBool();
        if (hasOutsideLosVoxelData) {
            auto numOutsideAreas = cursor.getCountOrIndex();
            for (unsigned int iOutArea = 0; iOutArea < numOutsideAreas; iOutArea++) {
                auto losArea = readLineOfSightArea(cursor, bsConfig.isSmallScene);
                payload->outsideLosAreas.push_back(std::move(losArea));
            }
        }
    }

    EAR_CHECK_REL(
        cursor.atLastByte(), "Unspecified error while parsing when parsing PortalPayload");

    return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
}

Bitstream::PayloadHandlerAndType makePortalPayloadHandler()
{
    return { std::make_shared<PortalPayloadHandler>(), Bitstream::PLD_PORTAL };
}

std::unique_ptr<LineOfSightArea> PortalPayloadHandler::readLineOfSightArea(
    const Bitstream::RawPayload& cursor, bool isSmallScene)
{
    auto losArea = std::make_unique<LineOfSightArea>();
    losArea->nVoxels[0] = cursor.getCountOrIndex();
    losArea->nVoxels[1] = cursor.getCountOrIndex();
    losArea->nVoxels[2] = cursor.getCountOrIndex();
    losArea->losAreaCornerP1 = cursor.getPosition(isSmallScene);
    losArea->losAreaCornerP2 = cursor.getPosition(isSmallScene);
    losArea->voxelSize = cursor.getCountOrIndex();
    auto aePortalCount = cursor.getCountOrIndex();
    auto portalBsIds = std::make_unique<std::vector<unsigned int>>();
    for (unsigned int iAEPortal = 0; iAEPortal < aePortalCount; iAEPortal++) {
        unsigned int portalBsId = cursor.getID();
        portalBsIds->push_back(portalBsId);
        unsigned int nLosBitMasks = cursor.getCountOrIndex();
        unsigned int iVoxel = 0;
        for (unsigned int iMask = 0; iMask < nLosBitMasks; iMask++) {
            unsigned int nBitmaskBytes = cursor.getCountOrIndex();
            auto voxels = decodeMask(nBitmaskBytes, cursor);
            for (auto voxel : voxels) {
                if (voxel) {
                    losArea->voxels[portalBsIds->at(iAEPortal)].insert(iVoxel);
                }
                iVoxel++;
            }
        }
    }
    return losArea;
}

int32_t PortalPayloadHandler::getSymbol(int32_t index)
{
    std::vector<int32_t> indexes;
    indexes.push_back(index);
    bool oorBypass = true;
    // Get the next symbol from the stream using the indicated cumulative distribution function
    auto decoded = m_ransDecoder->decode_stream(indexes, m_cdfs, m_cdfSizes, m_offsets, oorBypass);
    return decoded.at(0);
}

std::vector<bool> PortalPayloadHandler::decodeMask(
    int32_t length, const MPEGI::Bitstream::RawPayload& cursor)
{
    std::vector<bool> bitstream;
    std::size_t lengthWithoutPreamble = length - 4;
    int32_t width = cursor.readUInt(16);
    int32_t height = cursor.readUInt(16);
    std::vector<std::uint8_t> blob;
    for (unsigned int i = 0; i < lengthWithoutPreamble; i++) {
        blob.push_back(cursor.readUInt(8));
    }

    // Rens library requires input as an std::string which necessitates a bit of casting
    std::string dataAsString(reinterpret_cast<char*>(blob.data()), lengthWithoutPreamble);

    m_ransDecoder->set_stream(dataAsString);

    std::vector<std::vector<int32_t>> rec;
    rec.assign(height, std::vector<int32_t>(width, 0));

    std::vector<int32_t> prevRow(width, 0);
    unsigned int rowIdx{ 0 };
    unsigned int prevStart{ 0 };
    unsigned int prevEnd{ 0 };
    unsigned int vIdx{ 0 };
    // Decoding will go on as long as there is data in the input
    for (;;) {
        auto rowLl = getSymbol(0);

        for (unsigned int range = 0; range < rowLl; range++) {
            rec.at(rowIdx) = prevRow;
            rowIdx++;
        }

        if (rowIdx >= height) {
            break;
        }

        std::vector<int32_t> row(width, 0);
        int32_t startIdx = prevStart + getSymbol(1);
        prevStart = startIdx;
        int32_t segLen = getSymbol(0);
        int32_t endIdx = startIdx + segLen;

        int32_t pixIdx = startIdx;
        for (unsigned int idx = 0; idx < startIdx; idx++) {
            row[idx] = prevRow.at(idx);
        }
        int32_t prevV = prevRow.at(startIdx);

        while (pixIdx < endIdx) {
            int32_t ll = getSymbol(0);
            int32_t untilIdx = pixIdx + ll;
            for (unsigned int idx = pixIdx; idx < untilIdx; idx++) {
                row.at(idx) = 1 - prevV;
            }
            prevV = 1 - prevV;
            pixIdx += ll;
        }
        for (unsigned int idx = endIdx; idx < width; idx++) {
            row.at(idx) = prevRow.at(idx);
        }

        rec.at(rowIdx) = row;
        rowIdx += 1;
        if (rowIdx >= height) {
            break;
        }
        prevRow = row;
    }

    // Each byte represents a single boolean value. Flatten the result to a single vector.
    for (const auto& row : rec) {
        for (int32_t byte : row) {
            bitstream.push_back(byte);
        }
    }

    return bitstream;
}

} // namespace MPEGI
