/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include <MPEGI/Bitstream/RasterMapPayloadHandler.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/HuffmanDecoder.h>
#include <MPEGI/Bitstream/RasterMapPayload.h>

#include <EAR/Exception.h>

#include <algorithm>
#include <iostream>

namespace MPEGI {

void RasterMapPayloadHandler::unpackLayer(const Bitstream::RawPayload& bs,
    WaypointMatrix& waypointMatrix, int layer, std::vector<RasterMapCoord>& availableCoords)
{
    // Generic Codebooks
    HuffmanDecoder decoderVcX;
    HuffmanDecoder decoderVcY;
    HuffmanDecoder decoderWpX;
    HuffmanDecoder decoderWpY;
    decoderVcX.deserializeCodebook(bs);
    decoderVcY.deserializeCodebook(bs);
    decoderWpX.deserializeCodebook(bs);
    decoderWpY.deserializeCodebook(bs);

    int vcX = 0;
    int vcY = -1;
    int numRasterMapVoxels = bs.getID();
    availableCoords.reserve(availableCoords.size() + numRasterMapVoxels);
    for (int v = 0; v < numRasterMapVoxels; v++) {
        vcY += 1;
        bool hasVoxelCoordY = bs.readBool();
        if (hasVoxelCoordY) {
            vcY = decoderVcY.decodeSingle(bs);

            vcX += 1;
            bool hasVoxelCoordX = bs.readBool();
            if (hasVoxelCoordX) {
                vcX = decoderVcX.decodeSingle(bs);
            }
        }

        int mapChange, wpX, wpY, wpL;
        int refMode = bs.readUInt(2);
        switch (refMode) {
        case 0:
            mapChange = bs.readUInt(1);
            wpX = decoderWpX.decodeSingle(bs);
            wpY = decoderWpY.decodeSingle(bs);
            wpL = layer - mapChange;
            waypointMatrix.at(vcX, vcY, layer) = { wpX, wpY, wpL };
            break;
        case 1:
            waypointMatrix.at(vcX, vcY, layer) = waypointMatrix.at(vcX - 1, vcY, layer);
            break;
        case 2:
            waypointMatrix.at(vcX, vcY, layer) = waypointMatrix.at(vcX, vcY - 1, layer);
            break;
        case 3:
            waypointMatrix.at(vcX, vcY, layer) = waypointMatrix.at(vcX, vcY, layer - 1);
            break;
        }

        availableCoords.push_back({ vcX, vcY, layer });
    }
}

void RasterMapPayloadHandler::unpackSubscene(
    const Bitstream::RawPayload& bs, const BitstreamConfig& config, RasterMapPayload& payload)
{
    SubsceneData subscene;

    bool hasVoxSceneDiffractionRasterMap = bs.readBool();
    if (hasVoxSceneDiffractionRasterMap) {
        int rmWidth = bs.getCountOrIndex();
        int rmHeight = bs.getCountOrIndex();

        int numRasterMapSources = bs.getCountOrIndex();
        subscene.waypointMatrix.reserve(numRasterMapSources);
        for (int s = 0; s < numRasterMapSources; s++) {
            int rasterMapSourceIdLen = bs.getCountOrIndex();
            std::vector<int> rasterMapSourceId;
            rasterMapSourceId.reserve(rasterMapSourceIdLen);
            for (int i = 0; i < rasterMapSourceIdLen; i++) {
                rasterMapSourceId.push_back(bs.getID());
            }

            int rmLayers = bs.readUInt(3);

            WaypointMatrix waypointMatrix(rmWidth, rmHeight, rmLayers);
            std::vector<RasterMapCoord> availableCoords;
            for (int l = 0; l < rmLayers; l++) {
                unpackLayer(bs, waypointMatrix, l, availableCoords);
            }
            subscene.availableCoords.insert({ rasterMapSourceId[0], std::move(availableCoords) });

            subscene.waypointMatrix.push_back(std::move(waypointMatrix));
            for (int src_id : rasterMapSourceId) {
                subscene.source2entry.insert({ src_id, s });
            }
        }
    }

    payload.subsceneData.push_back(std::move(subscene));
}

//! Inherited from Bitstream::PayloadHandler
std::shared_ptr<const Bitstream::Payload> RasterMapPayloadHandler::unpack(
    const Bitstream::RawPayload& bs, const BitstreamConfig& config)
{
    auto payload = std::shared_ptr<RasterMapPayload>(new RasterMapPayload());

    int numSubscenes = bs.getCountOrIndex(); // num_rastermap_subscenes
    payload->subsceneData.reserve(numSubscenes);
    for (int n = 0; n < numSubscenes; n++) {
        unpackSubscene(bs, config, *payload);
    }

    EAR_CHECK_REL(
        bs.atLastByte(), "Unspecified error while parsing RasterMapPayload (too much data)");

    return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
}

Bitstream::PayloadHandlerAndType makeRasterMapPayloadHandler()
{
    return { std::make_shared<RasterMapPayloadHandler>(), Bitstream::PLD_RASTERMAP };
}

} // namespace MPEGI
