/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Bitstream/Globals.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Bitstream/ReverbPayloadHandler.h>

#include <EAR/Exception.h>
#include <MPEGI/Helpers/Config.h>
#include <iostream>
#include <math.h>

#include "DefaultAEPresets.h"

namespace MPEGI {

Bitstream::PayloadHandlerAndType makeReverbPayloadHandler()
{
    return { std::make_shared<ReverbPayloadHandler>(), Bitstream::PLD_REVERB };
}

enum class fgdMethodT : unsigned int
{
    Individual_Frequencies = 0u,
    Start_Hop_Amount,
    Default_Banding,
    Reserved
};

std::shared_ptr<const Bitstream::Payload> ReverbPayloadHandler::unpack(
    const Bitstream::RawPayload& rawPayload, const BitstreamConfig& config)
{
    auto payload = std::shared_ptr<ReverbPayload>(new ReverbPayload());

    // Control params
    const bool isLongRangeMode = false;
    const bool isSmallScene = true;
    const bool isHiPrecGain = true;

    unsigned int revNrElements = 0;

    unsigned numDelayLinesFromConfig = Config::getIntDef("Reverb:numDelayLines", EAR::Variant(0));

    // Read Acoustic Environment
    const bool revPresent = rawPayload.readBool();
    if (revPresent) {

        payload->freqGridData = config.GetFreqGrid();
        revNrElements = rawPayload.getCountOrIndex();

        std::vector<float> revRT60;
        std::vector<float> revDSR;

        for (int e = 0; e < revNrElements; ++e) {

            const unsigned int revAcEnvID = rawPayload.getID();
            const unsigned int revFreqGridIdx = rawPayload.getCountOrIndex();
            const float revPredelay = roundf(rawPayload.getDuration(isLongRangeMode) * 1000) / 1000;
            const bool revDelayLineCountPresent = rawPayload.readBool();
            unsigned revDelayLineCount = numDelayLinesFromConfig;
            if (revDelayLineCountPresent) {
                auto decodedDelayLineCount = rawPayload.readCodeWord("delayLineCountCode").value[0];
                if (!revDelayLineCount) {
                    revDelayLineCount = decodedDelayLineCount;
                }
            }
            if (!revDelayLineCount) {
                revDelayLineCount = REVERB_DEFAULT_NUMBER_OF_DELAY_LINE_LOOPS;
            }
            const auto& freqGridData = payload->freqGridData;
            const size_t sz = freqGridData[revFreqGridIdx]->nrBands;

            revRT60.resize(sz);
            for (int b = 0; b < sz; ++b) {
                revRT60[b] = roundf(rawPayload.getDuration(isLongRangeMode) * 1000) / 1000;
            }

            revDSR.resize(sz);
            for (int b = 0; b < sz; ++b) {
                revDSR[b] = pow(10, ((rawPayload.readCodeWord("dsrCode").value[0]) / 10.0));
            }

            payload->reverbParameters[revAcEnvID] = std::make_shared<ReverbParameters>(
                revAcEnvID, revFreqGridIdx, revPredelay, revDelayLineCount, revRT60, revDSR);
        }

        const bool revRenderControlPresent = rawPayload.readBool();
        if (revRenderControlPresent) {
            const float revDistanceGainDropDb = rawPayload.getGain(isHiPrecGain);
            payload->reverbDistanceGainDropDb = revDistanceGainDropDb;

            const float revMinDistance = rawPayload.getDistance(isSmallScene);
            payload->reverbMinimumDistance = revMinDistance;
        }

        const bool revMultiRoomConfigPresent = rawPayload.readBool();
        if (revMultiRoomConfigPresent) {
            const bool revFadeEnabled = rawPayload.readBool();
            payload->fadeInEnabled = revFadeEnabled;
            if (revFadeEnabled) {
                const float revFadeDistance = rawPayload.getDistance(isSmallScene);
                payload->reverbFadeInDistance = revFadeDistance;
            }
            const bool revHeadtrackingEnabled = rawPayload.readBool();
            payload->headTrackingEnabled = revHeadtrackingEnabled;
        }

        const bool revARConfigPresent = rawPayload.readBool();
        if (revARConfigPresent) {
            const float revLsdfTheoreticalRDRtoActualRDRdB = rawPayload.getGain(isHiPrecGain);
            payload->revLsdfTheoreticalRDRtoActualRDRdB = revLsdfTheoreticalRDRtoActualRDRdB;

            const float revLsdfRDRtoActualRDRdB = rawPayload.getGain(isHiPrecGain);
            payload->revLsdfRDRtoActualRDRdB = revLsdfRDRtoActualRDRdB;

            const bool revOverrideLSDF_RDR = rawPayload.readBool();
            payload->revOverrideLSDF_RDR = revOverrideLSDF_RDR;
        }
    }

    payload->revDefaultAEPresent = rawPayload.readBool();
    if (payload->revDefaultAEPresent) {

        const bool revHasCustomDefaultAE = rawPayload.readBool();
        if (revHasCustomDefaultAE) {

            std::vector<float> revDefaultAERT60;
            std::vector<float> revDefaultAEDSR;

            const unsigned int revDefaultAEFreqGridIdx = rawPayload.getCountOrIndex();
            float revDefaultAEPredelay =
                roundf(rawPayload.getDuration(isLongRangeMode) * 1000) / 1000;
            float predelayFromConfig = Config::getFloatDef(
                "Reverb:Default_Acoustic_Environment_Predelay_Override", EAR::Variant(0.0f));
            if (predelayFromConfig != 0.0f) {
                revDefaultAEPredelay = predelayFromConfig;
            }
            const bool revDelayLineCountPresent = rawPayload.readBool();
            unsigned revDelayLineCount = numDelayLinesFromConfig;
            if (revDelayLineCountPresent) {
                auto decodedDelayLineCount = rawPayload.readCodeWord("delayLineCountCode").value[0];
                if (!revDelayLineCount) {
                    revDelayLineCount = decodedDelayLineCount;
                }
            }
            if (!revDelayLineCount) {
                revDelayLineCount = REVERB_DEFAULT_NUMBER_OF_DELAY_LINE_LOOPS;
            }

            payload->revDefaultAEDelayRatio = rawPayload.readFloat();

            payload->revDefaultAEDelayMinMs = rawPayload.readFloat();

            const auto& freqGridData = payload->freqGridData;
            const size_t sz = freqGridData[revDefaultAEFreqGridIdx]->nrBands;

            revDefaultAERT60.resize(sz);
            revDefaultAEDSR.resize(sz);

            for (int b = 0; b < sz; ++b) {
                revDefaultAERT60[b] = roundf(rawPayload.getDuration(isLongRangeMode) * 1000) / 1000;
            }
            for (int b = 0; b < sz; ++b) {
                revDefaultAEDSR[b] =
                    pow(10, ((rawPayload.readCodeWord("dsrCode").value[0]) / 10.0));
            }
            payload->revDefaultAEDistanceGainDropDb = rawPayload.getGain(isHiPrecGain);
            payload->revDefaultAEMinimumDistance = rawPayload.getDistance(isSmallScene);

            payload->reverbParameters[revNrElements] =
                std::make_shared<ReverbParameters>(revNrElements, revDefaultAEFreqGridIdx,
                    revDefaultAEPredelay, revDelayLineCount, revDefaultAERT60, revDefaultAEDSR);

        } else {
            int revDefaultAEPreset = rawPayload.readUInt(6);
            int id = revDefaultAEPreset;
            // idFromConfig is used only if there is a defaultAE preset ID in the bitstream
            int idFromConfig = Config::getIntDef(
                "Reverb:Default_Acoustic_Environment_Preset_Override", EAR::Variant(-1));
            if (idFromConfig != -1) {
                id = idFromConfig;
            }
            const auto& preset = DEFAULT_ACOUSTIC_ENVIRONMENT_PRESETS[id];
            float revDefaultAEPredelay = preset.predelay;
            float predelayFromConfig = Config::getFloatDef(
                "Reverb:Default_Acoustic_Environment_Predelay_Override", EAR::Variant(0.0f));
            if (predelayFromConfig != 0.0f) {
                revDefaultAEPredelay = predelayFromConfig;
            }
            auto revDefaultAENumDelayLines = preset.numDelayLines;
            payload->revDefaultAEDelayMinMs = preset.delayMinMs;
            std::shared_ptr<FreqGrid> data =
                std::make_shared<FreqGrid>(preset.numBands, preset.centerFreq);
            payload->freqGridData.push_back(std::move(data));

            auto revDefaultAERT60 = preset.rt60;
            auto revDefaultAEDSR = preset.dsr;

            const bool revHasDefaultAEGain = rawPayload.readBool();
            if (revHasDefaultAEGain) {
                payload->revDefaultAEGain = pow(10.0f, (rawPayload.getGain(isHiPrecGain) / 20));
            }
            const bool revHasDefaultAEDelayRatio = rawPayload.readBool();
            if (revHasDefaultAEDelayRatio) {
                payload->revDefaultAEDelayRatio = rawPayload.readFloat();
            }
            const bool revHasDefaultAEDelayMinMs = rawPayload.readBool();
            if (revHasDefaultAEDelayMinMs) {
                payload->revDefaultAEDelayMinMs = rawPayload.readFloat();
            }
            const bool revHasDefaultAEDistanceGainDrop = rawPayload.readBool();
            if (revHasDefaultAEDistanceGainDrop) {
                payload->revDefaultAEDistanceGainDropDb = rawPayload.getGain(isHiPrecGain);
            }
            const bool revHasDefaultAEMinDist = rawPayload.readBool();
            if (revHasDefaultAEMinDist) {
                payload->revDefaultAEMinimumDistance = rawPayload.getDistance(isSmallScene);
            }

            const unsigned int revDefaultAEFreqGridIdx = payload->freqGridData.size() - 1;
            payload->reverbParameters[revNrElements] = std::make_shared<ReverbParameters>(
                revNrElements, revDefaultAEFreqGridIdx, revDefaultAEPredelay,
                revDefaultAENumDelayLines, revDefaultAERT60, revDefaultAEDSR);
        }
    }

    return std::dynamic_pointer_cast<Bitstream::Payload>(payload);
}

} // namespace MPEGI
