/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "MPEGI/Bitstream/VoxDataPayload.h"
#include "MPEGI/Helpers/Config.h"
#include <numeric>

//--------------------------------------------------------------------------------------------------
namespace {
//--------------------------------------------------------------------------------------------------
using MPEGI::Voxel::voxel_index_val_t;

//--------------------------------------------------------------------------------------------------
} // namespace
//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
VoxDataPayload::VoxDataPayload(Bitstream::RawPayload const& payload, const BitstreamConfig& config)
{
    auto const num_subscenes = 1
        + payload.readEscapedUInt(
            n_bits_num_subscenes_esc1, n_bits_num_subscenes_esc2, n_bits_num_subscenes_esc3);

    for (auto s = 0U; s < num_subscenes; ++s) {
        subscenes.emplace_back(payload, config);
    }

    set_global_parameters(payload);
    EAR_CHECK(payload.isConsumed(), "VoxDataPayload: Not all bits consumed");
}

void VoxDataPayload::set_global_parameters(Bitstream::RawPayload const& payload) const
{
    // Global configuration - Early reflection tuning parameters
    auto const has_voxer_num_origins = payload.readBool();
    if (has_voxer_num_origins) {
        Config::setDef("VoxelEarlyReflectionNumPoints",
            static_cast<int>(payload.readUInt(n_bits_vox_er_num_origins) + 2));
    }

    auto const has_voxer_num_clusters = payload.readBool();
    if (has_voxer_num_clusters) {
        Config::setDef("VoxelEarlyReflectionMaxNumReflections",
            static_cast<int>(payload.readUInt(n_bits_vox_er_num_clusters) + 2));
    }

    auto const has_voxer_indoor_outdoor_tuning = payload.readBool();
    if (has_voxer_indoor_outdoor_tuning) {
        auto get_gain = [](Bitstream::RawPayload const& payload) -> float {
            return std::pow(10.f, payload.getGain(true) * .05f);
        };

        Config::setDef("VoxelEarlyReflectionGainIndoor", get_gain(payload));
        Config::setDef("VoxelEarlyReflectionGainOutdoor", get_gain(payload));
    }

    Config::set("VoxelEarlyReflectionSectorDownmixEnabled", payload.readBool());
}

//--------------------------------------------------------------------------------------------------
namespace Voxel {
namespace Payload {
//--------------------------------------------------------------------------------------------------
std::vector<block<3>> subscene::readVoxSceneElements(
    Bitstream::RawPayload const& payload, Voxel::voxel_index_t const& res)
{
    auto const num_block_elements = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);

    std::vector<block<3>> blocks;
    blocks.reserve(num_block_elements);

    auto const n_bits_packed_coord = calc_bit_length(res);

    for (int el = 0; el < num_block_elements; ++el) {
        blocks.emplace_back(payload, res, n_bits_packed_coord);
    }

    return blocks;
}

//--------------------------------------------------------------------------------------------------
subscene::subscene(Bitstream::RawPayload const& payload, const BitstreamConfig& config)
{
    readVoxSceneHeader(payload, config);
    readVoxSceneElements(payload);
    readVoxSceneMaterials(payload, config);
    readVoxSceneInstructionsForOcclusion(payload);
    readVoxSceneInstructionsForDiffraction(payload);
    readVoxSceneUpdates(payload, config);
}

//--------------------------------------------------------------------------------------------------
void subscene::readVoxSceneHeader(
    Bitstream::RawPayload const& payload, const BitstreamConfig& config)
{
    scene.p0 = payload.getPosition(config.isSmallScene);
    scene.p1 = payload.getPosition(config.isSmallScene);

    for (auto& xyz : scene.dim) {
        xyz = payload.readUInt(n_bits_grid_resolution);
        if (!xyz) {
            throw std::logic_error("Scene-dimension must not be 0");
        }
    }
}

//--------------------------------------------------------------------------------------------------
void subscene::readVoxSceneElements(Bitstream::RawPayload const& payload)
{
    this->blocks = readVoxSceneElements(payload, scene.dim);
}

//--------------------------------------------------------------------------------------------------
void subscene::readVoxSceneMaterials(
    Bitstream::RawPayload const& payload, const BitstreamConfig& config)
{
    auto const num_materials = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
    materials.reserve(num_materials);

    for (auto i = 0; i < num_materials; ++i) {
        materials.emplace_back(payload, config, i + 1);
    }
}

//--------------------------------------------------------------------------------------------------
void subscene::readVoxSceneInstructionsForOcclusion(Bitstream::RawPayload const& payload)
{
    data_occlusion = std::make_shared<occlusion_data>(payload, materials);
}

//--------------------------------------------------------------------------------------------------
void subscene::readVoxSceneInstructionsForDiffraction(Bitstream::RawPayload const& payload)
{
    dmap_index_t const res{ scene.dim[0], scene.dim[1] };
    data_diffraction = std::make_shared<diffraction_data>(payload, materials, res);
}

//--------------------------------------------------------------------------------------------------
void subscene::readVoxSceneUpdates(
    Bitstream::RawPayload const& payload, const BitstreamConfig& config)
{
    auto const num_scene_updates = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
    scene_updates.reserve(num_scene_updates);

    for (auto i = 0U; i < num_scene_updates; ++i) {
        scene_updates.emplace_back(payload, config, materials, scene.dim);
    }
}

//--------------------------------------------------------------------------------------------------
scene_update::scene_update(Bitstream::RawPayload const& payload, const BitstreamConfig& config,
    std::vector<scene_material> const& materials, Voxel::voxel_index_t const& dim)
{
    id = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);

    auto const has_reference_point = payload.readUInt(n_bits_has_reference_point);
    if (has_reference_point) {
        p0 = payload.getPosition(config.isSmallScene);
    }

    auto const num_sequences = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
    sequence.reserve(num_sequences);

    for (auto s = 0U; s < num_sequences; ++s) {
        auto const duration = payload.readUInt(n_bits_update_duration);
        auto blocks = subscene::readVoxSceneElements(payload, dim);

        auto const occlusion_update = std::make_shared<occlusion_data>(payload, materials);

        dmap_index_t const res_dmap{ dim[0], dim[1] };
        auto const diffraction_update =
            std::make_shared<diffraction_data>(payload, materials, res_dmap);

        sequence.emplace_back(duration, std::move(blocks), occlusion_update, diffraction_update);
    }
}

//--------------------------------------------------------------------------------------------------
occlusion_data::occlusion_data(
    Bitstream::RawPayload const& payload, std::vector<scene_material> const& materials)
{
    read_instructions(payload, materials);
    read_source_material_dependency(payload);
}

//--------------------------------------------------------------------------------------------------
void occlusion_data::read_instructions(
    Bitstream::RawPayload const& payload, std::vector<scene_material> const& materials)
{
    instructions_occlusion.reserve(materials.size() + 1);

    // Add dummy instruction for sound-medium (mat:0) to avoid excessive checks later
    instructions_occlusion.emplace_back(0);

    // Fill with default values
    for (auto const& mat : materials) {
        instructions_occlusion.emplace_back(mat.id);
    }

    auto const has_instructions_occlusion = payload.readBool();
    if (has_instructions_occlusion) {
        // Skip instruction for sound-medium
        auto instr_it = std::begin(instructions_occlusion) + 1;
        while (instr_it != std::cend(instructions_occlusion)) {
            auto const has_recovery_factor = payload.readBool();
            if (has_recovery_factor) {
                instr_it->recovery = payload.readFloat();
            }
            ++instr_it;
        }
    }
}

//--------------------------------------------------------------------------------------------------
void occlusion_data::read_source_material_dependency(Bitstream::RawPayload const& payload)
{
    auto const has_source_material_mapping = payload.readBool();
    if (has_source_material_mapping) {
        auto const num_entries = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
        source_material_dependency.reserve(num_entries);

        for (auto i = 0U; i < num_entries; ++i) {
            source_material_dependency.emplace_back(payload);
        }
    }
}

//--------------------------------------------------------------------------------------------------
diffraction_data::diffraction_data(Bitstream::RawPayload const& payload,
    std::vector<scene_material> const& materials, dmap_index_t const& res)
    : resolution(res)
{
    read_instructions(payload, materials);
    read_effect_control(payload);

    auto const dmap_present = payload.readBool();
    if (dmap_present) {
        read_diffraction_map(payload, res);
    } else {
        read_diffraction_map_height(payload);
    }

    auto const pcpd_present = payload.readBool();
    if (pcpd_present) {
        read_pre_computed_path_data(payload, res);
    }
}

//--------------------------------------------------------------------------------------------------
void diffraction_data::read_instructions(
    Bitstream::RawPayload const& payload, std::vector<scene_material> const& materials)
{
    instructions_diffraction.reserve(materials.size() + 1);

    // Sound-medium - as above for occlusion
    instructions_diffraction.emplace_back(0, false);

    // Fill with default values
    for (auto const& mat : materials) {
        instructions_diffraction.emplace_back(mat.id);
    }

    auto const has_instructions_diffraction = payload.readBool();
    if (has_instructions_diffraction) {
        // Skip instruction for sound-medium
        auto instr_it = std::begin(instructions_diffraction) + 1;
        while (instr_it != std::cend(instructions_diffraction)) {
            instr_it->diffraction_enable = payload.readBool();
            ++instr_it;
        }
    }
}

//--------------------------------------------------------------------------------------------------
void diffraction_data::read_effect_control(Bitstream::RawPayload const& payload)
{
    auto const has_diffraction_effect_control = payload.readBool();
    if (has_diffraction_effect_control) {
        diffraction_effect_control = payload.readFloat();
    }
}

//--------------------------------------------------------------------------------------------------
void diffraction_data::read_diffraction_map_height(Bitstream::RawPayload const& payload)
{
    auto const has_diffraction_map_height = payload.readBool();
    if (has_diffraction_map_height) {
        diffraction_map_height = payload.readUInt(n_bits_diffraction_map_heigth);
    }
}

//--------------------------------------------------------------------------------------------------
void diffraction_data::read_diffraction_map(
    Bitstream::RawPayload const& payload, dmap_index_t const& res)
{

    auto const num_elements = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);
    blocks_dmap.reserve(num_elements);

    for (auto i = 0U; i < num_elements; ++i) {
        blocks_dmap.emplace_back(payload, res, calc_bit_length(res));
    }
}

//--------------------------------------------------------------------------------------------------
void diffraction_data::read_pre_computed_path_data(
    Bitstream::RawPayload const& payload, dmap_index_t const& res)
{
    auto decode_full = [this](Bitstream::RawPayload const& payload, dmap_index_t const& res) {
        if (!payload.readBool()) {
            return;
        }

        auto const bit_length = calc_bit_length(res);
        auto const start_xy = unpack(payload.readUInt(bit_length), res);

        std::pair<std::array<voxel_index_val_t, 2>, float> prev;
        for (auto x = 1U; x <= res[0]; ++x) {
            for (auto y = 1U; y <= res[1]; ++y) {
                pre_computed_path_data.emplace_back(payload, start_xy,
                    std::array<voxel_index_val_t, 2>{ x, y }, prev, res, bit_length);
            }
        }
    };

    auto decode_sparse = [this](Bitstream::RawPayload const& payload, dmap_index_t const& res) {
        auto const num_end_idx = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);

        auto const bit_length = calc_bit_length(res);
        auto const start_xy = unpack(payload.readUInt(bit_length), res);

        std::array<voxel_index_val_t, 2> prev_corner;
        for (auto i = 0U; i < num_end_idx; ++i) {
            pre_computed_path_data.emplace_back(payload, start_xy, prev_corner, res, bit_length);
        }
    };

    // Decode pre-computed path data
    auto const num_start_idx = payload.readEscapedUInt(n_bits_esc1, n_bits_esc2, n_bits_esc3);

    using decode_fn = std::function<void(Bitstream::RawPayload const&, dmap_index_t const&)>;
    auto const decode = payload.readBool() ? decode_fn(decode_full) : decode_fn(decode_sparse);

    for (auto i = 0U; i < num_start_idx; ++i) {
        decode(payload, res);
    }
}

//--------------------------------------------------------------------------------------------------
} // namespace Payload
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------