/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <MPEGI/Bitstream/BinCursor.h>

#include <array>
#include <iostream>
#include <memory>
#include <utility>
#include <vector>

using namespace MPEGI;

TEST_CASE("Binary Cursor Reading")
{
    SUBCASE("Read bool (true)")
    {
        std::uint8_t data = 0b10000000;
        auto cursor = BinCursor(&data, 1);

        bool readValue = cursor.readBool();

        CHECK(readValue == true);
    }

    SUBCASE("Read bool (false)")
    {
        std::uint8_t data = 0b01111111;
        auto cursor = BinCursor(&data, 1);

        bool readValue = cursor.readBool();

        CHECK(readValue == false);
    }

    SUBCASE("Read int")
    {
        std::uint8_t data = 0b11111111;
        auto cursor = BinCursor(&data, 1);

        int readValue = cursor.readUInt();

        CHECK(readValue == 255);
    }

    SUBCASE("Read int with length")
    {
        std::uint8_t data = 0b11111111;
        auto cursor = BinCursor(&data, 1);

        int readValue = cursor.readUInt(4);

        CHECK(readValue == 15);
    }

    SUBCASE("Read int with length longer than one byte")
    {
        const std::uint8_t data[2] = { 0xFF, 0xFF };
        auto cursor = BinCursor(data, 2);

        int readValue = cursor.readUInt(16);

        CHECK(readValue == 65535);
    }

    SUBCASE("Read int with length of non-integer multiple of eight")
    {
        const std::uint8_t data[2] = { 0xFF, 0xFF };
        auto cursor = BinCursor(data, 2);

        int readValue = cursor.readUInt(12);

        CHECK(readValue == 4095);
    }

    SUBCASE("Read signed int")
    {
        std::uint8_t data = 0b10000000;
        auto cursor = BinCursor(&data, 1);

        int readValue = cursor.readSInt(8);

        CHECK(readValue == -128);
    }

    SUBCASE("Read float")
    {
        const std::uint8_t data[4] = { 0x3F, 0xA0, 0x00, 0x00 };
        auto cursor = BinCursor(data, 4);

        float readValue = cursor.readFloat();

        CHECK(readValue == 1.25);
    }

    SUBCASE("Read Escaped Integer")
    {
        typedef std::pair<int, std::array<std::uint8_t, 2>> ResultAndData;
        auto data = std::vector<ResultAndData>{
            { 0, { 0x00, 0x00 } }, // "00"
            { 1, { 0x40, 0x00 } }, // "01"
            { 2, { 0x80, 0x00 } }, // "10"
            { 3, { 0xc0, 0x00 } }, // "11000"
            { 4, { 0xc8, 0x00 } }, // "11001"
            { 5, { 0xd0, 0x00 } }, // "11010"
            { 6, { 0xd8, 0x00 } }, // "11011"
            { 7, { 0xe0, 0x00 } }, // "11100"
            { 8, { 0xe8, 0x00 } }, // "11101"
            { 9, { 0xf0, 0x00 } }, // "11110"
            { 10, { 0xf8, 0x00 } }, // "111110000"
            { 11, { 0xf8, 0x80 } }, // "111110001"
            { 12, { 0xf9, 0x00 } }, // "111110010"
            { 13, { 0xf9, 0x80 } }, // "111110011"
            { 14, { 0xfa, 0x00 } }, // "111110100"
            { 15, { 0xfa, 0x80 } }, // "111110101"
            { 16, { 0xfb, 0x00 } }, // "111110110"
            { 17, { 0xfb, 0x80 } }, // "111110111"
            { 18, { 0xfc, 0x00 } }, // "111111000"
            { 19, { 0xfc, 0x80 } }, // "111111001"
            { 20, { 0xfd, 0x00 } }, // "111111010"
            { 21, { 0xfd, 0x80 } }, // "111111011"
            { 22, { 0xfe, 0x00 } }, // "111111100"
            { 23, { 0xfe, 0x80 } }, // "111111101"
            { 24, { 0xff, 0x00 } }, // "111111110"
            { 25, { 0xff, 0x80 } } // "111111111"
        };

        for (const auto& [expected_result, rawData] : data) {
            const std::uint8_t* dataPtr = rawData.data();
            auto cursor = BinCursor(dataPtr, 2);
            int value = cursor.readEscapedUInt(2, 3, 4);

            // https://github.com/doctest/doctest/issues/279
            const auto& expected_result_ = expected_result;
            CHECK(value == expected_result_);
        }
    }

    SUBCASE("Read blob (unique_ptr return)")
    {
        const std::uint8_t data[5] = { 0x00, 0x00, 0x00, 0x01, 0xFF };
        auto cursor = BinCursor(data, 5);
        std::size_t length = 0;

        std::unique_ptr<std::uint8_t[]> readValue = cursor.readBlob(length);

        CHECK(*readValue.get() == 0xFF);
        CHECK(length == 1);
    }

    SUBCASE("Read blob (shared_ptr return)")
    {
        const std::uint8_t data[5] = { 0x00, 0x00, 0x00, 0x01, 0xFF };
        auto cursor = BinCursor(data, 5);
        std::size_t length = 0;

        std::shared_ptr<std::vector<std::uint8_t>> readValue = cursor.readBlob();

        CHECK(readValue->size() == 1);
        CHECK(readValue->at(0) == 0xFF);
    }

    SUBCASE("Reading string")
    {
        const std::uint8_t data[4] = { 0x66, 0x6F, 0x6F, 0x00 };
        auto cursor = BinCursor(data, 4);

        std::string value = cursor.readString();

        CHECK(value == "foo");
    }

    SUBCASE("Read quantised float")
    {
        const std::uint8_t data[2] = { 0xFF, 0x9E };
        auto cursor = BinCursor(data, 2);
        auto qRange = QuantizationRange{ 0.f, 1.f, 8, QuantizationScaling::LINEAR };

        float firstValue = cursor.readQuantFloat(qRange);
        float secondValue = cursor.readQuantFloat(qRange);

        CHECK(firstValue == doctest::Approx(1.f));
        CHECK(secondValue == doctest::Approx(0.6196078431372549f));
    }

    SUBCASE("Byte Alignment Check")
    {
        std::uint8_t data = 0b00000000;
        auto cursor = BinCursor(&data, 1);

        CHECK(cursor.isByteAligned() == true);

        cursor.readUInt(3);

        CHECK(cursor.isByteAligned() == false);

        cursor.readUInt(5);

        CHECK(cursor.isByteAligned() == true);
    }

    SUBCASE("Reading beyond data should throw an exception")
    {
        std::uint8_t data = 0b10000000;
        auto cursor = BinCursor(&data, 1);

        CHECK_THROWS(cursor.readUInt(9));
        CHECK_THROWS(cursor.readUInt(309283));
    }

    SUBCASE("Move Cursor Right")
    {
        std::uint8_t data = 0b11110111;
        auto cursor = BinCursor(&data, 1);

        cursor.movePosRight(4);

        CHECK(cursor.readBool() == false);
    }

    SUBCASE("Move Cursor Left")
    {
        std::uint8_t data = 0b11101111;
        auto cursor = BinCursor(&data, 1);

        cursor.readUInt(4);
        cursor.movePosLeft(1);

        CHECK(cursor.readBool() == false);
    }

    SUBCASE("Moving Cursor ahead of data length should throw exception")
    {
        std::uint8_t data = 0b11111111;
        auto cursor = BinCursor(&data, 1);

        CHECK_THROWS(cursor.movePosRight(9));
    }

    SUBCASE("Moving Cursor behind data start should throw exception")
    {
        std::uint8_t data = 0b11111111;
        auto cursor = BinCursor(&data, 1);

        CHECK_THROWS(cursor.movePosLeft(1));
    }

    SUBCASE("get cursor position")
    {
        std::uint8_t data = 0b11111111;
        auto cursor = BinCursor(&data, 1);

        cursor.readUInt(3);
        cursor.readUInt(2);

        CHECK(cursor.getPosition() == 5);
    }

    SUBCASE("At last byte")
    {
        const std::uint8_t data[2] = { 0x00, 0xFF };
        auto cursor = BinCursor(data, 2);

        cursor.movePosRight(8);
        CHECK(cursor.atLastByte() == false);
        cursor.movePosRight(1);
        CHECK(cursor.atLastByte() == true);
    }

    SUBCASE("Copy Cursor")
    {
        const std::uint8_t data[2] = { 0x00, 0xFF };
        auto cursorA = BinCursor(data, 2);
        auto cursorB = cursorA;
        cursorA.movePosRight(8);

        CHECK(cursorA.readUInt(8) == 255);
        CHECK(cursorB.readUInt(8) == 0);
    }

    SUBCASE("Copy Cursor and check position")
    {
        std::uint8_t data = 0b10000000;
        auto cursorA = BinCursor(&data, 1);
        cursorA.movePosRight(1);
        auto cursorB = cursorA;

        CHECK(cursorB.getPosition() == 0);
        CHECK(cursorB.readUInt(7) == 0);
    }

    SUBCASE("Copy Cursor and throw exception beyond data end")
    {
        std::uint8_t data = 0b10000000;
        auto cursorA = BinCursor(&data, 1);
        cursorA.movePosRight(1);
        auto cursorB = cursorA;

        CHECK_THROWS(cursorB.readUInt(8));
    }

    SUBCASE("Readjust Length")
    {
        const std::uint8_t data[2] = { 0x00, 0xFF };
        auto cursorA = BinCursor(data, 2);
        cursorA.resize(1);
        cursorA.movePosRight(8);

        CHECK_THROWS(cursorA.readUInt(1));
    }

    SUBCASE("Read variable length using getID")
    {
        const std::uint8_t data[1] = { 0b01010100 };
        auto cursor = BinCursor(data, 1);

        float readValue = cursor.getID();

        CHECK(readValue == 42);
    }

    SUBCASE("Read variable length using getCountOrIndex")
    {
        SUBCASE("")
        {

            const std::uint8_t data[1] = { 0b11001010 };
            auto cursor = BinCursor(data, 1);

            float readValue = cursor.getCountOrIndex();

            CHECK(readValue == 42);
        }
        SUBCASE("")
        {
            const std::uint8_t data[1] = { 0b10000000 };
            auto cursor = BinCursor(data, 1);

            float readValue = cursor.getCountOrIndex();

            CHECK(readValue == 1);
        }
        SUBCASE("")
        {
            const std::uint8_t data[1] = { 0b10010010 };
            auto cursor = BinCursor(data, 1);

            float readValue = cursor.getCountOrIndex();

            CHECK(readValue == 65);
        }
    }

    SUBCASE("Read variable length using getDuration")
    {
        SUBCASE("isLongRangeMode=false")
        {
            bool isLongRangeMode = false;
            SUBCASE("")
            {
                const std::uint8_t data[2] = { 0b11001111, 0b0 };
                auto cursor = BinCursor(data, 2);

                float readValue = cursor.getDuration(isLongRangeMode);

                CHECK(std::abs(readValue - 18.0f) < 1);
            }
            SUBCASE("")
            {

                const std::uint8_t data[2] = { 0b01100111, 0b10000000 };
                auto cursor = BinCursor(data, 2);

                float readValue = cursor.getDuration(isLongRangeMode);

                CHECK(std::abs(readValue - 18.3) < 1);
            }

            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b01101111, 0b10101101, 0b11111100 };
                auto cursor = BinCursor(data, 3);

                float readValue = cursor.getDuration(isLongRangeMode);

                CHECK(std::abs(readValue - 18.30012) < 1);
            }

            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b01101010, 0b00110111, 0b10000000 };
                auto cursor = BinCursor(data, 3);

                float readValue = cursor.getDuration(isLongRangeMode);

                CHECK(std::abs(readValue - 18.327) < 1);
            }
        }
        SUBCASE("isLongRangeMode=true")
        {
            bool isLongRangeMode = true;
            SUBCASE("")
            {
                const std::uint8_t data[4] = { 0b01101010, 0b00110100, 0b00000100, 0b10100000 };
                auto cursor = BinCursor(data, 4);

                float readValue = cursor.getDuration(isLongRangeMode);

                CHECK(std::abs(readValue - 3618.327) < 1);
            }
        }
    }

    SUBCASE("Read variable length using getFrequency")
    {
        SUBCASE("")
        {
            const std::uint8_t data[1] = { 0b01001000 };
            auto cursor = BinCursor(data, 1);

            float readValue = cursor.getFrequency();

            CHECK(std::abs(readValue - 1250) < 1);
        }
        SUBCASE("")
        {
            const std::uint8_t data[2] = { 0b11001011, 0b0 };
            auto cursor = BinCursor(data, 2);

            float readValue = cursor.getFrequency();

            CHECK(std::abs(readValue - 1100) < 1);
        }
    }

    SUBCASE("Read variable length using getDistance")
    {
        SUBCASE("isSmallScene=true")
        {
            bool isSmallScene = true;
            SUBCASE("")
            {
                const std::uint8_t data[1] = { 0b10001000 };
                auto cursor = BinCursor(data, 1);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 50) < 0.0005f);
            }

            SUBCASE("")
            {
                const std::uint8_t data[2] = { 0b10001001, 0b00010000 };
                auto cursor = BinCursor(data, 2);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 50.42) < 0.0005f);
            }
            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b10001001, 0b00010001, 0b10000000 };
                auto cursor = BinCursor(data, 3);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 50.428) < 0.0005f);
            }
            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b10001000, 0b00000100, 0b10101000 };
                auto cursor = BinCursor(data, 3);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 50) < 0.0005f);

                readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 20) < 0.0005f);

                readValue = cursor.getID();

                CHECK(std::abs(readValue - 42) < 0.0005f);
            }
        }
        SUBCASE("isSmallScene=false")
        {

            bool isSmallScene = false;
            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b10001001, 0b11101000, 0b10001100 };
                auto cursor = BinCursor(data, 3);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 350.428) < 0.0005f);
            }
            SUBCASE("")
            {
                const std::uint8_t data[4] = { 0b10001001, 0b11111100, 0b10100010, 0b00110000 };
                auto cursor = BinCursor(data, 4);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 9350.428) < 0.0005f);
            }
            SUBCASE("")
            {
                const std::uint8_t data[5] = { 0b10001001, 0b11111100, 0b01110010, 0b10001000,
                    0b11000000 };
                auto cursor = BinCursor(data, 5);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 19350.428) < 0.0005f);
            }
            SUBCASE("")
            {
                const std::uint8_t data[4] = { 0b10001001, 0b00011100, 0b10100010, 0b00110000 };
                auto cursor = BinCursor(data, 4);

                float readValue = cursor.getDistance(isSmallScene);

                CHECK(std::abs(readValue - 9050.428) < 0.0005f);
            }
        }
    }
    SUBCASE("Read variable length using getGain")
    {
        SUBCASE("isHiPrecGain=false")
        {
            bool isHiPrecGain = false;
            SUBCASE("")
            {
                const std::uint8_t data[1] = { 0b01000000 };
                auto cursor = BinCursor(data, 1);

                float readValue = cursor.getGain(isHiPrecGain);

                CHECK(std::abs(readValue - 3) < 0.01f);
            }
            SUBCASE("Read variable length using getGain")
            {
                const std::uint8_t data[1] = { 0b11011110 };
                auto cursor = BinCursor(data, 1);

                float readValue = cursor.getGain(isHiPrecGain);

                CHECK(std::abs(readValue - 5) < 0.01f);
            }
            SUBCASE("Read variable length using getGain")
            {
                const std::uint8_t data[2] = { 0b01101011, 0b00111100 };
                auto cursor = BinCursor(data, 2);

                float readValue = cursor.getGain(isHiPrecGain);

                CHECK(std::abs(readValue - -42) < 0.01f);
            }
        }
        SUBCASE("isHiPrecGain=true")
        {
            bool isHiPrecGain = true;
            SUBCASE("")
            {
                const std::uint8_t data[2] = { 0b11011101, 0b10100000 };
                auto cursor = BinCursor(data, 2);

                float readValue = cursor.getGain(isHiPrecGain);

                CHECK(std::abs(readValue - 4.2) < 0.01f);
            }

            SUBCASE("")
            {
                const std::uint8_t data[2] = { 0b10111010, 0b01000000 };
                auto cursor = BinCursor(data, 2);

                float readValue = cursor.getGain(isHiPrecGain);

                CHECK(std::abs(readValue - -4.2) < 0.01f);
            }
        }
    }

    SUBCASE("Read variable length using getPosition")
    {
        SUBCASE("isSmallScene=true")
        {
            bool isSmallScene = true;
            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b00100100, 0b11000100, 0b11010000 };
                auto cursor = BinCursor(data, 3);

                auto readValue = cursor.getPosition(isSmallScene);
                auto x = readValue[0];
                auto y = readValue[1];
                auto z = readValue[2];

                CHECK(std::abs(x - 12) < 0.01f);
                CHECK(std::abs(y - 4) < 0.01f);
                CHECK(std::abs(z - 7) < 0.01f);
            }
            SUBCASE("")
            {
                const std::uint8_t data[3] = { 0b00100101, 0b11000100, 0b11010001 };
                auto cursor = BinCursor(data, 3);

                auto readValue = cursor.getPosition(isSmallScene);
                auto x = readValue[0];
                auto y = readValue[1];
                auto z = readValue[2];

                CHECK(std::abs(x - -12) < 0.01f);
                CHECK(std::abs(y - 4) < 0.01f);
                CHECK(std::abs(z - -7) < 0.01f);
            }

            SUBCASE("")
            {
                const std::uint8_t data[7] = { 0b10001001, 0b00010001, 0b10010110, 0b00101111,
                    0b00001101, 0b00111001, 0b01001100 };
                auto cursor = BinCursor(data, 7);

                auto readValue = cursor.getPosition(isSmallScene);
                auto x = readValue[0];
                auto y = readValue[1];
                auto z = readValue[2];

                CHECK(std::abs(x - -50.428) < 0.01f);
                CHECK(std::abs(y - 27.3) < 0.01f);
                CHECK(std::abs(z - -7.001) < 0.01f);
            }
        }
    }

    SUBCASE("Read variable length using getOrientation")
    {
        SUBCASE("")
        {
            const std::uint8_t data[1] = { 0b01000000 };
            auto cursor = BinCursor(data, 1);

            std::array<float, 3> readValue = cursor.getOrientation();
            auto x = readValue[0];
            auto y = readValue[1];
            auto z = readValue[2];

            CHECK(std::fmod(std::abs(x - 0), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(y - 0), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(z - 0), 360.0f) < 0.05f);
        }
        SUBCASE("")
        {
            const std::uint8_t data[1] = { 0b10100000 };
            auto cursor = BinCursor(data, 1);

            std::array<float, 3> readValue = cursor.getOrientation();
            auto x = readValue[0];
            auto y = readValue[1];
            auto z = readValue[2];

            CHECK(std::fmod(std::abs(x - 0), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(y - -90), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(z - 0), 360.0f) < 0.05f);
        }

        SUBCASE("")
        {
            const std::uint8_t data[3] = { 0b11101001, 0b00110010, 0b00100100 };
            auto cursor = BinCursor(data, 3);

            std::array<float, 3> readValue = cursor.getOrientation();
            auto x = readValue[0];
            auto y = readValue[1];
            auto z = readValue[2];

            CHECK(std::fmod(std::abs(x - 180.0f), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(y - 60), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(z - 180), 360.0f) < 0.05f);
        }

        SUBCASE("")
        {
            const std::uint8_t data[13] = { 0b10001000, 0b01100110, 0b10000000, 0b00000000,
                0b00001000, 0b01001110, 0b00000000, 0b00000000, 0b00001000, 0b01100110, 0b10000000,
                0b00000000, 0b00000000 };
            auto cursor = BinCursor(data, 13);

            std::array<float, 3> readValue = cursor.getOrientation();
            auto x = readValue[0];
            auto y = readValue[1];
            auto z = readValue[2];

            CHECK(std::fmod(std::abs(x - 180), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(y - 60), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(z - 180), 360.0f) < 0.05f);
        }

        SUBCASE("")
        {
            const std::uint8_t data[3] = { 0b11111001, 0b00110100, 0b01101000 };
            auto cursor = BinCursor(data, 3);

            std::array<float, 3> readValue = cursor.getOrientation();
            auto x = readValue[0];
            auto y = readValue[1];
            auto z = readValue[2];

            CHECK(std::fmod(std::abs(x - 60), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(y - 90), 360.0f) < 0.05f);
            CHECK(std::fmod(std::abs(z - 90), 360.0f) < 0.05f);
        }
    }
}