/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <algorithm>
#include <random>

#include <doctest/doctest.h>

#include <MPEGI/Bitstream/BinCursor.h>
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/HuffmanDecoder.h>
#include <iostream>

using namespace MPEGI;

constexpr size_t bitsPerByte{ 8 };

std::vector<uint8_t> bsVectorFromString(std::string dataString)
{
    while (dataString.size() % bitsPerByte != 0) {
        dataString += "0";
    }
    size_t vecTargetSize = dataString.size() / bitsPerByte;
    std::vector<uint8_t> bsVec;
    size_t stepSize = bitsPerByte;
    for (int i = 0; i < vecTargetSize; i++) {
        std::string substr = dataString.substr(i * stepSize, stepSize);
        uint8_t curr = std::stoi(substr, nullptr, 2);
        bsVec.push_back(curr);
    }
    return bsVec;
}

TEST_CASE("Test combined bitstream with 1 codebook")
{
    std::string codebookStr(
        "0100011111000010001010011001100101111001000000001100010100111010100100100011");
    std::string bsDataStr("010100010011101001111010100111101100010011110110011000100111101100110011"
                          "011010011110110011001101111010");
    std::string combinedStr = codebookStr + bsDataStr;
    Bitstream::RawPayloadForTesting rawPayload(combinedStr);
    BinCursor bitBuffer(rawPayload.getBuffer(), rawPayload.getSize());

    HuffmanDecoder decoder;
    decoder.deserializeCodebook(bitBuffer);

    auto list = decoder.decode(bitBuffer, combinedStr.length());

    std::vector<unsigned int> valuesProducedByPythonDec = { 97, 97, 115, 97, 115, 100, 97, 115, 100,
        102, 97, 115, 100, 102, 103, 97, 115, 100, 102, 103, 104, 97, 115, 100, 102, 103, 104, 106,
        97, 115, 100, 102, 103, 104, 106, 107 };

    CHECK_EQ(list.size(), valuesProducedByPythonDec.size());
    for (int i = 0; i < list.size(); i++) {
        CHECK_EQ(list.at(i), valuesProducedByPythonDec.at(i));
    }
}

TEST_CASE("Pseudo voxel data with 3 codebooks")
{
    std::string bs01(
        "110001011111000010100000011010110101000110000010100011011000111010110111001111100001001010"
        "011010111000101101010100110000001011001011010100111000110110100110000010100110000110110110"
        "010111111011110000001011111110110101000011101100111001011011010100101101111100101001010001"
        "001000000011010110100001111011101011111000110011110111110010000011011011000000000110011110"
        "101000111101110010000010110000000001110111111110000111110101001011110011010011110011100010"
        "011110100000010001011001100000100011110100111111011000101100011111001101100111101010100100"
        "110110100101101111100110101101010111111101011110001010100111000111101011101111010001101011"
        "010100101001001010010001111001001111101000111001110111001100100010000100111111011101011001"
        "010100101110001111100001010011100000101010111110001010100010110110000111111100000100111100"
        "100010101010111101100101011101000110010000000101011000110111000111000011101011110011111011"
        "000111011010000001000000100010110111000111111101110001100001010101011000001101011000001110"
        "101001100010010001100000011000100111010010001011100010101111101001011001110111101011000000"
        "111011100101010101010100010101");
    std::vector<unsigned int> targetVcX = { 31, 41, 37, 52, 30, 46, 42, 57, 32, 52, 44, 55, 56, 57,
        39, 48, 36, 46, 59, 47, 57, 30, 42, 33, 56, 59, 30, 30, 57, 31, 50, 36, 55, 49, 30, 42, 34,
        39, 54, 37, 52, 40, 43, 43, 43, 51, 41, 51, 47, 44, 47, 51, 48, 34, 34, 47, 57, 34, 39, 59,
        40, 58, 46, 54, 51, 35, 52, 45, 34, 38, 41, 35, 47, 30, 42, 58, 33, 58, 54, 50, 46, 35, 37,
        58, 32, 42, 51, 44, 47, 36, 42, 56, 53, 41, 35, 53, 46, 53, 51, 38 };
    std::vector<unsigned int> targetVcY = { 2, 2, 2, 4, 4, 3, 4, 3, 3, 4, 3, 2, 4, 3, 2, 2, 4, 3, 2,
        4, 3, 3, 2, 2, 4, 3, 4, 2, 3, 3, 4, 4, 3, 3, 4, 3, 2, 3, 4, 3, 2, 4, 4, 3, 2, 3, 3, 2, 3, 3,
        4, 2, 2, 2, 3, 3, 4, 2, 3, 2, 4, 2, 4, 2, 3, 2, 3, 3, 3, 4, 3, 2, 4, 2, 2, 2, 4, 3, 3, 2, 2,
        4, 3, 2, 4, 2, 4, 4, 2, 2, 2, 2, 2, 3, 4, 4, 2, 2, 2, 4 };
    std::vector<unsigned int> targetVcZ = { 13, 17, 15, 18, 11, 18, 12, 14, 12, 19, 15, 13, 16, 11,
        19, 14, 13, 13, 12, 18, 11, 18, 14, 18, 15, 15, 13, 15, 12, 13, 12, 18, 14, 16, 15, 14, 12,
        10, 13, 15, 12, 16, 11, 13, 13, 12, 14, 18, 17, 13, 17, 16, 17, 12, 10, 15, 18, 19, 10, 13,
        16, 12, 11, 16, 13, 19, 14, 19, 13, 17, 18, 10, 16, 17, 13, 10, 17, 10, 17, 14, 16, 10, 13,
        13, 17, 15, 14, 11, 16, 14, 11, 19, 15, 17, 18, 17, 18, 16, 17, 17 };
    Bitstream::RawPayloadForTesting rawPayload(bs01);
    BinCursor bitBuffer(rawPayload.getBuffer(), rawPayload.getSize());

    // read codebook headers from bitstream
    HuffmanDecoder decoderVcX;
    HuffmanDecoder decoderVcY;
    HuffmanDecoder decoderVcZ;
    decoderVcX.deserializeCodebook(bitBuffer);
    decoderVcY.deserializeCodebook(bitBuffer);
    decoderVcZ.deserializeCodebook(bitBuffer);

    // read actual voxel component data sequentially
    size_t num_bits = bs01.length();
    std::vector<unsigned int> decodedVcX;
    std::vector<unsigned int> decodedVcY;
    std::vector<unsigned int> decodedVcZ;
    while (bitBuffer.getPosition() < num_bits) {
        unsigned int x = decoderVcX.decodeSingle(bitBuffer);
        unsigned int y = decoderVcY.decodeSingle(bitBuffer);
        unsigned int z = decoderVcZ.decodeSingle(bitBuffer);
        decodedVcX.push_back(x);
        decodedVcY.push_back(y);
        decodedVcZ.push_back(z);
    }

    // check vcX
    CHECK_EQ(decodedVcX.size(), targetVcX.size());
    for (int i = 0; i < decodedVcX.size(); i++) {
        CHECK_EQ(decodedVcX.at(i), targetVcX.at(i));
    }

    // check vcY
    CHECK_EQ(decodedVcY.size(), targetVcY.size());
    for (int i = 0; i < decodedVcY.size(); i++) {
        CHECK_EQ(decodedVcY.at(i), targetVcY.at(i));
    }

    // check vcZ
    CHECK_EQ(decodedVcZ.size(), targetVcZ.size());
    for (int i = 0; i < decodedVcZ.size(); i++) {
        CHECK_EQ(decodedVcZ.at(i), targetVcZ.at(i));
    }
}
