/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include <optional>
#include <vector>

namespace MPEGI {

struct AirflowProfileParameters
{
    //! Enum class for the way in which custom profiles is stored
    enum class ProfileType : unsigned int
    {
        Freq_Gain_Pairs = 0u,
        Filter_Params,
        Reserved
    };

    //! Specifies which type of airflow profile this i
    enum class ProfileModification : unsigned int
    {
        speed = 0u,
        azimuth,
        elevation,
        Reserved
    };

    AirflowProfileParameters(ProfileType _profileType, unsigned int _freqGridIdx,
        ProfileModification _profileModifies, float _modificationValue,
        const std::vector<std::pair<float, float>>& _gain,
        const std::vector<std::pair<float, float>>& _q)
        : profileType(_profileType)
        , freqGridIdx(_freqGridIdx)
        , profileModifies(_profileModifies)
        , modificationValue(_modificationValue)
        , gain(_gain)
        , q(_q)
    {}

    const ProfileType profileType;
    const unsigned int freqGridIdx;
    const ProfileModification profileModifies;
    const float modificationValue;
    const std::vector<std::pair<float, float>> gain;
    const std::vector<std::pair<float, float>> q;
};

struct AirflowSource
{
    AirflowSource(unsigned int _id, unsigned int _generatorID,
        const std::string& _generatorStringID, unsigned int _defaultProfile, bool _hasValidRegions,
        const std::vector<AirflowProfileParameters>& _customAirflowProfiles,
        bool _activeInDefaultRegion, const std::vector<int>& _activeRegionIDs)
        : id(_id)
        , generatorID(_generatorID)
        , generatorStringID(_generatorStringID)
        , defaultProfile(_defaultProfile)
        , hasValidRegions(_hasValidRegions)
        , customAirflowProfiles(_customAirflowProfiles)
        , activeInDefaultRegion(_activeInDefaultRegion)
        , activeRegionIDs(_activeRegionIDs)
    {}

    bool isActiveInRegion(const std::optional<int>& regionId) const
    {
        if (regionId.has_value()) {
            // Return true if region ID is in the list of activeRegionIDs, false otherwise
            return std::find(activeRegionIDs.begin(), activeRegionIDs.end(), regionId.value())
                != activeRegionIDs.end();
        } else {
            // regionId does not have a value: region is default AE
            return activeInDefaultRegion;
        }
    }

    const unsigned int id;
    const unsigned int generatorID;
    const std::string generatorStringID;
    const unsigned int defaultProfile;
    const bool hasValidRegions;
    const std::vector<AirflowProfileParameters> customAirflowProfiles;
    const bool activeInDefaultRegion;
    const std::vector<int> activeRegionIDs;
};

} // namespace MPEGI
