/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_CLOCK
#define MPEGI_CLOCK

#include <chrono>
#include <mutex>

namespace MPEGI {

using ChronoClock = std::chrono::steady_clock;
using ChronoDuration = std::chrono::duration<double, std::ratio<1>>; // in seconds
using ChronoTimePoint = std::chrono::time_point<ChronoClock, ChronoDuration>;

//! Interface for a Clock that provides a continuous time since start was called
/*!
 * Make sure that every Clock implementation is thread-safe!
 */
class Clock
{
public:
    //! Destructor
    virtual ~Clock(){};

    //! Return the current time in seconds
    virtual double getCurrentTime() = 0;

    //! Start the Clock (resets the time if already running)
    virtual void start() = 0;

    //! Stop the Clock (resets the time)
    virtual void stop() = 0;

    //! Return true if clock is currently running
    virtual bool isRunning() = 0;

    //! Synchronize with external time source
    /*!
	 * \param t 		seconds since start() was called
	 */
    virtual void sync(double t) = 0;
};

//! ExternalClock provides a Clock that only proceeds, when it is synced with sync()
class ExternalClock : public Clock
{
public:
    //! Constructor
    ExternalClock()
        : running(false){};

    //! Destructor
    ~ExternalClock(){};

    //! Inherited from Clock
    double getCurrentTime() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        return t;
    };

    //! Inherited from Clock
    void start() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        t = 0.0;
        running = true;
    };

    //! Inherited from Clock
    void stop() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        t = 0.0;
        running = false;
    };

    //! Inherited from Clock
    bool isRunning() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        return running;
    };

    //! Inherited from Clock
    void sync(double nt) override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        t = nt;
    };

private:
    bool running;
    double t;
    std::mutex clockMutex;
};

//! InternalClock provides a steady Clock with the posibility of external synchronization
class InternalClock : public Clock
{
public:
    //! Constructor
    InternalClock()
        : running(false){};

    //! Destructor
    ~InternalClock(){};

    //! Inherited from Clock
    double getCurrentTime() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        if (!running) {
            return 0.0;
        }

        return ChronoDuration(ChronoClock::now() - startTime).count();
    };

    //! Inherited from Clock
    void start() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        startTime = ChronoClock::now();
        running = true;
    };

    //! Inherited from Clock
    void stop() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        running = false;
    };

    //! Inherited from Clock
    bool isRunning() override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        return running;
    };

    //! Inherited from Clock
    void sync(double t) override
    {
        std::lock_guard<std::mutex> guard(clockMutex);

        startTime = ChronoClock::now() - ChronoDuration(t);
    };

private:
    bool running;
    ChronoTimePoint startTime;

    std::mutex clockMutex;
};

} // namespace MPEGI

#endif // MPEGI_CLOCK
