/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_CONFIGURATION
#define MPEGI_CONFIGURATION

#include <MPEGI/Helpers/ConfigObserver.h>

#include <EAR/Exception.h>
#include <EAR/Variant.h>

#include <algorithm>
#include <map>
#include <string>
#include <unordered_map>
#include <vector>

namespace MPEGI {
// TODO: Full array support instead of float pointers

//! Configuration container to maintain a globally accessible config with file read/write support
/*!
 * Set default values in the constructor in the src file.
 */
class Config
{
public:
    //! Destructor
    virtual ~Config();

    //! Clear storage map and set some useful global defaults
    static void reset();

    //! Return a string serialized version of the current config
    /*!
     * Caution: Float pointer values are skipped and not serialized!
     */
    static std::string serialize();

    //! Set the config values as given in a serialized config
    /*!
     * Existing config keys are overwritten. Observers are called if config values change.
     *
     * \param str          string config as given by Config::serialize()
     */
    static void deserialize(std::string str);

    //! Access a value for a key and provide a default value (int)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline int getIntDef(const std::string& key, int defaultVal)
    {
        auto inst = getInstance();
        inst->setDef(key, EAR::Variant(defaultVal));
        return static_cast<int>(inst->storage.at(key));
    };

    //! Access a value for a key and provide a default value (bool)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline bool getBoolDef(const std::string& key, bool defaultVal)
    {
        auto inst = getInstance();
        inst->setDef(key, EAR::Variant(defaultVal));
        return static_cast<bool>(inst->storage.at(key));
    };

    //! Access a value for a key and provide a default value (float)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline float getFloatDef(const std::string& key, float defaultVal)
    {
        auto inst = getInstance();
        inst->setDef(key, EAR::Variant(defaultVal));
        return static_cast<float>(inst->storage.at(key));
    };

    //! Access a value for a key and provide a default value (string)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline const std::string& getStringDef(
        const std::string& key, const std::string& defaultVal)
    {
        auto inst = getInstance();
        inst->setDef(key, EAR::Variant(defaultVal));
        return static_cast<const std::string&>(inst->storage.at(key));
    };

    //! Access a value for a key and provide a default value (float*)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline const float* getFloatPtrDef(const std::string& key, const float* defaultVal)
    {
        auto inst = getInstance();
        inst->setDef(key, EAR::Variant(defaultVal));
        return static_cast<const float*>(inst->storage.at(key));
    };

    //! Access a value for a key and provide a default value (std::vector<float>)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline const std::vector<float>& getFloatVectorDef(
        const std::string& key, const std::vector<float>& defaultVal)
    {
        auto inst = getInstance();
        inst->setDef(key, EAR::Variant(defaultVal));
        return static_cast<const std::vector<float>&>(inst->storage.at(key));
    };

    //! Access a value for a key (casted to int directly)
    /*!
     * \param key          string key of the config value
     */
    static inline int getInt(const std::string& key)
    {
        auto inst = getInstance();
        EAR_CHECK_REL(inst->storage.find(key) != inst->storage.end(),
            "Key \"" + key + "\" not present in configuration.")
        return static_cast<int>(inst->storage.at(key));
    };

    //! Access a value for a key (casted to bool directly)
    /*!
     * \param key          string key of the config value
     */
    static inline bool getBool(const std::string& key)
    {
        auto inst = getInstance();
        EAR_CHECK_REL(inst->storage.find(key) != inst->storage.end(),
            "Key \"" + key + "\" not present in configuration.")
        return static_cast<bool>(inst->storage.at(key));
    };

    //! Access a value for a key (const, casted to float directly)
    /*!
     * \param key          string key of the config value
     */
    static inline float getFloat(const std::string& key)
    {
        auto inst = getInstance();
        EAR_CHECK_REL(inst->storage.find(key) != inst->storage.end(),
            "Key \"" + key + "\" not present in configuration.")
        return static_cast<float>(inst->storage.at(key));
    };

    //! Access a value for a key (const, casted to string directly)
    /*!
     * \param key          string key of the config value
     */
    static inline const std::string& getString(const std::string& key)
    {
        auto inst = getInstance();
        EAR_CHECK_REL(inst->storage.find(key) != inst->storage.end(),
            "Key \"" + key + "\" not present in configuration.")
        return static_cast<const std::string&>(inst->storage.at(key));
    };

    //! Access a value for a key (const, casted to float* directly)
    /*!
     * \param key          string key of the config value
     * \param defaultVal   value that should be used when no configuration is present
     *                     for the given key
     */
    static inline const float* getFloatPtr(const std::string& key)
    {
        auto inst = getInstance();
        EAR_CHECK_REL(inst->storage.find(key) != inst->storage.end(),
            "Key \"" + key + "\" not present in configuration.")
        return static_cast<const float*>(inst->storage.at(key));
    };

    //! Set a config value (creates if not exists)
    static inline void set(const std::string& key, const EAR::Variant& val)
    {
        auto inst = getInstance();
        inst->storage[key] = val;

        for (auto obs : inst->observers[key]) {
            obs->configValueChanged(key);
        }
    };

    //! Set a config value (only if it does not already exist)
    static inline void setDef(const std::string& key, const EAR::Variant& val)
    {
        auto inst = getInstance();
        if (inst->storage.find(key) == inst->storage.end()) {
            inst->storage[key] = val;

            for (auto obs : inst->observers[key]) {
                obs->configValueChanged(key);
            }
        }
    };

    //! Register observer for a specific key
    static inline void observeKey(const std::string& key, ConfigObserver* obs)
    {
        auto inst = getInstance();
        inst->observers[key].push_back(obs);
    };

    //! Remove an observer for a specific key
    static inline void removeObserver(const std::string& key, ConfigObserver* obs)
    {
        auto inst = getInstance();
        auto it = std::find(inst->observers[key].begin(), inst->observers[key].end(), obs);

        if (it != inst->observers[key].end()) {
            inst->observers[key].erase(it);
        }
    };

private:
    //! Private constructor (singleton impl)
    Config();

    static Config* instance;

    static inline Config* getInstance()
    {
        if (!instance) {
            instance = new Config();
            Config::reset();
        }
        return instance;
    }

    std::map<std::string, EAR::Variant> storage;
    std::unordered_map<std::string, std::vector<ConfigObserver*>> observers;
};

} // namespace MPEGI

#endif /* MPEGI_CONFIGURATION */
