/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_CONTROLLABLE
#define MPEGI_CONTROLLABLE

#include <EAR/Exception.h>

#include <atomic>
#include <chrono>
#include <memory>
#include <thread>

namespace MPEGI {

typedef std::chrono::duration<double> Seconds;

//! Controllable lets you receive a callback to the virtual function controlStep() at specified intervals
/*!
 * Note, that if you use member variables of your derived class in the controlStep() call, you have to stop the controller with stopController() in the derived destructor to avoid bad access!
 */
class Controllable
{
public:
    //! Destructor
    virtual ~Controllable()
    {
        // stop control thread if it is running
        if (controlThread && controlThread->joinable()) {
            controllerRunning.store(false);
            controlThread->join();
        }
    };

    //! Start a controller that is calling controlStep() at control rate
    /*!
     * Control rate defaults to 100 Hz and can be changed with setControlRate()
     */
    void startController()
    {
        controllerRunning.store(true);
        controlThread = std::make_unique<std::thread>(&Controllable::control, this);
    };

    //! Stop the controller
    void stopController()
    {
        controllerRunning.store(false);

        if (controlThread && controlThread->joinable()) {
            controlThread->join();
            controlThread.reset(nullptr);
        }
    };

    //! Change the control rate from its default value of 100 Hz
    /*!
     * \param f               frequency at which the Scene state will be updated [Hz]. Will throw an error for f <= 0 Hz.
     *                        Values above 1000 Hz effectively lead to maximum control rate, because the control interval is rounded to the next millisecond.
     */
    void setControlRate(double f)
    {
        if (f > 0) {
            controlInterval.store(Seconds(1 / f));
        } else {
            EAR_FAIL("Control rate has to be larger than 0 Hz. Call stopController() if you want "
                     "to stop it.");
        }
    };

    //! Return the time in seconds between two control steps
    double getControlInterval() const { return controlInterval.load().count(); };

    //! Return true if the controller is running
    bool isRunning() const { return controllerRunning.load(); };

    //! Return the runtime of the last controlStep() call in seconds (for debugging/profiling)
    double getLastControlRuntime() const { return lastRuntime.load(); }

    //! Can be called to manually trigger a control step (e.g. in offline mode)
    void triggerControlStep()
    {
        Seconds runtime;
        auto start = std::chrono::high_resolution_clock::now();

        controlStep();

        runtime =
            std::chrono::duration_cast<Seconds>(std::chrono::high_resolution_clock::now() - start);
        lastRuntime.store(runtime.count());
    };

    //! Callback of the actual implementation that is called at control rate
    virtual void controlStep() = 0;

protected:
    std::atomic<bool> controllerRunning{ false };

private:
    void control()
    {
        Seconds runtime;
        std::chrono::high_resolution_clock::time_point start;

        while (controllerRunning.load()) {
            start = std::chrono::high_resolution_clock::now();

            controlStep();

            runtime = std::chrono::duration_cast<Seconds>(
                std::chrono::high_resolution_clock::now() - start);
            lastRuntime.store(runtime.count());

            // one loop may take longer than controlInterval if update() is very slow
            const auto cInterval = controlInterval.load();
            if (runtime.count() < cInterval.count()) {
                std::this_thread::sleep_for(cInterval - runtime);
            }
        }
    };

    std::atomic<Seconds> controlInterval{ Seconds(1 / 100.0) }; // 100 Hz by default
    std::unique_ptr<std::thread> controlThread;
    std::atomic<double> lastRuntime;
};

} // namespace MPEGI

#endif /* MPEGI_CONTROLLABLE */
