/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include <array>
#include <vector>

#include "MPEGI/Helpers/VoxelTypes.h"

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {
//--------------------------------------------------------------------------------------------------
/**
 * Pathfinder Path (2D)
 */
struct path_2d
{
    using val_type = std::array<int, 2>;

    val_type get(std::size_t idx) const { return { path_.at(2 * idx), path_.at(2 * idx + 1) }; }

    void reserve(std::size_t max_path_length)
    {
        path_.reserve(max_path_length * 2); // *2 because x/y-interleaved storage
    }

    std::size_t size() const { return path_.size() / 2; }

    std::vector<int>& get() noexcept { return path_; }
    std::vector<int> const& get() const noexcept { return path_; }

protected:
    /// @todo Refactor path, e.g. add a "point" struct (x,y) instead of interleaved x/y. Needs adaption in cornder-finding algorithm
    std::vector<int> path_;
};

//--------------------------------------------------------------------------------------------------
/**
 * Pathfinder Interface
 */
class i_pathfinder
{
public:
    enum
    {
        pathfinder_ok = 0,
        pathfinder_init_error,
        pathfinder_no_path,
        pathfinder_no_direct_view
    };

    enum
    {
        smoothing_enabled = 1,
        smoothing_disabled = 0
    };

    virtual ~i_pathfinder() = default;

    /**
     * Find at path on the projection map
     *
     * @param from Start point [1..dim]
     * @param to End point     [1..dim]
     * @return pathfinder_ok on success, else error code
     */
    virtual int find_path(voxel_index_t const& from, voxel_index_t const& to) = 0;

    /**
     * Test direct view on the projection map
     *
     * @param from Start point [1..dim]
     * @param to End point     [1..dim]
     * @return pathfinder_ok on success, else error code
     */
    virtual int test_direct_view(voxel_index_t const& from, voxel_index_t const& to) = 0;

    /**
     * (Re-)initialize the pathfinder with a new projection map
     *
     * @param p_map The projection map
     * @param smooth_path If set, a smoothed path is calculated
     * @return pathfinder_ok on success, else error code
     */
    virtual int init(projection_map const& p_map, bool smooth_path) = 0;

    /**
     * Get the path
     *
     * @return A valid path if find_path returned pathfinder_ok, else undefined
     */
    virtual path_2d const& get_path() const = 0;

    /**
     * Get the path length
     *
     * @return The length of the path if find_path returned pathfinder_ok, else undefined
     */
    virtual float const& get_path_length() const = 0;
};

//--------------------------------------------------------------------------------------------------
class pathfinder_noop : public i_pathfinder
{
    int find_path(voxel_index_t const& from, voxel_index_t const& to) override
    {
        return pathfinder_no_path;
    }

    int init(projection_map const& p_map, bool smooth_path) override { return pathfinder_ok; }

    path_2d const& get_path() const override
    {
        static path_2d const p;
        return p;
    }
};

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------