/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include <memory>
#include <unordered_map>
#include <vector>

#include "MPEGI/Helpers/VoxelTypes.h"

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
struct occlusion_recovery
{
    using recovery_t = std::vector<std::pair<float, float>>;
    using shared = std::shared_ptr<occlusion_recovery>;

    enum
    {
        num_quant_points_recovery = 32
    };

    /**
     * Initialize occlusion recovery
     *
     * @param instructions Renderer instructions per material ID
     */
    void init(Voxel::renderer_instruction_map_t const& instructions);

    /**
     * Apply occlusion recovery
     * (Lis) <- t1 - | <- t0 - | <--- (Obj)
     *
     * @param material_id Material ID (to select recovery curve)
     * @param sqr_distance Squared distance to first occluding voxel
     * @param t0 Transmission behind the first occluder (for interpolation)
     * @param t1 Transmission in front of the first occluder (for interpolation), contains result
     * @param compensation_eq Compensation EQs for secondary sources (e.g. diffraction) to compensate for non-absorbed energy
     */
    void apply_recovery(int material_id, float sqr_distance, std::vector<float> const& t0,
        std::vector<float>& t1, std::vector<float>& compensation_eq) const;

    /**
     * Apply occlusion recovery, @sa apply_recovery(int, flat, std::vector<float> const&, std::vector<float>&, std::vector<float> const&)
     *
     * @param material_id Material ID
     * @param sqr_distance Squared distance to first occluding voxel
     * @param t Transmission
     */
    void apply_recovery(int material_id, float sqr_distance, std::vector<float>& t);

    /**
     * Check whether recovery is activated for given material
     *
     * @param material_id Material ID
     * @return True if recovery is activated, else false
     */
    bool has_recovery(int material_id) const;

protected:
    /**
     * Get recovery factor for material and distance
     *
     * @param material_id Material ID
     * @param sqr_magnitude Squared distance to occluding voxel
     * @return Recovery factor [0..1)
     */
    float get_recovery_factor(int material_id, float sqr_magnitude) const;

    /**
     * Initialize recovery table
     *
     * @tparam T Must be floating point
     * @param recovery_setting Setting from the bitstream
     * @param dr Table squared-distance -> recovery factor
     * @param num_quant_points Table size
     */
    template <typename T>
    void init_table(float recovery_setting, std::vector<std::pair<T, T>>& dr,
        int num_quant_points = num_quant_points_recovery)
    {
        static_assert(std::is_floating_point_v<T>, "Must be floating point");

        if (recovery_setting == 0.f) {
            // No recovery
            return;
        }

        dr.push_back(std::make_pair(0.f, 0.f));
        for (int i = 1; i <= num_quant_points; ++i) {
            auto const r_rev = i / static_cast<T>(num_quant_points + 1);
            auto const d_rev = -std::log2(T{ 2 } / (r_rev + 1) - 1) / recovery_setting;

            dr.push_back(std::make_pair(d_rev * d_rev, r_rev));
        }
    }

    std::unordered_map<float, recovery_t>
        _recovery_by_factor; //!< Recovery lookup per recovery control factor
    std::unordered_map<int, recovery_t> _recovery_by_material; //!< Recovery lookup per material ID
};

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------