/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "MPEGI/Renderer/RenderItemStore.h"
#include "MPEGI/Renderer/RenderList.h"

#include <memory>
#include <vector>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {
//--------------------------------------------------------------------------------------------------
/**
 * Class for managing secondary render items, e.g. diffraction
 */
template <typename T, typename P = std::unique_ptr<RenderItem>> struct secondary_item_storage
{
    using render_item_ptr = P;

    secondary_item_storage(secondary_item_storage const&) = delete;
    secondary_item_storage& operator=(secondary_item_storage const&) = delete;

    /**
     * Mute secondary item(s) associated with the given primary item
     * Mutes all secondary items if no secondary_item_idx is given
     *
     * @param primary_item_idx Index of the primary item
     * @param secondary_item_idx Index of the associated secondary item (if any)
     */
    void mute(int primary_item_idx, std::optional<int> const& secondary_item_idx = std::nullopt)
    {
        auto it = _storage.find(primary_item_idx);
        if (it == std::cend(_storage)) {
            // No secondary items for this primary item
            return;
        }

        if (secondary_item_idx.has_value()) {
            auto item_it = it->second.find(secondary_item_idx.value());
            if (item_it != std::cend(it->second)) {
                item_it->second->setEq(_zero);
            }
        } else {
            auto& items = it->second;
            for (auto& [idx, i] : items) {
                i->setEq(_zero);
            }
        }
    }

    /**
     * Trigger removal of secondary items associated to the given primary item index
     * Removes all secondary items if no secondary_item_idx is given
     *
     * @param primary_item_idx Index of the associated primary item
     * @param secondary_item_idx Index of the associated secondary item (if any)
     */
    void remove(int primary_item_idx, std::optional<int> const& secondary_item_idx = std::nullopt)
    {
        auto it = _storage.find(primary_item_idx);
        if (it == std::cend(_storage)) {
            // No secondary items for this primary item
            return;
        }

        if (secondary_item_idx.has_value()) {
            auto item_it = it->second.find(secondary_item_idx.value());
            if (item_it != std::cend(it->second)) {
                _store->remove(item_it->second);
            }
        } else {
            auto& items = it->second;
            for (auto& [idx, i] : items) {
                _store->remove(i);
            }
        }
    }

    /**
     * Remove all secondary items that have not been updated and are inactive, i.e., item->changed == ItemProperty::None
     * @param primary_item_idx Primary item index
     */
    void remove_unchanged(int primary_item_idx)
    {
        auto it = _storage.find(primary_item_idx);
        if (it == std::cend(_storage)) {
            // No secondary items for this primary item
            return;
        }

        for (auto const& [idx, item] : it->second) {
            if (item->changed == ItemProperty::None) {
                _store->remove(item);
            }
        }
    }

    /**
     * Update secondary render-item associated with given primary, sec_idx and location
     * The secondary item is created if it does not exist
     * @param primary The primary (parent) render item
     * @param sec_idx Secondary index
     * @param location Location of the secondary source
     * @return Returns the render item
     */
    RenderItem* update_item(
        render_item_ptr const& primary, std::size_t sec_idx, std::array<float, 3> const& location)
    {
        if (!primary) {
            return nullptr;
        }

        auto item = get_item(primary, sec_idx);
        if (!item) {
            return nullptr;
        }

        sync(*item, primary);

        item->setPosition(
            { { location[0], location[1], location[2] }, primary->position.orientation });

        return item;
    }

    /**
     * Add all active secondary items to render list
     *
     * @param render_list The render list
     */
    void add_items(RenderList& render_list)
    {
        update();
        for (auto const& [idx, items] : _storage) {
            for (auto const& [idx, item] : items) {
                if (!item->isInactive() && !item->isOrphaned()) {
                    render_list.addItemCopy(item);
                    item->resetChanged();
                }
            }
        }
    }

    /**
     * Get the secondary render-items by sector
     *
     * @param sec_idx Sector index
     * @return Vector of all secondary (reflection) render-items in given sector
     */
    std::vector<RenderItem*> get_secondary_by_sector(std::size_t sec_idx) const
    {
        std::vector<RenderItem*> sec_items;
        for (auto const& [_, sec_item_map] : _storage) {
            auto const it = sec_item_map.find(sec_idx);
            if (std::cend(sec_item_map) != it) {
                if (-1 != it->second->parentRI
                    && !(it->second->isInactive() || it->second->isOrphaned())) {
                    sec_items.push_back(it->second);
                }
            }
        }

        return sec_items;
    }

    /**
     * Object is valid and safe to use
     *
     * @return True if valid
     */
    bool valid() const { return _store != nullptr; }

protected:
    T const& derived() const { return static_cast<T const&>(*this); }

    /**
     * Get the item type of the secondary items to store
     * Must be implemented by derived classes
     * @param primary The associated primary render item
     * @return ItemType Item type
     */
    ItemType get_item_type(RenderItem const& primary) const
    {
        return derived().get_item_type(primary);
    }

    /**
     * Get the properties to sync from the primary item
     * Must be implemented by derived classes
     * @param primary The associated primary render item
     * @return Sync-flags
     */
    ItemProperty get_sync_flags(RenderItem const& primary) const
    {
        return derived().get_sync_flags(primary);
    }

    /**
     * Get the suffix for the secondary item (debugging only)
     * Must be implemented by derived classes
     * @return The suffix
     */
    std::string get_suffix() const { return derived().get_suffix(); }

protected:
    /**
     * Update item store and remove obsolete items
     */
    void update()
    {
        // Remove all storage entries that do not have secondary items
        auto it{ std::begin(_storage) };
        while (it != std::end(_storage)) {
            // Remove all secondary items that are no longer used (orphaned)
            auto sec_it = std::begin(it->second);
            while (sec_it != std::end(it->second)) {
                if (sec_it->second->isOrphaned()) {
                    sec_it = it->second.erase(sec_it);
                } else {
                    ++sec_it;
                }
            }

            if (it->second.empty()) {
                it = _storage.erase(it);
            } else {
                ++it;
            }
        }
    }

    void sync(RenderItem& secondary, render_item_ptr const& primary) const
    {
        secondary.sync(*primary, get_sync_flags(*primary));
    }

    RenderItem* make_item(render_item_ptr const& primary, std::size_t sec_idx)
    {
        auto item =
            _store->createItem(primary->objectName + get_suffix() + std::to_string(sec_idx));
        if (!item) {
            return nullptr;
        }

        item->parentRI = primary->idx;
        item->type = get_item_type(*primary);
        return item;
    }

    RenderItem* get_item(render_item_ptr const& primary, std::size_t sec_idx)
    {
        auto& sec_items = _storage[primary->idx];
        auto sec_it = sec_items.find(sec_idx);
        if (sec_it != std::cend(sec_items)) {
            if (sec_it->second->isOrphaned()) {
                sec_items.erase(sec_it);
            } else {
                return sec_it->second;
            }
        }

        auto& item = sec_items[sec_idx];
        item = make_item(primary, sec_idx);
        return item;
    }

protected:
    RenderItemStore* _store; //!< Render-item store to create new items

    using storage_t = std::unordered_map<int, std::unordered_map<int, RenderItem*>>;
    storage_t _storage; //!< To store 1..N secondary render items per primary

private:
    friend T;
    secondary_item_storage(RenderItemStore* store, unsigned int num_bands_eq)
        : _store(store)
        , _zero(num_bands_eq, 0.f)
    {}

    std::vector<float> const _zero; //!< Helper to mute the render-item by EQ
};

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
