/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPECULARREFLECTION
#define MPEGI_SPECULARREFLECTION

#include <EAR/BaseTypes.h>

#include <atomic>
#include <ostream>
#include <vector>

namespace MPEGI {

//! Represents a propagation path for a specular reflection.
class SpecularReflection
{

public:
    SpecularReflection(){};

    SpecularReflection(float ld, float lg, const EAR::Vector3& simg, int i,
        const std::vector<int>& s, const std::vector<int>& t, bool o,
        const std::vector<EAR::Vector3>& mp, long long int v, EAR::HomogenMatrix tform_,
        EAR::HomogenMatrix tformInv_, std::vector<float> eq_)
        : path_length_delay(ld)
        , path_length_gain(lg)
        , S_img(simg)
        , renderItemID(i)
        , surfaceSeq(s)
        , tform(tform_)
        , tformInv(tformInv_)
        , triangleSeq(t)
        , is_occluded(o)
        , mirrorPoints(mp)
        , value(v)
        , eq(eq_){};

    //! Explicit copy constructor
    SpecularReflection(const SpecularReflection& other) = default;

    //! Get sound propagation path length.
    /*!
     * \return      propagation path length in m
     */
    float getPathLengthDelay() const { return path_length_delay; }

    //! Set sound propagation path length.
    /*!
     * \param l     propagation path length in m
     */
    void setPathLengthDelay(float l) { path_length_delay = l; }

    //! Get apparent distance.
    /*!
     * \return      apparent distance in m
     */
    float getPathLengthGain() const { return path_length_gain; }

    //! Set apparent distance.
    /*!
     * \param l     apparent distance in m
     */
    void setPathLengthGain(float l) { path_length_gain = l; }

    //! Get image source location.
    /*!
     * \return      image source location
     */
    const EAR::Vector3& getImageSource() const { return S_img; }

    //! Set image source location.
    /*!
     * \param S     image source location
     */
    void setImageSource(const EAR::Vector3& S) { S_img = S; }

    //! Get render item ID.
    /*!
     * \return      unique render item ID
     */
    int getRenderItemID() const { return renderItemID; }

    //! Set render item ID.
    /*!
     * \param id    unique render item ID
     */
    void setRenderItemID(int id) { renderItemID = id; }

    //! Get sequence of surfaces on which the sound was reflected (from listener to source).
    /*!
     * \return      list of surface indices
     */
    const std::vector<int>& getSurfaceSeq() const { return surfaceSeq; }

    //! Set sequence of surfaces on which the sound was reflected (from listener to source).
    /*!
     * \param seq   list of surface indices
     */
    void setSurfaceSeq(const std::vector<int>& seq, int num_faces)
    {
        surfaceSeq = seq;
        value = 0;

        for (auto&& surfID : seq) {
            value = value * (num_faces + 1) + surfID + 1;
        }
    }

    //! Get sequence of triangles on which the sound was reflected (from listener to source).
    /*!
     * \return      list of triangle indices
     */
    const std::vector<int>& getTriangleSeq() const { return triangleSeq; }

    //! Set sequence of triangles on which the sound was reflected (from listener to source).
    /*!
     * \param seq   list of triangle indices
     */
    void setTriangleSeq(const std::vector<int>& seq) { triangleSeq = seq; }

    //! Get the occlusion flag.
    /*!
     * \return      false: visible; true: occluded
     */
    bool getIsOccluded() const { return is_occluded; }

    //! Set the occlusion flag.
    /*!
     * \param o     false: visible; true: occluded
     */
    void setIsOccluded(bool o) { is_occluded = o; }

    //! Get sequence of reflection points on the wall surfaces (from listener to source).
    /*!
     * \return      list of reflection points
     */
    const std::vector<EAR::Vector3>& getMirrorPoints() const { return mirrorPoints; }

    //! Set sequence of reflection points on the wall surfaces (from listener to source).
    /*!
     * \param seq   list of reflection points
     */
    void setMirrorPoints(const std::vector<EAR::Vector3>& mp) { mirrorPoints = mp; }

    //! Get the numerical value which is used to sort specular reflections.
    /*!
     * \return      sorting value
     */
    long long getValue() const { return value; }

    //! Set the numerical value which is used to sort specular reflections.
    /*!
     * \param v     sorting value
     */
    void setValue(long long v) { value = v; }

    //! Get the occlusion status of the Reflection
    bool isOccluded() const { return is_occluded; }

    //! Get the reflection matrix
    const EAR::HomogenMatrix getTransformationMatrix() const { return tform; }

    //! Get the inverse of reflection matrix
    const EAR::HomogenMatrix getInverseOfTransformationMatrix() const { return tformInv; }

    //! Get eq values
    const std::vector<float>& getEq() const { return eq; }

    //! numerical comparison according to the objects sorting value
    inline bool operator<(const SpecularReflection& rhs) const { return value < rhs.value; }
    inline bool operator>(const SpecularReflection& rhs) const { return value > rhs.value; }
    inline bool operator<=(const SpecularReflection& rhs) const { return value <= rhs.value; }
    inline bool operator>=(const SpecularReflection& rhs) const { return value >= rhs.value; }
    inline bool operator==(const SpecularReflection& rhs) const { return value == rhs.value; }
    inline bool operator!=(const SpecularReflection& rhs) const { return value != rhs.value; }

    friend std::ostream& operator<<(std::ostream& os, const SpecularReflection& ref)
    {
        using EAR::operator<<;
        return os << "SpecularReflection(" << std::endl
                  << "    path_length_delay = " << ref.path_length_delay << std::endl
                  << "    path_length_gain  = " << ref.path_length_gain << std::endl
                  << "    S_img             = " << ref.S_img << std::endl
                  << "    renderItemID      = " << ref.renderItemID << std::endl
                  << "    surfaceSeq        = " << ref.surfaceSeq << std::endl
                  << "    tform             = " << ref.tform << std::endl
                  << "    tformInv          = " << ref.tformInv << std::endl
                  << "    triangleSeq       = " << ref.triangleSeq << std::endl
                  << "    is_occluded       = " << ref.is_occluded << std::endl
                  << "    mirrorPoints      = " << ref.mirrorPoints << std::endl
                  << "    value             = " << ref.value << std::endl
                  << "    eq                = " << ref.eq << std::endl
                  << ")   // SpecularReflection";
    }

private:
    //! propagation path length in m
    float path_length_delay{ 1.f };

    //! apparent distance in m
    float path_length_gain{ 1.f };

    //! image source location
    EAR::Vector3 S_img;

    //! unique ID for the render item
    int renderItemID{ -1 };

    //! sequence of surfaces on which the sound was reflected (from listener to source)
    std::vector<int> surfaceSeq;

    //! position and orientation of the image source given as HomogenMatrix
    EAR::HomogenMatrix tform;

    //! Inverse of tfrom matrix
    EAR::HomogenMatrix tformInv;

    //! sequence of triangles on which the sound was reflected (from listener to source)
    std::vector<int> triangleSeq;

    //! false: visible; true: occluded
    bool is_occluded;

    //! list of reflection points on the wall surfaces
    std::vector<EAR::Vector3> mirrorPoints;

    //! sorting value
    long long int value{ -1 };

    //! eq values
    std::vector<float> eq;
};

} // namespace MPEGI

#endif /* MPEGI_SPECULARREFLECTION */
