/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPEEDINGTICKET
#define MPEGI_SPEEDINGTICKET

#include <array>

// #include <cmath>

namespace MPEGI {

#define SPEEDING_HISTORY_LENGTH 10
#define SPEEDING_GAIN_INCREMENT 0.1f

//! Data class for speeding tickets.
/*!
 * This class keeps records of exceeded speed limits with hysteresis.
 */
class SpeedingTicket
{
public:
    //! Default constructor: start in state "speed limit exceeded"
    SpeedingTicket()
        : pos(0)
        , sum(SPEEDING_HISTORY_LENGTH)
        , is_exceeded(true)
        , speed_limit(0.f)
        , gain(0.f)
    {
        history.fill(1);
    };

    //! Constructor
    /*!
     * \param speed             input: current speed
     * \param _speed_limit      input: speed limit
     */
    SpeedingTicket(float speed, float _speed_limit)
        : pos(0)
        , sum((speed > _speed_limit) ? SPEEDING_HISTORY_LENGTH : 0)
        , is_exceeded(speed > _speed_limit)
        , speed_limit(_speed_limit)
        , gain((speed > _speed_limit) ? 0.f : 1.f)
    {
        history.fill((speed > _speed_limit) ? 1 : 0);
    };

    //! Checks if the speed limit is exceeded incl. hysteresis and adds the result to the records.
    /*!
     * \param speed             input: current speed
     * \return                  false: speed limit not exceeded; true: speed limit exceeded
     */
    bool checkSpeed(float speed);

    //! Returns the speed specific render item gain
    /*!
     * \return                  linear gain
     */
    inline float getGain() const { return gain; }

    //! Adjust the speed limit
    /*!
     * \param _speed_limit      speed limit in m/s
     */
    inline void setSpeedLimit(float _speed_limit) { speed_limit = _speed_limit; }

private:
    //! maximum reflection order
    int pos;

    //! sum over all track records
    int sum;

    //! current state
    bool is_exceeded;

    //! speed limit
    float speed_limit;

    //! speed specific render item gain
    float gain;

    //! ring buffer for speeding history
    std::array<int, SPEEDING_HISTORY_LENGTH> history;
};

} // namespace MPEGI

#endif /* MPEGI_SPEEDINGTICKET */
