/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_TAPE_DECK
#define MPEGI_TAPE_DECK

#include <limits>
#include <map>
#include <mutex>
#include <vector>

namespace MPEGI {

//! TapeDeck pointer playback
/*!
 * A TapeDeck allows to register pointers to elements with a corresponding timestamp.
 * When a playback with another timestamp is requested, it returns all elements between the previous and the current playback call.
 *
 * The class does not handle the memory of the elements.
 */
template <class T> class TapeDeck
{
public:
    //! Constructor
    TapeDeck()
        : lt(-1 * std::numeric_limits<double>::infinity()){};

    //! Destructor
    ~TapeDeck(){};

    //! Add an element to the playback (thread-safe)
    /*!
     * Will only be played back if t is bigger than the previous playback t. (Negative ts will therefore be played back at playback(0)...)
     *
     * \param t     timestamp [s]
     * \param el    pointer to element (template type)
     */
    void addElement(double t, T* el)
    {
        const std::lock_guard<std::mutex> lock(elementsMutex);
        elements.insert({ t, el });
    };

    //! Return all elements that are between this and the previous playback call (thread-safe)
    std::vector<T*> playback(double t)
    {
        const std::lock_guard<std::mutex> lock(elementsMutex);
        std::vector<T*> result;

        auto lower = elements.lower_bound(lt);
        auto upper = elements.upper_bound(t);

        if (lower != upper) {
            for (auto it = lower; it != upper; it++) {
                result.push_back(it->second);
            }
        }

        lt = t;

        return result;
    };

    //! Reset all elements to be unplayed
    void resetPlayback() { lt = -1 * std::numeric_limits<double>::infinity(); };

private:
    double lt;
    std::multimap<double, T*> elements;

    std::mutex elementsMutex;
};

} // namespace MPEGI

#endif /* MPEGI_TAPE_DECK */
