/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_TYPES
#define MPEGI_TYPES

#include <EAR/Geometry.h>
#include <EAR/PchipInterpolator.h>
#include <array>
#include <map>
#include <optional>
#include <set>
#include <string>
#include <unordered_set>
#include <vector>

namespace EAR {

class HOASource;
class Portal;

} // namespace EAR

namespace MPEGI {

//! Indicator for different stages/conditions in the lifetime of the RenderItem
enum class ItemStatus
{
    //! This RenderItem appears for the first time
    /*!
     * New resources have to be allocated to incorporate this RI in the DSP processing.
     * A potential fade in is handled by the Spatializer.
     */
    New,

    //! This RenderItem is active
    /*!
     * A potential fade in if the previous state was Inactive is handled by the Spatializer.
     */
    Active,

    //! This RenderItem is fading out to be temporarily deactivated
    /*!
     * After a fade out handled by the Spatializer, the state will be set to Inactive.
     * No changes shall be made to the RI metadata, but the DSP processing shall continue.
     */
    FadeOutInactive,

    //! This RenderItem is temporarily deactivated
    /*!
     * It shall not be rendered now, but may return to the Active state later.
     * Resources needed for DSP processing should be kept to be reused.
     */
    Inactive,

    //! This RenderItem is fading out to be removed
    /*!
     * After a fade out handled by the Spatializer, the state will be set to Orphaned.
     * No changes shall be made to the RI metadata, but the DSP processing shall continue.
     */
    FadeOutOrphaned,

    //! This RenderItem is no longer needed
    /*!
     * It will be ultimately removed in the next control step.
     * A potential fade out is already complete (see state FadeOutOrphaned) and all resources can be deleted immediately.
     */
    Orphaned
};

//! Separates different types of RenderItem
enum class ItemType
{
    Primary, //!< Directly derived from a EAR::SceneObject
    HOA, //!< An HOA RenderItem
    Representative, //!< A representative RI derived from another RI (e.g. a cluster or an extended source representative)
    Reflection, //!< A secondary RI derived as a specular reflection of another RI
    Diffraction, //!< A secondary RI derived as a geometrically diffracted path of another RI
    DirectionalReverb, //!< A RI for reverb from different acoustic environments
    Nondiegetic, //!< A RI that is spatialized, but not part of the scene
    Downmix, //!< RI that is to be mixed directly to the output without additional processing
    Pannable, //!< RI that is to be panned into a virtual loudspeaker layout that is going to be spatialized
    Reverb, //!< RI that is generated as a reverb output for a reverb environment
    Airflow, //!< RI that is generated as a simulated airflow noise output
    ListenerVoice, //!< RI used to tune the user's voice
};

//! Separates different sub-types of RenderItem
enum class ItemSubtype
{
    None, //!< Indicates that the subtype has not been configured for the RI
    CoupledSrc, //!< RI that is generated from a portal extent representing a coupled source.
};

//! Separates different properties a RenderItem can have (e.g. for indicating that they changed)
/*!
 * Multiple changes can be concatenated with the bitwise | operator and checked for with bitwise &
 */
enum class ItemProperty
{
    None = 0, //!< indicator for no property (e.g. "no property has changed")
    Status = 1, //!< the RI status changed
    Position = 2, //!< the position (including Location, Orientation and/or Trajectory) changed
    Aparams = 4, //!< the aparams changed
    Directivity = 8, //!< the directivity pointer changed
    Extent = 16, //!< the extent geometry pointer changed
    Signal = 32, //!< the signal pointer changed (not the samples in the buffer itself)
    Gain = 64, //!< the gain value changed
    Eq = 128, //!< one or more eq values changed
    RefDist = 256, //!< reference distance changed
    Orientation = 512, //!< fine-granular version of position property
    Location = 1024, //!< fine-granular version of position property
    ParentRI = 2048, //!< parent RenderItem index
    Reverb = 4096, //!< room-related properties
    ReflectionInfo = 8192, //!< reflection information
    RayHits = 16384, //!< change in discovery rays hitting an extended source
    VDLMethod = 32768, //!< vdl method changed
    ReverbGain = 65536, //!< the reverb gain value changed
    NoReverb = 131072,
    Loudness = (1 << 18), // !< the loudness value changed
    GranularPos = (1 << 19), //!< granular position changed
    GranularGroupProb = (1 << 20), //!< granular group probability changed
    GranularDatabaseId = (1 << 21), //!< granular group probability changed
};

//! Auralization-related configuration for the specific RenderItem
enum class AParam
{
    None = 0,
    NoDoppler = 1,
    NoDistanceGain = 2,
    NoMediumAttenuation = 4,
    ForceNoDoppler = 8,
    noDirectionalFocus = 16,
    NoCulling = (1 << 5)
};

//! Separates different properties a Portal can have
enum class PortalType
{
    //! this portal creates a reverb-only downmix, no direct sound will be included
    ReverbOnly = 0,
    // //! this portal creates a downmix that includes all audio sources in the source cell,
    // /*!
    //  * The levels are independent of their distance to the portal.
    //  */
    // LocIndependent,
    // //! this portal creates a downmix that includes all audio sources in the source cell,
    // /*!
    //  * The relative levels between them are weighted based on their distance to the portal.
    //  */
    // LocDependent,
};

struct ItemPropertyProxy
{
    constexpr operator ItemProperty() { return this->prop; }
    constexpr operator bool() { return static_cast<bool>(this->prop); }
    ItemProperty prop;
};

//! Check if an ItemProperty is set in a concatenation of multiple properties
ItemPropertyProxy operator&(ItemProperty lhs, ItemProperty rhs);

//! Concatenate two ItemProperties
ItemProperty operator|(ItemProperty lhs, ItemProperty rhs);

//! Invert an ItemProperty
ItemProperty operator~(ItemProperty prop);

//! Add an ItemProperty to an existing concatenation
ItemProperty& operator|=(ItemProperty& lhs, ItemProperty rhs);

//! Bitwise and
ItemPropertyProxy operator&=(ItemProperty& lhs, ItemProperty rhs);

//! Check if an AParam is set in a concatenation of multiple properties
bool operator&(AParam lhs, AParam rhs);

//! Concatenate two AParams
AParam operator|(AParam lhs, AParam rhs);

//! Add an AParam to an existing concatenation
AParam& operator|=(AParam& lhs, AParam rhs);

//! Bitwise and
AParam& operator&=(AParam& lhs, AParam rhs);

//! A Reverb Environment defines a space where one FDN reverb is set up
struct ReverbEnv
{
    struct FrequencyData
    {
        float frequency;
    };

    struct FrequencyDecay : public FrequencyData
    {
        float rt60;
        std::optional<float> rdr;
    };

    struct FrequencyGain : public FrequencyData
    {
        float gain;
    };

    struct DirectionalGain
    {
        float azim;
        float elev;
        std::vector<FrequencyGain> frequencyResponse;
    };

    //! Unique identifier of this Reverb Environment (name of the acoustic environment)
    const std::string acousticEnvironmentId;

    //! The associated acoustic environment's bitstream id
    std::optional<int> acousticEnvironmentBsId;

    bool isSceneBasedAE;

    //! A geometry that describes the outer boundary of this space
    /*!
     * May be nullptr if the Reverb Environment extents to infinity
     */
    const EAR::Geometry* enclosure;

    //! List of all geometries that live inside this environment and potentially influence the sound field
    // std::vector<const EAR::Geometry*> geometriesInside;

    //! Pre-delay of the reverb [sec]. Defaults to geometrix approximation. In an AR scene context, this is the same as diffuseOnset.
    std::optional<float> predelay;

    //! Holds the Config predelayOffset if specified [sec]
    float predelayOffset = 0.f;

    //! ComplexityLevel of the reverb, default 0 (15-channel reverb)
    std::optional<int> complexityLevel;

    //! Number of delay lines used for this reverb
    std::optional<int> numDelayLines;

    //! Frequency dependent data
    std::vector<FrequencyDecay> frequencyDecayResponse;

    //! ReverbEnv originally authored in LSDF
    bool isLsdf;

    //! ReverbEnv authored with a RoomImpulseResponse in the LSDF
    bool isFromRIR;

    //! RoomImpulseResponse in the LSDF specifies the source and receiver locations for direct sound distance calc
    bool srcDistanceIsKnown;

    //! Add a new supporting point for frequency-dependent characteristics of the reverb environment
    /*!
     * \param f         Frequency value in Hz of the new supporting point
     * \param rt        RT60 in seconds at the given frequency
     * \param rdr       Reverberant-to-direct ratio (RDR) at the given frequency (optional)
     */
    void append(float f, float rt, std::optional<float> rdr);

    //! Return the RT60 in s for any given frequency
    /*!
     * TODO: deprecated and currently only used for 3DReverb
     *
     * \param f         frequency in Hz
     * \return          interpolated RT60 for the given frequency value
     */
    inline float getRtForFreq(float f) { return rtInterpolator->interpolate(f); }

    //! Constructor
    /*!
     * A unique reverb ID is assigned automatically.
     *
     * \param acousticEnvironmentId_  Unique identifier of this Reverb Environment (name of acoustic environment)
     * \param enclosure_              The outer boundary of this space
     */
    ReverbEnv(const std::string& acousticEnvironmentId_, const EAR::Geometry* enclosure_)
        : acousticEnvironmentId(acousticEnvironmentId_)
        , enclosure(enclosure_)
        , isLsdf(false)
        , isFromRIR(false)
        , srcDistanceIsKnown(false)
        , rtInterpolator(std::make_unique<EAR::PchipInterpolator>(EAR::PchipInterpolator(true))){};

private:
    // TODO: interpolator deprecated and currently only used for 3DReverb
    std::unique_ptr<EAR::PchipInterpolator> rtInterpolator;
};

//! A cell defines a smaller rendering unit inside the scene
struct Cell
{
    //! Unique identifier of this Cell
    const int cellId;

    //! Human-readable name (for debugging)
    const std::string cellName;

    //! A geometry that describes the outer boundary of the cell
    const EAR::Geometry* enclosure;

    //! the pointer to the reverb instance for current cell
    const ReverbEnv* reverbPtr;

    //! List of all geometries that live inside this cell and potentially influence the sound field
    std::vector<const EAR::Geometry*> geometriesInCell;

    //! List of all relevent audio sources for this cell
    std::vector<const EAR::SceneObject*> relevantSources;

    //! List of all visuble diffraction edges in this cell
    std::vector<int> visibleEdges;

    //! map of the relevant reflection surfaces when the listener is in this cell
    /*!
     * key : the id of the cell, in which the source is
     * value : the list of relevant surfaces
     */
    std::map<int, std::vector<std::vector<int>>> relevantSurfaces;

    Cell(const int id, const std::string name_)
        : cellName(name_)
        , cellId(id){};

private:
    //static int cellCounter;
};

//! Division of the scene to voxels that have line of sight data
struct LineOfSightArea
{
    std::array<int, 3> nVoxels;
    EAR::Vector3 losAreaCornerP1;
    EAR::Vector3 losAreaCornerP2;
    unsigned int voxelSize; // centimeters
    std::map<unsigned int, std::set<unsigned int>> voxels; // key: portalId
};

//! Intersection planes are used to ensure that a ray intersects with a specified plane.
struct IntersectPlane
{
    EAR::Vector3 planeCenter;
    EAR::Vector3 planeNormal;
    EAR::Vector3 lowerLeftCorner;
    EAR::Vector3 lowerRightCorner;
    EAR::Vector3 upperLeftCorner;
    EAR::Vector3 upperRightCorner;
    bool hasBounds;
    EAR::Vector3 u;
    EAR::Vector3 v;
    float uDotLL;
    float uDotLR;
    float vDotLL;
    float vDotUL;

    IntersectPlane(){};

    IntersectPlane(EAR::Vector3 _Center, EAR::Vector3 _Normal,
        EAR::Vector3 lowerLeftCorner = EAR::Vector3(),
        EAR::Vector3 lowerRightCorner = EAR::Vector3(),
        EAR::Vector3 upperLeftCorner = EAR::Vector3(),
        EAR::Vector3 upperRightCorner = EAR::Vector3(), bool _hasbounds = false)
        : planeCenter(_Center)
        , planeNormal(_Normal)
        , hasBounds(_hasbounds)
    {}

    void setBounds(EAR::Vector3 _lowerLeft, EAR::Vector3 _lowerRight, EAR::Vector3 _upperLeft,
        EAR::Vector3 _upperRight)
    {
        lowerLeftCorner = _lowerLeft;
        lowerRightCorner = _lowerRight;
        upperLeftCorner = _upperLeft;
        upperRightCorner = _upperRight;
        hasBounds = true;
        u = lowerRightCorner - lowerLeftCorner;
        v = upperLeftCorner - lowerLeftCorner;
        uDotLL = u.dot(lowerLeftCorner);
        uDotLR = u.dot(lowerRightCorner);
        vDotLL = v.dot(lowerLeftCorner);
        vDotUL = v.dot(upperLeftCorner);
    }
};
//! a Portal allows sound to transmit between Reverb Environments
struct Portal
{
    //! Unique id of this portal
    int portalId;

    std::string portalStringId;

    //! the id of the geometry that describes the extent of this portal
    std::optional<std::string> extentId;

    //! the id of the source environment this portal creates downmix from (fallback mode)
    std::string sourceId;

    //! the ids of the Acoustic Environments connected to the portal. Entry 0 corresponds to the AE in the direction of the normalVector, entry 1 to the opposite direction.
    std::array<std::string, 2> portalEnvironments;

    //! the type of this portal (location dependent, location independent or reverb only)
    PortalType type;

    //! material associated with the ID for calculation of energy passing through the portal. If no material was set, this is an empty string.
    //! NOTE: Users should use scenePortal->getMaterial/setMaterial instead, as that may be updated by the Scene mechanism.
    //! This value only contains the initial materialId as read from the bitstream, and is not updated.
    std::string materialId;

    //! position (location without orientation) of the portal
    EAR::Vector3 position;

    //! Equivalent size of the portal (Width, Height, Depth)
    std::optional<EAR::Vector3> equivalentPortalSize;

    //! Equivalent Box pointer
    std::unique_ptr<EAR::Box> equivalentBoxPointer = nullptr;

    //! normal vector of the portal.
    EAR::Vector3 normalVector;

    //! Pointer to the portal in the SceneState. This part can receive updates.
    EAR::Portal* scenePortal = nullptr;

    //! The portal height calculated from the bounding box of the portal extent perpendicular to the portal normal.
    float bbPortalHeight = 0.0f;

    //! The portal width calculated from the bounding box of the portal extent perpendicular to the portal normal.
    float bbPortalWidth = 0.0f;

    std::map<std::string, MPEGI::IntersectPlane> portalPlanes;
};

//! ReflectionInfo holds informations that may be relevant to subsequently handle reflected RenderItems
struct ReflectionInfo
{
    //! The points at which the sound is specularly reflected on geometry
    /*!
     * The first element is the first reflection as seen from the listener, last element is the first reflection as seen from the source.
     * The size of the vector corresponds to the order of reflection.
     */
    std::vector<EAR::Vector3> intersectionSequence;
};

//! Holds informations that may be relevant to subsequently handle occluded RenderItems
struct OcclusionInfo
{
    std::vector<std::shared_ptr<EAR::Material>> materials;
    bool operator<(const OcclusionInfo& other) const
    {
        if (materials.size() == other.materials.size()) {
            for (int i = 0; i < materials.size(); i++) {
                if (materials[i]->getBsId() < other.materials[i]->getBsId())
                    return true;
                else if (materials[i]->getBsId() > other.materials[i]->getBsId())
                    return false;
            }
            return false;
        } else {
            return materials.size() < other.materials.size();
        }
    }
};

//! The InputLayout gives hints for the positioning of the audio stream channels w.r.t. spatial extent of the source
struct ExtentLayout
{
    //! The alignment specifies the reference frame of the positioning
    enum class Alignment
    {
        User = 0, //!< Positionings are relative to the user's viewing direction
        Object //!< Positionings are fixed to the position/orientation of the ObjectSource
    };

    //! Positioning Mnemonics (see EIF specification for further details)
    enum class Positioning
    {
        TL = 0, //!< top left
        T, //!< top
        TR, //!< top right
        L, //!< left
        C, //!< center
        R, //!< right
        BL, //!< bottom left
        B, //!< bottom
        BR //!< bottom right
    };

    Alignment alignment; //!< Anchoring mode (see InputLayout::Alignment)
    std::vector<Positioning> positioning; //!< Positioning of the channels from the audio stream
};

//! A 3d int vector representation that supports basic algebraic expressions (+ and -).
/*!
 * In EAR, they are mainly used for voxel coordinates.
 *
 */
class VoxelCoord
{

public:
    //! Empty Constructor (zero-initializing)
    VoxelCoord() { set(0, 0, 0); };

    //! Constructor to initialize from single component values
    /*!
     * \param _x   first component
     * \param _y   second component
     * \param _z   third component
     */
    VoxelCoord(int _x, int _y, int _z) { set(_x, _y, _z); };

    //! Constructor to initialize from int array of length 3
    /*!
     * \param _val   int array of length 3
     */
    VoxelCoord(int _val[3]) { set(_val); };

    //! Copy constructor
    VoxelCoord(const VoxelCoord& rhs) = default;

    //! Destructor
    ~VoxelCoord(){};

    //! Set all components from outside
    /*!
     * \param _x   first component
     * \param _y   second component
     * \param _z   third component
     */
    void set(int _x, int _y, int _z);

    //! Set all components from outside
    /*!
     * \param _val   int array of length 3
     */
    void set(int _val[3]);

    /*VoxelCoord& operator=(const VoxelCoord& other); // copy
    VoxelCoord& operator=(VoxelCoord&& other); // move*/

    //! perform vector addition
    /*!
     * \param rhs   other VoxelCoord to be added
     */
    VoxelCoord& operator+=(const VoxelCoord& rhs);

    //! perform vector subtraction
    /*!
     * \param rhs   other VoxelCoord to be subtracted
     */
    VoxelCoord& operator-=(const VoxelCoord& rhs);

    //! direct access to components
    /*!
     * \param idx   index (0, 1, 2)
     */
    inline int& operator[](int idx) { return val[idx]; };

    //! direct const access to components
    /*!
     * \param idx   index (0, 1, 2)
     */
    inline int operator[](int idx) const { return val[idx]; };

    friend VoxelCoord operator+(VoxelCoord lhs, const VoxelCoord& rhs);
    friend VoxelCoord operator-(VoxelCoord lhs, const VoxelCoord& rhs);
    friend VoxelCoord operator-(const VoxelCoord& rhs);

    //! get raw pointer to underlying data
    inline int* data() { return &val[0]; }

    //! get const raw pointer to underlying data
    inline const int* data() const { return &val[0]; }

private:
    //! internal data store
    int val[3];
};

//! Holds information about the configuration for a HOA RenderItem
struct HoaInfo
{
    //! Affiliation to a group (-1 => no group)
    int group = -1;

    //! Whether the item should be internally or externally rendered (true => external)
    bool external = false;

    //! Whether the item should change the internal/external representation depending on the position
    /*!
     * If this is set to true and the listener is inside the extent geometry, the HOA item is always rendered internally.
     * Conversely, outside of the extent geometry, the HOA item is always rendered externally.
     */
    bool extentTransform = false;

    //! Width of the transition atrea when there is extentTransform
    float transitionDist = 1;

    //! The source the info is associated with
    EAR::HOASource* source{ nullptr };
};

//! Represents an informed audio source in MPHOA rendering
struct InformedSource
{
    InformedSource(EAR::Position position, bool listenerThresholdPresent,
        bool hoaSourceThresholdPresent, float listenerThreshold, float hoaSourceThreshold,
        int disableInformedSourceType, float disableInformedSourceThreshold, int priorityValue)
        : position{ position }
        , listenerThresholdPresent{ listenerThresholdPresent }
        , hoaSourceThresholdPresent{ hoaSourceThresholdPresent }
        , listenerThreshold{ listenerThreshold }
        , hoaSourceThreshold{ hoaSourceThreshold }
        , disableInformedSourceType{ disableInformedSourceType }
        , disableInformedSourceThreshold{ disableInformedSourceThreshold }
        , priorityValue{ priorityValue }
    {}

    //! The position of this source in space
    EAR::Position position;

    //! Turns on/off the threshold distance from the listener to the informed source
    bool listenerThresholdPresent;

    //! Turns on/off the threshold distance from HOA sources to the informed source
    bool hoaSourceThresholdPresent;

    //! Distance from the listener to the informed source where the informed source is active
    float listenerThreshold;

    //! Distance from the listener to HOA sources where the informed source is active
    float hoaSourceThreshold;

    //! Criteria used to disable active informed sources in the rendering
    int disableInformedSourceType;

    //! Distance threshold after the informed source is inactive
    float disableInformedSourceThreshold;

    //! Determines the processing priority between informed sources that are close to each other
    int priorityValue;
};

//! Represents a group of audio objects with high order ambisonics.
struct HOAGroup
{

    HOAGroup(EAR::Geometry* region, std::vector<EAR::HOASource*> const&& sources,
        std::vector<int> const&& complexityReductionParameters, int maxSimulInformedSources,
        std::vector<MPEGI::InformedSource> const&& informedSources, float distanceFactor,
        int panningGridOrder, float exteriorProjectionRadius, bool adaptiveExteriorProjectionRadius)
        : region{ region }
        , sources{ std::move(sources) }
        , complexityReductionParameters{ std::move(complexityReductionParameters) }
        , maxSimulInformedSources{ maxSimulInformedSources }
        , informedSources{ std::move(informedSources) }
        , distanceFactor{ distanceFactor }
        , panningGridOrder{ panningGridOrder }
        , exteriorProjectionRadius{ exteriorProjectionRadius }
        , adaptiveExteriorProjectionRadius{ adaptiveExteriorProjectionRadius }
    {}

    //! A geometry that describes the region of the group
    /*!
     * May be nullptr if the group has no region.
     */
    EAR::Geometry* region;

    //! The HOA sources belonging to the group
    std::vector<EAR::HOASource*> sources;

    //! Parameters used to reduce the complexity of MPHOA rendering
    std::vector<int> complexityReductionParameters;

    //! Maximum number of simultaneous active informed sources
    int maxSimulInformedSources = 10;

    //! Informed sources in the group (i.e. the known audio sources)
    std::vector<MPEGI::InformedSource> informedSources;

    //! Low complexity MPHOA renderer parameters
    float distanceFactor = 1.0f;
    int panningGridOrder = 0;

    //! Exterior rendering parameters
    float exteriorProjectionRadius;
    bool adaptiveExteriorProjectionRadius;
};

class AudioplusData
{
public:
    bool distanceGainModel;
    bool hasDiffuseness;
    std::optional<float> diffuseness;

    //! Empty Constructor (zero-initializing)
    AudioplusData()
    {
        distanceGainModel = 0;
        hasDiffuseness = 0;
    };

    AudioplusData(bool argDistanceGainModel, bool argHasDiffuseness)
    {
        distanceGainModel = argDistanceGainModel;
        hasDiffuseness = argHasDiffuseness;
    };

    AudioplusData(bool argDistanceGainModel, bool argHasDiffuseness, float argDiffuseness)
    {
        distanceGainModel = argDistanceGainModel;
        hasDiffuseness = argHasDiffuseness;
        diffuseness = argDiffuseness;
    };

    //! Destructor
    ~AudioplusData(){};
};

struct RayHit
{
    RayHit(unsigned int id, float dist, float w)
        : rayID(id)
        , refinementMask((unsigned int)-1)
        , distance(dist)
        , weight(w)
    {}
    unsigned int rayID;
    unsigned int refinementMask;
    float distance;
    float weight;
};
inline bool operator==(const RayHit& lhs, const RayHit& rhs)
{
    return lhs.rayID == rhs.rayID && lhs.distance == rhs.distance && lhs.weight == rhs.weight;
}

} // namespace MPEGI

#endif /* MPEGI_TYPES */
