/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "MPEGI/Helpers/VoxelTypes.h"

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {
//--------------------------------------------------------------------------------------------------
/**
 * Traverse the voxel-grid
 */
struct voxel_traversal
{
    /// Traversal predicate function. Return false to stop traversal
    using traversal_pred_t = std::function<bool(voxel_index_t const&)>;

    voxel_traversal() = default;
    virtual ~voxel_traversal() = default;
    voxel_traversal(voxel_traversal const&) = delete;
    voxel_traversal& operator=(voxel_traversal const&) = delete;

    /**
     * Reserve path memory to avoid dynamic allocations during runtime
     *
     * @param length Path length
     */
    void reserve(std::size_t length) { _path.reserve(length); }

    /**
     * Method using already converted/calculated coordinates and voxel indices
     * @todo Coordinate system harmonization. This method does NOT use OpenGL coordinate system
     *
     * @param scene_info Basic scene information
     * @param begin_xyz_vox Traversal start-point (x,y,z)
     * @param begin_idx Traversal start-point voxel indices
     * @param end_xyz_vox Traversal end-point (x,y,z)
     * @param end_idx Traversal end-point voxel indices
     * @param pred Optional predicate function to stop traversal early
     * @return True if a path was found, else false, i.e., !get_path().empty()
     */
    bool run(Voxel::voxel_scene_info_basic const& scene_info,
        std::array<float, 3> const& begin_xyz_vox, voxel_index_t const& begin_idx,
        std::array<float, 3> const& end_xyz_vox, voxel_index_t const& end_idx,
        traversal_pred_t const& pred = traversal_pred_t());

    /**
     * Method using raw coordinates
     *
     * @param scene_info Basic scene information
     * @param begin_xyz Traversal start-point (x,y,z, OpenGL)
     * @param end_xyz Traversal end-point (x,y,z, OpenGL)
     * @param pred Optional predicate function to stop traversal early
     * @return True if a path was found, else false, i.e., !get_path().empty()
     */
    bool run(Voxel::voxel_scene_info_basic const& scene_info, EAR::Vector3 const& begin_xyz,
        EAR::Vector3 const& end_xyz, traversal_pred_t const& pred = traversal_pred_t());

    /**
     * Get Object-to-listener path
     *
     * @return Traversed path (voxel indices)
     */
    voxel_traversal_path_t const& get_path() const { return _path; }

protected:
    voxel_traversal_path_t _path; //!< Object-to-listener path (voxel-indices)
};
//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------