/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include <algorithm>
#include <array>
#include <chrono>
#include <functional>
#include <map>
#include <memory>
#include <numeric>
#include <optional>
#include <string>
#include <unordered_map>
#include <valarray>
#include <vector>

#include "EAR/SceneObjectObserver.h"

#ifdef VOXEL_DEBUG
#define PM_DEBUG
#define PF_DEBUG
#define GRID_DEBUG
#endif

//--------------------------------------------------------------------------------------------------
class cPathFinder;
namespace EAR {
class Vector3;
}

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
struct RenderItem;
class RenderItemStore;
class VoxDataPayload;

//--------------------------------------------------------------------------------------------------
namespace Voxel {
//--------------------------------------------------------------------------------------------------
namespace Payload {
template <int N> struct block;
struct subscene;
struct scene_update;
struct scene_material;
struct sequence_point;
} // namespace Payload

using voxel_index_val_t = unsigned int;
/// Voxel index (x,y,z)
using voxel_index_t = std::array<voxel_index_val_t, 3>;
/// Diffraction map index (x,y)
using dmap_index_t = std::array<voxel_index_val_t, 2>;
/// Voxel index (x,y,z) / additional distance (>0 if position is outside voxel grid) pair
using voxel_index_pair_t = std::pair<voxel_index_t, float>;
/// Map item-index -> Voxel indices (x,y,z)
using index_storage_t = std::unordered_map<int, voxel_index_pair_t>;
/// Map material-id -> Material description
using scene_material_map_t = std::unordered_map<int, struct scene_material>;
/// Map material-id -> Renderer instructions
using renderer_instruction_map_t = std::unordered_map<int, struct renderer_instruction>;
/// Path returned by voxel-traversal algorithm (voxel-indices)
using voxel_traversal_path_t = std::vector<voxel_index_t>;

//--------------------------------------------------------------------------------------------------
/**
 * Basic class representing one voxel
 */
struct voxel
{
    using shared = std::shared_ptr<voxel>;

    voxel() = default;
    virtual ~voxel() = default;

    void set_material_id(std::uint16_t id)
    {
        _properties &= 0xffff0000;
        _properties |= id;
    }

    std::uint16_t get_material_id() const { return _properties & 0x0000ffff; }
    std::uint32_t get_properties() const { return _properties; }

    /**
     * Enable diffraction for this voxel
     *
     * @param status Enable/disable diffraction
     */
    void enable_diffraction(bool status) { set_property_bit(bit_diffraction_enable, status); }

    /**
     * Check whether diffraction is enabled for this voxel
     *
     * @return True, if voxel may contribute to diffraction, else false
     */
    bool diffraction_enabled() const { return get_property_bit(bit_diffraction_enable); }

    /**
     * Global occluder check
     *
     * @return true if this voxel is global occluder (Occluding material AND enabled for diffraction)
     */
    bool is_global_occluder() const { return diffraction_enabled(); }

    /**
     * Local occluder check
     *
     * @return true if this voxel is local occluder (Occluding material but not enabled for diffraction)
     */
    bool is_local_occluder() const { return !is_sound_medium() && !diffraction_enabled(); }

    void set_sound_medium(bool is_sound_medium)
    {
        set_property_bit(bit_sound_medium, is_sound_medium);
    }

    bool is_sound_medium() const { return get_property_bit(bit_sound_medium); }

protected:
    void set_property_bit(int bit, bool value)
    {
        // Clear if set
        _properties &= ~(1 << bit);
        _properties |= (value << bit);
    }

    bool get_property_bit(int bit) const { return (_properties >> bit) & 0x00000001; }

    enum
    {
        bit_diffraction_enable = 31,
        bit_sound_medium = 29
    };

    /** Bitfield storing the voxel properties
     *  Lower 16 bits to store material ID
     *  Bit 31: diffraction enable
     *  Bit 29: sound medium
     */
    std::uint32_t _properties{ (1 << bit_sound_medium) };
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing the voxel grid
 */
struct grid
{
    using shared = std::shared_ptr<grid>;

    /**
     * Constructs a voxel grid from scene geometry blocks
     *
     * @param blocks Scene geometry blocks
     * @param grid_dimension Scene grid dimensions
     * @param instructions Renderer instructions
     */
    grid(std::vector<Payload::block<3>> const& blocks, voxel_index_t const& grid_dimension,
        renderer_instruction_map_t const& instructions);

    /**
     * Get voxel at coordinate
     *
     * @param x x-coordinate (1..dim[0])
     * @param y y-coordinate (1..dim[1])
     * @param z z-coordinate (1..dim[2])
     * @return Valid voxel if in range, else undefined
     */
    voxel const& at(voxel_index_val_t x, voxel_index_val_t y, voxel_index_val_t z) const;
    voxel const& at(voxel_index_t const& idx) const { return at(idx[0], idx[1], idx[2]); }

    /**
     * Non-const version of voxel const& at(int,int,int) const;
     */
    voxel& at(voxel_index_val_t x, voxel_index_val_t y, voxel_index_val_t z)
    {
        return const_cast<voxel&>(static_cast<grid const*>(this)->at(x, y, z));
    }

    /**
     * Get the scene dimension in voxels [1..dim]
     */
    voxel_index_t get_dim() const { return _dim; }

    /**
     * Check wheter the given index is within the voxel-grid
     *
     * @param idx Voxel index (x,y,z) [1..dim]
     * @return True if inside, else false
     */
    bool in_range(voxel_index_t const& idx) const { return in_range(idx[0], idx[1], idx[2]); }

    /**
     * convert grid to scene geometry blocks
     */
    bool convert(std::vector<Payload::block<3>>& blocks);

#ifdef GRID_DEBUG
    /**
     * Access to the underlying vector (debugging)
     *
     * @return Vector containing the voxels
     */
    std::vector<voxel> const& get_voxels() const { return _voxels; }
#endif
    virtual ~grid() = default;

protected:
    /**
     * Checks whether x/y/z is within 1..dim[0/1/2]
     */
    bool in_range(voxel_index_val_t x, voxel_index_val_t y, voxel_index_val_t z) const
    {
        return (x > 0 && x <= _dim[0]) && (y > 0 && y <= _dim[1]) && (z > 0 && z <= _dim[2]);
    }

    /**
     * Update the grid
     */
    void update_grid(std::vector<Payload::block<3>> const& blocks,
        renderer_instruction_map_t const& instructions);

protected:
    voxel_index_t _dim;
    std::vector<voxel> _voxels;
};

//--------------------------------------------------------------------------------------------------
struct voxel_scene_info_basic
{
    using shared = std::shared_ptr<voxel_scene_info_basic>;

    std::array<float, 3> p0{ 0 }; //!< scene bounding box p0 (xyz)
    std::array<float, 3> p1{ 0 }; //!< scene bounding box p1 (xyz)
    Voxel::voxel_index_t dim{ 0 }; //!< voxel grid resolution (xyz)
    float voxel_size{ 0 }; //!< size of one voxel
    std::optional<int> id; //!< Subscene-ID, if the scene consists of more than one subscene

    voxel_scene_info_basic(std::array<float, 3> const& p0_, std::array<float, 3> const& p1_,
        Voxel::voxel_index_t const& dim_, std::optional<int> const& id_)
        : p0(p0_)
        , p1(p1_)
        , dim(dim_)
        , voxel_size(get_voxel_size())
        , id(id_)
    {}

    voxel_scene_info_basic(voxel_scene_info_basic const&) = default;

protected:
    Voxel::voxel_index_val_t to_index(float p, int idx) const
    {
        // Middle of the voxel, round to nearest integer
        return static_cast<Voxel::voxel_index_val_t>(((p - p0[idx]) / voxel_size + .5f) + .5f);
    }

    float get_voxel_size()
    {
        std::array<float, 3> const dim_meters = { p1[0] - p0[0], p1[1] - p0[1], p1[2] - p0[2] };
        /// @todo add sanity-check here?

        auto const& dim_voxels = dim;
        auto const dim_voxels_max = std::max_element(std::begin(dim_voxels), std::end(dim_voxels));
        auto const dim_meters_max = std::max_element(std::begin(dim_meters), std::end(dim_meters));

        return *dim_meters_max / *dim_voxels_max;
    }

    bool is_in_range(Voxel::voxel_index_t const& idx) const
    {
        auto const x = idx[0];
        auto const y = idx[1];
        auto const z = idx[2];

        return (x > 0 && x <= dim[0]) && (y > 0 && y <= dim[1]) && (z > 0 && z <= dim[2]);
    }

public:
    /**
     * Get voxel-index for given location
     *
     * @param location Object location (x,y,z - OpenGL)
     * @param usr_location Location of the user (only used if clip==false)
     * @param clamp Hard-clamp to scene boundaries
     * @return Voxel index, add'l distance pair
     */
    Voxel::voxel_index_pair_t to_voxel_grid(
        EAR::Vector3 const& location, EAR::Vector3 const& usr_location, bool clamp) const;

    /**
     * Get voxel-index for given location (clamped to grid if outside voxel-region)
     * @see to_voxel_grid(EAR::Vector3 const&, EAR::Vector3 const&, bool)
     */
    Voxel::voxel_index_pair_t to_voxel_grid(EAR::Vector3 const& location) const;

    /**
     * Get voxel-index for given location
     * @see to_voxel_grid(EAR::Vector3 const&, EAR::Vector3 const&, bool)
     */
    Voxel::voxel_index_pair_t to_voxel_grid(
        EAR::Vector3 const& location, EAR::Vector3 const& usr_location) const;

    /**
     * Map position to voxel indices (unsafe)
     * @param pos_xyz Position x,y,z coordinates
     * @return Voxel indices unclamped, might yield indices that are outside the valid region
     */
    Voxel::voxel_index_t xyz2idx_unsafe(std::array<float, 3> const& pos_xyz) const
    {
        return { to_index(pos_xyz[0], 0), to_index(pos_xyz[1], 1), to_index(pos_xyz[2], 2) };
    }

    /**
     * Map position to voxel indices
     * @param pos_xyz Position x,y,z coordinates
     * @return Voxel indices clamped to [1, dim]
     */
    Voxel::voxel_index_t xyz2idx(std::array<float, 3> const& pos_xyz) const
    {
        return { std::clamp(to_index(pos_xyz[0], 0), 1U, dim[0]),
            std::clamp(to_index(pos_xyz[1], 1), 1U, dim[1]),
            std::clamp(to_index(pos_xyz[2], 2), 1U, dim[2]) };
    }

    /**
     * Get mid-point of the voxel at given index
     * @param idx Voxel index x,y,z
     * @return Mid-point of the voxel
     */
    std::array<float, 3> idx2xyz(Voxel::voxel_index_t const& idx) const
    {
        return { static_cast<float>(p0[0] + voxel_size * (idx[0] - .5f)),
            static_cast<float>(p0[1] + voxel_size * (idx[1] - .5f)),
            static_cast<float>(p0[2] + voxel_size * (idx[2] - .5f)) };
    }
};

//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
struct projection_map
{
    /**
     * Construct a projection map and initialize with 0
     *
     * @param dim_x Map resolution x
     * @param dim_y Map resolution y
     */
    projection_map(voxel_index_val_t dim_x, voxel_index_val_t dim_y,
        std::optional<voxel_index_val_t> const& z_ = std::nullopt)
        : dim({ dim_x, dim_y })
        , z(z_)
    {
        map.resize(dim_y, std::valarray<char>('\0', dim_x));
    }

    /**
     * Construct a projection map and initialize with values from payload
     *
     * @param blocks_diffraction_map Diffraction map elements
     * @param res Resolution of the map (x,y)
     */
    projection_map(std::vector<Voxel::Payload::block<2>> const& blocks_diffraction_map,
        dmap_index_t const& res);

    /**
     * Construct a projection map and initialize with values from the grid
     *
     * @param grid Voxel grid
     * @param height Optional grid height (in voxels) at which the projection map will be constructed
     */
    projection_map(
        Voxel::grid const& grid, std::optional<unsigned int> const& height = std::nullopt);

    projection_map(projection_map const&) = default;
    projection_map& operator=(projection_map const&) = default;
    projection_map(projection_map&&) = default;
    projection_map& operator=(projection_map&&) = default;

    bool convert(std::vector<Payload::block<2>>& blocks) const;

    bool valid() const { return ((dim[1] == map.size()) && (dim[0] == map[0].size())); }
    char const& at(std::size_t x, std::size_t y) const { return map[y][x]; }

    char& at(std::size_t x, std::size_t y)
    {
        return const_cast<char&>(static_cast<projection_map const*>(this)->at(x, y));
    }

    void set(std::size_t x, std::size_t y, char value) { at(x, y) = value; }
    std::optional<voxel_index_val_t> const& get_height() const { return z; }

    std::array<voxel_index_val_t, 2> dim;

protected:
    std::valarray<std::valarray<char>> map;
    std::optional<voxel_index_val_t> z;
};

//--------------------------------------------------------------------------------------------------
struct index_storage
{
    using shared = std::shared_ptr<index_storage>;

    enum
    {
        listener_idx = -1 //!< Listener index in the storage (renderer item index start with 0)
    };

    index_storage_t storage; //!< To store item-idx to voxel-idices mapping

    /**
     * Update voxel index for renderer item
     *
     * @param render_item_idx Index of the renderer item (-1 for listener)
     * @param voxel_idx Voxel index (x,y,z), add'l distance pair
     * @return false if voxel_idx is equal to storage, else true
     */
    bool update_voxel_index(int render_item_idx, Voxel::voxel_index_pair_t&& voxel_idx);

    /**
     * Update the listener index
     */
    bool update_listener_voxel_index(Voxel::voxel_index_pair_t&& voxel_idx)
    {
        return update_voxel_index(listener_idx, std::move(voxel_idx));
    }

    /**
     * Get voxel-indices for render item
     *
     * @param render_item_idx Index of the render item
     * @return Stored voxel-indices; throws if render_item_idx is not in store
     */
    Voxel::voxel_index_t const& get_voxel_index(int render_item_idx) const
    {
        return storage.at(render_item_idx).first;
    }

    /**
     * Get add'l distance for render item
     *
     * @param render_item_idx Index of the render item
     * @return Stored distance, 0 if render item is within the voxel grid
     */
    float get_additional_distance(int render_item_idx) const
    {
        return storage.count(render_item_idx) ? storage.at(render_item_idx).second : 0.f;
    }

    /**
     * Get voxel-indices for listener
     */
    Voxel::voxel_index_t const& get_listener_voxel_index() const
    {
        return get_voxel_index(listener_idx);
    }
};

struct direct_view_map
{
    using shared = std::shared_ptr<direct_view_map>;

    std::unordered_map<int, std::uint8_t> dv_map; //!< Direct-view (0/1) per render item (index)

    /**
     * Set direct-view flag for a render item
     *
     * @param idx Render item index
     * @param direct_view Direct-view flag
     */
    void direct_view(int idx, bool direct_view) { dv_map[idx] = direct_view; }

    /**
     * Get direct-view flag for a render item
     *
     * @param idx Render item index
     * @return Direct-view flag if present, else 1 (object in direct-view)
     */
    bool direct_view(int idx) const { return dv_map.count(idx) ? dv_map.at(idx) : 1; }
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a scene element (block)
 */
struct scene_element
{
    scene_element(Payload::block<3> const& block);
    scene_element(int m_id, voxel_index_t const& p0, voxel_index_t p1);

    int m_id{ 0 }; //!< Material ID
    voxel_index_t p0{ 0 }; //!< Block corner 1 (voxel IDs, [1..dim])
    voxel_index_t p1{ 0 }; //!< Block corner 2 (voxel IDs, [1..dim])
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a scene material
 */
struct scene_material
{
    scene_material() = default;
    scene_material& operator=(Payload::scene_material const& material);

    int bs_id{ -1 }; //!< Bitstream ID (EIF) for material mappings
    std::vector<float> trans; //!< Transmission coefficients
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a renderer instruction
 */
struct renderer_instruction
{
    renderer_instruction() = default;

    struct
    {
        bool enabled{ false }; //!< Indicates whether diffraction is enabled

    } diffraction;

    struct
    {
        float recovery{ 0.f }; //!< Occlusion recovery setting

    } occlusion;
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a sequence point within a scene update
 */
struct sequence_point
{
    explicit sequence_point(scene_element const& el);
    explicit sequence_point(Voxel::Payload::sequence_point const& sp_pay);

    int duration{ 0 };
    std::vector<scene_element> scene_elements;
    std::optional<renderer_instruction_map_t> instructions;
    std::optional<unsigned int> diffraction_map_height;
    std::optional<Voxel::projection_map> diffraction_map;
};

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a scene update
 */
struct scene_update
{
    using milliseconds = std::chrono::milliseconds;

    scene_update(Payload::scene_update const& scene_update_pay);

    void advance(struct VoxelSceneState const& state, milliseconds const& elapsed);
    void trigger();
    void trigger(int m_id, Voxel::voxel_index_t const& beg, Voxel::voxel_index_t const& end);

    int id; //!< Update id
protected:
    bool ended() const { return _cur_sequence_point == std::cend(_sequence); }
    void next_sequence_point();

protected:
    /// Collection of sequence points for this update
    using scene_update_map_t = std::vector<sequence_point>;

    scene_update_map_t _sequence; //!< Scene update sequence (i.e. sequence of grid updates)
    scene_update_map_t::const_iterator _cur_sequence_point; //!< Current sequence point
    milliseconds _duration; //!< Duration [ms] until current sequence point is applied
};

//--------------------------------------------------------------------------------------------------
struct dmap_index_hash
{
    dmap_index_hash(dmap_index_t const& dim)
        : _x(dim[0])
    {}

    std::size_t operator()(dmap_index_t const& idx) const
    {
        return (idx[0] - 1) + (idx[1] - 1) * _x;
    }

protected:
    unsigned int _x;
};

//--------------------------------------------------------------------------------------------------
template <typename T, typename Hash = dmap_index_hash> struct cache
{
    using shared = std::shared_ptr<cache<T, Hash>>;

    template <typename... Args>
    cache(Args&&... args)
        : _hash(std::forward<Args>(args)...)
    {}

    /**
     * Store path in cache (copy)
     *
     * @param src Source index on the diffraction map (x,y)
     * @param dest Destination index on the diffraction map (x,y)
     * @param p Path information (T)
     */
    void set_path(dmap_index_t const& src, dmap_index_t const& dest, T const& p)
    {
        _cache[_hash(src)][_hash(dest)] = p;
    }

    /// Store path in cache (move)
    void set_path(dmap_index_t const& src, dmap_index_t const& dest, T&& p)
    {
        _cache[_hash(src)][_hash(dest)] = std::move(p);
    }

    /**
     * No path exists for given src/dest combination
     * Note: T must implement reset()
     * @param src Source diffraction map index (x,y)
     * @param dest Destination diffraction map index (x,y)
     */
    void set_no_path(dmap_index_t const& src, dmap_index_t const& dest)
    {
        _cache[_hash(src)][_hash(dest)].reset();
    }

    /**
     * Get info from cache
     *
     * @param src Source diffraction map index (x,y)
     * @param dest Destination diffraction map index (x,y)
     * @return Returns a path info if present, nullopt else
     */
    std::optional<T> get(dmap_index_t const& src, dmap_index_t const& dest) const
    {
        auto const path_map_it = _cache.find(_hash(src));
        if (path_map_it == std::cend(_cache)) {
            return std::nullopt;
        }

        auto const& path_map = path_map_it->second;
        auto const path_it = path_map.find(_hash(dest));
        if (path_it == std::cend(path_map)) {
            return std::nullopt;
        }

        return path_it->second;
    }

    /// Reset the cache
    void reset() { _cache.clear(); }

    using underlying_t = std::unordered_map<int, std::unordered_map<int, T>>;
    underlying_t const& get_lut() const { return _cache; }

    /**
     * Store path in cache
     *
     * @param src Source index on the diffraction map (x,y)
     * @param dest Destination index on the diffraction map (x,y)
     * @param length Path length
     * @param corner Corner index on the diffraction map (x,y)
     */
    void set_path(
        dmap_index_t const& src, dmap_index_t const& dest, float length, dmap_index_t const& corner)
    {
        auto const hash_src = _hash(src);
        auto const hash_dest = _hash(dest);

        auto& entry = _cache[hash_src][hash_dest];
        entry.length = length;
        entry.corner = corner;
    }

    unsigned int const get_lut_size() const
    {
        return std::accumulate(std::cbegin(_cache), std::cend(_cache), 0,
            [](auto const& lhs, auto const& rhs) { return lhs + rhs.second.size(); });
    }

protected:
    underlying_t _cache;
    Hash _hash;
};

//--------------------------------------------------------------------------------------------------
struct jps_info
{
    jps_info() = default;
    jps_info(float length_, dmap_index_t const& corner_)
        : length(length_)
        , corner(corner_)
    {}

    jps_info(jps_info const&) = default;
    jps_info(jps_info&&) = default;

    jps_info& operator=(jps_info const& other)
    {
        length = other.length;
        corner = other.corner;
        return *this;
    }

    jps_info& operator=(jps_info&& other)
    {
        length = std::move(other.length);
        corner = std::move(other.corner);
        return *this;
    }

    void reset() { length.reset(); }

    std::optional<float> length{ 0.f };
    dmap_index_t corner{ 0 };
};

using pathfinder_cache = cache<jps_info>;

//--------------------------------------------------------------------------------------------------
/**
 * Class representing a voxel subscene
 */
struct subscene
{
    using shared = std::shared_ptr<subscene>;
    subscene(Payload::subscene const& subscene_payload, std::optional<int> const& id);
    ~subscene();

    voxel_scene_info_basic::shared scene; //!< Basic voxel scene information
    grid::shared voxel_grid; //!< Voxel grid
    std::vector<scene_update> scene_updates; //!< Scene updates, e.g. updates to the voxel-grid
    scene_material_map_t scene_materials; //!< Scene materials by ID
    renderer_instruction_map_t renderer_instructions; //!< Instructinons by material ID
    Voxel::projection_map diffraction_map; //!< Diffraction map
    pathfinder_cache::shared cache_pathfinder; //!< Pathfinder (JPS) cache
    float diffraction_effect_control{ 1.f }; //!< Control diffraction "strength"

    /**
     * Set/Update diffraction map (move)
     *
     * @param dmap Diffraction map
     */
    void set_diffraction_map(projection_map&& dmap) { diffraction_map = std::move(dmap); }

    /**
     * Set/Update diffraction map (copy)
     *
     * @param dmap Diffraction map
     */
    void set_diffraction_map(projection_map const& dmap) { diffraction_map = dmap; }
};

//--------------------------------------------------------------------------------------------------
struct VoxelSceneState : public std::enable_shared_from_this<VoxelSceneState>,
                         public EAR::SceneObjectObserver
{
    using shared = std::shared_ptr<VoxelSceneState>;
    using const_shared = std::shared_ptr<VoxelSceneState const>;
    /// Map to store/exchange EQs per render item-index
    using eq_map = std::unordered_map<int, std::vector<float>>;
    using eq_map_ptr = std::shared_ptr<eq_map>;
    /// Scene update parameters
    using update_args_t = std::vector<std::pair<int, float>>;

    /**
     * Construct the voxel-scene state
     *
     * @param vox_data_payload Voxel payload
     */
    VoxelSceneState(MPEGI::VoxDataPayload const& vox_data_payload);

    VoxelSceneState(VoxelSceneState const&) = delete;
    VoxelSceneState& operator=(VoxelSceneState const&) = delete;

    /// On subscene-change callback signature
    using on_subscene_change_fn = std::function<void(const_shared const&)>;

    /**
     * Register for subscene-change notification
     *
     * @param on_subscene_change Callback
     */
    void on_subscene_change(on_subscene_change_fn const& on_subscene_change)
    {
        _on_subscene_change.push_back(on_subscene_change);
    }

    void notify_subscene_change() const
    {
        for (auto& on_subscene_change : _on_subscene_change) {
            on_subscene_change(shared_from_this());
        }
    }

    /// On scene update callback signature
    using on_scene_update_fn = std::function<void(const_shared const&)>;

    /**
     * Register for scene update notification
     *
     * @param on_scene_update Callback
     */
    void on_scene_update(on_scene_update_fn const& on_scene_update)
    {
        _on_scene_update.push_back(on_scene_update);
    }

    void notify_scene_update() const
    {
        for (auto const& on_scene_update : _on_scene_update) {
            on_scene_update(shared_from_this());
        }
    }

    /// Implementation of EAR::SceneObjectObserver
    void objectChanged(
        EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

    /// Get currently active subscene
    subscene::shared get_current_subscene() const { return current_subscene; }

    /// Get the voxel-grid
    grid::shared get_grid() const { return get_current_subscene()->voxel_grid; }

    /// Get the diffraction map
    Voxel::projection_map const& get_diffraction_map() const
    {
        return get_current_subscene()->diffraction_map;
    }

    /// Get the pathfinder cache for the current subscene
    pathfinder_cache::shared get_pathfinder_cache() const
    {
        return get_current_subscene()->cache_pathfinder;
    }

    /// Get basic voxel-scene information
    voxel_scene_info_basic::shared get_scene_info_basic() const
    {
        return get_current_subscene()->scene;
    }

    /// Get basic voxel-scene information for subscene index
    voxel_scene_info_basic::shared get_scene_info_basic(std::size_t index) const
    {
        if (subscenes.size() <= index) {
            return nullptr;
        }

        return subscenes[index]->scene;
    }

    /// Get index-storage
    index_storage::shared get_index_storage() const { return _voxel_index_storage; }

    /// Get direct-view map
    direct_view_map::shared get_direct_view_map() const { return _direct_view; }

    /// Get compensation EQ map
    eq_map_ptr get_eq_map() const { return _compensation_eq; }

    /// Trigger voxel-scene update with args
    void trigger_update(int id, update_args_t const& args) const;

    /**
     * Advance scene state
     *
     * @param elapsed Elapsed time since last call (milliseconds)
     */
    void tick(std::chrono::milliseconds const& elapsed) const;

    /// Check whether the scene is dynamic, i.e., supports voxel-geometry updates
    bool is_dynamic() const
    {
        return std::any_of(std::cbegin(subscenes), std::cend(subscenes),
            [](subscene::shared const& ss) { return !ss->scene_updates.empty(); });
    }

protected:
    /**
     * Update the voxel-scene state and switch sub-scene if necessary
     *
     * @param listener_location Listener location
     */
    void update(EAR::Vector3 const& listener_location);

    std::vector<subscene::shared> subscenes; //!< Collection of voxel-subscenes
    subscene::shared current_subscene; //!< Sub-scene the listener is currently located

    index_storage::shared _voxel_index_storage; //!< To share calculated voxel indices among stages
    direct_view_map::shared _direct_view; //!< Direct view indication per render item

    std::vector<on_subscene_change_fn> _on_subscene_change; //!< Observing sub-scene changes
    std::vector<on_scene_update_fn> _on_scene_update; //!< Observing scene updates

    eq_map_ptr
        _compensation_eq; //!< Exchange compensation EQs between Voxel occlusion and diffraction stage
};

//--------------------------------------------------------------------------------------------------
// Helper functions
//--------------------------------------------------------------------------------------------------
template <typename T> constexpr std::array<T, 3> opengl_to_vox(T x, T y, T z) noexcept
{
    return { z, x, y };
}

template <typename T>
constexpr std::array<T, 3> opengl_to_vox(std::array<T, 3> const& other) noexcept
{
    return opengl_to_vox(other[0], other[1], other[2]);
}

template <typename T> constexpr std::array<T, 3> opengl_to_vox(T const other[3]) noexcept
{
    return opengl_to_vox(other[0], other[1], other[2]);
}

//--------------------------------------------------------------------------------------------------
template <typename T> constexpr std::array<T, 3> vox_to_opengl(T x, T y, T z) noexcept
{
    return { y, z, x };
}

template <typename T>
constexpr std::array<T, 3> vox_to_opengl(std::array<T, 3> const& other) noexcept
{
    return vox_to_opengl(other[0], other[1], other[2]);
}

//--------------------------------------------------------------------------------------------------
template <typename T>
bool in_subscene(T const& location, voxel_scene_info_basic const& scene) noexcept
{
    auto const loc = opengl_to_vox(location.data());

    auto const x = loc[0];
    auto const y = loc[1];
    auto const z = loc[2];

    auto const& p0 = scene.p0;
    auto const& p1 = scene.p1;

    auto const x_min = p0[0];
    auto const y_min = p0[1];
    auto const z_min = p0[2];

    auto const x_max = p1[0];
    auto const y_max = p1[1];
    auto const z_max = p1[2];

    return x >= x_min && x <= x_max && y >= y_min && y <= y_max && z >= z_min && z <= z_max;
}

//--------------------------------------------------------------------------------------------------
inline projection_map make_corner_map(projection_map const& diffraction_map)
{
    projection_map corner_map(diffraction_map);
    for (auto x = 1U; x < diffraction_map.dim[0] - 1; ++x) {
        for (auto y = 1U; y < diffraction_map.dim[1] - 1; ++y) {
            if (diffraction_map.at(x, y)) {
                if (diffraction_map.at(x + 1, y) && diffraction_map.at(x - 1, y)) {
                    corner_map.at(x, y) = 0;
                    continue;
                }

                if (diffraction_map.at(x, y + 1) && diffraction_map.at(x, y - 1)) {
                    corner_map.at(x, y) = 0;
                }
            }
        }
    }

    return corner_map;
}

//--------------------------------------------------------------------------------------------------
#ifdef PM_DEBUG
void dump_pm(projection_map const&, std::string const&);
#endif
#ifdef GRID_DEBUG
void dump_grid(grid const&);
#endif
//--------------------------------------------------------------------------------------------------

void process_one_dimension(std::vector<std::vector<std::vector<uint16_t>>>& vox_matrix_id,
    uint16_t material_id, std::array<voxel_index_val_t, 3>& ps,
    std::array<voxel_index_val_t, 3>& pe, std::array<bool, 3>& checked_dim, voxel_index_t& dim);

void convert_matrix_to_blocks(std::vector<std::vector<std::vector<uint16_t>>>& voxel_matrix,
    std::array<unsigned int, 3>& dim, std::vector<Payload::block<3>>& voxel_blocks);

//--------------------------------------------------------------------------------------------------
void store_precomputed_path_data(pathfinder_cache const& cache_pathfinder,
    voxel_scene_info_basic const& scene, Voxel::projection_map const& diffraction_map);
//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
