/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDERER
#define MPEGI_RENDERER

#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStageObserver.h>
#include <MPEGI/Renderer/Spatializer.h>

#include <MPEGI/Scene.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Scene/SceneStateObserver.h>

#include <MPEGI/Stream/StreamManager.h>

#include <MPEGI/Helpers/Controllable.h>

#include <MPEGI/Helpers/AudioUtilization.h>

#include <EAR/LimiterIvas.h>
#include <EAR/SceneObjectObserver.h>

#include <atomic>
#include <cstddef>
#include <memory>
#include <mutex>
#include <utility>
#include <vector>

#ifdef MPEGI_MEASURE_STAGE_TIMES
#include <iostream>
#endif

//! The main MPEG-I namespace
namespace MPEGI {

// This is implemented in the auto-generated file RegisterStages.cpp:
void registerStages();

// fwd declares
class RenderStage;
struct RenderStageDspData;
struct RenderItem;
class RenderItemStore;

//! Renderer
class Renderer : public SceneStateObserver, public Controllable
{
public:
    //! Constructor
    /*
     * \param bitstream         a Bitstream instance to get the needed payloads and forward them into individual stages
     * \param state             pointer to a SceneState (must be active throughout the lifetime of the Renderer)
     *                          When the SceneState is detached while the Renderer is still alive, it will shutdown its rendering.
     * \param streamManager     a shared pointer to a StreamManager instance that manages the audio streams (fs, B, input stream to object mappings)
     */
    Renderer(const Bitstream& bitstream, const SceneState* state,
        std::shared_ptr<StreamManager> streamManager);

    //! Destructor
    ~Renderer();

    //! Add a new Spatializer
    /*!
     * The Spatializer(s) always terminate the pipeline. Multiple Spatializers can be run in parallel, working on the same read-only data created by the RenderStages.
     * See Spatializer documentation for rights and responsibilities.
     *
     * \param newSpatializer    shared_ptr to a Spatializer (shared ownership ensures that the Spatializer is alive for the whole Renderer lifetime)
     * \param bitstream         a bistream instance to retrieve relevant payloads for the initialization of this spatializer
     */
    void addSpatializer(std::shared_ptr<Spatializer> newSpatializer, const Bitstream& bitstream);

    //! Actual process function: Invokes a run through the pipeline at block rate
    /*!
     * The output channels given are shared between multiple Spatializers if available. The corresponding input samples are consumed by the (external) StreamManager.
     *
     * \param output        2D pointer to the output. The number of channels depends on the registered spatializers.
     */
    void processBlock(float** output);

    //! Inherited from SceneStateObserver
    void sceneStateAttached(const SceneState* state) override;

    //! Inherited from SceneStateObserver
    void sceneStateDetached() override;

    //! Inherited from SceneStateObserver
    void sceneObjectAdded(EAR::SceneObject* object) override;

    //! Inherited from SceneStateObserver
    void sceneObjectRemoved(EAR::SceneObject* object) override;

    AudioUtilization getAudioUtilization();

#ifdef MPEGI_MEASURE_STAGE_TIMES
    void printStageTimes(std::ostream& outStream = std::cout);
    void printStageTimesPerFrame(std::ostream& outStream = std::cout);
    void printStageTimesAvg(std::ostream& outStream = std::cout);
    void initStageTimePerFrame(int blocks = 0);
#endif

protected:
    // These protected members can be accessed by unit tests.
    // Derived classes can create arbitrary stages, but once constructed, the list of stages cannot be changed anymore.

    struct CtorArgs
    {
        const SceneState* sceneState;
        std::shared_ptr<StreamManager> streamManager;
        std::unique_ptr<RenderItemStore> itemStore;
        std::vector<std::shared_ptr<RenderStage>> stages;
    };

    static CtorArgs createStages(const Bitstream& bitstream, const SceneState* sceneState,
        std::shared_ptr<StreamManager> streamManager);

    Renderer(CtorArgs&& args);

    std::unique_ptr<RenderItemStore> itemStore;

private:
    //! Inherited from controllable
    void controlStep() override;

    //! New RenderList from primary items
    RenderList createFreshList();

    std::shared_ptr<StreamManager> streamManager;

    const SceneState* sceneState;

    RenderingStageObserver renderStageObserver;

    //! Flyweight helper to observe individual SceneObjects and assign a fixed RenderItem to them
    class SceneObjectAttachment : public EAR::SceneObjectObserver
    {
    public:
        //! Constructor automatically registers as observer and creates a new render item for the object
        /*!
         * \param obj              pointer to object that will be observed
         * \param parent           pointer to parent renderer (needed to lock list access, etc.)
         * \param parentObjectId   optional string id, that can be used to query the SceneState for teleport, etc.
         *                         (used for LoudspeakerSources)
         */
        SceneObjectAttachment(
            EAR::SceneObject* obj, Renderer* parent, const std::string& parentObjectId = "");

        //! Orphanes the render item and detaches the observer (MUST be called before destruction)
        void remove();

        //! Inherited from EAR::SceneObjectObserver
        void objectChanged(
            EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

        RenderItem* itemHandle;

    private:
        Renderer* parent;
        EAR::SceneObject* object;
        std::string objectId;
    };

    //! Flyweight helper to observe the Listener
    class ListenerAttachment : public EAR::SceneObjectObserver
    {
    public:
        //! Constructor automatically registers as observer and creates a new render item for the object if appropriate
        ListenerAttachment(EAR::SceneObject* obj, Renderer* parent);

        //! Destructor automatically orphanes the render item
        ~ListenerAttachment();

        void objectChanged(
            EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

        std::unique_ptr<RenderList::ListenerState> state;

    private:
        Renderer* parent;
        EAR::SceneObject* object;
    };

    std::vector<std::pair<EAR::SceneObject*, std::unique_ptr<SceneObjectAttachment>>>
        sceneObjectAttachments;
    std::unique_ptr<ListenerAttachment> listenerAttachment;

    friend SceneObjectAttachment;
    friend ListenerAttachment;

    std::mutex listAccessMutex;

    std::shared_ptr<RenderList> lineupList;
    std::size_t renderListCount;

    // NB: DspData uses triple buffering with atomic indices.
    using BufferSelector = unsigned int;
    BufferSelector updateIdx = 1;
    BufferSelector renderIdx = 0;
    std::atomic<BufferSelector> updateIdxAtomic = updateIdx;
    std::atomic<BufferSelector> renderIdxAtomic = renderIdx;
    static_assert(std::atomic<BufferSelector>::is_always_lock_free);
    using PerStageDspData = std::vector<std::unique_ptr<RenderStageDspData>>;
    PerStageDspData stageDspData[3];
    using PerSpatializerDspData = std::vector<std::unique_ptr<Spatializer::DspData>>;
    PerSpatializerDspData spatializerDspData[3];
    std::vector<int> orphanedItems[3];

    std::vector<std::shared_ptr<RenderStage>> stages;

    std::vector<std::shared_ptr<Spatializer>> spatializers;
    std::vector<std::unique_ptr<EAR::RealFrame>> outputBuffers;
    std::vector<std::unique_ptr<EAR::LimiterIvas>> limiters;

    std::vector<std::pair<StreamAccessBuffer*, SceneState::AudioStream>> signalChanges;
    std::atomic<bool> signalsToChange;
    std::mutex signalChangeMutex;

#ifdef MPEGI_MEASURE_STAGE_TIMES
    using Clock = std::chrono::steady_clock;
    Clock::duration spatializersConsumedUpdateTime = Clock::duration::zero();
    Clock::duration spatializersConsumedProcessTime = Clock::duration::zero();
    int spatializersUpdateNumBlocks = 0;
    int spatializersProcessNumBlocks = 0;
    std::vector<int> spatializersConsumedUpdateTimePerFrame;
    std::vector<int> spatializersConsumedProcessTimePerFrame;
    bool measureTimeFrameWise = false;
#endif
};

} // namespace MPEGI

#endif /* MPEGI_RENDERER */
