/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd. and Huawei

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. and Huawei retain full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDERSTAGES_MPHOABASESTAGE_HPP
#define MPEGI_RENDERSTAGES_MPHOABASESTAGE_HPP

#include <cstddef>
#include <memory>
#include <unordered_map>
#include <vector>

#include <EAR/BaseTypes.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <mpegcore/Config.hpp>
#include <mpegcore/HoaTranslator.hpp>

namespace mpegcore {

class HoaTranslator;

}

namespace EAR {

class RealBuffer;

} // namespace EAR

namespace MPEGI {

class MphoaPayload;
class StreamBuffer;

//! ...
/*! 
 * Inputs: 
 * - 
 *  
 * Outputs: 
 * - 
 *  
 * How it works: 
 * - 
 * 
 * Stages order:
 * - Required before: ?
 * - Required after: ?
 *  
 * Config options: 
 * -
 * 
 */
class MphoaBaseStage : public RenderStage
{
public:
    MphoaBaseStage();
    ~MphoaBaseStage();

    virtual std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;
    virtual void processBlock(DspContext ctx) override = 0;

    // Verify that groups in spatial metadata and in the scene state contain the same groups with the
    // same sources all in the same order.
    //
    // The verification is done based on the `SpatialMetadataHeader`. `InfoFrame`s or
    // `SpatialMetadataFrame`s are not verified as in the current metadata bitstream version they are
    // expected to have the same groups and sources in the same orders as the header has them (verified
    // by DEMUX in hoa-t).
    //
    // Later iterations of the spatial metadata may support dynamic changes to groups or sources and if
    // that happens then this check is not enough and another way to match the groups and sources to the
    // scene state (and render lists) may need to be implemented. Maybe at that time the metadata will
    // use numeric group and source ids that match those of the corresponding scene state entities.
    static void verifyGroups(std::shared_ptr<std::vector<std::uint8_t>> const& spatialMetadata,
        std::vector<std::unique_ptr<MPEGI::HOAGroup>> const& hoaGroups);

    static bool isMphoaSource(HOASource const& source);

    static bool isMphoaRenderItem(RenderItem const& item);

    static std::vector<mpegcore::HoaTranslatorConfig> createHoaTranslatorConfigs(
        std::vector<std::unique_ptr<MPEGI::HOAGroup>> const& hoaGroups,
        mpegcore::Config const& config, bool haveSpatialMetadata);

protected:
    friend class Renderer;

    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    virtual void initializeStageAdditional(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state, size_t translatorCount,
        const std::vector<size_t>& groupSourceCounts) = 0;

    // Return indices into `translatorData` for active HoaTranslators.
    std::vector<std::size_t> determineActiveHoaTranslators();

    struct DSPItem
    {
        // One buffer per HOASource in the group
        std::vector<std::shared_ptr<MPEGI::StreamBuffer>> inputBuffers;
    };

    struct DspData : RenderStage::DspData
    {
        // One DSPItem per group (i.e. translator) in the same order as in `translatorData`
        std::vector<DSPItem> dspItems;

        // Indices into `translatorData` for active HoaTranslators. This can change based on
        // position changes and therefore is updated per `update` call (i.e. per DSPList).
        std::vector<std::size_t> activeTranslatorIndices;
    };

    std::vector<RenderItem*> outputRenderItems;
    mpegcore::Config config;
    Position listenerPosition;

    struct TranslatorData
    {
        mpegcore::HoaTranslator translator{};
        MPEGI::HOAGroup* group{ nullptr };
        std::vector<std::string> hoaSourceIds;
    };

    // `TranslatorData` for all translators, active or not.
    std::vector<TranslatorData> translatorData;

    // Map to keep track of item properties for which we have received a change
    // ItemProperty -> HOASource index -> bool
    std::unordered_map<ItemProperty, std::unordered_map<std::size_t, bool>> itemPropertyChanged;

    std::shared_ptr<std::vector<std::uint8_t>> spatialMetadata;
    SceneState const* sceneState{ nullptr };
    std::size_t B{ 0 };
    float mphoaGain{ 0 };
    bool isInitialized{ false };
};

} // namespace MPEGI

#endif // MPEGI_RENDERSTAGES_MPHOASTAGE_HPP
