/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDER_ITEM
#define MPEGI_RENDER_ITEM

#include <algorithm>
#include <complex>
#include <cstddef>
#include <memory>
#include <string>
#include <vector>

#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Stream/StreamAccessBuffer.h>

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Trajectory.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/BaseTypes.h>
#include <EAR/DirectivityModel.h>
#include <EAR/SceneObject.h>

// fwd declares
namespace EAR {
class ComplexBuffer;
class Geometry;
} // namespace EAR

namespace MPEGI {

// This is defined outside of RenderItem to be able to forward declare
struct RenderItemExtraState
{
    // Default constructor is needed to allow implicit default constructor in derived class:
    RenderItemExtraState() = default;
    // Copy constructor is disabled to enforce usage of the object managed by the framework:
    RenderItemExtraState(const RenderItemExtraState&) = delete;
    // Move constructor is disabled to make sure the data stays in the originally allocated memory:
    RenderItemExtraState(RenderItemExtraState&&) = delete;
    virtual ~RenderItemExtraState() = default;
};

struct RenderItem
{

    //! Destructor
    ~RenderItem(){};

    //! Base class for stage-specific per-item state.
    //! \see getState(), getStateOrCreate(), unsafeRemoveState(), RenderItemStore::getItemStates().
    using ExtraState = RenderItemExtraState;

    //! Unique index of the RenderItem
    /*!
     * Will not be reused for an unrelated RenderItem during the lifetime of a Renderer and will not be shared across secondary and primary items.
     * However, with multiple Renderers, only the combination of idx and owner is truly unique.
     */
    const int idx;

    //! The owner is the RenderItemStore where this RI belongs to
    RenderItemStore* owner;

    //! Item status (see enum ItemStatus) – should be handled as READ-ONLY
    ItemStatus status;

    //! Item type (see enum ItemType)
    ItemType type;

    // TODO: Determine a unified approach for how to pass renderstage dependent data between stages.
    //       There's currently at least 4 different ways used for this: (1) String ids (i.e. pre- or postfixes put in the item name) (2) stage dependent member variables (3) ItemSubType (4) ExtraState.

    //! Item type (see enum ItemSubtype)
    ItemSubtype subtype;

    //! Indicate a change in the RenderItem configuration since the last time step
    ItemProperty changed;

    //! Combination of AParams (see enum in Types.h). Default behaviour is indicated by no flags.
    AParam aparams;

    //! RenderList index of the corresponding primary source if type is secondary (-1 if not secondary)
    int parentRI;

    //! Id of the Reverb environment this item is located in (name of the acoustic environment)
    std::string acousticEnvironmentId;

    //! Trajectory (provided for the audio thread to determine anticipated locations for every frame, e.g. for smooth doppler shifts)
    std::shared_ptr<Trajectory> trajectory;

    //! Teleport flag, indicating non-interpolated movement
    bool teleport;

    //! Position (location, orientation and cspace) contains the current values (sampled at the last update).
    EAR::Position position;

    //! The localCS can be used to transform other global coordinates into local coordinates relative to this RI (e.g. for directivity angles)
    EAR::HomogenMatrix localCS;

    //! Override for the distance gain and propagation delay calculated from the given position
    /*!
     * If this has a value != 0 (default: 0), this value is  added to the
     * computed distance used for calculating distance gain and propagation delay.
     *
     * A delta is used instead of "apparentDistance" to account for a possible constant-velocity trajectory.
     * If the velocity is not considered, the propagation delay cannot be smoothly updated in each audio block.
     */
    float apparentDistanceDelta;

    //! Directivity pointer - nullptr means it has no directivity
    std::shared_ptr<EAR::DirectivityModel> directivity;

    //! Directiveness value (intensity of the directivity filter)
    float directiveness;

    //! Reference distance for the distance model
    float refDistance;

    //! Portal distance compensation
    float portalDistComp;

    //! Extent Geometry pointer
    /*!
     * The Geometry object describes the extent of the source.
     */
    const EAR::Geometry* extent;

    //! List of rays that hit an extended source
    /*!
     * Entries are pairs of ray ID and ray length (relative to the length of the ray direction vector).
     *
     * There will only be hits for sources with an extent.
     *
     * The length can be negative, which means that the extended source has been hit from the inside.
     *
     * If a source has an extent but no rayHits, this means that the extended source is very small and/or far away
     * and should be handled like a point source, but using this->extent->position (this->position is ignored!).
     */
    std::vector<RayHit> rayHits;

    //! Extent channel description
    /*!
     * Describes how multiple channels of the signal are mapped to the source extent. See EIF for details.
     */
    const ExtentLayout* extentLayout;

    //! Information about the reflection sequence (only filled for ItemType::Reflection)
    std::shared_ptr<ReflectionInfo> reflectionInfo;

    //! Information about occlusion
    std::optional<OcclusionInfo> occlusionInfo;

    //! Positions for the channels (only filled for ItemType::Reverb)
    std::vector<EAR::Position> channelPositions;

    //! Information about the configuration if the RenderItem is a HOA item
    std::shared_ptr<HoaInfo> hoaInfo;

    //! Pointer to a StreamBuffer instance holding NxB samples, with N being the number of channels for this item and B being the blocksize
    /*!
     * If a MPEGI::StreamAccessBuffer is used, it is guaranteed by the StreamManager that the RenderStage or Spatializer finds the correct samples there on processBlock() call
     * A normal MPEGI::StreamBuffer can and has to be filled with samples by the creating stage.
     */
    std::shared_ptr<MPEGI::StreamBuffer> signal;

    //! Accumulated gain (factor, NOT dB)
    /*!
     * Including: source gains (from bitstream), RenderStage gains
     * Excluding: distance attenuation (will be realized from path length in audio thread)
     */
    float gain;

    //! Gain for reverb rendering (factor, NOT dB)
    /*!
     * Independent of other source gains and distance attenuation
     */
    float itemReverbGain;

    //! Flag for reverb rendering
    /*!
     * Independent of other source gains and distance attenuation
     */
    bool noReverb;

    //! Optional item loudness (factor, NOT dB)
    std::optional<float> loudness;

    //! Magnitude characteristics of the propagation path per band
    /*!
     * This field may be modified by RenderStages by multiplying the values with other magnitudes
     */
    std::vector<float> eq;

    //! Name of the SceneObject this is derived from (referring to the id from the EIF format, used for debug purposes)
    const std::string objectName;

    //! ID of the vdl method to be used for this item
    VDLMethod vdlMethod;

    //! ID of the parent portal when ItemType::DirectionalReverb
    int parentPortalID;

    bool isSoundEffect;

    //! Name of the granular database if the source is using granular synthesis
    std::string granularDatabaseId;

    //! Granular descriptor space target position
    EAR::Vector3 granularPos;

    //! Updated probability of the group specified by granularProbGroup
    EAR::Vector3 granularGroupProb;

    //! Helper function to set the Status and automatically add the correct changed flag
    inline void setStatus(ItemStatus newStatus, bool forceChanged = false)
    {
        if (status != newStatus || forceChanged) {
            EAR_CHECK_REL(this->status != ItemStatus::Orphaned, "Reanimating an Orphaned item!");
            status = newStatus;
            changed |= ItemProperty::Status;
        }
    };

    //! Reset changed flag (and item status if it was New)
    inline void resetChanged()
    {
        changed = ItemProperty::None;

        if (status == ItemStatus::New) {
            setStatus(ItemStatus::Active);
        }
        teleport = false;
    };

    //! Helper function to set the Aparams and automatically add the correct changed flag
    inline void setAParams(AParam newAparams)
    {
        aparams = newAparams;
        changed |= ItemProperty::Aparams; // TODO: only set if actually changed?
    };

    //! Helper function to set the Position and automatically add the correct changed flag
    /*!
     * This automatically updates the position and trajectory and allows to explicitly specify the new localCS. Update manually if that does not suit your needs (but don't forget the changed flag then).
     */
    inline void setPosition(const EAR::Position& newPosition, const EAR::HomogenMatrix& newLocalCS,
        std::shared_ptr<Trajectory> newTrajectory = std::shared_ptr<Trajectory>(nullptr))
    {
        if (newPosition.location != position.location) {
            changed |= ItemProperty::Location;
        }
        if (newPosition.orientation != position.orientation) {
            changed |= ItemProperty::Orientation;
        }

        if (newPosition.location != position.location
            || newPosition.orientation != position.orientation
            || newPosition.cSpace != position.cSpace || !(newLocalCS == localCS)) {
            changed |= ItemProperty::Position;
        }

        position = newPosition;
        localCS = newLocalCS;

        if (!newTrajectory) {
            double t = 0;
            if (trajectory) {
                t = trajectory->getTimestamp();
            }
            newTrajectory = std::make_shared<ConstLocation>(
                newPosition.location, t); // TODO: Update timestamp here?
        }
        trajectory = newTrajectory;
    };

    //! Helper function to set the Position and automatically add the correct changed flag
    /*!
     * This automatically updates the localCS, position and trajectory. Update manually if that does not suit your needs (but don't forget the changed flag then).
     */
    inline void setPosition(const EAR::Position& newPosition,
        std::shared_ptr<Trajectory> newTrajectory = std::shared_ptr<Trajectory>(nullptr))
    {
        if (newPosition.location != position.location) {
            changed |= ItemProperty::Location;
        }
        if (newPosition.orientation != position.orientation) {
            changed |= ItemProperty::Orientation;
        }

        if (newPosition.location != position.location
            || newPosition.orientation != position.orientation
            || newPosition.cSpace != position.cSpace) {
            changed |= ItemProperty::Position;
        }

        position = newPosition;
        localCS = EAR::HomogenMatrix(newPosition);

        if (!newTrajectory) {
            double t = 0;
            if (trajectory) {
                t = trajectory->getTimestamp();
            }
            newTrajectory = std::make_shared<ConstLocation>(
                newPosition.location, t); // TODO: Update timestamp here?
        }
        trajectory = newTrajectory;
    };

    //! Helper function to set the Directivity and automatically add the correct changed flag
    inline void setDirectivity(std::shared_ptr<EAR::DirectivityModel> newDirectivity)
    {
        if (directivity != newDirectivity) {
            directivity = newDirectivity;
            changed |= ItemProperty::Directivity;
        }
    };

    //! Helper function to set the Directivity and automatically add the correct changed flag
    inline void setDirectiveness(float newDirectiveness)
    {
        if (directiveness != newDirectiveness) {
            directiveness = newDirectiveness;
            changed |= ItemProperty::Directivity;
        }
    };

    //! Helper function to set the Extent with a layout and automatically add the correct changed flag
    inline void setExtent(const EAR::Geometry* newExtent, const ExtentLayout* newLayout = nullptr)
    {
        if (extent != newExtent) {
            extent = newExtent;
            extentLayout = newLayout;
            changed |= ItemProperty::Extent;
        }
    };

    //! Helper function to set the Signal and automatically add the correct changed flag
    inline void setSignal(std::shared_ptr<MPEGI::StreamBuffer> newSignal)
    {
        signal = newSignal;
        changed |= ItemProperty::Signal; // TODO: only set if actually changed?
    };

    //! Helper function to set the Gain and automatically add the correct changed flag
    inline void setGain(float g)
    {
        if (gain != g) {
            gain = g;
            changed |= ItemProperty::Gain;
        }
    };

    //! Helper function to set the reverb Gain and automatically add the correct changed flag
    inline void setReverbGain(float newItemReverbGain)
    {
        if (itemReverbGain != newItemReverbGain) {
            itemReverbGain = newItemReverbGain;
            // clamp newItemReverbGain value between -119 and +12 dB
            if (itemReverbGain > 3.9810717055349725077) {
                itemReverbGain = 3.9810717055349725077;
            }
            if (itemReverbGain < 0.00000112201845430196) {
                itemReverbGain = 0.00000112201845430196;
            }
            changed |= ItemProperty::ReverbGain;
        }
    };

    //! Helper function to set the reverb Gain and automatically add the correct changed flag
    inline void setNoReverb(bool newNoReverb)
    {
        if (noReverb != newNoReverb) {
            noReverb = newNoReverb;
            changed |= ItemProperty::NoReverb;
        }
    };

    //! Helper function to set the Gain and automatically add the correct changed flag
    inline void setVDLMethod(VDLMethod g)
    {
        if (vdlMethod != g) {
            vdlMethod = g;
            changed |= ItemProperty::VDLMethod;
        }
    };

    //! Helper function to set the Loudness and automatically add the correct changed flag
    inline void setLoudness(std::optional<float> const& l)
    {
        if (loudness != l) {
            loudness = l;
            changed |= ItemProperty::Loudness;
        }
    }

    //! Helper function to set the granular descriptor space position and automatically add the correct changed flag
    inline void setGranularDatabaseId(std::string name)
    {
        if (granularDatabaseId != name) {
            granularDatabaseId = name;
            changed |= ItemProperty::GranularDatabaseId;
        }
    };

    //! Helper function to set the granular descriptor space position and automatically add the correct changed flag
    inline void setGranularPos(EAR::Vector3 pos)
    {
        if (granularPos != pos) {
            granularPos = pos;
            changed |= ItemProperty::GranularPos;
        }
    };

    //! Helper function to set the probability of the group specified by granularProbGroupNo and automatically add the correct changed flag
    inline void setGranularGroupProb(EAR::Vector3 granGroupprob)
    {
        if (granularGroupProb != granGroupprob) {
            granularGroupProb = granGroupprob;
            changed |= ItemProperty::GranularGroupProb;
        }
    };

    //! Helper function to add to the Gain (by multiplicating the factors) and automatically add the correct changed flag
    inline void addGain(float g)
    {
        if (g != 1) {
            gain *= g;
            changed |= ItemProperty::Gain;
        }
    };

    //! Helper function to set the Eq and automatically add the correct changed flag
    inline void setEq(const std::vector<float>& newEq, bool change = true)
    {
        eq = newEq;
        if (change)
            changed |= ItemProperty::Eq; // TODO: only set, if actually changed?
    };

    //! Helper function to add to the Eq values (by multiplication of the factors) and automatically add the correct changed flag
    inline void addEq(const float* newEq, bool change = true)
    {
        std::transform(
            eq.begin(), eq.end(), newEq, eq.begin(), [](float l, const float r) { return l * r; });
        if (change)
            changed |= ItemProperty::Eq; // TODO: only set, if actually changed?
    };

    //! Helper function to add to the Eq values (by multiplication of the factors) and automatically add the correct changed flag (specialization for complex floats -> multiplying the real part)
    inline void addEq(const std::complex<float>* newEq, bool change = true)
    {
        std::transform(eq.begin(), eq.end(), newEq, eq.begin(),
            [](float l, const std::complex<float> r) { return l * r.real(); });
        if (change)
            changed |= ItemProperty::Eq; // TODO: only set, if actually changed?
    };

    //! Helper function to sync properties of two RenderItems
    /*!
     * The sync includes the changed status of the other RI.
     * \param other         Other RenderItem to copy the properties and their changed status from
     * \param props         A bitwise boolean concatenation of all properties that shall be synced
     */
    inline void sync(const RenderItem& other, ItemProperty props)
    {
        if (props & ItemProperty::Directivity) {
            this->changed |= other.changed & ItemProperty::Directivity;
            this->directivity = other.directivity;
            this->directiveness = other.directiveness;
        }

        if (props & ItemProperty::Gain) {
            this->changed |= other.changed & ItemProperty::Gain;
            this->gain = other.gain;
        }

        if (props & ItemProperty::Loudness) {
            this->changed |= other.changed & ItemProperty::Loudness;
            this->loudness = other.loudness;
        }

        if (props & ItemProperty::Eq) {
            this->changed |= other.changed & ItemProperty::Eq;
            this->eq = other.eq;
        }

        if (props & ItemProperty::Aparams) {
            this->changed |= other.changed & ItemProperty::Aparams;
            this->aparams = other.aparams;
        }

        if (props & ItemProperty::Position) {
            this->changed |= other.changed & ItemProperty::Position;
            if (this->position.location != other.position.location) {
                this->changed |= ItemProperty::Location;
                this->changed |= ItemProperty::Position;
            }
            if (this->position.orientation != other.position.orientation) {
                this->changed |= ItemProperty::Orientation;
                this->changed |= ItemProperty::Position;
            }
            this->position = other.position;
            this->localCS = other.localCS;
            this->trajectory = other.trajectory;
            this->teleport = other.teleport;
        }

        if (props & ItemProperty::Extent) {
            this->changed |= other.changed & ItemProperty::Extent;
            this->extent = other.extent;
            this->extentLayout = other.extentLayout;
        }

        if (props & ItemProperty::Signal) {
            this->changed |= other.changed & ItemProperty::Signal;
            this->signal = other.signal;
        }

        if (props & ItemProperty::Status) {
            EAR_CHECK_REL(this->status != ItemStatus::Orphaned, "Reanimating an Orphaned item!");
            this->changed |= other.changed & ItemProperty::Status;
            this->status = other.status;
        }

        if (props & ItemProperty::RefDist) {
            this->changed |= other.changed & ItemProperty::RefDist;
            this->refDistance = other.refDistance;
        }

        if (props & ItemProperty::VDLMethod) {
            this->changed |= other.changed & ItemProperty::VDLMethod;
            this->vdlMethod = other.vdlMethod;
        }

        if (props & ItemProperty::ParentRI) {
            this->changed |= other.changed & ItemProperty::ParentRI;
            this->parentRI = other.parentRI;
        }

        if (props & ItemProperty::Reverb) {
            this->changed |= other.changed & ItemProperty::Reverb;
            this->acousticEnvironmentId = other.acousticEnvironmentId;
        }

        if (props & ItemProperty::ReflectionInfo) {
            this->changed |= other.changed & ItemProperty::ReflectionInfo;
            this->reflectionInfo = other.reflectionInfo;
        }

        if (props & ItemProperty::RayHits) {
            this->changed |= other.changed & ItemProperty::RayHits;
            this->rayHits = other.rayHits;
        }

        if (props & ItemProperty::ReverbGain) {
            this->changed |= other.changed & ItemProperty::ReverbGain;
            this->itemReverbGain = other.itemReverbGain;
        }

        if (props & ItemProperty::NoReverb) {
            this->changed |= other.changed & ItemProperty::NoReverb;
            this->noReverb = other.noReverb;
        }

        if (props & ItemProperty::GranularPos) {
            this->changed |= other.changed & ItemProperty::GranularPos;
            this->granularPos = other.granularPos;
        }

        if (props & ItemProperty::GranularGroupProb) {
            this->changed |= other.changed & ItemProperty::GranularGroupProb;
            this->granularGroupProb = other.granularGroupProb;
        }
    }

    //! Helper to sync all properties without naming them explicitly
    /*!
     * \param other         Other RenderItem to copy the properties and their changed status from
     * \param exceptStatus  whether or not status should be skipped in the sync (default false, i.e. status WILL be synced)
     */
    inline void syncAll(const RenderItem& other, bool exceptStatus = false)
    {
        sync(other,
            ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Eq
                | ItemProperty::Aparams | ItemProperty::Position | ItemProperty::Extent
                | ItemProperty::Signal | ItemProperty::RefDist | ItemProperty::ParentRI
                | ItemProperty::Reverb | ItemProperty::ReflectionInfo | ItemProperty::RayHits
                | ItemProperty::VDLMethod | ItemProperty::ReverbGain | ItemProperty::NoReverb
                | ItemProperty::Loudness | ItemProperty::GranularDatabaseId
                | ItemProperty::GranularGroupProb | ItemProperty::GranularPos);

        if (!exceptStatus) {
            sync(other, ItemProperty::Status);
        }
    }

    bool isOrphaned() const { return status == ItemStatus::Orphaned; }
    bool isFadeOutOrphaned() const { return status == ItemStatus::FadeOutOrphaned; }
    bool isActive() const { return status == ItemStatus::Active; }
    bool isInactive() const { return status == ItemStatus::Inactive; }
    bool isFadeOutInactive() const { return status == ItemStatus::FadeOutInactive; }

    //! Get stage-specific item state.
    //! The template argument is an arbitrary type provided by the stage,
    //! it only has to be derived from RenderItem::ExtraState.
    //! \return A pointer to the requested state if it already exists, `nullptr` otherwise.
    //! \see getStateOrCreate(), unsafeRemoveState(), RenderItemStore::getItemStates()
    template <typename T> T* getState() { return this->owner->getState<T>(this->idx); }

    //! Get stage-specific item state, creating it if it doesn't exist yet.
    //! The first template argument is an arbitrary type provided by the stage,
    //! it only has to be derived from RenderItem::ExtraState.
    //! The lifetime of the data is managed by the framework,
    //! it is automatically destroyed when the item is destroyed.
    //! The stored data has a stable memory address,
    //! pointers/references to it can be passed to the audio thread
    //! (but data races still have to be avoided).
    //! \param args... Arguments forwarded to the constructor of `T`.
    //! \return A reference to the requested state.
    //! \see getState(), unsafeRemoveState(), RenderItemStore::getItemStates()
    template <typename T, typename... Args> T& getStateOrCreate(Args&&... args)
    {
        return this->owner->getStateOrCreate<T>(this->idx, std::forward<Args>(args)...);
    }

    //! Remove stage-specific state from item.
    //! It is named "unsafe" because is only allowed to be called while the audio thread is stopped!
    //! \see getState(), getStateOrCreate()
    template <typename T> std::size_t unsafeRemoveState()
    {
        return this->owner->unsafeRemoveState<T>(this->idx);
    }

private:
    friend RenderItemStore; // only RenderItemStore can call the Constructor and acts as a factory

    //! Factory function (accessible by RenderList)
    static inline std::unique_ptr<RenderItem> create(
        int idx, RenderItemStore* owner, const std::string& parentName)
    {
        return std::unique_ptr<RenderItem>(new RenderItem(idx, owner, parentName));
    };

    RenderItem(int idx, RenderItemStore* owner, const std::string& parentName)
        : idx(idx)
        , owner(owner)
        , status(ItemStatus::New)
        , type(ItemType::Primary)
        , subtype(ItemSubtype::None)
        , changed(ItemProperty::None)
        , aparams(AParam::None)
        , objectName(parentName)
        , parentRI(-1)
        , acousticEnvironmentId("") // NOTE: This should be the default AE id
        , hoaInfo(nullptr)
        , teleport(false)
        , localCS()
        , apparentDistanceDelta(0)
        , gain(1.f)
        , loudness(std::nullopt)
        , eq(Config::getInt("EQNBands"), 1)
        , portalDistComp(0.f)
        , extent(nullptr)
        , extentLayout(nullptr)
        , directivity(nullptr)
        , directiveness(1.f)
        , refDistance(1.f)
        , parentPortalID(-1)
        , vdlMethod(VDLMethod::spline)
        , itemReverbGain(1.f)
        , noReverb(false)
        , isSoundEffect(false)
        , granularDatabaseId("")
        , granularPos({ 0.0, 0.0, 0.0 })
        , granularGroupProb({ -1.0, 0.0, 0.0 }){};
};

} // namespace MPEGI

#endif /* MPEGI_RENDER_ITEM */
