/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDER_ITEM_STORE
#define MPEGI_RENDER_ITEM_STORE

#include <EAR/Exception.h>
#include <optional>

#include <cassert>
#include <memory>
#include <typeindex>
#include <unordered_map>
#include <vector>

namespace MPEGI {

class RenderItem;
struct RenderItemExtraState;

//! The RenderItem Store owns all Render Items to control their state and lifetime
class RenderItemStore
{
public:
    //! Create a new item
    /*!
     * Deleting a RI happens only indirectly by setting it to the status Orphaned or rather FadeOutOrphaned.
     * This is only possible for the parent of the RI.
     *
     * \param name          human-readable identifier (used for debugging, might be removed in the future)
     * 
     * \return              item handle (raw ptr, guaranteed to be valid until an item is orphaned)
     *                      the item handle shall only be kept and used by the owner of the item to 
     *                      update the "ground truth" and add it to a list
     */
    RenderItem* createItem(const std::string& name);

    //! Allow a spatializer to signal that it completed a fade out
    /*!
     * Once all active spatializers are done, the item status is updated to the successor state
     *
     * \param idx           index of the item
     */
    inline void fadeOutCompleted(int idx)
    {
        //EAR_CHECK_REL(fadeOutCounter[idx] < activeSpatializers,
        //    "The item with index '" + std::to_string(idx)
        //        + "' is already faded out by all Spatializers.");
        //fadeOutCounter[idx]++;

        fadeOutCounter[idx] = activeSpatializers;
        skipFade[idx] = false;
    };

    //! Allow a spatializer to signal that it completed a fade in
    /*!
     * \param idx           index of the item
     */
    inline void fadeInCompleted(int idx)
    {
        //EAR_CHECK_REL(fadeOutCounter[idx] > 0,
        //    "The item with index '" + std::to_string(idx)
        //        + "' is already faded in by all Spatializers.");
        //fadeOutCounter[idx]--;

        fadeOutCounter[idx] = 0;
        skipFade[idx] = false;
    };

    //! Mark the next fade of this item to be skipped
    /*!
     * \param idx           index of the item
     */
    inline void skipNextFade(int idx, bool skip = true) { skipFade[idx] = skip; };

    inline bool getSkipFade(int idx) { return skipFade[idx]; };

    //! Trigger fade out to ItemStatus::Inactive and keep it inactive afterwards
    /*!
     * \param item      ptr to the item to be deactivated (pointer changes every update(), but idx stays the same)
     */
    void deactivate(RenderItem* item);

    //! Trigger fade out to ItemStatus::Orphaned
    /*!
     * The actual status transition has to be handled by the RenderItemHandler
     *
     * \param idx       index of the item to be remove
     */
    void remove(RenderItem* item);

    //! Shallow copy getter for items map (no ownership)
    std::unordered_map<int, RenderItem*> getItems()
    {
        std::unordered_map<int, RenderItem*> newItems;
        for (auto& [k, v] : items) {
            newItems.emplace(k, v.item.get());
        }
        return newItems;
    };

    template <typename T> T* getState(int itemIdx)
    {
        auto entry = this->items.find(itemIdx);
        assert(entry != this->items.end());
        auto& [index, item] = *entry;
        auto& states = item.extraState;
        auto it = states.find(std::type_index(typeid(T)));
        if (it == states.end()) {
            return nullptr;
        }
        return static_cast<T*>(it->second.get());
    }

    template <typename T, typename... Args> T& getStateOrCreate(int itemIdx, Args&&... args)
    {
        auto entry = this->items.find(itemIdx);
        assert(entry != this->items.end());
        auto& [index, item] = *entry;
        auto& state = item.extraState[std::type_index(typeid(T))];
        if (state == nullptr) {
            state = std::make_unique<T>(std::forward<Args>(args)...);
        }
        return static_cast<T&>(*state);
    }

    // This is only allowed to be called while the audio thread is stopped!
    template <typename T> std::size_t unsafeRemoveState(int itemIdx)
    {
        auto entry = this->items.find(itemIdx);
        assert(entry != this->items.end());
        auto& [index, item] = *entry;
        auto& states = item.extraState;
        return states.erase(std::type_index(typeid(T)));
    }

    template <typename T> class ItemStateListIterator;

    template <typename T> class ItemStateList;

    //! Get an iterable object yielding stage-specific item states
    //! for all items where state of the requested type exists.
    //! \see RenderItem::getState(), RenderItem::getStateOrCreate()
    template <typename T> ItemStateList<T> getItemStates();

private:
    friend class Renderer;

    //! Destroy given items immediately and store currently orphaned items back into the list.
    void cleanup(std::vector<int>& orphanedItems);

    //! A RenderItem plus per-stage and per-spatializer state.
    struct ExtendedRenderItem
    {
        std::unique_ptr<RenderItem> item;
        // NB: On destruction, the extraState is destroyed first:
        std::unordered_map<std::type_index, std::unique_ptr<RenderItemExtraState>> extraState;
    };

    //! index -> item
    using RenderItemMap = std::unordered_map<int, ExtendedRenderItem>;
    RenderItemMap items;

    int itemIndex = 0; //!< counted up to have a unique ID for each render item

    int activeSpatializers = 0; //!< Updated by parent renderer to keep track of fade outs
    std::unordered_map<int, int> fadeOutCounter;

    std::unordered_map<int, bool> skipFade;
};

// This iterator has only the necessary features to be used in a range-based for loop.
template <typename T> class RenderItemStore::ItemStateListIterator
{
public:
    explicit ItemStateListIterator(RenderItemMap::iterator iter, RenderItemMap::iterator last)
        : _iter(iter)
        , _last(last)
    {
        _find_matching_state();
    }

    bool operator!=(const ItemStateListIterator<T>& rhs) const { return _iter != rhs._iter; }

    ItemStateListIterator<T>& operator++()
    {
        ++_iter;
        _find_matching_state();
        return *this;
    }

    T& operator*() { return *_cached_state; }

private:
    void _find_matching_state()
    {
        while (_iter != _last) {
            auto& [index, extendedRenderItem] = *_iter;
            auto& [item, states] = extendedRenderItem;
            auto it = states.find(std::type_index(typeid(T)));
            if (it != states.end()) {
                _cached_state = static_cast<T*>(it->second.get());
                break;
            }
            ++_iter;
        }
    }

    RenderItemMap::iterator _iter;
    RenderItemMap::iterator _last;
    T* _cached_state = nullptr;
};

template <typename T> class RenderItemStore::ItemStateList
{
public:
    using iterator = ItemStateListIterator<T>;

    explicit ItemStateList(RenderItemMap& items)
        : _items{ items }
    {}
    iterator begin() { return iterator{ _items.begin(), _items.end() }; }
    iterator end() { return iterator{ _items.end(), _items.end() }; }

private:
    RenderItemMap& _items;
};

template <typename T> RenderItemStore::ItemStateList<T> RenderItemStore::getItemStates()
{
    return ItemStateList<T>{ this->items };
}

} // namespace MPEGI

#endif /* MPEGI_RENDER_ITEM_STORE */
