/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDER_LIST
#define MPEGI_RENDER_LIST

#include <MPEGI/Renderer/DiffractionEdge.h>
#include <MPEGI/Renderer/RenderItem.h>

#include <MPEGI/Helpers/Trajectory.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/SceneObject.h>

#include <algorithm>
#include <cstddef>
#include <memory>
#include <vector>

// fwd declares
namespace EAR {
class Listener;
}

namespace MPEGI {

struct RenderList
{
    //! Unique ID of the RenderList instance.
    /*!
     * The ID is counted up by 1 every time a new RenderList is produced in an update step.
     * This is mostly used for debugging.
     */
    std::size_t id;

    //! List of RenderItems
    /*!
     * Every RenderStage can update elements in the list (e.g. their eq magnitudes or status) and add new elements.
     * Note, that changes to existing items and added items are not persistent over separate runs of the update process.
     * A stage always gets a pristine copy of the RenderList as output from the preceeding stage.
     */
    std::vector<std::unique_ptr<RenderItem>> items;

    //! Data structure with relevant information about the listener (updated by the Renderer itself)
    struct ListenerState
    {
        //! Id of the reverb environment the listener is located in (name of the acoustic environment)
        std::string acousticEnvironmentId;

        //! Listener Position (location, orientation and cspace) contains the current values (sampled at the last update).
        EAR::Position position;

        //! Position of the real listener in the listening room.
        EAR::Position realPosition;

        //! Trajectory (provided for the audio thread to determine anticipated locations for every frame, e.g. for smooth doppler shifts)
        std::shared_ptr<Trajectory> trajectory;

        //! The coordinate system can be used to transform other global coordinates into local coordinates relative to the listener (e.g. for HRIR angles)
        EAR::HomogenMatrix cs;

        //! Teleport flag for the listener (indicating non-interpolated movement)
        bool teleport;

        //! Flag if the listener moved since the last update step
        bool hasMoved;

        //! Flag if the listener has moved significantly in any positional direction
        bool coarseMoved;

        //! Flag if the listener has moved significantly on any rotation axis
        bool coarseRotated;
    };
    ListenerState listener;

    //! Constructor
    RenderList(std::size_t id)
        : id(id){};

    //! Add a deep copy of an existing item
    /*!
     * \param item      raw ptr to the item
     */
    inline void addItemCopy(RenderItem* item)
    {
        // only check for duplicates in debug mode for performance reasons
        EAR_CHECK(getItemByIndex(item->idx) == nullptr,
            "RenderItem with index already exists in RenderList.");

        items.push_back(std::make_unique<RenderItem>(*item));
    };

    //! Replace an existing item with a new one (if one with the same index exists)
    /*!
     * \param item      raw ptr to the item
     */
    inline void replaceItemWithCopy(RenderItem* item)
    {
        auto it = std::find_if(items.begin(), items.end(),
            [item](std::unique_ptr<RenderItem>& ri) { return ri->idx == item->idx; });

        EAR_CHECK_REL(
            it != items.end(), "Trying to replace a non-existing item in the RenderList.");

        *it = std::make_unique<RenderItem>(*item);
    };

    //! Find an item based on the index
    /*!
     * \param idx           index of the item
     * \return              a pointer to the corresponding RenderItem in primaryItems (nullptr if nothing found)
     */
    inline const RenderItem* getItemByIndex(int idx) const
    {
        auto it = std::find_if(items.begin(), items.end(),
            [idx](const std::unique_ptr<RenderItem>& ri) { return ri->idx == idx; });

        if (it != items.end()) {
            return it->get();
        }

        return nullptr;
    }

    //! Diffraction static paths described with edge center locations
    std::map<int, std::vector<std::vector<EAR::Vector3>>> diffractionPaths;
};

} // namespace MPEGI

#endif /* MPEGI_RENDER_LIST */
