/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDER_STAGE
#define MPEGI_RENDER_STAGE

#include <MPEGI/Renderer.h>
#include <MPEGI/Renderer/RenderStageObserver.h>
#include <MPEGI/Scene.h>

#include <MPEGI/Stream/StreamManager.h>

#include <MPEGI/Renderer/RenderList.h>

#include <atomic>
#include <map>
#include <mutex>
#include <vector>

namespace MPEGI {

// This is defined outside RenderStage to be able to forward-declare it:
struct RenderStageDspData
{
    // Default constructor is needed to allow implicit default constructor in derived class:
    RenderStageDspData() = default;
    // Copy constructor is disabled to enforce usage of the object managed by the framework:
    RenderStageDspData(const RenderStageDspData&) = delete;
    // Move constructor is disabled to make sure the data stays in the originally allocated memory:
    RenderStageDspData(RenderStageDspData&&) = delete;
    virtual ~RenderStageDspData() = default;
};

//! RenderStage Interface
class RenderStage
{
public:
    //! Destructor
    virtual ~RenderStage(){};

    //! Base class for data to be passed from update thread to audio thread
    using DspData = RenderStageDspData;

    //! Helper class that's passed to processBlock()
    struct DspContext
    {
        //! Base-class reference to data, use getData() to convert to derived type.
        DspData& data;

        //! Convert the data to the given type.
        //! This must be the exact same type as has been returned from update() (this is not checked!).
        template <typename T> T& getData() { return static_cast<T&>(this->data); }
    };

    //! Contribute to the creation of a new RenderList and reconfigure own internal state accordingly
    /*!
     * This method is running at CONTROL RATE, and therefore in a different thread than processBlock(). Do not share resources directly between these two methods.
     *
     * update() is called for all stages from first to last, in order to let every stage build up on the modifications of previous stages to the RenderList.
     *
     * \param renderList    a reference to the new RenderList as processed by the preceeding RenderStage
     * \return A pointer to data (derived from RenderStage::DspData) which is passed to processBlock() (via DspContext).
     *     If `nullptr` is returned, processBlock() is *not* called!
     */
    virtual std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) = 0;

    //! Process a new block of audio samples
    /*!
     * This method is running at BLOCK RATE. Make sure to minimize computational effort here.
     *
     * processBlock() is called for all stages from first to last, in order to produce the samples required for subsequent stages.
     * The pointers to the samples are known from the RenderList
     */
    virtual void processBlock(DspContext ctx) {}

    //! Registration callback
    /*!
     * This is only called once when the Renderer is constructed (guaranteed to be before the first call to update() or processBlock())
     *
     * \param bitstream             a bistream instance to retrieve relevant payloads for the initialization of this stage
     * \param itemStore             pointer to the RenderItemStore, where new RenderItems can be created and managed
     * \param streamManager         pointer to the StreamManager, where new or existing buffers and information about the current audio configuration may be requested
     * \param state                 pointer to the currently valid SceneState that allows requesting additional information like Geometry
     */
    virtual void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) = 0;

    RenderingStageObserver* renderingStageObserver;
    virtual void setObserver(RenderingStageObserver* renderStageObserver)
    {
        renderingStageObserver = renderStageObserver;
    }

#ifdef MPEGI_MEASURE_STAGE_TIMES
    std::chrono::steady_clock::duration consumedUpdateTime =
        std::chrono::steady_clock::duration::zero();
    std::chrono::steady_clock::duration consumedProcessTime =
        std::chrono::steady_clock::duration::zero();
    int updateNumBlocks = 0;
    int processNumBlocks = 0;
    std::vector<int> consumedUpdateTimePerFrame;
    std::vector<int> consumedProcessTimePerFrame;
#endif
};

class RenderStageNoOp : public RenderStage
{
protected:
    std::unique_ptr<RenderStage::DspData> update(RenderList&) override { return {}; }

    void initializeStage(
        const Bitstream&, RenderItemStore*, const StreamManager*, const SceneState*) override
    {}
};

} // namespace MPEGI

#endif /* MPEGI_RENDER_STAGE */
