/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RENDER_STAGE_REGISTRY
#define MPEGI_RENDER_STAGE_REGISTRY

#include <map>
#include <memory>
#include <string>
#include <vector>

#include <EAR/Exception.h>

namespace MPEGI {

// fwd declares
class Bitstream;
class RenderStage;
class Spatializer;

//! Create an instance of a Spatializer
template <class T> std::shared_ptr<Spatializer> createSpatializerInstance()
{
    return std::make_shared<T>();
}

//! Provide a factory method for all registered RenderStages by their name
class RenderStageRegistry
{
public:
    using CreateStageInstanceFunc = std::shared_ptr<RenderStage> (*)();

    //! Register a stage class and associate it with a string name
    static void registerStage(std::string const& name, CreateStageInstanceFunc func)
    {
        getInstance().availableStages[name] = func;
    }

    using CreateSpatializerInstanceFunc = std::shared_ptr<Spatializer> (*)();

    //! Register a spatializer type and associate it with a string name
    static void registerSpatializer(std::string const& name, CreateSpatializerInstanceFunc func)
    {
        getInstance().availableSpatializers[name] = func;
    }

    //! Create a new stage instance by name
    static std::shared_ptr<RenderStage> createStage(std::string const& name)
    {
        auto inst = getInstance();
        auto it = inst.availableStages.find(name);
        EAR_CHECK_REL(it != inst.availableStages.end(),
            "Trying to instanciate unknown RenderStage \"" + name + "\".");
        return it->second();
    }

    //! Create a new Spatializer instance by name
    static std::shared_ptr<Spatializer> createSpatializer(std::string const& name)
    {
        auto inst = getInstance();
        auto it = inst.availableSpatializers.find(name);
        EAR_CHECK_REL(it != inst.availableSpatializers.end(),
            "Trying to instanciate unknown Spatializer \"" + name + "\".");
        return it->second();
    }

    //! Return a list of the names of all registered stages
    static std::vector<std::string> getAvailableStages()
    {
        std::vector<std::string> result;
        for (auto& stage : getInstance().availableStages) {
            result.push_back(stage.first);
        }

        return result;
    }

    //! Return a list of the names of all registered spatializers
    static std::vector<std::string> getAvailableSpatializers()
    {
        std::vector<std::string> result;
        for (auto& stage : getInstance().availableSpatializers) {
            result.push_back(stage.first);
        }

        return result;
    }

private:
    static RenderStageRegistry& getInstance();

    std::map<std::string, CreateStageInstanceFunc> availableStages;
    std::map<std::string, CreateSpatializerInstanceFunc> availableSpatializers;
};

} // namespace MPEGI

#endif /* MPEGI_RENDER_STAGE_REGISTRY */
