/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPATIALIZER
#define MPEGI_SPATIALIZER

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <mutex>
#include <string>
#include <vector>

// fwd declares
namespace EAR {
class RealFrame;
}

namespace MPEGI {

// fwd declares
class Bitstream;
struct RenderList;
class Renderer;
class SceneState;

//! Spatializer Interface
class Spatializer
{
public:
    //! Destructor
    virtual ~Spatializer(){};

    //! Base class for data to be passed from update thread to audio thread
    struct DspData
    {
        virtual ~DspData() = default;
    };

    //! Helper class that's passed to requestBlock()
    struct DspContext
    {
        //! Base-class reference to data, use getData() to convert to derived type.
        DspData& data;

        //! Reference to the RealFrame where the output samples of the Spatializer should be written.
        //! The number of channels required varies by Spatializer type and can be found out with getNbrOfChannels().
        EAR::RealFrame& output;

        //! Convert the data to the given type.
        //! This must be the exact same type as has been returned from update() (this is not checked!).
        template <typename T> T& getData() { return static_cast<T&>(this->data); }
    };

    //! Provide a new RenderList instance
    /*!
     * This method is running at CONTROL RATE, and therefore in a different thread than requestBlock(). Do not share resources directly between these two methods.
     * Additionally, update() is called irregularily to reconfigure for a new Scene.
     *
     * \param newRenderList         a constant reference to a RenderList that is valid until it is replaced by a new one.
     * \return A pointer to data (derived from Spatializer::DspData) which is passed to requestBlock() (via DspContext).
     *     This must not be `nullptr`!
     */
    virtual std::unique_ptr<Spatializer::DspData> update(const RenderList& newRenderList) = 0;

    //! Request a new block of output audio samples
    /*!
     * This method is running at BLOCK RATE. Make sure to minimize computational effort here.
     *
     * The pointers to the input samples are known from the RenderList.
     *
     * \param ctx Contains DspData from the latest update() call and a reference to the output frame.
     */
    virtual void requestBlock(DspContext ctx) = 0;

    //! Return the number of output channels this spatializer wants to fill
    virtual std::size_t getNbrOfChannels() const = 0;

private:
    friend class Renderer;

    //! Registration callback (only accessible by Renderer)
    /*!
     * This may be called again to reinitialize the Spatializer when stream properties change
     * while the rendering is not running.
     *
     * \param bitstream             a bistream instance to retrieve relevant payloads for the initialization of this spatializer
     * \param fs                    sampling rate [Hz]
     * \param B                     block size [samples]
     * \param state                 pointer to SceneState
     */
    virtual void initializeSpatializer(
        const Bitstream& bitstream, float fs, std::size_t B, const SceneState* state) = 0;
};

//! Extension interface of Spatializer to allow loading HRIRs
class Binauralizer
{
public:
    //! Load a new HRIR set for binauralization
    /*!
     * \param path      full or relative path to a HRIR file
     *                  (format depends on spatializer, probably SOFA)
     */
    virtual void loadHRIRFile(const std::string& path) = 0;
};

} // namespace MPEGI

#endif /* MPEGI_SPATIALIZER */
