/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SCENE
#define MPEGI_SCENE

#include <LSDF/LSDFParser.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Scene/SceneStateObserver.h>
#include <MPEGI/Scene/Update.h>

#include <MPEGI/Helpers/Clock.h>
#include <MPEGI/Helpers/Controllable.h>
#include <MPEGI/Helpers/TapeDeck.h>

#include <atomic>
#include <memory>
#include <set>
#include <unordered_map>
#include <utility>
#include <vector>

// fwd declares
namespace EAR {
enum class CSpace;
class Geometry;
class Listener;
class Quaternion;
class SceneObject;
template <class Tx, class Ty> class Interpolator;
template <class Tx, class Ty> class LinearInterpolator;
class Vector3;
} // namespace EAR

namespace MPEGI {

// fwd declares
class Bitstream;
class OSCListener;

// shorthands for types
typedef EAR::LinearInterpolator<double, EAR::Vector3> LocationInterpolator;
typedef EAR::LinearInterpolator<double, EAR::Quaternion> OrientationInterpolator;
typedef EAR::LinearInterpolator<double, float> FloatInterpolator;

//! Scene
/*!
 * The Scene takes care to keep the SceneState up-to-date with updates from various sources.
 * While the SceneState has no "sense of time" and always reflects the current state of all SceneObjects,
 * the Scene holds the SceneClock and stays in sync with the audio samples that are played out.
 * The Scene may also hold an OscListener, that receives external updates via OSC messages.
 *
 * Note the following behavior regarding the 3 different types of updates (dynamic, timed, triggered):
 * - The behavior is undefined when a SceneObject property is changed with multiple types of updates simultaneously
 *   (i.e. when an ObjectSource is moving on a timed trajectory, it should not be moved by user interaction)
 * - Updates with “duration” are automatically assumed to be interpolating,
 *   independent of the “transition” attribute
 * - The “duration” attribute on a modification is ignored for non-interpolated values.
 *   Instead, the updates are executed immediately.
 * - Temporary interpolations are stopped when the scene time is looped.
 * - Start value for a temporary interpolation is always the value at the time,
 *   the update is triggered (not considering delay here).
 */
class Scene : SceneStateObserver, public Controllable
{
public:
    //! Constructor
    /*
     * \param bitstream           a Bitream instance to get the needed payloads
     * \param sceneClock          a MPEGI::Clock instance, that keeps track of the current scene time
     * \param offlineMode         if set to true, the scene will update its internal state on every advanceTime() call (default is false)
     * \param enableOscListener   if set to false, the Scene will not listen to OSC messages (default is true)
     * \param lsdf                optional pointer to an LSDFParser instance with an LSDF file for AR configuration (default nullptr)
     */
    Scene(const Bitstream& bitstream, std::shared_ptr<Clock> sceneClock, bool offlineMode = false,
        bool enableOscListener = true, const LSDF::LSDFParser* lsdf = nullptr);

    //! Destructor
    ~Scene();

    //! Set Scene time to a specific value in seconds
    /*!
     * This can be called from any thread and triggers a reconfiguration of the Scene at the next control step.
     * Currently running temporary interpolations are deleted.
     *
     * \param t              time in seconds
     */
    void setTime(double t);

    //! Return the current Scene time in seconds
    double getCurrentTime() const;

    //! Return pointer to the read-only SceneState managed by this Scene
    inline const SceneState* getState() const { return &state; };

    //! Trigger an update by its EIF integer id to be executed immediately
    /*!
     * \param idx                integer index like in EIF
     * \param placeholderValues  a vector of key-value pairs for placeholders as defined in the EIF update
     */
    void triggerUpdateByIndex(int idx, std::vector<std::pair<int, float>> placeholderValues = {});

    //! Trigger an update by its EIF id to be executed immediately
    /*!
     * \param id                 string id like in EIF
     * \param placeholderValues  a vector of key-value pairs for placeholders as defined in the EIF update
     */
    void triggerUpdateById(
        const std::string& id, std::vector<std::pair<int, float>> placeholderValues = {});

    //! Pass an arbitrary, externally constructed Update to be executed immediately
    void triggerUpdate(const Update& upd);

    //! Inherited from SceneStateObserver
    void sceneStateAttached(const SceneState* state) override{};

    //! Inherited from SceneStateObserver
    void sceneStateDetached() override{};

    //! Inherited from SceneStateObserver
    void sceneObjectAdded(EAR::SceneObject* object) override;

    //! Inherited from SceneStateObserver
    void sceneObjectRemoved(EAR::SceneObject* object) override;

    //! Return pointer to the OSC listener (e.g. to start and stop recording)
    inline OSCListener* getOSCListener() { return oscListener.get(); };

private:
    //! Inherited from Controllable
    void controlStep() override;

    inline void parsePlaceholders(const Modification& mod, EAR::Variant* val);
    inline void addUpdate(const BitstreamUpdate& upd);

    inline void checkProximityTriggers();

    SceneState state;

    std::unique_ptr<OSCListener> oscListener;
    friend OSCListener;

    bool offlineMode;

    std::shared_ptr<Clock> sceneClock;
    double pt; //!< previous update step Scene time
    //! duration until scene time loops around (used for OSC playbacks that wrap around)
    double sceneDuration;

    // storage for all updates and interpolators

    //! Data structure for permanent interpolators
    struct PermanentInterpolator
    {
        std::unique_ptr<LocationInterpolator> location;
        std::unique_ptr<OrientationInterpolator> orientation;
        std::unique_ptr<FloatInterpolator> gain;
        std::unique_ptr<FloatInterpolator> directiveness;
        std::unique_ptr<FloatInterpolator> itemReverbGain;
        std::unique_ptr<FloatInterpolator> loudness;
        std::unique_ptr<LocationInterpolator> granularPos;
        std::unique_ptr<LocationInterpolator> granularGroupProb;

        //! Constructor
        /*!
         * \param obj       pointer to scene object (needed to initialize start points)
         */
        PermanentInterpolator(EAR::SceneObject* obj);
    };
    std::unordered_map<std::string, std::unique_ptr<PermanentInterpolator>> permanentInterpolators;

    // Data structure for permanent interpolation (timed updates)
    std::vector<std::unique_ptr<Update>> timedUpdateStore;
    TapeDeck<Update> timedUpdates; //!< timed, non-interpolated updates

    //! Data structure for temporary interpolation (e.g. for triggered trajectory)
    struct TempInterpolator
    {
        std::string entity;
        std::string attribute;
        EAR::Variant targetVal;
        double duration;

        //! Constructor
        TempInterpolator(
            const std::string& e, const std::string& a, double d, const EAR::Variant& t)
            : entity(e)
            , attribute(a)
            , duration(d)
            , targetVal(t){ EAR_CHECK(
                  duration > 0, "A temporary interpolator should have a positive duration.") };
    };

    //! Data structure for a running temporary interpolation (keeping track of the remaining time)
    struct ActiveTempInterpolator
    {
        double endTime; //!< end timestamp, so it does NOT work across scene time loop boundaries

        const TempInterpolator* parent;

        // only one of the following is initialized (based on attribute)
        std::unique_ptr<LocationInterpolator> location;
        std::unique_ptr<OrientationInterpolator> orientation;
        std::unique_ptr<FloatInterpolator> gain;
        std::unique_ptr<FloatInterpolator> directiveness;
        std::unique_ptr<FloatInterpolator> itemReverbGain;
        std::unique_ptr<FloatInterpolator> loudness;
        std::unique_ptr<LocationInterpolator> granularPos;

        ActiveTempInterpolator(
            double st, const TempInterpolator* p, const EAR::Variant& initialVal);
    };
    std::vector<ActiveTempInterpolator> activeTempInterpolators;

    std::vector<std::string> previousMovingObjects;

    // //! Data structure for proximity-triggered updates
    struct ProximityTrigger
    {
        std::unique_ptr<Update> update;
        std::vector<std::unique_ptr<TempInterpolator>> tempInterpolators;

        const EAR::Geometry* region;
        //!< whether the update should be triggered when active changes from false to true or vice versa
        bool flank;
        double delay;
        bool active = false;

        //! Constructor
        ProximityTrigger(EAR::Geometry* r, bool f, double d)
            : region(r)
            , flank(f)
            , delay(d){};
    };
    std::vector<ProximityTrigger> proximityTriggers;

    //! Data structure for externally triggered updates
    struct ExternalTrigger
    {
        std::unique_ptr<Update> update;
        std::vector<std::unique_ptr<TempInterpolator>> tempInterpolators;
        std::map<int, float*> placeholders;

        double delay;

        //! Empty constructor
        ExternalTrigger(double d)
            : delay(d){};
    };
    std::vector<std::unique_ptr<ExternalTrigger>> externalTriggers;
    std::unordered_map<std::string, ExternalTrigger*> externalTriggersById;
    std::unordered_map<int, ExternalTrigger*> externalTriggersByIndex;

    std::set<std::string> idsWithUpdates; //!< used for checking if an object is static or not
};

} // namespace MPEGI

#endif /* MPEGI_SCENE */
