/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_OSC_LISTENER
#define MPEGI_OSC_LISTENER

#include <MPEGI/Helpers/Controllable.h>
#include <MPEGI/Helpers/TapeDeck.h>

#include <MPEGI/Scene.h>
#include <MPEGI/Scene/Update.h>

#include <EAR/Exception.h>
#include <EAR/OSC.h>
#include <EAR/Variant.h>

#include <chrono>
#include <string>
#include <vector>

namespace MPEGI {

//! The OSCListener handles incoming OSC messages/bundles and creates Scene Updates from them.
class OSCListener : public EAR::OSC
{

private:
    //! Helper class that starts a thread to play back the messages of one file
    class Playback : public Controllable
    {
    public:
        Playback(const std::string& serializedRecording, OSCListener* parent);
        ~Playback();

        void controlStep() override;

        void advanceTime(double dt);

        inline double getLastUpdateTime() { return lastUpdate; };

    private:
        OSCListener* parent;
        std::vector<std::pair<double, EAR::OSC::Bundle::Element>> messages;
        double seconds, lastUpdate;
        std::chrono::time_point<std::chrono::high_resolution_clock> startTime;
        TapeDeck<const EAR::OSC::Bundle::Element> tapedeck;
    };

public:
    //! Constructor
    /*!
     * \param scene         ptr to Scene that is modified by incoming messages or playbacks
     * \param offlineMode   boolean to enable "offline mode", where OSC playbacks are triggered externally instead in their own thread
     */
    OSCListener(Scene* scene, bool offlineMode);

    //! Destructor
    ~OSCListener();

    //! Loads an OSC recording and sends all messages to the SceneState according to the timestamps given in the file
    /*!
     * \param file      full string from a recording file as returned by EAR::OSC::serializeRecording()
     * \return          the playback length in seconds
     */
    double playbackRecording(const std::string& serializedRecording);

    //! Stops all running OSC recording playbacks
    void stopAllPlaybacks();

    //! Triggers a playback step with a specified time in "offline mode" (see constructor)
    /*!
     * This throws an error in Debug mode if it is called while NOT in offline mode.
     *
     * \param dt         how much time passed since last call (in seconds)
     */
    void triggerPlayback(double dt);

private:
    //! Parse and execute an OSC message that was received or played back
    void parseMessage(const EAR::OSC::Message& message);

    friend Playback;

    Scene* scene;
    std::vector<std::unique_ptr<Playback>> runningPlaybacks;

    const bool offlineMode;

    bool receivedTeleport; // needed for the weird AEP message /teleport
};

} // namespace MPEGI

#endif /* MPEGI_OSC_LISTENER */
