/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SCENE_STATE
#define MPEGI_SCENE_STATE

#include <LSDF/LSDFParser.h>

#include <EAR/DirectivityModel.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayload.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Helpers/VoxelTypes.h>
#include <MPEGI/Scene/SceneStateObserver.h>

#include <complex>
#include <map>
#include <mutex>
#include <string>
#include <unordered_map>
#include <vector>

// fwd declares
namespace EAR {
class AnchorObject;
class Box;
class ComplexBuffer;
class Cylinder;
class Geometry;
class HOASource;
class Listener;
class LoudspeakerLayout;
class Material;
class Mesh;
class PointSource;
class Portal;
class SceneObject;
class Sphere;
class AirflowGeneratorBase;
class GlobalAirflowGenerator;
class PointAirflowGenerator;
class ConeAirflowGenerator;
class PlaneAirflowGenerator;
} // namespace EAR

namespace MPEGI {

// fwd declare
class Scene;
class Update;

//! Scene state
class SceneState
{
public:
    //! A representation of the playback state of an AudioStream
    struct AudioStream
    {
        std::string id; //!< unique identifier from the EIF
        std::vector<std::size_t>
            channels; //!< an optional subset of the channels in the AudioStream with the given ID
        bool looping; //!< stream should restart after it ends
        bool playing; //!< stream should start playing immediately
        double seekTime; //!< stream should start at this time [s] (-1 indicates continued playback)

        //! Constructor
        AudioStream()
            : id("")
            , looping(true)
            , playing(true)
            , seekTime(0.0){};

        //! Constructor with id
        AudioStream(const std::string& id)
            : id(id)
            , looping(true)
            , playing(true)
            , seekTime(0.0){};
    };

    //! A representation of GranularSound
    struct GranularSound
    {
        std::string id; //!< unique identifier from the EIF
        int channels; //!< onumber of channels in GranularSound with the given ID

        //! Constructor
        GranularSound()
            : id(""){};

        //! Constructor with id
        GranularSound(const std::string& id)
            : id(id){};
    };

    //! Constructor
    /*
     * \param bitstream          a Bitstream instance to get the needed payloads
     * \param parent             Pointer to the parent scene (needed for access to the current scene time in interpolation)
     * \param lsdf               optional pointer to an LSDFParser instance with an LSDF file for AR configuration (default nullptr)
     */
    SceneState(
        const Bitstream& bitstream, const Scene* parent, const LSDF::LSDFParser* lsdf = nullptr);

    //! Destructor
    ~SceneState();

    //! Register a new SceneObserver (observing creation and deletion of SceneObjects)
    /*!
     * Make sure to call detach() (or to destruct the SceneState) before deleting the object that is observing the SceneState.
     *
     * \param observer        Pointer to a SceneObserver instance that will be notified if the SceneState changes
     */
    void attach(SceneStateObserver* observer) const;

    //! Remove a registered SceneObserver
    /*!
     * \param observer        Pointer to a SceneObserver instance to be removed from the list of observers. If this is not a valid observer, nothing happens.
     */
    void detach(SceneStateObserver* observer) const;

    //! Provide a new external update
    /*!
     * \param update          a timed update. If the timestamp is in the past according to the current Scene time, it is applied immediately.
     */
    void modify(const Update& update);

    //! Return true if the movement state of the object should NOT be interpolated ("teleport")
    /*!
     * \param id              id of the object of interest
     */
    bool isTeleporting(const std::string& id) const;

    //! get the current cell based on the point given
    /*!
     * Returns nullptr if the given position is not in any of the cells.
     */
    Cell* getCellForLocation(const EAR::Vector3& point);

    //! Get all HOAGroup representations in the SceneState
    inline const std::vector<std::unique_ptr<MPEGI::HOAGroup>>& getHOAGroups() const
    {
        return hoaGroups;
    }

    //! Get all HOASource representations in the SceneState
    inline const std::vector<std::unique_ptr<EAR::HOASource>>& getHOASources() const
    {
        return hoaSources;
    }

    //! Get all ObjectSource representations in the SceneState
    inline const std::vector<std::unique_ptr<EAR::PointSource>>& getObjectSources() const
    {
        return objectSources;
    }

    //! Get all anchors representations in the SceneState
    inline const std::vector<std::unique_ptr<EAR::AnchorObject>>& getAnchors() const
    {
        return anchors;
    }

    //! Get all LoudspeakerLayout representations in the SceneState
    inline const std::vector<std::unique_ptr<EAR::LoudspeakerLayout>>& getLoudspeakerLayouts() const
    {
        return lsLayouts;
    }

    inline const std::optional<LSDF::LSDFParser::ChannelSourceDesc>& getLoudspeakerSetup() const
    {
        return loudspeakerSetup;
    }

    inline const std::vector<std::unique_ptr<EAR::Portal>>& getPortals() const { return portals; }

    //! Get a Geometry by its ID
    /*!
     * Returns nullptr if no Geometry with this ID was found.
     *
     * \param id    id as in bitstream
     */
    inline EAR::Geometry* getGeometry(const std::string& id) const
    {
        EAR::Geometry* geo = nullptr;
        if (geometries.find(id) != geometries.end()) {
            geo = geometries.at(id);
        }
        return geo;
    }

    //! Get all Geometries in the SceneState (globally, not separated by reverb environment)
    inline std::vector<const EAR::Geometry*> getAllGeometry() const
    {
        std::vector<const EAR::Geometry*> geo;
        for (auto& pair : geometries) {
            geo.push_back(pair.second);
        }
        return geo;
    }

    //! Get all visible Geometries in the SceneState (globally without the ones used for acustic environments, not separated by reverb environment)
    inline const std::vector<EAR::Geometry*>& getAllVisibleGeometry() const
    {
        return visibleGeometries;
    }

    //! Get all Geometries in the SceneState which belong to the Early Reflection Geometry
    inline const std::vector<const EAR::Geometry*>& getEarlyReflectionGeometry() const
    {
        return earlyReflectionGeometries;
    }

    //! Get all Meshes in the SceneState
    inline const std::vector<std::unique_ptr<EAR::Mesh>>& getMeshes() const { return meshes; }

    //! Get combined Early Reflection Mesh
    inline const std::unique_ptr<EAR::Mesh>& getCombinedEarlyReflectionMesh() const
    {
        return combinedEarlyReflectionMesh;
    }

    //! Get combined Diffraction Mesh
    inline const std::unique_ptr<EAR::Mesh>& getCombinedDiffractionMesh() const
    {
        return combinedDiffractionMesh;
    }

    virtual inline const std::vector<std::unique_ptr<ReverbEnv>>& getReverbEnvs() const
    {
        return reverbEnvs;
    }

    //! Get all directivities
    inline const std::map<std::string, std::shared_ptr<EAR::DirectivityModel>>&
    getDirectivities() const
    {
        return directivities;
    }
    //! Get listener
    inline const EAR::Listener* getListener() const { return listener.get(); }

    //! Check if there is an AudioStream for a given AudioElement
    inline bool hasAudioStream(EAR::SceneObject* obj) const
    {
        if (audioStreamProperties.find(obj) == audioStreamProperties.end()) {
            return false;
        } else {
            return true;
        }
    }

    //! Get AudioStream for a given AudioElement
    inline const AudioStream& getAudioStream(EAR::SceneObject* obj) const
    {
        return audioStreamProperties.at(obj);
    }

    double getCurrentTime() const;

    //! Check if there is a GranularSound for a given AudioElement
    inline bool hasGranularSound(EAR::SceneObject* obj) const
    {
        if (GranularSoundProperties.find(obj) == GranularSoundProperties.end()) {
            return false;
        } else {
            return true;
        }
    }

    //! Get GranularSound for a given AudioElement
    inline const GranularSound* getGranularSound(EAR::SceneObject* obj) const
    {
        if (GranularSoundProperties.find(obj) == GranularSoundProperties.end()) {
            return nullptr;
        }
        return &(GranularSoundProperties.at(obj));
    }

    //! Get authoring parameters for a given AudioElement
    inline AParam getAParams(EAR::SceneObject* obj) const { return authoringParams.at(obj); }

    //! Get extent layout for a given AudioElement
    inline const ExtentLayout* getExtentLayout(EAR::SceneObject* obj) const
    {
        if (extentLayouts.find(obj) == extentLayouts.end()) {
            return nullptr;
        }

        return &(extentLayouts.at(obj));
    }

    inline Voxel::VoxelSceneState::shared getVoxelSceneState() const { return voxelSceneState; }

    inline const std::vector<std::shared_ptr<EAR::AirflowGeneratorBase>>&
    getAirflowGenerators() const
    {
        return airflowGenerators;
    }

    inline const Scene* getParentScene() const { return parentScene; }

protected:
    //! Constructor
    SceneState() noexcept;

    void createTransform(const EIFPayload::TransformDesc& tf);
    void createAnchor(const EIFPayload::AnchorDesc& a);
    void createObjectSource(const EIFPayload::ObjectSourceDesc& obj);
    void createChannelSource(const EIFPayload::ChannelSourceDesc& obj);
    void createHOAGroup(const EIFPayload::HoaGroupDesc& obj);
    void createHOASource(const EIFPayload::HoaSourceDesc& obj);
    void createDirectivity(
        const std::string& dirString, std::shared_ptr<EAR::DirectivityModel> dir);
    void createMaterial(const EIFPayload::AcousticMaterialDesc& mat);
    void createBox(const EIFPayload::BoxDesc& box);
    void createSphere(const EIFPayload::SphereDesc& sph);
    void createCylinder(const EIFPayload::CylinderDesc& cyl);
    void createMesh(const EIFPayload::MeshDesc& mesh);
    void createReverbEnv(const EIFPayload::AcousticEnvironmentDesc& env);
    void createPortal(MPEGI::Portal* portal);
    void createLSDFLocalSource(
        const LSDF::LSDFParser::LocalSourceDesc& obj, const std::string& directivityId = "");
    void createLSDFMaterial(const LSDF::LSDFParser::AcousticMaterialDesc& mat);
    void createLSDFMesh(const LSDF::LSDFParser::MeshDesc& mesh);
    void createLSDFReverbEnv(const LSDF::LSDFParser::AcousticEnvironmentDesc& env);
    void createVoxelSceneState(std::shared_ptr<class VoxDataPayload const> const& voxDataPayload);
    void createGlobalAirflowGenerator(const EIFPayload::GlobalAirflowGeneratorDesc& gen);
    void createPointAirflowGenerator(const EIFPayload::PointAirflowGeneratorDesc& gen);
    void createConeAirflowGenerator(const EIFPayload::ConeAirflowGeneratorDesc& gen);
    void createPlaneAirflowGenerator(const EIFPayload::PlaneAirflowGeneratorDesc& gen);

    void combineEarlyReflectionMesh();
    void combineDiffractionMesh();
    void combineMesh(std::vector<const EAR::Geometry*>& geometries, Mesh* meshHandle);

    AParam mapAuthoringParameters(const EIFPayload::AudioElementDesc::AuthoringParameter param);
    AParam mapLSDFAuthoringParameters(
        const LSDF::LSDFParser::AudioElement::AuthoringParameter param);

private:
    mutable std::vector<SceneStateObserver*> observers;

    const Scene* parentScene;

    friend Scene; // needs to access updateable to set them to non-static if needed

    std::map<std::string, EAR::SceneObject*> updateable;
    std::map<EAR::SceneObject*, AudioStream> audioStreamProperties;
    std::map<EAR::SceneObject*, GranularSound> GranularSoundProperties;
    std::map<std::string, bool> teleport;
    std::map<EAR::SceneObject*, ExtentLayout> extentLayouts;
    std::map<EAR::SceneObject*, AParam> authoringParams;

    std::unique_ptr<EAR::Listener> listener;

    std::vector<std::unique_ptr<EAR::PointSource>> objectSources;
    std::vector<std::unique_ptr<EAR::HOASource>> hoaSources;
    std::vector<std::unique_ptr<EAR::LoudspeakerLayout>> lsLayouts;

    std::optional<LSDF::LSDFParser::ChannelSourceDesc> loudspeakerSetup;

    std::map<std::string, std::shared_ptr<EAR::Material>> materials;
    std::map<std::string, EAR::Geometry*> geometries;
    std::map<std::string, EAR::AnchorObject*> anchorsById;
    std::map<std::string, EAR::AnchorObject*> anchorsByLSDFRef; // TODO: will be used in AR setup
    std::vector<std::unique_ptr<EAR::AnchorObject>> anchors;
    std::vector<std::unique_ptr<EAR::Mesh>> meshes;
    std::vector<std::unique_ptr<EAR::Box>> boxes;
    std::vector<std::unique_ptr<EAR::Sphere>> spheres;
    std::vector<std::unique_ptr<EAR::Cylinder>> cylinders;
    std::vector<std::unique_ptr<ReverbEnv>> reverbEnvs;
    std::map<std::string, std::unique_ptr<Cell>> cells;
    std::vector<std::unique_ptr<EAR::Portal>> portals;
    std::vector<EAR::Geometry*> visibleGeometries;
    std::vector<const EAR::Geometry*> earlyReflectionGeometries;
    std::vector<const EAR::Geometry*> diffractionGeometries;
    std::unique_ptr<EAR::Mesh> combinedEarlyReflectionMesh;
    std::unique_ptr<EAR::Mesh> combinedDiffractionMesh;

    std::vector<std::shared_ptr<EAR::AirflowGeneratorBase>> airflowGenerators;

    std::map<std::string, std::shared_ptr<EAR::DirectivityModel>> directivities;

    std::vector<std::string> hoaGroupIds; // used to create unique int ids
    std::vector<std::unique_ptr<MPEGI::HOAGroup>> hoaGroups;

    Voxel::VoxelSceneState::shared voxelSceneState;

    std::mutex sequentialUpdateMutex;
};

} // namespace MPEGI

#endif /* MPEGI_SCENE_STATE */
