/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_STREAM_ACCESS_BUFFER
#define MPEGI_STREAM_ACCESS_BUFFER

#include <atomic>
#include <cstddef>
#include <string>
#include <vector>

#include <MPEGI/Stream/StreamBuffer.h>

namespace MPEGI {

// fwd declares
class StreamManager;

//! A StreamAccessBuffer provides access to samples in a StreamManager
class StreamAccessBuffer : public StreamBuffer
{
public:
    //! Destructor
    ~StreamAccessBuffer();

    //! Change the AudioStream this AccessBuffer refers to (cross-faded over one block to avoid discontinuities)
    /*!
     * \param streamId      string id of the new AudioStream to be accessed
     * \param t             initial read cursor position [s]
     *                      (-1 indicates continuous playback, i.e. continuing the new stream where the previous stopped)
     */
    void setStream(const std::string& id, double t = 0.0);

    //! Start playing from the current cursor position (if not already playing)
    void play();

    //! Stop playing (AccessBuffer is filled with zeros)
    void stop();

    //! Move the read cursor to a specific time of the AudioStream (thread-safe)
    /*!
     * \param t             new read cursor position [s] (default 0)
     */
    void seekTo(double t);

    //! Set whether the AccessBuffer should restart from the beginning of the AudioStream if it reaches its end
    /*!
     * \param shouldLoop    come on...
     */
    void setLoop(bool shouldLoop);

    //! Get current read cursor position in samples
    std::size_t getReadPosition() const;

    //! Get current stream id
    inline const std::string& getStreamId() const { return streamId; };

private:
    // StreamManager implementation have to be introduced as friend explicitly
    friend class FileStreamManager;
    friend class MaxStreamManager;
    friend class VSTStreamManager;

    //! Clears the buffer, when the StreamManager is destroyed
    void unregister();

    //! Constructor (only accessible by StreamManager)
    /*!
     * \param parent        pointer to the parent StreamManager instance, where new samples are read from
     * \param streamId      string id of the AudioStream to be accessed (may be changed later)
     * \param t             initial read cursor position [s] (default 0)
     * \param playing       stream should start playing from the beginning
     * \param looping       stream should restart after it reached the end
     * \param channels      an optional subset of the channels in the AudioStream with the given ID
     */
    StreamAccessBuffer(StreamManager* parent, const std::string& streamId, double t = 0.0,
        bool playing = true, bool looping = true, const std::vector<std::size_t>& channels = {});

    StreamManager* parent;

    std::atomic<std::size_t> rc;

    std::string streamId; // TODO: not thread-safe!
    std::vector<std::size_t> channels;

    std::string prevStreamId;

    std::atomic<double> seekToTime;

    std::atomic<bool> playing;
    std::atomic<bool> looping;
    std::atomic<bool> playStateChange;
};

} // namespace MPEGI

#endif /* MPEGI_STREAM_ACCESS_BUFFER */
