/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_STREAM_MANAGER
#define MPEGI_STREAM_MANAGER

#include <MPEGI/Stream/StreamAccessBuffer.h>
#include <MPEGI/Stream/StreamManager.h>

#include <MPEGI/Scene/SceneState.h>

#include <cstddef>
#include <memory>

namespace MPEGI {

//! Stream manager Interface
class StreamManager
{
public:
    //! Constructor
    /*!
     * \param fs            sample rate [Hz]
     * \param B             block size [samples]
     */
    StreamManager(float fs, std::size_t B)
        : fs(fs)
        , B(B){};

    //! Destructor
    virtual ~StreamManager(){};

    //! Create a new empty buffer of type StreamBuffer with the right size B
    /*!
     * \param ch            number of channels in the new buffer (defaults to 1)
     */
    inline std::shared_ptr<StreamBuffer> createEmptyBuffer(std::size_t ch = 1) const
    {
        return std::make_shared<StreamBuffer>(ch, B);
    };

    //! Create a new StreamAccessBuffer instance
    /*!
     * \param streamId      string id of the AudioStream to be accessed (may be changed later)
     * \param t             initial read cursor position [s]
     * \return              shared pointer to a StreamAccessBuffer. New samples will be copied into it on every nextBlock() call.
     */
    virtual std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const std::string& streamId, double t = 0.0) = 0;

    //! Create a new StreamAccessBuffer instance
    /*!
     * \param stream        SceneState::AudioStream instance
     */
    virtual std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const SceneState::AudioStream& stream) = 0;

    //! Remove an obsolete StreamAccessBuffer
    /*!
     * \param buf           pointer to the instance to be removed
     */
    virtual void removeAccessBuffer(StreamAccessBuffer* buf) = 0;

    //! Copy the next block of samples into all active StreamAccessBuffers
    virtual void nextBlock() = 0;

    //! Return number of channels in a specified audio stream
    /*!
     * \param id            stream identifier
     */
    virtual std::size_t getNbrOfChannels(const std::string& id) const = 0;

    inline const float getFs() const { return fs; }

    inline const std::size_t getB() const { return B; }

protected:
    friend class StreamAccessBuffer;

    //! Sample rate [Hz]
    const float fs;
    //! Block size
    const std::size_t B;
};

} // namespace MPEGI

#endif /* MPEGI_STREAM_MANAGER */
