/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Scene/OSCListener.h>

#include <MPEGI/Helpers/Config.h>

#include <algorithm>
#include <utility>
#include <vector>
#ifdef USE_JSON_TRAJECTORY
#include <nlohmann/json.hpp>
#endif

namespace MPEGI {

OSCListener::Playback::Playback(const std::string& serializedRecording, OSCListener* parent)
    : parent(parent)
    , seconds(0.0)
    , lastUpdate(0.0)
{
#ifndef USE_JSON_TRAJECTORY
    // construct all EAR::OSC::Elements from the serialized recordings
    messages = EAR::OSC::deserializeRecording(serializedRecording);
#else
    {
        using json = nlohmann::json;
        auto traj = json::parse(serializedRecording);

        messages.clear();
        for (auto const& rec : traj["records"]) {
            OSC::Message msg(rec["address"]);

            for (auto const& arg : rec["args"]) {
                if (arg.type() == json::value_t::number_float) {
                    msg.addArgument(EAR::Variant(arg.template get<float>()));
                } else if (arg.type() == json::value_t::number_integer
                    || arg.type() == json::value_t::number_unsigned) {
                    msg.addArgument(EAR::Variant(arg.template get<int>()));
                }
            }

            auto const timestamp = rec["timestamp"];
            messages.push_back(std::make_pair(timestamp, msg));
        }
    }
#endif
    // register with tapedeck
    for (auto& msg : messages) {
        tapedeck.addElement(msg.first, &msg.second);

        if (lastUpdate < msg.first) {
            lastUpdate = msg.first;
        }
    }

    if (!parent->offlineMode) {
        setControlRate(Config::getIntDef("OSCControlRate", 300));
        startTime = std::chrono::high_resolution_clock::now();
        startController();
    }
}

OSCListener::Playback::~Playback()
{
    stopController();
}

void OSCListener::Playback::controlStep()
{
    if (!parent->offlineMode) {
        seconds = std::chrono::duration_cast<std::chrono::duration<double>>(
            std::chrono::high_resolution_clock::now() - startTime)
                      .count();
    }

    auto msgs = tapedeck.playback(seconds);

    for (auto msg : msgs) {
        if (msg->isMessage()) {
            parent->parseMessage(msg->getMessage());
        } else { // is bundle
            for (const auto& bundleElement : msg->getBundle()) {
                EAR_CHECK_REL(bundleElement.isMessage(),
                    "Nested bundles currently not supported by OSC playback (blame sws' "
                    "lazyness).");
                parent->parseMessage(bundleElement.getMessage());
            }
        }
    }

    // if all recordings passed, stop the thread
    if (seconds > lastUpdate) {
        controllerRunning.store(false);
        // TODO: also self-destruct object?
    }
}

void OSCListener::Playback::advanceTime(double dt)
{
    seconds += dt;
}

OSCListener::OSCListener(Scene* scene, bool offlineMode)
    : scene(scene)
    , offlineMode(offlineMode)
    , EAR::OSC(
          "127.0.0.1", Config::getIntDef("OSCPort", MPEGI_OSC_LISTENER_PORT), 1024, false, false)
    , receivedTeleport(false)
{
    if (!offlineMode) {
        addListener("/mpeg-i/update/position",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/mpeg-i/update/orientation",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/mpeg-i/update/gain",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener(
            "/mpeg-i/update", std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/mpeg-i/start-recording",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/mpeg-i/stop-recording",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/mpeg-i/set-config",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));

        // mirror AEP messages for compatibility
        addListener("/InteractionEventID",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener(
            "/userPose", std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/DynamicUpdates/Position",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/DynamicUpdates/Rotation",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener("/interactionEvent",
            std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));
        addListener(
            "/teleport", std::bind(&OSCListener::parseMessage, this, std::placeholders::_1));

        connect();
    }
}

OSCListener::~OSCListener()
{
    stopRecordingReceivedPackets();
    stopAllPlaybacks();
}

void OSCListener::parseMessage(const EAR::OSC::Message& message)
{
    auto args = message.getArguments();

    if (message.getOscAddress() == "/mpeg-i/update/position") {
        Update upd;

        EAR_CHECK(args.size() == 7 || args.size() == 8,
            "Invalid number of arguments for OSC position update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC position update has to be the entity ID.");

        bool interpolate = (args.size() == 8 && static_cast<int>(args[7]) == 0) || args.size() == 7;

        upd.addModification(
            args[0], "location", EAR::Vector3(args[1], args[2], args[3]), interpolate);
        upd.addModification(args[0], "velocity", EAR::Vector3(args[4], args[5], args[6]), false);

        scene->triggerUpdate(upd);
    } else if (message.getOscAddress() == "/mpeg-i/update/orientation") {
        Update upd;

        EAR_CHECK(args.size() == 4, "Invalid number of arguments for OSC orientation update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC orientation update has to be the entity ID.");

        upd.addModification(args[0], "orientation", EAR::Euler(args[1], args[2], args[3]), false);

        scene->triggerUpdate(upd);
    } else if (message.getOscAddress() == "/mpeg-i/update/gain") {
        Update upd;

        EAR_CHECK(args.size() == 2, "Invalid number of arguments for OSC gain update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC gain update has to be the entity ID.");

        upd.addModification(args[0], "gainDb", args[1], false);

        scene->triggerUpdate(upd);
    } else if (message.getOscAddress() == "/mpeg-i/update/reverbGainDb") {
        Update upd;

        EAR_CHECK(args.size() == 2, "Invalid number of arguments for OSC reverbGainDb update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC reverbGainDb update has to be the entity ID.");
        float itemReverbGainDb = args[1];
        float itemReverbGain = pow(10, itemReverbGainDb / 20);
        upd.addModification(args[0], "itemReverbGain", itemReverbGain, false);

        scene->triggerUpdate(upd);
    } else if (message.getOscAddress() == "/mpeg-i/update/noReverb") {
        Update upd;

        EAR_CHECK(args.size() == 2, "Invalid number of arguments for OSC noReverb update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC noReverb update has to be the entity ID.");

        upd.addModification(args[0], "noReverb", args[1], false);

        scene->triggerUpdate(upd);
    } else if (message.getOscAddress() == "/mpeg-i/update") {
        EAR_CHECK(args.size() == 1, "Invalid number of arguments for OSC external update");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "The update ID must be a string");

        auto strarg = static_cast<std::string>(args[0]);

        scene->triggerUpdateById(strarg);
    } else if (message.getOscAddress() == "/mpeg-i/start-recording") {
        startRecordingReceivedPackets();
    } else if (message.getOscAddress() == "/mpeg-i/stop-recording") {
        stopRecordingReceivedPackets();
    } else if (message.getOscAddress() == "/mpeg-i/set-config") {
        if (args.size() == 32) {
            std::vector<float> vecTemp(args.size() - 1, 0.f);
            for (int k = 0; k < args.size() - 1; k++) {
                if (args[k + 1].getVariantType() == EAR::Variant::Type::Int) {
                    vecTemp[k] = (int)args[k + 1];
                } else {
                    vecTemp[k] = (float)args[k + 1];
                }
            }
            Config::set(args[0], vecTemp);
        }
        // we just ignore invalid messages and don't throw an exception
        else if (args.size() >= 2 && args[0].getVariantType() == EAR::Variant::Type::String) {
            Config::set(args[0], args[1]);
        }

    } else if (message.getOscAddress() == "/InteractionEventID") {
        EAR_CHECK(args.size() == 1, "Invalid number of arguments for OSC external update");

        EAR_CHECK(
            args[0].getVariantType() == EAR::Variant::Type::Int, "The update ID must be an int");

        auto arg = static_cast<int>(args[0]);

        scene->triggerUpdateByIndex(arg);
    } else if (message.getOscAddress() == "/interactionEvent") {
        EAR_CHECK(
            args[0].getVariantType() == EAR::Variant::Type::Int, "The update ID must be an int");

        auto arg = static_cast<int>(args[0]);
        std::vector<std::pair<int, float>> placeholderValues;

        for (int i = 1; i < args.size(); i++) {
            EAR_CHECK(args[i].getVariantType() == EAR::Variant::Type::Float,
                "The update placeholder value must be of type float.");
            placeholderValues.push_back(std::make_pair(i - 1, static_cast<float>(args[i])));
        }

        scene->triggerUpdateByIndex(arg, placeholderValues);
    } else if (message.getOscAddress() == "/userPose") {
        Update upd;

        EAR_CHECK(args.size() >= 6, "Invalid number of arguments for OSC position update.");

        upd.addModification(
            "listener", "location", EAR::Vector3(args[0], args[1], args[2]), !receivedTeleport);
        upd.addModification("listener", "velocity", EAR::Vector3(0, 0, 0), false);
        upd.addModification(
            "listener", "orientation", EAR::Euler(args[3], args[4], args[5]), false);

        if (args.size() >= 12) {
            upd.addModification(
                "realListener", "location", EAR::Vector3(args[6], args[7], args[8]));
            upd.addModification(
                "realListener", "orientation", EAR::Euler(args[9], args[10], args[11]));
        }

        scene->triggerUpdate(upd);

        receivedTeleport = false;
    } else if (message.getOscAddress() == "/teleport") {
        receivedTeleport = true;
    } else if (message.getOscAddress() == "/DynamicUpdates/Position") {
        Update upd;

        EAR_CHECK(args.size() == 5, "Invalid number of arguments for OSC position update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC position update has to be the entity ID.");

        upd.addModification(args[0], "location", EAR::Vector3(args[1], args[2], args[3]),
            static_cast<bool>(args[4]));
        upd.addModification(args[0], "velocity", EAR::Vector3(0, 0, 0), false);

        scene->triggerUpdate(upd);
    } else if (message.getOscAddress() == "/DynamicUpdates/Rotation") {
        Update upd;

        EAR_CHECK(args.size() == 4, "Invalid number of arguments for OSC orientation update.");

        EAR_CHECK(args[0].getVariantType() == EAR::Variant::Type::String,
            "First argument of OSC orientation update has to be the entity ID.");

        upd.addModification(args[0], "orientation", EAR::Euler(args[1], args[2], args[3]), false);

        scene->triggerUpdate(upd);
    }
}

double OSCListener::playbackRecording(const std::string& serializedRecording)
{
    runningPlaybacks.emplace_back(std::make_unique<Playback>(serializedRecording, this));

    return runningPlaybacks.back()->getLastUpdateTime();
}

void OSCListener::stopAllPlaybacks()
{
    runningPlaybacks.clear();
}

void OSCListener::triggerPlayback(double dt)
{
    EAR_CHECK(offlineMode, "OSC playback can only be triggered in offline mode.");

    for (auto& pb : runningPlaybacks) {
        pb->advanceTime(dt);
        pb->controlStep();
    }
}

} // namespace MPEGI
