/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include <MPEGI/Helpers/OcclusionRecovery.h>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
void occlusion_recovery::init(Voxel::renderer_instruction_map_t const& instructions)
{
    for (auto const& [mat_id, i] : instructions) {
        if (!mat_id) {
            continue;
        }

        auto const recovery_setting = i.occlusion.recovery;
        if (!_recovery_by_factor.count(recovery_setting)) {
            init_table(recovery_setting, _recovery_by_factor[recovery_setting]);
        }

        _recovery_by_material[mat_id] = _recovery_by_factor[recovery_setting];
    }
}

//--------------------------------------------------------------------------------------------------
void occlusion_recovery::apply_recovery(int material_id, float sqr_distance,
    std::vector<float> const& t0, std::vector<float>& t1, std::vector<float>& compensation_eq) const
{
    auto const recovery_fac = get_recovery_factor(material_id, sqr_distance);

    // Calculate compensation EQs to limit energy of diffracted source, if any
    // Also compensate if recovery_fac == 0, to account for the energy that was not absorbed by the occluder
    compensation_eq.resize(t0.size());
    std::transform(std::cbegin(t0), std::cend(t0), std::cbegin(t1), std::begin(compensation_eq),
        [recovery_fac](
            auto const& lhs, auto const& rhs) { return (1 - recovery_fac) * (lhs - rhs); });

    if (recovery_fac > 0.f) {
        // Apply recovery factor and copy result into t1
        std::transform(std::cbegin(t0), std::cend(t0), std::cbegin(t1), std::begin(t1),
            [recovery_fac](auto const& lhs, auto const& rhs) {
                return recovery_fac * lhs + (1 - recovery_fac) * rhs;
            });
    }
}

//--------------------------------------------------------------------------------------------------
void occlusion_recovery::apply_recovery(int material_id, float sqr_distance, std::vector<float>& t)
{
    static std::vector<float> not_used;
    std::vector<float> const ones(t.size(), 1.f);

    apply_recovery(material_id, sqr_distance, ones, t, not_used);
}

//--------------------------------------------------------------------------------------------------
bool occlusion_recovery::has_recovery(int material_id) const
{
    auto const it = _recovery_by_material.find(material_id);
    return (it == std::cend(_recovery_by_material)) ? false : !it->second.empty();
}

//--------------------------------------------------------------------------------------------------
float occlusion_recovery::get_recovery_factor(int material_id, float sqr_magnitude) const
{
    if (!_recovery_by_material.count(material_id)) {
        return 0.f;
    }

    auto const& table = _recovery_by_material.at(material_id);
    if (table.empty()) {
        return 0.f;
    }

    auto const end = std::cend(table);
    auto const beg = std::cbegin(table);
    auto recovery_it = std::find_if(
        beg, end, [sqr_magnitude](auto const& dr) { return dr.first >= sqr_magnitude; });

    return (recovery_it == beg) ? beg->second : (recovery_it - 1)->second;
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------