/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "MPEGI/Helpers/VoxelTraversal.h"
#include <EAR/BaseTypes.h>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {

bool voxel_traversal::run(Voxel::voxel_scene_info_basic const& scene_info,
    std::array<float, 3> const& lis_xyz_vox, voxel_index_t const& lis_idx,
    std::array<float, 3> const& src_xyz_vox, voxel_index_t const& src_idx,
    traversal_pred_t const& pred)
{
    std::array<int, 3> step_xyz = { 0, 0, 0 };
    std::array<float, 3> tVoxel_xyz = { 0.0, 0.0, 0.0 };
    voxel_index_t xyz_idx;
    std::array<float, 3> vec_xyz;
    std::array<float, 3> dir_xyz;
    std::array<float, 3> voxelMax_xyz;
    std::array<float, 3> tMax_xyz;
    std::array<float, 3> tDelta;
    float norm;

    xyz_idx = lis_idx;

    vec_xyz[0] = src_xyz_vox[0] - lis_xyz_vox[0];
    vec_xyz[1] = src_xyz_vox[1] - lis_xyz_vox[1];
    vec_xyz[2] = src_xyz_vox[2] - lis_xyz_vox[2];

    norm = sqrtf(vec_xyz[0] * vec_xyz[0] + vec_xyz[1] * vec_xyz[1] + vec_xyz[2] * vec_xyz[2]);

    dir_xyz[0] = vec_xyz[0] / norm;
    dir_xyz[1] = vec_xyz[1] / norm;
    dir_xyz[2] = vec_xyz[2] / norm;

    for (int i = 0; i < 3; i++) {
        if (xyz_idx[i] == (scene_info.dim[i] + 1))
            xyz_idx[i] = xyz_idx[i] - 1;

        if (dir_xyz[i] >= 0) {
            tVoxel_xyz[i] = (double)xyz_idx[i] / scene_info.dim[i];
            step_xyz[i] = 1;
        } else {
            tVoxel_xyz[i] = (double)(xyz_idx[i] - 1) / scene_info.dim[i];
            step_xyz[i] = -1;
        }
    }

    voxelMax_xyz[0] = scene_info.p0[0] + tVoxel_xyz[0] * (scene_info.p1[0] - scene_info.p0[0]);
    voxelMax_xyz[1] = scene_info.p0[1] + tVoxel_xyz[1] * (scene_info.p1[1] - scene_info.p0[1]);
    voxelMax_xyz[2] = scene_info.p0[2] + tVoxel_xyz[2] * (scene_info.p1[2] - scene_info.p0[2]);

    tMax_xyz[0] = (voxelMax_xyz[0] - lis_xyz_vox[0]) / dir_xyz[0];
    tMax_xyz[1] = (voxelMax_xyz[1] - lis_xyz_vox[1]) / dir_xyz[1];
    tMax_xyz[2] = (voxelMax_xyz[2] - lis_xyz_vox[2]) / dir_xyz[2];

    tDelta[0] = scene_info.voxel_size / std::abs(dir_xyz[0]);
    tDelta[1] = scene_info.voxel_size / std::abs(dir_xyz[1]);
    tDelta[2] = scene_info.voxel_size / std::abs(dir_xyz[2]);

    _path.clear();

    while ((xyz_idx[0] * step_xyz[0] <= src_idx[0] * step_xyz[0])
        && (xyz_idx[1] * step_xyz[1] <= src_idx[1] * step_xyz[1])
        && (xyz_idx[2] * step_xyz[2] <= src_idx[2] * step_xyz[2])
        && (xyz_idx[0] <= scene_info.dim[0]) && (xyz_idx[0] >= 1)
        && (xyz_idx[1] <= scene_info.dim[1]) && (xyz_idx[1] >= 1)
        && (xyz_idx[2] <= scene_info.dim[2]) && (xyz_idx[2] >= 1)) {

        _path.push_back({ xyz_idx[0], xyz_idx[1], xyz_idx[2] });
        if (pred && !pred(xyz_idx)) {
            return !_path.empty();
        }

        if (tMax_xyz[0] < tMax_xyz[1]) {
            if (tMax_xyz[0] < tMax_xyz[2]) {
                xyz_idx[0] = xyz_idx[0] + step_xyz[0];
                tMax_xyz[0] = tMax_xyz[0] + tDelta[0];
            } else {
                xyz_idx[2] = xyz_idx[2] + step_xyz[2];
                tMax_xyz[2] = tMax_xyz[2] + tDelta[2];
            }
        } else {
            if (tMax_xyz[1] < tMax_xyz[2]) {
                xyz_idx[1] = xyz_idx[1] + step_xyz[1];
                tMax_xyz[1] = tMax_xyz[1] + tDelta[1];
            } else {
                xyz_idx[2] = xyz_idx[2] + step_xyz[2];
                tMax_xyz[2] = tMax_xyz[2] + tDelta[2];
            }
        }
    }

    return _path.size() > 0;
}

bool voxel_traversal::run(Voxel::voxel_scene_info_basic const& scene_info,
    EAR::Vector3 const& lis_xyz, EAR::Vector3 const& src_xyz, traversal_pred_t const& pred)
{
    auto lis_xyz_vox = Voxel::opengl_to_vox(lis_xyz.data());
    auto lis_idx = scene_info.to_voxel_grid(lis_xyz).first;
    auto src_xyz_vox = Voxel::opengl_to_vox(src_xyz.data());
    auto src_idx = scene_info.to_voxel_grid(src_xyz, lis_xyz).first;

    return run(scene_info, lis_xyz_vox, lis_idx, src_xyz_vox, src_idx, pred);
}

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
