/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
**************************************************************************/
#include <MPEGI/Bitstream/BinaryStream.h>
#include <MPEGI/Bitstream/VoxDataPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/VoxelTypes.h>
#include <MPEGI/Renderer/RenderItem.h>
#include <MPEGI/Renderer/RenderItemStore.h>

#include "SmitsAlgorithmC.hpp"

#include <climits>
#include <fstream>
#include <iomanip>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {
//--------------------------------------------------------------------------------------------------
#ifdef GRID_DEBUG
void dump_grid(grid const& g)
{
    static std::vector<int> const cycle = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };

    static int cycle_idx = 0;

    auto const file_path{ "/tmp/grid." + std::to_string((cycle_idx++ % cycle.size())) + ".bin" };

    std::ofstream os(file_path, std::ios_base::binary);
    if (!os.good())
        return;

    auto const dim_x = g.get_dim()[0];
    auto const dim_y = g.get_dim()[1];
    auto const dim_z = g.get_dim()[2];

    os.write(reinterpret_cast<char const*>(&dim_x), sizeof(dim_x));
    os.write(reinterpret_cast<char const*>(&dim_y), sizeof(dim_y));
    os.write(reinterpret_cast<char const*>(&dim_z), sizeof(dim_z));

    auto const& voxels = g.get_voxels();

    for (auto const& vox : g.get_voxels()) {
        auto const p = vox.get_properties();
    }
}
#endif
//--------------------------------------------------------------------------------------------------
#ifdef PM_DEBUG
void dump_pm(projection_map const& out, std::string const& out_file_path)
{
    auto const dim_x = out.dim[0];
    auto const dim_y = out.dim[1];

    std::ofstream pm_out(out_file_path);
    if (!pm_out.good())
        return;

    for (auto y = 0U; y < dim_y; ++y) {
        for (auto x = 0U; x < dim_x; ++x) {
            pm_out << static_cast<int>(out.at(x, y)) << " ";
        }

        pm_out << std::endl;
    }
}
#endif
//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
voxel const& grid::at(voxel_index_val_t x, voxel_index_val_t y, voxel_index_val_t z) const
{
    static voxel const invalid;

    auto const idx{ (z - 1) * _dim[0] * _dim[1] + (y - 1) * _dim[0] + (x - 1) };

    return (idx < _voxels.size()) ? _voxels[idx] : invalid;
}

//--------------------------------------------------------------------------------------------------
grid::grid(std::vector<Payload::block<3>> const& blocks, voxel_index_t const& grid_dimension,
    renderer_instruction_map_t const& instructions)
    : _dim(grid_dimension)
    , _voxels(_dim[0] * _dim[1] * _dim[2], (voxel()))
{
    update_grid(blocks, instructions);
#ifdef GRID_DEBUG
    dump_grid(*this);
#endif
}

//--------------------------------------------------------------------------------------------------
void grid::update_grid(
    std::vector<Payload::block<3>> const& blocks, renderer_instruction_map_t const& instructions)
{
    for (auto const& b : blocks) {
        auto const m_id = b.get_value();

        if (!instructions.count(m_id)) {
            throw std::logic_error(
                "[grid::update_grid] No instructions for material ID: " + std::to_string(m_id));
        }

        auto const& instr = instructions.at(m_id);
        auto const sound_medium = (m_id == 0);

        if (!in_range(b.p0[0], b.p0[1], b.p0[2]) || !in_range(b.p1[0], b.p1[1], b.p1[2])) {
            throw std::out_of_range("[grid::update_grid] Voxel-index out-of-range");
        }

        // Bitstream block indices are 1-based
        for (auto x = b.p0[0]; x <= b.p1[0]; ++x) {
            for (auto y = b.p0[1]; y <= b.p1[1]; ++y) {
                for (auto z = b.p0[2]; z <= b.p1[2]; ++z) {
                    // Set voxel properties
                    auto& vxl = at(x, y, z);
                    vxl.set_material_id(m_id);
                    vxl.enable_diffraction(instr.diffraction.enabled);
                    vxl.set_sound_medium(sound_medium);
                }
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
void process_one_dimension(std::vector<std::vector<std::vector<uint16_t>>>& vox_matrix_id,
    uint16_t material_id, std::array<unsigned int, 3>& ps, std::array<unsigned int, 3>& pe,
    std::array<bool, 3>& checked_dim, std::array<unsigned int, 3>& dim)
{
    voxel_index_t i_e;
    std::array<std::vector<voxel_index_val_t>, 3> D;
    std::array<std::vector<voxel_index_val_t>, 3> B;

    for (auto j = 0; j < 3; j++) {
        i_e[j] = 0;
        D[j].clear();
        B[j].clear();

        for (auto k = ps[j]; k <= pe[j]; k++) {
            B[j].push_back(k);
        }
    }

    for (auto k = 0; k < 3; k++) {
        if (checked_dim[k] == false) {
            for (auto i = ps[k]; i < dim[k]; i++) {
                i_e[k] = i + 1;
                auto P = std::min(i_e[k], dim[k] - 1);
                for (auto j = 0; j < 3; j++) {
                    D[j] = B[j];
                }
                D[k] = std::vector<voxel_index_val_t>(1, P);
                bool is_empty = false;
                for (auto x : D[0]) {
                    for (auto y : D[1]) {
                        for (auto z : D[2]) {
                            if (vox_matrix_id[x][y][z] != material_id) {
                                is_empty = true;
                                break;
                            }
                        }
                        if (is_empty == true) {
                            break;
                        }
                    }
                    if (is_empty == true) {
                        break;
                    }
                }
                if (is_empty == true)
                    break;
            }
        }
    }

    int max_val = INT_MIN;
    int max_dim = 0;

    for (auto j = 0; j < 3; j++) {
        int diff = (int)i_e[j] - (int)ps[j];
        if (diff > max_val) {
            max_val = diff;
            max_dim = j;
        }
    }

    pe[max_dim] = i_e[max_dim] - 1;
    B[max_dim].clear();
    for (auto k = ps[max_dim]; k <= pe[max_dim]; k++) {
        B[max_dim].push_back(k);
    }

    for (auto x : B[0]) {
        for (auto y : B[1]) {
            for (auto z : B[2]) {
                vox_matrix_id[x][y][z] = 0;
            }
        }
    }
    checked_dim[max_dim] = true;
}

void convert_matrix_to_blocks(std::vector<std::vector<std::vector<uint16_t>>>& voxel_matrix,
    std::array<unsigned int, 3>& dim, std::vector<Payload::block<3>>& voxel_blocks)
{
    std::map<std::uint16_t, std::uint16_t> materials;

    // determinate unique material ids
    for (auto x = 0; x < dim[0]; ++x) {
        for (auto y = 0; y < dim[1]; ++y) {
            for (auto z = 0; z < dim[2]; ++z) {
                auto material_id = voxel_matrix[x][y][z];
                if (material_id != 0) {
                    if (materials.count(material_id) == 0) {
                        materials[material_id] = material_id;
                    }
                }
            }
        }
    }

    // material selection
    for (auto material_id : materials) {
        std::array<unsigned int, 3> last_ps = { 0, 0, 0 };
        do {
            bool voxel_matrix_id_is_empty = true;
            unsigned int x, y, z;
            // order is important here !
            for (z = last_ps[2]; z < dim[2]; ++z) {
                for (y = last_ps[1]; y < dim[1]; ++y) {
                    for (x = last_ps[0]; x < dim[0]; ++x) {
                        if (voxel_matrix[x][y][z] == material_id.second) {
                            voxel_matrix_id_is_empty = false;
                            break;
                        }
                    }
                    last_ps[0] = 0;
                    if (voxel_matrix_id_is_empty == false) {
                        break;
                    }
                }
                last_ps[1] = 0;
                if (voxel_matrix_id_is_empty == false) {
                    break;
                }
            }
            last_ps[0] = 0;
            // check if we are done with this material_id
            if (voxel_matrix_id_is_empty)
                break;

            // block initialization
            std::array<unsigned int, 3> ps{ x, y, z };
            std::array<unsigned int, 3> pe = ps;
            last_ps = ps;
            std::array<bool, 3> checked_dim = { false, false, false };

            for (auto i = 0; i < 3; i++) {
                process_one_dimension(voxel_matrix, material_id.second, ps, pe, checked_dim, dim);
            }

            Payload::block<3> block = { material_id.first, ps, pe };

            voxel_blocks.push_back(block);
        } while (true);
    }
}

//--------------------------------------------------------------------------------------------------
bool grid::convert(std::vector<Payload::block<3>>& blocks)
{
    std::vector<std::vector<std::vector<uint16_t>>> voxel_matrix;
    std::array<unsigned int, 3> dimension = { _dim[0], _dim[1], _dim[2] };

    voxel_matrix.resize(dimension[0],
        std::vector<std::vector<uint16_t>>(dimension[1], std::vector<uint16_t>(dimension[2])));

    // set up working copy
    for (auto x = 0; x < dimension[0]; ++x) {
        for (auto y = 0; y < dimension[1]; ++y) {
            for (auto z = 0; z < dimension[2]; ++z) {
                auto& vxl = at(x + 1, y + 1, z + 1);
                auto material_id = vxl.get_material_id();
                voxel_matrix[x][y][z] = material_id;
            }
        }
    }

    convert_matrix_to_blocks(voxel_matrix, dimension, blocks);

    return true;
}

//--------------------------------------------------------------------------------------------------
bool index_storage::update_voxel_index(int render_item_idx, voxel_index_pair_t&& voxel_idx)
{
    auto& idx_stored = storage[render_item_idx];
    if (idx_stored == voxel_idx) {
        return false;
    }

    idx_stored = std::move(voxel_idx);
    return true;
}

//--------------------------------------------------------------------------------------------------
voxel_index_pair_t voxel_scene_info_basic::to_voxel_grid(
    EAR::Vector3 const& location, EAR::Vector3 const& usr_location, bool clamp) const
{
    auto const loc_vox{ opengl_to_vox(location[0], location[1], location[2]) };
    auto const loc_idx = clamp ? xyz2idx(loc_vox) : xyz2idx_unsafe(loc_vox);

    if (clamp || is_in_range(loc_idx)) {
        return std::make_pair(loc_idx, 0.f);
    }

    // Outside the grid - find intersection with the grid boundaries
    auto const eps = voxel_size / 10.f;
    // Apply saftey margin to grid
    EAR::Vector3 const grid_min(p0[0] + eps, p0[1] + eps, p0[2] + eps);
    EAR::Vector3 const grid_max(p1[0] - eps, p1[1] - eps, p1[2] - eps);

    auto const norm = (usr_location - location).normalize();
    auto const norm_vox{ opengl_to_vox(norm[0], norm[1], norm[2]) };

    float t_min = 0.f;
    SmitsAlgorithm(loc_vox.data(), norm_vox.data(), grid_min.data(), grid_max.data(), &t_min);

    // New location, now within the grid
    auto const new_loc = (location + norm * t_min);

    return std::make_pair(xyz2idx(opengl_to_vox(new_loc.data())), t_min);
}

voxel_index_pair_t voxel_scene_info_basic::to_voxel_grid(EAR::Vector3 const& location) const
{
    return to_voxel_grid(location, { 0, 0, 0 }, true);
}

voxel_index_pair_t voxel_scene_info_basic::to_voxel_grid(
    EAR::Vector3 const& location, EAR::Vector3 const& usr_location) const
{
    return to_voxel_grid(location, usr_location, false);
}

//--------------------------------------------------------------------------------------------------
VoxelSceneState::VoxelSceneState(MPEGI::VoxDataPayload const& vox_data_payload)
    : _voxel_index_storage(std::make_shared<struct index_storage>())
    , _direct_view(std::make_shared<direct_view_map>())
    , _compensation_eq(std::make_shared<eq_map>())
{
    auto const num_subscenes = vox_data_payload.subscenes.size();

    int ss_id = 0;
    for (auto const& ss : vox_data_payload.subscenes) {
        // Assign a sub-scene ID if there are more than one subscenes in the current scene
        subscenes.push_back(std::make_shared<Voxel::subscene>(
            ss, num_subscenes > 1 ? std::optional<int>(ss_id++) : std::nullopt));
    }

    current_subscene = subscenes.at(0);
}

//--------------------------------------------------------------------------------------------------
void VoxelSceneState::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    if (obj->getName() != "listener")
        return;

    if (EAR::SceneObjectObserver::Property::Position == modification) {
        update(obj->getPosition().location);
    }
}

//--------------------------------------------------------------------------------------------------
void VoxelSceneState::update(EAR::Vector3 const& location)
{
    if (!in_subscene(location, *get_current_subscene()->scene)) {
        auto it = std::find_if(
            std::cbegin(subscenes), std::cend(subscenes), [&location, this](auto const& sub_scene) {
                return in_subscene(location, *sub_scene->scene);
            });

        if (std::cend(subscenes) != it) {
            current_subscene = *it;
            notify_subscene_change();
        }
    }
}

//--------------------------------------------------------------------------------------------------
void VoxelSceneState::trigger_update(int id, update_args_t const& args) const
{
    auto& updates = get_current_subscene()->scene_updates;
    auto it = std::find_if(
        std::begin(updates), std::end(updates), [id](auto const& upd) { return id == upd.id; });

    if (std::end(updates) != it) {
        if (args.empty()) {
            it->trigger();
        } else {
            if (args.size() == 7) {
                auto to_idx = [](std::pair<int, float> const& p) {
                    return static_cast<voxel_index_val_t>(p.second);
                };

                auto const m_id = static_cast<int>(args[0].second);
                auto const beg = voxel_index_t{ to_idx(args[1]), to_idx(args[2]), to_idx(args[3]) };
                auto const end = voxel_index_t{ to_idx(args[4]), to_idx(args[5]), to_idx(args[6]) };

                it->trigger(m_id, beg, end);
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
void VoxelSceneState::tick(std::chrono::milliseconds const& elapsed) const
{
    auto& updates = get_current_subscene()->scene_updates;
    for (auto& upd : updates) {
        upd.advance(*this, elapsed);
    }
}

//--------------------------------------------------------------------------------------------------
subscene::subscene(Payload::subscene const& subscene_payload, std::optional<int> const& id)
    : scene(std::make_shared<Voxel::voxel_scene_info_basic>(
        subscene_payload.scene.p0, subscene_payload.scene.p1, subscene_payload.scene.dim, id))
    , diffraction_map(subscene_payload.scene.dim[0], subscene_payload.scene.dim[1])
    , cache_pathfinder(
          std::make_shared<pathfinder_cache>(subscene_payload.data_diffraction->resolution))
{
    scene_updates.reserve(subscene_payload.scene_updates.size());
    for (auto const& scene_update : subscene_payload.scene_updates) {
        scene_updates.emplace_back(scene_update);
    }

    scene_materials.reserve(subscene_payload.materials.size());
    for (auto const& scene_material : subscene_payload.materials) {
        scene_materials[scene_material.id] = scene_material;
    }

    renderer_instructions.reserve(subscene_payload.data_occlusion->instructions_occlusion.size());

    for (auto const& instr : subscene_payload.data_occlusion->instructions_occlusion) {
        renderer_instructions[instr.material_id].occlusion.recovery = instr.recovery;
    }

    for (auto const& instr : subscene_payload.data_diffraction->instructions_diffraction) {
        renderer_instructions[instr.material_id].diffraction.enabled = instr.diffraction_enable;
    }

    voxel_grid = std::make_shared<Voxel::grid>(
        subscene_payload.blocks, subscene_payload.scene.dim, renderer_instructions);

    if (subscene_payload.data_diffraction->blocks_dmap.empty()) {
        // No diffraction map transmitted in the bitstream, therefore derive from grid
        set_diffraction_map(
            projection_map(*voxel_grid, subscene_payload.data_diffraction->diffraction_map_height));
    } else {
        // Use diffraction map transmitted in the bitstream
        set_diffraction_map(projection_map(subscene_payload.data_diffraction->blocks_dmap,
            subscene_payload.data_diffraction->resolution));
    }

    diffraction_effect_control = subscene_payload.data_diffraction->diffraction_effect_control;

    // fill pathfinder cache with pre computed path data (if present)
    cache_pathfinder->reset();

    for (auto const& pre_computed_path_elem :
        subscene_payload.data_diffraction->pre_computed_path_data) {
        dmap_index_t corner = { 0, 0 };
        float length = .0f;

        if (pre_computed_path_elem.src_xy.has_value()
            && pre_computed_path_elem.length.has_value()) {

            dmap_index_t corner = pre_computed_path_elem.src_xy.value();
            float length = pre_computed_path_elem.length.value();
            cache_pathfinder->set_path(pre_computed_path_elem.start_xy,
                pre_computed_path_elem.end_xy, jps_info{ length, corner });
        } else {
            cache_pathfinder->set_no_path(
                pre_computed_path_elem.start_xy, pre_computed_path_elem.end_xy);
        }
    }
}

subscene::~subscene()
{
    bool store_pre_computed_path_data_at_shutdown =
        Config::getBoolDef("VoxelStorePreComputedPathDataAtShutdown", false);

    if (store_pre_computed_path_data_at_shutdown) {
        store_precomputed_path_data(*cache_pathfinder, *scene, diffraction_map);
    }
}

//--------------------------------------------------------------------------------------------------
projection_map::projection_map(
    std::vector<Voxel::Payload::block<2>> const& blocks_diffraction_map, dmap_index_t const& res)
    : projection_map(res[0], res[1])
{
    for (auto const& el : blocks_diffraction_map) {
        auto const x0 = el.p0[0];
        auto const y0 = el.p0[1];
        auto const x1 = el.p1[0];
        auto const y1 = el.p1[1];
        auto const value = el.get_value();

        // Indices from the bitstream are in the range [1..dim]
        for (int x = x0; x <= x1; x++) {
            for (int y = y0; y <= y1; y++) {
                this->at(x - 1, y - 1) = value;
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
projection_map::projection_map(
    Voxel::grid const& grid, std::optional<voxel_index_val_t> const& height)
    : projection_map(grid.get_dim()[0], grid.get_dim()[1], height)
{
    auto const& dim = grid.get_dim();
    if (!z.has_value() || (z.has_value() && z.value() > dim[2])) {
        z = dim[2] / 2;
    }

    // Projection map height
    auto const z = this->z.value();

    // JPS needs row-first ordering
    for (auto y = 1U; y <= dim[1]; ++y) {
        for (auto x = 1U; x <= dim[0]; ++x) {
            // Map voxel indices [1..dim] to map indices [0..dim-1]
            this->at(x - 1, y - 1) = grid.at(x, y, z).is_global_occluder();
        }
    }

    EAR_CHECK(this->valid(), "VoxelDiffractionStage: Invalid projection map");
#ifdef PM_DEBUG
    dump_pm(*this, "/tmp/pm.txt");
#endif
}

bool projection_map::convert(std::vector<Payload::block<2>>& blocks) const
{
    std::vector<std::vector<std::vector<uint16_t>>> voxel_matrix;
    std::vector<Payload::block<3>> voxel_blocks;
    std::array<unsigned int, 3> dimension = { dim[0], dim[1], 1 };

    voxel_matrix.resize(dimension[0],
        std::vector<std::vector<uint16_t>>(dimension[1], std::vector<uint16_t>(dimension[2])));

    // set up working copy
    for (auto x = 0; x < dimension[0]; ++x) {
        for (auto y = 0; y < dimension[1]; ++y) {
            uint16_t value = this->at(x, y);
            voxel_matrix[x][y][0] = value;
        }
    }

    convert_matrix_to_blocks(voxel_matrix, dimension, voxel_blocks);

    // drop unused third dimension for bitstream
    for (auto& voxel_block : voxel_blocks) {

        std::array<voxel_index_val_t, 2> p0, p1;

        for (auto i = 0; i < 2; i++) {
            p0[i] = voxel_block.p0[i] - 1;
            p1[i] = voxel_block.p1[i] - 1;
        }

        Payload::block<2> block(voxel_block.value, p0, p1);
        blocks.push_back(block);
    }
    return true;
}

//--------------------------------------------------------------------------------------------------
scene_element::scene_element(Payload::block<3> const& block)
    : m_id(block.get_value())
    , p0(block.p0)
    , p1(block.p1)
{}

scene_element::scene_element(int m_id_, voxel_index_t const& p0_, voxel_index_t p1_)
    : m_id(m_id_)
    , p0(p0_)
    , p1(p1_)
{}

//--------------------------------------------------------------------------------------------------
sequence_point::sequence_point(Voxel::Payload::sequence_point const& sp_pay)
    : duration(sp_pay.duration)
{
    scene_elements.reserve(sp_pay.blocks.size());
    for (auto const& b : sp_pay.blocks) {
        scene_elements.emplace_back(b);
    }

    if (!sp_pay.data_diffraction->blocks_dmap.empty()) {
        diffraction_map = projection_map(
            sp_pay.data_diffraction->blocks_dmap, sp_pay.data_diffraction->resolution);
    }

    if (!sp_pay.data_diffraction->instructions_diffraction.empty()) {
        instructions.emplace();
    }

    for (auto const& instr : sp_pay.data_diffraction->instructions_diffraction) {
        (*instructions)[instr.material_id].diffraction.enabled = instr.diffraction_enable;
    }

    this->diffraction_map_height = sp_pay.data_diffraction->diffraction_map_height;
}

sequence_point::sequence_point(scene_element const& el)
{
    scene_elements.push_back(el);
}

//--------------------------------------------------------------------------------------------------
scene_update::scene_update(Payload::scene_update const& scene_update_pay)
    : id(scene_update_pay.id)
{
    _sequence.reserve(scene_update_pay.sequence.size());
    for (auto const& seq : scene_update_pay.sequence) {
        _sequence.emplace_back(seq);
    }

    // Update sequence not active
    _cur_sequence_point = std::cend(_sequence);
}

//--------------------------------------------------------------------------------------------------
void scene_update::advance(VoxelSceneState const& state, milliseconds const& elapsed)
{
    using namespace std::chrono_literals;

    if (ended())
        return;

    _duration = std::max(0ms, _duration - elapsed);
    if (0ms != _duration) {
        return;
    }

    auto const& subscene = state.get_current_subscene();
    auto const& instructions = subscene->renderer_instructions;
    auto& grid_ = *subscene->voxel_grid;

    // Apply grid updates and notify observers
    for (auto const& ele : _cur_sequence_point->scene_elements) {
        if (!instructions.count(ele.m_id)) {
            throw std::logic_error("[scene_update::advance] No instructions for material ID: "
                + std::to_string(ele.m_id));
        }

        auto const& instr = instructions.at(ele.m_id);
        for (auto x = ele.p0[0]; x <= ele.p1[0]; ++x) {
            for (auto y = ele.p0[1]; y <= ele.p1[1]; ++y) {
                for (auto z = ele.p0[2]; z <= ele.p1[2]; ++z) {
                    auto& vox = grid_.at(x, y, z);

                    vox.set_material_id(ele.m_id);
                    vox.enable_diffraction(instr.diffraction.enabled);
                    vox.set_sound_medium(ele.m_id == 0);
                }
            }
        }
    }

    if (_cur_sequence_point->diffraction_map.has_value()) {
        subscene->set_diffraction_map(*_cur_sequence_point->diffraction_map);
    } else {
        auto const height = _cur_sequence_point->diffraction_map_height.has_value()
            ? _cur_sequence_point->diffraction_map_height
            : subscene->diffraction_map.get_height();
        subscene->set_diffraction_map(projection_map(grid_, height));
    }

    state.notify_scene_update();
    next_sequence_point();
}

//--------------------------------------------------------------------------------------------------
void scene_update::trigger()
{
    if (!_sequence.empty()) {
        // Update data was provided in the bitstream
        _cur_sequence_point = std::cbegin(_sequence);
        _duration = milliseconds(_cur_sequence_point->duration);
    }
}

void scene_update::trigger(
    int m_id, Voxel::voxel_index_t const& beg, Voxel::voxel_index_t const& end)
{
    using namespace std::chrono_literals;

    // Update is provided during runtime;
    _sequence.clear();
    _sequence.emplace_back(scene_element(m_id, beg, end));

    // Exectue immediately
    _cur_sequence_point = std::cbegin(_sequence);
    _duration = 0ms;
}

//--------------------------------------------------------------------------------------------------
void scene_update::next_sequence_point()
{
    using namespace std::chrono_literals;

    if (std::cend(_sequence) == _cur_sequence_point)
        return;

    ++_cur_sequence_point;
    _duration = (std::cend(_sequence) == _cur_sequence_point)
        ? 0ms
        : milliseconds(_cur_sequence_point->duration);
}

//--------------------------------------------------------------------------------------------------
auto scene_material::operator=(Payload::scene_material const& material) -> scene_material&
{
    bs_id = material.bs_id;
    return *this;
}

//--------------------------------------------------------------------------------------------------
void store_precomputed_path_data(pathfinder_cache const& cache_pathfinder,
    voxel_scene_info_basic const& scene, Voxel::projection_map const& diffraction_map)
{
    std::string out_file_path = Config::getString("VoxelPreComputedPathDataDirectory");
    out_file_path += "/pcpd";

    if (scene.id.has_value()) {
        out_file_path += "_";
        out_file_path += std::to_string(scene.id.value());
    }
    out_file_path += ".bin";

    MPEGI::binaryStream outstream(out_file_path);

    if (outstream.good()) {
        unsigned int bitpos = 7;
        std::bitset<8> bits;

        // generate compressed projection map
        std::vector<Payload::block<2>> converted_blocks;
        diffraction_map.convert(converted_blocks);

        // write out actual compressed projection map
        const unsigned int numberOfVoxDiffractionMapElements = converted_blocks.size();
        outstream.writeEscapedUInt(8, 16, 32, numberOfVoxDiffractionMapElements);

        auto const nBitsMap = Voxel::Payload::calc_bit_length(diffraction_map.dim);

        for (auto i = 0U; i < numberOfVoxDiffractionMapElements; ++i) {
            unsigned int voxDiffractionMapValue = 1;
            unsigned int voxDiffractionMapPosPackedS =
                Voxel::Payload::pack(converted_blocks[i].p0, diffraction_map.dim);
            unsigned int voxDiffractionMapPosPackedE =
                Voxel::Payload::pack(converted_blocks[i].p1, diffraction_map.dim);

            outstream.writeBitsUInt(1, voxDiffractionMapValue);
            outstream.writeBitsUInt(nBitsMap, voxDiffractionMapPosPackedS);
            outstream.writeBitsUInt(nBitsMap, voxDiffractionMapPosPackedE);
        }

        // write out actual pre computed path data
        const unsigned int numberOfVoxDiffractionPathData = cache_pathfinder.get_lut_size();
        outstream.writeEscapedUInt(8, 16, 32, numberOfVoxDiffractionPathData);

        auto lut = cache_pathfinder.get_lut();

        for (auto const& s : lut) {
            dmap_index_t src;
            src[0] = s.first % scene.dim[0] + 1;
            src[1] = s.first / scene.dim[0] + 1;

            unsigned int voxDiffractionPathStartVoxelPacked =
                Voxel::Payload::pack(src, diffraction_map.dim);

            for (auto const& d : s.second) {
                dmap_index_t dst;
                dst[0] = d.first % scene.dim[0] + 1;
                dst[1] = d.first / scene.dim[0] + 1;

                unsigned int voxDiffractionPathEndVoxelPacked =
                    Voxel::Payload::pack(dst, diffraction_map.dim);
                outstream.writeBitsUInt(nBitsMap, voxDiffractionPathStartVoxelPacked);
                outstream.writeBitsUInt(nBitsMap, voxDiffractionPathEndVoxelPacked);

                auto pi = d.second;

                unsigned int voxDiffractionPathDataExistFlag = pi.length.has_value();
                outstream.writeBitsUInt(voxDiffractionPathDataExistFlag, 1);

                if (voxDiffractionPathDataExistFlag) {
                    unsigned int voxDiffractionSourceDirectionPacked =
                        Voxel::Payload::pack(pi.corner, diffraction_map.dim);
                    float length = pi.length.value();
                    outstream.writeBitsUInt(nBitsMap, voxDiffractionPathEndVoxelPacked);
                    outstream.writeBitsUInt(32, *reinterpret_cast<unsigned int*>(&length));
                }
            }
        }
        // flush remaining bits
        outstream.flush();
    }
}

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
