/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Renderer.h>

#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageObserver.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Helpers/Trajectory.h>

#include <EAR/HOASource.h>
#include <EAR/Listener.h>
#include <EAR/ListenerVoiceSource.h>
#include <EAR/LoudspeakerLayout.h>
#include <EAR/PointSource.h>
#include <EAR/RealFrame.h>
#include <EAR/SceneObject.h>

#include <mpegcore/spdlog.hpp>

#include <algorithm>
#include <cassert>
#include <numeric>

namespace {
template <typename T> auto microseconds(T duration)
{
    return std::chrono::duration_cast<std::chrono::microseconds>(duration).count();
}
} // namespace

namespace MPEGI {

Renderer::CtorArgs Renderer::createStages(const Bitstream& bitstream, const SceneState* sceneState,
    std::shared_ptr<StreamManager> streamManager)
{
    auto itemStore = std::make_unique<RenderItemStore>();
    auto stages = std::vector<std::shared_ptr<RenderStage>>{};

    auto stageNames =
        std::istringstream(MPEGI::Config::getStringDef("RenderStages", MPEGI_DEFAULT_STAGES));

    for (std::string stageName; std::getline(stageNames, stageName, ',');) {
        auto newStage = MPEGI::RenderStageRegistry::createStage(stageName);
        newStage->initializeStage(bitstream, itemStore.get(), streamManager.get(), sceneState);
        stages.push_back(newStage);
    }
    return { sceneState, streamManager, std::move(itemStore), std::move(stages) };
}

Renderer::Renderer(const Bitstream& bitstream, const SceneState* state,
    std::shared_ptr<StreamManager> streamManager)
    : Renderer{ createStages(bitstream, state, streamManager) }
{}

Renderer::Renderer(CtorArgs&& args)
    : streamManager(std::move(args.streamManager))
    , itemStore(std::move(args.itemStore))
    , stages(std::move(args.stages))
    , sceneState(args.sceneState)
    , renderListCount(0)
    , signalsToChange{ false }
{
    // Initialize triple buffer
    this->stageDspData[0].resize(stages.size());
    this->stageDspData[1].resize(stages.size());
    this->stageDspData[2].resize(stages.size());

    // attaching at the end of constructor, so it does not happen when the initialization
    // causes an exception
    sceneState->attach(this);

    // register render-stage observer
    for (auto& stage : this->stages) {
        stage->setObserver(&renderStageObserver);
    }
}

Renderer::~Renderer()
{
    stopController();

    if (sceneState) {
        sceneState->detach(this);
    }
}

void Renderer::addSpatializer(
    std::shared_ptr<Spatializer> newSpatializer, const Bitstream& bitstream)
{
    EAR_CHECK_REL(!isRunning(),
        "Spatializers can only be added to the Renderer when the controller is not running.");

    newSpatializer->initializeSpatializer(
        bitstream, streamManager->getFs(), streamManager->getB(), sceneState);

    spatializers.push_back(newSpatializer);

    outputBuffers.push_back(std::make_unique<EAR::RealFrame>(
        newSpatializer->getNbrOfChannels(), streamManager->getB()));

    itemStore->activeSpatializers = spatializers.size();

    limiters.push_back(std::make_unique<EAR::LimiterIvas>(newSpatializer->getNbrOfChannels(),
        streamManager->getFs(), streamManager->getB(),
        Config::getFloatDef("Limiter_ivasAttackSeconds", 0.0025f),
        Config::getFloatDef("Limiter_ivasThreshold", 0.89f)));

    // Initialize triple buffer
    this->spatializerDspData[0].resize(spatializers.size());
    this->spatializerDspData[1].resize(spatializers.size());
    this->spatializerDspData[2].resize(spatializers.size());

    // TODO: implement removeSpatializer function
}

void Renderer::sceneStateDetached()
{
    listenerAttachment.reset(nullptr); // not listening to OSC anymore

    for (auto& pair : sceneObjectAttachments) {
        pair.second->remove();
    }

    sceneObjectAttachments.clear();
    sceneState = nullptr;

    stopController(); // we stop rendering at this point
}

void Renderer::sceneStateAttached(const SceneState* state)
{
    EAR_CHECK_REL(state == sceneState, "Renderer can only be attached to one SceneState.");
    EAR_CHECK_REL(!isRunning(),
        "Renderer was falsely attached to a scene state when it was already running.");
}

void Renderer::sceneObjectAdded(EAR::SceneObject* object)
{
    //! TODO: more elegant way to discriminate different SceneObject types (should be argument of observer callback)
    if (dynamic_cast<EAR::Listener*>(object) != nullptr) {
        listenerAttachment = std::make_unique<ListenerAttachment>(object, this);
    } else if (dynamic_cast<EAR::LoudspeakerLayout*>(object) != nullptr) {
        auto loudspeakers =
            dynamic_cast<EAR::LoudspeakerLayout*>(object)->getLoudspeakerPointSources();

        for (auto ls : loudspeakers) {
            sceneObjectAttachments.push_back(std::make_pair(ls.get(),
                std::make_unique<SceneObjectAttachment>(ls.get(), this, object->getName())));
        }
    } else if (dynamic_cast<EAR::PointSource*>(object) != nullptr) {
        sceneObjectAttachments.push_back(
            std::make_pair(object, std::make_unique<SceneObjectAttachment>(object, this)));

    } else if (dynamic_cast<EAR::HOASource*>(object) != nullptr) {
        sceneObjectAttachments.push_back(
            std::make_pair(object, std::make_unique<SceneObjectAttachment>(object, this)));
    }
    // ignore all other scene Objects (like geometry)
}

void Renderer::sceneObjectRemoved(EAR::SceneObject* object)
{
    // TODO: correctly remove loudspeaker sources

    auto it = std::find_if(sceneObjectAttachments.begin(), sceneObjectAttachments.end(),
        [object](auto& ref) { return ref.first == object; });

    if (it != sceneObjectAttachments.end()) {
        it->second->remove();
    }
}

void Renderer::controlStep()
{
    if (updateIdx == renderIdxAtomic.load(std::memory_order_acquire)) {
        // This should not happen very often, but once in a while it is OK:
        SPDLOG_INFO("Renderer::controlStep: update was too early, we have to wait ...");
        return;
    }

    // All items that are orphaned at this point are kept alive for one more round of triple buffering.
    itemStore->cleanup(orphanedItems[updateIdx]);

    // swap list instances and make clean copy
    listAccessMutex.lock();
    RenderList lineupList = createFreshList();
    listAccessMutex.unlock();

    // let stages contribute to lineup list
    for (std::size_t stageIdx = 0; stageIdx < stages.size(); ++stageIdx) {
#ifdef MPEGI_MEASURE_STAGE_TIMES
        auto start = Clock::now();
#endif
        auto& stage = stages[stageIdx];
        auto& data = stageDspData[updateIdx][stageIdx];
        // Old data is destroyed before update() is called:
        data.reset();
        data = stage->update(lineupList);
#ifdef MPEGI_MEASURE_STAGE_TIMES
        auto end = Clock::now();
        stage->consumedUpdateTime += end - start;
        stage->updateNumBlocks++;

        if (measureTimeFrameWise) {
            stage->consumedUpdateTimePerFrame.emplace_back(microseconds(end - start));
        }
#endif
    }

#ifdef MPEGI_MEASURE_STAGE_TIMES
    auto start = Clock::now();
#endif
    // new list is ready, so we let the spatializers know
    for (std::size_t spatializerIdx = 0; spatializerIdx < spatializers.size(); ++spatializerIdx) {
        auto& spat = spatializers[spatializerIdx];
        auto& data = spatializerDspData[updateIdx][spatializerIdx];
        // Old data is destroyed before update() is called:
        data.reset();
        data = spat->update(lineupList);
        if (data == nullptr) {
            EAR_FAIL("Returning nullptr from Spatializer::update() is not allowed");
        }
    }
#ifdef MPEGI_MEASURE_STAGE_TIMES
    auto end = Clock::now();
    spatializersConsumedUpdateTime += end - start;
    spatializersUpdateNumBlocks++;

    if (measureTimeFrameWise) {
        spatializersConsumedUpdateTimePerFrame.emplace_back(microseconds(end - start));
    }
#endif

    // signal to audio thread if it needs to change some audio streams (e.g. start to play)
    if (signalChanges.size() > 0) {
        signalsToChange.store(true);
    }

    updateIdx = (updateIdx + 1) % 3;
    // controlStep() is done, now processBlock() can take over:
    updateIdxAtomic.store(updateIdx, std::memory_order_release);
}

#ifdef MPEGI_MEASURE_STAGE_TIMES
void Renderer::printStageTimes(std::ostream& os)
{
    int stageCtr = 0;

    os << "Nr Stage Update ProcessBlock numUpdate numProcessBlock\n";

    auto stageNames = std::istringstream(MPEGI::Config::getStringDef("RenderStages", ""));
    std::string stageName;
    for (auto& stage : stages) {
        std::getline(stageNames, stageName, ',');

        os << stageCtr << " " << stageName << " ";
        os << microseconds(stage->consumedUpdateTime) << " ";
        os << microseconds(stage->consumedProcessTime) << " ";
        os << stage->updateNumBlocks << " " << stage->processNumBlocks << "\n";

        stageCtr++;
    }

    os << stageCtr << " Spatializers ";
    os << microseconds(spatializersConsumedUpdateTime) << " ";
    os << microseconds(spatializersConsumedProcessTime) << " ";
    os << spatializersUpdateNumBlocks << " ";
    os << spatializersProcessNumBlocks << "\n";
}

void Renderer::printStageTimesPerFrame(std::ostream& os)
{
    int stageCtr = 0;

    auto stageNames = std::istringstream(MPEGI::Config::getStringDef("RenderStages", ""));
    std::string stageName;
    for (auto& stage : stages) {
        std::getline(stageNames, stageName, ',');

        os << stageCtr << " 1 ";
        for (auto& stageTime : stage->consumedUpdateTimePerFrame) {
            os << ' ' << stageTime;
        }
        os << std::endl;
        os << stageCtr << " 2 ";
        for (auto& stageTime : stage->consumedProcessTimePerFrame) {
            os << ' ' << stageTime;
        }
        os << std::endl;

        stageCtr++;
    }

    os << stageCtr << " 1 ";
    for (auto stageTime : spatializersConsumedUpdateTimePerFrame) {
        os << ' ' << stageTime;
    }
    os << std::endl;

    os << stageCtr << " 2 ";
    for (auto stageTime : spatializersConsumedProcessTimePerFrame) {
        os << ' ' << stageTime;
    }
    os << std::endl;
}

void Renderer::printStageTimesAvg(std::ostream& os)
{
    auto sumAvgUpdate = Clock::duration::zero();
    auto sumAvgProcess = Clock::duration::zero();

    for (auto& stage : stages) {
        sumAvgUpdate += stage->consumedUpdateTime / stage->updateNumBlocks;
        if (stage->processNumBlocks != 0) {
            sumAvgProcess += stage->consumedProcessTime / stage->processNumBlocks;
        }
    }

    auto spatAvgUpdateTime = spatializersConsumedUpdateTime / spatializersUpdateNumBlocks;
    auto spatAvgProcessTime = spatializersConsumedProcessTime / spatializersProcessNumBlocks;
    sumAvgUpdate += spatAvgUpdateTime;
    sumAvgProcess += spatAvgProcessTime;

    os << "Stage avgUpdate(us) updateThread(%) avgProcessBlock(us) audioThread(%)\n";

    auto stageNames = std::istringstream(MPEGI::Config::getStringDef("RenderStages", ""));
    std::string stageName;
    for (auto& stage : stages) {
        std::getline(stageNames, stageName, ',');

        auto avgUpdateTime = stage->consumedUpdateTime / stage->updateNumBlocks;

        auto avgProcessTime = Clock::duration::zero();
        if (stage->processNumBlocks != 0) {
            avgProcessTime = stage->consumedProcessTime / stage->processNumBlocks;
        }

        os << stageName << " ";
        os << microseconds(avgUpdateTime) << " ";
        os << avgUpdateTime * 100 / sumAvgUpdate << "% ";
        os << microseconds(avgProcessTime) << " ";
        os << avgProcessTime * 100 / sumAvgProcess << "%\n";
    }

    os << "Spatializers ";
    os << microseconds(spatAvgUpdateTime) << " ";
    os << spatAvgUpdateTime * 100 / sumAvgUpdate << "% ";
    os << microseconds(spatAvgProcessTime) << " ";
    os << spatAvgProcessTime * 100 / sumAvgProcess << "%\n";

    os << "Total ";
    os << microseconds(sumAvgUpdate) << " 100% ";
    os << microseconds(sumAvgProcess) << " 100%\n";
}

void Renderer::initStageTimePerFrame(int blocks)
{
    measureTimeFrameWise = true;

    for (auto& stage : stages) {
        stage->consumedUpdateTimePerFrame.reserve(blocks);
        stage->consumedProcessTimePerFrame.reserve(blocks);
    }
    spatializersConsumedUpdateTimePerFrame.reserve(blocks);
    spatializersConsumedProcessTimePerFrame.reserve(blocks);
}
#endif

void Renderer::processBlock(float** output)
{
    // NB: The logic of the atomic updates requires that controlStep()
    //     has been completed at least once at this point!

    auto nextIdx = (renderIdx + 1) % 3;

    if (nextIdx == updateIdxAtomic.load(std::memory_order_acquire)) {
        // no new update, re-rendering old data
    } else {
        renderIdx = nextIdx;
        renderIdxAtomic.store(renderIdx, std::memory_order_release);

        // trigger DSP data swap in stages and spatializers if a new update step has finished

        // execute signal changes in sync with other DSP exchanges
        if (signalsToChange.load()) {
            std::lock_guard<std::mutex> guard(signalChangeMutex);
            for (auto& change : signalChanges) {
                change.first->setStream(change.second.id, change.second.seekTime);
                change.first->setLoop(change.second.looping);

                if (change.second.playing) {
                    change.first->play();
                } else {
                    change.first->stop();
                }
            }
            signalChanges.clear();
            signalsToChange.store(false);
        }
    }

    // get new samples from stream manager
    streamManager->nextBlock();

    for (std::size_t stageIdx = 0; stageIdx < stages.size(); ++stageIdx) {

        auto& data = stageDspData[renderIdx][stageIdx];
        if (data) {
#ifdef MPEGI_MEASURE_STAGE_TIMES
            auto start = Clock::now();
#endif
            auto& stage = stages[stageIdx];
            stage->processBlock(RenderStage::DspContext{ *data.get() });
#ifdef MPEGI_MEASURE_STAGE_TIMES
            auto end = Clock::now();
            stage->consumedProcessTime += end - start;
            stage->processNumBlocks++;

            if (measureTimeFrameWise) {
                stage->consumedProcessTimePerFrame.emplace_back(microseconds(end - start));
            }
#endif
        } else {
            // no data provided by update(), we don't call processBlock()
        }
    }

    int ch = 0;

#ifdef MPEGI_MEASURE_STAGE_TIMES
    auto start = Clock::now();
#endif
    for (int i = 0; i < spatializers.size(); i++) {
        auto& data = spatializerDspData[renderIdx][i];
        assert(data);
        auto& spatializer = spatializers[i];
        spatializer->requestBlock(Spatializer::DspContext{ *data.get(), *outputBuffers[i] });

        limiters[i]->process(outputBuffers[i]->data().get());

        // copy output to output pointers
        for (int k = 0; k < outputBuffers[i]->channels(); k++) {
            outputBuffers[i]->ch(k).read(output[ch], 0, streamManager->getB());
            ch++;
        }
    }
#ifdef MPEGI_MEASURE_STAGE_TIMES
    auto end = Clock::now();
    spatializersConsumedProcessTime += end - start;
    spatializersProcessNumBlocks++;

    if (measureTimeFrameWise) {
        spatializersConsumedProcessTimePerFrame.emplace_back(microseconds(end - start));
    }
#endif
}

RenderList Renderer::createFreshList()
{

    RenderList newList{ ++renderListCount };

    newList.listener = *listenerAttachment->state; // deep copy
    listenerAttachment->state->hasMoved = false; // state reset after copy
    listenerAttachment->state->coarseMoved = false;
    listenerAttachment->state->coarseRotated = false;
    listenerAttachment->state->teleport = false;

    for (auto it = sceneObjectAttachments.begin(); it != sceneObjectAttachments.end();) {
        if (it->second->itemHandle->status == ItemStatus::FadeOutInactive) {
            itemStore->deactivate(
                it->second->itemHandle); // try deactivation until all fade outs are completed
        } else if (it->second->itemHandle->status == ItemStatus::FadeOutOrphaned) {
            itemStore->remove(
                it->second->itemHandle); // try removal until all fade outs are completed
        }
        newList.addItemCopy(it->second->itemHandle);
        it->second->itemHandle->resetChanged();

        if (it->second->itemHandle->status == ItemStatus::Orphaned) {
            it = sceneObjectAttachments.erase(it);
        } else {
            it++;
        }
    }

    return newList;
}

Renderer::SceneObjectAttachment::SceneObjectAttachment(
    EAR::SceneObject* obj, Renderer* parent, const std::string& parentObjectId)
    : parent(parent)
    , object(obj)
{

    parent->listAccessMutex.lock();

    itemHandle = parent->itemStore->createItem(object->getName());

    objectId = parentObjectId == "" ? object->getName() : parentObjectId;

    if (parent->sceneState->hasAudioStream(object)) {
        const MPEGI::SceneState::AudioStream audioStream =
            parent->sceneState->getAudioStream(object);
        itemHandle->setSignal(parent->streamManager->createAccessBuffer(audioStream));
    } else if (parent->sceneState->hasGranularSound(object)) {
        const MPEGI::SceneState::GranularSound* granularSound =
            parent->sceneState->getGranularSound(object);
        // This object has no audio stream but has a granular sound connected to it, create a stream buffer where
        // the number of channels is borrowed from granularSound object connected to the objectSource
        itemHandle->setSignal(
            std::make_shared<StreamBuffer>(granularSound->channels, parent->streamManager->getB()));
    }
    itemHandle->setPosition(object->getGlobalPosition(), object->getLocalCSTransform());
    itemHandle->teleport = parent->sceneState->isTeleporting(objectId);

    auto psrc = dynamic_cast<EAR::PointSource*>(object);
    if (psrc) {
        itemHandle->setDirectivity(psrc->getDirectivity());
        itemHandle->setDirectiveness(psrc->getDirectiveness());
        itemHandle->setGain(psrc->getGainFactor());
        itemHandle->setLoudness(psrc->getLoudnessFactor());

        itemHandle->refDistance = psrc->getReferenceDistance();
        itemHandle->setAParams(parent->sceneState->getAParams(object));
        itemHandle->setVDLMethod(psrc->getVDLMethod());
        itemHandle->setReverbGain(psrc->getReverbGain());
        itemHandle->setNoReverb(psrc->getNoReverb());
        if (parent->sceneState->getGranularSound(object) != nullptr) {
            itemHandle->setGranularDatabaseId(parent->sceneState->getGranularSound(object)->id);
        }
        itemHandle->setGranularPos(psrc->getGranularPos());
        itemHandle->setGranularGroupProb(psrc->getGranularGroupProb());

        auto listenerVoice = dynamic_cast<EAR::ListenerVoiceSource*>(object);
        if (listenerVoice) {
            itemHandle->type = ItemType::ListenerVoice;
        } else {
            itemHandle->setExtent(psrc->getExtent(), parent->sceneState->getExtentLayout(object));
        }
    }

    auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
    if (hoasrc) {
        itemHandle->type = ItemType::HOA;
        itemHandle->setExtent(hoasrc->getExtent());
        itemHandle->setGain(hoasrc->getGainFactor());
        itemHandle->setLoudness(hoasrc->getLoudnessFactor());
        itemHandle->refDistance = hoasrc->getReferenceDistance();
        itemHandle->setAParams(parent->sceneState->getAParams(object));

        itemHandle->hoaInfo = std::make_shared<HoaInfo>();
        itemHandle->hoaInfo->group = hoasrc->getGroup();
        itemHandle->hoaInfo->external =
            (hoasrc->getRepresentation() == EAR::HOASource::Representation::Exterior);
        itemHandle->hoaInfo->extentTransform = hoasrc->doesExtentTransform();
        itemHandle->hoaInfo->transitionDist = hoasrc->getTransitionDistance();
        itemHandle->hoaInfo->source = hoasrc;
        itemHandle->setVDLMethod(hoasrc->getVDLMethod());
        itemHandle->setReverbGain(hoasrc->getReverbGain());
        itemHandle->setNoReverb(hoasrc->getNoReverb());
        if (parent->sceneState->getGranularSound(object) != nullptr) {
            itemHandle->setGranularDatabaseId(parent->sceneState->getGranularSound(object)->id);
        }
        itemHandle->setGranularPos(hoasrc->getGranularPos());
        itemHandle->setGranularGroupProb(hoasrc->getGranularGroupProb());
    }

    if (!object->isActive()) {
        itemHandle->status = ItemStatus::Inactive; // not fading in in the first place
    }

    parent->listAccessMutex.unlock();

    object->attach(this);
}

void Renderer::SceneObjectAttachment::remove()
{
    parent->itemStore->remove(itemHandle);
    object->detach(this);
}

void Renderer::SceneObjectAttachment::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    parent->listAccessMutex.lock();

    switch (modification) {
    case EAR::SceneObjectObserver::Property::Position: {
        double ts = object->getPositionTimestamp();
        itemHandle->teleport = parent->sceneState->isTeleporting(objectId);

        std::shared_ptr<Trajectory> traj =
            std::make_shared<ConstLocation>(object->getGlobalLocation(), ts);

        if (object->getGlobalVelocity().abs() > 0 && !itemHandle->teleport) {
            traj = std::make_shared<ConstVelocity>(
                object->getGlobalLocation(), object->getGlobalVelocity(), ts);
        }

        itemHandle->setPosition(object->getGlobalPosition(), object->getLocalCSTransform(), traj);

        break;
    }

    case EAR::SceneObjectObserver::Property::Activity: {
        if (object->isActive()) {
            // in any case, active items are set to active -> fade ins can be handled by spatializer alone
            itemHandle->setStatus(ItemStatus::Active);
        } else if (itemHandle->status == ItemStatus::Active
            || itemHandle->status == ItemStatus::New) {
            // item needs to be faded out (the completeness of this process is handled inside the render item)
            parent->itemStore->deactivate(itemHandle); // setStatus(ItemStatus::FadeOutInactive);
        }
        // no need to act otherwise

        break;
    }

    case EAR::SceneObjectObserver::Property::Directivity: {
        auto psrc = dynamic_cast<EAR::PointSource*>(object);

        if (psrc) {
            itemHandle->setDirectivity(psrc->getDirectivity());
            itemHandle->setDirectiveness(psrc->getDirectiveness());
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::Extent: {
        // TODO: Solve with extent decorator on SceneObject
        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setExtent(psrc->getExtent(), parent->sceneState->getExtentLayout(object));
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setExtent(hoasrc->getExtent());
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::Gain: {
        // TODO: Solve with extent decorator on SceneObject
        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setGain(psrc->getGainFactor());
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setGain(hoasrc->getGainFactor());
        }

        break;
    }
    case EAR::SceneObjectObserver::Property::Loudness: {
        // TODO: Solve with extent decorator on SceneObject
        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setLoudness(psrc->getLoudnessFactor());
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setLoudness(hoasrc->getLoudnessFactor());
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::ReverbGain: {
        // TODO: Solve with extent decorator on SceneObject
        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setReverbGain(psrc->getReverbGain());
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setReverbGain(hoasrc->getReverbGain());
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::NoReverb: {
        // TODO: Solve with extent decorator on SceneObject
        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setNoReverb(psrc->getNoReverb());
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setNoReverb(hoasrc->getNoReverb());
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::AudioStream: {
        auto signalptr = dynamic_cast<StreamAccessBuffer*>(itemHandle->signal.get());

        if (signalptr) {
            // this shall only take effect in synced with other updates
            // for that reason, we put this in a queue to be executed together with the render stage dsp changes
            std::lock_guard<std::mutex> guard(parent->signalChangeMutex);
            parent->signalChanges.push_back(
                std::make_pair(signalptr, parent->sceneState->getAudioStream(object)));
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::GranularPos: {

        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setGranularPos(psrc->getGranularPos());
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setGranularPos(hoasrc->getGranularPos());
        }

        break;
    }

    case EAR::SceneObjectObserver::Property::GranularGroupProb: {

        auto psrc = dynamic_cast<EAR::PointSource*>(object);
        if (psrc) {
            itemHandle->setGranularGroupProb(psrc->getGranularGroupProb());
        }

        auto hoasrc = dynamic_cast<EAR::HOASource*>(object);
        if (hoasrc) {
            itemHandle->setGranularGroupProb(hoasrc->getGranularGroupProb());
        }

        break;
    }

    default:
        break;
    }

    parent->listAccessMutex.unlock();
}

Renderer::ListenerAttachment::ListenerAttachment(EAR::SceneObject* obj, Renderer* parent)
    : parent(parent)
    , object(obj)
    , state(std::make_unique<RenderList::ListenerState>())
{
    object->attach(this);

    state->teleport = parent->sceneState->isTeleporting("listener");

    if (object->getGlobalVelocity().abs() > 0 && !state->teleport) {
        state->trajectory = std::make_shared<ConstVelocity>(
            object->getLocation(), object->getGlobalVelocity(), object->getPositionTimestamp());
    } else {
        state->trajectory =
            std::make_shared<ConstLocation>(object->getLocation(), object->getPositionTimestamp());
    }

    state->cs = object->getLocalCSTransform();
    state->position = object->getPosition();
    if (auto* listener = dynamic_cast<Listener*>(obj)) {
        state->realPosition = listener->realPosition;
    } else {
        EAR_FAIL("expected a listener");
    }

    state->hasMoved = true;
}

Renderer::ListenerAttachment::~ListenerAttachment()
{
    object->detach(this);
}

void Renderer::ListenerAttachment::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    if (modification == EAR::SceneObjectObserver::Property::Position) {
        parent->listAccessMutex.lock();

        state->teleport = parent->sceneState->isTeleporting("listener");

        if (object->getGlobalVelocity().abs() > 0 && !state->teleport) {
            state->trajectory = std::make_shared<ConstVelocity>(
                object->getLocation(), object->getGlobalVelocity(), object->getPositionTimestamp());
        } else {
            state->trajectory = std::make_shared<ConstLocation>(
                object->getLocation(), object->getPositionTimestamp());
        }

        state->cs = object->getLocalCSTransform();
        state->position = object->getPosition();
        if (auto* listener = dynamic_cast<Listener*>(obj)) {
            state->realPosition = listener->realPosition;
        } else {
            EAR_FAIL("expected a listener");
        }

        state->hasMoved = true;
        if (obj->hasCoarseMoved()) {
            // only set coarseMoved to true, it will be reset to false when the new renderList is created
            // Ensures that multiple listener position updates within a single frame don't set flag false before stages have been updated.
            state->coarseMoved = true;
        }
        if (obj->hasCoarseRotated()) {
            // only set coarseRotated to true, it will be reset to false when the new renderList is created
            state->coarseRotated = true;
        }

        parent->listAccessMutex.unlock();
    }
}

AudioUtilization Renderer::getAudioUtilization()
{
    AudioUtilization audioUtilization;
    for (const auto& source : sceneState->getObjectSources()) {
        if (!source->isActive()) {
            audioUtilization.notActiveSources.push_back(
                sceneState->getAudioStream(source.get()).id);
        }
    }

    for (const auto& source : sceneState->getLoudspeakerLayouts()) {
        if (!source->isActive()) {
            audioUtilization.notActiveSources.push_back(
                sceneState->getAudioStream(source.get()).id);
        }
    }

    for (const auto& source : sceneState->getHOASources()) {
        if (source->getGroup() == -1) {
            if (!(source->isActive())) {
                audioUtilization.notActiveSources.push_back(
                    sceneState->getAudioStream(source.get()).id);
            }
        }
    }
    audioUtilization.activeMPHOASources = renderStageObserver.getUtilization();
    return audioUtilization;
}

} // namespace MPEGI
