/***************************************************************************
This software module was originally developed jointly by

Nokia Technologies Ltd., Huawei, and Qualcomm Technologies, Inc. 

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd., Huawei, and Qualcomm Technologies, Inc. retain 
full right to modify and use the code for their own purpose, assign or donate 
the code to a third party without consent of and without accounting to 
third parties, and to inhibit third parties from using the code for 
products that do not conform to MPEG-related ISO/IEC International 
Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Renderer/MphoaBaseStage.h>

#include <algorithm>
#include <utility>

#include <EAR/AnchorObject.h>
#include <EAR/Exception.h>
#include <EAR/HOASource.h>
#include <EAR/Listener.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/MphoaPayload.hpp>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Stream/StreamBuffer.h>

#include <hoatranslation/MultiPointHoaTranslator.hpp>
#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/HoaTranslator.hpp>
#include <mpegcore/HoaTranslatorConfig.hpp>
#include <mpegcore/algorithm.hpp>
#include <mpegcore/spdlog.hpp>

using namespace mpegcore;

namespace MPEGI {

MphoaBaseStage::MphoaBaseStage()
    : config{ mpegcore::Config{} }
{
    mphoaGain = Config::getFloatDef("Mphoa:gain", EAR::Variant(1.F));
    config.hrirSofaFile = Config::getStringDef("HrirSofaFile", "");
}

MphoaBaseStage::~MphoaBaseStage() {}

#define PROPERTY_CHANGE_NOT_ALLOWED(x)                                                             \
    if (item->changed & (x)) {                                                                     \
        EAR_CHECK_REL(false, "MphoaStage: unsupported property change " #x);                       \
    }

#define PROPERTY_CHANGE_ALLOWED_ONCE(x)                                                            \
    if (item->changed & (x)) {                                                                     \
        if (itemPropertyChanged[x][hoaSourceIndex]) {                                              \
            EAR_CHECK_REL(false, "MphoaStage: multiple changes not supported for property " #x);   \
        } else {                                                                                   \
            itemPropertyChanged[x][hoaSourceIndex] = true;                                         \
        }                                                                                          \
    }

std::unique_ptr<RenderStage::DspData> MphoaBaseStage::update(RenderList& renderList)
{
    if (!isInitialized) {
        return {};
    }

    listenerPosition = renderList.listener.position;
    auto dspData = std::make_unique<MphoaBaseStage::DspData>();
    auto& dspList = *dspData;
    dspList.activeTranslatorIndices = determineActiveHoaTranslators();

    // Create a DSPItem for each group regardless of whether its associated translator is active or
    // not.
    auto const translatorCount = translatorData.size();
    dspList.dspItems.resize(translatorCount);

    auto indexOfHoaSource = [](auto const& sources, auto const* source) {
        auto it = std::find_if(sources.cbegin(), sources.cend(),
            [&source](auto const& s) { return s.get() == source; });
        EAR_CHECK_REL(it != sources.cend(), "MphoaStage: HOASource not found");
        return it - sources.cbegin();
    };

    auto const& hoaSources = sceneState->getHOASources();

    for (auto const& item : renderList.items) {
        if (!isMphoaRenderItem(*item)) {
            continue;
        }
        auto const hoaSourceIndex = indexOfHoaSource(hoaSources, item->hoaInfo->source);

        // Report an error if we get a property change we don't support.
        PROPERTY_CHANGE_NOT_ALLOWED(ItemProperty::Directivity);
        PROPERTY_CHANGE_NOT_ALLOWED(ItemProperty::Eq);
        PROPERTY_CHANGE_NOT_ALLOWED(ItemProperty::ParentRI);
        PROPERTY_CHANGE_NOT_ALLOWED(ItemProperty::RefDist);
        PROPERTY_CHANGE_NOT_ALLOWED(ItemProperty::ReflectionInfo);
        PROPERTY_CHANGE_NOT_ALLOWED(ItemProperty::Reverb);

        // Allow these properties to be "changed" once as they are sent initially when the Renderer
        // is starting up. We expect the changed value to be the same as the property had when the
        // stage was initialized and so we don't do anything based on the change. Report an error
        // for subsequent changes.
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Aparams);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Extent);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Gain);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Location);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Orientation);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Position);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Signal);
        PROPERTY_CHANGE_ALLOWED_ONCE(ItemProperty::Status);

        auto const group = item->hoaInfo->group;
        EAR_CHECK_REL(group < translatorCount, "MphoaStage: too high group group id");
        auto& dspItem = dspList.dspItems[group];

        // Verify that this item's `HOASource` is at the same index in its group as the signal frame
        // is being created in the DSP item. If this would fail then it could be that the render
        // items in the render list are created in a different order than the HOA sources were
        // placed into their groups and then the stage would need to keep another mapping between
        // them.
        auto const signalFrameIndex = dspItem.inputBuffers.size();
        auto const& groupSources = translatorData[group].group->sources;
        EAR_CHECK_REL(signalFrameIndex < groupSources.size(), "MphoaStage: too many signal frames");
        EAR_CHECK_REL(groupSources[signalFrameIndex] == item->hoaInfo->source,
            "MphoaStage: RenderList has sources in different order than they were initialized");

        dspItem.inputBuffers.emplace_back(item->signal);
    }

    for (auto* ri : outputRenderItems) {
        renderList.addItemCopy(ri);
        ri->resetChanged();
    }

    return dspData;
}

void MphoaBaseStage::verifyGroups(std::shared_ptr<std::vector<std::uint8_t>> const& spatialMetadata,
    std::vector<std::unique_ptr<MPEGI::HOAGroup>> const& hoaGroups)
{
    if (!spatialMetadata) {
        return;
    }

    decoderData d;
    auto stream = hoatranslation::ByteStream(spatialMetadata);
    read_spatial_metadata_info(stream, &d);

    auto const groupCount = d.num_hoa_groups;
    EAR_CHECK_REL(
        groupCount == hoaGroups.size(), "MphoaStage: Wrong number of spatial metadata groups");

    for (std::size_t groupIndex{ 0 }; groupIndex < groupCount; ++groupIndex) {
        auto const& smSources = d.hoaSourceIdStrs[groupIndex];
        auto const& hoaSources = hoaGroups[groupIndex]->sources;
        auto const sourceCount = smSources.size();
        EAR_CHECK_REL(sourceCount == hoaSources.size(),
            "MphoaStage: Wrong number of sources in a spatial metadata group");

        for (std::size_t sourceIndex{ 0 }; sourceIndex < sourceCount; ++sourceIndex) {
            EAR_CHECK_REL(smSources[sourceIndex] == hoaSources[sourceIndex]->getName(),
                "MphoaStage: Wrong (order of) sources in a spatial metadata group");
        }
    }
}

bool MphoaBaseStage::isMphoaSource(HOASource const& source)
{
    return source.getGroup() != -1
        && source.getRepresentation() == HOASource::Representation::Interior;
}

bool MphoaBaseStage::isMphoaRenderItem(RenderItem const& item)
{
    if (item.type != ItemType::HOA) {
        return false;
    }

    EAR_CHECK_REL(item.hoaInfo->source, "MphoaStage: HOA RenderItem is missing HOASource");
    return isMphoaSource(*item.hoaInfo->source);
}

std::vector<mpegcore::HoaTranslatorConfig> MphoaBaseStage::createHoaTranslatorConfigs(
    std::vector<std::unique_ptr<MPEGI::HOAGroup>> const& hoaGroups, mpegcore::Config const& config,
    bool haveSpatialMetadata)
{
    auto const groupCount = hoaGroups.size();
    std::vector<HoaTranslatorConfig> configs(groupCount);

    for (std::size_t i{ 0 }; i < groupCount; ++i) {
        configs[i].groupId = i;
        configs[i].sofaFile = config.hrirSofaFile;
        configs[i].useEncodedValues = haveSpatialMetadata;
        configs[i].exteriorProjectionRadius = hoaGroups[i]->exteriorProjectionRadius;
        configs[i].adaptiveExteriorProjectionRadius =
            hoaGroups[i]->adaptiveExteriorProjectionRadius;
    }

    return configs;
}

void MphoaBaseStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }

    auto const& hoaSources = state->getHOASources();
    if (std::none_of(hoaSources.cbegin(), hoaSources.cend(),
            [](auto const& s) { return isMphoaSource(*s); })) {
        return;
    }

    auto mphoaPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_MPHOA);
    EAR_CHECK_REL(mphoaPayloads.size() <= 1, "MphoaStage: too many mphoa payloads in bitstream");

    if (mphoaPayloads.size() == 1) {
        auto const mphoaPayload =
            std::dynamic_pointer_cast<MPEGI::MphoaPayload const>(mphoaPayloads[0]);
        spatialMetadata = mphoaPayload->getSpatialMetadata();
    }

    auto const& hoaGroups = state->getHOAGroups();
    if (hoaGroups.empty()) {
        return;
    }
    verifyGroups(spatialMetadata, hoaGroups);

    sceneState = state;
    B = streamManager->getB();

    // Buffer sizes etc get initiated based on config, so make sure we are running at same sized
    // blocks.
    EAR_CHECK_REL(B == config.processingSamplesPerBuffer,
        "MphoaStage: Mismatch of B and processingSamplesPerBuffer");

    auto hoaTranslatorConfigs = createHoaTranslatorConfigs(hoaGroups, config, !!spatialMetadata);
    auto const translatorCount = hoaTranslatorConfigs.size();
    translatorData = std::vector<TranslatorData>(translatorCount);
    std::vector<size_t> groupSourceCounts;

    // Create translator data for all translators
    for (std::size_t translatorIndex{ 0 }; translatorIndex < translatorCount; ++translatorIndex) {
        auto const& translatorConfig = hoaTranslatorConfigs[translatorIndex];

        // We created the group ids in `createHoaTranslatorConfigs` and so they should be fine but
        // this may change in the future if for example the group ids come from the bitstream.  The
        // configs are given to hoa-translation so they have to match what we here understand as the
        // group id. This check makes the current mapping explicit so that when we create an item to
        // `translatorData[translatorIndex]` that matches the config we give the corresponding
        // `HoaTranslator`. What is somewhat unfortunate in this is that even though this is called
        // `groupId`, it is not really an id but rather an index into `translatorData`.
        EAR_CHECK_REL(translatorConfig.groupId == translatorIndex,
            fmt::format("MphoaStage: hoaTranslatorConfig.groupId ({}) doesn't match the index of "
                        "the hoaTranslatorConfig itself ({})",
                translatorConfig.groupId, translatorIndex));

        // Group containing data on sources this HoaTranslator handles.
        // Indices in `HoaTranslator::Group` are used in the following way:
        // - sourceIndices - indices of the sources within the group [0..source count)
        // - regionIndex   - index into `translatorData`, through which the actual region is found
        HoaTranslator::Group translatorGroup;

        auto& sceneGroup = hoaGroups[translatorConfig.groupId];
        EAR_CHECK_REL(
            sceneGroup->sources.size() > 0, "MphoaStage: need at least one source in each group");

        // Check for auto-rescaling
        // - autoRescale - Enable the renderer to modify, based on the LSDF playback dimension
        //      and the HOA Group source dimension, a new location of the HOA Source.
        EAR::Vector3 rescale = { 1.0, 1.0, 1.0 };
        EAR::Vector3 rescaleOffset = { 0.0, 0.0, 0.0 };

        // Get parent of HOA Group to find the source space via its region
        auto hoaSourceParent = sceneGroup->sources[0]->getParent();

        if (hoaSourceParent != nullptr) {
            auto hoaParentName = hoaSourceParent->getName();

            auto const& anchors = state->getAnchors();

            for (auto const& thisAnchor : anchors) {
                if (thisAnchor->getName() == hoaParentName && thisAnchor->isAutoRescale()) {
                    auto& thisRegion = sceneGroup->region;
                    auto geos = sceneState->getAllGeometry();
                    // Issue: how to differentiate between LSDF mesh and scene meshes?
                    //      workaround is to assume only two regions per scene
                    for (auto thisGeo : geos) {
                        // Find LSDF listener space as a bounding box
                        if (thisRegion->getName() != thisGeo->getName()) {
                            //Update tolerance
                            auto const& tolerance = thisAnchor->getTolerance();
                            auto thisParent = thisGeo->getParent();
                            auto listenerBox = thisGeo->getBoundingBox();
                            auto thisRegionBox = thisRegion->getBoundingBox();
                            //Get offset in case center of listener position is not 0,0,0
                            auto offset = thisAnchor->getPosition(); // thisGeo->getPosition();
                            for (std::size_t i{ 0 }; i < thisRegionBox.size(); i++) {
                                rescale[i] = (listenerBox[i].high - listenerBox[i].low)
                                    / ((thisRegionBox[i].high * (1 - tolerance[i * 2]))
                                        - (thisRegionBox[i].low * (1 - tolerance[i * 2 + 1])));
                                if (rescale[i] > 1.0) {
                                    rescale[i] = 1.0;
                                }
                                rescaleOffset[i] = offset.location[i];
                            }
                        }
                    }
                }
            }
        }

        // Create source indices within the group. HoaTranslator uses these indices to indicate
        // which mic arrays are active, i.e. which sources in the group are going to be rendered.
        auto const groupSourceCount = sceneGroup->sources.size();
        groupSourceCounts.emplace_back(groupSourceCount);
        translatorGroup.sourceIndices.resize(groupSourceCount);
        std::iota(translatorGroup.sourceIndices.begin(),
            translatorGroup.sourceIndices.begin() + groupSourceCount, 0);

        transformToBack(sceneGroup->sources, translatorGroup.sourcePositions,
            [&](auto const* source) { return source->getGlobalPosition(); });
        transformToBack(sceneGroup->sources, translatorGroup.sourceLinearGains,
            [](auto const* source) { return source->getGainFactor(); });

        translatorGroup.complexityReductionParameters = sceneGroup->complexityReductionParameters;

        std::vector<HoaTranslator::InformedSourceConfig> informedSourceConfigs{};
        for (auto const src : sceneGroup->informedSources) {
            HoaTranslator::InformedSourceConfig ISconfig;
            ISconfig.sourceAwareObjectPosition = src.position;
            ISconfig.listenerThresholdPresent = src.listenerThresholdPresent;
            ISconfig.hoaSourceThresholdPresent = src.hoaSourceThresholdPresent;
            ISconfig.listenerThreshold = src.listenerThreshold;
            ISconfig.hoaSourceThreshold = src.hoaSourceThreshold;
            ISconfig.priorityValue = src.priorityValue;
            informedSourceConfigs.push_back(ISconfig);
            hoaTranslatorConfigs[translatorIndex].useSourceAwareMode = true;
        }
        hoaTranslatorConfigs[translatorIndex].maxSimulInformedSources =
            sceneGroup->maxSimulInformedSources;

        if (sceneGroup->region != nullptr) {
            // SceneState doesn't contain geometries in a single indexable vector so we'll use an
            // index to translatorData instead
            translatorGroup.regionIndex = translatorIndex;
        }

        // HoaTranslator expects all sources in the group to have the same number of channels so we
        // use the first source's channel count and don't have per source channel counts.
        auto const& streamId = state->getAudioStream(sceneGroup->sources[0]).id;
        translatorGroup.channelsPerSource = streamManager->getNbrOfChannels(streamId);

        for (int i = 0; i < sceneGroup->sources.size(); i++) {
            translatorData[translatorIndex].hoaSourceIds.push_back(
                state->getAudioStream(sceneGroup->sources[i]).id);
        }

        HoaParametersConfig hoaParametersConfig;
        hoaParametersConfig.isInputESD = true;

        translatorData[translatorIndex].translator.initialize(config,
            hoaTranslatorConfigs[translatorIndex], hoaParametersConfig, translatorGroup,
            spatialMetadata, informedSourceConfigs, rescale, rescaleOffset);
        EAR_CHECK_REL(translatorData[translatorIndex].translator.isInitialized(),
            "MphoaStage: Failed to initialize HOA translator");
        translatorData[translatorIndex].group = hoaGroups[translatorIndex].get();
    }

    initializeStageAdditional(
        bitstream, itemStore, streamManager, state, translatorCount, groupSourceCounts);

    // Create RenderItems for MphoaStage output
    outputRenderItems.push_back(itemStore->createItem("mphoa"));
    outputRenderItems.back()->type = ItemType::Downmix;
    constexpr std::size_t OUTPUT_CHANNEL_COUNT{ 2 };
    outputRenderItems.back()->setSignal(streamManager->createEmptyBuffer(OUTPUT_CHANNEL_COUNT));

    isInitialized = true;
}

std::vector<std::size_t> MphoaBaseStage::determineActiveHoaTranslators()
{
    std::vector<std::size_t> activeHoaTranslators;

    // Setting all HoaTranslators active, checking for listener in HoaGroup region
    // should be added at some point
    for (std::size_t i{ 0 }, end{ translatorData.size() }; i < end; ++i) {
        activeHoaTranslators.push_back(i);
    }

    return activeHoaTranslators;
}

} // namespace MPEGI
