/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Scene.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Scene/OSCListener.h>

#include <EAR/BaseTypes.h>
#include <EAR/Interpolator.h>
#include <EAR/LinearInterpolator.h>

#include <EAR/AirflowGeneratorBase.h>
#include <EAR/HOASource.h>
#include <EAR/Listener.h>
#include <EAR/LoudspeakerLayout.h>
#include <EAR/PointSource.h>
#include <EAR/SceneObject.h>

#include <algorithm>

namespace MPEGI {

namespace detail {
typedef Modification::TargetAttribute Attr;
// clang-format off
    static const std::map<Attr, std::string> attr2string = {
        {Attr::Position, "location"},
        {Attr::Orientation, "orientation"},
        {Attr::GainDb, "gainDb"},
        {Attr::CSpace, "cSpace"},
        {Attr::Active, "active"},
        {Attr::Signal, "signal"},
        {Attr::Extent, "extent"},
        {Attr::Directivity, "directivity"},
        {Attr::Directiveness, "directiveness"},
        {Attr::Play, "play"},
        {Attr::Group, "group"},
        {Attr::Region, "region"},
        {Attr::Size, "size"},
        {Attr::Speed, "speed"},
        {Attr::GranularPos, "granularPos"},
        {Attr::GranularGroupProb, "granularGroupProb"},
        {Attr::Material, "material"},
        {Attr::ReverbGain, "itemReverbGain"},
        {Attr::NoReverb, "noReverb"},
        {Attr::LoudnessDb, "loudnessDb"}
    };
// clang-format on
} // namespace detail

Scene::Scene(const Bitstream& bitstream, std::shared_ptr<Clock> sceneClock, bool offlineMode,
    bool enableOscListener, const LSDF::LSDFParser* lsdf)
    : pt(0)
    , state(bitstream, this, lsdf)
    , offlineMode(offlineMode)
    , sceneClock(sceneClock)
{

    if (enableOscListener) {
        oscListener = std::make_unique<OSCListener>(this, offlineMode);
    }

    state.attach(this);

    for (auto& upd : bitstream.getUpdates()) {
        addUpdate(upd);
    }

    // TODO: this static/non-static decision will move to the encoder
    // (and hopefully be a bit smarter)
    for (auto& pair : state.updateable) {
        if (pair.first == "listener") {
            continue;
        }
        pair.second->setStatic(true);
    }

    for (auto& pair : state.updateable) {
        if (idsWithUpdates.find(pair.first) != idsWithUpdates.end()) {
            pair.second->setStatic(false);
        }
    }

    auto eifPayload = std::dynamic_pointer_cast<const EIFPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));
    sceneDuration = eifPayload->getSceneInfo().duration;
}

Scene::~Scene()
{
    oscListener.reset(
        nullptr); // deleting the osc listener first to avoid rogue messages coming in late
    stopController();
    state.detach(this);
}

Scene::PermanentInterpolator::PermanentInterpolator(EAR::SceneObject* obj)
{
    location = std::make_unique<LocationInterpolator>();
    orientation = std::make_unique<OrientationInterpolator>();
    location->addSupportingPoint(0.0, obj->getLocation(), false);
    orientation->addSupportingPoint(0.0, EAR::Quaternion(obj->getOrientation()), false);

    if (auto ptr = dynamic_cast<EAR::PointSource*>(obj); ptr) {
        gain = std::make_unique<FloatInterpolator>();
        directiveness = std::make_unique<FloatInterpolator>();

        itemReverbGain = std::make_unique<FloatInterpolator>();
        granularPos = std::make_unique<LocationInterpolator>();
        granularGroupProb = std::make_unique<LocationInterpolator>();
        gain->addSupportingPoint(0.0, ptr->getGainDb(), false);
        directiveness->addSupportingPoint(0.0, ptr->getDirectiveness(), false);
        itemReverbGain->addSupportingPoint(0.0, ptr->getReverbGain(), false);

        loudness = std::make_unique<FloatInterpolator>();
        gain->addSupportingPoint(0.0, ptr->getGainDb(), false);
        directiveness->addSupportingPoint(0.0, ptr->getDirectiveness(), false);
        loudness->addSupportingPoint(0.0, ptr->getLoudnessDb(), false);
        granularPos->addSupportingPoint(0.0, ptr->getGranularPos(), false);
        granularGroupProb->addSupportingPoint(0.0, ptr->getGranularGroupProb(), false);
    }
    if (auto ptr = dynamic_cast<EAR::LoudspeakerLayout*>(obj); ptr) {
        itemReverbGain = std::make_unique<FloatInterpolator>();
        itemReverbGain->addSupportingPoint(0.0, ptr->getReverbGain(), false);
    }
    if (auto ptr = dynamic_cast<EAR::HOASource*>(obj); ptr) {
        gain = std::make_unique<FloatInterpolator>();
        itemReverbGain = std::make_unique<FloatInterpolator>();
        granularPos = std::make_unique<LocationInterpolator>();
        granularGroupProb = std::make_unique<LocationInterpolator>();
        gain->addSupportingPoint(0.0, ptr->getGainDb(), false);
        itemReverbGain->addSupportingPoint(0.0, ptr->getReverbGain(), false);
        granularPos->addSupportingPoint(0.0, ptr->getGranularPos(), false);
        granularGroupProb->addSupportingPoint(0.0, ptr->getGranularGroupProb(), false);
    }
    if (auto ptr = dynamic_cast<EAR::AirflowGeneratorBase*>(obj); ptr) {
        gain = std::make_unique<FloatInterpolator>();
        gain->addSupportingPoint(0.0, ptr->getGainDb(), false);
    }
}

Scene::ActiveTempInterpolator::ActiveTempInterpolator(
    double st, const TempInterpolator* p, const EAR::Variant& initialVal)
    : parent(p)
{

    endTime = st + parent->duration;

    if (parent->attribute == "location") {
        location = std::make_unique<LocationInterpolator>();
        location->addSupportingPoint(st, initialVal, false);
        location->addSupportingPoint(st + parent->duration, parent->targetVal, true);
    } else if (parent->attribute == "orientation") {
        orientation = std::make_unique<OrientationInterpolator>();
        orientation->addSupportingPoint(
            st, EAR::Quaternion(static_cast<EAR::Euler>(initialVal)), false);
        orientation->addSupportingPoint(st + parent->duration,
            EAR::Quaternion(static_cast<EAR::Euler>(parent->targetVal)), true);
    } else if (parent->attribute == "gainDb") {
        gain = std::make_unique<FloatInterpolator>();
        gain->addSupportingPoint(st, initialVal, false);
        gain->addSupportingPoint(st + parent->duration, parent->targetVal, true);
    } else if (parent->attribute == "itemReverbGain") {
        itemReverbGain = std::make_unique<FloatInterpolator>();
        itemReverbGain->addSupportingPoint(st, initialVal, false);
        itemReverbGain->addSupportingPoint(st + parent->duration, parent->targetVal, true);
    } else if (parent->attribute == "directiveness") {
        directiveness = std::make_unique<FloatInterpolator>();
        directiveness->addSupportingPoint(st, initialVal, false);
        directiveness->addSupportingPoint(st + parent->duration, parent->targetVal, true);
    } else if (parent->attribute == "loudnessDb") {
        loudness = std::make_unique<FloatInterpolator>();
        loudness->addSupportingPoint(st, initialVal, false);
        loudness->addSupportingPoint(st + parent->duration, parent->targetVal, true);
    } else if (parent->attribute == "granularPos") {
        granularPos = std::make_unique<LocationInterpolator>();
        granularPos->addSupportingPoint(st, initialVal, false);
        granularPos->addSupportingPoint(st + parent->duration, parent->targetVal, true);
    } else {
        EAR_FAIL("Unknown attribute for temporary interpolation.");
    }
}

void Scene::setTime(double t)
{
    double nt = sceneClock->getCurrentTime();
    sceneClock->sync(t);

    if (std::abs(nt - t) > 0.1) { // arbitrary minimum diff when we reset the scene interpolators
        // TODO: Thread safety
        if (nt > t) { // we're jumping backwards in time
            timedUpdates.resetPlayback();
        }
        activeTempInterpolators.clear();
    }

    if (offlineMode) {
        triggerControlStep();
    }
}

double Scene::getCurrentTime() const
{
    return sceneClock->getCurrentTime();
}

void Scene::controlStep()
{

    double nt = sceneClock->getCurrentTime();

    if (nt != pt) {

        if (offlineMode && oscListener) {
            double dt = nt - pt;
            if (dt < 0) { // scene time has looped
                dt += sceneDuration;
            }
            oscListener->triggerPlayback(dt);
        }

        // for objects that stopped moving in the last frame, we now set the moved flag back to false.
        for (std::string id : previousMovingObjects) {
            state.updateable[id]->resetMoved();
        }
        previousMovingObjects.clear();

        checkProximityTriggers();

        // active temporary interpolators
        Update tempUpdate(nt);

        for (auto it = activeTempInterpolators.begin(); it != activeTempInterpolators.end();) {

            if (it->location) {
                if (it->location->isChanging(nt, pt)) {
                    tempUpdate.addModification(it->parent->entity, "location",
                        it->location->interpolate(nt), it->location->isChanging(pt));
                    tempUpdate.addModification(
                        it->parent->entity, "velocity", it->location->getSlope(nt), false);
                } else if (it->location->isChanging(pt)) { // reset velocity if motion stopped
                    tempUpdate.addModification(
                        it->parent->entity, "velocity", EAR::Vector3(), false);
                }
            } else if (it->orientation && it->orientation->isChanging(nt, pt)) {
                tempUpdate.addModification(it->parent->entity, "orientation",
                    it->orientation->interpolate(nt).getEuler(), it->orientation->isChanging(pt));
            } else if (it->gain && it->gain->isChanging(nt, pt)) {
                tempUpdate.addModification(it->parent->entity, "gainDb", it->gain->interpolate(nt),
                    it->gain->isChanging(pt));
            } else if (it->itemReverbGain && it->itemReverbGain->isChanging(nt, pt)) {
                tempUpdate.addModification(it->parent->entity, "itemReverbGain",
                    it->itemReverbGain->interpolate(nt), it->itemReverbGain->isChanging(pt));
            } else if (it->directiveness && it->directiveness->isChanging(nt, pt)) {
                tempUpdate.addModification(it->parent->entity, "directiveness",
                    it->directiveness->interpolate(nt), it->directiveness->isChanging(pt));
            } else if (it->loudness && it->loudness->isChanging(nt, pt)) {
                tempUpdate.addModification(it->parent->entity, "loudnessDb",
                    it->loudness->interpolate(nt), it->loudness->isChanging(pt));
            } else if (it->granularPos && it->granularPos->isChanging(nt, pt)) {
                tempUpdate.addModification(it->parent->entity, "granularPos",
                    it->granularPos->interpolate(nt), it->granularPos->isChanging(pt));
            }

            // check if temp interpolator reached EOL
            if (nt > it->endTime) {
                if (it->location || it->orientation) {
                    // keep track of the objects which stopped moving in this frame so we can reset the moved flag.
                    // can't reset it here, as the modification still exists in tempUpdate which gets called to reset velocities.
                    previousMovingObjects.push_back(it->parent->entity);
                }
                it = activeTempInterpolators.erase(it);
            } else {
                ++it; // just continue with next
            }
        }

        if (tempUpdate.modifications.size() > 0) {
            state.modify(tempUpdate);
        }

        // timed updates
        Update timedUpdate(nt);

        for (auto& pair : permanentInterpolators) {
            if (pair.second->location->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "location",
                    pair.second->location->interpolate(nt), pair.second->location->isChanging(pt));
                timedUpdate.addModification(
                    pair.first, "velocity", pair.second->location->getSlope(nt), false);
            } else if (pair.second->location->isChanging(pt)) { // reset velocity if motion stopped
                timedUpdate.addModification(pair.first, "velocity", EAR::Vector3(), false);
            }

            if (pair.second->orientation->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "orientation",
                    pair.second->orientation->interpolate(nt).getEuler(),
                    pair.second->orientation->isChanging(pt));
            }

            if (pair.second->gain && pair.second->gain->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "gainDb",
                    pair.second->gain->interpolate(nt), pair.second->gain->isChanging(pt));
            }

            if (pair.second->itemReverbGain && pair.second->itemReverbGain->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "itemReverbGain",
                    pair.second->itemReverbGain->interpolate(nt),
                    pair.second->itemReverbGain->isChanging(pt));
            }

            if (pair.second->directiveness && pair.second->directiveness->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "gainDb",
                    pair.second->directiveness->interpolate(nt),
                    pair.second->directiveness->isChanging(pt));
            }
            if (pair.second->loudness && pair.second->loudness->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "loudnessDb",
                    pair.second->loudness->interpolate(nt), pair.second->loudness->isChanging(pt));
            }

            if (pair.second->granularPos && pair.second->granularPos->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "granularPos",
                    pair.second->granularPos->interpolate(nt),
                    pair.second->granularPos->isChanging(pt));
            }

            if (pair.second->granularGroupProb
                && pair.second->granularGroupProb->isChanging(nt, pt)) {
                timedUpdate.addModification(pair.first, "granularGroupProb",
                    pair.second->granularGroupProb->interpolate(nt),
                    pair.second->granularGroupProb->isChanging(pt));
            }
        }

        if (timedUpdate.modifications.size() > 0) {
            state.modify(timedUpdate);
        }

        auto updates = timedUpdates.playback(nt);

        for (auto u : updates) {
            state.modify(*u);
        }

        auto voxelSceneState = state.getVoxelSceneState();
        if (voxelSceneState) {
            std::chrono::duration<double> const elapsed(nt - pt);
            voxelSceneState->tick(std::chrono::duration_cast<std::chrono::milliseconds>(elapsed));
        }

        pt = nt;
    }
}

void Scene::triggerUpdateByIndex(int id, std::vector<std::pair<int, float>> placeholderValues)
{
    if (externalTriggersByIndex.find(id) != externalTriggersByIndex.end()) {
        auto upd = externalTriggersByIndex.at(id);

        // this is a work-around to avoid the following EAR_CHECK_REL message.
        // TODO: remove this if clause:
        if (upd->placeholders.size() != placeholderValues.size()) {
            return;
        }

        // fill placeholder values
        EAR_CHECK_REL(upd->placeholders.size() == placeholderValues.size(),
            "Not the right number of placeholder values for update with index '"
                + std::to_string(id) + "'.");
        for (auto& pair : placeholderValues) {
            *(upd->placeholders.at(pair.first)) = pair.second;
        }

        if (upd->update) {
            // has non-interpolated modifications
            if (upd->delay > 0.0) {
                timedUpdates.addElement(
                    sceneClock->getCurrentTime() + upd->delay, upd->update.get());
            } else {
                state.modify(*upd->update);
            }
        }

        // temporary interpolators
        for (auto& in : upd->tempInterpolators) {
            std::unique_ptr<EAR::Variant> initVal;
            if (in->attribute == "location") {
                initVal =
                    std::make_unique<EAR::Variant>(state.updateable[in->entity]->getLocation());
            } else if (in->attribute == "orientation") {
                initVal =
                    std::make_unique<EAR::Variant>(state.updateable[in->entity]->getOrientation());
            } else if (in->attribute == "gainDb") {
                if (auto ptr =
                        dynamic_cast<EAR::AirflowGeneratorBase*>(state.updateable[in->entity]);
                    ptr) {
                    initVal = std::make_unique<EAR::Variant>(ptr->getGainDb());
                } else {
                    initVal = std::make_unique<EAR::Variant>(
                        dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])->getGainDb());
                }
            } else if (in->attribute == "itemReverbGain") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])->getReverbGain());
            } else if (in->attribute == "directiveness") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                        ->getDirectiveness());
            } else if (in->attribute == "loudnessDb") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])->getLoudnessDb());
            } else if (in->attribute == "granularPos") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                        ->getGranularPos());
            }

            activeTempInterpolators.emplace_back(
                sceneClock->getCurrentTime() + upd->delay, in.get(), *initVal);
        }
    }

    auto voxelSceneState = state.getVoxelSceneState();
    if (voxelSceneState) {
        voxelSceneState->trigger_update(id, placeholderValues);
    }
}

void Scene::triggerUpdateById(
    const std::string& id, std::vector<std::pair<int, float>> placeholderValues)
{
    // throw error on missing ID in debug build
    EAR_CHECK(externalTriggersById.find(id) != externalTriggersById.end(),
        "Update with ID '" + id + "' not found.");

    if (externalTriggersById.find(id) != externalTriggersById.end()) {
        auto upd = externalTriggersById.at(id);

        // fill placeholder values
        EAR_CHECK_REL(upd->placeholders.size() == placeholderValues.size(),
            "Not the right number of placeholder values for update with ID '" + id + "'.");
        for (auto& pair : placeholderValues) {
            *(upd->placeholders.at(pair.first)) = pair.second;
        }

        if (upd->update) {
            // has non-interpolated modifications
            if (upd->delay > 0.0) {
                timedUpdates.addElement(
                    sceneClock->getCurrentTime() + upd->delay, upd->update.get());
            } else {
                state.modify(*upd->update);
            }
        }

        // temporary interpolators
        for (auto& in : upd->tempInterpolators) {
            std::unique_ptr<EAR::Variant> initVal;
            if (in->attribute == "location") {
                initVal =
                    std::make_unique<EAR::Variant>(state.updateable[in->entity]->getLocation());
            } else if (in->attribute == "orientation") {
                initVal =
                    std::make_unique<EAR::Variant>(state.updateable[in->entity]->getOrientation());
            } else if (in->attribute == "gainDb") {
                if (auto ptr =
                        dynamic_cast<EAR::AirflowGeneratorBase*>(state.updateable[in->entity]);
                    ptr) {
                    initVal = std::make_unique<EAR::Variant>(ptr->getGainDb());
                } else {
                    initVal = std::make_unique<EAR::Variant>(
                        dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])->getGainDb());
                }
            } else if (in->attribute == "itemReverbGain") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])->getReverbGain());
            } else if (in->attribute == "directiveness") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                        ->getDirectiveness());
            } else if (in->attribute == "loudnessDb") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])->getLoudnessDb());
            } else if (in->attribute == "granularPos") {
                initVal = std::make_unique<EAR::Variant>(
                    dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                        ->getGranularPos());
            }

            activeTempInterpolators.emplace_back(
                sceneClock->getCurrentTime() + upd->delay, in.get(), *initVal);
        }
    }
}

void Scene::triggerUpdate(const Update& upd)
{
    state.modify(upd);
}

void Scene::sceneObjectAdded(EAR::SceneObject* object)
{
    permanentInterpolators[object->getName()] = std::make_unique<PermanentInterpolator>(object);
}

void Scene::sceneObjectRemoved(EAR::SceneObject* object)
{
    permanentInterpolators.erase(object->getName());
}

inline void Scene::parsePlaceholders(const Modification& mod, EAR::Variant* val)
{
    for (auto& ph : mod.placeholders) {
        switch (ph.second) {
        // TODO: Find more elegant way to access data fields in Variant
        case Modification::Placeholder::PositionX:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::PositionY:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 1;
            break;
        case Modification::Placeholder::PositionZ:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 2;
            break;
        case Modification::Placeholder::OrientationY:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::OrientationP:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 1;
            break;
        case Modification::Placeholder::OrientationR:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 2;
            break;
        case Modification::Placeholder::GainDb:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::Speed:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::LoudnessDb:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::ReverbGain:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::GranularPosX:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::GranularPosY:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 1;
            break;
        case Modification::Placeholder::GranularPosZ:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 2;
            break;
        case Modification::Placeholder::GranularProbGroupNo:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData());
            break;
        case Modification::Placeholder::GranularGroupProb:
            externalTriggers.back()->placeholders[ph.first] =
                reinterpret_cast<float*>(val->getRawData()) + 1;
            break;
        }
    }
}

inline void Scene::addUpdate(const BitstreamUpdate& upd)
{
    typedef Modification::TargetAttribute Attr;

    if (upd.type == BitstreamUpdate::Type::Timed) {
        auto update = std::make_unique<Update>(upd.time + upd.delay);

        for (auto& mod : upd.modifications) {
            if (mod.attribute != Attr::LoudnessDb) {
                EAR_CHECK(permanentInterpolators[mod.id],
                    "Modifying entity '" + mod.id
                        + "' that has not been initialized for interpolation.");
                idsWithUpdates.insert(mod.id);
            }
            bool interpolate = (mod.duration > 0.f);
            double timestamp = upd.time + upd.delay;

            // add continuous attribute updates to permanent interpolator (gain, location, orientation, directiveness)
            // TODO: get rid of if cascade
            if (mod.attribute == Attr::Position) {
                EAR_CHECK(permanentInterpolators[mod.id]->location,
                    "Modifying position attribute of entity that has not been initialized for "
                    "interpolation.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    EAR::Vector3 startVal =
                        permanentInterpolators[mod.id]->location->interpolate(timestamp);
                    permanentInterpolators[mod.id]->location->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->location->addSupportingPoint(
                        timestamp + mod.duration, mod.targetValue, true);
                } else {
                    permanentInterpolators[mod.id]->location->addSupportingPoint(
                        timestamp, mod.targetValue, interpolate);
                }
            } else if (mod.attribute == Attr::Orientation) {
                EAR_CHECK(permanentInterpolators[mod.id]->orientation,
                    "Modifying orientation attribute of entity that has not been initialized for "
                    "interpolation.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    EAR::Quaternion startVal =
                        permanentInterpolators[mod.id]->orientation->interpolate(timestamp);
                    permanentInterpolators[mod.id]->orientation->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->orientation->addSupportingPoint(
                        timestamp + mod.duration,
                        EAR::Quaternion(static_cast<EAR::Euler>(mod.targetValue)), true);
                } else {
                    permanentInterpolators[mod.id]->orientation->addSupportingPoint(timestamp,
                        EAR::Quaternion(static_cast<EAR::Euler>(mod.targetValue)), interpolate);
                }

            } else if (mod.attribute == Attr::GainDb) {
                // EAR_CHECK_REL(permanentInterpolators[mod.id]->gain,
                //     "Modifying gainDb attribute of entity that is not a sound source.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    float startVal = permanentInterpolators[mod.id]->gain->interpolate(timestamp);
                    permanentInterpolators[mod.id]->gain->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->gain->addSupportingPoint(
                        timestamp + mod.duration, mod.targetValue, true);
                } else {
                    permanentInterpolators[mod.id]->gain->addSupportingPoint(
                        timestamp, mod.targetValue, interpolate);
                }
            } else if (mod.attribute == Attr::ReverbGain) {
                EAR_CHECK_REL(permanentInterpolators[mod.id]->itemReverbGain,
                    "Modifying reverbGain attribute of entity that is not a sound source.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    float startVal =
                        permanentInterpolators[mod.id]->itemReverbGain->interpolate(timestamp);
                    permanentInterpolators[mod.id]->itemReverbGain->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->itemReverbGain->addSupportingPoint(
                        timestamp + mod.duration, mod.targetValue, true);
                } else {
                    permanentInterpolators[mod.id]->itemReverbGain->addSupportingPoint(
                        timestamp, mod.targetValue, interpolate);
                }
            } else if (mod.attribute == Attr::Directiveness) {
                EAR_CHECK_REL(permanentInterpolators[mod.id]->directiveness,
                    "Modifying directiveness attribute of entity that is not a sound source.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    float startVal =
                        permanentInterpolators[mod.id]->directiveness->interpolate(timestamp);
                    permanentInterpolators[mod.id]->directiveness->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->directiveness->addSupportingPoint(
                        timestamp + mod.duration, mod.targetValue, true);
                } else {
                    permanentInterpolators[mod.id]->directiveness->addSupportingPoint(
                        timestamp, mod.targetValue, interpolate);
                }
            } else if (mod.attribute == Attr::LoudnessDb) {
                // get all sources where loudness need to be changed.
                for (auto& obj : state.objectSources) {
                    const SceneState::AudioStream& stream =
                        state.getAudioStream(dynamic_cast<EAR::SceneObject*>(obj.get()));
                    if (stream.id == mod.id) {
                        auto id = obj->getName();
                        idsWithUpdates.insert(id);
                        EAR_CHECK_REL(permanentInterpolators[id]->loudness,
                            "Modifying loudness attribute of entity that is not a sound source.");
                        if (mod.duration > 0.f) {
                            // add a starting point for the interpolation if duration is given
                            float startVal =
                                permanentInterpolators[id]->loudness->interpolate(timestamp);
                            permanentInterpolators[id]->loudness->addSupportingPoint(
                                timestamp, startVal, interpolate);
                            permanentInterpolators[id]->loudness->addSupportingPoint(
                                timestamp + mod.duration, mod.targetValue, true);
                        } else {
                            permanentInterpolators[id]->loudness->addSupportingPoint(
                                timestamp, mod.targetValue, interpolate);
                        }
                    }
                }
                for (auto& obj : state.hoaSources) {
                    const SceneState::AudioStream& stream =
                        state.getAudioStream(dynamic_cast<EAR::SceneObject*>(obj.get()));
                    if (stream.id == mod.id) {
                        auto id = obj->getName();
                        idsWithUpdates.insert(id);
                        EAR_CHECK_REL(permanentInterpolators[id]->loudness,
                            "Modifying loudness attribute of entity that is not a sound source.");
                        if (mod.duration > 0.f) {
                            // add a starting point for the interpolation if duration is given
                            float startVal =
                                permanentInterpolators[id]->loudness->interpolate(timestamp);
                            permanentInterpolators[id]->loudness->addSupportingPoint(
                                timestamp, startVal, interpolate);
                            permanentInterpolators[id]->loudness->addSupportingPoint(
                                timestamp + mod.duration, mod.targetValue, true);
                        } else {
                            permanentInterpolators[id]->loudness->addSupportingPoint(
                                timestamp, mod.targetValue, interpolate);
                        }
                    }
                }

            } else if (mod.attribute == Attr::GranularPos) {
                EAR_CHECK_REL(permanentInterpolators[mod.id]->granularPos,
                    "Modifying granular descriptor space position attribute of entity that is not "
                    "a sound source.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    EAR::Vector3 startVal =
                        permanentInterpolators[mod.id]->granularPos->interpolate(timestamp);
                    permanentInterpolators[mod.id]->granularPos->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->granularPos->addSupportingPoint(
                        timestamp + mod.duration, mod.targetValue, true);
                } else {
                    permanentInterpolators[mod.id]->granularPos->addSupportingPoint(
                        timestamp, mod.targetValue, interpolate);
                }
            } else if (mod.attribute == Attr::GranularGroupProb) {
                EAR_CHECK_REL(permanentInterpolators[mod.id]->granularGroupProb,
                    "Modifying granular probability group attribute of entity that is "
                    "not a sound source.");
                if (mod.duration > 0.f) {
                    // add a starting point for the interpolation if duration is given
                    EAR::Vector3 startVal =
                        permanentInterpolators[mod.id]->granularGroupProb->interpolate(timestamp);
                    startVal[0] = mod.targetValue.operator EAR::Vector3()[0];
                    permanentInterpolators[mod.id]->granularGroupProb->addSupportingPoint(
                        timestamp, startVal, interpolate);
                    permanentInterpolators[mod.id]->granularGroupProb->addSupportingPoint(
                        timestamp + mod.duration, mod.targetValue, true);
                } else {
                    permanentInterpolators[mod.id]->granularGroupProb->addSupportingPoint(
                        timestamp, mod.targetValue, interpolate);
                }
            } else if (mod.attribute == Attr::Signal || mod.attribute == Attr::Extent
                || mod.attribute == Attr::Directivity || mod.attribute == Attr::Group
                || mod.attribute == Attr::Region) {
                // this update is an int id but should internally be handled as a string
                update->addModification(
                    mod.id, detail::attr2string.at(mod.attribute), mod.targetValue, interpolate);
            } else {
                update->addModification(
                    mod.id, detail::attr2string.at(mod.attribute), mod.targetValue, interpolate);
            }
        }

        if (update->modifications.size() > 0) {
            timedUpdateStore.push_back(std::move(update));
            // add this update to TapeDeck
            timedUpdates.addElement(upd.time + upd.delay, timedUpdateStore.back().get());
        }

    } else if (upd.type == BitstreamUpdate::Type::Conditional) {
        // (only proximity conditions exist so far)
        proximityTriggers.emplace_back(state.getGeometry(upd.condition), upd.fireOn, upd.delay);

        auto update = std::make_unique<Update>();

        for (auto& mod : upd.modifications) {
            if (mod.attribute != Attr::LoudnessDb) {
                idsWithUpdates.insert(mod.id);
            }
            bool interpolate = (mod.duration > 0.f);

            // pre-create update with all modifications and potential temporary interpolators
            if (mod.duration > 0
                && (mod.attribute == Attr::Position || mod.attribute == Attr::Orientation
                    || mod.attribute == Attr::GainDb || mod.attribute == Attr::Directiveness
                    || mod.attribute == Attr::ReverbGain || mod.attribute == Attr::LoudnessDb
                    || mod.attribute == Attr::GranularPos)) {

                if (mod.attribute != Attr::LoudnessDb) {
                    proximityTriggers.back().tempInterpolators.emplace_back(
                        new TempInterpolator(mod.id, detail::attr2string.at(mod.attribute),
                            mod.duration, mod.targetValue));
                } else {
                    // get all sources where loudness need to be changed.
                    for (auto& obj : state.objectSources) {
                        const SceneState::AudioStream& stream =
                            state.getAudioStream(dynamic_cast<EAR::SceneObject*>(obj.get()));
                        if (stream.id == mod.id) {
                            auto id = obj->getName();
                            idsWithUpdates.insert(id);
                            proximityTriggers.back().tempInterpolators.emplace_back(
                                new TempInterpolator(id, detail::attr2string.at(mod.attribute),
                                    mod.duration, mod.targetValue));
                        }
                    }
                }
            } else if (mod.attribute == Attr::Signal || mod.attribute == Attr::Extent
                || mod.attribute == Attr::Directivity || mod.attribute == Attr::Group
                || mod.attribute == Attr::Region) {
                // this update is an int id but should internally be handled as a string
                update->addModification(
                    mod.id, detail::attr2string.at(mod.attribute), mod.targetValue, interpolate);
            } else {
                // non-interpolatable attributes (or those with duration 0)
                // can be handled with a simple state update
                update->addModification(
                    mod.id, detail::attr2string.at(mod.attribute), mod.targetValue, interpolate);
            }
        }

        if (update->modifications.size() > 0) {
            proximityTriggers.back().update = std::move(update);
        }
    } else { // triggered update implied
        externalTriggers.emplace_back(new ExternalTrigger(upd.delay));

        externalTriggers.back()->update = std::make_unique<Update>();
        auto update = externalTriggers.back()->update.get();
        // reserve right amount of memory to keep pointers to the elements valid (used for placeholders -> not a very nice solution...)
        // TODO: find better solution for this...
        update->modifications.reserve(upd.modifications.size());

        for (auto& mod : upd.modifications) {
            if (mod.attribute != Attr::LoudnessDb) {
                idsWithUpdates.insert(mod.id);
            }
            bool interpolate = (mod.duration > 0.f);

            EAR::Variant* val = NULL; // used for placeholders

            // pre-create update with all modifications and potential temporary interpolators

            if (mod.duration > 0
                && (mod.attribute == Attr::Position || mod.attribute == Attr::Orientation
                    || mod.attribute == Attr::GainDb || mod.attribute == Attr::Directiveness
                    || mod.attribute == Attr::ReverbGain || mod.attribute == Attr::LoudnessDb
                    || mod.attribute == Attr::GranularPos)) {

                if (mod.attribute != Attr::LoudnessDb) {
                    externalTriggers.back()->tempInterpolators.emplace_back(
                        new TempInterpolator(mod.id, detail::attr2string.at(mod.attribute),
                            mod.duration, mod.targetValue));
                    val = &(externalTriggers.back()->tempInterpolators.back()->targetVal);
                } else {
                    // get all sources where loudness need to be changed.
                    for (auto& obj : state.objectSources) {
                        const SceneState::AudioStream& stream =
                            state.getAudioStream(dynamic_cast<EAR::SceneObject*>(obj.get()));
                        if (stream.id == mod.id) {
                            auto id = obj->getName();
                            idsWithUpdates.insert(id);
                            externalTriggers.back()->tempInterpolators.emplace_back(
                                new TempInterpolator(id, detail::attr2string.at(mod.attribute),
                                    mod.duration, mod.targetValue));
                            val = &(externalTriggers.back()->tempInterpolators.back()->targetVal);
                            // save pointers to placeholders to quickly insert the right values when called
                            parsePlaceholders(mod, val);
                        }
                    }
                }
            } else if (mod.attribute == Attr::Signal || mod.attribute == Attr::Extent
                || mod.attribute == Attr::Directivity || mod.attribute == Attr::Group
                || mod.attribute == Attr::Region) {
                // this update is an int id but should internally be handled as a string
                update->addModification(
                    mod.id, detail::attr2string.at(mod.attribute), mod.targetValue, interpolate);
            } else {
                // non-interpolatable attributes (or those with duration 0)
                // can be handled with a simple state update

                if (mod.attribute != Attr::LoudnessDb) {
                    update->addModification(mod.id, detail::attr2string.at(mod.attribute),
                        mod.targetValue, interpolate);
                    val = &(update->modifications.back().targetValue);
                } else {
                    // get all sources where loudness need to be changed.
                    for (auto& obj : state.objectSources) {
                        const SceneState::AudioStream& stream =
                            state.getAudioStream(dynamic_cast<EAR::SceneObject*>(obj.get()));
                        if (stream.id == mod.id) {
                            auto id = obj->getName();
                            idsWithUpdates.insert(id);
                            update->addModification(id, detail::attr2string.at(mod.attribute),
                                mod.targetValue, interpolate);
                            val = &(update->modifications.back().targetValue);
                            // save pointers to placeholders to quickly insert the right values when called
                            parsePlaceholders(mod, val);
                        }
                    }
                }
            }

            if (mod.attribute != Attr::LoudnessDb) {
                // save pointers to placeholders to quickly insert the right values when called
                parsePlaceholders(mod, val);
            }
        }

        if (update->modifications.size() == 0) {
            // remove update if it is empty to not trigger it uselessly
            externalTriggers.back()->update = std::unique_ptr<Update>(nullptr);
        }

        externalTriggersById[upd.id] = externalTriggers.back().get();
        if (upd.index != -1) {
            externalTriggersByIndex[upd.index] = externalTriggers.back().get();
        }
    }
}

inline void Scene::checkProximityTriggers()
{
    EAR::Vector3 listenerPos = state.getListener()->getLocation();
    // TODO: only check when listener actually moved

    for (auto& cond : proximityTriggers) {
        bool isInside = cond.region->isInside(listenerPos);
        if (isInside && !cond.active) {
            cond.active = true;
            if (cond.flank) {
                if (cond.update) {
                    // has non-interpolated modifications
                    if (cond.delay > 0.0) {
                        timedUpdates.addElement(
                            sceneClock->getCurrentTime() + cond.delay, cond.update.get());
                    } else {
                        state.modify(*cond.update);
                    }
                }

                // temporary interpolators
                for (auto& in : cond.tempInterpolators) {
                    std::unique_ptr<EAR::Variant> initVal;
                    if (in->attribute == "location") {
                        initVal = std::make_unique<EAR::Variant>(
                            state.updateable[in->entity]->getLocation());
                    } else if (in->attribute == "orientation") {
                        initVal = std::make_unique<EAR::Variant>(
                            state.updateable[in->entity]->getOrientation());
                    } else if (in->attribute == "gainDb") {
                        if (auto ptr = dynamic_cast<EAR::AirflowGeneratorBase*>(
                                state.updateable[in->entity]);
                            ptr) {
                            initVal = std::make_unique<EAR::Variant>(ptr->getGainDb());
                        } else {
                            initVal = std::make_unique<EAR::Variant>(
                                dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                    ->getGainDb());
                        }
                    } else if (in->attribute == "itemReverbGain") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getReverbGain());
                    } else if (in->attribute == "directiveness") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getDirectiveness());
                    } else if (in->attribute == "loudnessDb") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getLoudnessDb());
                    } else if (in->attribute == "granularPos") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getGranularPos());
                    }

                    activeTempInterpolators.emplace_back(
                        sceneClock->getCurrentTime() + cond.delay, in.get(), *initVal);
                }
            }
        } else if (!isInside && cond.active) {
            cond.active = false;
            if (!cond.flank) {
                if (cond.update) {
                    // has non-interpolated modifications
                    if (cond.delay > 0.0) {
                        timedUpdates.addElement(
                            sceneClock->getCurrentTime() + cond.delay, cond.update.get());
                    } else {
                        state.modify(*cond.update);
                    }
                }

                // temporary interpolators
                for (auto& in : cond.tempInterpolators) {
                    std::unique_ptr<EAR::Variant> initVal;
                    if (in->attribute == "location") {
                        initVal = std::make_unique<EAR::Variant>(
                            state.updateable[in->entity]->getLocation());
                    } else if (in->attribute == "orientation") {
                        initVal = std::make_unique<EAR::Variant>(
                            state.updateable[in->entity]->getOrientation());
                    } else if (in->attribute == "gainDb") {
                        if (auto ptr = dynamic_cast<EAR::AirflowGeneratorBase*>(
                                state.updateable[in->entity]);
                            ptr) {
                            initVal = std::make_unique<EAR::Variant>(ptr->getGainDb());
                        } else {
                            initVal = std::make_unique<EAR::Variant>(
                                dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                    ->getGainDb());
                        }
                    } else if (in->attribute == "itemReverbGain") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getReverbGain());
                    } else if (in->attribute == "directiveness") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getDirectiveness());
                    } else if (in->attribute == "loudnessDb") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getLoudnessDb());
                    } else if (in->attribute == "granularPos") {
                        initVal = std::make_unique<EAR::Variant>(
                            dynamic_cast<EAR::PointSource*>(state.updateable[in->entity])
                                ->getGranularPos());
                    }

                    activeTempInterpolators.emplace_back(
                        sceneClock->getCurrentTime() + cond.delay, in.get(), *initVal);
                }
            }
        }
    }
}

} // namespace MPEGI
