/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Scene/SceneState.h>

#include <MPEGI/Bitstream/CompressedDirectivity.h>
#include <MPEGI/Bitstream/DirectivityPayload.h>
#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Bitstream/VoxDataPayload.h>

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/ListenerVoiceDefaultDirectivityData.hpp>

#include <MPEGI/Scene.h>
#include <MPEGI/Scene/Update.h>

#include <EAR/AirflowGeneratorBase.h>
#include <EAR/AnchorObject.h>
#include <EAR/BaseTypes.h>
#include <EAR/ConeAirflowGenerator.h>
#include <EAR/Exception.h>
#include <EAR/Geometry.h>
#include <EAR/GlobalAirflowGenerator.h>
#include <EAR/HOASource.h>
#include <EAR/Listener.h>
#include <EAR/ListenerVoiceSource.h>
#include <EAR/LoudspeakerLayout.h>
#include <EAR/PlaneAirflowGenerator.h>
#include <EAR/PointAirflowGenerator.h>
#include <EAR/PointSource.h>
#include <EAR/Portal.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>
#include <EAR/SceneObjectObserver.h>

#include <mpegcore/DirectionalDecay.hpp>
#include <mpegcore/ReverberatorParameters.hpp>

#include <algorithm>
#include <cassert>
#include <spdlog/fmt/fmt.h>

namespace MPEGI {

SceneState::SceneState(
    const Bitstream& bitstream, const Scene* parent, const LSDF::LSDFParser* lsdf)
    : parentScene(parent)
{
    auto eifPayload = std::dynamic_pointer_cast<const EIFPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));

    const bool& enableLocallyCapturedAudio = bitstream.getEnableLocallyCapturedAudio();
    const std::string& listenerVoiceDirectivityId = bitstream.getListenerVoiceDirectivityId();

    bool hasDefaultAe = false;
    auto reverbPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
    if (reverbPayloads.size() != 0) {
        auto bitstreamReverbPayload =
            std::dynamic_pointer_cast<const MPEGI::ReverbPayload>(reverbPayloads[0]);
        hasDefaultAe = bitstreamReverbPayload->revDefaultAEPresent;
    }

    auto directivityPayload = std::dynamic_pointer_cast<const DirectivityPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_DIRECTIVITY));

    listener = EAR::SceneObjectFactory<EAR::Listener>::getNewObject("listener");
    listener->setMoved(); // force initial filter design

    updateable["listener"] = listener.get();

    for (auto& mat : eifPayload->getAcousticMaterials()) {
        createMaterial(mat);
    }

    for (auto& transform : eifPayload->getTransforms()) {
        createTransform(transform);
    }

    for (auto& anchor : eifPayload->getAnchors()) {
        createAnchor(anchor);
    }

    for (auto& mesh : eifPayload->getMeshes()) {
        createMesh(mesh);
    }

    for (auto& box : eifPayload->getBoxes()) {
        createBox(box);
    }

    for (auto& sph : eifPayload->getSpheres()) {
        createSphere(sph);
    }

    for (auto& cyl : eifPayload->getCylinders()) {
        createCylinder(cyl);
    }

    for (auto& [id, directivity] : directivityPayload->getDirectivities()) {
        createDirectivity(id, directivity);
    }

    for (auto& obj : eifPayload->getObjectSources()) {
        createObjectSource(obj);
    }

    for (auto& src : eifPayload->getChannelSources()) {
        createChannelSource(src);
    }

    for (auto& src : eifPayload->getHoaSources()) {
        createHOASource(src);
    }

    for (auto& src : eifPayload->getHoaGroups()) {
        createHOAGroup(src);
    }

    for (auto env : eifPayload->getAcousticEnvironments()) {
        if (env.id == "defaultAE" && !hasDefaultAe) {
            continue;
        }
        createReverbEnv(env);
    }

    for (auto& gen : eifPayload->getGlobalAirflowGenerators()) {
        createGlobalAirflowGenerator(gen);
    }

    for (auto& gen : eifPayload->getPointAirflowGenerators()) {
        createPointAirflowGenerator(gen);
    }

    for (auto& gen : eifPayload->getConeAirflowGenerators()) {
        createConeAirflowGenerator(gen);
    }

    for (auto& gen : eifPayload->getPlaneAirflowGenerators()) {
        createPlaneAirflowGenerator(gen);
    }

    auto portalPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_PORTAL);
    EAR_CHECK_REL(portalPayloads.size() <= 1, "Too many portal payloads in bitstream");
    if (portalPayloads.size() == 1) {
        auto portalPayload = std::dynamic_pointer_cast<const PortalPayload>(
            bitstream.getOnlyPayloadOfType(Bitstream::PLD_PORTAL));
        if (portalPayload->getExplicitPortalMode()) {
            portals.reserve(portalPayload->getPortals().size());
            for (const auto& portal : portalPayload->getPortals()) {
                createPortal(portal.get());
            }
        }
    }

    // initialize update timestamps
    for (auto& pair : updateable) {
        pair.second->setPositionTimestamp(0.0);
        teleport[pair.first] = false;
    }

    // incorporate information from LSDF
    if (lsdf) {
        for (auto& mat : lsdf->getAcousticMaterials()) {
            createLSDFMaterial(mat);
        }

        for (auto& mesh : lsdf->getMeshes()) {
            createLSDFMesh(mesh);
        }

        for (auto& arAnchor : lsdf->getARAnchors()) {
            if (anchorsByLSDFRef.find(arAnchor.id) != anchorsByLSDFRef.end()) {
                anchorsByLSDFRef[arAnchor.id]->setPosition(arAnchor.position);
            }
        }

        for (auto env : lsdf->getAcousticEnivronments()) {
            createLSDFReverbEnv(env);
        }

        const auto& setups = lsdf->getLoudspeakerSetups();
        if (setups.size() > 0) {
            this->loudspeakerSetup = setups[0];
        } else if (setups.size() > 1) {
            EAR_FAIL("Only a single LoudspeakerSetup is allowed in LSDF");
        }

        if (enableLocallyCapturedAudio) {
            for (auto localSource : lsdf->getLocalSources()) {
                createLSDFLocalSource(localSource, listenerVoiceDirectivityId);
            }
        }
    }

    // initialize update timestamps
    for (auto& pair : updateable) {
        pair.second->setPositionTimestamp(0.0);
        teleport[pair.first] = false;
    }

    // Voxdata
    try {
        createVoxelSceneState(std::dynamic_pointer_cast<VoxDataPayload const>(
            bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_VOXDATA)));
    }
    /// @todo Need to catch EAR::Exception here as it does not inherit publicly from std::exception
    catch (EAR::Exception const&) {
        // No Voxdata-payload in the stream
        voxelSceneState.reset();
    }

    // combine early reflection geometries into a single static mesh
    combineEarlyReflectionMesh();

    // combine diffraction geometries into a single static mesh
    combineDiffractionMesh();
}

// Empty constructor
SceneState::SceneState() noexcept
    : parentScene{ nullptr }
{}

SceneState::~SceneState()
{
    // notify observers that here is nothing to observe anymore...
    for (auto o : observers) {
        o->sceneStateDetached();
    }
}

double SceneState::getCurrentTime() const
{
    return parentScene->getCurrentTime();
}

void SceneState::attach(SceneStateObserver* observer) const
{
    observers.push_back(observer);

    observer->sceneStateAttached(this);

    for (auto& a : anchors) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(a.get()));
    }

    for (auto& src : objectSources) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(src.get()));
    }

    for (auto& src : hoaSources) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(src.get()));
    }

    for (auto& src : lsLayouts) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(src.get()));
    }

    for (auto& geo : geometries) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(geo.second));
    }

    for (auto& gen : airflowGenerators) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(gen.get()));
    }

    for (auto& portal : portals) {
        observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(portal.get()));
    }

    observer->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(listener.get()));
}

void SceneState::detach(SceneStateObserver* observer) const
{
    auto it = std::find(observers.begin(), observers.end(), observer);

    if (it != observers.end()) {
        (*it)->sceneStateDetached();
        observers.erase(it);
    }
}

void SceneState::modify(const Update& update)
{
    std::lock_guard<std::mutex> lock(sequentialUpdateMutex);

    double ts = update.timestamp;
    //double nt = getCurrentTime();
    //if (ts < nt) {
    //    ts = nt;
    //}

    for (auto& mod : update.modifications) {

        // TODO: Why does the entityId contain gibberish in debug builds?

        SceneObject* sceneObject = nullptr;
        if (mod.entityId == "realListener") {
            if (mod.attributeName == "location") {
                listener->realPosition.location = mod.targetValue;
            } else if (mod.attributeName == "orientation") {
                listener->realPosition.orientation = mod.targetValue;
            } else {
                EAR_FAIL("Invalid attribute for realListener: " + mod.attributeName);
            }
            listener->notify(SceneObjectObserver::Property::Position);
            continue;
        } else if (auto it = updateable.find(mod.entityId); it != updateable.end()) {
            sceneObject = it->second;
        } else {
            continue; // silently ignore updates with unknown ID
        }
        assert(sceneObject);

        if (mod.attributeName == "location") {
            teleport[mod.entityId] = !mod.interpolate;
            sceneObject->setPositionTimestamp(ts);
            sceneObject->setLocation(mod.targetValue);
            continue;
        } else if (mod.attributeName == "velocity") {
            sceneObject->setPositionTimestamp(ts);
            sceneObject->setVelocity(mod.targetValue);
            continue;
        } else if (mod.attributeName == "orientation") {
            sceneObject->setPositionTimestamp(ts);
            sceneObject->setOrientation(mod.targetValue);
            continue;
        } else if (mod.attributeName == "cspace") {
            sceneObject->setCSpace(mod.targetValue);
            continue;
        } else if (mod.attributeName == "active") {
            sceneObject->setActive(mod.targetValue);
            continue;
        } else if (mod.attributeName == "signal") {
            audioStreamProperties[sceneObject].id = static_cast<std::string>(mod.targetValue);

            // TODO: Support channel sources? We would need to set ".channels" correctly.

            // make sure that the new signal is not played from the beginning
            // TODO: this should be a parameter of the update once we leave EIF phase
            if (audioStreamProperties[sceneObject].looping) {
                audioStreamProperties[sceneObject].seekTime = -1.0;
            }
            sceneObject->notify(EAR::SceneObjectObserver::Property::AudioStream);
        } else if (mod.attributeName == "play") {
            audioStreamProperties[sceneObject].playing = mod.targetValue;
            // play from beginning of stream
            audioStreamProperties[sceneObject].seekTime = 0.0;
            if (auto* cs = dynamic_cast<EAR::LoudspeakerLayout*>(sceneObject)) {
                for (auto& ls : cs->loudspeakers) {
                    audioStreamProperties[ls.get()].playing = mod.targetValue;
                    // play from beginning of stream
                    audioStreamProperties[ls.get()].seekTime = 0.0;
                }
            }
            sceneObject->notify(EAR::SceneObjectObserver::Property::AudioStream);
            if (mod.targetValue) {
                sceneObject->notify(EAR::SceneObjectObserver::Property::Play);
            } else {
                sceneObject->notify(EAR::SceneObjectObserver::Property::Stop);
            }
            continue;
        } else if (mod.attributeName == "mode") {
            bool looping = static_cast<std::string>(mod.targetValue) != "event";
            audioStreamProperties[sceneObject].looping = looping;
            if (auto* cs = dynamic_cast<EAR::LoudspeakerLayout*>(sceneObject)) {
                for (auto& ls : cs->loudspeakers) {
                    audioStreamProperties[ls.get()].looping = looping;
                }
            }
            sceneObject->notify(EAR::SceneObjectObserver::Property::AudioStream);
            continue;
        }

        auto psrc = dynamic_cast<EAR::PointSource*>(sceneObject);

        if (psrc) {
            if (mod.attributeName == "directivity") {
                psrc->setDirectivity(directivities[mod.targetValue]);
                continue;
            } else if (mod.attributeName == "directiveness") {
                psrc->setDirectiveness(mod.targetValue);
                continue;
            } else if (mod.attributeName == "gainDb") {
                psrc->setGainDb(mod.targetValue);
                continue;
            } else if (mod.attributeName == "itemReverbGain") {
                psrc->setReverbGain(mod.targetValue);
                continue;
            } else if (mod.attributeName == "noReverb") {
                psrc->setNoReverb(mod.targetValue);
            } else if (mod.attributeName == "loudnessDb") {
                psrc->setLoudnessDb(mod.targetValue);
                continue;
            } else if (mod.attributeName == "extent") {
                psrc->setExtent(geometries[mod.targetValue]);
            } else if (mod.attributeName == "signal") {
                psrc->setAudioStreamID(mod.targetValue);
                continue;
            } else if (mod.attributeName == "granularPos") {
                psrc->setGranularPos(mod.targetValue);
                continue;
            } else if (mod.attributeName == "granularGroupProb") {
                psrc->setGranularGroupProb(mod.targetValue);
            }
        }

        if (auto src = dynamic_cast<EAR::LoudspeakerLayout*>(sceneObject); src) {
            if (mod.attributeName == "gainDb") {
                src->setGainDb(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "itemReverbGain") {
                src->setReverbGain(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "noReverb") {
                src->setNoReverb(mod.targetValue);
                continue;
            }
        }

        if (auto src = dynamic_cast<EAR::HOASource*>(sceneObject); src) {
            if (mod.attributeName == "gainDb") {
                src->setGainDb(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "itemReverbGain") {
                src->setReverbGain(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "noReverb") {
                src->setNoReverb(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "group") {
                // Fail until there's a scene needing this and we know how to handle it
                EAR_FAIL("Changing HOASource group is not supported yet");
            }
            if (mod.attributeName == "granularPos") {
                src->setGranularPos(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "granularGroupProb") {
                src->setGranularGroupProb(mod.targetValue);
                continue;
            }
        }

        if (auto gen = dynamic_cast<EAR::AirflowGeneratorBase*>(updateable[mod.entityId]); gen) {
            if (mod.attributeName == "gainDb") {
                gen->setGainDb(mod.targetValue);
                continue;
            }
            if (mod.attributeName == "speed") {
                gen->setSpeed(mod.targetValue);
                continue;
            }
        }

        if (auto portal = dynamic_cast<EAR::Portal*>(updateable[mod.entityId]); portal) {
            if (mod.attributeName == "material") {
                portal->setMaterial(materials.at(mod.targetValue));
                continue;
            }
        }
    }
}

bool SceneState::isTeleporting(const std::string& id) const
{
    return teleport.at(id);
}

Cell* SceneState::getCellForLocation(const Vector3& point)
{
    for (auto& c : cells) {
        if (c.second->enclosure->isInside(point)) {
            return c.second.get();
        }
    }
    return nullptr;
}

void SceneState::createVoxelSceneState(std::shared_ptr<VoxDataPayload const> const& voxDataPayload)
{
    voxelSceneState = std::make_shared<Voxel::VoxelSceneState>(*voxDataPayload);

    // To receive listener position updates
    listener->attach(voxelSceneState.get());
}

void SceneState::createReverbEnv(const EIFPayload::AcousticEnvironmentDesc& env)
{
    EAR::Geometry* bounds = nullptr;
    if (env.region != nullptr) {
        bounds = geometries[env.region->id];
    }
    reverbEnvs.push_back(std::make_unique<ReverbEnv>(env.id, bounds));
    reverbEnvs.back()->acousticEnvironmentBsId = env.bsId;
    reverbEnvs.back()->isSceneBasedAE = 1;
}

inline void SceneState::createLSDFReverbEnv(
    const LSDF::LSDFParser::AcousticEnvironmentDesc& acousticEnv)
{
    EAR_CHECK_REL(acousticEnv.region != nullptr,
        "The renderer currently does not support unbounded AcousticEnvironments (this does NOT "
        "comply with the EIF spec).");
    EAR_CHECK_REL(
        std::find_if(reverbEnvs.begin(), reverbEnvs.end(),
            [&id = acousticEnv.id](auto const& env) { return env->acousticEnvironmentId == id; })
            == reverbEnvs.end(),
        "The renderer currently does not support LSDF environments sharing an ID with an "
        "environment specified in the scene file.");

    EAR::Geometry* bounds = geometries[acousticEnv.region->id];
    reverbEnvs.push_back(std::make_unique<ReverbEnv>(acousticEnv.id, bounds));
    reverbEnvs.back()->isLsdf = true;
    reverbEnvs.back()->isSceneBasedAE = 0;

    std::size_t numDelayLines = REVERB_DEFAULT_NUMBER_OF_DELAY_LINE_LOOPS;
    std::size_t numDelayLinesFromConfig =
        static_cast<std::size_t>(Config::getIntDef("Reverb:numDelayLines", EAR::Variant(0)));

    // If there is more than one acoustic parameter set, only the first is used
    // to initialize values for the LSDF reverb environment.
    if (!acousticEnv.acousticParameters.empty()) {
        mpegcore::printStatus("Using AcousticParameters");

        auto const& acousticDesc = acousticEnv.acousticParameters.front();
        for (auto& freqData : acousticDesc.values) {
            reverbEnvs.back()->append(freqData.f, freqData.rt60, freqData.rdr);
        }
        // In an AR scene context, predelay is defined as the diffuseOnset
        reverbEnvs.back()->predelay = acousticDesc.diffuseOnset;
        reverbEnvs.back()->predelayOffset = acousticDesc.predelayOffset;
        reverbEnvs.back()->complexityLevel = acousticDesc.complexityLevel;

        if (numDelayLinesFromConfig) {
            numDelayLines = numDelayLinesFromConfig;
        } else if (reverbEnvs.back()->complexityLevel.has_value()) {
            numDelayLines = mpegcore::complexityLevelToNumDelayLines(
                reverbEnvs.back()->complexityLevel.value());
        }
        reverbEnvs.back()->numDelayLines = numDelayLines;

    } else if (!acousticEnv.rirParameters.empty()) {

        mpegcore::printStatus("Using RoomImpulseResponse parameters");

        // Only 1 RIR is supported in an AcousticEnvironment
        auto const& rirDesc = acousticEnv.rirParameters.front();

        reverbEnvs.back()->isFromRIR = true;
        reverbEnvs.back()->predelayOffset = rirDesc.predelayOffset;
        reverbEnvs.back()->complexityLevel = rirDesc.complexityLevel;
        reverbEnvs.back()->srcDistanceIsKnown =
            rirDesc.sourcePosition.has_value() && rirDesc.micPosition.has_value();

        if (numDelayLinesFromConfig) {
            numDelayLines = numDelayLinesFromConfig;
        } else if (reverbEnvs.back()->complexityLevel.has_value()) {
            numDelayLines = mpegcore::complexityLevelToNumDelayLines(
                reverbEnvs.back()->complexityLevel.value());
        }
        reverbEnvs.back()->numDelayLines = numDelayLines;

        /* Decay analysis */

        // Input format - ordering and normalization
        mpegcore::HoaNormalization rirNormalization;
        std::string rirFormatString = rirDesc.format;
        if (rirFormatString == "ACN-SN3D") {
            rirNormalization = mpegcore::HoaNormalization::sn3d;
        } else {
            // "ACN-N3D", even for mono
            rirNormalization = mpegcore::HoaNormalization::n3d;
        }

        // Initialize and run RIR analysis
        mpegcore::SHDecay decayInfo{ rirDesc.path, rirNormalization, rirDesc.sourcePosition,
            rirDesc.micPosition, rirDesc.directivity, nullptr };
        decayInfo.init(*reverbEnvs.back()->numDelayLines);
        decayInfo.setRIR_file();
        rirlib_data* srirAnalysis = (rirlib_data*)decayInfo.pRirLib;
        // Mono RIR or analysis of SH omni component.
        decayInfo.analyze(RDR_DONE);

        // Populate reverbEnv with RIR analysis results (T60, RDR, directional gain)
        int numAnalysisBands = srirAnalysis->nBand;
        for (int ib = 0; ib < numAnalysisBands; ib++) {
            reverbEnvs.back()->append(srirAnalysis->centerFreqs[ib], srirAnalysis->t60Omni_bnd[ib],
                srirAnalysis->rdr_bnd[ib]);
        }
        if (srirAnalysis->dirGainDB_bnd_dir != NULL) {
            int numAnalysisDirs = srirAnalysis->nDir;

            std::vector<ReverbEnv::DirectionalGain> srirDirectionalGainData(numAnalysisDirs);

            const float dirGainFactor = static_cast<std::size_t>(
                Config::getFloatDef("Reverb:dirGainFactor", EAR::Variant(1.0f)));
            mpegcore::printStatus("Using dirGainFactor {}", dirGainFactor);

            for (int id = 0; id < numAnalysisDirs; id++) {

                std::vector<ReverbEnv::FrequencyGain> directionalFreqGains(numAnalysisBands);

                for (int ib = 0; ib < numAnalysisBands; ib++) {
                    float gain_dB = srirAnalysis->dirGainDB_bnd_dir[ib][id];
                    gain_dB = std::clamp(gain_dB * dirGainFactor, -12.f, 12.f);
                    // dB energy to linear (pressure) magnitude
                    const float gain_lin = std::pow(10.0F, gain_dB / 20.F);
                    directionalFreqGains[ib] = { srirAnalysis->centerFreqs[ib], gain_lin };
                };
                srirDirectionalGainData[id] = { srirAnalysis->sphDirsDeg[id][0], // azim
                    srirAnalysis->sphDirsDeg[id][1], // elev
                    directionalFreqGains };
            }
        }
        // In an AR scene context, predelay is defined as the diffuseOnset
        reverbEnvs.back()->predelay = srirAnalysis->diffuseOnsetSec;
    }
}

void SceneState::createTransform(const EIFPayload::TransformDesc& tf)
{
    anchors.push_back(EAR::SceneObjectFactory<EAR::AnchorObject>::getNewObject(tf.id, tf.position));
    anchorsById[tf.id] = anchors.back().get();
    anchors.back()->setMoved();

    // add to updateable IDs map
    updateable[tf.id] = anchors.back().get();

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(anchors.back().get()));
    }
}

void SceneState::createAnchor(const EIFPayload::AnchorDesc& a)
{
    anchors.push_back(EAR::SceneObjectFactory<EAR::AnchorObject>::getNewObject(a.id));
    anchorsById[a.id] = anchors.back().get();
    anchorsByLSDFRef[a.lsdfRef] = anchors.back().get();
    anchorsById[a.id]->setAutoRescale(a.autoRescale);
    anchorsById[a.id]->setTolerance(a.tolerance);
    anchorsByLSDFRef[a.lsdfRef]->setAutoRescale(a.autoRescale);
    anchorsByLSDFRef[a.lsdfRef]->setTolerance(a.tolerance);

    anchors.back()->setMoved();

    // add to updateable IDs map
    updateable[a.id] = anchors.back().get();

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(anchors.back().get()));
    }
}

void SceneState::createObjectSource(const EIFPayload::ObjectSourceDesc& obj)
{
    objectSources.push_back(EAR::SceneObjectFactory<EAR::PointSource>::getNewObject(
        obj.id, Position(), nullptr, obj.bsId));

    auto objHandle = objectSources.back().get();

    objHandle->setActive(obj.active);

    objHandle->setPosition(obj.position);
    if (obj.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (obj.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[obj.parentTransform->id]);
    } else if (obj.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[obj.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }

    objHandle->setGainDb(obj.gainDb);
    objHandle->setReverbGain(obj.itemReverbGain);
    objHandle->setNoReverb(obj.noReverb);
    objHandle->setDirectiveness(obj.directiveness);
    objHandle->setReferenceDistance(obj.refDistance);
    objHandle->setMoved();

    if (obj.directivity) {
        objHandle->setDirectivity(directivities[obj.directivity->id]);
    }

    if (obj.extent) {
        objHandle->setExtent(geometries[obj.extent->id]);
    }

    if (obj.inputLayout) {
        extentLayouts[objHandle].alignment =
            static_cast<ExtentLayout::Alignment>(obj.inputLayout->alignment);
        for (auto& val : obj.inputLayout->positioning) {
            extentLayouts[objHandle].positioning.push_back(
                static_cast<ExtentLayout::Positioning>(val));
        }
    }

    if (obj.signal) {
        audioStreamProperties[objHandle].id = obj.signal->id;
        audioStreamProperties[objHandle].looping =
            (obj.mode == EIFPayload::AudioElementDesc::Mode::Continuous);
        audioStreamProperties[objHandle].playing =
            (obj.mode == EIFPayload::AudioElementDesc::Mode::Continuous) ? true : obj.play;
        objHandle->setAudioStreamID(obj.signal->id);
    }

    // if the objSource is Event driven, cannot be active at start of scene

    // TODO: This gives differences in possibly early reflections compared to
    // having them active at the start of the scene and immediately disabling them in the effectactivator stage. Not sure why...
    // remove setActive above when fixed.

    // auto isActive = (obj.mode == EIFPayload::AudioElementDesc::Mode::Event) ? false : obj.active;
    // objHandle->setActive(isActive);

    if (obj.granularSignal) {
        GranularSoundProperties[objHandle].id = obj.granularSignal->id;
        GranularSoundProperties[objHandle].channels = obj.granularSignal->numChannels;
    }

    objHandle->setVDLMethod(obj.vdlMethod);

    authoringParams[objHandle] = AParam::None;
    for (auto ap : obj.aparams) {
        authoringParams[objHandle] |= mapAuthoringParameters(ap);
    }

    // add to updateable IDs map
    updateable[obj.id] = objHandle;

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createLSDFLocalSource(
    const LSDF::LSDFParser::LocalSourceDesc& description, const std::string& directivityId)
{
    PointSource* objHandle;

    if (description.listenerVoice) {
        objectSources.push_back(
            EAR::SceneObjectFactory<EAR::ListenerVoiceSource>::getNewObject(description.id));

        auto listenerVoice = dynamic_cast<EAR::ListenerVoiceSource*>(objectSources.back().get());
        listenerVoice->setDirectGainDb(description.directGainDB);
        listenerVoice->setERGainDb(description.erGainDB);
        if (directivityId.empty()) {
            auto defaultListenerVoiceDirectivity =
                std::make_shared<CompressedDirectivity>(LISTENER_VOICE_DEFAULT_DIRECTIVITY_DATA,
                    std::size(LISTENER_VOICE_DEFAULT_DIRECTIVITY_DATA));
            createDirectivity("dir:defaultListenerVoice", defaultListenerVoiceDirectivity);
            listenerVoice->setDirectivity(directivities.at("dir:defaultListenerVoice"));
        } else
            listenerVoice->setDirectivity(directivities.at(directivityId));
    } else {
        objectSources.push_back(
            EAR::SceneObjectFactory<EAR::PointSource>::getNewObject(description.id));
    }

    objHandle = objectSources.back().get();
    objHandle->setActive(description.active);

    objHandle->setPosition(description.position);
    if (description.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }

    objHandle->setGainDb(description.gainDb);
    objHandle->setReverbGain(std::pow(10.f, 0.05f * description.reverbGainDB));
    objHandle->setDirectiveness(description.directiveness);
    objHandle->setReferenceDistance(description.refDistance);
    objHandle->setMoved();

    audioStreamProperties[objHandle].id = description.signal->id;
    audioStreamProperties[objHandle].looping = true;
    audioStreamProperties[objHandle].playing = true;

    objHandle->setVDLMethod(description.vdlMethod);

    authoringParams[objHandle] = AParam::None;
    for (auto ap : description.aparams) {
        authoringParams[objHandle] |= mapLSDFAuthoringParameters(ap);
    }
    // add to updateable IDs map
    updateable[description.id] = objHandle;

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createChannelSource(const EIFPayload::ChannelSourceDesc& obj)
{
    lsLayouts.push_back(EAR::SceneObjectFactory<EAR::LoudspeakerLayout>::getNewObject(
        obj.inputLayout, 1.f, obj.id));
    auto objHandle = lsLayouts.back().get();
    objHandle->setActive(obj.active);
    objHandle->setPosition(obj.position);
    if (obj.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (obj.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[obj.parentTransform->id]);
    } else if (obj.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[obj.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }

    if (obj.granularSignal) {
        GranularSoundProperties[objHandle].id = obj.granularSignal->id;
        GranularSoundProperties[objHandle].channels = obj.granularSignal->numChannels;
    }

    // add to updateable IDs map
    updateable[obj.id] = objHandle;

    int i = 0;
    for (auto src : objHandle->getLoudspeakerPointSources()) {
        src->setLocation(obj.loudspeakers[i].position.location);
        src->setOrientation(obj.loudspeakers[i].position.orientation);
        if (obj.loudspeakers[i].directivity) {
            src->setDirectivity(directivities[obj.loudspeakers[i].directivity->id]);
        }
        src->setDirectiveness(obj.loudspeakers[i].directiveness);
        src->setGainDb(obj.gainDb);
        src->setReverbGain(obj.itemReverbGain);
        src->setNoReverb(obj.noReverb);

        audioStreamProperties[src.get()].id = obj.signal->id;
        audioStreamProperties[src.get()].channels = { static_cast<std::size_t>(i) };
        audioStreamProperties[src.get()].looping =
            (obj.mode == EIFPayload::AudioElementDesc::Mode::Continuous);
        audioStreamProperties[src.get()].playing =
            (obj.mode == EIFPayload::AudioElementDesc::Mode::Continuous) ? true : obj.play;

        src->setVDLMethod(obj.vdlMethod);

        authoringParams[src.get()] = AParam::None;
        for (auto ap : obj.aparams) {
            authoringParams[src.get()] = mapAuthoringParameters(ap);
        }

        teleport[obj.id + "_LS" + std::to_string(i + 1)] = false;

        i++;
    }

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createHOAGroup(const EIFPayload::HoaGroupDesc& group)
{
    EAR::Geometry* region = group.region ? geometries[group.region->id] : nullptr;
    std::vector<EAR::HOASource*> sources;
    sources.reserve(group.hoaSources.size());
    std::vector<MPEGI::InformedSource> informedSources;
    informedSources.reserve(group.informedSources.size());

    auto it = std::find(hoaGroupIds.begin(), hoaGroupIds.end(), group.id);
    EAR_CHECK_REL(it != hoaGroupIds.end(), "HOAGroup with id " + group.id + " not found");
    int groupId = it - hoaGroupIds.begin();

    for (auto const& source : hoaSources) {
        if (source->getGroup() == groupId) {
            sources.push_back(source.get());
        }
    }

    for (auto const& iS : group.informedSources) {
        auto iSrc = InformedSource(iS.position, iS.listenerThresholdPresent,
            iS.hoaSourceThresholdPresent, iS.listenerThreshold, iS.hoaSourceThreshold,
            iS.disableInformedSourceType, iS.disableInformedSourceThreshold, iS.priorityValue);
        informedSources.push_back(iSrc);
    }

    hoaGroups.push_back(std::make_unique<MPEGI::HOAGroup>(region, std::move(sources),
        std::move(group.complexityReductionParameters), group.maxSimulInformedSources,
        std::move(informedSources), group.distanceFactor, group.panningGridOrder,
        group.exteriorProjectionRadius, group.adaptiveExteriorProjectionRadius));
}

void SceneState::createHOASource(const EIFPayload::HoaSourceDesc& obj)
{
    int groupId = -1;
    if (obj.group) {
        auto it = std::find(hoaGroupIds.begin(), hoaGroupIds.end(), obj.group->id);
        groupId = it - hoaGroupIds.begin();
        if (it == hoaGroupIds.end()) {
            hoaGroupIds.push_back(obj.group->id);
        }
    }

    hoaSources.push_back(EAR::SceneObjectFactory<EAR::HOASource>::getNewObject(
        static_cast<EAR::HOASource::Representation>(obj.representation), groupId,
        obj.extentTransform, obj.transitionDistance, obj.id));
    auto objHandle = hoaSources.back().get();

    objHandle->setActive(obj.active);

    objHandle->setPosition(obj.position);
    if (obj.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (obj.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[obj.parentTransform->id]);
    } else if (obj.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[obj.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }

    objHandle->setGainDb(obj.gainDb);
    objHandle->setReverbGain(obj.itemReverbGain);
    objHandle->setNoReverb(obj.noReverb);
    objHandle->setReferenceDistance(obj.refDistance);
    objHandle->setMoved();

    if (obj.extent) {
        objHandle->setExtent(geometries[obj.extent->id]);
    }

    if (obj.signal) {
        audioStreamProperties[objHandle].id = obj.signal->id;
        audioStreamProperties[objHandle].looping =
            (obj.mode == EIFPayload::AudioElementDesc::Mode::Continuous);
        audioStreamProperties[objHandle].playing =
            (obj.mode == EIFPayload::AudioElementDesc::Mode::Continuous) ? true : obj.play;
    }

    if (obj.granularSignal) {
        GranularSoundProperties[objHandle].id = obj.granularSignal->id;
        GranularSoundProperties[objHandle].channels = obj.granularSignal->numChannels;
    }

    objHandle->setVDLMethod(obj.vdlMethod);

    authoringParams[objHandle] = AParam::None;
    for (auto ap : obj.aparams) {
        authoringParams[objHandle] = mapAuthoringParameters(ap);
    }

    // add to updateable IDs map
    updateable[obj.id] = objHandle;

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createDirectivity(
    const std::string& directivityString, std::shared_ptr<EAR::DirectivityModel> directivity)
{
    directivities[directivityString] = directivity;
}

void SceneState::createMaterial(const EIFPayload::AcousticMaterialDesc& mat)
{
    EAR_CHECK_REL(materials.find(mat.id) == materials.end(),
        "Duplicate material ids (probably between EIF and LSDF file).");
    materials[mat.id] = std::make_shared<EAR::Material>(mat.bsId, mat.id);

    materials[mat.id]->setName(mat.id);
    for (auto& freq : mat.frequencies) {
        materials[mat.id]->append(freq.f, freq.r, freq.s, freq.t, freq.c);
    }
}

void SceneState::createLSDFMaterial(const LSDF::LSDFParser::AcousticMaterialDesc& mat)
{
    EAR_CHECK_REL(materials.find(mat.id) == materials.end(),
        "Duplicate material ids (probably between EIF and LSDF file).");
    materials[mat.id] = std::make_shared<EAR::Material>();

    for (auto& freq : mat.frequencies) {
        materials[mat.id]->append(freq.f, freq.r, freq.s, freq.t, freq.c);
    }
}

void SceneState::createBox(const EIFPayload::BoxDesc& box)
{
    boxes.push_back(EAR::SceneObjectFactory<EAR::Box>::getNewObject(box.size, box.id));
    auto objHandle = boxes.back().get();

    if (box.material) {
        objHandle->setMaterial(materials[box.material->id]);
        // everything with a material is visible for diffraction, reflection, etc.
        visibleGeometries.push_back(objHandle);
    }

    objHandle->setPosition(box.position);
    if (box.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (box.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[box.parentTransform->id]);
    } else if (box.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[box.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }
    objHandle->resetMoved();

    // add to updateable IDs map
    updateable[box.id] = objHandle;
    // add to map of all geometry objects
    geometries[box.id] = objHandle;

    if (box.isEarlyReflectionGeometry) {
        earlyReflectionGeometries.push_back(objHandle->getMesh());
        diffractionGeometries.push_back(objHandle->getMesh());
    }

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createSphere(const EIFPayload::SphereDesc& sph)
{
    // TODO: allow ellipsoids instead of only perfect spheres
    // dividing by two as EIF specifies diameters and EAR wants radii
    spheres.push_back(EAR::SceneObjectFactory<EAR::Sphere>::getNewObject(sph.size[0] / 2, sph.id));
    auto objHandle = spheres.back().get();

    if (sph.material) {
        objHandle->setMaterial(materials[sph.material->id]);
        // everything with a material is visible for diffraction, reflection, etc.
        visibleGeometries.push_back(objHandle);
    }

    objHandle->setPosition(sph.position);
    if (sph.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (sph.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[sph.parentTransform->id]);
    } else if (sph.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[sph.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }
    objHandle->resetMoved();

    // add to updateable IDs map
    updateable[sph.id] = objHandle;
    // add to map of all geometry objects
    geometries[sph.id] = objHandle;

    if (sph.isEarlyReflectionGeometry) {
        earlyReflectionGeometries.push_back(objHandle->getMesh());
        diffractionGeometries.push_back(objHandle->getMesh());
    }

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createCylinder(const EIFPayload::CylinderDesc& cyl)
{
    // TODO: allow ellipsoid ground plane instead of only perfect circles
    // dividing by two as EIF specifies diameters and EAR wants radii
    cylinders.push_back(
        EAR::SceneObjectFactory<EAR::Cylinder>::getNewObject(cyl.size[0] / 2, cyl.size[1], cyl.id));
    auto objHandle = cylinders.back().get();

    if (cyl.material) {
        objHandle->setMaterial(materials[cyl.material->id]);
        // everything with a material is visible for diffraction, reflection, etc.
        visibleGeometries.push_back(objHandle);
    }

    objHandle->setPosition(cyl.position);
    if (cyl.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (cyl.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[cyl.parentTransform->id]);
    } else if (cyl.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[cyl.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }
    objHandle->resetMoved();

    // add to updateable IDs map
    updateable[cyl.id] = objHandle;
    // add to map of all geometry objects
    geometries[cyl.id] = objHandle;

    if (cyl.isEarlyReflectionGeometry) {
        earlyReflectionGeometries.push_back(objHandle->getMesh());
    }
    if (cyl.isStatic && cyl.material != nullptr) {
        diffractionGeometries.push_back(objHandle->getMesh());
    }

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createMesh(const EIFPayload::MeshDesc& mesh)
{
    meshes.push_back(EAR::SceneObjectFactory<EAR::Mesh>::getNewObject(mesh.id));
    auto objHandle = meshes.back().get();

    for (auto& v : mesh.vertices) {
        objHandle->addVertex(v.position, v.index);
    }

    bool anyFaceWithMaterial = false;
    for (auto& f : mesh.faces) {
        objHandle->addTriangle(f.v1, f.v2, f.v3,
            (f.material ? materials[f.material->id] : std::shared_ptr<EAR::Material>(nullptr)));
        if (f.material) {
            anyFaceWithMaterial = true;
        }
    }

    // TODO: if only a few faces have material, only those should be visible
    if (anyFaceWithMaterial) {
        // everything with a material is visible for diffraction, reflection, etc.
        visibleGeometries.push_back(objHandle);
    }

    objHandle->setPosition(mesh.position);
    if (mesh.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (mesh.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[mesh.parentTransform->id]);
    } else if (mesh.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[mesh.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }

    objHandle->lock();
    objHandle->resetMoved();

    // add to updateable IDs map
    updateable[mesh.id] = objHandle;
    // add to map of all geometry objects
    geometries[mesh.id] = objHandle;

    if (mesh.isEarlyReflectionGeometry) {
        earlyReflectionGeometries.push_back(objHandle->getMesh());
        diffractionGeometries.push_back(objHandle->getMesh());
    }

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void SceneState::createLSDFMesh(const LSDF::LSDFParser::MeshDesc& mesh)
{
    meshes.push_back(EAR::SceneObjectFactory<EAR::Mesh>::getNewObject(mesh.id));
    auto objHandle = meshes.back().get();

    for (auto& v : mesh.vertices) {
        objHandle->addVertex(v.position, v.index);
    }

    bool anyFaceWithMaterial = false;
    for (auto& f : mesh.faces) {
        objHandle->addTriangle(f.v1, f.v2, f.v3,
            (f.material ? materials[f.material->id] : std::shared_ptr<EAR::Material>(nullptr)));
        if (f.material) {
            anyFaceWithMaterial = true;
        }
    }

    // TODO: if only a few faces have material, only those should be visible
    if (anyFaceWithMaterial) {
        // everything with a material is visible for diffraction, reflection, etc.
        visibleGeometries.push_back(objHandle);
    }

    objHandle->setPosition(mesh.position);
    if (mesh.position.cSpace == CSpace::LCS) {
        objHandle->setCSpace(
            CSpace::LCS, listener.get()); // listener is parent of "user" cspace objects
    } else if (mesh.parentTransform) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[mesh.parentTransform->id]);
    } else if (mesh.parentAnchor) {
        objHandle->setCSpace(CSpace::RCS, anchorsById[mesh.parentAnchor->id]);
    } else {
        objHandle->setCSpace(CSpace::WCS);
    }

    objHandle->lock();

    // add to updateable IDs map
    updateable[mesh.id] = objHandle;
    // add to map of all geometry objects
    geometries[mesh.id] = objHandle;

    // notify observers
    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(objHandle));
    }
}

void createBaseFromDescription(
    AirflowGeneratorBase& object, const EIFPayload::AirflowGeneratorDesc& description)
{
    object.setPosition(description.position);
    object.setGainDb(description.gainDb);
    if (description.hasDistanceAttenuationFactor) {
        object.setDistanceAttenuationFactor(description.distanceAttenuationFactor);
    }

    object.setSpeed(description.speed);
    object.setBsId(description.bsId);
    object.setStringId(description.id);
}

inline void SceneState::createGlobalAirflowGenerator(
    const EIFPayload::GlobalAirflowGeneratorDesc& gen)
{
    std::shared_ptr<EAR::GlobalAirflowGenerator> generator =
        EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(gen.id);

    createBaseFromDescription(*generator, gen);
    updateable[gen.id] = generator.get();

    airflowGenerators.push_back(std::dynamic_pointer_cast<EAR::AirflowGeneratorBase>(generator));
}

inline void SceneState::createPointAirflowGenerator(
    const EIFPayload::PointAirflowGeneratorDesc& gen)
{
    std::shared_ptr<EAR::PointAirflowGenerator> generator =
        EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(gen.id);

    createBaseFromDescription(*generator, gen);

    if (gen.hasLimits) {
        generator->setLimits(gen.aziRange, gen.elevRange);
    }

    updateable[gen.id] = generator.get();

    airflowGenerators.push_back(std::dynamic_pointer_cast<EAR::AirflowGeneratorBase>(generator));
}

inline void SceneState::createConeAirflowGenerator(const EIFPayload::ConeAirflowGeneratorDesc& gen)
{
    std::shared_ptr<EAR::ConeAirflowGenerator> generator =
        EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject(gen.id);

    createBaseFromDescription(*generator, gen);

    generator->baseRadius = gen.baseRadius;
    generator->length = gen.length;
    if (gen.hasTopRadius) {
        generator->setTopRadius(gen.topRadius);
    }
    generator->edgeRollOff = gen.edgeRollOff;

    updateable[gen.id] = generator.get();

    airflowGenerators.push_back(std::dynamic_pointer_cast<EAR::AirflowGeneratorBase>(generator));
}

inline void SceneState::createPlaneAirflowGenerator(
    const EIFPayload::PlaneAirflowGeneratorDesc& gen)
{
    std::shared_ptr<EAR::PlaneAirflowGenerator> generator =
        EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(gen.id);

    createBaseFromDescription(*generator, gen);

    generator->width = gen.width;
    generator->height = gen.height;
    generator->edgeRollOff = gen.edgeRollOff;

    updateable[gen.id] = generator.get();

    airflowGenerators.push_back(std::dynamic_pointer_cast<EAR::AirflowGeneratorBase>(generator));
};

void SceneState::combineEarlyReflectionMesh()
{
    // create new mesh
    combinedEarlyReflectionMesh =
        EAR::SceneObjectFactory<EAR::Mesh>::getNewObject("mesh:_combinedEarlyReflectionMesh");
    auto objHandle = combinedEarlyReflectionMesh.get();

    // concatenate earlyReflectionGeometries into new mesh
    combineMesh(earlyReflectionGeometries, objHandle);
}

void SceneState::combineDiffractionMesh()
{
    // create new mesh
    combinedDiffractionMesh =
        EAR::SceneObjectFactory<EAR::Mesh>::getNewObject("mesh:_combinedDiffractionMesh");
    auto objHandle = combinedDiffractionMesh.get();

    // concatenate diffractionGeometries into new mesh
    combineMesh(diffractionGeometries, objHandle);
}

void SceneState::combineMesh(std::vector<const EAR::Geometry*>& geometries, Mesh* meshHandle)
{
    meshHandle->setCSpace(EAR::CSpace::WCS);
    meshHandle->setLocation({ 0.f, 0.f, 0.f });
    meshHandle->setOrientation({ 0.f, 0.f, 0.f });
    meshHandle->setStatic(true);

    for (const auto& geo : geometries) {
        const auto& mesh_new = geo->getMesh();
        const auto& vertices = meshHandle->getVertices();
        size_t v0 = vertices.size();

        // add vertices
        const auto& vertices_new = mesh_new->getVertices(EAR::Mesh::CSpace::Global);
        for (const auto& v : vertices_new) {
            meshHandle->addVertex(v);
        }

        // add triangles
        const auto& triangles_new = mesh_new->getTriangles();
        for (const auto& t : triangles_new) {
            meshHandle->addTriangle(v0 + t.v1, v0 + t.v2, v0 + t.v3, t.material);
        }
    }
}

const Geometry* getEnvironmentEnclosure(
    const std::vector<std::unique_ptr<MPEGI::ReverbEnv>>& environments,
    const std::optional<std::string>& environmentAcousticEnvironmentId)
{
    if (!environmentAcousticEnvironmentId.has_value()) {
        // No environment reverb id given, so no enclosure can be found
        return nullptr;
    }

    auto findLambda = [&environmentAcousticEnvironmentId](
                          const std::unique_ptr<MPEGI::ReverbEnv>& env) {
        return env->acousticEnvironmentId == environmentAcousticEnvironmentId;
    };

    const auto reverbEnvIter = std::find_if(environments.begin(), environments.end(), findLambda);

    if (reverbEnvIter == environments.end()) {
        // environmentAcousticEnvironmentId not found in environments
        return nullptr;
    } else {
        return (*reverbEnvIter)->enclosure;
    }
}

const EAR::Vector3 forwards(0.f, 0.f, -1.f);
const EAR::Vector3 backwards(0.f, 0.f, 1.f);

// Returns a pair containing position, opposite position & size of the portal extent
// Opposite position has opposite orientation from position.
std::tuple<EAR::Position, EAR::Position, EAR::Vector3, EAR::Geometry::BoundingBox>
getPortalExtentPositionAndSize(MPEGI::Portal* portal, Geometry* extent)
{
    // We need to calculate the centre point of the portal. Origin of a mesh may not be inside
    // of the portal, so we use the global bounding box
    EAR::Geometry::BoundingBox portalBox = extent->getGlobalBoundingBox();
    EAR::Vector3 portalLocation; //centre point of the portal, before axis alignment.
    portalLocation[0] = portalBox[0].low + (portalBox[0].high - portalBox[0].low) / 2;
    portalLocation[1] = portalBox[1].low + (portalBox[1].high - portalBox[1].low) / 2;
    portalLocation[2] = portalBox[2].low + (portalBox[2].high - portalBox[2].low) / 2;

    portal->position = portalLocation;

    // We rotate the portal so that the portalNormal is pointing forwards (forwards),
    // so that we can calculate the dimensions of the portal.
    EAR::Position calculatedPortalPosition;
    EAR::Position calculatedOppositePortalPosition;
    if ((portal->normalVector == forwards)) {
        portalBox = extent->getGlobalBoundingBox();
        calculatedPortalPosition = EAR::Position(portalLocation, EAR::Euler(.0f, .0f, .0f));
        calculatedOppositePortalPosition =
            EAR::Position(portalLocation, EAR::Euler(180.0f, .0f, .0f));
    } else if (portal->normalVector == backwards) {
        portalBox = extent->getGlobalBoundingBox();
        calculatedPortalPosition = EAR::Position(portalLocation, EAR::Euler(180.0f, .0f, .0f));
        calculatedOppositePortalPosition =
            EAR::Position(portalLocation, EAR::Euler(0.0f, .0f, .0f));
    } else {
        EAR::HomogenMatrix rotMatrix = EAR::rotateVectorTo(portal->normalVector, forwards);
        EAR::HomogenMatrix oppositeRotMatrix = EAR::rotateVectorTo(portal->normalVector, backwards);
        EAR::Euler currRotation = extent->getOrientation();

        // base geometry class has no setOrientation functin, so first cast to a temp ptr.
        // rotate the extent so that the portal normal is facing firwards, get the global
        // bounding box, and then rotate back to original orientation.
        if (extent->getType() == EAR::Geometry::Type::Box) {
            auto tempPtr = dynamic_cast<EAR::Box*>(extent);
            tempPtr->setOrientation(currRotation + rotMatrix.getRotation());
            portalBox = extent->getGlobalBoundingBox();
            tempPtr->setOrientation(currRotation);
        } else if (extent->getType() == EAR::Geometry::Type::Mesh) {
            auto tempPtr = dynamic_cast<EAR::Mesh*>(extent);
            tempPtr->setOrientation(currRotation + rotMatrix.getRotation());
            portalBox = extent->getGlobalBoundingBox();
            tempPtr->setOrientation(currRotation);
        } else {
            EAR_FAIL("Only Box and Mesh can be used to specify portals");
        }
        calculatedPortalPosition = EAR::Position(portalLocation, rotMatrix.inverse().getRotation());
        calculatedOppositePortalPosition =
            EAR::Position(portalLocation, oppositeRotMatrix.inverse().getRotation());
    }

    EAR::Vector3 portalSize;
    if (portal->equivalentPortalSize.has_value()) {
        // If specified, equivalentPortalSize will take precedence
        portalSize = portal->equivalentPortalSize.value();
    } else {
        // Otherwise, compute size from axis-aligned portalBox
        portalSize[0] = portalBox[0].high - portalBox[0].low;
        portalSize[1] = portalBox[1].high - portalBox[1].low;
        portalSize[2] = portalBox[2].high - portalBox[2].low;
    }
    return { calculatedPortalPosition, calculatedOppositePortalPosition, portalSize, portalBox };
}

void SceneState::createPortal(MPEGI::Portal* portal)
{
    const int portalSegments =
        3; // TODO: Switch back to 3 segments. When changed, re-enable tests in PortalExtent_Test.cpp - Line15
    const float backfaceOffset = 0.05f;

    EAR::Geometry* extent = nullptr;
    std::unique_ptr<EAR::Box> tempBox = nullptr;
    if (portal->extentId.has_value()) {
        extent = geometries.at(portal->extentId.value());
    } else {
        EAR_CHECK_REL(portal->equivalentPortalSize.has_value(),
            "Portal " + std::to_string(portal->portalId)
                + " has no extent id set AND no explicit portal dimensions. It needs to define "
                  "at least one of these.");

        // Create a Box from the explicit dimensions
        EIFPayload::BoxDesc box;
        box.id = (portal->portalStringId + " equivalent box");

        const EAR::Euler portalRotation =
            EAR::rotateVectorTo(forwards, portal->normalVector).getRotation();

        box.position = { portal->position, portalRotation };
        box.size = portal->equivalentPortalSize.value();
        box.isStatic = true;

        tempBox =
            std::unique_ptr<EAR::Box>(SceneObjectFactory<EAR::Box>::getNewObject(box.size, box.id));
        tempBox->setPosition(box.position);
        tempBox->setCSpace(CSpace::WCS);

        extent = tempBox.get();
        portal->equivalentBoxPointer = std::move(tempBox);
    }

    auto [calculatedPortalPosition, calculatedOppositePortalPosition, portalSize, portalBox] =
        getPortalExtentPositionAndSize(portal, extent);

    portal->bbPortalHeight = portalSize[1];
    portal->bbPortalWidth = portalSize[0];

    float midDepth = (portalBox[2].low + portalBox[2].high) / 2.0f;
    float midHeight = (portalBox[1].low + portalBox[1].high) / 2.0f;
    float midWidth = (portalBox[0].low + portalBox[0].high) / 2.0f;

    EAR::Vector3 tempBoxCenter; //centre point of the portal, before axis alignment.
    tempBoxCenter[0] = portalBox[0].low + (portalBox[0].high - portalBox[0].low) / 2;
    tempBoxCenter[1] = portalBox[1].low + (portalBox[1].high - portalBox[1].low) / 2;
    tempBoxCenter[2] = portalBox[2].low + (portalBox[2].high - portalBox[2].low) / 2;

    EAR::HomogenMatrix rotMatrix = EAR::rotateVectorTo(portal->normalVector, forwards);

    {
        // plane for AE 1
        auto tempMesh = std::unique_ptr<Mesh>(
            SceneObjectFactory<Mesh>::getNewObject("tempMesh", calculatedPortalPosition));
        tempMesh->addVertex(EAR::Vector3(portalBox[0].low, portalBox[1].low, portalBox[2].low)
            - tempBoxCenter); // 0 lowerLeft
        tempMesh->addVertex(EAR::Vector3(portalBox[0].high, portalBox[1].low, portalBox[2].low)
            - tempBoxCenter); // 1 lowerRight
        tempMesh->addVertex(EAR::Vector3(portalBox[0].low, portalBox[1].high, portalBox[2].low)
            - tempBoxCenter); // 2 upperLeft
        tempMesh->addVertex(EAR::Vector3(portalBox[0].high, portalBox[1].high, portalBox[2].low)
            - tempBoxCenter); // 3 upperRight

        // plane for AE 2
        tempMesh->addVertex(EAR::Vector3(portalBox[0].low, portalBox[1].low, portalBox[2].high)
            - tempBoxCenter); // 4 lowerLeft
        tempMesh->addVertex(EAR::Vector3(portalBox[0].high, portalBox[1].low, portalBox[2].high)
            - tempBoxCenter); // 5 lowerRight
        tempMesh->addVertex(EAR::Vector3(portalBox[0].low, portalBox[1].high, portalBox[2].high)
            - tempBoxCenter); // 6 upperLeft
        tempMesh->addVertex(EAR::Vector3(portalBox[0].high, portalBox[1].high, portalBox[2].high)
            - tempBoxCenter); // 7 upperRight

        tempMesh->addVertex(EAR::Vector3(midWidth, midHeight, portalBox[2].low)
            - tempBoxCenter); // 8 center plane 1
        tempMesh->addVertex(EAR::Vector3(midWidth, midHeight, portalBox[2].high)
            - tempBoxCenter); // 9 center plane 2

        tempMesh->lock();
        auto verts = tempMesh->getVertices(EAR::Mesh::CSpace::Global);
        tempMesh.reset();

        MPEGI::IntersectPlane plane;
        plane.planeNormal = portal->normalVector;
        plane.planeCenter = verts[8];
        plane.setBounds(verts[0], verts[1], verts[2], verts[3]);
        portal->portalPlanes.emplace(portal->portalEnvironments[0], plane);

        MPEGI::IntersectPlane plane2;
        plane2.planeNormal = portal->normalVector;
        plane2.planeCenter = verts[9];
        plane2.setBounds(verts[4], verts[5], verts[6], verts[7]);

        portal->portalPlanes.emplace(portal->portalEnvironments[1], plane2);
    }

    auto portalPtr = EAR::SceneObjectFactory<EAR::Portal>::getNewObject(
        portal->portalStringId, calculatedPortalPosition);

    portalPtr->setPortalId(portal->portalId);
    if (portal->materialId.empty()) {
        portalPtr->setMaterial(nullptr);
    } else {
        portalPtr->setMaterial(materials.at(portal->materialId));
    }

    // Create and set external reverb extents
    std::array<std::shared_ptr<PortalReverbExtent>, 2> externalReverbExtents;

    float defaultRevOffset = fmax(0.8f * fmin(portalSize[0], portalSize[1]), 0.3f);
    std::array<float, 2> reverbExtentOffset = { defaultRevOffset, defaultRevOffset };

    const std::array<EAR::Position, 2> portalPositions = { calculatedPortalPosition,
        calculatedOppositePortalPosition };
    for (int AEid = 0; AEid < 2; ++AEid)
    // Create reverb extents for both accoustic environments
    {
        const Geometry* constrainingGeometry =
            getEnvironmentEnclosure(getReverbEnvs(), portal->portalEnvironments[AEid]);

        auto reverbExtent = std::shared_ptr<EAR::PortalReverbExtent>(
            EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(portalSize,
                reverbExtentOffset[AEid], "PortalReverbExtent" + std::to_string(AEid + 1),
                portalPositions[AEid], portalSegments, backfaceOffset, constrainingGeometry));
        reverbExtent->setStatic(true);
        externalReverbExtents[AEid] = reverbExtent;
    }

    portalPtr->setExtentOffset(reverbExtentOffset);
    portalPtr->setPortalDims(portalSize);
    portalPtr->setExternalReverbExtents(externalReverbExtents);

    updateable[portalPtr->getName()] = portalPtr.get();

    for (auto obs : observers) {
        obs->sceneObjectAdded(dynamic_cast<EAR::SceneObject*>(portalPtr.get()));
    }

    // Set pointer to scene state in portal payload
    portal->scenePortal = portalPtr.get();

    portals.push_back(std::move(portalPtr));
}

AParam SceneState::mapAuthoringParameters(
    const EIFPayload::AudioElementDesc::AuthoringParameter param)
{
    typedef EIFPayload::AudioElementDesc::AuthoringParameter EIFAParam;
    switch (param) {
    case (EIFAParam::noDistance):
        return AParam::NoDistanceGain | AParam::NoMediumAttenuation;
    case (EIFAParam::noDoppler):
        return AParam::NoDoppler;
    case (EIFAParam::noDirectionalFocus):
        return AParam::noDirectionalFocus;
    case (EIFAParam::noCulling):
        return AParam::NoCulling;
    default:
        return AParam::None;
    }
}

inline AParam SceneState::mapLSDFAuthoringParameters(
    const LSDF::LSDFParser::AudioElement::AuthoringParameter param)
{
    typedef LSDF::LSDFParser::AudioElement::AuthoringParameter EIFAParam;
    switch (param) {
    case (EIFAParam::noDistance):
        return AParam::NoDistanceGain | AParam::NoMediumAttenuation;
    case (EIFAParam::noDoppler):
        return AParam::NoDoppler;
    default:
        return AParam::None;
    }
}

} // namespace MPEGI
