/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Stream/FileStreamManager.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayload.h>

#include <EAR/AudioFile.h>
#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <algorithm>
#include <cmath>

namespace MPEGI {

FileStreamManager::FileStreamManager(
    const Bitstream& bitstream, float fs, std::size_t B, const LSDF::LSDFParser* lsdf)
    : StreamManager(fs, B)
{
    auto eifPayload = std::dynamic_pointer_cast<const EIFPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));

    EAR::AudioFile inFile;

    for (auto& str : eifPayload->getAudioStreams()) {
        if (str.isLocallyCaptured) {
            if (bitstream.getEnableLocallyCapturedAudio()) {
                buffers[str.id] = std::make_unique<EAR::RealFrame>(1, B);
                overridableStreamBuffers[str.id] = &buffers[str.id]->ch(0);
            }
        } else {
            inFile.load(str.file);

            EAR_CHECK_REL(
                fs == inFile.getSampleRate(), "Sample rates of input files do not match.");

            std::size_t newLength = inFile.getFrame()->length();
            // round to next multiple of B for smooth wrap around when looping
            std::size_t overlap = B - (newLength % B);
            newLength += overlap;

            buffers[str.id] =
                std::make_unique<EAR::RealFrame>(inFile.getFrame()->channels(), newLength);

            for (int i = 0; i < inFile.getFrame()->channels(); i++) {
                buffers[str.id]->ch(i).write(
                    inFile.getFrame()->ch(i).data(), 0, inFile.getFrame()->length());
                if (overlap > 0) {
                    // fill potential overlap with first samples for smooth playback
                    buffers[str.id]->ch(i).write(
                        inFile.getFrame()->ch(i).data(), inFile.getFrame()->length(), overlap);
                }
            }
        }
    }

    // LSDF defined Streams
    if (lsdf && bitstream.getEnableLocallyCapturedAudio()) {
        for (auto& stream : lsdf->getLocalStreams()) {
            if (stream.overrides != "") {
                auto bufferToOverride = overridableStreamBuffers[stream.overrides];
                EAR_CHECK_REL(bufferToOverride,
                    "Failed to override signal " + stream.overrides
                        + ". It does not exist in the bitstream or is not flagged as "
                          "overridable")
                inputChannelToBuffer.emplace_back(stream.localInputChannel, bufferToOverride);

            } else {
                buffers[stream.id] = std::make_unique<EAR::RealFrame>(
                    1, B); //Support single channel input only for now
                inputChannelToBuffer.emplace_back(
                    stream.localInputChannel, &buffers[stream.id]->ch(0));
            }
        }
    }

    // predefine gain ramps for crossfading
    fadeIn = std::make_unique<EAR::RealBuffer>(B);
    fadeOut = std::make_unique<EAR::RealBuffer>(B);
    temp = std::make_unique<EAR::RealBuffer>(B);
    const float step = static_cast<float>(M_PI) / (2.0f * static_cast<float>(B));

    for (int i = 0; i < B; i++) {
        (*fadeIn)[i] = std::sin((float)i * step) * std::sin((float)i * step);
        (*fadeOut)[i] = std::cos((float)i * step) * std::cos((float)i * step);
    }
}

FileStreamManager::~FileStreamManager()
{
    for (auto acc : accessors) {
        acc->unregister();
    }
}

std::shared_ptr<StreamAccessBuffer> FileStreamManager::createAccessBuffer(
    const std::string& streamId, double t)
{

    EAR_CHECK_REL(buffers.find(streamId) != buffers.end(),
        "Invalid stream id provided to the StreamAccessBuffer");
    auto acc = std::shared_ptr<StreamAccessBuffer>(new StreamAccessBuffer(this, streamId, t));
    accessors.push_back(acc.get());
    return acc;
}

std::shared_ptr<StreamAccessBuffer> FileStreamManager::createAccessBuffer(
    const SceneState::AudioStream& stream)
{

    EAR_CHECK_REL(buffers.find(stream.id) != buffers.end(),
        "Invalid stream id provided to the StreamAccessBuffer");
    auto acc = std::shared_ptr<StreamAccessBuffer>(new StreamAccessBuffer(
        this, stream.id, stream.seekTime, stream.playing, stream.looping, stream.channels));
    accessors.push_back(acc.get());
    return acc;
}

void FileStreamManager::removeAccessBuffer(StreamAccessBuffer* buf)
{
    auto it = std::find(accessors.begin(), accessors.end(), buf);

    if (it != accessors.end()) {
        accessors.erase(it);
    }
}

void FileStreamManager::nextBlock()
{
    // TODO: Fade In/Out when properties change (with flag)

    for (auto acc : accessors) {

        std::size_t crc = acc->rc.load();

        if (acc->playing.load()) {
            EAR_CHECK(buffers.find(acc->streamId) != buffers.end(),
                "Invalid stream id provided to the StreamAccessBuffer");

            std::size_t prc;
            bool crossfade = false;

            // execute seeking if required (also 0 if stream id changed)
            double nt = acc->seekToTime.load();

            if (nt >= 0) {
                crossfade = true;

                // calculate next rc without seeking for crossfading
                prc = crc + B;

                if (acc->prevStreamId == "") { // no crossfade from silence
                    crossfade = false;
                    acc->prevStreamId = acc->streamId;
                }

                if (prc >= buffers[acc->prevStreamId]->length()) {
                    crossfade = false;
                }

                if (acc->playStateChange.load()) {
                    crossfade = false;
                    acc->playStateChange.store(false);
                }

                crc = nt * fs;
                acc->seekToTime.store(-1);
            } else if (acc->prevStreamId != acc->streamId) {
                // change stream without seeking
                crossfade = true;
                crc += B;
                prc = crc;

                if (prc >= buffers[acc->prevStreamId]->length()) {
                    crc %= buffers[acc->prevStreamId]->length();
                }
            } else { // nothing changed, just advance read cursor
                crc += B;
            }

            // wrap around or stop
            if (crc >= buffers[acc->streamId]->length() - B) {
                // reached the end of the current audio stream
                if (!acc->looping.load()) {
                    acc->stop();
                    acc->getFrame().zero();
                    continue;
                } else {
                    // no crossfade needed as this is a sample-perfect loop
                    // of the input stream
                    crc %= buffers[acc->streamId]->length();
                }
            }

            // copy new samples
            int k = 0; // index for access buffer channels (not increasing when one of i is skipped)

            for (std::size_t i : acc->channels) {
                acc->getFrame()[k].write(buffers[acc->streamId]->ch(i).data() + crc, 0, B);
                //TODO there could be a way to enhance live stream processing in this function
                //since the buffers are of the size of B

                if (crossfade) {
                    temp->zero();
                    temp->write(buffers[acc->prevStreamId]->ch(i).data() + prc, 0, B);

                    auto ptrA = acc->getFrame()[k].data();
                    auto ptrB = temp->data();
                    auto finPtr = fadeIn->data();
                    auto foutPtr = fadeOut->data();

                    for (int s = 0; s < B; s++) {
                        ptrA[s] = ptrB[s] * foutPtr[s] + ptrA[s] * finPtr[s];
                    }

                    acc->prevStreamId = acc->streamId;
                }

                k++;
            }
        } else {
            // not playing, so we'll clean the temp buffer
            acc->getFrame().zero(); // TODO: only do this once after stopping?

            // ... and advance the read cursor to emulate continuous playback of the streams
            // (only if not in event mode)
            if (acc->looping.load()) {
                crc += B;
                if (crc >= buffers[acc->streamId]->length() - B) {
                    // if statement is cheaper than modulo
                    crc %= buffers[acc->streamId]->length();
                }
            }
        }

        acc->rc.store(crc);
    }
}

std::size_t FileStreamManager::maxLength() const
{
    std::size_t max = 0;

    for (auto& buf : buffers) {
        if (buf.second->length() > max) {
            max = buf.second->length();
        }
    }

    return max;
}

std::size_t FileStreamManager::getNbrOfChannels(const std::string& id) const
{
    return buffers.at(id)->channels();
}

void FileStreamManager::resetTime()
{
    for (auto& acc : accessors) {
        acc->seekTo(0.0);
    }
}

void FileStreamManager::ingest(const float* const* input, int numChannels)
{
    for (auto& [inputChannel, buffer] : inputChannelToBuffer) {
        if (inputChannel < numChannels) {
            buffer->write(input[inputChannel], 0, B);
        }
    }
}

} // namespace MPEGI
