/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/Stream/VSTStreamManager.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayload.h>

#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <algorithm>

namespace MPEGI {

VSTStreamManager::VSTStreamManager(const Bitstream& bitstream, float fs, std::size_t B)
    : StreamManager(fs, B)
{
    auto eifPayload = std::dynamic_pointer_cast<const EIFPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));

    highestChannel = 0;

    for (auto& str : eifPayload->getAudioStreams()) {
        // create one Frame for every AudioStream in the EIF and remember their aepInputChannel inputs
        auto& channels = str.aepInputChannels;
        buffers[str.id] = std::make_unique<EAR::RealFrame>(channels.size(), B);
        for (int i = 0; i < channels.size(); i++) {
            if (channels[i] + 1 > highestChannel) {
                highestChannel = channels[i] + 1;
            }
            inputChannelToBuffer.emplace_back(channels[i], &buffers[str.id]->ch(i));
        }
    }
}

VSTStreamManager::~VSTStreamManager()
{
    for (auto acc : accessors) {
        acc->unregister();
    }
}

std::shared_ptr<StreamAccessBuffer> VSTStreamManager::createAccessBuffer(
    const std::string& streamId, double t)
{

    EAR_CHECK_REL(buffers.find(streamId) != buffers.end(),
        "Invalid stream id provided to the StreamAccessBuffer");
    auto acc = std::shared_ptr<StreamAccessBuffer>(new StreamAccessBuffer(this, streamId, t));
    accessors.push_back(acc.get());
    return acc;
}

std::shared_ptr<StreamAccessBuffer> VSTStreamManager::createAccessBuffer(
    const SceneState::AudioStream& stream)
{

    EAR_CHECK_REL(buffers.find(stream.id) != buffers.end(),
        "Invalid stream id provided to the StreamAccessBuffer");
    auto acc = std::shared_ptr<StreamAccessBuffer>(new StreamAccessBuffer(
        this, stream.id, stream.seekTime, stream.playing, stream.looping, stream.channels));
    accessors.push_back(acc.get());
    return acc;
}

void VSTStreamManager::removeAccessBuffer(StreamAccessBuffer* buf)
{
    auto it = std::find(accessors.begin(), accessors.end(), buf);

    if (it != accessors.end()) {
        accessors.erase(it);
    }
}

void VSTStreamManager::nextBlock()
{
    // TODO: Fade In/Out when properties change (with flag)

    for (auto acc : accessors) {
        if (acc->playing.load()) {
            EAR_CHECK(buffers.find(acc->streamId) != buffers.end(),
                "Invalid stream id provided to the StreamAccessBuffer");

            int ch = 0;
            for (std::size_t i : acc->channels) {
                acc->getFrame()[ch].copyFrom(buffers[acc->streamId]->ch(i));
                ch++;
            }

        } else {
            // not playing, so we'll clean the temp buffer
            acc->getFrame().zero(); // TODO: only do this once after stopping?
        }
    }
}

std::size_t VSTStreamManager::getNbrOfChannels(const std::string& id) const
{
    return buffers.at(id)->channels();
}

std::size_t VSTStreamManager::getNbrOfChannels() const
{
    return highestChannel;
}

void VSTStreamManager::ingest(const float* const* input)
{
    for (auto& pair : inputChannelToBuffer) {
        pair.second->write(input[pair.first], 0, B);
    }
}

} // namespace MPEGI
