/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef LSDF_EXCEPTION
#define LSDF_EXCEPTION

#include <exception>
#include <ostream>
#include <string>

namespace LSDF {

//! Custom LSDF exception
/*!
 * Inspired by ITA's VAException (virtualacoustics.org)
 */
class Exception : std::exception
{
public:
    //! Constructor for empty initialization
    Exception();

    //! Constructor with an error message
    /*!
     * \param errmsg 	Error message (defaults to empty string)
     */
    Exception(const std::string& errmsg = "");

    //! Destructor
    virtual ~Exception();

    //! Return the error message
    std::string getErrorMessage() const;

    //! Return the error message
    std::string toString() const;

    //! Return the error message
    virtual const char* what() const throw();

protected:
    std::string errorMessage;
};

//! Stream output operator for exceptions
std::ostream& operator<<(std::ostream& os, const LSDF::Exception& ex);

/*! \file Exception.h
    \brief Error Macros
*/

//! Macro for throwing an error (active in Release builds)
/*!
 * \param ERRORMSG      Error message to throw
 */
#define LSDF_FAIL(ERRORMSG)                                                                        \
    {                                                                                              \
        throw LSDF::Exception((ERRORMSG));                                                         \
    }

//! Macro for throwing an error when a condition is not matched (active in Release build)
/*!
 * \param COND          Condition to check
 * \param ERRORMSG      Error message to throw when condition fails
 */
#define LSDF_CHECK_REL(COND, ERRORMSG)                                                             \
    {                                                                                              \
        if (!(COND)) {                                                                             \
            throw LSDF::Exception((ERRORMSG));                                                     \
        }                                                                                          \
    }

#ifndef NDEBUG // debug mode active

//! Macro for throwing an error when a condition is not matched
/*!
 * BEWARE: Checks are disabled in Release builds to avoid performance drawbacks in highly critical situations
 *
 * \param COND 		  	Condition to check
 * \param ERRORMSG  	Error message to throw when condition fails
 */
#define LSDF_CHECK(COND, ERRORMSG)                                                                 \
    {                                                                                              \
        if (!(COND)) {                                                                             \
            throw LSDF::Exception((ERRORMSG));                                                     \
        }                                                                                          \
    }

#else // debug mode NOT active
#define LSDF_CHECK(COND, ERRORMSG)
#endif
} // namespace LSDF

#endif /* LSDF_EXCEPTION */
