/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef LSDF_READER
#define LSDF_READER

#ifdef LSDF_USE_EAR
// use EAR primitive types when available
#include <EAR/BaseTypes.h>
#include <EAR/SceneObject.h>
#include <EAR/Variant.h>

using namespace EAR;

#else // LSDF_USE_EAR
// use EIF primitive types as a fallback
#include <LSDF/Types.h>
#include <LSDF/Variant.h>

#endif // LSDF_USE_EAR

#include <LSDF/Exception.h>
#include <LSDF/XMLReader.h>

#include <filesystem>
#include <map>
#include <memory>
#include <optional>
#include <set>
#include <string>
#include <utility>
#include <vector>

#if defined(WIN32)
#define LSDF_DIRECTORY_SEPARATOR '\\'
#else
#define LSDF_DIRECTORY_SEPARATOR '/'
#endif

namespace LSDF {

//! Provide a accessible C++ representation of an LSDF file
/*!
 *  This class will parse a valid LSDF file that holds all of the relevant scene information in the format defined in N18807.
 */
class LSDFParser
{
public:
    // fwd declares
    struct AcousticMaterialDesc;
    struct AnchorDesc;
    struct AudioStreamDesc;
    struct FileStreamDesc;
    struct LocalStreamDesc;
    struct HoaSourceDesc;
    struct ListenerProximityConditionDesc;
    struct TransformDesc;
    struct SourceDirectivityDesc;
    struct VirtualLoudspeakerLayoutDesc;

    enum class SceneType
    {
        VR = 0,
        AR
    };

    //! Superclass of every description element
    struct Description
    {
        std::string id = ""; //!< globally unique string identifier
    };

    //! Representation of an audio element (ObjectSources, HOASources, and ChannelSources)
    struct AudioElement
    {
        //! Playback mode
        enum Mode
        {
            Continuous, //!< Will play in a loop from the start of the scene
            Event //!< Will play once when play is set to true. After playback is finished, play is set to false automatically
        };

        //! Configure behavior of the renderer for the individual AudioElement
        enum AuthoringParameter
        {
            noDoppler,
            noDistance
        };

        //! VDL interpolation method for delay lines defined for this source
        VDLMethod vdlMethod;

        //! If false, do not render this AudioElement
        bool active = true;

        //! Global gain in dB applied to the playback through this audio element
        float gainDb = 0.f;

        //! The distance in m at which the distance attenuation/gain is 0dB
        float refDistance = 1.f;

        //! Authoring parameters for this AudioElement (see enum AuthoringParameter)
        std::vector<AuthoringParameter> aparams;

        //! Playback mode (see enum Mode)
        Mode mode = Mode::Continuous;

        //! Play flag (has not effect in Mode::Continuous)
        bool play = false;
    };

    struct FileAudioElement : AudioElement
    {
        //! A pointer to the FileAudioStream description used for this audio element
        FileStreamDesc* signal;

        //! Parent transform if the entity is nested in a Transform node (nullptr if not nested)
        TransformDesc* parentTransform = nullptr;

        //! Parent anchor if the entity is nested in an Anchor node (nullptr if not nested)
        AnchorDesc* parentAnchor = nullptr;
    };
    struct LocalAudioElement : AudioElement
    {
        LocalStreamDesc* signal;
    };

    //! Representation of a geometry element (Primitives and Meshes)
    struct GeometryDesc : Description
    {
        //! Type of the geometry
        enum class GeometryType
        {
            Box = 1,
            Sphere,
            Cylinder,
            Mesh
        };

        //! Position (location, orientation and coordinate space) of the geometry (i.e. its center point)
        Position position;

        //! Parent transform if the entity is nested in a Transform node (nullptr if not nested)
        TransformDesc* parentTransform = nullptr;

        //! Parent anchor if the entity is nested in an Anchor node (nullptr if not nested)
        AnchorDesc* parentAnchor = nullptr;

        //! Type (see enum GeometryType)
        GeometryType geometryType;

        GeometryDesc(GeometryType type)
            : geometryType(type)
        {
            position.cSpace = CSpace::RCS; // default coordinate space is relative
        };
    };

    // Section 2.2
    //! Representation of an Update
    struct UpdateDesc : Description
    {

        //! Representation of a single Modification (a pair of ID and *one* modified parameter)
        struct Modification
        {

            //! Descriptor of the placeholder position in a dynamic update
            enum class Placeholder
            {
                PositionX = 0,
                PositionY,
                PositionZ,
                OrientationY,
                OrientationP,
                OrientationR,
                GainDb
            };

            std::string id; //!< ID of a modifiable entity

            float duration = 0.f; //!< Duration of interpolation in seconds (if interpolate = true)

            std::string
                attribute; //!< The name of the attribute being modified, e.g. "position" or "orientation"

            Variant
                targetValue; //!< The new value of the modification (may be empty for a dynamic update with placeholders)

            std::map<int, Placeholder>
                placeholders; //!< Placeholders ("parameter indicators") for dynamic updates (map key is the placeholder number)
        };

        //! Type of the update, inferred from the given attributes
        enum class Type
        {
            Undefined = -1, //!< Default value until specified by parser
            Timed = 0, //!< Update executed at specific time (required attributes: time)
            Conditional, //!< Update executed when condition is met (required attributes: condition, fireOn, delay)
            Triggered, //!< Update executed by external trigger, possibly with placeholder values for dynamic update (required attributes: id, index)
        };

        Type type = Type::Undefined; //!< Update type as described in enum UpdateDesc::Type

        int index =
            -1; //!< Globally unique identifier of this Update (required for triggered and dynamic updates, -1 indicates that id is not present)

        float time =
            -1; //!< Update execution time in seconds (negative value if not a timed update)

        ListenerProximityConditionDesc*
            condition; //!< Condition (nullptr if not a conditional update)

        bool fireOn = true; //!< Condition flank negation (only for conditional update)

        float delay = 0.f; //!< Postpone update after trigger in seconds

        std::vector<Modification>
            modifications; //!< List of all atomic modifications within this update
    };

    // Section 3.2
    //! Representation of an ObjectSource
    struct ObjectSourceDesc : Description, FileAudioElement
    {
        Position position; //!< The position of this source in space

        //! The InputLayout gives hints for the positioning of the audio stream channels w.r.t. spatial extent of the source
        struct InputLayout
        {
            //! The alignment specifies the reference frame of the positioning
            enum class Alignment
            {
                User = 0, //!< Positionings are relative to the user's viewing direction
                Object //!< Positionings are fixed to the position/orientation of the ObjectSource
            };

            //! Positioning Mnemonics (see EIF specification for further details)
            enum class Positioning
            {
                TL = 0, //!< top left
                T, //!< top
                TR, //!< top right
                L, //!< left
                C, //!< center
                R, //!< right
                BL, //!< bottom left
                B, //!< bottom
                BR //!< bottom right
            };

            Alignment alignment; //!< Anchoring mode (see InputLayout::Alignment)
            std::vector<Positioning>
                positioning; //!< Positioning of the channels from the audio stream
        };

        //! ObjectSource input layout (see ObjectSourceDesc::InputLayout)
        /*!
         * May be nullptr if no input layout is specified.
         */
        std::unique_ptr<InputLayout> inputLayout;

        //! Pointer to the Geometry that describes the spatial extent of this source (nullptr if no extent)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(extent->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(extent); }
         */
        GeometryDesc* extent;

        SourceDirectivityDesc*
            directivity; //!< Pointer to the SourceDirectivity (nullptr if no directivity)

        float directiveness =
            1.f; //!< Control the intensity of the directivity over all frequency bands

        ObjectSourceDesc()
        {
            position.cSpace = CSpace::RCS; // default coordinate space is relative
        };
    };

    //! Representation of a LocalSource
    struct LocalSourceDesc : Description, LocalAudioElement
    {

        Position position; //!< The position of this source in space

        SourceDirectivityDesc*
            directivity; //!< Pointer to the SourceDirectivity (nullptr if no directivity)

        float directiveness =
            1.f; //!< Control the intensity of the directivity over all frequency bands

        bool listenerVoice = false;

        float erGainDB = -13.0f;
        float directGainDB = -17.0f;
        float reverbGainDB = -6.0f;

        LocalSourceDesc()
        {
            //Fixed parameters these are not meant to be author
            play = true;
            mode = Mode::Continuous;
            //Default parameters, possible to author but with defined default values
            active = true;
            directiveness = 1.0f;
            vdlMethod = VDLMethod::spline;
        };
    };

    //! Representation of a HOA Group
    struct HoaGroupDesc : Description
    {
        //! Pointer to the Geometry that describes the region of validity (nullptr if valid everywhere)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(region->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(region); }
         */
        GeometryDesc* region;

        std::vector<HoaSourceDesc*>
            hoaSources; //!< List of HOA Source Representations that belong to this group
        std::vector<ObjectSourceDesc*>
            coSources; //!< List of ObjectSources sampled simultaneously with the HOA Sources
    };

    //! Representation of a Higher Order Ambisonics source
    struct HoaSourceDesc : Description, FileAudioElement
    {

        Position position; //!< The position of this source in space

        //! Pointer to corresponding HOAGroup (may be nullptr)
        HoaGroupDesc* group;

        //! Pointer to the Geometry that describes the spatial extent of this source (nullptr if no extent)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(extent->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(extent); }
         */
        GeometryDesc* extent;

        bool extentTransform =
            true; //!< Turns on/off the external rendering of an interior source or vice versa

        int representation; //!< Representation type; 1 => Exterior, otherwise interior

        float transitionDistance =
            0.f; //!< Determines the transition radius between external and internal representation [m]

        HoaSourceDesc()
        {
            position.cSpace = CSpace::RCS; // default coordinate space is relative
        }
    };

    //! Representation of a channel source
    struct ChannelSourceDesc : Description, FileAudioElement
    {

        //! Representation of a loudspeaker within a Channel Source
        struct LoudspeakerDesc
        {

            int channelIndex; //!< Index of channel from AudioStream that is played through this Loudspeaker

            Position position; //!< The position, always in RCS relative to the parent ChannelSource

            SourceDirectivityDesc*
                directivity; //!< Pointer to the SourceDirectivity (nullptr if no directivity)

            float directiveness =
                1.f; //!< Control the intensity of the directivity over all frequency bands
        };

        Position position; //!< The channel source position (LS positions are relative to this)

        int inputLayout; //!< CICP channel layout number as specified in ISO/IEC 23091-3:2018 CICP-Audio

        std::vector<LoudspeakerDesc>
            loudspeakers; //!< List of virtual loudspeakers (see struct LoudspeakerDesc)

        //! Reference point (only used in LoudspeakerSetup)
        struct ReferencePoint
        {
            //! Position relative to the channel source position
            Position position;
            //! Is the loudspaker system calibrated with regards to this reference point?
            bool calibrated = false;
        };
        std::optional<ReferencePoint> referencePoint;

        //! Loudspeaker directivity compensation (only used in LoudspeakerSetup)
        struct DirectivityCompensation
        {
            float eqFreq = 9000.0f;
            float eqGainDb = 6.0f;
            float eqAngleStart = 20.0f;
            float eqAngleRange = 30.0f;
        };
        std::optional<DirectivityCompensation> directivityCompensation;

        //! Loudspeaker level compensation (only used in LoudspeakerSetup)
        struct LevelCompensation
        {
            float slope1 = 8.0f;
            float slope2 = 0.0f;
            float criticalDistance = 2.7f;
            float beta = 5.0f;
        };
        std::optional<LevelCompensation> levelCompensation;
    };

    // Section 3.3
    //! Representation of a Transform node
    struct TransformDesc : Description
    {
        //! Position (location, orientation and coordinate space) of the transform
        Position position;
    };

    //! Representation of an Anchor node
    struct AnchorDesc : Description
    {
        //! Identifier of the corresponding LSDF anchor
        std::string lsdf_ref;
    };

    //! Representation of an ARAnchor node (only in LSDF mode)
    struct ARAnchorDesc : Description
    {
        //! Position (location, orientation and coordinate space) of the ARAnchor
        Position position;
    };

    // Section 3.4
    //! Representation of a Listener Proximity Condition
    struct ListenerProximityConditionDesc : Description
    {

        //! Pointer to the Geometry that describes the region where the condition is triggered
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(region->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(region); }
         */
        GeometryDesc* region;
    };

    // Section 3.5
    //! Representation of a file audio stream
    struct FileStreamDesc : Description
    {
        std::vector<int>
            aepInputChannels; //! Input channel indices where the audio for this stream is provided in the MPEG-I 6DoF AEP
        std::string filename; //!< File path to the audio file (relative to EIF file location)
    };
    //! Representation of a locally captured audio stream
    struct LocalStreamDesc : Description
    {
        int localInputChannel; //!< Input channel of locally capture stream
        std::string overrides; //!< Input channel of locally capture stream
    };

    // Section 3.6
    //! Representation of a directivity
    struct SourceDirectivityDesc : Description
    {

        //! File path to the directivity file relative to the EIF file
        std::string filename;
    };

    // Section 3.7
    //! Representation of a Box primitive geometry
    struct BoxDesc : GeometryDesc
    {

        Position position; //!< The global (WCS) position of the center point

        Vector3 size; //!< Spatial extent (in x, y, z direction)

        //! Pointer to the AcousticMaterial assigned to this geometry (may be nullptr)
        AcousticMaterialDesc* material;

        //! Default constructor
        BoxDesc()
            : GeometryDesc(GeometryType::Box){};
    };

    //! Representation of a Sphere primitive geometry
    /*!
     *  Different spatial dimensions (x, y, z) warp the sphere into an ellipsoid.
     */
    struct SphereDesc : GeometryDesc
    {

        //! Spatial extent (radius in x, y, z direction)
        Vector3 size;

        //! Pointer to the AcousticMaterial assigned to this geometry (may be nullptr)
        AcousticMaterialDesc* material;

        //! Default constructor
        SphereDesc()
            : GeometryDesc(GeometryType::Sphere){};
    };

    //! Representation of a Cylinder primitive geometry
    /*!
     *  The rotation axis is y.
     *  The size in y-direction specifies the height, whereas the size in x- and z-direction defines the cross section.
     *  Equal size for the x- and z-direction result in circular cross section, while different spatial dimensions for x and z warps the cross section to an ellipse.
     */
    struct CylinderDesc : GeometryDesc
    {

        //! Spatial extent (radius in x, z direction, height in y direction)
        Vector3 size;

        //! Pointer to the AcousticMaterial assigned to this geometry (may be nullptr)
        AcousticMaterialDesc* material;

        //! Default constructor
        CylinderDesc()
            : GeometryDesc(GeometryType::Cylinder){};
    };

    //! Representation of a Mesh geometry
    struct MeshDesc : GeometryDesc
    {

        //! Representation of a single Vertex
        struct VertexDesc
        {

            //! Unique index >= 0
            int index = -1;

            //! Local coordinates of the vertex
            Vector3 position;
        };

        //! Representation of a Mesh face
        /*!
         *  A face references 3 vertices. The order determines which side is front/back.
         */
        struct FaceDesc
        {

            //! Unique index >= 0
            int index;

            //! The 3 indexes of the previously declared Vertices
            int v1, v2, v3;

            //! Pointer to the AcousticMaterial assigned to this geometry (may be nullptr)
            AcousticMaterialDesc* material;
        };

        //! List of all Vertices of this Mesh
        std::vector<VertexDesc> vertices;

        //! List of all Faces of this Mesh
        std::vector<FaceDesc> faces;

        //! Default constructor
        MeshDesc()
            : GeometryDesc(GeometryType::Mesh){};
    };

    // Section 3.8
    //! Representation of an Acoustic Material
    struct AcousticMaterialDesc : Description
    {

        //! Representation of the material coefficients at a specific frequency
        /*!
         *  Note that r + s + t + c must be <= 1
         */
        struct FrequencyDesc
        {
            //! Frequency [Hz]
            float f;

            //! Specular reflection coefficient, range 0-1
            float r = 0;

            //! Diffuse scattering coefficient, range 0-1
            float s = 0;

            //! Transmission coefficient, range 0-1
            float t = 0;

            //! Coupling coefficient, range 0-1
            float c = 0;
        };

        //! List containing all coefficients across frequencies
        std::vector<FrequencyDesc> frequencies;
    };

    // Section 3.9
    //! Representation of an Acoustic Environment
    struct AcousticEnvironmentDesc : Description
    {

        //! Representation of acoustic parameters in a specific point within an acoustic environment
        struct AcousticParametersDesc : Description
        {

            //! Representation of all acoustic parameters at a specific frequency
            struct ValueDesc
            {

                //! Frequency [Hz]
                float f;

                //! RT60 [s], the time it takes for reverb at this frequency to decay by 60dB
                float rt60;

                //! Reverberant-to-direct ratio
                /*!
                 * Only valid with non-zero positive pre-delay.
                 */
                std::optional<float> rdr;
            };

            //! Diffuse onset of the reverb [s]. Defaults to geometrix approximation. Treated as predelay depending on context.
            std::optional<float> diffuseOnset;

            //! Holds Config::PredelayOffset if specified [sec]
            float predelayOffset = 0.f;

            //! ComplexityLevel of the reverb
            std::optional<int> complexityLevel = 0;

            //! List containing all parameters across frequencies
            std::vector<ValueDesc> values;

            //! The position in space at which the acoustic environment is described by the parameters
            Position position;

            //! Default constructor
            AcousticParametersDesc()
            {
                position.cSpace = CSpace::RCS; // default coordinate space is relative
            }
        };

        //! Representation of acoustic parameters in a specific point within an acoustic environment
        struct RoomImpulseResponseDesc : Description
        {

            std::string path;
            std::string format;

            //! The positions (and orientations) in space of the source and receiver, relative to the AudioScene orgin
            std::optional<Position> micPosition;
            std::optional<Position> sourcePosition;

            //! Source directivity: 0 is omni, 1 is nominal loudspeaker directivity
            int directivity = 1;

            //! Holds the Config::PredelayOffset if specified [sec]
            float predelayOffset = 0.f;

            //! ComplexityLevel of the reverb
            std::optional<int> complexityLevel = 0;

            //! Default constructor
            RoomImpulseResponseDesc()
            {
                path = "";
                format = "ACN-SN3D";
            }
        };

        //! Pointer to the Geometry that describes the region of validity (nullptr if valid everywhere)
        /*!
         *  Can be casted to the individual geometry type,
         *  e.g. if(region->geometryType == GeometryType::Mesh) { auto mesh = dynamic_cast<Mesh*>(region); }
         */
        GeometryDesc* region = nullptr;

        //! List of all AcousticParameterDesc instances in the region of validity
        /*!
         * See struct AcousticParametersDesc for details.
         */
        std::vector<AcousticParametersDesc> acousticParameters;

        //! List of all RoomImpulseResponseDesc instances in the region of validity
        /*!
         * See struct RoomImpulseResponseDesc for details.
         */
        std::vector<RoomImpulseResponseDesc> rirParameters;

        bool usingDefaultDirections = true;
    };

    //! Constructor with EIF file
    /*!
     * This may throw an exception when the given EIF file is not valid.
     *
     * \param scenefile         path to the scene.xml file
     * \param predelayOffset    offset applied to pre-delay parameter on the reverb stage
     * \param checkFiles        whether or not non-existent audio or directivity files cause an exception
     *                          (default true)
     * \param isLSDF            if true, the provided scenefile will be parsed as an LSDF file – see M53586 for details
     *                          (default false)
     */
    LSDFParser(const std::string& scenefile, float predelayOffset = 0.f, bool checkFiles = true,
        bool isLSDF = false);

    //! Returns the id of the AudioScene node
    inline const std::string& getAudioSceneId() const { return audioSceneId; }

    //! Returns the duration of the AudioScene
    inline float getAudioSceneDuration() const { return audioSceneDuration; }

    //! Returns the type of the AudioScene (VR or AR)
    inline SceneType getAudioSceneType() const { return audioSceneType; }

    //! Returns all FileStreams in the scene
    inline const std::vector<FileStreamDesc>& getFileStreams() const { return fileStreams; };

    //! Returns all LocalStreams in the scene
    inline const std::vector<LocalStreamDesc>& getLocalStreams() const { return localStreams; };

    //! Returns all SourceDirectivies in the scene
    inline const std::vector<SourceDirectivityDesc>& getSourceDirectivities() const
    {
        return directivities;
    };

    //! Returns all ObjectSources in the scene
    inline const std::vector<ObjectSourceDesc>& getObjectSources() const { return objectSources; };

    //! Returns all the LocalSources in the LSDF
    inline const std::vector<LocalSourceDesc>& getLocalSources() const { return localSources; };

    //! Returns all HOAGroups in the scene
    inline const std::vector<HoaGroupDesc>& getHoaGroups() const { return hoaGroups; }

    //! Returns all HOASources in the scene
    inline const std::vector<HoaSourceDesc>& getHoaSources() const { return hoaSources; };

    //! Returns all ChannelSources in the scene
    inline const std::vector<ChannelSourceDesc>& getChannelSources() const
    {
        return channelSources;
    };

    //! Returns all LoudspeakerSetup nodes in the scene (only in LSDF mode)
    inline const std::vector<ChannelSourceDesc>& getLoudspeakerSetups() const
    {
        return loudspeakerSetups;
    };

    //! Returns all AcousticMaterials in the scene
    inline const std::vector<AcousticMaterialDesc>& getAcousticMaterials() const
    {
        return materials;
    };

    //! Returns all Transform nodes in the scene
    inline const std::vector<TransformDesc>& getTransforms() const { return transforms; };

    //! Returns all Anchor nodes in the scene
    inline const std::vector<AnchorDesc>& getAnchors() const { return anchors; };

    //! Returns all ARAnchor nodes in the scene (only in LSDF mode)
    inline const std::vector<ARAnchorDesc>& getARAnchors() const { return arAnchors; };

    //! Returns all Boxes in the scene
    inline const std::vector<BoxDesc>& getBoxes() const { return boxes; };

    //! Returns all Spheres in the scene
    inline const std::vector<SphereDesc>& getSpheres() const { return spheres; };

    //! Returns all Cylinders in the scene
    inline const std::vector<CylinderDesc>& getCylinders() const { return cylinders; };

    //! Returns all Meshes in the scene
    inline const std::vector<MeshDesc>& getMeshes() const { return meshes; };

    //! Returns all Updates in the scene
    inline const std::vector<UpdateDesc>& getUpdates() const { return updates; };

    //! Returns all ListenerProximityConditions in the scene
    inline const std::vector<ListenerProximityConditionDesc>& getListenerProximityConditions() const
    {
        return conditions;
    };

    //! Returns all AcousticEnvironments in the scene
    inline const std::vector<AcousticEnvironmentDesc>& getAcousticEnivronments() const
    {
        return environments;
    };

protected:
    const EAR::Vector3 defaultListenerVoiceLocation = Vector3(0.0f, -0.05, 0.0f);
    const std::string sceneFilePath;
    std::filesystem::path basePath;

    std::unique_ptr<XMLReader> xmlFile;

    float audioSceneDuration;
    float predelayOffset;
    SceneType audioSceneType;

    std::string audioSceneId;
    std::set<std::string> ids;

    std::vector<FileStreamDesc> fileStreams;
    std::vector<LocalStreamDesc> localStreams;
    std::vector<SourceDirectivityDesc> directivities;
    std::vector<ObjectSourceDesc> objectSources;
    std::vector<LocalSourceDesc> localSources;
    std::vector<HoaGroupDesc> hoaGroups;
    std::vector<HoaSourceDesc> hoaSources;
    std::vector<ChannelSourceDesc> channelSources;
    std::vector<ChannelSourceDesc> loudspeakerSetups;
    std::vector<AcousticMaterialDesc> materials;
    std::vector<TransformDesc> transforms;
    std::vector<AnchorDesc> anchors;
    std::vector<ARAnchorDesc> arAnchors;
    std::vector<BoxDesc> boxes;
    std::vector<SphereDesc> spheres;
    std::vector<CylinderDesc> cylinders;
    std::vector<MeshDesc> meshes;
    std::vector<UpdateDesc> updates;
    std::vector<ListenerProximityConditionDesc> conditions;
    std::vector<AcousticEnvironmentDesc> environments;

    // Helper functions for the parsing
    inline void createAudioStreams(bool checkFiles);
    inline void createSourceDirectivities(bool checkFiles);
    inline void createAcousticMaterials();
    inline void createBoxes();
    inline void createSpheres();
    inline void createCylinders();
    inline void createMeshes();
    inline void createAcousticEnvironments();
    inline void createLocalSources();
    inline void createLocalStreams();
    inline void createAcousticEnvironments(bool checkFiles);
    inline void createObjectSources();
    inline void createHOAGroups();
    inline void createHOASources();
    inline void createLoudspeakers(
        const std::string& elementName, std::vector<ChannelSourceDesc>& list);
    inline void createListenerProximityConditions();
    inline void createUpdates();
    inline void createTransforms();
    inline void createAnchors();
    inline void createARAnchors();

    inline bool checkIfIdExists(const std::string& id) const;
    inline std::string checkIfFileExistsRelatively(
        const std::string& relativeFile, const std::string& id, bool throwIfNotFound = true) const;

    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        std::string& value, bool required = false, const std::string& eid = "");
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        std::string& value, const std::vector<std::string>& validValues, bool required = false,
        const std::string& eid = "");
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        bool& value, bool required = false, const std::string& eid = "");
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        int& value, bool required = false, const std::string& eid = "");
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        unsigned int& value, bool required = false, const std::string& eid = "");
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        float& value, bool required = false, const std::string& eid = "");
    inline void parseAttribute(
        const XMLReader::Node& node, const std::string& attribute, std::optional<int>& value);
    inline void parseAttribute(
        const XMLReader::Node& node, const std::string& attribute, std::optional<float>& value);
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        std::optional<EAR::Euler>& value);
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        float* value, bool required = false, const std::string& eid = "");
    inline void parseAttribute(const XMLReader::Node& node, const std::string& attribute,
        std::vector<int>& value, bool required = false, const std::string& eid = "");
    inline bool parseOptionalAttribute(
        const XMLReader::Node& node, const std::string& attribute, float* value);
    inline void parseCSpace(XMLReader::Node& node, CSpace& cspace, const std::string& id);
    inline void parseAuthParams(XMLReader::Node& node,
        std::vector<AudioElement::AuthoringParameter>& authParams, const std::string& id);
    void parseVDLMethod(XMLReader::Node& node, VDLMethod& vdlMethodDesc, const std::string& id);
    inline void parseMode(XMLReader::Node& node, AudioElement::Mode& mode, const std::string& id);
    inline std::vector<std::tuple<std::string, Variant, std::vector<int>>> parseModifications(
        XMLReader::Node& node, const std::string& updateId);
    inline void addConfigPredelayOffsetToDiffuseOnset(
        AcousticEnvironmentDesc::AcousticParametersDesc& aparamsRef);

    inline void findMaterial(XMLReader::Node& node, AcousticMaterialDesc*& ptr);
    inline void findCondition(XMLReader::Node& node, ListenerProximityConditionDesc*& ptr);
    inline void findGeometry(XMLReader::Node& node, const std::string& attributeName,
        const std::string& parentId, GeometryDesc*& geometryPointer, bool required = false);
    inline void findFileStream(
        XMLReader::Node& node, const std::string& parentId, FileStreamDesc*& audioStreamPointer);
    inline void findLocalStream(
        XMLReader::Node& node, const std::string& parentId, LocalStreamDesc*& audioStreamPointer);

    inline void findDirectivity(XMLReader::Node& node, SourceDirectivityDesc*& directivityPointer);
    inline void findHoaGroupAndRegister(XMLReader::Node& node, HoaSourceDesc& hoaSource);

    template <class DescType>
    void findPossibleParentTransform(XMLReader::Node& node, DescType& child);
    template <class DescType> void findPossibleParentAnchor(XMLReader::Node& node, DescType& child);

    inline std::vector<std::string> splitString(const std::string& s, char delimiter) const;
    inline void ltrim(std::string& s);
    inline void rtrim(std::string& s);
    inline void trim(std::string& s);
};

} // namespace LSDF

#endif /* LSDF_READER */
