/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef LSDF_XML_READER
#define LSDF_XML_READER

#include <memory>
#include <string>
#include <vector>

// tinyxml2 forward declares
namespace tinyxml2 {
class XMLElement;
class XMLDocument;
} // namespace tinyxml2

namespace LSDF {

//! A simple XML parser (slightly specialized for the requirements of the MPEG-I encoder input format)
/*!
 * This class loads a given XML file which can then be queried with a special query struct. See XMLReader::Query for documentation.
 */
class XMLReader
{
public:
    //! Currently implemented attribute types that can be read
    enum class SearchType
    {
        String = 0, //!< reads the attribute into a std::string
        CoordinateTriple, //!< parses three space-separated float values from the attribute
        IntegerList, //!< parses the attribute as a comma-separated list of integers
        Integer, //!< parses the attribute as a single integer value
        Float, //!< parses the attribute as a single float value
        Bool //!< parses the attribute as a single bool value (everything except "false", "0" and empty string evaluates to true)
    };

    //! Handle for a single Node in an XML element (may only be used in XMLReader::Query)
    struct Node
    {
        //! Return the name of the parent XML element
        std::string getParentName() const;

        //! Return a Node handle of the parent XML element
        Node getParentNode() const;

        friend class XMLReader;

    protected:
        tinyxml2::XMLElement* ptr;
        Node()
            : ptr(nullptr){};
        Node(tinyxml2::XMLElement* ptr)
            : ptr(ptr){};
    };

    //! Construct a query for an attribute of a specific node obtained by getNodes()
    struct Query
    {
        Node node; //!< the XMLReader::Node handle
        std::string attributeName; //!< the name of the attribute
        SearchType type; //!< the expected type of the attribute value
        bool found = false; //!< indicates wether the Node has an attribute with the given name
        std::vector<int>
            placeholders; //!< indicates if some of the found value(s) are declared as placeholders (-1 for no placeholder)

        //! Return the attribute value as string (only valid if SearchType is String)
        std::string& getStringResult() { return stringval; };

        //! Return the attribute value as a pointer to 3 float values (only valid if SearchType is CoordinateTriple)
        float* getCoordinateTripleResult() { return triple; };

        //! Return the attribute value as a reference to a vector of integers (only valid if SearchType is IntegerList)
        std::vector<int>& getIntegerListResult() { return intlist; };

        //! Return the attribute value as an integer value (only valid if SearchType is Integer)
        int getIntegerResult() { return intval; };

        //! Return the attribute value as a float value (only valid if SearchType is Float)
        float getFloatResult() { return floatval; };

        //! Return the attribute value as a bool value (only valid if SearchType is Bool)
        bool getBoolResult() { return boolval; };

        Query() = default;

        ~Query(){
            // clean up non-trivial union members
            /*if (type == string) {
                stringval.~string();
            } else if (type == intlist) {
                intlist.~vector();
            }*/
        };

        friend class XMLReader;

    protected:
        union
        {
            // char stringval[sizeof(std::string)];
            float triple[3];
            // char intlist[sizeof(std::vector<int>)];
            int intval;
            float floatval;
            bool boolval;
        };

        // may add non-trivial members into union when destructor can be called in time
        std::string stringval;
        std::vector<int> intlist;
    };

    //! Constructor
    /*!
     * \param file      full path to file
     */
    XMLReader(const std::string& file);

    //! Destructor
    ~XMLReader();

    //! Search for nodes and their children/attributes
    /*!
     * See XMLQuery for structure of a search query.
     * \param search        an XMLReader::Query instance with a valid XMLReader::Node instance, an attribute name and a XMLReader::SearchType set
     */
    void query(Query& search);

    //! Get number of child nodes
    /*!
     * \param name          name of node
     */
    std::size_t nodeCount(const std::string& name);

    //! Get a vector of all XMLReader::Node handles with a specific node name that are direct children of the root node
    /*!
     * \param name          name of node
     */
    std::vector<Node> getNodes(const std::string& name);

    //! Get a vector of all XMLReader::Node handles with a specific node name that are children of the root node
    /*!
     * In contrast to getNodes(), this traverses into siblings to find deeper level 
     * child nodes with the specified name.
     *
     * \param name          name of node
     */
    std::vector<Node> getNodesRecursive(const std::string& name);

    //! Get a vector of all XMLReader::Node handles with a specific node name that are direct children of the given XMLReader::Node
    /*!
     * \param name          name of node
     * \param parent        a XMLReader::Node handle
     */
    std::vector<Node> getNodes(const std::string& name, Node parent);

    //! Get the root node
    Node getRootNode();

    //! Get name of the root node
    std::string getRootNodeName();

private:
    //! Recursive traverse into child nodes to find elements by name
    std::vector<Node> findChildrenByName(const std::string& name, tinyxml2::XMLElement* parent);

    //! Parse a comma-separated list of integers
    /*!
     * \param input     char pointer to the attribute string
     * \param list      reference to a vector, where the parsed integers will be appended (the vector will not be emptied before)
     */
    void parseIntegerList(const char* input, std::vector<int>& list);

    //! Parse three space-separated float values (e.g. for position and orientation)
    /*!
     * \param input         char pointer to the attribute string
     * \param out           pointer to float array of length 3
     * \param placeholders  vector of ints that will be filled with placeholder values if some are found
     */
    void parseCoordinateTriple(const char* input, float* out, std::vector<int>& placeholders);

    //! Trim whitespace
    void ltrim(std::string& s);
    void rtrim(std::string& s);
    void trim(std::string& s);

    std::unique_ptr<tinyxml2::XMLDocument> xml;
    tinyxml2::XMLElement* rootNode;
};

} // namespace LSDF

#endif /* LSDF_XML_READER */
