/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <LSDF/LSDFParser.h>

#include <algorithm>
#include <cctype>
#include <fstream>
#include <iostream>
#include <optional>
#include <sstream>
#include <utility>

namespace LSDF {

namespace detail {
const std::map<std::string, LSDFParser::ObjectSourceDesc::InputLayout::Positioning>
    string2positioning = {
        { "TL", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::TL },
        { "T", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::T },
        { "TR", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::TR },
        { "L", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::L },
        { "C", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::C },
        { "R", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::R },
        { "BL", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::BL },
        { "B", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::B },
        { "BR", LSDFParser::ObjectSourceDesc::InputLayout::Positioning::BR },
    };
};

LSDFParser::LSDFParser(
    const std::string& scenefile, float predelayOffset, bool checkFiles, bool isLSDF)
    : sceneFilePath(scenefile)
    , xmlFile(std::make_unique<XMLReader>(scenefile))
    , predelayOffset(predelayOffset)
{

    basePath = std::filesystem::path(sceneFilePath).parent_path();

    parseAttribute(xmlFile->getRootNode(), "id", audioSceneId, true);

    audioSceneDuration = 0.f; // default value for optional duration attribute
    parseAttribute(xmlFile->getRootNode(), "duration", audioSceneDuration, false);

    audioSceneType = SceneType::VR; // default value for option type attribute
    std::string typeStr;
    parseAttribute(xmlFile->getRootNode(), "type", typeStr, false);
    trim(typeStr);
    if (typeStr == "AR") {
        audioSceneType = SceneType::AR;
    } else if (typeStr != "") {
        LSDF_CHECK_REL(typeStr == "VR", "Only \"VR\" and \"AR\" are valid AudioScene types.");
    }

    if (!isLSDF) {
        createTransforms();
        createAnchors();

        createAudioStreams(checkFiles);
        createSourceDirectivities(checkFiles);
        createAcousticMaterials();

        createBoxes();
        createSpheres();
        createCylinders();
        createMeshes();

        createAcousticEnvironments(checkFiles);

        createObjectSources();
        createHOAGroups();
        createHOASources();
        createLoudspeakers("ChannelSource", this->channelSources);

        createListenerProximityConditions();

        createUpdates();

        // assert that geometries used for AcousticEnvironments are not modified
        for (auto& acEnv : getAcousticEnivronments()) {
            if (acEnv.region) {
                const auto& id = acEnv.region->id;
                for (auto& upd : getUpdates()) {
                    auto it = std::find_if(upd.modifications.begin(), upd.modifications.end(),
                        [id](const auto& mod) { return mod.id == id; });
                    LSDF_CHECK_REL(it == upd.modifications.end(),
                        "The geometry '" + id
                            + "' is used as the region of an AcousticEnvironment and may not be "
                              "modified.");
                }
            }
        }
    } else { // is LSDF (only Nx AcousticMaterials 1x Mesh, 1x AcousticEnvironment and Mx ARAnchors allowed)
        createAcousticMaterials();

        createMeshes();
        // If there are no meshes then skip AR only checks
        if (meshes.size() != 0) {
            LSDF_CHECK_REL(meshes.size() == 1, "Only one Mesh is allowed in a LSDF file.");
            LSDF_CHECK_REL(
                meshes.back().vertices.size() <= 36, "The LSDF Mesh may only have <= 36 vertices.");
            LSDF_CHECK_REL(
                meshes.back().faces.size() <= 48, "The LSDF Mesh may only have <= 48 faces.");

            createAcousticEnvironments(checkFiles);
            LSDF_CHECK_REL(environments.size() == 1,
                "Only one AcousticEnvironment is allowed in a LSDF file.");
        }

        createARAnchors();
        createSourceDirectivities(checkFiles);
        createLoudspeakers("LoudspeakerSetup", this->loudspeakerSetups);
        createLocalStreams();
        createLocalSources();
    }
}

void LSDFParser::createAudioStreams(bool checkFiles)
{
    auto nodes = xmlFile->getNodes(
        "AudioStream"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        fileStreams.push_back(FileStreamDesc());

        parseAttribute(node, "id", fileStreams.back().id, true);

        LSDF_CHECK_REL(!checkIfIdExists(fileStreams.back().id),
            "AudioStream ID '" + fileStreams.back().id + "' already exists.");

        ids.insert(fileStreams.back().id);

        parseAttribute(node, "file", fileStreams.back().filename, false, fileStreams.back().id);
        fileStreams.back().filename = checkIfFileExistsRelatively(
            fileStreams.back().filename, fileStreams.back().id, checkFiles);

        parseAttribute(node, "aepInputChannels", fileStreams.back().aepInputChannels, true,
            fileStreams.back().id);

        // TODO: check input channel validity?
    }
}

void LSDFParser::createSourceDirectivities(bool checkFiles)
{
    auto nodes = xmlFile->getNodes(
        "SourceDirectivity"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        directivities.push_back(SourceDirectivityDesc());

        parseAttribute(node, "id", directivities.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(directivities.back().id),
            "Directivity ID '" + directivities.back().id + "' already exists.");
        ids.insert(directivities.back().id);

        parseAttribute(node, "file", directivities.back().filename, true, directivities.back().id);
        directivities.back().filename = checkIfFileExistsRelatively(
            directivities.back().filename, directivities.back().id, checkFiles);
    }
}

void LSDFParser::createAcousticMaterials()
{
    auto nodes = xmlFile->getNodes(
        "AcousticMaterial"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        materials.push_back(AcousticMaterialDesc());

        parseAttribute(node, "id", materials.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(materials.back().id),
            "Material ID '" + materials.back().id + "' already exists.");
        ids.insert(materials.back().id);

        auto freqs = xmlFile->getNodes("Frequency", node);
        LSDF_CHECK_REL(freqs.size() >= 1,
            "The Acoustic Material '" + materials.back().id
                + "' must have at least one Frequency child element.");

        for (auto& freq : freqs) {
            materials.back().frequencies.push_back(AcousticMaterialDesc::FrequencyDesc());
            auto& fr = materials.back().frequencies.back();

            parseAttribute(freq, "f", fr.f, true, materials.back().id);
            parseAttribute(freq, "r", fr.r, false);
            parseAttribute(freq, "s", fr.s, false);
            parseAttribute(freq, "t", fr.t, false);
            parseAttribute(freq, "c", fr.c, false);

            LSDF_CHECK_REL(fr.r + fr.s + fr.t + fr.c <= 1 && fr.r + fr.s + fr.t + fr.c >= 0,
                "The coefficients of Acoustic Material '" + materials.back().id + "' at frequency "
                    + std::to_string(fr.f) + "Hz should be >= 0 && <= 1.");
        }
    }
}

void LSDFParser::createBoxes()
{
    auto nodes = xmlFile->getNodesRecursive("Box");

    for (auto& node : nodes) {
        auto parentName = node.getParentName();

        boxes.push_back(BoxDesc());

        parseAttribute(node, "id", boxes.back().id, true);
        LSDF_CHECK_REL(
            !checkIfIdExists(boxes.back().id), "Box ID '" + boxes.back().id + "' already exists.");
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The box '" + boxes.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(boxes.back().id);

        parseAttribute(
            node, "position", boxes.back().position.location.data(), true, boxes.back().id);
        parseAttribute(node, "orientation", boxes.back().position.orientation.data(), false);
        parseCSpace(node, boxes.back().position.cSpace, boxes.back().id);
        findPossibleParentTransform<BoxDesc>(node, boxes.back());
        findPossibleParentAnchor<BoxDesc>(node, boxes.back());
        parseAttribute(node, "size", boxes.back().size.data(), true, boxes.back().id);
        findMaterial(node, boxes.back().material);
    }
}

void LSDFParser::createSpheres()
{
    auto nodes = xmlFile->getNodesRecursive("Sphere");

    for (auto& node : nodes) {
        auto parentName = node.getParentName();

        spheres.push_back(SphereDesc());

        parseAttribute(node, "id", spheres.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(spheres.back().id),
            "Sphere ID '" + spheres.back().id + "' already exists.");
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The sphere '" + spheres.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(spheres.back().id);

        parseAttribute(
            node, "position", spheres.back().position.location.data(), true, spheres.back().id);
        parseAttribute(node, "orientation", spheres.back().position.orientation.data(), false);
        parseCSpace(node, spheres.back().position.cSpace, spheres.back().id);
        findPossibleParentTransform<SphereDesc>(node, spheres.back());
        findPossibleParentAnchor<SphereDesc>(node, spheres.back());
        parseAttribute(node, "size", spheres.back().size.data(), true, spheres.back().id);
        findMaterial(node, spheres.back().material);
    }
}

void LSDFParser::createCylinders()
{
    auto nodes = xmlFile->getNodesRecursive("Cylinder");

    for (auto& node : nodes) {
        auto parentName = node.getParentName();

        cylinders.push_back(CylinderDesc());

        parseAttribute(node, "id", cylinders.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(cylinders.back().id),
            "Cylinder ID '" + cylinders.back().id + "' already exists.");
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The cylinder '" + cylinders.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(cylinders.back().id);

        parseAttribute(
            node, "position", cylinders.back().position.location.data(), true, cylinders.back().id);
        parseAttribute(node, "orientation", cylinders.back().position.orientation.data(), false);
        parseCSpace(node, cylinders.back().position.cSpace, cylinders.back().id);
        findPossibleParentTransform<CylinderDesc>(node, cylinders.back());
        findPossibleParentAnchor<CylinderDesc>(node, cylinders.back());
        parseAttribute(node, "size", cylinders.back().size.data(), true, cylinders.back().id);
        findMaterial(node, cylinders.back().material);
    }
}

void LSDFParser::createMeshes()
{
    auto nodes = xmlFile->getNodesRecursive("Mesh");

    for (auto& node : nodes) {
        auto parentName = node.getParentName();
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "<Mesh> nodes have to be direct children of <AudioScene>, <Transform> or <Anchor>.");

        meshes.push_back(MeshDesc());

        parseAttribute(node, "id", meshes.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(meshes.back().id),
            "Mesh ID '" + meshes.back().id + "' already exists.");
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The mesh '" + meshes.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(meshes.back().id);

        parseAttribute(node, "position", meshes.back().position.location.data(), false);
        parseAttribute(node, "orientation", meshes.back().position.orientation.data(), false);
        parseCSpace(node, meshes.back().position.cSpace, meshes.back().id);
        findPossibleParentTransform<MeshDesc>(node, meshes.back());
        findPossibleParentAnchor<MeshDesc>(node, meshes.back());

        auto vertices = xmlFile->getNodes("Vertex", node);
        LSDF_CHECK_REL(vertices.size() >= 1,
            "The mesh '" + meshes.back().id + "' must have at least one vertex.");

        for (auto& vertex : vertices) {
            meshes.back().vertices.push_back(MeshDesc::VertexDesc());

            int idx = -1;
            parseAttribute(vertex, "index", idx, true, meshes.back().id + " vertex");
            if (idx < 0) {
                LSDF_FAIL("A vertex of the mesh '" + meshes.back().id
                    + "' has an invalid index of '" + std::to_string(idx) + "'. ");
            }

            parseAttribute(vertex, "position", meshes.back().vertices.back().position.data(), true,
                meshes.back().id + " vertex idx=" + std::to_string(idx));

            for (auto v : meshes.back().vertices) {
                if (v.index == idx) {
                    LSDF_FAIL("The mesh'" + meshes.back().id + "' has duplicate vertex indices.");
                }
            }
            meshes.back().vertices.back().index = idx;
        }

        auto faces = xmlFile->getNodes("Face", node);
        LSDF_CHECK_REL(
            faces.size() >= 1, "The mesh '" + meshes.back().id + "' must have at least one face.");

        for (auto& face : faces) {
            int index;
            parseAttribute(face, "index", index, true, meshes.back().id + " face");
            for (auto& face : meshes.back().faces) {
                LSDF_CHECK_REL(face.index != index,
                    "Duplicate face ID of '" + std::to_string(face.index) + "' in mesh '"
                        + meshes.back().id + "'");
            }

            meshes.back().faces.push_back(MeshDesc::FaceDesc());
            meshes.back().faces.back().index = index;

            float temp[3];
            parseAttribute(face, "vertices", temp, true,
                meshes.back().id + " face idx=" + std::to_string(index));
            meshes.back().faces.back().v1 = static_cast<int>(temp[0]);
            meshes.back().faces.back().v2 = static_cast<int>(temp[1]);
            meshes.back().faces.back().v3 = static_cast<int>(temp[2]);

            // TODO: check if vertex ids exist

            findMaterial(face, meshes.back().faces.back().material);
        }
    }
}

void LSDFParser::createAcousticEnvironments(bool checkFiles)
{
    // only considering direct children of AudioScene, as nesting is not allowed
    auto nodes = xmlFile->getNodes("AcousticEnvironment");

    for (auto& node : nodes) {
        environments.push_back(AcousticEnvironmentDesc());

        parseAttribute(node, "id", environments.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(environments.back().id),
            "AcousticEnivronment ID '" + environments.back().id + "' already exists.");
        ids.insert(environments.back().id);

        findGeometry(node, "region", environments.back().id, environments.back().region, false);

        auto lsdfAcousticParamsNodes = xmlFile->getNodes("AcousticParameters", node);
        bool lsdfHasAcousticParams = lsdfAcousticParamsNodes.size() >= 1;
        auto lsdfRIRParamsNodes = xmlFile->getNodes("RoomImpulseResponse", node);
        bool lsdfHasRIRParams = lsdfRIRParamsNodes.size() >= 1;

        LSDF_CHECK_REL(lsdfHasAcousticParams != lsdfHasRIRParams, // xor
            "The AcousticEnvironment '" + environments.back().id
                + "' must have at least one AcousticParameter or RoomImpulseResponse child "
                  "element, and cannot have both.");

        if (lsdfHasAcousticParams) {
            for (auto& lsdfAcousticParams : lsdfAcousticParamsNodes) {
                environments.back().acousticParameters.push_back(
                    AcousticEnvironmentDesc::AcousticParametersDesc());

                auto& envAcousticParams = environments.back().acousticParameters.back();

                parseAttribute(lsdfAcousticParams, "position",
                    envAcousticParams.position.location.data(), true,
                    environments.back().id + " AcousticParameters");

                parseAttribute(lsdfAcousticParams, "diffuseOnset", envAcousticParams.diffuseOnset);
                LSDF_CHECK_REL(!envAcousticParams.diffuseOnset.has_value()
                        || envAcousticParams.diffuseOnset >= 0,
                    "The AcousticParameters in AcousticEnvironment '" + environments.back().id
                        + "' cannot have have a negative diffuseOnset value.");

                // Forward the user-specified Config predelayOffset to AcousticEnvironmentDesc
                envAcousticParams.predelayOffset = predelayOffset;

                parseAttribute(
                    lsdfAcousticParams, "complexityLevel", envAcousticParams.complexityLevel);
                LSDF_CHECK_REL(!envAcousticParams.complexityLevel.has_value()
                        || (envAcousticParams.complexityLevel <= 5
                            && envAcousticParams.complexityLevel >= 1),
                    "The AcousticParameters in AcousticEnvironment '" + environments.back().id
                        + "' cannot have have a complexityLevel value which is zero or less or "
                          "larger than 5.");

                auto freqs = xmlFile->getNodes("Frequency", lsdfAcousticParams);
                LSDF_CHECK_REL(freqs.size() >= 1,
                    "The AcousticParameters in AcousticEnvironment '" + environments.back().id
                        + "' must have at least one Frequency child element.");

                for (auto& freq : freqs) {
                    envAcousticParams.values.push_back(
                        AcousticEnvironmentDesc::AcousticParametersDesc::ValueDesc());

                    parseAttribute(freq, "f", envAcousticParams.values.back().f, true,
                        environments.back().id + " AcousticParameters");
                    LSDF_CHECK_REL(envAcousticParams.values.back().f > 0,
                        "The AcousticParameters in AcousticEnvironment '" + environments.back().id
                            + "' cannot have negative frequency value.");

                    parseAttribute(freq, "rt60", envAcousticParams.values.back().rt60, true,
                        environments.back().id + " AcousticParameters");
                    LSDF_CHECK_REL(envAcousticParams.values.back().rt60 >= 0,
                        "The AcousticParameters in AcousticEnvironment '" + environments.back().id
                            + "' cannot have have a negative RT60 value.");

                    parseAttribute(freq, "rdr", envAcousticParams.values.back().rdr);
                    LSDF_CHECK_REL(!envAcousticParams.values.back().rdr.has_value()
                            || envAcousticParams.values.back().rdr >= 0,
                        "The AcousticParameters in AcousticEnvironment '" + environments.back().id
                            + "' cannot have have a negative RDR value.");
                }
            }

        } else { // hasRIRParams
            for (auto& lsdfRIRParams : lsdfRIRParamsNodes) {
                environments.back().rirParameters.push_back(
                    AcousticEnvironmentDesc::RoomImpulseResponseDesc());

                auto& envRIRParams = environments.back().rirParameters.back();

                // parse and set attributes from LSDF file -> RoomImpulseResponseDesc
                parseAttribute(lsdfRIRParams, "id", envRIRParams.id, true);

                std::string rirPath;
                parseAttribute(lsdfRIRParams, "file", rirPath, true, envRIRParams.id);
                envRIRParams.path =
                    checkIfFileExistsRelatively(rirPath, envRIRParams.id, checkFiles);

                const std::vector<std::string> validFormats = { "ACN-N3D", "ACN-SN3D", "mono" };
                parseAttribute(lsdfRIRParams, "format", envRIRParams.format, validFormats, true,
                    envRIRParams.id);

                float inTriplet[3];

                bool micPositionFound =
                    parseOptionalAttribute(lsdfRIRParams, "micPosition", inTriplet);
                if (micPositionFound) {
                    envRIRParams.micPosition = Position(Vector3(inTriplet), // location
                        Euler(0.f, 0.f, 0.f) // default orientation
                    );
                    envRIRParams.micPosition->cSpace = CSpace::RCS;
                }
                bool micOrientationFound =
                    parseOptionalAttribute(lsdfRIRParams, "micOrientation", inTriplet);
                if (micOrientationFound) {
                    if (micPositionFound) {
                        envRIRParams.micPosition->orientation = Euler(inTriplet);
                    } else {
                        std::cerr << "LSDF specified a micOrientation but no micPosition, ignoring."
                                  << std::endl;
                    }
                }

                bool srcPositionFound =
                    parseOptionalAttribute(lsdfRIRParams, "sourcePosition", inTriplet);
                if (srcPositionFound) {
                    envRIRParams.sourcePosition = Position(Vector3(inTriplet), // location
                        Euler(180.f, 0.f, 0.f) // default orientation
                    );
                    envRIRParams.sourcePosition->cSpace = CSpace::RCS;
                }
                bool srcOrientationFound =
                    parseOptionalAttribute(lsdfRIRParams, "sourceOrientation", inTriplet);
                if (srcOrientationFound) {
                    if (srcPositionFound) {
                        envRIRParams.sourcePosition->orientation = Euler(inTriplet);
                    } else {
                        std::cerr
                            << "LSDF specified a sourceOrientation but no sourcePosition, ignoring."
                            << std::endl;
                    }
                }

                parseAttribute(lsdfRIRParams, "complexityLevel", envRIRParams.complexityLevel);
                LSDF_CHECK_REL(!envRIRParams.complexityLevel.has_value()
                        || (envRIRParams.complexityLevel <= 5 && envRIRParams.complexityLevel >= 1),
                    "The RoomImpulseResponse in AcousticEnvironment '" + environments.back().id
                        + "' cannot have have a complexityLevel value which is zero or less or "
                          "larger than 5.");

                // Forward the user-specified Config predelayOffset to AcousticEnvironmentDesc
                envRIRParams.predelayOffset = predelayOffset;

                parseAttribute(
                    lsdfRIRParams, "directivity", envRIRParams.directivity, false, envRIRParams.id);
            }
        }
    }
}

void LSDFParser::createObjectSources()
{
    auto nodes = xmlFile->getNodesRecursive("ObjectSource");

    for (auto& node : nodes) {
        auto parentName = node.getParentName();

        objectSources.push_back(ObjectSourceDesc());

        parseAttribute(node, "id", objectSources.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(objectSources.back().id),
            "ObjectSource ID '" + objectSources.back().id + "' already exists.");
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The ObjectSource '" + objectSources.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(objectSources.back().id);

        parseAttribute(node, "position", objectSources.back().position.location.data(), true,
            objectSources.back().id);
        parseAttribute(
            node, "orientation", objectSources.back().position.orientation.data(), false);
        parseCSpace(node, objectSources.back().position.cSpace, objectSources.back().id);
        parseAttribute(node, "active", objectSources.back().active, false);
        parseAttribute(node, "gainDb", objectSources.back().gainDb, false);
        parseAttribute(node, "refDistance", objectSources.back().refDistance, false);
        LSDF_CHECK_REL(objectSources.back().refDistance >= 0.f,
            "The reference distance of ObjectSource '" + objectSources.back().id
                + "' should be >= 0 meters.");

        // parse input layout
        auto inputLayouts = xmlFile->getNodes("InputLayout", node);
        LSDF_CHECK_REL(inputLayouts.size() <= 1,
            "The ObjectSource '" + objectSources.back().id + "' can have at most one InputLayout.");

        if (inputLayouts.size() == 1) {
            objectSources.back().inputLayout = std::make_unique<ObjectSourceDesc::InputLayout>();

            std::string temp;
            parseAttribute(inputLayouts[0], "positioning", temp, true,
                objectSources.back().id + " InputLayout");
            std::vector<std::string> tokenizedPositioning = splitString(temp, ' ');
            for (const auto& token : tokenizedPositioning) {
                LSDF_CHECK_REL(
                    detail::string2positioning.find(token) != detail::string2positioning.end(),
                    "'" + token + "' is an invalid label in the InputLayout positioning of "
                        + "ObjectSource '" + objectSources.back().id + "'");
                objectSources.back().inputLayout->positioning.push_back(
                    detail::string2positioning.at(token));
            }

            parseAttribute(
                inputLayouts[0], "alignment", temp, true, objectSources.back().id + " InputLayout");
            LSDF_CHECK_REL(temp == "user" || temp == "object",
                "The InputLayout alignment of ObjectSource '" + objectSources.back().id
                    + "' has to be either 'user' or 'object'.");
            // other option implied from previous check
            objectSources.back().inputLayout->alignment = (temp == "user")
                ? ObjectSourceDesc::InputLayout::Alignment::User
                : ObjectSourceDesc::InputLayout::Alignment::Object;
        }

        findFileStream(node, objectSources.back().id, objectSources.back().signal);
        findGeometry(node, "extent", objectSources.back().id, objectSources.back().extent);
        findDirectivity(node, objectSources.back().directivity);
        parseAttribute(node, "directiveness", objectSources.back().directiveness, false);
        LSDF_CHECK_REL(
            objectSources.back().directiveness >= 0.f && objectSources.back().directiveness <= 1.f,
            "ObjectSource '" + objectSources.back().id + "' directiveness value of '"
                + std::to_string(objectSources.back().directiveness) + "' is out of range. ");
        parseAuthParams(node, objectSources.back().aparams, objectSources.back().id);
        parseMode(node, objectSources.back().mode, objectSources.back().id);
        parseAttribute(node, "play", objectSources.back().play, false);
        findPossibleParentTransform<ObjectSourceDesc>(node, objectSources.back());
        findPossibleParentAnchor<ObjectSourceDesc>(node, objectSources.back());
    }
}

void LSDFParser::createLocalStreams()
{
    auto nodes = xmlFile->getNodes("AudioStream");

    for (auto& node : nodes) {
        localStreams.push_back(LocalStreamDesc());
        auto& localStream = localStreams.back();

        parseAttribute(node, "id", localStream.id, true);
        LSDF_CHECK_REL(!checkIfIdExists(localStream.id),
            "AudioStream ID '" + localStream.id + "' already exists.");
        ids.insert(localStream.id);
        parseAttribute(node, "overrides", localStream.overrides, false, localStream.id);

        parseAttribute(
            node, "localInputChannel", localStream.localInputChannel, false, localStream.id);
    }
}

void LSDFParser::createLocalSources()
{
    auto nodes = xmlFile->getNodes("ObjectSource");

    for (auto& node : nodes) {
        //Add new Source to list
        localSources.push_back(LocalSourceDesc());
        auto& localSource = localSources.back();

        //Store ID
        parseAttribute(node, "id", localSource.id, true);
        LSDF_CHECK_REL(!checkIfIdExists(localSource.id),
            "Local ObjectSource ID '" + localSource.id + "' already exists.");
        ids.insert(localSource.id);

        //Check parent
        LSDF_CHECK_REL(node.getParentName() == "AudioScene",
            "The LocalSource '" + localSource.id + "' has to be a direct child of <AudioScene>.");
        //Set Listener Voice specific parameters
        parseAttribute(node, "listenerVoice", localSource.listenerVoice, false, localSource.id);
        if (localSource.listenerVoice) {
            localSource.position.location = defaultListenerVoiceLocation;
            localSource.position.cSpace = CSpace::LCS;
            // localSource.aparams.push_back(AudioElement::AuthoringParameter::noDistance);
            localSource.vdlMethod = VDLMethod::linear;
            parseAttribute(
                node, "position", localSource.position.location.data(), false, localSource.id);
            parseAttribute(node, "erGainDb", localSource.erGainDB);
            parseAttribute(node, "directGainDb", localSource.directGainDB);
            parseAttribute(node, "reverbGainDb", localSource.reverbGainDB);
        } else {
            parseCSpace(node, localSource.position.cSpace, localSource.id);
            parseAttribute(
                node, "position", localSource.position.location.data(), true, localSource.id);
            parseAttribute(node, "refDistance", localSource.refDistance, false);
            LSDF_CHECK_REL(localSource.refDistance >= 0.f,
                "The reference distance of ObjectSource '" + localSource.id
                    + "' should be >= 0 meters.");
            parseAuthParams(node, localSource.aparams, localSource.id);
        }
        parseVDLMethod(node, localSource.vdlMethod, localSource.id);
        findLocalStream(node, localSource.id, localSource.signal);
        parseAttribute(node, "orientation", localSource.position.orientation.data(), false);
        parseAttribute(node, "active", localSource.active, false);
        parseAttribute(node, "gainDb", localSource.gainDb, false);

        parseAttribute(node, "directiveness", localSource.directiveness, false);
        LSDF_CHECK_REL(localSource.directiveness >= 0.f && localSource.directiveness <= 1.f,
            "ListenerVoiceSource directiveness value of "
                + std::to_string(localSource.directiveness) + " is out of range.");
    }
}

void LSDFParser::createHOAGroups()
{
    auto nodes = xmlFile->getNodes(
        "HOAGroup"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        hoaGroups.push_back(HoaGroupDesc());

        parseAttribute(node, "id", hoaGroups.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(hoaGroups.back().id),
            "HOAGroup ID '" + hoaGroups.back().id + "' already exists.");
        ids.insert(hoaGroups.back().id);

        findGeometry(node, "region", hoaGroups.back().id, hoaGroups.back().region);

        std::string objSources;
        parseAttribute(node, "coSources", objSources, false);
        std::vector<std::string> listOfObjSources = splitString(objSources, ',');

        for (auto& objId : listOfObjSources) {
            trim(objId);

            auto it = std::find_if(objectSources.begin(), objectSources.end(),
                [objId](const ObjectSourceDesc& subject) { return subject.id == objId; });
            if (it != objectSources.end()) {
                hoaGroups.back().coSources.push_back(&(*it));
            }
        }

        LSDF_CHECK_REL(listOfObjSources.size() == hoaGroups.back().coSources.size(),
            "One or multiple coSource IDs of HOAGroup '" + hoaGroups.back().id + "' do not exist.");
    }
}

void LSDFParser::createHOASources()
{
    auto nodes = xmlFile->getNodesRecursive("HOASource");

    for (auto& node : nodes) {
        auto parentName = node.getParentName();

        hoaSources.push_back(HoaSourceDesc());

        parseAttribute(node, "id", hoaSources.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(hoaSources.back().id),
            "HOASource ID '" + hoaSources.back().id + "' already exists.");
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The HOASource '" + hoaSources.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(hoaSources.back().id);

        parseAttribute(node, "position", hoaSources.back().position.location.data(), true,
            hoaSources.back().id);
        parseAttribute(node, "orientation", hoaSources.back().position.orientation.data(), false);
        parseCSpace(node, hoaSources.back().position.cSpace, hoaSources.back().id);
        parseAttribute(node, "active", hoaSources.back().active, false);
        parseAttribute(node, "gainDb", hoaSources.back().gainDb, false);
        parseAttribute(node, "refDistance", hoaSources.back().refDistance, false);
        LSDF_CHECK_REL(hoaSources.back().refDistance >= 0.f,
            "The reference distance of HOASource '" + hoaSources.back().id
                + "' should be >= 0 meters.");
        findFileStream(node, hoaSources.back().id, hoaSources.back().signal);
        findGeometry(node, "extent", hoaSources.back().id, hoaSources.back().extent);
        parseAttribute(node, "extentTransform", hoaSources.back().extentTransform, false);
        parseAttribute(
            node, "representation", hoaSources.back().representation, true, hoaSources.back().id);
        LSDF_CHECK_REL(
            hoaSources.back().representation == 1 || hoaSources.back().representation == 0,
            "HOA Source '" + hoaSources.back().id
                + "' should have a representation value of 0 or 1.");
        parseAttribute(node, "transitionDistance", hoaSources.back().transitionDistance, false);
        LSDF_CHECK_REL(hoaSources.back().transitionDistance >= 0.f,
            "The transition distance of HOASource '" + hoaSources.back().id
                + "' should be larger than 0 meters.");
        findHoaGroupAndRegister(node, hoaSources.back());
        findPossibleParentTransform<HoaSourceDesc>(node, hoaSources.back());
        findPossibleParentAnchor<HoaSourceDesc>(node, hoaSources.back());
        parseAuthParams(node, hoaSources.back().aparams, hoaSources.back().id);
        parseMode(node, hoaSources.back().mode, hoaSources.back().id);
        parseAttribute(node, "play", hoaSources.back().play, false);
    }
}

namespace {
//! Obtain orientation at a given position when looking at another given position.
Euler lookOrientation(Vector3 from, Vector3 to)
{
    auto minus_z = to - from; // vector in look direction, default look direction is -z
    auto z_length = minus_z.abs();
    if (z_length < 0.000001f) {
        LSDF_FAIL("lookOrientation: points are too close");
    }
    minus_z = minus_z.normalize(); // NB: division by scalar is not implemented yet
    Vector3 up{ 0.0f, 1.0f, 0.0f };
    auto up_ness = minus_z.dot(up);
    if (up_ness > 0.999999f) {
        // Looking up, remaining DoF is chosen arbitrarily. This is a singularity!
        return Euler{ 0, 90, 0 };
    } else if (up_ness < -0.999999f) {
        // Looking down, remaining DoF is chosen arbitrarily. This is a singularity!
        return Euler{ 0, -90, 0 };
    }
    auto x = minus_z.cross(up);
    x = x.normalize();
    auto y = x.cross(minus_z);
    return HomogenMatrix{ x, y, -minus_z, { 0, 0, 0 } }.getRotation();
}
} // namespace

void LSDFParser::createLoudspeakers(
    const std::string& elementName, std::vector<ChannelSourceDesc>& list)
{
    auto nodes = xmlFile->getNodesRecursive(elementName);

    for (auto& node : nodes) {
        auto parentName = node.getParentName();

        list.push_back(ChannelSourceDesc());

        if (elementName != "LoudspeakerSetup") {
            parseAttribute(node, "id", list.back().id, true);
            LSDF_CHECK_REL(!checkIfIdExists(list.back().id),
                elementName + " ID '" + list.back().id + "' already exists.");
        }
        LSDF_CHECK_REL(
            parentName == "AudioScene" || parentName == "Transform" || parentName == "Anchor",
            "The " + elementName + " '" + hoaSources.back().id
                + "' has to be a direct child of <AudioScene>, <Transform> or <Anchor>.");
        ids.insert(list.back().id);

        parseAttribute(
            node, "position", list.back().position.location.data(), true, list.back().id);
        parseAttribute(node, "orientation", list.back().position.orientation.data(), false);
        parseCSpace(node, list.back().position.cSpace, list.back().id);
        parseAttribute(node, "inputLayout", list.back().inputLayout, true, list.back().id);
        LSDF_CHECK_REL(list.back().inputLayout >= 0 && list.back().inputLayout <= 127,
            elementName + " '" + list.back().id
                + "' should have a CICP index in between 0 and 127 as an inputLayout.");
        parseAttribute(node, "active", list.back().active, false);
        parseAttribute(node, "gainDb", list.back().gainDb, false);
        parseAttribute(node, "refDistance", list.back().refDistance, false);
        LSDF_CHECK_REL(list.back().refDistance >= 0.f,
            "The reference distance of " + elementName + " '" + list.back().id
                + "' should be >= 0 meters.");
        if (elementName != "LoudspeakerSetup") {
            findFileStream(node, list.back().id, list.back().signal);
        }
        parseAuthParams(node, list.back().aparams, list.back().id);
        parseMode(node, list.back().mode, list.back().id);
        parseAttribute(node, "play", list.back().play, false);
        findPossibleParentTransform<ChannelSourceDesc>(node, list.back());
        findPossibleParentAnchor<ChannelSourceDesc>(node, list.back());

        if (elementName == "LoudspeakerSetup") {
            auto referencePoints = xmlFile->getNodes("ReferencePoint", node);
            if (referencePoints.size() == 0) {
                // nothing to do here
            } else if (referencePoints.size() == 1) {
                auto referencePoint = referencePoints[0];
                auto& target = list.back().referencePoint;
                target = ChannelSourceDesc::ReferencePoint{};
                parseAttribute(referencePoint, "position", target->position.location.data(), true,
                    "ReferencePoint");
                parseAttribute(referencePoint, "calibrated", target->calibrated, false);
            } else {
                LSDF_FAIL("There can be at most one ReferencePoint");
            }

            {
                auto directivityCompensations = xmlFile->getNodes("DirectivityCompensation", node);
                auto& target = list.back().directivityCompensation;
                target = ChannelSourceDesc::DirectivityCompensation{};
                if (directivityCompensations.size() == 0) {
                    // defaults are used
                } else if (directivityCompensations.size() == 1) {
                    auto directivityCompensation = directivityCompensations[0];
                    parseAttribute(directivityCompensation, "eqFreq", target->eqFreq, false);
                    parseAttribute(directivityCompensation, "eqGainDb", target->eqGainDb, false);
                    parseAttribute(
                        directivityCompensation, "eqAngleStart", target->eqAngleStart, false);
                    parseAttribute(
                        directivityCompensation, "eqAngleRange", target->eqAngleRange, false);
                } else {
                    LSDF_FAIL("There can be at most one DirectivityCompensation element");
                }
            }

            {
                auto levelCompensations = xmlFile->getNodes("LevelCompensation", node);
                auto& target = list.back().levelCompensation;
                target = ChannelSourceDesc::LevelCompensation{};
                if (levelCompensations.size() == 0) {
                    // defaults are used
                } else if (levelCompensations.size() == 1) {
                    auto levelCompensation = levelCompensations[0];
                    parseAttribute(levelCompensation, "slope1", target->slope1, false);
                    parseAttribute(levelCompensation, "slope2", target->slope2, false);
                    parseAttribute(
                        levelCompensation, "criticalDistance", target->criticalDistance, false);
                    parseAttribute(levelCompensation, "beta", target->beta, false);
                } else {
                    LSDF_FAIL("There can be at most one LevelCompensation element");
                }
            }
        }
        auto loudspeakers = xmlFile->getNodes("Loudspeaker", node);
        LSDF_CHECK_REL(loudspeakers.size() >= 1,
            "The " + elementName + " '" + list.back().id
                + "' must "
                  "have at least one loudspeaker within it");
        for (auto& loudspeaker : loudspeakers) {
            list.back().loudspeakers.push_back(ChannelSourceDesc::LoudspeakerDesc());
            auto& ldSpeaker = list.back().loudspeakers.back();

            parseAttribute(loudspeaker, "channel", ldSpeaker.channelIndex, true,
                list.back().id + " Loudspeaker");
            parseAttribute(loudspeaker, "position", ldSpeaker.position.location.data(), true,
                list.back().id + " channel=" + std::to_string(ldSpeaker.channelIndex));
            std::optional<Euler> orientation{};
            parseAttribute(loudspeaker, "orientation", orientation);
            if (orientation.has_value()) {
                ldSpeaker.position.orientation = orientation.value();
            } else {
                const auto& referencePoint = list.back().referencePoint;
                Vector3 sweetSpot{ 0.f, 0.f, 0.f };
                if (referencePoint) {
                    sweetSpot = referencePoint->position.location;
                } else {
                    // local origin {0, 0, 0} is used
                }
                ldSpeaker.position.orientation =
                    lookOrientation(ldSpeaker.position.location, sweetSpot);
            }
            findDirectivity(loudspeaker, ldSpeaker.directivity);
            parseAttribute(loudspeaker, "directiveness", ldSpeaker.directiveness, false);
            LSDF_CHECK_REL(ldSpeaker.directiveness >= 0.f && ldSpeaker.directiveness <= 1.f,
                "A loudspeaker of '" + list.back().id
                    + "' has an out of range directiveness value of '"
                    + std::to_string(ldSpeaker.directiveness) + "'. ");
        }
    }
}

void LSDFParser::createTransforms()
{
    auto nodes = xmlFile->getNodes(
        "Transform"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        transforms.push_back(TransformDesc());

        parseAttribute(node, "id", transforms.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(transforms.back().id),
            "Transform ID '" + transforms.back().id + "' already exists.");
        ids.insert(transforms.back().id);

        parseAttribute(node, "position", transforms.back().position.location.data(), false);
        parseAttribute(node, "orientation", transforms.back().position.orientation.data(), false);
        parseCSpace(node, transforms.back().position.cSpace, transforms.back().id);
    }
}

void LSDFParser::createAnchors()
{
    auto nodes = xmlFile->getNodes(
        "Anchor"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        anchors.push_back(AnchorDesc());

        parseAttribute(node, "id", anchors.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(anchors.back().id),
            "Anchor ID '" + anchors.back().id + "' already exists.");
        ids.insert(anchors.back().id);

        parseAttribute(node, "lsdf_ref", anchors.back().lsdf_ref, true, anchors.back().id);
    }
}

void LSDFParser::createARAnchors()
{
    auto nodes = xmlFile->getNodes(
        "ARAnchor"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        arAnchors.push_back(ARAnchorDesc());

        parseAttribute(node, "id", arAnchors.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(arAnchors.back().id),
            "ARAnchor ID '" + arAnchors.back().id + "' already exists.");
        ids.insert(arAnchors.back().id);

        parseAttribute(node, "position", arAnchors.back().position.location.data(), false);
        parseAttribute(node, "orientation", arAnchors.back().position.orientation.data(), false);
    }
}

void LSDFParser::createListenerProximityConditions()
{
    auto nodes = xmlFile->getNodes(
        "ListenerProximityCondition"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        conditions.push_back(ListenerProximityConditionDesc());

        parseAttribute(node, "id", conditions.back().id, true);
        LSDF_CHECK_REL(!checkIfIdExists(conditions.back().id),
            "Condition ID '" + conditions.back().id + "' already exists.");
        ids.insert(conditions.back().id);

        findGeometry(node, "region", conditions.back().id, conditions.back().region, true);
    }
}

void LSDFParser::createUpdates()
{
    // used for checking uniqueness
    std::set<int> updateIndices;
    std::set<float> updateTimes;
    std::set<std::pair<ListenerProximityConditionDesc*, bool>> conditions;

    auto nodes = xmlFile->getNodes(
        "Update"); // only considering direct children of AudioScene, as nesting is not allowed

    for (auto& node : nodes) {
        updates.push_back(UpdateDesc());

        parseAttribute(node, "id", updates.back().id, false);
        parseAttribute(node, "index", updates.back().index, false);
        findCondition(node, updates.back().condition);
        parseAttribute(node, "fireOn", updates.back().fireOn, false);
        parseAttribute(node, "delay", updates.back().delay, false);
        parseAttribute(node, "time", updates.back().time, false);

        if (updates.back().time >= 0) {
            LSDF_CHECK_REL(audioSceneDuration < 1e-5 || updates.back().time <= audioSceneDuration,
                "The Update '" + updates.back().id
                    + "' must have a time smaller or equal to the AudioScene duration.");
            LSDF_CHECK_REL(updates.back().condition == nullptr,
                "For timed updates, the attribute condition is not allowed (ID: '"
                    + updates.back().id + "').");
            LSDF_CHECK_REL(updates.back().delay <= 0,
                "For timed updates, the attribute delay is not allowed (ID: '" + updates.back().id
                    + "').");
            LSDF_CHECK_REL(updateTimes.find(updates.back().time) == updateTimes.end(),
                "An Update with time '" + std::to_string(updates.back().time)
                    + "' already exists.");
            updateTimes.insert(updates.back().time);
            updates.back().type = UpdateDesc::Type::Timed;
        }

        if (updates.back().condition) {
            LSDF_CHECK_REL(updates.back().type == UpdateDesc::Type::Undefined,
                "A triggered or timed update cannot have a condition attribute (ID: '"
                    + updates.back().id + "').");
            LSDF_CHECK_REL(updates.back().delay >= 0.f,
                "The Update '" + updates.back().id + "' must have a delay >= 0.");

            auto pair = std::make_pair(updates.back().condition, updates.back().fireOn);
            LSDF_CHECK_REL(conditions.find(pair) == conditions.end(),
                "Two Updates with the same condition/fireOn combination may not exist (ID: '"
                    + updates.back().id + "').");
            conditions.insert(pair);

            updates.back().type = UpdateDesc::Type::Conditional;
        } else if (updates.back().delay > 0) {
            LSDF_FAIL("Only conditional updates can have a delay attribute > 0 (ID: '"
                + updates.back().id + "').");
        }

        if (updates.back().index != -1) {
            LSDF_CHECK_REL(updateIndices.find(updates.back().index) == updateIndices.end(),
                "Update index '" + std::to_string(updates.back().index) + "' already exists.");
            updateIndices.insert(updates.back().index);
            if (updates.back().type == UpdateDesc::Type::Undefined) {
                LSDF_CHECK_REL(updates.back().id != "",
                    "For externally triggered updates, the attribute id is required (ID: '"
                        + updates.back().id + "').");
                LSDF_CHECK_REL(updates.back().time < 0,
                    "For externally triggered updates, the attribute time is not allowed (ID: '"
                        + updates.back().id + "').");
                LSDF_CHECK_REL(updates.back().condition == nullptr,
                    "For externally triggered updates, the attribute condition is not allowed (ID: "
                    "'" + updates.back().id
                        + "').");
                LSDF_CHECK_REL(updates.back().delay <= 0,
                    "For externally triggered updates, the attribute delay is not allowed (ID: '"
                        + updates.back().id + "').");
                updates.back().type = UpdateDesc::Type::Triggered;
            }
        }
        if (updates.back().id != "") {
            LSDF_CHECK_REL(!checkIfIdExists(updates.back().id),
                "Updates ID '" + updates.back().id + "' already exists.");
            ids.insert(updates.back().id);
            LSDF_CHECK_REL(updates.back().type != UpdateDesc::Type::Undefined,
                "For externally triggered updates, the attributes 'index' AND 'id' are required "
                "(ID: '"
                    + updates.back().id + "').");
        }

        LSDF_CHECK_REL(updates.back().type != UpdateDesc::Type::Undefined,
            "Something went wrong while determining the Update type (ID: '" + updates.back().id
                + "').");

        auto mods = xmlFile->getNodes("Modify", node);
        LSDF_CHECK_REL(mods.size() >= 1,
            "The Update '" + updates.back().id + "' must have one or more Modify child elements.");

        std::set<std::string> modifiedAttributes;
        for (auto& mod : mods) {
            std::string modId;
            parseAttribute(mod, "id", modId, true);
            LSDF_CHECK_REL(checkIfIdExists(modId),
                "Trying to modify non-existent target entity with id '" + modId + "' (Update ID: '"
                    + updates.back().id + "').");

            for (auto& modification : parseModifications(mod, updates.back().id)) {
                LSDF_CHECK_REL(modifiedAttributes.find(std::get<0>(modification) + modId)
                        == modifiedAttributes.end(),
                    "Attribute '" + std::get<0>(modification) + "' of '" + modId
                        + "' may not be modified more than once in Update '" + updates.back().id
                        + "'.");
                modifiedAttributes.insert(std::get<0>(modification) + modId);

                updates.back().modifications.emplace_back();
                updates.back().modifications.back().id = modId;

                parseAttribute(
                    mod, "duration", updates.back().modifications.back().duration, false);
                LSDF_CHECK_REL(updates.back().modifications.back().duration >= 0.f,
                    "Duration of modification '" + updates.back().modifications.back().id
                        + "' must be non-negative (Update ID: '" + updates.back().id + "').");

                updates.back().modifications.back().attribute = std::get<0>(modification);
                updates.back().modifications.back().targetValue = std::get<1>(modification);
                for (unsigned int i = 0; i < std::get<2>(modification).size();
                     i++) { // iterate over placeholders
                    int ph = std::get<2>(modification)[i];
                    if (ph != -1) {

                        // placeholder is present, so we check if it is allowed
                        LSDF_CHECK_REL(updates.back().type == UpdateDesc::Type::Triggered,
                            "Placeholders are only allowed in triggered updates.");
                        LSDF_CHECK_REL(updates.back().modifications.back().duration <= 0,
                            "Non-zero duration is not allowed for dynamic updates with "
                            "placeholders.");

                        if (std::get<0>(modification) == "gainDb") {
                            updates.back().modifications.back().placeholders[ph] =
                                UpdateDesc::Modification::Placeholder::GainDb;
                        } else if (std::get<0>(modification) == "position") {
                            if (i == 0) {
                                updates.back().modifications.back().placeholders[ph] =
                                    UpdateDesc::Modification::Placeholder::PositionX;
                            } else if (i == 1) {
                                updates.back().modifications.back().placeholders[ph] =
                                    UpdateDesc::Modification::Placeholder::PositionY;
                            } else { // i == 2 implied, because we definitely only have 3 values
                                updates.back().modifications.back().placeholders[ph] =
                                    UpdateDesc::Modification::Placeholder::PositionZ;
                            }
                        } else if (std::get<0>(modification) == "orientation") {
                            if (i == 0) {
                                updates.back().modifications.back().placeholders[ph] =
                                    UpdateDesc::Modification::Placeholder::OrientationY;
                            } else if (i == 1) {
                                updates.back().modifications.back().placeholders[ph] =
                                    UpdateDesc::Modification::Placeholder::OrientationP;
                            } else { // i == 2 implied, because we definitely only have 3 values
                                updates.back().modifications.back().placeholders[ph] =
                                    UpdateDesc::Modification::Placeholder::OrientationR;
                            }
                        } else {
                            LSDF_FAIL("Placeholder values are only allowed in gainDb, position and "
                                      "orientation attributes (Update ID: '"
                                + updates.back().id + "').")
                        }
                    }
                }
            }
        }

        // check that the placeholder keys are in a continuous range from 0 - N
        std::set<int> placeholderKeys;
        for (auto& mod : updates.back().modifications) {
            for (auto& ph : mod.placeholders) {
                LSDF_CHECK_REL(placeholderKeys.find(ph.first) == placeholderKeys.end(),
                    "Duplicate placeholder key in update with id '" + updates.back().id + "'.");
                placeholderKeys.insert(ph.first);
            }
        }
        for (unsigned int i = 0; i < placeholderKeys.size(); i++) {
            LSDF_CHECK_REL(placeholderKeys.find(i) != placeholderKeys.end(),
                "The placeholder keys of the update with id '" + updates.back().id
                    + "' should be consecutive integers from 0 to "
                    + std::to_string(placeholderKeys.size() - 1));
        }
    }
}

bool LSDFParser::checkIfIdExists(const std::string& id) const
{
    return ids.find(id) != ids.end();
}

std::string LSDFParser::checkIfFileExistsRelatively(
    const std::string& relativeFile, const std::string& id, bool throwIfNotFound) const
{
    auto fileToFind = basePath / relativeFile;
    std::ifstream fileStream(fileToFind.c_str());
    LSDF_CHECK_REL(!throwIfNotFound || fileStream.good(),
        "File '" + fileToFind.string() + "' not found in entity '" + id + "'.");

    return fileToFind.string();
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    std::string& value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::String;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        value = q.getStringResult();
    }
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    std::string& value, const std::vector<std::string>& validValues, bool required,
    const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::String;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        std::string& result = q.getStringResult();

        // Check for a valid match
        for (const std::string& validStr : validValues) {
            if (result == validStr) { // TODO: support case-insensitive?
                value = result;
                return; // match found
            }
        }

        // No match found...
        std::string errStr = "Attribute '" + attribute + "' for element with id '" + eid
            + "' must be one of the following";
        for (const std::string& str : validValues) {
            errStr += ", " + str;
        }
        errStr += ". You provided " + result + ".";
        LSDF_FAIL(errStr);
    }
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    bool& value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::Bool;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        value = q.getBoolResult();
    }
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    int& value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::Integer;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        value = q.getIntegerResult();
    }
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    unsigned int& value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::Integer;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        //value = q.getIntegerResult();
        if (q.getIntegerResult() < 0) {
            LSDF_FAIL("Attribute '" + attribute
                + "' is unsigned int and cannot be negative in element with id '" + eid + "'.");
        } else {
            value = q.getIntegerResult();
        }
    }
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    float& value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::Float;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        value = q.getFloatResult();
    }
}

void LSDFParser::parseAttribute(
    const XMLReader::Node& node, const std::string& attribute, std::optional<int>& value)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::Integer;
    xmlFile->query(q);

    if (q.found) {
        value = q.getIntegerResult();
    } else {
        value.reset();
    }
}

void LSDFParser::parseAttribute(
    const XMLReader::Node& node, const std::string& attribute, std::optional<float>& value)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::Float;
    xmlFile->query(q);

    if (q.found) {
        value = q.getFloatResult();
    } else {
        value.reset();
    }
}

void LSDFParser::parseAttribute(
    const XMLReader::Node& node, const std::string& attribute, std::optional<Euler>& value)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::CoordinateTriple;
    xmlFile->query(q);

    if (q.found) {
        value = Euler{ q.getCoordinateTripleResult() };
    } else {
        value.reset();
    }
}

bool LSDFParser::parseOptionalAttribute(
    const XMLReader::Node& node, const std::string& attribute, float* value)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::CoordinateTriple;
    xmlFile->query(q);

    if (q.found) {
        value[0] = q.getCoordinateTripleResult()[0];
        value[1] = q.getCoordinateTripleResult()[1];
        value[2] = q.getCoordinateTripleResult()[2];
    }

    return q.found;
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    float* value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::CoordinateTriple;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        value[0] = q.getCoordinateTripleResult()[0];
        value[1] = q.getCoordinateTripleResult()[1];
        value[2] = q.getCoordinateTripleResult()[2];
    }
}

void LSDFParser::parseAttribute(const XMLReader::Node& node, const std::string& attribute,
    std::vector<int>& value, bool required, const std::string& eid)
{
    XMLReader::Query q;
    q.node = node;
    q.attributeName = attribute;
    q.type = XMLReader::SearchType::IntegerList;
    xmlFile->query(q);

    if (required && !q.found) {
        LSDF_FAIL("Attribute '" + attribute + "' is required for element with id '" + eid + "'.");
    }

    if (q.found) {
        value = q.getIntegerListResult();
    }
}

void LSDFParser::parseCSpace(XMLReader::Node& node, CSpace& cspace, const std::string& id)
{
    std::string newCSpaceValue = "invalidParameter";
    parseAttribute(node, "cspace", newCSpaceValue, false);

    if (newCSpaceValue == "global") {
        cspace = CSpace::WCS;
    } else if (newCSpaceValue == "relative") {
        cspace = CSpace::RCS;
    } else if (newCSpaceValue == "user") {
        cspace = CSpace::LCS;
    } else if (newCSpaceValue != "invalidParameter") {
        LSDF_FAIL("The entity '" + id + "' has an invalid CSpace parameter of '" + newCSpaceValue
            + "'. "
              "'global' 'user' and 'relative' are the only valid parameters.");
    }
}

void LSDFParser::parseAuthParams(XMLReader::Node& node,
    std::vector<AudioElement::AuthoringParameter>& authParams, const std::string& id)
{
    std::string authParam;
    parseAttribute(node, "aparams", authParam, false);
    std::vector<std::string> listOfAuthParams = splitString(authParam, ',');

    for (auto& aparam : listOfAuthParams) {
        trim(aparam);
        if (aparam == "noDoppler") {
            authParams.push_back(AudioElement::AuthoringParameter::noDoppler);
        } else if (aparam == "noDistance") {
            authParams.push_back(AudioElement::AuthoringParameter::noDistance);
        } else {
            LSDF_FAIL(
                "The entity '" + id + "' has an invalid authoring parameter '" + aparam + "'. ");
        }
    }
}

void LSDFParser::parseVDLMethod(
    XMLReader::Node& node, VDLMethod& vdlMethodDesc, const std::string& id)
{
    std::string vdlMethod;
    parseAttribute(node, "vdlMethod", vdlMethod, false);
    if (vdlMethod != "") {
        if (vdlMethod == "spline") {
            vdlMethodDesc = VDLMethod::spline;
        } else if (vdlMethod == "linear") {
            vdlMethodDesc = VDLMethod::linear;
        } else if (vdlMethod == "spline_erLinear") {
            vdlMethodDesc = VDLMethod::spline_erLinear;
        } else {
            LSDF_FAIL("The entity '" + id + "' has an invalid VDL method '" + vdlMethod + "'. ");
        }
    }
}

void LSDFParser::parseMode(XMLReader::Node& node, AudioElement::Mode& mode, const std::string& id)
{
    std::string parsedMode;
    parseAttribute(node, "mode", parsedMode, false);

    if (parsedMode == "continuous") {
        mode = AudioElement::Mode::Continuous;
    } else if (parsedMode == "event") {
        mode = AudioElement::Mode::Event;
    } else if (parsedMode != "") {
        LSDF_FAIL(
            "The entity '" + id + "' has an invalid mode parameter of '" + parsedMode + "'. ");
    }
}

std::vector<std::tuple<std::string, Variant, std::vector<int>>> LSDFParser::parseModifications(
    XMLReader::Node& node, const std::string& updateId)
{
    std::vector<std::tuple<std::string, Variant, std::vector<int>>> mods;

    XMLReader::Query q;
    q.node = node;

    std::vector<std::string> attributes({ "position", "orientation", "cspace", "size", "active",
        "gainDb", "extent", "region", "directivity", "directiveness", "play", "signal" });

    for (const auto& attribute : attributes) {
        q.attributeName = attribute;

        if (attribute == "position" || attribute == "orientation" || attribute == "size") {
            q.type = XMLReader::SearchType::CoordinateTriple;
            xmlFile->query(q);

            if (q.found) {
                mods.push_back(std::make_tuple(
                    attribute, Variant(Vector3(q.getCoordinateTripleResult())), q.placeholders));
            }
        } else if (attribute == "extent" || attribute == "region" || attribute == "directivity"
            || attribute == "signal") {
            q.type = XMLReader::SearchType::String;
            xmlFile->query(q);

            if (q.found) {
                mods.emplace_back(
                    std::make_tuple(attribute, Variant(q.getStringResult()), q.placeholders));
            }
        } else if (attribute == "active" || attribute == "play") {
            q.type = XMLReader::SearchType::Bool;
            xmlFile->query(q);

            if (q.found) {
                mods.push_back(
                    std::make_tuple(attribute, Variant(q.getBoolResult()), q.placeholders));
            }
        } else if (attribute == "gainDb" || attribute == "directiveness") {
            q.type = XMLReader::SearchType::Float;
            xmlFile->query(q);

            if (q.found) {
                mods.push_back(
                    std::make_tuple(attribute, Variant(q.getFloatResult()), q.placeholders));
            }
        } else if (attribute == "cspace") {
            q.type = XMLReader::SearchType::String;
            xmlFile->query(q);

            if (q.found) {
                std::string val = q.getStringResult();
                CSpace cs = CSpace::WCS; // default value

                if (val == "user") {
                    cs = CSpace::LCS;
                } else if (val == "relative") {
                    cs = CSpace::RCS;
                }

                mods.push_back(
                    std::make_tuple(attribute, Variant(static_cast<int>(cs)), q.placeholders));
            }

        } else {
            LSDF_FAIL("The Update '" + updateId + "' is trying to modify the attribute '"
                + attribute + "', which does not exist or is not modifable.");
        }
    }

    return mods;
}

void LSDFParser::addConfigPredelayOffsetToDiffuseOnset(
    AcousticEnvironmentDesc::AcousticParametersDesc& aparamsRef)
{
    float predelay = predelayOffset;
    if (aparamsRef.diffuseOnset.has_value()) {
        predelay += aparamsRef.diffuseOnset.value();
    }
    if (predelay < 0.0f) {
        std::cerr << "(predelayOffset + diffuseOnset) value of " << predelay
                  << " is not allowed, using zero instead." << std::endl;
        aparamsRef.diffuseOnset = 0.0f;
    } else {
        aparamsRef.diffuseOnset = predelay;
    }
}

void LSDFParser::findMaterial(XMLReader::Node& node, AcousticMaterialDesc*& ptr)
{
    std::string mid;
    parseAttribute(node, "material", mid);

    auto it = std::find_if(materials.begin(), materials.end(),
        [mid](const AcousticMaterialDesc& subject) { return subject.id == mid; });

    if (it != materials.end()) {
        ptr = &(*it);
    } else {
        ptr = nullptr;

        LSDF_CHECK_REL(mid.size() == 0, "AcousticMaterial with id '" + mid + "' does not exist.");
    }
}

void LSDFParser::findCondition(XMLReader::Node& node, ListenerProximityConditionDesc*& ptr)
{
    std::string cid = "";
    parseAttribute(node, "condition", cid);

    auto it = std::find_if(conditions.begin(), conditions.end(),
        [cid](const ListenerProximityConditionDesc& subject) { return subject.id == cid; });

    if (it != conditions.end()) {
        ptr = &(*it);
    } else {
        ptr = nullptr;

        LSDF_CHECK_REL(cid.size() == 0, "Update condition with id '" + cid + "' does not exist.");
    }
}

void LSDFParser::findGeometry(XMLReader::Node& node, const std::string& attributeName,
    const std::string& parentId, GeometryDesc*& geometryPointer, bool required)
{
    std::string geometryId;
    parseAttribute(node, attributeName, geometryId, required, parentId);

    auto boxIt = std::find_if(boxes.begin(), boxes.end(),
        [geometryId](const BoxDesc& subject) { return subject.id == geometryId; });

    auto sphereIt = std::find_if(spheres.begin(), spheres.end(),
        [geometryId](const SphereDesc& subject) { return subject.id == geometryId; });

    auto cylinderIt = std::find_if(cylinders.begin(), cylinders.end(),
        [geometryId](const CylinderDesc& subject) { return subject.id == geometryId; });

    auto meshIt = std::find_if(meshes.begin(), meshes.end(),
        [geometryId](const MeshDesc& subject) { return subject.id == geometryId; });

    if (boxIt == boxes.end() && sphereIt == spheres.end() && cylinderIt == cylinders.end()
        && meshIt == meshes.end()) {

        if (required || geometryId.size() != 0) {
            LSDF_FAIL("The entity '" + parentId + "' is referring to Geometry '" + geometryId
                + "' which does not exist.");
        }
    }

    if (boxIt != boxes.end()) {
        geometryPointer = &(*boxIt);
    } else if (sphereIt != spheres.end()) {
        geometryPointer = &(*sphereIt);
    } else if (cylinderIt != cylinders.end()) {
        geometryPointer = &(*cylinderIt);
    } else if (meshIt != meshes.end()) {
        geometryPointer = &(*meshIt);
    } else {
        geometryPointer = nullptr;
    }
}

void LSDFParser::findFileStream(
    XMLReader::Node& node, const std::string& parentId, FileStreamDesc*& audioStreamPointer)
{
    std::string audioStreamId;
    parseAttribute(node, "signal", audioStreamId, true, parentId);

    auto it = std::find_if(fileStreams.begin(), fileStreams.end(),
        [audioStreamId](const FileStreamDesc& subject) { return subject.id == audioStreamId; });

    LSDF_CHECK_REL(it != fileStreams.end(),
        "The Source '" + parentId + "' refers to the File Audio Stream '" + audioStreamId
            + "' which does not exist.");
    audioStreamPointer = &(*it);
}

void LSDFParser::findLocalStream(
    XMLReader::Node& node, const std::string& parentId, LocalStreamDesc*& audioStreamPointer)
{
    std::string audioStreamId;
    parseAttribute(node, "signal", audioStreamId, true, parentId);

    auto it = std::find_if(localStreams.begin(), localStreams.end(),
        [audioStreamId](const LocalStreamDesc& subject) { return subject.id == audioStreamId; });

    LSDF_CHECK_REL(it != localStreams.end(),
        "The Source '" + parentId + "' refers to the Local Audio Stream '" + audioStreamId
            + "' which does not exist.");
    audioStreamPointer = &(*it);
}

void LSDFParser::findDirectivity(XMLReader::Node& node, SourceDirectivityDesc*& directivityPointer)
{
    std::string directivityId;
    parseAttribute(node, "directivity", directivityId);

    auto it = std::find_if(directivities.begin(), directivities.end(),
        [directivityId](
            const SourceDirectivityDesc& subject) { return subject.id == directivityId; });

    if (it != directivities.end()) {
        directivityPointer = &(*it);
    } else {
        directivityPointer = nullptr;

        LSDF_CHECK_REL(directivityId.size() == 0,
            "Directivity with ID '" + directivityId + "' was referenced, but does not exist.");
    }
}

void LSDFParser::findHoaGroupAndRegister(XMLReader::Node& node, HoaSourceDesc& hoaSource)
{
    std::string groupId;
    parseAttribute(node, "group", groupId, false);

    auto hoaGroupIt = std::find_if(hoaGroups.begin(), hoaGroups.end(),
        [&groupId](const HoaGroupDesc& hoaGroup) { return hoaGroup.id == groupId; });

    if (hoaGroupIt != hoaGroups.end()) {
        hoaSource.group = &(*hoaGroupIt);
        hoaGroupIt->hoaSources.push_back(&hoaSource);
    } else { // group not found
        hoaSource.group = nullptr;

        LSDF_CHECK_REL(groupId.size() == 0,
            "HOA Group with ID '" + groupId + "' was referenced, but does not exist.");
    }
}

template <class DescType>
void LSDFParser::findPossibleParentTransform(XMLReader::Node& node, DescType& child)
{
    if (node.getParentName() == "Transform") {
        std::string transformId;
        parseAttribute(node.getParentNode(), "id", transformId, true);

        auto transformIt = std::find_if(transforms.begin(), transforms.end(),
            [&transformId](const TransformDesc& transform) { return transform.id == transformId; });

        LSDF_CHECK_REL(transformIt != transforms.end(),
            "Impossible error while parsing the parent transform of '" + child.id + "'");
        child.parentTransform = &(*transformIt);
    }
}

template <class DescType>
void LSDFParser::findPossibleParentAnchor(XMLReader::Node& node, DescType& child)
{
    if (node.getParentName() == "Anchor") {
        std::string anchorId;
        parseAttribute(node.getParentNode(), "id", anchorId, true);

        auto anchorIt = std::find_if(anchors.begin(), anchors.end(),
            [&anchorId](const AnchorDesc& anchor) { return anchor.id == anchorId; });

        LSDF_CHECK_REL(anchorIt != anchors.end(),
            "Impossible error while parsing the parent anchor of '" + child.id + "'");
        child.parentAnchor = &(*anchorIt);
    }
}

std::vector<std::string> LSDFParser::splitString(const std::string& s, char delimiter) const
{
    std::vector<std::string> tokens;
    std::string token;
    std::istringstream tokenStream(s);

    while (std::getline(tokenStream, token, delimiter)) {
        tokens.push_back(token);
    }

    return tokens;
}

void LSDFParser::ltrim(std::string& s)
{
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](int ch) { return !std::isspace(ch); }));
}

void LSDFParser::rtrim(std::string& s)
{
    s.erase(std::find_if(s.rbegin(), s.rend(), [](int ch) { return !std::isspace(ch); }).base(),
        s.end());
}

void LSDFParser::trim(std::string& s)
{
    ltrim(s);
    rtrim(s);
}

} // namespace LSDF
