/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <LSDF/XMLReader.h>

#include <LSDF/Exception.h>

#include <tinyxml2.h>

#include <algorithm>
#include <cstddef>
#include <sstream>

namespace LSDF {

using namespace tinyxml2;

std::string XMLReader::Node::getParentName() const
{
    auto parentPtr = ptr->Parent()->ToElement();
    return (parentPtr != nullptr) ? parentPtr->Name() : "";
};

XMLReader::Node XMLReader::Node::getParentNode() const
{
    return Node(ptr->Parent()->ToElement());
};

XMLReader::XMLReader(const std::string& file)
{
    xml = std::make_unique<XMLDocument>();
    LSDF_CHECK_REL(xml->LoadFile(file.c_str()) == XML_SUCCESS,
        "The XML file \"" + file + "\" could not be opened. TinyXML2 error: " + xml->ErrorStr());

    rootNode = xml->RootElement();
}

XMLReader::~XMLReader() {}

void XMLReader::query(XMLReader::Query& search)
{
    search.placeholders.clear();
    switch (search.type) {
    case SearchType::String: {
        const char* stringResult;
        search.found =
            (search.node.ptr->QueryStringAttribute(search.attributeName.c_str(), &stringResult)
                == XML_SUCCESS);

        if (search.found) {
            search.stringval = stringResult;
            trim(search.stringval);
        }
    } break;

    case SearchType::CoordinateTriple: {
        const char* stringResult;
        search.found =
            (search.node.ptr->QueryStringAttribute(search.attributeName.c_str(), &stringResult)
                == XML_SUCCESS);

        if (search.found) {
            try {
                parseCoordinateTriple(stringResult, search.triple, search.placeholders);
            } catch (...) {
                search.found = false;
                throw;
            }
        }
    } break;

    case SearchType::IntegerList: {
        const char* stringResult;
        search.found =
            (search.node.ptr->QueryStringAttribute(search.attributeName.c_str(), &stringResult)
                == XML_SUCCESS);

        if (search.found) {
            try {
                parseIntegerList(stringResult, search.intlist);
            } catch (...) {
                search.found = false;
                throw;
            }
        }
    } break;

    case SearchType::Integer: {
        search.found =
            (search.node.ptr->QueryIntAttribute(search.attributeName.c_str(), &search.intval)
                == XML_SUCCESS);
    } break;

    case SearchType::Float: {
        // cannot use native float parsing of tinyxml, because we can expect placeholders in gain attribute...
        const char* stringResult = nullptr;
        search.found =
            (search.node.ptr->QueryStringAttribute(search.attributeName.c_str(), &stringResult)
                == XML_SUCCESS);

        search.placeholders.resize(1);
        if (search.found) {
            try {
                std::string val(stringResult);
                trim(val);
                // check if placeholder is present
                try {
                    if (val.rfind("$", 0) == 0) {
                        search.floatval = 0.f;
                        search.placeholders[0] = std::stoi(val.substr(1));
                    } else {
                        search.floatval = std::stof(val);
                        search.placeholders[0] = -1;
                    }
                } catch (...) {
                    LSDF_FAIL("Invalid format for a float attribute.");
                }
            } catch (...) {
                search.found = false;
                throw;
            }
        }
    } break;

    case SearchType::Bool: {
        search.found =
            (search.node.ptr->QueryBoolAttribute(search.attributeName.c_str(), &search.boolval)
                == XML_SUCCESS);
    } break;

    default:
        search.found = false;
        LSDF_FAIL("Unknown SearchType in XMLReader::Query.");
    }
}

std::size_t XMLReader::nodeCount(const std::string& name)
{
    auto elements = getNodes(name);
    return elements.size();
}

std::vector<XMLReader::Node> XMLReader::getNodesRecursive(const std::string& name)
{
    return findChildrenByName(name, rootNode);
}

std::vector<XMLReader::Node> XMLReader::findChildrenByName(
    const std::string& name, tinyxml2::XMLElement* parent)
{
    std::vector<XMLReader::Node> elements;
    XMLElement* currentNode = parent->FirstChildElement();

    while (currentNode != nullptr) {
        for (auto el : findChildrenByName(name, currentNode)) {
            elements.push_back(el);
        }

        if (name.compare(currentNode->Name()) == 0) {
            elements.push_back(XMLReader::Node(currentNode));
        }

        currentNode = currentNode->NextSiblingElement();
    }

    return elements;
}

std::vector<XMLReader::Node> XMLReader::getNodes(const std::string& name)
{
    std::vector<XMLReader::Node> elements;
    XMLElement* currentNode = rootNode->FirstChildElement(name.c_str());

    while (currentNode != nullptr) {
        elements.push_back(XMLReader::Node(currentNode));

        currentNode = currentNode->NextSiblingElement(name.c_str());
    }

    return elements;
}

std::vector<XMLReader::Node> XMLReader::getNodes(const std::string& name, XMLReader::Node parent)
{
    std::vector<XMLReader::Node> elements;
    XMLElement* currentNode = parent.ptr->FirstChildElement(name.c_str());

    while (currentNode != nullptr) {
        elements.push_back(XMLReader::Node(currentNode));

        currentNode = currentNode->NextSiblingElement(name.c_str());
    }

    return elements;
}

XMLReader::Node XMLReader::getRootNode()
{
    return XMLReader::Node(rootNode);
}

std::string XMLReader::getRootNodeName()
{
    return rootNode->Name();
}

void XMLReader::parseIntegerList(const char* input, std::vector<int>& list)
{
    std::istringstream stream(input);
    std::string val;

    while (std::getline(stream, val, ',')) {
        trim(val);

        std::string::size_type colonPos = val.find(':');
        if (colonPos != std::string::npos) {
            try {
                int start = std::stoi(val.substr(0, colonPos));
                int end = std::stoi(val.substr(colonPos + 1));
                for (int i = start; i <= end; i++) {
                    list.push_back(i);
                }
            } catch (...) {
                LSDF_FAIL("Invalid value in integer list.");
            }
        } else {
            try {
                list.push_back(std::stoi(val));
            } catch (...) {
                LSDF_FAIL("Invalid value in integer list.");
            }
        }
    }
}

void XMLReader::parseCoordinateTriple(const char* input, float* out, std::vector<int>& placeholders)
{
    std::string inputStr(input);
    trim(inputStr);
    std::istringstream stream(inputStr);
    std::string val;

    placeholders.resize(3);

    for (auto i = 0; i < 3; i++) {
        LSDF_CHECK_REL(std::getline(stream, val, ' '),
            "Invalid number of values for a coordinates attribute.");
        trim(val);

        // check if placeholder is present
        if (val.rfind("$", 0) == 0) {
            try {
                out[i] = 0.f;
                placeholders[i] = std::stoi(val.substr(1));
            } catch (...) {
                LSDF_FAIL("Invalid placeholder format for a coordinates attribute.");
            }
        } else {
            placeholders[i] = -1;

            try {
                out[i] = std::stof(val);
            } catch (...) {
                LSDF_FAIL("Invalid number format for a coordinates attribute.");
            }
        }
    }
}

void XMLReader::ltrim(std::string& s)
{
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](int ch) { return !std::isspace(ch); }));
}

void XMLReader::rtrim(std::string& s)
{
    s.erase(std::find_if(s.rbegin(), s.rend(), [](int ch) { return !std::isspace(ch); }).base(),
        s.end());
}

void XMLReader::trim(std::string& s)
{
    ltrim(s);
    rtrim(s);
}

} // namespace LSDF
