/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DIRECTIVITY_CODING
#define MPEGI_DIRECTIVITY_CODING

#include "flex_range_coder.h"
#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <memory>
#include <string>
#include <vector>

namespace MPEGI {

class DirectivityCoverQuant;

class DirectivityCover
{
public:
    float freq;

    float deg_step;
    float db_step;

    std::vector<float> cover;
    float max_cover;

    int int_per_90;
    int el_cnt;
    std::vector<int> az_cnt;
    int cover_width;

    void print(FILE* f = stdout, bool cover_only = false);
    void dump(FILE* f = stdout);

    void quantize(DirectivityCoverQuant& dcq, float db_step_req);
    void generateSphereGrid(float deg_step_req, int int_per_90_req = 0);

    float getDirectivityNearestPoint(float elevation, float azimuth) const;
    float getDirectivityInterpolated(float elevation, float azimuth) const;
};

class DirectivityCoverQuant
{
public:
    float freq;

    float deg_step;
    float db_step;

    std::vector<int> cover;

    int int_per_90;
    int el_cnt;
    std::vector<int> az_cnt;
    int cover_width;

    int db_step_idx;
    int min_val;
    int max_val;
    int min_pos_val; //!< minimum possible value for the current db_step
    int max_pos_val; //!< maximum possible value for the current db_step

    //! Used to store the optimal parameters found during estimation until the actual encoding
    int best_pred_order;
    float best_est_size;
    int best_min_res;
    int best_max_res;

    void print(FILE* f = stdout, bool cover_only = false);

    void dequantize(DirectivityCover& dc);
    void generateSphereGrid(int int_per_90_req);

    void circularResample(
        std::vector<int>& x, int start, int count, int new_count, std::vector<float>& xr);
    void predict(int pred_order, std::vector<int>& cover_res, DirectivityCoverQuant* prev);
    void unpredict(int pred_order, std::vector<int>& cover_res, DirectivityCoverQuant* prev);

    void encode(FlexRangeEncoder* fre, int freq_quant_type, DirectivityCoverQuant* prev);
    void decode(FlexRangeDecoder* frd, int freq_quant_type, DirectivityCoverQuant* prev);

    int encodeRawBaseline(FlexRangeEncoder* fre, int freq_quant_type, bool est_only);
    void decodeRawBaseline(FlexRangeDecoder* frd, int freq_quant_type);

    int encodeOptimizedV1(
        FlexRangeEncoder* fre, int freq_quant_type, bool est_only, DirectivityCoverQuant* prev);
    void decodeOptimizedV1(FlexRangeDecoder* frd, int freq_quant_type, DirectivityCoverQuant* prev);
};

//! Directivity cover list used for encoding, decoding, and accesing an entire directivity file
/*!
 * For encoding, see DirectivityEncodingTest.py for an example using the C wrapper in DirectivityEncodingWrapper.h
 * For decoding, after calling decodeFromBuffer function, getFrequency* and getDirectivity* functions are available
 */
class DirectivityCoverList
{
public:
    //! Add a new directivity sphere covering for the frequency freq to the list
    /*!
     * \param freq          frequency [Hz] (valid range 1.0 to 24000.0 inclusive)
     * \param deg_step_req  requested degree step [degrees] (valid range 2.0 to 90.0 inclusive)
     * \param cover_size    number of values in the sphere covering corresponding exactly to deg_step_req
     * \param cover         sphere covering values [dB] (see extractSphereCovering in DirectivityEncodingTest.py)
     * \param db_step_req   requested quantization step [dB] (valid range 0.25 to 3.0 inclusive)
     */
    void addDirectivityCover(
        float freq, float deg_step_req, int cover_size, float* cover, float db_step_req);

    //! Get the required buffer size that is always large enough to accommodate the entire encoded data
    /*!
     * \return              size in bytes of the required buffer size to be passed as buffer_data
     */
    int getRequiredBufferSize();

    //! Encode all the previously added directivity sphere coverings to the provided memory buffer
    /*!
     * \param buffer_size   size in bytes of the provided memory buffer
     * \param buffer_data   the provided memory buffer
     * \return              encoded data size in bytes, or its negated value if greater than buffer_size
     */
    int encodeToBuffer(int buffer_size, unsigned char* buffer_data);

    //! Decode the entire directivity sphere covering list from the provided memory buffer
    /*!
     * \param buffer_size   exact size in bytes of the encoded data contained in the provided memory buffer
     * \param buffer_data   the provided memory buffer
     * \return              whether the decoding was successful or an error occured during decoding
     */
    bool decodeFromBuffer(int buffer_size, unsigned char* buffer_data);

    //! Get the count of frequencies (the count of directivity sphere coverings)
    /*!
     * \return              count of frequencies (count of directivity sphere coverings)
     */
    int getFrequencyCount() const;

    //! Get the frequency corresponding to an index in the frequency list
    /*!
     * \param freq_idx      frequency index in the frequency list (directivity sphere covering list)
     * \return              frequency [Hz]
     */
    float getFrequency(int freq_idx) const;

    //! Get an approximate directivity response for a frequency index and azimuth and elevation angles
    /*!
     * It is fast and uses the nearest point on the sphere covering (constant time, division free)
     *
     * \param freq_idx      frequency index in the frequency list (directivity sphere covering list)
     * \param elevation     elevation angle [degrees] (valid range -90.0 to 90.0 inclusive)
     * \param azimuth       azimuth angle [degrees] (valid range 0.0 to 360.0 inclusive)
     * \return              directivity response [dB]
     */
    float getDirectivityNearestPoint(int freq_idx, float elevation, float azimuth) const;

    //! Get an accurate directivity response for a frequency index and azimuth and elevation angles
    /*!
     * It is slower and uses bilinear interpolation on the sphere covering (constant time, division free)
     *
     * \param freq_idx      frequency index in the frequency list (directivity sphere covering list)
     * \param elevation     elevation angle [degrees] (valid range -90.0 to 90.0 inclusive)
     * \param azimuth       azimuth angle [degrees] (valid range 0.0 to 360.0 inclusive)
     * \return              directivity response [dB]
     */
    float getDirectivityInterpolated(int freq_idx, float elevation, float azimuth) const;

    const std::vector<std::unique_ptr<DirectivityCover>>& getDirectivityCoverList() const;

    float getDirectivityMaxCover() const;

    //! Enable automated text dump of the entire directivity cover list after calling encodeToBuffer
    /*!
     * The file format of the generated file name is <file_prefix>-<suffix>.dir
     * where suffix is an integer number computed as (int) (list[0]->deg_step * 100.0f)
     *
     * \param file_prefix   prefix of the generated file name for the automated text dump
     */
    void enableAutomatedTextDump(const char* file_prefix);

    void print(FILE* f = stdout, bool cover_only = false);
    void dump(FILE* f = stdout);

private:
    std::vector<std::unique_ptr<DirectivityCover>> list;
    float max_cover; //!< overall maximum cover value

    bool enable_dump = false; //!< whether to enable automatied text dump after encoding
    std::string dump_file_prefix = ""; //!< file prefix used for the automated text dump
};

} // namespace MPEGI

#endif /* MPEGI_DIRECTIVITY_CODING */
