#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <algorithm>
#include <memory>
#include <optional>
#include <set>
#include <vector>

#include <MPEGI/GeometryCoding.h>

namespace py = pybind11;

struct Vertex
{
    Vertex(float _x, float _y, float _z)
        : x(_x)
        , y(_y)
        , z(_z)
    {}

    float x, y, z;
};

struct Face
{
    Face(std::optional<int> _m, int _v1, int _v2, int _v3)
        : matIdx(_m)
        , v1(_v1)
        , v2(_v2)
        , v3(_v3)
    {}

    std::optional<int> matIdx;
    int v1, v2, v3;
};

struct Mesh
{
    Mesh(std::vector<Vertex> v, std::vector<Face> f)
        : vertices(v)
        , faces(f)
    {}

    std::vector<Vertex> vertices;
    std::vector<Face> faces;
};

py::bytes encodeMesh(Mesh mesh, float quantStep, float vertexTol, float triTol, bool triRmNoMat)
{
    auto geoData = std::make_unique<MPEGI::GeometryData>();

    geoData->setEncoderOptions(quantStep, vertexTol, triTol,
        0.001, // not relevant for mesh coding
        triRmNoMat);

    int vertexCount = mesh.vertices.size();
    std::vector<float> x, y, z;
    for (auto vertex : mesh.vertices) {
        x.push_back(vertex.x);
        y.push_back(vertex.y);
        z.push_back(vertex.z);
    }

    int faceCount = mesh.faces.size();
    std::vector<int> materials;
    std::vector<int> v1, v2, v3;
    for (auto face : mesh.faces) {
        v1.push_back(face.v1);
        v2.push_back(face.v2);
        v3.push_back(face.v3);
        int matIdx = face.matIdx ? *face.matIdx : -1;
        materials.push_back(matIdx);
    }

    int sharedMaterialCount = *std::max_element(materials.begin(), materials.end()) + 1;
    geoData->addVertexList(vertexCount, x.data(), y.data(), z.data());
    geoData->addTriangleList(
        faceCount, v1.data(), v2.data(), v3.data(), sharedMaterialCount, materials.data());

    int requiredBufferSize = geoData->getRequiredBufferSize();
    auto buffer = std::make_unique<unsigned char[]>(requiredBufferSize);
    int actualBufferSize = geoData->encodeToBuffer(requiredBufferSize, buffer.get());
    const char* bufferPtr = (const char*)buffer.get();
    return py::bytes(bufferPtr, actualBufferSize);
}
