/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/CodingUtils.h>
#include <MPEGI/flex_range_coder.h>
#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstdlib>

using namespace std;

namespace MPEGI {

void encode_uint_compact(FlexRangeEncoder* fre, unsigned int n, int max_bits)
{
    assert((max_bits >= 1) && (max_bits <= 32));

    if (n <= 1) {
        fre_encode_uniform(fre, 0, max_bits); // special prefix of 0 used for values 0 and 1
        fre_encode_bits(fre, n, 1);
    } else {
        int bit_count = floor_log2(n); // number of bits after the most significant one bit
        assert(bit_count < max_bits);

        fre_encode_uniform(fre, bit_count, max_bits);
        fre_encode_bits_long(fre, n - (1u << bit_count), bit_count);
    }
}

unsigned int decode_uint_compact(FlexRangeDecoder* frd, int max_bits)
{
    assert((max_bits >= 1) && (max_bits <= 32));

    int n;
    int bit_count = frd_read_uniform(frd, max_bits);

    if (bit_count == 0) // special prefix of 0 used for values 0 and 1
    {
        n = frd_read_bits(frd, 1);
    } else {
        n = frd_read_bits_long(frd, bit_count) + (1u << bit_count);
    }

    return n;
}

float estimate_uint_compact(unsigned int n, int max_bits)
{
    assert((max_bits >= 1) && (max_bits <= 32));

    float est_bits = log2((float)max_bits);

    if (n <= 1) {
        est_bits += 1.0f;
    } else {
        int bit_count = floor_log2(n); // number of bits after the most significant one bit
        assert(bit_count < max_bits);

        est_bits += (float)bit_count;
    }

    return est_bits;
}

unsigned int interleave_int(int n)
{
    if (n < 0) {
        return 2 * (-(n + 1)) + 1;
    } else {
        return 2 * n;
    }
}

int deinterleave_int(unsigned int n)
{
    if ((n & 1) != 0) {
        return -(int)(n >> 1) - 1;
    } else {
        return (int)(n >> 1);
    }
}

void encode_int_magnitude_sign(FlexRangeEncoder* fre, int n, int bits)
{
    assert((bits >= 1) && (bits <= 32));
    // n must not be -(2 ^ 31), because it overflows in the sign-magnitude representation

    int magnitude = abs(n);
    int sign = (n < 0);

    fre_encode_bits_long(fre, magnitude, bits - 1);
    if (magnitude != 0) {
        fre_encode_bits(fre, sign, 1);
    }
}

int decode_int_magnitude_sign(FlexRangeDecoder* frd, int bits)
{
    assert((bits >= 1) && (bits <= 32));

    int magnitude = frd_read_bits_long(frd, bits - 1);

    if (magnitude != 0) {
        int sign = frd_read_bits(frd, 1);

        if (sign != 0) {
            magnitude = -magnitude;
        }
    }

    return magnitude;
}

void fre_write_byte_to_memory_buffer(void* output_state, unsigned char byte_val)
{
    ((MemoryBuffer*)output_state)->write_byte(byte_val);
}

int frd_read_byte_from_memory_buffer(void* input_state)
{
    return ((MemoryBuffer*)input_state)->read_byte();
}

} // namespace MPEGI
