/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <MPEGI/AdaptiveEstimator.h>
#include <MPEGI/CodingUtils.h>
#include <MPEGI/DirectivityCoding.h>
#include <MPEGI/flex_range_coder.h>
#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <memory>
#include <new>
#include <string>
#include <utility>
#include <vector>

#define DC_VERBOSE_RESULTS 0
#define DC_INTERNAL_DEVEL_OUTPUT 0
#define DC_INTERNAL_FORCE_SPECIFIC_CODING 0

using namespace std;

namespace MPEGI {

const int freq_1o_idx_min = -5; // -5 maps to 31.25
const int freq_1o_idx_max = 4; // 4 maps to 16000.00
const int freq_3o_idx_min = -17; // -17 maps to 19.69
const int freq_3o_idx_max = 13; // 13 maps to 20158.74
const int freq_6o_idx_min = -34; // -34 maps to 19.69
const int freq_6o_idx_max = 26; // 26 maps to 20158.74

int get_frequency_quant_type(vector<float> freqs)
{
    float max_abs_err_1o = 0.0f;
    float max_abs_err_3o = 0.0f;
    float max_abs_err_6o = 0.0f;

    for (int i = 0; i < (int)freqs.size(); ++i) {
        assert((freqs[i] >= 1.0f) && (freqs[i] <= 24000.0f));

        float freq_1o = log2(freqs[i] / 1000.0f);
        int freq_1o_idx = (int)lround(freq_1o);
        freq_1o_idx = max(freq_1o_idx_min, min(freq_1o_idx, freq_1o_idx_max));
        max_abs_err_1o = max(max_abs_err_1o, abs(freq_1o - (float)freq_1o_idx));

        float freq_3o = 3.0f * freq_1o;
        int freq_3o_idx = (int)lround(freq_3o);
        freq_3o_idx = max(freq_3o_idx_min, min(freq_3o_idx, freq_3o_idx_max));
        max_abs_err_3o = max(max_abs_err_3o, abs(freq_3o - (float)freq_3o_idx));

        float freq_6o = 6.0f * freq_1o;
        int freq_6o_idx = (int)lround(freq_6o);
        freq_6o_idx = max(freq_6o_idx_min, min(freq_6o_idx, freq_6o_idx_max));
        max_abs_err_6o = max(max_abs_err_6o, abs(freq_6o - (float)freq_6o_idx));
    }

    // normalize to match the scale of max_abs_err_6o
    max_abs_err_3o *= 2.0f;
    max_abs_err_1o *= 6.0f;

    int freq_quant_type = 0;
    if (max_abs_err_6o > 0.5f) {
        freq_quant_type = 3; // round to integer quantization
    } else if ((max_abs_err_1o - max_abs_err_3o < 0.01f)
        && (max_abs_err_1o - max_abs_err_6o < 0.01f)) {
        freq_quant_type = 0; // octave quantization
    } else if (max_abs_err_3o - max_abs_err_6o < 0.01f) {
        freq_quant_type = 1; // 3rd octave quantization
    } else {
        freq_quant_type = 2; // 6th octave quantization
    }

#if DC_INTERNAL_DEVEL_OUTPUT
    printf("max_abs_err_1o %4.2f max_abs_err_3o %4.2f max_abs_err_6o %4.2f freq_quant_type %1d\n",
        max_abs_err_1o, max_abs_err_3o, max_abs_err_6o, freq_quant_type);
#endif

    return freq_quant_type;
}

void encode_quant_frequency(FlexRangeEncoder* fre, float freq, int quant_type)
{
    assert((freq >= 1.0f) && (freq <= 24000.0f));

    if (quant_type == 0) // octave quantization
    {
        float freq_1o = log2(freq / 1000.0f);
        int freq_idx = (int)lround(freq_1o);
        freq_idx = max(freq_1o_idx_min, min(freq_idx, freq_1o_idx_max));
        fre_encode_uniform(fre, freq_idx - freq_1o_idx_min, freq_1o_idx_max - freq_1o_idx_min + 1);
    } else if (quant_type == 1) // 3rd octave quantization
    {
        float freq_3o = 3.0f * log2(freq / 1000.0f);
        int freq_idx = (int)lround(freq_3o);
        freq_idx = max(freq_3o_idx_min, min(freq_idx, freq_3o_idx_max));
        fre_encode_uniform(fre, freq_idx - freq_3o_idx_min, freq_3o_idx_max - freq_3o_idx_min + 1);
    } else if (quant_type == 2) // 6th octave quantization
    {
        float freq_6o = 6.0f * log2(freq / 1000.0f);
        int freq_idx = (int)lround(freq_6o);
        freq_idx = max(freq_6o_idx_min, min(freq_idx, freq_6o_idx_max));
        fre_encode_uniform(fre, freq_idx - freq_6o_idx_min, freq_6o_idx_max - freq_6o_idx_min + 1);
    } else // round to integer quantization
    {
        int freq_idx = (int)lround(freq);
        fre_encode_uniform(fre, freq_idx - 1, 24000);
    }
}

float estimate_quant_frequency(int quant_type)
{
    if (quant_type == 0) // octave quantization
    {
        return log2((float)(freq_1o_idx_max - freq_1o_idx_min + 1));
    } else if (quant_type == 1) // 3rd octave quantization
    {
        return log2((float)(freq_3o_idx_max - freq_3o_idx_min + 1));
    } else if (quant_type == 2) // 6th octave quantization
    {
        return log2((float)(freq_6o_idx_max - freq_6o_idx_min + 1));
    } else // round to integer quantization
    {
        return log2(24000.0f);
    }
}

float decode_quant_frequency(FlexRangeDecoder* frd, int quant_type)
{
    float freq;

    if (quant_type == 0) // octave quantization
    {
        int freq_idx =
            frd_read_uniform(frd, freq_1o_idx_max - freq_1o_idx_min + 1) + freq_1o_idx_min;
        freq = 1000.0f * pow(2.0f, (float)freq_idx);
    } else if (quant_type == 1) // 3rd octave quantization
    {
        int freq_idx =
            frd_read_uniform(frd, freq_3o_idx_max - freq_3o_idx_min + 1) + freq_3o_idx_min;
        freq = 1000.0f * pow(2.0f, (float)freq_idx / 3.0f);
    } else if (quant_type == 2) // 6th octave quantization
    {
        int freq_idx =
            frd_read_uniform(frd, freq_6o_idx_max - freq_6o_idx_min + 1) + freq_6o_idx_min;
        freq = 1000.0f * pow(2.0f, (float)freq_idx / 6.0f);
    } else // round to integer quantization
    {
        int freq_idx = frd_read_uniform(frd, 24000) + 1;
        freq = (float)freq_idx;
    }

    return freq;
}

void DirectivityCover::print(FILE* f, bool cover_only)
{
    if (!cover_only) {
        fprintf(f, "freq %7.1f\n", freq);

        fprintf(f, "deg_step %5.2f\n", deg_step);
        fprintf(f, "db_step %4.2f\n", db_step);

        fprintf(f, "int_per_90 %2d\n", int_per_90);
        fprintf(f, "el_cnt %3d\n", el_cnt);
        fprintf(f, "az_cnt:\n");
        for (int k = 0; k < el_cnt; ++k) {
            fprintf(f, " %3d", az_cnt[k]);
        }
        fprintf(f, "\n");
    }

    fprintf(f, "cover:\n");
    for (int k = 0; k < el_cnt; ++k) {
        for (int p = 0; p < az_cnt[k]; ++p) {
            fprintf(f, " %+6.2f", cover[k * cover_width + p]);
        }
        fprintf(f, "\n");
    }
}

void DirectivityCover::dump(FILE* f)
{
    fprintf(f, "%7.1f\n", freq);
    fprintf(f, "%5.2f\n", deg_step);

    for (int ei = 0; ei < el_cnt; ++ei) {
        for (int ai = 0; ai < az_cnt[ei]; ++ai) {
            float val_db = cover[ei * cover_width + ai];
            float val = pow(10.0f, val_db / 20.0f);

            fprintf(f, " %9.6f", val);
        }
        fprintf(f, "\n");
    }
}

void DirectivityCover::quantize(DirectivityCoverQuant& dcq, float db_step_req)
{
    assert((db_step_req >= 0.25f) && (db_step_req <= 3.0f));
    assert((freq >= 1.0f) && (freq <= 24000.0f));

    dcq.freq = freq;

    dcq.deg_step = deg_step;
    dcq.db_step = db_step;

    dcq.int_per_90 = int_per_90;
    dcq.el_cnt = el_cnt;
    dcq.az_cnt = az_cnt;
    dcq.cover_width = cover_width;

    dcq.cover.assign(el_cnt * cover_width, INT_MIN); // use INT_MIN as an undefined value

    // quantize db_step_req in 0.25 dB increments, rounding down
    dcq.db_step_idx = (int)(db_step_req / 0.25f) - 1;
    // quantized integer values are from 0 to 11 inclusive
    dcq.db_step = (float)(dcq.db_step_idx + 1) * 0.25f;

    dcq.min_val = INT_MAX;
    dcq.max_val = INT_MIN;
    dcq.min_pos_val = (int)lround(-128.0f / dcq.db_step);
    dcq.max_pos_val = (int)lround(127.0f / dcq.db_step);

    for (int k = 0; k < el_cnt; ++k) {
        for (int p = 0; p < az_cnt[k]; ++p) {
            float val = cover[k * cover_width + p];
            assert((val >= -128.0f) && (val <= 127.0f));

            int quant_val = (int)lround(val / dcq.db_step);

            dcq.cover[k * cover_width + p] = quant_val;
            dcq.min_val = min(quant_val, dcq.min_val);
            dcq.max_val = max(quant_val, dcq.max_val);
        }
    }

    // reset the storage of optimal parameters found during estimation until the actual encoding
    dcq.best_pred_order = -1;
    dcq.best_est_size = (float)INT_MAX;
    dcq.best_min_res = INT_MAX;
    dcq.best_max_res = INT_MIN;
}

void DirectivityCover::generateSphereGrid(float deg_step_req, int int_per_90_req)
{
    const float pi_over_180 = acos(-1.0f) / 180.0f;

    if (deg_step_req == 0.0f) // use int_per_90_req instead of deg_step_req
    {
        assert((int_per_90_req >= 1) && (int_per_90_req <= 45));

        int_per_90 = int_per_90_req;
    } else {
        assert((deg_step_req >= 2.0f) && (deg_step_req <= 90.0f));

        // add 0.01 to avoid reaching the next finer grid if deg_step_req
        // is supplied already rounded down using 2 fractional digits
        int_per_90 = (int)ceil(90.0f / (deg_step_req + 0.01f));
    }
    deg_step = 90.0f / (float)int_per_90;

    el_cnt = 2 * int_per_90 + 1;
    az_cnt.assign(el_cnt, 0);
    cover_width = 4 * int_per_90; // maximum number of azimuth points (at equator)

    for (int k = 0; k < el_cnt; ++k) {
        float el_ang = (float)(k - int_per_90) * deg_step;
        float el_len = cos(el_ang * pi_over_180);
        az_cnt[k] = max((int)lround(el_len * 4.0f * (float)int_per_90), 1);
    }
}

float DirectivityCover::getDirectivityNearestPoint(float elevation, float azimuth) const
{
    const float one_over_180 = 1.0f / 180.0f;

    assert((-90.0f <= elevation) && (elevation <= 90.0f));
    assert((0.0f <= azimuth) && (azimuth <= 360.0f));

    // compute the closest elevation index
    int ei = (int)lround(((elevation + 90.0f) * one_over_180) * (float)(el_cnt - 1));

    // compute the closest azimuth index for the elevation index ei
    int ai = (int)lround((azimuth * (one_over_180 * 0.5f)) * (float)az_cnt[ei]);
    ai = (ai == az_cnt[ei] ? 0 : ai); // reduce modulo az_cnt[ei]

    float directivity = cover[ei * cover_width + ai];

    return directivity;
}

float DirectivityCover::getDirectivityInterpolated(float elevation, float azimuth) const
{
    const float one_over_180 = 1.0f / 180.0f;

    assert((-90.0f <= elevation) && (elevation <= 90.0f));
    assert((0.0f <= azimuth) && (azimuth <= 360.0f));

    // compute elevation indexes ei_down and ei_up that bracket the real valued elevation "index"
    float ei_real = ((elevation + 90.0f) * one_over_180) * (float)(el_cnt - 1);
    int ei_down = min((int)ei_real, el_cnt - 2);
    int ei_up = ei_down + 1;
    float ei_frac = ei_real - (float)ei_down; // compute fractional part

    // linearly interpolate directivity across azimuth for elevation index down (ei_down)
    int az_down_cnt = az_cnt[ei_down];
    float ai_down_real = (azimuth * (one_over_180 * 0.5f)) * (float)az_down_cnt;
    int ai_down_left = (int)ai_down_real;
    int ai_down_right = ai_down_left + 1;
    ai_down_right = (ai_down_right == az_down_cnt ? 0 : ai_down_right); // reduce modulo az_down_cnt
    float ai_down_frac = ai_down_real - (float)ai_down_left; // compute fractional part
    float directivity_low = (1.0f - ai_down_frac) * cover[ei_down * cover_width + ai_down_left]
        + ai_down_frac * cover[ei_down * cover_width + ai_down_right];

    // linearly interpolate directivity across azimuth for elevation index up (ei_up)
    int az_up_cnt = az_cnt[ei_up];
    float ai_up_real = (azimuth * (one_over_180 * 0.5f)) * (float)az_up_cnt;
    int ai_up_left = (int)ai_up_real;
    int ai_up_right = ai_up_left + 1;
    ai_up_right = (ai_up_right == az_up_cnt ? 0 : ai_up_right); // reduce modulo az_up_cnt
    float ai_up_frac = ai_up_real - (float)ai_up_left; // compute fractional part
    float directivity_up = (1.0f - ai_up_frac) * cover[ei_up * cover_width + ai_up_left]
        + ai_up_frac * cover[ei_up * cover_width + ai_up_right];

    // linearly interpolate the resulting directivities for ei_down and ei_up across elevation
    float directivity = (1.0f - ei_frac) * directivity_low + ei_frac * directivity_up;

    return directivity;
}

void DirectivityCoverQuant::print(FILE* f, bool cover_only)
{
    if (!cover_only) {
        fprintf(f, "freq %7.1f\n", freq);

        fprintf(f, "deg_step %5.2f\n", deg_step);
        fprintf(f, "db_step %4.2f\n", db_step);

        fprintf(f, "int_per_90 %2d\n", int_per_90);
        fprintf(f, "el_cnt %3d\n", el_cnt);
        fprintf(f, "az_cnt:\n");
        for (int k = 0; k < el_cnt; ++k) {
            fprintf(f, " %3d", az_cnt[k]);
        }
        fprintf(f, "\n");
    }

    fprintf(f, "cover_quant:\n");
    for (int k = 0; k < el_cnt; ++k) {
        for (int p = 0; p < az_cnt[k]; ++p) {
            fprintf(f, " %+3d", cover[k * cover_width + p]);
        }
        fprintf(f, "\n");
    }

    if (!cover_only) {
        fprintf(f, "min_val %+3d max_val %+3d\n", min_val, max_val);
    }
}

void DirectivityCoverQuant::dequantize(DirectivityCover& dc)
{
    dc.freq = freq;

    dc.deg_step = deg_step;
    dc.db_step = db_step;

    dc.int_per_90 = int_per_90;
    dc.el_cnt = el_cnt;
    dc.az_cnt = az_cnt;
    dc.cover_width = cover_width;

    dc.cover.assign(el_cnt * cover_width, 0.0f);
    dc.max_cover = -128.0f;

    for (int ei = 0; ei < el_cnt; ++ei) {
        for (int ai = 0; ai < az_cnt[ei]; ++ai) {
            dc.cover[ei * cover_width + ai] = (float)cover[ei * cover_width + ai] * db_step;
            dc.max_cover = std::max(dc.max_cover, dc.cover[ei * cover_width + ai]);
        }
    }
}

void DirectivityCoverQuant::generateSphereGrid(int int_per_90_req)
{
    DirectivityCover dc;

    dc.generateSphereGrid(0.0f, int_per_90_req);

    // copy the generated values from dc
    deg_step = dc.deg_step;
    int_per_90 = dc.int_per_90;
    el_cnt = dc.el_cnt;
    az_cnt = dc.az_cnt;
    cover_width = dc.cover_width;
}

void DirectivityCoverQuant::circularResample(
    vector<int>& x, int start, int count, int new_count, vector<float>& xr)
{
    xr.assign(new_count, 0);

    float scale = (float)count / (float)new_count;

    for (int idx = 0; idx < new_count; ++idx) {
        float pos = (float)idx * scale;
        int pos_left = (int)pos;

        int pos_right = pos_left + 1;
        pos_right = (pos_right == count ? 0 : pos_right); // reduce modulo count

        float frac = pos - (float)pos_left; // compute fractional part

        xr[idx] = (1.0f - frac) * (float)x[start + pos_left] + frac * (float)x[start + pos_right];
    }
}

void DirectivityCoverQuant::predict(
    int pred_order, vector<int>& cover_res, DirectivityCoverQuant* prev)
{
    cover_res.assign(el_cnt * cover_width, INT_MIN); // use INT_MIN as an undefined value

    // predict across frequency, using the values from the previous frequency, only when it exists
    if (pred_order == 5) {
        assert(prev != nullptr);
        assert((int_per_90 == prev->int_per_90) && (db_step_idx == prev->db_step_idx));

        for (int ei = 0; ei < el_cnt; ++ei) {
            for (int ai = 0; ai < az_cnt[ei]; ++ai) {
                cover_res[ei * cover_width + ai] =
                    cover[ei * cover_width + ai] - prev->cover[ei * cover_width + ai];
            }
        }

        return;
    }

    // copy the original value at the South pole, cover[0], which is not predicted
    cover_res[0] = cover[0];

    // predict vertically, from the one after the South pole to the North pole, at azimuth 0
    for (int ei = 1; ei < el_cnt; ++ei) {
        int pred_v = 0;

        if ((pred_order == 1) || (ei == 1) || (pred_order == 3)) {
            pred_v = cover[(ei - 1) * cover_width];
        } else if ((pred_order == 2) || (pred_order == 4)) {
            pred_v = 2 * cover[(ei - 1) * cover_width] - cover[(ei - 2) * cover_width];
        } else {
            assert(false);
        }

        cover_res[ei * cover_width] = cover[ei * cover_width] - pred_v;

        // always use true order 1 or true order 2 horizontal prediction at the equator
        if (((pred_order == 3) || (pred_order == 4)) && (ei != int_per_90))
            continue;

        // predict horizontally, from azimuth 0 to the maximum azimuth
        for (int ai = 1; ai < az_cnt[ei]; ++ai) {
            int pred_h = 0;

            if ((pred_order == 1) || (ai == 1) || (pred_order == 3)) {
                pred_h = cover[ei * cover_width + (ai - 1)];
            } else if ((pred_order == 2) || (pred_order == 4)) {
                pred_h =
                    2 * cover[ei * cover_width + (ai - 1)] - cover[ei * cover_width + (ai - 2)];
            } else {
                assert(false);
            }

            cover_res[ei * cover_width + ai] = cover[ei * cover_width + ai] - pred_h;
        }
    }

    if ((pred_order == 3) || (pred_order == 4)) {
        vector<float> c_resample;
        c_resample.reserve(cover_width);

        // predict horizontally for each elevation, from the one following the equator to the South pole
        for (int ei = int_per_90 - 1; ei >= 1; --ei) {
            circularResample(cover, (ei + 1) * cover_width, az_cnt[ei + 1], az_cnt[ei], c_resample);

            for (int ai = 1; ai < az_cnt[ei]; ++ai) {
                float pred_h = (float)cover[ei * cover_width + (ai - 1)]
                    + (c_resample[ai] - c_resample[ai - 1]);
                cover_res[ei * cover_width + ai] =
                    cover[ei * cover_width + ai] - (int)lrint(pred_h);
            }
        }

        // predict horizontally for each elevation, from the one following the equator to the North pole
        for (int ei = int_per_90 + 1; ei < el_cnt - 1; ++ei) {
            circularResample(cover, (ei - 1) * cover_width, az_cnt[ei - 1], az_cnt[ei], c_resample);

            for (int ai = 1; ai < az_cnt[ei]; ++ai) {
                float pred_h = (float)cover[ei * cover_width + (ai - 1)]
                    + (c_resample[ai] - c_resample[ai - 1]);
                cover_res[ei * cover_width + ai] =
                    cover[ei * cover_width + ai] - (int)lrint(pred_h);
            }
        }
    }
}

void DirectivityCoverQuant::unpredict(
    int pred_order, vector<int>& cover_res, DirectivityCoverQuant* prev)
{
    cover.assign(el_cnt * cover_width, INT_MIN); // use INT_MIN as an undefined value

    // predict across frequency, using the values from the previous frequency, only when it exists
    if (pred_order == 5) {
        assert(prev != nullptr);
        assert((int_per_90 == prev->int_per_90) && (db_step_idx == prev->db_step_idx));

        for (int ei = 0; ei < el_cnt; ++ei) {
            for (int ai = 0; ai < az_cnt[ei]; ++ai) {
                cover[ei * cover_width + ai] =
                    cover_res[ei * cover_width + ai] + prev->cover[ei * cover_width + ai];
            }
        }

        return;
    }

    // copy the original value at the South pole, cover_res[0], which is not predicted
    cover[0] = cover_res[0];

    // predict vertically, from the one after the South pole to the North pole, at azimuth 0
    for (int ei = 1; ei < el_cnt; ++ei) {
        int pred_v = 0;

        if ((pred_order == 1) || (ei == 1) || (pred_order == 3)) {
            pred_v = cover[(ei - 1) * cover_width];
        } else if ((pred_order == 2) || (pred_order == 4)) {
            pred_v = 2 * cover[(ei - 1) * cover_width] - cover[(ei - 2) * cover_width];
        } else {
            assert(false);
        }

        cover[ei * cover_width] = cover_res[ei * cover_width] + pred_v;

        // always use true order 1 or true order 2 horizontal prediction at the equator
        if (((pred_order == 3) || (pred_order == 4)) && (ei != int_per_90))
            continue;

        // predict horizontally, from azimuth 0 to the maximum azimuth
        for (int ai = 1; ai < az_cnt[ei]; ++ai) {
            int pred_h = 0;

            if ((pred_order == 1) || (ai == 1) || (pred_order == 3)) {
                pred_h = cover[ei * cover_width + (ai - 1)];
            } else if ((pred_order == 2) || (pred_order == 4)) {
                pred_h =
                    2 * cover[ei * cover_width + (ai - 1)] - cover[ei * cover_width + (ai - 2)];
            } else {
                assert(false);
            }

            cover[ei * cover_width + ai] = cover_res[ei * cover_width + ai] + pred_h;
        }
    }

    if ((pred_order == 3) || (pred_order == 4)) {
        vector<float> c_resample;
        c_resample.reserve(cover_width);

        // predict horizontally for each elevation, from the one following the equator to the South pole
        for (int ei = int_per_90 - 1; ei >= 1; --ei) {
            circularResample(cover, (ei + 1) * cover_width, az_cnt[ei + 1], az_cnt[ei], c_resample);

            for (int ai = 1; ai < az_cnt[ei]; ++ai) {
                float pred_h = (float)cover[ei * cover_width + (ai - 1)]
                    + (c_resample[ai] - c_resample[ai - 1]);
                cover[ei * cover_width + ai] =
                    cover_res[ei * cover_width + ai] + (int)lrint(pred_h);
            }
        }

        // predict horizontally for each elevation, from the one following the equator to the North pole
        for (int ei = int_per_90 + 1; ei < el_cnt - 1; ++ei) {
            circularResample(cover, (ei - 1) * cover_width, az_cnt[ei - 1], az_cnt[ei], c_resample);

            for (int ai = 1; ai < az_cnt[ei]; ++ai) {
                float pred_h = (float)cover[ei * cover_width + (ai - 1)]
                    + (c_resample[ai] - c_resample[ai - 1]);
                cover[ei * cover_width + ai] =
                    cover_res[ei * cover_width + ai] + (int)lrint(pred_h);
            }
        }
    }
}

void DirectivityCoverQuant::encode(
    FlexRangeEncoder* fre, int freq_quant_type, DirectivityCoverQuant* prev)
{
    assert((min_pos_val >= -512) && (max_pos_val <= 511));
    assert((min_val >= min_pos_val) && (min_val <= max_pos_val));
    assert((max_val >= min_pos_val) && (max_val <= max_pos_val));
    assert(min_val <= max_val);

    assert((freq >= 1.0f) && (freq <= 24000.0f));
    assert((db_step_idx >= 0) && (db_step_idx <= 11));
    assert((int_per_90 >= 1) && (int_per_90 <= 45));

    int est_bits_rb = encodeRawBaseline(fre, freq_quant_type, true);
    int est_bits_o1 = encodeOptimizedV1(fre, freq_quant_type, true, prev);
#if DC_INTERNAL_DEVEL_OUTPUT
    printf(" est_bits_rb %5d est_bits_o1 %5d\n", est_bits_rb, est_bits_o1);
#endif

#if DC_INTERNAL_FORCE_SPECIFIC_CODING == 1
    est_bits_o1 = INT_MAX;
#elif DC_INTERNAL_FORCE_SPECIFIC_CODING == 2
    est_bits_rb = INT_MAX;
#endif

    if (est_bits_rb <= est_bits_o1) {
        fre_encode_bits(fre, 1, 1); // use_raw_baseline flag
        encodeRawBaseline(fre, freq_quant_type, false);
    } else {
        fre_encode_bits(fre, 0, 1); // use_raw_coding flag
        encodeOptimizedV1(fre, freq_quant_type, false, prev);
    }
}

int DirectivityCoverQuant::encodeRawBaseline(
    FlexRangeEncoder* fre, int freq_quant_type, bool est_only)
{
    int alph_max = max_pos_val - min_pos_val + 1;
    int alph = max_val - min_val + 1;

    if (!est_only) {
        encode_quant_frequency(fre, freq, freq_quant_type);

        fre_encode_uniform(fre, db_step_idx, 12);
        fre_encode_uniform(fre, int_per_90 - 1, 45);

        fre_encode_uniform(fre, min_val - min_pos_val, alph_max);
        fre_encode_uniform(fre, max_val - min_pos_val, alph_max);

        for (int k = 0; k < el_cnt; ++k) {
            for (int p = 0; p < az_cnt[k]; ++p) {
                int val = cover[k * cover_width + p] - min_val;

                fre_encode_uniform(fre, val, alph);
            }
        }
    }

    int total_points_count = 0;
    for (int k = 0; k < el_cnt; ++k) {
        total_points_count += az_cnt[k];
    }

    float est_bits = estimate_quant_frequency(freq_quant_type) + log2(12.0f) + log2(45.0f)
        + 2.0f * log2((float)alph_max) + log2((float)alph) * (float)total_points_count;

    return (int)ceil(est_bits);
}

float get_kt_parameters_bit_cost(int a, int n)
{
    // a is the alphabet size and n is the number of symbols in the sequence
    assert(a >= 1);
    assert(n >= 0);

    // approximated as the bit cost of the all zero sequence, which has zero empirical entropy
    // the initial probability is: product((1 / 2 + c) / (a / 2 + c), c = 0 .. n - 1)
    // the closed form formula is: [gamma(1 / 2 + n) / gamma(1 / 2)] * [gamma(a / 2) / gamma(a / 2 + n)]
    float bit_cost = lgamma(0.5f + (float)n) - 0.5f * log(acos(-1.0f)) + lgamma((float)a * 0.5f)
        - lgamma((float)a * 0.5f + (float)n);

    // negate and convert to base 2 logarithm to obtain the result in bits
    bit_cost = -bit_cost * log2(exp(1.0f));

    return bit_cost;
}

float get_range_and_kt_bits(vector<int>& vec, int& min_val, int& max_val, bool skip_first)
{
    min_val = INT_MAX;
    max_val = INT_MIN;

    for (int idx = (int)skip_first; idx < (int)vec.size();
         ++idx) // skip the first value if it is not predicted
    {
        int val = vec[idx];
        if (val == INT_MIN)
            continue; // skip undefined values which are located outside of the cover

        min_val = min(val, min_val);
        max_val = max(val, max_val);
    }

    vector<int> cnt;
    int alph = max_val - min_val + 1;
    cnt.assign(alph, 0);
    int total_cnt = 0;

    for (int idx = (int)skip_first; idx < (int)vec.size();
         ++idx) // skip the first value if it is not predicted
    {
        int val = vec[idx];
        if (val == INT_MIN)
            continue; // skip undefined values which are located outside of the cover

        cnt[val - min_val]++;
        total_cnt++;
    }

    float ent = 0.0f;
    float inv_total_cnt = 1.0f / (float)total_cnt;

    for (int sym = min_val; sym <= max_val; ++sym) {
        float p = (float)cnt[sym - min_val] * inv_total_cnt;

        if (p != 0.0f) {
            ent -= p * log2(p);
        }
    }

    // KT estimator approximate bit cost for entropy ent, total count total_cnt, and alphabet alph
    float kt_bits = ent * (float)total_cnt + get_kt_parameters_bit_cost(alph, total_cnt);

    return kt_bits;
}

int DirectivityCoverQuant::encodeOptimizedV1(
    FlexRangeEncoder* fre, int freq_quant_type, bool est_only, DirectivityCoverQuant* prev)
{
    int alph_max = max_pos_val - min_pos_val + 1;

    vector<int> cover_res;

    if (est_only) {
        best_pred_order = -1;
        best_est_size = (float)INT_MAX;
        best_min_res = INT_MAX;
        best_max_res = INT_MIN;
    }

    int total_points_count = 0;
    for (int k = 0; k < el_cnt; ++k) {
        total_points_count += az_cnt[k];
    }

    if (best_pred_order == -1) // during est_only or actual encoding without a previous est_only
    {
        for (int pred_order = 1; pred_order <= 5; ++pred_order) {
            if (pred_order == 5) {
                // skip prediction across frequency if prev is unavailable, like for the first frequency
                if (prev == nullptr)
                    continue;
                // also skip if the quantized deg_step (int_per_90) or db_step (db_step_idx) are different
                if ((int_per_90 != prev->int_per_90) || (db_step_idx != prev->db_step_idx))
                    continue;
            }

            predict(pred_order, cover_res, prev);

            // compute minimum and maximum of prediction residual and KT bits approximation
            int min_res;
            int max_res;

            bool skip_first = (pred_order != 5);
            float kt_bits = get_range_and_kt_bits(cover_res, min_res, max_res, skip_first);
#if DC_INTERNAL_DEVEL_OUTPUT
            printf(" pred_order %d alph %3d min_res %+3d max_res %+3d kt_bits %7.1f bps %4.2f\n",
                pred_order, max_res - min_res + 1, min_res, max_res, kt_bits,
                kt_bits / (float)(total_points_count - (int)skip_first));
#endif

            float est_size = (float)skip_first * log2((float)alph_max)
                + estimate_uint_compact(interleave_int(min_res), 10)
                + estimate_uint_compact(interleave_int(max_res), 10) + kt_bits;

            if (est_size < best_est_size) {
                best_pred_order = pred_order;
                best_est_size = est_size;
                best_min_res = min_res;
                best_max_res = max_res;
            }
        }
    }

    if (!est_only) {
        // recompute the prediction residual using the best prediction order
        predict(best_pred_order, cover_res, prev);

        int alph = best_max_res - best_min_res + 1;

        encode_quant_frequency(fre, freq, freq_quant_type);

        fre_encode_uniform(fre, db_step_idx, 12);
        fre_encode_uniform(fre, int_per_90 - 1, 45);

        fre_encode_uniform(fre, best_pred_order - 1, 5);

        bool skip_first = (best_pred_order != 5);
        if (skip_first) {
            // encode the original value at the South pole, cover_res[0], which is not predicted
            fre_encode_uniform(fre, cover_res[0] - min_pos_val, alph_max);
        }

        // encode best_min_res and best_max_res, which always fit into 10-bit signed integers
        encode_uint_compact(fre, interleave_int(best_min_res), 10);
        encode_uint_compact(fre, interleave_int(best_max_res), 10);

        AdaptiveEstimator ae;
        ae.create(alph, 32768);

        // elevation 0, at the South pole, has only one azimuth
        for (int ei = (int)skip_first; ei < el_cnt; ++ei) {
            for (int ai = 0; ai < az_cnt[ei]; ++ai) {
                int res = cover_res[ei * cover_width + ai];
                assert((best_min_res <= res) && (res <= best_max_res));

                ae.encodeUpdate(res - best_min_res, fre);
            }
        }
    }

    float est_bits = estimate_quant_frequency(freq_quant_type) + log2(12.0f) + log2(45.0f)
        + log2(5.0f) + best_est_size;

    return (int)ceil(est_bits);
}

void DirectivityCoverQuant::decode(
    FlexRangeDecoder* frd, int freq_quant_type, DirectivityCoverQuant* prev)
{
    int use_raw_baseline = frd_read_bit(frd);

    if (use_raw_baseline) {
        decodeRawBaseline(frd, freq_quant_type);
    } else {
        decodeOptimizedV1(frd, freq_quant_type, prev);
    }

    // compute minimum and maximum
    min_val = INT_MAX;
    max_val = INT_MIN;

    for (int k = 0; k < el_cnt; ++k) {
        for (int p = 0; p < az_cnt[k]; ++p) {
            int quant_val = cover[k * cover_width + p];

            min_val = min(quant_val, min_val);
            max_val = max(quant_val, max_val);
        }
    }
}

void DirectivityCoverQuant::decodeRawBaseline(FlexRangeDecoder* frd, int freq_quant_type)
{
    freq = decode_quant_frequency(frd, freq_quant_type);

    db_step_idx = frd_read_uniform(frd, 12);
    db_step = (float)(db_step_idx + 1) * 0.25f;
    int_per_90 = frd_read_uniform(frd, 45) + 1;
    generateSphereGrid(int_per_90);

    cover.assign(el_cnt * cover_width, INT_MIN); // use INT_MIN as an undefined value

    min_pos_val = (int)lround(-128.0f / db_step);
    max_pos_val = (int)lround(127.0f / db_step);

    min_val = frd_read_uniform(frd, max_pos_val - min_pos_val + 1) + min_pos_val;
    max_val = frd_read_uniform(frd, max_pos_val - min_pos_val + 1) + min_pos_val;

    int alph = max_val - min_val + 1;

    for (int k = 0; k < el_cnt; ++k) {
        for (int p = 0; p < az_cnt[k]; ++p) {
            int val = frd_read_uniform(frd, alph);

            cover[k * cover_width + p] = val + min_val;
        }
    }
}

void DirectivityCoverQuant::decodeOptimizedV1(
    FlexRangeDecoder* frd, int freq_quant_type, DirectivityCoverQuant* prev)
{
    freq = decode_quant_frequency(frd, freq_quant_type);

    db_step_idx = frd_read_uniform(frd, 12);
    db_step = (float)(db_step_idx + 1) * 0.25f;
    int_per_90 = frd_read_uniform(frd, 45) + 1;
    generateSphereGrid(int_per_90);

    int pred_order = frd_read_uniform(frd, 5) + 1;
    if (pred_order == 5) {
        assert(prev != nullptr);
        assert((int_per_90 == prev->int_per_90) && (db_step_idx == prev->db_step_idx));
        // TODO: should be be extended to return a bitstream error instead of triggering an assert
    }

    min_pos_val = (int)lround(-128.0f / db_step);
    max_pos_val = (int)lround(127.0f / db_step);
    int alph_max = max_pos_val - min_pos_val + 1;

    vector<int> cover_res;
    cover_res.assign(el_cnt * cover_width, INT_MIN); // use INT_MIN as an undefined value

    bool skip_first = (pred_order != 5);
    if (skip_first) {
        // decode the original value at the South pole, cover_res[0], which is not predicted
        cover_res[0] = frd_read_uniform(frd, alph_max) + min_pos_val;
    }

    // decode min_res and max_res, which always fit into 10-bit signed integers
    int min_res = deinterleave_int(decode_uint_compact(frd, 10));
    int max_res = deinterleave_int(decode_uint_compact(frd, 10));
    int alph = max_res - min_res + 1;

    AdaptiveEstimator ae;
    ae.create(alph, 32768);
    // elevation 0, at the South pole, has only one azimuth
    for (int ei = (int)skip_first; ei < el_cnt; ++ei) {
        for (int ai = 0; ai < az_cnt[ei]; ++ai) {
            int res = ae.decodeUpdate(frd) + min_res;

            assert((min_res <= res) && (res <= max_res));
            cover_res[ei * cover_width + ai] = res;
        }
    }

    // reconstruct the original values by undoing the prediction
    unpredict(pred_order, cover_res, prev);
}

void DirectivityCoverList::addDirectivityCover(
    float freq, float deg_step_req, int cover_size, float* cover, float db_step_req)
{
    auto dc = make_unique<DirectivityCover>();

    dc->freq = freq;

    dc->deg_step = deg_step_req;
    dc->db_step = db_step_req;

    dc->generateSphereGrid(deg_step_req);
    dc->cover.assign(dc->el_cnt * dc->cover_width, 0.0f);

    int pos = 0;
    for (int k = 0; k < dc->el_cnt; ++k) {
        for (int p = 0; p < dc->az_cnt[k]; ++p) {
            dc->cover[k * dc->cover_width + p] = cover[pos++];
        }
    }
    assert(pos == cover_size);

    list.push_back(move(dc));
}

int DirectivityCoverList::getRequiredBufferSize()
{
    int total_points_count = 0;

    for (int i = 0; i < (int)list.size(); ++i) {
        DirectivityCover& dc = *(list[i]);

        // add the azimuth count of points of each elevation for the current directivity cover
        for (int ei = 0; ei < dc.el_cnt; ++ei) {
            total_points_count += dc.az_cnt[ei];
        }
    }

    // compute the maximum theoretically possible required buffer size, including a safety margin
    int max_buffer_size = (int)list.size() * 8 + (total_points_count * 9) / 8 + 16;

    return max_buffer_size;
}

int DirectivityCoverList::encodeToBuffer(int buffer_size, unsigned char* buffer_data)
{
    vector<unique_ptr<DirectivityCoverQuant>> list_dcq;
    vector<float> freqs;

    for (int i = 0; i < (int)list.size(); ++i) {
        DirectivityCover& dc = *(list[i]);
        auto dcq = make_unique<DirectivityCoverQuant>();

        dc.quantize(*dcq, dc.db_step);

        freqs.push_back(dcq->freq);
        list_dcq.push_back(move(dcq));
    }

    // determine the optimal common quantization type to be used for all frequencies
    int freq_quant_type = get_frequency_quant_type(freqs);

    FlexRangeEncoder fre;
    MemoryBuffer mb;

    mb.setup(buffer_size, buffer_data);
    fre_init(&fre, fre_write_byte_to_memory_buffer, &mb);

    assert((int)list.size() < 64);
    fre_encode_bits(&fre, (int)list.size(), 6);

    fre_encode_bits(&fre, freq_quant_type, 2);

    int last_bits = fre_virtual_finish(&fre);

    for (int i = 0; i < (int)list.size(); ++i) {
        DirectivityCoverQuant& dcq = *(list_dcq[i]);
        DirectivityCoverQuant* dcq_prev = (i >= 1 ? list_dcq[i - 1].get() : nullptr);

        dcq.encode(&fre, freq_quant_type, dcq_prev);

        int current_bits = fre_virtual_finish(&fre);

#if DC_VERBOSE_RESULTS
        printf("[DirectivityCoverList::encodeToBuffer] idx %2d bits %6d\n", i,
            current_bits - last_bits);
#else
        (void)last_bits;
#endif

        last_bits = current_bits;
    }

    fre_finish(&fre);

    // if enabled, do automated text dump of the entire directivity cover list
    if (enable_dump) {
        int suffix_value = (int)(list[0]->deg_step * 100.0f);
        string file_name = dump_file_prefix + "-" + to_string(suffix_value) + ".dir";

#if DC_VERBOSE_RESULTS
        printf("[DirectivityCoverList::encodeToBuffer] generating text dump to %s\n",
            file_name.c_str());
#endif

        FILE* f = fopen(file_name.c_str(), "wb");
        assert(f != NULL);

        dump(f);

        fclose(f);
    }

    // in case of overflow, use a negative number to indicate the needed buffer size
    return !mb.has_overflown ? mb.pos : -mb.pos;
}

bool DirectivityCoverList::decodeFromBuffer(int buffer_size, unsigned char* buffer_data)
{
    vector<unique_ptr<DirectivityCoverQuant>> list_dcq;

    FlexRangeDecoder frd;
    MemoryBuffer mb;

    mb.setup(buffer_size, buffer_data);
    frd_init(&frd, -1 /* not available */, frd_read_byte_from_memory_buffer, &mb);

    list.clear();

    int directivity_count = frd_read_bits(&frd, 6);
    list.reserve(directivity_count);
    list_dcq.reserve(directivity_count);

    int freq_quant_type = frd_read_bits(&frd, 2);

    for (int i = 0; i < directivity_count; ++i) {
        auto dc = make_unique<DirectivityCover>();
        auto dcq = make_unique<DirectivityCoverQuant>();
        DirectivityCoverQuant* dcq_prev = (i >= 1 ? list_dcq[i - 1].get() : nullptr);

        dcq->decode(&frd, freq_quant_type, dcq_prev);

        dcq->dequantize(*dc);

        max_cover = std::max(max_cover, dc->max_cover);

        list.push_back(move(dc));
        list_dcq.push_back(move(dcq));
    }

    frd_finish(&frd);

#if DC_VERBOSE_RESULTS
    printf(
        "[DirectivityCoverList::decodeFromBuffer] bit_error_detected %d\n", frd.bit_error_detected);
#endif

    return !frd.bit_error_detected;
}

int DirectivityCoverList::getFrequencyCount() const
{
    return (int)list.size();
}

float DirectivityCoverList::getFrequency(int freq_idx) const
{
    assert((0 <= freq_idx) && (freq_idx < (int)list.size()));

    return list[freq_idx]->freq;
}

float DirectivityCoverList::getDirectivityNearestPoint(
    int freq_idx, float elevation, float azimuth) const
{
    assert((0 <= freq_idx) && (freq_idx < (int)list.size()));
    assert((-90.0f <= elevation) && (elevation <= 90.0f));
    assert((0.0f <= azimuth) && (azimuth <= 360.0f));

    float directivity = list[freq_idx]->getDirectivityNearestPoint(elevation, azimuth);

    return directivity;
}

float DirectivityCoverList::getDirectivityInterpolated(
    int freq_idx, float elevation, float azimuth) const
{
    assert((0 <= freq_idx) && (freq_idx < (int)list.size()));
    assert((-90.0f <= elevation) && (elevation <= 90.0f));
    assert((0.0f <= azimuth) && (azimuth <= 360.0f));

    float directivity = list[freq_idx]->getDirectivityInterpolated(elevation, azimuth);

    return directivity;
}

const std::vector<std::unique_ptr<DirectivityCover>>&
DirectivityCoverList::getDirectivityCoverList() const
{
    return list;
}

float DirectivityCoverList::getDirectivityMaxCover() const
{
    return max_cover;
}

void DirectivityCoverList::enableAutomatedTextDump(const char* file_prefix)
{
    enable_dump = true;
    dump_file_prefix = file_prefix;

    // remove extension suffix, if present
    dump_file_prefix = dump_file_prefix.substr(0, dump_file_prefix.find_last_of('.'));
}

void DirectivityCoverList::print(FILE* f, bool cover_only)
{
    for (int i = 0; i < (int)list.size(); ++i) {
        if (!cover_only) {
            fprintf(f, "\nidx %2d\n", i);
        }

        list[i]->print(f, cover_only);
    }
}

void DirectivityCoverList::dump(FILE* f)
{
    fprintf(f, "%d\n", (int)list.size());

    for (int i = 0; i < (int)list.size(); ++i) {
        list[i]->dump(f);
    }
}

} // namespace MPEGI
