/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifdef _MSC_VER
#define _CRT_SECURE_NO_WARNINGS // remove fscanf and fopen warnings (used only in the internal test code)
#endif

#include <MPEGI/AdaptiveEstimator.h>
#include <MPEGI/CodingUtils.h>
#include <MPEGI/GeometryCoding.h>
#include <MPEGI/flex_range_coder.h>
#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <map>
#include <memory>
#include <new>
#include <string>
#include <utility>
#include <vector>

#define GC_PRINT_VERTEX_LIST_QUANT 0
#define GC_PRINT_VERTEX_DUPLICATES 0
#define GC_PRINT_TRIANGLE_DUPLICATES 0
#define GC_PRINT_TRIANGLE_DEGENERATES 0
#define GC_PRINT_BOUNDING_BOX_QUANT 0
#define GC_PRINT_PREPROCESSOR_GCD 0
#define GC_PRINT_VERTEX_LIST_UNIQUE 0
#define GC_PRINT_VERTEX_UNIQUE_MTF 0
#define GC_PRINT_INTERNAL_DIAGNOSTICS 0
#define GC_PRINT_MACHINE_STATISTICS 0
#define GC_VERBOSE_RESULTS 0

#define REMOVE_VERTEX_DUPLICATES 1
#define REMOVE_TRIANGLE_DUPLICATES 1
#define REMOVE_TRIANGLE_DEGENERATES 1

#define BBQ_CODING_OPTIMIZE_LEVEL 0 // 0 for raw baseline
#define ENABLE_PREPROCESS_GCD 1 // 0 (disabled) for raw baseline
#define PPGCD_CODING_OPTIMIZE_LEVEL 0 // 0 for raw baseline
#define VLQ_CODING_OPTIMIZE_LEVEL 1 // 0 for raw baseline
#define GC_DEBUG_ADD_TRAILING_BITS 1
#define TL_CODING_OPTIMIZE_LEVEL 1 // 0 for raw baseline
#define AM_CODING_OPTIMIZE_LEVEL 1 // 0 for raw baseline

// TODO: set a meaningful value!
#define MPEGI_MAX_GRIDS 128

using namespace std;

namespace MPEGI {

static float distance2(VertexList& vl, int u, int v)
{
    float dist2 = (vl.x[v] - vl.x[u]) * (vl.x[v] - vl.x[u])
        + (vl.y[v] - vl.y[u]) * (vl.y[v] - vl.y[u]) + (vl.z[v] - vl.z[u]) * (vl.z[v] - vl.z[u]);

    return dist2;
}

void GeometryData::check_and_fix_mesh(float tol_v, float tol_t)
{
    assert(tol_v >= 0.0f);
    assert(tol_t >= 0.0f);

    vector<bool> vl_is_duplicate;
    vl_is_duplicate.assign(vl.count, false);
    // vertex index remapping, without duplicates and pointing to originals
    vector<int> vl_index_remap;
    vl_index_remap.assign(vl.count, INT_MIN); // invalid index value (meaning not initialized)

    vl.duplicates_removed = 0;
    vl.exact_duplicates_removed = 0;

#if REMOVE_VERTEX_DUPLICATES
    float tol2_v = tol_v * tol_v;

    // for indexes of original vertexes, which replace the duplicates
    vector<int> vl_deduplicated_indexes;
    // invalid index value (meaning not a duplicate)
    vl_deduplicated_indexes.assign(vl.count, INT_MIN);
    vector<int> vl_updated_indexes; // for updated vertex indexes, after removal of duplicates
    vl_updated_indexes.assign(vl.count, INT_MIN); // invalid index value (meaning a duplicate)

    if (vl.count >= 1) {
        vl_updated_indexes[0] = 0;
    }

    for (int k = 1; k < vl.count; ++k) {
        for (int j = 0; j < k; ++j) {
            float dist2_jk = distance2(vl, j, k);
            if (dist2_jk <= tol2_v) {
                bool exact_duplicate = (dist2_jk == 0.0f);
#if GC_PRINT_VERTEX_DUPLICATES
                if (log != NULL)
                    fprintf(log,
                        "duplicate_vertex: %d (like %d) exact %u dist %.5f delta_x %.5f delta_y "
                        "%.5f delta_z %.5f\n",
                        k, j, exact_duplicate, dist2_jk, vl.x[j] - vl.x[k], vl.y[j] - vl.y[k],
                        vl.z[j] - vl.z[k]);
#endif

                vl_is_duplicate[k] = true;
                vl.duplicates_removed++;
                vl_deduplicated_indexes[k] = j; // vertex k is a duplicate of vertex j
                vl.exact_duplicates_removed += exact_duplicate;
                break;
            }
        }

        if (!vl_is_duplicate[k]) {
            // k - vl.duplicate_count will be the updated index of vertex k, after removal of duplicates
            vl_updated_indexes[k] = k - vl.duplicates_removed;
        }
    }

    if (vl.duplicates_removed != 0) {
#if GC_PRINT_VERTEX_DUPLICATES
        if (log != NULL)
            fprintf(log,
                "check_duplicate_vertexes: marked %d duplicates, from which %d exact duplicates\n",
                vl.duplicates_removed, vl.exact_duplicates_removed);
#endif
    }

    // generate vertex index remapping, without duplicates and pointing to originals
    for (int k = 0; k < vl.count; ++k) {
        if (vl_is_duplicate[k]) {
            assert(vl_deduplicated_indexes[k] != INT_MIN);
            assert(vl_updated_indexes[vl_deduplicated_indexes[k]] != INT_MIN);

            vl_index_remap[k] = vl_updated_indexes[vl_deduplicated_indexes[k]];
        } else { // !vl.is_duplicate[k]
            assert(vl_deduplicated_indexes[k] == INT_MIN);
            assert(vl_updated_indexes[k] != INT_MIN);

            vl_index_remap[k] = vl_updated_indexes[k];
        }

        // the remapped alphabet size is vl.count - vl.duplicate_count
        assert(vl_index_remap[k] < vl.count - vl.duplicates_removed);
    }
#endif

    vector<bool> tl_is_duplicate;
    tl_is_duplicate.assign(tl.count, false);
    tl.duplicates_removed = 0;
    tl.exact_duplicates_removed = 0;

#if REMOVE_TRIANGLE_DUPLICATES
    for (int k = 1; k < tl.count; ++k) {
        for (int j = 0; j < k; ++j) {
            if ((tl.v1[j] == tl.v1[k]) && (tl.v2[j] == tl.v2[k]) && (tl.v3[j] == tl.v3[k])) {
#if GC_PRINT_TRIANGLE_DUPLICATES
                if (log != NULL)
                    fprintf(log, "duplicate_triangle: %d (like %d)\n", k, j);
#endif

                tl_is_duplicate[k] = true;
                tl.duplicates_removed++;
                break;
            }
        }
    }

    if (tl.duplicates_removed != 0) {
#if GC_PRINT_TRIANGLE_DUPLICATES
        if (log != NULL)
            fprintf(
                log, "check_duplicate_triangles: marked %d duplicates\n", tl.duplicates_removed);
#endif
    }
#endif

    vector<bool> tl_is_degenerate;
    tl_is_degenerate.assign(tl.count, false);
    tl.degenerates_removed = 0;
    tl.exact_degenerates_removed = 0;

#if REMOVE_TRIANGLE_DEGENERATES
    float tol2_t = tol_t * tol_t;

    for (int k = 0; k < tl.count; ++k) {
        float d2_12 = distance2(vl, tl.v1[k], tl.v2[k]);
        float d2_23 = distance2(vl, tl.v2[k], tl.v3[k]);
        float d2_31 = distance2(vl, tl.v3[k], tl.v1[k]);

        if ((d2_12 <= tol2_t) || (d2_23 <= tol2_t) || (d2_31 <= tol2_t)) {
            bool exact_degenerate = ((d2_12 == 0.0f) || (d2_23 == 0.0f) || (d2_31 == 0.0f));
#if GC_PRINT_TRIANGLE_DEGENERATES
            if (log != NULL)
                fprintf(log,
                    "degenerate_triangle: %d vertexes %d %d %d exact %u d12 %.5f d23 %.5f d31 "
                    "%.5f\n",
                    k, tl.v1[k], tl.v2[k], tl.v3[k], exact_degenerate, sqrt(d2_12), sqrt(d2_23),
                    sqrt(d2_31));
#endif

            tl_is_degenerate[k] = true;
            tl.degenerates_removed++;
            tl.exact_degenerates_removed += exact_degenerate;
            break;
        }
    }

    if (tl.degenerates_removed != 0) {
#if GC_PRINT_TRIANGLE_DEGENERATES
        if (log != NULL)
            fprintf(log,
                "check_degenerate_triangles: marked %d degenerates, from which %d exact "
                "degenerates\n",
                tl.degenerates_removed, tl.exact_degenerates_removed);
#endif
    }
#endif

    // copy all the vertexes without duplicates into new vectors and then swap the vectors
    vector<float> vl_x_new;
    vector<float> vl_y_new;
    vector<float> vl_z_new;
    int vl_actual_count = vl.count - vl.duplicates_removed;

    vl_x_new.reserve(vl_actual_count);
    vl_y_new.reserve(vl_actual_count);
    vl_z_new.reserve(vl_actual_count);

    for (int k = 0; k < vl.count; ++k) {
        if (vl_is_duplicate[k])
            continue;

        vl_x_new.push_back(vl.x[k]);
        vl_y_new.push_back(vl.y[k]);
        vl_z_new.push_back(vl.z[k]);
    }

    vl.x.swap(vl_x_new);
    vl.y.swap(vl_y_new);
    vl.z.swap(vl_z_new);
    vl.count = vl_actual_count;

    // copy all the triangles without duplicates into new vectors and then swap the vectors
    vector<int> v1_new;
    vector<int> v2_new;
    vector<int> v3_new;
    vector<int> mi_new;
    int tl_actual_count = tl.count - tl.duplicates_removed - tl.degenerates_removed;

    v1_new.reserve(tl_actual_count);
    v2_new.reserve(tl_actual_count);
    v3_new.reserve(tl_actual_count);
    mi_new.reserve(tl_actual_count);

    for (int k = 0; k < tl.count; ++k) {
        if (tl_is_duplicate[k] || tl_is_degenerate[k])
            continue;

        v1_new.push_back(vl_index_remap[tl.v1[k]]);
        v2_new.push_back(vl_index_remap[tl.v2[k]]);
        v3_new.push_back(vl_index_remap[tl.v3[k]]);
        mi_new.push_back(tl.mi[k]);
    }

    tl.v1.swap(v1_new);
    tl.v2.swap(v2_new);
    tl.v3.swap(v3_new);
    tl.mi.swap(mi_new);
    tl.count = tl_actual_count;
}

void VertexList::print(FILE* f)
{
    fprintf(f, "%d\n", count);

    for (int k = 0; k < count; ++k) {
        fprintf(f, "%d %.4f %.4f %.4f\n", k, x[k], y[k], z[k]);
    }
}

void VertexListQuant::print(FILE* f)
{
    fprintf(f, "%d\n", count);

    for (int k = 0; k < count; ++k) {
        fprintf(f, "%d %d %d %d\n", k, x[k], y[k], z[k]);
    }
}

void VertexListQuant::compute_unique_coordinates(FILE* f)
{
    xu.assign(count, 0);
    yu.assign(count, 0);
    zu.assign(count, 0);

    map<int, int> xuc;
    map<int, int> yuc;
    map<int, int> zuc;

    int pos = 0;
    for (int k = 0; k < count; ++k) {
        xu[pos] = x[k];
        yu[pos] = y[k];
        zu[pos] = z[k];
        pos++;

        xuc[x[k]]++;
        yuc[y[k]]++;
        zuc[z[k]]++;
    }
    assert(pos == count);

    sort(xu.begin(), xu.end());
    vector<int>::iterator last = unique(xu.begin(), xu.end());
    xu.erase(last, xu.end());

    sort(yu.begin(), yu.end());
    last = unique(yu.begin(), yu.end());
    yu.erase(last, yu.end());

    sort(zu.begin(), zu.end());
    last = unique(zu.begin(), zu.end());
    zu.erase(last, zu.end());

    if (f != NULL) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(f, "vlq: cnt %5d xu_cnt %5lu yu_cnt %5lu zu_cnt %5lu\n", count, xu.size(),
            yu.size(), zu.size());
#endif

#if GC_PRINT_VERTEX_LIST_UNIQUE
        fprintf(f, "xuc:\n");
        for (const auto& pair : xuc) {
            fprintf(f, "%+7d %5d\n", pair.first, pair.second);
        }

        fprintf(f, "yuc:\n");
        for (const auto& pair : yuc) {
            fprintf(f, "%+7d %5d\n", pair.first, pair.second);
        }

        fprintf(f, "zuc:\n");
        for (const auto& pair : zuc) {
            fprintf(f, "%+7d %5d\n", pair.first, pair.second);
        }
#endif
    }
}

void TriangleList::print(FILE* f)
{
    fprintf(f, "%d\n", count);

    for (int k = 0; k < count; ++k) {
        string mat_name =
            (mi[k] == -1 ? string("mat:none") : string("mat:idx_") + to_string(mi[k]));

        fprintf(f, "%d %d %d %d \"%s\"\n", k, v1[k], v2[k], v3[k], mat_name.c_str());
    }
}

void AcousticMaterial::print(FILE* f)
{
    fprintf(f, "\"%s\"\n", name.c_str());
    fprintf(f, "freq grid index: %d; count: %d\n", freq_grid_index, freq_count);

    for (int k = 0; k < freq_count; ++k) {
        fprintf(f, "%.3f %.3f %.3f %.3f\n", c1[k], c2[k], c3[k], c4[k]);
    }
}

void AcousticMaterialList::print(FILE* f)
{
    fprintf(f, "%d\n", count);

    for (int k = 0; k < (int)list.size(); ++k) {
        list[k]->print(f);
        fprintf(f, "\n");
    }
}

void VertexList::quantize(VertexListQuant& vlq, float step)
{
    assert(step > 0.0f);

    vlq.count = count;
    vlq.step = step;

    vlq.x.assign(count, 0);
    vlq.y.assign(count, 0);
    vlq.z.assign(count, 0);

    vlq.duplicates_removed = duplicates_removed;
    vlq.exact_duplicates_removed = exact_duplicates_removed;

    for (int k = 0; k < count; ++k) {
        // quantize all vertexes
        vlq.x[k] = (int)lround(x[k] / step);
        vlq.y[k] = (int)lround(y[k] / step);
        vlq.z[k] = (int)lround(z[k] / step);
    }

    vlq.xu.clear();
    vlq.yu.clear();
    vlq.zu.clear();
}

void VertexListQuant::dequantize(VertexList& vl)
{
    assert(step > 0.0f);
    assert(duplicates_removed == 0); // no duplicates are generated during decoding
    assert(exact_duplicates_removed == 0); // no duplicates are generated during decoding

    vl.count = count;

    vl.x.assign(count, 0.0f);
    vl.y.assign(count, 0.0f);
    vl.z.assign(count, 0.0f);

    vl.duplicates_removed = 0;
    vl.exact_duplicates_removed = 0;

    for (int k = 0; k < count; ++k) {
        // dequantize all vertexes
        vl.x[k] = (float)x[k] * step;
        vl.y[k] = (float)y[k] * step;
        vl.z[k] = (float)z[k] * step;
    }
}

void BoundingBoxQuant::compute(VertexListQuant& vlq)
{
    if (vlq.count == 0) {
        x_min = 0;
        x_max = 0;
        y_min = 0;
        y_max = 0;
        z_min = 0;
        z_max = 0;

        return;
    }

    x_min = vlq.x[0];
    x_max = vlq.x[0];
    y_min = vlq.y[0];
    y_max = vlq.y[0];
    z_min = vlq.z[0];
    z_max = vlq.z[0];

    for (int k = 1; k < vlq.count; ++k) {
        x_min = min(x_min, vlq.x[k]);
        x_max = max(x_max, vlq.x[k]);
        y_min = min(y_min, vlq.y[k]);
        y_max = max(y_max, vlq.y[k]);
        z_min = min(z_min, vlq.z[k]);
        z_max = max(z_max, vlq.z[k]);
    }
}

void BoundingBoxQuant::print(FILE* f)
{
    fprintf(f, "bbq: x_min %d x_max %d x_len %d\n", x_min, x_max, x_max - x_min);
    fprintf(f, "bbq: y_min %d y_max %d y_len %d\n", y_min, y_max, y_max - y_min);
    fprintf(f, "bbq: z_min %d z_max %d z_len %d\n", z_min, z_max, z_max - z_min);
}

int BoundingBoxQuant::encode_raw_baseline(FlexRangeEncoder* fre)
{
    int max_abs = max(abs(x_min), abs(x_max));
    max_abs = max(max_abs, abs(y_min));
    max_abs = max(max_abs, abs(y_max));
    max_abs = max(max_abs, abs(z_min));
    max_abs = max(max_abs, abs(z_max));

    // number of bits needed in the sign-magnitude representation
    int bits_needed = (max_abs == 0) ? 1 : floor_log2(max_abs) + 2;

    fre_encode_bits(fre, bits_needed - 1, 5);

    encode_int_magnitude_sign(fre, x_min, bits_needed);
    encode_int_magnitude_sign(fre, x_max, bits_needed);

    encode_int_magnitude_sign(fre, y_min, bits_needed);
    encode_int_magnitude_sign(fre, y_max, bits_needed);

    encode_int_magnitude_sign(fre, z_min, bits_needed);
    encode_int_magnitude_sign(fre, z_max, bits_needed);

    return bits_needed;
}

int BoundingBoxQuant::decode_raw_baseline(FlexRangeDecoder* frd)
{
    // number of bits needed in the sign-magnitude representation
    int bits_needed = frd_read_bits(frd, 5) + 1;

    x_min = decode_int_magnitude_sign(frd, bits_needed);
    x_max = decode_int_magnitude_sign(frd, bits_needed);

    y_min = decode_int_magnitude_sign(frd, bits_needed);
    y_max = decode_int_magnitude_sign(frd, bits_needed);

    z_min = decode_int_magnitude_sign(frd, bits_needed);
    z_max = decode_int_magnitude_sign(frd, bits_needed);

    return bits_needed;
}

int BoundingBoxQuant::encode_optimized_v1(FlexRangeEncoder* fre)
{
    int max_abs = max(abs(x_min), abs(x_max));
    max_abs = max(max_abs, abs(y_min));
    max_abs = max(max_abs, abs(y_max));
    max_abs = max(max_abs, abs(z_min));
    max_abs = max(max_abs, abs(z_max));

    // number of bits needed in the sign-magnitude representation
    int bits_needed = (max_abs == 0) ? 1 : floor_log2(max_abs) + 2;

    fre_encode_bits(fre, bits_needed - 1, 5);

    encode_int_magnitude_sign(fre, x_min, bits_needed);
    if (abs(x_min) == abs(x_max)) {
        fre_encode_bits(fre, 1, 1); // same absolute value flag
        fre_encode_bits(fre, (x_min == x_max), 1); // same value flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(stdout, "bbq: x_abs_equal with equal %u\n", (x_min == x_max));
#endif
    } else {
        fre_encode_bits(fre, 0, 1);
        encode_int_magnitude_sign(fre, x_max, bits_needed);
    }

    encode_int_magnitude_sign(fre, y_min, bits_needed);
    if (abs(y_min) == abs(y_max)) {
        fre_encode_bits(fre, 1, 1); // same absolute value flag
        fre_encode_bits(fre, (y_min == y_max), 1); // same value flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(stdout, "bbq: y_abs_equal with equal %u\n", (y_min == y_max));
#endif
    } else {
        fre_encode_bits(fre, 0, 1);
        encode_int_magnitude_sign(fre, y_max, bits_needed);
    }

    encode_int_magnitude_sign(fre, z_min, bits_needed);
    if (abs(z_min) == abs(z_max)) {
        fre_encode_bits(fre, 1, 1); // same absolute value flag
        fre_encode_bits(fre, (z_min == z_max), 1); // same value flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(stdout, "bbq: z_abs_equal with equal %u\n", (z_min == z_max));
#endif
    } else {
        fre_encode_bits(fre, 0, 1);
        encode_int_magnitude_sign(fre, z_max, bits_needed);
    }

    return bits_needed;
}

int BoundingBoxQuant::encode_optimized_v2(FlexRangeEncoder* fre)
{
    int max_abs = max(abs(x_min), abs(x_max));
    max_abs = max(max_abs, abs(y_min));
    max_abs = max(max_abs, abs(y_max));
    max_abs = max(max_abs, abs(z_min));
    max_abs = max(max_abs, abs(z_max));

    // number of bits needed in the sign-magnitude representation
    int bits_needed = (max_abs == 0) ? 1 : floor_log2(max_abs) + 2;

    fre_encode_bits(fre, bits_needed - 1, 5);

    bool any_same_abs =
        ((abs(x_min) == abs(x_max)) || (abs(y_min) == abs(y_max)) || (abs(z_min) == abs(z_max)));
    fre_encode_bits(fre, any_same_abs, 1); // any same absolute value

    encode_int_magnitude_sign(fre, x_min, bits_needed);
    if (abs(x_min) == abs(x_max)) {
        fre_encode_bits(fre, 1, 1); // same absolute value flag
        fre_encode_bits(fre, (x_min == x_max), 1); // same value flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(stdout, "bbq: x_abs_equal with equal %u\n", (x_min == x_max));
#endif
    } else {
        if (any_same_abs)
            fre_encode_bits(fre, 0, 1);
        encode_int_magnitude_sign(fre, x_max, bits_needed);
    }

    encode_int_magnitude_sign(fre, y_min, bits_needed);
    if (abs(y_min) == abs(y_max)) {
        fre_encode_bits(fre, 1, 1); // same absolute value flag
        fre_encode_bits(fre, (y_min == y_max), 1); // same value flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(stdout, "bbq: y_abs_equal with equal %u\n", (y_min == y_max));
#endif
    } else {
        if (any_same_abs)
            fre_encode_bits(fre, 0, 1);
        encode_int_magnitude_sign(fre, y_max, bits_needed);
    }

    encode_int_magnitude_sign(fre, z_min, bits_needed);
    if (abs(z_min) == abs(z_max)) {
        fre_encode_bits(fre, 1, 1); // same absolute value flag
        fre_encode_bits(fre, (z_min == z_max), 1); // same value flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        fprintf(stdout, "bbq: z_abs_equal with equal %u\n", (z_min == z_max));
#endif
    } else {
        if (any_same_abs)
            fre_encode_bits(fre, 0, 1);
        encode_int_magnitude_sign(fre, z_max, bits_needed);
    }

    return bits_needed;
}

int gcd_int(int m, int n)
{
    // std::gcd appears in <numeric> only from C++17
    while (n != 0) {
        int tmp = m % n;
        m = n;
        n = tmp;
    }

    return m;
}

void PreprocessorGCD::compute(VertexListQuant& vlq, BoundingBoxQuant& bbq)
{
    gcd_x = 0;
    gcd_y = 0;
    gcd_z = 0;

    for (int k = 0; k < vlq.count; ++k) {
        gcd_x = gcd_int(gcd_x, vlq.x[k] - bbq.x_min);
        gcd_y = gcd_int(gcd_y, vlq.y[k] - bbq.y_min);
        gcd_z = gcd_int(gcd_z, vlq.z[k] - bbq.z_min);
    }

    // set gcd_# values of 0 to 1, which happens when #_max - #_min == 0
    gcd_x = max(gcd_x, 1);
    gcd_y = max(gcd_y, 1);
    gcd_z = max(gcd_z, 1);
}

void PreprocessorGCD::preprocess(VertexListQuant& vlq, BoundingBoxQuant& bbq)
{
    // preprocess the coordinates of the vertexes
    for (int k = 0; k < vlq.count; ++k) {
        vlq.x[k] = (vlq.x[k] - bbq.x_min) / gcd_x;
        vlq.y[k] = (vlq.y[k] - bbq.y_min) / gcd_y;
        vlq.z[k] = (vlq.z[k] - bbq.z_min) / gcd_z;
    }
}

void PreprocessorGCD::adjust_bounding_box(BoundingBoxQuant& bbq)
{
    // the original must be saved if needed later
    bbq.x_max = (bbq.x_max - bbq.x_min) / gcd_x;
    bbq.x_min = 0;
    bbq.y_max = (bbq.y_max - bbq.y_min) / gcd_y;
    bbq.y_min = 0;
    bbq.z_max = (bbq.z_max - bbq.z_min) / gcd_z;
    bbq.z_min = 0;
}

void PreprocessorGCD::reverse_preprocess(VertexListQuant& vlq, BoundingBoxQuant& bbq)
{
    // reverse preprocess the coordinates of the vertexes
    for (int k = 0; k < vlq.count; ++k) {
        vlq.x[k] = vlq.x[k] * gcd_x + bbq.x_min;
        vlq.y[k] = vlq.y[k] * gcd_y + bbq.y_min;
        vlq.z[k] = vlq.z[k] * gcd_z + bbq.z_min;
    }
}

void PreprocessorGCD::print(FILE* f)
{
    if ((gcd_x != 1) || (gcd_y != 1) || (gcd_z != 1)) {
        fprintf(f, "vlq: gcd_x %d gcd_y %d gcd_z %d\n", gcd_x, gcd_y, gcd_z);
    }
}

void PreprocessorGCD::encode_raw_baseline(FlexRangeEncoder* fre, int bits_needed)
{
    if ((gcd_x != 1) || (gcd_y != 1) || (gcd_z != 1)) {
        fre_encode_bits(fre, 1, 1);

        // the maximum possible value of gcd_# is #_max - #_min <= 2 * bbq.max_abs
        // therefore, gcd_# can always be coded as unsigned using bits_needed bits
        fre_encode_bits(fre, (gcd_x != 1), 1);
        if (gcd_x != 1)
            fre_encode_bits_long(fre, gcd_x, bits_needed);
        fre_encode_bits(fre, (gcd_y != 1), 1);
        if (gcd_y != 1)
            fre_encode_bits_long(fre, gcd_y, bits_needed);
        fre_encode_bits(fre, (gcd_z != 1), 1);
        if (gcd_z != 1)
            fre_encode_bits_long(fre, gcd_z, bits_needed);
    } else {
        fre_encode_bits(fre, 0, 1);
    }
}

void PreprocessorGCD::decode_raw_baseline(FlexRangeDecoder* frd, int bits_needed)
{
    int gcd_any_present = frd_read_bits(frd, 1);
    gcd_x = 1;
    gcd_y = 1;
    gcd_z = 1;

    if (gcd_any_present) {
        // the maximum possible value of gcd_# is #_max - #_min <= 2 * bbq.max_abs
        // therefore, gcd_# can always be coded as unsigned using bits_needed bits
        int gcd_x_present = frd_read_bits(frd, 1);
        if (gcd_x_present) {
            gcd_x = frd_read_bits_long(frd, bits_needed);
        }
        int gcd_y_present = frd_read_bits(frd, 1);
        if (gcd_y_present) {
            gcd_y = frd_read_bits_long(frd, bits_needed);
        }
        int gcd_z_present = frd_read_bits(frd, 1);
        if (gcd_z_present) {
            gcd_z = frd_read_bits_long(frd, bits_needed);
        }
    }
}

void PreprocessorGCD::encode_optimized_v1(
    FlexRangeEncoder* fre, BoundingBoxQuant& bbq, int bits_needed)
{
    if ((gcd_x != 1) || (gcd_y != 1) || (gcd_z != 1)) {
        fre_encode_bits(fre, 1, 1);

        // the maximum possible value of gcd_# is #_max - #_min <= 2 * bbq.max_abs
        // therefore, gcd_# can always be coded as unsigned using bits_needed bits
        fre_encode_bits(fre, (gcd_x != 1), 1);
        if (gcd_x != 1) {
            if (gcd_x == bbq.x_max - bbq.x_min) {
                fre_encode_bits(fre, 1, 1); // gcd_x is equal with len_x flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
                fprintf(stdout, "vlq: gcd_x equal with len_x\n");
#endif
            } else {
                fre_encode_bits(fre, 0, 1);
                fre_encode_bits_long(fre, gcd_x, bits_needed);
            }
        }
        fre_encode_bits(fre, (gcd_y != 1), 1);
        if (gcd_y != 1) {
            if (gcd_y == bbq.y_max - bbq.y_min) {
                fre_encode_bits(fre, 1, 1); // gcd_y is equal with len_y flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
                fprintf(stdout, "vlq: gcd_y equal with len_y\n");
#endif
            } else {
                fre_encode_bits(fre, 0, 1);
                fre_encode_bits_long(fre, gcd_y, bits_needed);
            }
        }
        fre_encode_bits(fre, (gcd_z != 1), 1);
        if (gcd_z != 1) {
            if (gcd_z == bbq.z_max - bbq.z_min) {
                fre_encode_bits(fre, 1, 1); // gcd_z is equal with len_z flag
#if GC_PRINT_INTERNAL_DIAGNOSTICS
                fprintf(stdout, "vlq: gcd_z equal with len_z\n");
#endif
            } else {
                fre_encode_bits(fre, 0, 1);
                fre_encode_bits_long(fre, gcd_z, bits_needed);
            }
        }
    } else {
        fre_encode_bits(fre, 0, 1);
    }
}

double binary_entropy(int count_0, int count_total)
{
    double ent = 0.0;

    if ((count_0 != 0) && (count_0 != count_total) && (count_total != 0)) {
        double p0 = (double)count_0 / count_total;
        ent = -p0 * log2(p0) - (1.0 - p0) * log2(1.0 - p0);
    }

    return ent;
}

int VertexListQuant::encode_raw_baseline(
    FlexRangeEncoder* fre, BoundingBoxQuant& bbq, bool est_only)
{
    int alph_x = bbq.x_max - bbq.x_min + 1;
    int alph_y = bbq.y_max - bbq.y_min + 1;
    int alph_z = bbq.z_max - bbq.z_min + 1;

    if (!est_only) {
        for (int k = 0; k < count; ++k) {
            fre_encode_uniform_long(fre, x[k] - bbq.x_min, alph_x);
            fre_encode_uniform_long(fre, y[k] - bbq.y_min, alph_y);
            fre_encode_uniform_long(fre, z[k] - bbq.z_min, alph_z);
        }
    }

    double est_bits = (log2(alph_x) + log2(alph_y) + log2(alph_z)) * count;

    return (int)ceil(est_bits);
}

void VertexListQuant::decode_raw_baseline(
    FlexRangeDecoder* frd, BoundingBoxQuant& bbq, int vl_count)
{
    count = vl_count;

    x.assign(count, 0);
    y.assign(count, 0);
    z.assign(count, 0);

    duplicates_removed = 0;
    exact_duplicates_removed = 0;

    if (count == 0)
        return;

    int alph_x = bbq.x_max - bbq.x_min + 1;
    int alph_y = bbq.y_max - bbq.y_min + 1;
    int alph_z = bbq.z_max - bbq.z_min + 1;

    for (int k = 0; k < count; ++k) {
        x[k] = frd_read_uniform_long(frd, alph_x) + bbq.x_min;
        y[k] = frd_read_uniform_long(frd, alph_y) + bbq.y_min;
        z[k] = frd_read_uniform_long(frd, alph_z) + bbq.z_min;
    }
}

int VertexListQuant::encode_optimized_v1(
    FlexRangeEncoder* fre, BoundingBoxQuant& bbq, bool est_only)
{
    int alph_x = bbq.x_max - bbq.x_min + 1;
    int alph_y = bbq.y_max - bbq.y_min + 1;
    int alph_z = bbq.z_max - bbq.z_min + 1;

    if (!est_only) {
        map<int, int> xuc;
        map<int, int> yuc;
        map<int, int> zuc;

        AdaptiveEstimator sw_x;
        sw_x.create(2, 32768);
        AdaptiveEstimator sw_y;
        sw_y.create(2, 32768);
        AdaptiveEstimator sw_z;
        sw_z.create(2, 32768);

        // pre-insert the bounding box coordinates
        xuc[0] = 1;
        xuc[bbq.x_max - bbq.x_min] = 1;
        yuc[0] = 1;
        yuc[bbq.y_max - bbq.y_min] = 1;
        zuc[0] = 1;
        zuc[bbq.z_max - bbq.z_min] = 1;

        for (int k = 0; k < count; ++k) {
            int x_val = x[k] - bbq.x_min;
            int& x_count = xuc[x_val];
            if (x_count == 0) // key x_val was not found in the xuc map
            {
                // an insertion of new key x_val with value 0 was just made in the xuc map
                sw_x.encodeUpdate(0, fre); // value coded raw flag
                fre_encode_uniform_long(fre, x_val, alph_x);
            } else {
                // if all possible x_val are already in the xuc map, the flag is not necessary anymore
                if ((int)xuc.size() < alph_x)
                    sw_x.encodeUpdate(1, fre); // value coded with index flag
                int idx = (int)distance(xuc.begin(), xuc.find(x_val));
                fre_encode_uniform_long(fre, idx, (int)xuc.size());
            }
            x_count++;

            int y_val = y[k] - bbq.y_min;
            int& y_count = yuc[y_val];
            if (y_count == 0) {
                sw_y.encodeUpdate(0, fre);
                fre_encode_uniform_long(fre, y_val, alph_y);
            } else {
                if ((int)yuc.size() < alph_y)
                    sw_y.encodeUpdate(1, fre);
                int idx = (int)distance(yuc.begin(), yuc.find(y_val));
                fre_encode_uniform_long(fre, idx, (int)yuc.size());
            }
            y_count++;

            int z_val = z[k] - bbq.z_min;
            int& z_count = zuc[z_val];
            if (z_count == 0) {
                sw_z.encodeUpdate(0, fre);
                fre_encode_uniform_long(fre, z_val, alph_z);
            } else {
                if ((int)zuc.size() < alph_z)
                    sw_z.encodeUpdate(1, fre);
                int idx = (int)distance(zuc.begin(), zuc.find(z_val));
                fre_encode_uniform_long(fre, idx, (int)zuc.size());
            }
            z_count++;
        }
    }

    int xu_count = (int)xu.size();
    int yu_count = (int)yu.size();
    int zu_count = (int)zu.size();

    double kt_bits = 0.5 * log2(count + 1) + 0.325748; // KT binary estimator parameter cost

    double est_bits_x = 0.0;
    if (xu_count >= 2) {
        assert(alph_x >= 2);

        // using xu_count - 2 instead of xu_count because of pre-insertion of bounding box coordinates
        est_bits_x = log2(alph_x) * (xu_count - 2) + log2(xu_count) * (count - (xu_count - 2))
            + count * binary_entropy(xu_count - 2, count);
        if (alph_x > 2)
            est_bits_x += kt_bits; // no flags are are necessary if alph_x == 2
    }
    if (est_bits_x > log2(alph_x) * count) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("est_uniq_x %.1f est_raw_x %.1f\n", est_bits_x, log2(alph_x) * count);
#endif
        // approximate condition: log2(alph) * u + log2(u) * (n - u) + n * H(u, n) <= log2(alph) * n
        // where alph = alph_x, u = xu_count, and n = count
        // which is equivalent with: [log2(alph) - log2(u)] * (n - u) >= n * H(u, n)
    }

    double est_bits_y = 0.0;
    if (yu_count >= 2) {
        assert(alph_y >= 2);

        est_bits_y = log2(alph_y) * (yu_count - 2) + log2(yu_count) * (count - (yu_count - 2))
            + count * binary_entropy(yu_count - 2, count);
        if (alph_y > 2)
            est_bits_y += kt_bits;
    }
    if (est_bits_y > log2(alph_y) * count) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("est_uniq_y %.1f est_raw_y %.1f\n", est_bits_y, log2(alph_y) * count);
#endif
    }

    double est_bits_z = 0.0;
    if (zu_count >= 2) {
        assert(alph_z >= 2);

        est_bits_z = log2(alph_z) * (zu_count - 2) + log2(zu_count) * (count - (zu_count - 2))
            + count * binary_entropy(zu_count - 2, count);
        if (alph_z > 2)
            est_bits_z += kt_bits;
    }
    if (est_bits_z > log2(alph_z) * count) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("est_uniq_z %.1f est_raw_z %.1f\n", est_bits_z, log2(alph_z) * count);
#endif
    }

    double est_bits = est_bits_x + est_bits_y + est_bits_z;

    return (int)ceil(est_bits);
}

static void insert_into_unique_sorted(vector<int>& us, int value)
{
    auto it = lower_bound(us.begin(), us.end(), value);
    if (it != us.end()) {
        assert(*it != value); // vector contains unique values only
    }
    us.insert(it, value);
}

void VertexListQuant::decode_optimized_v1(
    FlexRangeDecoder* frd, BoundingBoxQuant& bbq, int vl_count)
{
    count = vl_count;

    x.assign(count, 0);
    y.assign(count, 0);
    z.assign(count, 0);

    duplicates_removed = 0;
    exact_duplicates_removed = 0;

    if (count == 0)
        return;

    int alph_x = bbq.x_max - bbq.x_min + 1;
    int alph_y = bbq.y_max - bbq.y_min + 1;
    int alph_z = bbq.z_max - bbq.z_min + 1;

    // unique values always sorted vectors for x, y, and z coordinates
    vector<int> xus;
    xus.reserve(count); // avoid any reallocations
    vector<int> yus;
    yus.reserve(count);
    vector<int> zus;
    zus.reserve(count);

    AdaptiveEstimator sw_x;
    sw_x.create(2, 32768);
    AdaptiveEstimator sw_y;
    sw_y.create(2, 32768);
    AdaptiveEstimator sw_z;
    sw_z.create(2, 32768);

    // pre-insert the bounding box coordinates
    insert_into_unique_sorted(xus, 0);
    if (bbq.x_max - bbq.x_min != 0)
        insert_into_unique_sorted(xus, bbq.x_max - bbq.x_min);
    insert_into_unique_sorted(yus, 0);
    if (bbq.y_max - bbq.y_min != 0)
        insert_into_unique_sorted(yus, bbq.y_max - bbq.y_min);
    insert_into_unique_sorted(zus, 0);
    if (bbq.z_max - bbq.z_min != 0)
        insert_into_unique_sorted(zus, bbq.z_max - bbq.z_min);

    for (int k = 0; k < count; ++k) {
        bool x_coded_with_index = true;
        // if not all possible x_val values are already in the xus vector
        if ((int)xus.size() < alph_x) {
            x_coded_with_index = (bool)sw_x.decodeUpdate(frd);
        }
        int x_val;
        if (x_coded_with_index) {
            int x_idx = frd_read_uniform_long(frd, (int)xus.size());
            x_val = xus[x_idx]; // get the value at index x_idx in the xus vector
        } else { // coded raw
            x_val = frd_read_uniform_long(frd, alph_x);
            insert_into_unique_sorted(xus, x_val); // insert the new unique value x_val into xus
        }
        x[k] = x_val + bbq.x_min;

        int y_coded_with_index = true;
        // if not all possible y_val values are already in the yus vector
        if ((int)yus.size() < alph_y) {
            y_coded_with_index = (bool)sw_y.decodeUpdate(frd);
        }
        int y_val;
        if (y_coded_with_index) {
            int y_idx = frd_read_uniform_long(frd, (int)yus.size());
            y_val = yus[y_idx]; // get the value at index y_idx in the yus vector
        } else { // coded raw
            y_val = frd_read_uniform_long(frd, alph_y);
            insert_into_unique_sorted(yus, y_val); // insert the new unique value y_val into yus
        }
        y[k] = y_val + bbq.y_min;

        bool z_coded_with_index = true;
        // if not all possible z_val values are already in the zus vector
        if ((int)zus.size() < alph_z) {
            z_coded_with_index = (bool)sw_z.decodeUpdate(frd);
        }
        int z_val;
        if (z_coded_with_index) {
            int z_idx = frd_read_uniform_long(frd, (int)zus.size());
            z_val = zus[z_idx]; // get the value at index z_idx in the zus vector
        } else { // coded raw
            z_val = frd_read_uniform_long(frd, alph_z);
            insert_into_unique_sorted(zus, z_val); // insert the new unique value z_val into zus
        }
        z[k] = z_val + bbq.z_min;
    }
}

class DirectMTF
{
public:
    vector<int> values;

    void create();
    void insert_new(int v);
    int insert_existing(int v);
};

void DirectMTF::create()
{
    values.clear();
}

void DirectMTF::insert_new(int v)
{
    values.insert(values.begin(), v);
}

int DirectMTF::insert_existing(int v)
{
    int pos = -1;

    for (auto it = values.begin(); it < values.end(); ++it) {
        if (*it == v) {
            pos = (int)distance(values.begin(), it);

            values.erase(it);
            values.insert(values.begin(), v);
            break;
        }
    }
    assert(pos != -1);

    return pos;
}

class EncoderMTF
{
public:
    static const int AE_MAX = 8;
    AdaptiveEstimator ae[AE_MAX];

    void create();
    void code(FlexRangeEncoder* fre, int idx, int alph);
};

void EncoderMTF::create()
{
    for (int i = 0; i < AE_MAX; ++i) {
        ae[i].create(2, 32768);
    }
}

void EncoderMTF::code(FlexRangeEncoder* fre, int idx, int alph)
{
    assert(idx < alph);

    /*
    when AE_MAX = 4, for a number, these binary sequences are coded, each using an ae vector entry
    0 -> 0
    1 -> 1 0
    2 -> 1 1 0
    3 -> 1 1 1 0
    4 -> 1 1 1 1 uniform(idx - 4, alph - 4)
    */

    int cnt = min(idx, AE_MAX - 1);
    for (int i = 0; i < cnt; ++i) {
        // flag to indicate whether idx > i
        ae[i].encodeUpdate(1, fre);
    }

    if (idx < AE_MAX) {
        ae[cnt].encodeUpdate(0, fre);
    } else { // idx >= AE_MAX
        ae[cnt].encodeUpdate(1, fre);
        fre_encode_uniform_long(fre, idx - AE_MAX, alph - AE_MAX);
    }
}

int VertexListQuant::encode_optimized_v2(
    FlexRangeEncoder* fre, BoundingBoxQuant& bbq, bool est_only)
{
    int alph_x = bbq.x_max - bbq.x_min + 1;
    int alph_y = bbq.y_max - bbq.y_min + 1;
    int alph_z = bbq.z_max - bbq.z_min + 1;

    if (!est_only) {
        map<int, int> xuc;
        map<int, int> yuc;
        map<int, int> zuc;

        AdaptiveEstimator sw_x;
        sw_x.create(2, 32768);
        AdaptiveEstimator sw_y;
        sw_y.create(2, 32768);
        AdaptiveEstimator sw_z;
        sw_z.create(2, 32768);

        DirectMTF mtf_x;
        mtf_x.create();
        DirectMTF mtf_y;
        mtf_y.create();
        DirectMTF mtf_z;
        mtf_z.create();
        EncoderMTF e_mtf_x;
        e_mtf_x.create();
        EncoderMTF e_mtf_y;
        e_mtf_y.create();
        EncoderMTF e_mtf_z;
        e_mtf_z.create();

        // pre-insert the bounding box coordinates
        xuc[0] = 1;
        xuc[bbq.x_max - bbq.x_min] = 1;
        yuc[0] = 1;
        yuc[bbq.y_max - bbq.y_min] = 1;
        zuc[0] = 1;
        zuc[bbq.z_max - bbq.z_min] = 1;

        if (bbq.x_max - bbq.x_min != 0)
            mtf_x.insert_new(bbq.x_max - bbq.x_min);
        mtf_x.insert_new(0);
        if (bbq.y_max - bbq.y_min != 0)
            mtf_y.insert_new(bbq.y_max - bbq.y_min);
        mtf_y.insert_new(0);
        if (bbq.z_max - bbq.z_min != 0)
            mtf_z.insert_new(bbq.z_max - bbq.z_min);
        mtf_z.insert_new(0);

        for (int k = 0; k < count; ++k) {
            int x_val = x[k] - bbq.x_min;
            int& x_count = xuc[x_val];
            if (x_count == 0) // key x_val was not found in the xuc map
            {
                // an insertion of new key x_val with value 0 was just made in the xuc map
                sw_x.encodeUpdate(0, fre); // value coded raw flag
                fre_encode_uniform_long(fre, x_val, alph_x);

                mtf_x.insert_new(x_val);
#if GC_PRINT_VERTEX_UNIQUE_MTF
                printf(" [%d]", (int)xuc.size() - 1);
#endif
            } else {
                // if all possible x_val are already in the xuc map, the flag is not necessary anymore
                if ((int)xuc.size() < alph_x)
                    sw_x.encodeUpdate(1, fre); // value coded with index flag

                int pos_x = mtf_x.insert_existing(x_val);
                assert(pos_x < (int)xuc.size());
#if GC_PRINT_VERTEX_UNIQUE_MTF
                printf(" %d", pos_x);
#endif

                e_mtf_x.code(fre, pos_x, (int)xuc.size());
            }
            x_count++;

            int y_val = y[k] - bbq.y_min;
            int& y_count = yuc[y_val];
            if (y_count == 0) {
                sw_y.encodeUpdate(0, fre);
                fre_encode_uniform_long(fre, y_val, alph_y);

                mtf_y.insert_new(y_val);
#if GC_PRINT_VERTEX_UNIQUE_MTF
                printf(" [%d]", (int)yuc.size() - 1);
#endif
            } else {
                if ((int)yuc.size() < alph_y)
                    sw_y.encodeUpdate(1, fre);

                int pos_y = mtf_y.insert_existing(y_val);
                assert(pos_y < (int)yuc.size());
#if GC_PRINT_VERTEX_UNIQUE_MTF
                printf(" %d", pos_y);
#endif

                e_mtf_y.code(fre, pos_y, (int)yuc.size());
            }
            y_count++;

            int z_val = z[k] - bbq.z_min;
            int& z_count = zuc[z_val];
            if (z_count == 0) {
                sw_z.encodeUpdate(0, fre);
                fre_encode_uniform_long(fre, z_val, alph_z);

                mtf_z.insert_new(z_val);
#if GC_PRINT_VERTEX_UNIQUE_MTF
                printf(" [%d]", (int)zuc.size() - 1);
#endif
            } else {
                if ((int)zuc.size() < alph_z)
                    sw_z.encodeUpdate(1, fre);

                int pos_z = mtf_z.insert_existing(z_val);
                assert(pos_z < (int)zuc.size());
#if GC_PRINT_VERTEX_UNIQUE_MTF
                printf(" %d", pos_z);
#endif

                e_mtf_z.code(fre, pos_z, (int)zuc.size());
            }
            z_count++;
        }
    }
#if GC_PRINT_VERTEX_UNIQUE_MTF
    printf("\n");
#endif

    int xu_count = (int)xu.size();
    int yu_count = (int)yu.size();
    int zu_count = (int)zu.size();

    double kt_bits = 0.5 * log2(count + 1) + 0.325748; // KT binary estimator parameter cost

    double est_bits_x = 0.0;
    if (xu_count >= 2) {
        assert(alph_x >= 2);

        // using xu_count - 2 instead of xu_count because of pre-insertion of bounding box coordinates
        est_bits_x = log2(alph_x) * (xu_count - 2) + log2(xu_count) * (count - (xu_count - 2))
            + count * binary_entropy(xu_count - 2, count);
        if (alph_x > 2)
            est_bits_x += kt_bits; // no flags are are necessary if alph_x == 2
    }
    if (est_bits_x > log2(alph_x) * count) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("est_uniq_x %.1f est_raw_x %.1f\n", est_bits_x, log2(alph_x) * count);
#endif
        // approximate condition: log2(alph) * u + log2(u) * (n - u) + n * H(u, n) <= log2(alph) * n
        // where alph = alph_x, u = xu_count, and n = count
        // which is equivalent with: [log2(alph) - log2(u)] * (n - u) >= n * H(u, n)
    }

    double est_bits_y = 0.0;
    if (yu_count >= 2) {
        assert(alph_y >= 2);

        est_bits_y = log2(alph_y) * (yu_count - 2) + log2(yu_count) * (count - (yu_count - 2))
            + count * binary_entropy(yu_count - 2, count);
        if (alph_y > 2)
            est_bits_y += kt_bits;
    }
    if (est_bits_y > log2(alph_y) * count) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("est_uniq_y %.1f est_raw_y %.1f\n", est_bits_y, log2(alph_y) * count);
#endif
    }

    double est_bits_z = 0.0;
    if (zu_count >= 2) {
        assert(alph_z >= 2);

        est_bits_z = log2(alph_z) * (zu_count - 2) + log2(zu_count) * (count - (zu_count - 2))
            + count * binary_entropy(zu_count - 2, count);
        if (alph_z > 2)
            est_bits_z += kt_bits;
    }
    if (est_bits_z > log2(alph_z) * count) {
#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("est_uniq_z %.1f est_raw_z %.1f\n", est_bits_z, log2(alph_z) * count);
#endif
    }

    double est_bits = est_bits_x + est_bits_y + est_bits_z;

    return (int)ceil(est_bits);
}

int TriangleList::encode_raw_baseline(
    FlexRangeEncoder* fre, int vl_count, int ml_count, bool est_only)
{
    if (count == 0)
        return 0;

    int alph = vl_count; // the number of vertexes
    int mat_alph = shared_material_count + 1;
    (void)ml_count;

    if (!est_only) {
        // encode the (shared) total number of acoustic materials
        assert(shared_material_count < (1 << 10));
        encode_uint_compact(fre, shared_material_count, 10);

        for (int k = 0; k < count; ++k) {
            fre_encode_uniform_long(fre, v1[k], alph);
            fre_encode_uniform_long(fre, v2[k], alph);
            fre_encode_uniform_long(fre, v3[k], alph);

            // a value of -1 is valid and represents "mat:none"
            fre_encode_uniform(fre, mi[k] + 1, mat_alph);
        }
    }

    double est_bits = estimate_uint_compact(shared_material_count, 10)
        + (3 * log2(max(alph, 1)) + log2(mat_alph)) * count;

    return (int)ceil(est_bits);
}

void TriangleList::decode_raw_baseline(
    FlexRangeDecoder* frd, int vl_count, int tl_count, int ml_count)
{
    count = tl_count;

    v1.assign(count, 0);
    v2.assign(count, 0);
    v3.assign(count, 0);
    mi.assign(count, 0);

    duplicates_removed = 0;
    exact_duplicates_removed = 0;
    degenerates_removed = 0;
    exact_duplicates_removed = 0;

    shared_material_count = 0; // valid value but actually unused, for the case count == 0

    if (count == 0)
        return;

    // decode the (shared) total number of acoustic materials
    shared_material_count = decode_uint_compact(frd, 10);

    int alph = vl_count; // the number of vertexes
    int mat_alph = shared_material_count + 1;
    (void)ml_count;

    for (int k = 0; k < count; ++k) {
        v1[k] = frd_read_uniform_long(frd, alph);
        v2[k] = frd_read_uniform_long(frd, alph);
        v3[k] = frd_read_uniform_long(frd, alph);

        // a value of -1 is valid and represents "mat:none"
        mi[k] = frd_read_uniform(frd, mat_alph) - 1;
    }
}

int TriangleList::encode_optimized_v1(
    FlexRangeEncoder* fre, int vl_count, int ml_count, bool est_only)
{
    if (count == 0)
        return 0;

    assert(vl_count >= 1);

    int alph = vl_count; // the number of vertexes
    int mat_alph = shared_material_count + 1;
    (void)ml_count;

    if (!est_only) {
        vector<bool> is_present;
        is_present.assign(mat_alph, false);
        int max_mi = 0;

        for (int k = 0; k < count; ++k) {
            assert((mi[k] >= -1) && (mi[k] < shared_material_count));

            is_present[mi[k] + 1] = true;
            max_mi = max(max_mi, mi[k] + 1);
        }

        // compute actual material index alphabet, without trailing material indexes not actually present
        int actual_mat_alph = max_mi + 1;
        // cmi = compacted material index, including only the material indexes actually present
        vector<int> mi_to_cmi;
        mi_to_cmi.assign(actual_mat_alph, 0);
        int cmi_alph = 0;

        for (int j = 0; j < actual_mat_alph; ++j) {
            mi_to_cmi[j] = cmi_alph;
            cmi_alph += is_present[j];
        }
        assert(cmi_alph >= 1);

        AdaptiveEstimator same_cmi_ae;
        same_cmi_ae.create(2, 32768);
        AdaptiveEstimator cmi_ae;
        cmi_ae.create(cmi_alph, 32768);
        int last_cmi = 0;

        // encode the actual material index alphabet instead of the (shared) total number of acoustic materials
        assert(shared_material_count < (1 << 10));
        assert(actual_mat_alph >= 1);
        assert(actual_mat_alph <= shared_material_count + 1);

        encode_uint_compact(fre, actual_mat_alph - 1, 10);

        // encode the presence bit mask for the material indexes
        for (int j = 0; j < actual_mat_alph; ++j) {
            if ((j == actual_mat_alph - 1) && (cmi_alph == 1) && (is_present[j] == 1)) {
                // all the previous is_present bits were 0, so the last is_present bit must be 1
            } else {
                fre_encode_bits(fre, is_present[j], 1);
            }
        }

#if GC_PRINT_INTERNAL_DIAGNOSTICS
        printf("tl: mat_alph %d actual_mat_alph %d cmi_alph %d is_present ", mat_alph,
            actual_mat_alph, cmi_alph);
        for (int j = 0; j < actual_mat_alph; ++j)
            printf("%d", (int)is_present[j]);
        printf("\n");
        int start_bits = fre_virtual_finish(fre);
#endif

        for (int k = 0; k < count; ++k) {
            fre_encode_uniform_long(fre, v1[k], alph);
            fre_encode_uniform_long(fre, v2[k], alph);
            fre_encode_uniform_long(fre, v3[k], alph);

            // a value of -1 is valid and represents "mat:none"
            int cmi = mi_to_cmi[mi[k] + 1];

            if (cmi_alph == 1) {
                assert(cmi == 0);

                // nothing to encode
                continue;
            }

            if (cmi == last_cmi) {
                same_cmi_ae.encodeUpdate(1, fre);
            } else // cmi != last_cmi
            {
                same_cmi_ae.encodeUpdate(0, fre);
                if (cmi_alph == 2) {
                    assert(cmi == 1 - last_cmi); // different, therefore it is the other symbol

                    // nothing to encode
                } else {
                    cmi_ae.encodeUpdate(cmi, fre);
                }

                last_cmi = cmi;
            }
        }

#if GC_PRINT_INTERNAL_DIAGNOSTICS
        int mi_bits = fre_virtual_finish(fre) - start_bits;
        mi_bits -= (int)ceil(3 * log2(alph) * count);
        int mi_baseline_bits = (int)ceil(log2(mat_alph) * count);
        printf("tl: count %d mi_baseline_bits %d mi_bits %d\n", count, mi_baseline_bits, mi_bits);
#endif
    }

    double est_bits = estimate_uint_compact(shared_material_count, 10)
        + (3 * log2(alph) + log2(mat_alph)) * count;

    return (int)ceil(est_bits);
}

void TriangleList::decode_optimized_v1(
    FlexRangeDecoder* frd, int vl_count, int tl_count, int ml_count)
{
    count = tl_count;

    v1.assign(count, 0);
    v2.assign(count, 0);
    v3.assign(count, 0);
    mi.assign(count, 0);

    duplicates_removed = 0;
    exact_duplicates_removed = 0;
    degenerates_removed = 0;
    exact_duplicates_removed = 0;

    shared_material_count = 0; // valid value but actually unused, for the case count == 0

    if (count == 0)
        return;

    // decode the actual material index alphabet
    int actual_mat_alph = decode_uint_compact(frd, 10) + 1;
    shared_material_count = actual_mat_alph - 1;

    int alph = vl_count; // the number of vertexes
    (void)ml_count;

    vector<bool> is_present;
    is_present.assign(actual_mat_alph, false);

    vector<int> cmi_to_mi;
    cmi_to_mi.reserve(actual_mat_alph); // avoid any reallocations
    int cmi_alph = 0;

    for (int j = 0; j < actual_mat_alph; ++j) {
        if ((j == actual_mat_alph - 1) && (cmi_alph == 0)) {
            // all the previous is_present bits were 0, so the last is_present bit must be 1
            is_present[j] = 1;
            // this also avoids a value of 0 for cmi_alph if a bitstream error makes all is_present bits 0
        } else {
            is_present[j] = frd_read_bits(frd, 1);
        }

        if (is_present[j]) {
            cmi_to_mi.push_back(j);
            cmi_alph++;
        }
    }
    assert(cmi_alph >= 1);

    AdaptiveEstimator same_cmi_ae;
    same_cmi_ae.create(2, 32768);
    AdaptiveEstimator cmi_ae;
    cmi_ae.create(cmi_alph, 32768);
    int last_cmi = 0;

    for (int k = 0; k < count; ++k) {
        v1[k] = frd_read_uniform_long(frd, alph);
        v2[k] = frd_read_uniform_long(frd, alph);
        v3[k] = frd_read_uniform_long(frd, alph);

        // a value of -1 is valid and represents "mat:none"
        int cmi;

        if (cmi_alph == 1) {
            // nothing to decode
            cmi = 0;
        } else {
            bool same_cmi = (bool)same_cmi_ae.decodeUpdate(frd);
            if (same_cmi) {
                cmi = last_cmi;
            } else // !same_cmi
            {
                if (cmi_alph == 2) {
                    // nothing to decode
                    cmi = 1 - last_cmi; // different, therefore it is the other symbol
                } else {
                    cmi = cmi_ae.decodeUpdate(frd);
                }

                last_cmi = cmi;
            }
        }

        mi[k] = cmi_to_mi[cmi] - 1;
    }
}

void TriangleList::sort_vertexes()
{
    for (int k = 0; k < count; ++k) {
        int vmin = min(min(v1[k], v2[k]), v3[k]);

        // if necessary, rotate triangle vertexes to make v1 the minimum one
        if (vmin == v3[k]) {
            // rotate right: v1 v2 v3 => v3 v1 v2
            int v3_tmp = v3[k];
            v3[k] = v2[k];
            v2[k] = v1[k];
            v1[k] = v3_tmp;
        } else if (vmin == v2[k]) {
            // rotate left: v1 v2 v3 => v2 v3 v1
            int v1_tmp = v1[k];
            v1[k] = v2[k];
            v2[k] = v3[k];
            v3[k] = v1_tmp;
        }
    }
}

void TriangleList::print_double_sorted(FILE* f)
{
    fprintf(f, "%d\n", count);
    int last_v1 = 0;

    for (int k = 0; k < count; ++k) {
        fprintf(f, "%5d %5d %5d %5d %5d %5d\n", v1[k], v2[k], v3[k], v1[k] - last_v1, v2[k] - v1[k],
            v3[k] - v1[k]);

        last_v1 = v1[k];
    }
}

void TriangleList::sort_triangles(int vl_count)
{
    long long cnt = vl_count;

    for (int k = 0; k < count - 1; ++k) {
        for (int j = k + 1; j < count; ++j) {
            long long val_j = v1[j] * cnt * cnt + v2[j] * cnt + v3[j];
            long long val_k = v1[k] * cnt * cnt + v2[k] * cnt + v3[k];

            if (val_k > val_j) {
                swap(v1[j], v1[k]);
                swap(v2[j], v2[k]);
                swap(v3[j], v3[k]);
                swap(mi[j], mi[k]);
            }
        }
    }
}

float AcousticMaterial::encode_raw_baseline(FlexRangeEncoder* fre, int cx_steps, bool est_only)
{
    if (!est_only) {
        fre_encode_uniform(fre, freq_grid_index, MPEGI_MAX_GRIDS);
        fre_encode_uniform(fre, freq_count - 1, 64);
        for (int k = 0; k < freq_count; ++k) {
            int c1_idx = (int)lround(c1[k] * (float)cx_steps);
            fre_encode_uniform(fre, c1_idx, cx_steps + 1);
            int c2_idx = (int)lround(c2[k] * (float)cx_steps);
            fre_encode_uniform(fre, c2_idx, cx_steps + 1);
            int c3_idx = (int)lround(c3[k] * (float)cx_steps);
            fre_encode_uniform(fre, c3_idx, cx_steps + 1);
            int c4_idx = (int)lround(c4[k] * (float)cx_steps);
            fre_encode_uniform(fre, c4_idx, cx_steps + 1);
        }
    }

    double est_bits = log2(64) + freq_count * (log2(24000) + 4 * log2(cx_steps + 1));

    return (float)est_bits;
}

float AcousticMaterial::encode_optimized_v1(FlexRangeEncoder* fre, int cx_steps, bool est_only)
{
    vector<int> c1_idx;
    vector<int> c2_idx;
    vector<int> c3_idx;
    vector<int> c4_idx;

    c1_idx.assign(freq_count, 0);
    c2_idx.assign(freq_count, 0);
    c3_idx.assign(freq_count, 0);
    c4_idx.assign(freq_count, 0);

    bool c1_is_0 = true;
    bool c2_is_0 = true;
    bool c3_is_0 = true;
    bool c4_is_0 = true;

    for (int k = 0; k < freq_count; ++k) {
        c1_idx[k] = (int)lround(c1[k] * (float)cx_steps);
        c2_idx[k] = (int)lround(c2[k] * (float)cx_steps);
        c3_idx[k] = (int)lround(c3[k] * (float)cx_steps);
        c4_idx[k] = (int)lround(c4[k] * (float)cx_steps);

        c1_is_0 &= (c1_idx[k] == 0);
        c2_is_0 &= (c2_idx[k] == 0);
        c3_is_0 &= (c3_idx[k] == 0);
        c4_is_0 &= (c4_idx[k] == 0);
    }

    if (!est_only) {
        fre_encode_uniform(fre, freq_grid_index, MPEGI_MAX_GRIDS);
        fre_encode_uniform(fre, freq_count - 1, 64);

        fre_encode_bits(fre, c1_is_0, 1);
        fre_encode_bits(fre, c2_is_0, 1);
        fre_encode_bits(fre, c3_is_0, 1);
        fre_encode_bits(fre, c4_is_0, 1);

        for (int k = 0; k < freq_count; ++k) {
            if (!c1_is_0)
                fre_encode_uniform(fre, c1_idx[k], cx_steps + 1);
            if (!c2_is_0)
                fre_encode_uniform(fre, c2_idx[k], cx_steps + 1);
            if (!c3_is_0)
                fre_encode_uniform(fre, c3_idx[k], cx_steps + 1);
            if (!c4_is_0)
                fre_encode_uniform(fre, c4_idx[k], cx_steps + 1);
        }
    }

    double est_bits =
        4 + freq_count * ((!c1_is_0 + !c2_is_0 + !c3_is_0 + !c4_is_0) * log2(cx_steps + 1));

    return (float)est_bits;
}

void AcousticMaterial::decode_raw_baseline(FlexRangeDecoder* frd, int cx_steps)
{
    float cx_steps_inverse = 1.0f / (float)cx_steps;

    freq_grid_index = frd_read_uniform(frd, MPEGI_MAX_GRIDS);
    freq_count = frd_read_uniform(frd, 64) + 1;

    name = string();
    c1.assign(freq_count, 0.0f);
    c2.assign(freq_count, 0.0f);
    c3.assign(freq_count, 0.0f);
    c4.assign(freq_count, 0.0f);

    for (int k = 0; k < freq_count; ++k) {
        int c1_idx = frd_read_uniform(frd, cx_steps + 1);
        c1[k] = (float)c1_idx * cx_steps_inverse;
        int c2_idx = frd_read_uniform(frd, cx_steps + 1);
        c2[k] = (float)c2_idx * cx_steps_inverse;
        int c3_idx = frd_read_uniform(frd, cx_steps + 1);
        c3[k] = (float)c3_idx * cx_steps_inverse;
        int c4_idx = frd_read_uniform(frd, cx_steps + 1);
        c4[k] = (float)c4_idx * cx_steps_inverse;
    }
}

void AcousticMaterial::decode_optimized_v1(FlexRangeDecoder* frd, int cx_steps)
{
    float cx_steps_inverse = 1.0f / (float)cx_steps;

    freq_grid_index = frd_read_uniform(frd, MPEGI_MAX_GRIDS);
    freq_count = frd_read_uniform(frd, 64) + 1;

    name = string();
    c1.assign(freq_count, 0.0f);
    c2.assign(freq_count, 0.0f);
    c3.assign(freq_count, 0.0f);
    c4.assign(freq_count, 0.0f);

    bool c1_is_0 = frd_read_bits(frd, 1);
    bool c2_is_0 = frd_read_bits(frd, 1);
    bool c3_is_0 = frd_read_bits(frd, 1);
    bool c4_is_0 = frd_read_bits(frd, 1);

    for (int k = 0; k < freq_count; ++k) {
        int c1_idx = 0;
        if (!c1_is_0)
            c1_idx = frd_read_uniform(frd, cx_steps + 1);
        c1[k] = (float)c1_idx * cx_steps_inverse;
        int c2_idx = 0;
        if (!c2_is_0)
            c2_idx = frd_read_uniform(frd, cx_steps + 1);
        c2[k] = (float)c2_idx * cx_steps_inverse;
        int c3_idx = 0;
        if (!c3_is_0)
            c3_idx = frd_read_uniform(frd, cx_steps + 1);
        c3[k] = (float)c3_idx * cx_steps_inverse;
        int c4_idx = 0;
        if (!c4_is_0)
            c4_idx = frd_read_uniform(frd, cx_steps + 1);
        c4[k] = (float)c4_idx * cx_steps_inverse;
    }
}

int AcousticMaterialList::encode_raw_baseline(FlexRangeEncoder* fre, bool est_only)
{
    double est_bits = 0.0;

    int step_idx = (int)lround(step * 10000.0f);
    assert((step_idx >= 1) && (step_idx <= 1000));
    // compute (int) lround(1.0f / step) more exactly, to avoid uncertainities for edge cases like step = 0.0032
    int cx_steps = (int)lround(10000.0f / (float)step_idx);
    assert((cx_steps >= 10) && (cx_steps <= 10000));

    // encode the material quantization step size
    if (count != 0) {
        if (!est_only) {
            encode_uint_compact(fre, step_idx, 10);
        }

        est_bits += estimate_uint_compact(step_idx, 10);
    }

    for (int k = 0; k < count; ++k) {
#if AM_CODING_OPTIMIZE_LEVEL == 1
        est_bits += list[k]->encode_optimized_v1(fre, cx_steps, est_only);
#else
        est_bits += list[k]->encode_raw_baseline(fre, cx_steps, est_only);
#endif
    }

    return (int)ceil(est_bits);
}

void AcousticMaterialList::decode_raw_baseline(FlexRangeDecoder* frd, int ml_count)
{
    count = ml_count;

    // decode the material quantization step size
    // valid value (matching step = 0.01f) but actually unused, for the case count == 0
    int step_idx = 100;
    if (count != 0) {
        step_idx = decode_uint_compact(frd, 10);
        assert((step_idx >= 1) && (step_idx <= 1000));
    }

    step = (float)step_idx / 10000.0f;
    // compute (int) lround(1.0f / step) more exactly, to avoid uncertainities for edge cases like step = 0.0032
    int cx_steps = (int)lround(10000.0f / (float)step_idx);
    assert((cx_steps >= 10) && (cx_steps <= 10000));

    for (int k = 0; k < count; ++k) {
        auto mat = make_unique<AcousticMaterial>();

#if AM_CODING_OPTIMIZE_LEVEL == 1
        mat->decode_optimized_v1(frd, cx_steps);
#else
        mat->decode_raw_baseline(frd, cx_steps);
#endif

        mat->name = string("mat:idx_") + to_string(k);

        list.push_back(move(mat));
    }
}

void GeometryData::addVertexList(int ver_count, float* x, float* y, float* z)
{
    assert((ver_count >= 0) && (ver_count < (1 << 18)));

    vl.count = ver_count;
    vl.x.insert(vl.x.end(), x, x + vl.count);
    vl.y.insert(vl.y.end(), y, y + vl.count);
    vl.z.insert(vl.z.end(), z, z + vl.count);

    vl.duplicates_removed = 0;
    vl.exact_duplicates_removed = 0;
}

void GeometryData::addTriangleList(
    int tri_count, int* v1_index, int* v2_index, int* v3_index, int shr_mat_count, int* mat_index)
{
    assert((tri_count >= 0) && (tri_count < (1 << 18)));
    assert((shr_mat_count >= 0) && (shr_mat_count < (1 << 10)));

    tl.count = tri_count;
    tl.shared_material_count = shr_mat_count;
    tl.v1.insert(tl.v1.end(), v1_index, v1_index + tl.count);
    tl.v2.insert(tl.v2.end(), v2_index, v2_index + tl.count);
    tl.v3.insert(tl.v3.end(), v3_index, v3_index + tl.count);
    tl.mi.insert(tl.mi.end(), mat_index, mat_index + tl.count);

    tl.duplicates_removed = 0;
    tl.exact_duplicates_removed = 0;
    tl.degenerates_removed = 0;
    tl.exact_degenerates_removed = 0;
}

void GeometryData::addAcousticMaterial(int mat_count, int mat_index, int freq_count,
    int freq_grid_index, float* c1, float* c2, float* c3, float* c4)
{
    assert((mat_count >= 0) && (mat_count < (1 << 10)));
    assert((mat_index >= 0) && (mat_index < mat_count));
    assert((freq_count >= 1) && (freq_count <= 64));

    if (ml.list.size() == 0) {
        ml.count = mat_count;
        ml.list.reserve(ml.count);
        for (int k = 0; k < ml.count; ++k) {
            auto mat = make_unique<AcousticMaterial>();

            ml.list.push_back(move(mat));
        }
    }

    assert(ml.count == mat_count);

    auto mat = ml.list[mat_index].get();

    mat->freq_count = freq_count;
    mat->freq_grid_index = freq_grid_index;
    mat->c1.insert(mat->c1.end(), c1, c1 + freq_count);
    mat->c2.insert(mat->c2.end(), c2, c2 + freq_count);
    mat->c3.insert(mat->c3.end(), c3, c3 + freq_count);
    mat->c4.insert(mat->c4.end(), c4, c4 + freq_count);

    mat->name = string("mat:idx_") + to_string(mat_index);
}

void GeometryData::setEncoderOptions(
    float quant_step, float vertex_tol, float triangle_tol, float mat_qnt_step, bool tri_rm_nomat)
{
    quantization_step = quant_step;
    vertex_tolerance = vertex_tol;
    triangle_tolerance = triangle_tol;
    material_quantization_step = mat_qnt_step;
    triangle_remove_nomaterial = tri_rm_nomat;

    // make the quantization step a multiple of the minimum quantization step (0.0001)
    int quant_step_idx = (int)lround(quantization_step * 10000.0f);
    assert((quant_step_idx >= 1) && (quant_step_idx <= 1000));
    quantization_step = (float)quant_step_idx / 10000.0f;

    // make the material quantization step a multiple of the minimum quantization step (0.0001)
    int mat_quant_step_idx = (int)lround(material_quantization_step * 10000.0f);
    assert((mat_quant_step_idx >= 1) && (mat_quant_step_idx <= 1000));
    material_quantization_step = (float)mat_quant_step_idx / 10000.0f;
}

int GeometryData::getRequiredBufferSize()
{
    if (!duplicates_and_degenerates_processed) {
        check_and_fix_mesh(vertex_tolerance, triangle_tolerance);
        duplicates_and_degenerates_processed = true;
    }

    int vertex_raw_bits = 32 + vl.count * 3 * 32;

    int raw_bits_vl_index = ((ceil_log2(max(vl.count, 1)) + 7) >> 3) << 3;
    int raw_bits_ml_index = ((ceil_log2(tl.shared_material_count + 1) + 7) >> 3) << 3;
    int triangle_raw_bits = 32 + tl.count * (3 * raw_bits_vl_index + raw_bits_ml_index);

    int material_raw_bits = 16;
    for (int mi = 0; mi < (int)ml.list.size(); ++mi) {
        material_raw_bits += 8 + ml.list[mi]->freq_count * (32 + 4 * 32);
    }

    int required_buffer_size =
        8192 + (vertex_raw_bits + 7 + triangle_raw_bits + 7 + material_raw_bits + 7) / 8;

#if GC_VERBOSE_RESULTS
    printf(
        "[GeometryData::getRequiredBufferSize] %d vl_raw_bits %d tl_raw_bits %d ml_raw_bits %d\n",
        required_buffer_size, vertex_raw_bits, triangle_raw_bits, material_raw_bits);
#endif

    return required_buffer_size;
}

int GeometryData::encodeToBuffer(int buffer_size, unsigned char* buffer_data)
{
    if (!duplicates_and_degenerates_processed) {
        check_and_fix_mesh(vertex_tolerance, triangle_tolerance);
        duplicates_and_degenerates_processed = true;
    }

    FlexRangeEncoder fre;
    MemoryBuffer mb;

    mb.setup(buffer_size, buffer_data);
    fre_init(&fre, fre_write_byte_to_memory_buffer, &mb);

    VertexListQuant vlq;
    BoundingBoxQuant bbq;
    PreprocessorGCD pp_gcd;

    vl.quantize(vlq, quantization_step);
#if GC_PRINT_VERTEX_LIST_QUANT
    vlq.print(log);
    fprintf(log, "\n");
#endif

    bbq.compute(vlq);
#if GC_PRINT_BOUNDING_BOX_QUANT
    bbq.print(log);
#endif

    pp_gcd.compute(vlq, bbq);
#if GC_PRINT_PREPROCESSOR_GCD
    pp_gcd.print(log);
#endif

    // encode the number of vertexes
    assert(vl.count < (1 << 18));
    encode_uint_compact(&fre, vl.count, 18);

    // encode the number of triangles
    assert(tl.count < (1 << 18));
    if (vl.count != 0) {
        encode_uint_compact(&fre, tl.count, 18);
    } else // vl.count == 0
    {
        assert(tl.count == 0);
    }

    // encode the number of acoustic materials
    assert(ml.count < (1 << 10));
    encode_uint_compact(&fre, ml.count, 10);

    int header_bits = fre_virtual_finish(&fre);
    (void)header_bits;

    // encode the quantization step size
    if (vl.count != 0) {
        int quant_step_idx = (int)lround(quantization_step * 10000.0f);
        assert((quant_step_idx >= 1) && (quant_step_idx <= 1000));
        encode_uint_compact(&fre, quant_step_idx, 10);
    }

    // encode the bounding box parameters
    int bbq_bits_needed = 0;
    if (vl.count != 0) {
#if BBQ_CODING_OPTIMIZE_LEVEL == 2
        bbq_bits_needed = bbq.encode_optimized_v2(&fre);
#elif BBQ_CODING_OPTIMIZE_LEVEL == 1
        bbq_bits_needed = bbq.encode_optimized_v1(&fre);
#else
        bbq_bits_needed = bbq.encode_raw_baseline(&fre);
#endif
    }

    // encode the GCD preprocessor parameters if it is enabled
#if ENABLE_PREPROCESS_GCD
    if (vl.count != 0) {
#if PPGCD_CODING_OPTIMIZE_LEVEL == 1
        pp_gcd.encode_optimized_v1(&fre, bbq, bbq_bits_needed);
#else
        pp_gcd.encode_raw_baseline(&fre, bbq_bits_needed);
#endif
    }

    pp_gcd.preprocess(vlq, bbq);
    pp_gcd.adjust_bounding_box(bbq);
#else
    (void)bbq_bits_needed;
#endif

#if (VLQ_CODING_OPTIMIZE_LEVEL == 1) || (VLQ_CODING_OPTIMIZE_LEVEL == 2)
    vlq.compute_unique_coordinates(log);
#else
    vlq.compute_unique_coordinates(NULL);
#endif

    int vertex_setup_bits = fre_virtual_finish(&fre) - header_bits;
    (void)vertex_setup_bits;
    int est_plain_bits = 0;
    (void)est_plain_bits;
    int est_unique_bits = 0;
    (void)est_unique_bits;

#if VLQ_CODING_OPTIMIZE_LEVEL == 2
    est_plain_bits = vlq.encode_raw_baseline(&fre, bbq, true); // estimation only, for comparison
    est_unique_bits = vlq.encode_optimized_v2(&fre, bbq, false);
#elif VLQ_CODING_OPTIMIZE_LEVEL == 1
    est_plain_bits = vlq.encode_raw_baseline(&fre, bbq, true); // estimation only, for comparison
    est_unique_bits = vlq.encode_optimized_v1(&fre, bbq, false);
#else
    est_plain_bits = vlq.encode_raw_baseline(&fre, bbq, false);
#endif

    int vertex_bits = fre_virtual_finish(&fre) - header_bits;

    // encode the triangle list
    int tl_est_plain_bits = 0;
    (void)tl_est_plain_bits;
    int tl_est_opt_bits = 0;
    (void)tl_est_opt_bits;

#if TL_CODING_OPTIMIZE_LEVEL == 1
    tl_est_plain_bits = tl.encode_raw_baseline(&fre, vl.count, ml.count, true); // estimation only
    tl_est_opt_bits = tl.encode_optimized_v1(&fre, vl.count, ml.count, false);
#else
    tl_est_plain_bits = tl.encode_raw_baseline(&fre, vl.count, ml.count, false);
#endif

    int triangle_bits = fre_virtual_finish(&fre) - vertex_bits - header_bits;
    (void)triangle_bits;

    // encode the acoustic material list
    ml.step = material_quantization_step;
    int ml_est_plain_bits = 0;
    (void)ml_est_plain_bits;
    ml_est_plain_bits = ml.encode_raw_baseline(&fre, false);

    int trailing_bits = 0;
    (void)trailing_bits;
#if GC_DEBUG_ADD_TRAILING_BITS
    // write 32 trailing bits for debugging to verify correct decoding
    fre_encode_bits_long(&fre, 0x31415926, 32);
    trailing_bits = 32;
#endif

    int output_bits = fre_finish(&fre) - trailing_bits;
    int material_bits = output_bits - triangle_bits - vertex_bits - header_bits;
    (void)material_bits;

#if GC_VERBOSE_RESULTS
    printf("[GeometryData::encodeToBuffer] vertex_bits %6d triangle_bits %6d material_bits %6d\n",
        vertex_bits, triangle_bits, material_bits);
#endif

#if GC_PRINT_MACHINE_STATISTICS
    // machine parsable output for automated computation of totals or statistics
    fprintf(log, "header_bits %d\n", header_bits);
    fprintf(log, "vertex_actual_count %d\n", vl.count);
    fprintf(log, "vertex_duplicates_removed %d\n", vl.duplicates_removed);
    fprintf(log, "vertex_raw_bits %d\n", 32 + vl.count * 3 * 32);
    fprintf(log, "vertex_baseline_bits %d\n", vertex_setup_bits + est_plain_bits);
    fprintf(log, "vertex_compressed_bits %d\n", vertex_bits);
    fprintf(log, "triangle_actual_count %d\n", tl.count);
    fprintf(log, "triangle_duplicates_removed %d\n", tl.duplicates_removed);
    fprintf(log, "triangle_degenerates_removed %d\n", tl.degenerates_removed);
    int raw_bits_vl_index = ((ceil_log2(max(vl.count, 1)) + 7) >> 3) << 3;
    int raw_bits_ml_index = ((ceil_log2(tl.shared_material_count + 1) + 7) >> 3) << 3;
    fprintf(log, "triangle_raw_bits %d\n",
        32 + 16 + tl.count * (3 * raw_bits_vl_index + raw_bits_ml_index));
    fprintf(log, "triangle_baseline_bits %d\n", tl_est_plain_bits);
    fprintf(log, "triangle_compressed_bits %d\n", triangle_bits);
    fprintf(log, "material_count %d\n", ml.count);
    int material_raw_bits = 16;
    for (int mi = 0; mi < (int)ml.list.size(); ++mi) {
        material_raw_bits += 8 + ml.list[mi]->freq_count * (32 + 4 * 32);
    }
    fprintf(log, "material_raw_bits %d\n", material_raw_bits);
    fprintf(log, "material_baseline_bits %d\n", ml_est_plain_bits);
    fprintf(log, "material_compressed_bits %d\n", material_bits);
#endif

    // if enabled, do automated text dump of the entire geometry data
    if (enable_dump) {
        string file_name = dump_file_prefix + "-" + to_string(vl.count) + "v-" + to_string(tl.count)
            + "t-" + to_string(ml.count) + "m-" + to_string(abs(mb.pos)) + ".geo";

#if GC_VERBOSE_RESULTS
        printf("[GeometryData::encodeToBuffer] generating text dump to %s\n", file_name.c_str());
#endif

        FILE* f = fopen(file_name.c_str(), "wb");
        assert(f != NULL);

        print(f);

        fclose(f);
    }

#if GC_VERBOSE_RESULTS
    printf("[GeometryData::encodeToBuffer] buffer_size %d buffer_needed %d buffer_overflow %u\n",
        buffer_size, mb.pos, (unsigned)mb.has_overflown);
#endif

    // in case of overflow, use a negative number to indicate the needed buffer size
    return !mb.has_overflown ? mb.pos : -mb.pos;
}

bool GeometryData::decodeFromBuffer(int buffer_size, unsigned char* buffer_data)
{
    FlexRangeDecoder frd;
    MemoryBuffer mb;

    mb.setup(buffer_size, buffer_data);
    frd_init(&frd, -1 /* not available */, frd_read_byte_from_memory_buffer, &mb);

    VertexListQuant vlq;
    BoundingBoxQuant bbq;
    // the original bounding box without preprocessing adjustments (if enabled)
    BoundingBoxQuant bbq_ori;

    // decode the number of vertexes
    int vl_count = decode_uint_compact(&frd, 18);

    // decode the number of triangles
    int tl_count = 0; // valid value, for the case vl_count == 0
    if (vl_count != 0) {
        tl_count = decode_uint_compact(&frd, 18);
    }

    // decode the number of acoustic materials
    int ml_count = decode_uint_compact(&frd, 10);

    // decode the quantization step size
    quantization_step = 0.001f; // valid value but actually unused, for the case vl_count == 0
    if (vl_count != 0) {
        int quant_step_idx = decode_uint_compact(&frd, 10);
        assert((quant_step_idx >= 1) && (quant_step_idx <= 1000));
        quantization_step = (float)quant_step_idx / 10000.0f;
    }

    vlq.step = quantization_step;

    // decode the bounding box parameters
    int bbq_bits_needed = 0;
    if (vl_count != 0) {
#if BBQ_CODING_OPTIMIZE_LEVEL == 2
#error BBQ_CODING_OPTIMIZE_LEVEL 2 not implemented
#elif BBQ_CODING_OPTIMIZE_LEVEL == 1
#error BBQ_CODING_OPTIMIZE_LEVEL 1 not implemented
#else
        bbq_bits_needed = bbq.decode_raw_baseline(&frd);
#endif
    }

    bbq_ori = bbq; // save the original bounding box

    // decode the GCD preprocessor parameters if it is enabled
#if ENABLE_PREPROCESS_GCD
    PreprocessorGCD pp_gcd;
    if (vl_count != 0) {
#if PPGCD_CODING_OPTIMIZE_LEVEL == 1
#error PPGCD_CODING_OPTIMIZE_LEVEL 1 not implemented
#else
        pp_gcd.decode_raw_baseline(&frd, bbq_bits_needed);
#endif
    }

    pp_gcd.adjust_bounding_box(bbq);
#else
    (void)bbq_bits_needed;
#endif

#if VLQ_CODING_OPTIMIZE_LEVEL == 2
#error VLQ_CODING_OPTIMIZE_LEVEL 2 not implemented
#elif VLQ_CODING_OPTIMIZE_LEVEL == 1
    vlq.decode_optimized_v1(&frd, bbq, vl_count);
#else
    vlq.decode_raw_baseline(&frd, bbq, vl_count);
#endif

#if ENABLE_PREPROCESS_GCD
    // use the original bounding box to reverse preprocessing
    pp_gcd.reverse_preprocess(vlq, bbq_ori);
#endif

    vlq.dequantize(vl);

    // decode the triangle list
#if TL_CODING_OPTIMIZE_LEVEL == 1
    tl.decode_optimized_v1(&frd, vl_count, tl_count, ml_count);
#else
    tl.decode_raw_baseline(&frd, vl_count, tl_count, ml_count);
#endif

    // decode the acoustic material list
    ml.decode_raw_baseline(&frd, ml_count);

#if GC_DEBUG_ADD_TRAILING_BITS
    // read 32 trailing bits for debugging to verify correct decoding
    unsigned int trailing_bits = frd_read_bits_long(&frd, 32);

    assert(trailing_bits == 0x31415926);
#endif

    frd_finish(&frd);

#if GC_VERBOSE_RESULTS
    printf("[GeometryData::decodeFromBuffer] bit_error_detected %d\n", frd.bit_error_detected);
#endif

    return !frd.bit_error_detected;
}

void GeometryData::enableAutomatedTextDump(const char* file_prefix)
{
    enable_dump = true;
    dump_file_prefix = file_prefix;

    // remove extension suffix, if present
    dump_file_prefix = dump_file_prefix.substr(0, dump_file_prefix.find_last_of('.'));
}

void GeometryData::print(FILE* f)
{
    vl.print(f);
    fprintf(f, "\n");

    tl.print(f);
    fprintf(f, "\n");

    ml.print(f);
}

} // namespace MPEGI
