/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef _MSC_VER /* predefined macro of Microsoft C/C++ compiler */
#include <intrin.h> /* for _BitScanReverse function */
#endif
#include <MPEGI/flex_range_coder.h>

#define FRE_DUMP_CALLS 0

int clz_u32(unsigned int n)
{
#if defined(_MSC_VER)
    unsigned long index;
#elif defined(__GNUC__)
#else
    int count;
#endif
    assert(1 <= n);

#if defined(_MSC_VER)
    (void)_BitScanReverse(&index, (unsigned long)n);
    return 31 - (int)index;
#elif defined(__GNUC__)
    return __builtin_clz(n);
#else
    for (count = 0; n < 0x80000000u; count++) {
        n <<= 1;
    }

    return count;
#endif
}

int floor_log2(unsigned int n)
{
    assert(1 <= n);

    /* floor(log2(n)) = 31 - clz_u32(n), for any n >= 1 */
    return 31 - clz_u32(n);
}

int ceil_log2(unsigned int n)
{
    assert(1 <= n);

    if (n >= 2) {
        /* ceil(log2(x)) = 1 + floor(log2(x - 1)) = 1 + [31 - clz_u32(x - 1)], for any x >= 2 */
        return 32 - clz_u32(n - 1);
    } else {
        return 0;
    }
}

/*-------------------------------------------------------------------*
 * Local function prototypes
 *-------------------------------------------------------------------*/

static void fre_shift(FlexRangeEncoder* fre);

/*-------------------------------------------------------------------*
 * fre_init()
 *
 * Initalize the range encoder
 *-------------------------------------------------------------------*/

void fre_init(FlexRangeEncoder* fre, /* i/o: Encoder state */
    fre_write_byte_type write_byte, void* output_state)
{
    fre->low = 0;
    fre->range = 0xFFFFFFFF;
    fre->pending = -1;
    fre->has_carry = 0;
    fre->carry_count = 0;

    fre->byte_count = 0;
    fre->last_byte_bit_count = -1;

    fre->write_byte = write_byte;
    fre->output_state = output_state;
}

/*-------------------------------------------------------------------*
 * fre_encode()
 *
 * Encode given cumulative count and symbol count
 *-------------------------------------------------------------------*/

void fre_encode(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int cum_count, /* i: Cumulative count up to symbol */
    unsigned int sym_count, /* i: Symbol count */
    unsigned int tot_count /* i: Total count */
)
{
    unsigned int low_offset;
    assert(tot_count <= 65536);

    fre->range /= tot_count;
    low_offset = fre->range * cum_count;
    fre->range *= sym_count;

    fre->low += low_offset;
    if (fre->low < low_offset) /* the last addition has wrapped around modulo 2 ^ 32 */
    {
        fre->has_carry = 1;
    }

    /* range was divided by up to 2 ^ 16, so at most two renormalizations are needed */
    if (fre->range < 0x01000000) {
        fre->range <<= 8;
        fre_shift(fre);
        if (fre->range < 0x01000000) {
            fre->range <<= 8;
            fre_shift(fre);
        }
    }
}

/*-------------------------------------------------------------------*
 * rc_enc_encode_fast()
 *
 * Encode given cumulative count and symbol count when total count is a power of 2
 *-------------------------------------------------------------------*/

void fre_encode_fast(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int cum_count, /* i: Cumulative count up to symbol */
    unsigned int sym_count, /* i: Symbol count */
    unsigned int tot_shift /* i: Total count as a power of 2 */
)
{
    unsigned int low_offset;
    assert(tot_shift <= 16);

    fre->range >>= tot_shift;
    low_offset = fre->range * cum_count;
    fre->range *= sym_count;

    fre->low += low_offset;
    if (fre->low < low_offset) /* the last addition has wrapped around modulo 2 ^ 32 */
    {
        fre->has_carry = 1;
    }

    /* range was shifted right by up to 16, so at most two renormalizations are needed */
    if (fre->range < 0x01000000) {
        fre->range <<= 8;
        fre_shift(fre);
        if (fre->range < 0x01000000) {
            fre->range <<= 8;
            fre_shift(fre);
        }
    }
}

/*-------------------------------------------------------------------*
 * fre_encode_symbol_fast()
 *
 * Encode an alphabet symbol when total count is a power of 2
 *-------------------------------------------------------------------*/

void fre_encode_symbol_fast(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int symbol, /* i: Symbol to encode */
    const unsigned short table[], /* i: Cumulative counts table */
    unsigned int tot_shift /* i: Total count as a power of 2 */
)
{
    /* an extra entry with value 2 ^ tot_shift must be present at the end of table */
    fre_encode_fast(fre, table[symbol], table[symbol + 1] - table[symbol], tot_shift);
}

/*-------------------------------------------------------------------*
 * fre_finish()
 *
 * Finalize the range encoder
 *-------------------------------------------------------------------*/

int fre_finish(/* o: Total number of bits produced */
    FlexRangeEncoder* fre /* i/o: Encoder state */
)
{
    int total_bit_count;
    unsigned int aligned_value, align_mask;
    int bits;
    assert(fre->range >= 0x01000000); /* range is normalized */
    assert(fre->last_byte_bit_count == -1); /* make sure finalization was not done already */

    bits = clz_u32(fre->range) + 1; /* bits = 32 - floor(log2(range)) */
    /* 32 - floor(log2(y)) = 32 - [31 - clz_u32(y)] = clz_u32(y) + 1 */

    bits++; /* conservative number of bits, because the decoder only has range available */

    assert((bits >= 2) && (bits <= 9)); /* depends on range, which is normalized */

    /* align low to the next multiple of 2 ^ (32 - bits) */
    align_mask = 0xFFFFFFFFu >> bits;
    aligned_value = (fre->low + align_mask) & ~align_mask;

    if (aligned_value < fre->low) /* the last addition has wrapped around modulo 2 ^ 32 */
    {
        fre->has_carry = 1;
    }

    fre->low = aligned_value;

    while (bits > 0) {
        fre_shift(fre);
        bits -= 8;
    }

    bits += 8;

    if (fre->carry_count > 0) {
        /* carry_count > 0, therefore the last call to fre_shift incremented carry_count */
        if (fre->pending
            >= 0) /* may actually be always true, but it is difficult to prove formally */
        {
            fre->write_byte(fre->output_state, (unsigned char)(fre->pending + fre->has_carry));
            fre->byte_count++;
        }

        while (fre->carry_count > 1) {
            fre->write_byte(fre->output_state, (unsigned char)(0xFF + fre->has_carry));
            fre->byte_count++;
            fre->carry_count--;
        }
        /* pack the last 1 to 8 bits into the MSB of the last byte, with zero padding into the LSB */
        fre->write_byte(
            fre->output_state, (unsigned char)((0xFF + fre->has_carry) & (0xFFu << (8 - bits))));
        fre->byte_count++;
        fre->last_byte_bit_count = bits;
    } else {
        /* carry_count == 0, therefore the last call to fre_shift wrote into pending */
        assert(fre->pending >= 0);

        /* pack the last 1 to 8 bits into the MSB of the last byte, with zero padding into the LSB */
        fre->write_byte(fre->output_state,
            (unsigned char)((fre->pending + fre->has_carry) & (0xFFu << (8 - bits))));
        fre->byte_count++;
        fre->last_byte_bit_count = bits;
    }

    assert((fre->last_byte_bit_count >= 1) && (fre->last_byte_bit_count <= 8));
    assert(fre->byte_count >= 1);

    total_bit_count = ((fre->byte_count - 1) << 3) + fre->last_byte_bit_count;
    assert(total_bit_count >= 2);

    return total_bit_count;
}

/*-------------------------------------------------------------------*
 * fre_save_state()
 *
 * Save the state of the range encoder (without output_state position)
 *-------------------------------------------------------------------*/

void fre_save_state(FlexRangeEncoder* fre, /* i: Encoder state */
    FlexRangeEncoder* dst /* o: Saved encoder state */
)
{
    assert(fre->last_byte_bit_count == -1); /* make sure finalization was not done already */

    dst->low = fre->low;
    dst->range = fre->range;
    dst->pending = fre->pending;
    dst->has_carry = fre->has_carry;
    dst->carry_count = fre->carry_count;

    dst->byte_count = fre->byte_count;
    dst->last_byte_bit_count = fre->last_byte_bit_count;

    dst->write_byte = fre->write_byte;
    dst->output_state = fre->output_state;

    /* the internal output_state write position must also be saved separately */
}

/*-------------------------------------------------------------------*
 * fre_restore_state()
 *
 * Restore the state of the range encoder (without output_state position)
 *-------------------------------------------------------------------*/

void fre_restore_state(FlexRangeEncoder* fre, /* o: Encoder state */
    FlexRangeEncoder* src /* i: Saved encoder state */
)
{
    assert(fre->last_byte_bit_count == -1); /* make sure finalization was not done already */

    fre->low = src->low;
    fre->range = src->range;
    fre->pending = src->pending;
    fre->has_carry = src->has_carry;
    fre->carry_count = src->carry_count;

    fre->byte_count = src->byte_count;
    fre->last_byte_bit_count = src->last_byte_bit_count;

    fre->write_byte = src->write_byte;
    fre->output_state = src->output_state;

    /* the internal output_state write position must also be restored separately */
}

/*-------------------------------------------------------------------*
 * fre_virtual_finish()
 *
 * Get the total number of bits that would be produced by finalization
 *-------------------------------------------------------------------*/

int fre_virtual_finish(/* o: Total number of bits produced */
    FlexRangeEncoder* fre /* i: Encoder state */
)
{
    assert(fre->range >= 0x01000000); /* range is normalized */
    assert(fre->last_byte_bit_count == -1); /* make sure finalization was not done already */
    assert(sizeof(fre->pending) == 4); /* ensure the sign bit computation is correct */

    /*
      byte_count bytes have already been written using the fre->write_byte function
      1 byte is pending if fre->pending >= 0, with value fre->pending or fre->pending + 1
      the number of pending bytes (0 or 1) is computed as 1 + (fre->pending >> 31)
      carry_count bytes are also pending, consisting of 0xFF or 0x00 values
      several bits will be additionally written during the finalization procedure
      the number of bits is computed as clz_u32(fre->range) + 2, as in fre_finish
    */
    return ((fre->byte_count + (fre->pending >> 31) + fre->carry_count) << 3) + clz_u32(fre->range)
        + 10;
}

/*-------------------------------------------------------------------*
 * fre_shift()
 *
 * Shift a byte out to bitstream (internal function)
 *-------------------------------------------------------------------*/

static void fre_shift(FlexRangeEncoder* fre /* i/o: Encoder state */
)
{
    if ((fre->low < 0xFF000000u) || fre->has_carry) {
        if (fre->pending >= 0) {
            fre->write_byte(fre->output_state, (unsigned char)(fre->pending + fre->has_carry));
            fre->byte_count++;
        }

        while (fre->carry_count > 0) {
            fre->write_byte(fre->output_state, (unsigned char)(0xFF + fre->has_carry));
            fre->byte_count++;
            fre->carry_count--;
        }

        fre->pending = (int)(fre->low >> 24);
        fre->has_carry = 0;
    } else {
        fre->carry_count++;
    }

    fre->low <<= 8;
}

/*-------------------------------------------------------------------*
 * fre_encode_bit_prob_fast()
 *
 * Encode one bit with probability count0 / 2 ^ tot_shift for symbol 0
 *-------------------------------------------------------------------*/

void fre_encode_bit_prob_fast(FlexRangeEncoder* fre, /* i/o: Encoder State */
    unsigned int value, /* i: Bit value to encode */
    int count0, /* i: Count for symbol 0 */
    unsigned int tot_shift /* i: Total count as a power of 2 */
)
{
    assert(value <= 1);

    if (value == 0) {
        fre_encode_fast(fre, 0, count0, tot_shift);
    } else {
        fre_encode_fast(fre, count0, (1 << tot_shift) - count0, tot_shift);
    }
}

/*-------------------------------------------------------------------*
 * fre_encode_bits()
 *
 * Encode up to 16 bits with uniform probability
 *-------------------------------------------------------------------*/

void fre_encode_bits(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int value, /* i: Value to encode */
    unsigned int bits /* i: Number of bits */
)
{
    unsigned int tmp;
    assert(bits <= 16);
    assert(value < (1u << bits));

#if FRE_DUMP_CALLS
    printf("encode_bits(%5u, %2u)\n", value, bits);
#endif

    fre->range >>= bits;
    tmp = fre->range * value;

    fre->low += tmp;
    if (fre->low < tmp) {
        fre->has_carry = 1;
    }

    /* range was shifted right by up to 16, so at most two renormalizations are needed */
    if (fre->range < 0x01000000) {
        fre->range <<= 8;
        fre_shift(fre);
        if (fre->range < 0x01000000) {
            fre->range <<= 8;
            fre_shift(fre);
        }
    }
}

/*-------------------------------------------------------------------*
 * fre_encode_bits_long()
 *
 * Encode up to 32 bits with uniform probability
 *-------------------------------------------------------------------*/

void fre_encode_bits_long(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int value, /* i: Value to encode */
    unsigned int bits /* i: Number of bits */
)
{
#if FRE_DUMP_CALLS
    printf("encode_bits_long(%9u, %2u)\n", value, bits);
#endif

    if (bits > 16) {
        fre_encode_bits(fre, value >> 16, bits - 16);
        fre_encode_bits(fre, value & 0xFFFF, 16);
    } else {
        fre_encode_bits(fre, value, bits);
    }
}

/*-------------------------------------------------------------------*
 * fre_encode_uniform()
 *
 * Encode value from an alphabet of size total with uniform distribution
 *-------------------------------------------------------------------*/

void fre_encode_uniform(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int value, /* i: Value to encode */
    unsigned int total /* i: Alphabet size */
)
{
    assert((total >= 1) && (total <= 65536));
    assert(value < total);

#if FRE_DUMP_CALLS
    printf("encode_uniform(%5u, %5u)\n", value, total);
#endif

    fre_encode(fre, value, 1, total);
}

/*-------------------------------------------------------------------*
 * fre_encode_uniform_long()
 *
 * Encode value from an alphabet of size total with uniform distribution
 *-------------------------------------------------------------------*/

void fre_encode_uniform_long(FlexRangeEncoder* fre, /* i/o: Encoder state */
    unsigned int value, /* i: Value to encode */
    unsigned int total /* i: Alphabet size */
)
{
    int n;
    unsigned int mask;
    unsigned int extra_bin;
    assert((total >= 1) && (total <= (1 << 24)));
    assert(sizeof(total) == 4); /* ensure the "pseudo" sign bit computation is correct */

#if FRE_DUMP_CALLS
    printf("encode_uniform_long(%9u, %9u)\n", value, total);
#endif

    if (total <= 256) {
        fre_encode(fre, value, 1, total);
    } else { /* 256 < total <= 2 ^ 24 */
        /* see ceil_log2(n) function for the derivation, valid for any 2 <= total */
        n = 32 - clz_u32(total - 1); /* n = ceil(log2(total)) */
        assert((n >= 9) && (n <= 24));

        n -= 8;
        mask = (1 << n) - 1;

        /* add an extra bin at the end if total is not a multiple of 2 ^ n */
        extra_bin = 1 - (((total & mask) - 1) >> 31);
        /* completely equivalent with extra_bin = ((total & mask) != 0) */

        /* split total into (total >> n) + extra_bin equal bins of size 2 ^ n */
        fre_encode(fre, value >> n, 1, (total >> n) + extra_bin);
        fre_encode_bits(fre, value & mask, n);
    }
}

/*-------------------------------------------------------------------*
 * fre_write_byte_default()
 *
 * Default callback to write byte_val byte to the FILE* output_state
 *-------------------------------------------------------------------*/

void fre_write_byte_default(void* output_state, unsigned char byte_val)
{
    int res = fputc(byte_val, (FILE*)output_state);
    assert(res != EOF); /* writing to the file failed */
}
