/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Binaural.h"

#include <MPEGI/Renderer.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Renderer/RenderList.h>

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/BaseTypes.h>
#include <EAR/ComplexBuffer.h>
#include <EAR/Exception.h>
#include <EAR/FFT.h>
#include <EAR/IRProvider.h>
#include <EAR/Listener.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>
#include <EAR/SofaHRIRProvider.h>

#include <EAR/FIRFilter.h>

#include <EAR/FixedHRIRProvider.h>

#include <chrono>
#include <cmath>
#include <numeric>
#include <thread>

#include <vector>

namespace MPEGI {

std::shared_ptr<Spatializer> makeBinauralSpatializer()
{
    return std::make_shared<BinauralSpatializer>();
}

const std::vector<float> templateSpatialDispersionFilterAzi4L = { 1.8851, 1.3414, -0.94643,
    -0.0022999, 0.98879, 1.4412, 0.90585, -0.20469, 1.8207, 0.59234, -0.49456, 0.75898, -0.71271,
    0.15531, 1.4836, 0.55975, 0.92137, 0.25426, -0.42576, 0.49106, -0.12198, 0.17224, 0.88671,
    0.21324, 0.14227, -0.93968, -0.066212, -0.0932, -0.66966, 1.9493, 0.91102, -0.36128, 0.66588,
    -0.68229, 1.4441, 1.9007, 0.28748, 0.78619, -2.142, -1.5365, 0.29466, -0.6012, -1.1083, -1.2338,
    -0.47357, 0.21661, 0.25799, 0.50719, -0.30746, 0.22622, 0.60989, 0.46916, -1.1549, -0.31592,
    -0.11128, -1.3765, 0.77415, -0.86679, 0.30029, 1.409, -0.32353, -0.44088, -0.92876, 0.25209,
    2.9322, -0.44763, -0.21962, 2.0233, -1.2859, -1.0554, 1.3965, 0.88539, -0.99837, -0.80886,
    -0.82709, -0.50952, -1.2025, 0.82418, 2.0891, 0.49768, 0.35544, 0.55602, 2.4414, 1.4896,
    -1.9328, -1.3886, 0.62531, 2.3308, 0.21226, -1.5505, 1.4585, -0.42583, -2.1545, -0.23733,
    -0.0047392, 0.99103, 0.15236, 0.41017, -0.71493, 0.67415, 1.477, -2.075, -0.91026, -0.80227,
    -0.64036, 0.50027, 0.4737, -0.037732, -1.4948, -1.1626, -0.97111, -0.23079, 2.158, -0.93907,
    0.62561, 2.8619, -1.2238, 0.68249, -0.22967, -2.3132, 0.54088, 0.22264, -0.14336, -1.8643,
    1.6966, 0.81241, -2.6671, -0.26208, -0.2364, 0.54183, 1.8319, 0.92692, -0.16398, -0.042301,
    1.0194, 0.59384, -0.17035, 0.42828, -0.46514, -1.5001, 1.1012, -0.67172, -1.6165, 0.16067,
    -0.43493, 0.36332, 0.17062, 0.31841, 0.010483, 1.7787, 0.46583, -0.74559, 0.1499, -0.062328,
    0.036548, -0.71394, -1.2466, -0.28818, 0.49365, 0.096309, -1.5605, -0.4777, 1.3364, -1.2375,
    -0.53021, 0.86708, -0.70674, -1.1301, -1.0483, -1.0366, 1.2537, 1.5829, 0.42856, 0.41962,
    0.33163, 0.87747, 0.79698, 0.58817, 0.42938, -0.54925, -1.7494, 0.57535, 2.1975, -0.30985,
    0.19281, 0.53724, 0.97838, -1.1005, -0.45414, 0.42717, -0.5864, -1.2376, -0.89444, 1.3916,
    1.0278, -1.2451, -0.31968, 1.004, -0.10647, 0.10862, 1.2345, 1.581, -0.24568, -1.2426, -0.16336,
    0.18966, 0.93218, 1.4373, -1.0878, -0.57004, 1.887, -0.52739, -0.74903, 1.1521, 1.0992, 1.1928,
    -0.80639, 0.025742, 1.0172, 0.58679, 0.094761, -0.2504, -0.40399, -1.5525, -0.1376, -1.1622,
    -0.50813, 0.063796, -1.7861, 1.0683, -0.086027, 0.61956, 1.3082, -1.755, -0.95426, 1.1732,
    1.513, 0.24195, -0.68367, -0.1487, -0.87765, -2.0024, -1.1004, 1.1408, 2.1504, 1.5583, -0.50169,
    -0.51793, -0.036382, -0.57687, -1.822, -1.6994, 0.28333, 0.048964, -0.12792, -2.4738, -1.5254,
    -0.22522, -0.46277, 1.2754, -1.6316, -1.4512, -0.29362, -1.5579, 1.7222, 1.0675, 0.39094,
    0.80198, -0.019308, 0.32201, -2.0857, -2.1224, -0.1415, -0.25619, 0.15049, 0.68237, -0.22002,
    0.50058, -0.2647, 0.95616, 2.1124, -0.47257, 0.53084, 3.0057, 0.52129, -0.71536, 1.6155,
    -0.13791, -0.46119, -0.93999, -2.8265, -1.4292, -1.1573, -0.26923, 0.39585, 0.39784, 1.511,
    1.3157, 1.3648, 0.34196, 0.14936, -0.64408, -0.40062, 0.59027, 0.91135, -0.45745, 0.91117,
    0.67185, -1.5215, 0.17215, 1.3194, -1.1268, -0.89613, 0.43674, 1.2104, 2.2666, -1.2025,
    -0.68112, 0.93081, 1.3322, -0.37393, -0.34653, 0.074589, -0.23802, 0.6611, -1.2193, -0.068246,
    0.19256, -0.56603, -1.5555, -1.4625, 0.49081, 0.22211, -0.58461, -1.3616, 0.22747, 1.273,
    0.13213, 1.2471, 0.42381, 0.44023, 0.30176, 0.1964, 0.63784, -0.46852, -0.78287, -0.51291,
    -1.1121, -1.533, 2.2722, 0.30738, -0.051575, 1.1377, -1.3299, 1.199, 0.27247, -0.16592,
    -0.32467, -2.1976, 0.79707, 0.45171, -0.11396, 1.5281, 0.74011, -1.3266, -1.2091, 1.9386,
    0.77081, 1.2352, 1.1044, 0.79787, 0.81993, -0.63923, 0.0014934, 1.0363, 0.26782, -2.1945,
    -0.82959, 1.1231, -0.91429, 0.56731, -0.3239, -0.90184, 1.216, 0.63775, -1.0125, -0.56066,
    0.30611, -0.26593, 0.66461, 0.96967, 1.0736, 1.6628, 0.81802, -0.22736, 0.20303, 0.89797,
    -0.4333, 0.69114, -0.2852, -1.4395, -0.32792, -1.0594, 0.036372, 0.31747, 0.16395, -0.62495,
    -0.2948, 0.52609, -0.72596, -1.4849, 1.3737, 2.2942, -0.83944, -0.33261, -0.67935, 0.2632,
    0.64378, 0.070675, -0.83187, -1.3621, 0.55969, -0.46518, -1.2697, -0.92838, -1.1305, -0.35009,
    -1.5699, -1.2397, -0.024887, -0.746, 0.78995, 0.11261, -0.7189, 0.27577, 1.6429, -0.04819,
    0.48135, 0.753, -0.89835, -0.93653, -0.67369, 0.29632, -0.41486, -0.78646, -0.16357, 1.6531,
    -1.6816, -0.023137, 1.2885, -0.22256, 1.3262, -0.31439, -0.83153, -0.8038, -0.16468, 0.73468,
    0.80747, -0.93496, -1.2389, 1.3676, 2.4885, 0.0075579, -3.3601, 0.42406, 0.95879, 0.18042,
    0.048067, -0.73255, 2.1972, 0.046881, -0.25773, 2.6769, 1.0329, 1.0814, 1.1006, -1.2349,
    -0.29112, 0.92271, 1.9965, 0.51026, -1.5237, -0.034136, 0.7268, 1.1458, 0.03347, 0.11994,
    1.1223, 0.22669, -0.15446, 0.47807, 0.33524, 0.036938, -0.12337, -1.1922, -0.082579, -0.54891,
    0.01796, 0.73568, -0.6512, -0.39291, 1.2906, 0.47413, -0.5478, -1.2419, 0.45106, 0.87875,
    0.44776, -0.32967, 0.077977, -0.18238, -0.14245, -0.31018, -2.4049, -1.9576, -1.5272, 1.4696,
    -0.066605, 0.35396, 3.6372, 0.30864, -0.60084, -0.74512, -1.0826, 0.45355, 1.0313, -0.27307,
    -0.41372, 0.66811, 1.3798, 1.6432, -0.61296, -1.8251, -1.3808, 0.15039, 0.44825, 0.1049,
    -0.19996, 1.5007, -0.074169, -0.47377, 2.324, 1.244, -0.9842, -1.3003, -1.2345, 0.021463,
    -0.19201, -0.057094, 0.47427, -0.18103, 0.36028, 0.89696, 0.0002241, -0.34101, -2.0678,
    -0.35073, 2.0354, -0.41844, 0.17491, -0.87469, -0.98279, 1.3595, -0.17279, -0.50881, 0.91399,
    -1.2112, 0.43936, 1.013, 0.78833, -0.19166, -0.56061, -0.35643, 1.1484, 0.34582, 0.13736,
    2.0375, -0.50189, -0.62823, 1.2324, 0.028023, -1.6654, 1.0996, 0.89769, -0.57706, 0.23164,
    0.502, 1.5745, -1.1387, -2.0202, 0.081976, -0.61196, -1.0216, -0.03905, -0.03142, 0.16238,
    -0.54101, 2.2796, 1.8255, 0.35508, 1.0244, -0.090408, -1.7015, -1.1112, -0.23899, 0.45999,
    0.32527, -1.743, 0.93785, 0.383, 0.0011851, 1.0963, -1.0476, 0.21869, 1.8064, 0.7521, 0.031923,
    -1.0341, 0.87797, -1.1746, -0.73876, 0.64769, 0.55481, 0.2874, -0.38528, -0.27698, -0.92587,
    2.7035, 0.10127, 0.10259, 0.84137, -1.4944, 1.1726, 1.0888, 0.73451, 1.1671, 0.76776, 0.82625,
    -0.6032, -1.26, -0.97156, -0.40604, -1.2246, -1.6539, -1.7923, -0.36546, -1.1928, -0.68116,
    2.009, -0.98855, -1.2862, 0.70962, -1.3275, 0.54642, 2.3815, -0.03765, -0.92236, -1.9742,
    -0.086957, 1.5332, -1.1357, 0.43163, 0.27794, -0.653, -0.025559, -0.33756, -0.11669, -1.0553,
    -0.097422, -0.46209, -1.7241, 0.37894, 1.4668, -3.0278, -0.59254, 0.17241, 0.14883, 0.39733,
    -0.88392, 0.97449, -0.71204, -0.044824, -0.9691, -0.78373, 0.097861, 1.1341, 1.17, -1.4466,
    0.23358, -0.10334, 0.75049, 0.91693, -0.83555, 0.73117, -0.21269, 0.79311, 1.2242, 0.88908,
    0.75052, 0.26755, 0.56912, -1.251, -0.16399, 0.34424, -1.2486, -0.66139, 0.8219, 0.68039,
    1.4426, -1.5162, -1.3877, 0.21503, -0.042725, 2.7976, 0.71211, 0.013552, -0.3893, -2.366,
    0.26273, 1.6795, 0.83604, -0.47941, -0.5707, 1.0274, 0.88531, -0.48918, 0.57961, 0.82575,
    1.3595, 0.16566, 1.1078, -0.33823, -1.404, 0.69607, -1.8622, 0.074717, 1.4778, -1.8682, -2.0655,
    0.7506, 1.3301, -0.6443, 0.23906, 0.52738, -0.85608, 1.362, -0.93056, 0.58876, 2.7884, -0.51813,
    -0.46731, 0.10754, -0.98527, -0.33371, -0.088166, -0.83007, 0.040104, 0.78619, -0.085668,
    -0.44826, 0.24394, -0.67234, 0.82832, 0.17496, 0.13062, 2.8271, 1.4975, -0.40908, 0.1359,
    0.9072, -1.7599, -0.5453, 2.1171, -0.014887, -2.5589, -0.57775, 1.6834, -0.080809, 0.093551,
    0.61804, -1.0751, -1.2765, -1.3712, -0.51046, -1.0615, -1.1314, 0.4715, -0.21498, 0.97776,
    -1.1221, -0.48521, -1.4482, -0.21924, 1.2974, -0.28956, 1.7851, 0.071719, 0.6046, -0.29271,
    -0.17605, 0.9617, -1.5625, -0.052854, -0.95211, -1.1408, -0.11444, 1.1291, -0.10331, -1.2746,
    0.3409, 1.1205, -0.007955, 0.72281, -0.5432, -1.0362, 0.70673, -1.4891, -1.4421, -0.79956,
    0.34391, 1.7518, -1.2119, -1.0638, 3.0957, 1.0524, -1.5077, 0.40659, -0.13836, 0.70665, 0.1415,
    -1.6408, -0.21954, -1.9064, -1.1034, -0.22287, -2.893, 0.64603, 0.54533, -0.37797, 2.0813,
    0.0096322, 1.4301, 1.544, 0.26251, -0.60614, -0.81926, -0.055482, -0.19908, -0.40061, -1.3979,
    -0.22105, -0.49838, -0.18867, 0.2221, 1.195, 0.92599, -0.47433, -1.0105, -1.0042, 0.5298,
    -1.0114, 0.12258, -0.55754, -2.2667, 1.9457, 1.1418, 0.47313, -0.16508, -1.1121, 1.1526,
    -1.5508, 0.085199, 1.257, -0.032102, 0.25237, 1.5262, 0.74626, -0.67467, 0.5987, -2.4428,
    -0.96545, 1.0493, 0.84523, -0.71383, 0.64297, -0.91477, -0.90476, -0.076182, -0.52942, 0.71457,
    -0.44154, 1.8578, 0.46729, 1.1348, 1.0283, -1.7163, 0.3485, -0.56561, -1.9252, -1.6894, 0.33806,
    1.1391, 0.019754, 0.075867, -0.40031, -1.2373, -0.55594, 1.8686, -0.18068, -0.32389, -1.1951,
    -1.3943, 1.0211, 0.86427, 0.41696, -0.064811, -0.52159, -1.4744, 0.10421, 1.4487, 1.3691,
    0.30712, 0.51162, 1.143, -0.69583, 0.99675, -0.7941, -1.7184, 0.49574, -0.5451, -0.40853,
    0.054156, 0.29137, 0.1873, 0.87522, -0.26537, -0.21206, 1.0851, 0.77455, -0.52056, 0.21701,
    0.55345, -0.26027, -0.21041, 0.72292, 0.74644, -0.92501, 0.95202, 0.33238, -1.3193, -0.60032,
    -0.58882, -0.36836, -0.45182, -0.11879, 1.2434, -1.0261, -1.6561, -0.58256, -1.8611, 1.5295,
    0.8271, -1.5577, 0.97718, 1.1691, 0.032562, 0.021012, -0.73012, -0.69482, -0.95646, 0.25551,
    0.57942, -0.020484, -1.4317, -1.0873, 0.93029, 0.68092, 1.3767, -0.4535, 0.031838, 0.10265,
    -0.062462, 0.1742, -0.76579, -1.0554, -1.7138, 0.92156, 0.5084, 0.61305, 0.38835, -1.4137,
    -0.16386, -0.60906, 0.99488, -0.15377, -0.43294, 0.63945, -0.93808, 0.046884, 0.28169, -0.35728,
    -0.0070352, -1.3972, 0.081111, -0.67337, -1.5621, 0.039233, 0.24375, -0.39277, 0.44538,
    -0.21229, 1.0072, 0.7884, -0.49646, 1.4641, 0.37058, -0.029677, 0.22157, 0.87804, 0.44536,
    0.19623, 1.4059, 0.13539, 1.9609, 0.90209, -0.01327, -1.9248, -1.9553, 1.6211, 0.99334, -0.2009,
    -1.9405, 1.3016, 0.2025, -0.23645, 0.90779, -0.87049, -1.2309, -0.48495, 1.0632, -0.77715,
    -0.43676, 1.1997, 2.1881, 0.66908, 1.3644, 1.4551, -1.3083, 0.45117, 0.098604, 0.19604, 0.48614,
    -2.0011, 1.4728, 1.9863, 0.32458, 1.5602, -0.73228, -0.90268, -0.18635, -1.4851, -0.47345,
    1.4504, -0.22424, -1.1756, 0.96195, 0.5339, 0.78421, 1.1679, -0.90154, 0.59978, 1.1049, -0.4953,
    -0.73007, 0.34732, -0.039512, -1.1357, 0.67044, -0.88231, 0.69941, 0.9471, -1.9209, -0.57087,
    -1.6958, -0.138, -0.56867, -0.85448, 0.32733, -1.104, -0.088407, -0.22002, -0.46841, 0.15759,
    -0.37281, -1.5106, 1.1484, -0.5359, 0.054622, 1.5979, -0.35963, 3.6665, 0.18604, -1.4568,
    -1.5638, -1.1508, 0.22391, -0.047055, 0.71316, -0.74877, 0.56384, 2.2421, 0.85364, 0.56248,
    -0.87945, -1.1682, 0.9451, 0.6915, -0.077629, -1.1818, -0.17716, -1.6785, -1.3966, -0.00312,
    2.2395, 1.3149, -0.18159, 0.57985, -0.94257, 0.66172, 1.7027, -0.31718, -0.94815, 0.64889,
    0.36113, -1.784, -1.1272, 2.5596, -0.18054, -1.2766, 1.9195, 0.63457, 1.5751, 2.1586, -1.1666,
    -2.3267, -1.5599, -0.16989, 0.23129, 0.10809, -0.099515, -0.1756, 0.39558, 2.3103, 1.1172,
    1.4088, -0.61327, -0.13879, 0.1258, -0.89995, 1.9482, 0.62827, 1.2225, 0.37326, -0.38694,
    0.77359, 0.37657, 0.47068, -1.0603, -1.0868, -1.1322, -0.40877, 0.24499, -2.3171, -0.12558,
    -0.25279, -1.3003, 1.2074, 1.0945, 1.8136, 2.5855, 1.6878, 0.24752, 0.69339, 2.0613, -0.85689,
    -1.4839, 0.49037, -1.4878, 1.6408, 0.98252, -3.1073, 1.4137, 0.59122, -1.7983, 0.95415,
    -0.59267, 0.028297, 0.075472, 0.69037, 1.7256, 1.2552, 1.4722, 0.53618, 0.65475, 1.8263, 1.3225,
    -0.50106, 0.34231, 0.98076, -1.2248, 0.086563, 0.70364, -1.1404, 1.4855, 0.63401, -1.0052,
    -0.56062, 1.4559, -0.18931, -1.3686, -1.2008, -0.33159, -0.28738, -0.5392, -1.3028, -3.2032,
    -1.2642, 0.28874, 1.2, -0.85951, -0.3494, -0.73112, -0.15607, 1.7478, 1.0546, 0.18679, -1.7647,
    1.2978, 0.47215, 0.51319, -2.0868, -2.4207, 0.079512, 0.16965, -1.0297, -0.84355, -0.73938,
    0.33954, 1.4896, 1.1705, 0.76244, 0.58082, 1.8582, -1.3445, -0.68385, 0.32823, 1.4576, 1.4928,
    -1.4179, 0.39287, 1.9265, -1.7712, -1.8243, 1.1277, 1.1727, 1.1575, 0.12053, 2.7316, 1.96,
    -0.5638, 0.8561, -0.028812, -1.0375, -1.3379, 0.86438, 0.10214, -1.4781, 0.21744, -0.04637,
    -2.485, -2.0821, 0.38918, -0.69387, -2.4561, -0.7664, 0.94953, -0.19945, -1.2875, 0.21535,
    2.0149, 1.7681, 1.0109, 0.053485, 0.88251, -0.14319, 0.16218, 0.038135, 0.70807, -0.13572,
    0.049937, -1.0621, -0.9522, 1.071, -2.0988, -0.46026, -0.162, -0.2635, -1.1812, -0.083951,
    -0.12916, -1.0836, -1.5345, -1.5286, 0.75613, -1.0262, 0.17711, 2.4183, 1.9178, 0.45965, 1.3011,
    -1.1437, -0.080648, 1.662, 1.3997, -0.97531, -0.33742, 2.3715, -1.5555, 0.061547, 0.29282,
    -0.14715, -0.21413, 1.491, 0.67859, -1.0014, -0.63877, 0.29502, -0.39921, -1.5649, -1.4655,
    -1.0633, 0.060291, 1.2752, 1.5975, 0.66559, 0.79621, -0.27771, 3.9322, -0.17422, -2.8988,
    -0.64256, -0.035866, 0.45182, 1.7466, -0.4418, -1.6681, 0.32632, 0.088848, -0.13264, -2.6827,
    -0.039628, 0.14426, -2.4108, -0.31075, 0.22664, 1.3976, 0.38042, -0.6491, 0.56804, -0.6117,
    -0.163, -1.4403, 0.6444, 0.32245, -0.50725, 1.6954, -1.4665, -0.74155, -1.5187, 0.48159,
    0.19066, -1.9828, 0.37601, -2.1533, -0.24562, -0.18498, 0.37735, 3.3855, -0.17819, 1.9374,
    -1.4737, -0.46226, 0.39557, 0.041312, 1.666, -0.15342, 0.082775, 0.2045, 0.29464, -0.36529,
    0.67643, 1.8706, 1.2919, -0.26429, -0.23554, -0.59964, -0.26342, -0.87437, -0.30131, 0.060096,
    0.994, 0.26277, 0.56049, 2.3882, -0.78876, 0.24919, 0.77415, 1.1389, -0.63739, -0.16345, 1.0443,
    -1.3892, -0.97766, 0.12252, -0.84555, -1.1002, -0.6454, -0.35156, 0.012707, 0.21588, -0.66209,
    1.0655, 1.6524, -0.98077, -0.23999, 0.4741, 0.87485, 0.72076, 0.88723, -0.43524 };
const std::vector<float> templateSpatialDispersionFilterAzi4R = { 0.33704, 1.8552, 1.5559, 1.1498,
    0.29595, -0.8019, 0.52389, -0.66144, 0.54709, 2.5769, -0.87282, -0.24696, 0.22304, -0.064749,
    1.5146, 0.88741, 0.56086, 0.16268, 1.1603, -0.60712, -0.34242, 1.5964, 0.72092, 0.39141,
    -0.84366, -0.84313, 0.74418, 0.053387, -0.10267, -0.82628, 0.68397, 2.3725, -0.55335, -0.41403,
    1.9029, 0.63753, -0.4045, 0.78201, 0.56852, -1.12, -1.7211, -0.47454, -1.3491, -1.0273, -0.279,
    0.53526, 0.52262, -0.52169, -0.29594, -0.30505, 0.4919, -0.21775, 0.15409, 0.85339, -0.50445,
    -1.1324, 0.092041, 0.58563, 0.39248, -1.4605, -0.20777, 1.5919, -0.86334, 0.84079, 1.5304,
    -0.94995, -0.63923, 2.609, -0.41717, -1.5121, 1.5954, -0.84186, 0.25923, 0.32429, -2.4553,
    -0.14961, 1.009, -0.66521, 1.4325, 1.4663, 0.80996, -1.0873, 0.82252, 0.54403, -1.2881, 1.0047,
    -0.41165, 1.266, 0.23886, -0.49206, 1.0506, -0.83075, -0.97906, -0.34056, -0.034318, 0.41944,
    0.11133, 0.90768, 0.62157, 0.39752, -0.13391, -0.57207, -1.8574, -0.23518, -0.51491, -0.46862,
    1.7918, -1.8901, -1.4628, -1.026, -0.65574, 1.0677, 0.77964, 0.47473, 1.3111, 1.0101, 0.016137,
    -0.30027, -1.5499, -1.531, 0.21709, 0.02692, 0.33537, -0.11021, -0.51926, 0.91416, -1.1898,
    -1.0827, -0.69013, 1.6389, 1.4012, 0.56894, 0.45598, 1.4704, 1.2893, -1.2394, -0.24188, -1.2206,
    0.11813, 0.38838, -0.50556, 0.26342, 0.03092, -0.35421, -0.58325, 0.24278, 1.3958, -0.29391,
    -1.1698, 0.98864, 0.95754, -0.28964, -1.1917, 0.58401, 1.6481, -0.20266, -2.0118, -0.80588,
    0.13727, 0.68899, 0.21898, 0.086773, -1.2007, -0.77953, 0.78799, -1.3044, -0.95748, -0.94668,
    -0.46495, -0.68115, -0.50627, 1.4451, 1.307, 0.83167, 0.49846, 0.34255, 1.053, -0.3802, -1.471,
    1.6847, 0.7785, -0.87307, 1.1198, -0.49134, -0.45272, 1.9656, -0.78851, -0.7673, 1.2983,
    0.36764, -1.3388, -1.3338, -0.14936, -0.53219, 1.2894, 0.71968, 0.77959, -0.47448, -1.7105,
    0.32192, 1.599, 1.0707, 0.019071, 0.62937, 0.0046704, -1.0743, -0.26611, 1.4551, -0.20943,
    1.1382, 1.0842, -1.2566, 1.9956, 0.51984, 0.019087, 0.78675, 0.3818, 0.00049256, 0.65056,
    0.50333, -1.5187, 1.7538, -1.161, -2.7199, 1.2423, -2.0084, -0.58767, 0.94025, -0.21432,
    0.33092, 0.029031, -0.8362, -0.52163, -1.1983, 0.70371, 1.9852, -0.78991, 1.5322, 0.41272,
    -1.975, -0.42063, -0.94821, -1.1931, 1.2805, 0.48645, 0.2869, 1.1124, -1.2244, -1.1608, 0.10938,
    0.33888, -1.639, -2.2469, 1.2495, 0.86219, -2.0408, -1.6087, -0.45979, -0.45242, -0.99223,
    -1.1851, -2.1266, -0.74243, 1.9099, 1.0189, 1.1119, -0.15854, -0.6576, 0.93811, -0.70401,
    -0.79916, -0.69329, -1.8685, -0.084733, -0.12283, -0.051298, 0.38029, -0.67276, 1.8527, 2.2858,
    -0.49767, 0.32174, 2.1044, 3.1248, -0.071749, -1.0253, 2.0693, -1.0513, -1.2872, -0.75784,
    -1.5602, 0.45591, -0.48604, -2.3631, -0.87964, 0.6573, 1.5335, 2.7387, 0.28443, 0.21852,
    0.10627, -0.12449, 0.20125, -0.028731, 0.65568, 0.1833, 1.4757, 0.24608, -0.73643, -0.98884,
    -0.24246, -0.72545, -0.095471, 1.2128, -0.47347, 2.1234, 1.6192, -1.2881, -0.083671, -0.2872,
    -1.5767, 1.184, -0.14623, -0.53203, 1.5546, -0.75153, 0.075176, -0.29992, -0.88358, -0.11278,
    0.29812, -0.68469, -3.0627, -0.49166, 0.19398, 0.068504, 1.2363, 0.26959, 0.025349, 0.33712,
    0.53274, 0.45818, 0.65563, 0.66046, 0.26902, -0.84743, -1.589, -0.086598, -0.011354, -0.88862,
    0.77249, -0.87089, 1.7522, 1.4763, -1.7303, 0.56167, -0.33832, 0.46719, -0.59106, -0.36334,
    1.3962, 0.28315, -1.1496, 0.63594, 1.517, -2.0234, 0.44827, 1.6261, 1.2468, 1.817, -0.95394,
    0.71779, 0.77603, 0.77744, 1.194, -0.69959, -0.93329, -0.92455, 0.30441, -0.049116, -0.13609,
    0.55469, -0.6371, -1.5834, 1.4743, -0.24954, -0.40993, 0.38456, -0.092218, 0.95002, -0.21552,
    -0.089123, 2.0048, 0.81859, -0.47682, 0.72501, 1.0292, 0.58896, -0.78309, -0.31489, -1.3425,
    0.54391, 1.2229, -1.3552, -0.61795, -0.061518, -0.49812, -1.732, 1.1944, 1.1786, -0.77453,
    0.44529, 0.89877, -0.48052, 0.23, -0.010488, -0.45732, -0.50447, 1.0143, 1.4025, -1.03, -1.2556,
    -0.36858, 0.91854, -2.3942, -2.6824, -0.31797, -1.8175, 0.51033, 0.22296, -0.75284, -0.67553,
    0.48786, -0.32356, 0.085113, 1.5534, -0.34831, -0.37006, 1.1978, -0.0027619, -1.4551, -0.81313,
    -0.95102, 0.61799, -0.088164, -0.9666, 0.0059797, -0.76823, 1.2901, -0.31528, -0.39783, 1.3573,
    -0.27944, -0.48675, 0.60418, 0.31995, 0.37287, -0.14901, -1.1554, -0.34517, 1.3302, 0.57795,
    0.092785, -0.55833, -0.34171, -0.22151, 0.66335, 0.49831, -0.80389, 1.4837, 0.59337, 0.23949,
    2.0932, 0.19687, -0.83747, 1.9598, 0.49214, 0.84913, 0.2026, -0.19694, 0.17104, -0.88549,
    1.2027, 2.1284, 0.78125, -0.17787, -1.1086, 0.95265, 1.6233, -1.1052, 0.011355, -0.88489,
    0.58229, -0.18906, -0.64274, 0.80716, -0.46034, -0.17627, -0.42372, 1.5244, 0.089254, -0.64681,
    0.0078003, -0.55643, 0.70197, -0.29594, -0.22835, 0.66337, 0.63272, 0.72066, -0.21104, -2.2308,
    0.29007, -1.0276, -1.6288, -0.18517, 0.34936, -0.28661, 0.12428, 2.7762, 0.062884, -0.25702,
    -1.0724, -1.4431, 1.5073, -0.44252, -0.97865, 0.40451, 1.4464, 0.24852, -0.013475, 1.105,
    -0.67392, 0.2782, -1.3171, -1.2994, 1.3763, -0.017518, 0.69087, 0.36684, 0.50159, 1.299, 0.5254,
    -0.62122, -1.3943, -1.7058, 0.52703, 0.74739, 0.0106, -0.044392, 0.0050638, 1.1411, -0.54949,
    -0.4152, 0.77568, -0.68106, 0.33867, 0.92296, -0.46335, -0.30179, -1.6231, -1.6172, 1.0904,
    -0.90463, 0.38892, 1.7374, -0.87816, 0.85408, 0.49735, 1.1521, -1.377, -1.2081, 0.29667,
    -0.40238, 1.3198, 1.2231, 1.1266, 1.1561, -0.15888, 0.35422, 0.79558, -0.42835, -1.2696,
    -0.29916, 0.64598, 0.075794, 0.34819, 0.64148, 0.17849, 1.093, -1.1214, -2.1301, -1.401,
    -0.5865, 0.487, 1.0779, 0.51937, 0.70704, 1.1146, 1.4153, 0.74479, -0.60093, 0.12953, -0.249,
    0.18788, -0.61264, -0.42081, 0.55264, -1.1526, -0.22557, -0.3344, 0.63683, 1.0571, 0.074912,
    0.82951, 0.19649, 0.37998, -0.32912, -0.50012, -0.087908, -1.8437, 0.18846, 1.6438, 0.70178,
    -0.68383, 0.32705, -0.59778, -0.069792, 0.49, 1.0508, 1.0392, -1.3976, 0.73296, 0.8593, 1.0681,
    0.62332, 0.2509, 1.4719, 0.38228, -0.22802, -1.5836, -1.967, -0.93725, -0.10295, -1.0985,
    -1.6327, -0.070733, -0.6365, 0.59223, -0.26226, -2.1457, -0.96183, 0.58327, 0.2515, 0.13263,
    0.72086, 0.38575, 0.73918, -0.75213, -0.55062, -0.99559, 0.89334, 0.92862, -0.39449, -2.1194,
    -1.1847, 1.6586, -0.5555, -1.8634, 0.13825, -0.41866, -1.7278, 1.4808, 0.65387, -2.5345,
    -1.7106, 2.0064, 1.1408, -2.0511, 0.59862, 0.21942, -1.9931, 0.76636, -0.69259, -1.1741, 0.5761,
    0.96149, -0.17797, -0.95562, 0.61296, 1.0413, -0.085247, -0.74634, 0.11808, 0.77916, 1.183,
    0.89734, 1.1209, 0.65507, 0.17496, 0.82561, -0.035875, 0.91258, 0.9509, -2.0915, -1.2121,
    -0.30108, -0.38053, 0.55359, -0.49203, -0.039222, 0.41694, 0.60789, 1.4381, 1.2282, -0.81418,
    -1.3535, -1.0761, 1.7111, -0.55229, 0.59931, 3.0443, -0.32315, -0.78287, -0.23602, 1.7729,
    0.93005, -0.3961, 1.1237, 1.0485, -0.80548, -0.027681, -1.1923, -0.40887, -0.79284, -0.55238,
    0.70318, 1.2776, -0.019774, -0.90922, 0.65507, -0.9209, 0.066807, 0.51621, -2.2343, 1.3425,
    0.8099, -0.95493, 1.0883, 0.14033, 1.2051, -0.094286, -0.38591, -0.74, -2.2652, 0.52709,
    0.40585, 0.16463, -0.54107, -1.0449, 1.2798, 0.71493, 0.84614, -0.17936, 0.46717, 1.6372,
    1.0552, 1.1256, 0.47366, -0.86937, -1.0255, 1.2103, 0.17531, -1.464, -1.2289, -0.56568, 0.37409,
    -0.041711, 1.5229, 0.47519, -0.16121, -1.6774, -1.9766, 0.86582, -1.0764, -1.0976, 0.55111,
    -1.5532, 0.43097, 0.78801, -2.3539, -1.667, 1.0349, 0.20383, 0.13893, 0.70712, -0.16359, 0.9384,
    -0.7775, 0.50179, 1.4519, -1.4373, -0.30179, -0.74156, -0.54495, -0.37917, 0.69651, -0.62169,
    1.2675, 0.79003, -1.271, -0.6222, 0.021504, 0.44991, -0.94106, -1.9431, 0.11631, 1.047, -1.1258,
    -0.18238, 0.58775, -0.23439, 0.33059, 1.8423, 0.34121, -0.81267, 0.05624, 0.15172, -1.5359,
    -0.55776, 0.1886, -0.27966, -1.2302, -1.86, -0.25435, -1.1575, -0.32568, 1.3999, -0.52897,
    0.81488, -0.0076271, 0.70344, -0.028607, -0.8689, 2.6373, -0.90456, -1.5622, 0.78061, -0.15256,
    -0.42505, 0.96154, -0.12014, -0.32566, 0.41901, -0.33607, 1.507, -0.77138, -1.5959, 1.0757,
    -0.97569, -0.91751, 0.15427, -1.4128, 0.64319, 0.77505, 0.35153, -0.068519, -0.13968, -0.38263,
    -1.197, 0.98241, 0.9884, -0.52542, 0.043785, 0.58224, 0.30032, 1.0727, 0.45644, -1.9095,
    -0.13341, 0.23684, -0.43355, 0.59424, -0.16034, 0.60438, 0.54342, -0.79794, -0.84868, 0.4583,
    -0.43446, -0.2027, 0.29193, 1.3834, 1.0838, 0.60274, -0.20582, -1.1924, 0.29398, -2.1465,
    -0.86608, -0.010905, -0.10809, 0.5573, 0.72338, 0.11933, -0.25956, -1.2342, -0.48209, -1.1702,
    0.8056, 1.3086, -2.0599, 0.17172, -0.811, 0.75286, 2.1094, -1.4289, 0.30489, -0.54365, -0.11663,
    1.8204, 1.5472, 1.3812, -0.23902, -0.75983, -0.56053, 0.19655, -0.11889, -0.024891, -1.0343,
    -1.3407, -0.34606, 0.078127, 1.787, 0.052711, -0.99729, 1.212, -0.71616, 0.7305, 0.92835,
    0.53485, 0.036136, -0.12721, 0.31347, -0.19024, 1.6087, -0.059717, 0.28888, 0.18797, -1.4319,
    -0.99523, -1.5278, -1.8524, 2.4909, 0.96329, -1.9591, -0.45869, -0.35718, -0.39169, -0.95629,
    -0.83167, -0.88986, 0.55599, 0.9224, 1.0862, 0.79843, 0.60553, -0.528, -0.54671, -0.38303,
    -0.9454, -0.078765, -0.95739, 1.075, 0.61077, -0.98378, -0.6026, -0.24061, 1.1547, 1.3901,
    -0.72881, -0.16609, -1.8277, -0.35833, 1.1332, -1.7111, 0.82531, 1.059, 0.13715, 0.75059,
    1.0543, -2.2335, -1.5594, -0.045526, 0.2567, 0.39568, -0.50714, 0.041196, 0.17957, -0.011558,
    -2.6584, -1.0323, 1.9567, -0.18367, -0.88119, -1.7847, -0.38159, 0.3368, 0.4363, 0.1516,
    -1.3689, 0.50722, 0.93187, 1.5275, 0.17389, 0.2949, 0.40645, 0.36684, 0.88846, 1.4203, 0.76952,
    0.10855, 0.26894, 2.3895, 0.25376, -1.6953, -0.12518, -0.93675, -0.2145, -0.16332, 0.41978,
    0.33395, -1.8554, 1.1659, 1.8946, -1.5105, -0.061268, 0.10337, -0.8616, -0.43875, -1.3936,
    1.3585, 2.1239, 1.1395, 0.82367, -0.21878, 1.0791, 0.38356, 0.34319, 0.70683, -0.62383,
    -0.056037, 0.42804, 0.14634, 0.69401, 1.227, 2.3897, -1.361, -0.14217, -0.52936, -1.656,
    -1.8544, 1.3218, 1.7732, -0.52906, 1.7429, -0.92599, 0.08697, 1.3273, 0.33381, 0.73708,
    -0.038159, -0.45331, 0.59648, -0.7201, 0.46711, 0.93125, -2.4249, 0.39834, 1.7964, -0.31333,
    -1.7536, -0.96653, -1.7653, -1.2899, 0.78943, -0.1746, -0.15463, -1.1354, -1.5331, -0.26345,
    -0.66208, 0.20485, -0.15815, -0.7595, 0.39711, -0.65171, 1.5399, 1.4928, -0.20195, 0.68991,
    0.28685, 0.098254, -1.7425, -1.4003, 0.35139, 0.26994, 1.1136, 0.44249, -0.65781, 0.63859,
    -0.96194, 1.6689, 0.30808, 0.11181, 1.3207, -2.2637, 1.0982, 0.37232, -0.86763, -0.28191,
    -1.0983, -0.41028, 1.1885, 1.1597, -0.85041, 1.0035, 0.050844, 0.38651, 0.79874, -0.12508,
    0.32351, -0.86048, 0.8073, -0.36111, -0.65969, 0.33931, -0.45306, 2.0425, 1.2442, -1.3958,
    1.0821, 0.49311, 0.036128, -0.88861, -2.01, 1.3543, 0.28695, -2.2478, -0.13565, -0.071995,
    1.7661, 1.6821, 0.33885, 0.57322, -0.0061708, 0.80808, 0.23508, -0.34033, 1.0234, 1.7307,
    -0.12226, 0.021101, 0.36503, 0.54197, 0.57415, -0.4611, -0.56372, -1.1684, 0.027439, 0.15537,
    -0.9306, -2.0143, -0.38769, -0.27463, -1.5039, 1.2383, 1.0166, 1.6031, 1.0427, 2.5844, 0.71155,
    -0.54751, 1.6813, 0.65732, -0.59946, -0.85848, 0.1142, 0.88508, 1.0068, -1.7088, -0.67397,
    0.36077, -0.033055, -0.47782, -0.33132, 1.6336, 1.4813, -0.85975, 1.8087, 1.3929, -0.13376,
    1.2894, 0.69987, 1.1426, 2.1059, -0.36684, -1.0949, 1.1906, 0.089333, 0.17633, 0.13397, 0.72654,
    2.2442, -1.654, 0.10799, -0.23958, -0.70555, -0.94461, -0.085478, 1.4369, -1.3882, -2.3178,
    -1.8045, -1.2788, -0.70669, 0.16231, -0.68577, 1.1149, -0.66789, -1.3089, -0.91223, -0.065167,
    0.42725, 1.064, 0.38384, 0.069534, 0.37599, 0.37596, -0.60273, -1.552, -1.4075, 0.0076051,
    0.2854, -1.1357, -0.84781, -0.44064, 1.0737, -0.22861, 0.73523, 0.77984, 0.5365, 0.45187,
    -1.4949, 1.2603, 1.8745, 0.19604, -0.067873, 1.1009, 1.2712, -0.74435, -0.2481, -0.49231,
    -0.21557, 0.79266, 1.0207, 1.5139, 1.6367, 0.84882, -0.54459, 1.3408, -0.28398, -1.4471,
    -1.9498, 1.0679, 1.4374, -1.0639, 0.82198, -1.242, -1.5331, -1.0809, -2.7287, -0.94559, -0.1898,
    -0.43659, -0.0082367, -1.1914, -0.42594, 1.1755, 0.91157, 1.3659, 0.66957, 0.55832, 0.63291,
    1.0257, 0.72418, -2.02, 0.003046, 1.9944, -0.90001, 0.043342, -1.2348, -0.27117, -0.94613,
    -0.25462, 0.077065, -2.0894, -0.221, -1.9531, -1.3847, 0.80325, -0.21443, -1.9434, -1.309,
    0.5086, 1.8988, 1.8301, 1.2468, -0.32472, 1.4931, 1.2705, -0.085228, -0.31084, -0.31576, 1.6693,
    -0.47475, 1.0311, 0.63006, -0.44175, 0.65529, -0.92325, 0.28645, 0.45158, 1.1932, 0.73085,
    -1.9236, 0.81218, -0.58016, -2.0183, -0.58721, -0.30354, 0.18068, -0.1505, 0.12707, 1.5491,
    1.4981, 0.62685, 2.0135, -1.1446, -0.39161, -0.259, -0.7951, 1.3221, -0.65827, 0.15617,
    -0.30129, -2.3034, 0.854, -0.43808, -1.6036, -0.68123, -2.0963, 1.892, -0.78227, -0.93879,
    -0.43246, 0.52054, 0.4836, -1.4507, 0.77917, -0.83623, -1.4031, 0.84757, 0.57253, -0.027712,
    1.3238, -1.5286, -0.64992, 0.8009, -0.43247, -0.45119, -1.0307, -0.96312, -2.1582, 0.37888,
    0.72913, 0.14049, 3.1725, 1.1284, 0.23249, -0.94105, -0.72276, 0.60969, -0.81174, 0.73446,
    2.0861, -0.3353, 0.39472, 0.87566, -0.09792, -1.1364, 1.9492, 2.0596, -0.57767, -0.19193,
    -0.9996, -0.69024, 0.37739, -0.45541, -0.37005, 1.3886, 0.33196, -0.0068382, 0.99834, 1.8246,
    0.37572, 0.76989, 0.3298, -1.0552, -0.17576, -0.37864, 0.21822, 0.0091368, -1.3222, -0.39459,
    -0.013082, -0.22519, -1.452, -1.4705, 0.21574, 0.48951, 1.2735, 0.99043, -0.68052, -0.27068,
    -0.10329, 1.0468, 0.40699, 0.74148, -1.1751 };

BinauralSpatializer::BinauralSpatializer()
    : processorInitialized(false)
    , forceFilterUpdate(true)
    , forceReinitialization(false)
    , currentHRIRPath("")
    , fs(0)
    , B(0)
{
    // set config values in one place so we don't call getInt and getIntDef in wrong order
    Config::setDef("Binaural_HRIRFilterLength_LoQ", EAR::Variant(256));
    Config::setDef("Dispersion_Early_Reflections", EAR::Variant(true));
    Config::setDef("Dispersion_Diffraction", EAR::Variant(false));
    Config::setDef("Dispersion_Filter_Length", EAR::Variant(.015f));
    Config::setDef("Dispersion_Gain_dB", EAR::Variant(0.f));
    Config::setDef("PanningHRIRconvolution", EAR::Variant(0));

    // create cosine window to shorten HRIRs if needed
    int filtL = Config::getInt("Binaural_HRIRFilterLength_LoQ");
    shortenedHRIRWindow = std::make_unique<EAR::RealBuffer>(filtL);
    float* win = shortenedHRIRWindow->data();
    for (int s = 0; s < filtL; s++) {
        win[s] = std::cos((M_PI / 2) * (s / filtL));
    }
}

BinauralSpatializer::~BinauralSpatializer() {}

void BinauralSpatializer::loadHRIRFile(const std::string& path)
{
    if (fs > 0) { // only if already registered with Renderer
        processorInitialized.store(false);
        std::this_thread::sleep_for(std::chrono::microseconds(static_cast<int>(B / fs * 1000000)));
        // TODO: find more elegant way to wait for potential requestBlock() run to finish

        hrirs.reset(nullptr);
        currentIR.reset(nullptr);

        if (path.length() == 0) {
            hrirs = std::make_unique<EAR::FixedHRIRProvider>();
        } else {
            hrirs = std::make_unique<EAR::SofaHRIRProvider>(path);
            runningMode = (PANNING_HRIR_CONVOLUTION_MODE)Config::getInt("PanningHRIRconvolution");
            if (runningMode == MODE_8CH_PANNING) {
                if (updateRep(path) == false) {
                    runningMode = MODE_CONVENTIONAL;
                }
            }
        }

        EAR_CHECK_REL(hrirs->getSampleRate() == fs,
            "The sample rate of the HRIR dataset does not match the scene's.");
        currentIR = std::make_unique<EAR::RealFrame>(2, hrirs->irLength());

        initProcessor();
    }

    currentHRIRPath = path;
}

namespace {
int getHrirLength(ItemType type, int length)
{
    int filtL = 0;
    if (type == ItemType::Reflection) {
        filtL = Config::getInt("Binaural_HRIRFilterLength_LoQ");
        if (filtL > length) {
            filtL = length;
        }
    } else {
        filtL = length;
    }
    return filtL;
}
} // anonymous namespace

BinauralSpatializer::ItemState::ItemState(ItemType type, std::size_t blockLength, int filterLength)
    : filter(blockLength, getHrirLength(type, filterLength))
{}

std::unique_ptr<Spatializer::DspData> BinauralSpatializer::update(const RenderList& renderList)
{

    if (!processorInitialized.load()) {
        return {};
    }

    if (runningMode == MODE_8CH_PANNING) {
        return updateForPanningHRIRconvolution(renderList);
    } else {
        return updateForConventionalHRIRconvolution(renderList);
    }
}

std::unique_ptr<Spatializer::DspData> BinauralSpatializer::updateForConventionalHRIRconvolution(
    const RenderList& renderList)
{
    auto dspData = std::make_unique<BinauralSpatializer::DspData>();
    auto& dspList = *dspData.get();

    std::size_t shortFilterLength = shortenedHRIRWindow->length();
    bool hrirNeedTrim = (hrirs->irLength() > shortFilterLength);

    for (auto& item : renderList.items) {
        if (item->type == ItemType::Downmix || item->type == ItemType::Airflow) {
            auto& frame = item->signal->getFrame();
            if (frame.channels() == getNbrOfChannels()) {
                dspList.downmixItems.emplace_back(&frame);
            }
            continue;
        }

        if (item->type != ItemType::Nondiegetic && item->type != ItemType::Primary
            && item->type != ItemType::ListenerVoice && item->type != ItemType::Reflection
            && item->type != ItemType::Representative && item->type != ItemType::Diffraction) {
            // skip unknown render item types
            continue;
        }

        // TODO: remove this when the SESS fade out mechanism is completed
        if (item->extent) {
            // extended sources with one input channel ("homogeneous SESSs") and
            // extended sources with multiple input channels ("heterogeneous SESSs")
            // have already been processed in previous stages, so we can skip anything with extent here
            continue;
        }

        // std::cout << item->objectName << " (" << item->idx << ") status: " << static_cast<int>(item->status) << std::endl;

        // create new resources if necessary
        if (forceReinitialization) {
            item->unsafeRemoveState<BinauralSpatializer::ItemState>();
        }
        auto& hrirFilter =
            item->getStateOrCreate<BinauralSpatializer::ItemState>(item->type, B, hrirs->irLength())
                .filter;

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        if (item->status == ItemStatus::Inactive) {
            // reset filters
            hrirFilter.flush();
            continue;
        }

        if ((item->changed & (ItemProperty::Position | ItemProperty::Status))
            || (item->position.cSpace != EAR::CSpace::LCS && renderList.listener.hasMoved)
            || forceFilterUpdate) {
            // replace filters if required

            EAR::Vector3 relLoc;
            if (item->position.cSpace == EAR::CSpace::LCS && item->type == ItemType::Nondiegetic) {
                // TODO: this is a hack for reverb loudspeakers - how would it be correct?
                relLoc = item->position.location;
            } else {
                relLoc = renderList.listener.cs.transform(item->position.location);
            }

            // TODO: move HRIR selection and exchange to block rate processing for less movement latency
            hrirs->getNearestIRForLocation(*currentIR, relLoc);

            if (hrirNeedTrim && item->type == ItemType::Reflection) { // apply window
                for (int ch = 0; ch < currentIR->channels(); ch++) {
                    currentIR->ch(ch).mul(*shortenedHRIRWindow, 0, 0, shortFilterLength);
                }
            }

            hrirFilter.setFilter(currentIR->data().get());
        }

        dspList.items.emplace_back(&hrirFilter,
            &item->signal->getFrame()[0], // TODO: handle multichannel?
            item->type, (bool*)NULL, (panningParms*)NULL, (panningParms*)NULL,
            (EAR::RealBuffer*)NULL);
    }

    // when we arrive here, all paths are updated if necessary
    forceFilterUpdate = false;
    forceReinitialization = false;

    return dspData;
}

void BinauralSpatializer::requestBlock(DspContext ctx)
{
    if (runningMode == MODE_8CH_PANNING) {
        return requestBlockForPanningHRIRconvolution(ctx);
    } else {
        return requestBlockForConventionalHRIRconvolution(ctx);
    }
}

void BinauralSpatializer::requestBlockForConventionalHRIRconvolution(DspContext ctx)
{
    auto& output = ctx.output;

    // clear output buffer (also when processing is not initialized)
    output.zero(); // TODO: make this optional for added flexibility?
    if (!processorInitialized.load()) {
        return;
    }

    auto& currentDSP = ctx.getData<BinauralSpatializer::DspData>();

    auto outL = &(output.ch(0));
    auto outR = &(output.ch(1));
    auto smL = &(smooth->ch(0));
    auto smR = &(smooth->ch(1));

    bool refOn = Config::getBool("Dispersion_Early_Reflections");
    bool difOn = Config::getBool("Dispersion_Diffraction");

    if (refOn || difOn)
        smooth->zero();

    if (false) {
        auto tempL = &(temp->ch(0));
        auto tempR = &(temp->ch(1));
        for (auto& item : currentDSP.items) {
            item.hrirFilter->process(item.signal->data(), temp->data().get());

            if (item.type == ItemType::Reflection && refOn) {
                smL->add(*tempL);
                smR->add(*tempR);
            } else if (item.type == ItemType::Diffraction && difOn) {
                smL->add(*tempL);
                smR->add(*tempR);
            } else {
                outL->add(*tempL);
                outR->add(*tempR);
            }
        } // dsp item loop
    } else {
        //faster DSP loop
        outFD->zero();
        outFDSidechainOut->zero();
        outFDSidechainIn->zero();

        if (refOn || difOn) {
            smoothFD->zero();
            smoothFDSidechainOut->zero();
            smoothFDSidechainIn->zero();
        }

        for (auto& item : currentDSP.items) {
            if ((item.type == ItemType::Reflection && refOn)
                || (item.type == ItemType::Diffraction && difOn)) {
                item.hrirFilter->processFD(item.signal->data(), smoothFD->data(),
                    smoothFDSidechainOut->data(), smoothFDSidechainIn->data());
            } else {
                item.hrirFilter->processFD(item.signal->data(), outFD->data(),
                    outFDSidechainOut->data(), outFDSidechainIn->data());
            }
        } // dsp item loop

        std::complex<float>* const tempPtr = tempFD->data();
        std::complex<float>* const tempSidechainOutPtr = tempFDSidechainOut->data();
        std::complex<float>* const tempSidechainInPtr = tempFDSidechainIn->data();
        float* finPtr = fadeIn->data();
        float* foutPtr = fadeOut->data();

        // time domain crossfade for output
        ifft->transform(outFD->data(), tempFD->data());
        ifft->transform(outFDSidechainOut->data(), tempFDSidechainOut->data());
        ifft->transform(outFDSidechainIn->data(), tempFDSidechainIn->data());
        float* const outLPtr = outL->data();
        float* const outRPtr = outR->data();
        for (int s = 0; s < B; s++) {
            outLPtr[s] += tempPtr[B + s].real() + tempSidechainOutPtr[B + s].real() * foutPtr[s]
                + tempSidechainInPtr[B + s].real() * finPtr[s];
            outRPtr[s] += tempPtr[B + s].imag() + tempSidechainOutPtr[B + s].imag() * foutPtr[s]
                + tempSidechainInPtr[B + s].imag() * finPtr[s];
        }

        // time domain crossfade for output
        ifft->transform(smoothFD->data(), tempFD->data());
        ifft->transform(smoothFDSidechainOut->data(), tempFDSidechainOut->data());
        ifft->transform(smoothFDSidechainIn->data(), tempFDSidechainIn->data());
        float* const smoothLPtr = smL->data();
        float* const smoothRPtr = smR->data();
        for (int s = 0; s < B; s++) {
            smoothLPtr[s] += tempPtr[B + s].real() + tempSidechainOutPtr[B + s].real() * foutPtr[s]
                + tempSidechainInPtr[B + s].real() * finPtr[s];
            smoothRPtr[s] += tempPtr[B + s].imag() + tempSidechainOutPtr[B + s].imag() * foutPtr[s]
                + tempSidechainInPtr[B + s].imag() * finPtr[s];
        }
    }

    if (m_smootherL && (refOn || difOn)) {
        // check for smoothing filter changes
        auto sm = Config::getFloat("Dispersion_Filter_Length");
        auto gn = Config::getFloat("Dispersion_Gain_dB");
        if (m_smoothLength != sm || m_smoothGain != gn) {
            m_smoothLength = sm;
            m_smoothGain = gn;
            configureDispersionFilter();
        }

        // convolve smooth buffer
        m_smootherL->process(smL->data(), smL->data());
        m_smootherR->process(smR->data(), smR->data());
        // add smooth buffer to output buffer
        outL->add(*smL);
        outR->add(*smR);
    }

    for (auto sig : currentDSP.downmixItems) {
        outL->add(sig->ch(0));
        outR->add(sig->ch(1));
    } // dsp downmix item loop

    if (headphoneEQOn) {
        // convolve headphoneEQ filter
        headphoneEQL->process(outL->data(), outL->data());
        headphoneEQR->process(outR->data(), outR->data());
    }

    // TODO: Limiter?
}

std::size_t BinauralSpatializer::getNbrOfChannels() const
{
    return 2; // not a magic number in a BInaural spatializer
}

void BinauralSpatializer::initializeSpatializer(
    const Bitstream& bitstream, float fs, std::size_t B, const SceneState* state)
{
    processorInitialized.store(false);
    std::this_thread::sleep_for(std::chrono::microseconds(static_cast<int>(B / fs * 1000000)));
    // TODO: find more elegant way to wait for potential requestBlock() run to finish

    this->fs = fs;
    this->B = B;

    loadHRIRFile(currentHRIRPath); // reload HRIRs to account for a possible sample rate change

    auto dispersionFilterPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_DISPF);
    EAR_CHECK_REL(
        dispersionFilterPayloads.size() <= 1, "DispersionFilter: too many payloads in bitstream");
    if (dispersionFilterPayloads.size() == 1) {
        dispersionFilterPayload = std::dynamic_pointer_cast<MPEGI::DispersionFilterPayload const>(
            dispersionFilterPayloads[0]);
        // set new defaults:
        bool dispER = dispersionFilterPayload->enableDispersionFilterER;
        bool dispDiffr = dispersionFilterPayload->enableDispersionFilterDiffr;
        Config::set("Dispersion_Early_Reflections", EAR::Variant(dispER));
        Config::set("Dispersion_Diffraction", EAR::Variant(dispDiffr));
        //only override default values if bitstream enables one or both options
        if (dispER || dispDiffr) {
            // in the bitstream the length is sent in milliseconds
            float codedSmoothLength =
                static_cast<float>(dispersionFilterPayload->codedDispersionFilterLength);
            float smoothLength = ((codedSmoothLength + 1) * 2.0f) / 1000.0f;
            m_smoothLength = smoothLength;
            Config::set("Dispersion_Filter_Length", EAR::Variant(smoothLength));

            float codedSmoothGain =
                static_cast<float>(dispersionFilterPayload->codedDispersionFilterGain);
            float smoothGain = codedSmoothGain - 21.0f;
            m_smoothGain = smoothGain;
            Config::set("Dispersion_Gain_dB", EAR::Variant(smoothGain));

            m_enableSpatialDispersion = dispersionFilterPayload->enableSpatialDispersion;
        }
    }

    configureDispersionFilter();
    if (hrirs) {
        initProcessor();
    }

    if (runningMode == MODE_8CH_PANNING) {
        // panning parameter calculation
        parameterCalc();
    }

    // get any available Headphone EQ data.
    ccHpEq = *bitstream.config
                  .GetContentCreatorHeadphoneEQ(); // content creator headphone data from bitstream
    localHpEq = getHeadphoneEQ("HeadphoneEq"); // local headphone data from the config file
    localCcHpEq = getHeadphoneEQ(
        "TargetHeadphoneEq"); // local target EQ from the config file that overrides the bitstream data.

    EAR::HeadphoneEQData targetEQ;
    EAR::HeadphoneEQData measuredEQ;

    if (localHpEq.specType != HeadphoneSpecType::unset) {
        measuredEQ = localHpEq;
    } else {
        // nothing given, so set a flat response.
        measuredEQ.specType = HeadphoneSpecType::freqgain;
        measuredEQ.centerFreq = { 0, 20000 };
        measuredEQ.gain = { 1, 1 };
    }

    //get spectral compensation settings from accessibility user config
    accessibilityEqLeft = getHeadphoneEQ("AccessibilityEqLeft");
    accessibilityEqRight = getHeadphoneEQ("AccessibilityEqRight");

    short accessibilityMode = 0;
    if (accessibilityEqLeft.specType != HeadphoneSpecType::unset) {
        accessibilityMode = 1;
        if (accessibilityEqRight.specType != HeadphoneSpecType::unset) {
            accessibilityMode = 2;
        }
    }

    if (!Config::getBoolDef("HeadphoneEQ:Disable", false)) {
        if (localCcHpEq.specType != HeadphoneSpecType::unset) {
            targetEQ = localCcHpEq;
        } else if (ccHpEq.specType != HeadphoneSpecType::unset) {
            targetEQ = ccHpEq;
        }

        if (targetEQ.specType == HeadphoneSpecType::unset && accessibilityMode == 0) {
            headphoneEQOn = false;
        } else {
            headphoneEQOn = true;
            if (targetEQ.specType == HeadphoneSpecType::unset) {
                // no headphone target EQ given, proceed with flat target response
                targetEQ.specType = HeadphoneSpecType::freqgain;
                targetEQ.centerFreq = { 0, 20000 };
                targetEQ.gain = { 1, 1 };
            }
            switch (accessibilityMode) {
            case 0: {
                auto filters = HeadphoneEQ::calcFilters(targetEQ, measuredEQ, fs, B);
                headphoneEQL = std::move(filters.first);
                headphoneEQR = std::move(filters.second);
                break;
            }
            case 1: {
                auto filters =
                    HeadphoneEQ::calcFilters(targetEQ, measuredEQ, accessibilityEqLeft, fs, B);
                headphoneEQL = std::move(filters.first);
                headphoneEQR = std::move(filters.second);
                break;
            }
            case 2: {
                auto filters = HeadphoneEQ::calcFilters(
                    targetEQ, measuredEQ, accessibilityEqLeft, accessibilityEqRight, fs, B);
                headphoneEQL = std::move(filters.first);
                headphoneEQR = std::move(filters.second);
                break;
            }
            default:
                headphoneEQOn = false; //should never reach this
            }
        }
    }
}

HeadphoneEQData BinauralSpatializer::getHeadphoneEQ(std::string prefix)
{
    std::string HpMode = Config::getStringDef(prefix + ":Mode", "");
    if (HpMode == "") {
        return HeadphoneEQData();
    };

    HeadphoneEQData HpEq = HeadphoneEQData();

    if (HpMode == "freq-gain") {
        HpEq.specType = HeadphoneSpecType::freqgain;
        auto temp = Config::getFloat(prefix + ":CenterFreq");
        HpEq.centerFreq = Config::getFloatVectorDef(prefix + ":CenterFreq", {});
        HpEq.gain = Config::getFloatVectorDef(prefix + ":Gain", {});
        EAR_CHECK_REL(HpEq.centerFreq.size() > 1,
            "Number of HeadphoneEq center frequencies in config must be greater than 1 in "
            "'freq-gain' mode.");
        EAR_CHECK_REL(HpEq.centerFreq.size() == HpEq.gain.size(),
            "Number of Headphone EQ center frequencies and gains do not match.");
    } else if (HpMode == "parametricFilters") {
        HpEq.specType = HeadphoneSpecType::parametricfilters;
        HpEq.centerFreq = Config::getFloatVectorDef(prefix + ":CenterFreq", {});

        auto filtTypes =
            std::istringstream(MPEGI::Config::getStringDef(prefix + ":filterType", ""));

        for (std::string filtType; std::getline(filtTypes, filtType, ',');) {
            if (filtType == "lowshelf") {
                HpEq.filterType.push_back(HeadphoneFilterType::lowshelf);
            } else if (filtType == "peak") {
                HpEq.filterType.push_back(HeadphoneFilterType::peak);
            } else if (filtType == "highshelf") {
                HpEq.filterType.push_back(HeadphoneFilterType::highshelf);
            } else {
                EAR_FAIL("Unknown headphone EQ parametric filter type specified");
            }
        }
        HpEq.gain = Config::getFloatVectorDef(prefix + ":Gain", {});
        HpEq.q = Config::getFloatVectorDef(prefix + ":Q", {});
        EAR_CHECK_REL(HpEq.centerFreq.size() > 1,
            "Number of HeadphoneEq center frequencies in config must be greater than 1 in "
            "'freq-gain' mode.");
        EAR_CHECK_REL(HpEq.centerFreq.size() == HpEq.gain.size(),
            "Number of Headphone EQ center frequencies and gains do not match.");
        EAR_CHECK_REL(HpEq.centerFreq.size() == HpEq.q.size(),
            "Number of Headphone EQ center frequencies and Q values do not match.");
        EAR_CHECK_REL(HpEq.centerFreq.size() == HpEq.filterType.size(),
            "Number of Headphone EQ center frequencies and filter type definitions do not match.");
    } else if (HpMode == "preset") {
        HpEq.specType = HeadphoneSpecType::preset;
        HpEq.preset = Config::getInt(prefix + ":Preset");
    } else {
        EAR_FAIL("Unknown Headphone EQ type specified, received: " + HpMode);
    }
    return HpEq;
}

void BinauralSpatializer::configureDispersionFilter()
{
    float makeupGainDispersionFilter = std::pow(10.0, m_smoothGain / 20);

    std::vector<float> template_l(
        templateSpatialDispersionFilterAzi4L.begin(), templateSpatialDispersionFilterAzi4L.end());
    std::vector<float> template_r(
        templateSpatialDispersionFilterAzi4R.begin(), templateSpatialDispersionFilterAzi4R.end());
    if (!m_enableSpatialDispersion) {
        template_l = template_r;
    }

    const int filterlength = 48000 * m_smoothLength;
    template_l.resize(filterlength);
    template_r.resize(filterlength);

    // normalizing the raw noise files
    float w_l = 0.0f;
    float w_r = 0.0f;
    for (int i = 0; i < filterlength; i++) {
        w_l = w_l + std::pow(template_l.at(i), 2);
        w_r = w_r + std::pow(template_r.at(i), 2);
    }
    w_l = 1 / sqrt(w_l);
    w_r = 1 / sqrt(w_r);

    //applying normalization and fade-out window
    for (int i = 0; i < filterlength; i++) {
        float g = (1.0 - sqrt(i / (float)filterlength)) / sqrt(filterlength / 4.0);
        template_l.at(i) = template_l.at(i) * (w_l * g);
        template_r.at(i) = template_r.at(i) * (w_r * g);
    }

    // checking energy again and apply make up gain
    w_l = 0.0f;
    w_r = 0.0f;
    for (int i = 0; i < filterlength; i++) {
        w_l = w_l + std::pow(template_l.at(i), 2);
        w_r = w_r + std::pow(template_r.at(i), 2);
    }
    float w = makeupGainDispersionFilter / sqrt((std::max)(w_l, w_r));

    for (int i = 0; i < filterlength; i++) {
        template_l.at(i) = template_l.at(i) * w;
        template_r.at(i) = template_r.at(i) * w;
    }
    m_smootherL = std::make_unique<EAR::FIRFilter>(template_l.data(), B, filterlength);
    m_smootherR = std::make_unique<EAR::FIRFilter>(template_r.data(), B, filterlength);
}

void BinauralSpatializer::initProcessor()
{
    temp = std::make_unique<EAR::RealFrame>(2, B);
    smooth = std::make_unique<EAR::RealFrame>(2, B);

    forceFilterUpdate = true;
    forceReinitialization = true;
    processorInitialized.store(true);

    const size_t K = 2 * B;
    if (runningMode == MODE_CONVENTIONAL) {
        tempFD = std::make_unique<EAR::ComplexBuffer>(K);
        tempFDSidechainOut = std::make_unique<EAR::ComplexBuffer>(K);
        tempFDSidechainIn = std::make_unique<EAR::ComplexBuffer>(K);
        outFD = std::make_unique<EAR::ComplexBuffer>(K);
        outFDSidechainOut = std::make_unique<EAR::ComplexBuffer>(K);
        outFDSidechainIn = std::make_unique<EAR::ComplexBuffer>(K);
        smoothFD = std::make_unique<EAR::ComplexBuffer>(K);
        smoothFDSidechainOut = std::make_unique<EAR::ComplexBuffer>(K);
        smoothFDSidechainIn = std::make_unique<EAR::ComplexBuffer>(K);
    } else {
        initProcessorForPanningHRIRconvolution();
    }

    ifft = EAR::IFFT::create(K);
    // predefine gain ramps for crossfading
    fadeIn = std::make_unique<RealBuffer>(B);
    fadeOut = std::make_unique<RealBuffer>(B);
    const float step = static_cast<float>(M_PI) / (2.0f * (float)(B - 1));

    for (int i = 0; i < B; i++) {
        (*fadeIn)[i] = sin((float)i * step) * sin((float)i * step);
        (*fadeOut)[i] = cos((float)i * step) * cos((float)i * step);
    }
}

} // namespace MPEGI
