/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_BINAURAL_SPATIALIZER
#define MPEGI_BINAURAL_SPATIALIZER

#include <MPEGI/Renderer/RenderItem.h>
#include <MPEGI/Renderer/Spatializer.h>

#include <MPEGI/Bitstream/DispersionFilterPayload.h>

#include <MPEGI/Helpers/Types.h>

#include <EAR/FFT.h>
#include <EAR/HeadphoneEQ.h>
#include <EAR/config.h>

#include <EAR/StereoFIRFilter.h>

#include <EAR/CartesianKDTree.h>

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>

#if defined(EAR_ENABLE_IPP)
#include <ipp.h>
#include <ipps.h>
#endif

// EAR fwd declares
namespace EAR {
class IRProvider;
class RealBuffer;
class ComplexBuffer;
class RealFrame;
class FIRFilter;
class IFFT;
} // namespace EAR

// Representative point
const size_t NUM_AZIMUTHS = 6; // [30, 90, 150, 210, 270, 330]
const size_t NUM_ELEVATIONS = 2; // [Top, Foot]
const size_t NUM_REPRESENTATIVE = (NUM_AZIMUTHS + NUM_ELEVATIONS);

namespace MPEGI {

// fwd declares
class Renderer;
struct RenderList;

//! Binaural rendering of the render items produced by the preceding pipeline
/*! 
 * Inputs: 
 * - Rendering inputs: ItemType::Nondiegetic, ItemType::Primary, ItemType::Reflection, ItemType::Representative, ItemType::Diffraction
 *      - RIs with "extent" property are skipped!
 * - RIs with ItemType::Downmix are added directly to the output buffer
 *  
 * Outputs: 
 * - Binauralized stereo output audio
 *  
 * How it works: 
 * - HRIRs are selected by the listener-relative position and convolved with the input signal
 * - HRIRs for ItemType::Reflection can be shortened for more efficient processing (see config "Binaural_HRIRFilterLength_LoQ")
 * 
 * Config options: 
 * - "Binaural_HRIRFilterLength_LoQ": how long the HRIR filters should be for the low quality rendering
 *   (default: 256 samples)
 * - (real-time adjustable) "Dispersion_Early_Reflections": enable dispersion filter for early reflections (default: true)
 * - (real-time adjustable) "Dispersion_Diffraction": enable dispersion filter for diffraction items (default: false)
 * - (real-time adjustable) "Dispersion_Filter_Length" length of dispersion filter (default: .015 seconds)
 * - (real-time adjustable) "Dispersion_Gain_dB" gain of dispersion filter (default: 0 decibel)
 *  
 */
class BinauralSpatializer : public Spatializer, public Binauralizer
{
public:
    //! Constructor
    BinauralSpatializer();

    //! Destructor
    ~BinauralSpatializer();

    //! Inherited from Binauralizer
    void loadHRIRFile(const std::string& path) override;

    //! Inherited from Spatializer
    std::unique_ptr<Spatializer::DspData> update(const RenderList& renderList) override;

    //! Inherited from Spatializer
    void requestBlock(DspContext ctx) override;

    //! Inherited from Spatializer
    std::size_t getNbrOfChannels() const override;

private:
    //! Inherited from Spatializer
    void initializeSpatializer(
        const Bitstream& bitstream, float fs, std::size_t B, const SceneState* state) override;

    //! Internal (re)configuration
    void initProcessor();
    void initProcessorForPanningHRIRconvolution();

    std::unique_ptr<Spatializer::DspData> updateForConventionalHRIRconvolution(
        const RenderList& renderList);
    std::unique_ptr<Spatializer::DspData> updateForPanningHRIRconvolution(
        const RenderList& renderList);

    void requestBlockForConventionalHRIRconvolution(DspContext ctx);
    void requestBlockForPanningHRIRconvolution(DspContext ctx);

    void parameterCalc();

    bool updateRep(const std::string path);

    std::vector<float> repAzimuth{ 30, 90, 150, 210, 270, 330 };
    std::vector<float> repElevation{ 90, -90 };

    std::vector<EAR::Polar> originRulerRepArray{
        EAR::Polar(repAzimuth[0], 0, 1.F),
        EAR::Polar(repAzimuth[1], 0, 1.F),
        EAR::Polar(repAzimuth[2], 0, 1.F),
        EAR::Polar(repAzimuth[3], 0, 1.F),
        EAR::Polar(repAzimuth[4], 0, 1.F),
        EAR::Polar(repAzimuth[5], 0, 1.F),
        EAR::Polar(0, repElevation[0], 1.F),
        EAR::Polar(0, repElevation[1], 1.F),
    };

    enum PANNING_HRIR_CONVOLUTION_MODE
    {
        MODE_CONVENTIONAL,
        MODE_8CH_PANNING
    };
    PANNING_HRIR_CONVOLUTION_MODE runningMode = MODE_CONVENTIONAL;

    float fs;
    std::size_t B;

    std::shared_ptr<const MPEGI::DispersionFilterPayload> dispersionFilterPayload{ nullptr };

    std::atomic<bool> processorInitialized;
    std::unique_ptr<EAR::RealFrame> temp;
    std::unique_ptr<EAR::RealFrame> smooth;

    std::unique_ptr<EAR::RealFrame> currentIR;
    std::unique_ptr<EAR::IRProvider> hrirs;
    std::unique_ptr<EAR::RealBuffer> shortenedHRIRWindow;
    bool forceFilterUpdate, forceReinitialization;

    std::string currentHRIRPath;

    //fast DSP loop members for ConventionalHRIRconvolution
    std::unique_ptr<EAR::ComplexBuffer> tempFD;
    std::unique_ptr<EAR::ComplexBuffer> tempFDSidechainOut;
    std::unique_ptr<EAR::ComplexBuffer> tempFDSidechainIn;
    std::unique_ptr<EAR::ComplexBuffer> outFD;
    std::unique_ptr<EAR::ComplexBuffer> outFDSidechainOut;
    std::unique_ptr<EAR::ComplexBuffer> outFDSidechainIn;
    std::unique_ptr<EAR::ComplexBuffer> smoothFD;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDSidechainOut;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDSidechainIn;

    //fast DSP loop members for PanningHRIRconvolution
    std::unique_ptr<EAR::ComplexBuffer> tempFDL;
    std::unique_ptr<EAR::ComplexBuffer> tempFDR;
    std::unique_ptr<EAR::ComplexBuffer> outFDL;
    std::unique_ptr<EAR::ComplexBuffer> outFDSidechainOutL;
    std::unique_ptr<EAR::ComplexBuffer> outFDSidechainInL;
    std::unique_ptr<EAR::ComplexBuffer> outFDR;
    std::unique_ptr<EAR::ComplexBuffer> outFDSidechainOutR;
    std::unique_ptr<EAR::ComplexBuffer> outFDSidechainInR;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDL;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDSidechainOutL;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDSidechainInL;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDR;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDSidechainOutR;
    std::unique_ptr<EAR::ComplexBuffer> smoothFDSidechainInR;

    //! complex to real IFFT processor
    std::unique_ptr<EAR::IFFT> ifft;
    //! fade envelopes (predefined on init)
    std::unique_ptr<EAR::RealBuffer> fadeIn, fadeOut;

    struct ItemState : RenderItem::ExtraState
    {
        ItemState(ItemType type, std::size_t blockLength, int filterLength);
        EAR::StereoFIRFilter filter;
    };

    std::vector<EAR::Vector3> prmPoints;
    std::unique_ptr<EAR::CartesianKDTree> lookupTree;
    std::vector<std::array<float, 15>> createSGtable;

    struct panningParms
    {
        std::array<float, NUM_REPRESENTATIVE> leftgainArray;
        std::array<float, NUM_REPRESENTATIVE> rightgainArray;
        std::array<float, NUM_REPRESENTATIVE> leftdelayArray;
        std::array<float, NUM_REPRESENTATIVE> rightdelayArray;

        void clear()
        {
#if defined(EAR_ENABLE_IPP)
            ippsZero_32f(leftgainArray.data(), NUM_REPRESENTATIVE);
            ippsZero_32f(rightgainArray.data(), NUM_REPRESENTATIVE);
            ippsZero_32f(leftdelayArray.data(), NUM_REPRESENTATIVE);
            ippsZero_32f(rightdelayArray.data(), NUM_REPRESENTATIVE);
#else
            std::fill(leftgainArray.begin(), leftgainArray.end(), 0.f);
            std::fill(rightgainArray.begin(), rightgainArray.end(), 0.f);
            std::fill(leftdelayArray.begin(), leftdelayArray.end(), 0.f);
            std::fill(rightdelayArray.begin(), rightdelayArray.end(), 0.f);
#endif
        }
    };

    std::unordered_map<int, std::unique_ptr<panningParms>> panParms;
    std::unordered_map<int, std::unique_ptr<panningParms>> prePanParms;

    std::array<std::array<std::unique_ptr<EAR::StereoFIRFilter>, NUM_REPRESENTATIVE>, 3>
        hrirFilterArrayL;
    std::array<std::array<std::unique_ptr<EAR::StereoFIRFilter>, NUM_REPRESENTATIVE>, 3>
        hrirFilterArrayR;

    std::unordered_map<int, std::unique_ptr<EAR::RealBuffer>> delayBuffs;
    std::unordered_map<int, std::unique_ptr<long>> representativePoints;
    std::unordered_map<int, std::unique_ptr<bool>> changeFlags;

    std::unique_ptr<EAR::RealFrame> buff;
    std::array<std::array<std::unique_ptr<EAR::RealFrame>, NUM_REPRESENTATIVE>, 3> outArray;
    std::array<std::array<std::unique_ptr<EAR::RealFrame>, NUM_REPRESENTATIVE>, 3> temp_Array;
    std::array<std::array<std::unique_ptr<EAR::RealFrame>, NUM_REPRESENTATIVE>, 3> preTemp_Array;

    struct DSPItem
    {
        EAR::StereoFIRFilter* hrirFilter;
        EAR::RealBuffer* signal;
        ItemType type;

        EAR::RealBuffer* delayBuff;
        bool* prmChg;
        panningParms* panParm;
        panningParms* prePanparm;
        DSPItem(EAR::StereoFIRFilter* hf, EAR::RealBuffer* sig, ItemType t, bool* _prmc,
            panningParms* _prm, panningParms* _prePrm, EAR::RealBuffer* _db)
            : hrirFilter(hf)
            , signal(sig)
            , type(t)
            , prmChg(_prmc)
            , panParm(_prm)
            , prePanparm(_prePrm)
            , delayBuff(_db)
        {}
    };

    struct DspData : Spatializer::DspData
    {
        std::vector<DSPItem> items;
        std::vector<EAR::RealFrame*> downmixItems;
    };

    std::unique_ptr<EAR::FIRFilter> m_smootherL;
    std::unique_ptr<EAR::FIRFilter> m_smootherR;
    float m_smoothLength{ .015f };
    float m_smoothGain{ 0.0f };
    bool m_enableSpatialDispersion{ true };
    void configureDispersionFilter();
    void frameClear();

    void firFilter(const float* input, float* taps, int tapsLen, float* output);
    void CrossCorrNorm(float* src1, float* src2, float* output);
    void dotProduct(const float* pSrc1, const float* pSrc2, int len, float* pDp);

#if !defined(EAR_ENABLE_IPP)
    inline void mulAdd(float* src, float* dst, float c, int length)
    {
        for (int n = 0; n < length; n++) {
            *dst++ += *src++ * c;
        }
    }
    inline void mulAdd(float* src1, float* c1, float* src2, float* c2, float* dst, int length)
    {
        for (int n = 0; n < length; n++) {
            *dst++ = *src1++ * (*c1++) + *src2++ * (*c2++);
        }
    }
#endif

    //! Accessibility: Spectral compensation configuration for left ear
    EAR::HeadphoneEQData accessibilityEqLeft;

    //! Accessibility: Spectral compensation configuration for right ear
    EAR::HeadphoneEQData accessibilityEqRight;

    //! Local Headphone EQ data from the config file.
    EAR::HeadphoneEQData localHpEq;

    //! Content-creator Headphone EQ data from the bitstream.
    EAR::HeadphoneEQData ccHpEq;

    //! Local override of content-creator specified target Headphone EQ data.
    EAR::HeadphoneEQData localCcHpEq;

    //! Method to load headphone EQ data from the config filel.
    EAR::HeadphoneEQData getHeadphoneEQ(std::string prefix);

    bool headphoneEQOn = false;

    std::unique_ptr<EAR::FIRFilter> headphoneEQL;
    std::unique_ptr<EAR::FIRFilter> headphoneEQR;
};

} // namespace MPEGI

#endif /* MPEGI_BINAURAL_SPATIALIZER */
