/***************************************************************************
This software module was originally developed jointly by

Panasonic Holdings Corporation and Akita Prefectual University

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Panasonic Holdings Corporation and Akita Prefectual University retains full
right to modify and use the code for its own purpose, assign or donate
the code to a third party without consent of and without accounting to third
parties, and to inhibit third parties from using the code for products that
do not conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Binaural.h"

#include <MPEGI/Renderer.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/SofaHRIRProvider.h>

#include <saf.h>
#include <saf_externals.h>

#include <EAR/FIRFilter.h>

namespace MPEGI {

bool BinauralSpatializer::updateRep(const std::string path)
{
    std::vector<float> nearestAzimuth(NUM_AZIMUTHS, NAN);
    float topElevation = NAN;
    float footElevation = NAN;

    SofaReader reader = SofaReader(path);
    const float* directions = reader.getDirectionsBuffer();
    const int AZIMUTH = 0;
    const int ELEVATION = 1;
    const int DISTANCE = 2;
    const int DIMENSION_SIZE = 3;

    unsigned long M = static_cast<unsigned long>(reader.getDimension("M"));

    // update paninngParameter Table
    createSGtable.resize(M);
    std::vector<float> absElevations(M);
    for (int mIndx = 0; mIndx < M; mIndx++) {
        createSGtable[mIndx][0] = directions[(DIMENSION_SIZE * mIndx) + AZIMUTH];
        createSGtable[mIndx][1] = directions[(DIMENSION_SIZE * mIndx) + ELEVATION];
        createSGtable[mIndx][2] = directions[(DIMENSION_SIZE * mIndx) + DISTANCE];
        absElevations[mIndx] = std::abs(createSGtable[mIndx][1]);
    }

    // get the elevation angle closest to 0 degrees
    float min_Elevation = *std::min_element(absElevations.begin(), absElevations.end());

    // get the Azimuth angle closest to 0 degrees
    std::vector<float> azimuths;
    for (int mIndx = 0; mIndx < M; mIndx++) {
        if (min_Elevation == createSGtable[mIndx][1]) {
            azimuths.push_back(std::abs(createSGtable[mIndx][0]));
        }
    }
    float min_Azimuth = *std::min_element(azimuths.begin(), azimuths.end());

    // correct representative points
    std::vector<int> azimuthIndex(NUM_AZIMUTHS, 0);
    for (int mIndx = 0; mIndx < M; mIndx++) {

        float elevation = createSGtable[mIndx][1];
        float azimuth = createSGtable[mIndx][0];

        if (azimuth == min_Azimuth) {
            if (topElevation < elevation || std::isnan(topElevation)) {
                topElevation = elevation;
            }

            if (footElevation > elevation || std::isnan(footElevation)) {
                footElevation = elevation;
            }
        }

        if (std::abs(elevation) == min_Elevation) {
            if (std::isnan(nearestAzimuth[0])) {
                nearestAzimuth[0] = azimuth;
                nearestAzimuth[1] = azimuth;
                nearestAzimuth[2] = azimuth;
                nearestAzimuth[3] = azimuth;
                nearestAzimuth[4] = azimuth;
                nearestAzimuth[5] = azimuth;
                continue;
            }

            for (int azimuthIdx = 0; azimuthIdx < NUM_AZIMUTHS; azimuthIdx++) {
                float target = repAzimuth[azimuthIdx];
                float diff = abs(target - azimuth);
                float nearestDiff = abs(target - nearestAzimuth[azimuthIdx]);

                if (diff < nearestDiff) {
                    nearestAzimuth[azimuthIdx] = azimuth;
                    azimuthIndex[azimuthIdx] = mIndx;

                } else if (diff == nearestDiff) {
                    if (target < 180) {
                        if (azimuth < nearestAzimuth[azimuthIdx]) {
                            nearestAzimuth[azimuthIdx] = azimuth;
                            azimuthIndex[azimuthIdx] = mIndx;
                        }
                    } else {
                        if (azimuth > nearestAzimuth[azimuthIdx]) {
                            nearestAzimuth[azimuthIdx] = azimuth;
                            azimuthIndex[azimuthIdx] = mIndx;
                        }
                    }
                }
            }
        }
    }

    // check
    decltype(azimuthIndex)::iterator result = std::unique(azimuthIndex.begin(), azimuthIndex.end());
    azimuthIndex.erase(result, azimuthIndex.end());
    bool bRtn = true;
    bool bAzimuthNan = false;
    for (int azimuthIdx = 0; azimuthIdx < NUM_AZIMUTHS; azimuthIdx++) {
        if (std::isnan(nearestAzimuth[azimuthIdx])) {
            bAzimuthNan = true;
            break;
        }
    }
    if (bAzimuthNan == true || std::isnan(topElevation) || std::isnan(footElevation)) {
        bRtn = false;
    } else if (azimuthIndex.size() != NUM_AZIMUTHS) {
        bRtn = false;
    } else if (topElevation <= 0) {
        bRtn = false;
    } else if (footElevation >= 0) {
        bRtn = false;
    }

    if (bRtn == true) {
        // update Azimuth and Elevation
        int idx;
        for (idx = 0; idx < NUM_AZIMUTHS; idx++) {
            repAzimuth[idx] = nearestAzimuth[idx];
        }
        repElevation[0] = topElevation;
        repElevation[1] = footElevation;

        // update Representative location
        for (idx = 0; idx < NUM_AZIMUTHS; idx++) {
            originRulerRepArray[idx] = EAR::Polar(repAzimuth[idx], 0, 1.F);
        }
        originRulerRepArray[idx++] = EAR::Polar(0, repElevation[0], 1.F);
        originRulerRepArray[idx++] = EAR::Polar(0, repElevation[1], 1.F);
    }

    return bRtn;
}

std::unique_ptr<Spatializer::DspData> BinauralSpatializer::updateForPanningHRIRconvolution(
    const RenderList& renderList)
{
    auto dspData = std::make_unique<BinauralSpatializer::DspData>();
    auto& dspList = *dspData.get();

    for (auto& item : renderList.items) {
        if (item->type == ItemType::Downmix || item->type == ItemType::Airflow) {
            auto& frame = item->signal->getFrame();
            if (frame.channels() == getNbrOfChannels()) {
                dspList.downmixItems.emplace_back(&frame);
            }
            continue;
        }

        if (item->type != ItemType::Nondiegetic && item->type != ItemType::Primary
            && item->type != ItemType::Reflection && item->type != ItemType::Representative
            && item->type != ItemType::Diffraction) {
            // skip unknown render item types
            continue;
        }

        // TODO: remove this when the SESS fade out mechanism is completed
        if (item->extent) {
            // extended sources with one input channel ("homogeneous SESSs") and
            // extended sources with multiple input channels ("heterogeneous SESSs")
            // have already been processed in previous stages, so we can skip anything with extent here
            continue;
        }

        // create new resources if necessary
        if (delayBuffs.find(item->idx) == delayBuffs.end() || forceReinitialization) {
            delayBuffs[item->idx] = std::make_unique<EAR::RealBuffer>(B);
        }
        if (panParms.find(item->idx) == panParms.end() || forceReinitialization) {
            panParms[item->idx] = std::make_unique<panningParms>();
        }
        if (representativePoints.find(item->idx) == representativePoints.end()
            || forceReinitialization) {
            representativePoints[item->idx] = std::make_unique<long>();
        }
        if (changeFlags.find(item->idx) == changeFlags.end() || forceReinitialization) {
            changeFlags[item->idx] = std::make_unique<bool>();
            changeFlags[item->idx].get()[0] = false;
        }
        if (prePanParms.find(item->idx) == prePanParms.end() || forceReinitialization) {
            prePanParms[item->idx] = std::make_unique<panningParms>();

            for (int i = 0; i < NUM_REPRESENTATIVE; i++) {
                panParms[item->idx]->leftdelayArray[i] = 0;
                panParms[item->idx]->rightdelayArray[i] = 0;
            }
        }
        if (forceReinitialization) {
            item->unsafeRemoveState<BinauralSpatializer::ItemState>();
        }
        auto& hrirFilter =
            item->getStateOrCreate<BinauralSpatializer::ItemState>(item->type, B, hrirs->irLength())
                .filter;

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        if (item->status == ItemStatus::Inactive) {
            // reset filters
            hrirFilter.flush();
            continue;
        }

        // update panning parameters
        if ((item->changed & (ItemProperty::Position | ItemProperty::Status))
            || (item->position.cSpace != EAR::CSpace::LCS && renderList.listener.hasMoved)
            || forceFilterUpdate) {
            // replace filters if required

            EAR::Vector3 relLoc;
            if (item->position.cSpace == EAR::CSpace::LCS && item->type == ItemType::Nondiegetic) {
                // TODO: this is a hack for reverb loudspeakers - how would it be correct?
                relLoc = item->position.location;
            } else {
                relLoc = renderList.listener.cs.transform(item->position.location);
            }

            long prmindex = lookupTree->getNearestIndex(relLoc);

            if (representativePoints[item->idx].get()[0] != prmindex) {
                representativePoints[item->idx].get()[0] = prmindex;

                panParms[item->idx]->clear();

                int idx1 = (NUM_AZIMUTHS - 1);
                float* azimuthTable = repAzimuth.data();
                for (int i = 0; i < (NUM_AZIMUTHS - 1); i++) {
                    if (createSGtable[prmindex][0] > azimuthTable[i]
                        && createSGtable[prmindex][0] <= azimuthTable[i + 1]) {
                        idx1 = i;
                        break;
                    }
                }
                int idx2;
                if (idx1 == (NUM_AZIMUTHS - 1)) {
                    idx2 = 0;
                } else {
                    idx2 = idx1 + 1;
                }
                panParms[item->idx]->leftgainArray[idx1] = createSGtable[prmindex][3];
                panParms[item->idx]->leftgainArray[idx2] = createSGtable[prmindex][4];
                panParms[item->idx]->rightgainArray[idx1] = createSGtable[prmindex][6];
                panParms[item->idx]->rightgainArray[idx2] = createSGtable[prmindex][7];
                panParms[item->idx]->leftdelayArray[idx1] = createSGtable[prmindex][9];
                panParms[item->idx]->leftdelayArray[idx2] = createSGtable[prmindex][10];
                panParms[item->idx]->rightdelayArray[idx1] = createSGtable[prmindex][12];
                panParms[item->idx]->rightdelayArray[idx2] = createSGtable[prmindex][13];

                int idx;
                if (createSGtable[prmindex][1] >= 0) {
                    idx = NUM_AZIMUTHS;
                } else {
                    idx = NUM_AZIMUTHS + 1;
                }
                panParms[item->idx]->leftgainArray[idx] = createSGtable[prmindex][5];
                panParms[item->idx]->rightgainArray[idx] = createSGtable[prmindex][8];
                panParms[item->idx]->leftdelayArray[idx] = createSGtable[prmindex][11];
                panParms[item->idx]->rightdelayArray[idx] = createSGtable[prmindex][14];
            }
            changeFlags[item->idx].get()[0] = true;
        }

        dspList.items.emplace_back(&hrirFilter,
            &item->signal->getFrame()[0], // TODO: handle multichannel?
            item->type, changeFlags[item->idx].get(), panParms[item->idx].get(),
            prePanParms[item->idx].get(), delayBuffs[item->idx].get());
    }

    // when we arrive here, all paths are updated if necessary
    forceFilterUpdate = false;
    forceReinitialization = false;

    return dspData;
}

void BinauralSpatializer::requestBlockForPanningHRIRconvolution(DspContext ctx)
{
    auto& output = ctx.output;

    // clear output buffer (also when processing is not initialized)
    output.zero(); // TODO: make this optional for added flexibility?
    if (!processorInitialized.load()) {
        return;
    }

    auto& currentDSP = ctx.getData<BinauralSpatializer::DspData>();

    auto outL = &(output.ch(0));
    auto outR = &(output.ch(1));
    auto smL = &(smooth->ch(0));
    auto smR = &(smooth->ch(1));

    bool refOn = Config::getBool("Dispersion_Early_Reflections");
    bool difOn = Config::getBool("Dispersion_Diffraction");

    if (refOn || difOn)
        smooth->zero();

    //faster DSP loop
    outFDL->zero();
    outFDSidechainOutL->zero();
    outFDSidechainInL->zero();
    outFDR->zero();
    outFDSidechainOutR->zero();
    outFDSidechainInR->zero();

    if (refOn || difOn) {
        smoothFDL->zero();
        smoothFDSidechainOutL->zero();
        smoothFDSidechainInL->zero();
        smoothFDR->zero();
        smoothFDSidechainOutR->zero();
        smoothFDSidechainInR->zero();
    }

    frameClear();

    //flag for itemType
    bool isRef = false;
    bool isDif = false;
    bool isOut = false;

    //flag for process representative points
    std::vector<std::vector<std::vector<bool>>> is_Array(
        2, std::vector<std::vector<bool>>(NUM_REPRESENTATIVE, std::vector<bool>(3, false)));

    // perform time shift, multiply by gain value,
    // and add to buffer at each representative position
    // (with crossfade)
    for (auto& item : currentDSP.items) {
        if (item.prmChg[0]) {

            int typeIdx;
            if (item.type == ItemType::Reflection && refOn) {
                typeIdx = 0;
                isRef = true;
            } else if (item.type == ItemType::Diffraction && difOn) {
                typeIdx = 1;
                isDif = true;
            } else {
                typeIdx = 2;
                isOut = true;
            }

            float* pbuff = buff->ch(0).data();
            int delay;
            float gain;
            for (int repIdx = 0; repIdx < NUM_REPRESENTATIVE; repIdx++) {
                for (int ch = 0; ch < 2; ch++) {
                    if (ch == 0) {
                        delay = (int)item.panParm->leftdelayArray[repIdx];
                        gain = item.panParm->leftgainArray[repIdx];
                    } else {
                        delay = (int)item.panParm->rightdelayArray[repIdx];
                        gain = item.panParm->rightgainArray[repIdx];
                    }
                    if (gain != 0.0f) {
#if defined(EAR_ENABLE_IPP)
                        ippsMulC_32f((const Ipp32f*)item.signal->data(), gain,
                            (Ipp32f*)pbuff + delay, B - delay);
                        ippsMulC_32f((const Ipp32f*)item.delayBuff->data() + B - delay, gain,
                            (Ipp32f*)pbuff, delay);
                        ippsAdd_32f_I((const Ipp32f*)pbuff,
                            (Ipp32f*)temp_Array[typeIdx][repIdx]->ch(ch).data(), B);
#else
                        buff->ch(0).copyFrom(*item.delayBuff, B - delay, 0, delay);
                        buff->ch(0).copyFrom(*item.signal, 0, delay, B - delay);
                        mulAdd(buff->ch(0).data(), temp_Array[typeIdx][repIdx]->ch(ch).data(), gain,
                            B);
#endif
                        is_Array[ch][repIdx][typeIdx] = true;
                    }

                    if (ch == 0) {
                        delay = (int)item.prePanparm->leftdelayArray[repIdx];
                        gain = item.prePanparm->leftgainArray[repIdx];
                    } else {
                        delay = (int)item.prePanparm->rightdelayArray[repIdx];
                        gain = item.prePanparm->rightgainArray[repIdx];
                    }
                    if (gain != 0.0f) {
#if defined(EAR_ENABLE_IPP)
                        ippsMulC_32f((const Ipp32f*)item.signal->data(), gain,
                            (Ipp32f*)pbuff + delay, B - delay);
                        ippsMulC_32f((const Ipp32f*)item.delayBuff->data() + B - delay, gain,
                            (Ipp32f*)pbuff, delay);
                        ippsAdd_32f_I((const Ipp32f*)pbuff,
                            (Ipp32f*)preTemp_Array[typeIdx][repIdx]->ch(ch).data(), B);
#else
                        buff->ch(0).copyFrom(*item.delayBuff, B - delay, 0, delay);
                        buff->ch(0).copyFrom(*item.signal, 0, delay, B - delay);
                        mulAdd(buff->ch(0).data(), preTemp_Array[typeIdx][repIdx]->ch(ch).data(),
                            gain, B);
#endif
                        is_Array[ch][repIdx][typeIdx] = true;
                    }
                }
            }

            for (int i = 0; i < NUM_REPRESENTATIVE; i++) {
                item.prePanparm->leftgainArray[i] = item.panParm->leftgainArray[i];
                item.prePanparm->rightgainArray[i] = item.panParm->rightgainArray[i];
                item.prePanparm->leftdelayArray[i] = item.panParm->leftdelayArray[i];
                item.prePanparm->rightdelayArray[i] = item.panParm->rightdelayArray[i];
            }

        } // prmChg
    } // dsp item loop

    // crossfade
    for (int typeIdx = 0; typeIdx < 3; typeIdx++) {
        if (typeIdx == 0 && !isRef) {
            continue;
        } else if (typeIdx == 1 && !isDif) {
            continue;
        } else if (typeIdx == 2 && !isOut) {
            continue;
        }

        for (int repIdx = 0; repIdx < NUM_REPRESENTATIVE; repIdx++) {
            for (int ch = 0; ch < 2; ch++) {
                if (is_Array[ch][repIdx][typeIdx]) {
#if defined(EAR_ENABLE_IPP)
                    ippsMul_32f_I((const Ipp32f*)fadeIn->data(),
                        (Ipp32f*)temp_Array[typeIdx][repIdx]->ch(ch).data(), B);
                    ippsMul_32f_I((const Ipp32f*)fadeOut->data(),
                        (Ipp32f*)preTemp_Array[typeIdx][repIdx]->ch(ch).data(), B);
                    ippsAdd_32f((const Ipp32f*)temp_Array[typeIdx][repIdx]->ch(ch).data(),
                        (const Ipp32f*)preTemp_Array[typeIdx][repIdx]->ch(ch).data(),
                        (Ipp32f*)outArray[typeIdx][repIdx]->ch(ch).data(), B);
#else
                    mulAdd(temp_Array[typeIdx][repIdx]->ch(ch).data(), fadeIn->data(),
                        preTemp_Array[typeIdx][repIdx]->ch(ch).data(), fadeOut->data(),
                        outArray[typeIdx][repIdx]->ch(ch).data(), B);
#endif
                }
            }
        }
    }

    // perform time shift, multiply by gain value,
    // and add to buffer at each representative position
    // (without crossfade)
    for (auto& item : currentDSP.items) {
        if (!item.prmChg[0]) {

            int typeIdx;
            if (item.type == ItemType::Reflection && refOn) {
                typeIdx = 0;
                isRef = true;
            } else if (item.type == ItemType::Diffraction && difOn) {
                typeIdx = 1;
                isDif = true;
            } else {
                typeIdx = 2;
                isOut = true;
            }

            float* pbuff = buff->ch(0).data();
            int delay;
            float gain;
            for (int repIdx = 0; repIdx < NUM_REPRESENTATIVE; repIdx++) {
                for (int ch = 0; ch < 2; ch++) {
                    if (ch == 0) {
                        delay = (int)item.prePanparm->leftdelayArray[repIdx];
                        gain = item.prePanparm->leftgainArray[repIdx];
                    } else {
                        delay = (int)item.prePanparm->rightdelayArray[repIdx];
                        gain = item.prePanparm->rightgainArray[repIdx];
                    }

                    if (gain != 0.0f) {
#if defined(EAR_ENABLE_IPP)
                        ippsMulC_32f((const Ipp32f*)item.signal->data(), gain,
                            (Ipp32f*)pbuff + delay, B - delay);
                        ippsMulC_32f((const Ipp32f*)item.delayBuff->data() + B - delay, gain,
                            (Ipp32f*)pbuff, delay);
                        ippsAdd_32f_I((const Ipp32f*)pbuff,
                            (Ipp32f*)outArray[typeIdx][repIdx]->ch(ch).data(), B);
#else
                        buff->ch(0).copyFrom(*item.delayBuff, B - delay, 0, delay);
                        buff->ch(0).copyFrom(*item.signal, 0, delay, B - delay);
                        mulAdd(
                            buff->ch(0).data(), outArray[typeIdx][repIdx]->ch(ch).data(), gain, B);
#endif
                        is_Array[ch][repIdx][typeIdx] = true;
                    }
                }
            }
        } // prmChg
        item.prmChg[0] = false;
#if defined(EAR_ENABLE_IPP)
        ippsCopy_32f((const Ipp32f*)item.signal->data(), (Ipp32f*)item.delayBuff->data(), B);
#else
        std::copy(item.signal->data(), item.signal->data() + B, item.delayBuff->data());
#endif
    } // dsp item loop

    // convolve HRIR at each representative position
    if (refOn && isRef) {
        for (int repIdx = 0; repIdx < NUM_REPRESENTATIVE; repIdx++) {
            if (is_Array[0][repIdx][0]) {
                hrirFilterArrayL[0][repIdx]->processFD(outArray[0][repIdx]->ch(0).data(),
                    smoothFDL->data(), smoothFDSidechainOutL->data(), smoothFDSidechainInL->data());
            }
            if (is_Array[1][repIdx][0]) {
                hrirFilterArrayR[0][repIdx]->processFD(outArray[0][repIdx]->ch(1).data(),
                    smoothFDR->data(), smoothFDSidechainOutR->data(), smoothFDSidechainInR->data());
            }
        }
    }
    if (difOn && isDif) {
        for (int repIdx = 0; repIdx < NUM_REPRESENTATIVE; repIdx++) {
            if (is_Array[0][repIdx][1]) {
                hrirFilterArrayL[1][repIdx]->processFD(outArray[1][repIdx]->ch(0).data(),
                    smoothFDL->data(), smoothFDSidechainOutL->data(), smoothFDSidechainInL->data());
            }
            if (is_Array[1][repIdx][1]) {
                hrirFilterArrayR[1][repIdx]->processFD(outArray[1][repIdx]->ch(1).data(),
                    smoothFDR->data(), smoothFDSidechainOutR->data(), smoothFDSidechainInR->data());
            }
        }
    }
    if (isOut) {
        for (int repIdx = 0; repIdx < NUM_REPRESENTATIVE; repIdx++) {
            if (is_Array[0][repIdx][2]) {
                hrirFilterArrayL[2][repIdx]->processFD(outArray[2][repIdx]->ch(0).data(),
                    outFDL->data(), outFDSidechainOutL->data(), outFDSidechainInL->data());
            }
            if (is_Array[1][repIdx][2]) {
                hrirFilterArrayR[2][repIdx]->processFD(outArray[2][repIdx]->ch(1).data(),
                    outFDR->data(), outFDSidechainOutR->data(), outFDSidechainInR->data());
            }
        }
    }

    // output
    std::complex<float>* const tempPtrL = tempFDL->data();
    std::complex<float>* const tempPtrR = tempFDR->data();
    ifft->transform(outFDL->data(), tempFDL->data());
    ifft->transform(outFDR->data(), tempFDR->data());
    float* const outLPtr = outL->data();
    float* const outRPtr = outR->data();
    for (int s = 0; s < B; s++) {
        outLPtr[s] += tempPtrL[B + s].real();
        outRPtr[s] += tempPtrR[B + s].imag();
    }
    ifft->transform(smoothFDL->data(), tempFDL->data());
    ifft->transform(smoothFDR->data(), tempFDR->data());
    float* const smoothLPtr = smL->data();
    float* const smoothRPtr = smR->data();
    for (int s = 0; s < B; s++) {
        smoothLPtr[s] += tempPtrL[B + s].real();
        smoothRPtr[s] += tempPtrR[B + s].imag();
    }

    if (m_smootherL && (refOn || difOn)) {
        // check for smoothing filter changes
        auto sm = Config::getFloat("Dispersion_Filter_Length");
        auto gn = Config::getFloat("Dispersion_Gain_dB");
        if (m_smoothLength != sm || m_smoothGain != gn) {
            m_smoothLength = sm;
            m_smoothGain = gn;
            configureDispersionFilter();
        }

        // convolve smooth buffer
        m_smootherL->process(smL->data(), smL->data());
        m_smootherR->process(smR->data(), smR->data());
        // add smooth buffer to output buffer
        outL->add(*smL);
        outR->add(*smR);
    }

    for (auto sig : currentDSP.downmixItems) {
        outL->add(sig->ch(0));
        outR->add(sig->ch(1));
    } // dsp downmix item loop

    if (headphoneEQOn) {
        // convolve headphoneEQ filter
        headphoneEQL->process(outL->data(), outL->data());
        headphoneEQR->process(outR->data(), outR->data());
    }

    // TODO: Limiter?
}

void BinauralSpatializer::initProcessorForPanningHRIRconvolution()
{
    const size_t K = 2 * B;
    buff = std::make_unique<EAR::RealFrame>(1, B);
    int smFiltL = Config::getInt("Binaural_HRIRFilterLength_LoQ");
    int filterLength;
    for (int i = 0; i < 3; i++) {
        if (i == 0) {
            filterLength = smFiltL;
        } else {
            filterLength = hrirs->irLength();
        }
        for (int j = 0; j < NUM_REPRESENTATIVE; j++) {
            outArray[i][j] = std::make_unique<EAR::RealFrame>(2, B);
            temp_Array[i][j] = std::make_unique<EAR::RealFrame>(2, B);
            preTemp_Array[i][j] = std::make_unique<EAR::RealFrame>(2, B);
            hrirFilterArrayL[i][j] = std::make_unique<EAR::StereoFIRFilter>(B, filterLength);
            hrirFilterArrayR[i][j] = std::make_unique<EAR::StereoFIRFilter>(B, filterLength);
        }
    }
    tempFDL = std::make_unique<EAR::ComplexBuffer>(K);
    outFDL = std::make_unique<EAR::ComplexBuffer>(K);
    outFDSidechainOutL = std::make_unique<EAR::ComplexBuffer>(K);
    outFDSidechainInL = std::make_unique<EAR::ComplexBuffer>(K);
    smoothFDL = std::make_unique<EAR::ComplexBuffer>(K);
    smoothFDSidechainOutL = std::make_unique<EAR::ComplexBuffer>(K);
    smoothFDSidechainInL = std::make_unique<EAR::ComplexBuffer>(K);
    tempFDR = std::make_unique<EAR::ComplexBuffer>(K);
    outFDR = std::make_unique<EAR::ComplexBuffer>(K);
    outFDSidechainOutR = std::make_unique<EAR::ComplexBuffer>(K);
    outFDSidechainInR = std::make_unique<EAR::ComplexBuffer>(K);
    smoothFDR = std::make_unique<EAR::ComplexBuffer>(K);
    smoothFDSidechainOutR = std::make_unique<EAR::ComplexBuffer>(K);
    smoothFDSidechainInR = std::make_unique<EAR::ComplexBuffer>(K);
    // get the HRIR of each representative point(for out and diffraction)
    EAR::Vector3 vecArray[NUM_REPRESENTATIVE];
    for (int i = 0; i < NUM_REPRESENTATIVE; i++) {
        vecArray[i] = polarToCartesian(originRulerRepArray[i]);
        hrirs->getNearestIRForLocation(*currentIR, vecArray[i]);
        hrirFilterArrayL[1][i]->setFilter(currentIR->data().get());
        hrirFilterArrayR[1][i]->setFilter(currentIR->data().get());
        hrirFilterArrayL[2][i]->setFilter(currentIR->data().get());
        hrirFilterArrayR[2][i]->setFilter(currentIR->data().get());
    }

    // get the HRIR of each representative point(for reflection)
    for (int i = 0; i < NUM_REPRESENTATIVE; i++) {
        hrirs->getNearestIRForLocation(*currentIR, vecArray[i]);
        for (int ch = 0; ch < currentIR->channels(); ch++) {
            currentIR->ch(ch).mul(
                *shortenedHRIRWindow, 0, 0, Config::getInt("Binaural_HRIRFilterLength_LoQ"));
        }
        hrirFilterArrayL[0][i]->setFilter(currentIR->data().get());
        hrirFilterArrayR[0][i]->setFilter(currentIR->data().get());
    }

    panParms.clear();
    prePanParms.clear();
}

void BinauralSpatializer::parameterCalc()
{
    const size_t K = 2 * B;
    int M = hrirs->irCount();

    prmPoints.resize(M);

    std::vector<std::vector<std::vector<float>>> panningGain(
        2, std::vector<std::vector<float>>(3, std::vector<float>(M)));

    std::vector<std::vector<std::vector<float>>> panningDelay(
        2, std::vector<std::vector<float>>(3, std::vector<float>(M)));

    std::vector<int> rindex(M);

    //LPF settings
    std::array<float, 13> LPFCoeff = { 0.031509543, -0.047208231, -0.050431148, 0.005297148,
        0.130896074, 0.266767171, 0.325763239, 0.266767171, 0.130896074, 0.005297148, -0.050431148,
        -0.047208231, 0.031509543 };
    int LPFCoeffLen = LPFCoeff.size();

    for (int mIndx = 0; mIndx < M; mIndx++) {
        // get location for LookupTree
        prmPoints[mIndx] = hrirs->getLocationForIndex(mIndx);

        // Calculate which representative point is included in the triangular element according to
        // the elevation and azimuth angles
        int thint = (int)createSGtable[mIndx][0];
        int phint = (int)createSGtable[mIndx][1];
        if (phint >= 0) {
            for (int azi = 0; azi < NUM_AZIMUTHS; azi++) {
                if (azi == NUM_AZIMUTHS - 1) {
                    rindex[mIndx] = azi;
                    break;
                } else if (thint > repAzimuth[azi] && thint <= repAzimuth[azi + 1]) {
                    rindex[mIndx] = azi;
                    break;
                }
            }
        } else {
            for (int azi = 0; azi < NUM_AZIMUTHS; azi++) {
                if (azi == NUM_AZIMUTHS - 1) {
                    rindex[mIndx] = NUM_AZIMUTHS + azi;
                    break;
                } else if (thint > repAzimuth[azi] && thint <= repAzimuth[azi + 1]) {
                    rindex[mIndx] = NUM_AZIMUTHS + azi;
                    break;
                }
            }
        }
    }
    //create LookupTree
    lookupTree = std::make_unique<CartesianKDTree>(prmPoints);

    //set representative point
    std::vector<long> repPosIndexAzi;
    std::vector<long> repPosIndexEle;
    EAR::Vector3 location;
    for (auto& azi : repAzimuth) {
        location = EAR::polarToCartesian({ azi, 0, 1.F });
        repPosIndexAzi.push_back(lookupTree->getNearestIndex(location));
    }
    for (auto& ele : repElevation) {
        location = EAR::polarToCartesian({ 0, ele, 1.F });
        repPosIndexEle.push_back(lookupTree->getNearestIndex(location));
    }

    std::vector<std::vector<long>> repPosTriangle;
    repPosTriangle.resize(NUM_AZIMUTHS * NUM_ELEVATIONS, std::vector<long>(3));
    int count = 0;
    for (int ele = 0; ele < NUM_ELEVATIONS; ele++) {
        for (int azi = 0; azi < NUM_AZIMUTHS; azi++) {
            repPosTriangle[count][0] = repPosIndexAzi[azi];
            repPosTriangle[count][2] = repPosIndexEle[ele];
            if (azi == NUM_AZIMUTHS - 1) {
                repPosTriangle[count][1] = repPosIndexAzi[0];
            } else {
                repPosTriangle[count][1] = repPosIndexAzi[azi + 1];
            }
            count++;
        }
    }

    // create filterKeyTable
    std::unordered_map<int, int> filterKeyTable;
    for (auto triangle : repPosTriangle) {
        for (auto triangleIndex : triangle) {
            // search
            int filterKey = -1;
            for (int i = 0; i < repPosIndexAzi.size(); i++) { // search from Azimuth
                if (repPosIndexAzi[i] == triangleIndex) {
                    filterKey = i;
                }
            }
            for (int i = 0; i < repPosIndexEle.size(); i++) { // search from Elevation
                if (repPosIndexEle[i] == triangleIndex) {
                    filterKey = NUM_AZIMUTHS + i;
                }
            }
            // set
            filterKeyTable[triangleIndex] = filterKey;
        }
    }

    std::vector<std::vector<std::vector<float>>> ir(
        2, std::vector<std::vector<float>>(3, std::vector<float>(B)));

    std::vector<std::vector<float>> ippOrgHRIRfilter(2, std::vector<float>(B));

    std::vector<std::vector<float>> ippxcorr(2, std::vector<float>((size_t)(K - 1)));

    // create filter
    std::vector<std::vector<std::vector<float>>> filteredRepHRIRfilter(
        2, std::vector<std::vector<float>>(NUM_REPRESENTATIVE, std::vector<float>(B)));

    float minDelay = 0;

    for (int ch = 0; ch < 2; ch++) {
        for (int i = 0; i < NUM_AZIMUTHS; i++) {
            firFilter(hrirs.get()->getIRPointerForIndex(repPosIndexAzi[i])[ch], LPFCoeff.data(),
                LPFCoeffLen, filteredRepHRIRfilter[ch][i].data());
        }
        for (int i = 0; i < NUM_ELEVATIONS; i++) {
            firFilter(hrirs.get()->getIRPointerForIndex(repPosIndexEle[i])[ch], LPFCoeff.data(),
                LPFCoeffLen, filteredRepHRIRfilter[ch][NUM_AZIMUTHS + i].data());
        }

        for (int mIndx = 0; mIndx < M; mIndx++) {
            int rindx = rindex[mIndx];

            //calculate delay
#if defined(EAR_ENABLE_IPP)
            ippsZero_32f(ippOrgHRIRfilter[ch].data(), B);
#else
            std::fill(ippOrgHRIRfilter[ch].begin(), ippOrgHRIRfilter[ch].end(), 0.f);
#endif
            firFilter(hrirs.get()->getIRPointerForIndex(mIndx)[ch], LPFCoeff.data(), LPFCoeffLen,
                ippOrgHRIRfilter[ch].data());

            for (int dindx = 0; dindx < 3; dindx++) {

                int filterKey = filterKeyTable[repPosTriangle[rindx][dindx]];

                //find the maximum cross-correlations
                CrossCorrNorm(ippOrgHRIRfilter[ch].data(),
                    filteredRepHRIRfilter[ch][filterKey].data(), ippxcorr[ch].data());

                float max = *std::max_element(ippxcorr[ch].begin(), ippxcorr[ch].end());

#if defined(EAR_ENABLE_IPP)
                ippsDivC_32f_I(max, ippxcorr[ch].data(), (int)(K - 1));
#else
                for (int n = 0; n < (int)(K - 1); n++) {
                    ippxcorr[ch][n] = ippxcorr[ch][n] / max;
                }
#endif

                int maxIdx = 0;
                float maxVal = 0.0f;

                for (int j = 0; j < (int)(K - 1); j++) {
                    if (ippxcorr[ch][j] > maxVal) {
                        maxIdx = j;
                        maxVal = ippxcorr[ch][j];
                    }
                }
                panningDelay[ch][dindx][mIndx] = (int)(B - 1) - maxIdx;

                if (minDelay > panningDelay[ch][dindx][mIndx]) {
                    minDelay = panningDelay[ch][dindx][mIndx];
                }

                //timeshift HRIR
                int delay = std::abs(panningDelay[ch][dindx][mIndx]);
                if (panningDelay[ch][dindx][mIndx] > 0) {
#if defined(EAR_ENABLE_IPP)
                    ippsZero_32f(ir[ch][dindx].data(), B);
                    ippsCopy_32f(filteredRepHRIRfilter[ch][filterKey].data(),
                        ir[ch][dindx].data() + delay, B - delay);
#else
                    std::fill(ir[ch][dindx].begin(), ir[ch][dindx].end(), 0.f);
                    std::copy(filteredRepHRIRfilter[ch][filterKey].data(),
                        filteredRepHRIRfilter[ch][filterKey].data() + (B - delay),
                        ir[ch][dindx].data() + delay);

#endif

                } else if (panningDelay[ch][dindx][mIndx] < 0) {
#if defined(EAR_ENABLE_IPP)
                    ippsZero_32f(ir[ch][dindx].data(), B);
                    ippsCopy_32f(filteredRepHRIRfilter[ch][filterKey].data() + delay,
                        ir[ch][dindx].data(), B - delay);
#else
                    std::fill(ir[ch][dindx].begin(), ir[ch][dindx].end(), 0.f);
                    std::copy(filteredRepHRIRfilter[ch][filterKey].data() + delay,
                        filteredRepHRIRfilter[ch][filterKey].data() + B, ir[ch][dindx].data());
#endif

                } else {
#if defined(EAR_ENABLE_IPP)
                    ippsCopy_32f(
                        filteredRepHRIRfilter[ch][filterKey].data(), ir[ch][dindx].data(), B);
#else
                    std::copy(filteredRepHRIRfilter[ch][filterKey].data(),
                        filteredRepHRIRfilter[ch][filterKey].data() + B, ir[ch][dindx].data());
#endif
                }
            }

            //calculate gain
            float sum_y11, sum_y12, sum_y13, sum_y22, sum_y23, sum_y33, sum_ori1, sum_ori2,
                sum_ori3;
            dotProduct(ir[ch][0].data(), ir[ch][0].data(), B, &sum_y11);
            dotProduct(ir[ch][0].data(), ir[ch][1].data(), B, &sum_y12);
            dotProduct(ir[ch][0].data(), ir[ch][2].data(), B, &sum_y13);
            dotProduct(ir[ch][1].data(), ir[ch][1].data(), B, &sum_y22);
            dotProduct(ir[ch][1].data(), ir[ch][2].data(), B, &sum_y23);
            dotProduct(ir[ch][2].data(), ir[ch][2].data(), B, &sum_y33);

            dotProduct(ippOrgHRIRfilter[ch].data(), ir[ch][0].data(), B, &sum_ori1);
            dotProduct(ippOrgHRIRfilter[ch].data(), ir[ch][1].data(), B, &sum_ori2);
            dotProduct(ippOrgHRIRfilter[ch].data(), ir[ch][2].data(), B, &sum_ori3);

            std::array<std::array<double, 3>, 3> mtrx = { { { sum_y11, sum_y12, sum_y13 },
                { sum_y12, sum_y22, sum_y23 }, { sum_y13, sum_y23, sum_y33 } } };
            std::array<std::array<float, 1>, 3> vecP = { { { sum_ori1 }, { sum_ori2 },
                { sum_ori3 } } };
            std::array<std::array<double, 3>, 3> invMtrx;

            utility_dinv(mtrx[0].data(), invMtrx[0].data(), 3);
            for (int i = 0; i < 3; i++) {
                for (int k = 0; k < 3; k++) {
                    panningGain[ch][i][mIndx] += invMtrx[i][k] * vecP[k][0];
                }
            }
        }
    }

    for (int mIndx = 0; mIndx < M; mIndx++) {
        for (int ch = 0; ch < 2; ch++) {
            createSGtable[mIndx][3 + ch * 3] = panningGain[ch][0][mIndx];
            createSGtable[mIndx][4 + ch * 3] = panningGain[ch][1][mIndx];
            createSGtable[mIndx][5 + ch * 3] = panningGain[ch][2][mIndx];

            createSGtable[mIndx][9 + ch * 3] = panningDelay[ch][0][mIndx] - minDelay;
            createSGtable[mIndx][10 + ch * 3] = panningDelay[ch][1][mIndx] - minDelay;
            createSGtable[mIndx][11 + ch * 3] = panningDelay[ch][2][mIndx] - minDelay;
        }
    }

    //set the delay amount at fixed times for elevation angles of 90 degrees and -90 degrees
    if (repElevation[1] == -90) {
        int index = repPosIndexEle[1];
        std::fill(&createSGtable[index][9], &createSGtable[index][14], -minDelay);
    }
    if (repElevation[0] == 90) {
        int index = repPosIndexEle[0];
        std::fill(&createSGtable[index][9], &createSGtable[index][14], -minDelay);
    }
}

void BinauralSpatializer::frameClear()
{
    bool difOn = Config::getBool("Dispersion_Diffraction");

    for (int i = 0; i < 3; i++) {
        if (i == 1) {
            if (difOn == false) {
                continue;
            }
        }
        for (int j = 0; j < NUM_REPRESENTATIVE; j++) {
            outArray[i][j]->zero();
            temp_Array[i][j]->zero();
            preTemp_Array[i][j]->zero();
        }
    }
}

void BinauralSpatializer::firFilter(const float* input, float* pTaps, int tapsLen, float* output)
{
#if defined(EAR_ENABLE_IPP)
    float* delayBuf = ippsMalloc_32f(tapsLen); // filter state
    ippsZero_32f(delayBuf, tapsLen); // delay buffer
    int inputPtr = 0; // state pointer

    float y = 0.0f;
    for (int n = 0; n < B; ++n) {
        delayBuf[inputPtr] = input[n]; // insert input into state
        y = std::inner_product(delayBuf + inputPtr, delayBuf + tapsLen, pTaps, 0.0f);
        y = std::inner_product(delayBuf, delayBuf + inputPtr, pTaps + (tapsLen - inputPtr), y);
        inputPtr = (inputPtr == 0 ? tapsLen - 1 : inputPtr - 1); // iterate state pointer in reverse
        output[n] = y;
    }
    ippsFree(delayBuf);
#else
    std::vector<float> delayBuf(tapsLen);
    std::fill(delayBuf.begin(), delayBuf.end(), 0.f);
    int inputPtr = 0; // state pointer

    float y = 0.0f;
    for (int n = 0; n < B; ++n) {
        delayBuf[inputPtr] = input[n]; // insert input into state
        auto iterator1 = std::next(delayBuf.begin(), inputPtr);
        y = std::inner_product(iterator1, delayBuf.end(), pTaps, 0.0f);
        y = std::inner_product(delayBuf.begin(), iterator1, pTaps + (tapsLen - inputPtr), y);
        inputPtr = (inputPtr == 0 ? tapsLen - 1 : inputPtr - 1); // iterate state pointer in reverse
        output[n] = y;
    }
#endif
}

void BinauralSpatializer::dotProduct(const float* pSrc1, const float* pSrc2, int len, float* pDp)
{
    (*pDp) = 0;
    for (int i = 0; i < len; i++) {
        (*pDp) += pSrc1[i] * pSrc2[i];
    }
}

//Cross-correlation
void BinauralSpatializer::CrossCorrNorm(float* src1, float* src2, float* output)
{
    const size_t K = 2 * B;
#if defined(EAR_ENABLE_IPP)
    IppStatus status;
    const int src1Len = B, src2Len = B, dstLen = K - 1;
    int lowLag = -(B - 1);
    int bufSizeNo = 0;
    Ipp8u* pBuffer;

    status = ippsCrossCorrNormGetBufferSize(
        src1Len, src2Len, dstLen, lowLag, ipp32f, ippAlgDirect, &bufSizeNo);
    if (status != ippStsNoErr) {
        return;
    }
    pBuffer = ippsMalloc_8u(bufSizeNo);

    status = ippsCrossCorrNorm_32f(
        src1, src1Len, src2, src2Len, output, dstLen, lowLag, ippAlgDirect, pBuffer);

    ippsFree(pBuffer);
#else
    int lim;
    bool negFLAG;
    const int dstLen = K - 1;
    std::fill(output, output + dstLen, 0.f);

    for (int m = 1; m <= dstLen; m++) {
        int arg = m - B;
        if (arg < 0) {
            negFLAG = true;
            lim = B + arg;
        } else {
            negFLAG = false;
            lim = B - arg;
        }
        for (int n = 1; n <= lim; n++) {
            if (negFLAG == false) {
                output[dstLen - m] += src1[arg + n - 1] * src2[n - 1];
            } else {
                output[dstLen - m] += src1[n - 1] * src2[n - arg - 1];
            }
        }
    }
#endif
}

} // namespace MPEGI
