/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "LoudspeakerSpatializer.h"

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Renderer.h>
#include <MPEGI/Renderer/RenderList.h>

#include <EAR/CICP_Layouts.h>
#include <EAR/Exception.h>
#include <EAR/RealFrame.h>

#include <cassert>

#include "trol/rendering_gd.h"
#include "trol/rendering_pan.h"

namespace {

void copyVector(EAR::Vector3 from, float* to)
{
    std::copy(from.data(), from.data() + 3, to);
}

} // namespace

namespace MPEGI {

std::shared_ptr<Spatializer> makeLoudspeakerSpatializer()
{
    return std::make_shared<LoudspeakerSpatializer>();
}

LoudspeakerSpatializer::LoudspeakerSpatializer()
    : pannerHandle(nullptr)
    , lsRendererHandle(nullptr)
{
    initPannerConfig = std::make_unique<rendering_pan_cfg_t>();
    realtimePannerConfig = std::make_unique<rendering_pan_rt_cfg_t>();

    initLsConfig = std::make_unique<rendering_gd_cfg_t>();
    realtimeLsConfig = std::make_unique<rendering_gd_rt_cfg_t>();

    this->panningEnabled = Config::getBoolDef("TROL_Panning", true);
    this->physicalCompensationEnabled = Config::getBoolDef("TROL_Physical_Compensation", true);
    Config::setDef("TROL_Delay_Compensation", true);
    Config::setDef("TROL_Gain_Compensation", true);
    Config::setDef("TROL_ref_proc", false);
}

LoudspeakerSpatializer::~LoudspeakerSpatializer()
{
    releaseTROL();
}

std::unique_ptr<Spatializer::DspData> LoudspeakerSpatializer::update(const RenderList& renderList)
{
    auto dspData = std::make_unique<LoudspeakerSpatializer::DspData>();
    auto& dspList = *dspData.get();

    auto virtListener = renderList.listener.position;
    auto realListener = renderList.listener.realPosition;
    dspList.listenerPosition = realListener;

    EAR::HomogenMatrix real, virt;
    real.setRotation(realListener.orientation);
    virt.setRotation(virtListener.orientation);
    auto offset_rotation = real * virt.inverse3x3();

    dspList.delayCompensation = Config::getBool("TROL_Delay_Compensation");
    dspList.gainCompensation = Config::getBool("TROL_Gain_Compensation");
    dspList.decay_1_dB = Config::getFloat("TROL_L1_slope1");
    dspList.decay_2_dB = Config::getFloat("TROL_L1_slope2");
    dspList.crit_dist_m = Config::getFloat("TROL_L1_criticalDistance");
    dspList.beta = Config::getFloat("TROL_L1_beta");
    dspList.ref_proc = Config::getBool("TROL_ref_proc");

    for (auto& item : renderList.items) {

        if (item->type == ItemType::HOA) {
            // HOA not processed here
            continue;
        }

        if (item->type == ItemType::Reverb) {
            // These items should be handled via Portals
            continue;
        }

        if (item->type == ItemType::Pannable) {
            // These items are handled by the PannerStage,
            // which creates new items of ItemType::Nondiegetic
            continue;
        }

        if (item->type == ItemType::DirectionalReverb) {
            // TODO: handle portals
            continue;
        }

        if (item->type == ItemType::Downmix) {
            // items of type downmix should be directly mixed into the individual loudspeaker signals
            auto& frame = item->signal->getFrame();
            if (frame.channels() == getNbrOfChannels()) {
                dspList.downmixSignals.emplace_back(&frame);
            }
            continue;
        }

        if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned) {
            auto iter = std::find(objectMapping, objectMapping + MPEGI_MAX_OBJECTS, item->idx);
            if (iter == objectMapping + MPEGI_MAX_OBJECTS) {
                // This might happen if the item is initially inactive or unexpectedly orphaned
                continue;
            }
            *iter = -1;
            continue;
        }

        // get object index if it already exists, find new available index if not

        auto iter = std::find(objectMapping, objectMapping + MPEGI_MAX_OBJECTS, item->idx);
        if (iter == objectMapping + MPEGI_MAX_OBJECTS) {
            iter = std::find(objectMapping, objectMapping + MPEGI_MAX_OBJECTS, -1);
            if (iter == objectMapping + MPEGI_MAX_OBJECTS) {
                // TODO: more graceful failure handling?
                EAR_FAIL("Too many objects for TROLSpatializer");
            }
            *iter = item->idx;
        }

        std::size_t idx = std::distance(objectMapping, iter);

        dspList.objectSignals[idx] = item->signal->getFrame()[0].data();
        if (item->signal->getFrame().channels() != 1) {
            EAR_FAIL("Only single-channel objects are supported in TROLSpatializer");
        }

        if (this->panningEnabled) {
            dspList.objectType[idx] = 1; // i.e. "normal" object source

            auto itemLocation = item->position.location;
            if (item->position.cSpace == EAR::CSpace::LCS) {
                itemLocation = renderList.listener.cs.inverse().transform(item->position.location);
            }

            auto compensated_location =
                offset_rotation * (itemLocation - virtListener.location) + realListener.location;
            copyVector(compensated_location, dspList.objectLocations[idx]);
        } else {
            // objectSignals will be directly copied to output channels
            // (up to the number of loudspeakers, the rest will be ignored).
        }
    }
    return dspData;
}

void LoudspeakerSpatializer::requestBlock(DspContext ctx)
{
    auto& output = ctx.output;
    // clear output buffer (also when processing is not initialized)
    output.zero();

    auto& currentDSP = ctx.getData<LoudspeakerSpatializer::DspData>();

    auto posListener = currentDSP.listenerPosition.location;

    // NB: this cannot be changed during processing, so it's safe to read the variable here
    if (this->panningEnabled) {
        copyVector(posListener, realtimePannerConfig->listener_pos);
        realtimePannerConfig->ref_proc = currentDSP.ref_proc;
        std::copy(&currentDSP.objectLocations[0][0],
            &currentDSP.objectLocations[0][0] + MPEGI_MAX_OBJECTS * 3,
            &realtimePannerConfig->obj_pos[0][0]);
        std::copy(currentDSP.objectType, currentDSP.objectType + MPEGI_MAX_OBJECTS,
            realtimePannerConfig->obj_type);
        int result = rendering_pan_updatecfg(pannerHandle, realtimePannerConfig.get());
        if (result != 0) {
            EAR_FAIL("rendering_pan_updatecfg failed: " + std::to_string(result));
        }
    }

    // NB: this cannot be changed during processing, so it's safe to read the variable here
    if (this->physicalCompensationEnabled) {
        copyVector(posListener, realtimeLsConfig->listener_pos);
        realtimeLsConfig->delay_on = 0; // Delays are applied "manually" below
        realtimeLsConfig->gain_on = currentDSP.gainCompensation;
        realtimeLsConfig->decay_1_dB = currentDSP.decay_1_dB;
        realtimeLsConfig->decay_2_dB = currentDSP.decay_2_dB;
        realtimeLsConfig->crit_dist_m = currentDSP.crit_dist_m;
        realtimeLsConfig->beta = currentDSP.beta;
        realtimeLsConfig->ref_proc = currentDSP.ref_proc;

        int result = rendering_gd_updatecfg(lsRendererHandle, realtimeLsConfig.get());
        if (result != 0) {
            EAR_FAIL("rendering_gd_updatecfg failed: " + std::to_string(result));
        }
    }

    float** inputSignals = currentDSP.objectSignals;
    float** outputSignals = output.dataRaw();

    if (this->panningEnabled) {
        rendering_pan_process(pannerHandle, inputSignals, outputSignals);
    } else {
        assert(output.channels() <= MPEGI_MAX_OBJECTS);
        for (std::size_t i = 0; i < output.channels(); i++) {
            if (float* in_data = inputSignals[i]; in_data != nullptr) {
                std::copy(in_data, in_data + initLsConfig->framesize, outputSignals[i]);
            }
            // If in_data == nullptr -> do nothing, outputSignals has been filled with zeros
        }
    }

    // NB: Switching this at runtime leads to artifacts!
    if (this->physicalCompensationEnabled) {
        rendering_gd_process(lsRendererHandle, outputSignals, outputSignals);
        if (currentDSP.delayCompensation) {
            float* delays = rendering_gd_get_delays(lsRendererHandle);

            for (unsigned int i = 0; i < output.channels(); i++) {
                auto& delayLine = *delayLines[i];
                auto& readCursor = *readCursors[i];
                float* in_channel = outputSignals[i];
                delayLine.write(&in_channel);

                auto delay = delays[i];

                if (delay < readCursor.getMinimumDelay()) {
                    delay = readCursor.getMinimumDelay();
                } else if (delay > initLsConfig->max_delay) {
                    delay = initLsConfig->max_delay;
                }

                delayLine.setDelay(readCursor, delay, /* teleport */ false);
                float* out_channel = output.ch(i).data();
                delayLine.read(readCursor, &out_channel);
            }
        }
    }

    for (auto sig : currentDSP.downmixSignals) {
        for (unsigned int i = 0; i < output.channels(); i++) {
            output.ch(i).add(sig->ch(i));
        }
    }
}

std::size_t LoudspeakerSpatializer::getNbrOfChannels() const
{
    return this->loudspeakerSetup.loudspeakers.size();
}

void LoudspeakerSpatializer::initializeSpatializer(
    const Bitstream& bitstream, float fs, std::size_t B, const SceneState* state)
{
    const auto& setup = state->getLoudspeakerSetup();
    if (!setup) {
        EAR_FAIL("TROLSpatializer: no loudspeaker setup available");
    }
    this->loudspeakerSetup = setup.value();

    auto layout = this->loudspeakerSetup.inputLayout;
    if (layout == 1 || layout == 8) {
        EAR_FAIL(
            "LoudspeakerSpatializer: inputLayout " + std::to_string(layout) + " is not supported");
    }

    initTROL(fs, B);
}

void LoudspeakerSpatializer::initTROL(float fs, std::size_t B)
{
    releaseTROL();

    const EAR::HomogenMatrix mat = this->loudspeakerSetup.position;
    const auto& loudspeakers = this->loudspeakerSetup.loudspeakers;

    if (loudspeakers.empty()) {
        EAR_FAIL("TROLSpatializer: no loudspeakers");
    }

    if (loudspeakers.size() > MPEGI_MAX_OUTPUT_CHANNELS) {
        EAR_FAIL("TROLSpatializer: too many loudspeakers");
    }

    if (!this->panningEnabled && loudspeakers.size() > MPEGI_MAX_OBJECTS) {
        EAR_FAIL("MPEGI_MAX_OBJECTS too small for TROL_Panning == false");
    }

    rendering_pan_defaultcfg(
        initPannerConfig.get(), realtimePannerConfig.get(), this->loudspeakerSetup.inputLayout);
    rendering_gd_defaultcfg(initLsConfig.get(), realtimeLsConfig.get());

    initPannerConfig->framesize = initLsConfig->framesize = B;
    initPannerConfig->sfreq_Hz = initLsConfig->sfreq_Hz = fs;
    initPannerConfig->nchan = initLsConfig->nchan = loudspeakers.size();

    initPannerConfig->nobj = MPEGI_MAX_OBJECTS;
    // Inactive objects are set to -1:
    std::fill(this->objectMapping, this->objectMapping + MPEGI_MAX_OBJECTS, -1);

    EAR::Vector3 referenceOffset{};
    if (auto referencePoint = this->loudspeakerSetup.referencePoint) {
        referenceOffset = referencePoint->position.location;
        realtimeLsConfig->cal_system = referencePoint->calibrated;
    }
    EAR::Vector3 pos = mat * referenceOffset;
    copyVector(pos, realtimeLsConfig->sweet_spot);
    copyVector(pos, realtimePannerConfig->sweet_spot);

    if (auto directivityCompensation = this->loudspeakerSetup.directivityCompensation) {
        realtimeLsConfig->eq_on = 1;
        realtimeLsConfig->eq_freq_Hz =
            Config::getFloatDef("TROL_L1_eqFreq", directivityCompensation->eqFreq);
        realtimeLsConfig->eq_gain_dB =
            Config::getFloatDef("TROL_L1_eqGainDb", directivityCompensation->eqGainDb);
        realtimeLsConfig->eq_angle1_deg =
            Config::getFloatDef("TROL_L1_eqAngleStart", directivityCompensation->eqAngleStart);
        realtimeLsConfig->eq_angle2_deg = realtimeLsConfig->eq_angle1_deg
            + Config::getFloatDef("TROL_L1_eqAngleRange", directivityCompensation->eqAngleRange);
    } else {
        EAR_FAIL("DirectivityCompensation must be populated by EIF parser");
    }

    if (auto levelCompensation = this->loudspeakerSetup.levelCompensation) {
        Config::setDef("TROL_L1_slope1", levelCompensation->slope1);
        Config::setDef("TROL_L1_slope2", levelCompensation->slope2);
        Config::setDef("TROL_L1_criticalDistance", levelCompensation->criticalDistance);
        Config::setDef("TROL_L1_beta", levelCompensation->beta);
    } else {
        EAR_FAIL("LevelCompensation must be populated by EIF parser");
    }

    int result =
        rendering_pan_init(&pannerHandle, initPannerConfig.get(), realtimePannerConfig.get());
    if (result != 0) {
        EAR_FAIL("rendering_pan_init failed: " + std::to_string(result));
    }
    result = rendering_gd_init(&lsRendererHandle, initLsConfig.get(), realtimeLsConfig.get());
    if (result != 0) {
        EAR_FAIL("rendering_gd_init failed: " + std::to_string(result));
    }

    EAR::Quaternion setup_rotation{ this->loudspeakerSetup.position.orientation };
    for (int ch = 0; ch < loudspeakers.size(); ch++) {
        EAR::Vector3 pos = mat * loudspeakers[ch].position.location;
        copyVector(pos, realtimeLsConfig->spk_pos[ch]);
        copyVector(pos, realtimePannerConfig->spk_pos[ch]);
        // Rotation in local coordinate system:
        EAR::Quaternion rot =
            EAR::Quaternion(loudspeakers[ch].position.orientation) * setup_rotation;

        EAR::Vector3 view_point{ 0, 0, -1 }; // default view direction
        view_point = rot.rotate(view_point); // apply loudspeaker rotation
        view_point += pos; // translate to loudspeaker position

        copyVector(view_point, realtimeLsConfig->spk_dir[ch]);
    }

    result = rendering_pan_updatecfg(pannerHandle, realtimePannerConfig.get());
    if (result != 0) {
        EAR_FAIL("rendering_pan_updatecfg failed: " + std::to_string(result));
    }
    result = rendering_gd_updatecfg(lsRendererHandle, realtimeLsConfig.get());
    if (result != 0) {
        EAR_FAIL("rendering_gd_updatecfg failed: " + std::to_string(result));
    }

    float* delays = rendering_gd_get_delays(lsRendererHandle);

    delayLines.clear();
    delayLines.reserve(loudspeakers.size());
    readCursors.clear();
    readCursors.reserve(loudspeakers.size());
    for (int ch = 0; ch < loudspeakers.size(); ch++) {
        delayLines.push_back(std::make_unique<EAR::VDL>(
            fs, B, initLsConfig->max_delay, /* channels */ 1, /* upsampling */ false));
        readCursors.push_back(
            std::make_unique<EAR::VDL::ReadCursor>(delayLines.back()->registerReadCursor(
                delays[ch], EAR::VDL::InterpolationNewton_Spline | EAR::VDL::QualityHigh)));
    }
}

void LoudspeakerSpatializer::releaseTROL()
{
    if (pannerHandle) {
        rendering_pan_done(pannerHandle);
        pannerHandle = nullptr;
    }

    if (lsRendererHandle) {
        rendering_gd_done(lsRendererHandle);
        lsRendererHandle = nullptr;
    }
}

} // namespace MPEGI
