/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_LOUDSPEAKER_SPATIALIZER
#define MPEGI_LOUDSPEAKER_SPATIALIZER

#include <LSDF/LSDFParser.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>
#include <MPEGI/Renderer/Spatializer.h>

#include <EAR/BaseTypes.h>
#include <EAR/VDL.h>

#include <deque>
#include <memory>
#include <vector>

// fwd declares
struct rendering_pan_cfg;
struct rendering_pan_rt_cfg;
struct rendering_gd_cfg;
struct rendering_gd_rt_cfg;
typedef rendering_pan_cfg rendering_pan_cfg_t;
typedef rendering_pan_rt_cfg rendering_pan_rt_cfg_t;
typedef rendering_gd_cfg rendering_gd_cfg_t;
typedef rendering_gd_rt_cfg rendering_gd_rt_cfg_t;

namespace MPEGI {

//! Loudspeaker Spatializer with TROL backend
class LoudspeakerSpatializer : public Spatializer
{
public:
    //! Constructor
    LoudspeakerSpatializer();

    ~LoudspeakerSpatializer();

    //! Inherited from Spatializer
    std::unique_ptr<Spatializer::DspData> update(const RenderList& renderList) override;

    //! Inherited from Spatializer
    void requestBlock(DspContext ctx) override;

    //! Inherited from Spatializer
    std::size_t getNbrOfChannels() const override;

private:
    //! Inherited from Spatializer
    void initializeSpatializer(
        const Bitstream& bitstream, float fs, std::size_t B, const SceneState* state) override;

    //! Initialize a new TROL renderer instance
    inline void initTROL(float fs, std::size_t B);

    //! Destroy an existing TROL renderer instance
    inline void releaseTROL();

    void* pannerHandle;
    std::unique_ptr<rendering_pan_cfg_t> initPannerConfig;
    std::unique_ptr<rendering_pan_rt_cfg_t> realtimePannerConfig;

    void* lsRendererHandle;
    std::unique_ptr<rendering_gd_cfg_t> initLsConfig;
    std::unique_ptr<rendering_gd_rt_cfg_t> realtimeLsConfig;

    struct DspData : Spatializer::DspData
    {
        //! Position of the real listener in the listening room
        EAR::Position listenerPosition;

        //! objects that should be panned
        //! 0: inactive, 1: active, (future values: hom. ext.)
        int objectType[MPEGI_MAX_OBJECTS] = {};
        float* objectSignals[MPEGI_MAX_OBJECTS] = {};
        float objectLocations[MPEGI_MAX_OBJECTS][3] = {};

        //! signals that should be directly mixed to the loudspeaker channels
        std::vector<EAR::RealFrame*> downmixSignals;

        bool delayCompensation;
        bool gainCompensation;
        float decay_1_dB;
        float decay_2_dB;
        float crit_dist_m;
        float beta;
        bool ref_proc;
    };

    //! Will be populated with render item indexes; -1 means empty!
    int objectMapping[MPEGI_MAX_OBJECTS];

    LSDF::LSDFParser::ChannelSourceDesc loudspeakerSetup;

    //! Panning layer (a.k.a object layer) can be deactivated by setting this to "false".
    bool panningEnabled = true;
    //! Physical compensation layer can be deactivated by setting this to "false".
    bool physicalCompensationEnabled = true;

    std::vector<std::unique_ptr<EAR::VDL>> delayLines;
    std::vector<std::unique_ptr<EAR::VDL::ReadCursor>> readCursors;
};

} // namespace MPEGI

#endif /* MPEGI_LOUDSPEAKER_SPATIALIZER */
