/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

/* ************************************************************************** *
 *
 *                            il_highshelving
 *
 * Description:
 * copy of functions from il_biquad filter library
 *     -> high shelving, 2nd order
 *     -> low shelving, 2nd order
 *     -> peaking, 2nd order
 *     -> low-pass, butterworth 2nd order
 *     -> high-pass, butterworth 2nd order
 *     -> il_biquad_filter
 *
 * ************************************************************************** */

#ifdef _WIN32
#   define _USE_MATH_DEFINES    /* for M_PI and the like on Windows */
#endif

#include <math.h>
#include "il_biquad.h"

/**
 * Design a unity filter of 2nd order.
 */
void
il_biquad_unity(
        il_biquad_cf_t          *coefs      /* out: filter coefficients */
)
{
    coefs->b[0] =   1.f;
    coefs->b[1] =   0.f;
    coefs->b[2] =   0.f;
    coefs->a[0] =   0.f;
    coefs->a[1] =   0.f;
}

/**
 * Design a high frequency shelving filter of 2nd order. */
void
il_biquad_hfshelving2(
        il_biquad_cf_t          *coefs,     /* out: filter coefficients */
        float                   fc,         /* in:  cut-off frequency [Hz] */
        float                   G,          /* in:  gain [dB] */
        float const             sfreq       /* in:  sampling rate [Hz] */
)
{
    double          wd, nrm, phi, sp, cp, V0;

    /* restrict range of cut-off frequency to Nyquist */
    fc =            fminf(fc, sfreq * 0.5f);

    /* shift cut filters cutoff frequency to have symmetric curves as boost filters */
    V0 =            pow(10.0, 0.05 * G);
    V0 =            pow(V0, 1.0/2.0);
    if (G < 0.f)
        wd =        tan(M_PI * fc / sfreq) * V0;
    else
        wd =        tan(M_PI * fc / sfreq);

    /* compute next normalized Butterworth polynomial: s^2 + 2*sin(phi)*s + 1 */
    phi =           M_PI * (2.0 * 0 + 1.0) / (2.0  * 2);
    sp =            sin(phi);
    cp =            cos(phi);

    /* normalization factor */
    nrm =           1.0 / ((wd + sp) * (wd + sp) + cp*cp);

    /* coefficient derivation */
    coefs->a[0] =   (float) (nrm * 2.0 * ((wd + sp) * (wd - sp) - cp*cp));
    coefs->a[1] =   (float) (nrm * ((wd - sp) * (wd - sp) + cp*cp));
    sp *=           V0;
    cp *=           V0;
    coefs->b[0] =   (float) (nrm * ((wd + sp) * (wd + sp) + cp*cp));
    coefs->b[1] =   (float) (nrm * 2.0 * ((wd + sp) * (wd - sp) - cp*cp));
    coefs->b[2] =   (float) (nrm * ((wd - sp) * (wd - sp) + cp*cp));
}


/**
 * Design a low frequency shelving filter of 2nd order.
 */
void
il_biquad_lfshelving2(
        il_biquad_cf_t          *coefs,     /* out: filter coefficients */
        float                   fc,         /* in:  cut-off frequency [Hz] */
        float                   G,          /* in:  gain [dB] */
        float const             sfreq       /* in:  sampling rate [Hz] */
)
{
    double          wd, nrm, phi, sp, cp, V0;

    /* restrict range of cut-off frequency to Nyquist */
    fc =            fminf(fc, sfreq * 0.5f);

    /* shift cut filters cutoff frequency to have symmetric curves as boost filters */
    V0 =            powf(10.0, 0.05 * G);
    V0 =            powf(V0, 1.0/2.0);
    if (G < 0.f)
        wd =        tanf(M_PI * fc / (sfreq)) / V0 ;
    else
        wd =        tanf(M_PI * fc / sfreq);

    /* compute next normalized Butterworth polynomial: s^2 + 2*sin(phi)*s + 1 */
    phi =           M_PI * (2.0 * 0 + 1.0) / (2.0  * 2);
    sp =            sin(phi);
    cp =            cos(phi);

    /* normalization factor */
    nrm =           1.0 / ((sp + 1.0/wd) * (sp + 1.0/wd) + cp*cp);

    /* coefficient derivation */
    coefs->a[0] =   (float) (nrm * 2.0 * ((sp + 1.0/wd) * (sp - 1.0/wd) + cp*cp));
    coefs->a[1] =   (float) (nrm * ((sp - 1.0/wd) * (sp - 1.0/wd) + cp*cp));
    sp *=           V0;
    cp *=           V0;
    coefs->b[0] =   (float) (nrm * ((sp + 1.0/wd) * (sp + 1.0/wd) + cp*cp));
    coefs->b[1] =   (float) (nrm * 2.0 * ((sp + 1.0/wd) * (sp - 1.0/wd) + cp*cp));
    coefs->b[2] =   (float) (nrm * ((sp - 1.0/wd) * (sp - 1.0/wd) + cp*cp));
}


/**
 * Design peaking filter of 2nd order.
 */
void
il_biquad_peaking(
        il_biquad_cf_t          *coefs,     /* out: filter coefficients */
        float                   fc,         /* cut-off frequency [Hz] */
        float                   G,          /* gain [dB] */
        float                   Q,          /* Q factor */
        float const             sfreq       /* sampling rate [Hz] */
)
{
    double   wd, gd, qd, nrm;

    /* restrict range of cut-off frequency to Nyquist */
    fc =                fminf(fc, sfreq * 0.5f);
    wd =                tanf(M_PI * fc / sfreq);
    gd =                powf(10.0, 0.05 * fabsf(G));
    qd =                1.0 / Q;

    if (G > 0.f) {
        /* normalization factor */
        nrm =           1.0 / (1.0 + wd*qd + wd*wd);
        /* coefficient derivation */
        coefs->b[0] =   (float) (nrm * (1.0 + gd*qd*wd + wd*wd));
        coefs->b[1] =   (float) (nrm * (2.0 * (wd*wd - 1.0)));
        coefs->b[2] =   (float) (nrm * (1.0 - gd*qd*wd + wd*wd));
        coefs->a[0] =   (float) (nrm * (2.0 * (wd*wd - 1.0)));
        coefs->a[1] =   (float) (nrm * (1.0 - qd*wd + wd*wd));
    }
    else if (G < 0.f) {
        /* normalization factor */
        nrm =           1.0 / (1.0 + gd*qd*wd + wd*wd);
        /* coefficient derivation */
        coefs->b[0] =   (float) (nrm * (1.0 + qd*wd + wd*wd));
        coefs->b[1] =   (float) (nrm * (2.0 * (wd*wd - 1.0)));
        coefs->b[2] =   (float) (nrm * (1.0 - qd*wd + wd*wd));
        coefs->a[0] =   (float) (nrm * (2.0 * (wd*wd - 1.0)));
        coefs->a[1] =   (float) (nrm * (1.0 - gd*qd*wd + wd*wd));
    }
    else {
        coefs->b[0] =   1.f;
        coefs->b[1] =   0.f;
        coefs->b[2] =   0.f;
        coefs->a[0] =   0.f;
        coefs->a[1] =   0.f;
    }
}


/**
 * Apply a biquad IIR filter.
 *
 * Note:
 * The following shows how to run a cascade of filters, using the correct state
 * every time. Note that:
 * - the state pointer is incremented after each call
 *
 *      il_biquad_cf_t    coefs;
 *      il_biquad_st_t    hist[3], *ptr;
 *      ...
 *      ptr =   hist;
 *      il_biquad_filter(coefs, ptr++, ...);
 *      il_biquad_filter(coefs, ptr++, ...);
 *      il_biquad_filter(coefs, ptr, ...);
 */
void
il_biquad_filter(
        il_biquad_cf_t const    coefs,      /* in:      filter coefficients */
        il_biquad_st_t          *state,     /* in/out:  filter memory */
        float const             *in,        /* in:      input signal buffer */
        float                   *out,       /* out:     output signal buffer */
        int const               nsize       /* in:      input signal buffer size */
)
{
    int         k;
    float       xn;
    float       yn, xnd, xndd, ynd, yndd;

    /* load state */
    xnd =       state->xnd;     /* x(n-1) */
    xndd =      state->xndd;    /* x(n-2) */
    ynd =       state->ynd;     /* y(n-1) */
    yndd =      state->yndd;    /* y(n-2) */

    for (k = 0; k < nsize; k++) {

        /* input */
        xn =    *in++;

        /* 2nd order IIR filter */
        yn =    xn * coefs.b[0];
        yn +=   xnd * coefs.b[1];
        yn +=   xndd * coefs.b[2];
        yn -=   ynd * coefs.a[0];
        yn -=   yndd * coefs.a[1];

        /* output */
        *out++ =    yn;

        /* prepare next iteration */
        xndd =  xnd;
        xnd =   xn;
        yndd =  ynd;
        ynd =   yn;
    }

    /* save state */
    state->xnd =    xnd;    /* x(n)   */
    state->xndd =   xndd;   /* x(n-1) */
    state->ynd =    ynd;    /* y(n)   */
    state->yndd =   yndd;   /* y(n-1) */
}
