/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

/* ************************************************************************** *
 *
 *                                 il_biquad
 *
 * Description:
 * copy of functions from il_biquad filter library
 *     -> high shelving, 2nd order
 *     -> low shelving, 2nd order
 *     -> peaking, 2nd order
 *     -> low-pass, butterworth 2nd order
 *     -> high-pass, butterworth 2nd order
 *     -> il_biquad_filter
 *
 * ************************************************************************** */

#ifndef IL_HIGHSHELVING_H
#define IL_HIGHSHELVING_H


/**
 * Coefficients for a biquad.
 */
typedef struct il_biquad_cf
{
    float       b[3];
    float       a[2];
} il_biquad_cf_t;

/**
 * State memory for a biquad.
 */
typedef struct il_biquad_st
{
    float       xnd;
    float       xndd;
    float       ynd;
    float       yndd;
} il_biquad_st_t;


/* API functions
 * ************************************************************************** */

/**
 * Design a unity filter of 2nd order.
 */
void
il_biquad_unity(
        il_biquad_cf_t          *coefs      /* out: filter coefficients */
);

/**
 * Design a high frequency shelving filter of 2nd order. */
void
il_biquad_hfshelving2(
        il_biquad_cf_t          *coefs,     /* out: filter coefficients */
        float                   fc,         /* in:  cut-off frequency [Hz] */
        float                   G,          /* in:  gain [dB] */
        float const             sfreq       /* in:  sampling rate [Hz] */
);

/**
 * Design a high frequency shelving filter of 2nd order. */
void
il_biquad_lfshelving2(
        il_biquad_cf_t          *coefs,     /* out: filter coefficients */
        float                   fc,         /* in:  cut-off frequency [Hz] */
        float                   G,          /* in:  gain [dB] */
        float const             sfreq       /* in:  sampling rate [Hz] */
);

/**
 * Design peaking filter of 2nd order.
 */
void
il_biquad_peaking(
        il_biquad_cf_t          *coefs,     /* out: filter coefficients */
        float                   fc,         /* cut-off frequency [Hz] */
        float                   G,          /* gain [dB] */
        float                   Q,          /* Q factor */
        float const             sfreq       /* sampling rate [Hz] */
);


/**
 * Apply a biquad IIR filter.
 *
 * Note:
 * The following shows how to run a cascade of filters, using the correct state
 * every time. Note that:
 * - the state pointer is incremented after each call
 *
 *      il_biquad_cf_t    coefs;
 *      il_biquad_st_t    hist[3], *ptr;
 *      ...
 *      ptr =   hist;
 *      il_biquad_filter(coefs, ptr++, ...);
 *      il_biquad_filter(coefs, ptr++, ...);
 *      il_biquad_filter(coefs, ptr, ...);
 */
void
il_biquad_filter(
        il_biquad_cf_t const    coefs,      /* in:      filter coefficients */
        il_biquad_st_t          *state,     /* in/out:  filter memory */
        float const             *in,        /* in:      input signal buffer */
        float                   *out,       /* out:     output signal buffer */
        int const               nsize       /* in:      input signal buffer size */
        );

#endif  /* IL_HIGHSHELVING_H */
