/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifdef _WIN32
#   define _USE_MATH_DEFINES    /* for M_PI and the like on Windows */
#endif

#include <stdlib.h>
#include <stdio.h>
#include "math.h"
#include "rendering_gd.h"
#include "il_biquad.h"

/* internal macros
 * ************************************************************************** */

#define SFREQ_MIN           (44100)
#define SFREQ_MAX           (48000)
#define VSOUND              (340.0)                     /* speed of sound in air [m/s] */
#define MAX_DELAY           (960)                       /* maximum delay [samples] */
#define STEREO_WIDTH        (1.0)                       /* default stereo width [m] */
#define OVERHEAD_GAIN       (0.25f)                     /* overhead [lin] */


/* Macros
 * ************************************************************************** */

#define ZEROMEM(p,s)                    \
    do {                                \
        char *q =   (char *)(p);        \
        char *e =   (char *)(p) + s;    \
        while (q < e) {                 \
            *q++ = '\0';                \
        }                               \
    } while (0)

#ifndef MIN
#ifdef __GNUC__
#   define MIN(A,B)     ({ __typeof__ (A) _A = (A); __typeof__ (B) _B = (B); _A < _B ? _A : _B; })
#else
#   define MIN(A,B)     (((A) < (B)) ? (A) : (B))
#endif
#endif

#ifndef MAX
#ifdef __GNUC__
#   define MAX(A,B)     ({ __typeof__ (A) _A = (A); __typeof__ (B) _B = (B); _A > _B ? _A : _B; })
#else
#   define MAX(A,B)     (((A) > (B)) ? (A) : (B))
#endif
#endif


/* data/tables
 * ************************************************************************** */



/* internal data structures
 * ************************************************************************** */

typedef struct {
    /* static parameters */
    float               sfreq_Hz;
    int                 nchan;
    int                 framesize;
    /* real-time parameters */
    int                 bypass_on;
    int                 gain_on;
    int                 eq_on;
    int                 ref_proc;
    float               gain_alpha;
    float               delay0[NCHANMAX];
    float               gain0[NCHANMAX];
    il_biquad_cf_t      iircoefs[NCHANMAX];
    /* state */
    float               gain[NCHANMAX];
    il_biquad_st_t      iirmem[NCHANMAX];
    /* for external delay line */
    float               delay_delta;
    float               delay_delta2;
    float               delay_sm[NCHANMAX];
} rendering_gd_data_t;


/* internal function with prototypes
 * ************************************************************************** */

/* headers */


/* API functions
 * ************************************************************************** */

/**
 * Get default parameters
 */
EXTC void STDCALL
rendering_gd_defaultcfg(
        rendering_gd_cfg_t           *cfg,
        rendering_gd_rt_cfg_t        *rt_cfg
        )
{
    int i;
    
    if (cfg != NULL) {
        cfg->sfreq_Hz =             48000;
        cfg->framesize =            512;
        cfg->nchan =                2;
        cfg->max_delay =            MAX_DELAY;
    }

    if (rt_cfg != NULL) {
        
        ZEROMEM(rt_cfg, sizeof(rt_cfg));

        rt_cfg->bypass_on =         0;
        rt_cfg->delay_on =          1;
        rt_cfg->delay_fade =        1;
        rt_cfg->delay_fract =       1;
        rt_cfg->gain_on =           1;
        rt_cfg->eq_on =             0;
        rt_cfg->ref_proc =          0;
        rt_cfg->cal_system =        0;

        rt_cfg->bound_on =          1;
        rt_cfg->bound_scale =       0.9f;

        rt_cfg->listener_pos[0] =   0;
        rt_cfg->listener_pos[1] =   0;
        rt_cfg->listener_pos[2] =   0;

        rt_cfg->decay_1_dB =        8.0f;
        rt_cfg->decay_2_dB =        0.0f;
        rt_cfg->crit_dist_m =       4.0f;
        rt_cfg->beta =              1.0f;

        rt_cfg->max_m_s =           1.0f;
        rt_cfg->max_m_s_s =         1.0f;
        rt_cfg->gain_ms =           40.0f;

        rt_cfg->eq_freq_Hz =        3000.f;
        rt_cfg->eq_gain_dB =        6.0f;
        rt_cfg->eq_angle1_deg =     30.f;
        rt_cfg->eq_angle2_deg =     60.f;
        rt_cfg->eq_warp =           1.f;
        rt_cfg->eq_dB =             0.f;
    }
}


/**
 * Initialize
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_gd_init(
        rendering_gd_data_h          *handle,
        rendering_gd_cfg_t           *cfg,
        rendering_gd_rt_cfg_t        *rt_cfg
        )
{
    int err = 0;

    rendering_gd_data_t *data = NULL;

    if ((cfg->sfreq_Hz < SFREQ_MIN) || (cfg->sfreq_Hz > SFREQ_MAX)) {
        return -7;
    }

    data = (rendering_gd_data_t *)calloc(1, sizeof(rendering_gd_data_t));
    if (data == NULL) {
        return -1;
    }
    *handle = data;

    /* zero out data */
    ZEROMEM(data, sizeof(rendering_gd_data_t));

    /* init */
    data->sfreq_Hz =    cfg->sfreq_Hz;
    data->nchan =       cfg->nchan;
    data->framesize =   cfg->framesize;
    data->ref_proc =    0;

    cfg->max_delay =    MAX_DELAY;

    /* set rt parameters */
    err = rendering_gd_updatecfg(data, rt_cfg);
    if (err)
        return err;

    /* reset state */
    rendering_gd_audioreset(data);

    return 0;
}


/**
 * Release
 */
EXTC void STDCALL
rendering_gd_done(
        rendering_gd_data_h          handle
        )
{
    rendering_gd_data_t   *data = (rendering_gd_data_t *)handle;

    free(data);
}


/**
 * Reset
 *
 * Flush all internal buffers.
 */
EXTC void STDCALL
rendering_gd_audioreset(
        rendering_gd_data_h         handle
        )
{
    int i;

    rendering_gd_data_t   *data = (rendering_gd_data_t *)handle;

    for (i = 0; i < data->nchan; i++) {
        data->gain[i] =     data->gain0[i];
        data->delay_sm[i] = data->delay0[i];
    }

    ZEROMEM(data->iirmem, sizeof(data->iirmem));
}


/**
 * Update real-time parameters
 *
 * This function should only be called after rendering_gd_init() has been called.
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_gd_updatecfg(
        rendering_gd_data_h          handle,
        rendering_gd_rt_cfg_t        *rt_cfg
        )
{
    int     i, err = 0;
    float   lpos_x, lpos_y, lpos_z, r_ref;

    rendering_gd_data_t   *data = (rendering_gd_data_t *)handle;

    data->bypass_on =   rt_cfg->bypass_on;
    data->gain_on =     rt_cfg->gain_on;
    data->eq_on =       rt_cfg->eq_on;

    /* bypass with delay */
    if (data->bypass_on) {
        data->bypass_on = 0; // no direct bypass
        data->gain_on = 0;
        data->eq_on = 0;
    }

    /* reference readius is first loudspeaker's distance */
    {
        float x, y, z;
        x = rt_cfg->sweet_spot[0] - rt_cfg->spk_pos[0][0];
        y = rt_cfg->sweet_spot[1] - rt_cfg->spk_pos[0][1];
        z = rt_cfg->sweet_spot[2] - rt_cfg->spk_pos[0][2];
        r_ref = sqrtf(x*x+y*y+z*z);
    }

    /* listener position */
    lpos_x = rt_cfg->listener_pos[0];
    lpos_y = rt_cfg->listener_pos[1];
    lpos_z = rt_cfg->listener_pos[2];

    if (rt_cfg->ref_proc) {
        /* set listener position to sweet spot, to below evoke sweet spot processing */
        lpos_x = rt_cfg->sweet_spot[0];
        lpos_y = rt_cfg->sweet_spot[1];
        lpos_z = rt_cfg->sweet_spot[2];
    }

    /* update gain delay calibration */
    for (i = 0; i < data->nchan; i++) {
        float r, x, y, z;

        /* vector: listener position - loudspeaker */
        x = lpos_x - rt_cfg->spk_pos[i][0];
        y = lpos_y - rt_cfg->spk_pos[i][1];
        z = lpos_z - rt_cfg->spk_pos[i][2];
        r = sqrt(x*x+y*y+z*z);

        /* negative of delay from spk to listener */
        data->delay0[i] = MAX_DELAY/2 - (r-r_ref) / VSOUND * data->sfreq_Hz;
        data->delay0[i] = fminf(data->delay0[i], MAX_DELAY-1);
        data->delay0[i] = fmaxf(data->delay0[i], 0.0);

        /* gain relative to 0dB at r_ref */
        {
            float pow_nf_cd, pow_ff_cd, pow_nf, pow_ff, pow_ref, scale;

            /* power at critical distance with decay 1 */
            pow_nf_cd = pow(r_ref / rt_cfg->crit_dist_m, 2.0f * rt_cfg->decay_1_dB/6.0);

            /* power at critical distance with decay 2 */
            pow_ff_cd = pow(r_ref / rt_cfg->crit_dist_m, 2.0f * rt_cfg->decay_2_dB/6.0);

            /* nearfield power at distance r */
            pow_nf = pow(r_ref / (r+1e-20f), 2.0 * rt_cfg->decay_1_dB/6.0);

            /* farfield power at distance r */
            pow_ff = pow(r_ref / (r+1e-20f), 2.0 * rt_cfg->decay_2_dB/6.0);

            /* scale pow_ff to be same as pow_nf at critical distance */
            scale = pow_nf_cd / pow_ff_cd;
            pow_ff *= scale;

            /* power at r_ref */
            pow_ref = powf(1.0f + powf(scale, rt_cfg->beta), (1.0f/rt_cfg->beta));
            pow_ref = fminf(pow_ref, 1.0f/(OVERHEAD_GAIN*OVERHEAD_GAIN));

            /* compenstation gain */
            data->gain0[i] = powf(powf(pow_ref,rt_cfg->beta) / (powf(pow_ff,rt_cfg->beta) + powf(pow_nf,rt_cfg->beta)), 0.5f/rt_cfg->beta) * OVERHEAD_GAIN;
        }
    }
    /* limit gains to 0dB (linked "limiter action") */
    {
        float mx = 0.0f;
        for (i = 0; i < data->nchan; i++) {
            mx = fmaxf(data->gain0[i], mx);
        }
        if (mx > 1.0f) {
            for (i = 0; i < data->nchan; i++) {
                data->gain0[i] /= mx;
            }
        }
    }

    /* samples per frame for specified speed */
    data->delay_delta = (float)data->framesize*rt_cfg->max_m_s/VSOUND;

    /* samples per frame delay change for specified acceleration */
    data->delay_delta2 = (float)data->framesize*rt_cfg->max_m_s_s/VSOUND;

    /* gain smoothing constant */
    data->gain_alpha  =  1.f - expf(- 1.f / (1e-3f * rt_cfg->gain_ms * data->sfreq_Hz));

    /* update eq */
    for (i = 0; i < data->nchan; i++) {
        float ux, uy, uz, ur, vx, vy, vz, vr, angle, gain, r;

        /* reference pos for angles */
        ux = rt_cfg->spk_dir[i][0] - rt_cfg->spk_pos[i][0];
        uy = rt_cfg->spk_dir[i][1] - rt_cfg->spk_pos[i][1];
        uz = rt_cfg->spk_dir[i][2] - rt_cfg->spk_pos[i][2];
        ur = sqrt(ux*ux+uy*uy+uz*uz);

        vx = lpos_x - rt_cfg->spk_pos[i][0];
        vy = lpos_y - rt_cfg->spk_pos[i][1];
        vz = lpos_z - rt_cfg->spk_pos[i][2];
        vr = sqrt(vx*vx+vy*vy+vz*vz);

        angle = (ux*vx+uy*vy+uz*vz)/(ur*vr+1e-20f);
        angle = fminf(angle, 1.0f); // otherwise with ref proc, can be 1.0000...0001, and acos on win undefined
        angle = acos(angle) * 180.0f / M_PI;

        //r = fminf(fabsf(angle),rt_cfg->eq_angle_deg)/(rt_cfg->eq_angle_deg + 1e-20f);
        r = (angle - rt_cfg->eq_angle1_deg)/(fmaxf(rt_cfg->eq_angle2_deg - rt_cfg->eq_angle1_deg, 1e-20f));
        r = fmaxf(fminf(r, 1.0f), 0.0f);
        r = powf(r, rt_cfg->eq_warp);
        gain = r * rt_cfg->eq_gain_dB;

        //printf("angle ch %d:   %f\n", i, angle);

        il_biquad_hfshelving2(&data->iircoefs[i], rt_cfg->eq_freq_Hz, gain, data->sfreq_Hz);
        gain = powf(10.0f, 0.05f*rt_cfg->eq_dB);
        data->iircoefs[i].b[0] *= gain;
        data->iircoefs[i].b[1] *= gain;
        data->iircoefs[i].b[2] *= gain;
    }

    /* make sweet spot action zero for calibrated systems */
    if (rt_cfg->cal_system) {
        float gain_ss, delay_ss; /* sweet spot gain/delay */

        /* set listener position to sweet spot, and compute below sweet spot gains/delays */
        lpos_x = rt_cfg->sweet_spot[0];
        lpos_y = rt_cfg->sweet_spot[1];
        lpos_z = rt_cfg->sweet_spot[2];
        for (i = 0; i < data->nchan; i++) {
            float r, x, y, z;

            /* vector: listener position - loudspeaker */
            x = lpos_x - rt_cfg->spk_pos[i][0];
            y = lpos_y - rt_cfg->spk_pos[i][1];
            z = lpos_z - rt_cfg->spk_pos[i][2];
            r = sqrt(x*x+y*y+z*z);

            /* negative of delay from spk to listener */
            delay_ss = - (r-r_ref) / VSOUND * data->sfreq_Hz;

            /* gain relative to 0dB at r_ref */
            {
                float pow_nf_cd, pow_ff_cd, pow_nf, pow_ff, pow_ref, scale;

                /* power at critical distance with decay 1 */
                pow_nf_cd = pow(r_ref / rt_cfg->crit_dist_m, 2.0f * rt_cfg->decay_1_dB/6.0);

                /* power at critical distance with decay 2 */
                pow_ff_cd = pow(r_ref / rt_cfg->crit_dist_m, 2.0f * rt_cfg->decay_2_dB/6.0);

                /* nearfield power at distance r */
                pow_nf = pow(r_ref / (r+1e-20f), 2.0 * rt_cfg->decay_1_dB/6.0);

                /* farfield power at distance r */
                pow_ff = pow(r_ref / (r+1e-20f), 2.0 * rt_cfg->decay_2_dB/6.0);

                /* scale pow_ff to be same as pow_nf at critical distance */
                scale = pow_nf_cd / pow_ff_cd;
                pow_ff *= scale;

                /* power at r_ref */
                pow_ref = powf(1.0f + powf(scale, rt_cfg->beta), (1.0f/rt_cfg->beta));
                pow_ref = fminf(pow_ref, 1.0f/(OVERHEAD_GAIN*OVERHEAD_GAIN));

                /* compenstation gain */
                gain_ss = powf(powf(pow_ref,rt_cfg->beta) / (powf(pow_ff,rt_cfg->beta) + powf(pow_nf,rt_cfg->beta)), 0.5f/rt_cfg->beta);
            }

            /* remove action from sweet spot (because calibrated system */
            data->delay0[i] -= delay_ss;
            data->delay0[i] = fminf(data->delay0[i], MAX_DELAY-1);
            data->delay0[i] = fmaxf(data->delay0[i], 0.0);
            data->gain0[i] /= gain_ss;
        }
        /* limit gains to 0dB */
        {
            float mx = 0.0f;
            for (i = 0; i < data->nchan; i++) {
                mx = fmaxf(data->gain0[i], mx);
            }
            if (mx > 1.0f) {
                for (i = 0; i < data->nchan; i++) {
                    data->gain0[i] /= mx;
                }
            }
        }
    }

    /* make ref_proc switching instantaneous */
    if (rt_cfg->ref_proc != data->ref_proc) {
        for (i = 0; i < data->nchan; i++) {
            data->gain[i] = data->gain0[i];
            data->delay_sm[i] = data->delay0[i];
        }
        data->ref_proc = rt_cfg->ref_proc;
    }

    return err;
}


/**
 * Audio processing
 *
 * Process one frame of audio, return one frame of audio.
 */
EXTC void STDCALL
rendering_gd_process(
        rendering_gd_data_h     handle,
        float                   **x,            /* in:  input audio */
        float                   **y             /* out: output audio */
        )
{
    int                     n, ch, framesize;
    rendering_gd_data_t     *data = (rendering_gd_data_t *)handle;

    framesize = data->framesize;

    /* bypass */

    if (data->bypass_on) {
        for (ch = 0; ch < data->nchan; ch++) {
            for (n = 0; n < framesize; n++) {
                y[ch][n] = x[ch][n] * OVERHEAD_GAIN;
            }
        }
    } else

    /* apply gain and eq per channel */
    for (ch = 0; ch < data->nchan; ch++) {
        float   *in, *out;
        int     i;

        in =    x[ch];
        out =   y[ch];

        /* apply gain with smoothing */
        {
            float alpha, g, gt;

            alpha = data->gain_alpha;
            if (data->gain_on) {
                gt = data->gain0[ch];
            } else {
                gt = OVERHEAD_GAIN;
            }
            g = data->gain[ch];
            for (i = 0; i < framesize; i++) {
                out[i] = g * in[i];
                g = alpha*gt + (1.0f-alpha)*g;
            }
            data->gain[ch] = g;
        }
        
        /* apply eqs */
        if (data->eq_on) {
            il_biquad_filter(data->iircoefs[ch], &data->iirmem[ch], out, out, framesize);
        }
    }
}


/**
 * Target delays
 *
 * Return a pointer to an array with the target delays for each channel (unit: samples)
 */
EXTC float* STDCALL
rendering_gd_get_delays(
        rendering_gd_data_h     handle
        )
{
    int                 ch;
    float               fdelay;
    rendering_gd_data_t *data = (rendering_gd_data_t *)handle;

    for (ch = 0; ch < data->nchan; ch++) {
        /* limit delay change */
        if (data->delay_sm[ch] > data->delay0[ch]) {
            fdelay = fmaxf(data->delay0[ch], data->delay_sm[ch]-data->delay_delta);
        } else {
            fdelay = fminf(data->delay0[ch], data->delay_sm[ch]+data->delay_delta);
        }
        /* limit delay change 2nd order */
        if (fabsf(fdelay - data->delay_sm[ch]) > data->delay_delta2) {
            if (fdelay > data->delay_sm[ch]) {
                fdelay = data->delay_sm[ch] + data->delay_delta2;
            } else {
                fdelay = data->delay_sm[ch] - data->delay_delta2;
            }
        }
        /* save current target values */
        data->delay_sm[ch] = fdelay;
    }

    return data->delay_sm;
}
