/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef RENDERING_GC_H
#define RENDERING_GC_H

/* use __stdcall calling convention for library functions on windows
 * ************************************************************************** */

#ifdef _WIN32
#   define STDCALL __stdcall
#else
#   define STDCALL
#endif


/* API functions use C linkage for C++ compatibility
 * ************************************************************************** */

#ifdef __cplusplus
#   define EXTC extern "C"
#else
#   define EXTC
#endif


/* Data structures
 * ************************************************************************** */

#ifndef NCHANMAX
#   define NCHANMAX    (16)
#endif

/**
 * Handle to data structure
 */
typedef void * rendering_gd_data_h;

/**
 * Initialization parameters
 */
typedef struct rendering_gd_cfg {
    int         framesize;              /* in:  number of samples per frame */
    float       sfreq_Hz;               /* in:  sampling frequency of input audio */
    int         nchan;                  /* in:  number of channels (loudspeakers) */
    float       max_delay;              /* out: maximum delay [samples] */
} rendering_gd_cfg_t;


/**
 * Real-time updatable parameters
 */
typedef struct rendering_gd_rt_cfg {
    int         bypass_on;              /* 0: normal operation, 1: bypass */
    int         bound_on;               /* 0: listener pos bound off, 1: on */
    float       bound_scale;            /* bounding polygon scaling (1=loudspeaker polygon) */
    int         delay_on;               /* 0: delay off, 1: on */
    int         delay_fract;            /* 0: fractional delays off, 1: on */
    int         delay_fade;             /* 0: delay fade off, 1: on */
    int         gain_on;                /* 0: gain off, 1: on */
    int         eq_on;                  /* 0: eq off, 1: on */
    int         ref_proc;               /* 0: normal operation, 1: processing like for sweet spot */
    int         cal_system;             /* 0: normal operation, 1: calibrated system */
    float       sweet_spot[3];          /* sweet spot position [m,m,m] */
    float       spk_pos[NCHANMAX][3];   /* loudspeaker coordinates [m,m,m] */
    float       spk_dir[NCHANMAX][3];  /* point of loudspeaker look-direction towards inside setup [m,m,m] */
    float       listener_pos[3];        /* listener coordinates [m,m,m] */
    float       decay_1_dB;             /* nearfield sound decay per distance doubling [dB] */
    float       decay_2_dB;             /* farfield sound decay per distance doubling [dB] */
    float       crit_dist_m;            /* critical distance [m] */
    float       beta;                   /* 1: default, >1: faster transition between nearfield and farfield */
    float       max_m_s;                /* maximum rate of delay change [v in m/s] */
    float       max_m_s_s;              /* maximum rate of delay 2nd order change [a in m/s] */
    float       gain_ms;                /* gain smoothing time constant [ms] */
    float       eq_freq_Hz;             /* eq: cut-off frequency [Hz] */
    float       eq_gain_dB;             /* eq: high shevling gain [dB] */
    float       eq_angle1_deg;          /* eq: angle [degree] */
    float       eq_angle2_deg;          /* eq: angle [degree] */
    float       eq_warp;                /* eq: non-linear warping for eq gain */
    float       eq_dB;                  /* eq: overall gain [dB] */
} rendering_gd_rt_cfg_t;


/* API functions
 * ************************************************************************** */

/**
 * Get default parameters
 */
EXTC void STDCALL
rendering_gd_defaultcfg(
        rendering_gd_cfg_t           *cfg,
        rendering_gd_rt_cfg_t        *rt_cfg
        );


/**
 * Initialize
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_gd_init(
        rendering_gd_data_h          *handle,
        rendering_gd_cfg_t           *cfg,
        rendering_gd_rt_cfg_t        *rt_cfg
        );


/**
 * Release
 */
EXTC void STDCALL
rendering_gd_done(
        rendering_gd_data_h          handle
        );


/**
 * Reset
 *
 * Flush all internal buffers.
 */
EXTC void STDCALL
rendering_gd_audioreset(
        rendering_gd_data_h          handle
        );


/**
 * Update real-time parameters
 *
 * This function should only be called after rendering_gd_init() has been called.
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_gd_updatecfg(
        rendering_gd_data_h          handle,
        rendering_gd_rt_cfg_t        *rt_cfg
        );


/**
 * Audio processing
 *
 * Process one frame of audio, return one frame of audio.
 */
EXTC void STDCALL
rendering_gd_process(
        rendering_gd_data_h     handle,
        float                   **x,             /* in:  input audio */
        float                   **y              /* out: output audio */
        );


/**
 * Target delays
 *
 * Return a pointer to an array with the target delays for each channel (unit: samples)
 */
EXTC float* STDCALL
rendering_gd_get_delays(
        rendering_gd_data_h     handle
        );

#endif  /* RENDERING_GC_H */
