/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifdef _WIN32
#   define _USE_MATH_DEFINES    /* for M_PI and the like on Windows */
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "rendering_pan.h"
#include "il_biquad.h"


/* internal macros
 * ************************************************************************** */

#define SFREQ_MIN           (44100)
#define SFREQ_MAX           (48000)
#define FRAMESIZE_MAX       (512)
#define NLAYERS             (2)


/* Macros
 * ************************************************************************** */

#define ZEROMEM(p,s)                    \
    do {                                \
        char *q =   (char *)(p);        \
        char *e =   (char *)(p) + s;    \
        while (q < e) {                 \
            *q++ = '\0';                \
        }                               \
    } while (0)

#ifndef MIN
#ifdef __GNUC__
#   define MIN(A,B)     ({ __typeof__ (A) _A = (A); __typeof__ (B) _B = (B); _A < _B ? _A : _B; })
#else
#   define MIN(A,B)     (((A) < (B)) ? (A) : (B))
#endif
#endif

#ifndef MAX
#ifdef __GNUC__
#   define MAX(A,B)     ({ __typeof__ (A) _A = (A); __typeof__ (B) _B = (B); _A > _B ? _A : _B; })
#else
#   define MAX(A,B)     (((A) > (B)) ? (A) : (B))
#endif
#endif

/* layer and bottom/top rendering bits */
#define BIT_L1      (1<<0)
#define BIT_L2      (1<<1)
#define BIT_BOTTOM  (1<<2)
#define BIT_TOP     (1<<3)
#define BIT_LFE     (0) // for now LFE is never used


/* data/tables
 * ************************************************************************** */


/* internal data structures
 * ************************************************************************** */

typedef struct {
    /* static parameters */
    float               sfreq_Hz;
    int                 nchan;
    int                 nobj;
    int                 framesize;
    /* real-time parameters */
    int                 bypass;
    int                 layer_spks[NLAYERS][NOBJMAX][2];
    float               layer_gains[NLAYERS][NOBJMAX][2];
    float               pan_gains0[NOBJMAX][NCHANMAX];
    float               gain_alpha;
    int                 bottom_top[NOBJMAX];
    float               pan_gains0_bottom_top[NOBJMAX][NCHANMAX];
    il_biquad_cf_t      iircoefs_top[5];
    il_biquad_cf_t      iircoefs_bottom[4];
    /* state */
    float               pan_gains[NOBJMAX][NCHANMAX];
    //float               obj_pos_intern[NOBJMAX][3];    /* object coordinates [m,m,m] */
    int                 obj_type[NOBJMAX];
    float               pan_gains_bottom_top[NOBJMAX][NCHANMAX];
    float               tmp[FRAMESIZE_MAX];
    il_biquad_st_t      iirmem_top[NOBJMAX][5];
    il_biquad_st_t      iirmem_bottom[NOBJMAX][4];
} rendering_pan_data_t;


/* internal function with prototypes
 * ************************************************************************** */

/* data types */
typedef struct {
    float x;
    float y;
    float z;
} vector_t;

/* headers */

static void object_flush(rendering_pan_data_h handle, int index);
static void tangent_law(float alpha, float alpha0, float* a1, float* a2);
static float vector_length(vector_t vector);
static float vector_vector_angle(vector_t a, vector_t b);


/* functions */

static void tangent_law(float alpha, float alpha0, float* a1, float* a2)
{
    float sinsum, sindif, den;

    // alpha0 => zero, gave very large gain, so added these safety range conditions
    alpha0 =    fmaxf(alpha0, 2.0);
    alpha =     fminf(alpha, alpha0);
    alpha =     fmaxf(alpha, -alpha0);

	alpha =     alpha * M_PI / 180.0;
    alpha0 =    alpha0 * M_PI / 180.0;

    sinsum =    sin(alpha0 + alpha);
    sindif =    sin(alpha0 - alpha);
    den =       sqrt(sinsum * sinsum + sindif * sindif);
    *a1 =       sinsum / den;
    *a2 =       sindif / den;
}

static void object_flush(
        rendering_pan_data_h     handle,
        int                      index
        )
{
    int ch;

    rendering_pan_data_t     *data = (rendering_pan_data_t *)handle;

    for (ch = 0; ch < data->nchan; ch++) {
        data->pan_gains0[index][ch] = 0.0f;
        data->pan_gains[index][ch] = 0.0f;
        data->pan_gains0_bottom_top[index][ch] = 0.0f;
        data->pan_gains_bottom_top[index][ch] = 0.0f;
    }
    for (ch = 0; ch < 5; ch++) {
        ZEROMEM(&data->iirmem_top[index][ch], sizeof(il_biquad_st_t));
    }
    for (ch = 0; ch < 4; ch++) {
        ZEROMEM(&data->iirmem_bottom[index][ch], sizeof(il_biquad_st_t));
    }
}

static float vector_length(vector_t vector)
{
  return sqrtf((vector.x*vector.x)+
              (vector.y*vector.y)+
              (vector.z*vector.z));
}

static float vector_vector_angle(vector_t a, vector_t b)
{
    float acosarg;
    
    acosarg = (a.x*b.x + a.y*b.y + a.z*b.z)  / (vector_length(a)*vector_length(b));
    acosarg = fminf(acosarg, 1.0f);

    return acosf(acosarg) * 180.0f / M_PI;
}


/* API functions
 * ************************************************************************** */

/**
 * Get default parameters
 */
EXTC void STDCALL
rendering_pan_defaultcfg(
        rendering_pan_cfg_t           *cfg,
        rendering_pan_rt_cfg_t        *rt_cfg,
        int                           cicp_layout   /* in:  MPEG spk layout (0 = not defined) */
        )
{
    int i;
    
    if (cfg != NULL) {
        cfg->sfreq_Hz =             48000;
        cfg->framesize =            256;
        cfg->nchan =                6;
        cfg->nobj =                 1;
    }

    if (rt_cfg != NULL) {
        
        ZEROMEM(rt_cfg, sizeof(rt_cfg));

        rt_cfg->bypass =            0;
        rt_cfg->gain_ms =           20.0f;
        rt_cfg->ref_proc =          0;
        rt_cfg->l_obj_min =         0.4;
        rt_cfg->ele_eq =            1;
        rt_cfg->ele_thr_deg =       0.0f;

        /* set layer assignments based on CICP layout */
        if (cicp_layout != 0) {
            /* reset layers */
            for (i = 0; i < NCHANMAX; i++) {
                rt_cfg->spk_layer[i] = 0;
            }
            /* set layers according to loudspeaker setup */
            switch (cicp_layout) {
            case 1:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 2:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 3:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 4:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 5:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 6:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 7:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_LFE;
                rt_cfg->spk_layer[3] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[6] = BIT_L1;
                rt_cfg->spk_layer[7] = BIT_L1;
                break;
            case 8:
                // seems to be not used in MPEG-I
                break;
            case 9:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 10:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[3] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 11:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 12:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[6] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[7] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                break;
            case 13:
                // bottom layer panning not supported, so we would use bottom layer for bottom rendering only
                // 24-channels, NCHANMAX too small, for now we do not define it
                break;
            case 14:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM | BIT_TOP;
                rt_cfg->spk_layer[6] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[7] = BIT_L2 | BIT_TOP;
                break;
            case 15:
                // bottom layer panning not supported, so we would use bottom layer for bottom rendering only
                rt_cfg->spk_layer[0] = BIT_L1;
                rt_cfg->spk_layer[1] = BIT_L1;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_BOTTOM;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[6] = BIT_BOTTOM;
                rt_cfg->spk_layer[7] = BIT_L1;
                rt_cfg->spk_layer[8] = BIT_L1;
                rt_cfg->spk_layer[9] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[10] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[11] = BIT_L2 | BIT_TOP;
                break;
            case 16:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[6] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[7] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[8] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[9] = BIT_L2 | BIT_TOP;
                break;
            case 17:
                // we do not support voice of god at this point, so it is just ignored for now
                // solution: set BIT_TOP for voice of god and clear for all channels
                // (but then eq has to be switched off, and TODO: testing)
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[2] = BIT_L1;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[6] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[7] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[8] = BIT_L2;
                rt_cfg->spk_layer[9] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[10] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[11] = 0;
                break;
            case 18:
                // we do not support voice of god at this point, so it is just ignored for now
                // solution: set BIT_TOP for voice of god and clear for all channels
                // (but then eq has to be switched off, and TODO: testing)
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[6] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[7] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[8] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[9] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[10] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[11] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[12] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[13] = 0;
                break;
            case 19:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[6] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[7] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[8] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[9] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[10] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[11] = BIT_L2 | BIT_TOP;
                break;
            case 20:
                rt_cfg->spk_layer[0] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[1] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[2] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[3] = BIT_LFE;
                rt_cfg->spk_layer[4] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[5] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[6] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[7] = BIT_L1 | BIT_BOTTOM;
                rt_cfg->spk_layer[8] = BIT_L2;
                rt_cfg->spk_layer[9] = BIT_L2;
                rt_cfg->spk_layer[10] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[11] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[12] = BIT_L2 | BIT_TOP;
                rt_cfg->spk_layer[13] = BIT_L2 | BIT_TOP;
                break;
            }
        }
    }
}


/**
 * Initialize
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_pan_init(
        rendering_pan_data_h          *handle,
        rendering_pan_cfg_t           *cfg,
        rendering_pan_rt_cfg_t        *rt_cfg
        )
{
    int i, err = 0;

    rendering_pan_data_t *data = NULL;

    if ((cfg->sfreq_Hz < SFREQ_MIN) || (cfg->sfreq_Hz > SFREQ_MAX)) {
        return -7;
    }

    if (cfg->framesize > FRAMESIZE_MAX) {
        return -1;
    }

    data = (rendering_pan_data_t *)calloc(1, sizeof(rendering_pan_data_t));
    if (data == NULL) {
        return -1;
    }
    *handle = data;

    /* zero out data */
    ZEROMEM(data, sizeof(rendering_pan_data_t));

    /* init */
    data->sfreq_Hz =    cfg->sfreq_Hz;
    data->nchan =       cfg->nchan;
    data->framesize =   cfg->framesize;
    data->nobj =        cfg->nobj;
    for (i = 0; i < NOBJMAX; i++) {
        data->obj_type[i] = 1;
    }

    /* set rt parameters */
    err = rendering_pan_updatecfg(data, rt_cfg);
    if (err) {
        return err;
    }

    /* reset state */
    rendering_pan_audioreset(data);

    return 0;
}


/**
 * Release
 */
EXTC void STDCALL
rendering_pan_done(
        rendering_pan_data_h          handle
        )
{
    rendering_pan_data_t   *data = (rendering_pan_data_t *)handle;

    free(data);
}


/**
 * Reset
 *
 * Flush all internal buffers.
 */
EXTC void STDCALL
rendering_pan_audioreset(
        rendering_pan_data_h         handle
        )
{
    int                     o, c;
    rendering_pan_data_t    *data = (rendering_pan_data_t *)handle;

    for (o = 0; o < data->nobj; o++) {
        for (c = 0; c < data->nchan; c++) {
            data->pan_gains[o][c] = data->pan_gains0[o][c];
            data->pan_gains_bottom_top[o][c] = data->pan_gains0_bottom_top[o][c];
        }
    }

    ZEROMEM(data->pan_gains, sizeof(data->pan_gains));
    ZEROMEM(data->pan_gains_bottom_top, sizeof(data->pan_gains_bottom_top));
    ZEROMEM(data->iirmem_top, sizeof(data->iirmem_top));
    ZEROMEM(data->iirmem_bottom, sizeof(data->iirmem_bottom));
}


/**
 * Update real-time parameters
 *
 * This function should only be called after rendering_pan_init() has been called.
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_pan_updatecfg(
        rendering_pan_data_h          handle,
        rendering_pan_rt_cfg_t        *rt_cfg
        )
{
    int     i, k, l, err = 0;
    float   lpos[3], r_ref;
    int     nspks_layer[NLAYERS+2];
    
    rendering_pan_data_t    *data = (rendering_pan_data_t *)handle;

    data->bypass =          rt_cfg->bypass;

    /* amplitude pan objects */
    for (i = 0; i < 3; i++) {
        if (rt_cfg->ref_proc) {
            lpos[i] = rt_cfg->sweet_spot[i];
        } else {
            lpos[i] = rt_cfg->listener_pos[i];
        }
    }

    /* gain smoothing constant */
    data->gain_alpha  = 1.f - expf(- 1.f / (1e-3f * rt_cfg->gain_ms * data->sfreq_Hz));

    /* switch objects on or off */
    for (i = 0; i < data->nobj; i++) {
        if (!rt_cfg->obj_type[i] && data->obj_type[i]) {
            /* flush objects which are switched off */
            object_flush(handle, i);
        }
        data->obj_type[i] = rt_cfg->obj_type[i];
    }

    /* reference readius for object gain is first loudspeaker's distance */
    {
        float x, y, z;
        x = rt_cfg->sweet_spot[0] - rt_cfg->spk_pos[0][0];
        y = rt_cfg->sweet_spot[1] - rt_cfg->spk_pos[0][1];
        z = rt_cfg->sweet_spot[2] - rt_cfg->spk_pos[0][2];
        r_ref = sqrt(x*x+y*y+z*z);
    }

    /* compute number of loudspeakers in each layer */
    for (l = 0; l < NLAYERS+2; l++) {
        nspks_layer[l] = 0;
        for (k = 0; k < data->nchan; k++) {
            if (rt_cfg->spk_layer[k] & (1<<l)) {
                nspks_layer[l]++;
            }
        }
    }

    ZEROMEM(data->pan_gains0, sizeof(data->pan_gains0));
    ZEROMEM(data->pan_gains0_bottom_top, sizeof(data->pan_gains0_bottom_top));
    for (i = 0; i < data->nobj; i++) {
        vector_t l_obj, l_spk[NCHANMAX];
        float   angle_spk_spk, angle_obj_spk;
        float   angles[NCHANMAX], angle1, angle2, a1, a2;
        int     idx1, idx2;

        if (!data->obj_type[i]) {
            /* skip inactive objects */
            continue;
        }

        data->bottom_top[i] = 0;  // 0: bottom is used, 1: top is used

        /* listener - object vector */
        l_obj.x = rt_cfg->obj_pos[i][0] - lpos[0];
        l_obj.y = rt_cfg->obj_pos[i][1] - lpos[1];
        l_obj.z = rt_cfg->obj_pos[i][2] - lpos[2];
        /* prevent issues when listener pos = object pos */
        if (vector_length(l_obj) < 1e-8) {
            l_obj.x = 1e-8;
        }
        /* as object gets closer to listener move it above / below through listener */
        if (vector_length(l_obj) < rt_cfg->l_obj_min) {
            float y;
            y = sqrtf(rt_cfg->l_obj_min*rt_cfg->l_obj_min - l_obj.x*l_obj.x - l_obj.z*l_obj.z);
            if (l_obj.y >= 0) {
                l_obj.y = y;
            } else {
                l_obj.y = -y;
            }
        }

        /* compute angles between listener-object, and listener-loudspeaker directions */
        for (k = 0; k < data->nchan; k++) {
            /* listener - loudspeaker vector */
            l_spk[k].x = rt_cfg->spk_pos[k][0] - lpos[0];
            l_spk[k].y = rt_cfg->spk_pos[k][1] - lpos[1];
            l_spk[k].z = rt_cfg->spk_pos[k][2] - lpos[2];
            /* prevent issues when listener pos = spk pos */
            if (vector_length(l_spk[k]) < 1e-10) {
                l_spk[k].x = 1e-8;
            }
            /* angle between listener-object, and listener-loudspeaker directions */
            {
                float l_spk_y, l_obj_y;
                /* projection on loudspeaker layer plane, set y coordinates to zero */
                l_spk_y = l_spk[k].y;
                l_obj_y = l_obj.y;
                l_spk[k].y = 0.0f;
                l_obj.y = 0.0f;
                angles[k] = vector_vector_angle(l_spk[k], l_obj);
                l_spk[k].y = l_spk_y;
                l_obj.y = l_obj_y;
            }
        }

        /* panning for each layer */
        for (l = 0; l < NLAYERS; l++) {
            /* find the two loudspeakers best matching object position */
            angle1 = 1e10;
            angle2 = 1e10;
            idx1 = -1;
            idx2 = -1;
            for (k = 0; k < data->nchan; k++) {
                if (!(rt_cfg->spk_layer[k] & (1<<l))) {
                    /* skip channels of other layers */
                    continue;
                }
                if (angles[k] < angle1) {
                    angle1 = angles[k];
                    idx1 = k;
                }
            }
            for (k = 0; k < data->nchan; k++) {
                if (!(rt_cfg->spk_layer[k] & (1<<l))) {
                    /* skip channels of other layers */
                    continue;
                }
                if (k == idx1) {
                    /* skip already selected loudspeaker */
                    continue;
                }
                if (angles[k] < angle2) {
                    {
                        float l_spk_y[2];
                        /* projection on loudspeaker layer plane, set y coordinates to zero */
                        l_spk_y[0] = l_spk[k].y;
                        l_spk_y[1] = l_spk[idx1].y;
                        l_spk[k].y = 0.0f;
                        l_spk[idx1].y = 0.0f;
                        angle_spk_spk = vector_vector_angle(l_spk[idx1], l_spk[k]);
                        l_spk[k].y = l_spk_y[0];
                        l_spk[idx1].y = l_spk_y[1];
                    }
                    //angle_obj_spk = vector_vector_angle(l_obj, l_spk[k]);
                    angle_obj_spk = angles[k];
                    //if ((angle_spk_spk >= angle1) && (angle_spk_spk >= angles[k])) {
                    if (angle_spk_spk+0.1f >= angle_obj_spk) { // +0.1 numerical
                        angle2 = angles[k];
                        idx2 = k;
                    }
                }
            }

            /* compute amplitude panning gains for object */
            if (idx2 >= 0) {
                float alpha0, alpha;
                alpha0 = 0.5*(angle1+angle2);
                alpha = alpha0 - angle1;

                tangent_law(alpha, alpha0, &a1, &a2);

                /* save layer's panning info */
                data->layer_spks[l][i][0] = idx1;
                data->layer_spks[l][i][1] = idx2;
                data->layer_gains[l][i][0] = a1;
                data->layer_gains[l][i][1] = a2;
            } else {
                /* indicate that layer has not valid panning */
                data->layer_spks[l][i][0] = -1;
            }
        }

        /* pan between layers */
        {
            if ((data->layer_spks[0][i][0] >= 0) && (data->layer_spks[1][i][0] < 0)) {
                /* first layer valid */
                float l_obj_l_ang_vertical, l_obj_l_ang;

                /* angle between object and panned object */
                {
                    vector_t l_obj_l;

                    /* compute for each layer: listener - pan-direction vector */
                    idx1 = data->layer_spks[0][i][0];
                    idx2 = data->layer_spks[0][i][1];
                    a1 = data->layer_gains[0][i][0];
                    a2 = data->layer_gains[0][i][1];
                    //
                    l_obj_l.x = a1*rt_cfg->spk_pos[idx1][0] + a2*rt_cfg->spk_pos[idx2][0] - lpos[0];
                    l_obj_l.y = a1*rt_cfg->spk_pos[idx1][1] + a2*rt_cfg->spk_pos[idx2][1] - lpos[1];
                    l_obj_l.z = a1*rt_cfg->spk_pos[idx1][2] + a2*rt_cfg->spk_pos[idx2][2] - lpos[2];
                    /* compute angle between panned object and object for layer */
                    l_obj_l_ang = vector_vector_angle(l_obj_l, l_obj);
                    /* threshold for elevation */
                    l_obj_l_ang = fmaxf(l_obj_l_ang-rt_cfg->ele_thr_deg, 0.0f);
                }
                /* angle between object and bottom/top direction */
                {
                    // VIVE: y=-1 = bottom direction!
                    vector_t vertical;
                    float   l_obj_l_ang_bottom, l_obj_l_ang_top;

                    // bottom
                    vertical.x = 0.0f;
                    vertical.y = -1.0f;
                    vertical.z = 0.0f;
                    l_obj_l_ang_bottom = vector_vector_angle(vertical, l_obj);
                    // top
                    vertical.x = 0.0f;
                    vertical.y = 1.0f;
                    vertical.z = 0.0f;
                    l_obj_l_ang_top = vector_vector_angle(vertical, l_obj);

                    // min angle
                    l_obj_l_ang_vertical = fminf(l_obj_l_ang_bottom, l_obj_l_ang_top);
                    data->bottom_top[i] = 0;
                    if (l_obj_l_ang_top < l_obj_l_ang_bottom) {
                        data->bottom_top[i] = 1;
                    }
                }
                /* compute panning gains */
                {
                    float alpha0, alpha, a_layer, a_vertical;

                    alpha0 = 0.5*(l_obj_l_ang+l_obj_l_ang_vertical);
                    alpha = alpha0 - l_obj_l_ang;
                    tangent_law(alpha, alpha0, &a_layer, &a_vertical);

                    data->pan_gains0[i][idx1] = a1 * a_layer;
                    data->pan_gains0[i][idx2] = a2 * a_layer;

                    /* contribution for vertical direction */
                    if (a_vertical > 0.01f) {
                        a_vertical /= sqrtf(nspks_layer[2]); /* distribute power equally to all first layer spks */
                        for (k = 0; k < data->nchan; k++) {
                            if (data->bottom_top[i] && (rt_cfg->spk_layer[k] & 4)) {
                                data->pan_gains0_bottom_top[i][k] += a_vertical;
                            } else if (!data->bottom_top[i] && (rt_cfg->spk_layer[k] & 8)) {
                                data->pan_gains0_bottom_top[i][k] += a_vertical;
                            }
                        }
                    }
                }
            } else if ((data->layer_spks[0][i][0] >= 0) && (data->layer_spks[1][i][0] >= 0)) {
                vector_t l_obj_l[NLAYERS];
                float   l_obj_l_ang[NLAYERS], angle_l_l, alpha, alpha0;
                float a_bottom, a_top;
                
                a_bottom = 0.0f;
                a_top = 0.0f;
                
                /* both layers valid */
                for (l = 0; l < NLAYERS; l++) {
                    /* compute for each layer: listener - pan-direction vector */
                    idx1 = data->layer_spks[l][i][0];
                    idx2 = data->layer_spks[l][i][1];
                    a1 = data->layer_gains[l][i][0];
                    a2 = data->layer_gains[l][i][1];
                    //
                    l_obj_l[l].x = a1*rt_cfg->spk_pos[idx1][0] + a2*rt_cfg->spk_pos[idx2][0] - lpos[0];
                    l_obj_l[l].y = a1*rt_cfg->spk_pos[idx1][1] + a2*rt_cfg->spk_pos[idx2][1] - lpos[1];
                    l_obj_l[l].z = a1*rt_cfg->spk_pos[idx1][2] + a2*rt_cfg->spk_pos[idx2][2] - lpos[2];
                    /* compute angle between panned object and object for layer */
                    l_obj_l_ang[l] = vector_vector_angle(l_obj_l[l], l_obj);
                }
                angle_l_l = vector_vector_angle(l_obj_l[0], l_obj_l[1]);
                /* if direction between layers: tangent law between layers */
                if ((angle_l_l > l_obj_l_ang[0]) && (angle_l_l > l_obj_l_ang[1])) {
                    alpha0 = 0.5*(l_obj_l_ang[0]+l_obj_l_ang[1]);
                    alpha = alpha0 - l_obj_l_ang[0];
                    tangent_law(alpha, alpha0, &a1, &a2);
                    
                    /* panning between layer 0 and bottom direction */
                } else if (l_obj_l_ang[0] < l_obj_l_ang[1]) {
                    /* threshold for elevation */
                    l_obj_l_ang[0] = fmaxf(l_obj_l_ang[0]-rt_cfg->ele_thr_deg, 0.0f);
                    l_obj_l_ang[1] = fmaxf(l_obj_l_ang[1]-rt_cfg->ele_thr_deg, 0.0f);
                    
                    // VIVE: y=-1 = bottom direction!
                    vector_t bottom;
                    float   l_obj_l_ang_bottom;
                    
                    bottom.x = 0.0f;
                    bottom.y = -1.0f;
                    bottom.z = 0.0f;
                    
                    l_obj_l_ang_bottom = vector_vector_angle(bottom, l_obj);
                    
                    alpha0 = 0.5*(l_obj_l_ang[0]+l_obj_l_ang_bottom);
                    alpha = alpha0 - l_obj_l_ang[0];
                    tangent_law(alpha, alpha0, &a1, &a_bottom);
                    a2 = 0.0f;
                    
                    /* panning between layer 1 and top direction */
                } else {
                    data->bottom_top[i] = 1;
                    // VIVE: y=1 = top direction!
                    vector_t top;
                    float   l_obj_l_ang_top;
                    
                    top.x = 0.0f;
                    top.y = 1.0f;
                    top.z = 0.0f;
                    
                    l_obj_l_ang_top = vector_vector_angle(top, l_obj);
                    
                    alpha0 = 0.5*(l_obj_l_ang[1]+l_obj_l_ang_top);
                    alpha = alpha0 - l_obj_l_ang[1];
                    tangent_law(alpha, alpha0, &a2, &a_top);
                    a1 = 0.0f;
                }
                {
                    int spk_1_L1, spk_2_L1, spk_1_L2, spk_2_L2;
                    spk_1_L1 = data->layer_spks[0][i][0];
                    spk_2_L1 = data->layer_spks[0][i][1];
                    spk_1_L2 = data->layer_spks[1][i][0];
                    spk_2_L2 = data->layer_spks[1][i][1];
                    /* pan between layers */
                    data->pan_gains0[i][spk_1_L1] = a1*data->layer_gains[0][i][0];
                    data->pan_gains0[i][spk_2_L1] = a1*data->layer_gains[0][i][1];
                    data->pan_gains0[i][spk_1_L2] = a2*data->layer_gains[1][i][0];
                    data->pan_gains0[i][spk_2_L2] = a2*data->layer_gains[1][i][1];
                    /* modify normalization when signal of two layers goes to same loudspeaker */
                    if ((a1 > 0.05f) && (a2 > 0.05f)) { /* only if both layers are used */
                        a1 = 1.0f / (a1 + a2);
                        a2 = 1.0f - a1;
                        if (spk_1_L1 == spk_1_L2) {
                            data->pan_gains0[i][spk_1_L1] = a1*data->layer_gains[0][i][0] + a2*data->layer_gains[1][i][0];;
                        }
                        if (spk_2_L1 == spk_2_L2) {
                            data->pan_gains0[i][spk_2_L1] = a1*data->layer_gains[0][i][1] + a2*data->layer_gains[1][i][1];;
                        }
                        if (spk_1_L1 == spk_2_L2) {
                            data->pan_gains0[i][spk_1_L1] = a1*data->layer_gains[0][i][0] + a2*data->layer_gains[1][i][1];;
                        }
                        if (spk_2_L1 == spk_1_L2) {
                            data->pan_gains0[i][spk_2_L1] = a1*data->layer_gains[0][i][1] + a2*data->layer_gains[1][i][0];;
                        }
                    }
                }
                /* contribution for bottom direction */
                if (a_bottom > 0.01f) {
                    a_bottom /= sqrtf(nspks_layer[2]); /* distribute power equally to all first layer spks */
                    for (k = 0; k < data->nchan; k++) {
                        if (rt_cfg->spk_layer[k] & 4) {
                            data->pan_gains0_bottom_top[i][k] += a_bottom;
                        }
                    }
                }
                /* contribution for top direction */
                if (a_top > 0.01f) {
                    a_top /= sqrtf(nspks_layer[3]); /* distribute power equally to all second layer spks */
                    for (k = 0; k < data->nchan; k++) {
                        if (rt_cfg->spk_layer[k] & 8) {
                            data->pan_gains0_bottom_top[i][k] += a_top;
                        }
                    }
                }
            }
        }
    }

    for (i = 0; i < 5; i++) {
        il_biquad_unity(&data->iircoefs_top[i]);
    }
    for (i = 0; i < 4; i++) {
        il_biquad_unity(&data->iircoefs_bottom[i]);
    }
    if (rt_cfg->ele_eq == 1) {
        float strength;
        /* top eq */
        il_biquad_hfshelving2(&data->iircoefs_top[0], 600.0f, -2.8f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_top[1], 8400.0f, 6.0f, 3.0f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_top[2], 1250.0f, 2.5f, 2.5f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_top[3], 720.0f, -4.0f, 2.5f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_top[4], 13000.0f, -7.0f, 2.5f, data->sfreq_Hz);
        /* bottom eq */
        strength = 4.0f/3.0f; // as bottom eq is defined for ele = -45
        il_biquad_lfshelving2(&data->iircoefs_bottom[0], 800.0f, -2.0f*strength, data->sfreq_Hz);
        il_biquad_hfshelving2(&data->iircoefs_bottom[1], 3000.0f, -3.0f*strength, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_bottom[2], 9000.0f, 4.0f*strength, 3.0f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_bottom[3], 6400.0f, -6.0f*strength, 4.0f, data->sfreq_Hz);
    } else if (rt_cfg->ele_eq == 2) {
        il_biquad_peaking(&data->iircoefs_top[0], 800.0f, -10.0f, 2.0f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_bottom[0], 6000.0f, -10.0f, 1.5f, data->sfreq_Hz);
    } else if (rt_cfg->ele_eq == 3) {
        il_biquad_peaking(&data->iircoefs_top[0], 800.0f, -6.0f, 2.0f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_top[1], 1300.0f, 4.0f, 2.5f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_top[2], 7500.0f, 5.0f, 2.0f, data->sfreq_Hz);
        il_biquad_hfshelving2(&data->iircoefs_bottom[0], 600.0f, -3.0f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_bottom[1], 900.0f, 4.0f, 1.0f, data->sfreq_Hz);
        il_biquad_peaking(&data->iircoefs_bottom[2], 8000.0f, 3.0f, 2.0f, data->sfreq_Hz);
    }

    return err;
}


/**
 * Audio processing
 *
 * Process one frame of audio, return one frame of audio.
 */
EXTC void STDCALL
rendering_pan_process(
        rendering_pan_data_h     handle,
        float                    **x,             /* in:  input audio */
        float                    **y              /* out: output audio */
        )
{
    int                     framesize;
    rendering_pan_data_t     *data = (rendering_pan_data_t *)handle;

    framesize = data->framesize;

    /* bypass: simply passing objects to output channels */
    if (data->bypass) {
        int c, i, nobj;

        nobj = data->nobj;
        nobj = MIN(nobj, data->nchan);

        for (c = 0; c < nobj; c++) {
            if (!data->obj_type[c]) {
                /* silence inactive objects */
                for (i = 0; i < data->framesize; i++) {
                    y[c][i] = 0.0f;
                }
            }
            for (i = 0; i < data->framesize; i++) {
                y[c][i] = x[c][i];
            }
        }
        for ( ; c < data->nchan; c++) {
            for (i = 0; i < data->framesize; i++) {
                y[c][i] = 0.0f;
            }
        }
        return;
    }

    /* zero output */
    {
        int c, i;
        for (c = 0; c < data->nchan; c++) {
            for (i = 0; i < data->framesize; i++) {
                y[c][i] = 0.0f;
            }
        }
    }

    /* render objects */
    {
        int c, i, o;
        for (o = 0; o < data->nobj; o++) {
            if (!data->obj_type[o]) {
                /* skip inactive objects */
                continue;
            }
            for (c = 0; c < data->nchan; c++) {
                for (i = 0; i < data->framesize; i++) {
                    y[c][i] += data->pan_gains[o][c] * x[o][i];
                    data->pan_gains[o][c] = data->gain_alpha*data->pan_gains0[o][c] + (1.0f-data->gain_alpha)*data->pan_gains[o][c];
                }
            }
        }
        for (o = 0; o < data->nobj; o++) {
            if (!data->obj_type[o]) {
                /* skip inactive objects */
                continue;
            }
            if (data->bottom_top[o]) {
                for (i = 0; i < 5; i++) {
                    il_biquad_filter(
                        data->iircoefs_top[i],
                        &data->iirmem_top[o][i],
                        i == 0 ? x[o] : data->tmp,
                        data->tmp, data->framesize);
                }
            } else {
                for (i = 0; i < 4; i++) {
                    il_biquad_filter(
                        data->iircoefs_bottom[i],
                        &data->iirmem_bottom[o][i],
                        i == 0 ? x[o] : data->tmp,
                        data->tmp, data->framesize);
                }
            }
            for (c = 0; c < data->nchan; c++) {
                for (i = 0; i < data->framesize; i++) {
                    y[c][i] += data->pan_gains_bottom_top[o][c] * data->tmp[i];
                    data->pan_gains_bottom_top[o][c] = data->gain_alpha*data->pan_gains0_bottom_top[o][c] + (1.0f-data->gain_alpha)*data->pan_gains_bottom_top[o][c];
                }
            }
        }
    }
}


/**
 * Get current rendering gains
 *
 * Usage:
 *    float   *gains[N];    // (N >= nobj)
 *
 *    rendering_get_rendering_gains(handle, &gains);  //(pointers to arrays with the gains will inserted into gains)
 */
EXTC void STDCALL
rendering_pan_get_rendering_gains(
        rendering_pan_data_h     handle,
        float                    **gains
        )
{
    int o;
    rendering_pan_data_t   *data = (rendering_pan_data_t *)handle;

    for (o = 0; o < data->nobj; o++) {
        gains[o] = data->pan_gains0[o];
    }
}

