/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef RENDERING_PAN_H
#define RENDERING_PAN_H


/* use __stdcall calling convention for library functions on windows
 * ************************************************************************** */

#ifdef _WIN32
#   define STDCALL __stdcall
#else
#   define STDCALL
#endif


/* API functions use C linkage for C++ compatibility
 * ************************************************************************** */

#ifdef __cplusplus
#   define EXTC extern "C"
#else
#   define EXTC
#endif


/* Data structures
 * ************************************************************************** */

#ifndef NCHANMAX
#   define NCHANMAX     (16)
#endif
#ifndef NOBJMAX
#   define NOBJMAX      (64)
#endif


/**
 * Handle to data structure
 */
typedef void * rendering_pan_data_h;

/**
 * Initialization parameters
 */
typedef struct rendering_pan_cfg {
    int         framesize;              /* in:  number of samples per frame */
    float       sfreq_Hz;               /* in:  sampling frequency of input audio */
    int         nchan;                  /* in:  number of channels (loudspeakers) */
    int         nobj;                   /* in:  number of objects to amplitude pan */
} rendering_pan_cfg_t;


/**
 * Real-time updatable parameters
 */
typedef struct rendering_pan_rt_cfg {
    int         bypass;                 /* 0: normal operation, 1: bypass */
    int         ref_proc;               /* 0: normal operation, 1: processing like for sweet spot */
    float       gain_ms;                /* gain smoothing time constant [ms] */
    float       sweet_spot[3];          /* sweet spot position [m,m,m] */
    float       listener_pos[3];        /* listener coordinates [m,m,m] */
    float       spk_pos[NCHANMAX][3];   /* loudspeaker coordinates [m,m,m] */
    /* spk_layer[NCHANMAX]:
        Each speaker can have these bits/flags
            bit0 (1):   loudspeakers for layer 1 panning
            bit1 (2):   loudspeakers for layer 2 panning
            bit2 (4):   loudspeakers for bottom direction panning (usually these have also bit0 and/or bit1 set)
            bit3 (8):   loudspeakers for top direction panning (usually also bit0 and/or bit1 are set)
            bit4 (16):  exclude loudspeaker for low-band processing
    */
    int         spk_layer[NCHANMAX];
    float       obj_pos[NOBJMAX][3];    /* object coordinates [m,m,m] */
    int         obj_type[NOBJMAX];      /* 0: inactive, 1: active, (future values: hom. ext.) */
    float       l_obj_min;              /* minimum distance listener-object [m]
                                           object will fly over/under listener at this distance,
                                           instead of going through listener */    /* top/bottom rendering */
    int         ele_eq;                 /* 0: off, 1: default top/bottom eq, 2-3: alternative eqs */
    float       ele_thr_deg;            /* 0..180 => elevation threshold (0=off, >=90=elevation off) */
} rendering_pan_rt_cfg_t;


/* API functions
 * ************************************************************************** */

/**
 * Get default parameters
 */
EXTC void STDCALL
rendering_pan_defaultcfg(
        rendering_pan_cfg_t           *cfg,
        rendering_pan_rt_cfg_t        *rt_cfg,
        int                           cicp_layout   /* in:  MPEG spk layout (0 = not defined) */
        );


/**
 * Initialize
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_pan_init(
        rendering_pan_data_h          *handle,
        rendering_pan_cfg_t           *cfg,
        rendering_pan_rt_cfg_t        *rt_cfg
        );


/**
 * Release
 */
EXTC void STDCALL
rendering_pan_done(
        rendering_pan_data_h          handle
        );


/**
 * Reset
 *
 * Flush all internal buffers.
 */
EXTC void STDCALL
rendering_pan_audioreset(
        rendering_pan_data_h          handle
        );


/**
 * Update real-time parameters
 *
 * This function should only be called after rendering_pan_init() has been called.
 *
 * Return:
 *  - 0 on success
 *  - error code on error
 */
EXTC int STDCALL
rendering_pan_updatecfg(
        rendering_pan_data_h          handle,
        rendering_pan_rt_cfg_t        *rt_cfg
        );


/**
 * Audio processing
 *
 * Process one frame of audio, return one frame of audio.
 */
EXTC void STDCALL
rendering_pan_process(
        rendering_pan_data_h     handle,
        float                    **x,             /* in:  input audio */
        float                    **y              /* out: output audio */
        );

/**
 * Get current rendering gains
 *
 * Usage:
 *    float   *gains[N];    // (N >= nobj)
 *
 *    rendering_get_rendering_gains(handle, &gains);  //(pointers to arrays with the gains will inserted into gains)
 */
EXTC void STDCALL
rendering_pan_get_rendering_gains(
        rendering_pan_data_h     handle,
        float                    **gains
        );

#endif  /* RENDERING_PAN_H */
