/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Airflow.h"
#include "SpeedProfile.h"

#include <EAR/AirflowGeneratorBase.h>
#include <EAR/PointSource.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <cmath>
#include <iostream>

namespace MPEGI {

std::shared_ptr<RenderStage> makeAirflowStage()
{
    return std::make_shared<AirflowStage>();
}

AirflowStage::AirflowStage()
    : itemStore(nullptr)
{}

AirflowStage::~AirflowStage() {}

std::unique_ptr<RenderStage::DspData> AirflowStage::update(RenderList& renderList)
{
    if (!itemStore) {
        return nullptr;
    }
    if (streamManager->getB() == 0) { // only update once initialized
        // Null pointer signals to audio thread that the stage is not ready
        return nullptr;
    }

    if (!bitstreamAirflowPayload) {
        // no airflow payload
        return nullptr;
    }

    auto dspData = std::make_unique<AirflowStage::DspData>();

    const Position listenerPosition = renderList.listener.position;
    const Vector3 listenerLocation = listenerPosition.location;
    const Euler listenerOrientation = listenerPosition.orientation;

    // Get listener environment bitstream id
    const std::optional<int> listenerEnvironmentBsId =
        environmentIdToBsIdMap.at(renderList.listener.acousticEnvironmentId);

    // For each source in the scene, check if it is active in the current listener AE.
    const auto& airflowSources = bitstreamAirflowPayload->GetAirflowSources();
    Vector3 airflow;
    for (const auto& [id, source] : bitstreamAirflowPayload->GetAirflowSources()) {
        auto& contents = items.at(id);
        auto& [renderItem, buffer, pinknoiseBuffer, generators, prevFilterGains] = contents;
        if (source->isActiveInRegion(listenerEnvironmentBsId)
            && generatorMap[source->generatorID]->isActive()) {
            airflow = generatorMap[source->generatorID]->getAirspeed(listenerLocation);

            // Convert airflow direction (direction the airflow is coming from) to relative polar coordinates compared to the listener.
            const Polar relativeAirflow = getAirflowOrigin(listenerOrientation, airflow);

            // Compute eq filter for each airflow source
            const auto& source = bitstreamAirflowPayload->GetAirflowSources().at(id);
            const auto& [speedProfiles, elevationProfiles, azimuthProfiles] =
                airflowProfilesMap.at(id);

            const auto gain = interpolateAndCombineProfiles(relativeAirflow[2], relativeAirflow[1],
                relativeAirflow[0], speedProfiles, elevationProfiles, azimuthProfiles);

            // Set and update prev filter gains
            if (prevFilterGains[0].empty() || prevFilterGains[1].empty()) {
                dspData->prevFilterGains[id] = { gain.first, gain.second };
            } else {
                dspData->prevFilterGains[id] = prevFilterGains;
            }

            prevFilterGains = { gain.first, gain.second };

            dspData->filterGains[id] = { gain.first, gain.second };
            dspData->generatorGain[id] = generatorMap[source->generatorID]->getGainFactor();
            renderItem->setGain(1.f);
            renderItem->setStatus(ItemStatus::Active);
            renderList.addItemCopy(renderItem);
        } else {
            renderItem->setGain(0.f);
            renderItem->setStatus(ItemStatus::Inactive);
        }
    }

    return dspData;
}

void AirflowStage::processBlock(DspContext ctx)
{
    auto& dspData = ctx.getData<AirflowStage::DspData>();
    frameCount++;

    for (const auto& [airflowSourceIndex, filterGains] : dspData.filterGains) {
        auto&& [renderItem, buffer, pinkNoiseBuffer, generators, prevGainsIgnore] =
            items.at(airflowSourceIndex);

        // NOTE: prevGains is currently both accessible from items and from dspData.prevFilterGains. The one is items should not be used to prevent a data race. This code should be refactored so this isn't possible at all.

        auto& prevGains = dspData.prevFilterGains.at(airflowSourceIndex);

        fillBuffersWithPinkNoise(pinkNoiseBuffer, generators);

        // Apply filters
        const size_t frameLength = buffer->getFrame()[0].length();
        auto overallGain = dspData.generatorGain.at(airflowSourceIndex);

        // loop over the left and right channels
        for (size_t ear = 0; ear < 2; ++ear) {
            // reset current ear output buffer to zero.
            buffer->getFrame()[ear].fill(0.0f);
            // loop through all filters
            for (size_t filterIndex = 0; filterIndex < filters[airflowSourceIndex][0].size();
                 ++filterIndex) {
                // loop through all samples in the buffer.
                for (size_t offset = 0; offset < frameLength; ++offset) {

                    float* in = pinkNoiseBuffer->getFrame()[ear].data() + offset;
                    float* out = buffer->getFrame()[ear].data() + offset;

                    // Interpolate filter gains over the dsp block
                    const float nextGainWeight = (float)(offset) / (float)(frameLength);
                    const float prevGainWeight = 1.f - nextGainWeight;
                    const float interpolatedGain = prevGainWeight * prevGains[ear][filterIndex]
                        + nextGainWeight * filterGains[ear][filterIndex];
                    // sum filter output to output buffer with gains applied
                    *out += (filters[airflowSourceIndex][ear][filterIndex].process(*in))
                        * interpolatedGain * overallGain;
                }
            }
        }
    }

    // Interpolation of gains has finished this block, so next processBlock (which may use the same dspData if no update() call has happened) must not do it again.
    dspData.prevFilterGains = dspData.filterGains;
}

using ProfileType = MPEGI::AirflowProfileParameters::ProfileType;
using ProfileModification = MPEGI::AirflowProfileParameters::ProfileModification;

float dBToLinear(float dB)
{
    return std::pow(10, dB / 20.f);
}

// Convenience function for converting frequency-gain pairs to third octave frequencies
std::array<std::vector<float>, 2> AirflowStage::convertFreqGainPairs(
    const MPEGI::AirflowPayload& payload, const MPEGI::AirflowProfileParameters& profile,
    const std::vector<float>& thirdOctaveCenterFreqs)
{
    std::array<std::vector<float>, 2> thirdOctaveSpectra;
    thirdOctaveSpectra[0].reserve(thirdOctaveCenterFreqs.size());
    thirdOctaveSpectra[1].reserve(thirdOctaveCenterFreqs.size());

    // Interpolate given frequency grid to third octave frequencies
    const auto& freqGrid = payload.GetFreqGrid()[profile.freqGridIdx];
    auto interpolator = std::array<PchipInterpolator, 2>{ true, true };

    EAR_CHECK_REL(freqGrid->centerFreq.size() == profile.gain.size(),
        "Frequency grid must be the same size as the gains");

    for (int i = 0; i < freqGrid->centerFreq.size(); ++i) {
        // Also converts from dB to linear gains
        interpolator[0].addSupportingPoint(
            freqGrid->centerFreq[i], dBToLinear(profile.gain[i].first));
        interpolator[1].addSupportingPoint(
            freqGrid->centerFreq[i], dBToLinear(profile.gain[i].second));
    }

    for (const auto& f : thirdOctaveCenterFreqs) {
        thirdOctaveSpectra[0].push_back(interpolator[0].interpolate(f));
        thirdOctaveSpectra[1].push_back(interpolator[1].interpolate(f));
    }

    return thirdOctaveSpectra;
}

std::array<float, 3> filtToPolynomial(float cf, float Q, float gain)
{
    // Prevent numerical issues.
    if (std::abs(cf) < 0.0001 || std::abs(Q) < 0.0001) {
        return { 0.f, 0.f, 0.f };
    }

    const float bw = cf / Q;

    const float x1 = cf - (bw / 2);
    const float x2 = cf;
    const float x3 = cf + (bw / 2);

    const float y1 = gain - 3;
    const float y2 = gain;
    const float y3 = gain - 3;

    const float a =
        (x1 * (y3 - y2) + x2 * (y1 - y3) + x3 * (y2 - y1)) / ((x1 - x2) * (x1 - x3) * (x2 - x3));
    const float b = ((y2 - y1) / (x2 - x1)) - a * (x1 + x2);
    const float c = y1 - (a * (x1 * x1)) - (b * x1);

    return { a, b, c };
}

std::vector<float> evaluatePolynomial(
    const std::array<float, 3>& poly, const std::vector<float>& freqs)
{
    std::vector<float> values;
    values.reserve(freqs.size());

    for (const auto& f : freqs) {
        const auto& [a, b, c] = poly;
        values.push_back(a * f * f + b * f + c);
    }

    return values;
}

// Convenience function for converting filter params to third octave frequencies
std::array<std::vector<float>, 2> AirflowStage::convertFilterParams(
    const MPEGI::AirflowPayload& payload, const MPEGI::AirflowProfileParameters& profile,
    const std::vector<float>& thirdOctaveCenterFreqs)
{
    std::array<std::vector<float>, 2> thirdOctaveSpectra;
    thirdOctaveSpectra[0].reserve(thirdOctaveCenterFreqs.size());
    thirdOctaveSpectra[1].reserve(thirdOctaveCenterFreqs.size());
    const auto& freqGrid = payload.GetFreqGrid()[profile.freqGridIdx];

    EAR_CHECK_REL(
        profile.q.size() == freqGrid->centerFreq.size() && profile.q.size() == profile.gain.size(),
        "Q, gain and frequency grid must be the same size");

    EAR_CHECK_REL(profile.q.size() > 0, "Q, gain and frequency must be non-empty");

    std::vector<std::array<std::vector<float>, 2>>
        singleFilterResponse; // One for each Q, cf, gain tuple
    singleFilterResponse.reserve(profile.q.size());
    for (int filter = 0; filter < profile.q.size(); ++filter) {
        const float Q1 = profile.q[filter].first;
        const float Q2 = profile.q[filter].second;
        const float cf = freqGrid->centerFreq[filter];
        const float gain1 = profile.gain[filter].first;
        const float gain2 = profile.gain[filter].second;

        const auto& poly1 = filtToPolynomial(cf, Q1, gain1);
        const auto& poly2 = filtToPolynomial(cf, Q2, gain2);

        singleFilterResponse.push_back({ evaluatePolynomial(poly1, thirdOctaveCenterFreqs),
            evaluatePolynomial(poly2, thirdOctaveCenterFreqs) });
    }

    // Compute the maximum over the filter responses
    thirdOctaveSpectra = singleFilterResponse[0];
    for (int filter = 1; filter < singleFilterResponse.size(); ++filter) {
        for (int ear = 0; ear < 2; ++ear) {
            for (int i = 0; i < thirdOctaveCenterFreqs.size(); ++i) {
                thirdOctaveSpectra[ear][i] =
                    std::max(thirdOctaveSpectra[ear][i], singleFilterResponse[filter][ear][i]);
            }
        }
    }

    // Convert from dB to linear
    for (int ear = 0; ear < 2; ++ear) {
        std::transform(thirdOctaveSpectra[ear].begin(), thirdOctaveSpectra[ear].end(),
            thirdOctaveSpectra[ear].begin(), dBToLinear);
    }

    return thirdOctaveSpectra;
}

std::map<unsigned int, std::tuple<SpeedProfiles, ElevationProfiles, AzimuthProfiles>>
AirflowStage::parseAirflowProfiles(const AirflowPayload& payload)
{
    // Get third octave frequencies and put into a vector
    const std::vector<float> thirdOctaveCenterFreqs{ MPEGI::Config::getFloatPtr("EQCenterFreqs"),
        MPEGI::Config::getFloatPtr("EQCenterFreqs") + MPEGI::Config::getInt("EQNBands") };

    EAR_CHECK(!thirdOctaveCenterFreqs.empty(), "No Third octave frequencies available from config");

    std::map<unsigned int, std::tuple<SpeedProfiles, ElevationProfiles, AzimuthProfiles>>
        profileMap; // Will be returned

    for (const auto& [airflowSourceId, airflowSource] : payload.GetAirflowSources()) {
        std::tuple<SpeedProfiles, ElevationProfiles, AzimuthProfiles>
            airflowProfiles; // Will be inserted into map for each source

        if (airflowSource->customAirflowProfiles.empty()) {
            // Use default profiles instead
            const int defaultIndex = airflowSource->defaultProfile;
            EAR_CHECK_REL(defaultSpeedProfiles.count(defaultIndex) > 0
                    || defaultElevationProfiles.count(defaultIndex) > 0
                    || defaultAzimuthProfiles.count(defaultIndex) > 0,
                "Default profile index is out of bounds");

            airflowProfiles = { defaultSpeedProfiles.at(defaultIndex),
                defaultElevationProfiles.at(defaultIndex),
                defaultAzimuthProfiles.at(defaultIndex) };
        } else {
            // Parse and convert custom profiles
            for (const auto& profile : airflowSource->customAirflowProfiles) {
                // Convert airflow profile into third octave spectra
                std::array<std::vector<float>, 2> thirdOctaveSpectra;
                switch (profile.profileType) {
                case ProfileType::Freq_Gain_Pairs: {
                    thirdOctaveSpectra =
                        convertFreqGainPairs(payload, profile, thirdOctaveCenterFreqs);
                    break;
                }
                case ProfileType::Filter_Params: {
                    thirdOctaveSpectra =
                        convertFilterParams(payload, profile, thirdOctaveCenterFreqs);
                    break;
                }
                default: {
                    EAR_FAIL("Unsupported airflow profile type");
                    break;
                }
                }

                // TODO: Should we check/enforce that speed and elevation profiles are equal for left and right ear?
                switch (profile.profileModifies) {
                case ProfileModification::speed: {
                    std::get<0>(airflowProfiles)[profile.modificationValue] = thirdOctaveSpectra[0];
                    break;
                }
                case ProfileModification::elevation: {
                    std::get<1>(airflowProfiles)[profile.modificationValue] = thirdOctaveSpectra[0];
                    break;
                }
                case ProfileModification::azimuth: {
                    // Only azimuth has different spectra for left and right ears
                    std::get<2>(airflowProfiles)[profile.modificationValue] = {
                        thirdOctaveSpectra[0], thirdOctaveSpectra[1]
                    };
                    break;
                }
                default: {
                    EAR_FAIL("Unsupported airflow profile modification type");
                    break;
                }
                }
            }
        }
        profileMap[airflowSourceId] = airflowProfiles;
    }

    return profileMap;
}

void AirflowStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }
    this->itemStore = itemStore;
    this->streamManager = streamManager;

    // Save environment name to bsId relationship for easy access later
    for (const auto& env : state->getReverbEnvs()) {
        environmentIdToBsIdMap[env->acousticEnvironmentId] = env->acousticEnvironmentBsId;
    }
    // Add an empty optional for empty id strings. This is for the default AE.
    environmentIdToBsIdMap[""] = std::optional<int>();

    auto airflowPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_AIRFLOW);

    EAR_CHECK_REL(airflowPayloads.size() <= 1, "Too many Airflow payloads in bitstream");
    if (airflowPayloads.size() == 1) {
        bitstreamAirflowPayload =
            std::dynamic_pointer_cast<const MPEGI::AirflowPayload>(airflowPayloads[0]);

        // For each airflow source, create a renderItem and buffer
        int noiseseed = 0; // Each channel has a different seed, so audio is uncorrelated.
        float ThirdOctaveQ = 4.3290f;
        const float* centerFreqs = Config::getFloatPtr("EQCenterFreqs");

        for (const auto& [id, source] : bitstreamAirflowPayload->GetAirflowSources()) {
            RenderItem* renderItem = itemStore->createItem("airflow_source_" + std::to_string(id));
            auto buffer = streamManager->createEmptyBuffer(2);
            renderItem->setSignal(buffer);
            renderItem->type = ItemType::Airflow;
            renderItem->resetChanged();

            auto pinknoiseBuffer = streamManager->createEmptyBuffer(2);

            items.insert({ id,
                { renderItem, buffer, pinknoiseBuffer,
                    { PinkNoise(noiseseed), PinkNoise(noiseseed + 1) }, {} } });
            noiseseed += 2;
            for (int i = 0; i < Config::getInt("EQNBands"); ++i) {
                for (int ear = 0; ear < 2; ++ear) {
                    filters[id][ear].emplace_back(bq_type_bandpass,
                        centerFreqs[i] / streamManager->getFs(), ThirdOctaveQ, 0.f);
                }
            }
        }

        airflowProfilesMap = parseAirflowProfiles(*bitstreamAirflowPayload);
    }

    // Save generators in a map for easy access later
    const auto& generators = state->getAirflowGenerators();
    for (const auto& generator : generators) {
        generatorMap[generator->getBsId()] = generator;
    }

    this->frameCount = 0;
}

void AirflowStage::fillBuffersWithPinkNoise(
    std::shared_ptr<StreamBuffer> buffer, std::array<PinkNoise, 2>& generators)
{
    for (size_t c = 0; c < buffer->getFrame().channels(); ++c) {
        for (size_t offset = 0; offset < buffer->getFrame()[c].length(); ++offset) {
            *(buffer->getFrame()[c].data() + offset) = generators[c].tick();
        }
    }
}

std::pair<std::vector<float>, std::vector<float>> interpolateAndCombineProfiles(float speed,
    float elevation, float azimuth, const SpeedProfiles& speedProfiles,
    const ElevationProfiles& elevationProfiles, const AzimuthProfiles& azimuthProfiles,
    float makeupGain)
{
    const auto& speedProfile = getSpeedGainProfile(speed, speedProfiles);
    const auto& elevationProfile = getElevationGainProfile(elevation, elevationProfiles);
    const auto& azimuthProfile = getAzimuthGainProfile(azimuth, azimuthProfiles);

    std::pair<std::vector<float>, std::vector<float>> combined;
    combined.first.resize(speedProfile.size());
    combined.second.resize(speedProfile.size());

    for (int j = 0; j < speedProfile.size(); ++j) {
        combined.first[j] = speedProfile[j] * elevationProfile[j] * azimuthProfile.first[j];
        combined.second[j] = speedProfile[j] * elevationProfile[j] * azimuthProfile.second[j];
    }

    return combined;
}

} // namespace MPEGI
