/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_AIRFLOW_STAGE
#define MPEGI_AIRFLOW_STAGE

#include "AzimuthProfile.h"
#include "ElevationProfile.h"
#include "PinkNoise.h"
#include "SpeedProfile.h"

#include <EAR/AirflowGeneratorBase.h>
#include <EAR/Biquad.h>
#include <EAR/ConeAirflowGenerator.h>
#include <EAR/GlobalAirflowGenerator.h>
#include <EAR/PlaneAirflowGenerator.h>
#include <EAR/PointAirflowGenerator.h>
#include <MPEGI/Bitstream/AirflowPayload.h>
#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <optional>
#include <utility>

namespace MPEGI {

//! Creates two pink noise sources at the listener. One for each ear. The spectrum and gain of the pink noise items are adjusted based on the direction and magnitude of the air flow.
/*! 
 * Inputs: 
 * - None required
 *  
 * Outputs: 
 * - Same renderitems as the input with two additional render items for the airflow. 
 *  
 * Stages order:
 * - Required before: RoomAssignment 
 * - Required after: BinauralSpatializer
 *  
 * Config options: 
 * - none
 * 
 */

// Interpolates and combines (multiplies) frequency profiles based on air speed, elevation and azimuth. The resulting profiles are also multiplied with makeupGain.
// See SpeedProfile.h, ElevationProfile.h and AzimuthProfile.h for details about the Profiles and the extrapolation behavior.
// Returns one profile for each ear in the order (left, right).
std::pair<std::vector<float>, std::vector<float>> interpolateAndCombineProfiles(float speed,
    float elevation, float azimuth, const SpeedProfiles& speedProfiles = defaultSpeedProfiles.at(0),
    const ElevationProfiles& elevationProfiles = defaultElevationProfiles.at(0),
    const AzimuthProfiles& azimuthProfiles = defaultAzimuthProfiles.at(0),
    float makeupGain = 4.3501f);

class AirflowStage : public RenderStage
{
public:
    AirflowStage();

    //! Destructor
    ~AirflowStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

protected:
    // Parses payload to generate airflow profiles (either custom profiles, or default profiles)
    std::map<unsigned int, std::tuple<SpeedProfiles, ElevationProfiles, AzimuthProfiles>>
    parseAirflowProfiles(const AirflowPayload& payload);

    // Converts Filter params (peak-Q filters) to third octave frequencies
    std::array<std::vector<float>, 2> convertFilterParams(const MPEGI::AirflowPayload& payload,
        const MPEGI::AirflowProfileParameters& profile,
        const std::vector<float>& thirdOctaveCenterFreqs);

    // Converts frequency-gain pairs to third octave frequencies
    std::array<std::vector<float>, 2> convertFreqGainPairs(const MPEGI::AirflowPayload& payload,
        const MPEGI::AirflowProfileParameters& profile,
        const std::vector<float>& thirdOctaveCenterFreqs);

    // Fills two buffers with pink noise. Both buffers should contain two channels (for both ears) and are of type Downmix, so they are added directly to the headphones without additional processing.
    // Normally, a render item can only have 1 eq and gain set, which is applied to all channels. In this case, we want to independently set the eq and gain for each ear. Therefore we split up the audio intended for the left and right ear into two render items and therefore two buffers.
    // The first buffer is intended for the left ear: the right ear channel is all zeros.
    // The second buffer is intended for the right ear: the left ear channel is all zeros.
    void fillBuffersWithPinkNoise(
        std::shared_ptr<StreamBuffer> buffer, std::array<PinkNoise, 2>& generators);

private:
    friend class Renderer;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    RenderItemStore* itemStore;
    const StreamManager* streamManager;

    struct ProcessItem
    {
        RenderItem* item;
        std::shared_ptr<StreamBuffer> buffer; // Buffer from RenderItem
        std::shared_ptr<StreamBuffer> pinknoiseBuffer; // Buffer for generating noise into
        std::array<PinkNoise, 2> generators; // For generating noise.
    };

    // Key: airflow bitstream id. Value: renderitem, buffer, pink noise buffer (only used internally), pink noise generators, previous filter gains
    std::unordered_map<unsigned int,
        std::tuple<RenderItem*, std::shared_ptr<StreamBuffer>, std::shared_ptr<StreamBuffer>,
            std::array<PinkNoise, 2>, std::array<std::vector<float>, 2>>>
        items;

    // Key: bitstream id
    std::unordered_map<int, std::shared_ptr<EAR::AirflowGeneratorBase>> generatorMap;

    std::unordered_map<std::string, std::optional<int>> environmentIdToBsIdMap;

    int frameCount = 0;

    std::shared_ptr<const MPEGI::AirflowPayload> bitstreamAirflowPayload;

    std::map<unsigned int, std::tuple<SpeedProfiles, ElevationProfiles, AzimuthProfiles>>
        airflowProfilesMap;

    std::map<unsigned int, std::array<std::vector<Biquad>, 2>> filters;

    struct DspData : RenderStage::DspData
    {
        // Key: airflow source id, value filter gain to apply
        std::unordered_map<int, std::array<std::vector<float>, 2>> prevFilterGains;
        std::unordered_map<int, std::array<std::vector<float>, 2>> filterGains;
        std::unordered_map<int, float> generatorGain;
    };
};

} // namespace MPEGI

#endif /* MPEGI_AIRFLOW_STAGE */
