/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "AzimuthProfile.h"
#include <EAR/Exception.h>
#include <cassert>
#include <cmath>

namespace MPEGI {

std::pair<std::vector<float>, std::vector<float>> getAzimuthGainProfile(
    float targetAzimuth, const AzimuthProfiles& profiles)
{
    if (profiles.empty()) {
        throw EAR::Exception("Tried to interpolate Azimuth with empty Azimuth profiles");
    }

    // Maps targetAzimuth into 0-360 degree range
    targetAzimuth = targetAzimuth - 360.f * std::floor(targetAzimuth / 360.f);

    // Find first element >= target Azimuth
    const auto next = profiles.lower_bound(targetAzimuth);

    if (next == profiles.end()) {
        // No element was found which was >= target Azimuth. Target Azimuth must be larger than anything in the AzimuthProfiles. This is not supported, so throw an error.
        throw EAR::Exception("Target azimuth is outside the range of the supplied AzimuthProfiles. "
                             "This likely indicates the supplied AzimuthProfiles does not cover "
                             "the entire 0-360 degrees space.");
    }

    if (next->first == targetAzimuth) {
        // Target Azimuth occurs exactly in the AzimuthProfiles. Return associated spectrum
        return next->second;
    }

    if (next == profiles.begin()) {
        // Target Azimuth is lower than every profile in the Azimuth profiles. Target Azimuth must be smaller than anything in the AzimuthProfiles. This is not supported, so throw an error.
        throw EAR::Exception("Target azimuth is outside the range of the supplied "
                             "AzimuthProfiles. This likely indicates the supplied AzimuthProfiles "
                             "does not cover the entire 0-360 degrees space.");
    }

    // Do linear interpolation between the two surrounding Azimuth profiles
    const auto previous = std::prev(next);

    const float prevAzimuth = previous->first;
    const float nextAzimuth = next->first;

    const float prevWeight = (nextAzimuth - targetAzimuth) / (nextAzimuth - prevAzimuth);
    const float nextWeight = 1.f - prevWeight;

    // Give some names to variables to disambiguate the .first and .second members
    const auto& previousProfiles = previous->second; // Contains 2 profiles: one for each ear
    const auto& previousLeft = previousProfiles.first;
    const auto& previousRight = previousProfiles.second;

    const auto& nextProfiles = next->second; // Contains 2 profiles: one for each ear
    const auto& nextLeft = nextProfiles.first;
    const auto& nextRight = nextProfiles.second;

    // Ensure spectra of previous and next are the same size
    assert(previousLeft.size() == previousRight.size());
    assert(nextLeft.size() == nextRight.size());
    assert(previousLeft.size() == nextLeft.size());

    const auto spectrumSize = nextLeft.size();

    std::pair<std::vector<float>, std::vector<float>> result;
    result.first.resize(spectrumSize);
    result.second.resize(spectrumSize);

    for (size_t i = 0; i < spectrumSize; ++i) {
        result.first[i] = prevWeight * previousLeft[i] + nextWeight * nextLeft[i];
        result.second[i] = prevWeight * previousRight[i] + nextWeight * nextRight[i];
    }

    return result;
}

} // namespace MPEGI
