/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ElevationProfile.h"
#include <EAR/Exception.h>
#include <cassert>
#include <cmath>

namespace MPEGI {

std::vector<float> getElevationGainProfile(float targetElevation, const ElevationProfiles& profiles)
{
    if (profiles.empty()) {
        throw EAR::Exception("Tried to interpolate elevation with empty elevation profiles");
    }

    // Find first element >= target Elevation
    const auto next = profiles.lower_bound(targetElevation);

    if (next == profiles.end()) {
        // No element was found which was >= target Elevation. Target Elevation must be larger than everything in the ElevationProfiles. Return last element from Elevation profiles
        return profiles.rbegin()->second;
    }

    if (next->first == targetElevation) {
        // Target Elevation occurs exactly in the ElevationProfiles. Return associated spectrum
        return next->second;
    }

    if (next == profiles.begin()) {
        // Target Elevation is lower than every profile in the Elevation profiles. Return last element from Elevation profiles
        return profiles.begin()->second;
    }

    // Do linear interpolation between the two surrounding Elevation profiles
    const auto previous = std::prev(next);

    const float prevElevation = previous->first;
    const float nextElevation = next->first;

    const float prevWeight = (nextElevation - targetElevation) / (nextElevation - prevElevation);
    const float nextWeight = 1.f - prevWeight;

    // Ensure spectra of previous and next are the same size
    assert(previous->second.size() == next->second.size());

    const auto spectrumSize = next->second.size();
    std::vector<float> result(spectrumSize);

    for (size_t i = 0; i < spectrumSize; ++i) {
        result[i] = prevWeight * previous->second[i] + nextWeight * next->second[i];
    }

    return result;
}

} // namespace MPEGI
