/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SpeedProfile.h"
#include <EAR/Exception.h>
#include <cassert>
#include <cmath>

namespace MPEGI {

std::vector<float> getSpeedGainProfile(float targetSpeed, const SpeedProfiles& profiles)
{
    if (profiles.empty()) {
        throw EAR::Exception("Tried to interpolate speed with empty speed profiles");
    }

    // Find first element >= target speed
    const auto next = profiles.lower_bound(targetSpeed);

    if (next == profiles.end()) {
        // No element was found which was >= target speed. Target speed must be larger than anything in the SpeedProfiles. Return last element from speed profiles
        return profiles.rbegin()->second;
    }

    if (next->first == targetSpeed) {
        // Target speed occurs exactly in the SpeedProfiles. Return associated spectrum
        return next->second;
    }

    if (next == profiles.begin()) {
        // Target speed is lower than every profile in the speed profiles. This case shouldn't occur, so throw an error.
        throw EAR::Exception("Target speed is lower than all values in speed profiles. This is a "
                             "logical error. Either SpeedProfiles does not contain a profile for "
                             "speed=0.f, or targetSpeed is negative.");
    }

    // Do linear interpolation between the two surrounding speed profiles
    const auto previous = std::prev(next);

    const float prevSpeed = previous->first;
    const float nextSpeed = next->first;

    const float prevWeight = (nextSpeed - targetSpeed) / (nextSpeed - prevSpeed);
    const float nextWeight = 1.f - prevWeight;

    // Ensure spectra of previous and next are the same size
    assert(previous->second.size() == next->second.size());

    const auto spectrumSize = next->second.size();
    std::vector<float> result(spectrumSize);

    for (size_t i = 0; i < spectrumSize; ++i) {
        result[i] = prevWeight * previous->second[i] + nextWeight * next->second[i];
    }

    return result;
}

} // namespace MPEGI
