/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <EAR/AirflowGeneratorBase.h>
#include <EAR/ConeAirflowGenerator.h>
#include <EAR/GlobalAirflowGenerator.h>
#include <EAR/PlaneAirflowGenerator.h>
#include <EAR/PointAirflowGenerator.h>

void compareAirflow(const EAR::Vector3& first, const EAR::Vector3& second)
{
    CHECK(first[0] == doctest::Approx(second[0]).epsilon(0.001));
    CHECK(first[1] == doctest::Approx(second[1]).epsilon(0.001));
    CHECK(first[2] == doctest::Approx(second[2]).epsilon(0.001));
}

void compareAirflow(const EAR::Polar& first, const EAR::Polar& second)
{
    // Convert to carthesian coordinates to solve non-unique polar angles for certain orientations (i.e. 180 degrees rotations)
    compareAirflow(EAR::polarToCartesian(first), EAR::polarToCartesian(second));
}

TEST_CASE("Global Generator")
{
    SUBCASE("Global generator is independent of listener position")
    {
        const EAR::Position generatorPosition = { { 0, 0, 0 }, { 0, 0, 0 } };

        const auto generator = EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        // Airflow is coming from the -z direction
        const EAR::Polar expectedOrigin = { 180, 0, 3 };

        const EAR::Euler listenerOrientation(0, 0, 0);
        const std::vector<EAR::Vector3> listenerLocations = { { 0, 0, 0 }, { 5, 0, 0 },
            { 0, -3, 0 }, { 0, 0, 2.5f }, { 1.8f, 2.9f, 10.11f }, { -3.14f, 5.f, -202.f } };

        for (const auto listenerLocation : listenerLocations) {
            const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
            const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
            compareAirflow(airflowOrigin, expectedOrigin);
        }
    }

    SUBCASE("No rotation")
    {
        const EAR::Position generatorPosition = { { 0, 0, 0 }, { 0, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        // Airflow is coming from the -z direction
        const EAR::Polar expectedOrigin = { 180, 0, 3 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("-90 degree yaw")
    {
        const EAR::Position generatorPosition = { { 0, 0, 0 }, { -90, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 90, 0, 3 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
    SUBCASE("90 degree yaw")
    {
        const EAR::Position generatorPosition = { { 0, 0, 0 }, { 90, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { -90, 0, 3 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
    SUBCASE("-45 degree yaw, 180 degrees pitch")
    {
        const EAR::Position generatorPosition = { { 0, 0, 0 }, { 180, -45, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { 0, 45, 3 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
    SUBCASE("45 degree pitch")
    {
        const EAR::Position generatorPosition = { { 0, 0, 0 }, { 0, -45, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::GlobalAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { 0, 135, 3 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
}

TEST_CASE("Point Source")
{
    SUBCASE("Point source without limits is independent of orientation")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { 0, 0, 0 } };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        // Take one orientation as reference
        const EAR::Vector3 refAirflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar refAirflowOrigin = getAirflowOrigin(listenerOrientation, refAirflow);

        const std::vector<EAR::Euler> generatorOrientations = { { 0, 0, 0 }, { 90, 0, 0 },
            { -45, 0, 0 }, { 0, 90, 0 }, { 45, 180, 0 }, { 22, -80, 135 } };
        for (const auto& orientation : generatorOrientations) {
            // Test that other orientations do not produce different results than the original orientation
            generator->setOrientation(orientation);
            const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
            const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, refAirflow);
            compareAirflow(airflowOrigin, refAirflowOrigin);
        }
    }

    SUBCASE("Source offset by (0,0,-2)")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { 0, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { 0, 0, 0.3333 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Source offset by (2,0,-2)")
    {
        const EAR::Position generatorPosition = { { 2, 0, -2 }, { 0, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { -45, 0, 0.2047 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Source offset by (2,0,-2), with distanceFactor of 2")
    {
        const EAR::Position generatorPosition = { { 2, 0, -2 }, { 0, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(6.f);
        generator->setDistanceAttenuationFactor(2.f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { -45, 0, 0.0279 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Source with limits: listener outside range")
    {
        const EAR::Position generatorPosition = { { 2, 0, -2 }, { 0, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);
        generator->setLimits(30, 45);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { 0, 0, 0 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Source with limits: listener inside range")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { 0, 180, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);
        generator->setLimits(60, 60);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { 0, 0, 0.3333 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Source with limits: listener outside range 2")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { 0, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PointAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(3.f);
        generator->setLimits(90, 90);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);
        const EAR::Polar expectedOrigin = { 0, 0, 0 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
}

TEST_CASE("Cone Source")
{
    SUBCASE("There is airflow inside the cone")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0.5, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(6.f);
        generator->length = 3;
        generator->baseRadius = 1;
        generator->edgeRollOff = 0;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        CHECK(airflowOrigin[2] > 0);
    }

    SUBCASE("Listener inside the cone")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(6.f);
        generator->length = 3;
        generator->baseRadius = 1;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 0.6667 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Listener inside the cone, gen has distance attenuation factor")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(12.f);
        generator->length = 3;
        generator->baseRadius = 1;
        generator->edgeRollOff = 100;
        generator->setDistanceAttenuationFactor(2.0f);

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 0.1481 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Listener in fade-out region of the cone")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 1.5, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(6.f);
        generator->length = 2;
        generator->baseRadius = 1;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 36.87, 0, 0.2449 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Cone with 270 degree yaw")
    {
        const EAR::Position generatorPosition = { { -2, 0, 0 }, { 270, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(6.f);
        // NOTE: Length set to 3 to prevent rounding issues with listener exactly on the cone boundary
        generator->length = 3;
        generator->baseRadius = 1;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 90, 0, 0.6667 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Cone with 270 degree yaw. Listener offset by (-2, 0, 1)")
    {
        const EAR::Position generatorPosition = { { -2, 0, 0 }, { 270, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 1 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(6.f);
        generator->length = 3;
        generator->setTopRadius(0.5f);
        generator->baseRadius = 1;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 75.96376, 0, 0.5121 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Cone with top, with 270 degree yaw.")
    {
        const EAR::Position generatorPosition = { { -2, 0, 0 }, { 270, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0.5 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator =
            EAR::SceneObjectFactory<EAR::ConeAirflowGenerator>::getNewObject("", generatorPosition);
        generator->setSpeed(6.f);

        generator->length = 3;
        generator->setTopRadius(1.f);
        generator->baseRadius = 1;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 90, 0, 0.6667 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
}

TEST_CASE("Plane generator")
{
    SUBCASE("Listener directly in front of plane")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(6.f);

        generator->height = 3;
        generator->width = 3;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 0.6667 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Listener directly in front of plane, gen has distance attenuation factor of 2")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(12.f);
        generator->setDistanceAttenuationFactor(2.f);

        generator->height = 3;
        generator->width = 3;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 0.1481 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Listener directly in front of plane, less than 1m dist, gen has distance attenuation "
            "factor of 2")
    {
        const EAR::Position generatorPosition = { { 0, 0, -0.5 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(12.f);
        generator->setDistanceAttenuationFactor(2.f);

        generator->height = 3;
        generator->width = 3;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 2.37 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Rotated plane, listener in fade-out region")
    {
        const EAR::Position generatorPosition = { { 2, 0, -2 }, { 135, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 0, 0, 3 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(6.f);

        generator->height = 3;
        generator->width = 3;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { -45, 0, 0.0993 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Listener behind plane")
    {
        const EAR::Position generatorPosition = { { 2, 0, -2 }, { 180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 2, 0, -4 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(6.f);
        generator->height = 3;
        generator->width = 3;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 0 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }

    SUBCASE("Listener offset inside edge rolloff")
    {
        const EAR::Position generatorPosition = { { 0, 0, -2 }, { -180, 0, 0 } };
        const EAR::Vector3 listenerLocation = { 2.25, 0, 0 };
        const EAR::Euler listenerOrientation = { 0, 0, 0 };

        const auto generator = EAR::SceneObjectFactory<EAR::PlaneAirflowGenerator>::getNewObject(
            "", generatorPosition);
        generator->setSpeed(6.f);

        generator->height = 3;
        generator->width = 3;
        generator->edgeRollOff = 100;

        const EAR::Vector3 airflow = generator->getAirspeed(listenerLocation);
        const EAR::Polar airflowOrigin = getAirflowOrigin(listenerOrientation, airflow);

        const EAR::Polar expectedOrigin = { 0, 0, 0.3333 };

        compareAirflow(airflowOrigin, expectedOrigin);
    }
}