/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <MPEGI/Bitstream/AirflowPayload.h>
#include <MPEGI/Bitstream/AirflowPayloadHandler.h>
#include <MPEGI/Bitstream/Bitstream.h>

#include <MPEGI/Helpers/AirflowSource.h>

#define MountainCabin_bitstream "EIF_MountainCabin_forUnitTests.bit"

#include "AirflowParsing_test_data_reference.h"

namespace AirflowParsing_test {

void registerPayloadHandlers()
{
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::AirflowPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_AIRFLOW };
    });
}

TEST_CASE("Airflow_parsing_test")
{
    registerPayloadHandlers();

    // MountainCabin_Simple
    {
        MPEGI::Bitstream bitstream(MountainCabin_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_AIRFLOW);
        auto payload = std::static_pointer_cast<MPEGI::AirflowPayload const>(payloads[0]);

        auto freqGridData = payload->GetFreqGrid();

        for (int i = 0; i < freqGridData[1]->centerFreq.size(); ++i) {
            CHECK(std::abs(freqGridData[1]->centerFreq[i] - MountainCabin_FreqGrid_0[i]) < 0.001);
        }

        for (int i = 0; i < freqGridData[2]->centerFreq.size(); ++i) {
            CHECK(std::abs(freqGridData[2]->centerFreq[i] - MountainCabin_FreqGrid_1[i]) < 0.001);
        }

        auto airflowSources = payload->GetAirflowSources();

        auto payloadProfile = airflowSources.find(5)->second->customAirflowProfiles[0];
        const auto& other = MountainCabin_Gain_Profile.at(0);

        CHECK(payloadProfile.profileType == other.profileType);
        CHECK(payloadProfile.freqGridIdx == other.freqGridIdx);
        CHECK(payloadProfile.profileModifies == other.profileModifies);
        CHECK(payloadProfile.modificationValue == other.modificationValue);
        for (int i = 0; i < payloadProfile.gain.size(); ++i) {
            CHECK(std::abs(payloadProfile.gain[i].first - other.gain[i].first) < 0.001f);
            CHECK(std::abs(payloadProfile.gain[i].second - other.gain[i].second) < 0.001f);
        }

        auto payloadAzimuth = airflowSources.find(6)->second->customAirflowProfiles[2];
        const auto& otherAzi = MountainCabin_Azimuth_Profile.at(0);
        CHECK(payloadAzimuth.profileType == otherAzi.profileType);
        CHECK(payloadAzimuth.freqGridIdx == otherAzi.freqGridIdx);
        CHECK(payloadAzimuth.profileModifies == otherAzi.profileModifies);
        CHECK(std::round(payloadAzimuth.modificationValue) == otherAzi.modificationValue);
        for (int i = 0; i < payloadAzimuth.gain.size(); ++i) {
            CHECK(std::abs(payloadAzimuth.gain[i].first - otherAzi.gain[i].first) < 0.001f);
            CHECK(std::abs(payloadAzimuth.gain[i].second - otherAzi.gain[i].second) < 0.001f);
            CHECK(std::abs(payloadAzimuth.q[i].first - otherAzi.q[i].first) < 0.001f);
            CHECK(std::abs(payloadAzimuth.q[i].second - otherAzi.q[i].second) < 0.001f);
        }

        const auto refAirflow = MountainCabin_Airflow;
        const auto& payloadAirflow = airflowSources.find(6)->second;
        CHECK(payloadAirflow->id == refAirflow.id);
        CHECK(payloadAirflow->generatorID == refAirflow.generatorID);
        CHECK(payloadAirflow->generatorStringID == refAirflow.generatorStringID);
        CHECK(payloadAirflow->defaultProfile == refAirflow.defaultProfile);
        CHECK(payloadAirflow->hasValidRegions == refAirflow.hasValidRegions);
        for (int i = 0; i < payloadAirflow->activeRegionIDs.size(); ++i) {
            CHECK(payloadAirflow->activeRegionIDs[i] == refAirflow.activeRegionIDs[i]);
        }
    }

    MPEGI::Bitstream::clearHandlerList();
}
} // namespace AirflowParsing_test
