/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include "../Airflow.h"
#include "../AzimuthProfile.h"
#include "../ElevationProfile.h"
#include "../SpeedProfile.h"
#include <EAR/AirflowGeneratorBase.h>

#include <filesystem>
#include <fstream>
#include <sstream>
#include <string>

void compareSpectrum(const std::vector<float>& computed, const std::vector<float>& expected)
{
    REQUIRE(computed.size() == expected.size());

    for (size_t i = 0; i < computed.size(); ++i) {
        CHECK(computed[i] == doctest::Approx(expected[i]).epsilon(0.0001));
    }
}

void compareLocation(const EAR::Vector3& first, const EAR::Vector3& second)
{

    CHECK(first[0] == doctest::Approx(second[0]).epsilon(0.0001));
    CHECK(first[1] == doctest::Approx(second[1]).epsilon(0.0001));
    CHECK(first[2] == doctest::Approx(second[2]).epsilon(0.0001));
}
TEST_CASE("Test Speed Profile interpolation")
{
    const std::vector<float> speedsToTest = { 0.0f, 0.5f, 1.0f, 1.5f, 2.0f, 2.5f, 3.0f, 3.5f, 4.0f,
        4.5f, 5.0f, 5.5f, 6.0f, 6.5f, 7.0f, 7.5f, 8.0f, 8.5f, 9.0f, 9.5f, 10.0f };

    // Expected values generated by MATLAB reference code
    const std::vector<std::vector<float>> expectedSpeedProfiles = {
        { 0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.008084766913484f, 0.010295975573624f, 0.013729904092457f,
            0.012813492043816f, 0.012096716527168f, 0.009176297888994f, 0.005856895880953f,
            0.003695451471943f, 0.005789852548199f, 0.003293574591197f, 0.003569994142735f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.016169533826969f, 0.020591951147248f, 0.027459808184915f,
            0.025626984087632f, 0.024193433054336f, 0.018352595777989f, 0.011713791761905f,
            0.007390902943886f, 0.011579705096398f, 0.006587149182393f, 0.007139988285469f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.024254300740453f, 0.030887926720872f, 0.041189712277372f,
            0.038440476131448f, 0.036290149581503f, 0.027528893666983f, 0.017570687642858f,
            0.011086354415828f, 0.017369557644597f, 0.009880723773590f, 0.010709982428204f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.032339067653937f, 0.041183902294495f, 0.054919616369830f,
            0.051253968175265f, 0.048386866108671f, 0.036705191555978f, 0.023427583523810f,
            0.014781805887771f, 0.023159410192796f, 0.013174298364787f, 0.014279976570938f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.040423834567422f, 0.051479877868119f, 0.068649520462287f,
            0.064067460219081f, 0.060483582635839f, 0.045881489444972f, 0.029284479404763f,
            0.018477257359714f, 0.028949262740995f, 0.016467872955984f, 0.017849970713673f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.048508601480906f, 0.061775853441743f, 0.082379424554745f,
            0.076880952262897f, 0.072580299163007f, 0.055057787333966f, 0.035141375285715f,
            0.022172708831657f, 0.034739115289195f, 0.019761447547180f, 0.021419964856407f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.056593368394390f, 0.072071829015367f, 0.096109328647202f,
            0.089694444306713f, 0.084677015690174f, 0.064234085222961f, 0.040998271166668f,
            0.025868160303600f, 0.040528967837394f, 0.023055022138377f, 0.024989958999142f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.070139084462248f, 0.090507477123967f, 0.118089572448238f,
            0.108815186848135f, 0.105403404013957f, 0.087391039811475f, 0.058108580886576f,
            0.037451487848548f, 0.045486527031371f, 0.026063580829527f, 0.027093166788576f,
            0.005997082297549f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.085050037818698f, 0.110978043366312f, 0.142132401176417f,
            0.129512742013958f, 0.128287210286893f, 0.114043158574869f, 0.078032244066223f,
            0.051006784411748f, 0.050236012886793f, 0.029000885545665f, 0.028829677989686f,
            0.013493435169485f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.099960991175149f, 0.131448609608656f, 0.166175229904597f,
            0.150210297179782f, 0.151171016559830f, 0.140695277338263f, 0.097955907245870f,
            0.064562080974947f, 0.054985498742215f, 0.031938190261804f, 0.030566189190796f,
            0.020989788041421f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.117721215930347f, 0.140863621347843f, 0.172389523480875f,
            0.159881733764572f, 0.170702563393792f, 0.177041411271448f, 0.141522357020686f,
            0.096945134951144f, 0.069762142663818f, 0.039726387699420f, 0.034110083223469f,
            0.025284814427649f, 0.005661610768983f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.137380954951378f, 0.142908263418260f, 0.166718126955885f,
            0.162202424628673f, 0.187999270601772f, 0.219850221984494f, 0.200850664525614f,
            0.141880027202671f, 0.091223558629541f, 0.050748513618020f, 0.038858899143853f,
            0.027445623156739f, 0.015097628717287f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.172770112737043f, 0.153673418197954f, 0.171355103158551f,
            0.165727162555532f, 0.204778364217990f, 0.259274811843423f, 0.260040963401617f,
            0.197748061209429f, 0.126017426366038f, 0.067809806427335f, 0.045955144639705f,
            0.031538413772820f, 0.024794009216137f, 0.002487818849098f, 0.002045590763071f,
            0.001681971969735f, 0.001823134686625f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.271076945581242f, 0.199320623814761f, 0.217225570271842f,
            0.174068088733425f, 0.219487003467161f, 0.285162518285884f, 0.318679227761919f,
            0.297348662237114f, 0.214141101185630f, 0.109027766799506f, 0.062441108437433f,
            0.043359131936864f, 0.035531839917170f, 0.014926913094590f, 0.012273544578425f,
            0.010091831818408f, 0.010938808119748f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.369383778425441f, 0.244967829431569f, 0.263096037385133f,
            0.182409014911317f, 0.234195642716331f, 0.311050224728346f, 0.377317492122220f,
            0.396949263264798f, 0.302264776005221f, 0.150245727171677f, 0.078927072235161f,
            0.055179850100907f, 0.046269670618203f, 0.027366007340081f, 0.022501498393779f,
            0.018501691667081f, 0.020054481552871f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.477073614212011f, 0.370251375728073f, 0.314215810723026f,
            0.239635424748786f, 0.292330591819263f, 0.381772305341680f, 0.498077900067154f,
            0.493130998658442f, 0.430969858195250f, 0.243025385863184f, 0.135703179832552f,
            0.077979403403711f, 0.057274631483355f, 0.035674259424836f, 0.029332897604376f,
            0.023293139846765f, 0.023872148055096f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.587109200734174f, 0.515444007194500f, 0.366647910617071f,
            0.309083205501150f, 0.361322118385635f, 0.463702979497733f, 0.634368843908246f,
            0.588458017643575f, 0.569820292227889f, 0.348695469134525f, 0.202551823379858f,
            0.103523665491205f, 0.068346374889537f, 0.042949800969406f, 0.035315158163784f,
            0.027179985109202f, 0.026365312824597f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 0.752236109033908f, 0.690974229869086f, 0.468834901548113f,
            0.392094250196060f, 0.419763887775116f, 0.514558142544963f, 0.698139736622779f,
            0.675437080548312f, 0.693472747526561f, 0.483124927357406f, 0.299492277463976f,
            0.157492236404053f, 0.098328188383980f, 0.062040892630304f, 0.051012667961423f,
            0.037521206971536f, 0.032229981971006f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 1.000000000000000f, 0.912010839355909f, 0.645654229034655f,
            0.495450190804790f, 0.462381021399260f, 0.518800038928961f, 0.653130552647472f,
            0.749894209332456f, 0.794328234724281f, 0.660693448007595f, 0.441570447353312f,
            0.254097270554930f, 0.156675107010815f, 0.098855309465694f, 0.081283051616410f,
            0.057543993733716f, 0.043151907682776f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
        { 0.000000000000000f, 1.000000000000000f, 0.912010839355909f, 0.645654229034655f,
            0.495450190804790f, 0.462381021399260f, 0.518800038928961f, 0.653130552647472f,
            0.749894209332456f, 0.794328234724281f, 0.660693448007595f, 0.441570447353312f,
            0.254097270554930f, 0.156675107010815f, 0.098855309465694f, 0.081283051616410f,
            0.057543993733716f, 0.043151907682776f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f, 0.000000000000000f,
            0.000000000000000f, 0.000000000000000f, 0.000000000000000f },
    };

    REQUIRE(speedsToTest.size() == expectedSpeedProfiles.size());

    for (size_t i = 0; i < speedsToTest.size(); ++i) {
        const float targetSpeed = speedsToTest[i];
        const auto computed = MPEGI::getSpeedGainProfile(targetSpeed);
        const auto expected = expectedSpeedProfiles[i];

        compareSpectrum(computed, expected);
    }
}

TEST_CASE("Test Elevation Profile interpolation")
{
    const std::vector<float> elevationsToTest = { -25.f, -20.f, -15.f, -10.f, -5.f, 0.f, 5.f, 10.f,
        15.f, 20.f, 25.f };

    const std::vector<std::vector<float>> expectedElevationProfiles = {
        { { 0.921766342602824f, 0.921766342602824f, 0.727344352617479f, 0.631476319213482f,
              0.763218137076511f, 0.724041143935932f, 0.685584218609851f, 0.710342659213271f,
              0.566617856976535f, 0.483059981935729f, 0.436422330906592f, 0.452727508875646f,
              0.661101308317979f, 0.823500159254814f, 0.747914062773237f, 0.628175226550336f,
              0.579897228673616f, 0.541648444997606f, 0.541648444997606f, 0.541648444997606f,
              0.541648444997606f, 0.541648444997606f, 0.541648444997606f, 0.541648444997606f,
              0.541648444997606f, 0.541648444997606f, 0.541648444997606f, 0.541648444997606f,
              0.541648444997606f, 0.541648444997606f, 0.541648444997606f },
            { 0.926376387939799f, 0.926376387939799f, 0.783347425181266f, 0.712850475795363f,
                0.881609068538256f, 0.862020571967966f, 0.842792109304925f, 0.855171329606635f,
                0.696274034191901f, 0.588597018791079f, 0.510231676488987f, 0.505562431835000f,
                0.720414853725071f, 0.885897461292877f, 0.840523320713004f, 0.721855815664793f,
                0.662694927591685f, 0.626794188124034f, 0.626794188124034f, 0.626794188124034f,
                0.626794188124034f, 0.626794188124034f, 0.626794188124034f, 0.626794188124034f,
                0.626794188124034f, 0.626794188124034f, 0.626794188124034f, 0.626794188124034f,
                0.626794188124034f, 0.626794188124034f, 0.626794188124034f },
            { 0.930986433276774f, 0.930986433276774f, 0.839350497745053f, 0.794224632377244f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                0.825930211407267f, 0.694134055646429f, 0.584041022071381f, 0.558397354794353f,
                0.779728399132162f, 0.948294763330940f, 0.933132578652771f, 0.815536404779250f,
                0.745492626509754f, 0.711939931250462f, 0.711939931250462f, 0.711939931250462f,
                0.711939931250462f, 0.711939931250462f, 0.711939931250462f, 0.711939931250462f,
                0.711939931250462f, 0.711939931250462f, 0.711939931250462f, 0.711939931250462f,
                0.711939931250462f, 0.711939931250462f, 0.711939931250462f },
            { 0.953990955517849f, 0.953990955517849f, 0.892900331830035f, 0.862816421584829f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                0.883953474271511f, 0.796089370430953f, 0.722694014714254f, 0.705598236529569f,
                0.853152266088108f, 0.965529842220627f, 0.955421719101847f, 0.877024269852833f,
                0.830328417673169f, 0.807959954166975f, 0.807959954166975f, 0.807959954166975f,
                0.807959954166975f, 0.807959954166975f, 0.807959954166975f, 0.807959954166975f,
                0.807959954166975f, 0.807959954166975f, 0.807959954166975f, 0.807959954166975f,
                0.807959954166975f, 0.807959954166975f, 0.807959954166975f },
            { 0.976995477758924f, 0.976995477758924f, 0.946450165915018f, 0.931408210792414f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                0.941976737135756f, 0.898044685215476f, 0.861347007357127f, 0.852799118264784f,
                0.926576133044054f, 0.982764921110313f, 0.977710859550924f, 0.938512134926417f,
                0.915164208836585f, 0.903979977083487f, 0.903979977083487f, 0.903979977083487f,
                0.903979977083487f, 0.903979977083487f, 0.903979977083487f, 0.903979977083487f,
                0.903979977083487f, 0.903979977083487f, 0.903979977083487f, 0.903979977083487f,
                0.903979977083487f, 0.903979977083487f, 0.903979977083487f },
            { 1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f, 1.000000000000000f,
                1.000000000000000f, 1.000000000000000f, 1.000000000000000f },
            { 0.974504573667948f, 0.974504573667948f, 0.920659871289413f, 0.893821560437001f,
                0.912885969675959f, 0.907558511483842f, 0.911191007916212f, 0.913453563788770f,
                0.902889621465195f, 0.878172377207283f, 0.851730731757656f, 0.825822330215837f,
                0.865423513179221f, 0.846273186416441f, 0.799657619496950f, 0.776269579846000f,
                0.766510695629087f, 0.767318594508808f, 0.767318594508808f, 0.767318594508808f,
                0.767318594508808f, 0.767318594508808f, 0.767318594508808f, 0.767318594508808f,
                0.767318594508808f, 0.767318594508808f, 0.767318594508808f, 0.767318594508808f,
                0.767318594508808f, 0.767318594508808f, 0.767318594508808f },
            { 0.949009147335895f, 0.949009147335895f, 0.841319742578825f, 0.787643120874001f,
                0.825771939351918f, 0.815117022967683f, 0.822382015832423f, 0.826907127577540f,
                0.805779242930389f, 0.756344754414566f, 0.703461463515313f, 0.651644660431673f,
                0.730847026358441f, 0.692546372832881f, 0.599315238993900f, 0.552539159691999f,
                0.533021391258173f, 0.534637189017616f, 0.534637189017616f, 0.534637189017616f,
                0.534637189017616f, 0.534637189017616f, 0.534637189017616f, 0.534637189017616f,
                0.534637189017616f, 0.534637189017616f, 0.534637189017616f, 0.534637189017616f,
                0.534637189017616f, 0.534637189017616f, 0.534637189017616f },
            { 0.923513721003843f, 0.923513721003843f, 0.761979613868237f, 0.681464681311002f,
                0.738657909027877f, 0.722675534451525f, 0.733573023748635f, 0.740360691366310f,
                0.708668864395584f, 0.634517131621849f, 0.555192195272969f, 0.477466990647510f,
                0.596270539537662f, 0.538819559249321f, 0.398972858490850f, 0.328808739537999f,
                0.299532086887260f, 0.301955783526424f, 0.301955783526424f, 0.301955783526424f,
                0.301955783526424f, 0.301955783526424f, 0.301955783526424f, 0.301955783526424f,
                0.301955783526424f, 0.301955783526424f, 0.301955783526424f, 0.301955783526424f,
                0.301955783526424f, 0.301955783526424f, 0.301955783526424f },
            { 0.949009147335895f, 0.949009147335895f, 0.841319742578825f, 0.698271738411507f,
                0.762102062925221f, 0.775090466566240f, 0.780336349780769f, 0.718387144446756f,
                0.670745628047801f, 0.578007994664372f, 0.506530557126370f, 0.452842242583902f,
                0.551266161692991f, 0.513319944212774f, 0.403646448722057f, 0.339521469689905f,
                0.332220468606964f, 0.329632054184705f, 0.329632054184705f, 0.329632054184705f,
                0.329632054184705f, 0.329632054184705f, 0.329632054184705f, 0.329632054184705f,
                0.329632054184705f, 0.329632054184705f, 0.329632054184705f, 0.329632054184705f,
                0.329632054184705f, 0.329632054184705f, 0.329632054184705f },
            { 0.974504573667948f, 0.974504573667948f, 0.920659871289413f, 0.715078795512012f,
                0.785546216822564f, 0.827505398680955f, 0.827099675812903f, 0.696413597527203f,
                0.632822391700018f, 0.521498857706895f, 0.457868918979770f, 0.428217494520294f,
                0.506261783848319f, 0.487820329176228f, 0.408320038953264f, 0.350234199841811f,
                0.364908850326668f, 0.357308324842986f, 0.357308324842986f, 0.357308324842986f,
                0.357308324842986f, 0.357308324842986f, 0.357308324842986f, 0.357308324842986f,
                0.357308324842986f, 0.357308324842986f, 0.357308324842986f, 0.357308324842986f,
                0.357308324842986f, 0.357308324842986f, 0.357308324842986f } }
    };

    REQUIRE(elevationsToTest.size() == expectedElevationProfiles.size());

    for (size_t i = 0; i < elevationsToTest.size(); ++i) {
        const float targetElevation = elevationsToTest[i];
        const auto computed = MPEGI::getElevationGainProfile(targetElevation);
        const auto expected = expectedElevationProfiles[i];

        compareSpectrum(computed, expected);
    }
}

TEST_CASE("Test Azimuth Profile interpolation")
{
    const std::vector<float> azimuthsToTest = { 0.f, 5.f, 15.f, 30.f, 60.f, 250.f, 340.f, 360.f };

    const std::vector<std::vector<std::vector<float>>> expectedAzimuthProfiles = { {
        { { { 0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f },
            { 0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f } } },
        { { { 0.294687530364207f, 0.294687530364207f, 0.322740440721184f, 0.289481034526964f,
                0.347287703814887f, 0.300797856744247f, 0.281588161566107f, 0.263403235439501f,
                0.243189797425611f, 0.239137851001778f, 0.254205709820075f, 0.278232141860921f,
                0.273761776431157f, 0.257993771157568f, 0.268452404865398f, 0.270606473315375f,
                0.258531277580158f, 0.259776252393670f, 0.259776252393670f, 0.259776252393670f,
                0.259776252393670f, 0.259776252393670f, 0.259776252393670f, 0.259776252393670f,
                0.259776252393670f, 0.259776252393670f, 0.259776252393670f, 0.259776252393670f,
                0.259776252393670f, 0.259776252393670f, 0.259776252393670f },
            { 0.261044373017057f, 0.261044373017057f, 0.284216851983735f, 0.261284757213917f,
                0.287362445579588f, 0.274905231601498f, 0.243727483631396f, 0.225128213269708f,
                0.215907721596028f, 0.214572546848716f, 0.220165004274018f, 0.236116249249723f,
                0.236944497573300f, 0.229093566051078f, 0.233047097743374f, 0.232009644023375f,
                0.228451054265429f, 0.229310214231886f, 0.229310214231886f, 0.229310214231886f,
                0.229310214231886f, 0.229310214231886f, 0.229310214231886f, 0.229310214231886f,
                0.229310214231886f, 0.229310214231886f, 0.229310214231886f, 0.229310214231886f,
                0.229310214231886f, 0.229310214231886f, 0.229310214231886f } } },
        { { { 0.424303849239818f, 0.424303849239818f, 0.508462580310750f, 0.408684361728089f,
                0.582104369591858f, 0.442634828379939f, 0.385005742845518f, 0.330450964465699f,
                0.269810650424030f, 0.257654811152530f, 0.302858387607423f, 0.374937683729959f,
                0.361526587440669f, 0.314222571619901f, 0.345598472743391f, 0.352060678093323f,
                0.315835090887671f, 0.319570015328206f, 0.319570015328206f, 0.319570015328206f,
                0.319570015328206f, 0.319570015328206f, 0.319570015328206f, 0.319570015328206f,
                0.319570015328206f, 0.319570015328206f, 0.319570015328206f, 0.319570015328206f,
                0.319570015328206f, 0.319570015328206f, 0.319570015328206f },
            { 0.323374377198369f, 0.323374377198369f, 0.392891814098403f, 0.324095529788948f,
                0.402328594885962f, 0.364956952951690f, 0.271423709041384f, 0.215625897956320f,
                0.187964422935282f, 0.183958898693345f, 0.200736270969251f, 0.248590005896366f,
                0.251074750867096f, 0.227521956300432f, 0.239382551377320f, 0.236270190217322f,
                0.225594420943483f, 0.228171900842854f, 0.228171900842854f, 0.228171900842854f,
                0.228171900842854f, 0.228171900842854f, 0.228171900842854f, 0.228171900842854f,
                0.228171900842854f, 0.228171900842854f, 0.228171900842854f, 0.228171900842854f,
                0.228171900842854f, 0.228171900842854f, 0.228171900842854f } } },
        { { { 0.618728327553234f, 0.618728327553234f, 0.787045789695099f, 0.587489352529776f,
                0.934329368257315f, 0.655390285833477f, 0.540132114764635f, 0.431022558004996f,
                0.309741929921658f, 0.285430251378658f, 0.375837404288445f, 0.519995996533516f,
                0.493173803954936f, 0.398565772313401f, 0.461317574560380f, 0.474241985260245f,
                0.401790810848940f, 0.409260659730011f, 0.409260659730011f, 0.409260659730011f,
                0.409260659730011f, 0.409260659730011f, 0.409260659730011f, 0.409260659730011f,
                0.409260659730011f, 0.409260659730011f, 0.409260659730011f, 0.409260659730011f,
                0.409260659730011f, 0.409260659730011f, 0.409260659730011f },
            { 0.416869383470336f, 0.416869383470336f, 0.555904257270404f, 0.418311688651495f,
                0.574777818845522f, 0.500034534976979f, 0.312968047156367f, 0.201372424986239f,
                0.146049474944163f, 0.138038426460288f, 0.171593171012100f, 0.267300640866331f,
                0.272270130807791f, 0.225164541674462f, 0.248885731828239f, 0.242661009508242f,
                0.221309470960564f, 0.226464430759306f, 0.226464430759306f, 0.226464430759306f,
                0.226464430759306f, 0.226464430759306f, 0.226464430759306f, 0.226464430759306f,
                0.226464430759306f, 0.226464430759306f, 0.226464430759306f, 0.226464430759306f,
                0.226464430759306f, 0.226464430759306f, 0.226464430759306f } } },
        { { { 0.701455298419971f, 0.701455298419971f, 1.000000000000000f, 0.836565595755810f,
                0.970509967245491f, 0.729457510254568f, 0.382384253658113f, 0.242381796455463f,
                0.178443198854037f, 0.194984459975804f, 0.216023003477475f, 0.281190083039894f,
                0.247742205763329f, 0.222843514927031f, 0.239056193534463f, 0.251767692775886f,
                0.223100220748196f, 0.233345806228100f, 0.233345806228100f, 0.233345806228100f,
                0.233345806228100f, 0.233345806228100f, 0.233345806228100f, 0.233345806228100f,
                0.233345806228100f, 0.233345806228100f, 0.233345806228100f, 0.233345806228100f,
                0.233345806228100f, 0.233345806228100f, 0.233345806228100f },
            { 0.368553041852479f, 0.368553041852479f, 0.698232404077171f, 0.542625251329014f,
                0.678422120498261f, 0.468813382145265f, 0.276375791917037f, 0.204173794466953f,
                0.142724983016489f, 0.135207256319428f, 0.149623565609444f, 0.183442516711773f,
                0.148593564228701f, 0.127350308101666f, 0.195659076793494f, 0.242661009508242f,
                0.279254384123734f, 0.269153480392692f, 0.269153480392692f, 0.269153480392692f,
                0.269153480392692f, 0.269153480392692f, 0.269153480392692f, 0.269153480392692f,
                0.269153480392692f, 0.269153480392692f, 0.269153480392692f, 0.269153480392692f,
                0.269153480392692f, 0.269153480392692f, 0.269153480392692f } } },
        { { { 0.406236106949981f, 0.406236106949981f, 0.577709219183740f, 0.452210611568191f,
                0.521033668664534f, 0.299956158914661f, 0.175708601417794f, 0.104648498228823f,
                0.093983840920633f, 0.100307923897053f, 0.167577017559822f, 0.220754311303833f,
                0.276454792509416f, 0.256513105962607f, 0.295477862463610f, 0.250812552755036f,
                0.262294865318029f, 0.252807403941057f, 0.252807403941057f, 0.252807403941057f,
                0.252807403941057f, 0.252807403941057f, 0.252807403941057f, 0.252807403941057f,
                0.252807403941057f, 0.252807403941057f, 0.252807403941057f, 0.252807403941057f,
                0.252807403941057f, 0.252807403941057f, 0.252807403941057f },
            { 0.172735218325844f, 0.172735218325844f, 0.279021838588788f, 0.196790251225053f,
                0.238338790869620f, 0.133517233148656f, 0.073348837709325f, 0.055049959011579f,
                0.038023998452372f, 0.052122318500831f, 0.071364476274881f, 0.124591702169851f,
                0.139055861428289f, 0.130792599144843f, 0.182409220364200f, 0.216759729465389f,
                0.246809591024992f, 0.247271688813355f, 0.247271688813355f, 0.247271688813355f,
                0.247271688813355f, 0.247271688813355f, 0.247271688813355f, 0.247271688813355f,
                0.247271688813355f, 0.247271688813355f, 0.247271688813355f, 0.247271688813355f,
                0.247271688813355f, 0.247271688813355f, 0.247271688813355f } } },
        { { { 0.354539379289024f, 0.354539379289024f, 0.447229295155737f, 0.355500916076464f,
                0.459811669539148f, 0.409982813626786f, 0.285271821746379f, 0.210874740299626f,
                0.173992773604909f, 0.168652074615659f, 0.191021904316867f, 0.254826884219688f,
                0.258139877513995f, 0.226736151425109f, 0.242550278194293f, 0.238400463314295f,
                0.224166104282510f, 0.227602744148338f, 0.227602744148338f, 0.227602744148338f,
                0.227602744148338f, 0.227602744148338f, 0.227602744148338f, 0.227602744148338f,
                0.227602744148338f, 0.227602744148338f, 0.227602744148338f, 0.227602744148338f,
                0.227602744148338f, 0.227602744148338f, 0.227602744148338f },
            { 0.489112008677623f, 0.489112008677623f, 0.601323650105533f, 0.468286025328651f,
                0.699512702480344f, 0.513553314197785f, 0.436714533485224f, 0.363974828978798f,
                0.283121076923239f, 0.266913291227906f, 0.327184726501097f, 0.423290454664478f,
                0.405408992945424f, 0.342336971851068f, 0.384171506682387f, 0.392787780482297f,
                0.344486997541427f, 0.349466896795475f, 0.349466896795475f, 0.349466896795475f,
                0.349466896795475f, 0.349466896795475f, 0.349466896795475f, 0.349466896795475f,
                0.349466896795475f, 0.349466896795475f, 0.349466896795475f, 0.349466896795475f,
                0.349466896795475f, 0.349466896795475f, 0.349466896795475f } } },
        { { { 0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f },
            { 0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f, 0.229879370926401f,
                0.229879370926401f, 0.229879370926401f, 0.229879370926401f } } },
    } };

    REQUIRE(azimuthsToTest.size() == expectedAzimuthProfiles.size());

    for (size_t i = 0; i < azimuthsToTest.size(); ++i) {
        const float targetAzimuth = azimuthsToTest[i];
        const auto computed =
            MPEGI::getAzimuthGainProfile(targetAzimuth, MPEGI::defaultAzimuthProfiles.at(2));
        const auto expected = expectedAzimuthProfiles[i];

        compareSpectrum(computed.first, expected[0]);
        compareSpectrum(computed.second, expected[1]);
    }
}

TEST_CASE("Test Speed, Elevation and Azimuth Profile interpolation")
{
    // Parse csv file to get testing data. NOTE: Minimal validation of the file format is performed.
    std::vector<float> speed, azimuth, elevation;
    std::vector<std::pair<std::vector<float>, std::vector<float>>> expectedSpectrum;

    std::ifstream f;
    f.open("./airflowTestData.csv");

    std::string currentLine;
    std::vector<float> leftEarSpectrum;

    // Skip first line: it contains header and frequency information which are not needed
    std::getline(f, currentLine);

    while (std::getline(f, currentLine)) {
        std::stringstream lineStream(currentLine);
        std::string speedString, azimuthString, elevationString, earString;

        std::getline(lineStream, speedString, ',');
        std::getline(lineStream, azimuthString, ',');
        std::getline(lineStream, elevationString, ',');
        std::getline(lineStream, earString, ',');

        if (earString == "left") {
            speed.push_back(std::stof(speedString));
            azimuth.push_back(std::stof(azimuthString));
            elevation.push_back(std::stof(elevationString));
        } else {
            // Ensure file contains alternating left and right ear data with the same speed, azimuth and elevation. Parser currently cannot handle anything else.
            CHECK(speed.back() == std::stof(speedString));
            CHECK(azimuth.back() == std::stof(azimuthString));
            CHECK(elevation.back() == std::stof(elevationString));
        }

        std::vector<float> spectrum;
        std::string spectrumString;
        while (std::getline(lineStream, spectrumString, ',')) {
            spectrum.push_back(std::stof(spectrumString));
        }

        if (earString == "left") {
            // File contains alternating left and right ear data. Save data untill right ear is parsed.
            leftEarSpectrum = spectrum;
        } else {
            // File contains alternating left and right ear data. Use previously parsed left ear data.
            expectedSpectrum.emplace_back(std::move(leftEarSpectrum), std::move(spectrum));
        }
    }

    CHECK(speed.size() == azimuth.size());
    CHECK(speed.size() == elevation.size());
    CHECK(speed.size() == expectedSpectrum.size());

    // End of file parsing. Start validation.
    for (int testRow = 0; testRow < speed.size(); ++testRow) {
        // Multiply profiles together

        const auto combined = MPEGI::interpolateAndCombineProfiles(speed[testRow],
            elevation[testRow], azimuth[testRow], MPEGI::defaultSpeedProfiles.at(2),
            MPEGI::defaultElevationProfiles.at(2), MPEGI::defaultAzimuthProfiles.at(2));

        CAPTURE(speed[testRow]);
        CAPTURE(elevation[testRow]);
        CAPTURE(azimuth[testRow]);

        compareSpectrum(combined.first, expectedSpectrum[testRow].first);
        compareSpectrum(combined.second, expectedSpectrum[testRow].second);
    }
}

TEST_CASE("Get relative position")
{
    SUBCASE("Reference at origin, offset with no rotation")
    {
        const EAR::Position reference({ 0, 0, 0 }, { 0, 0, 0 });
        const EAR::Vector3 target({ 3, 0, 0 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected = target;

        compareLocation(relative, expected);
    }

    SUBCASE("Reference and target not at origin, offset with no rotation")
    {
        const EAR::Position reference({ 0, 2, 0 }, { 0, 0, 0 });
        const EAR::Vector3 target({ 3, 0, 0 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected({ 3, -2, 0 });

        compareLocation(relative, expected);
    }

    SUBCASE("Reference at origin, rotated 180 degrees around yaw axis")
    {
        const EAR::Position reference({ 0, 0, 0 }, { 180, 0, 0 });
        const EAR::Vector3 target({ 3, 0, 0 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected = { -3, 0, 0 };

        compareLocation(relative, expected);
    }

    SUBCASE("Reference at origin, rotated 45 degrees around pitch axis")
    {
        const EAR::Position reference({ 0, 0, 0 }, { 0, 45, 0 });
        const EAR::Vector3 target({ 0, 0, 2 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected = { 0, std::sqrt(2.f), std::sqrt(2.f) };

        compareLocation(relative, expected);
    }

    SUBCASE("Reference at origin, rotated 90 degrees around roll axis")
    {
        const EAR::Position reference({ 0, 0, 0 }, { 0, 0, 90 });
        const EAR::Vector3 target({ 3, 0, 0 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected = { 0, -3, 0 };

        compareLocation(relative, expected);
    }

    SUBCASE("Reference at origin, rotated 45 degrees around pitch and roll axes")
    {
        const EAR::Position reference({ 0, 0, 0 }, { 0, 45, 45 });
        const EAR::Vector3 target({ 0, 2, 0 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected = { 1, 1, -std::sqrt(2.f) };

        compareLocation(relative, expected);
    }

    SUBCASE("Reference and target at same position")
    {
        const EAR::Position reference({ 0, 2, 0 }, { 45, 30, -20 });
        const EAR::Vector3 target({ 0, 2, 0 });

        const EAR::Vector3 relative = EAR::getRelativePosition(reference, target);

        const EAR::Vector3 expected = { 0, 0, 0 };

        compareLocation(relative, expected);
    }
}

class AirflowStageTester : public MPEGI::AirflowStage
{
public:
    // Make protected method public for testing
    std::array<std::vector<float>, 2> testConvertFilterParams(const MPEGI::AirflowPayload& payload,
        const MPEGI::AirflowProfileParameters& profile,
        const std::vector<float>& thirdOctaveCenterFreqs)
    {
        return MPEGI::AirflowStage::convertFilterParams(payload, profile, thirdOctaveCenterFreqs);
    }
};

class AirflowPayloadDummy : public MPEGI::AirflowPayload
{
public:
    // Make protected constructor of base public
    AirflowPayloadDummy() = default;

    // Make protected members of base public
    std::vector<std::shared_ptr<MPEGI::FreqGrid>>& getFreqGridData() { return freqGridData; }
    std::unordered_map<unsigned int, std::shared_ptr<const MPEGI::AirflowSource>>
    getAirflowSources()
    {
        return airflowSources;
    }
};

TEST_CASE("Convert peak Q filters to third octave frequencies")
{
    const std::vector<float> cf = { 50.f, 500.f, 4000.f };
    const std::vector<std::pair<float, float>> Q = { { 2, 2 }, { 2, 2 }, { 2, 2 } };
    const std::vector<std::pair<float, float>> gain = { { -5, -5 }, { 0, 0 }, { -15, -15 } };

    const std::vector<float> expectedThirdOctaveProfile = { 0.0769f, 0.1413f, 0.2639f, 0.4508f,
        0.5623f, 0.3870f, 0.0769f, 0.0291f, 0.0447f, 0.0777f, 0.1368f, 0.2512f, 0.4693f, 0.8017f,
        1.0000f, 0.6883f, 0.1368f, 0.0079f, 0.0131f, 0.0243f, 0.0447f, 0.0818f, 0.1386f, 0.1778f,
        0.1259f, 0.0286f, 0.0007f, 0.0000f, 0.0000f, 0.0000f, 0.0000f };

    // Get third octave frequencies and put into a vector
    const std::vector<float> thirdOctaveCenterFreqs{ MPEGI::Config::getFloatPtr("EQCenterFreqs"),
        MPEGI::Config::getFloatPtr("EQCenterFreqs") + MPEGI::Config::getInt("EQNBands") };

    // Create dummy payload with the above information
    AirflowPayloadDummy payload;
    const auto freqGrid = std::make_shared<MPEGI::FreqGrid>(3, cf);

    payload.getFreqGridData().push_back(freqGrid);

    MPEGI::AirflowProfileParameters customAirflowProfiles{
        MPEGI::AirflowProfileParameters::ProfileType::Filter_Params, 0,
        MPEGI::AirflowProfileParameters::ProfileModification::speed, 0.f, gain, Q
    };

    AirflowStageTester tester;
    std::array<std::vector<float>, 2> thirdOctaveProfile =
        tester.testConvertFilterParams(payload, customAirflowProfiles, thirdOctaveCenterFreqs);

    // Airflow stage has a (potentially different) spectrum for left and right ear. In this case, both have the same parameters, so both should be equal to the reference.
    compareSpectrum(thirdOctaveProfile[0], expectedThirdOctaveProfile);
    compareSpectrum(thirdOctaveProfile[1], expectedThirdOctaveProfile);
};