#include "Consolidation.h"

// many packet-related classes were renamed to "payload"
//#include <MPEGI/Bitstream/EIFPacket.h>
#include <MPEGI/Bitstream/EIFPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Stream/FileStreamManager.h>

#include <EAR/Exception.h>
#include <EAR/FIRFilter.h>
#include <EAR/PchipMinphaseFilterDesigner.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <cfloat>
#include <cstring>
#include <numeric>

#define OBJCLUST_MAX_NUM_OBJECTS (256)

/* weighting factors for perceptual coordinate system (PCS) */
/* distance in perceptual coordinates approximates number of just noticeable differences (JND) between two positions*/
#define PCS_WEIGHT_LEFTRIGHT (20.0f)
#define PCS_WEIGHT_FRONTBACK (20.0f)
#define PCS_WEIGHT_UPDOWN (20.0f)
#define PCS_WEIGHT_EQ (1.0f)
#define PCS_WEIGHT_ER (6.0f)
#define PCS_WEIGHT_DISTANCE (0.0f)
#define PCS_REF_DISTANCE (0.2f)

#define DISTANCE_INVALID (-2.0f)

namespace {

inline EAR::Vector3 dot_mul(const EAR::Vector3& x, const EAR::Vector3& y)
{
    return { x[0] * y[0], x[1] * y[1], x[2] * y[2] };
}

inline EAR::Vector3 dot_div(const EAR::Vector3& x, const EAR::Vector3& y)
{
    return { x[0] / y[0], x[1] / y[1], x[2] / y[2] };
}

// find maximum in an array, filtering out deactivated rows/cols
void findMaxFilt(float array[OBJCLUST_MAX_NUM_OBJECTS][OBJCLUST_MAX_NUM_OBJECTS], int sizeIni,
    int sizeInj, int* iOut, int* jOut, float* maxOut, int filti[OBJCLUST_MAX_NUM_OBJECTS],
    int filtj[OBJCLUST_MAX_NUM_OBJECTS])
{
    int iMax = 0, jMax = 0;
    float maxVal = -1e20f;
    for (int i = 0; i < sizeIni; i++) {
        if (filti[i])
            continue;
        for (int j = 0; j < sizeInj; j++) {
            if (filtj[j])
                continue;
            float tmpVal = array[i][j];
            if (tmpVal > maxVal) {
                maxVal = tmpVal;
                iMax = i;
                jMax = j;
            }
        }
    }
    *iOut = iMax;
    *jOut = jMax;
    *maxOut = maxVal;
}

} // namespace

namespace MPEGI {

std::shared_ptr<RenderStage> makeConsolidationStage()
{
    return std::make_shared<ConsolidationStage>();
}

ConsolidationStage::ConsolidationStage()
    : itemStore(nullptr)
{
    Config::setDef("Consolidation_NumClusters", EAR::Variant(OBJCLUST_MAX_NUM_OBJECTS));
}

ConsolidationStage::~ConsolidationStage() {}

std::unique_ptr<RenderStage::DspData> ConsolidationStage::update(RenderList& renderList)
{
    int consolidationThreshold = Config::getInt("Consolidation_Threshold");
    bool useLowComplexityER = Config::getBool("useLowComplexityER");

    int numItemsValid = 0;

    for (auto& item : renderList.items) {

        if (item->type == ItemType::Nondiegetic || item->status == ItemStatus::Inactive
            || item->type == ItemType::ListenerVoice || item->type == ItemType::HOA
            || item->type == ItemType::Pannable || item->type == ItemType::Reverb || item->extent
            || consolidationThreshold == 0 // zero threshold disables consolidation
            || item->status == ItemStatus::Orphaned) {
            // remove obsolete item state and clear the slot for further use
            if (itemStates.find(item->idx) != itemStates.end()) {
                freeSlots.push_front(itemStates[item->idx].cid);
            }
            itemStates.erase(item->idx);
            continue;
        }

        if (!itemStateIsValid(*item)) {
            if (itemStates.find(item->idx) != itemStates.end()) {
                freeSlots.push_front(itemStates[item->idx].cid);
            }
            continue;
        }

        numItemsValid++;

        auto& itemState = getItemState(*item);

        // convert to listener centric coordinates
        EAR::Vector3 relativePos = renderList.listener.cs.transform(item->position.location);

        // copy positions to clustering interface
        itemState.posIn->pos_xyz = relativePos;

        itemState.posIn->isER = 0;
        if (item->type == ItemType::Reflection) {
            itemState.posIn->isER = 1;
        }

        // copy offset compensated EQ curve to clustering interface
        if (Config::getBoolDef("Consolidation_CompareEQ", true)) {
            float meanEQ = 0.0f;
            for (int i = 0; i < item->eq.size(); i++) {
                float eq_db = 20.f * std::log10(item->eq[i]);
                meanEQ += eq_db;
                itemState.posIn->eq[i] = eq_db;
            }
            meanEQ /= item->eq.size();
            for (int i = 0; i < item->eq.size(); i++) {
                itemState.posIn->eq[i] -= meanEQ;
            }
            itemState.posIn->eqLength = item->eq.size();
        } else {
            itemState.posIn->eqLength = 0;
        }

        // use item gain to represent signal energy
        itemState.posIn->energy = item->gain * item->gain;
        // TODO: if available, get pre-calculated energy of current audio stream

        *(itemState.prevMembership) = *(itemState.outMembership);
    }

    // only process active slots
    int numObjIn = numItemsValid;

    // tuning parameters of perceptual models
    pcsCfg.weight_xyz[0] = PCS_WEIGHT_LEFTRIGHT;
    pcsCfg.weight_xyz[1] = PCS_WEIGHT_FRONTBACK;
    pcsCfg.weight_xyz[2] = PCS_WEIGHT_UPDOWN;
    pcsCfg.weightEq = PCS_WEIGHT_EQ;
    pcsCfg.weightER = PCS_WEIGHT_ER;
    pcsCfg.weightDistance = PCS_WEIGHT_DISTANCE;

    // threshold from bitstream
    pcsCfg.threshold = (float)consolidationThreshold;

    // clustering in perceptual coordinates
    int numClustersActive =
        clustering_PCS_wrapper(objPosIn.get(), numObjIn, prevNumClusters, clusterPosOut.get(),
            membershipOut.get(), bIsClustered.get(), membershipPrevious.get(), objectGains.get());

    //if (MPEGI::Config::getBoolDef("Consolidation_enablePrint", false)) {
    //    std::cerr << numObjIn << " " << numClustersActive << std::endl;
    //}

    int maxNumClust = std::max(prevNumClusters, numClustersActive);
    prevNumClusters = numClustersActive;

    // create a new DSPList that will be lined up for the process function
    auto dspList = std::make_unique<ConsolidationStage::DspData>();

    for (int i = 0; i < maxNumClust; i++) {
        auto& cluster = clusters[i];
        cluster.eqSum = 0.0f;
        cluster.eq = std::vector<float>(Config::getInt("EQNBands"), 0.0f);
        cluster.numItemsER = 0;
        cluster.numItemsNoER = 0;
    }

    // loop over items to assign into clusters
    for (auto& item : renderList.items) {
        // skip items that have been skipped above
        if (itemStates.find(item->idx) == itemStates.end()) {
            continue;
        }
        auto& itemState = getItemState(*item);

        // update data structure based on clustering operation results
        if (*(itemState.isClustered)) {

            if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned) {
                itemState.currCluster = nullptr;
            } else {
                itemState.currCluster = &(clusters[*(itemState.outMembership)]);
            }
            itemState.currCluster->objectCount++;

            // save acoustic environment id to enable low complexity ERs
            // TODO: consider to make this the loudest/closest item instead of the first item
            if (useLowComplexityER && (itemState.currCluster->objectCount == 1)) {
                itemState.currCluster->acousticEnvironmentId = item->acousticEnvironmentId;
            }

            if (itemState.posIn->isER == 1)
                itemState.currCluster->numItemsER++;
            else
                itemState.currCluster->numItemsNoER++;

            // accumulate average eq for target cluster (RMS weighted by item energy)
            for (int i = 0; i < item->eq.size(); i++) {
                float eq_db = 20.f * std::log10(item->eq[i]);
                itemState.currCluster->eq[i] += item->eq[i] * item->eq[i] * itemState.posIn->energy;
            }
            itemState.currCluster->eqSum += itemState.posIn->energy;

            // deactivate primary item (this has to be done in every update())
            itemStore->deactivate(item.get());
            if (itemState.currCluster == itemState.prevCluster) {
                item->changed &=
                    ~ItemProperty::Status; // temp workaround to not set the status all the time
            }

            // fade out render item if it is deactivated
            if (itemState.prevCluster == nullptr) {
                dspList->items.emplace_back(&(item->signal->getFrame()[0]),
                    &(item->signal->getFrame()[0]), CrossFadeState::ItemFadeOut, 1.0f, nullptr);
                item->owner->skipNextFade(item->idx);
            }

            // add downmix item to DSPList
            dspList->items.emplace_back(&(item->signal->getFrame()[0]),
                itemState.currCluster->buffer,
                (itemState.currCluster != itemState.prevCluster) ? CrossFadeState::CrossFadeIn
                                                                 : CrossFadeState::None,
                (*(itemState.gain) * item->gain), itemState.eqFilter.get());

            // fade out from previous cluster if assignment has changed
            if (itemState.currCluster != itemState.prevCluster
                && itemState.prevCluster != nullptr) {

                itemState.prevCluster->objectCount += 1;
                dspList->items.emplace_back(&(item->signal->getFrame()[0]),
                    itemState.prevCluster->buffer, CrossFadeState::CrossFadeOut,
                    (*(itemState.gainPrevious) * item->gain), itemState.eqFilter.get());
            }

        } else { //item is not clustered
            if (itemState.prevCluster != nullptr) {

                // reactive previously clustered item
                item->setStatus(ItemStatus::Active, true);
                // prevent (long) fade in FadeStage and add short manual crossfade
                item->owner->skipNextFade(item->idx);
                dspList->items.emplace_back(&(item->signal->getFrame()[0]),
                    &(item->signal->getFrame()[0]), CrossFadeState::ItemFadeIn, 1.0f, nullptr);

                // fade out from previous cluster => add item to DSPList one last time
                dspList->items.emplace_back(&(item->signal->getFrame()[0]),
                    itemState.prevCluster->buffer, CrossFadeState::CrossFadeOut,
                    (*(itemState.gainPrevious) * item->gain), itemState.eqFilter.get());
                itemState.prevCluster->objectCount += 1;
            }
            itemState.currCluster = nullptr;
        }
        *(itemState.gainPrevious) = *(itemState.gain);
        itemState.prevCluster = itemState.currCluster;
    }

    // update cluster positions and activity status
    EAR::HomogenMatrix inverseListenerCS = renderList.listener.cs.inverse();

    // loop over clusters to update positions and EQs
    for (int i = 0; i < maxNumClust; i++) {
        auto& cluster = clusters[i];
        // deactivate cluster if not needed
        if (cluster.objectCount == 0) {
            itemStore->deactivate(cluster.renderItem);
            cluster.renderItem->owner->skipNextFade(cluster.renderItem->idx);
        } else {
            if (cluster.renderItem->status == ItemStatus::Inactive) {
                cluster.renderItem->setStatus(ItemStatus::Active, true); // explicitly reactivate
                cluster.renderItem->owner->skipNextFade(cluster.renderItem->idx);
            }

            if (i < numClustersActive) {
                EAR::Position pos;
                // rotation does not matter as the clustered item does not have directivity
                pos.location = inverseListenerCS.transform(cluster.newPosition->pos_xyz);
                cluster.renderItem->setPosition(pos);

                // set EQ to weighted RMS over consolidated render items
                float eqW = 1 / ((float)cluster.eqSum);
                for (int j = 0; j < cluster.eq.size(); j++) {
                    cluster.eq[j] *= eqW;
                    cluster.eq[j] = std::sqrt(cluster.eq[j]);
                }
                cluster.renderItem->setEq(cluster.eq);
            }
        }

        // add cluster item to list
        if (cluster.renderItem->status != ItemStatus::Inactive
            && cluster.renderItem->status != ItemStatus::Orphaned) {
            dspList->destinations.push_back(cluster.buffer);
        }

        // determine item type of combined render items
        if (cluster.numItemsER == 0) {
            // no early reflections in cluster
            cluster.renderItem->type = ItemType::Representative;
        } else if (cluster.numItemsNoER == 0) {
            // only early reflections in cluster => use reflection to activate dispersion filter
            cluster.renderItem->type = ItemType::Reflection;
            //} else if (cluster.numItemsER > 2*cluster.numItemsNoER) {
            //    // majority of items are reflections => use reflection to activate dispersion filter
            //    cluster.renderItem->type = ItemType::Reflection;
        } else {
            // default: do not activate dispersion for mixed render items
            cluster.renderItem->type = ItemType::Representative;
        }

        if (useLowComplexityER) {
            cluster.renderItem->acousticEnvironmentId = cluster.acousticEnvironmentId;
        }

        renderList.addItemCopy(cluster.renderItem);
        cluster.objectCount = 0;
        cluster.renderItem->resetChanged();
    }

    return dspList;
}

void ConsolidationStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<ConsolidationStage::DspData>();

    bool bApplyItemGain = Config::getBoolDef("Consolidation_applyItemGain", true);

    // clear all destination buffers before adding new signals
    for (auto dst : currentDSP.destinations) {
        dst->zero();
    }

    for (auto& item : currentDSP.items) {

        // directly apply fade for fading in/out render items
        if (item.fade == CrossFadeState::ItemFadeIn) {
            // apply fade in to destination buffer
            item.dst->mul(*fadeIn);
            item.fade = CrossFadeState::None; // fade only one block
        } else if (item.fade == CrossFadeState::ItemFadeOut) {
            // apply fade out to destination buffer
            item.dst->mul(*fadeOut);
            item.fade = CrossFadeState::FadeOutComplete; // fade only one block
        } else if (item.fade == CrossFadeState::FadeOutComplete) {
            // faded out => do nothing
            continue;
        } else {
            // add source to destination buffers with crossfade if needed

            temp->copyFrom(*item.src);

            //apply gain
            if (bApplyItemGain) {
                temp->mul(item.gain);
            }

            // crossfade needed?
            if (item.fade == CrossFadeState::CrossFadeOut) {
                temp->mul(*fadeOut);
                // fade only one block and do not add afterwards
                item.fade = CrossFadeState::FadeOutComplete;
            } else if (item.fade == CrossFadeState::CrossFadeIn) {
                temp->mul(*fadeIn);
                // fade in only one block
                item.fade = CrossFadeState::None;
            }

            item.dst->add(*temp);
        }
    }
}

void ConsolidationStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (state) { // only init when renderer is ready

        eifPacket = std::dynamic_pointer_cast<const EIFPayload>(
            bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));

        this->itemStore = itemStore;
        this->B = streamManager->getB();

        // allocate memory for C-style function in-/output
        // (TODO: can this be simplified in C++ Style?)
        clusterPosOut = std::make_unique<ObjData[]>(Config::getInt("Consolidation_NumClusters"));
        objPosIn = std::make_unique<ObjData[]>(OBJCLUST_MAX_NUM_OBJECTS);
        membershipOut = std::make_unique<int[]>(OBJCLUST_MAX_NUM_OBJECTS);
        bIsClustered = std::make_unique<int[]>(OBJCLUST_MAX_NUM_OBJECTS);
        membershipPrevious = std::make_unique<int[]>(OBJCLUST_MAX_NUM_OBJECTS);
        objectGains = std::make_unique<float[]>(OBJCLUST_MAX_NUM_OBJECTS);
        previousGains = std::make_unique<float[]>(OBJCLUST_MAX_NUM_OBJECTS);

        freeSlots.resize(OBJCLUST_MAX_NUM_OBJECTS);
        std::iota(
            freeSlots.begin(), freeSlots.end(), 0); // fill with 0,1,...,OBJCLUST_MAX_NUM_OBJECTS

        // create RenderItems for clusters
        for (int i = 0; i < Config::getInt("Consolidation_NumClusters"); i++) {
            clusters.push_back(Cluster());
            clusters.back().renderItem = itemStore->createItem("Cluster_" + std::to_string(i));
            clusters.back().renderItem->type = ItemType::Representative;
            clusters.back().renderItem->signal = streamManager->createEmptyBuffer();
            clusters.back().buffer = &(clusters.back().renderItem->signal->getFrame()[0]);
            clusters.back().renderItem->setPosition(EAR::Position());
            clusters.back().newPosition = &clusterPosOut[i];
        }

        // create temp buffer for DSP processing
        temp = std::make_unique<EAR::RealBuffer>(B);

        // create pre-calculated fade curves
        fadeIn = std::make_unique<EAR::RealBuffer>(B);
        fadeOut = std::make_unique<EAR::RealBuffer>(B);

        // short one audio block crossfade to prevent clicks
        // window shape like fade stage
        for (auto i = 0; i < B; i++) {
            float x = static_cast<float>(i) / B;

            if (x < 0.f) {
                x = 0.f;
            } else if (x > 1.f) {
                x = 1.f;
            }

            fadeIn->at(i) = x * x * (3 - 2 * x);
            x = 1 - x;
            fadeOut->at(i) = x * x * (3 - 2 * x);
        }
    }
}

ConsolidationStage::ItemState& ConsolidationStage::getItemState(const RenderItem& item)
{
    if (itemStates.find(item.idx) != itemStates.end()) {
        return itemStates.at(item.idx);
    }

    EAR_CHECK_REL(freeSlots.size() > 0, "Consolidation: Too many objects to be clustered");

    auto& newState = itemStates[item.idx];
    newState.cid = freeSlots.front();
    freeSlots.pop_front();

    newState.posIn = &objPosIn[newState.cid];
    newState.outMembership = &membershipOut[newState.cid];
    newState.prevMembership = &membershipPrevious[newState.cid];
    newState.isClustered = &bIsClustered[newState.cid];
    newState.gain = &objectGains[newState.cid];
    newState.gainPrevious = &previousGains[newState.cid];

    return newState;
}

bool ConsolidationStage::itemStateIsValid(const RenderItem& item)
{
    if (itemStates.find(item.idx) != itemStates.end()) {
        return true;
    }
    if (freeSlots.size() > 0) {
        return true;
    } else {
        return false;
    }
}

/**************************************************************/
/* Helper Functions                                           */
/**************************************************************/

/* Convert listener centric coordinates to perceptual coordinate system (PCS) */
ConsolidationStage::ObjPositionPCS ConsolidationStage::pcsWeighting(ObjData objPosIn) const
{
    ObjPositionPCS objPosPCS;

    /* normalize to unit sphere */
    float r = objPosIn.pos_xyz.abs();

    /* make sure that distance is never smaller than the minimum distance */
    if (r < PCS_REF_DISTANCE) {
        r = PCS_REF_DISTANCE;
    }

    float c = 1.0f / r;

    /* convert radius to "JND for distance" => based on distance dependent SPL attenuation, JND = 1dB, reference distance 0.2 */
    float d = 20.f * (float)log10(r / PCS_REF_DISTANCE + FLT_MIN);

    /* OpenGL style coordinate alignment, right-handed: X>0 right, Y>0 up, Z>0 = out of screen */
    objPosPCS.pos_uvw = c * dot_mul(objPosIn.pos_xyz, pcsCfg.weight_xyz);

    objPosPCS.d = d;
    objPosPCS.energy = objPosIn.energy * objPosIn.energy;

    objPosPCS.eqLength = objPosIn.eqLength;
    if (objPosIn.eqLength > 0) {
        std::memcpy(&objPosPCS.eq[0], &objPosIn.eq[0], objPosIn.eqLength * sizeof(float));
    }

    objPosPCS.itemType = objPosIn.isER;

    return objPosPCS;
}

ConsolidationStage::ObjData ConsolidationStage::pcsWeightingInverse(ObjPositionPCS objPosIn) const
{
    ObjPositionPCS objPosTmp = objPosIn;
    ObjData objPosOut;

    // inverse weight
    objPosTmp.pos_uvw = dot_div(objPosTmp.pos_uvw, pcsCfg.weight_xyz);

    // restore radius
    float r = (float)pow(10.0f, objPosIn.d * 0.05f) * PCS_REF_DISTANCE;

    objPosTmp.pos_uvw *= r;
    objPosTmp.d = 0.0f;

    objPosOut.pos_xyz = objPosTmp.pos_uvw;
    objPosOut.energy = objPosTmp.energy;

    return objPosOut;
}

// decide if two render items may be combined
bool ConsolidationStage::isBelowJnd(ObjPositionPCS pos1, ObjPositionPCS pos2) const
{
    float eqDiff = 0.0f;
    float dOut = 0.0f;
    float du = 0.0f, dv = 0.0f, dw = 0.0f;
    int jnd = pcsCfg.threshold;

    if (pos1.itemType != pos2.itemType) {
        // add penalty for combining early reflections with primary items
        dOut += pcsCfg.weightER;
    }

    if (dOut < jnd) {
        dOut += (pos1.pos_uvw - pos2.pos_uvw).abs();
    }

    if (pcsCfg.weightDistance > 0.0f) {
        float dd = (float)fabs(pos1.d - pos2.d);
        dOut += dd * pcsCfg.weightDistance;
    }

    if (dOut < jnd) {
        /* calculate average eq difference */
        if ((pos1.eqLength > 0) && (pos1.eqLength == pos2.eqLength)) {

            for (int i = 0; i < pos1.eqLength; i++) {
                eqDiff += (float)fabs(pos1.eq[i] - pos2.eq[i]);
            }
            eqDiff /= (float)pos1.eqLength;
        }
    }

    dOut += eqDiff * pcsCfg.weightEq;

    if (dOut < jnd) {
        return true;
    } else {
        return false;
    }
}

int ConsolidationStage::jndbasedSimplification(int nObjs, int* memberships, int* isClustered,
    ObjPositionPCS* centroids, int* membershipPrevious, int* centroidAssignedCounter) const
{

    int minObj, maxObj, mini = 0, minj = 0, maxCounter, membership = -1, tmp;
    float minDist, enSum, enTmp, enTmpAcc;
    ObjPositionPCS outObjects[OBJCLUST_MAX_NUM_OBJECTS], newPosition;

    /* simple greedy search for combinable clusters, starting with lowest index*/
    for (int obj1 = 0; obj1 < nObjs; obj1++) {

        /* skip objects that are already assigned */
        if (centroidAssignedCounter[obj1] == 0) {
            continue;
        }

        for (int obj2 = obj1 + 1; obj2 < nObjs; obj2++) {

            if (centroidAssignedCounter[obj1] > 0 && centroidAssignedCounter[obj2] > 0
                && isBelowJnd(centroids[obj1], centroids[obj2])) {
                maxObj = obj1;
                minObj = obj2;

                memberships[minObj] = memberships[maxObj];
                isClustered[minObj] = 1;
                isClustered[maxObj] = 1;
                tmp = centroidAssignedCounter[minObj] + centroidAssignedCounter[maxObj];
                centroidAssignedCounter[minObj] = 0;
                centroidAssignedCounter[maxObj] = tmp;

                // updating the centroid
                enTmpAcc = centroids[maxObj].energy;
                enTmp = centroids[minObj].energy;
                newPosition.pos_uvw =
                    (centroids[maxObj].pos_uvw * enTmpAcc + centroids[minObj].pos_uvw * enTmp)
                    * (1.f / (enTmpAcc + enTmp));
                newPosition.d = (centroids[maxObj].d * enTmpAcc + centroids[minObj].d * enTmp)
                    / (enTmpAcc + enTmp);

                newPosition.energy = enTmpAcc + enTmp;
                newPosition.itemType = centroids[maxObj].itemType;

                centroids[maxObj] = newPosition;
            }
        }
        isClustered[obj1] = 1; // mark as processed
    }

    int nClusters = 0;
    for (int obj = 0; obj < nObjs; obj++) {
        if (centroidAssignedCounter[obj] > 0) {
            nClusters++;
        }
    }

    return nClusters;
}

/**************************************************************/
/* Main Clustering Function                                   */
/**************************************************************/
int ConsolidationStage::clustering_simple(ObjPositionPCS* objPosIn, int numObjIn,
    int prevNumClusters, ObjPositionPCS* clusterPosOut, int* membershipOut, int* bIsClustered,
    int* membershipPrevious) const
{

    ObjPositionPCS newPosition;
    ObjPositionPCS centroids[OBJCLUST_MAX_NUM_OBJECTS];
    int numCentroids = numObjIn;
    int centroidAssignedCounter[OBJCLUST_MAX_NUM_OBJECTS];
    int membership[OBJCLUST_MAX_NUM_OBJECTS], membershipCollected[OBJCLUST_MAX_NUM_OBJECTS];

    EAR_CHECK_REL(
        numObjIn <= OBJCLUST_MAX_NUM_OBJECTS, "ConsolidationStage: numObjIn out of bounds");

    /* initialize output values TODO: check if that is needed*/
    for (int i = 0; i < numObjIn; i++) {
        clusterPosOut[i] = objPosIn[i];
    }

    /* initialize centroids at original object positions */
    for (int i = 0; i < numObjIn; i++) {
        membership[i] = i;
        centroids[i] = objPosIn[i];
        centroidAssignedCounter[i] = 1;
        bIsClustered[i] = 0;
    }

    // greedy algorithm that clusters objects with distances smaller than the threshold
    numCentroids = jndbasedSimplification(
        numObjIn, membership, bIsClustered, centroids, membershipPrevious, centroidAssignedCounter);

    /* collect output positions */
    int j = 0;
    for (int i = 0; i < numObjIn; i++) {
        if (centroidAssignedCounter[i] > 0) {
            /* find valid centroids */
            clusterPosOut[j] = centroids[i];
            membershipCollected[j] = membership[i];
            j++;
        }
    }

    for (int i = 0; i < numObjIn; i++) {
        for (int j = 0; j < numCentroids; j++) {
            if (membership[i] == membershipCollected[j]) {
                /* re-assign cluster numbers */
                membershipOut[i] = j;
            }
        }
    }

    // optimize permutation order for improved temporal stability
    optimizePermutation(objPosIn, numObjIn, numCentroids, prevNumClusters, membership,
        membershipPrevious, membershipOut, centroids, clusterPosOut);

    return numCentroids;
}

/* wrapper for conversion from/to PCS */
int ConsolidationStage::clustering_PCS_wrapper(ObjData* objPosIn, int numObjIn, int prevNumClusters,
    ObjData* clusterPosOut, int* membershipOut, int* bIsClustered, int* membershipPrevious,
    float* objectGains) const
{
    ObjPositionPCS objPosPCS[OBJCLUST_MAX_NUM_OBJECTS], clusterPosPCS[OBJCLUST_MAX_NUM_OBJECTS];
    ObjPositionPCS clusterPosPCS_prev[OBJCLUST_MAX_NUM_OBJECTS];
    int i, numClustersOut;
    float energySum = 0.0f, thrMin = 0.0f;

    for (i = 0; i < numObjIn; i++) {
        /* Converting into PCS coordinates*/
        objPosPCS[i] = pcsWeighting(objPosIn[i]);
        energySum += objPosIn[i].energy;
    }

    numClustersOut = clustering_simple(objPosPCS, numObjIn, prevNumClusters, clusterPosPCS,
        membershipOut, bIsClustered, membershipPrevious);

    /* get the relative gain from the distance between object and centroid */
    for (i = 0; i < numObjIn; i++) {
        objectGains[i] = 1.0f;
    }

    /* convert back from PCS to world coordinates*/
    for (i = 0; i < numClustersOut; i++) {
        clusterPosOut[i] = pcsWeightingInverse(clusterPosPCS[i]);
    }

    return numClustersOut;
}

/* optimize permutation order of clusters to keep energy stable*/
void ConsolidationStage::optimizePermutation(ObjPositionPCS* objPosIn, int numObjIn, int k,
    int prevk, int* membership, int* membershipPrevious, int* membershipOut,
    ObjPositionPCS* centroids, ObjPositionPCS* clusterPosOut) const
{
    int permIsAssignedCurr[OBJCLUST_MAX_NUM_OBJECTS], permIsAssignedPrev[OBJCLUST_MAX_NUM_OBJECTS],
        permutation[OBJCLUST_MAX_NUM_OBJECTS], permutedClusters[OBJCLUST_MAX_NUM_OBJECTS];
    float energyPermutation[OBJCLUST_MAX_NUM_OBJECTS][OBJCLUST_MAX_NUM_OBJECTS];

    /* reset memory */
    for (int i = 0; i < k; i++) {
        permIsAssignedCurr[i] = 0;
        permIsAssignedPrev[i] = 0;
        for (int j = 0; j < prevk; j++) {
            energyPermutation[i][j] = 0.f;
        }
    }
    /* calculate how much energy is assigned to which cluster index in current vs previous frame*/
    for (int i = 0; i < numObjIn; i++) {
        if ((membershipOut[i] >= 0)
            && (membershipPrevious[i] >= 0)) { /* TODO: check for != MEMBERSHIP_CULL instead*/
            energyPermutation[membershipOut[i]][membershipPrevious[i]] +=
                objPosIn[i].energy + FLT_MIN;
        }
    }

    /* choose permutation that minimizes redistribution of energy between frames */
    int kmin = (prevk <= k) ? prevk : k;
    for (int i = 0; i < k; i++) {
        permutation[i] = i;
        permutedClusters[i] = 0;
    }

    // optimize permutations
    for (int i = 0; i < kmin; i++) {

        int iMax, jMax;
        float maxOut;
        findMaxFilt(energyPermutation, k, kmin, &iMax, &jMax, &maxOut, permIsAssignedCurr,
            permIsAssignedPrev);

        permutation[jMax] = iMax;
        permutedClusters[iMax] = 1;

        permIsAssignedCurr[iMax] = 1;
        permIsAssignedPrev[jMax] = 1;
    }

    // second loop only used current number of clusters is greater than the previous number of clusters
    for (int i = kmin; i < k; i++) {
        for (int j = 0; j < k; j++) {
            if (permutedClusters[j] == 0) {
                permutation[i] = j;
                permutedClusters[j] = 1;
                break;
            }
        }
    }

    /* apply permutation  */
    for (int i = 0; i < k; i++) {
        centroids[i] = clusterPosOut[permutation[i]];
        for (int j = 0; j < numObjIn; j++) {
            if (membershipOut[j] == permutation[i]) {
                membership[j] = i;
            }
        }
    }
    /* copy to output array */
    for (int i = 0; i < k; i++) {
        clusterPosOut[i] = centroids[i];
    }
    for (int i = 0; i < numObjIn; i++) {
        membershipOut[i] = membership[i];
    }
}

} // namespace MPEGI
