#ifndef MPEGI_CONSOLIDATION_STAGE
#define MPEGI_CONSOLIDATION_STAGE

#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <EAR/BaseTypes.h>
#include <EAR/FIRFilter.h>

#include <cstddef>
#include <deque>
#include <map>
#include <memory>
#include <vector>

namespace MPEGI {

class EIFPayload;

//! The ConsolidationStage combines render items
/*! 
 * Inputs: 
 * - all render items except for Nondiegetic, Pannable, Reverb, HOA
 * - config parameters from mpegiSceneConfig(): consolidationThreshold
 *  
 * Outputs: 
 * - a reduced number of consoldiated render items
 *  
 * How it works: 
 * - The goal is to consolidate render items in locations that are within
 *   perceptual thresholds of localization accuracy, in order to reduce the
 *   total number of render items in the pipeline and thus computational complexity
 * - All render items are converted into a user-relative, perceptual representation, to determine
 *   pairs of render items that are within perceptual thresholds
 * - Groups of render items within thresholds are selected by a clustering algorithm
 * - All render items within a group are combined into a representative render item
 *   in a combined position and with combined equalizer curves
 * - For render items that are consolidated, the original render items are deactivated
 * - The audio signals of all comined render items are downmixed into the representative
 *   render item, applying short crossfades when assignment changes to prevent discontinuities
 *
 *
 * Stages order: 
 * - Required before: DistanceStage
 * - Required after: EqStage, FadeStage
 *  
 * Config options: 
 * - (real-time adjustable) Consolidation_JND: Perceptual threshold with render items may be combined
 * - (real-time adjustable) PCS_weightLeftRight : Perceptual weight for left-right differences
 * - (real-time adjustable) PCS_weightFrontBack : Perceptual weight for front-back differences
 * - (real-time adjustable) PCS_weightUpDown : Perceptual weight for up- differences
 * - (real-time adjustable) PCS_weightDistance : Perceptual weight for radial distance differences
 * - (real-time adjustable) PCS_weightEq : Perceptual weight differences between equalizer curves
 * - (real-time adjustable) PCS_weightEr : Perceptual weight for differences between primary and reflection items
 
 */
class ConsolidationStage : public RenderStage
{
public:
    //! Default Constructur
    ConsolidationStage();

    //! Destructor
    ~ConsolidationStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

private:
    friend class Renderer;

    //--------------------------
    // Internal data structures
    //--------------------------

    //! data structure for render item object metadata
    struct ObjData
    {
        EAR::Vector3 pos_xyz;
        float energy;
        int eqLength;
        float eq[64];
        int isER;
    };

    //! data structure for perceptual coordinate system (PCS) representation
    struct ObjPositionPCS
    {
        EAR::Vector3 pos_uvw;
        float d;
        float energy;
        int eqLength;
        float eq[64];
        int itemType;
    };

    //! data structure for configuration of perceptual tuning parameters
    struct PcsConfig
    {
        EAR::Vector3 weight_xyz;
        float weightEq;
        float weightER;
        float weightDistance;
        float threshold;
    };

    //! data structure for cluster of combined render items
    struct Cluster
    {
        RenderItem* renderItem;
        EAR::RealBuffer* buffer; //!< points to renderItem->signal
        int objectCount{ 0 };

        ObjData* newPosition;

        std::vector<float> eq;
        float eqSum;

        int numItemsER;
        int numItemsNoER;

        //! Id of the Reverb environment this item is located in (name of the acoustic environment)
        std::string acousticEnvironmentId;
    };

    //! data structure for render item properties assigned to a cluster
    struct ItemState
    {
        int cid; //!< ID used by the clustering algorithm (!= item->idx)

        std::unique_ptr<EAR::FIRFilter> eqFilter;

        ObjData* posIn;
        int* outMembership;
        int* prevMembership;
        int* isClustered;
        float* gain;
        float* gainPrevious;

        Cluster* currCluster{ nullptr };
        Cluster* prevCluster{ nullptr };
    };

    enum class CrossFadeState
    {
        None = 0,
        CrossFadeIn = 1,
        CrossFadeOut = -1,
        ItemFadeIn = 2,
        ItemFadeOut = -2,
        FadeOutComplete = -3
    };

    //! Data structure for instructing the DSP thread
    struct DownmixItem
    {
        EAR::RealBuffer* src;
        EAR::RealBuffer* dst;

        CrossFadeState fade;

        float gain;
        EAR::FIRFilter* eqFilter;

        DownmixItem(EAR::RealBuffer* src, EAR::RealBuffer* dst, CrossFadeState fade, float gain,
            EAR::FIRFilter* eqFilter)
            : src(src)
            , dst(dst)
            , fade(fade)
            , gain(gain)
            , eqFilter(eqFilter){};
    };

    //! data to be passed from update thread to audio thread
    struct DspData : RenderStage::DspData
    {
        std::vector<DownmixItem> items;
        std::vector<EAR::RealBuffer*> destinations;
    };

private:
    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    //! Returns the render item state
    /*!
     * \param parentItem           parent render item
     * \return                     render item state
     */
    inline ItemState& getItemState(const RenderItem& item);

    //! checks for valid render item state
    /*!
     * \param parentItem           parent render item
     * \return                     render item state is valid
     */
    bool itemStateIsValid(const RenderItem& item);

    //! converts metadata from geometric coordinates into perceptual representation
    /*!
     * \param objPosIn      input: object position and metadata in geometric representation
     * \return              returns object position in perceptual representation
     */
    ObjPositionPCS pcsWeighting(ObjData objPosIn) const;

    //! inverts perceptual weighting to convert metadata back from perceptual to geometric representation
    /*!
     * \param objPosIn      input: object position in perceptual representation
     * \return              returns object position in geometric representation
     */
    ObjData pcsWeightingInverse(ObjPositionPCS objPosIn) const;

    //! determines if the perceptual difference of two objects is below a given threshold
    /*!
     * \param pos1          input: position of first object in perceptual representation
     * \param pos2          input: position of second object in perceptual representation
     * \return              returns true if distance is below threshold
     */
    bool isBelowJnd(ObjPositionPCS pos1, ObjPositionPCS pos2) const;

    //! uses a greedy clustering algorithm to determine groups of items within thresholds
    /*!
     * \param nObjs                     input: number of input objects
     * \param memberships               output: array of memberships of objects assigned to clusters
     * \param isClustered               output: array of flags indicating if an object was assigned to a cluster
     * \param centroids                 in/out: centroid positions - in: original out: clustered
     * \param membershipPrevious        input: array of previous object memberships
     * \param centroidAssignedCounter   output: counter of number of objects assigned to each cluster
     * \return                          returns number of consolidated objects
     */
    int jndbasedSimplification(int nObjs, int* memberships, int* isClustered,
        ObjPositionPCS* centroids, int* membershipPrevious, int* centroidAssignedCounter) const;

    //! consolidate render items by simple greedy clustering algorithm + permutation optimization
    /*!
     * \param objPosIn              input: object position in perceptual coordinates
     * \param numObjIn              input: number of imput objects
     * \param prevNumClusters       input: previous number of clusters
     * \param clusterPosOut         output: consolidated cluster positions in perceptual coordinates
     * \param membershipOut         output: array of memberships of objects assigned to clusters
     * \param bIsClustered          output: array of flags indicating if an object was assigned to a cluster
     * \param membershipPrevious    input: array of previous object memberships
     * \return                      returns number of consolidated objects
     */
    int clustering_simple(ObjPositionPCS* objPosIn, int numObjIn, int prevNumClusters,
        ObjPositionPCS* clusterPosOut, int* membershipOut, int* bIsClustered,
        int* membershipPrevious) const;

    //! wrapper function for perception-based clustering, converts into perceptual domain and calls actual clustering
    /*!
     * \param objPosIn              input: object positions in geometric coordinates
     * \param numObjIn              input: number of input objects
     * \param prevNumClusters       input: previous number of clusters
     * \param clusterPosOut         output: consolidated cluster positions in geometric coordinates
     * \param membershipOut         output: array of memberships of objects assigned to clusters
     * \param bIsClustered          output: array of flags indicating if an object was assigned to a cluster
     * \param membershipPrevious    input: array of previous object memberships
     * \param objectGains           output: downmix gains of consolidated objects
     * \return                      returns number of consolidated objects
     */
    int clustering_PCS_wrapper(ObjData* objPosIn, int numObjIn, int prevNumClusters,
        ObjData* clusterPosOut, int* membershipOut, int* bIsClustered, int* membershipPrevious,
        float* objectGains) const;

    //! optimizes permutation of cluster index by minimizing re-assignment of signal energy
    /*!
     * \param objPosIn              input: object position in perceptual coordinates
     * \param numObjIn              input: number of input objects
     * \param k                     input: current number of clusters
     * \param prevk                 input: previous number of clusters
     * \param membership            input: array of currently assigned object memberships
     * \param membershipPrevious    input: array of previously assigned object memberships
     * \param membershipOut         output: array of optimized  object memberships
     * \param centroids             input: cluster centroids in perceptual coordinates
     * \param clusterPosOut         output: optimized cluster centroids in perceptual coordinates
     */
    void optimizePermutation(ObjPositionPCS* objPosIn, int numObjIn, int k, int prevk,
        int* membership, int* membershipPrevious, int* membershipOut, ObjPositionPCS* centroids,
        ObjPositionPCS* clusterPosOut) const;

private:
    RenderItemStore* itemStore; //!< map of render items

    PcsConfig pcsCfg; //!< configuration parameters of perceptual coordinate system

    std::size_t B; //!< audio block size

    std::shared_ptr<const EIFPayload> eifPacket; //!< EIF payload packet

    int prevNumClusters = 0; //!< previous number of clusters

    std::vector<Cluster> clusters; //!< consolidated clusters

    std::unique_ptr<ObjData[]> objPosIn; //!< input object positions
    std::unique_ptr<ObjData[]> clusterPosOut; //!< output cluster positions

    std::unique_ptr<int[]> membershipOut; //!< membership of render item assignment
    std::unique_ptr<int[]> bIsClustered; //!< array of flags indicating if assigned to a cluster
    std::unique_ptr<int[]> membershipPrevious; //!< memory for previous membership

    std::unique_ptr<float[]> objectGains; //!< object downmix gains
    std::unique_ptr<float[]> previousGains; //!< previous object downmix gains

    std::deque<int>
        freeSlots; //!< slots in objPosIn etc. that can be used by new ItemState instances

    EAR::Vector3 listenerPrevPos; //!< previous listener position

    std::map<int, ItemState> itemStates; //!< map of item states

    std::unique_ptr<EAR::RealBuffer> temp; //!< temporary buffer for windowing
    std::unique_ptr<EAR::RealBuffer> fadeIn; //!< fade-in window
    std::unique_ptr<EAR::RealBuffer> fadeOut; //!< fade-out window
    std::unique_ptr<EAR::RealBuffer> tempEqFilter; //!< temporary eq filter
};

} // namespace MPEGI

#endif /* MPEGI_CONSOLIDATION_STAGE */
