/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DebugPrinting.h"

#include <iostream>

namespace MPEGI {

std::shared_ptr<RenderStage> makeDebugPrintingStage()
{
    return std::make_shared<DebugPrintingStage>();
}

std::unique_ptr<RenderStage::DspData> DebugPrintingStage::update(RenderList& renderList)
{
    std::cout << "Update using render list " << renderList.id << std::endl;

    const auto& listener = renderList.listener;
    if (listener.hasMoved) {
        std::cout << "Listener has moved to " << listener.position;
        if (listener.teleport) {
            std::cout << " using teleportation";
        }
        std::cout << std::endl;
    }

    for (auto& item : renderList.items) {
        std::cout << "item " << item->idx;
        if (!item->objectName.empty()) {
            std::cout << " (" << item->objectName << ")";
        }
        std::cout << "; ";
        switch (item->type) {
        case ItemType::Primary:
            std::cout << "Primary";
            break;
        case ItemType::HOA:
            std::cout << "HOA";
            break;
        case ItemType::Representative:
            std::cout << "Representative";
            break;
        case ItemType::Reflection:
            std::cout << "Reflection";
            break;
        case ItemType::DirectionalReverb:
            std::cout << "Portal";
            break;
        case ItemType::Nondiegetic:
            std::cout << "Nondiegetic";
            break;
        case ItemType::Downmix:
            std::cout << "Downmix";
            break;
        case ItemType::Pannable:
            std::cout << "Pannable";
            break;
        case ItemType::Reverb:
            std::cout << "Reverb";
            break;
        case ItemType::ListenerVoice:
            std::cout << "ListenerVoice";
            break;
        default:
            std::cout << "UNKNOWN(" << (int)item->type << ")";
        }
        std::cout << "; ";

        switch (item->status) {
        case ItemStatus::New:
            std::cout << "New";
            break;
        case ItemStatus::Active:
            std::cout << "Active";
            break;
        case ItemStatus::FadeOutInactive:
            std::cout << "FadeOutInactive";
            break;
        case ItemStatus::Inactive:
            std::cout << "Inactive";
            break;
        case ItemStatus::FadeOutOrphaned:
            std::cout << "FadeOutOrphaned";
            break;
        case ItemStatus::Orphaned:
            std::cout << "Orphaned";
            break;
        default:
            std::cout << "UNKNOWN(" << (int)item->status << ")";
        }
        if (item->changed & ItemProperty::Status) {
            std::cout << " (changed)";
        }
        std::cout << "; ";

        if (item->teleport) {
            std::cout << "teleport; ";
        }

        std::cout << std::endl;

        if (item->changed & ItemProperty::Aparams) {
            std::cout << " Aparams changed to [ ";
            if (item->aparams == AParam::None) {
                std::cout << "None ";
            }
            if (item->aparams & AParam::NoDoppler) {
                std::cout << "NoDoppler ";
            }
            if (item->aparams & AParam::NoDistanceGain) {
                std::cout << "NoDistanceGain ";
            }
            if (item->aparams & AParam::NoMediumAttenuation) {
                std::cout << "NoMediumAttenuation ";
            }
            if (item->aparams & AParam::noDirectionalFocus) {
                std::cout << "noDirectionalFocus ";
            }
            std::cout << "]" << std::endl;
        }

        if (item->changed
            & (ItemProperty::Position | ItemProperty::Location | ItemProperty::Orientation)) {
            if (item->changed & ItemProperty::Position) {
                std::cout << " Position";
            }
            if (item->changed & ItemProperty::Location) {
                std::cout << " Location";
            }
            if (item->changed & ItemProperty::Orientation) {
                std::cout << " Orientation";
            }
            std::cout << " changed to " << item->position << std::endl;
        }

        if (item->changed & ItemProperty::Directivity) {
            std::cout << " Directivity changed: TODO" << std::endl;
        }
        if (item->changed & ItemProperty::Extent) {
            std::cout << " Extent changed to ";
            if (item->extent) {
                std::cout << "\"" << item->extent->getName() << "\"";
                std::cout << " (object ID " << item->extent->getID() << ")";
                if (item->extent->isStatic()) {
                    std::cout << " (static)";
                } else {
                    std::cout << " (dynamic)";
                }
                std::cout << std::endl;
            } else {
                std::cout << "nullptr" << std::endl;
            }
        }
        if (item->extent && item->extent->hasMoved()) {
            std::cout << " Extent has moved to " << item->extent->getGlobalPosition() << std::endl;
        }
        if (item->changed & ItemProperty::RayHits) {
            std::cout << " Updated extended source ray hits: ";
            if (item->rayHits.empty()) {
                std::cout << " no hits";
            } else {
                std::cout << item->rayHits.size() << " hit(s) (ray ID, rel. dist.):";
                for (const auto& hit : item->rayHits) {
                    std::cout << " (" << hit.rayID << ", " << hit.distance << ", " << hit.weight
                              << ")";
                }
            }
            std::cout << std::endl;
        }
        if (item->changed & ItemProperty::Signal) {
            std::cout << " Signal changed: TODO" << std::endl;
        }
        if (item->changed & ItemProperty::Gain) {
            std::cout << "Gain changed" << std::endl;
            std::cout << " Gain: " << item->gain << std::endl;
        }
        if (item->changed & ItemProperty::Eq) {
            std::cout << "Eq changed" << std::endl;
            std::cout << " Eq: " << item->eq << std::endl;
        }
        if (item->changed & ItemProperty::RefDist) {
            std::cout << " RefDist changed: TODO" << std::endl;
        }
        if (item->changed & ItemProperty::ParentRI) {
            std::cout << " ParentRI changed to " << item->parentRI << std::endl;
        }
        if (item->changed & ItemProperty::Reverb) {
            std::cout << " Reverb changed: TODO" << std::endl;
        }
        if (item->changed & ItemProperty::GranularPos) {
            std::cout << " Granular position changed: TODO" << std::endl;
        }
        if (item->changed & ItemProperty::GranularGroupProb) {
            std::cout << " Granular group probability changed: TODO" << std::endl;
        }
        if (item->changed & ItemProperty::ReflectionInfo) {
            if (item->reflectionInfo) {
                // TODO: show name of associated primary source?
                const auto& seq = item->reflectionInfo->intersectionSequence;
                std::cout << " " << seq.size() << " reflection(s):";
                for (const auto& v : seq) {
                    std::cout << " " << v;
                }
                std::cout << std::endl;
            }
        }
    }
    return {};
}

void DebugPrintingStage::processBlock(DspContext ctx)
{
    std::cout << "Processing audio block " << this->processCounter++ << std::endl;
}

void DebugPrintingStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    std::cout << "Initializing DebugPrintingStage" << std::endl;
}

} // namespace MPEGI
