/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "AngleCalculation.h"

namespace MPEGI {

AngleCalculation::AngleCalculation(void)
    : V_origin(0, 0, 0)
    , V_bias(0, 0, 0)
    , V_origin2bias(0, 0, 0)
    , Ui_ref(0, 0, 0)
    , Uj_tar(0, 0, 0)
    , N0_Ui_Plane(0, 0, 0)
    , V_VIRTUAL(0, 0, 0)
    , primProperty(PrimativePropertiesforAngleCalculation::None)
    , CurrentCoorSys(CoordinateSystem::None)
    , ConstructorType(-1)
{}

AngleCalculation::AngleCalculation(
    const DiffractionEdge* _edge, int _faceIndicator, EAR::Vector3 _target_vertex_coord)
{
    this->V_origin = _edge->vtxCoords.first;
    this->V_bias = _edge->vtxCoords.second;
    this->V_origin2bias = this->V_bias - this->V_origin;

    if (_faceIndicator == 0) {
        this->Ui_ref = _edge->adjacentTriCentroids.first;
        this->N0_Ui_Plane = _edge->adjacentTriNormals.first;
    } else {
        this->Ui_ref = _edge->adjacentTriCentroids.second;
        this->N0_Ui_Plane = _edge->adjacentTriNormals.second;
    }

    this->Uj_tar = _target_vertex_coord;
    this->V_VIRTUAL = EAR::Vector3(0, 0, 0);
    this->primProperty = PrimativePropertiesforAngleCalculation::VertexToTriangle;
    this->CurrentCoorSys = CoordinateSystem::YZX_UNITY;
    this->ConstructorType = 0;
}

AngleCalculation::AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
    EAR::Mesh::Triangle* ref_tri, EAR::Vector3 target_vertex_coord)
    : V_origin(edgeVtxCoords.first)
    , V_bias(edgeVtxCoords.second)
    , V_origin2bias(V_bias - V_origin)
    , Ui_ref(ref_tri->getCentroid(EAR::Mesh::CSpace::Global))
    , Uj_tar(target_vertex_coord)
    , N0_Ui_Plane(ref_tri->getNormal(EAR::Mesh::CSpace::Global))
    , V_VIRTUAL(0, 0, 0)
    , primProperty(PrimativePropertiesforAngleCalculation::VertexToTriangle)
    , CurrentCoorSys(CoordinateSystem::YZX_UNITY)
    , ConstructorType(1)
{}

AngleCalculation::AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
    EAR::Vector3 ref_vertex_coord, EAR::Vector3 target_vertex_coord)
    : V_origin(edgeVtxCoords.first)
    , V_bias(edgeVtxCoords.second)
    , V_origin2bias(V_bias - V_origin)
    , Ui_ref(ref_vertex_coord)
    , Uj_tar(target_vertex_coord)
    , N0_Ui_Plane(0, 0, 0)
    , V_VIRTUAL(0, 0, 0)
    , primProperty(PrimativePropertiesforAngleCalculation::VertexToVertex)
    , CurrentCoorSys(CoordinateSystem::YZX_UNITY)
    , ConstructorType(2)
{}

AngleCalculation::AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
    EAR::Mesh::Triangle* tri_i, EAR::Mesh::Triangle* tri_j)
    : V_origin(edgeVtxCoords.first)
    , V_bias(edgeVtxCoords.second)
    , V_origin2bias(V_bias - V_origin)
    , Ui_ref(tri_i->getCentroid(EAR::Mesh::CSpace::Global))
    , Uj_tar(tri_j->getCentroid(EAR::Mesh::CSpace::Global))
    , N0_Ui_Plane(tri_i->getNormal(EAR::Mesh::CSpace::Global))
    , V_VIRTUAL(0, 0, 0)
    , primProperty(PrimativePropertiesforAngleCalculation::TriangleToTriangle)
    , CurrentCoorSys(CoordinateSystem::YZX_UNITY)
    , ConstructorType(3)
{}

AngleCalculation::AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
    EAR::Vector3 v0, EAR::Vector3 v2, EAR::Vector3 rot_srcloc)
    : primProperty(PrimativePropertiesforAngleCalculation::EdgeIntersection)
    , CurrentCoorSys(CoordinateSystem::YZX_UNITY)
    , ConstructorType(4)
{
    //! v0: lisloc, v2: intersec_p
    EAR::Vector3 _vec3_fir = v2 - edgeVtxCoords.first;
    EAR::Vector3 _vec3_sec = v2 - edgeVtxCoords.second;
    EAR::Vector3 v1;

    if (_vec3_fir.abs() > _vec3_sec.abs()) {
        v1 = edgeVtxCoords.second;
    } else {
        v1 = edgeVtxCoords.first;
    }

    V_VIRTUAL = v1;
    N0_Ui_Plane = (v0 - v1) * (v1 - v2);
    N0_Ui_Plane = N0_Ui_Plane.normalize();

    V_origin = V_VIRTUAL + N0_Ui_Plane;
    V_bias = V_VIRTUAL - N0_Ui_Plane;
    V_origin2bias = V_bias - V_origin;
    Ui_ref = v0;
    Uj_tar = rot_srcloc;
    N0_Ui_Plane = (v0 - V_origin) * (V_origin - V_bias);
}

AngleCalculation::~AngleCalculation(void) {}

float AngleCalculation::calVecAngleXY(EAR::Vector3 vec3)
{
    bool alongAxis = (abs(vec3[0]) == 0) && (abs(vec3[1]) == 0);
    if (alongAxis) {
        // when a target source is along the rotating axis: return invalid value -1;
        return -1.f;
    }

    if (abs(vec3[0]) == 0) {
        if (vec3[1] > 0) {
            return 90;
        } else if (vec3[1] < 0) {
            return 270;
        }
    }

    vec3.normalize();
    float angle = (float)(atan(abs(vec3[1]) / abs(vec3[0])) * (180 / M_PI));

    if ((vec3[0] <= 0) && (vec3[1] >= 0)) {
        return (180 - angle);
    } else if ((vec3[0] <= 0) && (vec3[1] <= 0)) {
        return (180 + angle);
    } else if ((vec3[0] > 0) && (vec3[1] < 0)) {
        return (360 - angle);
    } else {
        return (angle);
    }
}

bool AngleCalculation::setXYZCoordinates()
{
    if (CurrentCoorSys == CoordinateSystem::YZX_UNITY) {
        V_origin = { V_origin[2], V_origin[0], V_origin[1] };
        V_bias = { V_bias[2], V_bias[0], V_bias[1] };
        V_origin2bias = { V_origin2bias[2], V_origin2bias[0], V_origin2bias[1] };
        Ui_ref = { Ui_ref[2], Ui_ref[0], Ui_ref[1] };
        Uj_tar = { Uj_tar[2], Uj_tar[0], Uj_tar[1] };
        N0_Ui_Plane = { N0_Ui_Plane[2], N0_Ui_Plane[0], N0_Ui_Plane[1] };
        CurrentCoorSys = CoordinateSystem::XYZ;

        return true;
    } else {
        EAR_CHECK_REL(false, "[ERROR] CURRENT COORDINATE SYSTEM IS [NONE] OR [XYZ] ALREADY");

        return false;
    }
}

bool AngleCalculation::setYZXCoordinates()
{
    if (CurrentCoorSys == CoordinateSystem::XYZ) {
        V_origin = { V_origin[1], V_origin[2], V_origin[0] };
        V_bias = { V_bias[1], V_bias[2], V_bias[0] };
        V_origin2bias = { V_origin2bias[1], V_origin2bias[2], V_origin2bias[0] };
        Ui_ref = { Ui_ref[1], Ui_ref[2], Ui_ref[0] };
        Uj_tar = { Uj_tar[1], Uj_tar[2], Uj_tar[0] };
        N0_Ui_Plane = { N0_Ui_Plane[1], N0_Ui_Plane[2], N0_Ui_Plane[0] };
        CurrentCoorSys = CoordinateSystem::YZX_UNITY;

        return true;
    } else {
        EAR_CHECK_REL(false, "[ERROR] CURRENT COORDINATE SYSTEM IS [NONE] OR [YZX_UNITY] ALREADY");

        return false;
    }
}

float AngleCalculation::calculateInternalAngle()
{
    EAR_CHECK_REL((ConstructorType != -1),
        "[ERROR] CURRENT CONSTRUCTOR TYPE IS 0 BY DEFAULT: NOTHING HAS BEEN UPDATED FOR "
        "CALCULATION");

    if (!setXYZCoordinates()) {
        EAR_CHECK_REL(
            (CurrentCoorSys == CoordinateSystem::XYZ), "[ERROR] CURRENT COORDINATE SYSTEM IS NONE");
    }

    float faceAngle;
    // 1) initialization, 2) translation 3) X - Axis Rotation, 4) Y - Axis Rotation
    setAllMatricesAsIdentity();
    setTranslationMatrix(-1 * V_origin);
    EAR::Vector3 unbiased_vec = MatVecMul(V_bias, 0);
    setRotationMatrix(unbiased_vec, RotatingAxis::X_AXIS);
    setRotationMatrix(unbiased_vec, RotatingAxis::Y_AXIS);

    // 5) calcluate the transformation matrix to rotate given source location with respect to given rotating axis (e.g., an edge)
    getTransformationMatrix(3);

    // 6) calculate the rotated vectors
    EAR::Vector3 V_origin_rot = MatVecMul(V_origin, 7);
    EAR::Vector3 V_bias_rot = MatVecMul(V_bias, 7);
    EAR::Vector3 Ui_ref_rot = MatVecMul(Ui_ref, 7);
    EAR::Vector3 Uj_tar_rot = MatVecMul(Uj_tar, 7);
    EAR::Vector3 N0_Ui_Plane_rot = MatVecMul(N0_Ui_Plane + V_origin, 7);

    // 7) find the direction of rotation
    EAR::Vector3 Proj_V_rot_Ui_ref =
        V_bias_rot.dot(Ui_ref_rot) / V_bias_rot.abs() / V_bias_rot.abs() * V_bias_rot;
    EAR::Vector3 Otho_V_rot_Ui_ref = Ui_ref_rot - Proj_V_rot_Ui_ref;
    EAR::Vector3 Proj_V_rot_Uj_tar =
        V_bias_rot.dot(Uj_tar_rot) / V_bias_rot.abs() / V_bias_rot.abs() * V_bias_rot;
    EAR::Vector3 Otho_V_rot_Uj_tar = Uj_tar_rot - Proj_V_rot_Uj_tar;
    EAR::Vector3 CW_CCW_CHECK = Otho_V_rot_Ui_ref.cross(N0_Ui_Plane_rot);

    float ref_angle = calVecAngleXY(Otho_V_rot_Ui_ref);
    float tar_angle = calVecAngleXY(Otho_V_rot_Uj_tar);

    if (ref_angle == -1.f || tar_angle == -1.f) {
        // when a target source is along the rotating axis: return 0.f diffration angle
        return 0.f;
    }

    // 8) returns the internal angle between the reference plane and the target vector (defined by the target vertex)
    if (ref_angle >= 0) {
        if (CW_CCW_CHECK[2] > 0) {
            faceAngle = (float)fmod(tar_angle - ref_angle, 360);
        } else {
            faceAngle = (float)fmod(ref_angle - tar_angle, 360);
        }

        float thrsh = Config::getFloat("Diffraction_GeometricErrorThreshold");

        //! inevitable to deal with the geometrical error -> another way to deal with this?
        if (faceAngle < -thrsh) {
            faceAngle += 360;
        } else if ((faceAngle > -thrsh) && (faceAngle < thrsh)) {
            faceAngle = 0;
        } else if ((faceAngle > 360 - thrsh) && (faceAngle < 360 + thrsh)) {
            faceAngle = 0;
        } else if (faceAngle > 360 + thrsh) {
            faceAngle -= 360;
        }
    }

    if (!setYZXCoordinates()) {
        EAR_CHECK_REL((CurrentCoorSys != CoordinateSystem::YZX_UNITY),
            "[ERROR] CURRENT COORDINATE SYSTEM IS NOT YZX_UNITY");
    }

    return faceAngle;
}
} // namespace MPEGI
