/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_ANGLE_CALCULATION
#define MPEGI_ANGLE_CALCULATION

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Renderer/DiffractionEdge.h>

#include <EAR/Exception.h>
#include <EAR/Geometry.h>

#include "ArbitraryAxisRotation.h"
#include "Definitions.h"

namespace MPEGI {

// Angle calcuation class
/*!
* \usage: this class provides the interface for the calculation in different configuration/set-up
			1. when the input arguments are a point and a triangle, then it provides the interface to calculate the angle between two planes:
				one of two planes is a given plane by a triangle and the other of which is the triangular plane which is defined by a vertex and a given two edge end points
			2. when the input arguments are two points, then it provides the interface to calculate the angle between two planes:
				each triangular planes are defined by the each given point and a given edge's end points
			3. when the input arguments are two triangles, then it provides the interface to calculate the angle between two planes:
*/
class AngleCalculation : ArbitraryAxisRotation
{
public:
    // Constructor
    AngleCalculation(void);

    // Constructor for PrimativePropertiesforAngleCalculation::VertexToTriangle
    AngleCalculation(
        const DiffractionEdge* _edge, int _faceIndicator, EAR::Vector3 _target_vertex_coord);

    // Constructor for PrimativePropertiesforAngleCalculation::VertexToTriangle
    AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
        EAR::Mesh::Triangle* ref_tri, EAR::Vector3 target_vertex_coord);

    // Constructor for PrimativePropertiesforAngleCalculation::VertexToVertex
    AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
        EAR::Vector3 ref_vertex_coord, EAR::Vector3 target_vertex_coord);

    // Constructor for PrimativePropertiesforAngleCalculation::TriangleToTriangle
    AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords,
        EAR::Mesh::Triangle* tri_i, EAR::Mesh::Triangle* tri_j);

    // Constructor for PrimativePropertiesforAngleCalculation::EccentricityCalculation
    AngleCalculation(std::pair<EAR::Vector3, EAR::Vector3> edgeVtxCoords, EAR::Vector3 v0,
        EAR::Vector3 v2, EAR::Vector3 rot_srcloc);

    // Destructor
    ~AngleCalculation(void);

    // Return vector angle on the X-Y plane
    float calVecAngleXY(EAR::Vector3 vec3);

    // Set coordinate system as XYZ
    bool setXYZCoordinates();

    // Set coordinate system as YZX - OpenGL
    bool setYZXCoordinates();

    // Return the internal angle between adjacent triangles
    float calculateInternalAngle();

    const PrimativePropertiesforAngleCalculation& getPrimProperty() { return primProperty; }

    const EAR::Vector3& getV_origin() { return V_origin; }

    const EAR::Vector3& getV_bias() { return V_bias; }

    const EAR::Vector3& getV_origin2bias() { return V_origin2bias; }

    const EAR::Vector3& getUi_ref() { return Ui_ref; }

    const EAR::Vector3& getUj_tar() { return Uj_tar; }

    const EAR::Vector3& getN0_Ui_Plane() { return N0_Ui_Plane; }

    const EAR::Vector3& getV_VIRTUAL() { return V_VIRTUAL; }

private:
    EAR::Vector3 V_origin;
    EAR::Vector3 V_bias;
    EAR::Vector3 V_origin2bias;
    EAR::Vector3 Ui_ref;
    EAR::Vector3 Uj_tar;
    EAR::Vector3 N0_Ui_Plane;
    EAR::Vector3 V_VIRTUAL = EAR::Vector3(0, 0, 0);

    PrimativePropertiesforAngleCalculation primProperty;
    int ConstructorType; // redundant
    CoordinateSystem CurrentCoorSys;
};

} // namespace MPEGI

#endif /* MPEGI_ANGLE_CALCULATION */
