/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ArbitraryAxisRotation.h"

namespace MPEGI {

ArbitraryAxisRotation::ArbitraryAxisRotation(void)
{
    setAllMatricesAsIdentity();
}

ArbitraryAxisRotation::~ArbitraryAxisRotation(void) {}

void ArbitraryAxisRotation::setMatrixIdentity(int matrixIndex)
{
    EAR_CHECK_REL((matrixIndex >= 0 && matrixIndex <= 7),
        "[ERROR] SET_MATRIX_IDENTITY WITH UNSUPPORTED INDEX");

    if (matrixIndex == 7) {
        for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 4; j++) {
                matrix_transform[i][j] = 0.F;

                if (i == j) {
                    matrix_transform[i][j] = 1.0F;
                }
            }
        }
    } else {
        for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 4; j++) {
                matrix_phantom[matrixIndex][i][j] = 0.F;

                if (i == j) {
                    matrix_phantom[matrixIndex][i][j] = 1.F;
                }
            }
        }
    }
}

void ArbitraryAxisRotation::setAllMatricesAsIdentity()
{
    for (int i = 0; i < 8; i++) {
        setMatrixIdentity(i);
    }
}

void ArbitraryAxisRotation::setTranslationMatrix(EAR::Vector3 translationVector)
{
    setMatrixIdentity(0);
    matrix_phantom[0][0][3] = translationVector[0];
    matrix_phantom[0][1][3] = translationVector[1];
    matrix_phantom[0][2][3] = translationVector[2];

    setMatrixIdentity(6 - 0);
    matrix_phantom[6 - 0][0][3] = -translationVector[0];
    matrix_phantom[6 - 0][1][3] = -translationVector[1];
    matrix_phantom[6 - 0][2][3] = -translationVector[2];
}

void ArbitraryAxisRotation::setRotationMatrix(
    EAR::Vector3 rotatingAxis, RotatingAxis CoordAxis, float alpha)
{
    float A = rotatingAxis[0], B = rotatingAxis[1], C = rotatingAxis[2];
    float L = rotatingAxis.abs(), V = sqrt(B * B + C * C);

    if (CoordAxis == RotatingAxis::X_AXIS) {
        // Projection of unbiased rotating axis onto the XZ plane
        setMatrixIdentity(1);
        setMatrixIdentity(6 - 1);

        if (V > Config::getFloat("Diffraction_GeometricErrorThreshold")) {
            matrix_phantom[1][1][1] = C / V;
            matrix_phantom[1][2][1] = B / V;
            matrix_phantom[1][1][2] = -B / V;
            matrix_phantom[1][2][2] = C / V;

            matrix_phantom[6 - 1][1][1] = C / V;
            matrix_phantom[6 - 1][1][2] = B / V;
            matrix_phantom[6 - 1][2][1] = -B / V;
            matrix_phantom[6 - 1][2][2] = C / V;
        }
    } else if (CoordAxis == RotatingAxis::Y_AXIS) {
        // Projection of unbiased rotating axis along the Z axis
        setMatrixIdentity(2);
        setMatrixIdentity(6 - 2);

        if (L > Config::getFloat("Diffraction_GeometricErrorThreshold")) {
            matrix_phantom[2][0][0] = V / L;
            matrix_phantom[2][0][2] = -A / L;
            matrix_phantom[2][2][0] = A / L;
            matrix_phantom[2][2][2] = V / L;

            matrix_phantom[6 - 2][0][0] = V / L;
            matrix_phantom[6 - 2][0][2] = A / L;
            matrix_phantom[6 - 2][2][0] = -A / L;
            matrix_phantom[6 - 2][2][2] = V / L;
        }
    }
    // -> Z rotation
    else if (CoordAxis == RotatingAxis::Z_AXIS) {
        setMatrixIdentity(3);
        matrix_phantom[3][0][0] = (float)cos(alpha * M_PI / 180);
        matrix_phantom[3][0][1] = (float)-sin(alpha * M_PI / 180);
        matrix_phantom[3][1][0] = (float)sin(alpha * M_PI / 180);
        matrix_phantom[3][1][1] = (float)cos(alpha * M_PI / 180);
    }
}

void ArbitraryAxisRotation::MatMul(int matrixIndex)
{
    EAR_CHECK_REL((matrixIndex >= 0 && matrixIndex <= 6), "[ERROR] MAT_MUL WITH UNSUPPORTED INDEX");

    //! TODO: Do not use the product, but matrix_transform directly.
    float product[4][4] = {
        0.f,
    };
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            for (int k = 0; k < 4; k++) {
                product[i][j] += matrix_phantom[matrixIndex][i][k] * matrix_transform[k][j];
            }
        }
    }

    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            matrix_transform[i][j] = product[i][j];
        }
    }
}

EAR::Vector3 ArbitraryAxisRotation::MatVecMul(EAR::Vector3 p, int matrixIndex)
{
    EAR_CHECK_REL((matrixIndex >= 0 && matrixIndex <= 7), "[ERROR] MAT_MUL WITH UNSUPPORTED INDEX");

    float p_dot[4] = { p[0], p[1], p[2], 1 }; //! p-dot
    float p_tra[4] = {
        0,
    }; //! p-transformed (translated or rotated)

    if (matrixIndex == 7) {
        for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 4; j++) {
                p_tra[i] += matrix_transform[i][j] * p_dot[j];
            }
        }
    } else {
        for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 4; j++) {
                p_tra[i] += matrix_phantom[matrixIndex][i][j] * p_dot[j];
            }
        }
    }

    p.set(p_tra);
    return p;
}

void ArbitraryAxisRotation::getTransformationMatrix(int phantomMatrixLastIndex)
{
    EAR_CHECK_REL((phantomMatrixLastIndex >= 0 && phantomMatrixLastIndex <= 7),
        "[ERROR] SET_MATRIX_IDENTITY WITH UNSUPPORTED INDEX");

    setMatrixIdentity(7);
    for (int i = 0; i < phantomMatrixLastIndex; i++) {
        MatMul(i);
    }
}
} // namespace MPEGI
