/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Diffraction.h"

#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

#include <EAR/BaseTypes.h>
#include <EAR/EmbreeTracer.h>
#include <EAR/Exception.h>
#include <EAR/Geometry.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectObserver.h>

#include "AngleCalculation.h"
#include "ArbitraryAxisRotation.h"
#include "DiffractionItem.h"

#include <MPEGI/Renderer/PerceptualPointDistribution.h>

#include <algorithm>
#include <fstream>
#include <memory>

namespace {

void updateEdge(MPEGI::DiffractionEdge& edge, const EAR::Mesh* mesh)
{
    const auto& p0 = mesh->getVertex(edge.vtxIDs.first, EAR::Mesh::CSpace::Global);
    const auto& p1 = mesh->getVertex(edge.vtxIDs.second, EAR::Mesh::CSpace::Global);

    // re-construct edge vertex coordinates
    edge.vtxCoords = std::make_pair(p0, p1);

    // re-construct edge direction
    edge.edgeVector = p1 - p0;

    // re-construct normalized edge direction
    edge.normalizedEdgeDirection = edge.edgeVector.normalize();

    // re-construct edge length
    edge.edgeLength = edge.edgeVector.abs();

    const auto& face0 = mesh->getTriangle(edge.adjacentTriIDs.first);
    const auto& face1 = mesh->getTriangle(edge.adjacentTriIDs.second);

    // re-construct adjacent triangles
    edge.adjTris = std::make_pair(&face0, &face1);

    auto normal0 = face0.getNormal(EAR::Mesh::CSpace::Global).normalize();
    auto normal1 = face1.getNormal(EAR::Mesh::CSpace::Global).normalize();

    // re-construct adjacent triangle normals
    edge.adjacentTriNormals = std::make_pair(normal0, normal1);

    const auto center0 = face0.getCentroid(EAR::Mesh::CSpace::Global);
    const auto center1 = face1.getCentroid(EAR::Mesh::CSpace::Global);

    // re-construct adjacent triangle centroids
    edge.adjacentTriCentroids = std::make_pair(center0, center1);

    // re-construct edge angle
    float edgeAngleAbs = 180.f - acos(normal0.dot(normal1)) * 180. / M_PI;
    float d0 = normal0.dot(center0);
    float center1_plane0_dist = normal0.dot(center1) - d0; // use Hesse normal form
    float edgeAngleSgn = copysign(edgeAngleAbs, center1_plane0_dist);
    edge.edgeAngle = fmod(edgeAngleSgn + 360.f, 360.f); // expected value range: 0..360

    // re-construct edge center point; formally computed by edge.getEdgeCenterLocation(1e-3);
    edge.cp = 0.5f * (p0 + p1);

    // re-construct edge center point for visibility test; formally computed by edge.getEdgeCenterLocation(1e-3);
    float scale = 1e-3;
    auto normal_mean = (normal0 + normal1).normalize();
    edge.cp_vt = edge.cp + scale * normal_mean;
}

} // namespace

namespace MPEGI {

std::shared_ptr<RenderStage> makeDiffractionStage()
{
    return std::make_shared<DiffractionStage>();
}

DiffractionStage::DiffractionStage()
{
    // Set config variables
    Config::setDef("DiffractionItemNumber", EAR::Variant(10));
    Config::setDef("DiffractionFastSpeed", EAR::Variant(5.f));
    Config::setDef("DiffractionUpdateRate", EAR::Variant(10));
    Config::setDef("DiffractionFilterTuneParameter", EAR::Variant(0.5f));

    Config::setDef("DiffractionSourceExtentGainNSample", EAR::Variant(10));
    Config::setDef("DiffractionRegularizationAngle", EAR::Variant(5.f));
    Config::setDef("DiffractionSourceExtentGain", EAR::Variant(0.25f));
    Config::setDef("DiffractionSESSGain", EAR::Variant(1.5f));

    Config::setDef("DiffractionPathSelection", EAR::Variant(false));
    Config::setDef("DiffractionLinearPanning", EAR::Variant(false));

    Config::setDef("DiffractionCullingGain", EAR::Variant(0.1f));

    // Distance threshold for regarding two points as the same
    geoErrThreshold.store(
        Config::getFloatDef("Diffraction_GeometricErrorThreshold", 0.001f)); // [m]
    Config::observeKey("Diffraction_GeometricErrorThreshold", this);

    // Distance culling threshold
    cullingDist.store(Config::getFloatDef("DiffractionDistanceCullingThreshold", 64.f)); // [m]
    Config::observeKey("DiffractionDistanceCullingThreshold", this);

    Config::setDef("VisualizeDiffr", false);
}

DiffractionStage::~DiffractionStage()
{
    Config::removeObserver("Diffraction_GeometricErrorThreshold", this);
    Config::removeObserver("DiffractionDistanceCullingThreshold", this);
}

void DiffractionStage::configValueChanged(const std::string& key)
{
    if (key == "Diffraction_GeometricErrorThreshold") {
        geoErrThreshold.store(Config::getFloat("Diffraction_GeometricErrorThreshold"));
        return;
    }
    if (key == "DiffractionDistanceCullingThreshold") {
        cullingDist.store(Config::getFloat("DiffractionDistanceCullingThreshold"));
        return;
    }
}

void DiffractionStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    // Load payloads
    auto diffrPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_DIFFR);
    EAR_CHECK_REL(diffrPayloads.size() <= 1, "Too many diffraction payloads in bitstream");
    if (diffrPayloads.size() != 1) {
        // No diffraction payload, stage stays deactivated
        return;
    }
    std::shared_ptr<const MPEGI::DiffractionPayload> diffrPayload =
        std::dynamic_pointer_cast<const MPEGI::DiffractionPayload>(diffrPayloads[0]);
    eifPayload = std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE));

    // Initialize metadata
    diffrMetadata = std::make_shared<MPEGI::DiffractionMetadata>(std::move(diffrPayload));

    if (!state) {
        return;
    }

    this->itemStore = itemStore;

    listener = std::make_unique<Listener>();

    // Calculate intersection test samples
    calculateIntersectionTestSamples(state);

    auto isMeshUpdated = [&](const std::string& meshId) {
        for (auto& upd : bitstream.getUpdates()) {
            for (auto& mod : upd.modifications) {
                if (mod.id == meshId) {
                    return true;
                }
            }
        }
        return false;
    };

    const auto& combinedMesh = state->getCombinedDiffractionMesh();
    if (combinedMesh) {
        // re-construct edge data
        EAR::Mesh* mesh = combinedMesh.get();
        updateStaticEdges(mesh);

        meshes.push_back(combinedMesh.get());
    }

    const auto& tempGeometries = state->getAllVisibleGeometry();
    for (const auto& geo : tempGeometries) {
        const EAR::Mesh* mesh = geo->getMesh();
        if (!mesh || mesh->isStatic()) {
            continue;
        }

        // update meshes to initialize Embree tracer
        if (mesh->getParent()) {
            if (isMeshUpdated(mesh->getParent()->getName())) {
                continue;
            }
        }

        // re-construct edge data
        updateDynamicEdges(mesh);

        dynmeshes.push_back(mesh);

        // attach the non-static/dynamic geometries to geometryObservers
        geometryObservers.emplace_back(geo, this);
        isDynamicGeometryIncluded = true;
    }

    if (meshes.size() > 0) {
        // Tracer instantiated
        tracer = std::make_unique<EAR::EmbreeTracer>(
            meshes, true); // second argument sets tracer to ignore backface hits

        // Tracer instantiated
        dynTracer = std::make_unique<EAR::EmbreeTracer>(
            dynmeshes, true); // second argument sets tracer to ignore backface hits
    }

    // Reset diffraction-relevant member variables
    isDynamicGeometryRelocated =
        true; // let the update thread initialize the scene with the initial input data

    diffrItemInitialEQs.assign(Config::getInt("EQNBands"), 1.f);

    B = streamManager->getB();
    Fs = streamManager->getFs();

    // All dynamic sources are potentially relevant for diffraction.
    // The relevant static sources will be obtained from this->diffrPayload.
    for (const auto& s : eifPayload->getObjectSources()) {
        if (!s.isStatic) {
            dynamicObjectSources.insert(s.bsId);
        }
    }
}

std::unique_ptr<RenderStage::DspData> DiffractionStage::update(RenderList& renderList)
{
    std::lock_guard<std::mutex> guard(edgeListAccess);
    if (!tracer || (diffrMetadata->getEdgeList().size() == 0))
        return {};

    // Set update rate for complete path search and item counter
    unsigned int updateRate = Config::getInt("DiffractionUpdateRate");
    unsigned int itemCounter = 0;

    // Update listener
    listener->update(renderList.listener);
    for (auto& item : renderList.items) {
        if ((item->type != ItemType::Primary) && (item->type != ItemType::Representative)) {
            // Primary type includes point source and source extent currently
            continue;
        }

        if (item->position.cSpace == CSpace::LCS) {
            continue;
        }

        if (item->parentRI != -1) {
            continue;
        }

        auto& [secondaryItems] = item->getStateOrCreate<DiffractionStage::SecondaryItems>();

        // Remove diffraction items if the primary item becomes inactive
        if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned) {
            if (secondaryItems.size() > 0) {
                for (auto& secItem : secondaryItems) {
                    // try to remove secondary items until fade out is complete
                    itemStore->remove(secItem.second);
                }
            }
            continue;
        }

        // Initialize diffraction paths
        if (diffractionPaths.find(item->idx) != diffractionPaths.end()) {
            renderList.diffractionPaths.insert({ item->idx, diffractionPaths.at(item->idx) });
        } else {
            diffractionPaths.insert({ item->idx, std::vector<std::vector<EAR::Vector3>>() });
            renderList.diffractionPaths.insert(
                { item->idx, std::vector<std::vector<EAR::Vector3>>() });
        }

        auto& [source, rtDiffractionPathTracker] =
            item->getStateOrCreate<DiffractionStage::ItemState>(
                listener.get(), diffrMetadata, eifPayload);

        source.update(*item, B, Fs);
        if (!(source.sourceRelocated || listener->listenerRelocated
                || isDynamicGeometryRelocated)) { //
            continue;
        }

        // Check the direct path's occlusion status
        if (item->extent != nullptr) {
            bool isOccluded = false;
            for (auto& rItem : renderList.items) {
                if (rItem->type != ItemType::Representative)
                    continue;

                if ((rItem->objectName.find(item->objectName) != std::string::npos)
                    && (rItem->objectName.find("[extended&occluded]") != std::string::npos)) {

                    isOccluded = true;
                    break;
                }
            }

            source.isOcclusionStateChanged = (source.isPreviouslyOccluded != isOccluded);

            if (!isOccluded) {
                if (secondaryItems.size() > 0) {
                    for (auto& secItem : secondaryItems) {
                        // try to remove secondary items until fade out is complete
                        itemStore->remove(secItem.second);
                    }
                }

                // Update source's occluded state
                source.isPreviouslyOccluded = false;

                // Reset diffraction paths
                renderList.diffractionPaths.at(item->idx).clear();
                diffractionPaths.at(item->idx).clear();
                continue;
            } else {
                source.isPreviouslyOccluded = true;
            }
        } else if (item->occlusionInfo == std::nullopt) {
            if (secondaryItems.size() > 0) {
                for (auto& secItem : secondaryItems) {
                    // try to remove secondary items until fade out is complete
                    itemStore->remove(secItem.second);
                }
            }

            // Update source's occluded state
            source.isPreviouslyOccluded = false;

            // Reset diffraction paths
            diffractionPaths.at(item->idx).clear();
            renderList.diffractionPaths.at(item->idx).clear();

            continue;
        } else {
            // Update source's occluded state and changed flag which will be used to activate complete path search:
            // when a source starts to be occluded, a diffraction item should be promptly created accordingly
            source.isOcclusionStateChanged = (source.isPreviouslyOccluded != true);
            source.isPreviouslyOccluded = true;
        }

        // Resolve object name -> bitstream ID
        const std::string& srcName = source.sourceObjectName;
        int srcBitsId = -1;
        try {
            srcBitsId = diffrMetadata->getGeoBitsId(srcName);
        } catch (const std::out_of_range& err) {
            // skip unknown sources
            continue;
        }

        // Skip static sources which are not in the diffraction payload
        const auto& staticSrcIds = diffrMetadata->getStaticSourceIdxList();
        auto staticSrcIt = std::find(staticSrcIds.begin(), staticSrcIds.end(), srcBitsId);
        if (staticSrcIt == staticSrcIds.end()) {
            auto dynamicSrcIt =
                std::find(dynamicObjectSources.begin(), dynamicObjectSources.end(), srcBitsId);
            if (dynamicSrcIt == dynamicObjectSources.end()) {
                continue;
            }
        }

        if (item->extent) {
            // Initialize diffraction item's eq filter for source extent
            for (auto& rItem : renderList.items) {
                if (rItem->type != ItemType::Representative)
                    continue;
                if ((rItem->objectName.find(item->objectName) != std::string::npos)
                    && (rItem->objectName.find("[extended&occluded]") != std::string::npos)) {
                    for (unsigned int i = 0; i < Config::getInt("EQNBands"); i++) {
                        diffrItemInitialEQs[i] = std::fmax(1.f - rItem->eq[i], 0.f);
                    }
                    break;
                }
            }

            //! Use ray bundles to find occluded region center position and update diffraction gain
            updateOccludedLocationAndEqWithRay(item->objectName, source);

        } else {
            // Initialize diffraction item's eq filter for point source
            for (int i = 0; i < item->eq.size(); i++) {
                diffrItemInitialEQs[i] = std::fmax(1.0f - item->eq[i], 0.f);
            }
        }

        // It is implied that the primary item's occluded
        if ((updateCounter + itemCounter) % updateRate == 0 || source.isOcclusionStateChanged) {

            // Reset the (diffraction) paths and items per renderItem
            rtDiffractionPathTracker.resetPaths();
            rtDiffractionPathTracker.resetDiffractionItems();

            // Potential static path with or without occlusion by dynamic geometries
            updateValidPathIndexList(item->objectName, source, srcBitsId);
            rtDiffractionPathTracker.updatePathsToValidate(source.pathIndexList);

            // Brute force search for static sources to overcome coarse voxel grid resolution
            if (rtDiffractionPathTracker.getNofPaths() == 0) {
                auto it = std::find(diffrMetadata->getStaticSourceIdxList().begin(),
                    diffrMetadata->getStaticSourceIdxList().end(), srcBitsId);

                if (it != diffrMetadata->getStaticSourceIdxList().end()) {
                    runBruteForceSearchForStaticSources(source, srcBitsId);
                }

                rtDiffractionPathTracker.updatePathsToValidate(source.pathIndexList);
            }

            if (isDynamicGeometryIncluded) {
                rtDiffractionPathTracker.updateOccludedPathData(dynTracer);
            } else {
                rtDiffractionPathTracker.activateAllPathsToValidate();
            }

            // It is implied that all validated paths should be activated
            rtDiffractionPathTracker.selectPaths();
            rtDiffractionPathTracker.updateStaticDiffractionItems();
            rtDiffractionPathTracker.updateOccludedStaticDiffractionItems(tracer);

            if (Config::getBool("VisualizeDiffr")) {
                // Reset renderList's diffraction paths for visualization
                diffractionPaths.at(item->idx) =
                    updatePathsToVisualize(source, rtDiffractionPathTracker);
                renderList.diffractionPaths.at(item->idx) = diffractionPaths.at(item->idx);
            }

            // Potential dynamic paths with or without occlusion by static geometries
            if (isDynamicGeometryIncluded
                && source.speed < Config::getFloat("DiffractionFastSpeed")) {
                for (const auto& dynMesh : dynmeshes) {
                    const std::string& dynMeshName = dynMesh->getName();
                    int dynMeshBitsId = diffrMetadata->getGeoBitsId(dynMeshName);
                    const std::vector<std::pair<int, int>> dynamicPathIndexList =
                        validPathIndexList2(source, dynMeshBitsId);

                    if (dynamicPathIndexList.size() != 0) {
                        rtDiffractionPathTracker.updateDynamicPathsToValidate(
                            dynamicPathIndexList, dynMeshBitsId);
                        rtDiffractionPathTracker.updateOccludedDynamicPathData(
                            dynTracer, dynamicPathIndexList, dynMeshBitsId);
                        rtDiffractionPathTracker.updateDynamicDiffractionItems();
                    }
                }
            }
            rtDiffractionPathTracker.updateItemWeights();
        } else {
            renderList.diffractionPaths.at(item->idx) = diffractionPaths.at(item->idx);
        }

        // When complete path search is skipped, then previously stored diffraction items can be re-used to generate diffraction items
        // Create/remove secondary items per source
        auto& diffractionItems = rtDiffractionPathTracker.getDiffractionItems();
        std::vector<std::size_t> activeHashes;
        for (auto diffrItem : diffractionItems) {
            activeHashes.push_back(diffrItem->getID());
            bool itemOrphaned = false;
            auto it = secondaryItems.find(diffrItem->getID());
            const bool itemFound = (it != secondaryItems.end());
            if (itemFound && it->second->status == ItemStatus::Orphaned) {
                secondaryItems.erase(it); //invalidates it!!
                itemOrphaned = true;
            }
            if (!itemFound || itemOrphaned) {
                // create a secondary item for each path if it doesn't exist
                secondaryItems[diffrItem->getID()] = itemStore->createItem(
                    item->objectName + "_diffr_" + std::to_string(diffrItem->getID()));
                secondaryItems[diffrItem->getID()]->type = ItemType::Diffraction;
                secondaryItems[diffrItem->getID()]->parentRI = item->idx;
                secondaryItems[diffrItem->getID()]->isSoundEffect = item->isSoundEffect;
            }
        }

        for (auto& secItem : secondaryItems) {
            if (std::find(activeHashes.begin(), activeHashes.end(), secItem.first)
                == activeHashes.end()) {
                // mark secondary item as obsolete if the corresponding diffraction path no longer exists
                itemStore->remove(secItem.second);
            }
        }

        // Update secondary items per source
        for (auto& secItem : secondaryItems) {
            auto it = std::find_if(diffractionItems.begin(), diffractionItems.end(),
                [hash = secItem.first](const auto item) { return item->getID() == hash; });

            if (it == diffractionItems.end()) {
                continue; // this item is deprecated and does not need to be updated anymore
            }

            secItem.second->sync(*item,
                ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Loudness
                    | ItemProperty::Eq | ItemProperty::Aparams | ItemProperty::Position
                    | ItemProperty::RefDist | ItemProperty::Extent | ItemProperty::Reverb
                    | ItemProperty::Signal);

            if (secItem.second->status != ItemStatus::New) {
                secItem.second->sync(*item, ItemProperty::Status);
            }

            secItem.second->addGain((*it)->getGain());
            secItem.second->setEq(diffrItemInitialEQs);
            secItem.second->addEq((*it)->getDiffractionFilterBufferPointer());

            Position positionToUpdate = secItem.second->position;
            positionToUpdate.location = (*it)->getSourceRotatedLocation();
            positionToUpdate.orientation = (*it)->getSourceRotatedOrientation();

            // Force diffraction items to have the same distance as the one of the direct path (for No Doppler)
            float directPathDistance = (listener->listenerLocation - item->position.location).abs();
            Vector3 relativeLocation = positionToUpdate.location - listener->listenerLocation;
            relativeLocation *= directPathDistance / relativeLocation.abs();
            positionToUpdate.location = listener->listenerLocation + relativeLocation;

            secItem.second->setPosition(positionToUpdate);
        }

        itemCounter++;
    }
    updateCounter++;

    // add secondary render items to RenderList
    for (auto& state : itemStore->getItemStates<DiffractionStage::SecondaryItems>()) {
        //iterate over all child render items
        for (auto it = state.secondaryItems.begin(); it != state.secondaryItems.end();) {
            auto secItem = (*it).second;
            if (secItem->status == ItemStatus::FadeOutInactive) {
                itemStore->deactivate(
                    secItem); // try deactivation until all fade outs are completed
            } else if (secItem->status == ItemStatus::FadeOutOrphaned) {
                itemStore->remove(secItem); // try removal until all fade outs are completed
            }

            renderList.addItemCopy(secItem);
            secItem->resetChanged();

            // remove secondary item if orphaned fade out complete
            if (secItem->status == ItemStatus::Orphaned) {
                it = state.secondaryItems.erase(it); // returns next in line
            } else {
                it++;
            }
        }
    }

    isDynamicGeometryRelocated = false;
    return {};
}

void DiffractionStage::updateDynamicEdges(const EAR::Geometry* geometry)
{
    std::lock_guard<std::mutex> guard(edgeListAccess);

    EAR_CHECK_REL((geometry->isStatic() != true),
        "Exception: DiffractionStage::updateDynamicEdges, given geometry is static");

    const std::string& meshName = geometry->getName();
    int meshBitsId = diffrMetadata->getGeoBitsId(meshName);
    auto itr = diffrMetadata->getDynamicEdgeLists().find(meshBitsId);
    EAR_CHECK_REL((itr != diffrMetadata->getDynamicEdgeLists().end()),
        "EAR Exception: given dynamic mesh bitstream id is missing - updateDynamicEdges");

    auto mesh = geometry->getMesh();
    for (auto& edge : itr->second) {
        updateEdge(edge, mesh);
    }

    isDynamicGeometryRelocated = true;
}

void DiffractionStage::updateStaticEdges(const EAR::Geometry* geometry)
{
    std::lock_guard<std::mutex> guard(edgeListAccess);

    EAR_CHECK_REL((geometry->isStatic() == true),
        "Exception: DiffractionStage::updateStaticEdges, given geometry is dynamic");

    auto& edgeList = diffrMetadata->getEdgeList();

    auto mesh = geometry->getMesh();
    for (auto& edge : edgeList) {
        updateEdge(edge, mesh);
    }

    isDynamicGeometryRelocated = true;
}

void DiffractionStage::updateValidPathIndexList(std::string srcName, Source& source, int srcBitsId)
{
    bool isDynamicSourceInVR = false;
    if (eifPayload->getSceneInfo().type == EIFPayload::SceneInfo::SceneType::VR) {
        auto it = std::find(diffrMetadata->getStaticSourceIdxList().begin(),
            diffrMetadata->getStaticSourceIdxList().end(), srcBitsId);

        if (it == diffrMetadata->getStaticSourceIdxList().end())
            isDynamicSourceInVR = true;
    }

    // For all sources in AR
    if (eifPayload->getSceneInfo().type == EIFPayload::SceneInfo::SceneType::AR) {

        bool updateSrcEdgeIdList =
            (source.visibleEdgeIdList.size() == 0) || (source.sourceRelocated);

        source.visibleEdgeIdList.clear();
        source.visibleEdgeIdList =
            bruteForceEdgeSearch(source.sourceLocation, EAR::Vector3(), false);

        if (!listener->isUpdated) {
            listener->visibleEdgeIdList =
                bruteForceEdgeSearch(listener->listenerLocation, EAR::Vector3(), false);
            listener->isUpdated = true;
        }

        source.pathIndexList.clear();
        for (auto seId : source.visibleEdgeIdList) {
            EAR_CHECK_REL((diffrMetadata->getPathDictByEdgeIds().find(seId)
                              != diffrMetadata->getPathDictByEdgeIds().end()),
                "EAR Exception: the corresponding path dictionary is missing in AR");

            auto& pathList = diffrMetadata->getPathDictByEdgeIds().at(seId);
            for (auto leId : listener->visibleEdgeIdList) {
                int localIndex = 0;
                for (const auto eIdSeq : pathList) {
                    if (eIdSeq.back() == leId) {
                        source.pathIndexList.push_back(std::make_pair(seId, localIndex));
                    }
                    localIndex++;
                }
            }
        }
    } else if (isDynamicSourceInVR) {
        //! Requires high computation due to dynamic motion at run-time
        bool updateSrcEdgeIdList =
            (source.visibleEdgeIdList.size() == 0) || (source.sourceRelocated);

        if (updateSrcEdgeIdList) {
            auto isFast = source.speed > Config::getFloat("DiffractionFastSpeed");
            if (!source.sourceRelocated && source.visibleEdgeIdList.size() == 0) {
                source.visibleEdgeIdList =
                    bruteForceEdgeSearch(source.sourceLocation, listener->listenerLocation, isFast);
            } else {
                source.visibleEdgeIdList.clear();

                // source's visible edge list update using reachable voxel data
                source.visibleEdgeIdList =
                    visibleEdgeListFromReachableVoxels(source.sourceLocation);

                // if no information from voxels, run brute-force searching
                if (source.visibleEdgeIdList.size() == 0) {
                    source.visibleEdgeIdList = bruteForceEdgeSearch(
                        source.sourceLocation, listener->listenerLocation, isFast);
                }
            }
        }

        if (source.visibleEdgeIdList.size() == 0) {
            // return stored path index from the last process
            return;
        }

        if (listener->listenerRelocated) {
            listener->visibleEdgeIdList.clear();

            // listener visible edge list update using reachable voxel datda
            if (!listener->isUpdated) {
                listener->visibleEdgeIdList =
                    visibleEdgeListFromReachableVoxels(listener->listenerLocation);
                listener->isUpdated = true;

                // If still none of edges are visible from the listener
                if (listener->visibleEdgeIdList.size() == 0) {
                    visibleListenerEdgeList(listener->listenerLocation, source.visibleEdgeIdList);
                    listener->isUpdated = false;
                }
            }
        }

        if (updateSrcEdgeIdList || listener->listenerRelocated) {
            source.pathIndexList.clear();
            for (auto seId : source.visibleEdgeIdList) {
                EAR_CHECK_REL((diffrMetadata->getPathDictByEdgeIds().find(seId)
                                  != diffrMetadata->getPathDictByEdgeIds().end()),
                    "EAR Exception: the corresponding path dictionary is missing in VR with "
                    "moving "
                    "sources");

                auto& pathListByEdgeIds = diffrMetadata->getPathDictByEdgeIds().at(seId);

                if (source.speed > 0.f) {
                    // Up to 2nd order diffraction paths are allowed
                    auto iterPathIndexDict =
                        diffrMetadata->getStaticPathIndexListPerEdge2().find(seId);

                    for (auto pIndex : iterPathIndexDict->second) {
                        if (std::find(
                                source.pathIndexList.begin(), source.pathIndexList.end(), pIndex)
                            != source.pathIndexList.end())
                            continue;

                        for (auto leId : listener->visibleEdgeIdList) {
                            if (pathListByEdgeIds[pIndex.second].back() == leId) {
                                source.pathIndexList.push_back(pIndex);
                            }
                        }
                    }
                } else {
                    auto iterPathIndexDict =
                        diffrMetadata->getStaticPathIndexListPerEdge1().find(seId);

                    for (auto pIndex : iterPathIndexDict->second) {
                        if (std::find(
                                source.pathIndexList.begin(), source.pathIndexList.end(), pIndex)
                            != source.pathIndexList.end())
                            continue;

                        for (auto leId : listener->visibleEdgeIdList) {
                            if (pathListByEdgeIds[pIndex.second].back() == leId) {
                                source.pathIndexList.push_back(pIndex);
                            }
                        }
                    }
                }
            }
        }

    } else {

        auto lvIndex = pointToVoxelCoordinate(listener->listenerLocation);
        const auto& pathDict = diffrMetadata->getValidPathDict().at(srcBitsId);

        EAR::Vector3 posA = listener->listenerLocation;
        source.pathIndexList.clear();
        auto threshold = Config::getFloat("Diffraction_GeometricErrorThreshold");
        for (int dx = -5; dx < 6; dx += 1) {
            for (int dy = -5; dy < 6; dy += 1) {
                for (int dz = -5; dz < 6; dz += 1) {
                    uint64_t lvHashNeighbor =
                        PackVoxelCoords(lvIndex[0] + dx, lvIndex[1] + dy, lvIndex[2] + dz);

                    const auto lvHashNeighborIter = pathDict.find(lvHashNeighbor);
                    if (lvHashNeighborIter != pathDict.end()) {
                        for (const auto& pIndex : lvHashNeighborIter->second) {
                            if (diffrMetadata->getEdgeList()[pIndex.first].relevant == false)
                                continue;

                            if (std::find(source.pathIndexList.begin(), source.pathIndexList.end(),
                                    pIndex)
                                != source.pathIndexList.end())
                                continue;

                            // Check if the end point of the path is visible to current listener position
                            auto& edgeSequence =
                                diffrMetadata->getPathDictByEdges().at(pIndex.first)[pIndex.second];

                            auto& e = edgeSequence.back();
                            if (e->adjacentTriNormals.first.dot(posA - e->cp)
                                    * e->adjacentTriNormals.second.dot(posA - e->cp)
                                > 0)
                                continue;

                            std::vector<EAR::Vector3> samples = { e->cp_vt, e->vtxCoords.first,
                                e->vtxCoords.second };
                            bool isVisible = false;
                            for (auto& posB : samples) {
                                if (!tracer->isOccluded(posA, posB - posA, threshold)) {
                                    isVisible = true;
                                    break;
                                }
                            }

                            if (isVisible) {
                                source.pathIndexList.push_back(pIndex);
                            }
                        }
                    }
                }
            }
        }
    }

    std::sort(source.pathIndexList.begin(), source.pathIndexList.end());
    auto pIndexLast = std::unique(source.pathIndexList.begin(), source.pathIndexList.end());
    source.pathIndexList.erase(pIndexLast, source.pathIndexList.end());

    return;
}

std::vector<std::pair<int, int>> DiffractionStage::validPathIndexList2(
    Source& source, int dynMeshBitsId)
{
    auto itr = diffrMetadata->getDynamicEdgeLists().find(dynMeshBitsId);
    EAR_CHECK_REL((itr != diffrMetadata->getDynamicEdgeLists().end()),
        "EAR Exception: given dynamic mesh bitstream id is missing - validPathIndexList2");

    std::vector<int> sdynEdgeIdList;
    auto threshold = Config::getFloat("Diffraction_GeometricErrorThreshold");
    for (auto& e : itr->second) {
        EAR::Vector3 posA = source.sourceLocation;
        EAR::Vector3 posB = e.cp_vt;

        if (e.adjacentTriNormals.first.dot(posA - e.cp)
                * e.adjacentTriNormals.second.dot(posA - e.cp)
            > 0)
            continue;

        std::vector<EAR::Vector3> samples = { e.cp_vt, e.vtxCoords.first, e.vtxCoords.second };
        bool isVisible = false;
        for (auto& posB : samples) {
            if (!tracer->isOccluded(posA, posB - posA, threshold)) {
                isVisible = true;
                break;
            }
        }

        if (isVisible)
            sdynEdgeIdList.push_back(e.edgeID);
    }

    std::vector<int> ldynEdgeIdList;
    for (auto e : itr->second) {
        EAR::Vector3 posA = listener->listenerLocation;
        EAR::Vector3 posB = e.cp_vt;
        if (e.adjacentTriNormals.first.dot(posA - e.cp)
                * e.adjacentTriNormals.second.dot(posA - e.cp)
            > 0)
            continue;

        std::vector<EAR::Vector3> samples = { e.cp_vt, e.vtxCoords.first, e.vtxCoords.second };
        bool isVisible = false;
        for (auto& posB : samples) {
            if (!tracer->isOccluded(posA, posB - posA, threshold)) {
                isVisible = true;
                break;
            }
        }

        if (isVisible)
            ldynEdgeIdList.push_back(e.edgeID);
    }

    std::vector<std::pair<int, int>> pIndexList;
    for (auto seId : sdynEdgeIdList) {
        for (auto leId : ldynEdgeIdList) {

            int localIndex = 0;
            for (const auto eIdSeq :
                diffrMetadata->getDynamicPathDictByEdgeIds(dynMeshBitsId).at(seId)) {
                if (eIdSeq.back() == leId) {
                    pIndexList.push_back(std::make_pair(seId, localIndex));
                }
                localIndex++;
            }
        }
    }

    return pIndexList;
}

VoxelCoord DiffractionStage::pointToVoxelCoordinate(const EAR::Vector3& pos) const
{
    const EAR::Vector3& origin = diffrMetadata->getVoxelGridOrigin();
    const EAR::Vector3& pitch = diffrMetadata->getVoxelGridPitch();
    const VoxelCoord& shape = diffrMetadata->getVoxelGridShape();

    VoxelCoord vc;
    EAR::Vector3 shifted = pos - origin;
    vc[0] = int(shifted[0] / pitch[0] + 0.5);
    vc[1] = int(shifted[1] / pitch[1] + 0.5);
    vc[2] = int(shifted[2] / pitch[2] + 0.5);

    if ((vc[0] < 0) || (vc[0] >= shape[0]) || (vc[1] < 0) || (vc[1] >= shape[1]) || (vc[2] < 0)
        || (vc[2] >= shape[2])) {
        vc = VoxelCoord(-1, -1, -1);
    }

    return vc;
}

std::vector<int> DiffractionStage::visibleEdgeListFromReachableVoxels(
    EAR::Vector3 posA, int srcBitsId)
{
    std::vector<int> edgeIdList;

    // Voxel: 0.5 x 0.5 x 0.5
    auto vIndex = pointToVoxelCoordinate(posA);
    std::vector<int> svKey = { vIndex[0], vIndex[1], vIndex[2] };

    if (srcBitsId != -1) {
        auto svParentDataItr = diffrMetadata->getSourceVoxelDicts().find(srcBitsId);
        if (svParentDataItr == diffrMetadata->getSourceVoxelDicts().end()) {
            return edgeIdList;
        }

        auto svDataItr = svParentDataItr->second.find(svKey);
        if (svDataItr == svParentDataItr->second.end()) {
            return edgeIdList;
        }

        edgeIdList = svDataItr->second;
    } else {
        auto threshold = Config::getFloat("Diffraction_GeometricErrorThreshold");
        for (int dx = -1; dx < 2; dx++) {
            for (int dy = -1; dy < 2; dy++) {
                for (int dz = -1; dz < 2; dz++) {

                    std::vector<int> vKeydr = { svKey[0] + dx, svKey[1] + dy, svKey[2] + dz };
                    auto svDataItr = diffrMetadata->getListenerVoxelDict().find(vKeydr);
                    if (svDataItr == diffrMetadata->getListenerVoxelDict().end()) {
                        continue;
                    }

                    for (auto eid : svDataItr->second) {
                        if (std::find(edgeIdList.begin(), edgeIdList.end(), eid)
                            != edgeIdList.end()) {
                            continue;
                        }

                        auto& e = diffrMetadata->getEdgeList()[eid];
                        if (e.adjacentTriNormals.first.dot(posA - e.cp)
                                * e.adjacentTriNormals.second.dot(posA - e.cp)
                            > 0)
                            continue;

                        std::vector<EAR::Vector3> samples = { e.cp_vt, e.vtxCoords.first,
                            e.vtxCoords.second };

                        bool isVisible = false;
                        for (auto& posB : samples) {
                            if (!tracer->isOccluded(posA, posB - posA, threshold)) {
                                isVisible = true;
                                break;
                            }
                        }

                        if (isVisible) {
                            edgeIdList.push_back(e.edgeID);
                        }
                    }
                }
            }
        }
    }

    return edgeIdList;
}

std::vector<int> DiffractionStage::bruteForceEdgeSearch(
    EAR::Vector3 posA, EAR::Vector3 posB, bool isFast)
{
    std::vector<int> edgeIdList;
    auto threshold = Config::getFloat("Diffraction_GeometricErrorThreshold");
    for (const auto& e : diffrMetadata->getEdgeList()) {
        if (!isFast) {
            if ((posA - e.cp).abs() > cullingDist.load()) {
                continue;
            }
        } else {
            if ((posA - e.cp).abs() > (posA - posB).abs()) {
                continue;
            }
        }

        if (e.adjacentTriNormals.first.dot(posA - e.cp)
                * e.adjacentTriNormals.second.dot(posA - e.cp)
            > 0)
            continue;

        std::vector<EAR::Vector3> samples = { e.cp_vt, e.vtxCoords.first, e.vtxCoords.second };

        bool isVisible = false;
        for (auto& posB : samples) {
            if (!tracer->isOccluded(posA, posB - posA, threshold)) {
                isVisible = true;
                break;
            }
        }

        if (isVisible) {
            edgeIdList.push_back(e.edgeID);
        }
    }

    return edgeIdList;
}

void DiffractionStage::visibleListenerEdgeList(
    EAR::Vector3 posA, std::vector<int>& sourceEdgeIdList)
{
    auto threshold = Config::getFloat("Diffraction_GeometricErrorThreshold");
    for (auto seId : sourceEdgeIdList) {
        if (diffrMetadata->getVisibleEdgeDictionaryForListener().find(seId)
            == diffrMetadata->getVisibleEdgeDictionaryForListener().end())
            continue;

        auto& potentialVisibleEdgeIdsForListener =
            diffrMetadata->getVisibleEdgeDictionaryForListener().at(seId);

        for (const auto& eId : potentialVisibleEdgeIdsForListener) {
            if (std::find(
                    listener->visibleEdgeIdList.begin(), listener->visibleEdgeIdList.end(), eId)
                != listener->visibleEdgeIdList.end())
                continue;

            auto& e = diffrMetadata->getEdgeList()[eId];
            if (e.adjacentTriNormals.first.dot(posA - e.cp)
                    * e.adjacentTriNormals.second.dot(posA - e.cp)
                > 0)
                continue;

            std::vector<EAR::Vector3> samples = { e.cp_vt, e.vtxCoords.first, e.vtxCoords.second };

            bool isVisible = false;
            for (auto& posB : samples) {
                if (!tracer->isOccluded(posA, posB - posA, threshold)) {
                    isVisible = true;
                    break;
                }
            }

            if (isVisible) {
                listener->visibleEdgeIdList.push_back(e.edgeID);
            }
        }
    }
}

std::vector<std::vector<EAR::Vector3>> DiffractionStage::updatePathsToVisualize(
    Source& source, RTDiffractionPathTracker& rtDiffractionPathTracker)
{
    // Update static diffraction paths using edge center locations for visualization
    std::vector<std::vector<EAR::Vector3>> paths;
    for (const auto& diffItem : rtDiffractionPathTracker.getDiffractionItems()) {
        auto edgeSequence = diffrMetadata->getPathDictByEdges().at(
            diffItem->getPathIncluded()
                ->initialEdgeID)[diffItem->getPathIncluded()->localPathIndex];

        std::vector<EAR::Vector3> edgeLocations;
        edgeLocations.push_back(source.sourceLocation);
        for (auto& e : edgeSequence) {
            edgeLocations.push_back(e->cp_vt);
        }
        edgeLocations.push_back(listener->listenerLocation);
        paths.push_back(edgeLocations);
    }

    return paths;
}

void DiffractionStage::runBruteForceSearchForStaticSources(Source& source, int srcBitsId)
{
    source.visibleEdgeIdList = visibleEdgeListFromReachableVoxels(source.sourceLocation, srcBitsId);

    if (!listener->isUpdated) {
        listener->visibleEdgeIdList.clear();
        visibleListenerEdgeList(listener->listenerLocation, source.visibleEdgeIdList);

        //! Listener's list is updated for this given source
        listener->isUpdated = false;
    }

    source.pathIndexList.clear();
    for (auto seId : source.visibleEdgeIdList) {
        EAR_CHECK_REL((diffrMetadata->getPathDictByEdgeIds().find(seId)
                          != diffrMetadata->getPathDictByEdgeIds().end()),
            "EAR Exception: the corresponding path dictionary is missing in VR with moving "
            "sources");

        auto iterPathIndexDict = diffrMetadata->getStaticPathIndexListPerEdge1().find(seId);
        auto& pathListByEdgeIds = diffrMetadata->getPathDictByEdgeIds().at(seId);
        for (auto leId : listener->visibleEdgeIdList) {
            for (auto pIndex : iterPathIndexDict->second) {
                if (pathListByEdgeIds[pIndex.second].back() == leId) {
                    source.pathIndexList.push_back(pIndex);
                }
            }
        }
    }

    std::sort(source.pathIndexList.begin(), source.pathIndexList.end());
    auto pIndexLast = std::unique(source.pathIndexList.begin(), source.pathIndexList.end());
    source.pathIndexList.erase(pIndexLast, source.pathIndexList.end());

    return;
}

void DiffractionStage::updateOccludedLocationAndEqWithRay(std::string itemName, Source& source)
{
    int nofRayHits = 0;
    int nofRays = intersectionTestSamples.at(itemName).size();
    EAR::Vector3 occludedSrcLocation(0.f, 0.f, 0.f);
    for (auto& sample : intersectionTestSamples.at(itemName)) {
        if (tracer->isOccluded(
                listener->listenerLocation, sample - listener->listenerLocation, 0.001)) {
            nofRayHits++;
            occludedSrcLocation += sample;
        }
    }

    if (nofRayHits != 0) {
        for (int i = 0; i < 3; i++) {
            occludedSrcLocation[i] /= nofRayHits;
        }

        source.sourceLocation = occludedSrcLocation;
    }

    float tempGain = 0;
    if (nofRays == 0) {
        tempGain = Config::getFloat("DiffractionSourceExtentGain");
    } else {
        tempGain = Config::getFloat("DiffractionSourceExtentGain")
            * std::fmax(1.0f, 2 * float(nofRayHits / nofRays));
    }

    if (gainHistoryMap.find(itemName) == gainHistoryMap.end()) {
        gainHistoryMap.insert(
            { itemName, std::vector<float>(Config::getInt("DiffractionSourceExtentGainNSample")) });
    }

    gainHistoryMap.at(itemName).push_back(tempGain);
    gainHistoryMap.at(itemName).erase(gainHistoryMap.at(itemName).begin() + 0);

    float sum = std::accumulate(
        gainHistoryMap.at(itemName).begin(), gainHistoryMap.at(itemName).end(), 0.f);
    float gain = sum / Config::getInt("DiffractionSourceExtentGainNSample");

    for (int i = 0; i < diffrItemInitialEQs.size(); i++) {
        diffrItemInitialEQs[i] *= gain;
    }
}

void DiffractionStage::calculateIntersectionTestSamples(const SceneState* state)
{
    std::vector<std::string> srcIds;
    std::vector<std::string> srcExtentIds;
    std::vector<EAR::Vector3> srcExtentLocations;
    //intersectionTestSamples;
    for (auto& src : eifPayload->getObjectSources()) {
        if (src.extent) {
            srcIds.push_back(src.id);
            srcExtentIds.push_back(src.extent->id);
            srcExtentLocations.push_back(src.position.location);
            intersectionTestSamples.insert({ src.id, std::vector<EAR::Vector3>() });
        }
    }

    int i = 0;
    for (const auto& srcExtent : srcExtentIds) {
        for (const auto& geo : state->getAllGeometry()) {
            const EAR::Mesh* mesh = geo->getMesh();
            if (srcExtent == mesh->getName()) {
                auto tracerSourceExtent = std::make_unique<EAR::EmbreeTracer>(mesh, false);
                for (auto& rayDirection : PERCEPTUAL_POINT_DISTR_512) {

                    // To take rays in account only when they are on the horizontal plane
                    if (abs(rayDirection[1]) > 0.1)
                        continue;

                    EAR::Vector3 intersection = rayDirection;
                    tracerSourceExtent->firstHit(
                        srcExtentLocations[i], EAR::cartesianToPolar(rayDirection), true);
                    float distance = tracerSourceExtent->getDistanceToLastHit();
                    for (int j = 0; j < 3; j++) {
                        intersection[j] *= distance;
                        intersection[j] += srcExtentLocations[i][j];
                    }
                    intersectionTestSamples.at(srcIds[i]).push_back(intersection);
                }
                break;
            }
        }
        i++;
    }
}

} // namespace MPEGI
