/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DIFFRACTION_STAGE
#define MPEGI_DIFFRACTION_STAGE

#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Helpers/Config.h>

#include "DiffractionMetadata.h"
#include <MPEGI/Bitstream/DiffractionPayload.h>

#include <EAR/Geometry.h>
#include <EAR/SceneObject.h>

#include "RTDiffractionPathTracker.h"

#include <cstddef>
#include <deque>
#include <memory>
#include <mutex>

// fwd declares
namespace EAR {
class EmbreeTracer;
class RealBuffer;
} // namespace EAR

namespace MPEGI {

// fwd declares
struct Listener;

//! Add separate Diffraction items for RIs that are occluded by geometry
/*!
 * Inputs:
 * - Only items with ItemType::Primary and NOT in listener-relative coordinate space
 *
 * Outputs:
 * - New RIs with type ItemType::Diffraction, representing new paths that are diffracted around geometry edges
 *
 * How it works:
 * - When a primary RI is occluded, new paths are calculated using auxiliary information from the bitstream
 *
 * Stages order:
 * - Required before: OcclusionStage
 * - Required after: DistanceStage, EqStage, FadeStage
 *
 * Config options:
 * - (not real-time adjustable) "Diffraction_GeometricErrorThreshold": a threshold for comparison of geometric positions (default: 0.001 m)
 * - (not real-time adjustable) "Diffraction_MaxNumberOfPaths": maximum number of paths per diffraction item (default: 4, probably unused)
 *
 */
class DiffractionStage : public ConfigObserver, public RenderStage
{
public:
    // Constructor
    DiffractionStage();

    // Destructor
    ~DiffractionStage();

    // Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from ConfigObserver
    void configValueChanged(const std::string& key) override;

private:
    friend class Renderer;
    friend class GeometryObserver;

    // Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    //! compute the three-integer voxel coordinate for a Cartesian coordinate
    VoxelCoord pointToVoxelCoordinate(const EAR::Vector3& pos) const;

    // IDs of non-static object sources obtained from EIFPayload
    std::set<int> dynamicObjectSources;

    // Listener object
    std::unique_ptr<Listener> listener;

    // A list of the mesh pointers in the scene
    std::vector<const EAR::Mesh*> meshes;

    // A list of the mesh pointers in the scene
    std::vector<const EAR::Mesh*> dynmeshes;

    // Unique pointer of the instantiated EmbreeTracer
    std::unique_ptr<EAR::EmbreeTracer> tracer;

    // Unique pointer of the instantiated EmbreeTracer for dynamic geometries
    std::unique_ptr<EAR::EmbreeTracer> dynTracer;

    // Shared pointer of the diffraction payload
    std::shared_ptr<MPEGI::DiffractionMetadata> diffrMetadata{ nullptr };

    // Shared pointer of the eif payload
    std::shared_ptr<const MPEGI::EIFPayload> eifPayload{ nullptr };

    // Mutex object to offer exclusive ownership regarding edgeList access/update
    std::mutex edgeListAccess;

    // For secondaryItem
    // The pointer of instantiated RenderItemStore
    // needed to create new render items for secondary diffraction paths
    RenderItemStore* itemStore;

    struct ItemState : RenderItem::ExtraState
    {
        explicit ItemState(MPEGI::Listener* listener,
            std::shared_ptr<MPEGI::DiffractionMetadata> diffrMetadata,
            std::shared_ptr<const MPEGI::EIFPayload> eifPayload)
            : source{ listener }
            , rtDiffractionPathTracker{ &source, listener, diffrMetadata, eifPayload }
        {}

        Source source;
        RTDiffractionPathTracker rtDiffractionPathTracker;
    };

    struct SecondaryItems : RenderItem::ExtraState
    {
        // diffractionItem hash -> secondary item
        std::map<std::size_t, RenderItem*> secondaryItems;
    };

    // A list of deactivated primary items (Fade in/out)
    std::vector<int> deactivatedPrimaryItems;

    // Flag which indicates if a dynamic object is included
    bool isDynamicGeometryIncluded{ false };

    // Flag which indicates if a dynamic object has moved during a time frame
    bool isDynamicGeometryRelocated{ false };

    // Geometry observer
    /*!
	* \usage: this class provides the interface to monitor geometry changes/modifications
	*/
    class GeometryObserver : EAR::SceneObjectObserver
    {
    public:
        // Constructor
        GeometryObserver(EAR::Geometry* movingGeometry, DiffractionStage* parent)
            : observedGeometry(movingGeometry)
            , parent(parent)
        {
            observedGeometry->attach(this);
        }

        // Destructor
        ~GeometryObserver() { observedGeometry->detach(this); }

        // Inherited from EAR::SceneObjectObserver
        void objectChanged(
            EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override
        {
            EAR_CHECK_REL(obj == observedGeometry,
                "Exception: SceneObjectObserver::objectChanged, check input pointer");

            if (modification == EAR::SceneObjectObserver::Property::Position) {
                parent->updateDynamicEdges(observedGeometry);
            }
        }

    private:
        // The pointer of Diffraction Stage
        DiffractionStage* parent;

        // A pointer of a non-static/dynamically moving geometry
        EAR::Geometry* observedGeometry;
    };

    // A list of mesh observers which are created to monitor non-static geometries/meshes/primitives
    std::vector<GeometryObserver> geometryObservers;

    // Update dynamic edge's attributes when the motion is detected with primitive type geometry
    void updateDynamicEdges(const EAR::Geometry* geometry);

    // Update static edge's attributes
    void updateStaticEdges(const EAR::Geometry* geometry);

    // Return path index list in static path DBs
    void updateValidPathIndexList(std::string srcName, Source& source, int);

    // Return path index list in dynamic path DBs
    std::vector<std::pair<int, int>> validPathIndexList2(Source& source, int);

    // Return path index list using semi-brute force search
    void runBruteForceSearchForStaticSources(Source& source, int);

    // Return visible edge list using reachable voxel data
    std::vector<int> visibleEdgeListFromReachableVoxels(EAR::Vector3, int = -1);

    // Return visible edge list without any hint
    std::vector<int> bruteForceEdgeSearch(EAR::Vector3, EAR::Vector3, bool);

    // Return visible listener edge list from source edge list
    void visibleListenerEdgeList(EAR::Vector3, std::vector<int>&);

    // Update paths to visualize
    std::vector<std::vector<EAR::Vector3>> updatePathsToVisualize(
        Source& source, RTDiffractionPathTracker& rtDiffractionPathTracker);

    // Update source extent's occluded region position and gain with rayHits
    void updateOccludedLocationAndEqWithRay(std::string, Source& source);

    // Update intersection test samples
    void calculateIntersectionTestSamples(const SceneState* state);

    // Diffraction item initial EQs
    std::vector<float> diffrItemInitialEQs;

    // Diffraction relevant config values
    std::atomic<float> geoErrThreshold;
    std::atomic<float> cullingDist;

    // Block size and sampling frequency
    float B;
    float Fs;

    // Update counter to initiate heavy and light update
    unsigned int updateCounter{ 0 };

    // Gain history for source extent
    std::map<std::string, std::vector<float>> gainHistoryMap;

    //! Diffraction static paths described with edge center locations
    std::map<int, std::vector<std::vector<EAR::Vector3>>> diffractionPaths;

    // Intersection test samples per extended source
    std::map<std::string, std::vector<EAR::Vector3>> intersectionTestSamples;
};

} // namespace MPEGI

#endif /* MPEGI_DIFFRACTION_STAGE */
