/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DiffractionItem.h"
#include <cmath>

namespace MPEGI {

DiffractionItem::DiffractionItem(){};

DiffractionItem::~DiffractionItem(){};

bool DiffractionItem::updateStaticItem(void)
{
    edgeSequence = diffrMetadata->getPathDictByEdges().at(
        pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];
    faceIndicatorSequence = diffrMetadata->getPathDictByFaceIndicators().at(
        pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];
    angleSequence = diffrMetadata->getPathDictByAngles().at(
        pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];

    // Update diffraction item's location and orientation
    updateRotatedSourceLocation();
    updateRotatedSourceLocationForVisiblePath();
    float outwardMax =
        std::fmin(1.f + (listener->listenerLocation - source->sourceLocation).abs() / 20, 2.f);
    if (outwardDistance > outwardMax) {
        return false;
    }
    updateRotatedSourceOrientation();

    // Update diffraction item's filter buffer
    updateFilterBuffer();

    return true;
}

std::vector<std::vector<float>> DiffractionItem::calAngleSequencesByOccluder(EAR::Vector3 PosFront,
    EAR::Vector3 PosBack, std::unique_ptr<EAR::EmbreeTracer>& tracer, int dynMeshBitsId)
{
    auto itr = diffrMetadata->getDynamicEdgeLists().find(dynMeshBitsId);
    EAR_CHECK_REL((itr != diffrMetadata->getDynamicEdgeLists().end()),
        "EAR Exception: given dynamic mesh bitstream id is missing - calAngleSequencesByOccluder");

    std::vector<int> pfeList;
    for (auto& e : itr->second) {
        if (!tracer->isOccluded(PosFront, e.cp_vt - PosFront,
                Config::getFloat("Diffraction_GeometricErrorThreshold")))
            pfeList.push_back(e.edgeID);
    }

    std::vector<int> pbList;
    for (auto e : itr->second) {
        if (!tracer->isOccluded(PosBack, e.cp_vt - PosBack,
                Config::getFloat("Diffraction_GeometricErrorThreshold")))
            pbList.push_back(e.edgeID);
    }

    std::vector<std::pair<int, int>> pf2pb_dynamicPathIndexList;
    for (auto pfeId : pfeList) {
        for (auto pbeId : pbList) {
            int localIndex = 0;
            for (const auto eIdSeq :
                diffrMetadata->getDynamicPathDictByEdgeIds(dynMeshBitsId).at(pfeId)) {
                if (eIdSeq.back() == pbeId) {
                    pf2pb_dynamicPathIndexList.push_back(std::make_pair(pfeId, localIndex));
                }
                localIndex++;
            }
        }
    }

    std::vector<std::vector<float>> pf2pb_angleSequences;

    for (auto pIndex : pf2pb_dynamicPathIndexList) {
        const auto& edgeSequence =
            diffrMetadata->getDynamicPathDictByEdges(dynMeshBitsId).at(pIndex.first)[pIndex.second];
        const auto& faInSequence = diffrMetadata->getDynamicPathDictByFaceIndicators(dynMeshBitsId)
                                       .at(pIndex.first)[pIndex.second];
        const auto& anglSequence = diffrMetadata->getDynamicPathDictByAngles(dynMeshBitsId)
                                       .at(pIndex.first)[pIndex.second];

        // Check if the edge-to-edge visibility along the given dynamic geometry is guaranteed
        bool isE2EOccluded = false;
        for (size_t eIndex = 0; eIndex < edgeSequence.size() - 1; eIndex++) {
            const auto& eA = edgeSequence[eIndex + 0];
            const auto& eB = edgeSequence[eIndex + 1];

            EAR::Vector3 posA = eA->cp_vt;
            EAR::Vector3 posB = eB->cp_vt;
            bool isOccluded = tracer->isOccluded(
                posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

            if (isOccluded) {
                isE2EOccluded = true;
                pf2pb_angleSequences.clear();
                return pf2pb_angleSequences;
            }
        }

        float sourceAngle_UpperBound_InPath = anglSequence.front();
        float listenerAngle_LowerBound_InPath = anglSequence.back();
        float sourceAngle_UpperBound_InGeometry = edgeSequence.front()->edgeAngle;
        float listenerAngle_UpperBound_InGeometry = edgeSequence.back()->edgeAngle;

        AngleCalculation iac_object_src(
            edgeSequence.front(), faInSequence.front(), source->sourceLocation);
        float sourceAngle_InPath = iac_object_src.calculateInternalAngle();

        AngleCalculation iac_object_lis(
            edgeSequence.back(), faInSequence.back(), listener->listenerLocation);
        float listenerAngle_InPath = iac_object_lis.calculateInternalAngle();

        if ((sourceAngle_InPath > sourceAngle_UpperBound_InPath)
            || (sourceAngle_InPath > sourceAngle_UpperBound_InGeometry) || (sourceAngle_InPath < 0))
            continue;

        if ((listenerAngle_InPath < 0)
            || (listenerAngle_InPath > listenerAngle_UpperBound_InGeometry)
            || (listenerAngle_InPath < listenerAngle_LowerBound_InPath))
            continue;

        std::vector<float> angleSequence_pf2pb = anglSequence;
        if (edgeSequence.size() == 1) {
            if (listenerAngle_InPath - sourceAngle_InPath - 180.f < 0)
                continue;

            angleSequence_pf2pb[0] = listenerAngle_InPath - sourceAngle_InPath - 180.f;

        } else if (edgeSequence.size() > 1) {
            angleSequence_pf2pb.front() = sourceAngle_UpperBound_InPath - sourceAngle_InPath;
            angleSequence_pf2pb.back() = listenerAngle_InPath - listenerAngle_LowerBound_InPath;
        }

        pf2pb_angleSequences.push_back(angleSequence_pf2pb);
    }

    return pf2pb_angleSequences;
}

bool DiffractionItem::updateOccludedStaticItem(std::unique_ptr<EAR::EmbreeTracer>& tracer)
{
    edgeSequence = diffrMetadata->getPathDictByEdges().at(
        pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];
    faceIndicatorSequence = diffrMetadata->getPathDictByFaceIndicators().at(
        pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];
    angleSequence = diffrMetadata->getPathDictByAngles().at(
        pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];

    // Represent the influence of interruption by dynamic geometries as additional attenuation
    attenuationFilterBuffer.clear();
    for (int i = 0; i < (int)pathIncluded->whichPathOccluded.size(); i++) {
        // Find potential static paths from the lister to the source
        const std::string& dynMeshName = pathIncluded->intermediatePathOccluder[i]->getName();
        int dynMeshBitsId = diffrMetadata->getGeoBitsId(dynMeshName);
        auto angleSequencesByOccluder = calAngleSequencesByOccluder(
            edgeSequence.back()->cp_vt, listener->listenerLocation, tracer, dynMeshBitsId);

        if (angleSequencesByOccluder.size() != 0) {
            attenuationFilterBuffer.clear();
            break;
        }

        // Update attenuation buffer by other occluders
        updateAttenuationFilterBuffer(angleSequencesByOccluder);
    }

    // Update diffraction item's location and orientation
    updateRotatedSourceLocation();
    updateRotatedSourceLocationForVisiblePath();
    updateRotatedSourceOrientation();

    // Update diffraction item's filter buffer
    updateFilterBuffer();

    return true;
}

bool DiffractionItem::updateDynamicItem(int dynMeshBitsId)
{
    edgeSequence = diffrMetadata->getDynamicPathDictByEdges(dynMeshBitsId)
                       .at(pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];
    faceIndicatorSequence = diffrMetadata->getDynamicPathDictByFaceIndicators(dynMeshBitsId)
                                .at(pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];
    angleSequence = diffrMetadata->getDynamicPathDictByAngles(dynMeshBitsId)
                        .at(pathIncluded->initialEdgeID)[pathIncluded->localPathIndex];

    // Update diffraction item's location and orientation
    updateRotatedSourceLocation();
    updateRotatedSourceLocationForVisiblePath();
    updateRotatedSourceOrientation();

    // Update diffraction item's filter buffer
    updateFilterBuffer();

    return true;
}

EAR::Vector3& DiffractionItem::getSourceRotatedLocation()
{
    return rotatedSourceLocation;
}

EAR::Euler& DiffractionItem::getSourceRotatedOrientation()
{
    return rotatedSourceOrientation;
}

std::vector<float>& DiffractionItem::getDiffractionFilterBuffer()
{
    return diffractionFilterBuffer;
}

float* DiffractionItem::getDiffractionFilterBufferPointer()
{
    return &diffractionFilterBuffer.front();
}

void DiffractionItem::updateRotatedSourceLocation(void)
{
    EAR::Vector3 sourceLocation_temp, rotatedSourceLocation_temp;
    float angleToRotate;
    for (int i = 0; i < int(edgeSequence.size()); i++) {
        if (i == 0) {
            sourceLocation_temp = source->sourceLocation;
            angleToRotate = pathIncluded->diffractionAngleAtSource;
        } else if (i == int(edgeSequence.size()) - 1) {
            sourceLocation_temp = rotatedSourceLocation_temp;
            angleToRotate = pathIncluded->diffractionAngleAtListener;
        } else {
            sourceLocation_temp = rotatedSourceLocation_temp;
            angleToRotate = angleSequence[i]; // i in [0, ..., int(edgeSequence.size()) - 2]
        }
        angleCalculator =
            AngleCalculation(edgeSequence[i], faceIndicatorSequence[i], sourceLocation_temp);
        GeoPathAnalysis geoPathAnalyzer = GeoPathAnalysis(angleCalculator);

        rotatedSourceLocation_temp = geoPathAnalyzer.getRotatedTargetLocation(angleToRotate);
    }
    rotatedSourceLocation = rotatedSourceLocation_temp;
}

void DiffractionItem::updateRotatedSourceLocationForVisiblePath(void)
{
    angleCalculator =
        AngleCalculation(edgeSequence.back()->vtxCoords, listenerLocation, rotatedSourceLocation);
    GeoPathAnalysis geoPathAnalyzer = GeoPathAnalysis(angleCalculator);
    auto edgeIntersectionData = geoPathAnalyzer.getEdgeIntersection();

    EAR::Vector3 intersectionCoordinate = edgeIntersectionData.second.second;

    if (edgeIntersectionData.first) {
        outwardDistance = edgeIntersectionData.second.first;
    } else {
        outwardDistance = 0;
    }
}

void DiffractionItem::updateRotatedSourceOrientation(void)
{
    EAR::Euler sourceOrientation_temp, rotatedSourceOrientation_temp;
    float angleToRotate;
    for (int i = 0; i < int(edgeSequence.size()); i++) {
        if (i == 0) {
            sourceOrientation_temp = source->sourceOrientation;
            angleToRotate = pathIncluded->diffractionAngleAtSource;
        } else if (i == int(edgeSequence.size()) - 1) {
            sourceOrientation_temp = rotatedSourceOrientation_temp;
            angleToRotate = pathIncluded->diffractionAngleAtListener;
        } else {
            sourceOrientation_temp = rotatedSourceOrientation_temp;
            angleToRotate = angleSequence[i];
        }
        rotatedSourceOrientation_temp = rotateSourceOrientationPerEdge(edgeSequence[i],
            faceIndicatorSequence[i], nullptr, sourceOrientation_temp, angleToRotate);
    }

    rotatedSourceOrientation = rotatedSourceOrientation_temp;
}

EAR::Euler DiffractionItem::rotateSourceOrientationPerEdge(DiffractionEdge* edge,
    int _faceIndicator, EAR::Mesh::Triangle* triangle, EAR::Euler& sourceOrientation,
    float angleToRotate)
{
    float angleInRadian = angleToRotate * (M_PI / 180.f);

    EAR::Vector3 rotatingAxis = (edge->vtxCoords.second - edge->vtxCoords.first).normalize();
    EAR::Vector3 ecc = edge->cp;

    EAR::Vector3 tri2ecc = EAR::Vector3(0, 0, 0);
    EAR::Vector3 rot_dirc = EAR::Vector3(0, 0, 0);

    if (edge->isStatic) {
        if (_faceIndicator == 0) {
            tri2ecc = ecc - edge->adjacentTriCentroids.first;
            rot_dirc = edge->adjacentTriNormals.first * tri2ecc;
        } else if (_faceIndicator == 1) {
            tri2ecc = ecc - edge->adjacentTriCentroids.second;
            rot_dirc = edge->adjacentTriNormals.second * tri2ecc;
        }
    } else {
        throw EAR::Exception(
            ("NOT IMPLEMENTED - ERROR, ROTATE_SOURCE_ORIENTATION_PER_EDGE: [EDGE->isStatic = "
             "false], MOVABLE MESHES ARE NOT SUPPORTED BY THIS VERSION OF DIFFRACTION"));
    }

    bool sameDirec = (rotatingAxis.dot(rot_dirc) > 0);
    // to rotate a target orientation by given angleToRotate
    if (!sameDirec) {
        rotatingAxis = -1 * rotatingAxis;
    }

    // to get orientation data in Euler
    EAR::Quaternion currentSrcOriinQuarternion(sourceOrientation), rotatingMatrixinQuaternion,
        rotatedSrcOriinQuaternion;
    rotatingMatrixinQuaternion.set(cos(angleInRadian / 2), rotatingAxis[0] * sin(angleInRadian / 2),
        rotatingAxis[1] * sin(angleInRadian / 2), rotatingAxis[2] * sin(angleInRadian / 2));
    rotatedSrcOriinQuaternion = rotatingMatrixinQuaternion * currentSrcOriinQuarternion;
    EAR::Euler rotatedSrcOriinEuler = rotatedSrcOriinQuaternion.getEuler();

    return rotatedSrcOriinEuler;
}

void DiffractionItem::updateFilterBuffer(void)
{
    diffractionFilterBuffer = std::vector<float>(Config::getInt("EQNBands"), 1.f);

    // fade-out, if the diffraction point is outside of the edge
    float outwardMax =
        std::fmin(1.f + (listener->listenerLocation - source->sourceLocation).abs() / 20, 2.f);
    float spatialAttenuation = std::fmax(1.f - outwardDistance / outwardMax, 0.f);

    itemGain *= spatialAttenuation;
    angleSum = 0.f;

    // Update diffraction attenuation filter buffer by diffraction at the source side
    float tuneParameter = Config::getFloat("DiffractionFilterTuneParameter");

    // round to nearest integer
    angleSum += pathIncluded->diffractionAngleAtSource;
    int selectedFilter = int(pathIncluded->diffractionAngleAtSource + 0.5f);
    for (int k = 0; k < Config::getInt("EQNBands"); k++) {
        float updateValue =
            (1 - tuneParameter) + tuneParameter * Diffraction_FilterCoeffs[k][selectedFilter];
        diffractionFilterBuffer[k] *= updateValue;
    }

    // Update diffraction attenuation filter buffer by diffraction along internal paths
    for (int order = 1; order < int(angleSequence.size()) - 1; order++) {
        angleSum += angleSequence[order];
        for (int k = 0; k < Config::getInt("EQNBands"); k++) {
            float updateValue = 1 - tuneParameter
                + tuneParameter * Diffraction_FilterCoeffs[k][(int)(angleSequence[order])];
            diffractionFilterBuffer[k] *= updateValue;
        }
    }

    // Update diffraction attenuation filter buffer by diffraction at the listener side
    if (pathIncluded->diffractionAngleAtListener != -1) {
        angleSum += pathIncluded->diffractionAngleAtListener;
        for (int k = 0; k < Config::getInt("EQNBands"); k++) {
            float updateValue = 1 - tuneParameter
                + tuneParameter
                    * Diffraction_FilterCoeffs[k][(int)pathIncluded->diffractionAngleAtListener];
            diffractionFilterBuffer[k] *= updateValue;
        }
    }

    if (pathIncluded->pathType == DiffractionPathType::CombinedPath) {
        if (attenuationFilterBuffer.size() == 0) {
            for (int k = 0; k < Config::getInt("EQNBands"); k++) {
                diffractionFilterBuffer[k] = 0;
            }
        }

        for (int i = 0; i < attenuationFilterBuffer.size(); i++) {
            for (int k = 0; k < Config::getInt("EQNBands"); k++) {
                diffractionFilterBuffer[k] *= attenuationFilterBuffer[i][k];
            }
        }
    }
}

void DiffractionItem::updateAttenuationFilterBuffer(
    std::vector<std::vector<float>> angleSequencesByOccluder)
{
    int eqNBands = Config::getInt("EQNBands");
    // Create a vector of per-path attenuation coefficient buffers
    std::vector<std::vector<float>> attenuationFilterBuffers;
    for (auto angles : angleSequencesByOccluder) {
        std::vector<float> attenuationFilterBufferPerPath(eqNBands, 1);
        for (auto angle : angles) {
            for (int k = 0; k < eqNBands; k++) {
                attenuationFilterBufferPerPath[k] *= Diffraction_FilterCoeffs[k][(int)angle];
            }
        }
        attenuationFilterBuffers.push_back(attenuationFilterBufferPerPath);
    }

    // Calculate summed attenuation filter power
    // If there is no valid path, then there will be absolute silence such as diffraction path through a closed door
    std::vector<float> sFilter(eqNBands, 0);
    for (const auto& attenuationFilterBufferPerPath : attenuationFilterBuffers) {
        for (int k = 0; k < eqNBands; k++) {
            sFilter[k] += std::pow(attenuationFilterBufferPerPath[k], 2);
        }
    }

    for (int k = 0; k < eqNBands; k++) {
        if (sFilter[k] > 1.0f) {
            sFilter[k] = 1;
        } else {
            sFilter[k] = sqrt(sFilter[k]);
        }
    }

    attenuationFilterBuffer.push_back(sFilter);
}

} // namespace MPEGI
