/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DIFFRACTION_ITEM
#define MPEGI_DIFFRACTION_ITEM

#include <MPEGI/Bitstream/EIFPayload.h>
#include <MPEGI/Renderer/DiffractionEdge.h>

#include <EAR/Exception.h>

#include "Detail.h"
#include "DiffractionMetadata.h"
#include "GeometrywisePathAnalysis.h"
#include "ListenerStruct.h"
#include "PathToValidate.h"
#include "SourceStruct.h"

#include <EAR/Geometry.h>

namespace MPEGI {

// It provides diffraction item-relevant interfaces
class DiffractionItem : public AngleCalculation
{
public:
    // Constructor
    DiffractionItem();

    // Constructor
    DiffractionItem(const Source* sourceIn, const Listener* listenerIn,
        const PathToValidate* pathIn, std::shared_ptr<MPEGI::DiffractionMetadata> diffrMetadataIn,
        std::shared_ptr<const MPEGI::EIFPayload> eifPayloadIn)
        : source(sourceIn)
        , listener(listenerIn)
        , pathIncluded(pathIn)
        , diffrMetadata(diffrMetadataIn)
        , eifPayload(eifPayloadIn)
    {
        itemID = std::hash<std::string>{}(std::to_string(source->sourceID)
            + std::to_string(pathIncluded->initialEdgeID)
            + std::to_string(pathIncluded->localPathIndex));

        listenerLocation = listener->listenerLocation;
    };

    // Destructor
    ~DiffractionItem();

    // Update stored item
    bool updateStaticItem(void);

    // Update stored item
    bool updateOccludedStaticItem(std::unique_ptr<EAR::EmbreeTracer>&);

    // Update stored item
    bool updateDynamicItem(int dynMeshBitsId);

    // Return a pair of visible dynamic edge lists using dynMeshBitsId
    std::vector<std::vector<float>> calAngleSequencesByOccluder(
        EAR::Vector3, EAR::Vector3, std::unique_ptr<EAR::EmbreeTracer>&, int);

    // Return rotated source location by associated edges
    EAR::Vector3& getSourceRotatedLocation(void);

    // Return rotated source orientation by associated edges
    EAR::Euler& getSourceRotatedOrientation(void);

    // Return filter buffer
    std::vector<float>& getDiffractionFilterBuffer(void);

    // Return filter buffer pointer
    float* getDiffractionFilterBufferPointer(void);

    // Return itemId
    inline std::size_t getID() const { return itemID; };

    // Return itemGain
    inline float getGain() const { return itemGain; }

    // Set itemGain
    inline void setGain(float gain) { itemGain = gain; };

    // Return angleSum
    inline float getAngleSum() const { return angleSum; };

    // Set itemGain
    inline void addGain(float gain) { itemGain *= gain; };

    // Return the included path - for debug
    inline const PathToValidate* getPathIncluded(void) const { return pathIncluded; }

private:
    // Shared pointer of the diffraction metadata
    std::shared_ptr<MPEGI::DiffractionMetadata> diffrMetadata;

    // Shared pointer of the eif payload
    std::shared_ptr<const MPEGI::EIFPayload> eifPayload;

    // Source pointer derived from a renderItem
    const Source* source;

    // Listener pointer
    const Listener* listener;

    // Diffraction item ID hash from strings
    std::size_t itemID;

    // Fetched path sequence by edges
    std::vector<DiffractionEdge*> edgeSequence;

    // Fetched path sequence by face indicators (0: forward direction, 1: backward direction)
    std::vector<int> faceIndicatorSequence;

    // Fetched path sequence by corresponding triangles
    std::vector<EAR::Mesh::Triangle*> triangleSequence;

    // Fetched path sequence by diffraction and reference angles
    std::vector<float> angleSequence;

    // Attenuation Filter Buffer for unoccluded path
    std::vector<float> diffractionFilterBuffer;

    // Listener location which is added not to call listener pointer member variables
    EAR::Vector3 listenerLocation;

    // Rotated diffraction item location
    EAR::Vector3 rotatedSourceLocation;

    // Distance to mitigate finite-sized edge which is used for spatial smoothing weight
    float outwardDistance;

    // Rotated diffraction item orientation
    EAR::Euler rotatedSourceOrientation;

    // Included angleCalculator object used for angle calculation
    AngleCalculation angleCalculator;

    // Relevant validated path as parent
    const PathToValidate* pathIncluded;

    // Attenuation Buffer due to dynamic geometry interruption
    // Its size represents the number of occluded internal paths
    std::vector<std::vector<float>> attenuationFilterBuffer;

    // Linear gain for this item
    float itemGain{ 1.0 };

    // Angle sum
    float angleSum{ 0.f };

    // Private functions to generate diffraction item's position (location, orientation)
    void updateRotatedSourceLocation(void);
    void updateRotatedSourceLocationForVisiblePath(void);
    void updateRotatedSourceOrientation(void);
    EAR::Euler rotateSourceOrientationPerEdge(
        DiffractionEdge*, int, EAR::Mesh::Triangle*, EAR::Euler&, float);

    // Private function to update diffraction item EQ
    void updateFilterBuffer(void);
    void updateAttenuationFilterBuffer(std::vector<std::vector<float>>);
};

} // namespace MPEGI

#endif /* MPEGI_DIFFRACTION_ITEM */
