/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DiffractionMetadata.h"
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/DiffractionEdge.h>
#include <MPEGI/Renderer/DiffractionPath.h>

#include <EAR/BaseTypes.h>
#include <EAR/Exception.h>

#include <algorithm>
#include <cassert>
#include <utility>

namespace MPEGI {

//! Constructor
DiffractionMetadata::DiffractionMetadata(
    std::shared_ptr<const MPEGI::DiffractionPayload> diffrPayload)
    : diffrPayload(diffrPayload)
    , edgeList(diffrPayload->edgeList)
    , dynamicEdgeLists(diffrPayload->dynamicEdgeLists)
{
    // clone pathDict
    for (auto& [id, paths] : diffrPayload->pathDict) {

        for (auto& path : paths) {

            const auto& edgeIDSeq = path.getEdgeIDSequence();
            const auto& faceIndicatorSeq = path.getFaceIndicatorSequence();
            const auto& angleSeq = path.getAngleSequence();

            // number of edges in the path
            int nPoints = edgeIDSeq.size();

            std::vector<DiffractionPoint> pointSeq;
            pointSeq.reserve(nPoints);
            for (int k = 0; k < nPoints; k++) {
                int eId = edgeIDSeq[k];
                int fIndicator = faceIndicatorSeq[k];
                float angle = angleSeq[k];

                DiffractionPoint p = { eId, &edgeList[eId], fIndicator, angle };
                pointSeq.push_back(p);
            }
            DiffractionPath newPath(pointSeq);

            int pathDictKey = newPath.getKeyId(); // = the initial edge Id
            if (pathDict.find(pathDictKey) != pathDict.end()) {
                pathDict.at(pathDictKey).push_back(newPath);
            } else {
                std::vector<DiffractionPath> paths = { newPath };
                pathDict.insert({ pathDictKey, paths });
            }
        }
    }

    // initialize pathDictBy... and other path dict metadata
    updateStaticPathDict();

    // initialize dynamicPathDictsBy... and other dynamic path dict metadata
    updateDynamicPathDicts();
}

void DiffractionMetadata::updateStaticPathDict()
{
    for (auto pathPair : pathDict) {
        int pathKey = pathPair.first;

        int localPathCount = 0;
        for (auto pathInThisKey : pathPair.second) {

            // edgeSequence
            std::vector<DiffractionEdge*> edgeSequence; // by edge pointers
            for (auto edgeID : pathInThisKey.getEdgeIDSequence()) {
                edgeSequence.push_back(&edgeList[edgeID]);
            }

            // faceIndicatorSequence
            std::vector<int> faceIndicatorSequence = pathInThisKey.getFaceIndicatorSequence();

            // angleSequence
            std::vector<float> angleSequence = pathInThisKey.getAngleSequence();

            // push them into the three different maps for real-time process
            pathDictByEdges[pathKey].push_back(edgeSequence);
            pathDictByEdgeIds[pathKey].push_back(pathInThisKey.getEdgeIDSequence());
            pathDictByFaceIndicators[pathKey].push_back(faceIndicatorSequence);
            pathDictByAngles[pathKey].push_back(angleSequence);

            // path dictionaries propagation distances
            float distance = 0;
            for (int i = 0; i < int(edgeSequence.size()) - 1; i++) {
                const auto& e1 = edgeSequence[i + 0];
                const auto& e2 = edgeSequence[i + 1];

                distance += (e2->cp - e1->cp).abs();
            }
            pathDictByDistances[pathKey].push_back(distance);

            // Encoder: max_order_moving_source_DIFFR argument
            const auto pathIndex =
                std::make_pair(pathKey, int(pathDictByEdges[pathKey].size()) - 1);
            staticPathIndexListPerEdge1[pathKey].push_back(pathIndex);

            if (edgeSequence.size() < 3) {
                staticPathIndexListPerEdge2[pathKey].push_back(pathIndex);
            }
        }
    }

    // Update visible edge id dictionary for listener in case of moving sources
    for (const auto& pathDictPair : pathDictByEdgeIds) {
        int seId = pathDictPair.first;
        visibleEdgeDictionaryForListener.insert(
            std::pair<int, std::vector<int>>(seId, std::vector<int>()));

        std::vector<int> visibleEdgeIdListForListener;

        for (const auto& edgeIdSequence : pathDictPair.second) {
            if (std::find(visibleEdgeIdListForListener.begin(), visibleEdgeIdListForListener.end(),
                    edgeIdSequence.back())
                == visibleEdgeIdListForListener.end()) {
                visibleEdgeIdListForListener.push_back(edgeIdSequence.back());
            }
        }

        visibleEdgeDictionaryForListener.at(seId) = visibleEdgeIdListForListener;
    }
}

void DiffractionMetadata::updateDynamicPathDicts()
{
    for (auto& [geoIdIdx, pathDictToUpdate] : diffrPayload->dynamicPathDicts) {
        std::map<int, std::vector<std::vector<DiffractionEdge*>>> pathDictByEdgesToUpdate;
        std::map<int, std::vector<std::vector<int>>> pathDictByEdgeIdsToUpdate;
        std::map<int, std::vector<std::vector<int>>> pathDictByFaceIndicatorsToUpdate;
        std::map<int, std::vector<std::vector<float>>> pathDictByAnglesToUpdate;

        for (auto pathPair : pathDictToUpdate) {
            int pathKey = pathPair.first;

            int localPathCount = 0;
            for (auto pathInThisKey : pathPair.second) {

                // edgeSequence
                std::vector<DiffractionEdge*> edgeSequence; // by edge pointers
                for (auto edgeID : pathInThisKey.getEdgeIDSequence()) {
                    edgeSequence.push_back(&(dynamicEdgeLists.at(geoIdIdx)[edgeID]));
                }

                // faceIndicatorSequence
                std::vector<int> faceIndicatorSequence = pathInThisKey.getFaceIndicatorSequence();

                // angleSequence
                std::vector<float> angleSequence = pathInThisKey.getAngleSequence();

                // push them into the three different maps for real-time process
                pathDictByEdgesToUpdate[pathKey].push_back(edgeSequence);
                pathDictByEdgeIdsToUpdate[pathKey].push_back(pathInThisKey.getEdgeIDSequence());
                pathDictByFaceIndicatorsToUpdate[pathKey].push_back(faceIndicatorSequence);
                pathDictByAnglesToUpdate[pathKey].push_back(angleSequence);
            }
        }

        // Update dynamic Path dictionaries each of which is represented by edge pointers, edge Ids, face indicators and angles
        dynamicPathDictsByEdges.insert({ geoIdIdx, pathDictByEdgesToUpdate });
        dynamicPathDictsByEdgeIds.insert({ geoIdIdx, pathDictByEdgeIdsToUpdate });
        dynamicPathDictsByFaceIndicators.insert({ geoIdIdx, pathDictByFaceIndicatorsToUpdate });
        dynamicPathDictsByAngles.insert({ geoIdIdx, pathDictByAnglesToUpdate });
    }
}

const std::vector<DiffractionEdge>& DiffractionMetadata::getDynamicEdgeLists(int dynMeshBitsId)
{
    auto itr = dynamicEdgeLists.find(dynMeshBitsId);
    EAR_CHECK_REL(itr != dynamicEdgeLists.end(),
        "EAR Exception: given dynamic mesh bitstream id is missing - getDynamicEdgeLists");
    return itr->second;
}

std::map<int, std::vector<std::vector<DiffractionEdge*>>>&
DiffractionMetadata::getDynamicPathDictByEdges(int dynMeshBitsId)
{
    auto itr = dynamicPathDictsByEdges.find(dynMeshBitsId);
    EAR_CHECK_REL(itr != dynamicPathDictsByEdges.end(),
        "EAR Exception: String id in getDynamicPathDictByEdges()");
    return itr->second;
}

const std::map<int, std::vector<std::vector<int>>>&
DiffractionMetadata::getDynamicPathDictByEdgeIds(int dynMeshBitsId) const
{
    auto itr = dynamicPathDictsByEdgeIds.find(dynMeshBitsId);
    EAR_CHECK_REL(itr != dynamicPathDictsByEdgeIds.end(),
        "EAR Exception: String id in getDynamicPathDictByEdgeIds()");
    return itr->second;
}

const std::map<int, std::vector<std::vector<int>>>&
DiffractionMetadata::getDynamicPathDictByFaceIndicators(int dynMeshBitsId) const
{
    auto itr = dynamicPathDictsByFaceIndicators.find(dynMeshBitsId);
    EAR_CHECK_REL(itr != dynamicPathDictsByFaceIndicators.end(),
        "EAR Exception: String id in getDynamicPathDictByFaceIndicators()");
    return itr->second;
}

const std::map<int, std::vector<std::vector<float>>>&
DiffractionMetadata::getDynamicPathDictByAngles(int dynMeshBitsId) const
{
    auto itr = dynamicPathDictsByAngles.find(dynMeshBitsId);
    EAR_CHECK_REL(itr != dynamicPathDictsByAngles.end(),
        "EAR Exception: String id in getDynamicPathDictByAngles()");
    return itr->second;
}

} // namespace MPEGI
