/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DIFFRACTION_METADATA
#define MPEGI_DIFFRACTION_METADATA

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/DiffractionPayload.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/DiffractionEdge.h>
#include <MPEGI/Renderer/DiffractionPath.h>

#include <EAR/BaseTypes.h>
#include <EAR/Exception.h>

#include <utility>

namespace MPEGI {

// Provide a accessible C++ representation of an MPEG-I Diffraction type MHAS payload (temporary)
class DiffractionMetadata
{
    using VoxelPathDict = std::map<uint64_t, std::vector<std::pair<int, int>>>;
    using SourceVoxelDict = std::map<std::vector<int>, std::vector<int>>;

public:
    //! Constructor
    DiffractionMetadata(std::shared_ptr<const MPEGI::DiffractionPayload> diffrPayload);

    //! Constructor
    //DiffractionMetadata() = default;

    //! Destructor
    ~DiffractionMetadata() = default;

    //! Dynamic edge lists and path dictionaries in different forms
    //*! Return map of dynamic diffraction edge lists
    const std::vector<DiffractionEdge>& getDynamicEdgeLists(int dynMeshBitsId);

    //! Return all dynamic paths expressed by edges
    std::map<int, std::vector<std::vector<DiffractionEdge*>>>& getDynamicPathDictByEdges(
        int dynMeshBitsId);

    //! Return all dynamic paths expressed by edge Ids
    const std::map<int, std::vector<std::vector<int>>>& getDynamicPathDictByEdgeIds(
        int dynMeshBitsId) const;

    //! Return all dynamic paths expressed by face indicators
    const std::map<int, std::vector<std::vector<int>>>& getDynamicPathDictByFaceIndicators(
        int dynMeshBitsId) const;

    //! Return all dynamic paths expressed by angles
    const std::map<int, std::vector<std::vector<float>>>& getDynamicPathDictByAngles(
        int dynMeshBitsId) const;

    //! Return all edgeList parsed in the diffraction payload
    inline std::vector<DiffractionEdge>& getEdgeList() { return edgeList; };

    //! Return all paths parsed in the diffraction payload
    inline std::map<int, std::vector<DiffractionPath>>& getPathDict() { return pathDict; };

    //! Return all paths expressed by edges
    inline std::map<int, std::vector<std::vector<DiffractionEdge*>>>& getPathDictByEdges()
    {
        return pathDictByEdges;
    };

    //! Return all paths expressed by edge Ids
    inline const std::map<int, std::vector<std::vector<int>>>& getPathDictByEdgeIds() const
    {
        return pathDictByEdgeIds;
    };

    //! Return all paths expressed by face indicators
    inline const std::map<int, std::vector<std::vector<int>>>& getPathDictByFaceIndicators() const
    {
        return pathDictByFaceIndicators;
    };

    //! Return all paths expressed by angles
    inline const std::map<int, std::vector<std::vector<float>>>& getPathDictByAngles() const
    {
        return pathDictByAngles;
    };

    //! Return path distance dictionary
    inline const std::map<int, std::vector<float>>& getPathDistanceDict() const
    {
        return pathDictByDistances;
    };

    //! Return "static path index" list per edge : full support
    inline const std::map<int, std::vector<std::pair<int, int>>>&
    getStaticPathIndexListPerEdge1() const
    {
        return staticPathIndexListPerEdge1;
    };

    //! Return "static path index" list per edge : partial suppor, up-to-2nd-order
    inline const std::map<int, std::vector<std::pair<int, int>>>&
    getStaticPathIndexListPerEdge2() const
    {
        return staticPathIndexListPerEdge2;
    };

    //! Return visible edge Id dictionary for listener
    inline const std::map<int, std::vector<int>>& getVisibleEdgeDictionaryForListener(void) const
    {
        return visibleEdgeDictionaryForListener;
    };

    //! Return dynamicEdgeLists
    inline std::map<int, std::vector<DiffractionEdge>>& getDynamicEdgeLists()
    {
        return dynamicEdgeLists;
    };

    //! Return listenerVoxelDict
    inline const std::map<std::vector<int>, std::vector<int>>& getListenerVoxelDict() const
    {
        return diffrPayload->listenerVoxelDict;
    }

    //! Return sourceVoxelDict
    inline const std::map<int, SourceVoxelDict>& getSourceVoxelDicts() const
    {
        return diffrPayload->sourceVoxelDicts;
    }

    //! Return validPathDict
    inline const std::map<int, VoxelPathDict>& getValidPathDict() const
    {
        return diffrPayload->validPathDict;
    }

    //! Return static source Idx list
    inline const std::vector<int>& getStaticSourceIdxList(void) const
    {
        return diffrPayload->staticSourceIdxList;
    }

    //! Return voxel grid origin
    inline const EAR::Vector3& getVoxelGridOrigin(void) const
    {
        return diffrPayload->voxelGridOrigin;
    }

    //! Return voxel grid shape
    inline const EAR::Vector3& getVoxelGridPitch(void) const
    {
        return diffrPayload->voxelGridPitch;
    }

    //! Return voxel grid shape
    inline const VoxelCoord& getVoxelGridShape(void) const { return diffrPayload->voxelGridShape; }

    //! Resolve geometry name -> bitstream ID
    inline int getGeoBitsId(const std::string& geoName) const
    {
        return diffrPayload->geoName2BitsId.at(geoName);
    }

private:
    // Update static path dict metadata
    void updateStaticPathDict();

    // Update dynamic path dicts metadata
    void updateDynamicPathDicts();

private:
    // Shared pointer of the diffraction payload
    std::shared_ptr<const MPEGI::DiffractionPayload> diffrPayload{ nullptr };

    // List of diffraction edges
    std::vector<DiffractionEdge> edgeList;

    // For dynamic geometry handling
    // Map of dynamic diffraction edge lists
    std::map<int, std::vector<DiffractionEdge>> dynamicEdgeLists;

    // Dictionary of paths
    std::map<int, std::vector<DiffractionPath>> pathDict;

    // Dictionary of paths by edges
    std::map<int, std::vector<std::vector<DiffractionEdge*>>> pathDictByEdges;

    // Dictionary of paths by edge Id
    std::map<int, std::vector<std::vector<int>>> pathDictByEdgeIds;

    // Dictionary of paths by face indicators
    std::map<int, std::vector<std::vector<int>>> pathDictByFaceIndicators;

    // Dictionary of paths by angles
    std::map<int, std::vector<std::vector<float>>> pathDictByAngles;

    // Dictionary of propagation distances for paths
    std::map<int, std::vector<float>> pathDictByDistances;

    // "Static Path Index" List per edge for moving source case in VR: Class 1 - full
    std::map<int, std::vector<std::pair<int, int>>> staticPathIndexListPerEdge1;

    // "Static Path Index" List per edge for moving source case in VR: Class 2 - partial
    std::map<int, std::vector<std::pair<int, int>>> staticPathIndexListPerEdge2;

    // Dictionary of visible edge list for listener position in case of moving sources
    std::map<int, std::vector<int>> visibleEdgeDictionaryForListener;

    // Map of path dictionaries by edges
    std::map<int, std::map<int, std::vector<std::vector<DiffractionEdge*>>>>
        dynamicPathDictsByEdges;

    // Map of path dictionaries by edge Id
    std::map<int, std::map<int, std::vector<std::vector<int>>>> dynamicPathDictsByEdgeIds;

    // Map of path dictionaries by face indicators
    std::map<int, std::map<int, std::vector<std::vector<int>>>> dynamicPathDictsByFaceIndicators;

    // Map of path dictionaries by angles
    std::map<int, std::map<int, std::vector<std::vector<float>>>> dynamicPathDictsByAngles;
};

} // namespace MPEGI

#endif // MPEGI_DIFFRACTION_METADATA
