/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <EAR/Exception.h>
#include <EAR/Geometry.h>

#include "GeometrywisePathAnalysis.h"

namespace MPEGI {

GeoPathAnalysis::GeoPathAnalysis(void){};

GeoPathAnalysis::GeoPathAnalysis(AngleCalculation& angleCalculator)
    : geoAngleCalculator(angleCalculator){};

GeoPathAnalysis::~GeoPathAnalysis(void){};

EAR::Vector3 GeoPathAnalysis::getRotatedTargetLocation(float angleToRotate)
{
    EAR_CHECK_REL(
        geoAngleCalculator.getPrimProperty() != PrimativePropertiesforAngleCalculation::None,
        "[ERROR] GET_ROTATED_SOURCE_LOCATION::WILL DO NOTHING FOR NONE TYPE");

    // rotated target location as return type
    EAR::Vector3 rsl;

    if (geoAngleCalculator.setXYZCoordinates()) {
        // 1) initialization, 2) translation 3) X - Axis Rotation, 4) Y - Axis Rotation
        EAR::Vector3 unbiased_vec;
        setAllMatricesAsIdentity();
        setTranslationMatrix(-1 * geoAngleCalculator.getV_origin());
        unbiased_vec = MatVecMul(geoAngleCalculator.getV_bias(), 0);
        setRotationMatrix(unbiased_vec, RotatingAxis::X_AXIS);
        setRotationMatrix(unbiased_vec, RotatingAxis::Y_AXIS);

        // 5) calcluate the transformation matrix to rotate given source location with respect to given rotating axis (e.g., an edge)
        getTransformationMatrix(3);

        // 6) calculate the rotated vectors
        EAR::Vector3 V_origin_rot = MatVecMul(geoAngleCalculator.getV_origin(), 7);
        EAR::Vector3 V_bias_rot = MatVecMul(geoAngleCalculator.getV_bias(), 7);
        EAR::Vector3 Ui_ref_rot = MatVecMul(geoAngleCalculator.getUi_ref(), 7);
        EAR::Vector3 Uj_tar_rot = MatVecMul(geoAngleCalculator.getUj_tar(), 7);
        EAR::Vector3 N0_Ui_Plane_rot =
            MatVecMul(geoAngleCalculator.getN0_Ui_Plane() + geoAngleCalculator.getV_origin(), 7);

        // 7) find the direction of rotation
        EAR::Vector3 Proj_V_rot_Ui_ref =
            V_bias_rot.dot(Ui_ref_rot) / V_bias_rot.abs() / V_bias_rot.abs() * V_bias_rot;
        EAR::Vector3 Otho_V_rot_Ui_ref = Ui_ref_rot - Proj_V_rot_Ui_ref;
        EAR::Vector3 Proj_V_rot_Uj_tar =
            V_bias_rot.dot(Uj_tar_rot) / V_bias_rot.abs() / V_bias_rot.abs() * V_bias_rot;
        EAR::Vector3 Otho_V_rot_Uj_tar = Uj_tar_rot - Proj_V_rot_Uj_tar;
        EAR::Vector3 CW_CCW_CHECK = Otho_V_rot_Ui_ref.cross(N0_Ui_Plane_rot);

        if (CW_CCW_CHECK[2] > 0) {
            setRotationMatrix(unbiased_vec, RotatingAxis::Z_AXIS, angleToRotate);
        }

        if (CW_CCW_CHECK[2] < 0) {
            setRotationMatrix(unbiased_vec, RotatingAxis::Z_AXIS, -angleToRotate);
        }

        getTransformationMatrix(7);
        EAR::Vector3 temp_srcloc_rot_XYZ = MatVecMul(geoAngleCalculator.getUj_tar(), 7);

        rsl.set(0, 0, 0);
        rsl.set(temp_srcloc_rot_XYZ[1], temp_srcloc_rot_XYZ[2], temp_srcloc_rot_XYZ[0]);
    } else {
        throw "[ERROR] GET_ROTATED_SOURCE_LOCATION:: Coordinate system is not YZX before process";
        return geoAngleCalculator.getUj_tar();
    }

    geoAngleCalculator.setYZXCoordinates();

    return rsl;
}

std::pair<bool, std::pair<float, EAR::Vector3>> GeoPathAnalysis::getEdgeIntersection(void)
{

    std::pair<bool, std::pair<float, EAR::Vector3>> rType;
    EAR::Vector3 edgeIntersection;
    float outwardDistance = 0; // outward distance from the closest edge end point

    EAR::Vector3 p1 = geoAngleCalculator.getUi_ref(); // e.g., rotated source location by an edge
    EAR::Vector3 p2 = geoAngleCalculator.getUj_tar(); // e.g., listener's location
    EAR::Vector3 ed_vec = geoAngleCalculator.getV_origin2bias();
    EAR::Vector3 ed_ori = geoAngleCalculator.getV_origin();

    EAR::Vector3 src_vec = p1 - ed_ori;
    EAR::Vector3 proj_p1_to_ed_vec = src_vec.dot(ed_vec) / ed_vec.abs() / ed_vec.abs() * ed_vec;
    EAR::Vector3 otho_p1_to_ed_vec = src_vec - proj_p1_to_ed_vec;

    EAR::Vector3 tar_vec = p2 - ed_ori;
    EAR::Vector3 proj_p2_to_ed_vec = tar_vec.dot(ed_vec) / ed_vec.abs() / ed_vec.abs() * ed_vec;
    EAR::Vector3 otho_p2_to_ed_vec = tar_vec - proj_p2_to_ed_vec;

    float dist_src =
        otho_p1_to_ed_vec
            .abs(); // distance from the edge to the "rotated source location" perpendicularly
    float dist_tar =
        otho_p2_to_ed_vec
            .abs(); // distance from the edge to the "listener location" perpendicularly
    float ratio_src = dist_src / (dist_src + dist_tar);
    float ratio_tar = dist_tar / (dist_src + dist_tar);

    edgeIntersection[0] =
        proj_p1_to_ed_vec[0] * (1 - ratio_src) + proj_p2_to_ed_vec[0] * (1 - ratio_tar);
    edgeIntersection[1] =
        proj_p1_to_ed_vec[1] * (1 - ratio_src) + proj_p2_to_ed_vec[1] * (1 - ratio_tar);
    edgeIntersection[2] =
        proj_p1_to_ed_vec[2] * (1 - ratio_src) + proj_p2_to_ed_vec[2] * (1 - ratio_tar);

    float direc = ed_vec.dot(edgeIntersection);
    if (direc >= 0) {
        outwardDistance = std::max(edgeIntersection.abs() - ed_vec.abs(), 0.f);
        edgeIntersection += ed_ori;
        rType = std::make_pair(
            outwardDistance > 0.001f, std::make_pair(outwardDistance, edgeIntersection));
    } else {
        if (edgeIntersection.abs() > 0.001f) {
            outwardDistance = edgeIntersection.abs();
            edgeIntersection += ed_ori;
            rType = std::make_pair(true, std::make_pair(outwardDistance, edgeIntersection));
        } else {
            rType = std::make_pair(false, std::make_pair(outwardDistance, edgeIntersection));
        }
    }

    return rType;
}

EAR::Vector3 GeoPathAnalysis::alignSrcLocWithGivenEdge(float rotatingAngleToMakeSrcVisible)
{
    EAR_CHECK_REL((geoAngleCalculator.getPrimProperty()
                      == PrimativePropertiesforAngleCalculation::EdgeIntersection),
        "[ERROR] GET_ROTATED_SOURCE_LOCATION::WILL DO NOTHING IF IT IS NOT EDGE_INTERSECTION TYPE");

    EAR::Vector3 rsl;

    if (geoAngleCalculator.setXYZCoordinates()) {
        // 1) initialization, 2) translation 3) X - Axis Rotation, 4) Y - Axis Rotation
        EAR::Vector3 unbiased_vec;
        setAllMatricesAsIdentity();
        setTranslationMatrix(-1 * geoAngleCalculator.getV_origin());
        unbiased_vec = MatVecMul(geoAngleCalculator.getV_bias(), 0);
        setRotationMatrix(unbiased_vec, RotatingAxis::X_AXIS);
        setRotationMatrix(unbiased_vec, RotatingAxis::Y_AXIS);

        // 5) calcluate the transformation matrix to rotate given source location with respect to given rotating axis (e.g., an edge)
        getTransformationMatrix(3);

        if (rotatingAngleToMakeSrcVisible < 180) {
            rotatingAngleToMakeSrcVisible = 180 - rotatingAngleToMakeSrcVisible;
            setRotationMatrix(unbiased_vec, RotatingAxis::Z_AXIS, rotatingAngleToMakeSrcVisible);
        } else {
            rotatingAngleToMakeSrcVisible = rotatingAngleToMakeSrcVisible - 180;
            setRotationMatrix(unbiased_vec, RotatingAxis::Z_AXIS, -rotatingAngleToMakeSrcVisible);
        }

        getTransformationMatrix(7);
        EAR::Vector3 temp_srcloc_rot_XYZ = MatVecMul(geoAngleCalculator.getUj_tar(), 7);
        rsl = temp_srcloc_rot_XYZ;
    } else {
        throw "[ERROR] ALIGN_SOURCE_LOCATION_WITH_THE_GIVEN_EDGE:: Coordinate system is not YZX "
              "before";
        return geoAngleCalculator.getUj_tar();
    }

    geoAngleCalculator.setYZXCoordinates();

    return rsl;
}

} // namespace MPEGI
