/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "RTDiffractionPathTracker.h"
#include <math.h>

namespace MPEGI {

RTDiffractionPathTracker::RTDiffractionPathTracker(){};

RTDiffractionPathTracker::~RTDiffractionPathTracker(){};

bool RTDiffractionPathTracker::isTheDirectPathOccluded(std::unique_ptr<EAR::EmbreeTracer>& tracer)
{
    EAR::Vector3 relativeVectorFromSource = listener->listenerLocation - source->sourceLocation;
    bool occluded1 = false, occluded2 = false;

    occluded1 = tracer->isOccluded(source->sourceLocation, relativeVectorFromSource,
        Config::getFloat("Diffraction_GeometricErrorThreshold"));
    occluded2 = tracer->isOccluded(listener->listenerLocation, -1 * relativeVectorFromSource,
        Config::getFloat("Diffraction_GeometricErrorThreshold"));

    if (occluded1)
        source->sourceOccluder = tracer->firstHit(source->sourceLocation,
            EAR::cartesianToPolar(relativeVectorFromSource), relativeVectorFromSource.abs());
    if (occluded2)
        listener->listenerOccluder = tracer->firstHit(listener->listenerLocation,
            EAR::cartesianToPolar(-1 * relativeVectorFromSource), relativeVectorFromSource.abs());

    return (occluded1 || occluded2);
}

size_t RTDiffractionPathTracker::updatePathsToValidate(
    const std::vector<std::pair<int, int>>& pIndexList)
{
    pathsToValidate.clear();
    for (auto pIndex : pIndexList) {
        // Fetch corresponding path information
        const auto& edgeSequence =
            diffrMetadata->getPathDictByEdges().at(pIndex.first)[pIndex.second];
        const auto& faInSequence =
            diffrMetadata->getPathDictByFaceIndicators().at(pIndex.first)[pIndex.second];
        const auto& anglSequence =
            diffrMetadata->getPathDictByAngles().at(pIndex.first)[pIndex.second];

        // Check if the diffraction can occur from the given source around the initial edge
        float sourceAngle_UpperBound_InPath = anglSequence.front();
        float sourceAngle_UpperBound_InGeometry = edgeSequence.front()->edgeAngle;

        AngleCalculation iac_object_src(
            edgeSequence.front(), faInSequence.front(), source->sourceLocation);
        float sourceAngle_InPath = iac_object_src.calculateInternalAngle();

        float angleMargin = 3.0f;
        if ((sourceAngle_InPath < 0 - angleMargin)
            || (sourceAngle_InPath > sourceAngle_UpperBound_InGeometry + angleMargin)
            || (sourceAngle_InPath > sourceAngle_UpperBound_InPath + angleMargin)) {
            continue;
        }

        // Check if the diffraction can occur around the last edge to the listener position
        float listenerAngle_LowerBound_InPath = anglSequence.back();
        float listenerAngle_UpperBound_InGeometry = edgeSequence.back()->edgeAngle;

        AngleCalculation iac_object_lis(
            edgeSequence.back(), faInSequence.back(), listener->listenerLocation);
        float listenerAngle_InPath = iac_object_lis.calculateInternalAngle();

        if ((listenerAngle_InPath < 0 - angleMargin)
            || (listenerAngle_InPath > listenerAngle_UpperBound_InGeometry + angleMargin)
            || (listenerAngle_InPath < listenerAngle_LowerBound_InPath - angleMargin)) {
            continue;
        }

        //  Update pathsToValidate
        if (edgeSequence.size() == 1) {
            if ((listenerAngle_InPath - sourceAngle_InPath - 180.f) < 0 - angleMargin) {
                continue;
            }

            pathsToValidate.push_back(new PathToValidate());
            pathsToValidate.back()->pathType = DiffractionPathType::StaticPath;

            pathsToValidate.back()->diffractionAngleAtSource =
                std::fmax(listenerAngle_InPath - sourceAngle_InPath - 180.f, 0.f);

            pathsToValidate.back()->diffractionAngleAtListener = -1;
        } else if (edgeSequence.size() > 1) {
            pathsToValidate.push_back(new PathToValidate());
            pathsToValidate.back()->pathType = DiffractionPathType::StaticPath;
            pathsToValidate.back()->diffractionAngleAtSource =
                std::max(sourceAngle_UpperBound_InPath - sourceAngle_InPath, 0.f);
            pathsToValidate.back()->diffractionAngleAtListener =
                std::max(listenerAngle_InPath - listenerAngle_LowerBound_InPath, 0.f);
        }

        // update static-path data
        pathsToValidate.back()->geometryBitsId = -1;
        pathsToValidate.back()->initialEdgeID = edgeSequence.front()->edgeID;
        pathsToValidate.back()->lastEdgeID = edgeSequence.back()->edgeID;
        pathsToValidate.back()->localPathIndex = pIndex.second;

        EAR_CHECK_REL((diffrMetadata->getPathDistanceDict().find(pIndex.first)
                          != diffrMetadata->getPathDistanceDict().end()),
            "EAR Exception: key error when fetching path distance dictionary data");

        EAR_CHECK_REL(
            (diffrMetadata->getPathDistanceDict().at(pIndex.first).size() > pIndex.second),
            "EAR Exception: key error when fetching path distance dictionary data");

        pathsToValidate.back()->distance =
            diffrMetadata->getPathDistanceDict().at(pIndex.first)[pIndex.second]
            + (edgeSequence.front()->cp - source->sourceLocation).abs()
            + (listener->listenerLocation - edgeSequence.back()->cp).abs();

        // initialize occluded intermediate-path data
        pathsToValidate.back()->whichPathOccluded.clear();
        pathsToValidate.back()->intermediatePathOccluder.clear();
        pathsToValidate.back()->activated = false;
    }

    return pathsToValidate.size();
}

void RTDiffractionPathTracker::selectPaths(void)
{
    int itemNum = Config::getInt("DiffractionItemNumber");
    // Remove inactivated paths from pathsToValidate
    std::vector<int> indicesOfInactivePaths;
    for (int i = 0; i < (int)pathsToValidate.size(); i++) {
        if (!pathsToValidate[i]->activated)
            indicesOfInactivePaths.push_back(i);
    }

    for (auto it = indicesOfInactivePaths.rbegin(); it != indicesOfInactivePaths.rend(); ++it) {
        EAR_CHECK_REL((pathsToValidate.size() > *it),
            "EAR Exception: index error in a function call - selectPaths(void)");

        pathsToValidate.erase(pathsToValidate.begin() + *it);
    }

    nActivatedPaths = int(pathsToValidate.size());
    if (nActivatedPaths <= itemNum)
        return;

    // Update propagation distance measures to further reduce the number of paths
    std::vector<float> distances;
    for (auto& path : pathsToValidate) {
        EAR_CHECK_REL((path->distance > 0.f), "EAR Exception: path distance wrongly assigned");

        distances.push_back(path->distance);
    }

    // Update propagation distance measures to further reduce the number of paths
    std::sort(distances.begin(), distances.end(), [](const int a, const int b) { return a > b; });
    distanceThreshold = distances[int(distances.size()) - itemNum];

    std::vector<std::size_t> indices = getSortedIndices(distances);
    for (std::size_t i = 0; i != indices.size(); i++) {
        if (nActivatedPaths == itemNum) {
            break;
        }

        EAR_CHECK_REL((pathsToValidate.size() > i), "EAR Exception: path distances index error");

        auto& path = pathsToValidate[i];
        path->activated = false;
        nActivatedPaths--;
    }
}

// Function to returned sorted-out indices
std::vector<std::size_t> RTDiffractionPathTracker::getSortedIndices(const std::vector<float>& v)
{
    std::vector<std::size_t> idx(v.size());
    std::iota(idx.begin(), idx.end(), 0);
    std::stable_sort(
        idx.begin(), idx.end(), [&v](std::size_t i1, std::size_t i2) { return v[i1] > v[i2]; });
    return idx;
}

bool RTDiffractionPathTracker::updateOccludedPathData(std::unique_ptr<EAR::EmbreeTracer>& dyntracer)
{
    for (auto& path : pathsToValidate) {
        // Fetch corresponding path information
        const auto& edgeSequence =
            diffrMetadata->getPathDictByEdges().at(path->initialEdgeID)[path->localPathIndex];

        // Check if intermediate paths are occluded by dynamic geometries or not
        // Check the visibility from the last edge to the listener and update relevant occluded path information
        EAR::Vector3 posA = edgeSequence.back()->cp_vt;
        EAR::Vector3 posB = listener->listenerLocation;
        bool occluded = dyntracer->isOccluded(
            posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

        if (occluded) {
            // update occluded intermediate-path data
            path->intermediatePathOccluder.push_back(
                dyntracer->firstHit(posA, EAR::cartesianToPolar(posB - posA), (posB - posA).abs()));

            // Exceptional case: using adjacent voxel data it is possible
            if (path->intermediatePathOccluder.back()->isStatic()) {
                path->activated = false;
                continue;
            }

            path->whichPathOccluded.push_back(int(edgeSequence.size()));
            path->pathType = DiffractionPathType::CombinedPath;
            path->activated = false;
        } else {
            // To indicate this path is a static path which is ready to create/update diffraction items
            path->activated = true;
        }

        // Check the visibility for possible occluded intermediate paths
        posA = source->sourceLocation;
        posB = edgeSequence.front()->cp_vt;
        occluded = dyntracer->isOccluded(
            posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

        if (occluded) {
            path->whichPathOccluded.push_back(0);
            path->intermediatePathOccluder.push_back(
                dyntracer->firstHit(posA, EAR::cartesianToPolar(posB - posA), (posB - posA).abs()));
            path->pathType = DiffractionPathType::CombinedPath;
            path->activated = false;

            continue; // Go to the next path
        }

        for (size_t eIndex = 0; eIndex < edgeSequence.size() - 1; eIndex++) {
            const auto& eA = edgeSequence[eIndex + 0];
            const auto& eB = edgeSequence[eIndex + 1];

            posA = eA->cp_vt;
            posB = eB->cp_vt;
            occluded = dyntracer->isOccluded(
                posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

            if (occluded) {
                path->whichPathOccluded.push_back(eIndex + 1);
                path->intermediatePathOccluder.push_back(dyntracer->firstHit(
                    posA, EAR::cartesianToPolar(posB - posA), (posB - posA).abs()));
                path->pathType = DiffractionPathType::CombinedPath;
                path->activated = false;

                break; // Go to the next path
            }
        }
    }

    return true;
}

bool RTDiffractionPathTracker::updateOccludedDynamicPathData(
    std::unique_ptr<EAR::EmbreeTracer>& dyntracer,
    const std::vector<std::pair<int, int>>& pIndexList, int dynMeshBitsId)
{
    for (auto& path : pathsToValidate) {
        for (auto pIndex : pIndexList) {
            // Fetch corresponding path information
            // const auto& edgeSequence =
            //    diffrMetadata->getPathDictByEdges().at(path->initialEdgeID)[path->localPathIndex];
            auto& edgeSequence = diffrMetadata->getDynamicPathDictByEdges(dynMeshBitsId)
                                     .at(pIndex.first)[pIndex.second];
            // Check if intermediate paths are occluded by dynamic geometries or not
            // Check the visibility from the last edge to the listener and update relevant occluded     path information
            EAR::Vector3 posA = edgeSequence.back()->cp_vt;
            EAR::Vector3 posB = listener->listenerLocation;
            bool occluded = dyntracer->isOccluded(
                posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

            if (occluded) {
                // update occluded intermediate-path data
                path->intermediatePathOccluder.push_back(dyntracer->firstHit(
                    posA, EAR::cartesianToPolar(posB - posA), (posB - posA).abs()));

                // Exceptional case: using adjacent voxel data it is possible
                if (path->intermediatePathOccluder.back()->isStatic()) {
                    path->activated = false;
                    continue;
                }

                path->whichPathOccluded.push_back(int(edgeSequence.size()));
                path->pathType = DiffractionPathType::CombinedPath;
                path->activated = false;
            } else {
                // To indicate this path is a static path which is ready to create/update diffraction   items
                path->activated = true;
            }

            // Check the visibility for possible occluded intermediate paths
            posA = source->sourceLocation;
            posB = edgeSequence.front()->cp_vt;
            occluded = dyntracer->isOccluded(
                posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

            if (occluded) {
                path->whichPathOccluded.push_back(0);
                path->intermediatePathOccluder.push_back(dyntracer->firstHit(
                    posA, EAR::cartesianToPolar(posB - posA), (posB - posA).abs()));
                path->pathType = DiffractionPathType::CombinedPath;
                path->activated = false;

                continue; // Go to the next path
            }

            for (size_t eIndex = 0; eIndex < edgeSequence.size() - 1; eIndex++) {
                const auto& eA = edgeSequence[eIndex + 0];
                const auto& eB = edgeSequence[eIndex + 1];

                posA = eA->cp_vt;
                posB = eB->cp_vt;
                occluded = dyntracer->isOccluded(
                    posA, posB - posA, Config::getFloat("Diffraction_GeometricErrorThreshold"));

                if (occluded) {
                    path->whichPathOccluded.push_back(eIndex + 1);
                    path->intermediatePathOccluder.push_back(dyntracer->firstHit(
                        posA, EAR::cartesianToPolar(posB - posA), (posB - posA).abs()));
                    path->pathType = DiffractionPathType::CombinedPath;
                    path->activated = false;

                    break; // Go to the next path
                }
            }
        }
    }

    return true;
}

void RTDiffractionPathTracker::activateAllPathsToValidate(void)
{
    for (auto& path : pathsToValidate) {
        path->activated = true;
    }

    nActivatedPaths = int(pathsToValidate.size());
}

bool RTDiffractionPathTracker::updateDynamicPathsToValidate(
    const std::vector<std::pair<int, int>>& pIndexList, int dynMeshBitsId)
{
    for (auto pIndex : pIndexList) {

        auto itr = diffrMetadata->getDynamicPathDictByEdges(dynMeshBitsId).find(pIndex.first);
        EAR_CHECK_REL((itr != diffrMetadata->getDynamicPathDictByEdges(dynMeshBitsId).end()),
            "EAR Exception: path index with the given mesh bitstream id is not a valid key - "
            "updateDynamicPathsToValidate");

        auto& edgeSequence =
            diffrMetadata->getDynamicPathDictByEdges(dynMeshBitsId).at(pIndex.first)[pIndex.second];

        const auto& faInSequence = diffrMetadata->getDynamicPathDictByFaceIndicators(dynMeshBitsId)
                                       .at(pIndex.first)[pIndex.second];
        const auto& anglSequence = diffrMetadata->getDynamicPathDictByAngles(dynMeshBitsId)
                                       .at(pIndex.first)[pIndex.second];

        float sourceAngle_UpperBound_InPath = anglSequence.front();
        float listenerAngle_LowerBound_InPath = anglSequence.back();
        float sourceAngle_UpperBound_InGeometry = edgeSequence.front()->edgeAngle;
        float listenerAngle_UpperBound_InGeometry = edgeSequence.back()->edgeAngle;

        AngleCalculation iac_object_src(
            edgeSequence.front(), faInSequence.front(), source->sourceLocation);
        float sourceAngle_InPath = iac_object_src.calculateInternalAngle();

        AngleCalculation iac_object_lis(
            edgeSequence.back(), faInSequence.back(), listener->listenerLocation);
        float listenerAngle_InPath = iac_object_lis.calculateInternalAngle();

        if ((sourceAngle_InPath > sourceAngle_UpperBound_InPath)
            || (sourceAngle_InPath > sourceAngle_UpperBound_InGeometry)
            || (sourceAngle_InPath < 0)) {
            continue;
        }

        if ((listenerAngle_InPath < 0)
            || (listenerAngle_InPath > listenerAngle_UpperBound_InGeometry)
            || (listenerAngle_InPath < listenerAngle_LowerBound_InPath)) {
            continue;
        }

        if (edgeSequence.size() == 1) {
            if (listenerAngle_InPath - sourceAngle_InPath - 180.f < 0)
                continue;

            pathsToValidate.push_back(new PathToValidate());
            pathsToValidate.back()->pathType = DiffractionPathType::DynamicPath;
            pathsToValidate.back()->diffractionAngleAtSource =
                listenerAngle_InPath - sourceAngle_InPath - 180.f;
            pathsToValidate.back()->diffractionAngleAtListener = -1;
        } else if (edgeSequence.size() > 1) {
            pathsToValidate.push_back(new PathToValidate());
            pathsToValidate.back()->pathType = DiffractionPathType::DynamicPath;
            pathsToValidate.back()->diffractionAngleAtSource =
                sourceAngle_UpperBound_InPath - sourceAngle_InPath;
            pathsToValidate.back()->diffractionAngleAtListener =
                listenerAngle_InPath - listenerAngle_LowerBound_InPath;
        }

        pathsToValidate.back()->geometryBitsId = dynMeshBitsId;
        pathsToValidate.back()->initialEdgeID = pIndex.first;
        pathsToValidate.back()->lastEdgeID = edgeSequence.back()->edgeID;
        pathsToValidate.back()->localPathIndex = pIndex.second;

        pathsToValidate.back()->whichPathOccluded.clear();
        pathsToValidate.back()->intermediatePathOccluder.clear();
        pathsToValidate.back()->activated = true;
    }

    return (pathsToValidate.size() > 0);
} // namespace MPEGI

const std::vector<DiffractionItem*>& RTDiffractionPathTracker::getDiffractionItems(void) const
{
    return diffractionItems;
}

size_t RTDiffractionPathTracker::updateStaticDiffractionItems(void)
{
    //int nofActivatedPaths = 0;
    for (const auto& path : pathsToValidate) {
        if (!(path->activated)) {
            //std::cout << "(!(path->activated))" << std::endl;
            continue;
        }

        if (path->pathType != DiffractionPathType::StaticPath) {
            //std::cout << "(path->pathType != DiffractionPathType::StaticPath)" << std::endl;
            continue;
        }

        diffractionItems.push_back(
            new DiffractionItem(source, listener, path, diffrMetadata, eifPayload));
        bool isUpdated = diffractionItems.back()->updateStaticItem();
        if (!isUpdated) {
            diffractionItems.pop_back();
        } else {
            //nofActivatedPaths++;
        }
    }

    //std::cout << "pathsToValidate.size(), nofActivatedPaths: " << pathsToValidate.size() << ", "
    //          << nofActivatedPaths << std::endl;

    return diffractionItems.size();
}

size_t RTDiffractionPathTracker::updateOccludedStaticDiffractionItems(
    std::unique_ptr<EAR::EmbreeTracer>& tracer)
{
    //int nofActivatedPaths = 0;
    for (const auto& path : pathsToValidate) {
        if (!(path->activated)) {
            //std::cout << "(!(path->activated))" << std::endl;
            continue;
        }

        if (path->pathType == DiffractionPathType::StaticPath) {
            //std::cout << "(path->pathType != DiffractionPathType::StaticPath)" << std::endl;
            continue;
        }

        // It is implied that activated combined path will be validated
        diffractionItems.push_back(
            new DiffractionItem(source, listener, path, diffrMetadata, eifPayload));

        bool isUpdated = diffractionItems.back()->updateOccludedStaticItem(tracer);
        if (!isUpdated) {
            diffractionItems.pop_back();
        } else {
            //nofActivatedPaths++;
        }
    }

    //std::cout << "pathsToValidate.size(), nofActivatedPaths: " << pathsToValidate.size() << ", "
    //          << nofActivatedPaths << std::endl;

    return diffractionItems.size();
}

size_t RTDiffractionPathTracker::updateDynamicDiffractionItems(void)
{
    //int nofActivatedPaths = 0;
    for (const auto& path : pathsToValidate) {
        if (!(path->activated)) {
            //std::cout << "(!(path->activated))" << std::endl;
            continue;
        }

        if (path->pathType != DiffractionPathType::DynamicPath) {
            //std::cout << "(path->pathType != DiffractionPathType::DynamicPath)" << std::endl;
            continue;
        }

        diffractionItems.push_back(
            new DiffractionItem(source, listener, path, diffrMetadata, eifPayload));
        bool isUpdated = diffractionItems.back()->updateDynamicItem(path->geometryBitsId);
        if (!isUpdated) {
            diffractionItems.pop_back();
        } else {
            //nofActivatedPaths++;
        }
    }

    //std::cout << "pathsToValidate.size(), nofActivatedPaths: " << pathsToValidate.size() << ", "
    //<< nofActivatedPaths << std::endl;

    return diffractionItems.size();
}

void RTDiffractionPathTracker::updateItemWeights(void)
{
    if (diffractionItems.size() == 1) {
        return;
    }

    std::vector<float> angleVec;
    std::vector<int> itemIndex;
    float limit = Config::getFloat("DiffractionRegularizationAngle");
    for (int n = 0; n < diffractionItems.size(); n++) {
        auto& item = diffractionItems.at(n);
        if (item->getGain() > 0.f) {
            itemIndex.push_back(n);
            angleVec.push_back(item->getAngleSum());
        }
    }

    // compute gain values (without normalization)
    float sum{ 0.f };
    std::vector<float> gainVec;
    gainVec.reserve(angleVec.size());
    float reg = Config::getFloat("DiffractionRegularizationAngle");
    for (float angle : angleVec) {
        gainVec.push_back(1.f / (angle + reg));
        sum += gainVec.back();
    }

    // find maximum
    float max_val{ 0.f };
    for (float& gain : gainVec) {
        max_val = std::fmax(gain, max_val);
    }

    // apply gain culling
    float cullingGain = max_val * Config::getFloat("DiffractionCullingGain");
    for (float& gain : gainVec) {
        if (gain < cullingGain) {
            sum -= gain;
            gain = 0.f;
        }
    }

    // normalize gain to 1
    for (float& gain : gainVec) {
        gain /= sum;
    }

    // apply gain
    for (int n = 0; n < gainVec.size(); n++) {
        diffractionItems[n]->addGain(gainVec[n]);
    }
}

} // namespace MPEGI
