/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_RT_DIFFRACTION_PATH_TRACKER
#define MPEGI_RT_DIFFRACTION_PATH_TRACKER

#include <MPEGI/Bitstream/EIFPayload.h>

#include <EAR/Exception.h>
#include <EAR/Geometry.h>

#include "DiffractionItem.h"
#include "DiffractionMetadata.h"

namespace MPEGI {

// RT Diffraction Path Tracker/Handler
class RTDiffractionPathTracker
{
public:
    // Constructor
    RTDiffractionPathTracker();

    // Constructor
    RTDiffractionPathTracker(MPEGI::Source* _source, MPEGI::Listener* _listener,
        std::shared_ptr<MPEGI::DiffractionMetadata> diffrMetadata,
        std::shared_ptr<const MPEGI::EIFPayload> eifPayload)
        : source(_source)
        , listener(_listener)
        , diffrMetadata(diffrMetadata)
        , eifPayload(eifPayload)
    {}

    // Destructor
    ~RTDiffractionPathTracker();

    // Return true when the direct path is occluded
    bool isTheDirectPathOccluded(std::unique_ptr<EAR::EmbreeTracer>& tracer);

    // Update pathsToValidate private attribute without validation: only for static paths
    size_t updatePathsToValidate(const std::vector<std::pair<int, int>>&);

    // Return number of diffraction paths
    inline const std::size_t getNofPaths(void) const { return pathsToValidate.size(); }

    // Select a designated number of shortest paths out of the valid path list
    void selectPaths(void);

    // Return sorted indices
    std::vector<std::size_t> getSortedIndices(const std::vector<float>& v);

    // Update occluded intermediate data
    bool updateOccludedPathData(std::unique_ptr<EAR::EmbreeTracer>& dynTracer);

    // Update occluded intermediate dynamic data

    bool updateOccludedDynamicPathData(std::unique_ptr<EAR::EmbreeTracer>& dyntracer,
        const std::vector<std::pair<int, int>>& pIndexList, int);

    // Activate all the included paths
    void activateAllPathsToValidate(void);

    // Create new paths from path index list and update them considering positions of a source and the listener
    bool updateDynamicPathsToValidate(const std::vector<std::pair<int, int>>&, int);

    // Return the stored vector of diffraction item pointers
    const std::vector<DiffractionItem*>& getDiffractionItems(void) const;

    // Reset the stored vector of diffraction item pointers
    inline void resetDiffractionItems(void) { diffractionItems.clear(); }

    // Reset the stored vector of paths
    inline void resetPaths(void)
    {
        pathsToValidate.clear();
        nActivatedPaths = 0;
    }

    // Update diffraction items from activated and non-occluded static paths
    size_t updateStaticDiffractionItems(void);

    // Update diffraction items from activated and occluded static paths
    size_t updateOccludedStaticDiffractionItems(std::unique_ptr<EAR::EmbreeTracer>& tracer);

    // Update diffraction items from activated and non-occluded dynamic paths
    size_t updateDynamicDiffractionItems(void);

    // Normalize filters to prevent amplification by the multiple paths
    void updateItemWeights(void);

private:
    Source* source;
    Listener* listener;

    // Shared pointer of the diffraction metadata
    std::shared_ptr<MPEGI::DiffractionMetadata> diffrMetadata;

    // Shared pointer of the eif payload
    std::shared_ptr<const MPEGI::EIFPayload> eifPayload;

    // A list of pointers of validated paths
    std::vector<PathToValidate*> pathsToValidate;

    // A list of pointers of diffraction items
    std::vector<DiffractionItem*> diffractionItems;

    // Number of Activated paths = number of activated diffraction items
    int nActivatedPaths;

    // Distance threshold to decide which path can be rendered
    float distanceThreshold;
};

} // namespace MPEGI

#endif /* MPEGI_RT_DIFFRACTION_PATH_TRACKER */
