/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DirectionalFocus.h"

#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Stream/StreamManager.h>

#include <algorithm>

namespace MPEGI {

std::shared_ptr<RenderStage> makeDirectionalFocusStage()
{
    return std::make_shared<DirectionalFocusStage>();
}

DirectionalFocusStage::DirectionalFocusStage()
    : fs(0)
    , B(0)
{
    Config::getBoolDef("DirectionalFocus_enabled", false);
    Config::getFloatDef("DirectionalFocus_Aperture", EAR::Variant(30.f));
    Config::getFloatDef("DirectionalFocus_TransitionWidth", EAR::Variant(30.f));
    Config::getFloatDef("DirectionalFocus_StopbandAttenuationDb", EAR::Variant(20.f));
    Config::getBoolDef("DirectionalFocus_hasDirection", false);
    Config::getFloatDef("DirectionalFocus_directionYaw", EAR::Variant(0.f));
    Config::getFloatDef("DirectionalFocus_directionPitch", EAR::Variant(0.f));
}

DirectionalFocusStage::~DirectionalFocusStage() {}

std::unique_ptr<RenderStage::DspData> DirectionalFocusStage::update(RenderList& renderList)
{

    if (B == 0) { // only update once initialized
        return {};
    }

    std::set<StreamBuffer*> activeStreams;

    if (Config::getBool("DirectionalFocus_enabled")) {
        // Directional Focus Pre-processing
        float sinRefPitchRad = 0.f;
        float cosRefPitchRad = 1.f;
        float refYawRad = 0.f;
        float angDistDeg, startStopbandAngle, stopbandGainLinear, coefA, Aperture;
        EAR::Vector3 relLoc;
        EAR::Polar direction;
        Aperture = Config::getFloat("DirectionalFocus_Aperture");
        if (Config::getBool("DirectionalFocus_hasDirection")) {
            sinRefPitchRad =
                std::sin(Config::getFloat("DirectionalFocus_directionPitch") * (M_PI / 180.f));
            cosRefPitchRad =
                std::cos(Config::getFloat("DirectionalFocus_directionPitch") * (M_PI / 180.f));
            refYawRad = Config::getFloat("DirectionalFocus_directionYaw") * (M_PI / 180.f);
        }
        EAR_CHECK_REL(Aperture <= 180.f, "Aperture must be smaller than 180º");
        startStopbandAngle =
            std::min(180.f, Aperture + Config::getFloat("DirectionalFocus_TransitionWidth"));
        stopbandGainLinear =
            std::pow(10.f, -0.05 * Config::getFloat("DirectionalFocus_StopbandAttenuationDb"));
        coefA = -0.05 * Config::getFloat("DirectionalFocus_StopbandAttenuationDb")
            / (startStopbandAngle - Aperture);

        for (auto& item : renderList.items) {
            if (item->aparams & AParam::noDirectionalFocus) {
                continue;
            }

            if (item->type != ItemType::Primary && item->type != ItemType::Representative
                && item->type != ItemType::Reflection && item->type != ItemType::DirectionalReverb
                && item->type != ItemType::Diffraction
                && item->type != ItemType::Pannable /*&& item->type != ItemType::Reverb*/) {
                // only primary and directly derived render items processed here
                continue;
            }
            // skip the secondary items derived from extent and portal
            if ((item->extent && item->type == ItemType::Representative
                    && item->signal->getFrame().channels() == 1)
                || (item->type == ItemType::DirectionalReverb && item->parentRI != -1)) {
                continue;
            }

            if (item->status == ItemStatus::Orphaned) {
                continue;
            }

            if (item->status == ItemStatus::Inactive) {
                continue;
            }

            auto* itemState = item->getState<DirectionalFocusStage::ItemState>();
            // create new item-specific artifacts if needed
            if (itemState == nullptr || item->changed & ItemProperty::Signal) {
                itemState = &item->getStateOrCreate<DirectionalFocusStage::ItemState>();
                itemState->startGain = 1.f;
            }

            if ((!(item->aparams & AParam::NoDistanceGain)
                    || (item->aparams & AParam::NoDistanceGain
                        && item->type == ItemType::Representative))
                && item->position.cSpace != EAR::CSpace::LCS
                && (item->changed & ItemProperty::Position || renderList.listener.hasMoved
                    || item->changed & ItemProperty::Status)) {
                if (item->type != ItemType::Reflection) {

                    relLoc = renderList.listener.cs.transform(item->position.location);

                    direction[0] = static_cast<float>(std::atan2(-relLoc[0], -relLoc[2]));
                    direction[2] = relLoc.abs();

                    // avoiding sin(asin(x)) and cos(asin(x))
                    // https://www.wolframalpha.com/input/?i=cos(arcsin(x))

                    direction[1] =
                        static_cast<float>(relLoc[1] / direction[2]); //TOOO: prevent divZero?
                    angDistDeg =
                        static_cast<float>(std::acos(fmax(-1.f,
                            fmin(1.f,
                                sinRefPitchRad * direction[1]
                                    + cosRefPitchRad
                                        * std::sqrt(fmax(0.f, 1.f - direction[1] * direction[1]))
                                        * std::cos(refYawRad - direction[0])))))
                        * (180.f / M_PI);
                    itemState->startGain = 1.f;
                    item->changed |= ItemProperty::Gain;
                    if (angDistDeg > Aperture) {
                        if (angDistDeg >= startStopbandAngle) {
                            itemState->startGain = stopbandGainLinear;
                        } else {
                            itemState->startGain = std::pow(10.f, coefA * (angDistDeg - Aperture));
                        }
                    }
                }
            } else if (item->type == ItemType::Pannable && renderList.listener.hasMoved) {
                auto polar = EAR::cartesianToPolar(item->position.location);
                angDistDeg = std::acos(fmax(-1.f,
                                 fmin(1.f,
                                     std::cos(polar[1] * (M_PI / 180.f))
                                         * std::cos(polar[0] * (M_PI / 180.f)))))
                    * (180.f / M_PI);
                itemState->startGain = 1.f;
                item->changed |= ItemProperty::Gain;
                if (angDistDeg > Aperture) {
                    if (angDistDeg >= startStopbandAngle) {
                        itemState->startGain = stopbandGainLinear;
                    } else {
                        itemState->startGain = std::pow(10.f, coefA * (angDistDeg - Aperture));
                    }
                }
            }
            item->gain *= itemState->startGain;
        }
    }
    return {};
}

void DirectionalFocusStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    fs = streamManager->getFs();
    B = streamManager->getB();

    sceneState = state;
    this->streamManager = streamManager;
    //Config::setDef("DirectionalFocus_enabled", false);
}

} // namespace MPEGI
