/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DIRECTIONAL_FOCUS_STAGE
#define MPEGI_DIRECTIONAL_FOCUS_STAGE

#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <EAR/BaseTypes.h>
#include <EAR/SceneObject.h>

namespace MPEGI {

//! Modification of the gain values of individual RIs as specified for the directional focus, part of the accessibility feature.
/*!
 * Inputs:
 * - Items with ItemType::Primary, ItemType::Representative, ItemType::DirectionalReverb, ItemType::Diffraction, or , ItemType::Pannable
 *
 * Outputs:
 * - Same RIs, but with potentially updated gain field 
 *
 * How it works:
 * - The angular difference of the RI relative to the focus' main direction is computed 
 * - Based on this angular difference and the defined directional focus parameters, the gain value of the RI is modified
 *
 * Stages order:
 * - Required before: none
 * - Required after: EqStage
 *
 * Config options:
 * - (real-time adjustable) "DirectionalFocus_enabled": Bool to indicate if the directional focus is enabled
 * - (real-time adjustable) "DirectionalFocus_Aperture": Opening angle of the directional focus’ main lobe in degree.   
 * - (real-time adjustable) "DirectionalFocus_TransitionWidth": Width of the transition region between main lobe and stopband in degree.
 * - (real-time adjustable) "DirectionalFocus_StopbandAttenuationDb": Directional gain reduction in the stopband in decibel.
 * - (real-time adjustable) "DirectionalFocus_hasDirection": Bool to indicate if the directional focus has a non-default direction.
 * - (real-time adjustable) "DirectionalFocus_directionYaw": Yaw angle of the primary direction of the directional focus with respect to the frontal head orientation in degree.
 * - (real-time adjustable) "DirectionalFocus_directionPitch": Pitch angle of the primary direction of the directional focus with respect to the frontal head orientation in degree.
 *
 */
class DirectionalFocusStage : public RenderStage
{
public:
    //! Constructor
    DirectionalFocusStage();

    //! Destructor
    ~DirectionalFocusStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    //void processBlock(DspContext ctx) override;

private:
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    float fs;
    std::size_t B;

    const SceneState* sceneState;
    const StreamManager* streamManager;

    struct ItemState : RenderItem::ExtraState
    {
        float startGain;
    };
};

} // namespace MPEGI

#endif /* MPEGI_DIRECTIONAL_FOCUS_STAGE */
