/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Directivity.h"

#include <MPEGI/DirectivityCoding.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/ComplexBuffer.h>
#include <EAR/Exception.h>
#include <EAR/Listener.h>
#include <EAR/SofaReader.h>

#include <algorithm>
#include <complex>

namespace MPEGI {

std::shared_ptr<RenderStage> makeDirectivityStage()
{
    return std::make_shared<DirectivityStage>();
}

DirectivityStage::DirectivityStage() {}

DirectivityStage::~DirectivityStage() {}

std::unique_ptr<RenderStage::DspData> DirectivityStage::update(RenderList& renderList)
{
    const int numEqBands = Config::getInt("EQNBands");
    const float* centerFreqs = Config::getFloatPtr("EQCenterFreqs");
    const float cullingDeactivationThreshold =
        Config::getFloat("SimpleCulling_DeactivationThreshold");

    for (auto& item : renderList.items) {

        if (item->type != ItemType::Primary && item->type != ItemType::Representative
            && item->type != ItemType::Reflection && item->type != ItemType::Diffraction) {
            // only primary and directly derived render items processed here
            continue;
        }

        if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned) {
            continue;
        }

        if (item->directivity) {
            auto& magnitudes = item->getStateOrCreate<DirectivityStage::ItemState>().magnitudes;

            // set EQ change only if position or directivity/directiveness changed or skipeed last time
            if (item->changed & ItemProperty::Position
                || (item->position.cSpace != EAR::CSpace::LCS && renderList.listener.hasMoved)
                || item->changed & ItemProperty::Directivity
                || item->changed & ItemProperty::Status) {

                const EAR::Vector3& listenerLoc = renderList.listener.position.location;
                EAR::Vector3 listenerRelativePos = item->localCS.transform(listenerLoc);

                // find maximum EQ gain from prior stages

                float max_band_gain = *std::max_element(item->eq.begin(), item->eq.end());

                // check if we can skip the directivity calculation if culling stage would drop object
                // we use for this shortcut the maximum possible cover value for this item, and take
                // also the actual distance to object into account

                const float max_cover =
                    item->directivity->GetDirectivityCoverList().getDirectivityMaxCover();

                const float max_magnitude = std::pow(10.0f, max_cover / 20.f * item->directiveness);

                // calculate actual distance to render item
                const float distance = listenerRelativePos.abs();

                // calculate distiance gain, taking reference distance into account
                // avoid division by zero with a constant fraction expansion

                const float distanceGainFactor = max_band_gain * max_magnitude * item->gain;
                const float distanceGain =
                    distanceGainFactor * ((item->refDistance + 0.0001f) / (distance + 0.0001f));

                if (distanceGain < cullingDeactivationThreshold) {

                    // set magnitudes to zero
                    for (int i = 0; i < numEqBands; i++) {
                        magnitudes[i] = 0.0f;
                    }

                } else {
                    EAR::Polar coordinates = EAR::cartesianToPolar(listenerRelativePos);

                    // translate from "-180 to 180" range to "0 to 360"
                    if (std::abs(coordinates[0]) < 0.01f) {
                        coordinates[0] = 0.f;
                    } else if (coordinates[0] < 0.f) {
                        coordinates[0] = coordinates[0] + 360.f;
                    }

                    for (int i = 0; i < numEqBands; i++) {
                        float magnitude = item->directivity->get(centerFreqs[i], coordinates);
                        // account for possible directiveness
                        magnitudes[i] = std::pow(10.0f, magnitude / 20.f * item->directiveness);
                    };
                }
                item->changed |= ItemProperty::Eq;
            }

            item->addEq(magnitudes.data(), false);
        }
    }
    return {};
}

void DirectivityStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{}

} // namespace MPEGI
