/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DiscoverSess.h"
#include "EAR/GeometryBase.h"

#include <EAR/PointSource.h>
#include <EAR/Portal.h>
#include <MPEGI/Bitstream/PortalPayload.h>

#include <MPEGI/Renderer/PerceptualPointDistribution.h>

#include <cassert>

//HACK! for the empty rays
#include <unordered_set>

namespace MPEGI {

std::shared_ptr<RenderStage> makeDiscoverSessStage()
{
    return std::make_shared<DiscoverSessStage>();
}

using namespace MPEGI::EmbreeHelper;

// TODO: move to different place to share with ER stage
RTCDevice DiscoverSessStage::embreeDevice = nullptr;

DiscoverSessStage::DiscoverSessStage() {}

DiscoverSessStage::~DiscoverSessStage()
{
    for (auto& item : this->embreeMeshes) {
        rtcReleaseGeometry(item.second);
    }
    rtcReleaseScene(this->embreeScene);
}

// Anonymous namespace for local use
namespace {

void findSess(const struct RTCFilterFunctionNArguments* args);

struct IntersectContext
{
    IntersectContext()
    {
        rtcInitIntersectContext(&this->ctx);
        // incoherent is the default
        this->ctx.flags = RTC_INTERSECT_CONTEXT_FLAG_INCOHERENT;
        this->ctx.filter = findSess;
    }
    RTCIntersectContext ctx; // This must be the first field of the struct!
    std::unordered_map<unsigned int, float> hits; // Mapping from geomID to nearest hit distance
};

void findSess(const struct RTCFilterFunctionNArguments* args)
{
    assert(args->N == 1);

    // We want to get all hits, not only the first one:
    args->valid[0] = 0;

    auto* hit = (RTCHit*)args->hit;
    auto* ray = (RTCRay*)args->ray;
    auto* ctx = (IntersectContext*)args->context;
    auto* geom = (EAR::PointSource*)args->geometryUserPtr;

    auto [iter, inserted] = ctx->hits.try_emplace(geom->getID(), 0.0f);
    if (inserted || ray->tfar < std::abs(iter->second)) {
        // This is either the first hit for this extended source, or closer than the previous ones
        if (dot(ray, hit) <= 0) {
            // We hit the front side of the extended source
            iter->second = ray->tfar;
        } else {
            iter->second = -ray->tfar;
        }
    } else {
        // There was a closer hit with the same extended source already, we do nothing
    }
}

} // anonymous namespace

std::unique_ptr<RenderStage::DspData> DiscoverSessStage::update(RenderList& renderList)
{
    // NB: position/orientation of objectSource is ignored, only the extent counts!
    if (this->embreeMeshes.empty()) {
        // Nothing to do, there are no extended sources in the scene
        return {};
    }
    const size_t rayHitSubdivCount =
        size_t(std::max(0, Config::getIntDef("RayHitSubdivCount", 20)));

    // Go through all items (with extent) to see if Embree scene has to be updated
    bool updateScene = false;
    bool listenerMoved = (renderList.listener.coarseMoved || renderList.listener.coarseRotated);
    int numOfExtent = 0;
    for (auto& item : renderList.items) {

        if (item->extent == nullptr || item->status == ItemStatus::Inactive
            || item->type == ItemType::Primary && item->signal->getFrame().channels() > 1) {
            continue;
        }
        numOfExtent += 1;
        const EAR::Vector3 itemDir =
            item->position.location - renderList.listener.position.location;

        if ((listenerMoved && item->extent->getCSpace() == EAR::CSpace::LCS)
            || !item->extent->isStatic() && item->extent->hasMoved()) {

            RTCGeometry embreeMesh;
            try {
                embreeMesh = this->embreeMeshes.at(item->extent->getID());
            } catch (std::out_of_range&) {
                // Render item has extent, but wasn't added at initialization (e.g. HOA sources?)
                continue;
            }
            rtcUpdateGeometryBuffer(embreeMesh, RTC_BUFFER_TYPE_VERTEX, 0);
            rtcCommitGeometry(embreeMesh);
            updateScene = true;
        }
    }
    // no discovering is needed if there's no extent
    if (numOfExtent == 0) {
        return {};
    }

    // TODO: option to select the ray distribution
    const Vector3* rays = SESS_RAY_DISTRIBUTION;
    const auto rays_size = sizeof(SESS_RAY_DISTRIBUTION) / sizeof(Vector3);
    const int defaultExtraRayCount = 6;
    const float wt = 1.f / (defaultExtraRayCount + 1);

    if (updateScene) {
        rtcCommitScene(this->embreeScene);
        this->checkEmbreeDeviceError();
    }
    if (updateScene || listenerMoved) {
        for (auto& entry : this->rayHits) {
            // Clear "current" ray hits
            entry.second.second.clear();
        }

        IntersectContext ctx{};
        RTCRay ray;
        setOrg(ray, renderList.listener.position.location);
        EAR::Quaternion listenerRotation{ renderList.listener.position.orientation };
        std::unordered_map<EAR::ObjectID, RayHit> geomHits;
        //0.0534 is the average angle to the nearest ray for the current SESS_RAY_DISTRIBUTION
        const float r = 2.f * std::tan(0.0534) / 3.f;
        const float subrayNorm = std::sqrt(1.f + r * r);
        for (unsigned int rayId = 0; rayId < rays_size; ++rayId) {
            const EAR::Vector3 rayRotated = listenerRotation.rotate(rays[rayId]);
            setDir(ray, rayRotated);
            ray.tnear = 0;
            ray.tfar = std::numeric_limits<float>::infinity();
            ray.flags = 0;
            rtcOccluded1(this->embreeScene, &ctx.ctx, &ray);

            geomHits.clear();
            for (const auto& entry : ctx.hits) {
                const auto geomID = entry.first;
                const auto distance = entry.second;
                auto it = geomHits.find(geomID);
                if (it == geomHits.end()) {
                    geomHits.insert(std::make_pair(geomID, RayHit(rayId, distance, wt)));
                } else {
                    it->second.weight += wt;
                }
            }
            ctx.hits.clear();

            EAR::Vector3 axis0;
            if (rayId == 0) {
                axis0 = rays[rayId] - rays[rayId + 1];

            } else {
                axis0 = rays[rayId - 1] - rays[rayId];
            }
            axis0 = listenerRotation.rotate(axis0).normalize() * r;
            const EAR::Vector3 dir0 = rayRotated;
            const EAR::Vector3 axis1 = dir0.cross(axis0);
            for (int i = 0; i < defaultExtraRayCount; i++) {
                const float alpha = 2 * M_PI * i / defaultExtraRayCount;
                const float sa = std::sin(alpha);
                const float ca = std::cos(alpha);
                EAR::Vector3 dir1;
                for (int dim = 0; dim < 3; dim++)
                    dir1[dim] = (dir0[dim] + axis0[dim] * sa + axis1[dim] * ca) / subrayNorm;
                setDir(ray, dir1);
                ray.tnear = 0;
                ray.tfar = std::numeric_limits<float>::infinity();
                ray.flags = 0;
                rtcOccluded1(this->embreeScene, &ctx.ctx, &ray);

                for (const auto& entry : ctx.hits) {
                    const auto geomID = entry.first;
                    const auto distance = entry.second;
                    auto it = geomHits.find(geomID);
                    if (it == geomHits.end()) {
                        geomHits.insert(std::make_pair(geomID, RayHit(rayId, distance, wt)));
                    } else {
                        it->second.weight += wt;
                    }
                }
                ctx.hits.clear();
            }

            for (const auto& [geomID, rayHit] : geomHits) {
                if (rayHit.weight > 0.f)
                    this->rayHits[geomID].second.push_back(rayHit);
            }
        }
    } else {
        // No listener movement and no scene changes: previous rayHits are still valid
    }

    if (updateScene || listenerMoved) {
        // cast additional rays for the extents with few rayhits
        for (auto& item : renderList.items) {
            if (item->extent == nullptr || item->status == ItemStatus::Inactive) {
                continue;
            }
            try {
                const auto currID = item->extent->getID();
                auto& [previous, current] = this->rayHits.at(currID);

                if (!current.empty() && current.size() < rayHitSubdivCount) {
                    const int extraCounts[] = { 1, 6, 12 };
                    const int extraRaysCount = extraCounts[0] + extraCounts[1] + extraCounts[2];
                    IntersectContext ctx{};
                    RTCRay ray;
                    setOrg(ray, renderList.listener.position.location);
                    EAR::Quaternion listenerRotation{ renderList.listener.position.orientation };
                    for (RayHit& rh : current) {
                        EAR::Vector3 axis0;
                        if (rh.rayID == 0) {
                            axis0 = rays[rh.rayID] - rays[rh.rayID + 1];

                        } else {
                            axis0 = rays[rh.rayID - 1] - rays[rh.rayID];
                        }
                        const float r0 = 2.f * std::tan(0.0534) / 5.f;
                        axis0 = axis0.normalize();
                        const EAR::Vector3 axis1 = rays[rh.rayID].cross(axis0).normalize();
                        const EAR::Vector3 dir0 = rays[rh.rayID];
                        int hitCount = 0;
                        unsigned int subrayMask = 1;
                        unsigned int hitMask = 0;

                        for (int j = 0; j < 3; j++)
                            for (int i = 0; i < extraCounts[j]; i++) {
                                const float alpha = 2 * M_PI * i / extraCounts[j];
                                const EAR::Vector3 dir = dir0 + axis0 * r0 * j * std::sin(alpha)
                                    + axis1 * r0 * j * std::cos(alpha);
                                setDir(ray, listenerRotation.rotate(dir.normalize()));
                                ray.tnear = 0;
                                ray.tfar = std::numeric_limits<float>::infinity();
                                ray.flags = 0;
                                rtcOccluded1(this->embreeScene, &ctx.ctx, &ray);

                                for (const auto& entry : ctx.hits) {
                                    const auto geomID = entry.first;
                                    if (geomID == currID) {
                                        hitCount++;
                                        hitMask |= subrayMask;
                                        break;
                                    }
                                }
                                ctx.hits.clear();
                                subrayMask <<= 1;
                            }
                        rh.weight = float(hitCount) / float(extraRaysCount);
                        rh.refinementMask = hitMask;
                    }
                }

            } catch (std::out_of_range&) {
                // Render item has extent, but wasn't added at initialization (e.g. HOA sources?)
                continue;
            }
        }
    }

    for (auto& item : renderList.items) {
        if (item->extent == nullptr) {
            continue;
        }
        try {
            auto& [previous, current] = this->rayHits.at(item->extent->getID());
            item->rayHits = current;
            if (previous != current) {
                item->changed |= ItemProperty::RayHits;
            }

        } catch (std::out_of_range&) {
            // Render item has extent, but wasn't added at initialization (e.g. HOA sources?)
            continue;
        }
    }
    for (auto& entry : this->rayHits) {
        auto& [previous, current] = entry.second;
        previous = current;
    }
    return {};
}

void DiscoverSessStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    // TODO: code re-use with ER stage

    if (this->embreeDevice == nullptr) {
        this->embreeDevice = rtcNewDevice("threads=1,start_threads=1");
    }

    this->embreeScene = rtcNewScene(this->embreeDevice);

    bool dynamicScene = false;

    auto addExtent = [&](Geometry* extent) {
        const auto* mesh = extent->getMesh();
        if (mesh == nullptr) {
            EAR_FAIL("Only geometries with underlying meshes are supported as SESS");
        }

        RTCBuildQuality geometryBuildQuality;
        if (extent->isStatic() && extent->getCSpace() != EAR::CSpace::LCS) {
            geometryBuildQuality = RTC_BUILD_QUALITY_MEDIUM;
        } else {
            dynamicScene = true;
            geometryBuildQuality = RTC_BUILD_QUALITY_REFIT;
        }
        auto embreeMesh = addMesh(this->embreeScene, *mesh, geometryBuildQuality);

        rtcSetGeometryUserData(embreeMesh, extent);
        rtcCommitGeometry(embreeMesh);

        const auto id = extent->getID();
        this->rayHits[id]; // create empty entry
        this->embreeMeshes[id] = embreeMesh;
    };

    // Add primary extended sources
    for (const auto& source : state->getObjectSources()) {
        if (Geometry* extent = source->getExtent()) {
            // TODO: take only homogeneuous SESS?
            addExtent(extent);
        }
    }

    // Add portals
    auto portalPayload = std::dynamic_pointer_cast<const PortalPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_PORTAL));

    const bool isExplicitPortalMode = portalPayload->getExplicitPortalMode();
    for (auto& p : portalPayload->getPortals()) {
        if (isExplicitPortalMode) {
            // Add reverb extents
            addExtent((Geometry*)p->scenePortal->getExternalReverbExtents()[0]->getMesh());
            addExtent((Geometry*)p->scenePortal->getExternalReverbExtents()[1]->getMesh());

            // Add extent for coupling
            if (Config::getBoolDef("Portal_Enable_Coupling", true)) {
                addExtent((Geometry*)state->getGeometry(p->extentId.value())->getMesh());
            }
        } else {
            addExtent(state->getGeometry(p->extentId.value()));
        }
    }

    // TODO: add HOA sources from state->getHOASources()?

    auto sceneFlags = RTC_SCENE_FLAG_ROBUST | RTC_SCENE_FLAG_CONTEXT_FILTER_FUNCTION;
    if (dynamicScene) {
        rtcSetSceneBuildQuality(this->embreeScene, RTC_BUILD_QUALITY_LOW);
        sceneFlags = sceneFlags | RTC_SCENE_FLAG_DYNAMIC;
    } else {
        // TODO: RTC_BUILD_QUALITY_HIGH for better tracing performance?
        rtcSetSceneBuildQuality(this->embreeScene, RTC_BUILD_QUALITY_MEDIUM);
    }
    rtcSetSceneFlags(this->embreeScene, sceneFlags);

    rtcCommitScene(this->embreeScene);
    this->checkEmbreeDeviceError();
}

// TODO: code re-use with ER stage
void DiscoverSessStage::checkEmbreeDeviceError()
{
    switch (rtcGetDeviceError(this->embreeDevice)) {
    case RTC_ERROR_NONE:
        return;
    case RTC_ERROR_UNKNOWN:
        EAR_FAIL("Embree: An unknown error has occurred");
    case RTC_ERROR_INVALID_ARGUMENT:
        EAR_FAIL("Embree: An invalid argument was specified");
    case RTC_ERROR_INVALID_OPERATION:
        EAR_FAIL("Embree: The operation is not allowed for the specified object");
    case RTC_ERROR_OUT_OF_MEMORY:
        EAR_FAIL("Embree: There is not enough memory left to com- plete the operation");
    case RTC_ERROR_UNSUPPORTED_CPU:
        EAR_FAIL("Embree: The CPU is not supported as it does not support the lowest ISA "
                 "Embree is compiled for");
    case RTC_ERROR_CANCELLED:
        EAR_FAIL("Embree: The operation got canceled by a memory monitor callback or progress "
                 "monitor callback function");
    }
}

} // namespace MPEGI
