/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DISCOVER_SESS_STAGE
#define MPEGI_DISCOVER_SESS_STAGE

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include <MPEGI/Helpers/EmbreeHelper.h>

namespace MPEGI {

//! Send rays to discover spatially-extended sound sources (SESSs).
/*! 
 * Inputs: 
 * - Render items with extent, including portals
 *  
 * Outputs: 
 * - Modified render items with RenderItem::rayHits information
 *  
 * How it works: 
 * - Rays are sent into all directions in a virtual scene containing all extended sources
 * - Whenever they hit an extended source geometry,
 *   this information is stored in the corresponding render item
 * 
 * Stages order:
 * - Required before: none
 * - Required after: HomogeneousExtentStage (without it, extended sources are rendered as point sources)
 * - Must be before if used: PortalStage
 * - Must be after if used: OcclusionStage, DiffractionStage
 *  
 * Config options: 
 * - (real-time adjustable) "RayHitSubdivCount": ... (default: 20)
 * 
 */
class DiscoverSessStage : public RenderStage
{
public:
    DiscoverSessStage();
    ~DiscoverSessStage();

    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

private:
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    // TODO: code re-use with ER stage, move to some central place
    void checkEmbreeDeviceError();
    static RTCDevice embreeDevice;
    RTCScene embreeScene;

    using RayHitList = std::vector<RayHit>;
    // Pair of lists: previous hits and current hits
    std::unordered_map<unsigned int, std::pair<RayHitList, RayHitList>> rayHits;
    std::unordered_map<unsigned int, RTCGeometry> embreeMeshes;
};

} // namespace MPEGI

#endif /* MPEGI_DISCOVER_SESS_STAGE */
