/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Distance.h"

#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Stream/StreamManager.h>

#include <EAR/DistanceModels.h>
#include <EAR/FFT.h>
#include <EAR/Listener.h>
#include <EAR/RealBuffer.h>
#include <EAR/SceneObject.h>

#include <algorithm>

namespace MPEGI {

std::shared_ptr<RenderStage> makeDistanceStage()
{
    return std::make_shared<DistanceStage>();
}

DistanceStage::DistanceStage()
    : fs(0)
    , B(0)
    , enableDoppler(Config::getBoolDef("Distance_EnableDoppler", true))
{
    distanceModel = std::make_unique<EAR::DistanceGainVolumetric>(
        Config::getFloatDef("Distance_MinDistance", 0.2f));
    distanceModelListenerVoiceER = std::make_unique<EAR::DistanceGainVolumetric>(
        Config::getFloatDef("Distance_ERMinDistance", 1.0f));

    prepareAtmosphericAbsorption(Config::getFloatPtr("EQCenterFreqs"), Config::getInt("EQNBands"),
        Config::getFloat("Temperature"), Config::getFloat("Humidity"));

    if (enableDoppler) {
        smoothingFactor = Config::getFloatDef("Distance_SmoothingFactor", 0.5f);
    } else {
        smoothingFactor = 0.0f;
    }

    previousTeleportListener = true;
    previousListenerTrajTime = 0.0f;

    // Distance Gain Correction
    Config::getFloatDef("Distance_distExp", EAR::Variant(1.f));
}

DistanceStage::~DistanceStage() {}

std::unique_ptr<RenderStage::DspData> DistanceStage::update(RenderList& renderList)
{

    if (B == 0) { // only update once initialized
        return {};
    }

    float minDist = Config::getFloat("Distance_MinDistance");
    int nBand = Config::getInt("EQNBands");
    float distExp = fmax(0.f, Config::getFloat("Distance_distExp"));
    auto dspData = std::make_unique<DistanceStage::DspData>();

    dspData->listenerTrajectory = renderList.listener.trajectory;
    dspData->listenerTeleport = renderList.listener.teleport;
    dspData->ct = sceneState->getCurrentTime();

    EAR::Vector3 lPos = renderList.listener.trajectory->getLocation(dspData->ct);

    std::set<StreamBuffer*> activeStreams;

    float fsOverSpeedOfSound = fs / Config::getFloat("SpeedOfSound");
    for (auto& item : renderList.items) {

        if (item->type != ItemType::Primary && item->type != ItemType::Representative
            && item->type != ItemType::Reflection && item->type != ItemType::DirectionalReverb
            && item->type != ItemType::Diffraction) {
            // only primary and directly derived render items processed here
            continue;
        }
        // skip the secondary items derived from extent and portal
        if ((item->extent && item->type == ItemType::Representative
                && item->signal->getFrame().channels() == 1)
            || (item->type == ItemType::DirectionalReverb && item->parentRI != -1)) {
            continue;
        }

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        // signals of inactive items are also written to the delay buffer to be ready for reactivation
        activeStreams.insert(item->signal.get());
        // create new delay line for signal if it doesn't exist
        if (delayLines.find(item->signal.get()) == delayLines.end()) {
            delayLines[item->signal.get()] = std::make_unique<EAR::VDL>(fs, B,
                Config::getInt("Distance_MaxDelay"), item->signal->getFrame().channels(), false);
        }

        if (item->status == ItemStatus::Inactive) {
            continue;
        }

        EAR::Vector3 sPos;
        float dist, apparentWidth, apparentHeight;
        std::shared_ptr<Trajectory> trajectory;
        bool isSess = false;

        if (!item->extent) {
            trajectory = item->trajectory;

            if (item->type == ItemType::Diffraction) {
                // Find corresponding "direct"-signal
                auto direct = renderList.getItemByIndex(item->parentRI);
                if (direct) {
                    trajectory = direct->trajectory;
                }
            }

            sPos = trajectory->getLocation(dspData->ct);
            dist = (sPos - lPos).abs() + item->apparentDistanceDelta;

            apparentWidth = 0.f;
            apparentHeight = 0.f;

        } else {
            isSess = (item->type == ItemType::Primary && item->signal->getFrame().channels() == 1);
            EAR::Vector3 closestP;
            bool pointIsInside = false;
            item->extent->getApparentSize(renderList.listener.position, apparentWidth,
                apparentHeight, pointIsInside, closestP);

            if (pointIsInside) {
                sPos = lPos;
                dist = minDist;
            } else {
                sPos = closestP;
                dist = (sPos - lPos).abs();
            }

            dist += item->apparentDistanceDelta;

            // take velocity into account if it is available
            auto cv = std::dynamic_pointer_cast<ConstVelocity>(item->trajectory);
            if (cv) {
                trajectory =
                    std::make_shared<ConstVelocity>(sPos, cv->getVelocity(), cv->getTimestamp());
            } else {
                trajectory =
                    std::make_shared<ConstLocation>(sPos, item->trajectory->getTimestamp());
            }
        }

        auto* itemState = item->getState<DistanceStage::ItemState>();

        // create new item-specific artifacts if needed
        if (itemState == nullptr || item->changed & ItemProperty::Signal) {
            itemState = &item->getStateOrCreate<DistanceStage::ItemState>();

            // create new Eq arrays
            itemState->eq.resize(nBand);

            int flags = selectVdlMethod(*item, renderList);

            // create new read cursor for each item (regardless whether VDL is shared or not)
            itemState->readCursor = std::make_unique<EAR::VDL::ReadCursor>(
                delayLines[item->signal.get()]->registerReadCursor(
                    dist * fsOverSpeedOfSound, flags));
            // TODO: use cheaper algorithm for early reflections
        }

        if (!(item->aparams & AParam::NoDistanceGain)) {
            // calculate and apply distance gain
            if (item->changed & ItemProperty::Position
                || (item->position.cSpace != EAR::CSpace::LCS && renderList.listener.hasMoved)
                || item->changed & ItemProperty::Status) {

                // take care of the extra distance compensation of portal items
                float portalComp = 0.f;
                if ((item->type == ItemType::DirectionalReverb)
                    && (isExplicitPortalMode == false)) {
                    portalComp = item->portalDistComp;
                }

                //Use Distance model with longer minimun distance for ListenerVoice Reflections
                auto parent = renderList.getItemByIndex(item->parentRI);
                if (item->type == ItemType::Reflection && parent
                    && parent->type == ItemType::ListenerVoice) {
                    itemState->startGain =
                        distanceModelListenerVoiceER->getGainForDistance(dist, apparentWidth,
                            apparentHeight, item->refDistance, portalComp, isSess, distExp);
                } else {
                    itemState->startGain = distanceModel->getGainForDistance(dist, apparentWidth,
                        apparentHeight, item->refDistance, portalComp, isSess, distExp);
                }
                // indicate that gain is going to change
                item->changed |= ItemProperty::Gain;
            }
            item->gain *= itemState->startGain;
        }

        if (!(item->aparams & AParam::NoMediumAttenuation)) {
            // calculate and apply medium attenuation
            if (item->changed & ItemProperty::Position
                || (item->position.cSpace != EAR::CSpace::LCS && renderList.listener.hasMoved)
                || item->changed & ItemProperty::Status) {
                for (int i = 0; i < nBand; i++) {
                    // TODO: make cheaper with lookup
                    itemState->eq[i] = std::pow(10.0, (absorptionAlpha[i] * dist) / 20.0);
                }

                // indicate that eq is going to change
                item->changed |= ItemProperty::Eq;
            }

            item->addEq(itemState->eq.data(), false);
        }

        bool renderDoppler = enableDoppler && !(item->aparams & AParam::ForceNoDoppler);
        // && !(item->aparams & AParam::NoDoppler);

        auto& outBuffer = itemState->outBuffer;

        // split all items in separate buffers
        bool notifyChange = false;
        if (outBuffer == nullptr || item->changed & ItemProperty::Signal) {
            outBuffer = streamManager->createEmptyBuffer(item->signal->getFrame().channels());
            notifyChange = true;
        }

        // if we do not render doppler, we need to copy the original signal into the separate buffer
        EAR::RealFrame* copyPtr = nullptr;
        if (!renderDoppler) {
            copyPtr = &item->signal->getFrame();
        }

        // handle reactivation of items as teleport in doppler to avoid smoothing with deprecated position
        if (renderDoppler) {
            if ((item->changed & ItemProperty::Status) && (item->status == ItemStatus::Active)) {
                item->teleport = true;
            }
        }

        // save signal pointers for processBlock calls
        dspData->items.emplace_back(item->idx, delayLines[item->signal.get()].get(),
            itemState->readCursor.get(), &outBuffer->getFrame(), copyPtr, trajectory,
            item->refDistance, item->apparentDistanceDelta, renderDoppler, item->teleport);

        if (notifyChange) {
            item->setSignal(outBuffer);
        } else { // don't notify signal change if we did this before
            item->signal = outBuffer;
        }
    }

    for (auto ptr : activeStreams) {
        dspData->inputs.emplace_back(delayLines[ptr].get(), &(ptr->getFrame()));
        usedStreams.insert(ptr);
    }

    // every 10 times, remove unused delay lines here
    // TODO: if there is a better way, do it in cleanup function
    if (updateCount >= 20) { // TODO: magic number
        for (auto it = delayLines.begin(); it != delayLines.end();) {
            if (usedStreams.find(it->first) == usedStreams.end()) {
                it = delayLines.erase(it);
            } else {
                it++;
            }
        }
        usedStreams.clear();
        updateCount = 0;
    }
    updateCount++;

    return dspData;
}

void DistanceStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<DistanceStage::DspData>();

    int maxDelay = Config::getInt("Distance_MaxDelay");
    float speedOfSound = Config::getFloat("SpeedOfSound");
    bool detectedListenerTeleport = false;

    smoothingFactor = Config::getFloatDef("Distance_SmoothingFactor", 0.95f);
    float teleportThreshold = Config::getFloatDef("Distance_TeleportThreshold", 25.0f);

    if (currentDSP.listenerTrajectory) {
        // increase current time before to calculate positions at the end of the current block
        currentDSP.ct += tpb;

        listenerLocation = currentDSP.listenerTrajectory->getLocation(currentDSP.ct);

        // detect teleports from implausible listener speed, currentDSP.listenerTeleport flag is unreliable!
        if (previousListenerTraj != currentDSP.listenerTrajectory) {
            if (previousListenerTraj && (previousListenerTrajTime > 0.0f)) {
                float listenerMovement = (listenerLocation
                    - previousListenerTraj->getLocation(previousListenerTraj->getTimestamp()))
                                             .abs();
                float listenerTimeDiff = (float)(currentDSP.ct - previousListenerTrajTime);

                if (listenerTimeDiff > 0.0f) {
                    listenerMovement /= listenerTimeDiff;
                    if (listenerMovement > teleportThreshold) {
                        detectedListenerTeleport = true;
                    }
                }
            }
            // update time when listener position has last been updated
            // may have very low frequency in unity+AEP
            previousListenerTrajTime = currentDSP.ct;
        }

        for (auto& item : currentDSP.inputs) {
            item.delayLine->write(item.in->data().get());
        }

        int itemCtr = 0;
        for (auto& item : currentDSP.items) {

            itemLocation = item.trajectory->getLocation(currentDSP.ct);

            float dist = (itemLocation - listenerLocation).abs() + item.distComp;

            if (item.renderDoppler) { //IIR Smoothing of velocity
                float v = 0.0f;
                float vScene = 0.0f;
                double deltaTScene = 0.0;
                float distPrevScene = 0.0f;
                double t = currentDSP.ct;
                double deltaT = t - previousTimes[item.idx];
                float deltaTsamp = (float)(deltaT)*fs;
                float absVelocity = 0.0f;

                if (previousDistances[item.idx] && previousTrajectories[item.idx]) {
                    v = dist - previousDistances[item.idx];

                    auto cv = std::dynamic_pointer_cast<ConstVelocity>(item.trajectory);

                    // only update velocity when item state has updated
                    if (!cv) {
                        if ((item.trajectory->getTimestamp()
                                > previousTrajectories[item.idx]->getTimestamp())
                            || (currentDSP.listenerTrajectory != currentDSP.listenerTrajectory)) {
                            // estimate velocity for "constantPosition" items (where no velocity info is available)

                            EAR::Vector3 prevItemLoc = previousTrajectories[item.idx]->getLocation(
                                previousTrajectories[item.idx]->getTimestamp());

                            EAR::Vector3 prevListLoc = previousListenerTraj->getLocation(
                                previousListenerTraj->getTimestamp());

                            deltaTScene = (item.trajectory->getTimestamp()
                                - previousTrajectories[item.idx]->getTimestamp());

                            absVelocity = (itemLocation - prevItemLoc).abs() / deltaTScene;

                            // handle teleportation
                            if (!item.teleport && !detectedListenerTeleport) {
                                // normal case without teleport
                                distPrevScene = (prevItemLoc - prevListLoc).abs();
                            } else if (!item.teleport && detectedListenerTeleport) {
                                // listener teleported => consider only item movement
                                distPrevScene = (prevItemLoc - listenerLocation).abs();
                            } else if (item.teleport && !detectedListenerTeleport) {
                                /// item teleported => consider only listener movement
                                distPrevScene = (itemLocation - prevListLoc).abs();
                            } else {
                                // both teleported => no movement to consider
                                distPrevScene = dist;
                            }
                            // calculate movement speed between timestamps from scene update
                            // (NOT from dsp timestamp) for stable results and convert to block length
                            vScene = (dist - distPrevScene) / deltaTScene * (float)B / fs;

                            if (!smoothVelocities[item.idx]) {
                                if (vScene == 0.0f) {
                                    smoothVelocities[item.idx] = v;
                                } else {
                                    smoothVelocities[item.idx] = vScene;
                                }
                            }

                            // apply simple IIR smoothing to position to smooth out head tracking jitter etc., omit for teleport
                            if ((vScene != 0.0f) && !previousTeleportItem[item.idx]) {
                                smoothVelocities[item.idx] =
                                    smoothVelocities[item.idx] * smoothingFactor
                                    + vScene * (1.0f - smoothingFactor);
                            }

                            if (detectedListenerTeleport) {
                                smoothVelocities[item.idx] = vScene;
                            }

                            if (previousTimeDiff[item.idx]) {
                                if (deltaTScene > previousTimeDiff[item.idx] * 10.0f) {
                                    smoothDistances[item.idx] = dist;
                                    previousTimeDiff[item.idx] = deltaTScene;
                                    if (Config::getBoolDef(
                                            "Distance_ItemTeleportDetection", true)) {
                                        item.teleport = true;
                                    }
                                } else {
                                    previousTimeDiff[item.idx] = deltaTScene;
                                }

                            } else {
                                previousTimeDiff[item.idx] = deltaTScene;
                            }
                        }
                    } else {
                        // direct smoothing of velocity for "constantVelocity" items

                        if (!smoothVelocities[item.idx] || previousTeleportItem[item.idx]
                            || previousTeleportListener || currentDSP.listenerTeleport
                            || item.teleport) {
                            smoothVelocities[item.idx] = v;
                        }

                        // apply simple IIR smoothing to position to smooth out head tracking jitter etc., omit for teleport
                        if (!previousTeleportListener && !previousTeleportItem[item.idx]) {
                            if ((float)fabs(deltaTsamp - (float)B) < 10.0f) {
                                smoothVelocities[item.idx] =
                                    smoothVelocities[item.idx] * smoothingFactor
                                    + v * (1.0f - smoothingFactor);
                            }
                        }

                        absVelocity = cv->getVelocity().abs();
                    }

                    smoothDistances[item.idx] =
                        smoothDistances[item.idx] + smoothVelocities[item.idx];

                    // direct weighting of target distance to mitigate drift, TODO: tune smoothing constants
                    smoothDistances[item.idx] = smoothDistances[item.idx] * 0.995f + dist * 0.005f;

                    if (detectedListenerTeleport) {
                        smoothDistances[item.idx] = dist;
                    }

                    previousVelocities[item.idx] = v;
                } else {
                    smoothDistances[item.idx] = dist;
                }
                previousDistances[item.idx] = dist;
                previousTimes[item.idx] = t;

                // debug printout
                //fprintf(stderr, "%d %e %e %e %e %e %e\n", item.idx, dist, v, smoothDistances[item.idx],smoothVelocities[item.idx], t, item.trajectory->getTimestamp());
                //fprintf(stderr, "%d %e %e %e\n", item.idx, dist, smoothDistances[item.idx],
                //    detectedListenerTeleport * 1.0f + previousTeleportListener * 2.0f
                //        + currentDSP.listenerTeleport*4.0f);
                //    item.trajectory->getTimestamp());

                // update distance
                if (!item.teleport) {
                    dist = smoothDistances[item.idx];
                } else {
                    smoothDistances[item.idx] = dist;
                    smoothVelocities.erase(item.idx);
                }
                previousTeleportListener = detectedListenerTeleport;
                previousTeleportItem[item.idx] = item.teleport;
            }

            previousTrajectories[item.idx] = item.trajectory;
            previousListenerTraj = currentDSP.listenerTrajectory;

            // apply delay
            if (item.renderDoppler) {
                float dly = dist / speedOfSound * fs;

                if (dly < item.readCursor->getMinimumDelay()) {
                    dly = item.readCursor->getMinimumDelay();
                } else if (dly > maxDelay) {
                    //TODO: good idea to catch this here or rather throw an exception?
                    dly = static_cast<float>(maxDelay);
                    item.teleport = true;
                }

                item.delayLine->setDelay(*item.readCursor, dly,
                    item.teleport || currentDSP.listenerTeleport || detectedListenerTeleport);
                item.delayLine->read(*item.readCursor, item.out->data().get());

                item.teleport = false; // already handled this once...
                currentDSP.listenerTeleport = false;
            } else {
                // copy signal if no VDL is used
                item.out->copyFrom(*item.copySrc);
            }
        }
    }
}

void DistanceStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    fs = streamManager->getFs();
    B = streamManager->getB();
    tpb = B / fs;
    sceneState = state;
    this->streamManager = streamManager;
    updateCount = 0;

    auto portalPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_PORTAL);
    EAR_CHECK_REL(portalPayloads.size() <= 1, "Too many portal payloads in bitstream");
    if (portalPayloads.size() == 1) {
        isExplicitPortalMode =
            std::dynamic_pointer_cast<const MPEGI::PortalPayload>(portalPayloads[0])
                ->getExplicitPortalMode();
    }
}

void DistanceStage::prepareAtmosphericAbsorption(
    const float* freqs, int nfreq, double temp, double humidity, double pressure)
{
    // Ambient air temperature in Kelvin
    double tempKelv = temp + 273.15;

    // Reference air temperature (20C) in K
    double tempRef = 20.0 + 273.15;

    // Reference pressure at mean sea level (kPa)
    double pressureRef = 101.325;

    // Exponent C (Equation B.3)
    double C = -6.8346 * std::pow((273.16 / tempKelv), 1.261) + 4.6151;

    // Molar concentration of water vapour
    double h = humidity * std::pow(10.0, C) * (pressure / pressureRef);

    // Oxygen and nitrogen relaxation frequencies (Equation 3, 4)
    double frO = (pressure / pressureRef)
        * (24.0 + 4.04 * std::pow(10.0, 4.0) * h * (0.02 + h) / (0.391 + h));
    double exp = std::exp(-4.170 * (std::pow((tempKelv / tempRef), (-1.0 / 3.0)) - 1.0));
    double frN =
        (pressure / pressureRef) * std::pow((tempKelv / tempRef), -0.5) * (9.0 + 280.0 * h * exp);

    // Subexpressions in equation 5
    for (int i = 0; i < nfreq; i++) {
        float freq = freqs[i];
        double e1 = 8.686 * std::pow(freq, 2.0);
        double e2 = 0.0000000000184 * std::pow((pressure / pressureRef), -1.0)
            * std::pow((tempKelv / tempRef), 0.5);
        double e3 = std::pow((tempKelv / tempRef), (-5.0 / 2.0));
        double e4 = std::exp(-2239.1 / tempKelv);
        double e5 = std::pow(frO + std::pow(freq, 2.0) / frO, -1.0);
        double e6 = std::exp(-3352.0 / tempKelv);
        double e7 = std::pow(frN + std::pow(freq, 2.0) / frN, -1.0);

        absorptionAlpha.push_back(
            static_cast<float>(-1 * e1 * (e2 + e3 * (0.01275 * e4 * e5 + 0.1068 * e6 * e7))));
    }
}

int DistanceStage::selectVdlMethod(const MPEGI::RenderItem& item, MPEGI::RenderList& renderList)
{
    VDLMethod vdlMethod;
    int result = 0;

    if (item.type == ItemType::Reflection && -1 != item.parentRI) {
        int parentId = item.parentRI;
        auto parent = renderList.getItemByIndex(parentId);

        vdlMethod = parent->vdlMethod;
        if (vdlMethod == VDLMethod::spline_erLinear) {
            vdlMethod = VDLMethod::linear;
        }
    } else {
        vdlMethod = item.vdlMethod;
    }

    switch (vdlMethod) {
    case VDLMethod::spline:
        result = EAR::VDL::InterpolationNewton_Spline | EAR::VDL::QualityHigh;
        break;
    case VDLMethod::linear:
        result = EAR::VDL::InterpolationLinear;
        break;
    case VDLMethod::spline_erLinear:
        result = EAR::VDL::InterpolationNewton_Spline | EAR::VDL::QualityHigh;
        break;
    default:
        result = EAR::VDL::InterpolationNewton_Spline | EAR::VDL::QualityHigh;
    }

    return result;
}

} // namespace MPEGI
