/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_DISTANCE_STAGE
#define MPEGI_DISTANCE_STAGE

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Helpers/Trajectory.h>
#include <MPEGI/Renderer/RenderList.h>

#include <EAR/BaseTypes.h>
#include <EAR/SceneObject.h>
#include <EAR/VDL.h> // TODO: can be fwd declared?

#include <cstddef>
#include <map>
#include <memory>
#include <unordered_map>
#include <utility>

// fwd declares
namespace EAR {
class DistanceGainVolumetric;
class Listener;
class RealBuffer;
} // namespace EAR

namespace MPEGI {

// fwd declares
class SceneState;
class StreamBuffer;

//! Change the signal, gain and EQ of RIs based on their distance to the listener
/*!
 * Inputs:
 * - Items with ItemType::Primary, ItemType::Representative, ItemType::Reflection, ItemType::DirectionalReverb or ItemType::Diffraction
 *
 * Outputs:
 * - Same RIs, but with potentially updated EQ and gain field, as well as a delayed signal
 * - All output RIs of the considered types are guaranteed to have separate signal buffers
 *
 * How it works:
 * - The distance between source and listener is calculated
 * - The signal is delayed using variable delay lines (creating a doppler effect when the delay changes)
 * - The EQ is updated based on ISO 9613-1 medium attenuation
 * - The gain is updated based on a distance model that takes the size of the object into account
 *
 * Stages order:
 * - Required before: none
 * - Required after: EqStage
 *
 * Config options:
 * - (not real-time adjustable) "SpeedOfSound": default 340 m/s (relevant for the calculated delay)
 * - (not real-time adjustable) "Temperature": default: 20 deg celsius (relevant for medium attenuation)
 * - (not real-time adjustable) "Humidity": default: 40% (relevant for medium attenuation)
 * - (not real-time adjustable) "Distance_EnableDoppler": Whether or not the delay should be applied (default: true)
 * - (not real-time adjustable) "Distance_MinDistance": minimum distance to be rendered (default: 0.2m, lower bounded by the algorithmic delay of the delay line)
 *
 */
class DistanceStage : public RenderStage
{
public:
    //! Constructor
    DistanceStage();

    //! Destructor
    ~DistanceStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

private:
    friend class Renderer;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    //! Calculate the atmospheric absorption coefficient
    /*!
     *  This computation follows the methods given in ISO 9613-1
     *
     * \param freqs         pointer to frequencies of interest
     * \param nfreq         number of frequencies in freqs
     * \param temp          the given temperatutre [°C]
     * \param humidity      the relative humidity [%]
     * \param dist          the given distance [m]
     * \param pressure      the static pressure [kPa], default is pressure at mean sea level 101.325kPa
     */
    void prepareAtmosphericAbsorption(
        const float* freqs, int nfreq, double temp, double humidity, double pressure = 101.325f);

    //! Selects the VDL algorithm per render item
    int selectVdlMethod(const MPEGI::RenderItem& item, MPEGI::RenderList& renderList);

    std::vector<float> absorptionAlpha; // frequency dependent attenuation coefficients [dB / m]

    float fs;
    std::size_t B;
    float tpb; //!< time per block [s]

    bool enableDoppler;

    float smoothingFactor;

    //Distance model for RenderItems of type  Primary, Representative, Portal and Diffraction
    std::unique_ptr<EAR::DistanceGainVolumetric> distanceModel;
    //Distance model for RenderItems of type Reflection and child of type ListenerVoice
    std::unique_ptr<EAR::DistanceGainVolumetric> distanceModelListenerVoiceER;

    std::unordered_map<StreamBuffer*, std::unique_ptr<EAR::VDL>> delayLines;

    struct ItemState : RenderItem::ExtraState
    {
        std::unique_ptr<EAR::VDL::ReadCursor> readCursor;
        float startGain;
        std::vector<float> eq;
        std::shared_ptr<StreamBuffer> outBuffer;
    };

    std::set<StreamBuffer*> usedStreams;
    int updateCount; // used for removing old VDLs every N times

    struct DSPInput
    {
        EAR::VDL* delayLine;
        EAR::RealFrame* in;

        DSPInput(EAR::VDL* dl, EAR::RealFrame* in)
            : delayLine(dl)
            , in(in){};
    };

    struct DSPItem
    {
        int idx;
        EAR::VDL* delayLine;
        EAR::VDL::ReadCursor* readCursor;
        EAR::RealFrame* out;
        EAR::RealFrame* copySrc;
        std::shared_ptr<Trajectory> trajectory;
        float prevGain;
        float refDist;
        float distComp;
        bool renderDoppler;
        bool teleport;

        DSPItem(int idx, EAR::VDL* dl, EAR::VDL::ReadCursor* rc, EAR::RealFrame* out,
            EAR::RealFrame* cpy, std::shared_ptr<Trajectory> t, float rd, float dc, bool rDopp,
            bool teleport)
            : idx(idx)
            , delayLine(dl)
            , readCursor(rc)
            , out(out)
            , copySrc(cpy)
            , trajectory(t)
            , prevGain(1)
            , refDist(rd)
            , distComp(dc)
            , renderDoppler(rDopp)
            , teleport(teleport){};
    };

    std::unordered_map<int, std::shared_ptr<Trajectory>> previousTrajectories;
    std::shared_ptr<Trajectory> previousListenerTraj;
    double previousListenerTrajTime;

    std::unordered_map<int, float> previousVelocities;
    std::unordered_map<int, float> smoothVelocities;
    std::unordered_map<int, float> previousDistances;
    std::unordered_map<int, double> previousTimes;
    std::unordered_map<int, double> previousTimeDiff;
    std::unordered_map<int, float> smoothDistances;
    std::unordered_map<int, bool> previousTeleportItem;
    bool previousTeleportListener;

    struct DspData : RenderStage::DspData
    {
        std::vector<DSPInput> inputs;
        std::vector<DSPItem> items;

        std::shared_ptr<Trajectory> listenerTrajectory;
        bool listenerTeleport;
        double ct; //!< current scene time [s]
    };

    // following variables are used only in the DSP thread
    EAR::Vector3 itemLocation, listenerLocation;
    EAR::Vector3 forceVec;
    const SceneState* sceneState;
    const StreamManager* streamManager;
    bool isExplicitPortalMode = false;
};

} // namespace MPEGI

#endif /* MPEGI_DISTANCE_STAGE */
