/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "CylinderReflectorCalc.h"
#include <EAR/Exception.h>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES

#include <algorithm>
#include <cmath>

namespace MPEGI {

using namespace EAR;

/**************************************************************
 *                       public methods                       *
 **************************************************************/

CylinderReflectorCalc::CylinderReflectorCalc()
    : r(0.f)
    , len(0.f)
    , max_dim(0)
{}

CylinderReflectorCalc::CylinderReflectorCalc(const Cylinder& cylinder)
{

    updateCylinder(cylinder);
}

CylinderReflectorCalc::CylinderReflectorCalc(const Cylinder& cylinder, const HomogenMatrix& tform)
{

    updateTransformedCylinder(cylinder, tform);
}

void CylinderReflectorCalc::updateCylinder(const Cylinder& cylinder)
{

    auto circle_center = cylinder.getCenterOfBases();
    auto base = cylinder.getBasis();

    // starting point of the cylinder axis
    A = circle_center.first;

    // end point of the cylinder axis
    B = circle_center.second;

    Vector3 AB = B - A;
    Vector3 center = 0.5 * (A + B);

    // radius
    len = AB.abs();

    // third vector base: unit vector from A to B
    AB0 = AB * (1.f / len);

    // first vector base
    U = base[0] - center;
    U = U.normalize();

    // second vector base
    V = base[2] - center;
    V = V.normalize();

    // radius
    r = cylinder.getRadius();

    // dimension with greatest extend
    max_dim = 0;
    if (std::abs(AB0[1]) > std::abs(AB0[max_dim])) {
        max_dim = 1;
    }
    if (std::abs(AB0[2]) > std::abs(AB0[max_dim])) {
        max_dim = 2;
    }
}

void CylinderReflectorCalc::updateTransformedCylinder(
    const Cylinder& cylinder, const HomogenMatrix& tform)
{

    auto circle_center = cylinder.getCenterOfBases();
    auto base = cylinder.getBasis();

    // starting point of the cylinder axis
    A = tform * circle_center.first;

    // end point of the cylinder axis
    B = tform * circle_center.second;

    Vector3 AB = B - A;
    Vector3 center = 0.5 * (A + B);

    // radius
    len = AB.abs();

    // third vector base: unit vector from A to B
    AB0 = AB * (1.f / len);

    // first vector base
    U = tform * base[0] - center;
    U = U.normalize();

    // second vector base
    V = tform * base[2] - center;
    V = V.normalize();

    // radius
    r = cylinder.getRadius();

    // dimension with greatest extend
    max_dim = 0;
    if (std::abs(AB0[1]) > std::abs(AB0[max_dim])) {
        max_dim = 1;
    }
    if (std::abs(AB0[2]) > std::abs(AB0[max_dim])) {
        max_dim = 2;
    }
}

CVisible CylinderReflectorCalc::getReflection(
    const Vector3& S, const Vector3& L, Vector3& Y, Vector3& N0, float& apparent_dist) const
{

    // projection of the 2 points onto the cylinder axis
    Vector3 P_S, P_L; // FIXME: use Vector2
    float l_S, l_L;
    l_S = projectAxis(S, P_S);
    l_L = projectAxis(L, P_L);

    // S in local 2D coordinate system
    float u_S = U.dot(S - P_S);
    float v_S = V.dot(S - P_S);
    Vector3 S_uv(u_S, v_S, 0.f); // FIXME: use Vector2
    float az1 = std::atan2(v_S, u_S);
    float dist1 = S_uv.abs();

    // L in local 2D coordinate system
    float u_L = U.dot(L - P_L);
    float v_L = V.dot(L - P_L);
    Vector3 L_uv(u_L, v_L, 0.f); // FIXME: use Vector2
    float az2 = std::atan2(v_L, u_L);
    float dist2 = L_uv.abs();

    // check if listener or source are inside of the cylinder
    if ((dist1 <= r) || (dist2 <= r)) {
        apparent_dist = 0;
        return CVisible::NOT_VISIBLE;
    }

    // handle phase wrap-around
    if ((az2 - az1) > M_PI) {
        az1 += 2 * M_PI;
    } else if ((az1 - az2) > M_PI) {
        az2 += 2 * M_PI;
    }
    float az0 = (dist2 * az1 + dist1 * az2) / (dist1 + dist2);

    // Newton root search
    float az = az0;
    const float dx = 1e-3;
    float e;
    e = asymmetry(az, S_uv, L_uv);
    while (std::abs(e) > 1e-4) {
        constexpr float pi_2 = float(M_PI_2);
        float e2 = asymmetry(az + dx, S_uv, L_uv);

        // continue at a different angle, if we hit a local extremum
        if (e2 == e) {
            float az_plus = az + pi_2;
            float az_minus = az - pi_2;
            float e_plus = asymmetry(az_plus, S_uv, L_uv);
            float e_minus = asymmetry(az_minus, S_uv, L_uv);
            if (std::abs(e_plus) < std::abs(e_minus)) {
                az = az_plus;
                e = e_plus;
            } else {
                az = az_minus;
                e = e_minus;
            }
            continue;
        }

        float dy_dx = (e2 - e) / dx;
        float az_inc = -e / dy_dx;
        az_inc = std::max(az_inc, -pi_2);
        az_inc = std::min(az_inc, pi_2);
        az += az_inc;
        e = asymmetry(az, S_uv, L_uv);
    }

    Vector3 Y_uv = r * Vector3(cos(az), sin(az), 0.f); // FIXME: use Vector2

    // compute reflection point in global coordinates
    float len_S = (S_uv - Y_uv).abs();
    float len_L = (L_uv - Y_uv).abs();
    float mult_S = len_L / (len_S + len_L);
    float mult_L = 1.f - mult_S;
    auto P_Y = mult_S * P_S + mult_L * P_L;
    Y = P_Y + Y_uv[0] * U + Y_uv[1] * V;

    // compute surface normal
    N0 = Y - P_Y;
    N0 = N0.normalize();

    // check: reflection point is on the cylinder surface
    float l_Y = mult_S * l_S + mult_L * l_L;
    if ((l_Y < 0) || (l_Y > len)) {
        apparent_dist = 0;
        return CVisible::NOT_VISIBLE;
    }

    // check: reflection angle is smaller than 180°
    if (Y_uv.dot(S_uv - Y_uv) < 0) {
        Y.set(0.f, 0.f, 0.f);
        apparent_dist = 0;
        return CVisible::OCCLUDED;
    }

    // get magnification
    apparent_dist = getMagnification(S_uv, L_uv, Y_uv);

    return CVisible::VISIBLE;
}

/**************************************************************
 *                      private methods                       *
 **************************************************************/

float CylinderReflectorCalc::projectAxis(const Vector3& X, Vector3& P) const
{

    P = A + AB0 * AB0.dot(X - A);
    float l = (P[max_dim] - A[max_dim]) / AB0[max_dim];

    return l;
}

// FIXME: use Vector2 for S_uv, L_uv, and Y0
float CylinderReflectorCalc::asymmetry(float az, const Vector3& S_uv, const Vector3& L_uv) const
{

    Vector3 Y0(std::cos(az), std::sin(az), 0.f); // FIXME: use Vector2
    auto Y = r * Y0;
    auto X1 = S_uv - Y;
    auto X2 = L_uv - Y;
    float e = Y0.dot(X1.normalize()) - Y0.dot(X2.normalize());

    return e;
}

// FIXME: use Vector2 for S_uv, L_uv, and Y_uv
float CylinderReflectorCalc::getMagnification(
    const Vector3& S_uv, const Vector3& L_uv, const Vector3& Y_uv) const
{

    const float delta_az = 0.01; // small angular change on the circular reflector

    float dist_S = (S_uv - Y_uv).abs();
    float dist_L = (L_uv - Y_uv).abs();
    Vector3 dir_S = (S_uv - Y_uv) * (1.f / dist_S); // FIXME: use Vector2
    Vector3 dir_L = (L_uv - Y_uv) * (1.f / dist_L); // FIXME: use Vector2
    float az_refl = std::acos(std::clamp(dir_S.dot(dir_L), -1.0f, 1.0f)) / 2.f; // reflection angle

    // compute image scale factor on a planar reflector
    float scale_planar = (dist_S + dist_L) / dist_L;

    // compute image scale factor on a circular reflector
    float dY =
        r * std::tan(delta_az) * std::cos(0.5 * az_refl); // lateral shift on the circular reflector
    float dS = dist_S * std::tan(2 * delta_az) + scale_planar * dY; // lateral shift of the image
    float scale = dS / dY;

    // compute apparent distance
    float apparent_dist = std::sqrt(scale * scale_planar) * dist_L;

    return apparent_dist;
}

} // namespace MPEGI
