/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_CYLINDERREFLECTOR
#define MPEGI_CYLINDERREFLECTOR

#include "VisibilityCheck.h"
#include <EAR/Geometry.h>

// #include <cmath>

namespace MPEGI {

//! Class for computing reflections on cylinders.
/*!
 * This class implements a first order reflection on a cylinder.
 */
class CylinderReflectorCalc
{

public:
    //! Constructor
    CylinderReflectorCalc();

    //! Constructor
    /*!
     * \param cylinder          cylinder geometry
     */
    CylinderReflectorCalc(const EAR::Cylinder& cylinder);

    //! Constructor
    /*!
     * \param cylinder          cylinder geometry
     * \param tform             transformation matrix which is applied before the update
     */
    CylinderReflectorCalc(const EAR::Cylinder& cylinder, const EAR::HomogenMatrix& tform);

    //! update cylinder geometry and/or position
    /*!
     * \param cylinder          cylinder geometry
     */
    void updateCylinder(const EAR::Cylinder& cylinder);

    //! update transformed cylinder geometry and/or position
    /*!
     * \param cylinder          cylinder geometry
     * \param tform             transformation matrix which is applied before the update
     */
    void updateTransformedCylinder(const EAR::Cylinder& cylinder, const EAR::HomogenMatrix& tform);

    //! compute a specular reflection on this cylinder
    /*!
     * \param S                 input:  source location
     * \param L                 input:  listener location
     * \param Y                 output: reflection point
     * \param N0                output: normal vector at the reflection point
     * \param apparent_dist     output: apparent source distance
     * \return                  CVisible::NOT_VISIBLE | CVisible::VISIBLE | CVisible::OCCLUDED
     */
    CVisible getReflection(const EAR::Vector3& S, const EAR::Vector3& L, EAR::Vector3& Y,
        EAR::Vector3& N0, float& apparent_dist) const;

private:
    //! project a point on the axis of this cylinder
    /*!
     * \param X                 input:  point in Cartesian coordinates
     * \param P                 output: point on the cylinder axis in Cartesian coordinates
     * \return                  scalar which specifies where the projected point is located on the axis (should be between 0 and this::len)
     */
    float projectAxis(const EAR::Vector3& X, EAR::Vector3& P) const;

    //! function which is evaluated for Newton's root finding algorithm
    /*!
     * \param az                input: azimuth angle in radian
     * \param S_uv              input: source location in local 2D coordinate system {U,V}
     * \param L_uv              input: listener location in local 2D coordinate system {U,V}
     * \return                  difference / asymmetry value
     */
    float asymmetry(float az,
        const EAR::Vector3& S_uv, // FIXME: use Vector2
        const EAR::Vector3& L_uv) const; // FIXME: use Vector2

    //! compute apparent distance of a cylindrical reflection
    /*!
     * \param S_uv              input: source location in local 2D coordinate system {U,V}
     * \param L_uv              input: listener location in local 2D coordinate system {U,V}
     * \param Y_uv              input: reflection point in local 2D coordinate system {U,V}
     * \return                  apparent source distance
     */
    float getMagnification(const EAR::Vector3& S_uv, // FIXME: use Vector2
        const EAR::Vector3& L_uv, // FIXME: use Vector2
        const EAR::Vector3& Y_uv) const; // FIXME: use Vector2

private:
    //! starting point of the cylinder axis
    EAR::Vector3 A;

    //! end point of the cylinder axis
    EAR::Vector3 B;

    //! first base vector: unit vector from A to B
    EAR::Vector3 AB0;

    //! second base vector
    EAR::Vector3 U;

    //! third base vector
    EAR::Vector3 V;

    //! radius
    float r;

    //! radius
    float len;

    //! dimension with greatest extend
    unsigned int max_dim;
};

} // namespace MPEGI

#endif /* MPEGI_CYLINDERREFLECTOR */
