/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "EarlyReflection.h"

#include "SpecularReflectionPathFinder.h"
#include <MPEGI/Helpers/SpeedingTicket.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

#include <EAR/BaseTypes.h>
#include <EAR/Geometry.h>

#include <EAR/Buffer.h>
#include <EAR/ListenerVoiceSource.h>
#include <EAR/RealFrame.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectFactory.h>

#include <mutex>

namespace MPEGI {

namespace {
struct SpeedingTicketForItem : RenderItem::ExtraState
{
    SpeedingTicket ticket;
};
} // namespace

std::shared_ptr<RenderStage> makeEarlyReflectionStage()
{
    return std::make_shared<EarlyReflectionStage>();
}

EarlyReflectionStage::EarlyReflectionStage()
{
    Config::setDef("EarlyMaxOrderNoVoxel", EAR::Variant(1));
    Config::observeKey("EarlyMaxOrderNoVoxel", this);

    Config::observeKey("EarlyTuningGain", this);
    Config::observeKey("ListenerVoiceEarlyTuningGain", this);

    Config::setDef("EarlyLimitOrder", EAR::Variant(2));

    Config::setDef("EarlySpeedLimit", EAR::Variant(5.f));
    Config::observeKey("EarlySpeedLimit", this);

    Config::setDef("EarlyUpdateRate", EAR::Variant(4));
}

EarlyReflectionStage::~EarlyReflectionStage()
{
    Config::removeObserver("EarlyMaxOrderNoVoxel", this);
    Config::removeObserver("EarlyTuningGain", this);
    Config::removeObserver("ListenerVoiceEarlyTuningGain", this);
    Config::removeObserver("EarlySpeedLimit", this);
}

void EarlyReflectionStage::configValueChanged(const std::string& key)
{

    if (key == "EarlyTuningGain") {
        globalReflectionGain.store(Config::getFloat("EarlyTuningGain"));
        forceReconfiguration.store(true);
        return;
    }

    if (key == "EarlyMaxOrderNoVoxel") {
        std::scoped_lock lock(pathFinderMutex);

        pathFinder->delMesh(staticMesh);
        pathFinder->addMesh(Config::getInt("EarlyMaxOrderNoVoxel"), staticMesh, earlyPayload);

        forceReconfiguration.store(true);
        return;
    }

    if (key == "EarlySpeedLimit") {
        float speed_limit = Config::getFloat("EarlySpeedLimit");
        for (auto& speedTicket : itemStore->getItemStates<SpeedingTicketForItem>()) {
            speedTicket.ticket.setSpeedLimit(speed_limit);
        }
        return;
    }

    if (key == "ListenerVoiceEarlyTuningGain") {
        listenerVoiceERGain.store(Config::getFloat("ListenerVoiceEarlyTuningGain"));
        forceReconfiguration.store(true);
        return;
    }
}

std::unique_ptr<RenderStage::DspData> EarlyReflectionStage::update(RenderList& renderList)
{
    if (pathFinder == nullptr) {
        return {};
    }

    // generate a new constrained list of propagation paths for each primary render item, stored in paths_new
    updatePathsWithLimit(renderList);

    // create for each render item render items which
    // representatives for the image sources
    for (auto& parentItem : renderList.items) {

        auto& parentIdx = parentItem->idx;
        auto& childMap = renderItems[parentIdx];

        // skip irrelevant parent render items
        if (irrelevantRenderItem(*parentItem)) {
            continue;
        }

        // get item specific gain and skip speeding objects
        float itemGain = 0.f;
        if (auto* speedTicket = parentItem->getState<SpeedingTicketForItem>()) {
            itemGain = speedTicket->ticket.getGain();
        }
        if (itemGain == 0.f) {
            continue;
        }

        if (parentItem->status == ItemStatus::Inactive
            || parentItem->status == ItemStatus::FadeOutInactive) {
            // primary items are in different room or inactive themselves
            for (auto& [childIdx, childItem] : childMap) {
                itemStore->deactivate(childItem);
            }

            // empty common path list to read them correctly when reentering the room
            paths_old[parentIdx] = {};

            continue;
        }

        if (parentItem->status == ItemStatus::FadeOutOrphaned
            || parentItem->status == ItemStatus::Orphaned) {
            // trigger removal of secondary items of the obsolete primary
            for (auto& [childIdx, childItem] : childMap) {
                childItem->sync(*parentItem,
                    ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Loudness
                        | ItemProperty::Eq | ItemProperty::Aparams | ItemProperty::Status
                        | ItemProperty::RefDist);
            }
            continue;
        }

        std::set<SpecularReflection> pathsNewOnly;
        std::set<SpecularReflection> pathsCommon;
        std::set<SpecularReflection> pathsOldOnly;

        SpecularReflectionPathFinder::set_split(
            paths_new[parentIdx], paths_old[parentIdx], pathsNewOnly, pathsCommon, pathsOldOnly);

        paths_old[parentIdx] = paths_new[parentIdx]; // store the old paths

        // set render items which are no longer required to inactive
        for (const SpecularReflection& old_path : pathsOldOnly) {
            itemStore->deactivate(childMap[old_path.getValue()]);
        }

        // create new render items
        for (const SpecularReflection& path : pathsNewOnly) {

            auto childIdx = path.getValue();

            if (childMap.find(childIdx) == childMap.end()) {
                // create child render item, if it does not yet exist
                childMap[childIdx] = itemStore->createItem(parentItem->objectName
                    + "_earlyReflection_" + std::to_string(static_cast<int>(childIdx)));

                childMap[childIdx]->type = ItemType::Reflection;
                childMap[childIdx]->parentRI = parentIdx;
            } else {
                // item has been created before, but was inactive
                childMap[childIdx]->setStatus(ItemStatus::Active, true);
            }

            // pick the child render item which belongs to this reflection path
            auto& childItem = childMap[childIdx];

            // sync untouched properties of representatives with original item
            childItem->sync(*parentItem,
                ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Loudness
                    | ItemProperty::Eq | ItemProperty::Aparams | ItemProperty::Status
                    | ItemProperty::RefDist | ItemProperty::Signal);

            childItem->addEq(path.getEq().data());
            if (parentItem->type == ItemType::ListenerVoice) {
                childItem->gain *= itemGain * listenerVoiceERGain.load() * path.getPathLengthDelay()
                    / path.getPathLengthGain();
            } else {
                childItem->gain *= itemGain * globalReflectionGain.load()
                    * path.getPathLengthDelay() / path.getPathLengthGain();
            }
            // transformation matrix for world to local: y = localCS*imageR*x

            // create new ReflectionInfo with sequence of intersection points
            childItem->reflectionInfo = std::make_shared<ReflectionInfo>();
            childItem->reflectionInfo->intersectionSequence = path.getMirrorPoints();

            EAR::HomogenMatrix t = parentItem->localCS * path.getInverseOfTransformationMatrix();

            childItem->setPosition(
                EAR::Position(path.getImageSource(), EAR::Euler(NAN, NAN, NAN)), t);

            if (path.isOccluded()) {
                childItem->setStatus(ItemStatus::Inactive);
            }
        }

        //update the old but still necessary ones
        for (const SpecularReflection& path : pathsCommon) {

            // pick the child render item which belongs to this reflection path
            auto childIdx = path.getValue();
            auto& childItem = childMap[childIdx];

            // sync untouched properties of representatives with original item
            childItem->sync(*parentItem,
                ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Loudness
                    | ItemProperty::Eq | ItemProperty::Aparams | ItemProperty::Status
                    | ItemProperty::RefDist | ItemProperty::Signal);

            childItem->addEq(path.getEq().data());
            if (parentItem->type == ItemType::ListenerVoice) {
                childItem->gain *= itemGain * listenerVoiceERGain.load() * path.getPathLengthDelay()
                    / path.getPathLengthGain();
            } else {
                childItem->gain *= itemGain * globalReflectionGain.load()
                    * path.getPathLengthDelay() / path.getPathLengthGain();
            }

            // transformation matrix for world to local: y = localCS*imageR*x
            childItem->reflectionInfo->intersectionSequence = path.getMirrorPoints();

            EAR::HomogenMatrix t = parentItem->localCS * path.getInverseOfTransformationMatrix();
            childItem->setPosition(
                EAR::Position(path.getImageSource(), EAR::Euler(NAN, NAN, NAN)), t);
        }
    } // iteration render items

    std::vector<int> markedForRemoval;

    // add secondary render items to RenderList and DSP processing
    for (auto& [parentIdx, childMap] : renderItems) {

        auto parentItem = renderList.getItemByIndex(parentIdx);

        // iterate over all child render items
        for (auto& [childIdx, childItem] : childMap) {
            if (childItem->status == ItemStatus::FadeOutInactive) {
                // try deactivation until all fade outs are completed
                itemStore->deactivate(childItem);
            } else if (childItem->status == ItemStatus::FadeOutOrphaned) {
                // try removal until all fade outs are completed
                itemStore->remove(childItem);
            }
            renderList.addItemCopy(childItem);
            childItem->resetChanged();
        }

        if (parentItem->status == ItemStatus::Orphaned) {
            markedForRemoval.push_back(parentIdx);
        }
    }

    // delete obsolete secondary item handles
    for (int idx : markedForRemoval) {
        renderItems.erase(idx);
        paths_old.erase(idx);
    }

    forceReconfiguration.store(false);
    return {};
}

void EarlyReflectionStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    try {
        earlyPayload = std::dynamic_pointer_cast<const MPEGI::EarlyPayload>(
            bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_ER));
    } catch (std::exception&) {
        // Stage can work without payload PLD_ER
    }

    // renderer tuning parameters
    // 1. Default value
    // 2. Bitstream value
    // 3. Config value
    const float er_gain_lin = std::pow(10.f, BitstreamConfig::early_tuning_gain_default_db * .05f);
    globalReflectionGain.store(Config::getFloatDef("EarlyTuningGain", er_gain_lin));
    listenerVoiceERGain.store(Config::getFloatDef("ListenerVoiceEarlyTuningGain", er_gain_lin));
    configValueChanged("EarlyTuningGain");
    configValueChanged("ListenerVoiceEarlyTuningGain");

    if (!state) {
        return;
    }
    sceneState = state;

    this->itemStore = itemStore;

    std::vector<const EAR::Geometry*> geometries;
    const EAR::Mesh* mesh = state->getCombinedEarlyReflectionMesh()->getMesh();
    geometries.push_back(mesh);

    const auto& early_meshes = state->getEarlyReflectionGeometry();
    for (const auto geo : state->getAllVisibleGeometry()) {
        // if current geometry is not a early reflection static mesh, add it to path finder list
        if (std::find(early_meshes.begin(), early_meshes.end(), geo) == early_meshes.end()) {
            geometries.push_back(geo);
        }
    }
    for (const auto& source : state->getObjectSources()) {
        auto listenerVoice = dynamic_cast<EAR::ListenerVoiceSource*>(source.get());
        if (listenerVoice) {
            const float linGainERListenerVoice =
                std::pow(10.f, 0.05f * listenerVoice->getERGainDb());
            Config::set("ListenerVoiceEarlyTuningGain", EAR::Variant(linGainERListenerVoice));
            configValueChanged("ListenerVoiceEarlyTuningGain");
        }
    }
    createPathfinder(geometries);
}

void EarlyReflectionStage::createPathfinder(std::vector<const EAR::Geometry*> allGeometry)
{

    //first create path finder
    pathFinder = std::make_unique<SpecularReflectionPathFinder>();

    for (auto& geo : allGeometry) {
        //add cylinder
        if (geo->getType() == EAR::Geometry::Type::Cylinder) {
            pathFinder->addCylinder(static_cast<const EAR::Cylinder*>(geo), earlyPayload);
            continue;
        }

        //add sphere
        if (geo->getType() == EAR::Geometry::Type::Sphere) {
            //function yet not implemented
            continue;
        }

        // we need to modify some member variables of this mesh with
        // rebuildSurfaceStructureExtern() and rebuildRoundStructure()
        EAR::Mesh* mesh_rebuild =
            const_cast<EAR::Mesh*>(geo->getMesh()); //all other types are meshes
        EAR_CHECK_REL(mesh_rebuild != nullptr,
            "Geometric object is not a sphere nor a cylinder but has no mesh.");

        if (mesh_rebuild->getName() == "mesh:_combinedEarlyReflectionMesh") {
            EAR_CHECK_REL(mesh_rebuild->isStatic(), "sceneTrimeshER must be static");
            EAR_CHECK_REL(staticMesh == nullptr,
                "there can be only one \"mesh:_combinedEarlyReflectionMesh\" scene object");
            staticMesh = mesh_rebuild;
            if (earlyPayload != nullptr) {
                mesh_rebuild->rebuildSurfaceStructureExtern(earlyPayload->getSurfaceList(), true);
            } else {
                mesh_rebuild->rebuildSurfaceStructure();
            }
            mesh_rebuild->rebuildRoundStructure();
            pathFinder->addMesh(Config::getInt("EarlyMaxOrderNoVoxel"), mesh_rebuild, earlyPayload);
        } else {
            mesh_rebuild->rebuildSurfaceStructure();
            mesh_rebuild->rebuildRoundStructure();
            pathFinder->addMesh(Config::getInt("EarlyMaxOrderNoVoxel"), mesh_rebuild);
        }

    } // allGeometry loop
}

void EarlyReflectionStage::updatePathsWithLimit(const RenderList& renderList)
{
    std::scoped_lock lock(pathFinderMutex);

    // clear reflection path map
    paths_new.clear();

    // set order limitation
    unsigned int limit_order = Config::getInt("EarlyLimitOrder");
    pathFinder->setLimitOrder(limit_order);

    // get update rate for light- and heavy-weight updates
    unsigned int update_rate = Config::getInt("EarlyUpdateRate");

    // generate a list of visible reflection paths for all primary render items
    unsigned int itemCounter = 0;
    for (auto& parentItem : renderList.items) {

        const auto& parentIdx = parentItem->idx;
        const auto& sourceLocation = parentItem->position.location;
        const auto& listenerLocation = renderList.listener.position.location;

        // skip irrelevant parent render items
        if (irrelevantRenderItem(*parentItem)) {
            continue;
        }

        if (parentItem->status == ItemStatus::Inactive
            || parentItem->status == ItemStatus::FadeOutInactive
            || parentItem->status == ItemStatus::FadeOutOrphaned
            || parentItem->status == ItemStatus::Orphaned) {
            continue;
        }

        // update speeding ticket and skip speeding objects
        if (getItemGain(*parentItem) == 0.f) {
            continue;
        }

        if ((updateCounter + itemCounter) % update_rate == 0) {
            // heavy-weight update
            pathFinder->setId(parentIdx);
            paths_new[parentIdx] = pathFinder->getPaths(sourceLocation, listenerLocation);
        } else {
            // light-weight update
            for (auto spec : paths_old[parentIdx]) {
                Vector3 imgPos = spec.getTransformationMatrix() * sourceLocation;
                spec.setImageSource(imgPos);
                paths_new[parentIdx].insert(spec);
            }
        }

        itemCounter++;
    }

    // update counter for light- and heavy-weight updates
    updateCounter++;
}

bool EarlyReflectionStage::irrelevantRenderItem(const MPEGI::RenderItem& parentItem) const
{
    // image sources should only be rendered for primary and representative items
    if (parentItem.type != ItemType::Primary && parentItem.type != ItemType::Representative
        && parentItem.type != ItemType::ListenerVoice) {
        return true;
    }

    // skip listener-centric (channel) sources
    if (parentItem.position.cSpace == EAR::CSpace::LCS
        && parentItem.type != ItemType::ListenerVoice) {
        return true;
    }

    return false;
}

float EarlyReflectionStage::getItemGain(MPEGI::RenderItem& parentItem)
{
    float gain = 0.f;
    auto location = parentItem.position.location;
    auto timestamp = sceneState->getCurrentTime();
    int parentIdx = parentItem.idx;

    // check if we have a waypoint for this item
    auto* waypoint = parentItem.getState<Waypoint>();
    if (waypoint == nullptr) {
        // not found -> create waypoint map entry
        parentItem.getStateOrCreate<Waypoint>(location, timestamp);
        return gain;
    }

    auto* speedTicket = parentItem.getState<SpeedingTicketForItem>();

    if (timestamp == waypoint->trackedLocation.timestamp) {
        // same timestamp -> just update location and use latest gain
        waypoint->trackedLocation.location = location;
        if (speedTicket) {
            gain = speedTicket->ticket.getGain();
        }
        return gain;
    }

    // new timestamp -> compute speed
    Vector3 delta_X = location - waypoint->trackedLocation.location;
    float delta_t = float(timestamp - waypoint->trackedLocation.timestamp);
    float speed = delta_X.abs() / delta_t;

    // update waypoint
    waypoint->trackedLocation.location = location;
    waypoint->trackedLocation.timestamp = timestamp;

    // use speed ticketing system with hysteresis
    if (speedTicket == nullptr) {
        // speeding ticket does not exist yet -> create one
        // start with all sources muted till they are marked as being slower than the limit
        float speed_limit = Config::getFloat("EarlySpeedLimit");
        speedTicket = &parentItem.getStateOrCreate<SpeedingTicketForItem>();
        speedTicket->ticket.setSpeedLimit(speed_limit);
    } else {
        // check if the speed limit is exceeded and add the result to the records
        speedTicket->ticket.checkSpeed(speed);
    }

    gain = speedTicket->ticket.getGain();
    return gain;
}

} // namespace MPEGI
