/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_EARLY_REFLECTIONS
#define MPEGI_EARLY_REFLECTIONS

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include <MPEGI/Bitstream/EarlyPayload.h>

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/ConfigObserver.h>
#include <MPEGI/Helpers/SpecularReflection.h>

#include <EAR/RealBuffer.h>

#include <atomic>
#include <deque>
#include <map>
#include <mutex>
#include <utility>
#include <vector>

namespace EAR {
class Mesh;
class DegeneratedMesh;
class RealBuffer;
} // namespace EAR

namespace MPEGI {

class SpecularReflectionPathFinder;

struct TrackedLocation
{
    Vector3 location;
    double timestamp;
};

//! The EarlyReflectionStage adds specular reflections for sound sources in the scene
/*! 
 * Inputs: 
 * - All items of ItemType::Primary 
 * - All scene geometry 
 * - Extra "EarlyReflectionPacket" and "VoxelGridPacket" for acceleration 
 *  
 * Outputs: 
 * - New items of ItemType::Reflection, with the same properties and signal as the corresponding 
 *   primary item, except with a different position 
 *  
 * How it works: 
 * - All items of ItemType::Primary are reflected according to the geometry in the scene. For this, 
 *   reflection surfaces and round reflectors (cylinders, etc.) are detected in the encoder and 
 *   precalculated metadata is transmitted in the bitstream. 
 * - The stage adds new items of ItemType::Reflection for each specular reflection that is audible 
 *   based on the listener and source position, as well as geometry of the scene. The maximum order 
 *   is determined by the config value "SpecularReflectionOrder" with default 1. 
 * - The actual update of audible reflections is performed in a completely separate thread, as its 
 *   performance is currently not suitable for the Update thread. This way, the reflections and their 
 *   positions are updated less often than the listener position and rotation. For the reflection 
 *   position updates, however, atomicity is guaranteed, i.e. the reflections of all sources are  
 *   updated simultaneously. 
 * - The actual difference in propagation delay is then realized by the Distance stage. 
 *  
 * Stages order: 
 * - Required before: none 
 * - Required after: DirectivityStage, DistanceStage, EqStage, FadeStage
 *  
 * Config options: 
 * - (real-time adjustable) EarlyTuningGain : linear gain for all reflections (default: 0.4)
 * - (real-time adjustable) EarlyMaxOrderNoVoxel : maximum reflection order when voxel data are unavailable (brute force routine; default: 1)
 * - (real-time adjustable) EarlyLimitOrder : set reflection order limit for all reflections (default: 2)
 * - (real-time adjustable) EarlyFloorGainDB : gain in dB for floor reflections (default: -3.0)
 * - (real-time adjustable) EarlyFloorAngleDeg : angle in degree for identifying the floor (default: 15.0; valid range: ]0,90])
 * - (real-time adjustable) EarlySpeedLimit : speed limit in m/s (default: 5.0)
 * - (real-time adjustable) EarlyUpdateRate : update rate for heavyweight updates (default: 4)
 *
 */
class EarlyReflectionStage : public ConfigObserver, public RenderStage
{
public:
    //! Default Constructor
    EarlyReflectionStage();

    //! Destructor
    ~EarlyReflectionStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from ConfigObserver
    void configValueChanged(const std::string& key) override;

private:
    friend class Renderer;
    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    //! Creates a merged mesh := degeneratedMesh
    void createPathfinder(std::vector<const EAR::Geometry*> allGeometry);

    //! Generates a new constrained list of propagation paths for each primary render item.
    /*!
     * \param renderList           list of (primary) render items
     */
    void updatePathsWithLimit(const RenderList& renderList);

    //! Computes the EQ of each renderItem
    /*!
     * \param path                 path of renderItem
     * \param currentRoomId        current room Id --> room to be used for early reflection calculation
     */
    std::vector<float> calculateEQ(const SpecularReflection& path, int currentRoomId);

    //! Returns true, if given parent render item shall not be processed.
    /*!
     * \param parentItem           parent render item
     */
    bool irrelevantRenderItem(const MPEGI::RenderItem& parentItem) const;

    //! Returns the render item specific gain (speed limit etc.)
    /*!
     * \param parentItem           parent render item
     */
    float getItemGain(MPEGI::RenderItem& parentItem);

    const SceneState* sceneState{ nullptr };

    std::map<int, std::map<long long int, RenderItem*>> renderItems;
    // 1. map dimension: renderItem_idx         --> map of its SpecularReflection render Items
    // 2. map dimension: SpecularReflection_idx --> specific SpecularReflection

    RenderItemStore* itemStore;

    std::unique_ptr<SpecularReflectionPathFinder> pathFinder{ nullptr };
    EAR::Mesh* staticMesh{ nullptr };

    std::map<int, std::set<SpecularReflection>> paths_new;
    std::map<int, std::set<SpecularReflection>> paths_old;

    struct Waypoint : RenderItem::ExtraState
    {
        Waypoint(const Vector3& location, double timestamp)
            : trackedLocation{ location, timestamp }
        {}

        TrackedLocation trackedLocation;
    };

    std::mutex pathFinderMutex;
    std::atomic<bool> forceReconfiguration;

    std::atomic<float> globalReflectionGain;
    std::atomic<float> listenerVoiceERGain;

    std::shared_ptr<const MPEGI::EarlyPayload> earlyPayload{ nullptr };

    unsigned int updateCounter{ 0 };
};

} // namespace MPEGI

#endif /* MPEGI_EARLY_REFLECTIONS */
