/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SpecularCylinderReflector.h"
#include <EAR/Exception.h>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES

// #include <vector>

namespace MPEGI {

using namespace EAR;

SpecularCylinderReflector::SpecularCylinderReflector(
    const Cylinder* _cylinder, std::shared_ptr<const EarlyPayload> _earlyPayload)
    : SpecularReflector(ReflectorType::CYLINDER, 1)
    , cylinder(_cylinder)
    , cylinderCalc(*_cylinder)
    , max_dist_source(-1.f)
    , range_fadein_source(-1.f)
    , earlyPayload(_earlyPayload)
{

    EAR_CHECK_REL(cylinder->hasCircularBase(),
        "SpecularCylinderReflector expects a Cylinder with a circular base.");

    // set source culling distance
    if (_earlyPayload) {
        max_dist_source = float(_earlyPayload->getMaxDistSource1());
        range_fadein_source = max_dist_source / 5.0;
        if (range_fadein_source > 5.0) {
            range_fadein_source = 5.0; // maximum 5m fade-in range for source distance culling
        }
    }

    // initialize the cylinder calculator
    updateCylinder(_cylinder);
}

void SpecularCylinderReflector::updateCylinder(const Cylinder* _cylinder)
{

    cylinder = _cylinder;

    /* updeate cylinder calculator */
    cylinderCalc.updateCylinder(*_cylinder);

    // update total number of possible sorting values (only 1 reflection)
    num_values = 1;
}

void SpecularCylinderReflector::getPaths(const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<SpecularReflection>& path) const
{
    // perform source distance culling
    if ((max_dist_source > 0.f) && ((S - L).abs() > max_dist_source)) {
        return;
    }

    std::vector<int> seq = { 0 }; // skip direct sound
    std::vector<int> seqOccluded = {}; // not supported here
    Vector3 Y;
    Vector3 N0;
    float apparent_dist;

    // perform source distance culling
    if ((max_dist_source > 0.f) && ((S - L).abs() > max_dist_source)) {
        return;
    }

    CVisible visible = cylinderCalc.getReflection(S, L, Y, N0, apparent_dist);

    // SpecularReflection field elements
    int renderItemID = -1;
    std::vector<int> cylinderSeq;
    bool is_occluded = false;
    long long int value = -1;
    std::vector<Vector3> mirrorPoints = { Y };

    /* generate tform */
    float d = Y.dot(N0);
    Vector3 A = { 1.f, 0.f, 0.f };
    Vector3 B = { 0.f, 1.f, 0.f };
    Vector3 C = { 0.f, 0.f, 1.f };
    Vector3 O = { 0.f, 0.f, 0.f };
    A = A.mirror(N0, d);
    B = B.mirror(N0, d);
    C = C.mirror(N0, d);
    O = O.mirror(N0, d);
    HomogenMatrix tform = HomogenMatrix(A - O, B - O, C - O, O);
    HomogenMatrix tformInverse =
        tform; // first order reflection --> The inverse of tform equals tform

    value += value_start;

    /* FIXME: set renderItemID */

    // check if the other geometries occlude this image source
    if (visible == CVisible::VISIBLE) {
        visible = isVisible(S, L, mirrorPoints, reflectorMap, (SpecularReflector*)this);
        if (visible == CVisible::VISIBLE) {

            /* set cylinder ID */
            /* FIXME: set proper cylinder ID */

            float path_length = (L - Y).abs() + (S - Y).abs();

            Vector3 S_img = S.mirror(N0, d);
            is_occluded = false;

            float gain = fadeInSourceCulling(S, L) * path_length / apparent_dist;
            std::vector<float> eq = determineEq(seq, seqOccluded, gain);
            SpecularReflection spec(path_length, apparent_dist, S_img, renderItemID, seq,
                cylinderSeq, is_occluded, mirrorPoints, value, tform, tformInverse, eq);

            if (path.find(spec) == path.end()) {
                path.insert(spec);
            }
        }
    }
}

CVisible SpecularCylinderReflector::lineIntersect(
    const Vector3& P1, const Vector3& P2, std::vector<int>* triangleSeqOcclusion) const
{

    CVisible ret = CVisible::VISIBLE;

    /* auto base = cylinder->getBasis(); */
    auto circle_center = cylinder->getCenterOfBases();
    Vector3 axis = circle_center.second - circle_center.first;

    Vector3 corner = axis.cross(P2 - P1);
    corner = cylinder->getRadius() * corner.normalize();

    Vector3 A = circle_center.first + corner;
    Vector3 B = circle_center.first - corner;
    Vector3 C = circle_center.second - corner;
    Vector3 D = circle_center.second + corner;

    ret = line_triangle_intersect(P1, P2, A, B, C);
    if (ret == CVisible::VISIBLE)
        ret = line_triangle_intersect(P1, P2, C, D, A);

    return ret;
}

float SpecularCylinderReflector::fadeInSourceCulling(const Vector3& S, const Vector3& L) const
{
    float gain = 1.0f;

    if (max_dist_source <= 0.f) {
        return gain; // no source distance culling -> no fade-in
    }

    gain = (max_dist_source - (S - L).abs()) / range_fadein_source;

    // limit gain range to [0..1]
    if (gain > 1.0f) {
        gain = 1.0f;
    }
    if (gain < 0.0f) {
        gain = 0.0f;
    }

    return gain;
}

CVisible SpecularCylinderReflector::line_triangle_intersect(
    const Vector3& S, const Vector3& L, const Vector3& A, const Vector3& B, const Vector3& C) const
{

    Vector3 zero(0, 0, 0);
    HomogenMatrix M(B - A, C - A, S - L, zero);
    Vector3 X = M.inverse3x3() * (S - A);

    bool flag = (X[0] >= 0) && (X[1] >= 0) && (X[0] + X[1] <= 1) && (X[2] >= 0) && (X[2] <= 1);

    return (flag) ? CVisible::OCCLUDED : CVisible::VISIBLE;
}

std::vector<float> SpecularCylinderReflector::determineEq(
    const std::vector<int>& seq, const std::vector<int>& seqOcclusion, const float& gain) const
{
    (void)seq; // unused in this implementation, but kept to have consistent interface

    std::vector<float> eq(ThirdOctaveNBands, gain);

    std::shared_ptr<Material> material = cylinder->getMaterial();

    float f0 = 43 / cylinder->getRadius();
    unsigned int counter = 0;

    for (const float frequency : ThirdOctaveCenterFreqs) {
        //material filter (only first order specular reflection supported)
        if (material) {
            eq[counter] *= material->getReflectionCoeffForFreq(frequency);
        }
        //geometric filter (High-pass filter)
        eq[counter] *= std::sqrt(frequency * frequency / (f0 * f0 + frequency * frequency));
        ++counter;
    }

    return eq;
}

} // namespace MPEGI
