/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPECULARCYLINDERREFLECTOR
#define MPEGI_SPECULARCYLINDERREFLECTOR

#include <MPEGI/Bitstream/EarlyPayload.h>

#include "CylinderReflectorCalc.h"
#include "SpecularReflector.h"

#include <EAR/Geometry.h>

#include <map>

// #include <cmath>

namespace MPEGI {

//! Concrete class for reflecting planar geometries.
/*!
 * This class implements reflecting planar geometries.
 */
class SpecularCylinderReflector : public SpecularReflector
{

public:
    //! Constructor
    /*!
     * \param _cylinder         input: geometry data of the cylinder
     */
    SpecularCylinderReflector(
        const EAR::Cylinder* _cylinder, std::shared_ptr<const EarlyPayload> _earlyPayload = NULL);

    //! Update geometry data of the cylinder
    /*!
     * \param _cylinder         input: geometry data
     */
    void updateCylinder(const EAR::Cylinder* _cylinder) override;

    //! Get a list of visible or occluded image sources
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param path              output: list of specular reflections
     */
    void getPaths(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<SpecularReflection>& path) const override;

    //! Intersection test between a line and this reflecting object
    /*!
     * \param P1                    input: starting point of the line
     * \param P2                    input: end point of the line
     * \param triangleSeqOcclusion  output: end point of the line
     * \return                      VISIBLE: no intersection, OCCLUDED: intersection
     */
    CVisible lineIntersect(const EAR::Vector3& P1, const EAR::Vector3& P2,
        std::vector<int>* triangleSeqOcclusion = nullptr) const override;

    //! Get internal cylinder data
    /*!
    * \return                   geometry data
    */
    const EAR::Cylinder* getCylinder() { return cylinder; };

    //! Get linear gain for source distance culling fade-in
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \return                  linear gain
     */
    float fadeInSourceCulling(const EAR::Vector3& S, const EAR::Vector3& L) const;

private:
    CVisible line_triangle_intersect(const EAR::Vector3& S, const EAR::Vector3& L,
        const EAR::Vector3& A, const EAR::Vector3& B, const EAR::Vector3& C) const;

    std::vector<float> determineEq(const std::vector<int>& seq,
        const std::vector<int>& seqOcclusion, const float& gain) const override;

private:
    //! type of reflecting geometry
    const EAR::Cylinder* cylinder;

    //! computes reflections on cylyinders
    CylinderReflectorCalc cylinderCalc;

    //! source and triangle culling distances
    float max_dist_source;
    float range_fadein_source;

    std::shared_ptr<const EarlyPayload> earlyPayload{ NULL };

}; /* class SpecularCylinderReflector */

} /* namespace MPEGI */

#endif /* MPEGI_SPECULARCYLINDERREFLECTOR */
