/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SpecularPlanarReflector.h"
#include <EAR/Exception.h>
#include <MPEGI/Helpers/Config.h>
#include <algorithm>
#include <cassert>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES

#define EARLY_MAX_OCCLUSION 10

namespace MPEGI {

using namespace EAR;
using namespace MPEGI::EmbreeHelper;

RTCDevice SpecularPlanarReflector::embreeDevice = nullptr;

SpecularPlanarReflector::SpecularPlanarReflector(
    unsigned int _max_order, Mesh* _mesh, std::shared_ptr<const EarlyPayload> _earlyPayload)
    : SpecularReflector(ReflectorType::PLANAR, _max_order)
    , mesh(_mesh)
    , earlyPayload(_earlyPayload)
    , max_dist_triangle_1(-1.f)
    , max_dist_triangle_2(-1.f)
    , max_dist_source_1(-1.f)
    , max_dist_source_2(-1.f)
    , range_fadein_triangle_1(-1.0f)
    , range_fadein_triangle_2(-1.0f)
    , range_fadein_source_1(-1.0f)
    , range_fadein_source_2(-1.0f)
    , floor_thresh_y(0.0f)
    , floor_scale(0.0f)

{
    long long int ret = 0;
    long long int num_faces = this->mesh->getSurfaces().size();
    for (unsigned int n = 0; n < this->max_order; n++) {
        ret = (num_faces + 1) * ret + num_faces;
    }
    // update total number of possible sorting values (+1 for direct sound)
    this->num_values = ret + 1;

    // set source culling distance
    if (_earlyPayload) {
        max_dist_triangle_1 = float(_earlyPayload->getMaxDistTriangle1());
        max_dist_triangle_2 = float(_earlyPayload->getMaxDistTriangle2());
        max_dist_source_1 = float(_earlyPayload->getMaxDistSource1());
        max_dist_source_2 = float(_earlyPayload->getMaxDistSource2());
        range_fadein_triangle_1 = max_dist_triangle_1 / 5.0;
        range_fadein_triangle_2 = max_dist_triangle_2 / 5.0;
        range_fadein_source_1 = max_dist_source_1 / 5.0;
        range_fadein_source_2 = max_dist_source_2 / 5.0;
        if (range_fadein_triangle_1 > 5.0) {
            // maximum 5m fade-in range for triangle culling (1st order)
            range_fadein_triangle_1 = 5.0;
        }
        if (range_fadein_triangle_2 > 5.0) {
            // maximum 5m fade-in range for triangle culling (2nd order)
            range_fadein_triangle_2 = 5.0;
        }
        if (range_fadein_source_1 > 5.0) {
            // maximum 5m fade-in range for source distance culling (1st order)
            range_fadein_source_1 = 5.0;
        }
        if (range_fadein_source_2 > 5.0) {
            // maximum 5m fade-in range for source distance culling (2nd order)
            range_fadein_source_2 = 5.0;
        }
    }

    if (this->embreeDevice == nullptr) {
        this->embreeDevice = rtcNewDevice("threads=1,start_threads=1");
    }

    this->embreeScene = rtcNewScene(this->embreeDevice);

    // From the Embree manual:
    // If the RTC_BUILD_QUALITY_HIGH mode is set,
    // the filter functions may be called multiple times for the same primitive hit.
    // Further, rays hitting exactly the edge might also report two hits for the same surface.

    auto sceneFlags = RTC_SCENE_FLAG_ROBUST | RTC_SCENE_FLAG_CONTEXT_FILTER_FUNCTION;
    if (this->mesh->isStatic()) {
        // TODO: RTC_BUILD_QUALITY_HIGH for better tracing performance?
        rtcSetSceneBuildQuality(this->embreeScene, RTC_BUILD_QUALITY_MEDIUM);
    } else {
        rtcSetSceneBuildQuality(this->embreeScene, RTC_BUILD_QUALITY_LOW);
        sceneFlags = sceneFlags | RTC_SCENE_FLAG_DYNAMIC;

        // register as SceneObjectObserver
        this->mesh->attach(this);
    }
    rtcSetSceneFlags(this->embreeScene, sceneFlags);

    RTCBuildQuality geometryBuildQuality;
    if (this->mesh->isStatic()) {
        geometryBuildQuality = RTC_BUILD_QUALITY_MEDIUM;
    } else {
        geometryBuildQuality = RTC_BUILD_QUALITY_REFIT;
    }
    this->embreeMesh = addMesh(this->embreeScene, *this->mesh, geometryBuildQuality);
    rtcCommitScene(this->embreeScene);
    this->checkEmbreeDeviceError();

    // initialize config variables
    Config::setDef("EarlyFloorGainDB", EAR::Variant(-6.f));
    Config::setDef("EarlyFloorAngleDeg", EAR::Variant(15.f));
    updateFloorParameters();

    // register config observers
    Config::observeKey("EarlyFloorGainDB", this);
    Config::observeKey("EarlyFloorAngleDeg", this);

    // initialize transmission flags
    updateTransmissionFlags();
}

void SpecularPlanarReflector::checkEmbreeDeviceError()
{
    switch (rtcGetDeviceError(this->embreeDevice)) {
    case RTC_ERROR_NONE:
        return;
    case RTC_ERROR_UNKNOWN:
        EAR_FAIL("Embree: An unknown error has occurred");
    case RTC_ERROR_INVALID_ARGUMENT:
        EAR_FAIL("Embree: An invalid argument was specified");
    case RTC_ERROR_INVALID_OPERATION:
        EAR_FAIL("Embree: The operation is not allowed for the specified object");
    case RTC_ERROR_OUT_OF_MEMORY:
        EAR_FAIL("Embree: There is not enough memory left to com- plete the operation");
    case RTC_ERROR_UNSUPPORTED_CPU:
        EAR_FAIL("Embree: The CPU is not supported as it does not support the lowest ISA "
                 "Embree is compiled for");
    case RTC_ERROR_CANCELLED:
        EAR_FAIL("Embree: The operation got canceled by a memory monitor callback or progress "
                 "monitor callback function");
    }
}

SpecularPlanarReflector::~SpecularPlanarReflector()
{
    // unregister config observers
    Config::removeObserver("EarlyFloorGainDB", this);
    Config::removeObserver("EarlyFloorAngleDeg", this);

    this->mesh->detach(this);
    rtcReleaseGeometry(this->embreeMesh);
    rtcReleaseScene(this->embreeScene);
}

void SpecularPlanarReflector::configValueChanged(const std::string& key)
{
    if (key == "EarlyFloorGainDB" || key == "EarlyFloorAngleDeg") {
        updateFloorParameters();
    }
}

void SpecularPlanarReflector::updateMesh(unsigned int _max_order, Mesh* _mesh)
{
    EAR_CHECK_REL(_mesh == this->mesh, "invalid mesh");
    EAR_CHECK_REL(_max_order == this->max_order, "reflection order is not supposed to change");
    rtcUpdateGeometryBuffer(this->embreeMesh, RTC_BUFFER_TYPE_VERTEX, 0);
    rtcCommitGeometry(this->embreeMesh);
    rtcCommitScene(this->embreeScene);
    this->checkEmbreeDeviceError();

    // initialize transmission flags
    updateTransmissionFlags();
}

void SpecularPlanarReflector::updateOccludeOthers(
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap)
{
    this->mesh->resetOccludeOthers();
    for (auto& it : reflectorMap) {
        if (it.second->getType() == ReflectorType::PLANAR) {
            SpecularPlanarReflector& other = (SpecularPlanarReflector&)*it.second;
            if (other.getType() == ReflectorType::PLANAR) {
                this->mesh->updateOccludeOthers(other.getMesh());
            }
        }
    }
}

void SpecularPlanarReflector::getPaths(const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<SpecularReflection>& path) const
{
    bool use_voxeldata = earlyPayload != NULL;
    EarlyDataSet localDataSet;

    // perform source distance culling (1st order)
    if ((max_dist_source_1 > 0.f) && ((S - L).abs() > max_dist_source_1)) {
        return;
    }

    if (use_voxeldata) {
        try {
            localDataSet = earlyPayload->getLocalDataSet(L, S);
            return getPathsAccelerated(S, L, reflectorMap, localDataSet, path);
        } catch (std::out_of_range e) {
            // special case for finger snaps: try again with "source is listener" voxel data
            constexpr float listener_radius = 1.0f;
            if ((S - L).abs() < listener_radius) {
                try {
                    localDataSet = earlyPayload->getLocalDataSet(L, L);
                    return getPathsAccelerated(S, L, reflectorMap, localDataSet, path);
                } catch (std::out_of_range e) {
                    // do nothing - return getPathsBruteForce()
                }
            }
        }
    }
    return getPathsBruteForce(S, L, reflectorMap, path);
}

void findReflector(const struct RTCFilterFunctionNArguments* args);

struct IntersectContext
{
    IntersectContext()
    {
        rtcInitIntersectContext(&this->ctx);
        // incoherent is the default
        this->ctx.flags = RTC_INTERSECT_CONTEXT_FLAG_INCOHERENT;
        this->ctx.filter = findReflector;
    }
    RTCIntersectContext ctx; // This must be the first field of the struct!
    const Mesh::Surface* targetSurface = nullptr;
    // This has to be set to NOT_VISIBLE when a target surface is set
    CVisible visibility = CVisible::VISIBLE;
    std::optional<std::pair<unsigned int, unsigned int>> previousReflector = std::nullopt;
    const std::vector<bool>* isTransmissive = nullptr;
    std::array<unsigned int, EARLY_MAX_OCCLUSION> triangleSeqOcclusion;
    unsigned int num_occlusion = 0;
};

void findReflector(const struct RTCFilterFunctionNArguments* args)
{
    EAR_CHECK(args->N == 1, "only single rays are supported");

    auto* hit = (RTCHit*)args->hit;
    auto* ray = (RTCRay*)args->ray;
    auto* ctx = (IntersectContext*)args->context;

    EAR_CHECK(hit->geomID == 0, "only one geometry per scene");

    if (ctx->previousReflector
        && *ctx->previousReflector == std::pair{ hit->geomID, hit->primID }) {
        // The ray starts directly at the reflection surface,
        // so it might create a spurious hit which we ignore:
        args->valid[0] = 0;
        return;
    }

    auto dotProd = dot(ray, hit);

    if (ctx->targetSurface != nullptr) {
        for (auto id : ctx->targetSurface->triangleId) {
            if (id == hit->primID) {
                // We are hitting the expected surface

                if (dotProd >= 0) {
                    // We hit the back side of the reflector
                    ctx->visibility = CVisible::NOT_VISIBLE;
                    return;
                }

                if ((ctx->isTransmissive) && ((*ctx->isTransmissive)[hit->primID])) {
                    // the triangle is transmissive
                    ctx->visibility = CVisible::NOT_VISIBLE;
                    return;
                }

                ctx->visibility = CVisible::VISIBLE;
                return;
            }
        }
    }

    // ignore back side hits to allow sources in objects (front side from L->S is back side S->L)
    if (dotProd <= 0) {
        args->valid[0] = 0;
        return;
    }

    // check for occlusion
    if ((ctx->isTransmissive) && (ctx->num_occlusion < EARLY_MAX_OCCLUSION)
        && ((*ctx->isTransmissive)[hit->primID])) {
        // we hit a transmissive triangle
        ctx->triangleSeqOcclusion[ctx->num_occlusion] = hit->primID;
        ctx->num_occlusion++;
        args->valid[0] = 0;
        return;
    }

    // We hit the wrong triangle -> ray is occluded
    ctx->visibility = CVisible::OCCLUDED;
}

void SpecularPlanarReflector::iterateSurfaces(std::vector<Mesh::Surface>& surfaceSequence,
    const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<SpecularReflection>& path, std::set<int>& skipSurfaces) const
{
    unsigned int orderMax = std::min(this->max_order, this->limit_order);
    const auto& surfaces = this->mesh->getSurfaces();
    for (int id = 0; id < surfaces.size(); ++id) {
        if (std::isnan(surfaces[id].plane.d)) {
            // skip broken planes with colinear triangles
            continue;
        }
        if (!surfaceSequence.empty() && surfaceSequence.back().surfaceId == id) {
            // The same surface cannot be hit two times in a row
            continue;
        }
        surfaceSequence.push_back(surfaces[id]);
        this->traceReflectionPath(surfaceSequence, S, L, reflectorMap, path, skipSurfaces);

        if (surfaceSequence.size() < orderMax) {
            this->iterateSurfaces(surfaceSequence, S, L, reflectorMap, path, skipSurfaces);
        }
        surfaceSequence.pop_back();
    }
}

void SpecularPlanarReflector::traceReflectionPath(const std::vector<Mesh::Surface>& surfaceSequence,
    const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<SpecularReflection>& path, std::set<int>& skipSurfaces) const
{
    EAR_CHECK(surfaceSequence.size() >= 1, "empty surface sequence");

    // Create mirrored sources (using reverse surface order)

    std::vector<Vector3> mirroredSource{};
    mirroredSource.reserve(1 + surfaceSequence.size());
    mirroredSource.push_back(S);
    for (auto it = std::crbegin(surfaceSequence); it != std::crend(surfaceSequence); ++it) {
        mirroredSource.push_back(mirroredSource.back().mirror(it->plane.N0, it->plane.d));
    }

    Vector3 rayStart = L;
    IntersectContext ctx{};
    RTCRayHit rayHit;

    std::vector<Vector3> mirrorPoints;
    std::vector<int> triangleSeq;
    long long int value = -1;
    int renderItemID = -1;
    int triangleIdx = -1;

    // initialize occlusion parameters
    ctx.isTransmissive = &isTransmissive;

    // Ray from listener to highest-order mirror source.
    // If first surface is encountered, continue from hit point to next-lower mirror source.
    // And so on ... until the non-mirrored source is reached.

    for (auto i = 0;; ++i) {

        const Vector3 rayEnd = mirroredSource[mirroredSource.size() - 1 - i];

        setOrg(rayHit.ray, rayStart);
        setDir(rayHit.ray, rayEnd - rayStart);

        rayHit.ray.tnear = 0;
        rayHit.ray.tfar = 1.0; // fraction of "dir" vector length
        rayHit.ray.flags = 0;
        rayHit.hit.geomID = RTC_INVALID_GEOMETRY_ID;

        if (i < surfaceSequence.size()) {
            ctx.targetSurface = &surfaceSequence[i];
            // This will only become VISIBLE if we hit the expected surface:
            ctx.visibility = CVisible::NOT_VISIBLE;
        } else {
            ctx.targetSurface = nullptr;
        }

        rtcIntersect1(this->embreeScene, &ctx.ctx, &rayHit);

        if (ctx.visibility != CVisible::VISIBLE) {
            return;
        }
        if (ctx.targetSurface == nullptr) {
            // This was the last ray section, we can terminate the for loop
            break;
        }
        ctx.previousReflector.emplace(rayHit.hit.geomID, rayHit.hit.primID);
        rayStart = hitPosition(rayHit.ray);
        mirrorPoints.push_back(rayStart);
        triangleSeq.push_back(rayHit.hit.primID);
    }

    // check for occluding triangles
    std::vector<int> triangleSeqOcclusion;
    triangleSeqOcclusion.reserve(ctx.num_occlusion);
    for (int n = 0; n < ctx.num_occlusion; n++) {
        triangleSeqOcclusion.push_back(ctx.triangleSeqOcclusion[n]);
    }

    CVisible visible = CVisible::VISIBLE;

    // check if the other geometries occlude this image source
    visible = isVisible(
        S, L, mirrorPoints, reflectorMap, (SpecularReflector*)this, &triangleSeqOcclusion);

    if (visible == CVisible::VISIBLE) {
        Vector3 S_img = mirroredSource.back();

        // TODO: "seq" is used in legacy code, it should probably be removed at some point.
        std::vector<int> seq;
        seq.reserve(surfaceSequence.size());
        for (const auto& s : surfaceSequence) {
            seq.push_back(s.surfaceId);
        }
        unsigned int order = seq.size();

        value = getSortingValue(seq);

        float path_length = (L - S_img).abs();

        float floor_gain = 1.f;
        for (const auto& s : surfaceSequence) {
            floor_gain *= getFloorGain(s.plane.N0);
        }
        float gain = fadeInSourceCulling(S, L, order) * floor_gain;
        std::vector<float> eq = determineEq(triangleSeq, triangleSeqOcclusion, gain);

        // transformation matrix
        auto x = Vector3{ 1, 0, 0 };
        auto y = Vector3{ 0, 1, 0 };
        auto z = Vector3{ 0, 0, 1 };
        auto o = Vector3{ 0, 0, 0 };
        for (auto it = std::crbegin(surfaceSequence); it != std::crend(surfaceSequence); ++it) {
            x = x.mirror(it->plane.N0, it->plane.d);
            y = y.mirror(it->plane.N0, it->plane.d);
            z = z.mirror(it->plane.N0, it->plane.d);
            o = o.mirror(it->plane.N0, it->plane.d);
        }
        auto tform = HomogenMatrix{ x - o, y - o, z - o, o };

        // inverse transformation matrix
        x = Vector3{ 1, 0, 0 };
        y = Vector3{ 0, 1, 0 };
        z = Vector3{ 0, 0, 1 };
        o = Vector3{ 0, 0, 0 };
        for (auto it = std::cbegin(surfaceSequence); it != std::cend(surfaceSequence); ++it) {
            x = x.mirror(it->plane.N0, it->plane.d);
            y = y.mirror(it->plane.N0, it->plane.d);
            z = z.mirror(it->plane.N0, it->plane.d);
            o = o.mirror(it->plane.N0, it->plane.d);
        }
        auto tformInv = HomogenMatrix{ x - o, y - o, z - o, o };

        bool is_occluded = false;
        SpecularReflection spec(path_length, path_length, S_img, renderItemID, seq, triangleSeq,
            is_occluded, mirrorPoints, value, tform, tformInv, eq);

        if (path.find(spec) == path.end()) {
            path.insert(spec);

            // if this is a first order reflection, add it to the list
            if (seq.size() == 1) {
                skipSurfaces.insert(seq[0]);
            }
        }
    }
}

long long int SpecularPlanarReflector::getSortingValue(const std::vector<int>& seq) const
{
    constexpr float max_project = 0.5 * M_PI / 180.0; // corresponds to about 89.5 ... 90.5 degrees
    long long int value = 0;

    int num_faces = this->mesh->getSurfaces().size();
    unsigned int num_seq = seq.size();

    // check for 2nd order reflections at perpendicular edges
    if ((num_seq == 2) && (seq[0] > seq[1])) {
        const auto& surfaces = this->mesh->getSurfaces();
        auto N0 = surfaces[seq[0]].plane.N0;
        auto N1 = surfaces[seq[1]].plane.N0;
        auto project = N0.dot(N1);
        if (abs(project) < max_project) {
            for (unsigned int n = 0; n < num_seq; n++) {
                value = (num_faces + 1) * value + (seq[1 - n] + 1);
            }
            return value + value_start;
        }
    }

    for (unsigned int n = 0; n < num_seq; n++) {
        value = (num_faces + 1) * value + (seq[n] + 1);
    }

    return value + value_start;
}

void SpecularPlanarReflector::getPathsBruteForce(const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<SpecularReflection>& path) const
{
    if (this->limit_order == 0) {
        return;
    }

    std::set<int> skipSurfaces;

    std::vector<Mesh::Surface> surfaceSequence{};
    surfaceSequence.reserve(this->max_order);
    this->iterateSurfaces(surfaceSequence, S, L, reflectorMap, path, skipSurfaces);

    // get rounded edge reflections
    getPathsEdges(S, L, reflectorMap, skipSurfaces, path);

    // get rounded corner reflections
    getPathsCorners(S, L, reflectorMap, skipSurfaces, path);
}

void SpecularPlanarReflector::getPathsAccelerated(const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    EarlyDataSet earlyDataSet, std::set<SpecularReflection>& path) const
{
    std::set<int> skipSurfaces;
    auto surfaceList = earlyPayload->getSurfaceList();
    std::vector<Mesh::Surface> surfaceSequence{};
    for (auto& seq : earlyDataSet) {
        // apply order limitation
        if (seq.size() > this->limit_order) {
            continue;
        }
        surfaceSequence.clear();
        for (int idx : seq) {
            surfaceSequence.push_back(surfaceList[idx]);
        }
        this->traceReflectionPath(surfaceSequence, S, L, reflectorMap, path, skipSurfaces);
    }

    // get rounded edge reflections
    getPathsEdges(S, L, reflectorMap, skipSurfaces, path);

    // get rounded corner reflections
    getPathsCorners(S, L, reflectorMap, skipSurfaces, path);
}

CVisible SpecularPlanarReflector::lineIntersect(
    const Vector3& P1, const Vector3& P2, std::vector<int>* triangleSeqOcclusion) const
{

    IntersectContext ctx{};
    RTCRayHit rayHit;
    ctx.isTransmissive = &isTransmissive;
    setOrg(rayHit.ray, P1);
    setDir(rayHit.ray, P2 - P1);
    rayHit.ray.tnear = 0.001;
    rayHit.ray.tfar = 0.999; // fraction of "dir" vector length
    rayHit.ray.flags = 0;
    rayHit.hit.geomID = RTC_INVALID_GEOMETRY_ID;
    ctx.targetSurface = nullptr;
    rtcIntersect1(this->embreeScene, &ctx.ctx, &rayHit);

    // copy occluding triangles
    if (triangleSeqOcclusion) {
        for (int n = 0; n < ctx.num_occlusion; n++) {
            triangleSeqOcclusion->push_back(ctx.triangleSeqOcclusion[n]);
        }
    }

    return ctx.visibility;
}

void SpecularPlanarReflector::getPathsEdges(const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<int>& skipSurfaces, std::set<SpecularReflection>& path) const
{

    const auto& triangle = mesh->getTriangles();
    const auto& surface = mesh->getSurfaces();
    unsigned int num_triangle = triangle.size();
    std::vector<bool> available_side_S(num_triangle, false);
    std::vector<bool> available_side_L(num_triangle, false);
    std::vector<bool> side_S(num_triangle,
        false); // boolean lookup table: is the source on the visible side of the plane?
    std::vector<bool> side_L(num_triangle,
        false); // boolean lookup table: is the listener on the visible side of the plane?

    const auto& roundEdges = mesh->getRoundEdges();
    for (const auto& edge : roundEdges) {
        unsigned int vtxFrom = edge.vtxFrom; // first vertex index
        unsigned int vtxTo = edge.vtxTo; // second vertex index
        unsigned int tidLeft = edge.triLeft; // left triangle ID
        unsigned int tidRight = edge.triRight; // right triangle ID
        const auto& triLeft = mesh->getTriangle(tidLeft); // left triangle
        const auto& triRight = mesh->getTriangle(tidRight); // right triangle
        if (triLeft.surfaceId >= surface.size() || triRight.surfaceId >= surface.size()) {
            continue;
        }
        std::vector<unsigned int> surfaceIdList = { triLeft.surfaceId, triRight.surfaceId };

        const auto& planeLeft = surface[triLeft.surfaceId].plane;
        const auto& planeRight = surface[triRight.surfaceId].plane;
        Vector3 Simg_left;
        Vector3 Simg_right;
        Vector3 P_from;
        Vector3 P_to;
        float d_S, d_L, alpha;
        Vector3 Y_S, Y_L, Y;
        SpecularReflection spec;

        // skip transmissive edges
        if (isTransmissive[tidLeft] || isTransmissive[tidRight]) {
            continue;
        }

        // check if this edge reflection is already coverd by a planar 1st-order reflection
        bool alreadyCovered = false;
        for (unsigned int surfaceId : surfaceIdList) {
            if (skipSurfaces.find(surfaceId) != skipSurfaces.end()) {
                alreadyCovered = true;
                break;
            }
        }
        if (alreadyCovered) {
            continue;
        }

        // check if the source is on the visible side
        if (available_side_S[tidLeft] == false) {
            side_S[tidLeft] = mesh->pointPlaneDist(planeLeft, S) >= 0;
            available_side_S[tidLeft] = true;
        }
        if (available_side_S[tidRight] == false) {
            side_S[tidRight] = mesh->pointPlaneDist(planeRight, S) >= 0;
            available_side_S[tidRight] = true;
        }
        if (!side_S[tidLeft] && !side_S[tidRight]) {
            continue; // skip this edge, because the source is not on the reflecting side
        }

        // check if the listener is on the visible side
        if (available_side_L[tidLeft] == false) {
            side_L[tidLeft] = mesh->pointPlaneDist(planeLeft, L) >= 0;
            available_side_L[tidLeft] = true;
        }
        if (available_side_L[tidRight] == false) {
            side_L[tidRight] = mesh->pointPlaneDist(planeRight, L) >= 0;
            available_side_L[tidRight] = true;
        }
        if (!side_L[tidLeft] && !side_L[tidRight]) {
            continue; // skip this edge, because the listener is not on the reflecting side
        }

        // compute image source position mirrored on the left triangle
        if (side_S[tidLeft]) {
            Simg_left = S.mirror(planeLeft.N0, planeLeft.d);
        } else {
            Simg_left = S; // image source is occluded
        }

        // compute image source position mirrored on the right triangle
        if (side_S[tidRight]) {
            Simg_right = S.mirror(planeRight.N0, planeRight.d);
        } else {
            Simg_right = S; // image source is occluded
        }

        // create frustum
        P_from = mesh->getVertex(vtxFrom, Mesh::CSpace::Global);
        P_to = mesh->getVertex(vtxTo, Mesh::CSpace::Global);
        std::array<Mesh::Plane, 4> frustum = getEdgeFrustum(Simg_left, Simg_right, P_from, P_to);

        // compute reflection point Y, if L is located within the frustum
        if (!insideFrustum(L, frustum)) {
            continue;
        }
        d_S = pointLineDist(P_from, P_to, S, Y_S);
        d_L = pointLineDist(P_from, P_to, L, Y_L);
        alpha = d_L / (d_S + d_L);
        Y = Y_L + alpha * (Y_S - Y_L);

        if (getRoundSpec(S, L, Y, reflectorMap, tidLeft, triLeft.surfaceId, spec)) {
            path.insert(spec);
            skipSurfaces.insert(triLeft.surfaceId);
            skipSurfaces.insert(triRight.surfaceId);
        }
    }
}

void SpecularPlanarReflector::getPathsCorners(const Vector3& S, const Vector3& L,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
    std::set<int>& skipSurfaces, std::set<SpecularReflection>& path) const
{

    const auto& triangle = mesh->getTriangles();
    const auto& surface = mesh->getSurfaces();
    unsigned int num_triangle = triangle.size();
    std::vector<bool> available_side_S(num_triangle, false);
    std::vector<bool> available_side_L(num_triangle, false);
    std::vector<bool> side_S(num_triangle,
        false); // boolean lookup table: is the source on the visible side of the plane?
    std::vector<bool> side_L(num_triangle,
        false); // boolean lookup table: is the listener on the visible side of the plane?

    const auto& roundCorners = mesh->getRoundCorners();
    for (const auto& corner : roundCorners) {
        std::vector<unsigned int> surfaceIdList;
        for (unsigned int tid : corner.tidList) {
            Mesh::Triangle tri = mesh->getTriangle(tid);
            surfaceIdList.push_back(tri.surfaceId);
        }

        int num_correct_side = -1;
        std::vector<Vector3> Simg;
        Vector3 Y;
        std::vector<Mesh::Plane> frustum;
        int triangleId = -1;
        int surfaceId = -1;
        SpecularReflection spec;

        // check if this corner reflection is already coverd by a planar 1st-order reflection
        bool alreadyCovered = false;
        for (unsigned int surfaceId : surfaceIdList) {
            if (skipSurfaces.find(surfaceId) != skipSurfaces.end()) {
                alreadyCovered = true;
                break;
            }
        }
        if (alreadyCovered) {
            continue;
        }

        // check if the source is on the visible side
        num_correct_side = 0;
        for (unsigned int tid : corner.tidList) {
            if (available_side_S[tid] == false) {
                Mesh::Triangle tri = mesh->getTriangle(tid);
                Mesh::Plane plane;
                if (tri.surfaceId < size(surface)) {
                    plane = surface[tri.surfaceId].plane;
                } else {
                    plane = mesh->getPlane(tri);
                }
                side_S[tid] = mesh->pointPlaneDist(plane, S) >= 0;
                available_side_S[tid] = true;
                if (side_S[tid]) {
                    num_correct_side++;
                }
            }
        }
        if (num_correct_side == 0) {
            continue; // skip this corner, because the source is not on the reflecting side
        }

        // check if the listener is on the visible side
        num_correct_side = 0;
        for (unsigned int tid : corner.tidList) {
            if (available_side_L[tid] == false) {
                Mesh::Triangle tri = mesh->getTriangle(tid);
                Mesh::Plane plane;
                if (tri.surfaceId < size(surface)) {
                    plane = surface[tri.surfaceId].plane;
                } else {
                    plane = mesh->getPlane(tri);
                }
                side_L[tid] = mesh->pointPlaneDist(plane, L) >= 0;
                available_side_L[tid] = true;
                if (side_L[tid]) {
                    num_correct_side++;
                }
            }
        }
        if (num_correct_side == 0) {
            continue; // skip this corner, because the listener is not on the reflecting side
        }

        // check if the material is transmissive
        bool is_trans = false;
        for (unsigned int tid : corner.tidList) {
            if (isTransmissive[tid]) {
                is_trans = true;
                break;
            }
        }
        if (is_trans) {
            continue; // skip this corner, because it is transmissive
        }

        // compute image source position mirrored on the adjacent triangles
        for (unsigned int tid : corner.tidList) {
            Mesh::Triangle tri = mesh->getTriangle(tid);
            Mesh::Plane plane;
            if (tri.surfaceId < size(surface)) {
                plane = surface[tri.surfaceId].plane;
            } else {
                plane = mesh->getPlane(tri);
            }
            Vector3 S_new;
            if (side_S[tid]) {
                S_new = S.mirror(plane.N0, plane.d);
            } else {
                S_new = S; // image source is occluded
            }
            if ((Simg.size() == 0)
                || (Simg.back() != S_new)) { // identical points in subsequent order are not allowed
                Simg.push_back(S_new);
            }
        }
        if ((Simg.size() > 0)
            && (Simg.back()
                == Simg[0])) { // identical points in subsequent order are not allowed (wrap-around)
            Simg.erase(Simg.end() - 1);
        }

        // frustum must have a non-zero volume -> at least 3 points are needed
        if (Simg.size() < 3) {
            continue;
        }

        // create frustum
        Y = mesh->getVertex(corner.vtx, Mesh::CSpace::Global);
        frustum = getCornerFrustum(Simg, Y);

        // compute reflection point Y, if L is located within the frustum
        if (!insideFrustum(L, frustum)) {
            continue;
        }

        triangleId = corner.tidList[0];
        surfaceId = surfaceIdList[0];
        if (getRoundSpec(S, L, Y, reflectorMap, triangleId, surfaceId, spec)) {
            path.insert(spec);
            for (auto sid : surfaceIdList) {
                skipSurfaces.insert(sid);
            }
        }
    }
}

bool SpecularPlanarReflector::getRoundSpec(const Vector3& S, const Vector3& L, const Vector3& Y,
    const std::map<ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap, int triangleId,
    int surfaceId, SpecularReflection& spec) const
{

    // check occlusion by other geometric objects
    std::vector<int> triangleSeqOcclusion;
    std::vector<Vector3> mirrorPoints = { Y };
    CVisible visible = isVisible(S, L, mirrorPoints, reflectorMap, NULL, &triangleSeqOcclusion);

    if (visible != CVisible::VISIBLE) {
        return false;
    }

    // compute reflection plane in Hesse normal form
    Vector3 N0 = ((S - Y).normalize() + (L - Y).normalize()).normalize();
    float d = N0.dot(Y);

    // generate transformation matrix
    Vector3 A = Vector3(1, 0, 0).mirror(N0, d);
    Vector3 B = Vector3(0, 1, 0).mirror(N0, d);
    Vector3 C = Vector3(0, 0, 1).mirror(N0, d);
    Vector3 O = Vector3(0, 0, 0).mirror(N0, d);
    HomogenMatrix tform = HomogenMatrix(A - O, B - O, C - O, O);

    Vector3 S_img = S.mirror(N0, d);
    float path_length = (L - S_img).abs();
    int renderItemID = -1;
    std::vector<int> seq = { surfaceId };
    std::vector<int> triangleSeq = { triangleId };
    bool is_occluded = (visible == CVisible::VISIBLE) ? false : true;
    long long int value =
        value_start + surfaceId + 1; // see ImageRoomDatabase::seq2val() for further details
    int order = 1;
    float gain = fadeInSourceCulling(S, L, order) * getFloorGain(N0);
    std::vector<float> eq = determineEq(triangleSeq, triangleSeqOcclusion, gain);
    spec = SpecularReflection(path_length, path_length, S_img, renderItemID, seq, triangleSeq,
        is_occluded, mirrorPoints, value, tform, tform, eq);

    return true;
}

template <typename Container>
bool SpecularPlanarReflector::insideFrustum(const Vector3& P, const Container& frustum) const
{
    for (auto plane : frustum) {
        if (mesh->pointPlaneDist(plane, P) < threshInsideFrustum) {
            return false;
        }
    }
    return true;
}

std::array<Mesh::Plane, 4> SpecularPlanarReflector::getEdgeFrustum(const Vector3& Simg_left,
    const Vector3& Simg_right, const Vector3& P_from, const Vector3& P_to) const
{
    return {
        mesh->getPlane(Simg_right, P_from, P_to),
        mesh->getPlane(Simg_left, P_to, P_from),
        mesh->getPlane(P_to, Simg_right, Simg_left),
        mesh->getPlane(P_from, Simg_left, Simg_right),
    };
}

std::vector<Mesh::Plane> SpecularPlanarReflector::getCornerFrustum(
    const std::vector<Vector3>& Simg, const Vector3& Y) const
{
    std::vector<Mesh::Plane> frustum;

    int num_vtx = Simg.size();
    for (int v1 = 0; v1 < num_vtx; v1++) {
        int v2 = (v1 + 1) % num_vtx;
        frustum.push_back(mesh->getPlane(Y, Simg[v2], Simg[v1]));
    }

    return frustum;
}

float SpecularPlanarReflector::pointLineDist(
    const Vector3& A, const Vector3& B, const Vector3& P, Vector3& P_project) const
{

    Vector3 AB0 = (B - A).normalize();
    P_project = AB0.dot(P - A) * AB0 + A;
    float dist = (P - P_project).abs();

    return dist;
}

float SpecularPlanarReflector::fadeInSourceCulling(
    const Vector3& S, const Vector3& L, int order) const
{
    float gain = 1.0f;
    int max_dist_source = max_dist_source_1;
    float range_fadein_source = range_fadein_source_1;
    if (order > 1) {
        max_dist_source = max_dist_source_2;
        range_fadein_source = range_fadein_source_2;
    }

    if (max_dist_source <= 0.f) {
        return gain; // no source distance culling -> no fade-in
    }

    gain = (max_dist_source - (S - L).abs()) / range_fadein_source;

    // limit gain range to [0..1]
    if (gain > 1.0f) {
        gain = 1.0f;
    }
    if (gain < 0.0f) {
        gain = 0.0f;
    }

    return gain;
}

std::vector<float> SpecularPlanarReflector::determineEq(const std::vector<int>& triangleSeq,
    const std::vector<int>& triangleSeqOcclusion, const float& gain) const
{
    std::vector<float> eq(ThirdOctaveNBands, gain);
    unsigned int num_freq = sizeof(ThirdOctaveCenterFreqs) / sizeof(ThirdOctaveCenterFreqs[0]);

    // reflection filters
    for (int triangleIdx : triangleSeq) {
        std::shared_ptr<Material> material = mesh->getTriangle(triangleIdx).material;

        if (material) {
            for (unsigned int k = 0; k < num_freq; k++) {
                eq[k] *= material->getReflectionCoeffForFreq(ThirdOctaveCenterFreqs[k]);
            }
        }
    }

    // transmission filters
    for (int triangleIdx : triangleSeqOcclusion) {
        std::shared_ptr<Material> material = mesh->getTriangle(triangleIdx).material;

        if (material) {
            for (unsigned int k = 0; k < num_freq; k++) {
                eq[k] *= material->getTransmissionCoeffForFreq(ThirdOctaveCenterFreqs[k]);
            }
        }
    }

    //Geometric filter equals the identity element--> filter = {1,1,1,1,...}

    return eq;
}

void SpecularPlanarReflector::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    if (modification == EAR::SceneObjectObserver::Property::Position) {
        Mesh* geo = dynamic_cast<Mesh*>(obj);
        updateMesh(max_order, geo);
    }
}

void SpecularPlanarReflector::updateFloorParameters()
{
    float floor_detect_angle = Config::getFloat("EarlyFloorAngleDeg") * M_PI / 180.0;

    EAR_CHECK_REL(floor_detect_angle > 0.f, "EarlyFloorAngleDeg must be greater than 0 degree.");
    EAR_CHECK_REL(
        floor_detect_angle <= 90.f, "EarlyFloorAngleDeg must be less than or equal 90 degree.");

    // set member variables
    floor_thresh_y = std::cos(floor_detect_angle);
    floor_scale = Config::getFloat("EarlyFloorGainDB") / (1. - floor_thresh_y);
}

float SpecularPlanarReflector::getFloorGain(const EAR::Vector3 N0) const
{
    const float& y = N0[1];
    if (y < floor_thresh_y) {
        return 1.0; // no floor reflection -> return 0dB
    }

    float gain_dB = (y - floor_thresh_y) * floor_scale;
    float gain = pow(10., (gain_dB / 20.));

    return gain;
}

void SpecularPlanarReflector::updateTransmissionFlags()
{
    const auto& triangles = mesh->getTriangles();
    size_t num_triangles = triangles.size();
    isTransmissive.resize(num_triangles);

    for (size_t triangleIdx = 0; triangleIdx < num_triangles; triangleIdx++) {
        bool is_trans = false;
        float pwr_trans = 0.f;
        float pwr_refl = 0.f;
        std::shared_ptr<Material> material = mesh->getTriangle(triangleIdx).material;
        if (material) {
            for (const float frequency : ThirdOctaveCenterFreqs) {
                float trans = material->getTransmissionCoeffForFreq(frequency);
                float refl = material->getReflectionCoeffForFreq(frequency);
                pwr_trans += trans * trans;
                pwr_refl += refl * refl;
            }
            is_trans = pwr_trans > pwr_refl;
        }
        isTransmissive[triangleIdx] = is_trans;
    }
}

} // namespace MPEGI
