/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPECULARPLANARREFLECTOR
#define MPEGI_SPECULARPLANARREFLECTOR

#include <MPEGI/Bitstream/EarlyPayload.h>

#include "SpecularReflector.h"
#include <MPEGI/Helpers/ConfigObserver.h>
#include <MPEGI/Helpers/EmbreeHelper.h>

#include <EAR/BaseTypes.h>
#include <EAR/SceneObject.h>
#include <EAR/SceneObjectObserver.h>

#include <map>

namespace MPEGI {

//! Concrete class for reflecting planar geometries.
/*!
 * This class implements reflecting planar geometries.
 */
class SpecularPlanarReflector : public SpecularReflector,
                                public EAR::SceneObjectObserver,
                                public MPEGI::ConfigObserver
{

    const float threshInsideFrustum = -0.000001f;
    const int max_order_voxel = 2;

public:
    //! Constructor
    /*!
     * \param _max_order        input: maximum reflection order
     * \param _mesh             input: geometry data
     */
    SpecularPlanarReflector(unsigned int _max_order, EAR::Mesh* _mesh,
        std::shared_ptr<const EarlyPayload> _earlyPayload = NULL);

    ~SpecularPlanarReflector();

    //! Inherited from ConfigObserver
    void configValueChanged(const std::string& key) override;

    //! Update geometry data
    /*!
     * \param _max_order        input: maximum reflection order
     * \param _mesh             input: geometry data
     */
    void updateMesh(unsigned int _max_order, EAR::Mesh* _mesh) override;

    //! Updates the canOccludeOthers flags
    //! Warning: This modifies via a const_cast<Mesh*> the mesh which was passed to the constructor or to updateMesh()!
    /*!
    * \param reflectorMap       input: reflector map
    */
    void updateOccludeOthers(
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap);

    //! Get a list of visible or occluded image sources
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param path              output: list of specular reflections
     */
    void getPaths(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<SpecularReflection>& path) const override;

    //! Get a list of visible or occluded image sources by iterating over all surfaces
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param path              output: list of specular reflections
     */
    void getPathsBruteForce(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<SpecularReflection>& path) const;

    //! Get a list of visible or occluded image sources using a pre-computed voxel data
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param earlyDataSet      input: pre-computed voxel data set
     * \param path              output: list of specular reflections
     */
    void getPathsAccelerated(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        EarlyDataSet earlyDataSet, std::set<SpecularReflection>& path) const;

    //! Intersection test between a line and this reflecting object
    /*!
     * \param P1                    input: starting point of the line
     * \param P2                    input: end point of the line
     * \param triangleSeqOcclusion  output: end point of the line
     * \return                      VISIBLE: no intersection, OCCLUDED: intersection
     */
    CVisible lineIntersect(const EAR::Vector3& P1, const EAR::Vector3& P2,
        std::vector<int>* triangleSeqOcclusion = nullptr) const override;

    //! Get internal mesh
    /*!
    * \return                   geometry data
    */
    const EAR::Mesh* getMesh() { return mesh; };

    //! Determine combined reflection and occlusion EQ
    /*!
     * \param triangleSeq           input: sequence of reflecting triangles
     * \param triangleSeqOcclusion  input: sequence of occluding triangles
     * \param gain                  input: additional linear gain
     * \return                      VISIBLE: no intersection, OCCLUDED: intersection
     */
    std::vector<float> determineEq(const std::vector<int>& triangleSeq,
        const std::vector<int>& triangleSeqOcclusion, const float& gain) const override;

    //! Inherited from SceneObjectObserver
    void objectChanged(
        EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

private:
    //! Get a list of edge reflections
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param skipSurfaces      input/output: list of surfaces which are already taken into account
     * \param path              output: list of specular reflections
     */
    void getPathsEdges(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<int>& skipSurfaces, std::set<SpecularReflection>& path) const;

    //! Get a list of edge reflections
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param skipSurfaces      input/output: list of surfaces which are already taken into account
     * \param path              output: list of specular reflections
     */
    void getPathsCorners(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<int>& skipSurfaces, std::set<SpecularReflection>& path) const;

    //! Get a list of edge reflections
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param Y                 input: mirror point
     * \param reflectorMap      input: list of reflectors
     * \param triangleId        input: triangle id
     * \param surfaceId         input: surface id
     * \param spec              output: list of specular reflections
     */
    bool getRoundSpec(const EAR::Vector3& S, const EAR::Vector3& L, const EAR::Vector3& Y,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        int triangleId, int surfaceId, SpecularReflection& spec) const;

    //! Check if a point is located inside a frustum
    /*!
     * \param P                 input: point location
     * \param frustum           input: list of planes that define the frustum
     * \return                  false: outside frustum; true: inside frustum
     */
    template <typename Container>
    bool insideFrustum(const EAR::Vector3& P, const Container& frustum) const;

    //! Create an edge frustum
    /*!
     * \param Simg_left         input: image source from left triangle
     * \param Simg_right        input: image source from right triangle
     * \param P_from            input: starting point of the edge
     * \param P_to              input: end point of the edge
     * \return                  edge frustum
     */
    std::array<EAR::Mesh::Plane, 4> getEdgeFrustum(const EAR::Vector3& Simg_left,
        const EAR::Vector3& Simg_right, const EAR::Vector3& P_from, const EAR::Vector3& P_to) const;

    //! Create a corner frustum
    /*!
     * \param Simg              input: list of image sources
     * \param Y                 input: corner vertex
     * \return                  edge frustum
     */
    std::vector<EAR::Mesh::Plane> getCornerFrustum(
        const std::vector<EAR::Vector3>& Simg, const EAR::Vector3& Y) const;

    //! compute the perpendicular from a point to a line
    /*!
     * \param A                 input: starting point of the line
     * \param B                 input: end point of the line
     * \param P                 input: point location
     * \param P_project         output: P projected onto the line
     * \return                  distance
     */
    float pointLineDist(const EAR::Vector3& A, const EAR::Vector3& B, const EAR::Vector3& P,
        EAR::Vector3& P_project) const;

    //! Get linear gain for source distance culling fade-in
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param order             input: reflection order
     * \return                  linear gain
     */
    float fadeInSourceCulling(const EAR::Vector3& S, const EAR::Vector3& L, int order) const;

    void checkEmbreeDeviceError();

    void iterateSurfaces(std::vector<EAR::Mesh::Surface>& surfaceSequence, const EAR::Vector3& S,
        const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<SpecularReflection>& path, std::set<int>& skipSurfaces) const;

    void traceReflectionPath(const std::vector<EAR::Mesh::Surface>& surfaceSequence,
        const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<SpecularReflection>& path, std::set<int>& skipSurfaces) const;

    //! Get sorting value for a given reflection sequence
    /*!
     * \param seq               input: reflection sequence
     * \return                  sorting value
     */
    long long int getSortingValue(const std::vector<int>& seq) const;

    //! Updates the internal parameters needed for floor attenuation
    void updateFloorParameters();

    //! Updates the internal lookup table for transmissive triangles
    void updateTransmissionFlags();

    //! Computes a floor specific gain for a surface with a given normal vector
    /*!
     * \param N0                   unit normal vector of the surface (ideal floor: [0, 1, 0])
     * \return                     linear gain
     */
    float getFloorGain(const EAR::Vector3 N0) const;

private:
    //! type of reflecting geometry
    EAR::Mesh* mesh;

    //! source and triangle culling distances
    float max_dist_triangle_1;
    float max_dist_triangle_2;
    float max_dist_source_1;
    float max_dist_source_2;
    float range_fadein_triangle_1;
    float range_fadein_triangle_2;
    float range_fadein_source_1;
    float range_fadein_source_2;

    std::shared_ptr<const EarlyPayload> earlyPayload{ NULL };

    // TODO: common device that can be reused by all stages?
    static RTCDevice embreeDevice;

    RTCScene embreeScene;
    RTCGeometry embreeMesh;

    // attenuation parameters for floor reflection
    float floor_thresh_y;
    float floor_scale;

    std::vector<bool> isTransmissive;
}; /* class SpecularPlanarReflector */

} /* namespace MPEGI */

#endif /* MPEGI_SPECULARPLANARREFLECTOR */
