/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SpecularReflectionPathFinder.h"
#include <EAR/Exception.h>

#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif // !_USE_MATH_DEFINES

#include <algorithm>
#include <cmath>
#include <iostream>
#include <iterator>
#include <memory>
#include <set>
#include <vector>

namespace MPEGI {

using namespace EAR;

template <typename Tin, typename Tout>
void _set_split(Tin first1, Tin last1, Tin first2, Tin last2, Tout d_first, Tout i_first)
{
    while (first1 != last1) {
        if (first2 == last2) {
            std::copy(first1, last1, d_first);
            return;
        }

        if (*first1 < *first2) {
            *d_first++ = *first1++;
        } else {
            if (!(*first2 < *first1)) {
                auto ref = *first1;
                ref.setRenderItemID(first2->getRenderItemID());
                *i_first = std::move(ref);
                i_first++;
                first1++;
            }

            ++first2;
        }
    }
}

void SpecularReflectionPathFinder::addMesh(
    unsigned int _max_order, Mesh* _geometry, std::shared_ptr<const EarlyPayload> _earlyPayload)
{

    // add mesh to reflector map
    reflectorMap.insert(std::make_pair(_geometry->getID(),
        std::make_unique<SpecularPlanarReflector>(_max_order, _geometry, _earlyPayload)));

    // update value ranges
    long long int start = 0;
    for (auto& it : reflectorMap) {
        it.second->setValueStart(start);
        start += it.second->getNumValues();
    }
}

int SpecularReflectionPathFinder::delMesh(const Mesh* _geometry)
{

    // remove the item from the reflector map
    auto it = reflectorMap.find(_geometry->getID());
    if (it == reflectorMap.end()) {
        return -1;
    }
    reflectorMap.erase(it);

    // update value ranges
    long long int start = 0;
    for (auto& it : reflectorMap) {
        it.second->setValueStart(start);
        start += it.second->getNumValues();
    }

    return 0;
}

int SpecularReflectionPathFinder::updateMesh(unsigned int _max_order, Mesh* _geometry)
{

    auto it = reflectorMap.find(_geometry->getID());
    if (it == reflectorMap.end()) {
        return -1;
    }
    it->second->updateMesh(_max_order, _geometry);

    return 0;
}

void SpecularReflectionPathFinder::addCylinder(
    const Cylinder* _cylinder, std::shared_ptr<const EarlyPayload> _earlyPayload)
{

    // add mesh to reflector map
    reflectorMap.insert(std::make_pair(
        _cylinder->getID(), std::make_unique<SpecularCylinderReflector>(_cylinder, _earlyPayload)));

    // update value ranges
    long long int start = 0;
    for (auto& it : reflectorMap) {
        it.second->setValueStart(start);
        start += it.second->getNumValues();
    }
}

int SpecularReflectionPathFinder::delCylinder(const Cylinder* _cylinder)
{

    // remove the item from the reflector map
    auto it = reflectorMap.find(_cylinder->getID());
    if (it == reflectorMap.end()) {
        return -1;
    }
    reflectorMap.erase(it);

    // update value ranges
    long long int start = 0;
    for (auto& it : reflectorMap) {
        it.second->setValueStart(start);
        start += it.second->getNumValues();
    }

    return 0;
}

int SpecularReflectionPathFinder::updateCylinder(const Cylinder* _cylinder)
{

    auto it = reflectorMap.find(_cylinder->getID());
    if (it == reflectorMap.end()) {
        return -1;
    }
    it->second->updateCylinder(_cylinder);

    return 0;
}

void SpecularReflectionPathFinder::updateOccludeOthers()
{
    for (auto& it : reflectorMap) {
        if (it.second->getType() == ReflectorType::PLANAR) {
            SpecularPlanarReflector* refl = static_cast<SpecularPlanarReflector*>(it.second.get());
            refl->updateOccludeOthers(reflectorMap);
        }
    }
}

std::set<SpecularReflection> SpecularReflectionPathFinder::getPaths(
    const Vector3& S, const Vector3& L)
{

    std::set<SpecularReflection> path;

    for (auto& it : reflectorMap) {
        it.second->getPaths(S, L, reflectorMap, path);
    }

#ifdef MPEGI_ENABLE_MATLAB_PLOTTER

    HomogenMatrix tform;
    plotter.open("EAR_scene", id);

    /* plot listener */
    tform = HomogenMatrix({ 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 }, L);
    plotter.plot_listener(tform);

    /* plot source */
    tform = HomogenMatrix({ 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 }, S);
    plotter.plot_source(tform);

    /* plot geometries */
    int num_mesh = 0;
    int num_cylinder = 0;
    for (auto& it : reflectorMap) {
        SpecularPlanarReflector* reflPlanar = (SpecularPlanarReflector*)it.second.get();
        SpecularCylinderReflector* reflCylinder = (SpecularCylinderReflector*)it.second.get();
        Mesh* mesh = NULL;
        Cylinder* cylinder = NULL;
        ReflectorType type = it.second->getType();
        switch (type) {
        case ReflectorType::PLANAR:
            num_mesh++;
            plotter.plot_mesh(reflPlanar->getMesh(), "Mesh " + std::to_string(num_mesh));
            break;
        case ReflectorType::CYLINDER:
            num_cylinder++;
            plotter.plot_cylinder(
                reflCylinder->getCylinder(), "Cylinder " + std::to_string(num_cylinder));
            break;
        }
    }

    /* plot propagation paths */
    for (auto& it : path) {
        plotter.plot_path(it);
    }

    plotter.close();

#endif /* MPEGI_ENABLE_MATLAB_PLOTTER */

    return path;
}

void SpecularReflectionPathFinder::setLimitOrder(unsigned int _limit_order)
{
    for (auto& [objectId, reflector] : reflectorMap) {
        reflector->setLimitOrder(_limit_order);
    }
}

void SpecularReflectionPathFinder::set_split(const std::set<SpecularReflection>& setNew,
    const std::set<SpecularReflection>& setOld, std::set<SpecularReflection>& outNewOnly,
    std::set<SpecularReflection>& outCommon, std::set<SpecularReflection>& outOldOnly)
{

    _set_split(setNew.begin(), setNew.end(), setOld.begin(), setOld.end(),
        std::inserter(outNewOnly, outNewOnly.begin()), std::inserter(outCommon, outCommon.begin()));

    std::set_difference(setOld.begin(), setOld.end(), setNew.begin(), setNew.end(),
        std::inserter(outOldOnly, outOldOnly.begin()));
}

int SpecularReflectionPathFinder::idCounter = 1;

} // namespace MPEGI
