/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPECULARREFLECTIONPATHFINDER
#define MPEGI_SPECULARREFLECTIONPATHFINDER

#include "SpecularCylinderReflector.h"
#include "SpecularPlanarReflector.h"
#include "VisibilityCheck.h"
#include <MPEGI/Helpers/SpecularReflection.h>

#ifdef MPEGI_ENABLE_MATLAB_PLOTTER
#include "ScenePlotterMatlab.h"
#endif /* MPEGI_ENABLE_MATLAB_PLOTTER */

#include <EAR/BaseTypes.h>
#include <EAR/Geometry.h>

#include <map>
#include <memory>
#include <set>
#include <vector>

namespace MPEGI {

//! Factory for a set of visible or occluded SpecularReflection objects.
/*!
 * This class uses the ImageRoomDatabase class and VisibilityCheck class to
 * determine a list of visible or occluded specular reflections.
 */
class SpecularReflectionPathFinder
{

public:
    //! Constructor
    SpecularReflectionPathFinder()
        : id(idCounter++)
    {}

    //! Constructor
    SpecularReflectionPathFinder(EAR::Mesh* _geometry, unsigned int _max_order)
        : id(idCounter++)
    {

        addMesh(_max_order, _geometry);
    }

    //! Adds a room boundary / object geometry and re-computes the database.
    /*!
     * \param _max_order        input: maximum reflection order
     * \param _geometry         input: geometry data
     * \param _earlyPayload      input: pre-computed ER data for accelerated rendering
     * \param _voxelGridPayload  input: voxel grid for accelerated rendering
     */
    void addMesh(unsigned int _max_order, EAR::Mesh* _geometry,
        std::shared_ptr<const EarlyPayload> _earlyPayload = NULL);

    //! Removes a room boundary / object geometry and re-computes the database.
    /*!
     * \param _geometry         input: geometry data
     * \return                  0: no error; -1: error
     */
    int delMesh(const EAR::Mesh* _geometry);

    //! Updates the room boundary / object geometry and re-computes the database.
    /*!
     * \param _max_order        input: maximum reflection order
     * \param _geometry         input: geometry data
     * \return                  0: no error; -1: error
     */
    int updateMesh(unsigned int _max_order, EAR::Mesh* _geometry);

    //! Adds a cylindrical reflector.
    /*!
     * \param _cylinder         input: geometry data
     * \param _earlyPayload      input: pre-computed ER data for accelerated rendering
     */
    void addCylinder(
        const EAR::Cylinder* _cylinder, std::shared_ptr<const EarlyPayload> _earlyPayload);

    //! Removes a cylindrical reflector.
    /*!
     * \param _cylinder         input: geometry data
     * \return                  0: no error; -1: error
     */
    int delCylinder(const EAR::Cylinder* _cylinder);

    //! Updates a cylindrical reflector.
    /*!
     * \param _cylinder         input: geometry data
     * \return                  0: no error; -1: error
     */
    int updateCylinder(const EAR::Cylinder* _cylinder);

    //! Updates the canOccludeOthers flags
    //! Warning: This modifies via a const_cast<Mesh*> the mesh which was passed to the constructor or to addMesh()!
    void updateOccludeOthers();

    //! Get a list of visible or occluded image sources
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \return                  set of visible or occluded image sources
     */
    std::set<SpecularReflection> getPaths(const EAR::Vector3& S, const EAR::Vector3& L);

    //! Set identifier
    void setId(int _id) { id = _id; };

    //! Set current limit for reflection order
    /*!
     * \param _limit_order      input: limit for reflection order
     */
    void setLimitOrder(unsigned int _limit_order);

public: // static methods
    //! Split two sets of SpecularReflections into an intersection subset and two remaining subsets.
    /*!
     * \param setNew            input: set A with most recent values
     * \param setOld            input: set A with older values including a valid renderItemID
     * \param outNewOnly        output: non-common subset of setNew
     * \param outCommon         output: intersection set with renderItemID from setOld and all other values from setNew
     * \param outOldOnly        output: non-common subset of setOld
     */
    static void set_split(const std::set<SpecularReflection>& setNew,
        const std::set<SpecularReflection>& setOld, std::set<SpecularReflection>& outNewOnly,
        std::set<SpecularReflection>& outCommon, std::set<SpecularReflection>& outOldOnly);

private:
    //! map SceneObject pointers to SpecularReflector pointers
    std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>> reflectorMap;

    //! identifier
    int id;

    //! automatically increased to provide unique ID for every instance of this class
    static int idCounter;

#ifdef MPEGI_ENABLE_MATLAB_PLOTTER
    ScenePlotterMatlab plotter;
#endif /* MPEGI_ENABLE_MATLAB_PLOTTER */
};

} // namespace MPEGI

#endif /* MPEGI_SPECULARREFLECTIONPATHFINDER */
