/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_SPECULARREFLECTOR
#define MPEGI_SPECULARREFLECTOR

#include "VisibilityCheck.h"
#include <EAR/BaseTypes.h>
#include <EAR/Exception.h>
#include <EAR/Geometry.h>
#include <MPEGI/Helpers/SpecularReflection.h>
#include <map>
#include <set>
#include <sstream>

// #include <cmath>

namespace MPEGI {

//! Specifies the type of reflecting geometry
enum class ReflectorType
{
    PLANAR = int(0),
    CYLINDER
};

//constants #TODO: move them into a configFile!!??
const static int ThirdOctaveNBands = 31;
const static float ThirdOctaveCenterFreqs[] = { // [Hz]
    20, 25, 31.5, 40, 50, 63, 80, 100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250,
    1600, 2000, 2500, 3150, 4000, 5000, 6300, 8000, 10000, 12500, 16000, 20000
};

//! Base class for reflecting geometries.
/*!
 * This class defines a base class for reflecting geometries.
 */
class SpecularReflector
{

public:
    //! Constructor
    /*!
     * \param _type             input: type of reflecting geometry
     * \param _max_order        input: maximum reflection order
     */
    SpecularReflector(ReflectorType _type, unsigned int _max_order)
        : type(_type)
        , max_order(_max_order)
        , limit_order(_max_order)
        , num_values(0)
        , value_start(0){};

    //! Destructor
    virtual ~SpecularReflector(){};

    //! check if the given reflectors occlude the source
    /*!
     * \param S                     input: source location
     * \param L                     input: listener location
     * \param mirrorPoints          input: list of reflection points on the wall surfaces
     * \param reflectorMap          input: list of reflectors
     * \param skip                  input: listener location
     * \param triangleSeqOcclusion  output: end point of the line
     * \return
     */
    CVisible isVisible(const EAR::Vector3& S, const EAR::Vector3& L,
        std::vector<EAR::Vector3> mirrorPoints,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        SpecularReflector* skip, std::vector<int>* triangleSeqOcclusion = nullptr) const;

    //! get type of this reflector
    /*!
     * \return                  reflector type
     */
    ReflectorType getType() { return type; }

    //! get total number of possible sorting values
    /*!
     * \return                  total number of possible sorting values
     */
    long long int getNumValues() { return num_values; }

    //! set starting range of the sorting values
    /*!
     * \return                  starting range of the sorting values
     */
    void setValueStart(long long int start) { value_start = start; }

    //! Update geometry data
    /*!
     * \param _max_order        input: maximum reflection order
     * \param _mesh             input: geometry data
     */
    virtual void updateMesh(unsigned int _max_order, EAR::Mesh* _mesh)
    {
        std::stringstream msg;
        msg << "Invalid method: SpecularReflector::updateMesh(" << _max_order << ", " << _mesh
            << ")";
        EAR_FAIL(msg.str());
    };

    //! Update geometry data
    /*!
     * \param _cylinder         input: geometry data
     */
    virtual void updateCylinder(const EAR::Cylinder* _cylinder)
    {
        std::stringstream msg;
        msg << "Invalid method: SpecularReflector::updateCylinder(" << _cylinder << ")";
        EAR_FAIL(msg.str());
    };

    //! Get a list of visible or occluded image sources
    /*!
     * \param S                 input: source location
     * \param L                 input: listener location
     * \param reflectorMap      input: list of reflectors
     * \param path              output: list of specular reflections
     */
    virtual void getPaths(const EAR::Vector3& S, const EAR::Vector3& L,
        const std::map<EAR::ObjectID, std::unique_ptr<SpecularReflector>>& reflectorMap,
        std::set<SpecularReflection>& path) const = 0;

    //! Intersection test between a line and this reflecting object
    /*!
     * \param P1                    input: starting point of the line
     * \param P2                    input: end point of the line
     * \param triangleSeqOcclusion  input: end point of the line
     * \return                      VISIBLE: no intersection, OCCLUDED: intersection
     */
    virtual CVisible lineIntersect(const EAR::Vector3& P1, const EAR::Vector3& P2,
        std::vector<int>* triangleSeqOcclusion) const = 0;

    //! Set current limit for reflection order
    /*!
     * \param _limit_order      input: limit for reflection order
     */
    void setLimitOrder(unsigned int _limit_order) { limit_order = _limit_order; };

private:
    virtual std::vector<float> determineEq(const std::vector<int>& seq,
        const std::vector<int>& seqOcclusion, const float& gain) const = 0;

protected:
    //! type of reflecting geometry
    ReflectorType type;

    //! maximum reflection order
    unsigned int max_order;

    //! limit for reflection order which can be changed at runtime
    unsigned int limit_order;

    //! total number of possible sorting values
    long long int num_values;

    //! starting range of the sorting values
    long long int value_start;
};

} // namespace MPEGI

#endif /* MPEGI_SPECULARREFLECTOR */
