/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#include "EffectActivator.h"

#include <EAR/PointSource.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

namespace MPEGI {

std::shared_ptr<RenderStage> makeEffectActivatorStage()
{
    return std::make_shared<EffectActivator>();
}

EffectActivator::EffectActivator()
    : itemStore(nullptr)
{}

EffectActivator::~EffectActivator()
{
    for (auto& attached : attachedSources) {
        attached->detach(this);
    }
}

std::unique_ptr<RenderStage::DspData> EffectActivator::update(RenderList& renderList)
{
    if (!itemStore) {
        return {};
    }

    auto dspData = std::make_unique<EffectActivator::DspData>();

    const auto& objsources = sceneState->getObjectSources();

    for (auto& item : renderList.items) {
        if (item->status == ItemStatus::Inactive) {
            continue;
        }
        auto temp = item->signal;
        bool deactivate = false;

        // TODO: Can we access the name of the signal in item->signal? Would make this easier...

        if (objSrcMap.find(item->objectName) != objSrcMap.end()) {
            auto streamID = objsources[objSrcMap.at(item->objectName)]->getAudioStreamID();
            if (audioFileDurations.find(streamID) != audioFileDurations.end()) {
                // Object is a sound effect, should it be active?
                if (activeEffects.find(item->objectName) != activeEffects.end()) {
                    if (activeEffects.at(item->objectName) < frameCounter) {
                        deactivate = true;
                    }
                    item->isSoundEffect = true;
                } else {
                    // Is a sound effect, but not in the active list, so shouldn't be active here.
                    deactivate = true;
                }
            }
        }

        if (deactivate) {
            // remove from active effect list
            if (activeEffects.find(item->objectName) != activeEffects.end()) {
                activeEffects.erase(item->objectName);
            }

            // deactivate render item
            itemStore->deactivate(item.get());

            // deactivate scene object
            if (objSrcMap.find(item->objectName) != objSrcMap.end()) {
                objsources[objSrcMap.at(item->objectName)]->setActive(false);
            }
        }
    }

    return dspData;
}

void EffectActivator::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    if (modification == EAR::SceneObjectObserver::Property::Play) {
        std::string streamId = "";
        auto psrc = dynamic_cast<EAR::PointSource*>(obj);
        if (psrc) {
            streamId = psrc->getAudioStreamID();
        }
        if (activeEffects.find(obj->getName()) == activeEffects.end()) {
            activeEffects.emplace(obj->getName(), audioFileDurations.at(streamId) + frameCounter);
        } else {
            activeEffects.at(obj->getName()) = audioFileDurations.at(streamId) + frameCounter;
        }
        obj->setActive(true);
    } else if (modification == EAR::SceneObjectObserver::Property::Stop) {
        obj->setActive(false);
    }
}

void EffectActivator::processBlock(DspContext ctx)
{
    frameCounter += 1;
    return;
}

void EffectActivator::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    this->itemStore = itemStore;
    this->sceneState = state;
    this->frameCounter = 0;

    auto blockSize = streamManager->getB();
    auto fs = streamManager->getFs();
    auto blockRate = fs / blockSize;

    // delay the deactivation (default 0.025 seconds)
    int deactivationDelay =
        (int)std::ceil(Config::getFloatDef("EventActivator:delay", 1.0) * blockRate);

    auto eifPayload = std::dynamic_pointer_cast<const EIFPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));

    for (auto& stream : eifPayload->getAudioStreams()) {
        if (stream.isSoundEffect) {
            audioFileDurations.emplace(
                stream.id, (int)std::ceil(stream.duration * blockRate) + deactivationDelay);
        }
    }
    int idx = 0;

    for (const auto& objsource : state->getObjectSources()) {
        attachedSources.push_back(objsource.get());
        objsource->attach(this);
        objSrcMap.emplace(objsource->getName(), idx);
        idx++;
    }
}

} // namespace MPEGI
