/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#ifndef MPEGI_EFFECT_ACTIVATOR_STAGE
#define MPEGI_EFFECT_ACTIVATOR_STAGE

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include <map>

namespace MPEGI {

//! Activates and deactivates sound effects.
/*! 
 * Inputs: 
 * - All render items which are sound effects and which correspond to a point source.
 *  
 * Outputs: 
 * - The sources for every sound effect, are disabled once their duration has expired. The result is that when a new renderlist is created (for the next update() round), the corresponding RIs will be disabled.
 * - All other render items are passed through unchanged
 *  
 * How it works: 
 * - Listens to EAR::SceneObjectObserver::Property::Play/Stop signals for the scene state, and enables / disables sources based on it. Disables sources when sound effect duration is expired.
 * 
 * Stages order:
 * - Required before: none
 * - Required after: none
 * - This stage has no dependencies on other stages. It is recommended to use it as the first stage, so other stages don't need to process "empty" sound effect items.
 * Config options: 
 * - None
 */
class EffectActivator : public RenderStage, public SceneObjectObserver
{
public:
    EffectActivator();

    //! Destructor
    ~EffectActivator();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

    //! Inherited from SceneObjectObserver
    void objectChanged(
        EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

private:
    friend class Renderer;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    RenderItemStore* itemStore;
    const SceneState* sceneState;
    const StreamManager* streamManager;
    struct ItemState : RenderItem::ExtraState
    {
        bool deactivate = false;
    };
    std::vector<EAR::PointSource*> attachedSources; // used to detach in destructor

    std::map<std::string, int> audioFileDurations;
    std::map<std::string, int> activeEffects;
    std::map<std::string, int> objSrcMap;

    struct DspData : RenderStage::DspData
    {};

    std::size_t frameCounter;
};

} // namespace MPEGI

#endif /* MPEGI_EFFECT_ACTIVATOR_STAGE */
