/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Eq.h"

#include <EAR/FIRFilter.h>
#include <EAR/ListenerVoiceSource.h>
#include <EAR/PchipMinphaseFilterDesigner.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

namespace MPEGI {

std::shared_ptr<RenderStage> makeEqStage()
{
    return std::make_shared<EqStage>();
}

EqStage::EqStage()
{
    Config::setDef("Binaural_PathFilterLength_LoQ", EAR::Variant(256));
    Config::setDef("Binaural_PathFilterLength_HiQ", EAR::Variant(1024));
    Config::setDef("DirectSoundGain", EAR::Variant(1.f));
    Config::setDef("ListenerVoiceDirectSoundGain", EAR::Variant(0.1f));
    Config::setDef("DiffractionGain", EAR::Variant(1.f));
    Config::setDef("HOAGain", EAR::Variant(1.f));
    Config::setDef("DirectSoundEnabled", EAR::Variant(true));
    Config::setDef("EarlyReflectionsEnabled", EAR::Variant(true));
    Config::setDef("DiffractionEnabled", EAR::Variant(true));
    Config::setDef("Reverb:Enabled", EAR::Variant(true));
    Config::setDef("HOAEnabled", EAR::Variant(true));
    Config::setDef("PortalSoundEnabled", EAR::Variant(true));
    /* Accesssibility Control */
    Config::setDef("ErAttenuationDb", EAR::Variant(0.f));
    Config::setDef("ReverbAttenuationDb", EAR::Variant(0.f));
}

EqStage::~EqStage() {}

std::unique_ptr<RenderStage::DspData> EqStage::update(RenderList& renderList)
{
    auto dspData = std::make_unique<EqStage::DspData>();
    auto& dspList = dspData->dspList;

    float erGain = Config::getBool("EarlyReflectionsEnabled")
        * std::pow(10.f, -0.05 * Config::getFloat("ErAttenuationDb"));
    float reverbGain = Config::getBool("Reverb:Enabled")
        * std::pow(10.f, -0.05 * Config::getFloat("ReverbAttenuationDb"));
    int playPortal = Config::getBool("PortalSoundEnabled");
    float directGain = Config::getFloat("DirectSoundGain") * Config::getBool("DirectSoundEnabled");
    float directVoice =
        Config::getFloat("ListenerVoiceDirectSoundGain") * Config::getBool("DirectSoundEnabled");
    float diffrGain = Config::getFloat("DiffractionGain") * Config::getBool("DiffractionEnabled");
    float hoaGain = Config::getFloat("HOAGain") * Config::getBool("HOAEnabled");

    for (const auto& item : renderList.items) {
        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        if (item->type == ItemType::Airflow) {
            continue;
        }

        auto& itemState = item->getStateOrCreate<EqStage::ItemState>(*item);

        if (item->status == ItemStatus::Inactive) {
            // reset filters
            for (auto& filter : itemState.pathFilters) {
                if (filter) {
                    filter->flush();
                }
            }
            continue;
        }

        // create new path filter pointer for each channel if signal changed
        if (item->changed & (ItemProperty::Signal | ItemProperty::Status)
            && item->signal->getFrame().channels() > itemState.pathFilters.size()) {
            itemState.pathFilters.resize(item->signal->getFrame().channels());
        }

        // reset ramp start gain before fade in
        if (item->changed & (ItemProperty::Status) && item->status == ItemStatus::Active) {
            itemState.prevGain.store(item->gain);
        }

        // update filters if EQ changed
        if (item->changed & (ItemProperty::Eq | ItemProperty::Signal | ItemProperty::Status)) {
            // TODO: compare different filter structures (IIR/filterbank vs. FIR design)

            float eqSum = std::accumulate(item->eq.begin(), item->eq.end(), 0.f);
            if (!itemState.hasEq
                && std::abs(item->eq.size() - eqSum) <= 1) { // TODO: configurable threshold
                // only set to false if the item never had an EQ
                // TODO: can we turn this off smoothly when the EQ is temporarily 1?
                itemState.hasEq = false;
            } else {
                itemState.hasEq = true;

                int filterLen;
                if (item->type == ItemType::Reflection) {
                    filterLen = Config::getInt("Binaural_PathFilterLength_LoQ");
                    shortFilterDesigner->generate(item->eq.data(), *tempPathFilter);

                } else {
                    filterLen = Config::getInt("Binaural_PathFilterLength_HiQ");
                    filterDesigner->generate(item->eq.data(), *tempPathFilter);
                }

                for (auto& filter : itemState.pathFilters) {
                    if (!filter) {
                        filter = std::make_unique<EAR::FIRFilter>(B, filterLen);
                    }

                    filter->setFilter(tempPathFilter->data());
                }
            }
        }

        // create dsp item
        dspList.emplace_back();

        dspList.back().hasEq = itemState.hasEq;
        dspList.back().gain = item->gain;
        dspList.back().prevGain = &itemState.prevGain;

        if (item->type == ItemType::Primary) {
            dspList.back().gain *= directGain;
        } else if (item->type == ItemType::ListenerVoice) {
            dspList.back().gain *= directVoice;
        } else if (item->type == ItemType::Diffraction) {
            dspList.back().gain *= diffrGain;
        } else if (item->type == ItemType::Reflection) {
            dspList.back().gain *= erGain;
        } else if (item->type == ItemType::Pannable) {
            dspList.back().gain *= reverbGain;
        } else if (item->type == ItemType::HOA
            || (item->type == ItemType::Representative
                && item->objectName.find("output_speaker") != std::string::npos)) {
            dspList.back().gain *= hoaGain;
        } else if (item->type == ItemType::DirectionalReverb) {
            dspList.back().gain *= playPortal;
        }

        for (auto i = 0; i < item->signal->getFrame().channels(); i++) {
            dspList.back().signals.emplace_back(
                itemState.pathFilters[i].get(), &(item->signal->getFrame()[i]));
        }
    }

    return dspData;
}

void EqStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<EqStage::DspData>().dspList;
    for (auto& item : currentDSP) {
        float prevGain = item.prevGain->load();
        item.prevGain->store(item.gain);

        for (auto& sig : item.signals) {
            if (item.hasEq) {
                sig.first->process(sig.second->data(), sig.second->data());
            }
            sig.second->rampMul(prevGain, item.gain);
        }
    }
}

void EqStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }
    float fs = streamManager->getFs();
    this->B = streamManager->getB();

    filterDesigner = std::make_unique<EAR::PchipMinphaseFilterDesigner>(
        Config::getInt("Binaural_PathFilterLength_HiQ"), fs);
    shortFilterDesigner = std::make_unique<EAR::PchipMinphaseFilterDesigner>(
        Config::getInt("Binaural_PathFilterLength_LoQ"), fs);
    tempPathFilter =
        std::make_unique<EAR::RealBuffer>(Config::getInt("Binaural_PathFilterLength_HiQ"));
    for (const auto& source : state->getObjectSources()) {
        auto listenerVoice = dynamic_cast<EAR::ListenerVoiceSource*>(source.get());
        if (listenerVoice) {
            const float linGainListenerVoice =
                std::pow(10.f, 0.05f * listenerVoice->getDirectGainDb());
            Config::set("ListenerVoiceDirectSoundGain", linGainListenerVoice);
        }
    }
}

} // namespace MPEGI
