/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_EQ_STAGE
#define MPEGI_EQ_STAGE

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <utility>
#include <vector>

// fwd declares
namespace EAR {
class FilterDesigner;
class FIRFilter;
class RealBuffer;
} // namespace EAR

namespace MPEGI {

//! Apply EQ and gain of each RenderItem to the associated audio signal
/*! 
 * Inputs: 
 * - All render items
 *  
 * Outputs: 
 * - Same RIs, but with changed signal based on EQ and gain of the RI
 *  
 * How it works: 
 * - A minimum phase FIR filter is designed using the cepstrum windowing method
 * - Filter and gain are applied to the signal buffer
 * - Also applies the DirectSoundGain config value to all ItemType::Primary
 * 
 * Stages order:
 * - Required before: none
 * - Required after: none
 *  
 * Config options: 
 * - (not real-time adjustable) "Binaural_PathFilterLength_LoQ": low quality EQ filter length
 *   (used for all types except ItemType::Primary, default: 256)
 * - (not real-time adjustable) "Binaural_PathFilterLength_HiQ": high quality EQ filter length
 *   (used for all types except ItemType::Primary, default: 1024)
 * - (real-time adjustable) "DirectSoundGain": gain applied to ItemType::Primary (default: 1)
 * - (real-time adjustable) "ListenerVoiceDirectSoundGain": gain applied to ItemType::ListenerVoice (default: 0.1)
 * 
 */
class EqStage : public RenderStage
{
public:
    //! Default Constructor
    EqStage();

    //! Destructor
    ~EqStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

private:
    friend class Renderer;
    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    struct ItemState : RenderItem::ExtraState
    {
        explicit ItemState(RenderItem& item)
            : prevGain{ item.gain }
            , pathFilters(item.signal->getFrame().channels())
        {}

        std::vector<std::unique_ptr<EAR::FIRFilter>> pathFilters; // one for each channel
        std::atomic<float> prevGain; // directly accessed by audio thread
        bool hasEq = false;
    };

    struct DSPItem
    {
        std::vector<std::pair<EAR::FIRFilter*, EAR::RealBuffer*>> signals;
        bool hasEq;
        float gain;
        std::atomic<float>* prevGain;
    };

    struct DspData : RenderStage::DspData
    {
        std::vector<DSPItem> dspList;
    };

    std::unique_ptr<EAR::FilterDesigner> filterDesigner;
    std::unique_ptr<EAR::FilterDesigner> shortFilterDesigner;
    std::unique_ptr<EAR::RealBuffer> tempPathFilter;
    std::size_t B;
};

} // namespace MPEGI

#endif /* MPEGI_EQ_STAGE */
