/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Fade.h"

#include <MPEGI/Renderer/RenderItemStore.h>

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/Exception.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>

#include <algorithm>
#include <complex>

namespace MPEGI {

std::shared_ptr<RenderStage> makeFadeStage()
{
    return std::make_shared<FadeStage>();
}

FadeStage::FadeStage()
    : B(0)
{
    Config::setDef("Fade_Length", EAR::Variant(0.05f)); // in seconds
}

FadeStage::~FadeStage() {}

std::unique_ptr<RenderStage::DspData> FadeStage::update(RenderList& renderList)
{

    if (B == 0) { // only update once initialized
        return {};
    }

    auto dspData = std::make_unique<FadeStage::DspData>();
    auto& dspList = dspData->dspList;

    for (auto& item : renderList.items) {

        if (item->type == ItemType::Downmix || item->type == ItemType::Airflow) {
            // Downmix not processed here
            continue;
        }

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        auto* fade = item->getState<Fade>();
        if (item->status == ItemStatus::Inactive) {
            if (fade) {
                dspList.push_back({ fade, true, 0 });
            }
            continue;
        }

        // check if a new fade needs to be done
        int fadeDirection = 0;
        if ((item->status == ItemStatus::FadeOutInactive
                || item->status == ItemStatus::FadeOutOrphaned)
            && (item->changed & ItemProperty::Status)) {
            fadeDirection = -1; // fade out
        } else if (item->status == ItemStatus::New
            || (item->status == ItemStatus::Active && (item->changed & ItemProperty::Status))) {
            fadeDirection = 1; // fade in
        }

        //sess representative items are only used for book-keeping, they don't need fade ins/outs
        if (item->type == ItemType::Representative || item->owner->getSkipFade(item->idx)
            || (item->type == ItemType::DirectionalReverb && item->parentRI != -1)) {
            switch (fadeDirection) {
            case 1:
                item->owner->fadeInCompleted(item->idx);
                break;
            case -1:
                item->owner->fadeOutCompleted(item->idx);
                break;
            default:
                break;
            }
            item->owner->skipNextFade(item->idx, 0);
            continue;
        }

        // create data structure for new items
        if (!fade || (item->changed & ItemProperty::Signal)) {
            fade = &item->getStateOrCreate<Fade>(
                &(item->signal->getFrame()), this, fadeIn.get(), fadeOut.get());
        }

        // read the current state and report finished fades
        bool needsReset = false;
        const Fade::State fadeState = fade->getLastState();
        switch (fadeState) {
        case Fade::State::FadedIn:
            item->owner->fadeInCompleted(item->idx);
            needsReset = true;
            break;
        case Fade::State::FadedOut:
            item->owner->fadeOutCompleted(item->idx);
            if (item->status == ItemStatus::FadeOutInactive) {
                itemStore->deactivate(item.get());
            } else if (item->status == ItemStatus::FadeOutOrphaned) {
                itemStore->remove(item.get());
            }
            needsReset = true;
            break;
        // currently not used!
        case Fade::State::FadedOutDirChanged:
        case Fade::State::FadedInDirChanged:
            needsReset = true;
            break;
        default:
            break;
        }

        if (item->isSoundEffect && fadeDirection == 1) {
            // don't do fade in of sound effects, but do fade out.
            fadeDirection = 0;
        }

        dspList.push_back({ fade, needsReset, fadeDirection });
    }

    return dspData;
}

void FadeStage::processBlock(DspContext ctx)
{
    activeFades.clear();

    auto& currentDSP = ctx.getData<FadeStage::DspData>().dspList;

    for (auto& item : currentDSP) {
        item.fade->update(item.reset, item.direction);
        item.reset = false;
        activeFades.push_back(item.fade);
    }
    std::sort(activeFades.begin(), activeFades.end());
    for (auto it = activeFades.begin(), end = std::unique(activeFades.begin(), activeFades.end());
         it != end; it++) {
        (*it)->apply();
        (*it)->updateState();
    }
}

void FadeStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    this->itemStore = itemStore;
    this->B = streamManager->getB();

    activeFades.reserve(4096);

    // generate s fade curves (smoothstep)
    fadeLength = streamManager->getFs() * Config::getFloat("Fade_Length"); // implicit floor to int
    fadeIn = std::make_unique<EAR::RealBuffer>(fadeLength);
    fadeOut = std::make_unique<EAR::RealBuffer>(fadeLength);

    for (auto i = 0; i < fadeLength; i++) {
        float x = static_cast<float>(i) / fadeLength;

        if (x < 0.f) {
            x = 0.f;
        } else if (x > 1.f) {
            x = 1.f;
        }

        fadeIn->at(i) = x * x * (3 - 2 * x);
        x = 1 - x;
        fadeOut->at(i) = x * x * (3 - 2 * x);
    }
}

void FadeStage::Fade::apply()
{
    if (activeDirection == 0)
        return;

    std::size_t L = parent->B;
    if (samplesLeft < parent->B) {
        L = samplesLeft;
    }

    for (unsigned int i = 0; i < signal->channels(); i++) {
        signal->ch(i).mul(*activeEnvelope, parent->fadeLength - samplesLeft, 0, L);
    }

    if (L < parent->B && activeDirection == -1) { // set remaining samples to zero
        for (unsigned int i = 0; i < signal->channels(); i++) {
            signal->ch(i).zero(L, parent->B - L);
        }
    }

    samplesLeft -= parent->B;
}

void FadeStage::Fade::updateState()
{
    Fade::State currState = State::None;
    if (activeDirection != 0) {
        if (samplesLeft <= 0) {
            // fade in/out complete
            samplesLeft = 0;
            currState = activeDirection < 0 ? State::FadedOut : State::FadedIn;
        } else {
            currState = activeDirection < 0 ? State::FadingOut : State::FadingIn;
        }
    }
    state.store(currState);
}

void FadeStage::Fade::update(bool needsReset, int direction)
{
    if (needsReset) {
        //EAR_CHECK_REL(samplesLeft == 0, "Resetting a running fade!");
        activeDirection = 0;
        activeEnvelope = nullptr;
        samplesLeft = 0;
    }
    if (direction != 0 && direction != activeDirection) {
        // changing direction
        EAR_CHECK_REL(
            direction >= -1 && direction <= 1, "Invalid Direction provided to Fade::Update!");
        samplesLeft = parent->fadeLength - samplesLeft;
        activeDirection = direction;
        activeEnvelope = direction == 1 ? fadeInEnvelope : fadeOutEnvelope;
    }
}

FadeStage::Fade::State FadeStage::Fade::getLastState() const
{
    return state.load();
}

} // namespace MPEGI
