/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_FADE_STAGE
#define MPEGI_FADE_STAGE

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <vector>

// fwd declares
namespace EAR {
class RealBuffer;
class RealFrame;
}; // namespace EAR

namespace MPEGI {

//! Apply fade in and fade out to deactivated and reactivated RIs
/*! 
 * Inputs: 
 * - All RIs except ItemType::Downmix
 *  
 * Outputs: 
 * - Same RIs but potentially with fade applied to the signal
 *  
 * How it works: 
 * - A "smoothstep" fade curve is applied to each channel of the signal of fading render items
 * - ItemStatus::FadeOutInactive and ItemStatus::FadeOutOrphaned only turn into ItemStatus::Inactive
 *   and ItemStatus::Orphaned, respectively, once the fade out is complete
 * 
 * Stages order:
 * - Required before: DistanceStage
 * - Required after: none
 *  
 * Config options: 
 * - (not real-time adjustable) "Fade_Length": fade length in seconds (default: 0.05s)
 * 
 */
class FadeStage : public RenderStage
{
public:
    //! Constructor
    FadeStage();

    //! Destructor
    ~FadeStage();

    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

private:
    friend class Renderer;
    friend class Fade;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    RenderItemStore* itemStore;

    int consideredDirections;
    std::size_t B;
    int fadeLength;
    std::unique_ptr<EAR::RealBuffer> fadeIn, fadeOut;

    class Fade : public RenderItem::ExtraState
    {
    public:
        enum class State : int
        {
            None = 0,
            FadingIn = 1,
            FadingOut = 2,
            FadedIn = 3,
            FadedOut = 4,
            FadedInDirChanged = 5,
            FadedOutDirChanged = 6,
        };

        Fade(EAR::RealFrame* _signal, FadeStage* _parent, EAR::RealBuffer* _fadeInEnv,
            EAR::RealBuffer* _fadeOutEnv)
            : signal(_signal)
            , state(State::None)
            , samplesLeft(0)
            , activeDirection(0)
            , activeEnvelope(nullptr)
            , parent(_parent)
            , fadeInEnvelope(_fadeInEnv)
            , fadeOutEnvelope(_fadeOutEnv)
        {}

        void update(bool needsReset, int direction);
        State getLastState() const;
        void apply();
        void updateState();

    private:
        EAR::RealFrame* signal;
        std::atomic<State> state;
        int samplesLeft;
        int activeDirection;
        EAR::RealBuffer* activeEnvelope;

        FadeStage* const parent;
        EAR::RealBuffer* const fadeOutEnvelope;
        EAR::RealBuffer* const fadeInEnvelope;
    };

    struct DSPItem
    {
        Fade* fade;
        bool reset;
        int direction;
    };
    struct DspData : RenderStage::DspData
    {
        std::vector<DSPItem> dspList;
    };
    std::vector<Fade*> activeFades;
};

} // namespace MPEGI

#endif /* MPEGI_FADE_STAGE */
