/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#include "GranularSynth.h"
#include <algorithm>

namespace MPEGI {

// Local functions
bool areGrainPositionsComplementary(
    Vector3 targetDescrCoord, Vector3 coordsA, Vector3 coordsB, int dimensions, float companionThr);
float GetDistanceToLine(Vector3 refPoint, Vector3 point1, Vector3 point2);

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// GranularSynth class definitions
//
//////////////////////////////////////////////////////////////////////////////////////////////////////
GranularSynth::GranularSynth(MPEGI::ItemType itType, std::size_t blocksize, int samplerate)
    : B(blocksize)
    , Fs(samplerate)
    , itemType(itType)
    , smoothedTargetCoord({ 0.0, 0.0, 0.0 })
{
    granularSynthDspState =
        new GranularSynthDspState(itType, B, granularDatabase, grainDescrCoords);
}

GranularSynth::~GranularSynth()
{
    delete (granularSynthDspState);
}

void GranularSynth::InitState()
{
    granularSynthDspState->InitState();
}

bool GranularSynth::UpdateState(Vector3 descrCoor, int grainGroupIndex, float grainGroupProb)
{
    if (!granularDatabase)
        return true;

    // Set new target descriptor space coordinates
    float grainCoor0 = fmin(1.0f, fmax(0.0f, descrCoor[0]));
    float grainCoor1 = fmin(1.0f, fmax(0.0f, descrCoor[1]));
    float grainCoor2 = fmin(1.0f, fmax(0.0f, descrCoor[2]));
    Vector3 newCoords(grainCoor0, grainCoor1, grainCoor2);

    // Apply smoothing to avoid artefacts due to non-smooth changes in target descriptor coordinates
    smoothedTargetCoord[0] = smoothedTargetCoord[0]
            * granularSynthDspState->granularDatabase->smoothingTargetCoordinates[0]
        + (1.0f - granularSynthDspState->granularDatabase->smoothingTargetCoordinates[0])
            * newCoords[0];
    smoothedTargetCoord[1] = smoothedTargetCoord[1]
            * granularSynthDspState->granularDatabase->smoothingTargetCoordinates[1]
        + (1.0f - granularSynthDspState->granularDatabase->smoothingTargetCoordinates[1])
            * newCoords[1];
    smoothedTargetCoord[2] = smoothedTargetCoord[2]
            * granularSynthDspState->granularDatabase->smoothingTargetCoordinates[2]
        + (1.0f - granularSynthDspState->granularDatabase->smoothingTargetCoordinates[2])
            * newCoords[2];

    // Zero out dimensions not supported by the granular database and rescale the coordinates
    granularSynthDspState->targetDescrCoord[0] =
        smoothedTargetCoord[0] * granularSynthDspState->coordinateScaling[0];
    granularSynthDspState->targetDescrCoord[1] =
        smoothedTargetCoord[1] * granularSynthDspState->coordinateScaling[1];
    granularSynthDspState->targetDescrCoord[2] =
        smoothedTargetCoord[2] * granularSynthDspState->coordinateScaling[2];
    if (granularDatabase->dimensions < 3) {
        granularSynthDspState->targetDescrCoord[2] = 0.0;
    }
    if (granularDatabase->dimensions < 2) {
        granularSynthDspState->targetDescrCoord[1] = 0.0;
    }

    // Update grain group probabilities
    if (grainGroupIndex >= 0) {
        granularDatabase->probabilityGroupVal[grainGroupIndex] = grainGroupProb;
    }
    granularSynthDspState->firstUpdateReceived = true;

    // Update trend of target descriptor space coordinates
    float timeSinceLastTrendUpd = (float)granularSynthDspState->samplesSinceTrendUpdate / (float)Fs;
    Vector3 newtargetTrend(0.0, 0.0, 0.0);
    if (timeSinceLastTrendUpd > 0.0) {
        newtargetTrend = (smoothedTargetCoord - granularSynthDspState->lastTargetDescrCoord)
            * (1.0 / timeSinceLastTrendUpd);
        granularSynthDspState->smoothedTargetTrend =
            0.9f * granularSynthDspState->smoothedTargetTrend + 0.1f * newtargetTrend;
    }
    granularSynthDspState->lastTargetDescrCoord = smoothedTargetCoord;
    granularSynthDspState->samplesSinceTrendUpdate = 0;

    return true;
}

void GranularSynth::UpdateDspState()
{
    granularSynthDspState->UpdateState();
}

void GranularSynth::SetGranularDatabase(std::shared_ptr<GranularDatabase> granDatabase)
{
    granularDatabase = granDatabase;
    granularSynthDspState->granularDatabase = granDatabase;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// GranularSynthDspState class definitions
//
//////////////////////////////////////////////////////////////////////////////////////////////////////
void GranularSynthDspState::InitState()
{

    if (granularDatabase == nullptr)
        return;

    // Make sure that the overlap amount never exceeds 50%
    granularDatabase->grainOverlapAmount = fmin(0.5, granularDatabase->grainOverlapAmount);

    // Note wich weights that should be applied
    if (((granularDatabase->probabilityWeightPosition < 0.99)
            || (granularDatabase->probabilityWeightPosition > 1.01))
        && (granularDatabase->probabilityExponentPosition > 0.01)) {
        applyPositionWeight = true;
    }
    if (((granularDatabase->probabilityWeightHistory < 0.99)
            || (granularDatabase->probabilityWeightHistory > 1.01))
        && (granularDatabase->probabilityExponentHistory > 0.01)) {
        applyHistoryWeight = true;
    }
    if (((granularDatabase->probabilityWeightTrend < 0.99)
            || (granularDatabase->probabilityWeightTrend > 1.01))
        && (granularDatabase->probabilityExponentTrend > 0.01)
        && granularDatabase->descrTrendAvailable) {
        applyTrendWeight = true;
    }
    if (((granularDatabase->probabilityWeightOrigPos < 0.99)
            || (granularDatabase->probabilityWeightOrigPos > 1.01))
        && (granularDatabase->probabilityExponentOrigPos > 0.01)
        && granularDatabase->originalPositionAvailable) {
        applyOriginalPosWeight = true;
    }

    // Create KDTree structure for the granular database using scaled coordinates
    for (int loop = 0; loop < granularDatabase->grains.size(); loop++) {
        Vector3 scaledDescrCoord = granularDatabase->grains[loop].descriptorCoordinates;
        scaledDescrCoord[0] = coordinateScaling[0] * scaledDescrCoord[0];
        scaledDescrCoord[1] = coordinateScaling[1] * scaledDescrCoord[1];
        scaledDescrCoord[2] = coordinateScaling[2] * scaledDescrCoord[2];
        grainDescrCoords.emplace_back(scaledDescrCoord);
    }
    grainTree = std::make_unique<EAR::CartesianKDTree>(grainDescrCoords);

    // Create KDTree structure for the point specifying a mixin rule
    mixingRuleTree =
        std::make_unique<EAR::CartesianKDTree>(granularDatabase->grainMixingRuleCoordinates);

    // Go through all grains and set the mixing rule acoording to the closest point where
    // a mixing rule is specified
    for (int loop = 0; loop < granularDatabase->grains.size(); loop++) {
        std::size_t nearestPointIndex =
            mixingRuleTree->getNearestIndex(granularDatabase->grains[loop].descriptorCoordinates);
        granularDatabase->grains[loop].mixingRule =
            granularDatabase->grainMixRule[nearestPointIndex];
    }

    // Initialize the current grain
    currentGrainInd = grainTree->getNearestIndex(targetDescrCoord);
    interpGrainCurrent->InitInterpolatedGrain(granularDatabase, targetDescrCoord, 0, 0,
        granularDatabase->useInterpolationResampling,
        granularDatabase->grainInterpolationDistThreshold);
    int currentGrainOverlapLength =
        (int)floor((float)granularDatabase->grains[currentGrainInd].grainLen
            * granularDatabase->grainOverlapAmount)
        - 1;
    currentGrainCoreLen =
        granularDatabase->grains[currentGrainInd].samples[0].size() - 2 * currentGrainOverlapLength;
    currentGrainCoreStartSample = currentGrainOverlapLength;
    currentGrainCoreEndSample = currentGrainCoreStartSample + currentGrainCoreLen;
    currentSampleInd = currentGrainCoreStartSample;

    // Initialize the next grain
    nextGrainInd = grainTree->getNearestIndex(targetDescrCoord);
    interpGrainNext->InitInterpolatedGrain(granularDatabase, targetDescrCoord, 0, 0,
        granularDatabase->useInterpolationResampling,
        granularDatabase->grainInterpolationDistThreshold);
    int nextGrainOverlapLength = (int)floor((float)granularDatabase->grains[nextGrainInd].grainLen
                                     * granularDatabase->grainOverlapAmount)
        - 1;
    nextGrainCoreLen =
        granularDatabase->grains[nextGrainInd].samples[0].size() - 2 * nextGrainOverlapLength;
    nextGrainCoreStartSample = nextGrainOverlapLength;
    nextGrainCoreEndSample = nextGrainCoreStartSample + nextGrainCoreLen;
    nextSampleInd = nextGrainCoreStartSample;

    // Init the grain history
    for (int loop = 0; loop < granularDatabase->grainHistoryLength; loop++) {
        pastGrains.push_back(0);
    }

    overlapActive = false;

    if (granularDatabase->InterpolationMethod == 1) {
        hiQresampling = true;
    }
}

void GranularSynthDspState::UpdateState() {}

int GranularSynthDspState::SelectGrain(Vector3 targetCoordinate, Vector3* companionCoords)
{
    int selectedGrainInd = 0;

    // Select a set of candidate grains within the search radius from the target position
    std::vector<std::pair<std::size_t, float>> candidateGrains =
        grainTree->getAllIndicesAndDistancesWithinRadius(
            targetCoordinate, granularDatabase->searchRadius);

    int groupSize = candidateGrains.size();
    if (groupSize) {

        // Calculate a probability of each grain based on the position in the descriptor space
        // Start by finding the maxium distance within the group
        float maxDist = 0.0;
        for (const auto& pair : candidateGrains) {
            if (pair.second > maxDist) {
                maxDist = pair.second;
            }
        }

        // Calculate a probability of each grain in the group based on the distance
        // compared with the maximum distance
        std::vector<float> grainProbabilities;
        for (const auto& pair : candidateGrains) {
            if (pair.second != 0.0) {
                grainProbabilities.push_back(
                    pow(fmin(granularDatabase->probabilityWeightPosition, maxDist / pair.second),
                        granularDatabase->probabilityExponentPosition));
            } else {
                grainProbabilities.push_back(pow(granularDatabase->probabilityWeightPosition,
                    granularDatabase->probabilityExponentPosition));
            }
        }

        // Check if grains have been used in the past and if so, reduce the probability
        if (applyHistoryWeight) {
            CalculateWeigtBasedOnHistory(candidateGrains, grainProbabilities);
        }

        // Calculate a probability of each grain based on the coordinate trend
        if (applyTrendWeight) {
            CalculateWeigtBasedOnTrend(candidateGrains, grainProbabilities);
        }

        // Apply probability based on the grains positions in the original recordingy
        if (applyOriginalPosWeight) {
            CalculateWeigtBasedOnOriginalPos(candidateGrains, grainProbabilities);
        }

        // Apply individual grain and grain group probability
        if (granularDatabase->probabilityGroupsActive) {
            CalculateWeigtBasedOnGrainGroups(candidateGrains, grainProbabilities);
        }

        // If we are looking for a companion grain for interpolation, apply probability
        // based on a check if the position is complementary to that of the first grain
        if (companionCoords != NULL) {
            CalculateWeigtBasedOnComplementaryPosition(
                candidateGrains, grainProbabilities, (*companionCoords));
        }

        if (granularDatabase->randomSelection) {
            // Calculate the sum of probabilities
            float probSum = 0.0;
            for (const auto& prob : grainProbabilities) {
                probSum += prob;
            }

            int rndNumber = 0;
            if (probSum > 1.0) {
                rndNumber = rand() % ((int)probSum);
            }

            // Find the grain index corresponding to the random number
            probSum = 0.0;
            int grainIndex = 0;
            for (const auto& prob : grainProbabilities) {
                probSum += prob;
                if (probSum >= rndNumber) {
                    break;
                }
                grainIndex++;
            }
            if (grainIndex > grainProbabilities.size() - 1) {
                grainIndex = grainProbabilities.size()
                    - 1; // To avoid problem if some probability approaches infinity
            }
            selectedGrainInd = candidateGrains[grainIndex].first;
        } else {
            // Just choose the grain with the highest probability
            float highestProb = 0.0;
            int grainIndex = 0;
            for (int ind = 0; ind < grainProbabilities.size(); ind++) {
                if (grainProbabilities[ind] > highestProb) {
                    highestProb = grainProbabilities[ind];
                    grainIndex = ind;
                }
            }
            selectedGrainInd = candidateGrains[grainIndex].first;
        }

    } else {
        // Did not find any qualifying grain, just use the closest one
        selectedGrainInd = grainTree->getNearestIndex(targetCoordinate);
    }

    return selectedGrainInd;
}

void GranularSynthDspState::ScheduleNewGrain(float startPosInPerecent)
{
    if (granularDatabase == nullptr)
        return;
    if (granularDatabase->grains.size() == 0)
        return;

    int nextGrainOverlapLength = 0;

    // Select the next grain(s) from the database
    if (granularDatabase->useInterpolation) {
        // Select two grains at positions in the granular database that are complimentary
        // relative to the target position.
        int grainAindex = SelectGrain(targetDescrCoord, NULL);
        UpdateGrainHistory(grainAindex);

        // Check if the first grain is close enough that we don't need to do interpolation
        float distance =
            (granularDatabase->grains[grainAindex].descriptorCoordinates - targetDescrCoord).abs();
        int grainBindex = grainAindex;
        if (distance > granularDatabase->grainInterpolationDistThreshold) {
            // Select a second grain to do interpolation
            grainBindex = SelectGrain(
                targetDescrCoord, &granularDatabase->grains[grainAindex].descriptorCoordinates);
            UpdateGrainHistory(grainBindex);
        }

        // Create an interpolated grain using the two selected grains
        interpGrainNext->InitInterpolatedGrain(granularDatabase, targetDescrCoord, grainAindex,
            grainBindex, granularDatabase->useInterpolationResampling,
            granularDatabase->grainInterpolationDistThreshold);
        //nextGrainLen =
        //    (int)interpGrainNext->grainLen / (1.0f + 2.0 * granularDatabase->grainOverlapAmount);

        nextGrainOverlapLength =
            (int)floor((float)interpGrainNext->grainLen * granularDatabase->grainOverlapAmount) - 1;
        nextGrainCoreLen = interpGrainNext->grainLen - 2 * nextGrainOverlapLength;

        // Store the distance from the target pos to a line between the two selected grains
        targetToGrainDistance = GetDistanceToLine(targetDescrCoord,
            granularDatabase->grains[grainAindex].descriptorCoordinates,
            granularDatabase->grains[grainBindex].descriptorCoordinates);

    } else {
        nextGrainInd = SelectGrain(targetDescrCoord, NULL);
        nextGrainOverlapLength = (int)floor((float)granularDatabase->grains[nextGrainInd].grainLen
                                     * granularDatabase->grainOverlapAmount)
            - 1;
        nextGrainCoreLen =
            granularDatabase->grains[nextGrainInd].grainLen - 2 * nextGrainOverlapLength;
        UpdateGrainHistory(nextGrainInd);

        // Store the distance from the target pos to the grain
        targetToGrainDistance =
            (targetDescrCoord - granularDatabase->grains[nextGrainInd].descriptorCoordinates).abs();
    }

    // Initiate overlap add from the current to the next grain
    int currentGrainOverlapLength =
        (int)floor((float)granularDatabase->grains[currentGrainInd].grainLen
            * granularDatabase->grainOverlapAmount)
        - 1;
    overlapLength = fmin(currentGrainOverlapLength, nextGrainOverlapLength);
    overlapPosition = 0;
    overlapActive = true;

    nextGrainCoreStartSample = nextGrainOverlapLength;
    nextGrainCoreEndSample = nextGrainCoreStartSample + nextGrainCoreLen;
    nextSampleInd = nextGrainCoreStartSample - overlapLength
        + (int)floor(startPosInPerecent * (float)nextGrainCoreLen / 100.0);

    // Calculate the mixing rule for the next overlap
    grainMixingRule = fmax(granularDatabase->grains[currentGrainInd].mixingRule,
        granularDatabase->grains[nextGrainInd].mixingRule);
}

void GranularSynthDspState::CalculateWeigtBasedOnHistory(
    std::vector<std::pair<std::size_t, float>> grainsGroup, std::vector<float>& grainProbabilities)
{
    if (pastGrains.size() > 0) {
        for (int groupIndex = 0; groupIndex < grainsGroup.size() - 1; groupIndex++) {
            for (int pastInd = 0; pastInd < pastGrains.size(); pastInd++) {
                if (pastGrains[pastInd] == grainsGroup[groupIndex].first) {
                    // The grain was used recently, reduce its probability
                    float probability = 0.0;
                    if ((pastInd > 0) && (pastGrains.size() > 1)) {
                        float oneOverRatio = 1.0 / granularDatabase->probabilityWeightHistory;
                        probability = oneOverRatio
                            + ((float)pastInd - 1.0) / ((float)pastGrains.size() - 1.0)
                                * (1.0 - oneOverRatio);
                    }
                    grainProbabilities[groupIndex] = grainProbabilities[groupIndex]
                        * pow(probability, granularDatabase->probabilityExponentHistory);
                }
            }
        }
    }
}

void GranularSynthDspState::CalculateWeigtBasedOnTrend(
    std::vector<std::pair<std::size_t, float>> grainsGroup, std::vector<float>& grainProbabilities)
{
    if ((granularDatabase->descTrendMethod == 1) || (granularDatabase->descTrendMethod == 3)) {
        // Calculate the trend diff norms for each candidate grain
        std::vector<float> trendDiffNorms;
        for (int ind = 0; ind < grainsGroup.size(); ind++) {
            // Calculate difference between grain trend and target trend
            Vector3 trendDiff =
                granularDatabase->grains[grainsGroup[ind].first].coordTrend - smoothedTargetTrend;

            float trendDiffNorm = trendDiff.abs();
            trendDiffNorms.push_back(trendDiffNorm);
        }

        // Find the maximum trendDiff
        float maxTrendDiffNorm = 0.0;
        for (const auto& diffNorm : trendDiffNorms) {
            if (diffNorm > maxTrendDiffNorm) {
                maxTrendDiffNorm = diffNorm;
            }
        }

        // Set the probabilities of the candidate grains
        for (int ind = 0; ind < grainsGroup.size(); ind++) {
            float probability = 0.0;
            if (trendDiffNorms[ind] > 0) {
                probability = fmin(granularDatabase->probabilityWeightTrend,
                    maxTrendDiffNorm / trendDiffNorms[ind]);
            } else {
                probability = granularDatabase->probabilityWeightTrend;
            }
            grainProbabilities[ind] = grainProbabilities[ind]
                * pow(probability, granularDatabase->probabilityExponentTrend);
        }
    }
    if ((granularDatabase->descTrendMethod == 2) || (granularDatabase->descTrendMethod == 3)) {
        // Binary decision
        for (int groupIndex = 0; groupIndex < grainsGroup.size(); groupIndex++) {
            float trendProduct = smoothedTargetTrend.dot(
                granularDatabase->grains[grainsGroup[groupIndex].first].coordTrend);

            if (trendProduct < 0) {
                // This grain has the opposite coordinate trend, set a low probability
                grainProbabilities[groupIndex] = 0.00001 * grainProbabilities[groupIndex];
            }
        }
    }
}

void GranularSynthDspState::CalculateWeigtBasedOnOriginalPos(
    std::vector<std::pair<std::size_t, float>> grainsGroup, std::vector<float>& grainProbabilities)
{
    // The maximum distance is 1.0, which corresponds to that the grains
    // were either in different recordings or in different ends in the same one.

    // Set the probabilities of the candidate grains
    for (int ind = 0; ind < grainsGroup.size(); ind++) {
        float probability = 0.0;

        float pos1 = granularDatabase->grains[grainsGroup[ind].first].orgPosition;
        float pos2 = granularDatabase->grains[currentGrainInd].orgPosition;

        // Check if the grains come from the same recording or not
        float distance = 1.0;
        if (floor(pos1) == floor(pos2)) {
            // Grains where from the same recording segments, measure the distance within
            // that recording
            distance = fabs(pos1 - pos2);
        }

        if (distance > 0.0) {
            probability = fmin(granularDatabase->probabilityWeightOrigPos, 1.0 / distance);
        } else {
            probability = granularDatabase->probabilityWeightOrigPos;
        }
        grainProbabilities[ind] = grainProbabilities[ind]
            * pow(probability, granularDatabase->probabilityExponentOrigPos);
    }
}

void GranularSynthDspState::CalculateWeigtBasedOnGrainGroups(
    std::vector<std::pair<std::size_t, float>> grainsGroup, std::vector<float>& grainProbabilities)
{
    for (int ind = 0; ind < grainsGroup.size(); ind++) {
        // Get the probability of the group that the grain belongs to
        int groupInd = granularDatabase->grains[grainsGroup[ind].first].grainProbabilityGroup;

        if (groupInd < granularDatabase->probabilityGroupVal.size()) {
            float grpProb = granularDatabase->probabilityGroupVal[groupInd];
            grainProbabilities[ind] = grainProbabilities[ind] * grpProb;
        }

        // Apply individual grain probability
        grainProbabilities[ind] = grainProbabilities[ind]
            * granularDatabase->grains[grainsGroup[ind].first].perGrainProbabilityValue;
    }
}

void GranularSynthDspState::CalculateWeigtBasedOnComplementaryPosition(
    std::vector<std::pair<std::size_t, float>> grainsGroup, std::vector<float>& grainProbabilities,
    Vector3 companionCoords)
{
    // Calculate a probability of each grain based on how complementary its position is
    // with the companion coordinate relative to the target position.
    for (int groupIndex = 0; groupIndex < grainsGroup.size(); groupIndex++) {
        bool grainIsComplementary = true;
        Vector3 grainCoords =
            granularDatabase->grains[grainsGroup[groupIndex].first].descriptorCoordinates;

        grainIsComplementary =
            areGrainPositionsComplementary(targetDescrCoord, grainCoords, companionCoords,
                granularDatabase->dimensions, granularDatabase->grainInterpolationDistThreshold);

        if (!grainIsComplementary) {
            // This grain is not complementary to the companion grain
            grainProbabilities[groupIndex] = 0.0000001 * grainProbabilities[groupIndex];
        }
    }
}

void GranularSynthDspState::CompleteOverlap()
{
    if (granularDatabase->useInterpolation) {
        *interpGrainCurrent = *interpGrainNext;
    } else {
        currentGrainInd = nextGrainInd;
    }
    currentGrainCoreLen = nextGrainCoreLen;
    currentGrainCoreStartSample = nextGrainCoreStartSample;
    currentGrainCoreEndSample = nextGrainCoreEndSample;
    currentSampleInd = nextSampleInd;
    overlapActive = false;
}

void GranularSynthDspState::UpdateGrainHistory(int grainIndex)
{
    for (int ind = pastGrains.size() - 1; ind > 0; ind--) {
        pastGrains[ind] = pastGrains[ind - 1];
    }
    pastGrains[0] = grainIndex;
}

void GenerateGranularSynthSignals(
    GranularSynthDspState* DspState, EAR::RealFrame* outputSignal, std::size_t B)
{
    float weight1, weight2, overlapWindowValue;

    if (DspState->granularDatabase == nullptr)
        return;
    if (DspState->granularDatabase->grains.size() == 0)
        return;
    if (!DspState->firstUpdateReceived)
        return;

    // clear the RenderItem output buffers
    outputSignal->zero();

    double sample;
    if (DspState->granularDatabase->useInterpolation) {

        // Check if a fast grain-switch should be triggered
        if (DspState->granularDatabase->fastGrainSwitchActive && !DspState->overlapActive) {
            float newDistance = 0.0;
            if (DspState->interpGrainCurrent->grainA == DspState->interpGrainCurrent->grainB) {
                newDistance = (DspState->targetDescrCoord
                    - DspState->interpGrainCurrent->grainA->descriptorCoordinates)
                                  .abs();
            } else {
                newDistance = GetDistanceToLine(DspState->targetDescrCoord,
                    DspState->interpGrainCurrent->grainA->descriptorCoordinates,
                    DspState->interpGrainCurrent->grainB->descriptorCoordinates);
            }
            if (newDistance >= DspState->targetToGrainDistance
                    + DspState->granularDatabase->fastGrainSwitchThreshold) {
                // Distance has increased more than the threshold, trigger a fast switch!
                float grainPosInPercent = 0.0;
                if (true) {
                    // Synchronize phase of current and next grain
                    grainPosInPercent = 100.0
                        * (DspState->currentSampleInd - DspState->currentGrainCoreStartSample)
                        / DspState->currentGrainCoreLen;
                }
                // Check that we can do an overlap-add before the deafult overlap-add
                if (DspState->currentSampleInd + DspState->overlapLength
                    < DspState->currentGrainCoreEndSample - 1) {
                    DspState->ScheduleNewGrain(grainPosInPercent);
                }
            }
        }

        for (int i = 0; i < B; i++) {

            if (DspState->overlapLength > 0) {
                overlapWindowValue =
                    fmin(1.0, (float)DspState->overlapPosition / (float)DspState->overlapLength);
            } else {
                overlapWindowValue = 0.0;
            }

            weight1 = DspState->grainMixingRule * sqrt(overlapWindowValue)
                + (1.0 - DspState->grainMixingRule) * overlapWindowValue;
            weight2 = DspState->grainMixingRule * sqrt(1.0 - overlapWindowValue)
                + (1.0 - DspState->grainMixingRule) * (1.0 - overlapWindowValue);

            for (int chan = 0; chan < outputSignal->channels(); chan++) {
                if (DspState->overlapActive) {
                    sample = weight1
                            * DspState->interpGrainNext->GetInterpolatedSample(
                                DspState->targetDescrCoord, chan, DspState->nextSampleInd,
                                DspState->grainMixingRule, DspState->hiQresampling)
                        + weight2
                            * DspState->interpGrainCurrent->GetInterpolatedSample(
                                DspState->targetDescrCoord, chan, DspState->currentSampleInd,
                                DspState->grainMixingRule, DspState->hiQresampling);
                } else {
                    sample = DspState->interpGrainCurrent->GetInterpolatedSample(
                        DspState->targetDescrCoord, chan, DspState->currentSampleInd,
                        DspState->grainMixingRule, DspState->hiQresampling);
                }
                outputSignal->ch(chan).writeDbl(&sample, i, 1);
            }

            DspState->overlapPosition++;
            DspState->currentSampleInd++;
            DspState->nextSampleInd++;

            if (DspState->overlapActive && (DspState->overlapPosition >= DspState->overlapLength)) {
                // The overlap with the next grain has completed
                DspState->CompleteOverlap();
            }
            if (!DspState->overlapActive
                && (DspState->currentSampleInd >= DspState->currentGrainCoreEndSample)) {
                // We have reached the overlap region of the current grain, select grain to transition to
                // and initiate the overlap
                DspState->ScheduleNewGrain(0.0);
            }
        }
    } else { // No interpolation

        // Check if a fast grain-switch should be triggered
        if (DspState->granularDatabase->fastGrainSwitchActive && !DspState->overlapActive) {
            float newDistance = (DspState->targetDescrCoord
                - DspState->granularDatabase->grains[DspState->currentGrainInd]
                      .descriptorCoordinates)
                                    .abs();

            if (newDistance >= DspState->targetToGrainDistance
                    + DspState->granularDatabase->fastGrainSwitchThreshold) {
                // Distance has increased more than the threshold, trigger a fast switch!
                float grainPosInPercent = 0.0;
                if (true) {
                    // Synchronize phase of current and next grain
                    grainPosInPercent = 100.0
                        * (DspState->currentSampleInd - DspState->currentGrainCoreStartSample)
                        / DspState->currentGrainCoreLen;
                }

                // Check that we can do an overlap-add before the deafult overlap-add
                if (DspState->currentSampleInd + DspState->overlapLength
                    < DspState->currentGrainCoreEndSample - 1) {
                    DspState->ScheduleNewGrain(grainPosInPercent);
                }
            }
        }

        for (int i = 0; i < B; i++) {

            if (DspState->overlapLength > 0) {
                overlapWindowValue =
                    fmin(1.0, (float)DspState->overlapPosition / (float)DspState->overlapLength);
            } else {
                overlapWindowValue = 0.0;
            }

            weight1 = DspState->grainMixingRule * sqrt(overlapWindowValue)
                + (1.0 - DspState->grainMixingRule) * overlapWindowValue;
            weight2 = DspState->grainMixingRule * sqrt(1.0 - overlapWindowValue)
                + (1.0 - DspState->grainMixingRule) * (1.0 - overlapWindowValue);

            for (int chan = 0; chan < outputSignal->channels(); chan++) {
                if (DspState->overlapActive) {
                    sample = weight1
                            * DspState->granularDatabase->grains[DspState->nextGrainInd]
                                  .samples[chan][DspState->nextSampleInd]
                        + weight2
                            * DspState->granularDatabase->grains[DspState->currentGrainInd]
                                  .samples[chan][DspState->currentSampleInd];
                } else {
                    sample = DspState->granularDatabase->grains[DspState->currentGrainInd]
                                 .samples[chan][DspState->currentSampleInd];
                }
                outputSignal->ch(chan).writeDbl(&sample, i, 1);
            }

            DspState->overlapPosition++;
            DspState->currentSampleInd++;
            DspState->nextSampleInd++;

            if (DspState->overlapActive && (DspState->overlapPosition >= DspState->overlapLength)) {
                // The overlap with the next grain has completed
                DspState->CompleteOverlap();
            }
            if (!DspState->overlapActive
                && (DspState->currentSampleInd >= DspState->currentGrainCoreEndSample)) {
                // We have reached the overlap region of the current grain, select grain to transition to
                // and initiate the overlap
                DspState->ScheduleNewGrain(0.0);
            }
        }
    }

    DspState->samplesSinceTrendUpdate += B;

    return;
}

void InterpolatedGrain::InitInterpolatedGrain(std::shared_ptr<GranularDatabase> granDatabase,
    Vector3 targetDescrCoords, int grainAIndex, int grainBIndex, bool useResamp,
    float grainInterpDistThr)
{
    grainA = &granDatabase->grains[grainAIndex];
    grainB = &granDatabase->grains[grainBIndex];
    useResampling = useResamp;

    if (grainAIndex != grainBIndex) {
        CalculateGrainWeights(targetDescrCoords, grainInterpDistThr);

        if (useResampling) {
            // Calculate an interpolated length based on the lengths of grain A & B and their
            // relative positions to the target position.
            grainLen =
                (int)round(weightA * (float)grainA->grainLen + weightB * (float)grainB->grainLen);
        } else {
            // Set the length to the shorter of the grain A & B
            grainLen = (int)fmin(grainA->grainLen, grainB->grainLen);
        }
    } else {
        // A and B are the same grain, no interpolation needed
        grainLen = grainA->grainLen;
        weightA = 1.0;
        weightA_sqrt = 1.0;
        weightB = 0.0;
        weightB_sqrt = 0.0;
    }
}

void InterpolatedGrain::CalculateGrainWeights(Vector3 targetDescrCoords, float grainInterpDistThr)
{
    float distanceA = (grainA->descriptorCoordinates - targetDescrCoords).abs();
    float distanceB = (grainB->descriptorCoordinates - targetDescrCoords).abs();
    weightA = 1.0f;
    weightB = 1.0f;

    if ((distanceA == 0.0) || (distanceA < grainInterpDistThr)) {
        weightA = 1.0;
        weightB = 0.0;
    } else if ((distanceB == 0.0) || (distanceB < grainInterpDistThr)) {
        weightA = 0.0;
        weightB = 1.0;
    } else {
        weightA = distanceB / (distanceA + distanceB);
        weightB = distanceA / (distanceA + distanceB);
    }

    weightA_sqrt = sqrt(weightA);
    weightB_sqrt = sqrt(weightB);
}

float HammingKernel(int a, float x)
{
    float pi = 3.1415926535897932384626433832795;

    if (x == 0.0) {
        return 1.0;
    } else if ((x >= -a) && (x < a)) {
        return (float)sinf(pi * x) / x / pi * (0.54 - 0.46 * cosf(pi + pi * x / (float)a));
    } else {
        return 0.0;
    }
}

float LinearKernel(int a, float x)
{
    if (x == 0.0) {
        return 1.0;
    } else if ((x >= -1) && (x < 1)) {
        return (1.0 - abs(x));
    } else {
        return 0.0;
    }
}

float InterpolatedGrain::GetResampledGrainSample(
    Grain* grain, int channel, int sampleIndex, bool hiQ)
{
    float sample;
    float sampleIndFloat = ((float)sampleIndex / (float)grainLen) * (float)grain->grainLen;

    if (sampleIndFloat >= ((float)grain->grainLen - 1.0)) {
        sample = grain->samples[channel][grain->grainLen - 1];
    } else {
        int sampleIndFloor = (int)floor(sampleIndFloat);

        sample = 0.0;
        int kernelSize = 3;
        int i = sampleIndFloor;
        for (int iter = -kernelSize + 1; iter <= kernelSize; iter++) {
            i = sampleIndFloor + iter;
            float kernelSample;
            if (hiQ) {
                // High quality resampling
                kernelSample = HammingKernel(kernelSize, (sampleIndFloat - (float)i));
            } else {
                // Low complexity resampling
                kernelSample = LinearKernel(kernelSize, (sampleIndFloat - (float)i));
            }
            if (i > grain->grainLen - 1.0) {
                sample += kernelSample * grain->samples[channel][grain->grainLen - 1.0];
            } else if (i < 0) {
                sample += kernelSample * grain->samples[channel][0];
            } else {
                sample += kernelSample * grain->samples[channel][i];
            }
        }
    }

    return sample;
}

float InterpolatedGrain::GetInterpolatedSample(Vector3 targetDescrCoords, int channel,
    int sampleIndex, float grainMixingRule, bool hiQresmpling)
{
    // Calculate an interpolated sample based on the distance from the target position
    float sampleA = 0.0;
    float sampleB = 0.0;
    float sample = 0.0;
    if (useResampling) {
        if (weightA > 0.0) {
            sampleA = GetResampledGrainSample(grainA, channel, sampleIndex, hiQresmpling);
        }
        if (weightB > 0.0) {
            sampleB = GetResampledGrainSample(grainB, channel, sampleIndex, hiQresmpling);
        }
    } else {
        if (weightA > 0.0) {
            sampleA = grainA->samples[channel][sampleIndex];
        }
        if (weightB > 0.0) {
            sampleB = grainB->samples[channel][sampleIndex];
        }
    }

    // Calculate weight taking the mixing rule into account
    float wA = grainMixingRule * weightA_sqrt + (1.0 - grainMixingRule) * weightA;
    float wB = grainMixingRule * weightB_sqrt + (1.0 - grainMixingRule) * weightB;
    sample = wA * sampleA + wB * sampleB;

    return sample;
}

// Helper funtion that checks if a grain is in a complementary position in relation to another
// grain and the target position.
bool areGrainPositionsComplementary(
    Vector3 targetDescrCoord, Vector3 coordsA, Vector3 coordsB, int dimensions, float companionThr)
{
    bool positionIsComplementary = true;

    if ((coordsA[0] <= targetDescrCoord[0] - companionThr) && (coordsB[0] <= targetDescrCoord[0])) {
        positionIsComplementary = false;
    }
    if ((coordsA[0] >= targetDescrCoord[0] + companionThr) && (coordsB[0] >= targetDescrCoord[0])) {
        positionIsComplementary = false;
    }
    if (dimensions > 1) {
        if ((coordsA[1] <= targetDescrCoord[1] - companionThr)
            && (coordsB[1] <= targetDescrCoord[1])) {
            positionIsComplementary = false;
        }
        if ((coordsA[1] >= targetDescrCoord[1] + companionThr)
            && (coordsB[1] >= targetDescrCoord[1])) {
            positionIsComplementary = false;
        }
    }
    if (dimensions > 2) {
        if ((coordsA[2] <= targetDescrCoord[2] - companionThr)
            && (coordsB[2] <= targetDescrCoord[2])) {
            positionIsComplementary = false;
        }
        if ((coordsA[2] >= targetDescrCoord[2] + companionThr)
            && (coordsB[2] >= targetDescrCoord[2])) {
            positionIsComplementary = false;
        }
    }
    return positionIsComplementary;
}

// Helper function that calculates the distance from a reference position to a line
// between two points.
float GetDistanceToLine(Vector3 refPoint, Vector3 point1, Vector3 point2)
{
    Vector3 diffVec1 = refPoint - point1;
    Vector3 diffVec2 = point2 - point1;
    float dotProd1 = diffVec1.dot(diffVec2);
    float dotProd2 = diffVec2.dot(diffVec2);
    float t2 = dotProd1 / dotProd2;
    if (t2 > 1.0f) {
        t2 = 1.0f;
    }
    if (t2 < 0.0f) {
        t2 = 0.0f;
    }

    Vector3 closestPoint2 = point1 + diffVec2 * t2;
    Vector3 distVec = closestPoint2 - refPoint;
    float dist2 = distVec.abs();

    return dist2;
}

} // namespace MPEGI
