/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#ifndef MPEGI_GRANULARSYNTH
#define MPEGI_GRANULARSYNTH

#include <EAR/BaseTypes.h>
#include <EAR/CartesianKDTree.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/GranularPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Scene.h>

namespace EAR {
class RealBuffer;
class CartesianKDTree;
struct Position;
} // namespace EAR

namespace MPEGI {
class InterpolatedGrain;

class InterpolatedGrain
{
public:
    InterpolatedGrain()
    {
        grainA = NULL;
        weightA = 0.0;
        weightA_sqrt = 0.0;
        grainB = NULL;
        weightB = 0.0;
        weightB_sqrt = 0.0;
        grainLen = 0;
        useResampling = true;
    }

    void InitInterpolatedGrain(std::shared_ptr<GranularDatabase> granDatabase,
        Vector3 targetDescrCoords, int grainAIndex, int grainBIndex, bool useResampling,
        float grainInterpDistThr);
    void CalculateGrainWeights(Vector3 targetDescrCoords, float grainInterpDistThr);
    float GetInterpolatedSample(Vector3 targetDescrCoords, int channel, int sampleIndex,
        float grainMixingRule, bool hiQresmapling);
    float GetResampledGrainSample(Grain* grain, int channel, int sampleIndex, bool hiQresmapling);

    Grain* grainA;
    float weightA;
    float weightA_sqrt;
    Grain* grainB;
    float weightB;
    float weightB_sqrt;
    int grainLen;
    bool useResampling;
};

/*
     * GranularDspState
     *
     * This class holds information that is needed when rendering the output signals.
     * The variables that may change in the Update function of the rendering stage are copied
     * from the GranularSynthesisStage in order to avoid accessing these directly from the DSP thread.
     *
     */
class GranularSynthDspState
{
public:
    GranularSynthDspState(MPEGI::ItemType itType, std::size_t B,
        std::shared_ptr<GranularDatabase> granDatabase, std::vector<Vector3>& grainDescrCoordsRef)
        : B(B)
        , granularDatabase(granDatabase)
        , grainDescrCoords(grainDescrCoordsRef)
        , currentGrainInd(0)
        , currentSampleInd(0)
        , currentGrainCoreLen(0)
        , currentGrainCoreStartSample(0)
        , currentGrainCoreEndSample(0)
        , overlapActive(false)
        , overlapLength(1)
        , overlapPosition(0)
        , nextGrainInd(0)
        , nextSampleInd(0)
        , nextGrainCoreLen(0)
        , nextGrainCoreStartSample(0)
        , nextGrainCoreEndSample(0)
        , targetDescrCoord(0.0, 0.0, 0.0)
        , lastTargetDescrCoord(0.0, 0.0, 0.0)
        , smoothedTargetTrend(0.0, 0.0, 0.0)
        , targetToGrainDistance(0.0)
        , grainMixingRule(1.0)
        , coordinateScaling(1.0, 1.0, 1.0)
        , applyPositionWeight(false)
        , applyHistoryWeight(false)
        , applyTrendWeight(false)
        , applyOriginalPosWeight(false)
        , firstUpdateReceived(false)
        , samplesSinceTrendUpdate(0)
        , hiQresampling(0)
    {
        ScheduleNewGrain(0.0);
        interpGrainCurrent = new InterpolatedGrain();
        interpGrainNext = new InterpolatedGrain();
    }

    ~GranularSynthDspState() {}
    void InitState();
    void UpdateState();
    int SelectGrain(Vector3 targetCoord, Vector3* companionCoordinate);
    void ScheduleNewGrain(float startPosInPerecent);
    void CompleteOverlap();
    void UpdateGrainHistory(int grainIndex);
    void CalculateWeigtBasedOnHistory(std::vector<std::pair<std::size_t, float>> grainsGroup,
        std::vector<float>& grainProbabilities);
    void CalculateWeigtBasedOnTrend(std::vector<std::pair<std::size_t, float>> grainsGroup,
        std::vector<float>& grainProbabilities);
    void CalculateWeigtBasedOnOriginalPos(std::vector<std::pair<std::size_t, float>> grainsGroup,
        std::vector<float>& grainProbabilities);
    void CalculateWeigtBasedOnGrainGroups(std::vector<std::pair<std::size_t, float>> grainsGroup,
        std::vector<float>& grainProbabilities);
    void CalculateWeigtBasedOnComplementaryPosition(
        std::vector<std::pair<std::size_t, float>> grainsGroup,
        std::vector<float>& grainProbabilities, Vector3 companionCoords);

    Vector3 targetDescrCoord;
    Vector3 lastTargetDescrCoord;
    Vector3 smoothedTargetTrend;
    float targetToGrainDistance;

    // Current grain
    int currentGrainInd;
    int currentSampleInd;
    int currentGrainCoreLen; // The part excluding the overlap regions
    int currentGrainCoreStartSample;
    int currentGrainCoreEndSample;

    // Next grain
    int nextGrainInd;
    int nextSampleInd;
    int nextGrainCoreLen; // The part excluding the overlap regions
    int nextGrainCoreStartSample;
    int nextGrainCoreEndSample;

    // Overlap and add state
    bool overlapActive;
    int overlapLength;
    int overlapPosition;

    std::shared_ptr<GranularDatabase> granularDatabase;
    std::vector<int> pastGrains;
    std::vector<Vector3>& grainDescrCoords;
    std::unique_ptr<CartesianKDTree> grainTree;
    std::unique_ptr<CartesianKDTree> mixingRuleTree;
    std::size_t B;
    int Fs;
    float grainMixingRule;
    Vector3 coordinateScaling;

    // Grain interpolation
    InterpolatedGrain* interpGrainCurrent;
    InterpolatedGrain* interpGrainNext;

    // Probability weights
    bool applyPositionWeight;
    bool applyHistoryWeight;
    bool applyTrendWeight;
    bool applyOriginalPosWeight;

    bool firstUpdateReceived;
    int samplesSinceTrendUpdate;
    bool hiQresampling;

private:
};

/*
    * GranularSynth
    *
    * This class handles the generation of the granular synthesis signal used by one source.
    *
    */
class GranularSynth
{
public:
    //! Constructor
    GranularSynth(MPEGI::ItemType itType, std::size_t blocksize, int SampleRate);

    //! Destructor
    ~GranularSynth();

    //! Initiation
    void InitState();

    //! Function for updating the state of the GranularSynth
    bool UpdateState(Vector3 descrCoor, int grainGroupIndex, float grainGroupProb);

    //! Function for copying into to the GranularSynthDspState, which will be used in the processBlock function
    void UpdateDspState();

    //! Setting a pointer to the granular database to use for this source
    void SetGranularDatabase(std::shared_ptr<GranularDatabase> granularDatabase);

    MPEGI::ItemType itemType;
    GranularSynthDspState* granularSynthDspState;
    //std::vector<std::vector<double>> grains;
    std::vector<Vector3> grainDescrCoords;
    std::shared_ptr<GranularDatabase> granularDatabase;

private:
    std::size_t B; // Blocksize in samples
    int Fs; // SampleRate
    Vector3 smoothedTargetCoord;
};

void GenerateGranularSynthSignals(
    GranularSynthDspState* outSignalState, EAR::RealFrame* outputSignal, std::size_t B);

} // namespace MPEGI

#endif /* MPEGI_GRANULARSYNTH */
