/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "GranularSynthesisStage.h"
#include <EAR/SofaReader.h>

#include <MPEGI/Helpers/Types.h>

#include <EAR/BaseTypes.h>

namespace MPEGI {

std::shared_ptr<RenderStage> makeGranularSynthesisStage()
{
    return std::make_shared<GranularSynthesisStage>();
}

GranularSynthesisStage::GranularSynthesisStage()
    : B(0)
    , itemStore(NULL)
{}

GranularSynthesisStage::~GranularSynthesisStage() {}

std::unique_ptr<RenderStage::DspData> GranularSynthesisStage::update(RenderList& renderList)
{
    if (B == 0) { // only update once initialized
        return {};
    }
    EAR::Position pos;
    pos.cSpace = EAR::CSpace::WCS;

    auto dspData = std::make_unique<GranularSynthesisStage::DspData>();
    auto& dspItemList = dspData->dspItemList;

    for (auto& item : renderList.items) {
        // Check if this rendering item uses granular synthesis
        if (item->granularDatabaseId == "") {
            // skip irrelevant render items.
            continue;
        }

        auto& stageItem = item->getStateOrCreate<GranularSynthesisStageItem>(*item, this);

        // Update the descriptor space coordinates and probability groups
        stageItem.granularSynth->UpdateState(
            item->granularPos, item->granularGroupProb[0], item->granularGroupProb[1]);

        // Prepare DSP data
        dspItemList.emplace_back(stageItem.granularSynth->granularSynthDspState);

        dspItemList.back().granularSynthDspState = stageItem.granularSynth->granularSynthDspState;

        dspItemList.back().outputSignal = &item->signal->getFrame();
    }
    return dspData;
}

void GranularSynthesisStage::processBlock(DspContext ctx)
{
    auto& currentItemList = ctx.getData<GranularSynthesisStage::DspData>().dspItemList;
    for (auto dspItem : currentItemList) {
        GenerateGranularSynthSignals(dspItem.granularSynthDspState, dspItem.outputSignal, B);
    }
}

void GranularSynthesisStage::configValueChanged(const std::string& key) {}

void GranularSynthesisStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }

    this->itemStore = itemStore;
    B = streamManager->getB();
    Fs = streamManager->getFs();
    this->streamManager = streamManager;

    // Load granular synthesis payload
    try {
        granularPayload = std::dynamic_pointer_cast<const MPEGI::GranularPayload>(
            bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_GRANULAR));
        granularMap = granularPayload->getGranular();
    } catch (const EAR::Exception e) {
        // Do nothing
    }

    this->granStageItemCnt = 0;
}

} // namespace MPEGI
