/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_GRANULARSYNTHESISSTAGE
#define MPEGI_GRANULARSYNTHESISSTAGE

#include "GranularSynth.h"

#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/GranularPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/ConfigObserver.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <utility>

namespace EAR {
class RealBuffer;
struct Position;
} // namespace EAR

namespace MPEGI {

class GranularSynthDspState;
class GranularSynth;

//! Rendering stage generating granular synthesis signals for audio elements
/*!
* Inputs:
* - Items for which the source signal is to be generated with granular synthesis.
*
* Outputs:
* - Output audio data is generated for rendering items that should use granular synthesis signals.
*
* How it works:
* - 
*
* Stages order:
* - Required before: none
* - Required after: All stages that use audio input
*
* Config options:
* - none
*
*/
class GranularSynthesisStage : public ConfigObserver, public RenderStage
{
public:
    //! Constructor
    GranularSynthesisStage();

    //! Destructor
    ~GranularSynthesisStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

    //! Inherited from ConfigObserver
    void configValueChanged(const std::string& key) override;

    int granStageItemCnt;

private:
    friend class Renderer;
    RenderItemStore* itemStore;
    const StreamManager* streamManager;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState*) override;

    // Shared pointer of the GranularPayload payload
    std::shared_ptr<const MPEGI::GranularPayload> granularPayload{ nullptr };

    // The granularMap
    std::map<std::string, std::shared_ptr<GranularDatabase>> granularMap;

    /*
     *  DSPItem
     *
     *  This structure keeps a pointer to the DSP state of a granular synth along with
     *  pointers to the output signals. Copys of this structure is sent to the
     *  ProcessBlock function.
     */
    struct DSPItem
    {
        EAR::RealFrame* outputSignal;
        GranularSynthDspState* granularSynthDspState;

        DSPItem(GranularSynthDspState* granularSynthDspState_p)
            : granularSynthDspState(granularSynthDspState_p){};
    };

    struct GranularSynthesisStageItem : RenderItem::ExtraState
    {
        int idx;
        GranularSynthesisStage* parent;
        GranularSynth* granularSynth;
        std::shared_ptr<StreamBuffer> outBuffer;

        GranularSynthesisStageItem(const RenderItem& renderItem, GranularSynthesisStage* parent)
            : idx(renderItem.idx)
            , parent(parent)
            , outBuffer(nullptr)
        {
            granularSynth = new GranularSynth(renderItem.type, parent->B, parent->Fs);

            auto granularDatabase = parent->granularMap.find(renderItem.granularDatabaseId);
            if (granularDatabase != parent->granularMap.end()) {
                granularSynth->SetGranularDatabase(granularDatabase->second);
            }

            granularSynth->InitState();
        };
    };

    std::unordered_map<int, GranularSynthesisStageItem> granularSynthesisStageItems;

    struct DspData : RenderStage::DspData
    {
        std::vector<DSPItem> dspItemList;
    };

    std::size_t B;
    int Fs;
};

} // namespace MPEGI

#endif /* MPEGI_GRANULARSYNTHESISSTAGE */
