/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Heterogeneous.h"
#include <EAR/MpegiDecorrelator.h>
#include <algorithm>

#include <EAR/DistanceModels.h>
namespace MPEGI {

// Local geometry functions that should be removed later
void ESAR_SPATIAL_VecInit(float* const Pos_p, float const PosX, float const PosY, float const PosZ);
void ESAR_SPATIAL_VecCopy(float* const PosTarget_p, float const* const PosSource_p);
float ESAR_SPATIAL_VecInnProd(float const* const Vec1_p, float const* const Vec2_p);
void ESAR_SPATIAL_VecCrossProd(
    float const* const Vec1_p, float const* const Vec2_p, float* const CrossProd_p);
float ESAR_SPATIAL_VecNorm(float const* const Vec_p);
void ESAR_SPATIAL_VecNormalize(float const* const Vec_p, float* const VecNorm_p);
void ESAR_SPATIAL_VecScale(float const* const Vec_p, float const scaler, float* const VecScale_p);
void ESAR_SPATIAL_VecNeg(float const* const Vec_p, float* const VecNeg_p);
void ESAR_SPATIAL_VecAdd(
    float const* const Vec1_p, float const* const Vec2_p, float* const VecAdd_p);
void ESAR_SPATIAL_VecSub(
    float const* const Vec1_p, float const* const Vec2_p, float* const VecSub_p);
void ESAR_SPATIAL_VecProjError(
    float const* const Vec1_p, float const* const Vec2_p, float* const ProjErr_p);
bool ESAR_SPATIAL_VecEpsilonComparison(
    float const* const Vec1_p, float const* const Vec2_p, float maxRelDiff);
void ESAR_SPATIAL_ClosestPointOnPlane(float const* const referencePoint_p,
    float const* const planeVertex0_p, float const* const planeVertex1_p,
    float const* const planeVertex2_p, float* const closestPoint_p);
bool ESAR_SPATIAL_IsPointInTriangle(float const* const referencePoint_p,
    float const* const triangleVertex0_p, float const* const triangleVertex1_p,
    float const* const triangleVertex2_p);
void ESAR_SPATIAL_ClosestPointOnLine(float const* const referencePoint_p,
    float const* const lineStartPoint_p, float const* const lineEndPoint_p,
    float* const closestPoint_p);
void ESAR_SPATIAL_ClosestPointOnTriangle(float const* const referencePoint_p,
    float const* const triangleVertex0_p, float const* const triangleVertex1_p,
    float const* const triangleVertex2_p, float* const closestPoint_p, float* const distance_p);
void ESAR_SPATIAL_EvalOrthonormOrient(float* const FrontVecON_p, float* const UpVecON_p,
    float* const RightVecON_p, float const* const FrontVec_p, float const* const UpVec_p);
void ESAR_SPATIAL_MapPointToLocalSystem(const float* vertex, float* local_vertex,
    float* listenerPos, const float* const observationSystemFrontVec,
    const float* const observationSystemUpVec, const float* const observationSystemRightVec);

// Local help functions
float ESAR_SPATIAL_Smooth_angle(float previous_angle, float new_angle, float smoothing_coeff);
void ESAR_SPATIAL_FindMeshOutline(const EAR::Mesh* mesh, std::vector<int>* outlineVerticesInd);
void ESAR_SPATIAL_GetNeighborVertices(
    int referenceVertexInd, const EAR::Mesh* mesh, std::vector<int>* neighborList);
void ESAR_SPATIAL_GenerateOutlineNormalList(
    const EAR::Mesh* mesh, std::vector<OutlineNormal>* outlineNormals);
int ESAR_SPATIAL_FindClosestOutlineNormalIndex(
    std::vector<OutlineNormal> outlineNormals, Vector3 refPosition);
int ESAR_SPATIAL_FindClosestOutlineNormalIndex_LineSearch(const EAR::Mesh* mesh,
    std::vector<OutlineNormal> outlineNormals, Vector3 refPosition,
    float* pShortestSquaredDistance);
bool ESAR_SPATIAL_IsInsideOutline(std::vector<OutlineNormal> outlineNormals, Vector3 refPosition);
float ESAR_SPATIAL_FindDistanceToOutline(
    const EAR::Mesh* mesh, std::vector<OutlineNormal> outlineNormals, Vector3 refPosition);

// Local design constants
#define ESAR_SRC_SPATIAL_MAX_NUM_OF_NEIGHBOR_FACES 20
#define ESAR_SRC_SPATIAL_MAX_VERTICAL_FACE_ANGLE 15.0f
#define ESAR_SRC_SPATIAL_INTERIOR_DEADZONE_DISTANCE 10.0f
#define ESAR_SRC_SPATIAL_ROTATION_TRANSITION_LENGTH 10.0f
#define ESAR_SRC_SPATIAL_INTERIOR_FADE_DISTANCE 0.1f
#define ESAR_SRC_SPATIAL_INTERIOR_REAR_SUPPRESSION_DISTANCE 8.0f
#define ESAR_GAIN_THRESHOLD 1.0e-5

// Local geometric constants
#define ESAR_PI 3.141592653589793f // Pi
#define ESAR_PI_DIV_BY_180 0.017453292519943f // Pi/180
#define ESAR_180_DIV_BY_PI 57.295779513082323f // 180/Pi
#define ESAR_EPSILON_DISTANCE 0.000001

//////////////////////////////////////////////////////////////////////////////////////////////////////
// The positions of the twelve virtual loudspeakers used for rendering the interior representation.
//
// Order of speaker positions:
// Left, Right, Front, Rear,
// Front-left-high, Front-left-low, Front-right-high, Front-right-low,
// Rear-Left-hight, Rear-left-low, Rear-right-high, Rear-right-low
//////////////////////////////////////////////////////////////////////////////////////////////////////
static float intSpAz[12] = { -90.0, 90.0, 0.0, -180.0, -45.0, -45.0, 45.0, 45.0, -135.0, -135.0,
    135.0, 135.0 };
static float intSpEl[12] = { 0.0, 0.0, 0.0, 0.0, 45.0, -45.0, 45.0, -45.0, 45.0, -45.0, 45.0,
    -45.0 };

// Interior speakers mapping to output speakers
static int intToOutputSpMapping[12] = { 0, 1, 2, 3, 6, 7, 8, 9, 10, 11, 12, 13 };

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// HeterogeneouosSource class definitions
//
//////////////////////////////////////////////////////////////////////////////////////////////////////
HeterogeneousSource::HeterogeneousSource(MPEGI::ItemType itType,
    const ExtentLayout* itemExtentLayout, const EAR::Geometry* extent, float refDist,
    float transitionDist, std::size_t blocksize, int inputChannels, bool integr3DoFHoa)
    : B(blocksize)
    , itemType(itType)
    , numOfOutputSpeakers(0)
    , numOfInteriorSpeakers(0)
    , numOfExteriorSpeakers(0)
    , distanceGainModel(0)
    , diffuseness(1.0)
    , refDistance(refDist)
    , intSpeakerHorizRotation(0)
    , deltaDistanceChange(0.0)
    , previousListenerLocation(0.0, 0.0, 0.0)
    , previousSourceLocation(0.0, 0.0, 0.0)
    , previousSourceOrientation()
    , previousObservationFrontVector(0.00, 0.00, 1.00)
    , anchorPoint(0.0, 0.0, 0.0)
    , anchorPointInitialized(false)
    , leftEdgePoint(0.0, 0.0, 0.0)
    , leftEdgePointInitialized(false)
    , rightEdgePoint(0.0, 0.0, 0.0)
    , rightEdgePointInitialized(false)
    , maxAzim(0.0)
    , minAzim(0.0)
    , maxElev(0.0)
    , minElev(0.0)
    , planeWidth(0.0)
    , planeHeight(0.0)
    , topEdgePoint(0.0, 0.0, 0.0)
    , topEdgePointInitialized(false)
    , bottomEdgePoint(0.0, 0.0, 0.0)
    , bottomEdgePointInitialized(false)
    , distanceToExtent(0.0)
    , interiorExteriorWeight(0.0)
    , center_gain_comp(0.0)
    , distanceToExtentOutline(0.0)
    , closestNormalIndex(0)
    , rearHemisphereWeight(0.0)
    , topHemisphereWeight(0.0)
    , bottomHemisphereWeight(0.0)
    , integrated3DoFHoa(integr3DoFHoa)
    , ThreeDofHoaGain(0.0)
    , relativeHorizAngleExterior(0.0)
    , relativeVertAngleExterior(0.0)
    , relativeHorizAngleInterior(0.0)
    , relativeVertAngleInterior(0.0)
    , occlusion_gain_left(0.0)
    , occlusion_gain_middle(0.0)
    , occlusion_gain_right(0.0)
    , occludedExtentGain(0.0)
    , gridShiftCnt(0)
    , prevCentPointDir(0.0, 0.0, 0.0)
    , occlLeftAngleScaler(1.0)
    , occlRightAngleScaler(1.0)
    , occlTopAngleScaler(1.0)
    , occlBottomAngleScaler(1.0)
    , subareaOcclHorizRaysHigh(10)
    , subareaOcclVertRaysHigh(5)
    , subareaOcclHorizRaysLow(1)
    , subareaOcclVertRaysLow(1)
    , activeExtentRatio(1.0)
    , occlusionDetected(false)
{
    if (itemExtentLayout != NULL) {
        extentLayout.alignment = itemExtentLayout->alignment;
        extentLayout.positioning = itemExtentLayout->positioning;
    } else {
        if (itType == MPEGI::ItemType::HOA) {
            extentLayout.alignment = ExtentLayout::Alignment::Object;
        } else {
            extentLayout.alignment = ExtentLayout::Alignment::User;
        }
    }

    transitionDistance = transitionDist;
    EAR::Position pos;
    pos.location.set(0.0, 0.0, 0.0);
    pos.cSpace = EAR::CSpace::WCS;
    numOfExteriorSpeakers = HETEROGENEOUS_NUM_OF_EXTERIOR_SPEAKERS;
    numOfInteriorSpeakers = HETEROGENEOUS_NUM_OF_INTERIOR_SPEAKERS;
    numOfOutputSpeakers = HETEROGENEOUS_NUM_OF_OUTPUT_SPEAKERS;
    trans_fac[0] = 0.0f;
    trans_fac[1] = 0.0f;
    trans_fac[2] = (float)inputChannels;

    for (int loop = 0; loop < numOfOutputSpeakers; loop++) {
        outputSpeakerPosVec.push_back(EAR::Position(pos));
        outputSpeakerDistanceOffsets.push_back(0.0);
        outputSpeakerDistanceGains.push_back(1.0);
        outputSpeakerActive.push_back(true);
    }
    for (int loop = 0; loop < numOfExteriorSpeakers; loop++) {
        exteriorSpeakerPosVec.push_back(EAR::Position(pos));
        exteriorSpeakerGains.push_back(0.0);
    }
    for (int loop = 0; loop < numOfInteriorSpeakers; loop++) {
        interiorSpeakerPosVec.push_back(EAR::Position(pos));
        interiorSpeakerGains.push_back(0.0);
    }

    distanceModel = std::make_unique<EAR::DistanceGainVolumetric>(
        Config::getFloatDef("Distance_MinDistance", 0.2f));

    heterogeneousDspState =
        new HeterogeneousDspState(itType, itemExtentLayout, B, inputChannels, integrated3DoFHoa);

    previousSourceLocation = extent->getGlobalLocation();
    previousSourceOrientation = extent->getGlobalOrientation();
    ESAR_SPATIAL_GenerateOutlineNormalList(extent->getMesh(), &outlineNormals);

    // TODO: Make this dependent on the update rate rather than the frame length
    temporalDescrPointSmoothFact = 1.0 - ((float)B / 48000.0) * 2.0;

    eqnBands = Config::getInt("EQNBands");
    occlusion_eq_top_left.resize(eqnBands);
    occlusion_eq_top_mid.resize(eqnBands);
    occlusion_eq_top_right.resize(eqnBands);
    occlusion_eq_bottom_left.resize(eqnBands);
    occlusion_eq_bottom_mid.resize(eqnBands);
    occlusion_eq_bottom_right.resize(eqnBands);
    eq_centerFreqs.resize(eqnBands);
    speaker_occlusion_eq.resize(eqnBands);
    speaker_occlusion_eq_0.resize(eqnBands);
    speaker_occlusion_eq_1.resize(eqnBands);
    speaker_occlusion_eq_2.resize(eqnBands);
    speaker_occlusion_eq_4.resize(eqnBands);
    speaker_occlusion_eq_5.resize(eqnBands);
    auto centerFreqs = Config::getFloatPtr("EQCenterFreqs");
    for (int i = 0; i < Config::getInt("EQNBands"); i++) {
        eq_centerFreqs[i] = centerFreqs[i];
        occlusion_eq_top_left[i] = 1.0;
        occlusion_eq_top_mid[i] = 1.0;
        occlusion_eq_top_right[i] = 1.0;
        occlusion_eq_bottom_left[i] = 1.0;
        occlusion_eq_bottom_mid[i] = 1.0;
        occlusion_eq_bottom_right[i] = 1.0;
        speaker_occlusion_eq_0[i] = 1.0;
        speaker_occlusion_eq_1[i] = 1.0;
        speaker_occlusion_eq_2[i] = 1.0;
        speaker_occlusion_eq_4[i] = 1.0;
        speaker_occlusion_eq_5[i] = 1.0;
        speaker_occlusion_eq[i] = 1.0;
    }
    subareaEqMap.emplace(0, &occlusion_eq_top_left);
    subareaEqMap.emplace(1, &occlusion_eq_top_mid);
    subareaEqMap.emplace(2, &occlusion_eq_top_right);
    subareaEqMap.emplace(3, &occlusion_eq_bottom_left);
    subareaEqMap.emplace(4, &occlusion_eq_bottom_mid);
    subareaEqMap.emplace(5, &occlusion_eq_bottom_right);
    outputSpeakerEqMap.emplace(0, &speaker_occlusion_eq_0);
    outputSpeakerEqMap.emplace(1, &speaker_occlusion_eq_1);
    outputSpeakerEqMap.emplace(2, &speaker_occlusion_eq_2);
    outputSpeakerEqMap.emplace(4, &speaker_occlusion_eq_4);
    outputSpeakerEqMap.emplace(5, &speaker_occlusion_eq_5);
    occludedExtentPos = pos;
}

HeterogeneousSource::~HeterogeneousSource()
{
    delete (heterogeneousDspState);
}

bool HeterogeneousSource::UpdateState(const EAR::Geometry* extent, EAR::Position sourcePosition,
    EAR::Position listenerPosition, std::unique_ptr<EAR::EmbreeTracer>& tracer)
{
    EAR::Vector3 refPoint;
    float activeExtRatio = 1.0;

    if ((extent->getGlobalLocation() != previousSourceLocation)
        || (extent->getGlobalOrientation() != previousSourceOrientation)) {
        // Update the outline normal list only if the source has moved
        ESAR_SPATIAL_GenerateOutlineNormalList(extent->getMesh(), &outlineNormals);
    }

    FindExtentDescriptionPoints(extent, listenerPosition.location);

    RenderOcclusion(extent, tracer, listenerPosition.location, &activeExtRatio);

    CalculateEffectiveExtent(listenerPosition);

    UpdateExteriorRepresentationSpeakerSetup(listenerPosition);

    UpdateInteriorRepresentationSpeakerSetup(extent, listenerPosition);

    CloseToSurfaceModificationsOfInteriorRepresentationRendering(extent, listenerPosition);

    // Calculate the weight of the interior and exterior representations
    float modTransDist = activeExtRatio * transitionDistance;
    float extra_margin = modTransDist * 0.1f;
    if (distanceToExtent > modTransDist + extra_margin) {
        interiorExteriorWeight = 0.9f * interiorExteriorWeight;
    } else if (distanceToExtent <= 0.0f + extra_margin) {
        interiorExteriorWeight = 0.9f * interiorExteriorWeight + 0.1f;
    } else {
        interiorExteriorWeight = 0.9f * interiorExteriorWeight
            + 0.1f * (modTransDist - (distanceToExtent - extra_margin)) / (modTransDist);
    }
    // Round to zero to avoid denormal numbers and deactivate proccesing for silent speakers
    if (interiorExteriorWeight < 1.0e-5) {
        interiorExteriorWeight = 0.0;
    }

    // Set output positions
    float intWeight = interiorExteriorWeight;
    float extWeight = 1.0 - intWeight;
    // Exterior: bottom left & right, Interior: left & right
    outputSpeakerPosVec[0].location = intWeight * interiorSpeakerPosVec[0].location
        + extWeight * exteriorSpeakerPosVec[0].location;
    outputSpeakerPosVec[1].location = intWeight * interiorSpeakerPosVec[1].location
        + extWeight * exteriorSpeakerPosVec[1].location;
    // Exterior: middle, Interior: front
    outputSpeakerPosVec[2].location = intWeight * interiorSpeakerPosVec[2].location
        + extWeight * exteriorSpeakerPosVec[2].location;
    // Interior: rear
    outputSpeakerPosVec[3].location = interiorSpeakerPosVec[3].location;
    // Exterior: top left & right
    outputSpeakerPosVec[4].location = exteriorSpeakerPosVec[3].location;
    outputSpeakerPosVec[5].location = exteriorSpeakerPosVec[4].location;
    // Interior
    outputSpeakerPosVec[6].location = interiorSpeakerPosVec[4].location;
    outputSpeakerPosVec[7].location = interiorSpeakerPosVec[5].location;
    outputSpeakerPosVec[8].location = interiorSpeakerPosVec[6].location;
    outputSpeakerPosVec[9].location = interiorSpeakerPosVec[7].location;
    outputSpeakerPosVec[10].location = interiorSpeakerPosVec[8].location;
    outputSpeakerPosVec[11].location = interiorSpeakerPosVec[9].location;
    outputSpeakerPosVec[12].location = interiorSpeakerPosVec[10].location;
    outputSpeakerPosVec[13].location = interiorSpeakerPosVec[11].location;

    UpdateOutputSpeakerDistanceOffset(listenerPosition.location);

    // Calculate our own distance gain here
    for (int i = 0; i < numOfOutputSpeakers; i++) {
        float distGain;
        if (distanceGainModel == 0) {
            // Use volumetric distance gain model
            distGain = distanceModel->getGainForDistance(
                (outputSpeakerPosVec[i].location - listenerPosition.location).abs(), planeWidth,
                planeHeight, refDistance);
        } else {
            // Use point-source distance gain model
            distGain = distanceModel->getGainForDistance(
                (outputSpeakerPosVec[i].location - listenerPosition.location).abs(), 0.0, 0.0,
                refDistance);
        }

        if (distGain > 1.0)
            distGain = 1.0;
        outputSpeakerDistanceGains[i] = distGain;
    }

    // Update the angle between the observation vector and the orientation of the audio element
    if (extentLayout.alignment == ExtentLayout::Alignment::User) {
        relativeHorizAngleExterior = 0.0;
    } else {
        // Horizontal angle
        Vector3 relativeAncorPoint =
            extent->getGlobalPosition().location - listenerPosition.location;
        Polar polarObservationVector = cartesianToPolar(relativeAncorPoint);
        float objectJawAngle = sourcePosition.orientation[0];
        float newRelativeAngle = (polarObservationVector[0] + 180.0f)
            - objectJawAngle; // We compare with the negative observation angle
        float prevRelativeAngle = relativeHorizAngleExterior * ESAR_180_DIV_BY_PI;

        relativeHorizAngleExterior = ESAR_PI_DIV_BY_180
            * ESAR_SPATIAL_Smooth_angle(prevRelativeAngle, newRelativeAngle, 0.9f);

        // Vertical angle
        float objectPitchAngle = extent->getGlobalPosition().orientation[1];
        newRelativeAngle = -polarObservationVector[1]
            - objectPitchAngle; // We compare with the negative observation angle
        prevRelativeAngle = relativeVertAngleExterior * ESAR_180_DIV_BY_PI;

        relativeVertAngleExterior = ESAR_PI_DIV_BY_180
            * ESAR_SPATIAL_Smooth_angle(prevRelativeAngle, newRelativeAngle, 0.9f);

        // Update the relative horizontal angle between the interior speaker setup and the source's orientation
        newRelativeAngle = intSpeakerHorizRotation + 180.0f
            - objectJawAngle; // We compare with the negative observation angle
        relativeHorizAngleInterior = newRelativeAngle * ESAR_PI_DIV_BY_180;
    }

    ThreeDofHoaGain = interiorExteriorWeight * rearHemisphereWeight * rearHemisphereWeight;

    UpdateOutputSpeakerActivation();

    previousSourceLocation = extent->getGlobalLocation();
    previousSourceOrientation = extent->getGlobalOrientation();

    return true;
}

void HeterogeneousSource::UpdateDspState()
{
    heterogeneousDspState->UpdateState(&extentLayout, interiorExteriorWeight,
        relativeHorizAngleExterior, relativeVertAngleExterior, relativeHorizAngleInterior,
        relativeVertAngleInterior, rearHemisphereWeight, topHemisphereWeight,
        bottomHemisphereWeight, trans_fac, exteriorSpeakerGains, interiorSpeakerGains,
        occlLeftAngleScaler, occlRightAngleScaler, occlTopAngleScaler, occlBottomAngleScaler);
}

void HeterogeneousSource::FindExtentDescriptionPoints(
    const EAR::Geometry* extent, Vector3 listenerPosition)
{
    const Mesh* pMesh = extent->getMesh();
    int vertexInd, anchorPointInd;
    int faceInd, vertInd0, vertInd1, vertInd2, corner;
    float vertex_relative[3], anchorPoint_relative[3], anchorPoint_world[3], listenerPos[3],
        vertex_world[3];
    float faceNormal[3], vertA[3], vertB[3], faceVec1[3], faceVec2[3], closeFaceMidPoint[3],
        vectorToFaceMidPoint[3];
    float closestPointOnTriangle[3];
    float min_distance, distance, maxAbsNormalRayDot;
    bool bContinue;
    float facePoint0[3], facePoint1[3], facePoint2[3];
    int closeFaces[ESAR_SRC_SPATIAL_MAX_NUM_OF_NEIGHBOR_FACES];
    int numOfCloseFaces = 0;
    float normalRayDot, normalRayDots[ESAR_SRC_SPATIAL_MAX_NUM_OF_NEIGHBOR_FACES];

    float observationFrontVec[3], anchor_point_elev, observationUpVec[3], mapped_vertex[3];
    float observationSystemFrontVec[3], observationSystemUpVec[3], observationSystemRightVec[3];
    float minAzimVertex_local[3], maxAzimVertex_local[3], horizontalDirectionVec_local[3];
    float vertex_elev, vertex_azim, horizontalAngle, temp_angle;
    float max_elev_dist, min_elev_dist, verticalAngle;
    float minElevVertex_local[3], maxElevVertex_local[3], verticalDirectionVec_local[3];
    float displacement_left, displacement_right, geometric_width, width_scaler;
    float height_scaler, height_scaler_bottom;
    float observationSystemInverseFrontVec[3], observationSystemInverseUpVec[3],
        observationSystemInverseRightVec[3];
    float horizontalDirectionVec_relative[3], displacementVecLeft[3], leftEdgePos_relative[3];
    float displacementVecRight[3], rightEdgePos_relative[3], leftEdgePos_world[3],
        rightEdgePos_world[3], verticalDirectionVec_relative[3];
    float displacementVecTop[3], topEdgePos_relative[3], topEdgePos_world[3],
        displacementVecBottom[3], bottomEdgePos_relative[3], bottomEdgePos_world[3];
    float displacement_top, displacement_bottom;
    Vector3 previousDistanceVector, currentDistanceVector;
    int min_elev_index, max_elev_index;
    listenerPos[0] = listenerPosition[0];
    listenerPos[1] = listenerPosition[1];
    listenerPos[2] = listenerPosition[2];
    Vector3 midSpeakerPoint;

    std::vector vertices = pMesh->getVertices(Mesh::CSpace::Global);
    std::vector faces = pMesh->getTriangles();

    // Update the delta distance between the listener and the extent
    previousDistanceVector = previousListenerLocation - previousSourceLocation;
    currentDistanceVector = listenerPosition - extent->getGlobalLocation();
    deltaDistanceChange =
        0.9 * deltaDistanceChange + 0.1 * (currentDistanceVector - previousDistanceVector).abs();
    previousListenerLocation = listenerPosition;

    //
    // Find the anchor point of the extent
    //

    // Define the anchor point as the closest point to the listener
    ESAR_SPATIAL_VecInit(vertex_world, vertices[0][0], vertices[0][1], vertices[0][2]);
    ESAR_SPATIAL_VecSub(vertex_world, listenerPos, anchorPoint_relative);
    anchorPointInd = 0;
    // We avoid the square root of calculating the norm since we only want to find the shortest distance
    min_distance = ESAR_SPATIAL_VecInnProd(anchorPoint_relative, anchorPoint_relative);

    for (vertexInd = 0; vertexInd < vertices.size(); vertexInd++) {
        ESAR_SPATIAL_VecInit(
            vertex_world, vertices[vertexInd][0], vertices[vertexInd][1], vertices[vertexInd][2]);
        ESAR_SPATIAL_VecSub(vertex_world, listenerPos, vertex_relative);
        distance = ESAR_SPATIAL_VecInnProd(vertex_relative, vertex_relative);
        if (distance < min_distance) {
            ESAR_SPATIAL_VecCopy(anchorPoint_relative, vertex_relative);
            anchorPointInd = vertexInd;
            min_distance = distance;
        }
    }

    // Now calculate the real min distance
    min_distance = sqrtf(min_distance);

    // Now check the faces that the closest vertice belong to, to find if there is any point
    // within these that is even closer. Also check the normal of each face to see if the listener
    // position is front or the back of the face surface.
    bContinue = true;
    for (faceInd = 0; faceInd < faces.size(); faceInd++) {
        for (corner = 0; corner < 3; corner++) {
            if (corner == 0)
                vertexInd = faces[faceInd].v1;
            else if (corner == 1)
                vertexInd = faces[faceInd].v2;
            else
                vertexInd = faces[faceInd].v3;

            ESAR_SPATIAL_VecInit(vertex_world, vertices[vertexInd][0], vertices[vertexInd][1],
                vertices[vertexInd][2]);
            ESAR_SPATIAL_VecInit(anchorPoint_world, vertices[anchorPointInd][0],
                vertices[anchorPointInd][1], vertices[anchorPointInd][2]);
            if (ESAR_SPATIAL_VecEpsilonComparison(vertex_world, anchorPoint_world, (float)0.001)) {
                // Found a face that the closest point belongs to
                closeFaces[numOfCloseFaces++] = faceInd;
            }
            if (numOfCloseFaces >= ESAR_SRC_SPATIAL_MAX_NUM_OF_NEIGHBOR_FACES) {
                /*printf("Number faces that includes the closest point was more than the maximum "
                           "handled, skipping the rest!\n");*/
                bContinue = false;
                break;
            }
        }
        if (!bContinue) {
            break;
        }
    }

    maxAbsNormalRayDot = 0;
    for (faceInd = 0; faceInd < numOfCloseFaces; faceInd++) {
        // Calculate the normal of the face
        vertInd0 = faces[closeFaces[faceInd]].v1;
        vertInd1 = faces[closeFaces[faceInd]].v2;
        vertInd2 = faces[closeFaces[faceInd]].v3;

        ESAR_SPATIAL_VecInit(
            vertA, vertices[vertInd1][0], vertices[vertInd1][1], vertices[vertInd1][2]);
        ESAR_SPATIAL_VecInit(
            vertB, vertices[vertInd0][0], vertices[vertInd0][1], vertices[vertInd0][2]);
        ESAR_SPATIAL_VecSub(vertA, vertB, faceVec1);
        ESAR_SPATIAL_VecInit(
            vertA, vertices[vertInd2][0], vertices[vertInd2][1], vertices[vertInd2][2]);
        ESAR_SPATIAL_VecSub(vertA, vertB, faceVec2);
        ESAR_SPATIAL_VecCrossProd(faceVec1, faceVec2, faceNormal);
        ESAR_SPATIAL_VecNormalize(faceNormal, faceNormal);

        // Calculate the midpoint of the face
        ESAR_SPATIAL_VecInit(
            vertA, vertices[vertInd0][0], vertices[vertInd0][1], vertices[vertInd0][2]);
        ESAR_SPATIAL_VecInit(
            vertB, vertices[vertInd1][0], vertices[vertInd1][1], vertices[vertInd1][2]);
        ESAR_SPATIAL_VecAdd(vertA, vertB, closeFaceMidPoint);
        ESAR_SPATIAL_VecInit(
            vertB, vertices[vertInd2][0], vertices[vertInd2][1], vertices[vertInd2][2]);
        ESAR_SPATIAL_VecAdd(closeFaceMidPoint, vertB, closeFaceMidPoint);
        ESAR_SPATIAL_VecScale(closeFaceMidPoint, (float)(1.0 / 3.0), closeFaceMidPoint);

        // Check for a closer point on this face
        ESAR_SPATIAL_VecInit(
            facePoint0, vertices[vertInd0][0], vertices[vertInd0][1], vertices[vertInd0][2]);
        ESAR_SPATIAL_VecInit(
            facePoint1, vertices[vertInd1][0], vertices[vertInd1][1], vertices[vertInd1][2]);
        ESAR_SPATIAL_VecInit(
            facePoint2, vertices[vertInd2][0], vertices[vertInd2][1], vertices[vertInd2][2]);

        ESAR_SPATIAL_ClosestPointOnTriangle(
            listenerPos, facePoint0, facePoint1, facePoint2, closestPointOnTriangle, &distance);
        if (distance < min_distance) {
            min_distance = distance;
            ESAR_SPATIAL_VecSub(closestPointOnTriangle, listenerPos, anchorPoint_relative);
        }

        // Calculate the dot-product of vector from the listener position and the normal
        ESAR_SPATIAL_VecSub(
            closestPointOnTriangle /*closeFaceMidPoint*/, listenerPos, vectorToFaceMidPoint);
        ESAR_SPATIAL_VecNormalize(vectorToFaceMidPoint, vectorToFaceMidPoint);
        normalRayDot = ESAR_SPATIAL_VecInnProd(faceNormal, vectorToFaceMidPoint);

        normalRayDots[faceInd] = normalRayDot;

        // Store the dot product with the largest absolute value. This represent the face that is
        // most angled against (or away from) the listening position
        if (fabsf(normalRayDot) > fabsf(maxAbsNormalRayDot)) {
            maxAbsNormalRayDot = normalRayDot;
        }
    }

    ESAR_SPATIAL_VecAdd(anchorPoint_relative, listenerPos, anchorPoint_world);

    UpdateDescriptionPoint(stored_anchorPoint, anchorPoint_world, deltaDistanceChange,
        temporalDescrPointSmoothFact, stored_anchorPoint, &anchorPointInitialized);

    // Update the minimum distance
    ESAR_SPATIAL_VecSub(stored_anchorPoint, listenerPos, anchorPoint_relative);
    min_distance = ESAR_SPATIAL_VecNorm(anchorPoint_relative);

    if (maxAbsNormalRayDot > 0.0) {
        // The listener is inside the extent
        distanceToExtent = -min_distance;
    } else {
        distanceToExtent = min_distance;
    }

    // Handle special case when the listener position is coinciding with the anchor point
    if ((fabs(anchorPoint_relative[0]) == 0.0) && (fabs(anchorPoint_relative[1]) == 0.0)
        && (fabs(anchorPoint_relative[2]) == 0.0)) {
        distanceToExtent = 0;
    }

    if (distanceToExtent <= transitionDistance * 0.001f) {
        // Reset the smoothing of the description points since these should jump
        // instantly to their new positions when the listener goes out of the extent again
        anchorPointInitialized = false;
        leftEdgePointInitialized = false;
        rightEdgePointInitialized = false;
        topEdgePointInitialized = false;
        bottomEdgePointInitialized = false;

        return; // No need to calculate the exterior representation if we are inside the extent
    }

    //
    // Find the edge description points
    //

    // Front-vector
    ESAR_SPATIAL_VecNormalize(anchorPoint_relative, observationFrontVec);

    // Avoid the case where the observation vector is pointing straight up or down
    if (fabs(observationFrontVec[1]) > 0.9) {

        // Base the observation vector on the closest outline normal
        float normalAzimuth = atan2f(outlineNormals[closestNormalIndex].normal[0],
            outlineNormals[closestNormalIndex].normal[2]);
        ESAR_SPATIAL_VecInit(observationFrontVec, -sinf(normalAzimuth) / 100.0f,
            observationFrontVec[1] / fabsf(observationFrontVec[1]), -cosf(normalAzimuth) / 100.0f);
        ESAR_SPATIAL_VecNormalize(observationFrontVec, observationFrontVec);
    } else {
        previousObservationFrontVector.set(
            observationFrontVec[0], observationFrontVec[1], observationFrontVec[2]);
    }

    // Up-vector
    anchor_point_elev = asinf(observationFrontVec[1]);
    observationUpVec[0] = -sinf(anchor_point_elev) * observationFrontVec[0];
    observationUpVec[1] = cosf(anchor_point_elev);
    observationUpVec[2] = -sinf(anchor_point_elev) * observationFrontVec[2];
    ESAR_SPATIAL_VecNormalize(observationUpVec, observationUpVec);

    // Calculate orthonormal vectors for the observation coordinate system
    ESAR_SPATIAL_EvalOrthonormOrient(observationSystemFrontVec, observationSystemUpVec,
        observationSystemRightVec, observationFrontVec, observationUpVec);

    // Go through the mesh and find the points with maximum and minimum azimuth
    maxAzim = -1.0;
    minAzim = 1.0;
    maxElev = -1.0;
    minElev = 1.0;
    min_elev_index = 0;
    max_elev_index = 0;
    for (vertexInd = 0; vertexInd < vertices.size(); vertexInd++) {
        // Convert to world space coordinates

        ESAR_SPATIAL_VecInit(
            vertex_world, vertices[vertexInd][0], vertices[vertexInd][1], vertices[vertexInd][2]);
        ESAR_SPATIAL_VecSub(vertex_world, listenerPos, vertex_relative);

        mapped_vertex[0] = ESAR_SPATIAL_VecInnProd(vertex_relative, observationSystemFrontVec);
        mapped_vertex[1] = ESAR_SPATIAL_VecInnProd(vertex_relative, observationSystemUpVec);
        mapped_vertex[2] = ESAR_SPATIAL_VecInnProd(vertex_relative, observationSystemRightVec);

        // Evaluate the azimuth and elevation of the point in the observation coordinate system
        if ((mapped_vertex[0] == 0) && (mapped_vertex[1] == 0) && (mapped_vertex[2] == 0)) {
            vertex_elev = 0.0;
            vertex_azim = 0.0;
        } else {
            vertex_azim = ESAR_180_DIV_BY_PI * atan2f(mapped_vertex[2], mapped_vertex[0]);
            vertex_elev =
                ESAR_180_DIV_BY_PI * asinf(mapped_vertex[1] / ESAR_SPATIAL_VecNorm(mapped_vertex));
        }

        if ((vertex_azim < 90.0) && (vertex_azim > -90.0) && (vertex_elev < 89.0)
            && (vertex_elev > -89.0)) {
            if (vertex_azim > maxAzim) {
                maxAzim = vertex_azim;
                ESAR_SPATIAL_VecCopy(maxAzimVertex_local, mapped_vertex);
            }
            if (vertex_azim < minAzim) {
                minAzim = vertex_azim;
                ESAR_SPATIAL_VecCopy(minAzimVertex_local, mapped_vertex);
            }
            if (vertex_elev > maxElev) {
                maxElev = vertex_elev;
                max_elev_index = vertexInd;
                ESAR_SPATIAL_VecCopy(maxElevVertex_local, mapped_vertex);
            }
            if (vertex_elev < minElev) {
                minElev = vertex_elev;
                min_elev_index = vertexInd;
                ESAR_SPATIAL_VecCopy(minElevVertex_local, mapped_vertex);
            }
        }
    }
    // Check for min/max elevation along the edges that go through the found max/min elevation vertices
    for (faceInd = 0; faceInd < numOfCloseFaces; faceInd++) {
        int faceVertInd[3], vertex1, vertex2;
        float closestPointOnEdge_world[3], vertex_local[3];

        faceVertInd[0] = faces[faceInd].v1;
        faceVertInd[1] = faces[faceInd].v2;
        faceVertInd[2] = faces[faceInd].v3;

        // Check if any of the edges in this face includes the min/max elevation vertex
        for (int side = 0; side < 3; side++) {
            vertex1 = side;
            if (side == 2) {
                vertex2 = 0;
            } else {
                vertex2 = side + 1;
            }

            // Calculate the elevation for the closest point on the edges that include the min/max elevation vertex
            if ((faceVertInd[vertex1] == min_elev_index) || (faceVertInd[vertex2] == min_elev_index)
                || (faceVertInd[vertex1] == max_elev_index)
                || (faceVertInd[vertex2] == max_elev_index)) {
                ESAR_SPATIAL_ClosestPointOnLine(listenerPos, vertices[faceVertInd[vertex1]].data(),
                    vertices[faceVertInd[vertex2]].data(), closestPointOnEdge_world);
                ESAR_SPATIAL_MapPointToLocalSystem(closestPointOnEdge_world, vertex_local,
                    listenerPos, observationSystemFrontVec, observationSystemUpVec,
                    observationSystemRightVec);
                vertex_elev = ESAR_180_DIV_BY_PI
                    * atan2f(vertex_local[1],
                        sqrtf(
                            vertex_local[0] * vertex_local[0] + vertex_local[2] * vertex_local[2]));

                if ((vertex_elev > maxElev) && (vertex_elev < 89.0)) {
                    maxElev = vertex_elev;
                    ESAR_SPATIAL_VecCopy(maxElevVertex_local, vertex_local);
                }

                if ((vertex_elev < minElev) && (vertex_elev > -89.0)) {
                    minElev = vertex_elev;
                    ESAR_SPATIAL_VecCopy(minElevVertex_local, vertex_local);
                }
            }
        }
    }
    if (maxAzim < 0.0) {
        maxAzim = 0.0;
        ESAR_SPATIAL_VecInit(
            maxAzimVertex_local, min_distance, 0.0, 0.0); // Set it to the local anchor point
    }
    if (minAzim > 0.0) {
        minAzim = 0.0;
        ESAR_SPATIAL_VecCopy(minAzimVertex_local, mapped_vertex);
        ESAR_SPATIAL_VecInit(
            maxAzimVertex_local, min_distance, 0.0, 0.0); // Set it to the local anchor point
    }
    if (maxElev < 0.0) {
        maxElev = 0.0;
        ESAR_SPATIAL_VecInit(
            maxElevVertex_local, min_distance, 0.0, 0.0); // Set it to the local anchor point
    }

    if (minElev > 0.0) {
        minElev = 0.0;
        ESAR_SPATIAL_VecInit(
            minElevVertex_local, min_distance, 0.0, 0.0); // Set it to the local anchor point
    }

    //
    // Construct the 2-dimensional plane that will represent the extent
    //

    // Calculate the vector that describes the horizontal angle between the left and right ends of the extent
    ESAR_SPATIAL_VecSub(maxAzimVertex_local, minAzimVertex_local, horizontalDirectionVec_local);
    horizontalDirectionVec_local[1] = 0.0; // Project on the local horizontal plane
    ESAR_SPATIAL_VecNormalize(horizontalDirectionVec_local, horizontalDirectionVec_local);

    // Calulate the magnitude of the displacement along the horizontal direction, corresponding to the max and minimum
    // azimuth angles using the sine rule.
    horizontalAngle = ESAR_180_DIV_BY_PI
        * atan2f(horizontalDirectionVec_local[0], horizontalDirectionVec_local[2]);
    temp_angle = fmax(5.0,
        90.0f + horizontalAngle
            + minAzim); // Limit the angle of the displacement to avoid div by zero

    displacement_left = (-min_distance * sinf(-minAzim / ESAR_180_DIV_BY_PI))
        / (sinf(temp_angle / ESAR_180_DIV_BY_PI));
    temp_angle = fmax(5.0,
        90.0f - horizontalAngle
            - maxAzim); // Limit the angle of the displacement to avoid div by zero
    displacement_right = (min_distance * sinf(maxAzim / ESAR_180_DIV_BY_PI))
        / (sinf((temp_angle) / ESAR_180_DIV_BY_PI));
    // Calculate the vector that describes the vertical angle between the top and bottom ends of the extent
    ESAR_SPATIAL_VecSub(maxElevVertex_local, minElevVertex_local, verticalDirectionVec_local);
    verticalDirectionVec_local[2] =
        0.0; // Project on the local vertical plane, as defined by the front and up vectors
    ESAR_SPATIAL_VecNormalize(verticalDirectionVec_local, verticalDirectionVec_local);

    // Calulate the magnitude of the displacement along the vertical direction, corresponding to the max and minimum
    // elevation angles using the sine rule.
    verticalAngle = 90.0f
        - ESAR_180_DIV_BY_PI * atan2f(verticalDirectionVec_local[1], verticalDirectionVec_local[0]);
    temp_angle = fmax(0.1f,
        90.0f - verticalAngle
            + maxElev); // Limit the angle of the displacement to avoid div by zero
    temp_angle = 90.0f - verticalAngle + maxElev;
    max_elev_dist = ESAR_SPATIAL_VecNorm(maxElevVertex_local);
    displacement_top = (-min_distance * sinf(-maxElev / ESAR_180_DIV_BY_PI))
        / (sinf(temp_angle / ESAR_180_DIV_BY_PI));

    temp_angle = fmax(0.1,
        90.0f - verticalAngle
            + minElev); // Limit the angle of the displacement to avoid div by zero
    temp_angle = 90.0f - verticalAngle + minElev;
    min_elev_dist = ESAR_SPATIAL_VecNorm(minElevVertex_local);
    displacement_bottom = (min_distance * sinf(minElev / ESAR_180_DIV_BY_PI))
        / (sinf(temp_angle / ESAR_180_DIV_BY_PI));

    // Convert vector back to world coordinates
    ESAR_SPATIAL_VecInit(observationSystemInverseFrontVec, observationSystemFrontVec[0],
        observationSystemUpVec[0], observationSystemRightVec[0]);
    ESAR_SPATIAL_VecInit(observationSystemInverseUpVec, observationSystemFrontVec[1],
        observationSystemUpVec[1], observationSystemRightVec[1]);
    ESAR_SPATIAL_VecInit(observationSystemInverseRightVec, observationSystemFrontVec[2],
        observationSystemUpVec[2], observationSystemRightVec[2]);
    horizontalDirectionVec_relative[0] =
        ESAR_SPATIAL_VecInnProd(horizontalDirectionVec_local, observationSystemInverseFrontVec);
    horizontalDirectionVec_relative[1] =
        ESAR_SPATIAL_VecInnProd(horizontalDirectionVec_local, observationSystemInverseUpVec);
    horizontalDirectionVec_relative[2] =
        ESAR_SPATIAL_VecInnProd(horizontalDirectionVec_local, observationSystemInverseRightVec);
    ESAR_SPATIAL_VecNormalize(horizontalDirectionVec_relative, horizontalDirectionVec_relative);
    verticalDirectionVec_relative[0] =
        ESAR_SPATIAL_VecInnProd(verticalDirectionVec_local, observationSystemInverseFrontVec);
    verticalDirectionVec_relative[1] =
        ESAR_SPATIAL_VecInnProd(verticalDirectionVec_local, observationSystemInverseUpVec);
    verticalDirectionVec_relative[2] =
        ESAR_SPATIAL_VecInnProd(verticalDirectionVec_local, observationSystemInverseRightVec);
    ESAR_SPATIAL_VecNormalize(verticalDirectionVec_relative, verticalDirectionVec_relative);
    // The left and right edges of the extent are then calculated using the left and right displacement along the
    // horizontal direction from the anchor point
    ESAR_SPATIAL_VecScale(horizontalDirectionVec_relative, displacement_left, displacementVecLeft);
    ESAR_SPATIAL_VecAdd(anchorPoint_relative, displacementVecLeft, leftEdgePos_relative);
    ESAR_SPATIAL_VecScale(
        horizontalDirectionVec_relative, displacement_right, displacementVecRight);
    ESAR_SPATIAL_VecAdd(anchorPoint_relative, displacementVecRight, rightEdgePos_relative);
    ESAR_SPATIAL_VecScale(verticalDirectionVec_relative, displacement_top, displacementVecTop);
    ESAR_SPATIAL_VecAdd(anchorPoint_relative, displacementVecTop, topEdgePos_relative);
    ESAR_SPATIAL_VecScale(
        verticalDirectionVec_relative, displacement_bottom, displacementVecBottom);
    ESAR_SPATIAL_VecAdd(anchorPoint_relative, displacementVecBottom, bottomEdgePos_relative);
    ESAR_SPATIAL_VecAdd(leftEdgePos_relative, listenerPos, leftEdgePos_world);
    ESAR_SPATIAL_VecAdd(rightEdgePos_relative, listenerPos, rightEdgePos_world);
    ESAR_SPATIAL_VecAdd(topEdgePos_relative, listenerPos, topEdgePos_world);
    ESAR_SPATIAL_VecAdd(bottomEdgePos_relative, listenerPos, bottomEdgePos_world);

    UpdateDescriptionPoint(stored_leftEdgePoint, leftEdgePos_world, deltaDistanceChange * 2.0f,
        temporalDescrPointSmoothFact, &stored_leftEdgePoint[0], &leftEdgePointInitialized);
    UpdateDescriptionPoint(stored_rightEdgePoint, rightEdgePos_world, deltaDistanceChange * 2.0f,
        temporalDescrPointSmoothFact, &stored_rightEdgePoint[0], &rightEdgePointInitialized);
    UpdateDescriptionPoint(stored_topEdgePoint, topEdgePos_world, deltaDistanceChange * 2.0f,
        temporalDescrPointSmoothFact, &stored_topEdgePoint[0], &topEdgePointInitialized);
    UpdateDescriptionPoint(stored_bottomEdgePoint, bottomEdgePos_world, deltaDistanceChange * 2.0f,
        temporalDescrPointSmoothFact, &stored_bottomEdgePoint[0], &bottomEdgePointInitialized);

    anchorPoint.set(stored_anchorPoint);
    leftEdgePoint.set(stored_leftEdgePoint);
    rightEdgePoint.set(stored_rightEdgePoint);
    topEdgePoint.set(stored_topEdgePoint);
    bottomEdgePoint.set(stored_bottomEdgePoint);

    // Update the distance to extent so that is based on the smoothed anchor point, keeping the sign.
    distanceToExtent =
        distanceToExtent / fabs(distanceToExtent) * (anchorPoint - listenerPos).abs();

    planeWidth = fabsf(displacement_left) + fabsf(displacement_right);
    planeHeight = fabsf(displacement_top) + fabsf(displacement_bottom);

    return;
}

void HeterogeneousSource::CalculateEffectiveExtent(EAR::Position listenerPos)
{
    // Perceptual scaling of the width and height of the extent.
    float widthScaler = 1.0f;
    float heightScaler = 1.0f;
    float displacementTop = 0.0;
    float displacementBottom = 0.0;
    float effExtLimit = 6.0;
    float geometricWidth = 0.0;
    float geometricHeight = 0.0;

    if (activeExtentRatio < 1.0) {
        // Occlusion has modified the extent, don't use the anchorPoint since it
        // may be outside the new extent.
        geometricWidth = (rightEdgePoint - leftEdgePoint).abs();
        geometricHeight = (topEdgePoint - bottomEdgePoint).abs();
        if ((geometricWidth > 0) && (distanceToExtent > 0)
            && (distanceToExtent < geometricWidth / 6.0f)) {
            // At this distance the audio element can be seen as a line source
            widthScaler = 6.0f * distanceToExtent / geometricWidth;
            Vector3 newRightEdgePoint = 0.5f * (rightEdgePoint + leftEdgePoint)
                + widthScaler * 0.5f * (rightEdgePoint - leftEdgePoint);
            Vector3 newLeftEdgePoint = 0.5f * (rightEdgePoint + leftEdgePoint)
                - widthScaler * 0.5f * (rightEdgePoint - leftEdgePoint);
            rightEdgePoint = newRightEdgePoint;
            leftEdgePoint = newLeftEdgePoint;
        }

        displacementTop = (topEdgePoint - bottomEdgePoint).abs() / 2.0f;
        displacementBottom = (topEdgePoint - bottomEdgePoint).abs() / 2.0f;
        if ((displacementTop > 0) && (distanceToExtent > 0)
            && (distanceToExtent < displacementTop / 3.0f)) {
            heightScaler = 3.0f * distanceToExtent / displacementTop;
            Vector3 newtopEdgePoint = 0.5f * (bottomEdgePoint + topEdgePoint)
                + heightScaler * 0.5f * (topEdgePoint - bottomEdgePoint);
            topEdgePoint = newtopEdgePoint;
        }
        if ((displacementBottom > 0) && (distanceToExtent > 0)
            && (distanceToExtent < displacementBottom / 3.0f)) {
            heightScaler = 3.0f * distanceToExtent / displacementBottom;
            Vector3 newBottomEdgePoint = 0.5f * (bottomEdgePoint + topEdgePoint)
                - heightScaler * 0.5f * (topEdgePoint - bottomEdgePoint);
            bottomEdgePoint = newBottomEdgePoint;
        }
    } else {
        // Use the anchor point as the basis
        Vector3 horizontalDisplacementVec = (rightEdgePoint - leftEdgePoint).normalize();
        float displacementLeft = (leftEdgePoint - anchorPoint).abs();
        float displacementRight = (rightEdgePoint - anchorPoint).abs();
        geometricWidth = displacementLeft + displacementRight;
        if ((geometricWidth > 0) && (distanceToExtent > 0)
            && (distanceToExtent < geometricWidth / 6.0f)) {
            // At this distance the audio element can be seen as a line source
            widthScaler = 6.0f * distanceToExtent / geometricWidth;
            rightEdgePoint =
                anchorPoint + widthScaler * displacementRight * horizontalDisplacementVec;
            leftEdgePoint =
                anchorPoint - widthScaler * displacementLeft * horizontalDisplacementVec;
        }

        Vector3 verticalDisplacementVec = (topEdgePoint - anchorPoint).normalize();
        displacementTop = (topEdgePoint - anchorPoint).abs();
        displacementBottom = (bottomEdgePoint - anchorPoint).abs();
        if ((displacementTop > 0.0) && (distanceToExtent < displacementTop / 3.0)) {
            heightScaler = (float)3.0 * distanceToExtent / displacementTop;
            displacementTop = heightScaler * displacementTop;
            topEdgePoint = anchorPoint + displacementTop * verticalDisplacementVec;
        }
        if ((displacementBottom > 0.0) && (distanceToExtent < displacementBottom / 3.0)) {
            heightScaler = (float)3.0 * distanceToExtent / displacementBottom;
            displacementBottom = heightScaler * displacementBottom;
            bottomEdgePoint = anchorPoint - displacementBottom * verticalDisplacementVec;
        }
    }
}

void HeterogeneousSource::UpdateExteriorRepresentationSpeakerSetup(EAR::Position listenerPos)
{
    float ptLineAtt, ptLineAmp, lin2DimAtt, lin2DimAmp, gainCompPt, gainCompLine, gainComp2D;
    int numSpPt, numSpLine, numSp2D;
    Vector3 topLeftPoint, topRightPoint, bottomLeftPoint, bottomRightPoint;
    Vector3 leftListenerDelta, rightListenerDelta, midSpeakerPos;
    float leftToListenerDist, rightToListenerDist;
    float elev_ang, azim_ang, ang_st, ang_end;
    float midSpeakerGain = 0.5773;
    float sideSpeakerGain = 1.0;
    bool useGerzonUpmix = false;

    if (useGerzonUpmix) {
        midSpeakerGain = 0.5773;
        sideSpeakerGain = 0.78864;
    }

    //
    // Update the speaker positions
    //

    // Calculate the corner positions
    if (activeExtentRatio < 1.0) {
        // Occlusion has modified the extent, don't use the anchorPoint since it
        // may be outside the new extent.
        Vector3 vertDisplacementVecTop = 0.5 * (topEdgePoint - bottomEdgePoint);
        topLeftPoint = leftEdgePoint + vertDisplacementVecTop;
        topRightPoint = rightEdgePoint + vertDisplacementVecTop;
        bottomLeftPoint = leftEdgePoint - vertDisplacementVecTop;
        bottomRightPoint = rightEdgePoint - vertDisplacementVecTop;
    } else {
        // Use the anchor point as the basis
        Vector3 smoothedDisplacementVecTop = (topEdgePoint - anchorPoint);
        Vector3 smoothedDisplacementVecBottom = (bottomEdgePoint - anchorPoint);
        topLeftPoint = leftEdgePoint + smoothedDisplacementVecTop;
        topRightPoint = rightEdgePoint + smoothedDisplacementVecTop;
        bottomLeftPoint = leftEdgePoint + smoothedDisplacementVecBottom;
        bottomRightPoint = rightEdgePoint + smoothedDisplacementVecBottom;
    }

    // Update transition factors for adaptation of the number of speakers used
    ang_st = ESAR_PI / 32; // The angle where transition starts
    ang_end = ESAR_PI / 12; // The angle where the transition completes
    azim_ang = (abs(minAzim) + abs(maxAzim)) / 2;
    azim_ang = fmin(azim_ang, 90);
    azim_ang = azim_ang * ESAR_PI_DIV_BY_180;
    elev_ang = (abs(minElev) + abs(maxElev)) / 2;
    elev_ang = fmin(elev_ang, 90);
    elev_ang = elev_ang * ESAR_PI_DIV_BY_180;
    trans_fac[0] = (sinf(azim_ang) - sinf(ang_st)) / (sinf(ang_end) - sinf(ang_st));
    trans_fac[1] = (sinf(elev_ang) - sinf(ang_st)) / (sinf(ang_end) - sinf(ang_st));
    trans_fac[0] = fmin(fmax(trans_fac[0], 0), 1);
    trans_fac[1] = fmin(fmax(trans_fac[1], 0), 1);

    //transition from 1D to 2D
    if (trans_fac[0] > 0 && trans_fac[1] > 0) {
        exteriorSpeakerPosVec[0].location =
            trans_fac[1] * bottomLeftPoint + (1 - trans_fac[1]) * leftEdgePoint;
        exteriorSpeakerPosVec[1].location =
            trans_fac[1] * bottomRightPoint + (1 - trans_fac[1]) * rightEdgePoint;
    } else {
        exteriorSpeakerPosVec[0].location = leftEdgePoint;
        exteriorSpeakerPosVec[1].location = rightEdgePoint;
    }

    exteriorSpeakerPosVec[3].location = topLeftPoint;
    exteriorSpeakerPosVec[4].location = topRightPoint;

    // Adaptive placement of the middle speaker between the left and right edge points
    leftToListenerDist = (leftEdgePoint - listenerPos.location).abs();
    rightToListenerDist = (rightEdgePoint - listenerPos.location).abs();
    leftListenerDelta = leftEdgePoint * rightToListenerDist;
    rightListenerDelta = rightEdgePoint * leftToListenerDist;
    if ((leftToListenerDist + rightToListenerDist) > 0.0) {
        midSpeakerPos = (leftEdgePoint * rightToListenerDist + rightEdgePoint * leftToListenerDist)
            * (1.0f / (leftToListenerDist + rightToListenerDist));
    } else {
        midSpeakerPos = 0.5 * (leftEdgePoint + rightEdgePoint);
    }

    exteriorSpeakerPosVec[2].location = midSpeakerPos;

    //
    // Update gain for each speaker
    //
    numSpPt = 1; // Point-source representation
    numSpLine = 3; // Line-source representation
    numSp2D = 5; // Plane representation
    gainCompPt = sqrt((float)1.0 / numSpPt);
    gainCompLine = sqrt((float)1.0 / numSpLine);
    gainComp2D = sqrt((float)1.0 / numSp2D);

    ptLineAtt = gainCompPt - (gainCompPt - gainCompLine) * trans_fac[0];
    ptLineAmp = gainCompLine * trans_fac[0];

    if (trans_fac[0] > 0 && trans_fac[1] > 0) {
        lin2DimAtt = gainCompLine - (gainCompLine - gainComp2D) * trans_fac[1];
        lin2DimAmp = gainComp2D * trans_fac[1];
    } else {
        lin2DimAtt = gainCompLine;
        lin2DimAmp = 0.0;
    }

    exteriorSpeakerGains[0] = ptLineAmp * lin2DimAtt * sideSpeakerGain; // Left/Bottom-left
    exteriorSpeakerGains[1] = ptLineAmp * lin2DimAtt * sideSpeakerGain; // Right/Bottom-right
    exteriorSpeakerGains[2] = ptLineAtt * lin2DimAtt * midSpeakerGain; // Middle
    exteriorSpeakerGains[3] = lin2DimAmp * sideSpeakerGain; // Top-left
    exteriorSpeakerGains[4] = lin2DimAmp * sideSpeakerGain; // Top-right

    // Overall gain correction
    if (distanceToExtent < 2.0 * fmax(planeWidth, planeHeight)) {
        float distanceGainCorrection = CalculateDistanceGainCorrection(listenerPos);

        for (int i = 0; i < numOfExteriorSpeakers; i++) {
            exteriorSpeakerGains[i] = exteriorSpeakerGains[i] * distanceGainCorrection;
            if (exteriorSpeakerGains[i] < 1.0e-5) {
                exteriorSpeakerGains[i] =
                    0.0; // Set to zero for really low levels to bypass processing
            }
        }
    }
}

float HeterogeneousSource::CalculateDistanceGainCorrection(EAR::Position listenerPos)
{
    // Given the computed distance gains of the individual speakers, calculate a gain correction
    // in order to achieve the target overal distance gain of the audio element
    float distanceGainCorrection = 1.0;

    float distModelWidth = planeWidth;
    float distModelHeight = planeHeight;
    if (distanceGainModel == 1) {
        // Use point-source distance gain model
        distModelWidth = 0.0;
        distModelHeight = 0.0;
    }

    if (distanceToExtent > 0.0) {
        float targetDistGain = distanceModel->getGainForDistance(
            distanceToExtent, distModelWidth, distModelHeight, refDistance);

        // Never exceed the target gain of the reference distance
        targetDistGain = fmin(targetDistGain, 1.0);

        float gains_norm = 1.0;
        float max_gain =
            *std::max_element(exteriorSpeakerGains.begin(), exteriorSpeakerGains.end());
        if (max_gain > 0.0) {
            gains_norm = 1.0 / max_gain;

            float sumOfSquaredDistGains = 0.0;
            float sumOfSquaredNormGains = 0.0;
            for (int i = 0; i < numOfExteriorSpeakers; i++) {
                float speakerDistGain = gains_norm
                    * distanceModel->getGainForDistance(
                        (exteriorSpeakerPosVec[i].location - listenerPos.location).abs(),
                        distModelWidth, distModelHeight, refDistance);

                // Calculate the sum of squared distance gains of all speakers
                sumOfSquaredDistGains += exteriorSpeakerGains[i] * exteriorSpeakerGains[i]
                    * speakerDistGain * speakerDistGain;

                // Calculate the sum of the squared normalized speaker gains
                sumOfSquaredNormGains +=
                    exteriorSpeakerGains[i] * exteriorSpeakerGains[i] * gains_norm * gains_norm;
            }
            if ((sumOfSquaredDistGains > 0.0) && (sumOfSquaredNormGains > 0.0)) {
                float denominator = sqrtf(sumOfSquaredDistGains);
                float normalizer = sqrtf(sumOfSquaredNormGains);
                distanceGainCorrection = normalizer * (targetDistGain / denominator);
            }
        }
    }

    return distanceGainCorrection;
}

void HeterogeneousSource::UpdateInteriorRepresentationSpeakerSetup(
    const EAR::Geometry* extent, EAR::Position listenerPos)
{
    EAR::Vector3 localSpeakerLocation;
    float shortestSquaredDist = 100000000000000.0;
    float refJawAngle = -listenerPos.orientation[0];

    //
    // Set the speaker positions of the interior rendering setup
    //
    int closestNormalIndex = ESAR_SPATIAL_FindClosestOutlineNormalIndex_LineSearch(
        extent->getMesh(), outlineNormals, listenerPos.location, &shortestSquaredDist);
    distanceToExtentOutline = sqrtf(shortestSquaredDist);
    Polar polarNormalVector = cartesianToPolar(-outlineNormals[closestNormalIndex].normal);

    intSpeakerHorizRotation =
        ESAR_SPATIAL_Smooth_angle(intSpeakerHorizRotation, polarNormalVector[0], 0.99);

    for (int spInd = 0; spInd < numOfInteriorSpeakers; spInd++) {
        localSpeakerLocation.set(refDistance
                * -sin(ESAR_PI_DIV_BY_180 * (intSpeakerHorizRotation - intSpAz[spInd]))
                * cos(ESAR_PI_DIV_BY_180 * intSpEl[spInd]),
            refDistance * sin(ESAR_PI_DIV_BY_180 * intSpEl[spInd]),
            refDistance * -cos(ESAR_PI_DIV_BY_180 * (intSpeakerHorizRotation - intSpAz[spInd]))
                * cos(ESAR_PI_DIV_BY_180 * intSpEl[spInd]));
        interiorSpeakerPosVec[spInd].location = listenerPos.location + localSpeakerLocation;
    }

    // Set the gain of the interior representation speakers
    float intGainComp = sqrt(1.0 / numOfInteriorSpeakers);
    float rearGain = rearHemisphereWeight * rearHemisphereWeight;
    float topGain = topHemisphereWeight * topHemisphereWeight;
    float bottomGain = bottomHemisphereWeight * bottomHemisphereWeight;
    interiorSpeakerGains[0] = intGainComp; // Left
    interiorSpeakerGains[1] = intGainComp; // Right
    interiorSpeakerGains[2] = intGainComp; // Front
    interiorSpeakerGains[3] = intGainComp * rearGain; // Rear
    interiorSpeakerGains[4] = intGainComp * topGain; // Left-center-high
    interiorSpeakerGains[5] = intGainComp * bottomGain; // Left-center-low
    interiorSpeakerGains[6] = intGainComp * topGain; // Right-center-high
    interiorSpeakerGains[7] = intGainComp * bottomGain; // Right-center-low
    interiorSpeakerGains[8] = intGainComp * rearGain * topGain; // Left-back-high
    interiorSpeakerGains[9] = intGainComp * rearGain * bottomGain; // Left-back-low
    interiorSpeakerGains[10] = intGainComp * rearGain * topGain; // Right-back-high
    interiorSpeakerGains[11] = intGainComp * rearGain * bottomGain; // Right-back-low
}

void HeterogeneousSource::CloseToSurfaceModificationsOfInteriorRepresentationRendering(
    const EAR::Geometry* extent, EAR::Position listenerPos)
{
    //
    // Modify the interior rendering for listening positions close to the surface of the extent
    //
    bool isInsideOutline = ESAR_SPATIAL_IsInsideOutline(outlineNormals, listenerPos.location);

    float distToOutline =
        ESAR_SPATIAL_FindDistanceToOutline(extent->getMesh(), outlineNormals, listenerPos.location);

    //
    // Update gain of the rear hemisphere of the interior representation. This is controlled
    // by the distance to the horizontal outline of the extent.
    //
    float interiorTransisionDistance = transitionDistance;
    float newTargetRearHemisphereWeight = 0.0;
    if (!isInsideOutline) {
        // The user is outside the extent outline
        newTargetRearHemisphereWeight = 0.0;
    } else if (distToOutline >= interiorTransisionDistance) {
        // The user is further inside the extent outline, beyond the internal fade region
        newTargetRearHemisphereWeight = 1.0;
    } else {
        // The user is within the internal fade region
        newTargetRearHemisphereWeight =
            1.0 - (interiorTransisionDistance - distToOutline) / interiorTransisionDistance;
    }
    newTargetRearHemisphereWeight = fmax(0.5, newTargetRearHemisphereWeight);
    rearHemisphereWeight = 0.98 * rearHemisphereWeight + 0.02 * newTargetRearHemisphereWeight;

    // Avoid denormal numbers
    if (rearHemisphereWeight < 1e-20) {
        rearHemisphereWeight = 0.0;
    }

    //
    // Update gain of the top hemisphere of the interior representation.
    //

    // First find the topmost edgepoint of the simplified extent. It is not always the topEdgePoint.
    Vector3 topmostPoint = topEdgePoint;
    if (leftEdgePoint[1] > topmostPoint[1]) {
        topmostPoint = leftEdgePoint;
    }
    if (rightEdgePoint[1] > topmostPoint[1]) {
        topmostPoint = rightEdgePoint;
    }
    if (bottomEdgePoint[1] > topmostPoint[1]) {
        topmostPoint = bottomEdgePoint;
    }

    // Calculate height position of the topmost point relative to the listening position
    float topTransitionDistance = transitionDistance / 6.0;
    Vector3 vectorToTopPoint = topmostPoint - listenerPos.location;
    float topHeightDiff = vectorToTopPoint[1];
    float targetTopHemisphereWeight = 1.0;
    if (distanceToExtent > 0.0) {
        if (topHeightDiff > 0.0) {
            // The user is below the top of the extent
            targetTopHemisphereWeight = 1.0;
        } else if (topHeightDiff <= -topTransitionDistance) {
            // The user is above the transition region
            targetTopHemisphereWeight = 0.0;
        } else {
            // The user is outside and within the fading height
            targetTopHemisphereWeight = 1.0 + topHeightDiff / topTransitionDistance;
        }
    }
    targetTopHemisphereWeight = fmax(0.5, targetTopHemisphereWeight);
    topHemisphereWeight = 0.98 * topHemisphereWeight + 0.02 * targetTopHemisphereWeight;

    // Avoid denormal numbers
    if (topHemisphereWeight < 1e-20) {
        topHemisphereWeight = 0.0;
    }

    //
    // Update gain of the bottom hemisphere of the interior representation.
    //

    // First find the edgepoint furthest down of the simplified extent. It is not always the bottomEdgePoint.
    Vector3 bottomPoint = bottomEdgePoint;
    if (leftEdgePoint[1] < bottomPoint[1]) {
        bottomPoint = leftEdgePoint;
    }
    if (rightEdgePoint[1] < bottomPoint[1]) {
        bottomPoint = rightEdgePoint;
    }
    if (topEdgePoint[1] < bottomPoint[1]) {
        bottomPoint = topEdgePoint;
    }

    // Calculate height position of the bottom point relative to the listening position
    float bottomTransitionDistance = transitionDistance / 6.0;
    Vector3 vectorToBottomPoint = bottomPoint - listenerPos.location;
    float bottomHeightDiff = vectorToBottomPoint[1];
    float targetBottomHemisphereWeight = 1.0;
    if (distanceToExtent > 0.0) {
        if (bottomHeightDiff < 0.0) {
            // The user is above the bottom of the extent
            targetBottomHemisphereWeight = 1.0;
        } else if (bottomHeightDiff >= bottomTransitionDistance) {
            // The user is below the transition region
            targetBottomHemisphereWeight = 0.0;
        } else {
            // The user is outside and within the fading height
            targetBottomHemisphereWeight = 1.0 - bottomHeightDiff / bottomTransitionDistance;
        }
    }
    targetBottomHemisphereWeight = fmax(0.5, targetBottomHemisphereWeight);
    bottomHemisphereWeight = 0.98 * bottomHemisphereWeight + 0.02 * targetBottomHemisphereWeight;

    // Avoid denormal numbers
    if (bottomHemisphereWeight < 1e-20) {
        bottomHemisphereWeight = 0.0;
    }
}

void HeterogeneousSource::UpdateOutputSpeakerActivation()
{
    // Deactivate speakers that are silent

    // Speakers used for both exterior and interior representation
    if ((exteriorSpeakerGains[0] < ESAR_GAIN_THRESHOLD)
        && (interiorSpeakerGains[0] < ESAR_GAIN_THRESHOLD)) {
        outputSpeakerActive[0] = false;
    } else {
        outputSpeakerActive[0] = true;
    }
    if ((exteriorSpeakerGains[1] < ESAR_GAIN_THRESHOLD)
        && (interiorSpeakerGains[1] < ESAR_GAIN_THRESHOLD)) {
        outputSpeakerActive[1] = false;
    } else {
        outputSpeakerActive[1] = true;
    }
    if ((exteriorSpeakerGains[2] < ESAR_GAIN_THRESHOLD)
        && (interiorSpeakerGains[2] < ESAR_GAIN_THRESHOLD)) {
        outputSpeakerActive[2] = false;
    } else {
        outputSpeakerActive[2] = true;
    }

    // Speakers used only for exterior representation
    if (interiorExteriorWeight > 1.0 - ESAR_GAIN_THRESHOLD) {
        outputSpeakerActive[4] = false;
        outputSpeakerActive[5] = false;
    } else {
        outputSpeakerActive[4] = true;
        outputSpeakerActive[5] = true;
    }

    // Speakers used only for interior representation
    if (interiorExteriorWeight < ESAR_GAIN_THRESHOLD) {
        outputSpeakerActive[3] = false;
    } else {
        outputSpeakerActive[3] = true;
    }

    for (int spInd = 6; spInd < HETEROGENEOUS_NUM_OF_OUTPUT_SPEAKERS; spInd++) {
        if (interiorExteriorWeight < ESAR_GAIN_THRESHOLD) {
            outputSpeakerActive[spInd] = false;
        } else {
            outputSpeakerActive[spInd] = true;
        }
    }
}

void HeterogeneousSource::UpdateOutputSpeakerDistanceOffset(Vector3 listenerLocation)
{
    double speakerDist;
    // Setting a minimum distance can be used for matching the delay of another HOA rendering stage
    double internalDist = 0.0;

    // Use the distanceToExtent as the reference distance for the whole extent
    for (int index = 0; index < outputSpeakerPosVec.size(); index++) {
        speakerDist = (outputSpeakerPosVec[index].location - listenerLocation).abs();
        outputSpeakerDistanceOffsets[index] = fmax(distanceToExtent, internalDist) - speakerDist;
    }
}

void HeterogeneousSource::UpdateDescriptionPoint(float previousPos[3], float newPos[3],
    float maxPositionMovement, float temporalSmoothingFactor, float updatedPos[3],
    bool* pInitializedFlag)
{

    float positionMovement, tempVec[3], tempVec2[3];
    if (*pInitializedFlag) {
        // First do proportional interpolation based on the change of relative position

        // Calculate the direction of the movement
        ESAR_SPATIAL_VecSub(newPos, previousPos, tempVec);
        positionMovement = ESAR_SPATIAL_VecNorm(tempVec);
        if (positionMovement <= 1.0 * maxPositionMovement) {
            // The movement does not need to be restricted
            ESAR_SPATIAL_VecCopy(updatedPos, newPos);
        } else {
            // Move only the maxmimum distance in the right direction
            ESAR_SPATIAL_VecNormalize(tempVec, tempVec);
            ESAR_SPATIAL_VecScale(tempVec, maxPositionMovement, tempVec);
            ESAR_SPATIAL_VecCopy(tempVec2, previousPos);
            ESAR_SPATIAL_VecAdd(previousPos, tempVec, updatedPos);
        }

        // Then do temporal smoothing according to the provided smoothing factor
        ESAR_SPATIAL_VecScale(updatedPos, temporalSmoothingFactor, tempVec);
        ESAR_SPATIAL_VecScale(newPos, (1.0 - temporalSmoothingFactor), tempVec2);
        ESAR_SPATIAL_VecAdd(tempVec, tempVec2, updatedPos);

    } else {
        // This position has never been set a real value so skip interpolation
        ESAR_SPATIAL_VecCopy(updatedPos, newPos);
        *pInitializedFlag = true;
    }

    return;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// HeterogeneousDspState class definitions
//
//////////////////////////////////////////////////////////////////////////////////////////////////////

void HeterogeneousDspState::UpdateState(MPEGI::ExtentLayout* extLayout, float intExteriorWeight,
    float relHorAngleExt, float relVerAngleExt, float relHorAngleInt, float relVerAngleInt,
    float rearHemiGain, float topHemiGain, float bottomHemiGain, float trans_factor[],
    std::vector<float> extSpeakerGains, std::vector<float> intSpeakerGains, float occlLeftAngleSc,
    float occlRightAngleSc, float occlTopAngleSc, float occlBottomAngleSc)

{
    //extentLayout
    extentLayout.alignment = extLayout->alignment;
    extentLayout.positioning = extLayout->positioning;
    interiorExteriorWeight = intExteriorWeight;
    relativeHorizAngleExterior = relHorAngleExt;
    relativeVertAngleExterior = relVerAngleExt;
    relativeHorizAngleInterior = relHorAngleInt;
    relativeVertAngleInterior = relVerAngleInt;
    rearHemisphereGain = rearHemiGain;
    topHemisphereGain = topHemiGain;
    bottomHemisphereGain = bottomHemiGain;
    trans_fac[0] = trans_factor[0];
    trans_fac[1] = trans_factor[1];
    trans_fac[2] = trans_factor[2];
    exteriorSpeakerGains = extSpeakerGains;
    interiorSpeakerGains = intSpeakerGains;
    occlLeftAngleScaler = occlLeftAngleSc;
    occlRightAngleScaler = occlRightAngleSc;
    occlTopAngleScaler = occlTopAngleSc;
    occlBottomAngleScaler = occlBottomAngleSc;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// InteriorRepresentation class definitions
//
//////////////////////////////////////////////////////////////////////////////////////////////////////
InteriorRepresentation::InteriorRepresentation(InteriorRepresentationType intRepType,
    const MPEGI::ExtentLayout* extentLayout, std::size_t B, int inputChannels, int delay)
    : type(intRepType)
    , frameLength(B)
    , numOfLayers(1)
    , delay_in_samples(delay)
{
    int labelChanInd[HETEROGENEOUS_MAX_NO_INPUT_CHANNELS];
    bool topLayerRepresented = false;
    bool middleLayerRepresented = false;
    bool bottomLayerRepresented = false;
    int inputLayerInd = 0;

    for (int i = 0; i < HETEROGENEOUS_MAX_NO_INPUT_CHANNELS; i++) {
        labelChanInd[i] = -1; // -1 means that the label is not represented by any input channels
    }

    for (int i = 0; i < HETEROGENEOUS_INTERIOR_LAYER_CHANNELS * 3; i++) {
        layerSignalInputChanInd[i] = -1;
    }

    if (type == InteriorRepresentationType::LayeredQuadrasonic) {

        // Check what layers are represented in the list of labels
        if ((extentLayout == NULL) || (extentLayout->positioning.size() == 0)) {
            // No labels where specified, use default channel mapping
            switch (inputChannels) {
            case 1:
                numOfLayers = 1;
                labelChanInd[4] = 0;
                break;
            case 2:
                numOfLayers = 1;
                labelChanInd[3] = 0;
                labelChanInd[5] = 1;
                break;
            case 3:
                numOfLayers = 1;
                labelChanInd[3] = 0;
                labelChanInd[4] = 1;
                labelChanInd[5] = 2;
                break;
            case 4:
                numOfLayers = 2;
                labelChanInd[0] = 0;
                labelChanInd[2] = 1;
                labelChanInd[6] = 2;
                labelChanInd[8] = 3;
                break;
            case 5:
                numOfLayers = 3;
                labelChanInd[0] = 0;
                labelChanInd[2] = 1;
                labelChanInd[4] = 2;
                labelChanInd[6] = 3;
                labelChanInd[8] = 4;
                break;
            case 6:
                numOfLayers = 2;
                labelChanInd[0] = 0;
                labelChanInd[1] = 1;
                labelChanInd[2] = 2;
                labelChanInd[5] = 3;
                labelChanInd[6] = 4;
                labelChanInd[7] = 5;
                break;
            case 7:
                numOfLayers = 3;
                labelChanInd[0] = 0;
                labelChanInd[1] = 1;
                labelChanInd[2] = 2;
                labelChanInd[4] = 3;
                labelChanInd[5] = 4;
                labelChanInd[6] = 5;
                labelChanInd[7] = 6;
                break;
            case 8:
                numOfLayers = 3;
                labelChanInd[0] = 0;
                labelChanInd[1] = 1;
                labelChanInd[2] = 2;
                labelChanInd[3] = 3;
                labelChanInd[5] = 4;
                labelChanInd[6] = 5;
                labelChanInd[7] = 6;
                labelChanInd[8] = 7;
                break;
            case 9:
                numOfLayers = 3;
                labelChanInd[0] = 0;
                labelChanInd[1] = 1;
                labelChanInd[2] = 2;
                labelChanInd[3] = 3;
                labelChanInd[4] = 4;
                labelChanInd[5] = 5;
                labelChanInd[6] = 6;
                labelChanInd[7] = 7;
                labelChanInd[8] = 8;
                break;
            }
        } else {
            for (int i = 0; i < extentLayout->positioning.size(); i++) {
                MPEGI::ExtentLayout::Positioning label = extentLayout->positioning[i];
                labelChanInd[(int)label] = i;

                if ((label == MPEGI::ExtentLayout::Positioning::TL)
                    || (label == MPEGI::ExtentLayout::Positioning::T)
                    || (label == MPEGI::ExtentLayout::Positioning::TR)) {
                    topLayerRepresented = true;
                }
                if ((label == MPEGI::ExtentLayout::Positioning::L)
                    || (label == MPEGI::ExtentLayout::Positioning::C)
                    || (label == MPEGI::ExtentLayout::Positioning::R)) {
                    middleLayerRepresented = true;
                }
                if ((label == MPEGI::ExtentLayout::Positioning::BL)
                    || (label == MPEGI::ExtentLayout::Positioning::B)
                    || (label == MPEGI::ExtentLayout::Positioning::BR)) {
                    bottomLayerRepresented = true;
                }
            }

            // Choose the number of layers for the interior representation
            if (!topLayerRepresented && middleLayerRepresented && !bottomLayerRepresented) {
                numOfLayers = 1;
            } else if (topLayerRepresented && !middleLayerRepresented && bottomLayerRepresented) {
                numOfLayers = 2;
            } else {
                numOfLayers = 3;
            }
        }

        // Go through the active layers and set the input channel that each signal of the layer
        // should be taken from. Signals that were not explicitely represented in the
        // input layout positioning list need to reuse some other channel that corresponds
        // to a signal from the same elevation layer
        for (int i = 0; i < numOfLayers; i++) {

            if ((numOfLayers == 2) && (i == 1)) {
                // In the case that the interior representation has only two layers we
                // skip the middle layer of the input layout
                inputLayerInd = 2;
            } else if ((numOfLayers == 1) && (i == 0)) {
                // In the case that the interior representation has only one layer we
                // only use the middle layer of the input layout
                inputLayerInd = 1;
            } else {
                inputLayerInd = i;
            }

            // Front signal
            // If it was specified in the input layout, use that input channel, otherwise
            // it will be generated as a mix of the left and right
            if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 1]
                != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 1];
            }

            // Left signal
            // If not specified, use front or right
            if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER] != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS + 1] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER];
            } else if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 1]
                != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS + 1] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 1];
            } else if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 2]
                != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS + 1] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 2];
            }

            // Back signal is always generated using decorrelation

            // Right signal
            // If not specified, use front or left
            if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 2]
                != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS + 3] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 2];
            } else if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 1]
                != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS + 3] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER + 1];
            } else if (labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER]
                != -1) {
                layerSignalInputChanInd[i * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS + 3] =
                    labelChanInd[inputLayerInd * HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER];
            }
        }
    }

    corrEst = 0.5;

    // Allocate decorrelators and buffers for delay
    for (int i = 0; i < numOfLayers; i++) {
        layers.push_back(new EAR::RealFrame(
            HETEROGENEOUS_INTERIOR_LAYER_CHANNELS, frameLength + delay_in_samples));
        decorrelators.push_back(new EAR::MpegiDecorrelator(frameLength));
        decorrelators_front.push_back(new EAR::MpegiDecorrelator(frameLength));
    }

    delayBuffer = new EAR::RealFrame(HETEROGENEOUS_INTERIOR_LAYER_CHANNELS, delay_in_samples);

    decorrProcBuffer = std::make_unique<EAR::RealFrame>(2, frameLength);
    decorrProcBuffer_front = std::make_unique<EAR::RealFrame>(2, frameLength);
    esdConverter.initialize(frameLength);
}

void InteriorRepresentation::DeriveInteriorRepresentationFrame(
    EAR::RealFrame* inputSignal, MPEGI::ExtentLayout extentLayout)
{
    int layerIndex;

    if (type == InteriorRepresentationType::LayeredQuadrasonic) {

        // Channel order of each layer in the interior representation:
        // 0: Front, 1: Left, 2: Back, 3: Right
        for (short layer = 0; layer < numOfLayers; layer++) {

            layerIndex = layer * HETEROGENEOUS_INTERIOR_LAYER_CHANNELS;

            // Left channel and right channels
            if (layerSignalInputChanInd[layerIndex + 3]
                != layerSignalInputChanInd[layerIndex + 1]) {
                // Right channel
                layers[layer]->ch(1).copyFrom(
                    inputSignal->ch(layerSignalInputChanInd[layerIndex + 1]));
                layers[layer]->ch(3).copyFrom(
                    inputSignal->ch(layerSignalInputChanInd[layerIndex + 3]));
            } else {
                // The left & right channels are pointing to the same signal. In this case,
                // generate the left & right with decorrelation
                auto tmpPtr = decorrProcBuffer_front->data();
                decorrelators_front[layer]->process(
                    inputSignal->ch(layerSignalInputChanInd[layerIndex + 1]).data(), tmpPtr.get());
                layers[layer]->ch(1).copyFrom(decorrProcBuffer_front->ch(0));
                layers[layer]->ch(3).copyFrom(decorrProcBuffer_front->ch(1));

                // In this case don't use the specified front channel, use the mix
                // of the newly created left & right signals instead.
                layerSignalInputChanInd[layerIndex] = -1;
            }

            // Front signal, use provided input channel or mix left and right
            if (layerSignalInputChanInd[layerIndex] != -1) {
                layers[layer]->ch(0).copyFrom(inputSignal->ch(layerSignalInputChanInd[layerIndex]));
            } else {
                // Use mix of left and right signals
                layers[layer]->ch(0).copyFrom(layers[layer]->ch(1));
                layers[layer]->ch(0).add(layers[layer]->ch(3));
                layers[layer]->ch(0).mul(0.5f);
            }

            // Create the back signal by decorrelation of the front signal
            auto tmpPtr2 = decorrProcBuffer->data();
            decorrelators[layer]->process(layers[layer]->ch(0).data(), tmpPtr2.get());
            layers[layer]->ch(2).copyFrom(decorrProcBuffer->ch(0));
        }
    } else if (type == InteriorRepresentationType::FOA) {

        // Convert ESD-format to B-format
        for (auto i{ inputSignal->channels() }; i-- > 0;) {
            esdConverter.copyToInput(inputSignal->ch(i).data(), frameLength, i);
        }

        // We only need the FOA part of the signal
        auto const wasConversionOk = esdConverter.convertEsdToHoa(1);
        EAR_CHECK_REL(wasConversionOk, "ESD conversion failed");

        for (auto i{ 4 }; i-- > 0;) {
            std::copy_n(esdConverter.getOutputData(i), frameLength, layers[0]->ch(i).data());
        }
    }

    // Delay buffers
    for (short layer = 0; layer < numOfLayers; layer++) {
        for (short channel = 0; channel < 4; channel++) {
            // Store tail of samples from last frame
            delayBuffer->ch(channel).copyFrom(
                layers[layer]->ch(channel), frameLength, 0, delay_in_samples);

            // Move the frame forward
            layers[layer]->ch(channel).copyFrom(
                layers[layer]->ch(channel), 0, delay_in_samples, frameLength);

            // Place the tail at the start of the buffer
            layers[layer]->ch(channel).copyFrom(delayBuffer->ch(channel), 0, 0, delay_in_samples);
        }
    }
}

void InteriorRepresentation::CaptureFrameFromDirection(float horizontalAngle, float elevationAngle,
    EAR::RealBuffer* capturedFrame, float micPattern, float gain)
{
    float value;
    float layerGain[3];

    // Skip processing if the gain is below the threashold
    if (gain < 1.0e-5) {
        return;
    }

    if ((type == InteriorRepresentationType::LayeredQuadrasonic) && (numOfLayers == 1)) {
        // 0: Front, 1: Left, 2: Back, 3: Right
        float frontSigWeight = fmax(0.0f, cosf(horizontalAngle));
        float leftSigWeight = fmax(0.0f, cosf(horizontalAngle - ESAR_PI / 2.0f));
        float backSigWeight = fmax(0.0f, cosf(horizontalAngle - ESAR_PI));
        float rightSigWeight = fmax(0.0f, cosf(horizontalAngle - 3.0f * ESAR_PI / 2.0f));

        for (int i = 0; i < frameLength; i++) {
            value = gain
                * (frontSigWeight * layers[0]->ch(0)[i] + leftSigWeight * layers[0]->ch(3)[i]
                    + backSigWeight * layers[0]->ch(2)[i] + rightSigWeight * layers[0]->ch(1)[i]);
            capturedFrame->add(value, i, 1);
        }
    } else if (type == InteriorRepresentationType::LayeredQuadrasonic) {
        layerGain[0] = 0.0;
        layerGain[1] = 0.0;
        layerGain[2] = 0.0;
        float micElevOverlap = ESAR_PI / 8; // Controlled cross-feed of the elevation layers

        // Calculate gain of each layer first
        if (numOfLayers == 2) {
            layerGain[0] = fmax(0.0f, sinf(elevationAngle - ESAR_PI / 4.0f)); // Top
            layerGain[1] = fmax(0.0f, sinf(-elevationAngle + ESAR_PI / 4.0f)); // Bottom

            if (elevationAngle > ESAR_PI / 4.0f) {
                layerGain[0] = 1.0;
            } else if (elevationAngle < -ESAR_PI / 4.0f) {
                layerGain[1] = 1.0;
            } else {
                layerGain[0] = fmax(0.0f, sinf(elevationAngle + ESAR_PI / 4.0f)); // Top
                layerGain[1] = fmax(0.0f, sinf(-elevationAngle + ESAR_PI / 4.0f)); // Bottom
            }
        }
        if (numOfLayers == 3) {
            if (elevationAngle > ESAR_PI / 2.0f - micElevOverlap) {
                layerGain[0] = 1.0;
            } else if (elevationAngle < -(ESAR_PI / 2.0f - micElevOverlap)) {
                layerGain[2] = 1.0;
            } else {
                layerGain[0] = fmax(0.0f, sinf(elevationAngle + micElevOverlap)); // Top
                layerGain[2] = fmax(0.0f, sinf(-elevationAngle + micElevOverlap)); // Bottom
                layerGain[1] = sqrtf(pow(layerGain[0], 2) + pow(layerGain[2], 2)); // Middle
            }
        }
        // 0: Front, 1: Left, 2: Back, 3: Right
        float frontSigWeight = fmax(0.0f, cosf(horizontalAngle));
        float leftSigWeight = fmax(0.0f, cosf(horizontalAngle - ESAR_PI / 2.0f));
        float backSigWeight = fmax(0.0f, cosf(horizontalAngle - ESAR_PI));
        float rightSigWeight = fmax(0.0f, cosf(horizontalAngle - 3.0f * ESAR_PI / 2.0f));

        for (int layer = 0; layer < numOfLayers; layer++) {
            for (int i = 0; i < frameLength; i++) {
                value = gain * layerGain[layer]
                    * (frontSigWeight * layers[layer]->ch(0)[i]
                        + leftSigWeight * layers[layer]->ch(3)[i]
                        + backSigWeight * layers[layer]->ch(2)[i]
                        + rightSigWeight * layers[layer]->ch(1)[i]);
                capturedFrame->add(value, i, 1);
            }
        }
    }

    else if (type == InteriorRepresentationType::FOA) {

        float pattern = micPattern;
        float oneMinusPattern = 1.0 - pattern;
        float sinHorizontalAngle = sinf(horizontalAngle);
        float cosHorizontalAngle = cosf(horizontalAngle);
        float sinElevationAngle = sinf(elevationAngle);
        float cosElevationAngle = cosf(elevationAngle);

        for (int i = 0; i < frameLength; i++) {
            value = gain
                * (pattern * layers[0]->ch(0)[i]
                    + oneMinusPattern
                        * (cosElevationAngle * sinHorizontalAngle * layers[0]->ch(1)[i]
                            + sinElevationAngle * layers[0]->ch(2)[i]
                            + cosElevationAngle * cosHorizontalAngle * layers[0]->ch(3)[i]));
            capturedFrame->add(value, i, 1);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Local helper functions
//
//////////////////////////////////////////////////////////////////////////////////////////////////////

void CalculateHeterogeneousSourceOuputSignals(HeterogeneousDspState* DspState,
    EAR::RealFrame* inputSignal, std::vector<EAR::RealBuffer*> outputSignals, std::size_t B)
{
    double intWeight, extWeight;
    float intFadeGain = 0.0;
    float extMicPattern = 0.5;
    float intMicPattern = 0.30;

    intWeight = DspState->interiorExteriorWeight;
    extWeight = 1.0 - intWeight;
    float occlCenterHorizAngleScaler =
        (DspState->occlLeftAngleScaler - DspState->occlRightAngleScaler) / 2.0f;
    float occlCenterVertAngleScaler =
        (DspState->occlTopAngleScaler - DspState->occlBottomAngleScaler) / 2.0f;

    if (DspState->integrated3DoFHoa) {
        // Fade out the 6DoF HOA and hand over to the 3DoF HOA rendering stage
        intFadeGain = 1.0 - DspState->rearHemisphereGain * DspState->rearHemisphereGain;
    } else {
        intFadeGain = 1.0f;
    }

    // clear the RenderItem signal buffers
    for (auto sig : outputSignals) {
        sig->zero();
    }

    DspState->interiorRepresentation->DeriveInteriorRepresentationFrame(
        inputSignal, DspState->extentLayout);

    // Generate and add exterior representation signals
    DspState->interiorRepresentation->CaptureFrameFromDirection(
        DspState->relativeHorizAngleExterior - ESAR_PI / 2.0 * DspState->occlLeftAngleScaler,
        (DspState->trans_fac[1]) * (DspState->relativeVertAngleExterior - ESAR_PI / 4)
            * DspState->occlTopAngleScaler,
        outputSignals[0], extMicPattern, extWeight * DspState->exteriorSpeakerGains[0]);
    DspState->interiorRepresentation->CaptureFrameFromDirection(
        DspState->relativeHorizAngleExterior + ESAR_PI / 2.0 * DspState->occlRightAngleScaler,
        (DspState->trans_fac[1]) * (DspState->relativeVertAngleExterior - ESAR_PI / 4)
            * DspState->occlTopAngleScaler,
        outputSignals[1], extMicPattern, extWeight * DspState->exteriorSpeakerGains[1]);
    DspState->interiorRepresentation->CaptureFrameFromDirection(
        DspState->relativeHorizAngleExterior - ESAR_PI / 2.0 * occlCenterHorizAngleScaler,
        DspState->relativeVertAngleExterior + ESAR_PI / 4 * occlCenterVertAngleScaler,
        outputSignals[2], extMicPattern, extWeight * DspState->exteriorSpeakerGains[2]);
    DspState->interiorRepresentation->CaptureFrameFromDirection(
        DspState->relativeHorizAngleExterior - ESAR_PI / 2.0 * DspState->occlLeftAngleScaler,
        DspState->relativeVertAngleExterior + ESAR_PI / 4 * DspState->occlBottomAngleScaler,
        outputSignals[4], extMicPattern, extWeight * DspState->exteriorSpeakerGains[3]);
    DspState->interiorRepresentation->CaptureFrameFromDirection(
        DspState->relativeHorizAngleExterior + ESAR_PI / 2.0 * DspState->occlRightAngleScaler,
        DspState->relativeVertAngleExterior + ESAR_PI / 4 * DspState->occlBottomAngleScaler,
        outputSignals[5], extMicPattern, extWeight * DspState->exteriorSpeakerGains[4]);

    // Generate and add interior representation signals
    float intWeightTimesintFadeGain = intWeight * intFadeGain;
    for (int intSpInd = 0; intSpInd < HETEROGENEOUS_NUM_OF_INTERIOR_SPEAKERS; intSpInd++) {
        DspState->interiorRepresentation->CaptureFrameFromDirection(
            DspState->relativeHorizAngleInterior + intSpAz[intSpInd] * ESAR_PI_DIV_BY_180,
            DspState->relativeVertAngleExterior + intSpEl[intSpInd] * ESAR_PI_DIV_BY_180,
            outputSignals[intToOutputSpMapping[intSpInd]], intMicPattern,
            intWeightTimesintFadeGain * DspState->interiorSpeakerGains[intSpInd]);
    }

    return;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Local basic geometry functions that should be removed later
//
//////////////////////////////////////////////////////////////////////////////////////////////////////
void ESAR_SPATIAL_VecInit(float* const Pos_p, float const PosX, float const PosY, float const PosZ)
{
    Pos_p[0] = PosX;
    Pos_p[1] = PosY;
    Pos_p[2] = PosZ;
    return;
}

void ESAR_SPATIAL_VecCopy(float* const PosTarget_p, float const* const PosSource_p)
{
    PosTarget_p[0] = PosSource_p[0];
    PosTarget_p[1] = PosSource_p[1];
    PosTarget_p[2] = PosSource_p[2];
    return;
}

float ESAR_SPATIAL_VecInnProd(float const* const Vec1_p, float const* const Vec2_p)
{
    return Vec1_p[0] * Vec2_p[0] + Vec1_p[1] * Vec2_p[1] + Vec1_p[2] * Vec2_p[2];
}

void ESAR_SPATIAL_VecCrossProd(
    float const* const Vec1_p, float const* const Vec2_p, float* const CrossProd_p)
{
    CrossProd_p[0] = Vec1_p[1] * Vec2_p[2] - Vec1_p[2] * Vec2_p[1];
    CrossProd_p[1] = Vec1_p[2] * Vec2_p[0] - Vec1_p[0] * Vec2_p[2];
    CrossProd_p[2] = Vec1_p[0] * Vec2_p[1] - Vec1_p[1] * Vec2_p[0];
}

float ESAR_SPATIAL_VecNorm(float const* const Vec_p)
{
    return sqrtf(Vec_p[0] * Vec_p[0] + Vec_p[1] * Vec_p[1] + Vec_p[2] * Vec_p[2]);
}

void ESAR_SPATIAL_VecNormalize(float const* const Vec_p, float* const VecNorm_p)
{
    float scaler;

    if ((Vec_p[0] == 0) && (Vec_p[1] == 0) && (Vec_p[2] == 0)) {
        VecNorm_p[0] = 0.0;
        VecNorm_p[1] = 0.0;
        VecNorm_p[2] = 0.0;
        return;
    }

    scaler = 1.0f / sqrtf(Vec_p[0] * Vec_p[0] + Vec_p[1] * Vec_p[1] + Vec_p[2] * Vec_p[2]);
    VecNorm_p[0] = scaler * Vec_p[0];
    VecNorm_p[1] = scaler * Vec_p[1];
    VecNorm_p[2] = scaler * Vec_p[2];
}

void ESAR_SPATIAL_VecScale(float const* const Vec_p, float const scaler, float* const VecScale_p)
{
    VecScale_p[0] = scaler * Vec_p[0];
    VecScale_p[1] = scaler * Vec_p[1];
    VecScale_p[2] = scaler * Vec_p[2];
}

void ESAR_SPATIAL_VecNeg(float const* const Vec_p, float* const VecNeg_p)
{
    VecNeg_p[0] = -Vec_p[0];
    VecNeg_p[1] = -Vec_p[1];
    VecNeg_p[2] = -Vec_p[2];
}

void ESAR_SPATIAL_VecAdd(
    float const* const Vec1_p, float const* const Vec2_p, float* const VecAdd_p)
{
    VecAdd_p[0] = Vec1_p[0] + Vec2_p[0];
    VecAdd_p[1] = Vec1_p[1] + Vec2_p[1];
    VecAdd_p[2] = Vec1_p[2] + Vec2_p[2];
}

void ESAR_SPATIAL_VecSub(
    float const* const Vec1_p, float const* const Vec2_p, float* const VecSub_p)
{
    VecSub_p[0] = Vec1_p[0] - Vec2_p[0];
    VecSub_p[1] = Vec1_p[1] - Vec2_p[1];
    VecSub_p[2] = Vec1_p[2] - Vec2_p[2];
}

void ESAR_SPATIAL_VecProjError(
    float const* const Vec1_p, float const* const Vec2_p, float* const ProjErr_p)
{
    float ProjCoef;

    ProjCoef = ESAR_SPATIAL_VecInnProd(Vec1_p, Vec2_p) / ESAR_SPATIAL_VecInnProd(Vec2_p, Vec2_p);
    ProjErr_p[0] = Vec1_p[0] - ProjCoef * Vec2_p[0];
    ProjErr_p[1] = Vec1_p[1] - ProjCoef * Vec2_p[1];
    ProjErr_p[2] = Vec1_p[2] - ProjCoef * Vec2_p[2];
}

bool ESAR_SPATIAL_VecEpsilonComparison(
    float const* const Vec1_p, float const* const Vec2_p, float maxRelDiff)
{
    // Return true if the two vectors are close enough to be said to be the same
    float diff0 = Vec2_p[0] - Vec1_p[0];
    float max_0 = (Vec2_p[0] > Vec1_p[0]) ? Vec2_p[0] : Vec1_p[0];
    float diff1 = Vec2_p[1] - Vec1_p[1];
    float max_1 = (Vec2_p[1] > Vec1_p[1]) ? Vec2_p[1] : Vec1_p[1];
    float diff2 = Vec2_p[2] - Vec1_p[2];
    float max_2 = (Vec2_p[2] > Vec1_p[2]) ? Vec2_p[2] : Vec1_p[2];

    if (fabs(diff0) > fabs(max_0) * maxRelDiff)
        return false;
    if (fabs(diff1) > fabs(max_1) * maxRelDiff)
        return false;
    if (fabs(diff2) > fabs(max_2) * maxRelDiff)
        return false;

    return true;
}

void ESAR_SPATIAL_ClosestPointOnPlane(float const* const referencePoint_p,
    float const* const planeVertex0_p, float const* const planeVertex1_p,
    float const* const planeVertex2_p, float* const closestPoint_p)
{
    float vec1[3], vec2[3], normal[3], translVec[3], distance;

    // Calculate the normal of the plane
    ESAR_SPATIAL_VecSub(planeVertex1_p, planeVertex0_p, vec1);
    ESAR_SPATIAL_VecSub(planeVertex2_p, planeVertex0_p, vec2);
    ESAR_SPATIAL_VecCrossProd(vec1, vec2, normal);
    ESAR_SPATIAL_VecNormalize(normal, normal);
    ESAR_SPATIAL_VecSub(planeVertex0_p, referencePoint_p, translVec);
    distance = ESAR_SPATIAL_VecInnProd(normal, translVec);

    ESAR_SPATIAL_VecScale(normal, distance, closestPoint_p);
    ESAR_SPATIAL_VecAdd(closestPoint_p, referencePoint_p, closestPoint_p);
}

bool ESAR_SPATIAL_IsPointInTriangle(float const* const referencePoint_p,
    float const* const triangleVertex0_p, float const* const triangleVertex1_p,
    float const* const triangleVertex2_p)
{
    float vec0[3], vec1[3], vec2[3];
    float dotProd00, dotProd01, dotProd02, dotProd11, dotProd12;
    float invDenom, u, v;

    // Using Barycentric Technique, https://blackpawn.com/texts/pointinpoly/
    // Compute vectors
    ESAR_SPATIAL_VecSub(triangleVertex2_p, triangleVertex0_p, vec0);
    ESAR_SPATIAL_VecSub(triangleVertex1_p, triangleVertex0_p, vec1);
    ESAR_SPATIAL_VecSub(referencePoint_p, triangleVertex0_p, vec2);

    // Compute dot products
    dotProd00 = ESAR_SPATIAL_VecInnProd(vec0, vec0);
    dotProd01 = ESAR_SPATIAL_VecInnProd(vec0, vec1);
    dotProd02 = ESAR_SPATIAL_VecInnProd(vec0, vec2);
    dotProd11 = ESAR_SPATIAL_VecInnProd(vec1, vec1);
    dotProd12 = ESAR_SPATIAL_VecInnProd(vec1, vec2);

    // Compute barycentric coordinates
    invDenom = 1.0f / (dotProd00 * dotProd11 - dotProd01 * dotProd01);
    u = (dotProd11 * dotProd02 - dotProd01 * dotProd12) * invDenom;
    v = (dotProd00 * dotProd12 - dotProd01 * dotProd02) * invDenom;

    // Check if point is in triangle
    return (u >= 0) && (v >= 0) && (u + v < 1);
}

void ESAR_SPATIAL_ClosestPointOnLine(float const* const referencePoint_p,
    float const* const lineStartPoint_p, float const* const lineEndPoint_p,
    float* const closestPoint_p)
{
    float vec1[3], vec2[3], t, DotProd1, DotProd2;

    ESAR_SPATIAL_VecSub(referencePoint_p, lineStartPoint_p, vec1);
    ESAR_SPATIAL_VecSub(lineEndPoint_p, lineStartPoint_p, vec2);
    DotProd1 = ESAR_SPATIAL_VecInnProd(vec1, vec2);
    DotProd2 = ESAR_SPATIAL_VecInnProd(vec2, vec2);
    t = DotProd1 / DotProd2;
    if (t > 1.0f) {
        t = 1.0f;
    }
    if (t < 0.0f) {
        t = 0.0f;
    }

    ESAR_SPATIAL_VecScale(vec2, t, closestPoint_p);
    ESAR_SPATIAL_VecAdd(closestPoint_p, lineStartPoint_p, closestPoint_p);
}

void ESAR_SPATIAL_ClosestPointOnTriangle(float const* const referencePoint_p,
    float const* const triangleVertex0_p, float const* const triangleVertex1_p,
    float const* const triangleVertex2_p, float* const closestPoint_p, float* const distance_p)
{
    float closestPointOnLine1[3], closestPointOnLine2[3], closestPointOnLine3[3], distVec[3];
    float magnitude1, magnitude2, magnitude3, min_mag;

    ESAR_SPATIAL_ClosestPointOnPlane(
        referencePoint_p, triangleVertex0_p, triangleVertex1_p, triangleVertex2_p, closestPoint_p);

    if (ESAR_SPATIAL_IsPointInTriangle(
            closestPoint_p, triangleVertex0_p, triangleVertex1_p, triangleVertex2_p)) {
        ESAR_SPATIAL_VecSub(closestPoint_p, referencePoint_p, distVec);
        *distance_p = ESAR_SPATIAL_VecNorm(distVec);
        return;
    }

    ESAR_SPATIAL_ClosestPointOnLine(
        referencePoint_p, triangleVertex0_p, triangleVertex1_p, closestPointOnLine1);
    ESAR_SPATIAL_ClosestPointOnLine(
        referencePoint_p, triangleVertex1_p, triangleVertex2_p, closestPointOnLine2);
    ESAR_SPATIAL_ClosestPointOnLine(
        referencePoint_p, triangleVertex2_p, triangleVertex0_p, closestPointOnLine3);

    ESAR_SPATIAL_VecSub(closestPointOnLine1, referencePoint_p, distVec);
    magnitude1 = ESAR_SPATIAL_VecInnProd(distVec, distVec);
    ESAR_SPATIAL_VecSub(closestPointOnLine2, referencePoint_p, distVec);
    magnitude2 = ESAR_SPATIAL_VecInnProd(distVec, distVec);
    ESAR_SPATIAL_VecSub(closestPointOnLine3, referencePoint_p, distVec);
    magnitude3 = ESAR_SPATIAL_VecInnProd(distVec, distVec);

    min_mag = fmin(magnitude1, magnitude2);
    min_mag = fmin(min_mag, magnitude3);
    *distance_p = sqrt(min_mag);
    if (min_mag == magnitude1) {
        ESAR_SPATIAL_VecCopy(closestPoint_p, closestPointOnLine1);
        return;
    } else if (min_mag == magnitude2) {
        ESAR_SPATIAL_VecCopy(closestPoint_p, closestPointOnLine2);
        return;
    } else {
        ESAR_SPATIAL_VecCopy(closestPoint_p, closestPointOnLine3);
        return;
    }
}

void ESAR_SPATIAL_EvalOrthonormOrient(float* const FrontVecON_p, float* const UpVecON_p,
    float* const RightVecON_p, float const* const FrontVec_p, float const* const UpVec_p)
{
    // Evaluate the normalized front vector
    ESAR_SPATIAL_VecNormalize(FrontVec_p, FrontVecON_p);

    // Evaluate the orthonormal right vector
    // through the cross product of the front and the up vectors
    RightVecON_p[0] = FrontVecON_p[1] * UpVec_p[2] - FrontVecON_p[2] * UpVec_p[1];
    RightVecON_p[1] = FrontVecON_p[2] * UpVec_p[0] - FrontVecON_p[0] * UpVec_p[2];
    RightVecON_p[2] = FrontVecON_p[0] * UpVec_p[1] - FrontVecON_p[1] * UpVec_p[0];
    ESAR_SPATIAL_VecNormalize(RightVecON_p, RightVecON_p);

    // Evaluate the orthonormal up vector
    // through the cross product of the front and the right vectors
    UpVecON_p[0] = RightVecON_p[1] * FrontVecON_p[2] - RightVecON_p[2] * FrontVecON_p[1];
    UpVecON_p[1] = RightVecON_p[2] * FrontVecON_p[0] - RightVecON_p[0] * FrontVecON_p[2];
    UpVecON_p[2] = RightVecON_p[0] * FrontVecON_p[1] - RightVecON_p[1] * FrontVecON_p[0];
    ESAR_SPATIAL_VecNormalize(UpVecON_p, UpVecON_p);
}

float ESAR_SPATIAL_Smooth_angle(float previous_angle, float new_angle, float smoothing_coeff)
{
    float smoothed_angle;
    float anglediff = new_angle - previous_angle;

    if (anglediff < -180.0f) {
        smoothed_angle =
            smoothing_coeff * (previous_angle - 360.0f) + (1.0f - smoothing_coeff) * new_angle;
    } else if (anglediff > 180.0f) {
        smoothed_angle =
            smoothing_coeff * (previous_angle + 360.0f) + (1.0f - smoothing_coeff) * new_angle;
    } else {
        smoothed_angle = smoothing_coeff * previous_angle + (1.0f - smoothing_coeff) * new_angle;
    }

    return smoothed_angle;
}

void ESAR_SPATIAL_GetNeighborVertices(
    int referenceVertexInd, const EAR::Mesh* mesh, std::vector<int>* neighborList)
{
    int corner;
    float distanceSquared;
    bool refVertexFound = false;
    Vector3 distVector;
    std::vector vertices = mesh->getVertices(Mesh::CSpace::Global);
    std::vector triangles = mesh->getTriangles();

    for (int i = 0; i < triangles.size(); i++) {
        // Check if the reference point is part of this trangle
        refVertexFound = false;
        for (corner = 0; corner < 3; corner++) {

            if (corner == 0) {
                distVector = vertices[triangles[i].v1] - vertices[referenceVertexInd];
            } else if (corner == 1) {
                distVector = vertices[triangles[i].v2] - vertices[referenceVertexInd];
            } else {
                distVector = vertices[triangles[i].v3] - vertices[referenceVertexInd];
            }

            // Calculate the squared distance here to avoid the square root, we don't need it
            distanceSquared = distVector[0] * distVector[0] + distVector[1] * distVector[1]
                + distVector[2] * distVector[2];

            if (distanceSquared <= ESAR_EPSILON_DISTANCE) {
                refVertexFound = true;
                break;
            }
        }

        if (refVertexFound) {
            // Found a neigboring triangle, now add all vertices of this triangle to the list of neighbors
            // if they were not already there
            if (std::find(neighborList->begin(), neighborList->end(), triangles[i].v1)
                == neighborList->end()) {
                neighborList->emplace_back(triangles[i].v1);
            }
            if (std::find(neighborList->begin(), neighborList->end(), triangles[i].v2)
                == neighborList->end()) {
                neighborList->emplace_back(triangles[i].v2);
            }
            if (std::find(neighborList->begin(), neighborList->end(), triangles[i].v3)
                == neighborList->end()) {
                neighborList->emplace_back(triangles[i].v3);
            }
        }
    }
}

void ESAR_SPATIAL_FindMeshOutline(const EAR::Mesh* mesh, std::vector<int>* outlineVerticesInd)
{
    int i, loop;
    std::vector vertices = mesh->getVertices(Mesh::CSpace::Global);
    float max_x = vertices[0][0];
    int max_x_ind = 0;
    int currentVerticeInd = 0;
    int minNeigborAzimuthAngleIndex = 0;
    bool minAzimuthNeighborFound = false;
    float distanceSquared = 0.0f;
    float previousAzimuthAngle = 90.0f;
    Vector3 vectorToMinAzimNeighbor(1.0f, 0.0f, 0.0f);
    Vector3 frontVec(0.0f, 0.0f, -1.0f);
    Vector3 upVec(0.0f, 1.0f, 0.0f);
    Vector3 rightVec = (upVec * frontVec).normalize();

    // Find the vertex positioned furthest along the x-axis
    for (i = 0; i < vertices.size(); i++) {

        if (vertices[i][0] > max_x) {
            max_x = vertices[i][0];
            max_x_ind = i;
        } else if (vertices[i][0] == max_x) {
            // If there are more vertices with the same value on the x-axis, choose the one lowest on the y-axis
            if (vertices[i][1] < vertices[max_x_ind][1]) {
                max_x = vertices[i][0];
                max_x_ind = i;
            }
        }
    }
    outlineVerticesInd->emplace_back(max_x_ind);

    currentVerticeInd = max_x_ind;

    for (loop = 0; loop < vertices.size(); loop++) {
        // Find the neighboor vertice that has the smallest azimuth angle, as seen from the current vertice
        std::vector<int> neighborList;
        ESAR_SPATIAL_GetNeighborVertices(currentVerticeInd, mesh, &neighborList);

        float minNeigborAzimuthAngle = 360.0f;
        minNeigborAzimuthAngleIndex = 0;
        minAzimuthNeighborFound = false;

        for (i = 0; i < neighborList.size(); i++) {
            // Calculate the azimuth angle
            Vector3 vectorToNeighbor = vertices[neighborList[i]] - vertices[currentVerticeInd];
            vectorToNeighbor[1] = 0.0f; // Project onto world horizontal plane

            // Calculate the squared distance here to avoid the square root, we don't need it
            distanceSquared = vectorToNeighbor[0] * vectorToNeighbor[0]
                + vectorToNeighbor[2] * vectorToNeighbor[2];

            if (distanceSquared > ESAR_EPSILON_DISTANCE) {

                // Calculate the relative coordinates in the local coordinate system based on the angle between the previous points
                Vector3 localPoint;
                float angleEspilon = 0.01f;
                localPoint[0] = vectorToNeighbor.dot(frontVec);
                localPoint[1] = 0.0f;
                localPoint[2] = vectorToNeighbor.dot(rightVec);

                float neighborAzimuthAngle =
                    atan2(-localPoint[2], localPoint[0]) * ESAR_180_DIV_BY_PI;
                if (neighborAzimuthAngle < angleEspilon) {
                    neighborAzimuthAngle += 360.0f;
                }

                // Check if the azimuth angle of the vector to this vertex from the last one
                // is smallest found yet
                if (((neighborAzimuthAngle > 90.0f)
                        && (neighborAzimuthAngle < minNeigborAzimuthAngle))
                    || ((neighborAzimuthAngle > 90.0f)
                        && (neighborAzimuthAngle == minNeigborAzimuthAngle)
                        && (vertices[neighborList[i]][1]
                            < vertices[minNeigborAzimuthAngleIndex][1]))) {
                    minNeigborAzimuthAngle = neighborAzimuthAngle;
                    minNeigborAzimuthAngleIndex = neighborList[i];
                    vectorToMinAzimNeighbor = vectorToNeighbor;
                    minAzimuthNeighborFound = true;
                }
            }
        }
        if (minAzimuthNeighborFound == false) {
            minNeigborAzimuthAngleIndex = 0;
        }

        // Add the vertex to the outline if it was not already there
        if (std::find(
                outlineVerticesInd->begin(), outlineVerticesInd->end(), minNeigborAzimuthAngleIndex)
            == outlineVerticesInd->end()) {

            outlineVerticesInd->emplace_back(minNeigborAzimuthAngleIndex);
            currentVerticeInd = minNeigborAzimuthAngleIndex;
            previousAzimuthAngle = minNeigborAzimuthAngle;

            // Update coordinate system for the next neighbor evaluation
            frontVec = -vectorToMinAzimNeighbor;
            rightVec = (upVec * frontVec).normalize();
        } else {
            // We are trying to add a vertex that is alread included, so have gone full circle
            break;
        }
    }
}

void ESAR_SPATIAL_GenerateOutlineNormalList(
    const EAR::Mesh* mesh, std::vector<OutlineNormal>* outlineNormals)
{
    int i;
    Vector3 newNormalPos(0.0f, 0.0f, 0.0f);
    Vector3 newNormal(0.0f, 0.0f, 0.0f);
    std::vector vertices = mesh->getVertices(Mesh::CSpace::Global);
    std::vector<int> outlineVerticesInd;

    outlineNormals->clear();

    ESAR_SPATIAL_FindMeshOutline(mesh, &outlineVerticesInd);

    if (outlineVerticesInd.size() > 2) {

        for (i = 0; i < outlineVerticesInd.size() - 1; i++) {
            newNormalPos =
                vertices[outlineVerticesInd[i]] + vertices[outlineVerticesInd[i + (int)1]];
            newNormalPos.operator*=(0.5f);

            newNormal.set(
                (vertices[outlineVerticesInd[i + (int)1]][2] - vertices[outlineVerticesInd[i]][2]),
                0.0f,
                -(vertices[outlineVerticesInd[i + (int)1]][0]
                    - vertices[outlineVerticesInd[i]][0]));
            newNormal = newNormal.normalize();

            outlineNormals->emplace_back(
                newNormal, newNormalPos, outlineVerticesInd[i], outlineVerticesInd[i + 1]);
        }

        // Add the last normal
        newNormalPos = vertices[outlineVerticesInd[i]] + vertices[outlineVerticesInd[0]];
        newNormalPos.operator*=(0.5f);

        newNormal.set((vertices[outlineVerticesInd[0]][2] - vertices[outlineVerticesInd[i]][2]),
            0.0f, -(vertices[outlineVerticesInd[0]][0] - vertices[outlineVerticesInd[i]][0]));
        newNormal = newNormal.normalize();
        outlineNormals->emplace_back(
            newNormal, newNormalPos, outlineVerticesInd[i], outlineVerticesInd[0]);
    }
}

int ESAR_SPATIAL_FindClosestOutlineNormalIndex(
    std::vector<OutlineNormal> outlineNormals, Vector3 refPosition)
{
    float shortestSquaredDist = 100000000000.0f;
    int closestNormalIndex = 0;
    Vector3 projPoint = refPosition;
    projPoint[1] = 0.0f; // Project onto the horizontal plane

    for (int loop = 0; loop < outlineNormals.size(); loop++) {
        Vector3 normalPositionRelative = outlineNormals[loop].position - projPoint;
        normalPositionRelative[1] = 0.0f; // Project onto the horizontal plane
        float squaredDistance = normalPositionRelative[0] * normalPositionRelative[0]
            + normalPositionRelative[2] * normalPositionRelative[2];
        if (squaredDistance <= shortestSquaredDist) {
            shortestSquaredDist = squaredDistance;
            closestNormalIndex = loop;
        }
    }

    return closestNormalIndex;
}

int ESAR_SPATIAL_FindClosestOutlineNormalIndex_LineSearch(const EAR::Mesh* mesh,
    std::vector<OutlineNormal> outlineNormals, Vector3 refPosition, float* pShortestSquaredDistance)
{
    std::vector vertices = mesh->getVertices(Mesh::CSpace::Global);
    float shortestSquaredDist = 100000000000.0f;
    float squaredDistance;
    int closestNormalIndex = 0;
    Vector3 projPoint1;
    Vector3 projPoint2;
    Vector3 projRefPoint = refPosition;
    projRefPoint[1] = 0.0; // Project onto the horizontal plane

    for (int loop = 0; loop < outlineNormals.size(); loop++) {

        // Find the closest point on the line that the normal is situated at
        projPoint1 = vertices[outlineNormals[loop].vertex1];
        projPoint1[1] = 0.0;
        projPoint2 = vertices[outlineNormals[loop].vertex2];
        projPoint2[1] = 0.0;

        float p1[3], p2[3], ref[3], closest[3];
        ESAR_SPATIAL_VecInit(p1, projPoint1[0], 0.0, projPoint1[2]);
        ESAR_SPATIAL_VecInit(p2, projPoint2[0], 0.0, projPoint2[2]);
        ESAR_SPATIAL_VecInit(ref, projRefPoint[0], 0.0, projRefPoint[2]);

        ESAR_SPATIAL_ClosestPointOnLine(ref, p1, p2, closest);
        squaredDistance = (closest[0] - ref[0]) * (closest[0] - ref[0])
            + (closest[2] - ref[2]) * (closest[2] - ref[2]);

        if (squaredDistance <= shortestSquaredDist) {
            shortestSquaredDist = squaredDistance;
            closestNormalIndex = loop;
        }
    }

    *pShortestSquaredDistance = shortestSquaredDist;

    return closestNormalIndex;
}

bool ESAR_SPATIAL_IsInsideOutline(std::vector<OutlineNormal> outlineNormals, Vector3 refPosition)
{
    // Find the closest normal position
    int closestNormInd = 0;
    Vector3 projPoint = refPosition;
    projPoint[1] = 0.0f; // Project onto the horizontal plane
    closestNormInd = ESAR_SPATIAL_FindClosestOutlineNormalIndex(outlineNormals, projPoint);

    // See if the vector from the reference position to the closest normal is going along or against the normal
    float normRayDot = outlineNormals[closestNormInd].normal.dot(
        outlineNormals[closestNormInd].position - projPoint);

    if (normRayDot < 0.0) {
        // The point was outside the outline
        return false;
    } else {
        // The point was inside the outline
        return true;
    }
}

float ESAR_SPATIAL_FindDistanceToOutline(
    const EAR::Mesh* mesh, std::vector<OutlineNormal> outlineNormals, Vector3 refPosition)
{
    std::vector vertices = mesh->getVertices(Mesh::CSpace::Global);
    int closestNormInd = 0;
    Vector3 closestPoint;
    Vector3 projRefPoint = refPosition;
    Vector3 projPoint1;
    Vector3 projPoint2;
    float distance;
    projRefPoint[1] = 0.0f; // Project onto the horizontal plane

    // TODO: Check more normals than the closest one
    closestNormInd = ESAR_SPATIAL_FindClosestOutlineNormalIndex(outlineNormals, projRefPoint);

    // Find the closest point on the line that the normal is situated at
    projPoint1 = vertices[outlineNormals[closestNormInd].vertex1];
    projPoint1[1] = 0.0;
    projPoint2 = vertices[outlineNormals[closestNormInd].vertex2];
    projPoint2[1] = 0.0;

    float p1[3], p2[3], ref[3], closest[3];
    ESAR_SPATIAL_VecInit(p1, projPoint1[0], 0.0, projPoint1[2]);
    ESAR_SPATIAL_VecInit(p2, projPoint2[0], 0.0, projPoint2[2]);
    ESAR_SPATIAL_VecInit(ref, projRefPoint[0], 0.0, projRefPoint[2]);

    ESAR_SPATIAL_ClosestPointOnLine(ref, p1, p2, closest);
    distance = sqrtf((closest[0] - ref[0]) * (closest[0] - ref[0])
        + (closest[2] - ref[2]) * (closest[2] - ref[2]));

    return distance;
}

void ESAR_SPATIAL_MapPointToLocalSystem(const float* vertex, float* local_vertex,
    float* listenerPos, const float* const observationSystemFrontVec,
    const float* const observationSystemUpVec, const float* const observationSystemRightVec)
{
    float vertex_relative[3];

    ESAR_SPATIAL_VecSub(vertex, listenerPos, vertex_relative);
    local_vertex[0] = ESAR_SPATIAL_VecInnProd(vertex_relative, observationSystemFrontVec);
    local_vertex[1] = ESAR_SPATIAL_VecInnProd(vertex_relative, observationSystemUpVec);
    local_vertex[2] = ESAR_SPATIAL_VecInnProd(vertex_relative, observationSystemRightVec);
}

} // namespace MPEGI
