/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_HETEROGENEOUS
#define MPEGI_HETEROGENEOUS

#include <EAR/BaseTypes.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Helpers/VoxelTraversal.h>
#include <MPEGI/Scene.h>
#include <mpegcore/EsdConverter.hpp>

namespace EAR {
class RealBuffer;
struct Position;
class DistanceGainVolumetric;
class Polar;
class MpegiDecorrelator;
} // namespace EAR

namespace MPEGI {

#define HETEROGENEOUS_INTERIOR_LAYER_CHANNELS 4
#define HETEROGENEOUS_MAX_NO_INPUT_CHANNELS 9
#define HETEROGENEOUS_MAX_INPUT_CHANNELS_PER_LAYER 3
#define HETEROGENEOUS_NUM_OF_EXTERIOR_SPEAKERS 5
#define HETEROGENEOUS_NUM_OF_INTERIOR_SPEAKERS 12
#define HETEROGENEOUS_NUM_OF_OUTPUT_SPEAKERS 14
#define HETEROGENEOUS_NUM_OF_SUBAREAS 6
#define HETEROGENEOUS_EQ_SMOOTH 0.5f
#define HETEROGENEOUS_OCCL_GRID_RESOLUTION 5.0f
#define HETEROGENEOUS_OCCL_GRID_MIN_AZIM_RAYS 5
#define HETEROGENEOUS_OCCL_GRID_MAX_AZIM_RAYS 20
#define HETEROGENEOUS_OCCL_GRID_MIN_ELEV_RAYS 3
#define HETEROGENEOUS_OCCL_GRID_MAX_ELEV_RAYS 10
#define HETEROGENEOUS_OCCL_SUB_AREA_RESOLUTION 1.0f

// This value should match the delay difference between the 3DoF HOA and heterogeneous rendering stages
#define ESAR_SRC_SPATIAL_3DOF_HOA_DELAY 743

// Sub area occlusion EQ to output speaker EQ mapping matries
static const std::map<int, std::vector<float>> subareaEqToSpeakerEqMapPlane{
    { 0, { 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } }, // Top left
    { 1, { 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f } }, // Top right
    { 2, { 0.15f, 0.2f, 0.15f, 0.15f, 0.2f, 0.15f } }, // Middle
    { 4, { 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f } }, // Bottom left
    { 5, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f } } // Bottom right
};
static const std::map<int, std::vector<float>> subareaEqToSpeakerEqMapLine{
    { 0, { 0.5f, 0.0f, 0.0f, 0.5f, 0.0f, 0.0f } }, // Top left
    { 1, { 0.0f, 0.0f, 0.5f, 0.0f, 0.0f, 0.5f } }, // Top right
    { 2, { 0.15f, 0.2f, 0.15f, 0.15f, 0.2f, 0.15f } }, // Middle
    { 4, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } }, // Bottom left
    { 5, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } } // Bottom right
};
static const std::map<int, std::vector<float>> subareaEqToSpeakerEqMapPoint{
    { 0, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } }, // Top left
    { 1, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } }, // Top right
    { 2, { 0.1666f, 0.1666f, 0.1666f, 0.1666f, 0.1666f, 0.1666f } }, // Middle
    { 4, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } }, // Bottom left
    { 5, { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f } } // Bottom right
};

enum class InteriorRepresentationType
{
    LayeredQuadrasonic,
    FOA
};

// Occlusion states
enum class OcclusionState
{
    occluded = 0,
    not_occluded,
    outside_extent
};

struct CroppingOcclusionInfo
{
    float minUnoccludedAzim;
    float minUnoccludedAzimElev;
    float minUnoccludedAzimRange;
    float maxUnoccludedAzim;
    float maxUnoccludedAzimElev;
    float maxUnoccludedAzimRange;
    float minUnoccludedElev;
    float minUnoccludedElevAzim;
    float minUnoccludedElevRange;
    float maxUnoccludedElev;
    float maxUnoccludedElevAzim;
    float maxUnoccludedElevRange;
    bool occlusionDetected;
    int noOfRaysUsed;
};

/*
     *  InteriorRepresentation
     *
     *  This class holds the audio buffers of the interior representation of an heterogeneous
     *  audio element and has functions for generating the interior representation from the
     *  input signals as well as functions for extracting signals from different directions
     *  of the sounds field that is described by the interior representation.
     *
     */
class InteriorRepresentation
{
public:
    InteriorRepresentation(InteriorRepresentationType intRepType,
        const MPEGI::ExtentLayout* extentLayout, std::size_t B, int inputChannels, int delay);

    ~InteriorRepresentation()
    {
        layers.clear();
        decorrelators.clear();
    }

    void DeriveInteriorRepresentationFrame(
        EAR::RealFrame* inputSignal, MPEGI::ExtentLayout extentLayout);

    void CaptureFrameFromDirection(float horizontalAngle, float elevationAngle,
        EAR::RealBuffer* capturedFrame, float micPattern, float gain);

private:
    InteriorRepresentationType type;
    int frameLength;
    int numOfLayers;
    int layerSignalInputChanInd[HETEROGENEOUS_INTERIOR_LAYER_CHANNELS * 3];
    float corrEst;
    EAR::RealFrame* delayBuffer;
    int delay_in_samples;
    std::vector<EAR::RealFrame*> layers;
    std::vector<EAR::MpegiDecorrelator*> decorrelators;
    std::vector<EAR::MpegiDecorrelator*> decorrelators_front;
    std::unique_ptr<EAR::RealFrame> decorrProcBuffer;
    std::unique_ptr<EAR::RealFrame> decorrProcBuffer_front;
    mpegcore::EsdConverter esdConverter;
};

/*
     * HeterogeneousDspState
     *
     * This class holds information that is needed when calculating the output signals.
     * The variables that may change in the Update function of the rendering stage are copied
     * from the HeterogeneousSource in order to avoid accessing these directly from the DSP thread.
     *
     */
class HeterogeneousDspState
{
public:
    HeterogeneousDspState(MPEGI::ItemType itType, const MPEGI::ExtentLayout* extentLayout,
        std::size_t B, int inputChannels, bool integr3DoFHoa)
        : interiorExteriorWeight(0.0)
        , center_gain_comp(1.0)
        , relativeHorizAngleExterior(0.0)
        , relativeVertAngleExterior(0.0)
        , relativeHorizAngleInterior(0.0)
        , relativeVertAngleInterior(0.0)
        , B(B)
        , integrated3DoFHoa(integr3DoFHoa)
        , rearHemisphereGain(0.0)
        , topHemisphereGain(0.0)
        , bottomHemisphereGain(0.0)
        , occlLeftAngleScaler(1.0)
        , occlRightAngleScaler(1.0)
        , occlTopAngleScaler(1.0)
        , occlBottomAngleScaler(1.0)
    {
        if (itType == MPEGI::ItemType::HOA) {
            interiorRepresentation = new InteriorRepresentation(InteriorRepresentationType::FOA,
                extentLayout, B, inputChannels, ESAR_SRC_SPATIAL_3DOF_HOA_DELAY);

        } else {
            interiorRepresentation = new InteriorRepresentation(
                InteriorRepresentationType::LayeredQuadrasonic, extentLayout, B, inputChannels, 0);
        }

        for (int loop = 0; loop < HETEROGENEOUS_NUM_OF_EXTERIOR_SPEAKERS; loop++) {
            exteriorSpeakerGains.push_back(0.0);
        }
        for (int loop = 0; loop < HETEROGENEOUS_NUM_OF_INTERIOR_SPEAKERS; loop++) {
            interiorSpeakerGains.push_back(0.0);
        }

        trans_fac[0] = 0.0;
        trans_fac[1] = 0.0;
        trans_fac[2] = 0.0;
    }

    ~HeterogeneousDspState() { delete interiorRepresentation; }

    void UpdateState(MPEGI::ExtentLayout* extentLayout, float interiorExteriorWeight,
        float relHorAngleExt, float relVerAngleExt, float relHorAngleInt, float relVerAngleInt,
        float rearHemisphereGain, float topHemiGain, float bottomHemiGain, float trans_fac[],
        std::vector<float> extSpeakerGains, std::vector<float> intSpeakerGains,
        float occlLeftAngleScaler, float occlRightAngleScaler, float occlTopAngleScaler,
        float occlBottomAngleScaler);

    std::size_t B;
    InteriorRepresentation* interiorRepresentation;
    ExtentLayout extentLayout;
    std::vector<float> exteriorSpeakerGains;
    std::vector<float> interiorSpeakerGains;
    float interiorExteriorWeight;
    float center_gain_comp;
    bool integrated3DoFHoa;
    float rearHemisphereGain;
    float topHemisphereGain;
    float bottomHemisphereGain;
    float relativeHorizAngleExterior;
    float relativeVertAngleExterior;
    float relativeHorizAngleInterior;
    float relativeVertAngleInterior;
    float trans_fac[3];
    float occlLeftAngleScaler;
    float occlRightAngleScaler;
    float occlTopAngleScaler;
    float occlBottomAngleScaler;
};

/*
     * OutlineNormal
     * 
     * Small helper class to keep information belonging to each outline normal
     */
class OutlineNormal
{
public:
    OutlineNormal(EAR::Vector3 norm, EAR::Vector3 pos, int vert1, int vert2)
        : position(pos)
        , normal(norm)
        , vertex1(vert1)
        , vertex2(vert2){};

    Vector3 position;
    Vector3 normal;
    int vertex1;
    int vertex2;
};

/// Information needed for voxel-grid traversal
struct voxel_info
{
    voxel_info() = default;
    voxel_info(voxel_info const&) = delete;
    voxel_info& operator=(voxel_info const&) = delete;

    using shared = std::shared_ptr<voxel_info>;
    Voxel::voxel_scene_info_basic::shared sceneInfo; //!< Basic voxel-scene information
    Voxel::grid::shared grid; //!< Voxel grid
    Voxel::scene_material_map_t material_map; //!< Map of material IDs (voxel-scene)
    std::shared_ptr<struct occlusion_recovery> recovery; //!< Occlusion recovery
};

/*
     * HeterogeneousSource
     *
     * This handles the rendering of one heterogeneous source.
     *
     */
class HeterogeneousSource
{
public:
    //! Constructor
    HeterogeneousSource(MPEGI::ItemType itType, const ExtentLayout* extentLayout,
        const EAR::Geometry* extent, float refDistance, float transitionDist, std::size_t blocksize,
        int inputChannels, bool integr3DoFHoa);

    //! Destructor
    ~HeterogeneousSource();

    //! Function for updating the state of the heterogeneous source
    bool UpdateState(const EAR::Geometry* extent, EAR::Position sourcePosition,
        EAR::Position listenerPosition, std::unique_ptr<EAR::EmbreeTracer>& tracer);

    //! Function for copying into to the HeterogeneousDspState, which will be used in the processBlock function
    void UpdateDspState();

    int GetNumOfOutputSpeakers() { return numOfOutputSpeakers; }

    EAR::Position GetOutputSpeakerPosition(int index) { return outputSpeakerPosVec[index]; };
    double GetOutputSpeakerDistanceOffset(int index)
    {
        return outputSpeakerDistanceOffsets[index];
    };
    bool GetOutputSpeakerActivity(int index) { return outputSpeakerActive[index]; };
    double GetOutputSpeakerDistanceGain(int index) { return outputSpeakerDistanceGains[index]; };
    float Get3DofHoaGain() { return ThreeDofHoaGain; };
    bool GetIntegrated3DoFHoa() { return integrated3DoFHoa; };
    float* GetOcclusionEqData(int i);
    void SetDistanceModel(int distGainModel) { distanceGainModel = distGainModel; };
    void SetDiffuseness(float diffuse) { diffuseness = diffuse; };
    float GetOccludedExtentGain() { return occludedExtentGain; };
    EAR::Position GetOccludedExtentPos() { return occludedExtentPos; };
    void SetVoxelInfo(voxel_info::shared const& voxelInfo)
    {
        this->voxelInfo = voxelInfo;

        if (voxelInfo) {
            // Estimated maximum path length
            auto const& res = voxelInfo->sceneInfo->dim;
            voxelTraversal.reserve(res[0] + res[1] + res[2]);
        }
    }

    ExtentLayout extentLayout;
    MPEGI::ItemType itemType;
    HeterogeneousDspState* heterogeneousDspState;

private:
    std::size_t B; // Blocksize in samples
    float refDistance;
    float transitionDistance;
    int numOfOutputSpeakers;
    int numOfInteriorSpeakers;
    int numOfExteriorSpeakers;
    int distanceGainModel; // 0 = volumetric distance model, 1 = point source distance model
    float diffuseness;
    std::vector<EAR::Position> exteriorSpeakerPosVec;
    std::vector<EAR::Position> interiorSpeakerPosVec;
    std::vector<float> exteriorSpeakerGains;
    std::vector<float> interiorSpeakerGains;
    std::vector<EAR::Position> outputSpeakerPosVec;
    std::vector<double> outputSpeakerDistanceOffsets;
    std::vector<double> outputSpeakerDistanceGains;
    std::vector<bool> outputSpeakerActive;
    float intSpeakerHorizRotation;
    float deltaDistanceChange;
    float temporalDescrPointSmoothFact;
    Vector3 previousListenerLocation;
    Vector3 previousSourceLocation;
    EAR::Euler previousSourceOrientation;
    Vector3 previousObservationFrontVector;

    // Description points
    float stored_anchorPoint[3];
    bool anchorPointInitialized;
    float stored_leftEdgePoint[3];
    bool leftEdgePointInitialized;
    float stored_rightEdgePoint[3];
    bool rightEdgePointInitialized;
    float stored_topEdgePoint[3];
    bool topEdgePointInitialized;
    float stored_bottomEdgePoint[3];
    bool bottomEdgePointInitialized;

    // Simplified extent
    Vector3 anchorPoint;
    Vector3 leftEdgePoint;
    Vector3 rightEdgePoint;
    Vector3 topEdgePoint;
    Vector3 bottomEdgePoint;
    float maxAzim;
    float minAzim;
    float maxElev;
    float minElev;
    float planeWidth;
    float planeHeight;
    float distanceToExtent;

    float interiorExteriorWeight;
    float center_gain_comp;
    std::unique_ptr<EAR::DistanceGainVolumetric> distanceModel;
    float distanceToExtentOutline;
    int closestNormalIndex;
    float rearHemisphereWeight;
    float topHemisphereWeight;
    float bottomHemisphereWeight;
    bool integrated3DoFHoa;
    float ThreeDofHoaGain;
    float relativeHorizAngleExterior;
    float relativeVertAngleExterior;
    float relativeHorizAngleInterior;
    float relativeVertAngleInterior;
    std::vector<OutlineNormal> outlineNormals;
    float trans_fac[3];

    // Occlusion
    float occlusion_gain_left;
    float occlusion_gain_middle;
    float occlusion_gain_right;
    int eqnBands;
    std::vector<float> eq_centerFreqs;
    std::vector<float> occlusion_eq_top_left;
    std::vector<float> occlusion_eq_top_mid;
    std::vector<float> occlusion_eq_top_right;
    std::vector<float> occlusion_eq_bottom_left;
    std::vector<float> occlusion_eq_bottom_mid;
    std::vector<float> occlusion_eq_bottom_right;
    std::map<int, std::unique_ptr<std::vector<float>>> subareaEqMap;
    std::vector<float> speaker_occlusion_eq;
    std::vector<float> speaker_occlusion_eq_0;
    std::vector<float> speaker_occlusion_eq_1;
    std::vector<float> speaker_occlusion_eq_2;
    std::vector<float> speaker_occlusion_eq_4;
    std::vector<float> speaker_occlusion_eq_5;
    std::map<int, std::unique_ptr<std::vector<float>>> outputSpeakerEqMap;
    float occludedExtentGain;
    EAR::Position occludedExtentPos;
    int gridShiftCnt;
    Vector3 prevCentPointDir;
    Vector3 prevTopLeftPointDir;
    Vector3 prevTopRightPointDir;
    Vector3 prevBottomLeftPointDir;
    Vector3 prevBottomRightPointDir;
    float occlLeftAngleScaler;
    float occlRightAngleScaler;
    float occlTopAngleScaler;
    float occlBottomAngleScaler;
    CroppingOcclusionInfo cropOcclInfo;
    int noOfRaysSent;
    voxel_info::shared voxelInfo;
    Voxel::voxel_traversal voxelTraversal;
    int subareaOcclHorizRaysHigh;
    int subareaOcclVertRaysHigh;
    int subareaOcclHorizRaysLow;
    int subareaOcclVertRaysLow;
    float activeExtentRatio;
    bool occlusionDetected;

    // Heterogeneous rendering functions
    void FindExtentDescriptionPoints(const EAR::Geometry* extent, Vector3 listenerPosition);
    void UpdateExteriorRepresentationSpeakerSetup(EAR::Position listenerPos);
    void UpdateDescriptionPoint(float previousPos[3], float newPos[3], float maxPositionMovement,
        float temporalSmoothingFactor, float updatedPos[3], bool* pInitializedFlag);
    void CalculateEffectiveExtent(EAR::Position listenerPos);
    void UpdateInteriorRepresentationSpeakerSetup(
        const EAR::Geometry* extent, EAR::Position listenerPos);
    void CloseToSurfaceModificationsOfInteriorRepresentationRendering(
        const EAR::Geometry* extent, EAR::Position listenerPos);
    float CalculateDistanceGainCorrection(EAR::Position listenerPos);
    void UpdateOutputSpeakerActivation();
    void UpdateOutputSpeakerDistanceOffset(Vector3 listenerLocation);

    // Occlusion functions
    void RenderOcclusion(const EAR::Geometry* extent, std::unique_ptr<EAR::EmbreeTracer>& tracer,
        Vector3 listenerPosition, float* amountOfOcclusion);
    void CheckSubareaOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
        const EAR::Geometry* extent, Vector3 listenerPos, Vector3 topLeftPoint,
        Vector3 topRightPoint, Vector3 bottomLeftPoint, Vector3 bottomRightPoint,
        float subAreaMinAzimPos, float subAreaMaxAzimPos, float subAreaMinElevPos,
        float subAreaMaxElevPos, int numOfHorizDivisions, int numOfVertDivisions,
        std::vector<float>& occlusion_eq, int* numOfRays, int* numOfOccludedRays);
    float FindEdgeOfOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
        const EAR::Geometry* extent, Vector3 listenerPos, Vector3 topLeftPoint,
        Vector3 topRightPoint, Vector3 bottomLeftPoint, Vector3 bottomRightPoint,
        float startAzimPos, float startElevPos, float azimSearchRange, float elevSearchRange,
        float azimResolution, float elevResolution);
    bool CheckCroppingOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
        const EAR::Geometry* extent, Vector3 listenerPos, Vector3 topLeftPoint,
        Vector3 topRightPoint, Vector3 bottomLeftPoint, Vector3 bottomRightPoint, float azimPos,
        float elevPos, CroppingOcclusionInfo* cropOcIinf, float minAzimRange, float maxAzimRange,
        float minElevRange, float maxElevRange, OcclusionState* occlusionState);
    OcclusionState CheckRayCompleteOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
        const EAR::Geometry* extent, Vector3 listenerPos, EAR::Polar rayDir);
    OcclusionState CheckRayOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
        const EAR::Geometry* extent, Vector3 listenerPos, EAR::Polar rayDir,
        std::vector<float>& material_eq);
};

void CalculateHeterogeneousSourceOuputSignals(HeterogeneousDspState* outSignalState,
    EAR::RealFrame* inputSignal, std::vector<EAR::RealBuffer*> outputSignals, std::size_t B);

} // namespace MPEGI

#endif /* MPEGI_HETEROGENEOUS */
