/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Heterogeneous.h"
#include <MPEGI/Helpers/OcclusionRecovery.h>
#include <algorithm>

namespace MPEGI {

// Local functions
static void HetOclCalculatePointOnPlane(Vector3 topLeftPoint, Vector3 topRightPoint,
    Vector3 bottomLeftPoint, Vector3 bottomRightPoint, float azimPos, float elevPos,
    Vector3* pointOnPlane);

void HeterogeneousSource::RenderOcclusion(const EAR::Geometry* extent,
    std::unique_ptr<EAR::EmbreeTracer>& tracer, Vector3 listenerPos, float* amountOfOcclusion)
{
    EAR::Polar anchorDir = EAR::cartesianToPolar(anchorPoint - listenerPos);
    anchorDir[2] = 1000.5 * distanceToExtent;
    float eqIntF = 0.1;
    float active_extent_ratio = 1.0;
    float azimRange = maxAzim - minAzim;
    float elevRange = maxElev - minElev;

    noOfRaysSent = 0;

    if (distanceToExtent < 0.0) {
        // We are inside the extent, just interpolate the occlusion eq:s towards all ones
        for (int i = 0; i < eqnBands; i++) {
            occlusion_eq_top_left[i] = (1.0f - eqIntF) * occlusion_eq_top_left[i] + eqIntF;
            occlusion_eq_top_mid[i] = (1.0f - eqIntF) * occlusion_eq_top_mid[i] + eqIntF;
            occlusion_eq_top_right[i] = (1.0f - eqIntF) * occlusion_eq_top_right[i] + eqIntF;
            occlusion_eq_bottom_left[i] = (1.0f - eqIntF) * occlusion_eq_bottom_left[i] + eqIntF;
            occlusion_eq_bottom_mid[i] = (1.0f - eqIntF) * occlusion_eq_bottom_mid[i] + eqIntF;
            occlusion_eq_bottom_right[i] = (1.0f - eqIntF) * occlusion_eq_bottom_right[i] + eqIntF;
        }
        return;
    }

    // Calculate corner points of the simplified extent
    Vector3 displacementVecTop = (topEdgePoint - anchorPoint);
    Vector3 displacementVecBottom = (bottomEdgePoint - anchorPoint);
    Vector3 topLeftPoint = leftEdgePoint + displacementVecTop;
    Vector3 topRightPoint = rightEdgePoint + displacementVecTop;
    Vector3 bottomLeftPoint = leftEdgePoint + displacementVecBottom;
    Vector3 bottomRightPoint = rightEdgePoint + displacementVecBottom;

    // Calculate how many rays to cast in a first sparse grid
    int numOfAzimDivisions = fmin(HETEROGENEOUS_OCCL_GRID_MAX_AZIM_RAYS,
        fmax(
            HETEROGENEOUS_OCCL_GRID_MIN_AZIM_RAYS, azimRange / HETEROGENEOUS_OCCL_GRID_RESOLUTION));
    int numOfElevDivisions = fmin(HETEROGENEOUS_OCCL_GRID_MAX_ELEV_RAYS,
        fmax(
            HETEROGENEOUS_OCCL_GRID_MIN_ELEV_RAYS, elevRange / HETEROGENEOUS_OCCL_GRID_RESOLUTION));

    cropOcclInfo.minUnoccludedAzim = 1.0;
    cropOcclInfo.minUnoccludedAzimElev = 0.0;
    cropOcclInfo.maxUnoccludedAzim = 0.0;
    cropOcclInfo.maxUnoccludedAzimElev = 0.0;
    cropOcclInfo.minUnoccludedElev = 1.0;
    cropOcclInfo.minUnoccludedElevAzim = 0.0;
    cropOcclInfo.maxUnoccludedElev = 0.0;
    cropOcclInfo.maxUnoccludedElevAzim = 0.0;
    cropOcclInfo.occlusionDetected = false;
    cropOcclInfo.noOfRaysUsed = 0;
    cropOcclInfo.minUnoccludedAzimRange = -1.0f / (float)numOfAzimDivisions;
    cropOcclInfo.maxUnoccludedAzimRange = 1.0f / (float)numOfAzimDivisions;
    cropOcclInfo.minUnoccludedElevRange = -1.0f / (float)numOfElevDivisions;
    cropOcclInfo.maxUnoccludedElevRange = 1.0f / (float)numOfElevDivisions;

    OcclusionState occlusionState = OcclusionState::not_occluded;
    int azimShiftLim = 10;
    float gridShift = (float)gridShiftCnt / azimShiftLim;

    for (int v = 0; v < numOfElevDivisions; v++) {
        float elevPos = 0.01
            + 0.98 * (gridShift / (float)numOfElevDivisions + (float)v / (float)numOfElevDivisions);
        for (int h = 0; h < numOfAzimDivisions; h++) {
            float azimPos = 0.01
                + 0.98
                    * (gridShift / (float)numOfAzimDivisions
                        + (float)h / (float)numOfAzimDivisions);

            CheckCroppingOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
                bottomLeftPoint, bottomRightPoint, azimPos, elevPos, &cropOcclInfo,
                -2.0f / (float)numOfAzimDivisions, 2.0f / (float)numOfAzimDivisions,
                -2.0f / (float)numOfElevDivisions, 2.0f / (float)numOfElevDivisions,
                &occlusionState);
        }
    }

    // Update the shifting of the sparse grid
    gridShiftCnt++;
    if (gridShiftCnt > azimShiftLim) {
        gridShiftCnt = 0;
    }

    // Check points of the previous cropped extent
    CheckCroppingOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
        bottomLeftPoint, bottomRightPoint, prevCentPointDir[0], prevCentPointDir[1], &cropOcclInfo,
        -prevCentPointDir[0], 1.0 - prevCentPointDir[0], -prevCentPointDir[1],
        1.0 - prevCentPointDir[1], &occlusionState);
    CheckCroppingOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
        bottomLeftPoint, bottomRightPoint, prevTopLeftPointDir[0], prevTopLeftPointDir[1],
        &cropOcclInfo, -prevTopLeftPointDir[0], 1.0 - prevTopLeftPointDir[0],
        -prevTopLeftPointDir[1], 1.0 - prevTopLeftPointDir[1], &occlusionState);
    CheckCroppingOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
        bottomLeftPoint, bottomRightPoint, prevTopRightPointDir[0], prevTopRightPointDir[1],
        &cropOcclInfo, -prevTopRightPointDir[0], 1.0 - prevTopRightPointDir[0],
        -prevTopRightPointDir[1], 1.0 - prevTopRightPointDir[1], &occlusionState);
    CheckCroppingOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
        bottomLeftPoint, bottomRightPoint, prevBottomLeftPointDir[0], prevBottomLeftPointDir[1],
        &cropOcclInfo, -prevBottomLeftPointDir[0], 1.0 - prevBottomLeftPointDir[0],
        -prevBottomLeftPointDir[1], 1.0 - prevBottomLeftPointDir[1], &occlusionState);
    CheckCroppingOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
        bottomLeftPoint, bottomRightPoint, prevBottomRightPointDir[0], prevBottomRightPointDir[1],
        &cropOcclInfo, -prevBottomRightPointDir[0], 1.0 - prevBottomRightPointDir[0],
        -prevBottomRightPointDir[1], 1.0 - prevBottomRightPointDir[1], &occlusionState);

    // Store edge points as starting points for the next round
    prevTopLeftPointDir.set(
        cropOcclInfo.minUnoccludedAzim, cropOcclInfo.minUnoccludedAzimElev, 0.0);
    prevTopRightPointDir.set(
        cropOcclInfo.maxUnoccludedAzim, cropOcclInfo.maxUnoccludedAzimElev, 0.0);
    prevBottomLeftPointDir.set(
        cropOcclInfo.minUnoccludedElevAzim, cropOcclInfo.minUnoccludedElev, 0.0);
    prevBottomRightPointDir.set(
        cropOcclInfo.maxUnoccludedElevAzim, cropOcclInfo.maxUnoccludedElev, 0.0);

    // Refine the cropped extent with a binary search for each edge
    if (cropOcclInfo.occlusionDetected) {
        float refinedMaxUnoccludedAzimPos = 0.0f;
        float refinedMinUnoccludedAzimPos = 0.0f;
        float refinedMaxUnoccludedElevPos = 0.0f;
        float refinedMinUnoccludedElevPos = 0.0f;
        if (cropOcclInfo.maxUnoccludedAzimRange != 0) {
            refinedMaxUnoccludedAzimPos = FindEdgeOfOcclusion(tracer, extent, listenerPos,
                topLeftPoint, topRightPoint, bottomLeftPoint, bottomRightPoint,
                cropOcclInfo.maxUnoccludedAzim, cropOcclInfo.maxUnoccludedAzimElev,
                cropOcclInfo.maxUnoccludedAzimRange, 0.0f, 0.001f, 0.001f);
            cropOcclInfo.maxUnoccludedAzim = fmin(1.0f, refinedMaxUnoccludedAzimPos);
        }
        if (cropOcclInfo.minUnoccludedAzimRange != 0) {
            refinedMinUnoccludedAzimPos = FindEdgeOfOcclusion(tracer, extent, listenerPos,
                topLeftPoint, topRightPoint, bottomLeftPoint, bottomRightPoint,
                cropOcclInfo.minUnoccludedAzim, cropOcclInfo.minUnoccludedAzimElev,
                cropOcclInfo.minUnoccludedAzimRange, 0.0f, 0.001f, 0.001f);
            cropOcclInfo.minUnoccludedAzim = fmax(0.0, refinedMinUnoccludedAzimPos);
        }
        if (cropOcclInfo.minUnoccludedElevRange != 0) {
            refinedMinUnoccludedElevPos = FindEdgeOfOcclusion(tracer, extent, listenerPos,
                topLeftPoint, topRightPoint, bottomLeftPoint, bottomRightPoint,
                cropOcclInfo.minUnoccludedElevAzim, cropOcclInfo.minUnoccludedElev, 0.0f,
                cropOcclInfo.minUnoccludedElevRange, 0.001f, 0.001f);
            cropOcclInfo.minUnoccludedElev = fmax(0.0f, refinedMinUnoccludedElevPos);
        }
        if (cropOcclInfo.maxUnoccludedElevRange != 0) {
            refinedMaxUnoccludedElevPos = FindEdgeOfOcclusion(tracer, extent, listenerPos,
                topLeftPoint, topRightPoint, bottomLeftPoint, bottomRightPoint,
                cropOcclInfo.maxUnoccludedElevAzim, cropOcclInfo.maxUnoccludedElev, 0.0f,
                cropOcclInfo.maxUnoccludedElevRange, 0.001f, 0.001f);
            cropOcclInfo.maxUnoccludedElev = fmin(1.0f, refinedMaxUnoccludedElevPos);
        }
    } else {
        // No occlusion at all
        cropOcclInfo.minUnoccludedAzim = 0.0;
        cropOcclInfo.maxUnoccludedAzim = 1.0;
        cropOcclInfo.minUnoccludedElev = 0.0;
        cropOcclInfo.maxUnoccludedElev = 1.0;
    }

    // Store the middle point of the cropped extent as starting point for the next round
    prevCentPointDir.set((cropOcclInfo.maxUnoccludedAzim + cropOcclInfo.minUnoccludedAzim) * 0.5,
        (cropOcclInfo.maxUnoccludedElev + cropOcclInfo.minUnoccludedElev) * 0.5, 0.0);

    // Update left/right and top/bottom edges according to the cropping due to occlusion
    // If minUnoccludedAzim > maxUnoccludedAzim or minUnoccludedAzim < maxUnoccludedAzim
    // no rays reached the extent

    if ((azimRange > 0.0) && (elevRange > 0.0) && cropOcclInfo.occlusionDetected) {
        if ((cropOcclInfo.minUnoccludedAzim <= cropOcclInfo.maxUnoccludedAzim)
            && (cropOcclInfo.minUnoccludedElev <= cropOcclInfo.maxUnoccludedElev)) {

            Vector3 newTopLeftPoint;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, cropOcclInfo.minUnoccludedAzim, cropOcclInfo.maxUnoccludedElev,
                &newTopLeftPoint);
            Vector3 newTopRightPoint;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, cropOcclInfo.maxUnoccludedAzim, cropOcclInfo.maxUnoccludedElev,
                &newTopRightPoint);
            Vector3 newBottomLeftPoint;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, cropOcclInfo.minUnoccludedAzim, cropOcclInfo.minUnoccludedElev,
                &newBottomLeftPoint);
            Vector3 newBottomRightPoint;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, cropOcclInfo.maxUnoccludedAzim, cropOcclInfo.minUnoccludedElev,
                &newBottomRightPoint);

            // Update edge points
            leftEdgePoint = 0.5 * (newTopLeftPoint + newBottomLeftPoint);
            rightEdgePoint = 0.5 * (newTopRightPoint + newBottomRightPoint);
            topEdgePoint = 0.5 * (newTopLeftPoint + newTopRightPoint);
            bottomEdgePoint = 0.5 * (newBottomLeftPoint + newBottomRightPoint);

            topLeftPoint = newTopLeftPoint;
            topRightPoint = newTopRightPoint;
            bottomLeftPoint = newBottomLeftPoint;
            bottomRightPoint = newBottomRightPoint;

            // Calculate how big part of the original extent that was non-occluded
            active_extent_ratio = (cropOcclInfo.maxUnoccludedAzim - cropOcclInfo.minUnoccludedAzim)
                * (cropOcclInfo.maxUnoccludedElev - cropOcclInfo.minUnoccludedElev);

            // Calculate modified angles for the virtual microphones used to capture
            // sound representing the modified extent
            occlLeftAngleScaler = fmax(0.5f, 2.0f * (1.0f - cropOcclInfo.minUnoccludedAzim) - 1.0f);
            occlRightAngleScaler = fmax(0.5f, 2.0f * (cropOcclInfo.maxUnoccludedAzim) - 1.0f);
            occlTopAngleScaler = fmax(0.5f, 2.0f * (cropOcclInfo.maxUnoccludedElev) - 1.0f);
            occlBottomAngleScaler =
                fmax(0.5f, 2.0f * (1.0f - cropOcclInfo.minUnoccludedElev) - 1.0f);

            minAzim = (1.0f - cropOcclInfo.minUnoccludedAzim) * minAzim
                + cropOcclInfo.minUnoccludedAzim * maxAzim;
            maxAzim = (1.0f - cropOcclInfo.maxUnoccludedAzim) * minAzim
                + cropOcclInfo.maxUnoccludedAzim * maxAzim;
            minElev = (1.0f - cropOcclInfo.minUnoccludedElev) * minElev
                + cropOcclInfo.minUnoccludedElev * maxElev;
            maxElev = (1.0f - cropOcclInfo.maxUnoccludedElev) * minElev
                + cropOcclInfo.maxUnoccludedElev * maxElev;
            azimRange = maxAzim - minAzim;
            elevRange = maxElev - minElev;
        } else {
            // The whole extent is occluded
            active_extent_ratio = 0.0;
        }
    }

    if ((azimRange > 0.0) && (elevRange > 0.0)) {

        int horizRays = fmax(1.0, azimRange / HETEROGENEOUS_OCCL_SUB_AREA_RESOLUTION);
        int vertRays = fmax(1.0, elevRange / HETEROGENEOUS_OCCL_SUB_AREA_RESOLUTION);

        if (cropOcclInfo.occlusionDetected) {
            horizRays = fmin(horizRays, subareaOcclHorizRaysHigh);
            vertRays = fmin(vertRays, subareaOcclVertRaysHigh);
        } else {
            // Occlusion was not detected in the near past, use low resolution
            horizRays = subareaOcclHorizRaysLow;
            vertRays = subareaOcclVertRaysLow;
        }

        // Initialize the occlusion_eq:s to zeroes
        for (int i = 0; i < eqnBands; i++) {
            occlusion_eq_top_left[i] = 1e-6;
            occlusion_eq_top_mid[i] = 1e-6;
            occlusion_eq_top_right[i] = 1e-6;
            occlusion_eq_bottom_left[i] = 1e-6;
            occlusion_eq_bottom_mid[i] = 1e-6;
            occlusion_eq_bottom_right[i] = 1e-6;
        }

        float meanElev = 0.5f;
        float leftDivAngle = 1.0f / 3.0f;
        float rightDivAngle = 2.0f / 3.0f;

        int numOfRaysTopLeft, numOfRaysTopMid, numOfRaysTopRight, numOfRaysBottomLeft,
            numOfRaysBottomMid, numOfRaysBottomRight;
        int numOfOcclusionsTopLeft, numOfOcclusionsTopMid, numOfOcclusionsTopRight,
            numOfOcclusionsBottomLeft, numOfOcclusionsBottomMid, numOfOcclusionsBottomRight;

        CheckSubareaOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
            bottomLeftPoint, bottomRightPoint, 0.0, leftDivAngle, meanElev, 1.0f, 10, 10,
            occlusion_eq_top_left, &numOfRaysTopLeft, &numOfOcclusionsTopLeft);
        CheckSubareaOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
            bottomLeftPoint, bottomRightPoint, leftDivAngle, rightDivAngle, meanElev, 1.0f, 10, 10,
            occlusion_eq_top_mid, &numOfRaysTopMid, &numOfOcclusionsTopMid);
        CheckSubareaOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
            bottomLeftPoint, bottomRightPoint, rightDivAngle, 1.0, meanElev, 1.0f, 10, 10,
            occlusion_eq_top_right, &numOfRaysTopRight, &numOfOcclusionsTopRight);
        CheckSubareaOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
            bottomLeftPoint, bottomRightPoint, 0.0f, leftDivAngle, 0.0f, meanElev, 10, 10,
            occlusion_eq_bottom_left, &numOfRaysBottomLeft, &numOfOcclusionsBottomLeft);
        CheckSubareaOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
            bottomLeftPoint, bottomRightPoint, leftDivAngle, rightDivAngle, 0.0f, meanElev, 10, 10,
            occlusion_eq_bottom_mid, &numOfRaysBottomMid, &numOfOcclusionsBottomMid);
        CheckSubareaOcclusion(tracer, extent, listenerPos, topLeftPoint, topRightPoint,
            bottomLeftPoint, bottomRightPoint, rightDivAngle, 1.0f, 0.0f, meanElev, 10, 10,
            occlusion_eq_bottom_right, &numOfRaysBottomRight, &numOfOcclusionsBottomRight);

        // Scale the eq:s due to the cropping occlusion
        float eq_scaler = (active_extent_ratio > 0.0) ? sqrt(active_extent_ratio) : 0.0;
        for (int i = 0; i < eqnBands; i++) {
            occlusion_eq_top_left[i] = eq_scaler * occlusion_eq_top_left[i];
            occlusion_eq_top_mid[i] = eq_scaler * occlusion_eq_top_mid[i];
            occlusion_eq_top_right[i] = eq_scaler * occlusion_eq_top_right[i];
            occlusion_eq_bottom_left[i] = eq_scaler * occlusion_eq_bottom_left[i];
            occlusion_eq_bottom_mid[i] = eq_scaler * occlusion_eq_bottom_mid[i];
            occlusion_eq_bottom_right[i] = eq_scaler * occlusion_eq_bottom_right[i];
        }

        float minEqValue = 1.0;
        minEqValue = fmin(minEqValue,
            *std::min_element(occlusion_eq_top_left.begin(), occlusion_eq_top_left.end()));
        minEqValue = fmin(minEqValue,
            *std::min_element(occlusion_eq_top_mid.begin(), occlusion_eq_top_mid.end()));
        minEqValue = fmin(minEqValue,
            *std::min_element(occlusion_eq_top_right.begin(), occlusion_eq_top_right.end()));
        minEqValue = fmin(minEqValue,
            *std::min_element(occlusion_eq_bottom_left.begin(), occlusion_eq_bottom_left.end()));
        minEqValue = fmin(minEqValue,
            *std::min_element(occlusion_eq_bottom_left.begin(), occlusion_eq_bottom_left.end()));
        minEqValue = fmin(minEqValue,
            *std::min_element(occlusion_eq_bottom_right.begin(), occlusion_eq_bottom_right.end()));
        if (minEqValue < 1.0) {
            occlusionDetected = true;
        } else {
            occlusionDetected = false;
        }

        float totNumOfRays = numOfRaysTopLeft + numOfRaysTopMid + numOfRaysTopRight
            + numOfRaysBottomLeft + numOfRaysBottomMid + numOfRaysBottomRight;
        float subareasOcclusionRatio = 0.0f;
        if (totNumOfRays > 0) {
            subareasOcclusionRatio = (float)(numOfOcclusionsTopLeft + numOfOcclusionsTopMid
                                         + numOfOcclusionsTopRight + numOfOcclusionsBottomLeft
                                         + numOfOcclusionsBottomMid + numOfOcclusionsBottomRight)
                / ((float)totNumOfRays);
        }

        // Calculate gain and position of the occluded part of the extent, that might generate diffraction
        activeExtentRatio = active_extent_ratio * (1.0f - subareasOcclusionRatio);
        occludedExtentGain = (activeExtentRatio < 1.0) ? sqrt(1.0f - activeExtentRatio) : 0.0;
        occludedExtentPos.location = anchorPoint;

        *amountOfOcclusion = activeExtentRatio;
    }
}

OcclusionState HeterogeneousSource::CheckRayOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
    const EAR::Geometry* extent, Vector3 listenerPos, EAR::Polar rayDir,
    std::vector<float>& material_eq)
{
    int fid = 0;

    bool tracingFinished = (voxelInfo != nullptr);
    float rayLength = 0.0f;
    Vector3 startingPoint = listenerPos;
    OcclusionState occlusionStatus = OcclusionState::not_occluded;

    for (int i = 0; i < eqnBands; i++) {
        material_eq[i] = 1.0f;
    }

    if (voxelInfo) {
        auto const hit = tracer->firstHit(listenerPos, rayDir, false);
        if (hit != extent->getMesh()) {
            return OcclusionState::outside_extent;
        }

        rayDir[2] = tracer->getDistanceToLastHit();
        auto const dest = listenerPos + EAR::polarToCartesian(rayDir);
        auto const& grid = *voxelInfo->grid;
        auto const& scene_info = *voxelInfo->sceneInfo;
        auto const& material_map = voxelInfo->material_map;

        using hit_list_t = std::vector<std::pair<int, Voxel::voxel_index_t>>;
        hit_list_t hits;

        if (!voxelTraversal.run(scene_info, listenerPos, dest, [&](auto const& idx) {
                auto const mat_id = grid.at(idx).get_material_id();
                if (mat_id) {
                    hits.emplace_back(mat_id, idx);

                    occlusionStatus = OcclusionState::occluded;

                    auto trans_it = material_map.find(mat_id);
                    // Invalid bitstream. Should already fail earlier
                    EAR_CHECK(trans_it != std::cend(material_map),
                        "Invalid voxel material ID: " + std::to_string(mat_id));

                    auto const& trans = trans_it->second.trans;
                    EAR_CHECK(trans.size() == material_eq.size(), "Number of EQ bands mismatch");

                    std::transform(std::begin(material_eq), std::end(material_eq),
                        std::begin(trans), std::begin(material_eq),
                        [](auto const& lhs, auto const& rhs) { return std::min(rhs, lhs); });

                    if (!std::accumulate(std::cbegin(material_eq), std::cend(material_eq), 0)) {
                        return false;
                    }
                }

                return true;
            })) {
            // No path found - should never happen
            occlusionStatus = OcclusionState::occluded;
            material_eq.assign(material_eq.size(), 0.f);
        }

        // Apply occlusion recovery only if exactly one material was hit
        if (hits.size() == 1) {
            auto const& [mat_id, idx] = hits.front();

            if (voxelInfo->recovery->has_recovery(mat_id)) {
                // Location of the occluding voxel
                auto const occluder_xyz =
                    EAR::Vector3(Voxel::vox_to_opengl(scene_info.idx2xyz(idx)).data());

                // Distance from listener to occluding voxel (squared)
                auto const listener_to_occluder = occluder_xyz - listenerPos;
                auto const sqr_distance = listener_to_occluder.dot(listener_to_occluder);

                voxelInfo->recovery->apply_recovery(mat_id, sqr_distance, material_eq);
            }
        }
    }

    while (!tracingFinished) {
        const EAR::Mesh* rayhitMesh = tracer->firstHit(startingPoint, rayDir, false);
        noOfRaysSent++;

        if (rayhitMesh) {
            rayLength = tracer->getDistanceToLastHit();
            fid = tracer->getPrimitiveOfLastHit();
            if (fid >= 0) {
                if (rayhitMesh == extent->getMesh()) {
                    // We hit the extent of the source
                    tracingFinished = true;
                } else {
                    auto mat = rayhitMesh->getTriangle(fid).material;
                    if (mat) {
                        std::vector<float> transCoeffs = mat->getTransmissionCoeffs();
                        auto minElement = std::min_element(transCoeffs.begin(), transCoeffs.end());
                        // If the material was acoustically transparent, just continue
                        if (*minElement < 1.0) {
                            occlusionStatus = OcclusionState::occluded;
                            if (material_eq.size() > 0) {
                                for (int i = 0; i < eqnBands; i++) {
                                    material_eq[i] *=
                                        mat->getTransmissionCoeffForFreq(eq_centerFreqs[i]);
                                }
                            }
                        }
                    }
                    startingPoint = startingPoint
                        + EAR::polarToCartesian(
                            EAR::Polar(rayDir[0], rayDir[1], rayLength + 0.0001f));
                }
            }
        } else {
            // Never hit the extent of the source
            occlusionStatus = OcclusionState::outside_extent;
            tracingFinished = true;
        }
    }

    return occlusionStatus;
}

OcclusionState HeterogeneousSource::CheckRayCompleteOcclusion(
    std::unique_ptr<EAR::EmbreeTracer>& tracer, const EAR::Geometry* extent, Vector3 listenerPos,
    EAR::Polar rayDir)
{
    std::vector<float> material_eq;
    material_eq.resize(eqnBands);
    OcclusionState rayOcclusionState = OcclusionState::not_occluded;
    float materialEqEnergy = 0.0f;

    rayOcclusionState = CheckRayOcclusion(tracer, extent, listenerPos, rayDir, material_eq);
    if (rayOcclusionState == OcclusionState::occluded) {
        materialEqEnergy = std::accumulate(material_eq.begin(), material_eq.end(), 0.0);
    }

    if (rayOcclusionState == OcclusionState::not_occluded) {
        // Not occluded
        return OcclusionState::not_occluded;
    } else if (rayOcclusionState == OcclusionState::occluded && materialEqEnergy > 0.0) {
        // Soft occlusion, so not completely occluded
        return OcclusionState::not_occluded;
    } else if (rayOcclusionState == OcclusionState::outside_extent) {
        // The ray did not hit the extent, count as non-occluded
        return OcclusionState::outside_extent;
    } else {
        return OcclusionState::occluded;
    }
}

float HeterogeneousSource::FindEdgeOfOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
    const EAR::Geometry* extent, Vector3 listenerPos, Vector3 topLeftPoint, Vector3 topRightPoint,
    Vector3 bottomLeftPoint, Vector3 bottomRightPoint, float startAzimPos, float startElevPos,
    float azimSearchRange, float elevSearchRange, float azimResolution, float elevResolution)
{
    float testPos = 0.0f;
    float bound1 = 0.0f;
    float bound2 = 0.0f;
    OcclusionState occluded = OcclusionState::not_occluded;
    float boundedRange = 0.0f;
    float testAzimPos = 0.0f;
    float testElevPos = 0.0f;

    // Check input parameters
    if ((azimSearchRange != 0.0f) && (elevSearchRange != 0.0f)) {
        EAR_FAIL("HeterogeneousSource::FindEdgeOfOcclusion: Cannot search in both azim and elev!");
        return 0.0;
    }
    if (azimSearchRange != 0.0f) {
        // Search for the edge in azimuth dimension
        bound1 = 0.0f;
        bound2 = azimSearchRange;
        testPos = (bound1 + bound2) / 2.0f;
        boundedRange = fabs(bound2 - bound1);
        while (boundedRange > azimResolution) {
            testAzimPos = startAzimPos + testPos;
            testElevPos = startElevPos;
            Vector3 pointOnExt;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, testAzimPos, testElevPos, &pointOnExt);
            Vector3 rayDir = pointOnExt - listenerPos;
            Polar rayDirPolar = cartesianToPolar(rayDir);

            occluded = CheckRayCompleteOcclusion(tracer, extent, listenerPos, rayDirPolar);
            if (occluded == OcclusionState::occluded) {
                // We found a point that was occluded, now test the middle point of this and bound1
                bound2 = testPos;
            } else if (occluded == OcclusionState::not_occluded) {
                // We found a point that was not occluded, now test the middle point of this and bound2
                bound1 = testPos;
            } else {
                // The point did not hit the extent, try closer to start pos
                bound2 = testPos;
            }
            testPos = (bound1 + bound2) / 2.0f;
            boundedRange = fabs(bound2 - bound1);
        }
        return startAzimPos + testPos;
    } else if (elevSearchRange != 0.0f) {
        // Search for the edge in elevation dimension
        bound1 = 0.0f;
        bound2 = elevSearchRange;
        testPos = (bound1 + bound2) / 2.0f;
        boundedRange = fabs(bound2 - bound1);
        while (boundedRange > elevResolution) {
            testAzimPos = startAzimPos;
            testElevPos = startElevPos + testPos;
            Vector3 pointOnExt;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, testAzimPos, testElevPos, &pointOnExt);
            Vector3 rayDir = pointOnExt - listenerPos;
            Polar rayDirPolar = cartesianToPolar(rayDir);
            occluded = CheckRayCompleteOcclusion(tracer, extent, listenerPos, rayDirPolar);
            if (occluded == OcclusionState::occluded) {
                // We found a point that was occluded, now test the middle point of this and bound1
                bound2 = testPos;
            } else if (occluded == OcclusionState::not_occluded) {
                // We found a point that was not occluded, now test the middle point of this and bound2
                bound1 = testPos;
            } else {
                // The point did not hit the extent, try closer to start pos
                bound2 = testPos;
            }
            testPos = (bound1 + bound2) / 2.0f;
            boundedRange = fabs(bound2 - bound1);
        }
        return startElevPos + testPos;
    }

    // Search range is zero
    return 0.0;
}

void HeterogeneousSource::CheckSubareaOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
    const EAR::Geometry* extent, Vector3 listenerPos, Vector3 topLeftPoint, Vector3 topRightPoint,
    Vector3 bottomLeftPoint, Vector3 bottomRightPoint, float subAreaMinAzimPos,
    float subAreaMaxAzimPos, float subAreaMinElevPos, float subAreaMaxElevPos,
    int numOfHorizDivisions, int numOfVertDivisions, std::vector<float>& occlusion_eq,
    int* numOfRays, int* numOfOccludedRays)
{
    std::vector<float> material_eq;
    material_eq.resize(occlusion_eq.size());
    std::fill(occlusion_eq.begin(), occlusion_eq.end(), 0.0);
    int numOfOcclusions = 0;
    int TotNumOfSubareaRays = 0;
    float azimOffset = 0;

    if ((numOfHorizDivisions < 1) || (numOfVertDivisions < 1)) {
        return;
    }

    // Send a grid of rays towards the subarea and accumulate any occlusion EQs
    for (int elevLoop = 0; elevLoop < numOfVertDivisions; elevLoop++) {

        // Add an offset to the azimuth so we don't check the same azimuth many times
        azimOffset = (float)elevLoop / (float)numOfVertDivisions / (float)numOfHorizDivisions;

        for (int azimLoop = 0; azimLoop < numOfHorizDivisions; azimLoop++) {
            float subAreaAzimPos = (float)azimLoop / numOfHorizDivisions + azimOffset;
            float subAreaVertPos = (float)elevLoop / numOfVertDivisions;
            float azimPos =
                subAreaMinAzimPos * (1.0f - subAreaAzimPos) + subAreaMaxAzimPos * subAreaAzimPos;
            float elevPos =
                subAreaMinElevPos * (1.0f - subAreaVertPos) + subAreaMaxElevPos * subAreaVertPos;

            Vector3 pointOnExt;
            HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint,
                bottomRightPoint, azimPos, elevPos, &pointOnExt);

            Vector3 rayDir = pointOnExt - listenerPos;
            Polar rayDirPolar = cartesianToPolar(rayDir);

            OcclusionState rayOcclusionState =
                CheckRayOcclusion(tracer, extent, listenerPos, rayDirPolar, material_eq);
            if (rayOcclusionState == OcclusionState::occluded) {
                for (int i = 0; i < eqnBands; i++) {
                    occlusion_eq[i] += material_eq[i] * material_eq[i];
                }
                numOfOcclusions++;
                TotNumOfSubareaRays++;
            } else if (rayOcclusionState == OcclusionState::not_occluded) {
                for (int i = 0; i < eqnBands; i++) {
                    occlusion_eq[i] += 1.0;
                }
                TotNumOfSubareaRays++;
            }
        }
    }

    if (TotNumOfSubareaRays > 0) {
        for (int i = 0; i < eqnBands; i++) {
            // We assume a diffuse sound source, therefor use the square root here
            occlusion_eq[i] = sqrt(occlusion_eq[i] / (float)TotNumOfSubareaRays);
        }
    } else {
        for (int i = 0; i < eqnBands; i++) {
            occlusion_eq[i] = 1.0;
        }
    }

    *numOfRays = TotNumOfSubareaRays;
    *numOfOccludedRays = numOfOcclusions;
}

bool HeterogeneousSource::CheckCroppingOcclusion(std::unique_ptr<EAR::EmbreeTracer>& tracer,
    const EAR::Geometry* extent, Vector3 listenerPos, Vector3 topLeftPoint, Vector3 topRightPoint,
    Vector3 bottomLeftPoint, Vector3 bottomRightPoint, float azimPos, float elevPos,
    CroppingOcclusionInfo* cropOcIinf, float minAzimRange, float maxAzimRange, float minElevRange,
    float maxElevRange, OcclusionState* occlusionState)
{
    bool pointWasUsed = false;
    OcclusionState rayOcclusionState = OcclusionState::not_occluded;
    std::vector<float> material_eq;
    material_eq.resize(eqnBands);

    if ((azimPos < 0.0) || (azimPos > 1.0) || (elevPos < 0.0) || (elevPos > 1.0)) {
        // Outside of the extent
        return false;
    }

    Vector3 pointOnSimpExt;
    HetOclCalculatePointOnPlane(topLeftPoint, topRightPoint, bottomLeftPoint, bottomRightPoint,
        azimPos, elevPos, &pointOnSimpExt);

    Vector3 rayDir = pointOnSimpExt - listenerPos;
    Polar rayDirPolar = cartesianToPolar(rayDir);

    // Check if this point is completely occluded. If not, check if this point represents
    // the min/max azimuth or elevation that is not completely occluded.
    rayOcclusionState = CheckRayOcclusion(tracer, extent, listenerPos, rayDirPolar, material_eq);
    float materialEqEnergy = std::accumulate(material_eq.begin(), material_eq.end(), 0.0);
    if ((rayOcclusionState == OcclusionState::not_occluded)
        || ((rayOcclusionState == OcclusionState::occluded) && materialEqEnergy > 0.0)) {

        if (azimPos < cropOcIinf->minUnoccludedAzim) {
            cropOcIinf->minUnoccludedAzim = azimPos;
            cropOcIinf->minUnoccludedAzimElev = elevPos;
            cropOcIinf->minUnoccludedAzimRange = minAzimRange;
            pointWasUsed = true;
        }
        if (azimPos > cropOcIinf->maxUnoccludedAzim) {
            cropOcIinf->maxUnoccludedAzim = azimPos;
            cropOcIinf->maxUnoccludedAzimElev = elevPos;
            cropOcIinf->maxUnoccludedAzimRange = maxAzimRange;
            pointWasUsed = true;
        }
        if (elevPos < cropOcIinf->minUnoccludedElev) {
            cropOcIinf->minUnoccludedElev = elevPos;
            cropOcIinf->minUnoccludedElevAzim = azimPos;
            cropOcIinf->minUnoccludedElevRange = minElevRange;
            pointWasUsed = true;
        }
        if (elevPos > cropOcIinf->maxUnoccludedElev) {
            cropOcIinf->maxUnoccludedElev = elevPos;
            cropOcIinf->maxUnoccludedElevAzim = azimPos;
            cropOcIinf->maxUnoccludedElevRange = maxElevRange;
            pointWasUsed = true;
        }
    }
    *occlusionState = rayOcclusionState;

    if (pointWasUsed) {
        cropOcIinf->noOfRaysUsed++;
    }

    if (rayOcclusionState == OcclusionState::occluded) {
        cropOcIinf->occlusionDetected = true;
    }

    return pointWasUsed;
}

float* HeterogeneousSource::GetOcclusionEqData(int speakerInd)
{

    if ((speakerInd == 0) || (speakerInd == 1) || (speakerInd == 2) || (speakerInd == 4)
        || (speakerInd == 5)) {
        // Return an EQ that is calculated as a weighted sum of the sub area EQs that the
        // output speaker is associated with.
        for (int eqBand = 0; eqBand < eqnBands; eqBand++) {
            speaker_occlusion_eq[eqBand] = 0.0f;
        }

        // Calculate the weight of the mapping corresponding the plane, line and point speaker setups
        // This is controlled by the transition variables.

        // Restrict the vertical transition so that it will not exceed the horizontal transition
        float const_fac_1 = fmin(trans_fac[1], trans_fac[0]);
        float planeWeight = const_fac_1 * trans_fac[0];
        float lineWeight = (1.0f - const_fac_1) * trans_fac[0];
        float pointWeight = 1.0f - trans_fac[0];

        for (int eqBand = 0; eqBand < eqnBands; eqBand++) {

            if (occlusionDetected) {
                for (int subareaInd = 0; subareaInd < HETEROGENEOUS_NUM_OF_SUBAREAS; subareaInd++) {
                    speaker_occlusion_eq[eqBand] +=
                        (planeWeight * subareaEqToSpeakerEqMapPlane.at(speakerInd)[subareaInd]
                            + lineWeight * subareaEqToSpeakerEqMapLine.at(speakerInd)[subareaInd]
                            + pointWeight * subareaEqToSpeakerEqMapPoint.at(speakerInd)[subareaInd])
                        * subareaEqMap.at(subareaInd)->at(eqBand);
                }
            } else {
                // No occlusion, just set occlusion EQs to 1s
                speaker_occlusion_eq[eqBand] = 1.0f;
            }

            // Update the EQ of the output speaker with smoothing
            outputSpeakerEqMap.at(speakerInd)->at(eqBand) =
                HETEROGENEOUS_EQ_SMOOTH * outputSpeakerEqMap.at(speakerInd)->at(eqBand)
                + (1.0 - HETEROGENEOUS_EQ_SMOOTH) * speaker_occlusion_eq[eqBand];
        }
        return outputSpeakerEqMap.at(speakerInd)->data();
    } else {
        // This output speaker is not used for the exterior representation and has no occlusion EQ.
        // Return a "fully open" EQ just to avoid potential problems.
        for (int eqBand = 0; eqBand < eqnBands; eqBand++) {
            speaker_occlusion_eq[eqBand] = 1.0f;
        }
        return speaker_occlusion_eq.data();
    }
};

// Helper function that will calculated the world coordinates of a point on a plane
// given the corner points of the plane and the local coordinates of the point inside
// the plane given in the range [0.0, 1.0]. Azimuth is defined to increase to the right
// in this calculation.
static void HetOclCalculatePointOnPlane(Vector3 topLeftPoint, Vector3 topRightPoint,
    Vector3 bottomLeftPoint, Vector3 bottomRightPoint, float azimPos, float elevPos,
    Vector3* pointOnPlane)
{
    Vector3 azimPosAlongTopEdge = (1.0f - azimPos) * topLeftPoint + azimPos * topRightPoint;
    Vector3 azimPosAlongBottomEdge =
        (1.0f - azimPos) * bottomLeftPoint + azimPos * bottomRightPoint;
    *pointOnPlane = (1.0f - elevPos) * azimPosAlongBottomEdge + elevPos * azimPosAlongTopEdge;
}

} // namespace MPEGI
