/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "HeterogeneousStage.h"

#include <MPEGI/Helpers/OcclusionRecovery.h>
#include <MPEGI/Helpers/Types.h>

#include <EAR/BaseTypes.h>
#include <EAR/CICP_Layouts.h>
#include <EAR/HOASource.h>
#include <EAR/PointSource.h>

namespace MPEGI {

std::shared_ptr<RenderStage> makeHeterogeneousStage()
{
    return std::make_shared<HeterogeneousStage>();
}

HeterogeneousStage::HeterogeneousStage()
    : B(0)
    , itemStore(NULL)
{}

HeterogeneousStage::~HeterogeneousStage() {}

bool isMphoaSource(HOASource const& source)
{
    return source.getGroup() != -1
        && source.getRepresentation() == HOASource::Representation::Interior;
}

bool isMphoaRenderItem(RenderItem const& item)
{
    if (item.type != ItemType::HOA) {
        return false;
    }

    EAR_CHECK_REL(item.hoaInfo->source, "HeterogeneousStage: HOA RenderItem is missing HOASource");
    return isMphoaSource(*item.hoaInfo->source);
}

std::unique_ptr<RenderStage::DspData> HeterogeneousStage::update(RenderList& renderList)
{
    if (B == 0) { // only update once initialized
        return {};
    }
    EAR::Position pos;
    pos.cSpace = EAR::CSpace::WCS;

    auto dspData = std::make_unique<HeterogeneousStage::DspData>();
    auto& dspItemList = dspData->dspItemList;

    for (auto& item : renderList.items) {
        if ((item->extent == NULL) || (item->signal->getFrame().channels() <= 1)
            || item->type == ItemType::DirectionalReverb || isMphoaRenderItem(*item)) {
            // skip irrelevant render items.
            continue;
        }

        auto& heterItem = item->getStateOrCreate<HeterogeneousStageItem>(*item, this);

        heterItem.heterogeneousSource->UpdateState(
            item->extent, item->position, renderList.listener.position, tracer);

        // Update the positions and gain of the output rendering items
        for (int i = 0; i < heterItem.heterogeneousSource->GetNumOfOutputSpeakers(); i++) {
            pos.location = heterItem.heterogeneousSource->GetOutputSpeakerPosition(i).location;
            heterItem.outputItems[i]->setPosition(pos);
            heterItem.outputItems[i]->apparentDistanceDelta =
                heterItem.heterogeneousSource->GetOutputSpeakerDistanceOffset(i);
            heterItem.outputItems[i]->sync(*item,
                ItemProperty::Gain | ItemProperty::Loudness | ItemProperty::Eq
                    | ItemProperty::Aparams | ItemProperty::RefDist);
            heterItem.outputItems[i]->addGain(
                heterItem.heterogeneousSource->GetOutputSpeakerDistanceGain(i));
            heterItem.outputItems[i]->setAParams(heterItem.outputItems[i]->aparams
                | AParam::ForceNoDoppler
                | AParam::NoDistanceGain); // We calculate our own distance gain
            if (heterItem.heterogeneousSource->GetOutputSpeakerActivity(i) == false) {
                heterItem.outputItems[i]->addGain(
                    0.0); // Set gain to zero to deactivate further processing
            }
            if (i > 5) {
                // Add compensation EQ to the rendering items exclusively used for the interior representation
                heterItem.outputItems[i]->addEq(heterItem.eq.data(), false);
            } else if ((i == 0) || (i == 1) || (i == 2) || (i == 4) || (i == 5)) {
                // Add EQ from occlusion, only for the speaker of the exterior representation
                heterItem.outputItems[i]->addEq(
                    heterItem.heterogeneousSource->GetOcclusionEqData(i), false);
            }
            heterItem.heterogeneousSource->UpdateDspState();
        }

        if (heterItem.heterogeneousSource->GetIntegrated3DoFHoa()) {
            // Render the parent rendering items in a 3DoF HOA rendering stage
            item->addGain(heterItem.heterogeneousSource->Get3DofHoaGain());
        } else {
            // Don't render the parent rendering item
            itemStore->deactivate(item.get());
            item->addGain(0.0);
        }
    }

    // Copy the output RenderItems of each volumetric audio element into the RenderList and prepare DSP data
    for (auto& state : itemStore->getItemStates<HeterogeneousStageItem>()) {

        // Create a new DSPItem, add the output signals and fill in  the HeterogeneousDspState structure
        dspItemList.emplace_back(
            state.inputSignal, state.heterogeneousSource->heterogeneousDspState);

        for (auto* ri : state.outputItems) {

            // Store the audio buffers in the DSP items for use in the processBlock function
            dspItemList.back().outputSignals.push_back(&ri->signal->getFrame()[0]);

            renderList.addItemCopy(ri);
            ri->resetChanged();
        }

        dspItemList.back().heterogeneousDspState = state.heterogeneousSource->heterogeneousDspState;
    }

    return dspData;
}

void HeterogeneousStage::processBlock(DspContext ctx)
{
    auto& currentItemList = ctx.getData<HeterogeneousStage::DspData>().dspItemList;
    for (auto dspItem : currentItemList) {
        CalculateHeterogeneousSourceOuputSignals(
            dspItem.heterogeneousDspState, dspItem.inputSignal, dspItem.outputSignals, B);
    }
}

void HeterogeneousStage::configValueChanged(const std::string& key) {}

void HeterogeneousStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }

    this->itemStore = itemStore;
    B = streamManager->getB();

    // Load AudioplusPayload
    try {
        audioplusPayload = std::dynamic_pointer_cast<const MPEGI::AudioplusPayload>(
            bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_AUDIO_PLUS));
        audioplusMap = audioplusPayload->getAudioplus();
    } catch (const EAR::Exception e) {
        // Do nothing
    }

    // Setup information for voxel-traversal, if applicable
    auto const voxelSceneState = state->getVoxelSceneState();
    if (Config::getBoolDef("HeterogeneousVoxelOcclusion", true) && voxelSceneState) {
        auto init = [this](Voxel::VoxelSceneState::const_shared const& voxelSceneState) {
            this->voxelInfo = std::make_shared<voxel_info>();

            this->voxelInfo->grid = voxelSceneState->get_grid();
            this->voxelInfo->sceneInfo = voxelSceneState->get_scene_info_basic();
            this->voxelInfo->material_map =
                voxelSceneState->get_current_subscene()->scene_materials;
            this->voxelInfo->recovery = std::make_shared<occlusion_recovery>();
            this->voxelInfo->recovery->init(
                voxelSceneState->get_current_subscene()->renderer_instructions);
        };

        init(voxelSceneState);
        voxelSceneState->on_subscene_change(init);
    }

    // Get the mapping from bsId to string Id
    std::shared_ptr<const EIFPayload> eifPayload = std::dynamic_pointer_cast<const EIFPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_SCENE));

    // Initialize ray-tracer
    auto isTransformUpdated = [&](const std::string& tformId) {
        for (auto& upd : bitstream.getUpdates()) {
            for (auto& mod : upd.modifications) {
                if (mod.id == tformId) {
                    return true;
                }
            }
        }
        return false;
    };

    // TODO: add spheres and cylinders
    std::vector<const EAR::Mesh*> meshes;
    if (!voxelInfo) {
        for (const auto geo : state->getAllVisibleGeometry()) {
            if (geo->getMesh()) {
                if (geo->getMesh()->getParent()) {
                    if (isTransformUpdated(geo->getMesh()->getParent()->getName())) {
                        continue;
                    }
                }

                /*if (!geo->isStatic()) {
                const_cast<EAR::Geometry*>(geo)->attach(this);
                attachedGeometry.push_back(geo);
            }*/
                meshes.push_back(geo->getMesh());
            }
        }
    }

    // Add object sources with an extent
    for (const auto& objsource : state->getObjectSources()) {
        if (Geometry* extent = objsource->getExtent()) {
            // TODO: Exclude homogeneous sources here somehow
            meshes.push_back(extent->getMesh());
        }
    }

    // Add HOA sources with an extent and extent transform active
    for (const auto& hoasource : state->getHOASources()) {
        if (Geometry* extent = hoasource->getExtent()) {
            if (hoasource->doesExtentTransform()) {
                meshes.push_back(extent->getMesh());
            }
        }
    }

    if (true /*meshes.size() > 0*/) { // only run this stage when geometry is in the scene
        tracer = std::make_unique<EAR::EmbreeTracer>(meshes, true, true);
        //geometryDirty.store(true);
        //newMagnitudes.resize(Config::getInt("EQNBands"), 1.f);
    }

    B = streamManager->getB();
}

} // namespace MPEGI
