/***************************************************************************
This software module was originally developed by

Ericsson AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Ericsson AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_HETEROGENEOUSSTAGE
#define MPEGI_HETEROGENEOUSSTAGE

#include "Heterogeneous.h"

#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/AudioplusPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/ConfigObserver.h>
#include <MPEGI/Helpers/EmbreeHelper.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <utility>

#define HETEROGENEOUS_STAGE_INTEGRATE_WITH_3DOF_HOA

namespace EAR {
class RealBuffer;
class Decorrelator;
struct Position;
class EmbreeTracer;
class Geometry;
} // namespace EAR

namespace MPEGI {

class HeterogeneousDspState;
class HeterogeneousSource;

//! Rendering stage for rendering heterogeneous audio elements
/*!
* Inputs:
* - Items with an extent that has more than one channel and is not part of an HOA (multi-point) group.
*
* Outputs:
* - A set of new rendering items are created to represent the original source. The original
*   rendering item is deactivated.
*
* How it works:
* - Both an interior and exterior representation is created.
* - The exterior representation is based on a simplified extent.
* - Depending on the listening position, the interior or exterior representation is rendered
*   and within a transition region, there is a morph between these rendered.
*
* Stages order:
* - Required before: none
* - Required after: DistanceStage, EqStage
*
* Config options:
* - none
*
*/
class HeterogeneousStage : public ConfigObserver, public RenderStage
{
public:
    //! Constructor
    HeterogeneousStage();

    //! Destructor
    ~HeterogeneousStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

    //! Inherited from ConfigObserver
    void configValueChanged(const std::string& key) override;

private:
    friend class Renderer;
    RenderItemStore* itemStore;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState*) override;

    // Shared pointer of the AudioplusPayload payload
    std::shared_ptr<const MPEGI::AudioplusPayload> audioplusPayload{ nullptr };

    // The audioplusMap
    std::map<std::string, std::shared_ptr<AudioplusData>> audioplusMap;
    std::map<std::string, std::shared_ptr<AudioplusData>> audioplusStrIdMap;

    // Ray-tracing
    std::unique_ptr<EAR::EmbreeTracer> tracer;

    /// Voxel-info shared with heterogeneous sources
    voxel_info::shared voxelInfo;

    /*
     *  DSPItem
     *
     *  This structure keeps a pointer to the DSP state of a heterogeneous source along with 
     *  pointers to the input and output signals. Copys of this structure is sent to the 
     *  ProcessBlock function.
     */
    struct DSPItem
    {
        EAR::RealFrame* inputSignal;
        std::vector<EAR::RealBuffer*> outputSignals;
        HeterogeneousDspState* heterogeneousDspState;

        DSPItem(EAR::RealFrame* sig, HeterogeneousDspState* heterogeneousDspState_p)
            : inputSignal(sig)
            , heterogeneousDspState(heterogeneousDspState_p){};
    };

    struct HeterogeneousStageItem : RenderItem::ExtraState
    {
        int idx;
        HeterogeneousStage* parent;
        EAR::RealFrame* inputSignal;
        std::vector<RenderItem*> outputItems;
        HeterogeneousSource* heterogeneousSource;
        std::vector<float> eq;

        HeterogeneousStageItem(const RenderItem& renderItem, HeterogeneousStage* parent)
            : idx(renderItem.idx)
            , parent(parent)
            , inputSignal(&(renderItem.signal->getFrame()))
        {
            EAR::Position pos;
            pos.location = renderItem.extent->getGlobalLocation();
            pos.cSpace = EAR::CSpace::WCS;

            bool integrated3DoFHoa = false;
#ifdef HETEROGENEOUS_STAGE_INTEGRATE_WITH_3DOF_HOA
            if (renderItem.type == MPEGI::ItemType::HOA) {
                integrated3DoFHoa = true;
            }
#endif

            // Setup EQ used for compensation of low frequency buildup
            int eqnBands = Config::getInt("EQNBands");
            eq.resize(eqnBands);
            const float* eqFreqs = Config::getFloatPtr("EQCenterFreqs");
            for (int i = 0; i < eqnBands; i++) {
                if (eqFreqs[i] < 100.0) {
                    eq[i] = 0.2;
                } else if (eqFreqs[i] < 500.0) {
                    eq[i] = 0.4;
                } else if (eqFreqs[i] < 100.0) {
                    eq[i] = 0.6;
                }
                // Align frequency respons with SphoaStage
                else if (integrated3DoFHoa && (100.0 < eqFreqs[i]) && (eqFreqs[i] < 500.0)) {
                    eq[i] = 0.5;
                } else if (integrated3DoFHoa && (4000.0 < eqFreqs[i]) && (eqFreqs[i] < 10000.0)) {
                    eq[i] = 1.5;
                } else {
                    eq[i] = 1.0;
                }
            }

            float transitionDistance =
                0.04 * renderItem.extent->greatestDistance(); // Default transition distance
            heterogeneousSource = new HeterogeneousSource(renderItem.type, renderItem.extentLayout,
                renderItem.extent, renderItem.refDistance, transitionDistance, parent->B,
                inputSignal->channels(), integrated3DoFHoa);

            for (int loop = 0; loop < heterogeneousSource->GetNumOfOutputSpeakers(); loop++) {
                outputItems.push_back(parent->itemStore->createItem("output_speaker_"
                    + std::to_string(loop) + "_of_heterogeneous_source_" + std::to_string(idx)));
                outputItems.back()->type = ItemType::Representative;
                outputItems.back()->setSignal(std::make_shared<StreamBuffer>(1, parent->B));
                outputItems.back()->setPosition(renderItem.extent->getGlobalPosition());
                outputItems.back()->teleport = true;
                outputItems.back()->refDistance = renderItem.refDistance;
            }

            // Check for extra audio parameters for this item
            auto audioplus = parent->audioplusMap.find(renderItem.objectName);
            if (audioplus != parent->audioplusMap.end()) {
                heterogeneousSource->SetDistanceModel((audioplus->second->distanceGainModel));
                if (audioplus->second->hasDiffuseness) {
                    heterogeneousSource->SetDiffuseness(*audioplus->second->diffuseness);
                }
            }
        }

        /**
         * Forward the voxel-info to the underlying heterogeneous source
         *
         * @param voxelInfo Voxel-info shared-pointer (may be nullptr)
         */
        void SetVoxelInfo(voxel_info::shared const& voxelInfo)
        {
            if (heterogeneousSource) {
                this->heterogeneousSource->SetVoxelInfo(voxelInfo);
            }
        }
    };

    struct DspData : RenderStage::DspData
    {
        std::vector<DSPItem> dspItemList;
    };

    HeterogeneousStageItem& heterogeneousStageItemFor(RenderItem& ri);

    std::size_t B;
};

} // namespace MPEGI

#endif /* MPEGI_HETEROGENEOUSSTAGE */
