/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ExtentProcessor.h"

#include <MPEGI/Renderer/PerceptualPointDistribution.h>

#include <EAR/CartesianKDTree.h>
#include <EAR/ComplexFrame.h>
#include <EAR/EarPointCloud.h>
#include <EAR/Exception.h>
#include <EAR/FFT.h>
#include <EAR/FIRFilter.h>
#include <EAR/IRProvider.h>
#include <EAR/MpegiDecorrelator.h>
#include <EAR/PchipInterpolator.h>
#include <EAR/RealBuffer.h>
#include <EAR/StereoFIRFilter.h>
#include <nanoflann.hpp>

#include <MPEGI/Helpers/Config.h>

#include <algorithm>
#define _USE_MATH_DEFINES
#include <cmath>
#include <complex>

namespace MPEGI {

namespace detail {
static const std::complex<float> imag(0, 1);

void unwrap(float* phase, int L)
{
    float prev = phase[0];
    for (int i = 1; i < L; i++) {
        float d = phase[i] - prev;
        prev = phase[i];
        d = d > M_PI ? d - 2 * M_PI : (d < -M_PI ? d + 2 * M_PI : d); // wrap if distance is > pi
        phase[i] = phase[i - 1] + d;
    }
}
} // namespace detail

ExtentProcessor::LookupStruct::LookupStruct(const EAR::IRProvider* irPtr, int K)
    : hrirs(irPtr)
    , K(K)
    , N(irPtr->irLength())
    , fs(irPtr->getSampleRate()){};

ExtentProcessor::LookupStruct::~LookupStruct(){};

std::unique_ptr<const ExtentProcessor::LookupStruct> ExtentProcessor::extentLookupStruct = nullptr;

ExtentProcessor::ExtentProcessor(std::size_t B)
    : B(B)
{
    EAR_CHECK_REL(extentLookupStruct != nullptr, "the extent lookup table is not available");

    N = extentLookupStruct->N;
    hrirFFT = EAR::RFFT::create(N);
    hrirIFFT = EAR::IRFFT::create(N);
    K = hrirFFT->K;

    leftEarFilter = std::make_unique<EAR::StereoFIRFilter>(B, N);
    rightEarFilter = std::make_unique<EAR::StereoFIRFilter>(B, N);

    decorrelator = std::make_unique<EAR::MpegiDecorrelator>(B);

    tempSignalA = std::make_unique<EAR::RealFrame>(2, B);
    tempSignalB = std::make_unique<EAR::RealFrame>(2, B);
    tempDecor = std::make_unique<EAR::RealFrame>(2, B);
    hrtfPhase = std::make_unique<EAR::ComplexFrame>(2, K);
    iacc = std::make_unique<EAR::ComplexBuffer>(K);
    gains = std::make_unique<EAR::RealFrame>(2, K);
    H_alpha = std::make_unique<EAR::RealBuffer>(K);
    H_beta = std::make_unique<EAR::RealBuffer>(K);

    filtersLeft = std::make_unique<EAR::RealFrame>(2, N);
    filtersRight = std::make_unique<EAR::RealFrame>(2, N);
    tempHrir = std::make_unique<EAR::RealFrame>(2, N);
    tempFiltersFreq = std::make_unique<EAR::ComplexFrame>(4, K);

    virtualSourceWeight = 0.f;
    totalHitCount = 0;
    lastTotalHitCount = 0;
    insideHitCount = 0;

    hasSilentParts = false;
    isAudioSilent = false;
    wasAudioSilent = false;

    smoothingDecayConfig.store(-1.f);
}

ExtentProcessor::~ExtentProcessor() {}

void ExtentProcessor::clearRanges()
{
    virtualSourceWeight = 0.f;
    lastTotalHitCount = totalHitCount;
    totalHitCount = 0;
    insideHitCount = 0;

    gains->zero();
    iacc->zero();
    hrtfPhase->zero();
    wasAudioSilent = isAudioSilent;
    hasSilentParts = false;
    isAudioSilent = false;
}

void ExtentProcessor::addRange(const std::vector<RayHit>& rayHits, const std::vector<float>& eq)
{
    std::vector<RayHit> resultHits;
    resultHits.reserve(rayHits.size());

    //float hitWeight = 0.f;
    for (const auto& p : rayHits) {
        // To get a smooth transition in and out of the extent (180 to 360 degree), we skip the rays that
        // hit the inside of the extent but with a distance smaller than the virtual head size.
        if (p.distance > 0 || p.distance < -(extentLookupStruct->transitionDist)) {

            resultHits.push_back(p);
            //hitWeight += p.weight;
            if (p.distance < 0) {
                insideHitCount += 1;
            }
        }
    }

    const unsigned int resultSize = resultHits.size();
    totalHitCount += resultSize;

    // if this range is completely eq-ed out, get necessary data and skip it
    if (std::accumulate(eq.begin(), eq.end(), 0.f) == 0.f) {

        hasSilentParts = true;
        for (const auto& rh : resultHits) {
            auto tablePtr = extentLookupStruct->lookupTable[rh.rayID].get();
            virtualSourceWeight += tablePtr->numOfIndices * rh.weight;
        }

        return;
    }

    auto gainsData = gains->data();
    float* const gainsPtr[2] = { gainsData.get()[0], gainsData.get()[1] };
    std::complex<float>* const iaccPtr = iacc->data();

    //build an interpolator for the eq bands
    const int nBands = Config::getInt("EQNBands");
    const float* centerFreqs = Config::getFloatPtr("EQCenterFreqs");
    EAR::PchipInterpolator pci(centerFreqs, eq.data(), nBands, true);

    // add up all the relevant data (IACC, gain and total number of hrtf data points used) from lookup table
    // if all indices in the lookup table are included, use the already precalculated data for all hrtf points
    if (resultSize == extentLookupStruct->lookupTable.size()) {
        iacc->write(extentLookupStruct->allHrtfPoint->iacc->data(), 0, K);
        gains->copyFrom(*extentLookupStruct->allHrtfPoint->gains);
        virtualSourceWeight = extentLookupStruct->allHrtfPoint->numOfIndices;
    } else {
        float freqStep = extentLookupStruct->fs / 2 / (K - 1);
        // precaclulate bin gains, they are the same for each ray
        float binGains[1024];
        EAR_CHECK_REL(K <= 1024, "Extent bin count exceeded!");
        for (unsigned int i = 0; i < K; i++) {
            binGains[i] = pci.interpolate(i * freqStep);
        }
        for (const auto& rh : resultHits) {
            auto tablePtr = extentLookupStruct->lookupTable[rh.rayID].get();
            std::complex<float>* const iaccSource = tablePtr->iacc->data();
            float* const gainSource[2] = { tablePtr->gains->ch(0).data(),
                tablePtr->gains->ch(1).data() };
            virtualSourceWeight += tablePtr->numOfIndices * rh.weight;
            for (unsigned int i = 0; i < K; i++) {
                const float binGain = binGains[i] * binGains[i] * rh.weight * rh.weight;
                iaccPtr[i] += binGain * iaccSource[i];
                gainsPtr[0][i] += binGain * gainSource[0][i];
                gainsPtr[1][i] += binGain * gainSource[1][i];
            }
        }
    }
}

void ExtentProcessor::addPoint(const EAR::Vector3& dir, float gain, const std::vector<float>& eq)
{
    const float configGain = Config::getFloatDef("DiffractionSESSGain", 1.f);
    auto gainsPtr = gains->data();
    auto iaccPtr = iacc->data();
    auto hrtfPtr = hrtfPhase->data();
    auto tempHrirPtr = tempHrir->data();

    //build an interpolator for the eq bands
    const int nBands = Config::getInt("EQNBands");
    const float* centerFreqs = Config::getFloatPtr("EQCenterFreqs");
    EAR::PchipInterpolator pci(centerFreqs, eq.data(), nBands, true);

    float freqStep = extentLookupStruct->fs / 2 / (K - 1);
    // precaclulate bin gains, they are the same for each ray
    float binGains[1024];
    EAR_CHECK_REL(K <= 1024, "Extent bin count exceeded!");
    for (unsigned int i = 0; i < K; i++) {
        binGains[i] = pci.interpolate(i * freqStep);
    }

    // Compute phase info using the center hrir (or for no hrir in range case as well)
    extentLookupStruct->hrirs->getNearestIRForLocation(
        *tempHrir, dir * (extentLookupStruct->hrirDist / dir.abs()));
    hrirFFT->transform(tempHrirPtr[0], hrtfPtr[0]);
    hrirFFT->transform(tempHrirPtr[1], hrtfPtr[1]);
    const float normFactor = sqrt(extentLookupStruct->N);

    // compute IACC and gains
    for (unsigned int i = 0; i < K; i++) {
        const float binGain =
            /*gain*gain**/ configGain * configGain * binGains[i] * binGains[i]
            / (normFactor * normFactor);
        ;
        const float phaseDiff = std::arg(hrtfPtr[0][i]) - std::arg(hrtfPtr[1][i]);
        const float absL = std::abs(hrtfPtr[0][i]);
        const float absR = std::abs(hrtfPtr[1][i]);
        iaccPtr[i] += binGain * absL * absR * std::exp(detail::imag * phaseDiff);
        gainsPtr[0][i] += binGain * absL * absL;
        gainsPtr[1][i] += binGain * absR * absR;
    }
    virtualSourceWeight += 1.f;
}

void ExtentProcessor::calculateFilterState(EAR::Vector3 closestP, ExtentProcessorFilterState* state)
{
    const bool isInside = (totalHitCount == insideHitCount)
        && (totalHitCount >= extentLookupStruct->lookupTable.size() / 2);

    // if the listener is inside, the center point of the extent should be opposite of the closest point
    if (isInside) {
        closestP = -1.f * closestP;
    }
    // if the listener is right on the edge of the extent, use last frame's closest point
    if (closestP.abs() == 0.f) {
        closestP = savedClosestP;
    } else {
        // map the closest point onto the virtual sphere of the hrir points
        closestP = closestP * (extentLookupStruct->hrirDist / closestP.abs());
    }
    // work around for a discontinuity issue (when the source is far away and occluded)
    if (hasSilentParts && totalHitCount == 1 && wasAudioSilent) {
        state->isSilent = true;
    }
    // get the hrir closest to the center of the extent for phase calculation and when the extent
    // is far enough to be treated as point source
    extentLookupStruct->hrirs->getNearestIRForLocation(*tempHrir, closestP);

    savedClosestP = closestP;

    auto gainsPtr = gains->data();
    auto iaccPtr = iacc->data();
    auto phasePtr = hrtfPhase->data();
    auto tempHrirPtr = tempHrir->data();

    // Compute phase info using the center hrir (or for no hrir in range case as well)
    hrirFFT->transform(tempHrirPtr[0], phasePtr[0]);
    hrirFFT->transform(tempHrirPtr[1], phasePtr[1]);

    // we do not have any point in range for lookup table -> calculate date in place with nearest point
    if (virtualSourceWeight == 0.f) {

        float normFactor = sqrt(extentLookupStruct->N);

        EAR::ComplexFrame normalizedHrtf(2, K);
        normalizedHrtf.zero();
        auto hrtfPtr = normalizedHrtf.data();

        // compute IACC and gains
        // normalization of left channel
        std::transform(phasePtr[0], phasePtr[0] + K, hrtfPtr[0],
            [num = normFactor](std::complex<float> val) { return val / num; });

        // normalization of right channel
        std::transform(phasePtr[1], phasePtr[1] + K, hrtfPtr[1],
            [num = normFactor](std::complex<float> val) { return val / num; });

        // magnitude of left channel
        std::transform(hrtfPtr[0], hrtfPtr[0] + K, gainsPtr[0], [](std::complex<float>& val) {
            float aval = std::abs(val);
            return aval * aval;
        });

        // magnitude of right channel
        std::transform(hrtfPtr[1], hrtfPtr[1] + K, gainsPtr[1], [](std::complex<float>& val) {
            float aval = std::abs(val);
            return aval * aval;
        });

        // IACC
        for (unsigned int i = 0; i < K; i++) { // TODO: can we write that as a transform too?
            float phaseDiff = std::arg(hrtfPtr[0][i]) - std::arg(hrtfPtr[1][i]);
            iaccPtr[i] = std::abs(hrtfPtr[0][i]) * std::abs(hrtfPtr[1][i])
                * std::exp(detail::imag * phaseDiff);
        }
        // we have one hrir point now
        virtualSourceWeight = 1.f;
    }

    // normalize iacc and gain
    for (unsigned int i = 0; i < K; i++) { // TODO: can we write that as a transform too?
        iaccPtr[i] /= std::sqrt(gainsPtr[0][i] * gainsPtr[1][i]);
    }

    std::transform(gainsPtr[0], gainsPtr[0] + K, gainsPtr[0],
        [num = virtualSourceWeight](float val) { return std::sqrt(val / num); });

    std::transform(gainsPtr[1], gainsPtr[1] + K, gainsPtr[1],
        [num = virtualSourceWeight](float val) { return std::sqrt(val / num); });

    std::transform(phasePtr[0], phasePtr[0] + K, phasePtr[0],
        [](std::complex<float>& val) { return std::arg(val); });

    std::transform(phasePtr[1], phasePtr[1] + K, phasePtr[1],
        [](std::complex<float>& val) { return std::arg(val); });

    // compute cross-correlation magnitudes from intermediate results
    std::transform(iaccPtr, iaccPtr + K, H_beta->data(), [](std::complex<float>& val) {
        float aval = std::abs(val);
        return std::sqrt(0.5 * (1 - std::sqrt(std::fmax(1 - (aval * aval), 0))));
    });

    std::transform(H_beta->data(), H_beta->data() + K, H_alpha->data(),
        [](float val) { return std::sqrt(std::fmax(1 - (val * val), 0)); });

    // compute filters (two for each ear) and set them in the FIRFilter instances
    auto tempPtr = tempFiltersFreq->data();
    for (unsigned int i = 0; i < K; i++) { // TODO: can we write that as a transform too?
        tempPtr[0][i] = (H_alpha->at(i) * gainsPtr[0][i]) * std::exp(detail::imag * phasePtr[0][i]);
        tempPtr[1][i] = (H_beta->at(i) * gainsPtr[0][i]) * std::exp(detail::imag * phasePtr[0][i]);
        tempPtr[2][i] = (H_beta->at(i) * gainsPtr[1][i]) * std::exp(detail::imag * phasePtr[1][i]);
        tempPtr[3][i] = (H_alpha->at(i) * gainsPtr[1][i]) * std::exp(detail::imag * phasePtr[1][i]);
    }

    // decorrelated signal one should be filtered with H_alpha of left year + H_beta of the right year
    hrirIFFT->transform(tempPtr[0], filtersLeft->ch(0).data());
    hrirIFFT->transform(tempPtr[2], filtersLeft->ch(1).data());
    filtersLeft->mul(1.0f / sqrt(extentLookupStruct->N)); // normalization not part of IFFT
    filtersLeft->shiftLeft(filtersLeft->length() / 2);
    state->left.reset(new EAR::RealFrame(*filtersLeft));

    // decorrelated signal two should be filtered with H_beta of left year + H_alpha of the right year
    hrirIFFT->transform(tempPtr[1], filtersRight->ch(0).data());
    hrirIFFT->transform(tempPtr[3], filtersRight->ch(1).data());
    filtersRight->mul(1.0f / sqrt(extentLookupStruct->N)); // normalization not part of IFFT
    filtersRight->shiftLeft(filtersRight->length() / 2);
    state->right.reset(new EAR::RealFrame(*filtersRight));
}

void ExtentProcessor::setFilterState(const ExtentProcessorFilterState& state)
{
    //smoothing rate peak detector
    const float smoothingThres = 0.1f;
    const float smoothingDecay = smoothingDecayConfig.load();
    ;
    const float decayRate = std::pow(10.f, smoothingDecay);
    const float newSmoothing = state.smoothingWeight;
    const float currSW = leftEarFilter->getSmoothing();
    const float currentSmoothing =
        currSW > 0 ? -0.5f - std::log(1.f - currSW) / std::log(10.f) : 0.f;
    float finalSmoothing = currentSmoothing;
    if (newSmoothing > currentSmoothing)
        finalSmoothing = newSmoothing;
    else if (currentSmoothing < smoothingThres)
        finalSmoothing = 0.f;
    else
        finalSmoothing = currentSmoothing - decayRate * currentSmoothing;
    const float currSmoothingWeight =
        (finalSmoothing == 0.f) ? 0.f : 1.0f - std::pow(10.f, -(finalSmoothing + 0.5f));

    leftEarFilter->setSmoothing(currSmoothingWeight);
    leftEarFilter->setFilter(state.left->data().get());
    rightEarFilter->setSmoothing(currSmoothingWeight);
    rightEarFilter->setFilter(state.right->data().get());
}

void ExtentProcessor::process(const float* in, float** out)
{

    auto decorPtr = tempDecor->data();
    auto tempAPtr = tempSignalA->data();
    auto tempBPtr = tempSignalB->data();

    // process the signal with decorrelator
    decorrelator->process(in, decorPtr.get());

    // process the decorrelated signal with corresponding extent filter
    leftEarFilter->process(decorPtr[0], tempAPtr.get());
    rightEarFilter->process(decorPtr[1], tempBPtr.get());

    // mix the output of the filtered decorrelated signals
    std::transform(
        tempAPtr[0], tempAPtr[0] + B, tempBPtr[0], out[0], [](float a, float b) { return a + b; });

    std::transform(
        tempAPtr[1], tempAPtr[1] + B, tempBPtr[1], out[1], [](float a, float b) { return a + b; });
}

void ExtentProcessor::process(const float* inL, const float* inR, float** out)
{
    auto tempAPtr = tempSignalA->data();
    auto tempBPtr = tempSignalB->data();

    leftEarFilter->process(inL, tempAPtr.get());
    rightEarFilter->process(inR, tempBPtr.get());
    std::transform(
        tempAPtr[0], tempAPtr[0] + B, tempBPtr[0], out[0], [](float a, float b) { return a + b; });

    std::transform(
        tempAPtr[1], tempAPtr[1] + B, tempBPtr[1], out[1], [](float a, float b) { return a + b; });
}

void ExtentProcessor::setSourceSilentFlag(bool isSourceSilent)
{
    isAudioSilent = isSourceSilent;
}

void ExtentProcessor::updateHrirSet(const EAR::IRProvider* hrirs)
{

    // Prepare variables and containers
    std::unique_ptr<EAR::RFFT> hrirFFT = EAR::RFFT::create(hrirs->irLength());
    int K = hrirFFT->K;
    std::unique_ptr<EAR::ComplexFrame> hrtf = std::make_unique<EAR::ComplexFrame>(2, K);
    std::unique_ptr<LookupStruct> newStruct = std::make_unique<LookupStruct>(hrirs, K);
    float normFactor = sqrt(newStruct->N);

    // use a random point to test and temporarily store the common distance
    // from listener to the hrir points in this database for later sanity checks
    EAR::Vector3 tempLoc = polarToCartesian(EAR::Polar(0, 0, 2));
    float hrirDist = (hrirs->getNearestLocation(tempLoc)).abs();

    newStruct->transitionDist = Config::getFloatDef("virtualHeadSize", 0.1);
    newStruct->hrirDist = hrirDist;

    // Initialize the lookup point for when all hrtf points is included to render the extent
    newStruct->allHrtfPoint = std::make_unique<LookupPoint>(K);
    auto aGainsPtr = newStruct->allHrtfPoint->gains->data();
    auto aIaccPtr = newStruct->allHrtfPoint->iacc->data();
    newStruct->allHrtfPoint->gains->zero();
    newStruct->allHrtfPoint->iacc->zero();
    newStruct->allHrtfPoint->numOfIndices = hrirs->irCount();

    // Get the pre-determined tile center points
    const EAR::Vector3* rays = SESS_RAY_DISTRIBUTION;
    const auto rays_size = sizeof(SESS_RAY_DISTRIBUTION) / sizeof(EAR::Vector3);

    std::vector<EAR::Vector3> tileCenters;
    tileCenters.reserve(rays_size);
    newStruct->lookupTable.reserve(rays_size);

    // initialize all data points in the lookup table
    for (int r = 0; r < rays_size; r++) {

        // store the representation position of the tile
        tileCenters.push_back(rays[r] * hrirDist);

        // initialize data container for each tile
        std::unique_ptr<LookupPoint> newPoint = std::make_unique<LookupPoint>(K);
        newPoint->gains->zero();
        newPoint->iacc->zero();
        newPoint->numOfIndices = 0;
        newStruct->lookupTable.push_back(std::move(newPoint));
    }

    // initialize the lookupTree for all tile centers
    newStruct->lookupTree = std::make_unique<EAR::CartesianKDTree>(tileCenters);

    // Find the which tile the current hrir data point belongs to and store precalculated values
    for (int i = 0; i < hrirs->irCount(); i++) {

        EAR::Vector3 irLoc = hrirs->getLocationForIndex(i);
        EAR_CHECK_REL(
            std::abs(irLoc.abs() - hrirDist) <= 0.01f, "there are multiple layers of hrir points");

        int idx = newStruct->lookupTree->getNearestIndex(irLoc);
        auto gainsPtr = newStruct->lookupTable[idx]->gains->data();
        auto iaccPtr = newStruct->lookupTable[idx]->iacc->data();
        auto hrtfPtr = hrtf->data();

        std::unique_ptr<const float*[]> hrir = hrirs->getIRPointerForIndex(i);
        hrirFFT->transform(hrir[0], hrtfPtr[0]);
        hrirFFT->transform(hrir[1], hrtfPtr[1]);

        // compute IACC and gains
        // normalization of left channel
        std::transform(hrtfPtr[0], hrtfPtr[0] + K, hrtfPtr[0],
            [num = normFactor](std::complex<float> val) { return val / num; });

        // normalization of right channel
        std::transform(hrtfPtr[1], hrtfPtr[1] + K, hrtfPtr[1],
            [num = normFactor](std::complex<float> val) { return val / num; });

        // magnitude of left channel
        std::transform(hrtfPtr[0], hrtfPtr[0] + K, gainsPtr[0], gainsPtr[0],
            [](std::complex<float>& val, float prev) {
                float aval = std::abs(val);
                return prev + aval * aval;
            });

        // magnitude of right channel
        std::transform(hrtfPtr[1], hrtfPtr[1] + K, gainsPtr[1], gainsPtr[1],
            [](std::complex<float>& val, float prev) {
                float aval = std::abs(val);
                return prev + aval * aval;
            });

        // IACC
        for (unsigned int i = 0; i < K; i++) { // TODO: can we write that as a transform too?
            float phaseDiff = std::arg(hrtfPtr[0][i]) - std::arg(hrtfPtr[1][i]);
            iaccPtr[i] += std::abs(hrtfPtr[0][i]) * std::abs(hrtfPtr[1][i])
                * std::exp(detail::imag * phaseDiff);
        }

        // add data into the point precalculated for 'allHRTF included' case
        for (int i = 0; i < K; i++) {
            aGainsPtr[0][i] += gainsPtr[0][i];
            aGainsPtr[1][i] += gainsPtr[1][i];
            aIaccPtr[i] += iaccPtr[i];
        }

        newStruct->lookupTable[idx]->numOfIndices += 1;
    }

    newStruct->fs = hrirs->getSampleRate();

    //HACK Calculate the list of indices of rays without any hrtf samples assigned
    //std::vector<int> zeroIndices;
    //for (int r = 0; r < rays_size; r++) {
    //    if (newStruct->lookupTable[r]->numOfIndices == 0)
    //        zeroIndices.push_back(r);
    //}

    extentLookupStruct = std::move(newStruct);
}

float ExtentProcessor::getHitCountChange() const
{
    if (lastTotalHitCount == 0 && totalHitCount == 0)
        return 0;
    const float diff = std::fabs(float(lastTotalHitCount) - float(totalHitCount));
    const float sum = lastTotalHitCount + totalHitCount;
    return 2.f * diff / sum;
}

void ExtentProcessor::setSmoothingDecay(float smoothingDecay)
{
    smoothingDecayConfig.store(smoothingDecay);
}

} // namespace MPEGI
