/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_EXTENT_PROCESSOR
#define MPEGI_EXTENT_PROCESSOR

#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItem.h>

#include <EAR/BaseTypes.h>
#include <EAR/ComplexBuffer.h>
#include <EAR/Geometry.h>
#include <EAR/RealFrame.h>

#include <atomic>
#include <cstddef>
#include <memory>

// fwd declares
namespace EAR {
class CartesianKDTree;
class FIRFilter;
class RFFT;
class IRFFT;
class IRProvider;
class ComplexFrame;
class RealBuffer;
class MpegiDecorrelator;
class StereoFIRFilter;
} // namespace EAR

namespace MPEGI {

struct ExtentProcessorFilterState
{
    std::unique_ptr<EAR::RealFrame> left{ nullptr };
    std::unique_ptr<EAR::RealFrame> right{ nullptr };
    float smoothingWeight{ 0.f };
    bool isSilent{ false };
};

//! Binauralization of a single spatially extended sound source (SESS)
class ExtentProcessor : public RenderItem::ExtraState
{
public:
    //! Constructor
    /*!
	 * \param B 			block size in samples
	 */
    ExtentProcessor(std::size_t B);

    //! Destructor
    ~ExtentProcessor();

    //! Clears the state before receiving ranges of rayhits
    void clearRanges();

    //! Add a set of rays sharing a common eq curve
    /*
     * \param rayHits 	a vector of pair<index of rayhit direction, distance to the hit on extent>
     * \param eq        the eq curve corresponding to the materials hit by this batch of rays
     */
    void addRange(const std::vector<RayHit>& rayHits, const std::vector<float>& eq);

    void addPoint(const EAR::Vector3& dir, float gain, const std::vector<float>& eq);

    //! Finalize the filter calculations based on the rayhits submited and generate the filter state
    void calculateFilterState(EAR::Vector3 closestP, ExtentProcessorFilterState* state);

    // TODO: set the range with a given channel layout
    // void setRangeMultichannel(const EAR::Vector3& bottomLeft, const EAR::Vector3& topRight, Layout);

    //! Binauralize a SESS with a single input channel
    /*!
	 * Can also operate in place.
	 *
     * \param in 			single-channel input signal
     * \param out 			two-channel binauralized output signal
     */
    void process(const float* in, float** out);

    void setFilterState(const ExtentProcessorFilterState& state);

    //! Binauralize a SESS with stereo input channel
    /*!
     * Can also operate in place.
     *
     * \param inL 			the left input signal
     * \param inR 			the right input signal
     * \param out 			two-channel binauralized output signal
     */
    void process(const float* inL, const float* inR, float** out);

    void setSourceSilentFlag(bool isSourceSilent);

    //! remake the lookupStruct for extent stage everytime a new hrir database is loaded
    /*!
	 * Can also operate in place.
	 *
     * \param hrirs			the hrir database currently used by the renderer
     */
    static void updateHrirSet(const EAR::IRProvider* hrirs);

    float getHitCountChange() const;

    void setSmoothingDecay(float smoothingDecay);

private:
    //! This struct stores the data of each grid point in the extent lookupTable
    struct LookupPoint
    {
        LookupPoint(int K)
            : numOfIndices(0)
            , iacc(std::make_unique<EAR::ComplexBuffer>(K))
            , gains(std::make_unique<EAR::RealFrame>(2, K)){};

        int numOfIndices;
        std::unique_ptr<EAR::ComplexBuffer> iacc;
        std::unique_ptr<EAR::RealFrame> gains;
    };

    //! This struct stores all precalculated data for extent processing
    struct LookupStruct
    {
        //! Constructor
        LookupStruct(const EAR::IRProvider* irPtr, int K);
        //! Destructor
        ~LookupStruct();

        int K;
        int N;
        float transitionDist;
        float hrirDist;
        float fs;
        const EAR::IRProvider* hrirs;
        std::unique_ptr<EAR::CartesianKDTree> lookupTree;
        std::vector<std::unique_ptr<LookupPoint>> lookupTable;
        std::unique_ptr<LookupPoint> allHrtfPoint;
    };

    std::size_t B, N, K;

    std::unique_ptr<EAR::RFFT> hrirFFT;
    std::unique_ptr<EAR::IRFFT> hrirIFFT;

    std::unique_ptr<EAR::StereoFIRFilter> leftEarFilter;
    std::unique_ptr<EAR::StereoFIRFilter> rightEarFilter;
    std::unique_ptr<EAR::MpegiDecorrelator> decorrelator;

    std::unique_ptr<EAR::RealFrame> tempSignalA;
    std::unique_ptr<EAR::RealFrame> tempSignalB;
    std::unique_ptr<EAR::RealFrame> tempDecor;
    std::unique_ptr<EAR::ComplexFrame> hrtfPhase;
    std::unique_ptr<EAR::ComplexBuffer> iacc;
    std::unique_ptr<EAR::RealFrame> gains;
    std::unique_ptr<EAR::RealBuffer> H_alpha;
    std::unique_ptr<EAR::RealBuffer> H_beta;

    std::unique_ptr<EAR::ComplexFrame> tempFiltersFreq;
    std::unique_ptr<EAR::RealFrame> filtersLeft;
    std::unique_ptr<EAR::RealFrame> filtersRight;
    std::unique_ptr<EAR::RealFrame> tempHrir;

    std::atomic<bool> singleHRIR;
    std::vector<unsigned int> savedIndices;
    EAR::Vector3 savedClosestP;

    float virtualSourceWeight;

    unsigned int totalHitCount;
    unsigned int lastTotalHitCount;
    unsigned int insideHitCount;

    bool hasSilentParts, isAudioSilent, wasAudioSilent;

    std::atomic<float> smoothingDecayConfig;

    // a pointer to all precalculated data for extent processing
    static std::unique_ptr<const LookupStruct> extentLookupStruct;
};

} // namespace MPEGI

#endif
