/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "HomogeneousExtent.h"

#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

#include <EAR/BaseTypes.h>
#include <EAR/IRProvider.h>
#include <EAR/RealFrame.h>

#include <EAR/FixedHRIRProvider.h>
#include <EAR/SofaHRIRProvider.h>

#include <cassert>
#include <cmath>

#include "ExtentProcessor.h"

namespace MPEGI {

std::shared_ptr<RenderStage> makeHomogeneousExtentStage()
{
    return std::make_shared<HomogeneousExtentStage>();
}

HomogeneousExtentStage::HomogeneousExtentStage()
    : itemStore(nullptr)
{}

HomogeneousExtentStage::~HomogeneousExtentStage() {}

std::unique_ptr<RenderStage::DspData> HomogeneousExtentStage::update(RenderList& renderList)
{
    if (!itemStore || !hrirInitialized.load()) {
        return {};
    }
    const float sw =
        std::max(0.f, std::min(1.f, Config::getFloatDef("ExtentFilterSmoothing", 0.0f)));
    // NB: we use std::pow because std::powf is not available in GCC, see
    //     https://gcc.gnu.org/bugzilla/show_bug.cgi?id=79700
    const float smoothingWeight = (sw == 0.f) ? 0.f : 1.0f - std::pow(10.f, -(sw + 0.5f));
    const float smoothingDecay =
        std::max(-10.f, std::min(-1.f, Config::getFloatDef("SmoothingDecayRate", -2.0f)));
    const float hitCountMin =
        std::max(0.f, std::min(1.f, Config::getFloatDef("ExtentHitCountMin", 0.3f)));
    const float hitCountMax =
        std::max(0.f, std::min(1.f, Config::getFloatDef("ExtentHitCountMax", 0.6f)));
    const float hitCountRng = std::max(hitCountMax - hitCountMin, 0.01f);

    auto dspData = std::make_unique<HomogeneousExtentStage::DspData>();
    auto& dspList = *dspData;
    //make sure all homogeneous primary sources have valid extent processors
    //clear the extent processor states before adding ray ranges from the representative items
    for (auto& item : renderList.items) {
        if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned)
            continue;

        if (item->extent
            && ((item->type == ItemType::Primary && item->signal->getFrame().channels() == 1)
                || (item->type == ItemType::DirectionalReverb && item->parentRI == -1))) {
            auto& extentProcessor = item->getStateOrCreate<ExtentProcessor>(B);

            extentProcessor.clearRanges();

            // document if this extent is completely eq-ed out
            bool itemAudioSilent = false;
            if (item->eq[0] == 0.f) {
                itemAudioSilent = (std::accumulate(item->eq.begin(), item->eq.end(), 0.f) == 0.f);
            }
            extentProcessor.setSourceSilentFlag(itemAudioSilent);
        }
    }
    //collect rayhit bundles with their cumulative eqs
    for (auto& item : renderList.items) {
        if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned)
            continue;

        if (!(item->extent) || item->parentRI == -1
            || !(((item->type == ItemType::Primary || item->type == ItemType::Representative
                      || item->type == ItemType::Diffraction)
                     && item->signal->getFrame().channels() == 1)
                || item->type == ItemType::DirectionalReverb)) {
            // we only care for extended items with a single input audio channel ("homogeneuous SESS") or portals
            continue;
        }

        //if ((item->changed & (ItemProperty::Position | ItemProperty::Status))
        //    || (item->position.cSpace != EAR::CSpace::LCS && renderList.listener.hasMoved)) {
        if (item->type == ItemType::Representative || item->type == ItemType::DirectionalReverb) {
            auto* proc = itemStore->getState<ExtentProcessor>(item->parentRI);
            assert(proc);
            proc->addRange(item->rayHits, item->eq);

            //deactivate this representative render item
            itemStore->deactivate(item.get());
        }

        if (item->type == ItemType::Diffraction) {
            auto* proc = itemStore->getState<ExtentProcessor>(item->parentRI);
            assert(proc);
            proc->addPoint(
                renderList.listener.cs.transform(item->position.location), item->gain, item->eq);

            //deactivate this diffraction render item
            itemStore->deactivate(item.get());
        }
    }

    //generate updated filters for this frame
    for (auto& item : renderList.items) {
        if (item->status == ItemStatus::Inactive || item->status == ItemStatus::Orphaned)
            continue;

        if (item->extent
            && ((item->type == ItemType::Primary && item->signal->getFrame().channels() == 1)
                || (item->type == ItemType::DirectionalReverb && item->parentRI == -1))) {

            auto* processor = item->getState<ExtentProcessor>();
            if (processor == nullptr) {
                EAR_FAIL("Missing an extentProcessor for a primary extended source!");
            }
            EAR::Vector3 closestP;
            if (item->position.cSpace == EAR::CSpace::LCS && item->type == ItemType::Nondiegetic) {
                // listener-relative
                closestP = item->extent->closestPointTo(EAR::Vector3(0, 0, 0));
            } else {
                // global coordinates
                // By default uses the closest point on the extent mesh as to where it should be localized.
                // For concave shapes (i.e. portal extent meshes), or occluded shapes, this is often not
                // the visual centre of the mesh, so instead we use the centre of the mesh.
                // TODO: find proper fix for localization of occluded / concave meshes and remove this.
                if (item->type == ItemType::DirectionalReverb && isExplicitPortalMode) {
                    closestP = renderList.listener.cs.transform(item->extent->getCenterPosition());
                } else {
                    closestP = renderList.listener.cs.transform(
                        item->extent->closestPointTo(renderList.listener.position.location));
                }
            }

            auto filterState = ExtentProcessorFilterState{};
            processor->calculateFilterState(closestP, &filterState);
            processor->setSmoothingDecay(smoothingDecay);
            const float hitCountScale = processor->getHitCountChange();
            const float smoothingParam = (filterState.isSilent ? 0.f : sw)
                * std::fmin(std::fmax(hitCountScale - hitCountMin, 0.f), hitCountRng) / hitCountRng;
            filterState.smoothingWeight = smoothingParam;
            dspList.items.emplace_back(item->signal->getFrame(), processor, std::move(filterState));

            // deactivate primary render item
            itemStore->deactivate(item.get());
        }
    }

    if (dspList.items.size() != 0) {
        // send downmixout into processBlock
        dspList.output = &(extentDownmixItem->signal->getFrame());

        // add downmix item to the render list
        renderList.addItemCopy(extentDownmixItem);
        extentDownmixItem->resetChanged();
    }

    return dspData;
}

void HomogeneousExtentStage::processBlock(DspContext ctx)
{
    if (!hrirInitialized.load()) {
        return;
    }
    auto& currentDSP = ctx.getData<HomogeneousExtentStage::DspData>();

    if (!currentDSP.output) {
        return;
    }
    currentDSP.output->zero();

    auto tmpPtr = temp->data();
    auto outL = &(currentDSP.output->ch(0));
    auto outR = &(currentDSP.output->ch(1));

    for (auto& item : currentDSP.items) {
        item.extent->setFilterState(item.filterState);

        if (item.signal.channels() == 1) {
            item.extent->process((&item.signal[0])->data(), tmpPtr.get());
        } else {
            item.extent->process(
                (&item.signal[0])->data(), (&item.signal[1])->data(), tmpPtr.get());
        }

        if (item.filterState.isSilent) {
            continue;
        }
        // add extent signal to downmix output
        outL->add(temp->ch(0));
        outR->add(temp->ch(1));

    } // dsp item loop
}

void HomogeneousExtentStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    hrirInitialized.store(false);
    this->itemStore = itemStore;
    B = streamManager->getB();

    const std::string path = Config::getStringDef("HrirSofaFile", "");
    if (path == "") {
        hrirs = std::make_unique<EAR::FixedHRIRProvider>();
    } else {
        hrirs = std::make_unique<EAR::SofaHRIRProvider>(path);
    }

    ExtentProcessor::updateHrirSet(hrirs.get());

    // create downmix item for extent sources
    extentDownmixItem = itemStore->createItem("extents_downmix");
    extentDownmixItem->type = ItemType::Downmix;
    extentDownmixItem->setSignal(streamManager->createEmptyBuffer(2));
    temp = std::make_unique<EAR::RealFrame>(2, B);

    hrirInitialized.store(true);

    auto portal_payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_PORTAL);
    EAR_CHECK_REL(portal_payloads.size() <= 1, "Too many portal payloads in bitstream");
    if (!portal_payloads.empty()) {
        isExplicitPortalMode = std::dynamic_pointer_cast<PortalPayload const>(portal_payloads[0])
                                   ->getExplicitPortalMode();
    }
}

} // namespace MPEGI
