/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_HOMOGENEOUS_EXTENT_STAGE
#define MPEGI_HOMOGENEOUS_EXTENT_STAGE

#include <cstddef>
#include <map>
#include <memory>
#include <unordered_map>

#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include "ExtentProcessor.h"

namespace EAR {
class IRProvider;
class RealFrame;
} // namespace EAR

namespace MPEGI {

//! Render homogeneous extent (spatially extended sound sources with a single-channel input signal)
/*! 
 * Inputs: 
 * - All RIs with a geometry given for the "extent" property AND either of the following two options:
 *      - ItemType::Primary AND single-channel input signal
 *      - ItemType::DirectionalReverb
 * - Possibly external HRIR set given at runtime
 *  
 * Outputs: 
 * - The considered RIs are deactivated
 * - A new RI with ItemType::Downmix is created, that contains binauralized and downmixed signals
 *   for all considered RIs
 *  
 * How it works: 
 * - A 2D projection of the object extent onto sectors on the listener sphere is calculated
 *   by the DiscoverSess stage
 * - This stage then integrates all HRTFs from the covered sectors into two pairs of stereo filters
 *   which are applied to the input signal
 * - The filter output is downmixed into the stereo output
 * 
 * Stages order:
 * - Required before: DiscoverSessStage, DistanceStage, EqStage, FadeStage
 * - Required after: none
 *  
 * Config options: 
 * - none
 * 
 */
class HomogeneousExtentStage : public RenderStage
{
public:
    HomogeneousExtentStage();

    //! Destructor
    ~HomogeneousExtentStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

private:
    friend class Renderer;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    float fs;
    std::size_t B;
    RenderItemStore* itemStore;

    struct DSPItem
    {
        EAR::RealFrame& signal;
        ExtentProcessor* extent;
        ExtentProcessorFilterState filterState;

        DSPItem(EAR::RealFrame& sig, ExtentProcessor* e, ExtentProcessorFilterState&& fs)
            : signal(sig)
            , extent(e)
            , filterState(std::move(fs))
        {}
    };

    struct DspData : RenderStage::DspData
    {
        std::vector<DSPItem> items;
        EAR::RealFrame* output;
    };

    std::atomic<bool> hrirInitialized;
    std::unique_ptr<EAR::IRProvider> hrirs;
    std::unique_ptr<EAR::RealFrame> temp;

    RenderItem* extentDownmixItem;

    bool isExplicitPortalMode{ false };
};

} // namespace MPEGI

#endif /* MPEGI_HOMOGENEOUS_EXTENT_STAGE */
