/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "LCERConfigsHandler.h"

#include <MPEGI/Helpers/Config.h>

namespace MPEGI {

LCERConfigsHandler::LCERConfigsHandler(std::vector<LCERPayload::LCEREnvironment> envs)
    : envs(envs)
{
    uint32_t nEQNBands = Config::getInt("EQNBands");
    const float* eqCenterFreqs = Config::getFloatPtr("EQCenterFreqs");
    std::vector<float> centerFreqs(eqCenterFreqs, eqCenterFreqs + nEQNBands);

    isBitstreamConfigsHandler = true;

    for (auto& env : envs) {
        if (env.type != LCERPayload::LCEREnvironmentType::Outdoor) {
            return;
        }
        configs.push_back(std::make_unique<LCERConfig>(env.outdoorData));
        inverseMatricesForEnvironments.push_back(
            EAR::HomogenMatrix(env.outdoorData.boundingBox.position).inverse());
    }

    //for outdoor take the environment with the lowest priority as the default for now
    defaultConfig = std::make_unique<LCERConfig>(envs.back().outdoorData);
}

LCERConfigsHandler::LCERConfigsHandler(const ReverbEnv* reverbEnv)
{
    isBitstreamConfigsHandler = false;

    float defaultPredelay = Config::getFloatDef("LCEarlyDefaultPredelay", 15.f);
    float defaultRT60 = Config::getFloatDef("LCEarlyDefaultRT60", 0.0f);

    if (reverbEnv) {

        float pd = -1.0f;
        float epsilon = 0.0001;
        if (reverbEnv->predelay.has_value() && reverbEnv->predelay.value() > epsilon) {
            //s to ms
            pd = reverbEnv->predelay.value() * 1000.f;
        } else {
            //there is no predelay defined - default to 15 ms
            pd = defaultPredelay;
        }

        float rt60_1 = defaultRT60;
        float rt60_4 = defaultRT60;
        for (auto& freqData : reverbEnv->frequencyDecayResponse) {
            float epsilon = 0.0001;
            if (freqData.frequency - 1000 < epsilon) {
                rt60_1 = freqData.rt60;
            }
            if (freqData.frequency - 4000 < epsilon) {
                rt60_4 = freqData.rt60;
            }
        }

        defaultConfig = LCERConfig::spiralConfigFromEIFValues(pd, rt60_1, rt60_4);

        //for long narrow rooms update initDelay and ampfa
        auto boundingBox = reverbEnv->enclosure->getBoundingBox();
        auto sizeX = boundingBox.at(0).high - boundingBox.at(0).low;
        auto sizeZ = boundingBox.at(2).high - boundingBox.at(2).low;
        if ((sizeX < 3 || sizeZ < 3) && std::max(sizeX / sizeZ, sizeZ / sizeX) > 6) {
            defaultConfig->initDelay = 10.0f;
            defaultConfig->ampFac = 2.0;
        }
    } else {
        // default outdoor config pattern
        defaultConfig = LCERConfig::defaultOutdoorConfig();
    }
}

LCERConfig* LCERConfigsHandler::configForListenerLocation(EAR::Vector3 location)
{
    if (isBitstreamConfigsHandler) {
        for (int i = 0; i < envs.size(); i++) {
            auto boundingBox = envs.at(i).outdoorData.boundingBox;
            auto inverseMatrix = inverseMatricesForEnvironments.at(i);
            if (containsLocation(boundingBox, inverseMatrix, location)) {
                return configs.at(i).get();
            }
        }
    }
    return defaultConfig.get();
}

LCERConfig* LCERConfigsHandler::defaultConf()
{
    return defaultConfig.get();
}

bool LCERConfigsHandler::containsLocation(
    LCERPayload::BoundingBox& box, EAR::HomogenMatrix& inverseMatrix, EAR::Vector3 location)
{
    auto transform = inverseMatrix * location;
    if (transform[0] >= -box.size[0] / 2 && transform[0] <= box.size[0] / 2
        && transform[1] >= -box.size[1] / 2 && transform[1] <= box.size[1] / 2
        && transform[2] >= -box.size[2] / 2 && transform[2] <= box.size[2] / 2) {
        return true;
    }
    return false;
}

} // namespace MPEGI
