/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "LCERPatternProducer.h"
#include <cfloat>

namespace MPEGI {

constexpr std::size_t SPEED_OF_SOUND{ 343 };

LCERPatternProducer::LCERPatternProducer(LCERConfig* config)
    : config(config)
{
    Config::observeKey("LCER_OutdoorCompression", this);
    Config::observeKey("LCER_AmpFac", this);
    Config::observeKey("LCER_NumErSpiral", this);
    Config::observeKey("LCER_NumErOutdoor", this);
    Config::observeKey("LCER_Predelay", this);
}

LCERPatternProducer::~LCERPatternProducer()
{
    Config::removeObserver("LCER_OutdoorCompression", this);
    Config::removeObserver("LCER_AmpFac", this);
    Config::removeObserver("LCER_NumErSpiral", this);
    Config::removeObserver("LCER_NumErOutdoor", this);
    Config::removeObserver("LCER_Predelay", this);
}

void LCERPatternProducer::updateConfig(LCERConfig* newConfig, bool resetEnvCrossfadeFactor)
{
    std::lock_guard<std::mutex> lock(changeParamsMutex);
    if (resetEnvCrossfadeFactor) {
        envCrossfadeFactor = 1.0f;
    }
    config = newConfig;
}

LCERPatternProducer::Pattern LCERPatternProducer::process(
    EAR::Vector3 listenerLoc, EAR::Vector3 itemLoc)
{
    switch (config->algoType) {
    case LCERConfig::AlgoType::Spiral:
        return spiral(listenerLoc, itemLoc);
    case LCERConfig::AlgoType::Outdoor:
        return outdoor(listenerLoc, itemLoc);
    case LCERConfig::AlgoType::NoER:
    default:
        LCERPatternProducer::Pattern result;
        return result;
    }
}

void LCERPatternProducer::applyEnvCrossfadeFactor(float factor)
{
    const std::lock_guard<std::mutex> lock(changeParamsMutex);
    envCrossfadeFactor = factor;
}

size_t LCERPatternProducer::getNumberOfEarlyReflectionItems()
{
    const std::lock_guard<std::mutex> lock(changeParamsMutex);
    return config->numberOfERs;
}

LCERPatternProducer::Pattern LCERPatternProducer::spiral(
    EAR::Vector3 listenerLoc, EAR::Vector3 itemLoc)
{
    std::lock_guard<std::mutex> lock(changeParamsMutex);
    auto lsVector = itemLoc - listenerLoc;
    float sld = lsVector.abs();
    auto lsPolar = EAR::cartesianToPolar(lsVector);
    float yawOffset = lsPolar[0];

    float sourceListenerDistance = std::max(1.0f, sld);
    const float PHI = 1.85f; // growth factor fits to normal ER timing of a shoebox room
    const int ANGLE_OFFSET = -10; // [degree]
    const float AMPLITUDE_LIMIT = 0.7f;

    LCERPatternProducer::Pattern result;
    float angleOffsetrad = ANGLE_OFFSET * PI / 180.0f;

    float distoffset = 0.001 * config->initDelay * SPEED_OF_SOUND;

    // calculate pattern rotation offset depending on listener location
    float movementFac = Config::getFloatDef("LCEarlyMovementFac", 0.25f);
    float movementAngleOffs = listenerLoc[0] / distoffset * movementFac * PI
        + listenerLoc[2] / distoffset * movementFac * PI;

    movementAngleOffs = std::fmod(movementAngleOffs, 2.0f * PI);

    float beta1base = angleOffsetrad + PI / 3.0f;

    // calculate amplitude for first ER and apply limitation if necessary
    float beta1tmp = angleOffsetrad + PI / 3.0f;
    float r1base = config->distFac * std::pow(PHI, 2.0f * beta1base / PI);
    float r1tmp = distoffset + config->distFac * std::pow(PHI, 2.0f * beta1tmp / PI);
    r1tmp = r1tmp - config->distFac * std::pow(PHI, 2.0f * beta1base / PI); // to be checked

    // attenuate pattern to limite amplitude to 0.7 = -3dB for large slDistances
    float ampAtt = 1.0f;
    float tmpAmp = config->ampFac * (1.0f - config->absorption) * sourceListenerDistance / r1tmp;
    if (tmpAmp > AMPLITUDE_LIMIT) {
        ampAtt = AMPLITUDE_LIMIT * r1tmp / (sourceListenerDistance * (1.0f - config->absorption))
            / config->ampFac;
    }

    for (int i = 0; i < config->numberOfERs / 2; i++) {
        float beta1 = angleOffsetrad + (i + 1) * PI / 3.0f;
        float r1 = distoffset + config->distFac * std::pow(PHI, 2.0f * beta1 / PI) - r1base;

        beta1 = std::fmod(beta1, 2.0f * PI);
        ERPos pos;
        pos.azimuth = std::fmod(beta1 + movementAngleOffs, 2.0f * PI);
        pos.elevation = 0;
        pos.radius = r1;
        // distReducFactor must be calculated before level limitation
        pos.distReducFactor =
            ampAtt * config->ampFac * (1.0f - config->absorption) * sourceListenerDistance / r1;
        pos.distReducFactor = std::min(pos.distReducFactor, AMPLITUDE_LIMIT); // limit amplitude
        // envCrossfadeFactor is for crossfade between environments => apply after limitation
        pos.amplitude = pos.distReducFactor * envCrossfadeFactor;
        auto localERPolar = pos.polar();
        pos.globalPosition = listenerLoc + EAR::polarToCartesian(localERPolar);
        result.erpos.push_back(pos);

        float beta2 = beta1 + PI / 8.0f;
        float r2 = distoffset + config->distFac * std::pow(PHI, 2 * beta2 / PI);
        r2 = r2 - config->distFac * std::pow(PHI, 2 * beta1base / PI);

        beta2 = beta2 + PI;
        beta2 = std::fmod(beta2, 2.0f * PI);
        ERPos pos2;
        pos2.azimuth = std::fmod(beta2 + movementAngleOffs, 2.0f * PI);
        pos2.elevation = 0;
        pos2.radius = r2;
        pos2.distReducFactor =
            ampAtt * config->ampFac * (1 - config->absorption) * sourceListenerDistance / r2;
        pos2.distReducFactor = std::min(pos2.distReducFactor, AMPLITUDE_LIMIT);
        pos2.amplitude = pos2.distReducFactor * envCrossfadeFactor;
        auto localERPolar2 = pos2.polar();
        pos2.globalPosition = listenerLoc + EAR::polarToCartesian(localERPolar2);
        result.erpos.push_back(pos2);
    }

    return result;
}

LCERPatternProducer::Pattern LCERPatternProducer::outdoor(
    EAR::Vector3 listenerLoc, EAR::Vector3 itemLoc)
{
    std::lock_guard<std::mutex> lock(changeParamsMutex);
    auto lsVector = itemLoc - listenerLoc;
    float sld = lsVector.abs();
    auto lsPolar = EAR::cartesianToPolar(lsVector);
    float yawOffset = lsPolar[0];

    float sourceListenerDistance = std::max(1.0f, sld);

    float betaOutdoor[] = { 0.44f, 1.84f, 3.2f, 4.7f };
    float rstep = 0.2;

    float preDelay1 = config->preDelay1;
    float preDelay2 = config->preDelay2;
    float ampFac1 = config->ampFac1;
    float ampFac2 = config->ampFac2;

    // sanity check bitstream values -> nonzero predelay
    preDelay1 = std::max(preDelay1, 0.1f);
    preDelay2 = std::max(preDelay2, 0.1f);

    float rstart1 = 0.001 * preDelay1 * SPEED_OF_SOUND;
    float rstart2 = 0.001 * preDelay2 * SPEED_OF_SOUND;

    LCERPatternProducer::Pattern result;
    for (int i = 0; i < config->numberOfERs; i++) {
        float beta = betaOutdoor[i];
        ERPos pos;
        pos.azimuth = std::fmod(beta, 2.0f * PI);
        pos.elevation = 0;
        if (i % 2 == 0) {
            pos.radius = (0.7f + i * rstep) * rstart1;
            pos.distReducFactor = ampFac1;
        } else {
            pos.radius = (0.7f + i * rstep) * rstart2;
            pos.distReducFactor = ampFac2;
        }
        pos.distReducFactor *= envCrossfadeFactor;

        float reflDistGain = 1.0f / (pos.radius + FLT_MIN);
        pos.amplitude = pos.distReducFactor * reflDistGain;
        auto localERPolar = pos.polar();
        pos.globalPosition = listenerLoc + EAR::polarToCartesian(localERPolar);
        result.erpos.push_back(pos);
    }

    return result;
}

void LCERPatternProducer::configValueChanged(const std::string& key)
{
    if (key == "LCER_AmpFac") {
        const std::lock_guard<std::mutex> lock(changeParamsMutex);
        config->ampFac = Config::getFloat("LCER_AmpFac");
    }

    if (key == "LCER_NumErSpiral") {
        const std::lock_guard<std::mutex> lock(changeParamsMutex);
        int numberOfERInSpiral = static_cast<int>(Config::getFloat("LCER_NumErSpiral"));
        if (numberOfERInSpiral % 2 == 0) {
            config->numberOfERs = numberOfERInSpiral;
        }
    }

    if (key == "LCER_NumErOutdoor") {
        const std::lock_guard<std::mutex> lock(changeParamsMutex);
        config->numberOfERs = static_cast<int>(Config::getFloat("LCER_NumErOutdoor"));
    }

    if (key == "LCER_Predelay") {
        const std::lock_guard<std::mutex> lock(changeParamsMutex);
        config->preDelay = Config::getFloat("LCER_Predelay");
    }
}

} // namespace MPEGI
