/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "LCEarlyReflection.h"
#include "LCERConfigsHandler.h"
#include "LCERPatternProducer.h"

#include <MPEGI/Bitstream/EarlyPayload.h>
#include <MPEGI/Bitstream/PortalPayload.h>

#include <MPEGI/Renderer/RenderItemStore.h>

#include <EAR/BaseTypes.h>
#include <EAR/DistanceModels.h>

#include <cfloat>
#include <cmath>
#include <memory>

namespace MPEGI {

constexpr float MINIMUM_REFLITEM_DISTANCE{ 5.0 };

std::shared_ptr<RenderStage> makeLCEarlyReflectionStage()
{
    auto const useLowComplexityER = Config::getBool("useLowComplexityER");

    if (useLowComplexityER)
        return std::make_shared<LCEarlyReflectionStage>();
    else
        return std::make_shared<RenderStageNoOp>();
}

LCEarlyReflectionStage::LCEarlyReflectionStage()
{
    m_portalFadeDistance = Config::getFloatDef("LCEarlyPortalFadeDistance", 0.2f);
    m_lcEarlyMasterGain = Config::getFloatDef("LCEarlyMasterGain", .6f);
    m_printCurrentConfigOnce = false;
    Config::observeKey("LCEarlyPrintCurrentConfigOnce", this);
    Config::observeKey("LCEarlyPortalFadeDistance", this);
    Config::observeKey("LCEarlyMasterGain", this);
}

LCEarlyReflectionStage::~LCEarlyReflectionStage()
{
    Config::removeObserver("LCEarlyPrintCurrentConfigOnce", this);
    Config::removeObserver("LCEarlyPortalFadeDistance", this);
    Config::removeObserver("LCEarlyMasterGain", this);
}

std::unique_ptr<RenderStage::DspData> LCEarlyReflectionStage::update(RenderList& renderList)
{
    if (!itemStore) {
        return {};
    }

    auto dspData = std::make_unique<LCEarlyReflectionStage::DspData>();
    auto& dspLists = dspData->dspLists;

    float speedOfSound = Config::getFloat("SpeedOfSound");

    updatePatternProducers(renderList);

    //reset reverbState and create dsplist
    dspLists = std::vector<DSPList>();
    for (auto& reverbState : reverbStates) {
        reverbState.second.primaryItems.clear();
    }

    for (auto& item : renderList.items) {
        if (irrelevantRenderItem(*item)) {
            continue;
        }

        auto& itemState = item->getStateOrCreate<LCEarlyReflectionStage::ItemState>(*item.get());
        auto& reverbState = reverbStates[itemState.reverbId];
        if (item->status != ItemStatus::Inactive) {
            reverbState.primaryItems.push_back(item.get());
        }
    }

    const EAR::Vector3& lPos = renderList.listener.position.location;
    for (auto& reverbState : reverbStates) {
        auto primaryItems = reverbState.second.primaryItems;
        auto reflItems = reverbState.second.reflectionItems;

        //if the current reverbState does not contain active items, skip it
        if (primaryItems.size() == 0) {
            for (auto reflItem : reflItems) {
                itemStore->deactivate(reflItem);
            }
            continue;
        }

        //skip, if the listener is in a different reverb environment and also not
        //close to the current one
        bool isUserCloseToPortal = closeToPortalCheck(renderList.listener, reverbState.second);
        float epsilon = 0.001;
        if (m_portalFadeDistance < epsilon)
            isUserCloseToPortal = false;
        if (reverbState.first != renderList.listener.acousticEnvironmentId
            && !isUserCloseToPortal) {
            for (auto reflItem : reflItems) {
                itemStore->deactivate(reflItem);
            }
            continue;
        }

        //we need to handle this reverb state, so add a DSPList for it
        dspLists.push_back(DSPList());

        //create dsp items and reactivate reflection items if necessary
        for (auto reflItem : reflItems) {
            if (reflItem->status == ItemStatus::FadeOutInactive
                || reflItem->status == ItemStatus::Inactive) {
                reflItem->setStatus(ItemStatus::Active, true);
            }
            auto dspItem = DSPItem();
            dspLists.back().push_back(dspItem);
        }
        auto& patternProducer = reverbState.second.patternProducer;

        // look for closest render item and generate pattern
        float minSlDist = (lPos - primaryItems[0]->position.location).abs();
        auto closestItem = primaryItems[0];
        for (auto item : primaryItems) {
            float tmpSlDist = (lPos - item->position.location).abs();
            if (tmpSlDist < minSlDist) {
                minSlDist = tmpSlDist;
                closestItem = item;
            }
        }
        auto patternClosest = patternProducer->process(lPos, closestItem->position.location);

        for (auto item : primaryItems) {
            auto& xs = item->position.location;
            auto xsPolar = EAR::cartesianToPolar(xs - lPos);
            //just produce pattern for item and apply
            auto pattern = patternProducer->process(lPos, item->position.location);

            for (int i = 0; i < pattern.erpos.size(); i++) {
                auto globalPosition = pattern.erpos.at(i).globalPosition;
                auto& reflItem = reflItems.at(i);
                auto& reflItemPosition = reflItem->position.location;

                //this is a bit redundant for items with a fixed position relative to the listener:
                reflItem->setPosition(Position(globalPosition, Euler(0, 0, 0)));
                auto& dspItem = dspLists.back().at(i);
                dspItem.earlyReflectionsBufferIn.push_back(&item->signal->getFrame());

                float tmpGain = 1.0f;
                float slGain =
                    distanceModel->getGainForDistance((xs - lPos).abs(), 0, 0, item->refDistance);

                if (item->extent || (item->type == ItemType::Representative)) {
                    // extended sources representative items not gain adjusted in distance stage
                    // do not apply extra gain for clustered RI, TODO: better way than string compare?
                    if (item->objectName.compare(0, 8, "Cluster_")) {
                        tmpGain = slGain * Config::getFloatDef("LCEarlyExtendedSourceGain", 0.1f);
                    }
                } else if (item->gain > slGain) {
                    // general fallback: limit total item gain to distance-based gain
                    tmpGain = slGain / (item->gain + FLT_MIN);
                }

                dspItem.earlyReflectionsPrimaryGains.emplace_back(tmpGain);

                auto gainFactor = pattern.erpos.at(i).amplitude * m_lcEarlyMasterGain;

                dspItem.earlyReflectionsGainsIn.emplace_back(gainFactor);
                dspItem.earlyReflectionsOut = &reflItems.at(i)->signal->getFrame();

                float delaySeconds = pattern.erpos.at(i).radius / speedOfSound;
                int delaySamples = (int)(delaySeconds * fs);
                dspItem.earlyReflectionsIntegerDelay.emplace_back(delaySamples);
            }
        }

        //add to renderlist
        for (auto reflItem : reflItems) {
            if (reflItem->status == ItemStatus::Active
                || reflItem->status == ItemStatus::FadeOutInactive
                || reflItem->status == ItemStatus::FadeOutOrphaned) {
                renderList.addItemCopy(reflItem);
                reflItem->resetChanged();
            }
        }
    }

    return dspData;
}

void LCEarlyReflectionStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<LCEarlyReflectionStage::DspData>();

    int numBufferedSamples = 0;

    int itemCtr = 0;
    for (auto dspList : currentDSP.dspLists) {
        for (auto& dspListItem : dspList) {

            std::vector<EAR::RealFrame*>& buffersIn = dspListItem.earlyReflectionsBufferIn;
            std::vector<float>& gains = dspListItem.earlyReflectionsGainsIn;
            std::vector<float>& primaryGains = dspListItem.earlyReflectionsPrimaryGains;
            std::vector<int>& delaySamples = dspListItem.earlyReflectionsIntegerDelay;

            itemCtr++;

            if (buffersIn.size() > 0) {

                int bufferLenIn = buffersIn.at(0)->ch(0).length();

                if (itemCtr == 1) {
                    for (int i = 0; i < bufferLenIn; i++) {
                        ringBuffer[(ringBufferPos + i) % ringBufferLength] = 0.0f;
                    }
                    for (int primary = 0; primary < buffersIn.size(); primary++) {
                        float* inBuf = buffersIn.at(primary)->ch(0).data();
                        float w = primaryGains.at(primary);
                        for (int i = 0; i < bufferLenIn; i++) {
                            ringBuffer[(ringBufferPos + i) % ringBufferLength] += inBuf[i] * w;
                        }
                        numBufferedSamples = bufferLenIn;
                    }
                }

                auto buffersOut = dspListItem.earlyReflectionsOut;

                float* outBuf = buffersOut->ch(0).data();
                float w = gains.at(0);
                int tmpDelay = dspListItem.earlyReflectionsIntegerDelay.at(0);
                for (int i = 0; i < bufferLenIn; i++) {
                    //calculate read position index with double modulus to account for negative indices;
                    int readPos = ringBufferPos - tmpDelay + i;
                    readPos = ((readPos % ringBufferLength) + ringBufferLength) % ringBufferLength;

                    outBuf[i] = ringBuffer[readPos] * w;
                }
            }
        }
    }

    if (numBufferedSamples > 0) {
        ringBufferPos += numBufferedSamples;
        ringBufferPos = ringBufferPos % ringBufferLength;
    }
}

void LCEarlyReflectionStage::configValueChanged(const std::string& key)
{
    if (key == "LCEarlyPrintCurrentConfigOnce") {
        m_printCurrentConfigOnce =
            static_cast<bool>(Config::getFloat("LCEarlyPrintCurrentConfigOnce"));
    }

    if (key == "LCEarlyPortalFadeDistance") {
        m_portalFadeDistance = Config::getFloat("LCEarlyPortalFadeDistance");
    }

    if (key == "LCEarlyMasterGain") {
        m_lcEarlyMasterGain = Config::getFloat("LCEarlyMasterGain");
    }
}

void LCEarlyReflectionStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    auto lcERPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_LCER);
    EAR_CHECK_REL(lcERPayloads.size() <= 1, "LCEarlyReflection: too many payloads in bitstream");
    if (lcERPayloads.size() == 1) {
        lcEarlyPayload = std::dynamic_pointer_cast<MPEGI::LCERPayload const>(lcERPayloads[0]);
    }

    if (!state) {
        return;
    }

    // get tuning gain from HQ-ER Bitstream Payload
    std::shared_ptr<const MPEGI::EarlyPayload> earlyPayload{ nullptr };
    try {
        earlyPayload = std::dynamic_pointer_cast<const MPEGI::EarlyPayload>(
            bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_ER));
    } catch (std::exception&) {
        // Stage can work without payload PLD_ER
    }
    if (earlyPayload) {
        // renderer tuning parameters

        float gain_lin = Config::getFloatDef("EarlyTuningGain",
            std::pow(10.f, BitstreamConfig::early_tuning_gain_default_db * .05f));
        gain_lin *= Config::getFloatDef("LCEarlyRelativeTuningGain", 0.75f);
        Config::set("LCEarlyMasterGain", EAR::Variant(gain_lin));
        configValueChanged("LCEarlyMasterGain");
    }

    distanceModel = std::make_unique<EAR::DistanceGainVolumetric>(
        Config::getFloatDef("Distance_MinDistance", 0.2f));

    this->itemStore = itemStore;

    initializeReverbStates(state);

    initializeReflectionItemsPerReverbState(streamManager);

    // Add portals
    auto portalPayload = std::dynamic_pointer_cast<const PortalPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_PORTAL));
    for (auto& p : portalPayload->getPortals()) {
        aeToPortals[p->sourceId].push_back(state->getGeometry(p->extentId.value()));
    }

    fs = streamManager->getFs();
}

void LCEarlyReflectionStage::initializeReverbStates(const SceneState* state)
{
    //create default map entries for all acoustic environments + outdoor
    createDefaultReverbStates(state);

    maybeCreateBitstreamReverbStates(state);
}

void LCEarlyReflectionStage::createDefaultReverbStates(const SceneState* state)
{
    // we don't want to use the default acoustic environment at all: currently bitstream data is
    // mapped with the id "-1" (also used if no default ac env is available) to the respective user
    // regions.
    bool defaultAcousticEnvironmentFound = false;
    for (auto& r : state->getReverbEnvs()) {
        auto identifier = r.get()->acousticEnvironmentId;

        if (r->enclosure == nullptr) {
            // reverb environment without geometry is a default one
            addOutdoorDefaultReverbState(r->acousticEnvironmentId);
            defaultAcousticEnvironmentFound = true;
        } else {
            auto& newState = reverbStates[identifier];
            newState.reverbEnv = r.get();
            newState.configsHandler = std::make_unique<LCERConfigsHandler>(r.get());
            newState.patternProducer =
                std::make_unique<LCERPatternProducer>(newState.configsHandler->defaultConf());
            newState.distanceModelForReflectionItem = std::make_unique<OneOverRDistance>(1.0f,
                Config::getFloatDef("Distance_MaxDistance", 500.f),
                Config::getFloatDef("Distance_MaxDistance", 500.f) + 10, 0.85f);
        }
    }

    if (!defaultAcousticEnvironmentFound) {
        addOutdoorDefaultReverbState("");
    }
}

void LCEarlyReflectionStage::addOutdoorDefaultReverbState(const std::string& identifier)
{
    outdoorReverbId = identifier;
    auto& outdoorState = reverbStates[outdoorReverbId];
    outdoorState.reverbEnv = nullptr;
    outdoorState.configsHandler = std::make_unique<LCERConfigsHandler>(nullptr);
    outdoorState.patternProducer =
        std::make_unique<LCERPatternProducer>(outdoorState.configsHandler->defaultConf());
}

void LCEarlyReflectionStage::maybeCreateBitstreamReverbStates(const SceneState* state)
{
    if (lcEarlyPayload == nullptr) {
        return;
    }

    //go through bistream areas and overwrite if needed:
    for (auto& reverbState : reverbStates) {
        if (reverbState.second.reverbEnv
            && reverbState.second.reverbEnv->acousticEnvironmentId != "") {
            //ac env indoor
            auto acEnvId = reverbState.second.reverbEnv->acousticEnvironmentId;
            if (lcEarlyPayload->envs.find(acEnvId) != lcEarlyPayload->envs.end()) {
                auto envPayload = lcEarlyPayload->envs.at(acEnvId);
                reverbState.second.configsHandler =
                    std::make_unique<LCERConfigsHandler>(envPayload);
                reverbState.second.patternProducer = std::make_unique<LCERPatternProducer>(
                    reverbState.second.configsHandler->defaultConf());
                reverbState.second.distanceModelForReflectionItem =
                    std::make_unique<OneOverRDistance>(1.0f,
                        Config::getFloatDef("Distance_MaxDistance", 500.f),
                        Config::getFloatDef("Distance_MaxDistance", 500.f) + 10, 0.85f);
            }
        } else {
            if (reverbState.first == outdoorReverbId) {
                // bitstream outdoor data is saved with the identifier "-1"
                if (lcEarlyPayload->envs.find("") != lcEarlyPayload->envs.end()) {
                    auto envPayload = lcEarlyPayload->envs.at("");
                    reverbState.second.configsHandler =
                        std::make_unique<LCERConfigsHandler>(envPayload);
                    reverbState.second.patternProducer = std::make_unique<LCERPatternProducer>(
                        reverbState.second.configsHandler->defaultConf());
                }
            }
        }
    }
}

void LCEarlyReflectionStage::initializeReflectionItemsPerReverbState(
    const StreamManager* streamManager)
{
    for (auto& reverbState : reverbStates) {
        auto& reflectionItems = reverbState.second.reflectionItems;
        std::string acEnvName = "acEnv_" + reverbState.first;
        if (reverbState.first == outdoorReverbId) {
            acEnvName = "outdoor";
        }

        auto numERForReverbState =
            reverbState.second.patternProducer->getNumberOfEarlyReflectionItems();
        for (int i = 0; i < numERForReverbState; i++) {
            reflectionItems.push_back(
                itemStore->createItem(acEnvName + "_refl_" + std::to_string(i)));
            reflectionItems.back()->type = ItemType::Reflection;
            reflectionItems.back()->status = ItemStatus::Inactive;
            reflectionItems.back()->vdlMethod = VDLMethod::linear;
            reflectionItems.back()->setSignal(streamManager->createEmptyBuffer(1));
            reflectionItems.back()->setAParams(AParam::NoDistanceGain);
        }
    }
}

void LCEarlyReflectionStage::updatePatternProducers(RenderList& renderList)
{
    std::vector<std::shared_ptr<RenderItem>> activeItemsInCurrentAcEnv;
    for (auto& reverbState : reverbStates) {
        if (reverbState.second.patternProducer) {
            auto listenerLocation = renderList.listener.position.location;
            auto config =
                reverbState.second.configsHandler->configForListenerLocation(listenerLocation);
            reverbState.second.patternProducer->updateConfig(config);
            if (reverbState.first == renderList.listener.acousticEnvironmentId
                && m_printCurrentConfigOnce) {
                std::cout << "config: " << *config << std::endl;
                m_printCurrentConfigOnce = !m_printCurrentConfigOnce;
            }
        }
    }
}

bool LCEarlyReflectionStage::closeToPortalCheck(
    const RenderList::ListenerState& listener, ReverbState& reverbState)
{
    auto reverbEnv = reverbState.reverbEnv;
    if (reverbEnv && reverbEnv->acousticEnvironmentId != "") {
        auto acEnvId = reverbEnv->acousticEnvironmentId;

        if (aeToPortals.find(acEnvId) != aeToPortals.end()) {
            auto portalGeometriesToCheck = aeToPortals.at(acEnvId);
            for (auto& portalGeometry : portalGeometriesToCheck) {
                auto closestPoint = portalGeometry->closestPointTo(listener.position.location);
                auto distance = (listener.position.location - closestPoint).abs();
                const float distLimit = m_portalFadeDistance;
                if (distance <= distLimit) {
                    float scalingFactor = (distLimit - distance) / distLimit;
                    reverbState.patternProducer->applyEnvCrossfadeFactor(scalingFactor);
                    return true;
                }
            }
        }
    }
    return false;
}

bool LCEarlyReflectionStage::irrelevantRenderItem(const RenderItem& item) const
{
    // image sources should only be rendered for primary and representative items
    // also, feed diffraction items into LCER to avoid drop in occluded positions
    if (item.type != ItemType::Primary && item.type != ItemType::Representative
        && item.type != ItemType::Diffraction) {
        return true;
    }

    // skip listener-centric (channel) sources
    if (item.position.cSpace == EAR::CSpace::LCS) {
        return true;
    }

    return false;
}

} // namespace MPEGI
