/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_LC_EARLY_REFLECTION
#define MPEGI_LC_EARLY_REFLECTION

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include "OneOverRDistance.h"

#include <MPEGI/Bitstream/LCERPayload.h>

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/ConfigObserver.h>

#include <atomic>
#include <map>
#include <mutex>

#define LCER_RINGBUFFER_LENGTH 16000

// fwd declares
namespace EAR {
class DistanceGainVolumetric;
} // namespace EAR

namespace MPEGI {

class LCERConfigsHandler;
class LCERPatternProducer;

//! The LCEarlyReflectionStage applies a combined reflection pattern for all sound sources
// * to render a simple perceptual reflection effect.
/*!
 * Inputs:
 * - All items of ItemType::Primary, ItemType::Representative and ItemType::Diffusion
 *
 * Outputs:
 * - New items of ItemType::Reflection, with positions defined by the reflection pattern
 *   and signals from the combined input sound sources using multiple taps of an integer delay line 
 *
 * How it works:
 * - A single reflection pattern is used to represent the overall reflection characteristics
 *   of an environment. All sound sources within that environment are fed into the combined
 *   reflection pattern. Therefore, the number of reflections and computational complexity
 *   remains independent of the number of sound sources.
 * - This is done after the DistanceStage, therefore the individual source-listener
 *   distance dependent gain and delay are already applied to the signals feeding into the pattern
 * - The delays within the reflection pattern are realized by multiple taps in one integer delay line
 *   for low complexity, independent of the number of sound sources and reflections
 *
 * Stages order:
 * - Required before: DirectivityStage, DistanceStage, EqStage,
 * - Required after: FadeStage
 */
class LCEarlyReflectionStage : public ConfigObserver, public RenderStage
{
public:
    //! Default Constructor
    LCEarlyReflectionStage();

    //! Destructor
    ~LCEarlyReflectionStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

    //! Inherited from ConfigObserver
    void configValueChanged(const std::string& key) override;

private:
    friend class Renderer;
    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    void updatePatternProducers(RenderList& renderList);

    struct ItemState : RenderItem::ExtraState
    {
        explicit ItemState(const RenderItem& item)
            : reverbId{ item.acousticEnvironmentId }
        {}
        std::string reverbId;
    };

    struct ReverbState
    {
        const ReverbEnv* reverbEnv;
        std::vector<RenderItem*> primaryItems;
        std::vector<RenderItem*> reflectionItems;
        std::unique_ptr<LCERPatternProducer> patternProducer;
        std::unique_ptr<LCERConfigsHandler> configsHandler;
        std::unique_ptr<OneOverRDistance> distanceModelForReflectionItem;
    };

    RenderItemStore* itemStore;

    std::shared_ptr<const MPEGI::LCERPayload> lcEarlyPayload{ nullptr };

    std::map<std::string, std::vector<EAR::Geometry*>> aeToPortals;

    bool irrelevantRenderItem(const RenderItem& item) const;

    void initializeReverbStates(const SceneState* state);
    std::map<std::string, ReverbState> reverbStates;
    std::string outdoorReverbId;
    void createDefaultReverbStates(const SceneState* state);
    void addOutdoorDefaultReverbState(const std::string& identifier);
    void maybeCreateBitstreamReverbStates(const SceneState* state);
    void initializeReflectionItemsPerReverbState(const StreamManager* streamManager);

    struct DSPItem
    {
        std::vector<EAR::RealFrame*> earlyReflectionsBufferIn;
        std::vector<float> earlyReflectionsGainsIn;
        EAR::RealFrame* earlyReflectionsOut;

        std::vector<int> earlyReflectionsIntegerDelay;
        std::vector<float> earlyReflectionsPrimaryGains;
    };
    typedef std::vector<DSPItem> DSPList;
    struct DspData : RenderStage::DspData
    {
        std::vector<DSPList> dspLists;
    };
    std::unique_ptr<EAR::DistanceGainVolumetric> distanceModel;

    //! helper method to check if a listener is close to a portal
    // also applies a scaling factor to the amplitude
    bool closeToPortalCheck(const RenderList::ListenerState& listener, ReverbState& reverbState);

    std::atomic<bool> m_printCurrentConfigOnce;
    std::atomic<float> m_portalFadeDistance;
    std::atomic<float> m_lcEarlyMasterGain;

    // define ringbuffer for integer delay line
    // TODO: optimize by making buffer length scene size dependent
    const int ringBufferLength = LCER_RINGBUFFER_LENGTH;
    float ringBuffer[LCER_RINGBUFFER_LENGTH] = { 0.0f };
    int ringBufferPos = 0;

    float fs;
};

} // namespace MPEGI

#endif /* MPEGI_LC_EARLY_REFLECTION */
