/***************************************************************************
This software module was originally developed by

Huawei

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Huawei retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include "LowCompMphoaStage.h"
#include "MPEGI/Helpers/AudioUtilization.h"

namespace MPEGI {

std::shared_ptr<RenderStage> makeLowCompMphoaStage()
{
    return std::make_shared<LowCompMphoaStage>();
}

LowCompMphoaStage::LowCompMphoaStage()
    : MphoaBaseStage()
{}

LowCompMphoaStage::~LowCompMphoaStage() {}

std::unique_ptr<RenderStage::DspData> LowCompMphoaStage::update(RenderList& renderList)
{
    if (!isInitialized) {
        return {};
    }
    // First, call base class update method
    std::unique_ptr<RenderStage::DspData> dspData = MphoaBaseStage::update(renderList);

    listenerPosition.location[1] = 0.0f; // ignore vertical dimension (like with ESD sources)

    // Prepare rotation matrix to get the relative VLO positions
    float R[3][3], RT[3][3];
    LowCompMphoa::eulerAnglesToRotationMatrix(listenerPosition.orientation[0] * M_PI / 180.0f,
        listenerPosition.orientation[1] * M_PI / 180.0f,
        listenerPosition.orientation[2] * M_PI / 180.0f, R);
    LowCompMphoa::transposeMatrix(R, RT);

    std::vector<MPHOAAudioUtilization> audioUtilization;

    // Loop through all HOA groups and update VBAP weights for each one
    for (auto const translatorIndex :
        dynamic_cast<MphoaBaseStage::DspData*>(dspData.get())->activeTranslatorIndices) {

        auto& hoaTranslator = translatorData[translatorIndex].translator;

        hoaTranslator.preProcess(listenerPosition); // to update interpolation weights

        hoaTranslatorData* hoaTD = hoaTranslator.getHoaTranslatorData();
        float* interpWeights = hoaTD->currentInterpolationWeights;
        int shOrder = hoaTD->shOrder;

        // Loop through all ESD sources and update VBAP weights for each one
        for (auto& esdSource : esdSourceLists[translatorIndex]) {
            esdSource.updateVbapWeights(
                std::sqrt(interpWeights[esdSource.getIndex()]), RT, listenerPosition.location);
        }

        // audioUtilization information for this translator
        for (auto& esdSource : esdSourceLists[translatorIndex]) {
            if (interpWeights[esdSource.getIndex()] > 0.0) {
                audioUtilization.push_back(MPHOAAudioUtilization(
                    translatorData[translatorIndex].hoaSourceIds[esdSource.getIndex()], shOrder));
            }
        }
    }
    // set audio utilization for all translators
    renderingStageObserver->setUtilization(audioUtilization);

    return dspData;
}

void LowCompMphoaStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<LowCompMphoaStage::DspData>();

    if (!isInitialized || currentDSP.dspItems.empty()) {
        return;
    }

    // Binaural output will be accumulated here
    auto& outputBinauralFrame = outputRenderItems[0]->signal->getFrame();
    outputBinauralFrame.zero();

    // Loop through all HOA groups (translators) and update the binaural output on each iteration
    for (auto const translatorIndex : currentDSP.activeTranslatorIndices) {

        auto& inputStream = currentDSP.dspItems[translatorIndex].inputBuffers;

        // Loop through all VBAP nodes and update the binaural output on each iteration
        for (int indexVbapNode = 0;
             indexVbapNode < vbapHelpers[translatorIndex]->getNumberOfNodes(); indexVbapNode++) {

            // Signals panned to the current VBAP node will be accumulated here
            auto& vbapBuffer = tempFrame[0];
            vbapBuffer.zero();

            // Loop through all ESD sources and update the VBAP buffer on each iteration
            for (auto& esdSource : esdSourceLists[translatorIndex]) {
                auto& inputFrame = inputStream[esdSource.getIndex()]->getFrame();
                esdSource.panToVbapNode(inputFrame, indexVbapNode, vbapBuffer);
            }

            // Perform binaural rendering and update binaural output
            hrirFilters[translatorIndex][indexVbapNode]->process(
                vbapBuffer.data(), tempFrame.data().get());
            for (int ear = 0; ear < 2; ear++)
                outputBinauralFrame.ch(ear).add(tempFrame.ch(ear));
        }
    }

    // Apply EQ to match MphoaStage's frequency response
    for (int ch = 0; ch < 2; ch++) {
        for (int index_filter = 0; index_filter < NUM_BIQUADS; index_filter++) {
            timbreCorrectionFilters[ch][index_filter]->process(
                outputBinauralFrame.ch(ch).data(), outputBinauralFrame.ch(ch).data(), B);
        }
    }
    outputBinauralFrame.mul(timbreCorrection_gain);

    // Apply MPHOA-specific gain
    if (mphoaGain != 1.F) {
        outputBinauralFrame.mul(mphoaGain);
    }

    // Increase frame counter for each translator (active and inactive)
    for (auto const& data : translatorData) {
        data.translator.increaseFrameCounter();
    }
}

void LowCompMphoaStage::initializeStageAdditional(const Bitstream& bitstream,
    RenderItemStore* itemStore, const StreamManager* streamManager, const SceneState* state,
    size_t translatorCount, const std::vector<size_t>& groupSourceCounts)
{
    assert(translatorCount == groupSourceCounts.size());

    fs = streamManager->getFs();

    esdSourceLists.clear();

    // Loop through all translators (HOA groups) and initialize components for each one
    for (std::size_t translatorIndex{ 0 }; translatorIndex < translatorCount; ++translatorIndex) {

        // HOA group parameters
        auto& hoaTranslator = translatorData[translatorIndex].translator;
        auto h = hoaTranslator.getHoaTranslatorData();

        // Low-complexity-specific parameters
        int distanceFactor = translatorData[translatorIndex].group->distanceFactor;
        int panningGridOrder = translatorData[translatorIndex].group->panningGridOrder;

        // Generate VBAP lookup table
        int vbapOrder;
        if (panningGridOrder == 0) {
            vbapOrder = h->shOrder;
        } else {
            vbapOrder = panningGridOrder;
        }
        std::vector<std::vector<float>> vbapDirectionsTable =
            LowCompMphoa::getEsdSpeakerGrid(vbapOrder);
        vbapHelpers.emplace_back(std::make_unique<LowCompMphoa::VbapHelper>(vbapDirectionsTable));

        // Preprocess HRTF (keep VBAP directions with frequency-dependent time alignment)
        hrirFilters.emplace_back(LowCompMphoa::preprocessHrtf(h, vbapDirectionsTable, fs, B));

        // VLO factors
        float directivityFactor = LowCompMphoa::getDirectivityFactor(h->shOrder);
        std::vector<float> reverberationFactors =
            LowCompMphoa::getReverberationFactors(bitstream, state, hoaTranslator);

        // Make list of ESD sources, each with its corresponding list of VLOs
        esdSourceLists.emplace_back(std::vector<LowCompMphoa::EsdSource>());
        for (int indexSource = 0; indexSource < groupSourceCounts[translatorIndex]; indexSource++) {

            float medianDistanceToNeighbors = LowCompMphoa::findMedianDistanceToNeighbors(
                hoaTranslator, indexSource, groupSourceCounts[translatorIndex]);

            // Calculate VLO radius and generate esdSource struct
            float ri = medianDistanceToNeighbors * directivityFactor
                * reverberationFactors[indexSource] * distanceFactor;
            EAR::Position esdPosition = hoaTranslator.group().sourcePositions[indexSource];
            LowCompMphoa::EsdSource esdSource(
                esdPosition, ri, indexSource, h->shOrder, *vbapHelpers.back(), B, rampVec, tempVec);
            esdSourceLists.back().emplace_back(esdSource);
        }
    }

    // Allocate buffers
    tempFrame.resize(2, B);
    for (int i = 0; i < B; i++)
        rampVec[i] = (float)i / (float)B;

    // Prepare EQ filter
    for (int ch = 0; ch < 2; ch++) {
        timbreCorrectionFilters.emplace_back(std::vector<std::unique_ptr<EAR::IIRFilter>>());
        for (int index_biquad = 0; index_biquad < NUM_BIQUADS; index_biquad++) {
            timbreCorrectionFilters[ch].emplace_back(EAR::IIRFilter::create(
                timbreCorrection_b[index_biquad], 3, timbreCorrection_a[index_biquad], 3));
        }
    }
}

} // namespace MPEGI
