/***************************************************************************
This software module was originally developed by

Huawei

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Huawei retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#ifndef MPEGI_RENDERERSTAGES_LOWCOMPMPHOA_STAGE
#define MPEGI_RENDERERSTAGES_LOWCOMPMPHOA_STAGE

#include "LowCompMphoaUtilities.hpp"
#include "MPEGI/Renderer/MphoaBaseStage.h"

#include <EAR/IIRFilter.h>
#include <EAR/RealFrame.h>
#include <EAR/StereoFIRFilter.h>

namespace MPEGI {

//! Low complexity multi-point high order Ambisonics renderer
/*! 
 * Inputs: 
 * - One or more HOA groups (TODO: test for >1 group) 
 *  
 * Outputs: 
 * - Donwmix item with binaurally decoded signal
 *  
 * How it works: 
 * - Render HOA signals in the ESD domain by defining virtual loudspeaker objects (VLOs)
 * - Binaurally render VLOs via vector-based amplitude panning
 * 
 * Stages order:
 * - Required before: same as MPHOA stage
 * - Required after: same as MPHOA stage
 *  
 * Config options (set from bitstream): 
 * - distanceFactor: adjust spacing between VLOs (default: 1.0)
 * - panningGridOrder: ESD order of the VLO grid (default: 0 == use highest available order)
 * 
 */
class LowCompMphoaStage : public MphoaBaseStage
{
public:
    /**
    * Class constructor (calls MphoaBaseStage constructor).
    */
    LowCompMphoaStage();

    /**
    * Class destructor.
    */
    ~LowCompMphoaStage();

    /**
    * Implements MphoaBaseStage::update().
    */
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    /**
    * Implements MphoaBaseStage::processBlock().
    */
    void processBlock(DspContext ctx) override;

private:
    /**
    * Called from MphoaBaseStage::initializeStage().
    */
    void initializeStageAdditional(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state, size_t translatorCount,
        const std::vector<size_t>& groupSourceCounts);

    // Parameters
    float fs; // sampling rate in Hz

    // HRIR filters (num HOA groups * num VBAP nodes * 2)
    std::vector<std::vector<std::unique_ptr<EAR::StereoFIRFilter>>> hrirFilters;

    // VBAP helper objects (num HOA groups)
    std::vector<std::unique_ptr<LowCompMphoa::VbapHelper>> vbapHelpers;

    // List of ESD sources (num HOA grups)
    std::vector<std::vector<LowCompMphoa::EsdSource>> esdSourceLists;

    // Aux variables
    EAR::RealFrame tempFrame; // temp frame for VBAP and binaural rendering
    static const int MAX_B = 256; // maximum allowed buffer size (TODO: make larger?)
    float rampVec[MAX_B], tempVec[MAX_B]; // temp buffers for crossfading

    // IIR filter for timbre correction: 5 biquads + scalar gain
    static const int NUM_BIQUADS = 5;
    const float timbreCorrection_gain = 0.9046904377381193;
    const float timbreCorrection_b[NUM_BIQUADS][3] = { { 1.00609673, -1.90167555, 0.90622821 },
        { 0.98587878, -1.8330033, 0.87894348 }, { 1.02867848, -1.82110971, 0.86719006 },
        { 1.08722108, -1.52782705, 0.74551881 }, { 1.39971768, 0.80989364, -0.44128711 } };
    const float timbreCorrection_a[NUM_BIQUADS][3] = { { 1., -1.90167555, 0.91232494 },
        { 1., -1.8330033, 0.86482227 }, { 1., -1.82110971, 0.89586854 },
        { 1., -1.52782705, 0.83273989 }, { 1., 0.80989364, -0.04156943 } };
    std::vector<std::vector<std::unique_ptr<EAR::IIRFilter>>> timbreCorrectionFilters;
};
} // namespace MPEGI
#endif
