#include "LowCompMphoaUtilities.hpp"
#include "saf.h"

#include <EAR/ComplexBuffer.h> // for HRTF preprocessing
#include <EAR/FFT.h> // for HRTF preprocessing
#include <EAR/PchipMinphaseFilterDesigner.h> // for HRTF preprocessing

#include <MPEGI/Bitstream/ReverbPayload.h> // for reverberation factor

#if SAF_USE_INTEL_MKL
#include "mkl.h" // for vsLinearFrac, vsMul, vsAdd
#endif

namespace MPEGI {

namespace LowCompMphoa {

std::vector<std::vector<float>> getEsdSpeakerGrid(int order)
{
    // ESD directions as [inclination1, azimuth1, inclination2 ...], same as in EsdConverter.cpp
    std::vector<float> esdDirsRad{};
    if (order == 1) {
        esdDirsRad = { 0, 0, 1.910633, 0, 1.910633, 2.094395, 1.910633, -2.094 };
    } else if (order == 2) {
        esdDirsRad = { 0, 0, 2.361073, 0, 1.207589, -1.95668, 1.207589, 1.956682, 2.415178,
            -1.95668, 1.561039, -3.14159, 2.415178, 1.956681, 1.325668, 0.687124, 1.325667,
            -0.68712 };
    } else if (order == 3) {
        esdDirsRad = { 0, 0, 0.854098, 0, 2.031969, 1.119907, 2.605106, -0.25283, 1.078622,
            1.155586, 1.736608, 2.040481, 2.031968, -1.38118, 1.736609, 0.270692, 1.56888, -2.20417,
            0.917087, 2.297267, 0.917087, -2.80293, 1.763476, 3.010956, 2.649852, 2.154919,
            1.568881, -0.63529, 0.953962, -1.41973, 2.458122, -2.46809 };
    } else if (order == 4) {
        esdDirsRad = { 0, 0, 0.823218, 0, 1.73912, -2.00759, 0.724297, 1.927637, 1.336281, -1.41208,
            0.871631, -2.10001, 1.263705, 2.512927, 1.440147, 1.667633, 2.248313, 1.442383,
            1.433953, -0.60062, 2.888065, 0.329968, 2.003914, -1.18621, 1.80266, 2.983332, 1.396554,
            -2.69222, 2.170781, 0.507602, 1.952805, 2.208977, 1.58019, 0.952319, 2.584609, -1.71565,
            0.874597, 0.934402, 2.172935, -0.38654, 2.612717, 2.675958, 2.193907, -2.62842, 1.51674,
            0.165012, 0.715307, -1.02504, 0.762553, -3.13121 };
    } else {
        assert(false && "ESD arrangements only defined up to 4th order HOA.");
    }

    // Put into a more manageable format
    std::vector<std::vector<float>> esdDirsDeg;
    for (auto dir = esdDirsRad.begin(); dir != esdDirsRad.end(); dir += 2) {
        float incl = *dir;
        float azi = *(dir + 1);
        esdDirsDeg.push_back(
            std::vector<float>{ azi * 180.0f / (float)M_PI, 90.0f - incl * 180.0f / (float)M_PI });
    }
    return esdDirsDeg;
}

EAR::Vector3 matrixVectorMultiplication(float R[3][3], EAR::Vector3 v)
{
    EAR::Vector3 w;
    w[0] = R[0][0] * v[0] + R[0][1] * v[1] + R[0][2] * v[2];
    w[1] = R[1][0] * v[0] + R[1][1] * v[1] + R[1][2] * v[2];
    w[2] = R[2][0] * v[0] + R[2][1] * v[1] + R[2][2] * v[2];
    return w;
}

void eulerAnglesToRotationMatrix(float yaw_rad, float pitch_rad, float roll_rad, float R[3][3])
{
    float Rx[9]{ 1, 0, 0, 0, std::cos(pitch_rad), -std::sin(pitch_rad), 0, std::sin(pitch_rad),
        std::cos(pitch_rad) };
    float Ry[9]{ std::cos(yaw_rad), 0, std::sin(yaw_rad), 0, 1, 0, -std::sin(yaw_rad), 0,
        std::cos(yaw_rad) };
    float Rz[9]{ std::cos(roll_rad), -std::sin(roll_rad), 0, std::sin(roll_rad), std::cos(roll_rad),
        0, 0, 0, 1 };
    float temp1[9], temp2[9];
    matrixMult3x3(Rz, Rx, temp1);
    matrixMult3x3(temp1, Ry, temp2);
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            R[i][j] = temp2[3 * i + j];
}

void transposeMatrix(float A[3][3], float AT[3][3])
{
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            AT[i][j] = A[j][i];
}

std::vector<std::unique_ptr<EAR::StereoFIRFilter>> preprocessHrtf(
    hoaTranslatorData* h, std::vector<std::vector<float>> directionGrid, float fs, size_t B)
{
    // Prepare output
    std::vector<std::unique_ptr<EAR::StereoFIRFilter>> hrirFilters;

    // Constants (TODO: set from outside?)
    const float EAR_AZ[2] = { M_PI / 2.0, 3.0 * M_PI / 2.0 };
    const float EAR_COLAT[2] = { M_PI / 2.0, M_PI / 2.0 };
    const float R = 0.0875;
    const float C = 343;
    const float CUTOFF = 2000; // Hz
    const float SMOOTH_FACTOR = 1; // octaves

    // Get some parameters
    size_t irlen = h->hrirLen;
    int nears = 2;
    int nbands = irlen / 2 + 1;
    int ndirs = h->nHrtfDirs;
    int ndirsGrid = directionGrid.size();

    // HRTF grid
    float* gridDirsDeg = (float*)malloc(ndirsGrid * 2 * sizeof(float));
    for (int i = 0; i < ndirsGrid; i++) {
        gridDirsDeg[i * 2] = directionGrid[i][0];
        gridDirsDeg[i * 2 + 1] = directionGrid[i][1];
    }
    int* hrtfIdxGrid = (int*)malloc(ndirsGrid * sizeof(int));
    find_closest_grid_points(gridDirsDeg, h->hrtfDirsDeg, hrtfIdxGrid, ndirsGrid, ndirs);

    // Allocate buffers
    EAR::RealFrame currentIR(2, irlen);
    float_complex*** hrtfs = allocateFloatCmplxArray3d(nbands, h->nBinauralChannels, ndirs);
    float* hrtfMag = (float*)calloc1d(nbands, sizeof(float));
    float_complex* hrtf = (float_complex*)calloc1d(nbands, sizeof(float_complex));

    // Frequency vector
    float* freqVec = (float*)malloc(nbands * sizeof(float));
    for (int i = 0; i < nbands; i++)
        freqVec[i] = (float)(i)*fs / (nbands - 1) / 2.0f;

    // HRIR to HRTF and apply diffuse EQ (but NOT phase simplification)
    HRIRs2HRTFs(h->hrirs, ndirs, irlen, irlen, FLATTEN3D(hrtfs)); // instead of afSTFT
    int applyDiffuseEQ = 1;
    int applySimplePhase = 0; // we will do the minimum phase conversion later
    diffuseFieldEqualiseHRTFs(
        ndirs, h->itds, freqVec, nbands, NULL, applyDiffuseEQ, applySimplePhase, FLATTEN3D(hrtfs));

    // Prepare for FDTA and minimum phase
    std::unique_ptr<EAR::RFFT> rfft = EAR::RFFT::create(B);
    std::unique_ptr<EAR::IRFFT> irfft = EAR::IRFFT::create(B);
    EAR::ComplexBuffer complexBuffer(rfft->K);
    std::vector<float> freqBins = EAR::RFFT::binFrequencies(rfft->K, fs);
    float cutoffIndex = 0;
    for (size_t index_freq = 0; index_freq < rfft->K; index_freq++) {
        if (freqBins[index_freq] > CUTOFF)
            break;
        cutoffIndex = index_freq;
    }
    EAR::PchipMinphaseFilterDesigner filterDesigner(freqVec, nbands, irlen, fs);
    float maxItd = 0.0f;
    for (int dir = 0; dir < ndirs; dir++) {
        float thisItd = std::abs(h->itds[dir]);
        if (thisItd > maxItd)
            maxItd = thisItd;
    }
    float fixedDelay = maxItd / 2.0 + 2.0 / fs; // 2 samples safety

    // Loop through the HRIR directions and generate the filters
    for (int indexDirGrid = 0; indexDirGrid < ndirsGrid; indexDirGrid++) {
        // Find index
        int indexHrir = hrtfIdxGrid[indexDirGrid];
        // Generate filters
        float halfItd = h->itds[indexHrir] / 2.0f;
        for (int ear = 0; ear < nears; ear++) {
            // Minimum phase
            for (int band = 0; band < nbands; band++)
                hrtfMag[band] = std::abs(hrtfs[band][ear][indexHrir]);
            filterDesigner.generate(hrtfMag, currentIR.ch(ear));
            // FDTA
            float itdDelay;
            if (ear == 0)
                itdDelay = -halfItd;
            else
                itdDelay = halfItd;
            rfft->transform(currentIR.ch(ear).data(), complexBuffer.data());
            for (size_t indexFreq = 0; indexFreq < rfft->K; indexFreq++) {
                float f = freqBins[indexFreq];
                float alpha = 0.5f + (SMOOTH_FACTOR * std::log(f / CUTOFF)) / std::log(2.0f);
                if (alpha < 0.0f)
                    alpha = 0.0f;
                if (alpha > 1.0f)
                    alpha = 1.0f;
                float p = 2 * M_PI * f * (fixedDelay + (1 - alpha) * itdDelay);
                complexBuffer[indexFreq] *= std::exp(std::complex<float>(0.0f, -p));
            }
            irfft->transform(complexBuffer.data(), currentIR.ch(ear).data());
        }
        currentIR.mul(1.0f / (float)B);

        // Set filter
        hrirFilters.emplace_back(std::make_unique<EAR::StereoFIRFilter>(B, irlen));
        hrirFilters.back()->setFilter(currentIR.data().get());
    }

    free(hrtf);
    free(hrtfMag);
    free(hrtfs);
    free(freqVec);
    free(gridDirsDeg);
    free(hrtfIdxGrid);

    return hrirFilters;
}

float findMedianDistanceToNeighbors(
    mpegcore::HoaTranslator& hoaTranslator, int indexSource, int numEsdSources)
{
    auto h = hoaTranslator.getHoaTranslatorData();

    std::vector<bool> neighbourList(numEsdSources, false);
    std::vector<float> distances; // empty vector

    for (int tri = 0; tri < h->triangleFaces.size(); tri++) {
        bool triangleContainsSource = false;
        for (int vtx = 0; vtx < 3; vtx++) {
            if (h->triangleFaces[tri][vtx] == indexSource) {
                triangleContainsSource = true;
                break;
            }
        }
        if (triangleContainsSource) {
            for (int vtx = 0; vtx < 3; vtx++) {
                int neighbour = h->triangleFaces[tri][vtx];
                if (neighbour != indexSource && !neighbourList[neighbour]) {
                    EAR::Vector3 posSrc =
                        hoaTranslator.group().sourcePositions[indexSource].location;
                    EAR::Vector3 posNgb = hoaTranslator.group().sourcePositions[neighbour].location;
                    EAR::Vector3 srcToNgb = posNgb - posSrc;
                    float distance = srcToNgb.abs();
                    distances.push_back(distance);
                    neighbourList[neighbour] = true;
                }
            }
        }
    }

    std::sort(distances.begin(), distances.end());
    float medianDistance = distances[distances.size() / 2];

    return medianDistance;
}

float getDirectivityFactor(int order)
{
    if (order <= 1)
        return 1.5f;
    else if (order == 2)
        return 1.25f;
    else
        return 1.0f;
}

float Hz2ERBRateScale(float f)
{
    return 21.4f * log10(1.0f + 0.00437f * f);
}

float freqAverageERB(const std::vector<float> values, const std::vector<float> centerFreqs)
{
    assert(values.size() == centerFreqs.size());
    float totalWidthERB = 0.0f;
    std::vector<float> widthERB(values.size());
    for (int b = 0; b < values.size(); b++) {
        if (b == 0) {
            widthERB[b] = Hz2ERBRateScale(centerFreqs[b + 1]) - Hz2ERBRateScale(centerFreqs[b]);
        } else if (b == values.size() - 1) {
            widthERB[b] = Hz2ERBRateScale(centerFreqs[b]) - Hz2ERBRateScale(centerFreqs[b - 1]);
        } else {
            widthERB[b] =
                0.5 * (Hz2ERBRateScale(centerFreqs[b + 1]) - Hz2ERBRateScale(centerFreqs[b - 1]));
        }
        totalWidthERB += widthERB[b];
    }
    float weightedAvg = 0.0f;
    for (int b = 0; b < values.size(); b++) {
        weightedAvg += values[b] * widthERB[b] / totalWidthERB;
    }
    return weightedAvg;
}

std::vector<float> getReverberationFactors(const Bitstream& bitstream, const SceneState* state,
    const mpegcore::HoaTranslator& hoaTranslator)
{
    int numSources = hoaTranslator.group().sourcePositions.size();
    std::vector<float> rev(numSources, 1.0f); // default return value is 1.0

    auto reverbPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
    EAR_CHECK_REL(reverbPayloads.size() <= 1, "Too many reverb payloads in bitstream");

    if (state != nullptr && reverbPayloads.size() > 0) {

        auto bitstreamReverbPayload =
            std::dynamic_pointer_cast<const MPEGI::ReverbPayload>(reverbPayloads[0]);

        const std::vector<std::shared_ptr<FreqGrid>> freqGridVector =
            bitstreamReverbPayload->GetFreqGrid();

        auto reverbBsParameters = bitstreamReverbPayload->GetReverbParameters();

        auto const& reverbEnvs = state->getReverbEnvs();

        std::vector<float> dsr(numSources);
        float dsr_avg = 0.0f;
        int foundSources = 0;

        for (int src = 0; src < numSources; src++) {

            auto& posSrc = hoaTranslator.group().sourcePositions[src].location;

            for (const auto& r : reverbEnvs) {

                if (r->enclosure != nullptr && r->enclosure->isInside(posSrc)) {
                    if (r->acousticEnvironmentBsId.has_value()) {
                        auto& reverbParams = reverbBsParameters[r->acousticEnvironmentBsId.value()];
                        auto& freqGrid = freqGridVector[reverbParams->freqGridIdx];
                        dsr[src] = freqAverageERB(reverbParams->dsr, freqGrid->centerFreq);
                        dsr_avg += dsr[src];
                        foundSources++;
                    }
                    break;
                }
            }
        }

        // NOTES:
        // if r->enclosure == nullptr --> default AE
        // if posSrc is not inside any AE --> default AE
        // in both cases, dsr[src] == 0 and we assign a rev of 1.0f

        if (foundSources > 0) {
            dsr_avg /= foundSources;
            for (int src = 0; src < numSources; src++) {
                if (dsr[src] > 0.0f)
                    rev[src] = dsr[src] / dsr_avg;
            }
        }
    }

    return rev;
}

VbapHelper::VbapHelper(std::vector<std::vector<float>> vbapDirectionsTable)
    : outputWeights(vbapDirectionsTable.size())
{
    vbap_L = vbapDirectionsTable.size();
    float vbap_dirs[MAX_NUM_VBAP_NODES * 2];
    for (int indexVbap = 0; indexVbap < vbap_L; indexVbap++) {
        vbap_dirs[2 * indexVbap] = vbapDirectionsTable[indexVbap][0];
        vbap_dirs[2 * indexVbap + 1] = vbapDirectionsTable[indexVbap][1];
    }
    float* vbap_gtable;
    int vbap_N_gtable, vbap_nTriangles;
    generateVBAPgainTable3D(vbap_dirs, vbap_L, vbap_az_res, vbap_el_res, vbap_omitLargeTriangles,
        vbap_enableDummies, vbap_spread, &vbap_gtable, &vbap_N_gtable,
        &vbap_nTriangles); // this allocates memory
    VBAPgainTable2InterpTable(
        vbap_gtable, vbap_N_gtable, vbap_L); // re-normalize so that sum(gains) = 1
    vbap_gtableComp = (float*)malloc(vbap_N_gtable * 3 * sizeof(float));
    vbap_gtableIdx = (int*)malloc(vbap_N_gtable * 3 * sizeof(int));
    compressVBAPgainTable3D(vbap_gtable, vbap_N_gtable, vbap_L, vbap_gtableComp, vbap_gtableIdx);
    free(vbap_gtable); // we don't need the full table anymore
}

VbapHelper::~VbapHelper()
{
    free(vbap_gtableComp);
    free(vbap_gtableIdx);
}

std::vector<float>& VbapHelper::calculateWeights(EAR::Polar direction)
{
    int aziIndex = (int)(matlab_fmodf(direction[0] + 180.0f, 360.0f) / vbap_az_res + 0.5f);
    int elevIndex = (int)((direction[1] + 90.0f) / vbap_el_res + 0.5f);
    int idx3d = elevIndex * vbap_N_azi + aziIndex;

    for (int i = 0; i < outputWeights.size(); i++)
        outputWeights[i] = 0.0f;

    for (int i = 0; i < 3; i++) {
        // Omit repeated indices (e.g. when exactly at a node position)
        if (i > 0 && vbap_gtableIdx[idx3d * 3 + i] == vbap_gtableIdx[idx3d * 3 + i - 1])
            continue;
        if (i > 1 && vbap_gtableIdx[idx3d * 3 + i] == vbap_gtableIdx[idx3d * 3 + i - 2])
            continue;
        outputWeights[vbap_gtableIdx[idx3d * 3 + i]] = vbap_gtableComp[idx3d * 3 + i];
    }

    return outputWeights;
}

int VbapHelper::getNumberOfNodes() const
{
    return vbap_L;
}

Vlo::Vlo(EAR::Position _position, int _index, VbapHelper& _vbapHelper, std::size_t _B,
    float* _rampVec, float* _tempVec)
    : position(_position)
    , index(_index)
    , vbapHelper(_vbapHelper)
    , B(_B)
    , rampVec(_rampVec)
    , tempVec(_tempVec)
{
    for (int indexVbap = 0; indexVbap < VbapHelper::MAX_NUM_VBAP_NODES; indexVbap++) {
        targetVbapGains[indexVbap].store(0.0);
        currentVbapGains[indexVbap] = 0.0;
    }
}

Vlo::Vlo(const Vlo& other)
    : Vlo(other.position, other.index, other.vbapHelper, other.B, other.rampVec, other.tempVec)
{}

Vlo::~Vlo() {}

int Vlo::getIndex() const
{
    return index;
}

void Vlo::setAllVbapWeightsToZero()
{
    for (auto& targetVbapGain : targetVbapGains)
        targetVbapGain.store(0.0);
}

void Vlo::updateVbapWeights(float baryGain, float RT[3][3], EAR::Vector3& listenerLocation)
{
    if (baryGain <= 0.0f) {
        setAllVbapWeightsToZero();
    } else {
        EAR::Vector3 listenerToVlo(position.location - listenerLocation);
        listenerToVlo = matrixVectorMultiplication(RT, listenerToVlo); // head rotation
        EAR::Polar vloLocationPolar = EAR::cartesianToPolar(listenerToVlo);
        auto& vbapWeights = vbapHelper.calculateWeights(vloLocationPolar);
        for (int indexVbap = 0; indexVbap < vbapWeights.size(); indexVbap++)
            targetVbapGains[indexVbap].store(vbapWeights[indexVbap] * baryGain);
    }
}

void Vlo::panToVbapNode(
    const EAR::RealBuffer& inputBuffer, int indexVbapNode, EAR::RealBuffer& vbapBuffer)
{
    float curGain = currentVbapGains[indexVbapNode];
    float tarGain = targetVbapGains[indexVbapNode].load();

    if (curGain > 0.0 || tarGain > 0.0) {

#if SAF_USE_INTEL_MKL

        vsLinearFrac(B, rampVec, rampVec, tarGain - curGain, curGain, 0.0f, 1.0f, tempVec);
        vsMul(B, tempVec, inputBuffer.data(), tempVec); // smooth transition
        vsAdd(B, vbapBuffer.data(), tempVec, vbapBuffer.data());

#else // fallback method if MKL not available

        if (curGain != tarGain) {
            float gain_step = (tarGain - curGain) / (float)B;
            for (size_t i = 0; i < B; i++) {
                float gain = curGain + (float)(i)*gain_step;
                vbapBuffer[i] += inputBuffer[i] * gain;
            }
        } else {
            for (size_t i = 0; i < B; i++) {
                vbapBuffer[i] += inputBuffer[i] * curGain;
            }
        }

#endif
    }
    currentVbapGains[indexVbapNode] = tarGain; // update for next iteration
}

EsdSource::EsdSource(EAR::Position _position, float _radius, int _index, int esdOrder,
    VbapHelper& vbapHelper, std::size_t B, float* rampVec, float* tempVec)
    : position(_position)
    , radius(_radius)
    , index(_index)
{
    position.location[1] = 0.0f; // ignore vertical dimension

    std::vector<std::vector<float>> esdDirectionsTable = getEsdSpeakerGrid(esdOrder);

    // Get rotation matrix R (same as in Exterior.cpp)
    float sourceYawDeg = position.orientation[0];
    float sourcePitchDeg = position.orientation[1];
    float sourceRollDeg = position.orientation[2];
    float R[3][3]{};
    get_rotation_matrix_OpenGL(sourceYawDeg, sourcePitchDeg, sourceRollDeg, R);

    // Get location and orientation of each VLO
    for (int indexVlo = 0; indexVlo < esdDirectionsTable.size(); indexVlo++) {

        // Location: according to ESD lookup table
        EAR::Vector3 esdLocation = EAR::polarToCartesian(
            EAR::Polar(esdDirectionsTable[indexVlo][0], esdDirectionsTable[indexVlo][1], radius));
        esdLocation = matrixVectorMultiplication(R, esdLocation); // account for source rotation

        // Orientation: facing the source
        EAR::Polar esdLocationPolar = EAR::cartesianToPolar(-esdLocation);
        float esdYaw = fmod(180.0f + esdLocationPolar[0], 360.0f);
        float esdPitch = -esdLocationPolar[1];
        float esdRoll = 0.0f;
        EAR::Euler esdOrientation(esdYaw, esdPitch, esdRoll);

        EAR::Position vloPosition = EAR::Position(position.location + esdLocation, esdOrientation);

        vloList.emplace_back(Vlo(vloPosition, indexVlo, vbapHelper, B, rampVec, tempVec));
    }
}

EsdSource::~EsdSource() {}

int EsdSource::getIndex() const
{
    return index;
}

void EsdSource::updateVbapWeights(float baryGain, float RT[3][3], EAR::Vector3& listenerLocation)
{
    for (auto& vlo : vloList) {
        vlo.updateVbapWeights(baryGain, RT, listenerLocation);
    }
}

void EsdSource::panToVbapNode(
    const EAR::RealFrame& inputFrame, int indexVbapNode, EAR::RealBuffer& vbapBuffer)
{
    // Loop through all VLOs and update the VBAP buffer on each iteration
    for (auto& vlo : this->vloList) {
        auto& inputBuffer = inputFrame.ch(vlo.getIndex());
        vlo.panToVbapNode(inputBuffer, indexVbapNode, vbapBuffer);
    }
}

} // namespace LowCompMphoa

} // namespace MPEGI
