#pragma once

#include <atomic>

#include "MPEGI/Bitstream/Bitstream.h"
#include "MPEGI/Scene/SceneState.h"
#include <EAR/BaseTypes.h>
#include <EAR/RealBuffer.h>
#include <EAR/RealFrame.h>
#include <EAR/StereoFIRFilter.h>

#include <hoatranslation/MultiPointHoaTranslator.hpp> // hoaTranslatorData
#include <mpegcore/HoaTranslator.hpp> // mpegcore::HoaTranslator

namespace MPEGI {

namespace LowCompMphoa {

/**
* Helper function that returns a [(N+1)^2 , 2] table with azimuth and elevation in degrees
* for an ESD grid of order N.
*/
std::vector<std::vector<float>> getEsdSpeakerGrid(int order);

/**
* Helper function to multiply a 3x3 matrix with a 3x1 vector.
*/
EAR::Vector3 matrixVectorMultiplication(float R[3][3], EAR::Vector3 v);

/**
* Helper function that takes Euler angles and returns a  rotation matrix for pre-multiplication,
* openGL convention, yaw->pitch->roll.
*/
void eulerAnglesToRotationMatrix(float yaw_rad, float pitch_rad, float roll_rad, float R[3][3]);

/**
* Helper function to calculate the transpose of a 3x3 matrix.
*/
void transposeMatrix(float A[3][3], float AT[3][3]);

/**
* Helper function to preprocess an HRTF by selecting a few directions and applying
* frequency-dependent time alignment.
*/
std::vector<std::unique_ptr<EAR::StereoFIRFilter>> preprocessHrtf(
    hoaTranslatorData* h, std::vector<std::vector<float>> directionGrid, float fs, size_t B);

/**
* Helper function to calculate the median distance of an HOA/ESD source to its neighbors
*/
float findMedianDistanceToNeighbors(
    mpegcore::HoaTranslator& hoaTranslator, int indexSource, int numEsdSources);

/**
* Helper function to obtain the directivity factor based on HOA order.
*/
float getDirectivityFactor(int order);

/**
* Helper function to obtain the reverberation factor based on the reverb payload and source positions.
*/
std::vector<float> getReverberationFactors(const Bitstream& bitstream, const SceneState* state,
    const mpegcore::HoaTranslator& hoaTranslator);

class VbapHelper
{
public:
    /**
    * Class constructor.
    */
    VbapHelper(std::vector<std::vector<float>> vbapDirectionsTable);

    /**
    * Class destructor.
    */
    ~VbapHelper();

    /**
    * Calculates panning weights for a given polar direction. The result
    * can be retrieved with getWeights().
    */
    std::vector<float>& calculateWeights(EAR::Polar direction);

    /**
    * Returns the number of nodes.
    */
    int getNumberOfNodes() const;

    const static int MAX_NUM_VBAP_NODES = 25; // support up to 4th order

private:
    int vbap_L;
    const int vbap_az_res = 1;
    const int vbap_el_res = 1;
    const int vbap_N_azi = (int)(360.0f / vbap_az_res + 0.5f) + 1;
    const int vbap_omitLargeTriangles = 0;
    const int vbap_enableDummies = 0;
    const float vbap_spread = 0;
    float* vbap_gtableComp;
    int* vbap_gtableIdx;
    std::vector<float> outputWeights; // output weights are stored here
};

class Vlo
{
public:
    /**
    * Class constructor.
    * 
    * Inputs:
    *     _position: VLO location and orientation
    *     _index: VLO index
    *     _vbapHelper: reference to the VBAP helper object
    *     _B: buffer length in samples
    *     _rampVec: pointer to auxiliary buffer for crossfading calculation
    *     _tempVec: pointer to auxiliary buffer for crossfading calculation
    */
    Vlo(EAR::Position _position, int _index, VbapHelper& _vbapHelper, std::size_t _B,
        float* _rampVec, float* _tempVec);

    /**
    * Copy constructor.
    */
    Vlo(const Vlo& other);

    /**
    * Class destructor.
    */
    ~Vlo();

    /**
    * Returns the index of this VLO
    */
    int getIndex() const;

    /**
    * Sets all VBAP weights of this VLO to zero.
    */
    void setAllVbapWeightsToZero();

    /** 
    * Updates the VBAP weights for this VLO.
    * 
    * Inputs:
    *     baryGain: barycentric interpolation gain for the current ESD source
    *     RT: rotation matrix to account for the listener rotation
    *     listenerLocation: listener location
    */
    void updateVbapWeights(float baryGain, float RT[3][3], EAR::Vector3& listenerLocation);

    /** 
    * Takes the input buffer corresponding to this VLO, calculates the signal panned to
    * a VBAP node according to the VLO's panning weights (e.g. if panning weights are all zeroes, the panned signal
    * is all zeroes), and adds the panned signal to the VBAP node's current audio buffer.
    * 
    * Inputs:
    *     inputBuffer: input audio buffer corresponding to the VLO
    *     indexVbap: index of the VBAP node for which we want to obtained the panned signal
    *     vbapBuffer: audio buffer corresponding to the VBAP node (panned signal will be added to current contents)
    */
    void panToVbapNode(
        const EAR::RealBuffer& inputBuffer, int indexVbap, EAR::RealBuffer& vbapBuffer);

private:
    EAR::Position position;
    std::array<std::atomic<float>, VbapHelper::MAX_NUM_VBAP_NODES> targetVbapGains;
    std::array<float, VbapHelper::MAX_NUM_VBAP_NODES> currentVbapGains;
    int index;
    VbapHelper& vbapHelper;
    std::size_t B; // buffer size
    float* rampVec; // temp buffer for crossfading
    float* tempVec; // temp buffer for crossfading
};

class EsdSource
{
public:
    /**
    * Class constructor.
    * 
    * Inputs:
    *     _position: ESD source location and orientation
    *     _radius: radius used for the VLO grid
    *     _index: ESD source index
    *     esdOrder: order used for the VLO grid
    *     _vbapHelper: reference to the VBAP helper object
    *     _B: buffer length in samples
    *     _rampVec: pointer to auxiliary buffer for crossfading calculation
    *     _tempVec: pointer to auxiliary buffer for crossfading calculation
    */
    EsdSource(EAR::Position _position, float _radius, int _index, int esdOrder,
        VbapHelper& vbapHelper, std::size_t B, float* rampVec, float* tempVec);

    /**
    * Class destructor.
    */
    ~EsdSource();

    /**
    * Returns the index of this ESD source.
    */
    int getIndex() const;

    /** 
    * Updates the VBAP weights for this ESD source.
    * 
    * Inputs:
    *     baryGain: barycentric interpolation gain for the current ESD source
    *     RT: rotation matrix to account for the listener rotation
    *     listenerLocation: listener location
    */
    void updateVbapWeights(float baryGain, float RT[3][3], EAR::Vector3& listenerLocation);

    /** 
    * Takes the input audio frame corresponding to this ESD source, calculates the signal panned to
    * a VBAP node according to the ESD source's panning weights (e.g. if panning weights are all zeroes, the panned signal
    * is all zeroes), and adds the panned signal to the VBAP node's current audio buffer.
    * 
    * Inputs:
    *     inputFrame: input audio frame corresponding to the ESD source
    *     indexVbap: index of the VBAP node for which we want to obtained the panned signal
    *     vbapBuffer: audio buffer corresponding to the VBAP node (panned signal will be added to current contents)
    */
    void panToVbapNode(
        const EAR::RealFrame& inputFrame, int indexVbap, EAR::RealBuffer& vbapBuffer);

private:
    std::vector<Vlo> vloList;
    float radius;
    EAR::Position position;
    int index;
};

} // namespace LowCompMphoa

} // namespace MPEGI