/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "MphoaStage.hpp"

#include <algorithm>
#include <utility>

#include <EAR/AnchorObject.h>
#include <EAR/Exception.h>
#include <EAR/HOASource.h>
#include <EAR/Listener.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/MphoaPayload.hpp>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Stream/StreamBuffer.h>

#include <MPEGI/Helpers/AudioUtilization.h>
#include <MPEGI/Renderer/RenderStageObserver.h>
#include <hoatranslation/MultiPointHoaTranslator.hpp>
#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/HoaTranslator.hpp>
#include <mpegcore/HoaTranslatorConfig.hpp>
#include <mpegcore/algorithm.hpp>
#include <mpegcore/spdlog.hpp>

using namespace mpegcore;

namespace MPEGI {

std::shared_ptr<RenderStage> makeMphoaStage()
{
    return std::make_shared<MphoaStage>();
}

MphoaStage::MphoaStage()
    : MphoaBaseStage()
{}

MphoaStage::~MphoaStage() {}

std::unique_ptr<RenderStage::DspData> MphoaStage::update(RenderList& renderList)
{
    return MphoaBaseStage::update(renderList);
}

void MphoaStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<MphoaStage::DspData>();

    if (!isInitialized || currentDSP.dspItems.empty()) {
        return;
    }

    // Clear stage output
    outputRenderItems[0]->signal->getFrame().zero();

    std::vector<MPHOAAudioUtilization> audioUtilization;
    // Render translators
    for (auto const translatorIndex : currentDSP.activeTranslatorIndices) {
        auto& hoaTranslator = translatorData[translatorIndex].translator;

        // In principle we might run `preProcess` already in `MphoaStage::update` but that would
        // require taking care of possible data races for being able to update the translator's data
        // in one thread while accessing it in another. As the processing time of `preProcess` seems
        // to be quite miniscule compared to `HoaTranslator::render` this doesn't seem like a good
        // tradeoff. Might also add some bookkeeping so that `preProcess` would be run only when
        // `update` has been run but even that's not considered worthwhile.
        hoaTranslator.preProcess(listenerPosition);
        int fullHoaOrder = hoaTranslator.getHOAOrder();
        std::vector<std::string> fullOrderIdList;

        for (size_t fullOrderHOASource : hoaTranslator.getFullOrderHOASourceIndices()) {
            std::string fullOrderSourceID =
                translatorData[translatorIndex].hoaSourceIds[fullOrderHOASource];
            audioUtilization.push_back(MPHOAAudioUtilization(fullOrderSourceID, fullHoaOrder));
            fullOrderIdList.push_back(fullOrderSourceID);
        }
        for (size_t firstOrderHOASource : hoaTranslator.getFirstOrderHOASourceIndices()) {
            // check if first order source is already in full order list
            std::string firstOrderSourceID =
                translatorData[translatorIndex].hoaSourceIds[firstOrderHOASource];

            if (std::find(fullOrderIdList.begin(), fullOrderIdList.end(), firstOrderSourceID)
                == fullOrderIdList.end()) {
                audioUtilization.push_back(MPHOAAudioUtilization(firstOrderSourceID, 1));
            }
        }
        feedInputToHoaTranslator(hoaTranslator, currentDSP.dspItems[translatorIndex].inputBuffers);
        hoaTranslator.render();
    }
    renderingStageObserver->setUtilization(audioUtilization);

    // Copy translator outputs to stage output
    if (auto translatorIndexIt = currentDSP.activeTranslatorIndices.cbegin();
        translatorIndexIt != currentDSP.activeTranslatorIndices.cend()) {

        // Copy the first active translator's output
        auto& outputFrame = outputRenderItems[0]->signal->getFrame();
        auto const& translatorOutput = translatorData[*translatorIndexIt].translator.outputBuffer();
        for (auto channelIndex = outputFrame.channels(); channelIndex-- > 0;) {
            std::copy_n(translatorOutput[channelIndex].data(), B, outputFrame[channelIndex].data());
        }

        // Mix in other active translators' outputs
        for (auto end = currentDSP.activeTranslatorIndices.cend(); ++translatorIndexIt != end;) {
            auto const& translatorOutput =
                translatorData[*translatorIndexIt].translator.outputBuffer();

            for (auto channelIndex = outputFrame.channels(); channelIndex-- > 0;) {
                auto* src = translatorOutput[channelIndex].data();
                auto* dst = outputFrame[channelIndex].data();
                std::transform(dst, dst + B, src, dst, std::plus<float>());
            }
        }
    }

    if (mphoaGain != 1.F) {
        auto& outputFrame = outputRenderItems[0]->signal->getFrame();
        outputFrame.mul(mphoaGain);
    }

    // Increase frame counter for each translator (active and inactive)
    for (auto const& data : translatorData) {
        data.translator.increaseFrameCounter();
    }
}

void MphoaStage::feedInputToHoaTranslator(HoaTranslator& hoaTranslator,
    std::vector<std::shared_ptr<MPEGI::StreamBuffer>> const& inputBuffers)
{
    // Convert all processed inputs from ESD and feed the results to the translator
    for (auto const* hoaInput : hoaTranslator.getProcessedHoaInputs()) {
        auto const sourceIndex = hoaInput->sourceIndex;
        auto const channelCount = hoaInput->channelGains.size();

        for (std::size_t channelIndex{ 0 }; channelIndex < channelCount; ++channelIndex) {
            std::copy_n(inputBuffers[sourceIndex]->getFrame().ch(channelIndex).data(), B,
                esdConverter.inputAudioData(channelIndex));
        }
        auto const order = static_cast<std::size_t>(std::sqrt(channelCount)) - 1;
        esdConverter.convertEsdToHoa(order);

        hoaTranslator.readToInput(FLATTEN2D(esdConverter.getOutputPointerArray()),
            ORDER2NSH(order) * esdConverter.samplesPerConversion(), sourceIndex);
    }
}

void MphoaStage::initializeStageAdditional(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state, size_t translatorCount,
    const std::vector<size_t>& groupSourceCount)
{
    initializeEsdConversion(state->getHOASources(), *streamManager, *state);
}

void MphoaStage::initializeEsdConversion(
    std::vector<std::unique_ptr<EAR::HOASource>> const& hoaSources,
    StreamManager const& streamManager, SceneState const& state)
{
    std::size_t maxSourceChannelCount{ 0 };
    std::size_t minSourceChannelCount{ std::numeric_limits<std::size_t>::max() };

    // Determine the minimum and maximum channel counts of all mphoa sources.
    for (auto const& source : hoaSources) {
        if (!isMphoaSource(*source)) {
            continue;
        }
        auto const& streamId = state.getAudioStream(source.get()).id;
        auto const channelCount = streamManager.getNbrOfChannels(streamId);
        EAR_CHECK_REL(
            channelCount > 0, "MphoaStage: HOASources without channels are not supported");
        maxSourceChannelCount = std::max(maxSourceChannelCount, channelCount);
        minSourceChannelCount = std::min(minSourceChannelCount, channelCount);
    }

    esdConverter.initialize(B);
    EAR_CHECK_REL(esdConverter.isInitialized(), "MphoaStage: failed to initialize ESD conversion");

    // Verify that EsdConverter can handle all the sources (i.e. that the orders are acceptable).
    std::size_t maxOrder = (std::size_t)std::sqrt(maxSourceChannelCount) - 1;
    std::size_t minOrder = (std::size_t)std::sqrt(minSourceChannelCount) - 1;
    auto const maxEsdConversionOrder = esdConverter.maxConversionOrder();
    // We need minimum order of 1 as EsdConverter doesn't support 0th order yet.
    constexpr std::size_t MINIMUM_HOA_ORDER{ 1 };
    EAR_CHECK_REL(minOrder >= MINIMUM_HOA_ORDER,
        fmt::format("MphoaStage: minimum channel order ({}) is not acceptable (need {}-{})",
            minOrder, MINIMUM_HOA_ORDER, maxEsdConversionOrder));
    EAR_CHECK_REL(maxOrder <= maxEsdConversionOrder,
        fmt::format("MphoaStage: maximum channel order ({}) is not acceptable (need {}-{})",
            maxOrder, MINIMUM_HOA_ORDER, maxEsdConversionOrder));
}

} // namespace MPEGI
