/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Occlusion.h"

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

#include <EAR/BaseTypes.h>
#include <EAR/EmbreeTracer.h>
#include <EAR/Exception.h>
#include <EAR/Geometry.h>
#include <EAR/Listener.h>

#include <MPEGI/Helpers/RayPlaneIntersection.h>
#include <MPEGI/Renderer/PerceptualPointDistribution.h>

#include <algorithm>
#include <cassert>
#include <numeric>
#include <string>
#include <unordered_map>

namespace MPEGI {

std::shared_ptr<RenderStage> makeOcclusionStage()
{
    return std::make_shared<OcclusionStage>();
}

namespace {

inline bool vectorsEqual(const std::vector<float>& a, const std::vector<float>& b)
{
    return std::equal(a.begin(), a.end(), b.begin(), [](float v1, float v2) {
        constexpr float epsilon = 0.001f;
        return std::fabs(v1 - v2) < epsilon;
    });
}

// reset EQ buffer to multiply values for each material in the way
void setEq(const std::optional<OcclusionInfo>& occlusionInfo, std::vector<float>& eq)
{
    std::fill(eq.begin(), eq.end(), 1.f);
    if (occlusionInfo.has_value()) {
        for (auto mat : occlusionInfo->materials) {
            auto centerFreqs = Config::getFloatPtr("EQCenterFreqs");
            for (unsigned int i = 0; i < Config::getInt("EQNBands"); i++) {
                eq[i] *= mat->getTransmissionCoeffForFreq(centerFreqs[i]);
            }
        }
    }
}

} // anonymous namespace

OcclusionStage::OcclusionStage()
    : binaryMode(MPEGI::Config::getBoolDef("Occlusion_BinaryMode", false))
{}

OcclusionStage::~OcclusionStage()
{
    for (auto& ptr : attachedGeometry) {
        const_cast<EAR::Geometry*>(ptr)->detach(this);
    }
}

std::optional<OcclusionInfo> OcclusionStage::getOccludingMaterials(
    const EAR::Vector3& listenerLocation, const EAR::Vector3& sourceLocation)
{
    EAR::Polar direction = EAR::cartesianToPolar(listenerLocation - sourceLocation);
    float targetDist = direction[2];

    EAR::Vector3 nextStartPoint = sourceLocation;
    const EAR::Material* prevMaterial = nullptr;
    int prevFace = -1;
    const EAR::Mesh* hitMesh = tracer->firstHit(nextStartPoint, direction, true);

    std::optional<OcclusionInfo> result;

    if (hitMesh) { // is occluded?

        if (binaryMode) {
            result.emplace();
            result->materials.push_back(this->nonTransmittingMaterial);
        } else { // normal mode

            // process every distinct geometry that we hit (not just the first)
            while (hitMesh) {

                float d = tracer->getDistanceToLastHit();
                int fid = tracer->getPrimitiveOfLastHit();

                if (fid >= 0) {
                    auto mat = hitMesh->getTriangle(fid).material;

                    // only consider this hit when it has a material
                    // AND (is either a different material than before
                    //      OR more than a threshold distance away).
                    // This is done to avoid counting some surface multiple times if they are
                    // curved and hit from an acute angle

                    // TODO: tune the RepeatDist
                    if ((mat) && (mat.get() != prevMaterial)
                        || (d > Config::getFloatDef("Occlusion_RepeatDist", 1.f))) {

                        if (result == std::nullopt) {
                            result.emplace();
                        }
                        result->materials.push_back(mat);
                        prevMaterial = mat.get();
                    }
                }

                // adding 10 cm to the distance, making sure that we don't hit exactly the same face again
                d += 0.1;
                nextStartPoint = nextStartPoint
                    + EAR::polarToCartesian(EAR::Polar(direction[0], direction[1], d));
                targetDist -= d;
                direction[2] = targetDist;

                // repeat process to find the next potential occluder
                hitMesh = tracer->firstHit(nextStartPoint, direction, true);
            }
        }
    }
    return result;
}

std::unique_ptr<RenderStage::DspData> OcclusionStage::update(RenderList& renderList)
{

    if (!tracer) {
        return {};
    }

    const float sessDistThres = Config::getFloat("DiffractionSourceExtentDistanceThreshold");

    const std::vector<std::unique_ptr<MPEGI::Portal>>& portals =
        bitstreamPortalPayload->getPortals();

    for (auto& item : renderList.items) {

        if (item->type != ItemType::Primary && item->type != ItemType::Representative
            && item->type != ItemType::DirectionalReverb) {
            // only primary and directly derived render items and portals processed here
            continue;
        } else if (item->extent && item->signal->getFrame().channels() > 1
            && item->type != ItemType::DirectionalReverb) {
            // Skip heterogeneous items, except portals (which have 2 channels)
            continue;
        }

        if (item->position.cSpace == EAR::CSpace::LCS) {
            // Occlusion is ignored for listener-relative objects
            continue;
        }

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        //! To handle fast moving sources separately
        if (speedmap.find(item->objectName) == speedmap.end()) {
            speedmap.insert({ item->objectName, std::make_pair(item->position.location, 0.f) });
        } else {
            if (frameCounter % 5 == 0) {
                float distance =
                    (item->position.location - speedmap.at(item->objectName).first).abs();
                float speed = distance * 48000 / (256 * 5);
                speedmap.at(item->objectName) =
                    std::pair<EAR::Vector3, float>(item->position.location, speed);
            }
        }

        if (speedmap.at(item->objectName).second > Config::getFloat("OcclusionFastSpeed")) {
            //! Use Ray bundles
            EAR::Vector3 target = item->position.location;
            EAR::Vector3 origin = renderList.listener.position.location;
            EAR::Vector3 r = target - origin;
            EAR::Vector3 r_orth1 = r;
            EAR::Vector3 r_orth2 = r;

            float radius = speedmap.at(item->objectName).second / 50; // 100m/s -> 2m radius
            if (r[2] != 0) {
                r_orth1[2] = -1 * (r[0] + r[1]) / r[2];
            } else if (r[1] != 0) {
                r_orth1[1] = -1 * (r[0] + r[2]) / r[1];
            } else {
                r_orth1[0] = -1 * (r[1] + r[2]) / r[0];
            }
            r_orth1.normalize();
            r_orth2 = r.cross(r_orth1).normalize();

            unsigned int nofHits = 0;
            for (float angle = 0.f; angle < 360.f;) {
                EAR::Vector3 sample = { r_orth1[0] * (float)std::cos(angle * M_PI / 180.f)
                        + r_orth2[0] * (float)std::sin(angle * M_PI / 180.f),
                    r_orth1[0] * (float)std::cos(angle * M_PI / 180.f)
                        + r_orth2[0] * (float)std::sin(angle * M_PI / 180.f),
                    r_orth1[0] * (float)std::cos(angle * M_PI / 180.f)
                        + r_orth2[0] * (float)std::sin(angle * M_PI / 180.f) };

                //! Set speed-dependent bundle radius
                sample[0] *= radius;
                sample[1] *= radius;
                sample[2] *= radius;
                sample += target;

                //! Count hit-rays
                if (!tracer->isOccluded(
                        origin, sample - origin, Config::getFloat("GeometricErrorThreshold"))) {
                    nofHits++;
                }

                angle += 18.f;
            }

            float hitRatio = std::max((float)(nofHits / 20.f), 0.25f);

            // update item's eq with hitRatio
            std::vector<float> eq(Config::getInt("EQNBands"), hitRatio);
            item->addEq(&eq.front());
            continue;
        }

        auto& itemState = item->getStateOrCreate<ItemStatePoint>();

        itemState.eqChanged = false;

        if (item->extent && item->status == ItemStatus::Inactive) {
            this->removeOccludedItems(*item);
        } else if (item->extent
            && (renderList.listener.position.location - item->position.location).abs()
                > Config::getFloat("DiffractionSourceExtentDistanceThreshold")
            && item->type != ItemType::DirectionalReverb) {

            float distance =
                (renderList.listener.position.location - item->position.location).abs();
            float weight = std::fmax(1.f
                    - (distance - Config::getFloat("DiffractionSourceExtentDistanceThreshold"))
                        / 200.f,
                0.1f);
            std::vector<float> attenuation(itemState.magnitudes.size(), weight);
            // TODO: Not clear why EQ is reset. It should not be necessary. It interferes with
            //          coupled sources (added 2024-Q1), by resetting the EQ that was set in the portal
            //          stage. In order to prevent any issues with legacy operations, while fixing the
            //          issue for coupled sources, we will only skip this code for coupled sources. But
            //          it may not be necessary for any source....
            if (item->subtype == ItemSubtype::CoupledSrc) {
                item->addEq(attenuation.data(), true);
            } else {
                item->setEq(attenuation, true);
            }

        } else if (item->extent && !item->rayHits.empty()
            && (item->type == ItemType::DirectionalReverb
                || item->signal->getFrame().channels() == 1)) {
            // Items with homogeneous extent (including portals) with at least 1 ray hit

            // For explicit portals, find the plane that defines the portal opening.
            MPEGI::IntersectPlane portalPlane;
            if (item->type == ItemType::DirectionalReverb
                && bitstreamPortalPayload->getExplicitPortalMode()) {
                int portalId = item->parentPortalID;
                auto findLambda = [&portalId](const std::unique_ptr<MPEGI::Portal>& p) {
                    return p->portalId == portalId;
                };
                const auto portal = std::find_if(portals.begin(), portals.end(), findLambda);
                portalPlane = (*portal)->portalPlanes.at(item->acousticEnvironmentId);
            }

            // NB: Heterogeneous extended sources and homogeneous extended sources without ray hits
            //     are handled like point sources (in the "else" clause)

            // New entries are automatically created if needed:
            auto& currentOccludedItems = item->getStateOrCreate<ItemStateExtended>().secondaryItems;

            // TODO: check if extent geometry has moved
            // NB: change in item position doesn't matter if there is no change in ray hits
            if (geometryDirty.load()
                || item->changed & (ItemProperty::RayHits | ItemProperty::Status)
                || renderList.listener.hasMoved) {

                // TODO: option to select the ray distribution (used in multiple stages)
                const Vector3* rays = SESS_RAY_DISTRIBUTION;
                const auto rays_size = sizeof(SESS_RAY_DISTRIBUTION) / sizeof(Vector3);

                const int extraCounts[] = { 1, 6, 12 };
                const int extraRaysCount = extraCounts[0] + extraCounts[1] + extraCounts[2];
                EAR::Quaternion listenerRotation{ renderList.listener.position.orientation };

                SecondaryItems previousOccludedItems;
                currentOccludedItems.swap(previousOccludedItems);
                assert(currentOccludedItems.empty());

                for (const RayHit& rh : item->rayHits) {
                    EAR::Vector3 axis0;
                    if (rh.rayID == 0) {
                        axis0 = rays[rh.rayID] - rays[rh.rayID + 1];

                    } else {
                        axis0 = rays[rh.rayID - 1] - rays[rh.rayID];
                    }
                    const float r0 = 2.f * std::tan(0.0534) / 5.f;
                    axis0 = axis0.normalize();
                    const EAR::Vector3 axis1 = rays[rh.rayID].cross(axis0).normalize();
                    const EAR::Vector3 dir0 = rays[rh.rayID];
                    int hitCount = 0;
                    unsigned int subrayMask = 1;
                    const bool noRefinement = (rh.refinementMask == -1);
                    const int jmax = noRefinement ? 1 : 3;
                    const float wt = noRefinement ? 1.f : 1.f / extraRaysCount;
                    for (int j = 0; j < jmax; j++)
                        for (int i = 0; i < extraCounts[j]; i++) {
                            if (!(rh.refinementMask & subrayMask)) {
                                subrayMask <<= 1;
                                continue;
                            }
                            const float alpha = 2 * M_PI * i / extraCounts[j];
                            const EAR::Vector3 dir = dir0 + axis0 * r0 * j * std::sin(alpha)
                                + axis1 * r0 * j * std::cos(alpha);

                            std::optional<OcclusionInfo> occlusionInfo = std::nullopt;
                            if (rh.distance < 0.0) {
                                // We are inside the extended source, any occlusion is ignored
                                // and we use an empty list of materials.
                            } else {
                                const auto hitLocation = renderList.listener.position.location
                                    + listenerRotation.rotate(dir) * rh.distance;
                                occlusionInfo = getOccludingMaterials(
                                    renderList.listener.position.location, hitLocation);
                            }
                            SecondaryItems::iterator target;

                            if (item->type == ItemType::DirectionalReverb
                                && !occlusionInfo.has_value()
                                && item->acousticEnvironmentId
                                    != renderList.listener.acousticEnvironmentId
                                && bitstreamPortalPayload->getExplicitPortalMode()) {
                                // Ray is not occluded, the listener is not in the same romm as the extent mesh,
                                // so the ray should go through the portal plane.
                                // Only checked for explicit portal mode, not for fallback (shrunken mesh) mode.
                                std::pair<bool, EAR::Vector3> result =
                                    MPEGI::checkRayPlaneIntersection(
                                        renderList.listener.position.location,
                                        listenerRotation.rotate(dir), portalPlane);
                                if (!result.first) {
                                    // Ray didn't go through the portal plane, so ignore it.
                                    continue;
                                }
                            }

                            if (auto handle = previousOccludedItems.extract(occlusionInfo);
                                handle.empty()) {
                                auto [it, inserted] = currentOccludedItems.try_emplace(
                                    occlusionInfo, nullptr, RayHits{});
                                if (inserted) {
                                    std::string suffix = occlusionInfo.has_value()
                                        ? " [extended&occluded]"
                                        : " [extended]";
                                    auto* newItem =
                                        this->itemStore->createItem(item->objectName + suffix);

                                    // position should be ignored in favor of extent->position, but we set it anyway
                                    bool exceptStatus = true;
                                    newItem->syncAll(*item, exceptStatus);
                                    newItem->type = (item->type == ItemType::DirectionalReverb)
                                        ? ItemType::DirectionalReverb
                                        : ItemType::Representative;
                                    newItem->parentRI = item->idx;
                                    newItem->changed |= ItemProperty::ParentRI;

                                    // The new ray hits and "changed" information will be added afterwards
                                    newItem->rayHits.clear();
                                    newItem->changed &= ~ItemProperty::RayHits;

                                    std::vector<float> eq;
                                    eq.resize(Config::getInt("EQNBands"));
                                    setEq(occlusionInfo, eq);
                                    newItem->addEq(eq.data());
                                    it->second.first = newItem;
                                } else {
                                    // Nothing to do here, an item for this material sequence already exists
                                }
                                target = it;
                            } else {
                                // Child item for given material sequence already exists
                                auto result = currentOccludedItems.insert(std::move(handle));
                                assert(result.inserted);
                                target = result.position;
                            }

                            assert(target->second.first != nullptr);
                            target->second.first->occlusionInfo = occlusionInfo;
                            auto& newRayHits = target->second.second;
                            newRayHits.emplace_back(rh.rayID, rh.distance, wt);
                            subrayMask <<= 1;
                        }
                }

                for (auto& entry : currentOccludedItems) {
                    auto& [childItem, newRayHits] = entry.second;

                    // TODO: what if Eq of primary source changes?

                    //aggregate the rayhits
                    std::unordered_map<unsigned int, RayHit> rayWeights;
                    for (const auto& rh : newRayHits) {

                        auto it = rayWeights.find(rh.rayID);
                        if (it == rayWeights.end()) {
                            rayWeights.insert(
                                std::make_pair(rh.rayID, RayHit(rh.rayID, rh.distance, rh.weight)));
                        } else {
                            it->second.weight += rh.weight;
                        }
                    }
                    newRayHits.clear();
                    for (const auto& [id, rh] : rayWeights)
                        newRayHits.push_back(rh);

                    childItem->sync(*item,
                        ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Loudness
                            | ItemProperty::Aparams | ItemProperty::Position | ItemProperty::Extent
                            | ItemProperty::Signal | ItemProperty::RefDist | ItemProperty::Reverb);

                    // Ray hits will always be ordered by ray ID
                    if (childItem->rayHits == newRayHits) {
                        // Nothing to do, no change in ray hits
                        childItem->changed &= ~ItemProperty::RayHits;
                    } else {
                        const float totCount = (float)item->rayHits.size();
                        const float newCount = (float)newRayHits.size();
                        childItem->addGain(std::sqrt(newCount / totCount));
                        childItem->rayHits.swap(newRayHits);
                        childItem->changed |= ItemProperty::RayHits;
                    }
                    newRayHits.clear(); // Will be re-used in next update
                    assert(!childItem->rayHits.empty());
                }
                for (auto child : previousOccludedItems) {
                    this->itemsToRemove.push_back(child.second.first);
                }
            } else {
                // no changes, child items from previous run are used
            }

            // All ray hits have been moved to secondary render items
            item->rayHits.clear();
            // the extended source logic is taking care of occlusion
            itemState.isOccluded = false;
            // the primary item, however, is not deacivated, and its signal
            // feeds the filters generated based on the properties of the representative items
        } else { // no (homogeneous) extent or no ray hits

            // If an extended source has no ray hits, any previous child items have to be removed:
            this->removeOccludedItems(*item);

            if (geometryDirty.load() || item->changed & ItemProperty::Position
                || item->changed & ItemProperty::Status || renderList.listener.hasMoved) {
                // This would not be necessary for LCS items if the listener has moved,
                // but LCS items are completely ignored (see above).

                item->occlusionInfo = getOccludingMaterials(
                    renderList.listener.position.location, item->position.location);

                if (item->occlusionInfo.has_value()) {
                    setEq(item->occlusionInfo, newMagnitudes);

                    // only trigger an EQ filter update if the magnitudes actually changed
                    if (!vectorsEqual(newMagnitudes, itemState.magnitudes)) {
                        // TODO: more elegant way to find out if EQ changed?
                        std::copy(newMagnitudes.begin(), newMagnitudes.end(),
                            itemState.magnitudes.begin());
                        item->changed |= ItemProperty::Eq;
                        itemState.eqChanged = true;
                    }
                    itemState.isOccluded = true;
                } else {
                    // item moved and is not occluded
                    if (itemState.isOccluded) { // was occluded before, so update EQ filter
                        item->changed |= ItemProperty::Eq;
                        itemState.eqChanged = true;
                        // resetting magnitudes for a valid comparison when the next occlusion happens
                        std::fill(itemState.magnitudes.begin(), itemState.magnitudes.end(), 1.f);
                    }
                    itemState.isOccluded = false;
                }
            }
        } // if extent or not

        itemState.mainItemStatus = item->status;

        // Ray hits are always removed from the primary item, so they are never "changed"
        item->changed &= ~ItemProperty::RayHits;

        // We set the EQ of the main RenderItem even if it is deactivated
        // (to avoid artifacts when it is re-activated while still occluded)
        if (itemState.isOccluded) {
            item->addEq(itemState.magnitudes.data(), false);
        } else {
            // TODO: Don't we need this for non-extended items, too!?
            // TODO: Not clear why EQ is reset to ones. It should not be necessary. It interferes with
            //          coupled sources (added 2024-Q1), by resetting the EQ that was set in the portal
            //          stage. In order to prevent any issues with legacy operations, while fixing the
            //          issue for coupled sources, we will only skip this code for coupled sources. But
            //          it may not be necessary for any source....
            if (item->extent && item->subtype != ItemSubtype::CoupledSrc) {
                std::vector<float> ones(itemState.magnitudes.size(), 1.f);
                item->setEq(ones, true);
            }
        }
    } // for all render items

    // NB: we do this even if there have been no changes:
    for (auto& state : itemStore->getItemStates<ItemStateExtended>()) {
        auto& childItems = state.secondaryItems;
        for (auto& [occlusionInfo, data] : childItems) {
            auto& [childItem, rayHits] = data;
            assert(rayHits.empty());
            renderList.addItemCopy(childItem);
            childItem->resetChanged();
        }
    }

    std::vector<RenderItem*> temp;
    temp.swap(this->itemsToRemove);
    for (auto* item : temp) {
        assert(item != nullptr);
        if (item->status != ItemStatus::Orphaned) {
            this->itemStore->remove(item);
            this->itemsToRemove.push_back(item);
            renderList.addItemCopy(item);
        }
    }

    geometryDirty.store(false);
    frameCounter++;
    return {};
}

void OcclusionStage::removeOccludedItems(RenderItem& item)
{
    auto* state = item.getState<ItemStateExtended>();
    if (state) {
        for (auto& entry : state->secondaryItems) {
            this->itemsToRemove.push_back(entry.second.first);
        }
        state->secondaryItems.clear();
    }
}

void OcclusionStage::objectChanged(
    EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification)
{
    geometryDirty.store(true);
}

void OcclusionStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }

    Config::setDef("OcclusionFastSpeed", EAR::Variant(5.f));
    Config::setDef("DiffractionSourceExtentDistanceThreshold", EAR::Variant(32.f));
    Config::setDef("GeometricErrorThreshold", EAR::Variant(0.001f));

    this->itemStore = itemStore;

    auto eifPayload = std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE));

    auto portalPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_PORTAL);
    EAR_CHECK_REL(portalPayloads.size() <= 1, "Too many portal payloads in bitstream");
    if (portalPayloads.size() == 1) {
        bitstreamPortalPayload =
            std::dynamic_pointer_cast<const MPEGI::PortalPayload>(portalPayloads[0]);
    }

    B = streamManager->getB();
    Fs = streamManager->getFs();

    auto isTransformUpdated = [&](const std::string& tformId) {
        for (auto& upd : bitstream.getUpdates()) {
            for (auto& mod : upd.modifications) {
                if (mod.id == tformId) {
                    return true;
                }
            }
        }
        return false;
    };

    // TODO: add spheres and cylinders
    std::vector<const EAR::Mesh*> meshes;
    for (const auto geo : state->getAllVisibleGeometry()) {
        if (geo->getMesh()) {
            if (geo->getMesh()->getParent()) {
                if (isTransformUpdated(geo->getMesh()->getParent()->getName())) {
                    continue;
                };
            }

            if (!geo->isStatic()) {
                const_cast<EAR::Geometry*>(geo)->attach(this);
                attachedGeometry.push_back(geo);
            }
            meshes.push_back(geo->getMesh());
        }
    }

    if (meshes.size() > 0) { // only run this stage when geometry is in the scene
        tracer = std::make_unique<EAR::EmbreeTracer>(meshes, true, true);
        geometryDirty.store(true);
        newMagnitudes.resize(Config::getInt("EQNBands"), 1.f);
    }
}

} // namespace MPEGI
