/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_OCCLUSION_STAGE
#define MPEGI_OCCLUSION_STAGE

#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Renderer/RenderList.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/PortalPayload.h>

#include <atomic>
#include <cstddef>
#include <map>
#include <vector>

// fwd declaration
namespace EAR {
class EmbreeTracer;
class Geometry;
} // namespace EAR

namespace MPEGI {

//! Render (partial or total) occlusion of Render Items w.r.t. the listener position
/*! 
 * Inputs: 
 * - Only items with ItemType::Primary or ItemType::Representative
 * - The scene geometry (static and dynamic)
 *  
 * Outputs: 
 * - The considered RIs get a new EQ based on the "transmission" properties of the AcousticMaterials
 *   of the occluding objects
 *  
 * How it works: 
 * - A ray is traced between the source and listener position
 * - Any geometry surface that is hit from the normal direction is considered for occlusion
 * - If the hit surface has an AcousticMaterial, its "transmission" coefficient is applied as an EQ
 *   to the respective render item
 * 
 * Stages order:
 * - Required before: none
 * - Required after: EqStage, FadeStage
 * - Useful after: SimpleCullingStage
 * - Must be before if used: DiffractionStage, EarlyReflectionStage
 *  
 * Config options: 
 * - (not real-time adjustable) "Occlusion_BinaryMode": whether or not any occlusion immediately results in deactivation
 *   of the render item (default: false)
 * - (not real-time adjustable) "Occlusion_RepeatDist": distance after which a hit to the same geometry can be counted again
 *   (default: 1m). This can only happen with non-convex meshes.
 * 
 */
class OcclusionStage : public RenderStage, public SceneObjectObserver
{
public:
    OcclusionStage();

    //! Destructor
    ~OcclusionStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from SceneObjectObserver
    void objectChanged(
        EAR::SceneObject* obj, EAR::SceneObjectObserver::Property modification) override;

private:
    friend class Renderer;

    RenderItemStore* itemStore;

    const bool binaryMode; //!< always disable sources when direct path is obstructed

    // Block size and sampling frequency
    float B;
    float Fs;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    struct ItemStatePoint : RenderItem::ExtraState
    {
        explicit ItemStatePoint()
            : magnitudes(Config::getInt("EQNBands"), 1.f)
        {}

        std::vector<float> magnitudes;
        bool isOccluded = false;
        bool eqChanged;
        ItemStatus mainItemStatus = ItemStatus::Orphaned;
    };

    using RayHits = std::vector<RayHit>;
    /// Mapping from material list -> secondary render item
    using SecondaryItems = std::map<std::optional<OcclusionInfo>, std::pair<RenderItem*, RayHits>>;
    struct ItemStateExtended : RenderItem::ExtraState
    {
        SecondaryItems secondaryItems;
    };

    std::optional<OcclusionInfo> getOccludingMaterials(
        const EAR::Vector3& listenerLocation, const EAR::Vector3& sourceLocation);

    void removeOccludedItems(RenderItem& item);

    std::unique_ptr<EAR::EmbreeTracer> tracer;

    std::atomic<bool> geometryDirty; // set to true when a geometry moved
    std::vector<const EAR::Geometry*> attachedGeometry; // used to detach in destructor
    std::vector<float> newMagnitudes; // used for comparing old and new EQ

    std::vector<RenderItem*> itemsToRemove;

    /// Dummy material for binaryMode (default material has no transmission)
    std::shared_ptr<Material> nonTransmittingMaterial = std::make_shared<Material>();

    // Speed map for primary sources
    std::map<std::string, std::pair<EAR::Vector3, float>> speedmap;
    unsigned int frameCounter{ 0 };

    std::shared_ptr<const MPEGI::PortalPayload> bitstreamPortalPayload;
};

} // namespace MPEGI

#endif /* MPEGI_OCCLUSION_STAGE */
