/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "PannerStage.hpp"

#include <string>

#include <EAR/Exception.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Scene/SceneState.h>
#include <mpegcore/Panner.hpp>

namespace MPEGI {

std::shared_ptr<RenderStage> makePannerStage()
{
    return std::make_shared<PannerStage>();
}

namespace {

bool isPannableInputRenderItem(RenderItem const& item)
{
    return (item.type == ItemType::Pannable)
        && !(item.status == ItemStatus::Inactive || item.status == ItemStatus::Orphaned);
}

mpegcore::Panner::Config getPannerConfig()
{
    mpegcore::Panner::Config config;

    std::vector<float> defaultDirections = {
        // clang-format off
        // azi, ele,
        90, 0,
        -90, 0,
        114, 20,
        -60, -6,
        85, -44,
        -130, -21,
        49, 41,
        -67, 52,
        154, -9,
        -48, -55,
        19, -31,
        -162, 21,
        151, 73,
        -9, 20,
        180, -63,
        // clang-format on
    };

    auto const configDirFloats = Config::getFloatVectorDef("Panner:directions", defaultDirections);
    auto const floatCount = configDirFloats.size();
    EAR_CHECK_REL(floatCount > 0 && floatCount % 2 == 0,
        "Need an even number of panner floats for panner directions");
    config.directions.reserve(floatCount / 2);

    for (std::size_t i{ 0 }; i < floatCount; i += 2) {
        config.directions.push_back({ configDirFloats[i], configDirFloats[i + 1] });
    }

    config.oldGainWeight = Config::getFloatDef("Panner:oldGainWeight", 0.F);
    config.spreadAuxiliarySourceCount = Config::getIntDef("Panner:spreadAuxiliarySourceCount", 0);
    config.spreadRadiusDeg = Config::getFloatDef("Panner:spreadRadiusDeg", 0.F);
    config.spreadGainFraction = Config::getFloatDef("Panner:spreadGainFraction", 0.F);

    config.inputCount = 0; // We'll get the real input count in `update`.

    return config;
}

} // namespace

PannerStage::PannerStage() {}

PannerStage::~PannerStage() {}

std::unique_ptr<RenderStage::DspData> PannerStage::update(RenderList& renderList)
{
    if (B == 0) { // only update once initialized
        return {};
    }

    auto dspData = std::make_unique<PannerStage::DspData>();
    auto& dspList = *dspData;

    for (auto const& item : renderList.items) {
        if (!(isPannableInputRenderItem(*item))) {
            continue;
        }

        EAR_CHECK_REL(item->signal->getFrame().channels() == 1,
            "Multichannel inputs to panner not supported yet.")
        EAR_CHECK_REL(
            item->position.cSpace == CSpace::LCS, "Only LCS CSpace supported in PannerStage.")
        auto polar = EAR::cartesianToPolar(item->position.location);

        dspList.pannerInputs.push_back(
            { polar[0], polar[1], item->gain, item->signal->getFrame()[0].data() });
    }

    std::vector<float*> output;
    output.reserve(outputRenderItems.size());
    dspList.outputSignals.reserve(outputRenderItems.size());
    for (auto* ri : outputRenderItems) {
        dspList.outputSignals.push_back(&ri->signal->getFrame()[0]);
        output.push_back(dspList.outputSignals.back()->data());
        renderList.addItemCopy(ri);
        // Inactivate output if no inputs to skip spatializing silence.
        if (dspList.pannerInputs.empty()) {
            ri->setStatus(ItemStatus::Inactive);
        } else {
            ri->setStatus(ItemStatus::Active);
        }
        ri->resetChanged();
    }
    mpegcore::WritableAudioPointerVector out(output.size(), B, output.data());

    // Check that either we don't have a panner yet, the panner doesn't have any inputs (e.g. if the
    // listener previously was in an environment that generated Pannable items and then has moved to
    // an environment where they are not generated then the panner doesn't have inputs anymore), or
    // the input count is according to the configuration.
    EAR_CHECK_REL(!panner.has_value() || dspList.pannerInputs.size() == 0
            || pannerConfig.inputCount == dspList.pannerInputs.size(),
        "Panner input count change not supported");

    auto const inputCount = dspList.pannerInputs.size();
    if (!panner.has_value() && inputCount > 0) {
        pannerConfig.inputCount = inputCount;
        auto pannerMaybe = mpegcore::Panner::create(pannerConfig, out);
        EAR_CHECK_REL(pannerMaybe.has_value(), "Failed to create Panner.")
        panner.emplace(std::move(pannerMaybe.value()));
    }

    if (panner.has_value()) {
        dspList.panner = &panner.value();
    }

    return dspData;
}

void PannerStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<PannerStage::DspData>();

    if (currentDSP.panner == nullptr) {
        return;
    }

    currentDSP.panner->process(currentDSP.pannerInputs);
}

void PannerStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, [[maybe_unused]] const SceneState* state)
{
    B = streamManager->getB();

    if (outputRenderItems.empty()) {
        pannerConfig = getPannerConfig();
        pannerConfig.bufferSize = B;

        for (std::size_t i{ 0 }, end{ pannerConfig.directions.size() }; i < end; i++) {
            outputRenderItems.push_back(itemStore->createItem("panner_" + std::to_string(i)));
            auto* renderItem = outputRenderItems.back();
            renderItem->type = ItemType::Nondiegetic;
            renderItem->setSignal(streamManager->createEmptyBuffer());
            renderItem->setGain(1.F);

            // Place output channels to their designed positions.
            EAR::Position pos{};
            pos.location = EAR::polarToCartesian({ pannerConfig.directions[i].azimuthDeg,
                pannerConfig.directions[i].elevationDeg, 1.F });
            pos.cSpace = EAR::CSpace::LCS;
            renderItem->setPosition(pos);
        }
    }
}

} // namespace MPEGI
