/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <algorithm>
#include <random>

#include <doctest/doctest.h>

#include "../PannerStage.hpp"
#include <EAR/RealFrame.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Stream/StreamManager.h>

using namespace MPEGI;

namespace {

// StreamManager is abstract but it contains all the functionality needed for testing Panner. Any
// dummy that compiles is fine for Panner testing.
class DummyStreamManager : public StreamManager
{
public:
    DummyStreamManager(float fs, std::size_t B)
        : StreamManager(fs, B)
    {}

    std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const std::string& streamId, double t = 0.0) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const SceneState::AudioStream& stream) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    void removeAccessBuffer(StreamAccessBuffer* buf) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    void nextBlock() override { EAR_CHECK_REL(false, "Not expected to be called"); }

    std::size_t getNbrOfChannels(const std::string& id) const override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }
};

void whiteNoise(Buffer<float>& buffer)
{
    static std::default_random_engine engine;
    static std::uniform_real_distribution<> distribution(-0.99, 0.99);
    auto* data = buffer.data();
    std::transform(data, data + buffer.length(), data,
        []([[maybe_unused]] float value) { return static_cast<float>(distribution(engine)); });
}

} // namespace

TEST_CASE("PannerStage panning")
{
    // When other tests are added it probably makes sense to refactor creating RenderItemStore,
    // StreamManager, Bitstream and SceneState to a common helper. With stages that need SceneState
    // it would probably help if it was possible to construct SceneState in code without a real
    // Bitstream file.
    RenderItemStore store;
    DummyStreamManager streamManager(48000.F, 256U);
    Bitstream* bitstream{ nullptr };
    SceneState* sceneState{ nullptr };

    PannerStage pannerStage;
    pannerStage.initializeStage(*bitstream, &store, &streamManager, sceneState);
    auto items = store.getItems();
    constexpr std::size_t pannerItemCount{ 15 };
    CHECK(items.size() == pannerItemCount);

    auto* pannableItem = store.createItem("PannableItem");
    items = store.getItems();
    CHECK(items.size() == pannerItemCount + 1);
    pannableItem->type = ItemType::Pannable;
    pannableItem->setSignal(streamManager.createEmptyBuffer());
    whiteNoise(pannableItem->signal->getFrame()[0U]);
    CHECK(pannableItem->signal->getFrame()[0U].rms() > 0);
    pannableItem->setGain(1.F);
    EAR::Position position;
    // Set location of Pannable to first RenderItem of the Panner. After panning the first
    // RenderItem should contain audio while the others should not.
    position.location = items[0]->position.location;
    position.cSpace = EAR::CSpace::LCS;
    pannableItem->setPosition(position);

    // Check no audio in Panner output before rendering.
    for (std::size_t i{ 0 }; i < pannerItemCount; i++) {
        CHECK(items[i]->signal->getFrame()[0U].rms() == 0.F);
    }

    RenderList renderList(0);
    renderList.addItemCopy(pannableItem);
    auto dspData = pannerStage.update(renderList);
    pannerStage.processBlock(RenderStage::DspContext{ *dspData });
    // First Panner RenderItem should contain audio.
    CHECK(items[0]->signal->getFrame()[0U].rms() > 0.F);
    // Other Panner RenderItems should be silent.
    for (std::size_t i{ 1 }; i < pannerItemCount; i++) {
        CHECK(items[i]->signal->getFrame()[0U].rms() == 0.F);
    }
}
