/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Portal.h"
#include "../ReverbStage/ReverbStage.h"
#include "EAR/Exception.h"

#include <EAR/Portal.h>
#include <EAR/PortalReverbExtent.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

#include <algorithm>
#include <cstdarg>
#include <vector>

namespace MPEGI {

const float minTransmissionForRevExtent = 1e-6f;

std::shared_ptr<RenderStage> makePortalStage()
{
    return std::make_shared<PortalStage>();
}

PortalStage::PortalStage()
    : itemStore(nullptr)
{}

PortalStage::~PortalStage()
{
    for (auto [portalId, portal] : portals) {
        if (portal->scenePortal != nullptr) {
            portal->scenePortal->detach(this);
        }
    }
}

// Returns true when the maximum transmission coeff of the combination of portals with portalIds is larger than minTransmissionForRevExtent, and false otherwise.
// For this, it multiplies the transmission coeffs for all portals for every frequency for which they are defined.
// Assumes air if either the material, or the transmission coeffs of a portal are not set.

bool transmissionCoeffsAreLargeEnough(
    const std::vector<int>& portalIds, const std::map<int, MPEGI::Portal*>& portals)
{
    // Get all unique frequencies for the portal materials (unordered_set automatically removed duplicate frequencies)
    std::unordered_set<float> allFrequencies;
    for (int portalId : portalIds) {
        const auto& mat = portals.at(portalId)->scenePortal->getMaterial();
        if (mat) {
            const auto& frequencies = mat->getFrequencies();
            allFrequencies.insert(frequencies.begin(), frequencies.end());
        }
    }

    if (allFrequencies.empty()) {
        // No material set for any portal. Assume air for all of them.
        return true;
    }

    // Combine the transmission from all portals for all frequencies
    std::vector<float> combinedTransmission(allFrequencies.size(), 1.f);
    for (int portalId : portalIds) {
        const auto& mat = portals.at(portalId)->scenePortal->getMaterial();

        // If any material is missing then we assume air (i.e. transmission=1).
        if (mat) {
            const auto& tc = mat->getTransmissionCoeffs();
            if (tc.size() > 0) {
                int freqId = 0;
                for (const float f : allFrequencies) {
                    const float tc = mat->getTransmissionCoeffForFreq(f);
                    combinedTransmission[freqId] *= tc;
                    ++freqId;
                }
            } else {
                //Material was set, but no transmission coefficients were set. Assume material has no transmission (and is therefore not large enough)
                return false;
            }
        }
    }

    if (combinedTransmission.empty()) {
        // No materials are set for all portals: assume air.
        return true;
    } else {
        // At least one portal material is set and has transmission coeffs.
        const float maxTc =
            *std::max_element(combinedTransmission.begin(), combinedTransmission.end());

        return (maxTc > minTransmissionForRevExtent);
    }
}

bool PortalStage::isReverbExtentOfPortalActive(int portalId,
    const std::string& listenerAcousticEnvironmentId,
    const std::string& sourceAcousticEnvironmentId)
{
    // Is the portal directly connected to the AE of interest (listenerAcousticEnvironmentId)?
    if (portals[portalId]->portalEnvironments[0] == listenerAcousticEnvironmentId
        || portals[portalId]->portalEnvironments[1] == listenerAcousticEnvironmentId) {

        return transmissionCoeffsAreLargeEnough({ portalId }, portals);
    }

    // If no direct connection check all portals of the AE at the other side of the portal of interest

    // Determine what the AE at the other side of the portal is
    std::string otherSideAcousticEnvironmentId = portals[portalId]->portalEnvironments[0];
    if (portals[portalId]->portalEnvironments[0] == sourceAcousticEnvironmentId) {
        otherSideAcousticEnvironmentId = portals[portalId]->portalEnvironments[1];
    }

    // Check all connected portals
    for (int secondPortalId : portalMap[otherSideAcousticEnvironmentId]) {
        if (portals[secondPortalId]->portalEnvironments[0] == listenerAcousticEnvironmentId
            || portals[secondPortalId]->portalEnvironments[1] == listenerAcousticEnvironmentId) {
            return transmissionCoeffsAreLargeEnough({ portalId, secondPortalId }, portals);
        }
    }

    // No connection, directly or indirectly
    return false;
}

std::pair<bool, std::vector<std::string>> PortalStage::isCouplingActiveForPortal(
    int portalId, const std::string& listenerAcousticEnvironmentId)
{
    // Is the portal directly connected to the AE of interest (listenerAcousticEnvironmentId)?
    if (portals[portalId]->portalEnvironments[0] == listenerAcousticEnvironmentId) {
        return { true, { portals[portalId]->portalEnvironments[1] } };
    }

    if (portals[portalId]->portalEnvironments[1] == listenerAcousticEnvironmentId) {
        return { true, { portals[portalId]->portalEnvironments[0] } };
    }

    std::vector<std::string> couplingInputAEs;

    // If no direct connection check all portals of the two AEs the portal is connected to
    for (int secondPortalId : portalMap[portals[portalId]->portalEnvironments[0]]) {
        if (portals[secondPortalId]->portalEnvironments[0] == listenerAcousticEnvironmentId
            || portals[secondPortalId]->portalEnvironments[1] == listenerAcousticEnvironmentId) {
            couplingInputAEs.push_back(portals[portalId]->portalEnvironments[1]);
            break;
        }
    }

    // Check other AE the portal is connected to
    for (int secondPortalId : portalMap[portals[portalId]->portalEnvironments[1]]) {
        if (portals[secondPortalId]->portalEnvironments[0] == listenerAcousticEnvironmentId
            || portals[secondPortalId]->portalEnvironments[1] == listenerAcousticEnvironmentId) {
            couplingInputAEs.push_back(portals[portalId]->portalEnvironments[0]);
            break;
        }
    }

    return { !couplingInputAEs.empty(), couplingInputAEs };
}

bool PortalStage::isListenerOnActiveSide(const EAR::Vector3& listenerPos,
    const RenderItem* reverbExtentRI, const MPEGI::Portal* portal, bool isFirstPortalAE)
{
    const auto normalVec = portal->normalVector;

    // The portal position is at the edge of the portal, we need the point to be on the extent itself.
    EAR::Vector3 portalOffsetPosition;
    const float halfPortalDepth = portal->scenePortal->getPortalDims()[2] / 2.0f;
    if (isFirstPortalAE) {
        portalOffsetPosition = reverbExtentRI->position.location + normalVec * halfPortalDepth;
    } else {
        portalOffsetPosition = reverbExtentRI->position.location - normalVec * halfPortalDepth;
    }

    // Dot product to see if listener position is in same direction as normal vector
    // Normal vector always points into the first AE of the portal
    if (((listenerPos - portalOffsetPosition).dot(portal->normalVector) >= 0) != isFirstPortalAE) {
        return true;
    } else {
        return false;
    }
}

void applyImmersiveReverbXfadeGain(RenderList& renderList, float immersiveReverbXfadeGain)
{
    // Apply cross-fade gain to all immersive reverb items that the reverb stage has output
    for (auto& item : renderList.items) {
        if (item->type == MPEGI::ItemType::Pannable) {
            const auto* state = item->getState<MPEGI::ReverbStage::ItemState>();
            if (state && state->isImmersiveReverb) {
                item->addGain(immersiveReverbXfadeGain);
            }
        }
    }

    return;
}

// Converts material coupling coefficients to third octave frequencies, and set as eq for item.
void setCouplingCoeffsAsEq(RenderItem* item, const EAR::Material& material,
    const std::vector<float>& thirdOctaveCenterFreqs)
{
    if (material.getCouplingCoeffs().size() != 0) {

        std::vector<float> eq;
        eq.reserve(thirdOctaveCenterFreqs.size());
        for (const auto& f : thirdOctaveCenterFreqs) {
            eq.push_back(sqrt(material.getCouplingCoeffForFreq(f)));
        }

        item->setEq(eq);
    }
};

void PortalStage::objectChanged(SceneObject* obj, Property modification)
{
    if (modification == Property::Material) {
        auto* portal = dynamic_cast<EAR::Portal*>(obj);
        if (portal) {
            setCouplingCoeffsAsEq(couplingItems[portal->getPortalId()], *portal->getMaterial(),
                thirdOctaveCenterFreqs);
        }
    }
}

std::unique_ptr<RenderStage::DspData> PortalStage::update(RenderList& renderList)
{
    if (!itemStore) {
        return {};
    }

    auto dspData = std::make_unique<PortalStage::DspData>();

    const std::string& listenerAcousticEnvironmentId = renderList.listener.acousticEnvironmentId;
    bool noPortal = (portalMap.find(listenerAcousticEnvironmentId) == portalMap.end());

    // Aspect 1 using portals: Process reverb extent sources for playback of reverb from other environments

    // Deactivate all reverb extent RIs when no portal is associated with the listener's environment.
    if (noPortal) {
        for (auto [portalId, reverbExtentRenderItems] : reverbExtentRenderItemsMap) {
            for (auto reverbExtentRI : reverbExtentRenderItems) {
                // In implicit portal mode, there is only one render item
                if (reverbExtentRI == nullptr) {
                    continue;
                }

                // If there's no portal in currect AE, deactivate all portals
                itemStore->deactivate(reverbExtentRI);

                // Add portal render item to renderList
                renderList.addItemCopy(reverbExtentRI);
                reverbExtentRI->resetChanged();

                // When the listener switches from an AE to 'outside any AE' (and there is no portal to outside)
                // the deactivation triggers a fade-out of the directional reverb RI while no signal is updated.
                // To prevent artefacts, we keep feeding signal until it is faded out.
                if (reverbExtentRI->status != ItemStatus::Inactive) {
                    // Find corresponding reverb item, and assign signal
                    for (auto& item : renderList.items) {
                        if (item->type == ItemType::Reverb
                            && item->acousticEnvironmentId
                                == reverbExtentRI->acousticEnvironmentId) {
                            // Save input reverb + output portal render item combinations in dspData (so processBlock can process these).

                            auto [reverbExtentSrcOut, inserted] =
                                dspData->reverbExtentSrcOut.try_emplace(
                                    item->acousticEnvironmentId);
                            if (inserted) {
                                dspData->reverbIn[item->acousticEnvironmentId] =
                                    &item->signal->getFrame();
                            }
                            reverbExtentSrcOut->second.push_back(
                                &reverbExtentRI->signal->getFrame());
                            break;
                        }
                    }
                }
            }
        }
    } else {

        std::vector<MPEGI::RenderItem*> reverbExtentRIsToAdd;

        auto nrRIs = renderList.items.size();
        for (int itemIdx = 0; itemIdx < nrRIs; itemIdx++) {
            // Go over all immersive reverb RIs, and find which directional reverb RIs to activate and configure for it.
            if (renderList.items[itemIdx]->type != ItemType::Reverb) {
                continue;
            }
            // Do we have portal mapping data for this reverb?
            if (portalMap.find(renderList.items[itemIdx]->acousticEnvironmentId)
                == portalMap.end()) {

                continue;
            }

            PortalCrossFader crossfader(itemStore, initialDirectionalReverbGain);

            // For each reverb renderItem, find all portals related to that reverb
            for (auto portalId : portalMap[renderList.items[itemIdx]->acousticEnvironmentId]) {
                // Need to assign item every iteration because additions to the renderList in this loop may break the reference.
                auto& item = renderList.items[itemIdx];
                float reverbExtentOffset = 1.0f;
                EAR::Vector3 listenerPos = renderList.listener.position.location;

                // Consider only portal render items that are inside the reverb's AE
                auto reverbExtentRIPair = reverbExtentRenderItemsMap[portalId];
                RenderItem* reverbExtentRI = nullptr;
                bool isFirstPortalAE;

                if (isExplicitPortalMode) {
                    for (int AEid = 0; AEid < 2; ++AEid) {
                        if (reverbExtentRIPair[AEid]->acousticEnvironmentId
                            == item->acousticEnvironmentId) {
                            reverbExtentRI = reverbExtentRIPair[AEid];
                            reverbExtentOffset =
                                portals[portalId]->scenePortal->getExtentOffset()[AEid];
                            isFirstPortalAE = (AEid == 0);
                        }
                    }
                } else {
                    reverbExtentRI = reverbExtentRIPair[0];
                    isFirstPortalAE = true;
                }

                // Depending on whether the listener is in this reverb AE or not, we do something different
                if (item->acousticEnvironmentId == listenerAcousticEnvironmentId) {
                    // Listener is in this AE - Apply crossfade when listener is close enough to the portal
                    // For implicit portal mode, deactivate the reverb extent RI (when active)
                    if (isExplicitPortalMode) {
                        reverbExtentRI->setStatus(ItemStatus::Active);
                        crossfader.addItem(portalExtents[portalId], listenerPos, reverbExtentRI,
                            reverbExtentOffset);
                    } else {
                        // Deactivate the reverb extent RI (when active)
                        itemStore->deactivate(reverbExtentRI);
                    }
                } else {
                    // Listener is not in this AE - Only activate the reverb extent RI when the portal connects to the listener AE

                    if (isExplicitPortalMode) {
                        // Activate when the current portal connects to listener AE and the listener is on the right side of the reverb extent RI
                        if (isReverbExtentOfPortalActive(portalId, listenerAcousticEnvironmentId,
                                item->acousticEnvironmentId)
                            && isListenerOnActiveSide(
                                listenerPos, reverbExtentRI, portals[portalId], isFirstPortalAE)) {
                            if (reverbExtentRI->status != MPEGI::ItemStatus::Active) {
                                reverbExtentRI->setGain(initialDirectionalReverbGain);
                                reverbExtentRI->setStatus(ItemStatus::Active);
                            }

                        } else {
                            // Deactivate the reverb extent RI
                            itemStore->deactivate(reverbExtentRI);
                        }
                    } else {
                        if (reverbExtentRI->status != MPEGI::ItemStatus::Active) {
                            reverbExtentRI->setGain(initialDirectionalReverbGain);
                            reverbExtentRI->setStatus(ItemStatus::Active);
                        }
                    }
                }

                // Save input reverb + output portal render item combinations in dspData (so processBlock can process these).
                if (dspData->reverbExtentSrcOut.find(item->acousticEnvironmentId)
                    == dspData->reverbExtentSrcOut.end()) {
                    dspData->reverbIn[item->acousticEnvironmentId] = &item->signal->getFrame();
                    dspData->reverbExtentSrcOut[item->acousticEnvironmentId] =
                        std::vector<RealFrame*>();
                }

                // For all reverb extent RIs that are not completely inactive (i.e. also continue during fade-out)...
                if (reverbExtentRI->status != ItemStatus::Inactive) {
                    dspData->reverbExtentSrcOut[item->acousticEnvironmentId].push_back(
                        &reverbExtentRI->signal->getFrame());
                }

                // Add to list of RIs to be added to renderList. Gains may still be modified in case the listener is in two or more cross-fade regions
                reverbExtentRIsToAdd.emplace_back(reverbExtentRI);
            }

            if (isExplicitPortalMode) {
                crossfader.crossFade(renderList);
            }
        }

        // Add the reverb extent RIs to the renderList
        for (auto riItr : reverbExtentRIsToAdd) {
            renderList.addItemCopy(riItr);
            riItr->resetChanged();
        }
    }

    // Aspect 2 using portals: Process sources for coupled transmission through portals
    if (enableCoupling) {
        for (const auto& [portalId, portal] : portals) {
            float maxCouplingCoeff = 0.f;
            float maxTransmissionCoeff = 0.f;
            MPEGI::RenderItem* couplingItem = couplingItems[portalId];

            if (portal->scenePortal == nullptr || portal->scenePortal->getMaterial() == nullptr) {
                // No material is set. This code assumes air, and therefore no coupling.
                if (couplingItem->status != ItemStatus::Inactive) {
                    itemStore->deactivate(couplingItem);
                }
                continue;
            }
            const auto material = portal->scenePortal->getMaterial();

            if (material->getCouplingCoeffs().size() > 0) {
                maxCouplingCoeff = *std::max_element(
                    material->getCouplingCoeffs().begin(), material->getCouplingCoeffs().end());
            }

            if (material->getTransmissionCoeffs().size() > 0) {
                maxTransmissionCoeff = *std::max_element(material->getTransmissionCoeffs().begin(),
                    material->getTransmissionCoeffs().end());
            }

            // Only portals that connect to the listener room and 1 room further away are potentially active
            auto [isActive, couplingInputAcousticEnvironmentIds] =
                isCouplingActiveForPortal(portalId, renderList.listener.acousticEnvironmentId);

            // Deactivate all render items corresponding to materials with low coupling coeffs. Also deactivate coupling when the coupling coefficient is low compared to transmission, as that would be inaudible.
            isActive = isActive && (maxCouplingCoeff > couplingThreshold)
                && (maxCouplingCoeff > maxTransmissionCoeff * transmissionThreshold);

            if (!isActive) {
                // Newly created RIs are created as inactive. This check prevents them from being put into "FadeOutInactive" by the deactivate call.
                if (couplingItem->status != ItemStatus::Inactive) {
                    itemStore->deactivate(couplingItem);
                }
            }

            if (isActive || (couplingItem->status != ItemStatus::Inactive)) {
                // Process all items that are not inactive. This includes items during fadeout which have been deactived above but were active earlier.
                PortalStage::CouplingData data;
                data.portalId = portalId;
                data.couplingInputAEs = couplingInputAcousticEnvironmentIds;
                data.output = &couplingItem->signal->getFrame();

                // Find all render items with a portal contribution (was computed previously in the reverb stage)
                for (const auto& item : renderList.items) {
                    if (item->status == ItemStatus::Inactive) {
                        continue;
                    }

                    // Only consider renderitems from the correct reverb id (i.e. "behind" the portal wrt to the listener).
                    if (std::find(couplingInputAcousticEnvironmentIds.begin(),
                            couplingInputAcousticEnvironmentIds.end(), item->acousticEnvironmentId)
                        != couplingInputAcousticEnvironmentIds.end()) {
                        const auto state = item->getState<ReverbStage::ItemState>();
                        if (state && state->portalContributions.count(portalId) > 0) {
                            // Item has a portal contribution for this portal
                            PortalStage::CouplingInput input;
                            input.audioIn = item->signal;
                            input.gain = item->gain;
                            input.portalContribution = state->portalContributions.at(portalId);

                            data.inputs.push_back(input);
                        }
                    }
                }

                // If there are inputs for the coupling (either primary items, or reverb), add to renderList and dspData. Otherwise, ignore
                bool reverbForCoupling = false;
                for (const auto& couplingInputAE : data.couplingInputAEs) {
                    if (dspData->reverbIn.count(couplingInputAE) > 0) {
                        reverbForCoupling = true;
                        break;
                    }
                }

                if (!data.inputs.empty() || reverbForCoupling) {
                    dspData->couplingData.push_back(data);
                } else {
                    isActive = false;
                    if (couplingItem->status != ItemStatus::Inactive) {
                        itemStore->deactivate(couplingItem);
                    }
                }
            }

            if (isActive) {
                couplingItem->setStatus(MPEGI::ItemStatus::Active);
            }

            renderList.addItemCopy(couplingItem);
            couplingItem->resetChanged();
        }
    }

    return dspData;
}

void reverbDownmix(const EAR::RealFrame& reverbIn, EAR::RealFrame* output, bool downmixAllReverbCh)
{
    auto outputChannelCount = output->channels();
    EAR_CHECK_REL(
        outputChannelCount == 1 || outputChannelCount == 2, "Output channels must be 1 or 2");

    auto inputChannelCount = reverbIn.channels();
    if (downmixAllReverbCh) {
        if (outputChannelCount == 2) {
            // Last channel is mixed to both channels to preserve energy
            EAR::RealBuffer lastChannelCopy(reverbIn.ch(inputChannelCount - 1));
            lastChannelCopy.mul(0.5);
            output->ch(0).add(lastChannelCopy);
            output->ch(1).add(lastChannelCopy);
        } else if (outputChannelCount == 1) {
            output->ch(0).add(reverbIn.ch(inputChannelCount - 1));
        }
    }

    // leave out last channel as that was already mixed (or is to be omitted)
    for (int c = 0; c < inputChannelCount - 1; c++) {
        // Either alternatively add to channel 0 and 1 (for 2 output channels), or add all to channel 0 (1 output channel)
        output->ch(c % outputChannelCount).add(reverbIn.ch(c));
    }

    if (!downmixAllReverbCh) {
        // scale to restore energy if the last channel was omitted
        const float scaler = inputChannelCount / (inputChannelCount - 1.0f);
        for (int c = 0; c < outputChannelCount; ++c) {
            output->ch(c).mul(scaler);
        }
    }
}

void PortalStage::processBlock(DspContext ctx)
{
    auto& dspData = ctx.getData<PortalStage::DspData>();

    for (auto& reverbInPair : dspData.reverbIn) {
        const std::string& acousticEnvironmentId = reverbInPair.first;
        EAR::RealFrame& reverbInFrame = *reverbInPair.second;

        if (dspData.reverbExtentSrcOut[acousticEnvironmentId].size() > 0) {

            // Downmix input channels into 2 output channels
            // First downmix to first output
            auto firstOutputPtr = dspData.reverbExtentSrcOut[acousticEnvironmentId][0];

            firstOutputPtr->zero();
            reverbDownmix(reverbInFrame, firstOutputPtr, downmixAllReverbCh);

            // copy the downmixed reverb from first output into other output portal items
            for (int i = 1; i < dspData.reverbExtentSrcOut[acousticEnvironmentId].size(); i++) {
                dspData.reverbExtentSrcOut[acousticEnvironmentId][i]->copyFrom(*firstOutputPtr);
            }
        }
    }

    // Process coupling
    if (enableCoupling) {
        for (auto& coupling : dspData.couplingData) {
            coupling.output->zero();
            // Add primary render items
            for (const auto& input : coupling.inputs) {
                for (size_t inputChannel = 0; inputChannel < input.audioIn->getFrame().channels();
                     ++inputChannel) {
                    EAR_CHECK(input.audioIn->getFrame().ch(inputChannel).length()
                            == coupling.output->ch(0).length(),
                        "Error: Can't add buffers of unequal length");
                    float* outBegin = coupling.output->ch(0).data();
                    float* outEnd = outBegin + coupling.output->ch(0).length();
                    const float* inBegin = input.audioIn->getFrame().ch(inputChannel).data();

                    auto addAndApplyGains = [&input](float outputBuf, float inputBuf) {
                        return outputBuf
                            + inputBuf * input.gain * std::sqrt(input.portalContribution);
                    };

                    std::transform(outBegin, outEnd, inBegin, outBegin, addAndApplyGains);
                }
            }

            // Add reverb to coupling
            for (const auto& couplingInputAE : coupling.couplingInputAEs) {
                auto reverbExtentSrcOutPtr = dspData.reverbExtentSrcOut.find(couplingInputAE);

                if (reverbExtentSrcOutPtr != dspData.reverbExtentSrcOut.end()) {
                    // A reverb item exists in the right AE
                    auto [acousticEnvironmentId, reverbExtentSrcSig] = *reverbExtentSrcOutPtr;
                    // Reverb was previously downmixed to two channels for the portal reverb meshes. Downmix to single output channel for coupling.
                    if (reverbExtentSrcSig.size() > 0) {
                        for (size_t inputChannel = 0;
                             inputChannel < reverbExtentSrcSig[0]->channels(); ++inputChannel) {
                            coupling.output->ch(0).add(reverbExtentSrcSig[0]->ch(inputChannel));
                        }
                    } else {
                        // Inactive portals have an empty reverbExtentSrcOut, but do have reverbIn set. In that case, the downmix still needs to be done.
                        auto reverbInPtr = dspData.reverbIn.find(couplingInputAE);
                        if (reverbInPtr != dspData.reverbIn.end()) {
                            reverbDownmix(
                                *reverbInPtr->second, coupling.output, downmixAllReverbCh);
                        }
                    }
                }
            }
        }
    }
}

// Convenience function to handle setting up render items for portals
void setupRenderItem(RenderItem* item, const Geometry* portalExtent, float portalGain,
    const MPEGI::StreamManager* streamManager, std::string acousticEnvironmentId,
    const float reverbExtentOffset, int portalId)
{
    item->type = ItemType::DirectionalReverb;
    item->setExtent(portalExtent);
    item->setPosition(portalExtent->getPosition());
    item->acousticEnvironmentId = acousticEnvironmentId;
    item->setGain(portalGain);
    item->setStatus(ItemStatus::Inactive);
    item->setSignal(streamManager->createEmptyBuffer(2));
    item->setAParams(AParam::ForceNoDoppler);
    item->portalDistComp = std::fmin(std::sqrt(item->extent->getVolume()) * 0.05, 15.f);
    item->parentPortalID = portalId;

    item->refDistance = reverbExtentOffset;
}

void PortalStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    this->itemStore = itemStore;
    B = streamManager->getB();

    const auto portalPayload = std::dynamic_pointer_cast<const PortalPayload>(
        bitstream.getOnlyPayloadOfType(Bitstream::PLD_PORTAL));
    isExplicitPortalMode = portalPayload->getExplicitPortalMode();

    bool hasDefaultAe = false;
    auto reverbPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
    if (reverbPayloads.size() != 0) {
        auto bitstreamReverbPayload =
            std::dynamic_pointer_cast<const MPEGI::ReverbPayload>(reverbPayloads[0]);
        hasDefaultAe = bitstreamReverbPayload->revDefaultAEPresent;
    }

    // map AE's bitstream id to the acousticEnvironmentId
    // TODO: Find out if lsdfEnv needs to be handled
    std::vector<std::string> lsdfEnv;
    // TODO: How to handle default AE
    std::string defaultAEId = "";

    for (auto& reverb : state->getReverbEnvs()) {
        if (reverb->isLsdf) {
            lsdfEnv.push_back(reverb->acousticEnvironmentId);
        }

        // if there's a default AE in the scene, it should replace "outside"
        if (!reverb->enclosure && hasDefaultAe) {
            defaultAEId = reverb->acousticEnvironmentId;
        }
    }

    // initialize all portals
    const float directionalReverbGain = Config::getFloatDef("DirectionalReverb_Gain", 1.f);
    initialDirectionalReverbGain =
        directionalReverbGain; // TODO: Not real-time modifiable this way. Should it be? Perhaps best to remove from config.
    for (auto& p : portalPayload->getPortals()) {

        portals[p->portalId] = p.get();
        portals[p->portalId]->normalVector = portals[p->portalId]->normalVector.normalize();
        if (p->extentId.has_value()) {
            portalExtents[p->portalId] = state->getGeometry(p->extentId.value());
        } else {
            if (p->equivalentBoxPointer != nullptr) {
                portalExtents[p->portalId] = p->equivalentBoxPointer.get();
            } else {
                EAR_CHECK_REL(p->equivalentBoxPointer != nullptr,
                    "Portal " + p->portalStringId
                        + "has missing extentId and equivalentBoxPointer, one should be set.");
                portalExtents[p->portalId] = p->equivalentBoxPointer.get();
            }
        }

        if (isExplicitPortalMode) {
            const auto& reverbExtentOffsets = p->scenePortal->getExtentOffset();
            // Create two RenderItems for each portal. Each portal has an extent on each side of the portal.
            for (int AEid = 0; AEid < 2; ++AEid) {

                RenderItem* renderItem = itemStore->createItem("reverbExtentItem_"
                    + std::to_string(p->portalId) + "_forAE_" + p->portalEnvironments[AEid]);

                setupRenderItem(renderItem,
                    p->scenePortal->getExternalReverbExtents()[AEid]->getMesh(),
                    directionalReverbGain, streamManager, p->portalEnvironments[AEid],
                    reverbExtentOffsets[AEid], p->portalId);

                reverbExtentRenderItemsMap[p->portalId][AEid] = renderItem;
            }
        } else {
            // Create one RenderItem for each portal
            RenderItem* renderItem = itemStore->createItem(
                "reverbExtentItem_" + std::to_string(p->portalId) + "_forAE_" + p->sourceId);
            setupRenderItem(renderItem, (Geometry*)portalExtents[p->portalId]->getMesh(),
                directionalReverbGain, streamManager, p->sourceId, 1.0f, p->portalId);
            reverbExtentRenderItemsMap[p->portalId][0] = renderItem;
            reverbExtentRenderItemsMap[p->portalId][1] = nullptr;
        }
    }

    // Copy portal map
    portalMap = portalPayload->getPortalMap();

    downmixAllReverbCh = Config::getBoolDef("Portal_DownmixAll", true);

    enableCoupling = Config::getBoolDef("Portal_Enable_Coupling", true);

    if (enableCoupling) {

        thirdOctaveCenterFreqs = { MPEGI::Config::getFloatPtr("EQCenterFreqs"),
            MPEGI::Config::getFloatPtr("EQCenterFreqs") + MPEGI::Config::getInt("EQNBands") };

        // Initialize coupling render items
        for (auto& p : portalPayload->getPortals()) {
            RenderItem* item =
                itemStore->createItem("couplingItem_for_portal_" + std::to_string(p->portalId));

            item->type = ItemType::Primary;
            item->subtype = ItemSubtype::CoupledSrc;
            item->setExtent(portalExtents[p->portalId]->getMesh());
            item->setPosition(portalExtents[p->portalId]->getMesh()->getPosition());
            item->setStatus(ItemStatus::Inactive);
            item->setSignal(streamManager->createEmptyBuffer(1));
            item->setAParams(AParam::ForceNoDoppler);
            item->parentPortalID = p->portalId; // Probably unused
            item->resetChanged();

            couplingItems[p->portalId] = item;

            if (p->scenePortal != nullptr) {
                if (p->scenePortal->getMaterial() != nullptr) {
                    setCouplingCoeffsAsEq(
                        item, *p->scenePortal->getMaterial(), thirdOctaveCenterFreqs);
                }
                p->scenePortal->attach(this);
            }
        }
    }
}

EAR::Vector3 intersect(
    const EAR::Vector3& normal, const EAR::Vector3& rayPoint, const EAR::Vector3& planePoint)
{
    const EAR::Vector3 diff = rayPoint - planePoint;
    const double alpha = diff.dot(normal);
    return rayPoint - alpha * normal;
}

void getCrossFadeGains(const EAR::Vector3& ptOnExtent, const EAR::Vector3& ptOnPortal,
    const EAR::Vector3& ptListener, float& reverbXfadeGain, float& portalXfadeGain)
{
    // To determine the user pos relative to the portal and the extent,
    // first the closest points to the extent and portal are determined (ptOnExtent and ptOnPortal).
    // A line through these closest points is constructed
    // A plane through the listener position with a normal vector equal to the line direction vector is constructed
    // The intersection between the plane and the line determines the position where the listener is relative to the closest points
    // Two vectors pointing from the intersection to the closest points are constructed, and used to determine if the listener is between the closest points or outside these points

    // normal vector of a plane through p, pointing from p1 to p2
    const EAR::Vector3 n = (ptOnPortal - ptOnExtent).normalize();

    // intersection of a line through p1 and p2 and a plane trhough p with normal vector n
    const EAR::Vector3 pa = intersect(n, ptOnExtent, ptListener);

    double crossFadeCoeff = 0.0;

    // determine where exactly p is located along the line through p1 and p2
    const EAR::Vector3 d1 = ptOnExtent - pa;
    const EAR::Vector3 d2 = ptOnPortal - pa;
    if (d1.dot(d2) < 0) {
        // in between p1 and p2, crossfade
        double l1 = (ptOnExtent - ptListener).abs();
        double l2 = (ptOnPortal - ptListener).abs();
        crossFadeCoeff = l2 / (l1 + l2);
    } else {
        if (d1.abs() < d2.abs()) {
            // outside ptOnExtent
            crossFadeCoeff = 1.0;
        } else {
            // outside ptOnPortal
            crossFadeCoeff = 0.0;
        }
    }

    // sine law crossfade (constant power for uncorrelated signals)
    reverbXfadeGain = sin(crossFadeCoeff * M_PI / 2.0);
    portalXfadeGain = sin((1 - crossFadeCoeff) * M_PI / 2.0);
}

PortalCrossFader::PortalCrossFader(
    MPEGI::RenderItemStore* itemStore, float initialDirectionalReverbGain)
    : itemStore(itemStore)
    , initialDirectionalReverbGain(initialDirectionalReverbGain){};

void PortalCrossFader::addItem(const EAR::Geometry* portalExtent, const EAR::Vector3& listenerPos,
    MPEGI::RenderItem* reverbExtentRI, float reverbExtentOffset)
{
    float immersiveXfadeGain = 1.0f;
    float directionalXfadeGain = 0.0f;

    if (portalExtent->isInside(listenerPos)) {

        immersiveXfadeGain = 0.0f;
        directionalXfadeGain = 1.0f;

        crossFadeEntries.emplace_back(reverbExtentRI, directionalXfadeGain, immersiveXfadeGain);

        // Activate the reverb extent RI (when deactivated)
        if (reverbExtentRI->status != MPEGI::ItemStatus::Active) {
            reverbExtentRI->setStatus(ItemStatus::Active);
        }
    } else {
        // Get closest point on the portal extent
        EAR::Vector3 closestPointOnPortalExtent = portalExtent->closestPointTo(listenerPos);
        float listener2PortalDistance = (closestPointOnPortalExtent - listenerPos).abs();

        // If within the offset at which the reverb extent is
        if (listener2PortalDistance <= reverbExtentOffset) {

            // Get closest points on the reverb extent
            EAR::Vector3 closestPointOnReverbExtent =
                reverbExtentRI->extent->closestPointTo(listenerPos);

            // Calculate cross-fade gains when the listener is in between the two points
            getCrossFadeGains(closestPointOnReverbExtent, closestPointOnPortalExtent, listenerPos,
                immersiveXfadeGain, directionalXfadeGain);

            // Store data in case we have multiple active cross-fade regions
            crossFadeEntries.emplace_back(reverbExtentRI, directionalXfadeGain, immersiveXfadeGain);

            // Activate the reverb extent RI (when deactivated)
            if (reverbExtentRI->status != MPEGI::ItemStatus::Active) {
                reverbExtentRI->setStatus(ItemStatus::Active);
            }
        } else {
            // Deactivate the reverb extent RI (when active)
            itemStore->deactivate(reverbExtentRI);
        }
    }
}

void PortalCrossFader::crossFade(MPEGI::RenderList& renderList)
{
    if (crossFadeEntries.size() == 1) {
        crossFadeEntries[0].reverbExtentRI->setGain(
            initialDirectionalReverbGain * crossFadeEntries[0].directionalCoeff);
        if (crossFadeEntries[0].immersiveCoeff < 0.999f) {
            applyImmersiveReverbXfadeGain(renderList, crossFadeEntries[0].immersiveCoeff);
        }
    } else {
        float maxDirectionalPortion = 0.0f;
        float totalDirectionalPortion = 0.0f;
        for (auto& cfItr : crossFadeEntries) {
            maxDirectionalPortion = fmax(maxDirectionalPortion, cfItr.directionalCoeff);
            totalDirectionalPortion += cfItr.directionalCoeff;
        }
        // Normalize the directional reverb coefficients
        for (auto& cfItr : crossFadeEntries) {
            cfItr.reverbExtentRI->setGain(initialDirectionalReverbGain * cfItr.directionalCoeff
                * maxDirectionalPortion / totalDirectionalPortion);
        }

        applyImmersiveReverbXfadeGain(
            renderList, sqrt(1.0f - maxDirectionalPortion * maxDirectionalPortion));
    }
}

} // namespace MPEGI
