/***************************************************************************
This software module was originally developed jointly by

Fraunhofer IIS and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGI_PORTAL_STAGE
#define MPEGI_PORTAL_STAGE

#include "MPEGI/Renderer/RenderItem.h"
#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <EAR/SceneObjectObserver.h>

#include <array>
#include <map>

namespace EAR {
class RealFrame;
}

namespace MPEGI {

//! Render the late reverb of neighboring acoustic environments as a spatially extended sound source
//! Also adds coupling from rooms to portals (i.e. doors, windows).
/*!
 * Inputs (for reverb of neighboring environments):
 * - Render items with ItemType::Reverb, which contain the reverberated signal for each AcousticEnvironment
 * - PortalPayload from the bitstream
 *
 * Outputs (for reverb of neighboring environments):
 * - New render items with ItemType::DirectionalReverb, which are then processed by the HomogeneousExtentStage
 *
 * Inputs (coupling):
 * - Render items with ItemType::Primary and ItemType::Reverb.
 * - PortalPayload from the bitstream
 *
 * Outputs (coupling):
 * - New render items with ItemType::Primary and ItemSubType::CoupledSrc, which are then processed by the HomogeneousExtentStage
 * 
 * How it works:
 * - One or multiple portal positions with an extent geometry are defined in the bitstream
 * - The portals that are audible in the current listener position (also defined in the bitstream)
 *   are added to the render list and subsequently spatialized by the HomogeneousExtent stage
 * - For coupling, all primary and reverb items are downmixed to the portal mesh (different from the reverb extent mesh). The coupling coefficients of the portal material is set as the eq of this source.
 *
 * Stages order:
 * - Required before: RoomAssignmentStage, ReverbStage
 * - Required after: EqStage, FadeStage, HomogeneousExtentStage
 *
 * Config options:
 * - (not real-time adjustable) "DirectionalReverb_Gain": global gain for all directional reverbs (default: 1)
 * - "Portal_DownmixAll": toggle downmixin all reverb channels on/off. When off, one channel is omitted.
 * - (not real-time adjustable) "Portal_Enable_Coupling": When set to false, coupling is disabled. (default: true)
 */
class PortalStage : public RenderStage, EAR::SceneObjectObserver
{
public:
    PortalStage();

    //! Destructor
    ~PortalStage();

    //! Inherited from RenderStage
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;

    //! Inherited from RenderStage
    void processBlock(DspContext ctx) override;

    float getCouplingThreshold() { return couplingThreshold; };
    float getTransmissionThreshold() { return transmissionThreshold; };

    struct CouplingInput
    {
        std::shared_ptr<MPEGI::StreamBuffer> audioIn;
        float gain;
        float portalContribution;
    };

    struct CouplingData
    {
        int portalId;
        std::vector<std::string> couplingInputAEs;
        EAR::RealFrame* output;
        std::vector<CouplingInput> inputs;
    };

    struct DspData : RenderStage::DspData
    {
        std::map<std::string, EAR::RealFrame*> reverbIn;
        std::map<std::string, std::vector<EAR::RealFrame*>> reverbExtentSrcOut;

        std::vector<CouplingData> couplingData;
    };

private:
    friend class Renderer;
    friend class PortalStageTester;

    //! Inherited from RenderStage
    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    bool isReverbExtentOfPortalActive(int portalId,
        const std::string& listenerAcousticEnvironmentId,
        const std::string& sourceAcousticEnvironmentId);

    // isCouplingActive returns true if the portal with id portalId is either connected to listenerAcousticEnvironmentId (direct connection),
    // or one of the rooms it is connected to is connected (through another portal) with the listenerAcousticEnvironmentId (second order connection).
    // Return value: bool is wether coupling should be active for this portal. The string contains the reverb ids where sources should be participating in coupling (i.e. the reverb id "behind" the portal from the perspective of the listener.
    // NOTE: In some circumstances, a coupled source should include both AEs
    std::pair<bool, std::vector<std::string>> isCouplingActiveForPortal(
        int portalId, const std::string& listenerAcousticEnvironmentId);

    bool isListenerOnActiveSide(const EAR::Vector3& listenerPos, const RenderItem* reverbExtentRI,
        const MPEGI::Portal* portal, bool isFirstPortalAE);

    bool isExplicitPortalMode{ false };

    RenderItemStore* itemStore;
    std::map<int, MPEGI::Portal*> portals;
    std::map<int, EAR::Geometry*> portalExtents; // Extent of the portal, not the reverb extent.
    std::map<std::string, std::set<int>>
        portalMap; // Maps from Acoustic Environment name to the portal ids of the portals in that Acoustic Environment
    std::map<int, std::array<RenderItem*, 2>>
        reverbExtentRenderItemsMap; // Maps from portalId to the two renderItems associated with that portal. One RenderItem for each AE the portal is connected to. Each RenderItem / AE has it's own extent.
    float initialDirectionalReverbGain =
        1.0f; // Used to prevent cross-fading gains to permanently reduce the gains of the portal render items.

    std::size_t B;

    std::vector<float> thirdOctaveCenterFreqs;

    bool enableCoupling =
        true; // Will be read from Config at initialize, but is stored thereafter for easy access
    // Maps from portalId to the renderItem that is associated with that portal for coupling.
    std::map<int, RenderItem*> couplingItems;

    const float couplingThreshold = std::pow(10.f, -60.f / 10.f);
    const float transmissionThreshold = std::pow(10.f, -6.f / 10.f);

    // Will be set from config at init or default to true
    bool downmixAllReverbCh{ true };

    // Inherited via SceneObjectObserver
    // Listens for updates to portal materials, and updates the eq for coupling accordingly
    virtual void objectChanged(SceneObject* obj, Property modification) override;
};

struct crossFadeData
{
    MPEGI::RenderItem* reverbExtentRI = nullptr;
    float directionalCoeff = 0.0f;
    float immersiveCoeff = 0.0f;
    crossFadeData(
        MPEGI::RenderItem* reverbExtentRI, float directionalCoeffSet, float immersiveCoeffSet)
        : reverbExtentRI(reverbExtentRI)
        , directionalCoeff(directionalCoeffSet)
        , immersiveCoeff(immersiveCoeffSet)
    {}
};
// PortalCrossFader is responsible for keeping track of (possibly multiple) directional reverb extent RIs and crossfading it with the immersive reverb.
// Add reverb extent items with addItem, then call crossFade when finished to perform the final crossfade between all of them.
class PortalCrossFader
{
public:
    PortalCrossFader(MPEGI::RenderItemStore* itemStore, float initialDirectionalReverbGain);
    void addItem(const EAR::Geometry* portalExtent, const EAR::Vector3& listenerPos,
        MPEGI::RenderItem* reverbExtentRI, float reverbExtentOffset);
    void crossFade(MPEGI::RenderList& renderList);

private:
    MPEGI::RenderItemStore* itemStore;
    float initialDirectionalReverbGain = 1.0f;
    std::vector<crossFadeData> crossFadeEntries; // List of directional reverb RIs
};

} // namespace MPEGI

#endif /* MPEGI_PORTAL_STAGE */
