/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include <doctest/doctest.h>

#include "../../ReverbStage/ReverbStage.h"
#include "../Portal.h"
#include "PortalTest.h"
#include <EAR/Portal.h>
#include <MPEGI/Bitstream/EIFPayload.h>

#include <algorithm>
#include <string>
// This test uses the geometry below to test the portal stage. The rooms are connected through a portal with each other. Rooms are numbered for convenience. Asterisks (*) denote the location of sound sources. The listener will move to various places during the test.
//
//                 |------|
//                 |0     |
//                 |   *  |
//                 |      |
//   |------|------|--  --|------|
//   |1     |2     |3     |4     |
//   |   *      *      *         |
//   |      |      |      |      |
//   |------|------|--  --|------|
//                 |5     |                  Portal 0 is between rooms 0, 3
//                 |      |                  Portal 1 is between rooms 1, 2
//                 |      |                  Portal 2 is between rooms 2, 3
//     --x-->      |--  --|                  Portal 3 is between rooms 3, 4
//     |           |6     |                  Portal 4 is between rooms 3, 5
//     z           |   *  |                  Portal 5 is between rooms 5, 6
//     |           |      |
//     v           |------|

namespace MPEGI {

std::string PortalStageTester::createRoom(
    DummySceneState& state, const EAR::Vector3 location, const EAR::Vector3 size)
{
    EIFPayload::BoxDesc roomBox;
    roomBox.position = EAR::Position(location, {});
    roomBox.size = size;
    roomBox.bsId = boxIdCounter++;
    roomBox.id = std::to_string(roomBox.bsId);

    state.createBox2(roomBox);

    const std::string roomAcousticEnvironmentId = std::to_string(envIdCounter++);
    const EAR::Geometry* roomGeometry = state.getGeometry(roomBox.id);
    rooms.push_back(roomGeometry);
    state.addAcousticEnvironment(roomAcousticEnvironmentId, roomGeometry);

    return roomAcousticEnvironmentId;
}

void PortalStageTester::createPortalBetweenRooms(DummySceneState& state,
    std::shared_ptr<DummyPortalPayload> portalPayload,
    const std::vector<EAR::Vector3>& roomLocations, const std::vector<std::string>& roomIds,
    size_t roomIndex1, size_t roomIndex2, const std::string& materialId)
{
    const auto& room1Location = roomLocations[roomIndex1];
    const auto& room2Location = roomLocations[roomIndex2];

    const float portalWidth = 2.f;
    const float portalHeight = 2.f;
    const float portalDepth = 0.1f;

    const EAR::Vector3 portalLocation = (room1Location + room2Location) * 0.5f;
    portalLocations.push_back(portalLocation);
    const EAR::Vector3 portalNormal = (room1Location - room2Location).normalize();
    const EAR::Euler portalOrientation =
        EAR::rotateVectorTo({ 0, 0, -1 }, portalNormal).getRotation();

    EIFPayload::BoxDesc portalBox;
    portalBox.position = EAR::Position(portalLocation, portalOrientation);
    portalBox.size = { portalWidth, portalHeight, portalDepth };
    portalBox.bsId = boxIdCounter++;
    portalBox.id = std::to_string(portalBox.bsId);

    state.createBox2(portalBox);

    // Save portal boxes for later use in tests
    const EAR::Geometry* portalGeometry = state.getGeometry(portalBox.id);
    portalBoxes.push_back(portalGeometry);

    // Add to scene state and portal payload
    std::unique_ptr<Portal> portal = std::make_unique<Portal>();
    portal->portalId = portalIdCounter++;
    portal->portalStringId = "Portal" + std::to_string(portal->portalId);
    portal->portalEnvironments = { roomIds[roomIndex1], roomIds[roomIndex2] };
    portal->extentId = portalBox.id;
    portal->normalVector = portalNormal;
    portal->materialId = materialId;

    state.createPortal2(portal.get());
    portalPayload->addPortal(std::move(portal));
}

size_t countNumberActiveItems(const RenderList& list)
{
    auto isActive = [](const std::unique_ptr<RenderItem>& item) {
        return item->status == MPEGI::ItemStatus::Active;
    };

    return std::count_if(list.items.begin(), list.items.end(), isActive);
}

size_t countNumberActiveItems(const RenderList& list, MPEGI::ItemType type)
{
    auto isActiveAndType = [type](const std::unique_ptr<RenderItem>& item) {
        return (item->status == MPEGI::ItemStatus::Active) && (item->type == type);
    };

    return std::count_if(list.items.begin(), list.items.end(), isActiveAndType);
}

PortalStageTester::PortalStageTester(
    bool setPortalMaterial, float transmissionCoeff, float couplingCoeff)
    : sceneState(hasDefaultAE)
{
    std::string materialId;
    if (setPortalMaterial) {
        materialId = sceneState.addMaterial(transmissionCoeff, couplingCoeff);
    }

    // Constructor sets up geometry as mentioned above.

    // Build geometry of rooms with room 4 in the origin
    const float roomSide = 6;
    const float roomHeight = 6;
    const EAR::Vector3 roomSize = { roomSide, roomHeight, roomSide };

    roomLocations.reserve(7);

    roomLocations.push_back(roomSide * EAR::Vector3(0, 0, -roomSide)); // Room 0
    roomLocations.push_back(2 * roomSide * EAR::Vector3(-roomSide, 0, 0)); // Room 1
    roomLocations.push_back(roomSide * EAR::Vector3(-roomSide, 0, 0)); // Room 2
    roomLocations.push_back(EAR::Vector3(0, 0, 0)); // Room 3
    roomLocations.push_back(roomSide * EAR::Vector3(roomSide, 0, 0)); // Room 4
    roomLocations.push_back(roomSide * EAR::Vector3(0, 0, roomSide)); // Room 5
    roomLocations.push_back(2 * roomSide * EAR::Vector3(0, 0, roomSide)); // Room 6

    roomIds.reserve(7);
    for (const auto& location : roomLocations) {
        roomIds.push_back(createRoom(sceneState, location, roomSize));
    }

    // state should contain the 7 rooms (boxes) in the geometry
    CHECK(sceneState.getAllGeometry().size() == 7);

    // Create portals between rooms
    auto portalPayload = std::make_shared<DummyPortalPayload>();
    createPortalBetweenRooms(sceneState, portalPayload, roomLocations, roomIds, 0, 3, materialId);
    createPortalBetweenRooms(sceneState, portalPayload, roomLocations, roomIds, 1, 2, materialId);
    createPortalBetweenRooms(sceneState, portalPayload, roomLocations, roomIds, 2, 3, materialId);
    createPortalBetweenRooms(sceneState, portalPayload, roomLocations, roomIds, 3, 4, materialId);
    createPortalBetweenRooms(sceneState, portalPayload, roomLocations, roomIds, 3, 5, materialId);
    createPortalBetweenRooms(sceneState, portalPayload, roomLocations, roomIds, 5, 6, materialId);

    // State should contain 7 rooms (boxes) + 6 portal extents (also boxes) in the geometry
    CHECK(sceneState.getAllGeometry().size() == 13);

    bitstream.addPortalPayload(portalPayload);

    const std::array<int, 5> roomsWithSources = { 0, 1, 2, 3, 6 };
    // Create reverb sound sources in rooms 0, 1, 2, 3 & 6 (rooms 4 & 5 don't have sources)
    // and add to renderList
    for (const auto roomIndex : roomsWithSources) {
        auto item = store.createItem("reverb_re_" + std::to_string(roomIndex));
        item->position = { roomLocations[roomIndex], EAR::Euler() };
        item->acousticEnvironmentId = roomIds[roomIndex];
        item->type = MPEGI::ItemType::Reverb;
        item->setStatus(MPEGI::ItemStatus::Active);
        item->setSignal(streamManager.createEmptyBuffer());
        item->getStateOrCreate<MPEGI::ReverbStage::ItemState>(false);
        renderList.addItemCopy(item);
    }

    CHECK_EQ(countNumberActiveItems(renderList), 5);

    // Create immersive reverb signals
    for (const auto roomIndex : roomsWithSources) {
        auto item = store.createItem("reverb_ht_dl_" + std::to_string(roomIndex));
        item->position = { roomLocations[roomIndex], EAR::Euler() };
        item->acousticEnvironmentId = roomIds[roomIndex];
        item->type = MPEGI::ItemType::Pannable;
        item->setStatus(MPEGI::ItemStatus::Active);
        item->setSignal(streamManager.createEmptyBuffer());
        item->getStateOrCreate<MPEGI::ReverbStage::ItemState>(true);
        renderList.addItemCopy(item);
    }

    CHECK_EQ(countNumberActiveItems(renderList), 10);
    const std::unordered_map<int, std::vector<int>> roomIdToPortals = { { 0, { 0 } }, { 1, { 1 } },
        { 2, { 1, 2 } }, { 3, { 0, 2, 3, 4 } }, { 4, { 3 } }, { 5, { 4, 5 } }, { 6, { 5 } } };

    // For coupling, also create some primary sources in those rooms
    for (const auto roomIndex : roomsWithSources) {
        auto item = store.createItem("primary_" + std::to_string(roomIndex));
        item->position = { roomLocations[roomIndex], EAR::Euler() };
        item->acousticEnvironmentId = roomIds[roomIndex];
        item->type = MPEGI::ItemType::Primary;
        item->setStatus(MPEGI::ItemStatus::Active);
        item->setSignal(streamManager.createEmptyBuffer());
        auto& state = item->getStateOrCreate<MPEGI::ReverbStage::ItemState>(true);
        for (int portalId : roomIdToPortals.at(roomIndex)) {
            state.portalContributions[portalId] = 0.2f;
        }
        renderList.addItemCopy(item);
    }

    CHECK_EQ(countNumberActiveItems(renderList), 15);

    // Initialize portal stage with all the geometry
    portalStage.initializeStage(bitstream, &store, &streamManager, &sceneState);
}

// Macro which allows value parameterized tests. See https://github.com/doctest/doctest/blob/master/doc/markdown/parameterized-tests.md
#define DOCTEST_VALUE_PARAMETERIZED_DATA(data, data_container)                                     \
    static size_t _doctest_subcase_idx = 0;                                                        \
    std::for_each(data_container.begin(), data_container.end(), [&](const auto& in) {              \
        DOCTEST_SUBCASE(                                                                           \
            (std::string(#data_container "[") + std::to_string(_doctest_subcase_idx++) + "]")      \
                .c_str())                                                                          \
        {                                                                                          \
            data = in;                                                                             \
        }                                                                                          \
    });                                                                                            \
    _doctest_subcase_idx = 0

TEST_CASE("Check correct number of portals are active when in the center of the rooms")
{
    // Key: room index. Value: number of active portal items. This is the equal to the number of portals to adjoining rooms + the number of other portals in those adjoining rooms. Such a portal is only counted if the room it's leading to contains a source.
    const std::map<size_t, size_t> expectedNumberOfActivePortals = { { 0, 2 }, { 1, 2 }, { 2, 3 },
        { 3, 4 }, { 4, 3 }, { 5, 4 }, { 6, 1 } };

    PortalStageTester tester;

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active portals are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::DirectionalReverb);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::DirectionalReverb)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, expectedNumberOfActivePortals.at(roomIndex));
}

TEST_CASE("No portals are active when outside all rooms")
{
    PortalStageTester tester;

    const auto& outside = EAR::Vector3(200.f, 200.f, 200.f);
    tester.renderList.listener.position.location = outside;
    tester.renderList.listener.acousticEnvironmentId = "MiddleOfNowhere";

    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::DirectionalReverb);

    tester.portalStage.update(tester.renderList);

    CHECK_EQ(countNumberActiveItems(tester.renderList, MPEGI::ItemType::DirectionalReverb),
        previousNumberOfRenderItems);
}

TEST_CASE("Test that correct portal reverb meshes are active when inside all rooms")
{
    // Key: room index. Value: portal ids of portals which should be active.
    const std::map<size_t, std::vector<int>> expectedPortalIdsOfActivePortals = { { 0, { 0, 2 } },
        { 1, { 1, 2 } }, { 2, { 0, 1, 2 } }, { 3, { 0, 1, 2, 5 } }, { 4, { 0, 2, 3 } },
        { 5, { 0, 2, 4, 5 } }, { 6, { 4 } } };

    PortalStageTester tester;

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    tester.portalStage.update(tester.renderList);

    for (const auto& item : tester.renderList.items) {
        if (item->type == MPEGI::ItemType::DirectionalReverb
            && item->status == MPEGI::ItemStatus::Active) {

            // Check that the Portal is located at one of the portals which is known to be active.
            auto isCorrectPortal = [&](int portalId) {
                return item->position.location == tester.portalLocations[portalId];
            };

            CAPTURE(roomIndex);
            CHECK(std::any_of(expectedPortalIdsOfActivePortals.at(roomIndex).begin(),
                expectedPortalIdsOfActivePortals.at(roomIndex).end(), isCorrectPortal));
        }
    }
}

TEST_CASE("Portal crossfade check")
{
    PortalStageTester testerForLocations; // Used only to extract some positions for test setup.

    // Move listener in small steps from center of room 3 to the center of the portal between rooms 0 and 3.
    const EAR::Vector3 begin = testerForLocations.roomLocations[3];
    const EAR::Vector3 end = testerForLocations.portalLocations[0];

    const int numberOfSteps = 100;

    const EAR::Vector3 delta = (end - begin) * (1 / (float)(numberOfSteps));

    float prevReverbGain = 1.f;
    float prevPortalGain = 0.f;

    for (int i = 0; i <= numberOfSteps; ++i) {
        PortalStageTester tester;
        tester.renderList.listener.position.location = begin + i * delta;
        tester.renderList.listener.acousticEnvironmentId = tester.roomIds[3];

        tester.portalStage.update(tester.renderList);

        const auto& reverbItem = tester.renderList.items[8]; // Immersive reverb signal of room 3
        const auto& reverbExtentItem =
            tester.renderList
                .items[19]; // RenderItem corresponding to the portal between rooms 0 and 3.

        if (i == 0) {
            CHECK(reverbExtentItem->status
                != MPEGI::ItemStatus::Active); // Portal should start inactive
            CHECK(reverbItem->status == MPEGI::ItemStatus::Active); // Reverb should start active
            CHECK(reverbItem->gain == 1.0f);
        }

        if (i == numberOfSteps) {
            CHECK(
                reverbExtentItem->status == MPEGI::ItemStatus::Active); // Portal should end active
            // NOTE: We don't check that reverb item is inactive: that is the responsibility of the ReverbStage, which is not active in this test

            // Portal should have high gain at end
            CHECK(reverbExtentItem->gain == doctest::Approx(1.f).epsilon(0.01f));

            // Reverb should have low gain
            CHECK(reverbItem->gain == doctest::Approx(0.f).epsilon(0.01f));
        }

        // ReverbGain should decrease on each step
        CHECK(prevReverbGain >= reverbItem->gain);
        prevReverbGain = reverbItem->gain;
        // PortalGain should increase (when it is active)
        if (reverbExtentItem->status == MPEGI::ItemStatus::Active) {
            CAPTURE(i);
            CHECK(prevPortalGain <= reverbExtentItem->gain);
            prevPortalGain = reverbExtentItem->gain;
        }

        // For items which are inside the portal box, reverb extent gain should be 1.
        if (tester.portalBoxes[0]->isInside(tester.renderList.listener.position.location)) {
            CAPTURE(i);
            CHECK(reverbExtentItem->gain == doctest::Approx(1.0f).epsilon(0.01f));
        }
    }
}

TEST_CASE("Check no portals are active when transmission coeffs are very low")
{
    PortalStageTester tester(true, 1e-8f);

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active portals are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::DirectionalReverb);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::DirectionalReverb)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, 0);
}

TEST_CASE("Check no portals are active in adjoining rooms, when transmission coeffs of "
          "intermediate portal is very low")
{
    // In this test, we put the listener in room 1 and then set the portal material of the only portal to this room to a low transmission coeff.
    // This should turn off all portal reverb meshes
    PortalStageTester tester;

    // Add a new material to scene state with low transmission coeff
    const std::string lowTMatId = tester.sceneState.addMaterial(std::pow(10.f, -8.f), 0.4f);

    auto portalPayload = std::dynamic_pointer_cast<const DummyPortalPayload>(
        tester.bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PayloadType::PLD_PORTAL));

    for (auto& portalPtr : portalPayload->getPortals()) {
        if (portalPtr->portalId == 0) {
            // Set material of portal between room 0 and 3 so it has low transmission coeff.
            Update materialUpdate;
            materialUpdate.addModification(
                portalPtr->scenePortal->getName(), "material", lowTMatId, false);

            tester.sceneState.modify(materialUpdate);
        }
    }

    const int roomIndex = 0;

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active portals are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, ItemType::DirectionalReverb);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, ItemType::DirectionalReverb)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, 0);
}

TEST_CASE("Check portals are inactive when combined transmission of 2 portals is too low")
{
    // In this test, we put the listener in room 1 and then set the portal material of portals 0 and 2 to low values.
    // This value is chosen so individually each portal is above the threshold, but when combined they are below the threshold.
    // Therefore, portal 0 should be active, while portal 2 should be inactive (because the sound from portal 2 also has to pass through portal 0 to reach the listener).
    PortalStageTester tester;

    // Add a new material to scene state with low transmission coeff
    const std::string lowTMatId = tester.sceneState.addMaterial(std::pow(10.f, -4.f), 0.4f);

    auto portalPayload = std::dynamic_pointer_cast<const DummyPortalPayload>(
        tester.bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PayloadType::PLD_PORTAL));

    for (auto& portalPtr : portalPayload->getPortals()) {
        if (portalPtr->portalId == 0 || portalPtr->portalId == 2) {
            Update materialUpdate;
            materialUpdate.addModification(
                portalPtr->scenePortal->getName(), "material", lowTMatId, false);

            tester.sceneState.modify(materialUpdate);
        }
    }

    const int roomIndex = 0;

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active portals are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, ItemType::DirectionalReverb);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, ItemType::DirectionalReverb)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, 1);
}

TEST_CASE("Check coupling produces the right amount of active render items")
{
    const std::map<size_t, size_t> expectedNumberOfActivePortals = { { 0, 2 }, { 1, 2 }, { 2, 3 },
        { 3, 4 }, { 4, 3 }, { 5, 4 }, { 6, 1 } };

    PortalStageTester tester;

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active primary items are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, expectedNumberOfActivePortals.at(roomIndex));
}

TEST_CASE("Check that coupling produces render items with EQ set to the portal material coupling "
          "coefficients")
{
    const float couplingCoeff = 0.4f;
    PortalStageTester tester(true, 0.3f, couplingCoeff);

    const int roomIndex = 3;

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active primary items are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems = tester.renderList.items.size();

    tester.portalStage.update(tester.renderList);

    for (size_t RIIndex = previousNumberOfRenderItems; RIIndex < tester.renderList.items.size();
         ++RIIndex) {
        const auto& item = tester.renderList.items[RIIndex];

        if (item->status != MPEGI::ItemStatus::Inactive && item->type == MPEGI::ItemType::Primary) {
            // In this test setup, the material eq is constant for all frequencies
            for (const auto& eqVal : item->eq) {
                CHECK(eqVal == doctest::Approx(std::sqrt(couplingCoeff)).epsilon(0.001f));
            }
        }
    }
}

TEST_CASE("No coupling is active when outside all rooms")
{
    PortalStageTester tester;

    const auto& outside = EAR::Vector3(200.f, 200.f, 200.f);
    tester.renderList.listener.position.location = outside;
    tester.renderList.listener.acousticEnvironmentId = "MiddleOfNowhere";

    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary);

    tester.portalStage.update(tester.renderList);

    CHECK_EQ(countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary),
        previousNumberOfRenderItems);
}

TEST_CASE("Test that coupling uses the portal meshes")
{
    // Key: room index. Value: portal ids of portals which should be active.
    const std::map<size_t, std::vector<int>> expectedPortalIdsOfActivePortals = { { 0, { 0, 2 } },
        { 1, { 1, 2 } }, { 2, { 0, 1, 2 } }, { 3, { 0, 1, 2, 5 } }, { 4, { 0, 2, 3 } },
        { 5, { 0, 2, 4, 5 } }, { 6, { 4 } } };

    PortalStageTester tester;

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    tester.portalStage.update(tester.renderList);

    for (const auto& item : tester.renderList.items) {
        if (item->type == MPEGI::ItemType::Primary && item->status == MPEGI::ItemStatus::Active
            && (item->extent != nullptr)) {

            // Check that the coupling RI is located at one of the portals which is known to be active, and that it uses the portal mesh.
            auto isCorrectPortal = [&](int portalId) {
                return (item->position.location == tester.portalLocations[portalId])
                    && (item->extent->getMesh()->getVertices(EAR::Mesh::CSpace::Global)
                        == tester.portalBoxes[portalId]->getMesh()->getVertices(
                            EAR::Mesh::CSpace::Global));
            };

            CAPTURE(roomIndex);
            CHECK(std::any_of(expectedPortalIdsOfActivePortals.at(roomIndex).begin(),
                expectedPortalIdsOfActivePortals.at(roomIndex).end(), isCorrectPortal));
        }
    }
}

TEST_CASE("Coupling does not process portals without a material")
{
    // Create a tester that does not set any material for the portals
    PortalStageTester tester(false);

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active primary items are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, 0);
}

TEST_CASE("Coupling does not process portals with transmission coefficient below threshold")
{
    // Stage is only used to get the threshold value. PortalStage tester constructs it's own PortalStage for actual testing.
    float transmissionThreshold;
    float couplingThreshold;
    {
        PortalStage stage;
        transmissionThreshold = stage.getTransmissionThreshold();
        couplingThreshold = stage.getCouplingThreshold();
    }

    // Construct a tester with transmission coeff below the threshold, and coupling coeff above it
    PortalStageTester tester(true, transmissionThreshold / 2.f, couplingThreshold * 2.f);

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active primary items are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, 0);
}

TEST_CASE("Coupling does not process portals with coupling coefficient below threshold")
{
    // Stage is only used to get the threshold value. PortalStage tester constructs it's own PortalStage for actual testing.
    float transmissionThreshold;
    float couplingThreshold;
    {
        PortalStage stage;
        transmissionThreshold = stage.getTransmissionThreshold();
        couplingThreshold = stage.getCouplingThreshold();
    }

    // Construct a tester with transmission coeff above the threshold, and coupling coeff below it
    PortalStageTester tester(true, transmissionThreshold * 2.f, couplingThreshold / 2.f);

    int roomIndex;
    const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

    DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
        roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

    // Place listener in center of room
    const auto& centerOfRoom = tester.roomLocations[roomIndex];
    tester.renderList.listener.position.location = centerOfRoom;
    tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

    // Run the portalstage and check how many active primary items are added. Each active portal should add an active RenderItem to the RenderList.
    const size_t previousNumberOfRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary);

    tester.portalStage.update(tester.renderList);

    const size_t newlyAddedRenderItems =
        countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary)
        - previousNumberOfRenderItems;

    CAPTURE(roomIndex);
    CHECK_EQ(newlyAddedRenderItems, 0);
}

TEST_CASE("Coupling process function applies gain to single input")
{
    size_t bufferLength = 256;

    PortalStageTester tester;

    PortalStage::CouplingData coupling;
    coupling.portalId = 0;

    PortalStage::CouplingInput input;

    // Create input of all 1s.
    auto inputBuffer = std::make_shared<MPEGI::StreamBuffer>(1, bufferLength);
    inputBuffer->getFrame().fill(1.f);

    input.audioIn = inputBuffer;
    input.gain = 0.6f;
    input.portalContribution = 0.4f;

    coupling.inputs.push_back(input);

    EAR::RealFrame outputBuffer(1, bufferLength);
    outputBuffer.zero();

    coupling.output = &outputBuffer;

    PortalStage::DspData dspData;
    dspData.couplingData.push_back(coupling);

    MPEGI::RenderStage::DspContext dspContext{ dspData };

    tester.portalStage.processBlock(dspContext);

    const float* out = outputBuffer.ch(0).data();
    for (size_t i = 0; i < bufferLength; ++i) {
        CHECK(out[i] == doctest::Approx(0.6f * std::sqrt(0.4)).epsilon(0.001f));
    }
}

TEST_CASE("Coupling process function applies gain to two inputs")
{
    size_t bufferLength = 256;

    PortalStageTester tester;

    PortalStage::CouplingData coupling;
    coupling.portalId = 0;

    PortalStage::CouplingInput input;

    // Create input of all 1s.
    auto inputBuffer = std::make_shared<MPEGI::StreamBuffer>(1, bufferLength);
    inputBuffer->getFrame().fill(1.f);

    input.audioIn = inputBuffer;
    input.gain = 0.6f;
    input.portalContribution = 0.4f;

    coupling.inputs.push_back(input);

    // Second input
    auto inputBuffer2 = std::make_shared<MPEGI::StreamBuffer>(1, bufferLength);
    inputBuffer2->getFrame().fill(0.5f);

    PortalStage::CouplingInput input2;
    input2.audioIn = inputBuffer2;
    input2.gain = 0.8f;
    input2.portalContribution = 0.7f;

    coupling.inputs.push_back(input2);

    EAR::RealFrame outputBuffer(1, bufferLength);
    outputBuffer.zero();

    coupling.output = &outputBuffer;

    PortalStage::DspData dspData;
    dspData.couplingData.push_back(coupling);

    MPEGI::RenderStage::DspContext dspContext{ dspData };

    tester.portalStage.processBlock(dspContext);

    const float* out = outputBuffer.ch(0).data();
    for (size_t i = 0; i < bufferLength; ++i) {
        // = 1 * 0.6 * sqrt(0.4) + 0.5 * 0.8 * sqrt(0.7)
        CHECK(out[i] == doctest::Approx(0.714137f).epsilon(0.001f));
    }
}

TEST_CASE("Check coupling produces no render items when disabled by config")
{
    // Code wrapped in subcase, to ensure Config::reset is called even when subcase throws
    SUBCASE("")
    {
        // Disable coupling
        Config::set("Portal_Enable_Coupling", EAR::Variant(false));

        PortalStageTester tester;

        int roomIndex;
        const std::vector<int> roomIndices = { 0, 1, 2, 3, 4, 5, 6 };

        DOCTEST_VALUE_PARAMETERIZED_DATA(roomIndex,
            roomIndices); // This causes doctest to run the test multiple times. Each time, roomIndex will have a different value from roomIndices.

        // Place listener in center of room
        const auto& centerOfRoom = tester.roomLocations[roomIndex];
        tester.renderList.listener.position.location = centerOfRoom;
        tester.renderList.listener.acousticEnvironmentId = tester.roomIds[roomIndex];

        // Run the portalstage and check how many active primary items are added. Each active portal should add an active RenderItem to the RenderList.
        const size_t previousNumberOfRenderItems =
            countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary);

        tester.portalStage.update(tester.renderList);

        const size_t newlyAddedRenderItems =
            countNumberActiveItems(tester.renderList, MPEGI::ItemType::Primary)
            - previousNumberOfRenderItems;

        CAPTURE(roomIndex);
        CHECK_EQ(newlyAddedRenderItems, 0);
    }
    // Reset config to prevent config influencing other tests (particularly when tests are run out of order)
    Config::reset();
}

} // namespace MPEGI
