/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2023.
***************************************************************************/

#include <doctest/doctest.h>

#include "../Portal.h"
#include <MPEGI/Bitstream/EIFPayload.h>

#include <EAR/Box.h>
#include <EAR/Exception.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Renderer/RenderItem.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Stream/StreamManager.h>

#include <doctest/doctest.h>
#include <memory>
#include <mpegcore/datatypes_reverb.hpp>
#include <string>

// Setup Dummy classes needed for test

namespace MPEGI {

// Bitstream id number for default acoustic environment
static const unsigned int DEFAULT_AE_ID = 99;
static const std::string DEFAULT_AE_ID_NAME{ "default" };

// Bitstream id for an item that contributes to reverb from an external room
const unsigned int EXTERNAL_OBJ_SRC_ID = 13;

// StreamManager is abstract but it contains all the functionality needed for testing. Any
// dummy that compiles is fine for testing.
class DummyStreamManager : public StreamManager
{
public:
    DummyStreamManager(float fs, std::size_t B)
        : StreamManager(fs, B)
    {}

    std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const std::string& streamId, double t = 0.0) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const SceneState::AudioStream& stream) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    void removeAccessBuffer(StreamAccessBuffer* buf) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    void nextBlock() override { EAR_CHECK_REL(false, "Not expected to be called"); }

    std::size_t getNbrOfChannels(const std::string& id) const override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }
};

class DummyPortalPayload : public PortalPayload
{
public:
    DummyPortalPayload() { isExplicitPortalMode = true; };
    DummyPortalPayload& operator=(const DummyPortalPayload&) = delete;
    DummyPortalPayload(const DummyPortalPayload& other) = default;
    virtual ~DummyPortalPayload() {}
    void addPortal(std::unique_ptr<Portal>&& portal)
    {
        portalMap[portal->portalEnvironments[0]].insert(portal->portalId);
        portalMap[portal->portalEnvironments[1]].insert(portal->portalId);
        portals.push_back(std::move(portal));
    };
};

class DummyBitstream : public Bitstream
{
public:
    DummyBitstream() = default;
    DummyBitstream(std::shared_ptr<DummyPortalPayload> pp);
    DummyBitstream& operator=(const DummyBitstream&) = delete;
    DummyBitstream(const DummyBitstream&) = delete;
    virtual ~DummyBitstream() {}
    void addPortalPayload(std::shared_ptr<DummyPortalPayload> pp);
    virtual std::vector<std::shared_ptr<const Bitstream::Payload>> getAllPayloadsOfType(
        PayloadType type) const override;
    virtual std::shared_ptr<const Bitstream::Payload> getOnlyPayloadOfType(
        PayloadType type) const override;
    void clearPayloads();

private:
    std::vector<std::shared_ptr<const Bitstream::Payload>> portalPayloads;
};

class DummySceneState : public SceneState
{
public:
    DummySceneState(bool hasDefaultAe);
    DummySceneState& operator=(const DummySceneState&) = delete;
    DummySceneState(const DummySceneState&) = delete;
    virtual ~DummySceneState() {}
    virtual const std::vector<std::unique_ptr<ReverbEnv>>& getReverbEnvs() const override;
    // Adds a material to the scene state. Returns the id of the created material.
    std::string addMaterial(float transmissionCoeff, float couplingCoeff);
    void addAcousticEnvironment(std::optional<std::string> aeId, const EAR::Geometry* enclosure);
    void createBox2(const EIFPayload::BoxDesc& box);
    void createPortal2(MPEGI::Portal* portal);

private:
    std::vector<std::unique_ptr<ReverbEnv>> dummyReverbEnvs;
    int materialBsId = 0;
};

// PortalStageTester sets up all required (Dummy) state for a PortalStage of the required geometry
class PortalStageTester
{
public:
    // Create tester. Arguments determine wether a material will be set for the portals in the scene, and the coefficients of this material.
    PortalStageTester(
        bool createMaterials = true, float transmissionCoeff = 0.3f, float couplingCoeff = 0.4f);

    const float SAMPLING_FREQ{ 48000.F };
    const std::size_t B{ 256 };
    const std::size_t TEST_PROCESS_ITERATIONS{ 16 };

    DummySceneState sceneState;
    DummyBitstream bitstream;
    PortalStage portalStage;
    RenderList renderList{ 0 };

    std::vector<const EAR::Geometry*> rooms;
    std::vector<EAR::Vector3> roomLocations;
    std::vector<std::string> roomIds;

    std::vector<const EAR::Geometry*> portalBoxes; // Box of portal itself, not the reverb extent
    std::vector<EAR::Vector3> portalLocations;

private:
    // Creates a room and adds it to the scene state. Returns reverb ID of newly created room.
    std::string createRoom(
        DummySceneState& state, const EAR::Vector3 Position, const EAR::Vector3 size);

    // Creates a portal halfway between 2 rooms with index roomIndex1 and roomIndex2.
    void createPortalBetweenRooms(DummySceneState& state,
        std::shared_ptr<DummyPortalPayload> portalPayload,
        const std::vector<EAR::Vector3>& roomLocations, const std::vector<std::string>& roomIds,
        size_t roomIndex1, size_t roomIndex2, const std::string& materialId);

    RenderItemStore store;
    DummyStreamManager streamManager{ SAMPLING_FREQ, B };

    const bool hasDefaultAE = false;

    int boxIdCounter = 0;
    int envIdCounter = 0;
    int portalIdCounter = 0;
};

} // namespace MPEGI
