/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DesignIIR.h"
#include "quasiNewton.h"
#include <cstring>
#include <sys/timeb.h>

void minimax(Minimax* pIn_mm, Minimax* pOut_mm)
{

    /* ******************************************
    **   force is NULL for normal operations   **
    **   in UT force points to output arrays   **
    ** *************************************** */
    //  Minimax based on the Linf norm. (least pth minimax)
    //
    //    Inputs: tgtMagn - desired magnitude spectrum
    //             tgtFc - normalized frequency: 2f/fs
    //             iirOrder - order of the desired filter
    //            x - input inital parameter vector [length 2*N+1]
    //                  Filter coefficients expressed in a single vector.
    //                  x = [b01, b11, a01, a11, ...b0j, b1j, a0j, a1j H0]   [length 2*N+1]
    //             MIN_ERROR_DIFF - min. change in error for stopping criterion
    //             MAX_ITERATIONS - maximum number of iterations if stopping
    //             criterion not reached.
    //
    //     At each iteration attempts to minimize the maximum error across
    //     frequency.

    /*==========================================================================
          MATLAB CODE attributed to Patrick Kechichian
	      MATLAB module: designIIR2.m
	      C Transporter: Danilo Dara
	      Unit Tests: Danilo Dara
	      Last release: 17 May 2021
        ==========================================================================*/

    // pForce_mm = false;

    *pOut_mm = *pIn_mm;
    int32_t i;
    pOut_mm->iterCount.minimax++;

    pOut_mm->success = false;
    // init

    // Init inital error

    double prevIterErr;
    prevIterErr = 1 << 16;
    pOut_mm->err = 0;

    // Get normalized angular frquency
    double w[NR_FILTER_DESIGN_FREQ];
    for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        w[i] = pOut_mm->tgtFc[i] * M_PI;

    int32_t p = 2;
    int32_t m = 2;

    int32_t J = pOut_mm->iirOrder / 2;

    int32_t MAX_P = 64;

    int32_t fErr = -1;

    // Minimization
    QNewt* qnInterface = new QNewt();
    // prepare atomic parameters
    qnInterface->IIROrder = pOut_mm->iirOrder;
    qnInterface->p = p;
    qnInterface->MAX_ITERATIONS = pOut_mm->MAX_ITERATIONS;
    qnInterface->formatConfig = pOut_mm->formatConfig;
    qnInterface->iterCount = pOut_mm->iterCount;
    // prepare vectorial parameters
    int32_t qnVect_Size = (qnInterface->IIROrder * 2) + 1;
    std::memcpy(qnInterface->x0, pOut_mm->x, qnVect_Size * sizeof(float));
    std::memcpy(qnInterface->tgtMagn, pOut_mm->tgtMagn, NR_FILTER_DESIGN_FREQ * sizeof(float));
    std::memcpy(qnInterface->w, w, NR_FILTER_DESIGN_FREQ * sizeof(double));

    bool undershootIgnore[NR_FILTER_DESIGN_FREQ];
    for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        // Dirty coding, but we don't know fs here and CfP only works with 48 kHz.
        if (pOut_mm->tgtFc[i] > 15000.0f / 24000.0f || pOut_mm->tgtFc[i] < 50.0f / 24000.0f)
            undershootIgnore[i] = 1;
        else
            undershootIgnore[i] = 0;
    }

    for (int32_t iterIdx = 0; iterIdx != pOut_mm->MAX_ITERATIONS;
         iterIdx++) { // can get a "break" inside

        // printf("minimax iteration %d error band (fErr) %d\n", iterIdx, fErr);

        // Quasi-Newton finds local minima of error function
        qnInterface->p = p;

        quasi_Newton(qnInterface, qnInterface);
        std::memcpy(qnInterface->x0, qnInterface->x, qnVect_Size * sizeof(float));
        std::memcpy(pOut_mm->x, qnInterface->x, qnVect_Size * sizeof(float));

        // Calculate error properties

        float Magnitude[NR_FILTER_DESIGN_FREQ];
        get_magnitude_response(qnInterface->x, qnInterface->w, qnInterface->IIROrder, Magnitude);

        double e[NR_FILTER_DESIGN_FREQ];
        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            e[i] = Magnitude[i] - qnInterface->tgtMagn[i];

            if (undershootIgnore[i])
                e[i] = fmax(e[i], 0.0f);
        }

        // Normalized error power
        double errNorm[NR_FILTER_DESIGN_FREQ];
        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            errNorm[i] = pow(e[i] / pOut_mm->MIN_ERROR[i], 2);

        // Find worst band and use that as the error
        // MATLAB:      [~, fErr] = max(errNorm);
        double errNormWorstBand = -1; // any negative

        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            if (errNormWorstBand < errNorm[i]) {
                errNormWorstBand = errNorm[i];
                fErr = i;
            }

        pOut_mm->err = fabs(e[fErr]);
        // err is the maximum deviation

        //  Error power ratio vs maximum error within acceptable bounds for
        //  worst band.

        // Total error power ratio vs maximum error within acceptable bounds
        float num = 0;
        float den = 0;
        float errNormTotal;
        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            num += e[i] * e[i];
            den += pOut_mm->MIN_ERROR[i] * pOut_mm->MIN_ERROR[i];
        }
        errNormTotal = num / den;

        // 6 - Stop, when we..
        if ((pOut_mm->err < pOut_mm->MIN_ERROR[fErr]) || // ..reached the error threshold
            (errNormWorstBand < 1.2 && errNormTotal < 0.02)
            || // ..don't overrun the error threshold too much with the rest well within bounds
            (fabs(pOut_mm->err - prevIterErr) < pOut_mm->MIN_ERROR_DIFF)
            || // ..reached conversion target
            (iterIdx >= pOut_mm->MAX_ITERATIONS) // ..reached the maximum nr of iterations
        ) { // #6 Stop - do not iterate again
            // When we end because the error is acceptable, signal that to
            // calling function. Otherwise further iterations with different
            // start vectors are necessary.
            if ((pOut_mm->err < pOut_mm->MIN_ERROR[fErr])
                || (errNormWorstBand < 1.2 && errNormTotal < 0.02))
                pOut_mm->success = true;

            IIRFilterCoefficients* output = &pOut_mm->filterCoefficients;
            if (output->minimaxOutputType == IIRFilterType::SOSs) {
                output->filters.sosSequence.a[0].resize(J);
                output->filters.sosSequence.a[1].resize(J);
                output->filters.sosSequence.b[0].resize(J);
                output->filters.sosSequence.b[1].resize(J);
                output->filters.sosSequence.stageF.resize(J);
                output->filters.sosSequence.type = mpegcore::GraphicEqParameters::Type::cascade;
            }

            // Convert second order section coefficients into polynomial filter
            // coefficients.
            float h0;
            h0 = pOut_mm->x[(2 * pOut_mm->iirOrder)];
            for (int32_t sosIdx = 0; sosIdx != J; sosIdx++) {
                // Skip over second stage parameters
                int idx0;
                idx0 = sosIdx << 2;

                float bSos[3];
                bSos[0] = 1;
                bSos[1] = pOut_mm->x[idx0];
                bSos[2] = pOut_mm->x[idx0 + 1];

                // Check if any poles are outside the unit circle..
                // MATLAB:      rootage = x(idx0+3)*x(idx0+3)-4 * x(idx0+2);
                float p1r, p1i;
                float rootage, K;
                bool bComplex = false;
                rootage = pOut_mm->x[idx0 + 2] * pOut_mm->x[idx0 + 2] - 4 * pOut_mm->x[idx0 + 3];
                // MATLAB:      p1 = (-x(idx0+3) + sqrt(rootage))/2;
                if (rootage < 0) {
                    rootage = -rootage;
                    bComplex = true;
                }
                K = (float)sqrt((double)rootage);
                if (bComplex) {
                    p1r = -pOut_mm->x[idx0 + 2] / 2;
                    p1i = K / 2;
                } else {
                    p1r = (-pOut_mm->x[idx0 + 2] + K) / 2;
                    p1i = 0;
                }
                // MATLAB:      p2 = (-x(idx0+3) - sqrt(rootage))/2;
                float p2r, p2i;
                if (bComplex) {
                    p2r = -pOut_mm->x[idx0 + 2] / 2;
                    p2i = -K / 2;
                } else {
                    p2r = (-pOut_mm->x[idx0 + 2] - K) / 2;
                    p2i = 0;
                }

                // MATLAB:      if abs(p1) > 1
                if ((p1r * p1r) + (p1i * p1i) > 1) { // move back inside the circle
                    // MATLAB:          p1 = 1/p1;
                    float tmpr, tmpi;
                    tmpr = p1r / (p1r * p1r + p1i * p1i);
                    tmpi = p1i / (p1r * p1r + p1i * p1i);
                    p1r = tmpr;
                    p1i = tmpi;
                    // MATLAB:          h0 = h0*p1;
                    if (!bComplex) // When poles are complex conjugates, do update of h0 in second pole conversion
                        h0 *= p1r;
                }

                // MATLAB:      if abs(p2) > 1
                if ((p2r * p2r) + (p2i * p2i) > 1) { // move back inside the circle
                    // MATLAB:          p2 = 1/p2;
                    float tmpr, tmpi;
                    tmpr = p2r / (p2r * p2r + p2i * p2i);
                    tmpi = p2i / (p2r * p2r + p2i * p2i);
                    p2r = tmpr;
                    p2i = tmpi;
                    // MATLAB:          h0 = h0*p2;
                    if (bComplex) // When poles are complex conjugates, effect of both poles is applied here
                        h0 *= (p2r * p2r
                            + p2i * p2i); // Effect of multiplying complex conjugate poles with h0
                    else
                        h0 *= p2r;
                }

                float aSos[3];
                aSos[0] = 1;
                aSos[1] = -(p1r + p2r);
                aSos[2] = (p1r * p2r) - (p1i * p2i);

                // h0 is real

                // Update x vector, too, in case we fixed some poles.
                // MATLAB:      x(2*iirOrder+1) = h0;
                pOut_mm->x[pOut_mm->iirOrder * 2] = h0;

                // MATLAB:      x(idx0+2)       = -(p1+p2);
                pOut_mm->x[idx0 + 2] = aSos[1];

                // MATLAB:      x(idx0+3)       = p1*p2;
                pOut_mm->x[idx0 + 3] = aSos[2];

                if (output->minimaxOutputType == IIRFilterType::SOSs) {
                    output->filters.sosSequence.b[0][sosIdx] = bSos[1];
                    output->filters.sosSequence.b[1][sosIdx] = bSos[2];
                    output->filters.sosSequence.a[0][sosIdx] = aSos[1];
                    output->filters.sosSequence.a[1][sosIdx] = aSos[2];
                    output->filters.sosSequence.stageF[sosIdx] = 1.0f;
                }

            } // for sosIdx

            output->filters.sosSequence.F = h0;
            output->filters.sosSequence.levelDb = 0.0f;

            break;
        } // #6 - Stop if - stop
        else { // #6 - Stop if - don't stop

            p *= m;
            if (p > MAX_P)
                p = MAX_P;

        } // #6

        prevIterErr = pOut_mm->err;

    } // for iteration

    // store results
    pOut_mm->errFreqBand = fErr;

    delete qnInterface;
} // minimax

bool isStableFilter(IIRFilterCoefficients* filter, int32_t iirOrder)
{
    bool stable = true;

    float maxPoleAmpl = 0.995; // very close to the unit circle

    // Iterate over parameters
    for (int32_t sosIdx = 0; sosIdx != iirOrder / 2; sosIdx++) {

        // Get denominator coefficients.
        float a0 = filter->filters.sosSequence.a[0][sosIdx];
        float a1 = filter->filters.sosSequence.a[1][sosIdx];

        float discr = a0 * a0 - 4 * a1;
        if (discr < 0) {

            if ((float)sqrt((double)a1) > maxPoleAmpl) {
                stable = false;
            }
        } else {
            if ((fabs(0.5 * (-a0 - (float)sqrt((double)discr))) > maxPoleAmpl)
                || (fabs(0.5 * (-a0 + (float)sqrt((double)discr))) > maxPoleAmpl)) {
                // Real poles
                stable = false;
            }
        }

    } // iterate over parameters

    return (stable);
} // isStableFilterSP

void estimateIIR(EstimateIIR* pIn_eIIR, EstimateIIR* pOut_eIIR)
{

    //    Inputs:   EstimateIIR.Minimax.
    //                  x           Start vector of filter coefficients
    //                  tgtMagn     Target amplitude spectrum
    //                  tgtFc       Frequencies of target spectrum (normalized frequency: f / (fs/2))
    //                  iirOrder    Filter order
    //
    //    Loops through multiple random initializations to estimate the optimum
    //    set of IIR filter coefficients. Can be extended to also iterate over
    //    increasing filter order in increments of 2.
    //
    // ==========================================================================
    //
    //    Author: Patrick Kechichian
    //    email: patrick.kechichian@philips.com
    //    Last modified: May 9, 2020
    //
    // ==========================================================================
    timeb start, end;
    ftime(&start);

    *pOut_eIIR = *pIn_eIIR;

    float errOpt = 2e14; // in excess

    int32_t errOptFreqBand = 0;

    int32_t max_estimateIIR_iter = 64;

    pOut_eIIR->minimaxInterface.MAX_ITERATIONS = 64;

    pOut_eIIR->minimaxInterface.iterCount.estIIR++;

    int32_t nrIterations = 1;

    bool minimaxSuccess = 0;

    bool stableFilter = true;

    while ((((errOpt >= pOut_eIIR->minimaxInterface.MIN_ERROR[errOptFreqBand]) && (!minimaxSuccess))
               || !stableFilter)
        && (nrIterations <= max_estimateIIR_iter)) {
        if ((nrIterations > 1) && (nrIterations < 20))
            // For every iteration increase the error margins a bit so that there is a
            // better chance of finding a fit (fast).
            for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
                pOut_eIIR->minimaxInterface.MIN_ERROR[i] *= 1.05;

        // Filter coefficients expressed in a single vector.
        //   x = [b01, b11, a01, a11, ...b0j, b1j, a0j, a1j H0]   [length 2*N+1]
        //
        // Making N/2 second order sections normalized to have 4 non-unitary
        // coefficients and a common gain H0 for all concatenated sections.

        // random initialization as from Matlab is replaced with hardcoded injection of Matlab-generated values
        // for the sake of the unit test execution.
        // every first call to minimax() returns just the same filter
        // starting guess for T60 filter 1
        float rndX[LONG_QNVECT_SIZE] = { -0.203920067266352, 0.128787908879483, 0.079837810189545,
            0.099929196624988, -0.234180882464493, -0.215314119358636, -0.246411794226213,
            -0.172436052311266, -0.355835699346580, -0.334387138798732, 0.463930529067942,
            0.460226715285694, -0.311585344404065, -0.475693438370513, -0.295444453620049,
            0.199843614126557, 0.279514585555530,
            /* IIROrder == 8 */
            0.265907856480316, 0.018417987872943, -0.203199498423778, -0.312278771338748,
            -0.419258731235125, 0.238440296198970, 0.265907856480316, 0.018417987872943,
            /* IIROrder == 12 */
            -0.278006828910261, 0.370732306177376, -0.293280844660574, 0.418610907937922,
            -0.011588811205171, 0.111743862902646, 0.265907856480316, 0.018417987872943,
            -0.203199498423778, -0.312278771338748, -0.419258731235125, 0.238440296198970,
            -0.058690777104047, -0.341690132287349, 0.379937031201279, -0.225913538007775,
            -0.085764980918949, -0.203920067266352, 0.128787908879483, 0.079837810189545,
            0.099929196624988, -0.234180882464493, -0.215314119358636, -0.246411794226213,
            -0.172436052311266, -0.355835699346580, -0.334387138798732, 0.463930529067942,
            0.460226715285694, -0.311585344404065, -0.475693438370513, -0.295444453620049,
            0.199843614126557, 0.279514585555530, -0.477066907560919, 0.077662858129756 };
        /* IIROrder == 30 */
        if (nrIterations > 1) // let first iteration come from hardcoded values
        { // x = ApplyFormat(-0.5+rand(2*iirOrder+1,1), formatConfig);

            for (int32_t i = 0; i != LONG_QNVECT_SIZE; i++) {
                float rnd;
                rnd = DesignIIRRNG::rng(DesignIIRRNG::rngEngine);
                // rndX[i] = -0.5f + (rnd / RAND_MAX);
                rndX[i] = -0.5f + rnd;
            }
        }

        std::memcpy(pOut_eIIR->minimaxInterface.x, rndX, LONG_QNVECT_SIZE * sizeof(float));
        minimax(&pOut_eIIR->minimaxInterface, &pOut_eIIR->minimaxInterface);

        minimaxSuccess = pOut_eIIR->minimaxInterface.success;

        if (pOut_eIIR->minimaxInterface.err < errOpt) {
            errOpt = pOut_eIIR->minimaxInterface.err;

            errOptFreqBand = pOut_eIIR->minimaxInterface.errFreqBand;
        }

        printfAlways("estimateIIR() end of iteration %d -- Iteration error: %.4f / %.4f "
                     "(f=%d) -- Best error so far: %.4f\n",
            nrIterations, pOut_eIIR->minimaxInterface.err,
            pOut_eIIR->minimaxInterface.MIN_ERROR[errOptFreqBand],
            (int)((pIn_eIIR->minimaxInterface.tgtFc[errOptFreqBand] * 24000)), errOpt);

        stableFilter = isStableFilter(&(pOut_eIIR->minimaxInterface.filterCoefficients),
            pOut_eIIR->minimaxInterface.iirOrder);

        nrIterations++;

    } // while

    if (!stableFilter) {
        printfAlways("Filter design stopped without a stable filter.\n");
    }

    ftime(&end);
}

void designIIRFilter(EstimateIIR* pIn_IIR2, EstimateIIR* pOut_IIR2)
{

    //  wrapper design IIR method
    //
    //    Inputs:   EstimateIIR.Minimax.
    //                  x           Start vector of filter coefficients
    //                  tgtMagn     Target amplitude spectrum
    //                  tgtFc       Frequencies of target spectrum (normalized frequency: f / (fs/2))
    //                  iirOrder    Filter order
    //
    //
    // ==========================================================================
    //
    //    Author: Patrick Kechichian - revisited by Jeroen Koppens
    //    Last modified: Oct, 2021
    //
    // ==========================================================================

    *pOut_IIR2 = *pIn_IIR2;

    std::memset(&pOut_IIR2->minimaxInterface.iterCount, 0, sizeof(DbugSupport));

    pOut_IIR2->minimaxInterface.MIN_ERROR_DIFF = 1e-4;

    pOut_IIR2->minimaxInterface.MAX_ITERATIONS = 64;

    estimateIIR(pIn_IIR2, pOut_IIR2);
}
