/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include "mpegcore/datatypes_reverb.hpp"
#include <random>

#define MAX_FILTERORDER 30
#define LONG_QNVECT_SIZE (MAX_FILTERORDER * 2 + 1)
// # of coefficients useded in QN context: (2*IIROrder) + 1 w/ IIROrder = 12
#define MAX_COEFF_SIZE (MAX_FILTERORDER + 1)
// # of entries from fgrid vector after resampling and interpolation
#define NR_FILTER_DESIGN_FREQ 156

#define DEBUG_RUN 0
#if (DEBUG_RUN == 0)
// low level debugging messages
// errors and warnings
#define printfAlways // printf
#else
#define printfAlways printf
#endif

enum class IIRFilterType
{
    Polynomial = 0,
    SOSs
};

struct FormatConfig
{
    int32_t d2sMask; // bit mask
    int32_t d2sRoundMask; // bit mask
};

struct DbugSupport
{
    int32_t filter;
    int32_t estIIR;
    int32_t minimax;
    int32_t quasiNewton;
    int32_t qnFor;
    int32_t qnWhile;
    // all the arrays here are truely variable size!!!
    // size depends on the speed of convergence.
    double xList[130][LONG_QNVECT_SIZE];
    double fkList[130];
    double akList[130];
    double qnEndList[130];
    double errInMinimaxList;
    double errRatioTotal;
    double errRatioBand;
    double deltakList[130][LONG_QNVECT_SIZE];
    double gkList[63][LONG_QNVECT_SIZE];
};

struct IIRFilterCoefficients
{
    union U
    {
        struct PolynomialCoefs
        {
            float bOut[MAX_COEFF_SIZE];
            float aOut[MAX_COEFF_SIZE];
        } polynomial;

        mpegcore::GraphicEqParameters sosSequence;
        U() {}
        ~U() {}
    } filters;

    IIRFilterType minimaxOutputType;

    IIRFilterCoefficients(IIRFilterType type)
        : minimaxOutputType(type)
    {
        switch (type) {
        case IIRFilterType::Polynomial: {
            for (int i = 0; i < MAX_COEFF_SIZE; ++i) {
                filters.polynomial.aOut[i] = 0;
                filters.polynomial.bOut[i] = 0;
            }
        } break;
        case IIRFilterType::SOSs: {
            new (&filters.sosSequence) mpegcore::GraphicEqParameters();
        } break;
        default:
            break;
        }
    }
    IIRFilterCoefficients& operator=(const IIRFilterCoefficients& other)
    {
        minimaxOutputType = other.minimaxOutputType;
        switch (minimaxOutputType) {
        case IIRFilterType::Polynomial: {
            filters.polynomial = other.filters.polynomial;
        } break;
        case IIRFilterType::SOSs: {
            filters.sosSequence = other.filters.sosSequence;
        } break;
        default:
            break;
        }
        return *this;
    }
};

struct Minimax
{
    Minimax(IIRFilterType type)
        : filterCoefficients(type)
    {}
    // Input parms
    float x[LONG_QNVECT_SIZE];
    float tgtMagn[NR_FILTER_DESIGN_FREQ];
    double tgtFc[NR_FILTER_DESIGN_FREQ]; // double is mandatory, here.
    int32_t iirOrder{};
    int32_t MAX_ITERATIONS{};
    float MIN_ERROR[NR_FILTER_DESIGN_FREQ];
    float MIN_ERROR_DIFF{};
    FormatConfig formatConfig{};

    // output results
    bool success{};
    float err{};
    int32_t errFreqBand{};
    IIRFilterCoefficients filterCoefficients;

    // itercount is basically a debug support structure for Matlab. Does not have any algorithmic impact
    // Yet, we like to check that the iteration process is same from the two sides.
    DbugSupport iterCount{};
};

struct EstimateIIR
{
    EstimateIIR(IIRFilterType type)
        : minimaxInterface(type)
    {}
    struct Minimax minimaxInterface;
};

struct DesignIIRRNG
{
    static void ResetRng() { rngEngine.seed(2); }
    inline static std::default_random_engine rngEngine{ 2 };
    inline static std::uniform_real_distribution<float> rng{ 0, 1 };
};

void designIIRFilter(EstimateIIR* pIn_IIR2, EstimateIIR* pOut_IIR2);
