/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Reverb.hpp"

#include <algorithm>
#include <functional>
#include <memory>
#include <mpegcore/frontend/GraphicEqFilter.hpp>
#include <string>

namespace MPEGI {

Reverb::Reverb(mpegcore::ReverberatorParameters const& reverberatorParameters)
    : reverberatorParameters{ reverberatorParameters }
{
    using namespace mpegcore::frontend;

    if (auto const& eqFltParams = reverberatorParameters.eqFilterParams) {
        eqFilter = makeGraphicEqFilter(eqFltParams.value());
    }

    outputPositions.reserve(reverberatorParameters.delayLines.size());

    for (unsigned int i = 0; i < reverberatorParameters.delayLines.size(); i++) {
        EAR::Position pos{};
        auto const& delayLine = reverberatorParameters.delayLines[i];
        pos.location = EAR::polarToCartesian({ delayLine.azimuthDeg, delayLine.elevationDeg, 1.F });
        pos.cSpace = EAR::CSpace::LCS;
        outputPositions.push_back(pos);
    }
}

void Reverb::filterEQ() noexcept
{
    if (!eqFilter) {
        return;
    }

    // Sum all the input audio sample by sample to mixed predelay input
    std::fill_n(begin(mixedPredelayInput), B, 0.f);
    for (auto const& predelay : predelays) {
        std::transform(mixedPredelayInput.begin(), mixedPredelayInput.end(),
            predelay.predelayedInputAudio.begin(), mixedPredelayInput.begin(), std::plus());
    }

    eqFilter->filter(mixedPredelayInput);
}

void Reverb::processBlock() noexcept
{
    for (auto& predelay : predelays) {
        predelay.predelayBuffer->write(predelay.inputAudio);
        predelay.predelayBuffer->read(predelay.predelayedInputAudio);
    }
    filterEQ();
    fdn->process(static_cast<uint32_t>(B));
}

void Reverb::initialize(std::size_t B)
{
    using namespace mpegcore::frontend;
    this->B = B;

    if (!fdn) {
        fdn = FdnReverb(reverberatorParameters);
        outputAudio.init(outputPositions.size(), B);
        fdn->setOutputBuffer(outputAudio);
        mixedPredelayInput.resize(B);
        fdn->setInputBuffer(mixedPredelayInput);
        // Predelay for items enclosed in the same environment this reverberator models.
        // Items from connected environments may use different predelay parameters.
        initPredelayLine(reverberatorParameters.preDelayAdjustedInSamples.first);
    }
}

void Reverb::initPredelayLine(std::size_t length)
{
    auto it = std::find_if(predelays.begin(), predelays.end(),
        [length](PredelayData& predelayData) { return predelayData.predelayInSamples == length; });
    if (it == predelays.end()) {
        predelays.emplace_back(PredelayData{ length, mpegcore::AudioBuffer(B, 0),
            mpegcore::AudioBuffer(B, 0), std::make_unique<mpegcore::RingBuffer>(B, length) });
    }
}

mpegcore::AudioBuffer& Reverb::getInputBuffer(std::size_t length)
{
    auto it = std::find_if(predelays.begin(), predelays.end(),
        [length](PredelayData& predelayData) { return predelayData.predelayInSamples == length; });
    EAR_CHECK_REL(it != predelays.end(), "No predelay line for len=" + std::to_string(length));
    return it->inputAudio;
}

size_t Reverb::numberOfOutputs() const
{
    return outputPositions.size();
}

} // namespace MPEGI
