/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ReverbConfigCalc.h"
#include "DesignIIR.h"
#include "mpegcore/datatypes_reverb.hpp"
#include <MPEGI/Bitstream/Globals.h>

#include <cassert>

#include <algorithm>
#include <cmath>
#include <utility>

#include <mpegcore/primes.hpp>

void fillBaseConfig(SetBaseConfigStruct* baseConfig, unsigned int nrFbLoops, float predelayFactor,
    float delayRatio, float delayMinMs, const float fs)
{
    baseConfig->fs = fs;
    baseConfig->renderPredelay = .05f * fs;

    // intentionally constant parameters
    baseConfig->nrFbLoops = nrFbLoops;
    baseConfig->delayMinMs = 15;
    if (delayRatio != 0.0f) {
        baseConfig->delayRatio = delayRatio;
    } else {
        baseConfig->delayRatio = 3.2f;
    }
    if (delayMinMs != 0.0f) {
        baseConfig->delayMinMs = delayMinMs;
    }

    baseConfig->metadataPredelay = 0;
    baseConfig->predelayFactor = predelayFactor;
}

void fillBaseConfigOut(SetBaseConfigStruct* SBC, ReverbDSPCfgStruct* RevDSPCfg,
    const int16_t t60FltOrder, const int16_t colorFltOrder)
{
    // computed outputs
    for (int8_t i = 0; i != SBC->nrFbLoops; i++)
        RevDSPCfg->Reverberator.loopDelays[i] = SBC->loopDelays[i];

    // constants
    RevDSPCfg->Reverberator.nrLoops = SBC->nrFbLoops;

    RevDSPCfg->renderPredelay = SBC->renderPredelay;

    RevDSPCfg->Miscellanea.fs = SBC->fs;

    RevDSPCfg->Miscellanea.maxError = 0.01;

    RevDSPCfg->Miscellanea.T60In.t60FltOrder = t60FltOrder;
    RevDSPCfg->Miscellanea.ColorationIn.colorFltOrder = colorFltOrder;

    RevDSPCfg->Miscellanea.ColorationIn.responseSlopeLimit = 20;
    RevDSPCfg->Miscellanea.ColorationIn.slopeLimitPivotFreq = 1000;
    RevDSPCfg->Miscellanea.ColorationIn.nrFilterDesignFreq = NR_FILTER_DESIGN_FREQ;
    RevDSPCfg->Miscellanea.ColorationIn.maxError = -3; // dB
}

int32_t reverbSetBaseConfig(
    SetBaseConfigStruct* pSetIn, SetBaseConfigStruct* pSetOut, bool isDefaultAE)
{
#define INPUT_STRUCT_VALID false

    if (pSetIn == NULL)
        return 1;
    if (pSetOut == NULL)
        return 2;
    if (pSetIn->nrFbLoops > MAXNRFBLOOPS) {
        printf("invalid number of feedback loops. The branches might be incomplete.");
        return ERR_INVALID_NRFBLOOPS;
    }

    float roomSize_x = pSetIn->pRoomDimensions[0];
    float roomSize_y = pSetIn->pRoomDimensions[1];
    float roomSize_z = pSetIn->pRoomDimensions[2];

    if ((roomSize_x > 0.0f) && (roomSize_y > 0.0f) && (roomSize_z > 0.0f)) {
        float maxRoomDim = (roomSize_x > roomSize_y) ? roomSize_x : roomSize_y;
        maxRoomDim = (roomSize_z > maxRoomDim) ? roomSize_z : maxRoomDim;

        float minRoomDim = (roomSize_x < roomSize_y) ? roomSize_x : roomSize_y;
        minRoomDim = (roomSize_z < minRoomDim) ? roomSize_z : minRoomDim;

#define DELAYMINMS_FACTOR (1000.0f * 2.0f / 343.0f)
        pSetIn->delayMinMs = minRoomDim * DELAYMINMS_FACTOR;
        if (pSetIn->delayMinMs == 0)
            pSetIn->delayMinMs++; // fix for very small rooms
        float delayRatio = maxRoomDim / minRoomDim;
        // Use a minimum of 1.5 to ensure some variation in loop delays
        pSetIn->delayRatio = std::max(delayRatio, 1.5f);
    } else if (isDefaultAE) {
        // in case of default acoustic environment, delayRatio is being set independently of room dimensions
        float delayMinMs = pSetIn->delayMinMs;
        float predelayMs = ((float)pSetIn->metadataPredelay / (float)pSetIn->fs * 1000.0f);
        float delayRatio = pSetIn->delayRatio;
        pSetIn->delayRatio = std::max(delayRatio, 1.5f);
    }

    *pSetOut = *pSetIn;

    // Schroeder says: keep it limited to 1:1.5, Jot uses 1:2.5 in paper

    if (pSetIn->nrFbLoops > MAXNRFBLOOPS)
        return ERR_UNSOPPORTED_NR_FBACK_LOOPS;

    double r[MAXNRFBLOOPS] = { 0 };
    double loopDelay[MAXNRFBLOOPS] = { 0 };
    double minLoopDelay = 1e16;

    // Determine delay distribution
    double base = pow(0.14623044588361, 1.0 / (-(float)(pSetIn->nrFbLoops - 1)));

    // Matlab:  for n = 1:nrFbLoops
    // Matlab:    r(n) = base.^(- n + 1);
    // Matlab:  end
    for (int32_t n = 0; n != pSetIn->nrFbLoops; n++)
        r[n] = pow(base, -n);

    // Matlab:  % Make sure the delay-range is correct
    // Matlab:  alpha = (r(1) - delayRatio * r(nrFbLoops)) / (delayRatio - 1);
    // Matlab:  r = r + alpha;
    double alpha;
    alpha = (r[0] - pSetIn->delayRatio * r[pSetIn->nrFbLoops - 1]) / (pSetIn->delayRatio - 1);
    for (int32_t n = 0; n != pSetIn->nrFbLoops; n++)
        r[n] += alpha;

    // set loop delays
    {
        // Matlab:  % Determine delays
        // Matlab:  loopDelay(nrFbLoops) = delayMin*1e-3*fs;
        pSetOut->loopDelays[pSetIn->nrFbLoops - 1] = pSetIn->delayMinMs * 1e-3 * pSetIn->fs;
        // Matlab:  for n = nrFbLoops-1:-1:1
        // Matlab:    loopDelay(n) = loopDelay(nrFbLoops) * r(n) / r(nrFbLoops);
        // Matlab:  end
        for (int32_t n = pSetIn->nrFbLoops - 2; n != -1; n--)
            pSetOut->loopDelays[n] =
                pSetOut->loopDelays[pSetIn->nrFbLoops - 1] * r[n] / r[pSetIn->nrFbLoops - 1];
    }

    {

        // First loop delay is always the largest delay. Generate list of primes until first prime after largest delay.
        auto p = mpegcore::primesUpToNAfter(pSetOut->loopDelays[0], 1).value();
        auto largestPrime = p.back();

        // Matlab:  % Match to prime numbers
        // Matlab:  for k=1:length(loopDelay)
        // Matlab:    [d, idx] = min(abs(p-loopDelay(k)));
        // Matlab:    if idx == length(p)
        // Matlab:      loopDelay(k) = round(loopDelay(k));
        // Matlab:    else
        // Matlab:      loopDelay(k) = p(idx);
        // Matlab:    end
        // Matlab:  end
        for (int32_t k = 0; k != pSetIn->nrFbLoops; k++) {
            if (pSetOut->loopDelays[k] <= largestPrime) {
                int32_t min = largestPrime + 1, pos, diff;
                // linear search
                for (int32_t i = 1; i != p.size(); i++) {
                    diff = p[i] - pSetOut->loopDelays[k];
                    if (diff < 0)
                        diff = -diff;
                    if (diff < min) {
                        min = diff;
                        pos = i;
                    }
                }
                pSetOut->loopDelays[k] = p[pos];
            } else {
                pSetOut->loopDelays[k] = round(pSetOut->loopDelays[k]);
            }

            if (pSetOut->loopDelays[k] < minLoopDelay) {
                minLoopDelay = pSetOut->loopDelays[k];
            }
        } // end for
    }

    pSetOut->renderPredelay =
        std::fmax(pSetIn->metadataPredelay / pSetIn->predelayFactor, minLoopDelay);

    return 0;
} // reverbSetBaseConfig

void linspace(float d1, float d2, int16_t n, float* y)
{

    int16_t n1 = n - 1;

    // Matlab:    if isinf(c)
    // Matlab:    if isinf(d2 - d1) %opposite signs overflow
    // Matlab:    y = d1 + (d2./n1).*(0:n1) - (d1./n1).*(0:n1);
    // Matlab:    else
    // Matlab:    y = d1 + (0:n1).*((d2 - d1)./n1);
    // Matlab:    end
    // Matlab:    else
    // Matlab:    y = d1 + (0:n1).*(d2 - d1)./n1;
    // Matlab:    end
    int16_t i;
    for (i = 0; i <= n1; i++)
        y[i] = d1 + i * (d2 - d1) / n1;

    // Matlab:    if ~isempty(y)
    // Matlab:    if d1 == d2
    // Matlab:    y(:) = d1;
    // Matlab:    else
    // Matlab:    y(1) = d1;
    // Matlab:    y(end) = d2;
    // Matlab:    end
    // Matlab:    end
    y[0] = d1;
    y[i - 1] = d2;
} // linspace

void mpegInterp(
    float* fIn, float* valIn, float* fOut, float* valOut, int16_t lengthIn, int16_t lengthOut)
{
    if (lengthOut != NR_FILTER_DESIGN_FREQ)
        assert(false); // tested with this length, only

    for (int16_t i = 0; i != lengthOut; i++)
        valOut[i] = 0;

    int16_t i;
    for (i = 0; i != lengthOut; i++) {
        int16_t j;
        for (j = lengthIn - 1; j != -1; j--) {
            if (fIn[j] <= fOut[i])
                break;
        }

        if (j != -1) {

            if (j == lengthIn - 1)
                valOut[i] = valIn[lengthIn - 1];
            else {
                float x, x0, x1, y0, y1, y;
                x = fOut[i];
                x0 = fIn[j];
                x1 = fIn[j + 1];
                y0 = valIn[j];
                y1 = valIn[j + 1];
                y = y0 + ((x - x0) * ((y1 - y0) / (x1 - x0)));
                if (y < 0)
                    y = -25; // break point
                valOut[i] = y;
            } // else
        } else
            j++; // break point: why here

    } // for(int16_t i = 0; i != lengthOut; i++)
} // mpegInterp

void resampleToGrid(float* fIn, float* valIn, float* fOut, uint16_t fs, float* valOut,
    int16_t lengthIn, int16_t lengthOut)
{

    float fIn_internal[MPEGI::MAX_COUNT_OR_INDEX_VALUE + 2];
    float valIn_internal[MPEGI::MAX_COUNT_OR_INDEX_VALUE + 2];

    memcpy(fIn_internal, fIn, lengthIn * sizeof(float));
    memcpy(valIn_internal, valIn, lengthIn * sizeof(float));

    // Matlab:    if (fIn(1) ~= 0)
    if (fIn[0] != 0) {
        // Matlab:      fIn   = [0;         fIn(:)];
        // Matlab:      valIn = [valIn(1);  valIn(:)];
        for (int16_t i = 0; i != lengthIn; i++) {
            fIn_internal[i + 1] = fIn[i];
            valIn_internal[i + 1] = valIn[i];
        }
        fIn_internal[0] = 0;
        valIn_internal[0] = valIn[0];
        lengthIn++;
        // Matlab:    end
    }

    // Matlab:    if (fIn(end) ~= fs/2)
    if (fIn_internal[lengthIn - 1] != (float)fs / 2) {
        // Matlab:      fIn   = [fIn(:);    fs/2];
        fIn_internal[lengthIn] = (float)fs / 2;
        // Matlab:      valIn = [valIn(:);  valIn(end)];
        valIn_internal[lengthIn] = valIn_internal[lengthIn - 1];
        lengthIn++;
        // Matlab:    end
    }

    // length: GRID_DIM to GRID_DIM+2 = 24
    // or
    // 1025 to 1025 (getHRTFSetProperties)

    mpegInterp(fIn_internal, valIn_internal, fOut, valOut, lengthIn, lengthOut);
}

void gauss_f(float* vect_in, float* vect_out, int16_t length, float F, float Noct)
{
    float sigma;
    float vectOut_internal[NR_FILTER_DESIGN_FREQ];
    sigma = F / Noct;
    sigma /= M_PI;
    // Matlab:    g = exp(-(((f_x-F).^2)./(2.*(sigma^2)))); % Gaussian
    float num, den, sum = 0;
    den = 2 * (pow(sigma, 2));
    for (int16_t i = 0; i != length; i++) {
        num = -pow((vect_in[i] - F), 2);
        sum += (vectOut_internal[i] = exp(num / den));
    }
    // Matlab:    g = g./sum(g); % normalise magnitude
    for (int16_t i = 0; i != length; i++) {
        vectOut_internal[i] /= sum;
        if (vectOut_internal[i] < 1e-15)
            vectOut_internal[i] = 0; // get rid of denormalized numbers (#DEN)
    }

    memcpy(vect_out, vectOut_internal, NR_FILTER_DESIGN_FREQ * sizeof(float));

} // g_gauss

void smoothResponse(float* X, float* f, float* smoothed, int16_t length, float Noct)
{
    bool positiveX = true;
    float smoothedInternal[NR_FILTER_DESIGN_FREQ];
    for (int16_t i = 0; i != length; i++) {
        smoothedInternal[i] = X[i];
        if (X[i] < 0)
            positiveX = false;
    }

    if (Noct > 0) { // don't bother if no smoothing
        float g[NR_FILTER_DESIGN_FREQ];
        for (int16_t i = 0; i != length; i++)
            if (f[i] > 0) {
                gauss_f(f, g, NR_FILTER_DESIGN_FREQ, f[i], Noct);
                smoothedInternal[i] = 0;
                for (int16_t j = 0; j != length; j++)
                    smoothedInternal[i] += g[j] * X[j];
            }
        // Remove undershoot when X is positive
        if (positiveX) {
            for (int16_t i = 0; i != length; i++)
                if (smoothedInternal[i] < 0)
                    smoothedInternal[i] = 0;
        }
    }
    memcpy(smoothed, smoothedInternal, NR_FILTER_DESIGN_FREQ * sizeof(float));
} // smoothResponse

void preprocessReverbInputParameters(lateReverbInParms* pIn, lateReverb* pOut)
{

    // Create a perceptually relevant, sparse frequency grid that still works
    // well with filter design (i.e. doesn't create ripples between distant
    // points at high frequencies).

    // ERB spacing until 10 kHz, linear spacing above, 500 Hz apart.
    //    ERB stands for Equivalent Rectangular Bandwidth
    // Matlab:    fcFltDes = erb2freq([0:0.25:floor(freq2erb(10000))].');
    // Matlab:    fcFltDes = erb2freq([0.75 2 2.5 3.25:0.25:floor(freq2erb(10000))].');

    int16_t AcEnvLength = pIn->acousticParameters.dsrSize;
    float fcFltDes[NR_FILTER_DESIGN_FREQ];
    float RT60[NR_FILTER_DESIGN_FREQ];
    float DSR[NR_FILTER_DESIGN_FREQ];

    const float freq2erb = 34.545213651086577; /* freq2erb(10000), constant */
    const int16_t freq2erb_floor = (int16_t)(freq2erb - 0.5); // 34
    float step;
    int16_t i = 0;
    fcFltDes[i++] = 0.75;
    fcFltDes[i++] = 2;
    fcFltDes[i++] = 2.5;
    step = 3.25;

    while (step <= freq2erb_floor) {
        fcFltDes[i++] = step;
        step += 0.25;
    }

    int16_t j;
    for (j = 0; j < i; j++) {
        float partial;
        partial = exp(0.11 * fcFltDes[j]);
        fcFltDes[j] = (partial - 1) / 0.00437;
    }

    // entries up to 10 Khz are ready now

    // Matlab:    fcFltDes = [fcFltDes(1:end-1); linspace(fcFltDes(end), fs / 2, ceil((fs/2 - fcFltDes(end)) / 500)).'];

    float d1, d2;
    int16_t n;
    d1 = fcFltDes[j - 1];
    d2 = pIn->fs / 2;
    n = ceil((pIn->fs / 2 - fcFltDes[j - 1]) / 500);
    linspace(d1, d2, n, &fcFltDes[j - 1]);

    // move from arbitrary length to NR_FILTER_DESIGN_FREQ length

    resampleToGrid(pIn->acousticParameters.inputFrequencies, pIn->acousticParameters.t60, fcFltDes,
        pIn->fs, RT60, AcEnvLength, NR_FILTER_DESIGN_FREQ);

    resampleToGrid(pIn->acousticParameters.inputFrequencies, pIn->acousticParameters.dsr, fcFltDes,
        pIn->fs, DSR, AcEnvLength, NR_FILTER_DESIGN_FREQ);

    // Convert predelay from seconds to samples.

    // hrtf.left, .right and .iac come ready out of the box

    // ###########################
    // ####  Smooth the data  ####
    // ###########################

    float sum = 0;
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        sum += RT60[i];
    sum /= NR_FILTER_DESIGN_FREQ;
    if (sum < 0.3)
        smoothResponse(RT60, fcFltDes, RT60, NR_FILTER_DESIGN_FREQ, 1);
    else
        smoothResponse(RT60, fcFltDes, RT60, NR_FILTER_DESIGN_FREQ, 2);
    smoothResponse(DSR, fcFltDes, DSR, NR_FILTER_DESIGN_FREQ, 3);

    // Data at low frequencies is unreliable and often varies strongly.
    // When there is a clear slope up in the target response that can result
    // in unstable filters as it tends to put poles close to the unit circle
    // at 0 Hz to achieve that slope.
    sum = RT60[0] + RT60[1] + RT60[2];
    RT60[0] = sum / 3;

    RT60[1] = RT60[2] = RT60[0];
    sum = RT60[2] + RT60[4];
    RT60[3] = sum / 2;

    sum = DSR[0] + DSR[1] + DSR[2];

    DSR[0] = DSR[1] = DSR[2] = sum / 3;
    sum = DSR[2] + DSR[4];
    DSR[3] = sum / 2;

    float DSR_sm[NR_FILTER_DESIGN_FREQ];
    float RT60_sm[NR_FILTER_DESIGN_FREQ];

    smoothResponse(RT60, fcFltDes, RT60_sm, NR_FILTER_DESIGN_FREQ, 1);
    smoothResponse(DSR, fcFltDes, DSR_sm, NR_FILTER_DESIGN_FREQ, 1);
    memcpy(RT60, RT60_sm, 8 * sizeof(float));
    memcpy(DSR, DSR_sm, 8 * sizeof(float));

    // ####################################
    // ####  Create the output struct  ####
    // ####################################

    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        pOut->acousticRT60[i] = RT60[i];
        pOut->filterDesignFc[i] = fcFltDes[i];
        pOut->acousticDSR[i] = DSR[i];
    }

} // getLateReverbCharacteristics

void fillT60Config(ReverbDSPCfgStruct* RevDSPCfg, T60Params* T60Config)
{
    T60Config->Settings.nLoops = RevDSPCfg->Reverberator.nrLoops;

    for (int8_t i = 0; i != RevDSPCfg->Reverberator.nrLoops; i++)
        T60Config->Settings.loopDelays[i] = RevDSPCfg->Reverberator.loopDelays[i];

    T60Config->Settings.t60FltOrder = RevDSPCfg->Miscellanea.T60In.t60FltOrder;
    T60Config->fs = RevDSPCfg->Miscellanea.fs;

    T60Config->maxError = RevDSPCfg->Miscellanea.maxError;

    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        T60Config->fc[i] = RevDSPCfg->Miscellanea.fc[i];
    }
}

bool checkT60Configuration(T60Params* pIn)
{ // "correct configuration" means scalars verified by
    // appropriate input vector in unit test, only
    bool result = true;
    if (pIn->fs != 48000)
        result = false;
    if (pIn->Settings.nLoops > MAXNRFBLOOPS)
        result = false;
    if (pIn->Settings.t60FltOrder > MAX_FILTERORDER)
        result = false;
    return result;
} // checkT60Configuration

void reverbAddT60Filters(T60Params* pIn, T60Settings* pOut, IIRFilterType type)
{
    // targetT60, fc
    *pOut = pIn->Settings;
    int32_t fs = pIn->fs;
    int32_t nrFbLoops = pIn->Settings.nLoops;
    int32_t loopDelay[MAXNRFBLOOPS]; // [fltOrder]
    memcpy(loopDelay, pIn->Settings.loopDelays, MAXNRFBLOOPS * sizeof(uint32_t));

    // #### REVERBERATION-TIME FILTERS ####

    float h_b[MAXNRFBLOOPS][MAX_FILTERORDER + 1]; // [nrLoops][fltOrder + 1]
    memset(h_b, 0, sizeof(float) * MAXNRFBLOOPS * (MAX_FILTERORDER + 1));
    float h_a[MAXNRFBLOOPS][MAX_FILTERORDER + 1]; // [nrLoops][fltOrder + 1]
    memset(h_a, 0, sizeof(float) * MAXNRFBLOOPS * (MAX_FILTERORDER + 1));
    // Calculate the filters

    // Allow more errors at extreme frequencies.
    float freqDepLenience[NR_FILTER_DESIGN_FREQ];
    float fIn[4] = { 10, 100, 10000, (float)fs / 2 };
    float valIn[4] = { 1.5, 1, 1, 2.5 };
    float fFreq[NR_FILTER_DESIGN_FREQ];
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        fFreq[i] = pIn->fc[i];

    void mpegInterp(
        float* fIn, float* valIn, float* fOut, float* valOut, int16_t lengthIn, int16_t lengthOut);
    mpegInterp(fIn, valIn, fFreq, freqDepLenience, 4, NR_FILTER_DESIGN_FREQ);

    for (int32_t loopIdx = 0; loopIdx != nrFbLoops; loopIdx++) { // for each filter
        // Generate target spectrum
        // Matlab:      H_b = 10.^(-3 .* loopDelay(loopIdx) ./ (targetT60(:) .* fs));
        float H_b[NR_FILTER_DESIGN_FREQ];
        float expBase = (float)(-3 * loopDelay[loopIdx]) / fs;
        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            H_b[i] = pow(10, expBase / pIn->targetT60[i]);

        float MIN_ERROR[NR_FILTER_DESIGN_FREQ];
        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            MIN_ERROR[i] = (1 + pIn->maxError - H_b[i]) * 0.07;

        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            MIN_ERROR[i] *= freqDepLenience[i];

        // #noSingleAllowed
        double frequencies[NR_FILTER_DESIGN_FREQ];
        float K = 2.0f;
        K /= fs;
        for (int32_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            frequencies[i] = (pIn->fc[i] * K);

        EstimateIIR *DesignIn, *DesignOut;
        DesignIn = new EstimateIIR(type);
        DesignOut = new EstimateIIR(type);

        DesignIn->minimaxInterface.MAX_ITERATIONS = 64;
        DesignIn->minimaxInterface.MIN_ERROR_DIFF = 1e-4;
        DesignIn->minimaxInterface.err; // not set
        DesignIn->minimaxInterface.errFreqBand; // not set
        DesignIn->minimaxInterface.formatConfig; // not set

        DesignIn->minimaxInterface.iirOrder = pIn->Settings.t60FltOrder;

        memcpy(
            DesignIn->minimaxInterface.tgtFc, frequencies, NR_FILTER_DESIGN_FREQ * sizeof(double));
        memcpy(DesignIn->minimaxInterface.tgtMagn, H_b, NR_FILTER_DESIGN_FREQ * sizeof(float));
        memcpy(
            DesignIn->minimaxInterface.MIN_ERROR, MIN_ERROR, NR_FILTER_DESIGN_FREQ * sizeof(float));

        DesignIn->minimaxInterface.iterCount.filter = loopIdx;

        //char header[20];
        //sprintf(header, "t60_%d.dat",loopIdx+1);
        //void SaveDesignIIRLogs(EstimateIIR *DesignIn, char *header);
        // SaveDesignIIRLogs(DesignIn, header);

        designIIRFilter(DesignIn, DesignOut);

        IIRFilterCoefficients* output = &DesignOut->minimaxInterface.filterCoefficients;
        pOut->filters.sosSequence[loopIdx] = output->filters.sosSequence;

        delete DesignIn;
        delete DesignOut;

    } // for(int32_t loopIdx = 0; loopIdx != nrFbLoops; loopIdx++)

    // Changing filters to avoid phase differences
    // Matlab:   h_b = h_b .* repmat(sign(h_b(:,1)), 1, size(h_b, 2));
    // For SOS-sequence coefficients, we need to make sure that the global factor F
    // is non-negative to be equivalent to the polynomial correction.
    for (int32_t i = 0; i != nrFbLoops; i++)
        pOut->filters.sosSequence[i].F = std::abs(pOut->filters.sosSequence[i].F);

    // #### BUILD OUTPUT STRUCT ####
    pOut->t60FltOrder = pIn->Settings.t60FltOrder;

} // reverbAddT60Filters

void fillT60ConfigOut(T60Settings* T60Config, ReverbDSPCfgStruct* RevDSPCfg)
{
    RevDSPCfg->Reverberator.filterType = T60Config->outputType;
    RevDSPCfg->Reverberator.t60FltOrder = T60Config->t60FltOrder;
    for (int16_t loop = 0; loop != RevDSPCfg->Reverberator.nrLoops; loop++) {
        RevDSPCfg->Reverberator.filters.SOSs.t60Sos[loop] = T60Config->filters.sosSequence[loop];

        RevDSPCfg->Reverberator.loopDelays[loop] = T60Config->loopDelays[loop];
    }
}

void fillColorationConfig(ReverbDSPCfgStruct* RevDSPCfg, ColorParams* ColorConfig, float* DSR,
    float* RT60, uint16_t metadataPredelay)
{
    ColorConfig->nrLoops = RevDSPCfg->Reverberator.nrLoops;

    for (int32_t i = 0; i != ColorConfig->nrLoops; i++)
        ColorConfig->loopDelays[i] = RevDSPCfg->Reverberator.loopDelays[i];

    ColorConfig->colorFltOrder = RevDSPCfg->Miscellanea.ColorationIn.colorFltOrder;
    ColorConfig->MIN_ERROR = RevDSPCfg->Miscellanea.ColorationIn.maxError;

    ColorConfig->settings.fs = RevDSPCfg->Miscellanea.fs;
    ColorConfig->settings.responseSlopeLimit =
        RevDSPCfg->Miscellanea.ColorationIn.responseSlopeLimit;
    ColorConfig->settings.slopeLimitPivotFreq =
        RevDSPCfg->Miscellanea.ColorationIn.slopeLimitPivotFreq;
    ColorConfig->lateReverbCharact.nrFilterDesignFreq =
        RevDSPCfg->Miscellanea.ColorationIn.nrFilterDesignFreq;

    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        ColorConfig->lateReverbCharact.fc[i] = RevDSPCfg->Miscellanea.fc[i];
        ColorConfig->lateReverbCharact.acousticDSR[i] = DSR[i];
        ColorConfig->lateReverbCharact.acousticRT60[i] = RT60[i];
    }
    ColorConfig->lateReverbCharact.metadataPredelay = metadataPredelay;

} // fillColorationConfig

bool checkColorationInputConfiguration(ColorParams* pInColorationStruct)
{ // "correct configuration" means scalars verified by appropriate input vector in unit test, only
    bool result = true;
    if (pInColorationStruct->nrLoops > MAXNRFBLOOPS)
        result = false;
    if ((pInColorationStruct->settings.fs != 48000) && (pInColorationStruct->settings.fs != 64000))
        result = false; // two sampling frequencies used in the test vectors
    if (pInColorationStruct->colorFltOrder > MAX_FILTERORDER)
        result = false;

    return result;
} // checkColorationInputConfiguration

int32_t responseSlopeLimit(double X[], int32_t dimX, double* fc, int32_t T, int32_t pivotFreq)
{

    // The response slope of X expressed in dB per octave should be:
    //   20 * log10( (X(i) / X(i-1))^(1 / log2( f(i) / f(i-1) )) )
    //
    // In order to limit this slope to T dB/octave, we can write:
    //   20 * log10( (X(i) / X(i-1))^(1 / log2( f(i) / f(i-1) )) ) < T
    //               (X(i) / X(i-1))^(1 / log2( f(i) / f(i-1) ))   < 10^(T/20)
    //               (X(i) / X(i-1))                               < 10^(T/20 * log2( f(i) / f(i-1) ))

    // Prepare look-up table. Depends only on grid vector determined in
    // getLateReverbCharacteristics().
    if (dimX != NR_FILTER_DESIGN_FREQ)
        return 3; //ERR_INVALID_DIMENSION;

    double Tdiv20 = T / 20.0;
    double logFRatio[NR_FILTER_DESIGN_FREQ] = { 0 };

    for (int32_t i = 1; i != dimX; i++) {
        double prev = fc[i - 1];
        if (prev < 1)
            prev = 1;
        logFRatio[i] = log2(fc[i] / prev);
    }

    // Find frequency band closest to chosen pivot frequency.
    //  MATLAB:     [~, idxPivot] = min(abs(fc - pivotFreq));
    double min = 1000000;
    int32_t idxPivot;
    for (int32_t i = 0; i != dimX; i++) {
        double absdiff = fabs(fc[i] - pivotFreq);
        if (absdiff < min) {
            min = absdiff;
            idxPivot = i;
        }
    }

    double Y[NR_FILTER_DESIGN_FREQ] = { 0 };

    // Take over value at pivot frequency
    Y[idxPivot] = X[idxPivot];

    // Go up from pivot frequency and limit the slope to the maximum given by T.
    for (int32_t i = idxPivot + 1; i != dimX; i++) {
        //  MATLAB:     threshold = 10.^(T/20 * logFRatio(i));
        double threshold;
        threshold = pow(10, Tdiv20 * logFRatio[i]);

        double desiredChange;
        desiredChange = X[i] / Y[i - 1];

        double change = fabs(desiredChange);
        if (desiredChange >= 1)
            change = (change < threshold) ? change : threshold;
        else {
            double rhs = 1 / threshold;
            change = (change > rhs) ? change : rhs;
        }

        Y[i] = Y[i - 1] * change;
    } // for i

    // Go down from pivot frequency and limit the slope to the maximum given by T.
    for (int32_t i = idxPivot - 1; i != -1; i--) {
        //  MATLAB:      threshold = 10.^(T/20 * logFRatio(i));
        double threshold;
        threshold = pow(10, Tdiv20 * logFRatio[i]);

        double desiredChange;
        desiredChange = X[i] / Y[i + 1];

        double change = fabs(desiredChange);
        if (desiredChange >= 1)
            change = (change < threshold) ? change : threshold;
        else {
            double rhs = 1 / threshold;
            change = (change > rhs) ? change : rhs;
        }

        Y[i] = Y[i + 1] * change;
    } // for i

    for (int32_t i = 0; i != dimX; i++)
        X[i] = Y[i]; // in-place return of output vector

    return 0;
} // responseSlopeLimit

int32_t reverbCalculateLevelData(
    ColorParams* pInColorationStruct, ColorParams* pOutColorationStruct, bool isDefaultAE)
{
    // To get sufficient resolution for the low frequencies (~23 Hz)
    int32_t anaFFTSize = 2048;
    double* pT60 = pInColorationStruct->lateReverbCharact.acousticRT60;
    double* pfc = pInColorationStruct->lateReverbCharact.fc;

    *pOutColorationStruct = *pInColorationStruct;

    // MATLAB:  minDelayDiff = min(abs(diff(reverbConfig.loopDelays)));
    int32_t minDelayDiff = 99999;
    {
        int32_t i = 0;
        int32_t* plD = pOutColorationStruct->loopDelays;
        for (; i != (pOutColorationStruct->nrLoops - 1); i++, plD++) {
            int32_t d = abs(*plD - *(plD + 1));
            if (d < minDelayDiff && d > 0)
                minDelayDiff = d;
        }
    }

    // MATLAB:  minDelay = min(abs(reverbConfig.loopDelays));
    int32_t minDelay = 99999;
    int32_t maxDelay = 0;
    {
        int32_t i = 0;
        int32_t* plD = pOutColorationStruct->loopDelays;
        for (; i != (pOutColorationStruct->nrLoops); i++, plD++) {
            if (*plD < minDelay)
                minDelay = *plD;
            if (*plD > maxDelay)
                maxDelay = *plD;
        }
    }

    // Predict the frequency dependent energy of the reverb without coloration filter.
    // MATLAB:  	A0_square_est = 10. ^ (-3.*minDelay . / (T60.*fs)). ^ 2;

    // C decomposition: interm1 = T60.*fs
    // C decomposition: interm2 = (-3.*minDelay . / (T60.*fs));
    // C decomposition: interm3 = 10. ^ (-3.*minDelay . / (T60.*fs));
    // C result: A0_square_est = 10. ^ (-3.*minDelay . / (T60.*fs)). ^ 2;

    double A0_square_est[NR_FILTER_DESIGN_FREQ];
    {
        double interm1[NR_FILTER_DESIGN_FREQ],
            *ptrSrc = pOutColorationStruct->lateReverbCharact.acousticRT60, *ptrDest = interm1;
        int32_t fs = pOutColorationStruct->settings.fs;
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            *(ptrDest++) = *(ptrSrc++) * fs;
        }
        int32_t multK = minDelay * -3;
        double interm2[NR_FILTER_DESIGN_FREQ];
        ptrSrc = interm1, ptrDest = interm2;
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            *(ptrDest++) = multK / *(ptrSrc++);
        }
        double interm3[NR_FILTER_DESIGN_FREQ];
        ptrSrc = interm2, ptrDest = interm3;
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            *(ptrDest++) = pow(10.0, *(ptrSrc++));
        }
        ptrSrc = interm3, ptrDest = A0_square_est;
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++, ptrSrc++) {
            *(ptrDest++) = *(ptrSrc) * *(ptrSrc);
        }
    }

    // MATLAB:  alpha = log(10. ^ (-60 / 20)) . / -T60;
    double Num = -6.907755278982137 * -1; // embed negative denominator
    double alpha[NR_FILTER_DESIGN_FREQ];
    {
        double *ptrSrc = pOutColorationStruct->lateReverbCharact.acousticRT60, *ptrDest = alpha;
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            *(ptrDest++) = Num / (*ptrSrc++);
        }
    }

    double revPredNormEnergy[NR_FILTER_DESIGN_FREQ];
    {

        double minDelayMs = (double)minDelay / pOutColorationStruct->settings.fs * 1000;
        double delayRatio = (double)maxDelay / std::max((double)minDelay, 1.0);
        // Apply calibration
        double minDelayCalCoeff, delayRatioCalCoeff, T60CalCoeff, combinedOffsetCalCoeff;
        switch (pOutColorationStruct->nrLoops) {
        case 3:
            minDelayCalCoeff = -1.01887;
            delayRatioCalCoeff = -0.79637;
            T60CalCoeff = 0.02277;
            combinedOffsetCalCoeff =
                std::pow(10, 0.89629) * std::pow(10, 0.53578) * std::pow(10, -0.01255);
            break;
        case 7:
            minDelayCalCoeff = -1.07545;
            delayRatioCalCoeff = -0.87111;
            T60CalCoeff = 0.09588;
            combinedOffsetCalCoeff =
                std::pow(10, 1.42241) * std::pow(10, 0.58606) * std::pow(10, -0.05283);
            break;
        case 15:
            minDelayCalCoeff = -1.0962;
            delayRatioCalCoeff = -0.89751;
            T60CalCoeff = 0.11870;
            combinedOffsetCalCoeff =
                std::pow(10, 1.81447) * std::pow(10, 0.60383) * std::pow(10, -0.06541);
            break;
        case 31:
            minDelayCalCoeff = -1.11467;
            delayRatioCalCoeff = -0.90326;
            T60CalCoeff = 0.12366;
            combinedOffsetCalCoeff =
                std::pow(10, 2.17928) * std::pow(10, 0.60774) * std::pow(10, -0.06814);
            break;
        case 63:
            minDelayCalCoeff = -1.13477;
            delayRatioCalCoeff = -0.93452;
            T60CalCoeff = 0.11445;
            combinedOffsetCalCoeff =
                std::pow(10, 2.52982) * std::pow(10, 0.62875) * std::pow(10, -0.06615);
            break;
        default:
            assert(0); // TODO: properly exit with appropriate error message
        }
        double scalarComp = std::pow(minDelayMs, minDelayCalCoeff)
            * std::pow(delayRatio, delayRatioCalCoeff) * combinedOffsetCalCoeff;
        double fsNorm = pOutColorationStruct->settings.fs / 48000.0;
        double loopFactor = (15.0 / (double)pOutColorationStruct->nrLoops)
            * (15.0 / (double)pOutColorationStruct->nrLoops);
        double* ptrT60 = pOutColorationStruct->lateReverbCharact.acousticRT60;

        scalarComp *= fsNorm * loopFactor;

        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            revPredNormEnergy[i] = A0_square_est[i] / (2.0 * alpha[i]);
            revPredNormEnergy[i] *= scalarComp * std::pow(ptrT60[i], T60CalCoeff);
        }
    }

    // Adjust DSR for difference between render delay and DSR's predelay.
    double revEnergyFactor[NR_FILTER_DESIGN_FREQ];
    int32_t predelayDelta;
    predelayDelta = pOutColorationStruct->lateReverbCharact.metadataPredelay
        - pOutColorationStruct->renderPredelay;
    if (predelayDelta != 0 && !isDefaultAE) {
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            revEnergyFactor[i] =
                exp(2 * alpha[i] * predelayDelta / pOutColorationStruct->settings.fs);
    } else {
        for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
            revEnergyFactor[i] = 1.0;
    }

    double sqrtRevPNE[NR_FILTER_DESIGN_FREQ]; // intermediate array that comes useful more than once
    // #### CALCULATE THE REFERENCE SPECTRA FOR THE COLORATION FILTERS ####
    // MATLAB:          eqFilterTarget = sqrt(lateReverbCharacteristics.acousticDSR .* revEnergyFactor )
    // MATLAB:									./ sqrt(revPredNormEnergy);
    // C decomposition: parm1 = lateReverbCharacteristics.acousticDSR..* revEnergyFactor;
    // C decomposition: parm1 = sqrt(parm1);
    // C decomposition: sqrtRevPNE = sqrt(revPredNormEnergy);
    // C decomposition: eqFilterTarget = parm1 / sqrtRevPNE

    double parm1[NR_FILTER_DESIGN_FREQ];
    double* ptr_src1 = pOutColorationStruct->lateReverbCharact.acousticDSR;
    double* ptr_src3 = revEnergyFactor;
    double* ptr_dest = parm1;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        *(ptr_dest++) = *(ptr_src1++) * *(ptr_src3++);
    }

    double* ptr_src = ptr_dest = parm1; // in-place computation
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        *(ptr_dest++) = sqrt(*(ptr_src++));
    }

    ptr_src = revPredNormEnergy;
    ptr_dest = sqrtRevPNE;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        *(ptr_dest++) = sqrt(*(ptr_src++));
    }

    ptr_src1 = parm1;
    double* ptr_src2 = sqrtRevPNE;
    ptr_dest = pOutColorationStruct->eqFilterTarget;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        *(ptr_dest++) = *(ptr_src1++) / *(ptr_src2++);
    }

    // Make sure that target responses don't include too steep slopes. Filters
    // with limited order cannot represent steep slopes AND response features at
    // other frequencies.

    float sum = 0, ave;
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        sum += pT60[i];
    }
    ave = sum / NR_FILTER_DESIGN_FREQ;
    if (ave < 0.3) {
        float tempOutput[NR_FILTER_DESIGN_FREQ];
        float fFreq[NR_FILTER_DESIGN_FREQ];
        float spectrum[NR_FILTER_DESIGN_FREQ];
        for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            fFreq[i] = pfc[i];
            spectrum[i] = pOutColorationStruct->eqFilterTarget[i];
        }
        smoothResponse(spectrum, fFreq, tempOutput, NR_FILTER_DESIGN_FREQ, 1);
        for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
            pOutColorationStruct->eqFilterTarget[i] = tempOutput[i];
        }
    }

    if (responseSlopeLimit(pOutColorationStruct->eqFilterTarget, NR_FILTER_DESIGN_FREQ, pfc,
            pOutColorationStruct->settings.responseSlopeLimit,
            pOutColorationStruct->settings.slopeLimitPivotFreq))
        return 4; //ERR_INVALID_SIZE;

    // commonComponentClrFlt = mean(eqFilterTarget);
    pOutColorationStruct->commonComponentClrFlt = 0;
    for (int32_t i = 0; i != pOutColorationStruct->lateReverbCharact.nrFilterDesignFreq; i++) {
        pOutColorationStruct->commonComponentClrFlt += (pOutColorationStruct->eqFilterTarget[i]);
    }

    pOutColorationStruct->commonComponentClrFlt /=
        (pOutColorationStruct->lateReverbCharact.nrFilterDesignFreq);

    for (int32_t i = 0; i != pOutColorationStruct->lateReverbCharact.nrFilterDesignFreq; i++) {
        pOutColorationStruct->eqFilterTarget[i] /= pOutColorationStruct->commonComponentClrFlt;
    }

    return 0;
} // reverbCalculateLevelData

void MinError(float* MIN_ERROR, float* target, float* maxDev)
{
    // MATLAB: MIN_ERROR  = abs(Target  - Target  .* 10.^(maxDev/20));
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        float expon = maxDev[i] / 20.0f;
        float factor = pow(10.0f, expon);
        MIN_ERROR[i] = fabs(target[i] * (1.0f - factor));
    }
}

void MinError(float* MIN_ERROR, float* target, float maxDev)
{
    // MATLAB: MIN_ERROR  = abs(Target  - Target  .* 10.^(maxDev/20));
    float expon = maxDev / 20.0f;
    float factor = pow(10.0f, expon);
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        MIN_ERROR[i] = fabs(target[i] * (1.0f - factor));
    }
}

void reverbAddColorFilters(
    ColorParams* pInColorationStruct, ColorParams* pOutColorationStruct, IIRFilterType type)
{
    *pOutColorationStruct = *pInColorationStruct;

    float L[2][MAX_FILTERORDER + 1], R[2][MAX_FILTERORDER + 1];
    float target[NR_FILTER_DESIGN_FREQ];
    float MIN_ERROR[NR_FILTER_DESIGN_FREQ];

    EstimateIIR* In_ACF = new EstimateIIR(type);
    EstimateIIR* Out_ACF = new EstimateIIR(type);

    double fc[NR_FILTER_DESIGN_FREQ];
    memcpy(fc, pOutColorationStruct->lateReverbCharact.fc, sizeof(double) * NR_FILTER_DESIGN_FREQ);

    int32_t fNyq;
    fNyq = pOutColorationStruct->settings.fs / 2;

    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        fc[i] /= fNyq;

    // #### GENERATE THE FILTERS ####

    // Set the error function
    float maxDev[NR_FILTER_DESIGN_FREQ];
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        maxDev[i] = pOutColorationStruct->MIN_ERROR;

    // MATLAB:		[~, idxLF] = min(abs(fc - 100));
    int16_t idxLF;
    for (idxLF = 1; idxLF != NR_FILTER_DESIGN_FREQ - 1; idxLF++)
        if (pOutColorationStruct->lateReverbCharact.fc[idxLF] >= 100)
            break;
    float d1, d2;
    d1 = 100.0 - pOutColorationStruct->lateReverbCharact.fc[idxLF - 1];
    d2 = pOutColorationStruct->lateReverbCharact.fc[idxLF] - 100;
    if (d1 < d2)
        idxLF--;

    for (int16_t fIdx = 0; fIdx != idxLF; fIdx++) {
        float add;
        // MATLAB:		maxDev(fIdx) = maxDev(fIdx) + 2 * settings.iirFltDesign.MIN_ERROR .* (idxLF - fIdx) ./ idxLF;
        add = 2 * pOutColorationStruct->MIN_ERROR;
        add *= (idxLF - fIdx);
        add /= (idxLF + 1);
        maxDev[fIdx] += add;
    }

    // ########################
    // ####   EQ FILTER    ####
    // ########################
    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        target[i] = pOutColorationStruct->eqFilterTarget[i];

    MinError(MIN_ERROR, target, maxDev); // Left, here. Right, below.

    In_ACF->minimaxInterface.iirOrder = pOutColorationStruct->colorFltOrder; // parm 1
    memcpy(In_ACF->minimaxInterface.tgtFc, fc,
        sizeof(double) * NR_FILTER_DESIGN_FREQ); // parm 2
    memcpy(In_ACF->minimaxInterface.tgtMagn, target,
        sizeof(float) * NR_FILTER_DESIGN_FREQ); // parm 3

    memcpy(In_ACF->minimaxInterface.MIN_ERROR, MIN_ERROR,
        sizeof(float) * NR_FILTER_DESIGN_FREQ); // parm 4
    In_ACF->minimaxInterface.MAX_ITERATIONS = 64;
    void SaveDesignIIRLogs(EstimateIIR * DesignIn, char* header);
    // SaveDesignIIRLogs(In_ACF, "color_l.dat");
    Out_ACF = In_ACF;
    designIIRFilter(In_ACF, Out_ACF);
    IIRFilterCoefficients* output = &Out_ACF->minimaxInterface.filterCoefficients;

    // Ensure that filters are not opposite phase.
    pOutColorationStruct->filters.sosSequence[0] = output->filters.sosSequence;
    pOutColorationStruct->filters.sosSequence[0].F =
        std::abs(pOutColorationStruct->filters.sosSequence[0].F);

} // reverbAddColorFilters

void fillColorationConfigOut(ColorParams* ColorConfig, ReverbDSPCfgStruct* RevDSPCfg)
{
    RevDSPCfg->Reverberator.colorFltOrder = RevDSPCfg->Miscellanea.ColorationIn.colorFltOrder;

    RevDSPCfg->Reverberator.commonComponentClrFlt = ColorConfig->commonComponentClrFlt;

    int16_t fltIdx = 0;
    RevDSPCfg->Reverberator.filters.SOSs.colorSos[fltIdx] =
        ColorConfig->filters.sosSequence[fltIdx];

} // fillColorationConfigOut

mpegcore::GraphicEqParameters calcDirectivityFilter(ReverbDSPCfgStruct* RevDSPCfg,
    MPEGI::DirectivitySpectrum dirFltTarget, IIRFilterType type, uint16_t dirFltOrder,
    float maxErrordB)
{
    /* STEP 1 - Resample to filter design grid */
    int16_t nrFreqIn = dirFltTarget.frequencies.size();
    float* fIn = dirFltTarget.frequencies.data();
    float* tgtSpecBsGrid = dirFltTarget.coefficients.data();

    int16_t nrFreqOut = RevDSPCfg->Miscellanea.ColorationIn.nrFilterDesignFreq;
    float* fOut = RevDSPCfg->Miscellanea.fc;
    float tgtSpecFltDesGrid[NR_FILTER_DESIGN_FREQ];

    uint16_t fs = RevDSPCfg->Miscellanea.fs;

    resampleToGrid(fIn, tgtSpecBsGrid, fOut, fs, tgtSpecFltDesGrid, nrFreqIn, nrFreqOut);

    /* STEP 2 - Apply smoother to the spectrum */
    float tgtSpecFltDesGrid_sm[NR_FILTER_DESIGN_FREQ];

    smoothResponse(tgtSpecFltDesGrid, fOut, tgtSpecFltDesGrid_sm, NR_FILTER_DESIGN_FREQ, 1);

    /* STEP 3 - Create configuration for the filter design */
    EstimateIIR* iirDesignIn = new EstimateIIR(type);
    EstimateIIR* iirDesignOut; // = new EstimateIIR(type);

    iirDesignIn->minimaxInterface.iirOrder = dirFltOrder; // parm 1
    // memcpy(iirDesignIn->minimaxInterface.tgtFc, fOut,
    //     sizeof(double) * NR_FILTER_DESIGN_FREQ); // parm 2
    for (int frIdx = 0; frIdx < NR_FILTER_DESIGN_FREQ; frIdx++) { // parm 2
        iirDesignIn->minimaxInterface.tgtFc[frIdx] = fOut[frIdx] / ((double)fs / 2);
    }
    memcpy(iirDesignIn->minimaxInterface.tgtMagn, tgtSpecFltDesGrid_sm,
        sizeof(float) * NR_FILTER_DESIGN_FREQ); // parm 3

    // memcpy(iirDesignIn->minimaxInterface.MIN_ERROR, maxError,
    //     sizeof(float) * NR_FILTER_DESIGN_FREQ); // parm 4
    MinError(iirDesignIn->minimaxInterface.MIN_ERROR, tgtSpecFltDesGrid_sm, maxErrordB);
    iirDesignIn->minimaxInterface.MAX_ITERATIONS = 64;
    // void SaveDesignIIRLogs(EstimateIIR * DesignIn, char* header);
    // SaveDesignIIRLogs(In_ACF, "color_l.dat");
    iirDesignOut = iirDesignIn;

    /* STEP 4 - Design the filter */
    designIIRFilter(iirDesignIn, iirDesignOut);

    /* STEP 5 - Extract the filter */
    IIRFilterCoefficients* output = &iirDesignOut->minimaxInterface.filterCoefficients;

    mpegcore::GraphicEqParameters dirFlt = output->filters.sosSequence;

    //free(iirDesignIn);
    delete iirDesignOut;

    return dirFlt;
}
