/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include "DesignIIR.h"
#include "DirectivitySpectrum.h"

#include <array>
#include <cmath>
#include <vector>

#include <mpegcore/datatypes_reverb.hpp>
#include <mpegcore/primes.hpp>

#define NR_SOURCES_MAX 300
#define COEFFDIM 257 // block size + 1
#define MAXNRFBLOOPS 63 /* hard-coded as of now */

#define ERR_UNSOPPORTED_NR_FBACK_LOOPS 30
#define ERR_INVALID_FEEDBACK_MATRIX 31
#define ERR_MATRIX_NOT_INVERTIBLE 32
#define ERR_INVALID_NRFBLOOPS 33

// TSI Output 3: Reverberator
struct ReverberatorConfig
{
    int16_t blocksize; // Opt
    uint16_t nrLoops; // Must
    uint32_t loopDelays[MAXNRFBLOOPS]; // Must

    IIRFilterType filterType;

    int16_t t60FltOrder; // Must
    int16_t colorFltOrder; // Must

    float commonComponentClrFlt;

    union U
    {
        struct
        {
            float t60FltCoeffA[MAXNRFBLOOPS][MAX_FILTERORDER + 1]; // Must
            float t60FltCoeffB[MAXNRFBLOOPS][MAX_FILTERORDER + 1]; // Must

            float colorFltCoeffA[2][MAX_FILTERORDER + 1];
            float colorFltCoeffB[2][MAX_FILTERORDER + 1];
        } polynomial;

        struct
        {
            mpegcore::GraphicEqParameters t60Sos[MAXNRFBLOOPS];
            mpegcore::GraphicEqParameters colorSos[2];

        } SOSs;
        U() {}
        ~U() {}
    } filters;
};

struct SetBaseConfigStruct
{
    /* input parms */
    int32_t nrFbLoops;
    float delayMinMs;
    double delayRatio;
    int32_t fs;
    int32_t renderPredelay;
    int32_t metadataPredelay;
    float predelayFactor;
    float* pRoomDimensions;

    float loopDelays[MAXNRFBLOOPS];
};

struct ListenerPosition
{
    float x, y, z;
};

#define MAX_ROOMS 10

struct AcousticEnvironmentProperties
{
    std::string roomId; // Room identifier as from the SceneGraph
    float dimensions[3]; // Room dimensions in x, y and z direction
};

struct MiscConfig
{
    // everything useful that is not part of the Reverb Block output
    uint16_t fs;
    float fc[NR_FILTER_DESIGN_FREQ];
    double maxError;
    struct
    {
        float targetT60[NR_FILTER_DESIGN_FREQ];
        int16_t t60FltOrder;
    } T60In;
    struct
    {
        int32_t responseSlopeLimit;
        int32_t slopeLimitPivotFreq;
        int16_t nrFilterDesignFreq;
        int16_t t60FltOrder;
        int16_t colorFltOrder;
        float maxError;
    } ColorationIn;
};

struct UserPhysicalEnv
{
    bool dataUpdated; // true if it's first time for the user in the current room
    int16_t roomId; // 'Code' of the room
};

struct ReverbDSPCfgStruct
{
    // Reverb parameters computed by the compute_reverb block execute() method

    // Reverb rendering delay
    float renderPredelay;

    // BaseConfig Output
    ReverberatorConfig Reverberator;

    std::string acousticEnvironmentId;

    // Miscellanea
    MiscConfig Miscellanea;

    bool bStructInitialized;
};

struct lateReverb
{
    float filterDesignFc[NR_FILTER_DESIGN_FREQ]; // actual length is acousticParameters.dsrSize;
    float acousticRT60[NR_FILTER_DESIGN_FREQ]; // actual length is acousticParameters.dsrSize;
    float acousticDSR[NR_FILTER_DESIGN_FREQ]; // actual length is acousticParameters.dsrSize;
    int metadataPredelay;
};

struct lateReverbInParms
{
    uint16_t fs;
    struct
    {
        float inputFrequencies[NR_FILTER_DESIGN_FREQ];
        float t60[NR_FILTER_DESIGN_FREQ];
        float dsr[NR_FILTER_DESIGN_FREQ];
        int16_t dsrSize;
    } acousticParameters;
};

struct T60Settings
{
    int32_t nLoops;
    int32_t loopDelays[MAXNRFBLOOPS]; // filled in the first entries
    int32_t t60FltOrder;

    union U
    {
        double t60FltCoeff[MAXNRFBLOOPS][2
            * (MAX_FILTERORDER + 1)]; // [nrFbLoops][2*(order+1) : b + a, interleaved]

        mpegcore::GraphicEqParameters sosSequence[MAXNRFBLOOPS];
        U() {}
        ~U() {}
    } filters;

    IIRFilterType outputType;
    T60Settings(IIRFilterType type)
        : outputType(type)
    {
        switch (type) {
        case IIRFilterType::Polynomial: {
            for (int i = 0; i < MAXNRFBLOOPS; ++i) {
                for (int j = 0; j < 2 * (MAX_FILTERORDER + 1); ++j)
                    filters.t60FltCoeff[i][j] = 0;
            }
        } break;
        case IIRFilterType::SOSs: {
            for (int i = 0; i < MAXNRFBLOOPS; ++i)
                new (&filters.sosSequence[i]) mpegcore::GraphicEqParameters();
        } break;
        default:
            break;
        }
    }
    T60Settings(const T60Settings& other) { *this = other; }
    T60Settings& operator=(const T60Settings& other)
    {
        for (int i = 0; i < MAXNRFBLOOPS; ++i) {
            loopDelays[i] = other.loopDelays[i]; // filled in the first entries
        }
        t60FltOrder = other.t60FltOrder;
        outputType = other.outputType;
        nLoops = other.nLoops;
        switch (outputType) {
        case IIRFilterType::Polynomial: {
            for (int i = 0; i < MAXNRFBLOOPS; ++i)
                for (int j = 0; j < 2 * (MAX_FILTERORDER + 1); ++j)
                    filters.t60FltCoeff[i][j] = other.filters.t60FltCoeff[i][j];
        } break;
        case IIRFilterType::SOSs: {
            for (int i = 0; i < MAXNRFBLOOPS; ++i)
                filters.sosSequence[i] = other.filters.sosSequence[i];
        } break;
        default:
            break;
        }
        return *this;
    }
};

struct T60Params
{
    T60Params(IIRFilterType type)
        : Settings(type)
    {}
    T60Settings Settings;
    double maxError{};
    int32_t fs{};
    double targetT60[NR_FILTER_DESIGN_FREQ];
    double fc[NR_FILTER_DESIGN_FREQ];
};

struct ColorParams
{
    int32_t nrLoops;
    int32_t loopDelays[MAXNRFBLOOPS];
    int32_t colorFltOrder;

    struct
    {
        int32_t fs;
        int32_t DEBUG;
        int32_t blockSize;
        int32_t responseSlopeLimit;
        int32_t slopeLimitPivotFreq;
        int32_t mathPrecision;
    } settings;
    struct
    {
        int32_t nrFilterDesignFreq;
        double fc[NR_FILTER_DESIGN_FREQ];
        double acousticRT60[NR_FILTER_DESIGN_FREQ];
        double acousticDSR[NR_FILTER_DESIGN_FREQ];
        int32_t metadataPredelay;
    } lateReverbCharact;
    int32_t renderPredelay;
    float MIN_ERROR;

    double eqFilterTarget[NR_FILTER_DESIGN_FREQ];
    float commonComponentClrFlt;

    union U
    {
        double colorFltCoeff[2][MAX_FILTERORDER + 1]; // 0 = Den	1 = Num

        mpegcore::GraphicEqParameters sosSequence[2];
        U() {}
        ~U() {}
    } filters;

    IIRFilterType outputType;
    ColorParams() = delete;
    ColorParams(IIRFilterType type)
        : outputType(type)
    {
        switch (type) {
        case IIRFilterType::Polynomial: {
            for (int i = 0; i < 2; ++i) {
                for (int j = 0; j < MAX_FILTERORDER + 1; ++j) {
                    filters.colorFltCoeff[i][j] = 0.0f;
                }
            }
        } break;
        case IIRFilterType::SOSs: {
            for (int i = 0; i < 2; ++i)
                new (&filters.sosSequence[i]) mpegcore::GraphicEqParameters();
        } break;
        default:
            break;
        }
    }
    ColorParams(const ColorParams& other)
    {
        // Calculate level data copy
        for (int i = 0; i < NR_FILTER_DESIGN_FREQ; ++i) {
            eqFilterTarget[i] = other.eqFilterTarget[i];
        }
        commonComponentClrFlt = other.commonComponentClrFlt;

        nrLoops = other.nrLoops;
        for (int i = 0; i < MAXNRFBLOOPS; ++i) {
            loopDelays[i] = other.loopDelays[i];
        }

        colorFltOrder = other.colorFltOrder;

        settings = other.settings;
        lateReverbCharact = other.lateReverbCharact;

        renderPredelay = other.renderPredelay;
        MIN_ERROR = other.MIN_ERROR;

        outputType = other.outputType;
        switch (outputType) {
        case IIRFilterType::Polynomial: {
            for (int i = 0; i < 2; ++i) {
                for (int j = 0; j < MAX_FILTERORDER + 1; ++j) {
                    filters.colorFltCoeff[i][j] = other.filters.colorFltCoeff[i][j];
                }
            }
        } break;
        case IIRFilterType::SOSs: {
            for (int i = 0; i < 2; ++i)
                filters.sosSequence[i] = other.filters.sosSequence[i];
        } break;
        default:
            break;
        }
    }
    ColorParams& operator=(const ColorParams& other)
    {
        nrLoops = other.nrLoops;
        for (int i = 0; i < MAXNRFBLOOPS; ++i) {
            loopDelays[i] = other.loopDelays[i];
        }

        colorFltOrder = other.colorFltOrder;

        settings = other.settings;
        lateReverbCharact = other.lateReverbCharact;

        renderPredelay = other.renderPredelay;
        MIN_ERROR = other.MIN_ERROR;

        outputType = other.outputType;
        switch (outputType) {
        case IIRFilterType::Polynomial: {
            for (int i = 0; i < 2; ++i) {
                for (int j = 0; j < MAX_FILTERORDER + 1; ++j) {
                    filters.colorFltCoeff[i][j] = other.filters.colorFltCoeff[i][j];
                }
            }
        } break;
        case IIRFilterType::SOSs: {
            for (int i = 0; i < 2; ++i)
                filters.sosSequence[i] = other.filters.sosSequence[i];
        } break;
        default:
            break;
        }
        return *this;
    }
};

void fillBaseConfig(SetBaseConfigStruct* baseConfig, unsigned int nrFbLoops, float predelayFactor,
    float delayRatio, float delayMinMs, const float fs);
void fillBaseConfigOut(SetBaseConfigStruct* SBC, ReverbDSPCfgStruct* RevDSPCfg,
    const int16_t t60FltOrder, const int16_t colorFltOrder);

int32_t reverbSetBaseConfig(
    SetBaseConfigStruct* pSetIn, SetBaseConfigStruct* pSetOut, bool isDefaultAE);
void preprocessReverbInputParameters(lateReverbInParms* pIn, lateReverb* pOut);

void fillT60Config(ReverbDSPCfgStruct* RevDSPCfg, T60Params* T60Config);
bool checkT60Configuration(T60Params* pIn);
void reverbAddT60Filters(T60Params* pIn, T60Settings* pOut, IIRFilterType type);
void fillT60ConfigOut(T60Settings* T60Config, ReverbDSPCfgStruct* RevDSPCfg);
void fillColorationConfig(ReverbDSPCfgStruct* RevDSPCfg, ColorParams* ColorConfig, float* DSR,
    float* RT60, uint16_t predelay);
bool checkColorationInputConfiguration(ColorParams* pInColorationStruct);
int32_t reverbCalculateLevelData(
    ColorParams* pInColorationStruct, ColorParams* pOutColorationStruct, bool isDefaultAE);
void reverbAddColorFilters(
    ColorParams* pInColorationStruct, ColorParams* pOutColorationStruct, IIRFilterType type);
void fillColorationConfigOut(ColorParams* ColorConfig, ReverbDSPCfgStruct* RevDSPCfg);

void smoothResponse(float* X, float* f, float* smoothed, int16_t length, float Noct);

mpegcore::GraphicEqParameters calcDirectivityFilter(ReverbDSPCfgStruct* RevDSPCfg,
    MPEGI::DirectivitySpectrum dirFltTarget, IIRFilterType type, uint16_t dirFltOrder,
    float maxErrordB);
