/***************************************************************************
This software module was originally developed jointly by

Nokia Technologies Ltd. and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ReverbStage.h"
#include "Reverb.hpp"
#include "mpegcore/datatypes.hpp"
#include "mpegcore/datatypes_reverb.hpp"
#include <EAR/BaseTypes.h>
#include <EAR/Exception.h>
#include <EAR/Listener.h>
#include <EAR/ListenerVoiceSource.h>
#include <EAR/Log.h>
#include <EAR/PointSource.h>
#include <EAR/Portal.h>
#include <EAR/RealFrame.h>
#include <EAR/SceneObject.h>
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Scene/SceneState.h>
#include <MPEGI/Stream/StreamBuffer.h>
#include <algorithm>
#include <cmath>
#include <iostream>
#include <mpegcore/ReverberatorParameters.hpp>
#include <mpegcore/algorithm.hpp>
#include <mpegcore/chrono.hpp>
#include <mpegcore/frontend/GraphicEqFilter.hpp>
#include <mpegcore/frontend/neuralNetworkParameters.hpp>
#include <set>
#include <spdlog/fmt/fmt.h>
#include <string>
#include <tuple>
#include <unordered_map>
#include <utility>
#include <vector>

#if defined(EAR_ENABLE_IPP)
#include <ipp.h>
#include <ipps.h>
#endif

namespace MPEGI {

std::shared_ptr<RenderStage> makeReverbStage()
{
    return std::make_shared<ReverbStage>();
}

namespace {

inline float getMaterialGainOut(EAR::Material& material)
{
    return material.getTransmissionCoeffForFreq(1000) + material.getCouplingCoeffForFreq(1000);
}

inline float getMaterialGainIn(EAR::Material& material)
{
    return material.getTransmissionCoeffForFreq(1000);
}

bool matchEnvironmentIds(
    const std::array<std::string, 2>& lhs, const std::array<std::string, 2>& rhs)
{
    // Check if the two entries are exactly the same, or they are reversed.
    return (lhs[0] == rhs[0] && lhs[1] == rhs[1]) || (lhs[1] == rhs[0] && lhs[0] == rhs[1]);
}

bool matchOneEnvironmentId(const std::array<std::string, 2>& lhs, const std::string& rhs)
{
    if (lhs[0] == rhs) {
        return true;
    } else if (lhs[1] == rhs) {
        return true;
    } else {
        return false;
    }
}

bool isReverbInputRenderItem(RenderItem const& item)
{
    return (item.type == ItemType::Primary || item.type == ItemType::HOA
               || item.type == ItemType::ListenerVoice)
        && !(item.status == ItemStatus::Inactive || item.status == ItemStatus::Orphaned)
        && !(item.noReverb);
}

template <typename T> constexpr double delayInSamples(T const& delay, std::uint32_t sampleRateHz)
{
    return static_cast<double>(delay.count()) * sampleRateHz * T::period::num
        / static_cast<double>(T::period::den);
}

// Templates for using Intel IPP with standard C++ fallback implementations
void addProductCVectors(const float* srcPtr, float multiplier, float* dstPtr, std::size_t size)
{
#if defined(EAR_ENABLE_IPP)
    IppStatus status = ippsAddProductC_32f(static_cast<const Ipp32f*>(srcPtr),
        static_cast<Ipp32f>(multiplier), static_cast<Ipp32f*>(dstPtr), size);
    EAR_CHECK(status == ippStsNoErr, "ReverbStage: ippsAddProductC_32f failed");
#else
    for (std::size_t i{ 0 }; i < size; ++i) {
        dstPtr[i] += multiplier * srcPtr[i];
    }
#endif
}

void addFloatVectors(const float* srcPtr, float* dstPtr, std::size_t size)
{
#if defined(EAR_ENABLE_IPP)
    IppStatus status =
        ippsAdd_32f_I(static_cast<const Ipp32f*>(srcPtr), static_cast<Ipp32f*>(dstPtr), size);
    EAR_CHECK(status == ippStsNoErr, "ReverbStage: ippsAdd_32f_I failed");
#else
    for (std::size_t i{ 0 }; i < size; ++i) {
        dstPtr[i] += srcPtr[i];
    }
#endif
}

// This template operates on a std::vector in order to use std::fill as fallback solution
void zeroVector(std::vector<float>& dst)
{
#if defined(EAR_ENABLE_IPP)
    IppStatus status = ippsZero_32f(static_cast<Ipp32f*>(dst.data()), dst.size());
    EAR_CHECK(status == ippStsNoErr, "ReverbStage: ippsZero_32f failed");
#else
    std::fill(dst.begin(), dst.end(), 0.0f);
#endif
}

// Return the size of the predelay delay line and the minimum delay line length (in samples).
// The `predelay` argument here means the time that the first echo from the FDN is output.
// So the returned delay line size is `predelay` minus the length of the shortest among the delay lines.
// If predelay doesn't contain a value, it's 0.
mpegcore::Predelay adjustPredelay(std::optional<float> predelay,
    mpegcore::ReverberatorParameters const& reverbParams, float sampleRateHz)
{
    auto const minDelayLineDelay = minDelay(reverbParams.delayLines, sampleRateHz);
    std::size_t minDelayLineDelayInSamples = delayInSamples(minDelayLineDelay, sampleRateHz);
    if (!predelay.has_value()) {
        return std::make_pair(0, minDelayLineDelayInSamples);
    }
    auto const predelay_sec = predelay.value();
    // A zero predelay can mean either that a scene really had defined a predelay of zero or that
    // the encoder used a default value of zero which got propagated here. Either way we dismiss the
    // value.
    if (predelay_sec <= 0) {
        return std::make_pair(0, minDelayLineDelayInSamples);
    }

    auto const minDelayLine_nanosec = minDelay(reverbParams.delayLines, sampleRateHz);
    auto const predelay_nanosec = mpegcore::secondsToNanoseconds(predelay_sec);
    auto const predelayAdjusted_nanosec = predelay_nanosec - minDelayLine_nanosec;

    if (predelayAdjusted_nanosec.count() < 0) { // predelay < mindelayline
        // For the warning, tolerate rounding error within 1/100th of a ms
        if (round(predelay_nanosec.count() / 1e4) < round(minDelayLine_nanosec.count() / 1e4)) {
            const auto warning = fmt::format("Warning [ReverbStage::adjustPredelay]: "
                                             "Reverb rendering pre-delay {} ms is shorter than "
                                             "shortest loop delay {} ms, so predelay will be zero.",
                predelay_nanosec.count() / 1e6, minDelayLine_nanosec.count() / 1e6);
            std::cout << warning << std::endl;
        }
        return std::make_pair(0, minDelayLineDelayInSamples);
    }

    auto const predelayAdjusted_samples = delayInSamples(predelayAdjusted_nanosec, sampleRateHz);

    return std::make_pair(predelayAdjusted_samples, minDelayLineDelayInSamples);
}

// Return the size of the predelay delay line and the minimum delay line length (in samples).
// In contrast to adjustPredelay(), the requestedDiffuseOnsetSec argument means the time that the
// FDN's response is diffuse.
// So the returned delay line size is the difference between the requestedDiffuseOnsetSec and the
// FDN's diffuse onset (or zero if fdnDiffuseOnset > requestedDiffuseOnset), minus the length of
// the shortest among the delay lines.
mpegcore::Predelay getPredelayForDiffuseOnsetAlignment(float requestedDiffuseOnsetSec,
    mpegcore::ReverberatorParameters const& reverbParams, float sampleRateHz)
{
    float fdnDiffuseOnset_sec = reverbParams.fdnDiffuseOnsetSec;
    float diffuseOnsetDelayAdjustment_sec = requestedDiffuseOnsetSec - fdnDiffuseOnset_sec;

    auto const minDelay_nano = minDelay(reverbParams.delayLines, sampleRateHz);
    std::size_t minDelay_samp = delayInSamples(minDelay_nano, sampleRateHz);

    // If the requested diffuse onset is shorter than the FDN's inherent diffuse onet,
    // we can't achieve the requested diffuse onset, so return predelay of 0.
    if (diffuseOnsetDelayAdjustment_sec < 0) {
        const auto warning =
            fmt::format("FDN's diffuseOnset ({:.2f} ms) is later than the "
                        "requested diffuse onset ({:.2f} ms), so predelay line length will be 0.",
                fdnDiffuseOnset_sec * 1e3, requestedDiffuseOnsetSec * 1e3);
        std::cerr << warning << std::endl;
        return std::make_pair(0, minDelay_samp);

    } else {
        auto const renderPredelay_samp = static_cast<std::size_t>(delayInSamples(
            mpegcore::secondsToNanoseconds(diffuseOnsetDelayAdjustment_sec), sampleRateHz));

        return std::make_pair(renderPredelay_samp, minDelay_samp);
    }
}

std::size_t connectedPredelay(std::pair<std::size_t, std::size_t> predelayAndMinDelay)
{
    auto [predelay, minDelay] = predelayAndMinDelay;
    return std::max(minDelay, (predelay + minDelay) / 8);
}

float secondOrderReverbPredelay(const mpegcore::Predelay& subj, const mpegcore::Predelay& remote)
{
    return (subj.first + remote.first) / 2;
}

int getLosVoxelIndex(const EAR::Vector3& pos, const LineOfSightArea& area)
{
    // Calculate voxel index from item position and voxel area dimensions
    const auto& p1 = area.losAreaCornerP1;
    const auto& nVox = area.nVoxels;
    std::array<int, 3> idx{ -1, -1, -1 };
    for (int i = 0; i < 3; i++) {
        // Everything concerning positions is in integer millimeter arithmetic here
        // in order to avoid floating point rounding errors.
        constexpr float M_TO_MM{ 1000.f };
        constexpr float CM_TO_MM{ 10.f };
        int mmvoxelSize = area.voxelSize * CM_TO_MM;
        int mmpos = static_cast<int>(std::lround(pos[i] * M_TO_MM));
        int mmp1 = static_cast<int>(std::lround(p1[i] * M_TO_MM));
        int nearest = std::abs((mmpos - mmp1) / static_cast<float>(mmvoxelSize));
        // Round down to nearest voxel if at the very edge of the area.
        idx[i] = std::min(nearest, (nVox[i] - 1));
    }
    int voxIdx = idx[1] * nVox[2] * nVox[0] + idx[2] * nVox[0] + idx[0];
    return voxIdx;
}

// Returns a second order reverberation coefficient that represents the ratio of reverberation
// that would leak through a portal to another environment where the listener resides to the
// total reverberated energy in the source environment.
float getSecondOrderReverbCoefficient(std::vector<const MPEGI::Portal*> portals)
{
    float coeff{ 0.f };
    for (const auto* portal : portals) {
        if (portal == nullptr) {
            continue;
        }
        // Apply transmission coefficient of the portal material
        std::shared_ptr<EAR::Material> material = portal->scenePortal->getMaterial();
        float materialGainOut = 1.0f;
        if (material != nullptr) {
            // Energy going through the portal, out of the environment, includes coupling coefficient.
            materialGainOut = getMaterialGainOut(*material);
        }
        float area = portal->bbPortalHeight * portal->bbPortalWidth;
        coeff += materialGainOut * area / (16 * M_PI);
    }
    // Convert to signal gain by taking root and clamp value between 0.0 and 1.0.
    return std::max(std::min(std::sqrt(coeff), 1.f), 0.f);
}

} // namespace

ReverbStage::ReverbStage(
    const int16_t dirFltOrder, const int16_t t60FltOrder, const int16_t colorFltOrder)
    : dirFltOrder(dirFltOrder)
    , t60FltOrder(t60FltOrder)
    , colorFltOrder(colorFltOrder)
    , acousticEnvironmentData(DEFAULT_REVERB_EIF_PREDELAY_FACTOR,
          DEFAULT_REVERB_DISTANCE_GAIN_DROP_DB, DEFAULT_REVERB_MINIMUM_DISTANCE, 1.0f,
          DEFAULT_REVERB_DELAY_RATIO_IN_DEFAULT_AE, 0.0f)
    , defaultAcousticEnvironmentData(DEFAULT_REVERB_PREDELAY_FACTOR_IN_DEFAULT_AE,
          DEFAULT_REVERB_DISTANCE_GAIN_DROP_DB_IN_DEFAULT_AE,
          DEFAULT_REVERB_MINIMUM_DISTANCE_IN_DEFAULT_AE, DEFAULT_REVERB_GAIN_IN_DEFAULT_AE,
          DEFAULT_REVERB_DELAY_RATIO_IN_DEFAULT_AE, DEFAULT_REVERB_DELAY_MINIMUM_MS_IN_DEFAULT_AE)
{
    AcEnvProp.reserve(MAX_ROOMS);
}

void ReverbStage::processBlock(DspContext ctx)
{
    auto& currentDSP = ctx.getData<ReverbStage::DspData>();
    const auto& dspReverbData = currentDSP.reverbData;

    for (const auto& [acousticEnvironmentId, reverbUnit] : dspReverbData) {
        // Create a monomix of reverb inputs
        for (auto& predelay : reverbUnit.reverb->predelays) {
            // Input buffer for each predelay is cleared
            std::fill_n(begin(predelay.inputAudio), B, 0.F);
        }
        if (m_directivityFilterMap.find(acousticEnvironmentId) != m_directivityFilterMap.end()) {
            for (const auto& [ptr, filter] : m_directivityFilterMap.at(acousticEnvironmentId)) {
                filter->inputBuffer->getFrame()[0].fill(0.0f);
            }
        }
        directivityBusAdded.clear();
        for (auto& item : currentDSP.dspItems[acousticEnvironmentId]) {
            // Audio sources may have different predelay values depending on the reverb environment
            // that the audio emanates from. Sources are mixed with items sharing the same predelay
            // values to separate input buses.
            std::size_t predelay = reverbUnit.predelayInSamples.first;
            if (item.sourceAcousticEnvironmentId != acousticEnvironmentId) {
                auto const& it = reverbData.find(item.sourceAcousticEnvironmentId);
                if (it != reverbData.end()) {
                    predelay += connectedPredelay(it->second.predelayInSamples);
                }
            }
            auto* reverbInputBusPtr = reverbUnit.reverb->getInputBuffer(predelay).data();
            float const gain = getItemGainBasedOnAE(item, reverbUnit);
            for (std::size_t ch{ 0 }; ch < item.inputBuffer->getFrame().channels(); ++ch) {
                // All channels of multichannel inputs are added to the reverb bus
                // (assumes HOA signals to be in the ESD format)
                float const* srcPtr = item.inputBuffer->getFrame()[ch].data();
                // Sources without directivity filter are summed directly to reverb input bus
                if (item.directivityFilter == nullptr) {
                    for (std::size_t i{ 0 }; i < B; ++i) {
                        reverbInputBusPtr[i] += gain * srcPtr[i];
                    }
                } else {
                    float* directivityBusPtr =
                        item.directivityFilter->inputBuffer->getFrame()[0].data();
                    for (std::size_t i{ 0 }; i < B; ++i) {
                        directivityBusPtr[i] += gain * srcPtr[i];
                    }
                    // Making a note that directivity filter with specific predelay is in use
                    directivityBusAdded[item.directivityFilter].insert(predelay);
                }
            }
        }
        if (m_directivityFilterMap.find(acousticEnvironmentId) != m_directivityFilterMap.end()) {
            for (const auto& [directivityId, directivityFilter] :
                m_directivityFilterMap.at(acousticEnvironmentId)) {
                if (directivityBusAdded.count(directivityFilter)) {
                    auto& filter = directivityFilter->filter;
                    float const* directivityInBusPtr =
                        directivityFilter->inputBuffer->getFrame()[0].data();
                    // Apply directivity filter to each buffer that has a different predelay
                    for (std::size_t predelay : directivityBusAdded.at(directivityFilter)) {
                        auto* reverbInputBusPtr =
                            reverbUnit.reverb->getInputBuffer(predelay).data();
                        for (std::size_t i{ 0 }; i < B; ++i) {
                            reverbInputBusPtr[i] += filter->filter(directivityInBusPtr[i]);
                        }
                    }
                }
            }
        }

        bool sorEnabled = (acousticEnvironmentId == currentDSP.listenerAcousticEnvironmentId)
            || m_unrestrictedSor;
        if (currentDSP.sorGain > 0.f && sorEnabled) {
            // Apply second order reverberation from buffered outputs to listener's reverberator input
            for (const auto& [id, coeff] : *reverbUnit.connectedEnvironments) {
                const auto& connectedReverbIt = dspReverbData.find(id);
                if (connectedReverbIt == dspReverbData.end()) {
                    continue;
                }
                const auto& connectedEnv = connectedReverbIt->second;
                const std::size_t predelay = secondOrderReverbPredelay(
                    reverbUnit.predelayInSamples, connectedEnv.predelayInSamples);

                const float* source = connectedEnv.previousOutput.get()->data();
                float adjustedCoeff = coeff * currentDSP.sorGain;
                float* output = reverbUnit.reverb->getInputBuffer(predelay).data();
                addProductCVectors(source, adjustedCoeff, output, B);
            }
        }

        reverbUnit.reverb->processBlock();
    }

    // Store output for next update cycle for second order reverberation. This needs to be done
    // after all acoustic environments have been processed, but before gain is applied to any
    // of the reverb outputs.
    for (const auto& reverbData : dspReverbData) {
        savePreviousReverbOutput(reverbData.second);
    }

    for (const auto& [acousticEnvironmentId, reverbUnit] : dspReverbData) {
        // Copy reverb output to the corresponding output signal
        auto& frame = reverbUnit.outputRenderItem->signal->getFrame();
        for (auto ch{ reverbUnit.reverb->outputAudio.size() }; ch-- > 0;) {
            // Apply gain to output signal
            auto* output = reverbUnit.reverb->outputAudio[ch].data();
#if defined(EAR_ENABLE_IPP)
            ippsMulC_32f_I(static_cast<Ipp32f>(reverbGain), static_cast<Ipp32f*>(output), B);
#else // fallback multiply
            for (std::size_t i{ 0 }; i < B; ++i) {
                output[i] *= reverbGain;
            }
#endif
            frame[ch].write(output, 0, B);
        }
    }

    // Mix signals from all audible reverb environments to head tracked output
    // that is zeroed by default.
    for (const auto& output : headtrackedOutputRenderItems) {
        output->signal->getFrame()[0].fill(0.F);
    }
    // Reverbs are mixed together based on gain coefficients that are calculated during
    // every update phase.
    for (const auto& [acousticEnvironmentId, reverbUnit] : dspReverbData) {
        const float gain = reverbUnit.gain;
        if (gain > 0.0f) {
            auto const& reverbOutput = reverbUnit.reverb->outputAudio;
            std::vector<std::size_t> outputIndices = outputMapping[acousticEnvironmentId];
            for (auto ch{ reverbOutput.size() }; ch-- > 0;) {
                auto& output =
                    headtrackedOutputRenderItems[outputIndices[ch]]->signal->getFrame()[0];
                const auto& source = reverbOutput[ch];
                addProductCVectors(source.data(), gain, output.data(), B);
            }
        }
    }
}

ReverbStage::~ReverbStage() {}

void ReverbStage::SetAcousticEnvironmentProperties(const SceneState* state)
{
    AcEnvProp.resize(state->getReverbEnvs().size());

    unsigned int revParsIdx = 0;

    for (auto itr = reverbBsParameters.cbegin(); itr != reverbBsParameters.cend();
         ++itr, ++revParsIdx) {

        auto i = std::find_if(
            state->getReverbEnvs().begin(), state->getReverbEnvs().end(), [itr](const auto& v) {
                return v->acousticEnvironmentBsId.value() == (itr->second)->id;
            });
        if (i == state->getReverbEnvs().end())
            assert(false && "Acoustic environment mapping from P27 to P13 is wrong!");

        const std::unique_ptr<MPEGI::ReverbEnv>& revEnv = *i;

        AcEnvProp[revParsIdx].roomId = revEnv->acousticEnvironmentId;

        if (revEnv->enclosure != nullptr) {
            AcEnvProp[revParsIdx].dimensions[0] = revEnv->enclosure->getBoundingBox()[0].high
                - revEnv->enclosure->getBoundingBox()[0].low;
            AcEnvProp[revParsIdx].dimensions[1] = revEnv->enclosure->getBoundingBox()[1].high
                - revEnv->enclosure->getBoundingBox()[1].low;
            AcEnvProp[revParsIdx].dimensions[2] = revEnv->enclosure->getBoundingBox()[2].high
                - revEnv->enclosure->getBoundingBox()[2].low;
        }
    }
    // To be on the safe side, run max 10 iterations in case first update round(s)
    // would not be successful
    for (std::size_t id = 0; id < 10; id++) {
        DimensionUpdated();
    }
}

// Prevent integer ratios of room dimensions
bool ReverbStage::DimensionUpdated()
{
    bool updated = false;
    for (int i = 0; i < AcEnvProp.size(); i++) {
        float* roomDimi = AcEnvProp[i].dimensions;
        for (int j = 0; j < AcEnvProp.size(); j++) {
            if (j == i) {
                continue;
            }
            float* roomDimj = AcEnvProp[j].dimensions;
            for (int xid = 0; xid < 3; xid++) {
                float r = 1.0f;
                if (roomDimj[xid] > roomDimi[xid]) {
                    r = roomDimj[xid] / roomDimi[xid];
                } else {
                    r = roomDimi[xid] / roomDimj[xid];
                }
                float intRatio = roundf(r);
                if (fabs(r - intRatio) < 0.01) {
                    AcEnvProp[j].dimensions[xid] *= 1.02f;
                    updated = true;
                }
            }
        }
    }
    return updated;
}

std::pair<float, float> ReverbStage::GetSourceContribution(
    RenderItem& item, std::string acousticEnvironmentId)
{
    std::pair<float, float> sourceContributionPortion{ 0.f, 0.0f };

    if (acousticEnvironmentId == item.acousticEnvironmentId) {
        // Items inside the currently considered environment contribute fully to it (lost energy to external environments will be compensated later)
        sourceContributionPortion = { 1.f, 1.f };
    } else {
        // Calculate coefficient or get it from cache
        auto& cached = m_coeffCache[acousticEnvironmentId][item.idx];
        // Checking material change is complex so it is the last one to be checked after cache
        // existence and item position change checks.
        if (!cached.has_value() || item.changed & (ItemProperty::Position | ItemProperty::Status)
            || m_connectionResolver.portalMaterialChanged(
                item.acousticEnvironmentId, acousticEnvironmentId)) {
            // Items in other environment than the currently considered environment may contribute partially.
            sourceContributionPortion = getExternalSourceContribution(item, acousticEnvironmentId);
            cached = sourceContributionPortion;
        } else {
            // A stationary item will use cached value if portal materials have not changed.
            sourceContributionPortion = cached.value();
        }
    }

    return sourceContributionPortion;
}

std::pair<float, float> ReverbStage::sourceContributionToDownmixGain(
    float sourceContributionPortion, float dist, const MPEGI::RenderItem& item,
    const ReverbData& acousticEnv, const std::string& acousticEnvironmentId)
{ // If source contributes to reverberation, add it to the dspList for downmixing
    std::pair<float, float> gains = { 0.f, 0.f };
    if (sourceContributionPortion > 0.f) {
        // Find out if this is a VR environment. In this case we can find it from the list of revDSPCfg.
        auto reverb = RevDSPCfg.find(acousticEnv.bitstreamReverbId);
        const ReverbDSPCfgStruct* revDSPCfg = nullptr;
        if (reverb != RevDSPCfg.end()) {
            revDSPCfg = &reverb->second;
        }

        // Find out if this is a default acoustic environment
        bool isDefaultAE = false;
        if (acousticEnvironmentId == "defaultAE" && acousticEnv.enclosure == nullptr) {
            isDefaultAE = true;
        }

        std::optional<float> commonComponentClrFlt{ std::nullopt };
        if (revDSPCfg) {
            // Calculate the downmix gain for this item, consisting of accumulated gain in previous stages
            // (typically the source's pregain at this point), a (bitstream enabled) distance-dependent
            // gain where reverb contribution is reduced as sources are further away, common components
            // from filters transferred to the broadband gains.

            commonComponentClrFlt = revDSPCfg->Reverberator.commonComponentClrFlt
                * std::sqrt(sourceContributionPortion);
        }

        gains = calculateFilteredItemGains(dist, item, commonComponentClrFlt, isDefaultAE);
    }

    return gains;
}

std::pair<float, float> ReverbStage::getDownmixGainForInternalSources(RenderItem& item,
    RenderList& renderList, const ReverbData& acousticEnv, std::string acousticEnvironmentId,
    std::string listenerAcousticEnvironmentId, bool isExplicitPortalMode,
    std::vector<float>& totalExternalContributions, int itemIdx)
{
    std::pair<float, float> gains = { 0.0f, 0.0f };

    // Calculate which portion of the source's energy reaches the portal
    std::pair<float, float> sourceContributionPortion;
    if (isExplicitPortalMode) {
        sourceContributionPortion = GetSourceContribution(item, acousticEnvironmentId);
    } else {
        sourceContributionPortion = { 1.0f, 1.0f };
    }

    auto sourceLocation = item.position.location;
    // No distance gain for external sources, set to reference value by default.
    float dist = item.refDistance;

    // Distance gain attenuation is applied for internal sources.
    if (listenerAcousticEnvironmentId == item.acousticEnvironmentId) {
        // When item is in the same environment as listener, or the reverberated space is
        // other than listener's space, apply normal distance gain.
        dist = (sourceLocation - listenerLocation).abs();
    } else {
        // When item is in another environment, calculate distance gain based on item distance to nearest portal.
        const MPEGI::Portal* closestPortal =
            m_connectionResolver.closestPortal(listenerLocation, item.acousticEnvironmentId);
        if (closestPortal != nullptr) {
            dist = (closestPortal->position - sourceLocation).abs();
            // Sudden large changes to distance can be caused by listener moving to another environment.
            // Apply large changes to distance gradually by moving distance towards target in increments.
            auto cachedDist = m_itemDistGain.find(item.idx);
            if (cachedDist != m_itemDistGain.end()) {
                // Apply sudden distance changes over several update cycles. Save intermediate values in
                // m_itemDistGain.
                float prevDist = cachedDist->second;
                if (std::abs(dist - prevDist) > DISTANCE_GAIN_SMOOTHING_THRESHOLD) {
                    if (dist > prevDist) {
                        dist = prevDist + DISTANCE_GAIN_SMOOTHING_THRESHOLD;
                    } else if (dist < prevDist) {
                        dist = prevDist - DISTANCE_GAIN_SMOOTHING_THRESHOLD;
                    }
                } // Else: change smaller than step distance is applied instantly
            } // Else: no cached value present, apply distance instantly
        } else {
            // No portal connection to other environment, apply normal distance gain.
            dist = (sourceLocation - listenerLocation).abs();
        }
    }
    // Save distance for next update round in order to apply gradual change
    m_itemDistGain[item.idx] = dist;

    // Subtract all energy that has been lost to other environments
    sourceContributionPortion.second -= totalExternalContributions[itemIdx];

    gains = sourceContributionToDownmixGain(
        sourceContributionPortion.second, dist, item, acousticEnv, acousticEnvironmentId);

    return gains;
}

float max_withSign(const float x, const float y)
{
    // This function represents: max_withSign = sign(x) * max(abs(x), abs(y))
    const float y_abs = fabs(y);

    return (x < 0 ? fmin(x, -y_abs) : fmax(x, y_abs));
}

float computePortalContribution(const MPEGI::Portal& portal, MPEGI::RenderItem& item, float eps,
    const std::unordered_map<unsigned int, EAR::HomogenMatrix>& portalRotationMatrices)
{
    EAR::Vector3 portalPosition = portal.position;

    // Calculating vector from source to portal
    EAR::Vector3 srcPortalVec = item.position.location - portalPosition;

    // We operate on portals aligned with the XY-plane
    EAR::Vector3 portalNormal = { 0.f, 0.f, -1.f };
    // Flip the direction of the normal vector if the item is inside environment2 (by default the normal points into environment 1).
    if (item.acousticEnvironmentId.compare(portal.portalEnvironments[1]) == 0) {
        portalNormal *= -1.f;
    }

    // Rotate the source-portal vector with rotation that aligns the portal normal with [0 0 -1]
    const EAR::HomogenMatrix& rotMatrix = portalRotationMatrices.at(portal.portalId);
    srcPortalVec = rotMatrix * srcPortalVec;

    const float src2PortalDist_azi =
        fmax(sqrt(srcPortalVec[0] * srcPortalVec[0] + srcPortalVec[2] * srcPortalVec[2]), eps);
    const float src2PortalDist_ele =
        fmax(sqrt(srcPortalVec[1] * srcPortalVec[1] + srcPortalVec[2] * srcPortalVec[2]), eps);

    // Calculating cosine of the angle between the portal's normal and the source-portal vector in the horizontal plane.
    const float numA = srcPortalVec[0] * portalNormal[0] + srcPortalVec[2] * portalNormal[2];
    const float denA = src2PortalDist_azi
        * sqrt(portalNormal[0] * portalNormal[0] + portalNormal[2] * portalNormal[2]);
    float cosPhiA = copysignf(fmax(fabs(numA), eps), numA) / fmax(denA, eps);

    // Calculating cosine of the purely vertical angle between the portal's normal and the source-portal vector.
    const float numE = srcPortalVec[1] * portalNormal[1] + srcPortalVec[2] * portalNormal[2];
    const float denE = src2PortalDist_ele
        * sqrt(portalNormal[1] * portalNormal[1] + portalNormal[2] * portalNormal[2]);
    float cosPhiE = copysignf(fmax(fabs(numE), eps), numE) / fmax(denE, eps);

    // Ensure that acos(cosPhiA) and acos(cosPhiE) are in a range of 0..1/2*pi (0 .. 90 degrees).
    // Otherwise they are on the wrong side of the portal, clipping negative values will ensure 0 source contribution.
    cosPhiA = fmin(fmax(cosPhiA, 0.0f), 1.0f);
    cosPhiE = fmin(fmax(cosPhiE, 0.0f), 1.0f);

    // Calculate azimuthal and elevational angle ranges that the portal covers on a circle centered on
    // the source position and going through the portal position.

    const float W = portal.bbPortalWidth;
    const float H = portal.bbPortalHeight;

    float thetaA = atan(src2PortalDist_azi * W
        / max_withSign(src2PortalDist_azi * src2PortalDist_azi - 0.25f * W * W, eps * eps)
        * fmax(cosPhiA, eps));

    if (thetaA < 0.0f) {
        thetaA += M_PI;
    }
    float thetaE = atan(src2PortalDist_ele * H
        / max_withSign(src2PortalDist_ele * src2PortalDist_ele - 0.25f * H * H, eps * eps)
        * fmax(cosPhiE, eps));

    if (thetaE < 0.0f) {
        thetaE += M_PI;
    }

    // Calculate source energy contribution
    const float portalContribution = thetaA * sin(0.5 * thetaE) / (2 * M_PI);

    // Save portal contribution for later processing (Portal stage).
    auto& state = item.getStateOrCreate<ReverbStage::ItemState>(false);
    state.portalContributions[portal.portalId] = portalContribution;

    return portalContribution;
}

std::pair<float, float> computeSourceContribution(const MPEGI::Portal& portal,
    MPEGI::RenderItem& item, float eps,
    const std::unordered_map<unsigned int, EAR::HomogenMatrix>& portalRotationMatrices)
{
    // Calculate source energy contribution
    const float portalContribution =
        computePortalContribution(portal, item, eps, portalRotationMatrices);

    // Determing door material impact.
    // TODO: Consider creating a filter per material and configure that for the downmixer?
    std::shared_ptr<EAR::Material> material = portal.scenePortal->getMaterial();

    // NOTE: If the material can't be found (or none is specified), materialGain is set to 1.0f
    float materialGainOut = 1.0f;
    float materialGainIn = 1.0f;
    if (material != nullptr) {
        // Energy going through the portal, out of the environment, includes coupling coefficient.
        materialGainOut = getMaterialGainOut(*material);
        // Energy actually lost to the other environment excludes the coupling coefficient, as coupling energy contributes to both connected environments.
        materialGainIn = getMaterialGainIn(*material);
    }

    return { portalContribution * materialGainOut, portalContribution * materialGainIn };
}

std::unique_ptr<RenderStage::DspData> ReverbStage::update(RenderList& renderList)
{
    if (B == 0) { // only update once initialized
        return {};
    }

    setConfigurableVariables();

    auto dspData = std::make_unique<ReverbStage::DspData>();
    auto& dspList = *dspData;

    bool isExplicitPortalMode = false;
    if (bitstreamPortalPayload) {
        isExplicitPortalMode = bitstreamPortalPayload->getExplicitPortalMode();
    }

    listenerLocation = renderList.listener.position.location;
    listenerAcousticEnvironmentId = renderList.listener.acousticEnvironmentId;
    dspList.listenerAcousticEnvironmentId = listenerAcousticEnvironmentId;
    float gainThreshold = 1e-6;

    std::vector<float> totalExternalContributions;
    totalExternalContributions.resize(renderList.items.size(), 0.0f);

    // From the render items reaching this stage, creates a list of items that contribute to the reverberation.
    // First pass: calculate all reverb contributions from external sources (and keep track how much of the sources' energy is going to external environments)
    if (isExplicitPortalMode) {
        for (const auto& [acousticEnvironmentId, acousticEnv] : reverbData) {
            for (size_t itemIdx = 0; itemIdx < renderList.items.size(); ++itemIdx) {
                const auto& item = renderList.items[itemIdx];
                if (item != nullptr && !isReverbInputRenderItem(*item)) {
                    continue;
                }
                if (acousticEnvironmentId != item->acousticEnvironmentId) {
                    // Calculate which portion of the source's energy reaches the portal
                    std::pair<float, float> sourceContributionPortion;
                    if (isExplicitPortalMode) {
                        sourceContributionPortion =
                            GetSourceContribution(*item, acousticEnvironmentId);
                    } else {
                        sourceContributionPortion = { 1.0f, 1.0f };
                    }

                    totalExternalContributions[itemIdx] += sourceContributionPortion.second;
                }
            }
        }
    }

    // If there are portals to 'outside', and no reverb is associated with it, calculate how much
    // energy sources lose to that implicit environment.
    if (calcOutsideEnergy) {
        std::string outsideAcousticEnvironmentId = "";

        for (size_t itemIdx = 0; itemIdx < renderList.items.size(); ++itemIdx) {
            const auto& item = renderList.items[itemIdx];
            if (item->acousticEnvironmentId != outsideAcousticEnvironmentId) {
                // Calculate which portion of the source's energy reaches the portal
                std::pair<float, float> sourceContributionPortion =
                    GetSourceContribution(*item, outsideAcousticEnvironmentId);

                totalExternalContributions[itemIdx] += sourceContributionPortion.second;
            }
        }
    }

    if (isExplicitPortalMode) {
        for (size_t itemIdx = 0; itemIdx < renderList.items.size(); ++itemIdx) {
            const auto& item = renderList.items[itemIdx];
            if (item != nullptr && !isReverbInputRenderItem(*item)) {
                continue;
            }

            const auto& portals = bitstreamPortalPayload->getPortals();
            const auto& p2pFactors = bitstreamPortalPayload->getPortal2PortalFactors();

            // Calculate all reverb contribution to external AEs
            for (const auto& [acousticEnvironmentId, acousticEnv] : reverbData) {
                // Only external contributions
                if (item->acousticEnvironmentId != acousticEnvironmentId) {
                    // For every item + AE combination, compute all paths connecting the item to the AE and sum the contributions of each path
                    const auto allPaths =
                        getAllPaths(item->acousticEnvironmentId, acousticEnvironmentId, portals);
                    float totalContribution = 0.f;
                    for (const auto& path : allPaths) {
                        EAR_CHECK(path.size() > 1,
                            "Error: path of size 1 detected where it shouldn't be possible"); // Paths of size 1 are only possible if item->acousticEnvironmentId == acousticEnvironmentId
                        // Attenuation for sources with no line of sight to first portal on the path
                        float lineOfSightModifier = 1.0f;

                        // FirstSourceContribution relates to how much energy the first portal receives from the render item
                        // TODO: computePortalContribution is computed multiple times for the same item-portal combination.
                        float firstSourceContribution = 0.f;
                        if (isExplicitPortalMode) {
                            // Determine line of sight between the item and the first portal on the path
                            if (!m_connectionResolver.lineOfSightExists(item->position.location,
                                    item->acousticEnvironmentId,
                                    path[1].incomingPortal->portalId)) {
                                lineOfSightModifier = m_noLineOfSightCoeff;
                            }
                            if (lineOfSightModifier > 0.f) {
                                firstSourceContribution = lineOfSightModifier
                                    * computePortalContribution(*path[1].incomingPortal, *item,
                                        1e-9, portalRotationMatrices);
                            }
                        } else {
                            firstSourceContribution = 1.f;
                        }

                        // getHigherOrderSourceContribution tallies up the material properties of all the portals, as well as the portal2portal factors.
                        const float pathContribution = firstSourceContribution
                            * getHigherOrderSourceContribution(path, p2pFactors);
                        totalContribution += pathContribution;
                    }

                    const auto& downmixGains = sourceContributionToDownmixGain(totalContribution,
                        item->refDistance, *item, acousticEnv, acousticEnvironmentId);

                    if (downmixGains.first
                        > gainThreshold) { // If first gain is 0, the second is too.
                        const auto& directivity =
                            getDirectivityFilterForItem(*item, acousticEnvironmentId);
                        dspList.dspItems[acousticEnvironmentId].emplace_back(item->signal,
                            downmixGains.first, downmixGains.second, directivity,
                            item->acousticEnvironmentId);
                    }
                }
            }
        }
    }
    // This has to be done before updateMaterials is called.
    recalculateSecondOrderReverbCoefficients();

    m_connectionResolver.updateMaterials();

    // Second pass: calculate all reverb contributions from internal sources (and subtract each source's energy that went to external environments)
    for (const auto& [acousticEnvironmentId, acousticEnv] : reverbData) {
        int itemIdx = -1;
        for (auto const& item : renderList.items) {
            itemIdx++;
            if (item != nullptr && !isReverbInputRenderItem(*item)) {
                continue;
            }
            bool isExternalSource = (acousticEnvironmentId != item->acousticEnvironmentId);
            if (!isExternalSource) {
                auto downmixGains = getDownmixGainForInternalSources(*item, renderList, acousticEnv,
                    acousticEnvironmentId, listenerAcousticEnvironmentId, isExplicitPortalMode,
                    totalExternalContributions, itemIdx);

                if (downmixGains.first > gainThreshold) { // If first gain is 0, the second is too.
                    const auto& directivity =
                        getDirectivityFilterForItem(*item, acousticEnvironmentId);
                    dspList.dspItems[acousticEnvironmentId].emplace_back(item->signal,
                        downmixGains.first, downmixGains.second, directivity,
                        item->acousticEnvironmentId);
                }
            }
        }
    }

    bool immersiveReverbActive = calculateReverbEnvGains();

    // Place output channels to their designed positions. Delay lines are positioned around the
    // listener in listener coordinate space. For the headtracked reverb output we want the delay
    // line positions to stay "static within the scene" when the listener head rotates and thus we
    // rotate the delay line positions in inverse of the listener head rotation.

    if (m_headTrackingEnabled) {
        EAR::HomogenMatrix homogenMatrix;
        homogenMatrix.setRotation(renderList.listener.position.orientation);
        auto const inverseMatrix = homogenMatrix.inverse();
        for (unsigned int i{ 0 }; i < headtrackedOutputRenderItems.size(); ++i) {
            auto position = headtrackedOutputRenderItemDefaultPositions[i];
            if (m_headTrackingEnabled) {
                position.location = inverseMatrix.transform(position.location);
            }
            headtrackedOutputRenderItems[i]->setPosition(position);
        }
    }

    for (auto* ri : headtrackedOutputRenderItems) {
        // If no immersive reverb active, set the corresponding gain to 0, so that
        // panner stage does not cause discontinuous behavior when we enable the reverb later.
        if (isExplicitPortalMode) {
            if (!immersiveReverbActive) {
                ri->setGain(0.0f);
            } else {
                ri->setGain(1.0f);
            }
        }
        renderList.addItemCopy(ri);
        ri->resetChanged();
    }

    EAR_CHECK_REL(dspList.reverbData.empty(), "Reverb data map should be empty");

    for (const auto& [acousticEnvironmentId, reverbUnit] : reverbData) {
        const auto& item = reverbUnit.outputRenderItem;
        renderList.addItemCopy(item);
        item->resetChanged();
        dspList.reverbData.try_emplace(acousticEnvironmentId, reverbUnit);
    }

    dspList.sorGain = calculateSecondOrderReverbGain();
    previousListenerAcousticEnvironmentId = listenerAcousticEnvironmentId;
    return dspData;
}

std::pair<float, float> ReverbStage::getExternalSourceContribution(
    RenderItem& item, std::string acousticEnvironmentId) const
{
    std::pair<float, float> sourceContribution = { 0.0f, 0.0f };
    const float eps = 1e-9;

    if (bitstreamPortalPayload == nullptr) {
        return sourceContribution;
    }

    const std::vector<std::unique_ptr<MPEGI::Portal>>& portals =
        bitstreamPortalPayload->getPortals();

    for (const auto& portal : portals) {
        // If acousticEnvironmentId is one of the portal's environments, and the other is the same as the item's acousticEnvironmentId,
        // then we have a match to take action on.
        if ((acousticEnvironmentId.compare(portal->portalEnvironments[0]) == 0
                && item.acousticEnvironmentId.compare(portal->portalEnvironments[1]) == 0)
            || (acousticEnvironmentId.compare(portal->portalEnvironments[1]) == 0
                && item.acousticEnvironmentId.compare(portal->portalEnvironments[0]) == 0)) {
            float lineOfSightModifier = 1.0f;
            if (!m_connectionResolver.lineOfSightExists(
                    item.position.location, item.acousticEnvironmentId, portal->portalId)) {
                lineOfSightModifier = m_noLineOfSightCoeff;
            }
            const auto contribution =
                computeSourceContribution(*portal, item, eps, portalRotationMatrices);
            sourceContribution.first += lineOfSightModifier * contribution.first;
            sourceContribution.second += lineOfSightModifier * contribution.second;
        }
    }

    return sourceContribution;
}

// Returns true when path contains a given AE (ignores portals / directions)
bool pathContains(const ReverbStage::Path& path, const std::string& AE)
{
    auto nodeContainsAE = [AE](const ReverbStage::Node& node) {
        return node.acousticEnvironment == AE;
    };

    return std::find_if(path.begin(), path.end(), nodeContainsAE) != path.end();
}

std::vector<ReverbStage::Path> getAllPathsRecursive(const std::string& fromAE,
    const std::string& toAE,
    const std::unordered_map<std::string, std::vector<ReverbStage::Node>>& portalMap,
    const ReverbStage::Path& previousPath)
{
    // Recursive exit 1
    if (fromAE == toAE) {
        return { previousPath };
    }

    if (portalMap.count(fromAE) == 0) {
        // There is no portal in fromAE, and fromAE != toAE, so no connection is possible.
        return {};
    }

    std::vector<ReverbStage::Path> result;
    for (const auto& node : portalMap.at(fromAE)) {
        if (!pathContains(previousPath, node.acousticEnvironment)) {
            // Found a path through an AE that wasn't visited before (we only consider non-cyclical connections)
            // Add newly found path to previousPath, and recursively check if it ends up at toAE.
            ReverbStage::Path newPath = previousPath;
            newPath.push_back(node);

            const std::vector<ReverbStage::Path> paths =
                getAllPathsRecursive(node.acousticEnvironment, toAE, portalMap, newPath);

            for (const auto& path : paths) {
                result.push_back(path);
            }
        }
    }

    return result;
}

std::vector<ReverbStage::Path> ReverbStage::getAllPaths(const std::string& fromAE,
    const std::string& toAE, const std::vector<std::unique_ptr<Portal>>& portals)
{
    // Key = an AE, value is all possible Nodes that originate from that AE
    std::unordered_map<std::string, std::vector<ReverbStage::Node>> portalMap;

    for (const auto& portal : portals) {
        const std::string& AE0 = portal->portalEnvironments[0];
        const std::string& AE1 = portal->portalEnvironments[1];
        portalMap[AE0].emplace_back(portal.get(), AE1);
        portalMap[AE1].emplace_back(portal.get(), AE0);
    }

    return getAllPathsRecursive(fromAE, toAE, portalMap, { { nullptr, fromAE } });
}

float ReverbStage::getHigherOrderSourceContribution(
    const ReverbStage::Path& path, const std::map<std::pair<int, int>, float>& portal2portalFactors)
{
    float sourceContribution = 1.f;

    MPEGI::Portal* previousPortal = nullptr;
    for (const auto& node : path) {
        if (node.incomingPortal != nullptr) {
            const auto& mat = node.incomingPortal->scenePortal->getMaterial();
            if (mat) {
                sourceContribution *= getMaterialGainOut(*mat);
            }
            // If no material is set, getMaterialGainOut is assumed to be 1 (air)

            if (previousPortal != nullptr) {
                auto portal2portalIter = portal2portalFactors.find(
                    { previousPortal->portalId, node.incomingPortal->portalId });

                if (portal2portalIter != portal2portalFactors.end()) {
                    sourceContribution *= portal2portalIter->second;
                } else {
                    // Portal2portal factor was not provided for this combination of portals. Portal2Portal factors are optional in the bitstream, so this is an expected outcome.
                    // Assume portal2portal factor is 0, and hence the whole path has contribution 0.
                    return 0.f;
                }
            }
        }

        previousPortal = node.incomingPortal;
    }

    return sourceContribution;
}

void ReverbStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    IIRFilterType filterType = IIRFilterType::SOSs;

    if (!state) {
        return;
    }
    auto reverbPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
    auto portalPayloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_PORTAL);
    EAR_CHECK_REL(reverbPayloads.size() <= 1, "Too many reverb payloads in bitstream");
    if (reverbPayloads.size() == 1) {
        bitstreamReverbPayload =
            std::dynamic_pointer_cast<const MPEGI::ReverbPayload>(reverbPayloads[0]);
    }
    EAR_CHECK_REL(portalPayloads.size() <= 1, "Too many portal payloads in bitstream");
    if (portalPayloads.size() == 1) {
        bitstreamPortalPayload =
            std::dynamic_pointer_cast<const MPEGI::PortalPayload>(portalPayloads[0]);
        if (bitstreamPortalPayload->getExplicitPortalMode()) {
            calcOutsideEnergy = false;
            for (const auto& portal : bitstreamPortalPayload->getPortals()) {

                if (portal->extentId.has_value()) {
                    bitstreamPortalExtents[portal->portalId] =
                        state->getGeometry(portal->extentId.value());
                } else {
                    EAR_CHECK_REL(portal->equivalentBoxPointer != nullptr,
                        "Portal " + portal->portalStringId
                            + "has missing extentId and equivalentBoxPointer, one should be "
                              "set.");
                    bitstreamPortalExtents[portal->portalId] = portal->equivalentBoxPointer.get();
                }

                if (portal->scenePortal == nullptr) {
                    continue;
                }
                bitstreamPortalReverbOffsets[portal->portalId] =
                    std::max(portal->scenePortal->getExtentOffset()[0],
                        portal->scenePortal->getExtentOffset()[1]);

                EAR::Vector3 portalNormal = portal->normalVector;

                const EAR::Vector3 targetVector(0.f, 0.f, -1.f);
                EAR::HomogenMatrix rotMatrix = EAR::rotateVectorTo(portalNormal, targetVector);

                portalRotationMatrices[portal->portalId] = rotMatrix;

                // Verify that rotation is correct.
                portalNormal = rotMatrix * portalNormal;
                assert((portalNormal - targetVector).abs() < 1e-6f);

                if (portal->portalEnvironments[0].empty() || portal->portalEnvironments[1].empty())
                    calcOutsideEnergy = true;
            }

            if (bitstreamPortalPayload) {
                m_connectionResolver.readPortalPayload(bitstreamPortalPayload);
            }
        }
    }

    initializeConfigurationParameters();
    B = streamManager->getB();
    auto const fs = streamManager->getFs();
    EAR_CHECK_REL(fs == 48000.F, "Wrong sample rate: " + std::to_string(fs));

    auto const& reverbEnvs = state->getReverbEnvs();
    bool hasDefaultAE = false;
    if (bitstreamReverbPayload != nullptr) {
        hasDefaultAE = bitstreamReverbPayload->revDefaultAEPresent;
    }
    bool dropDefaultAE = false;
    std::optional<int> defaultAEId;

    for (auto const& re : reverbEnvs) {
        if (re->enclosure == nullptr) {
            defaultAEId = re->acousticEnvironmentBsId;
        }
    }
    auto reverbCount = reverbEnvs.size();
    if (hasDefaultAE && reverbCount >= (MAX_NUM_BOUNDED_AES_BEFORE_DROPPING_DEFAULT + 1)) {
        dropDefaultAE = true;
        reverbCount = reverbCount - 1;
    }

    // If default AE is being rendered, we don't need an extra calculation of reverb energy lost to 'outside'.
    if (hasDefaultAE && !dropDefaultAE) {
        calcOutsideEnergy = false;
    }

    // Calculate the configuration data (filters, delays) for the reverberator (FDN)
    // if a reverb payload is present in the bitstream.
    if (bitstreamReverbPayload) {
        calculateReverberatorConfig(
            bitstream, itemStore, streamManager, state, filterType, defaultAEId);
    }

    // Initialize source directivity eq mapping. Each acoustic environment has their own set of filters.
    auto sceneDirectivities = state->getDirectivities();
    std::unordered_map<std::string, mpegcore::GraphicEqParameters> dirFltParams;
    for (auto& [dirName, dirModel] : sceneDirectivities) {
        DirectivityPtr directivityPtr = dirModel;
        const auto& dirPat = (*dirModel).GetDirectivityCoverList();
        DirectivitySpectrum dirPatEffect(dirPat);
        FilterPtr filterPtr;
        if (RevDSPCfg.size() > 0) {
            // Calculate the filter
            float maxErrordB = 3;
            dirFltParams[dirName] = calcDirectivityFilter(
                &RevDSPCfg.begin()->second, dirPatEffect, filterType, dirFltOrder, maxErrordB);
        } else {
            // revDSPCfg not initialized, use neural network for directivity filter calculation
            EAR_CHECK_REL(directivityPtr != nullptr, "Directivity is null");
            dirFltParams[dirName] = calcNeuralNetDirectivityFilter(dirPatEffect, fs);
        }
    }

    for (auto& reverbEnv : reverbEnvs) {
        for (auto& [dirName, dirModel] : sceneDirectivities) {
            FilterPtr filterPtr = mpegcore::frontend::makeGraphicEqFilter(dirFltParams.at(dirName));
            auto directivityFilter = std::make_shared<DirectivityFilterData>(
                filterPtr, streamManager->createEmptyBuffer());
            m_directivityFilterMap[reverbEnv->acousticEnvironmentId].emplace(
                dirModel, directivityFilter);
        }
    }

    using namespace mpegcore;

    /* Configure reverb DSP for each reverb environment with config data from the bitstream and/or local calculation. */

    reverbData.reserve(reverbCount);
    std::size_t maxNumOutputChannels = 0;
    std::size_t numPannerChannels;
    bool hasAtLeastOneAREnvironment = false;

    for (auto& acousticEnv : reverbEnvs) {

        std::unique_ptr<Reverb> reverb;
        ReverberatorParameters reverbParams;
        int reverbID;

        // Create reverb depending whether the acoustic environment is a virtual/scene room, or a physical/AR room.
        if (acousticEnv->isSceneBasedAE) {
            reverbID = acousticEnv->acousticEnvironmentBsId.value();

            // Find the matching reverb config
            auto& virtReverbConfig = RevDSPCfg[reverbID];
            reverbParams.inputType = ReverberatorParameters::InputType::direct;
            reverbParams.eqFilterParams.emplace(
                virtReverbConfig.Reverberator.filters.SOSs.colorSos[0]);
            reverbParams.eqFilterParams->type =
                virtReverbConfig.Reverberator.filters.SOSs.colorSos->type;

            reverbParams.delayLines =
                ReverberatorCommon::DelayLineGroup(virtReverbConfig.Reverberator.nrLoops);

            std::vector<OutputDirection> outputChannelPositions;
            createOutputDirections(reverbParams.delayLines.size(), outputChannelPositions, true);

            for (unsigned int i = 0; i < reverbParams.delayLines.size(); ++i) {
                auto& delayLine = reverbParams.delayLines[i];
                auto& t60 = virtReverbConfig.Reverberator.filters.SOSs.t60Sos[i];

                delayLine.filterParams = t60;
                delayLine.filterParams.type = t60.type;
                delayLine.delayInSamples = virtReverbConfig.Reverberator.loopDelays[i];
                delayLine.azimuthDeg = outputChannelPositions[i].data()[0];
                delayLine.elevationDeg = outputChannelPositions[i].data()[1];
            }

            // Compensate for inherent FDN delay. Predelay-factor was already applied in CalculateBaseConfig(),
            // because it is needed in DSR scaling calculation for the eq/coloration/ratio filter.
            reverbParams.preDelayAdjustedInSamples =
                adjustPredelay(virtReverbConfig.renderPredelay / fs, reverbParams, fs);

            long long delayNs = (reverbParams.preDelayAdjustedInSamples.first * 1e9) / fs;
            reverbParams.preDelayAdjusted = std::chrono::nanoseconds{ delayNs };

            virtReverbConfig.acousticEnvironmentId = acousticEnv->acousticEnvironmentId;

            // save environment parameters in public struct to be visible for other stages
            acousticEnv->predelay = virtReverbConfig.renderPredelay / fs;

            for (auto [id, room] : reverbBsParameters) {
                unsigned int roomId = room->id;
                if (roomId == reverbID) {
                    auto& ACEVInputData = room;
                    const std::shared_ptr<FreqGrid> FreqGridInputData =
                        bitstreamReverbPayload->GetFreqGrid()[ACEVInputData->freqGridIdx];
                    for (int i = 0; i < room->t60.size(); i++) {
                        acousticEnv->frequencyDecayResponse.push_back(
                            { FreqGridInputData->centerFreq[i], room->t60[i] });
                    }
                }
            }

            m_connectionResolver.addAcousticEnvironmentBsId(
                acousticEnv->acousticEnvironmentId, reverbID);
        } else {
            /* AR scene */
            hasAtLeastOneAREnvironment = true;
            reverbID = ACOUSTIC_ENVIRONMENT_BITSTREAM_ID_UNDEFINED;

            // Generate ReverbParameters (including delay lines) from the acousticEnv
            reverbParams = generateReverbParameters_AR(acousticEnv, fs);
        }

        // Create the Reverb and init i/o buffers
        reverb = std::make_unique<Reverb>(reverbParams);
        reverb->initialize(B);

        std::size_t numberOfOutputChannels = reverb->numberOfOutputs();
        if (numberOfOutputChannels > maxNumOutputChannels) {
            maxNumOutputChannels = numberOfOutputChannels;
        }

        // Create render item for the reverb environment's non-headtracked output
        auto item = itemStore->createItem("reverb_re_" + acousticEnv->acousticEnvironmentId);
        item->type = ItemType::Reverb;
        item->setSignal(streamManager->createEmptyBuffer(numberOfOutputChannels));
        item->acousticEnvironmentId = acousticEnv->acousticEnvironmentId;
        // Note: reverb->outputPositions was initialized to the delay lines' azimuthDeg and elevationDeg, rho = 1.f
        item->channelPositions = reverb->outputPositions;
        item->getStateOrCreate<ReverbStage::ItemState>(false);

        reverbData[acousticEnv->acousticEnvironmentId].m_sorFilter =
            EAR::IIRFilter::create(SOR_FILTER_B, 3, SOR_FILTER_A, 3);

        reverbData[acousticEnv->acousticEnvironmentId].fill(item, std::move(reverb),
            acousticEnv->enclosure, reverbID, reverbParams.preDelayAdjustedInSamples);
    }

    // Prepare predelay lines for connected reverb and connections for second order reverb
    for (auto& [subjId, subj] : reverbData) {
        subj.previousOutput = std::make_unique<mpegcore::AudioBuffer>();
        subj.connectedEnvironments =
            std::make_unique<std::unordered_map<std::string, std::atomic<float>>>();
        subj.previousOutput->resize(B, 0.f);
        for (auto const& [srcId, src] : reverbData) {
            auto matchingPortals =
                m_connectionResolver.getPortalsBetweenEnvironments(subjId, srcId);
            if (matchingPortals.size() > 0) {
                // Store the name of the source environment and calculate a gain coefficient for it
                (*subj.connectedEnvironments)[srcId] =
                    getSecondOrderReverbCoefficient(matchingPortals);
                // Second order reverb has its own predelay line which is the average of both predelays
                subj.reverb->initPredelayLine(
                    secondOrderReverbPredelay(subj.predelayInSamples, src.predelayInSamples));
            }
            std::size_t subjPredelay = subj.predelayInSamples.first;
            std::size_t srcPredelay = connectedPredelay(src.predelayInSamples);
            std::size_t length = subjPredelay + srcPredelay;
            subj.reverb->initPredelayLine(length);
        }
    }

    /* Create head-tracked reverb render items */

    // Initialize output channel positions for immersive reverb (headtrackedOutputRenderItems) and Panner.
    // This amounts to busses encoded to fixed directions, and each render item's audio channels are sent
    // to the busses nearest to it's encoding directions.
    // Always use an immersive LS setup with Panner
    if (maxNumOutputChannels < REVERB_MIN_NUMBER_OF_LOUDSPEAKERS) {
        maxNumOutputChannels = REVERB_MIN_NUMBER_OF_LOUDSPEAKERS;
    }
    numPannerChannels = maxNumOutputChannels;

    // Create output encoding directions for the panner. These should coincide with the reverb's output
    // directions.
    std::vector<OutputDirection> pannerChanDirections_pol; // Polar(az, el, dist)
    createOutputDirections(
        numPannerChannels, pannerChanDirections_pol, !hasAtLeastOneAREnvironment);

    // Panner directions, polar -> cartesian coords
    std::vector<Vector3> pannerChanDirections_xyz;
    std::transform(pannerChanDirections_pol.begin(), pannerChanDirections_pol.end(),
        std::back_inserter(pannerChanDirections_xyz), [](const OutputDirection& dir_pol) {
            return polarToCartesian(static_cast<EAR::Polar>(dir_pol));
        });

    // Create an output mapping from reverb directions to output (panner) directions, for each reverb
    for (auto const& [revId, thisRevData] : reverbData) {

        // Note: outputRenderItem->channelPositions was set to reverb->outputPositions which is initialized
        // to the delay lines' azimuthDeg and elevationDeg, rho = 1.f
        std::vector<EAR::Position> revChanPositions =
            thisRevData.outputRenderItem->channelPositions;

        // Extract reverb's output encoding directions (cartesian coords) from Position
        std::vector<EAR::Vector3> revChanDirections_xyz;
        std::transform(revChanPositions.begin(), revChanPositions.end(),
            std::back_inserter(revChanDirections_xyz),
            [](const EAR::Position& revPos) { return revPos.location; });

        // Create a mapping that reassigns reverb output directions to the nearest panning output direction
        // Note: for AR that these directions should coincide.
        createOutputMapping(revId, revChanDirections_xyz, pannerChanDirections_xyz, outputMapping);
    }

    // Create output render items for headtracked reverb output. Need a render item per delay
    // line because they are individually positioned.
    headtrackedOutputRenderItems.reserve(numPannerChannels);
    headtrackedOutputRenderItemDefaultPositions.reserve(numPannerChannels);

    // Sequential (az, el) pairs stored in flat 1D vector
    std::vector<float> outChanAzimElevPairsFlat(maxNumOutputChannels * 2);

    for (unsigned int i = 0; i < numPannerChannels; i++) {

        auto item = itemStore->createItem("reverb_ht_dl_" + std::to_string(i));

        // TODO: make ItemType::Pannable use a multichannel StreamBuffer similar to ItemType::Reverb.
        // This would allow a single ItemType::Pannable render item for headtrackedOutputRenderItems
        // (instead of a vector of them), but requires PannerStage to be changed as it currently expects
        // RenderItems with a single channel.
        item->type = ItemType::Pannable;
        item->setSignal(streamManager->createEmptyBuffer());

        // Spatial positioning
        float azi = pannerChanDirections_pol[i].data()[0];
        float ele = pannerChanDirections_pol[i].data()[1];
        outChanAzimElevPairsFlat[i * 2] = azi;
        outChanAzimElevPairsFlat[i * 2 + 1] = ele;
        EAR::Vector3 location = polarToCartesian(pannerChanDirections_pol[i]);
        EAR::Position pos(location, EAR::Euler());
        pos.cSpace = EAR::CSpace::LCS;
        item->setPosition(pos);

        item->setAParams(item->aparams | AParam::ForceNoDoppler | AParam::NoDistanceGain);
        item->getStateOrCreate<ReverbStage::ItemState>(true);
        headtrackedOutputRenderItems.push_back(item);
        headtrackedOutputRenderItemDefaultPositions.push_back(pos);
    }
    Config::set("Panner:directions", EAR::Variant(outChanAzimElevPairsFlat));

    // Second order reverb fade in cycles depends on the update rate
    const auto* parentScene = state->getParentScene();
    if (parentScene != nullptr) {
        m_sorFade.second = static_cast<unsigned int>(
            SECOND_ORDER_REVERB_FADE_SECONDS / std::max(parentScene->getControlInterval(), 1e-8));
    }
    m_sorFilterBuffer.resize(B, 0.f);

    //Get LSDF ListenerVoice Reverb gain and store reference to listener voice SceneObject
    for (const auto& source : state->getObjectSources()) {
        listenerVoice = dynamic_cast<EAR::ListenerVoiceSource*>(source.get());
        if (listenerVoice) {
            Config::set("Reverb::ListenerVoiceGain", EAR::Variant(listenerVoice->getReverbGain()));
        }
    }
}

mpegcore::ReverberatorParameters ReverbStage::generateReverbParameters_AR(
    const std::unique_ptr<MPEGI::ReverbEnv>& acousticEnv, const float fs)
{
    using namespace mpegcore;

    AcousticParameters acousticParameters;

    // Populate T60 and RDR with initial acousticEnv data
    populateAcousParamDecayPropertiesFromRevEnv(acousticParameters, *acousticEnv);

    // Bounding box: for diffuseOnset and RDR (if not already specified), and delay line lengths
    double longestRoomDimension, roomVolume; // set by calculateRoomVolumeAndLongestDim
    EAR::Geometry::BoundingBox boundingBox = getAEBoundingBox(acousticEnv.get());
    setRoomVolumeAndLongestDim(boundingBox, roomVolume, longestRoomDimension);

    // If acousticEnv's predelay (diffuseOnset) is not set, approximate it
    if (!acousticEnv->predelay.has_value()) {
        auto geometricEIFPredelay = calculateEifPredelay(longestRoomDimension);
        // we know the EIF overestimates the diffuse onset by roughly a factor of 4
        const float diffuseOnsetApprox = geometricEIFPredelay / 4.f;
        acousticEnv->predelay = diffuseOnsetApprox;

        printStatus("Approx diffuse onset by theoretical model: {} sec", diffuseOnsetApprox);
    }
    const float diffusePredelay = acousticEnv->predelay.value();

    /*
     At this stage, acousticEnv->predelay is equivalent to the diffuse onset. Predelay
     informs the time adjustment of the FDN output and the calculation of RDR.
     User-specified predelayFactor and predelayOffset are NOT included prior to being
     used for RDR calculation. This is so that when a user sets the predelayFactor and
     predelayOffset they know they are affecting only the time alignment of the FDN.
     They have separate control over the RDR adjustment by config/bitstream parameters.
     The RDR is left to our "under the hood" estimation. */

    /* RDR adjustments by user configurations */
    updateRDRWithConfigSettings(acousticParameters, roomVolume, diffusePredelay);

    /* 
     acousticEnvironmentData.predelayFactor default value (4) anticipates the use of the
     EIF estimate of predelay, which overestimates the diffuseOnset roughly by a factor of 4.
     Now that it is corrected above (see geometricEIFPredelay), predelayFactor can defaul to
     1.0 and therefore can be set independently without assumptions of how the predelay or
     diffuseOnset was determined. */

    // The acoustic environment's predelay factor was initialized with the EIF predelay
    // factor default (which is used differently for VR), so update with AR predelayFactor.
    const float predelayFactor =
        Config::getFloatDef("Reverb:PredelayFactorAR", DEFAULT_REVERB_AR_PREDELAY_FACTOR);
    const float predelayOffset = acousticEnv->predelayOffset;
    validatePredelayFactor(diffusePredelay, predelayFactor, acousticParameters);
    acousticEnvironmentData.predelayFactor = predelayFactor;

    // TODO: predelayFactor would ideally be a _multiplier_ not a divisor.
    double requestedDiffuseOnsetSec =
        std::fmaxf(0.0, (diffusePredelay / predelayFactor) + predelayOffset);
    acousticParameters.preDelaySec = requestedDiffuseOnsetSec;

    /* Create reverb parameters (delay lines and RDR filters) */

    std::size_t numDelayLines;
    if (acousticEnv->numDelayLines.has_value())
        numDelayLines = acousticEnv->numDelayLines.value();
    else
        numDelayLines = REVERB_DEFAULT_NUMBER_OF_DELAY_LINE_LOOPS;
    if (numDelayLines < SECOND_ORDER_REVERB_DELAY_LINES) {
        m_unrestrictedSor = false;
    }

    // createReverbParameters() also creates delay lines
    auto reverbParams_maybe =
        createReverbParamsFromAcousticParams(acousticParameters, boundingBox, fs, numDelayLines);

    LSDF_CHECK_REL(reverbParams_maybe.has_value(),
        fmt::format(
            "Failed to create reverb parameters from acoustic parameters for (AR) reverb {}. "
            "Either number of delay lines is unsupported or delay lines failed to be created.",
            acousticEnv->acousticEnvironmentId));

    auto& reverbParams = reverbParams_maybe.value();
    reverbParams.inputType = ReverberatorParameters::InputType::direct;

    // Set the length of the predelay delay line (aligns the FDN diffuseOnset with the requested diffuse onset)
    reverbParams.preDelayAdjustedInSamples =
        getPredelayForDiffuseOnsetAlignment(requestedDiffuseOnsetSec, reverbParams, fs);
    long long delayNs = (reverbParams.preDelayAdjustedInSamples.first * 1e9) / fs;
    reverbParams.preDelayAdjusted = std::chrono::nanoseconds{ delayNs };

    printStatus("Rendering Predelay: {} samp ({:.4f} sec)",
        static_cast<float>(reverbParams.preDelayAdjustedInSamples.first),
        static_cast<float>(reverbParams.preDelayAdjustedInSamples.first) / fs);

    return reverbParams;
}

// AR: helper for generateReverbParameters_AR
void ReverbStage::populateAcousParamDecayPropertiesFromRevEnv(
    mpegcore::AcousticParameters& acousticParameters, const MPEGI::ReverbEnv& acousticEnv)
{
    // Polulate rt60 & rdr
    const bool isFromRIR = acousticParameters.isFromRIR = acousticEnv.isFromRIR;
    const bool srcDistanceIsKnown = acousticEnv.srcDistanceIsKnown;
    const bool predelayIsSpecifiedOrNotRequired = acousticEnv.predelay.has_value()
        || !Config::getBoolDef("Reverb:isPredelayRequiredForRdr", true);

    transformToBack(acousticEnv.frequencyDecayResponse, acousticParameters.frequencyDecayResponse,
        [&](auto const& revEnvFreqData) {
            mpegcore::AcousticParameters::FrequencyDecay outFreqData;
            outFreqData.frequency = revEnvFreqData.frequency;
            outFreqData.rt60 = revEnvFreqData.rt60;
            LSDF_CHECK_REL(predelayIsSpecifiedOrNotRequired || !revEnvFreqData.rdr.has_value(),
                "ReverbStage: can't have RDR without predelay");
            // RDR as measured from RIR is only reliable if the src-mic distance is known.
            // Otherwise rdr remains nullopt so renderer falls back on geometric approx of RDR.
            if (!isFromRIR || srcDistanceIsKnown) {
                outFreqData.rdr = revEnvFreqData.rdr;
            }
            return outFreqData;
        });
}

// AR: helper for generateReverbParameters_AR
void ReverbStage::updateRDRWithConfigSettings(mpegcore::AcousticParameters& acousticParameters,
    const double roomVolume, const double diffuseOnset)
{
    // frequencyDecayResponse may not have RDR values because they weren't specified in the
    // LSDF or RDR from RIR was not used because src->mic distance wasn't known.
    const bool lsdfHasRdrValues = hasRdrValues(acousticParameters.frequencyDecayResponse);
    mpegcore::printStatus("LSDF has RDR values? {}", lsdfHasRdrValues);

    float rdrOffset;
    if (!lsdfHasRdrValues || lsdfRDRHandlingData.revOverrideLSDF_RDR) {
        // Estimate and update RDR from theoretical model based on roomVolume and diffuseOnset
        setRdrValuesFromTheoreticalModel(
            acousticParameters.frequencyDecayResponse, roomVolume, diffuseOnset);
        rdrOffset = lsdfRDRHandlingData.revLsdfTheoreticalRDRtoActualRDRdB;

        mpegcore::printStatus("RDR source: calculated from theoretical model");
    } else {
        rdrOffset = lsdfRDRHandlingData.revLsdfRDRToActualRDRdB;
        if (acousticParameters.isFromRIR)
            mpegcore::printStatus("RDR source: calculated from RIR (LSDF)");
        else
            mpegcore::printStatus("RDR source: read in from LSDF");
    }
    mpegcore::printStatus("\tRDR offset {:.2f} dB", rdrOffset);

    addOffsetToRdrValues(acousticParameters.frequencyDecayResponse, rdrOffset);
}

// AR: helper for generateReverbParameters_AR
void ReverbStage::validatePredelayFactor(const float predelay, const float predelayFactor,
    const mpegcore::AcousticParameters& acousticParameters)
{
    // predelayFactor must be positive
    LSDF_CHECK_REL(predelayFactor > 0,
        "The predelayFactor must be greater than 0.0, check the "
        "Reverb:PredelayFactor configuration.");

    // Check for obvious error in Reverb:PredelayFactor value (predelayFactor as a divisor
    // can be easily misinterpreted as a multiplier)
    auto freqDecayWithMaxT60 = std::max_element(acousticParameters.frequencyDecayResponse.begin(),
        acousticParameters.frequencyDecayResponse.end(),
        [](const mpegcore::AcousticParameters::FrequencyDecay& a,
            const mpegcore::AcousticParameters::FrequencyDecay& b) { return a.rt60 < b.rt60; });
    const float maxRt60Sec = freqDecayWithMaxT60->rt60;

    LSDF_CHECK_REL((predelay / predelayFactor) < maxRt60Sec * 3,
        "The predelayFactor results in a predelay longer than three times the "
        "rt60, check the Reverb:PredelayFactor configuration.");
}

void ReverbStage::CalculateBaseConfig(std::shared_ptr<SetBaseConfigStruct>& baseConfigOut,
    unsigned int acEnvIdx, unsigned int nrFbLoops, std::shared_ptr<const ReverbParameters> room,
    bool isDefaultAE, const float fs)
{

    std::shared_ptr<SetBaseConfigStruct> baseConfig = std::make_shared<SetBaseConfigStruct>();
    baseConfigOut = std::make_shared<SetBaseConfigStruct>();

    if (isDefaultAE) {
        fillBaseConfig(baseConfig.get(), nrFbLoops, defaultAcousticEnvironmentData.predelayFactor,
            defaultAcousticEnvironmentData.delayRatio, defaultAcousticEnvironmentData.delayMinMs,
            fs);
    } else {
        float delayRatio = 0.0f;
        float delayMinMs = 0.0f;
        fillBaseConfig(baseConfig.get(), nrFbLoops, acousticEnvironmentData.predelayFactor,
            delayRatio, delayMinMs, fs);
    }
    baseConfig->metadataPredelay = (int32_t)(room->parOffset * (float)baseConfig->fs);
    baseConfig->pRoomDimensions = AcEnvProp[acEnvIdx].dimensions;
    reverbSetBaseConfig(baseConfig.get(), baseConfigOut.get(), isDefaultAE);
    fillBaseConfigOut(baseConfigOut.get(), &(RevDSPCfg[room->id]), t60FltOrder, colorFltOrder);
}

void ReverbStage::CalculateT60(std::shared_ptr<T60Settings>& T60ConfigOut,
    const unsigned int roomId, std::shared_ptr<lateReverb> lateRevOut, IIRFilterType type)
{
    std::shared_ptr<T60Params> T60ConfigIn = std::make_shared<T60Params>(type);
    T60ConfigOut = std::make_shared<T60Settings>(type);

    auto& ReverbConfig = RevDSPCfg[roomId];

    // T60 parameters

    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        // Note, the fc in Miscellanea is likely used for coloration filter
        ReverbConfig.Miscellanea.fc[i] = lateRevOut->filterDesignFc[i];
    }

    fillT60Config(&ReverbConfig, T60ConfigIn.get());

    for (int16_t i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        T60ConfigIn->targetT60[i] = lateRevOut->acousticRT60[i];
    }

    if (!checkT60Configuration(T60ConfigIn.get())) {
        assert(false);
    }

    reverbAddT60Filters(T60ConfigIn.get(), T60ConfigOut.get(), type);

    fillT60ConfigOut(T60ConfigOut.get(), &ReverbConfig);
}

void ReverbStage::CalculateLateReverb(std::shared_ptr<lateReverb>& lateRevOut,
    const std::shared_ptr<FreqGrid> FreqGridInputData, std::shared_ptr<const ReverbParameters> room)
{
    auto& ReverbConfig = RevDSPCfg[room->id];
    auto ACEVInputData = room;
    std::shared_ptr<lateReverbInParms> lateRevIn = std::make_shared<lateReverbInParms>();
    lateRevOut = std::make_shared<lateReverb>();
    for (int16_t i = 0; i < ACEVInputData->dsr.size(); i++) {
        lateRevIn->acousticParameters.dsr[i] = ACEVInputData->dsr[i];
    }
    lateRevIn->acousticParameters.dsrSize = ACEVInputData->dsr.size();
    for (int16_t i = 0; i < FreqGridInputData->nrBands; i++) {
        lateRevIn->acousticParameters.inputFrequencies[i] = FreqGridInputData->centerFreq[i];

        lateRevIn->acousticParameters.t60[i] = ACEVInputData->t60[i];
    }
    lateRevIn->fs = ReverbConfig.Miscellanea.fs;

    lateRevOut->metadataPredelay = (uint16_t)(ACEVInputData->parOffset * (float)lateRevIn->fs);

    preprocessReverbInputParameters(lateRevIn.get(), lateRevOut.get());
}

void ReverbStage::CalculateColoration(std::shared_ptr<ColorParams>& ColorOut,
    const unsigned int roomId, const std::shared_ptr<lateReverb> lateRevOut,
    const std::shared_ptr<SetBaseConfigStruct> baseConfigOut, IIRFilterType type, bool isDefaultAE)
{
    std::shared_ptr<ColorParams> ColorConfigIn = std::make_shared<ColorParams>(type);
    if (ColorOut == nullptr) {
        ColorOut = std::make_shared<ColorParams>(type);
    }

    auto& ReverbConfig = RevDSPCfg[roomId];

    fillColorationConfig(&ReverbConfig, ColorConfigIn.get(), lateRevOut.get()->acousticDSR,
        lateRevOut.get()->acousticRT60, lateRevOut.get()->metadataPredelay);

    if (!checkColorationInputConfiguration(ColorConfigIn.get())) {
        assert(false);
    }

    ColorConfigIn->renderPredelay = baseConfigOut->renderPredelay;

    reverbCalculateLevelData(ColorConfigIn.get(), ColorOut.get(), isDefaultAE);

    reverbAddColorFilters(ColorConfigIn.get(), ColorOut.get(), type);

    fillColorationConfigOut(ColorOut.get(), &ReverbConfig);
}

std::pair<float, float> ReverbStage::calculateFilteredItemGains(float distance,
    const RenderItem& item, std::optional<float> commonComponentClrFlt, bool isDefaultAE)
{
    auto result = std::make_pair<float, float>(1.f, 1.f);
    result = ReverbStage::calculateItemGains(distance, item);

    // Scaling is omitted in case the item is an AR environment.
    if (commonComponentClrFlt.has_value()) {
        // Include common component from coloration filter.
        // The sqrtNormalizedSignalScalefactor indicates how a source's PCM sample scales to
        // RMS source energy emitted into the room for an omnidirectional source.
        //      earSurf = 1e-3;
        //      sqrtNormalizedSignalScalefactor = sqrt(4 * M_PI / earSurf) = 112.0998;
        const float sqrtNormalizedSignalScalefactor = 112.0998f;
        float factor = sqrtNormalizedSignalScalefactor;
        factor *= commonComponentClrFlt.value();
        result.first *= factor;
        result.second *= factor;
    }

    // Include effect of reference distance on scaling factor for signal energy to get to
    // emitted source energy.
    result.first *= item.refDistance;
    if (isDefaultAE == false) {
        result.second *= item.refDistance;
    }
    // Individual reverb gain per source. Valid for bounded AEs and DefaultAE
    result.first *= item.itemReverbGain;
    result.second *= item.itemReverbGain;
    return result;
}

void ReverbStage::calculateReverberatorConfig(const Bitstream& bitstream,
    RenderItemStore* itemStore, const StreamManager* streamManager, const SceneState* state,
    IIRFilterType IIRFilterType, std::optional<int> defaultAEId)
{
    reverbBsParameters = bitstreamReverbPayload->GetReverbParameters();

    SetAcousticEnvironmentProperties(state);

    int acEnvIdx = -1;
    for (auto [id, room] : reverbBsParameters) {
        acEnvIdx++;
        unsigned int roomId = room->id;
        auto& ACEVInputData = room;
        auto& ReverbConfig = RevDSPCfg[roomId];
        ReverbConfig.acousticEnvironmentId = roomId;

        bool isDefaultAE = false;
        if (defaultAEId.has_value() && defaultAEId.value() == id) {
            isDefaultAE = true;
        }

        const std::shared_ptr<FreqGrid> FreqGridInputData =
            bitstreamReverbPayload->GetFreqGrid()[ACEVInputData->freqGridIdx];

        // Base Config
        std::shared_ptr<SetBaseConfigStruct> baseConfigOut;
        CalculateBaseConfig(baseConfigOut, acEnvIdx, room->delayLineCount, room, isDefaultAE,
            streamManager->getFs());
        // End Base Config

        // Late Reverb
        std::shared_ptr<lateReverb> lateRevOut;
        CalculateLateReverb(lateRevOut, FreqGridInputData, room);
        // End Late Reverb

        // T60
        std::shared_ptr<T60Settings> T60ConfigOut;
        CalculateT60(T60ConfigOut, roomId, lateRevOut, IIRFilterType);
        // End of T60

        // Coloration
        std::shared_ptr<ColorParams> ColorOut;
        CalculateColoration(
            ColorOut, roomId, lateRevOut, baseConfigOut, IIRFilterType, isDefaultAE);
        // End of Coloration

        if (room->delayLineCount < SECOND_ORDER_REVERB_DELAY_LINES) {
            m_unrestrictedSor = false;
        }
    }
}

std::optional<float> ReverbStage::calculateTheoreticalRdr(
    double rt60, double roomVolume, double predelay) const
{
    double rdrLog = -60.0;
    if (rt60 > 0.0) {
        rdrLog = 10.0 * std::log10(rt60 / roomVolume) - 60.0 * (predelay / rt60) + 25.0;
    }
    double rdrLin = rdrLogToRdrLin(rdrLog);
    return std::optional<float>(rdrLin);
}

double ReverbStage::rdrLogToRdrLin(double rdrLog) const
{
    return std::pow(10.0, rdrLog / 10.0);
}

double ReverbStage::calculateEifPredelay(double longestRoomDimension) const
{
    return 4.0 * longestRoomDimension / lsdfRDRHandlingData.speedOfSound;
}

void ReverbStage::setRoomVolumeAndLongestDim(EAR::Geometry::BoundingBox const& boundingBox,
    double& roomVolume, double& longestDimension) const
{
    auto const boundingBoxWidth = boundingBox[0].high - boundingBox[0].low;
    auto const boundingBoxHeight = boundingBox[1].high - boundingBox[1].low;
    auto const boundingBoxDepth = boundingBox[2].high - boundingBox[2].low;
    double temp = std::max(boundingBoxWidth, boundingBoxHeight);
    longestDimension = std::max(temp, static_cast<double>(boundingBoxDepth));
    roomVolume = boundingBoxWidth * boundingBoxHeight * boundingBoxDepth;
}

bool ReverbStage::calculateReverbEnvGains()
{
    // Default acoustic environment is only audible if the listener is not in any acoustic
    // environment.
    DefaultAcousticEnvironment defaultAE{ 1.0f, nullptr };
    bool immersiveReverbActive = false;
    for (auto& reverbUnit : reverbData) {
        auto& reverbEnv = reverbUnit.second;
        if (reverbEnv.enclosure != nullptr) {
            reverbEnv.gain = calculateReverbEnvGain(
                reverbEnv, reverbUnit.first == listenerAcousticEnvironmentId);
            // Default AE gain is faded in when the loudest bounded AE is faded out,
            // and vice versa.
            defaultAE.gain = std::min(defaultAE.gain, 1 - reverbEnv.gain);

            if (reverbEnv.gain > 0.0f) {
                immersiveReverbActive = true;
            }
        } else {
            EAR_CHECK_REL(!defaultAE.reverb, "ReverbStage: too many AEs without enclosure");
            defaultAE.reverb = &reverbEnv;
        }
    }

    // Gain of the default acoustic environment, if there is one, can only be determined
    // after going through all the other acoustic environments first.
    if (defaultAE.reverb != nullptr) {
        defaultAE.reverb->gain = defaultAE.gain * defaultAcousticEnvironmentData.gain;
        if (defaultAE.reverb->gain > 0.0f) {
            immersiveReverbActive = true;
        }
    }

    return immersiveReverbActive;
}

void ReverbStage::recalculateSecondOrderReverbCoefficients()
{
    // Update second order reverb coefficients in case portal materials have changed.
    for (auto& [subjId, subj] : reverbData) {
        for (auto& [srcId, gain] : *subj.connectedEnvironments) {
            if (m_connectionResolver.portalMaterialChanged(srcId, subjId)) {
                auto matchingPortals =
                    m_connectionResolver.getPortalsBetweenEnvironments(subjId, srcId);
                if (matchingPortals.size() > 0) {
                    gain = getSecondOrderReverbCoefficient(matchingPortals);
                }
            }
        }
    }
}

// Applies a gain ramp to second order reverb when moving across environs
float ReverbStage::calculateSecondOrderReverbGain()
{
    // Room has changed, apply SOR gain from zero to full across several update cycles
    if (previousListenerAcousticEnvironmentId != listenerAcousticEnvironmentId) {
        m_sorFade.first = m_sorFade.second;
    }

    float sorGain = m_defaultSorGain;
    if (!m_unrestrictedSor && m_sorFade.first > 0 && m_sorFade.second > 0) {
        float maxRounds = m_sorFade.second;
        sorGain *= (maxRounds - m_sorFade.first--) / maxRounds;
    }

    return sorGain;
}

float ReverbStage::calculateCrossfadeAttenuation(const EAR::Geometry& enclosure)
{
    // Default is full attenuation. Closer the listener gets to the enclosure of
    // the smaller the attenuation gets.
    float attenuation{ 1.0f };
    const float distanceToReverbEnv = listenerDistanceToEnv(enclosure);
    if (distanceToReverbEnv < m_crossfadeFadeInDistance) {
        attenuation = distanceToReverbEnv / m_crossfadeFadeInDistance;
    }
    return attenuation;
}

float ReverbStage::getItemGainBasedOnAE(const DSPItem& item, const ReverbData& reverbUnit)
{
    float gain{ 0.0f };
    if (reverbUnit.enclosure != nullptr) {
        gain = item.gainInBoundedAE;
    } else {
        gain = item.gainInDefaultAE;
    }
    return gain;
}

float ReverbStage::calculateReverbEnvGain(const ReverbData& reverbUnit, bool isListenerInEnv)
{
    float attenuation{ 1.0f };
    if (isListenerInEnv) {
        attenuation = 0.0f;
    } else {
        bool fade = m_crossfadeFadeInDistance > 0.0f;
        if (fade && isPortalNearby(listenerLocation)) {
            fade = false;
        }
        if (fade) {
            EAR_CHECK_REL(reverbUnit.enclosure, "ReverbStage: enclosure for bounded env is null");
            attenuation = calculateCrossfadeAttenuation(*reverbUnit.enclosure);
        } else {
            attenuation = 1.0f;
        }
    }
    return 1.0f - attenuation;
}

EAR::Geometry::BoundingBox ReverbStage::getAEBoundingBox(const ReverbEnv* re)
{
    // Null enclosures are allowed only for VR scenes that aren't processed here.
    // If the following check fails an AR scene has an AE without region.
    EAR_CHECK_REL(re->enclosure != nullptr, "ReverbStage: ReverbEnv's enclosure cannot be null");
    return re->enclosure->getBoundingBox();
}

void ReverbStage::savePreviousReverbOutput(const ReverbData& reverbUnit)
{
    // Save data for next round but clear the buffer first
    const std::unique_ptr<mpegcore::AudioBuffer>& prev = reverbUnit.previousOutput;

    zeroVector(m_sorFilterBuffer);
    zeroVector(*prev);

    // Create a mono mix of reverb output to use as input for next update cycle
    for (auto ch{ reverbUnit.reverb->outputAudio.size() }; ch-- > 0;) {
        auto const src = reverbUnit.reverb->outputAudio[ch];
        addFloatVectors(src.data(), m_sorFilterBuffer.data(), B);
    }

    reverbUnit.m_sorFilter->process(m_sorFilterBuffer.data(), prev->data(), B);
}

double ReverbStage::calculateDistanceGain(
    double distance, double refDistance, const AcousticEnvironmentData& env) const
{
    // refDistance is the distance at which the calculated attenuation is 0dB as per the EIF.
    double clampedDistance = std::max(env.minimumDistance, static_cast<float>(distance));
    double dbGain = env.distanceGainDbFactor * std::log10(refDistance / clampedDistance);
    double linearGain = std::pow(10.0, dbGain / 20.0);
    return linearGain;
}

std::pair<float, float> ReverbStage::calculateItemGains(float distance, const RenderItem& item)
{
    double const distanceGain =
        calculateDistanceGain(distance, item.refDistance, acousticEnvironmentData);
    double const itemGain = item.gain * distanceGain;
    double const distanceGainDefaultAE =
        calculateDistanceGain(distance, item.refDistance, defaultAcousticEnvironmentData);
    double const itemGainInDefaultAE = item.gain * distanceGainDefaultAE;

    return std::make_pair(itemGain, itemGainInDefaultAE);
}

void ReverbStage::setConfigurableVariables()
{
    // This function is performed every update cycle. Thus, it makes sense to add here only
    // those variables which can really be configured during runtime.
    reverbGain = Config::getFloatDef("Reverb:Gain", EAR::Variant(DEFAULT_REVERB_GAIN));
    if (listenerVoice) {
        listenerVoice->setReverbGain(
            Config::getFloatDef("Reverb::ListenerVoiceGain", EAR::Variant(DEFAULT_REVERB_GAIN)));
    }
    defaultAcousticEnvironmentData.gain =
        Config::getFloat("Reverb:Default_Acoustic_Environment_Gain");
}

inline float ReverbStage::listenerDistanceToEnv(const EAR::Geometry& reverbEnv)
{
    const auto& envClosestPoint = reverbEnv.closestPointTo(listenerLocation);
    return (listenerLocation - envClosestPoint).abs();
}

ReverbStage::DirectivityFilterPtr ReverbStage::getDirectivityFilterForItem(
    const RenderItem& item, const std::string& acousticEnvironmentId)
{
    DirectivityFilterPtr filter = nullptr;
    if (m_directivityFilterMap.find(acousticEnvironmentId) != m_directivityFilterMap.end()
        && item.directivity != nullptr) {
        auto it = m_directivityFilterMap.at(acousticEnvironmentId).find(item.directivity);
        if (it != m_directivityFilterMap.at(acousticEnvironmentId).end()) {
            filter = it->second;
        }
    }
    return filter;
}

/*
Values for variables used by the reverb stage are set in the following manner:
1. Default values are set during compilation.
2. If bitstream has reverb payload, it is used to replace default values.
3. Configuration file parameters take the highest precedence.
These values are not supposed to be changed after initializeStage() is called.
*/
void ReverbStage::initializeConfigurationParameters()
{
    float distanceGainDropDb{ DEFAULT_REVERB_DISTANCE_GAIN_DROP_DB };
    float minDistance{ DEFAULT_REVERB_MINIMUM_DISTANCE };
    float defaultAEGain{ DEFAULT_REVERB_GAIN_IN_DEFAULT_AE };
    float defaultAEDelayRatio{ DEFAULT_REVERB_DELAY_RATIO_IN_DEFAULT_AE };
    float defaultAEDelayMinMs{ DEFAULT_REVERB_DELAY_MINIMUM_MS_IN_DEFAULT_AE };
    float distanceGainDropDbDefaultAE{ DEFAULT_REVERB_DISTANCE_GAIN_DROP_DB_IN_DEFAULT_AE };
    float minDistanceDefaultAE{ DEFAULT_REVERB_MINIMUM_DISTANCE_IN_DEFAULT_AE };
    float fadeInDistance{ DEFAULT_REVERB_FADE_IN_DISTANCE };
    bool fadeInEnabled{ true };
    bool headtrackingEnabled{ false };
    float speedOfSound{ DEFAULT_REVERB_SPEED_OF_SOUND };
    float theoreticalRDRToActualRDR{ DEFAULT_REVERB_RDR_THEORETICAL_TO_RDR_ACTUAL };
    float lsdfRDRToActualRDR{ DEFAULT_REVERB_RDR_LSDF_TO_RDR_ACTUAL };
    bool overrideLSDFRDR{ DEFAULT_REVERB_LSDF_RDR_OVERRIDE };

    if (bitstreamReverbPayload) {
        distanceGainDropDb = bitstreamReverbPayload->reverbDistanceGainDropDb;
        minDistance = bitstreamReverbPayload->reverbMinimumDistance;
        defaultAEGain = bitstreamReverbPayload->revDefaultAEGain;
        defaultAEDelayRatio = bitstreamReverbPayload->revDefaultAEDelayRatio;
        defaultAEDelayMinMs = bitstreamReverbPayload->revDefaultAEDelayMinMs;
        distanceGainDropDbDefaultAE = bitstreamReverbPayload->revDefaultAEDistanceGainDropDb;
        minDistanceDefaultAE = bitstreamReverbPayload->revDefaultAEMinimumDistance;
        fadeInDistance = bitstreamReverbPayload->reverbFadeInDistance;
        fadeInEnabled = bitstreamReverbPayload->fadeInEnabled;
        headtrackingEnabled = bitstreamReverbPayload->headTrackingEnabled;
        theoreticalRDRToActualRDR = bitstreamReverbPayload->revLsdfTheoreticalRDRtoActualRDRdB;
        lsdfRDRToActualRDR = bitstreamReverbPayload->revLsdfRDRtoActualRDRdB;
        overrideLSDFRDR = bitstreamReverbPayload->revOverrideLSDF_RDR;
    }

    /* Acoustic Environment definitions. */
    acousticEnvironmentData = AcousticEnvironmentData(
        Config::getFloatDef(
            "Reverb:PredelayFactor", EAR::Variant(DEFAULT_REVERB_EIF_PREDELAY_FACTOR)),
        Config::getFloatDef("Reverb:DistanceGainDropDb", EAR::Variant(distanceGainDropDb)),
        Config::getFloatDef("Reverb:MinimumDistance", EAR::Variant(minDistance)), 1.0F, 0.0f, 0.0f);

    defaultAcousticEnvironmentData = AcousticEnvironmentData(
        Config::getFloatDef("Reverb:PredelayFactorDefaultAE",
            EAR::Variant(DEFAULT_REVERB_PREDELAY_FACTOR_IN_DEFAULT_AE)),
        Config::getFloatDef(
            "Reverb:DistanceGainDropDbDefaultAE", EAR::Variant(distanceGainDropDbDefaultAE)),
        Config::getFloatDef("Reverb:MinimumDistanceDefaultAE", EAR::Variant(minDistanceDefaultAE)),
        Config::getFloatDef(
            "Reverb:Default_Acoustic_Environment_Gain", EAR::Variant(defaultAEGain)),
        Config::getFloatDef(
            "Reverb:Default_Acoustic_Environment_Delay_Ratio", EAR::Variant(defaultAEDelayRatio)),
        Config::getFloatDef("Reverb:Default_Acoustic_Environment_Delay_Min_Ms_Override",
            EAR::Variant(defaultAEDelayMinMs)));

    /* Crossfade and headtracking definitions */
    m_crossfadeFadeInDistance =
        Config::getFloatDef("Reverb:Crossfade_Fade_In_Distance", EAR::Variant(fadeInDistance));
    m_headTrackingEnabled =
        Config::getBoolDef("Reverb:HeadTrackingEnabled", EAR::Variant(headtrackingEnabled));
    mpegcore::printStatus("Headtracking enabled: {}", m_headTrackingEnabled);

    /* Initializing lsdf handling. This is AR only. */
    lsdfRDRHandlingData.speedOfSound =
        Config::getFloatDef("Reverb:SpeedOfSound", EAR::Variant(speedOfSound));
    EAR_CHECK_REL(lsdfRDRHandlingData.speedOfSound > 0, "SpeedOfSound must be larger than 0");
    lsdfRDRHandlingData.revOverrideLSDF_RDR =
        Config::getBoolDef("Reverb:OverrideLSDF_RDR", EAR::Variant(overrideLSDFRDR));
    lsdfRDRHandlingData.revLsdfRDRToActualRDRdB =
        Config::getFloatDef("Reverb:LSDF_RDRToActualRDR", EAR::Variant(lsdfRDRToActualRDR));
    lsdfRDRHandlingData.revLsdfTheoreticalRDRtoActualRDRdB = Config::getFloatDef(
        "Reverb:TheoreticalRdrToActualRdr", EAR::Variant(theoreticalRDRToActualRDR));

    m_defaultSorGain =
        Config::getFloatDef("Reverb:SecondOrderReverbGain", EAR::Variant(DEFAULT_REVERB_SOR_GAIN));
    m_unrestrictedSor = Config::getBoolDef(
        "Reverb:SecondOrderReverbUnrestricted", EAR::Variant(DEFAULT_REVERB_SOR_UNRESTRICTED));
    m_noLineOfSightCoeff =
        Config::getFloatDef("Reverb:NoLineOfSightGain", EAR::Variant(DEFAULT_REVERB_NO_LOS_COEFF));
}

bool ReverbStage::isPortalNearby(const EAR::Vector3& location) const
{
    for (const auto& portal : bitstreamPortalExtents) {
        auto closestPoint = portal.second->closestPointTo(location);
        auto distanceToPortal = (location - closestPoint).abs();
        float reverbExtentOffset = bitstreamPortalReverbOffsets.at(portal.first);
        if (distanceToPortal < reverbExtentOffset) {
            return true;
        }
    }

    return false;
}

// Initializing a struct with atomic members is tricky, so this function helps with that
void ReverbStage::ReverbData::fill(RenderItem* item, std::unique_ptr<Reverb> reverb,
    const EAR::Geometry* enclosure, int bitstreamReverbId,
    const mpegcore::Predelay& predelayInSamples)
{
    this->outputRenderItem = item;
    this->reverb = std::move(reverb);
    this->enclosure = enclosure;
    this->bitstreamReverbId = bitstreamReverbId;
    this->predelayInSamples = predelayInSamples;
}

void ReverbStage::ConnectionResolver::readPortalPayload(const PortalPayloadPtr data)
{
    // Deep copy since payload data is discarded after initialization
    for (const auto& portal : data->getPortals()) {
        m_portals.push_back(portal.get());
    }
    for (const auto& [ae, area] : data->getLineOfSightAreas()) {
        m_losAreas[ae] = std::make_unique<LineOfSightArea>(*area);
    }
    for (const auto& area : data->getOutsideLineOfSightAreas()) {
        auto areaCopy = std::make_unique<LineOfSightArea>(*area);
        m_outsideLosAreas.push_back(std::move(areaCopy));
    }
}

std::vector<const MPEGI::Portal*> ReverbStage::ConnectionResolver::getPortalsBetweenEnvironments(
    const std::string& targetEnv, const std::string& remoteEnv)
{
    std::vector<const MPEGI::Portal*> matchingPortals;
    for (const auto& portal : m_portals) {
        if (matchEnvironmentIds(portal->portalEnvironments, { targetEnv, remoteEnv })) {
            matchingPortals.push_back(portal);
        }
    }
    return matchingPortals;
}

// Return whether there is a line of sight from position to portal
bool ReverbStage::ConnectionResolver::lineOfSightExists(
    const EAR::Vector3& location, const std::string& acousticEnv, unsigned int portalId) const
{
    const auto aeBsIdIt = aeNameToBsId.find(acousticEnv);
    if (aeBsIdIt != aeNameToBsId.end()) {

        unsigned int aeBsId = aeBsIdIt->second;
        const auto losAreaIt = m_losAreas.find(aeBsId);
        if (losAreaIt == m_losAreas.end()) {
            // LOS area not defined for this AE, don't apply occlusion in any position
            return true;
        }
        const LineOfSightArea& losArea = *losAreaIt->second;
        int losVoxelIndex = getLosVoxelIndex(location, losArea);

        if (losVoxelIndex >= 0) {
            const auto& losVoxelsPortalIt = losArea.voxels.find(portalId);
            if (losVoxelsPortalIt == losArea.voxels.end()) {
                return false;
            }
            if (losVoxelsPortalIt->second.count(losVoxelIndex) > 0) {
                return true;
            }
        }
    } else {
        // Check for outside areas
        for (unsigned int idx = 0; idx < m_outsideLosAreas.size(); idx++) {
            const auto& outside = m_outsideLosAreas.at(idx);
            const auto& p1 = outside->losAreaCornerP1;
            const auto& p2 = outside->losAreaCornerP2;
            if ((location[0] >= p1[0] && location[0] <= p2[0])
                && (location[1] <= p1[1] && location[1] >= p2[1])
                && (location[2] >= p1[2] && location[2] <= p2[2])) {
                unsigned int losVoxelIndex = getLosVoxelIndex(location, *outside);
                if (losVoxelIndex >= 0) {
                    const auto& losVoxelsPortalIt = outside->voxels.find(portalId);
                    if (losVoxelsPortalIt == outside->voxels.end()) {
                        return false;
                    }
                    if (losVoxelsPortalIt->second.count(losVoxelIndex) > 0) {
                        return true;
                    }
                }
            }
        }
    }
    return false;
};

// Detection for changes in portal materials that would require recalculation.
bool ReverbStage::ConnectionResolver::portalMaterialChanged(
    const std::string& srcAcousticEnv, const std::string& acousticEnv)
{
    for (const auto& portal : m_portals) {
        if (matchEnvironmentIds(portal->portalEnvironments, { acousticEnv, srcAcousticEnv })) {
            // Use a pointer comparison for checking whether the material has changed.
            const Material* portalMaterial = portal->scenePortal->getMaterial().get();
            if (m_portalMaterials.first.find(portal->portalId) == m_portalMaterials.first.end()
                || portalMaterial != m_portalMaterials.first.at(portal->portalId)) {
                m_portalMaterials.second[portal->portalId] = portalMaterial;
                return true;
            }
        }
    }
    return false;
}

const MPEGI::Portal* ReverbStage::ConnectionResolver::closestPortal(
    const EAR::Vector3& position, const std::string& subjectEnv) const
{
    std::optional<float> smallestDistance = std::nullopt;
    const MPEGI::Portal* closestPortal = nullptr;
    for (const auto portal : m_portals) {
        if (matchOneEnvironmentId(portal->portalEnvironments, subjectEnv)) {
            float distance = (portal->position - position).abs();
            if (smallestDistance == std::nullopt || distance < smallestDistance) {
                smallestDistance = distance;
                closestPortal = portal;
            }
        }
    }
    return closestPortal;
}

void ReverbStage::ConnectionResolver::updateMaterials()
{
    m_portalMaterials.first = m_portalMaterials.second;
}

void ReverbStage::ConnectionResolver::addAcousticEnvironmentBsId(
    const std::string& aeName, unsigned int bsId)
{
    aeNameToBsId[aeName] = bsId;
}

// Create a directivity filter with neural network algorithm for a particular directiveness model
mpegcore::GraphicEqParameters ReverbStage::calcNeuralNetDirectivityFilter(
    const MPEGI::DirectivitySpectrum& dirSpectrum, std::uint32_t sampleRateHz) const
{
    // Convert directivity spectrum 's float values to doubles in frequency response struct
    mpegcore::frontend::FrequencyResponse freqResp;
    auto toDouble = [&](auto f) { return static_cast<double>(f); };
    mpegcore::transformToBack(dirSpectrum.frequencies, freqResp.f, toDouble);
    mpegcore::transformToBack(dirSpectrum.coefficients, freqResp.response, toDouble);

    return calculateCascadeEqParameters(
        freqResp, mpegcore::frontend::octaveNeuralNetworkParameters(), sampleRateHz);
}

std::size_t ReverbStage::complexityLevelToNumDelayLines(int complexityLevel) const
{
    const std::size_t numDelayLineCounts = mpegcore::NUM_SUPPORTED_DELAY_LINE_COUNTS;
    std::array<size_t, numDelayLineCounts> delayLineCounts =
        mpegcore::getSupportedDelayLineCounts();
    switch (complexityLevel) {
    case 1:
        return delayLineCounts[0];
        break;
    case 2:
        return delayLineCounts[1];
        break;
    case 3:
        return delayLineCounts[2];
        break;
    case 4:
        return delayLineCounts[3];
        break;
    case 5:
        return delayLineCounts[4];
        break;
    default:
        return delayLineCounts[2];
        break;
    }
}

} // namespace MPEGI
