/***************************************************************************
This software module was originally developed jointly by

Nokia Technologies Ltd. and Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. and Koninklijke Philips N.V. retain full right to
modify and use the code for their own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include "Reverb.hpp"
#include "ReverbConfigCalc.h"

#include <atomic>
#include <cstddef>
#include <map>
#include <memory>
#include <optional>
#include <string>
#include <tuple>
#include <unordered_map>
#include <vector>

#include <MPEGI/Bitstream/PortalPayload.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Helpers/ReverbConstants.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderList.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>
#include <mpegcore/frontend/GraphicEqFilter.hpp>

#include <EAR/IIRFilter.h>
#include <EAR/ListenerVoiceSource.h>
#include <EAR/LoudspeakerLayout.h>
#include <EAR/PointSource.h>
#include <EAR/SceneObject.h>
#include <EAR/Uncopyable.h>
#include <MPEGI/Renderer/RenderStage.h>
#include <mpegcore/ReverberatorParameters.hpp>
#include <mpegcore/frontend/GraphicEqFilter.hpp>

namespace MPEGI {

class ReverbPayload;
class StreamBuffer;
struct Reverb;

//! Apply late reverberation effect to audible items in an acoustic environment
/*!
 * Inputs:
 *
 * Primary and higher-order ambisonics render items that are not inactive or orphaned.
 * Items with NoReverb configuration parameter are excluded.
 *
 * Outputs:
 *
 * ReverbStage creates two kinds of output render items: `ItemType::Pannable` and
 * `ItemType::Reverb`.
 *
 * - `ItemType::Pannable` items contains reverb output delay line audio for the reverb environment
 *   in which the listener is currently and the output positions are headtracked based on listener
 *   orientation so that the output directions remain static when the listener's head turns.
 *
 * - `ItemType::Reverb` items contain reverb output delay line audio for reverb environments
 *   regardless of whether the listener is located in the environment or not. The output positions
 *   are static, i.e.  they do not change based on the listener's head orientation.
 *
 * How it works:
 *
 * ReverbStage creates a late reverberation effect from its inputs. Each acoustic environment
 * in the scene has its own reverb environment with distinct inputs and outputs.
 *
 * Stages order:
 * - Required before: RoomAssignmentStage
 * - Required after: PannerStage, PortalStage
 *
 * Config options:
 *
 * Reverb:Gain (float) controls overall reverb gain in all acoustic environments.
 *
 */
class ReverbStage : public RenderStage
{
public:
    ReverbStage(const int16_t dirFltOrder = 10, const int16_t t60FltOrder = 8,
        const int16_t colorFltOrder = 12);
    ~ReverbStage();

    typedef std::shared_ptr<EAR::DirectivityModel> DirectivityPtr;
    typedef std::shared_ptr<mpegcore::frontend::GraphicEqFilter> FilterPtr;
    typedef std::shared_ptr<const PortalPayload> PortalPayloadPtr;

    struct DirectivityFilterData
    {
        FilterPtr filter{ nullptr };
        std::shared_ptr<StreamBuffer> inputBuffer{ nullptr };

        DirectivityFilterData(FilterPtr filter, std::shared_ptr<StreamBuffer> inputBuffer)
            : filter{ filter }
            , inputBuffer{ inputBuffer }
        {}
    };

    typedef std::shared_ptr<DirectivityFilterData> DirectivityFilterPtr;
    DirectivityFilterPtr getDirectivityFilterForItem(
        const RenderItem& item, const std::string& acousticEnvironmentId);

    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;
    void processBlock(DspContext ctx) override;

    struct ItemState : RenderItem::ExtraState
    {
        ItemState(bool isImmersiveReverb)
            : isImmersiveReverb(isImmersiveReverb)
        {}
        // True for items which are intended for the panner stage to simulate immersive reverb
        bool isImmersiveReverb;

        // Maps from portal id to portalContribution. The portalContributions relates to the amount of energy that reaches a portal from a source. This has not taken into account Transmission or Coupling coefficients yet.
        std::map<int, float> portalContributions;
    };

    struct Node
    {
        Portal* incomingPortal = nullptr;
        std::string acousticEnvironment = "";

        Node(Portal* p, const std::string& AE)
            : incomingPortal(p)
            , acousticEnvironment(AE){};
    };

    // A path contains a linear chain of nodes, which consist of acoustic environments together with the portal that connects them to the previous node. The first node in a path shall always have incomingPortal==nullptr, as there is no previous node.
    using Path = std::vector<Node>;

    // Find all paths from fromAE to toAE, given connections in portals.
    static std::vector<Path> getAllPaths(const std::string& fromAE, const std::string& toAE,
        const std::vector<std::unique_ptr<Portal>>& portals);

    // Accumulates the transmission, coupling and portal2portal contributions along a path.
    // portal2portal factors should have as key {fromPortalId, toPortalId} and value the portal2portal factor for that combination of portals.
    static float getHigherOrderSourceContribution(
        const Path& path, const std::map<std::pair<int, int>, float>& portal2portalFactors);

protected:
    struct ReverbData
    {
        void fill(RenderItem* item, std::unique_ptr<Reverb> reverb, const EAR::Geometry* enclosure,
            int bitstreamReverbId, const mpegcore::Predelay& predelayInSamples);

        RenderItem* outputRenderItem;
        std::unique_ptr<Reverb> reverb;
        const EAR::Geometry* enclosure;
        std::atomic<float> gain{ 0.0f };
        int bitstreamReverbId;
        mpegcore::Predelay predelayInSamples;

        // Output from previous block is used for second order reverb
        std::unique_ptr<mpegcore::AudioBuffer> previousOutput;

        // Second order reverb coefficients
        std::unique_ptr<std::unordered_map<std::string, std::atomic<float>>> connectedEnvironments;

        // Second order reverb filter
        std::unique_ptr<IIRFilter> m_sorFilter;
    };

    struct LSDFRDRHandlingData
    {
        float speedOfSound{ DEFAULT_REVERB_SPEED_OF_SOUND };
        float revLsdfRDRToActualRDRdB{ DEFAULT_REVERB_RDR_LSDF_TO_RDR_ACTUAL };
        float revLsdfTheoreticalRDRtoActualRDRdB{ DEFAULT_REVERB_RDR_THEORETICAL_TO_RDR_ACTUAL };
        bool revOverrideLSDF_RDR{ DEFAULT_REVERB_LSDF_RDR_OVERRIDE };
    };

    struct DSPItem
    {
        std::shared_ptr<StreamBuffer> inputBuffer{ nullptr };
        float gainInBoundedAE{ 0.0f };
        float gainInDefaultAE{ 0.0f };
        DirectivityFilterPtr directivityFilter;
        // Reverb id of the item's location. Used for connected reverb.
        std::string sourceAcousticEnvironmentId;

        DSPItem(std::shared_ptr<StreamBuffer> inputBuffer, float gain, float gainInDefaultAE,
            DirectivityFilterPtr directivityFilter, const std::string& sourceAcousticEnvironmentId)
            : inputBuffer{ inputBuffer }
            , gainInBoundedAE{ gain }
            , gainInDefaultAE{ gainInDefaultAE }
            , directivityFilter{ directivityFilter }
            , sourceAcousticEnvironmentId{ sourceAcousticEnvironmentId }
        {}
    };

    struct DspData : RenderStage::DspData
    {
        // Map acousticEnvironmentId -> DSPItems, one DSPItem per RenderList item that reverb should render
        std::unordered_map<std::string, std::vector<DSPItem>> dspItems;

        // Listener's current acousticEnvironmentId
        std::string listenerAcousticEnvironmentId = "";

        // Second order reverb gain
        float sorGain{ 0.f };

        // DspData has its own map of reverb data that is populated before lining up data.
        // ReverbData struct has pointers to items that are created at initialization
        // stage and will remain valid throughout the rendering process.
        std::unordered_map<std::string, const ReverbData&> reverbData;
    };

    // Crossfade distance between reverb environments
    float m_crossfadeFadeInDistance{ DEFAULT_REVERB_FADE_IN_DISTANCE };

    struct AcousticEnvironmentData
    {
        float predelayFactor{ DEFAULT_REVERB_EIF_PREDELAY_FACTOR };
        float distanceGainDropDb{ DEFAULT_REVERB_DISTANCE_GAIN_DROP_DB };
        float minimumDistance{ DEFAULT_REVERB_MINIMUM_DISTANCE };
        float distanceGainDbFactor{ 0.0f };
        float gain{ 0.0f };
        float delayRatio{ 0.0f };
        float delayMinMs{ 0.0f };

        AcousticEnvironmentData(float predelayFactor, float distanceGainDropDb,
            float minimumDistance, float gain, float delayRatio, float delayMinMs)
            : predelayFactor{ predelayFactor }
            , distanceGainDropDb{ distanceGainDropDb }
            , minimumDistance{ minimumDistance }
            , distanceGainDbFactor{ 0.0f }
            , gain{ gain }
            , delayRatio{ delayRatio }
            , delayMinMs{ delayMinMs }
        {
            distanceGainDbFactor = distanceGainDropDb / std::log10(2.0);
#ifdef MPEGI_CFP_SUBMISSION
            EAR_CHECK_REL(predelayFactor > 0, "Predelay factor needs to be positive.");
            EAR_CHECK_REL(distanceGainDropDb > 0, "Distance gain drop db needs to be positive.");
            EAR_CHECK_REL(minimumDistance > 0, "Minimum distance needs to be positive.");
#else
            float const minimumAllowedValue = 1e-10f;
            if (predelayFactor < minimumAllowedValue) {
                this->predelayFactor = minimumAllowedValue;
            }
            if (distanceGainDropDb < minimumAllowedValue) {
                this->distanceGainDropDb = minimumAllowedValue;
            }
            if (minimumDistance < minimumAllowedValue) {
                this->minimumDistance = minimumAllowedValue;
            }
#endif
        }
    };

    // Default acoustic environment is a special AE that does not have boundaries
    struct DefaultAcousticEnvironment
    {
        // Gain of default AE
        float gain{ 0 };

        // Reverberation parameters of the default AE
        ReverbData* reverb;
    };

    double rdrLogToRdrLin(double rdrLog) const;
    typedef std::pair<const std::string, std::map<int, std::pair<float, EAR::Vector3>>>
        PortalConnection;

    // Map from acoustic environment to connected environment
    typedef std::unordered_map<std::string, std::vector<std::string>> ReverbConnections;

    // Resolves which acoustic environments are audible through portals in the
    // listener's acoustic environment.
    class ConnectionResolver : Uncopyable
    {
    public:
        // Reads data needed by connected reverb from reverb payload
        void readPortalPayload(const PortalPayloadPtr data);

        void addAcousticEnvironmentBsId(const std::string& aeName, unsigned int bsId);

        // Return all portals between environments
        std::vector<const MPEGI::Portal*> getPortalsBetweenEnvironments(
            const std::string& targetEnv, const std::string& remoteEnv);

        // Return whether targetEnv is reverberated by sources in remoteEnv
        bool connectionExistsBetween(const std::string& targetEnv, const std::string& remoteEnv);

        // Return whether there is a line of sight from position to portal
        bool lineOfSightExists(const EAR::Vector3& location, const std::string& acousticEnv,
            unsigned int portalId) const;

        const MPEGI::Portal* closestPortal(
            const EAR::Vector3& position, const std::string& subjectEnv) const;

        // Detection for changes in portal materials that would require recalculation.
        bool portalMaterialChanged(
            const std::string& srcAcousticEnv, const std::string& acousticEnv);

        void updateMaterials();

    private:
        // Portals modelling object source direct propagation values over the scene
        std::vector<MPEGI::Portal*> m_portals;

        // Stored portal material pointers for each MPEGI::Portal id to detect changes in material
        // Using a raw pointer for fast comparison.
        std::pair<std::unordered_map<int, const Material*>,
            std::unordered_map<int, const Material*>>
            m_portalMaterials;

        // Mapping between acoustic environment names and bitstream identifiers
        std::unordered_map<std::string, int> aeNameToBsId;

        // Line of sight areas ordered by acoustic environment id
        std::map<unsigned int, std::unique_ptr<LineOfSightArea>> m_losAreas;

        // Line of sight areas for outside environments
        std::vector<std::unique_ptr<LineOfSightArea>> m_outsideLosAreas;
    };

    void setConfigurableVariables();

    float listenerDistanceToEnv(const EAR::Geometry& reverbEnv);

    float calculateCrossfadeAttenuation(const EAR::Geometry& enclosure);

    float getItemGainBasedOnAE(const DSPItem& item, const ReverbData& reverbUnit);

    float calculateReverbEnvGain(const ReverbData& reverbUnit, bool isListenerInEnv);

    void recalculateSecondOrderReverbCoefficients();
    float calculateSecondOrderReverbGain();

    double calculateDistanceGain(
        double distance, double refDistance, const AcousticEnvironmentData& env) const;

    std::pair<float, float> getExternalSourceContribution(
        RenderItem& item, std::string acousticEnvironmentId) const;

    // Augmented reality reverb calculation functions
    std::optional<float> calculateTheoreticalRdr(
        double rt60, double roomVolume, double eifPredelay) const;

    // Returns a pair where first gain is for bounded AEs and second gain is for default AE
    virtual std::pair<float, float> calculateItemGains(float distance, const RenderItem& item);

    // Store output for next update cycle for second order reverberation
    void savePreviousReverbOutput(const ReverbData& reverbUnit);

    EAR::Geometry::BoundingBox getAEBoundingBox(const ReverbEnv* re);

    template <typename TInput> bool hasRdrValues(TInput& frequencyData) const
    {
        bool rdrValuesGiven = false;
        for (std::size_t i{ 0 }; i < frequencyData.size(); ++i) {
            if (frequencyData[i].rdr.has_value()) {
                rdrValuesGiven = true;
            }
        }
        return rdrValuesGiven;
    }

    bool calculateReverbEnvGains();
    float reverbGain{ 1.0 };

    template <typename TInput>
    void setRdrValuesFromTheoreticalModel(
        TInput& frequencyData, double roomVolume, double predelay) const
    {
        for (std::size_t i{ 0 }; i < frequencyData.size(); ++i) {
            std::optional<float> rdr =
                calculateTheoreticalRdr(frequencyData[i].rt60, roomVolume, predelay);
            frequencyData[i].rdr = rdr;
        }
    }

    template <typename TInput>
    void addOffsetToRdrValues(TInput& frequencyData, float rdrValueOffsetLog) const
    {
        for (std::size_t i{ 0 }; i < frequencyData.size(); ++i) {
            if (frequencyData[i].rdr.has_value()) {
                float newRdrValueLog =
                    10.0 * std::log10(frequencyData[i].rdr.value()) + rdrValueOffsetLog;
                double newRdrValueLin = std::pow(10.0, newRdrValueLog / 10.0);
                frequencyData[i].rdr = std::optional<float>(newRdrValueLin);
            }
        }
    }

    std::size_t B{ 0 };
    void setRoomVolumeAndLongestDim(EAR::Geometry::BoundingBox const& boundingBox,
        double& roomVolume, double& longestDimension) const;
    bool m_headTrackingEnabled{ false };

    double calculateEifPredelay(double longestRoomDimension) const;

    std::unordered_map<std::string, ReverbData> reverbData;
    EAR::Vector3 listenerLocation;
    std::string listenerAcousticEnvironmentId;
    std::string previousListenerAcousticEnvironmentId;

    AcousticEnvironmentData defaultAcousticEnvironmentData;
    AcousticEnvironmentData acousticEnvironmentData;

    // Each reverb environment has their own collection of directivity filters
    using DirectivityFilterMap = std::unordered_map<DirectivityPtr, DirectivityFilterPtr>;
    std::unordered_map<std::string, DirectivityFilterMap> m_directivityFilterMap;

    // Set of different predelay values applicable for a directivity filter
    std::unordered_map<DirectivityFilterPtr, std::set<std::size_t>> directivityBusAdded;

    std::vector<RenderItem*> headtrackedOutputRenderItems;
    std::vector<EAR::Position> headtrackedOutputRenderItemDefaultPositions;
    mpegcore::OutputMapping outputMapping;

    LSDFRDRHandlingData lsdfRDRHandlingData;

    void CalculateBaseConfig(std::shared_ptr<SetBaseConfigStruct>& baseConfigOut,
        unsigned int buildingId, unsigned int nrFbLoops,
        std::shared_ptr<const ReverbParameters> room, bool isDefaultAE, const float fs);
    void CalculateT60(std::shared_ptr<T60Settings>& T60ConfigOut, const unsigned int roomId,
        std::shared_ptr<lateReverb> lateRevOut, IIRFilterType type);
    void CalculateLateReverb(std::shared_ptr<lateReverb>& lateRevOut,
        const std::shared_ptr<FreqGrid> FreqGridInputData,
        std::shared_ptr<const ReverbParameters> room);
    void CalculateColoration(std::shared_ptr<ColorParams>& ColorOut, const unsigned int roomId,
        const std::shared_ptr<lateReverb> lateRevOut,
        const std::shared_ptr<SetBaseConfigStruct> baseConfigOut, IIRFilterType type,
        bool isDefaultAE);

    void SetAcousticEnvironmentProperties(
        unsigned int buildingIdx, int roomId, float dx, float dy, float dz, int nrAcEnvsProvided)
    {
        AcousticEnvironmentProperties aep;
        aep.roomId = roomId;
        aep.dimensions[0] = dx;
        aep.dimensions[1] = dy;
        aep.dimensions[2] = dz;
        AcEnvProp[buildingIdx] = aep;
    };

    void initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state) override;

    inline const ReverbDSPCfgStruct& GetRevDSPCfg(unsigned int roomId) const
    {
        return RevDSPCfg.at(roomId);
    }

    std::size_t complexityLevelToNumDelayLines(int complexityLevel) const;

protected:
    std::vector<AcousticEnvironmentProperties> AcEnvProp;
    std::unordered_map<int, float> m_itemDistGain;

private:
    const int16_t dirFltOrder, t60FltOrder, colorFltOrder;

    std::shared_ptr<const MPEGI::ReverbPayload> bitstreamReverbPayload;
    std::shared_ptr<const MPEGI::PortalPayload> bitstreamPortalPayload;
    std::map<int, EAR::Geometry*> bitstreamPortalExtents;
    std::map<int, float> bitstreamPortalReverbOffsets;

    std::unordered_map<unsigned int, std::shared_ptr<const ReverbParameters>> reverbBsParameters;

    std::unordered_map<unsigned int, ReverbDSPCfgStruct> RevDSPCfg;

    // Cached source coefficient values for stationary items.
    // Key: render item index
    using CoeffCache = std::unordered_map<int, std::optional<std::pair<float, float>>>;

    // Key: acoustic environment name
    mutable std::unordered_map<std::string, CoeffCache> m_coeffCache;

    ConnectionResolver m_connectionResolver;

    std::unordered_map<unsigned int, EAR::HomogenMatrix> portalRotationMatrices;

    bool calcOutsideEnergy{ false };

    EAR::ListenerVoiceSource* listenerVoice = nullptr;

    friend class ReverbStageTester;

    void calculateReverberatorConfig(const Bitstream& bitstream, RenderItemStore* itemStore,
        const StreamManager* streamManager, const SceneState* state, IIRFilterType filterType,
        std::optional<int> defaultAEId);

    std::pair<float, float> calculateFilteredItemGains(float distance, const RenderItem& item,
        std::optional<float> commonComponentClrFlt, bool isDefaultAE);

    void SetAcousticEnvironmentProperties(const SceneState* state);

    void initializeConfigurationParameters();

    mpegcore::ReverberatorParameters generateReverbParameters_AR(
        const std::unique_ptr<MPEGI::ReverbEnv>& reverbEnv, const float fs);
    void populateAcousParamDecayPropertiesFromRevEnv(
        mpegcore::AcousticParameters& acousticParameters, const MPEGI::ReverbEnv& reverbEnv);
    void updateRDRWithConfigSettings(mpegcore::AcousticParameters& acousticParameters,
        const double roomVolume, const double diffuseOnset);
    void validatePredelayFactor(const float predelay, const float predelayFactor,
        const mpegcore::AcousticParameters& acousticParameters);

    mpegcore::GraphicEqParameters calcNeuralNetDirectivityFilter(
        const MPEGI::DirectivitySpectrum& dirSpectrum, std::uint32_t sampleRateHz) const;

    std::pair<float, float> getDownmixGainForInternalSources(RenderItem& item,
        RenderList& renderList, const ReverbData& acousticEnv, std::string acousticEnvironmentId,
        std::string listenerAcousticEnvironmentId, bool isExplicitPortalMode,
        std::vector<float>& totalExternalContributions, int itemIdx);

    std::pair<float, float> GetSourceContribution(
        RenderItem& item, std::string acousticEnvironmentId);

    bool isPortalNearby(const EAR::Vector3& location) const;

    std::pair<float, float> sourceContributionToDownmixGain(float sourceContributionPortion,
        float dist, const MPEGI::RenderItem& item, const ReverbData& acousticEnv,
        const std::string& acousticEnvironmentId);

    bool DimensionUpdated();

    // Current level of fading and the maximum level of fading as number of cycles
    std::pair<unsigned int, unsigned int> m_sorFade{ 0, 0 };

    // Unrestricted second order reverb means that is not restricted to listener's room
    bool m_unrestrictedSor;

    // Default gain for second order reverb
    float m_defaultSorGain;

    // Second order reverb filter buffer as a class member to avoid repeated memory allocations
    mpegcore::AudioBuffer m_sorFilterBuffer;

    // Attenuation for items out of line of sight
    float m_noLineOfSightCoeff;
};
} // namespace MPEGI
