/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "quasiNewton.h"

void get_magnitude_response(float* x, double* w, int IIROrder, float MResult[NR_FILTER_DESIGN_FREQ])
{
    //   returns the magnitude response of an even Nth order IIR filter with
    //   parameters x of form:
    //
    //   x = [b01, b11, a01, a11, ...b0j, b1j, a0j, a1j H0]      [length 2*N+1]
    //
    //   N/2 second order sections normalized to have 4 non-unitary coefficients
    //   and a common gain H0
    //
    //   where the transfer function is given by
    //
    //   H(z) = H0 * prod((b0j + b1jz + z^2)/(a0j + a1jz + z^2)) over J
    //
    //   the response is computed over normalized frequency points w in [0,0.5]
    //   w is a vector of length L. (w = 2*pi*f/fs)
    //
    //   J = N/2 where N is the even order of the filter response
    //
    //   Based on Sec. 16.2 of Digital Signal Processing by Antoniou
    //
    //   Returns the magnitude spectrum of length(w) = L.
    const int DynVectDim =
        (IIROrder * 2) + 1; // (IIROrder == 8)? SHORT_QNVECT_SIZE : LONG_QNVECT_SIZE;

    const int nrFreq = NR_FILTER_DESIGN_FREQ;

    int nrSos = IIROrder / 2; // Number of second order sections

    // H0 is the last element of the array x
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        MResult[i] = *(x + DynVectDim - 1);

    for (int freqIdx = 0; freqIdx != nrFreq; freqIdx++) {
        float cosw = cos(w[freqIdx]);
        float cos2w = cos(w[freqIdx] * 2);

        for (int sosIdx = 0; sosIdx != nrSos; sosIdx++) {
            // Matlab: idx0 = 4 * (j - 1) + 1; % skip over second stage parameters
            int idx0 = 4 * sosIdx;
            // Matlab: b0j = x(idx0);
            float b0 = *(x + idx0);
            // Matlab: b1j = x(idx0 + 1);
            float b1 = *(x + idx0 + 1);
            // Matlab: a0j = x(idx0 + 2);
            float a0 = *(x + idx0 + 2);
            // Matlab: a1j = x(idx0 + 3);
            float a1 = *(x + idx0 + 3);
            // //			float Num = 1.0 + b0 * b0 + b1 * b1 + (2 * b1 * ( 1 + b0 ) * cosw ) + ( 2 * b0 * cos2w );
            // //			float Den = 1.0 + a0 * a0 + a1 * a1 + (2 * a1 * ( 1 + a0 ) * cosw ) + ( 2 * a0 * cos2w );
            // // Adding brackets to force the calculation in the same order as in Matlab and get identical output
            // float Num = (1.0f
            //     + ((b0 * b0)
            //         + ((b1 * b1)
            //             + ((2.0f * (b1 * ((1.0f + b0) * cosw))) + (2.0f * (b0 * cos2w))))));
            // float Den = (1.0f
            //     + ((a0 * a0)
            //         + ((a1 * a1)
            //             + ((2.0f * (a1 * ((1.0f + a0) * cosw))) + (2.0f * (a0 * cos2w))))));

            //			float Num = 1.0 + b0 * b0 + b1 * b1 + (2 * b0 * ( 1 + b1 ) * cosw ) + ( 2 * b1 * cos2w );
            //			float Den = 1.0 + a0 * a0 + a1 * a1 + (2 * a0 * ( 1 + a1 ) * cosw ) + ( 2 * a1 * cos2w );
            // Adding brackets to force the calculation in the same order as in Matlab and get identical output
            float Num = (1.0f
                + ((b0 * b0)
                    + ((b1 * b1)
                        + ((2.0f * (b0 * ((1.0f + b1) * cosw))) + (2.0f * (b1 * cos2w))))));
            float Den = (1.0f
                + ((a0 * a0)
                    + ((a1 * a1)
                        + ((2.0f * (a0 * ((1.0f + a1) * cosw))) + (2.0f * (a1 * cos2w))))));
            if ((Num < 0) || (Den < 0)) {
                // Assert: Numerator or denominator should not be negative, but can
                //         happen due to numerical precision.
                //         Fix it and continue. Optimization should steer away from
                //         these points.

                if (Num < 0)
                    Num = 0; // will be increased later on
                if (Den < 0)
                    Den = 0; // will be increased later on
            } // if( (Num < 0 ) || ( Den < 0 ) )

            if (Den < 1e-7) {
                printfAlways("WARNING: Pole close to unit circle. den = %.4f\n", Den);
            }
            // Accumulate the effect of all second order sections, for each frequency
            // max() ensures no division by 0 is possible

            // Matlab: Magn(freqIdx) = Magn(freqIdx) * sqrt(num / max(den, 1e-10));
            Den = (Den < 1e-10) ? 1e-10 : Den;
            MResult[freqIdx] *= (float)sqrt((double)(Num / Den));
        } // for sosIdx

    } // iterate over frequency
} // get_magnitude_response

void get_error_gradient(float* x, double* w, float* e, float* M, int N,
    float outeg[][NR_FILTER_DESIGN_FREQ], int32_t order)
{
    //   Inputs: x - input parameter vector [length (2*iirOrder) +1]
    //           w - normalized frequency (w= 2 pi f / fs)
    //           e - error across elements in w
    //           Magn - estimated magnitude spectrum
    //           iirOrder - order of the desired filter
    //
    //   Function that returns the error gradient given the parameters in x and
    //   the magnitude response estimated over frequencies w at x.
    //
    //   e is the error across (normalized) frequency
    const int nrFreq = NR_FILTER_DESIGN_FREQ;
    const int nrSos = N / 2;
    const int DynVectDim = (order * 2) + 1;
    const int xlen = (2 * N) + 1;
    float s[NR_FILTER_DESIGN_FREQ];

    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        s[i] = (e[i] >= 0) ? 1 : -1;

    // Gradient vector is 2N + 1 x nrFreq, where nrFreq is number of freq points
    // double eg[VectDim][NR_FILTER_DESIGN_FREQ];

    // Matlab: eg(xlen, :) = 1 / (x(xlen)) *(s.*M).'; % H0 gradient, H0 is the last element of the array
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        outeg[DynVectDim - 1][i] = (1 / x[DynVectDim - 1]) * (s[i] * M[i]);
    }

    // Iterate over frequency
    for (int freqIdx = 0; freqIdx != NR_FILTER_DESIGN_FREQ; freqIdx++) {
        float cosw = cos(w[freqIdx]);
        float cos2w = cos(w[freqIdx] * 2);

        // Iterate over parameters
        for (int sosIdx = 0; sosIdx != nrSos; sosIdx++) {
            // Matlab: idx0 = 4 * (sosIdx - 1) + 1; % skip over second stage parameters
            int idx0 = 4 * sosIdx;

            // Matlab: b0 = x(idx0);
            float b0 = x[idx0];

            // Matlab: b1 = x(idx0 + 1);
            float b1 = x[idx0 + 1];

            // Matlab: a0 = x(idx0 + 2);
            float a0 = x[idx0 + 2];

            // Matlab: a1 = x(idx0 + 3);
            float a1 = x[idx0 + 3];

            // //			float Num = 1 + b0 * b0 + b1 * b1 + 2 * b1 * (1 + b0) * cosw + 2 * b0 * cos2w;
            // //			float Den = 1 + a0 * a0 + a1 * a1 + 2 * a1 * (1 + a0) * cosw + 2 * a0 * cos2w;
            // // Adding brackets to force same order of calculation as in Matlab
            // float Num = 1.0f
            //     + ((b0 * b0)
            //         + ((b1 * b1) + ((2.0f * (b1 * ((1.0f + b0) * cosw))) + (2.0f * (b0 * cos2w)))));
            // float Den = 1.0f
            //     + ((a0 * a0)
            //         + ((a1 * a1) + ((2.0f * (a1 * ((1.0f + a0) * cosw))) + (2.0f * (a0 * cos2w)))));

            //			float Num = 1 + b0 * b0 + b1 * b1 + 2 * b0 * (1 + b1) * cosw + 2 * b1 * cos2w;
            //			float Den = 1 + a0 * a0 + a1 * a1 + 2 * a0 * (1 + a1) * cosw + 2 * a1 * cos2w;
            // Adding brackets to force same order of calculation as in Matlab
            float Num = 1.0f
                + ((b0 * b0)
                    + ((b1 * b1) + ((2.0f * (b0 * ((1.0f + b1) * cosw))) + (2.0f * (b1 * cos2w)))));
            float Den = 1.0f
                + ((a0 * a0)
                    + ((a1 * a1) + ((2.0f * (a0 * ((1.0f + a1) * cosw))) + (2.0f * (a1 * cos2w)))));

            // Exceptions handling
            if (Num < 0 || Den < 0) {
                // Assert: Numerator or denominator should not be negative, but can
                //         happen due to numerical precision.
                //         Fix it and continue. Optimization should steer away from
                //         these points.
                Num = (Num > 1e-10) ? Num : 1e-10;
                Den = (Den > 1e-10) ? Den : 1e-10;
            }

            // max() ensures no division by 0 is possible
            float eps = 1.0e-10f;
            // Using fmax() seems to cause differences with Matlab.
            Num = (Num < eps) ? eps : Num;
            Den = (Den < eps) ? eps : Den;
            // Matlab: eg(idx0,     freqIdx) =  s(freqIdx) * (b0 + (1 + b1)     * cosw) / max(num, 1e-10) * Magn(freqIdx);
            outeg[idx0][freqIdx] = ((s[freqIdx] * (b0 + ((1 + b1) * cosw))) / Num) * M[freqIdx];

            // Matlab: eg(idx0 + 1, freqIdx) =  s(freqIdx) * (b1 + b0 * cosw + cos2w)   / max(num, 1e-10) * Magn(freqIdx);
            outeg[idx0 + 1][freqIdx] =
                ((s[freqIdx] * (b1 + ((b0 * cosw) + cos2w))) / Num) * M[freqIdx];

            // Matlab: eg(idx0 + 2, freqIdx) = -s(freqIdx) * (a0 + (1 + a1) * cosw) / max(den, 1e-10) * Magn(freqIdx);
            outeg[idx0 + 2][freqIdx] =
                ((-s[freqIdx] * (a0 + ((1 + a1) * cosw))) / Den) * M[freqIdx];

            // Matlab: eg(idx0 + 3, freqIdx) = -s(freqIdx) * (a1 + a0 * cosw + cos2w)         / max(den, 1e-10) * Magn(freqIdx);
            outeg[idx0 + 3][freqIdx] =
                ((-s[freqIdx] * (a1 + ((a0 * cosw) + cos2w))) / Den) * M[freqIdx];

        } // iterate over parameters

    } // iterate over frequency
} // get_error_gradient

float get_obj_func(QNewt* pIn, QNewt* pOut)
{
    float* x = pOut->x;
    double* w = pIn->w;
    int N = pIn->IIROrder;
    float* tgtMagn = pIn->tgtMagn;
    int p = pIn->p;

    float Mag[NR_FILTER_DESIGN_FREQ];
    get_magnitude_response(x, w, N, Mag);

    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        Mag[i] -= tgtMagn[i];

    float Eh = 0;
    float sumE = 0;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        float delta = Mag[i];
        if (delta < 0)
            delta = -delta;
        if (delta > Eh)
            Eh = delta;
        sumE += delta;
    }
    // Matlab:o = Eh * (sum((abs(e) ./ (Eh + 1e-8)).^p)). ^ (1 / p);
    float o;
    double oc = 0.0f;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        float delta = Mag[i];
        if (delta < 0)
            delta = -delta;
        if (0)
            oc += pow((delta / (Eh + 1.0e-8f)), p);
        else { // Needed to align with Matlab
            float r = 1.0f;
            for (int j = 0; j < p; j++)
                r = r * (delta / (Eh + 1.0e-8f));
            oc += r;
        }
    }

    o = pow((float)oc, 1.0f / p);
    o *= Eh;
    return o;
}

void get_obj_func_grad(float* x, double* w, int N, float* tgtMagn, int p, float* g, int32_t order)
{
    // Gets the Charalambous objective function gradient at x

    float Magn[NR_FILTER_DESIGN_FREQ];
    float e[NR_FILTER_DESIGN_FREQ];

    get_magnitude_response(x, w, N, Magn); // Quale é l' IIR Order da passare sotto?

    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        e[i] = Magn[i] - tgtMagn[i];

    const int DynVectDim = (order * 2) + 1;
    float ge[LONG_QNVECT_SIZE][NR_FILTER_DESIGN_FREQ];

    get_error_gradient(x, w, e, Magn, N, ge, order);

    // Matlab: Eh = max(abs(e));
    float Eh = 0;
    float sumE = 0;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        float delta = e[i];
        if (delta < 0)
            delta = -delta;
        if (delta > Eh)
            Eh = delta;
        sumE += delta;
    }

    // Matlab: g = sum((abs(e). / Eh).^p) ^ (1 / p - 1) * sum(repmat((abs(e.')./(Eh)).^(p-1),2*N+1,1).* ge,2);
    // Matlab decomposition:
    // I   : o = sum((abs(e)./Eh).^p)		double
    // II  : o = o ^ (1 / p - 1)			double
    // III : v1 = (abs(e.')./(Eh)).^(p-1)	double[NR_FILTER_DESIGN_FREQ]
    // IV  : v2 = repmat(v1,2*N+1,1)		double[VectDim][NR_FILTER_DESIGN_FREQ]
    // V   : v3 = sum(v2.* ge, 2)			double[VectDim]
    // VI  : g = o * v3						double[VectDim]

    // I   : o = sum((abs(e)./Eh).^p)		double
    float o;
    double oc = 0.0f;
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++) {
        float delta = e[i];
        if (delta < 0)
            delta = -delta;
        delta /= Eh;
        if (0)
            oc += pow(delta, p);
        else { // Needed to align with Matlab
            float r = 1.0f;
            for (int j = 0; j < p; j++)
                r = r * delta;
            oc += r;
        }
    }
    // II  : o = v1 ^ (1 / p - 1)			double
    o = pow((float)oc, (1.0f / p) - 1.0f);

    // III : v1 = (abs(e.')./(Eh)).^(p-1)	double[NR_FILTER_DESIGN_FREQ]
    float v1[NR_FILTER_DESIGN_FREQ];
    for (int i = 0; i != NR_FILTER_DESIGN_FREQ; i++)
        v1[i] = pow((((e[i] >= 0) ? e[i] : -e[i]) / Eh), (p - 1.0f));

    // IV  : v2 = repmat(v1,2*N+1,1)		double[VectDim][NR_FILTER_DESIGN_FREQ]
    // Optimized [D. Dara] into v2 =~ v1 with some changes in the subsequent matrix multiplication

    // V   : v3 = sum(v2.* ge, 2)			double[VectDim]
    // v2 is no longer a matrix but rather an array
    float prod[LONG_QNVECT_SIZE][NR_FILTER_DESIGN_FREQ];

    for (int i = 0; i != DynVectDim; i++) {
        for (int j = 0; j != NR_FILTER_DESIGN_FREQ; j++)
            prod[i][j] = v1[j] * ge[i][j]; // dot product
    }
    float v3[LONG_QNVECT_SIZE];

    for (int i = 0; i != DynVectDim; i++) {
        float sum = 0;
        for (int j = 0; j != NR_FILTER_DESIGN_FREQ; j++)
            sum += prod[i][j];
        v3[i] = sum;
    }

    // VI  : g = o * v3						double[VectDim]
    for (int i = 0; i != DynVectDim; i++)
        g[i] = o * v3[i];

} // get_obj_func_grad

float norm(float* vect, int length)
{
    float result = 0;
    for (int i = 0; i != length; i++) {
        float op = *(vect++);
        result += op * op;
    }
    return (float)sqrt((double)result);
}

void quasi_Newton(QNewt* pIn, QNewt* pOut)
{

    const int DynVectDim = (pIn->IIROrder * 2) + 1;

    float e1 = 1e-5;

    QNewt deltak;

    // x(k), the vector to be optimized
    for (int i = 0; i != DynVectDim; i++)
        pOut->x[i] = pIn->x0[i];

    int m = 0;

    // Line search variables - as suggested by the DSP book, section 16.4.3.
    float r = 0.1;
    float s = 0.7;
    float t = 0.1;
    float c = 0.75;
    int Mh = 600;
    float e2 = 1e-6; // % Termination criterion for line search optimization.
        // % 1e-6 seems to work better for single precision than 1e-10 from DSP book.

    // (1b) - Only S. Variable k is not used explicitly in the code
    int n = DynVectDim;

    // MATLAB: S = eye(n);
    float Sk[LONG_QNVECT_SIZE][LONG_QNVECT_SIZE] = { 0 };
    for (int i = 0; i != DynVectDim; i++)
        Sk[i][i] = 1.0;

    // Compute f(x0) and the corresponding gradient, g at x = x0;
    float fk;
    fk = get_obj_func(pIn, pOut);

    float gk[LONG_QNVECT_SIZE];
    get_obj_func_grad(pOut->x, pIn->w, pIn->IIROrder, pIn->tgtMagn, pIn->p, gk, pIn->IIROrder);

    m += 2;

    float fk_prev = fk;
    float dfk = fk;

    // start
    for (int k = 0; k != pIn->MAX_ITERATIONS; k++) {
        // Step 2 - (a) Set dk and (b) start finding alpha that minimizes f(x + a * dk), using the (c) line search.
        // (2a)

        float dk[LONG_QNVECT_SIZE];

        for (int i = 0; i != DynVectDim; i++) {
            float acc = 0;
            for (int k = 0; k != DynVectDim; k++) {
                acc += Sk[i][k] * gk[k];
            }
            dk[i] = -acc;
        }

        //  (2b)

        // Lower bound in line-search interval
        float aL = 0;

        // Upper bound on line - search interval, set to max value
        float aU = pow(2.0, 20);

        // Predicted change of f at x (= at lower boundary) related to x changing by dk.
        // Error gradient f at x (because aL * dk == 0 here) multiplied by dk
        // equals df for dx = dk.

        float xIn[LONG_QNVECT_SIZE];
        for (int i = 0; i != DynVectDim; i++)
            xIn[i] = pOut->x[i] + aL * dk[i];

        float gradOut[LONG_QNVECT_SIZE];
        double dfL;
        get_obj_func_grad(xIn, pIn->w, pIn->IIROrder, pIn->tgtMagn, pIn->p, gradOut, pIn->IIROrder);
        dfL = 0.0f;
        for (int i = 0; i != DynVectDim; i++)
            dfL += ((double)gradOut[i] * (double)dk[i]);

        float ak = 1;

        if (fabs(dfL) > e2)
            ak = -2 * (dfk / (float)dfL);

        if ((ak <= 0) || (ak > 1))
            ak = 1;

        //	 (2c)
        //	 Line search according to Fletcher inexact line search algorithm.
        //	 dk points towards decreasing direction. Search in range between
        //	 [aL * dk ... aU * dk] for minimum. The algorithm in the while loop
        //	 below determines a new, smaller search range around the minimum and
        //	 iterates. More details in DSP book, section 16.4.3.

        //	Error at lower boundary for line search algorithm
        float fL = fk;

        while (m < Mh) {

            if (ak == 0)
                printfAlways("warning: check value with ak\n");

            // Step 3 - (a) Set candidate deltak of line search, and (b) calculate fk = f(x + deltak).
            //          Still in step 2, but should not change when ending the while loop.
            // (3a)

            for (int i = 0; i != DynVectDim; i++)
                deltak.x[i] = ak * dk[i];

            // (3b)

            //  update fk by advancing x by deltak
            QNewt iteration;
            for (int i = 0; i != DynVectDim; i++)
                iteration.x[i] =
                    pOut->x[i] + deltak.x[i]; // this is the parm to be passed on to get_obj_func( )

            fk = get_obj_func(pIn, &iteration);
            m++;

            // % We have a function f(x) with a minimum in descent direction dk for x.
            // We have fL, the actual value of function f(x + aL * dk).
            // We have dfL, the slope of function f(x + aL * dk).
            // We have a first linear approximation of function f from known point aL.
            //   fL + aL * dfL
            //   This approximation is too steep.
            // We have r * dfL and s * dfL as modified slopes, with s > r.
            // We have a second linear approximation of function f from known point aL.
            //   fL + r * (a - aL) * dfL as a function of a.
            //   This approximation is too shallow and is going over the minimum.
            // The modified slope s * dfL represents a slope that is shallower
            //   than the first approximation but steeper than the second. It
            //   therefore represents a slope on the function f between
            //   x(k) + aL * dk and optimum x(k) + a_opt * dk.
            // We have a candidate ak for the optimum a_opt and a corresponding
            //   actual value of function fk = f(x(k) + ak * dk).
            // CONDITION 1
            //   If this actual value is less than the value predicted by the
            //   second (=shallow) approximation, then candidate ak is related
            //   to an x(k+1) = x(k) + ak * dk that lies before the place where
            //   the second approximation intersects with the rising function f
            //   beyond the minimum.
            // CONDITION 2
            //   If the actual slope of function f(x) at x(k) + ak * dk is larger
            //   (i.e. less steep) than the modified slope s * dfL, then
            //   candidate ak is related to an x(k+1) = x(k) + ak * dk that lies
            //   after the point where function f has slope s * dfL.
            // If both conditions are satisfied, ak is considered a valid,
            //   inexact candidate for a_opt and we can end the line search. The
            //   next iteration k+1 will continue with x(k+1) = x(k) + ak * dk
            // If condition 1 is violated, the candidate ak is beyond the
            //   intersection of the second approximation with rising f. It is
            //   considered an unacceptable candidate.
            //   A better candidate can be determined from an INTERPOLATION
            //   between aL and ak.
            // If condition 2 is violated, the candidate ak is between aL and
            //   the point where function f has the reduced slope s * dfL. It is
            //   considered an unacceptable candidate.
            //   A better candidate can be determined from an EXTRAPOLATION
            //   from ak. New aL can be set to ak.
            // Interpolation and extrapolation is done based on a quadratic
            //   representation for f(x) with known values and slopes at x and
            //   x + ak * dk.

            //  4 interpolation
            // When condition 1 is violated. More details in DSP book, section 16.4.3.
            if ((fk > (fL + r * (ak - aL) * dfL)) && (fabs(fL - fk) > e2) && (m < Mh)) {
                // Adjust upper limit aU to be ak (or keep aU if it is smaller)
                // a)
                if (ak < aU)
                    aU = ak;

                // Calculate new ak
                // b)
                float ah0;
                ah0 =
                    aL + (0.5 * (((ak - aL) * (ak - aL)) * dfL) / ((fL - fk) + ((ak - aL) * dfL)));

                // c)  Ensures ak moves at least a minimum amount (percentage of
                //     the range spanned by aL to aU) away from the boundary aL.

                float ah0L;
                ah0L = aL + (t * (aU - aL));

                if (ah0 < ah0L)
                    ah0 = ah0L;

                // d)  Ensures ak moves at least a minimum amount (percentage of
                //     the range spanned by aL to aU) away from the boundary aU.

                float ah0U;
                ah0U = aU - (t * (aU - aL));

                if (ah0 > ah0U)
                    ah0 = ah0U;

                // e)
                ak = ah0;

                continue;
            }

            // Predict change of error df at x + ak*dk related to x changing by dk, from gradient.

            for (int i = 0; i != DynVectDim; i++)
                xIn[i] = pOut->x[i] + ak * dk[i]; // parameter #1 to be passed
            get_obj_func_grad(
                xIn, pIn->w, pIn->IIROrder, pIn->tgtMagn, pIn->p, gradOut, pIn->IIROrder);
            dfk = 0;
            for (int i = 0; i != DynVectDim; i++)
                dfk += gradOut[i] * dk[i];

            m++;

            //	 EXTRAPOLATION
            //	 When condition 2 is violated. More details in DSP book, section 16.4.3.

            if ((dfk < (s * dfL)) && (fabs(fL - fk) > e2) && (m < Mh)) {
                // a) - Calculate delta for ak

                float da0;
                da0 = ((ak - aL) * dfk) / (dfL - dfk);

                // b) - If delta <= 0 it's not extrapolation.

                float ah0;
                if (da0 <= 0.0f)
                    ah0 = 2.0f * ak;
                else
                    // Otherwise add the delta
                    ah0 = ak + da0;

                // c
                // If the new ak is too close to the upper boundary, move it
                // back so that it is sufficiently far from the boundary.

                float ah0U;
                ah0U = ak + (c * (aU - ak));

                if (ah0 > ah0U)
                    ah0 = ah0U;

                // d - Adjust lower limit to be ak, and adjust fL and dfL accordingly

                aL = ak;
                fL = fk;
                dfL = dfk;

                // Set new ak to value just calculated
                ak = ah0;
                continue;

            } // if fp0 < s*fpL && abs(fL - f0) > e2 && m < Mh

            // When this point is reached, both conditions for a valid optimum
            // candidate are satisfied, a termination tolerance is reached or
            // the number of iterations exceeded the maximum.

            break; // break while
        } // while (m < Mh)

        //  Step 4 - (a) If L2 norm of deltak is below epsilon, stop (b) optimization and (c) output x = xk + deltak and fk.
        //	(4c)

        for (int i = 0; i != DynVectDim; i++)
            pOut->x[i] += deltak.x[i];

        dfk = fk_prev - fk;

        // (4a) - Stop if both the optimized vector and the error vector haven't
        //        changed much in the last iteration, or we spent too much time
        //        iterating in the line search.

        float fnorm = norm(deltak.x, DynVectDim);
        if (((fnorm < e1) && (fabs(dfk) < e1)) || (m >= Mh)) {
            if (m >= Mh)
                printfAlways("Ran out of line search iterations\n");
            // (4b)
            break;
        } else
            m = m; // breakpoint

        // Step 5 - (a) Compute gk for next iteration (k+1) with updated x and (b) set gamma_k = g(k+1) - g(k).

        fk_prev = fk;
        float gk_prev[LONG_QNVECT_SIZE];
        for (int32_t i = 0; i != DynVectDim; i++)
            gk_prev[i] = gk[i];

        // (5a)
        get_obj_func_grad(pOut->x, pIn->w, pIn->IIROrder, pIn->tgtMagn, pIn->p, gk, pIn->IIROrder);

        // (5b)
        float gammak[LONG_QNVECT_SIZE];
        for (int i = 0; i != DynVectDim; i++)
            gammak[i] = gk[i] - gk_prev[i];

        // Step 6 - Compute Sk matrix for next iteration (k+1): S(k+1) = S(k) + C
        // with correction matrix C

        float D = 0;
        for (int i = 0; i != DynVectDim; i++)
            D += (deltak.x[i] * gammak[i]);

        if (D <= 0) {
            // Step 7 - If S(k+1) is not positive definite, force it to become
            // positive definite.
            // MATLAB: Sk = eye(xLen);
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    Sk[i][j] = (i == j) ? 1 : 0;
        } else {
            //  C matrix according to Broyden-Fletcher-Goldfarb-Shanno (BFGS)
            //  From DSP book, chapter 16: A very important property of BFGS (and
            //  DFP) is that a positive definite matrix S(k) yields a positive
            //  definite S(k+1), provided that the line search in step 2 of the
            //  algorithm is exact. This property also holds in the case where an
            //  inexact line search is used in step 2, except that a scalar
            //  quantity inherent in the computation of Ck must be forced to
            //  remain positive.

            float nfactor = 0;
            for (int i = 0; i != DynVectDim; i++)
                nfactor += (deltak.x[i] * gammak[i]);

            // MATLAB: S = S + 1 / nfactor * ((1 + 1 / nfactor * gammak' * S * gammak) * (deltak * deltak') - (deltak*gammak'*S + S * gammak * deltak'));
            // MATLAB: C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * (deltak * deltak') - (deltak * gammak' * Sk + Sk * gammak * deltak'));

            // MATLAB Decomposition:
            // MM1 = deltak * gammak';	DynVectDim x DynVectDim	C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * (deltak * deltak') - (MM1 * Sk + Sk * gammak * deltak'));
            // MM4 = MM1 * Sk;			DynVectDim x DynVectDim	C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * (deltak * deltak') - (MM4 + Sk * gammak * deltak'));
            // MM1 = gammak * deltak';	DynVectDim x DynVectDim	C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * (deltak * deltak') - (MM4 + Sk * MM1));
            // MM2 = Sk * MM1;			DynVectDim x DynVectDim	C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * (deltak * deltak') - (MM4 + MM2));
            // MM1 = MM4 + MM2;			DynVectDim x DynVectDim	C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * (deltak * deltak') - MM1);
            // MM2 = deltak * deltak';	DynVectDim x DynVectDim	C = 1/nfactor*((1 + 1/nfactor * gammak' * Sk * gammak) * MM2 - MM1);
            // VV1 = Sk * gammak;		DynVectDim = [DynVectDim, DynVectDim] * DynVectDim  C = 1/nfactor*((1 + 1/nfactor * gammak' * VV1) * MM2 - MM1);
            // int1 = gammak' * VV1;	double 				C = 1/nfactor*((1 + 1/nfactor * int1) * MM2 - MM1);
            // int1 = (1 + 1/nfactor * int1);	double		C = 1/nfactor*(int1 * MM2 - MM1);
            // MM4 = int1 * MM2;		DynVectDim x DynVectDim	C = 1/nfactor*(MM4 - MM1);
            // MM4 -= MM1;				DynVectDim x DynVectDim	C = 1/nfactor*(MM4);
            // C = MM4 / nfactor;		DynVectDim x DynVectDim	C = 1/nfactor*(MM4);

            float MM1[LONG_QNVECT_SIZE][LONG_QNVECT_SIZE];
            float MM2[LONG_QNVECT_SIZE][LONG_QNVECT_SIZE];
            float MM4[LONG_QNVECT_SIZE][LONG_QNVECT_SIZE];
            float VV1[LONG_QNVECT_SIZE];
            float int1;

            // MM1 = deltak * gammak';	VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    MM1[i][j] = deltak.x[i] * gammak[j];

            // MM4 = MM1 * Sk;			VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++) {
                    float acc = 0;
                    for (int k = 0; k != DynVectDim; k++) {
                        acc += MM1[i][k] * Sk[k][j];
                    }
                    MM4[i][j] = acc;
                }
            MM1[0][0] = MM1[0][0]; // breakpoint

            // MM1 = gammak * deltak';	VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    MM1[i][j] = gammak[i] * deltak.x[j];

            // MM2 = Sk * MM1;			VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++) {
                    float acc = 0;
                    for (int k = 0; k != DynVectDim; k++) {
                        acc += Sk[i][k] * MM1[k][j];
                    }
                    MM2[i][j] = acc;
                }

            // MM1 = MM4 + MM2;				VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    MM1[i][j] = MM4[i][j] + MM2[i][j];

            MM1[0][0] = MM1[0][0]; // breakpoint;

            // MM2 = deltak * deltak';	VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    MM2[i][j] = deltak.x[i] * deltak.x[j];

            // VV1 = Sk * gammak;		VectDim = [VectDim, VectDim] * VectDim
            for (int i = 0; i != DynVectDim; i++) {
                float acc = 0;
                for (int k = 0; k != DynVectDim; k++) {
                    acc += Sk[i][k] * gammak[k];
                }
                VV1[i] = acc;
            }

            // int1 = gammak' * VV1;		VectDim x VectDim
            int1 = 0;
            for (int i = 0; i != DynVectDim; i++)
                int1 += gammak[i] * VV1[i];

            // int1 = (1 + 1/nfactor * int1);		double;
            int1 /= nfactor;
            int1 += 1.0;

            // MM4 = int1 * MM2;			VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    MM4[i][j] = MM2[i][j] * int1;

            // MM4 = MM4 - MM1;			VectDim x VectDim
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    MM4[i][j] = MM4[i][j] - MM1[i][j];

            // C = MM4 / nfactor;

            const float ratio = 1 / nfactor;
            float C[LONG_QNVECT_SIZE][LONG_QNVECT_SIZE];

            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    C[i][j] = (MM4[i][j] * ratio);

            // Update matrix S
            // MATLAB: 	Sk = Sk + C;
            for (int i = 0; i != DynVectDim; i++)
                for (int j = 0; j != DynVectDim; j++)
                    Sk[i][j] = Sk[i][j] + C[i][j];

            Sk[0][0] = Sk[0][0]; // breakpoint

        } // else (D > 0)

        // Step 8 - Set k = k+1 and go to step 2

    } // for (int k = 0; k != pIn->max_iterations; k++)

} // quasi_Newton
