/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2024.
***************************************************************************/

#pragma once

#include "../ReverbStage.h"
#include <EAR/Portal.h>
#include <MPEGI/Helpers/Types.h>
#include <doctest/doctest.h>

namespace MPEGI {

std::unique_ptr<Portal> createPortal(int id, const std::string& from, const std::string& to)
{
    auto portal = std::make_unique<Portal>();
    portal->portalId = id;
    portal->portalEnvironments = { from, to };
    return portal;
}

std::string pathToString(const ReverbStage::Path& path)
{
    std::string result;

    if (path.size() > 0) {
        result += path[0].acousticEnvironment;

        for (int id = 1; id < path.size(); ++id) {
            result += "->" + std::to_string(path[id].incomingPortal->portalId) + "->"
                + path[id].acousticEnvironment;
        }
    }

    return result;
}

// Returns true if paths contains a path which, when converted using pathToString equals pathString.
bool pathsContain(const std::vector<ReverbStage::Path>& paths, const std::string& pathString)
{
    auto correct = [pathString](
                       ReverbStage::Path path) { return pathToString(path) == pathString; };

    return std::find_if(paths.begin(), paths.end(), correct) != paths.end();
}

TEST_CASE("No path")
{
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    auto ADPaths = MPEGI::ReverbStage::getAllPaths("A", "D", portals);
    CHECK(ADPaths.size() == 0);
}

TEST_CASE("Linear path")
{
    // A -1- B - 2- C -3- D
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", "B"));
    portals.push_back(createPortal(2, "B", "C"));
    portals.push_back(createPortal(3, "C", "D"));

    {
        auto ADPaths = MPEGI::ReverbStage::getAllPaths("A", "D", portals);
        CHECK(ADPaths.size() == 1);
        CHECK(pathToString(ADPaths[0]) == "A->1->B->2->C->3->D");
    }

    {
        auto ACPaths = MPEGI::ReverbStage::getAllPaths("A", "C", portals);
        CHECK(ACPaths.size() == 1);
        CHECK(pathToString(ACPaths[0]) == "A->1->B->2->C");
    }
};

TEST_CASE("Linear path, but one AE is the empty string")
{
    // A -1- "" - 2- C -3- D
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", ""));
    portals.push_back(createPortal(2, "", "C"));
    portals.push_back(createPortal(3, "C", "D"));

    {
        auto ADPaths = MPEGI::ReverbStage::getAllPaths("A", "D", portals);
        CHECK(ADPaths.size() == 1);
        CHECK(pathToString(ADPaths[0]) == "A->1->->2->C->3->D");
    }

    {
        auto ACPaths = MPEGI::ReverbStage::getAllPaths("A", "C", portals);
        CHECK(ACPaths.size() == 1);
        CHECK(pathToString(ACPaths[0]) == "A->1->->2->C");
    }
};

TEST_CASE("Linear path, no portal to source or target")
{
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", "B"));
    portals.push_back(createPortal(2, "B", "C"));
    portals.push_back(createPortal(3, "C", "D"));

    auto allPaths = MPEGI::ReverbStage::getAllPaths("A", "ThisAeIsNotConnected", portals);
    CHECK(allPaths.size() == 0);

    auto allPaths2 = MPEGI::ReverbStage::getAllPaths("ThisAeIsNotConnected", "A", portals);
    CHECK(allPaths2.size() == 0);
};

TEST_CASE("Linear path, some portals reversed")
{
    // A -1- B - 2- C -3- D
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", "B"));
    portals.push_back(createPortal(2, "C", "B"));
    portals.push_back(createPortal(3, "D", "C"));

    auto ADPaths = MPEGI::ReverbStage::getAllPaths("A", "D", portals);

    CHECK(ADPaths.size() == 1);
    CHECK(pathToString(ADPaths[0]) == "A->1->B->2->C->3->D");
};

TEST_CASE("Disconnected path")
{
    // A -1- B - 2- C       D -3- E -4- F
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", "B"));
    portals.push_back(createPortal(2, "B", "C"));
    // NOTE: no path from C->D
    portals.push_back(createPortal(3, "D", "E"));
    portals.push_back(createPortal(4, "E", "F"));

    auto AFPaths = MPEGI::ReverbStage::getAllPaths("A", "F", portals);
    CHECK(AFPaths.size() == 0);
}

TEST_CASE("Path to self")
{
    // A -1- B - 2- C -3- D
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", "B"));
    portals.push_back(createPortal(2, "B", "C"));
    portals.push_back(createPortal(3, "C", "D"));

    {
        auto AAPaths = MPEGI::ReverbStage::getAllPaths("A", "A", portals);
        CHECK(AAPaths.size() == 1);
        CHECK(pathToString(AAPaths[0]) == "A");
    }
};

TEST_CASE("Triangle")
{
    // K --- 4
    // |     |
    // 5     L
    // |     |
    // M --- 6
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(4, "K", "L"));
    portals.push_back(createPortal(5, "K", "M"));
    portals.push_back(createPortal(6, "L", "M"));

    auto KMPaths = MPEGI::ReverbStage::getAllPaths("K", "M", portals);

    CHECK(KMPaths.size() == 2);

    CHECK(pathsContain(KMPaths, "K->5->M"));
    CHECK(pathsContain(KMPaths, "K->4->L->6->M"));
}

TEST_CASE("Triangle with legs")
{
    // P -7-  Q --- 9
    //        |     |
    //        8     R
    //        |     |
    // S -10- T ---11

    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(7, "P", "Q"));
    portals.push_back(createPortal(8, "Q", "T"));
    portals.push_back(createPortal(9, "Q", "R"));
    portals.push_back(createPortal(10, "S", "T"));
    portals.push_back(createPortal(11, "T", "R"));

    {
        auto PSPaths = MPEGI::ReverbStage::getAllPaths("P", "S", portals);

        CHECK(PSPaths.size() == 2);

        CHECK(pathsContain(PSPaths, "P->7->Q->8->T->10->S"));
        CHECK(pathsContain(PSPaths, "P->7->Q->9->R->11->T->10->S"));
    }
    {
        auto PTPaths = MPEGI::ReverbStage::getAllPaths("P", "T", portals);

        CHECK(PTPaths.size() == 2);

        CHECK(pathsContain(PTPaths, "P->7->Q->8->T"));
        CHECK(pathsContain(PTPaths, "P->7->Q->9->R->11->T"));
    }
}

TEST_CASE("Complex geometry")
{
    // Same as Triangle with legs, but with an additional room connected to S and R
    // P -7-  Q ----  9
    //        |       |
    //        8       |
    //        |       |
    // S -10- T -11-  R
    //  \             |
    //   12           |
    //    \-- U -----13

    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(7, "P", "Q"));
    portals.push_back(createPortal(8, "Q", "T"));
    portals.push_back(createPortal(9, "Q", "R"));
    portals.push_back(createPortal(10, "S", "T"));
    portals.push_back(createPortal(11, "T", "R"));
    portals.push_back(createPortal(12, "S", "U"));
    portals.push_back(createPortal(13, "U", "R"));

    {
        auto PSPaths = MPEGI::ReverbStage::getAllPaths("P", "S", portals);

        CHECK(PSPaths.size() == 4);

        CHECK(pathsContain(PSPaths, "P->7->Q->8->T->10->S"));
        CHECK(pathsContain(PSPaths, "P->7->Q->9->R->11->T->10->S"));
        CHECK(pathsContain(PSPaths, "P->7->Q->9->R->13->U->12->S"));
        CHECK(pathsContain(PSPaths, "P->7->Q->8->T->11->R->13->U->12->S"));
    }

    {
        auto PTPaths = MPEGI::ReverbStage::getAllPaths("P", "T", portals);

        CHECK(PTPaths.size() == 3);

        CHECK(pathsContain(PTPaths, "P->7->Q->8->T"));
        CHECK(pathsContain(PTPaths, "P->7->Q->9->R->11->T"));
        CHECK(pathsContain(PTPaths, "P->7->Q->9->R->13->U->12->S->10->T"));
    }
}

// Creates a material with given transmission and coupling coefficients. Has same t and c for every frequency.
std::shared_ptr<EAR::Material> createMaterial(float t, float c)
{
    auto mat = std::make_shared<EAR::Material>();
    mat->append(20, 0.f, 0.f, t, c);
    mat->append(20000, 0.f, 0.f, t, c);
    return mat;
}

TEST_CASE("Higher order source contribution test")
{
    // A -1- B - 2- C -3- D
    std::vector<std::unique_ptr<MPEGI::Portal>> portals;
    portals.push_back(createPortal(1, "A", "B"));
    portals.push_back(createPortal(2, "B", "C"));
    portals.push_back(createPortal(3, "C", "D"));

    // Create materials for the portals
    std::vector<std::shared_ptr<EAR::Material>> materials;
    materials.push_back(createMaterial(0.05f, 0.15f)); // Total energy going out = 0.2f
    materials.push_back(createMaterial(0.1f, 0.3f)); // Total energy going out = 0.4f
    materials.push_back(createMaterial(0.2f, 0.4f)); // Total energy going out = 0.6f

    // Create scene portals
    // NOTE: Portal Reverb extents aren't used in this test, but are only needed for initializing scene portals.
    std::array<std::shared_ptr<EAR::PortalReverbExtent>, 2> reverbExtents;
    float revOffset = 0.1f;
    reverbExtents[0] =
        EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(EAR::Vector3(), revOffset);
    reverbExtents[1] =
        EAR::SceneObjectFactory<EAR::PortalReverbExtent>::getNewObject(EAR::Vector3(), revOffset);

    std::vector<std::unique_ptr<EAR::Portal>> scenePortals;
    scenePortals.push_back(EAR::SceneObjectFactory<EAR::Portal>::getNewObject(
        "1", Position(), 1, materials[0], reverbExtents));
    scenePortals.push_back(EAR::SceneObjectFactory<EAR::Portal>::getNewObject(
        "2", Position(), 2, materials[1], reverbExtents));
    scenePortals.push_back(EAR::SceneObjectFactory<EAR::Portal>::getNewObject(
        "3", Position(), 3, materials[2], reverbExtents));

    // Assign scene portals to portals
    CHECK(portals.size() == scenePortals.size());
    for (int i = 0; i < portals.size(); ++i) {
        portals[i]->scenePortal = scenePortals[i].get();
    }

    const std::map<std::pair<int, int>, float> portal2portal = { { { 1, 2 }, 0.45f },
        { { 2, 3 }, 0.55f } };

    SUBCASE("Path to self")
    {
        auto AAPaths = MPEGI::ReverbStage::getAllPaths("A", "A", portals);
        CHECK(AAPaths.size() == 1);
        CHECK(pathToString(AAPaths[0]) == "A");

        float AASourceContrib =
            MPEGI::ReverbStage::getHigherOrderSourceContribution(AAPaths[0], portal2portal);

        // Path does not pass through any portals: source contribution is 1.
        CHECK(AASourceContrib == doctest::Approx(1.0f).epsilon(0.001));
    }

    SUBCASE("Path with 1 portal")
    {
        auto ABPaths = MPEGI::ReverbStage::getAllPaths("A", "B", portals);
        CHECK(ABPaths.size() == 1);
        CHECK(pathToString(ABPaths[0]) == "A->1->B");

        float ABSourceContrib =
            MPEGI::ReverbStage::getHigherOrderSourceContribution(ABPaths[0], portal2portal);

        // Source contrib is the total energy leaving Portal 1
        CHECK(ABSourceContrib == doctest::Approx(0.2f).epsilon(0.001));
    }

    SUBCASE("Path with 2 portals")
    {
        auto ACPaths = MPEGI::ReverbStage::getAllPaths("A", "C", portals);
        CHECK(ACPaths.size() == 1);
        CHECK(pathToString(ACPaths[0]) == "A->1->B->2->C");

        float ACSourceContrib =
            MPEGI::ReverbStage::getHigherOrderSourceContribution(ACPaths[0], portal2portal);

        // Multiply the total energy leaving each of the 2 portals together with the portal2portal factor.
        CHECK(ACSourceContrib == doctest::Approx(0.2f * 0.4f * 0.45f).epsilon(0.001));
    }

    SUBCASE("Path with 3 portals")
    {
        auto ADPaths = MPEGI::ReverbStage::getAllPaths("A", "D", portals);
        CHECK(ADPaths.size() == 1);
        CHECK(pathToString(ADPaths[0]) == "A->1->B->2->C->3->D");

        float ADSourceContrib =
            MPEGI::ReverbStage::getHigherOrderSourceContribution(ADPaths[0], portal2portal);

        // Multiply the total energy leaving each of the 3 portals together with the 2 portal2portal factors.
        CHECK(
            ADSourceContrib == doctest::Approx(0.2f * 0.4f * 0.6f * 0.45f * 0.55f).epsilon(0.001));
    }
}

} // namespace MPEGI