/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayloadHandler.h>
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Renderer.h>
#include <MPEGI/Stream/MaxStreamManager.h>

#include "../../RoomAssignmentStage/RoomAssignment.h"
#include "../ReverbStage.h"

#include "ReverbRoomsData.h"
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Helpers/Config.h>

#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Bitstream/PortalPayloadHandler.h>
#include <MPEGI/Bitstream/ReverbPayloadHandler.h>

#include <fstream>

#define cathedral_bitstream "reverb_replaced/EIF_Cathedral_LBR_P.bit"
#define hospital_bitstream "reverb_replaced/EIF_Hospital_LBR_P.bit"
#define basePath "reverb_replaced"

namespace MPEGI {
class DirectivityPayloadHandler : public Bitstream::PayloadHandler
{
public:
    //! Constructor
    DirectivityPayloadHandler() = default;

    //! Destructor
    ~DirectivityPayloadHandler() = default;

    //! Inherited from Bitstream::PayloadHandler
    std::shared_ptr<const Bitstream::Payload> unpack(
        Bitstream::RawPayload const& payload, const BitstreamConfig& config) override;
};
}; // namespace MPEGI

namespace ReverbDownmixCoefficient_test {

struct ReverbStage_test : public MPEGI::ReverbStage
{
    ReverbStage_test(EAR::Position _pos, unsigned int _acousticEnvironmentId)
        : pos(_pos)
        , acousticEnvironmentId(_acousticEnvironmentId)
    {
        AcEnvProp.resize(10);
    }

    void initialize(const MPEGI::Bitstream& bitstream, MPEGI::RenderItemStore* itemStore,
        const MPEGI::StreamManager* streamManager, const MPEGI::SceneState* state)
    {
        MPEGI::ReverbStage::initializeStage(bitstream, itemStore, streamManager, state);
    }

    std::unique_ptr<RenderStage::DspData> update(MPEGI::RenderList& renderList) override
    {
        renderList.listener.acousticEnvironmentId = acousticEnvironmentId;
        renderList.listener.position = pos;
        auto dspData = MPEGI::ReverbStage::update(renderList);
        lastDspData = static_cast<MPEGI::ReverbStage::DspData*>(dspData.get());
        return dspData;
    }

    // Keep a copy of the last generated dsp data for testing purposes
    MPEGI::ReverbStage::DspData* lastDspData = nullptr;

private:
    EAR::Position pos;
    unsigned int acousticEnvironmentId;
};

struct RoomAssignment_test : public MPEGI::RoomAssignmentStage
{
    RoomAssignment_test() {}

    void initialize(const MPEGI::Bitstream& bitstream, MPEGI::RenderItemStore* itemStore,
        const MPEGI::StreamManager* streamManager, const MPEGI::SceneState* state)
    {
        MPEGI::RoomAssignmentStage::initializeStage(bitstream, itemStore, streamManager, state);
    }
};

void registerPayloadHandlers()
{
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::DirectivityPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_DIRECTIVITY };
    });
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::ReverbPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_REVERB };
    });
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::PortalPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_PORTAL };
    });
}

std::shared_ptr<ReverbStage_test> stage = nullptr;

struct Renderer_test : public MPEGI::Renderer
{
    static CtorArgs createStages(const MPEGI::Bitstream& bitstream, const MPEGI::SceneState* state,
        std::shared_ptr<MPEGI::StreamManager> streamManager)
    {
        auto itemStore = std::make_unique<MPEGI::RenderItemStore>();
        auto stages = std::vector<std::shared_ptr<MPEGI::RenderStage>>{};
        auto roomAssignment = std::make_shared<RoomAssignment_test>();
        roomAssignment->initialize(bitstream, itemStore.get(), streamManager.get(), state);
        stages.push_back(roomAssignment);
        REQUIRE(stage != nullptr);
        stage->initialize(bitstream, itemStore.get(), streamManager.get(), state);
        stages.push_back(stage);
        return { state, streamManager, std::move(itemStore), std::move(stages) };
    }

    Renderer_test(const MPEGI::Bitstream& bitstream, const MPEGI::SceneState* state,
        std::shared_ptr<MPEGI::StreamManager> streamManager)
        : MPEGI::Renderer{ createStages(bitstream, state, streamManager) }
    {}
};

std::shared_ptr<MPEGI::RenderList> renderList;
std::unique_ptr<Renderer_test> renderer;
std::unique_ptr<MPEGI::Scene> scene;

void CleanUp()
{
    renderList = nullptr;
    renderer = nullptr;
    scene = nullptr;
    stage = nullptr;
}

void InitializeRenderer()
{
    DesignIIRRNG::ResetRng();

    registerPayloadHandlers();

    std::string str(basePath);

    MPEGI::Bitstream bitstream(hospital_bitstream);
    // MPEGI::Bitstream bitstream(cathedral_bitstream);
    bitstream.registerHandler(
        std::make_shared<MPEGI::EIFPayloadHandler>(basePath, false), MPEGI::Bitstream::PLD_SCENE);

    auto eifPayload = std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE));

    std::unique_ptr<LSDF::LSDFParser> lsdfObject = std::unique_ptr<LSDF::LSDFParser>(nullptr);
    LSDF::LSDFParser* lsdfPtr = nullptr;
    if (eifPayload->getSceneInfo().type == MPEGI::EIFPayload::SceneInfo::SceneType::AR) {
        lsdfPtr = lsdfObject.get();
    }

    const bool enableOSCListener = true;
    const bool offlineMode = true;
    std::shared_ptr<MPEGI::ExternalClock> clock = std::make_shared<MPEGI::ExternalClock>();
    scene =
        std::make_unique<MPEGI::Scene>(bitstream, clock, offlineMode, enableOSCListener, lsdfPtr);

    const float fs = 48000.0f;
    long B = 256;
    const int maxExternalInlets = 192; //Maximum number of connected inlets in the Max External
    std::shared_ptr<MPEGI::MaxStreamManager> streamManager =
        std::make_shared<MPEGI::MaxStreamManager>(bitstream, fs, B, maxExternalInlets);

    renderer = std::make_unique<Renderer_test>(bitstream, scene->getState(), streamManager);
    renderer->triggerControlStep();

    MPEGI::Bitstream::clearHandlerList();
}

std::vector<double> ConvertFromFilterCoefToMagnitudeSpectrum(
    const double* filterCoefficients, const size_t order)
{
    // w = 2 * PI * frequency / sample_freq;
    double w;
    double fs = 48000.0;
    std::vector<double> H_magn(240);
    const size_t _order = order + 1;
    // b = filterCoefficients[0 - _order]
    // a = filterCoefficients[_order+1 - 2*order]

    for (int i = 0, f = 20; f < 24000; f += 100, i++) {
        w = 2 * M_PI * f / fs;
        double sumAcos = 0.0;
        double sumBcos = 0.0;
        double sumAsin = 0.0;
        double sumBsin = 0.0;
        for (int j = 0; j < _order; ++j) {
            const double b = filterCoefficients[2 * j];
            const double a = filterCoefficients[(2 * j) + 1];
            sumBcos += ((b)*cos(j * w));
            sumBsin += ((b)*sin(j * w));
            sumAcos += ((a)*cos(j * w));
            sumAsin += ((a)*sin(j * w));
        }
        sumBcos = sumBcos * sumBcos;
        sumAcos = sumAcos * sumAcos;
        sumBsin = sumBsin * sumBsin;
        sumAsin = sumAsin * sumAsin;

        H_magn[i] = sqrt((sumBcos + sumBsin) / (sumAcos + sumAsin));
    }

    return H_magn;
}

#define NR_FILTER_TEST_FREQ (240)

void mpegInterp(
    float* fIn, float* valIn, float* fOut, float* valOut, int16_t lengthIn, int16_t lengthOut)
{
    if (lengthOut > NR_FILTER_TEST_FREQ)
        assert(false); // tested with this length, only

    for (int16_t i = 0; i != lengthOut; i++)
        valOut[i] = 0;

    int16_t i;
    for (i = 0; i != lengthOut; i++) {
        int16_t j;
        for (j = lengthIn - 1; j != -1; j--) {
            if (fIn[j] <= fOut[i])
                break;
        }

        if (j != -1) {

            if (j == lengthIn - 1)
                valOut[i] = valIn[lengthIn - 1];
            else {
                float x, x0, x1, y0, y1, y;
                x = fOut[i];
                x0 = fIn[j];
                x1 = fIn[j + 1];
                y0 = valIn[j];
                y1 = valIn[j + 1];
                y = y0 + ((x - x0) * ((y1 - y0) / (x1 - x0)));
                if (y < 0)
                    y = -25; // break point
                valOut[i] = y;
            } // else
        } else
            j++; // break point: why here

    } // for(int16_t i = 0; i != lengthOut; i++)
} // mpegInterp

void resampleToGrid(float* fIn, float* valIn, float* fOut, uint16_t fs, float* valOut,
    int16_t lengthIn, int16_t lengthOut)
{

    float fIn_internal[NR_FILTER_TEST_FREQ], valIn_internal[NR_FILTER_TEST_FREQ],
        fOut_internal[NR_FILTER_TEST_FREQ]; // in excess

    memcpy(fIn_internal, fIn, NR_FILTER_TEST_FREQ * sizeof(float));
    memcpy(valIn_internal, valIn, NR_FILTER_TEST_FREQ * sizeof(float));
    memcpy(fOut_internal, fOut, NR_FILTER_TEST_FREQ * sizeof(float));

    if (fIn[0] != 0) {
        // Matlab:      fIn   = [0;         fIn(:)];
        // Matlab:      valIn = [valIn(1);  valIn(:)];
        for (int16_t i = 0; i != lengthIn; i++) {
            fIn_internal[i + 1] = fIn[i];
            valIn_internal[i + 1] = valIn[i];
        }
        fIn_internal[0] = 0;
        valIn_internal[0] = valIn[0];
        lengthIn++;
    }

    if (fIn_internal[lengthIn] != (float)fs / 2) {
        // Matlab:      fIn   = [fIn(:);    fs/2];
        fIn_internal[lengthIn] = (float)fs / 2;
        // Matlab:      valIn = [valIn(:);  valIn(end)];
        valIn_internal[lengthIn] = valIn_internal[lengthIn - 1];
        lengthIn++;
    }

    // length: GRID_DIM to GRID_DIM+2 = 24
    // or
    // 1025 to 1025 (getHRTFSetProperties)

    mpegInterp(fIn_internal, valIn_internal, fOut, valOut, lengthIn, lengthOut);
}

void gauss_f(float* vect_in, float* vect_out, int16_t length, float F, float Noct)
{
    float sigma; // Standard deviation
    float vectOut_internal[NR_FILTER_TEST_FREQ];
    sigma = F / Noct;
    sigma /= M_PI;
    // Matlab:    g = exp(-(((f_x-F).^2)./(2.*(sigma^2)))); % Gaussian
    float num, den, sum = 0;
    den = 2 * (pow(sigma, 2));
    for (int16_t i = 0; i != length; i++) {
        num = -pow((vect_in[i] - F), 2);
        sum += (vectOut_internal[i] = exp(num / den));
    }
    // Matlab:    g = g./sum(g); % normalise magnitude
    for (int16_t i = 0; i != length; i++) {
        vectOut_internal[i] /= sum;
        if (vectOut_internal[i] < 1e-15)
            vectOut_internal[i] = 0; // get rid of denormalized numbers (#DEN)
    }

    memcpy(vect_out, vectOut_internal, NR_FILTER_TEST_FREQ * sizeof(float));

} // g_gauss

void smoothResponse(float* X, float* f, float* smoothed, int16_t length, float Noct)
{
    bool positiveX = true;
    float smoothedInternal[NR_FILTER_TEST_FREQ];
    for (int16_t i = 0; i != length; i++) {
        smoothedInternal[i] = X[i]; // Initial spectrum
        if (X[i] < 0)
            positiveX = false;
    }

    if (Noct > 0) {
        //don't bother if no smoothing
        float g[NR_FILTER_TEST_FREQ];
        for (int16_t i = 0; i != length; i++)
            if (f[i] > 0) {
                gauss_f(f, g, NR_FILTER_TEST_FREQ, f[i], Noct);
                smoothedInternal[i] = 0;
                for (int16_t j = 0; j != length; j++)
                    smoothedInternal[i] += g[j] * X[j];
            }
        // Remove undershoot when X is positive
        if (positiveX) {
            for (int16_t i = 0; i != length; i++)
                if (smoothedInternal[i] < 0)
                    smoothedInternal[i] = 0;
        }
    }
    memcpy(smoothed, smoothedInternal, NR_FILTER_TEST_FREQ * sizeof(float));
} // smoothResponse

TEST_CASE("Reverb_downmix_coefficients_test")
{
    // Disable connected reverb for this test
    MPEGI::Config::set("Reverb:ConnectedReverbGain", EAR::Variant(0.f));

    unsigned int acousticEnvironmentId = 1;
    EAR::Position userPos({ 0, 1.8, 13 }, { 0, 0, 0 });
    stage = std::make_shared<ReverbStage_test>(userPos, acousticEnvironmentId);

    InitializeRenderer();

    CHECK(stage->lastDspData != nullptr);
    auto& dspList = *(stage->lastDspData);

    // Reverberator  Source          DmxGain
    // ======================================
    // PatientRoom   FootstepsLeft   0.016588
    // PatientRoom   FootstepsRight  0.017000
    // PatientRoom   Doctor          0.025042
    // PatientRoom   PM1             0.007180
    // PatientRoom   PM2             0.028961
    // Corridor      PM1             0.000384
    // Corridor      PM2             0.001614
    // Corridor      FootstepsLeft   0.019151
    // Corridor      FootstepsRight  0.019313
    // Corridor      Doctor          0.027326
    // Corridor      Door1           0.004544
    // Corridor      Door2           0.004544
    // NurseRoom     Door3           0.086508
    // NurseRoom     Door4           0.102639
    const double referenceValue[23] = { 0.016588, 0.017000, 0.025042, 0.011896, 0.003432, 0.007178,
        0.028936, 0.003079, 0.003008, 0.004539, 0.000493, 0.003763, 0.083661, 0.099245, 0.000989,
        0.003540, 0.000384, 0.001614, 0.019148, 0.019309, 0.027320, 0.004544, 0.004544 };
    const std::vector<std::string> reverbOrder = { "PatientRoom", "NurseRoom", "Corridor" };
    int refIdx = 0;
    for (auto rItr : reverbOrder) {
        const auto& reverbInputSignalList = dspList.dspItems[rItr];

        for (auto reverbInputSignal = reverbInputSignalList.cbegin();
             reverbInputSignal != reverbInputSignalList.cend(); ++reverbInputSignal, ++refIdx) {
            CAPTURE(refIdx);
            CHECK(std::abs(20 * log10(reverbInputSignal->gainInBoundedAE / referenceValue[refIdx]))
                < 2);

            auto dirFlt = reverbInputSignal->directivityFilter;

            if (dirFlt != nullptr) {
                CHECK((refIdx == 2 || refIdx == 9 || refIdx == 20));

                auto fltPars = dirFlt->filter->getParameters();

                double fltCoeff[6];
                fltCoeff[0] = 1.0f;
                fltCoeff[1] = 1.0f;
                fltCoeff[2] = fltPars.b[0][0];
                fltCoeff[3] = fltPars.a[0][0];
                fltCoeff[4] = fltPars.b[1][0];
                fltCoeff[5] = fltPars.a[1][0];
                auto v1 = ConvertFromFilterCoefToMagnitudeSpectrum(fltCoeff, 2);
                for (auto sosIdx = 1; sosIdx < fltPars.stageF.size(); sosIdx++) {
                    fltCoeff[0] = 1.0f;
                    fltCoeff[1] = 1.0f;
                    fltCoeff[2] = fltPars.b[0][sosIdx];
                    fltCoeff[3] = fltPars.a[0][sosIdx];
                    fltCoeff[4] = fltPars.b[1][sosIdx];
                    fltCoeff[5] = fltPars.a[1][sosIdx];
                    auto v1x = ConvertFromFilterCoefToMagnitudeSpectrum(fltCoeff, 2);

                    for (auto f = 0; f < v1.size(); f++) {
                        v1[f] = v1[f] * v1x[f] * fltPars.stageF[sosIdx];
                    }
                }
                for (auto f = 0; f < v1.size(); f++) {
                    v1[f] = v1[f] * fltPars.F;
                }

                float refValML[] = { // From Matlab reference (i.e. full SOFA file representation)
                    1.1009f, 1.0212f, 0.7645f, 0.7087f, 0.9925f, 1.0197f, 1.0822f, 1.4298f, 1.1122f,
                    0.7416f, 0.4699f, 0.6054f, 0.6481f, 0.5647f, 0.4544f, 0.5810f, 0.4999f, 0.5476f
                };
                float refValP13[] = {
                    // Target for filter, derived from bitstream representation of the directivity pattern (i.e. sparser grid)
                    0.702597678, 0.653796256, 0.500820100, 0.474925220, 0.664544642, 0.715471685,
                    0.726162970, 0.968778849, 0.745780528, 0.494050235, 0.331647962, 0.430911839,
                    0.460442126, 0.447374314, 0.334054857, 0.418898970, 0.384585470, 0.413859695
                };
                float refFreq[] = { 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000,
                    2500, 3150, 4000, 5000, 6300, 8000 };
                float refValTestGrid[240] = { 0 };
                float refFreqTestGrid[240] = { 0 };

                for (int i = 0, f = 20; f < 24000; f += 100, i++) {
                    refFreqTestGrid[i] = f;
                }

                for (int j = 0; j < 18; ++j) {
                    refValP13[j] = std::sqrt(refValP13[j]);
                }

                resampleToGrid(refFreq, refValP13, refFreqTestGrid, 48000, refValTestGrid, 18, 240);

                float refValTestGrid_sm[NR_FILTER_TEST_FREQ];

                smoothResponse(
                    refValTestGrid, refFreqTestGrid, refValTestGrid_sm, NR_FILTER_TEST_FREQ, 1);

                // Check that raw filter target is close enough to Matlab reference
                for (int j = 0; j < 18; ++j) {
                    auto dBDifferenceValue = std::abs(log10(refValML[j] / refValP13[j])) * 20.0;
                    CAPTURE(j);
                    CHECK(dBDifferenceValue < 4);
                }

                // Check that filter response is close enough to target
                for (int j = 0; j < 239; ++j) {
                    auto dBDifferenceValue = std::abs(log10(v1[j] / refValTestGrid_sm[j])) * 20.0;
                    CAPTURE(j);
                    CHECK(dBDifferenceValue < 3);
                }
            }
        }
    }

    //Ensure all reference values were tested.
    CHECK(refIdx == 23);

    CleanUp();
}

} // namespace ReverbDownmixCoefficient_test
