/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/EIFPayloadHandler.h>
#include <MPEGI/Renderer.h>
#include <MPEGI/Stream/MaxStreamManager.h>

#include "../../RoomAssignmentStage/RoomAssignment.h"
#include "../ReverbStage.h"

#include "ReverbRoomsData.h"
#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Helpers/Config.h>

#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Renderer/RenderStageRegistry.h>

#include <MPEGI/Bitstream/PortalPayloadHandler.h>
#include <MPEGI/Bitstream/ReverbPayloadHandler.h>

#include <fstream>

#define cathedral_bitstream "reverb_replaced/EIF_Cathedral_LBR_P.bit"
#define basePath "reverb_replaced"

namespace MPEGI {
class DirectivityPayloadHandler : public Bitstream::PayloadHandler
{
public:
    //! Constructor
    DirectivityPayloadHandler() = default;

    //! Destructor
    ~DirectivityPayloadHandler() = default;

    //! Inherited from Bitstream::PayloadHandler
    std::shared_ptr<const Bitstream::Payload> unpack(
        Bitstream::RawPayload const& payload, const BitstreamConfig& config) override;
};
}; // namespace MPEGI

namespace ReverbExecuting_test {

std::unique_ptr<MPEGI::Renderer> renderer;
std::unique_ptr<MPEGI::Scene> scene;
std::shared_ptr<MPEGI::Spatializer> spatializer;

void CleanUp()
{
    renderer = nullptr;
    scene = nullptr;
    spatializer = nullptr;
}

void registerPayloadHandlers()
{
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::DirectivityPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_DIRECTIVITY };
    });
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::ReverbPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_REVERB };
    });
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::PortalPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_PORTAL };
    });
}

struct ReverbStage_test : public MPEGI::ReverbStage
{
    ReverbStage_test() { AcEnvProp.resize(10); }

    void initialize(const MPEGI::Bitstream& bitstream, MPEGI::RenderItemStore* itemStore,
        const MPEGI::StreamManager* streamManager, const MPEGI::SceneState* state)
    {
        MPEGI::ReverbStage::initializeStage(bitstream, itemStore, streamManager, state);
    }
};

struct RoomAssignment_test : public MPEGI::RoomAssignmentStage
{
    RoomAssignment_test() {}

    void initialize(const MPEGI::Bitstream& bitstream, MPEGI::RenderItemStore* itemStore,
        const MPEGI::StreamManager* streamManager, const MPEGI::SceneState* state)
    {
        MPEGI::RoomAssignmentStage::initializeStage(bitstream, itemStore, streamManager, state);
    }
};

struct Renderer_test : public MPEGI::Renderer
{
    static CtorArgs createStages(const MPEGI::Bitstream& bitstream, const MPEGI::SceneState* state,
        std::shared_ptr<MPEGI::StreamManager> streamManager)
    {
        auto itemStore = std::make_unique<MPEGI::RenderItemStore>();
        auto stages = std::vector<std::shared_ptr<MPEGI::RenderStage>>{};
        auto roomAssignment = std::make_shared<RoomAssignment_test>();
        roomAssignment->initialize(bitstream, itemStore.get(), streamManager.get(), state);
        stages.push_back(roomAssignment);
        auto stage = std::make_shared<ReverbStage_test>();
        stage->initialize(bitstream, itemStore.get(), streamManager.get(), state);
        stages.push_back(stage);
        return { state, streamManager, std::move(itemStore), std::move(stages) };
    }

    Renderer_test(const MPEGI::Bitstream& bitstream, const MPEGI::SceneState* state,
        std::shared_ptr<MPEGI::StreamManager> streamManager)
        : MPEGI::Renderer{ createStages(bitstream, state, streamManager) }
    {}
};

void InitializeRenderer()
{
    DesignIIRRNG::ResetRng();

    registerPayloadHandlers();

    std::string str(basePath);

    MPEGI::Bitstream bitstream(cathedral_bitstream);
    bitstream.registerHandler(
        std::make_shared<MPEGI::EIFPayloadHandler>(basePath, false), MPEGI::Bitstream::PLD_SCENE);

    auto eifPayload = std::dynamic_pointer_cast<const MPEGI::EIFPayload>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE));

    std::unique_ptr<LSDF::LSDFParser> lsdfObject = std::unique_ptr<LSDF::LSDFParser>(nullptr);
    LSDF::LSDFParser* lsdfPtr = nullptr;
    if (eifPayload->getSceneInfo().type == MPEGI::EIFPayload::SceneInfo::SceneType::AR) {
        lsdfPtr = lsdfObject.get();
    }

    const bool enableOSCListener = true;
    const bool offlineMode = true;
    std::shared_ptr<MPEGI::ExternalClock> clock = std::make_shared<MPEGI::ExternalClock>();
    scene =
        std::make_unique<MPEGI::Scene>(bitstream, clock, offlineMode, enableOSCListener, lsdfPtr);

    const float fs = 48000.0f;
    long B = 256;
    const int maxExternalInlets = 192; //Maximum number of connected inlets in the Max External
    std::shared_ptr<MPEGI::MaxStreamManager> streamManager =
        std::make_shared<MPEGI::MaxStreamManager>(bitstream, fs, B, maxExternalInlets);

    renderer = std::make_unique<Renderer_test>(bitstream, scene->getState(), streamManager);

    MPEGI::Bitstream::clearHandlerList();
}

TEST_CASE("Reverb_executing_test")
{
    InitializeRenderer();

    // run controller by hand once so we definitely have a first valid RenderList when the audio process starts
    renderer->triggerControlStep();

    CleanUp();
}
} // namespace ReverbExecuting_test
