/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "EAR/BaseTypes.h"
#include "ReverbUnitTest.h"
#include <MPEGI/Bitstream/EIFPayload.h>

namespace MPEGI {

std::vector<std::pair<float, float>> ReverbStageTester::getExternalSourceContribution(
    EAR::Vector3 portalPosition, EAR::Vector3 portalNormal, float portalHeight, float portalWidth,
    float portalDepth, const std::vector<EAR::Vector3>& sourcePositions)
{
    // Inialize reverbStage
    // Create scene with 2 rooms. For this test, the position & size of the rooms don't matter. The only important thing is that they have an accoustic environment.
    const auto room1AcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    const auto room2AcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    const float roomSize = 5.f;
    const EAR::Position room1Pos(EAR::Vector3(-roomSize / 2.f, 0.f, 0.f), EAR::Euler());
    const EAR::Position room2Pos(EAR::Vector3(roomSize / 2.f, 0.f, 0.f), EAR::Euler());
    const bool hasDefaultAE = false;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    auto reverbPayload = std::make_shared<DummyReverbPayload>();
    reverbPayload->addReverbParameters(std::stoi(room1AcousticEnvironmentId));
    reverbPayload->addReverbParameters(std::stoi(room2AcousticEnvironmentId));

    DummySceneState sceneState(hasDefaultAE);

    // Create 2 rooms / accoustic environments
    const EAR::Vector3 roomCube(roomSize, roomSize, roomSize);
    auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 1", room1Pos);
    auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube, "Room 2", room2Pos);
    sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
    sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

    HomogenMatrix rotMatrix = EAR::rotateVectorTo({ 0.f, 0.f, -1.f }, portalNormal);

    // Create portal extent
    EIFPayload::BoxDesc portalBox;
    portalBox.position = EAR::Position(portalPosition, rotMatrix.getRotation());
    portalBox.size = { portalWidth, portalHeight, portalDepth };
    portalBox.id = "0";
    portalBox.bsId = 0;

    sceneState.createBox2(portalBox);

    CHECK(sceneState.getAllGeometry().size() == 1);

    // Create portal between rooms
    auto portalPayload = std::make_shared<DummyPortalPayload>();
    {
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalNormal;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));
    }

    DummyBitstream bitstream(reverbPayload, portalPayload);

    // Initialize reverb stage
    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    RenderList list(0);

    std::vector<std::pair<float, float>> portalContributions;
    portalContributions.reserve(sourcePositions.size());
    for (const auto& position : sourcePositions) {
        auto item = store.createItem("Source");
        item->position = EAR::Position(position, EAR::Euler());
        item->acousticEnvironmentId = room2AcousticEnvironmentId;
        list.addItemCopy(item);

        portalContributions.push_back(reverbStage.getExternalSourceContribution(
            *list.items.back(), room1AcousticEnvironmentId));
    }

    return portalContributions;
}

std::vector<std::pair<float, float>> ReverbStageTester::getExternalSourceContribution_NR(
    EAR::Vector3 portalPosition, EAR::Vector3 portalNormal, float portalHeight, float portalWidth,
    float portalDepth, const std::vector<EAR::Vector3>& sourcePositions)
{
    // Inialize reverbStage
    // Create scene with 2 rooms. For this test, the position & size of the rooms don't matter. The only important thing is that they have an accoustic environment.
    const auto room1AcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    const auto room2AcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    const float roomSize = 5.f;
    const EAR::Position room1Pos(EAR::Vector3(-4.95f, 1.25f, 0.f), EAR::Euler());
    const EAR::Position room2Pos(EAR::Vector3(0.f, 1.25f, 0.f), EAR::Euler());
    const bool hasDefaultAE = false;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    auto reverbPayload = std::make_shared<DummyReverbPayload>();
    reverbPayload->addReverbParameters(std::stoi(room1AcousticEnvironmentId));
    reverbPayload->addReverbParameters(std::stoi(room2AcousticEnvironmentId));

    DummySceneState sceneState(hasDefaultAE);

    // Create 2 rooms / accoustic environments
    const EAR::Vector3 roomCube1(2.7f, 2.5f, 5.2f);
    const EAR::Vector3 roomCube2(7.2f, 2.5f, 5.2f);
    auto box1 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube1, "Room 1", room1Pos);
    auto box2 = EAR::SceneObjectFactory<EAR::Box>::getNewObject(roomCube2, "Room 2", room2Pos);
    sceneState.addAcousticEnvironment(room1AcousticEnvironmentId, box1.get());
    sceneState.addAcousticEnvironment(room2AcousticEnvironmentId, box2.get());

    HomogenMatrix rotMatrix = EAR::rotateVectorTo({ 0.f, 0.f, -1.f }, portalNormal);

    // Create portal extent
    EIFPayload::BoxDesc portalBox;
    portalBox.position = EAR::Position(portalPosition, rotMatrix.getRotation());
    portalBox.size = { portalWidth, portalHeight, portalDepth };
    portalBox.id = "0";
    portalBox.bsId = 0;

    sceneState.createBox2(portalBox);

    CHECK(sceneState.getAllGeometry().size() == 1);
    const EAR::Geometry* portalBoxPtr = sceneState.getAllGeometry()[0];

    // Create portal between rooms
    auto portalPayload = std::make_shared<DummyPortalPayload>();
    {
        std::unique_ptr<Portal> portal = std::make_unique<Portal>();
        portal->portalEnvironments = { room1AcousticEnvironmentId, room2AcousticEnvironmentId };
        portal->extentId = portalBox.id;
        portal->normalVector = portalNormal;

        sceneState.createPortal2(portal.get());
        portalPayload->addPortal(std::move(portal));
    }

    DummyBitstream bitstream(reverbPayload, portalPayload);

    // Initialize reverb stage
    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    RenderList list(0);

    std::vector<std::pair<float, float>> portalContributions;
    portalContributions.reserve(sourcePositions.size());
    for (const auto& position : sourcePositions) {
        auto item = store.createItem("Source");
        item->position = EAR::Position(position, EAR::Euler());
        item->acousticEnvironmentId = room1AcousticEnvironmentId;
        list.addItemCopy(item);

        portalContributions.push_back(reverbStage.getExternalSourceContribution(
            *list.items.back(), room2AcousticEnvironmentId));
    }

    return portalContributions;
}

void checkExternalContributions(EAR::Vector3 portalPos, EAR::Vector3 portalNormal,
    float portalHeight, float portalWidth, float portalDepth,
    const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults, bool useNR)
{
    const float epsilon = 1e-3;
    std::vector<EAR::Vector3> sourcePositions;
    sourcePositions.reserve(expectedResults.size());
    for (const auto& expectedResult : expectedResults) {
        sourcePositions.push_back(expectedResult.first);
    }

    // Compute portal contributions
    std::vector<std::pair<float, float>> portalContributions;
    if (useNR) {
        portalContributions = ReverbStageTester().getExternalSourceContribution_NR(
            portalPos, portalNormal, portalHeight, portalWidth, portalDepth, sourcePositions);
    } else {
        portalContributions = ReverbStageTester().getExternalSourceContribution(
            portalPos, portalNormal, portalHeight, portalWidth, portalDepth, sourcePositions);
    }

    // Compare to expected values
    for (int i = 0; i < expectedResults.size(); ++i) {
        CAPTURE(sourcePositions[i]);
        CHECK(portalContributions[i].first
            == doctest::Approx(expectedResults[i].second.first).epsilon(epsilon));
        CHECK(portalContributions[i].second
            == doctest::Approx(expectedResults[i].second.second).epsilon(epsilon));
    }
}

TEST_CASE("Reverb External Sources tests")
{
    SUBCASE("External Source Contribution centered")
    {
        // Tests a source approaching a portal from the center. The end point is the center of the portal.
        bool useNR = false;

        // Setup fixed geometry
        const EAR::Vector3 portalPos(0.f, 0.f, 0.f);
        const EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);

        const float portalHeight = 2.f;
        const float portalWidth = 1.f;
        const float portalDepth = 0.1f;

        const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { 20.0000f, 0.0000f, 0.0000f }, { 0.000397308f, 0.000397308f } },
            { { 10.0000f, 0.0000f, 0.0000f }, { 0.001582333f, 0.001582333f } },
            { { 5.0000f, 0.0000f, 0.0000f }, { 0.006221886f, 0.006221886f } },
            { { 2.0000f, 0.0000f, 0.0000f }, { 0.034873327f, 0.034873327f } },
            { { 1.0000f, 0.0000f, 0.0000f }, { 0.104357377f, 0.104357377f } },
            { { 0.5000f, 0.0000f, 0.0000f }, { 0.223606798f, 0.223606798f } },
            { { 0.1000f, 0.0000f, 0.0000f }, { 0.434997465f, 0.434997465f } },
            { { 0.0500f, 0.0000f, 0.0000f }, { 0.467690235f, 0.467690235f } },
            { { 0.0010f, 0.0000f, 0.0000f }, { 0.499363131f, 0.499363131f } },
            { { 0.0010f, 0.0000f, 0.0000f }, { 0.499363131f, 0.499363131f } },
            { { 0.0000f, 0.0000f, 0.0000f }, { 0.499999999f, 0.499999999f } },
        };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }

    SUBCASE("External Source Contribution off-center on edge of portal")
    {
        // Tests a source going away from a portal parallel to the portal normal. The start point is inside the portal, but off center.
        bool useNR = false;

        // Setup fixed geometry
        const EAR::Vector3 portalPos(0.f, 0.0f, 0.0f);
        const EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);

        const float portalHeight = 2.f;
        const float portalWidth = 2.0f;
        const float portalDepth = 0.1f;

        const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { 0.0000f, 0.0000f, 1.0000f }, { 0.250000000f, 0.250000000f } },
            { { 0.2000f, 0.0000f, 1.0000f }, { 0.229590454f, 0.229590454f } },
            { { 0.4000f, 0.0000f, 1.0000f }, { 0.202949704f, 0.202949704f } },
            { { 0.6000f, 0.0000f, 1.0000f }, { 0.174596888f, 0.174596888f } },
            { { 0.8000f, 0.0000f, 1.0000f }, { 0.147928201f, 0.147928201f } },
            { { 1.0000f, 0.0000f, 1.0000f }, { 0.124598007f, 0.124598007f } },
            { { 1.2000f, 0.0000f, 1.0000f }, { 0.104983561f, 0.104983561f } },
            { { 1.4000f, 0.0000f, 1.0000f }, { 0.088813163f, 0.088813163f } },
            { { 1.6000f, 0.0000f, 1.0000f }, { 0.075584020f, 0.075584020f } },
            { { 1.8000f, 0.0000f, 1.0000f }, { 0.064769641f, 0.064769641f } },
            { { 2.0000f, 0.0000f, 1.0000f }, { 0.055901699f, 0.055901699f } },
            { { 2.2000f, 0.0000f, 1.0000f }, { 0.048591603f, 0.048591603f } },
            { { 2.4000f, 0.0000f, 1.0000f }, { 0.042527320f, 0.042527320f } },
            { { 2.6000f, 0.0000f, 1.0000f }, { 0.037462052f, 0.037462052f } },
            { { 2.8000f, 0.0000f, 1.0000f }, { 0.033201707f, 0.033201707f } },
            { { 3.0000f, 0.0000f, 1.0000f }, { 0.029593708f, 0.029593708f } },
        };
        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }

    SUBCASE("External Source Contribution off-center inside portal")
    {
        // Tests a source going away from a portal parallel to the portal normal. The start point is inside the portal, but off center.
        bool useNR = false;

        // Setup fixed geometry
        const EAR::Vector3 portalPos(0.f, 0.0f, 0.0f);
        const EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);

        const float portalHeight = 2.0f;
        const float portalWidth = 2.4f;
        const float portalDepth = 0.1f;

        const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { 0.0000f, 0.0000f, 1.0000f }, { 0.499999999f, 0.499999999f } },
            { { 0.2000f, 0.0000f, 1.0000f }, { 0.353568985f, 0.353568985f } },
            { { 0.4000f, 0.0000f, 1.0000f }, { 0.274055898f, 0.274055898f } },
            { { 0.6000f, 0.0000f, 1.0000f }, { 0.221947348f, 0.221947348f } },
            { { 0.8000f, 0.0000f, 1.0000f }, { 0.182317980f, 0.182317980f } },
            { { 1.0000f, 0.0000f, 1.0000f }, { 0.150979039f, 0.150979039f } },
            { { 1.2000f, 0.0000f, 1.0000f }, { 0.125995159f, 0.125995159f } },
            { { 1.4000f, 0.0000f, 1.0000f }, { 0.106009629f, 0.106009629f } },
            { { 1.6000f, 0.0000f, 1.0000f }, { 0.089949128f, 0.089949128f } },
            { { 1.8000f, 0.0000f, 1.0000f }, { 0.076961973f, 0.076961973f } },
            { { 2.0000f, 0.0000f, 1.0000f }, { 0.066382528f, 0.066382528f } },
            { { 2.2000f, 0.0000f, 1.0000f }, { 0.057696118f, 0.057696118f } },
            { { 2.4000f, 0.0000f, 1.0000f }, { 0.050506505f, 0.050506505f } },
            { { 2.6000f, 0.0000f, 1.0000f }, { 0.044508483f, 0.044508483f } },
            { { 2.8000f, 0.0000f, 1.0000f }, { 0.039466127f, 0.039466127f } },
            { { 3.0000f, 0.0000f, 1.0000f }, { 0.035196074f, 0.035196074f } },
        };
        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }

    SUBCASE("External Source Contribution off-center outside portal")
    {
        // Tests a source going away from a portal from off-center, parallel to the portal normal but outside of the portal width/height.
        bool useNR = false;

        // Setup fixed geometry
        const EAR::Vector3 portalPos(0.f, 0.0f, 0.0f);
        const EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);

        const float portalHeight = 2.f;
        const float portalWidth = 2.4f;
        const float portalDepth = 0.1f;

        const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { 0.0000f, 0.0000f, 3.0000f }, { 0.000000000f, 0.000000000f } },
            { { 0.2000f, 0.0000f, 3.0000f }, { 0.009843616f, 0.009843616f } },
            { { 0.4000f, 0.0000f, 3.0000f }, { 0.018281901f, 0.018281901f } },
            { { 0.6000f, 0.0000f, 3.0000f }, { 0.024545369f, 0.024545369f } },
            { { 0.8000f, 0.0000f, 3.0000f }, { 0.028584551f, 0.028584551f } },
            { { 1.0000f, 0.0000f, 3.0000f }, { 0.030763282f, 0.030763282f } },
            { { 1.2000f, 0.0000f, 3.0000f }, { 0.031555172f, 0.031555172f } },
            { { 1.4000f, 0.0000f, 3.0000f }, { 0.031386919f, 0.031386919f } },
            { { 1.6000f, 0.0000f, 3.0000f }, { 0.030591364f, 0.030591364f } },
            { { 1.8000f, 0.0000f, 3.0000f }, { 0.029410279f, 0.029410279f } },
            { { 2.0000f, 0.0000f, 3.0000f }, { 0.028012270f, 0.028012270f } },
            { { 2.2000f, 0.0000f, 3.0000f }, { 0.026511685f, 0.026511685f } },
            { { 2.4000f, 0.0000f, 3.0000f }, { 0.024984211f, 0.024984211f } },
            { { 2.6000f, 0.0000f, 3.0000f }, { 0.023478594f, 0.023478594f } },
            { { 2.8000f, 0.0000f, 3.0000f }, { 0.022025088f, 0.022025088f } },
            { { 3.0000f, 0.0000f, 3.0000f }, { 0.020641386f, 0.020641386f } },
        };
        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }

    SUBCASE("External Sources Contribution, approaching centre from compound angle")
    {
        // Tests a source going away from a portal at an angle to the portal normal. The start point is the centre of the portal.
        bool useNR = false;

        // Setup fixed geometry
        const EAR::Vector3 portalPos(0.f, 0.0f, 0.0f);
        const EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);

        const float portalHeight = 0.8f;
        const float portalWidth = 1.5f;
        const float portalDepth = 0.1f;
        const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { 0.0000f, 0.0000f, 0.0000f }, { 0.500000000f, 0.500000000f } },
            { { 0.2000f, 0.0000f, 0.1000f }, { 0.371825188f, 0.371825188f } },
            { { 0.4000f, 0.0000f, 0.2000f }, { 0.237939413f, 0.237939413f } },
            { { 0.6000f, 0.0000f, 0.3000f }, { 0.149653516f, 0.149653516f } },
            { { 0.8000f, 0.0000f, 0.4000f }, { 0.097896160f, 0.097896160f } },
            { { 1.0000f, 0.0000f, 0.5000f }, { 0.067444996f, 0.067444996f } },
            { { 1.2000f, 0.0000f, 0.6000f }, { 0.048744294f, 0.048744294f } },
            { { 1.4000f, 0.0000f, 0.7000f }, { 0.036668017f, 0.036668017f } },
            { { 1.6000f, 0.0000f, 0.8000f }, { 0.028498414f, 0.028498414f } },
            { { 1.8000f, 0.0000f, 0.9000f }, { 0.022745644f, 0.022745644f } },
            { { 2.0000f, 0.0000f, 1.0000f }, { 0.018555261f, 0.018555261f } },
            { { 2.2000f, 0.0000f, 1.1000f }, { 0.015414609f, 0.015414609f } },
            { { 2.4000f, 0.0000f, 1.2000f }, { 0.013003140f, 0.013003140f } },
            { { 2.6000f, 0.0000f, 1.3000f }, { 0.011112949f, 0.011112949f } },
            { { 2.8000f, 0.0000f, 1.4000f }, { 0.009604787f, 0.009604787f } },
            { { 3.0000f, 0.0000f, 1.5000f }, { 0.008382724f, 0.008382724f } },
        };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }

    SUBCASE("External Sources Contribution, approaching centre from compound angle")
    {
        // Tests a source going away from a portal at an angle to the portal normal. The start point is to the side of the portal.
        bool useNR = false;

        // Setup fixed geometry
        const EAR::Vector3 portalPos(-1.f, 0.0f, 2.0f);
        const EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);

        const float portalHeight = 0.8f;
        const float portalWidth = 1.5f;
        const float portalDepth = 0.1f;
        const std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { 0.0000f, 0.0000f, 0.0000f }, { 0.019267647f, 0.019267647f } },
            { { 0.2000f, 0.0000f, 0.1000f }, { 0.019198910f, 0.019198910f } },
            { { 0.4000f, 0.0000f, 0.2000f }, { 0.018591335f, 0.018591335f } },
            { { 0.6000f, 0.0000f, 0.3000f }, { 0.017619729f, 0.017619729f } },
            { { 0.8000f, 0.0000f, 0.4000f }, { 0.016433506f, 0.016433506f } },
            { { 1.0000f, 0.0000f, 0.5000f }, { 0.015150198f, 0.015150198f } },
            { { 1.2000f, 0.0000f, 0.6000f }, { 0.013854400f, 0.013854400f } },
            { { 1.4000f, 0.0000f, 0.7000f }, { 0.012601840f, 0.012601840f } },
            { { 1.6000f, 0.0000f, 0.8000f }, { 0.011425660f, 0.011425660f } },
            { { 1.8000f, 0.0000f, 0.9000f }, { 0.010342730f, 0.010342730f } },
            { { 2.0000f, 0.0000f, 1.0000f }, { 0.009358952f, 0.009358952f } },
            { { 2.2000f, 0.0000f, 1.1000f }, { 0.008473291f, 0.008473291f } },
            { { 2.4000f, 0.0000f, 1.2000f }, { 0.007680651f, 0.007680651f } },
            { { 2.6000f, 0.0000f, 1.3000f }, { 0.006973813f, 0.006973813f } },
            { { 2.8000f, 0.0000f, 1.4000f }, { 0.006344699f, 0.006344699f } },
            { { 3.0000f, 0.0000f, 1.5000f }, { 0.005785135f, 0.005785135f } },
        };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }

    SUBCASE("Various portal orientations")
    {
        // Tests a fixed source but with different portal orientations.
        bool useNR = true;

        // Setup fixed geometry
        const float portalHeight = 2.0f;
        const float portalWidth = 1.2f;
        const float portalDepth = 0.2f;

        EAR::Vector3 portalPos(-3.6f, 1.f, 1.4f);
        EAR::Vector3 portalNormal(-1.f, 0.f, 0.f);
        std::vector<std::pair<EAR::Vector3, std::pair<float, float>>> expectedResults = {
            { { -4.193f, 0.883f, -1.976f }, { 0.00851957f, 0.00851957f } }
        };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);

        portalPos = { -3.6f, 1.f, 1.4f };
        portalNormal = { -std::sqrt(0.90f), std::sqrt(0.05f), -std::sqrt(0.05f) };
        expectedResults = { { { -4.193f, 0.883f, -1.976f }, { 0.01136188f, 0.01136188f } } };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);

        // Ceiling portal
        portalPos = { -4.6f, 2.5f, 1.4f };
        portalNormal = { 0.f, -1.f, 0.f };
        expectedResults = { { { -4.193f, 0.883f, -1.976f }, { 0.01306783f, 0.01306783f } } };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);

        // Ceiling portal at slight angle
        portalPos = { -4.6f, 2.5f, 1.4f };
        portalNormal = { 0.f, -std::sqrt(0.999f), std::sqrt(0.001f) };
        expectedResults = { { { -4.193f, 0.883f, -1.976f }, { 0.01295213f, 0.01295213f } } };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);

        // Ceiling portal at larger angle
        portalPos = { -4.6f, 2.5f, 1.4f };
        portalNormal = { 0.f, -std::sqrt(0.9f), -std::sqrt(0.1f) };
        expectedResults = { { { -4.193f, 0.883f, -1.976f }, { 0.01334023f, 0.01334023f } } };

        checkExternalContributions(portalPos, portalNormal, portalHeight, portalWidth, portalDepth,
            expectedResults, useNR);
    }
}
} // namespace MPEGI