/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include "../ReverbStage.h"
#include "ReverbRoomsData.h"
#include <MPEGI/Helpers/Config.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/ReverbPayloadHandler.h>

#include <fstream>
#include <iostream>

#define cathedral_bitstream "reverb_replaced/EIF_Cathedral_LBR_P.bit"

namespace ReverbInput_test {
struct ReverbStage_test : public MPEGI::ReverbStage
{
    const float fs = 48000.0f;

    ReverbStage_test(unsigned int roomId)
        : roomId(roomId)
    {
        AcEnvProp.resize(10);
    }
    void CalculateBaseConfig_test(std::shared_ptr<SetBaseConfigStruct>& baseConfigOut,
        const unsigned int buildingId, unsigned int nrFbLoops,
        std::shared_ptr<const MPEGI::ReverbParameters> room)
    {
        MPEGI::ReverbStage::CalculateBaseConfig(
            baseConfigOut, buildingId, nrFbLoops, room, false, fs);
    }
    void CalculateT60_test(std::shared_ptr<T60Settings>& T60ConfigOut, const unsigned int roomId,
        const std::shared_ptr<lateReverb> lateRevOut, IIRFilterType type)
    {
        MPEGI::ReverbStage::CalculateT60(T60ConfigOut, roomId, lateRevOut, type);
    }
    void CalculateLateReverb_test(std::shared_ptr<lateReverb>& lateRevOut,

        const std::shared_ptr<MPEGI::FreqGrid> FreqGridInputData,
        const std::shared_ptr<const MPEGI::ReverbParameters> room)
    {
        MPEGI::ReverbStage::CalculateLateReverb(lateRevOut, FreqGridInputData, room);
    }
    void CalculateColoration_test(std::shared_ptr<ColorParams>& ColorOut, const unsigned int roomId,
        const std::shared_ptr<lateReverb> lateRevOut,
        const std::shared_ptr<SetBaseConfigStruct> baseConfigOut, IIRFilterType type)
    {
        MPEGI::ReverbStage::CalculateColoration(
            ColorOut, roomId, lateRevOut, baseConfigOut, type, false);
    }

    void FillBuildings(unsigned int roomId, unsigned int buildingIdx)
    {
        using namespace ReverbRoomsHardcodedData;
        int nrRooms = 2;

        SetAcousticEnvironmentProperties(buildingIdx, roomId,
            buildings[roomId].corner2[0] - buildings[roomId].corner1[0],
            buildings[roomId].corner2[1] - buildings[roomId].corner1[1],
            buildings[roomId].corner2[2] - buildings[roomId].corner1[2], nrRooms);
    }

    inline const ReverbDSPCfgStruct& GetRevDSPCfg() const
    {
        return MPEGI::ReverbStage::GetRevDSPCfg(roomId);
    }

private:
    unsigned int roomId;
};

std::shared_ptr<MPEGI::Bitstream> bitstream;
std::shared_ptr<const MPEGI::ReverbPayload> bitstreamReverbPayload;
const float* LAV;
const float* RAV;
const float* IAC;

void registerPayloadHandlers()
{
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::ReverbPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_REVERB };
    });
}

void InitTest()
{
    DesignIIRRNG::ResetRng();

    registerPayloadHandlers();

    bitstream = std::make_shared<MPEGI::Bitstream>(cathedral_bitstream);

    auto reverbPayloads = bitstream->getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
    EAR_CHECK_REL(reverbPayloads.size() <= 1, "Too many reverb payloads in bitstream");
    if (reverbPayloads.size() == 1) {
        bitstreamReverbPayload =
            std::dynamic_pointer_cast<const MPEGI::ReverbPayload>(reverbPayloads[0]);
    } else {
        // Cannot initialize without data
        return;
    }

    MPEGI::Bitstream::clearHandlerList();
}
void WriteToFile(std::ofstream& oFile, std::vector<double> data)
{
    oFile << data.size() << std::endl;
    for (int i = 0; i < data.size(); ++i)
        oFile << data[i] << std::endl;
}

void WriteToFile(std::ofstream& oFile, const double* data, const size_t sz)
{
    oFile << sz << std::endl;
    for (int i = 0; i < sz; ++i)
        oFile << data[i] << std::endl;
}

void WriteToFile(std::ofstream& oFile, const float* data, const size_t sz)
{
    oFile << sz << std::endl;
    for (int i = 0; i < sz; ++i)
        oFile << data[i] << std::endl;
}
std::vector<double> ConvertFromFilterCoefToMagnitudeSpectrum(
    const float filterCoefficients[4][MAX_FILTERORDER + 1], const size_t order)
{
    // w = 2 * PI * frequency / sample_freq;
    double w;
    double fs = 48000.0;
    std::vector<double> H_magn(240);
    const size_t _order = order;
    // b = filterCoefficients[0 - _order]
    // a = filterCoefficients[_order+1 - 2*order]

    for (int i = 0, f = 20; f < 24000; f += 100, i++) {
        w = 2 * M_PI * f / fs;
        double sumAcos = 0.0;
        double sumBcos = 0.0;
        double sumAsin = 0.0;
        double sumBsin = 0.0;
        for (int j = 0; j < _order; ++j) {
            const double b = filterCoefficients[0][j];
            const double a = filterCoefficients[1][j];
            sumBcos += ((b)*cos(j * w));
            sumBsin += ((b)*sin(j * w));
            sumAcos += ((a)*cos(j * w));
            sumAsin += ((a)*sin(j * w));
        }
        sumBcos = sumBcos * sumBcos;
        sumAcos = sumAcos * sumAcos;
        sumBsin = sumBsin * sumBsin;
        sumAsin = sumAsin * sumAsin;

        H_magn[i] = sqrt((sumBcos + sumBsin) / (sumAcos + sumAsin));
    }

    return H_magn;
}
std::vector<double> ConvertFromFilterCoefToMagnitudeSpectrum(
    const float filterCoefficientsB[MAX_FILTERORDER + 1],
    const float filterCoefficientsA[MAX_FILTERORDER + 1], const size_t order)
{
    // w = 2 * PI * frequency / sample_freq;
    double w;
    double fs = 48000.0;
    std::vector<double> H_magn(240);
    const size_t _order = order;
    // b = filterCoefficients[0 - _order]
    // a = filterCoefficients[_order+1 - 2*order]

    for (int i = 0, f = 20; f < 24000; f += 100, i++) {
        w = 2 * M_PI * f / fs;
        double sumAcos = 0.0;
        double sumBcos = 0.0;
        double sumAsin = 0.0;
        double sumBsin = 0.0;
        for (int j = 0; j < _order; ++j) {
            const double b = filterCoefficientsB[j];
            const double a = filterCoefficientsA[j];
            sumBcos += ((b)*cos(j * w));
            sumBsin += ((b)*sin(j * w));
            sumAcos += ((a)*cos(j * w));
            sumAsin += ((a)*sin(j * w));
        }
        sumBcos = sumBcos * sumBcos;
        sumAcos = sumAcos * sumAcos;
        sumBsin = sumBsin * sumBsin;
        sumAsin = sumAsin * sumAsin;

        H_magn[i] = sqrt((sumBcos + sumBsin) / (sumAcos + sumAsin));
    }

    return H_magn;
}
std::vector<double> ConvertFromFilterCoefToMagnitudeSpectrum(
    const double* filterCoefficients, const size_t order)
{
    // w = 2 * PI * frequency / sample_freq;
    double w;
    double fs = 48000.0;
    std::vector<double> H_magn(240);
    const size_t _order = order + 1;
    // b = filterCoefficients[0 - _order]
    // a = filterCoefficients[_order+1 - 2*order]

    for (int i = 0, f = 20; f < 24000; f += 100, i++) {
        w = 2 * M_PI * f / fs;
        double sumAcos = 0.0;
        double sumBcos = 0.0;
        double sumAsin = 0.0;
        double sumBsin = 0.0;
        for (int j = 0; j < _order; ++j) {
            const double b = filterCoefficients[2 * j];
            const double a = filterCoefficients[(2 * j) + 1];
            sumBcos += ((b)*cos(j * w));
            sumBsin += ((b)*sin(j * w));
            sumAcos += ((a)*cos(j * w));
            sumAsin += ((a)*sin(j * w));
        }
        sumBcos = sumBcos * sumBcos;
        sumAcos = sumAcos * sumAcos;
        sumBsin = sumBsin * sumBsin;
        sumAsin = sumAsin * sumAsin;

        H_magn[i] = sqrt((sumBcos + sumBsin) / (sumAcos + sumAsin));
    }

    return H_magn;
}

std::vector<double> ConvertFromFilterCoefToMagnitudeSpectrum(
    const double filterCoefficients[4][MAX_FILTERORDER + 1], const size_t order)
{
    // w = 2 * PI * frequency / sample_freq;
    double w;
    double fs = 48000.0;
    std::vector<double> H_magn(240);
    const size_t _order = order;
    // b = filterCoefficients[0 - _order]
    // a = filterCoefficients[_order+1 - 2*order]

    for (int i = 0, f = 20; f < 24000; f += 100, i++) {
        w = 2 * M_PI * f / fs;
        double sumAcos = 0.0;
        double sumBcos = 0.0;
        double sumAsin = 0.0;
        double sumBsin = 0.0;
        for (int j = 0; j < _order; ++j) {
            const double b = filterCoefficients[0][j];
            const double a = filterCoefficients[1][j];
            sumBcos += ((b)*cos(j * w));
            sumBsin += ((b)*sin(j * w));
            sumAcos += ((a)*cos(j * w));
            sumAsin += ((a)*sin(j * w));
        }
        sumBcos = sumBcos * sumBcos;
        sumAcos = sumAcos * sumAcos;
        sumBsin = sumBsin * sumBsin;
        sumAsin = sumAsin * sumAsin;

        H_magn[i] = sqrt((sumBcos + sumBsin) / (sumAcos + sumAsin));
    }

    return H_magn;
}

void ValidateBaseConfig(
    const std::shared_ptr<SetBaseConfigStruct> baseConfigOut, const SetBaseConfigStruct& other)
{
    CHECK(baseConfigOut->nrFbLoops == other.nrFbLoops);
    CHECK(std::abs(baseConfigOut->delayMinMs - other.delayMinMs) <= 0.5);
    CHECK(baseConfigOut->fs == other.fs);

    CHECK(baseConfigOut->metadataPredelay == other.metadataPredelay);
    CHECK(baseConfigOut->nrFbLoops == other.nrFbLoops);

    CHECK(std::abs(baseConfigOut->delayRatio - other.delayRatio) < 0.001);

    float minLoopDelay = 9999999;
    for (int i = 0; i < baseConfigOut->nrFbLoops; ++i) {
        CHECK((baseConfigOut->loopDelays[i]) == other.loopDelays[i]);
        minLoopDelay = std::min(baseConfigOut->loopDelays[i], minLoopDelay);
    }

    if (baseConfigOut->renderPredelay < minLoopDelay) {
        CHECK(minLoopDelay == other.renderPredelay);
    } else {
        CHECK(baseConfigOut->renderPredelay == other.renderPredelay);
    }
}

void ValidateT60(const ReverbDSPCfgStruct& RevDSPCfg, const T60Settings& other)
{
    //std::ofstream f_v1("v1filters.bin");
    //std::ofstream f_v2("v2filters.bin");

    //std::ofstream f_input("inputfilters.bin");
    //std::ofstream f_mInput("matlabInputfilters.bin");

    for (int loopIdx = 0; loopIdx < RevDSPCfg.Reverberator.nrLoops; ++loopIdx) {
        double fltCoeff[6];
        fltCoeff[0] = 1.0f;
        fltCoeff[1] = 1.0f;
        fltCoeff[2] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].b[0][0];
        fltCoeff[3] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].a[0][0];
        fltCoeff[4] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].b[1][0];
        fltCoeff[5] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].a[1][0];
        auto v1 = ConvertFromFilterCoefToMagnitudeSpectrum(fltCoeff, 2);
        for (auto sosIdx = 1;
             sosIdx < RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].stageF.size(); sosIdx++) {
            fltCoeff[0] = 1.0f;
            fltCoeff[1] = 1.0f;
            fltCoeff[2] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].b[0][sosIdx];
            fltCoeff[3] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].a[0][sosIdx];
            fltCoeff[4] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].b[1][sosIdx];
            fltCoeff[5] = RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].a[1][sosIdx];
            auto v1x = ConvertFromFilterCoefToMagnitudeSpectrum(fltCoeff, 2);

            for (auto f = 0; f < v1.size(); f++) {
                v1[f] = v1[f] * v1x[f]
                    * RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].stageF[sosIdx];
            }
        }
        for (auto f = 0; f < v1.size(); f++) {
            v1[f] = v1[f] * RevDSPCfg.Reverberator.filters.SOSs.t60Sos[loopIdx].F;
        }
        auto v2 = ConvertFromFilterCoefToMagnitudeSpectrum(
            other.filters.t60FltCoeff[loopIdx], other.t60FltOrder);
        //WriteToFile(f_v1, v1);
        //WriteToFile(f_v2, v2);
        //WriteToFile(f_input, fltCoeff[i], sizeof(fltCoeff[i])/sizeof(fltCoeff[i][0]));
        //WriteToFile(f_mInput, other.t60FltCoeff[i], sizeof(other.t60FltCoeff[i])/sizeof(other.t60FltCoeff[i][0]));
        for (int j = 0; j < 240; ++j) {
            CAPTURE(j);
            CHECK(std::abs(log10(v1[j] / v2[j])) * 20.0 < 1);
        }
    }
}

void ValidateColoration(const ReverbDSPCfgStruct& RevDSPCfg, const ColorParams& other)
{
    unsigned int nrFlt = 1;
    for (int fltIdx = 0; fltIdx < nrFlt; ++fltIdx) {
        double fltCoeff[6];
        fltCoeff[0] = 1.0f;
        fltCoeff[1] = 1.0f;
        fltCoeff[2] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx]
                          .b[0][0]; // ColorOut->filters.sosSequence[fltIdx].b[0][0];
        fltCoeff[3] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].a[0][0];
        fltCoeff[4] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].b[1][0];
        fltCoeff[5] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].a[1][0];
        auto v1 = ConvertFromFilterCoefToMagnitudeSpectrum(fltCoeff, 2);
        for (auto sosIdx = 1;
             sosIdx < RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].stageF.size();
             sosIdx++) {
            fltCoeff[0] = 1.0f;
            fltCoeff[1] = 1.0f;
            fltCoeff[2] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].b[0][sosIdx];
            fltCoeff[3] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].a[0][sosIdx];
            fltCoeff[4] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].b[1][sosIdx];
            fltCoeff[5] = RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].a[1][sosIdx];
            auto v1x = ConvertFromFilterCoefToMagnitudeSpectrum(fltCoeff, 2);

            for (auto f = 0; f < v1.size(); f++) {
                v1[f] = v1[f] * v1x[f]
                    * RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].stageF[sosIdx];
            }
        }
        for (auto f = 0; f < v1.size(); f++) {
            v1[f] = v1[f] * RevDSPCfg.Reverberator.filters.SOSs.colorSos[fltIdx].F;
        }
        auto v2 = ConvertFromFilterCoefToMagnitudeSpectrum(
            (other.filters.colorFltCoeff + (fltIdx * 2)), 13);
        //WriteToFile(f_v1, v1);
        //WriteToFile(f_v2, v2);
        //WriteToFile(f_input, fltCoeff[i], sizeof(fltCoeff[i])/sizeof(fltCoeff[i][0]));
        //WriteToFile(f_mInput, other.t60FltCoeff[i], sizeof(other.t60FltCoeff[i])/sizeof(other.t60FltCoeff[i][0]));
        for (int j = 0; j < 238; ++j) {
            auto dBDifferenceValue = std::abs(log10(v1[j] / v2[j])) * 20.0;
            CAPTURE(j);
            CHECK(dBDifferenceValue < 6.1);
        }
    }
}

void ValidateLateReverb(const std::shared_ptr<lateReverb> lateReverbOut, const lateReverb& other)
{
    size_t sz = sizeof(lateReverbOut->filterDesignFc) / sizeof(lateReverbOut->filterDesignFc[0]);
    for (int i = 0; i < sz; ++i) {
        CHECK(std::abs(lateReverbOut->acousticDSR[i] - other.acousticDSR[i]) < 0.001);
        CHECK(std::abs(lateReverbOut->filterDesignFc[i] - other.filterDesignFc[i]) < 0.001);
        CHECK(std::abs(lateReverbOut->acousticRT60[i] - other.acousticRT60[i]) < 0.001);
        // CHECK(std::abs(lateReverbOut->hrtfAvgLeftPwrResponse[i] - other.hrtfAvgLeftPwrResponse[i])
        //     < 0.001);
        // CHECK(std::abs(lateReverbOut->hrtfAvgRightPwrResponse[i] - other.hrtfAvgRightPwrResponse[i])
        //     < 0.001);
        // CHECK(std::abs(lateReverbOut->hrtfIACoherence[i] - other.hrtfIACoherence[i]) < 0.001);
    }
    CHECK(lateReverbOut->metadataPredelay == other.metadataPredelay);
}

TEST_CASE("ReverbStage-input")
{
    InitTest();

    SUBCASE("ReverbInput_room_id_62_test_SOS")
    {

        auto reverbParameters = bitstreamReverbPayload->GetReverbParameters();
        const unsigned int roomId = 0;
        const unsigned int buildingId = 1;
        ReverbStage_test stage(roomId);
        const auto& room = reverbParameters[roomId];
        stage.FillBuildings(room->id, buildingId);
        const std::shared_ptr<MPEGI::FreqGrid> FreqGridData =
            bitstreamReverbPayload->GetFreqGrid()[room->freqGridIdx];
        IIRFilterType IIRFilterType = IIRFilterType::SOSs;
        const unsigned int nrFbLoops = 15;

        // BaseConfig test
        std::shared_ptr<SetBaseConfigStruct> baseConfigOut;
        stage.CalculateBaseConfig_test(baseConfigOut, buildingId, nrFbLoops, room);
        ValidateBaseConfig(
            baseConfigOut, ReverbRoomsHardcodedData::baseConfigCathedralRefenceRoomId62);
        // end BaseConfig test

        // lateReverb test
        std::shared_ptr<lateReverb> lateRevOut;
        stage.CalculateLateReverb_test(lateRevOut, FreqGridData, room);
        ValidateLateReverb(lateRevOut, ReverbRoomsHardcodedData::lateReverbRoomId62);
        // end lateReverb test

        // T60 test
        std::shared_ptr<T60Settings> T60ConfigOut;
        stage.CalculateT60_test(T60ConfigOut, roomId, lateRevOut, IIRFilterType);
        ValidateT60(stage.GetRevDSPCfg(), ReverbRoomsHardcodedData::t60SettingsRoomId62());
        // end T60 test

        // Coloration test
        std::shared_ptr<ColorParams> ColorOut;
        ColorOut = std::make_shared<ColorParams>(IIRFilterType);
        stage.CalculateColoration_test(ColorOut, roomId, lateRevOut, baseConfigOut, IIRFilterType);
        ValidateColoration(stage.GetRevDSPCfg(), ReverbRoomsHardcodedData::colorParamsRoomId62());
        // end Coloration test
    }

    SUBCASE("ReverbInput_room_id_63_test_SOS")
    {
        auto reverbParameters = bitstreamReverbPayload->GetReverbParameters();
        const unsigned int roomId = 1;
        const unsigned int buildingId = 0;
        ReverbStage_test stage(roomId);
        const auto& room = reverbParameters[roomId];
        stage.FillBuildings(room->id, buildingId);
        const std::shared_ptr<MPEGI::FreqGrid> FreqGridData =
            bitstreamReverbPayload->GetFreqGrid()[room->freqGridIdx];
        IIRFilterType IIRFilterType = IIRFilterType::SOSs;
        const unsigned int nrFbLoops = 15;

        // BaseConfig test
        std::shared_ptr<SetBaseConfigStruct> baseConfigOut;
        stage.CalculateBaseConfig_test(baseConfigOut, buildingId, nrFbLoops, room);
        ValidateBaseConfig(
            baseConfigOut, ReverbRoomsHardcodedData::baseConfigCathedralRefenceRoomId63);
        // end BaseConfig test

        // lateReverb test
        std::shared_ptr<lateReverb> lateRevOut;
        stage.CalculateLateReverb_test(lateRevOut, FreqGridData, room);
        ValidateLateReverb(lateRevOut, ReverbRoomsHardcodedData::lateReverbRoomId63);
        // end lateReverb test

        // T60 test
        std::shared_ptr<T60Settings> T60ConfigOut;
        stage.CalculateT60_test(T60ConfigOut, roomId, lateRevOut, IIRFilterType);
        ValidateT60(stage.GetRevDSPCfg(), ReverbRoomsHardcodedData::t60SettingsRoomId63());
        // end T60 test

        // Coloration test
        std::shared_ptr<ColorParams> ColorOut;
        ColorOut = std::make_shared<ColorParams>(IIRFilterType);
        stage.CalculateColoration_test(ColorOut, roomId, lateRevOut, baseConfigOut, IIRFilterType);
        ValidateColoration(stage.GetRevDSPCfg(), ReverbRoomsHardcodedData::colorParamsRoomId63());
        // end Coloration test
    }
}
} // namespace ReverbInput_test