/***************************************************************************
This software module was originally developed by

Koninklijke Philips N.V.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Koninklijke Philips N.V. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <doctest/doctest.h>

#include <MPEGI/Bitstream/Bitstream.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Bitstream/ReverbPayloadHandler.h>

#include <MPEGI/Helpers/ReverbParameters.h>

#define cathedral_bitstream "reverb_replaced/EIF_Cathedral_LBR_P.bit"
#define recreation_bitstream "reverb_replaced/EIF_Recreation_LBR_P.bit"
#define virtualBasketball_bitstream "reverb_replaced/EIF_VirtualBasketball_LBR_P.bit"
#define simple_maze_bitstream "reverb_replaced/EIF_SimpleMaze_LBR_P.bit"
#define singer_bitstream "reverb_replaced/EIF_SingerInTheLab_LBR_P.bit"
#define hospital_bitstream "reverb_replaced/EIF_Hospital_LBR_P.bit"

#define basePath "reverb_replaced"

#include "ReverbParsing_test_data_reference.h"

namespace ReverbParsing_test {

const std::array<float, 24> cathedral_general_data_reference{ 50, 63, 80, 100, 125, 160, 200, 250,
    315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000, 6300, 8000, 10000 };

void registerPayloadHandlers()
{
    MPEGI::Bitstream::introduceHandler([]() -> MPEGI::Bitstream::PayloadHandlerAndType {
        return { std::make_shared<MPEGI::ReverbPayloadHandler>(),
            MPEGI::Bitstream::PayloadType::PLD_REVERB };
    });
}

TEST_CASE("Reverb_parsing_test")
{
    registerPayloadHandlers();

    std::string str(basePath);
    // Cathedral
    {
        MPEGI::Bitstream bitstream(cathedral_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
        auto payload = std::static_pointer_cast<MPEGI::ReverbPayload const>(payloads[0]);
        auto freqGridData = payload->GetFreqGrid();
        auto reverbParameters = payload->GetReverbParameters();

        for (auto itr = reverbParameters.cbegin(); itr != reverbParameters.cend(); ++itr) {
            for (int j = 0; j < freqGridData[itr->second->freqGridIdx]->centerFreq.size(); ++j) {
                CHECK(std::abs(freqGridData[itr->second->freqGridIdx]->centerFreq[j]
                          - cathedral_general_data_reference[j])
                    < 0.001);
            }
        }

        for (auto itr = reverbParameters.cbegin(); itr != reverbParameters.cend(); ++itr) {
            const auto& other = cathedral_reference_data.at(itr->second->id);
            CHECK(itr->second->id == other.id);
            CHECK(itr->second->freqGridIdx == other.freqGridIdx);
            CHECK(std::abs(itr->second->parOffset - other.parOffset) < 0.001f);
            CHECK(itr->second->t60.size() == other.t60.size());
            for (int i = 0; i < itr->second->t60.size(); ++i)
                CHECK(std::abs(itr->second->t60[i] - other.t60[i]) < 0.001f);
            CHECK(itr->second->dsr.size() == other.dsr.size());
            for (int i = 0; i < itr->second->dsr.size(); ++i)
                CHECK(std::abs(itr->second->dsr[i] - other.dsr[i]) < 0.001f);
        }
    }

    // Hospital
    {
        MPEGI::Bitstream bitstream(hospital_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
        auto payload = std::static_pointer_cast<MPEGI::ReverbPayload const>(payloads[0]);
        auto reverbParameters = payload->GetReverbParameters();
        CHECK(reverbParameters.size() > 0);
    }

    // Recreation
    {
        MPEGI::Bitstream bitstream(recreation_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
        auto payload = std::static_pointer_cast<MPEGI::ReverbPayload const>(payloads[0]);
        auto reverbParameters = payload->GetReverbParameters();
        CHECK(reverbParameters.size() > 0);
    }

    // Simple Maze
    {
        /*
        MPEGI::Bitstream bitstream(simple_maze_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
        auto payload = std::static_pointer_cast<MPEGI::ReverbPayload const>(payloads[0]);
        auto reverbParameters = payload->GetReverbParameters();
        CHECK(reverbParameters.size() > 0);
        */
    }

    // Virtual Basketball
    {
        MPEGI::Bitstream bitstream(virtualBasketball_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
        auto payload = std::static_pointer_cast<MPEGI::ReverbPayload const>(payloads[0]);
        auto reverbParameters = payload->GetReverbParameters();
        CHECK(reverbParameters.size() > 0);
    }

    // Singer In The Lab
    {
        MPEGI::Bitstream bitstream(singer_bitstream);

        auto payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_REVERB);
        auto payload = std::static_pointer_cast<MPEGI::ReverbPayload const>(payloads[0]);
        auto reverbParameters = payload->GetReverbParameters();

        for (auto itr = reverbParameters.cbegin(); itr != reverbParameters.cend(); ++itr) {
            const auto& other = singer_reference_data.at(itr->second->id);
            CHECK(itr->second->id == other.id);
            CHECK(itr->second->freqGridIdx == other.freqGridIdx);
            CHECK(std::abs(itr->second->parOffset - other.parOffset) < 0.001f);
            CHECK(itr->second->t60.size() == other.t60.size());
            for (int i = 0; i < itr->second->t60.size(); ++i)
                CHECK(std::abs(itr->second->t60[i] - other.t60[i]) < 0.001f);
            CHECK(itr->second->dsr.size() == other.dsr.size());
            for (int i = 0; i < itr->second->dsr.size(); ++i)
                CHECK(std::abs(itr->second->dsr[i] - other.dsr[i]) < 0.001f);
        }
    }

    MPEGI::Bitstream::clearHandlerList();
}
} // namespace ReverbParsing_test