/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ReverbUnitTest.h"
#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>

namespace MPEGI {

DummyBitstream::DummyBitstream(bool hasDefaultAE)
{
    DummyReverbPayload rp;
    if (hasDefaultAE) {
        rp.addReverbParameters(DEFAULT_AE_ID);
    }
    reverbPayloads.push_back(std::make_shared<const DummyReverbPayload>(std::move(rp)));
}

DummyBitstream::DummyBitstream(const DummyReverbPayload& rp)
{
    reverbPayloads.push_back(std::make_shared<const DummyReverbPayload>(std::move(rp)));
}

DummyBitstream::DummyBitstream(
    std::shared_ptr<DummyReverbPayload> rp, std::shared_ptr<DummyPortalPayload> pp)
{
    reverbPayloads.push_back(rp);
    portalPayloads.push_back(pp);
}

void DummyBitstream::addExternalSources(
    const std::string& listenerAe, const std::string& itemAe, DummySceneState& sceneState)
{
    portalPayloads.emplace_back(
        std::make_shared<DummyPortalPayload>(listenerAe, itemAe, sceneState));
}

std::vector<std::shared_ptr<const Bitstream::Payload>> DummyBitstream::getAllPayloadsOfType(
    PayloadType type) const
{
    switch (type) {
    case PLD_REVERB:
        return reverbPayloads;
    case PLD_PORTAL:
        return portalPayloads;
    default:
        auto emptyPayload = std::vector<std::shared_ptr<const Bitstream::Payload>>();
        return emptyPayload;
    }
}

void DummyBitstream::clearPayloads()
{
    reverbPayloads.clear();
}

DummySceneState::DummySceneState(bool hasDefaultAe)
{
    if (hasDefaultAe) {
        ReverbEnv env = ReverbEnv(DEFAULT_AE_ID_NAME, nullptr);
        env.isLsdf = false;
        env.isSceneBasedAE = true;
        env.predelay = 0;
        env.acousticEnvironmentBsId = DEFAULT_AE_ID;
        dummyReverbEnvs.push_back(std::make_unique<ReverbEnv>(std::move(env)));
    }
}

void DummySceneState::addAcousticEnvironment(
    std::optional<std::string> aeId, EAR::Geometry* enclosure)
{
    ReverbEnv env = ReverbEnv(aeId.value_or("0"), enclosure);
    env.isLsdf = false;
    env.isSceneBasedAE = true;
    env.predelay = 0;
    env.acousticEnvironmentBsId = std::stoi(aeId.value_or("0"));
    dummyReverbEnvs.push_back(std::make_unique<ReverbEnv>(std::move(env)));
}

void DummySceneState::addLSDFEnvironment(std::optional<std::string> aeId, EAR::Geometry* enclosure)
{
    ReverbEnv env = ReverbEnv(aeId.value_or("0"), enclosure);
    env.isLsdf = true;
    env.isSceneBasedAE = false;
    env.predelay = 0;
    env.acousticEnvironmentBsId = std::nullopt;
    env.append(500.0F, 1.0F, std::nullopt);
    env.append(1000.0F, 1.0F, std::nullopt);
    env.append(2000.0F, 1.0F, std::nullopt);
    env.append(4000.0F, 1.0F, std::nullopt);
    env.append(8000.0F, 1.0F, std::nullopt);
    dummyReverbEnvs.push_back(std::make_unique<ReverbEnv>(std::move(env)));
}

// Create public method from private method of base class
void DummySceneState::createBox2(const EIFPayload::BoxDesc& boxDesc)
{
    SceneState::createBox(boxDesc);
}

// Create public method from private method of base class
void DummySceneState::createPortal2(MPEGI::Portal* portal)
{
    SceneState::createPortal(portal);
}

void DummySceneState::addPortalExtent(const std::string& extentId,
    const EAR::Vector3& portalPosition, const EAR::Vector3& portalBoxSize)
{
    // Create portal extent
    EIFPayload::BoxDesc portalBox;
    portalBox.position = EAR::Position(portalPosition, EAR::Euler());
    portalBox.size = portalBoxSize;
    portalBox.id = extentId;
    portalBox.bsId = 0;

    createBox2(portalBox);
}

const std::vector<std::unique_ptr<ReverbEnv>>& DummySceneState::getReverbEnvs() const
{
    return dummyReverbEnvs;
}

DummyReverbPayload::DummyReverbPayload()
{
    reverbFadeInDistance = 1.0f;
    reverbMinimumDistance = 1.0f;
    revDefaultAEGain = 1.0f;
    reverbDistanceGainDropDb = 1.5f;
    headTrackingEnabled = false;
    fadeInEnabled = false;

    std::vector<float> freqs = { 25.0f, 50.0f, 100.0f, 200.0f, 400.0f, 800.0f, 1600.0f, 3600.0f,
        7200.0f, 14400.0f };
    freqGridData.push_back(std::make_shared<FreqGrid>(10, freqs));
}

DummyPortalPayload::DummyPortalPayload()
{
    isExplicitPortalMode = true;
}

DummyPortalPayload::DummyPortalPayload(
    const std::string& listenerAe, const std::string& itemAe, DummySceneState& sceneState)
{
    isExplicitPortalMode = true;
    auto portal = std::make_unique<Portal>();
    const unsigned int portalId = 0;

    portal->portalId = portalId;
    portal->extentId = PORTAL_EXTENT_ID;
    portal->portalEnvironments = { listenerAe, itemAe };
    portal->normalVector = { std::sqrt(0.2f), std::sqrt(0.3f),
        std::sqrt(0.5f) }; // Random vector, assuming it doesn't matter
    sceneState.createPortal2(portal.get());

    this->portals.push_back(std::move(portal));
}

void DummyReverbPayload::addReverbParameters(int aeBsId)
{
    std::vector<float> dsr{ 2.51189e-8, 2.51189e-8, 3.16228e-8, 5.01187e-8, 5.01187e-8, 5.01187e-8,
        5.01187e-8, 6.30957e-8, 7.94328e-8, 7.94328e-8 };
    std::vector<float> t60{ 3.44, 3.67, 3.97, 4.32, 4.77, 4.85, 4.95, 5.07, 5.23, 5.44 };

    unsigned int freqGridIdx = 0;
    float parOffset = 0.f; // predelay coeff
    unsigned delayLineCount = 15;

    this->reverbParameters.emplace(aeBsId,
        std::make_shared<ReverbParameters>(
            aeBsId, freqGridIdx, parOffset, delayLineCount, t60, dsr));
}

int ReverbStageTester::acousticEnvironmentIdCounter = 1;

void ReverbStageTester::addPrimaryItem(
    std::string aeId, RenderItemStore& store, const StreamManager& streamManager)
{
    auto* primaryItem = store.createItem("PrimaryItem");
    renderItems = store.getItems();
    primaryItem->type = ItemType::Primary;
    primaryItem->acousticEnvironmentId = aeId;
    primaryItem->setSignal(streamManager.createEmptyBuffer());
    makeSound(primaryItem->signal->getFrame()[0U]);
    CHECK(primaryItem->signal->getFrame()[0U].rms() > 0);
    primaryItem->setGain(1.F);
    renderList.addItemCopy(primaryItem);
}

void ReverbStageTester::makeSound(EAR::Buffer<float>& buffer)
{
    auto* data = buffer.data();
    int i{ 0 };
    std::size_t BUF_LEN{ buffer.length() };
    const int DIV = 256;
    const int INC = DIV / 8 + 1;
    int counter = 0;
    while (i < BUF_LEN) {
        buffer[i++] = 2.f * (static_cast<float>(counter) / DIV) - 1.f;
        counter = (counter + INC) % DIV;
    }
}

bool ReverbStageTester::itemAudible(const RenderItem& item)
{
    bool audible = false;
    for (int ch = 0; ch < item.signal->getFrame().channels(); ch++) {
        if (item.signal->getFrame()[ch].rms() > 0.F) {
            audible = true;
        }
    }
    return audible;
}

void ReverbStageTester::checkIsAudible(const RenderItem& item, bool audible)
{
    bool soundDetected = itemAudible(item);
    if (!audible) {
        CHECK(!soundDetected);
    } else {
        CHECK(soundDetected);
    }
}

void ReverbStageTester::checkReverbOutputIsAudible(bool audible, std::optional<std::string> aeId)
{
    for (const auto& [idx, item] : renderItems) {
        if (item->type == ItemType::Reverb) {
            if (aeId.has_value() && aeId.value() == item->acousticEnvironmentId) {
                checkIsAudible(*item, audible);
            } else if (aeId == std::nullopt) {
                checkIsAudible(*item, audible);
            }
        }
    }
}

void ReverbStageTester::checkPannableOutputIsAudible(bool audible)
{
    // At least one of the pannable output items must be audible
    bool soundDetected = false;
    for (const auto& [idx, item] : renderItems) {
        if (item->type == ItemType::Pannable) {
            soundDetected = itemAudible(*item);
        }
        if (soundDetected) {
            break;
        }
    }
    if (!audible) {
        CHECK(!soundDetected);
    } else {
        CHECK(soundDetected);
    }
}

std::string ReverbStageTester::getAndIncAcousticEnvironmentId()
{
    return std::to_string(acousticEnvironmentIdCounter++);
}

void ReverbStageTester::testDefaultAcousticEnvironment()
{
    Config::setDef("Reverb:Default_Acoustic_Environment_Gain", EAR::Variant(1.f));

    // Initialize item store, scene state and reverb stage
    auto acousticEnvironmentId = "default";
    const bool hasDefaultAE = true;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    DummyBitstream bitstream(hasDefaultAE);
    DummySceneState sceneState(hasDefaultAE);
    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    // Create input render item and add it to render list
    renderItems = store.getItems();
    // 15 headtracked + 1 non-headtracked output
    constexpr std::size_t reverbItemCount{ 16 };
    CHECK(renderItems.size() == reverbItemCount);
    addPrimaryItem(acousticEnvironmentId, store, streamManager);
    CHECK(renderItems.size() == reverbItemCount + 1);
    renderList.listener.acousticEnvironmentId = acousticEnvironmentId;

    // Run one update round
    auto dspData = reverbStage.update(renderList);

    checkReverbOutputIsAudible(false);
    checkPannableOutputIsAudible(false);
    // Process reverb blocks for as many rounds as there are samples
    std::size_t processIterations{ TEST_PROCESS_ITERATIONS };
    while (processIterations--) {
        reverbStage.processBlock(RenderStage::DspContext{ *dspData.get() });
    }
    checkReverbOutputIsAudible(true);
    checkPannableOutputIsAudible(true);
}

void ReverbStageTester::testOneRoom()
{
    // Initialize item store, scene state and reverb stage
    auto defaultAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto roomAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    float pos = 4.0f;
    const bool hasDefaultAE = true;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    DummyReverbPayload reverbPayload;
    reverbPayload.addReverbParameters(DEFAULT_AE_ID);
    reverbPayload.addReverbParameters(std::stoi(roomAcousticEnvironmentId));
    DummyBitstream bitstream(reverbPayload);
    DummySceneState sceneState(hasDefaultAE);
    auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(EAR::Vector3(pos, pos, pos));
    sceneState.addAcousticEnvironment(roomAcousticEnvironmentId, box.get());
    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    // Create input render item and add it to render list
    renderItems = store.getItems();
    // 15 headtracked + 2 non-headtracked outputs
    constexpr std::size_t reverbItemCount{ 17 };
    CHECK(renderItems.size() == reverbItemCount);
    addPrimaryItem(roomAcousticEnvironmentId, store, streamManager);
    CHECK(renderItems.size() == reverbItemCount + 1);
    renderList.listener.acousticEnvironmentId = roomAcousticEnvironmentId;
    pos = pos / 4;
    renderList.listener.position = EAR::Position(EAR::Vector3(pos, pos, pos), EAR::Euler(0, 0, 0));

    // Run one update round
    auto dspData = reverbStage.update(renderList);

    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, roomAcousticEnvironmentId);
    checkPannableOutputIsAudible(false);
    // Process reverb blocks for as many rounds as there are samples
    std::size_t processIterations{ TEST_PROCESS_ITERATIONS };
    while (processIterations--) {
        reverbStage.processBlock(RenderStage::DspContext{ *dspData.get() });
    }
    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(true, roomAcousticEnvironmentId);
    checkPannableOutputIsAudible(true);
}

void ReverbStageTester::testCrossfade()
{
    // Explicit portal mode needs to be turned off for this test
    bool explicitPortalMode = Config::getBoolDef("ExplicitPortalMode", false);
    Config::set("ExplicitPortalMode", EAR::Variant(false));

    // Initialize item store, scene state and reverb stage
    auto defaultAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto roomAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    float pos = 4.0f;
    const bool hasDefaultAE = true;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    DummyReverbPayload reverbPayload;
    reverbPayload.addReverbParameters(DEFAULT_AE_ID);
    reverbPayload.addReverbParameters(std::stoi(roomAcousticEnvironmentId));
    DummyBitstream bitstream(reverbPayload);
    DummySceneState sceneState(hasDefaultAE);
    auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(EAR::Vector3(pos, pos, pos));
    sceneState.addAcousticEnvironment(roomAcousticEnvironmentId, box.get());
    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    // Create input render item and add it to render list
    renderItems = store.getItems();
    // 15 headtracked + 2 non-headtracked outputs
    constexpr std::size_t reverbItemCount{ 17 };
    CHECK(renderItems.size() == reverbItemCount);
    addPrimaryItem(roomAcousticEnvironmentId, store, streamManager);
    CHECK(renderItems.size() == reverbItemCount + 1);
    renderList.listener.acousticEnvironmentId = defaultAcousticEnvironmentId;
    // Position listener just outside the room to test crossfade
    pos = 2.02;
    renderList.listener.position = EAR::Position(EAR::Vector3(pos, pos, pos), EAR::Euler(0, 0, 0));

    // Run one update round
    auto dspData = reverbStage.update(renderList);

    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, roomAcousticEnvironmentId);
    checkPannableOutputIsAudible(false);
    // Process reverb blocks for as many rounds as there are samples
    std::size_t processIterations{ TEST_PROCESS_ITERATIONS };
    while (processIterations--) {
        reverbStage.processBlock(RenderStage::DspContext{ *dspData.get() });
    }
    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(true, roomAcousticEnvironmentId);
    checkPannableOutputIsAudible(true);

    Config::set("ExplicitPortalMode", EAR::Variant(explicitPortalMode));
}

void ReverbStageTester::textExternalSourceTwoRooms()
{
    MPEGI::Config::set("Reverb:SectorialDirectPropagationValue", EAR::Variant(false));
    // Initialize item store, scene state and reverb stage
    // Two rooms: sound source is in other room which is connected to listener's room.
    auto defaultAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto roomAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto otherRoomAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto aeBsId = std::stoi(roomAcousticEnvironmentId);
    auto otherAeBsId = std::stoi(otherRoomAcousticEnvironmentId);
    float pos = 4.0f;
    const bool hasDefaultAE = true;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);

    DummySceneState sceneState(hasDefaultAE);
    sceneState.addPortalExtent(PORTAL_EXTENT_ID, EAR::Vector3(0, 0, 0), EAR::Vector3(1, 1, 1));
    DummyReverbPayload reverbPayload;
    reverbPayload.addReverbParameters(DEFAULT_AE_ID);
    reverbPayload.addReverbParameters(aeBsId);
    reverbPayload.addReverbParameters(otherAeBsId);
    DummyBitstream bitstream(reverbPayload);
    bitstream.addExternalSources(
        roomAcousticEnvironmentId, otherRoomAcousticEnvironmentId, sceneState);
    auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(EAR::Vector3(pos, pos, pos));
    sceneState.addAcousticEnvironment(roomAcousticEnvironmentId, box.get());
    sceneState.addAcousticEnvironment(otherRoomAcousticEnvironmentId, box.get());
    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    // Create input render item and add it to render list
    renderItems = store.getItems();
    // 15 headtracked + 3 non-headtracked outputs
    constexpr std::size_t reverbItemCount{ 18 };
    CHECK(renderItems.size() == reverbItemCount);
    // Make the item external with regard to the listener's room so that the item's contribution
    // to the reverb is carried through a portal
    addPrimaryItem(otherRoomAcousticEnvironmentId, store, streamManager);
    CHECK(renderItems.size() == reverbItemCount + 1);
    renderList.listener.acousticEnvironmentId = roomAcousticEnvironmentId;
    pos = pos / 4;
    renderList.listener.position = EAR::Position(EAR::Vector3(pos, pos, pos), EAR::Euler(0, 0, 0));

    // Run one update round
    auto dspData = reverbStage.update(renderList);

    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, roomAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, otherRoomAcousticEnvironmentId);
    checkPannableOutputIsAudible(false);
    // Process reverb blocks for as many rounds as there are samples
    std::size_t processIterations{ TEST_PROCESS_ITERATIONS };
    while (processIterations--) {
        reverbStage.processBlock(RenderStage::DspContext{ *dspData.get() });
    }

    // The reverb should be audible in the room where it is located and also in the listener's room
    // because there is a connection. Default AE should remain silent.
    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(true, roomAcousticEnvironmentId);
    checkReverbOutputIsAudible(true, otherRoomAcousticEnvironmentId);
    checkPannableOutputIsAudible(true);
}

void ReverbStageTester::textExternalSourceThreeRooms()
{
    MPEGI::Config::set("Reverb:SectorialDirectPropagationValue", EAR::Variant(false));

    // Initialize item store, scene state and reverb stage
    // Three rooms: sound source is in source room which is connected to listener room.
    // The third room is not connected to the sound source.
    auto defaultAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto listenerAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto sourceAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto thirdRoomAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    auto listenerBsId = std::stoi(listenerAcousticEnvironmentId);
    auto sourceBsId = std::stoi(sourceAcousticEnvironmentId);
    auto thirdRoomBsId = std::stoi(thirdRoomAcousticEnvironmentId);
    float pos = 4.0f;
    const bool hasDefaultAE = true;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);

    DummySceneState sceneState(hasDefaultAE);
    sceneState.addPortalExtent(PORTAL_EXTENT_ID, EAR::Vector3(0, 0, 0), EAR::Vector3(1, 1, 1));
    DummyReverbPayload reverbPayload;
    reverbPayload.addReverbParameters(DEFAULT_AE_ID);
    reverbPayload.addReverbParameters(listenerBsId);
    reverbPayload.addReverbParameters(sourceBsId);
    reverbPayload.addReverbParameters(thirdRoomBsId);
    DummyBitstream bitstream(reverbPayload);
    bitstream.addExternalSources(
        listenerAcousticEnvironmentId, sourceAcousticEnvironmentId, sceneState);

    auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(EAR::Vector3(pos, pos, pos));
    sceneState.addAcousticEnvironment(listenerAcousticEnvironmentId, box.get());
    sceneState.addAcousticEnvironment(sourceAcousticEnvironmentId, box.get());
    sceneState.addAcousticEnvironment(thirdRoomAcousticEnvironmentId, box.get());

    ReverbStage reverbStage;
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    // Check that second order connections were correctly initialized
    for (const auto& reverbData : reverbStage.reverbData) {
        // Listener and source room are connected and coefficient is within bounds
        // Some clang versions seem to complain about using structured bindings with
        // the CHECK macros, so this is a workaround:
        const auto& id = reverbData.first;
        const auto& data = reverbData.second;
        if (id == listenerAcousticEnvironmentId) {
            CHECK(data.connectedEnvironments->size() == 1);
            CHECK(data.connectedEnvironments->count(sourceAcousticEnvironmentId) == 1);
            const auto& src = data.connectedEnvironments->find(sourceAcousticEnvironmentId);
            CHECK(src->second > 0.2);
            CHECK(src->second < 0.25);
        } else if (id == sourceAcousticEnvironmentId) {
            CHECK(data.connectedEnvironments->size() == 1);
            CHECK(data.connectedEnvironments->count(listenerAcousticEnvironmentId) == 1);
            const auto& src = data.connectedEnvironments->find(listenerAcousticEnvironmentId);
            CHECK(src->second > 0.2);
            CHECK(src->second < 0.25);
        } else if (id == thirdRoomAcousticEnvironmentId || id == defaultAcousticEnvironmentId) {
            // Third room and default ae are not connected to other AEs
            CHECK(data.connectedEnvironments->size() == 0);
        }
    }

    // Create input render item and add it to render list
    renderItems = store.getItems();
    // 15 headtracked + 4 non-headtracked outputs
    constexpr std::size_t reverbItemCount{ 19 };
    CHECK(renderItems.size() == reverbItemCount);
    // Make the item external with regard to the listener's room so that the item's contribution
    // to the reverb is carried through a portal
    addPrimaryItem(sourceAcousticEnvironmentId, store, streamManager);
    CHECK(renderItems.size() == reverbItemCount + 1);
    renderList.listener.acousticEnvironmentId = listenerAcousticEnvironmentId;
    pos = pos / 4;
    renderList.listener.position = EAR::Position(EAR::Vector3(pos, pos, pos), EAR::Euler(0, 0, 0));

    // Run one update round
    auto dspData = reverbStage.update(renderList);

    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, listenerAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, sourceAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, thirdRoomAcousticEnvironmentId);
    checkPannableOutputIsAudible(false);
    // Process reverb blocks for as many rounds as there are samples
    std::size_t processIterations{ TEST_PROCESS_ITERATIONS };
    while (processIterations--) {
        reverbStage.processBlock(RenderStage::DspContext{ *dspData.get() });
    }

    // The reverberation should be audible in the room where it is located and also in the listener's room
    // because there is a connection. Other room and default AE should remain silent.
    checkReverbOutputIsAudible(false, defaultAcousticEnvironmentId);
    checkReverbOutputIsAudible(true, listenerAcousticEnvironmentId);
    checkReverbOutputIsAudible(true, sourceAcousticEnvironmentId);
    checkReverbOutputIsAudible(false, thirdRoomAcousticEnvironmentId);
    checkPannableOutputIsAudible(true);
}

void ReverbStageTester::testLSDF()
{
    // Initialize item store, scene state and reverb stage
    const bool hasDefaultAE = false;
    auto lsdfAcousticEnvironmentId = getAndIncAcousticEnvironmentId();
    float pos = 4.0f;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    DummySceneState sceneState(false);
    auto box = EAR::SceneObjectFactory<EAR::Box>::getNewObject(EAR::Vector3(pos, pos, pos));
    sceneState.addLSDFEnvironment(lsdfAcousticEnvironmentId, box.get());
    ReverbStage reverbStage;
    reverbStage.initializeStage(DummyBitstream(hasDefaultAE), &store, &streamManager, &sceneState);

    // Create input render item and add it to render list
    renderItems = store.getItems();
    // 15 headtracked + 1 non-headtracked output
    constexpr std::size_t reverbItemCount{ 16 };
    CHECK(renderItems.size() == reverbItemCount);
    addPrimaryItem(lsdfAcousticEnvironmentId, store, streamManager);
    CHECK(renderItems.size() == reverbItemCount + 1);
    renderList.listener.acousticEnvironmentId = lsdfAcousticEnvironmentId;
    pos = pos / 4;
    renderList.listener.position = EAR::Position(EAR::Vector3(pos, pos, pos), EAR::Euler(0, 0, 0));

    // Run one update round
    auto dspData = reverbStage.update(renderList);

    checkPannableOutputIsAudible(false);
    // Process reverb blocks for as many rounds as there are samples
    std::size_t processIterations{ TEST_PROCESS_ITERATIONS };
    while (processIterations--) {
        reverbStage.processBlock(RenderStage::DspContext{ *dspData.get() });
    }
    checkPannableOutputIsAudible(true);
}

void ReverbStageTester::testMissingReverbPayload()
{
    // Initialize item store, scene state and reverb stage
    const bool hasDefaultAE = false;
    RenderItemStore store;
    DummyStreamManager streamManager(SAMPLING_FREQ, B);
    DummyBitstream bitstream(hasDefaultAE);
    DummySceneState sceneState(hasDefaultAE);
    // This removes reverb payloads from the bitstream
    bitstream.clearPayloads();
    ReverbStage reverbStage;
    // Initialization does nothing because there are no reverb payloads in bitstream
    reverbStage.initializeStage(bitstream, &store, &streamManager, &sceneState);

    renderItems = store.getItems();
    // 15 headtracked outputs
    constexpr std::size_t reverbItemCount{ 15 };
    CHECK(renderItems.size() == reverbItemCount);

    // Run one update round, but it does nothing because stage is not initialized
    auto dspData = reverbStage.update(renderList);

    // Function processBlock returns after doing nothing
    reverbStage.processBlock(RenderStage::DspContext{ *dspData });
}

TEST_CASE("Reverb stage")
{
    SUBCASE("One Room") { CHECK_NOTHROW(ReverbStageTester().testOneRoom()); }
    SUBCASE("Crossfade") { CHECK_NOTHROW(ReverbStageTester().testCrossfade()); }

    SUBCASE("External Source 2 Rooms Coefficient")
    {
        CHECK_NOTHROW(ReverbStageTester().textExternalSourceTwoRooms());
    }
    SUBCASE("External Source 3 Rooms Coefficient")
    {
        CHECK_NOTHROW(ReverbStageTester().textExternalSourceThreeRooms());
    }
    SUBCASE("LSDF") { CHECK_NOTHROW(ReverbStageTester().testLSDF()); }
    SUBCASE("Missing Reverb Payload")
    {
        CHECK_NOTHROW(ReverbStageTester().testMissingReverbPayload());
    }
    SUBCASE("Default Acoustic Environment")
    {
        CHECK_NOTHROW(ReverbStageTester().testDefaultAcousticEnvironment());
    }
}

} // namespace MPEGI
