/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#pragma once

#include "../ReverbStage.h"
#include <EAR/Box.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Bitstream/ReverbPayload.h>
#include <MPEGI/Renderer/RenderItem.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Stream/StreamManager.h>

#include <doctest/doctest.h>
#include <memory>
#include <mpegcore/datatypes_reverb.hpp>
#include <string>

namespace MPEGI {

// Bitstream id number for default acoustic environment
static const unsigned int DEFAULT_AE_ID = 99;
static const std::string DEFAULT_AE_ID_NAME{ "default" };
static const std::string PORTAL_EXTENT_ID{ "extentID" };

// StreamManager is abstract but it contains all the functionality needed for testing. Any
// dummy that compiles is fine for testing.
class DummyStreamManager : public StreamManager
{
public:
    DummyStreamManager(float fs, std::size_t B)
        : StreamManager(fs, B)
    {}

    std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const std::string& streamId, double t = 0.0) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    std::shared_ptr<StreamAccessBuffer> createAccessBuffer(
        const SceneState::AudioStream& stream) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    void removeAccessBuffer(StreamAccessBuffer* buf) override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }

    void nextBlock() override { EAR_CHECK_REL(false, "Not expected to be called"); }

    std::size_t getNbrOfChannels(const std::string& id) const override
    {
        EAR_CHECK_REL(false, "Not expected to be called");
    }
};

class DummySceneState : public SceneState
{
public:
    DummySceneState(bool hasDefaultAe);
    DummySceneState& operator=(const DummySceneState&) = delete;
    DummySceneState(const DummySceneState&) = delete;
    virtual ~DummySceneState() {}
    virtual const std::vector<std::unique_ptr<ReverbEnv>>& getReverbEnvs() const override;
    void addAcousticEnvironment(std::optional<std::string> aeId, EAR::Geometry* enclosure);
    void addLSDFEnvironment(std::optional<std::string> aeId, EAR::Geometry* enclosure);
    void addPortalExtent(const std::string& extentId, const EAR::Vector3& portalPosition,
        const EAR::Vector3& portalBoxSize);
    void createBox2(const EIFPayload::BoxDesc& box);
    void createPortal2(MPEGI::Portal* portal);

private:
    std::vector<std::unique_ptr<ReverbEnv>> dummyReverbEnvs;
};

class DummyReverbPayload : public ReverbPayload
{
public:
    DummyReverbPayload();
    DummyReverbPayload& operator=(const DummyReverbPayload&) = delete;
    DummyReverbPayload(const DummyReverbPayload& other) = default;
    virtual ~DummyReverbPayload() {}

    void addReverbParameters(int aeBsId);
};

class DummyPortalPayload : public PortalPayload
{
public:
    DummyPortalPayload();
    DummyPortalPayload(
        const std::string& listenerAe, const std::string& itemAe, DummySceneState& sceneState);
    DummyPortalPayload& operator=(const DummyPortalPayload&) = delete;
    DummyPortalPayload(const DummyPortalPayload& other) = default;
    virtual ~DummyPortalPayload() {}
    void addPortal(std::unique_ptr<Portal>&& portal) { portals.push_back(std::move(portal)); };
};

class DummyBitstream : public Bitstream
{
public:
    DummyBitstream(bool hasDefaultAE);
    DummyBitstream(const DummyReverbPayload& rp);
    DummyBitstream(std::shared_ptr<DummyReverbPayload> rp, std::shared_ptr<DummyPortalPayload> pp);
    DummyBitstream& operator=(const DummyBitstream&) = delete;
    DummyBitstream(const DummyBitstream&) = delete;
    virtual ~DummyBitstream() {}
    virtual std::vector<std::shared_ptr<const Bitstream::Payload>> getAllPayloadsOfType(
        PayloadType type) const override;
    void addReverbEnvironment(int aeBsId);
    void addExternalSources(
        const std::string& listenerAe, const std::string& itemAe, DummySceneState& sceneState);
    void clearPayloads();

private:
    std::vector<std::shared_ptr<const Bitstream::Payload>> reverbPayloads;
    std::vector<std::shared_ptr<const Bitstream::Payload>> portalPayloads;
};

class ReverbStageTester
{
public:
    ReverbStageTester()
        : renderList(0)
        , renderItems()
    {}

    void testDefaultAcousticEnvironment();
    void testOneRoom();
    void testCrossfade();
    void testLSDF();
    void testMissingReverbPayload();
    void textExternalSourceTwoRooms();
    void textExternalSourceThreeRooms();
    std::string getAndIncAcousticEnvironmentId();
    std::vector<std::pair<float, float>> getExternalSourceContribution(EAR::Vector3 portalPosition,
        EAR::Vector3 portalNormal, float portalHeight, float portalWidth, float portalDepth,
        const std::vector<EAR::Vector3>& sourcePositions);
    std::vector<std::pair<float, float>> getExternalSourceContribution_NR(
        EAR::Vector3 portalPosition, EAR::Vector3 portalNormal, float portalHeight,
        float portalWidth, float portalDepth, const std::vector<EAR::Vector3>& sourcePositions);

    const float SAMPLING_FREQ{ 48000.F };
    const std::size_t B{ 256 };
    const std::size_t TEST_PROCESS_ITERATIONS{ 16 };

private:
    void addPrimaryItem(
        std::string aeId, RenderItemStore& store, const StreamManager& streamManager);
    void checkReverbOutputIsAudible(bool audible, std::optional<std::string> aeId = std::nullopt);
    void checkPannableOutputIsAudible(bool audible);
    bool itemAudible(const RenderItem& item);
    void checkIsAudible(const RenderItem& item, bool audible);
    void makeSound(EAR::Buffer<float>& buffer);
    std::unordered_map<int, MPEGI::RenderItem*> renderItems;
    RenderList renderList;

    // Required to match ReverbEnv static counter
    static int acousticEnvironmentIdCounter;
};

} // namespace MPEGI