/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "RoomAssignment.h"

#include <algorithm>

namespace MPEGI {

std::shared_ptr<RenderStage> makeRoomAssignmentStage()
{
    return std::make_shared<RoomAssignmentStage>();
}

RoomAssignmentStage::RoomAssignmentStage() {}

RoomAssignmentStage::~RoomAssignmentStage() {}

std::unique_ptr<RenderStage::DspData> RoomAssignmentStage::update(RenderList& renderList)
{
    // determine the reverb environment of the listener

    std::string defaultAEId = ""; // NOTE: Should be init'ed default acoustic environment id
    renderList.listener.acousticEnvironmentId =
        ""; // NOTE: Should be init'ed default acoustic environment id
    for (const auto& r : reverbEnvs) {
        if (r->enclosure == nullptr) {
            // reverb environment without geometry is a default one
            defaultAEId = r->acousticEnvironmentId;
        }
        if (r->enclosure != nullptr
            && r->enclosure->isInside(renderList.listener.position.location)) {
            renderList.listener.acousticEnvironmentId = r->acousticEnvironmentId;
            break;
        }
    }
    // if listener is not in any enclosure use the default envionment
    if (renderList.listener.acousticEnvironmentId.empty()) {
        renderList.listener.acousticEnvironmentId = defaultAEId;
    }

    for (const auto& item : renderList.items) {

        // determine reverb environment of the RenderItem
        item->acousticEnvironmentId = "";
        for (const auto& r : reverbEnvs) {
            if (r->enclosure != nullptr && r->enclosure->isInside(item->position.location)) {
                item->acousticEnvironmentId = r->acousticEnvironmentId;
                break;
            }
        }
        if (item->acousticEnvironmentId == "") {
            // Due to quantization errors, items close to the enclosure might end up outside erroneously.
            // If an item isn't detected as "inside", we check if it is very close by.
            // This value must be larger than the potential quantization error,
            // but small enough not to include nearby items that are meant to be outside.
            constexpr float REVERB_BLEEDING_DISTANCE = 0.005f;

            auto closest = std::pair(REVERB_BLEEDING_DISTANCE, std::string(""));
            for (const auto& r : reverbEnvs) {
                if (r->enclosure != nullptr) {
                    auto distance = (r->enclosure->closestPointTo(item->position.location)
                        - item->position.location)
                                        .abs();
                    closest = std::min(closest, std::pair(distance, r->acousticEnvironmentId));
                }
            }
            item->acousticEnvironmentId = closest.second;
        }
        if (item->acousticEnvironmentId == "") {
            item->acousticEnvironmentId = defaultAEId;
        }
    } // iterating over render items ends
    return {};
}

void RoomAssignmentStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    for (auto& r : state->getReverbEnvs()) {
        reverbEnvs.push_back(r.get());
    }
}

} // namespace MPEGI
