/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SimpleCulling.h"

#include <MPEGI/Helpers/Config.h>
#include <MPEGI/Helpers/Types.h>
#include <MPEGI/Renderer/RenderItemStore.h>

namespace MPEGI {

std::shared_ptr<RenderStage> makeSimpleCullingStage()
{
    return std::make_shared<SimpleCullingStage>();
}

//A-weighting (linear, 20Hz-20kHz, 1/3 Octave band)
//IEC 61672-1:2013
const std::vector<float> coefAWeight = { 0.00299, 0.00582, 0.01072, 0.01862, 0.03090, 0.04898,
    0.07499, 0.11092, 0.15668, 0.21380, 0.28510, 0.37154, 0.46774, 0.57544, 0.69183, 0.80353,
    0.91201, 1.00000, 1.07152, 1.12202, 1.14815, 1.16145, 1.14815, 1.12202, 1.05925, 0.98855,
    0.88105, 0.74989, 0.60954, 0.46774, 0.34277 };

SimpleCullingStage::SimpleCullingStage()
    : itemStore(nullptr)
{}

SimpleCullingStage::~SimpleCullingStage() {}

std::unique_ptr<RenderStage::DspData> SimpleCullingStage::update(RenderList& renderList)
{
    if (!itemStore) {
        return {};
    }

    bool enableCullingReflRI = Config::getBool("EnableCullingReflectionRI");
    Vector3 lp = renderList.listener.position.location;
    float speedOfSound = Config::getFloat("SpeedOfSound");
    float thresh = Config::getFloat("SimpleCulling_DeactivationThreshold");

    float DeactivationThreshold = Config::getFloat("SimpleCulling_DeactivationThreshold");
    float DeactivationThresholdRel =
        Config::getFloatDef("SimpleCulling_DeactivationThresholdRel", 0.0);

    std::map<int, float> itemGain;

    // first loop, calculate item gains and max item gain (incl. loudness factor)

    float maxItemGain = 1.0e-10f;

    for (auto& item : renderList.items) {

        if (item->type != ItemType::Primary && item->type != ItemType::Representative
            && item->type != ItemType::Reflection && item->type != ItemType::DirectionalReverb
            && item->type != ItemType::HOA && item->type != ItemType::Diffraction) {
            // only HOA, portal, primary and directly derived render items processed here
            continue;
        }

        // do not touch extended items
        if (item->extent)
            continue;

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        if (item->aparams & AParam::NoCulling) {
            continue;
        }

        // calculate minimal attenuation of gain and EQ combined
        float maxBandGain = *std::max_element(item->eq.begin(), item->eq.end());
        itemGain[item->idx] = maxBandGain * item->gain;

        if (!item->loudness.has_value()) {
            continue;
        }

        if (item->status != ItemStatus::Inactive
            && itemGain[item->idx] * item->loudness.value() > maxItemGain) {
            maxItemGain = itemGain[item->idx] * item->loudness.value();
        }
    }

    // second loop, culling
    for (auto& item : renderList.items) {

        if (item->type != ItemType::Primary && item->type != ItemType::Representative
            && item->type != ItemType::Reflection && item->type != ItemType::DirectionalReverb
            && item->type != ItemType::HOA && item->type != ItemType::Diffraction) {
            // only HOA, portal, primary and directly derived render items processed here
            continue;
        }

        // do not touch extended items
        if (item->extent)
            continue;

        if (item->status == ItemStatus::Orphaned) {
            continue;
        }

        if (item->aparams & AParam::NoCulling) {
            continue;
        }

        auto& deactivate = item->getStateOrCreate<SimpleCullingStage::ItemState>().deactivate;

        bool freshActivation = false;
        if (item->changed
            & (ItemProperty::Gain | ItemProperty::Loudness | ItemProperty::Eq
                | ItemProperty::Status)) {

            // if attenuation is below threshold (default -60dB), or relative attenuation is below relative threshold,
            // deactivate the source

            if (item->status != ItemStatus::Inactive
                && (itemGain[item->idx] < DeactivationThreshold
                    || (item->loudness.has_value()
                        && itemGain[item->idx] * item->loudness.value() / maxItemGain
                            < DeactivationThresholdRel))) {

                deactivate = true;
            } else {
                if (deactivate) {
                    freshActivation = true;
                }
                deactivate = false;
            }
        }

        if (deactivate) {
            itemStore->deactivate(item.get());
        } else if (freshActivation) {
            item->setStatus(ItemStatus::Active, true); // explicitly reactivate
        }

        if (enableCullingReflRI && !voxelEREnable) {
            if (item->changed & (ItemProperty::Gain | ItemProperty::Eq | ItemProperty::Status)) {

                if (item->type == ItemType::Primary && item->status != ItemStatus::Inactive) {
                    auto& eqPSUM = item->getStateOrCreate<SimpleCullingStage::ItemState>().eqPSUM;
                    eqPSUM = 0.f;
                    //calculate averaged EQ of Primary in A-weighted domain
                    for (int i = 0; i < Config::getInt("EQNBands"); i++) {
                        eqPSUM += item->eq[i] * coefAWeight[i];
                    }
                } else if (item->type == ItemType::Reflection && item->status == ItemStatus::Active
                    && item->parentRI != -1) {

                    auto itemP = renderList.getItemByIndex(item->parentRI); //get the primary RI

                    //do not touch extended items or if the primary RI is Inactive
                    if (itemP->status == ItemStatus::Inactive
                        || itemP->type == ItemType::ListenerVoice || itemP->extent) {
                        continue;
                    }

                    if (itemP->gain != 0.f) {

                        Vector3 sp = item->position.location; //position of Reflection RI
                        Vector3 spP = itemP->position.location; //position of parent Primary RI

                        float dist_lp_sp = (sp - lp).abs();
                        float dist_lp_spP = (spP - lp).abs();

                        //calculate distances from Reflection and Primary RIs
                        float distDiff = dist_lp_sp - dist_lp_spP;

                        //time difference between Primary and Reflection RIs in ms
                        float timeDiff = (distDiff / speedOfSound) * 1000.f;

                        //culling is applied for timeDiff > 1.0 and timeDiff <= 40.0
                        if (timeDiff > 1.f && timeDiff <= 40.f) {

                            float gainDiff = 0.f; //ratio of gains of Reflection and Primary RIs
                            float gainThresh = 0.f; //Culling Threshold

                            //calculate DispersionFilter gain in linear
                            bool dispER = Config::getBoolDef("Dispersion_Early_Reflections", false);
                            float gnA = 1.f;

                            if (dispER) {
                                auto gn = Config::getFloat("Dispersion_Gain_dB");
                                gnA = pow(10.f, gn / 20.f);
                            }

                            float eqSUM = 0.f;
                            auto eqPSUM =
                                itemStore->getState<SimpleCullingStage::ItemState>(item->parentRI)
                                    ->eqPSUM;

                            //calculate averaged EQ in A-weighted domain
                            for (int i = 0; i < Config::getInt("EQNBands"); i++) {
                                eqSUM += item->eq[i] * coefAWeight[i];
                            }

                            //include DispersionFilter gain
                            if (eqSUM != 0.f && eqPSUM != 0.f) {
                                gainDiff = 20.f
                                    * log10((item->gain * eqSUM * gnA) / (itemP->gain * eqPSUM));

                            } else {
                                gainDiff = 20.f * log10((item->gain * gnA) / itemP->gain);
                            }

                            //calculate culling threshold
                            if (timeDiff < 10.f)
                                gainThresh = -0.1f * timeDiff - 17.5f;

                            else if (timeDiff < 20.f)
                                gainThresh = -0.55f * timeDiff - 13.f;

                            else if (timeDiff < 30.f)
                                gainThresh = -0.6f * timeDiff - 12.f;

                            else
                                gainThresh = -0.8f * timeDiff - 6.f;

                            //deactivate Reflection RI if gain difference is less than culling threshold
                            if (gainDiff < gainThresh) {
                                item->setStatus(ItemStatus::Inactive, true);
                            }
                        }
                    }
                }
            }
        }
    }

    return {};
}

void SimpleCullingStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    this->itemStore = itemStore;
    voxelEREnable = Config::getBoolDef("VoxelEarlyReflectionsEnabled", false);
}

} // namespace MPEGI
