/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SphoaStage.hpp"

#include <algorithm>
#include <cmath>
#include <string>
#include <utility>

#include <spdlog/fmt/fmt.h>

#include <EAR/Exception.h>
#include <EAR/HOASource.h>
#include <EAR/Listener.h>
#include <EAR/RealFrame.h>
#include <MPEGI/Renderer/RenderItemStore.h>
#include <MPEGI/Scene/SceneState.h>

using namespace mpegcore;

namespace MPEGI {

std::shared_ptr<RenderStage> makeSphoaStage()
{
    return std::make_shared<SphoaStage>();
}

namespace {

bool isSphoaSource(HOASource const& source)
{
    return source.getGroup() == -1
        && source.getRepresentation() == HOASource::Representation::Interior;
}

bool isSphoaRenderItem(RenderItem const& item)
{
    if (item.type != ItemType::HOA) {
        return false;
    }

    EAR_CHECK_REL(item.hoaInfo->source, "HOA RenderItem is missing HOASource");
    return isSphoaSource(*item.hoaInfo->source);
}

mpegcore::Rotation toCoreRotation(Euler const& euler) noexcept
{
    return { mpegcore::Rotation{ euler[0], -euler[1], -euler[2] } };
}

Euler relativeOrientation(Euler const& reference, Euler orientation) noexcept
{
    orientation[0] += reference[0];
    orientation[1] += reference[1];
    orientation[2] += reference[2];
    return orientation;
}

} // namespace

SphoaStage::SphoaStage()
    : config{ mpegcore::Config{} }
    , hoaBus{ config }
{
    sphoaGain = Config::getFloatDef("Sphoa:gain", EAR::Variant(1.F));
    config.hrirSofaFile = "";
    // TODO: use the file name that has been passed in:
    //config.hrirSofaFile = Config::getStringDef("HrirSofaFile", "");
}

SphoaStage::~SphoaStage() {}

std::unique_ptr<RenderStage::DspData> SphoaStage::update(RenderList& renderList)
{
    if (B == 0) { // only update once initialized
        return {};
    }

    auto dspData = std::make_unique<SphoaStage::DspData>();
    auto& dspItems = dspData->dspItems;
    auto const hoaInputCountAtBegin = hoaInputs.size();

    for (auto const& item : renderList.items) {

        if (!isSphoaRenderItem(*item)) {
            continue;
        }

        auto* input = hoaInputFor(*item);

        if ((item->changed & ItemProperty::Orientation) || item->changed & ItemProperty::Position
            || renderList.listener.hasMoved) {
            EAR_CHECK_REL(
                item->position.cSpace == CSpace::WCS, "Non-world spaced HoA items not handled");
            auto const rotation = toCoreRotation(relativeOrientation(
                renderList.listener.position.orientation, item->position.orientation));
            input->inputBus.setRotation(rotation);

            // Don't care about location changes as long as we consider Sphoa to always be
            // "noDoppler" and "noDistance".
        }

        if (item->changed & ItemProperty::Signal) {
            // We don't get any pre-indication of the signal being swapped so there's no time to
            // fade the previous signal out. The best we could do is fade the new signal in, but we
            // don't do that either for now. There's going to be an audio glitch since the previous
            // signal was abruptly cut.

            auto const itemSignalChannelCount = item->signal->getFrame().channels();

            // Changing signal's channel count is not supported.
            EAR_CHECK_REL(input->signalChannelCount == 0
                    || itemSignalChannelCount == input->signalChannelCount,
                "SphoaStage: changing HoA signal channel count not supported");

            input->signalChannelCount = itemSignalChannelCount;
        }

        // Dismiss changes that we get in the CfP scenes but don't really handle in any way.

        if (item->changed & ItemProperty::Gain) {
            // Gain changes are handled by EqStage which needs to be placed in front of SphoaStage.
        }

        if (item->changed & ItemProperty::Status) {
            // Don't care about status changes. Fading is handled by FadeStage which needs to be
            // placed in front of SphoaStage. For now we don't care about the performance hit so we
            // just keep processing all items regardless of their status.
        }

        if (item->changed & ItemProperty::Aparams) {
            // Sphoa is considered to be "noDistance" and "noDoppler",  even when not
            // explicitly specified to be so. Therefore we dismiss Aparams changes.
        }

        if (item->changed & ItemProperty::Location) {
            // Don't care about location changes as long as we consider Sphoa to always be
            // "noDoppler" and "noDistance".
        }

        if (item->changed & ItemProperty::Extent) {
            // We don't handle extent in Sphoa rendering.
        }

        // Bulk rejection of property changes we don't handle yet as they are not expected to happen
        // in CfP scenes.

#define NOT_SUPPORTED(x)                                                                           \
    if (item->changed & (x)) {                                                                     \
        EAR_CHECK_REL(false, "Sphoa doesn't support " #x);                                         \
    }

        NOT_SUPPORTED(ItemProperty::Directivity);
        NOT_SUPPORTED(ItemProperty::Eq);
        NOT_SUPPORTED(ItemProperty::RefDist);
        NOT_SUPPORTED(ItemProperty::ParentRI);
        NOT_SUPPORTED(ItemProperty::Reverb);
        NOT_SUPPORTED(ItemProperty::ReflectionInfo);

        auto& dspItem = dspItems.emplace_back();
        dspItem.hoaInput = input;
        dspItem.signalFrame = &item->signal->getFrame();
    }

    for (auto* ri : outputRenderItems) {
        renderList.addItemCopy(ri);
        ri->resetChanged();
    }

    auto const hoaInputCountAtEnd = hoaInputs.size();

    EAR_CHECK_REL(hoaInputCountAtBegin == 0 || hoaInputCountAtBegin == hoaInputCountAtEnd,
        "HoA inputs can't be changed dynamically");

    return dspData;
}

void SphoaStage::processBlock(DspContext ctx)
{
    if (suspendProcessing) {
        return;
    }

    auto& dspItems = ctx.getData<SphoaStage::DspData>().dspItems;

    if (dspItems.empty()) {
        return;
    }

    for (auto& dspItem : dspItems) {
        auto const channelCount = dspItem.hoaInput->inputAudios.size();

        for (auto i{ channelCount }; i-- > 0;) {
            esdConverter.copyToInput(dspItem.signalFrame->ch(i).data(), B, i);
        }

        auto const order = static_cast<std::size_t>(std::sqrt(channelCount) - 1);
        auto const wasConversionOk = esdConverter.convertEsdToHoa(order);
        EAR_CHECK_REL(wasConversionOk, "ESD conversion failed");

        for (auto i{ channelCount }; i-- > 0;) {
            copy(esdConverter.getOutputData(i), dspItem.hoaInput->inputAudios[i].audioBuffer);
        }
    }

    hoaDecoder.render();

    auto& outputSignalFrame = outputRenderItems[0]->signal->getFrame();
    auto const& outputArray = hoaDecoder.outputBuffer();

    for (auto i{ outputArray.size() }; i-- > 0;) {
        copy(outputArray[i], outputSignalFrame.ch(i).data());
    }

    // Once item gain and input fade handling has settled to a known shape, check if this stage gain
    // could be handled there for free instead of having to do this multiplication at the end.
    if (sphoaGain != 1.F) {
        outputSignalFrame.mul(sphoaGain);
    }
}

void SphoaStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    if (!state) {
        return;
    }

    auto const& hoaSources = state->getHOASources();
    if (hoaSources.empty()) {
        suspendProcessing = true;
        return;
    }
    suspendProcessing = false;

    B = streamManager->getB();

    // Buffer sizes etc get initiated based on config, so make sure we are running at same sized
    // blocks.
    EAR_CHECK_REL(B == config.processingSamplesPerBuffer,
        "SphoaStage: Mismatch of B and processingSamplesPerBuffer");

    esdConverter.initialize(B);

    outputRenderItems.push_back(itemStore->createItem("sphoa"));
    outputRenderItems.back()->type = ItemType::Downmix;
    constexpr std::size_t OUTPUT_CHANNEL_COUNT{ 2 };
    outputRenderItems.back()->setSignal(streamManager->createEmptyBuffer(OUTPUT_CHANNEL_COUNT));

    for (auto const& source : hoaSources) {
        if (!isSphoaSource(*source)) {
            continue;
        }
        int channelCount = 0;
        if (state->hasAudioStream(source.get())) {
            auto const& streamId = state->getAudioStream(source.get()).id;
            channelCount = streamManager->getNbrOfChannels(streamId);
        } else {
            const MPEGI::SceneState::GranularSound* granularSound =
                state->getGranularSound(source.get());
            channelCount = granularSound->channels;
        }
        createHoaInput(source->getName(), channelCount);
    }

    auto const isDecoderInitialized = hoaDecoder.initialize(config, hoaBus);
    EAR_CHECK_REL(isDecoderInitialized, "SphoaStage: failed to initialize HoaDecoder");
}

SphoaStage::HoaInput* SphoaStage::hoaInputFor(RenderItem const& ri)
{
    if (auto it = hoaInputs.find(ri.objectName); it != end(hoaInputs)) {
        return &it->second;
    }

    LSDF_FAIL(fmt::format("SphoaStage: Couldn't find an input for object {}", ri.objectName));
    return nullptr;
}

void SphoaStage::createHoaInput(std::string const& eifObjectId, std::size_t channelCount)
{
    auto& inputBus = hoaBus.createInputBus(shOrderFromChannelCount(channelCount));

    std::vector<InputAudio> inputAudios;
    inputAudios.reserve(channelCount);
    auto inputBusBufferIt = inputBus.buffers().begin();

    for (auto i{ channelCount }; i > 0; --i) {
        inputAudios.push_back({ *inputBusBufferIt++ });
    }

    auto [it, did] = hoaInputs.emplace(eifObjectId, HoaInput{ inputBus, std::move(inputAudios) });
    EAR_CHECK_REL(did, "Failed to add a HoaInput");
}

} // namespace MPEGI
