/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "MPEGI/Helpers/SecondaryItemStorage.h"
#include "MPEGI/Helpers/VoxelTypes.h"

#include <cmath>
#include <type_traits>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
class IVoxelDiffraction
{
public:
    virtual ~IVoxelDiffraction() = 0;
    IVoxelDiffraction(Voxel::VoxelSceneState::shared const& state, class RenderItemStore& store);

    /**
     * Process one frame
     * @param render_list Render-item list
     */
    virtual void process(RenderList& render_list) = 0;

protected:
    /**
     * Calculate diffracted source position relative to user location
     *
     * @param user_voxel_idx User voxel index
     * @paramitem_voxel_idx Item voxel index
     * @param pm_distance Estimated path length on the projection map
     * @param corner_idx Calculated corner index
     * @return Diffracted source position (az,el,r)
     */
    virtual std::array<float, 3> get_diffracted_source_aer_relative(
        Voxel::voxel_index_t user_voxel_idx, Voxel::voxel_index_t item_voxel_idx, float pm_distance,
        float additional_distance, Voxel::dmap_index_t const& corner_idx) const;

    /**
     * Get absolute diffracted source location
     *
     * @param src_aer Relative diffracted source location (az,el,r)
     * @return The diffracted source location (x,y,z)
     */
    virtual std::array<float, 3> get_diffracted_source_location(
        std::array<float, 3> const& src_aer) const;

    /**
     * Apply compensation EQs calculated by the VoxelOcclusionStage, if any
     *
     * @param eq EQ of the diffracted render-item
     * @param item_index Item index of the parent render-item
     */
    virtual void apply_compensation_eq(std::vector<float>& eq, int item_index) const;

    /**
     * Apply UDT EQs
     *
     * @param eq EQ of the diffracted source
     * @param diffraction_angle_deg Diffraction angle in deg
     * @param cb Callback function for further modifcation of UDT gains
     */
    virtual void apply_udt(
        std::vector<float>& eq, int diffraction_angle_deg,
        std::function<float(float)> const& cb = [](float val) { return val; }) const;

protected:
    /// Diffracted item store
    struct diffracted_item_store : public Voxel::secondary_item_storage<diffracted_item_store>
    {
        using shared = std::shared_ptr<diffracted_item_store>;
        diffracted_item_store(RenderItemStore* render_item_store, unsigned int num_bands_eq)
            : secondary_item_storage(render_item_store, num_bands_eq)
        {}

        ItemProperty get_sync_flags(RenderItem const&) const
        {
            return ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Aparams
                | ItemProperty::Position | ItemProperty::RefDist | ItemProperty::Reverb
                | ItemProperty::Signal | ItemProperty::Status;
        }

        ItemType get_item_type(RenderItem const&) const { return ItemType::Diffraction; }
        std::string get_suffix() const { return "_diffr_"; }
    };

protected:
    /// Basic scene info
    Voxel::voxel_scene_info_basic::shared _scene;

    /// To store/share calculated voxel indices
    Voxel::index_storage::shared _index_storage;

    /// Direct-view map (updated by VoxelOcclusionStage)
    Voxel::direct_view_map::shared _dv_map;

    /// Store diffracted items
    diffracted_item_store::shared _diffracted_item_store;

    /// Compensation EQs calculated in the VoxelOcclusionStage and applied here
    Voxel::VoxelSceneState::eq_map_ptr _compensation_eq;
};

//--------------------------------------------------------------------------------------------------
// Helpers
//--------------------------------------------------------------------------------------------------
inline bool diffraction_supported_for_item(MPEGI::RenderItem const& item) noexcept
{
    using EAR::CSpace;
    using MPEGI::ItemType;

    auto const i_type = item.type;
    auto const is_lcs = item.position.cSpace == CSpace::LCS;

    return (i_type == ItemType::Primary || (i_type == ItemType::Representative && !item.extent))
        && !is_lcs;
}

//--------------------------------------------------------------------------------------------------
inline float get_diffraction_angle(float r_direct, float r_diffracted) noexcept
{
    auto const angle_deg = 2.f * std::asin(std::clamp(r_direct / r_diffracted, 0.f, 1.f)) * 180.f
        / static_cast<float>(M_PI);

    return std::clamp(static_cast<int>(std::ceil(angle_deg)), 1, 180);
}

//--------------------------------------------------------------------------------------------------
template <typename T> std::array<T, 3> sph2cart(T az, T el, T r) noexcept
{
    // https://www.mathworks.com/help/matlab/ref/sph2cart.html
    static_assert(std::is_floating_point<T>::value, "Must be floating point");
    return { r * std::cos(el) * std::cos(az), r * std::cos(el) * std::sin(az), r * std::sin(el) };
}

template <typename T> std::array<T, 3> sph2cart(std::array<T, 3> const& sph) noexcept
{
    return sph2cart(sph[0], sph[1], sph[2]);
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------