/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "VoxelDiffraction.h"
#include "VoxelDiffractionJps.h"
#include "VoxelDiffractionRasterMap.h"

#include "MPEGI/Renderer/Udt.h"

#ifdef PM_DEBUG
#include <fstream>
#endif

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
std::shared_ptr<RenderStage> makeVoxelDiffractionStage()
{
    return std::make_shared<VoxelDiffractionStage>();
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    auto const has_raster_map_payload =
        !bitstream.getAllPayloadsOfType(Bitstream::PayloadType::PLD_RASTERMAP).empty();

    if (has_raster_map_payload && !state->getVoxelSceneState()->is_dynamic()) {
        _impl = std::make_unique<VoxelDiffractionRasterMap>(bitstream, *itemStore, *state);
    } else {
        _impl = std::make_unique<VoxelDiffractionJps>(bitstream, *itemStore, *state);
    }
}

//--------------------------------------------------------------------------------------------------
std::unique_ptr<RenderStage::DspData> VoxelDiffractionStage::update(RenderList& render_list)
{
    _impl->process(render_list);
    return {};
}

//--------------------------------------------------------------------------------------------------
#ifdef PF_DEBUG
void displayPathfinderResult(projection_map const& pm, Voxel::path_2d const& path,
    const Voxel::voxel_index_t& user_voxel_idx, const Voxel::voxel_index_t& item_voxel_idx)
{
    unsigned int i, x, y;
    unsigned char* disp;

    disp = new unsigned char[pm.dim[0] * pm.dim[1]];

    // MAP
    for (y = 0; y < pm.dim[1]; y++) {
        for (x = 0; x < pm.dim[0]; x++) {
            disp[x + y * pm.dim[0]] = pm.at(x, y) ? 219 : ' ';
        }
    }

    // OBJ
    x = item_voxel_idx[0] - 1;
    y = item_voxel_idx[1] - 1;

    disp[x + y * pm.dim[0]] = 'O';

    // USER
    x = user_voxel_idx[0] - 1;
    y = user_voxel_idx[1] - 1;
    disp[x + y * pm.dim[0]] = 'U';

    // PATH
    for (i = 0; i < path.size(); i++) {
        auto const pe{ path.get(i) };
        x = pe[0];
        y = pe[1];
        if (disp[x + y * pm.dim[0]] == ' ')
            disp[x + y * pm.dim[0]] = '*';
    }

    for (y = 0; y < pm.dim[1]; y++) {
        for (x = 0; x < pm.dim[0]; x++) {
            printf("%c", disp[(pm.dim[1] - 1 - y) * pm.dim[0] + x]);
        }
        printf("\n");
    }
    delete disp;
}
#endif

//--------------------------------------------------------------------------------------------------
// Voxel diffraction interface
//--------------------------------------------------------------------------------------------------
IVoxelDiffraction::IVoxelDiffraction(
    Voxel::VoxelSceneState::shared const& state, RenderItemStore& store)
    : _diffracted_item_store(
        std::make_shared<diffracted_item_store>(&store, Config::getInt("EQNBands")))
{
    EAR_CHECK_REL(state != nullptr, "VoxelOcclusionStage: Invalid VoxelSceneState");

    // Index storage and DV-map are sub-scene independent
    _index_storage = state->get_index_storage();
    _dv_map = state->get_direct_view_map();
    _scene = state->get_scene_info_basic();
    _compensation_eq = state->get_eq_map();
}

//--------------------------------------------------------------------------------------------------
IVoxelDiffraction::~IVoxelDiffraction() {}

//--------------------------------------------------------------------------------------------------
std::array<float, 3> IVoxelDiffraction::get_diffracted_source_aer_relative(
    Voxel::voxel_index_t user_voxel_idx, Voxel::voxel_index_t item_voxel_idx, float pm_distance,
    float additional_distance, Voxel::dmap_index_t const& corner_idx) const
{
    // Corner to user vector (2D)
    std::array<float, 2> const c2u = { static_cast<float>(corner_idx[0]) - user_voxel_idx[0],
        static_cast<float>(corner_idx[1]) - user_voxel_idx[1] };

    auto const azimuth = std::atan2(c2u[1], c2u[0]);

    // Source to user vector
    std::array<float, 3> const o2u = { static_cast<float>(item_voxel_idx[0]) - user_voxel_idx[0],
        static_cast<float>(item_voxel_idx[1]) - user_voxel_idx[1],
        static_cast<float>(item_voxel_idx[2]) - user_voxel_idx[2] };

    auto const o2u_xy =
        std::sqrt(o2u[0] * o2u[0] + o2u[1] * o2u[1]); // Object to User distance on PM
    auto const o2u_z = o2u[2]; // Object to user on z
    auto const elevation = std::atan2(o2u_z, o2u_xy);
    auto const o2u_dist = std::sqrt(pm_distance * pm_distance + o2u_z * o2u_z);

    // Radius
    auto const radius = _scene->voxel_size * o2u_dist + additional_distance;
    return { azimuth, elevation, radius };
}

//--------------------------------------------------------------------------------------------------
std::array<float, 3> IVoxelDiffraction::get_diffracted_source_location(
    std::array<float, 3> const& src_aer) const
{
    auto src_xyz = sph2cart(src_aer);
    auto const usr_xyz = _scene->idx2xyz(_index_storage->get_listener_voxel_index());
    src_xyz[0] += usr_xyz[0];
    src_xyz[1] += usr_xyz[1];
    src_xyz[2] += usr_xyz[2];

    return src_xyz;
}

//--------------------------------------------------------------------------------------------------
void IVoxelDiffraction::apply_compensation_eq(std::vector<float>& eq, int item_index) const
{
    auto& compensation_eq_map = *_compensation_eq;
    auto it = compensation_eq_map.find(item_index);
    if (std::cend(compensation_eq_map) == it) {
        return;
    }

    auto const& eq_comp = it->second;
    for (auto b = 0U; b < eq.size(); ++b) {
        eq[b] = std::min(eq[b], eq_comp[b]);
    }
}

void IVoxelDiffraction::apply_udt(
    std::vector<float>& eq, int diffraction_angle_deg, std::function<float(float)> const& cb) const
{
    auto const tuning_param = Config::getFloatDef("VoxelDiffractionFilterTuneParameter", 1.f);

    for (auto b = 0U; b < eq.size(); ++b) {
        // Apply UDT
        eq[b] = tuning_param * cb(Diffraction_FilterCoeffs[b][180 - diffraction_angle_deg]);
    }
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
