/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "VoxelDiffractionJps.h"
#include "CornerFindingAlgorithmC.h"
#include "jps_pathfinder_impl.h"

#include "MPEGI/Helpers/SecondaryItemStorage.h"
#include "MPEGI/Helpers/VoxelTypes.h"

//--------------------------------------------------------------------------------------------------
namespace {
//--------------------------------------------------------------------------------------------------
float wrapTo180(float phi)
{
    if (phi < -M_PI) {
        phi += 2 * M_PI;
    }
    if (phi > M_PI) {
        phi -= 2 * M_PI;
    }
    return (phi);
}

//--------------------------------------------------------------------------------------------------
} // namespace
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
//--------------------------------------------------------------------------------------------------
VoxelDiffractionJps::VoxelDiffractionJps(
    Bitstream const& bitstream, RenderItemStore& store, SceneState const& state)
    : IVoxelDiffraction(state.getVoxelSceneState(), store)
    , _pathfinder(std::make_unique<Voxel::jps_pathfinder>())
{
    // Default settings
    Config::setDef("VoxelDiffractionDisableCaching", EAR::Variant(false));
    Config::setDef("VoxelDiffractionDisableMultiPath", EAR::Variant(false));
    Config::setDef("VoxelDiffractionMultiPathAzimuthThreshold", EAR::Variant(30.0f));
    Config::setDef("VoxelStorePreComputedPathDataAtShutdown", EAR::Variant(false));
    Config::setDef("VoxelStorePreComputedPathDataByRequest", EAR::Variant(false));
    Config::setDef("VoxelPreComputedPathDataDirectory", EAR::Variant(std::string("/tmp")));

    Config::observeKey("VoxelStorePreComputedPathDataByRequest", this);

    auto const voxel_scene_state = state.getVoxelSceneState();
    EAR_CHECK_REL(voxel_scene_state != nullptr, "VoxelOcclusionStage: Invalid VoxelSceneState");

    // Initialize
    init(*voxel_scene_state);

    // Register callbacks
    voxel_scene_state->on_subscene_change(
        [this](Voxel::VoxelSceneState::const_shared const& voxel_scene_state) {
            std::lock_guard<std::mutex> lck(_update_mutex);
            reset();
            init(*voxel_scene_state);
        });

    voxel_scene_state->on_scene_update(
        [this](Voxel::VoxelSceneState::const_shared const& voxel_scene_state) {
            std::lock_guard<std::mutex> lck(_update_mutex);
            reset();
            init(*voxel_scene_state);
            _pathfinder_cache->reset();
        });
}

VoxelDiffractionJps::~VoxelDiffractionJps()
{
    Config::removeObserver("VoxelStorePreComputedPathDataByRequest", this);
}

//--------------------------------------------------------------------------------------------------
auto VoxelDiffractionJps::find_secondary_path(Voxel::dmap_index_t const& user_dmap_idx,
    Voxel::dmap_index_t const& item_dmap_idx, Voxel::dmap_index_t const& corner_primary,
    std::array<float, 3> const& dfr_location_rel_primary, float azimuth_threshold)
    -> sec_path_info_t
{
    std::array<float, 3> dfr_location_rel_secondary = { 0.0f, 0.0f, 0.0f };
    float secondary_source_weighting = 0.0f;

    sec_path_info_t sec_path_info;

    std::vector<std::array<int, 2>> close_voxel_neighbors = { { 0, 1 }, { 0, -1 }, { 1, 0 },
        { -1, 0 }, { 1, 1 }, { -1, 1 }, { 1, -1 }, { -1, -1 } };

    // update list of possible close voxel neighbors
    for (auto it = close_voxel_neighbors.begin(); it != close_voxel_neighbors.end();) {
        int new_x = (*it)[0] + (int)user_dmap_idx[0];
        int new_y = (*it)[1] + (int)user_dmap_idx[1];

        bool is_inside = (new_x > 0) && (new_x <= _projection_map->dim[0]) && (new_y > 0)
            && (new_y <= _projection_map->dim[1]);
        if (!is_inside) {
            it = close_voxel_neighbors.erase(it);
        } else {
            ++it;
        }
    }

    // check if user is close to single voxel occluding element and close to corner

    if (_single_voxel_occluder_neighbour_map->at(user_dmap_idx[0] - 1, user_dmap_idx[1] - 1)) {
        std::array<int, 2> const c2u = { static_cast<int>(corner_primary[0])
                - static_cast<int>(user_dmap_idx[0]),
            static_cast<int>(corner_primary[1]) - static_cast<int>(user_dmap_idx[1]) };
        // check if corner is close to user
        if (c2u[0] * c2u[0] + c2u[1] * c2u[1] < 4) {
            // get single voxel occluding element position
            for (auto neighbor : close_voxel_neighbors) {
                Voxel::dmap_index_t const occluder_dmap_idx = { user_dmap_idx[0] + neighbor[0],
                    user_dmap_idx[1] + neighbor[1] };

                if (_single_voxel_occluder_map->at(
                        occluder_dmap_idx[0] - 1, occluder_dmap_idx[1] - 1)) {
                    // Occluder to user vector
                    std::array<int, 2> const o2u = {
                        static_cast<int>(occluder_dmap_idx[0]) - static_cast<int>(user_dmap_idx[0]),
                        static_cast<int>(occluder_dmap_idx[1]) - static_cast<int>(user_dmap_idx[1])
                    };

                    auto const azimuth_user_occluder = std::atan2(o2u[1], o2u[0]);
                    auto const azimuth_user_corner = dfr_location_rel_primary[0];
                    const float sign_direction_arround_occluder =
                        (azimuth_user_occluder - azimuth_user_corner) < 0.0f ? -1.0f : 1.0f;

                    dfr_location_rel_secondary = dfr_location_rel_primary;
                    dfr_location_rel_secondary[0] =
                        azimuth_user_corner + M_PI_2 * sign_direction_arround_occluder;
                    secondary_source_weighting = 1.0f / 3.0f;
                    sec_path_info =
                        std::make_pair(dfr_location_rel_secondary, secondary_source_weighting);
                    break;
                }
            }
        }
    }

    // process all neighbors of user and try to find a second path with differnet corner

    if (!sec_path_info.has_value()) {

        for (auto neighbor : close_voxel_neighbors) {

            Voxel::dmap_index_t const user_dmap_idx_neighbor{ user_dmap_idx[0] + neighbor[0],
                user_dmap_idx[1] + neighbor[1] };

            auto path_info_opt = _pathfinder_cache->get(user_dmap_idx_neighbor, item_dmap_idx);

            if (path_info_opt.has_value()) {
                auto const& path_info = path_info_opt.value();

                if (!path_info.length.has_value()) {
                    // No path from user to item
                    continue;
                }

                float path_length_secondary = path_info.length.value();
                Voxel::dmap_index_t corner_secondary = path_info.corner;

                if (corner_secondary != corner_primary) {
                    const float azimuth_threshold_rad = azimuth_threshold * M_PI / 180.0f;

                    // Secondary corner to user vector (neighbor)
                    std::array<int, 2> const c2u = { static_cast<int>(corner_secondary[0])
                            - static_cast<int>(user_dmap_idx_neighbor[0]),
                        static_cast<int>(corner_secondary[1])
                            - static_cast<int>(user_dmap_idx_neighbor[1]) };

                    auto const azimuth_user_corner_primary = dfr_location_rel_primary[0];
                    auto const azimuth_user_corner_secondary = std::atan2(c2u[1], c2u[0]);

                    auto azimuth_delta = std::abs(
                        wrapTo180(azimuth_user_corner_primary - azimuth_user_corner_secondary));

                    if (azimuth_delta >= azimuth_threshold_rad) {
                        dfr_location_rel_secondary = dfr_location_rel_primary;
                        dfr_location_rel_secondary[0] = azimuth_user_corner_secondary;

                        if (std::abs(neighbor[0]) + std::abs(neighbor[1]) < 2)
                            secondary_source_weighting = 1.0f / 3.0f;
                        else
                            secondary_source_weighting = 1.0f / 5.0f;

                        sec_path_info =
                            std::make_pair(dfr_location_rel_secondary, secondary_source_weighting);
                        break;
                    }
                }
            }
        }
    }

    return sec_path_info;
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::populate_cache(Voxel::dmap_index_t const& user_dmap_idx,
    Voxel::dmap_index_t const& item_dmap_idx, Voxel::dmap_index_t const& corner_dmap_idx_primary,
    float path_length_primary)
{
    auto user_ap_dmap_idx = user_dmap_idx;
    auto path_length_pm_ap = path_length_primary;

    // Cash entry at current listener location
    _pathfinder_cache->set_path(user_dmap_idx, item_dmap_idx,
        Voxel::jps_info{ path_length_primary, corner_dmap_idx_primary });

    auto const& path = _pathfinder->get_path();
    for (int i = 1; i < path.size(); ++i) {
        auto const xy = path.get(i);
        // Voxel-indices start with 1
        auto const x = xy[0] + 1;
        auto const y = xy[1] + 1;

        auto const dx = std::abs(static_cast<int>(user_ap_dmap_idx[0] - x));
        auto const dy = std::abs(static_cast<int>(user_ap_dmap_idx[1] - y));
        EAR_CHECK(dx <= 1 && dy <= 1, "Path does not match user-position");

        user_ap_dmap_idx[0] = x;
        user_ap_dmap_idx[1] = y;

        if (user_ap_dmap_idx == corner_dmap_idx_primary) {
            break;
        }

        // Lateral or diagonal step
        path_length_pm_ap -= ((dx == 1 && dy == 1) ? M_SQRT2 : 1);
        _pathfinder_cache->set_path(user_ap_dmap_idx, item_dmap_idx,
            Voxel::jps_info{ path_length_pm_ap, corner_dmap_idx_primary });
    }
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::process(RenderList& render_list)
{
    auto const disable_caching = Config::getBool("VoxelDiffractionDisableCaching");
    auto const disable_multipath = Config::getBool("VoxelDiffractionDisableMultiPath");
    auto const azimuth_threshold = Config::getFloat("VoxelDiffractionMultiPathAzimuthThreshold");

    for (auto& item : render_list.items) {

        if (!diffraction_supported_for_item(*item)) {
            continue;
        }

        // If the primary item is orphaned (or fades-out to orphaned status), trigger removal of
        // all secondary (diffracted) items. Diffracted items are not actually removed here, but only
        // marked for removal
        if (item->isOrphaned() || item->isFadeOutOrphaned()) {
            _diffracted_item_store->remove(item->idx);
        }

        // If the primary item is inactive or visible, mute all secondary (diffracted) items
        if (_dv_map->direct_view(item->idx) || item->isInactive()) {
            _diffracted_item_store->mute(item->idx);
            continue;
        }

        auto user_voxel_idx = _index_storage->get_listener_voxel_index();
        auto const item_voxel_idx = _index_storage->get_voxel_index(item->idx);
        auto const additional_distance = _index_storage->get_additional_distance(item->idx);

        Voxel::dmap_index_t const user_dmap_idx{ user_voxel_idx[0], user_voxel_idx[1] };
        Voxel::dmap_index_t const item_dmap_idx{ item_voxel_idx[0], item_voxel_idx[1] };
        std::array<float, 3> dfr_location_rel_primary;
        float path_length_primary;
        Voxel::dmap_index_t corner_primary;

        float primary_source_weighting = 1.0f;

        // try to lookup
        /// @todo Check whether lookup needs to be synchronized as well
        auto path_info_opt = _pathfinder_cache->get(user_dmap_idx, item_dmap_idx);

        if (path_info_opt.has_value() && !disable_caching) {
            auto const& path_info = path_info_opt.value();

            if (!path_info.length.has_value()) {
                // No path from user to item
                _diffracted_item_store->mute(item->idx);
                continue;
            }
            path_length_primary = path_info.length.value();
            corner_primary = path_info.corner;

        } else {
            int res = Voxel::i_pathfinder::pathfinder_ok;
            {
                std::lock_guard<std::mutex> lck(_update_mutex);
                res = _pathfinder->find_path(user_voxel_idx, item_voxel_idx);
            }

            if (Voxel::i_pathfinder::pathfinder_ok != res) {
                /// No path:
                /// 1.) Either user or object position is on a wall
                /// 2.) No path has been found from user to object
                _pathfinder_cache->set_no_path(user_dmap_idx, item_dmap_idx);
                _diffracted_item_store->mute(item->idx);
                continue;
            }
#ifdef PF_DEBUG
            displayPathfinderResult(
                *_projection_map, _pathfinder->get_path(), user_voxel_idx, item_voxel_idx);
#endif
            path_length_primary = estimate_path_length();
            corner_primary = find_corner();

            populate_cache(user_dmap_idx, item_dmap_idx, corner_primary, path_length_primary);
        }

        dfr_location_rel_primary = get_diffracted_source_aer_relative(user_voxel_idx,
            item_voxel_idx, path_length_primary, additional_distance, corner_primary);

        // check for secondary path
        sec_path_info_t sec_path_info;
        if (!disable_multipath) {
            sec_path_info = find_secondary_path(user_dmap_idx, item_dmap_idx, corner_primary,
                dfr_location_rel_primary, azimuth_threshold);

            if (sec_path_info.has_value()) {
                auto const secondary_source_weighting = sec_path_info.value().second;
                primary_source_weighting = 1.0f - secondary_source_weighting;
            }
        }

        // To avoid discontinuities caused by different propagation delays of direct and
        // diffracted source, set diffracted to direct source distance
        auto const r_direct =
            (item->position.location - render_list.listener.position.location).abs();

        auto const r_diffracted_primary = dfr_location_rel_primary[2];
        dfr_location_rel_primary[2] = r_direct;

        // Get diffracted source position ...
        auto const dfr_location_xyz_primary =
            Voxel::vox_to_opengl(get_diffracted_source_location(dfr_location_rel_primary));

        // ... and update in item-store
        auto diffracted_item_primary =
            _diffracted_item_store->update_item(item, 0, dfr_location_xyz_primary);

        // Add EQ for primary diffracted source
        add_diffracted_source_eq(*diffracted_item_primary,
            get_diffraction_angle(r_direct, r_diffracted_primary),
            primary_source_weighting * r_direct / r_diffracted_primary);

        if (sec_path_info.has_value()) {
            auto& dfr_location_rel_secondary = sec_path_info.value().first;
            auto const secondary_source_weighting = sec_path_info.value().second;

            auto const r_diffracted_secondary = dfr_location_rel_secondary[2];
            dfr_location_rel_secondary[2] = r_direct;

            // Get diffracted source position ...
            auto const dfr_location_xyz_secondary =
                Voxel::vox_to_opengl(get_diffracted_source_location(dfr_location_rel_secondary));

            // ... and update in item-store
            auto diffracted_item_secondary =
                _diffracted_item_store->update_item(item, 1, dfr_location_xyz_secondary);

            // Add EQ for secondary diffracted source
            add_diffracted_source_eq(*diffracted_item_secondary,
                get_diffraction_angle(r_direct, r_diffracted_secondary),
                secondary_source_weighting * r_direct / r_diffracted_secondary);
        } else {
            // mute secondary diffracted source
            _diffracted_item_store->mute(item->idx, 1);
        }
    }

    // Add diffracted items to render-list
    _diffracted_item_store->add_items(render_list);
    return;
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::init(Voxel::VoxelSceneState const& voxel_scene_state)
{
    _scene = voxel_scene_state.get_scene_info_basic();

    _projection_map =
        std::make_shared<Voxel::projection_map>(voxel_scene_state.get_diffraction_map());
    build_corner_map();
    build_single_voxel_occluder_maps();
    init_pathfinder();
    _pathfinder_cache = voxel_scene_state.get_pathfinder_cache();

    Config::set("VoxelDiffractionFilterTuneParameter",
        voxel_scene_state.get_current_subscene()->diffraction_effect_control);
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::reset()
{
    _projection_map.reset();
    _corner_map.reset();
    _single_voxel_occluder_map.reset();
    _single_voxel_occluder_neighbour_map.reset();
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::build_corner_map()
{
    _corner_map = std::make_shared<Voxel::projection_map>(make_corner_map(*_projection_map));
    EAR_CHECK(_corner_map->valid(), "VoxelDiffractionStage: Invalid corner map");
#ifdef PM_DEBUG
    dump_pm(*_corner_map, "/tmp/cm.txt");
#endif
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::build_single_voxel_occluder_maps()
{
    _single_voxel_occluder_map =
        std::make_shared<Voxel::projection_map>(Voxel::projection_map(*_projection_map));

    for (int x = 0; x < _single_voxel_occluder_map->dim[0]; ++x) {
        for (int y = 0; y < _single_voxel_occluder_map->dim[1]; ++y) {
            _single_voxel_occluder_map->at(x, y) = 0;
        }
    }

    for (int x = 1; x < _single_voxel_occluder_map->dim[0] - 1; ++x) {
        for (int y = 1; y < _single_voxel_occluder_map->dim[1] - 1; ++y) {
            char mask[3][3] = { { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 0 } };

            bool flag = true;
            for (int xx = 0; xx < 3; xx++) {
                for (int yy = 0; yy < 3; yy++) {
                    if (_projection_map->at(x + xx - 1, y + yy - 1) != mask[xx][yy])
                        flag = false;
                }
            }
            if (flag)
                _single_voxel_occluder_map->at(x, y) = 1;
        }
    }
    EAR_CHECK(_single_voxel_occluder_map->valid(),
        "VoxelDiffractionStage: Invalid single voxel occluder map");
#ifdef PM_DEBUG
    dump_pm(*_single_voxel_occluder_map, "/tmp/single_voxel_occluder_map.txt");
#endif

    _single_voxel_occluder_neighbour_map =
        std::make_shared<Voxel::projection_map>(Voxel::projection_map(*_projection_map));

    for (int x = 0; x < _single_voxel_occluder_neighbour_map->dim[0]; ++x) {
        for (int y = 0; y < _single_voxel_occluder_neighbour_map->dim[1]; ++y) {
            _single_voxel_occluder_neighbour_map->at(x, y) = 0;
        }
    }

    for (int x = 1; x < _single_voxel_occluder_neighbour_map->dim[0] - 1; ++x) {
        for (int y = 1; y < _single_voxel_occluder_neighbour_map->dim[1] - 1; ++y) {
            bool flag = false;
            for (int xx = 0; xx < 3; xx++) {
                for (int yy = 0; yy < 3; yy++) {
                    if (_single_voxel_occluder_map->at(x + xx - 1, y + yy - 1) != 0)
                        flag = true;
                }
            }
            if (flag)
                _single_voxel_occluder_neighbour_map->at(x, y) = 1;
        }
    }

    EAR_CHECK(_single_voxel_occluder_neighbour_map->valid(),
        "VoxelDiffractionStage: Invalid single voxel occluder neighbour map");
#ifdef PM_DEBUG
    dump_pm(*_single_voxel_occluder_neighbour_map, "/tmp/single_voxel_occluder_neighbour_map.txt");
#endif
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::init_pathfinder()
{
    /// @todo For now only one pathfinder-instance
    auto const& pm = *_projection_map;
    _pathfinder->init(pm, Voxel::i_pathfinder::smoothing_enabled);
}

//--------------------------------------------------------------------------------------------------
float VoxelDiffractionJps::estimate_path_length() const
{
    return _pathfinder->get_path_length();
}

//--------------------------------------------------------------------------------------------------
Voxel::dmap_index_t VoxelDiffractionJps::find_corner()
{
    std::lock_guard<std::mutex> lck(_update_mutex);
    auto const& path_vec = _pathfinder->get_path().get();
    auto const& pm = *_corner_map;

    Voxel::dmap_index_t corner;
    CornerFindingAlgorithm(path_vec.size(), path_vec.data(), pm, corner.data());

    ++corner[0];
    ++corner[1];
    return corner;
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionJps::add_diffracted_source_eq(
    RenderItem& item, int diffraction_angle_deg, float fac) const
{
    std::vector<float> eq(Config::getInt("EQNBands"), 1.f);

    // Apply UDT
    apply_udt(eq, diffraction_angle_deg);

    // Add compensation calculated by the VoxelOcclusionStage, if any
    apply_compensation_eq(eq, item.parentRI);

    item.setEq(eq);
    // Compensate for additional distance of diffracted source
    item.addGain(std::clamp(fac, 0.f, 1.f));
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------