/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "IVoxelDiffraction.h"
#include "MPEGI/Helpers/IPathfinder.h"
#include "MPEGI/Renderer/RenderStage.h"
#include <mutex>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
/**
 * JPS-based voxel-diffraction
 */
class VoxelDiffractionJps : public IVoxelDiffraction, public MPEGI::ConfigObserver
{
    //@{
    /// Implementation of IVoxelDiffraction Interface
    void process(RenderList& render_list) override;
    //@}

public:
    VoxelDiffractionJps(
        Bitstream const& bitstream, RenderItemStore& store, SceneState const& state);
    ~VoxelDiffractionJps();
    VoxelDiffractionJps(VoxelDiffractionJps const&) = delete;
    VoxelDiffractionJps& operator=(VoxelDiffractionJps const&) = delete;

protected:
    /**
     * Subscene dependent initialization
     *
     * @param voxel_scene_state The voxel-scene state
     */
    void init(Voxel::VoxelSceneState const& voxel_scene_state);

    /**
     * Reset states, e.g. diffraction/corner maps
     */
    void reset();

    /**
     * Construct a corner-map for a projection map
     */
    void build_corner_map();

    /**
     * Construct single voxel occluder maps for a projection map
     */
    void build_single_voxel_occluder_maps();

    /**
     * Initializes the path finding algorithm
     */
    void init_pathfinder();

    /**
     * Estimate the lenght of the calculated path
     *
     * @return float Length of the path
     */
    float estimate_path_length() const;

    /**
     * Find the corner nearest to the listener on the calculated path
     *
     * @return Index (x/y) in the range [1..dim]
     */
    Voxel::dmap_index_t find_corner();

    /**
     * Add diffracted source EQ
     *
     * @param item The diffracted renderer item
     * @param diffraction_angle_deg Estimated diffraction angle in degrees [1,180]
     * @param fac Control factor (0..1) for the application of UDT and gain compensation
     */
    virtual void add_diffracted_source_eq(
        RenderItem& item, int diffraction_angle_deg, float fac) const;

    /// Secondary path info: secondary-item location, secondary source weight
    using sec_path_info_t = std::optional<std::pair<std::array<float, 3>, float>>;

    /**
     * Find secodary path to object
     *
     * @param user_dmap_idx User location on projection map
     * @param item_dmap_idx Object location on projection map
     * @param corner_primary Corner of primary path
     * @param dfr_location_rel_primary Relative primary diffracted source location (az,el,r)
     * @param azimuth_threshold Azimuth threshold for secondary path generation
     * @return Path info, std::nullopt if no secondary path was found
     */
    sec_path_info_t find_secondary_path(Voxel::dmap_index_t const& user_dmap_idx,
        Voxel::dmap_index_t const& item_dmap_idx, Voxel::dmap_index_t const& corner_primary,
        std::array<float, 3> const& dfr_location_rel_primary, float azimuth_threshold);

    /**
     * Populate pathfinder cache from the user position to the first corner
     *
     * @param user_dmap_idx User location on projection map
     * @param item_dmap_idx Object location on projection map
     * @param corner_dmap_idx_primary Corner of primary path on projection map
     * @param path_length_primary Lenght of the primary path on projection map (lateral + diagonal steps)
     */
    void populate_cache(Voxel::dmap_index_t const& user_dmap_idx,
        Voxel::dmap_index_t const& item_dmap_idx,
        Voxel::dmap_index_t const& corner_dmap_idx_primary, float path_length_primary);

    /**
     * Implements MPEGI::ConfigObserver
     */
    void configValueChanged(const std::string& key) override
    {
        static std::unordered_map<std::string,
            std::function<void(std::string const&, VoxelDiffractionJps&)>> const observed = {
            { "VoxelStorePreComputedPathDataByRequest",
                [](std::string const& key, VoxelDiffractionJps& jps) {
                    Voxel::store_precomputed_path_data(
                        *jps._pathfinder_cache, *jps._scene, *jps._projection_map);
                } }
        };

        auto const observed_it = observed.find(key);
        if (observed_it != std::cend(observed)) {
            observed_it->second(key, *this);
        }
    }

    /// Pathfinder handle
    std::unique_ptr<Voxel::i_pathfinder> _pathfinder;

    /// Projection map - either derived form the grid or transmitted in the bitstream
    std::shared_ptr<Voxel::projection_map> _projection_map;

    /// Corner map derived from corresponding projection map
    std::shared_ptr<Voxel::projection_map> _corner_map;

    /// Single voxel occluder maps derived from corresponding projection map
    std::shared_ptr<Voxel::projection_map> _single_voxel_occluder_map;
    std::shared_ptr<Voxel::projection_map> _single_voxel_occluder_neighbour_map;

    /// Pathfinder caching
    Voxel::pathfinder_cache::shared _pathfinder_cache;

    // Mutex to synchronize updates
    std::mutex _update_mutex;
}; // namespace MPEGI

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------