/***************************************************************************
This software module was originally developed by

Fraunhofer IIS

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Fraunhofer IIS retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "VoxelDiffractionRasterMap.h"

//--------------------------------------------------------------------------------------------------
namespace {
//--------------------------------------------------------------------------------------------------
/**
 * Get absolute diffracted source location (in OpenGL axis convention)
 *
 * @param src_aer Relative diffracted source location (az,el,r)
 * @param user_loc User location in OpenGL coordinates
 * @return The diffracted source location (x,y,z)
 */
std::array<float, 3> get_diffracted_source_location_opengl(
    std::array<float, 3> const& src_aer, const EAR::Vector3& user_loc)
{
    auto src_loc = MPEGI::Voxel::vox_to_opengl(MPEGI::sph2cart(src_aer));
    src_loc[0] += user_loc[0];
    src_loc[1] += user_loc[1];
    src_loc[2] += user_loc[2];

    return src_loc;
}

//--------------------------------------------------------------------------------------------------
} // namespace
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
VoxelDiffractionRasterMap::VoxelDiffractionRasterMap(
    Bitstream const& bitstream, RenderItemStore& store, SceneState const& state)
    : IVoxelDiffraction(state.getVoxelSceneState(), store)
{
    Config::setDef("VoxelDiffractionNormalizeDelay", EAR::Variant(true));
    Config::setDef("VoxelDiffractionEqStrength", EAR::Variant(.75f));

    auto pay = std::dynamic_pointer_cast<RasterMapPayload const>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_RASTERMAP));

    auto const vox_state = state.getVoxelSceneState();

    auto const num_subscenes = pay->getSubsceneData().size();
    _cache.reserve(num_subscenes);

    for (auto s = 0U; s < num_subscenes; ++s) {
        auto const scene = vox_state->get_scene_info_basic(s);
        EAR_CHECK_REL(scene != nullptr, "Subscene index out-of-range");

        auto const dmap_dim = Voxel::dmap_index_t{ scene->dim[0], scene->dim[1] };
        _cache.push_back(
            { std::make_shared<cache_t>(dmap_dim), std::make_shared<cache_t>(dmap_dim) });
    }

    // Fill caches
    transfer_rastermap(pay);

    // Register callbacks
    vox_state->on_subscene_change(
        [this](Voxel::VoxelSceneState::const_shared const& voxel_scene_state) {
            this->_scene = voxel_scene_state->get_scene_info_basic();
        });
}
//--------------------------------------------------------------------------------------------------

void VoxelDiffractionRasterMap::process(RenderList& render_list)
{
    // level normalization between pressure and power normalization:
    // -4.5dB for 2 active diffraction sources
    const float power_exp = float(4. / 3.);

    /// @todo Always activate for raster map, or should it stay configurable (VoxelDiffractionContinousListenerPosition)?
    auto const enable_continuous_listener_pos = true;

    for (auto& item : render_list.items) {

        if (!diffraction_supported_for_item(*item)) {
            continue;
        }

        // If the primary item is orphaned (or fades-out to orphaned status), trigger removal of
        // all secondary (diffracted) items. Diffracted items are not actually removed here, but only
        // marked for removal
        if (item->isOrphaned() || item->isFadeOutOrphaned()) {
            for (int layer = 0; layer < MAX_VOX_MULTI_PATH; layer++) {
                _diffracted_item_store->remove(item->idx, layer);
            }
        }

        // If the primary item is inactive or visible, mute all secondary (diffracted) items
        if (_dv_map->direct_view(item->idx) || item->isInactive()) {
            for (int layer = 0; layer < MAX_VOX_MULTI_PATH; layer++) {
                _diffracted_item_store->mute(item->idx, layer);
            }
            continue;
        }

        auto const user_voxel_idx = _index_storage->get_listener_voxel_index();
        auto const item_voxel_idx = _index_storage->get_voxel_index(item->idx);
        auto const additional_distance = _index_storage->get_additional_distance(item->idx);
        auto const idx = item->idx;

        std::array<float, 3> user_idx_float;
        EAR::Vector3 user_v3;
        EAR::Vector3 user_mid_v3;
        if (enable_continuous_listener_pos) {
            auto user_loc = render_list.listener.position.location;
            auto user_xyz = Voxel::opengl_to_vox(user_loc[0], user_loc[1], user_loc[2]);
            user_idx_float = xyz2idx_float(user_xyz);
            user_v3 = { user_idx_float[0], user_idx_float[1], user_idx_float[2] };
            user_mid_v3 = { float(user_voxel_idx[0]), float(user_voxel_idx[1]), user_idx_float[2] };
        }

        Voxel::dmap_index_t const user_dmap_idx{ user_voxel_idx[0], user_voxel_idx[1] };
        Voxel::dmap_index_t const item_dmap_idx{ item_voxel_idx[0], item_voxel_idx[1] };
        std::array<float, 3> dfr_location_rel;

        std::array<RenderItem*, MAX_VOX_MULTI_PATH> diffr_item_list;
        std::array<float, MAX_VOX_MULTI_PATH> diffr_gain;
        std::array<float, MAX_VOX_MULTI_PATH> diffr_dist;
        int num_layer = 0;
        auto& cache = get_cache();
        for (int layer = 0; layer < MAX_VOX_MULTI_PATH; layer++) {
            // layer specific gain values
            diffr_item_list[layer] = nullptr;
            diffr_gain[layer] = 0.f;

            // try to lookup
            /// @todo Revise caching
            auto path_info_opt = cache[layer]->get(user_dmap_idx, item_dmap_idx);

            if (path_info_opt.has_value()) {
                auto const& path_info = path_info_opt.value();

                if (!path_info.length.has_value()) {
                    // No path from user to item
                    _diffracted_item_store->mute(idx, layer);
                    continue;
                }

                // avoid path length discontinuities
                float length_offset = 0.f;
                if (enable_continuous_listener_pos) {
                    EAR::Vector3 corner_v3 = { float(path_info.corner[0]),
                        float(path_info.corner[1]), user_idx_float[2] };
                    length_offset = (corner_v3 - user_v3).abs() - (corner_v3 - user_mid_v3).abs();
                }

                dfr_location_rel = get_diffracted_source_aer_relative(user_voxel_idx,
                    item_voxel_idx, path_info.length.value() + length_offset, additional_distance,
                    path_info.corner);

                num_layer = layer + 1;

            } else {
                _diffracted_item_store->mute(idx, layer);
                continue;
            }

            // To avoid discontinuities caused by different propagation delays of direct and
            // diffracted source, set diffracted to direct source distance
            float r_direct;
            if (enable_continuous_listener_pos) {
                auto src_mid_gl = Voxel::vox_to_opengl(_scene->idx2xyz(item_voxel_idx));
                EAR::Vector3 src_mid_loc = { src_mid_gl[0], src_mid_gl[1], src_mid_gl[2] };
                r_direct = (src_mid_loc - render_list.listener.position.location).abs();
            } else {
                r_direct = (item->position.location - render_list.listener.position.location).abs();
            }
            auto const r_diffracted = dfr_location_rel[2];
            auto const normalize_delay = Config::getBool("VoxelDiffractionNormalizeDelay");
            if (normalize_delay) {
                dfr_location_rel[2] = r_direct;
                diffr_gain[layer] = r_direct / r_diffracted;
            } else {
                diffr_gain[layer] = 1.f;
            }

            // Get diffracted source position ...
            const auto& user_loc = render_list.listener.position.location;
            auto const dfr_location_xyz =
                get_diffracted_source_location_opengl(dfr_location_rel, user_loc);

            // ... and update in item-store
            auto diffracted_item =
                _diffracted_item_store->update_item(item, layer, dfr_location_xyz);
            diffr_item_list[layer] = diffracted_item;

            // store the diffraction path distance for this layer
            diffr_dist[layer] = r_diffracted;

            // Add EQ for diffracted source
            add_diffracted_source_eq(
                *diffracted_item, get_diffraction_angle(r_direct, r_diffracted), 1.f);

            //FIXME: remove this information for visualization
            Vector3 src = { dfr_location_xyz[0], dfr_location_xyz[1], dfr_location_xyz[2] };
            const Vector3& lis = render_list.listener.position.location;
            diffracted_item->reflectionInfo = std::make_shared<ReflectionInfo>();
            diffracted_item->reflectionInfo->intersectionSequence = { src, lis };
        }

        float ref_level = powf(diffr_gain[0], power_exp);

        // Apply multi-path cross-fading
        const auto& user_loc = render_list.listener.position.location;
        if (!apply_multipass_crossfade(
                user_loc, user_dmap_idx, item_dmap_idx, num_layer, diffr_gain)) {
            // perform fallback method, if multi-pass crossfading fails
            constexpr float fadeRange = 2.0;
            for (int layer = 1; layer < num_layer; layer++) {
                float diff = diffr_dist[layer] - diffr_dist[0];
                float fade_gain = 1.f - diff / fadeRange;
                if (fade_gain < 0.f) {
                    fade_gain = 0.f;
                }
                diffr_gain[layer] *= fade_gain;
            }
        }

        // level normalization
        float sum_level = ref_level;
        for (int layer = 1; layer < num_layer; layer++) {
            sum_level += powf(diffr_gain[layer], power_exp);
        }
        float norm_gain = powf(ref_level / sum_level, 1.f / power_exp);
        for (int layer = 0; layer < MAX_VOX_MULTI_PATH; layer++) {
            if (diffr_item_list[layer]) {
                diffr_item_list[layer]->addGain(norm_gain * diffr_gain[layer]);
            }
        }
    }
    // Add diffracted items to render-list
    _diffracted_item_store->add_items(render_list);
}

//--------------------------------------------------------------------------------------------------
bool VoxelDiffractionRasterMap::apply_multipass_crossfade(const Vector3& user_loc,
    const Voxel::dmap_index_t& user_dmap_idx, const Voxel::dmap_index_t& item_dmap_idx,
    int num_layer, std::array<float, MAX_VOX_MULTI_PATH>& diffr_gain) const
{
    if (num_layer < 2) {
        // At least 2 layers are needed for cross-fading
        return false;
    }

    // get multi-layer diffraction paths
    using seq_t = std::vector<Voxel::dmap_index_t>;
    std::vector<seq_t> seq_list(num_layer);
    std::vector<int> tail_idx(num_layer);
    for (int layer = 0; layer < num_layer; layer++) {
        seq_list[layer] = get_corner_seq(layer, user_dmap_idx, item_dmap_idx);

        // check for incomplete paths
        size_t len = seq_list[layer].size();
        if (len < 2) {
            return false;
        }

        // set index
        tail_idx[layer] = len - 1;
    }

    // find common tail
    bool has_more = true;
    bool equal_prev = true;
    bool equal_next = true;
    for (int layer = 0; layer < num_layer; layer++) {
        has_more &= tail_idx[layer] > 1;
    }
    for (int layer = 1; layer < num_layer; layer++) {
        int prev0 = tail_idx[0];
        int next0 = prev0 - 1;
        int prev = tail_idx[layer];
        int next = prev - 1;
        equal_prev &= seq_list[layer][prev] == seq_list[0][prev0];
        equal_next &= seq_list[layer][next] == seq_list[0][next0];
    }
    while (has_more && equal_prev && equal_next) {
        for (int layer = 0; layer < num_layer; layer++) {
            tail_idx[layer]--;
            has_more &= tail_idx[layer] > 1;
        }
        for (int layer = 1; layer < num_layer; layer++) {
            int prev0 = tail_idx[0];
            int next0 = prev0 - 1;
            int prev = tail_idx[layer];
            int next = prev - 1;
            equal_prev &= seq_list[layer][prev] == seq_list[0][prev0];
            equal_next &= seq_list[layer][next] == seq_list[0][next0];
        }
    }

    // all paths must be convex
    for (int layer = 0; layer < num_layer; layer++) {
        if (!path_is_convex(seq_list[layer], tail_idx[layer])) {
            return false;
        }
    }

    // Continuous voxel coordinate of the user
    auto user_xyz = Voxel::opengl_to_vox(user_loc[0], user_loc[1], user_loc[2]);
    auto user_idx_float = xyz2idx_float(user_xyz);

    std::vector<float> dist(num_layer);
    for (int layer = 0; layer < num_layer; layer++) {
        int prev = tail_idx[layer];
        int next = prev - 1;

        // we need signed int here
        auto& seq = seq_list[layer];
        std::array<int, 2> corner_0 = { int(seq[prev][0]), int(seq[prev][1]) };
        std::array<int, 2> corner_1 = { int(seq[next][0]), int(seq[next][1]) };

        // compute user <-> plane distance
        int dx = corner_1[0] - corner_0[0];
        int dy = corner_1[1] - corner_0[1];
        float len = sqrtf(dx * dx + dy * dy);
        float plane_Nx = -dy / len;
        float plane_Ny = dx / len;
        float plane_d = plane_Nx * corner_1[0] + plane_Ny * corner_1[1];
        dist[layer] = fabsf(plane_Nx * user_idx_float[0] + plane_Ny * user_idx_float[1] - plane_d);
        if (dist[layer] < 1.f) {
            dist[layer] = 1.f;
        }
    }

    // determine crossfade gains
    std::vector<float> fade_gain(num_layer);
    float gain_sum = 0.f;
    float dist_prod = 1.f;
    for (float d : dist) {
        dist_prod *= d;
    }
    for (int layer = 0; layer < num_layer; layer++) {
        float val = dist_prod / dist[layer];
        fade_gain[layer] = val;
        gain_sum += val;
    }
    for (float& gain : fade_gain) {
        gain /= gain_sum;
    }

    // apply crossfade gains
    for (int layer = 0; layer < num_layer; layer++) {
        diffr_gain[layer] *= fade_gain[layer];
    }

    return true;
}

//--------------------------------------------------------------------------------------------------
bool VoxelDiffractionRasterMap::path_is_convex(
    const std::vector<Voxel::dmap_index_t>& seq, int tail_idx) const
{
    if (tail_idx < 3) {
        return true;
    }

    int dx1 = int(seq[1][0]) - int(seq[2][0]);
    int dy1 = int(seq[1][1]) - int(seq[2][1]);
    int dx2 = int(seq[0][0]) - int(seq[1][0]);
    int dy2 = int(seq[0][1]) - int(seq[1][1]);
    int ref_val = dy1 * dx2 - dx1 * dy2;
    int ref_sign = (ref_val > 0) - (ref_val < 0);

    for (int n = 3; n <= tail_idx; n++) {
        dx1 = int(seq[n - 1][0]) - int(seq[n][0]);
        dy1 = int(seq[n - 1][1]) - int(seq[n][1]);
        dx2 = int(seq[n - 2][0]) - int(seq[n - 1][0]);
        dy2 = int(seq[n - 2][1]) - int(seq[n - 1][1]);
        int test_val = dy1 * dx2 - dx1 * dy2;
        int test_sign = (test_val > 0) - (test_val < 0);
        if ((test_sign != 0) && (test_sign != ref_sign)) {
            return false;
        }
    }

    return true;
}

//--------------------------------------------------------------------------------------------------
std::vector<dmap_index_t> VoxelDiffractionRasterMap::get_corner_seq(
    int layer, dmap_index_t const& user, dmap_index_t const& item) const
{
    std::vector<dmap_index_t> seq{ user };
    auto corner = user;
    auto const& cache = get_cache();

    while (true) {
        auto info = cache[layer]->get(corner, item);
        if (info.has_value()) {
            // only valid, if it has a length specification
            if (!info.value().length.has_value()) {
                break;
            }

            corner = info.value().corner;
            layer = info.value().corner_layer;
            seq.push_back(corner);

            // check if we reached the source item
            if (layer == 0 && corner == item) {
                break;
            }
        } else {
            seq.push_back(item);
            break;
        }
    }

    return seq;
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionRasterMap::add_diffracted_source_eq(
    RenderItem& item, int diffraction_angle_deg, float fac) const
{
    std::vector<float> eq(Config::getInt("EQNBands"), 1.f);

    // Apply UDT
    apply_udt(eq, diffraction_angle_deg,
        [eq_strength = Config::getFloat("VoxelDiffractionEqStrength")](
            float udt) { return std::pow(udt, eq_strength); });

    // Add compensation calculated by the VoxelOcclusionStage, if any
    apply_compensation_eq(eq, item.parentRI);

    item.setEq(eq);
    // Compensate for additional distance of diffracted source
    item.addGain(std::clamp(fac, 0.f, 1.f));
}

//--------------------------------------------------------------------------------------------------
void VoxelDiffractionRasterMap::transfer_rastermap(
    std::shared_ptr<RasterMapPayload const> rastermap_payload)
{
    if (rastermap_payload == nullptr) {
        return;
    }

    const std::vector<MPEGI::SubsceneData>& rmSubsceneList = rastermap_payload->getSubsceneData();

    for (size_t s = 0; s < rmSubsceneList.size(); s++) {
        // determine voxel size
        const MPEGI::SubsceneData& rmSubscene = rmSubsceneList.at(s);

        auto& cache = _cache[s];
        // clear cache layers
        for (auto& c : cache) {
            c->reset();
        }

        for (const auto& coordMap : rmSubscene.availableCoords) {
            int src_id = coordMap.first;
            const MPEGI::WaypointMatrix& wpMatrix = rmSubscene.getWaypointMatrix(src_id);
            for (const auto& vc : coordMap.second) {
                // get diffraction path
                std::vector<MPEGI::RasterMapCoord> path = wpMatrix.getPath(vc);
                if (path.size() == 0) {
                    continue;
                }

                // voxel coordinates for cache->set_path(); +1 for Matlab-style indices
                const auto& vcSrc = path.back();
                const auto& vcCorner = path.front();
                const MPEGI::Voxel::dmap_index_t item_dmap_idx = { (unsigned int)vcSrc[0] + 1,
                    (unsigned int)vcSrc[1] + 1 };
                const MPEGI::Voxel::dmap_index_t user_dmap_idx = { (unsigned int)vc[0] + 1,
                    (unsigned int)vc[1] + 1 };
                const MPEGI::Voxel::dmap_index_t corner = { (unsigned int)vcCorner[0] + 1,
                    (unsigned int)vcCorner[1] + 1 };

                // compute path length
                float path_length = wpMatrix.getPathLength(vc, path);

                // RasterMap layer index
                int layer = vc[2];

                // add path to cache
                cache[layer]->set_path(user_dmap_idx, item_dmap_idx,
                    raster_map_info{ path_length, corner, vcCorner[2] });
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
