/***************************************************************************
This software module was originally developed by

Fraunhofer IIS
Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "IVoxelDiffraction.h"
#include "MPEGI/Bitstream/RasterMapPayload.h"
#include "MPEGI/Renderer/RenderStage.h"

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
using Voxel::dmap_index_t;

/**
 * Raster-map based voxel-diffraction
 */
class VoxelDiffractionRasterMap : public IVoxelDiffraction
{
    static unsigned int const MAX_VOX_MULTI_PATH = 2;

    //@{
    /// Implementation of IVoxelDiffraction Interface
    void process(RenderList& render_list) override;
    //@}

public:
    /**
     * Constructor
     * @param bitstream Bitstream structure to retrieve payloads
     * @param store Render item store for secondary item storage in base class
     * @param state Scene state passed to the base class
     */
    VoxelDiffractionRasterMap(
        Bitstream const& bitstream, RenderItemStore& store, SceneState const& state);

protected:
    /**
     * Map position to continuous voxel indices (unsafe)
     * @param pos_xyz Position x,y,z coordinates
     * @return Continuous voxel indices
     */
    std::array<float, 3> xyz2idx_float(std::array<float, 3> const& pos_xyz) const
    {
        auto const& scene = *_scene;
        return { (pos_xyz[0] - scene.p0[0]) / scene.voxel_size + .5f,
            (pos_xyz[1] - scene.p0[1]) / scene.voxel_size + .5f,
            (pos_xyz[2] - scene.p0[2]) / scene.voxel_size + .5f };
    }

    /**
     * Apply multi-path cross-fading
     *
     * @param user_loc Location of the user (input)
     * @param user_dmap_idx Voxel coordinate of the user (input)
     * @param item_dmap_idx Voxel coordinate of the item (input)
     * @param num_layer Number of diffraction path layers (input)
     * @param diffr_gain List of linear gains (input/output)
     * @return Returns true if successful
     */
    bool apply_multipass_crossfade(const Vector3& user_loc,
        const Voxel::dmap_index_t& user_dmap_idx, const Voxel::dmap_index_t& item_dmap_idx,
        int num_layer, std::array<float, MAX_VOX_MULTI_PATH>& diffr_gain) const;

    /**
     * Check if a diffraction path is convex
     *
     * @param seq Diffraction path (corner sequence)
     * @param tail_idx Index of the last corner
     * @return Returns true if the given diffraction path is convex
     */
    bool path_is_convex(const std::vector<Voxel::dmap_index_t>& seq, int tail_idx) const;

    /**
     * Get corner sequence (complete diffraction path)
     *
     * @param layer Diffraction map layer
     * @param user User diffraction map index (x,y)
     * @param item Item diffraction map index (x,y)
     * @return Returns a vector of diffraction map indices (x,y)
     */
    std::vector<dmap_index_t> get_corner_seq(
        int layer, dmap_index_t const& user, dmap_index_t const& item) const;

    /**
     * Add diffracted source EQ
     *
     * @param item The diffracted renderer item
     * @param diffraction_angle_deg Estimated diffraction angle in degrees [1,180]
     * @param fac Control factor (0..1) for the application of UDT and gain compensation
     */
    void add_diffracted_source_eq(RenderItem& item, int diffraction_angle_deg, float fac) const;

    /**
     * Transfer RasterMap data into caches
     *
     * @param rastermap_payload RasterMap payload
     */
    void transfer_rastermap(std::shared_ptr<RasterMapPayload const> rastermap_payload);

protected:
    struct raster_map_info
    {
        raster_map_info() = default;
        raster_map_info(float length_, Voxel::dmap_index_t const& corner_, int corner_layer_)
            : length(length_)
            , corner(corner_)
            , corner_layer(corner_layer_)
        {}

        raster_map_info(raster_map_info const&) = default;
        raster_map_info(raster_map_info&&) = default;

        raster_map_info& operator=(raster_map_info const& other)
        {
            length = other.length;
            corner = other.corner;
            corner_layer = other.corner_layer;
            return *this;
        }

        raster_map_info& operator=(raster_map_info&& other)
        {
            length = std::move(other.length);
            corner = std::move(other.corner);
            corner_layer = std::move(other.corner_layer);
            return *this;
        }

        void reset() { length.reset(); }

        std::optional<float> length{ 0.f };
        Voxel::dmap_index_t corner{ 0 };
        int corner_layer{ 0 };
    };

    /// Allocate cache per subscene
    using cache_t = Voxel::cache<raster_map_info>;
    using layered_cache_t = std::array<cache_t::shared, MAX_VOX_MULTI_PATH>;
    std::vector<layered_cache_t> _cache;

    /// Get the cache for the current subscene (const)
    layered_cache_t const& get_cache() const
    {
        return _scene->id.has_value() ? _cache[_scene->id.value()] : _cache[0];
    }

    /// Get the cache for the current subscene (non-const)
    layered_cache_t& get_cache()
    {
        return const_cast<layered_cache_t&>(
            static_cast<VoxelDiffractionRasterMap const*>(this)->get_cache());
    }
};

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
