/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "jps_pathfinder_impl.h"
#include <cmath>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {
//--------------------------------------------------------------------------------------------------
jps_pathfinder::~jps_pathfinder() {}

//--------------------------------------------------------------------------------------------------
jps_pathfinder::jps_pathfinder()
    : m_grid_width(0)
    , m_grid_height(0)
{}

//--------------------------------------------------------------------------------------------------
int jps_pathfinder::init(projection_map const& p_map, bool smooth_path)
{

    m_projection_map = &p_map;
    m_smooth_path = smooth_path;

    if ((m_projection_map->dim[0] != m_grid_width) || (m_projection_map->dim[1] != m_grid_height)) {
        m_grid_width = p_map.dim[0];
        m_grid_height = p_map.dim[1];

        m_jps_grid_nodes.resize(m_grid_height, std::valarray<jps_node>(m_grid_width));
        for (int y = 0; y < m_grid_height; y++) {
            for (int x = 0; x < m_grid_width; x++) {
                m_jps_grid_nodes[y][x] = { x, y, 0, 0, 0, false, false, nullptr };
            }
        }
    }

    return pathfinder_ok;
}

//--------------------------------------------------------------------------------------------------
int jps_pathfinder::find_path(voxel_index_t const& from, voxel_index_t const& to)
{
    _path.get().clear();

    // Convert voxel - indices[1..dim] to JPS - indices[0..dim - 1]
    // Note: start end end are exchanged internally, but path will be returned
    // from start to end

    int startX = to[0] - 1;
    int startY = to[1] - 1;
    int endX = from[0] - 1;
    int endY = from[1] - 1;

    bool found = findPath(startX, startY, endX, endY, _path.get(), _path_length);

    return (found ? pathfinder_ok : pathfinder_no_path);
}

//--------------------------------------------------------------------------------------------------
int jps_pathfinder::test_direct_view(voxel_index_t const& from, voxel_index_t const& to)
{
    int x0, y0, x1, y1;
    bool direct_view;

    // Convert voxel - indices[1..dim] to JPS - indices[0..dim - 1]

    x0 = from[0] - 1;
    y0 = from[1] - 1;
    x1 = to[0] - 1;
    y1 = to[1] - 1;

    std::vector<std::pair<int, int>> line;
    interpolate_line(x0, y0, x1, y1, line);

    direct_view = line.size() != 0;

    return (direct_view ? pathfinder_ok : pathfinder_no_direct_view);
}

//--------------------------------------------------------------------------------------------------
Voxel::path_2d const& jps_pathfinder::get_path() const
{
    return _path;
}

//--------------------------------------------------------------------------------------------------
float const& jps_pathfinder::get_path_length() const
{
    return _path_length;
}

jps_node* jps_pathfinder::getNodeAt(int x, int y)
{
    return &m_jps_grid_nodes[y][x];
}

bool inline jps_pathfinder::isWalkableAt(int x, int y)
{
    return (isInside(x, y) && m_projection_map->at(x, y) == 0);
}

bool inline jps_pathfinder::isInside(int x, int y)
{
    return (x >= 0 && x < m_grid_width) && (y >= 0 && y < m_grid_height);
}

void jps_pathfinder::findNeighbors(jps_node* activeNode)
{
    jps_node* parent = activeNode->parent;
    int x = activeNode->x;
    int y = activeNode->y;

    m_neighbors.clear();

    if (parent) {
        int px, py, dx, dy;
        px = parent->x;
        py = parent->y;

        dx = (x - px) / std::max(abs(x - px), 1);
        dy = (y - py) / std::max(abs(y - py), 1);

        /* Diagonals */
        if (dx != 0 && dy != 0) {
            if (isWalkableAt(x, (y + dy))) {
                m_neighbors.push_back(std::make_pair(x, y + dy));
            }
            if (isWalkableAt((x + dx), y)) {
                m_neighbors.push_back(std::make_pair(x + dx, y));
            }
            if (isWalkableAt(x, (y + dy)) || isWalkableAt((x + dx), y)) {
                m_neighbors.push_back(std::make_pair(x + dx, y + dy));
            }
            if (!isWalkableAt((x - dx), y) && isWalkableAt(x, (y + dy))) {
                m_neighbors.push_back(std::make_pair(x - dx, y + dy));
            }
            if (!isWalkableAt(x, (y - dy)) && isWalkableAt((x + dx), y)) {
                m_neighbors.push_back(std::make_pair(x + dx, y - dy));
            }

            /* Horizontal / Vertical */
        } else {
            if (dx == 0) {
                if (isWalkableAt(x, (y + dy))) {
                    if (isWalkableAt(x, (y + dy))) {
                        m_neighbors.push_back(std::make_pair(x, y + dy));
                    }
                    if (!isWalkableAt((x + 1), y)) {
                        m_neighbors.push_back(std::make_pair(x + 1, y + dy));
                    }
                    if (!isWalkableAt((x - 1), y)) {
                        m_neighbors.push_back(std::make_pair(x - 1, y + dy));
                    }
                }
            } else {
                if (isWalkableAt((x + dx), y)) {
                    if (isWalkableAt((x + dx), y)) {
                        m_neighbors.push_back(std::make_pair(x + dx, y));
                    }
                    if (!isWalkableAt(x, (y + 1))) {
                        m_neighbors.push_back(std::make_pair(x + dx, y + 1));
                    }
                    if (!isWalkableAt(x, (y - 1))) {
                        m_neighbors.push_back(std::make_pair(x + dx, y - 1));
                    }
                }
            }
        }
    } else {
        bool d0 = false;
        bool d1 = false;
        bool d2 = false;
        bool d3 = false;

        /* UP */
        if (isWalkableAt(x, y - 1)) {
            m_neighbors.push_back(std::make_pair(x, y - 1));
            d0 = d1 = true;
        }

        /* RIGHT */
        if (isWalkableAt(x + 1, y)) {
            m_neighbors.push_back(std::make_pair(x + 1, y));
            d1 = d2 = true;
        }

        /* DOWN */
        if (isWalkableAt(x, y + 1)) {
            m_neighbors.push_back(std::make_pair(x, y + 1));
            d2 = d3 = true;
        }

        /* LEFT */
        if (isWalkableAt(x - 1, y)) {
            m_neighbors.push_back(std::make_pair(x - 1, y));
            d3 = d0 = true;
        }

        /* UP + LEFT */
        if (d0 && isWalkableAt(x - 1, y - 1)) {
            m_neighbors.push_back(std::make_pair(x - 1, y - 1));
        }

        /* UP + RIGHT */
        if (d1 && isWalkableAt(x + 1, y - 1)) {
            m_neighbors.push_back(std::make_pair(x + 1, y - 1));
        }

        /* DOWN + RIGHT */
        if (d2 && isWalkableAt(x + 1, y + 1)) {
            m_neighbors.push_back(std::make_pair(x + 1, y + 1));
        }

        /* DOWN + LEFT */
        if (d3 && isWalkableAt(x - 1, y + 1)) {
            m_neighbors.push_back(std::make_pair(x - 1, y + 1));
        }
    }
}

int jps_pathfinder::euclidean(int dx, int dy)
{
    float distance = std::sqrt((float)(dx * dx + dy * dy)) * 10.0f;
    int idistance = (int)distance;

    return idistance;
}

int jps_pathfinder::manhattan(int dx, int dy)
{
    int distance = (dx + dy) * 10;

    return distance;
}

bool jps_pathfinder::jump(int& x, int& y, int px, int py, int endX, int endY)
{
    int dx = x - px;
    int dy = y - py;

    if (!isWalkableAt(x, y)) {
        return false;
    } else if (x == endX && y == endY) {
        return true;
    }

    // check for forced neighbors
    // along the diagonal
    if (dx != 0 && dy != 0) {
        if ((isWalkableAt((x - dx), (y + dy)) && !isWalkableAt((x - dx), y))
            || (isWalkableAt((x + dx), (y - dy)) && !isWalkableAt(x, (y - dy)))) {
            return true;
        }
        // when moving diagonally, must check for vertical/horizontal jump points
        int xpdx = x + dx;
        int ypdy = y + dy;
        if (jump(xpdx, y, x, y, endX, endY) || jump(x, ypdy, x, y, endX, endY)) {
            return true;
        }
    }
    // horizontally/vertically
    else {
        if (dx != 0) {
            if ((isWalkableAt((x + dx), (y + 1)) && !isWalkableAt(x, (y + 1)))
                || (isWalkableAt((x + dx), (y - 1)) && !isWalkableAt(x, (y - 1)))) {
                return true;
            }
        } else {
            if ((isWalkableAt((x + 1), (y + dy)) && !isWalkableAt((x + 1), y))
                || (isWalkableAt((x - 1), (y + dy)) && !isWalkableAt((x - 1), y))) {
                return true;
            }
        }
    }

    px = x;
    py = y;

    x = x + dx;
    y = y + dy;

    return jump(x, y, px, py, endX, endY);
}

void jps_pathfinder::identifySuccessors(jps_node* activeNode, int endX, int endY)
{
    findNeighbors(activeNode);

    for (auto neighbor : m_neighbors) {
        int jx, jy;
        bool jumpPoint;
        jx = neighbor.first;
        jy = neighbor.second;

        jumpPoint = jump(jx, jy, activeNode->x, activeNode->y, endX, endY);

        if (jumpPoint) {
            int d, ng;
            jps_node* jumpNode;

            jumpNode = getNodeAt(jx, jy);
            if (jumpNode->closed) {
                continue;
            }

            d = euclidean(abs(jx - activeNode->x), abs(jy - activeNode->y));
            ng = activeNode->g + d;
            if (!jumpNode->opened || ng < jumpNode->g) {
                jumpNode->g = ng;
                if (jumpNode->h == 0)
                    jumpNode->h = manhattan(abs(jx - endX), abs(jy - endY));
                jumpNode->f = jumpNode->g + jumpNode->h;

                jumpNode->parent = activeNode;

                if (!jumpNode->opened) {
                    jumpNode->opened = true;
                    m_opened_list.push_back(jumpNode);
                }

                /* keep open list sorted by f value in descending order */
                for (auto i = m_opened_list.size() - 1; i > 0; i--) {
                    if (m_opened_list[i]->f > m_opened_list[i - 1]->f) {
                        std::swap(m_opened_list[i], m_opened_list[i - 1]);
                    }
                }
            }
        }
    }
}

void jps_pathfinder::interpolate_line(
    int x0, int y0, int x1, int y1, std::vector<std::pair<int, int>>& line)
{
    // interpolate from start to end using Bresenham's algorithm

    line.clear();
    int sx, sy, dx, dy, err, e2;

    dx = abs(x1 - x0);
    dy = abs(y1 - y0);

    sx = (x0 < x1) ? 1 : -1;
    sy = (y0 < y1) ? 1 : -1;

    err = dx - dy;

    while (true) {

        if (isWalkableAt(x0, y0)) {
            line.push_back(std::make_pair(x0, y0));
        } else {
            line.clear();
            break;
        }

        if (x0 == x1 && y0 == y1) {
            break;
        }

        e2 = 2 * err;
        if (e2 > -dy) {
            err = err - dy;
            x0 = x0 + sx;
        }
        if (e2 < dx) {
            err = err + dx;
            y0 = y0 + sy;
        }
    }
}

void jps_pathfinder::backtrace(jps_node* activeNode, std::vector<int>& path, float& path_length)
{
    std::vector<std::pair<int, int>> path_points;

    path_points.push_back(std::make_pair(activeNode->x, activeNode->y));

    while ((activeNode = activeNode->parent) != NULL) {
        path_points.push_back(std::make_pair(activeNode->x, activeNode->y));
    }

    if (!m_smooth_path) {

        path_length = (float)activeNode->g * 0.1f;

        for (auto point : path_points) {
            path.push_back(point.first); // x
            path.push_back(point.second); // y
        }
    } else {
        std::vector<std::pair<int, int>> smooth_path;

        auto it = path_points.begin();
        auto end = path_points.end();

        std::vector<std::pair<int, int>> lastJump;

        smooth_path.push_back(*it); // add first point

        while (it != end) {
            auto next = it + 1;

            // Jump as far as we can

            while (next != end) {
                std::vector<std::pair<int, int>> line;
                interpolate_line(it->first, it->second, next->first, next->second, line);

                if (line.size() != 0) // we can jump
                {
                    lastJump.clear();
                    std::copy(line.begin() + 1, line.end(), std::back_inserter(lastJump));
                    next++;

                } else {
                    next--;
                    break;
                }
            }

            for (auto&& point : lastJump)
                smooth_path.push_back(std::move(point));

            it = next;
        }

        for (auto point : smooth_path) {
            path.push_back(point.first); // x
            path.push_back(point.second); // y
        }
        // calculate path length
        path_length = 0.0f;

        for (auto i = 0; i < path.size() - 2; i += 2) {
            path_length +=
                (path[i] != path[i + 2]) && (path[i + 1] != path[i + 3]) ? 1.41421356f : 1.0f;
        }
    }
}

void jps_pathfinder::cleanUpNodes(void)
{
    // open list
    for (auto node : m_opened_list) {
        if (node) {
            node->f = 0;
            node->g = 0;
            node->h = 0;
            node->opened = false;
            node->closed = false;
            node->parent = NULL;
        }
    }
    //closed list
    for (auto node : m_closed_list) {
        if (node) {
            node->f = 0;
            node->g = 0;
            node->h = 0;
            node->opened = false;
            node->closed = false;
            node->parent = NULL;
        }
    }
}

bool jps_pathfinder::findPath(
    int startX, int startY, int endX, int endY, std::vector<int>& path, float& path_length)
{
    jps_node* startNode = getNodeAt(startX, startY);
    jps_node* endNode = getNodeAt(endX, endY);
    jps_node* activeNode;
    int counter = 0;
    bool res = false;

    path.clear();
    m_opened_list.clear();
    m_closed_list.clear();

    /* check if start or end node is blocked */
    if (isWalkableAt(startNode->x, startNode->y) && isWalkableAt(endNode->x, endNode->y)) {

        startNode->opened = true;
        m_opened_list.push_back(startNode);

        while (m_opened_list.size()) { /* List is empty when no successors have been found */

            /* close active node, move to closed list */
            activeNode = m_opened_list.back();
            activeNode->closed = true;

            m_opened_list.pop_back();
            m_closed_list.push_back(activeNode);

            if (activeNode->x == endNode->x && activeNode->y == endNode->y) {
                /* path found */
                backtrace(activeNode, path, path_length);
                res = true;
                break;
            }

            /* identifying successors */
            identifySuccessors(activeNode, endX, endY);

            counter++;
            if (counter >= 5000) {
                /* limit reached */
                break;
            }
        }
    }

    /* cleanup */
    cleanUpNodes();

    return res;
}

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------