/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "MPEGI/Helpers/IPathfinder.h"
#include <valarray>

//--------------------------------------------------------------------------------------------------

// This is an efficient C++ implementation of the JPS pathfinding algorithm.
// D. Harabor; A. Grastien (2011). Online Graph Pruning for Pathfinding on Grid Maps.
// 25th National Conference on Artificial Intelligence. AAAI.

namespace MPEGI {
namespace Voxel {
//--------------------------------------------------------------------------------------------------

// Contains all relevant information for a position in the grid
struct jps_node
{
    int x; //!< x coordinate of the node
    int y; //!< y coordinate of the node
    int g; //!< euclidean distance of the node to active node
    int h; //!< manhattan distance of the node to active node
    int f; //!< Sum of g and h distance
    bool opened; //!< node opended flag
    bool closed; //!< node closed flag
    struct jps_node* parent; //!< parent of node
};

class jps_pathfinder : public i_pathfinder
{

public:
    jps_pathfinder();
    virtual ~jps_pathfinder();

protected:
    //@{
    /// Implementation of i_pathfinder
    int init(projection_map const& p_map, bool smooth_path) override;
    int find_path(voxel_index_t const& from, voxel_index_t const& to) override;
    int test_direct_view(voxel_index_t const& from, voxel_index_t const& to) override;

    Voxel::path_2d const& get_path() const override;
    float const& get_path_length() const override;
    //@}

    path_2d _path; //!< Pathfinder results, i.e. the path x/y-indices interleaved
    float _path_length; //!< Pathfinder results, i.e. the path length

    // local data structures & functions

    int m_grid_width;
    int m_grid_height;
    projection_map const* m_projection_map;
    bool m_smooth_path;

    std::valarray<std::valarray<jps_node>> m_jps_grid_nodes;
    std::vector<jps_node*> m_opened_list; // opened node list
    std::vector<jps_node*> m_closed_list; // closed node list

    std::vector<std::pair<int, int>> m_neighbors; // neighbor list

    // Return a pointer to a node, identified by the x and y coordinates
    jps_node* getNodeAt(int x, int y);

    // Return true / false whether a node is walkable or not ( automatically performs isInside check )
    bool isWalkableAt(int x, int y);

    // Return True / False whether a node is inside the grid or not */
    bool isInside(int x, int y);

    // Find the neighbors for the given node. If the node has a parent,
    // prune the neighbors based on the jump point search algorithm, otherwise
    // return all available neighbors.
    void findNeighbors(jps_node* activeNode);

    // Computes Euclidean distance sqrt(dx * dx + dy * dy)
    int euclidean(int dx, int dy);
    // Computes Manhattan distance dx + dy
    int manhattan(int dx, int dy);

    // Jumps recursively into one direction until a suitable successor has been found.
    bool jump(int& x, int& y, int px, int py, int endX, int endY);

    // Finds successor nodes worth investigating
    void identifySuccessors(jps_node* activeNode, int endX, int endY);

    // Attempts to draw a straight line from start to end, using Bresenham's algorithm.
    // if this can be done, returns each of the line's points;
    // if not, returns an empty vector.
    void interpolate_line(int x0, int y0, int x1, int y1, std::vector<std::pair<int, int>>& line);

    // Once the goal has been reached - creates a list of all nodes passed to reach the goal
    // and return all the coordinates lying on the path formed by these coordinates.
    void backtrace(jps_node* activeNode, std::vector<int>& path, float& path_length);

    // Computes the path. Returns a list of X/Y coordinates that form the path.
    bool findPath(
        int startX, int startY, int endX, int endY, std::vector<int>& path, float& path_length);

    // Cleans node data changed during one pathfinder run
    void cleanUpNodes(void);
};

//--------------------------------------------------------------------------------------------------
} // namespace Voxel
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------