/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "VoxelEarlyReflection.h"
#include "MPEGI/Helpers/VoxelTraversal.h"
#include "mpegcore/spdlog.hpp"
#include "spdlog/sinks/basic_file_sink.h"

#include <algorithm>
#include <chrono>
#include <fstream>
#include <spdlog/cfg/env.h>

//--------------------------------------------------------------------------------------------------
#define LOG_INFO(logger, ...)                                                                      \
    do {                                                                                           \
        if (logger) {                                                                              \
            SPDLOG_LOGGER_INFO(logger, __VA_ARGS__);                                               \
        }                                                                                          \
    } while (0)

#define LOG_DEBUG(logger, ...)                                                                     \
    do {                                                                                           \
        if (logger) {                                                                              \
            SPDLOG_LOGGER_DEBUG(logger, __VA_ARGS__);                                              \
        }                                                                                          \
    } while (0)

#define LOG_TRACE(logger, ...)                                                                     \
    do {                                                                                           \
        if (logger) {                                                                              \
            SPDLOG_LOGGER_TRACE(logger, __VA_ARGS__);                                              \
        }                                                                                          \
    } while (0)

#define LOG_FLUSH(logger)                                                                          \
    do {                                                                                           \
        if (logger) {                                                                              \
            logger->flush();                                                                       \
        }                                                                                          \
    } while (0)

//--------------------------------------------------------------------------------------------------
namespace {

struct line_segment
{
    line_segment(EAR::Vector3 const& p0, EAR::Vector3 const& p1, std::size_t num_points)
    {
        _points.reserve(num_points);
        _points.push_back(p0);

        auto const d = p0 - p1;

        auto sign = [](float val) -> float {
            return static_cast<float>((0.f < val) - (val < 0.f));
        };

        auto const fac = 1.f / (num_points - 1);
        EAR::Vector3 x{ -sign(d[0]) * std::abs(d[0]) * fac, -sign(d[1]) * std::abs(d[1]) * fac,
            -sign(d[2]) * std::abs(d[2]) * fac };

        for (auto i = 1U; i < num_points - 1; ++i) {
            _points.push_back(p0 + x * i);
        }

        _points.push_back(p1);
    }

    std::vector<EAR::Vector3> const& get() const { return _points; }

protected:
    std::vector<EAR::Vector3> _points;
};

auto lerp = [](float a, float b, float t) { return a + t * (b - a); };

//--------------------------------------------------------------------------------------------------
} // namespace
//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
/**
 * Hit information on the voxel-grid
 * Cannot be in anon-namespace because it is needed to specialize the spdlog formatter-template
 */
struct hit_info_vox
{
    using shared = std::shared_ptr<hit_info_vox>;
    using delay_t = std::chrono::microseconds;
    using speed_of_sound_inv_t =
        std::chrono::duration<std::size_t, std::ratio<1, 343>>; //!< Inverse speed of sound sec/m

    hit_info_vox(int material_id_, MPEGI::Voxel::voxel_index_t const& hit_idx_,
        MPEGI::Voxel::voxel_index_t const& pre_hit_idx_)
        : material_id(material_id_)
        , hit_idx(hit_idx_)
        , pre_hit_idx(pre_hit_idx_)
    {}

    hit_info_vox(hit_info_vox const&) = default;
    hit_info_vox& operator=(hit_info_vox const&) = default;

    /**
     * Set the hit to listener distance and update the propagation delay
     * @param lis_idx Listener voxel-index
     * @param voxel_size Voxel size
     */
    void set_hit_to_listener_distance(MPEGI::Voxel::voxel_index_t const& lis_idx, float voxel_size)
    {
        hit_to_listener.dist = voxel_size * get_distance(hit_idx, lis_idx);
        hit_to_listener.delay = get_delay(hit_to_listener.dist);
    }

    /**
     * Set the item to hit distance and update the propagation delay
     * @param item_idx Item voxel-index
     * @param voxel_size Voxel size
     * @param addl_distance Additional distance, if item is outside the voxelized region
     */
    void set_item_to_hit_distance(
        MPEGI::Voxel::voxel_index_t const& item_idx, float voxel_size, float addl_distance)
    {
        item_to_hit.dist = voxel_size * get_distance(hit_idx, item_idx);
        item_to_hit.dist += addl_distance;
        item_to_hit.delay = get_delay(item_to_hit.dist);
    }

    /**
     * Set the item to listener (direct) distance
     * @param distance Distance
     */
    void set_item_to_listener_distance(float distance)
    {
        item_to_lis.dist = distance;
        item_to_lis.delay = get_delay(distance);
    }

    /**
     * Store the index of the corresponding primary item
     * @param index Primary item index
     */
    void set_primary_item_index(int index) { primary_idx = index; }

    /**
     * Get the item index of the corresponding primary item
     * @return Item index
     */
    int get_primary_item_index() const { return primary_idx; }

    /**
     * Get the overall propagation delay for this hit
     * @return Propagation delay in delay_t
     */
    delay_t get_delay() const { return item_to_hit.delay + hit_to_listener.delay; }

    /**
     * Get the distance corresponding to the overall propagation delay for this hit
     * @return Overall distance
     */
    float get_distance() const { return item_to_hit.dist + hit_to_listener.dist; }

    /**
     * Get direct path propagation delay (listener-to-item)
     * @return Direct path propagation delay
     */
    delay_t get_direct_delay() const { return item_to_lis.delay; }

    /**
     * Get the voxel-index of the hit
     * @return Voxel index (x,y,z) [1..dim]
     */
    MPEGI::Voxel::voxel_index_t const& get_hit_index() const { return hit_idx; }

    /**
     * Material ID of the voxel-hit
     * @return Voxel material ID
     */
    int get_material_id() const { return material_id; }

    /**
     * Get the voxel-index of the voxel right before the hit
     * @return Voxel index (x,y,z) [1..dim]
     */
    MPEGI::Voxel::voxel_index_t const& get_pre_hit_index() const { return pre_hit_idx; }

    /**
     * Set the hit location (azimuth, elevation, radius) relative to the listener
     *
     * @param hit_aer Relative hit location (az,el,r)
     * @param sector Sector ID
     */
    void set_hit_location_az_el_r(EAR::Polar const& hit_aer, int sector)
    {
        _sector = sector;
        _location_rel[0] = hit_aer[0];
        _location_rel[1] = hit_aer[1];
        _location_rel[2] = hit_aer[2];
    }

    /**
     * Get the sector for this hit
     * @return sector
     */
    int get_sector() const { return _sector; }

    /**
     * Get the hit location relative to the listener
     * @return Relative hit location
     */
    EAR::Polar const& get_hit_location_rel() const { return _location_rel; }

protected:
    float get_distance(
        MPEGI::Voxel::voxel_index_t const& from, MPEGI::Voxel::voxel_index_t const& to) const
    {
        EAR::Vector3 const from_(from);
        EAR::Vector3 const to_(to);

        return (from_ - to_).abs();
    }

    delay_t get_delay(float distance) const
    {
        return std::chrono::duration_cast<delay_t>(speed_of_sound_inv_t(1) * distance);
    }

private:
    friend std::ostream& operator<<(std::ostream&, hit_info_vox const&);
    friend struct fmt::formatter<hit_info_vox>;

protected:
    int material_id{ 0 }; //!< Voxel material ID
    int primary_idx{ -1 }; //!< Associated primary item index for this reflection
    MPEGI::Voxel::voxel_index_t hit_idx{ 0 }; //!< Voxel index of the collision
    MPEGI::Voxel::voxel_index_t pre_hit_idx{ 0 }; //!< Last voxel index before the collision
    int _sector{ -1 }; //!< Sector ID
    EAR::Polar _location_rel; //!< Hit location relative to the listener

    struct
    {
        float dist{ 0.f }; //!< Hit-to-listener distance
        delay_t delay; //!< Hit-to-listener delay

    } hit_to_listener;

    struct
    {
        float dist{ 0.f }; //!< Item-to-hit distance
        delay_t delay; //!< Item-to-hit delay

    } item_to_hit;

    struct
    {
        float dist{ 0.f }; //!< Item-to-listener distance (direct)
        delay_t delay; //!< Item-to-listener delay (direct)

    } item_to_lis;
};

//--------------------------------------------------------------------------------------------------
struct check_collision
{
    check_collision(MPEGI::Voxel::grid::shared const& grid)
        : _grid(grid)
    {}

    hit_info_vox::shared operator()(
        MPEGI::Voxel::voxel_index_t const& beg, std::array<int, 3> const& direction) const
    {
        auto pos = beg;
        auto prev_pos = pos;

        while (_grid->in_range(pos)) {
            auto const& vxl = _grid->at(pos);
            if (!vxl.is_sound_medium()) {
                return std::make_shared<hit_info_vox>(vxl.get_material_id(), pos, prev_pos);
            }

            prev_pos = pos;
            pos = next(pos, direction);
        }

        return nullptr;
    }

    MPEGI::Voxel::voxel_index_t& next(
        MPEGI::Voxel::voxel_index_t& pos, std::array<int, 3> const& direction) const
    {
        pos[0] += direction[0];
        pos[1] += direction[1];
        pos[2] += direction[2];

        return pos;
    }

protected:
    MPEGI::Voxel::grid::shared _grid;
};

//--------------------------------------------------------------------------------------------------
/**
 * Information for one sector downmix item
 */
using stream_buf_ptr = std::shared_ptr<MPEGI::StreamBuffer>;

struct dmx_info
{
    dmx_info(std::shared_ptr<StreamBuffer> const& buf,
        MPEGI::VoxelEarlyReflectionStage::fade_control fade)
        : _buf(buf)
        , _fade(fade)
    {}

    ~dmx_info() = default;
    dmx_info& operator=(dmx_info&&) = default;
    dmx_info(dmx_info&&) = default;

    stream_buf_ptr _buf; //!< Stream buffer to be considered for sector downmix
    MPEGI::VoxelEarlyReflectionStage::fade_control _fade;
};

/**
 * Downmix information for one sector
 */
struct sector_info
{
    sector_info(stream_buf_ptr const& dest)
        : _dest(dest)
    {}

    ~sector_info() = default;
    sector_info& operator=(sector_info&&) = default;
    sector_info(sector_info&&) = default;

    std::vector<dmx_info> _dmx_info; //< Downmix info for this sector
    stream_buf_ptr _dest; //!< Destination buffer
};

//--------------------------------------------------------------------------------------------------
/**
 * DSP data
 */
struct dsp_data : public RenderStage::DspData
{
    std::vector<sector_info> _sector_info; //!< Sector info for downmix
};

//--------------------------------------------------------------------------------------------------
std::shared_ptr<RenderStage> makeVoxelEarlyReflectionStage()
{
    return std::make_shared<VoxelEarlyReflectionStage>();
}

//--------------------------------------------------------------------------------------------------
VoxelEarlyReflectionStage::VoxelEarlyReflectionStage()
{
    static std::vector<float> const directions_default{ 0.f, 1.f, 0.f, 0.f, -1.f, 0.f, 1.f, 0.f,
        0.f, -1.f, 0.f, 0.f, 1.f, 1.f, 0.f, 1.f, -1.f, 0.f, -1.f, 1.f, 0.f, -1.f, -1.f, 0.f };

    static std::vector<float> const directions_default_26{ 0.f, 0.f, 1.f, 0.f, 0.f, -1.f, 0.f, 1.f,
        0.f, 0.f, -1.f, 0.f, 1.f, 0.f, 0.f, -1.f, 0.f, 0.f, 0.f, 1.f, 1.f, 0.f, -1.f, -1.f, 0.f,
        1.f, -1.f, 0.f, -1.f, 1.f, 1.f, 0.f, 1.f, -1.f, 0.f, -1.f, 1.f, 0.f, -1.f, -1.f, 0.f, 1.f,
        1.f, 1.f, 0.f, -1.f, -1.f, 0.f, 1.f, -1.f, 0.f, -1.f, 1.f, 0.f, 1.f, 1.f, 1.f, -1.f, -1.f,
        -1.f, 1.f, 1.f, -1.f, 1.f, -1.f, 1.f, -1.f, 1.f, 1.f, 1.f, -1.f, -1.f, -1.f, 1.f, -1.f,
        -1.f, -1.f, 1.f };

    Config::setDef("VoxelEarlyReflectionDirections", EAR::Variant(directions_default_26));
    Config::setDef("VoxelEarlyReflectionFilterByRuntime", true);
    Config::setDef("VoxelEarlyReflectionLogFilePath", std::string{ "" });
    Config::setDef("VoxelEarlyReflectionSectorDownmixEnabled", false);

    stage_configuration.num_points =
        Config::getIntDef("VoxelEarlyReflectionNumPoints", stage_configuration.num_points);
    stage_configuration.max_num_reflections = Config::getIntDef(
        "VoxelEarlyReflectionMaxNumReflections", stage_configuration.max_num_reflections);

    stage_configuration.runtime_filter.enabled =
        Config::getBool("VoxelEarlyReflectionFilterByRuntime");
    stage_configuration.runtime_filter.min_delay =
        std::chrono::microseconds(Config::getIntDef("VoxelEarlyReflectionMinDelayMicroSeconds",
            static_cast<unsigned int>(stage_configuration.runtime_filter.min_delay.count())));
    stage_configuration.runtime_filter.max_delay =
        std::chrono::microseconds(Config::getIntDef("VoxelEarlyReflectionMaxDelayMicroSeconds",
            static_cast<unsigned int>(stage_configuration.runtime_filter.max_delay.count())));
    stage_configuration.set_sector_gain(Config::getFloatDef("VoxelEarlyReflectionLevel",
        std::pow(10.f, BitstreamConfig::early_tuning_gain_default_db * .05f)));
    stage_configuration.sector_downmix_enabled =
        Config::getBool("VoxelEarlyReflectionSectorDownmixEnabled");

    stage_configuration.environment.gain_indoor =
        Config::getFloatDef("VoxelEarlyReflectionGainIndoor", 2.f);
    stage_configuration.environment.gain_outdoor =
        Config::getFloatDef("VoxelEarlyReflectionGainOutdoor", 1.f);

    Config::observeKey("VoxelEarlyReflectionLevel", this);
    Config::observeKey("VoxelEarlyReflectionMinDelayMicroSeconds", this);
    Config::observeKey("VoxelEarlyReflectionMaxDelayMicroSeconds", this);
    Config::observeKey("VoxelEarlyReflectionNumPoints", this);
    Config::observeKey("VoxelEarlyReflectionMaxNumReflections", this);
    Config::observeKey("VoxelEarlyReflectionGainIndoor", this);
    Config::observeKey("VoxelEarlyReflectionGainOutdoor", this);

    try {
        spdlog::cfg::load_env_levels();
        auto logger =
            spdlog::basic_logger_mt("VoxER", Config::getString("VoxelEarlyReflectionLogFilePath"));

        SPDLOG_LOGGER_INFO(
            logger, "VoxelEarlyReflectionNumPoints: {} points", stage_configuration.num_points);
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionDirections: {} directions",
            Config::getFloatVectorDef("VoxelEarlyReflectionDirections", {}).size() / 3);
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionMaxNumReflections: {} reflections per item",
            stage_configuration.max_num_reflections);
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionMinDelayMicroSeconds: {}us",
            stage_configuration.runtime_filter.min_delay.count());
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionMaxDelayMicroSeconds: {}us",
            stage_configuration.runtime_filter.max_delay.count());
        SPDLOG_LOGGER_INFO(
            logger, "VoxelEarlyReflectionLevel: {}", Config::getFloat("VoxelEarlyReflectionLevel"));
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionSectorDownmixEnabled: {}",
            Config::getBool("VoxelEarlyReflectionSectorDownmixEnabled"));
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionGainIndoor: {}",
            Config::getFloat("VoxelEarlyReflectionGainIndoor"));
        SPDLOG_LOGGER_INFO(logger, "VoxelEarlyReflectionGainOutdoor: {}",
            Config::getFloat("VoxelEarlyReflectionGainOutdoor"));

    } catch (const spdlog::spdlog_ex&) {
        // No logging
    }
}

//--------------------------------------------------------------------------------------------------
VoxelEarlyReflectionStage::~VoxelEarlyReflectionStage()
{
    Config::removeObserver("VoxelEarlyReflectionLevel", this);
    Config::removeObserver("VoxelEarlyReflectionMinDelayMicroSeconds", this);
    Config::removeObserver("VoxelEarlyReflectionMaxDelayMicroSeconds", this);
    Config::removeObserver("VoxelEarlyReflectionNumPoints", this);
    Config::removeObserver("VoxelEarlyReflectionMaxNumReflections", this);
    Config::removeObserver("VoxelEarlyReflectionGainIndoor", this);
    Config::removeObserver("VoxelEarlyReflectionGainOutdoor", this);
}

//--------------------------------------------------------------------------------------------------
void VoxelEarlyReflectionStage::initializeStage(const Bitstream& bitstream,
    RenderItemStore* itemStore, const StreamManager* streamManager, const SceneState* state)
{
    // Setup voxel-scene
    auto const voxel_scene_state = state->getVoxelSceneState();
    if (!voxel_scene_state) {
        return;
    }

    auto const materials = std::dynamic_pointer_cast<EIFPayload const>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE))
                               ->getAcousticMaterials();

    auto init = [this, materials](Voxel::VoxelSceneState::const_shared const& voxel_scene_state) {
        this->_scene = voxel_scene_state->get_scene_info_basic();
        this->_grid = voxel_scene_state->get_grid();

        // Map material IDs to voxel-material IDs
        auto const p_bands = Config::getFloatPtr("EQCenterFreqs");
        std::vector<float> const bands(p_bands, p_bands + Config::getInt("EQNBands"));

        for (auto const& [id, desc] : voxel_scene_state->get_current_subscene()->scene_materials) {
            auto const it = std::find_if(std::cbegin(materials), std::cend(materials),
                [bs_id = desc.bs_id](auto const& m) { return m.bsId == bs_id; });

            if (std::end(materials) == it) {
                continue;
            }

            EAR::Material ear_mat;
            for (auto const& f : it->frequencies) {
                ear_mat.append(f.f, f.r, f.s, f.t, f.c);
            }

            auto& coeff = _reflection_coeffs[id];
            coeff.clear();
            for (auto const& b : bands) {
                coeff.push_back(ear_mat.getReflectionCoeffForFreq(b));
            }
        }
    };

    init(voxel_scene_state);
    // Register callbacks
    voxel_scene_state->on_subscene_change(init);

    voxel_scene_state->on_scene_update(
        [this](Voxel::VoxelSceneState::const_shared const& voxel_scene_state) {
            scene_update = true;
        });

    // Setup directions
    auto const& directions = Config::getFloatVectorDef("VoxelEarlyReflectionDirections", {});
    EAR_CHECK_REL(!directions.empty() && !(directions.size() % 3),
        "VoxelEarlyReflectionDirections: invalid dimension. Must be non-empty and multiple of 3");

    _directions.reserve(directions.size() / 3);

    auto it = std::begin(directions);
    while (it != std::cend(directions)) {
        auto const x = static_cast<int>(*it++);
        auto const y = static_cast<int>(*it++);
        auto const z = static_cast<int>(*it++);

        _directions.push_back({ x, y, z });
    }

    // Item storage for ER secondary render-items
    _item_store =
        std::make_shared<early_reflection_item_store>(itemStore, Config::getInt("EQNBands"));
    EAR_CHECK_REL(_item_store != nullptr && _item_store->valid(),
        "VoxelEarlyReflectionStage: Failed to create item store");

    // Index storage
    _index_storage = voxel_scene_state->get_index_storage();
    EAR_CHECK_REL(
        _index_storage != nullptr, "VoxelEarlyReflectionStage: invalid voxel-index store");

    // Voxel traversal
    _vox_traversal = std::make_shared<Voxel::voxel_traversal>();
    EAR_CHECK_REL(
        _vox_traversal != nullptr, "VoxelEarlyReflectionStage: Failed to setup voxel traversal");

    // Sector downmix
    if (stage_configuration.sector_downmix_enabled) {
        setup_downmix(streamManager);
    }
}

//--------------------------------------------------------------------------------------------------
void VoxelEarlyReflectionStage::processBlock(DspContext ctx)
{
    auto const& data = ctx.getData<dsp_data>();
    if (data._sector_info.empty()) {
        return;
    }

    for (auto const& info : data._sector_info) {
        // Initialize destination buffer for current sector
        auto& buf = info._dest->getFrame();
        buf.fill(0.f);

        auto const sqrt_n_dmx = std::sqrt(static_cast<float>(info._dmx_info.size()));

        for (auto const& dmx : info._dmx_info) {
            auto const fac = 1.f / sqrt_n_dmx;
            auto const a = fade_control::fade_in == dmx._fade ? 0.f : 1.f;
            auto const b = fade_control::fade_out == dmx._fade ? 0.f : 1.f;

            for (auto ch = 0U; ch < buf.channels(); ++ch) {
                auto dest = buf.dataRaw()[ch];
                auto src = dmx._buf->getFrame().dataRaw()[ch];

                for (auto i = 0U; i < buf.length(); ++i) {
                    auto const t = i / static_cast<float>(buf.length() - 1);
                    dest[i] += src[i] * fac * lerp(a, b, t);
                }
            }
        }
    }
}

//--------------------------------------------------------------------------------------------------
std::unique_ptr<RenderStage::DspData> VoxelEarlyReflectionStage::update(RenderList& render_list)
{
    if (!initialized() || !enabled()) {
        return nullptr;
    }

    auto stage_dsp_data =
        stage_configuration.sector_downmix_enabled ? std::make_unique<dsp_data>() : nullptr;

    auto& idx_store = *_index_storage;
    auto& lis = render_list.listener.position.location;

    idx_store.update_listener_voxel_index(_scene->to_voxel_grid(lis));
    auto const lis_idx = idx_store.get_listener_voxel_index();

    auto const lis_has_moved = lis_idx != _last_lis_idx;
    _last_lis_idx = lis_idx;

    auto const lis_xyz_vox = Voxel::opengl_to_vox(lis.data());

    auto logger = spdlog::get("VoxER");
    LOG_DEBUG(logger, "Listener: voxel=({}, {}, {})", lis_idx[0], lis_idx[1], lis_idx[2]);

    check_collision cc(_grid);

    using aggregated_reflections_per_item_and_sector_t =
        std::unordered_map<int, std::unordered_map<int, aggregated_reflection>>;

    aggregated_reflections_per_item_and_sector_t aggregated_reflections_per_item_and_sector;

    // Discard all hits that are neither visible from the source nor from the listener
    auto is_visible = [this](hit_info_vox const& hit, std::array<float, 3> const& from_xyz,
                          Voxel::voxel_index_t const& from_idx) -> bool {
        auto const to_idx = hit.get_pre_hit_index();
        auto const to_xyz = _scene->idx2xyz(to_idx);

        if (_vox_traversal->run(*_scene, from_xyz, from_idx, to_xyz, to_idx)) {
            auto const& path = _vox_traversal->get_path();
            return !std::any_of(
                std::cbegin(path), std::cend(path), [grid = _grid](auto const& step) {
                    auto const is_sound_medium = grid->at(step).is_sound_medium();
                    return !is_sound_medium;
                });
        }

        return false;
    };

    // map voxel hit info to sector
    auto map_hit_info_vox_to_sector = [this](hit_info_vox& hit, EAR::Vector3& lis) {
        auto const num_reflections_max = stage_configuration.max_num_reflections;
        auto const sector_size_2_deg = 180.f / num_reflections_max;

        auto wrap_360 = [](float deg) -> float {
            while (deg < 0.f) {
                deg += 360.f;
            }
            return deg;
        };

        EAR::Vector3 const hit_xyz = Voxel::vox_to_opengl(_scene->idx2xyz(hit.get_hit_index()));
        auto hit_rel = EAR::cartesianToPolar(hit_xyz - lis);
        // replace radius withe total distance to item
        hit_rel[2] = hit.get_distance();
        // Center of first sector (id=0) is at 0 deg
        auto const az = wrap_360(hit_rel[0] + sector_size_2_deg);
        const int sector = (int)(az / (2 * sector_size_2_deg));

        // set relative location and sector
        hit.set_hit_location_az_el_r(hit_rel, sector);
    };

    // make a local copy of scene_update to handle asynchronous voxel grid updates
    auto const _scene_update = scene_update;

    // Indoor/outdoor estimation
    std::unordered_map<int, int> hit_count_by_item;
    std::unordered_map<int, unsigned int> num_points_by_item;
    std::unordered_map<int, float> hit_ratio_by_item;

    for (auto const& item : render_list.items) {

        /// @todo Check item types
        if (item->type != ItemType::Primary || item->position.cSpace == EAR::CSpace::LCS) {
            continue;
        }
        auto& item_state = item->getStateOrCreate<VoxelEarlyReflectionStage::ItemState>();

        if (item->isInactive() || item->isOrphaned()) {
            _item_store->mute(item->idx);
            continue;
        }

        idx_store.update_voxel_index(
            item->idx, _scene->to_voxel_grid(item->position.location, lis));

        hit_count_by_item[item->idx] = 0;
        num_points_by_item[item->idx] = stage_configuration.num_points;

        // check if item or listener has changed or moved. Recalculate reflections in this case
        bool recalculateReflections = ((item->changed & ItemProperty::Position)
            || (item->changed & ItemProperty::Status) || lis_has_moved || _scene_update);

        if (recalculateReflections) {
            item_state._aggregated_item_reflections.clear();
            auto const item_idx = idx_store.get_voxel_index(item->idx);
            auto const item_xyz_vox = Voxel::opengl_to_vox(item->position.location.data());
            auto const item_to_lis_dist{ (item->position.location - lis).abs() };

            LOG_DEBUG(logger, "Item: idx={}, voxel=({}, {}, {})", item->idx, item_idx[0],
                item_idx[1], item_idx[2]);
            // calculate line segment from listener voxel to item
            line_segment s(lis, item->position.location, stage_configuration.num_points);

            for (auto const& p : s.get()) {
                auto const [p_idx, addl_dist] = _scene->to_voxel_grid(p, lis);
                if (addl_dist > 0.f) {
                    // Point is outside the voxel-grid
                    --num_points_by_item[item->idx];
                    continue;
                }

                for (auto const& d : _directions) {
                    auto hit = cc(p_idx, d);
                    if (hit) {

                        hit->set_hit_to_listener_distance(lis_idx, _scene->voxel_size);
                        hit->set_item_to_hit_distance(item_idx, _scene->voxel_size,
                            idx_store.get_additional_distance(item->idx));
                        hit->set_item_to_listener_distance(item_to_lis_dist);
                        hit->set_primary_item_index(item->idx);

                        // Keep track of ALL hits for indoor/outdoor estimation
                        ++hit_count_by_item[item->idx];

                        // Discard all hits below/above the min/max propagation delay threshold

                        auto const delay_min = stage_configuration.runtime_filter.min_delay;
                        auto const delay_max = stage_configuration.runtime_filter.max_delay;

                        if (!stage_configuration.runtime_filter.enabled
                            || (hit->get_delay() <= (delay_max + hit->get_direct_delay())
                                && hit->get_delay() >= (delay_min + hit->get_direct_delay()))) {
                            // Check visibility from listener and item, test can be skipped if
                            // either listener or item is origin
                            bool is_visible_from_lis, is_visible_from_item;

                            is_visible_from_lis = true;
                            is_visible_from_item = true;
                            if (p_idx != lis_idx) {
                                is_visible_from_lis = is_visible(*hit, lis_xyz_vox, lis_idx);
                            }
                            if (p_idx != item_idx) {
                                is_visible_from_item = is_visible(*hit, item_xyz_vox, item_idx);
                            }

                            if (is_visible_from_lis && is_visible_from_item) {
                                map_hit_info_vox_to_sector(*hit, lis);
                                // Only one reflection is generated per sector, aggregate hit and reflection coeffs
                                auto sector = hit->get_sector();
                                auto location_rel = hit->get_hit_location_rel();
                                auto material_id = hit->get_material_id();

                                auto reflection_coeffs = _reflection_coeffs.find(material_id);

                                if (reflection_coeffs != std::cend(_reflection_coeffs)) {
                                    aggregated_reflections_per_item_and_sector
                                        [item->idx][hit->get_sector()]
                                            .add_reflection(hit->get_hit_location_rel(),
                                                reflection_coeffs->second);
                                } else {
                                    EAR_FAIL(
                                        "VoxelEarlyReflectionStage: no reflection coefficients "
                                        "for material id "
                                        + std::to_string(material_id));
                                }
                            }
                        }
                    }
                }
            }

            // Indoor/outdoor estimation
            hit_ratio_by_item[item->idx] = hit_count_by_item[item->idx]
                / static_cast<float>(num_points_by_item[item->idx] * _directions.size());

            item_state._hit_ratio = hit_ratio_by_item[item->idx];

            // finalize aggregated hits per sector
            if (aggregated_reflections_per_item_and_sector.count(item->idx) > 0) {

                for (auto& [sector, aggregated_reflection] :
                    aggregated_reflections_per_item_and_sector[item->idx]) {
                    aggregated_reflection.aggregate_reflections();
                }

                // update item state with new aggregated reflections
                item_state._aggregated_item_reflections =
                    aggregated_reflections_per_item_and_sector[item->idx];
            }
        } else {
            // use aggregated reflections from item state
            aggregated_reflections_per_item_and_sector[item->idx] =
                item_state._aggregated_item_reflections;
            hit_ratio_by_item[item->idx] = item_state._hit_ratio;
        }

        LOG_INFO(logger, "item={}, hit-ratio={}", item->idx, hit_ratio_by_item[item->idx]);
    }
    if (_scene_update == scene_update) {
        scene_update = false;
    }

    // loop through aggregated reflections per sector and update in the storage
    auto const sector_gain = stage_configuration.sector_gain;
    for (auto const& [item_idx, aggregated_reflections_per_sector] :
        aggregated_reflections_per_item_and_sector) {
        const RenderItem* primary = render_list.getItemByIndex(item_idx);
        EAR_CHECK_REL(
            primary != nullptr, "VoxelEarlyReflectionStage: reflection for non-existing item");

        LOG_DEBUG(logger, "{} active sectors for item {}", aggregated_reflections_per_sector.size(),
            item_idx);

        for (auto const& [sector, aggregated_reflection] : aggregated_reflections_per_sector) {

            auto hit_location = aggregated_reflection.get_reflection_location(lis);
            auto& r = aggregated_reflection.get_reflection_coeffs();

            // generate / update er secondary object
            auto i = _item_store->update_item(primary, sector, hit_location);

            i->setEq(r);
            i->setGain(sector_gain
                * lerp(stage_configuration.environment.gain_outdoor,
                    stage_configuration.environment.gain_indoor, hit_ratio_by_item[item_idx]));
        }
    }

    //----------------------------------------------------------------------------------------------
    // Sector downmix
    //----------------------------------------------------------------------------------------------
    if (stage_configuration.sector_downmix_enabled) {
        for (auto s_idx = 0U; s_idx < stage_configuration.max_num_reflections; ++s_idx) {
            // Get reflection items per sector
            auto er_items = _item_store->get_secondary_by_sector(s_idx);

            // Fade-logic for calculation of downmix
            {
                for (auto& item : er_items) {
                    item->getStateOrCreate<item_state>().fade(fade_control::no_fade);
                }

                // Current render items in this sector (sorted)
                std::set<RenderItem*> all_cur(std::cbegin(er_items), std::cend(er_items));

                // Previous render items in this sector (sorted)
                auto& all_prev = _prev_items[s_idx];
                for (auto& item : all_prev) {
                    item->getStateOrCreate<item_state>().fade(fade_control::no_fade);
                }

                // Fade-in: items that are in current but not in previous set
                std::vector<RenderItem*> fade_in;
                std::set_difference(std::cbegin(all_cur), std::cend(all_cur), std::cbegin(all_prev),
                    std::cend(all_prev), std::back_inserter(fade_in));

                for (auto& item : fade_in) {
                    item->getStateOrCreate<item_state>().fade(fade_control::fade_in);
                }

                // Fade-out: itmes that are in previous but not in current set
                std::vector<RenderItem*> fade_out;
                std::set_difference(std::cbegin(all_prev), std::cend(all_prev),
                    std::cbegin(all_cur), std::cend(all_cur), std::back_inserter(fade_out));

                for (auto& item : fade_out) {
                    item->getStateOrCreate<item_state>().fade(fade_control::fade_out);
                }

                // Save for next frame
                all_prev.clear();
                all_prev.insert(std::cbegin(er_items), std::cend(er_items));

                // Make sure that items are faded out properly
                std::copy_if(std::cbegin(fade_out), std::cend(fade_out),
                    std::back_inserter(er_items), [](auto const& ri) { return !ri->isOrphaned(); });
            }

            // Adding the buffer - buffer will be zeroed out if sector is inactive
            stage_dsp_data->_sector_info.reserve(_buf.size());
            auto& info = stage_dsp_data->_sector_info.emplace_back(_buf[s_idx]);

            LOG_DEBUG(logger, "{} individual reflection(s) in sector {}", er_items.size(), s_idx);
            if (er_items.empty()) {
                continue;
            }

            // Get sector info and fill DSP data
            auto it = std::begin(er_items);

            // Store in DSP data
            info._dmx_info.reserve(er_items.size());
            while (it != std::cend(er_items)) {
                auto& item = *it++;
                if (-1 != item->parentRI) {
                    // Item is not a downmix reflection item
                    info._dmx_info.emplace_back(
                        item->signal, item->getStateOrCreate<item_state>().fade());

                    // Sector downmix active, therefore de-activate contributing RIs
                    item->setStatus(MPEGI::ItemStatus::Inactive);
                }
            }

            // Generate sector-downmix RI
            // 1. Create item, if not alreday present. Downmix ER item does not have a parent
            auto dmx_item = _item_store->get_downmix_item(s_idx);
            if (!dmx_item) {
                dmx_item = _item_store->make_downmix_item(s_idx);
                dmx_item->signal = info._dest;
            }

            // 2. Calculate location of image source in the sector
            auto const pos_sec =
                std::accumulate(std::cbegin(er_items), std::cend(er_items), EAR::Vector3{},
                    [](EAR::Vector3 const& acc, RenderItem const* ri) {
                        return acc + ri->position.location;
                    })
                * (1.f / er_items.size());

            // 3. Average EQs
            std::fill(std::begin(dmx_item->eq), std::end(dmx_item->eq), 0.f);
            for (auto const& item : er_items) {
                std::transform(std::cbegin(item->eq), std::cend(item->eq), std::begin(dmx_item->eq),
                    std::begin(dmx_item->eq), std::plus<float>());
            }

            std::transform(std::cbegin(dmx_item->eq), std::cend(dmx_item->eq),
                std::begin(dmx_item->eq),
                [size = er_items.size()](auto const& val) { return val / size; });

            dmx_item->setPosition({ pos_sec, { 0, 0, 0 } });
            dmx_item->setGain(sector_gain);
            dmx_item->changed |= ItemProperty::Eq | ItemProperty::Gain;

            LOG_DEBUG(logger, "[S-{}] {}", s_idx, *dmx_item);
        }
    }

    LOG_FLUSH(logger);

    // Remove non-active reflections
    for (auto const& item : render_list.items) {
        if (item->type == ItemType::Primary) {
            _item_store->remove_unchanged(item->idx);
        }
    }

    // Remove non-active downmix-items, if any
    _item_store->remove_unchanged(-1);

    // Add to the list
    _item_store->add_items(render_list);
    return stage_dsp_data;
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------

template <> struct fmt::formatter<MPEGI::hit_info_vox> : fmt::formatter<std::string>
{
    auto format(MPEGI::hit_info_vox my, format_context& ctx) -> decltype(ctx.out())
    {
        return format_to(ctx.out(), "Hit: primary={}, material={}, voxel=({}, {}, {}), delay={}",
            my.get_primary_item_index(), my.get_material_id(), my.hit_idx[0], my.hit_idx[1],
            my.hit_idx[2], my.get_delay().count());
    }
};

template <> struct fmt::formatter<MPEGI::RenderItem> : fmt::formatter<std::string>
{
    auto format(MPEGI::RenderItem my, format_context& ctx) -> decltype(ctx.out())
    {
        std::array<float, 3> const loc = my.position.location;
        auto const& eq = my.eq;

        return format_to(ctx.out(), "RI-{}, loc=({}), eq=({}), gain={}", my.idx,
            fmt::join(loc, ", "), fmt::join(eq, ", "), my.gain);
    }
};