/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include "MPEGI/Helpers/SecondaryItemStorage.h"
#include "MPEGI/Helpers/VoxelTypes.h"
#include "MPEGI/Renderer/RenderStage.h"

#include <array>
#include <chrono>
#include <functional>
#include <map>
#include <vector>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
namespace Voxel {
struct voxel_traversal;
}

//--------------------------------------------------------------------------------------------------

class VoxelEarlyReflectionStage : public RenderStage, public MPEGI::ConfigObserver
{
public:
    VoxelEarlyReflectionStage();
    ~VoxelEarlyReflectionStage();

    enum class fade_control
    {
        no_fade, //!< No fading applied
        fade_in, //!< Fade-in, i.e., first frame contributing to downmix
        fade_out //!< Fade-out, i.e., last frame contributing to downmix
    };

protected:
    /// Deactivation threshold
    inline static float const stage_deactivation_threshold = std::pow(10.f, -100.f * .05f);

    //@{
    /// Implementation of RendererStage Interface
    std::unique_ptr<RenderStage::DspData> update(RenderList& renderList) override;
    void initializeStage(
        const Bitstream&, RenderItemStore*, const StreamManager*, const SceneState*) override;

    void processBlock(DspContext ctx) override;
    //@}

    /**
     * Check whether stage is initialized
     * @return True if initialized, else false
     */
    bool initialized() const { return _scene && _grid; }

    /**
     * @brief Check whether the stage should be enabled
     * Stage is disabled if maximum gains for both indoor and outdoor
     * are below the activation threshold (stage_deactivation_threshold_db)
     * @return True if enabled, else false
     */
    bool enabled() const
    {
        return stage_configuration.environment.gain_indoor > stage_deactivation_threshold
            || stage_configuration.environment.gain_outdoor > stage_deactivation_threshold;
    }

    /**
     * Set the up sector downmix
     * @param stream_manager Stream manager for buffer management
     */
    void setup_downmix(StreamManager const* stream_manager)
    {
        // Create stream-buffers for downmix items
        _buf.resize(stage_configuration.max_num_reflections);

        for (auto sec = 0; sec < stage_configuration.max_num_reflections; ++sec) {
            _buf.at(sec) = stream_manager->createEmptyBuffer();
        }

        _prev_items.assign(stage_configuration.max_num_reflections, {});
    }

protected:
    /// Early reflection item store
    struct early_reflection_item_store
        : public Voxel::secondary_item_storage<early_reflection_item_store, RenderItem const*>
    {
        using shared = std::shared_ptr<early_reflection_item_store>;
        early_reflection_item_store(RenderItemStore* render_item_store, unsigned int num_bands_eq)
            : secondary_item_storage(render_item_store, num_bands_eq)
        {}

        ItemProperty get_sync_flags(RenderItem const&) const
        {
            return ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Eq
                | ItemProperty::Loudness | ItemProperty::RefDist | ItemProperty::Signal
                | ItemProperty::Status;
        }

        ItemType get_item_type(RenderItem const&) const { return ItemType::Reflection; }
        std::string get_suffix() const { return "_er_"; }

        RenderItem* get_downmix_item(std::size_t sec_idx)
        {
            auto& sec_items = _storage[-1];
            auto sec_it = sec_items.find(sec_idx);
            if (sec_it != std::cend(sec_items)) {
                return sec_it->second;
            }

            return nullptr;
        }

        RenderItem* make_downmix_item(std::size_t sec_idx)
        {
            auto& sec_items = _storage[-1];
            auto& item = sec_items[sec_idx];

            item = _store->createItem("sector_dmx" + get_suffix() + std::to_string(sec_idx));
            item->type = get_item_type(*item);
            item->parentRI = -1;

            return item;
        }
    };

    /// Basic scene info
    Voxel::voxel_scene_info_basic::shared _scene;

    /// Voxel grid
    Voxel::grid::shared _grid;

    /// Directions for the voxel traversal
    std::vector<std::array<int, 3>> _directions;

    /// Storage for secondary ER items
    early_reflection_item_store::shared _item_store;

    /// Voxel traversal
    std::shared_ptr<Voxel::voxel_traversal> _vox_traversal;

    /// To store/share calculated voxel indices
    Voxel::index_storage::shared _index_storage;

    /// Reflection coefficients
    using reflection_coeffs_per_id_t = std::unordered_map<int, std::vector<float>>;
    reflection_coeffs_per_id_t _reflection_coeffs;

    /// Stream buffers for generation of downmix items
    std::vector<std::shared_ptr<MPEGI::StreamBuffer>> _buf;

    /// List of previous render items to control fading for downmix
    std::vector<std::set<RenderItem*>> _prev_items;

    /// Previous voxel index of the listener
    Voxel::voxel_index_t _last_lis_idx{ 0 };

    struct stage_configruation_t
    {
        inline static float const reflection_max_level = std::pow(10, .05f * 10);

        void set_sector_gain(float level)
        {
            sector_gain = 1 / std::sqrt(static_cast<float>(max_num_reflections))
                * std::clamp(level, 0.f, reflection_max_level);
        }

        unsigned int num_points{ 3 }; //!< Number of line segments + 1
        unsigned int max_num_reflections{
            16
        }; //!< Maximum number of reflections generated per item
        float sector_gain{ 0.f }; //!< Sector gain, derived from global reflection level

        struct
        {
            bool enabled{ false }; //!< Enable runtime filtering
            std::chrono::microseconds min_delay{ 100 }; //!< Minimum runtime
            std::chrono::microseconds max_delay{ 150000 }; //!< Maximum runtime

        } runtime_filter;

        struct
        {
            float gain_outdoor{ 1.f };
            float gain_indoor{ 2.f };

        } environment;

        bool sector_downmix_enabled{ false };

    } stage_configuration;

    struct item_state : RenderItem::ExtraState
    {
        void fade(fade_control ctrl) { _fade_ctrl = ctrl; }
        fade_control fade() const { return _fade_ctrl; }

    protected:
        fade_control _fade_ctrl{ fade_control::no_fade };
    };

    /**
     * Implements MPEGI::ConfigObserver
     */
    void configValueChanged(const std::string& key) override
    {
        static std::unordered_map<std::string,
            std::function<void(std::string, stage_configruation_t&)>> const observed = {
            { "VoxelEarlyReflectionLevel",
                [](std::string const& key, stage_configruation_t& config) {
                    config.set_sector_gain(Config::getFloat(key));
                } },

            { "VoxelEarlyReflectionMinDelayMicroSeconds",
                [](std::string const& key, stage_configruation_t& config) {
                    config.runtime_filter.min_delay =
                        std::chrono::microseconds(Config::getInt(key));
                } },

            { "VoxelEarlyReflectionMaxDelayMicroSeconds",
                [](std::string const& key, stage_configruation_t& config) {
                    config.runtime_filter.max_delay =
                        std::chrono::microseconds(Config::getInt(key));
                } },

            { "VoxelEarlyReflectionNumPoints",
                [](std::string const& key, stage_configruation_t& config) {
                    config.num_points = Config::getInt(key);
                } },

            { "VoxelEarlyReflectionMaxNumReflections",
                [](std::string const& key, stage_configruation_t& config) {
                    config.max_num_reflections = Config::getInt(key);
                } },
            { "VoxelEarlyReflectionGainIndoor",
                [](std::string const& key, stage_configruation_t& config) {
                    config.environment.gain_indoor = Config::getFloat(key);
                } },
            { "VoxelEarlyReflectionGainOutdoor",
                [](std::string const& key, stage_configruation_t& config) {
                    config.environment.gain_outdoor = Config::getFloat(key);
                } }
        };

        auto const observed_it = observed.find(key);
        if (observed_it != std::cend(observed)) {
            observed_it->second(key, this->stage_configuration);
        }
    }

    struct aggregated_reflection
    {
        void add_reflection(EAR::Polar const& location_rel, std::vector<float>& r)
        {
            _location_rel[0] += location_rel[0];
            _location_rel[1] += location_rel[1];
            _location_rel[2] += location_rel[2];

            if (_r_coeff.empty()) {
                _r_coeff = r;
            } else {
                std::transform(std::cbegin(_r_coeff), std::cend(_r_coeff), std::begin(r),
                    std::begin(_r_coeff), std::plus<float>());
            }
            _n_reflections++;
        }
        void aggregate_reflections()
        {
            if (_n_reflections != 0) {
                _location_rel[0] /= _n_reflections;
                _location_rel[1] /= _n_reflections;
                _location_rel[2] /= _n_reflections;

                std::transform(std::begin(_r_coeff), std::end(_r_coeff), std::begin(_r_coeff),
                    [n_reflections = _n_reflections](auto const& v) { return v / n_reflections; });
            }
        }
        /**
          * Get the reflection location (x,y,z, absolute)
          *
          * @param lis_xyz Listener location (x,y,z)
          * @return reflection location (x,y,z)
        */

        Vector3 get_reflection_location(EAR::Vector3 const& lis_xyz) const
        {
            return { EAR::polarToCartesian(_location_rel) + lis_xyz };
        }

        std::vector<float> const& get_reflection_coeffs() const { return _r_coeff; }

    private:
        EAR::Polar _location_rel; //!< Aggregated hit location relative to the listener;
        std::vector<float> _r_coeff; //!< Aggregated reflection coefficients
        int _n_reflections{ 0 }; //!< Number of aggregated reflections
    };

    bool scene_update = { false };

    /// Aggregated reflections for one primary item per sector
    struct ItemState : RenderItem::ExtraState
    {
        ItemState()
            : _aggregated_item_reflections()
        {}

        std::unordered_map<int, aggregated_reflection> _aggregated_item_reflections;
        float _hit_ratio{ 1.f }; /// Hit ratio for indoor/outdoor estimation
    };
};

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
