/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#include "VoxelOcclusion.h"
#include "MPEGI/Bitstream/PortalPayload.h"
#include "MPEGI/Bitstream/VoxDataPayload.h"
#include "MPEGI/Renderer/PerceptualPointDistribution.h"

//--------------------------------------------------------------------------------------------------
namespace {
//--------------------------------------------------------------------------------------------------
constexpr bool is_heterogeneous_source(MPEGI::RenderItem const& item) noexcept
{
    using MPEGI::ItemType;
    return item.type != ItemType::DirectionalReverb && item.extent
        && item.signal->getFrame().channels() > 1;
}

//--------------------------------------------------------------------------------------------------
constexpr bool is_occludable(MPEGI::RenderItem const& item) noexcept
{
    using EAR::CSpace;
    using MPEGI::ItemType;

    auto const t = item.type;
    auto const cspace = item.position.cSpace;

    return
        /// @todo Types from Occlusion.cpp, to be verified
        (t == ItemType::Primary || t == ItemType::Representative
            || t == ItemType::DirectionalReverb)
        && cspace != CSpace::LCS && !is_heterogeneous_source(item);
}

constexpr bool has_local_occlusion(int idx_local, int idx_global) noexcept
{
    // Apply local occlusion if:
    //  a) Both L and G but L is closer
    //  b) Only L
    return (idx_local >= 0) && (idx_local < idx_global || idx_global == -1);
}

constexpr bool has_global_occlusion(int idx_global) noexcept
{
    return idx_global >= 0;
}

constexpr bool is_occluded(int idx_local, int idx_global) noexcept
{
    return idx_global != -1 || idx_local != -1;
}

constexpr bool is_occluded(std::pair<int, int> const& occluder_info) noexcept
{
    return is_occluded(occluder_info.first, occluder_info.second);
}

//--------------------------------------------------------------------------------------------------
} // namespace
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
std::shared_ptr<RenderStage> makeVoxelOcclusionStage()
{
    return std::make_shared<VoxelOcclusionStage>();
}

//--------------------------------------------------------------------------------------------------
void VoxelOcclusionStage::initializeStage(const Bitstream& bitstream, RenderItemStore* itemStore,
    const StreamManager* streamManager, const SceneState* state)
{
    EAR_CHECK(state != nullptr, "VoxelOcclusionStage: Invalid SceneState");

    _extent_item_store =
        std::make_shared<extent_item_storage>(itemStore, Config::getInt("EQNBands"));
    EAR_CHECK(_extent_item_store && _extent_item_store->valid(),
        "VoxelOcclusionStage: Invalid extent-item storage");

    _voxel_traversal = std::make_unique<Voxel::voxel_traversal>();
    EAR_CHECK(_voxel_traversal != nullptr, "VoxelOcclusionStage: Failed to create voxel-traversal");

    auto const voxel_scene_state = state->getVoxelSceneState();
    EAR_CHECK(voxel_scene_state != nullptr, "VoxelOcclusionStage: Invalid VoxelSceneState");

    // Index storage and DV-map are sub-scene independent
    _index_storage = voxel_scene_state->get_index_storage();
    EAR_CHECK(_index_storage != nullptr, "VoxelOcclusionStage: Invalid voxel index storage");

    _dv_map = voxel_scene_state->get_direct_view_map();
    EAR_CHECK(_dv_map != nullptr, "VoxelOcclusionStage: Invalid direct-view map");

    _compensation_eq = voxel_scene_state->get_eq_map();
    EAR_CHECK(_compensation_eq != nullptr, "VoxelOcclusionStage: Invalid EQ map");

    auto const materials = std::dynamic_pointer_cast<EIFPayload const>(
        bitstream.getOnlyPayloadOfType(MPEGI::Bitstream::PLD_SCENE))
                               ->getAcousticMaterials();

    /// @todo This should be sub-scene dependent
    extract_transmission_coefficients(
        materials, voxel_scene_state->get_current_subscene()->scene_materials);
    init(*voxel_scene_state);

    voxel_scene_state->on_subscene_change(
        [this, materials](Voxel::VoxelSceneState::const_shared const& voxel_scene_state) {
            extract_transmission_coefficients(
                materials, voxel_scene_state->get_current_subscene()->scene_materials);
            init(*voxel_scene_state);
        });
    auto portal_payloads = bitstream.getAllPayloadsOfType(MPEGI::Bitstream::PLD_PORTAL);
    EAR_CHECK_REL(portal_payloads.size() <= 1, "Too many portal payloads in bitstream");
    if (!portal_payloads.empty()) {
        is_explicit_portal_mode = std::dynamic_pointer_cast<PortalPayload const>(portal_payloads[0])
                                      ->getExplicitPortalMode();
    }
};

//--------------------------------------------------------------------------------------------------
void VoxelOcclusionStage::init(Voxel::VoxelSceneState const& voxel_scene_state)
{
    _scene = voxel_scene_state.get_scene_info_basic();
    EAR_CHECK(_scene != nullptr, "VoxelOcclusionStage: Invalid voxel scene info");

    _grid = voxel_scene_state.get_grid();
    EAR_CHECK(_grid != nullptr, "VoxelOcclusionStage: Invalid grid");

    // Occlusion recovery
    _occlusion_recovery = std::make_shared<occlusion_recovery>();
    _occlusion_recovery->init(voxel_scene_state.get_current_subscene()->renderer_instructions);

    // Maximum distance for extent
    EAR::Vector3 p0(_scene->p0.data());
    EAR::Vector3 p1(_scene->p1.data());
    _extent_max_distance = (p1 - p0).abs();
}

//--------------------------------------------------------------------------------------------------
void VoxelOcclusionStage::extract_transmission_coefficients(
    std::vector<EIFPayload::AcousticMaterialDesc> const& materials,
    Voxel::scene_material_map_t& material_map)
{
    auto const p_bands = Config::getFloatPtr("EQCenterFreqs");
    std::vector<float> const bands(p_bands, p_bands + Config::getInt("EQNBands"));

    // Material IDs start with 1. ID 0 (sound-medium) is fully transparent
    _transmission[0].assign(bands.size(), 1.f);

    for (auto& [m_id, mat_desc] : material_map) {

        auto const m_bs_id = mat_desc.bs_id;

        auto const it = std::find_if(std::cbegin(materials), std::cend(materials),
            [m_bs_id](auto const& m) { return m.bsId == m_bs_id; });

        if (std::cend(materials) == it) {
            continue;
        }

        EAR::Material ear_mat;
        for (auto const& mat : it->frequencies) {
            ear_mat.append(mat.f, mat.r, mat.s, mat.t, mat.c);
        }

        auto& t = _transmission[m_id];
        t.reserve(bands.size());
        t.clear();

        for (auto const& b : bands) {
            t.push_back(ear_mat.getTransmissionCoeffForFreq(b));
        }

        // Store transmission coefficients to make it available to other stages
        mat_desc.trans = t;
    }
}

//--------------------------------------------------------------------------------------------------
std::vector<float> const& VoxelOcclusionStage::get_transmission_for_voxel(
    Voxel::voxel_index_t const& idx) const
{
    return get_transmission_for_voxel(_grid->at(idx));
}

//--------------------------------------------------------------------------------------------------
std::vector<float> const& VoxelOcclusionStage::get_transmission_for_voxel(
    Voxel::voxel const& vxl) const
{
    /// @todo Only apply if the material is relevant for the current source
    auto const m_id = vxl.get_material_id();
    if (!_transmission.count(m_id))
        return _transmission.at(0); // Unknown material ID, treat as transparent

    return _transmission.at(m_id);
}

//--------------------------------------------------------------------------------------------------
bool VoxelOcclusionStage::apply_transmission_for_voxel(
    Voxel::voxel const& vxl, render_item_ptr const& item) const
{
    auto const is_sound_medium = vxl.is_sound_medium();
    if (!is_sound_medium) {
        item->addEq(get_transmission_for_voxel(vxl).data());
    }

    return !is_sound_medium;
}

//--------------------------------------------------------------------------------------------------
std::vector<float> VoxelOcclusionStage::calc_transmission(
    EAR::Vector3 const& user_xyz, occluder_idx_t const& occluder_idx, int item_idx) const
{
    // We only consider occluding materials closest to the listener and the following cases:
    // 1: Both local and global, but local is closer to the listener
    // 2: Global only
    // 3: Local only

    auto const idx_local = occluder_idx.first;
    auto const idx_global = occluder_idx.second;
    auto const& path = _voxel_traversal->get_path();

    if (!is_occluded(idx_local, idx_global)) {
        return _transmission.at(0); // not occluded
    }

    auto const has_local = has_local_occlusion(idx_local, idx_global);
    auto const has_global = has_global_occlusion(idx_global);

    std::vector<float> trans;
    if (has_global && has_local) {
        // Case 1
        auto const& trans_local = get_transmission_for_voxel(path[idx_local]);
        auto const& trans_global = get_transmission_for_voxel(path[idx_global]);

        // Combine local and global transmission
        std::transform(std::cbegin(trans_global), std::cend(trans_global), std::cbegin(trans_local),
            std::back_inserter(trans), [](auto const& lhs, auto const& rhs) { return lhs * rhs; });

        // Apply recovery (fades towards global transmission with increasing distance)
        apply_recovery(user_xyz, idx_local, item_idx, trans_global, trans);

    } else {
        // Case 2/3: either local or global
        auto const idx = has_local ? idx_local : idx_global;
        trans = get_transmission_for_voxel(path[idx]);

        // Apply recovery (fades towards transparency with increasing distance)
        apply_recovery(user_xyz, idx, item_idx, _transmission.at(0), trans);
    }

    return trans;
}

//--------------------------------------------------------------------------------------------------
std::unique_ptr<RenderStage::DspData> VoxelOcclusionStage::update(RenderList& render_list)
{
    auto const& lis = render_list.listener;
    auto& istorage = *_index_storage;

    if (lis.hasMoved) {
        // Calculate voxel-index for listener (default clipping mode) and update in the storage
        (void)istorage.update_listener_voxel_index(_scene->to_voxel_grid(lis.position.location));
    }

    for (auto& item : render_list.items) {
        if (item->status == ItemStatus::Orphaned) {
            istorage.storage.erase(item->idx);
            continue;
        }

        if (!is_occludable(*item)) {
            continue;
        }

        auto const item_location =
            item->extent ? item->extent->getGlobalLocation() : item->position.location;

        (void)istorage.update_voxel_index(
            item->idx, _scene->to_voxel_grid(item_location, lis.position.location));

        // Check whether listener is inside an occluding material
        if (apply_transmission_for_voxel(_grid->at(istorage.get_listener_voxel_index()), item)) {
            continue;
        }

        if (!(item->type == ItemType::DirectionalReverb
                || item->subtype == ItemSubtype::CoupledSrc)) {

            // Check whether item is inside an occluding material
            if (apply_transmission_for_voxel(
                    _grid->at(istorage.get_voxel_index(item->idx)), item)) {
                continue;
            }

            // Find path between listener and source
            if (!run_voxel_traversal(lis.position.location, item_location, item->idx)) {
                // Did not find path (should never happen) - set dv-flag to true to disable
                // voxel-diffraction for this render-item
                _dv_map->direct_view(item->idx, true);
                continue;
            }

            // Only occluders closest to the listener are considered
            auto const occluder_idx = find_nearest_occluders();
            auto const l_idx = occluder_idx.first;
            auto const g_idx = occluder_idx.second;

            // Only check global occluder to signal whether diffraction is needed for this item
            _dv_map->direct_view(item->idx, g_idx < 0);

            // Calculate transmission based on occluding material combinations and listener position
            item->addEq(calc_transmission(lis.position.location, occluder_idx, item->idx).data());
        }

        // Handling of items with source-extent
        if (item->extent) {
            auto const user_object_dist = (lis.position.location - item_location).abs();

            if (item->status == ItemStatus::Inactive
                // Only consider source-extent within maximum distance
                || (_scene->id.has_value() && user_object_dist > _extent_max_distance)) {
                _extent_item_store->remove(item->idx);
                continue;
            }

            if (!item->rayHits.empty()) {
                // Get extent representative (created if it does not exist yet). Ray-hits are transferred to secondary item
                auto& extent_item =
                    *_extent_item_store->update_item(item, 0, item->position.location);

                // Remove ray-hits from the primary item
                item->rayHits.clear();
                item->changed &= ~ItemProperty::RayHits;

                // For "regular" extent we just use the location of the extent for occlusion check.
                if (item->type == ItemType::DirectionalReverb
                    || item->subtype == ItemSubtype::CoupledSrc) {
                    update_ray_hits(
                        extent_item, lis.position, item->subtype == ItemSubtype::CoupledSrc);
                    if (is_explicit_portal_mode) {
                        item->addGain(extent_item.gain);
                    } else {
                        item->setGain(extent_item.gain);
                    }
                }
            } else {
                _extent_item_store->remove(item->idx);
            }
        }
    }

    // Add extent-representatives to render list, if any
    _extent_item_store->add_items(render_list);
    return {};
};

//--------------------------------------------------------------------------------------------------
bool VoxelOcclusionStage::run_voxel_traversal(
    EAR::Vector3 const& lis_xyz, EAR::Vector3 const& src_xyz, int item_idx)
{
    auto const lis_xyz_vox = Voxel::opengl_to_vox(lis_xyz.data());
    auto const& lis_idx = _index_storage->get_listener_voxel_index();

    auto const src_xyz_vox = Voxel::opengl_to_vox(src_xyz.data());
    auto const& src_idx = _index_storage->get_voxel_index(item_idx);

    return _voxel_traversal->run(*_scene, lis_xyz_vox, lis_idx, src_xyz_vox, src_idx);
}

//--------------------------------------------------------------------------------------------------
auto VoxelOcclusionStage::find_nearest_occluders() const -> occluder_idx_t
{
    auto occluder_idx = std::make_pair(-1, -1);

    auto& idx_local = occluder_idx.first;
    auto& idx_global = occluder_idx.second;
    auto const& grid = *_grid;
    auto const& path = _voxel_traversal->get_path();

    for (int i = 1; i < path.size() - 1; ++i) {
        auto const& v{ grid.at(path[i]) };

        // Can't be both local and global occluder at the same time
        if (idx_global == -1 && v.is_global_occluder()) {
            idx_global = i;
            if (idx_local != -1)
                break;
        } else if (idx_local == -1 && v.is_local_occluder()) {
            idx_local = i;
            if (idx_global != -1)
                break;
        }
    }

    return occluder_idx;
}

//--------------------------------------------------------------------------------------------------
void VoxelOcclusionStage::apply_recovery(EAR::Vector3 const& user_xyz, int idx, int item_idx,
    std::vector<float> const& t0, std::vector<float>& t1) const
{
    auto const& path = _voxel_traversal->get_path();

    // Occluder position (voxel mid-point)
    auto const occluder_xyz = EAR::Vector3(Voxel::vox_to_opengl(_scene->idx2xyz(path[idx])).data());
    // Occluder to listener (squared) distance
    auto const occluder_to_user_xyz = occluder_xyz - user_xyz;

    auto const sqr_magnitude = occluder_to_user_xyz.dot(occluder_to_user_xyz);
    auto const m_id = _grid->at(path[idx]).get_material_id();

    // Calculate compensation factor to limit energy, in case diffracted source is active
    auto& compensation_eq = *_compensation_eq;

    // Apply recovery, result is in t1
    _occlusion_recovery->apply_recovery(m_id, sqr_magnitude, t0, t1, compensation_eq[item_idx]);
}

//--------------------------------------------------------------------------------------------------
void VoxelOcclusionStage::update_ray_hits(
    RenderItem& extent_item, EAR::Position const& listener_pos, bool isCoupledSrc)
{
    auto const n_hits = extent_item.rayHits.size();

    auto ray_it = std::begin(extent_item.rayHits);
    while (ray_it != std::cend(extent_item.rayHits)) {
        // For implicit portal support the ray hit point is moved further from the listener,
        // This is done to ensure that the voxels making up the AE geometry do not occlude the portal.
        // For coupled sources this should not be done.
        auto const hit_xyz = listener_pos.location
            + EAR::Quaternion(listener_pos.orientation).rotate(SESS_RAY_DISTRIBUTION[ray_it->rayID])
                * (ray_it->distance + (isCoupledSrc ? 0 : 1) * _scene->voxel_size * M_SQRT2);

        (void)_index_storage->update_voxel_index(
            extent_item.idx, _scene->to_voxel_grid(hit_xyz, listener_pos.location));

        if (!run_voxel_traversal(listener_pos.location, hit_xyz, extent_item.idx)
            || is_occluded(find_nearest_occluders())) {
            ray_it = extent_item.rayHits.erase(ray_it);
        } else {
            ++ray_it;
        }
    }

    extent_item.setGain(std::sqrt(extent_item.rayHits.size() / static_cast<float>(n_hits)));
    extent_item.rayHits.shrink_to_fit();
    extent_item.changed |= (ItemProperty::RayHits | ItemProperty::Eq);
}

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
