/***************************************************************************
This software module was originally developed by

Dolby International AB

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Dolby International AB retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/
#pragma once
//--------------------------------------------------------------------------------------------------
#include <unordered_map>
#include <vector>

#include <MPEGI/Bitstream/EIFPayload.h>
#include <MPEGI/Helpers/OcclusionRecovery.h>
#include <MPEGI/Helpers/SecondaryItemStorage.h>
#include <MPEGI/Helpers/VoxelTraversal.h>
#include <MPEGI/Renderer/RenderStage.h>

//--------------------------------------------------------------------------------------------------
namespace MPEGI {
//--------------------------------------------------------------------------------------------------
class VoxelOcclusionStage : public RenderStage
{
protected:
    //@{
    /// Implementation of RendererStage Interface
    std::unique_ptr<RenderStage::DspData> update(RenderList&) override;
    void initializeStage(
        const Bitstream&, RenderItemStore*, const StreamManager*, const SceneState*) override;
    //@}

    //@{
    /**
     * Traverse voxels
     *
     * @param lis_xyz Listener location (x,y,z), OpenGL
     * @param src_xyz Source location (x,y,z), OpenGL
     * @param item_idx Render item index
     * @return true if path was found, else false
     */
    bool run_voxel_traversal(
        EAR::Vector3 const& lis_xyz, EAR::Vector3 const& src_xyz, int item_idx);
    //@}

    /**
     * Subscene dependent initialization
     *
     * @param voxel_scene_state The voxel-scene state
     */
    void init(Voxel::VoxelSceneState const& voxel_scene_state);

    /**
     * Extract transmission coefficients for current subscene
     *
     * @param materials List of *all* materials in the scene
     * @param material_map Material map id -> scene material
     */
    void extract_transmission_coefficients(
        std::vector<EIFPayload::AcousticMaterialDesc> const& materials,
        Voxel::scene_material_map_t& material_map);

    /**
     * Get the transmission coefficients for given voxel index
     *
     * @param idx Voxel index
     * @return Transmission coefficients
     */
    std::vector<float> const& get_transmission_for_voxel(Voxel::voxel_index_t const& idx) const;

    /**
     * Get the transmission coefficients for given voxel
     *
     * @param vxl The voxel
     * @return Transmission coefficients
     */
    std::vector<float> const& get_transmission_for_voxel(Voxel::voxel const& vxl) const;

    using render_item_ptr = std::unique_ptr<MPEGI::RenderItem>;

    /**
     * Apply transmission for given voxel and render item if voxel is not sound-medium
     *
     * @param vxl The voxel
     * @param item The render item
     * @return true if transmission was applied and therefore the voxel is not sound-medium, else false
     */
    bool apply_transmission_for_voxel(Voxel::voxel const& vxl, render_item_ptr const& item) const;

    /// Occluder indices - first: local, second: global
    using occluder_idx_t = std::pair<int, int>;

    /**
     * Calculate transmission EQs for occlusion
     *
     * @param user_xyz Listener location (x,y,z, OpenGL)
     * @param occluder_idx Occluder indices local/global
     * @param item_idx Render item index
     * @return Transmission EQs
     */
    std::vector<float> calc_transmission(
        EAR::Vector3 const& user_xyz, occluder_idx_t const& occluder_idx, int item_idx) const;

    /**
     * Find nearest occluding voxel from the listener position.
     * Must not be called if run_voxel_traversal(EAR::Vector3 const&, EAR::Vector3 const&,i nt) returns false, i.e. no path was found
     *
     * @return Returns a pair with nearest local (first) and global (second) index (offset into _path array, -1 no occluder found)
     */
    occluder_idx_t find_nearest_occluders() const;

    /**
     * Update ray-hits for items with source-extent
     *
     * @param extent_item The secondary source-extent item
     * @param listener_pos Current listener position
     */
    void update_ray_hits(
        RenderItem& extent_item, EAR::Position const& listener_pos, bool isCoupledSrc);

protected:
    Voxel::grid::shared _grid; //!< The voxel-grid extracted from the bitstream
    Voxel::voxel_scene_info_basic::shared
        _scene; //!< Basic scene information extracted from the bitstream
    Voxel::index_storage::shared _index_storage; //!< To store/share calculated voxel indices
    Voxel::direct_view_map::shared
        _dv_map; //!< Direct-view map: indicates whether there is a direct view line between source and listener

    using trans_map_t = std::unordered_map<int, std::vector<float>>;
    trans_map_t _transmission; //!< Transmission per material ID

    std::unique_ptr<Voxel::voxel_traversal> _voxel_traversal; //!< Voxel-traversal algorithm

    //@{
    /// Local occlusion recovery
    occlusion_recovery::shared _occlusion_recovery;

    /// Compensation EQs calculated here and applied in VoxelDiffractionStage
    Voxel::VoxelSceneState::eq_map_ptr _compensation_eq;

    /**
     * Apply occlusion recovery, i.e. occlusion effect fades after a certain distance
     *
     * @param user_xyz User location (x,y,z, OpenGL)
     * @param idx Index of the closest occluder in _path
     * @param item_idx Render item index
     * @param t0 Transmission to interpolate
     * @param t1 Transmission to interpolate
     */
    void apply_recovery(EAR::Vector3 const& user_xyz, int idx, int item_idx,
        std::vector<float> const& t0, std::vector<float>& t1) const;

    //@}

    struct extent_item_storage : public Voxel::secondary_item_storage<extent_item_storage>
    {
        using shared = std::shared_ptr<extent_item_storage>;
        extent_item_storage(RenderItemStore* store, unsigned int num_bands_eq)
            : secondary_item_storage(store, num_bands_eq)
        {}

        ItemProperty get_sync_flags(RenderItem const&) const
        {
            return ItemProperty::Directivity | ItemProperty::Gain | ItemProperty::Loudness
                | ItemProperty::Aparams | ItemProperty::Position | ItemProperty::Extent
                | ItemProperty::Signal | ItemProperty::RefDist | ItemProperty::Reverb
                | ItemProperty::RayHits;
        }

        ItemType get_item_type(RenderItem const& primary) const
        {
            return (primary.type == ItemType::DirectionalReverb) ? ItemType::DirectionalReverb
                                                                 : ItemType::Representative;
        }
        std::string get_suffix() const { return "_extent_"; }
    };

    /// Item storage for items with source-extent
    extent_item_storage::shared _extent_item_store;

    /// Maximum distance from the listener where items with source-extent are considered
    float _extent_max_distance;

    bool is_explicit_portal_mode{ false };
};

//--------------------------------------------------------------------------------------------------
} // namespace MPEGI
//--------------------------------------------------------------------------------------------------
