/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_HOATRANSLATOR_HPP
#define MPEGCORE_HOATRANSLATOR_HPP

#include <memory>
#include <optional>
#include <vector>

#include <EAR/BaseTypes.h>

#include "HoaTranslatorConfig.hpp"
#include <hoatranslation/MultiPointHoaTranslator.hpp>
#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/Config.hpp>
#include <mpegcore/gain.hpp>

namespace mpegcore {

struct HoaParametersConfig;

class HoaTranslator
{
public:
    struct Group
    {
        // Indices in Group are unique in the system, but HoaTranslator shouldn't expect them to
        // refer to any particular data collection in the system. This allows us to keep the data
        // handled by HoaTranslator independent of the way scenes store the data.

        // Index of each HoaSource in this group.
        std::vector<std::size_t> sourceIndices;

        // Positions of each HoaSource in this group (in EIF global coordinate space).
        std::vector<EAR::Position> sourcePositions;

        // Linear gains of each HoaSource in this group.
        std::vector<float> sourceLinearGains;

        // Index of this group's region.
        std::optional<std::size_t> regionIndex;

        // Number of channels in each source in this group.
        std::size_t channelsPerSource{ 0 };

        // Parameters used to reduce the complexity of MPHOA rendering
        std::vector<int> complexityReductionParameters;
    };

    struct InformedSourceConfig
    {
        // Positions of each audio object in this group, that are known to source aware processing
        // (in EIF global coordinate space).
        EAR::Position sourceAwareObjectPosition;
        bool listenerThresholdPresent;
        bool hoaSourceThresholdPresent;
        float listenerThreshold;
        float hoaSourceThreshold;
        int priorityValue;
    };

    struct HoaInput
    {
        // Index of this source, using the same definition of Index as in the Group.
        std::size_t sourceIndex;

        // Gains for each of the channels of this source (for SN3D).
        std::vector<GainLinear> channelGains;
    };

    ~HoaTranslator();

    void initialize(Config const& config, HoaTranslatorConfig const& hoaTranslatorConfig,
        HoaParametersConfig const& hoaParametersConfig, HoaTranslator::Group group,
        std::shared_ptr<std::vector<std::uint8_t>> spatialMetadata,
        std::vector<InformedSourceConfig> informedSourceConfigs, EAR::Vector3 rescale,
        EAR::Vector3 rescaleOffset);

    void initializeHoaInputs(std::size_t channelCount);

    void preProcess(EAR::Position const& listenerPosition);

    void render();

    void readToInput(float* data, std::size_t amount, std::size_t sourceId);
    void readToInput(
        float* data, std::size_t amount, std::size_t sourceId, std::size_t channelIndex);

    AudioBufferArray const& outputBuffer() const;

    void updateProcessedHoaInputs();
    std::vector<HoaInput*> const& getProcessedHoaInputs() { return m_processedHoaInputs; };

    bool isInitialized() const { return m_isInitialized; };

    Group const& group() const noexcept { return m_group; };

    void increaseFrameCounter() const { ++m_hoaTranslatorData->currentFrame; }

    hoaTranslatorData* getHoaTranslatorData() { return m_hoaTranslatorData.get(); }

    std::vector<std::size_t> getFullOrderHOASourceIndices();
    std::vector<std::size_t> getFirstOrderHOASourceIndices();
    int getHOAOrder();

private:
    Group m_group;
    HoaTranslatorConfig m_config;
    std::vector<InformedSourceConfig> m_informedSourceConfigs;
    std::unique_ptr<hoaTranslatorData> m_hoaTranslatorData;
    AudioBufferArray m_outputBuffer;
    std::vector<HoaInput> m_hoaInputs;
    std::vector<HoaInput*> m_processedHoaInputs;
    bool m_isInputESD;
    bool m_isInputSN3D;
    bool m_isInitialized{ false };
    micArrayPositionsAndOrientations vectorizeMicArrays(
        Group const& group, EAR::Vector3 rescale, EAR::Vector3 rescaleOffset);
    std::vector<std::vector<float>> vectorizeSourceAwareObjects(
        std::vector<sourceAwareParametersPerSource>& sourceAwareParamsPerSource);
};

} // namespace mpegcore

#endif // MPEGCORE_HOATRANSLATOR_HPP
