/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_ROTATION_HPP
#define MPEGCORE_ROTATION_HPP

#include <array>

#include "algebra/lerp.hpp"

namespace mpegcore {

// Defined in w18159_(Encoder_Input_Format).docx
//
// Yaw-pitch-roll rotations are specific convention of Euler angles (i.e.
// rotations are three axis). Fig. 5 shows the three different axis rotations.
// Yaw-pitch-roll rotations are intrinsic, i.e. each rotation is relative to the
// previously rotated axis. This is achieved by executing the axis rotations in
// the following order
//
//     1. roll = rotation around the back axis (+z in OpenGL conventions)
//     2. pitch = rotation around the right axis (+x in OpenGL conventions)
//     3. yaw = rotation around the up axis (+y in OpenGL conventions)

struct Rotation
{
    float yawDeg{ 0 };
    float pitchDeg{ 0 };
    float rollDeg{ 0 };

    Rotation() = default;

    Rotation(float yawDeg, float pitchDeg, float rollDeg) noexcept
        : yawDeg{ yawDeg }
        , pitchDeg{ pitchDeg }
        , rollDeg{ rollDeg }
    {}

    Rotation(std::array<float, 3> const& array)
        : yawDeg{ array[0] }
        , pitchDeg{ array[1] }
        , rollDeg{ array[2] }
    {}
};

// Strict equality check, not considering float issues.
bool operator==(Rotation const& lhs, Rotation const& rhs) noexcept;
bool operator!=(Rotation const& lhs, Rotation const& rhs) noexcept;
Rotation operator+(Rotation const& lhs, Rotation const& rhs) noexcept;
Rotation operator-(Rotation const& lhs, Rotation const& rhs) noexcept;

bool areClose(Rotation const& r1, Rotation const& r2, float epsilon) noexcept;

namespace {

// Return `value` wrapped to range `[-MAX_VALUE, MAX_VALUE]`.
template <int MAX_VALUE> float toRange(float value) noexcept
{
    constexpr float max{ MAX_VALUE };
    while (value > max) {
        value -= 2 * max;
    }
    while (value < -max) {
        value += 2 * max;
    }
    return value;
}

// Return linear interpolation of yaw or roll angle in between `start` and `end`
// based on `targetTimeS` in `domain`.
template <typename TDomain>
constexpr float lerpYawOrRoll(
    float start, float end, TDomain const& domain, double targetTimeS) noexcept
{
    return toRange<180>(lerp(algebra::Range<float>{ start, end }, domain, targetTimeS));
}

// Return linear interpolation of pitch angle in between `start` and `end` based
// on `targetTimeS` in `domain`.
template <typename TDomain>
constexpr float lerpPitch(
    float start, float end, TDomain const& domain, double targetTimeS) noexcept
{
    return toRange<90>(lerp(algebra::Range<float>{ start, end }, domain, targetTimeS));
}

} // namespace

// Return linear interpolation of rotation in between `start` and `end` based on
// `targetTimeS` in `domain`.
template <typename TDomain>
constexpr Rotation lerp(
    Rotation const& start, Rotation const& end, TDomain const& domain, double targetTimeS) noexcept
{
    return Rotation{
        lerpYawOrRoll(start.yawDeg, end.yawDeg, domain, targetTimeS),
        lerpPitch(start.pitchDeg, end.pitchDeg, domain, targetTimeS),
        lerpYawOrRoll(start.rollDeg, end.rollDeg, domain, targetTimeS),
    };
}

} // namespace mpegcore

#endif // MPEGCORE_ROTATION_HPP
