/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_DATATYPES_REVERB_HPP
#define MPEGCORE_DATATYPES_REVERB_HPP

#include <array>
#include <chrono>
#include <cstddef>
#include <cstdint>
#include <iostream>
#include <optional>
#include <utility>
#include <vector>

namespace mpegcore {

struct GraphicEqParameters
{
    using Coefficients = std::vector<float>;
    std::array<Coefficients, 2> a;
    std::array<Coefficients, 2> b;
    Coefficients stageF;
    float F{ 0.F };
    float levelDb{ 0.F };

    enum class Type : std::uint8_t
    {
        uninitialized, // Use to initialize an uninitialized parameter set.
        cascade, // With all stage gains at 1.0.
        cascadeStaged, // With different stage gains (e.g. low-pass).
        gain, // Just `F`, no stages.
        parallel,
        silent,
    };
    Type type{ Type::uninitialized };

    GraphicEqParameters();
    GraphicEqParameters(Type type, std::size_t bandCount);

    std::size_t bandCount() const noexcept;
};

struct ReverberatorCommon
{
    struct DelayLineParams
    {
        std::size_t delayInSamples;
        float azimuthDeg;
        float elevationDeg;
        GraphicEqParameters filterParams;
        std::optional<GraphicEqParameters> direcGainFilterParams;
    };

    using DelayLineGroup = std::vector<DelayLineParams>;
};

// Pre-delay in samples and the length of the shortest delay line
using Predelay = std::pair<std::size_t, std::size_t>;

struct ReverberatorParameters : ReverberatorCommon
{
    DelayLineGroup delayLines;
    std::optional<GraphicEqParameters> eqFilterParams;

    // Pre-delay here has been adjusted so that it has absorbed as much of the delay line lengths of
    // the reverberator as possible. Not used in collab renderers.
    std::chrono::nanoseconds preDelayAdjusted{ 0 };

    // Pre-delay here has been adjusted so that it has absorbed as much of the
    // delay line lengths of the reverberator as possible. Integer number of samples and
    // the minimum delay line length.
    Predelay preDelayAdjustedInSamples{ 0, 0 };

    // Whereas the diffuseOnsetSec of an analyzed RIR is from t0, the FDN's diffuseOnsetSec is from
    // the beginning of the FDN response. I.e. the FDN is expected to receive the source signal with
    // no propagation delay.
    float fdnDiffuseOnsetSec{ 0.f };

    enum class InputType : std::uint8_t
    {
        direct,
        directAndEarlyReflections,
    };
    InputType inputType;
};

// Return the minimum of the delays in the delay line group.
std::chrono::nanoseconds minDelay(
    ReverberatorCommon::DelayLineGroup const& delayLineGroup, std::uint32_t sampleRateHz) noexcept;

std::size_t minDelaySamp(ReverberatorCommon::DelayLineGroup const& delayLineGroup) noexcept;

// Status posting for feedback during render or debugging
extern bool postStatus;
void printStatus(const std::string& message);
template <typename... Args> void printStatus(const std::string& formatMessage, Args&&... args);

} // namespace mpegcore

#endif // MPEGCORE_DATATYPES_REVERB_HPP
