/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_BYTESTREAM_HPP
#define HOA_TRANSLATION_BYTESTREAM_HPP

#include <cstddef>
#include <cstdint>
#include <fstream>
#include <memory>
#include <string>
#include <vector>

namespace hoatranslation {

// A big-endian byte oriented stream for reading binary spatial metadata
//
// Full stream is read into a memory buffer at construction so that accesses don't go to mass storage.
class ByteStream {
public:
   // Construct a byte stream initialized by from a vector
   ByteStream(std::shared_ptr<std::vector<std::uint8_t>> data);

    // Return an integer read from the stream. The amount of consumed data is the size of the
    // returned value.
    template<typename T> T read()
    {
        return static_cast<T>(readBytes(sizeof(T)));
    }

    // Read a value from the stream to `value`. The amount of consumed data is the size of `value`.
    template<typename T> void read(T& value)
    {
        value = static_cast<T>(readBytes(sizeof(T)));
    }

    // Read a value from the stream and push it to the end of `container`
    template<typename T> void readToBack(T& container) {
        container.push_back(read<typename T::value_type>());
    }

    // Return a spatial metadata string read from the stream
    std::string readString();

    // Read a spatial metadata string from the stream to `string`. Return the length of the string
    // as a convenience.
    std::size_t read(std::string& string);

    // Return the current byte read position of the stream
    std::size_t bytePosition() const;

    // Set the current byte read position of the stream
    void setBytePosition(std::size_t position);

    // Move the current byte read position of the stream forward by `byteCount`
    void skipBytes(std::size_t byteCount);

private:
    // Return `byteCount` bytes read from the stream. The bytes are the lowest bytes of the returned
    // value. The requested `byteCount` should not be bigger than can fit into the return value.
    std::uint64_t readBytes(std::size_t byteCount);

    std::shared_ptr<std::vector<std::uint8_t>> m_data;
    std::size_t m_bytePosition{0};
};

} // namespace hoatranslation

#endif // HOA_TRANSLATION_BYTESTREAM_HPP
