/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_CROSSFADE_HPP
#define HOA_TRANSLATION_CROSSFADE_HPP

#include "Utilities.hpp"

/* Crossfade two mic array signals, when the listener moves closer to a new array. The crossfade is done separately on frequency
 * bands, in a pre-determined random order. The process goes as follows:
 *  - one crossfade operation per each subframe (i.e. for each processing frame, do nFrames crossfade operations)
 *  - fadeOut frequency bands are updated every subframe
 *  - for fadeOut bands, use signals from the old closest mic array
 *  - for the remaining frequency bands, use signals from the new closest array
 *  - the randomized order for the frequency bands progression is determined by crossfadeBandSwaporder
 *  - for example, for crossFadeProgressIndx = 2, the bands outside fadeOut bands (= fadeIn bands) range is determined by:
 *    [ crossfadeBandLow[idx_0], crossfadeBandHigh[idx_0] ],
 *    [ crossfadeBandLow[idx_1], crossfadeBandHigh[idx_1] ]
 *    and [ crossfadeBandLow[idx_2], crossfadeBandHigh[idx_2] ],
 *    where idx_0 = crossfadeBandSwaporder[0], idx_1 = crossfadeBandSwaporder[1], idx_2 = crossfadeBandSwaporder[2],
 *    and crossfadeBandLow and crossfadeBandHigh determine the lower and upper freqeuncy band limits
 *  - during crossfade, the interpolation weights are treated as follows:
 *      * if both arrays are in the same triangle: use the most-up-to-date weights
 *      * if the arrays are in different triangles (i.e. the crossfade is done during a triangle switch):
 *          -> use the last weights in the old triangle for fadeOut bands
 *          -> for fadeIn bands, use the most up-to-date weights in the new triangle
 *
 * More info in doc/Band_by_band_crossfade_for_closest_mic_arrays.doc
 *
 * Output:
 *      arraysPartOfCrossfade    - mic arrays in the triangles, that are part of the crossfade
 *                                 (i.e. 3 arrays, when the crossfade is done within a triangle;
 *                                  6 arrays if the crossfade is between arrays from different triangles)
 */
void start_crossfade(hoaTranslatorData *h);

/* Get fade-out bands for crossfade process. In fade-out frequency bands, the HOA signals are obtained from the previously
 * closest mic array. Also increase progress index for crossfade (crossFadeProgressIndx).
 *
 * Output:
 *      fadeOutBands    - frequency bands, where the signal is used from the faded out mic array
 */
void get_fade_out_bands(hoaTranslatorData *h);

/* Check if crossfade is currently being processed:
 *  - if crossfade is active, get fadeOut frequency bands for the current processing frame
 *  - otherwise, determine the closest mic array, and start the crossfade process, if the closest array is switched
 */
void check_for_crossfade_start(hoaTranslatorData *h);

// Stop crossfade, when all the frequency bands have been processed
void check_for_crossfade_stop(hoaTranslatorData *h);

#endif //HOA_TRANSLATION_CROSSFADE_HPP
