/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "DEMUX.hpp"

#include <algorithm>
#include <cassert>
#include <cfloat>
#include <cmath>
#include <cstdint>
#include <memory>
#include <string>

#include "ByteStream.hpp"
#include "SpatialAnalysis.hpp"

void read_spatial_metadata_info(hoatranslation::ByteStream& stream, decoderData* d)
{
    // SpatialMetadataHeader()
    stream.read(d->audio_frame_duration);
    stream.read(d->subframes_duration);
    stream.read(d->sampling_rate);
    stream.read(d->num_frequency_bands);
    stream.read(d->useRecursiveAveraging);
    stream.read(d->num_hoa_groups);
    stream.read(d->num_frames);
    stream.read(d->azimuth_multiplier);
    stream.read(d->elevation_multiplier);
    stream.read(d->dtr_multiplier);
    stream.read(d->energy_multiplier);
    stream.read(d->position_multiplier);
    stream.read(d->rotation_multiplier);
    stream.read(d->exterior_multiplier);
    stream.read(d->exterior_projection_radius);
    stream.read(d->is_audio_uneven);
    d->bitStreamHeaderSize  = stream.bytePosition();


    if (d->useRecursiveAveraging) {
        d->subframes_per_frame = 1;
    } else {
        d->subframes_per_frame = static_cast<std::uint8_t>(d->audio_frame_duration / d->subframes_duration);
    }

    d->sourceMetadataFrameSize = (sizeof(d->hoa_render_meta_azimuth) + sizeof(d->hoa_render_meta_elevation) + sizeof(d->direct_to_total_energy_ratio) + sizeof(d->overall_energy)) * d->num_frequency_bands * d->subframes_per_frame;

    // static HOAGroupInfo (all HOAGroup ids and ids and HOA orders of sources in the groups
    d->hoa_source_group_id_lengths.resize(d->num_hoa_groups);
    d->hoaSourceGroupIdStrs.resize(d->num_hoa_groups);
    d->num_hoa_sources_in_group.resize(d->num_hoa_groups);
    d->groupMetadataFrameSizes.resize(d->num_hoa_groups);
    d->hoa_source_id_lengths.resize(d->num_hoa_groups);
    d->hoaSourceIdStrs.resize(d->num_hoa_groups);
    d->hoa_orders.resize(d->num_hoa_groups);

    for (int hoaGrp = 0; hoaGrp < d->num_hoa_groups; ++hoaGrp) {
        // HOAGroupData()
        auto const groupDataHeaderStartPosition = stream.bytePosition();

        d->hoa_source_group_id_lengths[hoaGrp] = stream.read(d->hoaSourceGroupIdStrs[hoaGrp]);
        stream.read(d->num_hoa_sources_in_group[hoaGrp]);

        d->groupDataHeaderSizes.push_back(stream.bytePosition() - groupDataHeaderStartPosition);

        d->groupMetadataFrameSizes[hoaGrp] = d->num_hoa_sources_in_group[hoaGrp] * d->sourceMetadataFrameSize;

        for (int hoaSrc = 0; hoaSrc < d->num_hoa_sources_in_group[hoaGrp]; ++hoaSrc) {
            // HOASourceInformation()
            auto const sourceId = stream.readString();
            d->hoa_source_id_lengths[hoaGrp].push_back(sourceId.size());
            d->hoaSourceIdStrs[hoaGrp].push_back(sourceId);

            // HOASourceData()
            stream.readToBack(d->hoa_orders[hoaGrp]);
        }
    }

    d->bitStreamHeaderSize  = stream.bytePosition();

    // InfoFrame()
    auto const infoFrameStartPosition = stream.bytePosition();
    auto const num_hoa_groups = stream.read<decltype(d->num_hoa_groups)>();
    assert(num_hoa_groups == d->num_hoa_groups); // dynamic number of groups not handled

    d->hoa_sources_pos_x.resize(num_hoa_groups);
    d->hoa_sources_pos_y.resize(num_hoa_groups);
    d->hoa_sources_pos_z.resize(num_hoa_groups);
    d->hoa_sources_rot_yaw.resize(num_hoa_groups);
    d->hoa_sources_rot_pitch.resize(num_hoa_groups);
    d->hoa_sources_rot_roll.resize(num_hoa_groups);

    for (int hoaGrp = 0; hoaGrp < num_hoa_groups; ++hoaGrp) {
        // hoa group id
        auto hoaGroupId = stream.readString();
        assert(hoaGroupId == d->hoaSourceGroupIdStrs[hoaGrp]);  // Hoa group name should be the same

        // number of sources
        auto const numSources = stream.read<typename decltype(d->num_hoa_sources_in_group)::value_type>();
        assert(numSources == d->num_hoa_sources_in_group[hoaGrp]); // Changing number of sources not handled

        for (int hoaSrc = 0; hoaSrc < d->num_hoa_sources_in_group[hoaGrp]; ++hoaSrc) {
            // HOASourceInformation()
            auto const hoaSourceId = stream.readString();
            assert(hoaSourceId == d->hoaSourceIdStrs[hoaGrp][hoaSrc]);  // Hoa source name should be the same
            // HOASourcePosition()
            stream.readToBack(d->hoa_sources_pos_x[hoaGrp]);
            stream.readToBack(d->hoa_sources_pos_y[hoaGrp]);
            stream.readToBack(d->hoa_sources_pos_z[hoaGrp]);
            stream.readToBack(d->hoa_sources_rot_yaw[hoaGrp]);
            stream.readToBack(d->hoa_sources_rot_pitch[hoaGrp]);
            stream.readToBack(d->hoa_sources_rot_roll[hoaGrp]);
        }
    }

    d->infoFrameSize  = stream.bytePosition() - infoFrameStartPosition;
}

void read_spatial_metadata_info(hoaTranslatorData *h,
                                bool assertValues)
{
    decoderData* d = &h->decoderData1;

    read_spatial_metadata_info(*h->spatialMetadataByteStream, d);

    // Check that the encoded values match with the translator counterparts
    if (assertValues) {
        assert(d->useRecursiveAveraging == h->useRecursiveAveraging);
        assert(d->sampling_rate == h->sampleRate);
        assert(d->num_hoa_sources_in_group[h->HOAGroupId] == h->nMicrophoneArrays);
        float errorMargin = 0.0001f;
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            float micX = d->hoa_sources_pos_x[h->HOAGroupId][mic] / (float)d->position_multiplier;
            float micY = d->hoa_sources_pos_y[h->HOAGroupId][mic] / (float)d->position_multiplier;
            float micZ = d->hoa_sources_pos_z[h->HOAGroupId][mic] / (float)d->position_multiplier;
            float micYaw = d->hoa_sources_rot_yaw[h->HOAGroupId][mic] / (float)d->rotation_multiplier;
            float micPitch = d->hoa_sources_rot_pitch[h->HOAGroupId][mic] / (float)d->rotation_multiplier;
            float micRoll = d->hoa_sources_rot_roll[h->HOAGroupId][mic] / (float)d->rotation_multiplier;
            assert(AnearB(micX, h->microphoneArrayPositions[mic][0], errorMargin));
            assert(AnearB(micY, h->microphoneArrayPositions[mic][1], errorMargin));
            assert(AnearB(micZ, h->microphoneArrayPositions[mic][2], errorMargin));
            assert(AnearB(micYaw, h->micArrayOrientations[mic][0], errorMargin));
            assert(AnearB(micPitch, h->micArrayOrientations[mic][1], errorMargin));
            assert(AnearB(micRoll, h->micArrayOrientations[mic][2], errorMargin));
        }
        if (h->useRecursiveAveraging) {
            assert(d->subframes_per_frame == 1);
        } else {
            assert(d->subframes_per_frame == h->nFrames);
        }
        assert(d->num_frequency_bands == h->nBands);
    }
}

void read_spatial_metadata_from_binary(hoaTranslatorData *h)
{
    auto copy_parameters_to_sub_frames = [&](auto const& mic) {
        for (int outputFrame = 1; outputFrame < h->nFrames; ++outputFrame) {
            std::copy_n(&h->azimuthData[mic][0][0], h->nBands, &h->azimuthData[mic][outputFrame][0]);
            std::copy_n(&h->elevationData[mic][0][0], h->nBands, &h->elevationData[mic][outputFrame][0]);
            std::copy_n(&h->DTRData[mic][0][0], h->nBands, &h->DTRData[mic][outputFrame][0]);
            std::copy_n(&h->energyData[mic][0][0], h->nBands, &h->energyData[mic][outputFrame][0]);
            if (h->useSphoa6DofExtensionMode) {
                std::copy_n(&h->singlePoint6DofDirGains[0][0], h->nBands, &h->singlePoint6DofDirGains[outputFrame][0]);
            }
        }
    };

    auto calculate_spatial_parameters_from_listener_to_projected_source = [&](int mic, int frame, int band, std::vector<float> &intensityVector,
                                                                              std::vector<float> &newDoaVector, std::vector<float> &interpolatedListenerPos) {
        // Calculate DOA angles from a new DOA vector pointing from the listener
        // to a projected source position. DTR and energy are calculated from the
        // original intensity vector.
        calculate_doa_vector_cartesian(h->azimuthData[mic][frame][band], h->elevationData[mic][frame][band], intensityVector);
        calculate_new_doa_vector(h, intensityVector.data(), newDoaVector, interpolatedListenerPos);
        calculate_sphoa_dir_gain(h, frame, band, newDoaVector);
        h->azimuthData[mic][frame][band]   = calculate_azimuth(newDoaVector);
        h->elevationData[mic][frame][band] = calculate_elevation(newDoaVector);
    };

    auto read_mphoa_spatial_metadata = [&]() {
        decoderData* d = &h->decoderData1;
        auto& stream = *h->spatialMetadataByteStream;
        for (int mic = 0; mic < d->num_hoa_sources_in_group[h->HOAGroupId]; mic++) {
            if (std::find(h->activeArraysForCurrentFrame.begin(), h->activeArraysForCurrentFrame.end(), mic) != h->activeArraysForCurrentFrame.end()) {
                //SpatialMetadataFrame()
                for (int subframe = 0; subframe < d->subframes_per_frame; subframe++) {
                    for (int band = 0; band < d->num_frequency_bands; band++) {
                        stream.read(d->hoa_render_meta_azimuth);
                        stream.read(d->hoa_render_meta_elevation);
                        stream.read(d->direct_to_total_energy_ratio);
                        stream.read(d->overall_energy);
                        h->azimuthData[mic][subframe][band] =
                            (float)d->hoa_render_meta_azimuth / (float)d->azimuth_multiplier * M_PI / 180.0f;
                        h->elevationData[mic][subframe][band] =
                            (float)d->hoa_render_meta_elevation / (float)d->elevation_multiplier * M_PI / 180.0f;
                        h->DTRData[mic][subframe][band] = (float)d->direct_to_total_energy_ratio / (float)d->dtr_multiplier;
                        h->energyData[mic][subframe][band] = ((float)d->overall_energy / (float)d->energy_multiplier) * h->sourceEnergyGains[mic];
                    }
                }
                if (h->useRecursiveAveraging) {
                    // In recursive averaging mode, same spatial values and energies are used in
                    // each sub-frame of the processing frame.
                    copy_parameters_to_sub_frames(mic);
                }
            } else {
                // skip over non-active sources
                stream.skipBytes(d->sourceMetadataFrameSize);
            }
        }
    };

    auto read_sphoa_spatial_metadata = [&](std::vector<float> &intensityVector, std::vector<float> &newDoaVector, std::vector<float> &interpolatedListenerPos) {
        decoderData* d = &h->decoderData1;
        auto& stream = *h->spatialMetadataByteStream;
        // In sphoa 6DoF extension mode the translator data arrays are initialized for only single
        // mic arrays. For example, the size of h->azimuthData here is something like 1 x nFrames
        // x nBands, so we can only access the first mic in the array.
        std::size_t const mic{ 0 };

        // seek to start of the nth encoded hoa
        int encodedMic = h->NthSphoaInEncodedMetadata - 1;
        stream.skipBytes(d->sourceMetadataFrameSize * encodedMic);
        //SpatialMetadataFrame()
        for (int subframe = 0; subframe < d->subframes_per_frame; subframe++) {
            get_interpolated_listener_position_in_array_space(h, mic, subframe, interpolatedListenerPos);
            for (int band = 0; band < d->num_frequency_bands; band++) {
                stream.read(d->hoa_render_meta_azimuth);
                stream.read(d->hoa_render_meta_elevation);
                stream.read(d->direct_to_total_energy_ratio);
                stream.read(d->overall_energy);
                h->azimuthData[mic][subframe][band] =
                    (float)d->hoa_render_meta_azimuth / (float)d->azimuth_multiplier * M_PI / 180.0f;
                h->elevationData[mic][subframe][band] =
                    (float)d->hoa_render_meta_elevation / (float)d->elevation_multiplier * M_PI / 180.0f;
                h->DTRData[mic][subframe][band] = (float)d->direct_to_total_energy_ratio / (float)d->dtr_multiplier;
                h->energyData[mic][subframe][band] = ((float)d->overall_energy / (float)d->energy_multiplier) * h->sourceEnergyGains[mic];
                calculate_spatial_parameters_from_listener_to_projected_source(mic, subframe, band, intensityVector, newDoaVector, interpolatedListenerPos);
            }
        }
        if (h->useRecursiveAveraging) {
            // In recursive averaging mode, same spatial values and energies are used in
            // each sub-frame of the processing frame.
            copy_parameters_to_sub_frames(mic);
        }
        // seek to end of sources
        stream.skipBytes(d->sourceMetadataFrameSize * (d->num_hoa_sources_in_group[h->HOAGroupId] - encodedMic - 1));
    };

    auto skip_over_groups = [&](double currentFrame) {
        decoderData* d = &h->decoderData1;
        auto& stream = *h->spatialMetadataByteStream;
        // Skip over rest of the groups to get start position for next frame unless we are at the last
        // frame in which case skipping would move the stream past its end. In that case the next read
        // (should) notice that the stream has looped and will find the then-current frame and set the
        // stream position accordingly.
        if (currentFrame < d->num_frames - 1) {
            for (int grp = h->HOAGroupId + 1; grp < d->num_hoa_groups; grp++) {
                stream.skipBytes(d->groupMetadataFrameSizes[grp]);
            }
        }
    };

    std::vector<float> intensityVector(3, 0.0f);
    std::vector<float> newDoaVector(3, 0.0f);
    std::vector<float> interpolatedListenerPos(3, 0.0f);

    // Determine the current reading frame
    decoderData* d = &h->decoderData1;
    std::uint8_t isInfoFrame;
    std::size_t startingIndex;
    auto& stream = *h->spatialMetadataByteStream;
    auto currentFrame = std::fmod(h->currentFrame, d->num_frames);
    // determine reading position

    // If we just moved to next frame, we can continue from last reading position but
    // otherwise need to go through bitstream to find position
    if (d->lastReadFrame != currentFrame - 1) {
        startingIndex = d->bitStreamHeaderSize + d->infoFrameSize;
        // go through frames until we are at currentFrame
        stream.setBytePosition(startingIndex);
        for (int i = 0; i < currentFrame; i++) {
            stream.read(isInfoFrame);
            startingIndex += sizeof(isInfoFrame);
            if (isInfoFrame == 1) {
                // InfoFrame()
                startingIndex += d->infoFrameSize;
                startingIndex += sizeof(isInfoFrame);
            }
            // regular frame
            for (int grp = 0; grp < d->num_hoa_groups; grp++) {
                startingIndex += d->groupMetadataFrameSizes[grp];
            }
            stream.setBytePosition(startingIndex);
        }
    } else {
        startingIndex = stream.bytePosition();
    }

    // Handle uneven frame counts
    if (d->is_audio_uneven) {
        if (currentFrame == (d->num_frames - 1)) {
            h->currentAudioLoop = (std::max)(0, (int)std::floor((h->currentFrame - d->num_frames) / d->num_frames));
            if (h->currentAudioLoop % 2 != 0) {
                ++h->currentFrame;
            }
        }
        if (currentFrame == 0 && h->currentAudioLoop > 0) {
            ++h->currentAudioLoop;
        }
    } 

    stream.read(isInfoFrame);
    startingIndex += sizeof(isInfoFrame);
    if (isInfoFrame == 1) {
        // InfoFrame()
        startingIndex += d->infoFrameSize;
        startingIndex += sizeof(isInfoFrame);
    }

    d->lastReadFrame = currentFrame;

    // find position of group
    for (int grp = 0; grp < h->HOAGroupId; grp++) {
        startingIndex += d->groupMetadataFrameSizes[grp];
    }
    stream.setBytePosition(startingIndex);

    if (!h->useSphoa6DofExtensionMode) {
        read_mphoa_spatial_metadata();
    } else {
        read_sphoa_spatial_metadata(intensityVector, newDoaVector, interpolatedListenerPos);
    }
    
    skip_over_groups(currentFrame);

    // Read the second subframe from the next frame, if the audio is read from the middle of the frame
    if (d->is_audio_uneven && (h->currentAudioLoop % 2 != 0) && (currentFrame < d->num_frames - 1)) {

        auto tempIndex = stream.bytePosition();
        startingIndex = stream.bytePosition();

        stream.read(isInfoFrame);
        startingIndex += sizeof(isInfoFrame);
        if (isInfoFrame == 1) {
            // InfoFrame()
            startingIndex += d->infoFrameSize;
            startingIndex += sizeof(isInfoFrame);
        }

        // find position of group
        for (int grp = 0; grp < h->HOAGroupId; grp++) {
            startingIndex += d->groupMetadataFrameSizes[grp];
        }
        stream.setBytePosition(startingIndex);

        for (int mic = 0; mic < d->num_hoa_sources_in_group[h->HOAGroupId]; mic++) {
            if (std::find(h->activeArraysForCurrentFrame.begin(), h->activeArraysForCurrentFrame.end(), mic) != h->activeArraysForCurrentFrame.end()) {
                //SpatialMetadataFrame()
                int subframe = 1;
                for (int band = 0; band < d->num_frequency_bands; band++) {
                    stream.read(d->hoa_render_meta_azimuth);
                    stream.read(d->hoa_render_meta_elevation);
                    stream.read(d->direct_to_total_energy_ratio);
                    stream.read(d->overall_energy);
                    h->azimuthData[mic][subframe][band] =
                        (float)d->hoa_render_meta_azimuth / (float)d->azimuth_multiplier * M_PI / 180.0f;
                    h->elevationData[mic][subframe][band] =
                        (float)d->hoa_render_meta_elevation / (float)d->elevation_multiplier * M_PI / 180.0f;
                    h->DTRData[mic][subframe][band] = (float)d->direct_to_total_energy_ratio / (float)d->dtr_multiplier;
                    h->energyData[mic][subframe][band] = ((float)d->overall_energy / (float)d->energy_multiplier) * h->sourceEnergyGains[mic];
                }
            } else {
                // skip over non-active sources
                stream.skipBytes(d->sourceMetadataFrameSize);
            }
        }
        
        stream.setBytePosition(tempIndex);
    }
}
