/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Exterior.hpp"

#include <algorithm>
#include <cfloat>
#include <cmath>
#include <iostream>
#include <numeric>
#include <vector>

#include "Rotator.hpp"
#include "SpatialAnalysis.hpp"
#include "VectorOperations.hpp"

void eraseElementFromVector(const std::vector<int>& element, std::vector<std::vector<int>>& vector)
{
    auto it = std::find(vector.begin(), vector.end(), element);
    if (it != vector.end()) {
        vector.erase(it);
    }
}

void find_original_and_omitted_exterior_edges(hoaTranslatorData *h)
{
    // Collect all triangle faces (valid and omitted ones)
    auto allTriangleFaces = h->triangleFaces;
    allTriangleFaces.insert(allTriangleFaces.end(), h->omittedTriangleFaces.begin(), h->omittedTriangleFaces.end());

    // Get all the edges as an edge list
    std::vector<std::vector<int>> edgeList{};
    for (auto const& tri : allTriangleFaces) {
        edgeList.push_back({tri[0], tri[1]});
        edgeList.push_back({tri[1], tri[2]});
        edgeList.push_back({tri[0], tri[2]});
    }

    // Find edges, that appear only once in edgeList, and add those to the original exterior edge list
    for (auto const& edge : edgeList) {
        if (std::count(edgeList.begin(), edgeList.end(), edge) == 1) {
            h->originalExteriorEdges.push_back(edge);
        }
    }

    // Find omitted exterior edges as the edges, that appear in the original exterior edge list, but not in the final exterior edge list
    for (auto const& edge : h->originalExteriorEdges) {
        if(std::find(h->exteriorEdges.begin(), h->exteriorEdges.end(), edge) == h->exteriorEdges.end()) {
            h->omittedExteriorEdges.push_back(edge);
        }
    }
}

// Find all the exterior edges
void find_exterior_edges(hoaTranslatorData *h)
{
    if (h->useMphoa1SourceMode) {
        h->exteriorMicArrays.push_back(0);
        h->nExteriorEdges = 0;
        return;
    }

    if (h->useMphoa2SourceMode) {
        h->exteriorEdges.push_back({0, 1});
        h->exteriorEdges.push_back({1, 0});
        h->originalExteriorEdges.push_back({0, 1});
        h->originalExteriorEdges.push_back({1, 0});
        h->exteriorMicArrays.push_back(0);
        h->exteriorMicArrays.push_back(1);
        h->nExteriorEdges = 2;
        return;
    }

    // Get all the edges from the valid triangles (long triangles already omitted)
    std::vector<std::vector<int>> edgeList{};
    for (auto const& tri : h->triangleFaces) {
        edgeList.push_back({tri[0], tri[1]});
        edgeList.push_back({tri[1], tri[2]});
        edgeList.push_back({tri[0], tri[2]});
    }

    // Find edges, that appear only once in edgeList, and add those to the exterior edge list
    auto leftoverEdges = h->edgesBetweenArrays;
    for (auto const& edge : edgeList) {
        if (std::count(edgeList.begin(), edgeList.end(), edge) == 1) {
            h->exteriorEdges.push_back(edge);
            // Collect all the exterior mic arrays as well
            for (auto const& mic : edge) {
                if (std::count(h->exteriorMicArrays.begin(), h->exteriorMicArrays.end(), mic) == 0) {
                    h->exteriorMicArrays.push_back(mic);
                }
            }
        }
        eraseElementFromVector(edge, leftoverEdges);
    }

    // Add all possible leftover edges to the bridge edge list. These edges consist of
    // possible edges, that connect two far away sub-groups of mic arrays.
    for (auto const& edge : leftoverEdges) {
        h->bridgeEdges.push_back(edge);
    }
    h->nExteriorEdges = h->exteriorEdges.size();
    h->nBridgeEdges = h->bridgeEdges.size();

    if (h->nBridgeEdges > 0) {
        h->useImprovedExteriorRendering = false;
        std::cerr << "WARNING: The scene includes individual edges, that are not part of any mic array triangles. "
                  << "The improved exterior rendering is disabled." << std::endl;
    }

    // Find also the original and omitted sets of exterior edges
    find_original_and_omitted_exterior_edges(h);
}

// Sort the exterior edges in traversing order.
// E.g. (m1, m2) - (m2, m5) - (m5, m6) ...
void sort_exterior_edges(hoaTranslatorData *h)
{
	if (h->useMphoa1SourceMode) {
        //no exterior edges in 1SourceMode
        return;
    }

    auto unsortedEdges = h->exteriorEdges;
    std::vector<std::vector<int>> sortedEdges{};

    // Start by taking the first edge from the list
    sortedEdges.push_back(unsortedEdges[0]);
    unsortedEdges.erase(unsortedEdges.begin());
    int secondMic = sortedEdges.back()[1];

    int nMaxLoops = h->nExteriorEdges + 10;
    int loopCounter = 0;
    while (!unsortedEdges.empty() || loopCounter < nMaxLoops) {
        for (auto const& edge : unsortedEdges) {
            if (edge[0] == secondMic || edge[1] == secondMic) {
                if (edge[0] == secondMic) {
                    // Edge starts with the previous last mic, add the edge to the sorted list
                    sortedEdges.push_back(edge);
                } else {
                    // Edge ends with the previous last mic, flip the edge and add to the sorted list
                    sortedEdges.push_back({edge[1], edge[0]});
                }
                eraseElementFromVector(edge, unsortedEdges);
                secondMic = sortedEdges.back()[1];
                break;
            }
        }
        ++loopCounter;
    }
    assert(h->nExteriorEdges == sortedEdges.size() && "Couldn't sort all the exterior edges!");
    h->exteriorEdges = sortedEdges;
}

bool is_non_convex_mic(hoaTranslatorData *h,
                       int mic)
{
    if (std::find(h->nonConvexMicArrays.begin(), h->nonConvexMicArrays.end(), mic) != h->nonConvexMicArrays.end()) {
        return true;
    }
    return false;
}

/* Check if edge contains a non-convex mic (i.e. if the edge is part of a non-convex hull).
 *
 * Return:
 *      * -1, if the edge does not contain a non-convex mic
 *      * mic index, if the edge contains a non-convex mic
 */
int search_non_convex_mic_in_edge(hoaTranslatorData *h,
                                  const std::vector<int>& edge)
{
    int mic = -1;
    if (is_non_convex_mic(h, edge[0])) {
        mic = edge[0];
    } else if (is_non_convex_mic(h, edge[1])) {
        mic = edge[1];
    }
    return mic;
}

std::vector<int> search_outermost_edge(hoaTranslatorData *h,
                                       int mic)
{
    // Look for the edge that contains the mic
    int edgeIndx = 0;
    for (int i = 0; i < h->nExteriorEdges; ++i) {
        if (h->exteriorEdges[i][0] == mic || h->exteriorEdges[i][1] == mic) {
            edgeIndx = i;
            break;
        }
    }

    // Check if one of the edge mics is part of the original convex hull
    int outermostEdgeMic1 = -1;
    int outermostEdgeMic2 = -1;
    if (std::find(h->nonConvexMicArrays.begin(), h->nonConvexMicArrays.end(), h->exteriorEdges[edgeIndx][0]) == h->nonConvexMicArrays.end()) {
        outermostEdgeMic1 = h->exteriorEdges[edgeIndx][0];
    }
    if (std::find(h->nonConvexMicArrays.begin(), h->nonConvexMicArrays.end(), h->exteriorEdges[edgeIndx][1]) == h->nonConvexMicArrays.end()) {
        outermostEdgeMic2 = h->exteriorEdges[edgeIndx][0];
    }

    // Traverse the edges backwards to find the first mic forming the outermost edge
    if (outermostEdgeMic1 < 0) {
        for (int traverseEdgeIndx = edgeIndx - 1; ; --traverseEdgeIndx) {
            if (traverseEdgeIndx < 0) {
                traverseEdgeIndx = h->nExteriorEdges - 1;
            }
            if (traverseEdgeIndx == edgeIndx) {
                assert(false && "Couldn't find mic array for an outermost edge!");
                break;
            }
            if (std::find(h->nonConvexMicArrays.begin(), h->nonConvexMicArrays.end(), h->exteriorEdges[traverseEdgeIndx][0]) == h->nonConvexMicArrays.end()) {
                outermostEdgeMic1 = h->exteriorEdges[traverseEdgeIndx][0];
                break;
            }
        }
    }

    // Traverse the edges forward to find the second mic forming the outermost edge
    if (outermostEdgeMic2 < 0) {
        for (int traverseEdgeIndx = edgeIndx + 1; ; ++traverseEdgeIndx) {
            if (traverseEdgeIndx > h->nExteriorEdges - 1) {
                traverseEdgeIndx = 0;
            }
            if (traverseEdgeIndx == edgeIndx) {
                assert(false && "Couldn't find mic array for an outermost edge!");
                break;
            }
            if (std::find(h->nonConvexMicArrays.begin(), h->nonConvexMicArrays.end(), h->exteriorEdges[traverseEdgeIndx][1]) == h->nonConvexMicArrays.end()) {
                outermostEdgeMic2 = h->exteriorEdges[traverseEdgeIndx][1];
                break;
            }
        }
    }

    return {outermostEdgeMic1, outermostEdgeMic2};
}

// Compute normals for the exterior edges
void compute_edge_normals(hoaTranslatorData *h)
{
    // For convex hulls, calculate the normal for the exterior edge
    auto calculate_normal_for_convex_hull = [&](const std::vector<float> &centerOfArrayMass, const hoatMatrix2D &rotationMatrix, const std::vector<int> &edge) {
        std::vector<float> comToMic1(3, 0.0f);
        std::vector<float> comToMic2(3, 0.0f);
        std::vector<float> edgeNormal(3, 0.0f);

        for (int i : {0, 2}) {
            // comToMic1 and 2 point from the center of mass to the edge mic arrays
            comToMic1[i] = h->microphoneArrayPositions[edge[0]][i] - centerOfArrayMass[i];
            comToMic2[i] = h->microphoneArrayPositions[edge[1]][i] - centerOfArrayMass[i];
        }
        // Compute the normalized edge normal
        auto micVec12 = subtractVectors(comToMic1, comToMic2);
        std::copy_n(&micVec12[0], 3, &h->temp3x1Matrix[0][0]);
        matrixMult(rotationMatrix, h->temp3x1Matrix, h->temp3x1Matrix2);
        std::copy_n(&h->temp3x1Matrix2[0][0], 3, &edgeNormal[0]);
        normalizeVector(edgeNormal);

        // Flip the normal, if it's pointing inwards the capturing area.
        // nMid points from the center of mass to the center point between the two edge mic arrays.
        // NOTE: only valid for convex hulls (might not work properly with non-convex hulls)
        auto micVec12Middle = sumVectors(comToMic1, comToMic2);
        micVec12Middle = multiplyVector(0.5f, micVec12Middle);
        normalizeVector(micVec12Middle);
        if (std::acos(dotProduct(micVec12Middle, edgeNormal)) > (M_PI / 2.0f)) {
            edgeNormal = multiplyVector(-1.0f, edgeNormal);
        }
        return edgeNormal;
    };

    // For non-convex hulls, use the outermost exterior edge normal (long edges included)
    auto calculate_normal_for_non_convex_hull = [&](const std::vector<float> &centerOfArrayMass, const hoatMatrix2D &rotationMatrix, int mic) {
        auto outermostEdge = search_outermost_edge(h, mic);
        return calculate_normal_for_convex_hull(centerOfArrayMass, rotationMatrix, outermostEdge);
    };

    if (h->useMphoa1SourceMode) {
        // no exterior edges in 1SourceMode
        return;
    }

    // Initializations
    float R[3][3]{};
    hoatMatrix2D rotationMatrix;
    createHoatMatrix(&rotationMatrix, 3, 3);
    std::vector<float> edgeNormal(3, 0.0f);

    // Compute center of mass for the mic arrays
    std::vector<float> centerOfArrayMass(3, 0.0f);
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int i = 0; i < 3; ++i) {
            centerOfArrayMass[i] += (h->microphoneArrayPositions[mic][i] / h->nMicrophoneArrays);
        }
    }

    // Get rotation matrix (90 degrees yaw)
    get_rotation_matrix_OpenGL(90, 0, 0, R);
    std::copy_n(&R[0][0], 3 * 3, &rotationMatrix[0][0]);

    // Compute edge normals (horizontal plane only)
    for (auto const& edge : h->exteriorEdges) {
        auto mic = search_non_convex_mic_in_edge(h, edge);
        if (mic >= 0) {
            edgeNormal = calculate_normal_for_non_convex_hull(centerOfArrayMass, rotationMatrix, mic);
        } else {
            edgeNormal = calculate_normal_for_convex_hull(centerOfArrayMass, rotationMatrix, edge);
        }
        h->edgeNormals.push_back(edgeNormal);
    }

    // Edge normals for the original triangles (long edges included)
    for (auto const& edge : h->originalExteriorEdges) {
        edgeNormal = calculate_normal_for_convex_hull(centerOfArrayMass, rotationMatrix, edge);
        h->originalEdgeNormals.push_back(edgeNormal);
    }

    // Free
    destroyHoatMatrix(&rotationMatrix);
}

// Compute normals for the exterior mic arrays. The normals are calculated as the average normals from the connecting edges.
void compute_mic_array_normals(hoaTranslatorData *h)
{
    if (h->useMphoa1SourceMode) {
        // no pre-computed mic-normals in the case of 1SourceMode
        return;
    }

    if (h->useMphoa2SourceMode) {
        // For 2 source mode, use normals parallel to the edge vector
        std::vector<float> m1{ h->microphoneArrayPositions[0][0], h->microphoneArrayPositions[0][1], h->microphoneArrayPositions[0][2] };
        std::vector<float> m2{ h->microphoneArrayPositions[1][0], h->microphoneArrayPositions[1][1], h->microphoneArrayPositions[1][2] };
        auto micVec21 = subtractVectors(m1, m2);
        normalizeVector(micVec21);
        h->micNormals.push_back(micVec21);
        h->micNormals.push_back(multiplyVector(-1.0f, micVec21));
        return;
    }

    for (int edgeIndx = 0; edgeIndx < h->nExteriorEdges; ++edgeIndx) {
        int firstNormalIndx = 0;
        if (edgeIndx == 0) {
            firstNormalIndx = h->nExteriorEdges - 1;
        } else {
            firstNormalIndx = edgeIndx - 1;
        }
        auto micNormal = sumVectors(h->edgeNormals[firstNormalIndx], h->edgeNormals[edgeIndx]);
        normalizeVector(micNormal);
        h->micNormals.push_back(micNormal);
    }
}

/* Calculate squared distance from a point 'l' to a line connected with m1 and m2.
 *
 * Example setup for calculate_squared_distance_to_line() and calculate_edge_weight():
 * (horizontal setup, where x is right, and y is up)
 *
 *      - l(x0, y0) is the listener position
 *      - m1(x1, y1) is the position of edge mic array 1
 *      - m2(x2, y2) is the position of edge mic array 2
 *      - e1 is a vector pointing from mic array 1 to mic array 2, i.e. m1m2
 *      - e2 is a vector pointing from mic array 1 to the listener, i.e. m1l
 *      - alpha is the angle between e1 and e2
 *      - l´ is the projected listener position on e1
 *
 *                        o l(x0, y0), listener
 *                      / |
 *                    /   |
 *               e2 /     | h
 *                /       |
 * micArray1,   / )alpha  |l´
 * m1(x1, y1) o-----------x--------------------o m2(x2, y2), micArray2
 *            |-----------------e1------------>
 *
 * The distance (h) from the listener (l) to the line (e1) can be calculated as (xy -horizontal plane):
 *
 *      h = |(x2 - x1)(y1 - y0) - (x1 - x0)(y2 - y1)| /
 *          sqrt((x2 - x1)^2 + (y2 - y1)^2)
 *
 * https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points
 */
float calculate_squared_distance_to_line(hoaTranslatorData *h, std::vector<float> m1, std::vector<float> m2)
{
    float x0 = h->currentListenerPosition[0];
    float y0 = -h->currentListenerPosition[2];
    float x1 = m1[0];
    float y1 = m1[1];
    float x2 = m2[0];
    float y2 = m2[1];
    float nominator = std::pow((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1), 2.0f);
    float denominator = std::pow(x2 - x1, 2.0f) + std::pow(y2 - y1, 2.0f);
    return (nominator / (denominator + FLT_MIN));
}

/* See above for the example setup image. The distance from mic array1 to the projected position p´ (=|v1p´|)
 * can be calculated as:
 *
 *      |v1p´| = cos(alpha) * |e2|
 *             = (dot(e1, e2) / |e1||e2|) * |e2|
 *             = dot(e1, e2) / |e1|
 *
 * The interpolation weight for mic array 1 is then obtained with:
 *
 *      weightV1 = 1.0 - |m1l´| / |e1|
 *               = 1.0 - dot(e1, e2) / |e1|^2
 *
 * http://www.sunshine2k.de/coding/java/PointOnLine/PointOnLine.html#step5
 */
float calculate_edge_weight(std::vector<float> e1, std::vector<float> e2)
{
    float nominator = std::inner_product(e1.begin(), e1.end(), e2.begin(), 0.0f);
    float denominator = std::pow(e1[0], 2.0f) + std::pow(e1[1], 2.0f);
    return (1.0f - nominator / (denominator + FLT_MIN));
}

/* Find the mic array, that is closest to the listener. The distance between two points is (xy -horizontal plane):
 *
 *      distance = sqrt((x2 - x1)^2 + (y2 - y1)^2)
 */
void find_closest_array_based_on_distance(hoaTranslatorData *h)
{
    for (auto const& mic : h->exteriorMicArrays) {
        std::vector<float> m1{h->microphoneArrayPositions[mic][0], -h->microphoneArrayPositions[mic][2]};
        float listenerToArrayDistanceSquared = std::pow(m1[0] - h->currentListenerPosition[0], 2.0f) +
                                               std::pow(m1[1] - (-h->currentListenerPosition[2]), 2.0f);
        if (listenerToArrayDistanceSquared < h->distToClosestArraySquared) {
            h->distToClosestArraySquared = listenerToArrayDistanceSquared;
            h->closestArrayByDistance = mic;
        }
    }
}

/* Try to find an edge, that is closest to the listener. The squared distance to an edge line is calculated with
 * calculate_squared_distance_to_line(). However, this distance also takes into account parts of the edge lines,
 * that extend beyond the connecting edge mic arrays. Take into account only those edges, that have edge weights
 * within [0 1].
 *
 * For example, below is an example scenario, where the listener is positioned in l, and two mic arrays at m1 and m2.
 * Two edges, e1 and e2, extend from m1. The distance from the listener to the line connecting m1 and m2 is |ll´|, where
 * l´ is the listeners projected position on line m1m2. Point l´ is not on the connecting edge e1, and needs to be
 * discarded somehow.
 *
 * The discard process uses the interpolation weights, that are calculated for m1 and m2 with calculate_edge_weight().
 * The edge weights for m1 and m2 are are calculated as a distance relation, for example weight for m1 is
 * w1 = 1.0 - |m1l´| / |e1|. As seen from the example scenario, weight for m2 will be below zero, because |m2l´| > |e1|.
 * When the weights for both edge arrays are within [0 1], the projected listener position is between the edge arrays,
 * and the edge is taken into account.
 *
 * In this example scenario, both edges are discarded, and no closest edge is found. In this case, the listener will be
 * later projected to the closest mic array (which is m1 here).
 *
 * Example scenario:
 *      l  - listener
 *      l´ - listeners projected position on e1
 *      m1 - mic array 1
 *      m2 - mic array 2
 *      e1 - edge connecting m1 and m2
 *      e2 - edge connecting m1 and some other mic array
 *
 *       l
 *       |
 *       l´ m1 ----------e1---- m2
 *         /
 *        /
 *       e2
 *      /
 */
void find_closest_edge(hoaTranslatorData *h)
{
    for (int edgeIndx = 0; edgeIndx < h->edgesBetweenArrays.size(); ++edgeIndx) {
        int mic1 = h->edgesBetweenArrays[edgeIndx][0];
        int mic2 = h->edgesBetweenArrays[edgeIndx][1];
        std::vector<float> m1{h->microphoneArrayPositions[mic1][0], -h->microphoneArrayPositions[mic1][2]};
        std::vector<float> m2{h->microphoneArrayPositions[mic2][0], -h->microphoneArrayPositions[mic2][2]};
        float distSquared = calculate_squared_distance_to_line(h, m1, m2);
        if (distSquared < h->distToClosestEdgeSquared && distSquared < h->distToClosestArraySquared) {
            std::vector<float> m1l{h->currentListenerPosition[0] - m1[0], -h->currentListenerPosition[2] - m1[1]};
            std::vector<float> m2l{h->currentListenerPosition[0] - m2[0], -h->currentListenerPosition[2] - m2[1]};
            std::vector<float> m1m2{m2[0] - m1[0], m2[1] - m1[1]};
            std::vector<float> m2m1{-m1m2[0], -m1m2[1]};
            float weight1 = calculate_edge_weight(m1m2, m1l);
            float weight2 = calculate_edge_weight(m2m1, m2l);
            // The listener is between mic arrays 1 and 2, if both weights are between [0 1]
            if (weight1 >= 0.0f && weight1 <= 1.0f && weight2 >= 0.0f && weight2 <= 1.0f) {
                h->distToClosestEdgeSquared = distSquared;
                h->edgeArrays[0] = mic1;
                h->edgeArrays[1] = mic2;
                h->edgeWeights[0] = weight1;
                h->edgeWeights[1] = 1.0f - h->edgeWeights[0];
                h->isClosestEdgeFound = true;
            }
        }
    }
}

std::vector<float> project_based_on_normal(hoaTranslatorData *h,
                                           const std::vector<int> &edge,
                                           const std::vector<float> &normal,
                                           const std::vector<float> &listenerPosition)
{
    // Form the edge vectors
    std::vector<float> mic1(3, 0.0f);
    std::vector<float> mic2(3, 0.0f);
    for (int i : {0, 2}) {
        mic1[i] = h->microphoneArrayPositions[edge[0]][i];
        mic2[i] = h->microphoneArrayPositions[edge[1]][i];
    }
    auto edgeVec = subtractVectors(mic2, mic1);
    float lenEdge = calculate_vector_norm(edgeVec);
    normalizeVector(edgeVec);

    // Horizontal listener position in mic1-space
    auto listenerPositionM1Space = subtractVectors(listenerPosition, mic1);
    listenerPositionM1Space[1] = 0.0f;

    // Calculate projection weights the normal and edge directions
    auto dotEdgeNormal = dotProduct(edgeVec, normal);
    float edgeWeight = 0.0f;
    float normalWeight = 0.0f;
    if (dotEdgeNormal == 0) {
        edgeWeight = dotProduct(listenerPositionM1Space, edgeVec);
        normalWeight = dotProduct(listenerPositionM1Space, normal);
    } else {
        edgeWeight = (dotProduct(listenerPositionM1Space, normal) - (dotProduct(listenerPositionM1Space, edgeVec) / dotEdgeNormal)) /
                     (dotEdgeNormal - (1.0f / dotEdgeNormal));
        normalWeight = (dotProduct(listenerPositionM1Space, edgeVec) - (dotProduct(listenerPositionM1Space, normal) / dotEdgeNormal)) /
                       (dotEdgeNormal - (1.0f / dotEdgeNormal));
    }
    edgeWeight = 1.0f - edgeWeight / lenEdge;

    return {edgeWeight, normalWeight};
}

/* Find the closest exterior edge parallel to the edge normal directions. First, project to the original exterior edges,
 * that include the long omitted edges. If the listener is projected to an omitted edge, project again to the final exterior
 * edges (long edges omitted) using the projected position as the original listener position.
 */
void find_closest_edge_based_on_normal(hoaTranslatorData *h)
{
    auto project_to_edge_and_check_if_closest = [&](const std::vector<int> &edge, const std::vector<float> &normal,
                                                    const std::vector<float> &listenerPosition, bool skipClosestArrayCheck) {
        auto edgeNormalWeights = project_based_on_normal(h, edge, normal, listenerPosition);
        auto edgeWeight = edgeNormalWeights[0];
        auto normalWeight = edgeNormalWeights[1];
        if (normalWeight > 0.0f) {
            auto distSquared = std::pow(normalWeight, 2.0f);
            if (distSquared < h->distToClosestEdgeSquared && (skipClosestArrayCheck || distSquared < h->distToClosestArraySquared) &&
                edgeWeight >= 0.0f && edgeWeight <= 1.0f) {
                h->distToClosestEdgeSquared = distSquared;
                h->edgeArrays[0] = edge[0];
                h->edgeArrays[1] = edge[1];
                h->edgeWeights[0] = edgeWeight;
                h->edgeWeights[1] = 1.0f - h->edgeWeights[0];
                h->isClosestEdgeFound = true;
                return true;
            }
        }
        return false;
    };

    bool isListenerProjectedToOmittedEdge = false;
    int originalEdgeIndx = 0;

    if (h->isListenerInsideOmittedTriangle) {
        // Project only to the final exterior edges (long edges omitted)
        for (int edgeIndx = 0; edgeIndx < h->nExteriorEdges; ++edgeIndx) {
            project_to_edge_and_check_if_closest(h->exteriorEdges[edgeIndx], h->edgeNormals[edgeIndx], h->currentListenerPosition, false);
        }
    } else {
        // Project to the original exterior edges (long edges included)
        for (int edgeIndx = 0; edgeIndx < h->originalExteriorEdges.size(); ++edgeIndx) {
            auto edge = h->originalExteriorEdges[edgeIndx];
            if (project_to_edge_and_check_if_closest(edge, h->originalEdgeNormals[edgeIndx], h->currentListenerPosition, false)) {
                // Check if the projected edge is omitted
                if(std::find(h->omittedExteriorEdges.begin(), h->omittedExteriorEdges.end(), edge) != h->omittedExteriorEdges.end()) {
                    isListenerProjectedToOmittedEdge = true;
                } else {
                    isListenerProjectedToOmittedEdge = false;
                }
                originalEdgeIndx = edgeIndx;
            }
        }
    }

    // If the projected edge is omitted, project again from the projected position
    if (isListenerProjectedToOmittedEdge) {
        // Reset the distance to the closest edge, use the normal of the originally projected edge
        h->distToClosestEdgeSquared = FLT_MAX;
        auto normal = h->originalEdgeNormals[originalEdgeIndx];

        // Get the projected listener position on the omitted edge
        std::vector<float> projectedListenerPosition(3, 0.0f);
        for (int i = 0; i < 3; ++i) {
            for (int mic = 0; mic < 2; ++mic) {
                projectedListenerPosition[i] += (h->microphoneArrayPositions[h->edgeArrays[mic]][i] * h->edgeWeights[mic]);
            }
        }

        // Project to the final exterior edges (long edges omitted)
        for (auto const &edge : h->exteriorEdges) {
            project_to_edge_and_check_if_closest(edge, normal, projectedListenerPosition, true);
        }
    }
}

// Project the listener position to the closest edge
void project_to_closest_edge(hoaTranslatorData *h)
{
    h->isListenerProjectedToClosestEdge = true;

    // Update the projected listener position
    std::fill(h->projectedListenerPosition.begin(), h->projectedListenerPosition.end(), 0.0f);
    for (int edgeArrayIdx = 0; edgeArrayIdx < 2; ++edgeArrayIdx) {
        for (int i = 0; i < 3; ++i) {
            h->projectedListenerPosition[i] += h->microphoneArrayPositions[h->edgeArrays[edgeArrayIdx]][i] * h->edgeWeights[edgeArrayIdx];
        }
    }

    // Determine true closest edge array
    int closestEdgeArray = h->edgeArrays[0];
    int secondClosestEdgeArray = h->edgeArrays[1];
    if (h->edgeWeights[0] < h->edgeWeights[1]) {
        closestEdgeArray = h->edgeArrays[1];
        secondClosestEdgeArray = h->edgeArrays[0];
    }

    // Update the triangle, in which the listener is projected to
    for (int tri = 0; tri < h->nTriangles; ++tri) {
        if (std::find(h->triangleFaces[tri].begin(), h->triangleFaces[tri].end(), closestEdgeArray) != h->triangleFaces[tri].end()) {
            // Triangle with the closest array was found
            h->activeTriangleIndex = tri;
            for (int i = 0; i < 3; ++i) {
                h->edgeTriangleArrays[i] = h->triangleFaces[tri][i];
            }
            if (std::find(h->triangleFaces[tri].begin(), h->triangleFaces[tri].end(), secondClosestEdgeArray) != h->triangleFaces[tri].end()) {
                // Triangle contains both arrays, use that for activeTriangleIndex and stop looping
                h->isClosestEdgePartOfTriangle = true;
                break;
            }
        }
    }

    // If the projected edge is not part of a triangle, find the triangle for the second closest array
    if (!h->isClosestEdgePartOfTriangle) {
        for (int tri = 0; tri < h->nTriangles; ++tri) {
            if (std::find(h->triangleFaces[tri].begin(), h->triangleFaces[tri].end(), secondClosestEdgeArray) != h->triangleFaces[tri].end()) {
                // Triangle with the second closest array was found
                for (int i = 0; i < 3; ++i) {
                    h->edgeTriangleArrays[i + 3] = h->triangleFaces[tri][i];
                }
            }
        }
    }
}

// Project the listener to the closest array
void project_to_closest_array(hoaTranslatorData *h)
{
    h->isListenerProjectedToClosestArray = true;

    // Update the projected listener position
    std::copy_n(&h->microphoneArrayPositions[h->closestArrayByDistance][0], 3, &h->projectedListenerPosition[0]);

    // Update the triangle, in which the listener is projected to
    for (int tri = 0; tri < h->nTriangles; ++tri) {
        if (std::find(h->triangleFaces[tri].begin(), h->triangleFaces[tri].end(), h->closestArrayByDistance) != h->triangleFaces[tri].end()) {
            // Triangle with the closest array was found
            h->activeTriangleIndex = tri;
            break;
        }
    }
}

void determine_exterior_edge_index(hoaTranslatorData *h)
{
    std::vector<int> edge1{ h->edgeArrays[0], h->edgeArrays[1] };
    std::vector<int> edge2{ h->edgeArrays[1], h->edgeArrays[0] };
    auto it1 = std::find(h->exteriorEdges.begin(), h->exteriorEdges.end(), edge1);
    auto it2 = std::find(h->exteriorEdges.begin(), h->exteriorEdges.end(), edge2);
    if (it1 != h->exteriorEdges.end()) {
        h->isEdgeFlipped = false;
        h->currentExteriorEdgeIndex = std::distance(h->exteriorEdges.begin(), it1);
    } else if (it2 != h->exteriorEdges.end()) {
        h->isEdgeFlipped = true;
        h->currentExteriorEdgeIndex = std::distance(h->exteriorEdges.begin(), it2);
    } else {
        assert(false && "Couldn't find exterior edge from exteriorEdges!");
    }
}

void determine_exterior_mic_index(hoaTranslatorData *h)
{
    for (int edgeIdx = 0; edgeIdx < h->nExteriorEdges; ++edgeIdx) {
        auto edge = h->exteriorEdges[edgeIdx];
        if (edge[0] == h->closestArrayByDistance) {
            h->currentExteriorMicIndex = edgeIdx;
            return;
        }
    }
}

void interpolate_mic_normals_at_projected_listener_position(hoaTranslatorData *h)
{

    if (h->useMphoa1SourceMode) {
        // listener normal points towards listener
        h->listenerExteriorNormal[0] = h->currentListenerPosition[0] - h->microphoneArrayPositions[0][0];
        h->listenerExteriorNormal[1] = h->currentListenerPosition[1] - h->microphoneArrayPositions[0][1];
        h->listenerExteriorNormal[2] = h->currentListenerPosition[2] - h->microphoneArrayPositions[0][2];
        normalizeVector(h->listenerExteriorNormal);
        return;
    }
    if (h->isListenerProjectedToClosestEdge) {
        float micMult1 = 0.0f;
        float micMult2 = 0.0f;
        auto micNormal1 = h->micNormals[0];
        auto micNormal2 = h->micNormals[0];

        // Obtain the mic normals and weights for the projected edge
        determine_exterior_edge_index(h);
        if (h->currentExteriorEdgeIndex == (h->nExteriorEdges - 1)) {
            micNormal1 = h->micNormals[h->nExteriorEdges - 1];
            micNormal2 = h->micNormals[0];
        } else {
            micNormal1 = h->micNormals[h->currentExteriorEdgeIndex];
            micNormal2 = h->micNormals[h->currentExteriorEdgeIndex + 1];
        }
        if (h->isEdgeFlipped) {
            micMult1 = h->edgeWeights[1];
            micMult2 = h->edgeWeights[0];
        } else  {
            micMult1 = h->edgeWeights[0];
            micMult2 = h->edgeWeights[1];
        }

       if (h->useMphoa2SourceMode) {
            // Interpolate the listener normal from the edge normal and the closest mic normal
            float micMult = micMult1;
            auto micNormal = micNormal1;
            auto edgeNormal = h->edgeNormals[h->currentExteriorEdgeIndex];
            if (micMult2 > micMult1) {
                micMult = micMult2;
                micNormal = micNormal2;
            }
            micMult = (micMult - 0.5f) / 0.5f;
            float edgeMult = 1.0f - micMult;
            micNormal = multiplyVector(micMult, micNormal);
            edgeNormal = multiplyVector(edgeMult, edgeNormal);
            h->listenerExteriorNormal = sumVectors(micNormal, edgeNormal);
        } else {
            // Interpolate the listener normal from the mic normals forming the edge
            micNormal1 = multiplyVector(micMult1, micNormal1);
            micNormal2 = multiplyVector(micMult2, micNormal2);
            h->listenerExteriorNormal = sumVectors(micNormal1, micNormal2);
        }

        normalizeVector(h->listenerExteriorNormal);

    } else {
        // Listener projected on top of a mic array, use the normal of that mic as the listener normal
        determine_exterior_mic_index(h);
        h->listenerExteriorNormal = h->micNormals[h->currentExteriorMicIndex];
    }
}

void reset_exterior_variables(hoaTranslatorData *h)
{
    h->distToClosestArraySquared = FLT_MAX;
    h->distToClosestEdgeSquared = FLT_MAX;
    h->isClosestEdgeFound = false;
    h->isListenerProjectedToClosestEdge = false;
    h->isListenerProjectedToClosestArray = false;
    h->isClosestEdgePartOfTriangle = false;
    std::fill(h->edgeTriangleArrays.begin(), h->edgeTriangleArrays.end(), 0);
}

/* If the listener is outside all the triangles, try to project the listener to the closest edge connecting two mic arrays.
 * In a corner section, the listener might also be projected to the closest mic array.
 *
 * NOTE: the edge might not be part of any triangle in triangleFaces (i.e. the edge might be connecting two sub-groups of
 * triangles, that are so far away from each other, that no triangles are formed between the sub-groups).
 *
 * Output:
 *      edgeArrays                 - the indexes of the two mic arrays, that connect the edge, where the listener is projected to
 *      edgeWeights                - interpolation weights for edgeArrays
 *      closestArrayByDistance     - the closest mic array index, when the listener is projected to an array position
 *      distanceToCapturingArea    - distance from the listener position to the projected position
 *      edgeTriangleArrays         - all the mic array indexes that are part of the triangle(s) of edgeArrays
 *                                   (3 if edgeArrays are in the same triangle, 6 if edgeArrays are part of different triangles)
 *      activeTriangleIndex        - updated active triangle index
 */
void project_to_closest_edge_or_array(hoaTranslatorData *h)
{
    // Find the closest array
    find_closest_array_based_on_distance(h);

    // Try to find the closest edge
    find_closest_edge_based_on_normal(h);

    if (h->isClosestEdgeFound) {
        project_to_closest_edge(h);
    } else {
        project_to_closest_array(h);
    }

    if (h->useImprovedExteriorRendering) {
        interpolate_mic_normals_at_projected_listener_position(h);
    }
}

void exterior_pre_processing(hoaTranslatorData *h)
{
    find_exterior_edges(h);
    sort_exterior_edges(h);
    compute_edge_normals(h);
    compute_mic_array_normals(h);
}

void adjust_metadata_outside_capturing_area(hoaTranslatorData *h, bool residualData)
{

	float **interpolatedAzimuthData = h->interpolatedAzimuthData;
	float **interpolatedElevationData = h->interpolatedElevationData;
	float **interpolatedDTRData = h->interpolatedDTRData;

	if (residualData) {
		interpolatedAzimuthData = h->interpolatedResidualAzimuthData;
		interpolatedElevationData = h->interpolatedResidualElevationData;
		interpolatedDTRData = h->interpolatedResidualDTRData;
	}

    if (!h->useSphoa6DofExtensionMode && !h->isListenerInsideTriangle && h->useImprovedExteriorRendering) {
        auto horizontalListenerPosition = h->currentListenerPosition;
        auto horizontalProjectedListenerPosition = h->projectedListenerPosition;
        horizontalListenerPosition[1] = 0.0f;
        horizontalProjectedListenerPosition[1] = 0.0f;
        // Apply metadata adjustments to the low frequency bins and to the
        // averaged high frequency band
        int vecLen = h->nDoaDtrBins + 1;
        if (vecLen > h->nBands) {
            vecLen = h->nBands;
        }
        if (h->adaptiveExteriorProjectionRadius) {
            adjust_projection_radius_based_on_src_positions(h, &h->exteriorProjectionRadius);
        }
        for (int frame = 0; frame < h->nFrames; ++frame) {

            // Form the original DOA vector
            calculate_doa_vectors_OpenGL(h,
                                         vecLen,
                                         &interpolatedAzimuthData[frame][0],
                                         &interpolatedElevationData[frame][0],
                                         h->originalDoaUnitX,
                                         h->originalDoaUnitY,
                                         h->originalDoaUnitZ);

            // Calculate exterior weights for the original and re-mapped DOA/DTR
            calculate_exterior_weights(h, vecLen);

            // Re-map DOA and DTR
            re_map_doa_and_dtr(h,
                               frame,
                               horizontalProjectedListenerPosition,
                               horizontalListenerPosition,
                               vecLen);

            // Apply exterior weighting
            apply_exterior_weighting(h, frame, vecLen);

            // Apply head rotation
            apply_rotation_to_weighted_doa(h, vecLen);

            // Collect the values
            calculate_azimuths_OpenGL(h, &h->reMappedDoaX[0], &h->reMappedDoaZ[0], &interpolatedAzimuthData[frame][0], vecLen);
            calculate_elevations_OpenGL(h, &h->reMappedDoaX[0], &h->reMappedDoaY[0], &h->reMappedDoaZ[0], &interpolatedElevationData[frame][0], vecLen);
            std::copy_n(&h->reMappedDtr[0], vecLen, &interpolatedDTRData[frame][0]);

        }

        // Calculate the diffuse exterior weight
        h->diffuseExteriorWeight = calculate_diffuse_exterior_weight(h->exteriorProjectionRadius,
                                                                     horizontalProjectedListenerPosition,
                                                                     horizontalListenerPosition);
    }
}

void adjust_projection_radius_based_on_src_positions(hoaTranslatorData *h,
                                                     float *exteriorProjectionRadius)
{
    if (h->useSourceAwareInitialFlag && (h->nOutsideSources > 0)) {
        float newRadius = FLT_MAX;
        int nActiveSources = h->activeSources.size();

        // Use the distance from the capturing area to the closest outside source as the new radius
        for (int i = 0; i < h->nOutsideSources; ++i) {
            int src = h->outsideSources[i];
            if (std::find(h->activeSources.begin(), h->activeSources.end(), src) != h->activeSources.end()) {
                if (h->outsideSrcDistances[i] < newRadius) {
                    newRadius = h->outsideSrcDistances[i];
                }
            }
        }
        // If no outside source is active, use the average distance to all outside sources
        if (nActiveSources == 0) {
            newRadius = h->averageOutsideSrcDistance;
        }
        *exteriorProjectionRadius = newRadius;
    }
}

void calculate_exterior_weights(hoaTranslatorData *h, int vecLen)
{
    float weightingSharpness = 2.0f;

    // Dot product for each DOA
    utility_svsmul(h->originalDoaUnitX, &h->listenerExteriorNormal[0], vecLen, h->dotX);
    utility_svsmul(h->originalDoaUnitY, &h->listenerExteriorNormal[1], vecLen, h->dotY);
    utility_svsmul(h->originalDoaUnitZ, &h->listenerExteriorNormal[2], vecLen, h->dotZ);
    for (int i = 0; i < vecLen; ++i) {
        h->dotXYZplus1[i] = h->dotX[i] + h->dotY[i] + h->dotZ[i] + 1.0f;
    }

    // w1 = 1/2^sharpness * (1 + dot(normal, doa))^sharpness
    float preTerm = 1.0f / std::pow(2.0f, weightingSharpness);
    powVector(weightingSharpness, h->dotXYZplus1, vecLen, h->dotXYZplus1);
    utility_svsmul(h->dotXYZplus1, &preTerm, vecLen, h->w1);

    // w2 = 1.0 - w1
    for (int i = 0; i < vecLen; ++i) {
        h->w2[i] = 1.0f - h->w1[i];
    }
}

float calculate_diffuse_exterior_weight(float exteriorProjectionRadius,
                                        const std::vector<float> &horizontalProjectedListenerPosition,
                                        const std::vector<float> &horizontalListenerPosition)
{
    auto projToListener = subtractVectors(horizontalListenerPosition, horizontalProjectedListenerPosition);
    float projToListenerDistance = calculate_vector_norm(projToListener);
    return (std::min)(1.0f, projToListenerDistance / exteriorProjectionRadius);
}

void re_map_doa_and_dtr(hoaTranslatorData *h,
                        int frame,
                        const std::vector<float> &horizontalProjectedListenerPosition,
                        const std::vector<float> &horizontalListenerPosition,
                        int vecLen)
{
    float lenRemappedVec = 0.0f;
    float newDtr = 0.0f;
    float projToListX = horizontalProjectedListenerPosition[0] - horizontalListenerPosition[0];
    float projToListY = horizontalProjectedListenerPosition[1] - horizontalListenerPosition[1];
    float projToListZ = horizontalProjectedListenerPosition[2] - horizontalListenerPosition[2];
    for (int i = 0; i < vecLen; ++i) {
        // Form re-mapped DOA
        h->reMappedDoaX[i] = h->exteriorProjectionRadius * h->originalDoaUnitX[i] + projToListX;
        h->reMappedDoaY[i] = h->exteriorProjectionRadius * h->originalDoaUnitY[i] + projToListY;
        h->reMappedDoaZ[i] = h->exteriorProjectionRadius * h->originalDoaUnitZ[i] + projToListZ;

        // Normalize
        lenRemappedVec = std::sqrt(std::pow(h->reMappedDoaX[i], 2.0f) + std::pow(h->reMappedDoaY[i], 2.0f) + std::pow(h->reMappedDoaZ[i], 2.0f));
        h->reMappedDoaX[i] /= lenRemappedVec;
        h->reMappedDoaY[i] /= lenRemappedVec;
        h->reMappedDoaZ[i] /= lenRemappedVec;

        // Re-map DTR
        newDtr = lenRemappedVec / h->exteriorProjectionRadius;
        h->reMappedDtr[i] = (std::min)(h->interpolatedDTRData[frame][i], newDtr);
    }
}

void apply_exterior_weighting(hoaTranslatorData *h, int frame, int vecLen)
{
    for (int i = 0; i < vecLen; ++i) {
        h->reMappedDoaX[i] = h->w1[i] * h->reMappedDoaX[i] + h->w2[i] * h->originalDoaUnitX[i];
        h->reMappedDoaY[i] = h->w1[i] * h->reMappedDoaY[i] + h->w2[i] * h->originalDoaUnitY[i];
        h->reMappedDoaZ[i] = h->w1[i] * h->reMappedDoaZ[i] + h->w2[i] * h->originalDoaUnitZ[i];
        h->reMappedDtr[i] = (std::min)(1.0f, h->w1[i] * h->reMappedDtr[i] + h->w2[i] * h->interpolatedDTRData[frame][i]);
     }
}

void apply_rotation_to_weighted_doa(hoaTranslatorData *h, int vecLen)
{
    // Rotating the head means that the source directions move to the opposite direction
    // (flip head rotation angles)
    float sourceYawDeg   = -h->yawDeg;
    float sourcePitchDeg = -h->pitchDeg;
    float sourceRollDeg  = -h->rollDeg;

    // Get head rotation matrix R
    float R[3][3]{};
    get_rotation_matrix_OpenGL(sourceYawDeg,
                               sourcePitchDeg,
                               sourceRollDeg,
                               R);

    // Apply head rotation to the weighted DOA vectors
    std::copy_n(&R[0][0], 3 * 3, &h->tempR[0][0]);
    std::copy_n(h->reMappedDoaX, vecLen, &h->temp3xnBandsMatrix[0][0]);
    std::copy_n(h->reMappedDoaY, vecLen, &h->temp3xnBandsMatrix[1][0]);
    std::copy_n(h->reMappedDoaZ, vecLen, &h->temp3xnBandsMatrix[2][0]);
    matrixMult(h->tempR, h->temp3xnBandsMatrix, h->temp3xnBandsMatrix2);
    std::copy_n(&h->temp3xnBandsMatrix2[0][0], vecLen, h->reMappedDoaX);
    std::copy_n(&h->temp3xnBandsMatrix2[1][0], vecLen, h->reMappedDoaY);
    std::copy_n(&h->temp3xnBandsMatrix2[2][0], vecLen, h->reMappedDoaZ);
}
