/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_EXTERIOR_HPP
#define HOA_TRANSLATION_EXTERIOR_HPP

#include "Utilities.hpp"

void find_closest_edge(hoaTranslatorData *h);

void reset_exterior_variables(hoaTranslatorData *h);

void project_to_closest_edge_or_array(hoaTranslatorData *h);

void exterior_pre_processing(hoaTranslatorData *h);

/* Project the listener to an edge based on normal direction. The normal might not be prependicular to the edge.
 * See [1] for more details on the math. Both weights for the edge and normal directions are calculated:
 *
 *      - edgeWeight = the weight for the edge vector direction (i.e. edgeWeight * edgeUnitVec = projected listener position on the edge)
 *      - normalWeight = the weight for the normal vector direction (i.e. normalWeight = distance from the projected listener position to the
 *                       original listener position along the normal direction)
 *
 * Returns:
 *      - the edge and normal direction weights as a vector: {edgeWeight, normalWeight}
 *
 *
 * [1] https://math.stackexchange.com/a/148218
 */
std::vector<float> project_based_on_normal(hoaTranslatorData *h,
                                           const std::vector<int> &edge,
                                           const std::vector<float> &normal,
                                           const std::vector<float> &listenerPosition);

/* Adjust the metadata, when the listener is positioned outside the capturing area. Re-map DOA and DTR values based on
 * the listener position, and the DOAs of the original sounds. Apply heavier modification to sounds, that are coming
 * outside the capturing area, and less modification to sounds coming from inside. The modifications in short:
 *
 *  - DOA:
 *      - The listener is projected to the edge of the capturing area.
 *      - The original DOA is calculated at the projected listener position.
 *      - Create a DOA projection circle, originating from the projected listener position.
 *      - Re-map the DOA for the original listener position. I.e. first find the point, where the original DOA vector
 *        crosses the projection circle (starting from the projected listener position). Then, take the vector from the
 *        original listener position to this crossing point, and use that as the re-mapped DOA.
 *      - The final modified DOA is a weighted combination of the re-mapped DOA, and the original DOA. The weighting
 *        increases the proportion of the re-mapped DOA, when the sound is coming outside the capturing area, and vice
 *        versa.
 *
 *  - DTR:
 *      - The diffuse proportion of the sound energy is increased, when the listener moves further away from the exterior
 *        edge.
 *      - Similarly as in the DOA modification, the DTR modification is weighted based on the original DOA (i.e. more
 *        modification for sounds coming from outside the capturing area).
 */
void adjust_metadata_outside_capturing_area(hoaTranslatorData *h, bool residualData);


/* Adjust the exterior projection radius based on the known source positions outside the capturing area. Use the distance
 * to the closest active outside source. If no outside source is active, use the average distance to all outside sources.
 *
 * Note: the average source distance is used even when the source aware mode is turned off during rendering (i.e. when the
 * listener moves far away from a known source position).
*/
void adjust_projection_radius_based_on_src_positions(hoaTranslatorData *h,
                                                     float *exteriorProjectionRadius);

/* Compute the exterior weights for each frequency band as:
 *
 *      w1 = 1/2^2 * (1 + dot(listenerExteriorNormal, originalDoa))^2
 *      w2 = 1.0 - w1
 *
 * Returns:
 *      - the exterior weights as a vector: {w1, w2}, (size w1 and w2: 1 x nBands)
 */
void calculate_exterior_weights(hoaTranslatorData *h, int vecLen);

/* Compute the diffuse exterior weight as:
 *
 *      w3 = min(1.0, projectionToListenerDistance / exteriorProjectionRadius)
 *      (w4 = 1.0 - w3)
 *
 * Returns:
 *      w3
 */
float calculate_diffuse_exterior_weight(float exteriorProjectionRadius,
                                        const std::vector<float> &horizontalProjectedListenerPosition,
                                        const std::vector<float> &horizontalListenerPosition);

/* Compute the re-mapped DOA (unit vector) and DTR for each frequency band as:
 *
 *      reMappedDoa = (projectedListenerPositionVec) + (projectionRadius * originalDoaUnit) - (listenerPositionVec)
 *      reMappedDtr = min( originalDtr, len(reMappedDoa) / exteriorProjectionRadius )
 *
 * Returns:
 *      - the re-mapped DOA (unit vector) and DTR as a vector: {reMappedDoaUnitX, reMappedDoaUnitY, reMappedDoaUnitZ, reMappedDtr}, (size: 4 x nBands)
 */
void re_map_doa_and_dtr(hoaTranslatorData *h,
                        int frame,
                        const std::vector<float> &horizontalProjectedListenerPosition,
                        const std::vector<float> &horizontalListenerPosition,
                        int vecLen);

/* Apply the exterior weights for each frequency band as:
 *
 *      weightedDoa = (exteriorWeight1 * reMappedDoa) + (exteriorWeight2 * originalDoa)
 *      weightedDtr = min( 1.0, (exteriorWeight1 * reMappedDtr) + (exteriorWeight2 * originalDtr) )
 *
 * Returns:
 *      - weighted DOA and weighted DTR as a vector: {weightedDoaX, weightedDoaY, weightedDoaZ, weightedDtr}, (size: 4 x nBands)
 */
void apply_exterior_weighting(hoaTranslatorData *h, int frame, int vecLen);

void apply_rotation_to_weighted_doa(hoaTranslatorData *h, int vecLen);

#endif //HOA_TRANSLATION_EXTERIOR_HPP
