/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "HoatMatrix.hpp"

#include <algorithm>
#include <iomanip>
#include <iostream>

#include <saf_externals.h>

const float_complex *hoatMatrixCmplx2D::operator[](int i) const{
    assert(i >= 0 && i < nRows);
    return data[i];
}

float_complex *&hoatMatrixCmplx2D::operator[](int i){
    assert(i >= 0 && i < nRows);
    return data[i];
}

const hoatMatrixCmplx2D hoatMatrixCmplx3D::operator[](int i) const {
    assert(i >= 0 && i < n2DMatrices);
    return matrices2D[i];
}

hoatMatrixCmplx2D &hoatMatrixCmplx3D::operator[](int i) {
    assert(i >= 0 && i < n2DMatrices);
    return matrices2D[i];
}

const float *hoatMatrix2D::operator[](int i) const {
    assert(i >= 0 && i < nRows);
    return data[i];
}

float *&hoatMatrix2D::operator[](int i) {
    assert(i >= 0 && i < nRows);
    return data[i];
}

const hoatMatrix2D hoatMatrix3D::operator[](int i) const {
    assert(i >= 0 && i < n2DMatrices);
    return matrices2D[i];
}

hoatMatrix2D &hoatMatrix3D::operator[](int i) {
    assert(i >= 0 && i < n2DMatrices);
    return matrices2D[i];
}

const hoatMatrix3D hoatMatrix4D::operator[](int i) const {
    assert(i >= 0 && i < n3DMatrices);
    return matrices3D[i];
}

hoatMatrix3D &hoatMatrix4D::operator[](int i) {
    assert(i >= 0 && i < n3DMatrices);
    return matrices3D[i];
}

void matrixMult2x2(const hoatMatrixCmplx2D &A,
                   const hoatMatrixCmplx2D &B,
                   hoatMatrixCmplx2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    float real = 0.0f;
    float imag = 0.0f;
    for (int ch1 = 0; ch1 < 2; ch1++) {
        for (int ch2 = 0; ch2 < 2; ch2++) {
            real  = A[ch1][0].real() * B[0][ch2].real() + A[ch1][1].real() * B[1][ch2].real();
            real -= A[ch1][0].imag() * B[0][ch2].imag() + A[ch1][1].imag() * B[1][ch2].imag();
            imag  = A[ch1][0].imag() * B[0][ch2].real() + A[ch1][1].imag() * B[1][ch2].real();
            imag += A[ch1][0].real() * B[0][ch2].imag() + A[ch1][1].real() * B[1][ch2].imag();
            C[ch1][ch2] = {real, imag};
        }
    }
}

void matrixMultConjNoTrans2x2(const hoatMatrixCmplx2D &A,
                              const hoatMatrixCmplx2D &B,
                              hoatMatrixCmplx2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    float real = 0.0f;
    float imag = 0.0f;
    for (int ch1 = 0; ch1 < 2; ch1++) {
        for (int ch2 = 0; ch2 < 2; ch2++) {
            real  =  A[0][ch1].real() * B[0][ch2].real() + A[1][ch1].real() * B[1][ch2].real();
            real -= -A[0][ch1].imag() * B[0][ch2].imag() - A[1][ch1].imag() * B[1][ch2].imag();
            imag  = -A[0][ch1].imag() * B[0][ch2].real() - A[1][ch1].imag() * B[1][ch2].real();
            imag +=  A[0][ch1].real() * B[0][ch2].imag() + A[1][ch1].real() * B[1][ch2].imag();
            C[ch1][ch2] = {real, imag};
        }
    }
}

void matrixMultNoTransConj2x2(const hoatMatrixCmplx2D &A,
                              const hoatMatrixCmplx2D &B,
                              hoatMatrixCmplx2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    float real = 0.0f;
    float imag = 0.0f;
    for (int ch1 = 0; ch1 < 2; ch1++) {
        for (int ch2 = 0; ch2 < 2; ch2++) {
            real  = A[ch1][0].real() *   B[ch2][0].real()  + A[ch1][1].real() *   B[ch2][1].real();
            real -= A[ch1][0].imag() * (-B[ch2][0].imag()) + A[ch1][1].imag() * (-B[ch2][1].imag());
            imag  = A[ch1][0].imag() *   B[ch2][0].real()  + A[ch1][1].imag() *   B[ch2][1].real();
            imag += A[ch1][0].real() * (-B[ch2][0].imag()) + A[ch1][1].real() * (-B[ch2][1].imag());
            C[ch1][ch2] = {real, imag};
        }
    }
}

void matrixMult(const hoatMatrixCmplx2D &A,
                const hoatMatrixCmplx2D &B,
                hoatMatrixCmplx2D &C)
{
    if (A.nRows == A.nCols && A.nRows == 2 && B.nRows == B.nCols && B.nRows == 2 && C.nRows == C.nCols && C.nRows == 2) {
        matrixMult2x2(A, B, C);
        return;
    }
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nRows == C.nRows && B.nCols == C.nCols && A.nCols == B.nRows);
    const float_complex calpha = float_complex(1.0f, 0.0f), cbeta = float_complex(0.0f, 0.0f);
    cblas_cgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, A.nRows, B.nCols, A.nCols, &calpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, &cbeta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMult(const hoatMatrix2D &A,
                const hoatMatrix2D &B,
                hoatMatrix2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nRows == C.nRows && B.nCols == C.nCols && A.nCols == B.nRows);
    const float alpha = 1.0f, beta = 0.0f;
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, A.nRows, B.nCols, A.nCols, alpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, beta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMultConjNoTrans(const hoatMatrixCmplx2D &A,
                           const hoatMatrixCmplx2D &B,
                           hoatMatrixCmplx2D &C)
{
    if (A.nRows == A.nCols && A.nRows == 2 && B.nRows == B.nCols && B.nRows == 2 && C.nRows == C.nCols && C.nRows == 2) {
        matrixMultConjNoTrans2x2(A, B, C);
        return;
    }
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nCols == C.nRows && B.nCols == C.nCols && A.nRows == B.nRows);
    const float_complex calpha = float_complex(1.0f, 0.0f), cbeta = float_complex(0.0f, 0.0f);
    cblas_cgemm(CblasRowMajor, CblasConjTrans, CblasNoTrans, A.nCols, B.nCols, A.nRows, &calpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, &cbeta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMultTransNoTrans(const hoatMatrixCmplx2D &A,
                            const hoatMatrixCmplx2D &B,
                            hoatMatrixCmplx2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nCols == C.nRows && B.nCols == C.nCols && A.nRows == B.nRows);
    const float_complex calpha = float_complex(1.0f, 0.0f), cbeta = float_complex(0.0f, 0.0f);
    cblas_cgemm(CblasRowMajor, CblasTrans, CblasNoTrans, A.nCols, B.nCols, A.nRows, &calpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, &cbeta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMultTransNoTrans(const hoatMatrix2D &A,
                            const hoatMatrix2D &B,
                            hoatMatrix2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nCols == C.nRows && B.nCols == C.nCols && A.nRows == B.nRows);
    const float alpha = 1.0f, beta = 0.0f;
    cblas_sgemm(CblasRowMajor, CblasTrans, CblasNoTrans, A.nCols, B.nCols, A.nRows, alpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, beta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMultNoTransConj(const hoatMatrixCmplx2D &A,
                           const hoatMatrixCmplx2D &B,
                           hoatMatrixCmplx2D &C)
{
    if (A.nRows == A.nCols && A.nRows == 2 && B.nRows == B.nCols && B.nRows == 2 && C.nRows == C.nCols && C.nRows == 2) {
        matrixMultNoTransConj2x2(A, B, C);
        return;
    }
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nRows == C.nRows && B.nRows == C.nCols && A.nCols == B.nCols);
    const float_complex calpha = float_complex(1.0f, 0.0f), cbeta = float_complex(0.0f, 0.0f);
    cblas_cgemm(CblasRowMajor, CblasNoTrans, CblasConjTrans, A.nRows, B.nRows, A.nCols, &calpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, &cbeta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMultNoTransTrans(const hoatMatrixCmplx2D &A,
                            const hoatMatrixCmplx2D &B,
                            hoatMatrixCmplx2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nRows == C.nRows && B.nRows == C.nCols && A.nCols == B.nCols);
    const float_complex calpha = float_complex(1.0f, 0.0f), cbeta = float_complex(0.0f, 0.0f);
    cblas_cgemm(CblasRowMajor, CblasNoTrans, CblasTrans, A.nRows, B.nRows, A.nCols, &calpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, &cbeta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMultNoTransTrans(const hoatMatrix2D &A,
                            const hoatMatrix2D &B,
                            hoatMatrix2D &C)
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    assert(A.nRows == C.nRows && B.nRows == C.nCols && A.nCols == B.nCols);
    const float alpha = 1.0f, beta = 0.0f;
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasTrans, A.nRows, B.nRows, A.nCols, alpha,
                FLATTEN2D(A.data), A.nCols,
                FLATTEN2D(B.data), B.nCols, beta,
                FLATTEN2D(C.data), C.nCols);
}

void matrixMult3x3(const float A[3 * 3],
                   const float B[3 * 3],
                   float C[3 * 3])
{
    assert(&A != &C && &B != &C && "C can not point to A or B");
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, 3, 3, 3, 1.0f,
                A, 3,
                B, 3, 0.0f,
                C, 3);
}

void matrixScalarMult2x2(float multiplier,
                         const hoatMatrixCmplx2D &A,
                         hoatMatrixCmplx2D &C)
{
    for (int ch1 = 0; ch1 < 2; ch1++) {
        for (int ch2 = 0; ch2 < 2; ch2++) {
            C[ch1][ch2] = {multiplier * A[ch1][ch2].real(), multiplier * A[ch1][ch2].imag()};
        }
    }
}

void matrixScalarMult(float multiplier,
                      const hoatMatrixCmplx2D &A,
                      hoatMatrixCmplx2D &C)
{
    assert(A.nRows == C.nRows && A.nCols == C.nCols);

    if (A.nRows == A.nCols && A.nRows == 2) {
        matrixScalarMult2x2(multiplier, A, C);
        return;
    }

    float_complex mult{multiplier, 0.0f};
    int multLen = A.nRows * A.nCols;
    utility_cvsmul(FLATTEN2D(A.data),
                   &mult,
                   multLen,
                   FLATTEN2D(C.data));
}

void matrixScalarMult(float multiplier,
                      const hoatMatrix2D &A,
                      hoatMatrix2D &C)
{
    assert(A.nRows == C.nRows && A.nCols == C.nCols);
    int multLen = A.nRows * A.nCols;
    utility_svsmul(FLATTEN2D(A.data),
                   &multiplier,
                   multLen,
                   FLATTEN2D(C.data));
}

void matrixMatrixSub(const hoatMatrixCmplx2D &A,
                     const hoatMatrixCmplx2D &B,
                     hoatMatrixCmplx2D &C)
{
    assert(A.nRows == B.nRows && A.nRows == C.nRows && A.nCols == B.nCols && A.nCols == C.nCols);
    for (int row = 0; row < A.nRows; ++row) {
        for (int col = 0; col < A.nCols; ++col) {
            C[row][col] = A[row][col] - B[row][col];
        }
    }
}

void matrixMatrixAdd(const hoatMatrixCmplx2D &A,
                     const hoatMatrixCmplx2D &B,
                     hoatMatrixCmplx2D &C)
{
    assert(A.nRows == B.nRows && A.nRows == C.nRows && A.nCols == B.nCols && A.nCols == C.nCols);
    int addLen = A.nRows * A.nCols;
    utility_cvvadd(FLATTEN2D(A.data),
                   FLATTEN2D(B.data),
                   addLen,
                   FLATTEN2D(C.data));
}

void matrixMatrixAdd(const hoatMatrixCmplx3D &A,
                     const hoatMatrixCmplx3D &B,
                     hoatMatrixCmplx3D &C)
{
    assert(A.n2DMatrices == B.n2DMatrices && A.n2DMatrices == C.n2DMatrices);
    for (int n = 0; n < A.n2DMatrices; ++n) {
        matrixMatrixAdd(A[n], B[n], C[n]);
    }
}

void matrixMatrixAdd(const hoatMatrix2D &A,
                     const hoatMatrix2D &B,
                     hoatMatrix2D &C)
{
    assert(A.nRows == B.nRows && A.nRows == C.nRows && A.nCols == B.nCols && A.nCols == C.nCols);
    int addLen = A.nRows * A.nCols;
    utility_svvadd(FLATTEN2D(A.data),
                   FLATTEN2D(B.data),
                   addLen,
                   FLATTEN2D(C.data));
}

void matrixMatrixSub(const hoatMatrix2D &A,
                     const hoatMatrix2D &B,
                     hoatMatrix2D &C)
{
    assert(A.nRows == B.nRows && A.nRows == C.nRows && A.nCols == B.nCols && A.nCols == C.nCols);
    for (int row = 0; row < A.nRows; ++row) {
        for (int col = 0; col < A.nCols; ++col) {
            C[row][col] = A[row][col] - B[row][col];
        }
    }
}

void normalize1dMatrix(hoatMatrix2D &M)
{
    assert(M.nRows == 1 || M.nCols == 1);
    float norm = 0.0f;
    for (int row = 0; row < M.nRows; ++row) {
        for (int col = 0; col < M.nCols; ++col) {
            norm += std::pow(M[row][col], 2.0f);
        }
    }
    norm = std::sqrt(norm);
    matrixScalarMult(1.0f / norm, M, M);
}

float sumDiagonalReal(hoatMatrixCmplx2D &M)
{
    assert(M.nRows == M.nCols);
    float sum = 0.0f;
    for (int i = 0; i < M.nRows; ++i) {
        sum += M[i][i].real();
    }
    return sum;
}

float sumDiagonal(hoatMatrix2D &M)
{
    assert(M.nRows == M.nCols);
    float sum = 0.0f;
    for (int i = 0; i < M.nRows; ++i) {
        sum += M[i][i];
    }
    return sum;
}

void copyDiagonal(const hoatMatrixCmplx2D &source, hoatMatrixCmplx2D &destination)
{
    assert(source.nRows == source.nCols && source.nRows == destination.nRows && source.nCols == destination.nCols);
    for (int i = 0; i < source.nRows; ++i) {
        destination[i][i] = source[i][i];
    }
}

void copy(hoatMatrixCmplx2D &source, hoatMatrixCmplx2D &destination)
{
    assert(source.nRows == destination.nRows && source.nCols == destination.nCols);
    std::copy_n(FLATTEN2D(source.data), source.nRows * source.nCols, FLATTEN2D(destination.data));
}

void copy(hoatMatrixCmplx3D &source, hoatMatrixCmplx3D &destination)
{
    assert(source.n2DMatrices == destination.n2DMatrices);
    for (int n = 0; n < source.n2DMatrices; ++n) {
        copy(source[n], destination[n]);
    }
}

void copy(hoatMatrix2D &source, hoatMatrix2D &destination)
{
    assert(source.nRows == destination.nRows && source.nCols == destination.nCols);
    std::copy_n(FLATTEN2D(source.data), source.nRows * source.nCols, FLATTEN2D(destination.data));
}

void fillWithZeros(hoatMatrixCmplx2D &M) {
    std::fill_n(FLATTEN2D(M.data), M.nRows * M.nCols, float_complex(0.0f, 0.0f));
}

void fillWithZeros(hoatMatrixCmplx3D &M)
{
    for (int n = 0; n < M.n2DMatrices; ++n) {
        fillWithZeros(M[n]);
    }
}

void fillWithZeros(hoatMatrix2D &M) {
    std::fill_n(FLATTEN2D(M.data), M.nRows * M.nCols, 0.0f);
}

void fillWithZeros(hoatMatrix3D &M)
{
    for (int n = 0; n < M.n2DMatrices; ++n) {
        fillWithZeros(M[n]);
    }
}

void fillWithZeros(hoatMatrix4D &M)
{
    for (int n = 0; n < M.n3DMatrices; ++n) {
        fillWithZeros(M[n]);
    }
}

void print(hoatMatrix2D &M)
{
    for (int row = 0; row < M.nRows; ++row) {
        for (int col = 0; col < M.nCols; ++col) {
            std::cout << std::setw(15) << M[row][col] << ", ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

void print(hoatMatrixCmplx2D &M)
{
    for (int row = 0; row < M.nRows; ++row) {
        for (int col = 0; col < M.nCols; ++col) {
            std::cout << "(" << std::setw(15) << M[row][col].real() << ", " << std::setw(15) << M[row][col].imag() << "), ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

void combineIntoCmplxMatrix(const hoatMatrix2D &realM, const hoatMatrix2D &imagM, hoatMatrixCmplx2D &cmplxM)
{
    assert(realM.nCols == imagM.nCols && realM.nRows == imagM.nRows && realM.nCols == cmplxM.nCols && realM.nRows == cmplxM.nRows);
    for (int row = 0; row < realM.nRows; ++row) {
        for (int col = 0; col < realM.nCols; ++col) {
            cmplxM[row][col] = {realM[row][col], imagM[row][col]};
		}
	}
}

void realToComplex(hoatMatrix2D &real, hoatMatrixCmplx2D &complex)
{
    assert(real.nRows == complex.nRows && real.nCols == complex.nCols);
    for (int row = 0; row < real.nRows; ++row) {
        for (int col = 0; col < real.nCols; ++col) {
            complex[row][col] = float_complex(real[row][col], 0.0f);
        }
    }
}

void createHoatMatrix(hoatMatrixCmplx2D* matrix, int nRows, int nCols)
{
    matrix->nRows = nRows;
    matrix->nCols = nCols;
    matrix->data = allocateFloatCmplxArray2d(matrix->nRows, matrix->nCols);
}

void createHoatMatrix(hoatMatrixCmplx3D* matrix, int n2DMatrices, int nRows, int nCols)
{
    matrix->n2DMatrices = n2DMatrices;
    matrix->matrices2D = new hoatMatrixCmplx2D[matrix->n2DMatrices];
    for (int n = 0; n < matrix->n2DMatrices; ++n) {
        createHoatMatrix(&(matrix->matrices2D[n]), nRows, nCols);
    }
}

void createHoatMatrix(hoatMatrix2D* matrix, int nRows, int nCols)
{
    matrix->nRows = nRows;
    matrix->nCols = nCols;
    matrix->data = allocateFloatArray2d(matrix->nRows, matrix->nCols);
}

void createHoatMatrix(hoatMatrix3D* matrix, int n2DMatrices, int nRows, int nCols)
{
    matrix->n2DMatrices = n2DMatrices;
    matrix->matrices2D = new hoatMatrix2D[matrix->n2DMatrices];
    for (int n = 0; n < matrix->n2DMatrices; ++n) {
        createHoatMatrix(&(matrix->matrices2D[n]), nRows, nCols);
    }
}

void createHoatMatrix(hoatMatrix4D* matrix, int n3DMatrices, int n2DMatrices, int nRows, int nCols)
{
    matrix->n2DMatrices = n2DMatrices;
    matrix->n3DMatrices = n3DMatrices;
    matrix->matrices3D = new hoatMatrix3D[matrix->n3DMatrices];
    for (int n = 0; n < matrix->n3DMatrices; ++n) {
        createHoatMatrix(&(matrix->matrices3D[n]), n2DMatrices, nRows, nCols);
    }
}

void destroyHoatMatrix(hoatMatrixCmplx2D* matrix){
    free(matrix->data);
}

void destroyHoatMatrix(hoatMatrixCmplx3D* matrix)
{
    for (int n = 0; n < matrix->n2DMatrices; ++n) {
        destroyHoatMatrix(&(matrix->matrices2D[n]));
    }
    delete[] matrix->matrices2D;
}

void destroyHoatMatrix(hoatMatrix2D* matrix)
{
    free(matrix->data);
}

void destroyHoatMatrix(hoatMatrix3D* matrix)
{
    for (int n = 0; n < matrix->n2DMatrices; ++n) {
        destroyHoatMatrix(&(matrix->matrices2D[n]));
    }
    delete[] matrix->matrices2D;
}

void destroyHoatMatrix(hoatMatrix4D* matrix)
{
    for (int n = 0; n < matrix->n3DMatrices; ++n) {
        destroyHoatMatrix(&(matrix->matrices3D[n]));
    }
    delete[] matrix->matrices3D;
}

float* allocateFloatArray1d(int dim1) {
    return (float*)calloc1d(dim1, sizeof(float));
}

float** allocateFloatArray2d(int dim1, int dim2) {
    return (float**)calloc2d(dim1, dim2, sizeof(float));
}

float*** allocateFloatArray3d(int dim1, int dim2, int dim3) {
    return (float***)calloc3d(dim1, dim2, dim3, sizeof(float));
}

int* allocateIntArray1d(int dim1) {
    return (int*)calloc1d(dim1, sizeof(int));
}

float_complex* allocateFloatCmplxArray1d(int dim1) {
    return (float_complex*)calloc1d(dim1, sizeof(float_complex));
}

float_complex** allocateFloatCmplxArray2d(int dim1, int dim2) {
    return (float_complex**)calloc2d(dim1, dim2, sizeof(float_complex));
}

float_complex*** allocateFloatCmplxArray3d(int dim1, int dim2, int dim3) {
    return (float_complex***)calloc3d(dim1, dim2, dim3, sizeof(float_complex));
}

float_complex**** allocateFloatCmplxArray4d(int dim1, int dim2, int dim3, int dim4) {
    auto temp = (float_complex****)calloc1d(dim1, sizeof(float_complex));
    for (int i = 0; i < dim1; ++i) {
        temp[i] = allocateFloatCmplxArray3d(dim2, dim3, dim4);
    }
    return temp;
}
