/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_HOATMATRIX_HPP
#define HOA_TRANSLATION_HOATMATRIX_HPP

#include <vector>

#include <saf.h>

/* hoatMatrixXXX structs are used to store matrix dimensions and data into a single struct.
 * The custom matrices are then used in custom matrix functions, that simplify the layer
 * between cblas-routines and some vector routines from SAF.
 * NOTE: the data in a 2D matrix is stored in a row-major order, i.e. the data can be accessed as data[rowIndx][colIndx]
 */
typedef struct hoatMatrixCmplx2D {
    int nRows;
    int nCols;
    float_complex **data;
    const float_complex *operator[](int i) const;
    float_complex *&operator[](int i);
}hoatMatrixCmplx2D;

typedef struct hoatMatrixCmplx3D {
    int n2DMatrices;
    hoatMatrixCmplx2D *matrices2D;
    const hoatMatrixCmplx2D operator[](int i) const;
    hoatMatrixCmplx2D &operator[](int i);
}hoatMatrixCmplx3D;

typedef struct hoatMatrix2D {
    int nRows;
    int nCols;
    float **data;
    const float *operator[](int i) const;
    float *&operator[](int i);
}hoatMatrix2D;

typedef struct hoatMatrix3D {
    int n2DMatrices;
    hoatMatrix2D *matrices2D;
    const hoatMatrix2D operator[](int i) const;
    hoatMatrix2D &operator[](int i);
}hoatMatrix3D;

typedef struct hoatMatrix4D {
    int n3DMatrices;
    int n2DMatrices;
    hoatMatrix3D *matrices3D;
    const hoatMatrix3D operator[](int i) const;
    hoatMatrix3D &operator[](int i);
}hoatMatrix4D;

// Matrix multiplication C = A * B
void matrixMult(const hoatMatrixCmplx2D &A,
                const hoatMatrixCmplx2D &B,
                hoatMatrixCmplx2D &C);

// Matrix multiplication C = A * B
void matrixMult(const hoatMatrix2D &A,
                const hoatMatrix2D &B,
                hoatMatrix2D &C);

// Matrix multiplication C = A^H * B
void matrixMultConjNoTrans(const hoatMatrixCmplx2D &A,
                           const hoatMatrixCmplx2D &B,
                           hoatMatrixCmplx2D &C);

// Matrix multiplication C = A^T * B
void matrixMultTransNoTrans(const hoatMatrixCmplx2D &A,
                            const hoatMatrixCmplx2D &B,
                            hoatMatrixCmplx2D &C);

// Matrix multiplication C = A^T * B
void matrixMultTransNoTrans(const hoatMatrix2D &A,
                            const hoatMatrix2D &B,
                            hoatMatrix2D &C);

// Matrix multiplication C = A * B^H
void matrixMultNoTransConj(const hoatMatrixCmplx2D &A,
                           const hoatMatrixCmplx2D &B,
                           hoatMatrixCmplx2D &C);

// Matrix multiplication C = A * B^T
void matrixMultNoTransTrans(const hoatMatrixCmplx2D &A,
                            const hoatMatrixCmplx2D &B,
                            hoatMatrixCmplx2D &C);

// Matrix multiplication C = A * B^T
void matrixMultNoTransTrans(const hoatMatrix2D &A,
                            const hoatMatrix2D &B,
                            hoatMatrix2D &C);

// Matrix multiplication C = A * B (real values, 3x3 matrices)
void matrixMult3x3(const float A[3 * 3],
                   const float B[3 * 3],
                   float C[3 * 3]);

// Matrix scalar multiplication for a 2x2 matrix C = multiplier .* A
void matrixScalarMult2x2(float multiplier,
                         const hoatMatrixCmplx2D &A,
                         hoatMatrixCmplx2D &C);

// Matrix scalar multiplication C = multiplier .* A
void matrixScalarMult(float multiplier,
                      const hoatMatrixCmplx2D &A,
                      hoatMatrixCmplx2D &C);

// Matrix scalar multiplication C = multiplier .* A
void matrixScalarMult(float multiplier,
                      const hoatMatrix2D &A,
                      hoatMatrix2D &C);

// Matrix matrix addition C = A + B
void matrixMatrixAdd(const hoatMatrixCmplx2D &A,
                     const hoatMatrixCmplx2D &B,
                     hoatMatrixCmplx2D &C);

// Matrix matrix addition C = A + B
void matrixMatrixAdd(const hoatMatrixCmplx3D &A,
                     const hoatMatrixCmplx3D &B,
                     hoatMatrixCmplx3D &C);

// Matrix matrix addition C = A + B
void matrixMatrixAdd(const hoatMatrix2D &A,
                     const hoatMatrix2D &B,
                     hoatMatrix2D &C);

// Matrix matrix subtraction C = A - B
void matrixMatrixSub(const hoatMatrixCmplx2D &A,
                     const hoatMatrixCmplx2D &B,
                     hoatMatrixCmplx2D &C);

// Matrix matrix subtraction C = A - B
void matrixMatrixSub(const hoatMatrix2D &A,
                     const hoatMatrix2D &B,
                     hoatMatrix2D &C);

// Normalize vector (1D matrix) M = M / norm(M)
void normalize1dMatrix(hoatMatrix2D &M);

// Return the sum of real diagonal values of M, i.e. trace(real(M))
float sumDiagonalReal(hoatMatrixCmplx2D &M);
float sumDiagonal(hoatMatrix2D &M);

// Miscellaneous operations
void copyDiagonal(const hoatMatrixCmplx2D &source, hoatMatrixCmplx2D &destination);
void copy(hoatMatrixCmplx2D &source, hoatMatrixCmplx2D &destination);
void copy(hoatMatrixCmplx3D &source, hoatMatrixCmplx3D &destination);
void copy(hoatMatrix2D &source, hoatMatrix2D &destination);
void fillWithZeros(hoatMatrixCmplx2D &M);
void fillWithZeros(hoatMatrixCmplx3D &M);
void fillWithZeros(hoatMatrix2D &M);
void fillWithZeros(hoatMatrix3D &M);
void fillWithZeros(hoatMatrix4D &M);
void combineIntoCmplxMatrix(const hoatMatrix2D &realM, const hoatMatrix2D &imagM, hoatMatrixCmplx2D &cmplxM);
void print(hoatMatrix2D &M);
void print(hoatMatrixCmplx2D &M);
void realToComplex(hoatMatrix2D &real, hoatMatrixCmplx2D &complex);

// Matrix creation and deletion
void createHoatMatrix(hoatMatrixCmplx2D* matrix, int nRows, int nCols);
void createHoatMatrix(hoatMatrixCmplx3D* matrix, int n2DMatrices, int nRows, int nCols);
void createHoatMatrix(hoatMatrix2D* matrix, int nRows, int nCols);
void createHoatMatrix(hoatMatrix3D* matrix, int n2DMatrices, int nRows, int nCols);
void createHoatMatrix(hoatMatrix4D* matrix, int n3DMatrices, int n2DMatrices, int nRows, int nCols);
void destroyHoatMatrix(hoatMatrixCmplx2D* matrix);
void destroyHoatMatrix(hoatMatrixCmplx3D* matrix);
void destroyHoatMatrix(hoatMatrix2D* matrix);
void destroyHoatMatrix(hoatMatrix3D* matrix);
void destroyHoatMatrix(hoatMatrix4D* matrix);

// Memory allocations
float* allocateFloatArray1d(int dim1);
float** allocateFloatArray2d(int dim1, int dim2);
float*** allocateFloatArray3d(int dim1, int dim2, int dim3);
int* allocateIntArray1d(int dim1);
float_complex* allocateFloatCmplxArray1d(int dim1);
float_complex** allocateFloatCmplxArray2d(int dim1, int dim2);
float_complex*** allocateFloatCmplxArray3d(int dim1, int dim2, int dim3);
float_complex**** allocateFloatCmplxArray4d(int dim1, int dim2, int dim3, int dim4);

#endif //HOA_TRANSLATION_HOATMATRIX_HPP
