/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "HrtfOperations.hpp"

#include <algorithm>
#include <numeric>

#include "SpatialAnalysis.hpp"

void get_HRTFs(hoaTranslatorData *h)
{
    if (!h->useMatlabFile || !load_HRTFs_from_matlab_file(h)) {
        if (h->sofaFilePath.empty() || saf_sofa_open(&h->safSofaContainer, h->sofaFilePath.data()) != SAF_SOFA_OK) {
            // Use default HRIR set from SAF
            loadSofaFile(NULL,
                         &(h->hrirs),
                         &(h->hrtfDirsDeg),
                         &(h->nHrtfDirs),
                         &(h->hrirLen),
                         &(h->hrirFs));
        } else {
            // Load HRIR data from SOFA file
            // NOTE: this hasn't been tested with a real SOFA file yet
            h->nHrtfDirs = h->safSofaContainer.nSources;
            h->hrirLen = h->safSofaContainer.DataLengthIR;
            h->hrirFs = (int)h->safSofaContainer.DataSamplingRate;
            h->hrirs = allocateFloatArray1d(h->nHrtfDirs * h->nBinauralChannels * h->hrirLen);
            h->hrtfDirsDeg = allocateFloatArray1d(h->nHrtfDirs * h->nBinauralChannels);

            std::copy_n(h->safSofaContainer.DataIR, h->nHrtfDirs * h->nBinauralChannels * h->hrirLen, h->hrirs);

            if (strcmp(h->safSofaContainer.SourcePositionType, "spherical") == 0) {
                // The data in SourcePosition is stored as [azimuth, elevation, radius], copy only
                // azimuth and elevation angles
                int pos = 0;
                for (int dir = 0; dir < h->nHrtfDirs * h->nBinauralChannels; dir += h->nBinauralChannels) {
                    h->hrtfDirsDeg[dir] = h->safSofaContainer.SourcePosition[pos];
                    h->hrtfDirsDeg[dir + 1] = h->safSofaContainer.SourcePosition[pos + 1];
                    pos += 3;
                }
            } else {
                // The data in SourcePosition is stored as [x, y ,z], convert them to
                // azimuth and elevation angles
                std::vector<float> sourcePos(3, 0.0f);
                int pos = 0;
                for (int dir = 0; dir < h->nHrtfDirs * h->nBinauralChannels; dir += h->nBinauralChannels) {
                    for (int i = 0; i < 3; ++i) {
                        sourcePos[i] = h->safSofaContainer.SourcePosition[pos + i];
                    }
                    h->hrtfDirsDeg[dir] = calculate_azimuth(sourcePos.data()) * 180.0f / M_PI;
                    h->hrtfDirsDeg[dir + 1] = calculate_elevation(sourcePos.data()) * 180.0f / M_PI;
                    pos += 3;
                }
            }
        }
        h->itds = allocateFloatArray1d(h->nHrtfDirs);
        estimateITDs(h->hrirs,
                     h->nHrtfDirs,
                     h->hrirLen,
                     h->hrirFs,
                     h->itds);
        h->hrtfFB = allocateFloatCmplxArray3d(h->nBands, h->nBinauralChannels, h->nHrtfDirs);
        HRIRs2HRTFs_afSTFT(h->hrirs, h->nHrtfDirs, h->hrirLen, h->hopSize, h->LDmode, h->hybridMode, FLATTEN3D(h->hrtfFB));
        diffuseFieldEqualiseHRTFs(h->nHrtfDirs,
                                  h->itds,
                                  h->freqVec,
                                  h->nBands,
                                  NULL,
                                  1,
                                  1,
                                  FLATTEN3D(h->hrtfFB));
    }
}

bool load_HRTFs_from_matlab_file(hoaTranslatorData *h)
{
    return read_matlab_dirs_file(h) && read_matlab_itd_file(h) && read_matlab_hrtf_file(h);
}

bool read_matlab_dirs_file(hoaTranslatorData *h)
{
    std::ifstream input(h->matlabDirsFile);
    if (!input) {
        assert(false && "Couldn't read matlabDirsFile.");
        return false;
    }

    // Read the header
    char data[100];
    input >> data;
    char *rest = data;
    char *token = strtok_x(data, ",", &rest);
    while (token != NULL) {
        char *end_token;
        char *token2 = strtok_x(token, ":", &end_token);
        if (strcmp(token2,"nDirections") == 0) {
            h->nHrtfDirs = std::stoi(strtok_x(NULL, ":", &end_token));
            break;
        }
        token = strtok_x(NULL, ",", &rest);
    }

    // Read the direction data
    h->hrtfDirsDeg = allocateFloatArray1d(h->nHrtfDirs * h->nBinauralChannels);
    int index = 0;
    float azi = 0;
    float elev = 0;
    input >> data;
    while (!input.eof()) {
        azi = std::stof(strtok_x(data, ",", &rest));
        elev = std::stof(strtok_x(NULL, ",", &rest));
        h->hrtfDirsDeg[index] = azi;
        h->hrtfDirsDeg[index + 1] = elev;
        input >> data;
        index += 2;
    }
    input.close();
    return true;
}

bool read_matlab_itd_file(hoaTranslatorData *h)
{
    std::ifstream input(h->matlabItdFile);
    if (!input) {
        assert(false && "Couldn't read matlabItdFile.");
        return false;
    }
    char data[100];
    input >> data; // flush the header
    h->itds = allocateFloatArray1d(h->nHrtfDirs);
    int index = 0;
    float itd = 0;
    input >> data;
    while (!input.eof()) {
        itd = std::stof(data);
        h->itds[index] = itd;
        input >> data;
        ++index;
    }
    input.close();
    return true;
}
bool read_matlab_hrtf_file(hoaTranslatorData *h)
{
    std::ifstream input(h->matlabHrtfFile);
    if (!input) {
        assert(false && "Couldn't read matlabHrtfFile.");
        return false;
    }

    // Read the header
    char data[100];
    input >> data;
    char *rest;
    char *token = strtok_x(data, ",", &rest);
    while (token != NULL) {
        char *end_token;
        char *token2 = strtok_x(token, ":", &end_token);
        if (strcmp(token2, "nBands") == 0) {
            int nBandsInHRTFfile = std::stoi(strtok_x(NULL, ":", &end_token));
            if (nBandsInHRTFfile != h->nBands) {
                input.close();
                assert(false &&
                       "The amount frequency bins used in the matlab HRTF file does not match with the amount of bins set for afSTFT. Check that 'matlabHRTFFile' is set correctly in the scene file.");
                return false;
            }
        } else if (strcmp(token2,"fs") == 0) {
            h->hrirFs = std::stoi(strtok_x(NULL, ":", &end_token));
        }
        token = strtok_x(NULL, ",", &rest);

    }

    // Read the HRTF data
    h->hrtfFB = allocateFloatCmplxArray3d(h->nBands, h->nBinauralChannels, h->nHrtfDirs);
    float real = 0;
    float imag = 0;
    for (int band = 0; band < h->nBands; ++band) {
        for (int ch = 0; ch < h->nBinauralChannels; ++ch) {
            for (int dir = 0; dir < h->nHrtfDirs; ++dir) {
                input >> data;
                real = std::stof(strtok_x(data, ",", &rest));
                imag = std::stof(strtok_x(NULL, ",", &rest));
                h->hrtfFB[band][ch][dir] = float_complex(real, imag);
            }
        }
    }
    input.close();
    return true;
}

// Returns the directions of a grid for a certain resolution
void get_grid_directions(int aziResDeg,
                         int elevResDeg,
                         int nAziAngles,
                         int nElevAngles,
                         float *dirs,
                         int aziFirst)
{
    std::vector<int> azi(nAziAngles);
    std::vector<int> elev(nElevAngles);
    int aziIndex = 0;
    for (int aziAngl = -180; aziAngl <= 180; aziAngl += aziResDeg) {
        azi[aziIndex] = aziAngl;
        ++aziIndex;
    }
    int elevIndex = 0;
    for (int elevAngl = -90; elevAngl <= 90; elevAngl += elevResDeg) {
        elev[elevIndex] = elevAngl;
        ++elevIndex;
    }
    int dirIndex = 0;
    if (aziFirst) {
        // Directions where azimuth angle changes first [(azi0 elev0) (azi1 elev0)...]
        for (int elevIndex = 0; elevIndex < nElevAngles; ++elevIndex) {
            for (int aziIndex = 0; aziIndex < nAziAngles; ++aziIndex) {
                dirs[dirIndex] = azi[aziIndex];
                dirs[dirIndex + 1] = elev[elevIndex];
                dirIndex += 2;
            }
        }
    } else {
        // Directions where elevation angle changes first [(azi0 elev0) (azi0 elev1)...]
        for (int aziIndex = 0; aziIndex < nAziAngles; ++aziIndex) {
            for (int elevIndex = 0; elevIndex < nElevAngles; ++elevIndex) {
                dirs[dirIndex] = azi[aziIndex];
                dirs[dirIndex + 1] = elev[elevIndex];
                dirIndex += 2;
            }
        }
    }
}

void nearest_neighbor_hrtf_interpolation(hoaTranslatorData *h)
{
    int aziFirst = 1; // 1: interpolated directions stored as [(azi0 elev0) (azi1 elev0)...]; 0: interpolated directions stored as [(azi0 elev0) (azi0 elev1)...]
    h->interpolatedHrtfAziResDeg = 2;
    h->interpolatedHrtfElevResDeg = 5;
    h->nInterpolatedHrtfAziAngles = std::round(360.0f / h->interpolatedHrtfAziResDeg) + 1;
    h->nInterpolatedHrtfElevAngles = std::round(180.0f / h->interpolatedHrtfElevResDeg) + 1;
    h->nInterpolatedHrtfDirs = h->nInterpolatedHrtfAziAngles * h->nInterpolatedHrtfElevAngles;

    // Generate a grid of the desired resolution, and generate a VBAP gain table for the grid
    float *gridDirsDeg = allocateFloatArray1d((h->nInterpolatedHrtfDirs * h->nBinauralChannels));
    h->hrtfIdxGrid = allocateIntArray1d(h->nInterpolatedHrtfDirs);
    get_grid_directions(h->interpolatedHrtfAziResDeg,
                        h->interpolatedHrtfElevResDeg,
                        h->nInterpolatedHrtfAziAngles,
                        h->nInterpolatedHrtfElevAngles,
                        gridDirsDeg,
                        aziFirst);
    find_closest_grid_points(gridDirsDeg,
                             h->hrtfDirsDeg,
                             h->hrtfIdxGrid,
                             h->nInterpolatedHrtfDirs,
                             h->nHrtfDirs);
    free(gridDirsDeg);
}

void find_closest_grid_points(float *gridDirsDeg,
                              float *hrtfDirsDeg,
                              int *hrtfIdxGrid,
                              int nGridDirs,
                              int nHrtfDirs)
{
    hoatMatrix2D gridDirsXYZ;
    hoatMatrix2D hrtfDirsXYZ;
    hoatMatrix2D dotDirs;
    createHoatMatrix(&gridDirsXYZ, nGridDirs, 3);
    createHoatMatrix(&hrtfDirsXYZ, nHrtfDirs, 3);
    createHoatMatrix(&dotDirs, nGridDirs, nHrtfDirs);
    unitSph2cart(gridDirsDeg, nGridDirs, 1, FLATTEN2D(gridDirsXYZ.data));
    unitSph2cart(hrtfDirsDeg, nHrtfDirs, 1, FLATTEN2D(hrtfDirsXYZ.data));
    float *maxDotDirs = nullptr;

    // Dot product between grid and HRTF directions
    matrixMultNoTransTrans(gridDirsXYZ, hrtfDirsXYZ, dotDirs);

    for (int gridDirIdx = 0; gridDirIdx < nGridDirs; ++gridDirIdx) {
        maxDotDirs = std::max_element(&dotDirs[gridDirIdx][0], &dotDirs[gridDirIdx][dotDirs.nCols - 1]);
        hrtfIdxGrid[gridDirIdx] = std::distance(&dotDirs[gridDirIdx][0], maxDotDirs);
    }
    destroyHoatMatrix(&gridDirsXYZ);
    destroyHoatMatrix(&hrtfDirsXYZ);
    destroyHoatMatrix(&dotDirs);
}
