/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_HRTFOPERATIONS_HPP
#define HOA_TRANSLATION_HRTFOPERATIONS_HPP

#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
#ifndef strtok_x
    #define strtok_x strtok_s
    #endif
#else
#ifndef strtok_x
#define strtok_x strtok_r
#endif
#endif

#include "Utilities.hpp"

/* Load a set of Head-related Impulse Responses (HRIR) and transform them into Head-related Transfer Functions (HRTF).
 * The HRIR data can be read either from a SOFA file, or from a default set inside SAF. The HRTFs can also be directly
 * read from a text file generated in Matlab.
 *
 * NOTE: this function allocates memory for hrtfFB, remember to free it later.
 *
 * Output:
 *      hrtfFB    - HRTF filter bank
 */
void get_HRTFs(hoaTranslatorData *h);

// Load HRTFs, source directions, and ITD data from text files generated in Matlab
bool load_HRTFs_from_matlab_file(hoaTranslatorData *h);

// Read HRTF related source direction data from a text file generated in Matlab
bool read_matlab_dirs_file(hoaTranslatorData *h);

// Read HRTF related Interaural Time Difference (ITD) data from a text file generated in Matlab
bool read_matlab_itd_file(hoaTranslatorData *h);

// Read HRTF data from a text file generated in Matlab
bool read_matlab_hrtf_file(hoaTranslatorData *h);

/* Returns the directions of a grid for a certain resolution,
 * adapted from the MATLAB library by Archontis Politis,
 * found here: https://github.com/polarch/Vector-Base-Amplitude-Panning.
 */
void get_grid_directions(int aziResDeg,
                         int elevResDeg,
                         int nAziAngles,
                         int nElevAngles,
                         float *dirs,
                         int aziFirst);

/* Creates a custom grid of points, and picks the closest HRTFs for each point. Sets h->hrtfIdxGrid, which can be used
 * as a table to find the closest HRTF index for a certain grid direction index. E.g. h->hrtfFB( h->hrtfIdxGrid(42) ),
 * would return the closest HRTF for grid direction index 42.
 */
void nearest_neighbor_hrtf_interpolation(hoaTranslatorData *h);

/* Find closest HRTFs for each each point in a custom grid of points.
 *
 *  Input:
 *      gridDirsDeg     - grid directions in degrees (size: nGridDirs * 2)
 *      hrtfDirsDeg     - HRTF directions in degrees (size: nHrtfDirs * 2)
 *      hrtfIdxGrid     - handle to store the closest HRTF indexes (size: nGridDirs)
 *      nGridDirs       - number of directions in the grid
 *      nHrtfDirs       - number of HRTF directions
 *
 *  Output:
 *      hrtfIdxGrid     - Closest HRTF indexes for each point in the input grid. For example, for a grid direction index
 *                        of 42, get the closest HRTF direction index as hrtfIndx = hrtfIdxGrid(42). Then, the actual
 *                        HRTF can be obtained from the HRTF list as hrtf(hrtfIndx).
 *
 * Note: allocates memory for hrtfIdxGrid, remember to free it later.
 */
void find_closest_grid_points(float *gridDirsDeg,
                              float *hrtfDirsDeg,
                              int *hrtfIdxGrid,
                              int nGridDirs,
                              int nHrtfDirs);

#endif //HOA_TRANSLATION_HRTFOPERATIONS_HPP
