/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "MetadataInterpolator.hpp"

#include <algorithm>
#include <vector>

#include "Exterior.hpp"
#include "Rotator.hpp"
#include "SpatialAnalysis.hpp"

#ifdef SAF_USE_INTEL_MKL
#include "mkl.h"
#elif SAF_USE_APPLE_ACCELERATE
#include <Accelerate/Accelerate.h>
#endif

void vectorize_metadata(hoaTranslatorData *h, bool residualData)
{
	float ***azimuthData = h->azimuthData;
	float ***elevationData = h->elevationData;
	float ***DTRData = h->DTRData;

	if (residualData) {
		azimuthData = h->residualAzimuthData;
		elevationData = h->residualElevationData;
		DTRData = h->residualDTRData;
	}

    int vecLen = h->nFrames * h->nDoaDtrBins;
    if (h->useRecursiveAveraging) {
        vecLen = h->nDoaDtrBins;
    }

    for (int mic : h->activeArraysForCurrentFrame) {
        calculate_doa_vectors_OpenGL(h,
                                     vecLen,
                                     &azimuthData[mic][0][0],
                                     &elevationData[mic][0][0],
                                     &h->tempV1[mic][0][0],
                                     &h->tempV2[mic][0][0],
                                     &h->tempV3[mic][0][0]);

        // Apply DTR
        utility_svvmul(&h->tempV1[mic][0][0], &DTRData[mic][0][0], vecLen, &h->tempV1[mic][0][0]);
        utility_svvmul(&h->tempV2[mic][0][0], &DTRData[mic][0][0], vecLen, &h->tempV2[mic][0][0]);
        utility_svvmul(&h->tempV3[mic][0][0], &DTRData[mic][0][0], vecLen, &h->tempV3[mic][0][0]);
    }
}

void rotate_metadata_vectors(hoaTranslatorData *h)
{
    float Rsource[3][3]{};
    float Rmic[3][3]{};
    float R[3][3]{};

    for (int mic : h->activeArraysForCurrentFrame) {
        // Get rotation matrix Rmic for the mic array rotation
        float micYawDeg   = h->micArrayOrientations[mic][0];
        float micPitchDeg = h->micArrayOrientations[mic][1];
        float micRollDeg  = h->micArrayOrientations[mic][2];
        get_rotation_matrix_OpenGL(micYawDeg,
                                   micPitchDeg,
                                   micRollDeg,
                                   Rmic);

        if (h->useSphoa6DofExtensionMode || h->isListenerInsideTriangle || !h->useImprovedExteriorRendering) {
            // Get rotation matrix Rsource for the head rotation.
            // Rotating the head means that the source directions move to the opposite direction
            // (flip head rotation angles)
            float sourceYawDeg   = -h->yawDeg;
            float sourcePitchDeg = -h->pitchDeg;
            float sourceRollDeg  = -h->rollDeg;
            get_rotation_matrix_OpenGL(sourceYawDeg,
                                      sourcePitchDeg,
                                      sourceRollDeg,
                                      Rsource);

            // Get rotation matrix R = Rsource * Rmic
            matrixMult3x3(*Rsource, *Rmic, *R);
            std::copy_n(&R[0][0], 3 * 3, &h->tempR[0][0]);
        } else {
            // Skip head rotation, when the listener is outside the capturing area, and the improved exterior rendering is used
            // (the head rotation is handled after the exterior rendering).
            std::copy_n(&Rmic[0][0], 3 * 3, &h->tempR[0][0]);
        }

        // Apply rotations
        for (int frame = 0; frame < h->nFrames; ++frame) {
            // Copy to matrix
            std::copy_n(&h->tempV1[mic][frame][0], h->nDoaDtrBins, &h->temp3xnBandsMatrix[0][0]);
            std::copy_n(&h->tempV2[mic][frame][0], h->nDoaDtrBins, &h->temp3xnBandsMatrix[1][0]);
            std::copy_n(&h->tempV3[mic][frame][0], h->nDoaDtrBins, &h->temp3xnBandsMatrix[2][0]);

            // Rotate
            matrixMult(h->tempR, h->temp3xnBandsMatrix, h->temp3xnBandsMatrix2);

            // Copy from matrix
            std::copy_n(&h->temp3xnBandsMatrix2[0][0], h->nDoaDtrBins, &h->tempV1[mic][frame][0]);
            std::copy_n(&h->temp3xnBandsMatrix2[1][0], h->nDoaDtrBins, &h->tempV2[mic][frame][0]);
            std::copy_n(&h->temp3xnBandsMatrix2[2][0], h->nDoaDtrBins, &h->tempV3[mic][frame][0]);

            // In recursive averaging mode, copy the first sub-frame values to rest of the sub-frames
            if (h->useRecursiveAveraging) {
                for (int outFrame = 1; outFrame < h->nFrames; ++outFrame) {
                    std::copy_n(&h->tempV1[mic][0][0], h->nDoaDtrBins, &h->tempV1[mic][outFrame][0]);
                    std::copy_n(&h->tempV2[mic][0][0], h->nDoaDtrBins, &h->tempV2[mic][outFrame][0]);
                    std::copy_n(&h->tempV3[mic][0][0], h->nDoaDtrBins, &h->tempV3[mic][outFrame][0]);
                }
                break;
            }
        }
    }
}


/* Average the metadata vectors as:
 *      v1 = sum(w(mic) * v1(mic))
 *      v2 = sum(w(mic) * v2(mic))
 *      v3 = sum(w(mic) * v3(mic))
 *
 * and the energy as:
 *      interpolated_energy = sum(w(mic) * energy(mic))
 *
 * Input:
 *      energyData              - energy (nMicrophoneArrays x nFrames x nBands)
 *      interpolatedEnergyData  - matrix to store interpolated energy (nFrames x nBands)
 *      residualData            - work on residual metadata (informed source processing)
 */
void average_metadata_vectors_and_energy(hoaTranslatorData *h, bool residualData)
{
    int metadataVecLen = h->nFrames * h->nDoaDtrBins;
    int energyVecLen = h->nFrames * h->nBands;

	float ***energyData = h->energyData;
	float **interpolatedEnergyData = h->interpolatedEnergyData;

	if (residualData) {
		energyData = h->residualEnergyData;
        interpolatedEnergyData = h->interpolatedResidualEnergyData;
	}

    // Reset vectors

    std::fill_n(&h->metadataV1[0][0], metadataVecLen, 0.0f);
    std::fill_n(&h->metadataV2[0][0], metadataVecLen, 0.0f);
    std::fill_n(&h->metadataV3[0][0], metadataVecLen, 0.0f);
    std::fill_n(&interpolatedEnergyData[0][0], energyVecLen, 0.0f);


    float weight = 0.0f;
    for (int mic : h->activeArraysForCurrentFrame) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            weight = h->chosenInterFrameWeights[mic][frame];
            if (h->isCrossfadeInProgress) {
                for (int band = 0; band < h->nBands; ++band) {
                    if (std::find(h->fadeOutBands[frame].begin(), h->fadeOutBands[frame].end(), band) != h->fadeOutBands[frame].end()) {
                        weight = h->fadeOutWeights[mic][frame];
                    } else {
                        weight = h->fadeInWeights[mic][frame];
                    }
                    h->weightedEnergy[mic][frame][band] = weight * energyData[mic][frame][band];
                    if (band < h->maxFreqLimitIndex) {
                        h->tempV1[mic][frame][band] *= weight;
                        h->tempV2[mic][frame][band] *= weight;
                        h->tempV3[mic][frame][band] *= weight;
                    }
                }
            } else {
                // v *= weight
                utility_svsmul(&h->tempV1[mic][frame][0], &weight, h->nDoaDtrBins, &h->tempV1[mic][frame][0]);
                utility_svsmul(&h->tempV2[mic][frame][0], &weight, h->nDoaDtrBins, &h->tempV2[mic][frame][0]);
                utility_svsmul(&h->tempV3[mic][frame][0], &weight, h->nDoaDtrBins, &h->tempV3[mic][frame][0]);

                // weightedEnergy = weight * energyData
                utility_svsmul(&energyData[mic][frame][0], &weight, h->nBands, &h->weightedEnergy[mic][frame][0]);
            }
        }
        // Sum the metadata from active microphone arrays together
        utility_svvadd(&h->metadataV1[0][0], &h->tempV1[mic][0][0], metadataVecLen, &h->metadataV1[0][0]);
        utility_svvadd(&h->metadataV2[0][0], &h->tempV2[mic][0][0], metadataVecLen, &h->metadataV2[0][0]);
        utility_svvadd(&h->metadataV3[0][0], &h->tempV3[mic][0][0], metadataVecLen, &h->metadataV3[0][0]);
        utility_svvadd(&interpolatedEnergyData[0][0], &h->weightedEnergy[mic][0][0], energyVecLen, &interpolatedEnergyData[0][0]);
    }
}

/* Calculate the interpolated metadata as:
 *      azi  = atan2(-v1, -v3)
 *      elev = atan2(v2, sqrt((-v1)^2 + (-v3)^2))
 *      dtr  = sqrt((-v1)^2 + v2^2 + (-v3)^2)
 *
 * Input:
 *      interpolatedAzimuthData    - matrix to store interpolated azimuth   (nFrames x nBands)
 *      interpolatedElevationData  - matrix to store interpolated elevation (nFrames x nBands)
 *      interpolatedDTRData        - matrix to store interpolated DTR       (nFrames x nBands)
 *      residualData               - work on residual metadata (informed source processing)
 */
void calculate_metadata_interpolation(hoaTranslatorData *h, bool residualData)
{
    int vecLen = h->nFrames * h->nDoaDtrBins;
    std::vector<float> v1SquaredVec(vecLen, 0.0f);
    std::vector<float> v2SquaredVec(vecLen, 0.0f);
    std::vector<float> v3SquaredVec(vecLen, 0.0f);
    std::vector<float> v13SquaredSumVec(vecLen, 0.0f);
    std::vector<float> v13sqrtSquaredSumVec(vecLen, 0.0f);
    std::vector<float> v123SquaredSumVec(vecLen, 0.0f);
    float *v1Squared = v1SquaredVec.data();
    float *v2Squared = v2SquaredVec.data();
    float *v3Squared = v3SquaredVec.data();
    float *v13SquaredSum = v13SquaredSumVec.data();
    float *v13sqrtSquaredSum = v13sqrtSquaredSumVec.data();
    float *v123SquaredSum = v123SquaredSumVec.data();

	float **interpolatedAzimuthData = h->interpolatedAzimuthData;
	float **interpolatedElevationData = h->interpolatedElevationData;
	float **interpolatedDTRData = h->interpolatedDTRData;

	if (residualData) {
		interpolatedAzimuthData = h->interpolatedResidualAzimuthData;
		interpolatedElevationData = h->interpolatedResidualElevationData;
		interpolatedDTRData = h->interpolatedResidualDTRData;
	}

    // Inverse of v1 and v3
    float inverseMult = -1.0f;
    utility_svsmul(&h->metadataV1[0][0], &inverseMult, vecLen, &h->metadataV1[0][0]);
    utility_svsmul(&h->metadataV3[0][0], &inverseMult, vecLen, &h->metadataV3[0][0]);

    #if defined(SAF_USE_INTEL_MKL)
    // Power of 2
    vsPowx(vecLen, &h->metadataV1[0][0], 2.0f, v1Squared);
    vsPowx(vecLen, &h->metadataV2[0][0], 2.0f, v2Squared);
    vsPowx(vecLen, &h->metadataV3[0][0], 2.0f, v3Squared);

    // v13SquaredSum = v1^2 + v3^2
    // v123SquaredSum = v1^2 + v2^2 + v3^2
    utility_svvadd(v1Squared, v3Squared, vecLen, v13SquaredSum);
    utility_svvadd(v13SquaredSum, v2Squared, vecLen, v123SquaredSum);

    // v13sqrtSquaredSum = sqrt(v1^2 + v3^2)
    vsHypot(vecLen, &h->metadataV1[0][0], &h->metadataV3[0][0], v13sqrtSquaredSum);

    // Interpolated azimuth, elevation, and DTR
    vsAtan2(vecLen, &h->metadataV1[0][0], &h->metadataV3[0][0], &interpolatedAzimuthData[0][0]);
    vsAtan2(vecLen, &h->metadataV2[0][0], v13sqrtSquaredSum, &interpolatedElevationData[0][0]);
    vsSqrt(vecLen, v123SquaredSum, &interpolatedDTRData[0][0]);

    #elif defined(SAF_USE_APPLE_ACCELERATE)
    // Power of 2
    vDSP_vsq(&h->metadataV1[0][0], 1, v1Squared, 1, vecLen);
    vDSP_vsq(&h->metadataV2[0][0], 1, v2Squared, 1, vecLen);
    vDSP_vsq(&h->metadataV3[0][0], 1, v3Squared, 1, vecLen);

    // v13SquaredSum = v1^2 + v3^2
    // v123SquaredSum = v1^2 + v2^2 + v3^2
    utility_svvadd(v1Squared, v3Squared, vecLen, v13SquaredSum);
    utility_svvadd(v13SquaredSum, v2Squared, vecLen, v123SquaredSum);

    // v13sqrtSquaredSum = sqrt(v1^2 + v3^2)
    vvsqrtf(v13sqrtSquaredSum, v13SquaredSum, &vecLen);

    // Interpolated azimuth, elevation, and DTR
    vvatan2f(&interpolatedAzimuthData[0][0], &h->metadataV1[0][0], &h->metadataV3[0][0], &vecLen);
    vvatan2f(&interpolatedElevationData[0][0], &h->metadataV2[0][0], v13sqrtSquaredSum, &vecLen);
    vvsqrtf(&interpolatedDTRData[0][0], v123SquaredSum, &vecLen);
    #endif
}

/* Metadata interpolation without MKL methods as:
 *      azi  = atan2(-v1, -v3)
 *      elev = atan2(v2, sqrt(v1^2 + v3^2))
 *      dtr  = sqrt(v1^2 + v2^2 + v3^2)
 *
 * Input:
 *      interpolatedAzimuthData    - matrix to store interpolated azimuth   (nFrames x nBands)
 *      interpolatedElevationData  - matrix to store interpolated elevation (nFrames x nBands)
 *      interpolatedDTRData        - matrix to store interpolated DTR       (nFrames x nBands)
 *      residualData               - work on residual metadata (informed source processing)
 */
void calculate_metadata_interpolation_without_mkl(hoaTranslatorData *h, bool residualData)
{

	float **interpolatedAzimuthData = h->interpolatedAzimuthData;
	float **interpolatedElevationData = h->interpolatedElevationData;
	float **interpolatedDTRData = h->interpolatedDTRData;

	if (residualData) {
		interpolatedAzimuthData = h->interpolatedResidualAzimuthData;
		interpolatedElevationData = h->interpolatedResidualElevationData;
		interpolatedDTRData = h->interpolatedResidualDTRData;
	}

    std::vector<float> tempMetadataVector{0.0f, 0.0f, 0.0f};
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int band = 0; band < h->nDoaDtrBins; ++band) {
            tempMetadataVector[0] = h->metadataV1[frame][band];
            tempMetadataVector[1] = h->metadataV2[frame][band];
            tempMetadataVector[2] = h->metadataV3[frame][band];
            interpolatedAzimuthData[frame][band] = calculate_azimuth_OpenGL(tempMetadataVector.data());
            interpolatedElevationData[frame][band] = calculate_elevation_OpenGL(tempMetadataVector.data());
            interpolatedDTRData[frame][band] = calculate_dtr_without_energy(tempMetadataVector.data());
        }
    }
}

/* For high frequencies, use one average value for DOA and DTR. The average value is calculated between frequency bins
 * lowAvgLimitIndex and maxFreqLimitIndex.
 */
void calculate_metadata_average_for_high_frequencies(hoaTranslatorData *h, bool residualData)
{
	float **interpolatedAzimuthData = h->interpolatedAzimuthData;
	float **interpolatedElevationData = h->interpolatedElevationData;
	float **interpolatedDTRData = h->interpolatedDTRData;

	if (residualData) {
		interpolatedAzimuthData = h->interpolatedResidualAzimuthData;
		interpolatedElevationData = h->interpolatedResidualElevationData;
		interpolatedDTRData = h->interpolatedResidualDTRData;
	}

    if (h->maxFreqLimitIndex < h->nBands) {
        int nAvgBands = h->maxFreqLimitIndex - h->lowAvgLimitIndex;
        for (int frame = 0; frame < h->nFrames; ++frame) {
            float avgAzi = 0.0f;
            float avgElev = 0.0f;
            float avgDtr = 0.0f;
            for (int band = h->lowAvgLimitIndex; band < h->maxFreqLimitIndex; ++band) {
                avgAzi += interpolatedAzimuthData[frame][band];
                avgElev += interpolatedElevationData[frame][band];
                avgDtr += interpolatedDTRData[frame][band];
            }
            avgAzi /= (float)nAvgBands;
            avgElev /= (float)nAvgBands;
            avgDtr /= (float)nAvgBands;
            interpolatedAzimuthData[frame][h->maxFreqLimitIndex] = avgAzi;
            interpolatedElevationData[frame][h->maxFreqLimitIndex] = avgElev;
            interpolatedDTRData[frame][h->maxFreqLimitIndex] = avgDtr;
        }
    }
}

void interpolate_metadata(hoaTranslatorData *h, bool interpolateResidual)
{
    // Transform metadata into vector format
    vectorize_metadata(h, interpolateResidual);

    // Rotate the metadata vectors based on head and mic array orientations.
    rotate_metadata_vectors(h);

    // Average active triangle vectors
    average_metadata_vectors_and_energy(h, interpolateResidual);

    // Interpolate azimuth, elevation, and DTR
    #if defined(SAF_USE_INTEL_MKL) || defined(SAF_USE_APPLE_ACCELERATE)
    calculate_metadata_interpolation(h, interpolateResidual);
    #else
    calculate_metadata_interpolation_without_mkl(h, interpolateResidual);
    #endif

    calculate_metadata_average_for_high_frequencies(h, interpolateResidual);

    adjust_metadata_outside_capturing_area(h, interpolateResidual);
}

void interpolate_mic_metadata(hoaTranslatorData *h)
{
    interpolate_metadata(h, false);
}

void interpolate_residual_metadata(hoaTranslatorData *h)
{
    interpolate_metadata(h, true);
}
