#include "OptimalMixing.hpp"

#include <algorithm>
#include <cmath>
#include <complex>

// Eigendecomposition for a 2x2 matrix C = U * D * U'
void eigendecomposition2x2(const hoatMatrixCmplx2D &C,
                           hoatMatrixCmplx2D &U,
                           hoatMatrixCmplx2D &D)
{
    float E1 = C[0][0].real();
    float E2 = C[1][1].real();
    fillWithZeros(U);
    float_complex C10 = C[1][0];
    float_complex C01 = C[0][1];

    /* Eigenvalues l for C:
     * det(C - l * I) = 0
     *  --> l^2 - (C00 + C11)l + (C00 * C11 - C01 * C10) = 0
     *  --> solutions: l = ((C00 + C11) / 2) +- (sqrt((C00 + C11)^2 - 4 * (C00 * C11 - C01 * C10)) / 2)
     *                   = add +- (sqrt(a) / 2)
     *                   = add +- pm
     *  , where
     *      add = (C00 + C11) / 2 = (E1 + E2) / 2
     *      a   = (C00 + C11)^2 - 4 * (C00 * C11 - C01 * C10)
     *          = (E1 + E2)^2 - 4 * (E1 * E2 - C01 * C10)
     *      pm  = sqrt(a) / 2
     */
    float crossSquare = (C10.real() * C10.real()) + (C10.imag() * C10.imag());
    float a = (E1 + E2) * (E1 + E2) - 4.0f * ((E1 * E2) - crossSquare);
    float pm =  0.5f * std::sqrt((std::max)(0.0f, a));
    float add = 0.5f * (E1 + E2);

    // D consists of the eigenvalues of C on the diagonal
    D[0][0] = {add + pm, 0.0f};
    D[1][1] = {(std::max)(0.0f, add - pm), 0.0f};

    // Numeric case, when input is practically zeros OR when input is near an identity matrix with a gain
    if ((D[0][0].real() < 1e-12f) || (pm < (1e-3f * add)))
    {
        U[0][0] = {1.0f, 0.0f};
        U[1][1] = {1.0f, 0.0f};
        return;
    }

    /* Eigenvectors U:
     *
     * Solving U from:
     *  (C - l * I) * U = 0
     * gives eigenvectors:
     *  U00 = C01 / (D00 - E1), if abs(D00 - E1) > abs(D00 - E2)
     *  OR
     *  U00 = (D00 - E2) / C10, if abs(D00 - E2) > abs(D00 - E1)
     *  U10 = 1
     *
     *  U11 = (D11 - E1) / C01, if abs(D11 - E1) > abs(D11 - E2)
     *  OR
     *  U11 = C10 / (D11 - E2), if abs(D11 - E2) > abs(D11 - E1)
     *  U01 = 1
     *
     * The vectors [U00 U10] and [U10 U11] are normalized by their vector length:
     *  1 / eigenVecLen = s * normVal
     */
    for (int ch = 0; ch < 2; ++ch)
    {
        if (std::abs(E2 - D[ch][ch].real()) > std::abs(E1 - D[ch][ch].real()))
        {
            float s = D[ch][ch].real() - E2;
            float normVal = std::sqrt((1.0f / (1e-12f + crossSquare + s*s)));
            U[0][ch] = {s * normVal, 0.0f};
            U[1][ch] = C10 * normVal;
        }
        else
        {
            float s = D[ch][ch].real() - E1;
            float normVal = std::sqrt((1.0f / (1e-12f + crossSquare + s*s)));
            U[1][ch] = {s * normVal, 0.0f};
            U[0][ch] = C01 * normVal;
        }
    }
}

// Cholesky decomposition for a 2x2 matrix C = out * out'
void choleskyDecomposition2x2(const hoatMatrixCmplx2D &C,
                              hoatMatrixCmplx2D &out)
{
    float E1 = C[0][0].real();
    float E2 = C[1][1].real();
    float_complex C10 = C[1][0];
    float_complex C01 = C[0][1];
    float sqrtVal = 0.0f;
    fillWithZeros(out);

    // Perform Cholesky decomposition according to louder channel first
    float crossSquare = (C10.real() * C10.real()) + (C10.imag() * C10.imag());
    if (E1 > E2)
    {
        out[0][0] = {std::sqrt(E1), 0.0f};
        out[1][0] = C10 / (1e-12f + out[0][0].real());
        sqrtVal = E2 - crossSquare / (1e-12f + E1);
        out[1][1] = {std::sqrt((std::max)(0.0f, sqrtVal)), 0.0f};
    }
    else
    {
        out[1][1] = {std::sqrt(E2), 0.0f};
        out[0][1] = C01 / (1e-12f + out[1][1].real());
        sqrtVal = E1 - crossSquare / (1e-12f + E2);
        out[0][0] = {std::sqrt((std::max)(0.0f, sqrtVal)), 0.0f};
    }
}

void calculate_KyGQKx(hoaTranslatorData *h,
                      const hoatMatrix2D &Q)
{
    float Ex1 = h->CxNormalized[0][0].real();
    float Ex2 = h->CxNormalized[1][1].real();
    float Ey1 = h->CyNormalized[0][0].real();
    float Ey2 = h->CyNormalized[1][1].real();

    // Prototype signal energy normalization (eq 7 in [1])
    h->GyHat[0][0] = std::sqrt((Ey1 / (1e-12f + (std::max)(Ex1, 0.001f * Ex2))));
    h->GyHat[1][0] = std::sqrt((Ey2 / (1e-12f + (std::max)(Ex2, 0.001f * Ex1))));

    // GyHatQ = GyHat * Q
    for (int ch = 0; ch < 2; ++ch)
    {
        h->GyHatQ[ch][0] = {Q[ch][0] * h->GyHat[ch][0], 0.0f};
        h->GyHatQ[ch][1] = {Q[ch][1] * h->GyHat[ch][0], 0.0f};
    }

    // temp = Ky' * GyHat * Q
    matrixMultConjNoTrans(h->Ky, h->GyHatQ, h->temp2x2Cmplx);

    // KyGQKx = Ky' * GyHat * Q * Kx
    matrixMult(h->temp2x2Cmplx, h->Kx, h->KyGQKx);
}

/* Find nearest orthonormal matrix P to KyGQKx = Ky' * G_hat * Q * Kx
 * For matrix A that is P = KyGQKx * (KyGQKx' * KyGQKx)^(-0.5)
 * http://people.csail.mit.edu/bkph/articles/Nearest_Orthonormal_Matrix.pdf
 * The following steps follow the procedure shown in the above document
 */
void formulate_optimal_P(hoaTranslatorData *h)
{
    // KyGQKx' * KyGQKx = Utemp * Dtemp * Utemp'
    matrixMultConjNoTrans(h->KyGQKx, h->KyGQKx, h->temp2x2Cmplx);
    eigendecomposition2x2(h->temp2x2Cmplx, h->Utemp, h->Dtemp);

    /* In the above document, it is stated that (KyGQKx' * KyGQKx)^(-0.5) has the same eigenvectors as (KyGQKx' * KyGQKx),
     * e.g. Utemp remains the same. The eigenvalues of (KyGQKx' * KyGQKx)^(-0.5) are Dtemp^(-0.5), i.e., the inverse of
     * the square roots the eigenvalues of (KyGQKx' * KyGQKx).
     *
     *  DsqrtInv = sqrt(1 / Dtemp) = Dtemp^(-0.5)
     */
    fillWithZeros(h->DsqrtInv);
    h->DsqrtInv[0][0] = {(std::min)(10000.0f, std::sqrt((1.0f / (1e-12f + h->Dtemp[0][0].real())))), 0.0f};
    h->DsqrtInv[1][1] = {(std::min)(10000.0f, std::sqrt((1.0f / (1e-12f + h->Dtemp[1][1].real())))), 0.0f};

    // temp2 = KyGQKx * Utemp * sqrt(1 / Dtemp)
    matrixMult(h->KyGQKx, h->Utemp, h->temp2x2Cmplx);
    matrixMult(h->temp2x2Cmplx, h->DsqrtInv, h->temp2x2Cmplx2);

    // P = KyGQKx * Utemp * sqrt(1 / Dtemp) * Utemp'
    matrixMultNoTransConj(h->temp2x2Cmplx2, h->Utemp, h->P);
}

/* Mixing matrix M = Ky * P * Kx^(-1) (eq 11 and 13 in [1]):
 *  Kx = Ux * sqrt(Sx)
 *  Ux^(-1) = Ux', because Ux is orthogonal
 *  Kx^(-1) = sqrt(Sx)^(-1) * Ux' = (1 / sqrt(Sx)) * Ux'
 *  M = Ky * P * (1 / sqrt(Sx)) * Ux'
 */
void formulate_mixing_matrix(hoaTranslatorData *h,
                             float regularizationCoeff,
                             hoatMatrixCmplx2D &M)
{

    // Regularize the diagonal Sx for matrix inversion (eq 14 in [1]),
    h->SxSqrtReg[0][0] = (std::max)(h->SxSqrt[0][0].real(), regularizationCoeff * h->SxSqrt[1][1].real());
    h->SxSqrtReg[1][1] = (std::max)(h->SxSqrt[1][1].real(), regularizationCoeff * h->SxSqrt[0][0].real());

    // P * (1 / SxSqrtReg)
    for (int ch1 = 0; ch1 < 2; ++ch1) {
        for (int ch2 = 0; ch2 < 2; ++ch2) {
            h->P[ch1][ch2] /= (h->SxSqrtReg[ch2][ch2] + 1e-12f);
        }
    }
    // M = Ky * P * (1 / SxSqrtReg) * Ux'
    matrixMult(h->Ky, h->P, h->temp2x2Cmplx);
    matrixMultNoTransConj(h->temp2x2Cmplx, h->Ux, M);
}

void form_residual_matrix_or_energy_compensate(hoaTranslatorData *h,
                                               const hoatMatrixCmplx2D &Cx,
                                               const hoatMatrixCmplx2D &Cy,
                                               bool useEnergyFlag,
                                               hoatMatrixCmplx2D &M,
                                               hoatMatrixCmplx2D &Cr,
                                               bool calculateCr)
{
    if (!useEnergyFlag && !calculateCr) {
        return;
    }

    // CyTilde = M * Cx * M';
    matrixMult(M, Cx, h->temp2x2Cmplx);
    matrixMultNoTransConj(h->temp2x2Cmplx, M, h->CyTilde);

    // Formulate residual covariance matrix Cr = Cy - CyTilde
    if (calculateCr){
        matrixMatrixSub(Cy, h->CyTilde, Cr);
        Cr[0][0] = {(std::max)(0.0f, Cr[0][0].real()), 0.0f};
        Cr[1][1] = {(std::max)(0.0f, Cr[1][1].real()), 0.0f};
    }

    /* Use energy compensation instead of residuals:
     *  M = G * M
     *  G = diag(sqrt(Cy / CyTilde))
     */
    if (useEnergyFlag) {
        fillWithZeros(h->Genergy);
        for (int ch = 0; ch < 2; ++ch) {
            h->Genergy[ch][ch] = std::sqrt(Cy[ch][ch] / (h->CyTilde[ch][ch] + 1e-20f));
        }
        matrixMult(h->Genergy, M, h->temp2x2Cmplx);
        copy(h->temp2x2Cmplx, M);
    }
}

void formulate_M_and_Cr_2x2(hoaTranslatorData *h,
                            const hoatMatrixCmplx2D &Cx,
                            const hoatMatrixCmplx2D &Cy,
                            const hoatMatrix2D &Q,
                            bool useEnergyFlag,
                            float regularizationCoeff,
                            hoatMatrixCmplx2D &M,
                            hoatMatrixCmplx2D &Cr,
                            bool calculateCr)
{
    // Normalize covariance matrices
    float maxEnergy = (std::max)({Cx[0][0].real(), Cx[1][1].real(), Cy[0][0].real(), Cy[1][1].real()});
    float mult = 1.0f / (maxEnergy + 1e-12f);
    matrixScalarMult(mult, Cx, h->CxNormalized);
    matrixScalarMult(mult, Cy, h->CyNormalized);

    // Cholesky decomposition of target / output covariance matrix Cy = Ky * Ky'
    choleskyDecomposition2x2(h->CyNormalized, h->Ky);

    // Eigendecomposition of input covariance matrix Cx = Ux * Sx * Ux'
    eigendecomposition2x2(h->CxNormalized, h->Ux, h->Sx);

    /* Eigendecomposition to Kx:
     * Cx = Ux * Sx * Ux'
     *    = Ux * sqrt(Sx) * sqrt(Sx)' * Ux'
     *    = Ux * sqrt(Sx) * (Ux * sqrt(Sx))'
     *    = Kx * Kx'
     *
     * --> Kx = Ux * sqrt(Sx)
     */
    h->SxSqrt[0][0] = std::sqrt(h->Sx[0][0]);
    h->SxSqrt[1][1] = std::sqrt(h->Sx[1][1]);
    matrixMult(h->Ux, h->SxSqrt, h->Kx);

    // KyGQKx = Ky' * G_hat * Q * Kx
    calculate_KyGQKx(h, Q);

    // P = KyGQKx * (KyGQKx' * KyGQKx)^(-0.5)
    formulate_optimal_P(h);

    // M = Ky * P * Kx^(-1)
    formulate_mixing_matrix(h, regularizationCoeff, M);

    // Cr = Cy - M * Cx * M'
    // MenergyCompensated = diag(sqrt(Cy / (M * Cx * M'))) * M
    form_residual_matrix_or_energy_compensate(h, Cx, Cy, useEnergyFlag, M, Cr, calculateCr);
}
