/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef NOMINMAX
#define NOMINMAX
#endif

#include "PositionPreProcessor.hpp"

#include <algorithm>
#include <cmath>
#include <tuple>
#include <vector>

#include "Exterior.hpp"
#include <delaunator.hpp>

/* Calculate the interpolation weights for the current listener position. The weights can be formed in three different ways:
 *
 *      - the listener is outside the triangles, and projected to an edge --> use weights for the two edge arrays, based on the projected position
 *      - the listener is outside the triangles, and projected to an array --> set the projected array weight to 1.0
 *      - the listener is inside a triangle --> use the barycentric coordinates as the weights
 *
 * Finally, the weights for the current processing frame are linearly interpolated with the previous frame weights.
 *
 * Output:
 *      currentInterpolationWeights    - the interpolation weights for each mic array for the current frame
 *      currentInterFrameWeights       - linearly interpolated weights from current and previous frame
 */
void calculate_interpolation_weights(hoaTranslatorData *h)
{
    // Reset the weights
    std::fill_n(&h->currentInterpolationWeights[0], h->nMicrophoneArrays, 0.0f);

    // Collect the new weights
    if (h->isListenerProjectedToClosestEdge) {
        h->currentInterpolationWeights[h->edgeArrays[0]] = h->edgeWeights[0];
        h->currentInterpolationWeights[h->edgeArrays[1]] = h->edgeWeights[1];
    } else if (h->isListenerProjectedToClosestArray) {
        h->currentInterpolationWeights[h->closestArrayByDistance] = 1.0f;
    } else {
        auto const& activeTriangle = h->triangleFaces[h->activeTriangleIndex];
        for (int i = 0; i < 3; ++i) {
            h->currentInterpolationWeights[activeTriangle[i]] = (std::max)(0.0f, h->barycentricCoordinates[i]);
        }
    }

    // Linear interpolation between the weights of this and previous frame
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->currentInterFrameWeights[mic][frame] = h->interpolator[frame] * h->currentInterpolationWeights[mic] +
                                                      (1.0f - h->interpolator[frame]) * h->previousInterpolationWeights[mic];
        }
    }
}

void pre_process_positions(hoaTranslatorData *h)
{
    /* Get a matrix of array positions in a triangle as (xy -horizontal plane):
     *
     *      x1  x2  x3
     *      y1  y2  y3
     *      1   1   1
     */
    auto get_triangle_position_matrix = [&](std::vector<std::vector<int>> triangleFaces, int triangleIndex, float trianglePosM[3 * 3]) {
        for (int i = 0; i < 3; ++i) {
            auto arrayPos = h->microphoneArrayPositions[triangleFaces[triangleIndex][i]];
            trianglePosM[i + (0 * 3)] = arrayPos[0];
            trianglePosM[i + (1 * 3)] = -arrayPos[2];
            trianglePosM[i + (2 * 3)] = 1.0f;
        }
    };

    auto is_positive = [](const float* vector, int size) {
        bool isPositive = true;
        for (int i = 0; i < size; ++i) {
            if (vector[i] < 0.0f) {
                isPositive = false;
                break;
            }
        }
        return isPositive;
    };

    // Reset variables
    h->isListenerInsideTriangle = false;
    h->isListenerInsideOmittedTriangle = false;
    h->activeTriangleIndex = 0;
    float trianglePosM[3 * 3]{};
    float listenerPos[3]{h->currentListenerPosition[0], -h->currentListenerPosition[2], 1.0f};
    reset_exterior_variables(h);

    // Compute barycentric coordinates for a triangle by solving a linear system:
    // A * X = B, or i.e.:
    // trianglePosM * barycentricCoordinates = listenerPosition
    // https://en.wikipedia.org/wiki/Barycentric_coordinate_system#Conversion_between_barycentric_and_Cartesian_coordinates
    if (!h->useMphoa2SourceMode && !h->useMphoa1SourceMode) {
        for (int tri = 0; tri < h->nTriangles; ++tri) {
            get_triangle_position_matrix(h->triangleFaces, tri, trianglePosM);
            utility_sglslv(trianglePosM, 3, listenerPos, 1, h->barycentricCoordinates);

            // The listener is inside a triangle, if all the barycentric coordinates are positive for the triangle
            if (is_positive(h->barycentricCoordinates, 3)) {
                h->isListenerInsideTriangle = true;
                h->activeTriangleIndex = tri;
                break;
            }
        }
    }

    // If the listener is not inside any triangle, find the closest point in all of the triangle arrays
    if (!h->isListenerInsideTriangle) {
        if (h->useImprovedExteriorRendering) {
            // Check if the listener is inside omitted triangles
            for (int tri = 0; tri < h->omittedTriangleFaces.size(); ++tri) {
                get_triangle_position_matrix(h->omittedTriangleFaces, tri, trianglePosM);
                utility_sglslv(trianglePosM, 3, listenerPos, 1, h->barycentricCoordinates);
                if (is_positive(h->barycentricCoordinates, 3)) {
                    h->isListenerInsideOmittedTriangle = true;
                    break;
                }
            }
        }
        project_to_closest_edge_or_array(h);
    }

    // Calculate the current interpolation weights, and linearly interpolate weights between current and previous frame
    calculate_interpolation_weights(h);
}

void delaunay_triangulation(hoaTranslatorData *h)
{
    auto collect_triangle_edges_and_faces = [&](hoaTranslatorData *h, delaunator::Delaunator &d) {
        h->edgesBetweenArrays.clear();
        h->omittedEdgesBetweenArrays.clear();
        h->nonConvexMicArrays.clear();
        h->triangleFaces.clear();
        h->omittedTriangleFaces.clear();
        for (int n = 0; n < d.triangles.size(); n += 3) {
            std::vector<int> triangleArrays{(int)d.triangles[n],
                                            (int)d.triangles[n + 1],
                                            (int)d.triangles[n + 2]};
            std::sort(triangleArrays.begin(), triangleArrays.end());
            std::vector<std::vector<int>> edgeList{{triangleArrays[0], triangleArrays[1]},
                                                   {triangleArrays[0], triangleArrays[2]},
                                                   {triangleArrays[1], triangleArrays[2]}};
            std::vector<int> micsNotPartOfEdge{triangleArrays[2], triangleArrays[1], triangleArrays[0]};
            bool addTriangle = true;
            for (int i = 0; i < edgeList.size(); ++i) {
                float x1 = h->microphoneArrayPositions[edgeList[i][0]][0];
                float y1 = -h->microphoneArrayPositions[edgeList[i][0]][2];
                float x2 = h->microphoneArrayPositions[edgeList[i][1]][0];
                float y2 = -h->microphoneArrayPositions[edgeList[i][1]][2];
                float edgeLength = std::sqrt((std::pow(x2 - x1, 2.0f)) + (std::pow(y2 - y1, 2.0f)));
                if (h->omitLongTriangleEdges && edgeLength > h->maxTriangleEdgeLength) {
                    addTriangle = false;
                    if (std::find(h->omittedEdgesBetweenArrays.begin(), h->omittedEdgesBetweenArrays.end(), edgeList[i]) == h->omittedEdgesBetweenArrays.end()) {
                        // Add new omitted edge
                        h->omittedEdgesBetweenArrays.push_back(edgeList[i]);
                        h->nonConvexMicArrays.push_back(micsNotPartOfEdge[i]);
                    }
                } else if (std::find(h->edgesBetweenArrays.begin(), h->edgesBetweenArrays.end(), edgeList[i]) == h->edgesBetweenArrays.end()) {
                    // Add new edge
                    h->edgesBetweenArrays.push_back(edgeList[i]);
                }
            }
            if (addTriangle) {
                h->triangleFaces.push_back(triangleArrays);
            } else {
                h->omittedTriangleFaces.push_back(triangleArrays);
            }
        }
        h->nTriangles = h->triangleFaces.size();
    };

    // Pick horizontal coordinates for the Delaunay triangulation
    std::vector<double> horizontalCoords(2 * h->nMicrophoneArrays);
    int coordIndex = 0;
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        horizontalCoords[coordIndex + 0] = h->microphoneArrayPositions[mic][0];
        horizontalCoords[coordIndex + 1] = -h->microphoneArrayPositions[mic][2];
        coordIndex += 2;
    }

    // Delaunay triangulation
    delaunator::Delaunator d(horizontalCoords);

    // Try to omit long triangles first
    h->omitLongTriangleEdges = true;
    h->maxTriangleEdgeLength = 4.0f;
    collect_triangle_edges_and_faces(h, d);

    // If no triangles are formed, include also long triangles
    if (h->nTriangles == 0) {
        h->omitLongTriangleEdges = false;
        collect_triangle_edges_and_faces(h, d);
    }
}

void triangulation_for_one_source(hoaTranslatorData *h)
{
    h->nTriangles = 1;
    h->triangleFaces.push_back({0, 0, 0});
}

void triangulation_for_two_sources(hoaTranslatorData *h)
{
    h->nTriangles = 1;
    h->edgesBetweenArrays.push_back({0, 1});
    h->edgesBetweenArrays.push_back({1, 0});
    h->triangleFaces.push_back({0, 1, 1});
}
