/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_POSITIONPREPROCESSOR_HPP
#define HOA_TRANSLATION_POSITIONPREPROCESSOR_HPP

#include "Utilities.hpp"

/* Pre process the current listener position.
 *
 * Try to find a triangle, in which the listener is currently located. The search is based on the barycentric coordinates
 * of the three mic arrays forming a triangle. If all three barycentric coordinates for a triangle are positive, the
 * listener is inside a triangle. In that case, the interpolation weights are formed directly from the barycentric
 * coordinates.
 *
 * If the listener is not inside any triangle, the listener position is projected to the border of the capturing area.
 * The projected position can lie on a triangle edge, on an edge between two mic arrays (that are not part of the same
 * triangle), or the position can be projected to the closest mic array.
 *
 * https://en.wikipedia.org/wiki/Barycentric_coordinate_system
 *
 * Output:
 *      activeTriangleIndex            - currently active triangle index
 *      currentInterpolationWeights    - the interpolation weights for each mic array for the current frame
 *      currentInterFrameWeights       - linearly interpolated weights from current and previous frame
 *
 * When the listener is outside the capturing area, output also:
 *      edgeArrays                     - the indexes of the two mic arrays, that connect the edge, where the listener is projected to
 *      distanceToCapturingArea        - distance from the listener position to the projected position
 *      edgeTriangleArrays             - all the mic array indexes that are part of the triangle(s) of edgeArrays
 *                                       (3 if edgeArrays are in the same triangle, 6 if edgeArrays are part of different triangles)
 */
void pre_process_positions(hoaTranslatorData *h);

/* Form triangles from microphone positions based on Delaunay triangulation.
 *
 * Output:
 *      edgesBetweenArrays      - mic array indexes forming edges between arrays
 *      triangleFaces           - triangle faces (i.e. mic array indexes forming the triangles)
 *      omittedTriangleFaces    - triangle faces that were omitted due to some edge length exceeding maxTriangleEdgeLength
 *      nTriangles              - number of triangles formed by Delaunay triangulation
 */
void delaunay_triangulation(hoaTranslatorData *h);

void triangulation_for_one_source(hoaTranslatorData *h);

void triangulation_for_two_sources(hoaTranslatorData *h);

#endif //HOA_TRANSLATION_POSITIONPREPROCESSOR_HPP
