/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Rotator.hpp"

#include <cmath>

// Rotation around z-axis
void get_z_rotation_matrix(float R[3][3],
                           float rollDeg)
{
    /* cos(deg)  -sin(deg)     0
     * sin(deg)   cos(deg)     0
     *    0          0         1
     */
    float rollRad = rollDeg * M_PI / 180.0f;
    R[0][0] = std::cos(rollRad);
    R[0][1] = -std::sin(rollRad);
    R[1][0] = std::sin(rollRad);
    R[1][1] = std::cos(rollRad);
    R[2][2] = 1.0f;
}

// Rotation around y-axis
void get_y_rotation_matrix(float R[3][3],
                           float yawDeg)
{
    /* cos(deg)  0   sin(deg)
     *     0       1       0
     * -sin(deg) 0   cos(deg)
     */
    float yawRad = yawDeg * M_PI / 180.0f;
    R[0][0] = std::cos(yawRad);
    R[0][2] = std::sin(yawRad);
    R[1][1] = 1.0f;
    R[2][0] = -std::sin(yawRad);
    R[2][2] = std::cos(yawRad);
}

// Rotation around x-axis
void get_x_rotation_matrix(float R[3][3],
                           float pitchDeg)
{
    /* 1        0           0
     * 0     cos(deg)  -sin(deg)
     * 0     sin(deg)   cos(deg)
     */
    float pitchRad = pitchDeg * M_PI / 180.0f;
    R[0][0] = 1.0f;
    R[1][1] = std::cos(pitchRad);
    R[1][2] = -std::sin(pitchRad);
    R[2][1] = std::sin(pitchRad);
    R[2][2] = std::cos(pitchRad);
}

void get_rotation_matrix_OpenGL(float yawDeg,
                                float pitchDeg,
                                float rollDeg,
                                float R[3][3])
{
    float R_yaw[3][3]{};
    float R_pitch[3][3]{};
    float R_roll[3][3]{};
    float tempR[3][3]{};

    get_y_rotation_matrix(R_yaw, yawDeg);
    get_x_rotation_matrix(R_pitch, pitchDeg);
    get_z_rotation_matrix(R_roll, rollDeg);

    // R = R_roll * R_pitch * R_yaw
    matrixMult3x3(*R_pitch, *R_yaw, *tempR);
    matrixMult3x3(*R_roll, *tempR, *R);
}

void get_rotation_matrix_classic(float yawDeg,
                                 float pitchDeg,
                                 float rollDeg,
                                 float R[3][3])
{
    float R_yaw[3][3]{};
    float R_pitch[3][3]{};
    float R_roll[3][3]{};
    float tempR[3][3]{};

    get_z_rotation_matrix(R_yaw, yawDeg);
    get_y_rotation_matrix(R_pitch, pitchDeg);
    get_x_rotation_matrix(R_roll, rollDeg);

    // R = R_roll * R_pitch * R_yaw
    matrixMult3x3(*R_pitch, *R_yaw, *tempR);
    matrixMult3x3(*R_roll, *tempR, *R);
}

void calculate_sh_rotation_matrix(hoaTranslatorData *h,
                                  int micArrayIndx,
                                  hoatMatrix2D &SHrotationMatrix)
{
    // Inverse head rotation, double flip pitch and roll, because positive rotation directions differ between OpenGL and SAF.
    float headYawDeg   = -h->yawDeg;
    float headPitchDeg = h->pitchDeg;
    float headRollDeg  = h->rollDeg;

    // Mic rotations
    float micYawDeg   = h->micArrayOrientations[micArrayIndx][0];
    float micPitchDeg = h->micArrayOrientations[micArrayIndx][1];
    float micRollDeg  = h->micArrayOrientations[micArrayIndx][2];

    // Get rotation matrix R = Rhead * Rmic
    float Rhead[3][3]{};
    float Rmic[3][3]{};
    float R[3][3]{};
    get_rotation_matrix_classic(headYawDeg,
                                headPitchDeg,
                                headRollDeg,
                                Rhead);

    get_rotation_matrix_classic(micYawDeg,
                                micPitchDeg,
                                micRollDeg,
                                Rmic);

    matrixMult3x3(*Rhead, *Rmic, *R);

    // Get spherical harmonics rotation matrix
    getSHrotMtxReal(R,
                    h->tempSHrotationMatrix,
                    h->shOrder);

    // Copy to complex currentSHrotationMatrix
    for (int row = 0; row < h->channelsPerArray; ++row) {
        for (int col = 0; col < h->channelsPerArray; ++col) {
            SHrotationMatrix[row][col] = h->tempSHrotationMatrix[row * h->channelsPerArray + col];
        }
    }
}

void get_sh_rotation_matrix(hoaTranslatorData *h)
{
    calculate_sh_rotation_matrix(h, h->currentClosestMicrophoneArray, h->currentSHrotationMatrix);
    if (h->isCrossfadeInProgress) {
        calculate_sh_rotation_matrix(h, h->previousClosestMicrophoneArray, h->previousSHrotationMatrix);
    }
}
