/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SignalInterpolator.hpp"

#include <algorithm>
#include <cmath>

#ifdef SAF_USE_INTEL_MKL
#include "mkl.h"
#elif SAF_USE_APPLE_ACCELERATE
#include <Accelerate/Accelerate.h>
#endif

// Add min values to energyData to avoid division by zero
void add_min_value_to_energy(hoaTranslatorData *h)
{
    const float minAddValue = FLT_MIN;
    utility_svsadd(&h->energyData[h->currentClosestMicrophoneArray][0][0], &minAddValue, h->nFrames * h->nBands, &h->energyData[h->currentClosestMicrophoneArray][0][0]);
    utility_svsadd(&h->energyData[h->previousClosestMicrophoneArray][0][0], &minAddValue, h->nFrames * h->nBands, &h->energyData[h->previousClosestMicrophoneArray][0][0]);
}

// Copy eqGains to float_complex array, so that we can use utility_cvvmul()
void transform_eq_gains_to_cmplx(hoaTranslatorData *h)
{
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int ch = 0; ch < h->channelsPerArray; ++ch) {
            for (int band = 0; band < h->nBands; ++band) {
                h->eqGainsCmplx[frame][ch][band] = float_complex(h->eqGains[frame][band], 0.0f);
            }
        }
    }
}

// eqGains = min( sqrt( interpolatedEnergyData / E_closest_mic_array ), interpolationEqGainMax )
// NOTE: interpolatedEnergyData = (w1 * E1) + (w2 * E2) + (w3 * E3)
void compute_eq_gains(hoaTranslatorData *h)
{
    const int nEqGains = h->nFrames * h->nBands;
    #if defined(SAF_USE_INTEL_MKL)
    vsDiv(nEqGains, h->interpolatedEnergyData[0], h->energyData[h->currentClosestMicrophoneArray][0], &h->eqGains[0][0]);
    vsSqrt(nEqGains, &h->eqGains[0][0], &h->eqGains[0][0]);
    vsFmin(nEqGains, &h->eqGains[0][0], h->eqGainsMax.data(), &h->eqGains[0][0]);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vDSP_vdiv(&h->energyData[h->currentClosestMicrophoneArray][0][0], 1, &h->interpolatedEnergyData[0][0], 1, &h->eqGains[0][0], 1, nEqGains);
    vvsqrtf(&h->eqGains[0][0], &h->eqGains[0][0], &nEqGains);
    vDSP_vmin(&h->eqGains[0][0], 1, h->eqGainsMax.data(), 1, &h->eqGains[0][0], 1, nEqGains);
    #endif
}

// interpolated_signal = eqGains * closest_mic_array_signal
void calculate_signal_interpolation(hoaTranslatorData *h) {
    utility_cvvmul(&h->eqGainsCmplx[0][0][0], &h->TFDInputFrameFullOrder[h->currentClosestMicrophoneArray][0][0][0], h->nFrames * h->channelsPerArray * h->nBands, &h->interpolatedTFDFrame[0][0][0]);
}

// Signal interpolation without MKL methods. Used during crossfade, and when MKL is not available
void calculate_signal_interpolation_without_mkl(hoaTranslatorData *h)
{
    float eqGain = 0;
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int ch = 0; ch < h->channelsPerArray; ++ch) {
            for (int band = 0; band < h->nBands; ++band) {
                int closestMicArray = h->currentClosestMicrophoneArray;
                if (h->isCrossfadeInProgress) {
                    if (std::find(h->fadeOutBands[frame].begin(), h->fadeOutBands[frame].end(), band) != h->fadeOutBands[frame].end()) {
                        closestMicArray = h->previousClosestMicrophoneArray;
                    }
                }
                eqGain = std::sqrt(h->interpolatedEnergyData[frame][band] /
                                   (h->energyData[closestMicArray][frame][band] + FLT_MIN));
                eqGain = (std::min)(h->interpolationEqGainMax, eqGain);
                h->interpolatedTFDFrame[frame][ch][band] = eqGain * h->TFDInputFrameFullOrder[closestMicArray][frame][ch][band];
            }
        }
    }
}

void interpolate_signals(hoaTranslatorData *h)
{
    if (h->isCrossfadeInProgress) {
        calculate_signal_interpolation_without_mkl(h);
        return;
    }
    #if defined(SAF_USE_INTEL_MKL) || defined(SAF_USE_APPLE_ACCELERATE)
    add_min_value_to_energy(h);
    compute_eq_gains(h);
    transform_eq_gains_to_cmplx(h);
    calculate_signal_interpolation(h);
    #else
    calculate_signal_interpolation_without_mkl(h);
    #endif
}

void interpolate_three_closest_arrays(hoaTranslatorData *h)
{
    // Clear interpolatedTFDFrame
    std::fill_n(&h->interpolatedTFDFrame[0][0][0], h->nFrames * h->channelsPerArray * h->nBands, float_complex(0.0f, 0.0f));

    // interpolated_signal = (w1 * sig1) + (w2 * sig2) + (w3 * sig3)
    float_complex weight{0.0f, 0.0f};
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            if (h->chosenInterFrameWeights[mic][frame] > 0.0f) {
                weight = float_complex(h->chosenInterFrameWeights[mic][frame], 0.0f);

                // tempSignal = interpolation_weight * input_signal
                utility_cvsmul(&h->TFDInputFrameFOA[mic][frame][0][0],
                               &weight,
                               h->channelsPerArray * h->nBands,
                               &h->tempSignal[0][0]);

                // interpolated_signal += tempSignal
                utility_cvvadd(&h->interpolatedTFDFrame[frame][0][0], &h->tempSignal[0][0], h->channelsPerArray * h->nBands, &h->interpolatedTFDFrame[frame][0][0]);
            }
        }
    }
}
