/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SourceAware.hpp"

#include <algorithm>
#include <cfloat>
#include <cmath>
#include <complex>
#include <numeric>
#include <vector>

#include "Exterior.hpp"
#include "MetadataInterpolator.hpp"
#include "Rotator.hpp"
#include "SpatialAnalysis.hpp"
#include "VectorOperations.hpp"


void calculate_mean_listener_position(hoaTranslatorData *h)
{
    for (int i = 0; i < 3; ++i) {
        h->meanListenerPosition[i] = (h->currentListenerPosition[i] + h->previousListenerPosition[i]) / 2.0f;
    }
}

// Calculate the vectors and distances from the mean listener position to each source
void calculate_listener_to_source_vectors(hoaTranslatorData *h)
{
    // Get head rotation matrix R
    float sourceYawDeg   = -h->yawDeg;
    float sourcePitchDeg = -h->pitchDeg;
    float sourceRollDeg  = -h->rollDeg;
    float R[3][3]{};
    get_rotation_matrix_OpenGL(sourceYawDeg,
                               sourcePitchDeg,
                               sourceRollDeg,
                               R);
    std::copy_n(&R[0][0], 3 * 3, &h->tempR[0][0]);

    // Calculate listener to source vectors and distances
    for (int src = 0; src < h->nSources; ++src) {

        // Form a vector from listener to source
        for (int i = 0; i < 3; ++i) {
            h->listenerToSrcVectorsOpenGL[src][i][0] = h->sourcePositionsOpenGL[src][i] - h->meanListenerPosition[i];
        }

        // Apply head rotation
        matrixMult(h->tempR, h->listenerToSrcVectorsOpenGL[src], h->temp3x1Matrix);
        std::copy_n(&h->temp3x1Matrix[0][0], 3, &h->listenerToSrcVectorsOpenGL[src][0][0]);

        // Calculate the distance
        h->listenerToSrcDistances[src] = calculate_vector_norm_for_first_n_dimensions(&h->listenerToSrcVectorsOpenGL[src][0][0], 3);
    }
}

void calculate_distances_to_outside_sources(hoaTranslatorData *h)
{
    auto update_outside_sources_and_closest_distance = [&](int src, bool &isOutsideSource, float edgeWeight, float src2edgeDistance, float &closestDistance) {
        // The source is outside the capturing area
        if (std::find(h->outsideSources.begin(), h->outsideSources.end(), src) == h->outsideSources.end()) {
            h->outsideSources.push_back(src);
            ++h->nOutsideSources;
            isOutsideSource = true;
        }
        // Update the closest distance to the capturing area
        if (edgeWeight >= 0.0f && edgeWeight <= 1.0f) {
            if (src2edgeDistance < closestDistance) {
                closestDistance = src2edgeDistance;
            }
        }
    };

    for (int src = 0; src < h->nSources; ++src) {
        bool isOutsideSource = false;
        float closestDistance = FLT_MAX;

        // Check if the source is outside the original capturing area
        for (int edgeIndx = 0; edgeIndx < h->originalExteriorEdges.size(); ++edgeIndx) {
            auto edgeNormalWeights = project_based_on_normal(h, h->originalExteriorEdges[edgeIndx], h->originalEdgeNormals[edgeIndx], h->sourcePositionsOpenGL[src]);
            auto edgeWeight = edgeNormalWeights[0];
            float src2edgeDistance = edgeNormalWeights[1];
            if (src2edgeDistance > 0.0f) {
                update_outside_sources_and_closest_distance(src, isOutsideSource, edgeWeight, src2edgeDistance, closestDistance);
            }
        }

        // Check if the source is in some omitted triangle
        for (int edgeIndx = 0; edgeIndx < h->nExteriorEdges; ++edgeIndx) {
            auto edgeNormalWeights = project_based_on_normal(h, h->exteriorEdges[edgeIndx], h->edgeNormals[edgeIndx], h->sourcePositionsOpenGL[src]);
            auto edgeWeight = edgeNormalWeights[0];
            float src2edgeDistance = edgeNormalWeights[1];
            if (src2edgeDistance > 0.0f && (edgeWeight >= 0.0f && edgeWeight <= 1.0f)) {
                update_outside_sources_and_closest_distance(src, isOutsideSource, edgeWeight, src2edgeDistance, closestDistance);
            }
        }

        if (h->useMphoa1SourceMode) {
            // all informed sources are outside sources in 1 source mode
            if (std::find(h->outsideSources.begin(), h->outsideSources.end(), src) == h->outsideSources.end()) {
                h->outsideSources.push_back(src);
                ++h->nOutsideSources;
                isOutsideSource = true;
            }
        }

        // Check if the closest mic array is closer than the closest edge (in the horizontal plane)
        if (isOutsideSource) {
            int closestMic = h->closestMicsForSources[src];
            auto srcPosHorizontal = {h->sourcePositionsOpenGL[src][0], h->sourcePositionsOpenGL[src][2]};
            auto micPosHorizontal = {h->microphoneArrayPositions[closestMic][0], h->microphoneArrayPositions[closestMic][2]};
            auto micToSrcVec = subtractVectors(srcPosHorizontal, micPosHorizontal);
            float micToSrcDistance = calculate_vector_norm(micToSrcVec);
            if (micToSrcDistance < closestDistance) {
                closestDistance = micToSrcDistance;
            }
            h->outsideSrcDistances.push_back(closestDistance);
        }
    }

    // Calculate the average distance to outside sources
    if (h->nOutsideSources > 0) {
        float sumDistance = 0.0f;
        for (const auto& dist : h->outsideSrcDistances) {
            sumDistance += dist;
        }
        h->averageOutsideSrcDistance = sumDistance / (float)h->nOutsideSources;
    }
}

/* Calculate DOA, distance, delay and attenuation gains for each mic array and source pair. Also calculate delays and
 * attenuation gains between arrays for each source.
 */
void calculate_source_geometry_parameters(hoaTranslatorData *h)
{
    hoatMatrix2D micToSrcVec;
    hoatMatrix2D micToSrcVecRotated;
    createHoatMatrix(&micToSrcVec, 1, 3);
    createHoatMatrix(&micToSrcVecRotated, 1, 3);
    float Rmic[3][3]{};
    for (int src = 0; src < h->nSources; ++src) {
        float minDistance = FLT_MAX;
        float minDistance2 = FLT_MAX;
        int closestMic = 0;
        int secondClosestMic = 0;
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {

            // Form a vector from mic array to source
            for (int i = 0; i < 3; ++i) {
                micToSrcVec[0][i] = h->sourcePositionsOpenGL[src][i] - h->microphoneArrayPositions[mic][i];
            }

            // Apply mic array rotation to the micToSrcVec
            float micYawDeg   = h->micArrayOrientations[mic][0];
            float micPitchDeg = h->micArrayOrientations[mic][1];
            float micRollDeg  = h->micArrayOrientations[mic][2];
            get_rotation_matrix_OpenGL(micYawDeg,
                                       micPitchDeg,
                                       micRollDeg,
                                       Rmic);
            std::copy_n(&Rmic[0][0], 3 * 3, &h->tempR[0][0]);
            matrixMult(micToSrcVec, h->tempR, micToSrcVecRotated);

            // Calculate DOA, distance, delay and attenuation gain for micToSrc vector
            h->micToSrcAziElevDeg[src][mic][0] = calculate_azimuth_OpenGL(&micToSrcVecRotated[0][0]) * 180.0f / M_PI;
            h->micToSrcAziElevDeg[src][mic][1] = calculate_elevation_OpenGL(&micToSrcVecRotated[0][0]) * 180.0f / M_PI;
            float micToSrcDistance = calculate_vector_norm_for_first_n_dimensions(&micToSrcVecRotated[0][0], 3);
            h->micToSrcDistances[src][mic] = micToSrcDistance;
            h->micToSrcDelaySamples[src][mic] = std::round((micToSrcDistance / h->speedOfSound) * h->sampleRate);
            h->micToSrcAttenuationGains[src][mic] = 1.0f / micToSrcDistance;

            // Update the closest mic indexes
            if (micToSrcDistance < minDistance) {
                minDistance2 = minDistance;
                minDistance = micToSrcDistance;
                secondClosestMic = closestMic;
                closestMic = mic;
            } else if (micToSrcDistance < minDistance2) {
                minDistance2 = micToSrcDistance;
                secondClosestMic = mic;
            }
        }

        // Collect the closest and second closest mic array for this source
        h->closestMicsForSources[src] = closestMic;
        h->secondClosestMicsForSources[src] = secondClosestMic;
        h->distanceToClosestMicForSource[src] = minDistance;
        h->distanceToSecondClosestMicForSource[src] = minDistance2;

        // Delays between each mic array for this source
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            for (int mic2 = 0; mic2 < h->nMicrophoneArrays; ++mic2) {
                h->micToMicDelaySamples[src][mic][mic2] = h->micToSrcDelaySamples[src][mic2] - h->micToSrcDelaySamples[src][mic];
                h->micToMicDelayHops[src][mic][mic2] = std::round((float)h->micToMicDelaySamples[src][mic][mic2] / (float)h->hopSize);
            }
        }

        // Delays and attenuation gains from the closest mic arrays to other arrays
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            h->closestMicToMicDelaySamples[src][mic] = h->micToMicDelaySamples[src][closestMic][mic];
            h->secondClosestMicToMicDelaySamples[src][mic] = h->micToMicDelaySamples[src][secondClosestMic][mic];
            h->closestMicToMicDelayHops[src][mic] = h->micToMicDelayHops[src][closestMic][mic];
            h->secondClosestMicToMicDelayHops[src][mic] = h->micToMicDelayHops[src][secondClosestMic][mic];
            h->closestMicToMicAttenuationGains[src][mic] = minDistance / h->micToSrcDistances[src][mic];
            h->secondClosestMicToMicAttenuationGains[src][mic] = minDistance2 / h->micToSrcDistances[src][mic];
            h->closestMicToMicAttenuationGainsSquared[src][mic] = std::pow(h->closestMicToMicAttenuationGains[src][mic], 2.0f);
            h->secondClosestMicToMicAttenuationGainsSquared[src][mic] = std::pow(h->secondClosestMicToMicAttenuationGains[src][mic], 2.0f);
        }
    }
    destroyHoatMatrix(&micToSrcVec);
    destroyHoatMatrix(&micToSrcVecRotated);
}

// yy^H for each source mic pair, where y = spherical harmonics from mic to src
void calculate_YY_for_mics_to_sources(hoaTranslatorData *h)
{
    auto copy_to_temp_matrix = [&](int src, int mic, float ***source) {
        for (int i = 0; i < 4; ++i) {
            h->temp4x1Matrix[i][0] = source[i][src][mic];
        }
    };

    int shOrder = 1;
    float ***Y = allocateFloatArray3d(ORDER2NSH(shOrder), h->nSources, h->nMicrophoneArrays);
    getRSH(shOrder, FLATTEN3D(h->micToSrcAziElevDeg), h->nSources * h->nMicrophoneArrays, FLATTEN3D(Y));
    for (int src = 0; src < h->nSources; ++src) {
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            copy_to_temp_matrix(src, mic, Y);
            matrixMultNoTransTrans(h->temp4x1Matrix, h->temp4x1Matrix, h->temp4x4Matrix);
            copy(h->temp4x4Matrix, h->micToSrcYY[src][mic]);
        }
    }
    free(Y);
}

// Beamformers from mic arrays to sources
void calculate_beams_from_mics_to_sources(hoaTranslatorData *h)
{
    // Get beamforming weights from closest mics to sources (hypercardioid/regular beamformers)
    int nSH = ORDER2NSH(h->shOrder);
    float *Y = allocateFloatArray1d(nSH);
    float *Y2 = allocateFloatArray1d(nSH);
    for (int src = 0; src < h->nSources; ++src) {
        getRSH(h->shOrder, &h->micToSrcAziElevDeg[src][h->closestMicsForSources[src]][0], 1, Y);
        getRSH(h->shOrder, &h->micToSrcAziElevDeg[src][h->secondClosestMicsForSources[src]][0], 1, Y2);
        for (int i = 0; i < nSH; ++i) {
            h->micToClosestSrcBeams[src][0][i] = float_complex(Y[i] / (float)nSH, 0.0f);
            h->micToSecondClosestSrcBeams[src][0][i] = float_complex(Y2[i] / (float)nSH, 0.0f);
        }
    }

    // Free memory
    free(Y);
    free(Y2);
}

void get_beamformed_source_signals(hoaTranslatorData *h)
{
    auto copy_to_temp_matrix = [&](int mic, int frame) {
        std::copy_n(&h->TFDInputFrameFullOrder[mic][frame][0][0], h->channelsPerArray * h->nBands, &h->tempInputFrameMatrix[0][0]);
    };

    auto copy_to_source_frame = [&](int src, int frame) {
        std::copy_n(&h->tempBandMatrix[0][0], h->nBands, &h->TFDSourceFrame[src][frame][0]);
    };

    for (int src : h->activeSources) {
        int mic = h->closestMicsForSources[src];
        auto const& beamWeights = h->micToClosestSrcBeams[src];
        for (int frame = 0; frame < h->nFrames; ++frame) {
            copy_to_temp_matrix(mic, frame);
            matrixMult(beamWeights, h->tempInputFrameMatrix, h->tempBandMatrix);
            copy_to_source_frame(src, frame);
        }
    }
}

// Compute energy for the TFDFrame (size: nFrames x nBands) for this frequency band.
float compute_energy(hoaTranslatorData *h, int band, float_complex **TFDFrame)
{
    float energy = 0.0f;
    for (int frame = 0; frame < h->nFrames; ++frame) {
        energy += (TFDFrame[frame][band] * std::conj(TFDFrame[frame][band])).real();
    }
    return energy;
}

// Compute energy for the first order input frame for this mic array and frequency band.
float compute_FOA_energy(hoaTranslatorData *h, int mic, int band)
{
    int nFOAch = 4;
    float energy = 0.0f;
    for (int ch = 0; ch < nFOAch; ++ch) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            energy += (h->TFDInputFrameFullOrder[mic][frame][ch][band] * std::conj(h->TFDInputFrameFullOrder[mic][frame][ch][band])).real();
        }
    }
    energy /= (float)nFOAch;
    return energy;
}

// Compute energies for source signals, when time-alignment is not used
void pre_compute_source_energies_without_time_alignment(hoaTranslatorData *h)
{
    for (int src : h->activeSources) {
        for (int band = 0; band < h->nBands; ++band) {
            h->Es_noTA[src][band] = compute_energy(h, band, h->TFDSourceFrame[src]);
        }
    }
}

void apply_post_filter(hoaTranslatorData *h, int src, int band, float postFilterGain) {
    for (int frame = 0; frame < h->nFrames; ++frame) {
        h->TFDSourceFrame[src][frame][band] *= postFilterGain;
    }
}

// Post-filter beamformed source signals based on cross pattern coherence (CroPaC).
void post_filter_source_signals_cropac_classic(hoaTranslatorData *h)
{
    for (int src : h->activeSources) {
        int mic = h->closestMicsForSources[src];
        std::vector<float> mic2srcDoa(3, 0.0f);
        calculate_doa_vector_cartesian(h->micToSrcAziElevDeg[src][mic][0] * M_PI / 180.0f, h->micToSrcAziElevDeg[src][mic][1] * M_PI / 180.0f, mic2srcDoa);
        for (int frame = 0; frame < h->nFrames; ++frame) {
            for (int band = 0; band < h->nBands; ++band) {
                // Figure-of-8 beamformer
                auto s_fig8 = mic2srcDoa[0] * h->TFDInputFrameFullOrder[mic][frame][3][band] +
                              mic2srcDoa[1] * h->TFDInputFrameFullOrder[mic][frame][1][band] +
                              mic2srcDoa[2] * h->TFDInputFrameFullOrder[mic][frame][2][band];
                float E_foa = 0.0f;
                for (int ch = 0; ch < 4; ++ch) {
                    E_foa += std::pow(h->TFDInputFrameFullOrder[mic][frame][ch][band].real(), 2.0f) + std::pow(h->TFDInputFrameFullOrder[mic][frame][ch][band].imag(), 2.0f);
                }
                E_foa /= 4.0f;

                auto c_cropac = s_fig8 * std::conj(h->TFDInputFrameFullOrder[mic][frame][0][band]) / (E_foa + FLT_MIN);
                float G = c_cropac.real();
                G = (std::min)(G, 1.0f);
                G = (std::max)(G, 0.0f);
                h->TFDSourceFrame[src][frame][band] *= G;
            }
        }
    }
}

/* Post-filter beamformed source signals based on source and total energy:
 *
 *      post_filter = E_src / E_total
 */
void post_filter_source_signals_energy(hoaTranslatorData *h)
{
    for (int src : h->activeSources) {
        int mic = h->closestMicsForSources[src];
        for (int band = 0; band < h->nBands; ++band) {

            // Total energy from the closest mic array
            float E_total = compute_FOA_energy(h, mic, band);

            // Source energy for the beamformer
            float E_src = compute_energy(h, band, h->TFDSourceFrame[src]);

            // Calculate post-filter
            float G = E_src / (E_total + FLT_MIN);
            G = (std::min)(G, 1.0f);

            // Apply the post-filter gain
            apply_post_filter(h, src, band, G);
        }
    }
}

void time_align_source_signals(hoaTranslatorData *h)
{
    // Shift buffer by nFrames, and copy current TFDSourceFrame to the end of the buffer
    auto update_source_frame_buffer = [&]() {
        for (int src : h->activeSources) {
            for (int frame = 0; frame < h->nSourceBufferFrames - h->nFrames; ++frame) {
                for (int band = 0; band < h->nBands; ++band) {
                    h->sourceFrameBuffer[src][frame][band] = h->sourceFrameBuffer[src][frame + h->nFrames][band];
                }
            }
            std::copy_n(&h->TFDSourceFrame[src][0][0], h->nFrames * h->nBands, &h->sourceFrameBuffer[src][h->nSourceBufferFrames - h->nFrames][0]);
        }
    };

    // Time-align source signals at each mic array position
    auto time_align_at_each_mic_array = [&]() {
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            for (int src : h->activeSources) {
                int delayHops = (std::min)(h->closestMicToMicDelayHops[src][mic], h->maxDelayHops);
                int startingHop = h->nSourceBufferFrames - h->nFrames - delayHops;
                std::copy_n(&h->sourceFrameBuffer[src][startingHop][0], h->nFrames * h->nBands, &h->TFDMicSrcTAFrame[mic][src][0][0]);
            }
        }
    };

    // Time-align source signals at the listener position. Use the delay from the closest mic array from the listener.
    auto time_align_at_listener_position = [&]() {
        for (int src : h->activeSources) {
            int delayHops = (std::min)(h->closestMicToMicDelayHops[src][h->currentClosestMicrophoneArray], h->maxDelayHops);
            int startingHop = h->nSourceBufferFrames - h->nFrames - delayHops;
            std::copy_n(&h->sourceFrameBuffer[src][startingHop][0], h->nFrames * h->nBands, &h->TFDListenerSrcTAFrame[src][0][0]);
        }
    };

    update_source_frame_buffer();
    time_align_at_each_mic_array();
    time_align_at_listener_position();
}

void estimate_mic_model_SCM(hoaTranslatorData *h)
{
    // Reset SCM
    fillWithZeros(h->Cam_src);

    float Es = 0.0f;
    float attenuation = 0.0f;
    float multiplier = 0.0f;
    for (int src : h->activeSources) {
        for (int mic : h->activeArraysForCurrentFrame) {
            attenuation = h->closestMicToMicAttenuationGainsSquared[src][mic];
            for (int band = 0; band < h->nBands; ++band) {
                // Compute source energy Es
                if (h->useSourceAwareTimeAlignment) {
                    Es = compute_energy(h, band, h->TFDMicSrcTAFrame[mic][src]);
                } else {
                    Es = h->Es_noTA[src][band];
                }

                // Cam_src = attenuation * Es * micToSrcYY
                multiplier = attenuation * Es;
                matrixScalarMult(multiplier, h->micToSrcYY[src][mic], h->temp4x4Matrix);
                matrixMatrixAdd(h->Cam_src[mic][band], h->temp4x4Matrix, h->Cam_src[mic][band]);
            }
        }
    }
}

void estimate_listener_model_SCM(hoaTranslatorData *h)
{
    // Reset SCMs
    fillWithZeros(h->Cyo_src);
    fillWithZeros(h->Cao_src);

    int shOrderListenerToSource = 1;
    float Es = 0.0f;
    float multiplier = 0.0f;
    std::vector<float> aziElev(2, 0.0f);
    for (int src : h->activeSources) {
        int mic = h->closestMicsForSources[src];

        float attenuation = (std::min)(std::pow(h->micToSrcDistances[src][mic] / h->listenerToSrcDistances[src], 2.0f), std::pow(h->micToSrcDistances[src][mic] / h->maxListenerModelAttenuationDistance, 2.0f));
        float aziRad = calculate_azimuth_OpenGL(&h->listenerToSrcVectorsOpenGL[src][0][0]);
        float elevRad = calculate_elevation_OpenGL(&h->listenerToSrcVectorsOpenGL[src][0][0]);
        int dirIndex = getHrtfDirIndex(h, aziRad, elevRad);
        aziElev[0] = aziRad * 180.0f / M_PI;
        aziElev[1] = elevRad * 180.0f / M_PI;
        // SHs from listener to source (FOA only)
        getRSH_recur(shOrderListenerToSource, aziElev.data(), 1, FLATTEN2D(h->temp4x1Matrix.data));

        for (int band = 0; band < h->nBands; ++band) {
            // Compute source energy Es
            if (h->useSourceAwareTimeAlignment) {
                Es = compute_energy(h, band, h->TFDListenerSrcTAFrame[src]);
            } else {
                Es = h->Es_noTA[src][band];
            }

            // Multiplier for matrix products
            multiplier = attenuation * Es;

            if (h->targetScmMode == SEPARATE_SCM) {
                // Binaural model SCM: Cyo_src = attenuation * Es * (HRTF * HRTF^H)
                copy_hrtf_to_matrix(h, band, dirIndex);
                matrixMultNoTransConj(h->tempHrtfMatrix, h->tempHrtfMatrix, h->tempHrtfProductMatrix);
                matrixScalarMult(multiplier, h->tempHrtfProductMatrix, h->tempHrtfProductMatrix);
                matrixMatrixAdd(h->Cyo_src[band], h->tempHrtfProductMatrix, h->Cyo_src[band]);

            } else if (h->targetScmMode == COMPOUND_PARAMS) {
                // Ambisonics (FOA) model SCM: Cao_src = attenuation * Es * (Y * Y^T)
                matrixMultNoTransTrans(h->temp4x1Matrix, h->temp4x1Matrix, h->temp4x4Matrix);
                matrixScalarMult(multiplier, h->temp4x4Matrix, h->temp4x4Matrix);
                matrixMatrixAdd(h->Cao_src[band], h->temp4x4Matrix, h->Cao_src[band]);
            }
        }
    }
}

void spatial_analysis_for_mic_model_SCM(hoaTranslatorData *h)
{
    std::vector<float> intensityVector(3, 0.0f);
    float energy = 0.0f;
    float energyNormalizer = (float)h->Cam_src[0][0].nRows * (float)h->nFrames;
    float intensityNormalizer = std::sqrt(3) * h->nFrames;

    for (int mic : h->activeArraysForCurrentFrame) {
        for (int band = 0; band < h->nBands; ++band) {

            // Calculate energy
            energy = sumDiagonal(h->Cam_src[mic][band]) / energyNormalizer;

            // Convert ACN channel order to cartesian x,y,z
            intensityVector[0] = h->Cam_src[mic][band][3][0] / intensityNormalizer;
            intensityVector[1] = h->Cam_src[mic][band][1][0] / intensityNormalizer;
            intensityVector[2] = h->Cam_src[mic][band][2][0] / intensityNormalizer;

            // Spatial analysis
            h->srcAzimuthData[mic][band]   = calculate_azimuth(intensityVector);
            h->srcElevationData[mic][band] = calculate_elevation(intensityVector);
            h->srcDTRData[mic][band]       = calculate_dtr(intensityVector, energy);
            h->srcEnergyData[mic][band]    = energy;
        }
    }
}

// Intensity vector = energy * dtr * doa
void calculate_intensity_vector(const float &azimuthRad, const float &elevationRad, const float &energy,
                                const float &dtr, std::vector<float> &intensityVector) {
    float multiplier = energy * dtr;
    calculate_doa_vector_cartesian(azimuthRad, elevationRad, intensityVector);
    utility_svsmul(intensityVector.data(), &multiplier, 3, NULL);
}

void calculate_intensity_vectors(hoaTranslatorData *h, const float *azimuthRads, const float *elevationRads, float *energies,
                                const float *dtrs, float *intensityX, float *intensityY, float *intensityZ) {
    utility_svvmul(energies, dtrs, h->nBands, h->intensityVectorMultipliers);
    calculate_doa_vectors_cartesian(h, h->nBands, azimuthRads, elevationRads, intensityX, intensityY, intensityZ);
    utility_svvmul(intensityX, h->intensityVectorMultipliers, h->nBands, intensityX);
    utility_svvmul(intensityY, h->intensityVectorMultipliers, h->nBands, intensityY);
    utility_svvmul(intensityZ, h->intensityVectorMultipliers, h->nBands, intensityZ);
}

/* Compute spatial parameters for the residual signals. The residual parameters are calculated from a residual intensity
 * vector and residual signal energy. These are obtained from:
 *
 *      residual_intensity = mic_intensity - src_intensity
 *      residual_energy = mic_energy - src_energy
 *
 * Note that the energy and the magnitude of the intensity of the source model are obtained from a whole processing
 * frame. I.e. the source model values need to be divided to sub-frames.
 *
 * Output:
 *      interpolatedResidualAzimuthData         - interpolated azimuth data for the residual signal   (nMicrohoneArrays x nFrames x nBands)
 *      interpolatedResidualElevationData       - interpolated elevation data for the residual signal (nMicrohoneArrays x nFrames x nBands)
 *      interpolatedResidualDTRData             - interpolated DTR data for the residual signal       (nMicrohoneArrays x nFrames x nBands)
 *      interpolatedResidualEnergyData          - interpolated energy data for the residual signal    (nMicrohoneArrays x nFrames x nBands)
 */
void compute_residual_parameters(hoaTranslatorData *h)
{
    // Intensity vectors for the beamformed signals from the mic array to source.
    auto calculate_src_intensity_vectors = [&](int mic) {
        calculate_intensity_vectors(h, h->srcAzimuthData[mic], h->srcElevationData[mic], h->srcEnergyData[mic], h->srcDTRData[mic], h->srcIntensityX, h->srcIntensityY, h->srcIntensityZ);
    };

    // Intensity vectors for the mic array
    auto calculate_mic_intensity_vectors = [&](int mic, int frame) {
        calculate_intensity_vectors(h, h->azimuthData[mic][frame], h->elevationData[mic][frame], h->energyData[mic][frame], h->DTRData[mic][frame], h->micIntensityX, h->micIntensityY, h->micIntensityZ);
    };

    // Intensity vectors for the residual signal = mic_intensity - src_intensity
    auto calculate_res_intensity_vectors = [&]() {
        subtractVectors(h->micIntensityX, h->srcIntensityX, h->resIntensityX, h->nBands);
        subtractVectors(h->micIntensityY, h->srcIntensityY, h->resIntensityY, h->nBands);
        subtractVectors(h->micIntensityZ, h->srcIntensityZ, h->resIntensityZ, h->nBands);
    };

    auto calculate_residual_energy = [&](int mic, int frame) {
        subtractVectors(h->energyData[mic][frame], h->srcEnergyData[mic], h->residualEnergyData[mic][frame], h->nBands);
        for (int i = 0; i < h->nBands; ++i) {
            h->residualEnergyData[mic][frame][i] = (std::max)(h->residualEnergyData[mic][frame][i], FLT_MIN);
        }
    };

    // In recursive averaging, all sub-frames have the same average data for the whole processing frame, so we can just
    // copy the first residual sub-frame data to the rest of the sub-frames.
    auto copy_to_subframes = [&](int mic) {
        for (int frame = 1; frame < h->nFrames; ++frame) {
            std::copy_n(&h->residualAzimuthData[mic][0][0],   h->nBands, &h->residualAzimuthData[mic][frame][0]);
            std::copy_n(&h->residualElevationData[mic][0][0], h->nBands, &h->residualElevationData[mic][frame][0]);
            std::copy_n(&h->residualDTRData[mic][0][0],       h->nBands, &h->residualDTRData[mic][frame][0]);
            std::copy_n(&h->residualEnergyData[mic][0][0],    h->nBands, &h->residualEnergyData[mic][frame][0]);
        }
    };

    std::vector<float> intensityVectorMic(3, 0.0f);
    std::vector<float> intensityVectorMicAvgSum(3, 0.0f);
    std::vector<float> intensityVectorSrc(3, 0.0f);
    std::vector<float> intensityVectorResidual(3, 0.0f);
    for (int mic : h->activeArraysForCurrentFrame) {
        calculate_src_intensity_vectors(mic);
        for (int frame = 0; frame < h->nFrames; ++frame) {
            calculate_mic_intensity_vectors(mic, frame);
            calculate_res_intensity_vectors();

            // Residual energy
            calculate_residual_energy(mic, frame);

            // Residual azi, elev and DTR
            calculate_azimuth(h->resIntensityX, h->resIntensityY, h->residualAzimuthData[mic][frame], h->nBands);
            calculate_elevation(h->resIntensityX, h->resIntensityY, h->resIntensityZ, h->tempBandVec1, h->tempBandVec2, h->residualElevationData[mic][frame], h->nBands);
            calculate_dtr(h->resIntensityX, h->resIntensityY, h->resIntensityZ, h->tempBandVec1, h->tempBandVec2, h->tempBandVec3, h->residualEnergyData[mic][frame], h->residualDTRData[mic][frame], h->nBands);

            if (h->useRecursiveAveraging) {
                copy_to_subframes(mic);
                break;
            }
        }
    }

    // Interpolate the residual parameters according to the listener position and orientation
    interpolate_residual_metadata(h);
}

void build_target_SCM_from_combined_parameters(hoaTranslatorData *h)
{
    std::vector<float> intensityVector(3, 0.0f);
    std::vector<float> intensityVectorSrc(3, 0.0f);
    std::vector<float> intensityVectorResidual(3, 0.0f);
    std::vector<float> intensityVectorResidualSum(3, 0.0f);
    int dirIndex = 0;
    float aziRad = 0.0f;
    float elevRad = 0.0f;
    float dtr = 0.0f;
    float energy = 0.0f;
    float energySrc = 0.0f;
    float energyResidual = 0.0f;
    float energyNormalizer = (float)h->Cao_src[0].nRows * (float)h->nFrames;
    float intensityNormalizer = std::sqrt(3) * (float)h->nFrames;
    fillWithZeros(h->newCyMatrix);

    for (int band = 0; band < h->nBands; ++band) {

        // Calculate energy
        energySrc = sumDiagonal(h->Cao_src[band]) / energyNormalizer;

        // Convert ACN channel order to cartesian x,y,z
        intensityVectorSrc[0] = h->Cao_src[band][3][0] / intensityNormalizer;
        intensityVectorSrc[1] = h->Cao_src[band][1][0] / intensityNormalizer;
        intensityVectorSrc[2] = h->Cao_src[band][2][0] / intensityNormalizer;

        for (int frame = 0; frame < h->nFrames; ++frame) {

            // Residual intensity vector and energy
            calculate_intensity_vector(h->interpolatedResidualAzimuthData[frame][band],
                                       h->interpolatedResidualElevationData[frame][band],
                                       h->interpolatedResidualEnergyData[frame][band],
                                       h->interpolatedResidualDTRData[frame][band],
                                       intensityVectorResidual);
            energyResidual = h->interpolatedResidualEnergyData[frame][band];

            // Combine source and residual intensity vectors and energies
            if (h->sourceResidualMode == SOURCE_ONLY) {
                energy = energySrc;
                std::copy_n(intensityVectorSrc.data(), 3, intensityVector.data());
            } else if (h->sourceResidualMode == RESIDUAL_ONLY) {
                energy = energyResidual;
                std::copy_n(intensityVectorResidual.data(), 3, intensityVector.data());
            } else if (h->sourceResidualMode == BOTH) {
                energy = energySrc + energyResidual;
                std::copy_n(intensityVectorSrc.data(), 3, intensityVector.data());
                utility_svvadd(intensityVector.data(), intensityVectorResidual.data(), 3, intensityVector.data());
            }

            // Calculate SCM from the combined parameters
            aziRad   = calculate_azimuth(intensityVector);
            elevRad  = calculate_elevation(intensityVector);
            dtr      = (std::min)(calculate_dtr(intensityVector, energy), 1.0f);
            dirIndex = getHrtfDirIndex(h, aziRad, elevRad);
            compute_direct_part_of_Cy(h, frame, band, dirIndex, energy, dtr, h->newCyMatrix);
            compute_diffuse_part_of_Cy(h, frame, band, energy, dtr, h->newCyMatrix);
        }
    }

    // Compute the target Cy
    for (int band = 0; band < h->nBands; ++band) {
        combine_covariance_matrices(h, h->Cy[band], h->previousCyMatrix[band], h->newCyMatrix[band]);
    }
}

void build_target_SCM_from_separate_SCMs(hoaTranslatorData *h)
{
    // Calculate residual binaural SCM
    if (h->sourceResidualMode == RESIDUAL_ONLY || h->sourceResidualMode == BOTH) {
        int dirIndex = 0;
        float energy = 0.0f;
        float dtr = 0.0f;
        fillWithZeros(h->Cyo_res);
        for (int frame = 0; frame < h->nFrames; ++frame) {
            for (int band = 0; band < h->nBands; ++band) {
                energy   = h->interpolatedResidualEnergyData[frame][band];
                dtr      = h->interpolatedResidualDTRData[frame][band];
                dirIndex = getHrtfDirIndex(h, h->interpolatedResidualAzimuthData[frame][band],
                                           h->interpolatedResidualElevationData[frame][band]);
                compute_direct_part_of_Cy(h, frame, band, dirIndex, energy, dtr, h->Cyo_res);
                compute_diffuse_part_of_Cy(h, frame, band, energy, dtr, h->Cyo_res);
            }
        }
    }

    // Build the new target SCM (newCy)
    if (h->sourceResidualMode == SOURCE_ONLY) {
        copy(h->Cyo_src, h->newCyMatrix);
    } else if (h->sourceResidualMode == RESIDUAL_ONLY) {
        copy(h->Cyo_res, h->newCyMatrix);
    } else if (h->sourceResidualMode == BOTH) {
        copy(h->Cyo_res, h->newCyMatrix);
        matrixMatrixAdd(h->newCyMatrix, h->Cyo_src, h->newCyMatrix);
    }

    // Compute the target Cy
    for (int band = 0; band < h->nBands; ++band) {
        combine_covariance_matrices(h, h->Cy[band], h->previousCyMatrix[band], h->newCyMatrix[band]);
    }
}

void determine_active_sources_and_update_track_record(hoaTranslatorData *h)
{
    auto listener_threshold_passed = [&](int src) {
        if (h->listenerThresholdPresentVec[src]) {
            return h->listenerToSrcDistances[src] < h->listenerThresholdVec[src];
        }
        else {
            return true;
        }
    };
    auto hoasource_threshold_passed = [&](int src) {
        if (h->hoaSourceThresholdPresentVec[src]) {
            return h->distanceToClosestMicForSource[src] < h->hoaSourceThresholdVec[src];
        }
        else {
            return true;
        }
    };

    // Find the active source candidates and their closest mic arrays
    h->activeSources.clear();
    h->activeSourceCandidates.clear();
    h->transformedSourceArrays.clear();

    if (h->maxSimulSources > 0) {
        // get sorted infored source indices in priority order
        std::vector<int> sortedPriorityValueIndices(h->nSources);
        std::iota(sortedPriorityValueIndices.begin(),sortedPriorityValueIndices.end(),0);
        sort(sortedPriorityValueIndices.begin(),sortedPriorityValueIndices.end(), [&](int i,int j){return h->priorityValueVec[i] > h->priorityValueVec[j];} );

        for (int i = 0; i < h->nSources; ++i) {
            // go through sources in priority order
            int src = sortedPriorityValueIndices[i];
            if (i > h->maxSimulSources - 1) {
                // maximum number of simultaneous informed sources reached
                break;
            }
            // check listener and hoa source thresholds
            if (hoasource_threshold_passed(src) && listener_threshold_passed(src)) {
                h->activeSourceCandidates.push_back(src);
                h->transformedSourceArrays.push_back(h->closestMicsForSources[src]);
            }
        }
    }

    // Update the track record for source mic arrays
    h->trackedSourceMicArrays[0] = h->transformedSourceArrays;

    // Look for source candidate mic arrays at the end of the track records (where the STFT buffers are flushed)
    auto lastTrackedSourceMicArrays = h->trackedSourceMicArrays[h->nTotalFramesTracked - 1];
    auto lastTrackedTriangleMicArrays = h->trackedCurrentTriangleArrays[h->nTotalFramesTracked - 1];
    for (int src : h->activeSourceCandidates) {
        int mic = h->closestMicsForSources[src];
        // The mic was found at the end of the triangle track record. Mark the source active, and fill the source mic track
        // record with the mic index.
        if (std::find(lastTrackedTriangleMicArrays.begin(), lastTrackedTriangleMicArrays.end(), mic) != lastTrackedTriangleMicArrays.end()) {
            for (auto&& trackVec : h->trackedSourceMicArrays) {
                trackVec.push_back(mic);
            }
            h->activeSources.push_back(src);
            continue;
        }
        // The mic was found at the end of source mic track record, mark the source active
        if (std::find(lastTrackedSourceMicArrays.begin(), lastTrackedSourceMicArrays.end(), mic) != lastTrackedSourceMicArrays.end()) {
            h->activeSources.push_back(src);
        }
    }

    // Shift the tracked values by one frame
    for (int i = h->nTotalFramesTracked - 1; i > 0; --i) {
        h->trackedSourceMicArrays[i] = h->trackedSourceMicArrays[i - 1];
    }
}

void check_for_source_aware_mode(hoaTranslatorData *h)
{
    // Disable source aware mode check, if the mode was disabled in the initialization
    if (!h->useSourceAwareInitialFlag) {
        return;
    }

    calculate_mean_listener_position(h);
    calculate_listener_to_source_vectors(h);
    determine_active_sources_and_update_track_record(h);
    if (h->activeSources.empty()) {
        h->useSourceAwareMode = false;
    } else {
        h->useSourceAwareMode = true;
    }
}

void pre_process_source_aware(hoaTranslatorData *h)
{
    calculate_source_geometry_parameters(h);
    calculate_YY_for_mics_to_sources(h);
    calculate_beams_from_mics_to_sources(h);
    calculate_distances_to_outside_sources(h);
}

void source_aware_processing(hoaTranslatorData *h)
{
    get_beamformed_source_signals(h);

    if (h->useSourceAwarePostFilter) {
        post_filter_source_signals_cropac_classic(h);
        //post_filter_source_signals_energy(h);
    }

    if (h->useSourceAwareTimeAlignment) {
        time_align_source_signals(h);
    } else {
        pre_compute_source_energies_without_time_alignment(h);
    }

    estimate_mic_model_SCM(h);
    estimate_listener_model_SCM(h);

    spatial_analysis_for_mic_model_SCM(h);

    compute_residual_parameters(h);

    if (h->targetScmMode == COMPOUND_PARAMS) {
        build_target_SCM_from_combined_parameters(h);
    } else if (h->targetScmMode == SEPARATE_SCM) {
        build_target_SCM_from_separate_SCMs(h);
    }
}
