/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SpatialAnalysis.hpp"

#include <algorithm>
#include <cmath>
#include <complex>
#include <vector>

#include "MetadataInterpolator.hpp"
#include "VectorOperations.hpp"

#ifdef SAF_USE_INTEL_MKL
#include "mkl.h"
#elif SAF_USE_APPLE_ACCELERATE
#include <Accelerate/Accelerate.h>
#endif

float calculate_azimuth(const float* xyzVec)
{
    return std::atan2(xyzVec[1],
                      xyzVec[0]);
}

float calculate_azimuth(const std::vector<float> &xyzVec)
{
    return std::atan2(xyzVec[1],
                      xyzVec[0]);
}

float calculate_elevation(const float* xyxVec)
{
    return std::atan2(xyxVec[2], calculate_vector_norm_for_first_n_dimensions(xyxVec, 2));
}

float calculate_elevation(const std::vector<float> &xyxVec)
{
    return std::atan2(xyxVec[2], calculate_vector_norm_for_first_n_dimensions(xyxVec, 2));
}

float calculate_dtr(const float* xyzVec, float energy)
{
    return calculate_vector_norm_for_first_n_dimensions(xyzVec, 3) / (energy + FLT_MIN);
}

float calculate_dtr(const std::vector<float> &xyzVec, float energy)
{
    return calculate_vector_norm(xyzVec) / (energy + FLT_MIN);
}

float calculate_dtr_without_energy(const float* xyzVec)
{
    return calculate_vector_norm_for_first_n_dimensions(xyzVec, 3);
}

void calculate_azimuth(const float* xVec, const float* yVec, float* azi, int vecLen)
{
    atan2Vector(yVec, xVec, azi, vecLen);
}

void calculate_elevation(float* xVec, float* yVec, const float* zVec, float* xVecPow, float* yVecPow, float* elev, int vecLen)
{
    calculate_vector_norm(xVec, yVec, xVecPow, yVecPow, elev, vecLen);
    atan2Vector(zVec, elev, elev, vecLen);
}

void calculate_dtr(float* xVec, float* yVec, float* zVec, float* xVecPow, float* yVecPow, float* zVecPow, const float* energy, float* dtr, int vecLen)
{
    // sqrt(x^2 + y^2 + z^2)
    calculate_vector_norm(xVec, yVec, zVec, xVecPow, yVecPow, zVecPow, dtr, vecLen);

    // dtr = sqrt(x^2 + y^2 + z^2) / energy
    #if defined(SAF_USE_INTEL_MKL)
    vsDiv(vecLen, dtr, energy, dtr);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vDSP_vdiv(energy, 1, dtr, 1, dtr, 1, vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
        dtr[i] /= energy[i];
    }
    #endif

    for (int i = 0; i < vecLen; ++i) {
        dtr[i] = (std::min)(dtr[i], 1.0f);
    }
}


float calculate_azimuth_OpenGL(const float* xyzVec)
{
    return std::atan2(-xyzVec[0],
                      -xyzVec[2]);
}

std::vector<float> calculate_azimuths_OpenGL(std::vector<float> &xVec, std::vector<float> &zVec)
{
    auto xVecInv = multiplyVector(-1.0f, xVec);
    auto zVecInv = multiplyVector(-1.0f, zVec);
    return atan2Vector(xVecInv.data(), zVecInv.data(), xVecInv.size());
}

void calculate_azimuths_OpenGL(hoaTranslatorData *h, float *xVec, float *zVec, float *aziVec, int vecLen)
{
    float invMult = -1.0f;
    if (vecLen <= h->nBands) {
        utility_svsmul(&xVec[0], &invMult, vecLen, &h->tempNbands1[0]);
        utility_svsmul(&zVec[0], &invMult, vecLen, &h->tempNbands2[0]);
        atan2Vector(h->tempNbands1, h->tempNbands2, aziVec, vecLen);
    } else {
        auto xVecInv = multiplyVector(-1.0f, xVec, vecLen);
        auto zVecInv = multiplyVector(-1.0f, zVec, vecLen);
        atan2Vector(xVecInv.data(), zVecInv.data(), aziVec, vecLen);
    }
}

float calculate_elevation_OpenGL(const float* xyxVec)
{
    return std::atan2(xyxVec[1],
                      std::sqrt(std::pow(xyxVec[0], 2.0f) +
                                std::pow(xyxVec[2], 2.0f)));
}

std::vector<float> calculate_elevations_OpenGL(std::vector<float> &xVec, std::vector<float> &yVec, std::vector<float> &zVec)
{
    auto xVecSqr = powVector(2.0f, xVec.data(), xVec.size());
    auto zVecSqr = powVector(2.0f, zVec.data(), zVec.size());
    auto xzSqrSum = sumVectors(xVecSqr, zVecSqr);
    auto xzSqrSumSqrt = sqrtVector(xzSqrSum.data(), xzSqrSum.size());

    return atan2Vector(yVec.data(), xzSqrSumSqrt.data(), yVec.size());
}

void calculate_elevations_OpenGL(hoaTranslatorData *h, float *xVec, float *yVec, float *zVec, float *elevVec, int vecLen)
{
    if (vecLen <= h->nBands) {
        powVector(2.0f, xVec, vecLen, h->tempNbands1);
        powVector(2.0f, zVec, vecLen, h->tempNbands2);
        utility_svvadd(&h->tempNbands1[0], &h->tempNbands2[0], vecLen, &h->tempNbands1[0]);
        sqrtVector(&h->tempNbands1[0], &h->tempNbands2[0], vecLen);
        atan2Vector(&yVec[0], h->tempNbands2, &elevVec[0], vecLen);
    } else {
        auto xVecSqr = powVector(2.0f, &xVec[0], vecLen);
        auto zVecSqr = powVector(2.0f, &zVec[0], vecLen);
        auto xzSqrSum = sumVectors(xVecSqr, zVecSqr);
        auto xzSqrSumSqrt = sqrtVector(xzSqrSum.data(), vecLen);
        atan2Vector(&yVec[0], xzSqrSumSqrt.data(), &elevVec[0], vecLen);
    }
}

void calculate_doa_vector_cartesian(float azimuthRad, float elevationRad, std::vector<float> &doa)
{
    assert(doa.size() == 3);
    doa[0] = std::cos(azimuthRad) * std::cos(elevationRad);
    doa[1] = std::sin(azimuthRad) * std::cos(elevationRad);
    doa[2] = std::sin(elevationRad);
}

void calculate_doa_vector_OpenGL(float azimuthRad, float elevationRad, std::vector<float> &doa)
{
    assert(doa.size() == 3);
    doa[0] = -(std::sin(azimuthRad) * std::cos(elevationRad));
    doa[1] = std::sin(elevationRad);
    doa[2] = -(std::cos(azimuthRad) * std::cos(elevationRad));
}

void calculate_doa_vectors_cartesian(hoaTranslatorData *h,
                                     int vecLen,
                                     const float *azimuthRads,
                                     const float *elevationRads,
                                     float *doasX,
                                     float *doasY,
                                     float *doasZ)
{
    assert(vecLen <= h->nMicrophoneArrays * h->nFrames * h->nBands && "vecLen exceeds the maximum length! Increase the sizes of sinAzi/Elev and cosAzi/Elev.");

    // Cosine and sine of azimuth and elevation DOAs
    get_sin_cos(h->sinAzi, h->cosAzi, azimuthRads, vecLen);
    get_sin_cos(h->sinElev, h->cosElev, elevationRads, vecLen);

    // x (front) = cos(azi) * cos(elev)
    utility_svvmul(h->cosAzi, h->cosElev, vecLen, doasX);

    // y (left) = sin(azi) * cos(elev)
    utility_svvmul(h->sinAzi, h->cosElev, vecLen, doasY);

    // z (up) = sin(elev)
    utility_svvcopy(h->sinElev, vecLen, doasZ);
}

void calculate_doa_vectors_OpenGL(hoaTranslatorData *h,
                                  int vecLen,
                                  const float *azimuthRads,
                                  const float *elevationRads,
                                  float *doasX,
                                  float *doasY,
                                  float *doasZ)
{

    float *cosAzi{};
    float *sinAzi{};
    float *cosElev{};
    float *sinElev{};
    if (vecLen <= h->nBands) {
        cosAzi = h->cosAziNbands;
        sinAzi = h->sinAziNbands;
        cosElev = h->cosElevNbands;
        sinElev = h->sinElevNbands;
    } else {
        std::vector<float> cosAziVec(vecLen, 0.0f);
        std::vector<float> sinAziVec(vecLen, 0.0f);
        std::vector<float> cosElevVec(vecLen, 0.0f);
        std::vector<float> sinElevVec(vecLen, 0.0f);
        cosAzi = cosAziVec.data();
        sinAzi = sinAziVec.data();
        cosElev = cosElevVec.data();
        sinElev = sinElevVec.data();
    }

    assert(vecLen <= h->nMicrophoneArrays * h->nFrames * h->nBands && "vecLen exceeds the maximum length! Increase the sizes of sinAzi/Elev and cosAzi/Elev.");

    float inverseMult = -1.0f;

    // Cosine and sine of azimuth and elevation DOAs
    get_sin_cos(h->sinAzi, h->cosAzi, azimuthRads, vecLen);
    get_sin_cos(h->sinElev, h->cosElev, elevationRads, vecLen);

    // x (right) = -sin(azi) * cos(elev)
    utility_svsmul(h->sinAzi, &inverseMult, vecLen, h->sinAzi);
    utility_svvmul(h->sinAzi, h->cosElev, vecLen, doasX);

    // y (up) = sin(elev)
    utility_svvcopy(h->sinElev, vecLen, doasY);

    // z (back) = -cos(azi) * cos(elev)
    utility_svsmul(h->cosAzi, &inverseMult, vecLen, h->cosAzi);
    utility_svvmul(h->cosAzi, h->cosElev, vecLen, doasZ);
}

void get_sin_cos(float *sinData, float *cosData, const float *sourceData, int vecLen)
{
    #if defined(SAF_USE_INTEL_MKL)
    vsSinCos(vecLen, sourceData, sinData, cosData);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vvsincosf(sinData, cosData, sourceData, &vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
        sinData[i] = std::sin(sourceData[i]);
        cosData[i] = std::cos(sourceData[i]);
    }
    #endif
}

void get_interpolated_listener_position_in_array_space(hoaTranslatorData *h, int mic, int frame, std::vector<float> &interpolatedListenerPos)
{
    float interpolator = h->interpolator[frame];
    if (h->useRecursiveAveraging) {
        interpolator = 0.5f;
    }
    float currentListPosMicSpace = 0.0f;
    float previousListPosMicSpace = 0.0f;
    std::vector<float> currentListenerPosCartesian(3, 0.0f);
    std::vector<float> previousListenerPosCartesian(3, 0.0f);
    std::vector<float> micArrayPosCartesian(3, 0.0f);
    openGL_to_cartesian_xFront(h->currentListenerPosition, currentListenerPosCartesian);
    openGL_to_cartesian_xFront(h->previousListenerPosition, previousListenerPosCartesian);
    openGL_to_cartesian_xFront(h->microphoneArrayPositions[mic], micArrayPosCartesian.data());
    for (int i = 0; i < 3; ++i) {
        currentListPosMicSpace = currentListenerPosCartesian[i] - micArrayPosCartesian[i];
        previousListPosMicSpace = previousListenerPosCartesian[i] - micArrayPosCartesian[i];
        interpolatedListenerPos[i] = interpolator * currentListPosMicSpace +
                                     (1.0f - interpolator) * previousListPosMicSpace;
    }
}

void calculate_new_doa_vector(hoaTranslatorData *h, const float* intensityVector,
                              std::vector<float> &newDoaVector, const std::vector<float> &interpolatedListenerPos)
{
    // Original DOA unit vector (= intensityVector / |intensityVector|)
    std::vector<float> doaVec(3, 0.0f);
    float intensityVecMagnitude = calculate_vector_norm_for_first_n_dimensions(intensityVector, 3);
    for (int i = 0; i < 3; ++i) {
        doaVec[i] = intensityVector[i] / (intensityVecMagnitude + FLT_MIN);
    }

    // New DOA vector (= vector from the listener to the projected source position)
    for (int i = 0; i < 3; ++i) {
        newDoaVector[i] = (doaVec[i] * h->sphoa6DofProjectionRadius) - interpolatedListenerPos[i];
    }
}

void calculate_sphoa_dir_gain(hoaTranslatorData *h, int frame, int band, const std::vector<float> &newDoaVector)
{
    float newDoaMagnitude = calculate_vector_norm(newDoaVector);
    float gain = h->sphoa6DofProjectionRadius / newDoaMagnitude;
    h->singlePoint6DofDirGains[frame][band] = (std::min)(gain, h->sphoa6DofMaxGain);
}

void spatial_analysis(hoaTranslatorData *h)
{
    auto calculate_doa_parameters = [&](int mic, int frame, int band, const std::vector<float> &vector) {
        h->azimuthData[mic][frame][band]   = calculate_azimuth(vector);
        h->elevationData[mic][frame][band] = calculate_elevation(vector);
    };

    auto reset_average_estimates = [&]() {
        std::fill_n(&h->intensityVectorAvg[0][0], h->nDoaDtrBins * 3, 0.0f);
        std::fill_n(&h->azimuthDataAvg[0],        h->nDoaDtrBins,     0.0f);
        std::fill_n(&h->elevationDataAvg[0],      h->nDoaDtrBins,     0.0f);
        std::fill_n(&h->DTRDataAvg[0],            h->nDoaDtrBins,     0.0f);
        std::fill_n(&h->energyDataAvg[0],         h->nBands,          0.0f);
    };

    auto normalize_average_spatial_parameters = [&]() {
        auto const& div = (float)h->nFrames;
        utility_svsdiv(&h->intensityVectorAvg[0][0], &div, h->nDoaDtrBins * 3, &h->intensityVectorAvg[0][0]);
        utility_svsdiv(&h->energyDataAvg[0], &div, h->nBands, &h->energyDataAvg[0]);
    };

    auto calculate_average_doa_parameters_sphoa = [&](std::vector<float> &newDoaVector, const std::vector<float> &interpolatedListenerPos) {
        for (int band = 0; band < h->nDoaDtrBins; ++band) {
            calculate_new_doa_vector(h, h->intensityVectorAvg[band], newDoaVector, interpolatedListenerPos);
            calculate_sphoa_dir_gain(h, 0, band, newDoaVector);
            h->azimuthDataAvg[band]   = calculate_azimuth(newDoaVector);
            h->elevationDataAvg[band] = calculate_elevation(newDoaVector);
        }
    };

    auto calculate_average_doa_parameters = [&]() {
        for (int band = 0; band < h->nDoaDtrBins; ++band) {
            h->azimuthDataAvg[band]   = calculate_azimuth(h->intensityVectorAvg[band]);
            h->elevationDataAvg[band] = calculate_elevation(h->intensityVectorAvg[band]);
        }
    };

    auto calculate_average_dtr = [&]() {
        for (int band = 0; band < h->nDoaDtrBins; ++band) {
            h->DTRDataAvg[band] = calculate_dtr(h->intensityVectorAvg[band], h->energyDataAvg[band]);
        }
    };

    auto copy_average_parameters_to_sub_frames = [&](int mic) {
        std::copy_n(&h->azimuthDataAvg[0],   h->nBands, &h->azimuthData[mic][0][0]);
        std::copy_n(&h->elevationDataAvg[0], h->nBands, &h->elevationData[mic][0][0]);
        std::copy_n(&h->DTRDataAvg[0],       h->nBands, &h->DTRData[mic][0][0]);
        for (int frame = 0; frame < h->nFrames; ++frame) {
            std::copy_n(&h->energyDataAvg[0],    h->nBands, &h->energyData[mic][frame][0]);
            if (h->useSphoa6DofExtensionMode) {
                std::copy_n(&h->singlePoint6DofDirGains[0][0], h->nBands, &h->singlePoint6DofDirGains[frame][0]);
            }
        }
    };

    auto doa_and_dtr_analysis = [&](int mic, int frame, int band, std::vector<float> &intensityVector, std::vector<float> &newDoaVector, std::vector<float> &interpolatedListenerPos) {
        // Calculate intensity vector (convert ACN channel order to cartesian x,y,z)
        intensityVector[0] = real(h->tempAnalysisVec[0] * std::conj(h->tempAnalysisVec[3]));
        intensityVector[1] = real(h->tempAnalysisVec[0] * std::conj(h->tempAnalysisVec[1]));
        intensityVector[2] = real(h->tempAnalysisVec[0] * std::conj(h->tempAnalysisVec[2]));

        if (h->useRecursiveAveraging) {
            for (int i = 0; i < intensityVector.size(); ++i) {
                h->intensityVectorAvg[band][i] += intensityVector[i];
            }
        } else if (h->useSphoa6DofExtensionMode) {
            // Calculate DOA angles from a new DOA vector pointing from the listener
            // to a projected source position. DTR and energy are calculated from the
            // original intensity vector.
            calculate_new_doa_vector(h, intensityVector.data(), newDoaVector, interpolatedListenerPos);
            calculate_sphoa_dir_gain(h, frame, band, newDoaVector);
            calculate_doa_parameters(mic, frame, band, newDoaVector);
            h->DTRData[mic][frame][band] = calculate_dtr(intensityVector, h->energyData[mic][frame][band]);
        } else {
            // Calculate DOA angles, DTR, and energy from the intensity vector
            calculate_doa_parameters(mic, frame, band, intensityVector);
            h->DTRData[mic][frame][band] = calculate_dtr(intensityVector, h->energyData[mic][frame][band]);
        }
    };

    std::vector<float> sectorBeamformingCoefficients{1.0f, 0.5774f, 0.5774f, 0.5774f};
    std::vector<float> intensityVector(3, 0.0f);
    std::vector<float> newDoaVector(3, 0.0f);
    std::vector<float> interpolatedListenerPos(3, 0.0f);
    float energy = 0.0f;
    float tempEnergy = 0.0f;
    float_complex ****TFDInputFrame;
    for (int mic : h->activeArraysForCurrentFrame) {
        if (std::find(h->transformedArraysForCurrentFrameFullOrder.begin(), h->transformedArraysForCurrentFrameFullOrder.end(), mic) != h->transformedArraysForCurrentFrameFullOrder.end()) {
            if (h->isCrossfadeInProgress && mic==h->currentClosestMicrophoneArray) {
                // this is the mic we are crossfading to, FullOrder afSTFT is not ready yet
                TFDInputFrame = h->TFDInputFrameFOA;
            } else {
                TFDInputFrame = h->TFDInputFrameFullOrder;
            }
        } else {
            TFDInputFrame = h->TFDInputFrameFOA;
        }
        if (h->useRecursiveAveraging) {
            reset_average_estimates();
        }
        for (int frame = 0; frame < h->nFrames; ++frame) {

            if (h->useSphoa6DofExtensionMode) {
                get_interpolated_listener_position_in_array_space(h, mic, frame, interpolatedListenerPos);
            }

            for (int band = 0; band < h->nBands; ++band) {
                for (int ch = 0; ch < h->nAnalysisChannels; ++ch) {
                    h->tempAnalysisVec[ch] = TFDInputFrame[mic][frame][ch][band] *
                                                   sectorBeamformingCoefficients[ch];
                }

                // Calculate energy
                tempEnergy = 0.0f;
                for (int ch = 0; ch < h->nAnalysisChannels; ++ch) {
                    tempEnergy += real(h->tempAnalysisVec[ch] * std::conj(h->tempAnalysisVec[ch]));
                }
                energy = 0.5f * tempEnergy;
                if (h->useRecursiveAveraging) {
                    h->energyDataAvg[band] += energy;
                } else {
                    h->energyData[mic][frame][band] = energy;
                }

                // Calculate DOA and DTR for frequency indexes below maxFreqLimitIndex
                if (band < h->maxFreqLimitIndex) {
                    doa_and_dtr_analysis(mic, frame, band, intensityVector, newDoaVector, interpolatedListenerPos);
                }
            }
        }
        if (h->useRecursiveAveraging) {
            // In recursive averaging mode, spatial parameters are obtained from an average intensity vector and energy,
            // which are averaged from the whole processing frame. The same spatial values and energies are used in
            // each sub-frame of the processing frame.
            normalize_average_spatial_parameters();
            if (h->useSphoa6DofExtensionMode) {
                calculate_average_doa_parameters_sphoa(newDoaVector, interpolatedListenerPos);
            } else {
                calculate_average_doa_parameters();
            }
            calculate_average_dtr();
            copy_average_parameters_to_sub_frames(mic);
        }
    }
}
