/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_SPATIAL_ANALYSIS_HPP
#define HOA_TRANSLATION_SPATIAL_ANALYSIS_HPP

#include "Utilities.hpp"

// Spatial metadata calculations, input vector xyzVec as: x (front), y (left), z (up)
// azimuth = atan2(y, x)
float calculate_azimuth(const float* xyzVec);
float calculate_azimuth(const std::vector<float> &xyzVec);

// elevation = atan2(z, sqrt(x^2 + y^2))
float calculate_elevation(const float* xyxVec);
float calculate_elevation(const std::vector<float> &xyxVec);

// DTR = sqrt(x^2 + y^2 + z^2) / energy
float calculate_dtr(const float* xyzVec, float energy);
float calculate_dtr(const std::vector<float> &xyzVec, float energy);

// DTR = sqrt(x^2 + y^2 + z^2)
float calculate_dtr_without_energy(const float* xyzVec);

// Calculate azimuth/elevation/DTR for multiple values. xVec, yVec and zVec pointers contain vecLen amount of
// data as: x (front), y (left), z (up). The calculated output values are stored in
// azi, elev and dtr pointers.
void calculate_azimuth(const float* xVec, const float* yVec, float* azi, int vecLen);
void calculate_elevation(float* xVec, float* yVec, const float* zVec, float* xVecPow, float* yVecPow, float* elev, int vecLen);
void calculate_dtr(float* xVec, float* yVec, float* zVec, float* xVecPow, float* yVecPow, float* zVecPow, const float* energy, float* dtr, int vecLen);

// Spatial metadata calculations, input vector xyzVec in OpenGL as: x (right), y (up), z (back)
// azimuth = atan2(-x, -z)
float calculate_azimuth_OpenGL(const float* xyzVec);
std::vector<float> calculate_azimuths_OpenGL(std::vector<float> &xVec, std::vector<float> &zVec);
void calculate_azimuths_OpenGL(hoaTranslatorData *h, float *xVec, float *zVec, float *aziVec, int vecLen);

// elevation = atan2(y, sqrt(x^2 + z^2))
float calculate_elevation_OpenGL(const float* xyxVec);
std::vector<float> calculate_elevations_OpenGL(std::vector<float> &xVec, std::vector<float> &yVec, std::vector<float> &zVec);
void calculate_elevations_OpenGL(hoaTranslatorData *h, float *xVec, float *yVec, float *zVec, float *elevVec, int vecLen);

// Calculate Direction-of-Arrival (DOA) vector from azimuth and elevation angles, DOA is given as: x (front), y (left), z (up)
void calculate_doa_vector_cartesian(float azimuthRad, float elevationRad, std::vector<float> &doa);

// Calculate Direction-of-Arrival (DOA) vector from azimuth and elevation angles, DOA is given as: x (right), y (up), z (back)
void calculate_doa_vector_OpenGL(float azimuthRad, float elevationRad, std::vector<float> &doa);

// Calculate multiple DOA vectors at once, DOA is given as: x (front), y (left), z (up)
void calculate_doa_vectors_cartesian(hoaTranslatorData *h,
                                     int vecLen,
                                     const float *azimuthRads,
                                     const float *elevationRads,
                                     float *doasX,
                                     float *doasY,
                                     float *doasZ);

// Calculate multiple DOA vectors at once, DOA is given as: x (right), y (up), z (back)
void calculate_doa_vectors_OpenGL(hoaTranslatorData *h,
                                  int vecLen,
                                  const float *azimuthRads,
                                  const float *elevationRads,
                                  float *doasX,
                                  float *doasY,
                                  float *doasZ);

// Calculate sine and cosine values for each sourceData element.
void get_sin_cos(float *sinData, float *cosData, const float *sourceData, int vecLen);

// Interpolated listener position in mic array coordinate space (as x (front), y (left), z (up),
// and (0,0,0) on top of the mic array).
void get_interpolated_listener_position_in_array_space(hoaTranslatorData *h, int mic, int frame, std::vector<float> &interpolatedListenerPos);

// Calculate a new DOA vector, that points from the listener to a projected source position
void calculate_new_doa_vector(hoaTranslatorData *h, const float* intensityVector,
                              std::vector<float> &newDoaVector, const std::vector<float> &interpolatedListenerPos);

// Calculate direct sound gain for the new DOA vector
void calculate_sphoa_dir_gain(hoaTranslatorData *h, int frame, int band, const std::vector<float> &newDoaVector);

/* Calculate Direction-of-arrival (DOA) angles, Direct-to-total sound ratio (DTR), and energy for the input signals.
 * The analysis is based on signal covariance matrix, that is calculated from the input time-frequency domain signal as:
 *
 *      covariance_matrix = input_signal * input_signal^H
 *
 * The input signal is in Ambisonics Channel Order (ACN), where the first 4 channels are:
 *      0 - omni,  W
 *      1 - left,  Y
 *      2 - up,    Z
 *      3 - front, X
 *
 * An intensity vector is obtained from the covariance matrix as (in cartesian order (x front, y left, z up)):
 *
 *      x = covariance_matrix(0, 3)
 *      y = covariance_matrix(0, 1)
 *      z = covariance_matrix(0, 2)
 *
 * The energy is obtained as the diagonal sum of the covariance matrix. The energy and intensity vector are further used
 * to obtain the DOA angles and DTR value for the mic array signal.
 *
 * NOTE: if recursive averaging mode is used, the spatial parameters are obtained from an average intensity vector and energy.
 * This means that instead of calculating spatial values for each sub-frame, the values are average values for the whole
 * processing frame.
 *
 * Output:
 *      azimuthData      - the azimuth DOA angles in radians
 *      elevationData    - the elevation DOA angles in radians
 *      DTRData          - the direct-to-total (DTR) energy ratios [0, 1]
 *      energyData       - the signal energies
 */
void spatial_analysis(hoaTranslatorData *h);

#endif //HOA_TRANSLATION_SPATIAL_ANALYSIS_HPP
