/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "SynthesisProcessor.hpp"

#ifndef SAF_ENABLE_SOFA_READER
#define SAF_ENABLE_SOFA_READER
#endif

#include <algorithm>
#include <cmath>
#include <complex>
#include <vector>

#include "OptimalMixing.hpp"
#include "Rotator.hpp"
#include "SourceAware.hpp"

const float __binauralEQ[133] =
        {0.9260,0.9125,0.9081,0.9125,0.9353,0.9539,0.9851,0.9864,1.0052,0.9861,0.8966,0.8599,0.8723,0.9705,1.1584,1.3672,1.4724,1.4113,1.3476,1.2923,1.2496,1.2379,1.2737,1.3830,1.5525,1.7644,1.8759,2.0243,2.1751,2.1952,2.0104,1.8608,1.7755,1.6988,1.5921,1.5094,1.4751,1.3455,1.2756,1.2593,1.1974,1.2166,1.2831,1.2675,1.2008,1.1575,1.1915,1.0990,1.1590,1.3169,1.5064,1.6291,1.5511,1.4354,1.2503,1.0908,1.0447,1.0882,1.0956,1.1305,1.2584,1.2681,1.3551,1.5122,1.5078,1.4253,1.5348,1.5593,1.5143,1.5152,1.6453,1.6941,1.6581,1.7024,1.4535,1.3868,1.3063,1.2892,1.4462,1.3555,1.3570,1.3459,1.3748,1.3809,1.4103,1.5158,1.5419,1.4814,1.4838,1.4973,1.5224,1.5630,1.5617,1.5790,1.6489,1.7983,1.9983,2.2378,2.2779,2.3001,1.8893,1.8594,1.5237,1.3058,1.2100,1.1955,1.2367,1.1854,1.2017,1.4122,1.6408,1.7691,1.5915,1.5407,1.5362,1.5532,1.5277,1.4606,1.5091,1.5231,1.6351,1.8617,1.9810,1.9883,1.8054,1.6002,1.5267,1.4320,1.3283,1.1892,1.0683,1.0110,0.9367};


void combine_bands(hoaTranslatorData *h, hoatMatrixCmplx3D &Cmatrix) {
    for (int band = 0; band < h->nComplexityReducedBands; band++) {
        int bottomBandIndx = h->bottomBandIndexes[band];
        int upperBandIndx = h->upperBandIndexes[band];
        int nCompressedBands = upperBandIndx - bottomBandIndx + 1;
        fillWithZeros(h->tempMatrixOutChSqrd);

        for (int i = bottomBandIndx; i <= upperBandIndx; ++i) {
            matrixMatrixAdd(h->tempMatrixOutChSqrd, Cmatrix[i], h->tempMatrixOutChSqrd);
        }
        matrixScalarMult(1.0f / (float)nCompressedBands, h->tempMatrixOutChSqrd, h->tempMatrixOutChSqrd);

        for (int i = bottomBandIndx; i <= upperBandIndx; ++i) {
            copy(h->tempMatrixOutChSqrd, Cmatrix[i]);
        }
    }
}

/* During crossfade, the interpolatedTFDFrame is formed from signals from two different mic arrays
 * (current closest and previously closest arrays). These arrays might have different orientations
 * in the scene, which means that they both need their own SHrotationMatrix. Here prototype signals
 * are created similarly as in generate_prototype_signals(), except the SHrotationMatrix is chosen
 * appropriately for each frequency band.
 */
void generate_prototype_signals_during_crossfade(hoaTranslatorData *h)
{
    auto choose_sh_rotation_matrix = [&](int band, int frame, hoatMatrix2D *&SHrotationMatrix) {
        SHrotationMatrix = &h->currentSHrotationMatrix;
        if (std::find(h->fadeOutBands[frame].begin(), h->fadeOutBands[frame].end(), band) != h->fadeOutBands[frame].end()) {
            SHrotationMatrix = &h->previousSHrotationMatrix;
        }
    };

    auto copy_input_to_matrix = [&](int band, int frame) {
        for (int ch = 0; ch < h->channelsPerArray; ++ch) {
            h->tempInputMatrixSmallReal[ch][0] = h->interpolatedTFDFrame[frame][ch][band].real();
            h->tempInputMatrixSmallImag[ch][0] = h->interpolatedTFDFrame[frame][ch][band].imag();
        }
    };

    auto copy_temp_to_binaural_matrix = [&](int band, int frame) {
        for (int ch = 0; ch < h->outChannels; ++ch) {
            h->currentBinauralMatrix[band][ch][frame] = h->tempBinauralMatrix[ch][0];
        }
    };

    hoatMatrix2D *SHrotationMatrix{};
    for (int band = 0; band < h->nBands; ++band) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            // Use rotation matrix either for the closest or previously closest array, depending on the stage of the crossfade
            choose_sh_rotation_matrix(band, frame, SHrotationMatrix);
            // Store input data into a matrix that is compatible with cblas routines
            copy_input_to_matrix(band, frame);
            // Apply rotation
            matrixMult(*SHrotationMatrix, h->tempInputMatrixSmallReal, h->tempRotatedMatrixSmallReal);
            matrixMult(*SHrotationMatrix, h->tempInputMatrixSmallImag, h->tempRotatedMatrixSmallImag);
            combineIntoCmplxMatrix(h->tempRotatedMatrixSmallReal, h->tempRotatedMatrixSmallImag, h->tempRotatedMatrixSmall);
            // Ambisonics to binaural decoding
            matrixMult(h->ambi2binDecodingMatrix[band], h->tempRotatedMatrixSmall, h->tempBinauralMatrix);
            // Copy temp binaural matrix from this frame to the binaural matrix
            copy_temp_to_binaural_matrix(band, frame);
        }
    }
}

void generate_prototype_signals(hoaTranslatorData *h)
{
    auto copy_input_to_matrix = [&](int band) {
        for (int ch = 0; ch < h->channelsPerArray; ++ch) {
            for (int frame = 0; frame < h->nFrames; ++frame) {
                h->tempInputMatrixReal[ch][frame] = h->interpolatedTFDFrame[frame][ch][band].real();
                h->tempInputMatrixImag[ch][frame] = h->interpolatedTFDFrame[frame][ch][band].imag();
            }
        }
    };

    // Get spherical harmonics rotation matrix, that is used to rotate the STFT signals
    get_sh_rotation_matrix(h);

    if (h->isCrossfadeInProgress) {
        generate_prototype_signals_during_crossfade(h);
        return;
    }

    for (int band = 0; band < h->nBands; ++band) {
        // Store input data into a matrix that is compatible with cblas routines
        copy_input_to_matrix(band);
        // Apply rotation
        matrixMult(h->currentSHrotationMatrix, h->tempInputMatrixReal, h->tempRotatedMatrixReal);
        matrixMult(h->currentSHrotationMatrix, h->tempInputMatrixImag, h->tempRotatedMatrixImag);
        combineIntoCmplxMatrix(h->tempRotatedMatrixReal, h->tempRotatedMatrixImag, h->tempRotatedMatrix);
        // Ambisonics to binaural decoding
        matrixMult(h->ambi2binDecodingMatrix[band], h->tempRotatedMatrix, h->currentBinauralMatrix[band]);
    }
}

void estimate_target_covariance_matrix(hoaTranslatorData *h)
{
    for(int band = 0; band < h->maxFreqLimitIndex; ++band) {
        // Compute the new input covariance matrix newCxMatrix
        matrixMultNoTransConj(h->currentBinauralMatrix[band], h->currentBinauralMatrix[band], h->newCxMatrix[band]);
        // Compute the estimated Cx, by combining the newCxMatrix with Cx data from previous frames
        combine_covariance_matrices(h, h->Cx[band], h->previousCxMatrix[band], h->newCxMatrix[band]);
        // Use the diagonal values from Cx as linear decoding Cy
        copyDiagonal(h->Cx[band], h->CyLinearDecoding[band]);
    }

    // For high frequencies, calculate only the diagonal
    hoatMatrixCmplx2D *binM{};
    for(int band = h->maxFreqLimitIndex; band < h->nBands; ++band) {
        binM = &h->currentBinauralMatrix[band];
        h->newCxMatrix[band][0][0] = ((*binM)[0][0] * std::conj((*binM)[0][0])) + ((*binM)[0][1] * std::conj((*binM)[0][1]));
        h->newCxMatrix[band][1][1] = ((*binM)[1][1] * std::conj((*binM)[1][1])) + ((*binM)[1][0] * std::conj((*binM)[1][0]));
        if (h->useRecursiveAveraging) {
            h->newCxMatrix[band][0][0] *= (1.0f - h->recursionSmoothingCoefficient);
            h->newCxMatrix[band][1][1] *= (1.0f - h->recursionSmoothingCoefficient);
            h->Cx[band][0][0] = h->Cx[band][0][0] * h->recursionSmoothingCoefficient + h->newCxMatrix[band][0][0];
            h->Cx[band][1][1] = h->Cx[band][1][1] * h->recursionSmoothingCoefficient + h->newCxMatrix[band][1][1];
        } else {
            h->Cx[band][0][0] = h->previousCxMatrix[band][0][0] + h->newCxMatrix[band][0][0];
            h->Cx[band][1][1] = h->previousCxMatrix[band][1][1] + h->newCxMatrix[band][1][1];
        }
        copyDiagonal(h->Cx[band], h->CyLinearDecoding[band]);
    }
}

void estimate_diffuse_field_covariance_matrix(hoaTranslatorData *h)
{
    float normalizer = 1.0f / float(h->nInterpolatedHrtfDirs);
    for (int band = 0; band < h->nBands; ++band) {
        for (int dirIndex = 0; dirIndex < h->nInterpolatedHrtfDirs; ++dirIndex) {
            copy_hrtf_to_matrix(h, band, h->hrtfIdxGrid[dirIndex]);
            matrixMultNoTransConj(h->tempHrtfMatrix, h->tempHrtfMatrix, h->tempHrtfProductMatrix);
            matrixScalarMult(normalizer, h->tempHrtfProductMatrix, h->tempHrtfProductMatrix);
            matrixMatrixAdd(h->Cdiffuse[band], h->tempHrtfProductMatrix, h->Cdiffuse[band]);
        }
        float oneOverTrace = 1.0f / sumDiagonalReal(h->Cdiffuse[band]);
        matrixScalarMult(oneOverTrace, h->Cdiffuse[band], h->Cdiffuse[band]);
    }
}

void determine_target_covariance_matrix(hoaTranslatorData *h)
{
    auto getHrtfDirIndex = [&](int frame, int band) {
        // The calculation of dirIndex assumes that the interpolatedHrtfFB values are stored in order of azi: -180...180, elev: -90...90
        int aziIndex = std::round(matlab_fmodf(h->interpolatedAzimuthData[frame][band] * 180.0f / M_PI + 180.0f, 360.0f) / h->interpolatedHrtfAziResDeg);
        int elevIndex = std::round(matlab_fmodf(h->interpolatedElevationData[frame][band] * 180.0f / M_PI + 90.0f, 180.0f) / h->interpolatedHrtfElevResDeg);
        int dirIndex = elevIndex * h->nInterpolatedHrtfAziAngles + aziIndex;  // [(azi0 elev0) (azi1 elev0)...]
        return h->hrtfIdxGrid[dirIndex];
    };

    auto computeDirectCy = [&](int frame, int band, float energy, float dtr) {
        // directPart = dtr * signalEnergy * (Hrtf * Hrtf^H)
        float multiplier = dtr * energy;
        if (h->useSphoa6DofExtensionMode) {
            multiplier *= std::pow(h->singlePoint6DofDirGains[frame][band], 2.0f);
        }
        // For high frequencies, calculate only the diagonal
        if (band < h->maxFreqLimitIndex) {
            matrixScalarMult(multiplier, h->tempHrtfProductMatrix, h->tempHrtfProductMatrix);
            matrixMatrixAdd(h->newCyMatrix[band], h->tempHrtfProductMatrix, h->newCyMatrix[band]);
        } else {
            h->newCyMatrix[band][0][0] += multiplier * h->tempHrtfProductMatrix[0][0];
            h->newCyMatrix[band][1][1] += multiplier * h->tempHrtfProductMatrix[1][1];
        }
    };

    auto computeDiffuseCyExterior = [&](int frame, int band, float energy, float dtr) {
        // Exterior rendering as: diffusePart = (1 - dtr) * signalEnergy * weightedDiffuseCovarianceMatrix,
        // where weightedDiffusCovarianceeMatrix = (w3 * DCMdiff) + (w4 * DCMambi).
        // DCMdiff is computed from the diffuse covariance matrix, Cdiffuse.
        // DCMambi is computed from the prototype binaural signal.

        // Compute weighted DCMdiff
        float w3 = h->diffuseExteriorWeight;
        float CdiffuseTrace = sumDiagonalReal(h->Cdiffuse[band]);
        float mult = w3 / (CdiffuseTrace * h->nFrames);
        if (band < h->maxFreqLimitIndex) {
            matrixScalarMult(mult, h->Cdiffuse[band], h->DCMdiff);
        } else {
            h->DCMdiff[0][0] = mult * h->Cdiffuse[band][0][0];
            h->DCMdiff[1][1] = mult * h->Cdiffuse[band][1][1];
        }

        // Compute weighted DCMambi
        float w4 = 1.0f - w3;
        float protoEnergyCh0 = std::norm(h->currentBinauralMatrix[band][0][frame]);
        float protoEnergyCh1 = std::norm(h->currentBinauralMatrix[band][1][frame]);
        float chEnergyBoth = protoEnergyCh0 + protoEnergyCh1 + FLT_MIN;
        h->DCMambi[0][0] = (w4 * protoEnergyCh0) / chEnergyBoth;
        h->DCMambi[1][1] = (w4 * protoEnergyCh1) / chEnergyBoth;

        // Form DCM from the diffuse signal energy, DCMdiff and DCMambi
        if (band < h->maxFreqLimitIndex) {
            matrixMatrixAdd(h->DCMdiff, h->DCMambi, h->DCM);
        } else {
            h->DCM[0][0] = h->DCMdiff[0][0] + h->DCMambi[0][0];
            h->DCM[1][1] = h->DCMdiff[1][1] + h->DCMambi[1][1];
        }
        float DCMtrace = sumDiagonalReal(h->DCM);
        mult = ((1.0f - dtr) * energy) / DCMtrace;
        if (band < h->maxFreqLimitIndex) {
            matrixScalarMult(mult, h->DCM, h->DCM);
        } else {
            h->DCM[0][0] *= (mult * h->DCM[0][0]);
            h->DCM[1][1] *= (mult * h->DCM[1][1]);
        }

        // Add DCM to newCy
        if (band < h->maxFreqLimitIndex) {
            matrixMatrixAdd(h->newCyMatrix[band], h->DCM, h->newCyMatrix[band]);
        } else {
            h->newCyMatrix[band][0][0] += h->DCM[0][0];
            h->newCyMatrix[band][1][1] += h->DCM[1][1];
        }
    };

    auto computeDiffuseCyInterior = [&](int frame, int band, float energy, float dtr) {
        float mult = (1.0f - dtr) * energy;
        if (h->diffuseMethodForTargetSCM == DIFFUSE_COVARIANCE_MATRIX) {
            // diffusePart = (1 - dtr) * signalEnergy * diffuseMatrix
            if (band < h->maxFreqLimitIndex) {
                matrixScalarMult(((1.0f - dtr) * energy), h->Cdiffuse[band], h->tempMatrixOutChSqrd);
                matrixMatrixAdd(h->newCyMatrix[band], h->tempMatrixOutChSqrd, h->newCyMatrix[band]);
            } else {
                h->newCyMatrix[band][0][0] += (mult * h->Cdiffuse[band][0][0]);
                h->newCyMatrix[band][1][1] += (mult * h->Cdiffuse[band][1][1]);
            }
        } else {
            // diffusePart = (1 - dtr) * signalEnergy * prototypeEnergySingleChannel / prototypeEnergyBothChannels
            float protoEnergyCh0 = std::norm(h->currentBinauralMatrix[band][0][frame]);
            float protoEnergyCh1 = std::norm(h->currentBinauralMatrix[band][1][frame]);
            float chEnergyBoth = protoEnergyCh0 + protoEnergyCh1 + FLT_MIN;
            h->newCyMatrix[band][0][0] += (mult * protoEnergyCh0 / chEnergyBoth);
            h->newCyMatrix[band][1][1] += (mult * protoEnergyCh1 / chEnergyBoth);
        }
    };

    auto computeDiffuseCy = [&](int frame, int band, float energy, float dtr) {
        if (!h->useSphoa6DofExtensionMode && !h->isListenerInsideTriangle && h->useImprovedExteriorRendering) {
            computeDiffuseCyExterior(frame, band, energy, dtr);
        } else {
            computeDiffuseCyInterior(frame, band, energy, dtr);
        }
    };

    // Compute newCy
    int hrtfDirIndex = 0;
    float energy = 0.0f;
    float dtr = 0.0f;
    fillWithZeros(h->newCyMatrix);
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int band = 0; band < h->nBands; ++band) {
            if (band <= h->maxFreqLimitIndex) {
                dtr = h->interpolatedDTRData[frame][band];
                hrtfDirIndex = getHrtfDirIndex(frame, band);
                copy_hrtf_to_matrix(h, band, hrtfDirIndex);
                matrixMultNoTransConj(h->tempHrtfMatrix, h->tempHrtfMatrix, h->tempHrtfProductMatrix);
            }
            energy = h->interpolatedEnergyData[frame][band];
            computeDirectCy(frame, band, energy, dtr);
            computeDiffuseCy(frame, band, energy, dtr);
        }
    }
    // Compute the target Cy
    for (int band = 0; band < h->nBands; ++band) {
        combine_covariance_matrices(h, h->Cy[band], h->previousCyMatrix[band], h->newCyMatrix[band]);
    }
}

void determine_mixing_rule(hoaTranslatorData *h)
{
    auto determine_computed_mixing_matrices = [&](int band, bool &useOptimalMixingForM, bool &useOptimalMixingForMr, bool &useEnergyFLAG) {
        if (h->freqVec[band] < h->lowerBandFreqLimit) {
            if (h->useDiffuseDecorrelation) {
                useOptimalMixingForM = true;
                useOptimalMixingForMr = true;
                useEnergyFLAG = false;
            } else {
                useOptimalMixingForM = true;
                useOptimalMixingForMr = false;
                useEnergyFLAG = true;
            }
        } else if (band < h->maxFreqLimitIndex) {
            useOptimalMixingForM = true;
            useOptimalMixingForMr = false;
            useEnergyFLAG = true;
        } else {
            useOptimalMixingForM = false;
            useOptimalMixingForMr = false;
        }
    };

    auto compute_mixing_matrix = [&](int band, bool useOptimalMixingForM, bool useEnergyFLAG) {
        if (useOptimalMixingForM) {
            formulate_M_and_Cr_2x2(h,
                                   h->Cx[band],
                                   h->Cy[band],
                                   h->Q,
                                   useEnergyFLAG,
                                   0.2f,
                                   h->currentM[band],
                                   h->Cr,
                                   !useEnergyFLAG);
        } else {
            fillWithZeros(h->currentM[band]);
            for (int i = 0; i < h->currentM[band].nRows; ++i) {
                h->currentM[band][i][i] = std::sqrt(h->Cy[band][i][i] / (float_complex(FLT_MIN, FLT_MIN) + h->CyLinearDecoding[band][i][i]));
            }
        }
    };

    auto compute_residual_mixing_matrix = [&](int band, bool useOptimalMixingForMr) {
        if (useOptimalMixingForMr) {
            formulate_M_and_Cr_2x2(h,
                                   h->CyLinearDecoding[band],
                                   h->Cr,
                                   h->Q,
                                   false,
                                   0.2f,
                                   h->currentMr[band],
                                   h->Cr,
                                   false);
        } else {
            fillWithZeros(h->currentMr[band]);
        }
    };

    bool useEnergyFLAG = false;
    bool useOptimalMixingForM = true;
    bool useOptimalMixingForMr = true;

    // Calculate mixing matrices only once per band
    if (h->useComplexityReduction) {
        for (int band = 0; band < h->nComplexityReducedBands; band++) {
            int bottomBandIndx = h->bottomBandIndexes[band];
            int upperBandIndx = h->upperBandIndexes[band];

            // Check which mixing matrices to compute
            determine_computed_mixing_matrices(bottomBandIndx, useOptimalMixingForM, useOptimalMixingForMr, useEnergyFLAG);

            // Compute M (and Cr)
            compute_mixing_matrix(bottomBandIndx, useOptimalMixingForM, useEnergyFLAG);

            // Compute Mr
            compute_residual_mixing_matrix(bottomBandIndx, useOptimalMixingForMr);

            // Copy to frequency bins
            for (int i = bottomBandIndx + 1; i <= upperBandIndx; ++i) {
                copy(h->currentM[bottomBandIndx], h->currentM[i]);
                copy(h->currentMr[bottomBandIndx], h->currentMr[i]);
            }
        }

        // Apply equalization
        if (h->useEqualizer) {
			for (int band = 0; band < h->nBands; ++band) {
                matrixScalarMult(__binauralEQ[band], h->currentM[band], h->currentM[band]);
                matrixScalarMult(__binauralEQ[band], h->currentMr[band], h->currentMr[band]);
            }
        }

    // Original processing, one mixing matrix per bin
    } else {
        for (int band = 0; band < h->nBands; ++band) {
            // Check which mixing matrices to compute
            determine_computed_mixing_matrices(band, useOptimalMixingForM, useOptimalMixingForMr, useEnergyFLAG);

            // Compute M (and Cr)
            compute_mixing_matrix(band, useOptimalMixingForM, useEnergyFLAG);

            // Compute Mr
            compute_residual_mixing_matrix(band, useOptimalMixingForMr);

            // Apply equalization
            if (h->useEqualizer) {
                matrixScalarMult(__binauralEQ[band], h->currentM[band], h->currentM[band]);
                matrixScalarMult(__binauralEQ[band], h->currentMr[band], h->currentMr[band]);
            }
        }
    }
}

void compute_decorrelated_frame(hoaTranslatorData *h)
{
    // Fetch a delayed prototype signal from inputXBlocksPrototype buffer, and copy the current prototype signal
    // to the end of the buffer.
    int delay = 0;
    int nBufferFrames = (h->nBlocksInDecorrelationBuffer - 1) * h->nFrames;
    for (int band = 0; band < h->nBands; ++band) {
        for (int ch = 0; ch < h->outChannels; ++ch) {
            delay = h->decorrelationDelays[band * h->outChannels + ch];
            auto const& xBlock = h->inputXBlocksPrototype[band][ch];

            // Fetch a delayed frame from the beginning of inputXBlocksPrototype
            std::copy_n(&xBlock[nBufferFrames - delay], h->nFrames, &h->decorrelatedTFDFrame[band][ch][0]);

            // Shift inputXBlocksPrototype by one whole processing frame frame
            std::copy_n(&xBlock[h->nFrames], nBufferFrames, &xBlock[0]);

            // Copy the input (prototype signal) to the end of inputXBlocksPrototype
            std::copy_n(&h->currentBinauralMatrix[band][ch][0], h->nFrames, &xBlock[nBufferFrames]);
        }
    }

    // Onset removal for the current prototype signal
    float alpha = 0.95f;          // default 0.95
    float beta = 0.995f;          // default 0.995
    float detectorEnergy = 0.0f;
    float transientEq = 0.0f;
    for (int band = 0; band < h->nBands; ++band) {
        for (int ch = 0; ch < h->outChannels; ++ch) {
            for (int frame = 0; frame < h->nFrames; ++frame) {
                detectorEnergy = std::norm(h->inputXBlocksPrototype[band][ch][nBufferFrames + frame]);
                h->transientDetector1[band][ch] *= alpha;
                if (h->transientDetector1[band][ch] < detectorEnergy) {
                    h->transientDetector1[band][ch] = detectorEnergy;
                }
                h->transientDetector2[band][ch] = (h->transientDetector2[band][ch] * beta) +
                                                  (h->transientDetector1[band][ch] * (1 - beta));
                if (h->transientDetector2[band][ch] > h->transientDetector1[band][ch]) {
                    h->transientDetector2[band][ch] = h->transientDetector1[band][ch];
                }
                transientEq = (std::min)(1.0f, 4.0f * h->transientDetector2[band][ch] / (FLT_MIN + h->transientDetector1[band][ch]));
                h->inputXBlocksPrototype[band][ch][nBufferFrames + frame] *= transientEq;
            }
        }
    }
}

void mix_signals(hoaTranslatorData *h)
{
    auto interpolate_mixing_matrix = [&](int band, int frame, const hoatMatrixCmplx3D &currentMixingMatrix, const hoatMatrixCmplx3D &previousMixingMatrix) {
        // Linear interpolation between mixing matrices of this and previous frame
        matrixScalarMult(h->interpolator[frame], currentMixingMatrix[band], h->tempMixingMatrix);
        matrixScalarMult((1.0f - h->interpolator[frame]), previousMixingMatrix[band], h->tempMatrixOutChSqrd);
        matrixMatrixAdd(h->tempMatrixOutChSqrd, h->tempMixingMatrix, h->tempMixingMatrix);
    };

    auto copy_to_temp_matrix = [&](int frame, float_complex **source) {
        for (int ch = 0; ch < h->tempSignalMatrix.nRows; ++ch) {
            h->tempSignalMatrix[ch][0] = source[ch][frame];
        }
    };

    auto copy_to_output = [&](int band, int frame, const hoatMatrixCmplx2D &source) {
        for (int ch = 0; ch < h->outChannels; ++ch) {
            h->outputTFDProcessingFrame[band][ch][frame] = source[ch][0];
        }
    };

    auto add_to_output = [&](int band, int frame, const hoatMatrixCmplx2D &source) {
        for (int ch = 0; ch < h->outChannels; ++ch) {
            h->outputTFDProcessingFrame[band][ch][frame] += source[ch][0];
        }
    };

    for (int band = 0; band < h->nBands; ++band) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            // Mix the prototype signal with the mixing matrix
            interpolate_mixing_matrix(band, frame, h->currentM, h->previousM);
            copy_to_temp_matrix(frame, h->previousBinauralMatrix[band].data);
            matrixMult(h->tempMixingMatrix, h->tempSignalMatrix, h->tempOutputMatrix);
            copy_to_output(band, frame, h->tempOutputMatrix);

            // Mix the decorrelated signal and add it to output
            if (h->useDiffuseDecorrelation) {
                interpolate_mixing_matrix(band, frame, h->currentMr, h->previousMr);
                copy_to_temp_matrix(frame, h->decorrelatedTFDFrame[band]);
                matrixMult(h->tempMixingMatrix, h->tempSignalMatrix, h->tempOutputMatrix);
                add_to_output(band, frame, h->tempOutputMatrix);
            }
        }
    }
}

void optimal_mixing(hoaTranslatorData *h)
{
    // Estimate target covariance matrix Cx
    estimate_target_covariance_matrix(h);

    // Determine target covariance matrix Cy
    if (h->useSourceAwareMode) {
        source_aware_processing(h);
    } else {
        determine_target_covariance_matrix(h);
    }

    // Combine matrices across frequency bands
    if (h->useComplexityReduction) {
        combine_bands(h, h->Cx);
        combine_bands(h, h->Cy);
        combine_bands(h, h->CyLinearDecoding);
    }

    // Compute mixing matrices M and Mr
    determine_mixing_rule(h);

    // Compute a decorrelated signal
    if (h->useDiffuseDecorrelation) {
        compute_decorrelated_frame(h);
    }

    // Mix prototype and decorrelated signal with M and Mr
    mix_signals(h);
}
