/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_SYNTHESISPROCESSOR_HPP
#define HOA_TRANSLATION_SYNTHESISPROCESSOR_HPP

#include "Utilities.hpp"

/* Generate prototype signals for the optimal mixing using Ambisonics to binaural decoding. The prototype signals are
 * computed from the interpolated STFT signals (interpolatedTFDFrame).
 *
 * Output:
 *      currentBinauralMatrix       - the binaural prototype signal in time-frequency domain (as hoatMatrixCmplx3D)
 *      currentTFDPrototypeFrame    - the binaural prototype signal in time-frequency domain (as float_complex ***)
 */
void generate_prototype_signals(hoaTranslatorData *h);

/* Estimate the target covariance matrix, Cx, for the optimal mixing. A new Cx matrix for the current frame is formed
 * from the covariance matrix of the prototype signal as:
 *
 *      newCx = prototypeSignal * prototypeSignal^H
 *
 * The matrix is then combined with a Cx matrix from the previous frame. The combination of the Cx matrices is done either
 * with a direct sum, or by recursively adding the current Cx to the previous Cx matrix.
 *
 * Output:
 *      Cx                  - estimate for the target covariance matrix
 *      CyLinearDecoding    - a matrix formed of the diagonal values of Cx
 */
void estimate_target_covariance_matrix(hoaTranslatorData *h);

/* Estimate a diffuse-field covariance matrix from an equally distributed set of HRTFs as:
 *
 *      Cdiffuse = (1 / nDirections) * sum(Hrtf * Hrtf^H)
 *
 * Output:
 *      Cdiffuse    - estimate for the diffuse field covariance matrix
 */
void estimate_diffuse_field_covariance_matrix(hoaTranslatorData *h);

/* Determine the target covariance matrix, Cy, based on the energy and DTR estimates.
 * The direct part of Cy is calculated with:
 *
 *      directPart = dtr * signalEnergy * (Hrtf * Hrtf^H)
 *
 * The diffuse part of Cy is calculated with:
 *
 *      if (diffuseMethodForTargetSCM == DIFFUSE_COVARIANCE_MATRIX):
 *          diffusePart = (1 - dtr) * signalEnergy * diffuseMatrix
 *
 *      if (diffuseMethodForTargetSCM == PROTOTYPE_ENERGY):
 *          diffusePart = (1 - dtr) * signalEnergy * prototypeEnergySingleChannel / prototypeEnergyBothChannels
 *
 * The matrix is then combined with a Cy matrix from the previous frame. The combination of the Cy matrices is done either
 * with a direct sum, or by recursively adding the current Cy to the previous Cy matrix.
 *
 * Output:
 *      Cy    - the target covariance matrix
 */
void determine_target_covariance_matrix(hoaTranslatorData *h);

/* Determine the optimal mixing matrix M, and the residual mixing matrix Mr.
 * The matrices M and Mr are calculated as:
 *
 *      if (frequencyBand < upperBandFreqLimit):
 *          M is calculated from Cx and Cy using optimal mixing
 *          Mr is calculated from diag(Cx) and Cr using optimal mixing (Cr is a residual output matrix from calculating M)
 *
 *      if (frequencyBand > upperBandFreqLimit):
 *          M = sqrt( diag(Cy) / diag(Cx) )
 *          Mr = 0
 *
 * Output:
 *      M     - mixing matrix for the prototype signal
 *      Mr    - mixing matrix for the decorrelated signal
 *
 * NOTE: Mr is only used, when useDiffuseDecorrelation = true
 */
void determine_mixing_rule(hoaTranslatorData *h);

/* Compute a decorrelated signal for diffuse decorrelation. The decorrelated signal is obtained, by removing transients
 * from a delayed prototype signal.
 *
 * Output:
 *      decorrelatedTFDFrame    - the decorrelated time-frequency signal
 */
void compute_decorrelated_frame(hoaTranslatorData *h);

/* Mix the prototype signals with the optimal mixing matrices, M and Mr, to get the output signals. The mixing matrices
 * are linearly interpolated between current and previous frames. The output signal is obtained as:
 *
 *      outputSignal = M * prototypeSignal
 *
 * If diffuse decorrelation is used, the decorrelated signal is added to the output as:
 *
 *      outputSignal += Mr * decorrelatedSignal
 *
 * Output:
 *      outputTFDProcessingFrame    - output signal in time-frequency domain
 */
void mix_signals(hoaTranslatorData *h);

/* Compute output time-frequency domain signal using optimal mixing. The stages in optimal
 * mixing include:
 *
 *      - Estimate target covariance matrix Cx
 *      - Determine target covariance matrix Cy
 *      - Compute mixing matrices M and Mr
 *      - Compute a decorrelated signal
 *      - Mix prototype and decorrelated signal with M and Mr
 *
 * Output:
 *      outputTFDProcessingFrame    - output signal in time-frequency domain
 */
void optimal_mixing(hoaTranslatorData *h);

#endif //HOA_TRANSLATION_SYNTHESISPROCESSOR_HPP
