/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Utilities.hpp"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>
#include <ios>
#include <numeric>

#include "Crossfade.hpp"
#include "DEMUX.hpp"
#include "Exterior.hpp"
#include "HrtfOperations.hpp"
#include "PositionPreProcessor.hpp"
#include "SourceAware.hpp"
#include "SpatialAnalysis.hpp"
#include "SynthesisProcessor.hpp"
#include "VectorOperations.hpp"

std::vector<int> get_closest_one_third_octave_band_indexes(hoaTranslatorData *h) {
    std::vector<float> thirdOctaveBandCenterFreqs{
            16.0f, 20.0f, 25.0f, 31.5f, 40.0f, 50.0f, 63.0f, 80.0f, 100.0f, 125.0f, 160.0f, 200.0f, 250.0f, 315.0f, 400.0f, 500.0f, 630.0f, 800.0f, 1000.0f, 1250.0f,
            1600.0f, 2000.0f, 2500.0f, 3150.0f, 4000.0f, 5000.0f, 6300.0f, 8000.0f, 10000.0f, 12500.0f, 16000.0f, 20000.0f
    };
    std::vector<int> thirdOctaveBandIndexes(thirdOctaveBandCenterFreqs.size(), 0);
    for (int i = 0; i < thirdOctaveBandIndexes.size(); ++i) {
        int closestIndex = 0;
        float smallestDiff = FLT_MAX;
        for (int band = 0; band < h->nBands; ++band) {
            float diff = std::abs(h->freqVec[band] - thirdOctaveBandCenterFreqs[i]);
            if (diff < smallestDiff) {
                smallestDiff = diff;
                closestIndex = band;
            }
        }
        thirdOctaveBandIndexes[i] = closestIndex;
    }
    return thirdOctaveBandIndexes;
}

/* Initialize band limits for processing as:
 *      - single bin bands: [0 ... highestSingleBinBandsIndex]
 *      - ERB bands with width intermediateBandsERBWidth: [(highestSingleBinBandsIndex + 1) ... lowestHighBandIndex]
 *      - single band: [(lowestHighBandIndex + 1) ... (nBands - 1)]
 */
void initialize_complexity_reduction(hoaTranslatorData *h, std::vector<int> const& complexityReductionParameters)
{
    assert(complexityReductionParameters.size() == 3 && "Too few/many complexity reduction parameters");
    int highestSingleBinBandsIndex = complexityReductionParameters[0];
    int lowestHighBandIndex = complexityReductionParameters[1];
    auto oneThirdOctaveBandIndexes = get_closest_one_third_octave_band_indexes(h);
    int intermediateBandsERBWidth = complexityReductionParameters[2];
    float maximumFrequencyLimit = h->freqVec[lowestHighBandIndex];
    std::vector<float> erbUpperFrequencies{};
    std::vector<int> erbUpperIndexes{ 0 };
    std::vector<int> erbBottomIndexes{ 0 };
    erbUpperFrequencies.push_back(h->freqVec[0]);
    float bandMultiplier = (std::pow(2.0f, 1.0f / 3.0f) + 1.0f) / 2.0f;
    for (int i = 1; i < h->nBands; ++i) {

        //Break if maximum frequency limit is exceeded
        if (erbUpperFrequencies.back() > maximumFrequencyLimit) {
            break;
        }

        // Break also if the next center frequency exceeds the limit
        int nextCenterFrequencyIndx = erbUpperIndexes.back() + 1;
        if ((nextCenterFrequencyIndx < h->nBands) && (h->freqVec[nextCenterFrequencyIndx] >= maximumFrequencyLimit)) {
            break;
        }

        // Calculate the upper frequency limit for this ERB band
        float erbUpperFrequency = (24.7 * intermediateBandsERBWidth) + erbUpperFrequencies[i - 1] * (1.0 + 0.108 * bandMultiplier * intermediateBandsERBWidth);
        erbUpperFrequency = (std::min)(erbUpperFrequency, maximumFrequencyLimit);

        // Find the frequency bin in the original centerFrequencies, that is
        // closest to the upper ERB frequency
        int closestBinIndex = 0;
        float minFrequencyDifference = INFINITY;
        for (int binIdx = 0; binIdx < h->nBands; ++binIdx){
            float frequencyDifference = abs(erbUpperFrequency - h->freqVec[binIdx]);
            if (frequencyDifference < minFrequencyDifference) {
                minFrequencyDifference = frequencyDifference;
                closestBinIndex = binIdx;
            }
        }
        // Ensure at least one bin difference between ERB bands
        if (closestBinIndex == erbUpperIndexes[i - 1]) {
            closestBinIndex += 1;
        }

        // Add the upper ERB frequency and the corresponing bin to arrays
        erbUpperFrequencies.push_back(h->freqVec[closestBinIndex]);
        erbUpperIndexes.push_back(closestBinIndex);
    }

    // Single bin bands for the low frequencies, and one band for the high frequencies
    std::vector<int>tempUpperIndexes(highestSingleBinBandsIndex + 1);
    std::iota(tempUpperIndexes.begin(), tempUpperIndexes.end(), 0);
    for (auto i : erbUpperIndexes) {
        if (i >= lowestHighBandIndex) {
            tempUpperIndexes.push_back(i);
            tempUpperIndexes.push_back(h->nBands - 1);
            break;
        }
        if (i > highestSingleBinBandsIndex) {
            tempUpperIndexes.push_back(i);
        }
    }
    erbUpperIndexes = tempUpperIndexes;

    // Reset the upper frequencies with the new low single bin bands
    erbUpperFrequencies.clear();
    for (auto i : erbUpperIndexes) {
        erbUpperFrequencies.push_back(h->freqVec[i]);
    }

    // Form the bottom indexes
    for (int i = 0; i < erbUpperIndexes.size() - 1; ++i) {
        erbBottomIndexes.push_back(erbUpperIndexes[i] + 1);
    }

    h->upperBandFrequencies = erbUpperFrequencies;
    h->upperBandIndexes = erbUpperIndexes;
    h->bottomBandIndexes = erbBottomIndexes;
    h->nComplexityReducedBands = erbUpperIndexes.size();
    h->maxFreqLimitIndex = h->bottomBandIndexes[h->nComplexityReducedBands - 1];
    h->nDoaDtrBins = h->maxFreqLimitIndex;

    // Determine low index for averaging high frequencies. Round maxFreqLimitIndex to the closest 3rd octave band limit,
    // and use the previous 3rd octave band limit as lowAvgLimitIndex.
    int closestIndex = 0;
    int smallestDiff = INT_MAX;
    for (int i = 0; i < oneThirdOctaveBandIndexes.size(); ++i) {
        int diff = std::abs(h->maxFreqLimitIndex - oneThirdOctaveBandIndexes[i]);
        if (diff < smallestDiff) {
            smallestDiff = diff;
            closestIndex = i;
        }
    }
    if (closestIndex > 0) {
        h->lowAvgLimitIndex = oneThirdOctaveBandIndexes[closestIndex - 1];
    } else if (h->maxFreqLimitIndex > 0) {
        h->lowAvgLimitIndex = h->maxFreqLimitIndex - 1;
    } else {
        h->lowAvgLimitIndex = 0;
    }
    if (h->maxFreqLimitIndex == h->lowAvgLimitIndex && h->maxFreqLimitIndex != 0) {
        h->lowAvgLimitIndex = h->maxFreqLimitIndex - 1;
    }
}

bool initialize_position_pre_processor(hoaTranslatorData *h)
{
    // Triangulation for the mic arrays
    if (h->useMphoa1SourceMode) {
        triangulation_for_one_source(h);
    } else if(h->useMphoa2SourceMode) {
        triangulation_for_two_sources(h);
    } else {
        try {
            delaunay_triangulation(h);
        } catch (const std::exception& e) {
            assert(false && "Delaunay triangulation failed for HOA translation.");
            return false;
        }
    }
    if (h->nTriangles == 0) {
        assert(false && "No triangles were formed as part of the Delaunay triangulation. Check the source positions and maxTriangleEdgeLength, if omitLongTriangleEdges is set to true.");
        return false;
    }
    return true;
}

void initialize_afSTFT(hoaTranslatorData *h)
{
    // afSTFT backward handle for the inverse transform
    afSTFT_create(&(h->afSTFTbackwardHandle), h->outChannels, h->outChannels, h->hopSize, h->LDmode, h->hybridMode, AFSTFT_BANDS_CH_TIME);

    // afSTFT forward handles for each microphone array
    h->afSTFTforwardHandlesFullOrder = new void *[h->nMicrophoneArrays];
    h->afSTFTforwardHandlesFOA = new void *[h->nMicrophoneArrays];

    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        afSTFT_create(&(h->afSTFTforwardHandlesFullOrder[mic]), h->channelsPerArray, h->outChannels, h->hopSize, h->LDmode, h->hybridMode, AFSTFT_TIME_CH_BANDS);
        afSTFT_create(&(h->afSTFTforwardHandlesFOA[mic]), 4, h->outChannels, h->hopSize, h->LDmode, h->hybridMode, AFSTFT_TIME_CH_BANDS);
    }
    h->nBands = afSTFT_getNBands(h->afSTFTforwardHandlesFullOrder[0]);
    h->freqVec = allocateFloatArray1d(h->nBands);
    afSTFT_getCentreFreqs(h->afSTFTforwardHandlesFullOrder[0], h->sampleRate, h->nBands, h->freqVec);
}

bool initialize_HRTFs(hoaTranslatorData* h,
                      inputFilesAndFlags inputFiles)
{
    h->sofaFilePath = h->resourcePath + inputFiles.sofaFile;
    h->useMatlabFile = inputFiles.useMatlabFile;
    h->matlabHrtfFile = h->resourcePath + inputFiles.matlabHrtfFile;
    h->matlabDirsFile = h->resourcePath + inputFiles.matlabDirsFile;
    h->matlabItdFile = h->resourcePath + inputFiles.matlabItdFile;

    // Load a set of HRIRs and transform them to HRTFs OR load a pre-calculated set of HRTFs
    get_HRTFs(h);

    // Get the closest HRTF indexes for a grid of points
    nearest_neighbor_hrtf_interpolation(h);

    return true;
}

void initialize_ambi2bin_decoder(hoaTranslatorData* h,
                                 BINAURAL_AMBI_DECODER_METHODS decodingMethod)
{
    h->decodingMethod = decodingMethod;
    float *weights = NULL;
    int enableDiffCovMatching = 0;
    int enableMaxReWeighting = 1;
    float_complex *tempDecoder = new float_complex[h->nBands * h->nBinauralChannels * ORDER2NSH(h->shOrder)]{};
    getBinauralAmbiDecoderMtx(FLATTEN3D(h->hrtfFB),
                              h->hrtfDirsDeg,
                              h->nHrtfDirs,
                              h->nBands,
                              h->decodingMethod,
                              h->shOrder,
                              h->freqVec,
                              h->itds,
                              weights,
                              enableDiffCovMatching,
                              enableMaxReWeighting,
                              tempDecoder);

    for (int band = 0; band < h->nBands; ++band) {
        for (int ch = 0; ch < h->outChannels; ++ch) {
            for (int sh = 0; sh < h->channelsPerArray; ++sh) {
                h->ambi2binDecodingMatrix[band][ch][sh] = tempDecoder[band * 2 * h->channelsPerArray + ch * h->channelsPerArray + sh];
            }
        }
    }
    delete[] tempDecoder;
}

void initialize_source_aware_processing(hoaTranslatorData *h,
                                        sourceAwareParameters sourceAwareParams,
                                        std::vector<sourceAwareParametersPerSource> const& sourceAwareParamsPerSource)
{
    h->nSources = sourceAwareParams.sourcePositionsOpenGL.size();
    h->useSourceAwarePostFilter = sourceAwareParams.useSourceAwarePostFilter;
    h->useSourceAwareTimeAlignment = sourceAwareParams.useSourceAwareTimeAlignment;
    h->sourceResidualMode = sourceAwareParams.sourceResidualMode;
    h->targetScmMode = sourceAwareParams.targetScmMode;
    h->maxSimulSources = sourceAwareParams.maxSimulSources;

    for (int src = 0; src < h->nSources; ++src) {
        h->listenerThresholdPresentVec.push_back(sourceAwareParamsPerSource[src].listenerThresholdPresent);
        h->hoaSourceThresholdPresentVec.push_back(sourceAwareParamsPerSource[src].hoaSourceThresholdPresent);
        h->listenerThresholdVec.push_back(sourceAwareParamsPerSource[src].listenerThreshold);
        h->hoaSourceThresholdVec.push_back(sourceAwareParamsPerSource[src].hoaSourceThreshold);
        h->priorityValueVec.push_back(sourceAwareParamsPerSource[src].priorityValue);
    }

    // float *
    h->intensityVectorMultipliers = allocateFloatArray1d(h->nBands);
    h->micIntensityX = allocateFloatArray1d(h->nBands);
    h->micIntensityY = allocateFloatArray1d(h->nBands);
    h->micIntensityZ = allocateFloatArray1d(h->nBands);
    h->resIntensityX = allocateFloatArray1d(h->nBands);
    h->resIntensityY = allocateFloatArray1d(h->nBands);
    h->resIntensityZ = allocateFloatArray1d(h->nBands);
    h->srcIntensityX = allocateFloatArray1d(h->nBands);
    h->srcIntensityY = allocateFloatArray1d(h->nBands);
    h->srcIntensityZ = allocateFloatArray1d(h->nBands);
    h->tempBandVec1 = allocateFloatArray1d(h->nBands);
    h->tempBandVec2 = allocateFloatArray1d(h->nBands);
    h->tempBandVec3 = allocateFloatArray1d(h->nBands);

    // float **
    h->interpolatedResidualAzimuthData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->interpolatedResidualDTRData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->interpolatedResidualElevationData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->interpolatedResidualEnergyData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->srcAzimuthData = allocateFloatArray2d(h->nMicrophoneArrays, h->nBands);
    h->srcDTRData = allocateFloatArray2d(h->nMicrophoneArrays, h->nBands);
    h->srcElevationData = allocateFloatArray2d(h->nMicrophoneArrays, h->nBands);
    h->srcEnergyData = allocateFloatArray2d(h->nMicrophoneArrays, h->nBands);

    // float ***
    h->micToSrcAziElevDeg = allocateFloatArray3d(h->nSources, h->nMicrophoneArrays, 2);
    h->residualAzimuthData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->residualDTRData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->residualElevationData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->residualEnergyData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);

    // float_complex
    h->TFDListenerSrcTAFrame = allocateFloatCmplxArray3d(h->nSources, h->nFrames, h->nBands);
    h->TFDMicSrcTAFrame = allocateFloatCmplxArray4d(h->nMicrophoneArrays, h->nSources, h->nFrames, h->nBands);
    h->TFDSourceFrame = allocateFloatCmplxArray3d(h->nSources, h->nFrames, h->nBands);

    h->nSourceBufferFrames = ((((h->maxDelayHops % h->nFrames) + h->maxDelayHops + h->nFrames - 1) / h->nFrames) + 1) * h->nFrames;
    h->sourceFrameBuffer = allocateFloatCmplxArray3d(h->nSources, h->nSourceBufferFrames, h->nBands);

    // vector
    h->closestMicsForSources.resize(h->nSources, 0);
    h->listenerToSrcDistances.resize(h->nSources, 0.0f);
    h->secondClosestMicsForSources.resize(h->nSources, 0);
    h->distanceToClosestMicForSource.resize(h->nSources, INFINITY);
    h->distanceToSecondClosestMicForSource.resize(h->nSources, INFINITY);

    for (int i = 0; i < h->nSources; ++i) {
        h->closestMicToMicDelayHops.emplace_back(h->nMicrophoneArrays, 0);
        h->closestMicToMicDelaySamples.emplace_back(h->nMicrophoneArrays, 0);
        h->micToMicDelayHops.emplace_back(h->nMicrophoneArrays, std::vector<int>(h->nMicrophoneArrays, 0));
        h->micToMicDelaySamples.emplace_back(h->nMicrophoneArrays, std::vector<int>(h->nMicrophoneArrays, 0));
        h->micToSrcDelaySamples.emplace_back(h->nMicrophoneArrays, 0);
        h->secondClosestMicToMicDelayHops.emplace_back(h->nMicrophoneArrays, 0);
        h->secondClosestMicToMicDelaySamples.emplace_back(h->nMicrophoneArrays, 0);
    }

    for (int i = 0; i < h->nTotalFramesTracked; ++i) {
        h->trackedSourceMicArrays.emplace_back(0, 0);
    }

    for (int i = 0; i < h->nSources; ++i) {
        h->Es_noTA.emplace_back(h->nBands, 0.0f);
        h->closestMicToMicAttenuationGains.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->closestMicToMicAttenuationGainsSquared.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->micToSrcAttenuationGains.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->micToSrcDistances.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->secondClosestMicToMicAttenuationGains.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->secondClosestMicToMicAttenuationGainsSquared.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->sourcePositionsOpenGL.emplace_back(3, 0.0f);
    }

    // hoatMatrixCmplx2D
    createHoatMatrix(&h->temp2x2MatrixCmplx, 2, 2);
    createHoatMatrix(&h->tempBandMatrix, 1, h->nBands);
    createHoatMatrix(&h->tempInputFrameMatrix, h->channelsPerArray, h->nBands);

    // hoatMatrixCmplx3D
    createHoatMatrix(&h->Cyo_res, h->nBands, 2, 2);
    createHoatMatrix(&h->Cyo_src, h->nBands, 2, 2);
    createHoatMatrix(&h->micToClosestSrcBeams, h->nSources, 1, ORDER2NSH(h->shOrder));
    createHoatMatrix(&h->micToSecondClosestSrcBeams, h->nSources, 1, ORDER2NSH(h->shOrder));

    // hoatMatrix
    createHoatMatrix(&h->Cam_src, h->nMicrophoneArrays, h->nBands, 4, 4);
    createHoatMatrix(&h->Cao_src, h->nBands, 4, 4);
    createHoatMatrix(&h->listenerToSrcVectorsOpenGL, h->nSources, 3, 1);
    createHoatMatrix(&h->micToSrcYY, h->nSources, h->nMicrophoneArrays, 4, 4);
    createHoatMatrix(&h->temp4x1Matrix, 4, 1);
    createHoatMatrix(&h->temp4x4Matrix, 4, 4);

    // Copy source positions
    for (int src = 0; src < h->nSources; ++src) {
        std::copy_n(&sourceAwareParams.sourcePositionsOpenGL[src][0],
                    3,
                    &h->sourcePositionsOpenGL[src][0]);
    }
}

void create_hoa_translator(hoaTranslatorData **phHoaTranslatorData,
                           int frameSize,
                           int sampleRate,
                           int hopSize,
                           int hybridMode,
                           int LDmode,
                           int outChannels,
                           int channelsPerArray,
                           int nAnalysisChannels,
                           std::string resourcePath,
                           DIFFUSE_METHOD_FOR_TARGET_SCM diffuseMethodForTargetSCM,
                           micArrayPositionsAndOrientations micPosAndOrientations,
                           std::shared_ptr<std::vector<std::uint8_t>> spatialMetadata,
                           sourceAwareParameters sourceAwareParams,
                           std::vector<sourceAwareParametersPerSource> const& sourceAwareParamsPerSource)
{
    hoaTranslatorData *h = *phHoaTranslatorData;
    h->spatialMetadataByteStream = std::make_unique<hoatranslation::ByteStream>(spatialMetadata);

    h->frameSize = frameSize;
    h->sampleRate = sampleRate;
    h->hopSize = hopSize;
    h->nFrames = h->frameSize / h->hopSize;
    h->hybridMode = hybridMode;
    h->LDmode = LDmode;
    h->outChannels = outChannels;
    h->channelsPerArray = channelsPerArray;
    h->shOrder = NSH2ORDER(h->channelsPerArray);
    h->nAnalysisChannels = nAnalysisChannels;
    h->resourcePath = resourcePath;
    h->diffuseMethodForTargetSCM = diffuseMethodForTargetSCM;
    h->nMicrophoneArrays = micPosAndOrientations.arrayPositionsOpenGL.size();
    if (h->nMicrophoneArrays == 1) {
        h->useMphoa1SourceMode = true;
    } else if (h->nMicrophoneArrays == 2) {
        h->useMphoa2SourceMode = true;
    }
    //// Initialize Alias-free Short-time Fourier transform (afSTFT)
    initialize_afSTFT(h);

    //// Memory allocations
    // float *
    h->DTRDataAvg = allocateFloatArray1d(h->nBands);
    h->azimuthDataAvg = allocateFloatArray1d(h->nBands);
    h->barycentricCoordinates = allocateFloatArray1d(3);
    h->cosAziNbands = allocateFloatArray1d(h->nBands);
    h->cosElevNbands = allocateFloatArray1d(h->nBands);
    h->cosAzi = allocateFloatArray1d(h->nMicrophoneArrays * h->nFrames * h->nBands);
    h->cosElev = allocateFloatArray1d(h->nMicrophoneArrays * h->nFrames * h->nBands);
    h->currentInterpolationWeights = allocateFloatArray1d(h->nMicrophoneArrays);
    h->dotX = allocateFloatArray1d(h->nBands);
    h->dotXYZplus1 = allocateFloatArray1d(h->nBands);
    h->dotY = allocateFloatArray1d(h->nBands);
    h->dotZ = allocateFloatArray1d(h->nBands);
    h->elevationDataAvg = allocateFloatArray1d(h->nBands);
    h->energyDataAvg = allocateFloatArray1d(h->nBands);
    h->inputSignalBuffer = allocateFloatArray1d(h->frameSize * h->channelsPerArray);
    h->interpolator = allocateFloatArray1d(h->nFrames);
    h->originalDoaUnitX = allocateFloatArray1d(h->nBands);
    h->originalDoaUnitY = allocateFloatArray1d(h->nBands);
    h->originalDoaUnitZ = allocateFloatArray1d(h->nBands);
    h->outputSignalBuffer = allocateFloatArray1d(h->frameSize * h->outChannels);
    h->previousInterpolationWeights = allocateFloatArray1d(h->nMicrophoneArrays);
    h->reMappedDoaX = allocateFloatArray1d(h->nBands);
    h->reMappedDoaY = allocateFloatArray1d(h->nBands);
    h->reMappedDoaZ = allocateFloatArray1d(h->nBands);
    h->reMappedDtr = allocateFloatArray1d(h->nBands);
    h->sinAziNbands = allocateFloatArray1d(h->nBands);
    h->sinElevNbands = allocateFloatArray1d(h->nBands);
    h->tempNbands1 = allocateFloatArray1d(h->nBands);
    h->tempNbands2 = allocateFloatArray1d(h->nBands);
    h->sinAzi = allocateFloatArray1d(h->nMicrophoneArrays * h->nFrames * h->nBands);
    h->sinElev = allocateFloatArray1d(h->nMicrophoneArrays * h->nFrames * h->nBands);
    h->tempSHrotationMatrix = allocateFloatArray1d(ORDER2NSH(h->shOrder) * ORDER2NSH(h->shOrder));
    h->w1 = allocateFloatArray1d(h->nBands);
    h->w2 = allocateFloatArray1d(h->nBands);

    // float **
    h->chosenInterFrameWeights = allocateFloatArray2d(h->nMicrophoneArrays, h->nFrames);
    h->currentInterFrameWeights = allocateFloatArray2d(h->nMicrophoneArrays, h->nFrames);
    h->eqGains = allocateFloatArray2d(h->nFrames, h->nBands);
    h->fadeInWeights = allocateFloatArray2d(h->nMicrophoneArrays, h->nFrames);
    h->fadeOutWeights = allocateFloatArray2d(h->nMicrophoneArrays, h->nFrames);
    h->intensityVectorAvg = allocateFloatArray2d(h->nBands, 3);
    h->interpolatedAzimuthData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->interpolatedDTRData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->interpolatedElevationData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->interpolatedEnergyData = allocateFloatArray2d(h->nFrames, h->nBands);
    h->metadataV1 = allocateFloatArray2d(h->nFrames, h->nBands);
    h->metadataV2 = allocateFloatArray2d(h->nFrames, h->nBands);
    h->metadataV3 = allocateFloatArray2d(h->nFrames, h->nBands);
    h->microphoneArrayPositions = allocateFloatArray2d(h->nMicrophoneArrays, 3);
    h->singlePoint6DofDirGains = allocateFloatArray2d(h->nFrames, h->nBands);
    h->timeDomainHopFrame = allocateFloatArray2d(h->channelsPerArray, h->hopSize);
    h->timeDomainOutputFrame = allocateFloatArray2d(h->outChannels, h->frameSize);
    h->transientDetector1 = allocateFloatArray2d(h->nBands, h->outChannels);
    h->transientDetector2 = allocateFloatArray2d(h->nBands, h->outChannels);

    // float ***
    h->azimuthData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->DTRData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->elevationData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->energyData = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->tempV1 = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->tempV2 = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->tempV3 = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);
    h->timeDomainInputFrame = allocateFloatArray3d(h->nMicrophoneArrays, h->channelsPerArray, h->frameSize);
    h->weightedEnergy = allocateFloatArray3d(h->nMicrophoneArrays, h->nFrames, h->nBands);

    // int *
    h->decorrelationDelays = allocateIntArray1d(h->nBands * h->outChannels);

    // vector
    h->nTotalFramesTracked = std::ceil(12 * h->hopSize / h->frameSize) + 1;
    for (int i = 0; i < h->nTotalFramesTracked; ++i) {
        h->trackedCurrentTriangleArrays.emplace_back(3, 0);
        h->trackedEdgeTriangleArrays.emplace_back(6, 0);
        h->trackedCurrentInterFrameWeights.emplace_back(h->nMicrophoneArrays * h->nFrames, 0.0f);
        h->trackedCurrentInterpolationWeights.emplace_back(h->nMicrophoneArrays, 0.0f);
        h->trackedPreviousInterpolationWeights.emplace_back(h->nMicrophoneArrays, 0.0f);
    }
    for (int frame = 0; frame < h->nFrames; ++frame) {
        h->fadeOutBands.emplace_back(h->nBands, 0);
    }
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        h->micArrayOrientations.emplace_back(3, 0.0f);
    }
    h->eqGainsMax.resize(h->nFrames * h->nBands, 0.0f);
    for (int ch = 0; ch < h->nAnalysisChannels; ++ch) {
        h->tempAnalysisVec.push_back({0.0f, 0.0f});
    }

    // Covariance domain framework handle, used in optimal mixing
    cdf4sap_cmplx_create(&(h->covarianceDomainFramework),
                         h->outChannels,
                         h->outChannels);

    // float_complex **
    h->tempSignal = allocateFloatCmplxArray2d(h->channelsPerArray, h->nBands);

    // float_complex ***
    h->decorrelatedTFDFrame = allocateFloatCmplxArray3d(h->nBands, h->outChannels, h->nFrames);
    h->eqGainsCmplx = allocateFloatCmplxArray3d(h->nFrames, h->channelsPerArray, h->nBands);
    h->interpolatedTFDFrame = allocateFloatCmplxArray3d(h->nFrames, h->channelsPerArray, h->nBands);
    h->outputTFDProcessingFrame = allocateFloatCmplxArray3d(h->nBands, h->outChannels, h->nFrames);
    h->tempTFDFrame = allocateFloatCmplxArray3d(h->nBands, h->outChannels, h->nFrames);
    h->TFDOutputFrame = allocateFloatCmplxArray3d(h->nFrames, h->outChannels, h->nBands);

    h->nSamplesInDecorrelationBuffer = 2 * 2048 + h->frameSize;
    h->nBlocksInDecorrelationBuffer = std::ceil(h->nSamplesInDecorrelationBuffer / h->frameSize);
    h->inputXBlocksPrototype = allocateFloatCmplxArray3d(h->nBands, h->outChannels, h->nBlocksInDecorrelationBuffer * h->nFrames);

    // float_complex ****
    h->TFDInputFrameFullOrder = (float_complex****)calloc1d(h->nMicrophoneArrays, sizeof(float_complex));
    h->TFDInputFrameFOA = (float_complex****)calloc1d(h->nMicrophoneArrays, sizeof(float_complex));
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        h->TFDInputFrameFullOrder[mic] = allocateFloatCmplxArray3d(h->nFrames, h->channelsPerArray, h->nBands);
        h->TFDInputFrameFOA[mic] = allocateFloatCmplxArray3d(h->nFrames, 4, h->nBands);
    }


    // hoatMatrixCmplx2D
    createHoatMatrix(&h->Cr, 2, 2);
    createHoatMatrix(&h->CxNormalized, 2, 2);
    createHoatMatrix(&h->CyNormalized, 2, 2);
    createHoatMatrix(&h->CyTilde, 2, 2);
    createHoatMatrix(&h->DCM, h->outChannels, h->outChannels);
    createHoatMatrix(&h->DCMambi, h->outChannels, h->outChannels);
    createHoatMatrix(&h->DCMdiff, h->outChannels, h->outChannels);
    createHoatMatrix(&h->DsqrtInv, 2, 2);
    createHoatMatrix(&h->Dtemp, 2, 2);
    createHoatMatrix(&h->Genergy, 2, 2);
    createHoatMatrix(&h->GyHatQ, 2, 2);
    createHoatMatrix(&h->Kx, 2, 2);
    createHoatMatrix(&h->Ky, 2, 2);
    createHoatMatrix(&h->KyGQKx, 2, 2);
    createHoatMatrix(&h->P, 2, 2);
    createHoatMatrix(&h->Sx, 2, 2);
    createHoatMatrix(&h->SxSqrt, 2, 2);
    createHoatMatrix(&h->Utemp, 2, 2);
    createHoatMatrix(&h->Ux, 2, 2);
    createHoatMatrix(&h->temp2x2Cmplx, 2, 2);
    createHoatMatrix(&h->temp2x2Cmplx2, 2, 2);
    createHoatMatrix(&h->tempAnalysisMatrix, h->nAnalysisChannels, 1);
    createHoatMatrix(&h->tempBinauralMatrix, h->outChannels, 1);
    createHoatMatrix(&h->tempHrtfMatrix, h->nBinauralChannels, 1);
    createHoatMatrix(&h->tempHrtfProductMatrix, h->nBinauralChannels, h->nBinauralChannels);
    createHoatMatrix(&h->tempInputMatrix, h->channelsPerArray, h->nFrames);
    createHoatMatrix(&h->tempInputMatrixSmall, h->channelsPerArray, 1);
    createHoatMatrix(&h->tempMatrixOutChSqrd, h->outChannels, h->outChannels);
    createHoatMatrix(&h->tempMixingMatrix, h->outChannels, h->outChannels);
    createHoatMatrix(&h->tempOutputMatrix, h->outChannels, 1);
    createHoatMatrix(&h->tempRotatedMatrix, h->channelsPerArray, h->nFrames);
    createHoatMatrix(&h->tempRotatedMatrixSmall, h->channelsPerArray, 1);
    createHoatMatrix(&h->tempSignalMatrix, h->outChannels, 1);

    // hoatMatrixCmplx3D
    createHoatMatrix(&h->ambi2binDecodingMatrix, h->nBands, h->outChannels, h->channelsPerArray);
    createHoatMatrix(&h->Cdiffuse, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->currentBinauralMatrix, h->nBands, h->outChannels, h->nFrames);
    createHoatMatrix(&h->currentM, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->currentMr, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->Cx, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->Cy, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->CyLinearDecoding, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->newCxMatrix, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->newCyMatrix, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->previousBinauralMatrix, h->nBands, h->outChannels, h->nFrames);
    createHoatMatrix(&h->previousCxMatrix, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->previousCyMatrix, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->previousM, h->nBands, h->outChannels, h->outChannels);
    createHoatMatrix(&h->previousMr, h->nBands, h->outChannels, h->outChannels);

    // hoatMatrix2D
    createHoatMatrix(&h->Aimag, 2, 2);
    createHoatMatrix(&h->Are, 2, 2);
    createHoatMatrix(&h->GyHat, 2, 1);
    createHoatMatrix(&h->KxImag, 2, 2);
    createHoatMatrix(&h->KxRe, 2, 2);
    createHoatMatrix(&h->KyImag, 2, 2);
    createHoatMatrix(&h->KyImagInv, 2, 2);
    createHoatMatrix(&h->KyRe, 2, 2);
    createHoatMatrix(&h->Q, 2, 2);
    createHoatMatrix(&h->SxSqrtReg, 2, 2);
    createHoatMatrix(&h->UxImag, 2, 2);
    createHoatMatrix(&h->UxRe, 2, 2);
    createHoatMatrix(&h->currentSHrotationMatrix, ORDER2NSH(h->shOrder), ORDER2NSH(h->shOrder));
    createHoatMatrix(&h->previousSHrotationMatrix, ORDER2NSH(h->shOrder), ORDER2NSH(h->shOrder));
    createHoatMatrix(&h->realSHrotationMatrix, ORDER2NSH(h->shOrder), ORDER2NSH(h->shOrder));
    createHoatMatrix(&h->temp3x1Matrix, 3, 1);
    createHoatMatrix(&h->temp3x1Matrix2, 3, 1);
    createHoatMatrix(&h->temp3xnBandsMatrix, 3, h->nBands);
    createHoatMatrix(&h->temp3xnBandsMatrix2, 3, h->nBands);
    createHoatMatrix(&h->tempImag, 2, 2);
    createHoatMatrix(&h->tempInputMatrixImag, h->channelsPerArray, h->nFrames);
    createHoatMatrix(&h->tempInputMatrixReal, h->channelsPerArray, h->nFrames);
    createHoatMatrix(&h->tempInputMatrixSmallImag, h->channelsPerArray, 1);
    createHoatMatrix(&h->tempInputMatrixSmallReal, h->channelsPerArray, 1);
    createHoatMatrix(&h->tempR, 3, 3);
    createHoatMatrix(&h->tempRe, 2, 2);
    createHoatMatrix(&h->tempRotatedMatrixImag, h->channelsPerArray, h->nFrames);
    createHoatMatrix(&h->tempRotatedMatrixReal, h->channelsPerArray, h->nFrames);
    createHoatMatrix(&h->tempRotatedMatrixSmallImag, h->channelsPerArray, 1);
    createHoatMatrix(&h->tempRotatedMatrixSmallReal, h->channelsPerArray, 1);

    // Copy mic array positions and orientations
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        std::copy_n(&micPosAndOrientations.arrayPositionsOpenGL[mic][0],
                    3,
                    &h->microphoneArrayPositions[mic][0]);
        std::copy_n(&micPosAndOrientations.arrayOrientationsOpenGL[mic][0],
                    3,
                    &h->micArrayOrientations[mic][0]);
    }

    // Initialize Q for optimal mixing
    h->Q[0][0] = 1.0f;
    h->Q[1][1] = 1.0f;
    h->Q[0][1] = 0.0f;
    h->Q[1][0] = 0.0f;

    // Source aware parameters
    h->useSourceAwareMode = sourceAwareParams.useSourceAwareMode;
    h->useSourceAwareInitialFlag = h->useSourceAwareMode;
    if (h->useSourceAwareMode) {
        initialize_source_aware_processing(h, sourceAwareParams, sourceAwareParamsPerSource);
    }
}

bool initialize_hoa_translator(hoaTranslatorData **phHoaTranslatorData,
                               int HOAGroupId,
                               bool useEncodedValues,
                               float interpolationEqGainMax,
                               BINAURAL_AMBI_DECODER_METHODS decodingMethod,
                               inputFilesAndFlags inputFiles,
                               optimalMixingParameters optimalMixingParams,
                               exteriorRenderingParameters exteriorRenderingParams,
                               sphoa6DofExtensionParameters sphoaParameters,
                               std::vector<float> const& linearSourceGains,
                               std::vector<int> const& complexityReductionParameters)
{
    hoaTranslatorData *h = *phHoaTranslatorData;

    // Input parameters
    h->HOAGroupId = HOAGroupId;
    h->useEncodedValues = useEncodedValues;
    h->interpolationEqGainMax = interpolationEqGainMax;
    std::fill(h->eqGainsMax.begin(), h->eqGainsMax.end(), h->interpolationEqGainMax);

    // Optimal mixing parameters
    h->lowerBandFreqLimit = optimalMixingParams.lowerBandFreqLimit;
    h->useDiffuseDecorrelation = optimalMixingParams.useDiffuseDecorrelation;
    h->useEqualizer = optimalMixingParams.useEqualizer;
    h->useRecursiveAveraging = optimalMixingParams.useRecursiveAveraging;
    h->recursionSmoothingCoefficient = optimalMixingParams.recursionSmoothingCoefficient;

    // Single-point HOA 6DoF extension parameters
    h->useSphoa6DofExtensionMode = sphoaParameters.useSphoa6DofExtension;
    h->sphoa6DofMaxGain = std::pow(10.0f, (sphoaParameters.sphoa6DofMaxGainDb / 20.0f));
    h->sphoa6DofProjectionRadius = sphoaParameters.sphoa6DofProjectionRadius;
    h->NthSphoaInEncodedMetadata = sphoaParameters.NthSphoaInEncodedMetadata;
    if (h->useSphoa6DofExtensionMode) {
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            h->transformedArraysForCurrentFrame.push_back(mic);
            h->activeArraysForCurrentFrame.push_back(mic);
        }
        std::fill_n(&h->chosenInterFrameWeights[0][0], h->nMicrophoneArrays * h->nFrames, 1.0f);
    }

    //// Initialize position pre-processor
    if(!h->useSphoa6DofExtensionMode && !initialize_position_pre_processor(h)) {
        return false;
    }

    //// Load or calculate an interpolated set of HRTFs
    if(!initialize_HRTFs(h, inputFiles)) {
        return false;
    }

    //// Compute a linear Ambisonics to binaural decoding matrix for the prototype signal
    initialize_ambi2bin_decoder(h, decodingMethod);

    //// Decorrelation delays for diffuse decorrelation
    int maxTFdelay = std::ceil((h->nSamplesInDecorrelationBuffer - h->frameSize) / h->hopSize);
    getDecorrelationDelays(h->outChannels,
                           h->freqVec,
                           h->nBands,
                           h->hrirFs,
                           maxTFdelay,
                           h->hopSize,
                           h->decorrelationDelays);

    //// Interpolator to linearly interpolate between frames
    for (int frame = 0; frame < h->nFrames; ++frame) {
        h->interpolator[frame] = ((float)frame + 1.0f) / float(h->nFrames);
    }

    //// Process loop variables
    h->currentFrame = 0;
    h->currentClosestMicrophoneArray = 0;
    h->previousClosestMicrophoneArray = 0;

    if (h->useEncodedValues) {
        read_spatial_metadata_info(h, true);
    }

    //// Crossfade
    h->crossfadeBandLow = {0, 2, 3, 4, 5, 6, 8, 10, 12, 14, 17, 20, 23, 27, 32, 37, 43, 50, 57, 66, 76, 88, 101, 116};
    h->crossfadeBandHigh = {1, 2, 3, 4, 5, 7, 9, 11, 13, 16, 19, 22, 26, 31, 36, 42, 49, 56, 65, 75, 87, 100, 115, h->nBands-1};
    h->crossfadeBandSwaporder = {7, 1, 2, 22, 21, 12, 18, 0, 5, 9, 8, 14, 13, 23, 15, 20, 19, 4, 16, 17, 10, 11, 3, 6};
    h->crossfadeLen = h->crossfadeBandLow.size();

    //// Improved processing outside capturing area
    h->useImprovedExteriorRendering = exteriorRenderingParams.useImprovedExteriorRendering;
    h->exteriorProjectionRadius = exteriorRenderingParams.exteriorProjectionRadius;
    h->adaptiveExteriorProjectionRadius = exteriorRenderingParams.adaptiveExteriorProjectionRadius;
    if (h->useEncodedValues) {
        h->exteriorProjectionRadius = (float)h->decoderData1.exterior_projection_radius / (float)h->decoderData1.exterior_multiplier;
    }

    //// Estimate diffuse field covariance matrix for the optimal mixing
    if (h->diffuseMethodForTargetSCM == DIFFUSE_COVARIANCE_MATRIX || h->useImprovedExteriorRendering || h->useSourceAwareMode) {
        estimate_diffuse_field_covariance_matrix(h);
    }

    //// Source energy gains
    h->sourceEnergyGains = std::vector<float>(h->nMicrophoneArrays);
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        h->sourceEnergyGains[mic] = std::pow(linearSourceGains[mic], 2.0f);
    }

    //// Exterior processing
    if (h->useImprovedExteriorRendering || h->useSourceAwareMode) {
        exterior_pre_processing(h);
    }

    //// Complexity reduction
    h->useComplexityReduction = !complexityReductionParameters.empty();
    if (h->useComplexityReduction) {
        initialize_complexity_reduction(h, complexityReductionParameters);
    } else {
        h->maxFreqLimitIndex = h->nBands;
        h->lowAvgLimitIndex = h->nBands;
        h->nDoaDtrBins = h->nBands;
    }

    //// Source aware
    if (h->useSourceAwareInitialFlag) {
        pre_process_source_aware(h);
    }

    return true;
}

void destroy_hoa_translator(hoaTranslatorData **phHoaTranslatorData)
{
    hoaTranslatorData *h = *phHoaTranslatorData;

    // float *
    free(h->DTRDataAvg);
    free(h->M_ESDtoHOA);
    free(h->azimuthDataAvg);
    free(h->barycentricCoordinates);
    free(h->cosAziNbands);
    free(h->cosElevNbands);
    free(h->cosAzi);
    free(h->cosElev);
    free(h->currentInterpolationWeights);
    free(h->dotX);
    free(h->dotXYZplus1);
    free(h->dotY);
    free(h->dotZ);
    free(h->elevationDataAvg);
    free(h->energyDataAvg);
    free(h->freqVec);
    free(h->hrirs);
    free(h->hrtfDirsDeg);
    free(h->inputSignalBuffer);
    free(h->interpolator);
    free(h->itds);
    free(h->originalDoaUnitX);
    free(h->originalDoaUnitY);
    free(h->originalDoaUnitZ);
    free(h->outputSignalBuffer);
    free(h->previousInterpolationWeights);
    free(h->reMappedDoaX);
    free(h->reMappedDoaY);
    free(h->reMappedDoaZ);
    free(h->reMappedDtr);
    free(h->sinAziNbands);
    free(h->sinElevNbands);
    free(h->tempNbands1);
    free(h->tempNbands2);
    free(h->sinAzi);
    free(h->sinElev);
    free(h->tempSHrotationMatrix);
    free(h->w1);
    free(h->w2);

    // float **
    free(h->chosenInterFrameWeights);
    free(h->currentInterFrameWeights);
    free(h->eqGains);
    free(h->fadeInWeights);
    free(h->fadeOutWeights);
    free(h->intensityVectorAvg);
    free(h->interpolatedAzimuthData);
    free(h->interpolatedDTRData);
    free(h->interpolatedElevationData);
    free(h->interpolatedEnergyData);
    free(h->metadataV1);
    free(h->metadataV2);
    free(h->metadataV3);
    free(h->microphoneArrayPositions);
    free(h->singlePoint6DofDirGains);
    free(h->timeDomainHopFrame);
    free(h->timeDomainOutputFrame);
    free(h->transientDetector1);
    free(h->transientDetector2);

    // float ***
    free(h->azimuthData);
    free(h->DTRData);
    free(h->elevationData);
    free(h->energyData);
    free(h->tempV1);
    free(h->tempV2);
    free(h->tempV3);
    free(h->timeDomainInputFrame);
    free(h->weightedEnergy);

    // int *
    free(h->decorrelationDelays);
    free(h->hrtfIdxGrid);

    // Covariance domain framework
    cdf4sap_cmplx_destroy(&(h->covarianceDomainFramework));

    // afSTFT
    afSTFT_destroy(&h->afSTFTbackwardHandle);
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        afSTFT_destroy(&h->afSTFTforwardHandlesFullOrder[mic]);
        afSTFT_destroy(&h->afSTFTforwardHandlesFOA[mic]);

    }
    delete[] h->afSTFTforwardHandlesFullOrder;
    delete[] h->afSTFTforwardHandlesFOA;


    // float_complex **
    free(h->tempSignal);

    // float_complex ***
    free(h->decorrelatedTFDFrame);
    free(h->eqGainsCmplx);
    free(h->hrtfFB);
    free(h->inputXBlocksPrototype);
    free(h->interpolatedTFDFrame);
    free(h->outputTFDProcessingFrame);
    free(h->tempTFDFrame);
    free(h->TFDOutputFrame);

    // float_complex ****
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        free(h->TFDInputFrameFOA[mic]);
        free(h->TFDInputFrameFullOrder[mic]);

    }
    free(h->TFDInputFrameFullOrder);
    free(h->TFDInputFrameFOA);


    // hoatMatrixCmplx2D
    destroyHoatMatrix(&h->Cr);
    destroyHoatMatrix(&h->CxNormalized);
    destroyHoatMatrix(&h->CyNormalized);
    destroyHoatMatrix(&h->CyTilde);
    destroyHoatMatrix(&h->DCM);
    destroyHoatMatrix(&h->DCMambi);
    destroyHoatMatrix(&h->DCMdiff);
    destroyHoatMatrix(&h->DsqrtInv);
    destroyHoatMatrix(&h->Dtemp);
    destroyHoatMatrix(&h->Genergy);
    destroyHoatMatrix(&h->GyHatQ);
    destroyHoatMatrix(&h->Kx);
    destroyHoatMatrix(&h->Ky);
    destroyHoatMatrix(&h->KyGQKx);
    destroyHoatMatrix(&h->P);
    destroyHoatMatrix(&h->Sx);
    destroyHoatMatrix(&h->SxSqrt);
    destroyHoatMatrix(&h->Utemp);
    destroyHoatMatrix(&h->Ux);
    destroyHoatMatrix(&h->temp2x2Cmplx);
    destroyHoatMatrix(&h->temp2x2Cmplx2);
    destroyHoatMatrix(&h->tempAnalysisMatrix);
    destroyHoatMatrix(&h->tempBinauralMatrix);
    destroyHoatMatrix(&h->tempHrtfMatrix);
    destroyHoatMatrix(&h->tempHrtfProductMatrix);
    destroyHoatMatrix(&h->tempInputMatrix);
    destroyHoatMatrix(&h->tempInputMatrixSmall);
    destroyHoatMatrix(&h->tempMatrixOutChSqrd);
    destroyHoatMatrix(&h->tempMixingMatrix);
    destroyHoatMatrix(&h->tempOutputMatrix);
    destroyHoatMatrix(&h->tempRotatedMatrix);
    destroyHoatMatrix(&h->tempRotatedMatrixSmall);
    destroyHoatMatrix(&h->tempSignalMatrix);

    // hoatMatrixCmplx3D
    destroyHoatMatrix(&h->ambi2binDecodingMatrix);
    destroyHoatMatrix(&h->Cdiffuse);
    destroyHoatMatrix(&h->currentBinauralMatrix);
    destroyHoatMatrix(&h->currentM);
    destroyHoatMatrix(&h->currentMr);
    destroyHoatMatrix(&h->Cx);
    destroyHoatMatrix(&h->Cy);
    destroyHoatMatrix(&h->CyLinearDecoding);
    destroyHoatMatrix(&h->newCxMatrix);
    destroyHoatMatrix(&h->newCyMatrix);
    destroyHoatMatrix(&h->previousBinauralMatrix);
    destroyHoatMatrix(&h->previousCxMatrix);
    destroyHoatMatrix(&h->previousCyMatrix);
    destroyHoatMatrix(&h->previousM);
    destroyHoatMatrix(&h->previousMr);

    // hoatMatrix2D
    destroyHoatMatrix(&h->Aimag);
    destroyHoatMatrix(&h->Are);
    destroyHoatMatrix(&h->GyHat);
    destroyHoatMatrix(&h->KxImag);
    destroyHoatMatrix(&h->KxRe);
    destroyHoatMatrix(&h->KyImag);
    destroyHoatMatrix(&h->KyImagInv);
    destroyHoatMatrix(&h->KyRe);
    destroyHoatMatrix(&h->Q);
    destroyHoatMatrix(&h->SxSqrtReg);
    destroyHoatMatrix(&h->UxImag);
    destroyHoatMatrix(&h->UxRe);
    destroyHoatMatrix(&h->currentSHrotationMatrix);
    destroyHoatMatrix(&h->previousSHrotationMatrix);
    destroyHoatMatrix(&h->realSHrotationMatrix);
    destroyHoatMatrix(&h->temp3x1Matrix);
    destroyHoatMatrix(&h->temp3x1Matrix2);
    destroyHoatMatrix(&h->temp3xnBandsMatrix);
    destroyHoatMatrix(&h->temp3xnBandsMatrix2);
    destroyHoatMatrix(&h->tempImag);
    destroyHoatMatrix(&h->tempInputMatrixImag);
    destroyHoatMatrix(&h->tempInputMatrixReal);
    destroyHoatMatrix(&h->tempInputMatrixSmallImag);
    destroyHoatMatrix(&h->tempInputMatrixSmallReal);
    destroyHoatMatrix(&h->tempR);
    destroyHoatMatrix(&h->tempRe);
    destroyHoatMatrix(&h->tempRotatedMatrixImag);
    destroyHoatMatrix(&h->tempRotatedMatrixReal);
    destroyHoatMatrix(&h->tempRotatedMatrixSmallImag);
    destroyHoatMatrix(&h->tempRotatedMatrixSmallReal);

    saf_sofa_close(&h->safSofaContainer);

    if (h->useSourceAwareMode) {
        // float *
        free(h->intensityVectorMultipliers);
        free(h->micIntensityX);
        free(h->micIntensityY);
        free(h->micIntensityZ);
        free(h->resIntensityX);
        free(h->resIntensityY);
        free(h->resIntensityZ);
        free(h->srcIntensityX);
        free(h->srcIntensityY);
        free(h->srcIntensityZ);
        free(h->tempBandVec1);
        free(h->tempBandVec2);
        free(h->tempBandVec3);

        // float **
        free(h->interpolatedResidualAzimuthData);
        free(h->interpolatedResidualDTRData);
        free(h->interpolatedResidualElevationData);
        free(h->interpolatedResidualEnergyData);
        free(h->srcAzimuthData);
        free(h->srcDTRData);
        free(h->srcElevationData);
        free(h->srcEnergyData);

        // float ***
        free(h->micToSrcAziElevDeg);
        free(h->residualAzimuthData);
        free(h->residualDTRData);
        free(h->residualElevationData);
        free(h->residualEnergyData);

        // float_complex
        free(h->TFDListenerSrcTAFrame);
        free(h->TFDSourceFrame);
        free(h->sourceFrameBuffer);
        free(h->TFDMicSrcTAFrame);

        // hoatMatrixCmplx2D
        destroyHoatMatrix(&h->temp2x2MatrixCmplx);
        destroyHoatMatrix(&h->tempBandMatrix);
        destroyHoatMatrix(&h->tempInputFrameMatrix);

        // hoatMatrixCmplx3D
        destroyHoatMatrix(&h->Cyo_res);
        destroyHoatMatrix(&h->Cyo_src);
        destroyHoatMatrix(&h->micToClosestSrcBeams);
        destroyHoatMatrix(&h->micToSecondClosestSrcBeams);

        // hoatMatrix
        destroyHoatMatrix(&h->temp4x1Matrix);
        destroyHoatMatrix(&h->temp4x4Matrix);
        destroyHoatMatrix(&h->Cao_src);
        destroyHoatMatrix(&h->listenerToSrcVectorsOpenGL);
        destroyHoatMatrix(&h->Cam_src);
        destroyHoatMatrix(&h->micToSrcYY);
    }
}

void determine_closest_microphone_array(hoaTranslatorData *h)
{
    auto calculate_closest_array_based_on_weights = [&]() {
        int currentClosestArray = 0;
        float biggestWeight = 0;
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            if (h->chosenInterFrameWeights[mic][h->nFrames - 1] > biggestWeight) {
                biggestWeight = h->chosenInterFrameWeights[mic][h->nFrames - 1];
                currentClosestArray = mic;
            }
        }
        return currentClosestArray;
    };

    auto are_closest_arrays_in_same_triangle = [&](int newClosestArrayCandidate) {
        auto const& activeTriangle = h->triangleFaces[h->activeTriangleIndex];
        return !(std::find(activeTriangle.begin(), activeTriangle.end(), newClosestArrayCandidate) == activeTriangle.end() ||
                 std::find(activeTriangle.begin(), activeTriangle.end(), h->previousClosestMicrophoneArray) == activeTriangle.end());
    };

    auto is_distance_threshold_exceeded = [&](int newClosestArrayCandidate, float distanceThreshold) {
        std::vector<float> listenerToCurrentClosestArray(3, 0.0f);
        std::vector<float> listenerToPreviousClosestArray(3, 0.0f);
        for (int i = 0; i < 3; ++i) {
            listenerToCurrentClosestArray[i]  = h->currentListenerPosition[i] - h->microphoneArrayPositions[newClosestArrayCandidate][i];
            listenerToPreviousClosestArray[i] = h->currentListenerPosition[i] - h->microphoneArrayPositions[h->previousClosestMicrophoneArray][i];
        }
        float distanceListenerToCurrentClosestArray = calculate_vector_norm(listenerToCurrentClosestArray);
        float distanceListenerToPreviousClosestArray = calculate_vector_norm(listenerToPreviousClosestArray);
        return (distanceListenerToPreviousClosestArray > (distanceThreshold * distanceListenerToCurrentClosestArray));
    };

    // Switch to new array, when: distanceToPreviousArray > (distanceThreshold * distanceToNewArray)
    float distanceThreshold = 1.2f;

    // Get candidate for the new closest array
    int newClosestArrayCandidate = calculate_closest_array_based_on_weights();

    // If the current and previous closest arrays are not part of the same triangle, update the closest array
    if (!are_closest_arrays_in_same_triangle(newClosestArrayCandidate)) {
        h->currentClosestMicrophoneArray = newClosestArrayCandidate;
        return;
    }

    // Switch the closest array only after a threshold distance
    if (is_distance_threshold_exceeded(newClosestArrayCandidate, distanceThreshold)) {
        h->currentClosestMicrophoneArray = newClosestArrayCandidate;
    } else {
        h->currentClosestMicrophoneArray = h->previousClosestMicrophoneArray;
    }
}

void save_tracked_values(hoaTranslatorData *h)
{
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->trackedCurrentInterFrameWeights[0][mic * h->nFrames + frame] = h->currentInterFrameWeights[mic][frame];
        }
        h->trackedCurrentInterpolationWeights[0][mic] = h->currentInterpolationWeights[mic];
        h->trackedPreviousInterpolationWeights[0][mic] = h->previousInterpolationWeights[mic];
    }
    h->trackedCurrentTriangleArrays[0] = h->triangleFaces[h->activeTriangleIndex];
    h->trackedEdgeTriangleArrays[0] = h->edgeTriangleArrays;
}

void post_rendering_actions(hoaTranslatorData *h)
{
    // Save values from this frame
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        h->previousInterpolationWeights[mic] = h->currentInterpolationWeights[mic];
    }
    copy(h->currentBinauralMatrix, h->previousBinauralMatrix);
    copy(h->currentM, h->previousM);
    copy(h->currentMr, h->previousMr);
    h->previousListenerPosition = h->currentListenerPosition;

    if (!h->useRecursiveAveraging) {
        copy(h->newCxMatrix, h->previousCxMatrix);
        copy(h->newCyMatrix, h->previousCyMatrix);
    }

    // Stop crossfade, when all the frequency bands have been processed
    check_for_crossfade_stop(h);

    // Save closest mic array
    if (!h->isCrossfadeInProgress) {
        h->previousClosestMicrophoneArray = h->currentClosestMicrophoneArray;
    }

    // Shift the tracked values by one frame
    for (int i = h->nTotalFramesTracked - 1; i > 0; --i) {
        h->trackedCurrentTriangleArrays[i] = h->trackedCurrentTriangleArrays[i - 1];
        h->trackedEdgeTriangleArrays[i] = h->trackedEdgeTriangleArrays[i - 1];
        h->trackedCurrentInterFrameWeights[i] = h->trackedCurrentInterFrameWeights[i - 1];
        h->trackedCurrentInterpolationWeights[i] = h->trackedCurrentInterpolationWeights[i - 1];
        h->trackedPreviousInterpolationWeights[i] = h->trackedPreviousInterpolationWeights[i - 1];
    }

    // Empty the tracked edge triangle arrays
    if (h->useImprovedExteriorRendering && !h->isListenerProjectedToClosestEdge) {
        for (int i = 0; i < h->nTotalFramesTracked; ++i) {
            std::fill(h->trackedEdgeTriangleArrays[i].begin(), h->trackedEdgeTriangleArrays[i].end(), 0);
        }
    }
}

void determine_active_arrays(hoaTranslatorData *h)
{
    h->transformedArraysForCurrentFrameFOA.clear();
    h->transformedArraysForCurrentFrameFullOrder.clear();
    h->activeArraysForCurrentFrame.clear();
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        bool isMicArrayTransformedFOA = false;
        bool isMicArrayTransformedFullOrder = false;
        bool isMicArrayActive = false;
        if (mic == h->currentClosestMicrophoneArray) {
            isMicArrayTransformedFullOrder = true;
        }
        for (int i = 0; i < h->nTotalFramesTracked; ++i) {
            // Check if mic array is found in the track record of current and (nTotalFramesTracked - 1) previous frames, and
            // do processing only for those arrays
            if (std::find(h->trackedCurrentTriangleArrays[i].begin(), h->trackedCurrentTriangleArrays[i].end(), mic) != h->trackedCurrentTriangleArrays[i].end()) {
                if (!isMicArrayTransformedFullOrder) {
                    isMicArrayTransformedFOA = true;
                }
                break;
            }
            // If the listener is outside the triangles and projected between two triangles, check the track record of the arrays in both edge triangles
            if (h->isListenerProjectedToClosestEdge && !h->isClosestEdgePartOfTriangle) {
                if (std::find(h->trackedEdgeTriangleArrays[i].begin(), h->trackedEdgeTriangleArrays[i].end(), mic) != h->trackedEdgeTriangleArrays[i].end()) {
                     if (!isMicArrayTransformedFullOrder) {
                        isMicArrayTransformedFOA = true;
                    }
                    if (mic == h->edgeArrays[0] || mic == h->edgeArrays[1]) {
                        // The metadata values are only needed for the edge arrays
                        isMicArrayActive = true;
                    }
                    break;
                }
            }
        }
        // If crossfade is in progress, check also the arrays part of the crossfade
        if (h->isCrossfadeInProgress) {
            if (mic == h->previousClosestMicrophoneArray) {
                isMicArrayTransformedFullOrder = true;
                // get ready for use after cross-fade
                isMicArrayTransformedFOA = true;
            }
            if (mic == h->currentClosestMicrophoneArray) {
                isMicArrayTransformedFullOrder = true;
                // FOA STFT needed until FullOrder is ready
                isMicArrayTransformedFOA = true;
            }

            if (std::find(h->arraysPartOfCrossfade.begin(), h->arraysPartOfCrossfade.end(), mic) != h->arraysPartOfCrossfade.end()) {
                if (!isMicArrayTransformedFullOrder) {
                    isMicArrayTransformedFOA = true;
                }

                isMicArrayActive = true;
            }
        }

        // Get spatial metadata only for the current active triangle
        if (!isMicArrayActive && (std::find(h->currentChosenTriangleArrays.begin(), h->currentChosenTriangleArrays.end(), mic) != h->currentChosenTriangleArrays.end())) {
            isMicArrayActive = true;
        }

        if (isMicArrayTransformedFullOrder) {
            h->transformedArraysForCurrentFrameFullOrder.push_back(mic);
        }
        if (isMicArrayTransformedFOA) {
            h->transformedArraysForCurrentFrameFOA.push_back(mic);
        }
        if (isMicArrayActive) {
            h->activeArraysForCurrentFrame.push_back(mic);
        }
    }

    // Add source aware mic arrays to the transformed arrays list
    for (auto const& mics : h->trackedSourceMicArrays) {
        for (auto const& mic : mics) {
            if (std::find(h->transformedArraysForCurrentFrame.begin(), h->transformedArraysForCurrentFrame.end(), mic) == h->transformedArraysForCurrentFrame.end()) {
                h->transformedArraysForCurrentFrameFullOrder.push_back(mic);
            }
        }
    }
    // Add source aware active mics to the active arrays list
    for (auto const& src : h->activeSources) {
        int mic = h->closestMicsForSources[src];
        if (std::find(h->activeArraysForCurrentFrame.begin(), h->activeArraysForCurrentFrame.end(), mic) == h->activeArraysForCurrentFrame.end()) {
            h->activeArraysForCurrentFrame.push_back(mic);
        }
    }
}

void choose_active_triangle_from_track_record(hoaTranslatorData *h)
{
    h->isTriangleSwitched = false;
    // Choose the active mic array triangle and the corresponding interpolation
    // weights from a set of tracked frames
    for (int i = 0; i < h->nTotalFramesTracked; ++i) {
        if (h->currentChosenTriangleArrays == h->trackedCurrentTriangleArrays[i]) {
            if (h->isListenerProjectedToClosestEdge && !h->isClosestEdgePartOfTriangle && (h->edgeTriangleArrays != h->trackedEdgeTriangleArrays[h->nTotalFramesTracked-1])) {
                // The listener is outside the triangles, and projected between two triangles. The arrays in the edge
                // triangles have not been processed long enough for the STFT buffers to flush. For the weights, use the last
                // ones, that were calculated before the projection.
                for (int n = 0; n < h->nTotalFramesTracked; ++n) {
                    if (h->edgeTriangleArrays != h->trackedEdgeTriangleArrays[n]) {
                        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
                            for (int frame = 0; frame < h->nFrames; ++frame) {
                                h->chosenInterFrameWeights[mic][frame] = h->trackedCurrentInterFrameWeights[n][mic * h->nFrames + h->nFrames - 1];
                            }
                        }
                        break;
                    }
                }
            } else {
                // The previously chosen mic array triangle was found in the set of
                // tracked frames, use that
                if (i == 0) {
                    // The listener hasn't moved between triangles, use the most current weights
                    std::copy_n(&h->currentInterFrameWeights[0][0], h->nMicrophoneArrays * h->nFrames, &h->chosenInterFrameWeights[0][0]);
                } else {
                    // The listener has recently moved to another triangle, use the last available weights from the previous triangle
                    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
                        for (int frame = 0; frame < h->nFrames; ++frame) {
                            h->chosenInterFrameWeights[mic][frame] = h->trackedCurrentInterFrameWeights[i][mic*h->nFrames + h->nFrames-1];
                        }
                    }
                }
            }
            // Update fadeOut and fadeIn weights for crossfade
            if (h->isCrossfadeInProgress && !h->isTriangleSwichedDuringCrossfade) {
                if (!h->isCrossfadeBetweenTriangles) {
                    std::copy_n(&h->chosenInterFrameWeights[0][0],h->nMicrophoneArrays * h->nFrames,&h->fadeOutWeights[0][0]);
                }
                std::copy_n(&h->chosenInterFrameWeights[0][0],h->nMicrophoneArrays * h->nFrames,&h->fadeInWeights[0][0]);
            }
            return;
        }
    }
    // The previously chosen triangle was not found in the set of tracked
    // frames, update the chosen triangle to the last one tracked, and
    // start the triangle switch
    if (!h->trackedCurrentTriangleArrays[h->nTotalFramesTracked - 1].empty()) {
        h->isTriangleSwitched = true;
        if (h->isCrossfadeInProgress) {
            h->isTriangleSwichedDuringCrossfade = true;
        }
        h->previousChosenTriangleArrays = h->currentChosenTriangleArrays;
        h->currentChosenTriangleArrays = h->trackedCurrentTriangleArrays[h->nTotalFramesTracked - 1];

        // Linearly interpolate the weights between the current and previous triangle (use the weights from the edge of the
        // two triangles, i.e. the last weights in the old triangle, and the first weights in the new triangle)
        auto currentTriangleWeights = h->trackedCurrentInterpolationWeights[h->nTotalFramesTracked - 1];
        auto previousTriangleWeights = h->trackedPreviousInterpolationWeights[h->nTotalFramesTracked - 1];
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            for (int frame = 0; frame < h->nFrames; ++frame) {
                h->chosenInterFrameWeights[mic][frame] = h->interpolator[frame] * currentTriangleWeights[mic]
                                                         + (1 - h->interpolator[frame]) * previousTriangleWeights[mic];
            }
        }
        std::copy_n(&currentTriangleWeights[0], h->nMicrophoneArrays, &h->currentInterpolationWeights[0]);
        std::copy_n(&previousTriangleWeights[0], h->nMicrophoneArrays, &h->previousInterpolationWeights[0]);
        std::copy_n(&h->currentInterpolationWeights[0], h->nMicrophoneArrays, &h->trackedCurrentInterpolationWeights[0][0]);
        std::copy_n(&h->previousInterpolationWeights[0], h->nMicrophoneArrays, &h->trackedPreviousInterpolationWeights[0][0]);
        h->trackedCurrentTriangleArrays[0] = h->currentChosenTriangleArrays;

    } else {
        // The last tracked triangle was empty, set the chosen triangle as
        // the current one
        h->currentChosenTriangleArrays = h->trackedCurrentTriangleArrays[0];
        std::copy_n(&h->currentInterFrameWeights[0][0], h->nMicrophoneArrays * h->nFrames, &h->chosenInterFrameWeights[0][0]);
    }
}

void openGL_to_cartesian_xFront(const float *openGLvec, float *cartesianVec)
{
    cartesianVec[0] = -openGLvec[2];
    cartesianVec[1] = -openGLvec[0];
    cartesianVec[2] =  openGLvec[1];
}

void openGL_to_cartesian_xFront(const std::vector<float> &openGLvec, std::vector<float> &cartesianVec)
{
    assert(openGLvec.size() == 3 && cartesianVec.size() == 3);
    cartesianVec[0] = -openGLvec[2];
    cartesianVec[1] = -openGLvec[0];
    cartesianVec[2] =  openGLvec[1];
}

bool AnearB(float A, float B, float errorMargin)
{
    return std::abs(A - B) < errorMargin;
}

std::shared_ptr<std::vector<std::uint8_t>> read_file_to_vector(std::string const& filepath) {
    std::ifstream stream;
    stream.open(filepath, std::ios_base::in | std::ios_base::binary | std::ios::ate);
    std::size_t const size = stream.is_open() ? static_cast<std::size_t>(stream.tellg()) : 0;
    auto data = std::make_shared<std::vector<std::uint8_t>>(size);
    stream.seekg(0);
    stream.read(reinterpret_cast<char*>(data->data()), size);

    if (!stream) {
        data.reset();
    }

    return data;
}

void combine_covariance_matrices(hoaTranslatorData *h,
                                 hoatMatrixCmplx2D &targetM,
                                 const hoatMatrixCmplx2D &previousM,
                                 hoatMatrixCmplx2D &newM)
{
    assert(targetM.nRows == previousM.nRows && targetM.nRows == newM.nRows && targetM.nCols == previousM.nCols && targetM.nCols == newM.nCols);
    if (h->useRecursiveAveraging) {
        combine_recursive_covariance_matrices(h, targetM, newM);
    } else {
        matrixMatrixAdd(previousM, newM, targetM);
    }
}

void combine_recursive_covariance_matrices(hoaTranslatorData *h,
                                           hoatMatrixCmplx2D &targetM,
                                           hoatMatrixCmplx2D &newM)
{
    matrixScalarMult(h->recursionSmoothingCoefficient, targetM, targetM);
    matrixScalarMult((1.0f - h->recursionSmoothingCoefficient), newM, newM);
    matrixMatrixAdd(targetM, newM, targetM);
}

void copy_hrtf_to_matrix(hoaTranslatorData *h, int band, int hrtfDirIndex) {
    for (int ch = 0; ch < h->nBinauralChannels; ++ch) {
        h->tempHrtfMatrix[ch][0] = h->hrtfFB[band][ch][hrtfDirIndex];
    }
}

int getHrtfDirIndex(hoaTranslatorData *h, float aziRad, float elevRad) {
    int aziIndex = std::round(matlab_fmodf(aziRad * 180.0f / M_PI + 180.0f, 360.0f) / h->interpolatedHrtfAziResDeg);
    int elevIndex = std::round(matlab_fmodf(elevRad * 180.0f / M_PI + 90.0f, 180.0f) / h->interpolatedHrtfElevResDeg);
    int dirIndex = elevIndex * h->nInterpolatedHrtfAziAngles + aziIndex;  // [(azi0 elev0) (azi1 elev0)...]
    return h->hrtfIdxGrid[dirIndex];
}

void compute_direct_part_of_Cy(hoaTranslatorData *h, int frame, int band, int dirIndex, float energy, float dtr, hoatMatrixCmplx3D &Cy) {
    float multiplier = dtr * energy;
    if (h->useSphoa6DofExtensionMode) {
        multiplier *= std::pow(h->singlePoint6DofDirGains[frame][band], 2.0f);
    }
    copy_hrtf_to_matrix(h, band, dirIndex);
    matrixMultNoTransConj(h->tempHrtfMatrix, h->tempHrtfMatrix, h->tempHrtfProductMatrix);
    matrixScalarMult(multiplier, h->tempHrtfProductMatrix, h->tempHrtfProductMatrix);
    matrixMatrixAdd(Cy[band], h->tempHrtfProductMatrix, Cy[band]);
}

void compute_diffuse_Cy_exterior(hoaTranslatorData *h, int frame, int band, float energy, float dtr, hoatMatrixCmplx3D &Cy) {
    // Exterior rendering as: diffusePart = (1 - dtr) * signalEnergy * weightedDiffuseCovarianceMatrix,
    // where weightedDiffusCovarianceeMatrix = (w3 * DCMdiff) + (w4 * DCMambi).
    // DCMdiff is computed from the diffuse covariance matrix, Cdiffuse.
    // DCMambi is computed from the prototype binaural signal.

    // Compute weighted DCMdiff
    float w3 = h->diffuseExteriorWeight;
    matrixScalarMult(w3, h->Cdiffuse[band], h->DCMdiff);

    // Compute weighted DCMambi
    float w4 = 1.0f - w3;
    float protoEnergyCh0 = std::norm(h->currentBinauralMatrix[band][0][frame]);
    float protoEnergyCh1 = std::norm(h->currentBinauralMatrix[band][1][frame]);
    float chEnergyBoth = protoEnergyCh0 + protoEnergyCh1 + FLT_MIN;
    h->DCMambi[0][0] = (w4 * protoEnergyCh0) / chEnergyBoth;
    h->DCMambi[1][1] = (w4 * protoEnergyCh1) / chEnergyBoth;

    // Form DCM from the diffuse signal energy, DCMdiff and DCMambi
    matrixMatrixAdd(h->DCMdiff, h->DCMambi, h->DCM);
    float DCMtrace = sumDiagonalReal(h->DCM);
    matrixScalarMult(((1.0f - dtr) * energy) / DCMtrace, h->DCM, h->DCM);

    // Add DCM to newCy
    matrixMatrixAdd(Cy[band], h->DCM, Cy[band]);
}

void compute_diffuse_Cy_interior(hoaTranslatorData *h, int frame, int band, float energy, float dtr, hoatMatrixCmplx3D &Cy) {
    if (h->diffuseMethodForTargetSCM == DIFFUSE_COVARIANCE_MATRIX || h->useSourceAwareMode) {
        // diffusePart = (1 - dtr) * signalEnergy * diffuseMatrix
        matrixScalarMult(((1.0f - dtr) * energy), h->Cdiffuse[band], h->tempMatrixOutChSqrd);
        matrixMatrixAdd(Cy[band], h->tempMatrixOutChSqrd, Cy[band]);
    } else {
        // diffusePart = (1 - dtr) * signalEnergy * prototypeEnergySingleChannel / prototypeEnergyBothChannels
        float protoEnergyCh0 = std::norm(h->currentBinauralMatrix[band][0][frame]);
        float protoEnergyCh1 = std::norm(h->currentBinauralMatrix[band][1][frame]);
        float chEnergyBoth = protoEnergyCh0 + protoEnergyCh1 + FLT_MIN;
        Cy[band][0][0] += ((1.0f - dtr) * energy * protoEnergyCh0 / chEnergyBoth);
        Cy[band][1][1] += ((1.0f - dtr) * energy * protoEnergyCh1 / chEnergyBoth);
    }
}

void compute_diffuse_part_of_Cy(hoaTranslatorData *h, int frame, int band, float energy, float dtr, hoatMatrixCmplx3D &Cy) {
    if (!h->useSphoa6DofExtensionMode && !h->isListenerInsideTriangle && h->useImprovedExteriorRendering && !h->useSourceAwareMode) {
        compute_diffuse_Cy_exterior(h, frame, band, energy, dtr, Cy);
    } else {
        compute_diffuse_Cy_interior(h, frame, band, energy, dtr, Cy);
    }
}