/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef HOA_TRANSLATION_UTILITIES_HPP
#define HOA_TRANSLATION_UTILITIES_HPP

#include <cstdint>
#include <fstream>
#include <memory>
#include <string>
#include <vector>

#include "ByteStream.hpp"
#include "HoatMatrix.hpp"
#include <saf.h>

// Different options for calculating the diffuse part of the target signal covariance matrix
enum DIFFUSE_METHOD_FOR_TARGET_SCM {
    // Estimate a diffuse-field covariance matrix from an equally distributed set of HRTFs. Use the
    // matrix to calculate the diffuse part as:
    //      diffPart = signalEnergy * (1 - dtr) * diffuseMatrix
    DIFFUSE_COVARIANCE_MATRIX,

    // Use the energy from the binaural prototype signal to calculate the diffuse part as:
    //      diffPart = signalEnergy * (1 - dtr) * prototypeEnergySingleChannel / prototypeEnergyBothChannels
    PROTOTYPE_ENERGY
};

// In source aware mode, which target SCMs (Cy) are used
enum SOURCE_RESIDUAL_MODE {
    // Use only the SCMs derived from the source model
    SOURCE_ONLY,

    // Use only the residual SCMs (= original mic - source model)
    RESIDUAL_ONLY,

    // Use both source model and residual SCMs
    BOTH
};

// In source aware mode, determine how the target SCM (Cy) is formed
enum TARGET_SCM_MODE {

    // First combine the source and residual parameters, and then calculate the target SCM (Cy) from the combined parameters
    COMPOUND_PARAMS,

    // Build target SCM (Cy) by combining separate source and residual SCMs together
    SEPARATE_SCM
};

struct inputFilesAndFlags {
    std::string matlabDirsFile;
    std::string matlabHrtfFile;
    std::string matlabItdFile;
    std::string sofaFile;
    std::string spatialMetadataBinaryFile;
    bool useMatlabFile{ false };
};

struct optimalMixingParameters {
    int lowerBandFreqLimit;
    int upperBandFreqLimit;
    bool useDiffuseDecorrelation;
    bool useEqualizer;
    bool useRecursiveAveraging;
    float recursionSmoothingCoefficient;
};

struct exteriorRenderingParameters {
    bool useImprovedExteriorRendering;
    float exteriorProjectionRadius;
    bool adaptiveExteriorProjectionRadius;
};

struct sphoa6DofExtensionParameters {
    bool useSphoa6DofExtension;
    float sphoa6DofMaxGainDb;
    float sphoa6DofProjectionRadius;

    // NthSphoaInEncodedMetadata can be used to read encoded metadata for a single HOA source from a metadata file
    // containing data for multiple HOAs. Range [1 N], where N is the number of HOAs in the metadata file (assuming only
    // single HOAGroup). For example, NthSphoaInEncodedMetadata = 2 reads the metadata for the second HOA in a metadata
    // file.
    int NthSphoaInEncodedMetadata;
};

struct sourceAwareParameters {
    bool useSourceAwareMode;
    bool useSourceAwarePostFilter;
    bool useSourceAwareTimeAlignment;
    SOURCE_RESIDUAL_MODE sourceResidualMode;
    TARGET_SCM_MODE  targetScmMode;
    std::vector<std::vector<float>> sourcePositionsOpenGL;
    int maxSimulSources;
};

struct sourceAwareParametersPerSource {
    bool listenerThresholdPresent;
    bool hoaSourceThresholdPresent;
    float listenerThreshold;
    float hoaSourceThreshold;
    int priorityValue;
};

/* Positions as OpenGL:
 *      - x (right)
 *      - y (up)
 *      - z (back)
 * Orientations as OpenGL:
 *      - yaw (rotation around y-axis)
 *      - pitch (rotation around x-axis)
 *      - roll (rotation around z-axis)
 */
struct micArrayPositionsAndOrientations {
    std::vector<std::vector<float>> arrayPositionsOpenGL;
    std::vector<std::vector<float>> arrayOrientationsOpenGL;
};

struct decoderData {
    // 6DoFBitStream()
    std::uint32_t audio_frame_duration;
    std::uint32_t subframes_duration;
    std::uint32_t num_frequency_bands;
    std::uint8_t useRecursiveAveraging;
    std::uint32_t sampling_rate;
    std::uint8_t num_hoa_groups;
    std::uint64_t num_frames;
    std::uint64_t azimuth_multiplier;
    std::uint64_t elevation_multiplier;
    std::uint64_t dtr_multiplier;
    std::uint64_t energy_multiplier;
    std::uint64_t position_multiplier;
    std::uint64_t rotation_multiplier;
    std::uint64_t exterior_multiplier;
    std::uint16_t exterior_projection_radius;
    std::uint8_t is_audio_uneven;
    std::size_t bitStreamHeaderSize;

    std::uint8_t subframes_per_frame;

    int lastReadFrame;

    // InfoFrame()
    std::size_t infoFrameSize;

	// SpatialMetadataFrame()
    std::size_t sourceMetadataFrameSize;
    std::int32_t hoa_render_meta_azimuth;
    std::int32_t hoa_render_meta_elevation;
    std::uint16_t direct_to_total_energy_ratio;
    std::uint32_t overall_energy;

    // HOAGroupData()
    std::vector<std::uint8_t> hoa_source_group_id_lengths;  // 1 x num_hoa_groups
    std::vector<std::string> hoaSourceGroupIdStrs;       // 1 x num_hoa_groups
    std::vector<std::uint8_t> num_hoa_sources_in_group; // 1 x num_hoa_groups
    std::vector<std::size_t> groupDataHeaderSizes;      // 1 x num_hoa_groups
	std::vector<std::size_t> groupMetadataFrameSizes; 

    // HOASourceInformationStruct()
    std::vector<std::vector<std::uint8_t>> hoa_source_id_lengths;  // num_hoa_groups x num_hoa_sources_in_group
    std::vector<std::vector<std::string>> hoaSourceIdStrs;               // num_hoa_groups x num_hoa_sources_in_group

    // HOASourcePositionStruct()
    std::vector<std::vector<std::int32_t>> hoa_sources_pos_x;       // num_hoa_groups x num_hoa_sources_in_group
    std::vector<std::vector<std::int32_t>> hoa_sources_pos_y;       // num_hoa_groups x num_hoa_sources_in_group
    std::vector<std::vector<std::int32_t>> hoa_sources_pos_z;       // num_hoa_groups x num_hoa_sources_in_group
    std::vector<std::vector<std::int32_t>> hoa_sources_rot_yaw;     // num_hoa_groups x num_hoa_sources_in_group
    std::vector<std::vector<std::int32_t>> hoa_sources_rot_pitch;   // num_hoa_groups x num_hoa_sources_in_group
    std::vector<std::vector<std::int32_t>> hoa_sources_rot_roll;    // num_hoa_groups x num_hoa_sources_in_group

    // HOASourceData()
    std::vector<std::vector<std::uint8_t>> hoa_orders;  // num_hoa_groups x num_hoa_sources_in_group
};

struct hoaTranslatorData {

    float diffuseExteriorWeight;
    float distToClosestArraySquared;
    float distToClosestEdgeSquared;
    float exteriorProjectionRadius;
    bool adaptiveExteriorProjectionRadius;
    float interpolationEqGainMax;         // default 4.0f
    float maxTriangleEdgeLength;
    float pitchDeg;                       // head rotation pitch (around right-axis)
    float recursionSmoothingCoefficient;
    float rollDeg;                        // head rotation roll (around back-axis)
    float sphoa6DofMaxGain;
    float sphoa6DofProjectionRadius;
    float yawDeg;                         // head rotation yaw (around up-axis)

    float *DTRDataAvg{};                  // nBands
    float *M_ESDtoHOA{};                  // ESD to HOA conversion matrix, size: flat (shOrder+1)^2 * nVirtualESDSpeakerDirs
    float *azimuthDataAvg{};              // nBands
    float *barycentricCoordinates{};      // 1 x 3

    float *cosAziNbands{};                // nBands
    float *cosElevNbands{};               // nBands
    float *cosAzi{};                      // h->nMicrophoneArrays * h->nFrames * h->nBands
    float *cosElev{};                     // h->nMicrophoneArrays * h->nFrames * h->nBands

    float *currentInterpolationWeights{}; // 1 x nMicrophoneArrays
    float *dotXYZplus1{};                 // nBands
    float *dotX{};                        // nBands
    float *dotY{};                        // nBands
    float *dotZ{};                        // nBands
    float *elevationDataAvg{};            // nBands
    float *energyDataAvg{};               // nBands
    float *freqVec{};                     // 1 x nBands
    float *hrirs{};                       // nHrtfDirs * nBinauralChannels * hrirLen
    float *hrtfDirsDeg{};                 // nHrtfDirs * nBinauralChannels
    float *inputSignalBuffer{};           // frameSize * channelsPerArrays
    float *interpolator{};                // 1 x nFrames
    float *itds{};                        // 1 x nHrtfDirs

    float *originalDoaUnitX{};            // nBands
    float *originalDoaUnitY{};            // nBands
    float *originalDoaUnitZ{};            // nBands
    float *outputSignalBuffer{};          // frameSize * outChannels
    float *previousInterpolationWeights{};// 1 x nMicrophoneArrays
    float *reMappedDoaX{};                // nBands
    float *reMappedDoaY{};                // nBands
    float *reMappedDoaZ{};                // nBands
    float *reMappedDtr{};                 // nBands
    float *sinAziNbands{};                // nBands
    float *sinElevNbands{};               // nBands
    float *tempNbands1{};                 // nBands
    float *tempNbands2{};                 // nBands

    float *sinAzi{};                      // h->nMicrophoneArrays * h->nFrames * h->nBands
    float *sinElev{};                     // h->nMicrophoneArrays * h->nFrames * h->nBands

    float *tempSHrotationMatrix{};        // (shOrder+1)^2 * (shOrder+1)^2
    float *w1{};                          // nBands
    float *w2{};                          // nBands

    float **chosenInterFrameWeights{};    // nMicrophoneArrays x nFrames
    float **currentInterFrameWeights{};   // nMicrophoneArrays x nFrames
    float **eqGains{};                    // nFrames x nBands
    float **fadeInWeights{};              // nMicrophoneArrays x nFrames
    float **fadeOutWeights{};             // nMicrophoneArrays x nFrames
    float **intensityVectorAvg{};         // nBands x 3
    float **interpolatedAzimuthData{};    // nFrames x nBands
    float **interpolatedDTRData{};        // nFrames x nBands
    float **interpolatedElevationData{};  // nFrames x nBands
    float **interpolatedEnergyData{};     // nFrames x nBands
    float **metadataV1{};                 // nFrames x nBands
    float **metadataV2{};                 // nFrames x nBands
    float **metadataV3{};                 // nFrames x nBands
    float **microphoneArrayPositions{};   // nMicrophoneArrays x 3 (as OpenGL: (x right, y up, z back))
    float **singlePoint6DofDirGains{};    // nFrames x nBands
    float **timeDomainHopFrame{};         // channelsPerArray x hopSize
    float **timeDomainOutputFrame{};      // outChannels x frameSize
    float **transientDetector1{};         // nBands x outChannels
    float **transientDetector2{};         // nBands x outChannels

    float ***azimuthData{};               // nMicrophoneArrays x nFrames x nBands
    float ***DTRData{};                   // nMicrophoneArrays x nFrames x nBands
    float ***elevationData{};             // nMicrophoneArrays x nFrames x nBands
    float ***energyData{};                // nMicrophoneArrays x nFrames x nBands
    float ***tempV1{};                    // nMicrophoneArrays x nFrames x nBands
    float ***tempV2{};                    // nMicrophoneArrays x nFrames x nBands
    float ***tempV3{};                    // nMicrophoneArrays x nFrames x nBands
    float ***timeDomainInputFrame{};      // nMicrophoneArrays x channelsPerArray x frameSize
    float ***weightedEnergy{};            // nMicrophoneArrays x nFrames x nBands

    bool isClosestEdgeFound;
    bool isClosestEdgePartOfTriangle{ false };
    bool isCrossfadeBetweenTriangles{ false };
    bool isCrossfadeInProgress{ false };
    bool isEdgeFlipped;
    bool isListenerInsideOmittedTriangle{ false };
    bool isListenerInsideTriangle{ false };
    bool isListenerProjectedToClosestArray{ false };
    bool isListenerProjectedToClosestEdge{ false };
    bool isTriangleSwichedDuringCrossfade{ false };
    bool isTriangleSwitched{ false };
    bool omitLongTriangleEdges{ true };
    bool useComplexityReduction{ false };
    bool useDiffuseDecorrelation;
    bool useEncodedValues;
    bool useEqualizer;
    bool useImprovedExteriorRendering;
    bool useMatlabFile;
    bool useMphoa1SourceMode{ false };
    bool useMphoa2SourceMode{ false };
    bool useRecursiveAveraging;
    bool useSphoa6DofExtensionMode;

    int HOAGroupId;                     // Which HOAGroup this translator corresponds to, used when reading the encoded metadata file
    int LDmode;                         // low-delay mode for afSTFT
    int NthSphoaInEncodedMetadata;
    int activeTriangleIndex;
    int channelsPerArray;
    int closestArrayByDistance;
    int crossFadeProgressIndx{ 0 };
    int crossfadeLen;
    int currentAudioLoop{ 0 };
    int currentClosestMicrophoneArray;
    int currentExteriorEdgeIndex;
    int currentExteriorMicIndex;
    int currentFrame;
    int frameSize;
    int hopSize;                        // default 128
    int hrirFs;
    int hrirLen;
    int hybridMode;                     // hybrid mode for afSTFT
    int interpolatedHrtfAziResDeg;      // default 2 degrees
    int interpolatedHrtfElevResDeg;     // default 5 degrees
    int lowAvgLimitIndex;
    int lowerBandFreqLimit;             // default 4000 Hz, used in optimal mixing
    int maxFreqLimitIndex;
    int nAnalysisChannels;              // default 4
    int nBands;
    int nBinauralChannels{ 2 };
    int nBlocksInDecorrelationBuffer;
    int nBridgeEdges;
    int nComplexityReducedBands;
    int nDoaDtrBins;
    int nExteriorEdges;
    int nFrames;                        // frameSize / hopSize
    int nHrtfDirs;
    int nInterpolatedHrtfAziAngles;     // 360/hrtfAziResDeg + 1
    int nInterpolatedHrtfDirs;
    int nInterpolatedHrtfElevAngles;    // 180/hrtfElevResDeg + 1
    int nMicrophoneArrays;
    int nSamplesInDecorrelationBuffer;
    int nTotalFramesTracked;            // current + (nTotalFramesTracked - 1) previous frames, that are being tracked
    int nTriangles;
    int outChannels;
    int previousClosestMicrophoneArray;
    int sampleRate;                     // default 48000
    int shOrder;

    int *decorrelationDelays{};         // nBands * outChannels
    int *hrtfIdxGrid{};                 // N directions in a grid of points

    int framesUntilCrossFadeStart;

    // In trackedXXX 2D vectors, the 0th vector corresponds to the current frame, the one after that to the previous frame, etc.
    std::vector<std::vector<int>> bridgeEdges{};                            // nBridgeEdges x 2
    std::vector<std::vector<int>> edgesBetweenArrays{};                     // N x 2 (N is the number of edges that have a length < maxTriangleEdgeLength)
    std::vector<std::vector<int>> exteriorEdges{};                          // nExteriorEdges x 2
    std::vector<std::vector<int>> fadeOutBands{};                           // nFrames x nBands
    std::vector<std::vector<int>> omittedEdgesBetweenArrays{};              // N x 2 (N is the number of omitted edges)
    std::vector<std::vector<int>> omittedExteriorEdges{};                   // N omitted exterior edges x 2
    std::vector<std::vector<int>> omittedTriangleFaces{};                   // N x 3 (N is the number of omitted triangles)
    std::vector<std::vector<int>> originalExteriorEdges{};                  // N original exterior edges x 2
    std::vector<std::vector<int>> trackedCurrentTriangleArrays{};           // nTotalFramesTracked x 3
    std::vector<std::vector<int>> trackedEdgeTriangleArrays{};              // nTotalFramesTracked x 6 (all the arrays from two triangles connected with an edge)
    std::vector<std::vector<int>> triangleFaces{};                          // nTriangles x 3

    std::vector<std::vector<float>> edgeNormals{};                          // nExteriorEdges x 3
    std::vector<std::vector<float>> micArrayOrientations{};                 // nMicrophoneArrays x 3 (yaw, pitch, roll in degrees)
    std::vector<std::vector<float>> micNormals{};                           // nExteriorEdges x 3
    std::vector<std::vector<float>> originalEdgeNormals{};                  // N original exterior edges x 3
    std::vector<std::vector<float>> trackedCurrentInterFrameWeights{};      // nTotalFramesTracked x (nMicrophoneArrays * nFrames)
    std::vector<std::vector<float>> trackedCurrentInterpolationWeights{};   // nTotalFramesTracked x nMicrophoneArrays
    std::vector<std::vector<float>> trackedPreviousInterpolationWeights{};  // nTotalFramesTracked x nMicrophoneArrays

    std::vector<int> arraysPartOfCrossfade{};                               // 6 mic arrays from previous and current triangle, that are being crossfaded
    std::vector<int> bottomBandIndexes{};                                   // nComplexityReducedBands
    std::vector<int> crossfadeBandHigh{};                                   // crossfadeLen x 1
    std::vector<int> crossfadeBandLow{};                                    // crossfadeLen x 1
    std::vector<int> crossfadeBandSwaporder{};                              // crossfadeLen x 1
    std::vector<int> currentChosenTriangleArrays{0, 1, 2};                  // 3 mic arrays in the triangle
    std::vector<int> edgeArrays{0, 0};                                      // 2 arrays connected with an edge
    std::vector<int> edgeTriangleArrays{0, 0, 0, 0, 0, 0};                  // all 6 arrays from both triangles connected with an edge (if edge between triangles)
    std::vector<int> exteriorMicArrays{};                                   // mic arrays part of the exterior hull
    std::vector<int> nonConvexMicArrays{};                                  // mic arrays, that fall inside non-convex hulls
    std::vector<int> previousChosenTriangleArrays{0, 1, 2};                 // 3 mic arrays in the triangle
    std::vector<int> upperBandIndexes{};                                    // nComplexityReducedBands

    std::vector<float> currentListenerPosition{0.0f, 0.0f, 0.0f};           // 1 x 3 (as OpenGL: (x right, y up, z back))
    std::vector<float> edgeWeights{0.0f, 0.0f};                             // weights for 2 arrays connected with an edge
    std::vector<float> eqGainsMax{};                                        // nFrames * nBands
    std::vector<float> upperBandFrequencies{};                               // nComplexityReducedBands
    std::vector<float> listenerExteriorNormal{0.0f, 0.0f, 0.0f};            // 1 x 3 (as OpenGL: (x right, y up, z back))
    std::vector<float> previousListenerPosition{0.0f, 0.0f, 0.0f};          // 1 x 3 (as OpenGL: (x right, y up, z back))
    std::vector<float> projectedListenerPosition{0.0f, 0.0f, 0.0f};         // 1 x 3 (as OpenGL: (x right, y up, z back))
    std::vector<float> sourceEnergyGains{};                                 // nMicrophoneArrays

    std::vector<float_complex> tempAnalysisVec{};                           // nAnalysisChannels

    std::vector<std::size_t> activeArraysForCurrentFrame{};                 // all the mic arrays, which are used in the processing (i.e. spatial metadata is needed for these arrays)
    std::vector<std::size_t> transformedArraysForCurrentFrame{};            // all the mic arrays, that are transformed with STFT
    std::vector<std::size_t> transformedArraysForCurrentFrameFullOrder{};   // mic arrays, for which STFT is run for full HOA order
    std::vector<std::size_t> transformedArraysForCurrentFrameFOA{};         // mic arrays, for which STFT is run for up to FOA

    void *afSTFTbackwardHandle{};
    void *covarianceDomainFramework{};

    void **afSTFTforwardHandlesFullOrder{};       // nMicrophoneArrays
    void **afSTFTforwardHandlesFOA{};             // nMicrophoneArrays


    float_complex **tempSignal{};                 // channelsPerArray x nBands

    float_complex ***decorrelatedTFDFrame{};      // nBands x outChannels x nFrames
    float_complex ***eqGainsCmplx{};              // nFrames x channelsPerArray x nBands
    float_complex ***hrtfFB{};                    // nBands x nBinauralChannels x nHrtfDirs
    float_complex ***inputXBlocksPrototype{};     // nBands x outChannels x (nBlocksInDecorrelationBuffer * nFrames)
    float_complex ***interpolatedTFDFrame{};      // nFrames x channelsPerArray x nBands
    float_complex ***outputTFDProcessingFrame{};  // nBands x outChannels x nFrames
    float_complex ***tempTFDFrame{};              // nBands x outChannels x nFrames
    float_complex ***TFDOutputFrame{};            // nFrames x outChannels x nBands

    float_complex ****TFDInputFrameFullOrder{};   // nMicrophoneArrays x nFrames x channelsPerArray x nBands
    float_complex ****TFDInputFrameFOA{};         // nMicrophoneArrays x nFrames x 4 x nBands

    hoatMatrixCmplx2D Cr;                         // 2 x 2
    hoatMatrixCmplx2D CxNormalized;               // 2 x 2
    hoatMatrixCmplx2D CyNormalized;               // 2 x 2
    hoatMatrixCmplx2D CyTilde;                    // 2 x 2
    hoatMatrixCmplx2D DCM;                        // outChannels x outChannels
    hoatMatrixCmplx2D DCMambi;                    // outChannels x outChannels
    hoatMatrixCmplx2D DCMdiff;                    // outChannels x outChannels
    hoatMatrixCmplx2D DsqrtInv;                   // 2 x 2
    hoatMatrixCmplx2D Dtemp;                      // 2 x 2
    hoatMatrixCmplx2D Genergy;                    // 2 x 2
    hoatMatrixCmplx2D GyHatQ;                     // 2 x 2
    hoatMatrixCmplx2D Kx;                         // 2 x 2
    hoatMatrixCmplx2D Ky;                         // 2 x 2
    hoatMatrixCmplx2D KyGQKx;                     // 2 x 2
    hoatMatrixCmplx2D P;                          // 2 x 2
    hoatMatrixCmplx2D Sx;                         // 2 x 2
    hoatMatrixCmplx2D SxSqrt;                     // 2 x 2
    hoatMatrixCmplx2D Utemp;                      // 2 x 2
    hoatMatrixCmplx2D Ux;                         // 2 x 2
    hoatMatrixCmplx2D temp2x2Cmplx2;              // 2 x 2
    hoatMatrixCmplx2D temp2x2Cmplx;               // 2 x 2
    hoatMatrixCmplx2D tempAnalysisMatrix;         // nAnalysisChannels x 1
    hoatMatrixCmplx2D tempBinauralMatrix;         // outChannels x 1
    hoatMatrixCmplx2D tempHrtfMatrix;             // nBinauralChannels x 1
    hoatMatrixCmplx2D tempHrtfProductMatrix;      // nBinauralChannels x nBinauralChannels
    hoatMatrixCmplx2D tempInputMatrix;            // channelsPerArray x nFrames
    hoatMatrixCmplx2D tempInputMatrixSmall;       // channelsPerArray x 1
    hoatMatrixCmplx2D tempMatrixOutChSqrd;        // outChannels x outChannels
    hoatMatrixCmplx2D tempMixingMatrix;           // outChannels x outChannels
    hoatMatrixCmplx2D tempOutputMatrix;           // outChannels x 1
    hoatMatrixCmplx2D tempRotatedMatrix;          // channelsPerArray x nFrames
    hoatMatrixCmplx2D tempRotatedMatrixSmall;     // channelsPerArray x 1
    hoatMatrixCmplx2D tempSignalMatrix;           // outChannels x 1

    hoatMatrixCmplx3D ambi2binDecodingMatrix;     // nBands x (outChannels x channelsPerArray)
    hoatMatrixCmplx3D Cdiffuse;                   // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D currentBinauralMatrix;      // nBands x (outChannels x nFrames)
    hoatMatrixCmplx3D currentM;                   // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D currentMr;                  // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D Cx;                         // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D Cy;                         // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D CyLinearDecoding;           // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D newCxMatrix;                // nBands x (outChannels x outChannels)
    hoatMatrixCmplx3D newCyMatrix;                // nBands x (outChannels x outChannels)
    hoatMatrixCmplx3D previousBinauralMatrix;     // nBands x (outChannels x nFrames)
    hoatMatrixCmplx3D previousCxMatrix;           // nBands x (outChannels x outChannels)
    hoatMatrixCmplx3D previousCyMatrix;           // nBands x (outChannels x outChannels)
    hoatMatrixCmplx3D previousM;                  // nBands x (outChannels * outChannels)
    hoatMatrixCmplx3D previousMr;                 // nBands x (outChannels * outChannels)

    hoatMatrix2D Aimag;                           // 2 x 2
    hoatMatrix2D Are;                             // 2 x 2
    hoatMatrix2D GyHat;                           // 2 x 1
    hoatMatrix2D KxImag;                          // 2 x 2
    hoatMatrix2D KxRe;                            // 2 x 2
    hoatMatrix2D KyImag;                          // 2 x 2
    hoatMatrix2D KyImagInv;                       // 2 x 2
    hoatMatrix2D KyRe;                            // 2 x 2
    hoatMatrix2D Q;                               // 2 x 2
    hoatMatrix2D SxSqrtReg;                       // 2 x 2
    hoatMatrix2D UxImag;                          // 2 x 2
    hoatMatrix2D UxRe;                            // 2 x 2
    hoatMatrix2D currentSHrotationMatrix;         // (shOrder+1)^2 x (shOrder+1)^2
    hoatMatrix2D previousSHrotationMatrix;        // (shOrder+1)^2 x (shOrder+1)^2
    hoatMatrix2D realSHrotationMatrix;            // (shOrder+1)^2 x (shOrder+1)^2
    hoatMatrix2D temp3x1Matrix2;                  // 3 x 1
    hoatMatrix2D temp3x1Matrix;                   // 3 x 1
    hoatMatrix2D temp3xnBandsMatrix2;             // 3 x nBands
    hoatMatrix2D temp3xnBandsMatrix;              // 3 x nBands
    hoatMatrix2D tempImag;                        // 2 x 2
    hoatMatrix2D tempInputMatrixImag;             // channelsPerArray x nFrames
    hoatMatrix2D tempInputMatrixReal;             // channelsPerArray x nFrames
    hoatMatrix2D tempInputMatrixSmallImag;        // channelsPerArray x 1
    hoatMatrix2D tempInputMatrixSmallReal;        // channelsPerArray x 1
    hoatMatrix2D tempR;                           // 3 x 3
    hoatMatrix2D tempRe;                          // 2 x 2
    hoatMatrix2D tempRotatedMatrixImag;           // channelsPerArray x nFrames
    hoatMatrix2D tempRotatedMatrixReal;           // channelsPerArray x nFrames
    hoatMatrix2D tempRotatedMatrixSmallImag;      // channelsPerArray x 1
    hoatMatrix2D tempRotatedMatrixSmallReal;      // channelsPerArray x 1

    std::string matlabDirsFile;
    std::string matlabHrtfFile;
    std::string matlabItdFile;
    std::string resourcePath;
    std::string sofaFilePath;

    std::unique_ptr<hoatranslation::ByteStream> spatialMetadataByteStream;

    saf_sofa_container safSofaContainer;

    BINAURAL_AMBI_DECODER_METHODS decodingMethod;
    DIFFUSE_METHOD_FOR_TARGET_SCM diffuseMethodForTargetSCM;

    decoderData decoderData1;

    /////////////////////// Source aware
    SOURCE_RESIDUAL_MODE sourceResidualMode;
    TARGET_SCM_MODE targetScmMode;

    int maxDelayHops{ 24 };
    int maxSimulSources;
    int nOutsideSources;
    int nSourceBufferFrames;
    int nSources;

    bool useSourceAwareInitialFlag;
    bool useSourceAwareMode;
    bool useSourceAwarePostFilter;
    bool useSourceAwareTimeAlignment;

    float averageOutsideSrcDistance;
    float maxListenerModelAttenuationDistance{ 0.5f };
    float maxSrcToListDist { 3.0f };
    float speedOfSound{ 343.0f };

    float *intensityVectorMultipliers{};             // nBands
    float *micIntensityX{};                          // nBands
    float *micIntensityY{};                          // nBands
    float *micIntensityZ{};                          // nBands
    float *resIntensityX{};                          // nBands
    float *resIntensityY{};                          // nBands
    float *resIntensityZ{};                          // nBands
    float *srcIntensityX{};                          // nBands
    float *srcIntensityY{};                          // nBands
    float *srcIntensityZ{};                          // nBands
    float *tempBandVec1{};                           // nBands
    float *tempBandVec2{};                           // nBands
    float *tempBandVec3{};                           // nBands

    float **interpolatedResidualAzimuthData{};       // nFrames x nBands
    float **interpolatedResidualDTRData{};           // nFrames x nBands
    float **interpolatedResidualElevationData{};     // nFrames x nBands
    float **interpolatedResidualEnergyData{};        // nFrames x nBands
    float **srcAzimuthData{};                        // nMicrophoneArrays x nBands
    float **srcDTRData{};                            // nMicrophoneArrays x nBands
    float **srcElevationData{};                      // nMicrophoneArrays x nBands
    float **srcEnergyData{};                         // nMicrophoneArrays x nBands

    float ***micToSrcAziElevDeg{};                   // nSources x nMicrophoneArrays x 2
    float ***residualAzimuthData{};                  // nMicrophoneArrays x nFrames x nBands
    float ***residualDTRData{};                      // nMicrophoneArrays x nFrames x nBands
    float ***residualElevationData{};                // nMicrophoneArrays x nFrames x nBands
    float ***residualEnergyData{};                   // nMicrophoneArrays x nFrames x nBands

    float_complex ***TFDListenerSrcTAFrame{};        // nSources x nFrames x nBands
    float_complex ***TFDSourceFrame{};               // nSources x nFrames x nBands
    float_complex ***sourceFrameBuffer{};            // nSources x nSourceBufferFrames x nBands
    float_complex ****TFDMicSrcTAFrame{};            // nMicrophoneArrays x nSources x nFrames x nBands

    std::vector<bool> listenerThresholdPresentVec{};
    std::vector<bool> hoaSourceThresholdPresentVec{};
    std::vector<float> listenerThresholdVec{};
    std::vector<float> hoaSourceThresholdVec{};
    std::vector<int> priorityValueVec{};

    std::vector<int> activeSourceCandidates{};                              // varying length of 0..nSources
    std::vector<int> activeSources{};                                       // varying length of 0..nSources
    std::vector<int> closestMicsForSources{};                               // nSources
    std::vector<float> distanceToClosestMicForSource{};                     // nSources
    std::vector<int> outsideSources{};                                      // nOutsideSources
    std::vector<int> secondClosestMicsForSources{};                         // nSources
    std::vector<float> distanceToSecondClosestMicForSource{};               // nSources
    std::vector<int> transformedSourceArrays{};                             // varying length of 0..nSources
    std::vector<std::vector<int>> closestMicToMicDelayHops{};               // nSources x nMicrophoneArrays
    std::vector<std::vector<int>> closestMicToMicDelaySamples{};            // nSources x nMicrophoneArrays
    std::vector<std::vector<int>> micToSrcDelaySamples{};                   // nSources x nMicrophoneArrays
    std::vector<std::vector<int>> secondClosestMicToMicDelayHops{};         // nSources x nMicrophoneArrays
    std::vector<std::vector<int>> secondClosestMicToMicDelaySamples{};      // nSources x nMicrophoneArrays
    std::vector<std::vector<int>> trackedSourceMicArrays{};                 // nTotalFramesTracked x N (N varying 0..nSources)
    std::vector<std::vector<std::vector<int>>> micToMicDelayHops{};         // nSources x nMicrophoneArrays x nMicrophoneArrays
    std::vector<std::vector<std::vector<int>>> micToMicDelaySamples{};      // nSources x nMicrophoneArrays x nMicrophoneArrays

    std::vector<float> listenerToSrcDistances{};                                        // nSources
    std::vector<float> meanListenerPosition{0.0f, 0.0f, 0.0f};
    std::vector<float> outsideSrcDistances{};                                           // nOutsideSources
    std::vector<std::vector<float>> Es_noTA;                                            // nSources x nBands
    std::vector<std::vector<float>> closestMicToMicAttenuationGainsSquared{};           // nSources x nMicrophoneArrays
    std::vector<std::vector<float>> closestMicToMicAttenuationGains{};                  // nSources x nMicrophoneArrays
    std::vector<std::vector<float>> micToSrcAttenuationGains{};                         // nSources x nMicrophoneArrays
    std::vector<std::vector<float>> micToSrcDistances{};                                // nSources x nMicrophoneArrays
    std::vector<std::vector<float>> secondClosestMicToMicAttenuationGainsSquared{};     // nSources x nMicrophoneArrays
    std::vector<std::vector<float>> secondClosestMicToMicAttenuationGains{};            // nSources x nMicrophoneArrays
    std::vector<std::vector<float>> sourcePositionsOpenGL{};                            // nSources x 3 (as OpenGL: (x right, y up, z back))

    hoatMatrixCmplx2D temp2x2MatrixCmplx;               // 2 x 2
    hoatMatrixCmplx2D tempBandMatrix;                   // 1 x nBands
    hoatMatrixCmplx2D tempInputFrameMatrix;             // channelsPerArray x nBands

    hoatMatrixCmplx3D Cyo_res;                          // nBands x 2 x 2
    hoatMatrixCmplx3D Cyo_src;                          // nBands x 2 x 2
    hoatMatrixCmplx3D micToClosestSrcBeams;             // nSources x (1 x (shOrder+1)^2)
    hoatMatrixCmplx3D micToSecondClosestSrcBeams;       // nSources x (1 x (shOrder+1)^2)

    hoatMatrix2D temp4x1Matrix;                         // 4 x 1
    hoatMatrix2D temp4x4Matrix;                         // 4 x 4
    hoatMatrix3D Cao_src;                               // nBands x 4 x 4
    hoatMatrix3D listenerToSrcVectorsOpenGL;            // nSources x 3 x 1
    hoatMatrix4D Cam_src;                               // nMicrophoneArrays x nBands x 4 x 4
    hoatMatrix4D micToSrcYY;                            // nSources x nMicrophoneArrays x 4 x 4
    //////////////////////////////////////////////////
};

void initialize_complexity_reduction(hoaTranslatorData *h, std::vector<int> const& complexityReductionParameters);

void create_hoa_translator(hoaTranslatorData **phHoaTranslatorData,
                           int frameSize,
                           int sampleRate,
                           int hopSize,
                           int hybridMode,
                           int LDmode,
                           int outChannels,
                           int channelsPerArray,
                           int nAnalysisChannels,
                           std::string resourcePath,
                           DIFFUSE_METHOD_FOR_TARGET_SCM diffuseMethodForTargetSCM,
                           micArrayPositionsAndOrientations micPosAndOrientations,
                           std::shared_ptr<std::vector<std::uint8_t>> spatialMetadata,
                           sourceAwareParameters sourceAwareParams,
                           std::vector<sourceAwareParametersPerSource> const& sourceAwareParamsPerSource);

bool initialize_hoa_translator(hoaTranslatorData **phHoaTranslatorData,
                               int HOAGroupId,
                               bool useEncodedValues,
                               float interpolationEqGainMax,
                               BINAURAL_AMBI_DECODER_METHODS decodingMethod,
                               inputFilesAndFlags inputFiles,
                               optimalMixingParameters optimalMixingParams,
                               exteriorRenderingParameters exteriorRenderingParams,
                               sphoa6DofExtensionParameters sphoaParameters,
                               std::vector<float> const& linearSourceGains,
                               std::vector<int> const& complexityReductionParameters);

void destroy_hoa_translator(hoaTranslatorData **phHoaTranslatorData);

/* Determine the closest mic array for this frame. The closest array candidate is determined by the biggest interpolation weights
 * in the last sample of this processing frame. The closest array is updated, when the distance from the previous closest
 * array exceeds a threshold limit, i.e.:
 *
 *      distance_to_previous_closest_array > (distance_threshold * distance_to_candidate_closest_array)
 *
 * Output:
 *      currentClosestMicrophoneArray    - closest mic array index
 */
void determine_closest_microphone_array(hoaTranslatorData *h);

/* Save currentInterFrameWeights, currentInterpolationWeights, previousInterpolationWeights and current triangle to the
 * beginning of the track records.
 */
void save_tracked_values(hoaTranslatorData *h);

/* After rendering a processing frame, do some actions:
 *  - save values from this frame
 *  - check if crossfade can be stopped
 *  - shift the tracked values by one frame
 *  - increase the counter for processed frames
 */
void post_rendering_actions(hoaTranslatorData *h);

/* Determine the arrays that are transformed with STFT, and the arrays that are used in the processing.
 *
 * Output:
 *      transformedArraysForCurrentFrame    - mic arrays, that are transformed with STFT
 *      activeArraysForCurrentFrame         - mic arrays, that are used in the processing for the current frame
 *                                            (i.e. spatial metadata is needed for these arrays)
 */
void determine_active_arrays(hoaTranslatorData *h);

/* Choose the active mic array triangle and the corresponding interpolation weights from a set of tracked frames. The number
 * of tracked frames is determined by nTotalFramesTracked (= current frame + (nTotalFramesTracked-1) previous frames).
 * The track record saves data from:
 * - active triangles (mic array indexes)
 * - interpolation weights
 *
 * The purpose of the track record is to delay the swap between triangles. This way the STFTs can be calculated only for
 * a limited amount of arrays (active triangle, and the triangles found in the track record) in order to diminish the
 * computational load.
 *
 * The selection of the active triangle goes as follows:
 *
 * - if the previously chosen active triangle is found in the track record, the chosen triangle is not swapped
 *    —> in that case, use the latest interpolation weights available for the previous triangle
 *    -> this indicates cases, where:
 *          * the previously chosen triangle is the current triangle (use current interpolation weights)
 *          * the listener moved to another triangle, but has only been in the new triangle less than nTotalFramesTracked
 *            frames time (i.e. the previous triangle is still found in the track record), (for interpolation weights, use
 *            the latest ones available from the previous triangle, i.e. the ones calculated on the border area before the
 *            listener switched triangles)
 *    -> NOTE: an exception to the above is, when the listener is projected to an edge connecting two triangles. In that case,
 *       all the arrays in both triangles need to be transformed long enough with STFT. If sufficient time has not passed,
 *       the interpolation weights are used from the latest frame, where the listener was not projected to the edge.
 *
 * - if the previously chosen active triangle is not found (i.e. the listener has not been in the previous triangle for
 *   at least nTotalFramesTracked frames), switch the chosen triangle to the one that has been tracked the longest (the
 *   triangle at the end of the track record list)
 *    —> this ensures that STFTs has been calculated for at least nTotalFramesTracked frames for the chosen HOAs
 *       (and the internal buffers of afSTFT should be flushed)
 *    —> for interpolation weights, do linear interpolation between the weights in current and previous chosen triangles
 *       (the used weights are from the edge of the two triangles, i.e. the last weights in the old triangle, and the
 *       first weights in the new triangle)
 *    -> after the switch, there are two options for the listener position:
 *          * the listener is still inside the newly switched triangle: do linear interpolation between the weights at
 *            the current position and the weights at the border/the first weights in the new triangle (i.e. the listener
 *            is treated as if he/she moved from the border to the current position in the new triangle)
 *          * the listener has moved to another triangle: do the usual checkup for the active triangle (i.e. check if the
 *            chosen triangle is still found in the track record) and start the switch process, if the listener has stayed
 *            in the new triangle long enough
 *
 * More info in doc/delayed_switch_between_mic_array_triangles.doc, doc/Band_by_band_crossfade_for_closest_mic_arrays.doc and
 * doc/Improvements_to_triangluation_and_processing_outside_capturing_area.doc
 *
 * Output:
 *      chosenInterFrameWeights    - interpolation weights for each mic array and each sub-frame (i.e. the weights for
 *                                   the active triangle)
 * When crossfade is in progress:
 *      fadeOutWeights             - interpolation weights for fade-out frequency bands
 *      fadeInWeights              - interpolation weights for fade-in frequency bands
 */
void choose_active_triangle_from_track_record(hoaTranslatorData *h);

// Transform 3D OpenGL vector (x right, y up, z back) to cartesian (x front, y left, z up)
void openGL_to_cartesian_xFront(const float *openGLvec, float *cartesianVec);
void openGL_to_cartesian_xFront(const std::vector<float> &openGLvec, std::vector<float> &cartesianVec);

bool AnearB(float A, float B, float errorMargin);

// Read and return the contents of file `filepath`. The vector may be empty if reading failed.
std::shared_ptr<std::vector<std::uint8_t>> read_file_to_vector(std::string const& filepath);

/* Form target covariance matrix targetM.
 *
 * When useRecursiveAveraging = true:
 *      targetM = recursionSmoothingCoefficient * targetM +
 *                (1 - recursionSmoothingCoefficient) * newM
 * else:
 *      targetM = previousM + newM
 */
void combine_covariance_matrices(hoaTranslatorData *h,
                                 hoatMatrixCmplx2D &targetM,
                                 const hoatMatrixCmplx2D &previousM,
                                 hoatMatrixCmplx2D &newM);

// targetM = recursionSmoothingCoefficient * targetM +
//           (1 - recursionSmoothingCoefficient) * newM
void combine_recursive_covariance_matrices(hoaTranslatorData *h,
                                           hoatMatrixCmplx2D &targetM,
                                           hoatMatrixCmplx2D &newM);

void copy_hrtf_to_matrix(hoaTranslatorData *h, int band, int hrtfDirIndex);

int getHrtfDirIndex(hoaTranslatorData *h, float aziRad, float elevRad);

void compute_direct_part_of_Cy(hoaTranslatorData *h, int frame, int band, int dirIndex, float energy, float dtr, hoatMatrixCmplx3D &Cy);

void compute_diffuse_part_of_Cy(hoaTranslatorData *h, int frame, int band, float energy, float dtr, hoatMatrixCmplx3D &Cy);


#endif //HOA_TRANSLATION_UTILITIES_HPP
