/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "VectorOperations.hpp"

#include <cassert>
#include <cmath>
#include <numeric>

#include <saf.h>

#ifdef SAF_USE_INTEL_MKL
#include "mkl.h"
#elif SAF_USE_APPLE_ACCELERATE
#include <Accelerate/Accelerate.h>
#endif

std::vector<float> sumVectors(std::vector<float> &A, const std::vector<float> &B)
{
    assert(A.size() == B.size());
    std::vector<float> sumVec(A.size(), 0.0f);
    utility_svvadd(A.data(), B.data(), A.size(), sumVec.data());
    return sumVec;
}

std::vector<float> subtractVectors(const std::vector<float> &A, const std::vector<float> &B)
{
    // NOTE: utility_svvsub() in SAF is broken for Apple Accelerate (in the current version).
    // Need to update SAF to use utility_svvsub().
    assert(A.size() == B.size());
    std::vector<float> subVec(A.size(), 0.0f);
    for (int i = 0; i < A.size(); ++i) {
        subVec[i] = A[i] - B[i];
    }
    return subVec;
}

void subtractVectors(const float *A, const float *B, float *C, int vecLen)
{
    #if defined(SAF_USE_INTEL_MKL)
    vsSub(vecLen, A, B, C);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vDSP_vsub(B, 1, A, 1, C, 1, vecLen);
    #else
	for (int i = 0; i < vecLen; ++i) {
		C[i] = A[i] - B[i];
	}
    #endif
}

void normalizeVector(std::vector<float> &vector)
{
    float normDiv = 1.0f / calculate_vector_norm(vector);
    utility_svsmul(vector.data(), &normDiv, vector.size(), NULL);
}

std::vector<float> multiplyVector(float multiplier, std::vector<float> &vector)
{
    std::vector<float> output(vector.size(), 0.0f);
    utility_svsmul(vector.data(), &multiplier, vector.size(), output.data());
    return output;
}

std::vector<float> multiplyVector(float multiplier, float *vector, int vecLen)
{
    std::vector<float> output(vecLen, 0.0f);
    utility_svsmul(&vector[0], &multiplier, vecLen, output.data());
    return output;
}

void addScalarToVector(float scalar, std::vector<float> &vector)
{
    utility_svsadd(vector.data(), &scalar, vector.size(), vector.data());
}

std::vector<float> powVector(float pow, float *vector, int vecLen)
{
    std::vector<float> powVec(vecLen, 0.0f);
    #if defined(SAF_USE_INTEL_MKL)
    vsPowx(vecLen, vector, pow, powVec.data());
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    if (std::fabs(pow - 2.0f) < FLT_EPSILON) {
        vDSP_vsq(vector, 1, powVec.data(), 1, vecLen);
    } else {
        for (int i = 0; i < vecLen; ++i) {
            powVec[i] = std::pow(vector[i], pow);
        }
    }
    #else
    for (int i = 0; i < vecLen; ++i) {
        powVec[i] = std::pow(vector[i], pow);
    }
    #endif
    return powVec;
}

void powVector(float pow, float *vector, int vecLen, float *powVector)
{
    #if defined(SAF_USE_INTEL_MKL)
    vsPowx(vecLen, vector, pow, powVector);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    if (std::fabs(pow - 2.0f) < FLT_EPSILON) {
        vDSP_vsq(vector, 1, powVector, 1, vecLen);
    } else {
        for (int i = 0; i < vecLen; ++i) {
            powVector[i] = std::pow(vector[i], pow);
        }
    }
    #else
    for (int i = 0; i < vecLen; ++i) {
        powVector[i] = std::pow(vector[i], pow);
    }
#endif
}


std::vector<float> sqrtVector(float *vector, int vecLen)
{
    std::vector<float> sqrtVec(vecLen, 0.0f);
    #if defined(SAF_USE_INTEL_MKL)
    vsSqrt(vecLen, vector, sqrtVec.data());
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vvsqrtf(sqrtVec.data(), vector, &vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
        sqrtVec[i] = std::sqrt(vector[i]);
    }
    #endif
    return sqrtVec;
}

void sqrtVector(float *vector, float *output, int vecLen)
{
    #if defined(SAF_USE_INTEL_MKL)
    vsSqrt(vecLen, vector, output);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vvsqrtf(output, vector, &vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
            output[i] = std::sqrt(vector[i]);
        }
    #endif
}

std::vector<float> atan2Vector(const float *vec1, const float *vec2, int vecLen)
{
    std::vector<float> atan2Vec(vecLen, 0.0f);
    #if defined(SAF_USE_INTEL_MKL)
    vsAtan2(vecLen, vec1, vec2, atan2Vec.data());
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vvatan2f(atan2Vec.data(), vec1, vec2, &vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
        atan2Vec[i] = std::atan2(vec1[i], vec2[i]);
    }
    #endif
    return atan2Vec;
}

void atan2Vector(const float *vec1, const float *vec2, float *outVec, int vecLen)
{
    #if defined(SAF_USE_INTEL_MKL)
        vsAtan2(vecLen, vec1, vec2, outVec);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
        vvatan2f(outVec, vec1, vec2, &vecLen);
    #else
        for (int i = 0; i < vecLen; ++i) {
            outVec[i] = std::atan2(vec1[i], vec2[i]);
        }
    #endif
}

float dotProduct(const std::vector<float> &A, const std::vector<float> &B)
{
    assert(A.size() == B.size());
    return std::inner_product(A.begin(), A.end(), B.begin(), 0.0f);
}

float calculate_vector_norm_for_first_n_dimensions(const float* vector, int n)
{
    return std::sqrt(std::inner_product(&vector[0], &vector[n], &vector[0], 0.0f));
}

float calculate_vector_norm_for_first_n_dimensions(const std::vector<float> &vector, int n)
{
    return std::sqrt(std::inner_product(&vector[0], &vector[n], &vector[0], 0.0f));
}

float calculate_vector_norm(const std::vector<float> &vector)
{
    return std::sqrt(std::inner_product(vector.begin(), vector.end(), vector.begin(), 0.0f));
}

void calculate_vector_norm(float* vec1, float* vec2, float* powVec1, float* powVec2, float* norm, int vecLen)
{
    // vec1^2 + vec2^2
    powVector(2.0f, vec1, vecLen, powVec1);
    powVector(2.0f, vec2, vecLen, powVec2);
    utility_svvadd(powVec1, powVec2, vecLen, norm);

    // norm = sqrt(vec1^2 + vec2^2)
    #if defined(SAF_USE_INTEL_MKL)
    vsSqrt(vecLen, norm, norm);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vvsqrtf(norm, norm, &vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
        norm[i] = std::sqrt(norm[i]);
    }
    #endif

}

void calculate_vector_norm(float* vec1, float* vec2, float* vec3, float* powVec1, float* powVec2, float* powVec3, float* norm, int vecLen)
{

    // vec1^2 + vec2^2 + vec3^2
    powVector(2.0f, vec1, vecLen, powVec1);
    powVector(2.0f, vec2, vecLen, powVec2);
    powVector(2.0f, vec3, vecLen, powVec3);
    utility_svvadd(powVec1, powVec2, vecLen, norm);
    utility_svvadd(powVec3, norm, vecLen, norm);

    // norm = sqrt(vec1^2 + vec2^2 + vec3^2)
    #if defined(SAF_USE_INTEL_MKL)
    vsSqrt(vecLen, norm, norm);
    #elif defined(SAF_USE_APPLE_ACCELERATE)
    vvsqrtf(norm, norm, &vecLen);
    #else
    for (int i = 0; i < vecLen; ++i) {
        norm[i] = std::sqrt(norm[i]);
    }
    #endif

}

