/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "gtest/gtest.h"

#include <algorithm>
#include <cmath>
#include <iomanip>
#include <iostream>


#include <hoatranslation/MultiPointHoaTranslator.hpp>

namespace hoatranslatortest {

class HoaTranslatorTest
{
public:

    enum MIC_ARRAY_ARRANGEMENT {
        DIAMOND,
        NON_CONVEX_HULL,
        DIAMOND_SOURCE_AWARE,
        TWO_SOURCES,
    };

    std::unique_ptr<hoaTranslatorData> m_hoaTranslatorData;

    ~HoaTranslatorTest()
    {
        if (m_hoaTranslatorData != nullptr) {
            auto h = &*m_hoaTranslatorData;
            destroy_hoa_translator(&h);
        }
    }

    HoaTranslatorTest() : HoaTranslatorTest(DIAMOND) {}

    explicit HoaTranslatorTest(const MIC_ARRAY_ARRANGEMENT micArrayArrangement)
    {
        m_hoaTranslatorData = std::make_unique<hoaTranslatorData>();
        auto h = &*m_hoaTranslatorData;

        int frameSize = 512;
        int sampleRate = 48000;
        int hopSize = 128;
        int hybridMode = 1;
        int LDmode = 0;
        int outChannels = 2;
        int channelsPerArray = 4;
        std::string resourcePath = "";
        DIFFUSE_METHOD_FOR_TARGET_SCM diffuseMethodForTargetSCM = DIFFUSE_COVARIANCE_MATRIX;
        int nAnalysisChannels = 4;
        float interpolationEqGainMax = 4.0F;
        BINAURAL_AMBI_DECODER_METHODS decodingMethod = BINAURAL_DECODER_MAGLS;
        bool useEncodedValues = false;
        int groupId = 0;

        inputFilesAndFlags inputFiles{ "",
                                       "",
                                       "",
                                       "",
                                       "data/test_6dof_metadata.bin",
                                       false };

        optimalMixingParameters optimalMixingParams{ 4000,
                                                     10000,
                                                     true,
                                                     true,
                                                     false,
                                                     0.9f };

        exteriorRenderingParameters exteriorRenderingParams{true,
                                                            4.0f};

        sphoa6DofExtensionParameters sphoaParameters{ false, 0.0f, 0.0f, 1 };

        sourceAwareParameters sourceAwareParams{false, false, false, BOTH, SEPARATE_SCM, {}, 10};
        std::vector<sourceAwareParametersPerSource> sourceAwareParamsPerSource{};

        int nMicrophoneArrays = 0;
        std::vector<std::vector<float>> arrayPositionsOpenGL{};
        std::vector<std::vector<float>> arrayOrientationsOpenGL{};
        if (micArrayArrangement == DIAMOND) {
            /* 5 FOA microphones (o) in a diamond shape arrangement.
             *
             *              o m4
             *           /  |  \
             *        /     |     \
             *     /        |        \
             * m2 o- - - - -o m0 - - - -o m1  -
             *     \        |        /        |
             *        \     |     /           1m
             *           \  |  /              |
             *              o m3              –
             *              |----1m---|
             *
             * NOTE: microphoneArrayPositions in hoaTranslatorData use a OpenGL coordinate system,
             *       where: x right, y up, z back
             */

            nMicrophoneArrays = 5;

            arrayPositionsOpenGL.push_back({ 0.0f, 1.7f,  0.0f}); // m0
            arrayPositionsOpenGL.push_back({ 1.0f, 1.7f,  0.0f}); // m1
            arrayPositionsOpenGL.push_back({-1.0f, 1.7f,  0.0f}); // m2
            arrayPositionsOpenGL.push_back({ 0.0f, 1.7f,  1.0f}); // m3
            arrayPositionsOpenGL.push_back({ 0.0f, 1.7f, -1.0f}); // m4

        } else if (micArrayArrangement == NON_CONVEX_HULL) {
            /* 8 FOA microphones (o) in an arrangement described below. Omitting edges that are >4m
             * result in a non-convex shape.
             *
             *  Before omitting long edges              After omitting long edges
             *     m0         m1                            m0         m1
             *     o - - - - o                  -            o - - - - o
             *     | \       | \                |              \       | \
             *     |    \    |    \             2m                \    |    \
             *     |       \ |       \          |                    \ |       \
             *     |      m7 o - - - - o m2     -                   m7 o - - - - o m2
             *     |      /  |        /|        |                      |        /|
             *     |     /   |     /   |        2m                     |     /   |
             *     |    /    |  /      |        |                      |  /      |
             *     |   /  m6 o - - - - o m3     -                   m6 o - - - - o m3
             *     |  /    / |       /          |                    / |       /
             *     | /  /    |    /             2m                /    |    /
             *     |//       | /                |              /       | /
             *     o - - - - o                  -            o - - - - o
             *     m5        m4                              m5        m4
             *     |----2m---|----2m---|
             */

            nMicrophoneArrays = 8;

            arrayPositionsOpenGL.push_back({0.0f, 1.7f, 0.0f}); // m0
            arrayPositionsOpenGL.push_back({2.0f, 1.7f, 0.0f}); // m1
            arrayPositionsOpenGL.push_back({4.0f, 1.7f, 2.0f}); // m2
            arrayPositionsOpenGL.push_back({4.0f, 1.7f, 4.0f}); // m3
            arrayPositionsOpenGL.push_back({2.0f, 1.7f, 6.0f}); // m4
            arrayPositionsOpenGL.push_back({0.0f, 1.7f, 6.0f}); // m5
            arrayPositionsOpenGL.push_back({2.0f, 1.7f, 4.0f}); // m6
            arrayPositionsOpenGL.push_back({2.0f, 1.7f, 2.0f}); // m7

        } else if (micArrayArrangement == DIAMOND_SOURCE_AWARE) {
            /* 5 FOA microphones (o) in a diamond shape arrangement.
             * 4 sources (x) next to the edge mic arrays.
             *
             *                          x s3
             *
             *                          o m4
             *                       /  |  \
             *                    /     |     \
             *                 /        |        \
             *       s1 x  m2 o- - - - -o m0 - - - -o m1  - x s0
             *                 \        |        /        |
             *                    \     |     /           1m
             *                       \  |  /              |
             *                          o m3              –
             *                          |----1m---|---1m----|
             *                          x s2
             */
            nMicrophoneArrays = 5;

            arrayPositionsOpenGL.push_back({0.0f, 1.7f, 0.0f});  // m0
            arrayPositionsOpenGL.push_back({1.0f, 1.7f, 0.0f});  // m1
            arrayPositionsOpenGL.push_back({-1.0f, 1.7f, 0.0f}); // m2
            arrayPositionsOpenGL.push_back({0.0f, 1.7f, 1.0f});  // m3
            arrayPositionsOpenGL.push_back({0.0f, 1.7f, -1.0f}); // m4

            std::vector<std::vector<float>> sourcePositionsOpenGL{ { 2.0f, 1.7f,  0.0f},
                                                                   {-2.0f, 1.7f,  0.0f},
                                                                   { 0.0f, 1.7f,  2.0f},
                                                                   { 0.0f, 1.7f, -2.0f} };

            sourceAwareParams = {true, false, false, BOTH, SEPARATE_SCM, sourcePositionsOpenGL};
 
            sourceAwareParamsPerSource.push_back({true, false, 2.0, 0.0, 0});
            sourceAwareParamsPerSource.push_back({true, false, 2.0, 0.0, 0});
            sourceAwareParamsPerSource.push_back({true, false, 2.0, 0.0, 0});
            sourceAwareParamsPerSource.push_back({true, false, 2.0, 0.0, 0});
        } else if (micArrayArrangement == TWO_SOURCES) {
            /* Two sources 1m apart from each other.
             *
             * m0 o- - - -o m1
             *    |--1m---|
             */
            nMicrophoneArrays = 2;
            arrayPositionsOpenGL.push_back({0.0f, 0.0f, 0.0f}); // m0
            arrayPositionsOpenGL.push_back({1.0f, 0.0f, 0.0f}); // m1

        }

        arrayOrientationsOpenGL.assign(nMicrophoneArrays, std::vector<float>(3, 0.0f));

        micArrayPositionsAndOrientations micPosAndOrientations {arrayPositionsOpenGL, arrayOrientationsOpenGL };

        std::vector<float> linearSourceGains(nMicrophoneArrays, 1.0f);

        std::vector<int> complexityReductionParameters{};

        create_hoa_translator(&h,
                              frameSize,
                              sampleRate,
                              hopSize,
                              hybridMode,
                              LDmode,
                              outChannels,
                              channelsPerArray,
                              nAnalysisChannels,
                              resourcePath,
                              diffuseMethodForTargetSCM,
                              micPosAndOrientations,
                              read_file_to_vector(resourcePath + inputFiles.spatialMetadataBinaryFile),
                              sourceAwareParams,
                              sourceAwareParamsPerSource);

        initialize_hoa_translator(&h,
                                  groupId,
                                  useEncodedValues,
                                  interpolationEqGainMax,
                                  decodingMethod,
                                  inputFiles,
                                  optimalMixingParams,
                                  exteriorRenderingParams,
                                  sphoaParameters,
                                  linearSourceGains,
                                  complexityReductionParameters);

        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
            h->transformedArraysForCurrentFrame.push_back(mic);
            h->transformedArraysForCurrentFrameFullOrder.push_back(mic);
            h->activeArraysForCurrentFrame.push_back(mic);
        }
    }
};

bool isTriangleEqualToA(const std::vector<int> &selectedTriangle,
                        const std::vector<int> &triangleA)
{
    return (selectedTriangle[0] == triangleA[0]
            && selectedTriangle[1] == triangleA[1]
            && selectedTriangle[2] == triangleA[2]);
}

bool isTriangleEqualToAorB(const std::vector<int> &selectedTriangle,
                           const std::vector<int> &triangleA,
                           const std::vector<int> &triangleB)
{
    return isTriangleEqualToA(selectedTriangle, triangleA)
           || isTriangleEqualToA(selectedTriangle, triangleB);
}

bool vectorAcontainsB(const std::vector<std::vector<int>> &A,
                      const std::vector<int> &B)
{
    for (auto const& element : A) {
        if (element == B) {
            return true;
        }
    }
    return false;
}

// Convert float vector to integer vector for comparison
// (use 2 decimals for rounding, i.e. convert 1.234f to 123)
void convertTo2DecimalIntVec(const std::vector<float> &input,
                             std::vector<int> &output)
{
    for (int i = 0; i < input.size(); ++i) {
        output[i] = std::lround(input[i] * 100.0f);
    }
}

// Convert float complex vector to integer vector for comparison
// (use 2 decimals for rounding, i.e. convert 1.234f to 123)
void convertTo2DecimalIntVec(const std::vector<float_complex> &input,
                             std::vector<std::vector<int>> &output)
{
    for (int i = 0; i < input.size(); ++i) {
        output[i][0] = std::lround(input[i].real() * 100.0f);
        output[i][1] = std::lround(input[i].imag() * 100.0f);
    }
}

// Convert float vector to integer vector for comparison
// (use N decimals for rounding, i.e. convert 1.234f to 123, if Ndecimals = 2)
void convertToNDecimalIntVec(const std::vector<float> &input,
                             std::vector<int> &output,
                             int Ndecimals)
{
    float intMultiplier = std::pow(10.0f, Ndecimals);
    for (int i = 0; i < input.size(); ++i) {
        output[i] = std::lround(input[i] * intMultiplier);
    }
}

std::string vecToString(const std::vector<int> &input)
{
    std::stringstream output;
    output << "{ ";
    for (const auto &value : input) {
        output << value << ", ";
    }
    output << "}";
    return output.str();
}

std::string vecToString(const std::vector<float> &input)
{
    std::stringstream output;
    output << std::fixed << std::setprecision(2);
    output << "{ ";
    for (const auto &value : input) {
        output << value << ", ";
    }
    output << "}";
    return output.str();
}

std::string vecToString(const std::vector<float_complex> &input)
{
    std::stringstream output;
    output << std::fixed << std::setprecision(2);
    output << "{ ";
    for (const auto &value : input) {
        output << "(" << value.real() << ", " << value.imag() << "), ";
    }
    output << "}";
    return output.str();
}

// Fill tracked vectors with given values
void fill_tracked_vectors(hoaTranslatorData *h,
                          std::vector<int> currentTriangle,
                          std::vector<float> currentInterFrameWeights,
                          std::vector<float> currentInterpolationWeights,
                          std::vector<float> previousInterpolationWeights)
{
    for (int frame = 0; frame < h->nTotalFramesTracked; ++frame) {
        h->trackedCurrentTriangleArrays[frame] = currentTriangle;
        h->trackedCurrentInterFrameWeights[frame] = currentInterFrameWeights;
        h->trackedCurrentInterpolationWeights[frame] = currentInterpolationWeights;
        h->trackedPreviousInterpolationWeights[frame] = previousInterpolationWeights;
    }
}

inline std::vector<int> noTriangle  = {};
inline std::vector<int> triangle013 = { 0, 1, 3 };
inline std::vector<int> triangle014 = { 0, 1, 4 };
inline std::vector<int> triangle023 = { 0, 2, 3 };
inline std::vector<int> triangle024 = { 0, 2, 4 };

struct PreProcessorTestData
{
    std::vector<float> listenerPosition;
    std::vector<float> expectedInterpolationWeights;
    bool isListenerInsideTriangle;
    std::vector<int> expectedTriangleA;
    std::vector<int> expectedTriangleB;
    friend std::ostream& operator<<(std::ostream& os, const PreProcessorTestData& data)
    {
        return os
                << "listenerPosition: " << vecToString(data.listenerPosition)
                << ", expectedInterpolationWeights: " << vecToString(data.expectedInterpolationWeights)
                << ", isListenerInsideTriangle: " << data.isListenerInsideTriangle
                << ", expectedTriangleA: " << vecToString(data.expectedTriangleA)
                << ", expectedTriangleB: " << vecToString(data.expectedTriangleB);
    }
};

class PreProcessorTest
        : public ::testing::TestWithParam<PreProcessorTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(PreProcessorTest,
       PreProcessor) // NOLINT
{
    /* Pre process different listener positions, and check that the
     * interpolation weights are correct. Also check that the correct mic array
     * triangle is selected.
     */

    auto params = PreProcessorTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    h->currentListenerPosition = params.listenerPosition;

    pre_process_positions(h);

    // Collect output values
    std::vector<int> selectedTriangle = h->triangleFaces[h->activeTriangleIndex];
    std::vector<float> expectedWeights = params.expectedInterpolationWeights;
    std::vector<float> processedWeights(h->nMicrophoneArrays, 0.0f);
    std::copy_n(&h->currentInterpolationWeights[0], h->nMicrophoneArrays, &processedWeights[0]);

    // Compare
    EXPECT_EQ(processedWeights, expectedWeights);
    EXPECT_EQ(h->isListenerInsideTriangle, params.isListenerInsideTriangle);
    if (!params.expectedTriangleB.empty()) {
        EXPECT_PRED3(isTriangleEqualToAorB, selectedTriangle, params.expectedTriangleA, params.expectedTriangleB);
    } else if (!params.expectedTriangleA.empty()) {
        EXPECT_PRED2(isTriangleEqualToA, selectedTriangle, params.expectedTriangleA);
    }
}

INSTANTIATE_TEST_SUITE_P(Default, PreProcessorTest,
                         testing::Values(
                         /* Test case: the listener in the center (on top of m0). The weight for m0
                          * should be 1.0, and the rest should be 0.0. The listener should be
                          * inside some triangle, but the triangle can be ambiguous, because the
                          * listener is on top of a mic array.
                          */
                         PreProcessorTestData{ std::vector<float> {0.0f, 1.7f, 0.0f},
                                               std::vector<float> {1.0f, 0.0f, 0.0f, 0.0f, 0.0f},
                                               true,
                                               noTriangle,
                                               noTriangle },
                         /* Test case: the listener is in the middle between m0 and m1. The weights
                          * for m0 and m1 should be 0.5, the rest should be 0.0. The listener
                          * should be inside triangle013 or triangle014: the choice is ambiguous,
                          * because the location is on the edge between two triangles.
                          */
                         PreProcessorTestData{ std::vector<float> {0.5f, 1.7f, 0.0f},
                                               std::vector<float> {0.5f, 0.5f, 0.0f, 0.0f, 0.0f},
                                               true,
                                               triangle013,
                                               triangle014 },
                         /* Test case: the listener is in the middle between m1 and m4. The weights
                          * for m1 and m4 should be 0.5, the rest should be 0.0. The listener
                          * should be inside triangle014.
                          */
                         PreProcessorTestData{ std::vector<float> {0.5f, 1.7f, -0.5f},
                                               std::vector<float> {0.0f, 0.5f, 0.0f, 0.0f, 0.5f},
                                               true,
                                               triangle014,
                                               noTriangle },
                         /* Test case: the listener is inside triangle023. The location is closer to
                          * m0, so the weight for it should be bigger than for m2 and m3. The weight
                          * for m0 should be 0.5, and for m2 and m3 the weights should be 0.25, the
                          * rest should be 0.0. The listener should be inside triangle023.
                          */
                         PreProcessorTestData{ std::vector<float> {-0.25f, 1.7f, 0.25f},
                                               std::vector<float> {0.5f, 0.0f, 0.25f, 0.25f, 0.0f},
                                               true,
                                               triangle023,
                                               noTriangle },
                         /* Test case: the listener is outside the triangles. The location lies on
                          * the front axis (+y in this case), and the listener position should be
                          * projected on top of m4. The weight for m4 should be 1.0, and the rest
                          * should be 0.0. The listener shouldn't be inside any triangle.
                          */
                         PreProcessorTestData{ std::vector<float> {0.0f, 1.7f, -2.0f},
                                               std::vector<float> {0.0f, 0.0f, 0.0f, 0.0f, 1.0f},
                                               false,
                                               noTriangle,
                                               noTriangle }
                         ));

struct MetadataInterpolatorTestData
{
    std::vector<float> interFrameWeights;
    std::vector<float> azimuthDataDeg;
    std::vector<float> elevationDataDeg;
    std::vector<float> DTRData;
    std::vector<float> energyData;
    float expectedAzimuthDeg;
    float expectedElevationDeg;
    float expectedDTR;
    float expectedEnergy;
    friend std::ostream& operator<<(std::ostream& os, const MetadataInterpolatorTestData& data)
    {
        return os
           << "interFrameWeights: " << vecToString(data.interFrameWeights)
           << ", azimuthDataDeg: " << vecToString(data.azimuthDataDeg)
           << ", elevationDataDeg: " << vecToString(data.elevationDataDeg)
           << ", DTRData: " << vecToString(data.DTRData)
           << ", energyData: " << vecToString(data.energyData)
           << ", expectedAzimuthDeg: " << data.expectedAzimuthDeg
           << ", expectedElevationDeg: " << data.expectedElevationDeg
           << ", expectedDTR: " << data.expectedDTR
           << ", expectedEnergy: " << data.expectedEnergy;
    }
};

class MetadataInterpolatorTest
        : public ::testing::TestWithParam<MetadataInterpolatorTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(MetadataInterpolatorTest,
       DISABLED_MetadataInterpolator) // NOLINT
{
    /* Interpolate metadata (DOA angles, DTR, energy), and check that the
     * interpolated values are correct.
     */

    auto params = MetadataInterpolatorTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    h->useImprovedExteriorRendering = false;
    float angleErrorMarginDegrees = 0.1f;
    float DTRerrorMargin = 0.01f;
    float energyErrorMargin = 0.1f;

    // Initialize input values
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        h->chosenInterFrameWeights[mic][0] = params.interFrameWeights[mic];
        h->azimuthData[mic][0][0] = params.azimuthDataDeg[mic] * M_PI / 180.0f;
        h->elevationData[mic][0][0] = params.elevationDataDeg[mic] * M_PI / 180.0f;
        h->DTRData[mic][0][0] = params.DTRData[mic];
        h->energyData[mic][0][0] = params.energyData[mic];
    }

    // Interpolate and collect output values
    interpolate_mic_metadata(h);
    float interpolatedAzimuth = h->interpolatedAzimuthData[0][0] * 180.0f / M_PI;
    float interpolatedElevation = h->interpolatedElevationData[0][0] * 180.0f / M_PI;
    float interpolatedDTR = h->interpolatedDTRData[0][0];
    float interpolatedEnergy = h->interpolatedEnergyData[0][0];

    // Compare
    EXPECT_NEAR(interpolatedAzimuth, params.expectedAzimuthDeg, angleErrorMarginDegrees);
    EXPECT_NEAR(interpolatedElevation, params.expectedElevationDeg, angleErrorMarginDegrees);
    EXPECT_NEAR(interpolatedDTR, params.expectedDTR, DTRerrorMargin);
    EXPECT_NEAR(interpolatedEnergy, params.expectedEnergy, energyErrorMargin);
}

INSTANTIATE_TEST_SUITE_P(Default, MetadataInterpolatorTest,
                         testing::Values(
                         /* Test case: the listener is positioned in the middle between m0 and m1.
                          * This is indicated by the 0.5 interpolation weights for m0 and m1. A
                          * direct sound source is located on top of m4, with 0 degrees of
                          * elevation. This sets the azimuth angle for m0 to 0 degrees, and the
                          * angle for m1 to 45 degrees. The current implementation of metadata
                          * interpolation takes the average of these angles, so the resulting
                          * interpolated azimuth angle should be 22.5. The elevation angle should
                          * remain intact (= 0). The direct-to-total (DTR) value is set to 1.0 for
                          * both m0 and m1. After calculations the interpolated DTR should result to 0.9239.
                          * The total energy should remain intact (= 1). The azimuth, elevation,
                          * DTR, and energy values for mic arrays m2..m4 shouldn't matter,
                          * because their interpolation weights are set to 0.
                          */
                         MetadataInterpolatorTestData{ std::vector<float> {0.5f, 0.5f,  0.0f,  0.0f, 0.0f},
                                                       std::vector<float> {0.0f, 45.0f, 30.0f, 4.0f, 53.0f},
                                                       std::vector<float> {0.0f, 0.0f,  3.0f,  5.0f, 1.0f},
                                                       std::vector<float> {1.0f, 1.0f,  0.5f,  0.4f, 0.9f},
                                                       std::vector<float> {1.0f, 1.0f,  30.0f, 2.0f, 500.0f},
                                                       45.0f/2.0f,
                                                       0.0f,
                                                       0.9239f,
                                                       1.0f },
                         /* Test case: the listener is positioned on top of m0. All the values for m0
                          * should remain intact after interpolation: azimuth 10, elevation 34,
                          * DTR 1, energy 1. The azimuth, elevation, DTR, and energy values for
                          * mic arrays m1..m4 shouldn't matter, because their interpolation weights
                          * are set to 0.
                          */
                         MetadataInterpolatorTestData{ std::vector<float> {1.0f,  0.0f,   0.0f,   0.0f,   0.0f},
                                                       std::vector<float> {10.0f, 40.0f, -30.0f,  120.0f, 180.0f},
                                                       std::vector<float> {34.0f, 3.0f,   56.0f, -54.0f,  34.0f},
                                                       std::vector<float> {1.0f,  0.3f,   0.4f,   1.0f,   0.8f},
                                                       std::vector<float> {1.0f,  2.0f,   65.0f,  23.0f,  14.0f},
                                                       10.0f,
                                                       34.0f,
                                                       1.0f,
                                                       1.0f }));

struct SignalInterpolatorTestData
{
    std::vector<float> interFrameWeights;
    std::vector<float> energyData;
    std::vector<float> realTFSignal;
    int expectedClosestMicArray;
    friend std::ostream& operator<<(std::ostream& os, const SignalInterpolatorTestData& data)
    {
        return os
                << "interFrameWeights: " << vecToString(data.interFrameWeights)
                << ", energyData: " << vecToString(data.energyData)
                << ", realTFSignal: " << vecToString(data.realTFSignal)
                << ", expectedClosestMicArray: " << data.expectedClosestMicArray;
    }
};

class SignalInterpolatorTest
        : public ::testing::TestWithParam<SignalInterpolatorTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(SignalInterpolatorTest,
       SignalInterpolator) // NOLINT
{
    // Interpolate signals, and check that the output is correct.

    auto params = SignalInterpolatorTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    float errorMargin = 0.1f;

    // Initialize input values
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->chosenInterFrameWeights[mic][frame] = params.interFrameWeights[mic];
        }
        h->energyData[mic][0][0] = params.energyData[mic];
        h->TFDInputFrameFullOrder[mic][0][0][0] = float_complex(params.realTFSignal[mic], 0.0f);
        h->interpolatedEnergyData[0][0] += h->chosenInterFrameWeights[mic][0] * h->energyData[mic][0][0];
    }

    // Interpolate
    determine_closest_microphone_array(h);
    interpolate_signals(h);

    // Calculate expected output and collect output values
    int expectedClosestMicrophoneArray = params.expectedClosestMicArray;
    float expectedEqGainMax = (std::min)(std::sqrt(h->interpolatedEnergyData[0][0] /
                                                   h->energyData[expectedClosestMicrophoneArray][0][0]), h->interpolationEqGainMax);
    float expectedSignal = expectedEqGainMax * h->TFDInputFrameFullOrder[expectedClosestMicrophoneArray][0][0][0].real();
    float interpolatedSignal = h->interpolatedTFDFrame[0][0][0].real();

    // Compare
    EXPECT_NEAR(interpolatedSignal, expectedSignal, errorMargin);
    EXPECT_EQ(h->currentClosestMicrophoneArray, expectedClosestMicrophoneArray);
}

INSTANTIATE_TEST_SUITE_P(Default, SignalInterpolatorTest,
                         testing::Values(
                         /* Test case: the listener is between m0 and m2, but closer to m2 (indicated
                          * by the interFrameWeights 0.25 and 0.75, respectively). Arrays m0 and m2
                          * have captured some signals (TFDInputFrame)
                          * with some energies (energyData). The energies are interpolated by
                          * multiplying them with their respective weights. Currently, the signal
                          * interpolator uses the signal from the closest mic array. In this case,
                          * the closest array should be m2, and the expected output values are
                          * calculated using that assumption. The values set for mic arrays m1, m3
                          * and m4 shouldn't matter, because their interpolation weights are set to
                          * 0.0.
                          */
                         SignalInterpolatorTestData{ std::vector<float> {0.25f, 0.0f, 0.75f, 0.0f, 0.0f},
                                                     std::vector<float> {1.4f,  3.0f, 3.7f,  4.0f, 5.0f},
                                                     std::vector<float> {13.0f, 0.5f, 42.0f, 1.8f, 300.0f},
                                                     2 }));

struct MetadataRotatorTestData
{
    std::vector<float> yawDegs;
    std::vector<float> pitchDegs;
    std::vector<float> rollDegs;
    float signalAzimuthDeg;
    float signalElevationDeg;
    float expectedAzimuthDeg;
    float expectedElevationDeg;
    friend std::ostream& operator<<(std::ostream& os, const MetadataRotatorTestData& data)
    {
        return os
                << "yawDegs: " << vecToString(data.yawDegs)
                << ", pitchDegs: " << vecToString(data.pitchDegs)
                << ", rollDegs: " << vecToString(data.rollDegs)
                << ", signalAzimuthDeg: " << data.signalAzimuthDeg
                << ", signalElevationDeg: " << data.signalElevationDeg
                << ", expectedAzimuthDeg: " << data.expectedAzimuthDeg
                << ", expectedElevationDeg: " << data.expectedElevationDeg;
    }
};

class MetadataRotatorTest
        : public ::testing::TestWithParam<MetadataRotatorTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(MetadataRotatorTest,
       DISABLED_RotateMetadata) // NOLINT
{
    // Rotate metadata, and check that the output is correct.

    auto params = MetadataRotatorTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    float errorMarginDegrees = 0.5F;

    // Initialize input
    int nRotations = params.yawDegs.size();
    h->azimuthData[0][0][0] = params.signalAzimuthDeg * M_PI / 180.0f;
    h->elevationData[0][0][0] = params.signalElevationDeg * M_PI / 180.0f;
    h->DTRData[0][0][0] = 1.0f;

    // Vectorize the metadata and rotate
    vectorize_metadata(h, false);
    for (int i = 0; i < nRotations; ++i) {
        h->yawDeg = params.yawDegs[i];
        h->pitchDeg = params.pitchDegs[i];
        h->rollDeg = params.rollDegs[i];
        rotate_metadata_vectors(h);
    }

    // Calculate expected output and collect output values
    std::vector<float> tempMetadataVector{h->tempV1[0][0][0],
                                          h->tempV2[0][0][0],
                                          h->tempV3[0][0][0]};
    float expectedAzimuth = params.expectedAzimuthDeg;
    float expectedElevation = params.expectedElevationDeg;
    float rotatedAzimuth = calculate_azimuth_OpenGL(tempMetadataVector.data()) * 180.0f / M_PI;
    float rotatedElevation = calculate_elevation_OpenGL(tempMetadataVector.data()) * 180.0f / M_PI;

    // Compare
    if (!isnan(expectedAzimuth)) {
        EXPECT_NEAR(rotatedAzimuth, expectedAzimuth, errorMarginDegrees);
    }
    if (!isnan(expectedElevation)) {
        EXPECT_NEAR(rotatedElevation, expectedElevation, errorMarginDegrees);
    }
}

INSTANTIATE_TEST_SUITE_P(Default, MetadataRotatorTest,
                         testing::Values(
                         /* Test case: a sound is coming from a direction of azimuth = -15.0 and
                          * elevation = 45.0 degrees. The listener rotates his/her head 78.0
                          * degrees in the horizontal plane (yaw rotation). The listener should now
                          * perceive that the sound is coming from a direction of
                          * azimuth = -78.0 + (-15.0) = -93.0 and elevation = 45.0.
                          */
                         MetadataRotatorTestData{ std::vector<float> {78.0f},
                                                  std::vector<float> {0.0f},
                                                  std::vector<float> {0.0f},
                                                  -15.0f,
                                                  45.0f,
                                                  -93.0f,
                                                  45.0f },
                         /* Test case: a sound is coming from a direction of azimuth = -15.0 and
                          * elevation = 90.0 degrees. The listener rotates his/her head around the
                          * horizontal plane (yaw rotation). The perceived direction of the sound
                          * should not change from the rotations, (because it's coming directly above
                          * the listener), i.e. the elevation direction should stay the same:
                          * elevation = 90.0. The azimuth direction is ambiguous, because the sound
                          * is coming from above the listener.
                          */
                         MetadataRotatorTestData{ std::vector<float> {42.0f},
                                                  std::vector<float> {0.0f},
                                                  std::vector<float> {0.0f},
                                                  -15.0f,
                                                  90.0f,
                                                  NAN,
                                                  90.0f },
                         /* Test case: a sound is coming from a direction of azimuth = 0.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head almost 90.0
                          * degrees in the median plane (pitch rotation). The listener should now
                          * perceive that the sound is coming from a direction of azimuth = 0.0 and
                          * elevation = -90.0.
                          */
                         MetadataRotatorTestData{ std::vector<float> {0.0f},
                                                  std::vector<float> {89.9f},
                                                  std::vector<float> {0.0f},
                                                  0.0f,
                                                  0.0f,
                                                  0.0f,
                                                  -89.9f },
                         /* Test case: a sound is coming from a direction of azimuth = -90.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head around the
                          * median plane (pitch rotation). The perceived direction of the sound
                          * should not change from the rotations, i.e. the direction should stay the
                          * same: azimuth = -90.0 and elevation = 0.0.
                          */
                         MetadataRotatorTestData{ std::vector<float> {0.0f},
                                                  std::vector<float> {-54.0f},
                                                  std::vector<float> {0.0f},
                                                  -90.0f,
                                                  0.0f,
                                                  -90.0f,
                                                  -0.0f },
                         /* Test case: a sound is coming from a direction of azimuth = -90.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head almost 90.0
                          * degrees in the frontal plane (roll rotation). The listener should now
                          * perceive that the sound is coming from a direction of elevation = -90.0.
                          * The azimuth direction is ambiguous, because the sound is coming from
                          * below the listener.
                          */
                         MetadataRotatorTestData{ std::vector<float> {0.0f},
                                                  std::vector<float> {0.0f},
                                                  std::vector<float> {89.9f},
                                                  -90.0f,
                                                  0.0f,
                                                  NAN,
                                                  -90.0f },
                         /* Test case: a sound is coming from a direction of azimuth = 0.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head around the
                          * frontal plane (roll rotation). The perceived direction of the sound
                          * should not change from the rotations, i.e. the direction should stay the
                          * same: azimuth = 0.0 and elevation = 0.0.
                          */
                         MetadataRotatorTestData{ std::vector<float> {0.0f},
                                                  std::vector<float> {0.0f},
                                                  std::vector<float> {-124.0f},
                                                  0.0f,
                                                  0.0f,
                                                  0.0f,
                                                  0.0f },
                         /* Test case: a sound is coming from a direction of azimuth = -90.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head 90.0 degrees
                          * in the frontal plane (roll rotation), and then -90.0 degrees in the
                          * median plane (pitch rotation). The listener should now perceive that the
                          * sound is coming from a direction of azimuth = 0.0 and elevation = 0.0
                          * (i.e. the sound source should be in front of him/her).
                          */
                         MetadataRotatorTestData{ std::vector<float> {0.0f},
                                                  std::vector<float> {-90.0f},
                                                  std::vector<float> {90.0f},
                                                  -90.0f,
                                                  0.0f,
                                                  0.0f,
                                                  0.0f },
                         /* Test case: a sound is coming from a direction of azimuth = 14.0 and
                          * elevation = -90.0 degrees. The listener rotates his/her head -90.0
                          * degrees in the frontal plane (roll rotation), and then -90.0 degrees in
                          * the horizontal plane (yaw rotation). The listener should now perceive
                          * that the sound is coming from a direction of azimuth = 0.0 and elevation = 0.0
                          * (i.e. the sound source should be in front of him/her).
                          */
                         MetadataRotatorTestData{ std::vector<float> {-90.0f},
                                                  std::vector<float> {0.0f},
                                                  std::vector<float> {-90.0f},
                                                  14.0f,
                                                  -90.0f,
                                                  0.0f,
                                                  0.0f },
                         /* Test case: a sound is coming from a direction of azimuth = 90.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head -90.0
                          * degrees in the median plane (pitch rotation), and then 90.0 degrees in
                          * the horizontal plane (yaw rotation). The listener should now perceive
                          * that the sound is coming from a direction of azimuth = 0.0 and elevation = 0.0
                          * (i.e. the sound source should be in front of him/her).
                          */
                         MetadataRotatorTestData{ std::vector<float> {90.0f},
                                                  std::vector<float> {-90.0f},
                                                  std::vector<float> {0.0f},
                                                  90.0f,
                                                  0.0f,
                                                  0.0f,
                                                  0.0f },
                         /* Test case: a sound is coming from a direction of azimuth = -90.0 and
                          * elevation = 0.0 degrees. The listener rotates his/her head 30.0 degrees
                          * in the frontal plane (roll rotation), then -90.0 degrees in the
                          * horizontal plane (yaw rotation), and finally -30.0 degrees in the median
                          * plane (pitch rotation). The listener should now perceive that the sound
                          * is coming from a direction of azimuth = 0.0 and elevation = 0.0 (i.e.
                          * the sound source should be in front of him/her).
                          */
                         MetadataRotatorTestData{ std::vector<float> {0.0f, -90.0f, 0.0f},
                                                  std::vector<float> {0.0f,  0.0f, -30.0f},
                                                  std::vector<float> {30.0f, 0.0f,  0.0f},
                                                  -90.0f,
                                                  0.0f,
                                                  0.0f,
                                                  0.0f }
                         ));

/* Ambisonics channel order (ACN):
 *      0 - omni,  W
 *      1 - left,  Y
 *      2 - up,    Z
 *      3 - front, X
 */
inline auto ACNleft   = std::vector<float_complex>{1.0f,  1.0f,  0.0f,  0.0f};
inline auto ACNright  = std::vector<float_complex>{1.0f, -1.0f,  0.0f,  0.0f};
inline auto ACNup     = std::vector<float_complex>{1.0f,  0.0f,  1.0f,  0.0f};
inline auto ACNdown   = std::vector<float_complex>{1.0f,  0.0f, -1.0f,  0.0f};
inline auto ACNfront  = std::vector<float_complex>{1.0f,  0.0f,  0.0f,  1.0f};
inline auto ACNback   = std::vector<float_complex>{1.0f,  0.0f,  0.0f, -1.0f};

struct SHRotatorTestData
{
    float headYawDeg;
    float headPitchDeg;
    float headRollDeg;
    std::vector<float_complex> ACNvector;
    std::vector<float_complex> expectedRotatedACNvector;
    friend std::ostream& operator<<(std::ostream& os, const SHRotatorTestData& data)
    {
        return os
                << "headYawDeg: " << data.headYawDeg
                << ", headPitchDeg: " << data.headPitchDeg
                << ", headRollDeg: " << data.headRollDeg
                << ", ACNvector: " << vecToString(data.ACNvector)
                << ", expectedRotatedACNvector: " << vecToString(data.expectedRotatedACNvector);
    }
};

class SHRotatorTest
        : public ::testing::TestWithParam<SHRotatorTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(SHRotatorTest,
       DISABLED_RotateSH) // NOLINT
{
    /* Test rotating the STFT signals with a spherical harmonic rotation matrix. The head
     * rotations follow OpenGL orientation (yaw around up-axis, pitch around right-axis,
     * roll around back-axis).
     */

    auto params = SHRotatorTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;

    // Acquire head rotations
    h->yawDeg = params.headYawDeg;
    h->pitchDeg = params.headPitchDeg;
    h->rollDeg = params.headRollDeg;

    // Initialize input
    for (int ch = 0; ch < params.ACNvector.size(); ++ch) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->tempInputMatrixReal[ch][frame] = params.ACNvector[ch].real();
            h->tempInputMatrixImag[ch][frame] = params.ACNvector[ch].imag();
        }
    }

    // Apply rotation
    get_sh_rotation_matrix(h);
    matrixMult(h->currentSHrotationMatrix, h->tempInputMatrixReal, h->tempRotatedMatrixReal);
    matrixMult(h->currentSHrotationMatrix, h->tempInputMatrixImag, h->tempRotatedMatrixImag);
    combineIntoCmplxMatrix(h->tempRotatedMatrixReal, h->tempRotatedMatrixImag, h->tempRotatedMatrix);

    // Collect rotated vector
    std::vector<float_complex> rotatedVec{h->tempRotatedMatrix[0][0],
                                          h->tempRotatedMatrix[1][0],
                                          h->tempRotatedMatrix[2][0],
                                          h->tempRotatedMatrix[3][0]};

    // Convert to integer vectors for comparison
    std::vector<std::vector<int>> rotatedIntVec(rotatedVec.size(), std::vector<int>(2, 0.0f));
    std::vector<std::vector<int>> expectedIntVec(rotatedVec.size(), std::vector<int>(2, 0.0f));
    convertTo2DecimalIntVec(rotatedVec, rotatedIntVec);
    convertTo2DecimalIntVec(params.expectedRotatedACNvector, expectedIntVec);

    //Compare
    EXPECT_EQ(expectedIntVec, rotatedIntVec);
}

INSTANTIATE_TEST_SUITE_P(Default, SHRotatorTest,
                         testing::Values(
                         // Test case: signal coming from the left, turn head 90 degrees yaw
                         // --> expect the rotated signal to be coming from the front
                         SHRotatorTestData{ 90.0f, 0.0f, 0.0f, ACNleft, ACNfront },

                         // Test case: signal coming from the left, turn head -90 degrees yaw
                         // --> expect the rotated signal to be coming from the back
                         SHRotatorTestData{ -90.0f, 0.0f, 0.0f, ACNleft, ACNback },

                         // Test case: signal coming from the front, turn head 90 degrees pitch
                         // --> expect the rotated signal to be coming from down
                         SHRotatorTestData{ 0.0f, 90.0f, 0.0f, ACNfront, ACNdown },

                         // Test case: signal coming from the front, turn head -90 degrees pitch
                         // --> expect the rotated signal to be coming from up
                         SHRotatorTestData{ 0.0f, -90.0f, 0.0f, ACNfront, ACNup },

                         // Test case: signal coming from the left, turn head 90 degrees roll
                         // --> expect the rotated signal to be coming from up
                         SHRotatorTestData{ 0.0f, 0.0f, 90.0f, ACNleft, ACNup },

                         // Test case: signal coming from the left, turn head -90 degrees roll
                         // --> expect the rotated signal to be coming from down
                         SHRotatorTestData{ 0.0f, 0.0f, -90.0f, ACNleft, ACNdown },

                         // Test case: signal coming from the left, turn head 90 degrees roll
                         // and 90 degrees pitch
                         // --> expect the rotated signal to be coming from front
                         SHRotatorTestData{ 0.0f, 90.0f, 90.0f, ACNleft, ACNfront },

                         // Test case: signal coming from the left, turn head 90 degrees pitch
                         // and -90 degrees yaw
                         // --> expect the rotated signal to be coming from front
                         SHRotatorTestData{ -90.0f, 90.0f, 0.0f, ACNright, ACNfront }
                         ));

// Right-handed OpenGL: x (right), y (up), z (back)
inline auto leftVecOpenGL  = std::vector<float>{-1.0f, 0.0f,  0.0f};
inline auto rightVecOpenGL = std::vector<float>{1.0f,  0.0f,  0.0f};
inline auto upVecOpenGL    = std::vector<float>{0.0f,  1.0f,  0.0f};
inline auto downVecOpenGL  = std::vector<float>{0.0f, -1.0f,  0.0f};
inline auto frontVecOpenGL = std::vector<float>{0.0f,  0.0f, -1.0f};
inline auto backVecOpenGL  = std::vector<float>{0.0f,  0.0f,  1.0f};

// Right-handed classic Cartesian: x (front), y (left), z (up)
inline auto leftVecClassic  = std::vector<float>{0.0f,  1.0f,  0.0f};
inline auto rightVecClassic = std::vector<float>{0.0f, -1.0f,  0.0f};
inline auto upVecClassic    = std::vector<float>{0.0f,  0.0f,  1.0f};
inline auto downVecClassic  = std::vector<float>{0.0f,  0.0f, -1.0f};
inline auto frontVecClassic = std::vector<float>{1.0f,  0.0f,  0.0f};
inline auto backVecClassic  = std::vector<float>{-1.0f, 0.0f,  0.0f};

// (yaw, pitch, roll) rotations in degrees in OpenGL form
inline auto noRotationOpenGL                   = std::vector<float>{0.0f,   0.0f,   0.0f};
inline auto leftRotationOpenGL                 = std::vector<float>{90.0f,  0.0f,   0.0f};
inline auto rightRotationOpenGL                = std::vector<float>{-90.0f, 0.0f,   0.0f};
inline auto upRotationOpenGL                   = std::vector<float>{0.0f,   90.0f,  0.0f};
inline auto downRotationOpenGL                 = std::vector<float>{0.0f,  -90.0f,  0.0f};
inline auto counterClockwiseRollRotationOpenGL = std::vector<float>{0.0f,   0.0f,   90.0f};
inline auto clockwiseRollRotationOpenGL        = std::vector<float>{0.0f,   0.0f,  -90.0f};

enum CoordinateSystem {
    ClassicCartesian,
    OpenGL
};

struct RotationMatrixTestData
{
    CoordinateSystem coordinateSystem;
    std::vector<float> yawPitchRollOpenGL;
    std::vector<float> directionVec;
    std::vector<float> expectedRotatedVec;
    friend std::ostream& operator<<(std::ostream& os, const RotationMatrixTestData& data)
    {
        return os
                << "coordinateSystem: " << data.coordinateSystem
                << ", yawPitchRoll: " << vecToString(data.yawPitchRollOpenGL)
                << ", directionVec: " << vecToString(data.directionVec)
                << ", expectedRotatedVec: " << vecToString(data.expectedRotatedVec);
    }
};

class RotationMatrixTest
        : public ::testing::TestWithParam<RotationMatrixTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(RotationMatrixTest,
       DISABLED_RotationMatrix) // NOLINT
{
    /* Rotate a direction vector with a rotation matrix. Both classic Cartesian and OpenGL axis systems are tested.
     * In a Cartesian system, the direction vector xyz follows:
     *      - x (front)
     *      - y (left)
     *      - z (up)
     * And the rotations:
     *      - yaw (around up-axis, z)
     *      - pitch (around left-axis, y)
     *      - roll (around front-axis, x)
     *
     * In OpenGL system, the direction vector xyz follows:
     *      - x (right)
     *      - y (up)
     *      - z (back)
     * And the rotations:
     *      - yaw (around up-axis, y)
     *      - pitch (around right-axis, x)
     *      - roll (around back-axis, z)
     *
     * NOTE: the input rotations are in OpenGL form.
     */

    auto params = RotationMatrixTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;

    // Create temp matrices
    hoatMatrix2D dirVec;
    hoatMatrix2D tempVec;
    createHoatMatrix(&dirVec, 3, 1);
    createHoatMatrix(&tempVec, 3, 1);

    // Collect rotations
    h->yawDeg = params.yawPitchRollOpenGL[0];
    h->pitchDeg = params.yawPitchRollOpenGL[1];
    h->rollDeg = params.yawPitchRollOpenGL[2];

    // Get rotator matrix
    float R[3][3]{};
    if (params.coordinateSystem == ClassicCartesian) {
        get_rotation_matrix_classic(h->yawDeg,
                                    h->pitchDeg,
                                    h->rollDeg,
                                    R);
    } else {
        get_rotation_matrix_OpenGL(h->yawDeg,
                                   h->pitchDeg,
                                   h->rollDeg,
                                   R);
    }

    // Rotate
    std::copy_n(&R[0][0], 3 * 3, &h->tempR[0][0]);
    std::copy_n(params.directionVec.data(), 3, &dirVec[0][0]);
    matrixMult(h->tempR, dirVec, tempVec);

    // Collect rotated vector
    std::vector<float> rotatedVec{tempVec[0][0], tempVec[1][0], tempVec[2][0]};

    // Convert to integer vectors for comparison
    std::vector<int> rotatedIntVec(rotatedVec.size(), 0.0f);
    std::vector<int> expectedIntVec(rotatedVec.size(), 0.0f);
    convertTo2DecimalIntVec(rotatedVec, rotatedIntVec);
    convertTo2DecimalIntVec(params.expectedRotatedVec, expectedIntVec);

    // Compare
    EXPECT_EQ(expectedIntVec, rotatedIntVec);

    // Destroy matrices
    destroyHoatMatrix(&dirVec);
    destroyHoatMatrix(&tempVec);
}

INSTANTIATE_TEST_SUITE_P(Default, RotationMatrixTest,
                         testing::Values(
                         // Test case: no rotation, expect same input and output
                         RotationMatrixTestData{ ClassicCartesian, noRotationOpenGL, frontVecClassic, frontVecClassic },
                         RotationMatrixTestData{ OpenGL, noRotationOpenGL, frontVecOpenGL, frontVecOpenGL },

                         // Test case: vector pointing to the front, turn 90 degrees yaw in OpenGL (left rotation)
                         // --> Cartesian output: left
                         // --> OpenGL output: left
                         RotationMatrixTestData{ ClassicCartesian, leftRotationOpenGL, frontVecClassic, leftVecClassic },
                         RotationMatrixTestData{ OpenGL, leftRotationOpenGL, frontVecOpenGL, leftVecOpenGL },

                         // Test case: vector pointing to the front, turn -90 degrees yaw in OpenGL (right rotation)
                         // --> Cartesian output: right
                         // --> OpenGL output: right
                         RotationMatrixTestData{ ClassicCartesian, rightRotationOpenGL, frontVecClassic, rightVecClassic },
                         RotationMatrixTestData{ OpenGL, rightRotationOpenGL, frontVecOpenGL, rightVecOpenGL },

                         // Test case: vector pointing to the front, turn 90 degrees pitch in OpenGL (up rotation)
                         // --> Cartesian output: down
                         // --> OpenGL output: up
                         RotationMatrixTestData{ ClassicCartesian, upRotationOpenGL, frontVecClassic, downVecClassic },
                         RotationMatrixTestData{ OpenGL, upRotationOpenGL, frontVecOpenGL, upVecOpenGL },

                         // Test case: vector pointing to the front, turn -90 degrees pitch in OpenGL (down rotation)
                         // --> Cartesian output: up
                         // --> OpenGL output: down
                         RotationMatrixTestData{ ClassicCartesian, downRotationOpenGL, frontVecClassic, upVecClassic },
                         RotationMatrixTestData{ OpenGL, downRotationOpenGL, frontVecOpenGL, downVecOpenGL },

                         // Test case: vector pointing to the left, turn 90 degrees roll in OpenGL (counter-clockwise roll)
                         // --> Cartesian output: up
                         // --> OpenGL output: down
                         RotationMatrixTestData{ ClassicCartesian, counterClockwiseRollRotationOpenGL, leftVecClassic, upVecClassic },
                         RotationMatrixTestData{ OpenGL, counterClockwiseRollRotationOpenGL, leftVecOpenGL, downVecOpenGL },

                         // Test case: vector pointing to the left, turn -90 degrees roll in OpenGL (clockwise roll)
                         // --> Cartesian output: down
                         // --> OpenGL output: up
                         RotationMatrixTestData{ ClassicCartesian, clockwiseRollRotationOpenGL, leftVecClassic, downVecClassic },
                         RotationMatrixTestData{ OpenGL, clockwiseRollRotationOpenGL, leftVecOpenGL, upVecOpenGL }
                         ));

struct MicArrayOrientationsTestData
{
    std::vector<float> headRotationOpenGL;
    std::vector<float> micArrayOrientationOpenGL;
    std::vector<float_complex> inputSignalACN;
    std::vector<float_complex> expectedRotatedACNvector;
    friend std::ostream& operator<<(std::ostream& os, const MicArrayOrientationsTestData& data)
    {
        return os
                << "headRotation: " << vecToString(data.headRotationOpenGL)
                << ", micArrayOrientationOpenGL: " << vecToString(data.micArrayOrientationOpenGL)
                << ", inputSignalACN: " << vecToString(data.inputSignalACN)
                << ", expectedRotatedACNvector: " << vecToString(data.expectedRotatedACNvector);
    }
};

class MicArrayOrientationsTest
        : public ::testing::TestWithParam<MicArrayOrientationsTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(MicArrayOrientationsTest,
       DISABLED_MicArrayOrientations) // NOLINT
{
    /* Initialize the mic arrays with some orientations.
     * Check that the array orientations are taken into account correctly, when
     * applying head rotations to metadata and input signals.
     */

    auto params = MicArrayOrientationsTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    float errorMarginDegrees = 0.5F;

    // Initialize head and mic array orientations
    h->yawDeg = params.headRotationOpenGL[0];
    h->pitchDeg = params.headRotationOpenGL[1];
    h->rollDeg = params.headRotationOpenGL[2];
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        std::copy_n(&params.micArrayOrientationOpenGL[0],
                    3,
                    &h->micArrayOrientations[mic][0]);
    }

    // Copy input signal to matrix
    for (int ch = 0; ch < params.inputSignalACN.size(); ++ch) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->tempInputMatrixReal[ch][frame] = params.inputSignalACN[ch].real();
            h->tempInputMatrixImag[ch][frame] = params.inputSignalACN[ch].imag();
        }
    }

    // Calculate DOA from the input and output vectors (transform to (front, left, up))
    std::vector<float> inputVec{ params.inputSignalACN[3].real(),
                                 params.inputSignalACN[1].real(),
                                 params.inputSignalACN[2].real() };
    std::vector<float> outputVec{ params.expectedRotatedACNvector[3].real(),
                                  params.expectedRotatedACNvector[1].real(),
                                  params.expectedRotatedACNvector[2].real() };
    h->azimuthData[0][0][0] = calculate_azimuth(inputVec.data());
    h->elevationData[0][0][0] = calculate_elevation(inputVec.data());
    h->DTRData[0][0][0] = 1.0f;
    float expectedAzimuth = calculate_azimuth(outputVec.data()) * 180.0f / M_PI;
    float expectedElevation = calculate_elevation(outputVec.data()) * 180.0f / M_PI;

    // Vectorize the metadata and rotate
    vectorize_metadata(h, false);
    rotate_metadata_vectors(h);

    // Get spherical harmonics rotation matrix and rotate the input signal
    get_sh_rotation_matrix(h);
    matrixMult(h->currentSHrotationMatrix, h->tempInputMatrixReal, h->tempRotatedMatrixReal);
    matrixMult(h->currentSHrotationMatrix, h->tempInputMatrixImag, h->tempRotatedMatrixImag);
    combineIntoCmplxMatrix(h->tempRotatedMatrixReal, h->tempRotatedMatrixImag, h->tempRotatedMatrix);

    // Collect output values
    std::vector<float> tempMetadataVector{h->tempV1[0][0][0],
                                          h->tempV2[0][0][0],
                                          h->tempV3[0][0][0]};
    float rotatedAzimuth = calculate_azimuth_OpenGL(tempMetadataVector.data()) * 180.0f / M_PI;
    float rotatedElevation = calculate_elevation_OpenGL(tempMetadataVector.data()) * 180.0f / M_PI;
    std::vector<float_complex> rotatedVec{h->tempRotatedMatrix[0][0],
                                          h->tempRotatedMatrix[1][0],
                                          h->tempRotatedMatrix[2][0],
                                          h->tempRotatedMatrix[3][0]};

    // Convert to integer vectors for comparison
    std::vector<std::vector<int>> rotatedIntVec(rotatedVec.size(), std::vector<int>(2, 0.0f));
    std::vector<std::vector<int>> expectedIntVec(rotatedVec.size(), std::vector<int>(2, 0.0f));
    convertTo2DecimalIntVec(rotatedVec, rotatedIntVec);
    convertTo2DecimalIntVec(params.expectedRotatedACNvector, expectedIntVec);

    // Compare (only compare azimuth in cases where the angle is not ambiguous)
    if (params.expectedRotatedACNvector != ACNup && params.expectedRotatedACNvector != ACNdown) {
        EXPECT_NEAR(rotatedAzimuth, expectedAzimuth, errorMarginDegrees);
    }
    EXPECT_NEAR(rotatedElevation, expectedElevation, errorMarginDegrees);
    EXPECT_EQ(rotatedIntVec, expectedIntVec);
}

INSTANTIATE_TEST_SUITE_P(Default, MicArrayOrientationsTest,
                         testing::Values(
                         /* Test case: no head rotation, the mic arrays point to the right.
                          * The arrays capture signal that is coming from their front.
                          * For the listener, the signal should be coming from the right.
                          */
                         MicArrayOrientationsTestData{ noRotationOpenGL,
                                                       rightRotationOpenGL,
                                                       ACNfront,
                                                       ACNright },
                         /* Test case: no head rotation, the mic arrays point to up.
                          * The arrays capture signal that is coming from their back.
                          * For the listener, the signal should be coming from down.
                          */
                         MicArrayOrientationsTestData{ noRotationOpenGL,
                                                       upRotationOpenGL,
                                                       ACNback,
                                                       ACNdown },
                         /* Test case: no head rotation, the mic arrays are rolled 90 degrees.
                          * The arrays capture signal that is coming from their right.
                          * For the listener, the signal should be coming from up.
                          */
                         MicArrayOrientationsTestData{noRotationOpenGL,
                                                      counterClockwiseRollRotationOpenGL,
                                                      ACNright,
                                                      ACNup },
                         /* Test case: the head points to left, the mic arrays point to the front.
                          * The arrays capture signal that is coming from their right.
                          * For the listener, the signal should be coming from back.
                          */
                         MicArrayOrientationsTestData{ leftRotationOpenGL,
                                                       noRotationOpenGL,
                                                       ACNright,
                                                       ACNback },
                         /* Test case: the head points to down, the mic arrays point to left.
                          * The arrays capture signal that is coming from their front.
                          * For the listener, the signal should be coming from left.
                          */
                         MicArrayOrientationsTestData{ downRotationOpenGL,
                                                       leftRotationOpenGL,
                                                       ACNfront,
                                                       ACNleft },
                         /* Test case: the head points to left, the mic arrays point to down.
                          * The arrays capture signal that is coming from their front.
                          * For the listener, the signal should be coming from down.
                          */
                         MicArrayOrientationsTestData{ leftRotationOpenGL,
                                                       downRotationOpenGL,
                                                       ACNfront,
                                                       ACNdown },
                         /* Test case: the head is rolled right 90 degrees, the mic arrays are rolled
                          * left 90 degrees.
                          * The arrays capture signal that is coming from their left.
                          * For the listener, the signal should be coming from right.
                          */
                         MicArrayOrientationsTestData{clockwiseRollRotationOpenGL,
                                                      counterClockwiseRollRotationOpenGL,
                                                      ACNleft,
                                                      ACNright }
                         ));

struct STFTTestData
{
    float inputSignal;
    float expectedOutput;
    friend std::ostream& operator<<(std::ostream& os, const STFTTestData& data)
    {
        return os
                << "inputSignal: " << data.inputSignal
                << ", expectedOutput: " << data.expectedOutput;
    }
};

class STFTTest
        : public ::testing::TestWithParam<STFTTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(STFTTest,
       STFT) // NOLINT
{
    /* Transform a time signal into time-frequency (TF) domain, and back to time
     * domain via Short-Time Fourier Transform (STFT). Compare input and
     * output signals. The transform process is calculated multiple times in this
     * test (stftCount), because there are processing delays in the STFT
     * functions.
     */
    auto params = STFTTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    float errorMargin = 0.1f;
    int sample = h->frameSize - 1;
    int stftCount = 5;
    float_complex ***tempFrame = allocateFloatCmplxArray3d(h->nBands, h->channelsPerArray, h->nFrames);
    h->trackedCurrentTriangleArrays[0].clear();
    h->trackedCurrentTriangleArrays[0] = {0};

    // Initialize time-domain input
    for (int n = 0; n < h->frameSize; ++n) {
        h->timeDomainInputFrame[0][0][n] = params.inputSignal;
    }

    // Transform to TF-domain and back to time-domain multiple times
    for (int i = 0; i < stftCount; ++i) {
        transform_to_time_frequency_domain(h);
        // Copy the TF-domain frame to tempFrame (to be compatible with transform_to_time_domain())
        for (int frame = 0; frame < h->nFrames; ++frame) {
            for (int ch = 0; ch < h->outChannels; ++ch) {
                for (int band = 0; band < h->nBands; ++band) {
                    tempFrame[band][ch][frame] = h->TFDInputFrameFullOrder[0][frame][ch][band];
                }
            }
        }
        transform_to_time_domain(h, tempFrame);
    }

    // Collect outputs
    float expectedOutput = params.expectedOutput;
    float processedOutput = h->timeDomainOutputFrame[0][sample];

    // Compare
    EXPECT_NEAR(processedOutput, expectedOutput, errorMargin);

    // Free temporary frame
    free(tempFrame);
}

INSTANTIATE_TEST_SUITE_P(Default, STFTTest,
                         testing::Values(
                         /* Test case: no additional processing is applied in time-frequency
                          * domain, so the output should be identical to the input.
                          */
                         STFTTestData {5.14f, 5.14f}));

struct SpatialAnalysisTestData
{
    std::vector<float_complex> channelTFSignals;
    float expectedAzimuthDeg;
    float expectedElevationDeg;
    friend std::ostream& operator<<(std::ostream& os, const SpatialAnalysisTestData& data)
    {
        return os
                << "channelTFSignals: " << vecToString(data.channelTFSignals)
                << ", expectedAzimuthDeg: " << data.expectedAzimuthDeg
                << ", expectedElevationDeg: " << data.expectedElevationDeg;
    }
};

class SpatialAnalysisTest
        : public ::testing::TestWithParam<SpatialAnalysisTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(SpatialAnalysisTest,
       SpatialAnalysis) // NOLINT
{
    /* Initalize the mic array channels with signals, and check that the spatial
     * analysis outputs correct Directions of Arrivals (DOAs), Direct to Total
     * Ratio (DTR), and energy values. The energy values are calculated with:
     * 0.5 * ((s0 * c0)^2 + ... + (s3 * c3)^2), where s0..s3 correspond to the
     * time-frequency signals of the respectful channels, and c0..c3 are the
     * sector beamforming coefficients of the respectful channels. The DTR is
     * calculated with: sqrt(sum(expectedIntensityVector^2)) / energy.
     *
     * NOTE: the mic array signals are converted from ACN channel order to
     * cartesian order (x, y, z), where x = front, y = left, z = up. I.e. the
     * first 4 active channels correspond to: 0 = omni, 1 = y = left, 2 = z =
     * up, 3 = x = front.
     */

    auto params = SpatialAnalysisTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    float errorMargin = 0.1f;
    float DTRerrorMargin = 0.01f;
    int firstOrderChannelCount = 4;
    std::vector<float> sectorBeamformingCoefficients = { 1.0f, 0.5774f, 0.5774f, 0.5774f };
    std::vector<float> expectedIntensityVector(3, 0.0f);
    std::vector<float_complex*> arraySignals(firstOrderChannelCount, nullptr);
    h->trackedCurrentTriangleArrays[0].clear();
    h->trackedCurrentTriangleArrays[0] = {0};

    // Initialize input
    for (std::size_t ch = 0; ch < firstOrderChannelCount; ++ch) {
        arraySignals[ch] = &h->TFDInputFrameFullOrder[0][0][ch][0];
        *arraySignals[ch] = params.channelTFSignals[ch];
    }

    spatial_analysis(h);

    // Multiply array signals with sector beamforming coefficients and
    // calculate expected energy
    float expectedEnergy = 0.0f;
    for (int ch = 0; ch < firstOrderChannelCount; ++ch) {
        *arraySignals[ch] *= sectorBeamformingCoefficients[ch];
        expectedEnergy += 0.5F * (std::pow((arraySignals[ch]->real()), 2.0f));
    }

    // Calculate expected output values and collect processed outputs
    expectedIntensityVector[0] = (arraySignals[0]->real()) * (arraySignals[3]->real());
    expectedIntensityVector[1] = (arraySignals[0]->real()) * (arraySignals[1]->real());
    expectedIntensityVector[2] = (arraySignals[0]->real()) * (arraySignals[2]->real());
    float expectedAzimuth = params.expectedAzimuthDeg;
    float expectedElevation = params.expectedElevationDeg;
    float expectedDTR = calculate_dtr(expectedIntensityVector.data(), expectedEnergy);
    float processedAzimuth = h->azimuthData[0][0][0] * 180.0f / M_PI;
    float processedElevation = h->elevationData[0][0][0] * 180.0f / M_PI;
    float processedEnergy = h->energyData[0][0][0];
    float processedDTR = h->DTRData[0][0][0];

    // Compare
    EXPECT_NEAR(processedEnergy, expectedEnergy, errorMargin);
    EXPECT_NEAR(processedDTR, expectedDTR, DTRerrorMargin);
    if (!isnan(expectedAzimuth)) {
        EXPECT_NEAR(processedAzimuth, expectedAzimuth, errorMargin);
    }
    if (!isnan(expectedElevation)) {
        EXPECT_NEAR(processedElevation, expectedElevation, errorMargin);
    }
}

INSTANTIATE_TEST_SUITE_P(Default, SpatialAnalysisTest,
                         testing::Values(
                         /* Test case: the first two channels are active (omni and left). This means
                          * that the signal is coming from a direction of azimuth = 90.0 and
                          * elevation = 0.0 degrees.
                          */
                         SpatialAnalysisTestData { ACNleft,
                                                   90.0f,
                                                   0.0f },
                         /* Test case: the first and third channels are active (omni and up). This
                          * means that the signal is coming from a direction of elevation = 90.0
                          * degrees. The azimuth direction is ambiguous, because the sound is coming
                          * from above.
                          */
                         SpatialAnalysisTestData { ACNup,
                                                   NAN,
                                                   90.0f },
                         /* Test case: the first and fourth channels are active (omni and front).
                          * This means that the signal is coming from a direction of azimuth = 0.0
                          * and elevation = 0.0 degrees.
                          */
                         SpatialAnalysisTestData { ACNfront,
                                                   0.0f,
                                                   0.0f }));

struct RecursiveAveragingSpatialAnalysisTestData{};

class RecursiveAveragingSpatialAnalysisTest
        : public ::testing::TestWithParam<RecursiveAveragingSpatialAnalysisTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(RecursiveAveragingSpatialAnalysisTest,
       SpatialAnalysis) // NOLINT
{
    /* In recursive averaging mode, the spatial metadata values (energy, DTR, DOA angles)
     * are calculated from the average signal values in the whole processing frame.
     * That is, there is only single energy value for all the sub-frames, for example.
     * This test checks that the values are calculated as expected, and that the values
     * in each sub-frame are the same.
     */

    auto h = &*hoaTranslator.m_hoaTranslatorData;
    int firstOrderChannelCount = 4;
    float errorMargin = 0.1f;
    std::vector<float> sectorBeamformingCoefficients = { 1.0f, 0.5774f, 0.5774f, 0.5774f };
    std::vector<std::vector<float>> expectedIntensityVector(h->nBands, std::vector<float>(3, 0.0f));
    std::vector<float> expectedEnergy(h->nBands, 0.0f);
    std::vector<float> inputSample(firstOrderChannelCount, 0.0f);
    auto const& input = &h->TFDInputFrameFullOrder[0];

    // Initialize input, and expected energy and intensity vector
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int band = 0; band < h->nBands; ++band) {
            for (int ch = 0; ch < firstOrderChannelCount; ++ch) {
                (*input)[frame][ch][band] = float_complex((ch * frame * band + 1), 0.0f);
                inputSample[ch] = (*input)[frame][ch][band].real() * sectorBeamformingCoefficients[ch];
                expectedEnergy[band] += (0.5f * inputSample[ch] * inputSample[ch]);
            }
            expectedIntensityVector[band][0] += (inputSample[0] * inputSample[3]);
            expectedIntensityVector[band][1] += (inputSample[0] * inputSample[1]);
            expectedIntensityVector[band][2] += (inputSample[0] * inputSample[2]);
        }
    }

    // Normalize energy and intensity vector
    for (int band = 0; band < h->nBands; ++band) {
        for (int i = 0; i < 3; ++i) {
            expectedIntensityVector[band][i] /= (float)h->nFrames;
        }
        expectedEnergy[band] /= (float)h->nFrames;
    }

    // Calculate expected values
    std::vector<float> expectedAzimuth(h->nBands, 0.0f);
    std::vector<float> expectedElevation(h->nBands, 0.0f);
    std::vector<float> expectedDTR(h->nBands, 0.0f);
    std::vector<std::vector<float>> expectedAzimuthVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> expectedElevationVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> expectedEnergyVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> expectedDTRVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    for (int band = 0; band < h->nBands; ++band) {
        expectedAzimuth[band] = calculate_azimuth(expectedIntensityVector[band].data());
        expectedElevation[band] = calculate_elevation(expectedIntensityVector[band].data());
        expectedDTR[band] = calculate_dtr(expectedIntensityVector[band].data(), expectedEnergy[band]);
    }
    // Copy expected energy values to each sub-frame
    for (int frame = 0; frame < h->nFrames; ++frame) {
        std::copy_n(&expectedEnergy[0], h->nBands, &expectedEnergyVec[frame][0]);
    }
    // For the rest of the metadata, copy only the first sub-frame, the rest of the sub-frames should be zero
    std::copy_n(&expectedAzimuth[0], h->nBands, &expectedAzimuthVec[0][0]);
    std::copy_n(&expectedElevation[0], h->nBands, &expectedElevationVec[0][0]);
    std::copy_n(&expectedDTR[0], h->nBands, &expectedDTRVec[0][0]);

    // Spatial analysis
    h->trackedCurrentTriangleArrays[0].clear();
    h->trackedCurrentTriangleArrays[0] = {0};
    h->useRecursiveAveraging = true;
    spatial_analysis(h);

    // Fetch the processed values from each sub-frame
    std::vector<std::vector<float>> processedAzimuthVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedElevationVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedEnergyVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedDTRVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    for (int frame = 0; frame < h->nFrames; ++frame) {
        std::copy_n(&h->azimuthData[0][frame][0], h->nBands, &processedAzimuthVec[frame][0]);
        std::copy_n(&h->elevationData[0][frame][0], h->nBands, &processedElevationVec[frame][0]);
        std::copy_n(&h->energyData[0][frame][0], h->nBands, &processedEnergyVec[frame][0]);
        std::copy_n(&h->DTRData[0][frame][0], h->nBands, &processedDTRVec[frame][0]);
    }

    // Compare
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int band = 0; band < h->nBands; ++band) {
            EXPECT_NEAR(processedAzimuthVec[frame][band], expectedAzimuthVec[frame][band], errorMargin);
            EXPECT_NEAR(processedElevationVec[frame][band], expectedElevationVec[frame][band], errorMargin);
            EXPECT_NEAR(processedEnergyVec[frame][band], expectedEnergyVec[frame][band], errorMargin);
            EXPECT_NEAR(processedDTRVec[frame][band], expectedDTRVec[frame][band], errorMargin);
        }
    }
}

INSTANTIATE_TEST_SUITE_P(Default, RecursiveAveragingSpatialAnalysisTest,
                         testing::Values(RecursiveAveragingSpatialAnalysisTestData{}));

struct CrossfadeTestData{};

class CrossfadeTest
        : public ::testing::TestWithParam<CrossfadeTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(CrossfadeTest,
       Crossfade) // NOLINT
{
    /* Check that the crossfade works as expected.
     * 1. crossfade bands are processed in expected order
     * 2. crossfade between triangles freezes the fadeOut weights
     * 3. triangle switch during the crossfade freezes both fadeOut and fadeIn weights
     */

    auto h = &*hoaTranslator.m_hoaTranslatorData;
    std::vector<int> expectedFadeInBands{};
    std::vector<int> expectedFadeOutBands{};
    std::vector<float> expectedFadeOutWeights{};
    std::vector<float> processedFadeOutWeights{};
    std::vector<float> expectedFadeInWeights{};
    std::vector<float> processedFadeInWeights{};
    h->currentChosenTriangleArrays = {0, 1, 2};
    std::vector<int> currentTriangle = h->currentChosenTriangleArrays;
    std::vector<float> currentInterFrameWeights(h->nMicrophoneArrays * h->nFrames, 10.0f);
    std::vector<float> currentInterpolationWeights(h->nMicrophoneArrays, 2.0f);
    std::vector<float> previousInterpolationWeights(h->nMicrophoneArrays, 3.0f);

    // 1. Test that the crossfade bands are processed in expected order
    int nProcessingFrames = 3;
    // Get fade-out frequency bands for nProcessingFrames (whole processing frames)
    for (int i = 0; i < nProcessingFrames; ++i) {
        get_fade_out_bands(h);
    }
    // Collect expected fade-in bands for nProcessingFrames
    for (int crossfadeIdx = 0; crossfadeIdx < nProcessingFrames * h->nFrames; ++crossfadeIdx) {
        int lowBand = h->crossfadeBandLow[h->crossfadeBandSwaporder[crossfadeIdx]];
        int highBand = h->crossfadeBandHigh[h->crossfadeBandSwaporder[crossfadeIdx]];
        for (int band = lowBand; band <= highBand; ++band) {
            expectedFadeInBands.push_back(band);
        }
    }
    // Collect expected fade-out bands (= all the bands which are not fade-in bands)
    for (int band = 0; band < h->nBands; ++band) {
        if (std::find(expectedFadeInBands.begin(), expectedFadeInBands.end(), band) == expectedFadeInBands.end()) {
            expectedFadeOutBands.push_back(band);
        }
    }
    // Sort and compare
    auto processedFadeOutBands = h->fadeOutBands[h->nFrames-1];
    std::sort(processedFadeOutBands.begin(), processedFadeOutBands.end());
    std::sort(expectedFadeOutBands.begin(), expectedFadeOutBands.end());
    EXPECT_EQ(processedFadeOutBands, expectedFadeOutBands);

    // 2. Test that a crossfade between triangles freezes the fade-out weights
    h->isCrossfadeInProgress = false;
    h->isTriangleSwichedDuringCrossfade = false;
    h->crossFadeProgressIndx = 0;
    for (int frame = 0; frame < h->nFrames; ++frame) {
        h->fadeOutBands[frame].clear();
    }
    h->currentChosenTriangleArrays = {0, 1, 2};
    currentTriangle = h->currentChosenTriangleArrays;
    // Fill the tracked values
    fill_tracked_vectors(h,
                         currentTriangle,
                         currentInterFrameWeights,
                         currentInterpolationWeights,
                         previousInterpolationWeights);
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->currentInterFrameWeights[mic][frame] = currentInterFrameWeights[mic];
        }
    }
    // Switch the triangle and the closest microphone array at the same time
    h->currentChosenTriangleArrays = {0, 2, 4};
    choose_active_triangle_from_track_record(h);
    EXPECT_TRUE(h->isTriangleSwitched);
    h->previousClosestMicrophoneArray = 1;
    determine_closest_microphone_array(h);
    start_crossfade(h);
    choose_active_triangle_from_track_record(h);
    EXPECT_TRUE(h->isCrossfadeInProgress);
    EXPECT_TRUE(h->isCrossfadeBetweenTriangles);
    // Collect the fade-in and fade-out weights (the expected fade-out weights should be frozen
    // to the previous frame values)
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            processedFadeOutWeights.push_back(h->fadeOutWeights[mic][frame]);
            processedFadeInWeights.push_back(h->fadeInWeights[mic][frame]);
            expectedFadeOutWeights.push_back(previousInterpolationWeights[mic]);
            expectedFadeInWeights.push_back(currentInterFrameWeights[mic]);
        }
    }
    EXPECT_EQ(processedFadeOutWeights, expectedFadeOutWeights);
    EXPECT_EQ(processedFadeInWeights, expectedFadeInWeights);
    processedFadeOutWeights.clear();
    processedFadeInWeights.clear();
    expectedFadeOutWeights.clear();
    expectedFadeInWeights.clear();

    // 3. Test that a triangle switch during the crossfade freezes both fadeOut and fadeIn weights
    h->isCrossfadeInProgress = false;
    h->isTriangleSwichedDuringCrossfade = false;
    h->crossFadeProgressIndx = 0;
    for (int frame = 0; frame < h->nFrames; ++frame) {
        h->fadeOutBands[frame].clear();
    }
    h->currentChosenTriangleArrays = {0, 1, 2};
    currentTriangle = h->currentChosenTriangleArrays;
    // Fill the tracked values
    fill_tracked_vectors(h,
                         currentTriangle,
                         currentInterFrameWeights,
                         currentInterpolationWeights,
                         previousInterpolationWeights);
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->currentInterFrameWeights[mic][frame] = currentInterFrameWeights[mic];
        }
    }
    // Start crossfade
    choose_active_triangle_from_track_record(h);
    h->previousClosestMicrophoneArray = 1;
    determine_closest_microphone_array(h);
    start_crossfade(h);
    EXPECT_TRUE(h->isCrossfadeInProgress);
    EXPECT_FALSE(h->isTriangleSwitched);
    EXPECT_FALSE(h->isCrossfadeBetweenTriangles);
    // Save the expected weights (= frozen weights), and update the current weights
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            expectedFadeOutWeights.push_back(currentInterFrameWeights[mic]);
            expectedFadeInWeights.push_back(currentInterFrameWeights[mic]);
            h->currentInterFrameWeights[mic][frame] = currentInterFrameWeights[mic] + 3.42f;
        }
    }
    // Switch triangle during the crossfade process
    std::vector<int> newTriangle{0, 2, 4};
    for (int i = 0; i < h->nTotalFramesTracked; ++i) {
        h->trackedCurrentTriangleArrays[0] = newTriangle;
        choose_active_triangle_from_track_record(h);
        // Shift the tracked values by one frame
        for (int i = h->nTotalFramesTracked - 1; i > 0; --i) {
            h->trackedCurrentTriangleArrays[i] = h->trackedCurrentTriangleArrays[i - 1];
            h->trackedCurrentInterFrameWeights[i] = h->trackedCurrentInterFrameWeights[i - 1];
            h->trackedCurrentInterpolationWeights[i] = h->trackedCurrentInterpolationWeights[i - 1];
            h->trackedPreviousInterpolationWeights[i] = h->trackedPreviousInterpolationWeights[i - 1];
        }
    }
    EXPECT_TRUE(h->isTriangleSwitched);
    // This should now choose newTriangle (which was already chosen previously),
    // and isTriangleSwitched should change to false
    choose_active_triangle_from_track_record(h);
    EXPECT_FALSE(h->isTriangleSwitched);
    // Collect the processed fade-out and fade-in weights
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            processedFadeOutWeights.push_back(h->fadeOutWeights[mic][frame]);
            processedFadeInWeights.push_back(h->fadeInWeights[mic][frame]);
        }
    }
    // Compare the weights (the weights should be frozen to values that were calculated
    // before the triangle switch process)
    EXPECT_EQ(processedFadeOutWeights, expectedFadeOutWeights);
    EXPECT_EQ(processedFadeInWeights, expectedFadeInWeights);
}

INSTANTIATE_TEST_SUITE_P(Default, CrossfadeTest,
                         testing::Values(
                         CrossfadeTestData {}));

struct TriangleSwitchTestData{};

class TriangleSwitchTest
        : public ::testing::TestWithParam<TriangleSwitchTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(TriangleSwitchTest,
       TriangleSwitch) // NOLINT
{
    /* Check that the delayed triangle switch works as expected.
     * 1. triangle is only switched after a delay
     * 2. the interpolation weights are processed as expected (interpolation between border weights, and after that
     * to the most up-to-date weights)
     */

    auto h = &*hoaTranslator.m_hoaTranslatorData;

    // 1. Test that the triangle is only switched after a delay
    h->currentChosenTriangleArrays = {0, 1, 2};
    std::vector<int> currentTriangle = h->currentChosenTriangleArrays;
    std::vector<float> currentInterFrameWeights(h->nMicrophoneArrays * h->nFrames, 10.0f);
    std::vector<float> currentInterpolationWeights(h->nMicrophoneArrays, 2.0f);
    std::vector<float> previousInterpolationWeights(h->nMicrophoneArrays, 3.0f);
    // Fill the tracked values
    fill_tracked_vectors(h,
                         currentTriangle,
                         currentInterFrameWeights,
                         currentInterpolationWeights,
                         previousInterpolationWeights);
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            h->currentInterFrameWeights[mic][frame] = currentInterFrameWeights[mic];
        }
    }
    // Switch triangle after a delay of nTotalFramesTracked
    std::vector<int> newTriangle = {0, 2, 4};
    for (int i = 0; i < h->nTotalFramesTracked - 1; ++i) {
        h->trackedCurrentTriangleArrays[0] = newTriangle;
        choose_active_triangle_from_track_record(h);
        EXPECT_FALSE(h->isTriangleSwitched);

        // Shift the tracked values by one frame
        for (int i = h->nTotalFramesTracked - 1; i > 0; --i) {
            h->trackedCurrentTriangleArrays[i] = h->trackedCurrentTriangleArrays[i - 1];
            h->trackedCurrentInterFrameWeights[i] = h->trackedCurrentInterFrameWeights[i - 1];
            h->trackedCurrentInterpolationWeights[i] = h->trackedCurrentInterpolationWeights[i - 1];
            h->trackedPreviousInterpolationWeights[i] = h->trackedPreviousInterpolationWeights[i - 1];
        }
    }
    choose_active_triangle_from_track_record(h);
    EXPECT_TRUE(h->isTriangleSwitched);

    // 2. Test that the interpolation weights are processed as expected after triangle switch
    // (interpolation between border weights, and after use the most up-to-date weights)
    std::vector<float> expectedWeightsAfterSwitch{};
    std::vector<float> processedWeightsAfterSwitch{};
    // The triangle was just switched, so the interpolation weights should be calculated from the weights from
    // the end of the track record (i.e. 'h->nTotalFramesTracked - 1' sub-frame)
    auto currentTriangleWeights = h->trackedCurrentInterpolationWeights[h->nTotalFramesTracked - 1];
    auto previousTriangleWeights = h->trackedPreviousInterpolationWeights[h->nTotalFramesTracked - 1];
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            expectedWeightsAfterSwitch.push_back(h->interpolator[frame] * currentTriangleWeights[mic] +
                                                 (1 - h->interpolator[frame]) * previousTriangleWeights[mic]);
            processedWeightsAfterSwitch.push_back(h->chosenInterFrameWeights[mic][frame]);
        }
    }
    EXPECT_EQ(processedWeightsAfterSwitch, expectedWeightsAfterSwitch);
    // The triangle switch is over, so the program should use the most up to date weights (currentInterFrameWeights)
    choose_active_triangle_from_track_record(h);
    EXPECT_FALSE(h->isTriangleSwitched);
    expectedWeightsAfterSwitch = currentInterFrameWeights;
    processedWeightsAfterSwitch.clear();
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            processedWeightsAfterSwitch.push_back(h->chosenInterFrameWeights[mic][frame]);
        }
    }
    EXPECT_EQ(processedWeightsAfterSwitch, expectedWeightsAfterSwitch);
}

INSTANTIATE_TEST_SUITE_P(Default, TriangleSwitchTest,
                         testing::Values(
                         TriangleSwitchTestData {}));

TEST(CustomMatrixTest,
     InitializationAndCopy) // NOLINT
{
    /* Test for custom matrix structs. Check that initialization and copy
     * works, and that changing copied values don't change the original matrix.
     * The matrix structs override operator[], test that its behaviour
     * works as intended.
     */

    int nRows = 5;
    int nCols = 3;
    int n2DMatrices = 4;
    float multiplier = 10.0f;
    hoatMatrixCmplx2D traditional2D;
    hoatMatrixCmplx2D short2D;
    hoatMatrixCmplx3D traditional3D;
    hoatMatrixCmplx3D short3D;
    hoatMatrixCmplx3D temp3D;
    createHoatMatrix(&traditional2D, nRows, nCols);
    createHoatMatrix(&short2D, nRows, nCols);
    createHoatMatrix(&traditional3D, n2DMatrices, nRows, nCols);
    createHoatMatrix(&short3D, n2DMatrices, nRows, nCols);
    createHoatMatrix(&temp3D, n2DMatrices, nRows, nCols);

    // Initialize matrices
    float_complex temp{0.0f, 0.0f};
    for (int n = 0; n < n2DMatrices; ++n) {
        for (int row = 0; row < nRows; ++row) {
            for (int col = 0; col < nCols; ++col) {
                temp = float_complex((n + 1.0f) * (row + 1.0f) * (col + 1.0f), 0.0f);
                traditional3D.matrices2D[n].data[row][col] = temp;
                short3D[n][row][col] = temp;
                if (n == 0) {
                    traditional2D.data[row][col] = temp;
                    short2D[row][col] = temp;
                }
            }
        }
    }

    // Test that the initialization was successful
    EXPECT_EQ(traditional3D.n2DMatrices, n2DMatrices);
    EXPECT_EQ(short3D.n2DMatrices, n2DMatrices);
    for (int n = 0; n < n2DMatrices; ++n) {
        EXPECT_EQ(traditional3D.matrices2D[n].nRows, nRows);
        EXPECT_EQ(traditional3D.matrices2D[n].nCols, nCols);
        EXPECT_EQ(short3D[n].nRows, nRows);
        EXPECT_EQ(short3D[n].nCols, nCols);
        for (int row = 0; row < nRows; ++row) {
            for (int col = 0; col < nCols; ++col) {
                EXPECT_EQ(traditional3D.matrices2D[n].data[row][col], short3D[n][row][col]);
                if (n == 0) {
                    EXPECT_EQ(short2D.nRows, nRows);
                    EXPECT_EQ(short2D.nCols, nCols);
                    EXPECT_EQ(traditional2D.data[row][col], short2D[row][col]);
                }
            }
        }
    }

    // Copy values with a function call
    copy(short3D, temp3D);

    // Change values in the copied matrix
    for (int n = 0; n < n2DMatrices; ++n) {
        for (int row = 0; row < nRows; ++row) {
            for (int col = 0; col < nCols; ++col) {
                temp3D[n][row][col] *= multiplier;
            }
        }
    }

    // Check that the values in the original matrix are unchanged
    for (int n = 0; n < n2DMatrices; ++n) {
        for (int row = 0; row < nRows; ++row) {
            for (int col = 0; col < nCols; ++col) {
                EXPECT_EQ(traditional3D.matrices2D[n].data[row][col], short3D[n][row][col]);
            }
        }
    }

    // Check that the values in the copied matrix are correct
    for (int n = 0; n < n2DMatrices; ++n) {
        for (int row = 0; row < nRows; ++row) {
            for (int col = 0; col < nCols; ++col) {
                EXPECT_EQ(temp3D[n][row][col], short3D[n][row][col] * multiplier);
            }
        }
    }

    // Destroy matrices
    destroyHoatMatrix(&traditional2D);
    destroyHoatMatrix(&short2D);
    destroyHoatMatrix(&traditional3D);
    destroyHoatMatrix(&short3D);
    destroyHoatMatrix(&temp3D);
}

TEST(CustomMatrixTest,
     MatrixMultiplication) // NOLINT
{
    //Test that the custom matrix multiplications work as intended.

    float errorMargin = 0.01f;
    int nRowsAB = 3;
    int nColsAB = 5;
    int nRowsColsC = 4;
    hoatMatrixCmplx2D A;
    hoatMatrixCmplx2D B;
    hoatMatrixCmplx2D C;
    hoatMatrixCmplx2D expected_C_C;
    hoatMatrixCmplx2D expected_AT_B;
    hoatMatrixCmplx2D expected_AH_B;
    hoatMatrixCmplx2D expected_A_BT;
    hoatMatrixCmplx2D expected_A_BH;
    hoatMatrixCmplx2D processed_C_C;
    hoatMatrixCmplx2D processed_AT_B;
    hoatMatrixCmplx2D processed_AH_B;
    hoatMatrixCmplx2D processed_A_BT;
    hoatMatrixCmplx2D processed_A_BH;
    createHoatMatrix(&A, nRowsAB, nColsAB);
    createHoatMatrix(&B, nRowsAB, nColsAB);
    createHoatMatrix(&C, nRowsColsC, nRowsColsC);
    createHoatMatrix(&expected_C_C, nRowsColsC, nRowsColsC);
    createHoatMatrix(&expected_AT_B, nColsAB, nColsAB);
    createHoatMatrix(&expected_AH_B, nColsAB, nColsAB);
    createHoatMatrix(&expected_A_BT, nRowsAB, nRowsAB);
    createHoatMatrix(&expected_A_BH, nRowsAB, nRowsAB);
    createHoatMatrix(&processed_C_C, nRowsColsC, nRowsColsC);
    createHoatMatrix(&processed_AT_B, nColsAB, nColsAB);
    createHoatMatrix(&processed_AH_B, nColsAB, nColsAB);
    createHoatMatrix(&processed_A_BT, nRowsAB, nRowsAB);
    createHoatMatrix(&processed_A_BH, nRowsAB, nRowsAB);

    // Initialize matrices A and B
    float_complex temp{0.0f, 0.0f};
    for (int row = 0; row < A.nRows; ++row) {
        for (int col = 0; col < A.nCols; ++col) {
            temp = float_complex((row - 2.0f) * (col + 1.0f), (row + 2.4f) * (col + 3.2f));
            A[row][col] = temp;
            temp *= 3.0f;
            B[row][col] = temp;
        }
    }
    // Initialize matrix C
    for (int row = 0; row < C.nRows; ++row) {
        for (int col = 0; col < C.nCols; ++col) {
            temp = float_complex((row + 5.23f) * (col + 4.3f), (row - 2.4f) * (col + 3.2f));
            C[row][col] = temp;
        }
    }

    // Expected output
    // C * C
    for (int row = 0; row < expected_C_C.nRows; ++row) {
        for (int col = 0; col < expected_C_C.nCols; ++col) {
            for (int i = 0; i < C.nRows; ++i) {
                expected_C_C[row][col] += C[row][i] * C[i][col];
            }
        }
    }
    // (A^T * B) and (A^H * B)
    for (int ATcol = 0; ATcol < A.nCols; ++ATcol) {
        for (int Bcol = 0; Bcol < B.nCols; ++Bcol) {
            for (int Brow = 0; Brow < B.nRows; ++Brow) {
                expected_AT_B[ATcol][Bcol] += A[Brow][ATcol] * B[Brow][Bcol];
                expected_AH_B[ATcol][Bcol] += std::conj(A[Brow][ATcol]) * B[Brow][Bcol];
            }
        }
    }
    // (A * B^T) and (A * B^H)
    for (int Arow = 0; Arow < A.nRows; ++Arow) {
        for (int BTcol = 0; BTcol < B.nCols; ++BTcol) {
            for (int BTrow = 0; BTrow < B.nRows; ++BTrow) {
                expected_A_BT[Arow][BTrow] += A[Arow][BTcol] * B[BTrow][BTcol];
                expected_A_BH[Arow][BTrow] += A[Arow][BTcol] * std::conj(B[BTrow][BTcol]);
            }
        }
    }

    // Processed output
    matrixMult(C, C, processed_C_C);
    matrixMultTransNoTrans(A, B, processed_AT_B);
    matrixMultConjNoTrans(A, B, processed_AH_B);
    matrixMultNoTransTrans(A, B, processed_A_BT);
    matrixMultNoTransConj(A, B, processed_A_BH);

    // Compare
    for (int row = 0; row < expected_C_C.nRows; ++row) {
        for (int col = 0; col < expected_C_C.nCols; ++col) {
            EXPECT_NEAR(expected_C_C[row][col].real(), processed_C_C[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_C_C[row][col].imag(), processed_C_C[row][col].imag(), errorMargin);
        }
    }
    for (int row = 0; row < expected_AT_B.nRows; ++row) {
        for (int col = 0; col < expected_AT_B.nCols; ++col) {
            EXPECT_NEAR(expected_AT_B[row][col].real(), processed_AT_B[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_AT_B[row][col].imag(), processed_AT_B[row][col].imag(), errorMargin);
            EXPECT_NEAR(expected_AH_B[row][col].real(), processed_AH_B[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_AH_B[row][col].imag(), processed_AH_B[row][col].imag(), errorMargin);
        }
    }
    for (int row = 0; row < expected_A_BT.nRows; ++row) {
        for (int col = 0; col < expected_A_BT.nCols; ++col) {
            EXPECT_NEAR(expected_A_BT[row][col].real(), processed_A_BT[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_A_BT[row][col].imag(), processed_A_BT[row][col].imag(), errorMargin);
            EXPECT_NEAR(expected_A_BH[row][col].real(), processed_A_BH[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_A_BH[row][col].imag(), processed_A_BH[row][col].imag(), errorMargin);
        }
    }

    // Destroy matrices
    destroyHoatMatrix(&A);
    destroyHoatMatrix(&B);
    destroyHoatMatrix(&C);
    destroyHoatMatrix(&expected_C_C);
    destroyHoatMatrix(&expected_AT_B);
    destroyHoatMatrix(&expected_AH_B);
    destroyHoatMatrix(&expected_A_BT);
    destroyHoatMatrix(&expected_A_BH);
    destroyHoatMatrix(&processed_C_C);
    destroyHoatMatrix(&processed_AT_B);
    destroyHoatMatrix(&processed_AH_B);
    destroyHoatMatrix(&processed_A_BT);
    destroyHoatMatrix(&processed_A_BH);
}

TEST(CustomMatrixTest,
     MatrixSumAndScalarMult) // NOLINT
{
    //Test that the custom matrix summations and scalar multiplication work as intended.

    float errorMargin = 0.01f;
    float scalarMultiplier = 2.47;
    int nRows = 5;
    int nCols = 5;
    hoatMatrixCmplx2D A;
    hoatMatrixCmplx2D B;
    hoatMatrixCmplx2D expected_sum_AB;
    hoatMatrixCmplx2D expected_s_A;
    hoatMatrixCmplx2D processed_sum_AB;
    hoatMatrixCmplx2D processed_s_A;
    createHoatMatrix(&A, nRows, nCols);
    createHoatMatrix(&B, nRows, nCols);
    createHoatMatrix(&expected_sum_AB, nRows, nCols);
    createHoatMatrix(&expected_s_A, nRows, nCols);
    createHoatMatrix(&processed_sum_AB, nRows, nCols);
    createHoatMatrix(&processed_s_A, nRows, nCols);

    // Initialize matrices A and B
    float_complex temp{0.0f, 0.0f};
    for (int row = 0; row < A.nRows; ++row) {
        for (int col = 0; col < A.nCols; ++col) {
            temp = float_complex((row - 2.0f) * (col + 1.0f), (row + 2.4f) * (col + 3.2f));
            A[row][col] = temp;
            temp *= 3.0f;
            B[row][col] = temp;
        }
    }

    // Expected output
    float expected_diagonal_sum = 0.0f;
    for (int row = 0; row < nRows; ++row) {
        for (int col = 0; col < nCols; ++col) {
            expected_sum_AB[row][col] = A[row][col] + B[row][col];
            expected_s_A[row][col] = A[row][col] * scalarMultiplier;
            if (col == row) {
                expected_diagonal_sum += A[row][col].real();
            }
        }
    }

    // Processed output
    matrixMatrixAdd(A, B, processed_sum_AB);
    matrixScalarMult(scalarMultiplier, A, processed_s_A);
    float processed_diagonal_sum = sumDiagonalReal(A);

    // Compare
    for (int row = 0; row < nRows; ++row) {
        for (int col = 0; col < nCols; ++col) {
            EXPECT_NEAR(expected_sum_AB[row][col].real(), processed_sum_AB[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_sum_AB[row][col].imag(), processed_sum_AB[row][col].imag(), errorMargin);
            EXPECT_NEAR(expected_s_A[row][col].real(), processed_s_A[row][col].real(), errorMargin);
            EXPECT_NEAR(expected_s_A[row][col].imag(), processed_s_A[row][col].imag(), errorMargin);
        }
    }
    EXPECT_NEAR(expected_diagonal_sum, processed_diagonal_sum, errorMargin);

    // Destroy matrices
    destroyHoatMatrix(&A);
    destroyHoatMatrix(&B);
    destroyHoatMatrix(&expected_sum_AB);
    destroyHoatMatrix(&expected_s_A);
    destroyHoatMatrix(&processed_sum_AB);
    destroyHoatMatrix(&processed_s_A);
}

TEST(VectorNormTest,
     VectorNorm) // NOLINT
{
    //Test that the vector norm calculations work as expected

    std::vector<float> vector1{-2.0f,  3.0f, 43.0f};
    std::vector<float> vector2{ 2.0f, -5.2f,  4.3f};
    std::vector<float> vector3{-4.2f, -1.0f,  0.0f};
    std::vector<float> vector4{ 3.4f,  2.1f,  2.0f};
    float expectedNorm1 = 0.0f;
    float expectedNorm2 = 0.0f;
    float expectedNorm3 = 0.0f;
    float expectedNorm4 = 0.0f;
    float expectedNorm4Short = 0.0f;
    float processedNorm1 = 0.0f;
    float processedNorm2 = 0.0f;
    float processedNorm3 = 0.0f;
    float processedNorm4 = 0.0f;
    float processedNorm4Short = 0.0f;
    float processedNorm1Pointer = 0.0f;
    float processedNorm2Pointer = 0.0f;
    float processedNorm3Pointer = 0.0f;
    float processedNorm4Pointer = 0.0f;

    // Expected norms
    for (int i = 0; i < vector1.size(); ++i) {
        expectedNorm1 += std::pow(vector1[i], 2.0f);
        expectedNorm2 += std::pow(vector2[i], 2.0f);
        expectedNorm3 += std::pow(vector3[i], 2.0f);
        expectedNorm4 += std::pow(vector4[i], 2.0f);
        if (i < 2) {
            expectedNorm4Short += std::pow(vector4[i], 2.0f);
        }
    }
    expectedNorm1 = std::sqrt(expectedNorm1);
    expectedNorm2 = std::sqrt(expectedNorm2);
    expectedNorm3 = std::sqrt(expectedNorm3);
    expectedNorm4 = std::sqrt(expectedNorm4);
    expectedNorm4Short = std::sqrt(expectedNorm4Short);

    // Processed norms
    processedNorm1 = calculate_vector_norm(vector1);
    processedNorm2 = calculate_vector_norm(vector2);
    processedNorm3 = calculate_vector_norm(vector3);
    processedNorm4 = calculate_vector_norm(vector4);
    processedNorm4Short = calculate_vector_norm_for_first_n_dimensions(vector4, 2);
    processedNorm1Pointer = calculate_vector_norm_for_first_n_dimensions(vector1.data(), 3);
    processedNorm2Pointer = calculate_vector_norm_for_first_n_dimensions(vector2.data(), 3);
    processedNorm3Pointer = calculate_vector_norm_for_first_n_dimensions(vector3.data(), 3);
    processedNorm4Pointer = calculate_vector_norm_for_first_n_dimensions(vector4.data(), 3);

    EXPECT_EQ(expectedNorm1, processedNorm1);
    EXPECT_EQ(expectedNorm2, processedNorm2);
    EXPECT_EQ(expectedNorm3, processedNorm3);
    EXPECT_EQ(expectedNorm4, processedNorm4);
    EXPECT_EQ(expectedNorm4Short, processedNorm4Short);
    EXPECT_EQ(expectedNorm1, processedNorm1Pointer);
    EXPECT_EQ(expectedNorm2, processedNorm2Pointer);
    EXPECT_EQ(expectedNorm3, processedNorm3Pointer);
    EXPECT_EQ(expectedNorm4, processedNorm4Pointer);
}

struct SphoaTestData{
    std::vector<float> currentListenerPosition;
    std::vector<float> previousListenerPosition;
    float projectionRadius;

    friend std::ostream& operator<<(std::ostream& os, const SphoaTestData& data)
    {
        return os
                << "currentListenerPosition: " << vecToString(data.currentListenerPosition)
                << ", previousListenerPosition: " << vecToString(data.previousListenerPosition)
                << ", projectionRadius: " << data.projectionRadius;
    }
};

class SphoaTest
        : public ::testing::TestWithParam<SphoaTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(SphoaTest,
       Sphoa6DofExtension) // NOLINT
{
    /* Test for the Single-point 6 Degree-of-Freedom HOA extension mode.
     * Test setting includes one mic array (m1 at (1, 1.7, 0) in OpenGL),
     * that captures sound from the frontal direction. The sphoa 6DoF
     * extension mode projects the source to a sphere determined by
     * projectionRadius. This test moves the listener around the field,
     * and checks that the new DOA of the incoming sound points towards
     * the projected source position on the sphere.
     */

    auto params = SphoaTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;

    // Single-point HOA 6DoF mode initializations
    float maxGainDb = 9.0f;
    h->useSphoa6DofExtensionMode = true;
    h->sphoa6DofMaxGain = std::pow(10.0f, (maxGainDb / 20.0f));
    h->sphoa6DofProjectionRadius = params.projectionRadius;

    // Only m1 array active (position (1, 1.7, 0) in OpenGL)
    int activeMicArray = 1;
    h->activeArraysForCurrentFrame.clear();
    h->activeArraysForCurrentFrame.push_back(activeMicArray);

    // Initialize input signal (sound coming from the front)
    int nFirstOrderChannels = 4;
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int ch = 0; ch < nFirstOrderChannels; ++ch) {
            for (int band = 0; band < h->nBands; ++band) {
                h->TFDInputFrameFullOrder[activeMicArray][frame][ch][band] = ACNfront[ch];
            }
        }
    }

    // Initialize listener positions
    h->currentListenerPosition = params.currentListenerPosition;
    h->previousListenerPosition = params.previousListenerPosition;

    // Expected unit intensity vector (= vector pointing front in cartesian)
    std::vector<float> expectedUnitIntensity{1.0f, 0.0f, 0.0f};

    // Expected original DOA
    std::vector<float> originalDoa(3, 0.0f);
    for (int i = 0; i < originalDoa.size(); ++i) {
        originalDoa[i] = expectedUnitIntensity[i] * h->sphoa6DofProjectionRadius;
    }

    // Interpolated listener position in mic array space
    std::vector<std::vector<float>> interpolatedListenerPositions(h->nFrames, std::vector<float>(3, 0.0f));
    for (int frame = 0; frame < h->nFrames; ++frame) {
        get_interpolated_listener_position_in_array_space(h, activeMicArray, frame, interpolatedListenerPositions[frame]);
    }

    // Expected new DOAs
    std::vector<std::vector<float>> newDoas(h->nFrames, std::vector<float>(3, 0.0f));
    for (int frame = 0; frame < h->nFrames; ++frame) {
        for (int i = 0; i < newDoas[frame].size(); ++i) {
            newDoas[frame][i] = originalDoa[i] - interpolatedListenerPositions[frame][i];
        }
    }

    // Expected azimuth and elevation DOA angles
    std::vector<float> expectedAzimuths(h->nFrames, 0.0f);
    std::vector<float> expectedElevations(h->nFrames, 0.0f);
    for (int frame = 0; frame < h->nFrames; ++frame) {
        expectedAzimuths[frame] = calculate_azimuth(newDoas[frame]);
        expectedElevations[frame] = calculate_elevation(newDoas[frame]);
    }

    // Estimate DOAs for the sphoa mode
    spatial_analysis(h);

    // Fetch the processed DOAs
    std::vector<float> processedAzimuthVec(h->nFrames, 0.0f);
    std::vector<float> processedElevationVec(h->nFrames, 0.0f);
    for (int frame = 0; frame < h->nFrames; ++frame) {
        processedAzimuthVec[frame] = h->azimuthData[activeMicArray][frame][0];
        processedElevationVec[frame] = h->elevationData[activeMicArray][frame][0];
    }

    // Compare
    EXPECT_EQ(expectedAzimuths, processedAzimuthVec);
    EXPECT_EQ(expectedElevations, processedElevationVec);
}

INSTANTIATE_TEST_SUITE_P(Default, SphoaTest,
                         testing::Values(
                                 // Test case: the listener is on top of the mic array
                                 SphoaTestData{ std::vector<float>{1.0f, 1.7f, 0.0f},
                                                std::vector<float>{1.0f, 1.7f, 0.0f},
                                                2.0f },

                                 // Test case: the listener moves to the right edge
                                 SphoaTestData{ std::vector<float>{2.0f, 1.7f, 0.0f},
                                                std::vector<float>{1.0f, 1.7f, 0.0f},
                                                2.0f },

                                 // Test case: the listener moves downwards
                                 SphoaTestData{ std::vector<float>{1.0f, -0.3f, 0.0f},
                                                std::vector<float>{1.0f, 1.7f, 0.0f},
                                                2.0f },

                                 // Test case: the listener moves to the left edge, large projection radius
                                 SphoaTestData{ std::vector<float>{-20.0f, 1.7f, 0.0f},
                                                std::vector<float>{1.0f, 1.7f, 0.0f},
                                                20.0f },

                                 // Test case: the listener moves to the left edge, small projection radius
                                 SphoaTestData{ std::vector<float>{-0.5f, 1.7f, 0.0f},
                                                std::vector<float>{1.0f, 1.7f, 0.0f},
                                                0.5f }
                                 ));

struct ExteriorTestData{};

class ExteriorTest
        : public ::testing::TestWithParam<ExteriorTestData>
{
public:
    void SetUp(const HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT &micArrayArrangement) {
        hoaTranslator = new HoaTranslatorTest(micArrayArrangement);
    }
    HoaTranslatorTest* hoaTranslator;
};

TEST_P(ExteriorTest,
       ExteriorGeometry) // NOLINT
{
    /* Tests for the exterior rendering geometry. The mic arrays are arranged as
     * described in the HoaTranslatorTest class in the beginning of this file
     * (5 FOA microphones (o) in a diamond shape arrangement). Test that:
     *
     *      - the number of exterior edges match
     *      - the exterior edges are formed from the correct mic arrays
     *      - the exterior edges are ordered in a traversing order
     *      - the edge and vertex normals point to expected directions
     */

    SetUp(HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT::DIAMOND);
    auto h = &*hoaTranslator->m_hoaTranslatorData;

    // Set the expected values
    int nExpectedExteriorEdges = 4;
    std::vector<std::vector<int>> expectedExteriorEdges{ {1, 3},
                                                         {1, 4},
                                                         {2, 3},
                                                         {2, 4} };
    float normalValue = 1.0f / std::sqrt(2.0f);
    std::vector<std::vector<float>> expectedEdgeNormals{ { normalValue,  0.0f,  normalValue},
                                                         { normalValue,  0.0f, -normalValue},
                                                         {-normalValue,  0.0f,  normalValue},
                                                         {-normalValue,  0.0f, -normalValue} };
    std::vector<int> vertexOrder{1, 2, 3, 4};
    std::vector<std::vector<float>> expectedVertexNormals{ { 1.0f, 0.0f,  0.0f},
                                                           {-1.0f, 0.0f,  0.0f},
                                                           { 0.0f, 0.0f,  1.0f},
                                                           { 0.0f, 0.0f, -1.0f} };

    // Test that there are expected amount of exterior edges
    EXPECT_EQ(nExpectedExteriorEdges, h->nExteriorEdges);

    // Test that the expected exterior edges are found in the exterior edge list (in some order)
    for (auto edge : h->exteriorEdges) {
        std::sort(edge.begin(), edge.end());
        EXPECT_TRUE(vectorAcontainsB(expectedExteriorEdges, edge));
    }

    // Test that the exterior edge list is sorted in a traversing order
    int previousLastVertex = 0;
    int currentFirstVertex = 0;
    int previousEdgeIdx = 0;
    for (int edgeIdx = 0; edgeIdx < h->nExteriorEdges; ++edgeIdx) {
        if (edgeIdx == 0) {
            previousEdgeIdx = h->nExteriorEdges - 1;
        } else {
            previousEdgeIdx = edgeIdx - 1;
        }
        previousLastVertex = h->exteriorEdges[previousEdgeIdx][1];
        currentFirstVertex = h->exteriorEdges[edgeIdx][0];
        EXPECT_EQ(previousLastVertex, currentFirstVertex);
    }

    // Collect indexes of corresponding edges and vertices in the expected normal vectors
    std::vector<int> expectedEdgeIdxList{};
    std::vector<int> expectedVertexIdxList{};
    for (int edgeIdx = 0; edgeIdx < h->nExteriorEdges; ++edgeIdx) {
        auto edge = h->exteriorEdges[edgeIdx];
        std::sort(edge.begin(), edge.end());
        auto expectedEdgeIt = std::find(expectedExteriorEdges.begin(), expectedExteriorEdges.end(), edge);
        assert(expectedEdgeIt != expectedExteriorEdges.end() && "Couldn't find edge in the expected exterior edge vector!");
        int expectedEdgeIdx = std::distance(expectedExteriorEdges.begin(), expectedEdgeIt);
        expectedEdgeIdxList.push_back(expectedEdgeIdx);

        int vertex = h->exteriorEdges[edgeIdx][0];
        auto vertexIt = std::find(vertexOrder.begin(), vertexOrder.end(), vertex);
        assert(vertexIt != vertexOrder.end() && "Couldn't find vertex in the expected vertex order vector!");
        int vertexIdx = std::distance(vertexOrder.begin(), vertexIt);
        expectedVertexIdxList.push_back(vertexIdx);
    }

    // Test that the edge normals point to the expected directions
    std::vector<float> edgeNormalVec(3, 0.0f);
    std::vector<int> edgeNormalInt(3, 0);
    std::vector<int> expectedEdgeNormalInt(3, 0);
    for (int edgeIdx = 0; edgeIdx < h->nExteriorEdges; ++edgeIdx) {
        edgeNormalVec = h->edgeNormals[edgeIdx];
        auto expectedEdgeNormal = expectedEdgeNormals[expectedEdgeIdxList[edgeIdx]];

        // Convert to int and compare
        convertTo2DecimalIntVec(edgeNormalVec, edgeNormalInt);
        convertTo2DecimalIntVec(expectedEdgeNormal, expectedEdgeNormalInt);
        EXPECT_EQ(expectedEdgeNormalInt, edgeNormalInt);
    }

    // Test that the vertex normals point to expected directions
    std::vector<float> vertexNormalVec(3, 0.0f);
    std::vector<int> vertexNormalInt(3, 0);
    std::vector<int> expectedVertexNormalInt(3, 0);
    for (int vertexIdx = 0; vertexIdx < h->nExteriorEdges; ++vertexIdx) {
        vertexNormalVec =h->micNormals[vertexIdx];
        auto expectedVertexNormal = expectedVertexNormals[expectedVertexIdxList[vertexIdx]];

        // Convert to int and compare
        convertTo2DecimalIntVec(vertexNormalVec, vertexNormalInt);
        convertTo2DecimalIntVec(expectedVertexNormal, expectedVertexNormalInt);
        EXPECT_EQ(expectedVertexNormalInt, vertexNormalInt);
    }
}

TEST_P(ExteriorTest,
       InterpolatedNormals) // NOLINT
{
    // Test that the exterior normals are interpolated correctly at the projected listener positions

    SetUp(HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT::DIAMOND);
    auto h = &*hoaTranslator->m_hoaTranslatorData;

    // Set the listener positions
    std::vector<std::vector<float>> listenerPositions{ {0.6f, 1.7f, -0.6f},
                                                       {0.9f, 1.7f, -0.3f},
                                                       {0.0f, 1.7f, -1.2f} };

    // Set the expected values
    std::vector<std::vector<float>> expectedVertexNormals{ { 1.0f, 0.0f,  0.0f},
                                                           {-1.0f, 0.0f,  0.0f},
                                                           { 0.0f, 0.0f,  1.0f},
                                                           { 0.0f, 0.0f, -1.0f} };
    std::vector<std::vector<float>> expectedListenerNormals(listenerPositions.size(), std::vector<float>(3, 0.0f));

    // The first projected listener position is between m1-m4, and the normal should point to the middle of the edge
    expectedListenerNormals[0] = sumVectors(expectedVertexNormals[0], expectedVertexNormals[3]);
    normalizeVector(expectedListenerNormals[0]);

    // The second projected listener position is closer to m1 than m4, and the normal should point more towards the normal direction of m1
    h->distToClosestArraySquared = FLT_MAX;
    h->distToClosestEdgeSquared = FLT_MAX;
    h->isClosestEdgeFound = false;
    h->currentListenerPosition = listenerPositions[1];
    find_closest_edge(h);
    auto normalM1 = expectedVertexNormals[0];
    auto normalM4 = expectedVertexNormals[3];
    normalM1 = multiplyVector((std::max)(h->edgeWeights[0], h->edgeWeights[1]), normalM1);
    normalM4 = multiplyVector((std::min)(h->edgeWeights[0], h->edgeWeights[1]), normalM4);
    expectedListenerNormals[1] = sumVectors(normalM1, normalM4);
    normalizeVector(expectedListenerNormals[1]);

    // The third projected listener position is on top of m4, and the normal should also be the same as with m4
    expectedListenerNormals[2] = expectedVertexNormals[3];

    // Loop over the listener positions; calculate the interpolated normals at the projected listener positions
    std::vector<float> processedListenerNormal(3, 0.0f);
    std::vector<int> processedListenerNormalInt(3, 0);
    std::vector<int> expectedListenerNormalInt(3, 0);
    for (int i = 0; i < listenerPositions.size(); ++i) {
        h->currentListenerPosition = listenerPositions[i];
        reset_exterior_variables(h);
        project_to_closest_edge_or_array(h);
        processedListenerNormal = h->listenerExteriorNormal;


        // Convert to int and compare
        convertTo2DecimalIntVec(processedListenerNormal, processedListenerNormalInt);
        convertTo2DecimalIntVec(expectedListenerNormals[i], expectedListenerNormalInt);
        EXPECT_EQ(processedListenerNormalInt, expectedListenerNormalInt);
    }
}

TEST_P(ExteriorTest,
       AdjustMetadata) // NOLINT
{
    /* Test for adjusting DTR and DOA angles outside the capturing area. The different test cases are formed by setting
     * different listener positions, source DOAs and original DTRs. The expected exterior weights, weighted DOA angles
     * and weighted DTR values are first manually calculated. These values are also calculated with the library functions,
     * using the expected DOAs, normals and listener positions as input. Comparing the output to manually calculated
     * values assures, that the individual functions work properly. Finally the whole processing pipeline is tested, by
     * comparing the processing output to the expected output values.
     *
     * The weights are calculated as:
     *      - w1 = 1/2^2 * (1 + dot(originalDoaVec, listenerExteriorNormal))^2
     *      - w2 = 1.0 - w1
     *
     * The re-mapped DOA is calculated as (all in horizontal plane):
     *      - re-mapped DOA = (projectedListenerPositionVec) + (projectionRadius * originalDoaVec) - (listenerPositionVec)
     *
     * The re-mapped DTR is calculated as:
     *      - re-mapped DTR = min(1.0, w1 * min(originalDtr, (reMappedDoaLen / projectionRadius)) + w2 * originalDTR )
     *
     * Projection radius = 4
     *
     * Test case 1:
     *      - listener position (0.6, 1.7 -0.6), middle of m1 and m4
     *      - source DOA (-45, 0) degrees
     *      - original DTR = 1.0
     *      - originalDoaVec = (1/sqrt(2), -1/sqrt(2))
     *      - listenerExteriorNormal = (1/sqrt(2), -1/sqrt(2))
     *      - projectedListenerPositionVec = (0.5, -0.5)
     *      - listenerPositionVec = (0.6, -0.6)
     *      - expected weights: w1 = 1.0, w2 = 0.0
     *      - re-mapped DOA vector = (0.5, -0.5) + 4*(1/sqrt(2), -1/sqrt(2)) - (0.6, -0.6) = (2.728, -2.728)
     *      - re-mapped DOA length = sqrt(2.728^2 + (-2.728)^2) = 3.8586
     *      - normalized re-mapped DOA vector = (2.728, -2.728) / 3.8586 = (1/sqrt(2), -1/sqrt(2))
     *      - expected re-mapped DTR = (1.0 * 3.8586 / 4) + (0.0 * 1.0) = 0.9646
     *      - expected re-mapped DOA (-45, 0) (no change expected)
     *
     * Test case 2:
     *      - listener position (1.0, 1.7 -1.0), middle of m1 and m4
     *      - source DOA (0, 0) degrees
     *      - original DTR = 1.0
     *      - originalDoaVec = (0, -1)
     *      - listenerExteriorNormal = (1/sqrt(2), -1/sqrt(2))
     *      - projectedListenerPositionVec = (0.5, -0.5)
     *      - listenerPositionVec = (1.0, -1.0)
     *      - expected weights: w1 = 0.7286, w2 = 0.2714
     *      - re-mapped DOA vector = (0.5, -0.5) + (0, -4.0) - (1.0, -1.0) = (-0.5, -3.5)
     *      - re-mapped DOA length = sqrt((-0.5)^2 + (-3.5)^2) = 3.5355
     *      - normalized re-mapped DOA vector = (-0.5, -3.5) / 3.5355 = (-0.1414, -0.9900)
     *      - final re-mapped DOA vector = 0.7286 * (-0.1414, -0.9900) + 0.2714 * (0, -1) = (-0.1030, -0.9927)
     *      - expected re-mapped DTR = (0.7286 * 3.5355 / 4) + (0.2714 * 1.0) = 0.9154
     *      - expected re-mapped DOA = ( atan2(0.1030, 0.9927), 0) = (5.9, 0)
     *
     * Test case 3:
     *      - listener position (0.3, 1.7 -2.0), projected to m4
     *      - source DOA (-160, 0) degrees
     *      - original DTR = 1.0
     *      - originalDoaVec = (0.342, 0.940)
     *      - listenerExteriorNormal = (0.0, -1.0)
     *      - projectedListenerPositionVec = (0.0, -1.0)
     *      - listenerPositionVec = (0.3, -2.0)
     *      - expected weights: w1 = 0.0009, w2 = 0.9991
     *      - re-mapped DOA vector = (0.0, -1.0) + 4*(0.342, 0.940) - (0.3, -2.0) = (1.068, 4.760)
     *      - re-mapped DOA length = sqrt((-1.068)^2 + 4.760^2) = 4.8783
     *      - normalized re-mapped DOA vector = (1.068, 4.760) / 4.8783 = (0.2189, 0.9757)
     *      - final re-mapped DOA vector = 0.0009 * (0.2189, 0.9757) + 0.9991 * (0.342, 0.940) = (0.3419, 0.940)
     *      - expected re-mapped DTR = (0.0009 * min(1.0f, 4.8783 / 4)) + (0.9991 * 1.0) = 1.0
     *      - expected re-mapped DOA = ( atan2(-0.3419, -0.940), 0) = (-160.0125, 0)
     */

    SetUp(HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT::DIAMOND);
    auto h = &*hoaTranslator->m_hoaTranslatorData;

    // Set the listener positions, DOA angles, DTR values and projection radius
    std::vector<std::vector<float>> listenerPositions{ {0.6f, 1.7f, -0.6f},
                                                       {1.0f, 1.7f, -1.0f},
                                                       {0.0f, 1.7f, -1.2f} };
    std::vector<std::vector<float>> initialDoaAnglesDeg{ { -45.0f, 0.0f},
                                                         {   0.0f, 0.0f},
                                                         {-160.0f, 0.0f} };
    std::vector<float> initialDtrValues{ 1.0f, 1.0f, 1.0f };
    h->exteriorProjectionRadius = 4.0f;

    // Set the expected values
    std::vector<std::vector<float>> expectedOriginalDoaVecs{ {1.0f/std::sqrt(2.0f), 0.0f, -1.0f/std::sqrt(2.0f)},
                                                             {0.0f,   0.0f, -1.0f},
                                                             {0.342f, 0.0f,  0.94f} };
    std::vector<std::vector<float>> expectedListenerNormals{ {1.0f/std::sqrt(2.0f), 0.0f, -1.0f/std::sqrt(2.0f)},
                                                             {1.0f/std::sqrt(2.0f), 0.0f, -1.0f/std::sqrt(2.0f)},
                                                             {0.0f, 0.0f, -1.0f} };
    std::vector<std::vector<float>> expectedProjectedListenerPositions{ {0.5f, 0.0f, -0.5f},
                                                                        {0.5f, 0.0f, -0.5f},
                                                                        {0.0f, 0.0f, -1.0f} };

    // Set hand-calculated values
    std::vector<std::vector<float>> handCalculatedExteriorWeights{ {1.0f,    0.0f},
                                                                   {0.7286f, 0.2714f},
                                                                   {0.0009f, 0.9991f} };
    std::vector<std::vector<float>> handCalculatedDoaAnglesDeg{ {-45.0f, 0.0f},
                                                                {  5.9f, 0.0f},
                                                                {-160.0125f, 0.0f} };
    std::vector<float> handCalculatedDtrValues{ 0.9646f, 0.9154f, 1.0f };

    std::vector<float> expectedAzimuthRad(h->nBands, 0.0f);
    std::vector<float> expectedElevationRad(h->nBands, 0.0f);
    std::vector<float> expectedDtr(h->nBands, 0.0f);
    std::vector<float> exteriorWeights1(h->nBands, 0.0f);
    std::vector<float> exteriorWeights2(h->nBands, 0.0f);
    std::vector<float> horizontalListenerPosition(3, 0.0f);
    std::vector<float> reMappedDoaVec(3, 0.0f);
    std::vector<float> expectedOriginalDoaX(h->nBands, 0.0f);
    std::vector<float> expectedOriginalDoaY(h->nBands, 0.0f);
    std::vector<float> expectedOriginalDoaZ(h->nBands, 0.0f);
    std::vector<float> initialDtrValuesVec(h->nBands, 0.0f);
    std::vector<float> handCalculatedw1Vec(h->nBands, 0.0f);
    std::vector<float> handCalculatedw2Vec(h->nBands, 0.0f);
    std::vector<float> handCalculatedAziVec(h->nBands, 0.0f);
    std::vector<float> handCalculatedElevVec(h->nBands, 0.0f);
    std::vector<float> handCalculatedDtrVec(h->nBands, 0.0f);
    std::vector<float> processedAzimuthDegVec(h->nBands, 0.0f);
    std::vector<float> processedElevationDegVec(h->nBands, 0.0f);
    std::vector<float> processedDtrVec(h->nBands, 0.0f);
    std::vector<int> expectedOriginalDoaXInt(h->nBands, 0);
    std::vector<int> expectedOriginalDoaYInt(h->nBands, 0);
    std::vector<int> expectedOriginalDoaZInt(h->nBands, 0);
    std::vector<int> initialDtrValuesVecInt(h->nBands, 0);
    std::vector<int> handCalculatedw1VecInt(h->nBands, 0.);
    std::vector<int> handCalculatedw2VecInt(h->nBands, 0);
    std::vector<int> handCalculatedAziVecInt(h->nBands, 0);
    std::vector<int> handCalculatedElevVecInt(h->nBands, 0);
    std::vector<int> handCalculatedDtrVecInt(h->nBands, 0);
    std::vector<int> exteriorWeightsw1Int(h->nBands, 0);
    std::vector<int> exteriorWeightsw2Int(h->nBands, 0);
    std::vector<int> expectedAzimuthDegInt(h->nBands, 0);
    std::vector<int> expectedElevationDegInt(h->nBands, 0);
    std::vector<int> expectedDtrInt(h->nBands, 0);
    std::vector<int> processedAzimuthDegVecInt(h->nBands, 0);
    std::vector<int> processedElevationDegVecInt(h->nBands, 0);
    std::vector<int> processedDtrVecInt(h->nBands, 0);

    for (int i = 0; i < listenerPositions.size(); ++i) {

        // Fill vectors
        std::fill_n(expectedOriginalDoaX.data(),  h->nBands, expectedOriginalDoaVecs[i][0]);
        std::fill_n(expectedOriginalDoaY.data(),  h->nBands, expectedOriginalDoaVecs[i][1]);
        std::fill_n(expectedOriginalDoaZ.data(),  h->nBands, expectedOriginalDoaVecs[i][2]);
        std::fill_n(initialDtrValuesVec.data(),   h->nBands, initialDtrValues[i]);
        std::fill_n(handCalculatedw1Vec.data(),   h->nBands, handCalculatedExteriorWeights[i][0]);
        std::fill_n(handCalculatedw2Vec.data(),   h->nBands, handCalculatedExteriorWeights[i][1]);
        std::fill_n(handCalculatedAziVec.data(),  h->nBands, handCalculatedDoaAnglesDeg[i][0]);
        std::fill_n(handCalculatedElevVec.data(), h->nBands, handCalculatedDoaAnglesDeg[i][1]);
        std::fill_n(handCalculatedDtrVec.data(),  h->nBands, handCalculatedDtrValues[i]);

        // Convert to integers
        convertTo2DecimalIntVec(expectedOriginalDoaX,  expectedOriginalDoaXInt);
        convertTo2DecimalIntVec(expectedOriginalDoaY,  expectedOriginalDoaYInt);
        convertTo2DecimalIntVec(expectedOriginalDoaZ,  expectedOriginalDoaZInt);
        convertTo2DecimalIntVec(initialDtrValuesVec,   initialDtrValuesVecInt);
        convertTo2DecimalIntVec(handCalculatedw1Vec,   handCalculatedw1VecInt);
        convertTo2DecimalIntVec(handCalculatedw2Vec,   handCalculatedw2VecInt);
        convertToNDecimalIntVec(handCalculatedAziVec,  handCalculatedAziVecInt, 1);
        convertToNDecimalIntVec(handCalculatedElevVec, handCalculatedElevVecInt, 1);
        convertTo2DecimalIntVec(handCalculatedDtrVec,  handCalculatedDtrVecInt);

        // Set the horizontal listener position
        horizontalListenerPosition = listenerPositions[i];
        horizontalListenerPosition[1] = 0.0f;

        // Set the initial values for processing
        h->currentListenerPosition = listenerPositions[i];
        std::fill_n(&h->interpolatedAzimuthData[0][0], h->nFrames * h->nBands, initialDoaAnglesDeg[i][0] * M_PI / 180.0f);
        std::fill_n(&h->interpolatedElevationData[0][0], h->nFrames * h->nBands, initialDoaAnglesDeg[i][1] * M_PI / 180.0f);
        std::fill_n(&h->interpolatedDTRData[0][0], h->nFrames * h->nBands, initialDtrValues[i]);
        std::copy_n(&expectedListenerNormals[i][0], 3, &h->listenerExteriorNormal[0]);
        std::copy_n(expectedOriginalDoaX.data(), expectedOriginalDoaX.size(), &h->originalDoaUnitX[0]);
        std::copy_n(expectedOriginalDoaY.data(), expectedOriginalDoaY.size(), &h->originalDoaUnitY[0]);
        std::copy_n(expectedOriginalDoaZ.data(), expectedOriginalDoaZ.size(), &h->originalDoaUnitZ[0]);

        // Compute the exterior weights and compare to hand-calculated values
        calculate_exterior_weights(h,
                                   expectedOriginalDoaX.size());
        std::copy_n(&h->w1[0], exteriorWeights1.size(), exteriorWeights1.data());
        std::copy_n(&h->w2[0], exteriorWeights2.size(), exteriorWeights2.data());
        convertTo2DecimalIntVec(exteriorWeights1, exteriorWeightsw1Int);
        convertTo2DecimalIntVec(exteriorWeights2, exteriorWeightsw2Int);
        EXPECT_EQ(exteriorWeightsw1Int, handCalculatedw1VecInt);
        EXPECT_EQ(exteriorWeightsw2Int, handCalculatedw2VecInt);

        // Re-map DOA and DTR
        re_map_doa_and_dtr(h,
                           0,
                           expectedProjectedListenerPositions[i],
                           horizontalListenerPosition,
                           expectedOriginalDoaX.size());

        // Apply exterior weighting
        apply_exterior_weighting(h, 0, expectedOriginalDoaX.size());

        // Collect the values and compare to hand-calculated values
        std::copy_n(&h->reMappedDtr[0], expectedDtr.size(), expectedDtr.data());
        calculate_azimuths_OpenGL(h, h->reMappedDoaX, h->reMappedDoaZ, expectedAzimuthRad.data(), expectedAzimuthRad.size());
        calculate_elevations_OpenGL(h, h->reMappedDoaX, h->reMappedDoaY, h->reMappedDoaZ, expectedElevationRad.data(), expectedElevationRad.size());
        auto expectedAzimuthDeg = multiplyVector(180.0f / M_PI, expectedAzimuthRad);
        auto expectedElevationDeg = multiplyVector(180.0f / M_PI, expectedElevationRad);
        convertToNDecimalIntVec(expectedAzimuthDeg, expectedAzimuthDegInt, 1);
        convertToNDecimalIntVec(expectedElevationDeg, expectedElevationDegInt, 1);
        convertTo2DecimalIntVec(expectedDtr, expectedDtrInt);
        EXPECT_EQ(expectedAzimuthDegInt, handCalculatedAziVecInt);
        EXPECT_EQ(expectedElevationDegInt, handCalculatedElevVecInt);
        EXPECT_EQ(expectedDtrInt, handCalculatedDtrVecInt);

        // Re-initialize values for processing
        std::fill_n(&h->interpolatedAzimuthData[0][0], h->nFrames * h->nBands, initialDoaAnglesDeg[i][0] * M_PI / 180.0f);
        std::fill_n(&h->interpolatedElevationData[0][0], h->nFrames * h->nBands, initialDoaAnglesDeg[i][1] * M_PI / 180.0f);
        std::fill_n(&h->interpolatedDTRData[0][0], h->nFrames * h->nBands, initialDtrValues[i]);
        std::copy_n(expectedOriginalDoaX.data(), expectedOriginalDoaX.size(), &h->originalDoaUnitX[0]);
        std::copy_n(expectedOriginalDoaY.data(), expectedOriginalDoaY.size(), &h->originalDoaUnitY[0]);
        std::copy_n(expectedOriginalDoaZ.data(), expectedOriginalDoaZ.size(), &h->originalDoaUnitZ[0]);

        // Process
        reset_exterior_variables(h);
        project_to_closest_edge_or_array(h);
        adjust_metadata_outside_capturing_area(h, false);

        // Collect the processed values and compare to the expected values
        for (int frame = 0; frame < h->nFrames; ++frame) {
            std::copy_n(&h->interpolatedAzimuthData[frame][0], h->nBands, processedAzimuthDegVec.data());
            std::copy_n(&h->interpolatedElevationData[frame][0], h->nBands, processedElevationDegVec.data());
            std::copy_n(&h->interpolatedDTRData[frame][0], h->nBands, processedDtrVec.data());
            processedAzimuthDegVec = multiplyVector(180.0f / M_PI, processedAzimuthDegVec);
            processedElevationDegVec = multiplyVector(180.0f / M_PI, processedElevationDegVec);
            convertToNDecimalIntVec(processedAzimuthDegVec, processedAzimuthDegVecInt, 1);
            convertToNDecimalIntVec(processedElevationDegVec, processedElevationDegVecInt, 1);
            convertTo2DecimalIntVec(processedDtrVec, processedDtrVecInt);
            EXPECT_EQ(expectedAzimuthDegInt, processedAzimuthDegVecInt);
            EXPECT_EQ(expectedElevationDegInt, processedElevationDegVecInt);
            EXPECT_EQ(expectedDtrInt, processedDtrVecInt);
        }
    }
}

TEST_P(ExteriorTest,
       NonConvexHulls) // NOLINT
{
    /* Test that exterior rendering works as expected, when some of the mic triangle edges are omitted.
     * Omitting triangle edges results in a non-convex hull shape for the exterior area, see below.
     * The exterior rendering in a non-convex hull case works as following:
     *      - Project the listener to a closest original exterior edge. (e.g. m0-m5)
     *      - If the projected edge is an omitted one, further project from the projected position
     *        to the final exterior edges. (e.g. from some position on edge m0-m5 to m5-m6)
     *      - Both of these projections are done parallel to the normal of the original exterior edge.
     *        (e.g. project parallel to the normal of edge m0-m5)
     *      - The normal of the final projected edge is set to be the same as the omitted edge normal,
     *        from where the listener was projected from. (e.g. edge normal of m5-m6 is set to be the same
     *        as m0-m5 edge normal)
     *
     * 8 FOA microphones (o) in an arrangement described below. The max edge length is set to 4m.
     * This restriction omits edges m0-m5 and m5-m7. When the listener is one the left side of edge m0-m5,
     * the listener should first be projected to m0-m5 edge. After that, the listener is further projected
     * to some edge (m5-m6, m6-m7, m7-m0), using the projected position on edge m0-m5 as the starting listener
     * position. Both projections are done parallel to the normal direction of edge m0-m5.
     *
     *  Before omitting long edges              After omitting long edges
     *     m0         m1                            m0         m1
     *     o - - - - o                  -            o - - - - o
     *     | \       | \                |              \       | \
     *     |    \    |    \             2m                \    |    \
     *     |       \ |       \          |                    \ |       \
     *     |      m7 o - - - - o m2     -                   m7 o - - - - o m2
     *     |      /  |        /|        |                      |        /|
     *     |     /   |     /   |        2m                     |     /   |
     *     |    /    |  /      |        |                      |  /      |
     *     |   /  m6 o - - - - o m3     -                   m6 o - - - - o m3
     *     |  /    / |       /          |                    / |       /
     *     | /  /    |    /             2m                /    |    /
     *     |//       | /                |              /       | /
     *     o - - - - o                  -            o - - - - o
     *     m5        m4                              m5        m4
     *     |----2m---|----2m---|
     *
     * Test cases:
     *  Test 1:
     *      - listener position: (-6.0, 1.7, -5.0)
     *      - expected weights: m5(0.5), m6(0.5) (expect to project to middle of edge m5-m6)
     *
     *  Test 2:
     *      - listener position: (-6.0, 1.7, 2.5)
     *      - expected weights: m6(0.25), m7(0.75) (expect to project to edge m6-m7, 0.5m away from m7)
     *
     *  Test 3:
     *      - listener position: (-6.0, 1.7, 2.0)
     *      - expected weights: m7(1.0) (expect to project to m7)
     *
     *  Test 4:
     *      - listener position: (0.5, 1.7, 1.5)
     *      - expected weights: m0(0.25), m7(0.75) (inside omitted triangle, expect to project to edge m0-m7,
     *        1/4 length away from m7 on edge m0-m7)
     *
     *  Test 5:
     *      - listener position: (-1.0, 1.7, -1.0)
     *      - expected weights: m0(1.0) (expect to project to m0)
     *
     *  Test 6:
     *      - listener position: (4.0, 1.7, 0.0)
     *      - expected weights: m1(0.5), m2(0.5) (expect to project to middle of edge m1-m2)
     */

    SetUp(HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT::NON_CONVEX_HULL);
    auto h = &*hoaTranslator->m_hoaTranslatorData;

    std::vector<float> processedWeights(h->nMicrophoneArrays, 0.0f);
    std::vector<int> processedWeightsInt(h->nMicrophoneArrays, 0.0f);
    std::vector<int> expectedWeightsInt(h->nMicrophoneArrays, 0.0f);

    std::vector<std::vector<float>> listenerPositions{ {-6.0f, 1.7f,  5.0f},
                                                       {-6.0f, 1.7f,  2.5f},
                                                       {-6.0f, 1.7f,  2.0f},
                                                       { 0.5f, 1.7f,  1.5f},
                                                       {-1.0f, 1.7f, -1.0f},
                                                       { 4.0f, 1.7f,  0.0f} };

    //                                                  m0     m1    m2    m3    m4    m5    m6     m7
    std::vector<std::vector<float>> expectedWeights{ { 0.0f,  0.0f, 0.0f, 0.0f, 0.0f, 0.5f, 0.5f,  0.0f  },
                                                     { 0.0f,  0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.25f, 0.75f },
                                                     { 0.0f,  0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,  1.0f  },
                                                     { 0.25f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,  0.75f },
                                                     { 1.0f,  0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,  0.0f  },
                                                     { 0.0f,  0.5f, 0.5f, 0.0f, 0.0f, 0.0f, 0.0f,  0.0f  } };

    for (int i = 0; i < listenerPositions.size(); ++i) {
        h->currentListenerPosition = listenerPositions[i];
        pre_process_positions(h);
        std::copy_n(&h->currentInterpolationWeights[0], h->nMicrophoneArrays, &processedWeights[0]);
        convertTo2DecimalIntVec(processedWeights, processedWeightsInt);
        convertTo2DecimalIntVec(expectedWeights[i], expectedWeightsInt);
        EXPECT_EQ(processedWeightsInt, expectedWeightsInt);
    }
}

INSTANTIATE_TEST_SUITE_P(Default, ExteriorTest,
                         testing::Values(ExteriorTestData {}));

struct DecoderTestData{
    bool useRecursiveAveraging;
};

class DecoderTest
        : public ::testing::TestWithParam<DecoderTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(DecoderTest,
       DISABLED_Decoder) // NOLINT
{
    /* Decode a test binary file, and check that the decoded
     * values match with the expected ones. The test binary
     * file is created with create_test_bitstream_for_hoa_translator.m
     * in mpeg-i-development/matlab/EIF repository.
     */

    auto params = DecoderTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    decoderData* d = &h->decoderData1;

    h->useRecursiveAveraging = params.useRecursiveAveraging;
    if (h->useRecursiveAveraging) {
        h->spatialMetadataByteStream = std::make_unique<hoatranslation::ByteStream>(read_file_to_vector("data/test_6dof_metadata_recursive_averaging.bin"));
    }
    std::vector<std::vector<float>> expectedAzimuthVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> expectedElevationVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> expectedEnergyVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> expectedDTRVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedAzimuthVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedElevationVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedEnergyVec(h->nFrames, std::vector<float>(h->nBands,0.0f));
    std::vector<std::vector<float>> processedDTRVec(h->nFrames, std::vector<float>(h->nBands,0.0f));

    float errorMargin = 0.01f;
    h->useEncodedValues = true;
    h->trackedCurrentTriangleArrays[0].clear();
    h->trackedCurrentTriangleArrays[0] = {0, 1, 2, 3, 4};
    int nDecodedArrays = 11;
    h->activeArraysForCurrentFrame.clear();
    for (int mic = 0; mic < nDecodedArrays; ++mic) {
        h->activeArraysForCurrentFrame.push_back(mic);
    }
    read_spatial_metadata_info(h, false);

    // Update values based on test metadata
    h->nMicrophoneArrays = d->num_hoa_sources_in_group[h->HOAGroupId];
    h->nFrames = h->useRecursiveAveraging ? h->nFrames : d->subframes_per_frame;
    h->nBands = d->num_frequency_bands;
    read_spatial_metadata_from_binary(h);

    // 6DoFBitStream()
    std::uint32_t expected_audio_frame_duration = 512;
    std::uint32_t expected_subframes_duration = 128;
    std::uint32_t expected_sampling_rate = 48000;
    std::uint8_t expected_num_hoa_groups = 3;
    EXPECT_EQ(d->audio_frame_duration, expected_audio_frame_duration);
    EXPECT_EQ(d->subframes_duration, expected_subframes_duration);
    EXPECT_EQ(d->sampling_rate, expected_sampling_rate);
    EXPECT_EQ(d->num_hoa_groups, expected_num_hoa_groups);

    // HOAGroupData()
    std::vector<std::uint8_t> expected_hoa_source_group_id_length = {11, 11, 11};
    std::vector<std::string> expected_hoaSourceGroupIdStrs = {"grp:HOAGRP1","grp:HOAGRP2","grp:HOAGRP3"};

    std::vector<std::uint8_t> expected_num_hoa_sources_in_groups = {3, 5, 3};
    EXPECT_EQ(d->hoa_source_group_id_lengths, expected_hoa_source_group_id_length);
    EXPECT_EQ(d->hoaSourceGroupIdStrs, expected_hoaSourceGroupIdStrs);
    EXPECT_EQ(d->num_hoa_sources_in_group, expected_num_hoa_sources_in_groups);

    // HOASourceData()
    std::vector<std::uint8_t> expected_hoa_orders_group0 = {3, 3, 3};
    std::vector<std::uint8_t> expected_hoa_orders_group1 = {3, 3, 3, 3, 3};
    std::vector<std::uint8_t> expected_hoa_orders_group2 = {3, 3, 3};
    EXPECT_EQ(d->hoa_orders[0], expected_hoa_orders_group0);
    EXPECT_EQ(d->hoa_orders[1], expected_hoa_orders_group1);
    EXPECT_EQ(d->hoa_orders[2], expected_hoa_orders_group2);

    // HOASourceInformationStruct()
    std::vector<std::string> expected_hoaSourceIdStrs_group0 = {"src:HOA1","src:HOA2","src:HOA3"};
    std::vector<std::string> expected_hoaSourceIdStrs_group1 = {"src:HOA4","src:HOA5","src:HOA6","src:HOA7","src:HOA8"};
    std::vector<std::string> expected_hoaSourceIdStrs_group2 = {"src:HOA9","src:HOA10","src:HOA11"};
    EXPECT_EQ(d->hoaSourceIdStrs[0], expected_hoaSourceIdStrs_group0);
    EXPECT_EQ(d->hoaSourceIdStrs[1], expected_hoaSourceIdStrs_group1);
    EXPECT_EQ(d->hoaSourceIdStrs[2], expected_hoaSourceIdStrs_group2);

    // HOASourcePositionStruct()
    std::uint64_t expected_position_multiplier = std::pow(10.0f, 4.0f);
    std::uint64_t expected_rotation_multiplier = std::pow(2.0f, 16.0f) - 1;
    EXPECT_EQ(d->position_multiplier, expected_position_multiplier);
    EXPECT_EQ(d->rotation_multiplier, expected_rotation_multiplier);
    float expected_pos_rot_grp_values[3][6]
    {
            { 123.0f, -42.0f, 33.0f, 89.0f, -65.0f, 94.0f},
            { 124.0f, -43.0f, 30.0f, 55.0f,   3.0f, 94.0f},
            { 125.0f, -44.0f, 29.0f, -3.0f,   2.0f, 95.0f}
    };
    std::vector<std::vector<std::int32_t>> pos_xyz(3, std::vector<std::int32_t>{});
    std::vector<std::vector<std::int32_t>> rotations(3, std::vector<std::int32_t>{});
    std::int32_t pos_value = 0;
    std::int32_t rot_value = 0;
    std::int32_t nSources = 0;
    for (int group = 0; group < expected_num_hoa_groups; ++group) {
        nSources = expected_num_hoa_sources_in_groups[group];
        for (int i = 0; i < 3; ++i) {
            pos_xyz[i].clear();
            rotations[i].clear();
            pos_value = expected_position_multiplier * expected_pos_rot_grp_values[group][i];
            rot_value = expected_rotation_multiplier * expected_pos_rot_grp_values[group][i + 3];
            pos_xyz[i] = std::vector<std::int32_t>(nSources, pos_value);
            rotations[i] = std::vector<std::int32_t>(nSources, rot_value);
        }
        EXPECT_EQ(d->hoa_sources_pos_x[group], pos_xyz[0]);
        EXPECT_EQ(d->hoa_sources_pos_y[group], pos_xyz[1]);
        EXPECT_EQ(d->hoa_sources_pos_z[group], pos_xyz[2]);
        EXPECT_EQ(d->hoa_sources_rot_yaw[group], rotations[0]);
        EXPECT_EQ(d->hoa_sources_rot_pitch[group], rotations[1]);
        EXPECT_EQ(d->hoa_sources_rot_roll[group], rotations[2]);
    }

    // HOA6DOFRenderingMetadata()
    std::uint64_t expected_num_frames = 100;
    std::uint64_t expected_azimuth_multiplier = std::pow(2.0f, 16.0f);
    std::uint64_t expected_elevation_multiplier = std::pow(2.0f, 16.0f);
    std::uint64_t expected_dtr_multiplier = std::pow(2.0f, 16.0f) - 1;
    std::uint64_t expected_energy_multiplier = std::pow(2.0f, 25.0f);
    EXPECT_EQ(d->num_frames, expected_num_frames);
    EXPECT_EQ(d->azimuth_multiplier, expected_azimuth_multiplier);
    EXPECT_EQ(d->elevation_multiplier, expected_elevation_multiplier);
    EXPECT_EQ(d->dtr_multiplier, expected_dtr_multiplier);
    EXPECT_EQ(d->energy_multiplier, expected_energy_multiplier);

    // HOA6DOFRenderingMetadataFrame()
    std::uint32_t expected_subframes_per_frame = h->useRecursiveAveraging ? 1 : 4;
    EXPECT_EQ(d->subframes_per_frame, expected_subframes_per_frame);

    // HOA6DOFRenderingMetadataInformationSample()
    std::uint8_t expected_num_frequency_bands = 5;
    EXPECT_EQ(d->num_frequency_bands, expected_num_frequency_bands);

    // Initializations for the tests below
    std::vector<std::vector<float>> expected_hoa_render_meta_azimuth(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<float>> expected_hoa_render_meta_elevation(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<float>> expected_hoa_render_meta_dtr(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<float>> expected_hoa_render_meta_energy(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<int>> expected_hoa_render_meta_azi_int(h->nFrames, std::vector<int>(h->nBands, 0.0f));
    std::vector<std::vector<int>> expected_hoa_render_meta_elev_int(h->nFrames, std::vector<int>(h->nBands, 0.0f));
    std::vector<std::vector<int>> expected_hoa_render_meta_dtr_int(h->nFrames, std::vector<int>(h->nBands, 0.0f));
    std::vector<std::vector<int>> expected_hoa_render_meta_energy_int(h->nFrames, std::vector<int>(h->nBands, 0.0f));
    std::vector<std::vector<float>> azimuth_values(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<float>> elevation_values(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<float>> dtr_values(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<float>> energy_values(h->nFrames, std::vector<float>(h->nBands, 0.0f));
    std::vector<std::vector<int>> azi_values_int(h->nFrames, std::vector<int>(h->nBands, 0));
    std::vector<std::vector<int>> elev_values_int(h->nFrames, std::vector<int>(h->nBands, 0));
    std::vector<std::vector<int>> dtr_values_int(h->nFrames, std::vector<int>(h->nBands, 0));
    std::vector<std::vector<int>> energy_values_int(h->nFrames, std::vector<int>(h->nBands, 0));
    float value = 0.0f;
    float frame0value = 0.0f;
    float groupIncrement = 0.001f;
    float sourceIncrement = 0.001f;
    float frameIncrement = 0.001f;
    float increment = 0.0001f;
    float degMultiplier = 180.0f / M_PI;
    int NdecimalConversion = 4;

    auto get_expected_and_processed_metadata_values = [&](int mic, float &value, float &frame0value) {
        for (int frame = 0; frame < h->nFrames; ++frame) {

            // Store the 0th sub-frame value, and use that, when recursive averaging is used
            if (frame == 0) {
                frame0value = value;
            } else if (h->useRecursiveAveraging && frame != 0) {
                value = frame0value;
            }

            // Form the expected metadata values
            for (int band = 0; band < h->nBands; ++band) {
                expected_hoa_render_meta_azimuth[frame][band]   = value;
                expected_hoa_render_meta_elevation[frame][band] = value + increment;
                expected_hoa_render_meta_dtr[frame][band]       = value + 2.0f * increment;
                expected_hoa_render_meta_energy[frame][band]    = value + 3.0f * increment;
                value += 4.0f * increment;
            }

            // Fetch the processed metadata values, and convert radians to degrees
            std::copy_n(&h->azimuthData[mic][frame][0], h->nBands, azimuth_values[frame].data());
            std::copy_n(&h->elevationData[mic][frame][0], h->nBands, elevation_values[frame].data());
            std::copy_n(&h->DTRData[mic][frame][0], h->nBands, dtr_values[frame].data());
            std::copy_n(&h->energyData[mic][frame][0], h->nBands, energy_values[frame].data());
            utility_svsmul(azimuth_values[frame].data(), &degMultiplier, h->nBands, NULL);
            utility_svsmul(elevation_values[frame].data(), &degMultiplier, h->nBands, NULL);

            // Round to N decimals and transform to integer for comparison
            convertToNDecimalIntVec(azimuth_values[frame], azi_values_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(elevation_values[frame], elev_values_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(dtr_values[frame], dtr_values_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(energy_values[frame], energy_values_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(expected_hoa_render_meta_azimuth[frame], expected_hoa_render_meta_azi_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(expected_hoa_render_meta_elevation[frame], expected_hoa_render_meta_elev_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(expected_hoa_render_meta_dtr[frame], expected_hoa_render_meta_dtr_int[frame], NdecimalConversion);
            convertToNDecimalIntVec(expected_hoa_render_meta_energy[frame], expected_hoa_render_meta_energy_int[frame], NdecimalConversion);
        }
    };

    /// Simulate the effect of using a different translator with a different group
    for (h->HOAGroupId = 0; h->HOAGroupId < expected_num_hoa_groups; ++h->HOAGroupId) {
        h->nMicrophoneArrays = d->num_hoa_sources_in_group[h->HOAGroupId];
        read_spatial_metadata_from_binary(h);
        for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {

            // Set the initial value
            value = (h->HOAGroupId + 1) * groupIncrement + (mic + 1) * sourceIncrement;

            for (h->currentFrame = 0; h->currentFrame < d->num_frames; ++h->currentFrame) {

                // Add frameIncrement to differentiate values between processing frames
                value += frameIncrement;

                read_spatial_metadata_from_binary(h);
                get_expected_and_processed_metadata_values(mic, value, frame0value);

                // Compare
                EXPECT_EQ(azi_values_int, expected_hoa_render_meta_azi_int);
                EXPECT_EQ(elev_values_int, expected_hoa_render_meta_elev_int);
                EXPECT_EQ(dtr_values_int, expected_hoa_render_meta_dtr_int);
                EXPECT_EQ(energy_values_int, expected_hoa_render_meta_energy_int);
            }
        }
    }

    //// Simulate the effect of increasing currentFrame outside the translator
    h->HOAGroupId = 0;
    int mic = 0;
    h->nMicrophoneArrays = d->num_hoa_sources_in_group[h->HOAGroupId];
    int nSkippedFrames = 98;

    // Set the expected initial value based on the amount skipped frames
    if (h->useRecursiveAveraging) {
        value = (h->HOAGroupId + 1) * groupIncrement + (mic + 1) * sourceIncrement + (float)nSkippedFrames * ((float)h->nBands * 4.0f * increment + frameIncrement) + frameIncrement;
    } else {
        value = (h->HOAGroupId + 1) * groupIncrement + (mic + 1) * sourceIncrement + (float)nSkippedFrames * ((float)h->nFrames * (float)h->nBands * 4.0f * increment + frameIncrement) + frameIncrement;
    }

    // Read the bitstream from the beginning, and then from the skipped frame position
    h->currentFrame = 0;
    read_spatial_metadata_from_binary(h);
    h->currentFrame = nSkippedFrames;
    read_spatial_metadata_from_binary(h);

    get_expected_and_processed_metadata_values(mic, value, frame0value);

    // Compare
    EXPECT_EQ(azi_values_int, expected_hoa_render_meta_azi_int);
    EXPECT_EQ(elev_values_int, expected_hoa_render_meta_elev_int);
    EXPECT_EQ(dtr_values_int, expected_hoa_render_meta_dtr_int);
    EXPECT_EQ(energy_values_int, expected_hoa_render_meta_energy_int);
}

INSTANTIATE_TEST_SUITE_P(Default, DecoderTest,
                         testing::Values(
                         // Test case: block based processing
                         DecoderTestData {false},
                         // Test case: recursive averaging mode
                         DecoderTestData {true}
                         ));


struct OptimalMixingTestData{};

class OptimalMixingTest
        : public ::testing::TestWithParam<OptimalMixingTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(OptimalMixingTest,
       MoreEfficientOptimalMixing) // NOLINT
{
    // Test that the new optimal mixing provides near similar output as the SAF counterpart.

    auto params = OptimalMixingTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;

    std::vector<float_complex> Cr(h->outChannels * h->outChannels, float_complex(0.0f, 0.0f));
    std::vector<float_complex> Q(h->outChannels * h->outChannels, float_complex(0.0f, 0.0f));
    Q[0] = float_complex(1.0f, 0.0f);
    Q[1] = float_complex(0.0f, 0.0f);
    Q[2] = float_complex(0.0f, 0.0f);
    Q[3] = float_complex(1.0f, 0.0f);
    hoatMatrixCmplx2D safM;
    hoatMatrixCmplx2D safMr;
    hoatMatrixCmplx2D newM;
    hoatMatrixCmplx2D newMr;
    createHoatMatrix(&safM, 2, 2);
    createHoatMatrix(&safMr, 2, 2);
    createHoatMatrix(&newM, 2, 2);
    createHoatMatrix(&newMr, 2, 2);

    float errorPercent = 5.0f;
    bool useEnergyFLAG = false;

    for (int band = 0; band < h->nBands; ++band) {

        // Initialize input values
        float_complex channelX1 = {band * 0.5f + 0.1f, band * 0.1f + 0.05f};
        float_complex channelX2 = {(h->nBands - band) * 0.5f + 0.1f, (h->nBands - band) * 0.1f + 0.05f};
        h->Cx[band][0][0] = channelX1.real() * channelX1.real() + channelX1.imag() * channelX1.imag();
        h->Cx[band][1][1] = channelX2.real() * channelX2.real() + channelX2.imag() * channelX2.imag();
        h->Cx[band][1][0] = channelX1 * std::conj(channelX2);
        h->Cx[band][0][1] = channelX2 * std::conj(channelX1);

        float_complex channelY1 = channelX1 + float_complex{-band * 0.5f, band * 0.5f};
        float_complex channelY2 = channelX1 + float_complex{band * 0.5f, -band * 0.5f};
        h->Cy[band][0][0] = channelY1.real() * channelY1.real() + channelY1.imag() * channelY1.imag();
        h->Cy[band][1][1] = channelY2.real() * channelY2.real() + channelY2.imag() * channelY2.imag();
        h->Cy[band][1][0] = channelY1 * std::conj(channelY2);
        h->Cy[band][0][1] = channelY2 * std::conj(channelY1);

        h->CyLinearDecoding[band][0][0] = h->Cx[band][0][0];
        h->CyLinearDecoding[band][1][1] = h->Cx[band][1][1];

        // SAF processing
        formulate_M_and_Cr_cmplx(h->covarianceDomainFramework,
                                 FLATTEN2D(h->Cx[band].data),
                                 FLATTEN2D(h->Cy[band].data),
                                 Q.data(),
                                 useEnergyFLAG,
                                 0.2f,
                                 FLATTEN2D(safM.data),
                                 Cr.data());

        formulate_M_and_Cr_cmplx(h->covarianceDomainFramework,
                                 FLATTEN2D(h->CyLinearDecoding[band].data),
                                 Cr.data(),
                                 Q.data(),
                                 0,
                                 0.2f,
                                 FLATTEN2D(safMr.data),
                                 NULL);

        // New more efficient processing
        formulate_M_and_Cr_2x2(h,
                               h->Cx[band],
                               h->Cy[band],
                               h->Q,
                               useEnergyFLAG,
                               0.2f,
                               newM,
                               h->Cr,
                               !useEnergyFLAG);

        formulate_M_and_Cr_2x2(h,
                               h->CyLinearDecoding[band],
                               h->Cr,
                               h->Q,
                               false,
                               0.2f,
                               newMr,
                               h->Cr,
                               false);

        // Compare outputs
        float minError = 1e-2f;
        for (int ch1 = 0; ch1 < 2; ++ch1) {
            for (int ch2 = 0; ch2 < 2; ++ch2) {
                float biggerMRealAbs = std::max(std::abs(safM[ch1][ch2].real()), std::abs(newM[ch1][ch2].real()));
                float biggerMImagAbs = std::max(std::abs(safM[ch1][ch2].imag()), std::abs(newM[ch1][ch2].imag()));
                float biggerMrRealAbs = std::max(std::abs(safMr[ch1][ch2].real()), std::abs(newMr[ch1][ch2].real()));
                float biggerMrImagAbs = std::max(std::abs(safMr[ch1][ch2].imag()), std::abs(newMr[ch1][ch2].imag()));
                float MRealError = (errorPercent / 100.0f) * biggerMRealAbs;
                float MImagError = (errorPercent / 100.0f) * biggerMImagAbs;
                float MrRealError = (errorPercent / 100.0f) * biggerMrRealAbs;
                float MrImagError = (errorPercent / 100.0f) * biggerMrImagAbs;
                if (MRealError >= minError) {
                    EXPECT_NEAR(safM[ch1][ch2].real(), newM[ch1][ch2].real(), MRealError);
                }
                if (MImagError >= minError) {
                    EXPECT_NEAR(safM[ch1][ch2].imag(), newM[ch1][ch2].imag(), MImagError);
                }
                if (MrRealError >= minError) {
                    EXPECT_NEAR(safMr[ch1][ch2].real(), newMr[ch1][ch2].real(), MrRealError);
                }
                if (MrImagError >= minError) {
                    EXPECT_NEAR(safMr[ch1][ch2].imag(), newMr[ch1][ch2].imag(), MrImagError);

                }
            }
        }
    }
    destroyHoatMatrix(&safM);
    destroyHoatMatrix(&safMr);
    destroyHoatMatrix(&newM);
    destroyHoatMatrix(&newMr);
}

INSTANTIATE_TEST_SUITE_P(Default, OptimalMixingTest,
                         testing::Values(
                                 OptimalMixingTestData{}
                         ));


struct LowComplexityTestData{};

class LowComplexityTest
        : public ::testing::TestWithParam<LowComplexityTestData>
{
public:
    HoaTranslatorTest hoaTranslator;
};

TEST_P(LowComplexityTest,
       LowComplexityInit) // NOLINT
{
    // Test that initializing the complexity reduction works as intended.

    auto params = LowComplexityTest::GetParam();
    auto h = &*hoaTranslator.m_hoaTranslatorData;
    int highestSingleBinBandsIndex = 15;
    int intermediateBandsERBWidth = 2;

    // Test when the high band index is exactly at 3rd octave band tick
    int lowestHighBandIndex = 89;
    std::vector<int> complexityReductionParameters{highestSingleBinBandsIndex, lowestHighBandIndex, intermediateBandsERBWidth};
    initialize_complexity_reduction(h, complexityReductionParameters);
    EXPECT_EQ(h->lowAvgLimitIndex, 71);
    EXPECT_EQ(h->maxFreqLimitIndex, 90);

    // Test when the high band index is slightly below 3rd octave band tick
    lowestHighBandIndex = 85;
    complexityReductionParameters = {highestSingleBinBandsIndex, lowestHighBandIndex, intermediateBandsERBWidth};
    initialize_complexity_reduction(h, complexityReductionParameters);
    EXPECT_EQ(h->lowAvgLimitIndex, 71);

    // Test when the high band index is slightly above 3rd octave band tick
    lowestHighBandIndex = 95;
    complexityReductionParameters = {highestSingleBinBandsIndex, lowestHighBandIndex, intermediateBandsERBWidth};
    initialize_complexity_reduction(h, complexityReductionParameters);
    EXPECT_EQ(h->lowAvgLimitIndex, 71);

    // Test when the high band index is low
    highestSingleBinBandsIndex = 0;
    lowestHighBandIndex = 3;
    complexityReductionParameters = {highestSingleBinBandsIndex, lowestHighBandIndex, intermediateBandsERBWidth};
    initialize_complexity_reduction(h, complexityReductionParameters);
    EXPECT_EQ(h->lowAvgLimitIndex, h->maxFreqLimitIndex - 1);

    // Test when the high band index is zero
    highestSingleBinBandsIndex = 0;
    lowestHighBandIndex = 0;
    complexityReductionParameters = {highestSingleBinBandsIndex, lowestHighBandIndex, intermediateBandsERBWidth};
    initialize_complexity_reduction(h, complexityReductionParameters);
    EXPECT_EQ(h->lowAvgLimitIndex, 0);
}

INSTANTIATE_TEST_SUITE_P(Default, LowComplexityTest,
                         testing::Values(
                                 LowComplexityTestData{}
                         ));

struct SourceAwareTestData{

    friend std::ostream& operator<<(std::ostream& os, const SourceAwareTestData& data)
    {
        return os;
        //<< "currentListenerPosition: " << vecToString(data.currentListenerPosition)
        //<< ", previousListenerPosition: " << vecToString(data.previousListenerPosition)
        //<< ", projectionRadius: " << data.projectionRadius;
    }
};

class SourceAwareTest
        : public ::testing::TestWithParam<SourceAwareTestData>
{
public:
    void SetUp(const HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT &micArrayArrangement) {
        hoaTranslator = new HoaTranslatorTest(micArrayArrangement);
    }
    HoaTranslatorTest* hoaTranslator;
};

TEST_P(SourceAwareTest,
       SourceAware) // NOLINT
{

    /* 5 FOA microphones (o) in a diamond shape arrangement.
     * 4 sources (x) next to the edge mic arrays.
     *
     *                          x s3
     *
     *                          o m4
     *                       /  |  \
     *                    /     |     \
     *                 /        |        \
     *       s1 x  m2 o- - - - -o m0 - - - -o m1  - x s0
     *                 \        |        /        |
     *                    \     |     /           1m
     *                       \  |  /              |
     *                          o m3              –
     *                          |----1m---|---1m----|
     *                          x s2
     */


    SetUp(HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT::DIAMOND_SOURCE_AWARE);
    auto h = &*hoaTranslator->m_hoaTranslatorData;

    auto params = SourceAwareTest::GetParam();

    EXPECT_EQ(h->nSources, 4);

    // Test geometrical values
    std::vector<int> expectedClosestMicsForSources{1, 2, 3, 4};
    std::vector<float> expectedClosestMicToSrcAziDeg{-90.0f, 90.0f, 180.0f, 0.0f};
    float expectedClosestMicToSrcElevDeg = 0.0f;
    float expectedClosestMicToSrcDistance = 1.0f;
    float errorMargin = 0.1f;
    for (int src = 0; src < h->nSources; ++src) {
        int mic = h->closestMicsForSources[src];
        EXPECT_EQ(h->closestMicsForSources[src], expectedClosestMicsForSources[src]);
        if (src == 2) {
            // Source (s2) azimuth direction can be either 180 or -180 degrees
            EXPECT_NEAR(std::abs(h->micToSrcAziElevDeg[src][mic][0]), expectedClosestMicToSrcAziDeg[src], errorMargin);
        } else {
            EXPECT_NEAR(h->micToSrcAziElevDeg[src][mic][0], expectedClosestMicToSrcAziDeg[src], errorMargin);
        }
        EXPECT_NEAR(h->micToSrcAziElevDeg[src][mic][1], expectedClosestMicToSrcElevDeg, errorMargin);
        EXPECT_NEAR(h->micToSrcDistances[src][mic], expectedClosestMicToSrcDistance, errorMargin);
    }

    // Initialize input signals (sound coming from the closest sources)
    int nFirstOrderChannels = 4;
    for (int mic = 0; mic < h->nMicrophoneArrays; ++mic) {
        for (int frame = 0; frame < h->nFrames; ++frame) {
            for (int ch = 0; ch < nFirstOrderChannels; ++ch) {
                for (int band = 0; band < h->nBands; ++band) {
                    if (mic == 0) {
                        // Omni only
                        if (ch == 0) {
                            h->TFDInputFrameFullOrder[mic][frame][ch][band] = 1.0f;
                        }
                    } else if (mic == 1) {
                        h->TFDInputFrameFullOrder[mic][frame][ch][band] = ACNright[ch];
                    } else if (mic == 2) {
                        h->TFDInputFrameFullOrder[mic][frame][ch][band] = ACNleft[ch];
                    } else if (mic == 3) {
                        h->TFDInputFrameFullOrder[mic][frame][ch][band] = ACNback[ch];
                    } else if (mic == 4) {
                        h->TFDInputFrameFullOrder[mic][frame][ch][band] = ACNfront[ch];
                    }
                }
            }
        }
    }
    // Activate sources
    check_for_source_aware_mode(h);
    h->activeSources.clear();
    for (int src = 0; src < h->nSources; ++src) {
        h->activeSources.push_back(src);
    }

    source_aware_processing(h);

    // Test source spatial values (for mic arrays m1-m4)
    for (int mic = 1; mic < h->nMicrophoneArrays; ++mic) {
        for (int band = 0; band < h->nBands; ++band) {
            if (mic == 3) {
                // Mic array (m3) azimuth direction can be either 180 or -180 degrees
                EXPECT_NEAR(std::abs(h->srcAzimuthData[mic][band]) * 180.0f / M_PI, expectedClosestMicToSrcAziDeg[mic - 1], errorMargin);
            } else {
                EXPECT_NEAR(h->srcAzimuthData[mic][band] * 180.0f / M_PI, expectedClosestMicToSrcAziDeg[mic - 1], errorMargin);
            }
            EXPECT_NEAR(h->srcElevationData[mic][band] * 180.0f / M_PI, expectedClosestMicToSrcElevDeg, errorMargin);
        }
    }

    // Test vectors from the listener (0,0,0) to sources
    auto expectedListenerToSrcVectors = h->sourcePositionsOpenGL;
    float expectedListenerToSrcDistance = calculate_vector_norm(expectedListenerToSrcVectors[0]);
    for (int src = 0; src < h->nSources; ++src) {
        for (int i = 0; i < 3; ++i) {
            EXPECT_NEAR(h->listenerToSrcVectorsOpenGL[src][i][0], expectedListenerToSrcVectors[src][i], errorMargin);
        }
        EXPECT_NEAR(h->listenerToSrcDistances[src], expectedListenerToSrcDistance, errorMargin);
    }
}

INSTANTIATE_TEST_SUITE_P(Default, SourceAwareTest,
                         testing::Values(
                                 SourceAwareTestData{}
                         ));

struct TwoSourceTestData{};

class TwoSourceTest
        : public ::testing::TestWithParam<TwoSourceTestData>
{
public:
    void SetUp(const HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT &micArrayArrangement) {
        hoaTranslator = new HoaTranslatorTest(micArrayArrangement);
    }
    HoaTranslatorTest* hoaTranslator;
};

TEST_P(TwoSourceTest,
       TwoSources) // NOLINT
{
    /* Test for two HOA sources. Test that the listener is projected to expected position. Also test the mic and edge
     * normal directions and the interpolated listener normal direction.
     *
     * Test scene:
     *      - microphone arrays m0 (0,0) and m1 (1,0)
     *      - listener position l_0 .. l_3
     *      - mic normals n_m0 and n_m1
     *      - edge normals n_e0 and n_e1
     *      - projected listener position normals on edge n_l2 and n_l3
     *        (normals for l_0 and l_1 are obtained directly from n_m1 and n_m0)
     *
     *                                   x l_2
     *
     *                           ^ n_e0      n_l2
     *                           |          /
     *                m0         |        /    m1
     * l_1 x  n_m0 <-- o- - x - - - - - -x- - -o --> n_m1       x l_0
     *                     /      |
     *                    /       |
     *                  n_l3      \/ n_e1
     *
     *                      x l_3
     *
     *                 |----------1m-----------|
     */

    SetUp(HoaTranslatorTest::MIC_ARRAY_ARRANGEMENT::TWO_SOURCES);
    auto h = &*hoaTranslator->m_hoaTranslatorData;

    std::vector<std::vector<float>> listenerPositions{{ 2.0f,  0.0f,   0.0f },
                                                      {-1.0f,  0.0f,   0.0f },
                                                      { 0.75f, 0.0f,   1.0f },
                                                      { 0.25f, 0.0f,  -1.0f }};

    std::vector<std::vector<float>> expectedProjectedListenerPositions{{ h->microphoneArrayPositions[1][0], h->microphoneArrayPositions[1][1], h->microphoneArrayPositions[1][2] },
                                                                       { h->microphoneArrayPositions[0][0], h->microphoneArrayPositions[0][1], h->microphoneArrayPositions[0][2] },
                                                                       { 0.75f, 0.0f, 0.0f },
                                                                       { 0.25f, 0.0f, 0.0f }};

    std::vector<std::vector<bool>> expectedArrayEdgeProjection{{ true,  false },
                                                               { true,  false },
                                                               { false, true  },
                                                               { false, true  }};

    std::vector<std::vector<float>> expectedEdgeNormals{{ 0.0f,  0.0f,   1.0f },
                                                        { 0.0f,  0.0f,  -1.0f }};

    std::vector<std::vector<float>> expectedMicNormals{{ -1.0f,  0.0f,   0.0f },
                                                       {  1.0f,  0.0f,   0.0f }};

    // Normals at the projected listener positions
    auto listPos2EdgeVec = multiplyVector(0.5f, expectedEdgeNormals[0]);
    auto listPos2MicVec = multiplyVector(0.5f, expectedMicNormals[1]);
    auto listPos3EdgeVec = multiplyVector(0.5f, expectedEdgeNormals[1]);
    auto listPos3MicVec = multiplyVector(0.5f, expectedMicNormals[0]);
    auto listPos2Vec = sumVectors(listPos2EdgeVec, listPos2MicVec);
    auto listPos3Vec = sumVectors(listPos3EdgeVec, listPos3MicVec);
    normalizeVector(listPos2Vec);
    normalizeVector(listPos3Vec);
    std::vector<std::vector<float>> expectedListenerNormals{ expectedMicNormals[1],
                                                             expectedMicNormals[0],
                                                             listPos2Vec,
                                                             listPos3Vec };

    // Check edge and mic normals
    std::vector<int> edgeNormalInt(3, 0);
    std::vector<int> micNormalInt(3, 0);
    std::vector<int> expectedEdgeNormalInt(3, 0);
    std::vector<int> expectedMicNormalInt(3, 0);
    for (int i = 0; i < expectedEdgeNormals.size(); ++i) {
        convertTo2DecimalIntVec(h->edgeNormals[i], edgeNormalInt);
        convertTo2DecimalIntVec(h->micNormals[i], micNormalInt);
        convertTo2DecimalIntVec(expectedEdgeNormals[i], expectedEdgeNormalInt);
        convertTo2DecimalIntVec(expectedMicNormals[i], expectedMicNormalInt);
        EXPECT_EQ(edgeNormalInt, expectedEdgeNormalInt);
        EXPECT_EQ(micNormalInt, expectedMicNormalInt);
    }

    // Check projected listener position and listener normal
    std::vector<int> projectedListenerPositionInt(3, 0);
    std::vector<int> expectedProjectedListenerPositionInt(3, 0);
    std::vector<int> listenerNormalInt(3, 0);
    std::vector<int> expectedListenerNormalInt(3, 0);

    for (int i = 0; i < listenerPositions.size(); ++i) {
        h->currentListenerPosition = listenerPositions[i];
        pre_process_positions(h);

        convertTo2DecimalIntVec(h->projectedListenerPosition, projectedListenerPositionInt);
        convertTo2DecimalIntVec(expectedProjectedListenerPositions[i], expectedProjectedListenerPositionInt);
        convertTo2DecimalIntVec(h->listenerExteriorNormal, listenerNormalInt);
        convertTo2DecimalIntVec(expectedListenerNormals[i], expectedListenerNormalInt);

        EXPECT_EQ(projectedListenerPositionInt, expectedProjectedListenerPositionInt);
        EXPECT_EQ(h->isListenerProjectedToClosestArray, expectedArrayEdgeProjection[i][0]);
        EXPECT_EQ(h->isListenerProjectedToClosestEdge, expectedArrayEdgeProjection[i][1]);
        EXPECT_EQ(listenerNormalInt, expectedListenerNormalInt);
    }
}

INSTANTIATE_TEST_SUITE_P(Default, TwoSourceTest,
                         testing::Values(TwoSourceTestData{}));


} // namespace hoatranslatortest
