/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_AMBISTORE_HPP
#define MPEGCORE_AMBISTORE_HPP

#include <cstddef>
#include <functional>
#include <optional>
#include <vector>

namespace mpegcore {

namespace {
constexpr std::size_t CLEAR_BUFFER_ROUNDS{ 10 };
constexpr std::size_t CLEAR_BUFFER_SIZE{ 512 };

template <typename T, typename U>
std::optional<std::size_t> findFreeHandle(T const& params, std::vector<U> const& ambis)
{
    for (std::size_t i{ 0 }; i < ambis.size(); i++) {
        auto& ambi = ambis.at(i);
        if (!ambi.isInUse && params == ambi.params) {
            return i;
        }
    }
    return {};
}
} // namespace

// Initializing SAF decoders is very slow. By storing them in vectors for later
// reuse cuts down test execution time almost by an order of magnitude.
template <typename Ambi, typename AmbiParams> class AmbiStore
{
public:
    ~AmbiStore()
    {
        for (auto&& ambi : m_ambis) {
            if (ambi.ambi != nullptr) {
                // We need a custom destroy method as vector uses destructor for
                // existing instances when adding new.
                ambi.destroy();
                ambi.ambi = nullptr;
            }
        }
    }

    std::optional<std::size_t> reserveAmbi(AmbiParams const& params)
    {
        auto handle = findFreeHandle(params, m_ambis);
        if (handle.has_value()) {
            m_ambis[handle.value()].isInUse = true;
            return handle.value();
        }
        auto& ambi = m_ambis.emplace_back(Ambi{ params, nullptr, true });
        auto const initOk = ambi.init();

        if (!initOk) {
            return {};
        }

        ambi.isInUse = true;
        return m_ambis.size() - 1;
    }

    Ambi& getAmbi(std::size_t index) { return m_ambis.at(index); }

    void freeAmbi(std::size_t index)
    {
        auto& ambi = m_ambis[index];
        ambi.deinit(CLEAR_BUFFER_ROUNDS, CLEAR_BUFFER_SIZE);
        ambi.isInUse = false;
    }

private:
    std::vector<Ambi> m_ambis;
};

} // namespace mpegcore

#endif // MPEGCORE_AMBISTORE_HPP
