/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <iostream>

#include "DirectionalDecay.hpp"

namespace mpegcore {

void SHDecay::init(std::size_t const numDirections)
{
    SPHDESIGN_PRESETS sphDesign;

    switch (numDirections) {
    case 15:
        sphDesign = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_15;
        break;
    case 16:
        sphDesign = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_16;
        break;
    case 31:
        sphDesign = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_31;
        break;
    case 32:
        sphDesign = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_32;
        break;
    case 63:
        sphDesign = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_63;
        break;
    case 64:
        sphDesign = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_64;
        break;
    default:
        EAR_FAIL("SHDecay::init: numDirections must be 15, 31, or 63.")
    }

    rirlib_create(&pRirLib, sphDesign);

    rirlib_data* rirlib = (rirlib_data*)pRirLib;

    // Set source and receiver positions: source defaults to 1m in front of the mic
    EAR::Vector3 sPos, rPos;
    if (sourcePosition.has_value()) {
        sPos = sourcePosition.value().location;
    } else {
        sPos = EAR::Vector3(0.f, 1.5f, -1.f); // openGL coords
    };
    if (micPosition.has_value()) {
        rPos = micPosition.value().location;
    } else {
        rPos = EAR::Vector3(0.f, 1.5f, 0.f); // openGL coords
    }
    // Note arg order: OpenGL -> ambisonic coords
    rirlib_setSrcPosition(rirlib, sPos[2] * -1.f, sPos[0] * -1.f, sPos[1]);
    rirlib_setRecPosition(rirlib, rPos[2] * -1.f, rPos[0] * -1.f, rPos[1]);

    rirlib->srcDirectivityFlag = srcDirectivityFlag;
}

void SHDecay::setRIR_file()
{
    if (srirPath.empty()) {
        EAR_FAIL("[SHDecay] setRIR_file called without setting the srirPath.\n")
    }
    rirlib_data* rirlib = (rirlib_data*)pRirLib;

    srir.load(srirPath);

    NORMALIZATION_TYPES normType;
    switch (normalization) {
    case mpegcore::HoaNormalization::n3d:
        normType = N3D_NORM;
        break;
    case mpegcore::HoaNormalization::sn3d:
        normType = SN3D_NORM;
        break;
    default:
        EAR_FAIL("Invalid normalization provided, can't convert to rirlib NORMALIZATION_TYPES.")
        break;
    }

    rirlib_setRIR(rirlib, (const float**)srir.getFrame()->dataRaw(),
        static_cast<int>(srir.getFrame()->channels()), static_cast<int>(srir.getFrame()->length()),
        static_cast<int>(srir.getSampleRate()), static_cast<int>(normType));

    // Wait for buffers to allocate and load
    while (rirlib->analysisStage < ANALYSIS_BUFS_LOADED) {
        SAF_SLEEP(10);
    };
}

void SHDecay::setRIR_buffer(const float** bufPtr, std::size_t const nChan, std::size_t const nSamp,
    std::uint32_t const sampleRate)
{
    rirlib_data* rirlib = (rirlib_data*)pRirLib;

    NORMALIZATION_TYPES normType;

    switch (normalization) {
    case mpegcore::HoaNormalization::n3d:
        normType = N3D_NORM;
        break;
    case mpegcore::HoaNormalization::sn3d:
        normType = SN3D_NORM;
        break;
    default:
        EAR_FAIL("Invalid normalization provided, can't convert to rirlib NORMALIZATION_TYPES.");
        break;
    }

    rirlib_setRIR(rirlib, bufPtr, static_cast<int>(nChan), static_cast<int>(nSamp), sampleRate,
        static_cast<int>(normType));

    // Wait for buffers to allocate and load
    while (rirlib->analysisStage < ANALYSIS_BUFS_LOADED) {
        SAF_SLEEP(10);
    };
}

bool SHDecay::isInitialized()
{
    if (pRirLib != nullptr) {
        rirlib_data* pHOSirr = (rirlib_data*)pRirLib;

        return pHOSirr->analysisStage >= ANALYSIS_BUFS_LOADED;
    } else {
        return false;
    }
}

void SHDecay::analyze(ANALYSIS_STAGE endStage = DIRGAIN_DONE)
{
    if (isInitialized()) {
        rirlib_data* pRirData = (rirlib_data*)pRirLib;
        rirlib_processRIR(pRirData, endStage);

        // Wait for analysis to complete
        while (pRirData->analysisStage < endStage) {
            SAF_SLEEP(10);
        };
    } else {
        EAR_FAIL("SHDecay::analyze called before analysis object has been initialized.")
    }
}

SHDecay::~SHDecay()
{
    rirlib_destroy(&pRirLib);
    pRirLib = nullptr;
}

} // namespace mpegcore
