/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "EsdConverter.hpp"

#include <algorithm>
#include <cassert>

#include <saf.h>
#include <saf_externals.h>

namespace mpegcore {

namespace {

// Virtual loudspeaker directions for ESD/HOA conversion as [inclination1 azimuth1; inclination2
// azimuth2;...] in radians Reference (p.17->):
// https://www.etsi.org/deliver/etsi_TS/126200_126299/126260/15.00.00_60/ts_126260v150000p.pdf
const std::vector<float> lsDirs1stOrderRad{ 0, 0, 1.910633, 0, 1.910633, 2.094395, 1.910633,
    -2.094 };
const std::vector<float> lsDirs2ndOrderRad{ 0, 0, 2.361073, 0, 1.207589, -1.95668, 1.207589,
    1.956682, 2.415178, -1.95668, 1.561039, -3.14159, 2.415178, 1.956681, 1.325668, 0.687124,
    1.325667, -0.68712 };
const std::vector<float> lsDirs3rdOrderRad{ 0, 0, 0.854098, 0, 2.031969, 1.119907, 2.605106,
    -0.25283, 1.078622, 1.155586, 1.736608, 2.040481, 2.031968, -1.38118, 1.736609, 0.270692,
    1.56888, -2.20417, 0.917087, 2.297267, 0.917087, -2.80293, 1.763476, 3.010956, 2.649852,
    2.154919, 1.568881, -0.63529, 0.953962, -1.41973, 2.458122, -2.46809 };
const std::vector<float> lsDirs4thOrderRad{ 0, 0, 0.823218, 0, 1.73912, -2.00759, 0.724297,
    1.927637, 1.336281, -1.41208, 0.871631, -2.10001, 1.263705, 2.512927, 1.440147, 1.667633,
    2.248313, 1.442383, 1.433953, -0.60062, 2.888065, 0.329968, 2.003914, -1.18621, 1.80266,
    2.983332, 1.396554, -2.69222, 2.170781, 0.507602, 1.952805, 2.208977, 1.58019, 0.952319,
    2.584609, -1.71565, 0.874597, 0.934402, 2.172935, -0.38654, 2.612717, 2.675958, 2.193907,
    -2.62842, 1.51674, 0.165012, 0.715307, -1.02504, 0.762553, -3.13121 };

/* Convert input signals from Equivalent spatial domain (ESD) representation to Higher-Order
 * Ambisonics (HOA). HOA format is obtained from ESD by multiplying the ESD signals with a
 * conversion matrix M_ESDtoHOA. The conversion matrix contains real spherical harmonics for
 * pre-determined directions:
 *
 *      HOA = M_ESDtoHOA * ESD,
 *
 *      where matrix sizes are:
 *          - HOA: nChannels x frameSize
 *          - M_ESDtoHOA: nChannels x nChannels
 *          - ESD: nChannels x frameSize
 *
 * For more details about the conversion, see pages 7 and 18 in:
 * https://www.etsi.org/deliver/etsi_TS/126200_126299/126260/15.00.00_60/ts_126260v150000p.pdf
 *
 * See also section F.1.3 onwards in:
 * ISO_IEC_FDIS_23008-3_2018(E)_(MPEG-H_3DA_2nd_ed)_20180319
 *
 * Input:
 *      nChannels       - number of channels in one HOA input signal (= (order+1)^2)
 *      framesize       - number of samples in one processing frame
 *      M_ESDtoHOA      - ESD to HOA conversion matrix (size: nChannels * nChannels)
 *      input           - ESD input signal (size: nChannels * framesize)
 *      output          - pre-allocated handle for the HOA output signal (size: nChannels *
 *                        framesize)
 *
 * Output:
 *      output          - input signal in HOA format
 */
void convert_esd_to_hoa(
    int nChannels, int framesize, float* M_ESDtoHOA, float* input, float* output)
{
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, nChannels, framesize, nChannels, 1.0f,
        M_ESDtoHOA, nChannels, input, framesize, 0.0f, output, framesize);
}

/* Convert HOA signals to ESD as:
 *
 *      ESD = M_HOAtoESD * HOA
 *          = inv(M_ESDtoHOA) * HOA
 *
 * Input:
 *      nChannels       - number of channels in one HOA input signal (= (order+1)^2)
 *      framesize       - number of samples in one processing frame
 *      M_HOAtoESD      - HOA to ESD conversion matrix (size: nChannels * nChannels)
 *      input           - HOA input signal (size: nChannels * framesize)
 *      output          - pre-allocated handle for the ESD output signal (size: nChannels *
 *                        framesize)
 *
 * Output:
 *      output          - input signal in ESD format
 */
void convert_hoa_to_esd(
    int nChannels, int framesize, float* M_HOAtoESD, float* input, float* output)
{
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, nChannels, framesize, nChannels, 1.0f,
        M_HOAtoESD, nChannels, input, framesize, 0.0f, output, framesize);
}

} // namespace

EsdConverter::EsdConverter(std::uint32_t samplesPerBuffer)
{
    initialize(samplesPerBuffer);
}

void EsdConverter::initialize(std::uint32_t samplesPerBuffer)
{
    m_samplesPerConversion = samplesPerBuffer;
    m_EsdtoHoaMatrices = std::vector<float**>(m_maxConversionOrder);
    m_HoatoEsdMatrices = std::vector<float**>(m_maxConversionOrder);
    for (std::size_t orderIdx{ 0 }; orderIdx < m_maxConversionOrder; ++orderIdx) {
        std::size_t numChannels = ORDER2NSH(orderIdx + 1);
        m_channelsPerConversion.push_back(numChannels);
        m_EsdtoHoaMatrices[orderIdx] = (float**)calloc2d(numChannels, numChannels, sizeof(float));
        if (!get_esd_conversion_matrix((int)numChannels, FLATTEN2D(m_EsdtoHoaMatrices[orderIdx]))) {
            return;
        }
        // Matrix inversion
        m_HoatoEsdMatrices[orderIdx] = (float**)calloc2d(numChannels, numChannels, sizeof(float));
        utility_sinv(FLATTEN2D(m_EsdtoHoaMatrices[orderIdx]),
            FLATTEN2D(m_HoatoEsdMatrices[orderIdx]), numChannels);
    }
    m_input =
        (float**)calloc2d(ORDER2NSH(m_maxConversionOrder), m_samplesPerConversion, sizeof(float));
    m_output =
        (float**)calloc2d(ORDER2NSH(m_maxConversionOrder), m_samplesPerConversion, sizeof(float));

    m_isInitialized = true;
}

EsdConverter::~EsdConverter()
{
    for (auto m : m_EsdtoHoaMatrices) {
        free(m);
    }
    for (auto m : m_HoatoEsdMatrices) {
        free(m);
    }
    free(m_input);
    free(m_output);
}

bool EsdConverter::convertEsdToHoa(std::size_t order)
{
    if (order > m_maxConversionOrder) {
        return false;
    }

    convert_esd_to_hoa((int)m_channelsPerConversion[order - 1], m_samplesPerConversion,
        FLATTEN2D(m_EsdtoHoaMatrices[order - 1]), FLATTEN2D(m_input), FLATTEN2D(m_output));
    return true;
}

bool EsdConverter::convertHoaToEsd(std::size_t order)
{
    if (order > m_maxConversionOrder) {
        return false;
    }

    convert_hoa_to_esd((int)m_channelsPerConversion[order - 1], m_samplesPerConversion,
        FLATTEN2D(m_HoatoEsdMatrices[order - 1]), FLATTEN2D(m_input), FLATTEN2D(m_output));
    return true;
}

float* EsdConverter::inputAudioData(std::size_t channelIndex) noexcept
{
    return m_input[channelIndex];
}

void EsdConverter::copyToInput(AudioBuffer const& source, std::size_t channelIndex)
{
    copy(source, m_input[channelIndex]);
}

void EsdConverter::copyToInput(float const* data, std::size_t sampleCount, std::size_t channelIndex)
{
    std::copy_n(data, sampleCount, m_input[channelIndex]);
}

float const* EsdConverter::getOutputData(std::size_t channelIndex) const noexcept
{
    return m_output[channelIndex];
}

bool get_esd_conversion_matrix(const int nChannels, float* M_ESDtoHOA)
{
    // Obtain the virtual loudspeaker directions for the corresponding order
    int order = NSH2ORDER(nChannels);
    std::vector<float> lsDirsRad{};
    if (order < 1) {
        return false;
    } else if (order == 1) {
        lsDirsRad = lsDirs1stOrderRad;
    } else if (order == 2) {
        lsDirsRad = lsDirs2ndOrderRad;
    } else if (order == 3) {
        lsDirsRad = lsDirs3rdOrderRad;
    } else if (order == 4) {
        lsDirsRad = lsDirs4thOrderRad;
    } else {
        assert(false && "ESD to HOA conversion currently supports only up to 4th order HOA.");
        return false;
    }

    // Convert the virtual loudspeaker directions to [azimuth1 elevation1; ...] in degrees
    std::vector<float> lsDirsDeg{};
    for (auto dir = lsDirsRad.begin(); dir != lsDirsRad.end(); dir += 2) {
        float incl = *dir;
        float azi = *(dir + 1);
        lsDirsDeg.push_back(azi * 180.0f / M_PI);
        lsDirsDeg.push_back((M_PI / 2.0f - incl) * 180.0f / M_PI);
    }
    int nVirtualESDSpeakerDirs = lsDirsDeg.size() / 2;
    if (nVirtualESDSpeakerDirs != nChannels) {
        assert(false
            && "The number of virtual loudspeaker directions does not match with the number of "
               "channels.");
        return false;
    }

    // Real spherical harmonics for the speaker directions (= ESD to HOA conversion matrix)
    getRSH(order, lsDirsDeg.data(), nChannels, M_ESDtoHOA);
    return true;
}

} // namespace mpegcore
