/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include <mpegcore/HoaTranslator.hpp>

#include <algorithm>
#include <cmath>
#include <iostream>

#include <mpegcore/PlaybackMode.hpp>
#include <mpegcore/spdlog.hpp>

namespace mpegcore {

HoaTranslator::~HoaTranslator()
{
    if (m_hoaTranslatorData) {
        auto h = &*m_hoaTranslatorData;
        destroy_hoa_translator(&h);
    }
}

void HoaTranslator::initialize(Config const& config, HoaTranslatorConfig const& hoaTranslatorConfig,
    HoaParametersConfig const& hoaParametersConfig, HoaTranslator::Group group,
    std::shared_ptr<std::vector<std::uint8_t>> spatialMetadata,
    std::vector<InformedSourceConfig> informedSourceConfigs, EAR::Vector3 rescale,
    EAR::Vector3 rescaleOffset)
{
    // Check configs
    m_config = hoaTranslatorConfig;
    if (!m_config.isActive) {
        return;
    } else if (config.playbackMode != PlaybackMode::headphones) {
        SPDLOG_ERROR("HOA translation only supports playback for headphones, "
                     "the translator was not initialized.");
        return;
    }

    m_group = group;
    m_isInputESD = hoaParametersConfig.isInputESD;
    m_isInputSN3D = m_config.isInputSN3D;

    m_hoaTranslatorData = std::make_unique<hoaTranslatorData>();
    auto h = &*m_hoaTranslatorData;
    int outChannels = 2;

    auto micPosAndOrientations = vectorizeMicArrays(group, rescale, rescaleOffset);

    inputFilesAndFlags inputFiles{ m_config.matlabDirsFile, m_config.matlabHRTFFile,
        m_config.matlabITDFile, m_config.sofaFile, m_config.spatialMetadataBinaryFile,
        m_config.useMatlabFile };

    optimalMixingParameters optimalMixingParams{ m_config.lowerBandFreqLimit,
        m_config.upperBandFreqLimit, m_config.useDiffuseDecorrelation, m_config.useEqualizer,
        m_config.useRecursiveAveraging, m_config.recursionSmoothingCoefficient };

    exteriorRenderingParameters exteriorRenderingParams{ m_config.useImprovedExteriorRendering,
        m_config.exteriorProjectionRadius, m_config.adaptiveExteriorProjectionRadius };

    sphoa6DofExtensionParameters sphoaParameters{ m_config.useSphoa6DofExtension,
        m_config.sphoa6DofMaxGainDb, m_config.sphoa6dofProjectionRadius,
        m_config.NthSphoaInEncodedMetadata };

    m_informedSourceConfigs = informedSourceConfigs;
    std::vector<sourceAwareParametersPerSource> sourceAwareParamsPerSource{};
    auto sourcePositionsOpenGL = vectorizeSourceAwareObjects(sourceAwareParamsPerSource);
    sourceAwareParameters sourceAwareParams{ m_config.useSourceAwareMode,
        m_config.useSourceAwarePostFilter, m_config.useSourceAwareTimeAlignment,
        m_config.sourceResidualMode, m_config.targetScmMode, sourcePositionsOpenGL,
        m_config.maxSimulInformedSources };

    create_hoa_translator(&h, config.processingSamplesPerBuffer, config.sampleRateHz,
        m_config.hopSize, m_config.hybridMode, m_config.LDmode, outChannels,
        static_cast<int>(group.channelsPerSource), m_config.nAnalysisChannels,
        m_config.resourcePath, m_config.diffuseMethodForTargetSCM, micPosAndOrientations,
        spatialMetadata, sourceAwareParams, sourceAwareParamsPerSource);

    if (!initialize_hoa_translator(&h, m_config.groupId, m_config.useEncodedValues,
            m_config.interpolationEqGainMax, BINAURAL_DECODER_MAGLS, inputFiles,
            optimalMixingParams, exteriorRenderingParams, sphoaParameters, group.sourceLinearGains,
            group.complexityReductionParameters)) {
        destroy_hoa_translator(&h);
        SPDLOG_ERROR("The initialization of HOA translator failed.");
        return;
    }

    m_outputBuffer.init(h->outChannels, h->frameSize);

    initializeHoaInputs(group.channelsPerSource);

    m_isInitialized = true;
}

void HoaTranslator::initializeHoaInputs(std::size_t channelCount)
{
    m_hoaInputs.reserve(m_group.sourceIndices.size());
    for (auto const& hoaSourceId : m_group.sourceIndices) {
        HoaInput hoaInput{};
        hoaInput.sourceIndex = hoaSourceId;

        // Set up normalization gains for each of the channels if needed.
        if (!m_isInputESD && m_isInputSN3D) {
            hoaInput.channelGains.reserve(channelCount);
            for (std::size_t channelIndex{ 0 }; channelIndex < channelCount; ++channelIndex) {
                float SHOrder = std::ceil(std::sqrt(channelIndex + 1.0F)) - 1;
                hoaInput.channelGains.push_back({ std::sqrt(2.0F * SHOrder + 1.0F) });
            }
        } else {
            hoaInput.channelGains.resize(channelCount, { 1.F });
        }

        m_hoaInputs.push_back(std::move(hoaInput));
    }
}

void HoaTranslator::preProcess(EAR::Position const& listenerPosition)
{
    hoaTranslatorData* h = &*m_hoaTranslatorData;

    // Update listener position and orientation
    h->currentListenerPosition[0] = listenerPosition.location[0];
    h->currentListenerPosition[1] = listenerPosition.location[1];
    h->currentListenerPosition[2] = listenerPosition.location[2];
    h->yawDeg = listenerPosition.orientation[0];
    h->pitchDeg = listenerPosition.orientation[1];
    h->rollDeg = listenerPosition.orientation[2];

    // Skip the rest of pre-processing, if single-point 6DoF extension mode is used
    if (h->useSphoa6DofExtensionMode) {
        updateProcessedHoaInputs();
        return;
    }

    // Collect the interpolation weights corresponding to this frame/listener position
    pre_process_positions(h);

    // Save the current weights and triangle to the beginning of the track
    // record
    save_tracked_values(h);

    // Select active triangle and interpolation weights from a set of current
    // and previous frames
    choose_active_triangle_from_track_record(h);

    check_for_crossfade_start(h);

    check_for_source_aware_mode(h);

    determine_active_arrays(h);

    updateProcessedHoaInputs();
}

void HoaTranslator::render()
{
    hoaTranslatorData* h = &*m_hoaTranslatorData;

    // Transform to time-frequency domain
    transform_to_time_frequency_domain(h);

    // Obtain DOA angles, DTR, and energy values (metadata)
    if (h->useEncodedValues) {
        read_spatial_metadata_from_binary(h);
    } else {
        spatial_analysis(h);
    }

    //// Metadata interpolator
    interpolate_mic_metadata(h);

    //// Signal interpolator
    if (m_config.nonParametricOnly) {
        interpolate_three_closest_arrays(h);
    } else {
        interpolate_signals(h);
    }

    //// Synthesis processor
    generate_prototype_signals(h);

    if (m_config.nonParametricOnly) {
        transform_to_time_domain(h, h->currentBinauralMatrix);
    } else {
        optimal_mixing(h);
        transform_to_time_domain(h, h->outputTFDProcessingFrame);
    }

    // Write the current frame to output
    for (int ch = 0; ch < h->outChannels; ++ch) {
        copy(h->timeDomainOutputFrame[ch], m_outputBuffer[ch]);
    }

    post_rendering_actions(h);
}

void HoaTranslator::readToInput(float* data, std::size_t amount, std::size_t sourceId)
{
    auto const& hoaSourceIds = m_group.sourceIndices;
    auto it = std::find(hoaSourceIds.begin(), hoaSourceIds.end(), sourceId);
    if (it != hoaSourceIds.end()) {
        std::size_t micArrayIndex = it - hoaSourceIds.begin();
        std::copy_n(data, amount, &m_hoaTranslatorData->timeDomainInputFrame[micArrayIndex][0][0]);
    }
}

void HoaTranslator::readToInput(
    float* data, std::size_t amount, std::size_t sourceId, std::size_t channelIndex)
{
    auto const& hoaSourceIds = m_group.sourceIndices;
    auto it = std::find(hoaSourceIds.begin(), hoaSourceIds.end(), sourceId);
    if (it != hoaSourceIds.end()) {
        std::size_t micArrayIndex = it - hoaSourceIds.begin();
        std::copy_n(data, amount,
            &m_hoaTranslatorData->timeDomainInputFrame[micArrayIndex][channelIndex][0]);
    }
}

AudioBufferArray const& HoaTranslator::outputBuffer() const
{
    return m_outputBuffer;
}

void HoaTranslator::updateProcessedHoaInputs()
{
    m_processedHoaInputs.clear();
    for (auto const& i : m_hoaTranslatorData->transformedArraysForCurrentFrameFOA) {
        m_processedHoaInputs.push_back(&m_hoaInputs[i]);
    }
    for (auto const& i : m_hoaTranslatorData->transformedArraysForCurrentFrameFullOrder) {
        m_processedHoaInputs.push_back(&m_hoaInputs[i]);
    }
}

micArrayPositionsAndOrientations HoaTranslator::vectorizeMicArrays(
    Group const& group, EAR::Vector3 rescale, EAR::Vector3 rescaleOffset)
{
    auto const nMicrophoneArrays = group.sourcePositions.size();
    std::vector<std::vector<float>> arrayPositionsOpenGL(
        nMicrophoneArrays, std::vector<float>(3, 0.0F));
    std::vector<std::vector<float>> arrayOrientationsOpenGL(
        nMicrophoneArrays, std::vector<float>(3, 0.0F));
    for (int mic = 0; mic < nMicrophoneArrays; ++mic) {
        auto const& sourcePosition = group.sourcePositions[mic];
        arrayPositionsOpenGL[mic][0] =
            ((sourcePosition.location[0] - rescaleOffset[0]) * rescale[0]) + rescaleOffset[0];
        arrayPositionsOpenGL[mic][1] =
            ((sourcePosition.location[1] - rescaleOffset[1]) * rescale[1]) + rescaleOffset[1];
        arrayPositionsOpenGL[mic][2] =
            ((sourcePosition.location[2] - rescaleOffset[2]) * rescale[2]) + rescaleOffset[2];
        arrayOrientationsOpenGL[mic][0] = sourcePosition.orientation[0];
        arrayOrientationsOpenGL[mic][1] = sourcePosition.orientation[1];
        arrayOrientationsOpenGL[mic][2] = sourcePosition.orientation[2];
    }
    return micArrayPositionsAndOrientations{ arrayPositionsOpenGL, arrayOrientationsOpenGL };
}

std::vector<std::vector<float>> HoaTranslator::vectorizeSourceAwareObjects(
    std::vector<sourceAwareParametersPerSource>& sourceAwareParamsPerSource)
{
    auto const nSources = m_informedSourceConfigs.size();
    std::vector<std::vector<float>> sourcePositionsOpenGL(nSources, std::vector<float>(3, 0.0F));
    for (int src = 0; src < nSources; ++src) {
        auto const& sourcePosition = m_informedSourceConfigs[src].sourceAwareObjectPosition;
        sourcePositionsOpenGL[src][0] = sourcePosition.location[0];
        sourcePositionsOpenGL[src][1] = sourcePosition.location[1];
        sourcePositionsOpenGL[src][2] = sourcePosition.location[2];

        sourceAwareParametersPerSource params;
        params.listenerThresholdPresent = m_informedSourceConfigs[src].listenerThresholdPresent;
        params.hoaSourceThresholdPresent = m_informedSourceConfigs[src].hoaSourceThresholdPresent;
        params.listenerThreshold = m_informedSourceConfigs[src].listenerThreshold;
        params.hoaSourceThreshold = m_informedSourceConfigs[src].hoaSourceThreshold;
        params.priorityValue = m_informedSourceConfigs[src].priorityValue;
        sourceAwareParamsPerSource.push_back(params);
    }
    return sourcePositionsOpenGL;
}

std::vector<std::size_t> HoaTranslator::getFullOrderHOASourceIndices()
{
    hoaTranslatorData* h = m_hoaTranslatorData.get();
    std::vector<std::size_t> fullOrderHOASourceIndices;
    for (std::vector<std::size_t>::iterator it =
             h->transformedArraysForCurrentFrameFullOrder.begin();
         it != h->transformedArraysForCurrentFrameFullOrder.end(); ++it) {
        fullOrderHOASourceIndices.push_back(m_group.sourceIndices[*it]);
    }
    return fullOrderHOASourceIndices;
}

std::vector<std::size_t> HoaTranslator::getFirstOrderHOASourceIndices()
{
    hoaTranslatorData* h = m_hoaTranslatorData.get();
    std::vector<std::size_t> firstOrderHOASourceIndices;
    for (std::vector<std::size_t>::iterator it = h->transformedArraysForCurrentFrameFOA.begin();
         it != h->transformedArraysForCurrentFrameFOA.end(); ++it) {
        firstOrderHOASourceIndices.push_back(m_group.sourceIndices[*it]);
    }
    return firstOrderHOASourceIndices;
}

int HoaTranslator::getHOAOrder()
{
    hoaTranslatorData* h = &*m_hoaTranslatorData;
    return h->shOrder;
}

} // namespace mpegcore
