/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Panner.hpp"

#define _USE_MATH_DEFINES
#include <cassert>
#include <cmath>
#include <math.h>

namespace mpegcore {

namespace {
bool smoothAndNormalizeGains(
    std::vector<float>& targetGains, std::vector<float> const& previousGains, float oldGainWeight)
{
    float energySum{ 0.0F };

    // Smooth
    for (std::size_t ch{ 0 }, end{ targetGains.size() }; ch < end; ch++) {
        targetGains[ch] =
            oldGainWeight * previousGains[ch] + (1.F - oldGainWeight) * targetGains[ch];
        energySum += targetGains[ch] * targetGains[ch];
    }

    if (energySum == 0.F) {
        // VBAP is not working as expected if it produces zero gains. Can't normalize because zero
        // energy sum will produce zero division.
        return false;
    }

    // Normalize
    float normalizationGain = 1.F / std::sqrt(energySum);

    for (auto& targetGain : targetGains) {
        targetGain *= normalizationGain;
    }
    return true;
}

bool isConfigValid(Panner::Config const& config)
{
    if (config.bufferSize == 0) {
        return false;
    }
    if (config.oldGainWeight < 0.F || config.oldGainWeight >= 1.F) {
        return false;
    }
    if (config.spreadRadiusDeg < 0.F || config.spreadRadiusDeg > 90.F) {
        return false;
    }
    if (config.spreadGainFraction < 0.F || config.spreadGainFraction > 1.F) {
        return false;
    }
    return true;
}
} // namespace

std::optional<Panner> Panner::create(
    Panner::Config const& config, WritableAudioPointerVector const& output)
{
    if (!isConfigValid(config)) {
        return {};
    }
    auto vbapExpected = Vbap::create(config.directions);
    if (!vbapExpected.hasValue()) {
        return {};
    }
    return Panner{ config, output, vbapExpected.value() };
}

bool Panner::process(std::vector<Panner::Input>& inputs)
{
    m_output.fill(0.F);
    if (inputs.size() != m_config.inputCount) {
        assert(false && "Input count mismatch is invalid usage.");
        return false;
    }
    for (std::size_t objectIndex{ 0 }, end{ std::min(inputs.size(), m_config.inputCount) };
         objectIndex < end; objectIndex++) {
        auto const& input = inputs[objectIndex];
        for (std::size_t spreadIndex{ 0 }, end{ m_config.spreadAuxiliarySourceCount + 1U };
             spreadIndex < end; spreadIndex++) {
            // Determine VBAP gains for the current object and spread.
            float spreadGain = 0.F;
            if (input.gain > 0.0f) {
                float azi = input.azimuthDeg;
                float ele = input.elevationDeg;
                bool const isSpreadCenter{ spreadIndex == 0 };
                if (isSpreadCenter) {
                    spreadGain = 1.F - m_config.spreadGainFraction;
                } else {
                    auto angle = static_cast<float>(2 * M_PI * spreadIndex
                        / static_cast<double>(m_config.spreadAuxiliarySourceCount));
                    azi += m_config.spreadRadiusDeg * std::cos(angle);
                    ele += m_config.spreadRadiusDeg * std::sin(angle);
                    spreadGain = m_config.spreadGainFraction
                        / static_cast<float>(m_config.spreadAuxiliarySourceCount);
                }
                m_vbap.determineGains(azi, ele, m_targetChannelGains);
            } else {
                // This prevents discontinuities for signals with zero gain that are switched on/off
                // or faded in/out.
                m_targetChannelGains.assign(m_targetChannelGains.size(), 0.0f);
                // TODO: Why do we even run the panner for a RI with gain = 0 beyond the first
                //       frame after it became 0?
            }
            auto&& previousChannelGains = m_vbapSpreadGains[spreadIndex][objectIndex];
            if (m_config.oldGainWeight > 0.F) {
                if (!smoothAndNormalizeGains(
                        m_targetChannelGains, previousChannelGains, m_config.oldGainWeight)) {
                    assert(false && "Normalization failed.");
                    return false;
                }
            }

            // Fade channel audio from previous gain to current gain.
            for (std::size_t ch{ 0 }, end{ m_targetChannelGains.size() }; ch < end; ch++) {
                auto deltaGain = (m_targetChannelGains[ch] - previousChannelGains[ch])
                    / static_cast<float>(m_config.bufferSize);
                float gain = previousChannelGains[ch];
                deltaGain *= spreadGain;
                gain *= spreadGain;
                auto* output = m_output.data.at(ch);
                for (std::size_t i = 0; i < m_config.bufferSize; i++) {
                    gain += deltaGain;
                    output[i] += input.audioData[i] * gain;
                }
                previousChannelGains[ch] = m_targetChannelGains[ch];
            }
        }
    }
    return true;
}

Panner::Panner(
    Panner::Config const& config, WritableAudioPointerVector const& output, Vbap const& vbap)
    : m_config(config)
    , m_output(output)
    , m_vbap(vbap)
{
    m_targetChannelGains.resize(m_config.directions.size(), 0.F);
    m_vbapSpreadGains.resize(m_config.spreadAuxiliarySourceCount + 1);
    for (auto& objectGains : m_vbapSpreadGains) {
        objectGains.resize(m_config.inputCount);
        for (auto& channelGains : objectGains) {
            channelGains.resize(m_config.directions.size(), 0.F);
        }
    }
}

} // namespace mpegcore
