/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_PANNER_HPP
#define MPEGCORE_PANNER_HPP

#include <cstdint>
#include <optional>
#include <vector>

#include <mpegcore/AudioPointerVector.hpp>
#include <mpegcore/Vbap.hpp>

namespace mpegcore {

class Panner
{
public:
    struct Config
    {
        // Speaker node directions as defined by Vbap, layout needs to be accepted by Vbap.
        std::vector<Vbap::SpeakerNodeDirection> directions;
        // Number of samples per processed buffer. Greater than 0.
        std::uint16_t bufferSize;
        // Count of inputs to be processed on each call to process.
        std::size_t inputCount;
        // Weight of previous processing gain to smoothen transition. Greater or equal to 0, less
        // than 1.
        float oldGainWeight;
        // Count of auxiliary sources generated around an input to make the input sound wider and to
        // avoid an input localizing to a single speaker. The auxiliary sources are placed around
        // the actual input in a circle evenly.
        std::uint16_t spreadAuxiliarySourceCount;
        // Radius of the auxiliary circle in degrees. Greater or equal to 0, less or equal than 90.
        float spreadRadiusDeg;
        // Gain fraction for the auxiliary sources in total. Gain of the actual input will be
        // reduced accordingly. Each auxiliary source will get the gain fraction divided by the
        // count of auxiliary sources. Greater or equal to 0, less or equal to 1.
        float spreadGainFraction;
    };

    // Data pointer of output must remain valid, pointer vector does not need to. If config is
    // invalid return value does not contain a Panner.
    std::optional<Panner> static create(
        Config const& config, WritableAudioPointerVector const& outputBuffer);

    struct Input
    {
        float azimuthDeg;
        float elevationDeg;
        float gain;
        float* audioData;
    };

    // Count of inputs must match configured inputCount in create. Returns false if there is an
    // input count mismatch or on failed normalization, true on success.
    bool process(std::vector<Input>& inputs);

private:
    Panner(Config const& config, WritableAudioPointerVector const& output, Vbap const& vbap);

    using ChannelGains = std::vector<float>;
    using ObjectGains = std::vector<ChannelGains>;

    Config m_config;
    WritableAudioPointerVector m_output;
    std::vector<float> m_targetChannelGains;
    Vbap m_vbap;
    std::vector<ObjectGains> m_vbapSpreadGains;
};

} // namespace mpegcore

#endif // MPEGCORE_PANNER_HPP
