/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ReverberatorParameters.hpp"

#include <algorithm>
#include <array>
#include <cassert>
#include <cmath>
#include <complex>
#include <cstddef>
#include <cstdint>
#include <iostream>
#include <iterator>
#include <numeric>
#include <optional>
#include <vector>

#include <EAR/BaseTypes.h>
#include <EAR/FFT.h>

#include "dblinear.hpp"
#include "frontend/neuralNetworkParameters.hpp"
#include "frontend/neurallyControlledGraphicEqualizer.hpp"
#include "primes.hpp"
#include <hoa-t/src/HoatMatrix.hpp>
#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/DirectionalDecay.hpp>
#include <mpegcore/algorithm.hpp>
#include <mpegcore/constants.hpp>
#include <mpegcore/datatypes.hpp>
#include <mpegcore/frontend/FdnReverb.hpp>
#include <mpegcore/matlab.hpp>
#include <mpegcore/spdlog.hpp>
#include <rirlib/src/rirlib_internal.h>

namespace mpegcore {

namespace {

constexpr std::size_t OCTAVE_BAND_COUNT{ 10 };
constexpr std::size_t THIRD_OCTAVE_BAND_COUNT{ 31 };
constexpr std::array<size_t, NUM_SUPPORTED_DELAY_LINE_COUNTS> SUPPORTED_OUTPUT_COUNTS = { 3, 7, 15,
    31, 63 };

template <typename T> T mean(std::vector<T> values)
{
    auto const sum = std::accumulate(cbegin(values), cend(values), 0.);
    auto const mean = sum / values.size();
    return mean;
}

template <typename T> T maxRt60(std::vector<T> rt60Values)
{
    auto maxVal = 0.0f;
    auto const count = rt60Values.size();
    for (std::size_t i{ 0 }; i < count; ++i) {
        if (rt60Values[i] > maxVal) {
            maxVal = rt60Values[i];
        }
    }
    return maxVal;
}

struct Mode
{
    float x;
    float y;
    float z;
};

std::vector<double> bandCenterFrequencies(std::size_t bandCount)
{
    if (bandCount == 10) {
        return std::vector<double>{ 31.25, 62.5, 125, 250, 500, 1000, 2000, 4000, 8000, 16000 };
    } else if (bandCount == 31) {
        // Values created in Matlab using generating expression:
        // [1000./(2.^((1/3)*(17:-1:0))) 1000.*(2.^((1/3)*(1:13)))].';
        return std::vector<double>{ 19.6862664046074, 24.8031414370031, 31.25, 39.3725328092148,
            49.6062828740062, 62.5, 78.7450656184296, 99.2125657480125, 125, 157.490131236859,
            198.425131496025, 250, 314.980262473718, 396.85026299205, 500, 629.960524947437,
            793.7005259841, 1000, 1259.92104989487, 1587.4010519682, 2000, 2519.84209978975,
            3174.8021039364, 4000, 5039.68419957949, 6349.6042078728, 8000, 10079.368399159,
            12699.2084157456, 16000, 20158.736798318 };
    }
    SPDLOG_ERROR("Can't create band center frequencies for {} bands", bandCount);
    assert(false && "Wrong frequency band count");
    return std::vector<double>(bandCount);
}

template <typename T_FreqData>
std::size_t closestFrequencyIndex(
    double targetFrequency, std::vector<T_FreqData> const& frequencyResponse)
{
    auto const count = frequencyResponse.size();
    std::vector<double> distances(count);
    for (std::size_t i{ 0 }; i < count; ++i) {
        distances[i] = std::abs(frequencyResponse[i].frequency - targetFrequency);
    }
    return std::min_element(begin(distances), end(distances)) - begin(distances);
}

template <typename T_FreqData> struct MappedFrequencyResponse
{
    std::vector<T_FreqData> response;
    std::vector<std::size_t> inputIndices;
};

template <typename T_FreqData>
MappedFrequencyResponse<T_FreqData> mapFreqRespToBandCenterFreqs(
    std::vector<double> centerFrequencies, std::vector<T_FreqData> const& inputFrequencyResponse)
{
    MappedFrequencyResponse<T_FreqData> mapped;
    auto const centerCount = centerFrequencies.size();
    mapped.response.reserve(centerCount);
    mapped.inputIndices.reserve(centerCount);

    for (auto f : centerFrequencies) {
        auto const inputIndex = closestFrequencyIndex(f, inputFrequencyResponse);
        auto response = inputFrequencyResponse[inputIndex];
        response.frequency = f;
        mapped.response.push_back(response);
        mapped.inputIndices.push_back(inputIndex);
    }
    return mapped;
}

template <typename T_FreqData>
bool canLowerResolution(MappedFrequencyResponse<T_FreqData> const& response)
{
    // In order to output the response at the requested resolution we require
    // that at least six adjacent frequency points map to different bands
    // (somewhere in the spectrum). This is indicated by a trail of adjacent
    // ones in the index difference. In this case the given response actually
    // utilizes the denser scale. Otherwise the resolution can be lowered.
    auto d = response.inputIndices;
    std::adjacent_difference(begin(d), end(d), begin(d));
    std::size_t count{ 0 };
    std::size_t maxCount{ 0 };

    for (std::size_t i{ 1 }, end{ d.size() }; i < end; ++i) {
        if (d[i] == 1) {
            ++count;
        } else {
            maxCount = std::max(count, maxCount);
            count = 0;
        }
    }
    maxCount = std::max(count, maxCount);

    constexpr std::size_t REQUIRED_ADJACENT_ONES_COUNT{ 6 };
    return maxCount < REQUIRED_ADJACENT_ONES_COUNT;
}

template <typename T_FreqData>
auto calculateFreqRespAtBandCenterFreqs(std::vector<T_FreqData> const& frequencyResponse)
{
    auto mappedFrequencyResponse = mapFreqRespToBandCenterFreqs<T_FreqData>(
        bandCenterFrequencies(THIRD_OCTAVE_BAND_COUNT), frequencyResponse);

    if (canLowerResolution(mappedFrequencyResponse)) {
        mappedFrequencyResponse = mapFreqRespToBandCenterFreqs<T_FreqData>(
            bandCenterFrequencies(OCTAVE_BAND_COUNT), frequencyResponse);
    }
    return mappedFrequencyResponse.response;
}

std::vector<double> pickRt60(
    std::vector<AcousticParameters::FrequencyDecay> const& frequencyResponse)
{
    std::vector<double> result;
    transformToBack(frequencyResponse, result, [](auto r) { return r.rt60; });
    return result;
}

std::vector<double> pickGain(
    std::vector<AcousticParameters::FrequencyGain> const& frequencyResponse)
{
    std::vector<double> result;
    transformToBack(frequencyResponse, result, [](auto r) { return r.gain; });
    return result;
}

std::vector<AcousticParameters::FrequencyDecay> selectFreqResponseWithRdr(
    std::vector<AcousticParameters::FrequencyDecay> const& freqDecayVec)
{
    std::vector<AcousticParameters::FrequencyDecay> rdrResponse;
    rdrResponse.reserve(freqDecayVec.size());
    std::copy_if(cbegin(freqDecayVec), cend(freqDecayVec), std::back_inserter(rdrResponse),
        [](auto fdata) { return fdata.rdr.has_value(); });
    return rdrResponse;
}

template <typename T_FreqData>
std::vector<double> pickFrequency(std::vector<T_FreqData> const& freqDataVec)
{
    std::vector<double> returnFreqs;
    transformToBack(freqDataVec, returnFreqs, [](auto fdata) { return fdata.frequency; });
    return returnFreqs;
}

GraphicEqParameters createAttenuationFilter(
    std::vector<AcousticParameters::FrequencyGain> const& frequencyResponse,
    NeuralNetworkParameters const& networkParameters, std::uint32_t sampleRateHz)
{
    frontend::FrequencyResponse controlGainLinearFreqResponse;
    controlGainLinearFreqResponse.f = pickFrequency(frequencyResponse);
    controlGainLinearFreqResponse.response = pickGain(frequencyResponse);
    auto const eq = calculateCascadeEqParameters(
        controlGainLinearFreqResponse, networkParameters, sampleRateHz);
    return eq;
}

void calculateDelays(std::vector<double>& delays, const std::size_t startIndex,
    const std::vector<Mode>& roomModes, EAR::Geometry::BoundingBox const& boundingBox,
    std::uint32_t sampleRateHz)
{
    auto boundingBoxWidth = boundingBox[0].high - boundingBox[0].low;
    auto boundingBoxHeight = boundingBox[1].high - boundingBox[1].low;
    auto boundingBoxDepth = boundingBox[2].high - boundingBox[2].low;

    // Map the bounding box dimensions to the dimensions of an equivalent volume bounding box with
    // favorable ratios regarding room resonances
    std::array<float, 3> dimensionRatios = { 1.0F, 1.3F, 1.9F };
    auto boundingBoxVolume = boundingBoxWidth * boundingBoxHeight * boundingBoxDepth;
    auto tmp = boundingBoxVolume / (dimensionRatios[0] * dimensionRatios[1] * dimensionRatios[2]);
    float dim = pow(tmp, 1.0f / 3);
    boundingBoxWidth = dim * dimensionRatios[0];
    boundingBoxHeight = dim * dimensionRatios[1];
    boundingBoxDepth = dim * dimensionRatios[2];

    for (std::size_t i{ 0 }; (i < roomModes.size()) && ((startIndex + i) < delays.size()); ++i) {
        auto xMode = roomModes[i].x / boundingBoxWidth;
        auto yMode = roomModes[i].y / boundingBoxHeight;
        auto zMode = roomModes[i].z / boundingBoxDepth;
        xMode *= xMode;
        yMode *= yMode;
        zMode *= zMode;
        auto const resonanceFreq =
            0.5 * constants::SPEED_OF_SOUND_M_PER_S * std::sqrt(xMode + yMode + zMode);
        delays[startIndex + i] = sampleRateHz / resonanceFreq;
    }
}

std::vector<Mode> generateRoomModes(const std::vector<Mode>& currentModes)
{
    float minFreq = 1e10;
    float maxFreq = 0;
    for (std::size_t i{ 0 }; i < currentModes.size(); ++i) {
        auto mode = currentModes[i];
        float freq = std::sqrt(mode.x * mode.x + mode.y * mode.y + mode.z * mode.z);
        if (freq < minFreq) {
            minFreq = freq;
        } else if (freq > maxFreq) {
            maxFreq = freq;
        }
    }
    std::vector<Mode> roomModes;
    std::array<float, 6> steps = { 0.0F, 0.31F, 0.74F, 1.11F, 1.754F, 1.943F };
    for (auto xit = steps.begin(); xit != steps.end(); ++xit) {
        for (auto yit = steps.begin(); yit != steps.end(); ++yit) {
            for (auto zit = steps.begin(); zit != steps.end(); ++zit) {
                auto x = (*xit);
                auto y = (*yit);
                auto z = (*zit);
                auto f = std::sqrt(x * x + y * y + z * z);
                if (f >= minFreq && f <= maxFreq) {
                    roomModes.push_back(Mode{ x, y, z });
                }
            }
        }
    }
    return roomModes;
}

std::vector<double> determineDelayLineLengths(EAR::Geometry::BoundingBox const& boundingBox,
    std::uint32_t sampleRateHz, std::size_t numDelayLines)
{
    auto const roomW = boundingBox[0].high - boundingBox[0].low;
    auto const roomH = boundingBox[1].high - boundingBox[1].low;
    auto const roomD = boundingBox[2].high - boundingBox[2].low;

    // Index of highest frequency mode for any axis
    std::uint8_t maxMode = 2;

    /* Generate all mode permutations */
    // Number of mode permutations/resonant delay values
    auto numResDelays = (std::pow(maxMode + 1, 3)) - 1; // -1 accounts for removal of mode [0,0,0]
    std::vector<std::array<std::size_t, 3>> modePermutations;
    for (std::size_t i{ 0 }; i <= maxMode; ++i) {
        for (std::size_t j{ 0 }; j <= maxMode; ++j) {
            for (std::size_t k{ 0 }; k <= maxMode; ++k) {
                if (i > 0 || j > 0 || k > 0) // don't use mode [0,0,0]
                    modePermutations.emplace_back(std::array<std::size_t, 3>{ i, j, k });
            }
        }
    }

    // Delays for all mode permutations, in samples (float)
    std::vector<float> resonDelays_fsamp(numResDelays);
    for (std::size_t ires{ 0 }; ires < numResDelays; ++ires) {
        auto xMode = modePermutations[ires][0] / roomW;
        auto yMode = modePermutations[ires][1] / roomH;
        auto zMode = modePermutations[ires][2] / roomD;
        auto const resonanceFreq = 0.5 * constants::SPEED_OF_SOUND_M_PER_S
            * std::sqrt(xMode * xMode + yMode * yMode + zMode * zMode);
        resonDelays_fsamp[ires] = sampleRateHz / resonanceFreq;
    }
    // Sort the delays in ascending order
    std::sort(resonDelays_fsamp.begin(), resonDelays_fsamp.end());

    /* Generate log-spaced sample delays from min to max freq */

    std::vector<double> logDelays_fsamp(numDelayLines);

    //  Matlab:
    //    fs = 48e3; N = 15; % number of delays
    //    flow = 40; fhigh = 120; % lowest, highest freq
    //    st = fs/fhigh; en = fs/flow;
    //    % explicit version of: logspace( log10(fs/fhigh), log10(fs/flow), N);
    //    for i = 0:N-1   % 0-based indexing
    //        out(i+1) = st * (en/st).^(i/(N-1));
    //    end

    float minResFreq = sampleRateHz / resonDelays_fsamp.back();
    float maxResFreq = sampleRateHz / resonDelays_fsamp.front();
    minResFreq = std::max(5.f, minResFreq);
    maxResFreq = std::min(300.f, maxResFreq);

    float start = sampleRateHz / maxResFreq;
    float end = sampleRateHz / minResFreq;
    float base = end / start;
    for (std::size_t i{ 0 }; i < numDelayLines; ++i) {
        logDelays_fsamp[i] = start * std::pow(base, static_cast<float>(i) / (numDelayLines - 1));
    }

    // Nudge logfreqs to neighboring eigenfreqs

    std::vector<std::size_t> nearestLogDelIdx(numResDelays);

    // For each resonant freq's delay, find the index of the nearest log-spaced delay
    for (std::size_t ires{ 0 }; ires < numResDelays; ++ires) {
        float resDel = resonDelays_fsamp[ires];
        // initialize
        std::size_t nearestIdx = 0;
        float minDist = std::numeric_limits<float>::max();
        for (std::size_t ilog = 0; ilog < numDelayLines; ++ilog) {
            float dist = std::abs(logDelays_fsamp[ilog] - resDel);
            if (dist < minDist) {
                minDist = dist;
                nearestIdx = ilog;
            }
        }
        nearestLogDelIdx[ires] = nearestIdx;
    }

    for (std::size_t ilog{ 0 }; ilog < numDelayLines; ++ilog) {
        float logDel = logDelays_fsamp[ilog]; // log-spaced delay to be (possibly) replaced
        float minDist = std::numeric_limits<float>::max();

        for (std::size_t ires{ 0 }; ires < numResDelays; ++ires) {
            if (nearestLogDelIdx[ires] == ilog) {
                float resonDel = resonDelays_fsamp[ires];
                float dist = std::abs(logDel - resonDel);
                if (dist < minDist) {
                    minDist = dist;
                    logDelays_fsamp[ilog] = resonDel;
                }
            }
        }
    }

    // These will be passed on to be made mutually-prime integers
    return logDelays_fsamp;
}

std::optional<std::vector<Prime>> createDelayLineLengths(
    EAR::Geometry::BoundingBox const& boundingBox, std::uint32_t sampleRateHz,
    std::size_t numDelayLines)
{
    std::vector<double> delays =
        determineDelayLineLengths(boundingBox, sampleRateHz, numDelayLines);

    // Sort delay lengths shortest->longest (corresponds to output channel order)
    std::sort(begin(delays), end(delays));

    // Make delays prime (if possible)
    std::optional<std::vector<Prime>> primeDelays = mapToPrimeIntegers(delays);

    return primeDelays;
}

// Note: this method of creating the delay lines is used only in the AR initialization path, not VR
std::optional<ReverberatorCommon::DelayLineGroup> createDelayLines(
    AcousticParameters const& acousParams, EAR::Geometry::BoundingBox const& boundingBox,
    NeuralNets const& neuralNets, std::uint32_t sampleRateHz, std::size_t numDelayLines)
{
    ReverberatorCommon::DelayLineGroup delayLines(numDelayLines);

    bool hasDirectionalFreqResponse = acousParams.directionalFrequencyResponse.has_value();

    auto const primeDelaysOpt = createDelayLineLengths(boundingBox, sampleRateHz, numDelayLines);
    if (!primeDelaysOpt.has_value()) {
        return std::nullopt;
    }

    /* Encoding directions */
    std::vector<OutputDirection> encodingDirections_pol;
    if (hasDirectionalFreqResponse) {
        // RIR analysis directions
        transformToBack(*acousParams.directionalFrequencyResponse, encodingDirections_pol,
            [](AcousticParameters::DirectionalGain dFrqResp) {
                return OutputDirection(dFrqResp.azim, dFrqResp.elev, 1.f);
            });
    } else {
        // Default t-design
        mpegcore::createOutputDirections(numDelayLines, encodingDirections_pol, false);
    }

    // To Cartesian coords
    std::vector<EAR::Vector3> encodingDirections_xyz;

    transformToBack(encodingDirections_pol, encodingDirections_xyz,
        [](auto pol) { return polarToCartesian(pol); });

    EAR::Vector3 boundingDim(boundingBox[0].high - boundingBox[0].low,
        boundingBox[1].high - boundingBox[1].low, boundingBox[2].high - boundingBox[2].low);
    boundingDim *= 0.5;

    // Scale unit lenth directions to (roughly) ellipsoid dimensions fit to the room, then find the
    // distance from the origin
    std::vector<float> ellipseDists;
    for (const auto& xyz : encodingDirections_xyz) {
        EAR::Vector3 const xyz2ellipse(
            xyz[0] * boundingDim[0], xyz[1] * boundingDim[1], xyz[2] * boundingDim[2]);
        float const distance = xyz2ellipse.abs();
        ellipseDists.push_back(distance);
    }

    // Get the vector of indices that would sort the distances, ascending
    std::vector<size_t> sortedIdcs = sortIndices(ellipseDists);

    // Populate delay line properties: delayInSamples, filterParams, azimuthDeg, elevationDeg
    for (std::size_t delayIdx{ 0 }; delayIdx < numDelayLines; ++delayIdx) {

        /* Delay time */
        delayLines[delayIdx].delayInSamples = (*primeDelaysOpt)[delayIdx];

        /* Decay attenuation */
        std::vector<AcousticParameters::FrequencyGain> decayAttenuationResponse;

        // Convert rt60 to delay-proportional attentuation gain for FDN decay
        transformToBack(acousParams.frequencyDecayResponse, decayAttenuationResponse,
            [&](auto const& decayData) {
                float const attenPerSample_dB = -60.f / (sampleRateHz * decayData.rt60);
                float const attenuation_dB = attenPerSample_dB * (*primeDelaysOpt)[delayIdx];
                AcousticParameters::FrequencyGain freqGain;
                freqGain.frequency = decayData.frequency;
                freqGain.gain = dbToLinear(attenuation_dB);
                return freqGain;
            });

        auto const bandFreqResponse_decay =
            calculateFreqRespAtBandCenterFreqs(decayAttenuationResponse);
        auto const numBands_decay = bandFreqResponse_decay.size();
        auto const netIter_decay = neuralNets.find(numBands_decay);

        if (netIter_decay == neuralNets.end()) {
            SPDLOG_ERROR("Erroneous decay filter band count {}", numBands_decay);
            return std::nullopt;
        }

        auto const filterParams_decay =
            createAttenuationFilter(bandFreqResponse_decay, *netIter_decay->second, sampleRateHz);

        delayLines[delayIdx].filterParams = std::move(filterParams_decay);

        /* Directional assignmentx */
        // Assign the directions so that they corresponding to the shortest-to-longest delay
        // times/distances, according to an ellipsoid fit to the room geometry

        size_t direcIdx = sortedIdcs[delayIdx];
        // size_t direcIdx = delayIdx; // to bypass delay fitting to bounding box ellipsoid

        delayLines[delayIdx].azimuthDeg = encodingDirections_pol[direcIdx].data()[0];
        delayLines[delayIdx].elevationDeg = encodingDirections_pol[direcIdx].data()[1];

        /* Directional gain */
        if (hasDirectionalFreqResponse) {

            // Convert freqs of the analyzed bands to freqs of filter bands
            // Be sure to used the directional index from the above order sorted by ascending
            // delay/distance
            auto const directionGain_bnd = calculateFreqRespAtBandCenterFreqs(
                (*acousParams.directionalFrequencyResponse)[direcIdx].frequencyResponse);
            auto const numBands_dGain = directionGain_bnd.size();
            auto const netIter_dGain = neuralNets.find(numBands_dGain);

            if (netIter_dGain == neuralNets.end()) {
                SPDLOG_ERROR("Invalid directional gain filter band count: {}", numBands_dGain);
                return std::nullopt;
            }

            auto const filterParams_dGain =
                createAttenuationFilter(directionGain_bnd, *netIter_dGain->second, sampleRateHz);
            delayLines[delayIdx].direcGainFilterParams = std::move(filterParams_dGain);

        } else {
            // No directional gain
            delayLines[delayIdx].direcGainFilterParams = std::nullopt;
        }
    }

    return delayLines;
}

AudioBufferArray renderImpulse(ReverberatorParameters const& reverbParams,
    std::size_t numSamples_tail, std::size_t numSamples_predelay)
{
    /* numSamples_predelay denotes the sample index of when the caller expects the
     first echo to be output from the FDN. The input buffer is sized accordingly and
     the input impulse is positioned in the input buffer with the appropriate delay
     so that the inpulseIndex + shortestDelayLineInSamples = numSamples_predelay. The
     minimum sample delay will always be the length of the shortest delay line, so
     numSamples_predelay less than can still be requested but won't be realized.
     */

    std::size_t numSamples_total = numSamples_tail + numSamples_predelay;
    AudioBufferArray output(reverbParams.delayLines.size(), numSamples_total);
    std::vector<float> input(numSamples_total, 0.F);
    std::size_t impulseIndex{ 0 };

    if (numSamples_predelay > 0) {
        // The first delay line has the shortest delay
        auto const numSamples_shortestDelayLine = reverbParams.delayLines[0].delayInSamples;
        if (numSamples_predelay > numSamples_shortestDelayLine) {
            impulseIndex = numSamples_predelay - numSamples_shortestDelayLine;
        } else {
            SPDLOG_WARN("[renderImpulse] Requested pre-delay ({} samples) is shorter than the "
                        "minimum delay line length ({} samples). Pre-delay can't be reached.",
                numSamples_predelay, numSamples_shortestDelayLine);
        }
    }
    input[impulseIndex] = 1.F;
    frontend::FdnReverb reverb(reverbParams);
    reverb.setInputBuffer(input);
    reverb.setOutputBuffer(output);
    reverb.process(static_cast<std::uint32_t>(numSamples_total));
    return output;
}

std::vector<float> sumBuffers(AudioBufferArray const& input)
{
    auto const sampleCount{ input[0].size() };
    std::vector<float> sum(sampleCount, 0.F);
    for (std::size_t i{ 0 }; i < sampleCount; ++i) {
        for (auto const& buffer : input) {
            sum[i] += buffer[i];
        }
    }
    return sum;
}

// Based on https://jameshfisher.com/2018/03/30/round-up-power-2
std::uint64_t nextPow2(std::uint64_t x)
{
    auto nextPow2m1 = [](std::uint64_t x) {
        x |= x >> 1;
        x |= x >> 2;
        x |= x >> 4;
        x |= x >> 8;
        x |= x >> 16;
        x |= x >> 32;
        return x;
    };

    return nextPow2m1(x - 1) + 1;
}

bool isPowerOfTwo(std::size_t n)
{
    // A power of two has only one bit set in its binary representation.
    // So, if n is a power of two, n & (n - 1) will be 0.
    return (n > 0) && ((n & (n - 1)) == 0);
}

std::vector<float> abs(std::vector<std::complex<float>> const& data)
{
    std::vector<float> abs;
    transformToBack(data, abs,
        [](auto const c) { return std::sqrt(c.real() * c.real() + c.imag() * c.imag()); });

    return abs;
}

std::vector<float> singleSidedSpectralEnergyDensity(std::vector<float> const& spectrum)
{
    auto const halfCount{ spectrum.size() / 2 + 1 };
    auto const energyScaler = 1.0 / spectrum.size();
    std::vector<float> singleSidedEnergyDensity(halfCount);
    // DC and nyqyist frequency handled separately
    singleSidedEnergyDensity[0] = static_cast<float>(energyScaler * spectrum[0] * spectrum[0]);
    auto const nyquistFreqIndex = halfCount - 1;
    singleSidedEnergyDensity[nyquistFreqIndex] =
        static_cast<float>(energyScaler * spectrum[nyquistFreqIndex] * spectrum[nyquistFreqIndex]);
    // add energy from negative frequencies to preserve total energy
    for (std::size_t i{ 1 }; i < halfCount - 1; ++i) {
        singleSidedEnergyDensity[i] =
            static_cast<float>(2 * energyScaler * spectrum[i] * spectrum[i]);
    }
    return singleSidedEnergyDensity;
}

std::vector<float> linSpace(double x1, double x2, std::size_t N)
{
    std::vector<float> linSpace(N);
    auto const step = (x2 - x1) / (N - 1);
    for (std::size_t i{ 0 }; i < N; ++i) {
        linSpace[i] = static_cast<float>(x1 + i * step);
    }
    return linSpace;
}

struct SpectralEnergyDensity
{
    std::vector<float> frequencies;
    std::vector<float> spectralEnergyDensity;
};

SpectralEnergyDensity calculateSpectralEnergyDensity(std::vector<float> const& impulseResponse,
    std::uint32_t sampleRateHz, std::uint32_t& outNfft, const std::size_t startSamp = 0)
{
    std::size_t irLength = impulseResponse.size();
    if (!isPowerOfTwo(irLength)) {
        SPDLOG_ERROR(
            "calculateSpectralEnergyDensity: impulseResponse needs to be a power of two. {}",
            irLength);
    }
    if (startSamp > impulseResponse.size() - 1) {
        SPDLOG_ERROR(
            "calculateSpectralEnergyDensity: startSamp {} exceeds the size of impulseResponse {}",
            startSamp, impulseResponse.size());
    }
    std::vector<float> irDiffuseOnly(irLength, 0.f);
    std::copy(impulseResponse.begin() + startSamp, impulseResponse.end(),
        irDiffuseOnly.begin() + startSamp);

    std::vector<std::complex<float>> fftOutput(irLength);
    auto fftr = EAR::RFFT::create(static_cast<int>(irLength));
    fftr->transform(irDiffuseOnly.data(), fftOutput.data());

    auto const fftOutputAbs = abs(fftOutput);
    auto const positiveHalfSpectralEnergyDensity = singleSidedSpectralEnergyDensity(fftOutputAbs);

    auto const spectrumValueCount = positiveHalfSpectralEnergyDensity.size();
    auto const dFreq = static_cast<double>(sampleRateHz) / irLength;

    SpectralEnergyDensity response;
    response.spectralEnergyDensity = positiveHalfSpectralEnergyDensity;
    response.frequencies = linSpace(0, dFreq * (spectrumValueCount - 1), spectrumValueCount);
    outNfft = static_cast<std::uint32_t>(irLength);

    return response;
}

SpectralEnergyDensity calculateSpectralEnergyDensityOfAUnityImpulse(
    const std::vector<float>& frequencies, std::uint32_t nfft)
{
    SpectralEnergyDensity response;
    response.spectralEnergyDensity = std::vector<float>(frequencies.size());
    std::fill(
        response.spectralEnergyDensity.begin(), response.spectralEnergyDensity.end(), 2.0f / nfft);
    response.frequencies = frequencies;
    return response;
}

struct BandFrequencies
{
    std::vector<double> lowHz;
    std::vector<double> centerHz;
    std::vector<double> highHz;
};

BandFrequencies bandFrequencies(std::size_t bandCount, std::uint32_t sampleRateHz)
{
    BandFrequencies bands;
    assert(bandCount == OCTAVE_BAND_COUNT || bandCount == THIRD_OCTAVE_BAND_COUNT);
    auto const RATIO =
        bandCount == THIRD_OCTAVE_BAND_COUNT ? std::pow(2., 1. / 6) : std::pow(2., 1. / 2);
    bands.centerHz = bandCenterFrequencies(bandCount);
    transformToBack(bands.centerHz, bands.lowHz, [&](auto f) { return f / RATIO; });
    transformToBack(bands.centerHz, bands.highHz, [&](auto f) { return f * RATIO; });
    bands.lowHz.front() = 0.0;
    bands.highHz.back() = 0.5 * sampleRateHz;
    return bands;
}

std::vector<float> calculateBandEnergies(SpectralEnergyDensity const& spectralEnergyDensity,
    std::size_t bandCount, std::uint32_t sampleRateHz)
{
    auto const bands = bandFrequencies(bandCount, sampleRateHz);
    std::vector<float> bandEnergyDensity(bandCount, 0.F);
    std::vector<std::size_t> perBandResponseCount(bandCount, 0);
    std::size_t responseIndex{ 0 };
    auto const responseCount = spectralEnergyDensity.frequencies.size();

    // Drop spectral energy density values whose frequency is lower than the lowest band's
    // lowest frequency. Then the main loop can just keep summing each band up
    // to it's highest frequency.
    while (responseIndex < responseCount
        && spectralEnergyDensity.frequencies[responseIndex] < bands.lowHz[0]) {
        ++responseIndex;
    }

    for (std::size_t bandIndex{ 0 }; bandIndex < bandCount; ++bandIndex) {
        auto const bandHighHz = bands.highHz[bandIndex];
        while (responseIndex < responseCount
            && spectralEnergyDensity.frequencies[responseIndex] <= bandHighHz) {
            bandEnergyDensity[bandIndex] +=
                spectralEnergyDensity.spectralEnergyDensity[responseIndex];
            ++responseIndex;
        }
    }

    return bandEnergyDensity;
}

std::optional<GraphicEqParameters> createRdrFilter(AcousticParameters const& acousticParameters,
    std::vector<float> const& fdnImpulseResponse, NeuralNets const& neuralNets,
    std::uint32_t sampleRateHz, const std::size_t lateStartSamp = 0)
{
    auto inputRdr = selectFreqResponseWithRdr(acousticParameters.frequencyDecayResponse);
    if (inputRdr.size() < 1) {
        return std::nullopt;
    }
    auto const inputRdr_banded = calculateFreqRespAtBandCenterFreqs(inputRdr);
    auto const bandCount = inputRdr_banded.size();

    // Late spectral energy density as _measured_ from a rendered IR using the
    // reverbParams (info contained in DelayLines)
    std::uint32_t nfft = 0;
    auto const lateSpectralEnergyDensity =
        calculateSpectralEnergyDensity(fdnImpulseResponse, sampleRateHz, nfft, lateStartSamp);
    if (nfft == 0) {
        SPDLOG_ERROR("Invalid FFT size {}!", nfft);
    }

    auto const lateSpectralEnergyDensity_banded =
        calculateBandEnergies(lateSpectralEnergyDensity, bandCount, sampleRateHz);

    auto const directSpectralEnergyDensity =
        calculateSpectralEnergyDensityOfAUnityImpulse(lateSpectralEnergyDensity.frequencies, nfft);

    auto const directSpectralEnergyDensity_banded =
        calculateBandEnergies(directSpectralEnergyDensity, bandCount, sampleRateHz);

    // Frequency response of the RDR correction filter
    frontend::FrequencyResponse rdrCorrection;
    rdrCorrection.f = pickFrequency(inputRdr_banded);
    rdrCorrection.response.reserve(bandCount);

    for (std::size_t i{ 0 }; i < bandCount; ++i) {
        if (directSpectralEnergyDensity_banded[i] <= 0) {
            SPDLOG_ERROR(
                "Erroneous direct energy {} at band {}", directSpectralEnergyDensity_banded[i], i);
        }

        auto measuredRdr =
            lateSpectralEnergyDensity_banded[i] / directSpectralEnergyDensity_banded[i];
        if (measuredRdr < 1e-10) {
            SPDLOG_WARN("Clamping measured RDR of band {} to a small value.", i);
            measuredRdr = static_cast<float>(1e-10);
        }

        // Ratio of the magnitudes (sqrt) of the input (requested) RDR to measured RDR
        float rdrGain = std::sqrt(inputRdr_banded[i].rdr.value()) / std::sqrt(measuredRdr);
        rdrCorrection.response.push_back(rdrGain);
    }

    auto const netIt = neuralNets.find(bandCount);
    if (netIt == neuralNets.end()) {
        SPDLOG_ERROR("Erroneous filter band count {}", bandCount);
        return std::nullopt;
    }

    auto const params = calculateCascadeEqParameters(rdrCorrection, *netIt->second, sampleRateHz);

    return params;
}

std::vector<EAR::Vector3> polarToCartesianVectors(const std::vector<OutputDirection>& in)
{
    std::vector<EAR::Vector3> outputDirectionVectors;
    for (std::size_t l = 0; l < in.size(); l++) {
        outputDirectionVectors.push_back(polarToCartesian(in[l]));
    }
    return outputDirectionVectors;
}

std::vector<EAR::Polar> cartesianToPolarCoordinates(const std::vector<EAR::Vector3>& in)
{
    std::vector<EAR::Polar> outputDirectionVectors;
    for (std::size_t l = 0; l < in.size(); l++) {
        outputDirectionVectors.push_back(cartesianToPolar(in[l]));
    }
    return outputDirectionVectors;
}

// Create N points nearly uniformly distributed on the sphere
// https://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere
// Original Matlab implementation: x forward, y left, z upward
// EAR/OpenGL: x right, y upward, z back
// Note: fibonacciSphere() produces a less-uniform design than fibonacciSphere2() on account
// of points being inserted on the lateral poles regardless of design size.
std::vector<EAR::Polar> fibonacciSphere(
    std::size_t numDirections, std::vector<EAR::Vector3>& outputDirections)
{
    outputDirections = std::vector<EAR::Vector3>(numDirections);
    auto phi = M_PI * (3.0f - sqrt(5.0f));
    for (std::size_t id = 0; id < numDirections; id++) {
        auto y = 1.0f - (id / static_cast<float>(numDirections - 1.0f)) * 2.0f;
        auto radius = sqrt(1.0f - y * y);
        auto theta = phi * id;
        auto x = cos(theta) * radius;
        auto z = sin(theta) * radius;
        outputDirections[id] = EAR::Vector3(-y, z, -x);
    }
    std::vector<EAR::Polar> outputPolarDirections = cartesianToPolarCoordinates(outputDirections);

    return outputPolarDirections;
}

// Á. González, “Measurement of areas on a sphere using Fibonacci and latitude-longitude lattices,”
// Math Geosci, vol. 42, no. 1, pp. 49–64, Jan. 2010, doi: 10.1007/s11004-009-9257-x.
// This publication notes that the algorithm is valid for odd numbers of points, but also works for
// even point counts.
std::vector<EAR::Polar> fibonacciSphere2(
    std::size_t numDirections, std::vector<EAR::Vector3>& outputDirections)
{
    std::vector<EAR::Polar> outputPolarDirections(numDirections);
    const float phi = (1.f + std::sqrt(5.f)) / 2.f; // golden ratio
    const float N = (numDirections - 1) / 2.f;
    float n_iter = -1.f * N;

    for (std::size_t id = 0; id < numDirections; id++) {
        float el = std::asin(2.f * n_iter / (2 * N + 1)) * 180.f / M_PI;
        float az = std::fmod(n_iter, phi) * 360.f / phi;
        if (az < -180.f)
            az += 360.f;
        else if (az > 180.f)
            az -= 360.f;

        outputPolarDirections[id] = EAR::Polar(az, el, 1.f);
        n_iter += 1.f;
    }

    return outputPolarDirections;
}

EAR::Polar roundAziEle(const EAR::Polar& in)
{
    return EAR::Polar(std::roundf(in[0]), std::roundf(in[1]), in[2]);
}

} // namespace

void createOutputDirections(std::size_t numDirections,
    std::vector<OutputDirection>& outputDirectionsPolar, // azim (deg), elev (deg), rho
    bool forceLegacyDirections // use fibonacciSphere() to generate directions (def: false)
)
{
    outputDirectionsPolar.resize(numDirections);

    // t-designs matching the SUPPORTED_OUTPUT_COUNTS >= 15, which align with those used in the RIR
    // analysis. Note there are more designs available in the rirlib if needed.
    // These designs have had two stages of optimization:
    //  1. A rotation that places two adjacent points on the horizontal plane, symmetric to forward
    //  axis (the Panner requires that there be two directions on the horizontal plane separated by
    //  less than 180 degrees). Hence the name "FWPAIR", i.e. forward pair.
    //  2. A permutation of the design indices such that successive indices are maximally distant
    //  from one another, to distribute the sorted ascending delay lengths around the sphere. This
    //  was done by brute force search of the longest connecting path over 10e6 iterations. See RIR
    //  lib for details.
    // NOTE: For the proper application of directional gain, the selected design must align with
    // that used in SHDecay::init().
    bool loadTDesign = !forceLegacyDirections;
    SPHDESIGN_PRESETS tDesignPreset;
    switch (numDirections) {
    case 15:
        tDesignPreset = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_15;
        break;
    case 16:
        tDesignPreset = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_16;
        break;
    case 31:
        tDesignPreset = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_31;
        break;
    case 32:
        tDesignPreset = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_32;
        break;
    case 63:
        tDesignPreset = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_63;
        break;
    case 64:
        tDesignPreset = SPHDESIGN_PRESET_T_DESIGN_FWPAIR_64;
        break;
    default:
        // otherwise generate a fibonacci sphere, losing correspondence with directional gain
        // analysis directions
        loadTDesign = false;
    }

    if (loadTDesign && numDirections <= RIRLIB_MAX_NUM_OUTPUTS) {
        float dirsDeg_azel[RIRLIB_MAX_NUM_OUTPUTS][2];
        int outNumChan;
        loadSphDesignPreset(tDesignPreset, dirsDeg_azel, &outNumChan);
        for (std::size_t id = 0; id < numDirections; ++id) {
            outputDirectionsPolar[id] = EAR::Polar(dirsDeg_azel[id][0], dirsDeg_azel[id][1], 1.f);
        }
        // Do no reordering, directional gain requires the assigned directions to remain unchanged.

    } else {
        // Non-optimized catch-all to support any number of directions (and allow use of "legacy"
        // directions). Note that directions assigned here will deviate from the spherical designs
        // used in RIR directional analysis and the use of directional gain (if enabled).

        std::vector<EAR::Vector3> outputDirs_xyz;
        std::vector<EAR::Polar> outputDirsFib_polar(numDirections);

        // Note: fibonacciSphere() produces a less-uniform design than fibonacciSphere2(). However,
        // it passes the Panning initialization criteria which requires two points on the horizontal
        // plane separated by 180 deg or less. fibonacciSphere2() would require a rotation for this.
        outputDirsFib_polar = fibonacciSphere(numDirections, outputDirs_xyz);

        // Because delay line outputs are sorted by ascending delay length order, we assign every
        // other length delay lines across the median plane, creating some simple separation of
        // similar delays.
        std::size_t cnt = 0;
        for (std::size_t id_even = 0; id_even < numDirections; id_even += 2) {
            outputDirectionsPolar[id_even] = roundAziEle(outputDirsFib_polar[cnt]);
            if ((id_even + 1) < numDirections) {
                outputDirectionsPolar[id_even + 1] =
                    roundAziEle(outputDirsFib_polar[numDirections - 1 - cnt]);
            }
            cnt += 1;
        }
    }
}

float euclideanDistance(const EAR::Vector3& v1, const EAR::Vector3& v2)
{
    auto xdiff = v1.data()[0] - v2.data()[0];
    auto ydiff = v1.data()[1] - v2.data()[1];
    auto zdiff = v1.data()[2] - v2.data()[2];
    auto d = std::sqrt(xdiff * xdiff + ydiff * ydiff + zdiff * zdiff);
    return d;
}

Matrix2d distanceMatrix(
    const std::vector<EAR::Vector3>& directions1, const std::vector<EAR::Vector3>& directions2)
{
    // Pairwise distances between all output positions
    std::size_t M = directions1.size();
    std::size_t N = directions2.size();
    Matrix2d distMtx(M, N, 0.0);
    for (std::size_t r = 0; r < M; r++) {
        for (std::size_t c = 0; c < N; c++) {
            float d = euclideanDistance(directions1[r], directions2[c]);
            distMtx.setValue(r, c, d);
        }
    }
    return distMtx;
}

std::vector<size_t> closestColIndices_exclusive(const Matrix2d& distanceMatrix)
{
    std::vector<size_t> closestIndices(distanceMatrix.m);
    std::unordered_map<std::size_t, bool> selected;
    for (std::size_t r = 0; r < distanceMatrix.m; r++) {
        std::size_t closestIdx = 0;
        double minDist = 1e10;
        for (std::size_t c = 0; c < distanceMatrix.n; c++) {
            if (distanceMatrix.getValue(r, c) < minDist && selected.find(c) == selected.end()) {
                minDist = distanceMatrix.getValue(r, c);
                closestIdx = c;
            }
        }
        closestIndices[r] = closestIdx;
        selected.insert({ closestIdx, true });
    }
    return closestIndices;
}

std::vector<size_t> closestColIndices(const Matrix2d& distanceMatrix)
{
    std::vector<size_t> closestIndices(distanceMatrix.m);
    for (std::size_t r = 0; r < distanceMatrix.m; r++) {
        std::size_t closestIdx = 0;
        double minDist = 1e10;
        for (std::size_t c = 0; c < distanceMatrix.n; c++) {
            double thisDist = distanceMatrix.getValue(r, c);
            if (thisDist < minDist) {
                minDist = thisDist;
                closestIdx = c;
            }
        }
        closestIndices[r] = closestIdx;
    }
    return closestIndices;
}

template <typename T> std::vector<size_t> sortIndices(const std::vector<T>& valuesToSort)
{
    // Create a vector of pairs where each pair consists of a value and its index
    std::vector<std::pair<T, size_t>> indexed_values;
    indexed_values.reserve(valuesToSort.size());
    for (size_t i = 0; i < valuesToSort.size(); ++i) {
        indexed_values.push_back(std::make_pair(valuesToSort[i], i));
    }

    // Sort the vector of pairs based on the values
    std::sort(indexed_values.begin(), indexed_values.end());

    // Extract the indices from the sorted pairs
    std::vector<size_t> indices;
    indices.reserve(valuesToSort.size());
    for (const auto& pair : indexed_values) {
        indices.push_back(pair.second);
    }

    return indices;
}

void createOutputMapping(const std::string reverbID,
    const std::vector<EAR::Vector3>& reverbDirections_xyz,
    const std::vector<EAR::Vector3>& outputDirections_xyz, OutputMapping& outputMapping)
{
    std::size_t numOutputDirections = outputDirections_xyz.size();
    std::size_t numReverbDirections = reverbDirections_xyz.size();

    if (numOutputDirections < SUPPORTED_OUTPUT_COUNTS.front()
        || numOutputDirections > SUPPORTED_OUTPUT_COUNTS.back()) {
        return;
    }
    if (numReverbDirections < SUPPORTED_OUTPUT_COUNTS.front()
        || numReverbDirections > SUPPORTED_OUTPUT_COUNTS.back()) {
        return;
    }

    // Map the reverb directions to the nearest output directions.
    auto distMatrix = distanceMatrix(reverbDirections_xyz, outputDirections_xyz);
    auto nearestOutputIdcs = closestColIndices(distMatrix);

    outputMapping.insert({ reverbID, nearestOutputIdcs });

    return;
}

/* Generate ReverberatorParameters (including delay lines) from AcousticParameters (AR) */
std::optional<ReverberatorParameters> createReverbParamsFromAcousticParams(
    AcousticParameters const& acousticParameters, EAR::Geometry::BoundingBox boundingBox,
    std::uint32_t sampleRateHz, std::size_t numDelayLines)
{
    if (std::find(SUPPORTED_OUTPUT_COUNTS.begin(), SUPPORTED_OUTPUT_COUNTS.end(), numDelayLines)
        == SUPPORTED_OUTPUT_COUNTS.end()) {
        return std::nullopt;
    }

    ReverberatorParameters reverbParams;
    auto const nnParams10 = mpegcore::frontend::octaveNeuralNetworkParameters();
    auto const nnParams31 = mpegcore::frontend::thirdOctaveNeuralNetworkParameters();
    NeuralNets neuralNets{ { 10, &nnParams10 }, { 31, &nnParams31 } };

    // Create and assign delay lines
    auto delayLines =
        createDelayLines(acousticParameters, boundingBox, neuralNets, sampleRateHz, numDelayLines);
    if (!delayLines.has_value()) {
        return std::nullopt;
    }
    reverbParams.delayLines = delayLines.value();

    /* Render an FDN impulse response to measure its RDR and diffuse onset */

    auto const maxRt60Sec = maxRt60(pickRt60(acousticParameters.frequencyDecayResponse));
    auto const preDelaySec = acousticParameters.preDelaySec.value_or(0);

    // Because we're measuring the diffuse onset of the FDN only, not including any predelay line,
    // we don't want to delay the input impulse.
    auto const inputDelaySampleCount = 0;

    // For FFT (Ooura used by default in EAR::RFFT) the sample count must be a power of two, so the
    // tail length must be a power of two.
    auto const tailSampleCount =
        nextPow2(static_cast<std::uint64_t>(std::round(maxRt60Sec * sampleRateHz)));

    // Multichannel FDN output
    AudioBufferArray fdnIR_multichan =
        renderImpulse(reverbParams, tailSampleCount, inputDelaySampleCount);
    std::vector<float> fdnIR_mono = sumBuffers(fdnIR_multichan);
    std::size_t fdnDiffuseOnsetIdx =
        getFDNDiffuseOnsetIdx(fdnIR_multichan, reverbParams, sampleRateHz);

    reverbParams.fdnDiffuseOnsetSec = static_cast<float>(fdnDiffuseOnsetIdx) / sampleRateHz;
    reverbParams.eqFilterParams = createRdrFilter(
        acousticParameters, fdnIR_mono, neuralNets, sampleRateHz, fdnDiffuseOnsetIdx);

    printStatus("Diffuse onset (FDN): {:.5f} sec", reverbParams.fdnDiffuseOnsetSec);

    return reverbParams;
}

std::size_t getFDNDiffuseOnsetIdx(AudioBufferArray& multichanFdnIR,
    ReverberatorParameters const& reverbParams, std::uint32_t sampleRateHz)
{
    printStatus("Analyzing FDN response for diffuse onset...");

    const std::size_t nFdnSamp = multichanFdnIR.samplesPerBuffer();
    const std::size_t nFdnChan = multichanFdnIR.size();
    const int shOrder_diff = 1; // analysis happens at first order
    const int nSH_diff = ORDER2NSH(shOrder_diff);

    mpegcore::SHDecay fdnDecayInfo{
        "", // path not needed for internal buffer analysis
        mpegcore::HoaNormalization::n3d // FDN encoded with N3D normalization
    };
    fdnDecayInfo.srcDirectivityFlag = 0; // omni directivity
    fdnDecayInfo.init(reverbParams.delayLines.size());

    // Matrix: FDN IR channels
    hoatMatrix2D fdnChannelMtx;
    fdnChannelMtx.nRows = static_cast<int>(nFdnChan);
    fdnChannelMtx.nCols = static_cast<int>(nFdnSamp);
    fdnChannelMtx.data = multichanFdnIR.rawPointerArray();

    // Inject a unit impulse at the head of the FDN IR, from a single direction, for the "direct
    // sound". This helps disambiguate the direct onset from the first echo
    fdnChannelMtx.data[0][0] = 1.f;

    // Matrix: FOA-encoded FDN IR
    hoatMatrix2D foaFdnMtx;
    createHoatMatrix(&foaFdnMtx, nSH_diff, fdnChannelMtx.nCols);

    // Matrix: Beamforming coefficients
    hoatMatrix2D beamformMtx;
    createHoatMatrix(&beamformMtx, fdnChannelMtx.nRows, nSH_diff);

    /* Get beamforming coeffs for the encoding spherical design */
    // Coeffs of m = 0 components
    float* c_l = static_cast<float*>(malloc1d((shOrder_diff + 1) * sizeof(float)));
    beamWeightsHypercardioid2Spherical(shOrder_diff, c_l);

    rirlib_data* pAnalysis = (rirlib_data*)fdnDecayInfo.pRirLib;

    for (int idir = 0; idir < nFdnChan; idir++) {
        rotateAxisCoeffsReal(shOrder_diff, c_l,
            (SAF_PI / 2.0f) - (pAnalysis->sphDirsDeg[idir][1] * SAF_PI / 180.0f),
            pAnalysis->sphDirsDeg[idir][0] * SAF_PI / 180.0f, &beamformMtx.data[idir][0]);
    }

    free(c_l);

    /* Encode FDN channels to FOA (ACN-N3D) via transposed decoding matrix */

    // Note: Using cblas_sgemm matrix mul wasn't successful, so doing explicit matrix mul :(
    // matrixMultTransNoTrans(decoderMtx, fdnChannelMtx, foaFdn);
    for (std::size_t ismp = 0; ismp < nFdnSamp; ismp++) {
        for (std::size_t ish = 0; ish < nSH_diff; ish++) {
            float sum = 0.f;
            for (std::size_t ich = 0; ich < nFdnChan; ich++) {
                sum += fdnChannelMtx.data[ich][ismp] * beamformMtx.data[ich][ish];
            }
            foaFdnMtx.data[ish][ismp] = sum;
        }
    }

    // foaFdnMtx needs to be const float **
    const float* constPtrArray[nSH_diff];
    for (int i = 0; i < nSH_diff; i++) {
        constPtrArray[i] = foaFdnMtx.data[i];
    }

    fdnDecayInfo.setRIR_buffer(constPtrArray, nSH_diff, nFdnSamp, sampleRateHz);

    // We only need to analyze through the diffuse onset measurement stage
    fdnDecayInfo.analyze(DIFFUSENESS_ONSET_FOUND);

    bool diffuseOnsetIsValid =
        true; // could add further check here on validity of the diffuse onset calc
    std::size_t fdnDiffuseOnsetIdx;
    if (diffuseOnsetIsValid) {
        // Use the diffuse onset _index_ because it references the beginning of the FDN's IR,
        // whereas diffuseOnsetSec references a measured t0, which accounts for direct onset
        // measurement and src->rec distance, which are not relevant to the FDN IR.
        fdnDiffuseOnsetIdx = static_cast<std::size_t>(pAnalysis->diffuseOnsetIdx);
    } else {
        // Fall back: diffuse onset is the time of the FDN's last echo delay
        fdnDiffuseOnsetIdx = reverbParams.delayLines.back().delayInSamples;
    }

    return fdnDiffuseOnsetIdx;
}

std::array<size_t, NUM_SUPPORTED_DELAY_LINE_COUNTS> getSupportedDelayLineCounts()
{
    return SUPPORTED_OUTPUT_COUNTS;
}

std::size_t complexityLevelToNumDelayLines(std::size_t complexityLevel)
{
    const std::size_t numDelayLineCounts = NUM_SUPPORTED_DELAY_LINE_COUNTS;
    std::array<size_t, numDelayLineCounts> delayLineCounts = getSupportedDelayLineCounts();
    switch (complexityLevel) {
    case 1:
        return delayLineCounts[0];
        break;
    case 2:
        return delayLineCounts[1];
        break;
    case 3:
        return delayLineCounts[2];
        break;
    case 4:
        return delayLineCounts[3];
        break;
    case 5:
        return delayLineCounts[4];
        break;
    default:
        return delayLineCounts[2];
        break;
    }
}

} // namespace mpegcore
