/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ShBinDecoder.hpp"

#include <algorithm>
#include <cassert>
#include <unordered_map>

#include <ambi_bin.h>

#include "AmbiStore.hpp"
#include "saf.hpp"
#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/Hoa.hpp>
#include <mpegcore/spdlog.hpp>

namespace mpegcore {

bool operator==(BinAmbiParams const& a, BinAmbiParams const& b)
{
    return a.order == b.order && a.sampleRateHz == b.sampleRateHz;
};

namespace {
AmbiStore<BinAmbi, BinAmbiParams> store;
}

ShBinDecoder::ShBinDecoder(BinAmbiParams params)
{
    m_ambiHandle = store.reserveAmbi(params);
}

ShBinDecoder::ShBinDecoder(ShBinDecoder&& other) noexcept
{
    this->m_ambiHandle = other.m_ambiHandle;
    other.m_ambiHandle = {};
}

ShBinDecoder::~ShBinDecoder()
{
    if (m_ambiHandle.has_value()) {
        store.freeAmbi(m_ambiHandle.value());
        m_ambiHandle = {};
    }
}

bool ShBinDecoder::isInitialized()
{
    return m_ambiHandle.has_value();
}

void ShBinDecoder::decode(AudioBufferArray const& inputs, AudioBufferArray& outputs)
{
    assert(inputs[0].size() == outputs[0].size());
    assert(m_ambiHandle.has_value());
    if (!m_ambiHandle.has_value()) {
        SPDLOG_ERROR("no ambi handle, cannot decode.");
        return;
    }
    ambi_bin_process(store.getAmbi(m_ambiHandle.value()).ambi,
        const_cast<float** const>(inputs.rawPointerArray()), outputs.rawPointerArray(),
        static_cast<int>(inputs.size()), static_cast<int>(outputs.size()),
        static_cast<int>(inputs[0].size()));
}

bool BinAmbi::init()
{
    ambi_bin_create(&ambi);
    ambi_bin_setSofaFilePath(ambi, params.hrirSofaFile.c_str());
    ambi_bin_setNormType(ambi, toSafHoaNormalization(params.normalization));
    ambi_bin_setInputOrderPreset(ambi, toSafShOrders(params.order));
    ambi_bin_setDecodingMethod(ambi, DECODING_METHOD_MAGLS);
    ambi_bin_setHRIRsPreProc(ambi, HRIR_PREPROC_OFF);
    ambi_bin_setEnableDiffuseMatching(ambi, 1);
    ambi_bin_init(ambi, static_cast<int>(params.sampleRateHz));

#ifdef MPEGCORE_DECMTX_H
    if (params.order < ShOrder::first || params.order > ShOrder::fourth) {
        SPDLOG_ERROR("Precomputed matrix missing for order: {}. Compile ShBinDecoder.cpp without "
                     "MPEGCORE_DECMTX_H or add a precomputed matrix to DECMTX.h.",
            params.order);
        return false;
    }
    ShBinDecoder_initCodec(ambi);
#else
    ambi_bin_initCodec(ambi);
#endif

    // If we provided ambi_bin a custom SOFA file and it is still using the default HRIRs then we
    // take it that loading the custom SOFA file failed.
    if (ambi_bin_getUseDefaultHRIRsflag(ambi)) {
        return params.hrirSofaFile.empty();
    }

    return true;
}

void BinAmbi::deinit(std::size_t clearBufferRounds, std::size_t clearBufferSize)
{
    AudioBufferArray inputs(shChannelCount(params.order), clearBufferSize);
    AudioBufferArray outputs(2, clearBufferSize);
    std::size_t clearRounds{ clearBufferRounds };
    while (clearRounds-- > 0) {
        ambi_bin_process(ambi, const_cast<float** const>(inputs.rawPointerArray()),
            outputs.rawPointerArray(), static_cast<int>(inputs.size()),
            static_cast<int>(outputs.size()), static_cast<int>(inputs[0].size()));
    }
}

void BinAmbi::destroy()
{
    ambi_bin_destroy(&ambi);
}

void BinAmbi::ShBinDecoder_initCodec(void* const hAmbi)
{
    ambi_bin_data* pData = (ambi_bin_data*)(hAmbi);
    ambi_bin_codecPars* pars = pData->pars;
    int i, j, nSH, order, band;

    pData->reinit_hrtfsFLAG = 0; // do not load the HR filter sofa file
    int decMtxFLAG = 0; // do not calculate the decMtx

    if (pData->codecStatus != CODEC_STATUS_NOT_INITIALISED)
        return; /* re-init not required, or already happening */
    while (pData->procStatus == PROC_STATUS_ONGOING) {
        /* re-init required, but we need to wait for the current processing loop to end */
        pData->codecStatus = CODEC_STATUS_INITIALISING; /* indicate that we want to init */
        SAF_SLEEP(10);
    }

    /* for progress bar */
    pData->codecStatus = CODEC_STATUS_INITIALISING;
    strcpy(pData->progressBarText, "Preparing HRIRs");
    pData->progressBar0_1 = 0.0f;

    /* (Re)Initialise afSTFT */
    order = pData->new_order;
    nSH = (order + 1) * (order + 1);
    if (pData->hSTFT == NULL)
        /* low delay mode enabled, hybrid-mode disabled for afSTFT */
        afSTFT_create(&(pData->hSTFT), nSH, NUM_EARS, HOP_SIZE, 1, 0, AFSTFT_BANDS_CH_TIME);
    else if (pData->nSH != nSH) { /* Or change the number of channels */
        afSTFT_channelChange(pData->hSTFT, nSH, NUM_EARS);
        afSTFT_clearBuffers(pData->hSTFT);
    }
    pData->nSH = nSH;

    /* get new decoder */
    strcpy(pData->progressBarText, "Computing Decoder");
    pData->progressBar0_1 = 0.95f;
    float_complex* decMtx;
    switch (order) {
    case 1:
        decMtx = (float_complex*)DECMTX_FIRST;
        break;
    case 2:
        decMtx = (float_complex*)DECMTX_SECOND;
        break;
    case 3:
        decMtx = (float_complex*)DECMTX_THIRD;
        break;
    case 4:
        decMtx = (float_complex*)DECMTX_FOURTH;
        break;
    default:
        SPDLOG_ERROR("Precomputed matrix missing for order: {}", order);
        strcpy(pData->progressBarText, "Decoder initialization failed");
        pData->codecStatus = CODEC_STATUS_NOT_INITIALISED;
        return;
    }

    /* replace current decoder */
    memset(pars->M_dec, 0, HYBRID_BANDS * NUM_EARS * MAX_NUM_SH_SIGNALS * sizeof(float_complex));
    for (band = 0; band < HYBRID_BANDS; band++)
        for (i = 0; i < NUM_EARS; i++)
            for (j = 0; j < nSH; j++)
                pars->M_dec[band][i][j] = decMtx[band * NUM_EARS * nSH + i * nSH + j];
    if (decMtxFLAG) {
        free(decMtx);
    }

    pData->order = order;

    /* done! */
    strcpy(pData->progressBarText, "Done!");
    pData->progressBar0_1 = 1.0f;
    pData->codecStatus = CODEC_STATUS_INITIALISED;
}

} // namespace mpegcore
