/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "ShLsDecoder.hpp"

#include <algorithm>
#include <cassert>
#include <unordered_map>

#include <ambi_dec.h>

#include "AmbiStore.hpp"
#include "saf.hpp"
#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/Hoa.hpp>
#include <mpegcore/spdlog.hpp>

namespace mpegcore {

namespace {

AmbiStore<LsAmbi, LsAmbiParams> store;

} // namespace

bool operator==(LsAmbiParams const& a, LsAmbiParams const& b)
{
    return a.order == b.order && a.sampleRateHz == b.sampleRateHz && a.cicpLayout == b.cicpLayout;
};

ShLsDecoder::ShLsDecoder(LsAmbiParams params)
{
    m_channelConfig = mpegcore::channelConfiguration(params.cicpLayout);
    auto normalChannelCount = channelCount(params.cicpLayout) - lfeCount(params.cicpLayout);
    m_decoderOutputs.resize(static_cast<std::size_t>(normalChannelCount));
    m_ambiHandle = store.reserveAmbi(params);
}

ShLsDecoder::ShLsDecoder(ShLsDecoder&& other) noexcept
    : m_decoderOutputs(std::move(other.m_decoderOutputs))
    , m_channelConfig(std::move(other.m_channelConfig))
{
    this->m_ambiHandle = other.m_ambiHandle;
    other.m_ambiHandle = {};
}

ShLsDecoder::~ShLsDecoder()
{
    if (m_ambiHandle.has_value()) {
        store.freeAmbi(m_ambiHandle.value());
        m_ambiHandle = {};
    }
}

void ShLsDecoder::decode(AudioBufferArray const& inputs, AudioBufferArray& outputs)
{
    assert(inputs[0].size() == outputs[0].size());
    assert(m_ambiHandle.has_value());
    std::size_t decoderOutputCount{ 0 };
    for (std::size_t i{ 0 }; i < m_channelConfig.size(); i++) {
        auto& channel = m_channelConfig.at(i);
        if (channel.type == cicp::ChannelType::normal) {
            m_decoderOutputs[decoderOutputCount++] = outputs.rawPointerArray()[i];
        } else {
            std::fill(outputs[i].begin(), outputs[i].end(), 0.F);
        }
    }
    if (!m_ambiHandle.has_value()) {
        SPDLOG_ERROR("no ambi handle, cannot decode.");
        return;
    }
    ambi_dec_process(store.getAmbi(m_ambiHandle.value()).ambi,
        const_cast<float** const>(inputs.rawPointerArray()), m_decoderOutputs.data(),
        static_cast<int>(inputs.size()), static_cast<int>(m_decoderOutputs.size()),
        static_cast<int>(inputs[0].size()));
}

bool LsAmbi::init()
{
    auto channelConfig = mpegcore::channelConfiguration(params.cicpLayout);
    ambi_dec_create(&ambi);
    ambi_dec_setNormType(ambi, toSafHoaNormalization(params.normalization));
    ambi_dec_init(ambi, static_cast<int>(params.sampleRateHz));
    auto const shOrders = toSafShOrders(params.order);
    ambi_dec_setMasterDecOrder(ambi, shOrders);
    ambi_dec_setDecOrderAllBands(ambi, shOrders);
    auto normalChannelCount = channelCount(params.cicpLayout) - lfeCount(params.cicpLayout);
    ambi_dec_setNumLoudspeakers(ambi, static_cast<std::int32_t>(normalChannelCount));
    ambi_dec_setDecMethod(ambi, 0, DECODING_METHOD_ALLRAD);
    ambi_dec_setDecMethod(ambi, 1, DECODING_METHOD_ALLRAD);

    int i{ 0 };
    for (auto&& channel : channelConfig) {
        if (channel.type == cicp::ChannelType::lfe) {
            continue;
        }
        ambi_dec_setLoudspeakerAzi_deg(ambi, i, channel.azimuthDeg);
        ambi_dec_setLoudspeakerElev_deg(ambi, i, channel.elevationDeg);
        i++;
    }
    ambi_dec_initCodec(ambi);
    return true;
}

void LsAmbi::deinit(std::size_t clearBufferRounds, std::size_t clearBufferSize)
{
    AudioBufferArray inputs(shChannelCount(params.order), clearBufferSize);
    AudioBufferArray outputs(channelCount(params.cicpLayout), clearBufferSize);
    std::size_t clearRounds{ clearBufferRounds };
    while (clearRounds-- > 0) {
        ambi_dec_process(ambi, const_cast<float** const>(inputs.rawPointerArray()),
            outputs.rawPointerArray(), static_cast<int>(inputs.size()),
            static_cast<int>(outputs.size()), static_cast<int>(inputs[0].size()));
    }
}

void LsAmbi::destroy()
{
    ambi_dec_destroy(&ambi);
}

} // namespace mpegcore
