/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "Vbap.hpp"

#define _USE_MATH_DEFINES

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstddef>
#include <cstdlib>
#include <math.h>
#include <set>
#include <utility>

#include "algorithm.hpp"

namespace mpegcore {
namespace {

using Vector3 = Vbap::Vector3;
using Matrix3x3 = Vbap::Matrix3x3;

double const EVS_PI{ M_PI };
float const MIN_ELEVATION_DEG{ -90.F };
float const MAX_ELEVATION_DEG{ 90.F };
float const FULL_CIRCLE_DEG{ 360.F };
float const NORMALIZATION_DEFAULT_SUM_INIT{ 1e-12F };

// 128 is maximum num_speaker_nodes number. This relates to memory optimization
// and maximum of triplets:
//   - triplet indices are std::uint8_t (see Vbap data structures)
//     --> max triplets is 256
//   - internalSpeakerNodeCount <= speakerNodeCount + 2 (potential virtual
//     node channels, bottom and top)
//   - maxTripletCount = 256 = (internalSpeakerNodeCount - 2) * 2
//     = (speakerNodeCount) * 2
//     --> max_num_ls = 128
std::int16_t const MAX_NUM_SPEAKER_NODES{ 128 };
std::int16_t const MAX_NUM_CONNECTIONS{ MAX_NUM_SPEAKER_NODES * 3 };

// The fairly large epsilon is for detecting planes etc and accounts for
// rounding issues.
float const EPSILON{ 0.001F };

std::int16_t const MAX_PLANE_COUNT{ 50 };
float const MAX_HORIZONTAL_GAP_FOR_PLANE_DETECTION{ 140.0F };

// If a speaker node is found
// - above NO_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG, no virtual node is used
// - above DISTRIBUTE_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG, energy-spreading
//   virtual node is used
// - not above DISTRIBUTE_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG, energy-omitting
//   virtual node is used Same applies for both elevations and inclinations,
//   i.e., the two half-spheres.
float const NO_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG{ 45.0F };
float const DISTRIBUTE_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG{ 20.0F };

struct Direction
{
    std::int16_t aziDeg{ 0 };
    std::int16_t eleDeg{ 0 };
};

struct TripletNodeIndices
{
    std::uint16_t nodeAIndex{ 0 };
    std::uint16_t nodeBIndex{ 0 };
    std::uint16_t nodeCIndex{ 0 };
};

struct Connection
{
    std::uint16_t nodeAIndex{ 0 };
    std::uint16_t nodeBIndex{ 0 };
};

using Connections = std::vector<std::optional<Connection>>;

enum class ConnectionClass : std::uint8_t
{
    regular,
    elevatedPlaneThinTriangle,
    speakerNodeBehind
};

// Defines a single speaker node
struct SpeakerNode
{
    Vector3 unitVec;
    float aziDeg;
    float eleDeg;
    Vbap::SpeakerNodeGroup group;
};

// Return a copy of `vec` with negative elements replaced with 0.0f.
Vector3 negativesZeroed(Vector3&& vec)
{
    Vector3 out;
    for (auto i{ vec.size() }; i-- > 0;) {
        auto const v = vec[i];
        out[i] = v >= 0.F ? v : 0.F;
    }

    return out;
};

// Return normalized `vec`. Default `sumInit` prevents division by zero for
// cases where all elements of `vec` can be zeros.
Vector3 normalized(Vector3 vec, float sumInit)
{
    Vector3 out;
    float const squareSum = accumulate(vec, sumInit, [](auto sum, auto v) { return sum + v * v; });
    float const normValue{ 1.0F / sqrt(squareSum) };

    for (auto i{ vec.size() }; i-- > 0;) {
        out[i] = vec[i] * normValue;
    }

    return out;
};

// Return {aziIndex, eleIndex} pair based on given step counts.
std::pair<std::size_t, std::size_t> angleIndicesFromStepCounts(
    float aziStepCount, float eleStepCount) noexcept
{
    auto aziIndex = static_cast<std::int64_t>(aziStepCount);
    float const HORITONTAL_LEVEL_DEG{ 90.F };
    auto eleIndex =
        static_cast<std::int64_t>(eleStepCount + HORITONTAL_LEVEL_DEG / Vbap::ELE_RESOLUTION_DEG);

    std::int64_t const FULL_CIRCLE_DEG{ 360 };
    while (aziIndex < 0) {
        aziIndex += FULL_CIRCLE_DEG / Vbap::AZI_RESOLUTION_DEG;
    }
    while (aziIndex >= FULL_CIRCLE_DEG / Vbap::AZI_RESOLUTION_DEG) {
        aziIndex -= FULL_CIRCLE_DEG / Vbap::AZI_RESOLUTION_DEG;
    }
    if (eleIndex < 0) {
        eleIndex = 0;
    }
    std::int64_t const MAX_ELEVATION_INDEX{ 180 / Vbap::ELE_RESOLUTION_DEG };
    if (eleIndex > static_cast<std::int64_t>(MAX_ELEVATION_INDEX)) {
        eleIndex = MAX_ELEVATION_INDEX;
    }

    return { aziIndex, eleIndex };
}

// Return {aziDeg, eleDeg} between ranges -180,180 and -90,90.
std::pair<float, float> aziEleToExpectedRange(float aziDeg, float eleDeg)
{
    eleDeg = std::fmod(eleDeg, 360.F);
    if (eleDeg <= -180.F) {
        eleDeg += 360.F;
    } else if (eleDeg > 180.F) {
        eleDeg -= 360.F;
    }
    if (eleDeg < -90) {
        eleDeg = -180.F - eleDeg;
        aziDeg += 180.F;
    } else if (eleDeg > 90) {
        eleDeg = 180.F - eleDeg;
        aziDeg += 180.F;
    }
    aziDeg = std::fmod(aziDeg, 360.F);
    if (aziDeg < -180.F) {
        aziDeg += 360.F;
    } else if (aziDeg > 180.F) {
        aziDeg -= 360.F;
    }
    return { aziDeg, eleDeg };
}

// Return `deg` in radians.
template <typename T> float deg2rad(T deg)
{
    double const DEG_TO_RAD_FACTOR{ EVS_PI / 180.0 };
    return static_cast<float>(static_cast<double>(deg) * DEG_TO_RAD_FACTOR);
}

// Return dot product of `vec1` and `vec2`
float dotProduct(Vector3 const& vec1, Vector3 const& vec2) noexcept
{
    float p{ 0.0F };

    if (vec1.size() != vec2.size()) {
        return p;
    }

    for (std::size_t i{ 0 }, end{ vec1.size() }; i < end; ++i) {
        p += vec1[i] * vec2[i];
    }

    return p;
}

// Return vector cross product of `vec1` and `vec2`
template <typename T> auto crossProduct(T const& vec1, T const& vec2) noexcept
{
    Vector3 product;
    product[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
    product[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
    product[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

    return product;
}

// Return the result of multiplying `vector` with `matrix`.
Vector3 vectorMatrixMultiply3x3(Vector3 const& vector, Matrix3x3 const& matrix) noexcept
{
    Vector3 result;
    for (std::size_t k{ 0 }; k < 3; k++) {
        result[k] = 0;
        for (std::size_t l{ 0 }; l < 3; l++) {
            result[k] += vector[l] * matrix[l][k];
        }
    }

    return result;
}

// Return the index of the triplet that best matches `aziRad` and `eleRad`.
// Index is to the vector `triplets`.
std::uint8_t determineBestTriplet(
    float aziRad, float eleRad, std::vector<Vbap::VirtualSurfaceTriplet> const& triplets) noexcept
{
    Vector3 panningUnitVec{
        std::cos(aziRad) * std::cos(eleRad),
        std::sin(aziRad) * std::cos(eleRad),
        std::sin(eleRad),
    };

    std::optional<float> bestMinGain;
    std::uint8_t bestTriplet{ 0 };
    for (std::size_t tr{ 0 }, end{ triplets.size() }; tr < end; ++tr) {
        auto const unnormalizedGains =
            vectorMatrixMultiply3x3(panningUnitVec, triplets[tr].inverseMatrix);

        std::optional<float> minGain;
        for (auto gain : unnormalizedGains) {
            if ((!minGain || gain < minGain) && !std::isnan(gain)) {
                minGain = gain;
            }
        }

        if (minGain > bestMinGain) {
            bestMinGain = *minGain;
            bestTriplet = static_cast<std::uint8_t>(tr);
        }
    }

    return bestTriplet;
}

// Determine how the virtual node's (`virtualSpeakerNodeIndex`) gains are
// distributed to real nodes.
// Gains are output to `virtualNodeDivisionGains`.
void determineVirtualSpeakerNodeDivisionGains(std::optional<std::uint16_t> virtualSpeakerNodeIndex,
    std::vector<float>& virtualNodeDivisionGains, Connections const& connections,
    Vbap::VirtualSpeakerNodeType type) noexcept
{
    if (type != Vbap::VirtualSpeakerNodeType::distributeEnergy) {
        return;
    }

    // When node type is DISTRIBUTE_ENERGY, the gains of the virtual node
    // are distributed to all neighboring real speaker nodes. An
    // amplitude-division instead of energy division is utilized just in
    // case to avoid excessive emphasis on the coherent distributed sound.

    for (auto&& connection : connections) {
        if (!connection) {
            continue;
        }

        if (connection->nodeAIndex == virtualSpeakerNodeIndex) {
            virtualNodeDivisionGains[connection->nodeBIndex] = 1.0F;
        } else if (connection->nodeBIndex == virtualSpeakerNodeIndex) {
            virtualNodeDivisionGains[connection->nodeAIndex] = 1.0F;
        }
    }

    auto const sum = accumulate(virtualNodeDivisionGains, 0.F);

    for (auto ch{ virtualNodeDivisionGains.size() }; ch-- > 0;) {
        virtualNodeDivisionGains[ch] /= sum;
        virtualNodeDivisionGains[ch] = std::pow(virtualNodeDivisionGains[ch], 0.8f);
    }
}

// Creates the internal speaker node data based on the speakers' azimuth and
// elevation data.
//
// Returns the created vector of `SpeakerNodes`.
//
// Exceptions
// - may throw on memory allocation
std::vector<SpeakerNode> createInternalSpeakerNodes(
    std::vector<Vbap::SpeakerNodeDirection> const& internalSpeakerNodeDirections)
{
    auto const internalSpeakerNodeCount = internalSpeakerNodeDirections.size();
    std::vector<SpeakerNode> speakerNodes(internalSpeakerNodeCount);

    for (std::size_t i{ 0 }, end{ internalSpeakerNodeCount }; i < end; ++i) {
        auto& node = speakerNodes[i];
        node.aziDeg = internalSpeakerNodeDirections[i].azimuthDeg;
        float eleRad{ 0.F };
        auto const eleDeg = internalSpeakerNodeDirections[i].elevationDeg;

        float const HORIZONTAL_BAND_THRESHOLD_DEG{ 5.0F };
        if ((eleDeg >= -HORIZONTAL_BAND_THRESHOLD_DEG)
            && (eleDeg <= HORIZONTAL_BAND_THRESHOLD_DEG)) {
            node.eleDeg = 0.0F;
            node.group = Vbap::SpeakerNodeGroup::horizontal;
        } else {
            eleRad = deg2rad(eleDeg);
            node.eleDeg = eleDeg;
            node.group =
                eleRad < 0 ? Vbap::SpeakerNodeGroup::bottomHalf : Vbap::SpeakerNodeGroup::topHalf;
        }

        auto aziRad = deg2rad(node.aziDeg);
        node.unitVec = {
            std::cos(aziRad) * std::cos(eleRad),
            std::sin(aziRad) * std::cos(eleRad),
            std::sin(eleRad),
        };
    }

    return speakerNodes;
}

// Return the inverse of `matrix`.
auto matrixInverse3x3(Matrix3x3 const& matrix) noexcept
{
    Matrix3x3 inverse;
    auto const crossVec = crossProduct(matrix[1], matrix[2]);
    auto const determinant = dotProduct(matrix[0], crossVec);

    auto inverseColumn = [&determinant, &inverse](std::size_t column, auto const& crossVec) {
        for (std::size_t k{ inverse.size() }; k-- > 0;) {
            inverse[k][column] = crossVec.at(k) / determinant;
        }
    };

    inverseColumn(0, crossVec);
    inverseColumn(1, crossProduct(matrix[2], matrix[0]));
    inverseColumn(2, crossProduct(matrix[0], matrix[1]));

    return inverse;
}

// Update `nodeAConnectsTo` based on how `nodeAIndex` can be directly reached
// using `connections` (1st degree).
void updateNodeAConnections(std::uint16_t nodeAIndex, Connections const& connections,
    std::array<bool, MAX_NUM_CONNECTIONS>& nodeAConnectsTo) noexcept
{
    fill(nodeAConnectsTo, false);

    for (auto&& connection : connections) {
        if (connection) {
            if (connection->nodeAIndex == nodeAIndex) {
                nodeAConnectsTo.at(connection->nodeBIndex) = true;
            }
            if (connection->nodeBIndex == nodeAIndex) {
                nodeAConnectsTo.at(connection->nodeAIndex) = true;
            }
        }
    }
}

// Find next possible speaker node triplet.
//
// Return the triplet indices if a possible triplet was found, or null option
// otherwise.
std::optional<TripletNodeIndices> getNextPossibleTriplet(
    TripletNodeIndices const& tripletNodeIndices, std::size_t speakerNodeCount,
    Connections const& connections, std::array<bool, MAX_NUM_CONNECTIONS>& nodeAConnectsTo) noexcept
{
    bool needConnectionSearchTableUpdate{ false };
    auto [nodeAIndex, nodeBIndex, nodeCIndex] = tripletNodeIndices;

    while (true) {
        // First round configuration, simply sets data appropriately for the
        // following loops.
        if (nodeCIndex == 0) {
            nodeBIndex = 1;
            nodeCIndex = 1;
            needConnectionSearchTableUpdate = true;
        }

        // Increment to next nodeAIndex-nodeBIndex-nodeCIndex pair where
        // nodeCIndex > nodeBIndex > nodeAIndex
        if (nodeCIndex++ == speakerNodeCount - 1) {
            if (nodeBIndex++ == speakerNodeCount - 2) {
                if (nodeAIndex++ == speakerNodeCount - 3) {
                    return {};
                }
                nodeBIndex = static_cast<std::uint16_t>(nodeAIndex + 1);
                // nodeAIndex incremented: Update possible connections table
                needConnectionSearchTableUpdate = true;
            }
            nodeCIndex = static_cast<std::uint16_t>(nodeBIndex + 1);
        }

        if (needConnectionSearchTableUpdate) {
            updateNodeAConnections(nodeAIndex, connections, nodeAConnectsTo);
            needConnectionSearchTableUpdate = false;
        }

        // Continue searching unless there is a connection from nodeA to
        // both nodeB and nodeC
        if (!(nodeAConnectsTo.at(nodeBIndex) && nodeAConnectsTo.at(nodeCIndex))) {
            continue;
        }

        // At this stage, nodeA is connected to nodeB and nodeC.
        // Find if there is the third connection, between nodeB and nodeC,
        // forming the triplet.
        for (auto&& connection : connections) {
            if (((connection->nodeAIndex == nodeBIndex) && (connection->nodeBIndex == nodeCIndex))
                || ((connection->nodeBIndex == nodeBIndex)
                    && (connection->nodeAIndex == nodeCIndex))) {
                return TripletNodeIndices{ nodeAIndex, nodeBIndex, nodeCIndex };
            }
        }
        // If third connection not found, keep searching
    }

    // Should not end here
    return {};
}

// Check through all speaker nodes that none of them are within the triplet.
// Node within the triplet is identified by that all three panning gains are
// positive.
//
// Return true if the triplet identified by `tripletNodeIndices` contains a
// speaker node (from `speakerNodes`), or false otherwise.
bool tripletContainsSpeakerNode(std::vector<SpeakerNode> const& speakerNodes,
    TripletNodeIndices const& tripletNodeIndices, Matrix3x3 const& inverseMatrix) noexcept
{
    bool foundSpeakerNodeInsideTriplet{ false };
    auto const [nodeAIndex, nodeBIndex, nodeCIndex] = tripletNodeIndices;

    for (std::size_t nodeIndex{ speakerNodes.size() }; nodeIndex-- > 0;) {
        if (nodeIndex == nodeAIndex || nodeIndex == nodeBIndex || nodeIndex == nodeCIndex) {
            continue;
        }
        auto const unnormalizedGains =
            vectorMatrixMultiply3x3(speakerNodes[nodeIndex].unitVec, inverseMatrix);

        // Epsilon-condition is for some small rounding issues.
        if (unnormalizedGains[0] > EPSILON && unnormalizedGains[1] > EPSILON
            && unnormalizedGains[2] > EPSILON) {
            foundSpeakerNodeInsideTriplet = true;
            break;
        }
    }

    return foundSpeakerNodeInsideTriplet;
}

// Determine virtual surface triples that are used for panning based on
// `speakerNodes` and `connections`.
//
// Return the triplets.
//
// Exceptions
// - may throw on memory allocation
std::vector<Vbap::VirtualSurfaceTriplet> determineVirtualSurfaceTriplets(
    std::vector<SpeakerNode> const& speakerNodes, Connections const& connections)
{
    auto const speakerNodeCount = speakerNodes.size();
    auto const maxTripletCount = (speakerNodeCount - 2) * 2;
    std::vector<Vbap::VirtualSurfaceTriplet> triplets;
    triplets.reserve(maxTripletCount);
    TripletNodeIndices tripletNodeIndices;
    std::array<bool, MAX_NUM_CONNECTIONS> nodeAConnectsTo = {};

    auto nextTripletMaybe = [&connections, &nodeAConnectsTo, &tripletNodeIndices,
                                &speakerNodeCount]() {
        return getNextPossibleTriplet(
            tripletNodeIndices, speakerNodeCount, connections, nodeAConnectsTo);
    };

    for (auto tripletMaybe{ nextTripletMaybe() }; tripletMaybe.has_value();
         tripletMaybe = nextTripletMaybe()) {
        // Determine inverse matrix for VBAP formulation
        tripletNodeIndices = tripletMaybe.value();
        auto const [nodeAIndex, nodeBIndex, nodeCIndex] = tripletNodeIndices;
        auto const inverseMatrix = matrixInverse3x3({
            speakerNodes[nodeAIndex].unitVec,
            speakerNodes[nodeBIndex].unitVec,
            speakerNodes[nodeCIndex].unitVec,
        });
        auto const foundSpeakerNodeInsideTriplet =
            tripletContainsSpeakerNode(speakerNodes, tripletNodeIndices, inverseMatrix);

        // No speaker node inside triplet -> appropriate triplet found, save
        // data.
        if (!foundSpeakerNodeInsideTriplet) {
            Vbap::VirtualSurfaceTriplet triplet;
            triplet.speakerNodeIndices = {
                static_cast<std::uint8_t>(nodeAIndex),
                static_cast<std::uint8_t>(nodeBIndex),
                static_cast<std::uint8_t>(nodeCIndex),
            };
            triplet.inverseMatrix = inverseMatrix;
            triplets.push_back(triplet);
        }
    }

    return triplets;
}

// Return gap to the next speaker node at the same plane (at elevation
// `planeEleDeg`).
std::optional<float> maxGapBetweenSpeakersOnPlane(
    std::vector<SpeakerNode> const& speakerNodes, float planeEleDeg) noexcept
{
    std::optional<float> maxGap;
    for (auto&& nodeA : speakerNodes) {
        if (fabs(nodeA.eleDeg - planeEleDeg) >= EPSILON) {
            continue;
        }

        std::optional<float> gapToNextNode;
        for (auto&& nodeB : speakerNodes) {
            if (&nodeB == &nodeA || fabs(nodeB.eleDeg - planeEleDeg) >= EPSILON) {
                continue;
            }
            float gap = nodeB.aziDeg - nodeA.aziDeg;
            while (gap < 0) {
                gap += FULL_CIRCLE_DEG;
            }
            gapToNextNode = gapToNextNode ? std::min(*gapToNextNode, gap) : gap;
        }

        maxGap = std::max(maxGap, gapToNextNode);
    }

    return maxGap;
}

// Return elevations of non-crossing planes.
//
// Exceptions
// - may throw on memory allocation
std::vector<float> determineNonCrossingPlanes(std::vector<SpeakerNode> const& speakerNodes)
{
    std::vector<float> nonCrossingPlaneElesDeg;
    nonCrossingPlaneElesDeg.reserve(MAX_PLANE_COUNT);

    // For each plane, check if a non-crossing plane should be determined
    std::optional<float> planeEleDeg;
    while (planeEleDeg < MAX_ELEVATION_DEG) {

        std::optional<float> nextPlaneEleDeg;
        // Find next node elevation that is not in horizontal plane
        for (auto&& node : speakerNodes) {
            if ((node.group != Vbap::SpeakerNodeGroup::horizontal)
                && (!planeEleDeg || node.eleDeg > *planeEleDeg + EPSILON)
                && (!nextPlaneEleDeg || node.eleDeg < *nextPlaneEleDeg - EPSILON)) {
                nextPlaneEleDeg = node.eleDeg;
            }
        }
        planeEleDeg = nextPlaneEleDeg;

        if (!planeEleDeg) {
            // When no next node elevation found, break loop
            break;
        }

        auto maxGap = maxGapBetweenSpeakersOnPlane(speakerNodes, *planeEleDeg);

        // If maximum gap is small enough, then a non-crossing plane is
        // detected
        if (maxGap.has_value()
            && maxGap.value() < (MAX_HORIZONTAL_GAP_FOR_PLANE_DETECTION + EPSILON)) {
            nonCrossingPlaneElesDeg.push_back(*planeEleDeg);

            if (nonCrossingPlaneElesDeg.size() == nonCrossingPlaneElesDeg.capacity()) {
                // Does not happen with any real speaker node configuration. Triangulation succeeds
                // even if number of non crossing planes are limited. So we can stop here.
                break;
            }
        }
    }

    return nonCrossingPlaneElesDeg;
}

// Get next half-sphere connection such that the connected nodes are either both
// in `group` or one of the is in `group` and the other is horizontal.
//
// Return the connection if one was found, or null option otherwise.
std::optional<Connection> getNextPossibleHalfSphereConnection(Connection const& connection,
    std::vector<SpeakerNode> const& speakerNodes, Vbap::SpeakerNodeGroup group) noexcept
{
    auto [nodeAIndex, nodeBIndex] = connection;

    while (true) {
        // Increment to next nodeAIndex-nodeBIndex pair where nodeBIndex >
        // nodeAIndex
        auto const speakerNodeCount = speakerNodes.size();
        if (nodeBIndex++ == speakerNodeCount - 1) {
            if (nodeAIndex++ == speakerNodeCount - 2) {
                // No new potential connections to test
                return {};
            }
            nodeBIndex = static_cast<std::uint16_t>(nodeAIndex + 1);
        }

        // At least nodeAIndex or nodeBIndex must be at the same group
        // (BOTTOM or TOP half), and the other must be at the same group or
        // at the horizontal plane. Horizontal-horizontal connections are
        // determined at another function.
        auto canConnect = [&group](auto node1, auto node2) {
            return node1.group == group
                && (node2.group == group || node2.group == Vbap::SpeakerNodeGroup::horizontal);
        };
        auto const& nodeA = speakerNodes[nodeAIndex];
        auto const& nodeB = speakerNodes[nodeBIndex];

        if (canConnect(nodeA, nodeB) || canConnect(nodeB, nodeA)) {
            return Connection{ nodeAIndex, nodeBIndex };
        }
    }

    /* Should not end here */
    return {};
}

// Return the connection class of `connection`.
ConnectionClass determineConnectionClass(std::vector<SpeakerNode> const& speakerNodes,
    Vbap::SpeakerNodeGroup group, Connection const& connection) noexcept
{
    // This loop checks the nodeAIndex-nodeBIndex connection with respect to
    // all loudspeakers:
    // - in case there is a node behind or nearly behind the connection
    //   line. These connections need to be discarded.
    // - in case there is a node that is closer to the connection line than
    //   1/5 of the connection length AND at the same horizontal plane. These
    //   connections need to be weighted with a penalty (a special case, for
    //   example avoiding elevated L,R,C triplet)
    auto const [nodeAIndex, nodeBIndex] = connection;
    for (std::size_t nodeIndex{ 0 }, end{ speakerNodes.size() }; nodeIndex < end; ++nodeIndex) {

        // Select speaker_node only within TOP or BOTTOM sphere half, not
        // being part of nodeAIndex-nodeBIndex pair
        if ((group == speakerNodes[nodeIndex].group) && (nodeIndex != nodeAIndex)
            && (nodeIndex != nodeBIndex)) {
            // The following lines formulate the point on the
            // nodeAIndex-nodeBIndex-connection that is nearest to the
            // origo-node-line
            auto const& nodeA = speakerNodes[nodeAIndex];
            auto const& nodeB = speakerNodes[nodeBIndex];
            auto const& p1 = nodeA.unitVec;
            Vector3 v1;
            for (std::size_t k{ 0 }; k < 3; ++k) {
                v1[k] = nodeB.unitVec[k] - nodeA.unitVec[k];
            }
            auto const& v2 = speakerNodes[nodeIndex].unitVec;
            auto const v1v1 = dotProduct(v1, v1);
            auto const v1v2 = dotProduct(v1, v2);
            auto const v2v2 = 1.0F; // dotProduct(v2, v2) is always 1.
            auto const v1p1 = dotProduct(v1, p1);
            auto const v2p1 = dotProduct(v2, p1);
            auto const determinant = (v1v1) * (-v2v2) + (v1v2 * v1v2);
            // Norm distance = distance parameter on line
            // nodeAIndex-nodeBIndex, determines point that is nearest to
            // origo-node line. Distance 0 means nodeAIndex and distance 1
            // means nodeBIndex. Can be outside this region as well.
            auto const normDistanceOnV1 =
                (1.0F / determinant) * ((v2v2) * (v1p1) + (-v1v2) * (v2p1));

            // Continue only if the nearest point is between nodeAIndex and
            // nodeBIndex
            if (normDistanceOnV1 > 0.0F && normDistanceOnV1 < 1.0F) {
                // Formulate vTarget, that is an unit vector that goes
                // through the determined point on nodeAIndex - nodeBIndex
                // connection
                Vector3 vTarget;
                Vector3 vecDiff;
                for (std::size_t k{ 0 }; k < 3; k++) {
                    vTarget[k] = p1.at(k) + normDistanceOnV1 * v1[k];
                    vecDiff[k] = vTarget[k] - v2.at(k);
                }
                vTarget = normalized(vTarget, 0.F);

                // A check if the angle between vTarget and
                // speakerNodes[nodeIndex].unit_vec is less than 1 degree.
                // Essentially reveals if there is a speaker node too close
                // "behind" the connection. Such connections should be
                // rejected.
                float const BEHIND_THRESHOLD_RAD{ 0.9998F };
                if (dotProduct(vTarget, v2) > BEHIND_THRESHOLD_RAD) {
                    return ConnectionClass::speakerNodeBehind;
                }

                // A special case, mainly accounting for ELEVATED L,R,C
                // speaker nodes. A triplet between these nodes is not
                // desired if there is a top node, a penalty is implemented
                // to take care of this.
                float const VECDIFF_TO_V1_FACTOR{ 5.0F };
                if (sqrt(v1v1) > VECDIFF_TO_V1_FACTOR * sqrt(dotProduct(vecDiff, vecDiff))) {
                    if (fabs(nodeB.unitVec[2] - nodeA.unitVec[2]) < EPSILON) {
                        return ConnectionClass::elevatedPlaneThinTriangle;
                    }
                }
            }
        }
    }

    return ConnectionClass::regular;
}

// Output connections between `speakerNodes` in the horizontal plane to
// `connections`.
void formulateHorizontalConnections(
    std::vector<SpeakerNode> const& speakerNodes, Connections& connections) noexcept
{
    auto speakerNodeCount = static_cast<std::uint16_t>(speakerNodes.size());

    for (std::uint16_t aIndex{ 0 }; aIndex < speakerNodeCount; ++aIndex) {
        // Find next horizontal speaker node
        auto const& nodeA = speakerNodes[aIndex];
        if (nodeA.group == Vbap::SpeakerNodeGroup::horizontal) {
            std::uint16_t nextIndex{ 0 };
            std::optional<float> minArcDiff;
            bool foundConnection{ false };

            for (std::uint16_t bIndex{ 0 }; bIndex < speakerNodeCount; ++bIndex) {
                auto const& nodeB = speakerNodes[bIndex];

                if ((aIndex != bIndex) && (nodeB.group == Vbap::SpeakerNodeGroup::horizontal)) {
                    float arcDiff = nodeB.aziDeg - nodeA.aziDeg;

                    while (arcDiff < 0.0F) {
                        arcDiff += FULL_CIRCLE_DEG;
                    }

                    if (!minArcDiff || arcDiff < minArcDiff) {
                        minArcDiff = arcDiff;
                        nextIndex = bIndex;
                        foundConnection = true;
                    }
                }
            }

            if (!foundConnection) {
                continue;
            }

            // Find free slot and store the connection
            auto it = findIf(connections, [](auto&& connection) { return !connection; });
            if (it != connections.end()) {
                *it = { aIndex, nextIndex };
            }
        }
    }
}

// Return `true` if the connection between elevations `ele1Deg` and `ele2Deg`
// crosses a non-crossing plane, and `false` otherwise.
bool crossesNonCrossingPlane(
    float ele1Deg, float ele2Deg, std::vector<float> const& nonCrossingPlaneElesDeg) noexcept
{
    float const SEPARATION_DEG{ 1.0F };
    for (auto planeEleDeg : nonCrossingPlaneElesDeg) {
        if ((ele1Deg > planeEleDeg + SEPARATION_DEG) && (ele2Deg < planeEleDeg - SEPARATION_DEG)) {
            return true;
        }

        if ((ele2Deg > planeEleDeg + SEPARATION_DEG) && (ele1Deg < planeEleDeg - SEPARATION_DEG)) {
            return true;
        }
    }

    return false;
}

// Study if the crossing is between the connection arc, if so, a crossing is
// detected.
// Connections, i.e., node-pairs determine a plane. The crossing can be
// determined by studying the intersection of these planes.
//
// The function takes in quite a lot of parameters, but its main purpose as it
// is is to pull some of the complexity out of `formulateHalfSphereConnections`
// function. More extensive reworking of that function would be needed to make
// the interface to this helper better.
//
// Return `true` if a crossing is detercted, `false` otherwise.
bool detectCrossing(std::vector<SpeakerNode> const& speakerNodes, Connection connection,
    Connection oldConnection, Vector3 const& connectionCross, float connectionArc,
    Vector3 const& newConnectionCross, float newArc) noexcept
{
    auto planeCrossingVec = normalized(
        crossProduct(connectionCross, newConnectionCross), NORMALIZATION_DEFAULT_SUM_INIT);

    // If the plane intersection is between both connections, then the two
    // connections cross. Study first if the crossing is between arc
    // nodeAIndex-nodeBIndex
    auto angleCmp =
        std::acos(dotProduct(planeCrossingVec, speakerNodes[connection.nodeAIndex].unitVec));
    angleCmp +=
        std::acos(dotProduct(planeCrossingVec, speakerNodes[connection.nodeBIndex].unitVec));

    float const MAX_ANGLE_RAD{ 0.01F };
    auto const TWO_PI = static_cast<float>(2.0F * EVS_PI);
    if (fabs(newArc - angleCmp) < MAX_ANGLE_RAD) {
        // Crossing is within the arc. Proceed to next check.
    } else if (fabs(newArc - (TWO_PI - angleCmp)) < MAX_ANGLE_RAD) {
        // In this case, the plane crossing vector is inverted.  The inverse
        // is another plane-crossing vector, and detected to be between
        // nodeAIndex-nodeBIndex connection.
        for (auto& e : planeCrossingVec) {
            e *= -1.0F;
        }
    } else {
        // Crossing is not within the arc, reject it.
        return false;
    }

    angleCmp =
        std::acos(dotProduct(planeCrossingVec, speakerNodes[oldConnection.nodeAIndex].unitVec));
    angleCmp +=
        std::acos(dotProduct(planeCrossingVec, speakerNodes[oldConnection.nodeBIndex].unitVec));

    return fabs(connectionArc - angleCmp) < MAX_ANGLE_RAD;
}

// Output connections between `speakerNodes` belonging to `group` into
// `connections`.
//
// Exceptions
// - may throw on memory allocation
void formulateHalfSphereConnections(std::vector<SpeakerNode> const& speakerNodes,
    Vbap::SpeakerNodeGroup group, Connections& connections,
    std::vector<float> const& nonCrossingPlaneElesDeg)
{
    auto const maxConnectionCount = connections.size();
    std::vector<float> connectionArc(maxConnectionCount, 0.F);
    std::vector<float> connectionArcWeighted(maxConnectionCount, 0.F);
    std::vector<Vector3> connectionCross(maxConnectionCount, { 0, 0, 0 });

    Connection connection;
    auto nextConnectionMaybe = [&connection, &speakerNodes, &group]() {
        return getNextPossibleHalfSphereConnection(connection, speakerNodes, group);
    };

    // The following loop goes through all reasonable nodeAIndex - nodeBIndex
    // pairs for the half-sphere

    for (auto connectionMaybe{ nextConnectionMaybe() }; connectionMaybe.has_value();
         connectionMaybe = nextConnectionMaybe()) {
        connection = connectionMaybe.value();
        auto const connectionClass = determineConnectionClass(speakerNodes, group, connection);

        if (connectionClass != ConnectionClass::speakerNodeBehind) {
            auto const [nodeAIndex, nodeBIndex] = connection;
            auto const& nodeA = speakerNodes[nodeAIndex];
            auto const& nodeB = speakerNodes[nodeBIndex];
            float newArc = std::acos(dotProduct(nodeA.unitVec, nodeB.unitVec));
            float newArcWeighted = newArc;
            float const ARC_WEIGHT_PENALTY_FACTOR{ 2.F };

            if (connectionClass == ConnectionClass::elevatedPlaneThinTriangle) {
                newArcWeighted *= ARC_WEIGHT_PENALTY_FACTOR;
            }

            // If the connection passes a pre-determined plane of speaker nodes,
            // then add further penalty
            if (crossesNonCrossingPlane(nodeA.eleDeg, nodeB.eleDeg, nonCrossingPlaneElesDeg)) {
                newArcWeighted *= ARC_WEIGHT_PENALTY_FACTOR;
            }

            // Determine if new connection between nodeAIndex and nodeBIndex is
            // valid
            auto const newConnectionCross = crossProduct(nodeA.unitVec, nodeB.unitVec);
            auto isNewConnectionValid{ true };
            for (std::size_t connectionIndex{ 0 }; connectionIndex < maxConnectionCount;
                 ++connectionIndex) {
                // Compare against all previously defined connections
                auto& oldConnection = connections[connectionIndex];
                if (!oldConnection) {
                    continue;
                }
                // The connections are compared only if they don't involve same
                // speaker nodes
                if ((oldConnection->nodeAIndex != nodeAIndex)
                    && (oldConnection->nodeBIndex != nodeAIndex)
                    && (oldConnection->nodeBIndex != nodeBIndex)) {
                    auto const isCrossing = detectCrossing(speakerNodes, connection, *oldConnection,
                        connectionCross[connectionIndex], connectionArc[connectionIndex],
                        newConnectionCross, newArc);

                    if (isCrossing) {
                        if (newArcWeighted <= connectionArcWeighted[connectionIndex]) {
                            // Prior connection not valid, remove from data,
                            // and keep studying prior connections compared
                            // to the new connection.
                            oldConnection = std::nullopt;
                        } else {
                            // New connection not valid, stop studying it
                            isNewConnectionValid = false;
                            break;
                        }
                    }
                }
            }

            if (isNewConnectionValid) {
                for (std::size_t c{ 0 }; c < maxConnectionCount; ++c) {
                    // Find first free slot and store data
                    if (!connections[c]) {
                        connections[c] = { nodeAIndex, nodeBIndex };
                        connectionArc[c] = newArc;
                        connectionArcWeighted[c] = newArcWeighted;
                        connectionCross[c] = newConnectionCross;
                        break;
                    }
                }
            }
        }
    }
}

// Return all valid connections between all `speakerNodes`.
//
// Exceptions
// - may throw on memory allocation
Connections determineConnections(std::vector<SpeakerNode> const& speakerNodes)
{
    Connections connections;

    auto const maxConnectionCount = (speakerNodes.size() - 2) * 3;
    assert(maxConnectionCount <= MAX_NUM_CONNECTIONS);
    connections.resize(maxConnectionCount, std::nullopt);

    // Determine some prominent elevated planes that are favoured in making
    // node-node connections.
    auto nonCrossingPlaneElesDeg = determineNonCrossingPlanes(speakerNodes);

    // The node-node connections are determined in three stages: top, horizontal
    // and bottom.
    formulateHalfSphereConnections(
        speakerNodes, Vbap::SpeakerNodeGroup::topHalf, connections, nonCrossingPlaneElesDeg);

    formulateHorizontalConnections(speakerNodes, connections);

    formulateHalfSphereConnections(
        speakerNodes, Vbap::SpeakerNodeGroup::bottomHalf, connections, nonCrossingPlaneElesDeg);

    return connections;
}

bool haveDuplicates(std::vector<Vbap::SpeakerNodeDirection> directions)
{
    struct Cmp
    {
        bool operator()(
            Vbap::SpeakerNodeDirection const& d1, Vbap::SpeakerNodeDirection const& d2) const
        {
            return d1.azimuthDeg < d2.azimuthDeg
                || (d1.azimuthDeg == d2.azimuthDeg && d1.elevationDeg < d2.elevationDeg);
        }
    };
    std::set<Vbap::SpeakerNodeDirection, Cmp> set;

    for (auto const d : directions) {
        set.insert(d);
    }

    return set.size() < directions.size();
}

} // namespace

Expected<Vbap, Vbap::Error> Vbap::create(
    std::vector<SpeakerNodeDirection> const& speakerNodeDirections) noexcept
{
    Vbap vbap;
    auto const speakerNodeCount = speakerNodeDirections.size();

    // Basic init checks
    if (speakerNodeCount > MAX_NUM_SPEAKER_NODES || speakerNodeCount <= 3) {
        return Error::illegalSpeakerNodeCount;
    }
    if (speakerNodeDirections.empty()) {
        return Error::missingSpeakerNodeData;
    }
    if (haveDuplicates(speakerNodeDirections)) {
        return Error::duplicateSpeakerNodes;
    }

    try {
        vbap.m_topVirtualSpeakerNodeDivisionGains.resize(speakerNodeCount, 0.F);
        vbap.m_bottomVirtualSpeakerNodeDivisionGains.resize(speakerNodeCount, 0.F);
    } catch (...) {
        return Error::outOfMemory;
    }

    // Check if the speaker node setup needs a virtual top or bottom node.  This
    // increments m_internalSpeakerNodeCount when necessary, so this needs to be
    // done before later operations that depend on the internal speaker count
    // may proceed.
    auto const virtualBottomType =
        vbap.checkNeedOfVirtualSpeakerNode(speakerNodeDirections, SpeakerNodeGroup::bottomHalf);
    auto const virtualTopType =
        vbap.checkNeedOfVirtualSpeakerNode(speakerNodeDirections, SpeakerNodeGroup::topHalf);

    auto const internalSpeakerNodeCount = speakerNodeCount
        + (virtualTopType != VirtualSpeakerNodeType::none ? 1 : 0)
        + (virtualBottomType != VirtualSpeakerNodeType::none ? 1 : 0);

    // Init internal speaker node configuration, which is the original
    // configuration potentially appended with virtual top and/or bottom
    // loudspeakers.
    std::vector<SpeakerNodeDirection> internalSpeakerNodeDirections;
    try {
        internalSpeakerNodeDirections.resize(internalSpeakerNodeCount, { 0.F, 0.F });
    } catch (...) {
        return Error::outOfMemory;
    }

    copy(speakerNodeDirections.cbegin(), speakerNodeDirections.cend(),
        internalSpeakerNodeDirections.begin());

    // Insert virtual speaker nodes if needed.
    auto currentSpeakerCount = static_cast<std::uint16_t>(speakerNodeCount);
    if (virtualBottomType != VirtualSpeakerNodeType::none) {
        vbap.m_bottomVirtualSpeakerNodeIndex = currentSpeakerCount;
        internalSpeakerNodeDirections[currentSpeakerCount].azimuthDeg = 0.0F;
        internalSpeakerNodeDirections[currentSpeakerCount].elevationDeg = MIN_ELEVATION_DEG;
        ++currentSpeakerCount;
    }
    if (virtualTopType != VirtualSpeakerNodeType::none) {
        vbap.m_topVirtualSpeakerNodeIndex = currentSpeakerCount;
        internalSpeakerNodeDirections[currentSpeakerCount].azimuthDeg = 0.0F;
        internalSpeakerNodeDirections[currentSpeakerCount].elevationDeg = MAX_ELEVATION_DEG;
        ++currentSpeakerCount;
    }

    std::vector<SpeakerNode> internalSpeakerNodes;
    Connections connections;
    try {
        internalSpeakerNodes = createInternalSpeakerNodes(internalSpeakerNodeDirections);
        connections = determineConnections(internalSpeakerNodes);
        vbap.m_triplets = determineVirtualSurfaceTriplets(internalSpeakerNodes, connections);
    } catch (...) {
        return Error::outOfMemory;
    }

    // Determine how the virtual node gains should be distributed to real nodes,
    // if necessary.
    determineVirtualSpeakerNodeDivisionGains(vbap.m_topVirtualSpeakerNodeIndex,
        vbap.m_topVirtualSpeakerNodeDivisionGains, connections, virtualTopType);
    determineVirtualSpeakerNodeDivisionGains(vbap.m_bottomVirtualSpeakerNodeIndex,
        vbap.m_bottomVirtualSpeakerNodeDivisionGains, connections, virtualBottomType);

    vbap.createTripletSelectionTable();

    // Requires speakers on horizontal plane to have a separation <= 180 deg
    auto maxGap = maxGapBetweenSpeakersOnPlane(internalSpeakerNodes, 0.F);
    if (!maxGap.has_value() || maxGap.value() > 180.F) {
        return Error::tooLargeHorizontalPlaneGap;
    }

    return vbap;
}

bool Vbap::determineGains(
    float anyAziDeg, float anyEleDeg, std::vector<float>& gains) const noexcept
{
    if (gains.size() < m_topVirtualSpeakerNodeDivisionGains.size()) {
        return false;
    }
    if (std::isnan(anyAziDeg) || std::isnan(anyEleDeg)) {
        return false;
    }
    auto const [aziDeg, eleDeg] = aziEleToExpectedRange(anyAziDeg, anyEleDeg);

    auto const aziStepCount = std::round(aziDeg / static_cast<float>(AZI_RESOLUTION_DEG));
    auto const eleStepCount = std::round(eleDeg / static_cast<float>(ELE_RESOLUTION_DEG));

    // Get gains based on pre-formulated triplet inverse matrix and the panning
    // unit vector
    auto const aziRad = deg2rad(aziStepCount * AZI_RESOLUTION_DEG);
    auto const eleRad = deg2rad(eleStepCount * ELE_RESOLUTION_DEG);
    Vector3 const panningUnitVec{
        std::cos(aziRad) * std::cos(eleRad),
        std::sin(aziRad) * std::cos(eleRad),
        std::sin(eleRad),
    };
    auto const& triplet = m_triplets[tripletIndex(aziStepCount, eleStepCount)];

    // Normalize gains to unit energy
    auto const gainTriplet =
        normalized(negativesZeroed(vectorMatrixMultiply3x3(panningUnitVec, triplet.inverseMatrix)),
            NORMALIZATION_DEFAULT_SUM_INIT);

    fill(gains, 0.F);

    // Map gain triplet (internal speaker node configuration) to speaker node
    // output (actual speaker node configuration)
    for (std::size_t i{ 0 }; i < 3; ++i) {
        auto const speakerNodeIndex = triplet.speakerNodeIndices.at(i);

        auto const tripletGain = gainTriplet.at(i);
        if (speakerNodeIndex == m_bottomVirtualSpeakerNodeIndex) {
            for (auto j{ m_bottomVirtualSpeakerNodeDivisionGains.size() }; j-- > 0;) {
                gains[j] += m_bottomVirtualSpeakerNodeDivisionGains[j] * tripletGain;
            }
        } else if (speakerNodeIndex == m_topVirtualSpeakerNodeIndex) {
            for (auto j{ m_topVirtualSpeakerNodeDivisionGains.size() }; j-- > 0;) {
                gains[j] += m_topVirtualSpeakerNodeDivisionGains[j] * tripletGain;
            }
        } else {
            gains.at(speakerNodeIndex) += tripletGain;
        }
    }

    return true;
}

// Return the triplet index given the `aziStepCount` and `eleStepCount`.
std::uint8_t Vbap::tripletIndex(float aziStepCount, float eleStepCount) const noexcept
{
    auto const [aziIndex, eleIndex] = angleIndicesFromStepCounts(aziStepCount, eleStepCount);
    assert(aziIndex < m_tripletSelect.size());
    assert(eleIndex < m_tripletSelect.at(aziIndex).size());
    return m_tripletSelect.at(aziIndex)[eleIndex];
}

// Return the type of virtual speaker node needed for `group`.
Vbap::VirtualSpeakerNodeType Vbap::checkNeedOfVirtualSpeakerNode(
    std::vector<SpeakerNodeDirection> const& speakerNodeDirections, SpeakerNodeGroup group) noexcept
{
    float maxElevationDeg{ 0.0F };

    for (auto speakerNodeDirection : speakerNodeDirections) {
        if (group == SpeakerNodeGroup::topHalf) {
            if (speakerNodeDirection.elevationDeg > maxElevationDeg) {
                maxElevationDeg = speakerNodeDirection.elevationDeg;
            }
        } else {
            if (-speakerNodeDirection.elevationDeg > maxElevationDeg) {
                maxElevationDeg = -speakerNodeDirection.elevationDeg;
            }
        }
    }

    if (maxElevationDeg > NO_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG - EPSILON) {
        return VirtualSpeakerNodeType::none;
    }

    if (maxElevationDeg > DISTRIBUTE_VIRTUAL_SPEAKER_NODE_ELE_LIMIT_DEG - EPSILON) {
        return VirtualSpeakerNodeType::distributeEnergy;
    }

    return VirtualSpeakerNodeType::distributeEnergy;
}

// Create the internal triplet selection table.
void Vbap::createTripletSelectionTable() noexcept
{
    std::size_t const FULL_CIRCLE_DEG{ 360 };
    for (std::size_t aziDeg{ 0 }, aziIndex{ 0 }; aziDeg < FULL_CIRCLE_DEG;
         aziDeg += AZI_RESOLUTION_DEG, ++aziIndex) {
        auto const aziRad = deg2rad(aziDeg);
        std::size_t eleIndex{ 0 };
        for (float eleDeg{ MIN_ELEVATION_DEG }; eleDeg <= MAX_ELEVATION_DEG;) {
            auto const eleRad = deg2rad(eleDeg);
            assert(aziIndex < m_tripletSelect.size());
            assert(eleIndex < m_tripletSelect.at(aziIndex).size());
            m_tripletSelect.at(aziIndex)[eleIndex] =
                determineBestTriplet(aziRad, eleRad, m_triplets);
            ++eleIndex;
            eleDeg += ELE_RESOLUTION_DEG;
        }
    }
}

} // namespace mpegcore
