/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_VBAP_HPP
#define MPEGCORE_VBAP_HPP

#include <array>
#include <cstdint>
#include <optional>
#include <vector>

#include <mpegcore/Expected.hpp>

namespace mpegcore {

class Vbap
{
public:
    // Resolutions for triplet selection table.
    std::int16_t const static AZI_RESOLUTION_DEG{ 2 };
    std::int16_t const static ELE_RESOLUTION_DEG{ 5 };

    using Vector3 = std::array<float, 3>;
    using Matrix3x3 = std::array<Vector3, 3>;

    enum class Error : std::uint8_t
    {
        ok,
        illegalSpeakerNodeCount,
        missingSpeakerNodeData,
        outOfMemory,
        tooLargeHorizontalPlaneGap,
        duplicateSpeakerNodes,
    };

    enum class VirtualSpeakerNodeType : std::uint8_t
    {
        none,
        discardEnergy,
        distributeEnergy
    };

    enum class SpeakerNodeGroup : std::uint8_t
    {
        bottomHalf,
        horizontal,
        topHalf
    };

    // Defines a single virtual surface triplet of loudspeakers
    // with a precalculated inverse matrix
    struct VirtualSurfaceTriplet
    {
        Matrix3x3 inverseMatrix;
        std::array<std::uint8_t, 3> speakerNodeIndices;
    };

    struct SpeakerNodeDirection
    {
        // Positive left.
        float azimuthDeg;
        // Positive up.
        float elevationDeg;
    };

    // Initialize VBAP data structure given the azimuths and elevations of
    // speakers.
    //
    // Return the initialized `Vbap` or an error if one couldn't be initalized.
    Expected<Vbap, Error> static create(
        std::vector<SpeakerNodeDirection> const& speakerNodeDirections) noexcept;

    // Update `gains` with gains for all speaker nodes for azimuth `aziDeg` and
    // elevation `eleDeg`.
    //
    // Return `true` if gains were successfully determined, `false` otherwise
    // (e.g. if `gains` is shorter than the number of speaker nodes the `Vbap`
    // was initialized with).
    bool determineGains(float aziDeg, float eleDeg, std::vector<float>& gains) const noexcept;

private:
    Vbap::VirtualSpeakerNodeType checkNeedOfVirtualSpeakerNode(
        std::vector<SpeakerNodeDirection> const& speakerNodeDirections,
        SpeakerNodeGroup group) noexcept;

    void createTripletSelectionTable() noexcept;
    std::uint8_t tripletIndex(float aziStepCount, float eleStepCount) const noexcept;

    std::optional<std::uint16_t> m_topVirtualSpeakerNodeIndex{};
    std::optional<std::uint16_t> m_bottomVirtualSpeakerNodeIndex{};

    std::array<std::array<std::uint8_t, 180 / ELE_RESOLUTION_DEG + 1>, 360 / AZI_RESOLUTION_DEG>
        m_tripletSelect;
    std::vector<VirtualSurfaceTriplet> m_triplets;

    std::vector<float> m_bottomVirtualSpeakerNodeDivisionGains;
    std::vector<float> m_topVirtualSpeakerNodeDivisionGains;
};

} // namespace mpegcore

#endif // MPEGCORE_VBAP_HPP
