/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_ALGORITHM_HPP
#define MPEGCORE_ALGORITHM_HPP

// Some utility functions for working with STL's algorithms.

#include <algorithm>
#include <cassert>
#include <iterator>
#include <numeric>
#include <utility>

namespace mpegcore {

template <typename T, typename V> auto accumulate(T&& container, V init)
{
    return std::accumulate(container.begin(), container.end(), init);
}

template <typename T, typename V, typename BinaryOperation>
auto accumulate(T&& container, V init, BinaryOperation op)
{
    return std::accumulate(container.begin(), container.end(), init, op);
}

template <typename T, typename V> auto fill(T&& container, V const& value)
{
    return std::fill(container.begin(), container.end(), value);
}

template <typename T, typename P> auto findIf(T&& container, P&& predicate)
{
    return std::find_if(container.begin(), container.end(), predicate);
}

// Apply `operation` to each element in `input` and store the result at
// corresponding element in `output`.
//
// This works only with containers, and `input` and `output` must be of
// the same size.
template <typename TInput, typename TOutput, typename UnaryOperation>
auto transform(TInput const& input, TOutput& output, UnaryOperation&& operation)
{
    assert(input.size() == output.size());
    return std::transform(
        cbegin(input), cend(input), begin(output), std::forward<UnaryOperation>(operation));
}

// Apply `operation` to each element in `input` and insert the results to the
// back of `output`.
template <typename TInput, typename TOutput, typename UnaryOperation>
auto transformToBack(TInput const& input, TOutput& output, UnaryOperation&& operation)
{
    output.reserve(output.size() + input.size());
    return std::transform(cbegin(input), cend(input), std::back_inserter(output),
        std::forward<UnaryOperation>(operation));
}

} // namespace mpegcore

#endif // MPEGCORE_ALGORITHM_HPP
