/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "loudspeakerlayout.hpp"

#include <array>
#include <cmath>
#include <limits>
#include <numeric>

namespace mpegcore {

namespace {

auto const LFE{ cicp::ChannelType::lfe };
auto constexpr NaN{ std::numeric_limits<float>::quiet_NaN() };

auto const& CHANNEL_CONFIGURATIONS() noexcept
{
    // Channel configurations per ISO/IEC DIS 23091-3:2017 (E) Table 3
    using Configurations =
        std::array<cicp::ChannelConfiguration, cicp::CHANNEL_CONFIGURATION_COUNT>;
    Configurations const static CONFIGURATIONS{
        // clang-format off

        // Layout 0 (any setup)
        cicp::ChannelConfiguration{ },

        // Layout 1 (1/0.0)
        {
            {    0,    0 }, // centre front
        },

        // Layout 2 (2/0.0)
        {
            {   30,    0 }, // left front
            {  -30,    0 }, // right front
        },

        // Layout 3 (3/0.0)
        {
            {    0,    0 }, // centre front
            {   30,    0 }, // left front
            {  -30,    0 }, // right front
        },

        // Layout 4 (3/1.0)
        {
            {    0,    0 }, // centre front
            {   30,    0 }, // left front
            {  -30,    0 }, // right front
            {  180,    0 }, // rear centre
        },

        // Layout 5 (3/2.0)
        {
            {    0,    0 }, // centre front
            {   30,    0 }, // left front
            {  -30,    0 }, // right front
            {  110,    0 }, // left surround
            { -110,    0 }, // right surround
        },

        // Layout 6 (3/2.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {    0,  -15, LFE }, // LFE
        },

        // Layout 7 (5/2.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front centre
            {  -30,    0      }, // right front centre
            {   45,    0      }, // left front
            {  -45,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {    0,  -15, LFE }, // LFE
        },

        // Layout 8 (1+1)
        {
            { NaN, NaN },
            { NaN, NaN },
        },

        // Layout 9 (2/1.0)
        {
            {   30,    0 }, // left front
            {  -30,    0 }, // right front
            {  180,    0 }, // rear centre
        },

        // Layout 10 (2/2.0)
        {
            {   30,    0 }, // left front
            {  -30,    0 }, // right front
            {  110,    0 }, // left surround
            { -110,    0 }, // right surround
        },

        // Layout 11 (3/3.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {  180,    0      }, // rear centre
            {    0,  -15, LFE }, // LFE
        },

        // Layout 12 (3/4.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {  135,    0      }, // rear surround left
            { -135,    0      }, // rear surround right
            {    0,  -15, LFE }, // LFE
        },

        // Layout 13 (11/11.2)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front centre
            {  -30,    0      }, // right front centre
            {   60,    0      }, // left front
            {  -60,    0      }, // right front
            {   90,    0      }, // left side surround
            {  -90,    0      }, // right side surround
            {  135,    0      }, // rear surround left
            { -135,    0      }, // rear surround right
            {  180,    0      }, // rear centre
            {   45,  -15, LFE }, // LFE
            {  -45,  -15, LFE }, // LFE2
            {    0,   35      }, // centre front vertical height
            {   45,   35      }, // left front vertical height
            {  -45,   35      }, // right front vertical height
            {   90,   35      }, // left vertical height side surround
            {  -90,   35      }, // right vertical height side surround
            {    0,   90      }, // top centre surround
            {  135,   35      }, // left surround vertical height rear
            { -135,   35      }, // right surround vertical height rear
            {  180,   35      }, // centre vertical height rear
            {    0,  -15      }, // centre front vertical bottom
            {   45,  -15      }, // left front vertical bottom
            {  -45,  -15      }, // right front vertical bottom
        },

        // Layout 14 (5/2.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {   45,  -15, LFE }, // LFE
            {   30,   35      }, // left front vertical height
            {  -30,   35      }, // right front vertical height
        },

        // Layout 15 (5/5.2)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {   90,    0      }, // left side surround
            {  -90,    0      }, // right side surround
            {  135,    0      }, // left surround
            { -135,    0      }, // right surround
            {   45,   35      }, // left front vertical height
            {  -45,   35      }, // right front vertical height
            {  180,   45      }, // centre vertical height rear
            {   45,  -15, LFE }, // LFE
            {  -45,  -15, LFE }, // LFE2
        },

        // Layout 16 (5/4.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {   45,  -15, LFE }, // LFE
            {   30,   30      }, // left front vertical height
            {  -30,   30      }, // right front vertical height
            {  110,   30      }, // left vertical height surround
            { -110,   30      }, // right vertical height surround
        },

        // Layout 17 (6/5.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {    0,  -15, LFE }, // LFE
            {   30,   30      }, // left front vertical height
            {  -30,   30      }, // right front vertical height
            {    0,   30      }, // centre front vertical height
            {  110,   30      }, // left vertical height surround
            { -110,   30      }, // right vertical height surround
            {    0,   90      }, // top centre surround
        },

        // Layout 18 (6/7.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {  110,    0      }, // left surround
            { -110,    0      }, // right surround
            {  150,    0      }, // left back surround
            { -150,    0      }, // right back surround
            {    0,  -15, LFE }, // LFE
            {   30,   30      }, // left front vertical height
            {  -30,   30      }, // right front vertical height
            {    0,   30      }, // centre front vertical height
            {  110,   30      }, // left vertical height surround
            { -110,   30      }, // right vertical height surround
            {    0,   90      }, // top centre surround
        },

        // Layout 19 (5/6.1)
        {
            {    0,    0      }, // centre front
            {   30,    0      }, // left front
            {  -30,    0      }, // right front
            {   90,    0      }, // left side surround
            {  -90,    0      }, // right side surround
            {  135,    0      }, // rear surround left
            { -135,    0      }, // rear surround right
            {    0,  -15, LFE }, // LFE
            {   30,   30      }, // left front vertical height
            {  -30,   30      }, // right front vertical height
            {  135,   30      }, // left surround vertical height rear
            { -135,   30      }, // right surround vertical height rear
        },

        // clang-format on
    };

    return CONFIGURATIONS;
}

} // namespace

std::size_t channelCount(cicp::LayoutIndex inputLayout) noexcept
{
    return channelConfiguration(inputLayout).size();
}

std::size_t lfeCount(cicp::LayoutIndex inputLayout) noexcept
{
    return lfeIndices(inputLayout).size();
}

std::vector<std::size_t> lfeIndices(cicp::LayoutIndex inputLayout) noexcept
{
    std::vector<std::size_t> indices;
    try {
        auto const& config = channelConfiguration(inputLayout);
        for (std::size_t i{ 0 }, end{ config.size() }; i < end; ++i) {
            if (config[i].type == cicp::ChannelType::lfe) {
                indices.push_back(i);
            }
        }
    } catch (...) {
        indices.clear();
    }
    return indices;
}

cicp::ChannelConfiguration const& channelConfiguration(cicp::LayoutIndex inputLayout) noexcept
{
    if (!isValid(inputLayout)) {
        static cicp::ChannelConfiguration const empty{};
        return empty;
    }

    return CHANNEL_CONFIGURATIONS().at(inputLayout);
}

bool isValid(cicp::LayoutIndex inputLayout) noexcept
{
    return inputLayout < cicp::CHANNEL_CONFIGURATION_COUNT;
}

} // namespace mpegcore
