/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "FdnReverb.hpp"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <iostream>
#include <numeric>
#include <vector>

#include <mpegcore/AudioBufferArray.hpp>
#include <mpegcore/constants.hpp>
#include <mpegcore/spdlog.hpp>

namespace mpegcore {
namespace frontend {

FdnReverb::FdnReverb(ReverberatorParameters const& params)
{
    m_outputCount = params.delayLines.size();
    m_oneOverM = 1.0f / std::sqrt(m_outputCount + 1);
    switch (m_outputCount) {
    case 3:
        m_hasPositiveFeedbackGain = {
            1,
        };
        m_hasNegativeFeedbackGain = {
            0,
            2,
        };
        m_alpha = -0.1666666667f;
        break;
    case 7:
        m_hasPositiveFeedbackGain = {
            1,
            2,
            4,
        };
        m_hasNegativeFeedbackGain = {
            0,
            3,
            5,
            6,
        };
        m_alpha = -0.0923495156f;
        break;
    case 15:
        m_hasPositiveFeedbackGain = {
            1,
            2,
            3,
            5,
            6,
            9,
            11,
        };
        m_hasNegativeFeedbackGain = {
            0,
            4,
            7,
            8,
            10,
            12,
            13,
            14,
        };
        m_alpha = -0.0500000000f;
        break;
    case 31:
        m_hasPositiveFeedbackGain = {
            1,
            2,
            3,
            4,
            6,
            7,
            9,
            12,
            13,
            19,
            20,
            21,
            24,
            28,
            30,
        };
        m_hasNegativeFeedbackGain = {
            0,
            5,
            8,
            10,
            11,
            14,
            15,
            16,
            17,
            18,
            22,
            23,
            25,
            26,
            27,
            29,
        };
        m_alpha = -0.0265555905f;
        break;
    case 63:
        m_hasPositiveFeedbackGain = {
            1,
            2,
            3,
            4,
            5,
            7,
            8,
            9,
            10,
            13,
            14,
            15,
            17,
            19,
            20,
            25,
            27,
            28,
            29,
            33,
            34,
            36,
            37,
            39,
            42,
            46,
            49,
            50,
            53,
            55,
            57,
        };
        m_hasNegativeFeedbackGain = {
            0,
            6,
            11,
            12,
            16,
            18,
            21,
            22,
            23,
            24,
            26,
            30,
            31,
            32,
            35,
            38,
            40,
            41,
            43,
            44,
            45,
            47,
            48,
            51,
            52,
            54,
            56,
            58,
            59,
            60,
            61,
            62,
        };
        m_alpha = -0.0138888889f;
        break;
    default:
        SPDLOG_ERROR("Can't configure FdnReverb to have {} outputs.", m_outputCount);
        assert(false && "Can't configure FdnReverb");
    }
    m_delayLines = std::vector<DelayLine>(m_outputCount);
    m_filters = std::vector<std::unique_ptr<GraphicEqFilter>>(m_outputCount);
    m_v = std::vector<float>(m_outputCount * 2, 0.0F);

    for (std::size_t i{ 0 }; i < m_outputCount; ++i) {
        auto const delay = params.delayLines[i].delayInSamples;
        m_delayLines[i] = DelayLine{ delay, delay };
        m_filters[i] = makeGraphicEqFilter(params.delayLines[i].filterParams);
    }
}

void FdnReverb::setInputBuffer(const std::vector<float>& input)
{
    m_inputBuffer = &input;
}

void FdnReverb::setOutputBuffer(AudioBufferArray& output)
{
    assert(output.size() == FdnReverb::m_outputCount);
    m_outputBuffer = &output;
}

void FdnReverb::process(std::uint32_t sampleCount)
{
    m_outputBuffer->fill(0.F);
    float outputScaler{ 1.0f / m_outputCount };
    std::vector<float> updated_v(m_outputCount);
    const size_t numPositiveItems = m_hasPositiveFeedbackGain.size();
    const size_t numNegativeItems = m_hasNegativeFeedbackGain.size();

    for (std::size_t sampleIndex{ 0 }; sampleIndex < sampleCount; sampleIndex++) {
        auto const d = std::accumulate(m_v.begin(), m_v.begin() + m_outputCount, 0.F);
        auto const alphaTimesD = m_alpha * d;
        std::fill(updated_v.begin(), updated_v.end(), 0.0f);

        for (std::size_t delayLineIndex{ 0 }; delayLineIndex < m_outputCount; delayLineIndex++) {
            // Read position starts from the beginning of the ring buffer and moves backwards (from
            // the beginning of the second half which repeats the first half).
            float* s = &m_v[m_outputCount - delayLineIndex];

            float b = 0;
            float c = 0;
            size_t* positiveIds = &m_hasPositiveFeedbackGain[0];
            size_t* negativeIds = &m_hasNegativeFeedbackGain[0];
            for (std::size_t dlId2{ 0 }; dlId2 < numPositiveItems; dlId2++) {
                b += s[*positiveIds++];
            }
            for (std::size_t dlId3{ 0 }; dlId3 < numNegativeItems; dlId3++) {
                c += s[*negativeIds++];
            }
            float a = m_oneOverM * (b - c) + alphaTimesD;

            auto const delayedSample =
                m_delayLines[delayLineIndex].getDelayedSample(m_inputBuffer->at(sampleIndex) + a);
            auto filteredSample = m_filters[delayLineIndex]->filter(delayedSample);
            updated_v[delayLineIndex] = filteredSample;

            (*m_outputBuffer)[delayLineIndex][sampleIndex] += filteredSample;
        }

        std::copy_n(&updated_v[0], m_outputCount, &m_v[0]);
        std::copy_n(&updated_v[0], m_outputCount, &m_v[m_outputCount]);

        for (std::size_t outputChannelIndex = 0; outputChannelIndex < m_outputCount;
             outputChannelIndex++) {
            (*m_outputBuffer)[outputChannelIndex][sampleIndex] *= outputScaler;
        }
    }
}

} // namespace frontend
} // namespace mpegcore
