/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "GraphicEqFilter.hpp"

#include <cassert>

#include <mpegcore/spdlog.hpp>

namespace mpegcore::frontend {

namespace {

class GraphicEqParallelFilter : public GraphicEqFilter
{
public:
    GraphicEqParallelFilter(GraphicEqParameters const& params)
        : GraphicEqFilter()
    {
        auto const bandCount = params.bandCount();
        m_a1.resize(bandCount);
        m_a2.resize(bandCount);
        m_b0.resize(bandCount);
        m_b1.resize(bandCount);
        m_wPast0.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast1.resize(bandCount, vcl::Vec8f(0.F));

        for (auto i{ bandCount }; i-- > 0;) {
            m_a1[i] = vcl::Vec8f(params.a[0][i]);
            m_a2[i] = vcl::Vec8f(params.a[1][i]);
            m_b0[i] = vcl::Vec8f(params.b[0][i]);
            m_b1[i] = vcl::Vec8f(params.b[1][i]);
        }
        m_F = params.F;
        m_level = GainDb{ -params.levelDb }.toLinear().value;
    }

    std::size_t bandCount() const noexcept override { return m_wPast0.size(); }

    GraphicEqParameters::Type type() const noexcept override
    {
        return GraphicEqParameters::Type::parallel;
    }

    void copyState(std::size_t toLaneIndex, GraphicEqFilter const& fromFilter,
        std::size_t fromLaneIndex) override
    {
        auto& from = *static_cast<GraphicEqParallelFilter const*>(&fromFilter);
        auto const count = bandCount();
        auto const fli = static_cast<int>(fromLaneIndex);
        std::size_t const tli{ toLaneIndex };

        for (std::size_t i{ 0 }; i < count; ++i) {
            setVec8(tli, m_a1[i], from.m_a1[i][fli]);
            setVec8(tli, m_a2[i], from.m_a2[i][fli]);
            setVec8(tli, m_b0[i], from.m_b0[i][fli]);
            setVec8(tli, m_b1[i], from.m_b1[i][fli]);
            setVec8(tli, m_wPast0[i], from.m_wPast0[i][fli]);
            setVec8(tli, m_wPast1[i], from.m_wPast1[i][fli]);
        }
        setVec8(toLaneIndex, m_F, from.m_F[fli]);
        setVec8(toLaneIndex, m_level, from.m_level[fli]);
    }

    vcl::Vec8f filter(vcl::Vec8f sample, vcl::Vec8f a1, vcl::Vec8f a2, vcl::Vec8f b0, vcl::Vec8f b1,
        vcl::Vec8f& wPast0, vcl::Vec8f& wPast1)
    {
        auto const w = sample - a1 * wPast0 - a2 * wPast1;
        auto const y = b0 * w + b1 * wPast0;

        wPast1 = wPast0;
        wPast0 = w;

        return y;
    }

    float filter(float sample) noexcept override
    {
        auto const level = m_level;
        auto const F = m_F;
        auto filteredSample = F * sample;
        for (std::size_t i{ 0 }, end{ m_a1.size() }; i < end; ++i) {
            auto const y =
                filter(sample, m_a1[i], m_a2[i], m_b0[i], m_b1[i], m_wPast0[i], m_wPast1[i]);
            filteredSample += y;
        }
        auto const result = level * filteredSample;
        return result[0];
    }

    void filter(AudioBuffer& buffer) noexcept override
    {
        for (auto& sample : buffer) {
            sample = filter(sample);
        }
    }

    void filter(AudioVectorBuffer& buffer) noexcept override
    {
        std::size_t sampleIndex{ 0 };
        auto const level = m_level;
        auto const F = m_F;
        for (auto sample : buffer) {
            auto filteredSample = F * sample;
            for (std::size_t i{ 0 }, end{ m_a1.size() }; i < end; ++i) {
                auto const y =
                    filter(sample, m_a1[i], m_a2[i], m_b0[i], m_b1[i], m_wPast0[i], m_wPast1[i]);
                filteredSample += y;
            }
            buffer[sampleIndex++] = level * filteredSample;
        }
    }

    void setParameters(std::size_t laneIndex, GraphicEqParameters const& params) override
    {
        auto const currentBandCount = bandCount();
        auto const bandCount = params.bandCount();
        if (currentBandCount > 0 && currentBandCount != bandCount) {
            SPDLOG_ERROR("Can't change a filter's band count (old {}, new {}). "
                         "Filter parameters unchanged.",
                currentBandCount, bandCount);
            assert(false && "Can't change a filter's band count");
            return;
        }

        m_b0.resize(bandCount, vcl::Vec8f(0.F));
        m_b1.resize(bandCount, vcl::Vec8f(0.F));
        m_a1.resize(bandCount, vcl::Vec8f(0.F));
        m_a2.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast0.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast1.resize(bandCount, vcl::Vec8f(0.F));

        set(laneIndex, m_a1, params.a[0]);
        set(laneIndex, m_a2, params.a[1]);
        set(laneIndex, m_b0, params.b[0]);
        set(laneIndex, m_b1, params.b[1]);
        clear(laneIndex, m_wPast0);
        clear(laneIndex, m_wPast1);
        setVec8(laneIndex, m_F, params.F);
        setVec8(laneIndex, m_level, GainDb{ -params.levelDb }.toLinear().value);
    }

private:
    std::vector<vcl::Vec8f> m_a1;
    std::vector<vcl::Vec8f> m_a2;
    std::vector<vcl::Vec8f> m_b0;
    std::vector<vcl::Vec8f> m_b1;
    std::vector<vcl::Vec8f> m_wPast0;
    std::vector<vcl::Vec8f> m_wPast1;
    vcl::Vec8f m_F;
    vcl::Vec8f m_level;
};

class GraphicEqSilentFilter : public GraphicEqFilter
{
public:
    GraphicEqSilentFilter()
        : GraphicEqFilter()
    {}

    std::size_t bandCount() const noexcept override { return 0; }

    GraphicEqParameters::Type type() const noexcept override
    {
        return GraphicEqParameters::Type::silent;
    }

    void copyState(std::size_t toLaneIndex, GraphicEqFilter const& fromFilter,
        std::size_t fromLaneIndex) override
    {}

    float filter(float sample) noexcept override { return 0.F; }

    void filter(AudioBuffer& buffer) noexcept override
    {
        std::fill_n(buffer.begin(), buffer.size(), 0.F);
    }

    void filter(AudioVectorBuffer& buffer) noexcept override
    {
        std::fill_n(buffer.begin(), buffer.size(), 0.F);
    }

    void setParameters(std::size_t laneIndex, GraphicEqParameters const& params) override {}
};

class GraphicEqCascadeFilter : public GraphicEqFilter
{
public:
    GraphicEqCascadeFilter(GraphicEqParameters const& params)
        : GraphicEqFilter()
    {
        auto const bandCount = params.bandCount();
        m_a1.resize(bandCount);
        m_a2.resize(bandCount);
        m_b1.resize(bandCount);
        m_b2.resize(bandCount);
        m_wPast0.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast1.resize(bandCount, vcl::Vec8f(0.F));

        for (auto i{ bandCount }; i-- > 0;) {
            m_a1[i] = vcl::Vec8f(params.a[0][i]);
            m_a2[i] = vcl::Vec8f(params.a[1][i]);
            m_b1[i] = vcl::Vec8f(params.b[0][i]);
            m_b2[i] = vcl::Vec8f(params.b[1][i]);
        }
        m_levelAndFFactor = params.F * GainDb{ -params.levelDb }.toLinear().value;
    }

    std::size_t bandCount() const noexcept override { return m_wPast0.size(); }

    GraphicEqParameters::Type type() const noexcept override
    {
        return GraphicEqParameters::Type::cascade;
    }

    void copyState(std::size_t toLaneIndex, GraphicEqFilter const& fromFilter,
        std::size_t fromLaneIndex) override
    {
        auto& from = *static_cast<GraphicEqCascadeFilter const*>(&fromFilter);
        auto const count = bandCount();
        auto const fli = static_cast<int>(fromLaneIndex);
        std::size_t const tli{ toLaneIndex };

        for (std::size_t i{ 0 }; i < count; ++i) {
            setVec8(tli, m_a1[i], from.m_a1[i][fli]);
            setVec8(tli, m_a2[i], from.m_a2[i][fli]);
            setVec8(tli, m_b1[i], from.m_b1[i][fli]);
            setVec8(tli, m_b2[i], from.m_b2[i][fli]);
            setVec8(tli, m_wPast0[i], from.m_wPast0[i][fli]);
            setVec8(tli, m_wPast1[i], from.m_wPast1[i][fli]);
        }
        setVec8(toLaneIndex, m_levelAndFFactor, from.m_levelAndFFactor[fli]);
    }

    vcl::Vec8f filter(vcl::Vec8f sample, vcl::Vec8f a1, vcl::Vec8f a2, vcl::Vec8f b1, vcl::Vec8f b2,
        vcl::Vec8f& wPast0, vcl::Vec8f& wPast1)
    {
        auto const w = sample - a1 * wPast0 - a2 * wPast1;
        sample = w + b1 * wPast0 + b2 * wPast1;

        wPast1 = wPast0;
        wPast0 = w;

        return sample;
    }

    float filter(float sample) noexcept override
    {
        auto filteredSample = sample;
        for (std::size_t i{ 0 }, end{ m_a1.size() }; i < end; ++i) {
            filteredSample = filter(
                filteredSample, m_a1[i], m_a2[i], m_b1[i], m_b2[i], m_wPast0[i], m_wPast1[i])[0];
        }
        filteredSample = m_levelAndFFactor[0] * filteredSample;
        return filteredSample;
    }

    void filter(AudioBuffer& buffer) noexcept override
    {
        for (auto& sample : buffer) {
            sample = filter(sample);
        }
    }

    void filter(AudioVectorBuffer& buffer) noexcept override
    {
        std::size_t sampleIndex{ 0 };
        auto const factor = m_levelAndFFactor;
        for (auto sample : buffer) {
            auto filteredSample = sample;
            for (std::size_t i{ 0 }, end{ m_a1.size() }; i < end; ++i) {
                filteredSample = filter(
                    filteredSample, m_a1[i], m_a2[i], m_b1[i], m_b2[i], m_wPast0[i], m_wPast1[i]);
            }
            buffer[sampleIndex++] = factor * filteredSample;
        }
    }

    void setParameters(std::size_t laneIndex, GraphicEqParameters const& params) override
    {
        auto const currentBandCount = bandCount();
        auto const bandCount = params.bandCount();
        if (currentBandCount > 0 && currentBandCount != bandCount) {
            SPDLOG_ERROR("Can't change a filter's band count (old {}, new {}). "
                         "Filter parameters unchanged.",
                currentBandCount, bandCount);
            assert(false && "Can't change a filter's band count");
            return;
        }

        m_a1.resize(bandCount, vcl::Vec8f(0.F));
        m_a2.resize(bandCount, vcl::Vec8f(0.F));
        m_b1.resize(bandCount, vcl::Vec8f(0.F));
        m_b2.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast0.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast1.resize(bandCount, vcl::Vec8f(0.F));

        set(laneIndex, m_a1, params.a[0]);
        set(laneIndex, m_a2, params.a[1]);
        set(laneIndex, m_b1, params.b[0]);
        set(laneIndex, m_b2, params.b[1]);
        clear(laneIndex, m_wPast0);
        clear(laneIndex, m_wPast1);
        setVec8(
            laneIndex, m_levelAndFFactor, params.F * GainDb{ -params.levelDb }.toLinear().value);
    }

    GraphicEqParameters getParameters() override
    {
        auto fltOrder = m_a1.size();
        GraphicEqParameters params(GraphicEqParameters::Type::cascade, fltOrder);
        for (int sosIdx = 0; sosIdx < fltOrder; sosIdx++) {
            params.a[0][sosIdx] = m_a1[sosIdx].get_high().extract(0);
            params.a[1][sosIdx] = m_a2[sosIdx].get_high().extract(0);
            params.b[0][sosIdx] = m_b1[sosIdx].get_high().extract(0);
            params.b[1][sosIdx] = m_b2[sosIdx].get_high().extract(0);
        }

        params.F = m_levelAndFFactor.extract(0);

        return params;
    }

private:
    std::vector<vcl::Vec8f> m_a1;
    std::vector<vcl::Vec8f> m_a2;
    std::vector<vcl::Vec8f> m_b1;
    std::vector<vcl::Vec8f> m_b2;
    std::vector<vcl::Vec8f> m_wPast0;
    std::vector<vcl::Vec8f> m_wPast1;
    vcl::Vec8f m_levelAndFFactor;
};

class GraphicEqCascadeStagedFilter : public GraphicEqFilter
{
public:
    GraphicEqCascadeStagedFilter(GraphicEqParameters const& params)
        : GraphicEqFilter()
    {
        auto const bandCount = params.bandCount();
        m_a1.resize(bandCount);
        m_a2.resize(bandCount);
        m_b1.resize(bandCount);
        m_b2.resize(bandCount);
        m_F.resize(bandCount);
        m_wPast0.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast1.resize(bandCount, vcl::Vec8f(0.F));

        for (auto i{ bandCount }; i-- > 0;) {
            m_a1[i] = vcl::Vec8f(params.a[0][i]);
            m_a2[i] = vcl::Vec8f(params.a[1][i]);
            m_b1[i] = vcl::Vec8f(params.b[0][i]);
            m_b2[i] = vcl::Vec8f(params.b[1][i]);
            m_F[i] = vcl::Vec8f(params.stageF[i]);
        }
        m_levelAndF = params.F * GainDb{ -params.levelDb }.toLinear().value;
    }

    std::size_t bandCount() const noexcept override { return m_wPast0.size(); }

    GraphicEqParameters::Type type() const noexcept override
    {
        return GraphicEqParameters::Type::cascadeStaged;
    }

    void copyState(std::size_t toLaneIndex, GraphicEqFilter const& fromFilter,
        std::size_t fromLaneIndex) override
    {
        auto& from = *static_cast<GraphicEqCascadeStagedFilter const*>(&fromFilter);
        auto const count = bandCount();
        auto const fli = static_cast<int>(fromLaneIndex);
        std::size_t const tli{ toLaneIndex };

        for (std::size_t i{ 0 }; i < count; ++i) {
            setVec8(tli, m_a1[i], from.m_a1[i][fli]);
            setVec8(tli, m_a2[i], from.m_a2[i][fli]);
            setVec8(tli, m_b1[i], from.m_b1[i][fli]);
            setVec8(tli, m_b2[i], from.m_b2[i][fli]);
            setVec8(tli, m_wPast0[i], from.m_wPast0[i][fli]);
            setVec8(tli, m_wPast1[i], from.m_wPast1[i][fli]);
            setVec8(tli, m_F[i], from.m_F[i][fli]);
        }
        setVec8(toLaneIndex, m_levelAndF, from.m_levelAndF[fli]);
    }

    vcl::Vec8f filter(vcl::Vec8f sample, vcl::Vec8f a1, vcl::Vec8f a2, vcl::Vec8f b1, vcl::Vec8f b2,
        vcl::Vec8f& wPast0, vcl::Vec8f& wPast1)
    {
        auto const w = sample - a1 * wPast0 - a2 * wPast1;
        sample = w + b1 * wPast0 + b2 * wPast1;

        wPast1 = wPast0;
        wPast0 = w;

        return sample;
    }

    float filter(float sample) noexcept override
    {
        auto filteredSample = sample;
        for (std::size_t i{ 0 }, end{ m_a1.size() }; i < end; ++i) {
            filteredSample = filter(
                filteredSample, m_a1[i], m_a2[i], m_b1[i], m_b2[i], m_wPast0[i], m_wPast1[i])[0];
            filteredSample *= m_F[i][0];
        }
        filteredSample = m_levelAndF[0] * filteredSample;
        return filteredSample;
    }

    void filter(AudioBuffer& buffer) noexcept override
    {
        for (auto& sample : buffer) {
            sample = filter(sample);
        }
    }

    void filter(AudioVectorBuffer& buffer) noexcept override
    {
        std::size_t sampleIndex{ 0 };
        auto const factor = m_levelAndF;
        for (auto sample : buffer) {
            auto filteredSample = sample;
            for (std::size_t i{ 0 }, end{ m_a1.size() }; i < end; ++i) {
                filteredSample = filter(
                    filteredSample, m_a1[i], m_a2[i], m_b1[i], m_b2[i], m_wPast0[i], m_wPast1[i]);
                filteredSample *= m_F[i];
            }
            buffer[sampleIndex++] = factor * filteredSample;
        }
    }

    void setParameters(std::size_t laneIndex, GraphicEqParameters const& params) override
    {
        auto const currentBandCount = bandCount();
        auto const bandCount = params.bandCount();
        if (currentBandCount > 0 && currentBandCount != bandCount) {
            SPDLOG_ERROR("Can't change a filter's band count (old {}, new {}). "
                         "Filter parameters unchanged.",
                currentBandCount, bandCount);
            assert(false && "Can't change a filter's band count");
            return;
        }

        m_a1.resize(bandCount, vcl::Vec8f(0.F));
        m_a2.resize(bandCount, vcl::Vec8f(0.F));
        m_b1.resize(bandCount, vcl::Vec8f(0.F));
        m_b2.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast0.resize(bandCount, vcl::Vec8f(0.F));
        m_wPast1.resize(bandCount, vcl::Vec8f(0.F));
        m_F.resize(bandCount, vcl::Vec8f(0.F));

        set(laneIndex, m_a1, params.a[0]);
        set(laneIndex, m_a2, params.a[1]);
        set(laneIndex, m_b1, params.b[0]);
        set(laneIndex, m_b2, params.b[1]);
        set(laneIndex, m_F, params.stageF);
        clear(laneIndex, m_wPast0);
        clear(laneIndex, m_wPast1);
        setVec8(laneIndex, m_levelAndF, params.F * GainDb{ -params.levelDb }.toLinear().value);
    }

private:
    std::vector<vcl::Vec8f> m_a1;
    std::vector<vcl::Vec8f> m_a2;
    std::vector<vcl::Vec8f> m_b1;
    std::vector<vcl::Vec8f> m_b2;
    std::vector<vcl::Vec8f> m_wPast0;
    std::vector<vcl::Vec8f> m_wPast1;
    std::vector<vcl::Vec8f> m_F;
    vcl::Vec8f m_levelAndF;
};

class GraphicEqGainFilter : public GraphicEqFilter
{
public:
    GraphicEqGainFilter(GraphicEqParameters const& params)
        : GraphicEqFilter()
    {
        m_F = params.F;
    }

    std::size_t bandCount() const noexcept override { return 0; }

    GraphicEqParameters::Type type() const noexcept override
    {
        return GraphicEqParameters::Type::gain;
    }

    void copyState(std::size_t toLaneIndex, GraphicEqFilter const& fromFilter,
        std::size_t fromLaneIndex) override
    {
        auto& from = *static_cast<GraphicEqGainFilter const*>(&fromFilter);
        auto const fli = static_cast<int>(fromLaneIndex);
        setVec8(toLaneIndex, m_F, from.m_F[fli]);
    }

    float filter(float sample) noexcept override { return m_F[0] * sample; }

    void filter(AudioBuffer& buffer) noexcept override
    {
        for (auto& sample : buffer) {
            sample = filter(sample);
        }
    }

    void filter(AudioVectorBuffer& buffer) noexcept override
    {
        auto const factor = m_F;
        for (auto& sample : buffer) {
            sample *= factor;
        }
    }

    void setParameters(std::size_t laneIndex, GraphicEqParameters const& params) override
    {
        setVec8(laneIndex, m_F, params.F);
    }

private:
    vcl::Vec8f m_F;
};

} // namespace

GraphicEqFilter::GraphicEqFilter()
{
    disableDenormals();
}

std::unique_ptr<GraphicEqFilter> makeGraphicEqFilter(GraphicEqParameters const& parameters) noexcept
{
    switch (parameters.type) {
    case GraphicEqParameters::Type::cascade:
        return std::make_unique<GraphicEqCascadeFilter>(parameters);
    case GraphicEqParameters::Type::cascadeStaged:
        return std::make_unique<GraphicEqCascadeStagedFilter>(parameters);
    case GraphicEqParameters::Type::gain:
        return std::make_unique<GraphicEqGainFilter>(parameters);
    case GraphicEqParameters::Type::parallel:
        return std::make_unique<GraphicEqParallelFilter>(parameters);
    case GraphicEqParameters::Type::silent:
        return std::make_unique<GraphicEqSilentFilter>();
    case GraphicEqParameters::Type::uninitialized:
        SPDLOG_WARN("Can't make a graphic eq with uninitialized parameters, "
                    "using silent filter instead");
        return std::make_unique<GraphicEqSilentFilter>();
    default:
        SPDLOG_WARN("Unknown graphic eq type {}, using silent filter",
            static_cast<std::size_t>(parameters.type));
        return std::make_unique<GraphicEqSilentFilter>();
    }
}

} // namespace mpegcore::frontend
