/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#ifndef MPEGCORE_FRONTEND_GRAPHICEQFILTER_HPP
#define MPEGCORE_FRONTEND_GRAPHICEQFILTER_HPP

#include <cstddef>
#include <memory>
#include <stdexcept>
#include <vector>

#include <mpegcore/AudioBuffer.hpp>
#include <mpegcore/AudioVectorBuffer.hpp>
#include <mpegcore/datatypes_reverb.hpp>
#include <mpegcore/gain.hpp>

#if defined(_WIN32) || defined(_WIN64)
#include <intrin.h>
#endif

namespace mpegcore {
namespace frontend {

class GraphicEqFilter
{
public:
    GraphicEqFilter();
    virtual ~GraphicEqFilter() = default;

    virtual std::size_t bandCount() const noexcept = 0;

    virtual float filter(float sample) = 0;

    virtual void filter(AudioBuffer& buffer) = 0;

    virtual void filter(AudioVectorBuffer& buffer) = 0;

    virtual void copyState(
        std::size_t toLaneIndex, GraphicEqFilter const& fromFilter, std::size_t fromLaneIndex) = 0;

    virtual void setParameters(std::size_t laneIndex, GraphicEqParameters const& parameters) = 0;

    virtual GraphicEqParameters getParameters() { throw std::logic_error("Not implemented."); };

    virtual GraphicEqParameters::Type type() const noexcept = 0;

    static void setVec8(std::size_t laneIndex, vcl::Vec8f& vec8, float value) noexcept
    {
        vec8.insert(static_cast<int>(laneIndex), value);
    }

    void set(std::size_t laneIndex, std::vector<vcl::Vec8f>& vec8s,
        std::vector<float> const& floats) noexcept
    {
        for (std::size_t i{ vec8s.size() }; i-- > 0;) {
            vec8s[i].insert(static_cast<int>(laneIndex), floats[i]);
        }
    }

    void clear(std::size_t laneIndex, std::vector<vcl::Vec8f>& vec8s) noexcept
    {
        for (std::size_t i{ vec8s.size() }; i-- > 0;) {
            vec8s[i].insert(static_cast<int>(laneIndex), 0.F);
        }
    }

    static void disableDenormals() noexcept
    {
#if defined(_WIN32) || defined(_WIN64) || !defined(MPEGI_ARM_BUILD)
        // With suitable audio data filtering can result in denormals. If not
        // handled they will make filtering unusable performance-wise. Turn FTZ
        // and DAZ on so that denormals don't need explicit handling.

        constexpr unsigned int CSR_DAZ{ 1 << 6 }; // Denormals are zeros
        constexpr unsigned int CSR_FTZ{ 1 << 15 }; // Flush to zero
        constexpr unsigned int CSR_SET_BITS = CSR_DAZ | CSR_FTZ;
#endif

#if defined(_WIN32) || defined(_WIN64)
        unsigned int csr = _mm_getcsr();
        csr |= CSR_SET_BITS;
        _mm_setcsr(csr);
#elif !defined(MPEGI_ARM_BUILD)
        unsigned int csr = __builtin_ia32_stmxcsr();
        csr |= CSR_SET_BITS;
        __builtin_ia32_ldmxcsr(csr);
#endif
    }
};

std::unique_ptr<GraphicEqFilter> makeGraphicEqFilter(
    GraphicEqParameters const& parameters) noexcept;

} // namespace frontend
} // namespace mpegcore

#endif // MPEGCORE_FRONTEND_GRAPHICEQFILTER_HPP
