/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#define _USE_MATH_DEFINES

#include "neurallyControlledGraphicEqualizer.hpp"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstddef>
#include <functional>
#include <limits>
#include <math.h>
#include <numeric>

#include "../dblinear.hpp"

namespace mpegcore {
namespace frontend {

namespace {

using NeuralGainVector = NeuralNetworkParameters::Vector;

struct FilterGains : public NeuralGainVector
{};

struct ControlGainsDb : public NeuralGainVector
{};

using Matrix = NeuralNetworkParameters::Matrix;

template <typename F> Vector transform(Vector const& lhs, Vector const& rhs, F&& f) noexcept
{
    Vector result;
    auto const count = lhs.size();
    if (count != rhs.size()) {
        assert(false && "Need the same number of elements");
        return result;
    }
    result.resize(count);
    std::transform(lhs.cbegin(), lhs.cend(), rhs.cbegin(), result.begin(), std::move(f));
    return result;
}

Vector operator+(Vector const& lhs, Vector const& rhs) noexcept
{
    return transform(lhs, rhs, std::plus());
}

Vector operator-(Vector const& lhs, Vector const& rhs) noexcept
{
    return transform(lhs, rhs, std::minus());
}

template <typename F> Vector transform(Vector const& v, F&& f) noexcept
{
    Vector result(v.size());
    std::transform(v.cbegin(), v.cend(), result.begin(), std::move(f));
    return result;
}

Vector operator+(Vector const& lhs, double rhs) noexcept
{
    return transform(lhs, [&rhs](auto l) { return l + rhs; });
}

Vector operator-(Vector const& lhs, double rhs) noexcept
{
    return lhs + -rhs;
}

Vector operator*(double lhs, Vector const& rhs) noexcept
{
    return transform(rhs, [&lhs](auto r) { return lhs * r; });
}

Vector operator*(Matrix const& lhs, Vector const& rhs) noexcept
{
    Vector result;
    auto const columnCount = rhs.size();
    auto const rowCount = lhs.size();
    result.resize(rowCount);
    for (std::size_t i{ 0 }; i < rowCount; ++i) {
        result[i] = std::inner_product(lhs[i].cbegin(), lhs[i].cend(), rhs.cbegin(), 0.);
    }
    return result;
}

Vector tanh(Vector const& in) noexcept
{
    return transform(in, [](auto x) { return std::tanh(x); });
}

Vector hadamardDivision(Vector const& lhs, Vector const& rhs) noexcept
{
    return transform(lhs, rhs, std::divides());
}

Vector hadamard(Vector const& lhs, Vector const& rhs) noexcept
{
    return transform(lhs, rhs, std::multiplies());
}

Vector dbToLinear(Vector dbs) noexcept
{
    return transform(dbs, [](auto db) { return ::mpegcore::dbToLinear(db); });
}

double prod(Vector vector) noexcept
{
    return std::accumulate(
        vector.cbegin(), vector.cend(), 1., std::multiplies<Vector::value_type>());
}

ControlGainsDb makeControlGains(FrequencyResponse const& response)
{
    return { transform(response.response, [](auto linear) { return linearToDb(linear); }) };
}

constexpr std::size_t PAR_EQ_COEFFICIENT_COUNT{ 3 };

GraphicEqParameters::Coefficients coefficients(Matrix const& matrix, std::size_t index) noexcept
{
    assert(index < PAR_EQ_COEFFICIENT_COUNT);

    GraphicEqParameters::Coefficients result;
    auto const bandCount = matrix.size();
    result.resize(bandCount);

    for (std::size_t i{ 0 }; i < bandCount; ++i) {
        result[i] = static_cast<decltype(result)::value_type>(matrix[i][index]);
    }

    return result;
}

// Turn zeros in the input into the smallest non-zero value among the input FrequencyResponse
// (linear magnitudes). Otherwise coefficient calculations may produce NaNs.
FrequencyResponse fixZeroCoefficientsWithSmallestValue(FrequencyResponse const& in)
{
    FrequencyResponse out{ in };

    auto minGain = std::numeric_limits<double>::max();
    bool foundNonZeroMin = false;
    for (auto value : in.response) {
        if (value != 0. && value < minGain) {
            foundNonZeroMin = true;
            minGain = value;
        }
    }

    // all input FrequencyResponse values were zero, clamp to -100 dB
    if (!foundNonZeroMin) {
        minGain = 0.00001; // -100 dB
    }

    for (auto& value : out.response) {
        value = std::max(value, minGain);
    }

    return out;
}

std::size_t closestFrequencyIndex(double targetFrequency, std::vector<double> const& frequencies)
{
    std::size_t minIndex{ 0 };
    double minDiff{ std::numeric_limits<double>::max() };
    for (std::size_t i{ 0 }, end{ frequencies.size() }; i < end; ++i) {
        if (auto const diff = std::abs(frequencies[i] - targetFrequency); diff < minDiff) {
            minDiff = diff;
            minIndex = i;
        }
    }
    return minIndex;
}

FrequencyResponse calculateFrequencyResponseAtBandCenterFrequencies(
    Vector const& centerFrequencies, FrequencyResponse const& inputFrequencyResponse)
{
    FrequencyResponse centerResponse;
    auto const bandCount = centerFrequencies.size();
    centerResponse.f.reserve(bandCount);
    centerResponse.response.reserve(bandCount);

    using Frequency = decltype(centerResponse.f)::value_type;
    for (std::size_t i{ 0 }; i < bandCount; ++i) {
        centerResponse.f.push_back(static_cast<Frequency>(centerFrequencies[i]));
        auto const index = closestFrequencyIndex(centerFrequencies[i], inputFrequencyResponse.f);
        centerResponse.response.push_back(inputFrequencyResponse.response[index]);
    }

    return centerResponse;
}

struct GeqFilterParameters
{
    NeuralNetworkParameters::Matrix a;
    NeuralNetworkParameters::Matrix b;
    double G0;

    GeqFilterParameters(std::size_t count)
    {
        a.resize(count);
        b.resize(count);
    }
};

GeqFilterParameters makeGeqFilterParameters(FilterGains const& filterGains, double c,
    Vector const& B, std::size_t bandCount, Vector const& wc)
{

    auto const& g = filterGains;
    auto const G = dbToLinear(g);
    auto const gB = c * g;
    auto const GB = dbToLinear(gB);

    GeqFilterParameters result(bandCount);
    Vector b0;
    b0.resize(bandCount);

    for (std::size_t i{ 0 }; i < bandCount; ++i) {
        auto const parEqParams = pareq2Filter(G[i], GB[i], wc[i], B[i]);
        result.a[i] = parEqParams.den;
        result.b[i] = parEqParams.num;
        b0[i] = parEqParams.b0;
    }

    result.G0 = prod(b0);

    return result;
}

GeqFilterParameters geqFilterParameters(FilterGains const& filterGains, std::uint32_t sampleRateHz)
{
    constexpr std::size_t BAND_COUNT{ 10 };
    auto const wc = 2 * M_PI / sampleRateHz * bandCenterFrequencies(BAND_COUNT);
    constexpr double c{ 0.3 };

    Vector B(BAND_COUNT);
    for (std::size_t i{ 0 }, end{ BAND_COUNT }; i < end; ++i) {
        B[i] = 1.5 * wc[i];
    }
    B[7] *= 0.93;
    B[8] *= 0.78;
    B[9] *= 0.5067;

    return makeGeqFilterParameters(filterGains, c, B, BAND_COUNT, wc);
}

GeqFilterParameters geq3FilterParameters(FilterGains const& filterGains, std::uint32_t sampleRateHz)
{
    constexpr std::size_t BAND_COUNT{ 31 };
    auto const wc = 2 * M_PI / sampleRateHz * bandCenterFrequencies(BAND_COUNT);
    constexpr double c{ 0.4 };

    auto const B = 2 * M_PI / sampleRateHz
        * Vector{ 9.178, 11.56, 14.57, 18.36, 23.13, 29.14, 36.71, 46.25, 58.28, 73.43, 92.51,
              116.6, 146.9, 185.0, 233.1, 293.7, 370.0, 466.2, 587.4, 740.1, 932.4, 1175, 1480,
              1865, 2350, 2846, 3502, 4253, 5038, 5689, 5570 };

    return makeGeqFilterParameters(filterGains, c, B, BAND_COUNT, wc);
}

FilterGains ngeqFilterGains(
    ControlGainsDb const& controlGainsDb, NeuralNetworkParameters const& parameters) noexcept
{
    auto const& p = parameters;
    auto const g = 2.0 * hadamardDivision(controlGainsDb - p.xMin, p.xMax - p.xMin) - 1;
    auto const o = tanh(p.W1 * g + p.theta1);
    auto const go = p.W2 * o + p.theta2;
    auto const gains = hadamard((p.tMax - p.tMin), 0.5 * (go + 1.0)) + p.tMin;
    return { gains };
}

FilterGains ngeq3FilterGains(
    ControlGainsDb const& controlGainsDb, NeuralNetworkParameters const& parameters) noexcept
{
    auto const& p = parameters;
    auto const g = 2.0 * hadamardDivision(controlGainsDb - p.xMin, p.xMax - p.xMin) - 1;
    auto const o1 = tanh(p.W1 * g + p.theta1);
    auto const o2 = tanh(p.W2 * o1 + p.theta2);
    auto const go = p.W3 * o2 + p.theta3;
    auto const gains = hadamard((p.tMax - p.tMin), 0.5 * (go + 1.0)) + p.tMin;
    return { gains };
}

double levelDbFromControlGains(ControlGainsDb const& gains)
{
    auto const sum = std::accumulate(gains.cbegin(), gains.cend(), 0.);
    auto const mean = sum / gains.size();
    return -mean;
}

GraphicEqParameters makeGraphicEqParameters(GeqFilterParameters const& filterParams, double levelDb)
{
    GraphicEqParameters result;
    result.type = GraphicEqParameters::Type::cascade;
    result.a[0] = coefficients(filterParams.a, 1);
    result.a[1] = coefficients(filterParams.a, 2);
    result.b[0] = coefficients(filterParams.b, 1);
    result.b[1] = coefficients(filterParams.b, 2);
    result.stageF = GraphicEqParameters::Coefficients(result.a[0].size(), 1.F);
    result.F = static_cast<float>(filterParams.G0);
    result.levelDb = static_cast<float>(levelDb);
    return result;
}

} // namespace

GraphicEqParameters calculateCascadeEqParameters(FrequencyResponse const& inputFrequencyResponse,
    NeuralNetworkParameters const& networkParameters, std::uint32_t sampleRateHz) noexcept
{
    // We expect `inputFrequencyResponse` to describe a (linear) magnitude spectrum (in contrast to
    // EIF which gives energy values).
    auto const zeroFixedInput = fixZeroCoefficientsWithSmallestValue(inputFrequencyResponse);
    auto const bandCount = networkParameters.theta2.size();
    auto const targetFrequencyResponse = calculateFrequencyResponseAtBandCenterFrequencies(
        bandCenterFrequencies(bandCount), zeroFixedInput);
    auto const controlGainsNoOffsetDb = makeControlGains(targetFrequencyResponse);
    auto const levelDb = levelDbFromControlGains(controlGainsNoOffsetDb);
    ControlGainsDb controlGainsDb{ controlGainsNoOffsetDb + levelDb };

    if (bandCount == 31) {
        auto const filterGains = ngeq3FilterGains(controlGainsDb, networkParameters);
        auto const filterParams = geq3FilterParameters(filterGains, sampleRateHz);
        return makeGraphicEqParameters(filterParams, levelDb);
    }
    // bandCount == 10
    auto const filterGains = ngeqFilterGains(controlGainsDb, networkParameters);
    auto const filterParams = geqFilterParameters(filterGains, sampleRateHz);
    return makeGraphicEqParameters(filterParams, levelDb);
}

Vector bandCenterFrequencies(std::size_t bandCount)
{
    Vector fc;
    fc.reserve(bandCount);

    switch (bandCount) {
    case 10:
        for (std::size_t i{ 0 }, end{ bandCount }; i < end; ++i) {
            fc.push_back(16000. / (1 << (end - i - 1)));
        }
        break;
    case 31:
        for (int i{ 17 }; i >= 0; --i) {
            fc.push_back(1000. / std::pow(2., i / 3.));
        }
        for (std::size_t i{ 1 }; i <= 13; ++i) {
            fc.push_back(1000. * std::pow(2., i / 3.));
        }
        break;
    default:
        std::fill_n(fc.begin(), bandCount, 0);
        assert(false && "Unallowed band count for center frequencies");
        break;
    }

    return fc;
}

ParEqParameters pareq2Filter(double G, double GB, double wc, double B) noexcept
{
    ParEqParameters result;

    double beta{ std::tan(B / 2) };
    if (G != 1.0) {
        beta *= std::sqrt(std::abs(GB * GB - 1) / std::abs(G * G - GB * GB));
    }

    result.b0 = (1 + G * beta) / (1 + beta);
    result.num = {
        1,
        -2 * std::cos(wc) / (1 + G * beta),
        (1 - G * beta) / (1 + G * beta),
    };
    result.den = {
        1,
        -2 * std::cos(wc) / (1 + beta),
        (1 - beta) / (1 + beta),
    };

    return result;
}

} // namespace frontend
} // namespace mpegcore
