/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#define _USE_MATH_DEFINES

#include <mpegcore/matlab.hpp>

#include <algorithm>
#include <cstddef>
#include <vector>

#include <mpegcore/algorithm.hpp>

#if defined(__APPLE__)
#include <Accelerate/Accelerate.h>
#else
#include <mkl.h>
#endif

namespace mpegcore {

#if defined(__APPLE__)
using mlDouble = __CLPK_doublereal;
using mlInt = __CLPK_integer;
#else
using mlDouble = double;
using mlInt = MKL_INT;
#endif

namespace {

std::size_t determineOrder(std::vector<double> b, std::vector<double> a) noexcept
{
    return std::max(b.size(), a.size());
}

Matrix2d eye(std::size_t N)
{
    Matrix2d m(N, N, 0.0);

    for (std::size_t i{ 0 }; i < N; ++i) {
        m.data[i + i * N] = 1.0;
    }

    return m;
}

// Implement the LU solver variant (square A) mldivide Matlab function.
//
// This function takes in both the number of rows and the number of columns but checks that they are
// equal.
//
// Parameters:
//
//     A - M-by-N row-major matrix
//     B - 1-by-N row-major matrix (i.e. N-vector)
//     M - number of rows
//     N - number of columns
//
// Return:
//
//     false if an error occurred, true otherwise
//
// Post-condition:
//
//     B contains 1-by-M column-major matrix (i.e. M-vector) of the result
//
bool mldivideLU(double* A, double* B, std::size_t M, std::size_t N)
{
    if (M != N) {
        return false;
    }
    auto m = static_cast<mlInt>(N);
    auto n = static_cast<mlInt>(M);
    std::vector<mlInt> ipiv(n);
    mlInt info;

    dgetrf_(&m, &n, A, &m, ipiv.data(), &info);
    if (info < 0) {
        return false;
    }

    mlInt ONE{ 1 };
    dgetrs_("T", &m, &ONE, A, &m, ipiv.data(), B, &n, &info);
    if (info < 0) {
        return false;
    }

    return true;
}

void filterForward(std::vector<double> const& X, std::vector<double> const& a,
    std::vector<double> const& b, std::vector<double>& Z, std::vector<double>& Y)
{
    auto const orderM1 = determineOrder(b, a) - 1;
    for (std::size_t i{ 0 }; i < X.size(); ++i) {
        auto const Xi = X[i];
        auto const Yi = b[0] * Xi + Z[0];

        for (std::size_t j{ 1 }; j < orderM1; ++j) {
            Z[j - 1] = b[j] * Xi + Z[j] - a[j] * Yi;
        }

        Z[orderM1 - 1] = b[orderM1] * Xi - a[orderM1] * Yi;
        Y[i] = Yi;
    }
}

void filterBackward(std::vector<double> const& X, std::vector<double> const& a,
    std::vector<double> const& b, std::vector<double>& Z, std::vector<double>& Y)
{
    // Process signal X backwards. Other parameters are as for `filterForward`.
    auto const orderM1 = determineOrder(b, a) - 1;
    for (std::size_t i{ X.size() }; i-- > 0;) {
        auto const Xi = X[i];
        auto const Yi = b[0] * Xi + Z[0];

        for (std::size_t j{ 1 }; j < orderM1; ++j) {
            Z[j - 1] = b[j] * Xi + Z[j] - a[j] * Yi;
        }

        Z[orderM1 - 1] = b[orderM1] * Xi - a[orderM1] * Yi;
        Y[i] = Yi;
    }
}

} // namespace

std::vector<double> filtfilt(std::vector<double> b, std::vector<double> a, std::vector<double> X)
{
    std::vector<double> Y(X.size());
    auto const order = determineOrder(b, a);

    if (order < 2) {
        return Y;
    }

    auto const orderM1 = order - 1;
    a.resize(order, 0.0);
    b.resize(order, 0.0);

    auto K = eye(order - 1);
    K.setColumn(0, a.cbegin() + 1, a.cbegin() + order);
    K.data[0] += 1.0;

    for (std::size_t i{ 1 }; i < order - 1; ++i) {
        K.data[i + (i - 1) * (order - 1)] = -1.0;
    }

    std::vector<double> Kdivisor(K.m);
    for (std::size_t i{ 0 }; i < K.m; ++i) {
        Kdivisor[i] = b[1 + i] - a[1 + i] * b[0];
    }
    auto const initialConditionsOk = mldivideLU(K.data.data(), Kdivisor.data(), K.m, K.n);
    if (!initialConditionsOk) {
        return Y;
    }

    auto const x1_2 = 2 * X[0];
    auto const xf_2 = 2 * *X.crbegin();
    auto const nEdge = 3 * (order - 1);
    std::vector<double> Xi(nEdge);
    std::vector<double> Xf(nEdge);

    if (X.size() <= nEdge) {
        return Y;
    }

    // Create reflected signals at front and back to remove DC offset.
    for (std::size_t i{ 0 }; i < nEdge; ++i) {
        Xi[i] = x1_2 - X[nEdge - i];
        Xf[i] = xf_2 - X[X.size() - 2 - i];
    }

    // Filter the front reflected signal
    std::vector<double> Z(order - 1);
    for (std::size_t i{ 0 }; i < order - 1; ++i) {
        Z[i] = Kdivisor[i] * Xi[0];
    }
    filterForward(Xi, a, b, Z, Y); // Y is dummy, only care of Z

    // Filter the actual signal and the back reflected signal
    std::vector<double> Ys(Y.size());
    filterForward(X, a, b, Z, Ys);
    std::vector<double> Yf(Xi.size());
    filterForward(Xf, a, b, Z, Yf);

    // Filter the back reflected signal and the actual signal in reverse
    for (std::size_t i{ 0 }; i < order - 1; ++i) {
        Z[i] = Kdivisor[i] * Yf[nEdge - 1];
    }
    filterBackward(Yf, a, b, Z, Y); // Y is dummy, only care of Z
    filterBackward(Ys, a, b, Z, Y);

    return Y;
}

std::vector<float> filtfilt(std::vector<double> b, std::vector<double> a, std::vector<float> X)
{
    // This is only used a few times at system startup, so don't mind a couple of transforms.

    std::vector<double> x;
    transformToBack(X, x, [](auto f) { return static_cast<double>(f); });
    auto const Y = filtfilt(b, a, x);
    std::vector<float> y;
    transformToBack(Y, y, [](auto d) { return static_cast<float>(d); });

    return y;
}

} // namespace mpegcore
