/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "primes.hpp"

namespace mpegcore {

// Return all prime numbers up to `N` primes beyond `value`.
//
// This is used only at system startup and for quite small numbers, so we care more about simplicity
// than performance, and so this uses the sieve of Sundaram instead of some faster but more complex
// algorithm.
//
// For reference, an approximation of the number of primes up to some value x is x/ln(x).
// Plotting up to x = 10k (1229 primes), shows a more accurate approximation is 1.1327 * x/ln(x).
// For further reference, the numbe of primes up to x = 48k, is 4946.
//
// See https://en.wikipedia.org/wiki/Sieve_of_Sundaram
std::optional<std::vector<Prime>> primesUpToNAfter(float value, std::size_t N)
{
    std::vector<Prime> primes;

    // Given maximal gaps between primes (https://primes.utm.edu/notes/GapsTable.html) we can adjust
    // `value` so that we'll generate up to `N` primes that are at least as large as `value`.
    //
    // The cutoff point is arbitrary but since the primes are used as sample counts for delay line
    // lengths, 3842610773 would allow delays of over 22 hours (at 48000 Hz) which should be enough.

    constexpr std::size_t MAX_GAP{ 319 };
    constexpr std::size_t PRIME_AT_NEXT_MAX_GAP{ 3842610773 };
    static_assert(std::numeric_limits<Prime>::max() > PRIME_AT_NEXT_MAX_GAP);
    auto const maxValue = value + MAX_GAP * (N + 1);
    if (maxValue > PRIME_AT_NEXT_MAX_GAP) {
        return {};
    }

    auto k = static_cast<std::size_t>((std::ceil(maxValue) - 2) / 2);
    std::vector<std::uint8_t> isPrime(k + 1, true);

    // The main sieve
    for (std::size_t i{ 1 }; i <= k; ++i) {
        auto j = i;
        while (i + j + 2 * i * j <= k) {
            isPrime[i + j + 2 * i * j] = false;
            ++j;
        }
    }

    // Collect up to the desired number of primes. We reserve (much) more but this way we don't have
    // to calculate the number of primes before.

    primes.reserve(isPrime.size());
    primes.push_back(2);
    std::size_t primeCountAtValue{ 0 };
    std::size_t primeCount{ primes.size() };

    for (std::size_t i{ 1 }; i < isPrime.size(); ++i) {
        if (isPrime[i]) {
            auto const prime = 2 * i + 1;
            primes.push_back(static_cast<Prime>(prime));
            ++primeCount;

            if (primeCountAtValue == 0 && prime >= value) {
                primeCountAtValue = primeCount;
            }

            if (primeCountAtValue > 0 && primeCount == primeCountAtValue + N - 1) {
                break;
            }
        }
    }

    if (primeCountAtValue == 0 || primeCount != primeCountAtValue + N - 1) {
        return {};
    }

    return primes;
}

std::optional<std::vector<Prime>> convertToMutuallyPrimeIntegers(std::vector<double> values)
{
    auto N = values.size();
    std::vector<Prime> primed(N);
    auto const maxValue = *std::max_element(cbegin(values), cend(values));
    auto const primesMaybe = primesUpToNAfter(maxValue, N);

    if (!primesMaybe.has_value()) {
        return {}; // no primes found in requested range
    }
    auto primes = primesMaybe.value();

    // Want N primes above the max value to be sure that we'll get one for each value. Not likely to
    // happen, so this is just being very careful.
    if (*(primes.crbegin() + N - 1) < maxValue) {
        return {};
    }

    std::vector<double> absDiffs;
    // Slightly broken algorithm in that we don't necessarily get the absolutely closest prime for
    // each value because we remove used primes from the available set of primes. Thus a later value
    // could have been closer to a prime that was already used. This is also quite a slow algorithm
    // but it's simpler than a more focused search would have been. Despite these issues this should
    // be ok for the current purposes, with relatively few primes and values, and only being used at
    // the system setup time.
    for (std::size_t i{ 0 }, end{ values.size() }; i < end; ++i) {
        auto const v = values[i];
        transformToBack(primes, absDiffs, [&](auto p) { return std::abs(p - v); });
        auto const minIt = std::min_element(cbegin(absDiffs), cend(absDiffs));
        auto const minIndex = minIt - cbegin(absDiffs);
        primed[i] = primes[minIndex];
        primes[minIndex] = std::numeric_limits<Prime>::max();
        absDiffs.clear();
    }
    return primed;
}

// Similar to convertToMutuallyPrimeIntegers, but a mapping is made from the available primes in the
// requested range, to the input values, not the other way around. Each prime will have a nearest
// input value. An input value may be the nearest to multiple primes. The returned value
// corresponding to that input is the nearest among those primes. If an input value is not nearest
// to any primes, it is unchanged (rare).
std::optional<std::vector<Prime>> mapToPrimeIntegers(std::vector<double> values)
{
    auto N = values.size();
    std::vector<Prime> primed(N);
    auto const maxValue = *std::max_element(cbegin(values), cend(values));
    // return all primes up to the max input value, padded by 3 primes (approx) beyond the max just
    // in case
    auto const primesOptional = primesUpToNAfter(maxValue, 3);

    if (!primesOptional.has_value()) {
        return {}; // no primes found in requested range
    }
    const std::vector<Prime>& primes = *primesOptional;
    std::size_t numAvailablePrimes = primes.size();

    std::vector<float> absDiffs(N);
    std::vector<std::size_t> nearestInValIdc(numAvailablePrimes);

    // For each available prime, find the (index of the) input value closest to it.
    for (std::size_t i{ 0 }, end{ numAvailablePrimes }; i < end; ++i) {
        Prime prime = primes[i];
        for (std::size_t j{ 0 }, end{ N }; j < end; ++j) {
            absDiffs[j] = std::abs(prime - values[j]);
        }

        auto const minIt = std::min_element(cbegin(absDiffs), cend(absDiffs));
        auto const minIdx = minIt - cbegin(absDiffs);
        nearestInValIdc[i] = minIdx;
    }

    // For each input value for which it is the nearest to one or more prime integers,
    // set it's corresponding return prime value to the nearest of those prime integers
    for (std::size_t i{ 0 }, end{ N }; i < end; ++i) {
        auto const val = values[i];
        primed[i] = val;
        float minDistToVal = std::numeric_limits<float>::max();
        for (std::size_t j{ 0 }, end{ numAvailablePrimes }; j < end; ++j) {
            if (nearestInValIdc[j] == i) {
                float distToVal = std::abs(val - primes[j]);
                if (distToVal < minDistToVal) {
                    primed[i] = primes[j];
                    minDistToVal = distToVal;
                }
            }
        }
    }

    return primed;
}

} // namespace mpegcore
