/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.
***************************************************************************/

#include "processing.hpp"

#include <algorithm>
#include <cassert>

#include <mpegcore/AudioBufferArray.hpp>

namespace mpegcore {

namespace {

constexpr float fadePartial(
    AudioBuffer& buffer, std::size_t begin, std::size_t end, float multiplier, float step) noexcept
{
    for (std::size_t index{ begin }; index < end; ++index) {
        buffer[index] *= multiplier;
        multiplier += step;
    }
    return multiplier;
}

} // namespace

void processGain(mpegcore::AudioBuffer& buffer, mpegcore::GainLinear gain)
{
    auto const linear = gain.value;
    for (std::size_t i{ 0 }; i < buffer.size(); i++) {
        buffer[i] *= linear;
    }
}

void processMute(mpegcore::AudioBuffer& buffer)
{
    std::fill(buffer.begin(), buffer.end(), 0.F);
}

void add(const AudioBufferArray& source, AudioBufferArray& target)
{
    assert(source.size() == target.size());
    if (source.size() == 0) {
        return;
    }
    for (std::size_t buffer{ 0 }; buffer < source.size(); buffer++) {
        add(source[buffer], target[buffer]);
    }
}

void add(const AudioBuffer& source, AudioBuffer& target)
{
    assert(source.size() == target.size());
    if (source.size() == 0) {
        return;
    }
    for (std::size_t sample{ 0 }, size(source.size()); sample < size; sample++) {
        target[sample] += source[sample];
    }
}

void fade(FadeDirection direction, AudioBuffer& buffer)
{
    fade(direction, buffer, 0, { buffer.size(), buffer.size(), 0 });
}

void fade(
    FadeDirection direction, AudioBuffer& buffer, std::size_t begin, PartialFade const& partialFade)
{
    assert(partialFade.processLength + partialFade.processStartIndex <= partialFade.totalLength);
    float const step{ (direction == FadeDirection::fadeIn ? 1.F : -1.F) / partialFade.totalLength };
    auto const firstBlockSize = std::min(partialFade.processLength, buffer.size() - begin);
    auto progress{ static_cast<float>(partialFade.processStartIndex) / partialFade.totalLength };
    float multiplier{ direction == FadeDirection::fadeIn ? progress : (1.F - progress) };

    multiplier = fadePartial(
        buffer, begin, std::min(begin + firstBlockSize, buffer.size()), multiplier, step);
    fadePartial(buffer, 0, partialFade.processLength - firstBlockSize, multiplier, step);
}

void toMono(AudioBufferArray& bufferArray)
{
    auto const bufferCount = bufferArray.size();
    if (bufferCount <= 1) {
        return;
    }
    auto const bufferSize = bufferArray[0].size();
    for (std::size_t sampleIndex{ 0 }; sampleIndex < bufferSize; sampleIndex++) {
        float value{ 0.F };
        for (auto&& buffer : bufferArray) {
            value += buffer[sampleIndex];
        }
        value /= bufferCount;
        for (auto&& buffer : bufferArray) {
            buffer[sampleIndex] = value;
        }
    }
}

} // namespace mpegcore
