/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.

This software module derives its structure, with substantial modification of
functionality, from the HO-SIRR library released under the following license:

BSD 3-Clause License

Copyright (c) 2019, Leo McCormack
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
***************************************************************************/

#ifndef __RIRLIB_H_INCLUDED__
#define __RIRLIB_H_INCLUDED__

#include <stdarg.h>
#include <stdio.h>

// Define logging levels
#define LOG_LEVEL_NONE 0
#define LOG_LEVEL_DATA 1
#define LOG_LEVEL_DEBUG1 2
#define LOG_LEVEL_DEBUG2 3

// Define default logging level
#ifndef LOG_LEVEL_RUNTIME
#define LOG_LEVEL_RUNTIME LOG_LEVEL_NONE
#endif

// Define common logging macro
#if defined(__cplusplus)
#include <iostream>
#define LIBLOG(level, msg, ...)                                                                    \
    do {                                                                                           \
        if (level <= LOG_LEVEL_RUNTIME) {                                                          \
            std::cout << msg << std::endl;                                                         \
        }                                                                                          \
    } while (0)
#else
#define LIBLOG(level, msg, ...)                                                                    \
    do {                                                                                           \
        if (level <= LOG_LEVEL_RUNTIME) {                                                          \
            fprintf(stderr, msg "\n", ##__VA_ARGS__);                                              \
        }                                                                                          \
    } while (0)
#endif

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

    /* ========================================================================== */
    /*                             Presets + Constants                            */
    /* ========================================================================== */

    // Available static beamforming approaches
    typedef enum
    {
        CARDIOID_BEAM = 1, /* cardioid */
        HYPERCARDIOID_BEAM, /* hyper-cardioid */
        MAX_EV_BEAM /* hyper-cardioid with max_rE weighting */
    } BEAM_TYPES;

    typedef enum _CH_ORDER
    {
        ACN_ORDER = 1, // Ambisonic Channel Numbering (ACN)
        FUMA_ORDER // Furse-Malham/B-format (WXYZ) (not supported)
    } CH_ORDERING;

// Number of channel ordering options
#define RIRLIB_NUM_CH_ORDERINGS (2)

    typedef enum _NORM_TYPES
    {
        N3D_NORM = 1, // orthonormalised (N3D)
        SN3D_NORM, // Schmidt semi-normalisation (SN3D)
        FUMA_NORM // Furse-Malham/B-format (not supported)
    } NORMALIZATION_TYPES;

// Number of normalisation options
#define RIRLIB_NUM_NORM_TYPES (3)

    /**
     * Available SPHERICAL DESIGN presets
     */
    typedef enum _SPHDESIGN_PRESETS
    {
        SPHDESIGN_PRESET_DEFAULT = 1,
        SPHDESIGN_PRESET_T_DESIGN_4,
        SPHDESIGN_PRESET_T_DESIGN_12,
        SPHDESIGN_PRESET_T_DESIGN_15,
        SPHDESIGN_PRESET_T_DESIGN_FWPAIR_15,
        SPHDESIGN_PRESET_SPH_COVER_15,
        SPHDESIGN_MPEGFDN_15,
        SPHDESIGN_PRESET_T_DESIGN_16,
        SPHDESIGN_PRESET_T_DESIGN_FWPAIR_16,
        SPHDESIGN_PRESET_SPH_COVER_16,
        SPHDESIGN_PRESET_T_DESIGN_24,
        SPHDESIGN_PRESET_T_DESIGN_31,
        SPHDESIGN_PRESET_T_DESIGN_FWPAIR_31,
        SPHDESIGN_PRESET_T_DESIGN_32,
        SPHDESIGN_PRESET_T_DESIGN_FWPAIR_32,
        SPHDESIGN_PRESET_T_DESIGN_36,
        SPHDESIGN_PRESET_T_DESIGN_48,
        SPHDESIGN_PRESET_T_DESIGN_60,
        SPHDESIGN_PRESET_T_DESIGN_63,
        SPHDESIGN_PRESET_T_DESIGN_FWPAIR_63,
        SPHDESIGN_PRESET_T_DESIGN_64,
        SPHDESIGN_PRESET_T_DESIGN_FWPAIR_64

    } SPHDESIGN_PRESETS;

#define RIRLIB_MAX_SH_ORDER (4)

    // SRIR analysis stages
    typedef enum
    {
        RIR_NOT_LOADED = 1, /* No RIR is set/loaded */
        RIR_LOADED, /* RIR is read into member buffer */
        FILTERS_INTITIALIZED, /* Filterbank is initialized */
        ANALYSIS_BUFS_LOADED, /* Staging buffers for intermediate processing are loaded */
        BANDS_SPLIT, /* RIR is split into bands */
        DIRECT_ONSETS_FOUND, /* First arrival  position is found */
        DIFFUSENESS_ONSET_FOUND, /* T60s are calculated for each beam/band */
        EDC_OMNI_DONE, /* EDCs are calculated for each band of the omni channel */
        T60_OMNI_DONE, /* T60s are calculated for each band of the omni channel */
        RDR_DONE, /* RDR is calculated for each band (omni) */
        BEAMFORMED, /* SHD signal has been beamformed into directional signals (by band)  */
        EDC_DIR_DONE, /* EDCs are calculated for each beam/band */
        DIRGAIN_DONE /* Directional gain has been calculated between FDN and RIR channels/bands */
    } ANALYSIS_STAGE;

    // Logging level
    typedef enum _LOG_LEVEL
    {
        LGDATA = 1, /* Output results from analysis */
        LGDBG1, /* Output basic debugging */
        LGDBG2 /* Output verbose debugging */
    } LOG_LEVEL;

#define RIRLIB_MAX_NUM_OUTPUTS (64)

    /* Main Functions */

    // phRL (&) address of rirlib handle
    void rirlib_create(void** const phRL, SPHDESIGN_PRESETS sphDesignPreset);
    void rirlib_destroy(void** const phRL);

    /* Set/calc Functions */

    void rirlib_setRIR(
        void* const hRL, const float** H, int nChan, int nSamp, int sampleRate, int inNormInt);
    void rirlib_initBandProcessing(void* const hRL, ANALYSIS_STAGE thisStage);
    void rirlib_allocProcBufs(void* const hRL, ANALYSIS_STAGE thisStage);
    void rirlib_setUninitialized(void* const hRL);
    void rirlib_processRIR(void* const hRL, ANALYSIS_STAGE endStage);
    void rirlib_splitBands(void* const hRL, float** const shSig, float*** const shSig_bnd_sh,
        int removeFiltDelay, ANALYSIS_STAGE thisStage);
    void rirlib_setDirectOnsetIndices(void* const hRL, float* const sigOmniBrdbnd,
        float*** const shSig_bnd_sh, const float thresh_dB, ANALYSIS_STAGE thisStage);
    void rirlib_setDiffuseOnsetIndex_shd(void* const hRL, float** const shSig,
        const float thresh_fac, const int nWin_smooth, ANALYSIS_STAGE thisStage);
    void rirlib_setDiffuseOnsetIndex_mono(void* const hRL, float* const sig, const float thresh_fac,
        const int winSize, const int hopSize, const int startSamp, ANALYSIS_STAGE thisStage);
    void rirlib_calcRDR(void* const hRL, float*** const shSig_bnd_sh, float* const rdr_bnd,
        const int nBand, const int nSamp, const int diffuseOnsetIdx, int* const directOnsetIdx_bnd,
        float* const t60Omni_bnd, const int srcDirectivityFlag, int* const validSlope_bnd,
        ANALYSIS_STAGE thisStage);
    void rirlib_calcEDC_omni(void* const hRL, float*** const shSig_bnd_sh,
        float** const edcOmni_bnd, const int nBand, const int nSamp, ANALYSIS_STAGE thisStage);
    void rirlib_calcT60_omni(void* const hRL, float** const edc_bnd, float* const t60_bnd,
        int** const measureBoundsBuf, int* const validSlopeBoolBuf, const int nBand,
        const int nSamp, const int startBoundLow_idx, const int startBoundHigh_idx,
        const int nStartBoundSteps, const float spanBoundLow_db, const float spanBoundHigh_db,
        const float spanBoundStep_db, ANALYSIS_STAGE thisStage);

    /* Helpers */

    int getDirectOnset_1ch(
        const float* const sig, float* const tmp, const float thresh_dB, const int nSamp);
    void rirlib_calcEDC_1ch(float* const edc, const int nSamp);
    int rirlib_findDecayBounds(float* const edc, const int beginIdx, const int nSamp,
        const float start_db, const float span_db, int* const stEndIdc);
    float rirlib_T60_lineFit(float* const edc, float* x_slopeBuf, float* y_edc0mBuf,
        float* stageBuf, const int startIdx, const int endIdx, float* fitCorr, const float fs);
    int rirlib_firstIndexLessThan(float* vec, int startIdx, int endIdx, float thresh);
    int rirlib_firstIndexGreaterThan(float* vec, int startIdx, int endIdx, float thresh);
    void rirlib_setSrcPosition(void* const hRL, const float x, const float y, const float z);
    void rirlib_setRecPosition(void* const hRL, const float x, const float y, const float z);
    float rirlib_getSrcRecDistance(void* const hRL);
    void checkProperProcessingOrder(
        void* const hRL, ANALYSIS_STAGE currentStage, const char* funcName, LOG_LEVEL logLevel);
    void rirlib_validateOnsetTiming(void* const hRL);

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#endif /* __RIRLIB_H_INCLUDED__ */
