/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.

This software module derives its structure, with substantial modification of
functionality, from the HO-SIRR library released under the following license:

BSD 3-Clause License

Copyright (c) 2019, Leo McCormack
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
***************************************************************************/

#include "rirlib_internal.h"

void rirlib_create(void** const phRL, SPHDESIGN_PRESETS sphDesignPreset)
{
    rirlib_data* pData = (rirlib_data*)malloc1d(sizeof(rirlib_data));
    *phRL = (void*)pData;

    // depend only on nDir
    pData->encBeamCoeffs = NULL; // nSH x nDir
    pData->decBeamCoeffs = NULL; // nDir x nSH
    pData->dirGainDB_bnd_dir = NULL; // nDir x nBand
    pData->t60Omni_bnd = NULL; // nBand x 1
    pData->rdr_bnd = NULL; // nBand x 1
    pData->directOnsetIdx_bnd = NULL; // nBand x 1
    pData->validSlope_bnd = NULL; // nBand x 1
    pData->t60Bounds_bnd = NULL; // nBand x 2

    // depend on output design (nDir) AND input RIR (nSamp)
    pData->rir_sh = NULL; // nSH x nSamp
    pData->rir_bnd_sh = NULL; // nBand x nSH x nSamp
    pData->rir_bnd_dir = NULL; // nBand x nDir x nSamp
    pData->edc_bnd = NULL; // nBand x nSamp
    pData->edc_bnd_dir = NULL; // nBand x nDir x nSamp

    pData->H_bandFilt = NULL; // nBand x filtOrder+1
    pData->centerFreqs = NULL; // 1 x nBand
    pData->xOverFreqs = NULL; // 1 x nBand-1
    pData->srcDirectivity_bnd = NULL; // 1 x nBand

    /* Constants, defaults */

    // If diffuseness never crosses this threshold, diffuse onset defaults to
    // direct onset +5ms.
    pData->diffuseMin = 0.3f;
    pData->diffuseOnsetFallbackDelay = 0.015;
    for (int i = 0; i < 3; i++) {
        pData->srcPosition[i] = 0.f;
        pData->recPosition[i] = 0.f;
    }
    pData->srcPosition[0] = 1.f; // source defaults to 1m in front of receiver
    pData->srcDirectivityFlag = 1; // default to "loudspeaker directivity"

    // Zero out the state of buffer resources
    rirlib_setUninitialized(pData);

    // Beam shape for analysis decomposition
    pData->beamType = HYPERCARDIOID_BEAM;

    // Initialize spherical design for directional analysis
    loadSphDesignPreset(sphDesignPreset, pData->sphDirsDeg, &(pData->nDir));

    pData->windowLength = DEFAULT_WINDOW_LENGTH;

    /* NOTE: Don't initialize filters yet. RIR must first be set for
       the filterbank constants (fs) set in initBandFilters(). */
}

void rirlib_destroy(void** const phRL)
{
    rirlib_data* pData = (rirlib_data*)(*phRL);

    if (pData != NULL) {
        // depend only on output design (nDir)
        free(pData->encBeamCoeffs);
        free(pData->decBeamCoeffs);
        free(pData->dirGainDB_bnd_dir);
        free(pData->t60Omni_bnd);
        free(pData->rdr_bnd);
        free(pData->directOnsetIdx_bnd);
        free(pData->validSlope_bnd);
        free(pData->t60Bounds_bnd);
        // depend on output design (nDir) AND input RIR (nSamp)
        free(pData->rir_sh);
        free(pData->rir_bnd_sh);
        free(pData->rir_bnd_dir);
        free(pData->edc_bnd);
        free(pData->edc_bnd_dir);
        // unchanging after init
        free(pData->H_bandFilt);
        free(pData->centerFreqs);
        free(pData->xOverFreqs);
        free(pData->srcDirectivity_bnd);

        free(pData);
        pData = NULL;
    }
}

void rirlib_setInputNorm(void* const hRL, int newType)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    if ((NORMALIZATION_TYPES)newType != FUMA_NORM) // FUMA not currently supported
    {
        if (newType < 4) { // only 3 input norm types
            pData->inputNorm = (NORMALIZATION_TYPES)newType;
        } else {
            rirlib_print_error("Input norm specification is outside the valid range (1..3).");
        }
    } else {
        rirlib_print_warning("FUMA is not currently supported as an input format.");
        return;
    }
}

void rirlib_setRIR(void* const hRL, const float** H, int nChan, int nSamp, int sampleRate,
    int inNormInt // N3D, SN3D, FUMA : 1, 2, 3
)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    // Check channel count to see if input is actually in the SHD
    if (fabsf(sqrtf((float)nChan) - floorf(sqrtf((float)nChan))) > 0.0001f) {
        rirlib_print_error(
            "setRIR: Based on numChannels, input file doesn't appear to be a SHD file.");
    }

    pData->nSH = nChan;
    pData->nSamp = nSamp;
    pData->shOrder = RIRLIB_MIN(sqrt(nChan - 1), RIRLIB_MAX_SH_ORDER);
    pData->fs = (float)sampleRate;
    pData->duration = nSamp / (float)sampleRate;

    NORMALIZATION_TYPES inNorm = (NORMALIZATION_TYPES)inNormInt;
    rirlib_setInputNorm(pData, inNorm);

    // (Re)alloc and copy in input RIR
    pData->rir_sh = (float**)realloc2d((void**)pData->rir_sh, nChan, nSamp, sizeof(float));

    // Copy input file in and convert to N3D if needed
    switch (pData->inputNorm) {
    case N3D_NORM: // Already in N3D, just copy it in
        LIBLOG(LGDATA, "Processing N3D");

        for (int i = 0; i < nChan; i++)
            utility_svvcopy(H[i], nSamp, pData->rir_sh[i]);
        break;

    case SN3D_NORM: // Convert SN3D to N3D
        LIBLOG(LGDATA, "Processing SN3D");

        for (int n = 0; n < pData->shOrder + 1; n++) {
            int numOrderChans = n * 2 + 1;
            int orderBaseIdx = n * n;
            for (int i = 0; i < numOrderChans; i++) {
                float orderScale = sqrtf(2.0f * (float)n + 1.0f);
                utility_svsmul((float*)H[orderBaseIdx + i], &orderScale, nSamp,
                    pData->rir_sh[orderBaseIdx + i]);
            }
        }
        break;

    case FUMA_NORM: // Not supported
        rirlib_print_error("FUMA_NORM is not supported as an input format");
        break;
    }
    pData->analysisStage = RIR_LOADED;

    // Initialize band-processing filters
    rirlib_initBandProcessing(pData, FILTERS_INTITIALIZED);

    // Alloc processing resources
    rirlib_allocProcBufs(pData, ANALYSIS_BUFS_LOADED);
}

void rirlib_setSrcPosition(void* const hRL, const float x, const float y, const float z)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    float* sPos = pData->srcPosition;
    sPos[0] = x;
    sPos[1] = y;
    sPos[2] = z;
}

void rirlib_setRecPosition(void* const hRL, const float x, const float y, const float z)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    float* rPos = pData->recPosition;
    rPos[0] = x;
    rPos[1] = y;
    rPos[2] = z;
}

float rirlib_getSrcRecDistance(void* const hRL)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    float* rPos = pData->recPosition;
    float* sPos = pData->srcPosition;
    float sqDiff[3];
    for (int i = 0; i < 3; i++) {
        float diff = sPos[i] - rPos[i];
        sqDiff[i] = diff * diff;
    }
    return sqrtf(sumf(sqDiff, 3));
}

void rirlib_setUninitialized(void* const hRL)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    pData->nSH = -1;
    pData->nSamp = -1;
    pData->shOrder = -1;
    pData->fs = -1.f;
    pData->directOnsetIdx_brdbnd = -1;
    pData->diffuseOnsetIdx = -1;
    pData->diffuseOnsetSec = 0.f;
    pData->t0Sec = -1;
    pData->t0Idx = 0;
    pData->duration = 0.0f;
    pData->analysisStage = RIR_NOT_LOADED;
}

void rirlib_initBandProcessing(void* const hRL, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDBG1);

    /* Initialize octave band filters
     * bandCenterFreqs are "center freqs" used to determine crossover freqs.
     * The lowest and highest bands are LP and HP filters, "center freq" for
     * those is a bit of a misnomer. Xover freqs are center freqs * sqrt(2)
     * (octave bands), see rirlib_initBandProcessing().
     * LPF cutoff is 125 * sqrt(2) = 177 Hz.
     * HPF cuton is 8k * sqrt(2) = 11313 Hz.
     * Must be the same size as pData->nBand. */
    pData->nBand = 8;
    pData->bandFiltOrder = 400;
    float bandCenterFreqs[8] = { 125.f, 250.f, 500.f, 1000.f, 2000.f, 4000.f, 8000.f, 16000.f };

    // Bandwise directivity factor of a loudspeaker (Genelec_8351A) calculated via SHD with
    // https://github.com/AppliedAcousticsChalmers/sound-source-directivities
    float srcLSDirectivity[8] = { 1.1216f, 1.8731f, 3.6854f, 5.4908f, 5.1385f, 6.8701f, 5.1735f,
        6.3951f };

    // Memory alloc
    pData->centerFreqs = malloc1d(pData->nBand * sizeof(float));
    pData->xOverFreqs = malloc1d((pData->nBand - 1) * sizeof(float));
    pData->srcDirectivity_bnd = malloc1d(pData->nBand * sizeof(float));

    // Populate the member vars for external access
    utility_svvcopy(bandCenterFreqs, pData->nBand, pData->centerFreqs);
    utility_svvcopy(srcLSDirectivity, pData->nBand, pData->srcDirectivity_bnd);

    // Set xover freqs
    for (int ib = 0; ib < pData->nBand - 1; ib++) {
        pData->xOverFreqs[ib] = bandCenterFreqs[ib] * sqrtf(2.f);
        // LIBLOG(LGDATA, "\tXOver band %d %.1f", ib, pData->bandXOverFreqs[ib]);
    }

    // Create the filterbank
    if (pData->H_bandFilt == NULL) { // if this is the first time this function is called ...
        // Allocate filter coefficients (nBand x bandFiltOrder + 1)
        pData->H_bandFilt = (float**)realloc2d(
            (void**)pData->H_bandFilt, pData->nBand, pData->bandFiltOrder + 1, sizeof(float));

        // Compute FIR Filterbank coefficients
        rirlib_FIRFilterbank(pData->bandFiltOrder, pData->xOverFreqs, pData->nBand - 1, pData->fs,
            WINDOWING_FUNCTION_HAMMING, 1, FLATTEN2D(pData->H_bandFilt));
    }

    pData->analysisStage = thisStage;
}

// (re)allocate the buffers used for storing intermediate processing data
void rirlib_allocProcBufs(void* const hRL, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDBG1);

    const int nSH = pData->nSH;
    const int nDir = pData->nDir;
    const int nBand = pData->nBand;
    const int nSamp = pData->nSamp;

    // Optim: these don't all need to be allocated in the case of a mono RIR

    // These members depend on output design (nDir) AND input RIR (nSH, nSamp)
    pData->encBeamCoeffs =
        (float**)realloc2d((void**)pData->encBeamCoeffs, nSH, nDir, sizeof(float));
    pData->decBeamCoeffs =
        (float**)realloc2d((void**)pData->decBeamCoeffs, nDir, nSH, sizeof(float));
    pData->rir_bnd_sh =
        (float***)realloc3d((void***)pData->rir_bnd_sh, nBand, nSH, nSamp, sizeof(float));
    pData->rir_bnd_dir =
        (float***)realloc3d((void***)pData->rir_bnd_dir, nBand, nDir, nSamp, sizeof(float));
    pData->edc_bnd = (float**)realloc2d((void**)pData->edc_bnd, nBand, nSamp, sizeof(float));
    pData->edc_bnd_dir =
        (float***)realloc3d((void***)pData->edc_bnd_dir, nBand, nDir, nSamp, sizeof(float));

    /* (Re)allocate buffers that depend only on the number of output channels and bands
     * Optim: These could alternatively only be reallocated when the out design
     * changes, but it's not heavy and it's more concise */
    pData->dirGainDB_bnd_dir =
        (float**)realloc2d((void**)pData->dirGainDB_bnd_dir, nBand, nDir, sizeof(float));
    pData->t60Omni_bnd = (float*)realloc1d((void*)pData->t60Omni_bnd, nBand * sizeof(float));
    pData->rdr_bnd = (float*)realloc1d((void*)pData->rdr_bnd, nBand * sizeof(float));
    pData->directOnsetIdx_bnd =
        (int*)realloc1d((void*)pData->directOnsetIdx_bnd, nBand * sizeof(int));
    pData->validSlope_bnd = (int*)realloc1d((void*)pData->validSlope_bnd, nBand * sizeof(int));
    pData->t60Bounds_bnd = (int**)realloc2d((void**)pData->t60Bounds_bnd, nBand, 2, sizeof(int));

    pData->analysisStage = thisStage;
}

void rirlib_processRIR(void* const hRL, ANALYSIS_STAGE endStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    ANALYSIS_STAGE requiredPreviousStage = ANALYSIS_BUFS_LOADED;
    checkProperProcessingOrder(pData, requiredPreviousStage, __func__, LGDBG1);

    // TODO: set constants elsewhere?
    // Direct arrival onset threashold (dB below omni peak)
    const float directOnsetThreshDb = -6.f;
    // Diffuse onset threashold (normalized scalar below diffuseness peak)
    const float diffuseOnsetThresh_shd = 0.7071f;
    // Diffuse onset threashold (normalized scalar below diffuseness peak)
    const float diffuseOnsetThresh_mono = 0.7071f;
    // Number of analysis frames to smooth diffuseness value (framesize: 128, hop: 64)
    const int nWin_smooth = 2;
    // Maximum ajustment of directional gain allowed (absolute value)
    const int maxDirGainAdjustment = 12;

    const int nBand = pData->nBand;
    const int nDir = pData->nDir;
    const int nSamp = pData->nSamp;
    const float fs = pData->fs;

    /* Omni, bandwise analysis */

    if (pData->analysisStage + 1 > endStage)
        return;

    rirlib_splitBands(pData, pData->rir_sh, pData->rir_bnd_sh, 1, BANDS_SPLIT);
    if (pData->analysisStage + 1 > endStage)
        return;

    rirlib_setDirectOnsetIndices(
        pData, &pData->rir_sh[0][0], pData->rir_bnd_sh, directOnsetThreshDb, DIRECT_ONSETS_FOUND);
    if (pData->analysisStage + 1 > endStage)
        return;

    if (pData->shOrder > 0) {
        // SHD SRIR
        rirlib_setDiffuseOnsetIndex_shd(
            pData, pData->rir_sh, diffuseOnsetThresh_shd, nWin_smooth, DIFFUSENESS_ONSET_FOUND);
    } else {
        // Mono RIR
        rirlib_setDiffuseOnsetIndex_mono(pData,
            &pData->rir_sh[0][0], // ptr to head of first channel (0th-order shd (2D) signal)
            diffuseOnsetThresh_mono, 1024, 64, // window size, hop size
            pData->directOnsetIdx_brdbnd, // search start sample (center of first window)
            DIFFUSENESS_ONSET_FOUND);
    }
    // After calculating diffuse onset, perform sanity checks for all onsets
    rirlib_validateOnsetTiming(pData);
    if (pData->analysisStage + 1 > endStage)
        return;

    rirlib_calcEDC_omni(pData, pData->rir_bnd_sh, pData->edc_bnd, nBand, nSamp, EDC_OMNI_DONE);
    if (pData->analysisStage + 1 > endStage)
        return;

    // T60 measurement bounds parameters
    // NOTE: requires directOnsetIdx_brdbnd and diffuseOnsetIdx to have been calculated!
    const int directOnsetIdx_brdbnd = pData->directOnsetIdx_brdbnd;
    const int diffuseOnsetIdx = pData->diffuseOnsetIdx;
    const int startBoundLow_idx = (int)(directOnsetIdx_brdbnd
        + (0.004 * pData->fs)); // earliest decay span start time, x ms after direct onset
    const int startBoundHigh_idx = pData->diffuseOnsetIdx; // latest decay span start time
    const int nStartBoundSteps = 5; // number of intermediate start times to try
    const float spanBoundLow_db = 10.f; // shortest decay span to measure
    const float spanBoundHigh_db = 60.f; // longest decay span to measure
    const float spanBoundStep_db = 2.f; // decay measurement span step size

    LIBLOG(LGDBG2,
        "\tStarting T60 measurement (values should be ascending):\n\t"
        "\tdirection onset: %d samp (%.3f sec)\n\t"
        "\tfirst t60 measurement start: %d samp (%.3f sec)\n\t"
        "\tlast t60 measurement start: %d samp (%.3f sec)\n\t"
        "\tdiffuse onset: %d samp (%.3f sec)\n\t",
        directOnsetIdx_brdbnd, directOnsetIdx_brdbnd / fs, startBoundLow_idx,
        startBoundLow_idx / fs, startBoundHigh_idx, startBoundHigh_idx / fs, diffuseOnsetIdx,
        diffuseOnsetIdx / fs);

    rirlib_calcT60_omni(pData, pData->edc_bnd, pData->t60Omni_bnd, pData->t60Bounds_bnd,
        pData->validSlope_bnd, nBand, nSamp, startBoundLow_idx, startBoundHigh_idx,
        nStartBoundSteps, spanBoundLow_db, spanBoundHigh_db, spanBoundStep_db, T60_OMNI_DONE);

    if (pData->analysisStage + 1 > endStage)
        return;

    // Requires: omni broadband diffuse onset, omni bandwise direct onsets, omni bandwise T60s
    rirlib_calcRDR(pData, pData->rir_bnd_sh, pData->rdr_bnd, nBand, nSamp, diffuseOnsetIdx,
        pData->directOnsetIdx_bnd, pData->t60Omni_bnd, pData->srcDirectivityFlag,
        pData->validSlope_bnd, RDR_DONE);
    if (pData->analysisStage + 1 > endStage)
        return;
}

/* Find the direct onset
 * The naive approach is to consider the index of absolute max value in the
 * buffer. However, with synthetic RIRs, like from a shoebox, coincident
 * reflections may have greater magnitude than the direct arrival.
 * So this returns the index of the first value that crosses the
 * (max abs value - threshold).
 *
 * thresh_dB: Threshold (pressure, dB) below the absolute max value in the
 * buffer, above which the onset is considered to have occured. */
void rirlib_setDirectOnsetIndices(void* const hRL,
    float* const sigOmniBrdbnd, // 1 x nSamp
    float*** const shSig_bnd_sh, // nBand x nSH x nSamp
    const float thresh_dB, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDATA);

    const int nSamp = pData->nSamp;
    const int nBand = pData->nBand;
    const float fs = pData->fs;

    // Buffer to store signal absolute values(1ch at a time)
    float* const vabs_tmp = malloc1d(nSamp * sizeof(float));

    const float srcRecDist = rirlib_getSrcRecDistance(hRL);
    int directOnsetIdx = getDirectOnset_1ch(sigOmniBrdbnd, vabs_tmp, thresh_dB, nSamp);
    pData->directOnsetIdx_brdbnd = RIRLIB_MAX(directOnsetIdx, 0);

    // T0: set based on broadband direct onset
    const float t0 = ((float)pData->directOnsetIdx_brdbnd / fs) - (srcRecDist / 343.f);
    pData->t0Sec = t0; // sec
    pData->t0Idx = (int)(t0 * fs); // A negative t0Idx means t0 is before the file begins (RIR can
                                   // be trimmed ahead of the direct arrival)

    // Debug logging
    LIBLOG(LGDATA, "\tsrc->rec dist: %.3f m", srcRecDist);
    LIBLOG(LGDATA, "\tdirect onset index: %d\t(%.4f sec)", pData->directOnsetIdx_brdbnd,
        (float)pData->directOnsetIdx_brdbnd / pData->fs);
    LIBLOG(LGDATA, "\tt0 index: %d\t(%.4f sec)", pData->t0Idx, t0);
    LIBLOG(LGDBG1, "\tbandwise direct onsets:");

    // Bandwise direct onsets
    for (int ib = 0; ib < nBand; ib++) {
        directOnsetIdx =
            getDirectOnset_1ch(&shSig_bnd_sh[ib][0][0], // omni band: nBand x nSH x nSamp
                vabs_tmp, thresh_dB, nSamp);
        pData->directOnsetIdx_bnd[ib] = RIRLIB_MAX(directOnsetIdx, 0);

        LIBLOG(LGDBG1, "\t\tb%d\t%d \t(%.4f sec)", ib, pData->directOnsetIdx_bnd[ib],
            (float)pData->directOnsetIdx_bnd[ib] / pData->fs);
    }

    pData->analysisStage = thisStage;
    free(vabs_tmp);
}

// returns -1 on fail
int getDirectOnset_1ch(const float* const sig,
    float* tmp, // buffer to hold copied channel data, NULL uses local memory
    const float thresh_dB, const int nSamp)
{
    int freeLocalBuf;
    if (tmp == NULL) {
        // tmp buffer is allocated locally, free it when done
        freeLocalBuf = 1;
        tmp = malloc1d(nSamp * sizeof(float));
    } else {
        // tmp buffer has been passed in, don't free it
        freeLocalBuf = 0;
    }
    // Absolute values of the omni channel
    int maxIdx;
    utility_svabs(sig, nSamp, tmp); // abs
    utility_simaxv(tmp, nSamp, &maxIdx); // index of max(abs(omni))

    // Index of first index above threshold
    const float maxVal = tmp[maxIdx];
    const float onsetThresh = maxVal * powf(10.f, thresh_dB / 20.f);
    const int directOnsetIdx = rirlib_firstIndexGreaterThan(tmp, 0, nSamp - 1, onsetThresh);

    if (freeLocalBuf)
        free(tmp);

    return directOnsetIdx;
}

/* thresh_fac: threshold (normalized scalar) below the absolute max value in the
 *             buffer, above which the onset is considered to have occured.
 * NOTE: rirBuf_sh is expected to be ACN-N3D */
void rirlib_setDiffuseOnsetIndex_shd(void* const hRL,
    float** const shSig, // nsh x nsamp
    const float thresh_fac, const int nWin_smooth, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDATA);

    // Take a local copy of current configuration to be thread safe
    const float fs = pData->fs;
    const int nSamp = pData->nSamp;
    const int winSize = pData->windowLength;
    const int nPV = 4; // P-V channels: WXYZ only
    const int directOnsetIdx_brdbnd = pData->directOnsetIdx_brdbnd;
    const int t0Idx = pData->t0Idx;

    // Inits
    const int hopSize = winSize / 2; // time-domain hop (for input)
    const int fftSize = winSize * 2; // fftsize = 2*winsize (input is zero-padded)
    const int nBins_anl = winSize / 2 + 1; // nBins used for analysis
    const int nBins_syn = fftSize / 2 + 1; // nBins used for synthesis
    const int pad_head = winSize / 2;
    const int pad_tail = winSize * 2;
    const int lenSig_pad = pad_head + nSamp + pad_tail;
    float_complex pvCOV[4][4];

    /* Smoothing constant. A refactoring of:
     * tau          = (hopsize * nwin_smooth) / fs; <-- smoothing coefficient
     * smooth_const = expf(-1.f / (tau * fs / (float)hopsize)); <-- "decay" constant */
    const float smooth_const = expf(-1.f / (float)nWin_smooth);

    // Max freq bin to calculate diffuseness
    float nearestVal = 10e5f;
    int maxDiffFreq_idx = 0;
    for (int i = 0; i < nBins_anl; i++) {
        // bin_idx * binwidth (Hz)
        float tmp = fabsf((float)i * (fs / (float)fftSize) - MAX_DIFF_FREQ_HZ);
        // float tmp = fabsf((float)i * (fs / (float)winsize) - MAX_DIFF_FREQ_HZ);
        if (tmp < nearestVal) {
            nearestVal = tmp;
            maxDiffFreq_idx = i;
        }
    }

    // Debug logging
    LIBLOG(LGDBG1, "\tsmoothing wins: %d, coeff: %.3f", nWin_smooth, smooth_const);
    LIBLOG(LGDBG1, "\tnum bins used for diffuseness: %d / %d", maxDiffFreq_idx + 1, nBins_anl);

    float *pvir, *pvir_pad, *win, *insig_win, *diff_frames;
    float_complex *inspec_anl, *inspec_syn, *pvspec_win;
    void* hFFT_syn; // for FFT

    // Make local copy of the SH RIR, from ACN to pressure-velocity (WXYZ) ordering
    pvir = malloc1d(nPV * nSamp * sizeof(float));
    int acnOrder[4] = { 0, 3, 1, 2 };
    for (int i = 0; i < nPV; i++) {
        int reorderIdx = acnOrder[i]; // w y z x -> w x y z
        memcpy(&pvir[i * nSamp], &shSig[reorderIdx][0], nSamp * sizeof(float));
    };

    // Scale XYZ to normalized velocity from N3D
    float velScale = 1.f / sqrtf(3.f);
    utility_svsmul(&pvir[1 * nSamp], &velScale, (nPV - 1) * nSamp, &pvir[1 * nSamp]);

    // Normalise so the peak of the omni is 1
    int peakIdx;
    utility_simaxv(pvir, nSamp, &peakIdx); // index of max(abs(omni))
    float peakNorm = 1.0f / fabsf(pvir[peakIdx]);
    utility_svsmul(pvir, &peakNorm, nPV * nSamp, pvir);

    // Zero pad the signal's start and end for STFT: winsize/2 at head, winsize*2 at tail
    pvir_pad = calloc1d(nPV * lenSig_pad, sizeof(float));
    for (int i = 0; i < nPV; i++) {
        memcpy(&pvir_pad[i * lenSig_pad + pad_head], &(pvir[i * nSamp]), nSamp * sizeof(float));
    }

    // Transform window (symmetric Hann - 'hanning' in matlab)
    win = malloc1d(winSize * sizeof(float));
    for (int i = 0; i < winSize; i++)
        win[i] = powf(sinf((float)i * (M_PI / (float)winSize)), 2.0f);

    // Mem alloc for diffuseness of each window
    int nDiffFrames = (int)((nSamp + (2 * winSize)) / hopSize + 0.5f);
    diff_frames = calloc1d(nDiffFrames, sizeof(float));

    // Mem alloc for a single window of processing
    insig_win = calloc1d(fftSize, sizeof(float)); // for single-channel fft
    inspec_syn = calloc1d(nPV * nBins_syn, sizeof(float_complex)); // store 4-ch fft
    inspec_anl = calloc1d(nPV * nBins_anl, sizeof(float_complex));
    pvspec_win = malloc1d(nPV * nBins_anl * sizeof(float_complex));

    saf_rfft_create(&hFFT_syn, fftSize);

    // Initialize 'previous' intensity and energy vars such that initital diffuseness = 0
    // (appropriate in context of SRIR analysis)
    float prev_ixyz_smooth[3] = { 1.f, 0.f, 0.f };
    float prev_energy_smooth = 1.f;

    /* Window processing */

    int irIdx = 0; // sample frame index into pvir_pad, increments by hopsize
    int hopCount = 0;
    while (irIdx + winSize < nSamp + 2 * winSize) {
        // Transform to frequency domain, one channel at a time
        for (int ipv = 0; ipv < nPV; ipv++) {
            // window the input, placing it into a zero-padded buffer insig_win
            for (int j = 0; j < winSize; j++)
                insig_win[j] = win[j] * pvir_pad[(ipv * lenSig_pad) + irIdx + j];

            // full fft, put in inspec_syn
            saf_rfft_forward(hFFT_syn, insig_win, &inspec_syn[ipv * nBins_syn]);

            for (int j = 0, k = 0; j < nBins_anl; j++, k += fftSize / winSize)
                inspec_anl[(ipv * nBins_anl) + j] = inspec_syn[(ipv * nBins_syn) + k];
        }

        // Copy spectrum of windowed input channels into pvspec_win (nPV * nBins_anl)
        for (int i = 0; i < nPV; i++)
            memcpy(&pvspec_win[i * nBins_anl], &inspec_anl[i * nBins_anl],
                nBins_anl * sizeof(float_complex));

        // Compute broadband covariance matrix
        const float_complex calpha = cmplxf(1.0f, 0.0f), cbeta = cmplxf(0.0f, 0.0f);
        cblas_cgemm(CblasRowMajor, CblasNoTrans, CblasConjTrans, 4, 4, maxDiffFreq_idx + 1, &calpha,
            pvspec_win, nBins_anl, pvspec_win, nBins_anl, &cbeta, FLATTEN2D(pvCOV), 4);

        // Compute broadband intensity and energy
        float intensity[3], energy;
        for (int i = 0; i < 3; i++)
            intensity[i] = crealf(pvCOV[1 + i][0]);
        energy = 0.0f;
        for (int i = 0; i < 4; i++)
            energy += crealf(pvCOV[i][i]) * 0.5f;

        // Estimating and time averaging of broadband diffuseness
        float ixyz_smooth[3], energy_smooth, iaMag_smooth;
        iaMag_smooth = 0.0f;
        for (int i = 0; i < 3; i++) {
            ixyz_smooth[i] =
                ((1.0f - smooth_const) * intensity[i]) + (smooth_const * prev_ixyz_smooth[i]);
            prev_ixyz_smooth[i] = ixyz_smooth[i];
            iaMag_smooth += powf(ixyz_smooth[i], 2.0f);
        }
        iaMag_smooth = sqrtf(iaMag_smooth);
        energy_smooth = ((1.0f - smooth_const) * energy) + (smooth_const * prev_energy_smooth);
        prev_energy_smooth = energy_smooth;

        // Store broadband diffuseness value and advance
        diff_frames[hopCount] = 1.0f - (iaMag_smooth / (energy_smooth + 2.23e-10f));
        irIdx += hopSize;
        hopCount++;
    }

    /* Diffuse onset */

    /* Begin search for max diffuseness at the first window containing the
     * direct arrival, where we can expect low diffuseness */
    int directStartIdx = RIRLIB_MAX(directOnsetIdx_brdbnd - hopSize + pad_head, 0);
    int directHopIdx = (int)ceilf((float)directStartIdx / hopSize);
    int maxHopIdx;
    utility_simaxv(&diff_frames[directHopIdx], nDiffFrames - directHopIdx, &maxHopIdx);
    maxHopIdx += directHopIdx; // add the starting position back

    // Index of first window above threshold
    const float diffuseMax = diff_frames[maxHopIdx];
    const float onsetThresh = diffuseMax * thresh_fac;
    const int onsetHopIdx =
        rirlib_firstIndexGreaterThan(diff_frames, directHopIdx, nDiffFrames - 1, onsetThresh);

    /* NOTE: Normally we'd assume the sample onset index is in the middle of the
     * window. I.e. onsetWinIdx * hopsize + winsize/2. But because pvsig was
     * zero-padded at the head by winsize/2, we can omit that offset here. */
    int diffuseOnsetIdx = onsetHopIdx * hopSize;

    if (diffuseMax < pData->diffuseMin) {
        rirlib_print_warning("Diffuseness didn't exceed the valid threshold.\nFalling back on "
                             "directOnset + diffuseOnsetFallbackDelay.");
        diffuseOnsetIdx = directOnsetIdx_brdbnd + (int)(pData->diffuseOnsetFallbackDelay * fs);
    }

    // Diffuse onset sample index.
    // This is the integer sample position in the the analyzed buffer.
    pData->diffuseOnsetIdx = diffuseOnsetIdx;

    /* Diffuse onset time.
     * This is the acoustic property of the room. It is relative to t0, the time
     * at which the sound is emitted from the source (which can be before
     * beginning of the provided buffer, e.g. if the silence preceding the
     * direct arrival was trimmed from the recording). */
    pData->diffuseOnsetSec = ((float)diffuseOnsetIdx / fs) - pData->t0Sec;

    LIBLOG(LGDBG2, "\tbegin search hop idx: %d", directHopIdx);
    LIBLOG(LGDBG2, "\tdiff max hop idx: %d", maxHopIdx);
    LIBLOG(LGDBG1, "\tdiff max value: %.3f", diffuseMax);
    LIBLOG(LGDBG2, "\tdiff onset thresh: %.3f", onsetThresh);
    LIBLOG(LGDBG1, "\tdiff onset win idx: %d", onsetHopIdx);
    LIBLOG(LGDATA, "\tdiff onset: %d samps (%.3f sec)\n\tt0-adjusted diff onset: %.3f sec",
        diffuseOnsetIdx, (float)diffuseOnsetIdx / pData->fs, pData->diffuseOnsetSec);

    pData->analysisStage = thisStage;

    // Cleanup
    free(pvir);
    free(pvir_pad);
    free(win);
    free(diff_frames);
    free(insig_win);
    free(inspec_anl);
    free(inspec_syn);
    free(pvspec_win);
    saf_rfft_destroy(&hFFT_syn);
}

/* Diffuseness based on echo density:
 * Abel & Huang 2006, "A simple, robust measure of reverberation echo
 * density", In: Proc. of the 121st AES Convention, San Francisco */
void rirlib_setDiffuseOnsetIndex_mono(void* const hRL, float* const sig, const float thresh_fac,
    const int winSize, const int hopSize,
    const int startSamp, // center of first window
    ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDATA);

    const float fs = pData->fs;
    const int nSamp = pData->nSamp;
    const int wsize = winSize - (winSize % 2); // make sure it's even (rounds down)
    const int halfWsize = wsize / 2;
    const float erfcNorm = 1.f / 0.317310507862914f; // 1 / erfc(1/sqrt(2))
    const int nHops = (int)((nSamp - startSamp - wsize - halfWsize) / hopSize);

    if (nHops < 1)
        rirlib_print_error(
            "[setDiffuseOnsetIndex_mono] input buffer is smaller than the window size.");

    float* const window = malloc1d(wsize * sizeof(float));
    getWindowingFunction(WINDOWING_FUNCTION_HANN, wsize, window);
    float winnorm = 1.f / sumf(window, wsize);
    utility_svsmul(window, &winnorm, wsize, NULL);

    int foundOnset = 0;
    int hopIdx = 0;
    int winStart = startSamp - halfWsize; // center the window at startSamp (OK to start at 0)
    int winOffset;
    double echoDens;
    for (int ih = 0; ih < nHops; ih++) {
        winStart = winStart + hopSize;

        int thisWsize, bufOffset;
        if (winStart < 0) { // window starts at negative index
            thisWsize = wsize + winStart;
            bufOffset = 0;
            winOffset = abs(winStart);
        } else {
            thisWsize = wsize;
            bufOffset = winStart;
            winOffset = 0;
        }

        // accumulate windowed, squared signal for standard deviation
        double sum_sd = 0.f;
        for (int iw = 0; iw < thisWsize; iw++) {
            const float x = sig[bufOffset + iw];
            sum_sd += x * x * window[winOffset + iw];
        }
        double sd = sqrt(sum_sd);

        // sum window scaled by tip count for echo density
        double sum_ed = 0.f;
        for (int iw = 0; iw < thisWsize; iw++) {
            if (fabsf(sig[bufOffset + iw]) > sd)
                sum_ed += window[winOffset + iw];
        }

        // normalized echo densty
        echoDens = sum_ed * erfcNorm;

        // Debug logging
        if (LOG_LEVEL_RUNTIME >= LGDBG2) {
            LIBLOG(LGDBG2,
                "\tDiffuseness win data:\n\t\twinOffset: %d\n\t\tbufOffset: %d\n\t\tsum_sd: %.3f",
                winOffset, bufOffset, sum_sd);
            LIBLOG(LGDBG2, "\thop %d; winStart: %d, sampAt: %d, echoDens: %.3f", ih, winStart,
                winStart + halfWsize, echoDens);
        }

        if (echoDens >= thresh_fac) {
            foundOnset = 1;
            break;
        }

        hopIdx++;
    }

    int diffuseOnsetIdx;
    if (foundOnset) {
        // We're calling the onset the center of the window where the measure is
        // most "sensitive". Could alternatively linearly interp between prev
        // window peak and this one.
        diffuseOnsetIdx = winStart + halfWsize;
    } else {
        rirlib_print_warning("Echo density (diffuseness) didn't exceed the threshold. "
                             "\nFalling back on directOnset + diffuseOnsetFallbackDelay.");
        diffuseOnsetIdx =
            pData->directOnsetIdx_brdbnd + (int)(pData->diffuseOnsetFallbackDelay * fs);
    }

    // Diffuse onset sample index in the _input buffer_
    pData->diffuseOnsetIdx = diffuseOnsetIdx;

    // Diffuse onset time in the _room_, from t0, which can be before beginning
    // of input the buffer (e.g. an RIR file is trimmed to direct arrival)
    pData->diffuseOnsetSec = ((float)diffuseOnsetIdx / fs) - pData->t0Sec;

    pData->analysisStage = thisStage;
    free(window);

    // Data logging
    LIBLOG(LGDATA, "\tdiff onset: %d, %.3f sec\n\t\tt0-adjusted %.3f sec", pData->diffuseOnsetIdx,
        (float)pData->diffuseOnsetIdx / pData->fs, pData->diffuseOnsetSec);
    // Debug logging
    if (LOG_LEVEL_RUNTIME >= LGDBG2) {
        LIBLOG(LGDBG2, "\tdiff max value: %.3f", echoDens);
        LIBLOG(LGDBG2, "\tdiff onset thresh: %.3f", thresh_fac);
        LIBLOG(LGDBG2, "\tdiff onset hop idx: %d", hopIdx);
    }
}

void rirlib_validateOnsetTiming(void* const hRL)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    const int t0Idx = pData->t0Idx;
    const int directOnsetIdx_brdbnd = pData->directOnsetIdx_brdbnd;
    const int diffuseOnsetIdx = pData->diffuseOnsetIdx;
    const int nSamp = pData->nSamp;

    // A very late diffuse onset could be sign of an error, warn
    float lateOnsetWarnThresh = 0.5f; // factor of time span between direct arrival and end of file
    int diffuseOnsetMax =
        directOnsetIdx_brdbnd + (int)((nSamp - directOnsetIdx_brdbnd) * lateOnsetWarnThresh);
    if (diffuseOnsetIdx > diffuseOnsetMax) {
        rirlib_print_warning("Diffuse onset occurs very late in the RIR, detection threshold may "
                             "need to be updated.")
    }

    // Time events should be properly sequential
    if ((t0Idx >= directOnsetIdx_brdbnd) || (directOnsetIdx_brdbnd > diffuseOnsetIdx)) {
        // NOTE: could rirlib_print_error instead to throw on bad onset order
        rirlib_print_warning("Order of analyzed events isn't valid. "
                             "Should be t0Idx < directOnsetIdx < diffuseOnsetIdx.");
        LIBLOG(0, "t0Idx = %d \ndirectOnsetIdx_brdbnd = %d \ndiffuseOnsetIdx = %d", t0Idx,
            directOnsetIdx_brdbnd, diffuseOnsetIdx);

        if (directOnsetIdx_brdbnd > diffuseOnsetIdx) {
            rirlib_print_warning("Adjusting diffuseOnset to fallback value.");
            pData->diffuseOnsetIdx =
                directOnsetIdx_brdbnd + (int)(pData->diffuseOnsetFallbackDelay * pData->fs);
            pData->diffuseOnsetSec = ((float)(pData->diffuseOnsetIdx) / pData->fs) - pData->t0Sec;
        }
    }

    // Diffuse onset from t0 should be positive
    if (pData->diffuseOnsetSec < 0)
        rirlib_print_error("Diffuse onset is negative.");
}

/* Freq-domain band filtering via FD convolution
 * removeFiltDelay : 1 will truncate the head and tail of the filtered
 *                   signal, instead of (only) the tail. So either way, the
 *                   written output will always be pData->nSamp for each channel. */
void rirlib_splitBands(void* const hRL, float** const shSig, // nSH x nSamp
    float*** const shSig_bnd_sh, int removeFiltDelay, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDBG1);

    const int nInSamp = pData->nSamp;
    const int filtOrder = pData->bandFiltOrder;

    int startCopyIdx;
    if (removeFiltDelay) {
        startCopyIdx = (int)(filtOrder / 2.f); // length of filter delay (floor)
    } else {
        startCopyIdx = 0;
    }

    /* Apply filterbank to rir_bands
     * Because of fftconv's indexing, it expects filter coefficients
     * for each channel. So we do one band/channel at a time.
     * The output length will be nSamp + filtOrder.
     * See fftconv(): y_len = x_len + h_len - 1; */
    float* stage = malloc1d((nInSamp + filtOrder) * sizeof(float));

    for (int ish = 0; ish < pData->nSH; ish++) {
        for (int ib = 0; ib < pData->nBand; ib++) {
            fftconv(&shSig[ish][0], // input: 1 channel at a time
                &pData->H_bandFilt[ib][0], // band filter coeffs
                nInSamp, // input length
                filtOrder + 1, // filter length
                1, // 1 channel at a time
                stage); // write to staging buffer

            // Copy staging buffer to out
            memcpy((void*)&shSig_bnd_sh[ib][ish][0], &stage[startCopyIdx], nInSamp * sizeof(float));
        }
    }

    pData->analysisStage = thisStage;
    free(stage);
}

// Calc bandwise RDR of the omni channel
void rirlib_calcRDR(void* const hRL,
    float*** const shSig_bnd_sh, // nband x nsh x nsamp
    float* const rdr_bnd, const int nBand, const int nSamp, const int diffuseOnsetIdx,
    int* const directOnsetIdx_bnd, float* const t60Omni_bnd,
    const int srcDirectivityFlag, // 0: omni, 1: "regular" loudspeaker
    int* const validSlope_bnd, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDATA);

    // TODO: could make these configs higher-level parameters
    // Size of the direct onset window as multiple of wavelength of band center freq
    const int nCyclePerWin = 2;
    // Limit minimum window size (esp. high frequencies)
    const float minWinSizeSec = 0.0025f;
    // Where to align the window relative to the direct onset. 0.0 window
    // begins at onset, 0.5 window centered around the onset
    const float winAlignFac = 0.4f;
    // Source distance low-clipped at 25cm
    const float srcRecDist = RIRLIB_MAX(rirlib_getSrcRecDistance(hRL), 0.25);

    // Data logging
    LIBLOG(LGDATA, "\tsource directivity FLAG: %d", srcDirectivityFlag);
    // Debug logging
    if (LOG_LEVEL_RUNTIME >= LGDBG2 && srcDirectivityFlag) {
        for (int ib = 0; ib < nBand; ib++)
            LIBLOG(LGDBG2, "\tband %d: %.5f (%.4f dB)", ib, pData->srcDirectivity_bnd[ib],
                10.f * log10f(pData->srcDirectivity_bnd[ib]));
    }

    for (int ib = 0; ib < nBand; ib++) {
        rdr_bnd[ib] = 0.f; // intialize

        if (validSlope_bnd[ib]) { // proceed only if T60 estimate was successful

            /* Direct energy */
            // Measure signal energy in a rectangular window centered on the direct arrival peak

            // Winsize specified in cycles, clip to be at least minWinSizeSec
            float winSizeSec = 1.f / pData->centerFreqs[ib];
            winSizeSec = winSizeSec * nCyclePerWin;
            winSizeSec = RIRLIB_MAX(winSizeSec, minWinSizeSec);
            int winSize_direct = (int)(winSizeSec * pData->fs + 0.5f);
            const int winSize_preOnset = (int)(winSize_direct * winAlignFac);

            int winStart_direct = directOnsetIdx_bnd[ib] - winSize_preOnset;
            if (winStart_direct < 0) {
                // The direct arrival is early in the signal so that the full window can't be
                // centered on the peak, so it's either clipped at index 0 (the window shrinks) or
                // it simply starts at index zero and maintains its size. NOTE: matlab version
                // doesn't shrink the window

                // winSize_direct = winSize_direct + winStart_direct; // shrink the window
                winStart_direct = 0;
                LIBLOG(LGDATA,
                    "Direct arrival occurs early in the signal, early energy window adjusted to "
                    "start from first index (band %d)",
                    ib);
            }

            // Window end sample (exclusive)
            int winEnd_direct = winStart_direct + winSize_direct;
            if (winEnd_direct > nSamp)
                winEnd_direct = nSamp;

            double directEnergy_sum = 0.f;
            for (int is = winStart_direct; is < winEnd_direct; is++) {
                // Omni pressure from this band, scaled to bring source to 1m
                float directPress = shSig_bnd_sh[ib][0][is] * srcRecDist;
                directEnergy_sum += directPress * directPress;
            }

            /* Diffuse energy */

            // Diffuse energy measured to -30 dB time after the diffuse onset to avoid noise floor
            float decayTime_30dB = t60Omni_bnd[ib] * 0.5f;
            int winSize_diffuse = (int)(decayTime_30dB * pData->fs);
            int winStart_diffuse = RIRLIB_MAX(diffuseOnsetIdx, winEnd_direct);
            int winEnd_diffuse = RIRLIB_MIN(winStart_diffuse + winSize_diffuse, nSamp);

            // Sum omni energy post diffuseOnset
            double diffuseEnergy_sum = 0.f;
            for (int is = winStart_diffuse; is < winEnd_diffuse; is++) {
                diffuseEnergy_sum += shSig_bnd_sh[ib][0][is] * shSig_bnd_sh[ib][0][is];
            }

            /* RDR */

            // RDR for an omni source
            rdr_bnd[ib] = diffuseEnergy_sum / directEnergy_sum;

            // Apply affect of source directivity
            if (srcDirectivityFlag) {
                rdr_bnd[ib] = rdr_bnd[ib] * pData->srcDirectivity_bnd[ib];
            }

            // Data logging
            LIBLOG(LGDATA, "\tb%d rdr: %.4f (%.2f dB)", ib, rdr_bnd[ib], 10 * log10f(rdr_bnd[ib]));
            // Debug logging
            if (LOG_LEVEL_RUNTIME >= LGDBG2) {
                LIBLOG(
                    LGDBG2, "\t\tdiff / dir sum: %.5f / %.5f", diffuseEnergy_sum, directEnergy_sum);
                LIBLOG(LGDBG2, "\t\tdirect st/en : [%d  %d] : %d samp (%.1f ms), %d min",
                    winStart_direct, winEnd_direct, winEnd_direct - winStart_direct,
                    (winEnd_direct - winStart_direct) / pData->fs * 1000,
                    (int)(minWinSizeSec * pData->fs));
                LIBLOG(LGDBG2, "\t\tdiffuse st/en: [%d  %d] : %d samp (%.1f ms)", winStart_diffuse,
                    winEnd_diffuse, winEnd_diffuse - winStart_diffuse,
                    (winEnd_diffuse - winStart_diffuse) / pData->fs * 1000);
                LIBLOG(LGDBG2, "\t\tt30 %.3f, wsize diffuse (pre nSamp clip) %d", decayTime_30dB,
                    winSize_diffuse);
            }
        }
    }

    // Fill in RDR for any degenerate bands
    for (int ib = 0; ib < nBand; ib++) {
        if (!validSlope_bnd[ib]) {
            // Band was determined to be degenerate in T60 estimate, likely from noise floor
            // contamination, so extend trend of neighboring bands. Note: this method assumes stable
            // bands 2 and 3 (250, 500 Hz)
            if (ib == 0) {
                rdr_bnd[0] = rdr_bnd[1] + RIRLIB_MAX(0, (rdr_bnd[1] - rdr_bnd[2]));
            } else if (ib > 2) {
                rdr_bnd[ib] = rdr_bnd[ib - 1] - RIRLIB_MAX(0, (rdr_bnd[ib - 2] - rdr_bnd[ib - 1]));
            }
            LIBLOG(LGDATA, "\tDegenerate band (%d) rdr: %.3f", ib, rdr_bnd[0]);
        }
    }

    pData->analysisStage = thisStage;
}

// Calc bandwise EDCs of the omni channel
void rirlib_calcEDC_omni(void* const hRL,
    float*** const shSig_bnd_sh, // nband x nsh x nsamp
    float** const edcOmni_bnd, // nband x nsamp
    const int nBand, const int nSamp, ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);

    checkProperProcessingOrder(pData, thisStage, __func__, LGDBG1);

    // Copy the RIR's omni bands into EDC buffer
    for (int ib = 0; ib < nBand; ib++) {
        utility_svvcopy(&shSig_bnd_sh[ib][0][0], nSamp, &edcOmni_bnd[ib][0]);
    }

    // Calculate EDC in-place, one band at a time
    for (int ib = 0; ib < nBand; ib++) {
        rirlib_calcEDC_1ch(&edcOmni_bnd[ib][0], nSamp);
    }
    pData->analysisStage = thisStage;
}

/* Calc EDC of a single channel
 * NOTE: in-place operation, so copy data into edcBuf before calling.
 * OPTIM: vectorize */
void rirlib_calcEDC_1ch(float* const edc, const int nSamp)
{
    double sum = 0.0;
    for (int i = nSamp - 1; i >= 0; i--) {
        sum += edc[i] * edc[i]; // energy
        edc[i] = (float)(10.0 * log10(sum)); // store in dB
    }
}

// Calc T60s of bandwise omni channel, and store the selected measurement span endpoints.
// The search for the best T60 fit happens over multiple measurement spans of the EDC. Each span has
// a starting point and a decay span (dB) that determines the endpoint. Spans are generated with
// different start points, numbering nStartBoundSteps, from startBoundLow_idx to startBoundHigh_idx.
// For each start point multiple spans are generated from spanBoundLow_db to spanBoundHigh_db in
// step sizes of spanBoundStep_db. So the total number of searched spans is nStartBoundSteps *
// nSpanSteps. The final T60 is chosen from the span with the best linear fit (the highest
// correlation coefficient).
void rirlib_calcT60_omni(void* const hRL, float** const edc_bnd, float* const t60_bnd,
    int** const t60Bounds, int* const validSlope, const int nBand, const int nSamp,
    const int startBoundLow_idx, const int startBoundHigh_idx, const int nStartBoundSteps,
    const float spanBoundLow_db, const float spanBoundHigh_db, const float spanBoundStep_db,
    ANALYSIS_STAGE thisStage)
{
    rirlib_data* pData = (rirlib_data*)(hRL);
    const float fs = pData->fs;

    checkProperProcessingOrder(pData, thisStage, __func__, LGDATA);

    const int nSpanSteps = (int)((spanBoundHigh_db - spanBoundLow_db) / spanBoundStep_db);
    const int nIter = nStartBoundSteps * nSpanSteps;

    const float startBoundStep =
        (float)(startBoundHigh_idx - startBoundLow_idx) / RIRLIB_MAX((nStartBoundSteps - 1), 1);

    // Memory alloc
    float* const x_slope1 = malloc1d(nSamp * sizeof(float)); // vector with slope of 1/samp
    float* const y_edc0m = malloc1d(nSamp * sizeof(float)); // zero-mean EDC
    float* const stage = malloc1d(nSamp * sizeof(float)); // staging buffer
    // Start and end samples to measure t60 in each channel
    int*** const st_end_meas_tmp = (int***)malloc3d(nIter, nBand, 2, sizeof(int));
    int** const spanOK = (int**)malloc2d(nIter, nBand, sizeof(int));

    // Find the start and end points of the measurement
    const float start_db = 0.f; // start at 0 dB: the level of beginIdx

    for (int istart = 0; istart < nStartBoundSteps; istart++) {
        const int beginIdx = (int)(startBoundLow_idx + (istart * startBoundStep));

        for (int ispan = 0; ispan < nSpanSteps; ispan++) {
            // Evaluate spans high to low
            const float span_db = spanBoundHigh_db - (ispan * spanBoundStep_db);
            const int iterIdx = istart * nSpanSteps + ispan;

            for (int ibnd = 0; ibnd < nBand; ibnd++) {
                spanOK[iterIdx][ibnd] = rirlib_findDecayBounds(&edc_bnd[ibnd][0], beginIdx, nSamp,
                    start_db, span_db, &st_end_meas_tmp[iterIdx][ibnd][0]);
            }
        }
    }

    for (int ibnd = 0; ibnd < nBand; ibnd++) {
        float fitCorrMax = 0.f;
        int bestStartStep = 0; // for logging
        float bestSpan = spanBoundHigh_db;
        int wasValidSpan = 0;

        for (int istart = 0; istart < nStartBoundSteps; istart++) {
            for (int ispan = 0; ispan < nSpanSteps; ispan++) {
                float t60, fitCorr;
                const int iterIdx = istart * nSpanSteps + ispan;
                const int thisStartIdx = st_end_meas_tmp[iterIdx][ibnd][0];
                const int thisEndIdx = st_end_meas_tmp[iterIdx][ibnd][1];

                t60 = rirlib_T60_lineFit(&edc_bnd[ibnd][0], x_slope1, y_edc0m, stage, thisStartIdx,
                    thisEndIdx, &fitCorr, pData->fs);

                // Store t60 with highest line fit correlation
                if (fitCorr > fitCorrMax) {
                    const float thisSpan = spanBoundHigh_db - (ispan * spanBoundStep_db);
                    // For correlations over 95%, keep the one with the longest measurement span
                    if (fitCorrMax < 0.95f || thisSpan > bestSpan) {
                        t60_bnd[ibnd] = t60;
                        t60Bounds[ibnd][0] = thisStartIdx;
                        t60Bounds[ibnd][1] = thisEndIdx;
                        fitCorrMax = fitCorr;
                        bestStartStep = istart;
                        bestSpan = thisSpan;
                        wasValidSpan = spanOK[iterIdx][ibnd];
                    }
                }
            }
        }

        if (!wasValidSpan) {
            // Only warn if the chosen span was truncated
            LIBLOG(LGDATA,
                "! RIRLIB WARNING: findDecayBounds found no end point below the "
                "decay span, so resulting t60 is from start measurement index to "
                "0.85 of the remaining buffer length.");
        }

        LIBLOG(LGDATA,
            "\tb %.0f t60:  %.2f sec (max corr: %.3f @ startStep: %d, %.1f dB span from %d to %d)",
            pData->centerFreqs[ibnd], t60_bnd[ibnd], fitCorrMax, bestStartStep, bestSpan,
            t60Bounds[ibnd][0], t60Bounds[ibnd][1]);
    }

    // Error checking: first and upper bands
    // Note this approach assumes a stable second and third band (250, 500 Hz)
    for (int ibnd = 0; ibnd < nBand; ibnd++) {

        validSlope[ibnd] = 1; // default to valid slope

        // Lowest band is susceptible to noise floor contamination so check that it's reasonable,
        // not more than double the second-lowest band, limit if so
        if (ibnd == 2 && t60_bnd[0] > t60_bnd[1] * 2) {
            t60_bnd[0] = t60_bnd[1] + RIRLIB_MAX(0, (t60_bnd[1] - t60_bnd[2]));
            validSlope[0] = 0;
            rirlib_print_warning("Measured T60 for lowest band is longer than twice the next band. "
                                 "Limiting to a back projection from bands 2 and 3.");
            LIBLOG(0, "\tUpdated b0 t60:  %.3f sec", t60_bnd[0]);
        }

        // Upper bands, if a band's T60 is > 3x the previous, we've likely overestimated, so extend
        // trend of previous bands
        if (ibnd >= 2 && t60_bnd[ibnd] > t60_bnd[ibnd - 1] * 3) {
            t60_bnd[ibnd] =
                t60_bnd[ibnd - 1] - RIRLIB_MAX(0, (t60_bnd[ibnd - 2] - t60_bnd[ibnd - 1]));
            validSlope[ibnd] = 0;
            rirlib_print_warning("Measured T60 for this band is longer than 3x the previous band. "
                                 "Limiting to a forward projection from previous bands.");
            LIBLOG(0, "\tUpdated b%d t60:  %.2f sec", ibnd, t60_bnd[ibnd]);
        }

        // T60 must be positive
        if (t60_bnd[ibnd] < 0.f) {
            validSlope[ibnd] = 0;
            rirlib_print_error("T60 for this band is negative.");
        }
    }

    // Error check: warn of very long T60 estimates
    for (int ibnd = 0; ibnd < nBand; ibnd++) {
        if (t60_bnd[ibnd] * fs > nSamp * 2) {
            // Note: could also flag as invalid slope calculation
            rirlib_print_warning(
                "Measured T60 for this band is longer than twice the duration of the buffer. "
                "If this is unexpected, double check that your start_db and span_db are "
                "appropriate.");
            LIBLOG(0, "\t->(Band %d, t60 %.1f sec)\n", ibnd, t60_bnd[ibnd]);
        }
    }

    free(x_slope1);
    free(y_edc0m);
    free(stage);
    free(st_end_meas_tmp);
    free(spanOK);
    pData->analysisStage = thisStage;
}

/* Measure the T60 by the line of best fit
 * x0m: zero-mean vector of a line with a slope of 1/sample
 * y0m: vector of edc values (over measurement span), with mean removed
 *      x0m      = vec_idc - mean(vec_idc);
 *      y0m      = edc_span - mean(edc_span);
 *      dc_db    = sum(x0m .* y0m) / sum(x0m.^2); // decay constant (dB/sample)
 *      t60_meas = (-60 / dc_db) / fs;
 */
float rirlib_T60_lineFit(float* const edc, float* x_slopeBuf, float* y_edc0mBuf, float* stageBuf,
    const int startIdx, const int endIdx, float* fitCorr, const float fs)
{
    int nSamp_meas = endIdx - startIdx + 1;

    // Remove mean from EDC, within the measurement span
    float y_mean = sumf(&edc[startIdx], nSamp_meas) / nSamp_meas;
    utility_svssub(&edc[startIdx], &y_mean, nSamp_meas, y_edc0mBuf);

    // Construct a vector with a slope of 1:samp with zero mean
    float firstVal = (nSamp_meas - 1) / -2.f;
    for (int i = 0; i < nSamp_meas; i++) {
        x_slopeBuf[i] = firstVal + i;
    };

    // Covariances: x*y and x*x
    float c_xy, c_xx, c_yy, dbPerSamp;
    utility_svvmul(x_slopeBuf, y_edc0mBuf, nSamp_meas, stageBuf);
    c_xy = sumf(stageBuf, nSamp_meas);
    utility_svvmul(x_slopeBuf, x_slopeBuf, nSamp_meas, stageBuf);
    c_xx = sumf(stageBuf, nSamp_meas);

    // Correlation coefficient to rate the fit
    if (fitCorr != NULL) {
        utility_svvmul(y_edc0mBuf, y_edc0mBuf, nSamp_meas, stageBuf);
        c_yy = sumf(stageBuf, nSamp_meas);
        *fitCorr = (c_xy * c_xy) / (c_xx * c_yy);
    }
    dbPerSamp = c_xy / c_xx; // slope

    return (-60.f / dbPerSamp) / fs;
}

int rirlib_findDecayBounds(float* const edc, // 1 x nSamp
    const int beginIdx, // start search at this index
    const int nSamp, // number of samples in edcBuf
    const float start_db, // <= 0
    const float span_db, // > 0
    int* const stEndIdc // output: 1 x 2, [startIdx, endIdx]
)
{
    // Check start and end points of T60 measurement
    float edcMax = edc[beginIdx];
    LIBLOG(LGDBG2, "\tbeginIdx: %d\tbuflength: %d", beginIdx, nSamp);

    stEndIdc[0] = beginIdx;

    if (start_db < 0.f) { // no need to search if start_db is 0
        int startIdx_t60 = rirlib_firstIndexLessThan(edc, beginIdx, nSamp - 1,
            edcMax + start_db); // start_db is <= 0
        if (startIdx_t60 < 0) {
            LIBLOG(LGDATA,
                "! RIRLIB WARNING: findDecayBounds found no value below the "
                "start level, start index will be the provided beginIdx.");
        } else {
            stEndIdc[0] = startIdx_t60;
        }
    }

    int end_t60 = rirlib_firstIndexLessThan(edc, stEndIdc[0], nSamp - 1,
        edcMax + start_db - span_db); // start_db is negative, span_db is positive

    int spanSuccess;
    if (end_t60 < 0) {
        // No value found below the requested span, fall back to near the end of the EDC.
        stEndIdc[1] = stEndIdc[0] + (int)(0.85f * (nSamp - stEndIdc[0]));
        spanSuccess = 0;
    } else {
        stEndIdc[1] = end_t60;
        spanSuccess = 1;
    }

    // Debug logging
    if (LOG_LEVEL_RUNTIME >= LGDBG2) {
        LIBLOG(LGDBG2, "\tst end idx: %d %d (%d samples)", stEndIdc[0], stEndIdc[1],
            stEndIdc[1] - stEndIdc[0]);
        LIBLOG(LGDBG2, "\t\tedcMax: %.4f", edcMax);
    }

    return spanSuccess;
}

// Returns first index of vec below thresh, or  -1 on fail. Search starts at startIdx.
int rirlib_firstIndexLessThan(float* vec, int startIdx, int endIdx, float thresh)
{
    for (int i = startIdx; i < endIdx + 1; i++) {
        if (vec[i] < thresh)
            return i;
    }
    return -1;
}

// Returns first index of vec above thresh, or  -1 on fail. Search starts at startIdx.
int rirlib_firstIndexGreaterThan(float* vec, int startIdx, int endIdx, float thresh)
{
    for (int i = startIdx; i < endIdx + 1; i++) {
        if (vec[i] >= thresh)
            return i;
    }
    return -1;
}

void checkProperProcessingOrder(
    void* const hRL, ANALYSIS_STAGE currentStage, const char* funcName, LOG_LEVEL logLevel)
{
    LIBLOG(logLevel, "> %s", funcName);

    rirlib_data* pData = (rirlib_data*)(hRL);

    if (pData->analysisStage < currentStage - 1) {
        fprintf(stderr, "ERROR [srirlib]");
        fprintf(stderr, " %s (%d)", funcName, (int)currentStage);
        fprintf(stderr,
            " was called before required processing stages were completed (currently at %d).\n",
            (int)pData->analysisStage);
        exit(EXIT_FAILURE);
    }
}
