/***************************************************************************
This software module was originally developed by

Nokia Technologies Ltd.

in the course of development of ISO/IEC 23090-4 for reference purposes and
its performance may not have been optimized.  This software module is an
implementation of one or more tools as specified by the ISO/IEC 23090-4
standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
copyright license to copy, distribute, and make derivative works of this
software module or modifications thereof for use in implementations or
products claiming conformance to ISO/IEC standards incorporating ISO/IEC
23090-4 reference software and which satisfy any specified conformance
criteria of that standard.

Those intending to use this software module in products are advised that its
use may infringe existing patents and that the license in the previous
paragraph grants no licenses under such patents.  ISO/IEC and the developers
of this software module have no liability for use of this software module or
modifications thereof.

Copyright is not released for products that do not conform to an ISO/IEC
standard.

Nokia Technologies Ltd. retains full right to
modify and use the code for its own purpose, assign or donate the code to
a third party without consent of and without accounting to third parties,
and to inhibit third parties from using the code for products that do not
conform to MPEG-related ISO/IEC International Standards.

This copyright notice must be included in all copies or derivative works.

Copyright (c) ISO/IEC 2022.

This software module derives its structure, with substantial modification of
functionality, from the HO-SIRR library released under the following license:

BSD 3-Clause License

Copyright (c) 2019, Leo McCormack
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
***************************************************************************/

#ifndef __RIRLIB_INTERNAL_H_INCLUDED__
#define __RIRLIB_INTERNAL_H_INCLUDED__

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "rirlib.h"
#include "saf.h"
#include "saf_externals.h" // to also include saf dependencies (cblas etc.)

#ifdef __cplusplus
#include <cmath> // For std::exp
#include <complex> // For C++ complex types
#else
#include <complex.h> // For C99 complex types
#endif

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/* Internal Analysis Parameters */

// To help avoid traingulation errors
#define MIN_NUM_LOUDSPEAKERS (4)
#define DEFAULT_WINDOW_LENGTH (128)
#define MAX_WINDOW_LENGTH (256)
#define MIN_WINDOW_LENGTH (32)
#define MAX_DIFF_FREQ_HZ (3000)
#define ALPHA_DIFF_COEFF (0.7165f)

    /* SAF SUBSTITUTES */
    /* To resolve conflicting versions of SAF, define local versions of these funcs */

#ifndef RIRLIB_MIN
#define RIRLIB_MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef RIRLIB_MAX
#define RIRLIB_MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef RIRLIB_CLAMP
#define RIRLIB_CLAMP(a, min, max) (RIRLIB_MAX(min, RIRLIB_MIN(max, a)))
#endif

    /* Warning/error message formatting with the file name and line number */

#define rirlib_print_warning(message)                                                              \
    {                                                                                              \
        fprintf(stdout, "! RIRLIB WARNING: %s [%s LINE %u] \n", message, __FILE__, __LINE__);      \
    }

#define rirlib_print_error(message)                                                                \
    {                                                                                              \
        fprintf(stderr, "! RIRLIB ERROR: %s [%s LINE %u] \n", message, __FILE__, __LINE__);        \
        exit(EXIT_FAILURE);                                                                        \
    }

    /* Structures */

    // Main structure for rirlib
    typedef struct _rirlib
    {
        // depend only on nDir
        float** encBeamCoeffs; // nSH x nDir
        float** decBeamCoeffs; // nDir x nSH
        float** dirGainDB_bnd_dir; // nBand x nDir
        float* t60Omni_bnd; // nBand x 1
        float* rdr_bnd; // nBand x 1
        int* directOnsetIdx_bnd; // nBand x 1
        int* validSlope_bnd; // nBand x 1
        int** t60Bounds_bnd; // nBand x 2

        // depend on output design (nDir) AND input RIR (nSamp)
        float** rir_sh; // nSH x nSamp (input RIR)
        float*** rir_bnd_sh; // nBand x nSH x nSamp
        float*** rir_bnd_dir; // nBand x nDir x nSamp
        float** edc_bnd; // (omni) nBand x nSamp
        float*** edc_bnd_dir; // nBand x nDir x nSamp

        float** H_bandFilt; // nBand x bandFiltOrder + 1
        float* centerFreqs; // size should match nBand.
        float* xOverFreqs; // 1 x nBand-1
        float* srcDirectivity_bnd; // 1 x nBand
        float srcPosition[3]; // [X, Y, Z], used for src-rec distance calc
        float recPosition[3]; // [X, Y, Z]

        int nSH, nSamp, shOrder;
        float fs;
        int nDir, nBand, bandFiltOrder;
        int directOnsetIdx_brdbnd; // direct arrival onset index within the input RIR
        int diffuseOnsetIdx; // diffuse onset sample index within the input RIR
        float diffuseOnsetSec; // diffuse onset from t0, in seconds
        float t0Sec; // time 0: when the sound leaves the source (directOnset - sourceDistance/343)
        int t0Idx; // sample index of t0 (can be negative)
        float duration; // seconds
        float diffuseMin; // minimum diffuseness to detect onset, otherwise it's direct onset +10ms
        float diffuseOnsetFallbackDelay; // delay (sec) after direct onset, to substitute
                                         // calculation of diffuseOnset in case of diffuseness
                                         // threshold isn't reached.
        int srcDirectivityFlag; // boolean: 0: omni, 1: "regular" loudspeaker directivity

        ANALYSIS_STAGE analysisStage;
        BEAM_TYPES beamType;
        NORMALIZATION_TYPES inputNorm;

        /* user parameters */
        int windowLength;
        float sphDirsDeg[RIRLIB_MAX_NUM_OUTPUTS][2];

    } rirlib_data;

    /* Internal Functions */

    void loadSphDesignPreset(
        SPHDESIGN_PRESETS preset, float dirsDeg[RIRLIB_MAX_NUM_OUTPUTS][2], int* nChanOut);

    /* ========================================================================== */
    /*                            FIR Filter Functions                            */
    /* The three functions                                                        */
    /*      rirlib_applyWindowingFunction, rirlib_FIRCoeffs, rirlib_FIRFilterbank */
    /* mirror those of the same name in SAF, correcting band crossovers.          */
    /* The patch is made here to avoid having to bump the version of the current  */
    /* SAF dependency. If the SAF dependency is updated in the future, these can  */
    /* be removed, provided this PR is accepted:                                  */
    /* https://github.com/leomccormack/Spatial_Audio_Framework/pull/61            */
    /* ========================================================================== */

    static void rirlib_applyWindowingFunction(
        WINDOWING_FUNCTION_TYPES type, int winLength, float* x)
    {
        int i, N;

        if (!(winLength % 2 == 0))
            // Odd winlength: symmetric window (mid index has value=1)
            N = winLength - 1;
        else
            // Even winlength: (index: winlength/2+1 = 1.0, but first value != last value)
            N = winLength;

        switch (type) {
        case WINDOWING_FUNCTION_RECTANGULAR:
            break;

        case WINDOWING_FUNCTION_HAMMING:
            for (i = 0; i < winLength; i++) {
                // more wide-spread coefficient values
                x[i] *= 0.54f - 0.46f * (cosf(2.0f * SAF_PI * (float)i / (float)N));
                // Optimal equiripple coefficient values:
                // x[i] *= 0.53836f - 0.46164f * (cosf(2.0f * SAF_PI * (float)i / (float)N));
            }
            break;

        case WINDOWING_FUNCTION_HANN:
            for (i = 0; i < winLength; i++)
                x[i] *= 0.5f - 0.5f * (cosf(2.0f * SAF_PI * (float)i / (float)N));
            break;

        case WINDOWING_FUNCTION_BARTLETT:
            for (i = 0; i < winLength; i++)
                x[i] *= 1.0f - 2.0f * fabsf((float)i - ((float)N / 2.0f)) / (float)N;
            break;

        case WINDOWING_FUNCTION_BLACKMAN:
            for (i = 0; i < winLength; i++) {
                x[i] *= 0.42659f - 0.49656f * cosf(2.0f * SAF_PI * (float)i / (float)N)
                    + 0.076849f * cosf(4.0f * SAF_PI * (float)i / (float)N);
            }
            break;

        case WINDOWING_FUNCTION_NUTTALL:
            for (i = 0; i < winLength; i++) {
                x[i] *= 0.355768f - 0.487396f * cosf(2.0f * SAF_PI * (float)i / (float)N)
                    + 0.144232f * cosf(4.0f * SAF_PI * (float)i / (float)N)
                    - 0.012604f * cosf(6.0f * SAF_PI * (float)i / (float)N);
            }
            break;

        case WINDOWING_FUNCTION_BLACKMAN_NUTTALL:
            for (i = 0; i < winLength; i++) {
                x[i] *= 0.3635819f - 0.4891775f * cosf(2.0f * SAF_PI * (float)i / (float)N)
                    + 0.1365995f * cosf(4.0f * SAF_PI * (float)i / (float)N)
                    + 0.0106411f * cosf(4.0f * SAF_PI * (float)i / (float)N);
            }
            break;

        case WINDOWING_FUNCTION_BLACKMAN_HARRIS:
            for (i = 0; i < winLength; i++) {
                x[i] *= 0.35875f - 0.48829f * cosf(2.0f * SAF_PI * (float)i / (float)N)
                    + 0.14128f * cosf(4.0f * SAF_PI * (float)i / (float)N)
                    + 0.01168f * cosf(4.0f * SAF_PI * (float)i / (float)N);
            }
            break;
        }
    }

    static void rirlib_FIRCoeffs(FIR_FILTER_TYPES filterType, int order, float fc1,
        float fc2, // Only needed for band-pass/stop
        float fs, WINDOWING_FUNCTION_TYPES windowType, int scalingFLAG, float* h_filt)
    {
        int i, h_len;
        float ft1, ft2, h_sum, f0;
        float_complex h_z_sum;

        h_len = order + 1;

        // Corrected, was ft1 = fc1/(fs*2.0f)
        // see §8.1.1, J. G. Proakis and D. G. Manolakis, Digital Signal Processing: Principles,
        // Algorithms, and Applications.
        ft1 = fc1 / fs;

        /* Compute filter weights */

        if (order % 2 == 0) {
            // Filter order is even
            switch (filterType) {
            case FIR_FILTER_LPF:
                for (i = 0; i < h_len; i++)
                    h_filt[i] = i == order / 2 ? 2.0f * ft1
                                               : sinf(2.0f * SAF_PI * ft1 * (float)(i - order / 2))
                            / (SAF_PI * (float)(i - order / 2));
                break;

            case FIR_FILTER_HPF:
                for (i = 0; i < h_len; i++)
                    h_filt[i] = i == order / 2 ? 1.0f - 2.0f * ft1
                                               : -sinf(2.0f * ft1 * SAF_PI * (float)(i - order / 2))
                            / (SAF_PI * (float)(i - order / 2));
                break;

            case FIR_FILTER_BPF:
                ft2 = fc2 / fs; // corrected, was fc2/(fs*2.0f);
                for (i = 0; i < h_len; i++) {
                    h_filt[i] = i == order / 2 ? 2.0f * (ft2 - ft1)
                                               : sinf(2.0f * SAF_PI * ft2 * (float)(i - order / 2))
                                / (SAF_PI * (float)(i - order / 2))
                            - sinf(2.0f * SAF_PI * ft1 * (float)(i - order / 2))
                                / (SAF_PI * (float)(i - order / 2));
                }
                break;

            case FIR_FILTER_BSF:
                ft2 = fc2 / fs; // corrected, was fc2/(fs*2.0f);
                for (i = 0; i < h_len; i++) {
                    h_filt[i] = i == order / 2 ? 1.0f - 2.0f * (ft2 - ft1)
                                               : sinf(2.0f * SAF_PI * ft1 * (float)(i - order / 2))
                                / (SAF_PI * (float)(i - order / 2))
                            - sinf(2.0f * SAF_PI * ft2 * (float)(i - order / 2))
                                / (SAF_PI * (float)(i - order / 2));
                }
                break;
            }
        } else
            rirlib_print_error("Please specify an even value for the filter 'order' argument");

        // Apply windowing function
        rirlib_applyWindowingFunction(windowType, h_len, h_filt);

        // Scaling, to ensure pass-band is truely at 1 (0dB).
        // See "Programs for Digital Signal Processing", IEEE Press John Wiley & Sons, 1979,
        // pg. 5.1-2.
        if (scalingFLAG) {
            switch (filterType) {
            case FIR_FILTER_LPF:
            case FIR_FILTER_BSF:
                h_sum = 0.0f;
                for (i = 0; i < h_len; i++)
                    h_sum += h_filt[i];
                for (i = 0; i < h_len; i++)
                    h_filt[i] /= h_sum;
                break;
#ifdef __cplusplus
            // :( ifdef: colliding complex types/functions when compiling in C++ project!
            case FIR_FILTER_HPF:
                f0 = 1.0f;
                h_z_sum = cmplxf(0.0f, 0.0f);
                for (i = 0; i < h_len; i++)
                    h_z_sum = ccaddf(h_z_sum,
                        crmulf(std::exp(cmplxf(0.0f, -2.0f * SAF_PI * (float)i * f0 / 2.0f)),
                            h_filt[i]));
                h_sum = std::abs(h_z_sum);
                for (i = 0; i < h_len; i++)
                    h_filt[i] /= h_sum;
                break;

            case FIR_FILTER_BPF:
                f0 = fc1 / fs + fc2 / fs; // corrected, was (fc1/fs+fc2/fs)/2.0f;
                h_z_sum = cmplxf(0.0f, 0.0f);
                for (i = 0; i < h_len; i++)
                    h_z_sum = ccaddf(h_z_sum,
                        crmulf(std::exp(cmplxf(0.0f, -2.0f * SAF_PI * (float)i * f0 / 2.0f)),
                            h_filt[i]));
                h_sum = std::abs(h_z_sum);
                for (i = 0; i < h_len; i++)
                    h_filt[i] /= h_sum;
                break;
#else
        case FIR_FILTER_HPF:
            f0 = 1.0f;
            h_z_sum = cmplxf(0.0f, 0.0f);
            for (i = 0; i < h_len; i++)
                h_z_sum = ccaddf(h_z_sum,
                    crmulf(cexpf(cmplxf(0.0f, -2.0f * SAF_PI * (float)i * f0 / 2.0f)), h_filt[i]));
            h_sum = cabsf(h_z_sum);
            for (i = 0; i < h_len; i++)
                h_filt[i] /= h_sum;
            break;

        case FIR_FILTER_BPF:
            f0 = fc1 / fs + fc2 / fs; // corrected, was (fc1/fs+fc2/fs)/2.0f;
            h_z_sum = cmplxf(0.0f, 0.0f);
            for (i = 0; i < h_len; i++)
                h_z_sum = ccaddf(h_z_sum,
                    crmulf(cexpf(cmplxf(0.0f, -2.0f * SAF_PI * (float)i * f0 / 2.0f)), h_filt[i]));
            h_sum = cabsf(h_z_sum);
            for (i = 0; i < h_len; i++)
                h_filt[i] /= h_sum;
            break;
#endif
            }
        }
    }

    static void rirlib_FIRFilterbank(int order, float* fc, // Cut-off frequencies; nCutoffFreq x 1
        int nCutoffFreq, float sampleRate, WINDOWING_FUNCTION_TYPES windowType, int scalingFLAG,
        float* filterbank // (nCutoffFreq+1) x (order+1)
    )
    {
        int k, nFilt;

        // Number of filters returned is always one more than the number of cut-off frequencies
        nFilt = nCutoffFreq + 1;

        // First and last bands are low-pass and high pass filters, using the first and last cut-off
        // frequencies in vector 'fc', respectively.
        rirlib_FIRCoeffs(
            FIR_FILTER_LPF, order, fc[0], 0.0f, sampleRate, windowType, scalingFLAG, filterbank);
        rirlib_FIRCoeffs(FIR_FILTER_HPF, order, fc[nCutoffFreq - 1], 0.0f, sampleRate, windowType,
            scalingFLAG, &filterbank[(nFilt - 1) * (order + 1)]);

        // In-between bands are then band-pass filters
        if (nCutoffFreq > 1) {
            for (k = 1; k < nFilt - 1; k++)
                rirlib_FIRCoeffs(FIR_FILTER_BPF, order, fc[k - 1], fc[k], sampleRate, windowType,
                    scalingFLAG, &filterbank[k * (order + 1)]);
        }
    }

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#endif /* __RIRLIB_INTERNAL_H_INCLUDED__ */
