# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import Any, Dict, List, Type

from mpegi.framework import Bitstream, Inventory, MHASPacketType, Packet, Payload, serialize_payload

from . import _payloads


class ScenePayloadPacket(Packet):
    """Packet to include bulky metadata for rendering."""

    def __init__(self, args: Dict[str, Any]) -> None:
        self.payloads_to_serialize: List[Type[Payload]] = [
            _payloads.ScenePayload,
            _payloads.DirectivityPayload,
            _payloads.ERPayload,
            _payloads.DiffractionPayload,
            _payloads.PortalPayload,
            _payloads.LateReverbPayload,
            _payloads.AudioplusPayload,
            _payloads.VoxdataPayload,
            _payloads.GranularDatabasePayload,
            _payloads.AirflowPayload,
            _payloads.RasterMapPayload,
            _payloads.LCERPayload,
        ]

        if args["disableDIFFR"]:
            diffr_index = self.payloads_to_serialize.index(_payloads.DiffractionPayload)
            self.payloads_to_serialize.pop(diffr_index)

        if args["disableER"] or args["useLowComplexityER"]:
            er_index = self.payloads_to_serialize.index(_payloads.ERPayload)
            self.payloads_to_serialize.pop(er_index)

        if not args["useLowComplexityER"]:
            lcer_index = self.payloads_to_serialize.index(_payloads.LCERPayload)
            self.payloads_to_serialize.pop(lcer_index)

    @staticmethod
    def get_packet_type() -> MHASPacketType:
        """Return PACTYPE_PLD MHAS structure."""
        return MHASPacketType.PACTYP_PLD

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        """Return a list of all required arguments."""
        return [
            "disableDIFFR",
            "disableER",
            "useLowComplexityER",
        ]

    def serialize(self, inventory: Inventory) -> List[Bitstream]:
        """Return a list of payloads from a scene payload packet."""
        payloads: List[Bitstream] = []
        num_of_payloads = len(self.payloads_to_serialize)
        for payload_to_serialize in self.payloads_to_serialize:
            pld_as_bstream = serialize_payload(
                pld_id=0,
                pld_type=payload_to_serialize,
                inv=inventory,
            )

            payloads.append(pld_as_bstream)

            if pld_as_bstream.getLength() == 0:
                num_of_payloads -= 1

        bstream = Bitstream()
        # TODO: Think of a better to id payload packets if we need it
        bstream.addInt(0, width=16)
        bstream.addInt(num_of_payloads, width=16)
        for payload in payloads:
            bstream.appendBitstream(payload)

        return [bstream]
