# This software module was originally developed jointly by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code
# for its own purpose, assign or donate the code to a third party without
# consent of and without accounting to third parties, and to inhibit third
# parties from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.


import logging
from enum import Enum
from typing import Any, Dict, List, Optional

from mpegi.framework import Bitstream, BitstreamComposer, Payload, ScenePayloadPacketType, tokens
from .. import _constants

airflowSyntax: List[tokens.Token] = [
    tokens.writeCountOrIndex("numberOfAirflowSources"),
    tokens.Repeat(
        "numberOfAirflowSources",
        syntax=[
            tokens.writeID("airflowSourceID"),
            tokens.Flag("hasCustomProfile"),
            tokens.Conditional(
                "hasCustomProfile",
                true=[
                    tokens.UnsignedInteger("profileType", width=2),
                    tokens.Switch(
                        "profileType",
                        cases=[
                            tokens.SwitchCase(
                                0,
                                syntax=[
                                    tokens.writeCountOrIndex("numSpeedProfiles"),
                                    tokens.writeCountOrIndex("numAziProfiles"),
                                    tokens.writeCountOrIndex("numElevProfiles"),
                                    tokens.Repeat(
                                        "numSpeedProfiles",
                                        syntax=[
                                            tokens.writeCountOrIndex("freqGridIdx"),
                                            tokens.QuantizedFloat(
                                                "refSpeed",
                                                quant_range=_constants.airflow_speed_range,
                                            ),
                                            tokens.Repeat(
                                                "fgdNrBands",
                                                syntax=[
                                                    tokens.writeGain(
                                                        "speedFreqGain", isHiPrecGain=True
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                    tokens.Repeat(
                                        "numAziProfiles",
                                        syntax=[
                                            tokens.writeCountOrIndex("freqGridIdx"),
                                            tokens.QuantizedFloat(
                                                "refAzimuth",
                                                quant_range=_constants.airflow_angle_range,
                                            ),
                                            tokens.Repeat(
                                                "fgdNrBands",
                                                syntax=[
                                                    tokens.writeGain(
                                                        "azimuthFreqGainL", isHiPrecGain=True
                                                    ),
                                                    tokens.writeGain(
                                                        "azimuthFreqGainR", isHiPrecGain=True
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                    tokens.Repeat(
                                        "numElevProfiles",
                                        syntax=[
                                            tokens.writeCountOrIndex("freqGridIdx"),
                                            tokens.QuantizedFloat(
                                                "refElevation",
                                                quant_range=_constants.airflow_angle_range,
                                            ),
                                            tokens.Repeat(
                                                "fgdNrBands",
                                                syntax=[
                                                    tokens.writeGain(
                                                        "elevationFreqGain", isHiPrecGain=True
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                            tokens.SwitchCase(
                                1,
                                syntax=[
                                    tokens.writeCountOrIndex("numSpeedProfiles"),
                                    tokens.writeCountOrIndex("numAziProfiles"),
                                    tokens.writeCountOrIndex("numElevProfiles"),
                                    tokens.Repeat(
                                        "numSpeedProfiles",
                                        syntax=[
                                            tokens.writeCountOrIndex("freqGridIdx"),
                                            tokens.QuantizedFloat(
                                                "refSpeed",
                                                quant_range=_constants.airflow_speed_range,
                                            ),
                                            tokens.Repeat(
                                                "fgdNrBands",
                                                syntax=[
                                                    tokens.writeGain(
                                                        "speedFilterGain", isHiPrecGain=True
                                                    ),
                                                    tokens.QuantizedFloat(
                                                        "speedFilterQ",
                                                        quant_range=_constants.airflow_filter_Q_range,
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                    tokens.Repeat(
                                        "numAziProfiles",
                                        syntax=[
                                            tokens.writeCountOrIndex("freqGridIdx"),
                                            tokens.QuantizedFloat(
                                                "refAzimuth",
                                                quant_range=_constants.airflow_angle_range,
                                            ),
                                            tokens.Repeat(
                                                "fgdNrBands",
                                                syntax=[
                                                    tokens.writeGain(
                                                        "azimuthFilterGainL", isHiPrecGain=True
                                                    ),
                                                    tokens.writeGain(
                                                        "azimuthFilterGainR", isHiPrecGain=True
                                                    ),
                                                    tokens.QuantizedFloat(
                                                        "azimuthFilterQL",
                                                        quant_range=_constants.airflow_filter_Q_range,
                                                    ),
                                                    tokens.QuantizedFloat(
                                                        "azimuthFilterQR",
                                                        quant_range=_constants.airflow_filter_Q_range,
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                    tokens.Repeat(
                                        "numElevProfiles",
                                        syntax=[
                                            tokens.writeCountOrIndex("freqGridIdx"),
                                            tokens.QuantizedFloat(
                                                "refElevation",
                                                quant_range=_constants.airflow_angle_range,
                                            ),
                                            tokens.Repeat(
                                                "fgdNrBands",
                                                syntax=[
                                                    tokens.writeGain(
                                                        "elevationFilterGain", isHiPrecGain=True
                                                    ),
                                                    tokens.QuantizedFloat(
                                                        "elevationFilterQ",
                                                        quant_range=_constants.airflow_filter_Q_range,
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
                false=[tokens.UnsignedInteger("defaultProfile", width=2)],
            ),
            tokens.writeID("generatorID"),
            tokens.Flag("activeInDefaultRegion"),
            tokens.Flag("hasValidRegions"),
            tokens.Conditional(
                "hasValidRegions",
                false=[],
                true=[
                    tokens.writeCountOrIndex("numRegions"),
                    tokens.Repeat(
                        "numRegions",
                        syntax=[
                            tokens.writeID("airflowRegion"),
                        ],
                    ),
                ],
            ),
        ],
    ),
]


class AirflowProfileTypes(Enum):
    FREQ_GAIN_PAIRS = 0
    FILTER_PARAMS = 1


class AirflowPayload(Payload):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Serialises a AirflowPacket packet. The packet is comprised of AirflowSource parameters."""
        super().__init__(args, artifacts)

        self.skipSerializingPacket = False

        airflowSources: Optional[airflowSources] = artifacts["airflow_sources"]

        if not airflowSources:
            logging.info("Skipping airflow packet creation because there are no airflow sources")
            self.skipSerializingPacket = True
            return

        composer = BitstreamComposer(airflowSyntax)
        self._writeAirflowPayload(airflowSources, artifacts, composer)

        self.payload_as_bitstream = composer.get_bitstream()

    def _writeAirflowPayload(
        self, airflowSources, artifacts: Dict[str, Any], composer: BitstreamComposer
    ):
        profiles = artifacts["airflow_frequency_grids"]
        freqGridMapping = artifacts["frequency_grid_mapping"]
        composer.push(len(airflowSources))
        for src in airflowSources:
            composer.push(artifacts["bs_ids"].anyToIntIdMap[src.id])
            hasCustomProfile = src.AirflowProfile is not None
            composer.push(hasCustomProfile)
            if hasCustomProfile:
                speedProfiles = [x for x in src.AirflowProfile if x.speed is not None]
                azimuthProfiles = [x for x in src.AirflowProfile if x.azimuth is not None]
                elevationProfiles = [x for x in src.AirflowProfile if x.elevation is not None]
                profileType = speedProfiles[0].profileType
                composer.push(profileType)
                if profileType == AirflowProfileTypes.FREQ_GAIN_PAIRS.value:
                    composer.push(len(speedProfiles))
                    composer.push(len(azimuthProfiles))
                    composer.push(len(elevationProfiles))
                    for spd in speedProfiles:
                        fgdIdx = freqGridMapping[spd.id]
                        fgdNrBands = profiles[spd.id].fgdNrBands
                        composer.scope.update({"fgdNrBands": fgdNrBands})
                        composer.push(fgdIdx)
                        composer.push(spd.speed)
                        for i in range(0, fgdNrBands):
                            composer.push(profiles[spd.id].normGain[i])
                    for azi in azimuthProfiles:
                        fgdIdx = freqGridMapping[azi.id]
                        fgdNrBands = profiles[azi.id].fgdNrBands
                        composer.scope.update({"fgdNrBands": fgdNrBands})
                        composer.push(fgdIdx)
                        composer.push(azi.azimuth)
                        for i in range(0, fgdNrBands):
                            composer.push(profiles[azi.id].normGain[i][0])
                            composer.push(profiles[azi.id].normGain[i][1])
                    for elev in elevationProfiles:
                        fgdIdx = freqGridMapping[elev.id]
                        fgdNrBands = profiles[elev.id].fgdNrBands
                        composer.scope.update({"fgdNrBands": fgdNrBands})
                        composer.push(fgdIdx)
                        composer.push(elev.elevation)
                        for i in range(0, fgdNrBands):
                            composer.push(profiles[elev.id].normGain[i])
                elif profileType == AirflowProfileTypes.FILTER_PARAMS.value:
                    composer.push(len(speedProfiles))
                    composer.push(len(azimuthProfiles))
                    composer.push(len(elevationProfiles))
                    for spd in speedProfiles:
                        fgdIdx = freqGridMapping[spd.id]
                        fgdNrBands = profiles[spd.id].fgdNrBands
                        composer.scope.update({"fgdNrBands": fgdNrBands})
                        composer.push(fgdIdx)
                        composer.push(spd.speed)
                        for i in range(0, len(spd.gainDb)):
                            composer.push(profiles[spd.id].normGain[i])
                            composer.push(spd.q[i])
                    for azi in azimuthProfiles:
                        fgdIdx = freqGridMapping[azi.id]
                        fgdNrBands = profiles[azi.id].fgdNrBands
                        composer.scope.update({"fgdNrBands": fgdNrBands})
                        composer.push(fgdIdx)
                        composer.push(azi.azimuth)
                        for i in range(0, len(azi.gainDb)):
                            composer.push(profiles[azi.id].normGain[i][0])
                            composer.push(profiles[azi.id].normGain[i][1])
                            composer.push(azi.q[i][0])
                            composer.push(azi.q[i][1])
                    for elev in elevationProfiles:
                        fgdIdx = freqGridMapping[elev.id]
                        fgdNrBands = profiles[elev.id].fgdNrBands
                        composer.scope.update({"fgdNrBands": fgdNrBands})
                        composer.push(fgdIdx)
                        composer.push(elev.elevation)
                        for i in range(0, len(elev.gainDb)):
                            composer.push(profiles[elev.id].normGain[i])
                            composer.push(elev.q[i])
            else:
                composer.push(src.defaultAirflowProfile)

            composer.push(artifacts["bs_ids"].anyToIntIdMap[src.airflowGenerator])
            composer.push(src.activeOutsideAEs)
            hasValidRegions = src.ActiveEnvironment is not None
            composer.push(hasValidRegions)
            if hasValidRegions:
                numRegions = len(src.ActiveEnvironment)
                composer.push(numRegions)
                for region in src.ActiveEnvironment:
                    composer.push(artifacts["bs_ids"].anyToIntIdMap[region.AcEnvId])
            pass

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.AIRFLOW

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "airflow_sources",
            "airflow_generators",
            "airflow_frequency_grids",
            "frequency_grid_mapping",
            "bs_ids",
        ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return []

    def get_bitstream(self) -> Bitstream:
        return Bitstream() if self.skipSerializingPacket else self.payload_as_bitstream
