# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import json
from typing import List

from mpegi.coding._listenerVoxelDiffrEncoder import ListenerVoxelDiffrEncoder
from mpegi.coding._validPathDiffrEncoder import ValidPathDiffrEncoder
from mpegi.encoder.components.bitstreamids import BitstreamIdentifiers
from mpegi.framework import Bitstream, BitstreamComposer, Payload, ScenePayloadPacketType, tokens
from mpegi.metadata.diffraction import Edge
from mpegi.framework.generic_codebook import (
    SymbolHistogram,
    get_encoder_config,
)

from .. import _constants

voxel_grid_syntax: List[tokens.Token] = [
    tokens.writePosition("voxelGridOrigin"),
    tokens.writeDistance("voxelGridPitchX"),
    tokens.writeDistance("voxelGridPitchY"),
    tokens.writeDistance("voxelGridPitchZ"),
    tokens.writeID("voxelGridShapeX"),
    tokens.writeID("voxelGridShapeY"),
    tokens.writeID("voxelGridShapeZ"),
]

edges_syntax: List[tokens.Token] = [
    tokens.CodebookConfig("codebookEdgeID"),
    tokens.CodebookConfig("codebookVtxID"),
    tokens.CodebookConfig("codebookTriID"),
    tokens.writeID("edgesCount"),
    tokens.Repeat(
        "edgesCount",
        syntax=[
            tokens.CodebookSymbol("edgeId", config="codebookEdgeID"),
            tokens.CodebookSymbol("vertex0Id", config="codebookVtxID"),
            tokens.CodebookSymbol("vertex1Id", config="codebookVtxID"),
            tokens.CodebookSymbol("adjacentTri0Id", config="codebookTriID"),
            tokens.CodebookSymbol("adjacentTri1Id", config="codebookTriID"),
            tokens.Flag("roundedEdge"),
            tokens.Flag("relevant"),
        ],
    ),
]

static_edges_syntax: List[tokens.Token] = [
    tokens.Flag("hasStaticEdgeData"),
    tokens.Conditional("hasStaticEdgeData", true=edges_syntax, false=[]),
]

paths_syntax: List[tokens.Token] = [
    tokens.CodebookConfig("codebookEdgeIDSeqLen"),
    tokens.CodebookConfig("codebookEdgeIDSeq"),
    tokens.CodebookConfig("codebookAngleSeq"),
    tokens.UnsignedInteger("numBitsAngle", width=6),
    tokens.writeID("relevantEdgeCount"),
    tokens.Repeat(
        "relevantEdgeCount",
        syntax=[
            tokens.writeID("pathCount"),
            tokens.Repeat(
                "pathCount",
                syntax=[
                    tokens.CodebookSymbol("edgesInPathCount", config="codebookEdgeIDSeqLen"),
                    tokens.Repeat(
                        "edgesInPathCount",
                        syntax=[
                            tokens.CodebookSymbol("edgeId", config="codebookEdgeIDSeq"),
                            tokens.Flag("faceIndicator"),
                            tokens.CodebookSymbol("angle", config="codebookAngleSeq"),
                        ],
                    ),
                ],
            ),
        ],
    ),
]

static_paths_syntax: List[tokens.Token] = [
    tokens.Flag("hasStaticPathsData"),
    tokens.Conditional("hasStaticPathsData", true=paths_syntax, false=[]),
]

listener_voxels_syntax: List[tokens.Token] = [
    tokens.Flag("hasListenerVoxelData"),
    tokens.Conditional(
        "hasListenerVoxelData",
        true=[
            tokens.CodebookConfig("codebookVcX"),
            tokens.CodebookConfig("codebookVcY"),
            tokens.CodebookConfig("codebookVcZ"),
            tokens.CodebookConfig("codebookNumEdges"),
            tokens.CodebookConfig("codebookEdgeId"),
            tokens.CodebookConfig("codebookIndicesRemoved"),
            tokens.writeID("listenerVoxelCount"),
            tokens.Repeat(
                "listenerVoxelCount",
                syntax=[
                    tokens.Flag("hasVoxelCoordZ"),
                    tokens.Conditional(
                        "hasVoxelCoordZ",
                        true=[
                            tokens.CodebookSymbol("voxelCoordZ", config="codebookVcZ"),
                            tokens.Flag("hasVoxelCoordY"),
                            tokens.Conditional(
                                "hasVoxelCoordY",
                                true=[
                                    tokens.CodebookSymbol("voxelCoordY", config="codebookVcY"),
                                    tokens.Flag("hasVoxelCoordX"),
                                    tokens.Conditional(
                                        "hasVoxelCoordX",
                                        true=[
                                            tokens.CodebookSymbol(
                                                "voxelCoordX", config="codebookVcX"
                                            ),
                                        ],
                                        false=[],
                                    ),
                                ],
                                false=[],
                            ),
                        ],
                        false=[],
                    ),
                    tokens.UnsignedInteger("voxelEncodingMode", width=2),
                    tokens.Conditional(
                        "voxelEncodingMode",
                        true=[
                            tokens.Assign("do_loop", tokens.Value(1)),
                            tokens.While(
                                "do_loop",
                                syntax=[
                                    tokens.CodebookSymbol(
                                        "listIndicesRemovedIncrement",
                                        config="codebookIndicesRemoved",
                                    ),
                                    tokens.Assign(
                                        "do_loop", tokens.Lookup("listIndicesRemovedIncrement")
                                    ),
                                ],
                            ),
                        ],
                        false=[],
                    ),
                    tokens.CodebookSymbol("edgeInVoxelCount", config="codebookNumEdges"),
                    tokens.Repeat(
                        "edgeInVoxelCount",
                        syntax=[
                            tokens.CodebookSymbol("edgeId", config="codebookEdgeId"),
                        ],
                    ),
                ],
            ),
        ],
        false=[],
    ),
]

source_voxels_syntax: List[tokens.Token] = [
    tokens.Flag("hasSourceVoxelData"),
    tokens.Conditional(
        "hasSourceVoxelData",
        true=[
            tokens.writeID("staticSourceCount"),
            tokens.Repeat(
                "staticSourceCount",
                syntax=[
                    tokens.writeID("sourceId"),
                    tokens.writeID("voxelPerSourceCount"),
                    tokens.Repeat(
                        "voxelPerSourceCount",
                        syntax=[
                            tokens.writeID("voxelGridIndexX"),
                            tokens.writeID("voxelGridIndexY"),
                            tokens.writeID("voxelGridIndexZ"),
                            tokens.writeID("edgesInVoxelCount"),
                            tokens.Repeat(
                                "edgesInVoxelCount",
                                syntax=[tokens.writeID("edgeId")],
                            ),
                        ],
                    ),
                ],
            ),
        ],
        false=[],
    ),
]

valid_path_syntax: List[tokens.Token] = [
    tokens.Flag("hasValidPaths"),
    tokens.Conditional(
        "hasValidPaths",
        true=[
            tokens.writeID("staticSourceCount"),
            tokens.Repeat(
                "staticSourceCount",
                syntax=[
                    tokens.writeID("sourceId"),
                    tokens.CodebookConfig("codebookVcX"),
                    tokens.CodebookConfig("codebookVcY"),
                    tokens.CodebookConfig("codebookVcZ"),
                    tokens.CodebookConfig("codebookNumPaths"),
                    tokens.CodebookConfig("codebookEdgeId"),
                    tokens.CodebookConfig("codebookPathId"),
                    tokens.CodebookConfig("codebookIndicesRemoved"),
                    tokens.writeID("listenerVoxelCount"),
                    tokens.Repeat(
                        "listenerVoxelCount",
                        syntax=[
                            tokens.Flag("hasVoxelCoordZ"),
                            tokens.Conditional(
                                "hasVoxelCoordZ",
                                true=[
                                    tokens.CodebookSymbol("voxelCoordZ", config="codebookVcZ"),
                                    tokens.Flag("hasVoxelCoordY"),
                                    tokens.Conditional(
                                        "hasVoxelCoordY",
                                        true=[
                                            tokens.CodebookSymbol(
                                                "voxelCoordY", config="codebookVcY"
                                            ),
                                            tokens.Flag("hasVoxelCoordX"),
                                            tokens.Conditional(
                                                "hasVoxelCoordX",
                                                true=[
                                                    tokens.CodebookSymbol(
                                                        "voxelCoordX", config="codebookVcX"
                                                    ),
                                                ],
                                                false=[],
                                            ),
                                        ],
                                        false=[],
                                    ),
                                ],
                                false=[],
                            ),
                            tokens.UnsignedInteger("validPathMode", width=2),
                            tokens.Conditional(
                                "validPathMode",
                                true=[
                                    tokens.Assign("do_loop", tokens.Value(1)),
                                    tokens.While(
                                        "do_loop",
                                        syntax=[
                                            tokens.CodebookSymbol(
                                                "listIndicesRemovedIncrement",
                                                config="codebookIndicesRemoved",
                                            ),
                                            tokens.Assign(
                                                "do_loop",
                                                tokens.Lookup("listIndicesRemovedIncrement"),
                                            ),
                                        ],
                                    ),
                                ],
                                false=[],
                            ),
                            tokens.CodebookSymbol(
                                "pathsPerSourceListenerPairCount", config="codebookNumPaths"
                            ),
                            tokens.Repeat(
                                "pathsPerSourceListenerPairCount",
                                syntax=[
                                    tokens.CodebookSymbol("initialEdgeId", config="codebookEdgeId"),
                                    tokens.CodebookSymbol("pathIndex", config="codebookPathId"),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
        false=[],
    ),
]

dynamic_edge_syntax: List[tokens.Token] = [
    tokens.Flag("hasDynamicEdgeData"),
    tokens.Conditional(
        "hasDynamicEdgeData",
        true=[
            tokens.writeID("dynamicEdgeCount"),
            tokens.Repeat(
                "dynamicEdgeCount",
                syntax=[
                    tokens.writeID("GeometryIndex"),
                    *edges_syntax,
                ],
            ),
        ],
        false=[],
    ),
]

dynamic_path_syntax: List[tokens.Token] = [
    tokens.Flag("hasDynamicPaths"),
    tokens.Conditional(
        "hasDynamicPaths",
        true=[
            tokens.writeID("dynamicPathCount"),
            tokens.Repeat(
                "dynamicPathCount",
                syntax=[
                    tokens.writeID("GeometryIndex"),
                    *paths_syntax,
                ],
            ),
        ],
        false=[],
    ),
]

diffraction_syntax: List[tokens.Token] = [
    *voxel_grid_syntax,
    *static_edges_syntax,
    *static_paths_syntax,
    *listener_voxels_syntax,
    *source_voxels_syntax,
    *valid_path_syntax,
    *dynamic_edge_syntax,
    *dynamic_path_syntax,
]


class DiffractionPayload(Payload):
    def __init__(self, args, artifacts):
        """initialize Diffraction Packet attributes and call methods on demand"""
        super().__init__(args, artifacts)
        composer = BitstreamComposer(diffraction_syntax)
        isSmallScene = artifacts["scene_info"]["isSmallScene"]

        _serialize_voxel_grid(composer, artifacts["earLevelVoxelGrid"], isSmallScene)

        _serialize_static_edges(composer, artifacts["diffraction_edges"])

        _serialize_static_paths(composer, artifacts["diffraction_paths"])

        _serialize_listener_voxels(composer, artifacts["diffraction_listenerVoxels"])

        _serialize_source_voxels(
            composer, artifacts["bs_ids"], artifacts["diffraction_sourceVoxels"]
        )

        _serialize_valid_paths(
            composer, artifacts["bs_ids"], artifacts["diffraction_validPathDict"]
        )

        _serialize_dynamic_edges(composer, artifacts["bs_ids"], artifacts["diffraction_edgeDict"])

        _serialize_dynamic_paths(composer, artifacts["bs_ids"], artifacts["diffraction_pathDict"])

        self.payload_as_bitstream = composer.get_bitstream()

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.DIFFRACTION

    @staticmethod
    def getArtifactSubscription():
        return [
            "bs_ids",
            "earLevelVoxelGrid",
            "diffraction_edges",
            "diffraction_paths",
            "diffraction_sourceVoxels",
            "diffraction_listenerVoxels",
            "diffraction_cell",
            "diffraction_edgeDict",
            "diffraction_pathDict",
            "diffraction_validPathDict",
            "scene_info",
            "scene_trimesh_DIFFR_unproc_as_eif_mesh",  # FIXME: remove this!
            "meshes",  # FIXME: remove this!
            "boxes",  # FIXME: remove this!
            "cylinders",  # FIXME: remove this!
        ]

    @staticmethod
    def getArgumentSubscription():
        return []

    def get_bitstream(self) -> Bitstream:
        return self.payload_as_bitstream


def _serialize_voxel_grid(composer: BitstreamComposer, voxelGrid, isSmallScene: bool) -> None:
    origin = [0.0, 0.0, 0.0]
    pitch = [0.0, 0.0, 0.0]
    shape = [0, 0, 0]

    if voxelGrid is not None:
        origin = voxelGrid.origin
        pitch = voxelGrid.pitch
        shape = voxelGrid.shape

    composer.push({"position": origin, "isSmallScene": isSmallScene})

    composer.push({"distance": pitch[0], "isSmallScene": isSmallScene})
    composer.push({"distance": pitch[1], "isSmallScene": isSmallScene})
    composer.push({"distance": pitch[2], "isSmallScene": isSmallScene})

    composer.push(int(shape[0]))
    composer.push(int(shape[1]))
    composer.push(int(shape[2]))


def _serialize_edges(composer: BitstreamComposer, edge_list: List[Edge]) -> None:

    # generate symbol lists
    list_edge_ID = list()
    list_vtx_ID = list()
    list_tri_ID = list()

    for e in edge_list:
        list_edge_ID.append(int(e.eID))
        list_vtx_ID.append(int(e.vtxIDs[0]))
        list_vtx_ID.append(int(e.vtxIDs[1]))
        list_tri_ID.append(int(e.fIDs[0]))
        list_tri_ID.append(int(e.fIDs[1]))

    # serialize codebooks
    histogramEdgeID = SymbolHistogram.from_plain(list_edge_ID)
    histogramVtxID = SymbolHistogram.from_plain(list_vtx_ID)
    histogramTriID = SymbolHistogram.from_plain(list_tri_ID)
    encoderEdgeID = get_encoder_config(histogramEdgeID)
    encoderVtxID = get_encoder_config(histogramVtxID)
    encoderTriID = get_encoder_config(histogramTriID)
    composer.push(encoderEdgeID)
    composer.push(encoderVtxID)
    composer.push(encoderTriID)

    composer.push(len(edge_list))

    for e in edge_list:
        composer.push(int(e.eID))

        composer.push(int(e.vtxIDs[0]))
        composer.push(int(e.vtxIDs[1]))

        composer.push(int(e.fIDs[0]))
        composer.push(int(e.fIDs[1]))

        composer.push(e.rounded)
        composer.push(e.relevant)


def _serialize_static_edges(composer: BitstreamComposer, edge_list: List[Edge]) -> None:

    has_static_edge_data = len(edge_list) > 0
    composer.push(has_static_edge_data)

    if not has_static_edge_data:
        return

    _serialize_edges(composer, edge_list)


def _serialize_paths(composer: BitstreamComposer, paths_dict) -> None:

    # quantization constant
    num_bits_angle = _constants.bitsForDiffrAngle

    # generate symbol lists
    list_edgeIDSeq_len = list()
    list_edgeIDSeq = list()
    list_angleSeq = list()

    for path_list in paths_dict.values():
        for path in path_list:
            list_edgeIDSeq_len.append(len(path.edgeIDSeq))
            for edgeID in path.edgeIDSeq:
                list_edgeIDSeq.append(int(edgeID))
            for angle in path.angleSeq:
                angle_quant = _quantize_angle(angle, num_bits_angle)
                list_angleSeq.append(int(angle_quant))

    # serialize codebooks
    histogramEdgeIDSeqLen = SymbolHistogram.from_plain(list_edgeIDSeq_len)
    histogramEdgeIDSeq = SymbolHistogram.from_plain(list_edgeIDSeq)
    histogramAngleSeq = SymbolHistogram.from_plain(list_angleSeq)
    encoderEdgeIDSeqLen = get_encoder_config(histogramEdgeIDSeqLen)
    encoderEdgeIDSeq = get_encoder_config(histogramEdgeIDSeq)
    encoderAngleSeq = get_encoder_config(histogramAngleSeq)

    composer.push(encoderEdgeIDSeqLen)
    composer.push(encoderEdgeIDSeq)
    composer.push(encoderAngleSeq)

    composer.push(num_bits_angle)

    # relevantEdgeCount
    composer.push(len(paths_dict))

    for path_list in paths_dict.values():
        # pathCount
        composer.push(len(path_list))

        for path in path_list:
            # edgesInPathCount
            edgesInPathCount = len(path.edgeIDSeq)
            composer.push(edgesInPathCount)

            for iPoint in range(edgesInPathCount):
                composer.push(int(path.edgeIDSeq[iPoint]))
                composer.push(path.faceIndicatorSeq[iPoint] > 0)
                angle = path.angleSeq[iPoint]
                angle_quant = _quantize_angle(angle, num_bits_angle)
                composer.push(int(angle_quant))


def _serialize_static_paths(composer: BitstreamComposer, static_paths) -> None:

    has_static_paths_data = len(static_paths) > 0
    composer.push(has_static_paths_data)

    if not has_static_paths_data:
        return

    _serialize_paths(composer, static_paths)


def _serialize_listener_voxels(composer: BitstreamComposer, listener_voxels: dict) -> None:

    has_listener_voxel_data = len(listener_voxels) > 0
    composer.push(has_listener_voxel_data)

    if not has_listener_voxel_data:
        return

    voxelEncoder = ListenerVoxelDiffrEncoder()

    bitCounter = voxelEncoder.encode(listener_voxels)
    for token in bitCounter.tokenList:
        composer.push(token)


def _serialize_source_voxels(
    composer: BitstreamComposer,
    bs_ids: BitstreamIdentifiers,
    source_voxels,
) -> None:
    has_source_voxel_data = len(source_voxels) > 0
    composer.push(has_source_voxel_data)

    if not has_source_voxel_data:
        return

    composer.push(len(source_voxels))

    for source_id, iSVD in source_voxels.items():
        source_id = bs_ids.getId(source_id)
        composer.push(source_id)
        composer.push(len(iSVD))

        for source_voxel_indices_as_str, source_voxel_edge_ids in iSVD.items():
            source_voxel_indices = [int(val) for val in source_voxel_indices_as_str.split(",")]
            composer.push(source_voxel_indices[0])
            composer.push(source_voxel_indices[1])
            composer.push(source_voxel_indices[2])

            composer.push(len(source_voxel_edge_ids))
            for edge_id in source_voxel_edge_ids:
                composer.push(int(edge_id))


def _serialize_valid_paths(
    composer: BitstreamComposer, bs_ids: BitstreamIdentifiers, paths: dict
) -> None:

    has_valid_paths_data = len(paths) > 0
    composer.push(has_valid_paths_data)

    if not has_valid_paths_data:
        return

    composer.push(len(paths))
    for source_id, path_dict in paths.items():
        source_bs_id = bs_ids.getId(source_id)
        composer.push(source_bs_id)

        voxelEncoder = ValidPathDiffrEncoder()

        bitCounter = voxelEncoder.encode(path_dict)
        for token in bitCounter.tokenList:
            composer.push(token)


def _serialize_dynamic_edges(
    composer: BitstreamComposer, bs_ids: BitstreamIdentifiers, dynamic_edges
) -> None:

    num_lists = sum(1 for geo_id in dynamic_edges.keys() if geo_id != "sceneTrimesh_DIFFR")

    has_dynamic_edge_data = num_lists > 0
    composer.push(has_dynamic_edge_data)

    if not has_dynamic_edge_data:
        return

    composer.push(num_lists)

    for geo_id, edge_list in dynamic_edges.items():
        # Fixme: why is static "sceneTrimesh_DIFFR" contained in dynamic_edges?!?
        if geo_id == "sceneTrimesh_DIFFR":
            continue

        geo_bs_id = bs_ids.getId(geo_id)
        composer.push(geo_bs_id)
        _serialize_edges(composer, edge_list)


def _serialize_dynamic_paths(
    composer: BitstreamComposer, bs_ids: BitstreamIdentifiers, dynamic_paths
) -> None:

    num_lists = sum(1 for geo_id in dynamic_paths.keys() if geo_id != "sceneTrimesh_DIFFR")

    has_dynamic_paths = num_lists > 0
    composer.push(has_dynamic_paths)

    if not has_dynamic_paths:
        return

    composer.push(num_lists)

    for geometry_id, path_dict in dynamic_paths.items():
        if geometry_id == "sceneTrimesh_DIFFR":
            continue

        geometry_bs_id = bs_ids.getId(geometry_id)
        composer.push(geometry_bs_id)
        _serialize_paths(composer, path_dict)


def _quantize_angle(value, num_bits):
    """Quantize an angle with the given number of bits
    :param value:       float, angle in degree
    :param num_bits:    int, number of bits
    :return:            int, quantization index
    """
    num_steps = 1 << num_bits

    fval = value * num_steps / 360.0
    nval = int(round(fval))

    while nval < 0:
        nval += num_steps
    while nval >= num_steps:
        nval -= num_steps

    return nval
