# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import Any, Dict, List

from mpegi.framework import (
    Bitstream,
    BitstreamComposer,
    NotEnoughDataPushedError,
    Payload,
    ScenePayloadPacketType,
    tokens,
)

from .. import _constants

syntax: List[tokens.Token] = [
    tokens.CommentToken("DIRECTIVITY PAYLOAD"),
    tokens.writeCountOrIndex("numberOfDirectivities"),
    tokens.Repeat(
        "numberOfDirectivities",
        [
            tokens.writeID("directivityID"),
            tokens.Blob("directivityCoded"),
        ],
    ),
]


class DirectivityPayload(Payload):
    """A payload that contains coded source directivities.

    Source directivities include magnitude coefficients per 3rd band octave per
    direction where the direction is given in Polar coordinates.
    """

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Construct a Directivity Payload."""
        super().__init__(args, artifacts)
        composer = BitstreamComposer(syntax)

        coded_directivities = artifacts["directivities_coded"]
        composer.push(len(coded_directivities.items()))
        for directivity_id, blob in coded_directivities.items():
            bs_id = artifacts["bs_ids"].getId(directivity_id)
            composer.push(bs_id)
            composer.push(blob)

        try:
            self.payload_as_bitstream = composer.get_bitstream()
        except NotEnoughDataPushedError as exc:
            raise NotEnoughDataPushedError(f"DirectivityPayload: {exc}") from exc

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        """Return the type of this Payload."""
        return ScenePayloadPacketType.DIRECTIVITY

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        """Return the artifact subscription for this Payload."""
        return ["directivities_coded", "bs_ids"]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        """Return the argument subscription for this Payload."""
        return []

    def get_bitstream(self) -> Bitstream:
        """Return the payload as a bitstream."""
        return self.payload_as_bitstream
