# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import logging
from typing import List

import numpy as np

from mpegi.framework import (
    Bitstream,
    BitstreamComposer,
    Payload,
    ScenePayloadPacketType,
    tokens,
)
from mpegi.framework.generic_codebook import (
    SymbolHistogram,
    encode_symbols,
    get_encoder_config,
    serialize_codebook,
)
from mpegi.coding import VoxelDataEarlyEncoder

from .. import _constants

distance_culling_syntax: List[tokens.Token] = [
    tokens.UnsignedInteger("triangleCullingDistance_1Order", width=8),
    tokens.UnsignedInteger("triangleCullingDistance_2Order", width=8),
    tokens.UnsignedInteger("triangleSourceDistance_1Order", width=8),
    tokens.UnsignedInteger("triangleSourceDistance_2Order", width=8),
]

voxel_grid_syntax: List[tokens.Token] = [
    tokens.writePosition("voxelGridOrigin"),
    tokens.writeDistance("voxelGridPitchX"),
    tokens.writeDistance("voxelGridPitchY"),
    tokens.writeDistance("voxelGridPitchZ"),
    tokens.UnsignedInteger("voxelGridShapeX", width=_constants.bitsForVoxelGridShapeX),
    tokens.UnsignedInteger("voxelGridShapeY", width=_constants.bitsForVoxelGridShapeY),
    tokens.UnsignedInteger("voxelGridShapeZ", width=_constants.bitsForVoxelGridShapeZ),
]

surfaces_syntax: List[tokens.Token] = [
    tokens.Flag("hasSurfaceData"),
    tokens.Conditional("hasSurfaceData", true=[tokens.Blob("surfaceData")], false=[]),
]

voxel_data_syntax: List[tokens.Token] = [
    tokens.Flag("hasVoxelDataEarly"),
    tokens.Conditional("hasVoxelDataEarly", true=[tokens.Blob("voxelDataEarly")], false=[]),
]


syntax: List[tokens.Token] = [
    tokens.CommentToken("ER PAYLOAD"),
    *distance_culling_syntax,
    *voxel_grid_syntax,
    *surfaces_syntax,
    *voxel_data_syntax,
]


class ERPayload(Payload):
    def __init__(self, args, artifacts):
        super().__init__(args, artifacts)

        er_module_data = artifacts["ER_data"]
        voxel_grid = artifacts["reachable_voxels"]
        isSmallScene = artifacts["scene_info"]["isSmallScene"]

        composer = BitstreamComposer(syntax)

        _serialize_distance_culling(composer, args)
        _serialize_voxel_grid(composer, voxel_grid, isSmallScene)
        _serialize_surface(composer, er_module_data)
        _serialize_voxel_data_early(composer, er_module_data, voxel_grid)

        self.payload_as_bitstream = composer.get_bitstream()

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.EARLY_RELFECTION

    @staticmethod
    def getArtifactSubscription():
        return ["ER_data", "reachable_voxels", "scene_info", "acoustic_tuning_params"]

    @staticmethod
    def getArgumentSubscription():
        return [
            "max_dist_ER_triangle_1",
            "max_dist_ER_triangle_2",
            "max_dist_ER_source_1",
            "max_dist_ER_source_2",
        ]

    def get_bitstream(self) -> Bitstream:
        return self.payload_as_bitstream


def _serialize_distance_culling(composer: BitstreamComposer, args) -> None:
    max_dist_triangle_1 = np.round(args["max_dist_ER_triangle_1"]).astype(int)
    max_dist_triangle_2 = np.round(args["max_dist_ER_triangle_2"]).astype(int)
    max_dist_source_1 = np.round(args["max_dist_ER_source_1"]).astype(int)
    max_dist_source_2 = np.round(args["max_dist_ER_source_2"]).astype(int)
    composer.push(int(max_dist_triangle_1))
    composer.push(int(max_dist_triangle_2))
    composer.push(int(max_dist_source_1))
    composer.push(int(max_dist_source_2))


def _serialize_voxel_grid(composer: BitstreamComposer, voxelGrid, isSmallScene) -> None:
    origin = [0.0, 0.0, 0.0]
    pitch = [0.0, 0.0, 0.0]
    shape = [0.0, 0.0, 0.0]

    if voxelGrid is not None:
        origin = voxelGrid.origin
        pitch = voxelGrid.pitch
        shape = voxelGrid.shape

    composer.push({"position": origin, "isSmallScene": isSmallScene})

    composer.push({"distance": pitch[0], "isSmallScene": isSmallScene})
    composer.push({"distance": pitch[1], "isSmallScene": isSmallScene})
    composer.push({"distance": pitch[2], "isSmallScene": isSmallScene})

    composer.push(shape[0])
    composer.push(shape[1])
    composer.push(shape[2])


def _serialize_surface(composer: BitstreamComposer, er_module_data) -> None:
    # self.addLabel("EARLY REFLECTION SURFACE DATA")

    has_surface_data = len(er_module_data.surface) > 0
    composer.push(has_surface_data)

    if not has_surface_data:
        logging.info("ER: size(SurfaceData) = 0 bytes")
        return

    surfaceBlob = Bitstream()

    # generate codebooks
    list_length = list()
    list_face = list()
    list_azi = list()
    list_ele = list()
    list_dist = list()
    for s, surf in enumerate(er_module_data.surface):
        # double check that surfaceIdx is not screwed up
        assert surf.surfaceIdx == s

        # symbol list for faceList codebooks
        list_length.append(len(surf.faceIdx))
        for face in surf.faceIdx:
            list_face.append(face)

        # symbol list for N0
        sph = _cart2sph(surf.N0)
        quant_azi = _quantize_azi(sph[0], _constants.bitsForNormalAzi)
        quant_ele = _quantize_ele(sph[1], _constants.bitsForNormalEle)
        list_azi.append(quant_azi)
        list_ele.append(quant_ele)

        # symbol list for d
        quant_d = _quantize_dist(surf.d)
        list_dist.append(quant_d)

    # apply distance offset
    dist_offset = min(list_dist)
    list_dist = list(np.array(list_dist) - dist_offset)

    # serialize codebooks for faceList
    histogramLengthFaceIdx = SymbolHistogram.from_plain(list_length)
    histogramFaceIdx = SymbolHistogram.from_plain(list_face)
    encoderLengthFaceIdx = get_encoder_config(histogramLengthFaceIdx)
    encoderFaceIdx = get_encoder_config(histogramFaceIdx)
    surfaceBlob.appendBitstream(serialize_codebook(encoderLengthFaceIdx))
    surfaceBlob.appendBitstream(serialize_codebook(encoderFaceIdx))

    # serialize codebooks for N0
    histogramAzi = SymbolHistogram.from_plain(list_azi)
    histogramEle = SymbolHistogram.from_plain(list_ele)
    encoderAzi = get_encoder_config(histogramAzi)
    encoderEle = get_encoder_config(histogramEle)
    surfaceBlob.appendBitstream(serialize_codebook(encoderAzi))
    surfaceBlob.appendBitstream(serialize_codebook(encoderEle))

    # serialize codebook for d
    histogramDist = SymbolHistogram.from_plain(list_dist)
    encoderDist = get_encoder_config(histogramDist)
    surfaceBlob.appendBitstream(serialize_codebook(encoderDist))

    # write distance offset
    surfaceBlob.addSignedInt(dist_offset, _constants.bitsForPlaneDist)

    # write number of surfaces
    num_surfaces = len(er_module_data.surface)
    surfaceBlob.addInt(num_surfaces, _constants.bitsForMaxFaces)

    for s, surf in enumerate(er_module_data.surface):
        # serialize faceList
        surfaceBlob.appendBitstream(encode_symbols(encoderLengthFaceIdx, [list_length[s]]))
        for face in surf.faceIdx:
            surfaceBlob.appendBitstream(encode_symbols(encoderFaceIdx, [face]))

        # serialize N0
        surfaceBlob.appendBitstream(encode_symbols(encoderAzi, [list_azi[s]]))
        surfaceBlob.appendBitstream(encode_symbols(encoderEle, [list_ele[s]]))

        # serialize d
        surfaceBlob.appendBitstream(encode_symbols(encoderDist, [list_dist[s]]))

    logging.info("ER: size(SurfaceData) = " + str(surfaceBlob.getLength() / 8.0) + " bytes")
    composer.push(surfaceBlob)


def _serialize_voxel_data_early(composer: BitstreamComposer, er_module_data, voxel_grid) -> None:
    has_voxel_data = len(er_module_data.vde.lut) > 0
    composer.push(has_voxel_data)

    if not has_voxel_data:
        logging.info("ER: size(VoxelDataEarly) = 0 bytes")
        return

    encoder = VoxelDataEarlyEncoder()
    vdeBlob = encoder.encode(er_module_data.vde, voxel_grid)

    logging.info("ER: size(VoxelDataEarly) = " + str(vdeBlob.getLength() / 8.0) + " bytes")
    composer.push(vdeBlob)


def _cart2sph(cart):
    """Convert Cartesian coordinates into spherical coordinates
    :param sph:         np.array(3,), Cartesian coordinates
    :return:            np.array(3,), spherical coordinates [azi, ele, dist] in radian
    """
    x, y, z = (cart[0], cart[1], cart[2])
    dist = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    xz = np.sqrt(x ** 2 + z ** 2)
    azi = np.arctan2(-x, -z)
    ele = np.arctan2(y, xz)
    return np.array([azi, ele, dist])


def _sph2cart(sph):
    """Convert spherical coordinates into Cartesian coordinates
    :param sph:         np.array(3,), spherical coordinates [azi, ele, dist] in radian
    :return:            np.array(3,), Cartesian coordinates
    """
    azi, ele, dist = (sph[0], sph[1], sph[2])
    x = -1.0 * dist * np.sin(azi) * np.cos(ele)
    y = dist * np.sin(ele)
    z = -1.0 * dist * np.cos(azi) * np.cos(ele)
    return np.array([x, y, z])


def _quantize_azi(value, num_bits):
    """Quantize an azimuth angle with the given number of bits
    :param value:       float, azimuth angle in radian
    :param num_bits:    int, number of bits
    :return:            int, quantization index
    """
    max_steps = 1 << num_bits
    num_steps = 144 * (max_steps // 144)

    fval = value * num_steps / (2 * np.pi)
    nval = int(round(fval))

    while nval < 0:
        nval += num_steps
    while nval >= num_steps:
        nval -= num_steps

    return nval


def _quantize_ele(value, num_bits):
    """Quantize an elevation angle with the given number of bits
    :param value:       float, elevation angle in radian, range: [-pi/2,pi/2]
    :param num_bits:    int, number of bits
    :return:            int, quantization index
    """
    max_steps = 1 << num_bits
    num_steps = 72 * (max_steps // 72)

    fval = value * num_steps / np.pi + num_steps / 2
    nval = int(round(fval))

    assert nval >= 0
    assert nval <= num_steps
    assert nval < max_steps

    return nval


def _quantize_dist(value):
    """Quantize a plane distance
    :param value:       float, plane distance in m
    :return:            int, quantization index
    """

    return int(round(1000 * value))
