# This software module was originally developed by
#
# Ericsson AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Ericsson AB retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2023.

from typing import Any, Dict, List

from mpegi.framework import (
    Bitstream,
    BitstreamComposer,
    BitstreamDecomposer,
    Payload,
    ScenePayloadPacketType,
    tokens,
    NotEnoughDataPushedError,
)

import logging
from .. import _constants

per_channel_syntax: List[tokens.Token] = [
    tokens.Repeat(
        "grainLen",
        syntax=[tokens.QuantizedFloat("grainval", quant_range=_constants.granularDB_qrange)],
    ),
]

desc_trend_syntax: List[tokens.Token] = [
    tokens.Repeat(
        "Dimensions",
        syntax=[
            tokens.QuantizedFloat(
                "DescriptorCoordinateTrend",
                quant_range=_constants.granularDB_descriptor_trend_qrange,
            ),
        ],
    )
]

orig_pos_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat(
        "PositionWithinAudioFile", quant_range=_constants.granularDB_pos_within_audio_file_qrange
    )
]

grain_prob_group_syntax: List[tokens.Token] = [
    tokens.writeCountOrIndex("grainProbabilityGroup"),
    tokens.QuantizedFloat(
        "perGrainProbabilityValue", quant_range=_constants.grainProbabilityValue_qrange
    ),
]

per_grain_syntax: List[tokens.Token] = [
    tokens.Repeat(
        "Dimensions",
        syntax=[
            tokens.QuantizedFloat(
                "DescriptorCoordinates", quant_range=_constants.granularDB_descriptor_qrange
            ),
        ],
    ),
    tokens.UnsignedInteger("grainLen", width=int(18)),
    tokens.Conditional("DescTrend", true=desc_trend_syntax, false=[]),
    tokens.Conditional("OrigPosition", true=orig_pos_syntax, false=[]),
    tokens.Conditional("probabilityGroups", true=grain_prob_group_syntax, false=[]),
    tokens.Repeat("Channels", syntax=[*per_channel_syntax]),
]


grain_overlap_amount_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat("GrainOverlapAmount", quant_range=_constants.grain_overlap_amount_qrange)
]

grainMixingRule_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat("grainMixingRule", quant_range=_constants.granularMixRule_qrange),
    tokens.Repeat(
        "Dimensions",
        syntax=[
            tokens.QuantizedFloat(
                "DescriptorCoordinates", quant_range=_constants.granularDB_descriptor_qrange
            ),
        ],
    ),
]

desc_trend_global_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat("probabilityExponentTrend", quant_range=_constants.granularDBprobExp),
    tokens.QuantizedFloat("probabilityWeightTrend", quant_range=_constants.granularDBprobWeight),
    tokens.writeCountOrIndex("descrTrendMethod"),
]
orig_pos_global_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat(
        "probabilityExponentOriginalPos", quant_range=_constants.granularDBprobExp
    ),
    tokens.QuantizedFloat(
        "probabilityWeightOriginalPos", quant_range=_constants.granularDBprobWeight
    ),
]

probability_group_syntax: List[tokens.Token] = [
    tokens.writeCountOrIndex("P"),
    tokens.Repeat(
        "P",
        syntax=[
            tokens.QuantizedFloat(
                "probabilityGroupVal", quant_range=_constants.grainProbabilityValue_qrange
            )
        ],
    ),
]

grainInterp_syntax: List[tokens.Token] = [
    tokens.QuantizedFloat(
        "grainInterpDistThreshold", quant_range=_constants.granularDBgrainInterpDistThreshold
    ),
    tokens.Flag("useInterpolationResampling"),
    tokens.writeCountOrIndex("InterpolationMethod"),
    tokens.Repeat("Dimensions", syntax=[tokens.Flag("pitchDimension")]),
]

uncompressed_syntax: List[tokens.Token] = [tokens.Repeat("numGrains", syntax=[*per_grain_syntax])]


granular_database_syntax: List[tokens.Token] = [
    tokens.CommentToken("GRANULAR DATABASE PAYLOAD"),
    tokens.writeCountOrIndex(
        "numberOfGranularDatabases"
    ),  # number of granular databases in the EIF
    tokens.Repeat(
        "numberOfGranularDatabases",
        [
            tokens.writeID("granularId"),  # id of the granular database read from the EIF
            tokens.writeCountOrIndex(
                "GranularAudioEncodingMode"
            ),  # 0 is for uncompressed PCM, 1 for encoding mode with external audio grains, rest are reserved
            tokens.writeCountOrIndex("Dimensions"),
            # tokens.CommentToken("channels")
            tokens.writeCountOrIndex("Channels"),
            ##--------------------##
            tokens.UnsignedInteger("numGrains", width=int(20)),
            ##--------------------##
            # other parameters:
            tokens.Repeat(
                "Dimensions",
                syntax=[
                    tokens.QuantizedFloat(
                        "coordinateScaling", quant_range=_constants.granularDBcoordScaling_range
                    )
                ],
            ),
            tokens.Repeat(
                "Dimensions",
                syntax=[
                    tokens.QuantizedFloat(
                        "smoothingTargetCoordinates",
                        quant_range=_constants.granularDBsmoothingTargetCoordinates_range,
                    )
                ],
            ),
            tokens.QuantizedFloat(
                "searchRadius", quant_range=_constants.granularDBsearchRadius
            ),  # change
            tokens.writeCountOrIndex("grainHistoryLength"),
            tokens.writeCountOrIndex("NumMixRule"),
            tokens.Repeat("NumMixRule", syntax=grainMixingRule_syntax),
            tokens.Flag("fastGrainSwitchThresholdAvailable"),
            tokens.Conditional(
                "fastGrainSwitchThresholdAvailable",
                true=[
                    tokens.QuantizedFloat(
                        "fastGrainSwitchThreshold",
                        quant_range=_constants.granularDBfastGrainSwitchThreshold,
                    )
                ],
                false=[],
            ),
            # ---------------------------------------------------
            tokens.QuantizedFloat(
                "probabilityExponentPosition", quant_range=_constants.granularDBprobExp
            ),
            tokens.QuantizedFloat(
                "probabilityExponentHistory", quant_range=_constants.granularDBprobExp
            ),
            tokens.QuantizedFloat(
                "probabilityWeightPosition", quant_range=_constants.granularDBprobWeight
            ),
            tokens.QuantizedFloat(
                "probabilityWeightHistory", quant_range=_constants.granularDBprobWeight
            ),
            ##----metadata regarding descriptor trend and original position in audio file--##
            tokens.Flag("DescTrend"),
            tokens.Conditional("DescTrend", true=desc_trend_global_syntax, false=[]),
            tokens.Flag("OrigPosition"),
            tokens.Conditional("OrigPosition", true=orig_pos_global_syntax, false=[]),
            # --------------------------------------------------
            # probability groups:
            tokens.Flag("probabilityGroups"),
            tokens.Conditional("probabilityGroups", true=probability_group_syntax, false=[]),
            # --------------------------------------------------
            # random selection
            tokens.Flag("randomSelection"),
            # --------------------------------------------------
            # interpolation:
            tokens.Flag("useInterpolation"),
            tokens.Conditional("useInterpolation", true=grainInterp_syntax, false=[]),
            tokens.Flag("GrainOverlapAmount"),
            tokens.Conditional("GrainOverlapAmount", true=grain_overlap_amount_syntax, false=[]),
            tokens.Conditional(
                "GranularAudioEncodingMode", true=[], false=uncompressed_syntax
            ),  # nothing written for cases the encoding mode is 1 or more.
        ],
    ),
]


class GranularDatabasePayload(Payload):
    """Payload with granular database containing audio grains, descriptor coordinates and other metadata required for rendering. The database has many encoding modes. Each grain has a header-descriptor coordinate and grain length. All the grains are serialized/stored according to this syntax"""

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Construct a Granular DB Payload."""
        super().__init__(args, artifacts)
        grain_database = artifacts["grain_coded"]  # this is all the databases in a particular scene

        if not grain_database:
            logging.info(
                "Skipping granular database packet creation because there are no granular databases"
            )
            self.skipDatabase = True
            return

        composer = BitstreamComposer(granular_database_syntax)
        self.skipDatabase = False

        composer.push(len(grain_database.items()))  # number of granular databases in the scene

        for grain_id, blobs in grain_database.items():
            bs_id = artifacts["bs_ids"].getId(grain_id)  # id of the granular database "granularId"
            composer.push(bs_id)

            composer.push(blobs["GranularAudioEncodingMode"])

            descrCoord = blobs["DescriptorCoordinates"]  # descriptor coordinates as an array
            grains = blobs["Grains"]
            grainLens = blobs["grainLens"]
            channel = blobs["Channels"]
            dim = blobs["Dimensions"]
            # --------------------
            composer.push(int(blobs["Dimensions"]))  # dimension of descriptor coordinates
            # push channels:
            composer.push(int(blobs["Channels"]))
            # numGrains
            composer.push(int(blobs["numGrains"]))
            # ------push global parameters-----------#
            coordScaling = blobs["coordinateScaling"]
            coordScaling = coordScaling.flatten()
            for ind in range(dim):
                composer.push(float(coordScaling[ind]))

            smoothingTargetCoordinates = blobs["smoothingTargetCoordinates"]
            smoothingTargetCoordinates = smoothingTargetCoordinates.flatten()
            for ind in range(dim):
                composer.push(float(smoothingTargetCoordinates[ind]))

            composer.push(float(blobs["searchRadius"]))
            composer.push(int(blobs["grainHistoryLength"]))

            numMixRule = int(blobs["NumMixRule"])
            composer.push(numMixRule)

            grainMixingRule = blobs["grainMixingRule"]
            grainMixingRuleCoordinates = blobs["grainMixingRuleCoordinates"]
            for ind in range(numMixRule):
                composer.push(grainMixingRule[ind])
                for coord_ind in range(dim):
                    composer.push(grainMixingRuleCoordinates[ind, coord_ind])

            fastGrainSwitchThresholdAvailable = blobs["fastGrainSwitchThresholdAvailable"]
            composer.push(fastGrainSwitchThresholdAvailable)

            if fastGrainSwitchThresholdAvailable:
                composer.push(float(blobs["fastGrainSwitchThreshold"]))

            # ---exponents and weights for position, history--:
            composer.push(float(blobs["probabilityExponentPosition"]))
            composer.push(float(blobs["probabilityExponentHistory"]))
            composer.push(float(blobs["probabilityWeightPosition"]))
            composer.push(float(blobs["probabilityWeightHistory"]))

            # -desc trend and orig pos---
            composer.push(blobs["DescTrend"])
            if blobs["DescTrend"]:
                descrCoordtrend = blobs["DescriptorCoordinatesTrend"]
                probabilityExponentTrend = float(blobs["probabilityExponentTrend"])
                composer.push(probabilityExponentTrend)
                probabilityWeightTrend = float(blobs["probabilityWeightTrend"])
                composer.push(probabilityWeightTrend)
                descrTrendMethod = int(blobs["descrTrendMethod"])
                composer.push(descrTrendMethod)

            composer.push(blobs["OrigPosition"])
            if blobs["OrigPosition"]:
                OrigPos = blobs["PositionWithinAudioFile"]
                probabilityExponentOriginalPos = float(blobs["probabilityExponentOriginalPos"])
                composer.push(probabilityExponentOriginalPos)
                probabilityWeightOriginalPos = float(blobs["probabilityWeightOriginalPos"])
                composer.push(probabilityWeightOriginalPos)

            # -------probability groups----#
            probabilityGroups = blobs["probabilityGroups"]
            composer.push(probabilityGroups)  # flags

            if probabilityGroups:
                P = int(blobs["P"])
                probabilityGroupVal = blobs["probabilityGroupVal"]
                grainProbabilityGroup = blobs["grainProbabilityGroup"]
                perGrainProbVal = blobs["perGrainProbabilityVal"]

                composer.push(P)
                for ind in range(P):
                    composer.push((probabilityGroupVal[ind]))

            # ----random selection----#
            composer.push(blobs["randomSelection"])

            # -interpolation----#
            useInterpolation = blobs["useInterpolation"]
            composer.push(useInterpolation)

            if useInterpolation:
                composer.push(float(blobs["grainInterpDistThreshold"]))
                composer.push(blobs["useInterpolationResampling"])
                interpMethod = int(blobs["interpMethod"])
                composer.push(interpMethod)
                pitchDimension = blobs["pitchDimension"]
                for d in range(dim):
                    composer.push(bool(pitchDimension[d]))

            composer.push(blobs["grain_overlap_flag"])
            # --grainOverlapAmount--#
            if blobs["grain_overlap_flag"]:
                composer.push(blobs["GrainOverlapAmount"])

            # -------------write grain---------------------------#
            for ind in range(blobs["numGrains"]):

                descriptor_coord = descrCoord[ind]

                for d in range(dim):
                    composer.push(descriptor_coord[d])

                composer.push(grainLens[ind])

                if blobs["DescTrend"]:
                    descriptor_coord_trend = descrCoordtrend[ind]
                    for d in range(dim):
                        composer.push(descriptor_coord_trend[d])

                if blobs["OrigPosition"]:
                    composer.push(OrigPos[ind])

                if blobs["probabilityGroups"]:
                    composer.push(int(grainProbabilityGroup[ind]))
                    composer.push(perGrainProbVal[ind])

                if channel > 1:
                    audio_grain = grains[ind]

                    for ch_ind in range(channel):
                        audio_grain_ch = audio_grain[ch_ind]

                        if blobs["GranularAudioEncodingMode"] == 0:
                            # uncompressed PCM
                            for grainInd in range(len(audio_grain_ch)):
                                composer.push(audio_grain_ch[grainInd])
                else:
                    audio_grain = grains[ind]

                    if blobs["GranularAudioEncodingMode"] == 0:
                        for grainInd in range(len(audio_grain)):
                            composer.push(audio_grain[grainInd])

                # -------------------------------------------------------#

        try:
            self.payload_as_bitstream = composer.get_bitstream()
            # bitstream decomposer and write:
            graindb = BitstreamDecomposer(self.payload_as_bitstream, granular_database_syntax)

        except NotEnoughDataPushedError as exc:
            raise NotEnoughDataPushedError(f"GranularDatabasePayload: {exc}") from exc

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        """Return the type of this Payload."""
        return ScenePayloadPacketType.GRANULAR

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        """Return the artifact subscription for this Payload."""
        return ["grain_coded", "bs_ids"]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        """Return the argument subscription for this Payload."""
        return []

    def get_bitstream(self) -> Bitstream:
        """Return the payload as a bitstream."""
        return Bitstream() if self.skipDatabase else self.payload_as_bitstream
