# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List

from mpegi.framework import Bitstream, BitstreamComposer, Payload, ScenePayloadPacketType, tokens
from mpegi.metadata.lcearly import (
    LCEREnvironmentType,
    LCEROutdoorData,
)

from .. import _constants

bounding_box_syntax: List[tokens.Token] = [
    tokens.writePosition("boundingBoxPos"),
    tokens.writeOrientation("boundingBoxOri"),
    tokens.writePosition("boundingBoxSize"),
]


outdoor_environment_syntax: List[tokens.Token] = [
    tokens.writeGain("ampFac1", isHiPrecGain=True),
    tokens.writeGain("ampFac2", isHiPrecGain=True),
    tokens.UnsignedInteger("predelay1", width=16),
    tokens.UnsignedInteger("predelay2", width=16),
    *bounding_box_syntax,
]

syntax: List[tokens.Token] = [
    tokens.CommentToken("LC ER PAYLOAD"),
    tokens.UnsignedInteger("environmentCount", width=_constants.bitsForMaxEnvs),
    tokens.Repeat(
        "environmentCount",
        syntax=[
            tokens.UnsignedInteger("environmentType", width=8),
            tokens.UnsignedInteger("priority", width=8),
            tokens.Switch(
                "environmentType",
                cases=[
                    tokens.SwitchCase(case=2, syntax=outdoor_environment_syntax),
                ],
            ),
        ],
    ),
]


class LCERPayload(Payload):
    def __init__(self, args, artifacts):
        super().__init__(args, artifacts)
        composer = BitstreamComposer(syntax)
        isSmallScene = artifacts["scene_info"]["isSmallScene"]

        lcer_component_data = artifacts["lcerComponentData"]
        composer.push(len(lcer_component_data.environments))
        for env in lcer_component_data.environments:
            composer.push(env.type.value)
            composer.push(env.priority)
            if env.type == LCEREnvironmentType.Outdoor:
                self.add_outdoor_environment(composer, env.outdoorData, isSmallScene)

        self.payload_as_bitstream = composer.get_bitstream()

    @staticmethod
    def add_outdoor_environment(
        composer: BitstreamComposer, outdoor_data: LCEROutdoorData, isSmallScene: bool
    ) -> None:
        composer.push(outdoor_data.ampFac1)
        composer.push(outdoor_data.ampFac2)
        composer.push(outdoor_data.predelay1)
        composer.push(outdoor_data.predelay2)
        composer.push({"position": outdoor_data.boundingBox.position, "isSmallScene": isSmallScene})
        composer.push(outdoor_data.boundingBox.orientation)
        composer.push({"position": outdoor_data.boundingBox.size, "isSmallScene": isSmallScene})

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.LC_EARLY_REFLECTION

    @staticmethod
    def getArtifactSubscription():
        return ["bs_ids", "lcerComponentData", "scene_info"]

    @staticmethod
    def getArgumentSubscription():
        return []

    def get_bitstream(self) -> Bitstream:
        return self.payload_as_bitstream
