# This software module was originally developed by
#
# Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Koninklijke Philips N.V. retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import Any, Dict, List, Optional

from mpegi.framework import Bitstream, BitstreamComposer, Payload, ScenePayloadPacketType, tokens
from mpegi.metadata.eifplus import Portal

LosVoxelDataStruct: List[tokens.Token] = [
    tokens.writeCountOrIndex("n_voxels_x"),
    tokens.writeCountOrIndex("n_voxels_y"),
    tokens.writeCountOrIndex("n_voxels_z"),
    tokens.writePosition("areaCornerP1"),
    tokens.writePosition("areaCornerP2"),
    tokens.writeCountOrIndex("voxelSize"),
    tokens.writeCountOrIndex("areaPortalCount"),
    tokens.Repeat(
        "areaPortalCount",
        syntax=[
            tokens.writeID("portalID"),
            tokens.writeCountOrIndex("nLosBitMasks"),
            tokens.Repeat(
                "nLosBitMasks",
                syntax=[
                    tokens.writeCountOrIndex("nBitmaskBytes"),
                    tokens.Repeat(
                        "nBitmaskBytes", syntax=[tokens.UnsignedInteger("rleByte", width=8)]
                    ),
                ],
            ),
        ],
    ),
]

AELosVoxelData: List[tokens.Token] = [
    tokens.writeCountOrIndex("nAes"),
    tokens.Repeat(
        "nAes",
        syntax=[tokens.writeCountOrIndex("aeId"), *LosVoxelDataStruct],
    ),
]

outsideLosVoxelData: List[tokens.Token] = [
    tokens.writeCountOrIndex("nOutsideLosAreas"),
    tokens.Repeat(
        "nOutsideLosAreas",
        syntax=[*LosVoxelDataStruct],
    ),
]

from math import log10

syntax: List[tokens.Token] = [
    tokens.CommentToken("PORTAL PAYLOAD"),
    tokens.writeCountOrIndex("portalsInSceneCount"),
    tokens.Conditional(
        "portalsInSceneCount",
        false=[],
        true=[
            tokens.Flag("isExplicitPortalMode"),
            tokens.Repeat(
                "portalsInSceneCount",
                syntax=[
                    tokens.writeID("portalId"),
                    tokens.Flag("hasExtent"),
                    tokens.Conditional(
                        "hasExtent",
                        false=[],
                        true=[
                            tokens.writeID("portalExtentId"),
                        ],
                    ),
                    tokens.Conditional(
                        "isExplicitPortalMode",
                        true=[
                            tokens.Flag("isConnectedToAE"),
                            tokens.Conditional(
                                "isConnectedToAE",
                                false=[],
                                true=[
                                    tokens.writeID("portalEnvironment1"),
                                    tokens.Flag("isConnectedToSecondAE"),
                                    tokens.Conditional(
                                        "isConnectedToSecondAE",
                                        false=[],
                                        true=[
                                            tokens.writeID("portalEnvironment2"),
                                        ],
                                    ),
                                ],
                            ),
                            tokens.Flag("hasMaterial"),
                            tokens.Conditional(
                                "hasMaterial",
                                false=[],
                                true=[
                                    tokens.writeID("portalMaterialID"),
                                ],
                            ),
                            tokens.writePosition("portalNormal"),
                            tokens.Flag("explicitPortalProperties"),
                            tokens.Conditional(
                                "explicitPortalProperties",
                                false=[],
                                true=[
                                    tokens.writePosition("portalCenterPos"),
                                    tokens.writeDistance("equivalentPortalWidth"),
                                    tokens.writeDistance("equivalentPortalHeight"),
                                    tokens.writeDistance("equivalentPortalDepth"),
                                ],
                            ),
                            tokens.Flag("hasPortal2PortalFactors"),
                            tokens.Conditional(
                                "hasPortal2PortalFactors",
                                false=[],
                                true=[
                                    tokens.Repeat(
                                        "numPossiblePortalFactors",
                                        syntax=[
                                            tokens.Conditional(
                                                "arePortalsConnected",
                                                false=[],
                                                true=[
                                                    tokens.Flag("hasAcousticLink"),
                                                    tokens.Conditional(
                                                        "hasAcousticLink",
                                                        false=[],
                                                        true=[
                                                            tokens.writeValueFromLUT(
                                                                "portal2portalFactor1",
                                                                tableName="portal_to_portal_factor_codes",
                                                            ),
                                                            tokens.writeValueFromLUT(
                                                                "portal2portalFactor2",
                                                                tableName="portal_to_portal_factor_codes",
                                                            ),
                                                        ],
                                                    ),
                                                ],
                                            ),
                                        ],
                                    )
                                ],
                            ),
                        ],
                        false=[
                            tokens.writeID("portalParentEnvironment"),
                        ],
                    ),
                ],
            ),
            tokens.Conditional(
                "isExplicitPortalMode",
                true=[
                    tokens.Flag("hasAELosVoxelData"),
                    tokens.Conditional("hasAELosVoxelData", false=[], true=[*AELosVoxelData]),
                    tokens.Flag("hasOutsideLosVoxelData"),
                    tokens.Conditional(
                        "hasOutsideLosVoxelData", false=[], true=[*outsideLosVoxelData]
                    ),
                ],
                false=[],
            ),
        ],
    ),
]


def arePortalsConnected(p1: Portal, p2: Portal):
    match0 = (
        p1.acousticEnvironment1 == p2.acousticEnvironment1
        or p1.acousticEnvironment1 == p2.acousticEnvironment2
    )
    match1 = (
        p1.acousticEnvironment2 == p2.acousticEnvironment1
        or p1.acousticEnvironment2 == p2.acousticEnvironment2
    )

    return sum([match0, match1]) == 1


class PortalPayload(Payload):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Serialises a Portal Payload."""
        super().__init__(args, artifacts)
        composer = BitstreamComposer(syntax)
        isExplicitPortalMode = True

        portals = artifacts["eif_scene"].portals
        reverbParams = artifacts["reverb_params"]
        isSmallScene = artifacts["scene_info"]["isSmallScene"]
        nPortals = len(portals)
        p2pFactors = artifacts["portal_to_portal_factors"]
        #  check if we have to switch to implicit portal mode
        if nPortals == 0:
            portals = artifacts["portal_info"].allPortals
            isExplicitPortalMode = False
            # Update implicit portal length
            nPortals = len(portals)

        composer.push(nPortals)
        if nPortals > 0:
            composer.push(isExplicitPortalMode)

        for i in range(0, nPortals):
            p = portals[i]
            if isExplicitPortalMode:
                composer.push(artifacts["bs_ids"].getId(p.id))
                hasExtent = p.portalExtent != None
            else:
                composer.push(p.id)
                hasExtent = p.extent != None
            composer.push(hasExtent)
            if hasExtent:
                if isExplicitPortalMode:
                    composer.push(artifacts["bs_ids"].getId(p.portalExtent))
                else:
                    # bsId_portal
                    composer.push(artifacts["bs_ids"].getId(p.extent))
            if isExplicitPortalMode:
                isConnectedToAE = p.acousticEnvironment1 is not None
                isConnectedToSecondAE = p.acousticEnvironment2 is not None
                composer.push(isConnectedToAE)
                if isConnectedToAE:
                    composer.push(artifacts["bs_ids"].getId(p.acousticEnvironment1))
                    composer.push(isConnectedToSecondAE)
                    if isConnectedToSecondAE:
                        composer.push(artifacts["bs_ids"].getId(p.acousticEnvironment2))

                hasMaterial = p.material is not None
                composer.push(hasMaterial)
                if hasMaterial:
                    composer.push(artifacts["bs_ids"].getId(p.material))
                composer.push({"position": p.portalNormal, "isSmallScene": True})

                # isExplicitPortal = all(p.portalPosition != None) and all(p.portalSize != None)
                composer.push(p.hasExplicitPortal)
                if p.hasExplicitPortal:
                    composer.push({"position": p.portalPosition, "isSmallScene": isSmallScene})
                    composer.push({"distance": p.portalSize[0], "isSmallScene": isSmallScene})
                    composer.push({"distance": p.portalSize[1], "isSmallScene": isSmallScene})
                    composer.push({"distance": p.portalSize[2], "isSmallScene": isSmallScene})

                composer.scope.update({"numPossiblePortalFactors": i})
                hasPortal2PortalFactors = (
                    p2pFactors is not None and i != 0 and any(p2pFactors[i][:i])
                )
                composer.push(hasPortal2PortalFactors)
                if hasPortal2PortalFactors:
                    for j in range(0, i):
                        # don't include links that are below -95dB.
                        connected = arePortalsConnected(p, portals[j])
                        composer.scope.update({"arePortalsConnected": connected})
                        if connected:
                            hasAcousticLink = (p2pFactors[i][j] is not None) and (
                                float(p2pFactors[i][j]) > 0.000000000316
                            )
                            composer.push(hasAcousticLink)
                            if hasAcousticLink:
                                composer.push(10 * log10(p2pFactors[i][j]))
                                composer.push(10 * log10(p2pFactors[j][i]))
                        else:
                            # If they are not connected, remove the tokens from the stack.
                            composer.token_generator.token_stack.pop()

            else:
                # bsId_fromWhichAE
                composer.push(artifacts["bs_ids"].getId(p.sourceId))

        if isExplicitPortalMode:
            aeLosVoxelData = reverbParams.aeLosVoxelData
            outsideLosVoxelData = reverbParams.outsideLosVoxelData
            hasAELosVoxelData = aeLosVoxelData != None
            composer.push(hasAELosVoxelData)

            if hasAELosVoxelData:
                n_aes = len(aeLosVoxelData)
                composer.push(n_aes)
                for ae_id, ae_los_voxel_data in aeLosVoxelData.items():
                    ae_los_voxels_rle_encoded = ae_los_voxel_data["portal_los_voxels_rle_encoded"]
                    ae_portals = ae_los_voxel_data["ae_portals"]
                    ae_portal_count = len(ae_portals)
                    los_voxel_side_len_cm = ae_los_voxel_data["los_voxel_side_len_cm"]
                    ae_corner_p1 = ae_los_voxel_data["ae_corner_points"][0]
                    ae_corner_p2 = ae_los_voxel_data["ae_corner_points"][1]
                    n_voxels_x, n_voxels_y, n_voxels_z = ae_los_voxel_data["ae_n_voxels"]

                    composer.push(artifacts["bs_ids"].getId(ae_id))
                    composer.push(n_voxels_x)
                    composer.push(n_voxels_y)
                    composer.push(n_voxels_z)
                    composer.push({"position": ae_corner_p1, "isSmallScene": isSmallScene})
                    composer.push({"position": ae_corner_p2, "isSmallScene": isSmallScene})
                    composer.push(los_voxel_side_len_cm)

                    composer.push(ae_portal_count)
                    for i in range(ae_portal_count):
                        portal = ae_portals[i]
                        composer.push(artifacts["bs_ids"].getId(portal.id))
                        portal_los_data = ae_los_voxels_rle_encoded[portal.id]
                        n_masks = len(portal_los_data)
                        composer.push(n_masks)
                        for j in range(n_masks):
                            los_bitmask = portal_los_data[j]
                            n_bytes = len(los_bitmask)
                            composer.push(n_bytes)
                            for byte in los_bitmask:
                                composer.push(byte)

            hasOutsideLosVoxelData = outsideLosVoxelData != None
            composer.push(hasOutsideLosVoxelData)
            if hasOutsideLosVoxelData:
                nOutsideLosVoxelAreas = len(outsideLosVoxelData)
                composer.push(nOutsideLosVoxelAreas)
                for outside_ae_id, outside_los_voxel_data in outsideLosVoxelData.items():
                    outside_los_voxels_rle_encoded = outside_los_voxel_data[
                        "portal_los_voxels_rle_encoded"
                    ]
                    area_portals = outside_los_voxel_data["area_portals"]
                    area_portal_count = len(area_portals)
                    los_voxel_side_len_cm = outside_los_voxel_data["los_voxel_side_len_cm"]
                    area_corner_p1 = outside_los_voxel_data["diag_corner_points"][0]
                    area_corner_p2 = outside_los_voxel_data["diag_corner_points"][1]
                    n_voxels_x, n_voxels_y, n_voxels_z = outside_los_voxel_data["n_voxels"]

                    composer.push(n_voxels_x)
                    composer.push(n_voxels_y)
                    composer.push(n_voxels_z)
                    composer.push({"position": area_corner_p1, "isSmallScene": isSmallScene})
                    composer.push({"position": area_corner_p2, "isSmallScene": isSmallScene})
                    composer.push(los_voxel_side_len_cm)

                    composer.push(area_portal_count)
                    for i in range(area_portal_count):
                        portal = area_portals[i]
                        composer.push(artifacts["bs_ids"].getId(portal.id))
                        portal_los_data = outside_los_voxels_rle_encoded[portal.id]
                        n_masks = len(portal_los_data)
                        composer.push(n_masks)
                        for j in range(n_masks):
                            los_bitmask = portal_los_data[j]
                            n_bytes = len(los_bitmask)
                            composer.push(n_bytes)
                            for byte in los_bitmask:
                                composer.push(byte)

        self.payload_as_bitstream = composer.get_bitstream()

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.PORTAL

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "bs_ids",
            "portal_info",
            "scene_info",
            "eif_scene",
            "reverb_params",
            "portal_to_portal_factors",
        ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return []

    def get_bitstream(self) -> Bitstream:
        return self.payload_as_bitstream
