# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2023.

from typing import Any, Dict, List
import logging
import numpy as np

from mpegi.framework import (
    Bitstream,
    BitstreamComposer,
    Payload,
    ScenePayloadPacketType,
    tokens,
)
from mpegi.framework.generic_codebook import (
    SymbolHistogram,
    get_encoder_config,
)

from .. import _constants

rastermap_layer_syntax: List[tokens.Token] = [
    tokens.CodebookConfig("codebookVcX"),
    tokens.CodebookConfig("codebookVcY"),
    tokens.CodebookConfig("codebookWpX"),
    tokens.CodebookConfig("codebookWpY"),
    tokens.writeID("numRasterMapVoxels"),
    tokens.Repeat(
        "numRasterMapVoxels",
        syntax=[
            tokens.Flag("hasVoxelCoordY"),
            tokens.Conditional(
                "hasVoxelCoordY",
                true=[
                    tokens.CodebookSymbol("voxelCoordY", config="codebookVcY"),
                    tokens.Flag("hasVoxelCoordX"),
                    tokens.Conditional(
                        "hasVoxelCoordX",
                        true=[
                            tokens.CodebookSymbol("voxelCoordX", config="codebookVcX"),
                        ],
                        false=[],
                    ),
                ],
                false=[],
            ),
            tokens.UnsignedInteger("refMode", width=2),
            tokens.Conditional(
                "refMode",
                true=[],
                false=[
                    tokens.UnsignedInteger("mapChange", width=1),
                    tokens.CodebookSymbol("waypointX", config="codebookWpX"),
                    tokens.CodebookSymbol("waypointY", config="codebookWpY"),
                ],
            ),
        ],
    ),
]

rastermap_subscene_syntax: List[tokens.Token] = [
    tokens.Flag("hasVoxSceneDiffractionRasterMap"),
    tokens.Conditional(
        "hasVoxSceneDiffractionRasterMap",
        true=[
            tokens.writeCountOrIndex("rasterMapWidth"),
            tokens.writeCountOrIndex("rasterMapHeight"),
            tokens.writeCountOrIndex("numRasterMapSources"),
            tokens.Repeat(
                "numRasterMapSources",
                syntax=[
                    tokens.writeCountOrIndex("rasterMapSourceIdLen"),
                    tokens.Repeat(
                        "rasterMapSourceIdLen",
                        syntax=[
                            tokens.writeID("rasterMapSourceId"),
                        ],
                    ),
                    tokens.UnsignedInteger("numRasterMapLayers", width=3),
                    tokens.Repeat(
                        "numRasterMapLayers",
                        syntax=rastermap_layer_syntax,
                    ),
                ],
            ),
        ],
        false=[],
    ),
]

syntax: List[tokens.Token] = [
    tokens.CommentToken("RASTERMAP PAYLOAD"),
    tokens.writeCountOrIndex("num_rastermap_subscenes"),
    tokens.Repeat("num_rastermap_subscenes", syntax=rastermap_subscene_syntax),
]


class RasterMapPayload(Payload):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Serialises a Portal Payload."""
        super().__init__(args, artifacts)

        self.isSmallScene = artifacts["scene_info"]["isSmallScene"]
        self.rasterMapList = artifacts["rastermap"]

        # handle empty RasterMap data
        if self.rasterMapList is None:
            logging.getLogger("RasterMapPayload").warning("Skipping empty rastermap payload")
            composer = BitstreamComposer([])
            self.payload_as_bitstream = composer.get_bitstream()
            return

        composer = BitstreamComposer(syntax)

        # FIXME: remove this
        self.entry_counter = 0

        num_subscenes = len(self.rasterMapList)
        composer.push(num_subscenes)
        for rasterMapDicts in self.rasterMapList:
            self._serialize_subscene(composer, rasterMapDicts)

        self.payload_as_bitstream = composer.get_bitstream()

        # FIXME: remove this debug code
        print("")
        print("Total number of RasterMap entries: " + str(self.entry_counter))
        print("")

    def _serialize_subscene(self, composer: BitstreamComposer, rasterMapDicts: List) -> None:
        numRasterMapSources = len(rasterMapDicts)

        hasVoxSceneDiffractionRasterMap = numRasterMapSources > 0
        composer.push(hasVoxSceneDiffractionRasterMap)
        if hasVoxSceneDiffractionRasterMap:

            rmMatrix = rasterMapDicts[0]["wp_matrix"]
            rasterMapWidth = rmMatrix.shape[0]
            rasterMapHeight = rmMatrix.shape[1]
            composer.push(rasterMapWidth)
            composer.push(rasterMapHeight)

            composer.push(numRasterMapSources)
            for rasterMapDict in rasterMapDicts:
                srcIds = rasterMapDict["src_id"]
                rmMatrix = rasterMapDict["wp_matrix"]
                rasterMapSourceIdLen = len(srcIds)
                composer.push(rasterMapSourceIdLen)
                for srcId in srcIds:
                    composer.push(srcId)

                # FIXME: remove this debug code
                print("")
                print("Source IDs: " + str(srcIds))

                numLayers = rmMatrix.shape[2]
                composer.push(numLayers)
                for layer in range(numLayers):
                    self._serialize_layer(composer, rmMatrix, layer)

    def _serialize_layer(self, composer: BitstreamComposer, rmMatrix: np.ndarray, layer: int):
        tokenList = []
        symbolListVcX = []
        symbolListVcY = []
        symbolListWpX = []
        symbolListWpY = []

        dimX = rmMatrix.shape[0]
        dimY = rmMatrix.shape[1]

        numRasterMapVoxels = 0

        vcX = 0
        vcY = -1
        for x in range(dimX):
            for y in range(dimY):
                wp = rmMatrix[x, y, layer, :]
                if (wp < 0).any():
                    continue

                numRasterMapVoxels += 1

                vcY += 1
                hasVoxelCoordY = vcY != y or vcX != x
                tokenList.append(hasVoxelCoordY)
                if hasVoxelCoordY:
                    vcY = y
                    tokenList.append(vcY)
                    symbolListVcY.append(vcY)

                    vcX += 1
                    hasVoxelCoordX = vcX != x
                    tokenList.append(hasVoxelCoordX)
                    if hasVoxelCoordX:
                        vcX = x
                        tokenList.append(vcX)
                        symbolListVcX.append(vcX)

                refMode = 0
                if x > 0 and tuple(rmMatrix[x - 1, y, layer, :]) == tuple(wp):
                    refMode = 1
                elif y > 0 and tuple(rmMatrix[x, y - 1, layer, :]) == tuple(wp):
                    refMode = 2
                elif layer > 0 and tuple(rmMatrix[x, y, layer - 1, :]) == tuple(wp):
                    refMode = 3
                tokenList.append(refMode)

                if not refMode:
                    mapChange = int(layer - wp[2])
                    tokenList.append(mapChange)
                    tokenList.append(wp[0])
                    tokenList.append(wp[1])
                    symbolListWpX.append(wp[0])
                    symbolListWpY.append(wp[1])

        histogramVcX = SymbolHistogram.from_plain(symbolListVcX)
        histogramVcY = SymbolHistogram.from_plain(symbolListVcY)
        histogramWpX = SymbolHistogram.from_plain(symbolListWpX)
        histogramWpY = SymbolHistogram.from_plain(symbolListWpY)
        encoderVcX = get_encoder_config(histogramVcX)
        encoderVcY = get_encoder_config(histogramVcY)
        encoderWpX = get_encoder_config(histogramWpX)
        encoderWpY = get_encoder_config(histogramWpY)
        composer.push(encoderVcX)
        composer.push(encoderVcY)
        composer.push(encoderWpX)
        composer.push(encoderWpY)

        composer.push(numRasterMapVoxels)

        for token in tokenList:
            composer.push(token)

        # FIXME: remove this debug code
        print("layer " + str(layer) + ": " + str(numRasterMapVoxels) + " RasterMap entries")
        self.entry_counter += numRasterMapVoxels

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.RASTERMAP

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "scene_info",
            "rastermap",
        ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return []

    def get_bitstream(self) -> Bitstream:
        return self.payload_as_bitstream
