# This software module was originally developed jointly by
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V.
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS, Nokia and Koninklijke Philips N.V. retain full right to
# modify and use the code for their own purpose, assign or donate the code
# to a third party without consent of and without accounting to third
# parties, and to inhibit third parties from using the code for products
# that do not conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

import itertools
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

import numpy as np

import mpegi.framework._lookup_tables as LUT
import mpegi.metadata.reverb.constants as constants
import mpegi.encoder._constants as _constants
from mpegi.framework import Bitstream, BitstreamComposer, Payload, ScenePayloadPacketType, tokens
from mpegi.metadata.eifplus import AcousticEnvironment
from mpegi.metadata.eifplus.scene import SceneType

ReverbFreqGridData: List[tokens.Token] = [
    tokens.writeCountOrIndex("fgdNrGrids"),
    tokens.Repeat(
        "fgdNrGrids",
        syntax=[
            tokens.UnsignedInteger("fgdMethod", width=2),
            tokens.Switch(
                "fgdMethod",
                cases=[
                    tokens.SwitchCase(
                        0,
                        syntax=[
                            tokens.writeCountOrIndex("fgdNrBands", storeValue=True),
                            tokens.Repeat(
                                "fgdNrBands", syntax=[tokens.writeFrequency("fgdCenterFreq")]
                            ),
                        ],
                    ),  # individual freq
                    tokens.SwitchCase(
                        1,
                        syntax=[
                            tokens.writeCountOrIndex("fgdNrBands", storeValue=True),
                            tokens.writeFrequency("fgdCenterFreq"),
                            tokens.writeValueFromLUT(
                                "frequencyHop", tableName="frequency_hop_codes"
                            ),
                        ],
                    ),  # start-hop
                    tokens.SwitchCase(
                        2,
                        syntax=[
                            tokens.UnsignedInteger("fgdDefaultGrid", width=4),
                            tokens.Flag("fgdIsSubGrid"),
                            tokens.Conditional(
                                "fgdIsSubGrid",
                                false=[],
                                true=[
                                    tokens.UnsignedInteger("fgdDefaultGridOffset", width=3),
                                    tokens.UnsignedInteger("fgdDefaultGridNrBands", width=6),
                                ],
                            ),
                        ],
                    ),  # Default Banding
                    tokens.SwitchCase(3, syntax=[]),  # reserved
                ],
            ),
        ],
    ),
]

reverbSyntax: List[tokens.Token] = [
    tokens.Flag("revPresent"),
    tokens.Conditional(
        "revPresent",
        false=[],
        true=[
            tokens.writeCountOrIndex("revNrElements"),
            tokens.Repeat(
                "revNrElements",
                syntax=[
                    tokens.writeID("revAcEnvId"),
                    tokens.writeCountOrIndex("revFreqGridIdx"),
                    tokens.writeDuration(
                        "revPredelay", isLongRangeMode=constants.default_use_long_range_mode
                    ),
                    tokens.Flag("revDelayLineCountPresent"),
                    tokens.Conditional(
                        "revDelayLineCountPresent",
                        false=[],
                        true=[
                            tokens.writeValueFromLUT(
                                "revDelayLineCount", tableName="delay_line_count_codes"
                            ),
                        ],
                    ),
                    tokens.Repeat(
                        "fgdNrBands",
                        # "revFreqGridIdx",
                        syntax=[
                            tokens.writeDuration(
                                "revRT60", isLongRangeMode=constants.default_use_long_range_mode
                            )
                        ],
                    ),
                    tokens.Repeat(
                        "fgdNrBands",
                        # "revFreqGridIdx",
                        syntax=[tokens.writeValueFromLUT("revDSR", tableName="dsr_code")],
                    ),
                ],
            ),
            tokens.Flag("revRenderControlPresent"),
            tokens.Conditional(
                "revRenderControlPresent",
                false=[],
                true=[
                    tokens.writeGain(
                        "revDistanceGainDropDb", isHiPrecGain=constants.default_is_hi_prec_gain
                    ),
                    tokens.writeDistance(
                        "revMinDistance", isSmallScene=constants.default_is_small_scene
                    ),
                ],
            ),
            tokens.Flag("revMultiRoomConfigPresent"),
            tokens.Conditional(
                "revMultiRoomConfigPresent",
                false=[],
                true=[
                    tokens.Flag("revFadeEnabled"),
                    tokens.Conditional(
                        "revFadeEnabled",
                        false=[],
                        true=[
                            tokens.writeDistance(
                                "revFadeDistance", isSmallScene=constants.default_is_small_scene
                            )
                        ],
                    ),
                    tokens.Flag("revHeadTrackingEnabled"),
                ],
            ),
            tokens.Flag("revARConfigPresent"),
            tokens.Conditional(
                "revARConfigPresent",
                false=[],
                true=[
                    tokens.writeGain(
                        "revLsdfTheoreticalRDRtoactualRDRdB",
                        isHiPrecGain=constants.default_is_hi_prec_gain,
                    ),
                    tokens.writeGain(
                        "revLsdfRDRtoactualRDRdB",
                        isHiPrecGain=constants.default_is_hi_prec_gain,
                    ),
                    tokens.Flag("revOverrideLSDF_RDR"),
                ],
            ),
        ],
    ),
    tokens.Flag("revDefaultAEPresent"),
    tokens.Conditional(
        "revDefaultAEPresent",
        false=[],
        true=[
            tokens.Flag("revHasCustomDefaultAE"),
            tokens.Conditional(
                "revHasCustomDefaultAE",
                true=[
                    tokens.writeCountOrIndex("revDefaultAEFreqGridIdx"),
                    tokens.writeDuration("revDefaultAEPredelay"),
                    tokens.Flag("revDefaultAEDelayLineCountPresent"),
                    tokens.Conditional(
                        "revDefaultAEDelayLineCountPresent",
                        false=[],
                        true=[
                            tokens.writeValueFromLUT(
                                "revDefaultAEDelayLineCount", tableName="delay_line_count_codes"
                            ),
                        ],
                    ),
                    tokens.Float32("revDefaultAEDelayRatio"),
                    tokens.Float32("revDefaultAEDelayMinMs"),
                    tokens.Repeat(
                        "fgdNrBands",
                        syntax=[
                            tokens.writeDuration(
                                "revDefaultAERT60",
                                isLongRangeMode=constants.default_use_long_range_mode,
                            )
                        ],
                    ),
                    tokens.Repeat(
                        "fgdNrBands",
                        syntax=[tokens.writeValueFromLUT("revDefaultAEDSR", tableName="dsr_code")],
                    ),
                    tokens.writeGain(
                        "revDefaultAEDistanceGainDropDb",
                        isHiPrecGain=constants.default_is_hi_prec_gain,
                    ),
                    tokens.writeDistance(
                        "revDefaultAEMinimumDistance", isSmallScene=constants.default_is_small_scene
                    ),
                ],
                false=[
                    tokens.UnsignedInteger("revDefaultAEPreset", width=6),
                    tokens.Flag("revHasDefaultAEGain"),
                    tokens.Conditional(
                        "revHasDefaultAEGain",
                        false=[],
                        true=[
                            tokens.writeGain(
                                "revDefaultAEGain", isHiPrecGain=constants.default_is_hi_prec_gain
                            )
                        ],
                    ),
                    tokens.Flag("revHasDefaultAEDelayRatio"),
                    tokens.Conditional(
                        "revHasDefaultAEDelayRatio",
                        false=[],
                        true=[tokens.Float32("revDefaultAEDelayRatio")],
                    ),
                    tokens.Flag("revHasDefaultAEDelayMinMs"),
                    tokens.Conditional(
                        "revHasDefaultAEDelayMinMs",
                        false=[],
                        true=[tokens.Float32("revDefaultAEDelayMinMs")],
                    ),
                    tokens.Flag("revHasDefaultAEDistanceGainDrop"),
                    tokens.Conditional(
                        "revHasDefaultAEDistanceGainDrop",
                        false=[],
                        true=[
                            tokens.writeGain(
                                "revDefaultAEDistanceGainDropDb",
                                isHiPrecGain=constants.default_is_hi_prec_gain,
                            )
                        ],
                    ),
                    tokens.Flag("revHasDefaultAEMinDist"),
                    tokens.Conditional(
                        "revHasDefaultAEMinDist",
                        false=[],
                        true=[tokens.writeDistance("revDefaultAEMinimumDistance")],
                    ),
                ],
            ),
        ],
    ),
]


@dataclass
class FrequencyGrid:
    acousticEnvironmentId: str
    rt60: List[float]
    dsr: List[float]
    fgdNrBands: int
    frequencies: List[float]
    preDelay: float
    complexityLevel: int


class LateReverbPayload(Payload):
    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Serialises a ReverbPacket packet. The packet is comprised of reverberator parameters."""
        super().__init__(args, artifacts)

        self.skipSerializingPacket = False

        reverbParams: Optional[ReverbParameters] = artifacts["reverb_params"]

        if not reverbParams or (
            len(reverbParams.acousticEnvironments) == 0 and reverbParams.defaultAE.enabled == False
        ):
            logging.info("Skipping reverberator packet creation because there are no reverb params")
            self.skipSerializingPacket = True
            return

        composer = BitstreamComposer(reverbSyntax)
        self._writeReverbPayload(reverbParams, artifacts, composer)

        self.payload_as_bitstream = composer.get_bitstream()

    def _writeReverbPayload(
        self, reverbParams, artifacts: Dict[str, Any], composer: BitstreamComposer
    ):
        freqGridData = artifacts["reverb_frequency_grids"]
        acousticEnvironments = reverbParams.acousticEnvironments

        revNrElements = len(acousticEnvironments)
        DAE = reverbParams.defaultAE
        revPresent = revNrElements > 0 or DAE.environment != None
        revDefaultAEPresent = DAE.enabled

        revRendererControl = reverbParams.revRendererControl
        revRendererControlPresent = (
            True
            if revRendererControl != None
            and (
                revRendererControl.minDist != None
                or revRendererControl.distanceDropDb != None
                or revRendererControl.fadeInDistance
            )
            else False
        )

        isExplicitPortalMode = True
        if len(artifacts["eif_scene"].portals) == 0:
            isExplicitPortalMode = False

        composer.push(revPresent)

        if revPresent:
            composer.push(revNrElements)

            for e in range(revNrElements):
                ae = acousticEnvironments[e]
                composer.push(artifacts["bs_ids"][ae.id])

                aeGrid = freqGridData[ae.id]
                gridIdx = artifacts["frequency_grid_mapping"][ae.id]
                composer.push(gridIdx)

                # Reverb predelay duration
                composer.push(round(aeGrid.preDelay, 3))
                composer.scope.update({"fgdNrBands": aeGrid.fgdNrBands})

                hasComplexityLevel = aeGrid.complexityLevel != None
                composer.push(hasComplexityLevel)
                if hasComplexityLevel:
                    delayLineCount = self.complexityLevelToDelayLineCount(aeGrid.complexityLevel)
                    composer.push(delayLineCount)

                for b in range(aeGrid.fgdNrBands):
                    composer.push(round(aeGrid.rt60[b], 3))

                for b in range(aeGrid.fgdNrBands):
                    dsr = self.dsrLinToLog(aeGrid.dsr[b])
                    composer.push(round(dsr))

            fade_in_distance = constants.default_fade_in_distance
            fade_in_enabled = constants.default_fade_in_enabled

            composer.push(revRendererControlPresent)
            if revRendererControlPresent:
                composer.push(
                    revRendererControl.distanceDropDb
                    if revRendererControl.distanceDropDb != None
                    else constants.default_reverb_distance_gain_drop_db
                )
                composer.push(
                    revRendererControl.minDist
                    if revRendererControl.minDist != None
                    else constants.default_reverb_minimum_distance
                )
                if revRendererControl.fadeInDistance != None:
                    fade_in_enabled = True
                    fade_in_distance = revRendererControl.fadeInDistance

            if not isExplicitPortalMode:
                fade_in_distance = constants.default_fade_in_distance_fallback_mode

            composer.push(constants.default_multi_room_control_present)
            if constants.default_multi_room_control_present:
                composer.push(fade_in_enabled)
                if fade_in_enabled:
                    composer.push(fade_in_distance)

                composer.push(constants.default_headtracking_enabled)

            # lsdf parameters
            revARConfigPresent = reverbParams.sceneType == SceneType.AR
            composer.push(revARConfigPresent)
            if revARConfigPresent:
                composer.push(constants.lsdf_theoretical_rdr_to_actual_rdr)
                composer.push(constants.lsdf_rdr_to_actual_rdr)
                composer.push(constants.lsdf_override_lsdf_ddr)

        composer.push(revDefaultAEPresent)
        if revDefaultAEPresent:
            customDefaultAE = DAE.environment != None
            composer.push(customDefaultAE)
            if customDefaultAE:
                aeGrid = freqGridData[DAE.environment.id]
                gridIdx = artifacts["frequency_grid_mapping"][DAE.environment.id]
                composer.push(gridIdx)

                # Reverb predelay duration
                composer.push(round(aeGrid.preDelay, 3)),

                hasComplexityLevel = aeGrid.complexityLevel != None
                composer.push(hasComplexityLevel)
                if hasComplexityLevel:
                    delayLineCount = self.complexityLevelToDelayLineCount(aeGrid.complexityLevel)
                    composer.push(delayLineCount)

                composer.push(
                    DAE.density if DAE.density != None else constants.default_ae_delay_density
                ),

                composer.push(
                    DAE.delayMinMs
                    if DAE.delayMinMs != None
                    else constants.default_ae_delay_delayMinMs
                ),
                composer.scope.update({"fgdNrBands": aeGrid.fgdNrBands})
                for b in range(aeGrid.fgdNrBands):
                    composer.push(round(aeGrid.rt60[b], 3))

                for b in range(aeGrid.fgdNrBands):
                    dsr = self.dsrLinToLog(aeGrid.dsr[b])
                    if DAE.gainDb != None:
                        dsr += DAE.gainDb
                    composer.push(round(dsr))
                composer.push(
                    DAE.distanceDropDb
                    if DAE.distanceDropDb != None
                    else constants.default_ae_distance_gain_drop_db
                ),
                composer.push(
                    DAE.minDist if DAE.minDist != None else constants.default_ae_minimum_distance
                )
            else:
                try:
                    defaultAEPreset = LUT.default_AE_codes[DAE.preset]
                except:
                    raise ValueError(f"Unknown Default AE Preset: {DAE.preset}")
                composer.push(defaultAEPreset)

                hasDefaultAeGain = DAE.gainDb != None
                composer.push(hasDefaultAeGain)
                if hasDefaultAeGain:
                    composer.push(DAE.gainDb)

                hasDelayRatio = DAE.density != None
                composer.push(hasDelayRatio)
                if hasDelayRatio:
                    composer.push(DAE.density)

                hasDelayMinMs = DAE.delayMinMs != None
                composer.push(hasDelayMinMs)
                if hasDelayMinMs:
                    composer.push(DAE.delayMinMs)

                hasDistGainDrop = DAE.distanceDropDb != None
                composer.push(hasDistGainDrop)
                if hasDistGainDrop:
                    composer.push(DAE.distanceDropDb)

                hasMinDist = DAE.minDist != None
                composer.push(hasMinDist)
                if hasMinDist:
                    composer.push(DAE.minDist)

    def dsrLinToLog(self, dsrValue):
        dsrMin = min(LUT.dsr_code.keys())
        dsrMax = max(LUT.dsr_code.keys())
        if dsrValue == float("-inf") or dsrValue == 0:
            LogDSR = dsrMin
            logging.warning(
                "Reverb: DSR Value {} less than minimum code, clamping to {}".format(LogDSR, dsrMin)
            )
        else:
            LogDSR = round(10 * np.log10(dsrValue))

        if LogDSR < dsrMin:
            logging.warning(
                "Reverb: DSR Value {} less than minimum code, clamping to {}".format(LogDSR, dsrMin)
            )
            LogDSR = dsrMin
        if LogDSR > dsrMax:
            logging.warning(
                "Reverb: DSR Value {} greater than maximum code, clamping to {}".format(
                    LogDSR, dsrMax
                )
            )
            LogDSR = dsrMax
        return LogDSR

    def complexityLevelToDelayLineCount(self, complexityLevel):
        defaultNumberOfDl = 15
        if complexityLevel == 1:
            return 3
        elif complexityLevel == 2:
            return 7
        elif complexityLevel == 3:
            return defaultNumberOfDl
        elif complexityLevel == 4:
            return 31
        elif complexityLevel == 5:
            return 63
        else:
            logging.warning(
                "Reverb: complexityLevel value {} unsupported, fallback to default number of DL {}".format(
                    complexityLevel, defaultNumberOfDl
                )
            )
            return defaultNumberOfDl

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.LATE_REVERB

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "reverb_params",
            "bs_ids",
            "reverb_frequency_grids",
            "frequency_grids",
            "frequency_grid_mapping",
            "eif_scene",
        ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return []

    def get_bitstream(self) -> Bitstream:
        return Bitstream() if self.skipSerializingPacket else self.payload_as_bitstream
