# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import Any, Dict, List

from mpegi.framework import Bitstream, BitstreamComposer, Payload, ScenePayloadPacketType, tokens
from mpegi.metadata.eifplus import SceneMetadata

from .. import _constants
from ..components import BitstreamIdentifiers
from . import _scene_serializers


write_scene_payload = tokens.Function(
    "writeScenePayload",
    syntax=[
        tokens.QuantizedFloat("sceneDuration", quant_range=_constants.scene_duration_qrange),
        tokens.Flag("sceneType"),
        tokens.NewLine(),
        _scene_serializers.write_transforms,
        _scene_serializers.write_anchors,
        _scene_serializers.write_audio_streams,
        _scene_serializers.write_granular_sounds,
        _scene_serializers.write_materials,
        _scene_serializers.write_directivities,
        _scene_serializers.write_meshes,
        _scene_serializers.write_primitives,
        _scene_serializers.write_environments,
        _scene_serializers.write_object_sources,
        _scene_serializers.write_hoa_groups,
        _scene_serializers.write_hoa_sources,
        _scene_serializers.write_channel_sources,
        _scene_serializers.write_airflow_generators,
    ],
)


class ScenePayload(Payload):
    """This packet contains all of the Scene data required for an MPEG-I
    scene. This includes the base EIF information as well as any additional
    metadata.

    The Early Reflection and Diffraction component reconstruct the combined
    static mesh from the individual geometries in the following order:
    1) meshes
    2) primitive boxes
    3) primitive spheres
    4) primitive cylinders
    5) (acoustic environments)
    """

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        super().__init__(args, artifacts)
        composer = BitstreamComposer([write_scene_payload])

        metadata: SceneMetadata = artifacts["scene_metadata"]
        bs_ids: BitstreamIdentifiers = artifacts["bs_ids"]
        isSmallScene = artifacts["scene_info"]["isSmallScene"]
        enableLocallyCapturedAudio = artifacts["scene_info"]["enableLocallyCapturedAudio"]

        composer.push(float(artifacts["scene_info"]["duration"]))
        composer.push(bool(artifacts["scene_info"]["type"]))

        _scene_serializers.serialize_transforms(
            composer, artifacts["transforms"], bs_ids, isSmallScene
        )
        _scene_serializers.serialize_anchors(composer, artifacts["anchors"], bs_ids)

        streams = artifacts["audio_streams"]
        streams = [s for s in streams if metadata.isAudioStreamUsed(s.id)]
        _scene_serializers.serialize_audio_streams(
            composer,
            streams,
            bs_ids,
            enableLocallyCapturedAudio,
            metadata.eventStreams,
            metadata.eventDurations,
        )

        granular_sound = artifacts["granular_sound"]
        granular_sound = [s for s in granular_sound if metadata.isGranularSoundUsed(s.id)]
        _scene_serializers.serialize_granular_sound(composer, granular_sound, bs_ids)

        materials = artifacts["acoustic_materials"]
        materials = [mat for mat in materials if metadata.isMaterialUsed(mat.id)]
        _scene_serializers.serialize_materials(
            composer, materials, artifacts["frequency_grid_mapping"], bs_ids
        )
        _scene_serializers.serialize_direcitvities(composer, artifacts["directivities"], bs_ids)

        """The Early Reflection and Diffraction component reconstruct the
        static mesh from the individual geometries in the following order:
        1) meshes
        2) primitive boxes
        3) primitive spheres
        4) primitive cylinders
        5) further non-material geometries like URRs and portals
        """

        meshes = [*artifacts["meshes"]]
        # add meshes generated by automatic detection of portals in implicit portal mode
        if artifacts["portal_info"] != []:
            meshes = meshes + [*artifacts["portal_info"].portalGeometries]

        meshes = [m for m in meshes if len(m.vertices) > 0 and len(m.faces) > 0]
        _scene_serializers.serialize_meshes(composer, meshes, bs_ids, metadata, isSmallScene)

        primitives = [*artifacts["boxes"], *artifacts["spheres"], *artifacts["cylinders"]]
        if args["serializeURR"]:
            primitives.extend(artifacts["user_regions"])
        _scene_serializers.serialize_primitives(
            composer, primitives, bs_ids, metadata, isSmallScene
        )

        environments = artifacts["acoustic_environments"]
        dae = artifacts["default_acoustic_environment"]
        if dae and dae.environment != None:
            for ae in environments:
                if ae.id == dae.environment.id:
                    environments.remove(ae)
                    break

        _scene_serializers.serialize_environments(composer, environments, bs_ids, isSmallScene)

        object_sources = artifacts["object_sources"]
        object_sources = [o for o in object_sources if metadata.isAudioSourceRendered(o.id)]
        _scene_serializers.serialize_object_sources(
            composer, object_sources, bs_ids, metadata, isSmallScene
        )

        _scene_serializers.serialize_hoa_groups(
            composer, artifacts["hoa_groups"], bs_ids, isSmallScene
        )

        hoa_sources = artifacts["hoa_sources"]
        hoa_sources = [h for h in hoa_sources if metadata.isAudioSourceRendered(h.id)]
        _scene_serializers.serialize_hoa_sources(
            composer, hoa_sources, bs_ids, metadata, isSmallScene
        )

        channel_sources = artifacts["channel_sources"]
        channel_sources = [cs for cs in channel_sources if metadata.isAudioSourceRendered(cs.id)]
        _scene_serializers.serialize_channel_sources(
            composer, channel_sources, bs_ids, metadata, isSmallScene
        )

        airflow_generators = artifacts["airflow_generators"]
        _scene_serializers.serialize_airflow_generators(
            composer, airflow_generators, bs_ids, metadata, isSmallScene
        )

        self.payload_as_bitstream = composer.get_bitstream()

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.SCENE

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "scene_info",
            "acoustic_environments",
            "acoustic_materials",
            "airflow_generators",
            "anchors",
            "audio_streams",
            "boxes",
            "channel_sources",
            "cylinders",
            "directivities",
            "default_acoustic_environment",
            "granular_sound",
            "hoa_groups",
            "hoa_sources",
            "meshes",
            "object_sources",
            "spheres",
            "transforms",
            "user_regions",
            "bs_ids",
            "scene_metadata",
            "portal_info",
            "frequency_grids",
            "frequency_grid_mapping",
        ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return ["serializeURR", "disableDefaultAE"]

    def get_bitstream(self) -> Bitstream:
        return self.payload_as_bitstream
