# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List
from enum import IntEnum

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import SceneMetadata, AirflowGenerator

from ... import _constants

from ...components import BitstreamIdentifiers


class GeneratorType(IntEnum):
    Global = 0
    Point = 1
    Cone = 2
    Plane = 3


write_airflow_generators = tokens.Function(
    "writeAirflowGenerators",
    syntax=[
        tokens.writeCountOrIndex("airflowGeneratorCount"),
        tokens.Repeat(
            "airflowGeneratorCount",
            syntax=[
                tokens.writeID("generatorID"),
                tokens.UnsignedInteger("generatorType", width=3),
                tokens.QuantizedFloat("speed", quant_range=_constants.airflow_speed_range),
                tokens.writeOrientation("orientation"),
                tokens.writeGain("gainDb", isHiPrecGain=False),
                tokens.Flag("active"),
                tokens.Switch(
                    "generatorType",
                    cases=[
                        # global source, no additional data
                        tokens.SwitchCase(0, syntax=[]),
                        # Point source
                        tokens.SwitchCase(
                            1,
                            syntax=[
                                tokens.writePosition("position"),
                                tokens.Flag("hasLimits"),
                                tokens.Conditional(
                                    "hasLimits",
                                    false=[],
                                    true=[
                                        tokens.QuantizedFloat(
                                            "aziLimit",
                                            quant_range=_constants.airflow_aziElevLimit_range,
                                        ),
                                        tokens.QuantizedFloat(
                                            "elevLimit",
                                            quant_range=_constants.airflow_aziElevLimit_range,
                                        ),
                                    ],
                                ),
                            ],
                        ),
                        # cone source
                        tokens.SwitchCase(
                            2,
                            syntax=[
                                tokens.writePosition("position"),
                                tokens.writeDistance("baseRadius"),
                                tokens.writeDistance("length"),
                                tokens.Flag("hasTopRadius"),
                                tokens.Conditional(
                                    "hasTopRadius",
                                    false=[],
                                    true=[tokens.writeDistance("topRadius")],
                                ),
                                tokens.UnsignedInteger("edgeRolloff", width=10),
                            ],
                        ),
                        tokens.SwitchCase(
                            3,
                            syntax=[
                                tokens.writePosition("position"),
                                tokens.writeDistance("width"),
                                tokens.writeDistance("height"),
                                tokens.UnsignedInteger("edgeRolloff", width=10),
                            ],
                        ),
                    ],
                ),
                tokens.Flag("hasDistanceAttenuationFactor"),
                tokens.Conditional(
                    "hasDistanceAttenuationFactor",
                    false=[],
                    true=[
                        tokens.QuantizedFloat(
                            "distanceAttenuationFactor",
                            quant_range=_constants.airflow_distanceAttenuationFactor_range,
                        )
                    ],
                ),
            ],
        ),
    ],
)


def serialize_airflow_generators(
    composer: BitstreamComposer,
    airflowGenerators: List[AirflowGenerator],
    bs_ids: BitstreamIdentifiers,
    metadata: SceneMetadata,
    isSmallScene: bool,
) -> None:
    """Push object source data to a bitstream composer."""
    composer.push(len(airflowGenerators))
    for gen in airflowGenerators:
        composer.push(bs_ids[gen])
        composer.push(GeneratorType[gen.generatorType].value)
        composer.push(gen.speed)
        composer.push(gen.orientation)
        composer.push(gen.gainDb)
        composer.push(gen.active)
        if gen.generatorType == "Point":
            composer.push({"position": gen.position, "isSmallScene": isSmallScene})
            hasLimits = gen.azimuthLimit is not None or gen.elevationLimit is not None
            composer.push(hasLimits)
            if hasLimits:
                composer.push(gen.azimuthLimit if gen.azimuthLimit is not None else 360.0)
                composer.push(gen.elevationLimit if gen.elevationLimit is not None else 360.0)
        elif gen.generatorType == "Cone":
            composer.push({"position": gen.position, "isSmallScene": isSmallScene})
            composer.push({"distance": gen.baseRadius, "isSmallScene": isSmallScene})
            composer.push({"distance": gen.length, "isSmallScene": isSmallScene})
            hasTopRadius = gen.topRadius is not None
            composer.push(hasTopRadius)
            if hasTopRadius:
                composer.push({"distance": gen.topRadius, "isSmallScene": isSmallScene})
            composer.push(gen.edgeRolloff)
        elif gen.generatorType == "Plane":
            composer.push({"position": gen.position, "isSmallScene": isSmallScene})
            composer.push({"distance": gen.width, "isSmallScene": isSmallScene})
            composer.push({"distance": gen.height, "isSmallScene": isSmallScene})
            composer.push(gen.edgeRolloff)
        hasdistanceAttenuationFactor = gen.distanceAttenuationFactor != 1.0
        composer.push(hasdistanceAttenuationFactor)
        if hasdistanceAttenuationFactor:
            composer.push(gen.distanceAttenuationFactor)
