# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import ChannelSource, SceneMetadata

from ... import _constants
from ...components import BitstreamIdentifiers
from ._util import (
    audio_source_syntax,
    location_syntax,
    serialize_audio_source,
    serialize_location,
    serialize_spatial_transform,
    spatial_transform_syntax,
    granular_pos_syntax,
    serialize_granular_position,
)

loudspeaker_syntax: List[tokens.Token] = [
    tokens.writeID("loudspeakerId"),
    tokens.writePosition("loudspeakerPosition"),
    tokens.writeOrientation("loudspeakerOrientation"),
    tokens.writeCountOrIndex("loudspeakerChannel"),
    tokens.Flag("hasDirectivity"),
    tokens.Conditional(
        "hasDirectivity",
        false=[],
        true=[tokens.writeID("directivityId")],
    ),
    tokens.QuantizedFloat("directiveness", quant_range=_constants.directiveness_qrange),
]

write_channel_sources = tokens.Function(
    "writeChannelSources",
    syntax=[
        tokens.writeCountOrIndex("channelSourcesCount"),
        tokens.Repeat(
            "channelSourcesCount",
            syntax=[
                tokens.writeCountOrIndex("loudspeakersCount"),
                tokens.Repeat("loudspeakersCount", syntax=loudspeaker_syntax),
                tokens.writeID("channelSourceId"),
                *location_syntax,
                tokens.UnsignedInteger("inputLayout", width=7),
                *audio_source_syntax,
                tokens.Flag("noReverb"),
                tokens.Flag("noDoppler"),
                tokens.Flag("noDistance"),
                tokens.Flag("noCulling"),
                tokens.Flag("noDirectionalFocus"),
                tokens.Flag("mode"),
                tokens.Flag("play"),
                tokens.UnsignedInteger("vdlMethod", width=3),
                *spatial_transform_syntax,
                tokens.Flag("isStatic"),
                *granular_pos_syntax,
            ],
        ),
    ],
)


def serialize_channel_sources(
    composer: BitstreamComposer,
    channel_sources: List[ChannelSource],
    bs_ids: BitstreamIdentifiers,
    metadata: SceneMetadata,
    isSmallScene: bool,
) -> None:
    """Push channel sources data into bitstreamc composer."""
    composer.push(len(channel_sources))
    for chan_src in channel_sources:
        composer.push(len(chan_src.loudspeakers))
        for loudspeaker in chan_src.loudspeakers:
            composer.push(0)  # loudspeakerId - not currently in use

            composer.push({"position": loudspeaker.position, "isSmallScene": isSmallScene})
            # composer.push(loudspeaker.position[1])
            # composer.push(loudspeaker.position[2])

            composer.push(loudspeaker.orientation)
            # composer.push(loudspeaker.orientation[1])
            # composer.push(loudspeaker.orientation[2])

            composer.push(loudspeaker.channel)

            has_directivity = loudspeaker.directivity is not None
            composer.push(has_directivity)
            if has_directivity:
                composer.push(bs_ids[loudspeaker.directivity])

            composer.push(loudspeaker.directiveness)

        # --= END LOUDSPEAKER SERIALISATION =-- #

        composer.push(bs_ids[chan_src])

        serialize_location(
            composer, chan_src.position, chan_src.orientation, chan_src.cspace, isSmallScene
        )

        composer.push(chan_src.inputLayout)
        serialize_audio_source(
            composer,
            chan_src.active,
            chan_src.gainDb,
            chan_src.hasReverbGain,
            chan_src.reverbGainDb,
            bs_ids[chan_src.signal],
            chan_src.refDistance,
            isSmallScene,
        )

        composer.push(chan_src.noReverb)
        composer.push(chan_src.aparams.noDoppler)
        composer.push(chan_src.aparams.noDistance)
        composer.push(chan_src.aparams.noCulling)
        composer.push(chan_src.aparams.noDirectionalFocus)
        composer.push(bool(chan_src.mode))
        composer.push(chan_src.play)
        composer.push(chan_src.vdlMethod)

        tform = bs_ids.getId(chan_src.transform) if chan_src.transform is not None else None
        anchor = bs_ids.getId(chan_src.anchor) if chan_src.anchor is not None else None
        serialize_spatial_transform(composer, tform, anchor)

        composer.push(metadata.isEntityStatic(chan_src.id))

        has_granularPos = chan_src.granularPos is not None
        composer.push(has_granularPos)
        if has_granularPos:
            serialize_granular_position(composer, chan_src.granularPos)
