# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List

from mpegi.framework import BitstreamComposer, tokens, QuantizationRange
from mpegi.metadata.eifplus import HOAGroup, InformedSource

from ... import _constants
from ...components import BitstreamIdentifiers

distance_threshold_qrange = QuantizationRange(min=0, max=100.0, num_bits=16)
distance_factor_range = QuantizationRange(min=0.25, max=4.0, num_bits=8)

write_hoa_groups = tokens.Function(
    "writeHoaGroups",
    syntax=[
        tokens.writeCountOrIndex("hoaGroupsCount"),
        tokens.Repeat(
            "hoaGroupsCount",
            syntax=[
                tokens.writeID("hoaGroupId"),
                tokens.Flag("hasRegion"),
                tokens.Conditional(
                    "hasRegion",
                    false=[],
                    true=[tokens.writeID("regionId")],
                ),
                tokens.writeCountOrIndex("coSourceCount"),
                tokens.Repeat(
                    "coSourceCount",
                    syntax=[tokens.writeID("hoaSourceId")],
                ),
                tokens.Flag("hasHOAGroupHasComplexityReductionParameters"),
                tokens.Conditional(
                    "hasHOAGroupHasComplexityReductionParameters",
                    false=[],
                    true=[
                        tokens.UnsignedInteger("highestSingleBinBandsIndex", width=8),
                        tokens.UnsignedInteger("lowestHighBandIndex", width=8),
                        tokens.UnsignedInteger("intermediateBandsERBWidth", width=8),
                    ],
                ),
                tokens.writeDistance("exteriorProjectionRadius"),
                tokens.Flag("hasInformedSources"),
                tokens.Conditional(
                    "hasInformedSources",
                    false=[],
                    true=[
                        tokens.writeCountOrIndex("informedSourceCount"),
                        tokens.UnsignedInteger("maxSimultaneousInformedSources", width=8),
                        tokens.Repeat(
                            "informedSourceCount",
                            syntax=[
                                tokens.writeID("informedSourceId"),
                                tokens.writePosition("informedSourcePos"),
                                tokens.Flag("isListenerThersholdPresent"),
                                tokens.Flag("isHoaSourceThersholdPresent"),
                                tokens.Conditional(
                                    "isListenerThersholdPresent",
                                    false=[],
                                    true=[
                                        tokens.QuantizedFloat(
                                            "listenerThreshold",
                                            quant_range=distance_threshold_qrange,
                                        )
                                    ],
                                ),
                                tokens.Conditional(
                                    "isHoaSourceThersholdPresent",
                                    false=[],
                                    true=[
                                        tokens.QuantizedFloat(
                                            "hoaSourceThreshold",
                                            quant_range=distance_threshold_qrange,
                                        )
                                    ],
                                ),
                                tokens.UnsignedInteger("priorityValue", width=6),
                            ],
                        ),
                        tokens.Flag("adaptiveExteriorProjectionRadius"),
                    ],
                ),
                tokens.Flag("hasLowCompProfileParameters"),
                tokens.Conditional(
                    "hasLowCompProfileParameters",
                    false=[],
                    true=[
                        tokens.QuantizedFloat("distanceFactor", quant_range=distance_factor_range),
                        tokens.UnsignedInteger("panningGridOrder", width=8),
                    ],
                ),
            ],
        ),
    ],
)


def serialize_hoa_groups(
    composer: BitstreamComposer,
    hoa_groups: List[HOAGroup],
    bs_ids: BitstreamIdentifiers,
    isSmallScene: bool,
) -> None:
    hasComplexityReductionParameters = True
    highestSingleBinBandsIndex = 15
    lowestHighBandIndex = 89
    intermediateBandsERBWidth = 2
    hasLowCompProfileParameters = True
    distanceFactor = 1.0
    panningGridOrder = 0

    """Push hoa groups into bitstream composer."""
    composer.push(len(hoa_groups))

    for grp in hoa_groups:
        composer.push(bs_ids[grp])

        has_region = grp.region is not None
        composer.push(has_region)
        if has_region:
            composer.push(bs_ids[grp.region])

        composer.push(len(grp.coSources))
        for coSrc in grp.coSources:
            composer.push(bs_ids[coSrc])

        composer.push(hasComplexityReductionParameters)
        if hasComplexityReductionParameters:
            composer.push(highestSingleBinBandsIndex)
            composer.push(lowestHighBandIndex)
            composer.push(intermediateBandsERBWidth)

        composer.push({"distance": grp.exteriorProjectionRadius, "isSmallScene": isSmallScene})

        has_informed_sources = len(grp.informedSources) > 0
        composer.push(has_informed_sources)
        if has_informed_sources:
            composer.push(len(grp.informedSources))
            max_simultaneous_informed_sources = 10
            composer.push(max_simultaneous_informed_sources)
            for src in grp.informedSources:
                serializeInformedSource(composer, bs_ids, src, isSmallScene)
            composer.push(grp.adaptiveExteriorProjectionRadius)

        composer.push(hasLowCompProfileParameters)
        if hasLowCompProfileParameters:
            composer.push(distanceFactor)
            composer.push(panningGridOrder)


def serializeInformedSource(
    composer: BitstreamComposer,
    bs_ids: BitstreamIdentifiers,
    src: InformedSource,
    isSmallScene: bool,
) -> None:
    """Push an InformedSource into bitstream composer."""
    composer.push(bs_ids[src])
    composer.push(
        {
            "position": [src.position[0], src.position[1], src.position[2]],
            "isSmallScene": isSmallScene,
        }
    )
    is_listener_threshold_present = src.listenerThreshold != 0.0
    is_hoa_source_threshold_present = src.hoaSourceThreshold != 0.0
    composer.push(is_listener_threshold_present)
    composer.push(is_hoa_source_threshold_present)
    if is_listener_threshold_present:
        composer.push(src.listenerThreshold)
    if is_hoa_source_threshold_present:
        composer.push(src.hoaSourceThreshold)

    composer.push(src.priorityValue)
