# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import HOASource, SceneMetadata

from ... import _constants
from ...components import BitstreamIdentifiers
from ._util import (
    audio_source_syntax,
    location_syntax,
    serialize_audio_source,
    serialize_location,
    serialize_spatial_transform,
    spatial_transform_syntax,
    granular_pos_syntax,
    serialize_granular_position,
)

write_hoa_sources = tokens.Function(
    "writeHoaSources",
    syntax=[
        tokens.writeCountOrIndex("hoaSourcesCount"),
        tokens.Repeat(
            "hoaSourcesCount",
            syntax=[
                tokens.writeID("hoaSourceId"),
                *location_syntax,
                *audio_source_syntax,
                tokens.Flag("hasExtent"),
                tokens.Conditional(
                    "hasExtent",
                    false=[],
                    true=[
                        tokens.writeID("extentId"),
                    ],
                ),
                tokens.Flag("extentTransform"),
                tokens.Flag("representation"),
                tokens.Flag("hasGroup"),
                tokens.Conditional(
                    "hasGroup",
                    false=[],
                    true=[tokens.writeID("hoaGroupId")],
                ),
                tokens.Flag("noReverb"),
                tokens.Flag("noDoppler"),
                tokens.Flag("noDistance"),
                tokens.Flag("noCulling"),
                tokens.Flag("noDirectionalFocus"),
                tokens.Flag("mode"),
                tokens.Flag("play"),
                tokens.UnsignedInteger("vdlMethod", width=3),
                *spatial_transform_syntax,
                tokens.Flag("isStatic"),
                *granular_pos_syntax,
            ],
        ),
    ],
)


def serialize_hoa_sources(
    composer: BitstreamComposer,
    hoa_sources: List[HOASource],
    bs_ids: BitstreamIdentifiers,
    metadata: SceneMetadata,
    isSmallScene: bool,
) -> None:
    """Push hoa sources into bitstream composer."""
    composer.push(len(hoa_sources))
    for src in hoa_sources:
        composer.push(bs_ids[src])
        serialize_location(composer, src.position, src.orientation, src.cspace, isSmallScene)
        serialize_audio_source(
            composer,
            src.active,
            src.gainDb,
            src.hasReverbGain,
            src.reverbGainDb,
            bs_ids[src.signal],
            src.refDistance,
            isSmallScene,
        )

        has_extent = src.extent is not None
        composer.push(has_extent)
        if has_extent:
            composer.push(bs_ids[src.extent])

        composer.push(src.extentTransform)
        composer.push(bool(src.representation))

        has_group = src.group is not None
        composer.push(has_group)
        if has_group:
            composer.push(bs_ids[src.group])

        composer.push(src.noReverb)
        composer.push(src.aparams.noDoppler)
        composer.push(src.aparams.noDistance)
        composer.push(src.aparams.noCulling)
        composer.push(src.aparams.noDirectionalFocus)

        composer.push(bool(src.mode))
        composer.push(src.play)
        composer.push(src.vdlMethod)

        tform = bs_ids.getId(src.transform) if src.transform is not None else None
        anchor = bs_ids.getId(src.anchor) if src.anchor is not None else None
        serialize_spatial_transform(composer, tform, anchor)

        composer.push(metadata.isEntityStatic(src.id))

        # Note, why is granularPos put into a tuple here? This is not the case for the objectsources.
        has_granularPos = src.granularPos[0] is not None
        composer.push(has_granularPos)
        if has_granularPos:
            serialize_granular_position(composer, src.granularPos[0])
