# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List

from mpegi.coding import GeometryCoder
from mpegi.framework import Bitstream, BitstreamComposer, tokens
from mpegi.metadata.eifplus import Mesh, SceneMetadata

from ... import _constants
from ...components import BitstreamIdentifiers
from ._util import (
    location_syntax,
    serialize_location,
    serialize_spatial_transform,
    spatial_transform_syntax,
)

write_meshes = tokens.Function(
    "writeMeshes",
    syntax=[
        tokens.writeCountOrIndex("meshesCount"),
        tokens.Repeat(
            "meshesCount",
            syntax=[
                tokens.writeID("meshID"),
                tokens.Blob("meshCoded"),
                *location_syntax,
                *spatial_transform_syntax,
                tokens.Flag("isMeshStatic"),
                tokens.Flag("isEarlyReflectionMesh"),
            ],
        ),
    ],
)


def serialize_meshes(
    composer: BitstreamComposer,
    meshes: List[Mesh],
    bs_ids: BitstreamIdentifiers,
    metadata: SceneMetadata,
    isSmallScene: bool,
) -> None:
    """Code mesh and push data into bitstream composer.

    :param payload: The payload to serialize meshes into
    :param meshes: Meshes to seralize
    :param bsIds: BitstreamIds object
    :param metadata: SceneMetadata object
    """
    composer.push(len(meshes))
    for mesh in meshes:
        composer.push(bs_ids[mesh])

        coder = GeometryCoder()
        material_ids = [bs_ids.getId(mat) if mat is not None else -1 for mat in mesh.materials]
        mesh_as_bytes = coder.code_mesh(mesh, material_ids)
        mesh_as_bstream = Bitstream.fromBytes(mesh_as_bytes)
        composer.push(mesh_as_bstream)

        serialize_location(composer, mesh.position, mesh.orientation, mesh.cspace, isSmallScene)

        tform = bs_ids.getId(mesh.transform) if mesh.transform is not None else None
        anchor = bs_ids.getId(mesh.anchor) if mesh.anchor is not None else None
        serialize_spatial_transform(composer, tform, anchor)

        composer.push(metadata.isEntityStatic(mesh.id))

        is_early_reflection_mesh = mesh.id in metadata.earlyStaticMeshIds
        composer.push(is_early_reflection_mesh)
