# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import ObjectSource, SceneMetadata

from ... import _constants
from ...components import BitstreamIdentifiers
from ._util import (
    audio_source_syntax,
    location_syntax,
    serialize_audio_source,
    serialize_location,
    serialize_spatial_transform,
    spatial_transform_syntax,
    granular_pos_syntax,
    serialize_granular_position,
)

write_object_sources = tokens.Function(
    "writeObjectSources",
    syntax=[
        tokens.writeCountOrIndex("objectSourcesCount"),
        tokens.Repeat(
            "objectSourcesCount",
            syntax=[
                tokens.Flag("hasInputLayout"),
                tokens.Conditional(
                    "hasInputLayout",
                    false=[],
                    true=[
                        tokens.Flag("inputLayoutAlignment"),
                        tokens.Flag("inputLayoutTL"),
                        tokens.Flag("inputLayoutT"),
                        tokens.Flag("inputLayoutTR"),
                        tokens.Flag("inputLayoutL"),
                        tokens.Flag("inputLayoutC"),
                        tokens.Flag("inputLayoutR"),
                        tokens.Flag("inputLayoutBL"),
                        tokens.Flag("inputLayoutB"),
                        tokens.Flag("inputLayoutBR"),
                    ],
                ),
                tokens.writeID("objectSourceId"),
                *location_syntax,
                *audio_source_syntax,
                tokens.Flag("hasExtent"),
                tokens.Conditional(
                    "hasExtent",
                    false=[],
                    true=[tokens.writeID("extentId")],
                ),
                tokens.Flag("hasDirectivity"),
                tokens.Conditional(
                    "hasDirectivity",
                    false=[],
                    true=[tokens.writeID("directivityId")],
                ),
                tokens.QuantizedFloat("directiveness", quant_range=_constants.directiveness_qrange),
                tokens.Flag("noReverb"),
                tokens.Flag("noDoppler"),
                tokens.Flag("noDistance"),
                tokens.Flag("noCulling"),
                tokens.Flag("noDirectionalFocus"),
                tokens.Flag("mode"),
                tokens.Flag("play"),
                tokens.UnsignedInteger("vdlMethod", width=3),
                *spatial_transform_syntax,
                tokens.Flag("isStatic"),
                *granular_pos_syntax,
            ],
        ),
    ],
)


def serialize_object_sources(
    composer: BitstreamComposer,
    objectSources: List[ObjectSource],
    bs_ids: BitstreamIdentifiers,
    metadata: SceneMetadata,
    isSmallScene: bool,
) -> None:
    """Push object source data to a bitstream composer."""
    composer.push(len(objectSources))
    for src in objectSources:
        has_input_layout = src.inputLayout is not None
        composer.push(has_input_layout)
        if has_input_layout:
            composer.push(bool(src.inputLayout.alignment))

            composer.push(src.inputLayout.positioning.TL)
            composer.push(src.inputLayout.positioning.T)
            composer.push(src.inputLayout.positioning.TR)
            composer.push(src.inputLayout.positioning.L)
            composer.push(src.inputLayout.positioning.C)
            composer.push(src.inputLayout.positioning.R)
            composer.push(src.inputLayout.positioning.BL)
            composer.push(src.inputLayout.positioning.B)
            composer.push(src.inputLayout.positioning.BR)

        composer.push(bs_ids[src])
        serialize_location(composer, src.position, src.orientation, src.cspace, isSmallScene)
        serialize_audio_source(
            composer,
            src.active,
            src.gainDb,
            src.hasReverbGain,
            src.reverbGainDb,
            bs_ids[src.signal],
            src.refDistance,
            isSmallScene,
        )

        has_extent = src.extent is not None
        composer.push(has_extent)
        if has_extent:
            composer.push(bs_ids[src.extent])

        has_directivity = src.directivity is not None
        composer.push(has_directivity)
        if has_directivity:
            composer.push(bs_ids[src.directivity])

        composer.push(src.directiveness)

        composer.push(src.noReverb)
        composer.push(src.aparams.noDoppler)
        composer.push(src.aparams.noDistance)
        composer.push(src.aparams.noCulling)
        composer.push(src.aparams.noDirectionalFocus)

        composer.push(bool(src.mode))
        composer.push(src.play)
        composer.push(src.vdlMethod)

        tform = bs_ids.getId(src.transform) if src.transform is not None else None
        anchor = bs_ids.getId(src.anchor) if src.anchor is not None else None
        serialize_spatial_transform(composer, tform, anchor)

        composer.push(metadata.isEntityStatic(src.id))

        has_granularPos = src.granularPos is not None
        composer.push(has_granularPos)
        if has_granularPos:
            serialize_granular_position(composer, src.granularPos)
