# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import Primitive, SceneMetadata

from ... import _constants
from ...components import BitstreamIdentifiers
from ._util import (
    location_syntax,
    serialize_location,
    serialize_spatial_transform,
    spatial_transform_syntax,
)

write_primitives = tokens.Function(
    "writePrimtives",
    syntax=[
        tokens.writeCountOrIndex("primitivesCount"),
        tokens.Repeat(
            "primitivesCount",
            syntax=[
                tokens.UnsignedInteger("primitiveType", width=2),
                tokens.writeID("primitiveId"),
                *location_syntax,
                tokens.writeDistance("primitiveSizeX"),
                tokens.writeDistance("primitiveSizeY"),
                tokens.writeDistance("primitiveSizeZ"),
                tokens.Flag("primitiveHasMaterial"),
                tokens.Conditional(
                    "primitiveHasMaterial",
                    false=[],
                    true=[tokens.writeID("materialID")],
                ),
                *spatial_transform_syntax,
                tokens.Flag("isPrimitiveStatic"),
                tokens.Flag("isEarlyReflectionPrimitive"),
            ],
        ),
    ],
)


def serialize_primitives(
    composer: BitstreamComposer,
    primitives: List[Primitive],
    bs_ids: BitstreamIdentifiers,
    metadata: SceneMetadata,
    isSmallScene: bool,
) -> None:
    """Push geometry data into bitstream composer.

    :param payload: The payload to serialize geometries into.
    :param primitives: All primitives to serialize.
    :param bsIds: BitstreamIds object.
    :param metadata: SceneMetadata object.
    """
    composer.push(len(primitives))
    for prim in primitives:
        composer.push(int(prim.geoType))
        composer.push(bs_ids[prim])

        serialize_location(composer, prim.position, prim.orientation, prim.cspace, isSmallScene)

        composer.push({"distance": prim.size[0], "isSmallScene": isSmallScene})
        composer.push({"distance": prim.size[1], "isSmallScene": isSmallScene})
        composer.push({"distance": prim.size[2], "isSmallScene": isSmallScene})

        material = prim.material if hasattr(prim, "material") else None
        has_material = material is not None
        composer.push(has_material)
        if has_material:
            composer.push(bs_ids[material])

        tform = bs_ids.getId(prim.transform) if prim.transform is not None else None
        anchor = bs_ids.getId(prim.anchor) if prim.anchor is not None else None
        serialize_spatial_transform(composer, tform, anchor)

        is_static = metadata.isEntityStatic(prim.id) if "userregion" not in prim.id else True
        composer.push(is_static)

        is_early_reflection_primitive = prim.id in metadata.earlyStaticMeshIds
        composer.push(is_early_reflection_primitive)
