# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List, Dict

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import AudioStream

from ... import _constants
from ...components import BitstreamIdentifiers

file_audio_stream_syntax: List[tokens.Token] = [
    tokens.CString("audioStreamFile"),
    tokens.writeCountOrIndex("aepInputChannelsCount"),
    tokens.Repeat(
        "aepInputChannelsCount",
        syntax=[tokens.writeCountOrIndex("aepInputChannel")],
    ),
    tokens.Flag("audioStreamIsSoundEffect"),
    tokens.Conditional(
        "audioStreamIsSoundEffect",
        false=[],
        true=[
            tokens.writeDuration("soundeffectDuration"),
        ],
    ),
]

write_audio_streams = tokens.Function(
    "writeAudioStreams",
    syntax=[
        tokens.writeCountOrIndex("audioStreamsCount"),
        tokens.Repeat(
            "audioStreamsCount",
            syntax=[
                tokens.writeID("audioStreamId"),
                tokens.Conditional(
                    "enableLocallyCapturedAudio",
                    true=[
                        tokens.Flag("isLocallyCaptured"),
                        tokens.Conditional(
                            "isLocallyCaptured",
                            false=file_audio_stream_syntax,
                            true=[],
                        ),
                    ],
                    false=file_audio_stream_syntax,
                ),
            ],
        ),
    ],
)


def serialize_audio_streams(
    composer: BitstreamComposer,
    streams: List[AudioStream],
    bs_ids: BitstreamIdentifiers,
    enableLocallyCapturedAudioFromConfigPacket: bool,
    eventStreams: Dict[str, bool],
    streamDurations: Dict[str, float],
) -> None:
    def push_stream_info():
        composer.push(stream.file)
        composer.push(len(stream.aepInputChannels))
        for channel in stream.aepInputChannels:
            composer.push(channel)
        composer.push(eventStreams[stream.id])
        if eventStreams[stream.id]:
            composer.push(streamDurations[stream.id])

    """Push streams into bitstream composer."""
    composer.push(len(streams))
    for stream in streams:
        composer.scope.update(
            {"enableLocallyCapturedAudio": enableLocallyCapturedAudioFromConfigPacket}
        )
        composer.push(bs_ids.getId(stream))
        if enableLocallyCapturedAudioFromConfigPacket:
            composer.push(stream.isLocallyCaptured)
            if not stream.isLocallyCaptured:
                push_stream_info()
        else:
            push_stream_info()
