# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.

from typing import List, Optional

from mpegi.framework import BitstreamComposer, tokens
from mpegi.metadata.eifplus import CSpace

from ... import _constants

location_syntax: List[tokens.Token] = [
    tokens.writePosition("positions x,y,z"),
    tokens.writeOrientation("Orientation Yaw,Pitch,Roll", highResolution=True),
    tokens.Flag("cspace"),
]

audio_source_syntax: List[tokens.Token] = [
    tokens.Flag("active"),
    tokens.writeGain("gainDb", isHiPrecGain=True),
    tokens.Flag("hasReverbGain"),
    tokens.Conditional(
        "hasReverbGain",
        false=[],
        true=[
            tokens.writeGain("reverbGainDb", isHiPrecGain=True),
        ],
    ),
    tokens.writeDistance("refDistance"),
    tokens.writeID("signalId"),
]

spatial_transform_syntax: List[tokens.Token] = [
    tokens.Flag("hasSpatialTransform"),
    tokens.Conditional(
        "hasSpatialTransform",
        false=[],
        true=[
            tokens.Flag("hasAnchor"),
            tokens.Conditional(
                "hasAnchor",
                false=[tokens.writeID("parentTransformId")],
                true=[tokens.writeID("parentAnchorId")],
            ),
        ],
    ),
]

granular_pos_syntax: List[tokens.Token] = [
    tokens.Flag("hasGranularPos"),
    tokens.Conditional(
        "hasGranularPos",
        false=[],
        true=[
            tokens.Repeat(
                3,
                syntax=[
                    tokens.QuantizedFloat(
                        "granularPosValue", quant_range=_constants.granularDB_descriptor_qrange
                    ),
                ],
            )
        ],
    ),
]


def serialize_location(
    composer: BitstreamComposer,
    position: List[float],
    orientation: List[float],
    cspace: CSpace,
    isSmallScene: bool,
) -> None:
    """Push position, orientation, and cspace to composer."""
    composer.push({"position": position, "isSmallScene": isSmallScene})
    composer.push(orientation)
    composer.push(bool(cspace))


def serialize_audio_source(
    composer: BitstreamComposer,
    active: bool,
    gainDb: float,
    hasReverbGain: bool,
    reverbGainDb: float,
    signalId: int,
    refDistance: float,
    isSmallScene: bool,
):
    """Add active, signal, gainDb, hasReverbGain, reverbGainDb, and reference distance to composer."""
    composer.push(active)
    composer.push(gainDb)
    composer.push(hasReverbGain)
    if hasReverbGain:
        composer.push(reverbGainDb)
    composer.push({"distance": refDistance, "isSmallScene": isSmallScene})
    composer.push(signalId)


def serialize_spatial_transform(
    composer: BitstreamComposer,
    transform: Optional[int],
    anchor: Optional[int],
) -> None:
    """Push either an anchor or a transform into bitstream composer."""
    assert not (transform is not None and anchor is not None)

    has_either_transform_or_anchor = transform is not None or anchor is not None
    composer.push(has_either_transform_or_anchor)
    if has_either_transform_or_anchor:
        has_anchor = anchor is not None
        composer.push(has_anchor)
        if has_anchor:
            composer.push(anchor)
        else:
            composer.push(transform)


def serialize_granular_position(
    composer: BitstreamComposer,
    granular_position: List[float],
) -> None:
    """Push granular position to composer."""
    has_granular_position = granular_position is not None
    if has_granular_position:
        composer.push(granular_position[0])
        composer.push(granular_position[1])
        composer.push(granular_position[2])
