# This software module was originally developed by
#
# Dolby International AB
#
# in the course of development of ISO/IEC 23090-4 for reference purposes and
# its performance may not have been optimized.  This software module is an
# implementation of one or more tools as specified by the ISO/IEC 23090-4
# standard.  ISO/IEC gives you a royalty-free, worldwide, non-exclusive,
# copyright license to copy, distribute, and make derivative works of this
# software module or modifications thereof for use in implementations or
# products claiming conformance to ISO/IEC standards incorporating ISO/IEC
# 23090-4 reference software and which satisfy any specified conformance
# criteria of that standard.
#
# Those intending to use this software module in products are advised that its
# use may infringe existing patents and that the license in the previous
# paragraph grants no licenses under such patents.  ISO/IEC and the developers
# of this software module have no liability for use of this software module or
# modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Dolby International AB retains full right to
# modify and use the code for its own purpose, assign or donate the code to
# a third party without consent of and without accounting to third parties,
# and to inhibit third parties from using the code for products that do not
# conform to MPEG-related ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative works.
#
# Copyright (c) ISO/IEC 2022.
from typing import List, Dict, Any, Optional
from mpegi.framework import tokens, Bitstream, BitstreamComposer, ScenePayloadPacketType

import logging
import math
import sys
import pathlib
import pytest
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element

from mpegi.framework.payload import Payload

from ._vox_diffraction import vox_diffraction

"""
Voxel scene info
    p0/p1 : bounding box of the voxel scene
    dim   : dimension of the voxel scene
"""
voxel_scene_info: List[tokens.Token] = [
    tokens.writePosition("p0"),
    tokens.writePosition("p1"),
    tokens.UnsignedInteger("dim_x", width=16),
    tokens.UnsignedInteger("dim_y", width=16),
    tokens.UnsignedInteger("dim_z", width=16),
]

"""
Voxel block
    material_id : Material ID associated with this block
    beg         : block indices 'from'
    end         : block indices 'to' (inclusive)
"""
voxel_block: List[tokens.Token] = [
    tokens.EscapedUnsignedInteger("material_id", widths=[4, 8, 32]),
    tokens.PackedVoxelCoord("beg_xyz"),
    tokens.PackedVoxelCoord("end_xyz"),
]

"""
Scene material
    material_bsid : Material bitstream ID for material mapping
"""
voxel_scene_material: List[tokens.Token] = [
    tokens.EscapedUnsignedInteger("material_bsid", widths=[8, 16, 32])
]

"""
Source-material dependency (occlusion)
"""
source_material_dependency_occlusion: List[tokens.Token] = [
    tokens.Flag("has_source_material_dependency"),
    tokens.Conditional(
        "has_source_material_dependency",
        false=[],
        true=[
            tokens.EscapedUnsignedInteger(
                "num_source_material_dependency_entries", widths=[8, 16, 32]
            ),
            tokens.Repeat(
                "num_source_material_dependency_entries",
                syntax=[
                    tokens.EscapedUnsignedInteger(
                        "source_material_dependency_object_id", widths=[8, 16, 32]
                    ),
                    tokens.EscapedUnsignedInteger(
                        "source_material_dependency_material_id", widths=[8, 16, 32]
                    ),
                ],
            ),
        ],
    ),
]


"""
Renderer instructions (occlusion)
"""
voxel_renderer_instructions_occlusion: List[tokens.Token] = [
    tokens.Flag("has_instructions_occlusion"),
    tokens.Conditional(
        "has_instructions_occlusion",
        false=[],
        true=[
            tokens.Repeat(
                "num_materials",
                syntax=[
                    tokens.Flag("has_recovery"),
                    tokens.Conditional(
                        "has_recovery", false=[], true=[tokens.Float32("recovery_setting")]
                    ),
                ],
            )
        ],
    ),
    *source_material_dependency_occlusion,
]

"""
Pre computed path data
"""
voxel_pre_computed_path_data: List[tokens.Token] = [
    tokens.EscapedUnsignedInteger("num_of_vox_diffraction_path_data", widths=[8, 16, 32]),
    tokens.Repeat(
        "num_of_vox_diffraction_path_data",
        syntax=[
            tokens.PackedVoxelCoord("vox_diffraction_path_start_xy"),
            tokens.PackedVoxelCoord("vox_diffraction_path_end_xy"),
            tokens.Flag("vox_diffraction_path_data_exist"),
            tokens.Conditional(
                "vox_diffraction_path_data_exist",
                false=[],
                true=[
                    tokens.PackedVoxelCoord("vox_diffraction_source_direction_xy"),
                    tokens.Float32("vox_diffraction_path_length"),
                ],
            ),
        ],
    ),
]

"""
Diffraction map
"""
voxel_diffraction_map: List[tokens.Token] = [
    tokens.EscapedUnsignedInteger("num_blocks_dfr_map", widths=[8, 16, 32]),
    tokens.Repeat(
        "num_blocks_dfr_map",
        syntax=[
            tokens.Flag("diffraction_enabled"),
            tokens.PackedVoxelCoord("dfr_map_beg_xy"),
            tokens.PackedVoxelCoord("dfr_map_end_xy"),
        ],
    ),
]


"""
Renderer instructions (diffraction)
"""
voxel_renderer_instructions_diffraction: List[tokens.Token] = [
    tokens.Flag("has_instructions_diffraction"),
    tokens.Conditional(
        "has_instructions_diffraction",
        false=[],
        true=[
            tokens.Repeat(
                "num_materials",
                syntax=[
                    tokens.Flag("has_diffraction"),
                ],
            )
        ],
    ),
    tokens.Flag("has_diffraction_effect_control"),
    tokens.Conditional(
        "has_diffraction_effect_control",
        true=[tokens.Float32("diffraction_effect_control_value")],
        false=[],
    ),
    tokens.Flag("has_diffraction_map"),
    tokens.Conditional(
        "has_diffraction_map",
        true=voxel_diffraction_map,
        false=[
            tokens.Flag("has_diffraction_map_height"),
            tokens.Conditional(
                "has_diffraction_map_height",
                true=[tokens.UnsignedInteger("diffraction_map_height", width=16)],
                false=[],
            ),
        ],
    ),
]

"""
Updates
"""
voxel_scene_update: List[tokens.Token] = [
    tokens.EscapedUnsignedInteger("update_id", widths=[8, 16, 32]),
    tokens.Flag("has_reference_point"),
    tokens.Conditional(
        "has_reference_point",
        false=[],
        true=[tokens.writePosition("p0_upd")],
    ),
    tokens.EscapedUnsignedInteger("num_sequences", widths=[8, 16, 32]),
    tokens.Repeat(
        "num_sequences",
        syntax=[
            tokens.UnsignedInteger("duration", width=16),
            tokens.EscapedUnsignedInteger("num_blocks_upd", widths=[8, 16, 32]),
            tokens.Repeat("num_blocks_upd", syntax=voxel_block),
            *voxel_renderer_instructions_occlusion,
            *voxel_renderer_instructions_diffraction,
        ],
    ),
]

"""
Voxel subscene
"""
voxel_subscene: List[tokens.Token] = [
    *voxel_scene_info,
    tokens.EscapedUnsignedInteger("num_blocks", widths=[8, 16, 32]),
    tokens.Repeat("num_blocks", syntax=voxel_block),
    tokens.EscapedUnsignedInteger("num_materials", widths=[8, 16, 32]),
    tokens.Repeat("num_materials", syntax=voxel_scene_material),
    *voxel_renderer_instructions_occlusion,
    *voxel_renderer_instructions_diffraction,
    tokens.EscapedUnsignedInteger("num_scene_updates", widths=[8, 16, 32]),
    tokens.Repeat("num_scene_updates", syntax=voxel_scene_update),
]

"""
Tuning parameters
"""
voxel_tuning_parameters: List[tokens.Token] = [
    tokens.Flag("has_tuning_num_origins"),
    tokens.Conditional(
        "has_tuning_num_origins", true=[tokens.UnsignedInteger("num_origins", width=5)], false=[]
    ),
    tokens.Flag("has_tuning_num_clusters"),
    tokens.Conditional(
        "has_tuning_num_clusters", true=[tokens.UnsignedInteger("num_clusters", width=5)], false=[]
    ),
    tokens.Flag("has_indoor_outdoor_tuning"),
    tokens.Conditional(
        "has_indoor_outdoor_tuning",
        true=[
            tokens.writeGain("tuning_gain_indoor", isHiPrecGain=True),
            tokens.writeGain("tuning_gain_outdoor", isHiPrecGain=True),
        ],
        false=[],
    ),
    tokens.Flag("combine_reflection_per_cluster"),
]


voxdata_payload: List[tokens.Token] = [
    tokens.EscapedUnsignedInteger("num_subscenes", widths=[8, 16, 32]),
    tokens.Assign("num_subscenes", tokens.Add("num_subscenes", 1)),
    tokens.Repeat("num_subscenes", syntax=voxel_subscene),
    *voxel_tuning_parameters,
]


# ---------------------------------------------------------------------------------------------------
class NoVoxelPayloadException(Exception):
    """
    Raised if no voxel-payload is present in the EIF
    """

    pass


# ---------------------------------------------------------------------------------------------------
class VoxdataPayload(Payload):
    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _push_instructions_diffraction(
        composer: BitstreamComposer, instructions: Dict[str, bool], materials: List[str]
    ) -> None:
        for mat in materials[1:]:
            diffraction_enable: bool = instructions[mat] if mat in instructions else True
            composer.push(diffraction_enable)

    @staticmethod
    def _push_instructions_occlusion(
        composer: BitstreamComposer, instructions: Dict[str, float], materials: List[str]
    ) -> None:
        for mat in materials[1:]:
            has_instruction_for_mat = mat in instructions
            composer.push(has_instruction_for_mat)
            if has_instruction_for_mat:
                composer.push(instructions[mat])

    @staticmethod
    def _push_diffraction_effect_control(
        composer: BitstreamComposer, diffraction_effect_control: float
    ) -> None:
        composer.push(diffraction_effect_control)

    @staticmethod
    def _get_delta_code(delta: float) -> Optional[int]:
        if delta == pytest.approx(1, rel=0.01):
            return 0
        elif delta == pytest.approx(math.sqrt(2) - 1, rel=0.01):
            return 1
        elif delta == pytest.approx(1 - math.sqrt(2), rel=0.01):
            return 2
        elif delta == pytest.approx(-1, rel=0.01):
            return 3
        else:
            logging.getLogger("VoxdataPayload").warning("Unexpected delta")
            return None

    @staticmethod
    def _push_diffraction_map(
        composer: BitstreamComposer, diffraction_map: Dict[str, Any], diffraction_map_dim: List[int]
    ):
        composer.push(len(diffraction_map))

        for el in diffraction_map:
            composer.push(bool(int(el["enabled"])))
            composer.push({"value": el["p0"], "dimensions": diffraction_map_dim})
            composer.push({"value": el["p1"], "dimensions": diffraction_map_dim})

    @staticmethod
    def _push_pre_computed_path_data(
        composer: BitstreamComposer,
        vox_config: Dict[str, Any],
        diffraction_map: Dict[str, Any],
        diffraction_map_height: Optional[int],
        pre_computed_path_data: Optional[Dict[str, Any]],
        voxel_dimensions: List[int],
        objects_vox: List[List[int]],
        blocks_vox: Dict[str, Any],
    ) -> None:
        logger: logging.Logger = logging.getLogger("VoxdataPayload")

        if vox_config["mode"] == "low_bitrate":
            if pre_computed_path_data:
                logger.warning(
                    "Low-bitrate mode requested, pre-calculated path data will be discarded"
                )

                pre_computed_path_data.clear()
        else:
            if len(pre_computed_path_data) > 0 and len(diffraction_map) == 0:
                logger.warning(
                    f"Voxconf: 'pre_computed_path_data' specified without 'diffraction_map'. Discarding 'pre_computed_path_data' specified in voxconf"
                )
                pre_computed_path_data.clear()

            if not vox_diffraction and len(pre_computed_path_data) == 0:
                logger.warning(
                    "Vox diffraction DLL not found! Functionality to generate pre computed path data not available"
                )

            if vox_diffraction and len(pre_computed_path_data) == 0:
                logger.warning(
                    "Generating pre computed path data. Depending on the complexity of the scene, this can take a while"
                )

                # Setup projection map
                pm = vox_diffraction.projection_map(voxel_dimensions[0], voxel_dimensions[1])

                if diffraction_map == []:
                    logger.warning(
                        "No diffraction map provided. Generating diffraction map from voxel blocks"
                    )

                    # build projection map from voxel blocks
                    z = (
                        diffraction_map_height
                        if diffraction_map_height is not None
                        else int(voxel_dimensions[2] / 2)
                    )

                    for block in blocks_vox:
                        start: List[int] = block["grid_coordinate_s"]
                        end: List[int] = block["grid_coordinate_e"]
                        value = block["diffraction_enabled"]

                        for x in range(start[0] - 1, end[0]):
                            for y in range(start[1] - 1, end[1]):
                                if start[2] - 1 <= z < end[2]:
                                    pm.set(x, y, chr(value))
                else:
                    for el in diffraction_map:
                        value = bool(int(el["enabled"]))
                        x0 = int(el["p0"][0])
                        y0 = int(el["p0"][1])
                        x1 = int(el["p1"][0])
                        y1 = int(el["p1"][1])

                        # Indices from the bitstream are in the range [1..dim]
                        for x in range(x0, x1 + 1, 1):
                            for y in range(y0, y1 + 1, 1):
                                pm.set(x - 1, y - 1, chr(value))
                # Generate pre computed pathdata for all user positions
                pf = vox_diffraction.pathfinder()
                pf.init(pm, True)
                cf = vox_diffraction.cornerfinder(pm)

                for object_vox in objects_vox:
                    object_x = object_vox[0]
                    object_y = object_vox[1]
                    for user_x in range(1, voxel_dimensions[0] + 1, 1):
                        for user_y in range(1, voxel_dimensions[1] + 1, 1):
                            value = pm.get(user_x - 1, user_y - 1)
                            if value != "\x01":
                                user_vox = [user_x, user_y, 1]
                                if not pf.test_direct_view(user_vox, object_vox) == 0:
                                    if pf.find_path(user_vox, object_vox) == 0:
                                        path_length = pf.get_path_length()
                                        if (
                                            cf.find_corner(pf.get_path().length(), pf.get_path())
                                            == 0
                                        ):
                                            corner = cf.corner
                                            pre_computed_path_data.append(
                                                {
                                                    "p0": [user_x, user_y],
                                                    "p1": [object_x, object_y],
                                                    "corner": [corner[0] + 1, corner[1] + 1],
                                                    "length": path_length,
                                                }
                                            )
                                    else:
                                        pre_computed_path_data.append(
                                            {"p0": [user_x, user_y], "p1": [object_x, object_y]}
                                        )

        bitstream: Bitstream = composer.bitstream
        pre_computed_path_data_present: bool = len(pre_computed_path_data) > 0
        bitstream.addBool(pre_computed_path_data_present)

        if pre_computed_path_data_present:
            logger.info(f"Encoding pre-computed path data, mode={vox_config['mode']}")
            bit_cnt_init: int = bitstream.getLength()

            # Number of start-positions
            start_positions: List[List[int]] = []
            for obj in objects_vox:
                if [obj[0], obj[1]] not in start_positions:
                    start_positions.append([obj[0], obj[1]])

            bitstream.addEscapedInt(len(start_positions), [8, 16, 32])

            pcpd_full: bool = (
                vox_config["mode"] == "jps_full" or vox_config["mode"] == "multi_path_or_jps"
            )
            bitstream.addBool(pcpd_full)

            voxel_dimensions2D = voxel_dimensions.copy()
            voxel_dimensions2D.pop()
            if pcpd_full:
                for pos in start_positions:

                    # Get all entries for current object
                    data: List[Dict[str, Any]] = [
                        el for el in pre_computed_path_data if el["p1"] == pos
                    ]

                    data_present_for_object: bool = len(data) > 0
                    bitstream.addBool(data_present_for_object)

                    if not data_present_for_object:
                        continue

                    # Voxel coord of object
                    pv = tokens.PackedVoxelCoord("", voxel_dimensions2D)
                    bs = pv.to_bitstream(pos)
                    bitstream.appendBitstream(bs)

                    for x in range(1, voxel_dimensions2D[0] + 1):
                        # Reset previous when advancing to the next column
                        prev: Dict[str, Any] = {"corner": [], "length": -1}

                        for y in range(1, voxel_dimensions2D[1] + 1):
                            path = next((p for p in data if p["p0"] == [x, y]), None)

                            if path and "corner" in path:
                                bitstream.addBool(True)

                                use_prev: bool = False
                                delta_code: Optional[int] = None

                                if prev["corner"] == path["corner"]:
                                    delta_code: Optional[int] = VoxdataPayload._get_delta_code(
                                        path["length"] - prev["length"]
                                    )

                                    use_prev = not delta_code is None

                                bitstream.addBool(use_prev)
                                if not use_prev:
                                    bitstream.appendBitstream(pv.to_bitstream(path["corner"]))
                                    bitstream.addFloat32(path["length"])
                                    prev["corner"] = path["corner"]
                                else:
                                    bitstream.addInt(delta_code, 2)

                                prev["length"] = path["length"]

                            else:
                                bitstream.addBool(False)  # No path
                                prev = {"corner": [], "length": -1}
            else:
                for pos in start_positions:

                    # Get all entries for current object
                    data: List[Dict[str, Any]] = [
                        el for el in pre_computed_path_data if el["p1"] == pos
                    ]

                    # Number of end-coordinates
                    bitstream.addEscapedInt(len(data), [8, 16, 32])

                    # Start-coordinate
                    pv = tokens.PackedVoxelCoord("", voxel_dimensions2D)
                    bitstream.appendBitstream(pv.to_bitstream(pos))

                    prev_corner: List[int] = []
                    for d in data:
                        bitstream.appendBitstream(pv.to_bitstream(d["p0"]))

                        exists: bool = "length" in d
                        bitstream.addBool(exists)
                        if exists:
                            use_prev: bool = d["corner"] == prev_corner
                            bitstream.addBool(use_prev)

                            if not use_prev:
                                bitstream.appendBitstream(pv.to_bitstream(d["corner"]))

                            bitstream.addFloat32(d["length"])
                            prev_corner = d["corner"]

            logger.info(
                f"Pre-computed path data: encoded {bitstream.getLength() - bit_cnt_init} bits"
            )

    # ----------------------------------------------------------------------------------------------

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        super().__init__(args, artifacts)

        self._bitstream: Bitstream = Bitstream()
        self._isSmallScene = artifacts["scene_info"]["isSmallScene"]
        self._bs_ids = artifacts["bs_ids"]
        self._object_sources = artifacts["object_sources"]
        self._voxel_config = artifacts["voxel_config"]
        self._voxel_blocks = []

        if not "subscenes" in self._voxel_config:
            self._voxel_config["subscenes"] = []

        if artifacts["voxel_mode"]:
            self._voxel_config["mode"] = artifacts["voxel_mode"]
        elif "mode" not in self._voxel_config:
            # Default mode
            self._voxel_config["mode"] = "low_bitrate"

            # Tuning parameters from the EIF or default values
        self._set_tuning_parameters(
            artifacts["acoustic_tuning_params"][0].voxOccCtrl,
            artifacts["acoustic_tuning_params"][0].voxDiffrCtrl,
        )

        # RasterMap data
        self._rastermap = artifacts["rastermap"]

        try:
            root = ET.parse(args["scene"]).getroot()

            voxel_subscenes = root.findall("VoxelScene")
            if not voxel_subscenes:
                raise NoVoxelPayloadException

            composer = BitstreamComposer(voxdata_payload)

            composer.push(len(voxel_subscenes) - 1)

            for subscene in voxel_subscenes:
                # collect objects voxel coords for this subscene
                objects_vox: List[List[int]] = []
                for object in self._object_sources:
                    voxel_coord = self._position_to_voxel_coord(subscene, object.position)
                    if len(voxel_coord):
                        objects_vox.append(voxel_coord)

                resolution: List[int] = self._push_voxel_scene_info(composer, subscene)

                blocks = subscene.findall("VoxelBlock")
                materials = self._get_subscene_materials(blocks)
                self._push_voxel_blocks(composer, subscene, blocks, materials)

                self._push_material_mapping(composer, materials)

                # Subscene config
                subscene_config: Optional[Dict[str, Any]] = next(
                    filter(
                        lambda ss: ss["id"] == subscene.attrib["id"],
                        self._voxel_config["subscenes"],
                    ),
                    {},
                )

                # Diffraction map height for voxel-diffraction
                if "diffraction_map_height" not in subscene_config:
                    height: Optional[int] = int(subscene.attrib.get("diffraction_map_height", 0))
                    if height:
                        subscene_config["diffraction_map_height"] = height

                # Scene instructions
                if "instructions_diffraction" not in subscene_config:
                    subscene_config["instructions_diffraction"] = {}

                instructions: List[Element] = subscene.findall("./Instructions/VoxelDiffraction")
                for instr in instructions:
                    mat: str = instr.attrib.get("material", "")
                    enabled: bool = "true" == instr.attrib.get("enabled", False)
                    subscene_config["instructions_diffraction"][mat] = enabled

                self._push_renderer_instructions(
                    composer,
                    resolution,
                    subscene_config,
                    materials,
                    objects_vox,
                    self._voxel_blocks,
                    self._occlusion_effect_control,
                    self._diffraction_effect_control,
                )

                updates = [upd for upd in root.findall("Update") if "index" in upd.attrib]
                self._push_voxel_scene_updates(
                    composer,
                    subscene,
                    subscene_config,
                    updates,
                    materials,
                    objects_vox,
                    self._voxel_blocks,
                )

            # Global control parameters
            vox_early_ctrl = artifacts["acoustic_tuning_params"][0].voxEarlyCtrl
            if vox_early_ctrl:
                has_num_origins: bool = vox_early_ctrl.num_origins is not None
                composer.push(has_num_origins)

                if has_num_origins:
                    composer.push(vox_early_ctrl.num_origins - 2)

                has_num_clusters: bool = vox_early_ctrl.num_clusters is not None
                composer.push(has_num_clusters)

                if has_num_clusters:
                    composer.push(vox_early_ctrl.num_clusters - 2)

                has_indoor_outdoor_tuning: bool = (
                    vox_early_ctrl.indoor_gain_db is not None
                    and vox_early_ctrl.outdoor_gain_db is not None
                )
                composer.push(has_indoor_outdoor_tuning)

                if has_indoor_outdoor_tuning:
                    composer.push(vox_early_ctrl.indoor_gain_db)
                    composer.push(vox_early_ctrl.outdoor_gain_db)

                composer.push(vox_early_ctrl._combine_reflections_per_cluster)
            else:
                composer.push(False)
                composer.push(False)
                composer.push(False)
                composer.push(False)

            self._bitstream = composer.get_bitstream()

        except NoVoxelPayloadException:
            pass
        except Exception as exc:
            logging.getLogger("VoxdataPayload").exception("Error writing VoxdataPayload")

    def _set_tuning_parameters(
        self,
        occlusion_effect_control: Optional[Dict[str, float]],
        diffraction_effect_control: Optional[float],
    ) -> None:
        """
        Set voxel-occlusion and diffraction effect control for all sub-scenes
        """
        self._occlusion_effect_control = occlusion_effect_control
        self._diffraction_effect_control = (
            diffraction_effect_control if diffraction_effect_control != 1.0 else None
        )

    def _push_voxel_scene_info(self, composer: BitstreamComposer, subscene: Element) -> List[int]:
        """
        Basic voxel scene information, i.e., scene bounding box, voxel dimensions, returns scene resolution
        """
        pos_s: List[float] = [float(val) for val in subscene.attrib["position_s"].split()]
        composer.push({"position": pos_s, "isSmallScene": self._isSmallScene})

        pos_e: List[float] = [float(val) for val in subscene.attrib["position_e"].split()]
        composer.push({"position": pos_e, "isSmallScene": self._isSmallScene})

        res: List[int] = [int(val) for val in subscene.attrib["vox_resolution"].split()]
        for xyz in res:
            composer.push(xyz)

        return res

    def _push_voxel_blocks(
        self, composer: BitstreamComposer, subscene, blocks: List[Element], materials: List[str]
    ) -> None:
        """
        Voxel blocks composing the scene geometry. A block defines a region in the voxel-scene
        that has the same acoustic properties (e.g. material, sound-medium)
        """

        composer.push(len(blocks))

        voxel_dimensions: List[int] = [
            int(val) for val in subscene.attrib["vox_resolution"].split()
        ]

        self._voxel_blocks = []

        for block in blocks:
            mat: str = block.attrib["material"]
            id = materials.index(mat)
            composer.push(id)

            voxel_coord_s: List[int] = [
                int(val) for val in block.attrib["grid_coordinate_s"].split()
            ]
            composer.push({"value": voxel_coord_s, "dimensions": voxel_dimensions})

            voxel_coord_e: List[int] = [
                int(val) for val in block.attrib["grid_coordinate_e"].split()
            ]
            composer.push({"value": voxel_coord_e, "dimensions": voxel_dimensions})

            # Store voxel-blocks in case they are needed to build the diffraction map later on
            self._voxel_blocks.append(
                {
                    "material": mat,
                    "grid_coordinate_s": voxel_coord_s,
                    "grid_coordinate_e": voxel_coord_e,
                    "diffraction_enabled": bool(mat),
                }
            )

    def _get_subscene_materials(self, blocks: List[Element]) -> List[str]:
        """
        Get all materials in the subscene and store for material mapping
        """
        materials: List[str] = [""]
        for block in blocks:
            if block.attrib["material"] not in materials:
                materials.append(block.attrib["material"])
        return sorted(materials)

    def _position_to_voxel_coord(self, subscene: Element, position: List[float]) -> List[int]:
        pos_s: List[float] = [float(val) for val in subscene.attrib["position_s"].split()]
        pos_e: List[float] = [float(val) for val in subscene.attrib["position_e"].split()]
        voxel_res: List[int] = [int(val) for val in subscene.attrib["vox_resolution"].split()]
        voxel_size: List[float] = [
            (pos_e[0] - pos_s[0]) / voxel_res[0],
            (pos_e[1] - pos_s[1]) / voxel_res[1],
            (pos_e[2] - pos_s[2]) / voxel_res[2],
        ]

        vox_position: List[float] = [
            position[2],
            position[0],
            position[1],
        ]

        voxel_coord: List[int] = []
        if (
            vox_position[0] >= pos_s[0]
            and vox_position[1] >= pos_s[1]
            and vox_position[2] >= pos_s[2]
            and vox_position[0] <= pos_e[0]
            and vox_position[1] <= pos_e[1]
            and vox_position[2] <= pos_e[2]
        ):
            for index, xyz in enumerate(vox_position):
                voxel_coord.append(
                    int(((vox_position[index] - pos_s[index]) / voxel_size[index] + 0.5) + 0.5)
                )
        return voxel_coord

    def _push_material_mapping(self, composer: BitstreamComposer, materials: List[str]) -> None:
        """
        Write the material mapping, using the bitstream ids
        """
        composer.push(len(materials) - 1)
        for mat in materials[1:]:
            id = self._bs_ids.getId(mat)
            composer.push(id)

    def _push_renderer_instructions(
        self,
        composer: BitstreamComposer,
        voxel_dimensions: List[int],
        config: Optional[Dict[str, Any]],
        materials: List[str],
        objects_vox: List[List[int]],
        blocks_vox: Dict[str, Any],
        effect_control_occlusion: Optional[Dict[str, float]] = None,
        effect_control_diffraction: Optional[float] = None,
    ) -> None:
        """
        Renderer instructions for occlusion and diffraction
        """

        # Instructions for occlusion (config takes precedence over EIF)
        instructions_occlusion: List[str, float] = config.get(
            "instructions_occlusion", effect_control_occlusion
        )

        has_instructions_occlusion: bool = bool(instructions_occlusion)
        composer.push(has_instructions_occlusion)

        if has_instructions_occlusion:
            self._push_instructions_occlusion(composer, instructions_occlusion, materials)

        # No source-material dependency for now
        composer.push(False)

        # Instructions for diffraction
        instructions_diffraction: List[str, bool] = config.get("instructions_diffraction", None)

        has_instructions_diffraction: bool = bool(instructions_diffraction)
        composer.push(has_instructions_diffraction)

        if has_instructions_diffraction:
            self._push_instructions_diffraction(composer, instructions_diffraction, materials)
            for b in self._voxel_blocks:
                mat: str = b["material"]
                if mat:
                    b["diffraction_enabled"]: bool = (
                        instructions_diffraction[mat] if mat in instructions_diffraction else True
                    )

        # Diffraction effect control (config takes precedence over EIF)
        diffraction_effect_control = config.get(
            "diffraction_effect_control", effect_control_diffraction
        )

        has_diffraction_effect_control: bool = not diffraction_effect_control is None
        composer.push(has_diffraction_effect_control)

        if has_diffraction_effect_control:
            self._push_diffraction_effect_control(composer, diffraction_effect_control)

        # Diffraction data (diffraction map and pre-computed path data)
        diffraction_map: Dict[str, Any] = config.get("diffraction_map", [])

        has_diffraction_map: bool = bool(diffraction_map)

        # Bitstream may not contain pre-computed path data AND RasterMap data
        has_diffraction_map &= self._rastermap is None

        composer.push(has_diffraction_map)

        if has_diffraction_map:
            self._push_diffraction_map(composer, diffraction_map, voxel_dimensions[0:2])

        else:
            has_diffraction_map_height: bool = "diffraction_map_height" in config
            composer.push(has_diffraction_map_height)

            if has_diffraction_map_height:
                composer.push(config["diffraction_map_height"])

        pre_computed_path_data: List[Dict[str, Any]] = config.get("pre_computed_path_data", [])

        self._push_pre_computed_path_data(
            composer,
            self._voxel_config,
            diffraction_map,
            config.get("diffraction_map_height", None),
            pre_computed_path_data,
            voxel_dimensions,
            objects_vox,
            blocks_vox,
        )

    def _push_voxel_scene_updates(
        self,
        composer: BitstreamComposer,
        subscene: Element,
        subscene_config: Optional[Dict[str, Any]],
        updates: List[Element],
        materials: List[str],
        objects_vox: List[List[int]],
        blocks_vox: Dict[str, Any],
    ) -> None:
        """
        Voxel scene updates. At the moment only triggered updates (i.e., L2/3) are supported
        """
        subscene_id = subscene.attrib["id"]

        # Only consider updates for the current subscene
        updates: List[Element] = list(
            filter(
                lambda upd: len(
                    [mod for mod in upd.findall("Modify") if mod.attrib["id"] == subscene_id]
                )
                > 0,
                updates,
            )
        )

        voxel_dimensions: List[int] = [
            int(val) for val in subscene.attrib["vox_resolution"].split()
        ]

        update_config: List[Dict[str, Any]] = (
            subscene_config["updates"]
            if not subscene_config is None and "updates" in subscene_config
            else []
        )

        composer.push(len(updates))
        for upd in updates:

            composer.push(int(upd.attrib["index"]))
            composer.push(False)  # For now no reference point updates

            modify: List[Element] = [
                mod
                for mod in upd.findall("Modify")
                if mod.attrib["id"] == subscene_id and mod.get("duration") is not None
            ]

            composer.push(len(modify))
            for mod in modify:
                duration = int(float(mod.attrib["duration"]) * 1000)
                composer.push(duration)

                # Only one block per modify, @todo modifies with same index can be merged
                composer.push(1)

                mat: str = mod.attrib["material"]
                id = materials.index(mat)
                composer.push(int(id))

                voxel_coord_s: List[int] = [
                    int(val) for val in mod.attrib["grid_coordinate_s"].split()
                ]
                composer.push({"value": voxel_coord_s, "dimensions": voxel_dimensions})

                voxel_coord_e: List[int] = [
                    int(val) for val in mod.attrib["grid_coordinate_e"].split()
                ]
                composer.push({"value": voxel_coord_e, "dimensions": voxel_dimensions})

                config: Optional[Dict[str, Any]] = next(
                    filter(
                        lambda upd_cfg: upd_cfg["index"] == int(upd.attrib["index"]), update_config
                    ),
                    {},
                )
                self._push_renderer_instructions(
                    composer, voxel_dimensions, config, materials, objects_vox, blocks_vox
                )

    @staticmethod
    def get_payload_type() -> ScenePayloadPacketType:
        return ScenePayloadPacketType.VOXDATA

    @staticmethod
    def getArtifactSubscription() -> List[str]:
        return [
            "scene_info",
            "bs_ids",
            "object_sources",
            "voxel_config",
            "voxel_mode",
            "acoustic_tuning_params",
            "rastermap",
        ]

    @staticmethod
    def getArgumentSubscription() -> List[str]:
        return ["scene"]

    def get_bitstream(self) -> Bitstream:
        return self._bitstream


# ---------------------------------------------------------------------------------------------------
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage {} <path to EIF>".format(sys.argv[0]))
        exit(1)

    try:
        voxdata = VoxdataPayload({"scene": sys.argv[1]}, {})

        bs: Bitstream = voxdata.get_bitstream()
        bs.pad()
        bs.writeToFile(pathlib.Path(sys.argv[1]).stem + ".VoxData")

    except Exception as exc:
        print(exc)
