# This software module was originally developed by
#
# Fraunhofer IIS
#
# in the course of development of ISO/IEC 23090-4 for reference purposes
# and its performance may not have been optimized. This software module
# is an implementation of one or more tools as specified by the ISO/IEC
# 23090-4 standard. ISO/IEC gives you a royalty-free, worldwide,
# non-exclusive, copyright license to copy, distribute, and make
# derivative works of this software module or modifications thereof for
# use in implementations or products claiming conformance to ISO/IEC
# standards incorporating ISO/IEC 23090-4 reference software and which
# satisfy any specified conformance criteria of that standard.
#
# Those intending to use this software module in products are advised that
# its use may infringe existing patents and that the license in the
# previous paragraph grants no licenses under such patents. ISO/IEC and
# the developers of this software module have no liability for use of this
# software module or modifications thereof.
#
# Copyright is not released for products that do not conform to an ISO/IEC
# standard.
#
# Fraunhofer IIS retains full right to modify and use the code for its own
# purpose, assign or donate the code to a third party without consent of
# and without accounting to third parties, and to inhibit third parties
# from using the code for products that do not conform to MPEG-related
# ISO/IEC International Standards.
#
# This copyright notice must be included in all copies or derivative
# works.
#
# Copyright (c) ISO/IEC 2022.
"""Use this to for easier payload creation.

The payload class has some abstract methods, that, when overriden, can be used
by the framework to easily collect all of the data required to construct a
payload.

There's also an enum describing which ScenePayloadPacketTypes are available in
an MPEG-I Bitstream.
"""

import enum
from abc import abstractmethod
from typing import Any, Dict, List, Type

from .bitstream import Bitstream
from .inventory import Inventory


class ScenePayloadPacketType(enum.IntEnum):
    """MPEG-I ScenePayloadPacket types.

    An MPEG-I ScenePayloadPacket can have multiple different types. This enum
    is to help denote the types of payloads that can be loaded into a
    ScenePayloadPacket.
    """

    DIRECTIVITY = 6
    DIFFRACTION = 10
    EARLY_RELFECTION = 11
    PORTAL = 12
    LATE_REVERB = 14
    AUDIO_PLUS = 15
    LC_EARLY_REFLECTION = 16
    DISPERSION = 17
    VOXDATA = 18
    AIRFLOW = 19
    RASTERMAP = 20
    GRANULAR = 22

    # These types are used during development and should be removed in the future
    SCENE = 31


class Payload:
    """Super class to make Payload elements for a ScenePayloadPacket."""

    def __init__(self, args: Dict[str, Any], artifacts: Dict[str, Any]):
        """Construct payload from subscribed args and artifacts.

        :param args: The subscribed arguments
        :param artifacts: The subscribed artifacts
        """

    @staticmethod
    @abstractmethod
    def get_payload_type() -> ScenePayloadPacketType:
        """Return type of this payload."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def getArtifactSubscription() -> List[str]:
        """Return a list of all required artifacts for this payload."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def getArgumentSubscription() -> List[str]:
        """Return a list of all required arguments for this payload."""
        raise NotImplementedError()

    @abstractmethod
    def get_bitstream(self) -> Bitstream:
        """Return a bitstream that this payload will create."""
        return NotImplementedError()


def serialize_payload(
    *,
    pld_id: int,
    pld_type: Type[Payload],
    inv: Inventory,
) -> Bitstream:
    """Construct a payload with an inventory and serialize it into a bitstream.

    If the content of the payload would be zero, then an empty Bitstream is
    returned instead of a regular payload with a header and empty content
    information.lauge

    :param pld_id: The id for this payload
    :param pld_type: The abstract type of the Payload, subclass of Payload
    :param inv: An inventory already filled with artifacts
    :return: The payload in Bitstream form
    """
    try:
        args, artifacts = inv.getArgsAndArtifacts(pld_type)
    except KeyError:
        raise KeyError(
            f"Payload '{pld_type.__name__}' is requesting a non-existent artifact/argument."
        )

    concrete_payload = pld_type(args, artifacts)
    payload_as_bitstream = concrete_payload.get_bitstream()
    payload_as_bitstream.pad()

    if payload_as_bitstream.getLength() == 0:
        return Bitstream()  # Payloads with 0 info should be skipped

    payload_bstream = Bitstream()
    payload_bstream.addInt(int(pld_type.get_payload_type()), width=11)
    payload_bstream.addInt(pld_id, width=2)
    payload_size = int(payload_as_bitstream.getLength() / 8)
    payload_bstream.addInt(payload_size, width=59)
    payload_bstream.appendBitstream(payload_as_bitstream)

    return payload_bstream
